# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import argparse
import csv
import itertools
import json
import math
import sys
import urllib
import urllib2
import warnings

from oslo_config import cfg

from ceilometer.i18n import _

cfg.CONF.import_opt('host', 'ceilometer.api',
                    group='api')
cfg.CONF.import_opt('port', 'ceilometer.api',
                    group='api')

CONF = cfg.CONF

PYCADF_REQUEST_PATTERN = '%s/v2/meters/http.request%s'
PYCADF_RESPONSE_PATTERN = '%s/v2/meters/http.response%s'

KEYSTONE_PATTERN = '%s/v2/events%s'
KEYSTONE_REQUEST_FILTER = [{'q.field': 'event_type',
                            'q.op': 'eq',
                            'q.value': 'audit.http.request'}]
KEYSTONE_RESPONSE_FILTER = [{'q.field': 'event_type',
                             'q.op': 'eq',
                             'q.value': 'audit.http.response'}]

BASE_URL = 'http://%s:%s' % (CONF.api.host, CONF.api.port)


def build_filter_query_string(filters):
    '''Builds a query string given the filters.

    :param filters: List of dictionary objects
    :returns: a url compatible query string
    :raises TypeError when filters is not a list of dictionary objects
    :raise Exception when filter q.op filter is invalid
    '''
    query_filter = ''
    separator = '?'
    for filter_dict in filters:

        if 'q.op' in filter_dict and filter_dict['q.op'] not in ['eq', 'gt',
                                                                 'lt']:
            raise Exception('q.op filter field must be one of "eq", "lt" or '
                            '"gt"')
        query_filter += separator + urllib.urlencode(filter_dict)
        separator = '&'
    return query_filter


def combine_query_filters(query_filter_filename=None, extra_filter_obj=[]):
    '''Combine query filter from file and extra filter.

    :param query_filter_filename: Qualified path to the filter file
    :param extra_filter_obj: Extra query filter
    :returns: query in string
    '''
    query_filter = ''

    if query_filter_filename:
        with open(query_filter_filename, 'r') as f:
            filter_json = f.read().strip()
        filter_obj = json.loads(filter_json) + extra_filter_obj
        # TODO(dperaza) catch ValueError here and raise new InvalidFilterFile
        query_filter = build_filter_query_string(filter_obj)
    else:
        query_filter = build_filter_query_string(extra_filter_obj)
    return query_filter


def do_request(auth_token, request_url):
    '''Builds and performs the ceilometer meters request.

    :param auth_token: The authentication token
    :param request_url: The url of request
    :returns: HTTP Response object
    :raises ValueError when json in query_filter_filename is not well formatted
    '''
    headers = {'X-Auth-Token': auth_token,
               'Content-Type': 'application/json',
               'Accept': 'application/json',
               'User-Agent': 'ceilometer-audit-export'}

    req = urllib2.Request(request_url, headers=headers)
    try:
        resp = urllib2.urlopen(req)
    except urllib2.HTTPError as ex:
        if ex.code == 500:
            query_msg = ('Double check your query filters, the request url is '
                         '%s') % request_url
            raise Exception('There was an error querying the audit '
                            'records. %s' % query_msg)
        raise
    except urllib2.URLError as ex:
        if '[Errno 111]' in str(ex):
            raise Exception('There was an error connecting to audit '
                            'host %s at port %i.' % (CONF.api.host,
                                                     CONF.api.port))
        raise
    return resp


def make_meters_get_request(auth_token, query_filter_filename=None):
    '''Builds and performs the ceilometer meters request.

    :param auth_token: The authentication token
    :param query_filter_filename: Qualified path to the filter file
    :returns: HTTP Response object
    '''
    query_filter = combine_query_filters(query_filter_filename)
    request_url = PYCADF_REQUEST_PATTERN % (BASE_URL, query_filter)
    return do_request(auth_token, request_url)


def make_meters_get_response(auth_token, query_filter_filename=None):
    '''Builds and performs the ceilometer meters response.

    '''
    query_filter = combine_query_filters(query_filter_filename)
    request_url = PYCADF_RESPONSE_PATTERN % (BASE_URL, query_filter)
    return do_request(auth_token, request_url)


def make_events_get_request(auth_token, query_filter_filename=None):
    '''Builds and performs the ceilometer events request.

    '''
    query_filter = combine_query_filters(query_filter_filename,
                                         KEYSTONE_REQUEST_FILTER)
    request_url = KEYSTONE_PATTERN % (BASE_URL, query_filter)
    return do_request(auth_token, request_url)


def make_events_get_response(auth_token, query_filter_filename=None):
    '''Builds and performs the ceilometer events response.

    '''
    query_filter = combine_query_filters(query_filter_filename,
                                         KEYSTONE_RESPONSE_FILTER)
    request_url = KEYSTONE_PATTERN % (BASE_URL, query_filter)
    return do_request(auth_token, request_url)


def merge_and_limit(req_objs, res_objs, top_n=None):
    '''Merges two CADF dictionary lists while taking into account the

    top n upper bound.

    :param req_objs: The http.request list
    :param res_objs: The http.response list
    :param top_n: Upper bound integer
    :returns: Combined list of CADF dictionaries
    '''

    if top_n:
        # converting to float to get decimals after division
        slice_n = float(top_n) / 2
        # always round up to ceiling to make sure we get pairs of requests and
        # responses when we combine.
        slice_n = int(math.ceil(slice_n))
        req_objs = itertools.islice(req_objs, slice_n)
        res_objs = itertools.islice(res_objs, slice_n)

    combined_iter = itertools.izip(req_objs, res_objs)

    combined = []

    for req_obj, res_obj in combined_iter:
        combined.append(req_obj)
        combined.append(res_obj)

    return combined


def get_response_object(response):
    '''Takes a response object and returns the response dictionary

    list representing the json in the response.

    :param response: urllib2 response object
    :returns: Response dictionary
    :raises Exception when response code is not 200
    '''
    res_text = response.read()
    r_code = response.getcode()
    if r_code != 200:
        raise Exception('request returned %i, not the expected 200 '
                        'response code. Here is the response '
                        'text: %s' % (r_code, res_text))
    res_objs = json.loads(res_text)

    return res_objs


def query_and_combine(auth_token, method,
                      query_filter_filename=None, top_n=None):
    '''Combines the result of two http request.

    Makes two queries to ceilometer meters resource one for requests and
    one for responses. It then combines the two into single list.

    :param auth_token_filename: Authentication token file name
    :param query_filter_filename: Query filter file name
    :returns: Combined list of CADF dictionaries
    '''

    req = None
    res = None
    if method == 'keystone':
        req = make_events_get_request(auth_token, query_filter_filename)
        res = make_events_get_response(auth_token, query_filter_filename)
    else:
        req = make_meters_get_request(auth_token, query_filter_filename)
        res = make_meters_get_response(auth_token, query_filter_filename)

    req_objs = get_response_object(req)
    res_objs = get_response_object(res)

    combined = merge_and_limit(req_objs, res_objs, top_n=top_n)

    return combined


def csv_get_header(combined):
    '''Return the header line.

    :param combined: the list of combined requests and responses
    :returns: header line list.
    '''
    headers = []
    for cadf in combined:
        resource_metadata = cadf['resource_metadata']
        for key in resource_metadata:
            if not key.startswith('request.CADF_EVENT'):
                continue
            if key in headers:
                continue
            headers.append(key)
    headers.sort()
    return headers


def export_to_file(combined, output_filename, output_format):
    '''Exports results to a file taking into account the output format.

    :param combined: Combined list of requests and responses
    :param args: Command Line arguments
    :raises Exception when output format in args is not expected
    '''
    if output_format == 'json':
        cadf_json = []
        out_json = None

        for cadf in combined:
            resource_metadata = cadf['resource_metadata']
            record = {}
            for key, value in resource_metadata.items():
                if key.startswith('request.CADF_EVENT'):
                    record[key] = value
            cadf_json.append(record)
        out_json = json.dumps(cadf_json,
                              sort_keys=True,
                              indent=4,
                              separators=(',', ': '))
        with open(output_filename, 'w') as f:
            f.write(out_json)
    elif output_format == 'text':
        out_text = json.dumps(combined,
                              sort_keys=True,
                              indent=4,
                              separators=(',', ': '))
        with open(output_filename, 'w') as f:
            f.write(out_text)
    elif output_format == 'csv':
        writer = csv.writer(file(output_filename, 'w'))
        headers = csv_get_header(combined)
        writer.writerow(headers)

        for cadf in combined:
            resource_metadata = cadf['resource_metadata']
            record = []
            for key in headers:
                field = resource_metadata.get(key, '')
                record.append(field)
            writer.writerow(record)


def format_warning(message, category, filename, lineno, line=None):
    # ignore the file name and line number
    return "%s: %s\n" % (category.__name__, message)


def check_arguments(args):
    '''Check if the arguments are correct.

    :param args: Parsed command line arguments
    :raises Exception when format or data in args is not expected
    '''
    if args.output_format not in ['json', 'text', 'csv']:
        raise Exception(_('Output format must be either "json" or "csv"'))

    if args.output_format == 'text':
        warnings.simplefilter('always', DeprecationWarning)
        warnings.formatwarning = format_warning
        warnings.warn('"text" is deprecated and will be removed in the next '
                      'release, "json" is recommended.', DeprecationWarning)

    if args.top_n:
        try:
            top_n = int(args.top_n)
            if top_n < 1:
                raise Exception(_('Number of records must be larger than 0'))
        except ValueError:
            raise Exception(_('Number of records must be a integer'))

    if args.method not in ['pycadf', 'keystone']:
        raise Exception(_('Audit method must be either "pycadf" '
                          'or "keystone"'))


def main():
    try:
        parser = argparse.ArgumentParser()
        parser.add_argument('-a', '--auth_token',
                            metavar='FILENAME',
                            help=('File containing authentication'
                                  ' token, by default, will read from '
                                  'standard input'))
        parser.add_argument('-x', '--output_format', metavar='STRING',
                            help=('The output format, choose between "csv" and'
                                  ' "json", "json" is default'),
                            default='json')
        parser.add_argument('-n', '--top_n', metavar='INTEGER',
                            help=('Upper bound for number of records to export'
                                  ', by default, the number of records is not'
                                  ' bound'))
        parser.add_argument('-o', '--output', metavar='FILENAME',
                            help=('Output file, defaults to export.audit.csv '
                                  'if output format is csv otherwise it '
                                  'defaults to export.audit'))
        parser.add_argument('-f', '--filter', metavar='FILENAME',
                            help=('JSON file containing all possible filters, '
                                  'by default, there is no filter'))
        parser.add_argument('-m', '--method', metavar='STRING',
                            help=('Audit filter method, choose between '
                                  '"pycadf" and "keystone", "pycadf" is'
                                  ' default'),
                            default='pycadf')

        args = parser.parse_args()

        check_arguments(args)

        output_file = args.output
        if not output_file:
            output_file = 'export.audit'
            if args.output_format == 'csv':
                output_file = '.'.join((output_file, args.output_format))

        if args.auth_token:
            with open(args.auth_token, 'r') as f:
                auth_token = f.read().strip()
        else:
            # auth_token from stdin
            auth_token = raw_input("Authentication Token: ")

        combined = query_and_combine(auth_token, args.method,
                                     args.filter, top_n=args.top_n)
        export_to_file(combined, output_file, args.output_format)
    except Exception as e:
        print(e)
        sys.exit(1)
