# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Tests for Openstack Audit Export Tool."""
import os
import urllib2

import mock
from oslo_config import cfg

from ceilometer.ibm.audit_export import os_audit_export
from ceilometer.tests import base as test_base

MOD_PATH = path = os.path.dirname(__file__)

CONF = cfg.CONF


class FakeResponse(object):

    def __init__(self, filename):
        self.filename = filename

    def read(self):
        filepath = os.path.join(MOD_PATH, self.filename)
        with open(filepath, 'r') as f:
            text = f.read().strip()

        return text

    def getcode(self):
        return 200


class FakeArgs(object):

    def __init__(self, auth_token, output_format, top_n, output,
                 filter_j, method='pycadf'):
        self.auth_token = auth_token
        self.output_format = output_format
        self.top_n = top_n
        self.output = output
        self.filter = filter_j
        self.method = method


class AuditExportTestCase(test_base.BaseTestCase):
    """Test case for Audit Export."""

    filter_path = os.path.join(MOD_PATH, 'export_test.filter')
    auth_token = os.path.join(MOD_PATH, 'auth_token_test')
    method = 'pycadf'

    @staticmethod
    def _parse_sort_query_filter(query_filter):
        # NOTE(jecarey) hardcoded to three binary operators
        sorted_query_filter = []
        sorted_query_filter.append(query_filter[0])
        rest_query_filter = query_filter[1:].split('&')
        sorted_query_filter.extend(sorted(rest_query_filter[0:3]))
        sorted_query_filter.extend(sorted(rest_query_filter[3:6]))
        sorted_query_filter.extend(sorted(rest_query_filter[6:]))
        return sorted_query_filter

    def test_build_query_filter(self):
        filter_j = [{"q.field": "metadata.request.CADF_EVENT.initiator.name",
                     "q.op": "eq", "q.value": "admin"},
                    {"q.field": "metadata.request.CADF_EVENT.target.name",
                     "q.op": "eq", "q.value": "nova"},
                    {"q.field": "timestamp", "q.op": "gt",
                     "q.value": "2013-09-03T22:11:02.134000"}]
        query_text = os_audit_export.build_filter_query_string(filter_j)
        # NOTE(jecarey): Ordering is not guaranteed when processing
        # dictionaries so filter parts can be in any order.
        expected = ('?q.op=eq&q.value=admin&q.field=metadata.request.'
                    'CADF_EVENT.initiator.name&q.op=eq&q.value=nova&q.field'
                    '=metadata.request.CADF_EVENT.target.name&q.op=gt&q.value'
                    '=2013-09-03T22%3A11%3A02.134000&q.field=timestamp')
        self.assertEqual(len(expected), len(query_text))
        self.assertEqual(self._parse_sort_query_filter(expected),
                         self._parse_sort_query_filter(query_text))

    def test_build_query_filter_bad_op(self):
        filter_j = [{"q.field": "metadata.request.CADF_EVENT.initiator.name",
                     "q.op": "blah", "q.value": "admin"}]

        ex = self.assertRaises(Exception,
                               os_audit_export.build_filter_query_string,
                               filter_j)
        expected = ('q.op filter field must be one of "eq", "lt" or "gt"')
        self.assertEqual(expected, str(ex))

    def test_build_query_bad_filter(self):
        filter_j = {"que vuelta": "aqui"}
        self.assertRaises(TypeError, os_audit_export.build_filter_query_string,
                          filter_j)

    def test_csv_get_header(self):
        combined_json = [
            {"resource_metadata": {
                "event_type": "http.request",
                "host": "me",
                "request.CADF_EVENT:action": "read",
                "request.CADF_EVENT:id": "blah",
                "request.CADF_EVENT:name": "bluff"}},
            {"resource_metadata": {
                "event_type": "http.response",
                "host": "me",
                "request.CADF_EVENT:action": "write",
                "request.CADF_EVENT:id": "blue",
                "request.CADF_EVENT:name": "bloom"}}]
        csv_text = os_audit_export.csv_get_header(combined_json)
        expected = ['request.CADF_EVENT:action', 'request.CADF_EVENT:id',
                    'request.CADF_EVENT:name']
        self.assertEqual(expected, csv_text)

    def test_merge_and_bind(self):
        req_json = [
            {"request.CADF_EVENT:action": "read"},
            {"request.CADF_EVENT:action": "create"}]
        res_json = [
            {"request.CADF_EVENT:action": "write"},
            {"request.CADF_EVENT:action": "delete"}]
        combined = os_audit_export.merge_and_limit(req_json, res_json)

        expected_4 = [
            {"request.CADF_EVENT:action": "read"},
            {"request.CADF_EVENT:action": "write"},
            {"request.CADF_EVENT:action": "create"},
            {"request.CADF_EVENT:action": "delete"}]
        self.assertEqual(expected_4, combined)

        combined = os_audit_export.merge_and_limit(req_json, res_json, 2)

        expected_2 = [
            {"request.CADF_EVENT:action": "read"},
            {"request.CADF_EVENT:action": "write"}]

        self.assertEqual(expected_2, combined)

        combined = os_audit_export.merge_and_limit(req_json, res_json, 3)
        self.assertEqual(expected_4, combined)

    @staticmethod
    def _fake_urlopen(req):
        if 'http.request' in req.get_full_url():
            return FakeResponse('http.request.test')
        return FakeResponse('http.response.test')

    @staticmethod
    def _fake_bad_url_filter_open(req):
        raise urllib2.HTTPError('fakeurl', 500, 'Internal Server Error',
                                None, None)

    @staticmethod
    def _fake_bad_url_connect_open(req):
        raise urllib2.URLError('[Errno 111] Connection refused')

    def test_export_bad_url_filter(self):
        with mock.patch('urllib2.urlopen', self._fake_bad_url_filter_open):

            ex = self.assertRaises(Exception,
                                   os_audit_export.make_meters_get_request,
                                   self.auth_token,
                                   self.filter_path)

            expected = ('There was an error querying the audit records. '
                        'Double check your query filters')
            self.assertTrue(expected in str(ex))

    def test_export_bad_url_conn(self):
        with mock.patch('urllib2.urlopen', self._fake_bad_url_connect_open):
            CONF.api.host = 'ibmhost'
            CONF.api.port = 4567
            ex = self.assertRaises(Exception,
                                   os_audit_export.make_meters_get_request,
                                   self.auth_token,
                                   query_filter_filename=self.filter_path)

            expected = ('There was an error connecting to audit host '
                        'ibmhost at port 4567.')
            self.assertEqual(expected, str(ex))

    def _run_export_test(self, args, expected):
        with mock.patch('urllib2.urlopen', self._fake_urlopen):
            combined_json = os_audit_export.query_and_combine(args.auth_token,
                                                              args.filter)
            os_audit_export.export_to_file(combined_json, args.output,
                                           args.output_format)
            with open(args.output, 'r') as f:
                text = f.read().strip()
            self.assertEqual(expected, text)
            os.remove(args.output)

    def test_export_text(self):
        args = FakeArgs(self.auth_token, 'text', None, 'export_test.txt.test',
                        self.filter_path)
        filepath = os.path.join(MOD_PATH, 'export.txt.test')
        with open(filepath, 'r') as f:
            expected = f.read().strip()
        self._run_export_test(args, expected)

    def test_export_json(self):
        args = FakeArgs(self.auth_token, 'json', None, 'export_test.audit',
                        self.filter_path)
        filepath = os.path.join(MOD_PATH, 'export.audit')
        with open(filepath, 'r') as f:
            expected = f.read().strip()
        self._run_export_test(args, expected)

    def test_export_csv(self):
        args = FakeArgs(self.auth_token, 'csv', None, 'export_test.csv',
                        self.filter_path)
        filepath = os.path.join(MOD_PATH, 'export.csv')
        with open(filepath, 'r') as f:
            expected = f.read().strip()
        self._run_export_test(args, expected)

    def test_export_text_with_keystone(self):
        args = FakeArgs(self.auth_token, 'text', None, 'export_test.txt.test',
                        self.filter_path, 'keystone')
        filepath = os.path.join(MOD_PATH, 'export.txt.test')
        with open(filepath, 'r') as f:
            expected = f.read().strip()
        self._run_export_test(args, expected)

    def test_export_json_with_keystone(self):
        args = FakeArgs(self.auth_token, 'json', None, 'export_test.audit',
                        self.filter_path, 'keystone')
        filepath = os.path.join(MOD_PATH, 'export.audit')
        with open(filepath, 'r') as f:
            expected = f.read().strip()
        self._run_export_test(args, expected)

    def test_export_csv_with_keystone(self):
        args = FakeArgs(self.auth_token, 'csv', None, 'export_test.csv',
                        self.filter_path, 'keystone')
        filepath = os.path.join(MOD_PATH, 'export.csv')
        with open(filepath, 'r') as f:
            expected = f.read().strip()
        self._run_export_test(args, expected)

    def test_check_arguments(self):
        args = FakeArgs(self.auth_token, 'csv', -1, 'export_test.csv',
                        self.filter_path)
        ex = self.assertRaises(Exception, os_audit_export.check_arguments,
                               args)
        expected = ('Number of records must be larger than 0')
        self.assertEqual(expected, ex.message.translate('en_US'))

        args = FakeArgs(self.auth_token, 'csv', 'sdf', 'export_test.csv',
                        self.filter_path)
        ex = self.assertRaises(Exception, os_audit_export.check_arguments,
                               args)
        expected = ('Number of records must be a integer')
        self.assertEqual(expected, ex.message.translate('en_US'))

        args = FakeArgs(self.auth_token, 'scv', None, 'export_test.csv',
                        self.filter_path)
        ex = self.assertRaises(Exception, os_audit_export.check_arguments,
                               args)
        expected = ('Output format must be either "json" or "csv"')
        self.assertEqual(expected, ex.message.translate('en_US'))

        args = FakeArgs(self.auth_token, 'csv', None, 'export_test.csv',
                        self.filter_path, 'fake_method')
        ex = self.assertRaises(Exception, os_audit_export.check_arguments,
                               args)
        expected = ('Audit method must be either "pycadf" or "keystone"')
        self.assertEqual(expected, ex.message.translate('en_US'))
