# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from nova.api.openstack import extensions
from nova.scheduler.ibm.ego import ego_rpcapi

from oslo_log import log as logging
import oslo_messaging as messaging
from webob import exc

LOG = logging.getLogger(__name__)

# TODO(zhiwei) unify the entry_points and use this ALIAS
ALIAS = "ibm-ego-allocation"
authorize = extensions.os_compute_authorizer(ALIAS)


class AllocationController(object):

    def __init__(self):
        super(AllocationController, self).__init__()
        self.scheduler_api = ego_rpcapi.EGOSchedulerAPI()

    def update(self, req, id, body):
        """Update an ego allocation."""
        context = req.environ['nova.context']
        authorize(context, action='update')
        sub_demands = body.get('sub_demands')
        try:
            placement_decisions = (
                self.scheduler_api.ego_alloc_resize(context, id, sub_demands))
        except messaging.RemoteError as e:
            if (e.exc_type == 'EGOExpResizeAllocFailed' or
                    e.exc_type == 'EGOExpConnLost'):
                raise exc.HTTPInternalServerError(explanation=e.value)
            elif (e.exc_type == 'InvalidSubdemandInput'):
                raise exc.HTTPBadRequest(explanation=e.value)
        except Exception as ex:
            LOG.exception(ex)
            raise exc.HTTPBadRequest(explanation=ex.value)
        return {'allocation': placement_decisions}

    def create(self, req, body):
        """Create an ego allocation."""
        context = req.environ['nova.context']
        authorize(context, action='create')
        alloc_name = body.get('name')
        inter_sub_policies = body.get('inter_sub_policies')
        sub_demands = body.get('sub_demands')
        try:
            placement_decisions = (
                self.scheduler_api.ego_alloc_create(context, alloc_name,
                                                         inter_sub_policies,
                                                         sub_demands))
        except messaging.RemoteError as e:
            if (e.exc_type == 'EGOExpCreateAllocFailed' or
                    e.exc_type == 'EGOExpConnLost'):
                raise exc.HTTPInternalServerError(explanation=e.value)
            elif (e.exc_type == 'InvalidSubdemandInput'):
                raise exc.HTTPBadRequest(explanation=e.value)
        except Exception as ex:
            LOG.exception(ex)
            raise exc.HTTPBadRequest(explanation=ex.value)
        return {'allocation':
                    {'placement_decisions': placement_decisions}}

    def delete(self, req, id):
        """Delete an ego allocation."""
        context = req.environ['nova.context']
        authorize(context, action='delete')
        try:
            self.scheduler_api.ego_alloc_free(context, id)
        except Exception as ex:
            raise exc.HTTPBadRequest(explanation=ex.value)

    def getReleaseCandidates(self, req, id, body):
        context = req.environ['nova.context']
        authorize(context, action='getReleaseCandidates')
        sub_demand_name = body.get('sub_demand_name')
        num_unit = body.get('num_unit')
        try:
            placement_decisions = (
                self.scheduler_api.ego_get_release_candidates(context,
                                       id, sub_demand_name, num_unit))
        except Exception as ex:
            raise exc.HTTPBadRequest(explanation=ex.value)
        return {'allocation': placement_decisions}


class Allocation(extensions.V21APIExtensionBase):
    """Ego Allocation support."""

    name = "IBM EGO Allocation"
    alias = ALIAS
    namespace = ("http://ibm.com/compute/ext/"
                 "ego/prs/allocation/api/v3")
    version = 1

    def get_resources(self):
        """Returns a list of ego allocation objects."""
        resources = []

        res = extensions.ResourceExtension('ego/prs/allocation',
                             AllocationController(),
                             member_actions={'getReleaseCandidates': 'POST'})
        resources.append(res)

        return resources

    def get_controller_extensions(self):
        return []
