# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
from webob import exc

from nova.api.openstack import extensions
from nova.ibm.ego import ego_constants
from nova.ibm.ego import ego_gettextutils
from nova.ibm.ego import hapolicy
from nova.ibm.ego.hapolicy import utils as ha_utils
from nova import objects
from nova.scheduler.ibm.ego import ego_resreq_utils

from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_serialization import jsonutils

LOG = logging.getLogger(__name__)

_ = ego_gettextutils._


ALIAS = "ibm-prs-hypervisor-maintenance"
authorize = extensions.os_compute_authorizer(ALIAS)


class HypervisorMaintenanceController(object):
    """
    The HA Policy API controller for the OpenStack API.
    """

    def __init__(self):
        super(HypervisorMaintenanceController, self).__init__()

    def update(self, req, id, body):
        """
        :param req: GET /v2.1/{tenant_id}/ego/prs/hypervisor_maintenance/{id}

        Return maintenance related information of the hypervisor.
        """
        context = req.environ['nova.context']
        authorize(context, action='update')

        if not body:
            msg = _('No Request Body')
            raise exc.HTTPBadRequest(explanation=msg)

        if body.get('status') not in ['enable', 'disable']:
            msg = _("The status must be either 'enable' or 'disable'")
            raise exc.HTTPBadRequest(explanation=msg)

        nodes = objects.ComputeNodeList.get_all(context)
        hypervisor = None
        for node in nodes:
            if node['hypervisor_hostname'] == id:
                hypervisor = node
                break

        if not hypervisor:
            raise exc.HTTPNotFound(
                explanation=_("Hypervisor %s could not be found.") % id)

        hv_name = hypervisor['hypervisor_hostname']
        kwds = dict(for_maintenance=True)
        if (not ha_utils.hypervisor_is_up(context, hv_name, **kwds) and
                "enable" == body['status']):
            err_msg = (_("The nova-compute service for hypervisor %s is "
                         "unavailable. Please remedy the situation "
                         "and try again.") % hv_name)
            raise exc.HTTPBadRequest(explanation=err_msg)

        htype = hypervisor['hypervisor_type']
        # Skip ESXi hosts which be maintained by vCenter
        if htype == ego_constants.HYPERVISOR_TYPE_ESX_DB:
            metrics = jsonutils.loads(hypervisor['metrics'])
            metrics_map = ego_resreq_utils.list2map(metrics, 'name')
            hypervisor_state = metrics_map.get('hypervisor.state')
            if hypervisor_state and hypervisor_state.get('inMaintenanceMode'):
                msg = (_("Cannot run this command on host %s "
                         "because it is maintained by vCenter. Please exit "
                         "maintenance mode from vCenter and try again.")
                       % hypervisor['hypervisor_hostname'])
                raise exc.HTTPBadRequest(explanation=msg)
        # For docker and VMware vCenter, only supports migrate mode to be none
        # when enabling maintenance
        if htype in [ego_constants.HYPERVISOR_TYPE_VMWARE_DB,
                     ego_constants.DOCKER_HYPERVISOR_TYPE]:
            migrate_mode = body.get("migrate")
            if migrate_mode and migrate_mode != "none":
                msg = (_("The %(htype)s only supports migrate mode to be "
                         "'none' when enabling maintenance.")
                       % {'htype': htype})
                raise exc.HTTPBadRequest(explanation=msg)

        target = body.get('target_host')
        if target:
            is_valid_target = False
            for node in nodes:
                if node.host == target:
                    is_valid_target = True
                    break
            if not is_valid_target:
                reason = (_("Target host %s could not be found.")
                          % target)
                raise exc.HTTPNotFound(explanation=reason)

        rpc_api = hapolicy.API()
        try:
            if "enable" == body['status']:
                rpc_api.set_maintenance(context, id, body['status'],
                                        body.get('migrate', 'none'),
                                        body.get('target_host'))
            else:
                rpc_api.set_maintenance(context, id, body['status'])
        except messaging.MessagingTimeout:
            reason = (_("Unable to enable maintenance mode. Verify that "
                        "the Platform Resource Scheduler services are enabled "
                        "properly."))
            raise exc.HTTPRequestTimeout(explanation=reason)
        except messaging.RemoteError as e:
            if e.exc_type == 'HostMaintenanceError':
                raise exc.HTTPBadRequest(explanation=e.value)
            raise
        except Exception as e:
            LOG.exception(e)
            raise

        mt_info = {'hypervisor_hostname': hv_name,
                   'status': body['status']}
        if 'enable' == body['status']:
            mt_info.update({'migrate': body.get('migrate', 'none'),
                            'target-host': body.get('target_host')})
        return {'hypervisor_maintenance': mt_info}


class Hypervisormaintenance(extensions.V21APIExtensionBase):
    """List status of hypervisors."""

    name = "IBM HA Hypervisor Maintenance"
    alias = ALIAS
    namespace = ("http://ibm.com/compute/ext/"
                 "ego/prs/hypervisor_maintenance/api/v3")
    version = 1

    def get_resources(self):
        """Returns hypervisor maintenance."""
        resources = []

        res = extensions.ResourceExtension('ego/prs/hypervisor_maintenance',
                                           HypervisorMaintenanceController(),
                                           member_actions={})
        resources.append(res)

        return resources

    def get_controller_extensions(self):
        return []
