# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
import six
from webob import exc

from nova.api.openstack import extensions
from nova.ibm.ego import ego_constants
from nova.ibm.ego import ego_gettextutils
from nova.ibm.ego import ego_utils
from nova import objects
from nova.scheduler.ibm.ego import ego_resreq_utils

from oslo_log import log as logging

LOG = logging.getLogger(__name__)

_ = ego_gettextutils._


ALIAS = "ibm-prs-hypervisor-status"
authorize = extensions.os_compute_authorizer(ALIAS)


def get_hypervisor_ha_state(status, node):
    htype = node['hypervisor_type']
    if htype in [ego_constants.HYPERVISOR_TYPE_VMWARE_DB,
                 ego_constants.DOCKER_HYPERVISOR_TYPE]:
        state = ego_constants.NOT_SUPPORTED
    else:
        state = status.get_ha(node['hypervisor_hostname'])
    return state


def get_hypervisor_mt_state(status, node):
    return status.get_maintenance(node['hypervisor_hostname'])


class HypervisorStatusController(object):
    """
    The hypervisor status API controller for the OpenStack API
    """

    def __init__(self):
        super(HypervisorStatusController, self).__init__()

    def index(self, req):
        """
        :param req: GET /v2.1/{tenant_id}/ego/prs/hypervisor_status

        Return a list of status of all nodes.
        """
        context = req.environ['nova.context']
        authorize(context, action='index')
        nodes = objects.ComputeNodeList.get_all(context)
        hypervisor_status = []
        status = ego_utils.get_compute_status()
        for node in nodes:
            hypervisor_status.append({'id': node['id'],
                'hypervisor_hostname': node['hypervisor_hostname'],
                'hypervisor_type': node['hypervisor_type'],
                'ha_status': get_hypervisor_ha_state(status, node),
                'maintenance_status': get_hypervisor_mt_state(status, node)})
        return {'hypervisor_status': hypervisor_status}

    def show(self, req, id):
        """
        :param req: GET /v2.1/{tenant_id}/ego/prs/hypervisor_status/{id}
        :param id: hypervisor name

        Return details of a specified policy
        """
        context = req.environ['nova.context']
        authorize(context, action='show')

        if ego_resreq_utils.is_positive_integer(id):
            hyper = objects.ComputeNode.get_by_id(context, id)
        else:
            nodes = objects.ComputeNodeList.get_by_hypervisor(context, id)
            hyper = None
            for node in nodes:
                if node['hypervisor_hostname'] == id:
                    hyper = node
                    break

        if not hyper:
            raise exc.HTTPNotFound(
                explanation=_("Hypervisor '%s' could not be found.") % id)

        try:
            hv_name = hyper['hypervisor_hostname']
            status = ego_utils.get_compute_status()
            hypervisor_status = {'id': hyper['id'],
                'hypervisor_hostname': hyper['hypervisor_hostname'],
                'hypervisor_type': hyper['hypervisor_type'],
                'ha_status': get_hypervisor_ha_state(status, hyper),
                'ha_updated_at':
                   status.get_ha_timestamp(hv_name).get('updated_at'),
                'ha_created_at':
                   status.get_ha_timestamp(hv_name).get('created_at'),
                'maintenance_status': get_hypervisor_mt_state(status, hyper),
                'maintenance_updated_at':
                   status.get_maintenance_timestamp(hv_name).get('updated_at'),
                'maintenance_created_at':
                   status.get_maintenance_timestamp(hv_name).get('created_at'),
                'target-host': status.get_target_host(hv_name),
                'migrate': status.get_migrate(hv_name)}
        except Exception as e:
            raise exc.HTTPNotFound(explanation=six.text_type(e))
        return {'hypervisor_status': hypervisor_status}


class Hypervisorstatus(extensions.V21APIExtensionBase):
    """List status of hypervisors."""

    name = "IBM HA Hypervisor Status"
    alias = ALIAS
    namespace = ("http://ibm.com/compute/ext/"
                 "ego/prs/hypervisor_status/api/v3")
    version = 1

    def get_resources(self):
        """Returns a list of hyperivsor status."""
        resources = []

        res = extensions.ResourceExtension('ego/prs/hypervisor_status',
                                           HypervisorStatusController(),
                                           member_actions={})
        resources.append(res)

        return resources

    def get_controller_extensions(self):
        return []
