# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_utils import timeutils
from webob import exc

from nova.api.openstack import extensions
from nova.ibm.ego import ego_gettextutils
from nova.ibm.ego import slapolicy


LOG = logging.getLogger(__name__)

_ = ego_gettextutils._


ALIAS = "ibm-sla-policy"


authorize = extensions.os_compute_authorizer(ALIAS)


def _translate_policy_view(policy):
    return {'sla_policy': policy}


def _translate_policies_view(policies):
    return {'sla_policies':
             [_translate_policy_view(po)['sla_policy']
              for po in policies]}


class SLAPolicyController(object):
    """
    The SLA Policy API controller for the OpenStack API.
    """

    def __init__(self):
        self.slapolicy_api = slapolicy.API()
        super(SLAPolicyController, self).__init__()

    def index(self, req):
        """
        :param req: GET /v21/ego/policy/sla

        Return a list of sla policies.
        """
        context = req.environ['nova.context']
        authorize(context, 'index')
        policies = self.slapolicy_api.get_all(context)
        return _translate_policies_view(policies)

    def show(self, req, id):
        """
        :param req: GET /v21/ego/policy/sla/{id}
        :param id: sla policy id

        Return details of a specified sla policy.
        """
        context = req.environ['nova.context']
        authorize(context, 'show')
        try:
            policy = self.slapolicy_api.get(context, id)
        except messaging.RemoteError as e:
            if e.exc_type == 'SLAPolicyNotFound':
                msg = _("Cannot get SLA policy %(id)s due to "
                        "%(reason)s") % {'id': id, 'reason': e.value}
                raise exc.HTTPNotFound(explanation=msg)
        except Exception as e:
            LOG.exception(e)
            raise
        return _translate_policy_view(policy)

    def update(self, req, id, body):
        """
        :param req: PUT /v21/ego/policy/sla/{id}
        :param id: sla policy id
        :param body: {sla_policy':
                        {'description': 'example',
                         'xxx': 'xxx'
                        }
                     }

        Update the specified sla policy and return details
        of the [updated] sla policy.
        """
        context = req.environ['nova.context']
        authorize(context, 'update')
        try:
            policy = body['sla_policy']
        except KeyError:
            msg = _("SLA policy update body should be sla_policy")
            raise exc.HTTPBadRequest(explanation=msg)

        slapolicy = None
        try:
            slapolicy = self.slapolicy_api.update(context, id, policy)
        except messaging.RemoteError as e:
            if e.exc_type == 'SLAPolicyNotFound':
                raise exc.HTTPNotFound(explanation=e.value)
            elif (e.exc_type == 'InvalidSLAPolicyOperation' or
                  e.exc_type == 'InvalidSLAPolicyState' or
                  e.exc_type == 'InvalidSLAPolicyInput' or
                  e.exc_type == 'InvalidSLAPolicyAction' or
                  e.exc_type == 'InvalidSLAPolicyExtra' or
                  e.exc_type == 'SLAPolicyEGOConnLost'):
                raise exc.HTTPBadRequest(explanation=e.value)
        except Exception as e:
            LOG.exception(e)
            raise
        return _translate_policy_view(slapolicy)

    def slapolicy_hypervisor_list(self, req):
        search_opts = {}
        search_opts.update(req.GET)

        context = req.environ['nova.context']
        authorize(context, 'slapolicy_hypervisor_list')
        try:
            sla_hypervisors = self.slapolicy_api.hypervisor_list(context,
                                                               search_opts)
        except messaging.RemoteError as e:
            raise exc.HTTPBadRequest(explanation=e.value)
        except Exception as e:
            LOG.exception(e)
            raise
        return {'slapolicy_hypervisors': sla_hypervisors}

    def slapolicy_event_list(self, req):
        search_opts = {}
        search_opts.update(req.GET)

        context = req.environ['nova.context']
        authorize(context, 'slapolicy_event_list')

        int_limit = None
        isotime_marker = None
        if 'limit' in search_opts:
            try:
                int_limit = int(search_opts.pop('limit'))
            except ValueError:
                msg = _('parameter limit must be an integer')
                raise exc.HTTPBadRequest(explanation=msg)
            if int_limit < 0:
                msg = _('parameter limit must be positive')
                raise exc.HTTPBadRequest(explanation=msg)
        if 'marker' in search_opts:
            try:
                timeutils.parse_isotime(search_opts['marker'])
                isotime_marker = search_opts.pop('marker')
            except ValueError:
                msg = _('parameter marker must be a time in ISO 8601 format')
                raise exc.HTTPBadRequest(explanation=msg)
        try:
            sla_events = self.slapolicy_api.event_list(context, search_opts,
                                                  isotime_marker, int_limit)
        except messaging.RemoteError as e:
            raise exc.HTTPBadRequest(explanation=e.value)
        except Exception as e:
            LOG.exception(e)
            raise
        return {'slapolicy_events': sla_events}


class Slapolicy(extensions.V21APIExtensionBase):
    """SLA Policy."""

    name = "IBM SLA Policy"
    alias = ALIAS
    namespace = "http://ibm.com/compute/ext/ego/policy/sla/api/v21"
    version = 1

    def get_resources(self):
        """Returns a list of policy objects."""
        resources = []
        coll_actions = {
                'slapolicy_hypervisor_list': 'GET',
                'slapolicy_event_list': 'GET'
                }
        res = extensions.ResourceExtension(
                        'ego/policy/sla',
                        SLAPolicyController(),
                        collection_actions=coll_actions,
                        member_actions={})
        resources.append(res)

        return resources

    def get_controller_extensions(self):
        return []
