# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
from nova.api.openstack import extensions
from nova.ibm.ego import ego_gettextutils
from nova.ibm.ego import resource

from oslo_log import log as logging
from webob import exc

_ = ego_gettextutils._

LOG = logging.getLogger(__name__)

ALIAS = "ibm-prs-consumers"


def authorize(context, action_name):
    action = '%s:%s' % (ALIAS, action_name)
    extensions.extension_authorizer('compute', action)(context)


class ConsumerController(object):
    """
    Consumer(Tenant) Openstack API.
    """

    def __init__(self):
        self.resourceAPI = resource.API()
        super(ConsumerController, self).__init__()

    def index(self, req):
        context = req.environ['nova.context']
        authorize(context, 'index')
        return {"consumers":
                self.resourceAPI.get_consumer_list(context)}

    def show(self, req, id):
        context = req.environ['nova.context']
        authorize(context, 'show')
        try:
            consumer = self.resourceAPI.get_consumer_by_id(
                context, id)
        except Exception as ex:
            LOG.exception(ex)
            raise exc.HTTPBadRequest(explanation=ex.value)

        if not consumer:
            msg = _("The system cannot get consumer "
                    "%(id)s") % {'id': id}
            raise exc.HTTPNotFound(explanation=msg)
        else:
            return {"consumer": consumer}

    def create(self, req, body):
        """
        :param req: POST /v2/{tenant_id}/ego/prs/consumers
        :param body: the consumer tenant ID.

        Return None
        """
        if not isinstance(body, basestring):
            raise exc.exc.HTTPBadRequest()

        context = req.environ['nova.context']
        authorize(context, 'create')
        try:
            self.resourceAPI.create_consumer(context, body)
        except Exception as ex:
            LOG.exception(ex)
            raise exc.HTTPBadRequest(explanation=ex.value)
        return {}

    def delete(self, req, id):
        context = req.environ['nova.context']
        authorize(context, 'delete')
        try:
            self.resourceAPI.delete_consumer(context, id)
        except Exception as ex:
            LOG.exception(ex)
            raise exc.HTTPBadRequest(explanation=ex.value)
        return {}

    def update(self, req, id, body):
        """
        :param req: POST /v2/{tenant_id}/ego/prs/consumers
        :param body: A dict contains the properties of consumer.
        A sample is
        {'lendout_grace_period': '300',
         'reclaim_grace_period': '200',
         'rebalance': 'false'}

        Return None
        """
        if not isinstance(body, dict):
            raise exc.HTTPBadRequest()

        context = req.environ['nova.context']
        authorize(context, 'update')
        try:
            self.resourceAPI.set_consumer(context, {id: body})
        except Exception as ex:
            LOG.exception(ex)
            raise exc.HTTPBadRequest(explanation=ex.value)
        return {}


class Consumer(extensions.ExtensionDescriptor):
    """Consumer(Tenant)"""
    name = "IBM EGO Consumer"
    alias = ALIAS
    namespace = ("http://ibm.com/compute/ext/ego/prs/consumers/api/v2.0")
    updated = "2013-02-24T00:00:00+00:00"

    def get_resources(self):
        """Returns a list of policy objects."""
        resources = []

        res = extensions.ResourceExtension('ego/prs/consumers',
                                           ConsumerController(),
                                           member_actions={})
        resources.append(res)

        return resources
