# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_utils import timeutils
from webob import exc

from nova.api.openstack import extensions
from nova import compute
from nova.ibm.ego import ego_gettextutils
from nova.ibm.ego import hapolicy

LOG = logging.getLogger(__name__)

_ = ego_gettextutils._


ALIAS = "ibm-ha-policy"


def authorize(context, action_name):
    action = '%s:%s' % (ALIAS, action_name)
    extensions.extension_authorizer('compute', action)(context)


class HAPolicyController(object):
    """
    The HA Policy API controller for the OpenStack API.
    """

    def __init__(self):
        self.compute_api = compute.API()
        self.hapolicy_api = hapolicy.API()
        super(HAPolicyController, self).__init__()

    def index(self, req):
        """
        :param req: GET /v2/{tenant_id}/ego/policy/ha

        Return a list of HA policies allocated to a project.
        """
        context = req.environ['nova.context']
        authorize(context, 'index')
        policies = self.hapolicy_api.get_all(context)
        return {'policies': policies}

    def show(self, req, id):
        """
        :param req: GET /v2/{tenant_id}/ego/policy/ha/{id}
        :param id: HA policy id

        Return details of a specified policy
        """
        context = req.environ['nova.context']
        authorize(context, 'show')
        try:
            policy = self.hapolicy_api.get(context, id)
        except messaging.RemoteError as e:
            if e.exc_type == 'HAPolicyNotFound':
                msg = _("The system cannot get policy %(id)s due to "
                        "%(reason)s") % {'id': id, 'reason': e.value}
                raise exc.HTTPNotFound(explanation=msg)
        except Exception as e:
            LOG.exception(e)
            raise
        return {'policy': policy}

    def update(self, req, id, body):
        """
        :param req: GET /v2/{tenant_id}/ego/policy/ha/{id}
        :param id: policy id
        :param body: {policy':
                        {'description': 'example',
                         'xxx': 'xxx'
                        }
                     }
        Update one specified HA policy and return details
        of the specified policy
        """
        context = req.environ['nova.context']
        authorize(context, 'update')
        try:
            policy = body['policy']
        except KeyError:
            msg = _("Policy update body should be policy")
            raise exc.HTTPBadRequest(explanation=msg)

        try:
            self.hapolicy_api.update(context, id, policy)
        except messaging.RemoteError as e:
            if e.exc_type == 'HAPolicyNotFound':
                raise exc.HTTPNotFound(explanation=e.value)
            elif e.exc_type in ['InvalidHAPolicyAction',
                                'InvalidHAPolicyState',
                                'InvalidHAPolicyInput']:
                raise exc.HTTPBadRequest(explanation=e.value)
        except Exception as e:
            LOG.exception(e)
            raise
        return {'policy': policy}

    def hapolicy_hypervisor_list(self, req):
        search_opts = {}
        search_opts.update(req.GET)

        context = req.environ['nova.context']
        authorize(context, 'hapolicy_hypervisor_list')
        try:
            ha_hypervisors = self.hapolicy_api.hypervisor_list(context,
                                                               search_opts)
        except messaging.RemoteError as e:
            raise exc.HTTPBadRequest(explanation=e.value)
        except Exception as e:
            LOG.exception(e)
            raise
        return {'hapolicy_hypervisors': ha_hypervisors}

    def hapolicy_event_list(self, req):
        search_opts = {}
        search_opts.update(req.GET)

        context = req.environ['nova.context']
        authorize(context, 'hapolicy_event_list')

        int_limit = None
        isotime_marker = None
        if 'limit' in search_opts:
            try:
                int_limit = int(search_opts.pop('limit'))
            except ValueError:
                msg = _('limit param must be an integer')
                raise exc.HTTPBadRequest(explanation=msg)
            if int_limit < 0:
                msg = _('limit param must be positive')
                raise exc.HTTPBadRequest(explanation=msg)
        if 'marker' in search_opts:
            try:
                timeutils.parse_isotime(search_opts['marker'])
                isotime_marker = search_opts.pop('marker')
            except ValueError:
                msg = _('marker param must be time in ISO 8601 format')
                raise exc.HTTPBadRequest(explanation=msg)
        try:
            ha_events = self.hapolicy_api.event_list(context, search_opts,
                                                  isotime_marker, int_limit)
        except messaging.RemoteError as e:
            raise exc.HTTPBadRequest(explanation=e.value)
        except Exception as e:
            LOG.exception(e)
            raise
        return {'hapolicy_events': ha_events}


class Hapolicy(extensions.ExtensionDescriptor):
    """Resource Optimization Policy."""

    name = "IBM HA Policy"
    alias = ALIAS
    namespace = ("http://ibm.com/compute/ext/ego/policy/ha/api/v2.0")
    updated = "2013-02-24T00:00:00+00:00"

    def get_resources(self):
        """Returns a list of policy objects."""
        resources = []

        coll_actions = {
                'hapolicy_hypervisor_list': 'GET',
                'hapolicy_event_list': 'GET'
                }
        res = extensions.ResourceExtension('ego/policy/ha',
                                           HAPolicyController(),
                                           collection_actions=coll_actions,
                                           member_actions={})
        resources.append(res)

        return resources
