# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2014 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
from nova.api.openstack import extensions
from nova.ibm.ego import ego_gettextutils
from nova.ibm.ego import rtpolicy

from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_utils import timeutils
from webob import exc

LOG = logging.getLogger(__name__)

_ = ego_gettextutils._

ALIAS = "ibm_policy_placement"


def authorize(context, action_name):
    action = '%s:%s' % (ALIAS, action_name)
    extensions.extension_authorizer('compute', action)(context)


def _translate_placement_policy_view(rtpolicy):
    return {'placement_policy': rtpolicy}


def _translate_placement_policies_view(placement_policies):
    return {'placement_policies':
             [_translate_placement_policy_view(po)['placement_policy']
              for po in placement_policies]}


class RTPolicyController(object):
    """
    The Placement Policy API controller for the OpenStack API.
    """

    def __init__(self):
        self.rtpolicy_api = rtpolicy.API()
        super(RTPolicyController, self).__init__()

    def index(self, req):
        """
        :param req: GET /v2/{tenant_id}/ego/policy/placement

        Return a list of placement policies.
        """
        context = req.environ['nova.context']
        authorize(context, 'index')
        placement_policies = self.rtpolicy_api.get_all(context)
        return _translate_placement_policies_view(placement_policies)

    def show(self, req, id):
        """
        :param req: GET /v2/{tenant_id}/ego/policy/placement/{id}
        :param id: placement policy id

        Return details of a specified placement policy.
        """
        context = req.environ['nova.context']
        authorize(context, 'show')
        try:
            policy = self.rtpolicy_api.get(context, id)
        except messaging.RemoteError as e:
            if e.exc_type == 'PlacementPolicyNotFound':
                msg = _("Cannot get placement policy %(id)s due to "
                        "%(reason)s") % {'id': id, 'reason': e.value}
                raise exc.HTTPNotFound(explanation=msg)
        except Exception as e:
            LOG.exception(e)
            raise
        return _translate_placement_policy_view(policy)

    def update(self, req, id, body):
        """
        :param req: PUT /v2/{tenant_id}/ego/policy/placement/{id}
        :param id: placement policy id
        :param body: {placement_policy':
                        {'description': 'example',
                         'xxx': 'xxx'
                        }
                     }

        Update the specified placement policy and return details
        of the [updated] placement policy.
        """
        context = req.environ['nova.context']
        authorize(context, 'update')
        try:
            placement_policy = body['placement_policy']
        except KeyError:
            msg = _("Placement policy update body should be placement_policy")
            raise exc.HTTPBadRequest(explanation=msg)

        if ('id' in body['placement_policy'] and
            int(body['placement_policy']['id']) != int(id)):
            msg = _("Cannot change the placement policy's id "
                    "because it is immutable")
            raise exc.HTTPBadRequest(explanation=msg)

        rtpolicy = None
        try:
            rtpolicy = self.rtpolicy_api.update(context, id,
                                                placement_policy)
        except messaging.RemoteError as e:
            if e.exc_type == 'PlacementPolicyNotFound':
                raise exc.HTTPNotFound(explanation=e.value)
            elif (e.exc_type == 'InvalidPlacementPolicyOperation' or
                  e.exc_type == 'InvalidPlacementPolicyState' or
                  e.exc_type == 'InvalidPlacementPolicyGoal' or
                  e.exc_type == 'InvalidPlacementPolicyInput' or
                  e.exc_type == 'InvalidPlacementPolicyAction' or
                  e.exc_type == 'PlacementPolicyEGOConnLost'):
                raise exc.HTTPBadRequest(explanation=e.value)
        except Exception as e:
            LOG.exception(e)
            raise
        return _translate_placement_policy_view(rtpolicy)

    def rtpolicy_hypervisor_list(self, req):
        search_opts = {}
        search_opts.update(req.GET)

        context = req.environ['nova.context']
        authorize(context, 'rtpolicy_hypervisor_list')
        try:
            rt_hypervisors = self.rtpolicy_api.hypervisor_list(context,
                                                               search_opts)
        except messaging.RemoteError as e:
            raise exc.HTTPBadRequest(explanation=e.value)
        except Exception as e:
            LOG.exception(e)
            raise
        return {'rtpolicy_hypervisors': rt_hypervisors}

    def rtpolicy_event_list(self, req):
        search_opts = {}
        search_opts.update(req.GET)

        context = req.environ['nova.context']
        authorize(context, 'rtpolicy_event_list')

        int_limit = None
        isotime_marker = None
        if 'limit' in search_opts:
            try:
                int_limit = int(search_opts.pop('limit'))
            except ValueError:
                msg = _('limit param must be an integer')
                raise exc.HTTPBadRequest(explanation=msg)
            if int_limit < 0:
                msg = _('limit param must be positive')
                raise exc.HTTPBadRequest(explanation=msg)
        if 'marker' in search_opts:
            try:
                timeutils.parse_isotime(search_opts['marker'])
                isotime_marker = search_opts.pop('marker')
            except ValueError:
                msg = _('marker param must be time in ISO 8601 format')
                raise exc.HTTPBadRequest(explanation=msg)
        try:
            rt_events = self.rtpolicy_api.event_list(context, search_opts,
                                                  isotime_marker, int_limit)
        except messaging.RemoteError as e:
            raise exc.HTTPBadRequest(explanation=e.value)
        except Exception as e:
            LOG.exception(e)
            raise
        return {'rtpolicy_events': rt_events}


class Rtpolicy(extensions.ExtensionDescriptor):
    """Resource Optimization Policy."""

    name = "Placement"
    alias = ALIAS
    namespace = ("http://ibm.com/compute/ext/"
                 "ego/policy/placement/api/v2.0")
    updated = "2013-02-24T00:00:00+00:00"

    def get_resources(self):
        """Returns a list of policy placement objects."""
        resources = []

        coll_actions = {
                'rtpolicy_hypervisor_list': 'GET',
                'rtpolicy_event_list': 'GET'
                }
        res = extensions.ResourceExtension('ego/policy/placement',
                         RTPolicyController(),
                         collection_actions=coll_actions,
                         member_actions={})
        resources.append(res)

        return resources
