# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from nova.api.openstack import extensions
from nova.api.openstack import wsgi
from nova import exception as nova_exception
from nova.ibm.ego import ego_gettextutils
from nova.ibm.ego.wrg import api as wrg_api
from nova.ibm.ego.wrg import constants as wrg_constants
from nova.ibm.ego.wrg import exception as wrg_exception

from oslo_log import log as logging
import oslo_messaging as messaging
from webob import exc

_ = ego_gettextutils._

LOG = logging.getLogger(__name__)

ALIAS = "ibm-ego-workload-resource-group"


def _authorize_context(req, action):
    context = req.environ['nova.context']
    action = '%s:%s' % ('wrg', action)
    extensions.extension_authorizer('compute', action)(context)
    return context


class WRGController(object):

    def __init__(self):
        super(WRGController, self).__init__()
        self.wrg_api = wrg_api.API()

    def index(self, req):
        """
        :param req: GET /v2/{tenant_id}/ego/prs/wrg

        Return a list of workload resource group.
        """
        context = _authorize_context(req, 'index')
        try:
            return self.wrg_api.get_all(context)
        except messaging.RemoteError as err:
            msg = "%(err_type)s: %(err_msg)s" % {'err_type': err.exc_type,
                                                 'err_msg': err.value}
            raise exc.HTTPBadRequest(explanation=msg)
        except messaging.MessagingTimeout:
            msg = (_('Timed out waiting for nova-ibm-ego-workload-resource-'
                     'group. Is it running? Or did this service start before '
                     'nova-ibm-ego-workload-resource-group?'))
            raise exc.HTTPBadRequest(explanation=msg)
        except Exception:
            raise

    def show(self, req, id):
        """
        :param req: GET /v2/{tenant_id}/ego/prs/wrg/{wrg_id}
        :param id: workload resource group uuid

        Return details of a specified workload resource group.
        """
        context = _authorize_context(req, 'show')
        try:
            return {'wrg': self.wrg_api.get(context, id)}
        except wrg_exception.WRGNotFoundByUUID as ex:
            raise exc.HTTPBadRequest(explanation=ex.format_message())
        except messaging.RemoteError as ex:
            msg = "%(err_type)s: %(err_msg)s" % {'err_type': ex.exc_type,
                                                 'err_msg': ex.value}
            raise exc.HTTPBadRequest(explanation=msg)
        except messaging.MessagingTimeout:
            msg = (_('Timed out waiting for nova-ibm-ego-workload-resource-'
                     'group. Is it running? Or did this service start before '
                     'nova-ibm-ego-workload-resource-group?'))
            raise exc.HTTPBadRequest(explanation=msg)
        except Exception:
            raise

    def update(self, req, id, body):
        """
        :param id: workload resource group uuid
        """
        if (not (wsgi.Controller.is_valid_body(body, wrg_constants.WRG_INFO)
           or wsgi.Controller.is_valid_body(body, wrg_constants.PERF_POLICY))):
            raise exc.HTTPUnprocessableEntity()
        context = _authorize_context(req, 'update')
        try:
            return {'wrg': self.wrg_api.update(context, id, body)}
        except nova_exception.InvalidInput as ex:
            raise exc.HTTPBadRequest(explanation=ex.format_message())
        except messaging.RemoteError as ex:
            msg = "%(err_type)s: %(err_msg)s" % {'err_type': ex.exc_type,
                                                 'err_msg': ex.value}
            raise exc.HTTPBadRequest(explanation=msg)
        except messaging.MessagingTimeout:
            msg = (_('Timed out waiting for nova-ibm-ego-workload-resource-'
                     'group. Is it running? Or did this service start before '
                     'nova-ibm-ego-workload-resource-group?'))
            raise exc.HTTPBadRequest(explanation=msg)
        except Exception:
            raise

    def create(self, req, body):
        """Create a workload resource group."""
        if not wsgi.Controller.is_valid_body(body, wrg_constants.WRG_INFO):
            raise exc.HTTPUnprocessableEntity()
        context = _authorize_context(req, 'create')
        try:
            return {'wrg': self.wrg_api.create(context, body)}
        except nova_exception.InvalidInput as ex:
            raise exc.HTTPBadRequest(explanation=ex.format_message())
        except messaging.RemoteError as ex:
            msg = "%(err_type)s: %(err_msg)s" % {'err_type': ex.exc_type,
                                                 'err_msg': ex.value}
            raise exc.HTTPBadRequest(explanation=msg)
        except messaging.MessagingTimeout:
            msg = (_('Timed out waiting for nova-ibm-ego-workload-resource-'
                     'group. Is it running? Or did this service start before '
                     'nova-ibm-ego-workload-resource-group?'))
            raise exc.HTTPBadRequest(explanation=msg)
        except Exception:
            raise

    def delete(self, req, id):
        """
        :param id: workload resource group uuid
        """
        context = _authorize_context(req, 'delete')
        try:
            self.wrg_api.delete(context, id)
        except wrg_exception.WRGNotFoundByUUID as error:
            raise exc.HTTPBadRequest(explanation=error.format_message())
        except messaging.RemoteError as err:
            msg = "%(err_type)s: %(err_msg)s" % {'err_type': err.exc_type,
                                                 'err_msg': err.value}
            raise exc.HTTPBadRequest(explanation=msg)
        except messaging.MessagingTimeout:
            msg = (_('Timed out waiting for nova-ibm-ego-workload-resource-'
                     'group. Is it running? Or did this service start before '
                     'nova-ibm-ego-workload-resource-group?'))
            raise exc.HTTPBadRequest(explanation=msg)
        except Exception:
            raise


class Workloadresourcegroup(extensions.ExtensionDescriptor):
    """Workload resource group support."""

    name = "IBM EGO workload resource group"
    alias = ALIAS
    namespace = ("http://ibm.com/compute/ext/"
                 "ego/prs/wrg/api/v2.0")
    updated = "2014-04-16T06:09:00+08:00"

    def get_resources(self):
        resources = []

        res = extensions.ResourceExtension('ego/prs/wrg',
                             WRGController(),
                             member_actions={})
        resources.append(res)

        return resources
