# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import re
import six

from oslo_log import log as logging

from nova.compute.ibm.ovf import descriptor
from nova.compute.ibm.ovf import ovf_env
from nova.i18n import _


LOG = logging.getLogger(__name__)


class InvalidPropertyName(Exception):
    """This exception is thrown if a property name is not valid because
       it's not defined in the descriptor.
    """
    pass


def _clean_ovf_props_matching_compresed_key(env_properties, compressed_targets,
                                            map_properties):
    env_props_text = str(env_properties)
    for target in compressed_targets:
        # . is a regular expression reserved char, escaping for accuracy
        target = target.replace('.', '\.')
        # this will also take care of product sections with no instance
        target = target.replace('\.*', '\.*\d*')
        target = target.replace('\.*\.', '\.\d+\.')
        match_list = re.findall(target, env_props_text)

        for prop in match_list:
            if prop not in map_properties and prop in env_properties:
                env_properties.pop(prop)


def _calc_all_properties(descriptor_obj, properties, compressed_targets=None):
    # calculate properties, which includes both
    # 1) all properties from the OVF descriptor, with either the specified
    #    default value or an empty string if no default is provided.
    # 2) the overridden/new values in properties

    all_properties = {}

    descriptor_properties = descriptor_obj.calc_property_values()

    # Initialize property values from descriptor.  Properties without a
    # default value are set to an empty string.  This matches the
    # behavior of existing tools, and allows additional processing to
    # occur based on the existence of properties.
    for key, value in six.iteritems(descriptor_properties):
        if value is None:
            value = ''
        all_properties[key] = value

    if compressed_targets:
        _clean_ovf_props_matching_compresed_key(all_properties,
                                                compressed_targets,
                                                properties)

    # update properties from descriptor with new properties,
    # overrides default value if there is one.
    for key, prop_info in six.iteritems(properties):
        is_expanded = prop_info.get('expanded', False)
        if key not in descriptor_properties and not is_expanded:
            if prop_info['required']:
                raise InvalidPropertyName(_("Invalid property '%s' in OVF"
                    " properties. The key is not defined in the descriptor.") %
                    key)

            # The property is not required.
            LOG.debug("Ignoring optional OVF property '%(key)s' because it's "
                      "not defined in the OVF descriptor" % dict(key=key))
            continue

        all_properties[key] = prop_info['value']

    return all_properties


def calc_ovf_env(descriptor_obj, properties, compressed_targets=None):
    '''Calculates the ovf-env file given the descriptor and the properties to
       override.

    :param descriptor_obj: OVF Environment descriptor object
    :param properties: OVF Environment properties
    :param compressed_targets: Compressed OVF mapping targets
    '''

    all_properties = _calc_all_properties(descriptor_obj, properties,
                                    compressed_targets=compressed_targets)

    ovf_env_obj = ovf_env.OvfEnv(descriptor_obj.system_id)
    ovf_env_obj.set_properties(all_properties)

    return ovf_env_obj


def calc_ovf_env_text(descriptor_raw_text_or_obj, properties,
                      compressed_targets=None):
    '''Calculates the contents of the ovf-env.xml file given the
       raw descriptor text (XML .ovf file) and the properties to override.

    :param descriptor_raw_text_or_obj: OVF Environment xml text or an instance
                                       of descriptor.Descriptor instance.
    :param properties: OVF Environment properties
    :param compressed_targets: Compressed OVF mapping targets
    '''

    if isinstance(descriptor_raw_text_or_obj, descriptor.Descriptor):
        descriptor_obj = descriptor_raw_text_or_obj
    else:
        descriptor_raw_text = descriptor_raw_text_or_obj
        descriptor_obj = descriptor.Descriptor.parse(descriptor_raw_text)

    ovf_env_obj = calc_ovf_env(descriptor_obj, properties,
                               compressed_targets=compressed_targets)
    ovf_env_contents = ovf_env_obj.calc_raw()
    return ovf_env_contents
