# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""PRS common utility file"""

import os
import sys

from oslo_serialization import jsonutils
import six

from nova.compute import task_states
from nova import context
from nova import exception
from nova.ibm.ego import ego_gettextutils
from nova.ibm.ego.hapolicy import constants as ha_constants
from nova.ibm.ego.hapolicy import utils as ha_utils
from nova.ibm.ego import hv_status
from nova import objects
from nova.objects import base as base_obj

from oslo_config import cfg
from oslo_log import log as logging

_ = ego_gettextutils._
_LI = ego_gettextutils._LI
_LW = ego_gettextutils._LW

LOG = logging.getLogger(__name__)

VERSION = hv_status.VERSION
EMPTY_JSON = {'version': VERSION, 'maintenance': {}, 'ha': {}}

CONF = cfg.CONF
CONF.import_opt('prs_compute_status_file', 'nova.ibm.ego.ego_flags')


def get_compute_status():
    """Get compute node status"""
    if not os.path.isfile(CONF.prs_compute_status_file):
        return hv_status.HypervisorStatus(EMPTY_JSON)

    try:
        with open(CONF.prs_compute_status_file, 'r') as json_file:
            status_json = jsonutils.load(json_file)

    except Exception as ex:
        msg = (_LW('The state file (%(file_name_path)s) is corrupt '
                 'and could not be loaded (%(reason)s). '
                 'This will be remedied by effectively loading '
                 'an empty JSON structure that will simply '
                 'overwrite the corrupted file.')
               % {'file_name_path': CONF.prs_compute_status_file,
                  'reason': six.text_type(ex)})
        LOG.warning(msg)
        return hv_status.HypervisorStatus(EMPTY_JSON)

    if 'version' in status_json:
        if status_json['version'] != VERSION:
            upgrade_compute_status(status_json)
        ha_utils.check_hv_status_json_keys(status_json)
        return hv_status.HypervisorStatus(status_json)
    else:
        return hv_status.HypervisorStatus(EMPTY_JSON)


def set_compute_status(status_json):
    """
    Store the status_json to /etc/nova/prs_compute_node_status.json
    """
    if not os.path.isfile(CONF.prs_compute_status_file):
        return

    with open(CONF.prs_compute_status_file, 'w') as json_file:
        jsonutils.dump(status_json, json_file)


def upgrade_compute_status(status_json):
    """
    Upgrade /etc/nova/prs_compute_node_status.json as needed.
    This only supports upgrading from version n-1 to n.
    """

    if status_json['version'] != '2.0':
        # if the old version is not 2.0, we just clear the hypervisor status
        status_json = EMPTY_JSON
        set_compute_status(status_json)
        return

    ha_status = status_json.get('ha', {})
    mt_status = status_json.get('maintenance', {})
    # In version 2.0, moving_inst is of the format {inst_uuid: target_host},
    # but version 3 adds some additional details about the movement type;
    # e.g., whether it was a migration or a rebuild operation.
    cxt = context.get_admin_context()
    for status in [ha_status, mt_status]:
        for hv_name in status:
            inst_hosts = status[hv_name].get('moving_insts')
            if not inst_hosts:
                continue
            new_inst_hosts = {}
            for (inst_uuid, target_host) in inst_hosts.items():
                if status == ha_status:
                    op = ha_constants.HA_ACTION_REBUILD
                else:
                    op = ha_constants.HA_ACTION_LIVE_MIGRATION
                try:
                    inst = objects.Instance.get_by_uuid(cxt, inst_uuid)
                    if inst.task_state == task_states.MIGRATING:
                        op = ha_constants.HA_ACTION_LIVE_MIGRATION
                    elif inst.task_state in [task_states.RESIZE_MIGRATING,
                                             task_states.RESIZE_MIGRATED,
                                             task_states.RESIZE_FINISH]:
                        op = ha_constants.HA_ACTION_COLD_MIGRATION
                except Exception:
                    # If the instance was deleted and we hit an exception,
                    # just ignore and let the HA service handle the error.
                    pass
                new_inst_hosts.update({inst_uuid: {'target-host': target_host,
                                                   'action': op}})
            status[hv_name]['moving_insts'] = new_inst_hosts

    status_json['version'] = VERSION
    status_json['ha'] = ha_status
    status_json['maintenance'] = mt_status
    set_compute_status(status_json)
    LOG.info(_LI('The compute status file has been upgraded from version '
               '%(file_version)s to version %(cur_version)s.') %
              {'file_version': '2.0',
               'cur_version': VERSION})


def safe_remove(data_struct, value):
    """
    Make sure value not in data_struct after calling this function
    """
    if isinstance(data_struct, list) or isinstance(data_struct, set):
        while value in data_struct:
            data_struct.remove(value)

    if isinstance(data_struct, dict):
        if value in data_struct:
            data_struct.pop(value)


def encrypt_password(str_need_encrypted):
    """
    Encrypt a string in EGO user password encryption way.
    It is required to install EGO and run
    'source /opt/ibm/ego/profile.platform' before call this method.
    """
    ego_lib_path = os.environ.get('EGO_LIBDIR')
    if not ego_lib_path:
        raise exception.EncryptionFailure(reason=(_("Could not find "
                        "EGO_LIBDIR in the environment; please ensure EGO is "
                        "installed and then retry the request.")))
    sys.path.insert(0, ego_lib_path)
    ego = __import__('ego')
    ego_handler = ego.ego()
    return ego_handler.encryptPassword(str_need_encrypted)


def convert_spec_obj_to_legacy_dict(context, spec_obj, inst_props):
    """
    Convert RequestSpec Object to request_spec and filter_properties dict.
    """
    # In OSCE mitaka release, nova conductor parse RequestSpec object
    # instead of request_spec/filter_properties to 'select_destination'.
    # (related OSCE patch: https://review.openstack.org/#/c/203078/)
    # To align with this, PRS should convert RequestSpec Object to legacy
    # request_spec/filter_properties dict.
    request_spec = spec_obj.to_legacy_request_spec_dict()
    filter_properties = spec_obj.to_legacy_filter_properties_dict()
    instance_uuid = request_spec.get('instance_properties', {}).get('uuid')
    if not inst_props:
        # Force the loading of the metadata attributes so that they are put
        # into the dictionary via the `obj_to_primitive` call.
        instance = objects.Instance.get_by_uuid(
            context, instance_uuid,
            expected_attrs=['metadata', 'system_metadata'])
        inst_props = base_obj.obj_to_primitive(instance)
    request_spec['instance_properties'].update(inst_props)
    # Get operation_type from scheduler_hints and append it to request_spec
    # dict, it will help HA service to trigger instance action later
    scheduler_hints = filter_properties.get('scheduler_hints', {})
    request_spec['operation_type'] = scheduler_hints.get('operation_type')
    filter_properties['request_spec'] = request_spec
    filter_properties['instance_type'] = request_spec['instance_type']
    filter_properties['context'] = context
    return (request_spec, filter_properties)
