# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Base engine for HA and maintenance."""

import eventlet
import six

from nova import compute
from nova.compute import instance_actions
from nova.compute import rpcapi as compute_rpcapi
from nova.compute import task_states
from nova.compute import vm_states
from nova import context
from nova import exception
from nova.ibm.ego import ego_constants
from nova.ibm.ego import ego_gettextutils
from nova.ibm.ego import ego_utils
from nova.ibm.ego.hapolicy import constants as ha_constants
from nova.ibm.ego.hapolicy import utils as ha_utils
from nova import objects
from nova.objects import instance as instance_obj
from nova import rpc
from nova.scheduler.ibm.ego import ego_resreq_utils
from nova.scheduler import rpcapi as scheduler_rpcapi
from nova.scheduler import utils as scheduler_utils

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import timeutils

_ = ego_gettextutils._
_LI = ego_gettextutils._LI
_LW = ego_gettextutils._LW

CONF = cfg.CONF
CONF.import_opt('policy_per_host_aggregate', 'nova.ibm.ego.ego_flags')
CONF.import_opt('prs_ha_timeout_seconds', 'nova.ibm.ego.ego_flags')
CONF.import_opt('prs_mt_migration_preferred_order', 'nova.ibm.ego.ego_flags')

LOG = logging.getLogger(__name__)


class BasePolicyEngine(object):
    def __init__(self, ego_api, inst_driver, prs_driver):
        self.hv_map = {}
        self.compute_api = compute.API()
        self.compute_rpcapi = compute_rpcapi.ComputeAPI()
        self.scheduler_rpcapi = scheduler_rpcapi.SchedulerAPI()
        self.ego_api = ego_api
        self.ego_attr_name = None
        self.notifier = rpc.get_notifier('hapolicy')
        self.event_types = {}
        self.inst_driver = inst_driver
        self.reset()
        self._configure_hv_status()
        # Retrieve the context in a robust manner; self.ctxt is only used
        # during initialization anyway and then the context is retrieved
        # at each periodic task cycle (i.e., to keep it refreshed).
        try:
            self.ctxt = prs_driver.get_context()
        except Exception as e:
            LOG.warning(_LW("Encountered exception fetching driver extension "
                            "context; falling back to standard admin context; "
                            "exception message is: %(message)s")
                        % {'message': six.text_type(e)})
            self.ctxt = context.get_admin_context()

    def reset(self):
        self.violated_hvs = set()
        self.in_action_hvs = set()
        self.idle_hvs = set()
        self.error_insts = {}

    def _configure_hv_status(self):
        self.start_status = None
        self.migrating_status = None
        self.rebuilding_status = None
        self.end_status = None
        self.error_status = None

    def select_instances(self, context, hv_name, hv_is_up):
        raise NotImplementedError()

    def update_hv_state(self, context, hv_name, state, extra_msg=None):
        """
        Update status of the hypervisor $hv_name to $state.
        """
        extra_msg = extra_msg or {}
        if not state:
            LOG.debug("Cannot set hypervisor %(hv)s state to none!"
                      % {'hv': hv_name})
            return
        old_state = self.get_status(hv_name)
        if old_state == state:
            LOG.debug("Ignoring request to set hypervisor %(hv)s state to "
                      "%(state)s since it's already set to that value."
                      % {'hv': hv_name, 'state': state})
            return
        try:
            self._store_event(context, hv_name,
                              self.event_types[state]['start'],
                              extra_msg)
            if self.ego_attr_name:
                self.ego_api.ego_update_resource_attr(hv_name,
                                     {self.ego_attr_name: state})
            self.set_status(hv_name, state)
            self._store_event(context, hv_name,
                              self.event_types[state]['end'],
                              extra_msg)
            LOG.info(_LI("Successfully updated the state of hypervisor %(hv)s "
                       "from %(old_state)s to %(state)s.")
                     % {'hv': hv_name,
                        'old_state': old_state,
                        'state': state})
        except Exception as ex:
            LOG.warning(_LW("Failed to update the state of hypervisor %(hv)s "
                       "due to: %(reason)s")
                     % {'reason': six.text_type(ex),
                        'hv': hv_name})
            raise

    def get_status(self, hv_name):
        raise NotImplementedError()

    def set_status(self, hv_name, state):
        raise NotImplementedError()

    def get_moving_insts(self, hv_name, action=None):
        raise NotImplementedError()

    def set_moving_insts(self, hv_name, inst_host):
        raise NotImplementedError()

    def set_move_start_time(self, hv_name, inst_uuid, start=None):
        raise NotImplementedError

    def get_move_start_time(self, hv_name, inst_uuid):
        raise NotImplementedError

    def run(self, *args, **kwargs):
        """
        The engine main function that acts as an entry point for callers.

        Subclasses of BasePolicyEngine should implement this method with any
        necessary arguments.

        The purpose of this method in the BasePolicyEngine is to document the
        intended entry point for users of engines.
        """
        raise NotImplementedError()

    def _sync_up_hv_status(self, context):
        # mark all hypervisor to end status when all instances have moved
        # to the destinations; try to mark hypervisors to 'error' status if
        # moving instances failed and make the instance's vm_state ERROR.
        for hv_name in self.idle_hvs:
            if hv_name not in self.error_insts:
                self.error_insts[hv_name] = {}
            error_insts = self.error_insts[hv_name]
            inst_hosts = self.get_moving_insts(hv_name) or {}

            temp_inst_hosts = inst_hosts.copy()
            timeout = CONF.prs_ha_timeout_seconds
            all_moved = True
            for inst_uuid in inst_hosts:
                # 1) remove the instance info in the engine if the instance
                #    has been deleted in the database.
                host = inst_hosts[inst_uuid]['target-host']
                try:
                    new_inst = objects.Instance.get_by_uuid(context, inst_uuid)
                except Exception:
                    temp_inst_hosts.pop(inst_uuid)
                    error_insts.pop(inst_uuid, None)
                    continue

                if new_inst.host == self.hv_map[hv_name]['host']:
                    # if the VM is still on the original host, the related
                    # hypervisor cannot enter end state ('on' for MT and
                    # 'evacuated' for HA) yet.
                    all_moved = False
                    # 2) if the engine made the instance ERROR for automated
                    #    migration/rebuild, try to update the hypervisor's
                    #    state to 'error'.
                    if new_inst.vm_state == vm_states.ERROR:
                        if (error_insts.get(inst_uuid, 0) <
                                    ha_constants.MAX_RETRY_TIMES):
                            msg = (_LW("Automated migration/rebuild of "
                                 "instance %(hostname)s (%(inst_uuid)s) to "
                                 "%(dest)s from %(source)s failed. For "
                                 "detailed information about the failure, "
                                 "see the nova-compute log file (usually in "
                                 "/var/log/nova/nova-compute-%(dest)s.log) "
                                 "for host %(dest)s.")
                               % {'inst_uuid': inst_uuid,
                                  'hostname': new_inst.hostname,
                                  'dest': host,
                                  'source': hv_name})
                            LOG.warning(msg)
                            error_insts[inst_uuid] = (
                                ha_constants.MAX_RETRY_TIMES)
                            self.__handle_error(context, hv_name,
                                            {'uuid': inst_uuid,
                                             'hostname': new_inst.hostname},
                                            msg)

                    # 3) if the instance is still not moved away, try to update
                    #    the hypervisor to error state if the move operation
                    #    times out.
                    if (error_insts.get(inst_uuid, 0) <
                                    ha_constants.MAX_RETRY_TIMES):
                        # mark timed out hypervisors to error state
                        now = timeutils.utcnow()
                        start = self.get_move_start_time(hv_name, inst_uuid)
                        start_time = timeutils.parse_strtime(start)
                        diff = timeutils.delta_seconds(start_time, now)
                        if diff > timeout:
                            msg = (_LW("Automated migration/rebuild of "
                                     "instance %(name)s (%(inst)s) on "
                                     "hypervisor %(hv_name)s failed with the "
                                     "following error: operation has been "
                                     "running for %(diff)s seconds, which "
                                     "exceeds the timeout of %(timeout)s "
                                     "seconds.")
                                     % {'name': new_inst['hostname'],
                                        'inst': inst_uuid, 'hv_name': hv_name,
                                        'diff': diff, 'timeout': timeout})
                            LOG.warning(msg)
                            error_insts[inst_uuid] = (
                                ha_constants.MAX_RETRY_TIMES)
                            self.__handle_error(
                                context, hv_name,
                                {'uuid': inst_uuid,
                                 'hostname': new_inst.hostname}, msg)
                            ego_utils.set_compute_status(
                                          self.status.get_json())
                else:
                    # 4) remove the instance moving info from the engine if it
                    #    has been moved away (automatically or manually).
                    temp_inst_hosts.pop(inst_uuid)
                    error_insts.pop(inst_uuid, None)

            self.set_moving_insts(hv_name, temp_inst_hosts)
            if (isinstance(self, MaintenanceEngine) and
                    self.get_status(hv_name) == self.error_status):
                continue
            # 5) check if the failed instances for HA have been moved away
            #    manually; if yes, update the failed instances' record so that
            #    the hypervisor's HA state can change to 'evacuated' from
            #    'error'
            elif isinstance(self, HAEngine):
                failed_insts = self.status.get_ha_failed_instances(hv_name)
                temp_failed_insts = failed_insts[:]
                for failed_inst in failed_insts:
                    try:
                        inst = objects.Instance.get_by_uuid(context,
                                                            failed_inst,
                                                            expected_attrs=[])
                        if (inst.host !=
                                self.hv_map[hv_name]['host']):
                            # the instance has been moved away manually.
                            temp_failed_insts.remove(failed_inst)
                            error_insts.pop(failed_inst, None)
                    except exception.InstanceNotFound:
                        # Instance has been removed from OpenStack.
                        temp_failed_insts.remove(failed_inst)
                        error_insts.pop(failed_inst, None)
                self.status.set_ha_failed_instances(hv_name, temp_failed_insts)
                if temp_failed_insts:
                    continue
            # 6) update the hypervisor's state to end status if all the
            #    instances on it have been moved away.
            if all_moved:
                self.update_hv_state(context, hv_name, self.end_status)
        ego_utils.set_compute_status(self.status.get_json())

    def _move_instances(self, context, hv_name):
        """Migrates or rebuilds instances."""
        raise NotImplementedError()

    def __update_moving_insts(self, hv_name, inst, host, action):
        moving_insts = self.get_moving_insts(hv_name)
        moving_inst = {inst['uuid']: {'target-host': host,
                                      'action': action}}
        if moving_insts:
            moving_insts.update(moving_inst)
        else:
            moving_insts = moving_inst
        self.set_moving_insts(hv_name, moving_insts)
        ego_utils.set_compute_status(self.status.get_json())

    def _migrate_instances(self, context, hv_name):
        """Live or cold migrates instances based on their state."""
        inst_tuples = self.select_instances(context, hv_name, True)
        if not inst_tuples:
            if hv_name not in self.idle_hvs:
                # No more instances to move and we're not idle yet, so this
                # means that we should consider our work done!
                self._action_finished(context, hv_name)
            return

        if hv_name not in self.in_action_hvs:
            self.in_action_hvs.add(hv_name)
            ego_utils.safe_remove(self.idle_hvs, hv_name)
        is_finished = True
        filters = {'task_state': [task_states.MIGRATING,
                                  task_states.RESIZE_MIGRATING],
                   'host': self.hv_map[hv_name]['host']}

        for inst, action in inst_tuples:
            if hv_name in self.idle_hvs:
                return

            if inst['task_state']:
                is_finished = False
                continue

            migration_count = len(objects.InstanceList.get_by_filters(
                                  context, filters))
            if migration_count >= self.__get_parallel_migration(
                    self.hv_map[hv_name]):
                is_finished = False
                # hit max; no need to look at any more instances this time
                break
            try:
                if inst['vm_state'] == vm_states.ACTIVE:
                    if self.inst_driver.is_pre_operation_done(
                        context, inst, ha_constants.HA_ACTION_LIVE_MIGRATION):
                        success = self.__live_migration(context, hv_name, inst)
                    else:
                        self.inst_driver.do_pre_operation(
                            context, inst,
                            ha_constants.HA_ACTION_LIVE_MIGRATION)
                        if self.inst_driver.is_pre_operation_done(
                            context, inst,
                            ha_constants.HA_ACTION_LIVE_MIGRATION):
                            success = self.__live_migration(context, hv_name,
                                                            inst)
                    if not success:
                        is_finished = False
                elif inst['vm_state'] == vm_states.STOPPED:
                    if self.inst_driver.is_pre_operation_done(
                        context, inst, ha_constants.HA_ACTION_COLD_MIGRATION):
                        success = self.__cold_migration(context, hv_name, inst)
                    else:
                        self.inst_driver.do_pre_operation(
                            context, inst,
                            ha_constants.HA_ACTION_COLD_MIGRATION)
                        if self.inst_driver.is_pre_operation_done(
                            context, inst,
                            ha_constants.HA_ACTION_COLD_MIGRATION):
                            success = self.__cold_migration(context, hv_name,
                                                            inst)
                    if not success:
                        is_finished = False
                else:
                    is_finished = False
                    LOG.warning(_LW("The system cannot migrate instance "
                               "%(name)s (%(instance_uuid)s) in the %(state)s "
                               "state; the HA service only migrates active or "
                               "stopped instances.")
                             % {'name': inst['hostname'],
                                'instance_uuid': inst['uuid'],
                                'state': inst['vm_state']})
            except Exception:
                is_finished = False

        if is_finished:
            self._action_finished(context, hv_name)

    def _rebuild_instances(self, context, hv_name):
        """Rebuilds instances on a failed hypervisor."""
        inst_tuples = self.select_instances(context, hv_name, False)
        if not inst_tuples:
            if hv_name not in self.idle_hvs:
                # No more instances to move and we're not idle yet, so this
                # means that we should consider our work done!
                self._action_finished(context, hv_name)
            return

        compute_node = self.hv_map[hv_name]
        htype = compute_node['hypervisor_type']

        # if the hypervisor cannot support 'rebuild' operation, mark
        # the hypervisor to 'error' state.
        if htype in ego_constants.UNSUPPORTED_HYPERVISORS:
            if self.get_status(hv_name) != self.error_status:
                self._action_finished(context, hv_name)
                err_msg = (_("The operation failed on the host (%(host)s) "
                             "because its hypervisor (%(hv)s) does not "
                             "support the 'rebuild' operation.")
                           % {'hv': hv_name,
                              'host': compute_node['host']})
                self.update_hv_state(context, hv_name, self.error_status,
                                     {'error_msg': err_msg})
            return

        if hv_name not in self.in_action_hvs:
            self.in_action_hvs.add(hv_name)
            ego_utils.safe_remove(self.idle_hvs, hv_name)

        is_finished = True

        rebuild_states = [task_states.REBUILDING,
                          task_states.REBUILD_SPAWNING,
                          task_states.REBUILD_BLOCK_DEVICE_MAPPING]
        for inst, action in inst_tuples:
            if hv_name in self.idle_hvs:
                return

            # if the VM is being rebuilt, then we shouldn't try to call rebuild
            # again; otherwise, let's try our best to rebuild it regardless of
            # its task state since the host is dead.
            if inst['task_state'] in rebuild_states:
                is_finished = False
                continue

            # NOTE(jwcroppe): This active/stopped/error state check is imposed
            # by OpenStack, so we check it here to avoid problems calling the
            # evacuate API since we know it would otherwise fail.
            if inst['vm_state'] not in [vm_states.ACTIVE, vm_states.STOPPED,
                                        vm_states.ERROR]:
                is_finished = False
                LOG.warning(_LW("The system cannot rebuild instance %(name)s "
                           "(%(instance_uuid)s) in the %(state)s state "
                           "because OpenStack restricts the valid initial "
                           "states to active, stopped and error.")
                         % {'name': inst['hostname'],
                            'instance_uuid': inst['uuid'],
                            'state': inst['vm_state']})
                continue

            num_insts_rebuilding = self._get_num_insts_currently_moving(
                context, hv_name, rebuild_states,
                action=ha_constants.HA_ACTION_REBUILD)

            LOG.debug('Number of instances currently rebuilding on '
                      'hypervisor %(hv)s: %(num)s.' %
                      {'hv': hv_name,
                       'num': num_insts_rebuilding})

            max_parallel_rebuilds = self.__get_parallel_rebuild(compute_node)
            if num_insts_rebuilding >= max_parallel_rebuilds:
                LOG.debug('Hypervisor %(hv)s has %(current_num)s instances '
                          'currently rebuilding, but it can only handle '
                          '%(max_num)s in parallel. Waiting to rebuild the '
                          'rest of the instances until the next cycle.' %
                          {'hv': hv_name,
                           'current_num': num_insts_rebuilding,
                           'max_num': max_parallel_rebuilds})
                is_finished = False
                # hit max; no need to look at any more instances this time
                break

            try:
                if self.inst_driver.is_pre_operation_done(
                    context, inst, ha_constants.HA_ACTION_REBUILD):
                    success = self.__rebuild(context, hv_name, inst)
                else:
                    self.inst_driver.do_pre_operation(
                        context, inst, ha_constants.HA_ACTION_REBUILD)
                    if self.inst_driver.is_pre_operation_done(
                        context, inst, ha_constants.HA_ACTION_REBUILD):
                        success = self.__rebuild(context, hv_name, inst)
                    else:
                        success = False
                if not success:
                    is_finished = False
            except Exception:
                is_finished = False

        if is_finished:
            self._action_finished(context, hv_name)

    def _get_num_insts_currently_moving(self, context, hv_name, states,
                                        action=None):
        """
        Returns the number of instances deemed to have task states in the list
        of states passed as a parameter for a specific hypervisor.
        """
        # Only consider the moving instances initiated by this engine
        # since the instance.host attribute only changes to the target host
        # in the database once the movement call actually finishes (which is
        # why we can't just count the instances based on the VM state).
        # Because of the above reasoning, we work backwards from what
        # instances we know had movements started and deduce which instances
        # are still moving.
        instances_no_longer_moving = []
        instances_moving = self.get_moving_insts(hv_name, action=action) or {}
        for inst_uuid in instances_moving:
            try:
                temp_inst = objects.Instance.get_by_uuid(context,
                                                         inst_uuid)
            except Exception:
                # Instance got deleted, so don't consider it as moving
                instances_no_longer_moving.append(inst_uuid)
                continue

            # If the instance's task state is no longer one of the specified
            # task states, then consider the instance as no longer moving.
            currently_moving = temp_inst.task_state in states
            if not currently_moving:
                instances_no_longer_moving.append(inst_uuid)

        return len([inst_uuid for inst_uuid in instances_moving
                    if inst_uuid not in instances_no_longer_moving])

    def __revert_allocation(self, context, inst_uuid, source_node):
        # scheduler service will revert instance's allocation to
        # source_node when accepting this notification.
        payload = {'instance_id': inst_uuid, 'node': source_node}
        self.notifier.info(context,
                    ego_constants.EVENT_TYPE_PRS_HA_MT_INST_AUTO_MOVE_ABORT,
                    payload)

    def __live_migration(self, context, hv_name, inst):
        """Live migrates an instance."""
        img = ha_utils.get_image_from_inst(inst)
        request_spec = scheduler_utils.build_request_spec(context, img, [inst])
        svc_host = self.hv_map[hv_name]['host']
        filter_properties = {'ignore_hosts': [svc_host]}
        target_hosts = self._get_target_hosts(
            context, ha_constants.HA_ACTION_LIVE_MIGRATION, hv_name)

        if target_hosts:
            filter_properties['force_hosts'] = target_hosts
        filter_properties[ego_constants.HA_MAINTENANCE_STATUS] = (
                                                    self.status.get_json())
        # set filter working scope
        scope = ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION
        filter_properties['scheduler_hints'] = {'operation_type': scope}

        # Construct RequestSpec Object
        spec_obj = objects.RequestSpec.from_primitives(
            context, request_spec, filter_properties)

        try:
            scheduler_utils.setup_instance_group(context, request_spec,
                                                 filter_properties)
            destinations = self.scheduler_rpcapi.select_destinations(
                context, spec_obj)
            host = destinations[0]['host']
            inst = objects.Instance.get_by_uuid(
                    context, inst['uuid'],
                    expected_attrs=instance_obj.INSTANCE_DEFAULT_FIELDS)
            self.compute_api.live_migrate(context.elevated(), inst, False,
                                          False, host)
            if self.get_status(hv_name) != self.migrating_status:
                self.update_hv_state(context, hv_name, self.migrating_status)

            action_info = {'vm_name': inst['hostname'],
                           'vm_uuid': inst['uuid'],
                           'action_name': 'live-migration',
                           'target_host': host}
            self._store_action_info(context, hv_name, action_info)
            self.__update_moving_insts(hv_name, inst, host,
                                       ha_constants.HA_ACTION_LIVE_MIGRATION)
            self.set_move_start_time(hv_name, inst['uuid'])
            LOG.info(_LI("Live migration of instance %(name)s (%(instance)s) "
                       "to host %(host)s has started.")
                     % {'name': inst['hostname'],
                        'instance': inst['uuid'],
                        'host': host})
            return True
        except Exception as ex:
            if isinstance(ex, exception.NoValidHost):
                # give a more friendly message for NoValidHost cases
                msg = (_LW("Automated live migration of instance %(name)s "
                         "(%(instance)s) failed because no valid destination "
                         "host(s) could be found. Make sure you have enough "
                         "resources in your relocation domain and try again.")
                       % {'name': inst['hostname'],
                          'instance': inst['uuid']})
            else:
                msg = (_LW("Automated live migration of instance %(name)s "
                         "(%(instance)s) failed with the following error: "
                         "%(reason)s")
                       % {'name': inst['hostname'],
                          'instance': inst['uuid'],
                          'reason': six.text_type(ex)})
                self.__revert_allocation(context, inst['uuid'], hv_name)
            LOG.warning(msg)
            self.__handle_error(context, hv_name,
                                {'uuid': inst['uuid'],
                                 'hostname': inst['hostname']}, msg)
            raise

    def __cold_migration(self, context, hv_name, inst):
        """Cold migrates an instance."""
        img = ha_utils.get_image_from_inst(inst)

        flavor = objects.Flavor.get_by_id(context, inst['instance_type_id'])
        request_spec = scheduler_utils.build_request_spec(context, img, [inst],
                                                          instance_type=flavor)
        svc_host = self.hv_map[hv_name]['host']
        filter_properties = {'ignore_hosts': [svc_host]}
        target_hosts = self._get_target_hosts(
            context, ha_constants.HA_ACTION_COLD_MIGRATION, hv_name)

        if target_hosts:
            filter_properties['force_hosts'] = target_hosts
        filter_properties[ego_constants.HA_MAINTENANCE_STATUS] = (
                                                    self.status.get_json())
        # set filer working scope
        scope = ego_constants.FILTER_WORKING_SCOPE_COLD_MIGRATION
        filter_properties['scheduler_hints'] = {'operation_type': scope}

        # Construct RequestSpec Object
        spec_obj = objects.RequestSpec.from_primitives(
            context, request_spec, filter_properties)

        try:
            scheduler_utils.setup_instance_group(context, request_spec,
                                                 filter_properties)
            destinations = self.scheduler_rpcapi.select_destinations(
                context, spec_obj)

            scheduler_utils.populate_filter_properties(
                filter_properties, destinations[0])
            # context is not serializable
            filter_properties.pop('context', None)

            # TODO(timello): originally, instance_type in request_spec
            # on compute.api.resize does not have 'extra_specs', so we
            # remove it for now to keep tests backward compatibility.
            request_spec['instance_type'].pop('extra_specs')

            inst = objects.Instance.get_by_uuid(
                    context, inst['uuid'],
                    expected_attrs=instance_obj.INSTANCE_DEFAULT_FIELDS)

            inst.task_state = task_states.RESIZE_PREP
            inst.progress = 0
            inst.save(expected_task_state=[None])

            context.request_id = (
                ha_constants.PREFIX_PRS_HA + context.request_id)
            objects.InstanceAction.action_start(context,
                                                inst['uuid'],
                                                instance_actions.RESIZE,
                                                want_result=False)

            (host, node) = (destinations[0]['host'],
                            destinations[0]['nodename'])
            self.compute_rpcapi.prep_resize(
                            context.elevated(),
                            img,
                            inst,
                            flavor,
                            host,
                            None,
                            request_spec=request_spec,
                            filter_properties=filter_properties,
                            node=node)
            if self.get_status(hv_name) != self.migrating_status:
                self.update_hv_state(context, hv_name, self.migrating_status)

            action_info = {'vm_name': inst['hostname'],
                           'vm_uuid': inst['uuid'],
                           'action_name': 'cold-migration',
                           'target_host': host}
            self._store_action_info(context, hv_name, action_info)
            self.__update_moving_insts(hv_name, inst, host,
                                       ha_constants.HA_ACTION_COLD_MIGRATION)
            self.set_move_start_time(hv_name, inst['uuid'])
            LOG.info(_LI("Cold migration of instance %(name)s (%(instance)s) "
                       "to host %(host)s has started.")
                     % {'name': inst['hostname'],
                        'instance': inst['uuid'],
                        'host': host})
            return True
        except Exception as ex:
            if isinstance(ex, exception.NoValidHost):
                # give a more friendly message for NoValidHost cases
                msg = (_LW("Automated cold migration of instance %(name)s "
                         "(%(instance)s) failed because no valid destination "
                         "host(s) could be found. Make sure you have enough "
                         "resources in your relocation domain and try again.")
                       % {'name': inst['hostname'],
                          'instance': inst['uuid']})
            else:
                msg = (_LW("Automated cold migration of instance %(name)s "
                         "(%(instance)s) failed with the following error: "
                         "%(reason)s")
                       % {'name': inst['hostname'],
                          'instance': inst['uuid'],
                          'reason': six.text_type(ex)})
                self.__revert_allocation(context, inst['uuid'], hv_name)
            LOG.warning(msg)
            self.__handle_error(context, hv_name,
                                {'uuid': inst['uuid'],
                                 'hostname': inst['hostname']}, msg)
            raise

    def __rebuild(self, context, hv_name, inst):
        """Rebuilds an instance."""
        img = ha_utils.get_image_from_inst(inst)
        request_spec = scheduler_utils.build_request_spec(context, img, [inst])
        # set filer working scope
        svc_host = self.hv_map[hv_name]['host']
        filter_properties = {'ignore_hosts': [svc_host]}
        target_hosts = self._get_target_hosts(
            context, ha_constants.HA_ACTION_REBUILD, hv_name)

        if target_hosts:
            filter_properties['force_hosts'] = target_hosts
        filter_properties[ego_constants.HA_MAINTENANCE_STATUS] = (
                                                self.status.get_json())
        scope = ego_constants.FILTER_WORKING_SCOPE_EVACUATE
        filter_properties['scheduler_hints'] = {'operation_type': scope}

        # Construct RequestSpec Object
        spec_obj = objects.RequestSpec.from_primitives(
            context, request_spec, filter_properties)

        try:
            scheduler_utils.setup_instance_group(context, request_spec,
                                                 filter_properties)
            destinations = self.scheduler_rpcapi.select_destinations(
                context, spec_obj)
            host = destinations[0]['host']
            target_hv_name = destinations[0]['nodename']

            # check if the parallel rebuild instances number on the destination
            # host have exceeded the max number; if yes, skip this rebuild
            rebuild_states = [task_states.REBUILDING,
                              task_states.REBUILD_SPAWNING,
                              task_states.REBUILD_BLOCK_DEVICE_MAPPING]
            filters = {'task_state': rebuild_states, 'host': host}
            compute_node = self.hv_map.get(target_hv_name)
            if not compute_node:
                compute_node = objects.ComputeNode.get_by_host_and_nodename(
                                              context, host, target_hv_name)
            rebuild_num = len(objects.InstanceList.get_by_filters(
                              context, filters))
            if rebuild_num >= self.__get_parallel_dest_rebuild(compute_node):
                # skip this rebuild and recover EGO allocation
                self.__revert_allocation(context, inst['uuid'], hv_name)
                LOG.info(_LI("The HA engine selected destination host "
                             "%(host)s to rebuild %(vm_name)s (%(uuid)s), "
                             "but the host is currently serving as the "
                             "destination host for %(rebuild_num)s rebuild "
                             "operation(s) and can't handle any more; the HA "
                             "engine will wait until the next cycle to try "
                             "and find a valid host for the instance.")
                         % {'host': host, 'vm_name': inst['hostname'],
                            'uuid': inst['uuid'], 'rebuild_num': rebuild_num})
                return False

            inst = objects.Instance.get_by_uuid(
                    context, inst['uuid'],
                    expected_attrs=instance_obj.INSTANCE_DEFAULT_FIELDS)
            self.compute_api.evacuate(context.elevated(), inst, host, True)
            if self.get_status(hv_name) != self.rebuilding_status:
                self.update_hv_state(context, hv_name, self.rebuilding_status)

            action_info = {'vm_name': inst['hostname'],
                           'vm_uuid': inst['uuid'],
                           'action_name': 'evacuate',
                           'target_host': host}
            self._store_action_info(context, hv_name, action_info)
            self.__update_moving_insts(hv_name, inst, host,
                                       ha_constants.HA_ACTION_REBUILD)
            self.set_move_start_time(hv_name, inst['uuid'])
            LOG.info(_LI("Successfully initiated the rebuild processing for "
                         "instance %(name)s (%(instance)s) on host %(host)s.")
                     % {'name': inst['hostname'],
                        'instance': inst['uuid'],
                        'host': host})
            # Give instance.host a moment to update via RPC for accurate counts
            eventlet.sleep(1)
            return True
        except Exception as ex:
            if isinstance(ex, exception.NoValidHost):
                # give a more friendly message for NoValidHost cases
                msg = (_LW("Automated rebuild of instance %(name)s "
                         "(%(instance)s) failed because no valid destination "
                         "host(s) could be found. Make sure you have enough "
                         "resources in your relocation domain and try again.")
                       % {'name': inst['hostname'],
                          'instance': inst['uuid']})
            else:
                msg = (_LW("Automated rebuild of instance %(name)s "
                         "(%(instance)s) failed with the following error: "
                         "%(reason)s")
                       % {'name': inst['hostname'],
                          'instance': inst['uuid'],
                          'reason': six.text_type(ex)})
                self.__revert_allocation(context, inst['uuid'], hv_name)
            LOG.warning(msg)
            self.__handle_error(context, hv_name,
                                {'uuid': inst['uuid'],
                                 'hostname': inst['hostname']}, msg)
            raise

    def _store_event(self, context, hv_name, event_type, extra_msg=None):
        raise NotImplementedError()

    def _store_action_info(self, context, hv_name, action_info):
        return

    def _action_finished(self, context, hv_name):
        ego_utils.safe_remove(self.in_action_hvs, hv_name)
        self.idle_hvs.add(hv_name)

    def __handle_error(self, context, hv_name, inst_dict, err_msg):
        """Common error handler for the engine to handle retries, etc."""
        inst_uuid = inst_dict['uuid']
        if hv_name not in self.error_insts:
            self.error_insts[hv_name] = {}
        self.error_insts[hv_name].update({inst_uuid:
                 self.error_insts.get(hv_name, {}).get(inst_uuid, 0) + 1})
        if (self.error_insts[hv_name][inst_uuid] >=
                ha_constants.MAX_RETRY_TIMES):
            payload = {'hypervisor_hostname': hv_name,
                       'hypervisor_id': self.hv_map[hv_name]['id'],
                       'instance_uuid': inst_uuid,
                       'error_msg': err_msg}
            self.notifier.info(context,
                               ha_constants.ETYPE_MOVING_INSTANCE_FAILED,
                               payload)
            LOG.warning(_LW('Failed to automatically move instance '
                           '%(hostname)s (%(inst_uuid)s) away from '
                           '%(hv_name)s; please move it manually.')
                        % {'hv_name': hv_name, 'inst_uuid': inst_uuid,
                           'hostname': inst_dict['hostname']})
            # 1. for the HA case, only when all the instances which the HA
            #    HA selected have failed, the hypervisor's HA status will
            #    become 'error'.
            # 2. for the maintenance case, once one instance fails to move
            #    away, the hypervisor's maintenance status will become
            #    'error'.
            if isinstance(self, HAEngine):
                failed_insts = self.status.get_ha_failed_instances(hv_name)
                if inst_uuid not in failed_insts:
                    failed_insts.append(inst_uuid)
                    self.status.set_ha_failed_instances(hv_name, failed_insts)
            elif isinstance(self, MaintenanceEngine):
                if self.get_status(hv_name) != self.error_status:
                    self.update_hv_state(context, hv_name, self.error_status,
                                         {'error_msg': err_msg})
                    self._action_finished(context, hv_name)

    def __get_parallel_migration(self, compute_node):
        stats = compute_node.stats or {}
        return int(stats.get(ha_constants.PARALLEL_MIGRATE,
                             ha_constants.DEFAULT_PARALLEL_MIGRATE))

    def __get_parallel_rebuild(self, compute_node):
        stats = compute_node.stats or {}
        return int(stats.get(ha_constants.PARALLEL_REBUILD,
                             ha_constants.DEFAULT_PARALLEL_REBUILD))

    def __get_parallel_dest_rebuild(self, compute_node):
        stats = compute_node.stats or {}
        return int(stats.get(ha_constants.PARALLEL_DEST_REBUILD,
                             ha_constants.DEFAULT_PARALLEL_DEST_REBUILD))

    def _should_stop_move(self, context, hv_name):
        raise NotImplementedError()

    def _get_target_hosts(self, context, ha_action, hv_name):
        """When moving instances, specify target hosts."""
        return None

    def move_inst_fail(self, context, hv_name, payload, is_live_migration):
        """
        This method is called when compute service sends notification about
        live-migration roll back or cold migration/rebuild error.

        :param context: OpenStack nova client context
        :param hv_name: the hypervisor name which the instance was moved from
        :param payload: the notification message payload
        :param is_live_migration: True if the instance was live migrated
                                  False if the instance was cold
                                        migrated/rebuild
        :return: None
        """
        inst_id = payload.get('instance_id')
        if (self.error_insts.get(hv_name, {}).get(inst_id, 0) >=
                ha_constants.MAX_RETRY_TIMES):
            # for the rebuild/cold migrate fail case, if sync up logic has
            # done this work, skip it here.
            return
        hostname = payload.get('hostname')
        inst_hosts = self.get_moving_insts(hv_name)
        msg = (_LW("Automated movement of instance %(hostname)s "
                 "(%(inst_uuid)s) to %(dest)s from %(source)s failed. "
                 "For detailed information about the failure, see the "
                 "nova-compute log file (usually in "
                 "/var/log/nova/nova-compute-%(dest)s.log) for host %(dest)s.")
               % {'inst_uuid': inst_id,
                  'hostname': hostname,
                  'dest': inst_hosts.get(inst_id, {}).get('target-host'),
                  'source': hv_name})
        LOG.warning(msg)
        if is_live_migration:
            # for live-migration roll back case, the instance moving lifecycle
            # is over and we should add it into moving_inst so we try to move
            # it in the next cycle.
            ego_utils.safe_remove(self.idle_hvs, hv_name)
            ego_utils.safe_remove(inst_hosts, inst_id)
            self.in_action_hvs.add(hv_name)
            self.set_moving_insts(hv_name, inst_hosts)
        else:
            if hv_name not in self.error_insts:
                self.error_insts[hv_name] = {}
            # if the VM has been automatically cold migrated or evacuated
            # away, even if something wrong occurred on the target host
            # to make the VM enter error state, just log a warnning message
            # since the VM is not on original hypervisor anymore.
            is_away = False
            try:
                inst = objects.Instance.get_by_uuid(context, inst_id,
                                                    expected_attrs=[])
                if inst.node != hv_name:
                    is_away = True
            except exception.InstanceNotFound:
                # This only happens when the notification about the VM
                # transitioning to the error state lags too much (e.g.,
                # the VM has since been deleted).
                is_away = True
            if is_away:
                ego_utils.safe_remove(inst_hosts, inst_id)
                self.set_moving_insts(hv_name, inst_hosts)
                self.error_insts[hv_name].pop(inst_id, None)
                LOG.warning(_LW("The instance %(hostname)s (%(inst_uuid)s) "
                                "is no longer on hypervisor %(hv_name)s, "
                                "so the engine will not do any action on "
                                "it anymore and take the automated movement "
                                "of the instance as successful one.")
                            % {'inst_uuid': inst_id,
                               'hostname': hostname,
                               'hv_name': hv_name})
                return
            # for rebuild/cold migrate error, the VM has errored out;
            # we should mark the hypervisor as 'error' immediately.
            self.error_insts[hv_name].update({inst_id:
                              ha_constants.MAX_RETRY_TIMES})

        self.__handle_error(context, hv_name,
                 {'uuid': inst_id, 'hostname': hostname}, msg)

    def get_hv_moving_inst_map(self):
        """
        This method returns a dict which the hypervisor name is key and the
        instance uuid list which this engine moved is value.
        """
        hv_moving_inst_map = {}
        for hv_name in self.hv_map:
            moving_insts = self.get_moving_insts(hv_name)
            if moving_insts:
                inst_uuid_list = moving_insts.keys()
                hv_moving_inst_map.update({hv_name: inst_uuid_list})
        return hv_moving_inst_map


class HAEngine(BasePolicyEngine):
    """
    This class orchestrates all HA operations. The engine's run method takes
    in a list of hypervisors that it is expected to manage as well as a
    list of hypervisors that have been deemed violated. It then marks those
    hypervisors as violated and begins moving instances off of them.
    """

    def __init__(self, ego_api=None, status=None, hv_driver=None,
                 inst_driver=None, prs_driver=None, ha_managed_hvs=None):
        super(HAEngine, self).__init__(ego_api, inst_driver, prs_driver)
        self.status = status
        self.hv_driver = hv_driver
        self.inst_driver = inst_driver
        self.ego_attr_name = 'ha_status'
        self.event_types = ha_constants.HA_EVENT_TYPES

        # Format: {request_id1: {action_name: xx, vm_name: xx, vm_uuid: xx,
        #                        target_host: xx},
        #          request_id12: {}, ...}
        self.hapolicy_action = {}

        self.__load_hv_state_from_file(ha_managed_hvs)

    def _configure_hv_status(self):
        super(HAEngine, self)._configure_hv_status()
        self.start_status = ha_constants.STATE_HA_STARTED
        self.migrating_status = ha_constants.STATE_HA_MIGRATING
        self.rebuilding_status = ha_constants.STATE_HA_REBUILDING
        self.end_status = ha_constants.STATE_HA_EVACUATED
        self.error_status = ha_constants.STATE_HA_ERROR

    def _get_target_hosts(self, context, ha_action, hv_name):
        """When moving instances, specify target hosts."""
        target_host = self.hv_driver.get_target_service_host(
            context, ha_action, hv_name)
        return [target_host] if target_host else None

    def run(self, context,
            all_nodes,
            managed_hypervisors=None,
            violated_hypervisors=None):
        """
        The engine main function.

        :param all_nodes: the detailed info dict of all the nodes in the
                          system
        :param managed_hypervisors: names of all hypervisors that the engine
                                    should manage
        :param violated_hypervisors: names of all hypervisors that have been
                                     deemed violated by the caller

        :returns: None
        """
        self.set_managed_hypervisors(context, all_nodes, managed_hypervisors)

        # If there are hypervisors that were previously violated but are now
        # no longer considered violated, then recover those hypervisors
        recovered_hypervisors = list(
            set(self.violated_hvs) - set(violated_hypervisors))
        if recovered_hypervisors:
            self._recover_hypervisors(context, recovered_hypervisors)

        self._mark_violated_hvs(context, violated_hypervisors)

        for hv_name in self.violated_hvs:
            try:
                self._move_instances(context, hv_name)
            except Exception as e:
                LOG.warning(_LW('Failed to move instances away from '
                                'hypervisor %(hv_name)r due to: '
                                '%(reason)s')
                            % {'reason': six.text_type(e),
                               'hv_name': hv_name})
                continue

        self._sync_up_hv_status(context)

    def set_managed_hypervisors(self, context, all_nodes,
                                managed_hypervisors):
        """
        Set the hypervisors that the engine manages

        :param context: User context
        :param managed_hypervisors: Hypervisors that the engine manages
        """
        # convert the list of managed hypervisor names to list of dictionaries
        # with hypervisor name / node info key / value dicts
        temp_hv_map = {}
        for hv_name in managed_hypervisors:
            for node in all_nodes:
                if node['hypervisor_hostname'] == hv_name:
                    temp_hv_map[hv_name] = node
                    break

        # for every hypervisor that we were previously managing but no longer
        # are, set its state to OK and remove it from all of the VM lists
        old_hvs = list(set(self.hv_map.keys()) - set(temp_hv_map.keys()))
        self._recover_hypervisors(context, old_hvs)

        self.hv_map = temp_hv_map

    def select_instances(self, context, hv_name, hv_is_up):
        svc_host = self.hv_map[hv_name]['host']
        inst_tuples = self.inst_driver.select_instances(context, svc_host,
                                                        hv_is_up)

        # For the HA engine, PRS will not try to migrate instances that
        # previously failed to move.
        failed_inst_list = self.status.get_ha_failed_instances(hv_name)
        for inst_uuid in failed_inst_list:
            for inst, action in inst_tuples:
                if inst_uuid == inst.uuid:
                    inst_tuples.remove((inst, action))
                    break

        if not inst_tuples:
            # if the engine has no other instances that need to migrate except
            # previously failed ones, the hypervisor's HA status should
            # change to the 'error' state.
            if (self.get_status(hv_name) != self.error_status and
                failed_inst_list):
                self.update_hv_state(context, hv_name, self.error_status)
                self._action_finished(context, hv_name)
        return inst_tuples

    def get_status(self, hv_name):
        return self.status.get_ha(hv_name)

    def set_status(self, hv_name, state):
        self.status.set_ha(hv_name, state)
        ego_utils.set_compute_status(self.status.get_json())

    def get_moving_insts(self, hv_name, action=None):
        return self.status.get_ha_moving_insts(hv_name, action=action)

    def set_moving_insts(self, hv_name, inst_host):
        self.status.set_ha_moving_insts(hv_name, inst_host)
        ego_utils.set_compute_status(self.status.get_json())

    def get_move_start_time(self, hv_name, inst_uuid):
        return self.status.get_ha_moving_inst_start(hv_name, inst_uuid)

    def set_move_start_time(self, hv_name, inst_uuid, start=None):
        if not start:
            start = timeutils.strtime()
        self.status.set_ha_moving_inst_start(hv_name, inst_uuid, start)
        ego_utils.set_compute_status(self.status.get_json())

    def move_inst_fail(self, context, hv_name, payload, is_live_migration):
        super(HAEngine, self).move_inst_fail(context, hv_name, payload,
                                             is_live_migration)
        action_info = self.hapolicy_action.pop(context.request_id, None)
        if action_info:
            action_info['action_result'] = 'failed'
            self._store_event(context, hv_name,
                              ha_constants.ETYPE_HA_ACTION_END, action_info)

    def __load_hv_state_from_file(self, ha_managed_hvs):
        if not ha_managed_hvs:
            return
        # for the first time this engine runs, the HA status of managed
        # hypervisors should be persisted from COMPUTE_STATS_FILE
        for hv_name in ha_managed_hvs:
            ha_state = self.status.get_ha(hv_name)
            if ha_state != ha_constants.STATE_OK:
                self.violated_hvs.add(hv_name)
                # if the hypervisor's HA status is not 'ok', the related
                # EGO RPI resource should marked as 'UNVAIL'
                ha_utils.mark_node_unavailable(self.ego_api, hv_name)
            if ha_state in [self.migrating_status,
                            self.rebuilding_status]:
                self.in_action_hvs.add(hv_name)
            elif ha_state in [self.end_status,
                              self.error_status]:
                self.idle_hvs.add(hv_name)

    def _move_instances(self, context, hv_name):
        if self._should_stop_move(context, hv_name):
            self._action_finished(context, hv_name)
            return
        if ha_utils.hypervisor_is_up(context, hv_name):
            self._migrate_instances(context, hv_name)
        else:
            self._rebuild_instances(context, hv_name)

    def _should_stop_move(self, context, hv_name):
        # If the hypervisor is entering maintenance or has maintenance
        # enabled, the HA engine shouldn't operate on the hypervisor
        # since it's already being operated on or expected to be down.
        # If the MT engine is in error state (i.e., something went wrong
        # during maintenance) *and* the hypervisor is down, then
        # the HA engine should be able to rebuild the VMs as necessary.
        # In all other cases, the HA engine should be able to operate
        # on the hypervisor.
        mt_status = self.status.get_maintenance(hv_name)
        if mt_status in [ha_constants.STATE_MT_ENTERING,
                         ha_constants.STATE_MT_ON]:
            return True
        elif mt_status == ha_constants.STATE_MT_ERROR:
            return ha_utils.hypervisor_is_up(context, hv_name)
        else:
            return False

    def _mark_violated_hvs(self, context, violated_hypervisors):
        self.violated_hvs = violated_hypervisors
        for hv_name in violated_hypervisors:
            if self.get_status(hv_name) == ha_constants.STATE_OK:
                # if the hypervisor's HA status is 'OK', mark it 'started'
                self.update_hv_state(context, hv_name,
                                     self.start_status)
                ha_utils.mark_node_unavailable(self.ego_api, hv_name)

    def _recover_hypervisors(self, context, recovered_hypervisors):
        for hv_name in recovered_hypervisors:
            self.update_hv_state(context,
                                 hv_name,
                                 ha_constants.STATE_OK)
            ego_utils.safe_remove(self.violated_hvs, hv_name)
            ego_utils.safe_remove(self.in_action_hvs, hv_name)
            ego_utils.safe_remove(self.idle_hvs, hv_name)
            ego_utils.safe_remove(self.error_insts, hv_name)
        LOG.debug("Recovered hypervisors: %s" % recovered_hypervisors)

    def _store_event(self, context, hv_name, event_type, extra_msg=None):
        extra_msg = extra_msg or {}
        payload = {}
        if event_type != ha_constants.ETYPE_HA_ACTION_END:
            # for prs.ha.ha_action.end event type, the related payload
            # should have been stored in self.hapolicy_action cache; we
            # can get it from the cache.

            payload = {'hypervisor_hostname': hv_name,
                       'hypervisor_id': self.hv_map[hv_name]['id'],
                       'status': self.get_status(hv_name),
                       'policy_type': 'hapolicy'}
        payload.update(extra_msg)
        if event_type == ha_constants.ETYPE_HA_ACTION_START:
            self.hapolicy_action[context.request_id] = payload
        self.notifier.info(context, event_type, payload)

    def _store_action_info(self, context, hv_name, action_info):
        self._store_event(context, hv_name, ha_constants.ETYPE_HA_ACTION_START,
                          action_info)

    def move_inst_success(self, context):
        action_info = self.hapolicy_action.pop(context.request_id, None)
        if action_info:
            action_info['action_result'] = 'success'
            self._store_event(context, action_info.get('hypervisor_hostname'),
                              ha_constants.ETYPE_HA_ACTION_END,
                              action_info)


class MaintenanceEngine(BasePolicyEngine):
    def __init__(self, hv_name, migrate, status, ego_api, target_host,
                 inst_driver=None, prs_driver=None):
        super(MaintenanceEngine, self).__init__(ego_api, inst_driver,
                                                prs_driver)
        self.hv_name = hv_name
        self.migrate = migrate
        self.status = status
        self.target_host = target_host
        self.ego_attr_name = 'maintenance_status'
        self.event_types = ha_constants.MT_EVENT_TYPES

        self.update_managed_hvs(self.ctxt,
                                objects.ComputeNodeList.get_all(self.ctxt))

    def run(self, context, all_nodes, update_hvs_state=True,
            do_action=True):
        """
        The engine main function.

        :param all_nodes: the detailed info dict of all the nodes in the
                          system
        :param update_hvs_state: update the hypervisors' state if true
        :param do_action: do moving instance action based on hypervisors'
                          state if true

        :returns: None
        """
        if update_hvs_state:
            self.update_managed_hvs(context, all_nodes)
            self.update_hvs_state(context)
        if do_action:
            for hv_name in self.violated_hvs:
                # If the maintenance status is either ON or ERROR, that means
                # all instances have been moved away from this host or an
                # error occured. Either way, there is nothing left to do, so
                # just return.
                mt_status = self.status.get_maintenance(hv_name)
                if mt_status in [ha_constants.STATE_MT_ON,
                                 ha_constants.STATE_MT_ERROR]:
                    LOG.debug('Maintenance status of %(hv)s is %(status)s, '
                              'so there is no need to continue trying to '
                              'move instances.' % {'hv': hv_name,
                                                   'status': mt_status})
                    if hv_name not in self.idle_hvs:
                        # This can occur if the ha-service is restarted. Go
                        # ahead and add the host back to self.idle_hvs since
                        # there's nothing to do with it.
                        self._action_finished(context, hv_name)
                    return
                self._move_instances(context, hv_name)
        self._sync_up_hv_status(context)

    def _configure_hv_status(self):
        super(MaintenanceEngine, self)._configure_hv_status()
        self.start_status = ha_constants.STATE_MT_ENTERING
        self.end_status = ha_constants.STATE_MT_ON
        self.error_status = ha_constants.STATE_MT_ERROR

    def update_managed_hvs(self, context, all_nodes):
        for node in all_nodes:
            if node['hypervisor_hostname'] == self.hv_name:
                self.hv_map[node['hypervisor_hostname']] = node
                break

    def update_hvs_state(self, context):
        """
        Put the hypervisors that violated the policy into the violated list
        and change their status to self.start_status so as to trigger the next
        set of actions on them.
        """
        self.violated_hvs.add(self.hv_name)
        if self.get_status(self.hv_name) == ha_constants.STATE_OK:
            self.status.set_migrate(self.hv_name, self.migrate)
            self.status.set_target_host(self.hv_name,
                                        self.target_host)
            self.update_hv_state(context, self.hv_name,
                                 self.start_status)

    def select_instances(self, context, hv_name, hv_is_up):
        def _inst_comp(inst, htype):
            if (CONF.prs_mt_migration_preferred_order ==
                ego_constants.MOST_MEM):
                return -inst['memory_mb']
            elif (CONF.prs_mt_migration_preferred_order ==
                  ego_constants.LEAST_MEM):
                return inst['memory_mb']
            elif (CONF.prs_mt_migration_preferred_order ==
                  ego_constants.MOST_VCPUS):
                vcpus = ego_resreq_utils.get_instance_vcpus(inst, htype)
                return -vcpus
            elif (CONF.prs_mt_migration_preferred_order ==
                  ego_constants.LEAST_VCPUS):
                vcpus = ego_resreq_utils.get_instance_vcpus(inst, htype)
                return vcpus
            else:
                LOG.warning(_LW("Unrecognized value for nova.conf option "
                           "'prs_mt_migration_preferred_order' (%(val)s); "
                           "assuming effective value of %(default)s "
                           "instead.")
                         % {'val': CONF.prs_mt_migration_preferred_order,
                            'default': ego_constants.MOST_MEM})
                return -inst['memory_mb']

        if hv_name in self.idle_hvs:
            return []

        inst_tuples = []
        states = None
        if self.migrate == 'all':
            states = [vm_states.ACTIVE, vm_states.RESIZED, vm_states.BUILDING,
                      vm_states.PAUSED, vm_states.RESCUED, vm_states.ERROR,
                      vm_states.SUSPENDED, vm_states.STOPPED,
                      vm_states.SOFT_DELETED, vm_states.SHELVED]
        elif self.migrate == 'active-only':
            states = [vm_states.ACTIVE, vm_states.RESIZED, vm_states.BUILDING,
                      vm_states.PAUSED, vm_states.RESCUED, vm_states.ERROR]
        if states:
            svc_host = self.hv_map[self.hv_name]['host']
            insts = objects.InstanceList.get_by_host(context, svc_host,
                                expected_attrs=['system_metadata'])
            inst_tuples = [(inst, None) for inst in insts
                           if inst['vm_state'] in states]

        compute_node = self.hv_map.get(hv_name)
        htype = compute_node['hypervisor_type']
        inst_tuples.sort(key=lambda inst_tuple: _inst_comp(inst_tuple[0],
                                                            htype))
        return inst_tuples

    def get_status(self, hv_name):
        return self.status.get_maintenance(hv_name)

    def set_status(self, hv_name, state):
        self.status.set_maintenance(hv_name, state)
        ego_utils.set_compute_status(self.status.get_json())

    def get_moving_insts(self, hv_name, action=None):
        return self.status.get_maintenance_moving_insts(hv_name)

    def set_moving_insts(self, hv_name, inst_host):
        self.status.set_maintenance_moving_insts(hv_name, inst_host)
        ego_utils.set_compute_status(self.status.get_json())

    def get_move_start_time(self, hv_name, inst_uuid):
        return self.status.get_mt_moving_inst_start(hv_name, inst_uuid)

    def set_move_start_time(self, hv_name, inst_uuid, start=None):
        if not start:
            start = timeutils.strtime()
        self.status.set_mt_moving_inst_start(hv_name, inst_uuid, start)
        ego_utils.set_compute_status(self.status.get_json())

    def _move_instances(self, context, hv_name):
        if self._should_stop_move(context, hv_name):
            self._action_finished(context, hv_name)
            return

        kwds = dict(for_maintenance=True)
        if ha_utils.hypervisor_is_up(context, hv_name, **kwds):
            self._migrate_instances(context, hv_name)
        else:
            err_msg = _("The hypervisor failed to enter "
                        "maintenance mode because it became inactive "
                        "during the operation; please remedy the situation "
                        "and re-enable maintenance mode.")
            self.update_hv_state(context, self.hv_name,
                                 self.error_status,
                                 {'error_msg': err_msg})
            self._action_finished(context, hv_name)

    def _should_stop_move(self, context, hv_name):
        return hv_name in self.idle_hvs

    def _get_target_hosts(self, context, ha_action, hv_name):
        return [self.target_host] if self.target_host else None

    def _store_event(self, context, hv_name, event_type, extra_msg=None):
        extra_msg = extra_msg or {}
        payload = {'hypervisor_hostname': hv_name,
                   'hypervisor_id': self.hv_map[hv_name]['id']}
        payload.update(extra_msg)
        self.notifier.info(context, event_type, payload)
