# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from collections import defaultdict
import six
import uuid

from nova.ibm.ego import ego_constants
from nova.ibm.ego import ego_gettextutils
from nova import objects
from nova.scheduler.ibm.ego import ego_resreq_utils

from oslo_config import cfg
from oslo_log import log as logging

_ = ego_gettextutils._
_LI = ego_gettextutils._LI
_LW = ego_gettextutils._LW

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class HAMonitor(object):

    """
    Class that monitors hypervisors of interest and determines if said
    hypervisors are "violated" based on the run_interval and the stabilization
    count.
    """

    def __init__(self, run_interval, stabilization, aggregate_id=None,
                 hypervisor_driver=None):
        self.run_interval = int(run_interval)
        self.current_interval = 0
        self.stabilization = int(stabilization)
        self.aggregate_id = None
        if aggregate_id is not None:
            self.aggregate_id = str(aggregate_id)
        self.hypervisor_driver = hypervisor_driver
        self.monitored_hypervisors = {}
        self.violation_counts = defaultdict(int)
        self.id = uuid.uuid4()

        LOG.info(_LI('Monitor with ID %(id)s initialized with run_interval of '
                     '%(run_interval)d, stabilization of %(stabilization)d, '
                     'and aggregate ID of %(agg_id)s.') %
                 {'id': self.id,
                  'run_interval': self.run_interval,
                  'stabilization': self.stabilization,
                  'agg_id': self.aggregate_id})

    def set_stabilization(self, stabilization):
        self.stabilization = int(stabilization)
        LOG.info(_LI('Monitor with ID %(id)s stabilization is now '
                     '%(stabilization)d.') %
                     {'id': self.id, 'stabilization': self.stabilization})

    def set_run_interval(self, run_interval):
        self.run_interval = int(run_interval)
        LOG.info(_LI('Monitor with ID %(id)s run_interval is now '
                     '%(run_interval)d.') %
                 {'id': self.id, 'run_interval': self.run_interval})

    def reset(self):
        self.current_interval = 0
        self.monitored_hypervisors.clear()
        self.violation_counts.clear()
        LOG.info(_LI('Monitor with ID %(id)s was reset.') % {'id': self.id})

    def get_monitored_hypervisors(self):
        """
        Returns a list of names of all of the hypervisors that this monitor
        finds relevant (e.g. those which belong to the aggregate, or all).
        """
        return self.monitored_hypervisors.keys()

    def get_violated_hypervisors(self):
        """
        Returns the names of all violated hypervisors, where a violated
        hypervisor is one that has been found to be down for stabilization
        number of times with run_interval between each check.
        """
        violated_hypervisors = [
            hypervisor for hypervisor, violation_count in
            six.iteritems(self.violation_counts) if
            violation_count >= self.stabilization]

        return violated_hypervisors

    def get_alarming_hypervisors(self):
        """
        Returns the names of all hypervisors that have been marked down at
        least once (meaning their violation_count is 1 or more).
        """
        alarming_hypervisors = [
            hypervisor for hypervisor, violation_count in
            six.iteritems(self.violation_counts) if
            violation_count > 0]

        return alarming_hypervisors

    def get_violation_count(self, hypervisor_name):
        return self.violation_counts[hypervisor_name]

    def monitor(self, context, all_nodes):
        self.current_interval += 1

        self.update_monitored_hypervisors(context, all_nodes)
        if self.current_interval >= int(self.run_interval):
            self.current_interval = 0
            self.update_violation_counts(context)

    def update_monitored_hypervisors(self, context, all_nodes):
        nodes_to_monitor = []
        if CONF.policy_per_host_aggregate:
            if self.aggregate_id:
                # Get all the hosts for the aggregate
                service_hosts = objects.Aggregate.get_by_id(
                    context, self.aggregate_id).hosts

                # For every compute node, if the host is in the list of hosts
                # for the aggregate, then add it to the list of hypervisors
                # that should be monitored
                nodes_to_monitor = [node for node in all_nodes if
                                    node['host'] in service_hosts]
        else:
            # If we're not running in aggregate mode, then add all compute
            # node hosts to the list of hypervisors we are managing
            nodes_to_monitor = all_nodes

        temp_monitored_hypervisors = {}
        for node in nodes_to_monitor:
            temp_monitored_hypervisors[node['hypervisor_hostname']] = node

        # For any hypervisors that we were previously monitoring but no
        # longer are, remove them from the violation counts map
        for hypervisor in (set(self.monitored_hypervisors.keys()) -
                           set(temp_monitored_hypervisors.keys())):
            if hypervisor in self.violation_counts:
                self.violation_counts.pop(hypervisor)

        self.monitored_hypervisors = temp_monitored_hypervisors

        LOG.debug('Monitored hypervisors for monitor with ID %(id)s: %(hvs)s' %
                  {'id': self.id, 'hvs': self.monitored_hypervisors.keys()})

    def update_violation_counts(self, context):
        down_hypervisors = self._get_currently_down_hypervisors(context)

        for hypervisor in down_hypervisors:
            self.violation_counts[hypervisor] += 1

        # Get the list of hypervisors that were deemed down the last time this
        # was called but are now no longer marked as down and remove them from
        # the violation counts
        recovered_hypervisors = list(
            set(self.violation_counts.keys()) - set(down_hypervisors))
        for hypervisor in recovered_hypervisors:
            self.violation_counts.pop(hypervisor)

        LOG.debug('Violation counts for monitor with ID %(id)s: %(counts)s' %
                  {'id': self.id, 'counts': self.violation_counts})

    def _get_currently_down_hypervisors(self, context):
        down_hypervisors = self.hypervisor_driver.select_hypervisors(
            context,
            aggregate_id=self.aggregate_id)

        # If we aren't managing the hypervisor, then we should remove it
        # from the down_hypervisors list
        down_hypervisors = list(
            set(down_hypervisors) & set(self.monitored_hypervisors.keys()))

        for hypervisor_name, node in (
                six.iteritems(self.monitored_hypervisors)):
            if hypervisor_name not in down_hypervisors:
                continue
            stats = node.stats or {}
            hv_type = node.hypervisor_type

            if ((hv_type == ego_constants.HYPERVISOR_TYPE_VMWARE_DB) or
                    ego_resreq_utils.is_powervc_driver_host(hv_type, stats)):
                down_hypervisors.remove(hypervisor_name)

        LOG.debug('Down hypervisors for monitor with ID %s: %s' %
                  (self.id, down_hypervisors))

        return down_hypervisors
