# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from nova.ibm.ego import ego_gettextutils
from nova.ibm.ego.hapolicy import constants as ha_constants
from nova.ibm.ego.hapolicy import policy_exception
from nova.ibm.ego.hapolicy import utils as hautils

_ = ego_gettextutils._


def get_value(node):
    if len(node.childNodes) > 0:
        return str(node.childNodes[0].data)
    else:
        raise policy_exception.InvalidHAPolicyInput(_("HA policy configuration"
                  " attribute (%s) cannot be blank.") % node.tagName)


def validate_xml(policy):
    if set(policy.keys()) != set(ha_constants.KEYS):
        raise policy_exception.InvalidHAPolicyInput(_("Missing argument in "
               "policy configuration; it needs id, name, description, state, "
               "run_interval and stabilization."))

    hautils.validate_policy_parameters(policy)


class HAPolicy(dict):
    """Custom object for HA policy; each policy is an object of this class."""

    def __init__(self, node):
        self.node = node
        for child in node.childNodes:
            if hasattr(child, 'tagName'):
                if child.tagName in ha_constants.KEYS:
                    self[str(child.tagName)] = get_value(child)

        validate_xml(self)

    def __setitem__(self, key, value):
        if key not in ha_constants.KEYS:
            raise IndexError
        dict.__setitem__(self, key, value)
        node = self.node.getElementsByTagName(key)[0]
        node.childNodes[0].data = value

    def __getitem__(self, key):
        if key == 'aggregates':
            return dict.__getitem__(self, 'aggregates')
        if key not in ha_constants.KEYS:
            raise IndexError
        return dict.__getitem__(self, key)

    def update(self, policy):
        for key, value in policy.items():
            try:
                self.__setitem__(key, value)
            except IndexError:
                continue

    def add_aggregates(self, aggregates):
        """
        If HA service is in aggregate mode, then the policy should have
        aggregate information.
        """
        dict.__setitem__(self, 'aggregates', aggregates)

    def remove_state(self):
        """
        If HA service is in aggregate mode, then the policy should not
        have its 'state' defined.
        """
        if ha_constants.KEY_STATE in self.keys():
            dict.__delitem__(self, ha_constants.KEY_STATE)

    def remove_stabilization(self):
        """
        If HA service is in aggregate mode, then the policy should not
        have its 'stabilization' defined.
        """
        if ha_constants.KEY_STABILIZATION in self.keys():
            dict.__delitem__(self, ha_constants.KEY_STABILIZATION)

    def remove_run_interval(self):
        """
        If HA service is in aggregate mode, then the policy should not
        have its 'run interval' defined.
        """
        if ha_constants.KEY_RUN_INTERVAL in self.keys():
            dict.__delitem__(self, ha_constants.KEY_RUN_INTERVAL)
