# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Policy Engine Manager."""

import codecs
import os

import six

from xml.dom import minidom

from nova import compute
from nova.compute import vm_states
from nova import context
from nova import exception
from nova.ibm.ego import ego_constants
from nova.ibm.ego import ego_gettextutils
from nova.ibm.ego import ego_utils
from nova.ibm.ego.hapolicy import constants as ha_constants
from nova.ibm.ego.hapolicy import engine as engines
from nova.ibm.ego.hapolicy import ha_monitor
from nova.ibm.ego.hapolicy import hapolicy
from nova.ibm.ego.hapolicy import policy_exception
from nova.ibm.ego.hapolicy import utils as ha_utils
from nova.ibm.ego import policy_ceilometer_client
from nova import objects
from nova.objects import instance as instance_obj
from nova.scheduler.ibm.ego import ego_host_aggregate
from nova.scheduler.ibm.ego import ego_manager
from nova.scheduler.ibm.ego import ego_rpcapi

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils

_ = ego_gettextutils._
_LI = ego_gettextutils._LI
_LW = ego_gettextutils._LW
_LE = ego_gettextutils._LE

driver_opts = [
  cfg.StrOpt('ha_policy_hypervisor_driver',
             default='nova.ibm.ego.hapolicy.hypervisor_driver'
                     '.driver.HAHypervisorDriver',
             help='The default driver for HA service to get failed '
                  'hypervisors'),
  cfg.StrOpt('ha_policy_instance_driver',
             default='nova.ibm.ego.hapolicy.instance_driver.driver.'
                     'HAInstanceDriver',
             help='The default driver for the HA service to get instances '
                  'to migrate or rebuild'),
]

CONF = cfg.CONF
CONF.register_opts(driver_opts)
CONF.import_opt('policy_per_host_aggregate', 'nova.ibm.ego.ego_flags')

LOG = logging.getLogger(__name__)


class PolicyEngineManager(object):
    """A daemon that serves the HA policy."""

    def __init__(self, policy_conf_path, prs_driver):
        self.all_policies = []
        self.enabled_policy = None
        self.ha_monitors = []
        self.mt_engines = []
        self.status = ego_utils.get_compute_status()
        self.ego_api = ego_manager.EGOManager()
        self.scheduler_api = ego_rpcapi.EGOSchedulerAPI()
        self.agg_container = ego_host_aggregate.EGOHostAggregate()
        self.compute_api = compute.API()
        self.ceilometer_api = policy_ceilometer_client.API()
        self.host_api = compute.HostAPI()
        self.policy_conf = policy_conf_path
        self.hv_driver = importutils.import_object(
                               CONF.ha_policy_hypervisor_driver)
        self.inst_driver = importutils.import_object(
                                   CONF.ha_policy_instance_driver)
        self.prs_driver = prs_driver
        self.__policy_init()

    @property
    def is_global_mode(self):
        return not CONF.policy_per_host_aggregate

    def __policy_init(self):
        """Perform any necessary policy initialization tasks."""
        if not os.path.isfile(self.policy_conf):
            raise policy_exception.HAPolicyConfFileNotFoundPath(
                                           policy_conf=self.policy_conf)

        LOG.debug('The policy configuration file path is: %s'
                  % self.policy_conf)

        self.__get_all_policies_from_xml()

        if self.is_global_mode:
            for policy in self.all_policies:
                if policy['state'] == ha_constants.ENABLED:
                    self.enabled_policy = policy
                    monitor = ha_monitor.HAMonitor(
                        policy['run_interval'],
                        policy['stabilization'],
                        hypervisor_driver=self.hv_driver)
                    self.ha_monitors.append(monitor)
                    break
        else:
            self.__sync_up_all_aggregates_and_policies()
            LOG.debug("Host aggregate based policies were enabled "
                      "during HA service engine initialization.")

        self.__init_maintenance_engines()
        self.__init_ha_engine()

    def __write_policies_to_xml(self):
        doc = minidom.parseString("<policies></policies>")
        policies = doc.childNodes[0]

        for policy in self.all_policies:
            policies.appendChild(policy.node)

        with codecs.open(self.policy_conf, 'w', 'utf-8') as f:
            doc.writexml(f, addindent='', encoding='utf-8')

    def __sync_up_all_aggregates_and_policies(self):
        id_list = [policy["id"] for policy in self.all_policies]
        aggregates = self.agg_container.get_host_aggregate_metadata()
        # 1) update HA monitors based on aggregate metadata
        for aggr_id in aggregates:
            metadata = aggregates[aggr_id]
            if ha_constants.KEY_HA_ID not in metadata:
                self.__remove_ha_monitor(aggr_id)
                continue
            if (ha_constants.KEY_HA_RUN_INTERVAL not in metadata or
                ha_constants.KEY_HA_STABILIZATION not in metadata):
                LOG.warning(_LW("Missing argument in metadata for aggregate "
                           "ID %(aid)s.  Provide the hapolicy-stabilization "
                           "and hapolicy-run_interval values as well.")
                         % {'aid': aggr_id})
                self.__remove_ha_monitor(aggr_id)
                continue
            try:
                ha_utils.validate_policy_parameters(metadata, id_list=id_list)
            except Exception as ex:
                LOG.warning(_LW("Cannot apply HA policy for aggregate "
                           "ID %(aid)s. Reason: %(reason)s")
                         % {'aid': aggr_id,
                            'reason': six.text_type(ex)})
                self.__remove_ha_monitor(aggr_id)
                continue
            self.__update_or_add_ha_monitor(
                metadata[ha_constants.KEY_HA_RUN_INTERVAL],
                metadata[ha_constants.KEY_HA_STABILIZATION],
                aggr_id)
        # 2) remove HA monitors whose aggregate has been removed.
        ha_monitors_copy = self.ha_monitors[:]
        aggregate_ids = aggregates.keys()
        for monitor in ha_monitors_copy:
            if long(monitor.aggregate_id) not in aggregate_ids:
                self.ha_monitors.remove(monitor)
                LOG.info(_LI("Removed aggregate ID %(aid)s from the HA "
                               "service.") % {'aid': monitor.aggregate_id})

    def __update_or_add_ha_monitor(self, run_interval, stabilization,
                                   aggregate_id):
        existing_monitor = self.__find_ha_monitor(aggregate_id)
        if existing_monitor:
            value_changed = False
            if int(stabilization) != existing_monitor.stabilization:
                existing_monitor.set_stabilization(stabilization)
                value_changed = True
            if int(run_interval) != existing_monitor.run_interval:
                existing_monitor.set_run_interval(run_interval)
                value_changed = True
            if value_changed:
                existing_monitor.reset()
                LOG.info(_LI("Updated aggregate ID %(aid)s for the HA "
                           "service.") % {'aid': aggregate_id})
        else:
            monitor = ha_monitor.HAMonitor(
                run_interval,
                stabilization,
                aggregate_id=aggregate_id,
                hypervisor_driver=self.hv_driver)
            self.ha_monitors.append(monitor)
            LOG.info(_LI("Added aggregate ID %(aid)s to the HA service.")
                     % {'aid': aggregate_id})

    def __find_ha_monitor(self, aggregate_id):
        """
        Finds the first monitor in self.ha_monitors that is configured to
        monitor the aggregate with the specific ID.
        """
        monitors = [m for m in self.ha_monitors if
                    m.aggregate_id == str(aggregate_id)]
        return monitors[0] if monitors else None

    def __remove_ha_monitor(self, aggregate_id):
        monitor = self.__find_ha_monitor(aggregate_id)
        if monitor:
            self.ha_monitors.remove(monitor)
            LOG.info(_LI("Removed aggregate ID %(aid)s from the HA "
                         "service.") % {'aid': aggregate_id})

    def __get_aggregates(self, policy_id):
        host_aggregates_metadata = (
            self.agg_container.get_host_aggregate_metadata())

        def _is_valid(pid, metadata):
            try:
                ha_utils.validate_policy_parameters(metadata)
            except Exception:
                return False

            if (metadata.get(ha_constants.KEY_HA_ID, None) != pid or
                ha_constants.KEY_HA_RUN_INTERVAL not in metadata or
                    ha_constants.KEY_HA_STABILIZATION not in metadata):
                return False

            return True

        host_aggregates = []
        for aggr_id, metadata in host_aggregates_metadata.items():
            if metadata.get(ha_constants.KEY_HA_ID, None) == policy_id:
                m = {}
                if ha_constants.KEY_HA_RUN_INTERVAL in metadata:
                    m[ha_constants.KEY_RUN_INTERVAL] = (
                        metadata[ha_constants.KEY_HA_RUN_INTERVAL])
                if ha_constants.KEY_HA_STABILIZATION in metadata:
                    m[ha_constants.KEY_STABILIZATION] = (
                        metadata[ha_constants.KEY_HA_STABILIZATION])
                m[ego_constants.POLICY_VALID_FLAG] = (
                                _is_valid(policy_id, metadata))
                host_aggregates.append({'policy_id': policy_id,
                                        'aggregate_id': aggr_id,
                                        'metadata': m})

        return host_aggregates

    def __get_all_policies_from_xml(self):
        doc = minidom.parse(self.policy_conf)
        nodes = doc.getElementsByTagName('hapolicy')
        for node in nodes:
            policy = hapolicy.HAPolicy(node)
            exist = False
            for p in self.all_policies:
                if p.get(ha_constants.KEY_ID) == policy.get(
                                        ha_constants.KEY_ID):
                    exist = True
                    break
            if exist:
                LOG.error(_LE('Policy %(policy)s with same ID already exists;'
                          ' this one will be ignored.') % {'policy': policy})
                continue
            self.all_policies.append(policy)

    def get_all(self):
        """Get all runtime policies from the hapolicy_conf.xml file."""
        for policy in self.all_policies:
            if not self.is_global_mode:
                policy.remove_state()
                policy.remove_stabilization()
                policy.remove_run_interval()
        return [dict(policy) for policy in self.all_policies]

    def get(self, policy_id):
        policy = None
        for po in self.all_policies:
            if po.get('id', None) == policy_id:
                policy = po
        if not policy:
            raise policy_exception.HAPolicyNotFound(policy_id=policy_id)

        if not self.is_global_mode:
            aggregates = self.__get_aggregates(policy_id)
            policy.remove_state()
            policy.remove_stabilization()
            policy.remove_run_interval()
            policy.add_aggregates(aggregates)

        return dict(policy)

    def hypervisor_list(self, context, search_opts):
        """
        Returns a list hypervisors' hapolicy info based on any search
        options specified.
        """
        alert = search_opts.get('alert')
        aggregate_id = search_opts.get('aggregate_id')
        temp_ha_monitors = []
        if aggregate_id:
            if not CONF.policy_per_host_aggregate:
                raise exception.InvalidInput(_('Cannot support aggregate-id '
                                               'option in global mode (policy'
                                               '_per_host_aggregate = False is'
                                               ' defined in the nova.conf file'
                                               ').'))
            else:
                self.__validate_aggregate_id(aggregate_id)

            monitor = self.__find_ha_monitor(aggregate_id)
            if monitor:
                temp_ha_monitors.append(monitor)
        else:
            temp_ha_monitors = self.ha_monitors

        hypervisors = {}
        for monitor in temp_ha_monitors:
            hv_names = []
            if alert:
                hv_names = monitor.get_alarming_hypervisors()
            else:
                hv_names = monitor.get_monitored_hypervisors()

            for hv_name in hv_names:
                if (monitor.monitored_hypervisors[hv_name]['hypervisor_type']
                       in [ego_constants.HYPERVISOR_TYPE_VMWARE_DB,
                           ego_constants.DOCKER_HYPERVISOR_TYPE]):
                    continue

                hv_id = monitor.monitored_hypervisors[hv_name]['id']
                hv_info = hypervisors.get(hv_id)
                if hv_info is None:
                    hv_info = {
                        'id': hv_id,
                        'hypervisor_hostname': hv_name,
                        'status': self.ha_engine.get_status(hv_name),
                        'monitors': []}

                # add this monitor's info
                monitor_info = {
                    'run_interval': monitor.run_interval,
                    'stabilization': monitor.stabilization,
                    'consecutive_alert_times': (
                        monitor.get_violation_count(hv_name))}
                if monitor.aggregate_id is not None:
                    monitor_info['aggregate_id'] = monitor.aggregate_id
                hv_info['monitors'].append(monitor_info)

                hypervisors[hv_id] = hv_info

        return sorted(hypervisors.values(), key=lambda k: k['id'])

    def event_list(self, context, search_opts, marker=None, limit=None):
        """
        Returns a list of hapolicy events based on search options specified.
        """
        hypervisor_id = search_opts.get('hypervisor_id')
        if hypervisor_id:
            self.__validate_hypervisor_id(hypervisor_id)
            hypervisor_id = long(hypervisor_id)
            search_opts['hypervisor_id'] = hypervisor_id
        return self.__load_events(context, search_opts, marker, limit)

    def __load_events(self, context, search_opts=None, marker=None,
                      limit=None):
        query = []
        query.append({'field': 'policy_type', 'op': 'eq', 'value': 'hapolicy',
                      'type': 'string'})
        if search_opts:
            if search_opts.get('hypervisor_id'):
                query.append({'field': 'hypervisor_id', 'op': 'eq',
                              'value': search_opts['hypervisor_id'],
                              'type': 'integer'})

        return self.ceilometer_api.get_events(query, marker, limit)

    def __validate_hypervisor_id(self, hypervisor_id):
        if not hypervisor_id:
            raise exception.InvalidInput(_('Specify a valid hypervisor ID.'))
        try:
            int(hypervisor_id)
        except Exception:
            raise exception.InvalidInput(_('Hypervisor (%s) is not valid. '
                                           'Specify a numeric hypervisor '
                                           'ID.') % hypervisor_id)

    def __validate_aggregate_id(self, aggregate_id):
        if not aggregate_id:
            raise exception.InvalidInput(_('Specify a valid aggregate ID.'))
        try:
            int(aggregate_id)
        except Exception:
            raise exception.InvalidInput(_('Aggregate (%s) is not valid. '
                                           'Specify a numeric aggregate '
                                           'ID.') % aggregate_id)

    def __enable_policy(self, policy):
        self.enabled_policy = policy

        self.ha_monitors = []
        monitor = ha_monitor.HAMonitor(
            policy['run_interval'],
            policy['stabilization'],
            hypervisor_driver=self.hv_driver)
        self.ha_monitors.append(monitor)

        LOG.info(_LI("The policy engine manager successfully changed "
                   "the policy's (%(policy_id)s) state (%(state)s).")
                 % {'policy_id': policy['id'],
                    'state': 'enabled'})

    def __disable_policy(self, policy):
        if self.enabled_policy and self.enabled_policy['id'] == policy['id']:
            self.enabled_policy = None
            self.ha_monitors = []
            LOG.info(_LI("The policy engine manager successfully changed "
                       "the policy's (%(policy_id)s) state (%(state)s).")
                     % {'policy_id': policy['id'],
                        'state': 'disabled'})

    def __update_policy(self, updates, policy):
        policy.update(updates)

        for po in self.all_policies:
            if po.get('id', None) == policy['id']:
                po.update(policy)
                break

        self.__write_policies_to_xml()

        if not self.is_global_mode:
            return

        if 'state' in updates:
            if updates['state'] == ha_constants.ENABLED:
                self.__enable_policy(policy)
            else:
                self.__disable_policy(policy)
        elif (ha_constants.KEY_RUN_INTERVAL in updates or
              ha_constants.KEY_STABILIZATION in updates):
            # if update run_interval or stabilization for current running
            # policy, it must be in global mode; then update it
            if (self.enabled_policy and
                    self.enabled_policy['id'] == policy['id'] and
                    self.ha_monitors):
                if ha_constants.KEY_RUN_INTERVAL in updates:
                    self.ha_monitors[0].set_run_interval(
                        updates.get(ha_constants.KEY_RUN_INTERVAL))
                if ha_constants.KEY_STABILIZATION in updates:
                    self.ha_monitors[0].set_stabilization(
                        updates.get(ha_constants.KEY_STABILIZATION))

    def update(self, policy_id=None, updates=None):
        """Update specified runtime policy from the hapolicy_conf.xml file."""
        policy = self.get(policy_id)
        LOG.info(_LI("The policy engine manager will update policy (%s).")
                 % policy_id)
        updates = updates['updates']

        # 1. check validation of input arguments
        if not self.is_global_mode:
            if (ha_constants.KEY_STATE in updates or
                ha_constants.KEY_RUN_INTERVAL in updates or
                    ha_constants.KEY_STABILIZATION in updates):
                raise policy_exception.InvalidHAPolicyAction(
                          action='update',
                          policy_id=policy_id,
                          reason='policy_per_host_aggregate = True '
                                 'is defined in the nova.conf file')
        else:
            if (ha_constants.KEY_STATE in updates and
                updates[ha_constants.KEY_STATE] == ha_constants.ENABLED and
                self.enabled_policy and
                    self.enabled_policy['id'] != policy['id']):
                raise policy_exception.InvalidHAPolicyAction(
                        action='update',
                        policy_id=policy['id'],
                        reason='already has an enabled policy')

        ha_utils.validate_policy_parameters(updates)

        # 2. update XML and runtime memory
        self.__update_policy(updates, policy)

        return policy

    def policy_function(self, context):
        """
        1. Verify instances HA/host maintenance migrations
        2. Run maintenance actions
        3. Check if HA service is disabled
        4. If possible, run HA actions
        5. Refresh compute_status.
            a. Remove non-existent hosts from maintenance
            b. Remove non-existent and non-managed hosts from HA
        """
        all_nodes = objects.ComputeNodeList.get_all(context)

        # 1) confirm resize the instances which were migrated automatically
        #    by HA/Maintenance
        self.__verify_migration(context)

        # 2) remove the Maintenance engine for the hypervisor which has been
        #    removed.
        all_hv_names = [node['hypervisor_hostname'] for node in all_nodes]
        obsolete_mts = ([mt for mt in self.mt_engines if mt.hv_name not in
                             all_hv_names])
        for obslt_mt in obsolete_mts:
            self.__disable_maintenance(context, obslt_mt.hv_name)

        # 3) do maintenance periodic task
        for mte in self.mt_engines:
            mte.run(context, all_nodes)

        # 4) sync up HA monitors with aggregates' metadata related to HA
        if not self.is_global_mode:
            self.__sync_up_all_aggregates_and_policies()

        # hypervisors that are managed by HA
        ha_managed_hvs = set()
        service_disabled = ha_utils.is_service_disabled_by_topic(context,
                                                                 'hapolicy')
        # 5) do HA periodic task
        if not service_disabled:
            for monitor in self.ha_monitors:
                monitor.monitor(context, all_nodes)

            ha_managed_hvs = self._get_all_ha_monitored_hypervisors()
            violated_hvs = self._get_all_ha_violated_hypervisors()
            try:
                self.ha_engine.run(
                    context,
                    all_nodes,
                    managed_hypervisors=ha_managed_hvs,
                    violated_hypervisors=violated_hvs)
            except Exception as e:
                LOG.warning(_LW('%s') % six.text_type(e))

        else:
            ha_managed_hvs = self._get_all_ha_monitored_hypervisors()
            for monitor in self.ha_monitors:
                monitor.reset()
            self.ha_engine.reset()

        # 6) make the data persist.
        self.status.refresh(all_nodes, ha_managed_hvs)
        ego_utils.set_compute_status(self.status.get_json())

    def _get_all_ha_violated_hypervisors(self):
        """
        Returns a set of violated hypervisors (as names) across all of the
        HA monitors.
        """
        violated_hypervisors = set()
        for m in self.ha_monitors:
            violated_hypervisors |= set(m.get_violated_hypervisors())

        LOG.debug('Set of violated hypervisors across all monitors: %s' %
                  violated_hypervisors)

        return violated_hypervisors

    def _get_all_ha_monitored_hypervisors(self, need_update_monitored=False):
        """
        Returns a set of monitored hypervisors (as names) across all of the
        HA monitors.
        """
        monitored_hypervisors = set()
        if need_update_monitored:
            try:
                ctxt = self.prs_driver.get_context()
            except Exception as e:
                LOG.warning(_LW("Encountered exception fetching driver "
                            "extension context; falling back to standard "
                            "admin context; exception message is: %(message)s")
                        % {'message': six.text_type(e)})
                ctxt = context.get_admin_context()
            all_nodes = objects.ComputeNodeList.get_all(ctxt)
            for m in self.ha_monitors:
                m.update_monitored_hypervisors(ctxt, all_nodes)
                monitored_hypervisors |= set(m.get_monitored_hypervisors())
        else:
            for m in self.ha_monitors:
                monitored_hypervisors |= set(m.get_monitored_hypervisors())

        LOG.debug('Set of monitored hypervisors across all monitors: %s' %
                  monitored_hypervisors)

        return monitored_hypervisors

    def __verify_migration(self, context):
        """
        To make sure HA just verify instances we migrated, the HA service needs
        to perform the following tasks:
        1. Get all resized finished migrations
        2. Iterate over migrations
        3. Get actions of an instance
        4. Check request_id of instance action
        5. If it starts with 'prs-ha-' and it is migrated by HA
        """
        filters = {'status': 'finished'}
        migrations = objects.MigrationList.get_by_filters(context, filters)
        for mig in migrations:
            if mig['old_instance_type_id'] == mig['new_instance_type_id']:
                actions = objects.InstanceActionList.get_by_instance_uuid(
                            context, mig['instance_uuid'])
                if len(actions) > 0:
                    request_id = actions[0]['request_id']
                    if not request_id.startswith(ha_constants.PREFIX_PRS_HA):
                        continue
                inst = objects.Instance.get_by_uuid(
                        context, mig['instance_uuid'],
                        expected_attrs=instance_obj.INSTANCE_DEFAULT_FIELDS)
                try:
                    self.compute_api.confirm_resize(context, inst)
                    LOG.info(_LI("Successfully confirmed resize of instance "
                               "%(name)s (%(instance)s).")
                             % {'name': inst.name,
                                'instance': inst.uuid})
                except Exception as ex:
                    LOG.warning(_LW("Failed to confirm resize of instance "
                               "%(name)s (%(instance)s) due to: %(reason)s")
                             % {'name': inst.name,
                                'instance': inst.uuid,
                                'reason': six.text_type(ex.message)})

    def get_hv_status(self):
        """Return the hosts' HA and maintenance status."""
        return self.status.get_json()

    def set_maintenance(self, context, hv_name, status, migrate=None,
                        target_host=None):
        if status == 'enable':
            self.__enable_maintenance(context, hv_name, migrate,
                                      target_host)
        else:
            self.__disable_maintenance(context, hv_name)

    def __enable_maintenance(self, context, hv_name, migrate,
                             target_host):
        # check if this hypervisor can enter maintenance mode
        result = self.hv_driver.can_enter_maintenance(context,
                                                      hv_name, migrate)
        if not result[0]:
            if len(result) == 2:
                # i.e., the exploiter provided their own custom message
                reason = result[1]
            else:
                # otherwise just generate a default message
                reason = _("The HA driver prevented the hypervisor (%s) from "
                           "entering maintenance mode.") % hv_name
            raise policy_exception.HostMaintenanceError(reason=reason)

        state = self.status.get_maintenance(hv_name)
        if state != ha_constants.STATE_OK:
            msg = (_("The hypervisor (%s) is already in maintenance mode.")
                   % hv_name)
            raise policy_exception.HostMaintenanceError(reason=msg)

        # Also disable the nova-compute to prevent the hypervisor from showing
        # up in the live migration target list.
        nodes = objects.ComputeNodeList.get_all(context)
        for node in nodes:
            if node['hypervisor_hostname'] == hv_name:
                try:
                    self.host_api.service_update(
                        context, node['host'],
                        'nova-compute', {'disabled': True, 'disabled_reason':
                        'The nova-compute service is disabled since the '
                        'service host is in maintenance mode.'})
                    LOG.info(_LI("Successfully disabled nova-compute for "
                               "(%(name)s) when enabling maintenance mode.")
                             % {'name': hv_name})
                    break
                except exception.NovaException as e:
                    msg = (_LE("Failed to disable nova-compute for "
                             "(%(name)s) when enabling maintenance "
                             "mode due to: (%(err)s).")
                           % {'name': hv_name,
                              'err': e.format_message()})
                    LOG.error(msg)
                    raise policy_exception.HostMaintenanceError(reason=msg)

        mte = engines.MaintenanceEngine(hv_name=hv_name,
                                        migrate=migrate,
                                        status=self.status,
                                        ego_api=self.ego_api,
                                        target_host=target_host,
                                        inst_driver=self.inst_driver,
                                        prs_driver=self.prs_driver)
        # NOTE(jwcroppe): we just append the maintenance engine to the list
        # without any action except marking its MT status to 'entering',
        # so that the periodic task will pick it up during the next cycle;
        # this avoids potentially long-running synchronous REST API calls
        # that would otherwise block waiting for X migrations to be called.
        mte.run(context, nodes, do_action=False)
        ego_utils.set_compute_status(self.status.get_json())
        self.mt_engines.append(mte)
        self.scheduler_api.ego_sync_up_hosts(context)

    def __disable_maintenance(self, context, hv_name):
        state = self.status.get_maintenance(hv_name)
        if state not in ha_constants.STATE_LIST_MT:
            msg = _("Cannot disable maintenance mode for the hypervisor (%s) "
                    "because it is not in maintenance mode.") % hv_name
            raise policy_exception.HostMaintenanceError(reason=msg)
        match_found = False
        nodes = objects.ComputeNodeList.get_all(context)
        # Enable the nova-compute since it's been disabled when entering
        # maintenance mode.
        for node in nodes:
            if node['hypervisor_hostname'] == hv_name:
                try:
                    self.host_api.service_update(
                        context, node['host'],
                        'nova-compute', {'disabled': False, 'disabled_reason':
                                                                        None})
                    LOG.info(_LI("Successfully enabled nova-compute for "
                               "(%(name)s) when disabling maintenance mode.")
                             % {'name': hv_name})
                    break
                except exception.NovaException as e:
                    msg = (_LE("Failed to enable nova-compute for "
                             "(%(name)s) when disabling maintenance mode "
                             "due to: (%(err)s).")
                           % {'name': hv_name,
                              'err': e.format_message()})
                    LOG.error(msg)
                    raise policy_exception.HostMaintenanceError(reason=msg)

        for mte in self.mt_engines:
            if mte.hv_name == hv_name:
                mte.update_hv_state(context, hv_name, ha_constants.STATE_OK)
                self.mt_engines.remove(mte)
                match_found = True
                self.scheduler_api.ego_sync_up_hosts(context)
                break
        if not match_found:
            # just a safety net to handle unexpected cases
            LOG.error(_LE("Didn't find a matching hypervisor for %s in the "
                        "maintenance mode engines; the disable operation "
                        "will likely fail and the HA service will need to "
                        "be restarted to remedy this!") % hv_name)
        else:
            LOG.info(_LI("Successfully disabled maintenance mode for "
                       "hypervisor %(hv_name)s.") % {'hv_name': hv_name})
            ego_utils.set_compute_status(self.status.get_json())

    def notify_policy_resource_status(self, context, msg):
        """
        Notify HA service to take related actions based on
        notification message
        """
        event_type = msg['event_type']
        payload = msg['payload']
        vm_state = payload.get('state')

        if vm_state == vm_states.ERROR:
            self.__move_inst_fail(context, payload, False)
        elif vm_state == vm_states.ACTIVE:
            if event_type == (
                ego_constants.EVENT_TYPE_COMPUTE_INST_LIVE_MIG_ROLLBACK_END):
                self.__move_inst_fail(context, payload)
        elif event_type == ego_constants.EVENT_TYPE_COMPUTE_NODE_DELETE_END:
            hv_name = payload.get('hypervisor_hostname')
            if hv_name in [mt.hv_name for mt in self.mt_engines]:
                self.__disable_maintenance(context, hv_name)
        if event_type in (
            [ego_constants.EVENT_TYPE_COMPUTE_INST_RESIZE_END,
             ego_constants.EVENT_TYPE_COMPUTE_INST_LIVE_MIGRATION_POST_END,
             ego_constants.EVENT_TYPE_COMPUTE_INST_REBUILD_END]):
            if context.request_id in self.ha_engine.hapolicy_action:
                self.ha_engine.move_inst_success(context)

        if self.is_global_mode:
            return
        # Aggregate metadata delete: delete policy engine
        if event_type == ego_constants.EVENT_TYPE_AGGREGATE_DELETE_END:
            LOG.debug("Host aggregate was deleted with message: %s" % msg)
            agg_id = payload['aggregate_id']
            self.__remove_ha_monitor(agg_id)
        # Aggregate metadata update: update policy engine
        elif event_type == ego_constants.EVENT_TYPE_AGGREGATE_METADATA_SET_END:
            LOG.debug("Host aggregate metadata updated with message: %s"
                      % msg)
            agg_id = payload['aggregate_id']
            self.__update_aggregate_metadata(agg_id)
        elif event_type in (
                 [ego_constants.EVENT_TYPE_AGGREGATE_HOST_ADD_END,
                  ego_constants.EVENT_TYPE_AGGREGATE_HOST_REMOVE_END]):
            self.__update_monitored_hvs(context, payload['aggregate_id'])

    def __update_monitored_hvs(self, context, aggregate_id):
        monitor = self.__find_ha_monitor(aggregate_id)
        if monitor:
            monitor.update_monitored_hypervisors(
                context, objects.ComputeNodeList.get_all(context))

    def __move_inst_fail(self, context, payload, is_live_migration=True):
        instid = payload.get('instance_id')
        engines = set([self.ha_engine]) | set(self.mt_engines)
        for eg in engines:
            hv_moving_inst_map = eg.get_hv_moving_inst_map()
            for hv, inst_list in hv_moving_inst_map.items():
                if instid in inst_list:
                    eg.move_inst_fail(context, hv, payload, is_live_migration)
                    return

    def __update_aggregate_metadata(self, aggregate_id):
        # get current metadata from aggregate instead of payload data
        # due to it may be obsolete
        aggregates = self.agg_container.get_host_aggregate_metadata()
        aggr_meta = aggregates[long(aggregate_id)]

        # Check if id is valid in metadata
        id_list = [policy["id"] for policy in self.all_policies]

        # validate metadata first
        try:
            ha_utils.validate_policy_parameters(aggr_meta, id_list=id_list)
        except Exception as ex:
            LOG.warning(_LW("Cannot apply HA policy for aggregate ID %(aid)s. "
                       "Reason: %(reason)s")
                     % {'aid': aggregate_id,
                        'reason': six.text_type(ex)})
            self.__remove_ha_monitor(aggregate_id)
            return

        # Flag to indicate whether this policy should be removed.
        need_to_remove = False

        # if hapolicy-id exists, check metadata, otherwise don't check.
        # if any metadata is missing, remove this policy engine; this can
        # happen if a user deletes hapolicy-stabilization or
        # hapolicy-run_interval
        if ha_constants.KEY_HA_ID not in aggr_meta:
            need_to_remove = True
        else:
            if (ha_constants.KEY_HA_RUN_INTERVAL not in aggr_meta or
                ha_constants.KEY_HA_STABILIZATION not in aggr_meta):
                LOG.warning(_LW("Missing argument in metadata for aggregate "
                           "ID %(aid)s. Provide the hapolicy-stabilization "
                           "and hapolicy-run_interval values.")
                         % {'aid': aggregate_id})
                need_to_remove = True

        if need_to_remove:
            self.__remove_ha_monitor(aggregate_id)
            return

        self.__update_or_add_ha_monitor(
            aggr_meta[ha_constants.KEY_HA_RUN_INTERVAL],
            aggr_meta[ha_constants.KEY_HA_STABILIZATION],
            aggregate_id)

    def __init_maintenance_engines(self):
        hvs = self.status.get_maintenance_hvs()
        for hv_name in hvs:
            migrate = self.status.get_migrate(hv_name)
            target_host = self.status.get_target_host(hv_name)
            mtengine = engines.MaintenanceEngine(hv_name=hv_name,
                                          migrate=migrate,
                                          status=self.status,
                                          ego_api=self.ego_api,
                                          target_host=target_host,
                                          inst_driver=self.inst_driver,
                                          prs_driver=self.prs_driver)
            self.mt_engines.append(mtengine)

    def __init_ha_engine(self):
        ha_managed_hvs = self._get_all_ha_monitored_hypervisors(
                                         need_update_monitored=True)
        self.ha_engine = engines.HAEngine(
            ego_api=self.ego_api,
            status=self.status,
            hv_driver=self.hv_driver,
            inst_driver=self.inst_driver,
            prs_driver=self.prs_driver,
            ha_managed_hvs=ha_managed_hvs)
