# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Client side of the hapolicy RPC API."""

from oslo_config import cfg
import oslo_messaging as messaging

from nova.objects import base as objects_base
from nova import rpc

rpcapi_opts = [
    cfg.StrOpt('hapolicy_topic',
               default='hapolicy',
               help='The topic on which the hapolicy service listens'),
]

CONF = cfg.CONF
CONF.register_opts(rpcapi_opts)

rpcapi_cap_opt = cfg.StrOpt('hapolicy',
            help='Set a version cap for messages sent to hapolicy service')
CONF.register_opt(rpcapi_cap_opt, 'upgrade_levels')


class HAPolicyAPI(object):
    '''Client side of the hapolicy RPC API.

    API version history:

        1.0 - Initial version.

    '''

    VERSION_ALIASES = {
        'havana': '1.0'
    }

    def __init__(self, topic=None):
        super(HAPolicyAPI, self).__init__()
        target = messaging.Target(topic=CONF.hapolicy_topic, version="1.0")
        version_cap = self.VERSION_ALIASES.get(CONF.upgrade_levels.hapolicy,
                                               CONF.upgrade_levels.hapolicy)
        serializer = objects_base.NovaObjectSerializer()
        self.client = rpc.get_client(target,
                                     version_cap=version_cap,
                                     serializer=serializer)

    def get_all(self, context):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'get_all')

    def get(self, context, policy_id):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'get', policy_id=policy_id)

    def update(self, context, policy_id, updates):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'update', policy_id=policy_id,
                                             updates=updates)

    def set_maintenance(self, context, hv_name, status, migrate=None,
                        target_host=None):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'set_maintenance', hv_name=hv_name,
                          status=status, migrate=migrate,
                          target_host=target_host)

    def get_hv_status(self, context):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'get_hv_status')

    def hypervisor_list(self, context, search_opts):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'hypervisor_list', search_opts=search_opts)

    def event_list(self, context, search_opts, marker, limit):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'event_list', search_opts=search_opts,
                          marker=marker, limit=limit)
