# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Utilities for the HA policy service."""

from nova.ibm.ego import ego_constants
from nova.ibm.ego import ego_gettextutils
from nova.ibm.ego.hapolicy import constants as ha_constants
from nova.ibm.ego.hapolicy import policy_exception
from nova.scheduler.ibm.ego import ego_resreq_utils

from nova import objects
from nova import utils

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils

_ = ego_gettextutils._
_LW = ego_gettextutils._LW
_hv_drvr = None

LOG = logging.getLogger(__name__)


def get_image_from_inst(instance):
    # Get the system metadata from the instance
    system_meta = utils.instance_sys_meta(instance)

    # Convert the system metadata to image metadata
    return utils.get_image_from_system_metadata(system_meta)


def hypervisor_is_up(context, hv_name, **kwargs):
    """Check whether the hypervisor is up."""

    global _hv_drvr
    # we do this here to avoid circular imports
    if _hv_drvr is None:
        CONF = cfg.CONF
        CONF.import_opt('ha_policy_hypervisor_driver',
                        'nova.ibm.ego.hapolicy.policy_engine_manager')
        _hv_drvr = importutils.import_object(CONF.ha_policy_hypervisor_driver)
    return _hv_drvr.hypervisor_is_up(context, hv_name, **kwargs)


def validate_policy_parameters(policy, id_list=None):
    def _check_parameter(policy, key1, key2):
        key = None
        if key1 in policy:
            key = key1
        elif key2 in policy:
            key = key2
        else:
            return

        try:
            if (int(policy[key]) < ego_constants.MIN_POSITIVE_INT or
                    int(policy[key]) > ego_constants.MAX_INT_32):
                raise Exception()
        except Exception:
            raise policy_exception.InvalidHAPolicyInput(
                                    value=policy[key],
                                    attribute=key,
                                    min=ego_constants.MIN_POSITIVE_INT,
                                    max=ego_constants.MAX_INT_32)

    # Check if id is valid in metadata
    if id_list:
        if (("hapolicy-id" in policy) and
               (policy["hapolicy-id"] not in id_list)):
            raise policy_exception.HAPolicyNotFound(
                    policy_id=policy["hapolicy-id"])

    _check_parameter(policy, ha_constants.KEY_ID, ha_constants.KEY_HA_ID)
    _check_parameter(policy, ha_constants.KEY_RUN_INTERVAL,
                     ha_constants.KEY_HA_RUN_INTERVAL)
    _check_parameter(policy, ha_constants.KEY_STABILIZATION,
                     ha_constants.KEY_HA_STABILIZATION)

    # Test state if exists
    key_state = ha_constants.KEY_STATE
    if key_state in policy:
        if (policy[key_state] not in [ha_constants.ENABLED,
                                      ha_constants.DISABLED]):
            raise policy_exception.InvalidHAPolicyState()


def is_service_disabled_by_topic(context, topic):
    service_disabled = True
    services = objects.ServiceList.get_by_topic(context, topic)
    for service in services:
        if not service['disabled']:
            service_disabled = False
            break
    return service_disabled


def mark_node_unavailable(ego_api, node_name):
    ego_host = ego_resreq_utils.get_rpi_res_name(node_name)
    cstr = ego_resreq_utils.generate_res_state_and_attr_str(ego_host,
                ego_constants.HOST_RES_UNAVAIL)
    ego_api.ego_set_resource_info_str(cstr)


def check_hv_status_json_keys(json_stat):
    ha_fixed_key = ['state', 'created_at', 'updated_at']
    mt_fixed_key = ['state', 'migrate', 'created_at', 'updated_at',
                    'target_host']

    def check_keys(json_status, func_item, check_item):
        hv_names = json_status[func_item].keys()
        for hv_name in hv_names:
            diff_item = (set(check_item) -
                set(json_status[func_item][hv_name].keys()))
            if len(diff_item) != 0:
                for diff_keys in diff_item:
                    LOG.warning(_LW('The %(f_item)s state key for hypervisor '
                               '%(hyper_name)s is missing the key %(st_key)s.')
                             % {'f_item': func_item,
                                'hyper_name': hv_name,
                                'st_key': diff_keys})
                json_status[func_item].pop(hv_name)
                LOG.warning(_LW('The %(f_item)s state key for hypervisor '
                           '%(hyper_name)s was deleted because it was '
                           'unexpectedly corrupted and this deletion '
                           'will remedy the situation.')
                         % {'f_item': func_item, 'hyper_name': hv_name})
    check_keys(json_stat, 'ha', ha_fixed_key)
    check_keys(json_stat, 'maintenance', mt_fixed_key)
