# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import six

from oslo_utils import timeutils

from nova.ibm.ego.hapolicy import constants as ha_constants

VERSION = '3.0'


class HypervisorStatus(object):
    def __init__(self, json):
        self.mt_hvs = json['maintenance']
        self.ha_hvs = json['ha']

    def get_ha(self, hv_name):
        """
        Get HA status of a hypervisor
        """
        if hv_name not in self.ha_hvs:
            return ha_constants.STATE_OK
        return self.ha_hvs[hv_name]['state']

    def get_maintenance(self, hv_name):
        """
        Get maintenance status of a hypervisor
        """
        if hv_name not in self.mt_hvs:
            return ha_constants.STATE_OK
        hv = self.mt_hvs[hv_name]
        return hv.get('state', ha_constants.STATE_OK)

    def get_migrate(self, hv_name):
        if hv_name not in self.mt_hvs:
            return '-'
        return self.mt_hvs[hv_name]['migrate']

    def get_maintenance_hvs(self):
        return self.mt_hvs.keys()

    def get_ha_moving_insts(self, hv_name, action=None):
        if hv_name not in self.ha_hvs:
            return None

        moving_insts = self.ha_hvs[hv_name].get('moving_insts')
        if action and moving_insts:
            moving_insts = {k: v for k, v in six.iteritems(moving_insts) if
                            v.get('action') == action}

        return moving_insts

    def get_maintenance_moving_insts(self, hv_name):
        if hv_name not in self.mt_hvs:
            return None
        return self.mt_hvs[hv_name].get('moving_insts')

    def get_ha_moving_inst_start(self, hv_name, inst_uuid):
        if hv_name not in self.ha_hvs:
            return None
        return self.ha_hvs[hv_name].get('inst_start', {}).get(inst_uuid)

    def get_mt_moving_inst_start(self, hv_name, inst_uuid):
        if hv_name not in self.mt_hvs:
            return None
        return self.mt_hvs[hv_name].get('inst_start', {}).get(inst_uuid)

    def get_ha_timestamp(self, hv_name):
        """Get ha timestamp (created_at, updated_at) of the hypervisor."""
        if hv_name not in self.ha_hvs:
            ha_ts = {'created_at': '-', 'updated_at': '-'}
        else:
            ha_ts = {'created_at': self.ha_hvs[hv_name]['created_at'],
                     'updated_at': self.ha_hvs[hv_name]['updated_at']
                    }
        return ha_ts

    def get_maintenance_timestamp(self, hv_name):
        """
        Get maintenance timestamp (created_at, updated_at) of the hypervisor
        """
        if hv_name not in self.mt_hvs:
            mt_ts = {'created_at': '-', 'updated_at': '-'}
        else:
            mt_ts = {'created_at': self.mt_hvs[hv_name]['created_at'],
                     'updated_at': self.mt_hvs[hv_name]['updated_at']
                    }
        return mt_ts

    def get_ha_failed_instances(self, hv_name):
        """Get the instances which ha failed to move away."""
        if hv_name not in self.ha_hvs:
            failed_instances = []
        else:
            failed_instances = (self.ha_hvs[hv_name].get('failed_instances') or
                                [])
        return failed_instances

    def set_ha_failed_instances(self, hv_name, inst_uuid_list):
        """Set the instances ha failed to move away for a hypervisor."""
        if hv_name in self.ha_hvs:
            self.ha_hvs[hv_name]['failed_instances'] = inst_uuid_list

    def set_ha(self, hv_name, status):
        """Set ha status to a hypervisor."""
        if status == ha_constants.STATE_OK:
            self.ha_hvs.pop(hv_name)
        else:
            if hv_name in self.ha_hvs:
                self.ha_hvs[hv_name].update(
                                    {'state': status,
                                     'updated_at': timeutils.strtime()})
            else:
                self.ha_hvs[hv_name] = {'state': status,
                                        'created_at': timeutils.strtime(),
                                        'updated_at': timeutils.strtime()}

    def set_maintenance(self, hv_name, status):
        """Set maintenance status to a hypervisor."""
        if status == ha_constants.STATE_OK:
            self.mt_hvs.pop(hv_name)
        else:
            if hv_name in self.mt_hvs:
                self.mt_hvs[hv_name].update(
                                    {'state': status,
                                     'updated_at': timeutils.strtime()})
            else:
                self.mt_hvs[hv_name] = {'state': status,
                                        'created_at': timeutils.strtime(),
                                        'updated_at': timeutils.strtime()}

    def set_migrate(self, hv_name, migrate):
        if hv_name in self.mt_hvs:
            self.mt_hvs[hv_name].update(
                                {'migrate': migrate,
                                 'updated_at': timeutils.strtime()})
        else:
            self.mt_hvs[hv_name] = {'migrate': migrate,
                                    'created_at': timeutils.strtime(),
                                    'updated_at': timeutils.strtime()}

    def set_ha_moving_insts(self, hv_name, inst_host):
        if hv_name in self.ha_hvs:
            self.ha_hvs[hv_name].update({'moving_insts': inst_host})

    def set_maintenance_moving_insts(self, hv_name, inst_host):
        if hv_name in self.mt_hvs:
            self.mt_hvs[hv_name].update({'moving_insts': inst_host})

    def set_mt_moving_inst_start(self, hv_name, inst_uuid, start):
        if hv_name in self.mt_hvs:
            if 'inst_start' in self.mt_hvs[hv_name]:
                self.mt_hvs[hv_name]['inst_start'].update({inst_uuid: start})
            else:
                self.mt_hvs[hv_name].update({'inst_start':
                                             {inst_uuid: start}})

    def set_ha_moving_inst_start(self, hv_name, inst_uuid, start):
        if hv_name in self.ha_hvs:
            if 'inst_start' in self.ha_hvs[hv_name]:
                self.ha_hvs[hv_name]['inst_start'].update({inst_uuid: start})
            else:
                self.ha_hvs[hv_name].update({'inst_start':
                                             {inst_uuid: start}})

    def set_target_host(self, hv_name, tgt_host):
        if hv_name in self.mt_hvs:
            self.mt_hvs[hv_name].update(
                                {'target_host': tgt_host,
                                 'updated_at': timeutils.strtime()})
        else:
            self.mt_hvs[hv_name] = {'target_host': tgt_host,
                                    'created_at': timeutils.strtime(),
                                    'updated_at': timeutils.strtime()}

    def get_target_host(self, hv_name):
        if hv_name not in self.mt_hvs:
            return None
        return self.mt_hvs[hv_name]['target_host']

    def refresh(self, all_nodes, ha_manage_hosts):
        """
        Refresh status of all hypervisors:
            a. remove non-existent hypervisors from MT
            b. remove non-existent and non-managed hypervisors from HA
        """
        curr_hvs = set([node['hypervisor_hostname'] for node in all_nodes])
        ha_hvs = set(self.ha_hvs.keys())
        mt_hosts = set(self.mt_hvs.keys())

        for key in mt_hosts - curr_hvs:
            self.mt_hvs.pop(key)

        for key in ha_hvs - ha_manage_hosts:
            self.ha_hvs.pop(key)

    def get_json(self):
        """Return full json object."""
        return {'version': VERSION,
                'ha': self.ha_hvs,
                'maintenance': self.mt_hvs}
