# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
import copy

import six

from nova.ibm.ego import ego_constants
from nova.ibm.ego import ego_gettextutils

from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_messaging.notify import notifier

_ = ego_gettextutils._
_LE = ego_gettextutils._LE

LOG = logging.getLogger(__name__)

notification_topic_opt = cfg.ListOpt(
    'ibm_notification_topics', default=['notifications', ],
    help='AMQP topic used for OpenStack PRS notifications')

CONF = cfg.CONF
CONF.register_opt(notification_topic_opt)


def _sanitize(message):
    """Sanitizes the message before logging."""
    SANITIZE = ['_context_auth_token', 'auth_token', 'new_pass']

    def _fix_passwords(d):
        """Sanitizes the password fields in the dictionary."""
        for k in six.iterkeys(d):
            if k.lower().find('password') != -1:
                d[k] = '<SANITIZED>'
            elif k.lower() in SANITIZE:
                d[k] = '<SANITIZED>'
            elif isinstance(d[k], dict):
                _fix_passwords(d[k])
        return d

    return _fix_passwords(copy.deepcopy(message))


class NotifierDispatcher(notifier.Driver):

    """
    Send notifications using the 1.0 message format.
    """

    def __init__(self, conf, topics, transport, version=1.0):
        super(NotifierDispatcher, self).__init__(conf, topics, transport)
        self.version = version
        self.topics = CONF.ibm_notification_topics

    def notify(self, ctxt, message, priority, retry):
        # Filter the message which does not contain
        # the event type we care about
        event_type = message.get('event_type', '')
        if event_type not in ego_constants.EVENT_TYPE_LIST:
            return

        priority = priority.lower()
        for topic in self.topics:
            target = messaging.Target(topic='%s.%s' % (topic, priority))
            try:
                self.transport._send_notification(target,
                                                  ctxt,
                                                  message,
                                                  version=self.version,
                                                  retry=retry)
                LOG.debug("Sent notification to notification service; "
                          "topic: %(topic)s, msg: %(msg)s" %
                            {'topic': topic, 'msg': _sanitize(message)})
            except Exception:
                LOG.exception(_LE("Failed to send notification; topic: "
                              "%(topic)s, msg: %(message)s"),
                              dict(topic=topic, message=_sanitize(message)))
