# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
import datetime
import re

from ceilometerclient import client as ceilometer_client

from nova.ibm.ego import ego_gettextutils

from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import sslutils
from oslo_utils import timeutils


CONF = cfg.CONF
CONF.register_opts(
    [cfg.IntOpt('prs_instance_cpu_ut_interval',
                default=60,
                help='The length of time (in minutes) over which the CPU '
                     'utilization of an instance is calculated')])
CONF.import_opt('username', 'nova.ibm.ego.ego_flags', 'ibm_prs_ceilometer')
CONF.import_opt('password', 'nova.ibm.ego.ego_flags', 'ibm_prs_ceilometer')
CONF.import_opt('project_name', 'nova.ibm.ego.ego_flags', 'ibm_prs_ceilometer')
CONF.import_opt('auth_url', 'nova.ibm.ego.ego_flags', 'ibm_prs_ceilometer')
CONF.import_opt('service_type', 'nova.ibm.ego.ego_flags', 'ibm_prs_ceilometer')
CONF.import_opt('insecure', 'nova.ibm.ego.ego_flags', 'ibm_prs_ceilometer')
CONF.import_opt('project_domain_name', 'nova.ibm.ego.ego_flags',
                'ibm_prs_ceilometer')
CONF.import_opt('user_domain_name', 'nova.ibm.ego.ego_flags',
                'ibm_prs_ceilometer')
sslutils.is_enabled(CONF)
CONF.import_opt('prs_policy_event_time_window_hours', 'nova.ibm.ego.ego_flags')

LOG = logging.getLogger(__name__)

_ = ego_gettextutils._


def get_ceilometer_client():
    """
    Return a Ceilometer client.
    """
    os_auth_url = CONF.ibm_prs_ceilometer.auth_url.strip()
    service_type = CONF.ibm_prs_ceilometer.service_type.strip()
    project_name = CONF.ibm_prs_ceilometer.project_name.strip()
    username = CONF.ibm_prs_ceilometer.username.strip()
    os_password = CONF.ibm_prs_ceilometer.password.strip()
    user_domain_name = CONF.ibm_prs_ceilometer.user_domain_name
    project_domain_name = CONF.ibm_prs_ceilometer.project_domain_name
    insecure = CONF.ibm_prs_ceilometer.insecure
    ca_file = CONF.ssl.ca_file
    cert_file = CONF.ssl.cert_file
    key_file = CONF.ssl.key_file

    auth_version = "v3" if re.search(r'/v3', os_auth_url) else "v2.0"
    args = {
        'os_auth_url': os_auth_url,
        'service_type': service_type,
        'os_tenant_name': project_name,
        'os_username': username,
        'os_password': os_password,
        'ca_file': ca_file,
        'cert_file': cert_file,
        'key_file': key_file,
        'insecure': insecure}
    if auth_version == "v3":
        args.update({
        'os_user_domain_name': user_domain_name,
        'os_project_domain_name': project_domain_name})

    cclient = ceilometer_client.get_client('2', **args)
    return cclient


class API(object):

    def get_instance_ut(self, instance, meter):
        """
        Get the instance utilization from Ceilometer.
        :param instance: Instance for which utilization will be returned
        :param meter: The meter for which the value will be returned

        :return: The instance's CPU utilization as computed as an average over
        the configured time window. See Ceilometer's cpu_util for details.
        """
        resource_id = instance['uuid']
        start_time = timeutils.parse_isotime(timeutils.strtime())
        start_time = start_time.replace(tzinfo=None)
        interval = CONF.prs_instance_cpu_ut_interval
        query_start = start_time - datetime.timedelta(minutes=interval)
        query = [{'field': 'resource_id', 'op': 'eq',
                  'value': resource_id},
                 {'field': 'timestamp',
                  'op': 'ge',
                  'value': query_start}]
        metrics = get_ceilometer_client().samples.list(meter, query)

        utilization_sum = 0.0
        count = 0
        for meter in metrics:
            utilization_sum += meter.counter_volume
            count += 1

        if count == 0:
            instance_utilization = 0
        else:
            instance_utilization = utilization_sum / count
        LOG.debug("The average %(meter)s value of instance %(name)s "
                  "(%(uuid)s) is %(utilization)s (%(num)d data sample(s))."
                  % {'meter': meter, 'name': instance['display_name'],
                     'uuid': instance['uuid'], 'num': count,
                     'utilization': instance_utilization})
        return instance_utilization

    def get_events(self, query=None, marker=None, limit=None):
        """
        Get events from Ceilometer.
        :param query: Filter arguments for which events to return (optional).
        :param marker: The events to be returned should be older than
                       the marker specified timestamp (optional).
        :param limit: Maximum number of events to return (optional).

        :return: events list.
        """
        event_list = []
        if query is None:
            query = []
        # There is a ceilometer bug, only one ceilometer query can return
        # correct results, so only pass 1 condition to ceilometer, other
        # condition filter will implement in PRS
        q_db = query[:1]
        q_prs = query[1:]
        if marker:
            q_db.append({'field': 'end_timestamp', 'op': 'eq',
                          'value': marker, 'type': 'datetime'})
        start_time = timeutils.parse_isotime(timeutils.strtime())
        start_time = start_time.replace(tzinfo=None)
        query_start = start_time - datetime.timedelta(
                             hours=CONF.prs_policy_event_time_window_hours)
        start_timestamp = timeutils.isotime(query_start)
        q_db.append({'field': 'start_timestamp', 'op': 'eq',
                      'value': start_timestamp, 'type': 'datetime'})

        # the events would be sorted from older to newer by default,
        # we should reverse it.
        # Because the ceilometer bug which is ceilometer not return latest
        # events from api, we not pass limit in the events.list
        # once community fix the issue, we will use ceilometer limit
        events = get_ceilometer_client().events.list(q_db)
        events.sort(key=lambda event: event.generated, reverse=True)
        for event in events:
            if marker and event.generated == marker:
                continue
            if (limit is not None) and (len(event_list) >= limit):
                break
            trait_list = [(trait['name'], trait['value'])
                                  for trait in event.traits]
            q_prs_list = [(q['field'], str(q['value'])) for q in q_prs]
            if (set(q_prs_list) - set(trait_list)):
                continue
            event_dict = {}
            event_dict['time'] = event.generated
            event_dict['event_type'] = event.event_type
            event_dict['description'] = dict(trait_list)
            event_list.append(event_dict)
        return event_list
