# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Resource Manager Service Engine"""

from nova.ibm.ego import ego_constants
from nova.ibm.ego.resource import constants
from nova.ibm.ego.resource import keystone_client
from nova.ibm.ego.resource import utils
from nova.scheduler.ibm.ego import ego_exception
from nova.scheduler.ibm.ego import ego_manager


class ResourceEngineManager(object):

    """A daemon that serves the Resource manager Service."""
    def __init__(self, *args, **kwargs):
        super(ResourceEngineManager, self).__init__(*args, **kwargs)
        self.ego_manager = ego_manager.EGOManager()
        self.keystone_client = keystone_client.KeystoneClient()

    def _get_resource_plan(self):
        return utils.parse_xml(
            self.ego_manager.ego_get_resource_plan(
                constants.MD_RESOURCE_PLAN_NAME,
                "md",
                constants.VEM_POLICY_FULL_PLAN,
                constants.EGO_WITHOUT_VERSION))

    def _get_consumer_tree(self):
        return utils.parse_xml(
            self.ego_manager.ego_get_consumer_tree(
                constants.VEM_GETCONSUMERTREE_AS_CONSUMER_ADMIN))

    def _update_resource_plan(self, res_plan):
        self.ego_manager.ego_set_resource_plan_md(
            constants.MD_RESOURCE_PLAN_NAME,
            res_plan.toxml(),
            constants.RESOURCE_PLAN_FLAG_UPDATE)

    def get_resource_plan_properties(self, context):
        return utils.get_properties(self._get_resource_plan())

    def set_resource_plan_properties(self, context, properties):
        res_plan = self._get_resource_plan()
        utils.set_properties(
            res_plan, properties)
        self._update_resource_plan(res_plan)

    def get_consumer_list(self, context):
        """
        Get EGO consumer list
        from EGO multi dimension resource plan.
        Return a list of EGO consumer list.
        Sample:
        [{"lendout_grace_period": "3", "rebalance": "true",
        "consumer_name": "224799a223594a19a9e9fa39eecc8674",
        "reclaim_grace_period": "3", "tenant_name": "service"},
        {"lendout_grace_period": "9", "rebalance": "true",
        "consumer_name": "d6a196966c984112a13eb5819559a3d1",
        "reclaim_grace_period": "9", "tenant_name": "admin"}]
        """
        part1 = utils.get_consumer_properties_from_consumer_tree(
            self._get_consumer_tree())

        part2 = utils.get_consumer_properties_from_md(
            self._get_resource_plan())

        # md only contains openstack tenants but
        # consumer tree contains more.
        for consumer, properties in part2.items():
            properties.update(part1.get(consumer, {}))

        # fill the tenant name.
        tenant_info_list = self.keystone_client.get_tenants()
        tenant_info_dict = {}
        for info in tenant_info_list:
            tenant_info_dict[info._info.get('id')] = info._info

        prop_list = []
        for id, prop in part2.items():
            prop[constants.ATTR_CONSUMER_NAME] = id
            prop[constants.ATTR_TENANT_NAME] = tenant_info_dict.get(
                id, {}).get(
                'name', constants.TAG_ORPHAN_CONSUMER + id)
            prop_list.append(prop)

        return prop_list

    def get_consumer_by_id(self, context, tenant_id):
        part1 = utils.get_consumer_properties_from_consumer_tree_by_consumer(
            self._get_consumer_tree(),
            tenant_id)
        part2 = utils.get_consumer_properties_from_md_by_consumer(
            self._get_resource_plan(),
            tenant_id)
        part1.update(part2)
        part1[constants.ATTR_CONSUMER_NAME] = tenant_id
        part1[constants.ATTR_TENANT_NAME] = (
            self.keystone_client.get_tenant_by_id(tenant_id))

        return part1

    def get_consumer_set(self):
        """
        Get OpenStack consumer list
        from EGO multi dimension resource plan.
        Return a SET of Openstack consumer list.
        """
        return set(utils.get_consumers(self._get_resource_plan()))

    def create_consumer(self, consumer):
        # add the consumer to consumer tree.
        self.ego_manager.ego_create_consumer(
            consumer, constants.VEM_EXEC_USER_EGO_ADMIN,
            '/', 'false', 0, ['Admin'],
            [constants.MD_RESOURCE_GROUP_NAME])

        # add the consumer element to the MDPlan.xml.
        self.ego_manager.ego_add_remove_consumer_for_md(
            '/' + consumer, consumer, constants.MD_RESOURCE_PLAN_NAME,
            constants.POLICY_CONSUMER_ADDONLY)

        # add the resource group attribute in
        # the MDPlan.xml in consumer element.
        res_plan = self._get_resource_plan()
        utils.set_resource_group_for_md_plan(
            res_plan,
            constants.MD_RESOURCE_GROUP_NAME,
            consumer)
        self._update_resource_plan(res_plan)

    def delete_consumer(self, consumer):
        try:
            # Delete the comsumer.
            # The delete consumer operation will
            # fail when there is any allocation
            # on the consumer. In this case we
            # still treat it a the operation succeeded.
            self.ego_manager.ego_delete_consumer('/' + consumer)

        except ego_exception.EGOExpDeleteConsumerFailed:
            pass

    def set_consumer(self, context, properties):
        for consumer, prop in properties.items():
            if constants.ATTR_LENDOUT in prop:
                res_plan = self._get_resource_plan()
                utils.set_lendout_grace_period(
                    res_plan,
                    consumer,
                    prop.get(constants.ATTR_LENDOUT, None))
                self._update_resource_plan(res_plan)
            if (constants.ATTR_REBALANCE in prop or
               constants.ATTR_RECLAIM in prop):
                self.ego_manager.ego_set_consumer(
                    '/' + consumer,
                    consumer,
                    prop.get(constants.ATTR_REBALANCE, 'true'),
                    int(prop.get(constants.ATTR_RECLAIM, '0') or '0'))

    def sync_up_ego_consumer(self):
        """
        Check if the OpenStack tenants is mapping with the ego consumers.
        If not, sync them.
        """
        # check EGO connection
        if self.ego_manager.ego_hdl.is_conn_lost():
            return

        tenants = set([info._info.get(
            "id", "") for info in self.keystone_client.get_tenants()])
        consumers = self.get_consumer_set()

        for consumer in tenants - consumers:
            self.create_consumer(consumer)

        for consumer in consumers - tenants:
            self.delete_consumer(consumer)

    def notify_resource_status(self, context, msg):
        """
        Do sync up task when accepting tenant delete/create notification.
        """
        event_type = msg['event_type']
        payload = msg['payload']
        if event_type not in (
            [ego_constants.EVENT_TYPE_IDENTITY_PROJECT_CREATED,
             ego_constants.EVENT_TYPE_IDENTITY_PROJECT_DELETED]):
            return
        tenant_id = payload['resource_info']
        consumers = self.get_consumer_set()
        if event_type == ego_constants.EVENT_TYPE_IDENTITY_PROJECT_CREATED:
            if tenant_id not in consumers:
                self.create_consumer(tenant_id)
        else:
            if tenant_id in consumers:
                self.delete_consumer(tenant_id)

    def get_plan_details(self, context):
        return utils.get_plan_details(
            self._get_resource_plan(), self.keystone_client)

    def get_plan_detail_by_consumer(self, context, consumer):
        return utils.get_plan_detail_by_consumer(
            self._get_resource_plan(),
            self.keystone_client,
            consumer)

    def set_plan_details(self, context, details):
        if details:
            res_plan = self._get_resource_plan()
            utils.set_plan_details(
                res_plan, details)
            self._update_resource_plan(res_plan)

    def get_services_info(self, context):
        hosts, total_resource = utils.get_host_info(self._get_resource_plan())
        instances = self.ego_manager.get_service_instance_info(
            constants.MD_RESOURCE_PLAN_NAME)

        stat = {}

        # Calculate the usage info for each host.
        used_cpu = used_mem = total_used_cpu = total_used_mem = inst_num = 0
        for inst in instances:
            used_cpu = stat.get(inst['host_name'], {}).get(
                'used_cpu', 0) + inst.get('used_cpu', 0)
            used_mem = stat.get(inst['host_name'], {}).get(
                'used_mem', 0) + inst.get('used_mem', 0)
            inst_num = stat.get(inst['host_name'], {}).get(
                'inst_num', 0) + 1
            total_used_cpu += inst.get('used_cpu', 0)
            total_used_mem += inst.get('used_mem', 0)
            stat[inst['host_name']] = {'used_cpu': used_cpu,
                                       'used_mem': used_mem,
                                       'inst_num': inst_num}

        # update the host info.
        for host in hosts:
            host.update(
                stat.get(host['host_name'], {'used_cpu': 0,
                         'used_mem': 0, 'inst_num': 0}))

        return {'host_info': hosts,
                'instance_info': instances,
                'usage': {'total_used_cpu': total_used_cpu,
                          'total_used_mem': total_used_mem,
                          'total_cpu': total_resource.get('ncpus', 0),
                          'total_mem': total_resource.get('maxmem', 0)}}
