# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""
Client side of the resource RPC API.
"""
import oslo_messaging as messaging

from nova.objects import base as objects_base
from nova import rpc

from oslo_config import cfg

resourceapi_opts = [
    cfg.StrOpt('resource_topic',
               default='ego_resource',
               help='The topic on which the resource service listens'),
]

CONF = cfg.CONF
CONF.register_opts(resourceapi_opts)

rpcapi_cap_opt = cfg.StrOpt('resource',
            help='Set a version cap for messages sent to resource service')
CONF.register_opt(rpcapi_cap_opt, 'upgrade_levels')


class ResourceAPI(object):
    """Client side of the resource RPC API.

    API version history:

        1.0 - Initial version.

    """
    VERSION_ALIASES = {
        'havana': '1.0'
    }

    def __init__(self, topic=None):
        super(ResourceAPI, self).__init__()
        target = messaging.Target(topic=CONF.resource_topic, version="1.0")
        version_cap = self.VERSION_ALIASES.get(CONF.upgrade_levels.resource,
                                               CONF.upgrade_levels.resource)
        serializer = objects_base.NovaObjectSerializer()
        self.client = rpc.get_client(target,
                                     version_cap=version_cap,
                                     serializer=serializer)

    def get_resource_plan_properties(self, context):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'ego_get_resource_plan_properties')

    def set_resource_plan_properties(self, context, properties):
        cctxt = self.client.prepare()
        return cctxt.call(
            context,
            'ego_set_resource_plan_properties',
            properties=properties)

    def get_consumer_list(self, context):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'ego_get_consumer_list')

    def get_consumer_by_id(self, context, tenant_id):
        cctxt = self.client.prepare()
        return cctxt.call(
            context, 'ego_get_consumer_by_id', tenant_id=tenant_id)

    def create_consumer(self, context, tenant_id):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'ego_create_consumer', tenant_id=tenant_id)

    def delete_consumer(self, context, tenant_id):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'ego_delete_consumer', tenant_id=tenant_id)

    def set_consumer(self, context, properties):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'ego_set_consumer', properties=properties)

    def get_plan_details(self, context):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'ego_get_plan_details')

    def get_plan_detail_by_consumer(self, context, tenant_id):
        cctxt = self.client.prepare()
        return cctxt.call(
            context, 'ego_get_plan_detail_by_consumer', tenant_id=tenant_id)

    def set_plan_details(self, context, details):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'ego_set_plan_details', details=details)

    def get_services_info(self, context):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'get_services_info')
