# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from xml.dom import minidom

from nova.ibm.ego.resource import constants


def parse_xml(xml_str):
    return minidom.parseString(xml_str)


def get_properties(res_plan):
    properties = {}

    for node in res_plan.getElementsByTagName('ConsumableResource'):
        child = node._get_firstChild()
        if (node.TEXT_NODE == child.nodeType and 'ncpus' == child.data):
            properties['cpu_overcommit_ratio'] = node.getAttribute(
                'OverCommitRatio')
        elif (node.TEXT_NODE == child.nodeType and 'maxmem' == child.data):
            properties['memory_overcommit_ratio'] = node.getAttribute(
                'OverCommitRatio')

    for node in res_plan.getElementsByTagName('PolicyParameter'):
        if 'DefaultGracePeriod' == node.getAttribute('ParameterName'):
            child = node._get_firstChild()
            if child and node.TEXT_NODE == child.nodeType:
                properties['share_pool_grace_period'] = child.data

    return properties


def set_properties(res_plan, properties):
    for node in res_plan.getElementsByTagName('ConsumableResource'):
        child = node._get_firstChild()
        if (child and node.TEXT_NODE == child.nodeType and
           'ncpus' == child.data and'cpu_overcommit_ratio' in properties):
            node.setAttribute(
                'OverCommitRatio', properties['cpu_overcommit_ratio'])
        elif (child and node.TEXT_NODE == child.nodeType and
              'maxmem' == child.data and
              'memory_overcommit_ratio' in properties):
            node.setAttribute(
                'OverCommitRatio', properties['memory_overcommit_ratio'])

    for node in res_plan.getElementsByTagName('PolicyParameter'):
        if 'DefaultGracePeriod' == node.getAttribute('ParameterName'):
            child = node._get_firstChild()
            if child and node.TEXT_NODE == child.nodeType:
                child.data = properties['share_pool_grace_period']


def get_consumers(res_plan):
    consumers = []
    for node in res_plan.getElementsByTagName('Consumer'):
        consumer = node.getAttribute('ConsumerName')
        if consumer:
            consumers.append(consumer)

    return consumers


def _get_ownership_type(res_plan):
    for ownership_policy in res_plan.getElementsByTagName('OwnershipPolicy'):
        resource = ownership_policy.getElementsByTagName('Resource')
        if len(resource) > 0:
            return resource[0].getAttribute('Type')

    return constants.ATTR_OWNERSHIP_TYPE_ABUSOLUTE


def _construct_absolute_ownership_nodes(
        res_plan, owned_host, reserved_host, lendout_grace_period):
    owned_hosts = res_plan.createElement('NumHosts')
    owned_hosts.setAttribute('Type', constants.ATTR_OWNERSHIP_TYPE_ABUSOLUTE)
    owned_hosts.appendChild(res_plan.createTextNode(
        str(owned_host)))
    host_selection = res_plan.createElement(
        'HostSelection')
    host_selection.setAttribute('Type', 'NumHosts')
    host_selection.appendChild(owned_hosts)
    lendout = res_plan.createElement('Lendout')
    if lendout_grace_period is not None:
        lendout.setAttribute('GracePeriod', lendout_grace_period)
    if reserved_host is not None:
        reserve_hosts = res_plan.createElement('NumHosts')
        reserve_hosts.setAttribute(
            'Type',
            constants.ATTR_OWNERSHIP_TYPE_ABUSOLUTE)
        reserve_hosts.appendChild(res_plan.createTextNode(
            str(reserved_host)))
        reserve = res_plan.createElement('Reserve')
        reserve.appendChild(reserve_hosts)
        lendout.appendChild(reserve)
    return (host_selection, lendout)


def _construct_dimension(res_plan, cpu, mem, flag):
    if cpu is None and mem is None:
        return None
    dimension = res_plan.createElement('Dimension')
    if cpu is not None:
        cpu_node = res_plan.createElement('Resource')
        cpu_node.setAttribute('Name', 'ncpus')
        cpu_node.setAttribute('Type', flag)
        cpu_node.appendChild(res_plan.createTextNode(str(float(cpu))))
        dimension.appendChild(cpu_node)
    if mem is not None:
        mem_node = res_plan.createElement('Resource')
        mem_node.setAttribute('Name', 'maxmem')
        mem_node.setAttribute('Type', flag)
        mem_node.appendChild(res_plan.createTextNode(str(float(mem))))
        dimension.appendChild(mem_node)
    return dimension


def _construct_percentage_ownership_node(
        res_plan, cpu, mem, reserved_cpu, reserved_mem, lendout_grace_period):
    perhost = res_plan.createElement('PerHost')
    perhost.setAttribute('Type', 'dimension')
    perhost.appendChild(_construct_dimension(
        res_plan, cpu, mem, constants.ATTR_OWNERSHIP_TYPE_PERCENTAGE))
    lendout = res_plan.createElement('Lendout')
    if lendout_grace_period is not None:
        lendout.setAttribute('GracePeriod', lendout_grace_period)
    if reserved_cpu is not None or reserved_mem is not None:
        reserve_perhost = res_plan.createElement('PerHost')
        reserve_perhost.setAttribute('Type', 'dimension')
        reserve_perhost.appendChild(_construct_dimension(
            res_plan, reserved_cpu or 0, reserved_mem or 0,
            constants.ATTR_OWNERSHIP_TYPE_PERCENTAGE))
        reserve = res_plan.createElement('Reserve')
        reserve.appendChild(reserve_perhost)
        lendout.appendChild(reserve)
    return (perhost, lendout)


def set_resource_group_for_md_plan(res_plan, res_group, consumer):
    resource_group = res_plan.createElement('ResourceGroup')
    resource_group.setAttribute('Name', res_group)
    resource_group.setAttribute('PreferenceLevel', '5')
    resource_group.setAttribute('GetFreeFromNextLevelBeforeReclaim', 'Y')
    if constants.ATTR_OWNERSHIP_TYPE_ABUSOLUTE == _get_ownership_type(
            res_plan):
        childs = _construct_absolute_ownership_nodes(res_plan, 0, None, None)
    else:
        childs = _construct_percentage_ownership_node(
            res_plan, 0, 0, None, None, None)

    for child in childs:
        resource_group.appendChild(child)

    for node in res_plan.getElementsByTagName('Consumer'):
        consumer_name = node.getAttribute('ConsumerName')
        if consumer_name == consumer:
            node.getElementsByTagName(
                'OwnershipPolicy')[0].appendChild(resource_group)
            break


def set_lendout_grace_period(res_plan, consumer, lendout):
    for node in res_plan.getElementsByTagName('Consumer'):
        if node.getAttribute('ConsumerName') == consumer:
            ownership_policy = node.getElementsByTagName('OwnershipPolicy')[0]
            lendout_nodes = ownership_policy.getElementsByTagName('Lendout')
            if len(lendout_nodes) == 0:
                break
            lendout_node = lendout_nodes[0]
            if lendout:
                lendout_node.setAttribute('GracePeriod', lendout)
            else:
                lendout_node.removeAttribute('GracePeriod')
            break


def _get_lendout(node):
    prop = {constants.ATTR_LENDOUT: None}
    lendout = node.getElementsByTagName('Lendout')
    if len(lendout) > 0:
        prop[constants.ATTR_LENDOUT] = lendout[0].getAttribute(
            'GracePeriod')
    return prop


def get_consumer_properties_from_md_by_consumer(res_plan, consumer):
    for node in res_plan.getElementsByTagName('Consumer'):
        if node.getAttribute('ConsumerName') == consumer:
            return _get_lendout(node)


def get_consumer_properties_from_md(res_plan):
    properties = {}
    for node in res_plan.getElementsByTagName('Consumer'):
        properties[node.getAttribute('ConsumerName')] = _get_lendout(node)
    return properties


def _get_reclaim_reblance(node):
    properties = {constants.ATTR_RECLAIM: None,
                  constants.ATTR_REBALANCE: 'true'}
    reclaim = node.getElementsByTagName('ReclamationTimeout')
    if len(reclaim) > 0:
        child = reclaim[0]._get_firstChild()
        properties[constants.ATTR_RECLAIM] = child.data

    rebalance = node.getElementsByTagName('Rebalance')
    if len(rebalance) > 0:
        child = rebalance[0]._get_firstChild()
        properties[constants.ATTR_REBALANCE] = child.data

    return properties


def get_consumer_properties_from_consumer_tree(tree):
    properties = {}
    for node in tree.getElementsByTagName('Consumer'):
        properties[node.getAttribute(
            'ConsumerName')] = _get_reclaim_reblance(node)
    return properties


def get_consumer_properties_from_consumer_tree_by_consumer(tree, consumer):
    for node in tree.getElementsByTagName('Consumer'):
        if node.getAttribute('ConsumerName') == consumer:
            return _get_reclaim_reblance(node)


def _get_share_plan(node, keystone_client):
    share_plan = {
        constants.ATTR_CPU: None,
        constants.ATTR_MEM: None
    }
    consumer = node.getAttribute('ConsumerName')
    share_plan[constants.ATTR_CONSUMER_NAME] = consumer
    share_plan[constants.ATTR_TENANT_NAME] = (keystone_client.
                                              get_tenant_by_id(consumer))
    ratio = node.getElementsByTagName('Shares')
    share_plan[constants.ATTR_SHARE_RATIO] = ratio[0]._get_firstChild().data

    share = node.getElementsByTagName('SharingPolicy')
    share_plan[constants.ATTR_PRIORITY_GROUP] = (share[
        0].getAttribute('Priority'))
    for resouce in share[0].getElementsByTagName('Resource'):
        if 'ncpus' == resouce.getAttribute('Name'):
            share_plan[constants.ATTR_CPU] = resouce._get_firstChild().data
        elif 'maxmem' == resouce.getAttribute('Name'):
            share_plan[constants.ATTR_MEM] = resouce._get_firstChild().data

    return share_plan


def get_share_plans(res_plan, keystone_client):
    share_plans = []
    for node in res_plan.getElementsByTagName('Consumer'):
        share_plans.append(_get_share_plan(node, keystone_client))

    return share_plans


def get_share_plan_by_consumer(res_plan, keystone_client, consumer):
    for node in res_plan.getElementsByTagName('Consumer'):
        if node.getAttribute('ConsumerName') == consumer:
            return _get_share_plan(node, keystone_client)


def set_share_plans(res_plan, share_plans):
    # change the share_plan from list to dict to
    # simplify the following process.
    share_plan_dict = {}
    for plan in share_plans:
        share_plan_dict[plan[constants.ATTR_CONSUMER_NAME]] = plan

    for node in res_plan.getElementsByTagName('Consumer'):
        consumer = node.getAttribute('ConsumerName')
        if consumer not in share_plan_dict:
            continue

        detail = share_plan_dict[consumer]
        share_plan = node.getElementsByTagName('SharingPolicy')
        share_plan[0].setAttribute(
            'Priority', detail[constants.ATTR_PRIORITY_GROUP])
        share_ratio = share_plan[0].getElementsByTagName('Shares')[0]
        share_ratio._get_firstChild().data = detail[constants.ATTR_SHARE_RATIO]

        limit_old = (share_plan[0].getElementsByTagName('Limit') or [None])[0]
        limit_new = None
        dimension = _construct_dimension(
            res_plan,
            detail[constants.ATTR_CPU],
            detail[constants.ATTR_MEM],
            constants.ATTR_OWNERSHIP_TYPE_ABUSOLUTE)
        if dimension:
            limit_new = res_plan.createElement('Limit')
            limit_new.setAttribute('Type', 'dimension')
            limit_new.appendChild(dimension)

        if not limit_old:
            if limit_new:
                share_plan[0].appendChild(limit_new)
        else:
            if limit_new:
                share_plan[0].replaceChild(limit_new, limit_old)
            else:
                share_plan[0].removeChild(limit_old)


def _get_ownership_plan(node, keystone_client):
    ownership_plan = {}
    consumer = node.getAttribute('ConsumerName')
    ownership_plan[constants.ATTR_CONSUMER_NAME] = consumer
    ownership_plan[constants.ATTR_TENANT_NAME] = (keystone_client.
                                                  get_tenant_by_id(consumer))
    rank = node.getElementsByTagName('Rank')
    ownership_plan[constants.ATTR_RANK] = rank[0]._get_firstChild().data

    host_selection = node.getElementsByTagName('HostSelection')
    if len(host_selection) > 0:
        ownership_plan[constants.ATTR_OWNERSHIP_TYPE] = (
            constants.ATTR_OWNERSHIP_TYPE_ABUSOLUTE)
        numhosts = host_selection[0].getElementsByTagName('NumHosts')
        ownership_plan[constants.ATTR_OWNED_HOST] = (
            numhosts[0]._get_firstChild().data)
        reserved = node.getElementsByTagName('Reserve')
        if len(reserved) > 0:
            numhosts = reserved[0].getElementsByTagName('NumHosts')
            ownership_plan[constants.ATTR_RESERVED_HOST] = (
                numhosts[0]._get_firstChild().data)
        else:
            ownership_plan[constants.ATTR_RESERVED_HOST] = None
    else:
        ownership_plan[constants.ATTR_OWNERSHIP_TYPE] = (
            constants.ATTR_OWNERSHIP_TYPE_PERCENTAGE)
        ownership_plan[constants.ATTR_CPU] = None
        ownership_plan[constants.ATTR_MEM] = None
        ownership_plan[constants.ATTR_RESERVED_CPU] = None
        ownership_plan[constants.ATTR_RESERVED_MEM] = None
        for perhost in node.getElementsByTagName('PerHost'):
            if 'ResourceGroup' == perhost.parentNode.nodeName:
                for resouce in perhost.getElementsByTagName('Resource'):
                    if 'ncpus' == resouce.getAttribute('Name'):
                        ownership_plan[constants.ATTR_CPU] = (
                            resouce._get_firstChild().data)
                    elif 'maxmem' == resouce.getAttribute('Name'):
                        ownership_plan[constants.ATTR_MEM] = (
                            resouce._get_firstChild().data)
            else:
                for resouce in perhost.getElementsByTagName('Resource'):
                    if 'ncpus' == resouce.getAttribute('Name'):
                        ownership_plan[constants.ATTR_RESERVED_CPU] = (
                            resouce._get_firstChild().data)
                    elif 'maxmem' == resouce.getAttribute('Name'):
                        ownership_plan[constants.ATTR_RESERVED_MEM] = (
                            resouce._get_firstChild().data)

    return ownership_plan


def get_ownership_plans(res_plan, keystone_client):
    ownership_plans = []
    for node in res_plan.getElementsByTagName('Consumer'):
        ownership_plans.append(_get_ownership_plan(node, keystone_client))

    return ownership_plans


def get_ownership_plan_by_consumer(res_plan, keystone_client, consumer):
    for node in res_plan.getElementsByTagName('Consumer'):
        if node.getAttribute('ConsumerName') == consumer:
            return _get_ownership_plan(node, keystone_client)


def set_ownership_plans(res_plan, ownership_plans):
    # change the share_plan from list to dict to
    # simplify the following process.
    ownership_plan_dict = {}
    for plan in ownership_plans:
        ownership_plan_dict[plan[constants.ATTR_CONSUMER_NAME]] = plan

    local_ownership_type = _get_ownership_type(res_plan)
    ownership_type = ownership_plans[0][constants.ATTR_OWNERSHIP_TYPE]
    ownership_type_changed = False
    if local_ownership_type != ownership_type:
        ownership_type_changed = True

    for node in res_plan.getElementsByTagName('Consumer'):
        consumer = node.getAttribute('ConsumerName')
        if consumer in ownership_plan_dict or ownership_type_changed:
            detail = ownership_plan_dict.get(consumer, {})
            ownership_policy = node.getElementsByTagName('OwnershipPolicy')[0]
            resource_group = ownership_policy.getElementsByTagName(
                'ResourceGroup')[0]

            lendout_grace_period = None
            childs = resource_group.childNodes
            while childs.length > 0:
                child = childs.pop()
                if 'Lendout' == child.nodeName:
                    lendout_grace_period = child.getAttribute(
                        'GracePeriod') or None

            if constants.ATTR_OWNERSHIP_TYPE_ABUSOLUTE == ownership_type:
                if (constants.ATTR_RANK in detail and
                   detail[constants.ATTR_RANK] is not None):
                    rank = node.getElementsByTagName('Rank')[0]
                    rank._get_firstChild().data = detail[constants.ATTR_RANK]

                childs = _construct_absolute_ownership_nodes(
                    res_plan,
                    detail.get(constants.ATTR_OWNED_HOST, 0) or 0,
                    detail.get(constants.ATTR_RESERVED_HOST, None),
                    lendout_grace_period)
            else:
                childs = _construct_percentage_ownership_node(
                    res_plan,
                    detail.get(constants.ATTR_CPU, 0) or 0,
                    detail.get(constants.ATTR_MEM, 0) or 0,
                    detail.get(constants.ATTR_RESERVED_CPU, None),
                    detail.get(constants.ATTR_RESERVED_MEM, None),
                    lendout_grace_period)

            for child in childs:
                resource_group.appendChild(child)


def get_plan_details(res_plan, keystone_client):
    return {
        'share': get_share_plans(res_plan, keystone_client),
        'ownership': get_ownership_plans(res_plan, keystone_client)
    }


def get_plan_detail_by_consumer(res_plan, keystone_client, consumer):
    return {
        'share': get_share_plan_by_consumer(
            res_plan, keystone_client, consumer),
        'ownership': get_ownership_plan_by_consumer(
            res_plan, keystone_client, consumer)
    }


def set_plan_details(res_plan, plan_details):
    if plan_details.get('share', None):
        set_share_plans(res_plan, plan_details['share'])
    if plan_details.get('ownership', None):
        set_ownership_plans(res_plan, plan_details['ownership'])


def get_host_info(res_plan):
    host_info = []
    resource_group = res_plan.getElementsByTagName(
        'OwnedInConf')[0].getElementsByTagName('ResourceGroup')[0]
    for child in resource_group.childNodes:
        if child.nodeName == 'HostList':
            for host in child.getElementsByTagName('Host'):
                host_dict = {}
                host_dict['host_name'] = host.getAttribute('Name')
                for resource in host.getElementsByTagName('Resource'):
                    host_dict[resource.getAttribute(
                        'Name')] = resource._get_firstChild().data
                host_info.append(host_dict)
        elif child.nodeName == 'Dimension':
            total_resource = {}
            for resource in child.getElementsByTagName('Resource'):
                    total_resource[resource.getAttribute(
                        'Name')] = resource._get_firstChild().data
    return (host_info, total_resource)
