# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Policy Engine Manager."""

import codecs
import copy
import os
import six

from xml.dom import minidom

from nova import db
from nova import exception
from nova.ibm.ego import ego_constants
from nova.ibm.ego import ego_gettextutils
from nova.ibm.ego import ego_utils
from nova.ibm.ego.hapolicy import constants as ha_constants
from nova.ibm.ego import policy_ceilometer_client
from nova.ibm.ego.rtpolicy import policy_exception
from nova.ibm.ego.rtpolicy import rtpolicy_resource_group_status
from nova import objects
from nova import rpc
from nova.scheduler.ibm.ego import ego_exception
from nova.scheduler.ibm.ego import ego_host_aggregate
from nova.scheduler.ibm.ego import ego_manager
from nova.scheduler.ibm.ego import ego_resreq_utils
from nova import servicegroup

from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils

_ = ego_gettextutils._
_LI = ego_gettextutils._LI
_LW = ego_gettextutils._LW
_LE = ego_gettextutils._LE

policy_engine_manager_opts = [
    cfg.StrOpt('rtpolicy_driver',
               default='nova.ibm.ego.rtpolicy.rtpolicy_driver.RTPolicyDriver',
               help='The default driver for resource optimization extensions'),
    cfg.ListOpt('rtpolicy_optimizers',
                default=[
                    'nova.ibm.ego.rtpolicy.rtpolicy_optimizer.'
                    'RTPolicyMigrationOptimizer'
                ],
                help='Which optimizers will be used for resource optimization')
]

CONF = cfg.CONF
CONF.register_opts(policy_engine_manager_opts)
CONF.import_opt('disk_allocation_ratio', 'nova.scheduler.filters.disk_filter')
CONF.import_opt('scheduler_default_filters', 'nova.scheduler.host_manager')
CONF.import_opt('policy_per_host_aggregate', 'nova.ibm.ego.ego_flags')

DEFAULT_PARALLEL_MIGRATE = 5
PARALLEL_MIGRATE = 'prs:max_parallel_migrations'

LOG = logging.getLogger(__name__)


class PolicyEngineManager(object):
    """A daemon that serves the runtime optimization policy."""

    def __init__(self, rtpolicy_conf_path):
        self.all_policies = {}
        self.enabled_policy = {}
        # Format: {aggregate-id: RTPolicyResourceGroupStatus}
        self.aggregate_statuses = {}
        self.ego_api = ego_manager.EGOManager()
        self.ceilometer_api = policy_ceilometer_client.API()
        self.servicegroup_api = servicegroup.API()
        self.rtpolicy_driver = importutils.import_object(CONF.rtpolicy_driver)
        self.aggregate_container = ego_host_aggregate.EGOHostAggregate(
            policy_actions=self.rtpolicy_driver.get_policy_actions())
        self.rtpolicy_conf = rtpolicy_conf_path
        self.aggid_policy_dict = {}
        self.notifier = rpc.get_notifier('rtpolicy')
        self.need_set_resource_plan = True
        # Format: {ego_host:
        #          {'svc_host': svc_host, 'node_name': hypervisor_hostname}}
        self.ego_host_dict = {}
        self.optimizers = []

        # Instantiate the optimizer objects
        for optimizer in CONF.rtpolicy_optimizers:
            optimizer_obj = importutils.import_object(
                optimizer, policy_engine_manager=self,
                rtpolicy_driver=self.rtpolicy_driver)
            self.optimizers.append(optimizer_obj)
        # Sort the optimizers (lowest number --> highest priority)
        self.optimizers = sorted(self.optimizers,
                                 key=lambda optimizer: optimizer.priority)

        self.__policy_init()

    #################################################################
    #                 Begin Policy Management                       #
    #################################################################

    def __policy_init(self):
        """Perform initialization."""
        if not os.path.isfile(self.rtpolicy_conf):
            raise policy_exception.PlacementPolciyConfFileNotFoundPath(
                rtpolicy_conf=self.rtpolicy_conf)

        LOG.debug('The policy configuration file path is: %s'
                  % self.rtpolicy_conf)
        rtpolicies = self.get_all()
        for rtpolicy in rtpolicies:
            self.all_policies[rtpolicy['id']] = rtpolicy

        if not CONF.policy_per_host_aggregate:
            for id in self.all_policies.keys():
                if self.all_policies[id]['state'] == 'enabled':
                    self.enabled_policy = self.all_policies[id]
                    break
            if self.enabled_policy.get('goal'):
                try:
                    self.__set_ego_resource_plan(
                        self.enabled_policy.get('goal'))
                except (ego_exception.EGOExpSetResourcePlanFailed,
                        ego_exception.EGOExpConnLost) as exc:
                    LOG.error(_LE('The policy engine manager could not set '
                                'EGO resource plan for initializing '
                                'policy %(policy_id)s due to: %(reason)s')
                              % {'policy_id': self.enabled_policy.get('id'),
                                 'reason': six.text_type(exc)})
                LOG.debug("The enabled global policy is: %s"
                          % self.enabled_policy)
        else:
            # Set host aggregate based on initial placement policy
            self.aggregate_container.sync_up_aggregate_policy(
                set(self.all_policies.keys()))
            self.__update_aggregate_and_policy_info()
            self.__set_resource_plan_per_aggregate()
            LOG.debug("Host aggregate based policies have been enabled "
                      "during runtime service initialization.")

    def __set_resource_plan_per_aggregate(self):
        self.aggid_policy_dict = copy.deepcopy(
            self.aggregate_container.aggregate_id_policy_dict)
        try:
            # Set resource plan based on self.aggid_policy_dict
            self.__set_ego_resource_plan()
            self.need_set_resource_plan = False
        except (ego_exception.EGOExpSetResourcePlanFailed,
                ego_exception.EGOExpConnLost) as exc:
            LOG.error(_LE('The policy engine manager could not set '
                        'the EGO resource plan due to: %(reason)s')
                      % {'reason': six.text_type(exc)})

    def __validate_policy_parameters(self, values_dict, policy_id=''):
        def _validate_within_bounds(values_dict, attr, max_value, min_value):
            if attr not in values_dict:
                return
            try:
                value = values_dict.get(attr)
                if not (min_value <= int(value) <= max_value):
                    raise ValueError
            except ValueError:
                raise policy_exception.InvalidPlacementPolicyInput(
                    value=value, attribute=attr, min=min_value, max=max_value)

        bounded_parameters = [ego_constants.POLICY_RUN_INTERVAL,
                              ego_constants.POLICY_STABILIZATION,
                              ego_constants.POLICY_MAX_PARALLEL,
                              ego_constants.POLICY_THRESHOLD]
        for param in bounded_parameters:
            _validate_within_bounds(values_dict,
                                    param,
                                    ego_constants.MAX_INT_32,
                                    ego_constants.MIN_POSITIVE_INT)

        if ego_constants.POLICY_GOAL in values_dict:
            goal_val = values_dict.get(ego_constants.POLICY_GOAL)
            if goal_val == '' or goal_val.isspace():
                raise policy_exception.InvalidPlacementPolicyGoal(
                    rtpolicy_id=policy_id)
        if ego_constants.POLICY_STATE in values_dict:
            state_value = values_dict.get(ego_constants.POLICY_STATE)
            if state_value not in ['enabled', 'disabled']:
                raise policy_exception.InvalidPlacementPolicyState(
                    rtpolicy_id=policy_id, state=state_value)
        if ego_constants.POLICY_ACTION in values_dict:
            action_value = values_dict.get(ego_constants.POLICY_ACTION)
            if action_value not in self.rtpolicy_driver.get_policy_actions():
                raise policy_exception.InvalidPlacementPolicyAction(
                    policy_action=ego_constants.POLICY_ACTION)

    def __get_node_value(self, node):
        return node.childNodes[0].nodeValue if node.hasChildNodes() else ''

    def __set_node_value(self, doc, node, value):
        if node[0].hasChildNodes():
            node[0].childNodes[0].nodeValue = value
        else:
            text_node = doc.createTextNode(value)
            node[0].appendChild(text_node)

    def __get_nodes(self, node, name):
        return node.getElementsByTagName(name) if node else []

    def __updatexml(self, doc):
        with codecs.open(self.rtpolicy_conf, 'w', 'utf-8') as f:
            doc.writexml(f, addindent='', encoding='utf-8')

    def __get_host_aggregates_by_policy_id_and_type(self, policy_id,
                                                    policy_type):
        """
        Get filtered metadata of aggregates directly for API request.  The flag
        'is_valid' will denote whether the parameters are valid or not.
        """
        host_aggregates = []
        filtered_meta = self.aggregate_container.aggid_policy_no_check
        for agg_id, policies in filtered_meta.items():
            policy = copy.deepcopy(policies.get(policy_type, {}))
            temp_id = policy.get(ego_constants.POLICY_ID)
            if policy and temp_id == str(policy_id):
                policy[ego_constants.POLICY_ID] = agg_id
                tmp = self.aggid_policy_dict.get(agg_id, {})
                pid = tmp.get(policy_type, {}).get(ego_constants.POLICY_ID)
                if pid and pid == policy_id:
                    policy[ego_constants.POLICY_VALID_FLAG] = True
                else:
                    policy[ego_constants.POLICY_VALID_FLAG] = False
                host_aggregates.append(policy)
        if not self.enabled_policy and host_aggregates:
            return {'host_aggregates': host_aggregates}

    def __make_rtpolicy(self, node):
        rtpolicy_attribute_show = [
            (ego_constants.POLICY_ID, ego_constants.ATTRIBUTE_TYPE_INT),
            (ego_constants.POLICY_NAME, ego_constants.ATTRIBUTE_TYPE_STR),
            (ego_constants.POLICY_DESCRIPTION,
             ego_constants.ATTRIBUTE_TYPE_STR),
            (ego_constants.POLICY_GOAL, ego_constants.ATTRIBUTE_TYPE_STR)
        ]
        # Global policy mode needs to retrieve all policy parameters
        if not CONF.policy_per_host_aggregate:
            rtpolicy_attribute_show += [
                (ego_constants.POLICY_STATE,
                 ego_constants.ATTRIBUTE_TYPE_STR),
                (ego_constants.POLICY_ACTION,
                 ego_constants.ATTRIBUTE_TYPE_STR),
                (ego_constants.POLICY_THRESHOLD,
                 ego_constants.ATTRIBUTE_TYPE_STR),
                (ego_constants.POLICY_STABILIZATION,
                 ego_constants.ATTRIBUTE_TYPE_STR),
                (ego_constants.POLICY_RUN_INTERVAL,
                 ego_constants.ATTRIBUTE_TYPE_STR),
                (ego_constants.POLICY_MAX_PARALLEL,
                 ego_constants.ATTRIBUTE_TYPE_STR)
            ]

        obj = {}
        policy_id = 0
        for attribute in rtpolicy_attribute_show:
            node_obj = self.__get_nodes(node, attribute[0])
            if not node_obj:
                continue
            text_value = self.__get_node_value(node_obj[0])
            if attribute[1] == ego_constants.ATTRIBUTE_TYPE_STR:
                obj[attribute[0]] = text_value
            elif attribute[1] == ego_constants.ATTRIBUTE_TYPE_INT:
                value = int(text_value)
                obj[attribute[0]] = value
            else:
                raise exception.InvalidInput(_('The policy engine manager '
                                               'does not support %(type)s '
                                               'for the %(name)s policy '
                                               'configuration parameter.')
                                             % {'type': attribute[1],
                                                'name': attribute[0]})
            if 'id' == attribute[0]:
                policy_id = obj[attribute[0]]
            if not CONF.policy_per_host_aggregate:
                self.__validate_policy_parameters(obj, obj['id'])
        obj[ego_constants.POLICY_KEY_HOST_AGGREGATE_INITIAL] = (
            self.__get_host_aggregates_by_policy_id_and_type(
                policy_id, ego_constants.INITIAL_POLICY))
        obj[ego_constants.POLICY_KEY_HOST_AGGREGATE_RUNTIME] = (
            self.__get_host_aggregates_by_policy_id_and_type(
                policy_id, ego_constants.RUNTIME_POLICY))
        return obj

    def __get_rtpolicies(self, doc):
        return self.__get_nodes(doc.documentElement, 'placement_policy')

    def __get_rtpolicy(self, doc, rtpolicy_id):
        rtpolicy_nodes = self.__get_rtpolicies(doc)
        for rtpolicy_node in rtpolicy_nodes:
            node_id = self.__get_nodes(rtpolicy_node, 'id')
            rt_id = self.__get_node_value(node_id[0])
            if rt_id == rtpolicy_id:
                return rtpolicy_node

        raise policy_exception.PlacementPolicyNotFound(
            rtpolicy_id=rtpolicy_id)

    def get_violation_count(self, host, aggregate_id=None):
        """
        Returns the number of consecutive times the host has been violated
        """
        if not CONF.policy_per_host_aggregate:
            aggregate_id = ego_constants.RESOURCE_GROUP_DEFAULT
        aggregate_status = self.aggregate_statuses.get(aggregate_id)
        if not aggregate_status:
            return None
        return aggregate_status.get_violation_count(host)

    def get_all(self):
        """
        Get all runtime policies from rtpolicy_conf.xml for global mode.
        Get all runtime policies from metadata for aggregate mode.
        """
        rtpolicy_list = []
        doc = minidom.parse(self.rtpolicy_conf)
        rtpolicy_nodes = self.__get_rtpolicies(doc)
        for rtpolicy_node in rtpolicy_nodes:
            rtpolicy = None
            try:
                rtpolicy = self.__make_rtpolicy(rtpolicy_node)
            except (policy_exception.InvalidPlacementPolicyInput,
                    policy_exception.InvalidPlacementPolicyState,
                    policy_exception.InvalidPlacementPolicyAction,
                    Exception) as e:
                LOG.error(e)
                continue
            if rtpolicy:
                exist = False
                for policy in rtpolicy_list:
                    if policy.get(ego_constants.POLICY_ID) == rtpolicy.get(
                                                    ego_constants.POLICY_ID):
                        exist = True
                        break
                if exist:
                    LOG.error(_LE('Policy %(policy)s with the same ID already '
                                  'exists; this one will be ignored.')
                              % {'policy': rtpolicy})
                    continue
                rtpolicy_list.append(rtpolicy)
        return rtpolicy_list

    def __validate_rtpolicy_id(self, rtpolicy_id):
        if not rtpolicy_id:
            raise exception.InvalidInput(_('Specify a valid policy ID.'))
        try:
            int(rtpolicy_id)
        except Exception:
            raise exception.InvalidInput(_('Policy (%s) is not valid. '
                                           'Specify a numeric policy ID.')
                                         % rtpolicy_id)

    def get(self, rtpolicy_id):
        """
        Get specified runtime policy from rtpolicy_conf.xml for global mode.
        Get specified runtime policy from metadata for aggregate mode.
        """
        self.__validate_rtpolicy_id(rtpolicy_id)
        LOG.debug("The policy engine manager successfully "
                  "retrieved policy: %s" % rtpolicy_id)
        doc = minidom.parse(self.rtpolicy_conf)
        rtpolicy_node = self.__get_rtpolicy(doc, rtpolicy_id)
        return self.__make_rtpolicy(rtpolicy_node)

    def event_list(self, context, search_opts, marker=None, limit=None):
        """
        Returns a list of rtpolicy event based on search options specified.
        """
        hypervisor_id = search_opts.get('hypervisor_id')
        rtpolicy_id = search_opts.get('rtpolicy_id')
        aggregate_id = search_opts.get('aggregate_id')
        if hypervisor_id:
            self.__validate_hypervisor_id(hypervisor_id)
            hypervisor_id = long(hypervisor_id)
            search_opts['hypervisor_id'] = hypervisor_id
        if rtpolicy_id:
            self.__validate_rtpolicy_id(rtpolicy_id)
            rtpolicy_id = int(rtpolicy_id)
            search_opts['rtpolicy_id'] = rtpolicy_id
        if aggregate_id:
            self.__validate_aggregate_id(aggregate_id)
            aggregate_id = long(aggregate_id)
            search_opts['aggregate_id'] = aggregate_id
        return self.__load_events(context, search_opts, marker, limit)

    def hypervisor_list(self, context, search_opts):
        """
        Returns a list hypervisors's rtpolicy info based on the search
        options specified.
        """
        alert = search_opts.get('alert')
        rtpolicy_id = search_opts.get('rtpolicy_id')
        aggregate_id = search_opts.get('aggregate_id')
        hyvsr_list = []

        # Only return hypervisors in the aggregates, which would depend on
        # both rtpolicy_id and aggregated options. If it is None, return all
        # the hypervisors; if it is [], do not return any hypervisor; if it
        # is [ag1, ag2,...], return hypervisors in either in ag1, or in ag2...
        aggregates = None

        if rtpolicy_id:
            self.__validate_rtpolicy_id(rtpolicy_id)
            rtpolicy_id = int(rtpolicy_id)
            aggregates = (
                self.__get_aggregates_policy_valid_by_policy_id_and_type(
                         rtpolicy_id, ego_constants.RUNTIME_POLICY))
        if aggregate_id:
            if not CONF.policy_per_host_aggregate:
                raise exception.InvalidInput(_('Cannot support aggregate-id '
                                               'option in global mode (policy'
                                               '_per_host_aggregate = True is '
                                               'defined in the nova.conf file'
                                               ').'))
            else:
                self.__validate_aggregate_id(aggregate_id)
                aggregate_id = long(aggregate_id)
                if aggregates is None:
                    if aggregate_id not in self.aggid_policy_dict:
                        LOG.warning(_LW('Aggregate %s could not be found.')
                                    % aggregate_id)
                        aggregates = []
                    elif (ego_constants.RUNTIME_POLICY not in
                              self.aggid_policy_dict[aggregate_id]):
                        LOG.warning(_LW('Aggregate %s is not set with a valid '
                                        'runtime policy.') % aggregate_id)
                        aggregates = []
                    else:
                        aggregates = [aggregate_id]
                else:
                    if aggregate_id in aggregates:
                        aggregates = [aggregate_id]
                    else:
                        aggregates = []
        if aggregates or (aggregates is None):
            hyvsr_list = self.__hypervisor_list(context, aggregates, alert)
        return hyvsr_list

    def __hypervisor_list(self, context, aggregate_list=None, alert=False):
        """
        Get hypervisors' rtpolicy info list

        :param aggregate_list: the hypervisor's aggregate's runtime
                               policy is valid and the aggregate should
                               be in aggregate_list
        :param alert: the hypervisors' rtpolicy should has been
                      alerting if True
        :returns: hypervisors' rtpolicy info list specified by aggregate_list
                  and alert option
        """
        hyvsr_list = []
        if not self.ego_host_dict:
            self.__construct_ego_host_dict(context)
        if aggregate_list is None:
            if not CONF.policy_per_host_aggregate:
                if self.enabled_policy:
                    aggregate_list = [ego_constants.RESOURCE_GROUP_DEFAULT]
                else:
                    aggregate_list = []
            else:
                aggregate_list = []
                for aggregate_id, policy in six.iteritems(
                                                self.aggid_policy_dict):
                    if ego_constants.RUNTIME_POLICY in policy:
                        aggregate_list.append(aggregate_id)
        agg_rtpolicy_map = self.__get_aggregate_policy_map(aggregate_list,
                                                ego_constants.RUNTIME_POLICY)
        for agg in aggregate_list:
            if agg not in agg_rtpolicy_map:
                # The aggregate did not set rtpolicy
                continue

            aggregate_status = self.aggregate_statuses.get(agg)

            if alert:
                egohost_list = []
                if aggregate_status:
                    egohost_list = aggregate_status.alarming_resources
            else:
                egohost_list = []
                if agg != ego_constants.RESOURCE_GROUP_DEFAULT:
                    try:
                        svc_hosts = db.aggregate_host_get_all(context, agg)
                    except exception.AggregateNotFound:
                        # This can happen due to some race conditions
                        continue
                    for ego_host, h_dict in six.iteritems(self.ego_host_dict):
                        if h_dict['svc_host'] in svc_hosts:
                            egohost_list.append(ego_host)
                else:
                    egohost_list = self.ego_host_dict.keys()
            for egohost in egohost_list:
                hyvsr = {}
                hyvsr['id'] = self.ego_host_dict[egohost]['node_id']
                hyvsr['hypervisor_hostname'] = (
                        self.ego_host_dict[egohost]['node_name'])
                if aggregate_status:
                    hyvsr['consecutive_alert_times'] = (
                        aggregate_status.get_violation_count(egohost))
                else:
                    hyvsr['consecutive_alert_times'] = 0
                hyvsr['stabilization'] = (
                        agg_rtpolicy_map[agg]['stabilization'])
                hyvsr['rtpolicy_id'] = agg_rtpolicy_map[agg]['id']
                hyvsr['rtpolicy_name'] = agg_rtpolicy_map[agg]['name']
                hyvsr['is_violated'] = 'No'
                if (aggregate_status and
                        aggregate_status.stably_violated(egohost)):
                    hyvsr['is_violated'] = 'Yes'
                if CONF.policy_per_host_aggregate:
                    hyvsr['aggregate_id'] = agg
                hyvsr_list.append(hyvsr)
        hyvsr_list.sort(key=lambda hyvsr: hyvsr['id'])
        return hyvsr_list

    def store_event(self, context, event_type, payload):
        extra_payload_info = None
        try:
            extra_payload_info = self.rtpolicy_driver.handle_notification(
                context, event_type, payload)
        except Exception as e:
            # Just log the exception and move on
            LOG.exception(e)
        if extra_payload_info is not None:
            # Add extension-specific information into the payload copy
            payload = payload.copy()
            payload.update(extra_payload_info)
        self.notifier.info(context, event_type, payload)

    def __load_events(self, context, search_opts=None, marker=None,
                      limit=None):
        query = []
        query.append({'field': 'policy_type', 'op': 'eq', 'value': 'rtpolicy',
                      'type': 'string'})
        if search_opts:
            if search_opts.get('hypervisor_id'):
                query.append({'field': 'hypervisor_id', 'op': 'eq',
                              'value': search_opts['hypervisor_id'],
                              'type': 'integer'})
            if search_opts.get('rtpolicy_id'):
                query.append({'field': 'rtpolicy_id', 'op': 'eq',
                              'value': search_opts['rtpolicy_id'],
                              'type': 'integer'})
            if search_opts.get('aggregate_id'):
                query.append({'field': 'aggregate_id', 'op': 'eq',
                              'value': search_opts['aggregate_id'],
                              'type': 'integer'})

        return self.ceilometer_api.get_events(query, marker, limit)

    def __get_aggregate_policy_map(self, aggregate_list, policy_type):
        agg_rtpolicy_map = {}
        for aggregate_id in aggregate_list:
            policy = None
            if aggregate_id == ego_constants.RESOURCE_GROUP_DEFAULT:
                policy = self.enabled_policy
            else:
                policy = (self.aggid_policy_dict.get(aggregate_id, {}).get(
                              policy_type))
            if policy:
                policy_id = policy[ego_constants.POLICY_ID]
                # Get policy name from all policies by id
                policy_name = self.all_policies[policy_id].get(
                    ego_constants.POLICY_NAME)
                agg_rtpolicy_map[aggregate_id] = (
                    {'id': policy_id,
                     'name': policy_name,
                     'stabilization': policy[
                                        ego_constants.POLICY_STABILIZATION]})
        return agg_rtpolicy_map

    def __get_aggregates_policy_valid_by_policy_id_and_type(self, policy_id,
                                                            policy_type):
        """
        Get host aggregates whose policy_type is valid and this policy id
        is policy_id specified.
        """
        aggregates = []
        # For global mode, only one policy can be enabled at most
        if not CONF.policy_per_host_aggregate:
            if (self.enabled_policy and
                   self.enabled_policy['id'] == policy_id):
                aggregates = [ego_constants.RESOURCE_GROUP_DEFAULT]
        # For aggregate mode, the same rtpolicy might be applied to multiple
        # host aggregates
        else:
            host_aggregates = (
                self.__get_host_aggregates_by_policy_id_and_type(
                         policy_id, policy_type))
            if host_aggregates:
                for host_aggregate in host_aggregates['host_aggregates']:
                    if host_aggregate[ego_constants.POLICY_VALID_FLAG]:
                        aggregates.append(
                            host_aggregate[ego_constants.POLICY_ID])
        return aggregates

    def __validate_aggregate_id(self, aggregate_id):
        if not aggregate_id:
            raise exception.InvalidInput(_('Specify a valid aggregate ID.'))
        try:
            int(aggregate_id)
        except Exception:
            raise exception.InvalidInput(_('Aggregate (%s) is not valid. '
                                           'Specify a numeric aggregate '
                                           'ID.') % aggregate_id)

    def __validate_hypervisor_id(self, hypervisor_id):
        if not hypervisor_id:
            raise exception.InvalidInput(_('Specify a valid hypervisor ID.'))
        try:
            int(hypervisor_id)
        except Exception:
            raise exception.InvalidInput(_('Hypervisor (%s) is not valid. '
                                           'Specify a numeric hypervisor '
                                           'ID.') % hypervisor_id)

    def __update_policy_state(self, doc, rtpolicy_node, rtpolicy, state):
        state_node = self.__get_nodes(rtpolicy_node, 'state')
        self.__set_node_value(doc, state_node, state)
        rtpolicy['state'] = state

    def __enable_policy(self, doc, rtpolicy_node, rtpolicy, goal):
        if (self.enabled_policy and
                self.enabled_policy['id'] != rtpolicy['id']):
            raise policy_exception.InvalidPlacementPolicyOperation(
                operation='update', rtpolicy_id=rtpolicy['id'],
                reason=_('already has an enabled policy'))

        if (goal and goal == rtpolicy['goal'] and
                rtpolicy['state'] == 'enabled'):
            return

        # Clean up global variables
        self.aggregate_statuses = {}
        # Reset the optimizers
        for optimizer in self.optimizers:
            optimizer.reset_optimizer()

        # Update EGO resource plan for global policy mode
        if not CONF.policy_per_host_aggregate:
            self.__validate_policy_parameters(rtpolicy, rtpolicy['id'])
            try:
                if goal:
                    self.__set_ego_resource_plan(goal)
                else:
                    self.__set_ego_resource_plan(rtpolicy['goal'])
            except (ego_exception.EGOExpSetResourcePlanFailed,
                    ego_exception.EGOExpConnLost) as exc:
                LOG.error(_LE('The policy engine manager could not %(action)s '
                              'policy %(rtpolicy_id)s due to: %(reason)s')
                          % {'action': 'enable',
                             'rtpolicy_id': rtpolicy['id'],
                             'reason': six.text_type(exc)})
                raise

        # Update XML
        self.__update_policy_state(doc, rtpolicy_node, rtpolicy, 'enabled')
        self.enabled_policy = rtpolicy

        LOG.info(_LI('The policy engine manager successfully changed policy '
                     '(%(rtpolicy_id)s) to state: %(state)s')
                 % {'rtpolicy_id': rtpolicy['id'],
                    'state': 'enabled'})

    def __disable_policy(self, doc, rtpolicy_node, rtpolicy):
        # Do not update policy state if already disabled
        if rtpolicy['state'] == 'disabled':
            return
        # Set EGO resource plan for global policy mode
        if not CONF.policy_per_host_aggregate:
            self.__set_ego_resource_plan()

        self.aggregate_statuses = {}
        for optimizer in self.optimizers:
            optimizer.reset_optimizer()

        # Update XML
        self.__update_policy_state(doc, rtpolicy_node, rtpolicy, 'disabled')

        # Clean up global variables
        self.enabled_policy = None
        LOG.info(_LI('The policy engine manager successfully changed policy '
                     '(%(rtpolicy_id)s) to state: %(state)s')
                 % {'rtpolicy_id': rtpolicy['id'],
                    'state': 'disabled'})

    def __is_update_operation_valid(self, param_dict):
        invalid_parameter_of_aggregate_mode = [
            ego_constants.POLICY_ACTION,
            ego_constants.POLICY_MAX_PARALLEL,
            ego_constants.POLICY_THRESHOLD,
            ego_constants.POLICY_STABILIZATION,
            ego_constants.POLICY_STATE,
            ego_constants.POLICY_RUN_INTERVAL
        ]
        for item in param_dict.items():
            if item[0] in invalid_parameter_of_aggregate_mode:
                return False
        return True

    def update(self, rtpolicy_id, kwargs):
        """
        Update specified runtime policy from rtpolicy_conf.xml
        """
        self.__validate_rtpolicy_id(rtpolicy_id)
        rtpolicy = self.get(rtpolicy_id)
        LOG.info(_LI("The policy engine manager is going to "
                   "update policy: %s")
                 % rtpolicy_id)
        if CONF.policy_per_host_aggregate:
            if not self.__is_update_operation_valid(kwargs):
                raise policy_exception.InvalidPlacementPolicyOperation(
                    operation='update', rtpolicy_id=rtpolicy_id,
                    reason=_('policy_per_host_aggregate = True '
                             'is defined in the nova.conf file'))
        # Check validation of input arguments
        self.__validate_policy_parameters(kwargs, rtpolicy_id)
        doc = minidom.parse(self.rtpolicy_conf)
        rtpolicy_node = self.__get_rtpolicy(doc, rtpolicy_id)
        state = kwargs.get(ego_constants.POLICY_STATE)
        goal = kwargs.get(ego_constants.POLICY_GOAL)
        if state:
            # Check EGO connection
            if self.ego_api.ego_hdl.is_conn_lost():
                raise policy_exception.PlacementPolicyEGOConnLost()
            if state == 'enabled':
                self.__enable_policy(doc, rtpolicy_node, rtpolicy, goal)
            else:
                self.__disable_policy(doc, rtpolicy_node, rtpolicy)
        rtpolicy_attribute_update = [ego_constants.POLICY_DESCRIPTION,
                                     ego_constants.POLICY_GOAL,
                                     ego_constants.POLICY_NAME,
                                     ego_constants.POLICY_ACTION,
                                     ego_constants.POLICY_MAX_PARALLEL,
                                     ego_constants.POLICY_RUN_INTERVAL,
                                     ego_constants.POLICY_STABILIZATION,
                                     ego_constants.POLICY_STATE,
                                     ego_constants.POLICY_THRESHOLD]
        for key, value in six.iteritems(kwargs):
            if key in rtpolicy_attribute_update:
                key_node = self.__get_nodes(rtpolicy_node, key)
                self.__set_node_value(doc, key_node, value)
                rtpolicy[key] = value
                LOG.debug("The policy engine manager successfully updated "
                          "key %(key)s for policy %(rtpolicy_id)s "
                          "to: %(value)s"
                          % {'key': key,
                             'rtpolicy_id': rtpolicy_id,
                             'value': value})
        self.__updatexml(doc)
        # If the updated rtpolicy is the enabled policy, update enabled policy
        if (self.enabled_policy
                and self.enabled_policy.get('id') == rtpolicy.get('id')):
            self.enabled_policy = rtpolicy
        # The policy within all_policies should be updated as well
        self.all_policies[rtpolicy['id']] = rtpolicy
        # Set resource plan based on different working mode
        if goal:
            if CONF.policy_per_host_aggregate:
                self.__update_aggregate_and_policy_info()
                self.__set_resource_plan_per_aggregate()
            else:
                self.__set_ego_resource_plan(goal)
        return rtpolicy

    #################################################################
    #                    End Policy Management                      #
    #################################################################

    def __generate_group_resreq(self, aggregate_id):
        return "%s <in> %s" % (str(aggregate_id),
                               ego_constants.RPI_KEY_HOST_AGGREGATE_ID)

    def __clean_internal_aggregate_data(self, agg_id):
        """Clean up internal structures when deleting an aggregate."""
        if agg_id in self.aggregate_statuses:
            del self.aggregate_statuses[agg_id]

    def __resource_optimization_per_aggregate(self, context,
                                              aggregate_id, policy):
        """
        Check and trigger the resource optimization per host aggregate.
        """
        run_interval = int(policy[ego_constants.POLICY_RUN_INTERVAL])
        stabilization = int(policy[ego_constants.POLICY_STABILIZATION])

        if aggregate_id not in self.aggregate_statuses:
            aggregate_status = (
                rtpolicy_resource_group_status.RTPolicyResourceGroupStatus(
                    run_interval, stabilization, aggregate_id))
            self.aggregate_statuses[aggregate_id] = aggregate_status
        else:
            aggregate_status = self.aggregate_statuses[aggregate_id]
            aggregate_status.run_interval = run_interval
            aggregate_status.stabilization = stabilization

        if not CONF.policy_per_host_aggregate:
            LOG.debug("Current interval count is %(current)d/%(total)s." %
                      {'current': aggregate_status.current_interval,
                       'total': run_interval})
        else:
            LOG.debug("Current interval count is %(current)d/%(total)s for "
                      "aggregate with ID of %(agg_id)s." %
                      {'agg_id': aggregate_id,
                       'current': aggregate_status.current_interval,
                       'total': run_interval})

        do_not_continue = aggregate_status.current_interval_lt_run_interval()
        aggregate_status.increment_current_interval()

        if do_not_continue:
            return

        goal = self.__get_goal_str(policy[ego_constants.POLICY_ID])
        threshold = int(policy[ego_constants.POLICY_THRESHOLD])
        policy_id = policy[ego_constants.POLICY_ID]
        policy_name = self.all_policies[policy_id][ego_constants.POLICY_NAME]
        policy_max_migrations = int(policy[ego_constants.POLICY_MAX_PARALLEL])
        # Prepare the policy_info dict that will be passed to the policy driver
        policy_info = dict(aggregate_id=aggregate_id,
                           policy_id=policy_id,
                           policy_name=policy_name,
                           policy_goal=goal,
                           policy_action=policy[ego_constants.POLICY_ACTION],
                           policy_threshold=threshold,
                           policy_stabilization=stabilization,
                           policy_run_interval=run_interval,
                           policy_max_migrations=policy_max_migrations)

        for optimizer in self.optimizers:
            optimizer.run_interval_ticker(context, policy_info)

        # Check whether optimization is enabled for the host aggregate
        result = self.rtpolicy_driver.is_aggregate_optimization_enabled(
            context, aggregate_id, policy_info)
        if not result:
            LOG.debug("The policy driver prevented the host aggregate (%s) "
                      "from being optimized because its optimization has been "
                      "disabled." % aggregate_id)
            # If optimization is disabled, none of its hosts should be violated
            aggregate_status.reset_violation_counts()
            return

        LOG.debug('Platform Resource Scheduler has determined that the '
                  'load on these hosts exceeds the threshold '
                  '(%(threshold)d): %(host)s'
                  % {'threshold': threshold,
                     'host': aggregate_status.violation_counts})

        # Check whether optimization is allowed for the host aggregate
        result = self.rtpolicy_driver.can_perform_aggregate_optimization(
            context, aggregate_id, policy_info)
        if not result[0]:
            if len(result) == 2:
                # i.e., the extension provided its own message
                reason = result[1]
            else:
                reason = (("The policy driver prevented the host aggregate "
                           "(%s) from being considered for ongoing "
                           "optimization.") % aggregate_id)
            LOG.debug(reason)
            # NOTE(jwcroppe): We don't reset the violated hosts (i.e.,
            # self.aggregate_statuses[aggregate_id].reset_violation_counts())
            # because this may be a more transient condition
            return

        # Get violated_hosts with the given host aggregate scope
        violated_hosts = self.__get_violated_hosts(context, policy_info)

        valid_violated_hosts = []
        for host in violated_hosts:
            result = self.rtpolicy_driver.can_perform_host_optimization(
                context, host[0], policy_info)
            if not result[0]:
                if len(result) == 2:
                    # i.e., the extension provided its own message
                    reason = result[1]
                else:
                    reason = (("The policy driver prevented the host (%s) "
                               "from being considered for ongoing "
                               "optimization.") % host[0])
                LOG.debug(reason)
            else:
                valid_violated_hosts.append(host)

        optimized_hosts = []
        for optimizer in self.optimizers:
            ret = optimizer.optimize_hosts(context, valid_violated_hosts,
                                           optimized_hosts, policy_info)
            # Stop subsequent optimizations for this cycle if instructed to
            if ((ret['code'] & ego_constants.OPTIMIZE_HOSTS_WAIT_MASK) ==
                ego_constants.OPTIMIZE_HOSTS_WAIT_MASK):
                break
            new_optimized_hosts = ret['optimized_hosts'] or []
            optimized_hosts = list(set(optimized_hosts + new_optimized_hosts))

    def __check_resource_plan_change(self):
        """
        Check and update if host aggregate based initial policies are changed.
        Note that it will be available only when running in aggregate mode.
        """
        # Retrieve the latest policies, compare with the self.aggid_policy_dict
        # to see if there are changes for initial policy; only check for
        # aggregate mode
        if (not CONF.policy_per_host_aggregate) or self.need_set_resource_plan:
            return
        for aggregate_id, policy in (
                six.iteritems(
                    self.aggregate_container.aggregate_id_policy_dict)):
            if aggregate_id not in self.aggid_policy_dict:
                self.need_set_resource_plan = True
                break
            else:
                current_policy = self.aggid_policy_dict[aggregate_id]
                if ego_constants.INITIAL_POLICY in policy:
                    # If the latest one has the initial policy, check
                    # whether the current policy as the same, otherwise,
                    # need to set resource plan
                    new_policy_init = policy[ego_constants.INITIAL_POLICY]
                    if ego_constants.INITIAL_POLICY in current_policy:
                        current_policy_init = (
                            current_policy[ego_constants.INITIAL_POLICY])
                        if current_policy_init[ego_constants.POLICY_ID] != (
                                new_policy_init[ego_constants.POLICY_ID]):
                            self.need_set_resource_plan = True
                            break
                    else:
                        self.need_set_resource_plan = True
                        break
                else:
                    if ego_constants.INITIAL_POLICY in current_policy:
                        self.need_set_resource_plan = True
                        break

    def __policy_function_global_mode(self, context):
        """
        Set resource plan and trigger on-going policy for global policy mode.
        """
        if self.need_set_resource_plan:
            try:
                goal = (self.enabled_policy.get('goal', '')
                        if self.enabled_policy else '')
                self.__set_ego_resource_plan(goal)
                self.need_set_resource_plan = False
            except (ego_exception.EGOExpSetResourcePlanFailed,
                    ego_exception.EGOExpConnLost) as exc:
                LOG.error(_LE('The policy engine manager could not set '
                              'the EGO resource plan for policy '
                              '%(policy_id)s due to: %(reason)s')
                          % {'policy_id': self.enabled_policy.get('id'),
                             'reason': six.text_type(exc)})
        if self.enabled_policy:
            self.__resource_optimization_per_aggregate(
                context, ego_constants.RESOURCE_GROUP_DEFAULT,
                self.enabled_policy)

    def __update_aggregate_and_policy_info(self):
        """
        Internal wrapper to update latest aggregate and policy information.
        """
        self.aggid_policy_dict = copy.deepcopy(
            self.aggregate_container.aggregate_id_policy_dict)

    @lockutils.synchronized('prs-aggregate-lock', 'prs-aggregate-lock-')
    def __policy_function_aggregate_mode(self, context):
        # Sync up the relationship between host and aggregate
        self.aggregate_container.sync_up_aggregate_policy(
            set(self.all_policies.keys()))
        self.__check_resource_plan_change()
        self.__update_aggregate_and_policy_info()
        if self.need_set_resource_plan:
            self.__set_resource_plan_per_aggregate()

        self.aggregate_container.sync_aggregates_and_metadatas()

        # Go through all host aggregates to trigger on-going resource
        # optimization policy based on their settings
        for aggregate_id, policy in six.iteritems(self.aggid_policy_dict):
            if ego_constants.RUNTIME_POLICY in policy:
                self.__resource_optimization_per_aggregate(
                    context, aggregate_id,
                    policy[ego_constants.RUNTIME_POLICY])

    def __construct_ego_host_dict(self, context):
        """
        Get map from EGO host to service host and hypervisor name.
        This should run at the beginning of each periodic task.
        """
        def get_parallel_migration(compute_node):
            """Retrieve the max # of parallel migrations supported by host."""
            # TODO(jwcroppe): consider consolidating this with the HA logic;
            # need to be mindful of compatibility concerns when re-factoring.
            stats = compute_node.stats or {}
            return int(stats.get(PARALLEL_MIGRATE, DEFAULT_PARALLEL_MIGRATE))

        compute_nodes = objects.ComputeNodeList.get_all(context)
        self.ego_host_dict.clear()
        for compute_node in compute_nodes:
            svc_host = compute_node['host']
            node_name = compute_node['hypervisor_hostname']
            htype = compute_node['hypervisor_type']
            ego_host = ego_resreq_utils.get_rpi_res_name(node_name)
            max_migr = get_parallel_migration(compute_node)
            self.ego_host_dict[ego_host] = {'svc_host': svc_host,
                                            'node_name': node_name,
                                            'node_id': compute_node['id'],
                                            'max_migrations': max_migr,
                                            'htype': htype}

    def policy_function(self, context):
        """
        Periodic task (called via the manager) that's called every 60 seconds.
        It serves two main purposes:

        1. set resource plan per resource groups as initial placement policy
        2. trigger ongoing optimization policy based on the goal and threshold
           of each host aggregates

        For the backward compatibility, the global policy is still supported;
        as a general rule, if there is any policy is enabled, then it means it
        is treated as global policy.
        """
        # Check EGO connection
        if self.ego_api.ego_hdl.is_conn_lost():
            self.need_set_resource_plan = True
            return

        self.__construct_ego_host_dict(context)
        for optimizer in self.optimizers:
            optimizer.run_ticker(context)

        # Has global policy enabled; no need to do on-going resource
        # optimization policy
        if CONF.policy_per_host_aggregate:
            self.__policy_function_aggregate_mode(context)
        else:
            self.__policy_function_global_mode(context)

    def __get_violated_hosts(self, context, policy_info):
        """
        Retrieve a list of violated hosts from EGO.

        :param policy_info: a dict containing all of the policy information,
                            including the aggregate ID, policy id, policy name,
                            policy goal, policy threshold, policy run interval,
                            policy stabilization and policy max migrations.
        :returns: a list of violated hosts that are eligible for optimization
        """
        aggregate_id = policy_info['aggregate_id']
        policy_id = policy_info['policy_id']
        policy_name = policy_info['policy_name']
        threshold = policy_info['policy_threshold']
        goal = policy_info['policy_goal']
        stabilization = policy_info['policy_stabilization']
        run_interval = policy_info['policy_run_interval']
        hosts_to_return = []

        goal = goal + "*100" if goal.find('.percent') >= 0 else goal
        if goal == ego_constants.PACKING_POLICY_GOAL:
            # Packing policy: not enough VMs on host, threshold = minimum VMs
            cstr = ('(%s+1+%s>0)&&(%s)' %
                    (goal, threshold,
                     self.__generate_group_resreq(aggregate_id)))
        elif goal == ego_constants.STRIPING_POLICY_GOAL:
            # Striping policy: too many VMs on host, threshold = maximum VMs
            cstr = ('(%s-1>%s)&&(%s)' %
                    (goal, threshold,
                     self.__generate_group_resreq(aggregate_id)))
        else:
            cstr = ('(%s>%s)&&(%s)' %
                    (goal, threshold,
                     self.__generate_group_resreq(aggregate_id)))

        # 1. powervc_driver_host != 1 filter out the PowerVM sync'd by the
        #    PowerVC driver from PowerVC; these hosts don't need to involve
        #    rtpolicy.
        # 2. the containers on Docker hypervisor do not support live migration,
        #    so Docker hypervisors don't need to involve rtpolicy.
        cstr_select = ('select(%(cstr)s&&(hypervisor_type!=%(vmware)s)&&'
                       '((! cpu_mgmt)||(cpu_mgmt==off))&&'
                       '(hypervisor_type!=%(docker)s)&&(activeCompute==1)'
                       '&&(powervc_driver_host!=1))order(0-(%(goal)s))' %
                       {'cstr': cstr,
                        'vmware': ego_constants.HYPERVISOR_TYPE_VMWARE_RPI,
                        'docker': ego_constants.DOCKER_HYPERVISOR_TYPE,
                        'goal': goal})
        violated_hosts = self.__get_ego_resources(cstr_select)
        LOG.debug('The policy engine manager successfully '
                  'retrieved violated hosts from EGO using '
                  'the RPI string: %(cstr)s:%(host)s'
                  % {'cstr': cstr,
                     'host': violated_hosts})
        optimization_disabled = {}

        aggr_status = self.aggregate_statuses[aggregate_id]
        prev_stably_violated_hosts = copy.deepcopy(
            aggr_status.stably_violated_resources)

        for index, hostname in enumerate(violated_hosts):
            # Check whether optimization for the host has been disabled
            result = self.rtpolicy_driver.is_host_optimization_enabled(
                context, hostname, policy_info)
            if not result:
                LOG.debug("The policy driver prevented the host (%s) from "
                          "being optimized because its optimization has been "
                          "disabled." % hostname)
                optimization_disabled[hostname] = True
                # Hosts with optimization disabled should have their violation
                # counts reset
                aggr_status.reset_violation_count(hostname)
                continue

            aggr_status.increment_violation_count(hostname)

            host_dict = self.ego_host_dict[hostname]
            payload = {'hypervisor_id': host_dict['node_id'],
                       'hypervisor_hostname': host_dict['node_name'],
                       'policy_type': 'rtpolicy',
                       'rtpolicy_id': policy_id,
                       'rtpolicy_name': policy_name,
                       'threshold': threshold,
                       'stabilization': stabilization,
                       'run_interval': run_interval,
                       'is_violated': (
                           'Yes' if aggr_status.stably_violated(hostname)
                           else 'No'),
                       'alert_times': aggr_status.get_violation_count(
                           hostname)}
            if CONF.policy_per_host_aggregate:
                payload['aggregate_id'] = aggregate_id
            self.store_event(context,
                             ego_constants.PRS_RTPOLICY_HYPERVISOR_ALERT,
                             payload)
            # The host will perform rtpolicy action every <stabilization> times
            # if the rtpolicy alert continues.
            if aggr_status.get_violation_count(hostname) % stabilization == 0:
                LOG.debug('The violated host (%(host)s) reached the '
                          'stabilization period (%(period)s).'
                          % {'host': hostname, 'period': stabilization})
                hosts_to_return.append((hostname, index))
                if aggr_status.get_violation_count(hostname) == stabilization:
                    self.store_event(
                        context,
                        ego_constants.PRS_RTPOLICY_HYPERVISOR_VIOLATED,
                        payload)

        # Recovered hosts are those that are no longer violating the rtpolicy.
        # This is formed by computing the set difference between the set of
        # hosts in the RTPolicyResourceGroupStatus object that are "alarming"
        # (i.e., have violation counts > 0) and the set of hosts from this
        # cycle's violated hosts, thus leaving the recovered hosts in the
        # resulting set.
        recovered_hosts = (
            set(aggr_status.alarming_resources) - set(violated_hosts))

        if recovered_hosts:
            # Fetch all nodes with activeCompute==0 so that we can determine
            # if all "recovered" hosts are actually recovered in reality
            cstr = 'select(activeCompute==0)'
            inactive_compute_hosts = self.__get_ego_resources(cstr)
            LOG.debug('The policy engine manager successfully '
                      'retrieved hosts with inactive compute from EGO using '
                      'the RPI string: %(cstr)s:%(hosts)s'
                      % {'cstr': cstr,
                         'hosts': inactive_compute_hosts})

        for hostname in recovered_hosts:
            aggr_status.reset_violation_count(hostname)

            if hostname in optimization_disabled:
                # If optimization is disabled, don't issue any notifications
                continue
            host_dict = self.ego_host_dict.get(hostname)
            if not host_dict:
                # This can happen when a host that was violated is removed from
                # the host aggregate; just move on since it doesn't make sense
                # to generate a notification for the host as it's no longer a
                # member of the aggregate.
                continue
            payload = {'hypervisor_id': host_dict['node_id'],
                       'hypervisor_hostname': host_dict['node_name']}
            if CONF.policy_per_host_aggregate:
                payload['aggregate_id'] = aggregate_id
            # Only issue the "recovered notification" if the host was truly
            # violated (i.e., the stabilization period was encountered).
            if hostname in prev_stably_violated_hosts:
                if hostname in inactive_compute_hosts:
                    # Node has an activeCompute value of 0, hence why it is
                    # no longer returned as a violated host.  This could mean
                    # it is down or it could mean it is in maintenance mode.
                    # Therefore the current status of this node is invalidated
                    if ego_utils.get_compute_status().get_maintenance(
                            host_dict['node_name']) != ha_constants.STATE_OK:
                        LOG.info(_LI('Node %(node_name)s is in maintenance '
                                     'mode; sending invalidated event.')
                                 % {'node_name': host_dict['node_name']})
                    else:
                        LOG.info(_LI('Node %(node_name)s is currently down, '
                                     'so it cannot be determined whether or '
                                     'not the node is actually no longer '
                                     'violated; sending invalidated event.')
                                 % {'node_name': host_dict['node_name']})
                    event = ego_constants.PRS_RTPOLICY_HYPERVISOR_INVALIDATED
                else:
                    payload.update({
                        'policy_type': 'rtpolicy',
                        'rtpolicy_id': policy_id,
                        'rtpolicy_name': policy_name,
                        'threshold': threshold,
                        'stabilization': stabilization,
                        'run_interval': run_interval,
                        'is_violated': 'No',
                        'alert_times': 0})
                    event = ego_constants.PRS_RTPOLICY_HYPERVISOR_RECOVERED

                self.store_event(context, event, payload)

        LOG.debug('The policy engine manager determined the latest set of '
                  'violated hosts that will potentially require policy action '
                  'to be: %s' %
                  self.aggregate_statuses[aggregate_id].violation_counts)
        # hosts_to_return are the hosts whose threshold is above the
        # configured threshold and we need to migrate some VMs off this
        # node to reduce the overall load on the host(s)
        LOG.debug('The policy engine manager determined the latest set of '
                  'violated hosts that require policy action to be: %s'
                  % dict(hosts_to_return).keys())
        return hosts_to_return

    def __get_ego_resources(self, cstr):
        hostnames = []
        try:
            host_list = self.ego_api.ego_get_resource_list(cstr)
            hostnames = [host['name'] for host in host_list]
        except (ego_exception.EGOExpGetResourceInfoFailed,
                ego_exception.EGOExpConnLost) as exc:
            LOG.error(_LE('The policy engine manager could not retrieve RPI '
                          'resources from EGO using resource requirements '
                          '(%(cstr)s) due to: %(reason)s')
                      % {'cstr': cstr,
                         'reason': six.text_type(exc)})
        return hostnames

    def __generate_resource_plan_for_group(self, rg_name, goal=""):
        goal = 'order(%s)' % goal if len(goal) > 0 else ''
        xmlstr = ("<ResourceGroup ResourceGroupName=\"%(rg_name)s\">"
                  "    <PlacementPolicy>%(goal)s</PlacementPolicy>"
                  "</ResourceGroup>" %
                  {'rg_name': rg_name, 'goal': goal})
        return xmlstr

    def __get_goal_str(self, policy_id):
        return self.all_policies[policy_id]['goal']

    def __set_ego_resource_plan(self, goal=''):
        """
        Set EGO resource plan; usage:
        1. When policy engine is initialized
        2. When enable global policy (if goal is specified)
        3. When set host aggregate based policy

        :param goal: resource requirement string or empty string
           a. for CPU sensitive load balance, the goal is:
              (vcpuConsumed + vcpuRequest) / vcpu
           b. for memory sensitive load balance the goal is:
              (memSizeConsumed + memSizeRequest) / memSize
           c. for striping policy the goal is:
              running_vms + 1
           d. for packing policy the goal is:
              0 - (running_vms + 1)
           e. for cpu_ut policy the goal is:
              cpu_ut
           f. for mem_ut policy the goal is:
              mem_ut
        """
        # Set default group first
        res_group_xml = self.__generate_resource_plan_for_group(
            ego_constants.RESOURCE_GROUP_DEFAULT, goal)
        for aggregate_id, policy in six.iteritems(self.aggid_policy_dict):
            # Global goal or host aggregate based input goal
            if goal:
                res_group_xml += self.__generate_resource_plan_for_group(
                    str(aggregate_id), goal)
            else:
                policy_goal = ""
                if ego_constants.INITIAL_POLICY in policy:
                    init_policy = policy[ego_constants.INITIAL_POLICY]
                    policy_id = init_policy[ego_constants.POLICY_ID]
                    policy_goal = self.__get_goal_str(policy_id)
                res_group_xml += self.__generate_resource_plan_for_group(
                    str(aggregate_id), policy_goal)

        xmlstr = ("<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
                  "<DistributionPolicies>"
                  "    <Scheduler>placetorus</Scheduler>"
                  "    <InitialPlacement>%s</InitialPlacement>"
                  "</DistributionPolicies>" % (res_group_xml))
        LOG.debug('The policy engine manager set the EGO resource plan '
                  'to the following XML string: %s' % xmlstr)
        self.ego_api.ego_set_resource_plan(xmlstr)

    #################################################################
    #                Begin Notification Handlers                    #
    #################################################################

    def notify_policy_resource_status(self, context, msg):
        """
        Notify the resource optimization policy service about resource status
        updates such that appropriate action(s) can be taken.
        """
        event_type = msg['event_type']
        payload = msg['payload']

        for optimizer in self.optimizers:
            optimizer.receive_notification(context, msg)

        # If we're not running in aggregate mode, then there is no need to pass
        # along any of the additional notifications
        if not CONF.policy_per_host_aggregate:
            return

        # Handle host aggregate notifications
        if event_type == ego_constants.EVENT_TYPE_AGGREGATE_CREATE_END:
            LOG.debug("Aggregate create end message: %s" % msg)
            self.__notify_aggregate_create(payload)
        elif event_type == ego_constants.EVENT_TYPE_AGGREGATE_DELETE_END:
            LOG.debug("Aggregate delete end message: %s" % msg)
            self.__notify_aggregate_delete(payload)
        elif event_type == ego_constants.EVENT_TYPE_AGGREGATE_METADATA_SET_END:
            LOG.debug("Aggregate metadata update end message: %s" % msg)
            self.__notify_aggregate_set_metadata(context, payload)
        elif event_type == ego_constants.EVENT_TYPE_AGGREGATE_HOST_REMOVE_END:
            LOG.debug("Aggregate remove host end message: %s" % msg)
            self.__notify_aggregate_remove_host(context, payload)

    @lockutils.synchronized('prs-aggregate-lock', 'prs-aggregate-lock-')
    def __notify_aggregate_create(self, payload):
        # Just create resource group accordingly and update policy's
        # dictionary; no need to set resource plan here
        if int(payload['aggregate_id']) not in self.aggid_policy_dict:
            self.aggregate_container.sync_up_aggregate_policy(
                set(self.all_policies.keys()))
            self.aggid_policy_dict = copy.deepcopy(
                self.aggregate_container.aggregate_id_policy_dict)

    @lockutils.synchronized('prs-aggregate-lock', 'prs-aggregate-lock-')
    def __notify_aggregate_delete(self, payload):
        # Just delete resource group accordingly and update policy's
        # dictionary; no need to set resource plan here
        agg_id = int(payload['aggregate_id'])
        if agg_id in self.aggid_policy_dict:
            self.aggregate_container.sync_up_aggregate_policy(
                set(self.all_policies.keys()))
            self.aggid_policy_dict = copy.deepcopy(
                self.aggregate_container.aggregate_id_policy_dict)
        self.__clean_internal_aggregate_data(agg_id)

    @lockutils.synchronized('prs-aggregate-lock', 'prs-aggregate-lock-')
    def __notify_aggregate_set_metadata(self, context, payload):
        agg_id = int(payload['aggregate_id'])
        # Grab the original metadata for the aggregate (i.e., current settings)
        orig_agg_data = self.__get_aggregate_policy_map(
            [agg_id], ego_constants.RUNTIME_POLICY).get(agg_id)

        # Update policy's dictionary and set resource plan if needed
        self.aggregate_container.sync_up_aggregate_policy(
            set(self.all_policies.keys()), agg_id)
        self.__set_resource_plan_per_aggregate()

        # Grab the new metadata for the aggregate
        new_agg_data = self.__get_aggregate_policy_map(
            [agg_id], ego_constants.RUNTIME_POLICY).get(agg_id)

        if not new_agg_data and agg_id in self.aggregate_statuses:
            self.aggregate_statuses[agg_id].reset_violation_counts()

        # An unexpected case happened, so just return
        if not (orig_agg_data and new_agg_data):
            return

        # Invalidate hosts if necessary
        self.__invalidate_hosts_if_needed(context, agg_id, orig_agg_data,
                                          new_agg_data)

    def __invalidate_hosts_if_needed(self, context, agg_id, orig_agg_data,
                                     new_agg_data):
        """Invalidate host violations if necessary based on changes."""
        def get_ego_host_data(svc_host):
            ret = (None, None)
            for ego_host, h_dict in six.iteritems(self.ego_host_dict):
                if h_dict is not None and h_dict.get('svc_host') == svc_host:
                    ret = (ego_host, h_dict)
                    break
            return ret

        invalidate_hosts = False

        if orig_agg_data.get('id') != new_agg_data.get('id'):
            # The rtpolicy was changed; invalidate results
            invalidate_hosts = True
        elif (orig_agg_data.get('stabilization') !=
              new_agg_data.get('stabilization')):
            # The stabilization count changed; invalidate results
            invalidate_hosts = True

        # Don't send any events if we don't need to
        if not invalidate_hosts:
            return

        # Wipe the violation counts and invalidate the violation
        if agg_id in self.aggregate_statuses:
            self.aggregate_statuses[agg_id].reset_violation_counts()
        # Load the service hosts for the aggregate
        svc_hosts = [k for (k, v) in
                     six.iteritems(self.aggregate_container.host_aggregates)
                     if agg_id in v]
        for svc_host in svc_hosts:
            (ego_host, host_dict) = get_ego_host_data(svc_host)
            if ego_host is None or host_dict is None:
                continue
            payload = {'hypervisor_id': host_dict['node_id'],
                       'hypervisor_hostname': host_dict['node_name'],
                       'aggregate_id': agg_id}
            self.store_event(context,
                             ego_constants.PRS_RTPOLICY_HYPERVISOR_INVALIDATED,
                             payload)

    @lockutils.synchronized('prs-aggregate-lock', 'prs-aggregate-lock-')
    def __notify_aggregate_remove_host(self, context, payload):
        # Remove the host from the respective aggregate's alert counter; this
        # prevents erroneous recovery notifications from occurring after we
        # have removed a host from an aggregate
        def get_ego_host_data(svc_host):
            ret = (None, None)
            for ego_host, h_dict in six.iteritems(self.ego_host_dict):
                if h_dict is not None and h_dict.get('svc_host') == svc_host:
                    ret = (ego_host, h_dict)
                    break
            return ret

        (ego_host, host_dict) = get_ego_host_data(payload['host_name'])
        if ego_host is None or host_dict is None:
            return
        agg_id = int(payload['aggregate_id'])
        aggregate_status = self.aggregate_statuses.get(agg_id)
        if aggregate_status:
            if ego_host in aggregate_status.violation_counts:
                aggregate_status.reset_violation_count(ego_host)
        payload = {'hypervisor_id': host_dict['node_id'],
                   'hypervisor_hostname': host_dict['node_name'],
                   'aggregate_id': agg_id}
        self.store_event(context,
                         ego_constants.PRS_RTPOLICY_HYPERVISOR_INVALIDATED,
                         payload)

    #################################################################
    #                 End Notification Handlers                     #
    #################################################################
