# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2014 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""
Client side of the rtpolicy RPC API.
"""

from oslo_config import cfg
import oslo_messaging as messaging

from nova.objects import base as objects_base
from nova import rpc

rpcapi_opts = [
    cfg.StrOpt('rtpolicy_topic',
               default='rtpolicy',
               help='The topic on which the rtpolicy service listens'),
]

CONF = cfg.CONF
CONF.register_opts(rpcapi_opts)

rpcapi_cap_opt = cfg.StrOpt('rtpolicy',
            help='Set a version cap for messages sent to rtpolicy')
CONF.register_opt(rpcapi_cap_opt, 'upgrade_levels')


class RTPolicyAPI(object):
    '''Client side of the rtpolicy RPC API.

    API version history:

        1.0 - Initial version.

    '''

    VERSION_ALIASES = {
        'havana': '1.0'
    }

    def __init__(self, topic=None):
        super(RTPolicyAPI, self).__init__()
        target = messaging.Target(topic=CONF.rtpolicy_topic, version="1.0")
        version_cap = self.VERSION_ALIASES.get(CONF.upgrade_levels.rtpolicy,
                                               CONF.upgrade_levels.rtpolicy)
        serializer = objects_base.NovaObjectSerializer()
        self.client = rpc.get_client(target,
                                     version_cap=version_cap,
                                     serializer=serializer)

    def get_all(self, context):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'get_all')

    def get(self, context, rtpolicy_id):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'get', rtpolicy_id=rtpolicy_id)

    def delete(self, context, rtpolicy_id):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'delete', rtpolicy_id=rtpolicy_id)

    def update(self, context, rtpolicy_id, rtpolicy, **kwargs):
        args = kwargs
        for key in rtpolicy.keys():
            args[key] = rtpolicy[key]
        args['rtpolicy_id'] = rtpolicy_id
        cctxt = self.client.prepare()
        return cctxt.call(context, 'update', **kwargs)

    def hypervisor_list(self, context, search_opts):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'hypervisor_list', search_opts=search_opts)

    def event_list(self, context, search_opts, marker, limit):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'event_list', search_opts=search_opts,
                          marker=marker, limit=limit)
