# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from nova.ibm.ego import ego_constants
from nova.ibm.ego import policy_ceilometer_client


class RTPolicyDriver(object):
    """Extension point for customization of ongoing optimization behaviors."""

    def __init__(self, *args, **kwargs):
        # The Ceilometer API that is used for retrieving VM utilization
        self._ceilometer_api = policy_ceilometer_client.API()

    def get_instance_cpu_util(self, instance_obj):
        """
        Retrieves the effective CPU utilization percentage for the specified
        instance object.

        :param instance_obj: instance object (i.e., objects.Instance)
        :returns: a floating point representing the instance's CPU utilization
        percentage (e.g., 20.4% would be returned as 20.4)
        """
        return self._ceilometer_api.get_instance_ut(
            instance_obj, ego_constants.CEILOMETER_CPU_UT_KEY)

    def get_instance_cpu_range(self, instance_obj):
        """
        Retrieves the [min, max] values for the instance's virtual CPUs values.
        This is used in the context of VM right-sizing to determine the
        boundary values (inclusive).

        :param instance_obj: instance object (i.e., objects.Instance)
        :returns: a 2-tuple of floating point values representing the min and
        max values for virtual CPUs (inclusive)
        """
        return (1.0, float(ego_constants.MAX_INT_32))

    def get_instance_cpu_resize_min_delta(self, instance_obj):
        """
        Retrieves the smallest delta in virtual CPUs when performing a resize
        operation.  This is used in the context of VM right-sizing to determine
        the finest level of allowed granularity.

        :param instance_obj: instance object (i.e., objects.Instance)
        :returns: a floating point value representing the smallest virtual CPU
        delta allowed when performing a VM resize operation
        """
        return 1.0

    def get_instance_memory_util(self, instance_obj):
        """
        Retrieves the effective memory utilization percentage for the specified
        instance object.

        :param instance_obj: instance object (i.e., objects.Instance)
        :returns: a floating point representing the instance's memory
        utilization percentage (e.g., 20.4% would be returned as 20.4)
        """
        return self._ceilometer_api.get_instance_ut(
            instance_obj, ego_constants.CEILOMETER_MEM_UT_KEY)

    def get_instance_memory_range(self, instance_obj):
        """
        Retrieves the [min, max] values for the instance's virtual memory
        values (units are MBs).  This is used in the context of VM right-sizing
        to determine the boundary values (inclusive).

        :param instance_obj: instance object (i.e., objects.Instance)
        :returns: a 2-tuple of integer values representing the min and
        max values (in MBs) for virtual memory (inclusive)
        """
        return (1, ego_constants.MAX_INT_32)

    def get_instance_memory_resize_min_delta(self, instance_obj):
        """
        Retrieves the smallest delta in virtual memory (in MBs) when performing
        a resize operation.  This is used in the context of VM right-sizing to
        determine the finest level of allowed granularity.

        :param instance_obj: instance object (i.e., objects.Instance)
        :returns: an integer value representing the smallest virtual memory
        delta (in MBs) allowed when performing a VM resize operation
        """
        return 1

    def is_aggregate_optimization_enabled(self, context, agg_id, policy_info):
        """
        Determines whether optimization is enabled for the host aggregate.  If
        optimization is disabled, no optimization-related notifications will be
        generated for the respective aggregate or its child hosts or VMs.

        :param context: nova context object
        :param agg_id: host aggregate id
        :param policy_info: a dict containing all of the policy information,
                            including the aggregate ID, policy id, policy name,
                            policy goal, policy threshold, policy run interval,
                            policy stabilization and policy max migrations.
        :returns: True if enabled and False otherwise
        """
        return True

    def can_perform_aggregate_optimization(self, context, agg_id, policy_info):
        """
        Determines whether optimization will be performed on the host
        aggregate.  If it's disabled, then PRS will NOT evaluate any of
        the member hosts to determine if they're violating the runtime
        policy, and thus, will NEVER invoke the policy action (e.g., migrate)
        on the member hosts/VMs.

        If optimization is not permitted, the return value's tuple
        will include a reason (string) with further explanation.

        :param context: nova context object
        :param agg_id: host aggregate id
        :param policy_info: a dict containing all of the policy information,
                            including the aggregate ID, policy id, policy name,
                            policy goal, policy threshold, policy run interval,
                            policy stabilization and policy max migrations.
        :returns: (boolean, reason_string_if_false)
        """
        return (True, )

    def is_host_optimization_enabled(self, context, hv_name, policy_info):
        """
        Determines whether optimization is enabled for the host.  If
        optimization is disabled, no optimization-related notifications will be
        generated for the respective host or its child VMs.

        :param context: nova context object
        :param hv_name: name of a hypervisor
        :param policy_info: a dict containing all of the policy information,
                            including the aggregate ID, policy id, policy name,
                            policy goal, policy threshold, policy run interval,
                            policy stabilization and policy max migrations.
        :returns: True if enabled and False otherwise
        """
        return True

    def can_perform_host_optimization(self, context, hv_name, policy_info):
        """
        Determines whether optimization will be performed on the hypervisor.
        If optimization is not permitted, the return value's tuple will include
        a reason (string) with further explanation.

        :param context: nova context object
        :param hv_name: name of a hypervisor
        :param policy_info: a dict containing all of the policy information,
                            including the aggregate ID, policy id, policy name,
                            policy goal, policy threshold, policy run interval,
                            policy stabilization and policy max migrations.
        :returns: (boolean, reason_string_if_false)
        """
        return (True, )

    def is_instance_optimization_enabled(self, context, instance_obj,
                                         policy_info):
        """
        Determines whether optimization is enabled for the instance.  If
        optimization is disabled, no optimization-related notifications will be
        generated for the respective VM.

        :param context: nova context object
        :param instance_obj: instance object (i.e., objects.Instance)
        :param policy_info: a dict containing all of the policy information,
                            including the aggregate ID, policy id, policy name,
                            policy goal, policy threshold, policy run interval,
                            policy stabilization and policy max migrations.
        :returns: True if enabled and False otherwise
        """
        return True

    def can_perform_instance_optimization(self, context, instance_obj,
                                          policy_info):
        """
        Determines whether optimization will be performed on the instance.  If
        optimization is not permitted, the return value's tuple will include a
        reason (string) with further explanation.

        :param context: nova context object
        :param instance_obj: instance object (i.e., objects.Instance)
        :param policy_info: a dict containing all of the policy information,
                            including the aggregate ID, policy id, policy name,
                            policy goal, policy threshold, policy run interval,
                            policy stabilization and policy max migrations.
        :returns: (boolean, reason_string_if_false)
        """
        return (True, )

    def should_perform_instance_migration(self, context, instance_obj,
                                          instance_rank, num_instances,
                                          policy_info):
        """
        Determines whether an instance should be migrated during an
        optimization cycle.  By the time this is called, the instance has been
        deemed "OK" to be subjected to migration-based optimization, but this
        provides the ability to determine whether the migration is allowed.
        For example, if an instance isn't even contributing 1% CPU utilization
        to the overall host utilization, it is probably wise to bypass the
        migration.

        It is also implied that if this function returns False for an instance
        whose rank is x, then any instance whose rank is > x must also return
        False (i.e., this means that the migration optimizer will stop trying
        to migrate VMs once this returns False).

        :param context: nova context object
        :param instance_obj: instance object (i.e., objects.Instance)
        :param instance_rank: an integer between [1, n] where `n` represents
                              the number of instances on the host that are
                              subject to migration-based optimizations (note
                              that the instance whose rank is 1 is attempted
                              to be migrated first)
        :param num_instances: the number of instances on the host that are
                              subject to migration-based optimizations
        :param policy_info: a dict containing all of the policy information,
                            including the aggregate ID, policy id, policy name,
                            policy goal, policy threshold, policy run interval,
                            policy stabilization and policy max migrations.
        :returns: (boolean, reason_string_if_false)
        """
        return (True, )

    def sort_instances_for_optimization(self, context, instances, policy_info):
        """
        Performs an in-place sort of the instances once a violated host is
        selected for optimization.  The default order is based on the ongoing
        optimization policy (e.g., for cpu.util, the instances with the lowest
        CPU utilization will be migrated first).

        :param context: nova context object
        :param instances: a list of instances on the hypervisor that are
                          eligible for optimization (i.e.,
                          can_perform_instance_optimization(...) is True)
        :param policy_info: a dict containing all of the policy information,
                            including the aggregate ID, policy id, policy name,
                            policy goal, policy threshold, policy run interval,
                            policy stabilization and policy max migrations.
        """
        raise NotImplementedError()

    def handle_notification(self, context, event_type, payload):
        """
        Serves as a notification handler for all notifications that are
        generated by the optimization service.

        :param context: nova context object
        :param event_type: the notification's event type
        :param payload: the notification's payload
        :returns: a dict of additional information to add to the payload before
                  it is sent out via the message bus (or `None` if nothing)
        """
        return None

    def get_policy_actions(self):
        """
        Returns a list of policy actions that this driver supports.  If a user
        tries to set a policy action to a value not contained within this list,
        the resource optimization process will not run.

        :returns: a list of policy actions supported by this driver
        """
        return list(ego_constants.POLICY_ACTION_LIST)

    def get_advise_policy_actions(self):
        """
        Returns a list of policy actions that are considered to be "advise
        only", which means the optimizations will not actually be performed.
        The optimizers use this information to determine what action(s) to
        take during their execution.

        :returns: a list of policy actions that are considered to be "advise
                  only"
        """
        return [ego_constants.POLICY_ACTION_MIGRATE_VM_ADVISE_ONLY]
