# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import eventlet
import six

from nova import compute
from nova.compute import task_states
from nova.compute import vm_states
from nova import exception
from nova.ibm.ego import ego_constants
from nova.ibm.ego import ego_gettextutils
from nova import objects
from nova.objects import instance as instance_obj
from nova.scheduler.ibm.ego import ego_exception
from nova.scheduler.ibm.ego import ego_resreq_utils
from nova.scheduler.ibm.ego import ego_rpcapi
from nova.scheduler import utils as scheduler_utils
from nova import utils as nova_utils

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import timeutils

_ = ego_gettextutils._
_LI = ego_gettextutils._LI
_LW = ego_gettextutils._LW
_LE = ego_gettextutils._LE

CONF = cfg.CONF
CONF.import_opt('policy_per_host_aggregate', 'nova.ibm.ego.ego_flags')
CONF.import_opt('prs_rtpolicy_instance_recovery_time',
                'nova.ibm.ego.ego_flags')

LOG = logging.getLogger(__name__)


class RTPolicyOptimizer(object):
    """Abstract class for all optimizers."""

    def __init__(self, policy_engine_manager, rtpolicy_driver):
        self._policy_engine_manager = policy_engine_manager
        self._rtpolicy_driver = rtpolicy_driver
        self._priority = ego_constants.OPTIMIZER_DEFAULT_PRIORITY

    @property
    def priority(self):
        return self._priority

    def receive_notification(self, context, msg):
        """Notification handler."""
        raise NotImplementedError()

    def reset_optimizer(self):
        """Reset the optimizer (e.g., when enabled or disabled)."""
        raise NotImplementedError()

    def run_ticker(self, context):
        """Invoked once per minute for general tasks."""
        raise NotImplementedError()

    def run_interval_ticker(self, context, policy_info):
        """
        Invoked every <run interval> minutes to perform general tasks (e.g.,
        any necessary housekeeping tasks, etc.) when optimization is enabled.
        """
        raise NotImplementedError()

    def optimize_hosts(self, context, violated_hosts, already_optimized_hosts,
                       policy_info):
        """
        Performs optimization on the violated hosts.

        :param context: nova context object
        :param violated_hosts: the list of hosts violating their optimization
                               policy (that the rtpolicy driver's method
                               can_perform_host_optimization() also deemed
                               acceptable for optimization)
        :param already_optimized_hosts: the list of hosts that have already
                                        been optimized during this cycle or
                                        [] if nothing has been optimized
        :param policy_info: a dict containing all of the policy information,
                            including the aggregate ID, policy id, policy name,
                            policy goal, policy threshold, policy run interval,
                            policy stabilization and policy max migrations.
        :return: a recommendation to the optimization orchestrator on the next
                 set of suggested steps; represented by a dict of the form:
                 {'code': int,
                  'optimized_hosts': list of hypervisor hostnames}

                where dict['code'] must be one of:

                 OPTIMIZE_HOSTS_PERFORMED_WAIT - performed optimization, but
                 no further optimizations should be performed this cycle

                 OPTIMIZE_HOSTS_PERFORMED_CONTINUE - performed optimization
                 and further optimizations should be performed

                 OPTIMIZE_HOSTS_NOT_PERFORMED_WAIT - no optimization was
                 performed and no further optimizations should be performed
                 this cycle

                 OPTIMIZE_HOSTS_NOT_PERFORMED_CONTINUE - no optimization was
                 performed and further optimizations should be performed

                and dict['optimized_hosts'] is a list of hypervisor hostnames
                that were optimized by this optimizer during this cycle
        """
        raise NotImplementedError()


class RTPolicyMigrationOptimizer(RTPolicyOptimizer):
    """Implementation of the migration optimizer."""
    def __init__(self, *args, **kwargs):
        super(RTPolicyMigrationOptimizer, self).__init__(*args, **kwargs)
        # Represents the list of (host, inst_uuid) tuples that are currently in
        # use for an optimization cycle (where the 'host' represents a
        # destination target and 'inst_uuid' represents the instance moving).
        # Entries are deleted once the instance is no longer migrating.
        # Format: [(compute_host_name, instance_id)]
        self._used_hosts = []
        # Format: {request_id: {hypervisor_id: xx, hypervisor_hostname: xx,
        #                       rtpolicy_id: xx, aggregate_id: xx,
        #                       action_name: xx, vm_name: xx, vm_uuid: xx,
        #                       target_host: xx}}
        self._rtpolicy_action = {}
        # Migration failed instances metadata
        self._failed_instances = {}
        self._compute_api = compute.API()
        self._scheduler_api = ego_rpcapi.EGOSchedulerAPI()

    ####################################################
    #                 Internal Methods                 #
    ####################################################

    def __check_optimization_host_prereqs(self, context, ego_host):
        """
        Check any prerequisites required to be satisfied prior to allowing
        ongoing optimization to occur on the hypervisor.

        :param context: nova context object
        :param ego_host: a hypervisor's EGO resource name
        :return: True if optimization is allowed and False if not allowed
        """

        host_dict = self._policy_engine_manager.ego_host_dict.get(ego_host)
        # 1. Ignore the violated host that has been deleted from the system
        if not host_dict:
            LOG.info(_LI("Ignoring violated host (%s) because it has been "
                         "deleted from the management system.") % ego_host)
            return False

        # 2. Ignore the violated host who is the target host in prior cycle
        if ego_host in dict(self._used_hosts):
            LOG.info(_LI("Ignoring violated host (%s) because it's still "
                         "being used as part of a previous optimization "
                         "cycle (i.e., a migration is still ongoing).")
                     % ego_host)
            return False

        # 3. Ensure we don't issue too many migrations for the host
        svc_host = host_dict.get('svc_host')
        host_migrate_num = len(self.__get_instance_by_task_state(
            context, [task_states.MIGRATING], svc_host=svc_host))
        max_migrations = host_dict['max_migrations']
        if host_migrate_num >= max_migrations:
            # At max # of migrations; try the next violated host
            LOG.debug("The source host (%(host)s) is performing its "
                      "maximum number of live migrations "
                      "(%(migrate_num)d); skipping host."
                      % {'host': ego_host,
                         'migrate_num': host_migrate_num})
            return False

        return True

    def __get_instance_load(self, instance):
        """Return instance load according to metrics type."""
        if self._aggregate_goal_for_sort.find('vcpu') != -1:
            ego_host = ego_resreq_utils.get_rpi_res_name(instance['node'])
            ego_host_dict = self._policy_engine_manager.ego_host_dict
            htype = ego_host_dict.get(ego_host, {}).get('htype')
            return ego_resreq_utils.get_instance_vcpus(instance, htype)
        elif self._aggregate_goal_for_sort.find('memSize') != -1:
            return instance['memory_mb']
        elif self._aggregate_goal_for_sort.find(
                ego_constants.CPU_UTILIZATION_METRICS_NAME) != -1:
            return self._rtpolicy_driver.get_instance_cpu_util(instance)
        elif self._aggregate_goal_for_sort.find(
                ego_constants.MEM_UTILIZATION_METRICS_NAME) != -1:
            return self._rtpolicy_driver.get_instance_memory_util(instance)
        else:
            return 0

    def __compare_instance_load(self, instance1, instance2):
        """Comparator used to order instances for migration priority."""
        inst_load_1 = self.__get_instance_load(instance1)
        inst_load_2 = self.__get_instance_load(instance2)

        if inst_load_1 < inst_load_2:
            return -1
        elif inst_load_1 > inst_load_2:
            return 1
        else:
            return 0

    def __get_host_cpu_freqency(self, context, hostname):
        """
        Get a host's CPU frequency from compute node metrics.

        :param context: nova context
        :param hostname: which host's frequency will be returned
        :return: the host's CPU frequency as retrieved from the DB
        """
        value = 0
        computes = objects.ComputeNodeList.get_all_by_host(context, hostname)
        metrics = computes[0]['metrics']
        metrics = jsonutils.loads(metrics)
        for metric in metrics:
            if metric['name'] == ego_constants.HOST_CPU_FREQUENCY_METRIC_KEY:
                value = metric['value']
                break
        return value

    def __get_instance_by_task_state(self, context, task_state, svc_host=None):
        """
        Get the specified instances based on the task state and the optionally
        specified service host.
        """
        try:
            filters = {'task_state': task_state, 'deleted': 0}
            # If a host was specified, create the filter entry
            if svc_host:
                filters['host'] = svc_host
            return objects.InstanceList.get_by_filters(context, filters)
        except Exception as exc:
            LOG.error(_LE("Could not retrieve the instances due to: %s")
                      % six.text_type(exc))
            return []

    def __get_instance_by_task_state_based_on_aggregate(self, context,
                                                        task_state,
                                                        aggregate_id):
        """
        Retrieve all instances first and then filter out the instances that
        aren't a member of the host aggregate.
        """
        all_instances = self.__get_instance_by_task_state(context, task_state)

        if CONF.policy_per_host_aggregate:
            agg_container = self._policy_engine_manager.aggregate_container
            instances = [ins for ins in all_instances
                         if aggregate_id in
                         agg_container.host_aggregates.get(ins['host'], {})]
        else:
            # For global policy mode, just return all instances
            instances = all_instances
        return instances

    def __get_instances_for_compute_node(self, context, ego_host, policy_info):
        """
        Get migratable instances (i.e., active and not performing any tasks)
        of hosts from the database; results are sorted by the order in which
        migrations should be attempted.

        self._aggregate_goal_for_sort MUST be set before sorting instances.
        """
        host_dict = self._policy_engine_manager.ego_host_dict.get(ego_host)
        if not host_dict:
            return []

        instances = objects.InstanceList.get_by_host(
                            context, host_dict['svc_host'],
                            expected_attrs=['metadata', 'system_metadata'])
        # Filter instances that are on the proper compute node that are also
        # "migratable" (i.e., active and not performing any tasks)
        hv_insts = []
        for inst in instances:
            result = self._rtpolicy_driver.is_instance_optimization_enabled(
                context, inst, policy_info)
            if not result:
                LOG.debug("Instance %(name)s (%(inst_id)s) was skipped "
                          "because its optimization has been disabled."
                          % {'name': inst['display_name'],
                             'inst_id': inst['uuid']})
                continue
            elif inst['node'] != host_dict['node_name']:
                LOG.info(_LI("Instance %(name)s (%(inst_id)s) was skipped "
                             "because its node (%(node)s) does not match the "
                             "current PRS node (%(prs_node)s).")
                         % {'name': inst['display_name'],
                            'inst_id': inst['uuid'],
                            'node': inst['node'],
                            'prs_node': host_dict['node_name']})
                continue
            elif inst['vm_state'] != vm_states.ACTIVE:
                LOG.info(_LI("Instance %(name)s (%(inst_id)s) was skipped "
                             "because its VM state (%(state)s) is not active.")
                         % {'name': inst['display_name'],
                            'inst_id': inst['uuid'],
                            'state': inst['vm_state']})
                continue
            elif inst['task_state']:
                LOG.info(_LI("Instance %(name)s (%(inst_id)s) was skipped "
                             "because it is performing a task (%(task)s).")
                         % {'name': inst['display_name'],
                            'inst_id': inst['uuid'],
                            'task': inst['task_state']})
                continue
            else:
                # Check if the instance previously failed to migrate
                metadata = self._failed_instances.get(inst['uuid'], {})
                failed_at = metadata.get('prs_rtpolicy_migr_failed_tstamp')
                # Check to see if sufficient recovery time has elapsed
                if failed_at:
                    failed_at_time = timeutils.parse_strtime(failed_at)
                    now = timeutils.utcnow()
                    diff = timeutils.delta_seconds(failed_at_time, now)
                    recovery_time = CONF.prs_rtpolicy_instance_recovery_time
                    if diff < recovery_time:
                        LOG.info(_LI("Instance %(name)s (%(inst_id)s) was "
                                     "skipped because it previously "
                                     "failed to migrate. The elapsed "
                                     "time since the previous failure "
                                     "is %(diff)s seconds and does not "
                                     "exceed the recovery time (%(rec_time)s "
                                     "seconds).")
                                 % {'name': inst['display_name'],
                                    'inst_id': inst['uuid'],
                                    'diff': diff,
                                    'rec_time': recovery_time})
                        continue
            result = self._rtpolicy_driver.can_perform_instance_optimization(
                context, inst, policy_info)
            if not result[0]:
                if len(result) == 2:
                    # i.e., the extension provided its own message
                    reason = result[1]
                else:
                    reason = (("The policy driver prevented instance %(name)s "
                               "(%(uuid)s) from being considered for ongoing "
                               "optimization.")
                              % {'name': inst['display_name'],
                                 'uuid': inst['uuid']})
                LOG.debug(reason)
                continue
            hv_insts.append(inst)

        try:
            self._rtpolicy_driver.sort_instances_for_optimization(
                context, hv_insts, policy_info)
        except NotImplementedError:
            # Fallback to the default sorting algorithm
            hv_insts.sort(self.__compare_instance_load)
        return hv_insts

    def __get_possible_host(self, context, goal, instance, threshold,
                            aggregate_id, violated_hosts):
        """Get a list of hosts from the scheduler and send through filters."""
        possible_host_list = []
        try:
            goal = self.__transform_policy_goal(context, goal, instance)
            system_metadata = nova_utils.instance_sys_meta(instance)
            dec_id = ego_resreq_utils.get_decision_id(instance)
            image = nova_utils.get_image_from_system_metadata(system_metadata)
            request_spec = scheduler_utils.build_request_spec(
                context, image, [instance])

            request_spec['operation_type'] = (
                ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION)
            request_spec['invoked_by_rtpolicy'] = True
            filter_properties = {'request_spec': request_spec}
            # Possible hosts are sorted by EGO automatically
            possible_host_list = (
                self._scheduler_api.ego_get_possible_hosts(
                    context, dec_id, [aggregate_id],
                    [goal], filter_properties))
        except ego_exception.EGOExpGetPossibleHostsFailed as exc:
            LOG.error(_LE("The optimizer could not retrieve the host for "
                          "instance %(name)s (%(uuid)s) due to: %(reason)s")
                      % {'name': instance['display_name'],
                         'uuid': instance['uuid'],
                         'reason': six.text_type(exc)})
            raise
        except (ego_exception.EGOExpConnLost,
                ego_exception.EGOExpNoAllocID,
                ego_exception.EGOExpGetPossibleHostsFailed) as exc:
            LOG.error(_LE("The optimizer could not find a suitable host for "
                          "instance %(name)s (%(uuid)s) due to: %(reason)s")
                      % {'name': instance['display_name'],
                         'uuid': instance['uuid'],
                         'reason': six.text_type(exc)})
            return None

        if len(possible_host_list) == 0:
            LOG.info(_LI("The optimizer could not find a suitable host for "
                         "instance %(name)s (%(uuid)s) because the cluster "
                         "does not have enough resource capacity or there "
                         "is no suitable host.")
                     % {'name': instance['display_name'],
                        'uuid': instance['uuid']})
            return None
        LOG.debug("The optimizer found candidate host(s) for instance "
                  "%(name)s (%(uuid)s): %(host)s"
                  % {'name': instance['display_name'],
                     'uuid': instance['uuid'],
                     'host': possible_host_list})

        possible_host = None

        if goal != ego_constants.PACKING_POLICY_GOAL:
            used_hosts_list = dict(self._used_hosts).keys()
            for host in possible_host_list:
                if host[0] not in used_hosts_list:
                    possible_host = host
                    break
        else:
            # For the packing policy, there is no need to check the possible
            # host list because we always want the migrations going to the
            # same host.
            possible_host = possible_host_list[0]

        if not possible_host:
            LOG.info(_LI("The optimizer could not find a suitable host "
                         "for instance %(name)s (%(uuid)s) because all of "
                         "the candidate host(s) were selected for past "
                         "optimization cycles (and still ongoing).")
                     % {'name': instance['display_name'],
                        'uuid': instance['uuid']})
            return None
        else:
            LOG.debug("The optimizer compared the possible hosts "
                      "(%(possible_host_list)s) and the used hosts "
                      "(%(used_hosts)s) and found this host to use: "
                      "%(possible_host)s"
                      % {'possible_host_list': possible_host_list,
                         'used_hosts': self._used_hosts,
                         'possible_host': possible_host})

        if self.__validate_possible_host(possible_host, instance, threshold,
                                         goal, violated_hosts):
            LOG.debug("The host (%(possible_host)s) is an acceptable "
                      "migration target for instance %(name)s (%(uuid)s)."
                      % {'possible_host': possible_host,
                         'name': instance['display_name'],
                         'uuid': instance['uuid']})
            return possible_host
        else:
            LOG.info(_LI("The host (%(possible_host)s) is not suitable for "
                         "migrating instance %(name)s (%(uuid)s) because it "
                         "will cause the load on %(possible_host)s to exceed "
                         "its threshold.")
                     % {'possible_host': possible_host,
                        'name': instance['display_name'],
                        'uuid': instance['uuid']})
            return None

    def __validate_possible_host(self, possible_host, instance, threshold,
                                 goal, violated_hosts):
        LOG.debug("Checking the available host's (%(possible_host)s) load "
                  "to verify whether it exceeds the threshold "
                  "(%(threshold)d) for instance %(name)s (%(uuid)s)."
                  % {'possible_host': possible_host,
                     'threshold': threshold,
                     'name': instance['display_name'],
                     'uuid': instance['uuid']})
        if goal == ego_constants.PACKING_POLICY_GOAL:
            # For packing policy, the possible host should have more instances
            # than source host, so it should behind the source host in this
            # violated host list
            hosts_dict = dict(violated_hosts)
            if possible_host[0] in hosts_dict:
                node_name = instance['node']
                violated_host = ego_resreq_utils.get_rpi_res_name(node_name)
                return hosts_dict[possible_host[0]] > hosts_dict[violated_host]
            else:
                return True
        else:
            return possible_host[1] <= threshold

    def __migrate_vm(self, context, instance, ego_host, payload):
        """Start performing live migration of a particular instance."""
        inst = objects.Instance._from_db_object(
            context, objects.Instance(), instance,
            expected_attrs=instance_obj.INSTANCE_DEFAULT_FIELDS)
        host_dict = self._policy_engine_manager.ego_host_dict.get(ego_host)
        if not host_dict:
            LOG.warning(_LW("The selected target host (%(ego_host)s) is "
                            "currently unavailable in the DB; cannot live "
                            "migrate instance %(name)s (%(uuid)s).")
                        % {'ego_host': ego_host,
                           'name': inst['display_name'],
                           'uuid': inst['uuid']})
            raise exception.ComputeHostNotFound(host=ego_host)
        self._policy_engine_manager.store_event(
            context, ego_constants.PRS_RTPOLICY_ACTION_START, payload)
        self._rtpolicy_action[context.request_id] = payload
        self._compute_api.live_migrate(
            context.elevated(), inst, False, False, host_dict['svc_host'])

    def __transform_policy_goal(self, context, goal, instance):
        """
        As utilization policy's goal should consider the VM utilization.
        So transform the goal here and add the VM utilization from Ceilometer.

        :param goal: The primitive key
        :param instance: instance to optimize
        :return: Transformed goal
        """
        if goal == ego_constants.CPU_UTILIZATION_METRICS_NAME:
            instance_cpu_ut = 0.0
            try:
                instance_cpu_ut = self._rtpolicy_driver.get_instance_cpu_util(
                    instance)
            except Exception as e:
                LOG.warning(_LW("The optimizer failed to transform the policy "
                                "goal (%(goal)s) for instance %(name)s "
                                "(%(uuid)s) due to: %(reason)s")
                            % {'goal': goal, 'name': instance['display_name'],
                               'uuid': instance['uuid'],
                               'reason': six.text_type(e)})
                raise
            # Ceilometer will return 20 if the real utilization is 20%
            instance_cpu_ut = instance_cpu_ut / ego_constants.FIXED_UT_MULTIPLE
            ego_host = ego_resreq_utils.get_rpi_res_name(instance['node'])
            htype = self._policy_engine_manager.ego_host_dict.get(
                ego_host, {}).get('htype')
            instance_total_vcpus = ego_resreq_utils.get_instance_vcpus(
                instance, htype)
            # For CPU utilization, transform the goal:
            # goal = dest[cpu.percent] + VMut * instance['vcpus'] *
            # source[cpu.frequency] / dest[vcpus] * dest[cpu.frequency]
            # ego_get_possible_hosts return host which cpu.percent
            # is the minimum host
            source_cpu_frequency = self.__get_host_cpu_freqency(
                context, instance['host'])
            instance_total_cpu_frequency = (instance_cpu_ut *
                                            source_cpu_frequency *
                                            instance_total_vcpus)
            LOG.debug("Forming goal for (%(host)s, %(inst)s): instance util="
                      "%(instance_ut)s; source CPU freq=%(freq)s; "
                      "instance vcpus=%(vcpus)s" %
                      {'host': instance['host'],
                       'inst': instance['display_name'],
                       'instance_ut': instance_cpu_ut,
                       'freq': source_cpu_frequency,
                       'vcpus': instance_total_vcpus})
            instance_total_cpu_frequency = format(instance_total_cpu_frequency,
                                                  '.16f')
            goal = "".join(["(", ego_constants.CPU_UTILIZATION_METRICS_NAME,
                            "+", str(instance_total_cpu_frequency),
                           "/(vcpus*cpu.frequency))*100"])
        elif goal == ego_constants.MEM_UTILIZATION_METRICS_NAME:
            instance_mem_ut = 0.0
            try:
                instance_mem_ut = (
                    self._rtpolicy_driver.get_instance_memory_util(instance))
            except Exception as e:
                LOG.warning(_LW("The optimizer failed to transform the policy "
                                "goal (%(goal)s) for instance %(name)s "
                                "(%(uuid)s) due to: %(reason)s")
                            % {'goal': goal, 'name': instance['display_name'],
                               'uuid': instance['uuid'],
                               'reason': six.text_type(e)})
                raise
            instance_mem_ut = instance_mem_ut / ego_constants.FIXED_UT_MULTIPLE
            instance_total_memory = instance['memory_mb']
            # For mem_ut, transform the goal:
            # goal = mem_ut + VMut * instance[memory] / host[memory]
            # ego_get_possible_hosts return host which mem_ut is the minimum.
            goal = "".join(["(", ego_constants.MEM_UTILIZATION_METRICS_NAME,
                            "+", str(instance_mem_ut), "*",
                           str(instance_total_memory), "/memory_mb)*100"])
        LOG.debug("The policy goal after transformation is: %s" % goal)
        return goal

    ####################################################
    #          Abstract Class Implementation           #
    ####################################################

    def receive_notification(self, context, msg):
        event_type = msg['event_type']
        payload = msg['payload']
        instance_id = payload.get('instance_id')

        # Handle VM notifications for tracking whether the live migrations
        # completed successfully

        if event_type in (
            [ego_constants.EVENT_TYPE_COMPUTE_INST_ROCKBACK_LIVE_MIG_END,
             ego_constants.EVENT_TYPE_COMPUTE_INST_LIVE_MIG_ROLLBACK_END,
             ego_constants.EVENT_TYPE_CHECK_MIGRATE_DEST]):
            # Only tag the rtpolicy-triggered migrations that failed
            request_id = str(context.request_id)
            if request_id.startswith(ego_constants.RTPOLICY_PREFFIX):
                now = timeutils.strtime()
                metadata = {"prs_rtpolicy_migr_failed_tstamp": now}
                if instance_id not in self._failed_instances:
                    self._failed_instances[instance_id] = metadata

            action_info = self._rtpolicy_action.pop(context.request_id, None)
            if action_info:
                action_info['action_result'] = 'failed'
                self._policy_engine_manager.store_event(context,
                                 ego_constants.PRS_RTPOLICY_ACTION_END,
                                 action_info)
        elif event_type == (
             ego_constants.EVENT_TYPE_COMPUTE_INST_LIVE_MIGRATION_POST_END):
            action_info = self._rtpolicy_action.pop(context.request_id, None)
            if action_info:
                action_info['action_result'] = 'success'
                self._policy_engine_manager.store_event(context,
                                 ego_constants.PRS_RTPOLICY_ACTION_END,
                                 action_info)

    def reset_optimizer(self):
        LOG.info(_LI("Resetting the optimizer."))
        self._used_hosts = []
        self._rtpolicy_action = {}
        self._failed_instances = {}

    def run_ticker(self, context):
        """Clear the 'expired' failed migration instances."""
        for key, metadata in self._failed_instances.items():
            failed_at = metadata.get('prs_rtpolicy_migr_failed_tstamp')
            failed_at_time = timeutils.parse_strtime(failed_at)
            now = timeutils.utcnow()
            diff = timeutils.delta_seconds(failed_at_time, now)
            recovery_time = CONF.prs_rtpolicy_instance_recovery_time
            if diff > recovery_time:
                self._failed_instances.pop(key)

    def run_interval_ticker(self, context, policy_info):
        """Remove entries from 'used_hosts' for completed migrations."""
        migr_insts = self.__get_instance_by_task_state(context,
                                                       [task_states.MIGRATING])
        migrating_instance_ids = [instance['uuid'] for instance in migr_insts]
        used_hosts_prime = []
        for host_instance in self._used_hosts:
            if host_instance[1] not in migrating_instance_ids:
                LOG.debug("Removing %s from the 'used_hosts' list because its "
                          "migration completed." %
                          six.text_type(host_instance))
            else:
                used_hosts_prime.append(host_instance)
        self._used_hosts = used_hosts_prime
        LOG.debug("The following hosts are being used for migrations and are "
                  "unavailable: %s" % self._used_hosts)

    def optimize_hosts(self, context, violated_hosts, already_optimized_hosts,
                       policy_info):
        """Optimize the hosts via live migrations."""
        LOG.debug("The following hosts have been optimized this cycle and "
                  "will not undergo migration-based optimization: %s"
                  % six.text_type(already_optimized_hosts))
        # Do not consider the already optimized hosts
        violated_hosts = [host for host in violated_hosts
                          if host[0] not in already_optimized_hosts]

        goal = policy_info['policy_goal']
        agg_id = policy_info['aggregate_id']
        max_migrations = policy_info['policy_max_migrations']
        threshold = policy_info['policy_threshold']
        policy_action = policy_info['policy_action']
        ret = dict(code=ego_constants.OPTIMIZE_HOSTS_NOT_PERFORMED_CONTINUE,
                   optimized_hosts=[])

        # This member variable is used for instance sort only
        self._aggregate_goal_for_sort = goal
        # Query which actions are "advise only" actions
        advise_actions = self._rtpolicy_driver.get_advise_policy_actions()

        # Return early if there are no hosts in the aggregate that are not
        # violated
        if len(self._get_nonviolated_hosts(context, agg_id)) == 0:
            LOG.info(_LI("The optimizer cannot optimize any of the violated "
                         "hosts (%(violated_hosts)s) because there are no "
                         "potential hosts to serve as a destination host for "
                         "a migration. All of the hosts in the aggregate "
                         "(%(agg_id)s) are violated.")
                     % {'violated_hosts': violated_hosts,
                        'agg_id': agg_id})
            return ret

        for host in violated_hosts:
            # Ensure the host passes optimization prerequisites
            if not self.__check_optimization_host_prereqs(context, host[0]):
                continue

            # Ensure we don't issue too many cluster-wide migrations
            agg_migr_num = len(
                self.__get_instance_by_task_state_based_on_aggregate(
                    context, [task_states.MIGRATING], agg_id))
            if agg_migr_num >= max_migrations:
                LOG.info(_LI("The current number of aggregate-wide migrations "
                             "(%(migrating_num)s) for aggregate (%(agg_id)s) "
                             "has reached or exceeded its maximum allowed "
                             "value (%(max_parallel)s); additional migrations "
                             "for this aggregate will be bypassed until the "
                             "number of active migrations decreases below the "
                             "threshold.")
                         % {'migrating_num': agg_migr_num,
                            'agg_id': agg_id,
                            'max_parallel': max_migrations})
                break

            best_target_host = None
            best_target_instance = None
            instances = self.__get_instances_for_compute_node(
                context, host[0], policy_info)
            inst_rank = 0
            num_instances = len(instances)

            # Start trying to find a VM to migrate
            for inst in instances:
                inst_rank += 1
                # Check whether the VM should be migrated
                res = self._rtpolicy_driver.should_perform_instance_migration(
                    context, inst, inst_rank, num_instances, policy_info)
                if not res[0]:
                    # The VM should not be migrated; halt optimization for
                    # this host
                    if len(res) == 2:
                        # i.e., the extension provided its own message
                        reason = res[1]
                    else:
                        reason = (("The optimizer determined that instance "
                                   "%(name)s (%(uuid)s) with rank %(rank)d/"
                                   "%(total)d should not be migrated; halting "
                                   "optimization for host %(host)s and no "
                                   "additional instances will be attempted to "
                                   "be migrated.") %
                                  {'name': inst.display_name,
                                   'uuid': inst.uuid,
                                   'rank': inst_rank,
                                   'total': num_instances,
                                   'host': host[0]})
                    LOG.debug(reason)
                    break

                # best_possible_host is of the form [host_name,
                # evaluatedGoal_value]
                best_possible_host = self.__get_possible_host(context, goal,
                                                              inst, threshold,
                                                              agg_id,
                                                              violated_hosts)
                if best_possible_host:
                    best_target_instance = inst
                    best_target_host = best_possible_host[0]
                    break

            if best_target_host and best_target_instance:
                try:
                    context.request_id = (ego_constants.RTPOLICY_PREFFIX +
                                          context.request_id)
                    ego_host_dict = self._policy_engine_manager.ego_host_dict
                    alert_times = (self._policy_engine_manager.
                                   get_violation_count(host[0], agg_id))
                    host_dict = ego_host_dict[host[0]]
                    payload = {'vm_name': best_target_instance['display_name'],
                               'vm_uuid': best_target_instance['uuid'],
                               'target_host': ego_host_dict.get(
                                        best_target_host, {}).get('svc_host'),
                               'action_name': str(policy_action),
                               'hypervisor_id': host_dict['node_id'],
                               'hypervisor_hostname': host_dict['node_name'],
                               'target_host_evaluated_goal':
                                            best_possible_host[1],
                               'policy_type': 'rtpolicy',
                               'rtpolicy_id': policy_info['policy_id'],
                               'rtpolicy_name': policy_info['policy_name'],
                               'threshold': policy_info['policy_threshold'],
                               'stabilization':
                                  policy_info['policy_stabilization'],
                               'run_interval':
                                  policy_info['policy_run_interval'],
                               'is_violated': 'Yes',
                               'alert_times': alert_times}
                    if CONF.policy_per_host_aggregate:
                        payload['aggregate_id'] = agg_id
                    # An optimization was performed (or advised) on the host
                    ret['code'] = (
                        ego_constants.OPTIMIZE_HOSTS_PERFORMED_CONTINUE)
                    ret['optimized_hosts'].append(host[0])
                    if policy_action in advise_actions:
                        # We're only advising what to do; omit migration
                        LOG.info(_LI("The optimizer is in %(advise)s mode; "
                                     "instance %(dn)s (%(uuid)s) is advised "
                                     "to be live migrated from %(ori_host)s "
                                     "to %(tar_host)s.")
                                 % {'advise': policy_action,
                                    'dn': best_target_instance['display_name'],
                                    'uuid': best_target_instance['uuid'],
                                    'ori_host': host[0],
                                    'tar_host': best_target_host})
                        # Now generate notifications for advise only mode
                        self._policy_engine_manager.store_event(context,
                                    ego_constants.PRS_RTPOLICY_ACTION_START,
                                    payload)
                        # Always assume success as nothing can go wrong
                        payload['action_result'] = 'success'
                        self._policy_engine_manager.store_event(context,
                                     ego_constants.PRS_RTPOLICY_ACTION_END,
                                     payload)
                        # Move on to the next violated host
                        continue
                    # Otherwise, migration was requested
                    self.__migrate_vm(context,
                                      best_target_instance,
                                      best_target_host, payload)
                    self._used_hosts.append((best_target_host,
                                             best_target_instance['uuid']))
                    # Allow EGO to migrate the alloc from the previous instance
                    eventlet.sleep(2)
                    LOG.info(_LI("The optimizer successfully initiated live "
                                 "migration of instance %(name)s (%(uuid)s) "
                                 "from %(ori_host)s to %(tar_host)s.")
                             % {'name': best_target_instance['display_name'],
                                'uuid': best_target_instance['uuid'],
                                'ori_host': host[0],
                                'tar_host': best_target_host})
                except exception.ComputeHostNotFound:
                    pass
                except Exception as exc:
                    action_info = self._rtpolicy_action.pop(context.request_id,
                                                            None)
                    if action_info:
                        action_info['action_result'] = 'failed'
                        self._policy_engine_manager.store_event(context,
                                     ego_constants.PRS_RTPOLICY_ACTION_END,
                                     action_info)
                    LOG.error(_LE("The optimizer could not live migrate "
                                  "instance %(name)s (%(uuid)s) from "
                                  "%(ori_host)s to %(tar_host)s due to: "
                                  "%(reason)s")
                              % {'name': best_target_instance['display_name'],
                                 'uuid': best_target_instance['uuid'],
                                 'ori_host': host[0],
                                 'tar_host': best_target_host,
                                 'reason': six.text_type(exc)})
        return ret

    def _get_nonviolated_hosts(self, context, agg_id=None):
        """
        Returns all hosts which are not deemed violated by the policy engine
        manager.

        :param context: Nova context object
        :param agg_id: optional parameter for the aggregate ID
        :return: list of host names in the aggregate that are not violated
        """
        search_options = {}
        if CONF.policy_per_host_aggregate and agg_id:
            search_options = {'aggregate_id': agg_id}

        hosts = self._policy_engine_manager.hypervisor_list(
            context, search_options)

        nonviolated_hosts = [
            host['hypervisor_hostname'] for host in hosts if
            str(host['is_violated']).lower() == 'no'
        ]
        return nonviolated_hosts
