# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from collections import defaultdict


class RTPolicyResourceGroupStatus(object):

    """Helper class to maintain the status of a group of rtpolicy resources"""

    def __init__(self, run_interval, stabilization, group_id):
        # Run interval associated with the RT policy for this group
        self.run_interval = int(run_interval)
        # Stabilization associated with the RT policy for this group
        self.stabilization = int(stabilization)

        # Current interval in reference to the run interval. In practice this
        # loops between [1, self.run_interval] for the policy engine manager
        self._current_interval = 1

        # Dictionary with format: {resource: number of consecutive times the
        #                                    resource has been violated} where
        # 'resource' is an identifier to the resource, such as an EGO hostname
        # or a VM UUID
        self._violation_counts = defaultdict(int)

        # ID of this group of resources, such as an aggregate ID
        self.group_id = str(group_id)

    @property
    def violation_counts(self):
        """Returns the violation counts for all resources"""
        return self._violation_counts

    def increment_violation_count(self, resource):
        self._violation_counts[resource] += 1

    def reset_violation_count(self, resource):
        """Resets the consecutive violation count for a specific resource"""
        if resource in self._violation_counts:
            del self._violation_counts[resource]

    def reset_violation_counts(self):
        """Resets the consecutive violation count for all resources"""
        self._violation_counts.clear()

    def get_violation_count(self, resource):
        """
        Returns the consecutive number of times a resource has been violated
        """
        return self._violation_counts[resource]

    @property
    def alarming_resources(self):
        """
        Returns a list of resources that have been violated at least once
        """
        return [resource for resource in self._violation_counts if
                self._violation_counts[resource] > 0]

    @property
    def stably_violated_resources(self):
        """
        Returns a list of resources that have been violated <stabilization>
        number of times consecutively and are therefore considered 'stably'
        violated
        """
        return [resource for resource in self._violation_counts if
                self.stably_violated(resource)]

    def stably_violated(self, resource):
        """
        Returns True if the resource is stably violated, meaning it has been
        violated <stabilization> number of times consecutively, and False
        otherwise
        """
        return self._violation_counts[resource] >= self.stabilization

    @property
    def current_interval(self):
        """Returns the current interval in reference to the run interval"""
        return self._current_interval

    def increment_current_interval(self):
        """
        Increments the current interval by one, and if the current interval is
        greater than run_interval, then reset the current interval to 1
        """
        self._current_interval += 1
        if self._current_interval > self.run_interval:
            self._current_interval = 1

    def reset_current_interval(self):
        """Resets the current interval back to 1"""
        self._current_interval = 1

    def current_interval_lt_run_interval(self):
        """
        Returns True if the current interval is less than run_interval, and
        False otherwise.
        """
        return self._current_interval < self.run_interval
