# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Base engine."""

import eventlet
import six

from nova import compute
from nova.compute import task_states
from nova import objects

from oslo_config import cfg
from oslo_log import log as logging

from nova.ibm.ego import ego_constants
from nova.ibm.ego import ego_gettextutils
from nova.ibm.ego.slapolicy import constants as sla_constants
from nova.scheduler.ibm.ego import ego_rpcapi


_ = ego_gettextutils._
_LI = ego_gettextutils._LI
_LW = ego_gettextutils._LW
_LE = ego_gettextutils._LE

CONF = cfg.CONF
CONF.import_opt('prs_slapolicy_instance_recovery_time',
                'nova.ibm.ego.ego_flags')
CONF.import_opt('prs_slapolicy_instance_failed_times',
                'nova.ibm.ego.ego_flags')

LOG = logging.getLogger(__name__)


class BasePolicyEngine(object):
    """Basic class for all engines."""

    def __init__(self, policy_engine_manager):
        self.ego_host_dict = {}
        self._scheduler_api = ego_rpcapi.EGOSchedulerAPI()
        self._compute_api = compute.API()
        self._policy_engine_manager = policy_engine_manager

    ####################################################
    #                    Class Methods                 #
    ####################################################
    @classmethod
    def get_hosts_filter_resreq(cls):
        """
        Extend method for get the specific hosts and its info from EGO.
        :param cls: class method default input
        :return return all the active host resreq.
        """

        return '(($activeCompute==1)&&($maintenance_status==ok))'

    @classmethod
    def get_optimization_actions(cls):
        """
        Extend method for get the specific actions for optimization.
        :param cls: class method default input
        :return return all the actions for optimization.
        """

        return ['migrate_vm']

    ####################################################
    #             Methods need to implement            #
    ####################################################

    def get_violated_hosts(self, context,
                           aggregate_id, policy, *args, **kwargs):
        """
        Extend method for getting a list of violated hosts from EGO.
        :param context: nova context object
        :param aggregate_id: aggregate id
        :param policy: policy info
        :param *args: extend parameters, type: tuple
        :param **kwargs: extend parameters, type: dict
        """

        raise NotImplementedError()

    def get_instances(self, context, ego_host, policy, *args, **kwargs):
        """
        Extend method for getting a list of instance from a host.
        :param context: nova context object
        :param ego_host: ego host name
        :param policy: policy info
        :param *args: extend parameters, type: tuple
        :param **kwargs: extend parameters, type: dict
        """

        raise NotImplementedError()

    def get_possible_hosts(self, context, policy, instance, aggregate_id,
                           *args, **kwargs):
        """
        Extend method for getting a list of possible hosts from EGO.
        :param context: nova context object
        :param policy: policy info
        :param instance: instance for migration
        :param aggregate_id: aggregate id
        :param *args: extend parameters, type: tuple
        :param **kwargs: extend parameters, type: dict
        """

        raise NotImplementedError()

    def optimization_action(self, context, instance, target_host_dict,
                            policy, *args, **kwargs):
        """
        Extend method for running policy optimization action.
        :param context: nova context object
        :param instance: instance for migration
        :param target_host_dict: target host dict
        :param *args: extend parameters, type: tuple
        :param **kwargs: extend parameters, type: dict
        """

        raise NotImplementedError()

    ####################################################
    #           Methods can rewrite optionally         #
    ####################################################

    def sort_violated_hosts(self, context, violated_hosts, *args, **kwargs):
        """
        Extend method for sorting the violated hosts.
        :param context: nova context object
        :param violated_hosts: violated hosts
        :param *args: extend parameters, type: tuple
        :param **kwargs: extend parameters, type: dict
        :return: sorted violated hosts
        """

        return violated_hosts

    def sort_instances_for_optimization(self, context, instances, policy,
                                        *args, **kwargs):
        """
        Extend method for sorting the instances.
        :param context: nova context object
        :param instances: instances on host
        :param *args: extend parameters, type: tuple
        :param **kwargs: extend parameters, type: dict
        :return: return sorted instances
        """

        return instances

    def select_proper_instances(self, instances, ego_host,
                                default_failed_times, recovery_time,
                                *args, **kwargs):
        """
        Extend method for select the proper instances.
        :param: instances: instances for migration
        :param: ego_host: the host which the instance running on
        :param: default_failed_times: the instance migration failed times
        :param: recovery_time: the time interval about considering the vm again
        :param *args: extend parameters, type: tuple
        :param **kwargs: extend parameters, type: dict
        :return: return proper instances
        """

        return instances

    def sort_possible_hosts(self, context, possible_hosts, policy, instance,
                            aggregate_id, *args, **kwargs):
        """
        Extend method for sorting the possible hosts.
        :param context: nova context object
        :param possible_hosts: possible hosts
        :param *args: extend parameters, type: tuple
        :param **kwargs: extend parameters, type: dict
        :return return sorted possible hosts
        """

        return possible_hosts

    def get_best_possible_host(self, possible_host_list, instance,
                               *args, **kwargs):
        """
        Extend method for getting the best possible host from the
        possible host list. Default is return the first one.
        :param possible_host_list: possible host list
        :param instance: instance for migration
        :return: best possible host
        """

        if len(possible_host_list) == 0:
            return None

        possible_host = possible_host_list[0]

        # Check whether the possible is qualified
        return (
            possible_host if self.is_validate_possible_host(
                possible_host, instance) else None)

    def is_validate_possible_host(self, possible_host, instance,
                                  *args, **kwargs):
        """
        Extend method for checking whether the possible is qualified.
        :param possible_host: best possible host
        :param instance: instance for migration
        :param *args: extend parameters, type: tuple
        :param **kwargs: extend parameters, type: dict
        """

        return True

    def build_engine_ego_host_dict(self, pem_ego_host_dict, *args, **kwargs):
        """
        for building engine ego host dict.
        :param pem_ego_host_dict: policy engine manager ego host dict.
        :param *args: extend parameters, type: tuple
        :param **kwargs: extend parameters, type: dict
        """

        self.ego_host_dict.clear()

        for hostname, attribute in six.iteritems(pem_ego_host_dict):
            svc_host = attribute['host']
            node_name = attribute['hypervisor_hostname']
            node_id = attribute['id']
            htype = attribute['hypervisor_type']
            self.ego_host_dict[hostname] = {'svc_host': svc_host,
                                        'node_name': node_name,
                                        'node_id': node_id,
                                        'htype': htype}

    def run(self, context, policy, aggregate_id):
        """
        Run the policy logics.
        :param context: nova context object
        :param policy: policy info
        :param aggregate_id: aggregate id
        """

        try:
            if not self._is_reach_policy_interval(policy, aggregate_id):
                # Set the running flag to False, because this cycle
                # is completed
                self._policy_engine_manager.aggregate_policy_is_running_dict[
                    aggregate_id] = False
                return

            self.build_engine_ego_host_dict(
                self._policy_engine_manager.ego_host_dict)

            self._sync_used_hosts(context)

            violated_hosts = self._get_violated_hosts_with_monitor(
                context, aggregate_id, policy)

            if not violated_hosts:
                LOG.info(_LI("There is not violated hosts in aggregate "
                          "%(agg_id)s for policy %(policy)s.")
                          % {'agg_id': aggregate_id,
                             'policy': policy['name']}
                          )

            violated_hosts = self.sort_violated_hosts(context, violated_hosts)

            for host in violated_hosts:
                # Ensure the host passes optimization prerequisites
                if not self._is_match_optimization_prereqs(context, host):
                    continue

                # Ensure we don't issue too many cluster-wide migrations
                # on one host
                if self._is_over_max_parallel(context, policy, aggregate_id):
                    break

                best_target_host = None
                best_target_instance = None

                instances = self.get_instances(context, host, policy)
                proper_instances = self.select_proper_instances(
                    instances, host, CONF.prs_slapolicy_instance_failed_times,
                    CONF.prs_slapolicy_instance_recovery_time)
                sorted_instances = self.sort_instances_for_optimization(
                    context, proper_instances, policy)

                for inst in sorted_instances:
                    possible_hosts = self.get_possible_hosts(
                        context, policy, inst, aggregate_id)

                    if not possible_hosts:
                        LOG.debug("There is no available possible hosts for "
                                  "instance %(instance)s to migrate."
                                  % {'instance': inst['display_name']}
                                  )
                        continue

                    best_possible_host = self.get_best_possible_host(
                        possible_hosts, inst)
                    if best_possible_host:
                        best_target_instance = inst
                        best_target_host = best_possible_host[0]
                        break

                if best_target_host and best_target_instance:
                    try:
                        self._optimization_action_with_monitor(context,
                                          policy,
                                          best_target_instance,
                                          best_target_host,
                                          host,
                                          aggregate_id)
                        # sleep Allow EGO to migrate the alloc from the
                        # previous instance
                        eventlet.sleep(2)
                    except Exception as exc:
                        action_info = (
                            self._policy_engine_manager.slapolicy_action.pop(
                                context.request_id, None))
                        if action_info:
                            action_info['action_result'] = 'failed'
                            self._policy_engine_manager.store_event(
                                context,
                                sla_constants.PRS_SLAPOLICY_ACTION_END,
                                action_info)
                        LOG.error(_LE("The optimizer could not live migrate "
                                      "instance %(name)s (%(uuid)s) from "
                                      "%(ori_host)s to %(tar_host)s due to: "
                                      "%(reason)s")
                                  % {'name': best_target_instance[
                                        'display_name'],
                                     'uuid': best_target_instance['uuid'],
                                     'ori_host': host,
                                     'tar_host': best_target_host,
                                     'reason': six.text_type(exc)})

        except Exception as e:
            LOG.error(_LE("SLA engines running into an error, the reason is "
                          "%(reason)s")
                       % {'reason': six.text_type(e)})

        # Set the running flag to False, because this cycle is completed
        self._policy_engine_manager.aggregate_policy_is_running_dict[
            aggregate_id] = False
    ####################################################
    #                 Internal Methods                 #
    ####################################################

    def _optimization_action_with_monitor(self, context, policy,
                                         instance_for_migration,
                                         target_host,
                                         source_host, aggregate_id):
        """
        Wrapper method for running the optimization with policy monitor.
        :param: context: nova context object
        :param: policy: policy info
        :param: instance_for_migration: instance for migration
        :param: target_host: best target host
        :param: source_host: violated host
        :param: aggregate_id: aggregate id
        """

        context.request_id = (sla_constants.SLAPOLICY + context.request_id)
        target_host_dict = self.ego_host_dict.get(target_host)
        source_host_dict = self.ego_host_dict.get(source_host)
        payload = {'vm_name': instance_for_migration['display_name'],
                   'vm_uuid': instance_for_migration['uuid'],
                   'target_host': target_host_dict['svc_host'],
                   'action_name': str(policy['action']),
                   'hypervisor_id': source_host_dict['node_id'],
                   'hypervisor_hostname': source_host_dict['node_name'],
                   'policy_type': 'slapolicy',
                   'slapolicy_id': policy['id'],
                   'slapolicy_name': policy['name'],
                   'is_violated': 'Yes',
                   'stabilization': policy['stabilization'],
                   'alert_times': self._policy_engine_manager.
                        aggr_hosts_alert[aggregate_id][source_host]
                   }
        if CONF.policy_per_host_aggregate:
            payload['aggregate_id'] = aggregate_id

        self._policy_engine_manager.store_event(
            context, sla_constants.PRS_SLAPOLICY_ACTION_START, payload)
        self._policy_engine_manager.slapolicy_action[context.request_id] = (
            payload)

        self.optimization_action(context, instance_for_migration,
                                 target_host_dict, policy)

        self._policy_engine_manager.used_hosts.append(
            (target_host, instance_for_migration['uuid']))

    def _get_violated_hosts_with_monitor(self, context, aggregate_id, policy):
        """
        Wrapper method for getting a list of violated hosts from EGO with
        run policy event monitor logics.
        :param context: nova context object
        :param aggregate_id: aggregate id
        :param policy: policy info
        :return: violated hosts like [host1, host2, ...]
        """

        hosts_to_return = []
        violated_hosts = self.get_violated_hosts(
            context, aggregate_id, policy)

        stabilization = policy[ego_constants.POLICY_STABILIZATION]
        host_and_count = {}
        for hostname in violated_hosts:
            host_and_count[hostname] = (
                self._policy_engine_manager.aggr_hosts_alert.get(
                    aggregate_id, {}).get(hostname, 0) + 1)

            host_dict = self.ego_host_dict.get(hostname)
            # If host dict did NOT have the host, this mean the host is new
            # add, so ignore in this cycle and will consider in next cycle
            if not host_dict:
                continue

            payload = {'hypervisor_id': host_dict['node_id'],
                       'hypervisor_hostname': host_dict['node_name'],
                       'policy_type': 'slapolicy',
                       'slapolicy_id': policy['id'],
                       'slapolicy_name': policy['name'],
                       'stabilization': stabilization,
                       'extra': policy.get('extra', {}),
                       'is_violated': ('Yes'
                            if (host_and_count[hostname] >= stabilization)
                            else 'No'),
                       'alert_times': host_and_count[hostname]}

            if CONF.policy_per_host_aggregate:
                payload['aggregate_id'] = aggregate_id

            # The host will perform policy action every <stabilization> times
            # If the policy alert continues.
            self._policy_engine_manager.store_event(context,
                             sla_constants.PRS_SLAPOLICY_HYPERVISOR_ALERT,
                             payload)
            # TODO(zhaoyim) should porting liberty fixing in here
            if host_and_count[hostname] % stabilization == 0:
                LOG.debug('The violated host (%(host)s) reached the '
                          'stabilization period (%(period)s).'
                          % {'host': hostname, 'period': stabilization})
                hosts_to_return.append(hostname)
                if host_and_count[hostname] == stabilization:
                    self._policy_engine_manager.store_event(context,
                             sla_constants.PRS_SLAPOLICY_HYPERVISOR_VIOLATED,
                             payload)

        # i.e., the hosts that are no longer violating the policy; this is
        # formed by computing the "set difference" between the previous cycle's
        # potentially violated hosts and this cycle's potentially violated
        # hosts, thus leaving the "recovered hosts" in the resulting set.
        prev_cycle_host_and_count = self._policy_engine_manager.\
            aggr_hosts_alert.get(aggregate_id, {})
        recovered_hosts = (
                set(prev_cycle_host_and_count.keys()) -
                    set(host_and_count.keys()))

        for hostname in recovered_hosts:
            host_dict = self.ego_host_dict.get(hostname)
            # If the host is not sync between EGO and nova DB, ignore it
            if not host_dict:
                continue
            payload = {'hypervisor_id': host_dict['node_id'],
                       'hypervisor_hostname': host_dict['node_name'],
                       'policy_type': 'slapolicy',
                       'slapolicy_id': policy['id'],
                       'slapolicy_name': policy['name'],
                       'stabilization': stabilization,
                       'extra': policy.get('extra', {}),
                       'is_violated': 'No',
                       'alert_times': 0}
            if CONF.policy_per_host_aggregate:
                payload['aggregate_id'] = aggregate_id
            # Only issue the "recovered notification" if the host was truly
            # violated (i.e., the stabilization period was encountered).
            if prev_cycle_host_and_count.get(hostname, 0) >= stabilization:
                self._policy_engine_manager.store_event(
                    context, sla_constants.PRS_SLAPOLICY_HYPERVISOR_RECOVERED,
                    payload)

        self._policy_engine_manager.\
            aggr_hosts_alert[aggregate_id] = host_and_count

        return hosts_to_return

    def _is_over_max_parallel(self, context, policy, aggregate_id):
        """
        Check one host max parallel migrations, ensure we don't issue
        too many cluster-wide migrations on one host.
        :param context: nova context object
        :param policy: policy info
        :param aggregate_id: aggregate id
        :return: return True if reach the max parallel number, otherwise False
        """
        policy_max_migrations = policy['max_parallel']
        agg_migr_num = len(
            self._get_instance_by_task_state_based_on_aggregate(
                context, [task_states.MIGRATING], aggregate_id))
        if agg_migr_num >= policy_max_migrations:
            LOG.info(_LI("The current number of aggregate-wide migrations "
                         "(%(migrating_num)s) for aggregate (%(agg_id)s) "
                         "has reached or exceeded its maximum allowed "
                         "value (%(max_parallel)s); additional migrations "
                         "for this aggregate will be ignored until the "
                         "number of active migrations decreases below the "
                         "threshold.")
                     % {'migrating_num': agg_migr_num,
                        'agg_id': aggregate_id,
                        'max_parallel': policy_max_migrations})
            return True
        else:
            return False

    def _is_reach_policy_interval(self, policy, aggregate_id):
        """
        Run the aggregate internal ticker to control the policy running
        cycle time interval.
        :param policy: policy info
        :param aggregate_id: aggregate id
        :return: match the policy interval, return True, otherwise False
        """

        # If the aggregate is not in the interval cache, add it
        if (aggregate_id not in
            self._policy_engine_manager.aggregates_interval_dict):
            self._policy_engine_manager.\
                aggregates_interval_dict[aggregate_id] = 0
        run_interval = int(policy[ego_constants.POLICY_RUN_INTERVAL])
        self._policy_engine_manager.\
            aggregates_interval_dict[aggregate_id] += 1
        if CONF.policy_per_host_aggregate:
            LOG.debug("Current interval count is %(current)d/%(total)s for"
                      " aggregate with ID of %(agg_id)s." %
                      {'agg_id': aggregate_id,
                       'current': self._policy_engine_manager.
                            aggregates_interval_dict[aggregate_id],
                       'total': run_interval})
        else:
            LOG.debug("Current interval count is %(current)d/%(total)s." %
                      {'current': self._policy_engine_manager.
                            aggregates_interval_dict[aggregate_id],
                       'total': run_interval})
        if self._policy_engine_manager.\
               aggregates_interval_dict[aggregate_id] < run_interval:
            return False
        else:
            # When the interval match the policy, set back to 0 for next cycle
            self._policy_engine_manager.\
                aggregates_interval_dict[aggregate_id] = 0
            return True

    def _get_instance_by_task_state(self, context, task_state, svc_host=None):
        """
        Get the specified instances based on the task state and the optionally
        specified service host.
        :param context: nova context object
        :param task_state: instance task state
        :param svc_host: name of host
        :return: instances based on task state
        """

        try:
            filters = {'task_state': task_state, 'deleted': 0}
            # If a host was specified, create the filter entry
            if svc_host:
                filters['host'] = svc_host
            return objects.InstanceList.get_by_filters(context, filters)
        except Exception as exc:
            LOG.error(_LE("Could not retrieve the instances due to: %s")
                      % six.text_type(exc))
            return []

    def _get_instance_by_task_state_based_on_aggregate(
        self, context, task_state, aggregate_id):
        """
        Retrieve all instances first and then filter out the instances that
        aren't a member of the host aggregate.
        :param context: nova context object
        :param task_state: instance task state
        :param aggregate_id: aggregate id
        :return: instances based on task state and aggregate id
        """

        all_instances = self._get_instance_by_task_state(context, task_state)

        if CONF.policy_per_host_aggregate:
            instances = [ins for ins in all_instances
                         if aggregate_id in
                         self._policy_engine_manager.
                            aggregate_container.host_aggregates.get(
                                ins['host'], {})]
        else:
            # For global mode, just return all instances
            instances = all_instances
        return instances

    def _sync_used_hosts(self, context):
        """
        Ticker for removing entries from 'used_hosts' for completed
        migrations.
        :param context: nova context object
        """
        migr_insts = self._get_instance_by_task_state(
            context, [task_states.MIGRATING])
        migrating_instance_ids = [instance['uuid'] for instance in migr_insts]
        used_hosts_prime = []
        for host_instance in self._policy_engine_manager.used_hosts:
            if host_instance[1] not in migrating_instance_ids:
                LOG.debug("Removing %s from the 'used_hosts' list because its "
                          "migration completed." %
                          six.text_type(host_instance))
            else:
                used_hosts_prime.append(host_instance)
        self._policy_engine_manager.used_hosts = used_hosts_prime
        LOG.debug("The following hosts are being used for migrations and are "
                  "unavailable: %s" % self._policy_engine_manager.used_hosts)

    def _is_match_optimization_prereqs(self, context, ego_host):
        """
        Check any prerequisites required to be satisfied prior to allowing
        ongoing optimization to occur on the hypervisor.

        :param context: nova context object
        :param ego_host: name of a hypervisor
        :return: True if optimization is allowed and False if not allowed
        """

        host_dict = self.ego_host_dict.get(ego_host)
        # 1. Ignore the violated host that has been deleted from the system
        if not host_dict:
            LOG.info(_LI("Ignoring violated host (%s) because it has been "
                         "deleted from the management system.") % ego_host)
            return False

        # 2. Ignore the violated host who is the target host in prior cycle
        if ego_host in dict(self._policy_engine_manager.used_hosts):
            LOG.info(_LI("Ignoring violated host (%s) because it is still "
                         "being used as part of a previous optimization "
                         "cycle (i.e., a migration is still ongoing).")
                     % ego_host)
            return False

        return True
