# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""SLA Policy Engine Manager."""

import copy
import os
import six

import eventlet
from nova import db
from nova import exception
from nova import rpc

from nova.ibm.ego import ego_constants
from nova.ibm.ego import ego_gettextutils
from nova.ibm.ego import ego_utils
from nova.ibm.ego import policy_ceilometer_client
from nova.ibm.ego.slapolicy import constants as sla_constants
from nova.ibm.ego.slapolicy import policy_exception
from nova.ibm.ego.slapolicy import utils as slapolicy_utils
from nova.scheduler.ibm.ego import ego_exception
from nova.scheduler.ibm.ego import ego_host_aggregate
from nova.scheduler.ibm.ego import ego_manager
from nova.scheduler.ibm.ego import ego_resreq_utils

from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import timeutils

from xml.dom import minidom

CONF = cfg.CONF
CONF.import_opt('prs_slapolicy_instance_recovery_time',
                'nova.ibm.ego.ego_flags')


_ = ego_gettextutils._
_LI = ego_gettextutils._LI
_LW = ego_gettextutils._LW
_LE = ego_gettextutils._LE

LOG = logging.getLogger(__name__)


class PolicyEngineManager(object):
    """A daemon that serves the SLA policy."""

    def __init__(self, slapolicy_conf_path):
        self.ego_api = ego_manager.EGOManager()
        self.aggregate_container = ego_host_aggregate.EGOHostAggregate()
        self.notifier = rpc.get_notifier(sla_constants.SLAPOLICY)
        self.ceilometer_api = policy_ceilometer_client.API()

        self.slapolicy_conf = slapolicy_conf_path
        self.all_policies = {}
        self.enabled_policy = {}
        self.need_set_resource_plan = True

        self.aggid_policy_dict = {}
        self.aggid_policy_dict_old = {}
        self.aggid_policy_no_check = {}
        self.aggregates_interval_dict = {}
        self.used_hosts = []
        self.failed_instances = {}
        self.aggr_hosts_alert = {}
        self.ego_host_dict = {}

        self.slapolicy_action = {}
        # This cache is used to mark whether have the policy running in
        # an aggregate, it like {agg_id1: False, agg_id2: True}, True means
        # there is a policy engine is running in the aggregate.
        self.aggregate_policy_is_running_dict = {}
        # This cache is used to cache all the engines that are configured
        # both in the policy conf xml file and implemented engine and register
        # with stevedore in the entry point
        self.policy_engine_dict = {}
        self.pool = eventlet.GreenPool()

        self.__policy_init()

    def __policy_init(self):
        """Perform SLA policy initialization tasks."""

        if not os.path.isfile(self.slapolicy_conf):
            raise policy_exception.SLAPolciyConfFileNotFoundPath(
                slapolicy_conf=self.slapolicy_conf)

        # Get all SLA policies info from the policy conf file
        slapolicies = self.get_all()
        for slapolicy in slapolicies:
            self.all_policies[slapolicy[ego_constants.POLICY_ID]] = slapolicy

        # Load policy engines here, because the engine mapping need
        # all policies cache
        self.__load_policy_engines()

        if not CONF.policy_per_host_aggregate:
            self.aggregate_policy_is_running_dict = {
                ego_constants.RESOURCE_GROUP_DEFAULT: False}
            for id in self.all_policies.keys():
                if (self.all_policies[id][
                        ego_constants.POLICY_STATE] == 'enabled'):
                    self.enabled_policy = self.all_policies[id]
                    break
            if self.enabled_policy.get(ego_constants.POLICY_GOAL):
                try:
                    self.__set_ego_resource_plan(
                        self.enabled_policy.get(ego_constants.POLICY_GOAL))
                except (ego_exception.EGOExpSetResourcePlanFailed,
                        ego_exception.EGOExpConnLost) as exc:
                    LOG.error(_LE('The SLA policy engine manager could not '
                                'set EGO resource plan for initializing '
                                'policy %(policy_id)s due to: %(reason)s')
                              % {'policy_id': self.enabled_policy.get(
                                    ego_constants.POLICY_ID),
                                 'reason': six.text_type(exc)})
        else:
            # Set host aggregate based on initial sla policy
            self.__init_aggregate_policy_is_running_dict()
            self.__sync_up_aggregate_policy(set(self.all_policies.keys()))
            self.aggid_policy_dict_old = copy.deepcopy(self.aggid_policy_dict)
            self.__set_resource_plan_per_aggregate()

    def __load_policy_engines(self):
        engines_mapping = slapolicy_utils.get_mapping('sla_policy_plugins')
        for policy_name, engine in engines_mapping:
            for policy_info in self.all_policies.values():
                if policy_name == policy_info['name']:
                    self.policy_engine_dict[policy_name] = engine

    def __get_policy_engine(self, policy):
        """
        Policy engine factory to generate the policy engine instance
        :param policy: sla policy
        :return: policy engine instance
        """

        engine = self.policy_engine_dict.get(
            self.all_policies[policy['id']]['name'])
        if engine:
            return engine(policy_engine_manager=self)
        else:
            LOG.warning(_LW("Cannot find the (%s) policy engine. Please "
                            "verify the settings in SLA policy conf file or "
                            "SLA engine plug-in entry point.")
                        % self.all_policies[policy['id']]['name'])
            return None

    def __init_aggregate_policy_is_running_dict(self):
        """
        Initial the aggregate policy running map.
        """

        # Get the aggregate id and metadata dict from DB
        aggregate_id_metadata_dict = (
            self.aggregate_container.get_host_aggregate_metadata())
        # Initial the policy running map
        for agg_id in aggregate_id_metadata_dict.keys():
            if agg_id not in self.aggregate_policy_is_running_dict:
                self.aggregate_policy_is_running_dict[agg_id] = False

    def __clear_running_map_and_interval(self, aggregate_id):
        """
        Delete the aggregate from policy running map and
        delete the aggregate from interval dict.
        """
        ego_utils.safe_remove(
            self.aggregate_policy_is_running_dict, aggregate_id)
        ego_utils.safe_remove(self.aggregates_interval_dict, aggregate_id)

    def __parse_aggregate_metadata_for_policy_show(self, metadata):
        """
        Parse metadata of aggregate with policy-related attributes.
        Do not check the validity of parameters.
        e.g., {'slainitialpolicy':{'id':'9527'},
               'slaruntimepolicy':{'id':'jd', 'run_interval':'ncmm',
               'max_parallel':'-3', 'threshold':'3x', 'action':'migrate_vm'}}
        """
        policy = {}
        runtime_prefix_len = (
            len(sla_constants.SLARUNTIMEPOLICY) +
            len(ego_constants.POLICY_DELIMITER))
        # initial policy
        sla_initial_policy = {}
        if sla_constants.SLAINITIALPOLICY_ID in metadata:
            text_value = metadata[sla_constants.SLAINITIALPOLICY_ID]
            sla_initial_policy[ego_constants.POLICY_ID] = text_value
            policy[sla_constants.SLAINITIALPOLICY] = sla_initial_policy

        # runtime policy
        sla_runtime_policy = {}
        for attribute in sla_constants.SLA_RUNTIME_POLICY_ATTRIBUTES:
            if attribute[0] in metadata:
                text_value = metadata[attribute[0]]
                attr_key = attribute[0][runtime_prefix_len:]
                if attr_key == sla_constants.POLICY_EXTRA:
                    try:
                        text_value = slapolicy_utils.parse_extra_value(
                                        text_value)
                    except Exception:
                        sla_runtime_policy[attr_key] = text_value
                        continue
                sla_runtime_policy[attr_key] = text_value

        if sla_runtime_policy:
            policy[sla_constants.SLARUNTIMEPOLICY] = sla_runtime_policy
        return policy

    def __sync_up_aggregate_policy(self, valid_policy_id_set):
        """
        Sync up the aggregate metadata policy info with engine manager
        cache.
        :param valid_policy_id_set: valid policy ids
        """

        aggid_policy_dict = {}
        aggid_policy_nc = {}
        aggregates_metadata = (
            self.aggregate_container.get_host_aggregate_metadata())
        for agg_id, metadata in six.iteritems(aggregates_metadata):
            aggid_policy_dict[agg_id] = slapolicy_utils.get_policies_info(
                metadata, agg_id, valid_policy_id_set, self.all_policies,
                self.policy_engine_dict)
            # Parse aggregate metadata to pick up all sla initial and
            # runtime policies for each aggregate even through the policy
            # parameters are invalid
            aggid_policy_nc[agg_id] = (
                self.__parse_aggregate_metadata_for_policy_show(metadata))
        self.aggid_policy_dict = aggid_policy_dict
        self.aggid_policy_no_check = aggid_policy_nc

    def __set_ego_resource_plan(self, goal=''):
        """
        Set ego resource plan; usage:
        1. When policy engine is initialized
        2. When enable global policy (if goal is specified)
        3. When set host aggregate based policy
        :param goal: The policy goal
        """

        # Set default group first
        res_group_xml = slapolicy_utils.generate_resource_plan_for_group(
            ego_constants.RESOURCE_GROUP_DEFAULT, goal)
        for aggregate_id, policy in six.iteritems(self.aggid_policy_dict):
            # Global goal or host aggregate based input goal
            if goal:
                res_group_xml += (
                    slapolicy_utils.generate_resource_plan_for_group(
                        str(aggregate_id), goal))
            else:
                policy_goal = ""
                if sla_constants.SLAINITIALPOLICY in policy:
                    init_policy = policy[sla_constants.SLAINITIALPOLICY]
                    policy_id = init_policy[ego_constants.POLICY_ID]
                    policy_goal = self.all_policies[policy_id][
                        ego_constants.POLICY_GOAL]
                res_group_xml += (
                    slapolicy_utils.generate_resource_plan_for_group(
                        str(aggregate_id), policy_goal))

        xmlstr = ("<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
                  "<DistributionPolicies>"
                  "    <Scheduler>placetorus</Scheduler>"
                  "    <InitialPlacement>%s</InitialPlacement>"
                  "</DistributionPolicies>" % (res_group_xml))
        LOG.debug('The sla policy engine manager set the EGO resource plan '
                  'to the following XML string: %s' % xmlstr)
        self.ego_api.ego_set_resource_plan(xmlstr)

    def __set_resource_plan_per_aggregate(self):
        try:
            # Set resource plan based on self.aggid_policy_dict
            self.__set_ego_resource_plan()
            self.need_set_resource_plan = False
        except (ego_exception.EGOExpSetResourcePlanFailed,
                ego_exception.EGOExpConnLost) as exc:
            LOG.error(_LE('The SLA policy engine manager could not set '
                        'the EGO resource plan due to: %(reason)s')
                      % {'reason': six.text_type(exc)})

    def __get_host_aggregates_by_policy_id_and_type(self, policy_id,
                                                    policy_type):
        """
        Get filtered metadata of aggregates directly for API request. The flag
        'is_valid' will denote whether the parameters are valid or not.
        """

        host_aggregates = []
        for agg_id, slartpolicy in six.iteritems(self.aggid_policy_no_check):
            agg_policy = copy.deepcopy(slartpolicy.get(policy_type, {}))
            agg_policy_id = agg_policy.get(ego_constants.POLICY_ID)
            if agg_policy and str(agg_policy_id) == str(policy_id):
                # Overwrite policy id by aggregate id for CLI show
                agg_policy[ego_constants.POLICY_ID] = agg_id
                valid_policy = self.aggid_policy_dict.get(agg_id, {})
                valid_policy_id = valid_policy.get(
                                policy_type, {}).get(ego_constants.POLICY_ID)
                if valid_policy_id and valid_policy_id == policy_id:
                    agg_policy[ego_constants.POLICY_VALID_FLAG] = True
                else:
                    agg_policy[ego_constants.POLICY_VALID_FLAG] = False
                host_aggregates.append(agg_policy)
        if not self.enabled_policy and host_aggregates:
            return {'host_aggregates': host_aggregates}

    def __validate_policy_parameters(self, values_dict, policy_id=''):
        def _validate_within_bounds(values_dict, attr, max_value, min_value):
            if attr not in values_dict:
                return
            try:
                value = values_dict.get(attr)
                if not (min_value <= int(value) <= max_value):
                    raise ValueError
            except ValueError:
                raise policy_exception.InvalidSLAPolicyInput(
                    value=value, attribute=attr, min=min_value, max=max_value)
        bounded_parameters = [ego_constants.POLICY_ID,
                              ego_constants.POLICY_RUN_INTERVAL,
                              ego_constants.POLICY_STABILIZATION,
                              ego_constants.POLICY_MAX_PARALLEL]
        for param in bounded_parameters:
            _validate_within_bounds(values_dict,
                                    param,
                                    ego_constants.MAX_INT_32,
                                    ego_constants.MIN_POSITIVE_INT)

        if ego_constants.POLICY_GOAL in values_dict:
            goal_val = values_dict.get(ego_constants.POLICY_GOAL)
            if goal_val == '' or goal_val.isspace():
                raise policy_exception.InvalidSLAPolicyGoal(
                    slapolicy_id=policy_id)
        if ego_constants.POLICY_STATE in values_dict:
            state_value = values_dict.get(ego_constants.POLICY_STATE)
            if state_value not in ['enabled', 'disabled']:
                raise policy_exception.InvalidSLAPolicyState(
                    slapolicy_id=policy_id, state=state_value)
        if ego_constants.POLICY_ACTION in values_dict:
            action_value = values_dict.get(ego_constants.POLICY_ACTION)
            if action_value not in list(ego_constants.POLICY_ACTION_LIST):
                raise policy_exception.InvalidSLAPolicyAction(
                    policy_action=ego_constants.POLICY_ACTION)

    def __make_slapolicy(self, node):
        slapolicy_attribute_show = [
            (ego_constants.POLICY_ID, ego_constants.ATTRIBUTE_TYPE_INT),
            (ego_constants.POLICY_NAME, ego_constants.ATTRIBUTE_TYPE_STR),
            (ego_constants.POLICY_DESCRIPTION,
             ego_constants.ATTRIBUTE_TYPE_STR),
            (ego_constants.POLICY_GOAL, ego_constants.ATTRIBUTE_TYPE_STR)
        ]

        # Global policy mode needs to retrieve all policy parameters
        if not CONF.policy_per_host_aggregate:
            slapolicy_attribute_show += [
                (ego_constants.POLICY_STATE,
                 ego_constants.ATTRIBUTE_TYPE_STR),
                (ego_constants.POLICY_ACTION,
                 ego_constants.ATTRIBUTE_TYPE_STR),
                (ego_constants.POLICY_STABILIZATION,
                 ego_constants.ATTRIBUTE_TYPE_INT),
                (ego_constants.POLICY_RUN_INTERVAL,
                 ego_constants.ATTRIBUTE_TYPE_INT),
                (ego_constants.POLICY_MAX_PARALLEL,
                 ego_constants.ATTRIBUTE_TYPE_INT),
                (sla_constants.POLICY_EXTRA,
                 sla_constants.ATTRIBUTE_TYPE_ELEMENT)
            ]

        obj = {}
        policy_id = 0
        for attribute in slapolicy_attribute_show:
            node_obj = slapolicy_utils.get_nodes(node, attribute[0])
            if not node_obj:
                LOG.error(_LE('Missing attribute %(attribute)s in the SLA '
                            'configuration file. Please check the attributes '
                            'in the file: %(file)s .')
                          % {'attribute': attribute[0],
                             'file': self.slapolicy_conf})
                raise policy_exception.MissedSLAPolicyAttributes(
                    attribute=attribute[0])

            if attribute[1] == ego_constants.ATTRIBUTE_TYPE_STR:
                text_value = slapolicy_utils.get_node_value(node_obj[0])
                obj[attribute[0]] = str(text_value)
            elif attribute[1] == ego_constants.ATTRIBUTE_TYPE_INT:
                text_value = slapolicy_utils.get_node_value(node_obj[0])
                # Check all integer type's policy parameters
                try:
                    obj[attribute[0]] = int(text_value)
                except ValueError:
                    raise exception.InvalidInput(_('The value %(value)r for '
                                        '%(name)r is invalid. '
                                        'Specify a numeric value.')
                                        % {'name': attribute[0],
                                           'value': text_value})
            elif attribute[1] == sla_constants.ATTRIBUTE_TYPE_ELEMENT:
                extra_dict = {}
                nodes = node_obj[0].childNodes
                for node in nodes:
                    if node.nodeType == node.ELEMENT_NODE:
                        value = slapolicy_utils.get_node_value(node)
                        extra_dict[node.nodeName] = value
                obj[attribute[0]] = extra_dict
            else:
                raise exception.InvalidInput(_('The SLA policy engine manager '
                                               'does not support %(type)s '
                                               'for the %(name)s policy '
                                               'configuration parameter.')
                                             % {'type': attribute[1],
                                                'name': attribute[0]})
            if not CONF.policy_per_host_aggregate:
                self.__validate_policy_parameters(
                    obj, obj[ego_constants.POLICY_ID])

        policy_id = obj[ego_constants.POLICY_ID]
        obj[ego_constants.POLICY_KEY_HOST_AGGREGATE_INITIAL] = (
            self.__get_host_aggregates_by_policy_id_and_type(
                policy_id, sla_constants.SLAINITIALPOLICY))
        obj[ego_constants.POLICY_KEY_HOST_AGGREGATE_RUNTIME] = (
            self.__get_host_aggregates_by_policy_id_and_type(
                policy_id, sla_constants.SLARUNTIMEPOLICY))
        return obj

    @lockutils.synchronized('prs-slapolicy-aggregate-lock')
    def __notify_aggregate_create(self, payload):
        # Just create resource group accordingly and update policy's
        # dictionary;  no need to set resource plan here
        if int(payload['aggregate_id']) not in self.aggid_policy_dict:
            # Initial the aggregate lockers and interval again, because the
            # aggregate created
            self.__init_aggregate_policy_is_running_dict()
            self.__sync_up_aggregate_policy(set(self.all_policies.keys()))
            self.aggid_policy_dict_old = copy.deepcopy(self.aggid_policy_dict)

    @lockutils.synchronized('prs-slapolicy-aggregate-lock')
    def __notify_aggregate_delete(self, payload):
        # Just delete resource group accordingly and update policy's
        # dictionary; no need to set resource plan here
        if int(payload['aggregate_id']) in self.aggid_policy_dict:
            # Delete the aggregate lockers and interval again, because the
            # aggregate deleted
            self.__clear_running_map_and_interval(
                int(payload['aggregate_id']))
            self.__sync_up_aggregate_policy(set(self.all_policies.keys()))
            self.aggid_policy_dict_old = copy.deepcopy(self.aggid_policy_dict)

    @lockutils.synchronized('prs-slapolicy-aggregate-lock')
    def __notify_aggregate_set_metadata(self, payload):
        # Update policy's dictionary and set resource plan if needed
        self.__sync_up_aggregate_policy(set(self.all_policies.keys()))
        self.aggid_policy_dict_old = copy.deepcopy(self.aggid_policy_dict)
        self.__set_resource_plan_per_aggregate()

    @lockutils.synchronized('prs-slapolicy-aggregate-lock')
    def __policy_function_aggregate_mode(self, context):
        self.__sync_up_aggregate_policy(set(self.all_policies.keys()))
        self.__check_resource_plan_change()
        if self.need_set_resource_plan:
            # If need to set resource plan, update the old policy
            # dict to the current policy dict
            self.aggid_policy_dict_old = copy.deepcopy(self.aggid_policy_dict)
            self.__set_resource_plan_per_aggregate()
        # Use for instance state checking in the engine
        self.aggregate_container.sync_aggregates_and_metadatas()

        for aggregate_id, policy in six.iteritems(self.aggid_policy_dict):
            if sla_constants.SLARUNTIMEPOLICY in policy:
                if not self.aggregate_policy_is_running_dict[aggregate_id]:
                    self.aggregate_policy_is_running_dict[aggregate_id] = True
                    policy_engine = self.__get_policy_engine(
                        policy.get(sla_constants.SLARUNTIMEPOLICY))

                    if not policy_engine:
                        continue

                    slaruntimepolicy = policy[sla_constants.SLARUNTIMEPOLICY]
                    id = slaruntimepolicy[ego_constants.POLICY_ID]
                    goal = self.all_policies[id][ego_constants.POLICY_GOAL]
                    name = (self.all_policies[id][ego_constants.POLICY_NAME])
                    slaruntimepolicy[ego_constants.POLICY_GOAL] = goal
                    slaruntimepolicy[ego_constants.POLICY_NAME] = name

                    self.pool.spawn_n(policy_engine.run,
                                      context,
                                      slaruntimepolicy,
                                      aggregate_id)

    def __policy_function_global_mode(self, context):
        if self.enabled_policy:
            # If True, set the resource plan which is used for initial
            # policy otherwise the resource plan already set or the
            # policy not changed
            if self.need_set_resource_plan:
                try:
                    goal = (self.enabled_policy.get(
                        ego_constants.POLICY_GOAL, '')
                            if self.enabled_policy else '')
                    self.__set_ego_resource_plan(goal)
                    self.need_set_resource_plan = False
                except (ego_exception.EGOExpSetResourcePlanFailed,
                        ego_exception.EGOExpConnLost) as exc:
                    LOG.error(_LE('The policy engine manager could not set '
                                  'the EGO resource plan for policy '
                                  '%(policy_id)s due to: %(reason)s')
                              % {'policy_id': self.enabled_policy.get(
                                    ego_constants.POLICY_ID),
                                 'reason': six.text_type(exc)})

            policy_engine = self.__get_policy_engine(self.enabled_policy)
            if policy_engine:
                # If the locker is True, that means already has the policy
                # cycle running, will not run the policy again until the
                # cycle complete
                if not self.aggregate_policy_is_running_dict[
                    ego_constants.RESOURCE_GROUP_DEFAULT]:
                    self.aggregate_policy_is_running_dict[
                        ego_constants.RESOURCE_GROUP_DEFAULT] = True
                    policy_engine.run(context,
                                      self.enabled_policy,
                                      ego_constants.RESOURCE_GROUP_DEFAULT)

    def __clear_expired_failed_instances(self):
        """
        Clear the expired failed migration instances
        """

        for key, metadata in self.failed_instances.items():
            fail_at = metadata.get('prs_slapolicy_migr_failed_tstamp')
            fail_at_time = timeutils.parse_strtime(fail_at)
            now = timeutils.utcnow()
            diff = timeutils.delta_seconds(fail_at_time, now)
            recovery_time = CONF.prs_slapolicy_instance_recovery_time
            if diff > recovery_time:
                self.failed_instances.pop(key)

    def __check_resource_plan_change(self):
        """
        Check whether host aggregate based initial policies are changed.
        Note it will be available for global policy disabled mode.
        """

        if (not CONF.policy_per_host_aggregate) or self.need_set_resource_plan:
            return

        for aggregate_id, new_policy in six.iteritems(self.aggid_policy_dict):
            if (new_policy.
                    get(sla_constants.SLAINITIALPOLICY, {}).
                    get(ego_constants.POLICY_ID) !=
                        self.aggid_policy_dict_old.
                        get(aggregate_id, {}).
                        get(sla_constants.SLAINITIALPOLICY, {}).
                        get(ego_constants.POLICY_ID)):
                self.need_set_resource_plan = True
                return

    def __construct_ego_host_dict(self):
        """
        Construct method for getting host info from EGO.
        This should run at the beginning of each periodic task.
        """

        if not self.policy_engine_dict:
            return
        select = ''
        for engine in self.policy_engine_dict.values():
            resreq = engine.get_hosts_filter_resreq()
            select = resreq if select == '' else '%s||%s' % (select, resreq)
        cstr_select = 'select(%(select)s)' % {'select': select}
        ego_nodes = self.get_ego_resources(cstr_select,
                                           is_attr_required=True)
        self.ego_host_dict.clear()

        for node in ego_nodes:
            node_attr = node['attributes']
            node_name = node_attr['hypervisor_hostname']
            ego_host = ego_resreq_utils.get_rpi_res_name(node_name)
            self.ego_host_dict[ego_host] = node_attr

    def get_ego_resources(self, cstr, state=None, is_attr_required=False):
        """
        Invoke EGO get resource list API.
        :param cstr: EGO resreq string
        :return: hosts name which match the EGO resreq string
        """

        try:
            return self.ego_api.ego_get_resource_list(
                cstr, state, is_attr_required)
        except (ego_exception.EGOExpGetResourceInfoFailed,
                ego_exception.EGOExpConnLost) as exc:
            LOG.error(_LE('The SLA policy engine manager could not retrieve '
                          'RPI resources from EGO using resource '
                          'requirements (%(cstr)s) due to: %(reason)s')
                      % {'cstr': cstr,
                         'reason': six.text_type(exc)})
            raise

    def policy_function(self, context):
        # Check EGO connection, if the EGO connection lost, need to set the
        # resource plan again, in case the policy changed during the EGO
        # connection lost period
        if self.ego_api.ego_hdl.is_conn_lost():
            self.need_set_resource_plan = True
            return

        self.__construct_ego_host_dict()
        # Clear the expired and failed migration instances
        self.__clear_expired_failed_instances()

        if CONF.policy_per_host_aggregate:
            self.__policy_function_aggregate_mode(context)
        else:
            self.__policy_function_global_mode(context)

    def store_event(self, context, event_type, payload):
        self.notifier.info(context, event_type, payload)

    def __load_events(self, context, search_opts=None, marker=None,
                      limit=None):
        query = []
        query.append({'field': 'policy_type', 'op': 'eq', 'value': 'slapolicy',
                      'type': 'string'})
        if search_opts:
            if search_opts.get('hypervisor_id'):
                query.append({'field': 'hypervisor_id', 'op': 'eq',
                              'value': search_opts['hypervisor_id'],
                              'type': 'integer'})
            if search_opts.get('slapolicy_id'):
                query.append({'field': 'slapolicy_id', 'op': 'eq',
                              'value': search_opts['slapolicy_id'],
                              'type': 'integer'})
            if search_opts.get('aggregate_id'):
                query.append({'field': 'aggregate_id', 'op': 'eq',
                              'value': search_opts['aggregate_id'],
                              'type': 'integer'})

        return self.ceilometer_api.get_events(query, marker, limit)

    def __get_slapolicies(self, doc):
        return slapolicy_utils.get_nodes(doc.documentElement,
                                         sla_constants.SLAPOLICY)

    def __get_slapolicy(self, doc, slapolicy_id):
        slapolicy_nodes = self.__get_slapolicies(doc)
        for slapolicy_node in slapolicy_nodes:
            node_id = slapolicy_utils.get_nodes(slapolicy_node, 'id')
            sla_id = slapolicy_utils.get_node_value(node_id[0])
            if sla_id == slapolicy_id:
                return slapolicy_node

        raise policy_exception.SLAPolicyNotFound(
             slapolicy_id=slapolicy_id)

    def get(self, slapolicy_id):
        """
        Get specified runtime policy from slapolicy_conf.xml for global mode.
        Get specified runtime policy from metadata for aggregate mode.
        """
        slapolicy_utils.validate_id(slapolicy_id, 'policy')
        doc = minidom.parse(self.slapolicy_conf)
        slapolicy_node = self.__get_slapolicy(doc, slapolicy_id)
        LOG.debug("The sla policy engine manager successfully "
                  "retrieved policy: %s" % slapolicy_id)
        return self.__make_slapolicy(slapolicy_node)

    def get_all(self):
        """
        Get all SLA policies information from slapolicy_conf.xml file.
        """

        slapolicy_list = []

        doc = minidom.parse(self.slapolicy_conf)
        slapolicy_nodes = self.__get_slapolicies(doc)

        for slapolicy_node in slapolicy_nodes:
            try:
                slapolicy = self.__make_slapolicy(slapolicy_node)
            except (policy_exception.InvalidSLAPolicyInput,
                    policy_exception.InvalidSLAPolicyState,
                    policy_exception.InvalidSLAPolicyAction,
                    policy_exception.MissedSLAPolicyAttributes,
                    Exception) as e:
                LOG.error(e)
                continue
            if slapolicy:
                exist = slapolicy_utils.is_policy_exist(
                    slapolicy, slapolicy_list)
                if exist:
                    LOG.error(_LE('A policy %(policy)s with the same ID '
                                  'already exists; this one will be ignored.')
                              % {'policy': slapolicy})
                    continue
            slapolicy_list.append(slapolicy)
        return slapolicy_list

    def __update_policy_state(self, doc, slapolicy_node, slapolicy, state):
        state_node = slapolicy_utils.get_nodes(slapolicy_node,
                                               ego_constants.POLICY_STATE)
        slapolicy_utils.set_node_value(doc, state_node[0], state)
        slapolicy['state'] = state

    def __enable_policy(self, doc, slapolicy_node, slapolicy):
        if (self.enabled_policy and
                self.enabled_policy['id'] != slapolicy['id']):
            raise policy_exception.InvalidSLAPolicyOperation(
                operation='update', slapolicy_id=slapolicy['id'],
                reason=_('already has an enabled policy'))

        # Clean up global variables
        self.aggr_hosts_alert = {}

        # Update EGO resource plan for global policy mode
        self.__validate_policy_parameters(slapolicy, slapolicy['id'])
        try:
            self.__set_ego_resource_plan(slapolicy['goal'])
        except (ego_exception.EGOExpSetResourcePlanFailed,
                ego_exception.EGOExpConnLost) as exc:
            LOG.error(_LE('The SLA policy engine manager could not '
                          '%(action)s policy %(slapolicy_id)s '
                          'due to: %(reason)s')
                      % {'action': 'enable',
                         'slapolicy_id': slapolicy['id'],
                         'reason': six.text_type(exc)})
            raise

        # Update XML
        self.__update_policy_state(doc, slapolicy_node, slapolicy, 'enabled')
        self.enabled_policy = slapolicy

        LOG.info(_LI('The SLA policy engine manager successfully changed '
                     'policy (%(slapolicy_id)s) to state: %(state)s')
                 % {'slapolicy_id': slapolicy['id'],
                    'state': 'enabled'})

    def __disable_policy(self, doc, slapolicy_node, slapolicy):
        # Do not update policy state if already disabled
        if slapolicy['state'] == 'disabled':
            return
        # Set EGO resource plan for global policy mode
        self.__set_ego_resource_plan()

        self.aggr_hosts_alert = {}

        # Update XML
        self.__update_policy_state(doc, slapolicy_node, slapolicy, 'disabled')

        # Clean up global variables
        self.enabled_policy = {}
        LOG.info(_LI('The SLA policy engine manager successfully changed '
                     'policy (%(slapolicy_id)s) to state: %(state)s')
                 % {'slapolicy_id': slapolicy['id'],
                    'state': 'disabled'})

    def update(self, slapolicy_id, kwargs):
        """
        Update specified SLA Policy from slapolicy_conf.xml
        """
        slapolicy_utils.validate_id(slapolicy_id, 'policy')
        slapolicy = self.get(slapolicy_id)
        LOG.info(_LI("The SLA policy engine manager is about to "
                   "update policy: %s")
                 % slapolicy_id)
        if CONF.policy_per_host_aggregate:
            if not (slapolicy_utils.
                    is_update_operation_valid_for_aggregate_mode(kwargs)):
                raise policy_exception.InvalidSLAPolicyOperation(
                    operation='update', slapolicy_id=slapolicy_id,
                    reason=_('policy_per_host_aggregate = True '
                             'is defined in the nova.conf file'))
        # Check validation of input arguments
        self.__validate_policy_parameters(kwargs, slapolicy_id)
        doc = minidom.parse(self.slapolicy_conf)
        slapolicy_node = self.__get_slapolicy(doc, slapolicy_id)
        state = kwargs.get(ego_constants.POLICY_STATE)
        if state:
            # Check EGO connection
            if self.ego_api.ego_hdl.is_conn_lost():
                raise policy_exception.SLAPolicyEGOConnLost()
            if state == 'enabled':
                self.__enable_policy(doc, slapolicy_node, slapolicy)
            else:
                self.__disable_policy(doc, slapolicy_node, slapolicy)
        slapolicy_attribute_update = [ego_constants.POLICY_DESCRIPTION,
                                      ego_constants.POLICY_GOAL,
                                      ego_constants.POLICY_NAME,
                                      ego_constants.POLICY_ACTION,
                                      ego_constants.POLICY_MAX_PARALLEL,
                                      ego_constants.POLICY_RUN_INTERVAL,
                                      ego_constants.POLICY_STABILIZATION,
                                      ego_constants.POLICY_STATE,
                                      sla_constants.POLICY_EXTRA]
        for key, value in six.iteritems(kwargs):
            if key in slapolicy_attribute_update:
                key_node = slapolicy_utils.get_nodes(slapolicy_node, key)
                if key == sla_constants.POLICY_EXTRA:
                    # If extra value is not dict, convert it to dict
                    if not isinstance(value, dict):
                        value = slapolicy_utils.parse_extra_value(value)
                    # If value is '{}', remove all extra items.
                    # Otherwise, refresh all extra items
                    slapolicy_utils.remove_child_nodes(key_node[0])
                    if value:
                        for k, v in six.iteritems(value):
                            slapolicy_utils.create_new_node(doc,
                                                        key_node[0], k, v)
                else:
                    slapolicy_utils.set_node_value(doc, key_node[0], value)
                slapolicy[key] = value
                LOG.debug("The sla policy engine manager successfully updated "
                          "key %(key)s for policy %(slapolicy_id)s "
                          "to: %(value)s"
                          % {'key': key,
                             'slapolicy_id': slapolicy_id,
                             'value': value})
        slapolicy_utils.updatexml(doc, self.slapolicy_conf)
        # If the updated slapolicy is the enabled policy, update enabled policy
        if (self.enabled_policy
                and self.enabled_policy.get('id') == slapolicy.get('id')):
            self.enabled_policy = slapolicy
        # The policy within all_policies should be updated as well
        self.all_policies[slapolicy_id] = slapolicy
        return slapolicy

    def event_list(self, context, search_opts, marker=None, limit=None):
        """
        Returns a list of slapolicy event, based on search options specified.
        """
        hypervisor_id = search_opts.get('hypervisor_id')
        slapolicy_id = search_opts.get('slapolicy_id')
        aggregate_id = search_opts.get('aggregate_id')
        if hypervisor_id:
            slapolicy_utils.validate_id(hypervisor_id, 'hypervisor')
            hypervisor_id = long(hypervisor_id)
            search_opts['hypervisor_id'] = hypervisor_id
        if slapolicy_id:
            slapolicy_utils.validate_id(slapolicy_id, 'policy')
            slapolicy_id = long(slapolicy_id)
            search_opts['slapolicy_id'] = slapolicy_id
        if aggregate_id:
            slapolicy_utils.validate_id(aggregate_id, 'aggregate')
            aggregate_id = long(aggregate_id)
            search_opts['aggregate_id'] = aggregate_id
        return self.__load_events(context, search_opts, marker, limit)

    def __get_aggregates_policy_valid_by_policy_id_and_type(self, policy_id,
                                                            policy_type):
        """
        Get host aggregates whose policy_type is valid and this policy id
        is policy_id specified.
        """
        aggregates = []
        # For global mode, only one policy can be enabled at most
        if not CONF.policy_per_host_aggregate:
            if (self.enabled_policy and
                   self.enabled_policy['id'] == policy_id):
                aggregates = [ego_constants.RESOURCE_GROUP_DEFAULT]
        # For aggregate mode, the same slapolicy might be applied to multiple
        # host aggregates
        else:
            host_aggregates = (
                self.__get_host_aggregates_by_policy_id_and_type(
                         policy_id, policy_type))
            if host_aggregates:
                for host_aggregate in host_aggregates['host_aggregates']:
                    if host_aggregate[ego_constants.POLICY_VALID_FLAG]:
                        aggregates.append(
                            host_aggregate[ego_constants.POLICY_ID])
        return aggregates

    def __get_aggregate_policy_map(self, aggregate_list, policy_type):
        agg_slapolicy_map = {}
        for aggregate_id in aggregate_list:
            policy = None
            if aggregate_id == ego_constants.RESOURCE_GROUP_DEFAULT:
                policy = self.enabled_policy
            else:
                policy = (self.aggid_policy_dict.get(aggregate_id, {}).get(
                              policy_type))
            policy_id = policy[ego_constants.POLICY_ID]
            # Get policy name from all policies by id
            policy_name = self.all_policies[
                                    policy_id][ego_constants.POLICY_NAME]
            if policy:
                agg_slapolicy_map[aggregate_id] = (
                    {'id': policy_id,
                     'name': policy_name,
                     'stabilization': policy[
                                        ego_constants.POLICY_STABILIZATION]})
        return agg_slapolicy_map

    def __hypervisor_list(self, context, aggregate_list=None, alert=False):
        """
        Get hypervisors' slapolicy info list

        :param aggregate_list: the hypervisor's aggregate's sla
                               policy is valid and the aggregate should
                               be in aggregate_list
        :param alert: the hypervisors' slapolicy should has been
                      alerting if True
        :return: hypervisors' slapolicy info list specified by aggregate_list
                 and alert option
        """
        hyvsr_list = []
        if not self.ego_host_dict:
            self.__construct_ego_host_dict()
        if aggregate_list is None:
            if not CONF.policy_per_host_aggregate:
                # Could get hypervisor list from each engine
                if self.enabled_policy:
                    aggregate_list = [ego_constants.RESOURCE_GROUP_DEFAULT]
                else:
                    aggregate_list = []
            else:
                aggregate_list = []
                for aggregate_id, policy in six.iteritems(
                                                self.aggid_policy_dict):
                    if sla_constants.SLARUNTIMEPOLICY in policy:
                        aggregate_list.append(aggregate_id)
        agg_slapolicy_map = self.__get_aggregate_policy_map(aggregate_list,
                                                sla_constants.SLARUNTIMEPOLICY)
        for agg in aggregate_list:
            if agg not in agg_slapolicy_map:
                # The aggregate did not set slapolicy
                continue
            if alert:
                egohost_list = self.aggr_hosts_alert.get(agg, {}).keys()
            else:
                egohost_list = []
                if agg != ego_constants.RESOURCE_GROUP_DEFAULT:
                    svc_hosts = db.aggregate_host_get_all(context, agg)
                    for ego_host, h_dict in six.iteritems(self.ego_host_dict):
                        if (h_dict['host'] in svc_hosts and (
                            str(agg) in h_dict['host_aggregate_ids'].split(
                                ','))):
                            egohost_list.append(ego_host)
                else:
                    egohost_list = self.ego_host_dict.keys()
            for egohost in egohost_list:
                hyvsr = {}
                hyvsr['id'] = self.ego_host_dict[egohost]['id']
                hyvsr['hypervisor_hostname'] = (
                        self.ego_host_dict[egohost]['hypervisor_hostname'])
                hyvsr['consecutive_alert_times'] = (
                        self.aggr_hosts_alert.get(agg, {}).get(egohost, 0))
                hyvsr['stabilization'] = (
                        agg_slapolicy_map[agg]['stabilization'])
                hyvsr['slapolicy_id'] = agg_slapolicy_map[agg]['id']
                hyvsr['slapolicy_name'] = agg_slapolicy_map[agg]['name']
                hyvsr['is_violated'] = ('Yes'
                    if (int(hyvsr['consecutive_alert_times'])
                        >= int(hyvsr['stabilization']))
                    else "No")
                if CONF.policy_per_host_aggregate:
                    hyvsr['aggregate_id'] = agg
                hyvsr_list.append(hyvsr)
        hyvsr_list.sort(key=lambda hyvsr: hyvsr['id'])
        return hyvsr_list

    def hypervisor_list(self, context, search_opts):
        """
        Returns a list hypervisors's slapolicy info, based on any search
        options specified.
        """
        alert = search_opts.get('alert')
        slapolicy_id = search_opts.get('slapolicy_id')
        aggregate_id = search_opts.get('aggregate_id')
        hyvsr_list = []

        # Only return hypervisors in the aggregates, which would depend on
        # both slapolicy_id and aggregated options. If it is None, return all
        # the hypervisors; if it is [], do not return any hypervisor; if it
        # is [ag1, ag2,...], return hypervisors in either in ag1, or in ag2...
        aggregates = None

        if slapolicy_id:
            slapolicy_utils.validate_id(slapolicy_id, 'policy')
            slapolicy_id = int(slapolicy_id)
            aggregates = (
                self.__get_aggregates_policy_valid_by_policy_id_and_type(
                         slapolicy_id, sla_constants.SLARUNTIMEPOLICY))
        if aggregate_id:
            if not CONF.policy_per_host_aggregate:
                raise exception.InvalidInput(_("Cannot support aggregate-id "
                                               "option in global mode because "
                                               "('policy_per_host_aggregate = "
                                               "True' is defined in the "
                                               "nova.conf file)."))
            else:
                slapolicy_utils.validate_id(aggregate_id, 'aggregate')
                aggregate_id = int(aggregate_id)
                if aggregates is None:
                    if aggregate_id not in self.aggid_policy_dict:
                        LOG.warning(_LW('Aggregate %s could not be found.')
                                    % aggregate_id)
                        aggregates = []
                    elif (sla_constants.SLARUNTIMEPOLICY not in
                              self.aggid_policy_dict[aggregate_id]):
                        LOG.warning(_LW('Aggregate %s is not set with a valid '
                                        'SLA runtime policy.') % aggregate_id)
                        aggregates = []
                    else:
                        aggregates = [aggregate_id]
                else:
                    if aggregate_id in aggregates:
                        aggregates = [aggregate_id]
                    else:
                        aggregates = []
        if aggregates or (aggregates is None):
            hyvsr_list = self.__hypervisor_list(context, aggregates, alert)
        return hyvsr_list

    def notify_policy_resource_status(self, context, msg):
        """
        Notify resource optimization policy service take related
        action based on notification message.
        This routine only works for aggregate policy mode and
        should verify whether the notification is obsolete or not.

        Aggregate create: Create EGO resource group accordingly
        Aggregate delete: Delete EGO resource group accordingly
        Set Aggregate metadata: Update policy's dictionary and set resource
                                plan if needed
        """

        event_type = msg['event_type']
        payload = msg['payload']
        instance_id = payload.get('instance_id')

        # Handle all mode notification here
        if event_type in (
            [ego_constants.EVENT_TYPE_COMPUTE_INST_ROCKBACK_LIVE_MIG_END,
             ego_constants.EVENT_TYPE_COMPUTE_INST_LIVE_MIG_ROLLBACK_END,
             ego_constants.EVENT_TYPE_CHECK_MIGRATE_DEST]):
            # Only tag the slapolicy-triggered migrations that failed
            request_id = str(context.request_id)
            if request_id.startswith(sla_constants.SLAPOLICY):
                now = timeutils.strtime()
                metadata = {"prs_slapolicy_migr_failed_tstamp": now}
                if instance_id not in self.failed_instances:
                    metadata['failed_times'] = 1
                    self.failed_instances[instance_id] = metadata
                else:
                    self.failed_instances[instance_id][
                        'prs_slapolicy_migr_failed_tstamp'] = now
                    self.failed_instances[instance_id]['failed_times'] += 1

            action_info = self.slapolicy_action.pop(context.request_id, None)
            if action_info:
                action_info['action_result'] = 'failed'
                self.store_event(context,
                                   sla_constants.PRS_SLAPOLICY_ACTION_END,
                                   action_info)
        elif event_type == (
             ego_constants.EVENT_TYPE_COMPUTE_INST_LIVE_MIGRATION_POST_END):
            action_info = self.slapolicy_action.pop(context.request_id, None)
            if action_info:
                action_info['action_result'] = 'success'
                self.store_event(context,
                                   sla_constants.PRS_SLAPOLICY_ACTION_END,
                                   action_info)

        if not CONF.policy_per_host_aggregate:
            return

        # Handle aggregate mode only notification here
        if event_type == ego_constants.EVENT_TYPE_AGGREGATE_CREATE_END:
            LOG.debug("Aggregate create end message: %s" % msg)
            self.__notify_aggregate_create(payload)
        elif event_type == ego_constants.EVENT_TYPE_AGGREGATE_DELETE_END:
            LOG.debug("Aggregate delete end message: %s" % msg)
            self.__notify_aggregate_delete(payload)
        elif event_type == ego_constants.EVENT_TYPE_AGGREGATE_METADATA_SET_END:
            LOG.debug("Aggregate metadata update end message: %s" % msg)
            self.__notify_aggregate_set_metadata(payload)
