# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""
Client side of the slapolicy RPC API.
"""

from oslo_config import cfg
import oslo_messaging as messaging

from nova.objects import base as objects_base
from nova import rpc

rpcapi_opts = [
    cfg.StrOpt('slapolicy_topic',
               default='slapolicy',
               help='The topic on which the slapolicy service listens'),
]

CONF = cfg.CONF
CONF.register_opts(rpcapi_opts)

rpcapi_cap_opt = cfg.StrOpt('slapolicy',
            help='Set a version cap for messages sent to slapolicy')
CONF.register_opt(rpcapi_cap_opt, 'upgrade_levels')


class SLAPolicyAPI(object):
    '''Client side of the slapolicy RPC API.

    API version history:

        1.0 - Initial version.

    '''

    VERSION_ALIASES = {
        'liberty': '1.0'
    }

    def __init__(self, topic=None):
        super(SLAPolicyAPI, self).__init__()
        target = messaging.Target(topic=CONF.slapolicy_topic, version="1.0")
        version_cap = self.VERSION_ALIASES.get(CONF.upgrade_levels.slapolicy,
                                               CONF.upgrade_levels.slapolicy)
        serializer = objects_base.NovaObjectSerializer()
        self.client = rpc.get_client(target,
                                     version_cap=version_cap,
                                     serializer=serializer)

    def get_all(self, context):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'get_all')

    def get(self, context, slapolicy_id):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'get', slapolicy_id=slapolicy_id)

    def update(self, context, slapolicy_id, slapolicy, **kwargs):
        for key in slapolicy.keys():
            kwargs[key] = slapolicy[key]
        kwargs['slapolicy_id'] = slapolicy_id
        cctxt = self.client.prepare()
        return cctxt.call(context, 'update', **kwargs)

    def hypervisor_list(self, context, search_opts):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'hypervisor_list', search_opts=search_opts)

    def event_list(self, context, search_opts, marker, limit):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'event_list', search_opts=search_opts,
                          marker=marker, limit=limit)
