# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""SLAPolicy Utility."""

import codecs
import six

from nova import exception

from nova.ibm.ego import ego_constants
from nova.ibm.ego import ego_gettextutils
from nova.ibm.ego.slapolicy import constants as sla_constants
from nova.ibm.ego.slapolicy import policy_exception

from stevedore import extension

from oslo_log import log as logging

_ = ego_gettextutils._
_LI = ego_gettextutils._LI
_LW = ego_gettextutils._LW
_LE = ego_gettextutils._LE

LOG = logging.getLogger(__name__)


def get_mapping(namespace):
    mgr = extension.ExtensionManager(
        namespace=namespace, invoke_on_load=False)
    return [[name, mgr[name].plugin] for name in mgr.names()]


def get_nodes(node, name):
    """
    Get the child nodes by name under parent node.
    :param node: parent node object
    :param name: child node name
    :return: child nodes if parent not None, otherwise []
    """

    return node.getElementsByTagName(name) if node else []


def get_node_value(node):
    """
    Get the XML node value.
    :param node: XML node object
    :return: node value
    """

    return node.childNodes[0].nodeValue if node.hasChildNodes() else ''


def set_node_value(doc, node, value):
    """
    Set the XML node with a value
    :param: doc: XML document object
    :param: node: XML node object
    :param: value: XML node value
    """

    if not value:
        # If parameter 'value' is None, delete the related XML node
        parent_node = node.parentNode
        parent_node.removeChild(node)
    else:
        value = str(value)
        if node.hasChildNodes():
            node.childNodes[0].nodeValue = value
        else:
            text_node = doc.createTextNode(value)
            node.appendChild(text_node)


def create_new_node(doc, parentNode, nodename, value):
    """
    Create new XML node with a specified nodename and value
    :param doc: XML document object
    :param parentNode: XML node object
    :param nodename: XML node name
    :param value: XML node value
    """

    new_node = doc.createElement(nodename)
    text_node = doc.createTextNode(str(value))
    new_node.appendChild(text_node)
    parentNode.appendChild(new_node)


def remove_child_nodes(node):
    """
    Remove all child nodes under specified node
    :param node: XML node value
    """

    child_nodes = get_nodes(node, "*")
    for child_node in child_nodes:
        node.removeChild(child_node)


def updatexml(doc, conf_path):
    """
    Update XML document that specified by 'conf_path'
    :param doc:
    :param conf_path:
    :return:
    """

    with codecs.open(conf_path, 'w', 'utf-8') as f:
        doc.writexml(f, addindent='', encoding='utf-8')


def validate_id(id_value, id_name):
    """
    Common function to validate policy/hypervisor/aggregate id.
    :param id_value: policy/hypervisor/aggregate id
    :param id_name: id type, shoubld be 'policy', 'hypervisor'
                     or 'aggregate'
    """

    if not id_value:
        raise exception.InvalidInput(_('Specify a valid %s ID.')
                                         % id_name)
    try:
        int(id_value)
    except Exception:
        raise exception.InvalidInput(_('%(type)s %(id)s is not valid. '
                                        'Specify a numeric %(type)s ID.')
                                        % {'type': id_name, 'id': id_value})


def is_update_operation_valid_for_aggregate_mode(param_dict):
    invalid_parameter_of_aggregate_mode = [
        ego_constants.POLICY_ACTION,
        ego_constants.POLICY_MAX_PARALLEL,
        sla_constants.POLICY_EXTRA,
        ego_constants.POLICY_STABILIZATION,
        ego_constants.POLICY_STATE,
        ego_constants.POLICY_RUN_INTERVAL
    ]
    for item in param_dict.keys():
        if item in invalid_parameter_of_aggregate_mode:
            return False
    return True


def is_policy_exist(current, exists_list):
    """
    Check whether the current policy is already exist in the policy list.
    :param current: current policy
    :param exists_list: existing policies list
    :return: if already exist in the list return True otherwise False
    """

    for policy in exists_list:
        if policy.get(ego_constants.POLICY_ID) == current.get(
            ego_constants.POLICY_ID):
            return True
    return False


def generate_resource_plan_for_group(rg_name, goal=""):
    """
    Generate EGO resource plan XML.
    :param rg_name: resource group name
    :param goal: resource plan goal
    :return: EGO resource plan XML
    """

    goal = 'order(%s)' % goal if len(goal) > 0 else ''
    xmlstr = ("<ResourceGroup ResourceGroupName=\"%(rg_name)s\">"
              "    <PlacementPolicy>%(goal)s</PlacementPolicy>"
              "</ResourceGroup>" %
              {'rg_name': rg_name, 'goal': goal})
    return xmlstr


def parse_extra_value(value):
    """
    Convert extra value from string to dict
    """
    value = value or '{}'
    try:
        extra = eval(value)
        # handle invalid extra str like "{'aa'}"
        if isinstance(extra, set):
            raise Exception
        return dict(extra)
    except Exception:
        raise policy_exception.InvalidSLAPolicyExtra(extra=value)


def get_policies_info(metadata, agg_id, valid_policy_id_set, all_policies,
                      policy_engine_dict):
    """
    Parse the aggregate metadata and get policy info.
    :param metadata: aggregate metadata
    :param agg_id: aggregate id
    :param valid_policy_id_set: all valid policy ids
    :param all_policies: all policy information
    :param policy_engine_dict: policy name and policy engine mapping
    :return: return policy info
    """

    def _runtime_policy_int_check(value, attr, agg_id):
        if (value < ego_constants.MIN_INT_32
             or value > ego_constants.MAX_INT_32):
            attribute = '%s-%s' % (sla_constants.SLARUNTIMEPOLICY, attr)
            msg = (_LW('The value (%(val)s) of attribute (%(attr)s) '
                    'in metadata is invalid for host aggregate '
                    '(%(agg_id)s). It should be within the range '
                    '(%(val_min)s - %(val_max)s).'
                    ) % {'attr': attribute,
                         'val': value,
                         'val_min': ego_constants.MIN_INT_32,
                         'val_max': ego_constants.MAX_INT_32,
                         'agg_id': agg_id})
            LOG.warning(msg)
            raise exception.Invalid()
    policy = {}
    if sla_constants.SLAINITIALPOLICY_ID in metadata:
        slainitial_policy = {}
        try:
            tmp_initialpolicy_id = metadata[sla_constants.SLAINITIALPOLICY_ID]
            initialpolicy_id = int(tmp_initialpolicy_id)
            if initialpolicy_id in valid_policy_id_set:
                slainitial_policy[ego_constants.POLICY_ID] = initialpolicy_id
                policy[sla_constants.SLAINITIALPOLICY] = slainitial_policy
            else:
                valid_ids_str = ','.join(
                        [str(pid) for pid in valid_policy_id_set])
                msg = (_LW('The specified initial policy ID (%(value)s) '
                         'for attribute (%(attr)s) is invalid for '
                         'aggregate (%(agg_id)s). '
                         'Specify one of the following policy IDs '
                         'instead: (%(valid_ids)s).') %
                       {'value': initialpolicy_id,
                        'attr': sla_constants.SLAINITIALPOLICY_ID,
                        'valid_ids': valid_ids_str,
                        'agg_id': agg_id})
                LOG.warning(msg)
        except Exception:
            msg = (_LW('The value (%(value)s) of attribute (%(attr)s) '
                    'in metadata is not the expected type (%(type)s) '
                    'for aggregate (%(agg_id)s).'
                    ) % {'value': tmp_initialpolicy_id,
                         'attr': sla_constants.SLAINITIALPOLICY_ID,
                         'type': ego_constants.ATTRIBUTE_TYPE_INT,
                         'agg_id': agg_id})
            LOG.warning(msg)

    if (sla_constants.SLARUNTIMEPOLICY_RUN_INTERVAL in metadata
        or sla_constants.SLARUNTIMEPOLICY_STABILIZATION in metadata
        or sla_constants.SLARUNTIMEPOLICY_EXTRA in metadata
        or sla_constants.SLARUNTIMEPOLICY_ACTION in metadata
        or sla_constants.SLARUNTIMEPOLICY_ID in metadata
        or sla_constants.SLARUNTIMEPOLICY_MAX_PARALLEL in metadata):
        slaruntime_policy = {}
        try:
            for attribute in sla_constants.SLA_RUNTIME_POLICY_ATTRIBUTES:
                if attribute[0] in metadata:
                    try:
                        value = metadata[attribute[0]]
                        key = attribute[0][len(
                            sla_constants.SLARUNTIMEPOLICY +
                                ego_constants.POLICY_DELIMITER):]
                        if attribute[1] == ego_constants.ATTRIBUTE_TYPE_INT:
                            value = int(value)
                            if (attribute[0] ==
                                 sla_constants.SLARUNTIMEPOLICY_ID
                                 and value not in valid_policy_id_set):
                                valid_ids_str = ','.join(
                                    [str(pid) for pid in valid_policy_id_set])
                                msg = (_LW('The specified runtime policy ID '
                                         '(%(value)s) for attribute '
                                         '(%(attr)s) is invalid for aggregate'
                                         ' (%(agg_id)s). Specify one of the '
                                         'following policy IDs instead: '
                                         '(%(valid_ids)s).') %
                                       {'value': value,
                                        'attr':
                                            sla_constants.SLARUNTIMEPOLICY_ID,
                                        'valid_ids': valid_ids_str,
                                        'agg_id': agg_id})
                                LOG.warning(msg)
                                raise policy_exception.InvalidSLAPolicyID(
                                    value=value,
                                    attr=sla_constants.SLARUNTIMEPOLICY_ID,
                                    agg_id=agg_id,
                                    valid_ids=valid_ids_str)
                            _runtime_policy_int_check(
                                value, attribute[0], agg_id)
                            slaruntime_policy[key] = value
                        elif attribute[1] == ego_constants.ATTRIBUTE_TYPE_STR:
                            slaruntime_policy[key] = value
                        elif (attribute[1] ==
                            sla_constants.ATTRIBUTE_TYPE_ELEMENT):
                            slaruntime_policy[key] = parse_extra_value(value)
                    except policy_exception.InvalidSLAPolicyID:
                        raise
                    except (NameError, ValueError, SyntaxError):
                        msg = (_LW('The value (%(value)s) of '
                                'attribute (%(attr)s) in metadata '
                                'is not the expected type (%(type)s) '
                                'for aggregate (%(agg_id)s).'
                                ) % {'value': value,
                                     'attr': attribute[0],
                                     'type': attribute[1],
                                     'agg_id': agg_id})
                        LOG.warning(msg)
                        raise
                    except Exception as e:
                        msg = (_LW('Unknown error on parsing host aggregate '
                                   'ID (%(agg_id)s) metadata. The reason is'
                                   ' (%(reason)s).')
                                % {'agg_id': agg_id,
                                   'reason': six.text_type(e)})
                        LOG.warning(msg)
                        raise
                else:
                    if attribute[0] != sla_constants.SLARUNTIMEPOLICY_EXTRA:
                        msg = (_LW('Host aggregate ID (%(agg_id)s) does '
                                'not have attribute (%(attr)s) in metadata.'
                                ) % {'attr': attribute[0], 'agg_id': agg_id})
                        LOG.warning(msg)
                        raise exception.InvalidMetadata(reason=msg)
            policy_id = slaruntime_policy[ego_constants.POLICY_ID]
            policy_name = all_policies[policy_id][ego_constants.POLICY_NAME]
            policy_engine = policy_engine_dict[policy_name]
            engine_actions = policy_engine.get_optimization_actions()
            if slaruntime_policy[
                ego_constants.POLICY_ACTION] not in engine_actions:
                valid_actions_str = ','.join(engine_actions)
                msg = (_LW('In aggregate %(aggregate)s, the %(policy_action)s '
                           'is not a valid policy action, please specify '
                           'a valid action %(engine_actions)r for the '
                           'SLA policy.') %
                       {'aggregate': agg_id,
                        'policy_action': slaruntime_policy[
                            ego_constants.POLICY_ACTION],
                        'engine_actions': valid_actions_str})
                LOG.warning(msg)
                raise policy_exception.InvalidSLAPolicyAction(
                    policy_action=valid_actions_str)
        except Exception:
            return policy
        policy[sla_constants.SLARUNTIMEPOLICY] = slaruntime_policy

    return policy
