# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from nova.ibm.ego.wrg.clients import BaseStoreClient
from nova.ibm.ego.wrg.clients.etcd.client import Client
from nova.ibm.ego.wrg.clients.etcd import EtcdException
from nova.ibm.ego.wrg.clients.etcd import EtcdKeyNotFound
from nova.ibm.ego.wrg import exception as wrg_exception

from oslo_config import cfg


_etcd_opts = [
  cfg.StrOpt('etcd_host_ip',
             default='127.0.0.1',
             help='IP to connect to etcd'),
  cfg.IntOpt('etcd_host_port',
             default=4001,
             help='Port used to connect to etcd.'),
  cfg.StrOpt('etcd_url_version',
             default='/v2',
             help='Url or version prefix in etcd url'),
  cfg.StrOpt('etcd_url_protocol',
             default='http',
             help='Protocol used to connect to etcd.'),
  cfg.StrOpt('etcd_cert_file',
             default='',
             help='The cert file name'),
  cfg.StrOpt('etcd_key_file',
             default='',
             help='The key file name'),
  cfg.StrOpt('etcd_ca_cert',
             default='',
             help='The ca certificate.')
]


def etcd_error_handler(func):
    def handler(*args, **kwdargs):
        try:
            return func(*args, **kwdargs)
        except EtcdKeyNotFound:
            raise wrg_exception.StoreKeyNotFound()
        except EtcdException:
            raise wrg_exception.StoreClientError()
    return handler


class EtcdClient(BaseStoreClient):
    def __init__(self, host='127.0.0.1', port=4001,
                 version_prefix='/v2', read_timeout=60,
                 allow_redirect=True, protocol='http',
                 cert_file=None, key_file=None,
                 ca_cert=None, allow_reconnect=False,
                 use_proxies=False, expected_cluster_id=None):
        if key_file:
            cert = (cert_file, key_file)
        else:
            cert = cert_file
        self.client = Client(host, port, version_prefix, read_timeout,
                             allow_redirect, protocol, cert, ca_cert,
                             allow_reconnect, use_proxies, expected_cluster_id)

    @etcd_error_handler
    def read(self, key, **kwdargs):
        results = {}
        node = self.client.read(key, **kwdargs)
        for child in node.leaves:
            results[child.key] = child.value

        return results

    @etcd_error_handler
    def write(self, key, value, **kwdargs):
        self.client.write(key, value, **kwdargs)

    @etcd_error_handler
    def delete(self, key, **kwdargs):
        self.client.delete(key, **kwdargs)

    @etcd_error_handler
    def update(self, obj):
        self.client.update(obj)

    @classmethod
    def get_instance(cls, conf):
        conf.register_opts(_etcd_opts)
        return cls(host=conf.etcd_host_ip, port=conf.etcd_host_port,
                   version_prefix=conf.etcd_url_version,
                   protocol=conf.etcd_url_protocol,
                   cert_file=conf.etcd_cert_file,
                   key_file=conf.etcd_key_file,
                   ca_cert=conf.etcd_ca_cert)
