# =================================================================
# Licensed Materials _ Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights _ Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Workload resource group service."""

import copy
import six

from glanceclient import client as glance

import keystoneclient.auth.identity.generic.password as password_auth
import keystoneclient.session as keystonesession

from nova import exception
from nova.ibm.ego import ego_constants as ego_consts
from nova.ibm.ego import ego_gettextutils
from nova.ibm.ego.wrg import clients
from nova.ibm.ego.wrg import constants as wrg_consts
from nova.ibm.ego.wrg import exception as wrg_exception
from nova.ibm.ego.wrg import utils as wrg_utils
from nova import manager
import nova.objects as nova_objects
from nova import rpc

from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_serialization import jsonutils
from oslo_service import sslutils

_ = ego_gettextutils._

_LI = ego_gettextutils._LI
_LW = ego_gettextutils._LW
_LE = ego_gettextutils._LE

CONF = cfg.CONF

LOG = logging.getLogger(__name__)

glance_options = [
    cfg.StrOpt('username',
               default='',
               help='Glance service user name'),
    cfg.StrOpt('password',
               secret=True,
               default='',
               help='Glance service password'),
    cfg.StrOpt('project_name',
               default='',
               help='Glance service tenant name'),
    cfg.StrOpt('auth_url',
               default='',
               help='Keystone authentication URL'),
    cfg.BoolOpt('insecure',
                default=False,
                help='Verify HTTPS connections')]

CONF.register_opts(glance_options, 'ibm_prs_glance')
sslutils.is_enabled(CONF)


class WorkloadResourceGroupManager(manager.Manager):
    """Manage the workload resource group CRUD operations."""

    target = messaging.Target(version="1.0")

    def __init__(self, *args, **kwargs):
        super(WorkloadResourceGroupManager, self).__init__(
                  service_name='Workload_Resource_Group', *args, **kwargs)
        self.store_client = clients.get_store_client(CONF)
        self.notifier = rpc.get_notifier("workload_resource_group")
        self.glanceclient = self.__create_glanceclient()
        self.__sync_up_wrg()

    def create(self, context, wrg_body):
        """create a new WRG with wrg_body
        """
        self.notifier.info(context, wrg_consts.PRS_WRG_CREATE_START, wrg_body)
        try:
            wrg_utils.validate_wrg(wrg_body, create=True)
            wrg_name = wrg_body[wrg_consts.WRG_INFO][wrg_consts.WRG_NAME]
            wrg_uuid = wrg_body[wrg_consts.WRG_INFO][wrg_consts.WRG_CID]

            wrgs = self.__get_all_wrg()
            for wrg in wrgs.values():
                # WRG name must be unique
                if wrg[wrg_consts.WRG_INFO][wrg_consts.WRG_NAME] != wrg_name:
                    continue
                LOG.error(_LE("Workload resource group %(name)r "
                              "already exists.") % {'name': wrg_name})
                raise wrg_exception.WRGNameExists(name=wrg_name)
            self.__create_wrg(wrg_uuid, wrg_body)
            wrgs[wrg_uuid] = wrg_body
            self.__sync_metadata_defs(context, wrgs)
            self.notifier.info(context, wrg_consts.PRS_WRG_CREATE_END,
                               wrg_body)
            return wrg_body
        except Exception:
            self.notifier.info(context, wrg_consts.PRS_WRG_CREATE_ERROR,
                               wrg_body)
            raise

    def get(self, context, wrg_uuid):
        """return the specified WRG
        """
        return self.__get_wrg_by_uuid(wrg_uuid)

    def get_all(self, context):
        """return all WRGs
        """
        wrgs = self.__get_all_wrg()
        all_wrg = {"workload_resource_groups": []}
        for wrg in wrgs.values():
            all_wrg["workload_resource_groups"].append(wrg)
        return all_wrg

    def update(self, context, wrg_uuid, wrg_body):
        """update the WRG specified by wrg_uuid with wrg_body
        """
        old_wrg = self.__get_wrg_by_uuid(wrg_uuid)
        self.notifier.info(context, wrg_consts.PRS_WRG_UPDATE_START, wrg_body)
        try:
            if wrg_consts.PERF_POLICY not in wrg_body:
                # if not plan to update performance policy, keep it unchanged
                wrg_body[wrg_consts.PERF_POLICY] = (
                            old_wrg[wrg_consts.PERF_POLICY])
            else:
                # if plan to update performance policy, check and generate a
                # valid one
                wrg_utils.validate_policy(wrg_body[wrg_consts.PERF_POLICY],
                                          create=True)
            if wrg_consts.WRG_INFO not in wrg_body:
                wrg_body[wrg_consts.WRG_INFO] = {}
            if wrg_consts.WRG_DESC not in wrg_body[wrg_consts.WRG_INFO]:
                wrg_body[wrg_consts.WRG_INFO][wrg_consts.WRG_DESC] = (
                    old_wrg[wrg_consts.WRG_INFO][wrg_consts.WRG_DESC])
            for const_attr in [wrg_consts.WRG_CID, wrg_consts.WRG_NAME]:
                # both WRG UUID and name cannot be changed
                if const_attr not in wrg_body[wrg_consts.WRG_INFO]:
                    wrg_body[wrg_consts.WRG_INFO][const_attr] = (
                        old_wrg[wrg_consts.WRG_INFO][const_attr])
                else:
                    if (wrg_body[wrg_consts.WRG_INFO][const_attr] !=
                            old_wrg[wrg_consts.WRG_INFO][const_attr]):
                        msg = (_LE('Cannot update workload resource group %s')
                               % const_attr)
                        LOG.error(msg)
                        raise exception.InvalidInput(reason=msg)

            wrg_utils.validate_wrg(wrg_body)
            self.__update_wrg(wrg_uuid, wrg_body)
            self.notifier.info(context, wrg_consts.PRS_WRG_UPDATE_END,
                               wrg_body)
            return wrg_body
        except Exception:
            self.notifier.info(context, wrg_consts.PRS_WRG_UPDATE_ERROR,
                               wrg_body)
            raise

    def delete(self, context, wrg_uuid):
        """delete the specified WRG
        """
        if wrg_uuid == self.default_wrg_uuid:
            msg = (_LW("Deleting the default workload resource group "
                       "%(uuid)r is prohibited."
                       ) % {'uuid': wrg_uuid})
            LOG.warning(msg)
            raise wrg_exception.WRGDeleteFailed(uuid=wrg_uuid, reason=msg)
        wrg_dict = self.__get_wrg_by_uuid(wrg_uuid)
        wrg_name = wrg_dict[wrg_consts.WRG_INFO][wrg_consts.WRG_NAME]
        inst_wrg_map = self.__get_all_inst_wrg_map()
        associated_instances = []
        for wrg_info in six.itervalues(inst_wrg_map):
            if wrg_name in [wrg_info.get(wrg_consts.WRG_NAME),
                            wrg_info.get(wrg_consts.OLD_WRG)]:
                associated_instances.append(wrg_info.get(wrg_consts.VS_NAME))
        if associated_instances:
            msg = (_LW("Deleting a workload resource group containing a "
                       "VM instances %(name)r is prohibited."
                        ) % {'name': associated_instances})
            LOG.warning(msg)
            raise wrg_exception.WRGDeleteFailed(uuid=wrg_uuid, reason=msg)

        mapped_flavors = self.__get_mapped_flavor(context, wrg_name)
        if mapped_flavors:
            flavor_names = []
            for flavor in mapped_flavors:
                flavor_names.append(str(flavor.name))
            msg = (_LW("It is not permitted to delete a WRG because "
                       "flavors %(name)r associated with it."
                       ) % {'name': flavor_names})
            LOG.warning(msg)
            raise wrg_exception.WRGDeleteFailed(uuid=wrg_uuid, reason=msg)

        self.notifier.info(context, wrg_consts.PRS_WRG_DELETE_START,
                           {wrg_consts.WRG_CID: wrg_uuid})
        try:
            self.__delete_wrg(wrg_uuid)
            self.__sync_metadata_defs(context)
            self.notifier.info(context, wrg_consts.PRS_WRG_DELETE_END,
                               wrg_dict)
        except Exception:
            self.notifier.info(context, wrg_consts.PRS_WRG_DELETE_ERROR,
                               {wrg_consts.WRG_CID: wrg_uuid})
            raise

    def __get_wrg_by_uuid(self, wrg_uuid):
        try:
            raw_wrg = self.store_client.read(
                      wrg_consts.WRG_STORE_PREFIX + wrg_uuid)
            wrg = wrg_utils.filter_map_from_store_base_prefix(raw_wrg,
                                                 wrg_consts.WRG_STORE_PREFIX)
        except wrg_exception.StoreKeyNotFound:
            LOG.error(_LE("Workload resource group %(uuid)r cannot be "
                          "found.") % {'uuid': wrg_uuid})
            raise wrg_exception.WRGNotFoundByUUID(uuid=wrg_uuid)
        except wrg_exception.StoreClientError:
            LOG.error(_LE("Failed to read workload resource group "
                          "information from data store."))
            raise wrg_exception.WRGSyncFailed()
        else:
            return wrg[wrg_uuid]

    def __delete_wrg(self, wrg_uuid):
        """delete workload resource group from persistence layer
        """
        try:
            self.store_client.delete(wrg_consts.WRG_STORE_PREFIX + wrg_uuid)
            LOG.info(_LI('Deleted workload resource group %(uuid)r.')
                     % {'uuid': wrg_uuid})
        except Exception:
            msg = (_LE("Failed to delete workload resource group %(uuid)r "
                       "from persistence.") % {'uuid': wrg_uuid})
            LOG.error(msg)
            raise wrg_exception.WRGDeleteFailed(uuid=wrg_uuid, reason=msg)

    def __create_wrg(self, wrg_uuid, wrg_body):
        """create workload resource group in persistence layer
        """
        try:
            self.store_client.write(wrg_consts.WRG_STORE_PREFIX + wrg_uuid,
                                    jsonutils.dumps(wrg_body))
            LOG.info(_LI('Created workload resource group %(uuid)r: %(wrg)s.')
                     % {'uuid': wrg_uuid, 'wrg': wrg_body})
        except Exception:
            wrg_name = wrg_body[wrg_consts.WRG_INFO][wrg_consts.WRG_NAME]
            LOG.error(_LE("Failed to persist workload resource group "
                          "%(name)r.") % {'name': wrg_name})
            raise wrg_exception.WRGStoreFailed(name=wrg_name)

    def __update_wrg(self, wrg_uuid, wrg_body):
        """update workload resource group in persistence layer
        """
        try:
            self.store_client.write(wrg_consts.WRG_STORE_PREFIX + wrg_uuid,
                                    jsonutils.dumps(wrg_body), prevExist=True)
            LOG.info(_LI('Updated workload resource group %(uuid)r: %(wrg)s.')
                     % {'uuid': wrg_uuid, 'wrg': wrg_body})
        except Exception:
            LOG.error(_LE("Failed to update workload resource group "
                          "%(uuid)r.") % {'uuid': wrg_uuid})
            raise wrg_exception.WRGUpdateFailed(uuid=wrg_uuid)

    def __get_all_wrg(self):
        """Get all WRGs from persistence layer."""
        try:
            store_result = self.store_client.read(wrg_consts.WRG_STORE_PREFIX)
            wrgs = wrg_utils.filter_map_from_store_base_prefix(store_result,
                                                 wrg_consts.WRG_STORE_PREFIX)
            return wrgs
        except wrg_exception.StoreKeyNotFound:
            return {}
        except wrg_exception.StoreClientError:
            LOG.error(_LE("Failed to read workload resource group "
                          "information from data store."))
            raise wrg_exception.WRGSyncFailed()

    def __sync_up_wrg(self):
        """Check if PRSDefaultWorkloadResourceGroup has been stored in
        persistence layer. If not, persist it.
        """
        found_default = False
        wrgs = self.__get_all_wrg()

        for wrg_uuid, wrg_body in six.iteritems(wrgs):
            if not found_default and \
                    wrg_body[wrg_consts.WRG_INFO][wrg_consts.WRG_NAME] ==\
                        wrg_consts.DEFAULT_WRG_NAME:
                self.default_wrg_uuid = wrg_uuid
                found_default = True

        if not found_default:
            # Create the default workload.
            try:
                wrg_body = copy.deepcopy(wrg_consts.DEFAULT_WRG)
                wrg_utils.validate_wrg(wrg_body, create=True)
                wrg_uuid = \
                    wrg_body[wrg_consts.WRG_INFO][wrg_consts.WRG_CID]
                self.__create_wrg(wrg_uuid, wrg_body)
                self.default_wrg_uuid = wrg_uuid
                wrgs[wrg_uuid] = wrg_body
            except wrg_exception.WRGStoreFailed:
                raise wrg_exception.WRGDefaultCreateFailed()

        self.__sync_metadata_defs(wrgs=wrgs, create=True)

    def __get_all_inst_wrg_map(self):
        """Get all insts' association with WRG from persistence layer."""
        try:
            store_result = self.store_client.read(wrg_consts.VS_STORE_PREFIX)
            return wrg_utils.filter_map_from_store_base_prefix(store_result,
                                                  wrg_consts.VS_STORE_PREFIX)
        except wrg_exception.StoreKeyNotFound:
            return {}
        except wrg_exception.StoreClientError:
            LOG.error(_LE("Failed to read virtual server instances' "
                          "association with workload resource group "
                          "from data store during sync."))
            raise wrg_exception.InstWRGMapSyncFailed()

    def __get_inst_wrg_map_by_uuid(self, uuid):
        """Get specified uuid inst's association with WRG from persistence
        layer.
        """
        try:
            store_result = self.store_client.read(
                                             wrg_consts.VS_STORE_PREFIX + uuid)
            return wrg_utils.filter_map_from_store_base_prefix(store_result,
                                                  wrg_consts.VS_STORE_PREFIX)
        except wrg_exception.StoreKeyNotFound:
            return {}
        except wrg_exception.StoreClientError:
            LOG.error(_LE("Failed to read virtual server instances' "
                          "association with workload resource group from "
                          "data store during sync."))
            raise wrg_exception.InstWRGMapSyncFailed()

    def associate_inst_with_wrg(self, context, inst_uuid, wrg_name, inst_name,
                                **kwargs):
        if (wrg_name is None) and (wrg_consts.OLD_WRG not in kwargs):
            # if wrg_name is None, and kwargs do not have the old wrg name
            # we will delete the instance's WRG info.
            try:
                self.store_client.delete(
                    wrg_consts.VS_STORE_PREFIX + inst_uuid)
                LOG.info(_LI('Deleted workload resource group association of '
                             'virtual server %r.')
                          % inst_uuid)
                return
            except Exception:
                raise wrg_exception.DissociateVSWRGFailed(uuid=inst_uuid)
        else:
            if wrg_name:
                # It is allowed to associate a VM instance with a WRG only when
                # the WRG exists.
                try:
                    wrgs = self.__get_all_wrg()
                except Exception:
                    raise wrg_exception.AssociateVSWRGFailed(uuid=inst_uuid)
                for wrg in wrgs.values():
                    if (wrg[wrg_consts.WRG_INFO][wrg_consts.WRG_NAME] ==
                            wrg_name):
                        break
                else:
                    raise wrg_exception.WRGNotFoundByName(name=wrg_name)
                body = {wrg_consts.VS_NAME: inst_name,
                        wrg_consts.WRG_NAME: wrg_name}
            else:
                body = {wrg_consts.VS_NAME: inst_name}
            if kwargs:
                body.update(kwargs)
            try:
                self.store_client.write(wrg_consts.VS_STORE_PREFIX + inst_uuid,
                                    jsonutils.dumps(body))
                LOG.info(_LI('Update workload resource group association of '
                             'virtual server %(uuid)r to %(body)r.')
                         % {'uuid': inst_uuid, 'body': body})
                return body
            except Exception:
                raise wrg_exception.AssociateVSWRGFailed(uuid=inst_uuid)

    def get_inst_wrg_map(self, context):
        return self.__get_all_inst_wrg_map()

    def __create_glanceclient(self):
        auth_url = CONF.ibm_prs_glance.auth_url
        if not auth_url:
            LOG.warning(_LW('Skipping glance client creation due to '
                            'no auth url specified'))
            return None

        auth = password_auth.Password(
            auth_url=CONF.ibm_prs_glance.auth_url,
            username=CONF.ibm_prs_glance.username,
            password=CONF.ibm_prs_glance.password,
            project_name=CONF.ibm_prs_glance.project_name)
        if CONF.ibm_prs_glance.insecure:
            verify = False
        else:
            verify = CONF.ssl.ca_file or True

        cert_file = CONF.ssl.cert_file
        key_file = CONF.ssl.key_file
        if cert_file and key_file:
            cert = (cert_file, key_file)
        else:
            cert = None

        try:
            return glance.Client(
                version='2',
                session=keystonesession.Session(
                    auth=auth,
                    cert=cert,
                    verify=verify))
        except Exception:
            LOG.warning(_LW('Could not obtain glance client'), exc_info=True)
            return None

    def __get_mapped_flavor(self, context, wrg_name):
        """Get all flavors that are associated with a given workload.
        """
        flavors = []
        for flavor in nova_objects.FlavorList.get_all(context):
            if flavor.extra_specs.get(ego_consts.ZHPM_WRG_NAME) == wrg_name:
                flavors.append(flavor)
        return flavors

    def __sync_metadata_defs(self, context=None, wrgs=None, create=False):
        """syncs the list of workloads in the metadata namespace with what
           is known in etcd, creating a new namespace if necessary.
        """
        if not self.glanceclient:
            LOG.info(_LI('No glance client available. Ignoring metadata '
                         'sync'))
            return

        if not wrgs:
            wrgs = self.__get_all_wrg()

        wrg_names = [wrg_body[wrg_consts.WRG_INFO][wrg_consts.WRG_NAME]
                     for wrg_body in six.itervalues(wrgs)]
        propertydef = {
            'enum': wrg_names,
            'default': wrg_consts.DEFAULT_WRG_NAME,
            'title': 'Workload Resource Group',
            'type': 'string'
        }
        if create:
            try:
                namespace = self.glanceclient.metadefs_namespace.get(
                    wrg_consts.WRG_NAMESPACE)
            except Exception:
                LOG.info(_LI('No namespace found, creating.'))
            else:
                if namespace:
                    create = False

        try:
            if create:
                namespacedef = {
                    'namespace': wrg_consts.WRG_NAMESPACE,
                    'resource_type_associations': [{
                        'name': 'OS::Nova::Flavor'
                    }],
                    'properties': {
                        ego_consts.ZHPM_WRG_NAME: propertydef
                    },
                    'display_name': 'Workload Resource Group Properties',
                    'description': 'Properties related to Workload '
                                   'Resource Group scheduling.',
                    'visibility': 'public',
                    'protected': True
                }
                self.glanceclient.metadefs_namespace.create(**namespacedef)
            else:
                self.glanceclient.metadefs_property.update(
                    wrg_consts.WRG_NAMESPACE,
                    ego_consts.ZHPM_WRG_NAME,
                    **propertydef)
        except Exception:
            LOG.warning(_LW('Cannot sync metadata namespace'), exc_info=True)
            self.glance_endpoint = None
        else:
            LOG.info(_LI('Synchronized metadata namespace'))
