# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""
Client side of the workload resource group RPC API.
"""

from oslo_config import cfg
import oslo_messaging as messaging

from nova.ibm.ego.wrg import constants as wrg_constants
from nova.objects import base as objects_base
from nova import rpc

wrg_opts = [
    cfg.StrOpt('topic',
               default=wrg_constants.WRG_SERVICE_TOPIC,
               help='The topic on which the workload resource group service '
                    'listens'),
    cfg.IntOpt('workers',
               help='Number of workers for workload resource group service. '
                    'The default will be the number of CPUs available.')
]
wrg_group = cfg.OptGroup(name='workload_resource_group',
                         title='Workload resource group Options')

CONF = cfg.CONF
CONF.register_group(wrg_group)
CONF.register_opts(wrg_opts, wrg_group)

rpcapi_cap_opt = cfg.StrOpt('workload_resource_group',
            help='Set a version cap for messages sent to workload resource'
                 'group service')
CONF.register_opt(rpcapi_cap_opt, 'upgrade_levels')


class WRGAPI(object):
    '''Client side of the workload resource group RPC API.

    API version history:

        1.0 - Initial version.

    '''

    VERSION_ALIASES = {
        'liberty': '1.0'
    }

    def __init__(self, topic=None):
        super(WRGAPI, self).__init__()
        target = messaging.Target(topic=CONF.workload_resource_group.topic,
                                  version="1.0")
        version_cap = self.VERSION_ALIASES.get(
                               CONF.upgrade_levels.workload_resource_group,
                               CONF.upgrade_levels.workload_resource_group)
        serializer = objects_base.NovaObjectSerializer()
        self.client = rpc.get_client(target,
                                     version_cap=version_cap,
                                     serializer=serializer)

    def create(self, context, wrg_body):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'create', wrg_body=wrg_body)

    def get_all(self, context):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'get_all')

    def get(self, context, wrg_uuid):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'get', wrg_uuid=wrg_uuid)

    def delete(self, context, wrg_uuid):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'delete', wrg_uuid=wrg_uuid)

    def update(self, context, wrg_uuid, wrg_body):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'update', wrg_uuid=wrg_uuid,
                          wrg_body=wrg_body)

    def associate_inst_with_wrg(self, context, inst_uuid, wrg_name, inst_name,
                                **kwargs):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'associate_inst_with_wrg',
                          inst_uuid=inst_uuid, wrg_name=wrg_name,
                          inst_name=inst_name, **kwargs)

    def get_inst_wrg_map(self, context):
        cctxt = self.client.prepare()
        return cctxt.call(context, 'get_inst_wrg_map')
