# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import copy
import re
import six
import uuid

from nova import exception
from nova.ibm.ego import ego_gettextutils
from nova.ibm.ego.wrg import constants as wrg_consts
from nova.ibm.ego.wrg import exception as wrg_exception
from nova import utils as nova_utils

from oslo_log import log as logging
from oslo_serialization import jsonutils

_ = ego_gettextutils._
_LW = ego_gettextutils._LW
_LE = ego_gettextutils._LE

LOG = logging.getLogger(__name__)


def validate_str_attr(data_dict, key, key_scope, value_scope=None,
                      min_length=None, max_length=None):
    """validate the key in data_dict
    :param data_dict: the dict which the key is in
    :param key: the key which need be validated
    :param key_scope: the scope which this key belongs to
    :param value_scope: the range of the key's corresponding value
    :param min_length: validate the min_length of the value if not None
    :param max_length: validate the min_length of the value if not None
    """
    key_name = " ".join([key_scope, key])
    if key not in data_dict:
        msg = _LE("%s is not defined") % key_name
        LOG.error(msg)
        raise exception.InvalidInput(reason=msg)
    value = data_dict[key]
    if isinstance(value, six.string_types):
        value = value.strip()
    if min_length is not None:
        nova_utils.check_string_length(value, key_name,
               min_length=min_length, max_length=max_length)
    if value_scope and value not in value_scope:
        if value_scope == wrg_consts.SC_CPU_CRITICAL_LIST:
            value_scope = '[true, false] boolean value'
        msg = (_LE("Invalid %(key_name)s: %(value)s, which must be one of "
                 "the following: %(value_scope)s.")
               % {'key_name': key_name, 'value': value,
                  'value_scope': value_scope})
        LOG.error(msg)
        raise exception.InvalidInput(reason=msg)


def pop_invalid_key(data_dict, data_name, valid_key_list):
    """pop keys which are not in valid_key_list from data_dict
    """
    invalid_key_set = set(data_dict.keys()) - set(valid_key_list)
    for invalid_key in invalid_key_set:
        data_dict.pop(invalid_key)
    if invalid_key_set:
        LOG.warning(_LW("Keys dropped for %(data_name)s: %(invalid_keys)s")
                    % {'data_name': data_name,
                       'invalid_keys': str(list(invalid_key_set))})


def is_regEx_exact_match(pattern, target_str):
    """check if the target string matches the regex pattern exactly
    """
    result = re.match(pattern, target_str)
    if result and result.group() == target_str:
        return True
    return False


def validate_type(inst, inst_name, cls):
    if not isinstance(inst, cls):
        msg = (_LE('Invalid type of %(inst_name)s: should be %(cls)s type')
               % {'inst_name': inst_name, 'cls': cls.__name__})
        LOG.error(msg)
        raise exception.InvalidInput(reason=msg)


def validate_wrg(wrg_dict, create=False):
    """validate workload resource group
    """
    wrg_info = wrg_dict[wrg_consts.WRG_INFO]
    wrg_scope = 'Workload resource group'
    validate_str_attr(wrg_info, wrg_consts.WRG_NAME, wrg_scope,
                      min_length=1,
                      max_length=wrg_consts.WRG_NAME_MAX_LENGTH)
    if not is_regEx_exact_match(r'[\w_\- ]+', wrg_info[wrg_consts.WRG_NAME]):
        raise wrg_exception.InvalidWRGName(name=wrg_info[wrg_consts.WRG_NAME])

    if create:
        if wrg_consts.WRG_CID in wrg_info:
            msg = (_LE("When creating a workload resource group, %r cannot "
                       "be specified."
                       ) % wrg_consts.WRG_CID)
            LOG.error(msg)
            raise exception.InvalidInput(reason=msg)
        wrg_info[wrg_consts.WRG_CID] = str(uuid.uuid4())
    validate_str_attr(wrg_info, wrg_consts.WRG_CID, wrg_scope)
    if wrg_consts.WRG_DESC not in wrg_info:
        wrg_info[wrg_consts.WRG_DESC] = ""
    validate_str_attr(wrg_info, wrg_consts.WRG_DESC, wrg_scope,
                      min_length=0,
                      max_length=wrg_consts.WRG_DESCRIPTION_MAX_LENGTH)
    pop_invalid_key(wrg_info, ": ".join([wrg_scope, wrg_consts.WRG_INFO]),
                    wrg_consts.WRG_INFO_ATTR_LIST)

    if wrg_dict.get(wrg_consts.PERF_POLICY):
        validate_policy(wrg_dict[wrg_consts.PERF_POLICY], create)
    else:
        wrg_dict[wrg_consts.PERF_POLICY] = (
                          copy.deepcopy(wrg_consts.DEFAULT_PERF_POLICY))
        for sc in wrg_dict[wrg_consts.PERF_POLICY][wrg_consts.SVC_CLASSES]:
            sc[wrg_consts.SC_CID] = str(uuid.uuid4())
    pop_invalid_key(wrg_dict, wrg_scope, wrg_consts.WRG_ATTR_LIST)


def validate_policy(policy_dict, create=False):
    """validate performance policy of workload resource group
    """
    validate_type(policy_dict, wrg_consts.PERF_POLICY, dict)
    if not policy_dict:
        policy_dict = copy.deepcopy(wrg_consts.DEFAULT_PERF_POLICY)
    policy_info = policy_dict.get(wrg_consts.PERF_POLICY_INFO)
    if not policy_info:
        msg = _LE("Invalid %s") % wrg_consts.PERF_POLICY
        LOG.error(msg)
        raise exception.InvalidInput(reason=msg)
    validate_type(policy_info, wrg_consts.PERF_POLICY_INFO, dict)
    policy_scope = "Workload resource group's performance policy"
    validate_str_attr(policy_info, wrg_consts.POLICY_NAME, policy_scope,
                      min_length=1,
                      max_length=wrg_consts.POLICY_NAME_MAX_LENGTH)
    if wrg_consts.POLICY_DESC not in policy_info:
        policy_info[wrg_consts.POLICY_DESC] = ""
    validate_str_attr(policy_info, wrg_consts.POLICY_DESC, policy_scope,
                      min_length=0,
                      max_length=wrg_consts.POLICY_DESCRIPTION_MAX_LENGTH)
    validate_str_attr(policy_info, wrg_consts.POLICY_BUSNS_IMPTC, policy_scope,
                      value_scope=wrg_consts.BUSINESS_IMPORTANCE_LIST)
    pop_invalid_key(policy_info,
                    ": ".join([policy_scope, wrg_consts.PERF_POLICY_INFO]),
                    wrg_consts.POLICY_INFO_ATTR_LIST)

    if policy_dict.get(wrg_consts.SVC_CLASSES):
        validate_service_class(policy_dict[wrg_consts.SVC_CLASSES], create)
    else:
        policy_dict[wrg_consts.SVC_CLASSES] = (
                [wrg_consts.DEFAULT_SVC_CLASS.copy()])
        for sc in policy_dict[wrg_consts.SVC_CLASSES]:
            sc[wrg_consts.SC_CID] = str(uuid.uuid4())
    pop_invalid_key(policy_dict, policy_scope,
                    wrg_consts.PERF_POLICY_ATTR_LIST)


def validate_service_class(sc_list, create=False):
    """validate service classes of performance policy
    """
    def __validate_vs_filter(sc):
        if wrg_consts.SC_VS_NAME_FILTERS not in sc:
            msg = (_("%s is not defined in the service class")
                   % wrg_consts.SC_VS_NAME_FILTERS)
            raise exception.InvalidInput(reason=msg)
        validate_type(sc[wrg_consts.SC_VS_NAME_FILTERS],
                      wrg_consts.SC_VS_NAME_FILTERS, list)
        for filter_str in sc[wrg_consts.SC_VS_NAME_FILTERS]:
            try:
                re.compile(filter_str)
            except Exception:
                msg = (_LE('Invalid regular expression in %s ')
                       % wrg_consts.SC_VS_NAME_FILTERS)
                LOG.error(msg)
                raise exception.InvalidInput(reason=msg)

    validate_type(sc_list, wrg_consts.SVC_CLASSES, list)
    sc_scope = "Performance policy's service class"
    svc_names = []
    default_svc_index = -1
    for index, sc in enumerate(sc_list):
        validate_type(sc, "service class", dict)
        validate_str_attr(sc, wrg_consts.SC_NAME, sc_scope,
                          min_length=1,
                          max_length=wrg_consts.SC_NAME_MAX_LENGTH)
        if sc[wrg_consts.SC_NAME] in svc_names:
            msg = (_LE("It is not permitted to make service class's name "
                       "duplicate in a performance policy: service class "
                       "named %s is duplicate.")
                       % sc[wrg_consts.SC_NAME])
            LOG.error(msg)
            raise exception.InvalidInput(reason=msg)
        else:
            svc_names.append(sc[wrg_consts.SC_NAME])
        # It is not recommended that a user defines a service class named
        # PRSDefaultServiceClass, and once the user defines it, it is asked
        # the service class should keep same definition as PRS default one.
        if (sc[wrg_consts.SC_NAME] ==
                wrg_consts.DEFAULT_SVC_CLASS[wrg_consts.SC_NAME]):
            if __is_default_svc_class(sc):
                default_svc_index = index
            else:
                msg = (_LE("Service class %(prs_default_sc_name)s is constant "
                           "and must be %(prs_default_sc)s.")
                       % {"prs_default_sc_name":
                            wrg_consts.DEFAULT_SVC_CLASS[wrg_consts.SC_NAME],
                          "prs_default_sc": wrg_consts.DEFAULT_SVC_CLASS})
                LOG.error(msg)
                raise exception.InvalidInput(reason=msg)

        if wrg_consts.SC_DESC not in sc:
            sc[wrg_consts.SC_DESC] = ""
        validate_str_attr(sc, wrg_consts.SC_DESC, sc_scope,
                          min_length=0,
                          max_length=wrg_consts.SC_DESCRIPTION_MAX_LENGTH)
        validate_str_attr(sc, wrg_consts.SC_BUSNS_IMPTC, sc_scope,
                          value_scope=wrg_consts.BUSINESS_IMPORTANCE_LIST)
        validate_str_attr(sc, wrg_consts.SC_VEL_GOAL, sc_scope,
                          value_scope=wrg_consts.VELOCITY_GOAL_LIST)
        validate_str_attr(sc, wrg_consts.SC_CPU_CRITICAL, sc_scope,
                          value_scope=wrg_consts.SC_CPU_CRITICAL_LIST)
        __validate_vs_filter(sc)

        if create:
            if wrg_consts.SC_CID in sc:
                msg = (_LE("When creating/updating a workload resource group, "
                           "service class %r cannot be specified.")
                       % wrg_consts.SC_CID)
                LOG.error(msg)
                raise exception.InvalidInput(reason=msg)
            sc[wrg_consts.SC_CID] = str(uuid.uuid4())
        pop_invalid_key(sc, sc_scope, wrg_consts.SC_ATTR_LIST)

    # make sure the last service class must be PRS default service class
    if default_svc_index == -1:
        default_sc = wrg_consts.DEFAULT_SVC_CLASS.copy()
        default_sc[wrg_consts.SC_CID] = str(uuid.uuid4())
        sc_list.append(default_sc)
    elif default_svc_index != (len(sc_list) - 1):
        default_sc = sc_list.pop(default_svc_index)
        sc_list.append(default_sc)


def __is_default_svc_class(sc):
    tmp = sc.copy()
    tmp.pop(wrg_consts.SC_CID, None)
    return tmp == wrg_consts.DEFAULT_SVC_CLASS


def filter_map_from_store_base_prefix(store_result, prefix):
    prefix_len = len(prefix)
    filtered = {}
    for key, value in six.iteritems(store_result):
        if key.startswith(prefix) and key != prefix:
            k = key[prefix_len:]
            if '/' not in k:
                filtered[k] = jsonutils.loads(value)
    return filtered
