# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
import eventlet
from eventlet import queue

from nova.ibm.ego import ego_constants
from nova.ibm.ego import ego_gettextutils
from nova import manager
from nova.objects import base as objects_base
from nova import rpc

from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging as messaging

_ = ego_gettextutils._
_LI = ego_gettextutils._LI

LOG = logging.getLogger(__name__)

CONF = cfg.CONF
CONF.import_opt('scheduler_topic', 'nova.scheduler.rpcapi')
CONF.import_opt('rtpolicy_topic', 'nova.ibm.ego.rtpolicy.rpcapi')
CONF.import_opt('ibm_notification_topics',
                'nova.ibm.ego.notifier.rpc_notifier')
CONF.import_opt('hapolicy_topic', 'nova.ibm.ego.hapolicy.rpcapi')
CONF.import_opt('resource_topic', 'nova.ibm.ego.resource.rpcapi')
CONF.import_opt('slapolicy_topic', 'nova.ibm.ego.slapolicy.rpcapi')


class NotificationManager(manager.Manager):
    """
    Implement a notification dispatcher.
    This NotificationManager listens to the AMQP server's notification topics,
    handles these messages and dispatches notifications to PRS scheduler
    service, resource optimization service and HA service when needed.
    """

    def __init__(self, *args, **kwargs):
        super(NotificationManager, self).__init__(*args, **kwargs)
        self.notification_queue = queue.Queue()

        serializer = objects_base.NovaObjectSerializer()
        target_schd = messaging.Target(topic=CONF.scheduler_topic)
        self.client_schd = rpc.get_client(target_schd, serializer=serializer)

        target_opt = messaging.Target(topic=CONF.rtpolicy_topic)
        self.client_opt = rpc.get_client(target_opt, serializer=serializer)

        target_ha = messaging.Target(topic=CONF.hapolicy_topic)
        self.client_ha = rpc.get_client(target_ha, serializer=serializer)

        target_rs = messaging.Target(topic=CONF.resource_topic)
        self.client_rs = rpc.get_client(target_rs, serializer=serializer)

        target_sla = messaging.Target(topic=CONF.slapolicy_topic)
        self.client_sla = rpc.get_client(target_sla, serializer=serializer)

        self.topics = CONF.ibm_notification_topics

        eventlet.spawn_after(1, self.__start_ntf_svr)
        eventlet.spawn_after(1, self.__start_consumer_thread)

    def __start_ntf_svr(self):
        LOG.info(_LI("Begin to listen topics: %s") % self.topics)

        class NotificationProxy(object):
            def __init__(self, ntf_queue):
                self.ntf_queue = ntf_queue

            def audit(self, ctxt, publisher_id, event_type,
                      payload, metadata):
                self.notification_dispatch(ctxt, publisher_id, event_type,
                                           payload, metadata)

            def debug(self, ctxt, publisher_id, event_type,
                      payload, metadata):
                self.notification_dispatch(ctxt, publisher_id, event_type,
                                           payload, metadata)

            def info(self, ctxt, publisher_id, event_type,
                     payload, metadata):
                self.notification_dispatch(ctxt, publisher_id, event_type,
                                           payload, metadata)

            def warn(self, ctxt, publisher_id, event_type,
                     payload, metadata):
                self.notification_dispatch(ctxt, publisher_id, event_type,
                                           payload, metadata)

            def error(self, ctxt, publisher_id, event_type,
                      payload, metadata):
                self.notification_dispatch(ctxt, publisher_id, event_type,
                                           payload, metadata)

            def critical(self, ctxt, publisher_id, event_type,
                         payload, metadata):
                self.notification_dispatch(ctxt, publisher_id, event_type,
                                           payload, metadata)

            def sample(self, ctxt, publisher_id, event_type,
                       payload, metadata):
                self.notification_dispatch(ctxt, publisher_id, event_type,
                                           payload, metadata)

            def notification_dispatch(self, ctxt, publisher_id, event_type,
                                      payload, metadata):
                # Ignore the messages that we are not interested in.
                if event_type not in ego_constants.EVENT_TYPE_LIST:
                    return

                msg = {'publisher_id': publisher_id,
                       'event_type': event_type,
                       'payload': payload,
                       'metadata': metadata}
                LOG.debug('Received a notification from compute node; '
                          'ctxt: %(ctxt)s, msg: %(msg)s.'
                          % {'ctxt': ctxt, 'msg': msg})
                self.ntf_queue.put({'ctxt': ctxt, 'msg': msg})
                LOG.debug('The current queue size is %s.'
                          % str(self.ntf_queue.qsize()))

        targets = []
        for topic in self.topics:
            targets.append(messaging.Target(topic=topic))
        endpoints = [NotificationProxy(self.notification_queue)]
        serializer = rpc.RequestContextSerializer(None)

        rpc.init(CONF)
        server = messaging.get_notification_listener(rpc.TRANSPORT, targets,
                                                     endpoints,
                                                     serializer=serializer)
        server.start()
        server.wait()

    def ntf_hdl_scheduler_service(self, ctxt, msg):
        """
        Send notification to nova scheduler
        """
        LOG.debug('Notify the Nova scheduler with message: %s.' % msg)

        cctxt = self.client_schd.prepare(version='4.2')
        cctxt.cast(ctxt, 'notify_scheduler_resource_status', msg=msg)

    def ntf_hdl_opt_service(self, ctxt, msg):
        """
        Send notification to nova-ibm-ego-resource-optimization service
        """
        LOG.debug('Notify the Nova IBM EGO Resource Optimization '
                  'with message: %s.' % msg)

        cctxt = self.client_opt.prepare(version='1.0')
        cctxt.cast(ctxt, 'notify_policy_resource_status', msg=msg)

    def ntf_hdl_ha_service(self, ctxt, msg):
        """
        Send notification to nova-ibm-ego-ha-service
        """

        LOG.debug('Notify the Nova IBM EGO High Availability service '
                  'with message %s.' % msg)

        cctxt = self.client_ha.prepare(version='1.0')
        cctxt.cast(ctxt, 'notify_policy_resource_status', msg=msg)

    def ntf_hdl_resource_service(self, ctxt, msg):
        """
        Send notification to nova-ibm-ego-resource service
        """

        LOG.debug('Notify the Nova IBM EGO Resource service '
                  'with message %s.' % msg)

        cctxt = self.client_rs.prepare(version='1.0')
        cctxt.cast(ctxt, 'notify_resource_status', msg=msg)

    def ntf_hdl_sla_service(self, ctxt, msg):
        """
        Send notification to nova-ibm-ego-slapolicy service
        """
        LOG.debug('Notify the Nova IBM EGO SLA Policy '
                  'with message: %s.' % msg)

        cctxt = self.client_sla.prepare(version='1.0')
        cctxt.cast(ctxt, 'notify_policy_resource_status', msg=msg)

    def __start_consumer_thread(self):
        """
        Get the notification and run the notification handler.
        """

        while True:
            obj = self.notification_queue.get()
            if obj:
                LOG.debug('The current notification queue size is: %s.'
                          % str(self.notification_queue.qsize()))
                self.ntf_hdl_scheduler_service(obj['ctxt'], obj['msg'])
                self.ntf_hdl_opt_service(obj['ctxt'], obj['msg'])
                self.ntf_hdl_ha_service(obj['ctxt'], obj['msg'])
                self.ntf_hdl_resource_service(obj['ctxt'], obj['msg'])
                self.ntf_hdl_sla_service(obj['ctxt'], obj['msg'])
            else:
                eventlet.sleep(0.1)
