import collections
import six

from nova.compute import vm_states
from nova import exception
from nova.ibm.ego import ego_constants
from nova.ibm.ego import ego_gettextutils
from nova.ibm.ego import ego_utils
from nova.ibm.ego.wrg import constants as wrg_consts
from nova.ibm.ego.wrg import exception as wrg_exception
from nova.ibm.ego.wrg import rpcapi as wrg_rpcapi
from nova import objects
from nova.objects import base as obj_base
from nova.scheduler.ibm.ego import ego_exception
from nova.scheduler.ibm.ego.ego_extension_manager import base
from nova.scheduler.ibm.ego import ego_resreq_utils

from oslo_config import cfg
from oslo_log import log as logging
from oslo_messaging import exceptions as oslo_exceptions

_ = ego_gettextutils._
_LI = ego_gettextutils._LI
_LW = ego_gettextutils._LW
_LE = ego_gettextutils._LE

zHPM_opts = [
    cfg.FloatOpt('zhpm_vm_cpu_capacity_ratio',
                 default=1,
                 help='Ratio used to estimate instance demand capacity, the '
                      'value of ratio is between 0 and 1; when '
                      'zhpm_vm_cpu_capacity_ratio=1, instance is full demand.')
]

CONF = cfg.CONF
CONF.register_opts(zHPM_opts)

LOG = logging.getLogger(__name__)


class zHPMEGOExtentionManager(base.BaseEGOExtentionManager):
    def __init__(self, ego_manager):
        super(zHPMEGOExtentionManager, self).__init__(ego_manager)
        self.wrg_rpcapi = wrg_rpcapi.WRGAPI()
        # Cache for Workload resource group definitions
        # {key: wrg_name, value: wrg definition}
        self.wrg_def_map = {}
        # cache for association of inst with WRG
        self.inst_wrg_map = collections.defaultdict(dict)
        self.zHPM_hosts = set()
        self.unfinished_sync_hosts = set()
        self.__validate_opts()

    def __validate_opts(self):
        ratio = CONF.zhpm_vm_cpu_capacity_ratio
        if ratio <= 0 or ratio > 1:
            LOG.warning(_LW('Invalid value %(ratio)r for '
                       '"zhpm_vm_cpu_capacity_ratio" whose range should be '
                       '(0, 1]. Platform Resource Scheduler will apply it '
                       'with its default value: 1')
                       % {'ratio': ratio})
            CONF.zhpm_vm_cpu_capacity_ratio = 1
        CONF.scheduler_default_filters.append(ego_constants.ZHPM_FILTER)
        CONF.scheduler_default_filters = list(
                                           set(CONF.scheduler_default_filters))

    def sync_up(self):
        self.__ego_construct_wrg_def_map()
        self.__ego_construct_inst_wrg_map()

    def instance_sync_up(self, instance_list):
        inst_uuids = set([vm['uuid'] for vm in instance_list])
        inst_uuids_in_wrg = set(self.inst_wrg_map.keys())
        dissociation_inst = {}
        for inst_deleted in (inst_uuids_in_wrg - inst_uuids):
            # Dissociate WRG for instance which has been deleted in DB.
            inst_name = self.inst_wrg_map[inst_deleted].get(
                                                wrg_consts.VS_NAME)
            dissociation_inst[inst_deleted] = inst_name
        for inst in instance_list:
            inst_uuid = inst['uuid']
            if (inst_uuid not in self.ego_manager.inst_allocid_map and
                inst['vm_state'] != vm_states.SHELVED_OFFLOADED and
                inst_uuid in inst_uuids_in_wrg):
                # Dissociate WRG for instance whose allocation has been
                # deleted except 'shelved_offloaded' instance.
                dissociation_inst[inst_uuid] = (
                    self.inst_wrg_map[inst_uuid].get(wrg_consts.VS_NAME))
        for inst_uuid, inst_name in six.iteritems(dissociation_inst):
            try:
                self.__associate_inst_with_wrg(inst_uuid, None, inst_name)
                LOG.debug('Dissociated workload resource group for instance '
                          '%(vm_name)s (%(vm_uuid)s)' %
                          {'vm_name': inst_name, 'vm_uuid': inst_uuid})
            except Exception as e:
                LOG.warning(_LW("Platform Resource Scheduler could not "
                            "dissociate workload resource group for instance "
                            "%(name)s (%(uuid)s) due to: (%(reason)s)")
                          % {"reason": six.text_type(e),
                             "name": inst_name, "uuid": inst_uuid})
                raise

    def host_sync_up(self):
        zHPM_hosts = set([node['name'] for node in
                self.ego_manager.ego_get_resource_list(
                    'select(($cpu_mgmt==on)&&($zhpm_cooperative_mode==on))')])
        if zHPM_hosts == self.zHPM_hosts and not self.unfinished_sync_hosts:
            return
        zHPM_to_on = zHPM_hosts - self.zHPM_hosts
        zHPM_to_down = self.zHPM_hosts - zHPM_hosts
        self.zHPM_hosts = zHPM_hosts
        need_sync_up_hosts = (
                    zHPM_to_on | zHPM_to_down | self.unfinished_sync_hosts)
        self.unfinished_sync_hosts.clear()
        for egohost in need_sync_up_hosts:
            host_dict = self.ego_manager.all_egohost_attrs.get(egohost)
            if not host_dict:
                LOG.warning(_LW('Cannot find record of %(host)s in DB; Ignore '
                                'it during sync up host info.')
                            % {'host': egohost})
                ego_utils.safe_remove(self.zHPM_hosts, egohost)
                continue
            svc_host = host_dict['service']['host']
            insts = objects.InstanceList.get_by_host(
                                    self.ego_manager.ctxt, svc_host)
            LOG.debug('Host %(host)r has changed to zHPM %(mode)r '
                      'mode.' %
                      {'host': egohost,
                       'mode': 'on' if egohost in self.zHPM_hosts else 'off'})
            for inst in insts:
                # if PRS has not created allocation for the instance,
                # ignore it here and instance sync up will handle it.
                if inst.uuid not in self.ego_manager.inst_allocid_map:
                    continue
                if inst.task_state:
                    # do sync up in next cycle
                    self.unfinished_sync_hosts.add(egohost)
                    continue
                resreq = self.ego_manager.inst_resreq_map.get(inst.uuid)
                is_zHPM_resreq = ego_resreq_utils.is_zHPM_resreq(resreq)
                if ((is_zHPM_resreq and egohost not in self.zHPM_hosts) or
                         (not is_zHPM_resreq and egohost in self.zHPM_hosts)):
                    try:
                        self.__ego_update_decision(inst)
                    except Exception as e:
                        LOG.warning(_LW("Platform Resource Scheduler could "
                                "not update EGO allocation of instance "
                                "%(name)s (%(uuid)s) for zHPM mode of host "
                                "%(host)s changed to %(mode)r due to: "
                                "(%(reason)s)")
                           % {"reason": six.text_type(e),
                              "name": inst.hostname,
                              "uuid": inst.uuid,
                              "mode":
                                 'on' if egohost in self.zHPM_hosts else 'off',
                              "host": egohost})
                        self.unfinished_sync_hosts.add(egohost)

    def handle_for_schedule_info(self, filter_properties):
        spec = filter_properties.get('request_spec', {})
        operation_type = spec.get('operation_type')
        inst_uuid = spec.get('instance_id')

        wrg_name = None
        if operation_type in ([ego_constants.FILTER_WORKING_SCOPE_CREATE,
                               ego_constants.FILTER_WORKING_SCOPE_UNSHELVE,
                               ego_constants.FILTER_WORKING_SCOPE_RESIZE]):
            instance_type = spec.get('instance_type') or {}
            extra_specs = instance_type.get('extra_specs') or {}
            wrg_name = extra_specs.get(ego_constants.ZHPM_WRG_NAME)
        elif not operation_type:
            # for PRS current cases, when operation_type is None, force_hosts
            # is always a non-empty list and the VM has been on the
            # force_hosts[0].
            force_hosts = filter_properties.get('force_hosts')
            if force_hosts[0] not in self.zHPM_hosts:
                return
            wrg_from_flavor = spec.get('instance_type', {}).get(
                            'extra_specs', {}).get(ego_constants.ZHPM_WRG_NAME)
            wrg_name = wrg_from_flavor or wrg_consts.DEFAULT_WRG_NAME
        else:
            if (self.ego_manager.inst_egohost_map.get(inst_uuid) in
                                                        self.zHPM_hosts):
                wrg_name = self.inst_wrg_map.get(inst_uuid).get(
                                                   wrg_consts.WRG_NAME)
        if wrg_name:
            wrg_def = self._get_wrg_def(wrg_name)
            spec['wrg_def'] = wrg_def
            spec[ego_constants.STAND_ALONE] = False

    def handle_for_compute_info(self, compute):
        metrics = compute.get('metrics')
        hypervisor_hostname = compute.get('hypervisor_hostname')
        ego_host_name = ego_resreq_utils.get_rpi_res_name(
                                                hypervisor_hostname)
        if not metrics:
            return
        zhpm_cpu_mgmt = (
            metrics.get(ego_constants.ZHPM_CPU_MGMT_METRICS_NAME))
        zhpm_cooperative_mode = (
            metrics.get(ego_constants.ZHPM_COOPERATIVE_MODE_METRICS_NAME))
        if zhpm_cpu_mgmt and zhpm_cooperative_mode:
            # Convert metrics
            ego_resreq_utils.convert_zhpm_placement_metrics(metrics)
            if (zhpm_cpu_mgmt.get(ego_constants.ZHPM_CPU_MGMT) == 'on' and
                        zhpm_cooperative_mode.get(
                            ego_constants.ZHPM_COOPERATIVE_MODE) == 'on'):
                # Construct 'inst_schedule_info_map' to store the map between
                # instance uuid and its real capacity in each importance level
                # for current hypervisor host, sample as below:
                # {'highestHighest': {'11111':2.0, "22222":1.0},
                # 'highestHigh': {'33333':1.0},
                # ...}
                inst_schedule_info_map = {}
                resreq_map = self.ego_manager.inst_resreq_map
                maxslots_map = self.ego_manager.inst_maxslots_map
                egohost_map = self.ego_manager.inst_egohost_map
                for inst in resreq_map:
                    if (not maxslots_map.get(inst) or
                                egohost_map.get(inst) != ego_host_name):
                        continue
                    consume_str = (ego_resreq_utils.get_criteria_str_by_resreq(
                        resreq_map.get(inst), 'consume'))
                    (importance, capacity) = (
                        ego_resreq_utils.parse_zhpm_importance_level(
                                consume_str))
                    if importance in ego_constants.IMPORTANCE_LEVELS:
                        inst_cap = {inst: capacity}
                        if inst_schedule_info_map.get(importance):
                            inst_schedule_info_map[importance].update(inst_cap)
                        else:
                            inst_schedule_info_map[importance] = inst_cap
                # Adjust metrics
                ego_resreq_utils.adjust_zhpm_placement_metrics(
                                    metrics, inst_schedule_info_map)

    def consume_inst_from_egohost(self, instid, ego_host,
                                  resreq, old_resreq, compute,
                                  consume_free_flag):
        """
        Adjust the CPU total capacity for zhpm case once instance get
        removed from a host, since ego is fully reserved the request
        resource on the host, but actually it just consume the real capacity
        so nova scheduler should adjust the total capacity based on below
        formula :
            remove/migrate from host:
                total_capacity = total_capacity - (request - real)
            resize to same or different host:
                total_capacity = total_capacity - (old_request - request)
        """
        if (not consume_free_flag and ego_host in self.zHPM_hosts):
            metrics = compute.get('metrics')
            if not metrics:
                return
            cons_str = ego_resreq_utils.get_criteria_str_by_resreq(
                                                        resreq, 'consume')
            (impt_level, request) = (
                ego_resreq_utils.parse_zhpm_importance_level(cons_str))
            if old_resreq:
                # Resize to same or different host
                old_cons_str = ego_resreq_utils.get_criteria_str_by_resreq(
                                                        old_resreq, 'consume')
                (old_impt_level, old_request) = (
                    ego_resreq_utils.parse_zhpm_importance_level(old_cons_str))
                if old_impt_level not in ego_constants.IMPORTANCE_LEVELS:
                    return
                current_host = self.ego_manager.inst_egohost_map[instid]
                if (current_host == ego_host and old_impt_level == impt_level
                        and old_request == request):
                    # For resize to same host case, if importance level and
                    # capacity request not change, no need to adjust total
                    # capacity
                    return
                index = ego_constants.IMPORTANCE_LEVELS.index(
                        old_impt_level)
                real_cap = self._get_inst_real_capcity_from_metrics(metrics,
                                                    old_impt_level, instid)
                delta_cap = float(old_request) - float(real_cap)
            else:
                # Remove/migrate from host
                index = ego_constants.IMPORTANCE_LEVELS.index(impt_level)
                real_cap = self._get_inst_real_capcity_from_metrics(metrics,
                                                    impt_level, instid)
                delta_cap = float(request) - float(real_cap)
            for impl in ego_constants.IMPORTANCE_LEVELS[index:]:
                metrics[impl] -= delta_cap

    def _get_inst_real_capcity_from_metrics(self, metrics, importance_level,
                                            instance_id):
        real_caps = metrics.get(
                importance_level + ego_constants.VS_CPU_CAPACITY_SUFFIX)
        if real_caps:
            real_caps_dict = ego_resreq_utils.convert_vs_caps_to_dict(
                 real_caps)
            return real_caps_dict.get(instance_id, 0)
        return 0

    def instance_create(self, filter_properties):
        spec = filter_properties['request_spec']
        wrg_def = spec.get("wrg_def")
        inst_uuid = spec['instance_id']
        instance_name = spec['instance_properties']['hostname']
        if wrg_def:
            wrg_name = wrg_def[wrg_consts.WRG_INFO][wrg_consts.WRG_NAME]
            self.__associate_inst_with_wrg(inst_uuid,
                                    wrg_name, instance_name)

    def instance_delete(self, instance_uuid):
        if instance_uuid in self.inst_wrg_map:
            # Only delete the instance's WRG info when the instance has been
            # deleted in DB. For shelve_offload case, we must make sure the
            # WRG which this VM specified when booting cannot be deleted,
            # otherwise it will fail to unsheve since its WRG has been deleted.
            # We can make it by leaving the inst WRG map of shelve_offload VM.
            try:
                # We can delete a VM's inst and WRG map if the WRG is PRS
                # default WRG since this WRG cannot be deleted even there is
                # no VM assigned it.
                if (self.inst_wrg_map[instance_uuid].get(
                          wrg_consts.WRG_NAME) != wrg_consts.DEFAULT_WRG_NAME):
                    inst = objects.Instance.get_by_uuid(self.ego_manager.ctxt,
                                                 instance_uuid)
                    if inst.vm_state == vm_states.SHELVED_OFFLOADED:
                        return
            except exception.InstanceNotFound:
                pass
            try:
                inst_wrg = self.inst_wrg_map[instance_uuid]
                inst_name = inst_wrg[wrg_consts.VS_NAME]
                self.__associate_inst_with_wrg(instance_uuid, None, inst_name)
            except Exception:
                LOG.error(_LE('Failed to dissociate workload resource group '
                              'for virtual instance %r') % instance_uuid)

    def instance_update(self, filter_properties):
        spec = filter_properties['request_spec']
        inst_uuid = spec['instance_id']
        instance_name = spec['instance_properties']['hostname']
        operation_type = spec.get("operation_type")
        wrg_def = spec.get("wrg_def", {})
        inst_wrg = self.inst_wrg_map.get(inst_uuid, {})
        vm_state = spec['instance_properties']['vm_state']
        if not (wrg_def or (vm_state != vm_states.RESIZED and
                                wrg_consts.OLD_WRG in inst_wrg)):
            return
        new_wrg_name = wrg_def.get(wrg_consts.WRG_INFO, {}).get(
                                                  wrg_consts.WRG_NAME)
        if operation_type == ego_constants.FILTER_WORKING_SCOPE_RESIZE:
            old_wrg_name = inst_wrg.get(wrg_consts.WRG_NAME, '')
        elif vm_state == vm_states.RESIZED:
            old_wrg_name = inst_wrg.get(wrg_consts.OLD_WRG, '')
        else:
            old_wrg_name = None
        try:
            if old_wrg_name is None:
                self.__associate_inst_with_wrg(inst_uuid,
                                           new_wrg_name,
                                           instance_name)
            else:
                self.__associate_inst_with_wrg(inst_uuid,
                                           new_wrg_name,
                                           instance_name,
                                           old_wrg=old_wrg_name)
        except Exception:
            LOG.error(_LE("Failed to update association of workload resource "
                          "group for virtual instance %r") % inst_uuid)
            raise

    def handle_notification(self, msg):
        event_type = msg['event_type']
        payload = msg['payload']
        if event_type == wrg_consts.PRS_WRG_CREATE_END:
            self.__wrg_create(payload)
        elif event_type == wrg_consts.PRS_WRG_DELETE_END:
            self.__wrg_delete(payload)
        elif event_type == wrg_consts.PRS_WRG_UPDATE_END:
            self.__wrg_update(payload)
        elif event_type == ego_constants.EVENT_TYPE_RESIZE_CONFIRM_END:
            self.__resize_confirm(payload)
        elif event_type == ego_constants.EVENT_TYPE_COMPUTE_INST_DELETE_END:
            # just for delete a shelved_offloaded VM case whose allocation
            # has been deleted when shelve it.
            self.instance_delete(payload['uuid'])

    def __wrg_create(self, wrg_def):
        wrg_name = wrg_def[wrg_consts.WRG_INFO][wrg_consts.WRG_NAME]
        self.wrg_def_map[wrg_name] = wrg_def

    def __wrg_update(self, wrg_def):
        wrg_name = wrg_def[wrg_consts.WRG_INFO][wrg_consts.WRG_NAME]
        # 1. update local WRG definition cache.
        self.wrg_def_map[wrg_name] = wrg_def
        # 2. update all instances' allocation consume str in the WRG
        #    if consume str should change.
        for inst_uuid, wrg_info in six.iteritems(self.inst_wrg_map):
            if (wrg_info[wrg_consts.WRG_NAME] != wrg_name or
                    self.ego_manager.inst_egohost_map.get(inst_uuid) not in
                        self.zHPM_hosts):
                continue
            try:
                inst = objects.Instance.get_by_uuid(self.ego_manager.ctxt,
                                inst_uuid,
                                expected_attrs=['metadata', 'system_metadata'])
                self.__ego_update_decision(inst)
            except exception.InstanceNotFound:
                # The inst may have been deleted in DB; PRS will handle
                # it when accepting instance delete end notification.
                continue
            except Exception as e:
                LOG.warning(_LW('Platform Resource Scheduler could not '
                                'update EGO resource decision for instance '
                                '(%(instanceid)s) due to: (%(reason)s)')
                              % {"instanceid": inst_uuid,
                                 "reason": six.text_type(e)})

    def __wrg_delete(self, wrg_def):
        wrg_name = wrg_def[wrg_consts.WRG_INFO][wrg_consts.WRG_NAME]
        ego_utils.safe_remove(self.wrg_def_map, wrg_name)

    def __resize_confirm(self, instance):
        inst_uuid = instance['uuid']
        inst_wrg = self.inst_wrg_map.get(inst_uuid, {})
        if inst_wrg:
            wrg_name = inst_wrg.get(wrg_consts.WRG_NAME)
            old_wrg_name = inst_wrg.get(wrg_consts.OLD_WRG)
            if old_wrg_name:
                try:
                    self.__associate_inst_with_wrg(inst_uuid, wrg_name,
                                           instance['hostname'])
                except Exception:
                    LOG.error(_LE('Failed to dissociate workload resource '
                                  'group %(old_wrg)s for virtual instance '
                                  '%(vs_name)r (%(uuid)s).')
                              % {'uuid': inst_uuid,
                                 'old_wrg': old_wrg_name,
                                 'vs_name': instance['hostname']})
                    raise

    def __associate_inst_with_wrg(self, inst_uuid, wrg_name,
                                  instance_name, **kwargs):
        if wrg_name:
            body = {wrg_consts.VS_NAME: instance_name,
                    wrg_consts.WRG_NAME: wrg_name}
        else:
            body = {wrg_consts.VS_NAME: instance_name}
        if kwargs:
            body.update(kwargs)
        if self.inst_wrg_map.get(inst_uuid) == body:
            return
        try:
            updated_inst_wrg = (
                self.wrg_rpcapi.associate_inst_with_wrg(self.ego_manager.ctxt,
                                                    inst_uuid, wrg_name,
                                                    instance_name, **kwargs))
        except oslo_exceptions.MessagingTimeout:
            LOG.error(_LE("Service 'nova-ibm-ego-workload-resource-group' "
                          "RPC response time out."))
            raise wrg_exception.AssociateVSWRGFailed(uuid=inst_uuid)
        except Exception as e:
            if e.exc_type == 'WRGNotFoundByName':
                # this exception can occur when WRG related notification
                # missed, so it need do sync up operation.
                self.sync_up()
                raise wrg_exception.AssociateVSWRGFailed(uuid=inst_uuid)
            raise
        if updated_inst_wrg is None:
            self.inst_wrg_map.pop(inst_uuid)
        else:
            self.inst_wrg_map[inst_uuid] = updated_inst_wrg

    def __ego_construct_wrg_def_map(self):
        """
        Used for sync up Workload Resource Group definition with
        zHPM when scheduler start/restart
        """
        service_status = ego_resreq_utils.get_service_status_by_topic(
                                            self.ego_manager.ctxt,
                                            CONF.workload_resource_group.topic)
        if not (service_status and service_status['is_up']):
            LOG.error(_LE("Service 'nova-ibm-ego-workload-resource-group' "
                          "is not available now; Please start it for "
                          "zKVM hypervisors with zHPM enabled in the system."))
            raise exception.ServiceUnavailable()
        else:
            all_wrg = {}
            try:
                all_wrg = self.wrg_rpcapi.get_all(self.ego_manager.ctxt)
            except oslo_exceptions.MessagingTimeout:
                LOG.error(_LE("Service 'nova-ibm-ego-workload-resource-group' "
                              "RPC response time out."))
                raise exception.ServiceUnavailable()
            if all_wrg:
                self.wrg_def_map.clear()
                for wrg in all_wrg['workload_resource_groups']:
                    wrg_name = wrg[wrg_consts.WRG_INFO][wrg_consts.WRG_NAME]
                    self.wrg_def_map[wrg_name] = wrg

    def __ego_construct_inst_wrg_map(self):
        service_status = ego_resreq_utils.get_service_status_by_topic(
                                            self.ego_manager.ctxt,
                                            CONF.workload_resource_group.topic)
        if not (service_status and service_status['is_up']):
            LOG.error(_LE("Service 'nova-ibm-ego-workload-resource-group' "
                          "is not available now; Please start it for "
                          "zKVM hypervisors with zHPM enabled in the system."))
            raise exception.ServiceUnavailable()
        else:
            try:
                self.inst_wrg_map = self.wrg_rpcapi.get_inst_wrg_map(
                                                        self.ego_manager.ctxt)
            except oslo_exceptions.MessagingTimeout:
                LOG.error(_LE("Service 'nova-ibm-ego-workload-resource-group' "
                              "RPC response time out."))
                raise exception.ServiceUnavailable()

    def __ego_update_decision(self, inst):
        egohost = ego_resreq_utils.get_rpi_res_name(inst.node)
        inst_uuid = inst.uuid
        max_slot = self.ego_manager.inst_maxslots_map.get(inst_uuid, 1)
        instance_type = obj_base.obj_to_primitive(inst.get_flavor())
        inst = obj_base.obj_to_primitive(inst)
        filter_properties = {
                            'force_hosts': [egohost],
                            'max_slot': max_slot,
                            'force_realloc_migrate': True,
                            'ignore_filters': True,
                            'request_spec':
                              {'instance_id': inst_uuid,
                               'instance_type': instance_type,
                               'instance_properties': inst}}
        self.ego_manager.ego_update_decision(self.ego_manager.ctxt,
                                             filter_properties)

    def _get_wrg_def(self, wrg_name):
        if wrg_name not in self.wrg_def_map:
            self.__ego_construct_wrg_def_map()
        wrg = self.wrg_def_map.get(wrg_name)
        if wrg is None:
            errmsg = (_LE('Workload resource group %(name)s '
                          'does not exist') % {'name': wrg_name})
            LOG.error(errmsg)
            raise ego_exception.InvalidWorkloadResourceGroup(
                name=wrg_name, reason=errmsg)
        else:
            return wrg

    def check_subdemands_for_heat(self, ctxt, operation_type, sub_demands):
        for sub_demand in sub_demands:
            if sub_demand['filter_properties']['request_spec'][
                    'instance_type']['extra_specs'].\
                    get(ego_constants.ZHPM_WRG_NAME):
                flavor_name = sub_demand[
                    'filter_properties']['request_spec']['instance_type'].\
                        get('name')
                error_msg = (_LE("The heat resource 'IBM::Policy::Group' "
                                 "doesn't support zHPM technology, please "
                                 "remove the WorkLoad Resource Group name "
                                 "from flavor (%s)'s metadata and try it "
                                 "again.") % (flavor_name))
                LOG.error(error_msg)
                if operation_type == 'create':
                    raise ego_exception.EGOExpCreateAllocFailed(
                        reason=error_msg)
                else:
                    raise ego_exception.EGOExpResizeAllocFailed(
                        reason=error_msg)
