# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
import collections

import six

from nova import exception
from nova.ibm.ego import ego_constants
from nova.ibm.ego import ego_gettextutils
from nova.ibm.ego import ego_utils
from nova.scheduler.ibm.ego import ego_resreq_utils
from nova.scheduler.ibm.ego import filters
from nova.scheduler.ibm.ego.filters import json_filter

from oslo_config import cfg
from oslo_log import log as logging

_ = ego_gettextutils._

LOG = logging.getLogger(__name__)

CONF = cfg.CONF
CONF.import_opt('scheduler_available_filters', 'nova.scheduler.host_manager')


class EGOFilterDriver(object):
    """
    EGO Filter Driver that used to construct ego resource requirement
    """
    def __init__(self):
        self.all_filters = []
        # self.resreq_cache will persist the filter constraints for multi
        # instances creation case, the data structure as below:
        # {'request-id': {
        #   'filtername':{'samehost':[],'diffhost':[],'ignore_hosts':[],
        #                 'force_hosts':[],'select':[]},
        #   'filtername':{...},...}}
        self.resreq_cache = {}
        # self.schedule_cache will persist the schedule cache for
        # instance, the data structure is like:
        # {instance_id:
        #     {'scope':
        #         {
        #             "force_hosts": []
        #             "schedule time 1":
        #                 {
        #                  'filter_reqs' : {fitler1:{},
        #                                   filter2:{},
        #                                  ...
        #                                 },
        #                  'consume_req' : "",
        #                  'reserveon_req' : "",
        #                  'select_host' : [],
        #                  'error_msg': "",
        #                  'resreq': "",
        #                  'select_host': []
        #                 },
        #             "schedule time 2" : {},
        #             ...
        #         }
        #     }
        # }
        self.schedule_cache = {}
        self.json_filter_utils = json_filter.JsonFilter()
        self.filter_handler = filters.HostFilterHandler()
        # Only load the user defined filter classes when the value of the
        # scheduler_available_filters is not the default value
        if (['nova.scheduler.filters.all_filters'] !=
              CONF.scheduler_available_filters):
            self.all_filters.extend(self.filter_handler.get_matching_classes(
                    CONF.scheduler_available_filters))
        self.all_filters.extend(self.filter_handler.get_matching_classes(
                ["nova.scheduler.ibm.ego.filters.all_filters"]))
        self.all_filters = self.all_filters

    def choose_host_filters(self, filters, scope):
        if filters is None:
            filters = CONF.scheduler_default_filters[:]
        if not isinstance(filters, list):
            filters = [filters]
        # Add the built-in filters.
        builtin_filters = [ego_constants.MAINTENANCE_FILTER,
                           ego_constants.INSTANCE_METADATA_FILTER,
                           ego_constants.VMWARE_DATACENTER_FILTER]
        filters = list(set(filters + builtin_filters))
        host_filters = self.__build_host_filters(filters, scope)
        LOG.debug("Host filters for scheduling are (%s)." % host_filters)
        return host_filters

    def __build_host_filters(self, filters, scope):
        host_filters = []
        for f in filters:
            for cls in self.all_filters:
                if cls.__name__ == f:
                    obj = cls()
                    filter_scope = obj.filter_scope
                    if scope in filter_scope:
                        filter_func = getattr(obj, 'parse', None)
                        if filter_func:
                            host_filters.append((filter_func, obj))
        return host_filters

    def schedule_request_completed(self, ctxt):
        """Called when all scheduling tasks of the same request have finished
        """
        ego_utils.safe_remove(self.resreq_cache, ctxt.request_id)

    def __parse_filters(self, ctxt, filter_properties, all_egohost_attrs,
                        host_egohost, resreq, scope, index=0):
        """
        This function will do following things:
          1. Get the VM constraints in the filter that were defined
             in nova.conf scheduler_default_filters
          2. Convert the JSON syntax resreq to EGO syntax resreq if need
          3. Merge resreq for 'select', 'samehost' and 'diffhost' statements
        """
        def __convert_to_ego_hosts(hosts):
            ego_hosts = []
            for host in hosts:
                if host in host_egohost:
                    ego_hosts.extend(list(host_egohost[host]))
                elif host in all_egohost_attrs:
                    ego_hosts.append(host)
                else:
                    for egohost, attrs in six.iteritems(all_egohost_attrs):
                        if attrs.get('hypervisor_hostname') == host:
                            ego_hosts.append(
                                ego_resreq_utils.get_rpi_res_name(host))
                            break
                    else:
                        ego_hosts.append(host)
            return ego_hosts

        select_str = ''
        samehost_str = ''
        diffhost_str = ''
        extra_select_str = ''
        ret_resreq = {'select': resreq.get('select', []),
                      'samehost': resreq.get('samehost', []),
                      'diffhost': resreq.get('diffhost', []),
                      'extra_select': resreq.get('extra_select', [])}
        all_egohosts = all_egohost_attrs.keys()
        force_hosts = filter_properties.get('force_hosts') or []
        force_nodes = filter_properties.get('force_nodes') or []
        force_hosts.extend(force_nodes)
        ignore_hosts = filter_properties.get('ignore_hosts') or []
        if force_hosts:
            force_hosts = __convert_to_ego_hosts(force_hosts)
            if not (set(all_egohosts) & set(force_hosts)):
                raise exception.NoValidHost(
                          reason=ego_constants.FILTER_ERROR_MESSAGE,
                          code=ego_constants.NOT_ENOUGH_RESOURCES_CODE)
        if ignore_hosts:
            ignore_hosts = __convert_to_ego_hosts(ignore_hosts)
            if not (set(all_egohosts) - set(ignore_hosts)):
                raise exception.NoValidHost(
                          reason=ego_constants.FILTER_ERROR_MESSAGE,
                          code=ego_constants.NOT_ENOUGH_RESOURCES_CODE)

        # go through all filters
        if index == 0:
            self.resreq_cache[ctxt.request_id] = {}
        resreq_cache = self.resreq_cache[ctxt.request_id]
        filter_fns = self.choose_host_filters(None, scope)
        for filter_fn, obj in filter_fns:
            extra_select_list = []
            select_list = []
            filter_name = filter_fn.im_class.__name__
            # if the current filter will filter all the hosts based on
            # an initial assessment, no_valid_host_flag will be True and
            # exception.NoValidHost would be raised
            no_valid_host_flag = False
            if index == 0 or not obj.run_filter_once_per_request:
                try:
                    LOG.debug('Attempting to run filter: %(filter)s'
                              % {'filter': filter_name})
                    filter_fn(filter_properties, all_egohost_attrs)
                except Exception as ex:
                    LOG.error(_('Ran filter %(filter)s, but encountered '
                                'an error: %(reason)s')
                              % {'filter': filter_name,
                                 'reason': six.text_type(ex)})
                    raise
                filter_resreq = obj.filter_resreq
                filter_hosts = obj.filter_hosts
                # get the ignore_hosts and force_hosts from user-defined
                # filter and merge to ignore_hosts and force_hosts
                filter_ignore_hosts = filter_hosts.get('ignore_hosts')
                if filter_ignore_hosts:
                    filter_ignore_hosts = __convert_to_ego_hosts(
                                              filter_ignore_hosts)
                    ignore_hosts.extend(filter_ignore_hosts)
                if not (set(all_egohosts) - set(ignore_hosts)):
                    no_valid_host_flag = True
                if not (set(all_egohosts) - set(filter_ignore_hosts)):
                    no_valid_host_flag = True

                filter_force_hosts = filter_hosts.get('force_hosts')
                if isinstance(filter_force_hosts, list):
                    # force_hosts is the empty set, which indicates that no
                    # hosts satisfied this filter
                    if not filter_force_hosts:
                        no_valid_host_flag = True
                    else:
                        filter_force_hosts = __convert_to_ego_hosts(
                                                    filter_force_hosts)
                        if force_hosts:
                            mixed_force_hosts = list(
                               set(force_hosts) & set(filter_force_hosts))
                            if mixed_force_hosts:
                                force_hosts = mixed_force_hosts
                            else:
                                no_valid_host_flag = True
                        else:
                            force_hosts.extend(filter_force_hosts)
                        if not (set(all_egohosts) & set(force_hosts)):
                            no_valid_host_flag = True

                def convert_list(src_list):
                    dst_list = []
                    if src_list:
                        # The syntax of resreq
                        # case 1: ['[">=", "$free_ram_mb", 512]']
                        # case 2: ['[">=", "$vcpu", 4]',
                        #          '[">=","$memSize", 512]']
                        # case 3: ['activeCompute==1']
                        # case 4: ['["not","xhu_node1","xhu_node2"]',
                        #          'disk_total > 5', 'disk_available > 6']
                        for s in src_list:
                            if s.startswith('['):
                                # convert JSON syntax resreq to EGO syntax
                                filter_properties['json_query'] = s
                                self.json_filter_utils.parse(filter_properties,
                                                             all_egohost_attrs)
                                dst_list.extend(
                                    self.json_filter_utils.filter_resreq.get(
                                                                'select', []))
                            else:
                                dst_list.append(s)
                    return dst_list
                if (not obj.extra_resreq_for_create and
                        scope == ego_constants.FILTER_WORKING_SCOPE_CREATE):
                    extra_select_list = filter_resreq.get('select', [])
                    extra_select_list = convert_list(extra_select_list)
                else:
                    select_list = filter_resreq.get('select', [])
                    select_list = convert_list(select_list)
                samehost_list = filter_resreq.get('samehost', [])
                diffhost_list = filter_resreq.get('diffhost', [])

                ret_resreq['select'].extend(select_list)
                ret_resreq['extra_select'].extend(extra_select_list)
                ret_resreq['samehost'].extend(samehost_list)
                ret_resreq['diffhost'].extend(diffhost_list)

                LOG.debug('The VM constraints in (%(scope)s) '
                          'filter (%(filter)s) are: '
                          'ignore hosts list is (%(ignore_hosts)s), '
                          'force hosts list is (%(force_hosts)s), '
                          'select string is (%(select)s), '
                          'extra select string is (%(extra_select)s).'
                          % {'scope': scope,
                             'filter': filter_name,
                             'ignore_hosts': filter_ignore_hosts,
                             'force_hosts': filter_force_hosts,
                             'select': select_list,
                             'extra_select': extra_select_list})

                resreq_cache[filter_name] = {}
                resreq_cache[filter_name]['ignore_hosts'] = (
                    filter_ignore_hosts)
                resreq_cache[filter_name]['force_hosts'] = (
                    filter_force_hosts)
                select_list.extend(extra_select_list)
                resreq_cache[filter_name]['select'] = (
                    select_list)
                resreq_cache[filter_name]['samehost'] = (
                    samehost_list)
                resreq_cache[filter_name]['diffhost'] = (
                    diffhost_list)
            else:
                # i.e., index != 0 and obj.run_filter_once_per_request is True
                ignore_hosts.extend(
                    resreq_cache[filter_name]['ignore_hosts'])
                filter_force_hosts_cache = (
                    resreq_cache[filter_name]['force_hosts'])
                if force_hosts and filter_force_hosts_cache:
                    force_ego_hosts = (
                            ego_resreq_utils.convert_hosts_to_egohosts(
                                                force_hosts, host_egohost))
                    mixed_force_hosts = list(
                        set(force_ego_hosts) & set(filter_force_hosts_cache))
                    if mixed_force_hosts:
                        force_hosts = mixed_force_hosts
                    else:
                        no_valid_host_flag = True
                elif isinstance(filter_force_hosts_cache, list):
                    force_hosts.extend(filter_force_hosts_cache)
                ret_resreq['select'].extend(
                    resreq_cache[filter_name]['select'])
                ret_resreq['samehost'].extend(
                    resreq_cache[filter_name]['samehost'])
                ret_resreq['diffhost'].extend(
                    resreq_cache[filter_name]['diffhost'])

            # If no_valid_host_flag is True, save troubleshooting message and
            # raise exception.NoValidHost
            if no_valid_host_flag:
                filter_reqs = {'filter_reqs': {}}
                filter_reqs['filter_reqs'].update(resreq_cache)
                self.schedule_cache_update(filter_properties, filter_reqs)
                raise exception.NoValidHost(
                          reason=obj.filter_error_message,
                          code=ego_constants.NOT_ENOUGH_RESOURCES_CODE)
        # End of loop "for filter_fn, obj in filter_fns" - i.e., done iterating
        # over the filters

        # Update schedule cache by filters' requirement saved in
        # resreq_cache
        filter_reqs = {'filter_reqs': {}}
        filter_reqs['filter_reqs'].update(resreq_cache)
        self.schedule_cache_update(filter_properties, filter_reqs)

        if ignore_hosts:
            ignore_hosts = list(set(ignore_hosts))
            filter_properties['ignore_hosts'] = ignore_hosts
        if force_hosts:
            force_hosts = list(set(force_hosts))
            filter_properties['force_hosts'] = force_hosts
        if ret_resreq['select']:
            ret_resreq['select'] = list(set(ret_resreq['select']))
        if ret_resreq['extra_select']:
            ret_resreq['extra_select'] = list(set(ret_resreq['extra_select']))
        if ret_resreq['samehost']:
            ret_resreq['samehost'] = list(set(ret_resreq['samehost']))
        if ret_resreq['diffhost']:
            ret_resreq['diffhost'] = list(set(ret_resreq['diffhost']))

        select_str = ego_resreq_utils.get_resreq_str(
            ret_resreq.get('select', []), 'select')
        extra_select_str = ego_resreq_utils.get_resreq_str(
            ret_resreq.get('extra_select', []), 'select')
        samehost_str = ego_resreq_utils.get_resreq_str(
            ret_resreq.get('samehost', []), 'samehost')
        diffhost_str = ego_resreq_utils.get_resreq_str(
            ret_resreq.get('diffhost', []), 'diffhost')

        return (select_str, samehost_str, diffhost_str, extra_select_str)

    def __convert_host_attrs(self, host_attrs_old):
        """
        :param host_attrs_old: {host1: {'compute': {k : v}, 'cstr': cstr},
                                host2: {}}
        :returns: {host1: {k : v}, host2: {}}
        """
        host_attrs_new = collections.defaultdict(dict)
        for ego_host, attrs in host_attrs_old.items():
            for value in six.itervalues(attrs):
                if isinstance(value, dict):
                    host_attrs_new[ego_host].update(value)
        return host_attrs_new

    def get_ego_filter_res_req(self, ctxt, filter_properties,
                               all_egohost_attrs, host_egohost, scope,
                               valid_aggregate_ids=None, index=0):
        """
        Construct EGO resreq according to filters
        :param filter_properties: filter properties
        :param all_egohost_attrs: All hosts attribute in the cluster
                              {host1: {'compute': {k : v}, 'cstr': cstr}
                              host2: {}}
        :param host_egohost: map from service host to ego host set
        :param scope: filter working scope, such as 'create',
                      'live-migration', 'cold-migration' and 'resize'
        :param valid_aggregate_ids: all valid aggregate ids in the system
        :param index: The index of the instance when boot multi instances
                      using '--num-instances' option.

        :returns: EGO resource requirement string,
                  reserveon(OpenStackHosts)consume(vcpu=2:memSize=2048)
                  select(('vcpu'>4)&&('location'=='Xian'))
                  samehost('029bea32-f82d-4258-afcc-7283abb4b66d')
                  diffhost('5fb154c0-846a-40bf-bc88-437cbee62d49')
                  select((usable_disk_mb>=0))
        """
        newalloc_list = [ego_constants.FILTER_WORKING_SCOPE_CREATE,
                         ego_constants.FILTER_WORKING_SCOPE_UNSHELVE]

        valid_aggregate_ids = valid_aggregate_ids or []
        all_egohost_attrs = self.__convert_host_attrs(all_egohost_attrs)
        # Check for PowerVM specification
        hypervisor_data = ego_resreq_utils.get_hypervisor_data(
            filter_properties)
        hypervisor_type = hypervisor_data['hypervisor_type']
        img_arch = hypervisor_data['architecture']

        # For ESXi host case
        image_prop = filter_properties.get('request_spec',
            {}).get('image', {}).get('properties', {})
        prs_hypervisor_type = image_prop.get('prs_hypervisor_type')
        if (prs_hypervisor_type == ego_constants.HYPERVISOR_TYPE_ESX_IMAGE):
            resreq = {
                    'select':
                    ['$hypervisor_type=='
                    + ego_constants.HYPERVISOR_TYPE_ESX_RPI],
                    'samehost': [], 'diffhost': [], 'extra_select': []}
        else:
            if hypervisor_type in ego_constants.POWERVM_HYPERVISOR_TYPES:
                resreq = {
                    'select':
                    ["$hypervisor_type <in> '%s'"
                     % ",".join(ego_constants.POWERVM_HYPERVISOR_TYPES)],
                    'samehost': [], 'diffhost': [], 'extra_select': []}
            elif (hypervisor_type ==
                  ego_constants.HYPERVISOR_TYPE_VMWARE_IMAGE):
                resreq = {
                        'select':
                        ['$hypervisor_type=='
                        + ego_constants.HYPERVISOR_TYPE_VMWARE_RPI],
                        'samehost': [], 'diffhost': [], 'extra_select': []}
            elif (hypervisor_type == ego_constants.HYPERVISOR_TYPE_QEMU and
                    img_arch == ego_constants.IMAGE_ARCH_PPC64):
                # for PowerKVM case
                select = ['$hypervisor_type=='
                         + ego_constants.HYPERVISOR_TYPE_QEMU]
                image_str = "".join(("'" + img_arch + "'",
                                    ' <in> $supported_instances'))
                select.append(image_str)
                resreq = {'select': select, 'samehost': [], 'diffhost': [],
                          'extra_select': []}
            else:
                resreq = {'select': [], 'samehost': [], 'diffhost': [],
                          'extra_select': []}

        samehost_str = ''
        diffhost_str = ''
        select_str = ''
        extra_select_str = ''
        ego_res_req = ''
        reserveon_str = ''
        consume_str = ''

        # Don't create reserveon_str and consume_str if RetryFilter is working.
        if scope in newalloc_list and (
                   not filter_properties.get('retry4create', False)):
            reserveon_str = ego_resreq_utils.get_reserveon_str(
                filter_properties, valid_aggregate_ids)
            consume_str = ego_resreq_utils.get_consume_str(
                                            filter_properties)

        scheduler_hints = filter_properties.get('scheduler_hints') or {}
        native_ego_resreq = scheduler_hints.get('resreq', [])

        # Only support one resreq
        if native_ego_resreq or len(native_ego_resreq) > 0:
            # handle native ego resource requirement
            if isinstance(native_ego_resreq, six.string_types):
                if (native_ego_resreq != 'samehost' and
                      native_ego_resreq != 'diffhost'):
                    ego_res_req = native_ego_resreq
                else:
                    (select_str, samehost_str, diffhost_str,
                            extra_select_str) = (self.__parse_filters(
                                                ctxt,
                                                filter_properties,
                                                all_egohost_attrs,
                                                host_egohost,
                                                resreq,
                                                scope,
                                                index))
            else:
                msg = _('The EGO filter does not support more than one '
                        'native resource requirement and will ignore '
                        'all additional resource requirements.')
                LOG.warn(msg)
        else:
            (select_str, samehost_str, diffhost_str, extra_select_str) = (
                        self.__parse_filters(ctxt, filter_properties,
                                             all_egohost_attrs,
                                             host_egohost,
                                             resreq,
                                             scope,
                                             index))

        resreqstr = "".join((reserveon_str,
                             consume_str,
                             select_str,
                             samehost_str,
                             diffhost_str,
                             ego_res_req))
        # to make sure 'resreqstr' is persisted str and 'extra_select_str'
        # not persisted
        if scope not in newalloc_list:
            resreqstr, extra_select_str = extra_select_str, resreqstr
        LOG.debug("After passing all OpenStack filters, the final "
                  "resource requirement expression string for "
                  "EGO is (%(resreqstr)s), extra select str is "
                  "(%(extra_select_str)s)." %
                    {"resreqstr": resreqstr,
                     "extra_select_str": extra_select_str})
        return resreqstr, extra_select_str

    def schedule_cache_update(self, filter_properties,
                              trouble_shooting_info=None):
        """
        Initialize self.schedule_cache with instance id, scope, num_attempts
        from filter_properties and update the schedule cache with
        trouble_shooting_info
        :param filter_properties: same with nova filters used
        :param trouble_shooting_info: for updating self.schedule_cache
        """
        trouble_shooting_info = trouble_shooting_info or {}
        if not filter_properties:
            return
        spec = filter_properties.get('request_spec', {})
        if not spec:
            return
        scope = spec.get('operation_type', '')
        if not scope:
            return
        instance_id = spec.get('instance_id')
        if not instance_id:
            return
        retry = filter_properties.get('retry', {})
        if not retry:
            num_attempts = 1
        else:
            num_attempts = retry.get('num_attempts')
        if not num_attempts:
            return
        if instance_id not in self.schedule_cache:
            self.schedule_cache[instance_id] = {}
        if scope not in self.schedule_cache[instance_id]:
            self.schedule_cache[instance_id][scope] = {}
        if num_attempts not in self.schedule_cache[instance_id][scope]:
            self.schedule_cache[instance_id][scope][num_attempts] = {}

        # This 'force_hosts' is specified by user and would not change with
        # schedule attempt
        if 'force_hosts' in trouble_shooting_info:
            force_hosts = trouble_shooting_info.pop('force_hosts')
            self.schedule_cache[instance_id][scope].update(
                {'force_hosts': force_hosts})

        if 'source_host' in trouble_shooting_info:
            source_host = trouble_shooting_info.pop('source_host')
            self.schedule_cache[instance_id][scope].update(
                {'source_host': source_host})

        tshoot = self.schedule_cache[instance_id][scope][num_attempts]
        if 'host_error' in trouble_shooting_info:
            pre_tshoot = self.schedule_cache[instance_id][scope].setdefault(
                (num_attempts - 1), {})
            pre_tshoot['error_msg'] = trouble_shooting_info['host_error']
            trouble_shooting_info.pop('host_error')
        if 'resreq' in trouble_shooting_info:
            consume_str = ego_resreq_utils.get_criteria_str_by_resreq(
                              trouble_shooting_info['resreq'], 'consume')
            reserveon_str = ego_resreq_utils.get_criteria_str_by_resreq(
                                trouble_shooting_info['resreq'], 'reserveon')
            trouble_shooting_info['consume'] = consume_str
            trouble_shooting_info['reserveon'] = reserveon_str
        tshoot.update(trouble_shooting_info)
