# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import os
import sys
import time

import six

from nova.ibm.ego import ego_gettextutils
from nova.scheduler.ibm.ego import ego_exception

from oslo_config import cfg
from oslo_log import log as logging

_ = ego_gettextutils._

LOG = logging.getLogger(__name__)

CONF = cfg.CONF
CONF.import_opt('ego_master_list', 'nova.ibm.ego.ego_flags')
CONF.import_opt('ego_kd_port', 'nova.ibm.ego.ego_flags')
CONF.import_opt('ego_lib_path', 'nova.ibm.ego.ego_flags')
CONF.import_opt('ego_user', 'nova.ibm.ego.ego_flags')
CONF.import_opt('ego_password', 'nova.ibm.ego.ego_flags')
CONF.import_opt('ego_enable_ipv6', 'nova.ibm.ego.ego_flags')


ego = None


def conn_check(func):
    def decorator(self, *args, **kwargs):
        if self.conn_lost:
            LOG.error(_('Platform Resource Scheduler is no longer connected '
                        'to EGO. Platform Resource Scheduler will reconnect '
                        'to EGO.'))
            raise ego_exception.EGOExpConnLost()
        return func(self, *args, **kwargs)
    return decorator


class EGOHandler(object):
    """
    EGO C API Python Wrapper
    """
    def __init__(self, client_name=None):
        self.__check_ego_conf()
        self.__set_ego_env()

        self.conn_and_logon(client_name)

        LOG.debug("Platform Resource Scheduler successfully "
                  "initialized the EGO handler.")

    def conn_and_logon(self, client_name):
        try:
            self.__open_ego_connection()
            if client_name:
                self.__logon_and_register(client_name)
            else:
                self.__logon()
            self.conn_lost = False
        except (ego_exception.EGOExpOpenConnFaild,
                ego_exception.EGOExpLogonOrRegFailed,
                ego_exception.EGOExpLogonFailed):
            self.conn_lost = True

    def __check_ego_conf(self):
        self._master_list = CONF.ego_master_list.strip()
        self._kd_port = CONF.ego_kd_port.strip()
        self._lib_path = CONF.ego_lib_path.strip()
        self._user = CONF.ego_user
        self._password = CONF.ego_password

        def _check_exist(key, value):
            if value == '':
                LOG.error(_("The EGO handler cannot be initialized because "
                            "the system cannot locate a value for the %(key)s"
                            " parameter within the nova.conf file. "
                            "Set the %(key)s value and restart the "
                            "EGO scheduler.") % {'key': key})
                sys.exit(-1)

        _check_exist('ego_master_list', self._master_list)
        _check_exist('ego_kd_port', self._kd_port)
        _check_exist('ego_lib_path', self._lib_path)

        if (not self._kd_port.isdigit() or
                int(self._kd_port) <= 1024 or int(self._kd_port) >= 65535):
            LOG.error(_("The value(%s) of 'ego_kd_port' configured in "
                        "nova.conf is bad. EGO handler can't be initialized. "
                        "It should be in range of (1024 - 65535). "
                        "Please configure it correctly before restart.") %
                        self._kd_port)
            sys.exit(-1)

    def __set_ego_env(self):
        os.environ['EGO_CONFDIR'] = self._lib_path
        os.environ['EGO_LIBDIR'] = self._lib_path
        os.environ['EGO_SEC_CONF'] = self._lib_path
        # ego master list, you can configure it
        # as "host1 host2 ..."
        os.environ['EGO_MASTER_LIST'] = self._master_list
        os.environ['EGO_KD_PORT'] = self._kd_port
        # Enable IPv6 if ego_enable_ipv6==True
        if CONF.ego_enable_ipv6:
            os.environ['EGO_ENABLE_SUPPORT_IPV6'] = 'y'

        global ego
        if ego is None:
            sys.path.insert(0, self._lib_path)
            ego = __import__('ego')

    def __open_ego_connection(self):
        self.ego_handler = ego.ego()
        try:
            self.ego_handler.open()
        except ego.error as e:
            LOG.error(_('Platform Resource Scheduler could not connect to EGO '
                        'due to (%s). Start up EGO if EGO is not running, '
                        'otherwise wait one or two minutes to allow EGO '
                        'to start up.') % six.text_type(e))
            raise ego_exception.EGOExpOpenConnFaild(reason=six.text_type(e))

    def __logon(self):
        try:
            self.ego_handler.logon(self._user, self._password)
        except ego.error as e:
            LOG.error(_('Platform Resource Scheduler could not log on '
                        'due to (%s).') % six.text_type(e))
            raise ego_exception.EGOExpLogonFailed(reason=six.text_type(e))

    def __logon_and_register(self, client):
        try:
            self.ego_handler.logonAndRegister(self._user,
                                              self._password, client)
        except ego.error as e:
            LOG.error(_('Platform Resource Scheduler could not log on '
                        'or register due to (%s).') % six.text_type(e))
            raise ego_exception.EGOExpLogonOrRegFailed(reason=six.text_type(e))

    def is_conn_lost(self):
        return self.conn_lost

    def ego_uname(self):
        """
        Get EGO cluster info
        """
        try:
            self.ego_handler.uname()
        except ego.error as e:
            LOG.error(_('Platform Resource Scheduler could not get EGO '
                        'cluster info due to (%s).') % six.text_type(e))
            self.conn_lost = True

    def close_ego_connection(self):
        self.ego_handler.close()

    @conn_check
    def ego_get_all_alloc(self):
        try:
            return self.ego_handler.getAllAlloc()
        except ego.error as e:
            LOG.error(_('Platform Resource Scheduler could not get '
                        'information of all allocations due to (%(reason)s).')
                      % {"reason": six.text_type(e)})
            raise ego_exception.EGOExpGetAllAllocFailed(
                                        reason=six.text_type(e))

    @conn_check
    def ego_set_resource_info_str(self, cstr):
        """
        Set nova-compute metrics as EGO RPI resource

        :param cstr: RPI Resource string
        <plugin_name>, <version>,  <number_of_resources>,
        [resource_name, resource_type, resource_state,
        number_attributes, [attribute, type, value, ...] ...]
        [resource_name, resource_type, resource_state,
        number_attributes, [attribute, type, value, ...] ...]

        The following is a sample output of openstack host
        information for an EGO cluster:
        dynamic 1 2
        devstack19 host 0 2 memSize 2 40960 cpuSize 1 16
        devstack11 host 0 2 memSize 2 10240 cpuSize 1 8

        :returns: None
                  Throw exception if failed
        """
        try:
            self.ego_handler.setResourceInfo(cstr)
        except ego.error as e:
            LOG.error(_('Platform Resource Scheduler could not set RPI '
                        'resources using resource string (%(cstr)s) '
                        'due to (%(reason)s).')
                      % {"cstr": cstr, "reason": six.text_type(e)})
            raise ego_exception.EGOExpSetResourceInfoFailed(cstr=cstr,
                                              reason=six.text_type(e))

    @conn_check
    def ego_set_resource_plan(self, cstr):
        """
        Set EGO resource plan

        :param cstr: global placement policy

        <?xml version="1.0" encoding="UTF-8"?>
        <DistributionPolicies>
            <Scheduler>placetorus</Scheduler>
            <InitialPlacement>
                <PlacementPolicy>order(0-ut)</PlacementPolicy>
            </InitialPlacement>
            <OveruseRatio>
                <vcpu>4</vcpu>
                <memSize>3.5</memSize>
            <OveruseRatio>
        </DistributionPolicies>

        :returns: None
                  Throw exception if failed
        """
        try:
            self.ego_handler.setResourcePlan(cstr)
        except ego.error as e:
            LOG.error(_('Platform Resource Scheduler could not set the EGO '
                        'resource plan using XML string (%(cstr)s) due '
                        'to (%(reason)s).')
                      % {"cstr": cstr, "reason": six.text_type(e)})
            raise ego_exception.EGOExpSetResourcePlanFailed(cstr=cstr,
                                               reason=six.text_type(e))

    @conn_check
    def ego_get_resource_info(self, cstr, is_attr_required=False):
        """
        Get EGO RPI resources info with resource requirement

        :param cstr: EGO resource requirement
        :param is_attr_required: EGO resource's attributes will be included
                                 in the returns if True

        :returns: [{'name': nodename1, 'state': 1, 'attributes': {...}},
                   {'name': nodename2, 'state': 0, 'attributes': {...}}, ...]
                  Throw exception if failed
        """
        try:
            return self.ego_handler.getResourceInfo(cstr, is_attr_required)
        except ego.error as e:
            LOG.error(_('Platform Resource Scheduler could not get RPI '
                        'resources using resource requirement '
                        '(%(cstr)s) due to (%(reason)s).')
                      % {"cstr": cstr, "reason": six.text_type(e)})
            raise ego_exception.EGOExpGetResourceInfoFailed(cstr=cstr,
                                              reason=six.text_type(e))

    @conn_check
    def ego_get_decisions_by_filter(self, filter):
        """
        :param filter: get decisions by filter

        :returns: decision_list, its element is a tuple like below
            decision[0]: resource name
            decision[1]: decision id
            deicsion[2]: subdemand name
            deicsion[3]: slot number
            deicsion[4]: resreq
            deicsion[5]: resource group
        """
        try:
            return self.ego_handler.getDecisionsByFilter(filter)
        except ego.error as e:
            LOG.error(_('Platform Resource Scheduler could not find '
                        'decisions by filter(%(filter)s) due to '
                        '(%(reason)s).') % {'filter': str(filter),
                                            'reason': six.text_type(e)})
            raise ego_exception.EGOExpGetDecisionsByFilterFailed(
                             filter=str(filter), reason=six.text_type(e))

    @conn_check
    def ego_create_alloc(self, name, sub_demands,
                           inter_sub_policies=None):
        """
        Create EGO allocation when create instance

        :param name: allocation name, for openstack Resource
                     Optimization, it is the id of the instance
        :param sub_demands: sub-demands of this allocation
        :param inter_sub_policies: inter sub-demands policies

        :returns: return (allocId, decid, egohost) if succeed
                  Throw exception if failed
        """
        # TODO(gyliu) use "admin" as default EGO consumer
        # Will update this part when tenant map to EGO consumer
        inter_sub_policies = inter_sub_policies or []
        tenant = 'admin'

        try:
            (ret, allocId) = self.ego_handler.createalloc(name, tenant,
                                                          inter_sub_policies,
                                                          sub_demands)
            while 1:
                while 1:
                    rc = self.ego_handler.select(1000)
                    if rc < 0:
                        time.sleep(1)
                        continue
                    break
                allocInfo = self.ego_handler.read()
                retval = allocInfo[0]

                # EGO message includes multiple types:ADD_RESOURCE,
                # RESOURCE_RECLAI, CONTAINER_STATE_CHG, HOST_STATE_CHANGE
                # Ignore non-ADD_RESOURCE message, which is nothing to do
                # with openstack Resource Optimization
                # retval = 0 : get ADD_RESOURCE message
                # retval = 1 : get non-ADD_RESOURCE message

                if retval:
                    continue
                (retval, allocId, consumer, decisionlist, clientname) = (
                                                                     allocInfo)
                # decision_list is a list, it's element is a tuple like below
                # decision[0]: resource name
                # decision[1]: decision id
                # deicsion[2]: subdemand name
                # deicsion[3]: slot number
                # deicsion[4]: resreq
                # deicsion[5]: resource group
                return (allocId, consumer, decisionlist)
        except ego.error as e:
            LOG.error(_('Platform Resource Scheduler could not create an '
                        'EGO resource allocation due to (%s); refer '
                        'to /var/log/nova/nova-scheduler.log for '
                        'troubleshooting report details.') % six.text_type(e))
            raise ego_exception.EGOExpCreateAllocFailed(
                                            reason=six.text_type(e))

    @conn_check
    def ego_update_decision(self, decid, allocid, resreq, extra_resreq,
                            prefer_same_host=0, force_realloc_migrate=False,
                            max_slot=1, retry_resize_flag=False,
                            warning_if_no_resource_group=False):
        """
        Resize EGO allocation when resize instance

        :param decid: decision ID
        :param allocid: allocation ID
        :param resreq: new resource requirement for resize
        :param extra_resreq: select((vcpu==4))
        :param prefer_same_host: prefer to resize on same node
        :param max_slot: max slot for the allocation, 0 means released
        :param force_realloc_migrate: force realloc or migration without
                                      considering resource requirement
                                      and resource quota
        :param retry_resize_flag: whether it is a retry resize option
        :param warning_if_no_resource_group: Set True to log warning if
                                             specified resource group does not
                                             exist

        :returns: return (allocId, decid, egohost) if succeed
                  Throw exception if failed
        """
        opt_name = 're-create' if retry_resize_flag else 'resize'

        try:
            (ret, allocId) = (
                self.ego_handler.updateDecision(allocid, decid, resreq,
                                                extra_resreq,
                                                prefer_same_host,
                                                force_realloc_migrate,
                                                max_slot))
            while 1:
                while 1:
                    rc = self.ego_handler.select(1000)
                    if rc < 0:
                        time.sleep(1)
                        continue
                    break
                allocInfo = self.ego_handler.read()
                retval = allocInfo[0]

                # EGO message includes multiple types:ADD_RESOURCE,
                # RESOURCE_RECLAI, CONTAINER_STATE_CHG, HOST_STATE_CHANGE
                # Ignore non-ADD_RESOURCE message, which is nothing to do
                # with openstack Resource Optimization
                # retval = 0 : get ADD_RESOURCE message
                # retval = 1 : get non-ADD_RESOURCE message

                if retval:
                    continue
                (retval, allocId, consumer, decision, clientname) = allocInfo
                if len(decision) == 0:
                    errmsg = (_('Platform Resource Scheduler cannot find a '
                               'suitable host for instance decision '
                               '(%(decid)s) to %(opt_name)s EGO '
                               'allocation (%(allocId)s) with resource '
                               'requirement (%(resreq)s) and extra '
                               'resource requirement (%(extra_resreq)s); '
                               'refer to /var/log/nova/nova-scheduler.log for '
                               'troubleshooting report details.')
                            % {'decid': decid, 'opt_name': opt_name,
                               'allocId': allocId, 'resreq': resreq,
                               'extra_resreq': extra_resreq})
                    if warning_if_no_resource_group:
                        LOG.warning(errmsg)
                    else:
                        LOG.error(errmsg)
                    raise ego_exception.EGOExpUpdateDecFailed(
                                decid=decid, reason=errmsg)
                (egohost, decid, sub_demand_name, max_slots, resreq1,
                        resource_group) = decision.pop()
                # Update instance host map
                if max_slots == 0:
                    LOG.debug("Platform Resource Scheduler successfully "
                              "released resource for EGO allocation "
                              "with ID (%s)" % (allocId))
                else:
                    LOG.debug("Platform Resource Scheduler "
                              "successfully resized allocation "
                              "ID (%(allocId)s) and decision "
                              "(%(decid)s) on host "
                              "(%(egohost)s)." %
                                {'allocId': allocId, 'decid': decid,
                                 'egohost': egohost})
                return (allocId, decid, egohost, resource_group)
        except ego.error as e:
            LOG.error(_('Platform Resource Scheduler could not %(opt_name)s '
                        'the EGO allocation with resreq (%(resreq)s) and '
                        'extra resreq (%(extra_resreq)s) due to (%(e)s)'
                        'for decision (%(decid)s); refer to '
                        '/var/log/nova/nova-scheduler.log for '
                        'troubleshooting report details.') %
                      {'opt_name': opt_name, 'e': six.text_type(e),
                       'resreq': resreq, 'extra_resreq': extra_resreq,
                       'decid': decid})
            raise ego_exception.EGOExpUpdateDecFailed(decid=decid,
                                                      reason=six.text_type(e))

    @conn_check
    def ego_resize_alloc(self, allocid, subdemands):
        try:
            (ret, allocId) = self.ego_handler.resizealloc(allocid, subdemands)
            while 1:
                while 1:
                    rc = self.ego_handler.select(1000)
                    if rc < 0:
                        time.sleep(1)
                        continue
                    break
                allocInfo = self.ego_handler.read()
                retval = allocInfo[0]

                # EGO message includes multiple types:ADD_RESOURCE,
                # RESOURCE_RECLAI, CONTAINER_STATE_CHG, HOST_STATE_CHANGE
                # Ignore non-ADD_RESOURCE message, which is nothing to do
                # with openstack Resource Optimization
                # retval = 0 : get ADD_RESOURCE message
                # retval = 1 : get non-ADD_RESOURCE message

                if retval:
                    continue
                # decision_list = allocInfo[3]
                # decision_list is a list, it's element is a tuple like below
                # decision[0]: resource name
                # decision[1]: decision id
                # deicsion[2]: subdemand name
                # deicsion[3]: slot number
                # deicsion[4]: resreq
                # deicsion[5]: resource group
                return allocInfo
        except ego.error as e:
            LOG.error(_('Platform Resource Scheduler could not resize an '
                        'EGO resource allocation due to (%s); refer '
                        'to /var/log/nova/nova-scheduler.log for '
                        'troubleshooting report details.') % six.text_type(e))
            raise ego_exception.EGOExpResizeAllocFailed(
                                                  reason=six.text_type(e))

    @conn_check
    def ego_free_alloc(self, allocid):
        try:
            self.ego_handler.freealloc(allocid)
        except ego.error as e:
            LOG.error(_('Platform Resource Scheduler could not release EGO '
                        'resource allocation for allocation id (%(allocid)s) '
                        'due to (%(reason)s).')
                      % {"allocid": allocid, "reason": six.text_type(e)})
            raise ego_exception.EGOExpFreeAllocFailed(reason=six.text_type(e))

    @conn_check
    def ego_migrate_decision(self, decid, allocid,
                             force_flag=False, extra_resreq=''):
        """
        :param decid:    decision ID
        :param allocid:  allocation ID
        :param force_flag: force realloc or migration without
                           considering resource requirement
                           and resource quota
        :param extra_resreq: extra resource reqiurement

        :returns: None or exception if failed
        """

        try:
            self.ego_handler.migratedecision(decid, allocid,
                                             force_flag, extra_resreq)
        except ego.error as e:
            LOG.error(_('Platform Resource Scheduler could not migrate '
                        'decision (%(decid)s) with extra resreq '
                        '(%(extra_resreq)s) due to (%(reason)s); refer '
                        'to /var/log/nova/nova-scheduler.log for '
                        'troubleshooting report details.')
                      % {"decid": decid, "reason": six.text_type(e),
                         "extra_resreq": extra_resreq})
            raise ego_exception.EGOExpMigrateDecisionFailure(
                              decid=decid, reason=six.text_type(e))

    @conn_check
    def ego_get_possible_hosts(self, decid, resreq):
        """
        Get possible hosts for one instance

        :param decid: decision id, for openstack Resource Optimization,
                           it is the decision id of the instance
        :param resreq: reserveon and order string with EGO format

        :returns: return possible host list which have the same
                  resource requirement with current decision, the
                  return value is in the fomant of ['host1', 'host2', ...]
                  Throw exception if failed
        """
        try:
            return self.ego_handler.getPossibleHosts(decid, resreq)
        except ego.error as e:
            LOG.error(_('Platform Resource Scheduler could not find '
                        'suitable hosts for instance decision (%(decid)s) '
                        'due to (%(reason)s).')
                      % {"decid": decid, "reason": six.text_type(e)})
            raise ego_exception.EGOExpGetPossibleHostsFailed(
                                        decid=decid,
                                        reason=six.text_type(e))

    @conn_check
    def ego_create_resource_group(self, group_name, resreq):
        """
        Create EGO dynamic resource group

        :param group_name: EGO dynamic resource group name
        :param resreq: EGO resource requirement for resource group

        :returns: 0 success
                  throw exception if failed
        """
        try:
            self.ego_handler.createDynamicGroup(group_name, resreq)
            LOG.debug('Platform Resource Scheduler successfully created '
                      'EGO resource group %(group)s with resource '
                      'requirement %(resreq)s.' %
                      {"group": group_name, "resreq": resreq})
        except ego.error:
            raise ego_exception.EGOExpCreateResourceGroupFailure(
                                             group=group_name, resreq=resreq)

    @conn_check
    def ego_delete_resource_group(self, group_name):
        """
        Delete EGO dynamic resource group

        :param group_name: EGO dynamic resource group name
        :return 0 success
                throw exception if failed
        """
        try:
            self.ego_handler.deleteDynamicGroup(group_name)
            LOG.debug('Platform Resource Scheduler successfully deleted '
                      'EGO resource group %(group)s.' %
                      {"group": group_name})
        except ego.error:
            raise ego_exception.EGOExpDeleteResourceGroupFailure(
                    group=group_name)

    @conn_check
    def ego_get_all_resource_groups(self):
        """
        Get all EGO resource group details

        :param None

        :returns: resource group list
                  [(resource group name1, [host list for resource group1]),
                   (resource group name2, [host list for resource group2]),
                   ...]
                  [('rg1, ['host11', 'host12']'),
                   ('rg2, ['host21', 'host21']')]
        """
        try:
            return self.ego_handler.getAllResourceGroups()
        except ego.error:
            raise ego_exception.EGOExpGetResourceGroupFailure()

    @conn_check
    def ego_release_decision(self, allocid, decid):
        try:
            self.ego_handler.releaseDecision(allocid, decid)
        except ego.error as e:
            raise ego_exception.EGOExpReleaseDecisionFailed(decid=decid,
                                                reason=six.text_type(e))

    @conn_check
    def ego_get_resource_plan(self, name, scheduler, flag, with_version):
        """
        Get EGO multi dimension resource plan.

        :returns: resource plan string
        """
        try:
            ret, xml_str = self.ego_handler.getResourcePlan(
                name, scheduler, flag, with_version)
            return xml_str
        except ego.error as e:
            LOG.error(_('Platform Resource Scheduler could not get the EGO '
                        'resource plan due to (%(reason)s).')
                      % {"reason": six.text_type(e)})
            raise ego_exception.EGOException(
                reason=six.text_type(e))

    @conn_check
    def ego_set_resource_plan_md(self, name, cstr, flag):
        """
        Set EGO resource plan

        :param cstr: multi dimension resource plan string

        :returns: None
                  Throw exception if failed
        """
        try:
            self.ego_handler.setResourcePlanMD(name, cstr, flag)
        except ego.error as e:
            LOG.error(_('Platform Resource Scheduler could not set the EGO '
                        'resource plan using XML string (%(cstr)s) due '
                        'to (%(reason)s).')
                      % {"cstr": cstr, "reason": six.text_type(e)})
            raise ego_exception.EGOExpSetResourcePlanFailed(
                cstr=cstr, reason=six.text_type(e))

    @conn_check
    def ego_create_consumer(self, consumer_name, exec_user, parent_path,
                            rebalance, reclaim_time, users, plans):
        """
        Create a new consumer in ego.

        :returns: None
                  Throw exception if failed
        """
        try:
            self.ego_handler.addConsumer(
                consumer_name, exec_user, parent_path, rebalance,
                reclaim_time, users, plans)
        except ego.error as e:
            LOG.error(_('Platform Resource Scheduler could not add the '
                        'consumer (%(name)s) due to (%(reason)s).')
                      % {"name": consumer_name, "reason": six.text_type(e)})
            raise ego_exception.EGOExpCreateConsumerFailed(
                name=consumer_name, reason=six.text_type(e))

    @conn_check
    def ego_add_remove_consumer_for_md(self, consumer_path, consumer_name,
                                       md_plan_name, operation):
        """
        Add/Remove consumer section in MD resource plan.

        :returns: None
                  Throw exception if failed
        """
        try:
            self.ego_handler.addRemoveConsumerForMD(
                consumer_path, consumer_name, md_plan_name, operation)
        except ego.error as e:
            LOG.error(_('Platform Resource Scheduler could not modify the '
                        'consumer (%(name)s) in MD resource plan due '
                        'to (%(reason)s). Operation is (%(operation)s)')
                      % {"name": consumer_name,
                         "reason": six.text_type(e),
                         "operation": operation})
            raise ego_exception.EGOExpAddRemoveConsumerFailed(
                name=consumer_name, reason=six.text_type(e))

    @conn_check
    def ego_set_consumer(self, consumer_path, consumer_name,
                         rebalance, reclaim_timeout):
        """
        Set the rebalance or reclaim timeout
        for the specified consumer in ego.

        :returns: None
                  Throw exception if failed
        """
        try:
            self.ego_handler.setConsumer(
                consumer_path, consumer_name,
                rebalance, reclaim_timeout)
        except ego.error as e:
            LOG.error(_('Platform Resource Scheduler could not set the '
                        'consumer (%(name)s) due to (%(reason)s).')
                      % {"name": consumer_path, "reason": six.text_type(e)})
            raise ego_exception.EGOExpSetConsumerFailed(
                name=consumer_name, reason=six.text_type(e))

    @conn_check
    def ego_get_consumer_tree(self, role):
        """
        Get the consumer tree without plans.

        :returns: xml str for consumer tree.
        """
        try:
            return self.ego_handler.getConsumerTree(role)
        except ego.error as e:
            LOG.error(_('Platform Resource Scheduler could not get the '
                        'consumer tree due to (%(reason)s).')
                      % {"reason": six.text_type(e)})
            raise ego_exception.EGOExpGetConsumerTreeFailed(
                reason=six.text_type(e))

    @conn_check
    def ego_delete_consumer(self, consumer_path):
        """
        Delete the specified consumer in ego.

        :returns: None
                  Throw exception if failed
        """
        try:
            self.ego_handler.delConsumer(consumer_path)
        except ego.error as e:
            LOG.error(_('Platform Resource Scheduler could not delete the '
                        'consumer (%(name)s) due to (%(reason)s).')
                      % {"name": consumer_path, "reason": six.text_type(e)})
            raise ego_exception.EGOExpDeleteConsumerFailed(
                name=consumer_path, reason=six.text_type(e))

    @conn_check
    def get_service_instance_info(self, resource_plan):
        try:
            return self.ego_handler.getInstanceInfo(
                self._user, self._password, resource_plan)
        except ego.error as e:
            LOG.error(_('Platform Resource Scheduler could not get the '
                        'service instance info due to (%(reason)s).')
                      % {"reason": six.text_type(e)})
            raise ego_exception.EGOExpGetServiceInstanceInfoFailed(
                reason=six.text_type(e))
