# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""
Construct EGO RPI host.
All info of nova host can be mapped to EGO RPI host attribute.
"""

from nova import availability_zones
from nova import context as nova_context
from nova import exception
from nova.ibm.ego import ego_constants
from nova.ibm.ego import ego_gettextutils
from nova.ibm.ego.hapolicy import constants as ha
from nova import objects
from nova.scheduler.filters import trusted_filter
from nova.scheduler.ibm.ego import ego_resreq_utils
from nova import servicegroup

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
import six

_ = ego_gettextutils._
_LW = ego_gettextutils._LW

CONF = cfg.CONF
CONF.register_opts([
    cfg.IntOpt('prs_ut_data_sample_num', default=10,
               help='The number of cached metric values that will be used to '
               'compute the average value for the respective metric'),
])
CONF.import_opt('disk_allocation_ratio', 'nova.scheduler.filters.disk_filter')
CONF.import_opt('scheduler_default_filters', 'nova.scheduler.host_manager')
CONF.import_opt('reserved_host_memory_mb', 'nova.compute.resource_tracker')
CONF.import_opt('reserved_host_disk_mb', 'nova.compute.resource_tracker')

LOG = logging.getLogger(__name__)


class EGOHostAttrString(object):
    """wrapper for the attribute strings. protect it from
    writing manaully.
    """
    def __init__(self):
        self.__attrnum = 0
        self.__attrstr = ''
        self.__attr_keys = []

    def get_attr_number(self):
        return self.__attrnum

    def get_attr_string(self):
        return self.__attrstr

    def get_attr_keys(self):
        return self.__attr_keys

    def append_res_str(self, key, val, is_expanded=True, is_dynamic=False):
        """append resource attribute string as 'key type value'
        This should be the first choice if a new attribute need to
        be added.

        :param key: resource attribute key
        :param val: resource attribute value
        :param is_expanded: if true, expand dict value or
        dict-like six.string_types value

        :return: resource attribute string
        """
        res_str = ''
        if key not in self.__attr_keys:

            # To fixed the cpu info message not right issue:
            if isinstance(val, six.string_types):
                val = val.replace('null', 'None')

            attr_type, attr_str = ego_resreq_utils.generate_attr_str(val)
            if not attr_type or not attr_str:
                attr_type = '2'
                attr_str = '""'
            if isinstance(val, (six.string_types, dict)):
                if isinstance(val, dict) and is_expanded:
                    for k in sorted(val):
                        v = val[k]
                        self.append_res_str(k, v, is_expanded)
                if isinstance(val, six.string_types) and is_expanded:
                    try:
                        sub_dict = dict(eval(val))
                        for k in sorted(sub_dict):
                            v = sub_dict[k]
                            self.append_res_str(k, v, is_expanded)
                    except Exception:
                        # any exception in eval() and dict() will be ignored
                        # such as ValueError, NameError
                        pass

            if is_dynamic:
                attr_type = '3'

            res_str = ' ' + str(key) + ' ' + attr_type + ' ' + attr_str
            self.__attrnum = self.__attrnum + 1
            self.__attrstr = self.__attrstr + res_str
            self.__attr_keys.append(key)
        return res_str


class EGOHostAttrs(object):
    """Construct EGO RPI resource attribute list."""
    def __init__(self,
                 ego_host,
                 compute=None,
                 metadatas=None,
                 flavors=None,
                 node_stats=None,
                 host_aggregate_ids=None,
                 service=None,
                 context=None):
        compute = compute or {}
        metadatas = metadatas or {}
        flavors = flavors or []
        node_stats = node_stats or {}
        host_aggregate_ids = host_aggregate_ids or []
        service = service or {}
        self.ego_host = ego_host
        self.compute = compute
        self.metadatas = metadatas
        self.flavors = flavors
        self.node_stats = node_stats
        self.host_aggregate_ids = host_aggregate_ids
        self.service = service
        self.context = context
        self.attr_str = EGOHostAttrString()
        self.servicegroup_api = servicegroup.API()
        self.resource_state = ego_constants.HOST_RES_OK

    def __get_powervm_fixed(self):
        # Get content of self.node_stats which
        # is related with the current compute

        # Track number of process units
        proc_units = float(self.node_stats.get('proc_units', 0))
        # Track number of unreturned cores
        proc_units_unreturned = float(self.node_stats.get(
            'cod_unreturned_mobile_cores', 0))
        # Track number of reserved process units
        proc_units_reserved = (
            float(self.node_stats.get('proc_units_reserved', 0)))
        # Track number of reserved memory
        host_memory_reserved = (
            float(self.node_stats.get('host_memory_reserved', 0)))

        return (proc_units, proc_units_reserved, proc_units_unreturned,
                host_memory_reserved)

    def get_ego_fixed(self, cpu_over_commit_ratio, mem_over_commit_ratio):
        """Add RPI attributes of EGO fixed info."""
        # Check for process unit of Power machine
        # Here we suppose over commit ratio always be 1 for Demo purpose

        egofixed_str = self.attr_str.append_res_str('vcpuRequest', 0.00)
        egofixed_str += self.attr_str.append_res_str('memSizeRequest', 0.00)
        egofixed_str += self.attr_str.append_res_str(
            'realvcpu', self.compute['vcpus'])
        egofixed_str += self.attr_str.append_res_str(
            'realmemSize', self.compute['memory_mb'])

        reserved_host_memory_mb = CONF.reserved_host_memory_mb
        htype = str(self.compute['hypervisor_type'])
        if ((htype in ego_constants.POWERVM_HYPERVISOR_TYPES) and
            (not ego_resreq_utils.is_powervc_driver_host(htype,
                                                         self.node_stats))):
            # Only for PowerVM case: vcpu and memSize came from node_stats
            if ('proc_units' not in self.node_stats or
                'proc_units_reserved' not in self.node_stats or
                    'host_memory_reserved' not in self.node_stats):
                LOG.warning(_LW("Platform Resource Scheduler cannot currently "
                                "schedule VMs to hypervisor %(ego_host)s "
                                "because it is missing required host stats; "
                                "please wait for a few more minutes.")
                            % {'ego_host': self.ego_host})
                return
            (proc_units,
             proc_units_reserved,
             proc_units_unreturned,
             memory_mb_reserved) = self.__get_powervm_fixed()
            vcpus_val = float(proc_units * cpu_over_commit_ratio -
                (proc_units_reserved + proc_units_unreturned))
            mem_val = (self.compute['memory_mb'] * mem_over_commit_ratio -
                memory_mb_reserved - reserved_host_memory_mb)
        else:
            vcpus_val = self.compute['vcpus'] * cpu_over_commit_ratio
            mem_val = (self.compute['memory_mb'] * mem_over_commit_ratio -
                reserved_host_memory_mb)

        cpu_mem_str = self.attr_str.append_res_str('vcpu', vcpus_val)
        cpu_mem_str += self.attr_str.append_res_str('memSize', mem_val)

        LOG.debug("After calculating the ratio, the resource "
                  "metrics expression for EGO is %(egofixed_str)s "
                  "%(cpu_mem_str)s." %
                    {'egofixed_str': egofixed_str,
                     'cpu_mem_str': cpu_mem_str})

    def __get_attrs_keys(self):
        return self.attr_str.get_attr_keys()

    def get_supported_instances(self):
        # Convert [['x86_64', 'kvm', 'hvm'], ['x86', 'power', 'hvm']]
        # to "x86_64,kvm,hvm;x86,power,hvm"
        # 'supported_instances' is used for ImagePropertiesFilter
        if not self.compute.get('supported_hv_specs'):
            return
        supported_inst_list = [hvs.to_list() for hvs in
                                   self.compute['supported_hv_specs']]

        supported_inst_str = self.attr_str.append_res_str(
            'supported_instances',
            supported_inst_list)

        LOG.debug("The supported instance resource metrics "
                  "expression string for EGO is (%s)." %
                  supported_inst_str)

    def get_overcommit(self, cpu_over_commit_ratio, mem_over_commit_ratio):
        """Add RPI attributes for over commit from db."""

        over_commit_str = self.attr_str.append_res_str(
            'memory_mb',
            self.compute['memory_mb'] * mem_over_commit_ratio)

        over_commit_str += self.attr_str.append_res_str(
            'vcpus',
            self.compute['vcpus'] * cpu_over_commit_ratio)
        LOG.debug("The over commit resource metrics "
                  "expression string for EGO is (%s)."
                  % over_commit_str)

    def __get_host_average_metric_value(self, hostname, metric, value,
                                        host_metrics):
        """
        Get metric cache list and update the metrics list. The cache only holds
        as many data points as the nova.conf option allows and returns the
        computed average value.
        """
        if not value:
            value = 0.0
        metrics_list = host_metrics.get(metric)
        if not metrics_list:
            metrics_list = [value]
        else:
            if len(metrics_list) >= CONF.prs_ut_data_sample_num:
                metrics_list.pop(0)
            metrics_list.append(value)
        host_metrics.update({metric: metrics_list})

        sample_size = len(metrics_list)
        avalue = sum(metrics_list) / float(sample_size)

        LOG.debug('The average %(metric)s value of %(host)s is %(avalue)s '
                  '(computed from %(sample_size)d data points).'
                  % {'metric': metric, 'host': hostname, 'avalue': avalue,
                     'sample_size': sample_size})
        return avalue

    def get_compute_node_metrics(self, metrics, host_metrics):
        """
        Add useful metrics in compute node to rpi resource
        Compute node metrics format as follow:
        [{'timestamp': '2013-09-17T03:00:16.823606',
        'name': 'cpu.user.percent',
        'value': 0.022713533120303997,
        'source': 'libvirt.LibvirtDriver'},
        {'timestamp': '2013-09-17T03:00:16.823606',
        'name': 'mem.percent',
        'value': 4614680000000,
        'source': 'libvirt.LibvirtDriver'}]

        :param metrics: The metrics of compute node get from db
        :param hostname: The host name of the metrics
        :param host_metrics: The host metrics cache

        :return: Add the compute node's metrics to rpi
        """
        if not metrics:
            return

        for key in sorted(metrics):
            val = metrics.get(key)
            # cpu utilization need to calculate the average value.
            if (key == ego_constants.CPU_UTILIZATION_METRICS_NAME or
                  key == ego_constants.MEM_UTILIZATION_METRICS_NAME):
                if val is not None:
                    cvalue = self.__get_host_average_metric_value(
                                 self.ego_host, key, val, host_metrics)
                    self.attr_str.append_res_str(key, cvalue, False)
            else:
                val = "null" if val is None else val
                self.attr_str.append_res_str(key, val, True)

    def get_compute_node(self, host_metrics):
        """
        Add RPI attributes from compute node info.

        :param host_metrics: The host metrics cache.
        """
        compute_str = ''
        for key in sorted(self.compute):
            # ignore supported_hv_specs, the format here is not correct.
            # get_supported_instances() will deal this
            # get_compute_node_stats will handle stats
            is_expanded = True
            val = self.compute[key]
            if val is None:
                # skip this host metric if the value is None, update it
                # to EGO RPI until it has non-null value by periodic task
                continue
            if key in ['supported_hv_specs', 'stats']:
                continue
            elif (key == 'hypervisor_type' and
                  val == ego_constants.HYPERVISOR_TYPE_VMWARE_DB):
                # the attribute value which contains blank space can not
                # be used to selectQ in resreq, select('a a'==key) is invalid
                val = ego_constants.HYPERVISOR_TYPE_VMWARE_RPI
            elif (key == 'hypervisor_type' and
                  val == ego_constants.HYPERVISOR_TYPE_ESX_DB):
                # For ESXi host case
                val = ego_constants.HYPERVISOR_TYPE_ESX_RPI
            elif key == 'metrics':
                self.get_compute_node_metrics(val, host_metrics)
                continue
            elif key == 'numa_topology':
                # as for numa_topology, its val is like
                # {'nova_object.name': ...}, if PRS expands it, when rpc call
                # 'ego_get_resource_list' will fail for taking a value
                # 'nova_object.name' as a object when deserializing. Since PRS
                # does not use numa_topology, we need not expand it now.
                is_expanded = False
            compute_str += self.attr_str.append_res_str(key, val, is_expanded)
        LOG.debug("The compute resource metrics expression "
                  "string for EGO is (%s)."
                   % compute_str)

    def get_service(self):
        """Add RPI attributes from service info."""
        dict_date = dict(self.service)
        service_str = ''
        for key in sorted(dict_date):
            val = dict_date[key]
            if key in ['compute_node']:
                continue
            if key in ['created_at', 'updated_at']:
                val = str(val)
            service_str += self.attr_str.append_res_str(key, val)
        LOG.debug("The service resource metrics expression "
                  "string for EGO is (%s)." %
                  service_str)

    def get_host_aggregate_ids(self):
        """
        Add RPI attributes from aggregate id info
        If there is no aggregate id is found, the
        default group will be set
        """
        ha_ids = ego_constants.RESOURCE_GROUP_DEFAULT
        if self.host_aggregate_ids:
            ha_ids = self.host_aggregate_ids

        ha_attr_str = self.attr_str.append_res_str(
            ego_constants.RPI_KEY_HOST_AGGREGATE_ID,
            ha_ids)
        LOG.debug("The host aggregate ids resource metrics expression "
                  "string for EGO is (%s)." % ha_attr_str)

    def get_compute_node_stats(self):
        """Add PRI attributes from compute node stats info."""
        node_stats_optional_str = ''
        for (k, v) in self.node_stats.items():
            stats_val = ego_resreq_utils.convert_str(str(v))
            node_stats_optional_str += (
                self.attr_str.append_res_str(str(k), stats_val, True))
        # add default host states
        node_stats_optional_str += self.attr_str.append_res_str(
            'num_instances',
            int(self.node_stats.get('num_instances', 0)))

        node_stats_optional_str += self.attr_str.append_res_str(
            'io_workload',
            int(self.node_stats.get('io_workload', 0)))
        node_stats_optional_str += self.attr_str.append_res_str(
            'num_vm_error',
            int(self.node_stats.get('num_vm_error', 0)))

        LOG.debug("The host stats metrics expression "
                  "for EGO is (%s)." % node_stats_optional_str)

    def get_availability_zone(self, context, svc_host):
        """Add RPI Resource attribute from availability zone info."""
        if 'availability_zone' in self.attr_str.get_attr_keys():
            return
        availability_zone = availability_zones.get_host_availability_zone(
                                            context, svc_host)
        self.attr_str.append_res_str('availability_zone', availability_zone,
                                     is_dynamic=False)

    def get_compute(self):
        """Add RPI Resource attribute from compute info."""
        self.attr_str.append_res_str('activeCompute', self.is_active_compute())

    def is_active_compute(self):
        # active_compute is for ComputeFilter, if active_compute is 1, it
        # indicates that the host is ready for scheduling; if active_compute
        # is 0, it indicates that the host is not ready for scheduling.
        active_compute = 1
        htype = str(self.compute['hypervisor_type'])
        ha_status = self.node_stats.get('ha_status')
        if not self.service:
            active_compute = 0
            self.resource_state = ego_constants.HOST_RES_UNAVAIL
        elif not self.servicegroup_api.service_is_up(self.service):
            active_compute = 0
            self.resource_state = ego_constants.HOST_RES_UNAVAIL
        elif self.service['disabled']:
            active_compute = 0
        elif htype in ego_constants.POWERVM_HYPERVISOR_TYPES:
            if not ego_resreq_utils.is_powervc_driver_host(htype,
                                                           self.node_stats):
                if 'operating' != self.node_stats.get('hypervisor_state'):
                    active_compute = 0

        # Update RPISTAT according to ha_status, if a host is in ha_status,
        # change resource_state to UNAVAIL
        if ha_status and ha_status != ha.STATE_OK:
            self.resource_state = ego_constants.HOST_RES_UNAVAIL
        return active_compute

    def get_powervc_driver_host(self):
        powervc_driver_host = 0
        htype = str(self.compute['hypervisor_type'])
        if ego_resreq_utils.is_powervc_driver_host(htype, self.node_stats):
            powervc_driver_host = 1

        self.attr_str.append_res_str(
            'powervc_driver_host', powervc_driver_host)

    def get_aggregate_metadata(self):
        """
        Add RPI attributes from aggregate metadata info
        """
        find_instance_type = False
        for key, vals in six.iteritems(self.metadatas):
            if key == 'instance_type':
                find_instance_type = True
            if key == 'availability_zone':
                # since prs-generate-availability-zone-topology treats
                # this attribute as string type, so does here
                self.attr_str.append_res_str(key, vals, is_dynamic=False)
            else:
                self.attr_str.append_res_str(key, vals, is_dynamic=True)

        # set instance_type=dynamic as default if host aggregation
        # do not have metadata as instance_type
        if not find_instance_type:
            self.attr_str.append_res_str(
                'instance_type', 'dynamic', is_dynamic=True)

    def get_disk(self, disk_allocation_ratio):
        """
        Add RPI attributes of disk usage
        """
        reserved_host_disk_mb = CONF.reserved_host_disk_mb
        total_usable_disk_mb = self.compute['local_gb'] * 1024
        disk_mb_limit = (total_usable_disk_mb * disk_allocation_ratio -
            reserved_host_disk_mb)
        free_disk_gb = self.compute['free_disk_gb'] or 0
        least_disk_gb = self.compute['disk_available_least']
        if least_disk_gb is not None:
            if least_disk_gb > free_disk_gb:
                LOG.warning(_LW("Host has more disk space than database "
                           "expected (%(physical)sgb > %(database)sgb)") %
                         {'physical': least_disk_gb,
                          'database': free_disk_gb})
            free_disk_gb = min(least_disk_gb, free_disk_gb)
        free_disk_mb = free_disk_gb * 1024
        used_disk_mb = total_usable_disk_mb - free_disk_mb
        usable_disk_mb = disk_mb_limit - used_disk_mb

        self.attr_str.append_res_str('usable_disk_mb', usable_disk_mb)

    def get_type_affinity(self):
        """
        Add RPI attributes of instance type running on the host
        """

        if len(self.flavors) == 0:
            self.attr_str.append_res_str('run_instance_type', 'dynamic')
        else:
            self.attr_str.append_res_str('run_instance_type',
                                         list(set(self.flavors)))

    def get_trusted(self):
        """
        Add RPI attributes of host trusted
        """

        attestation_service = trusted_filter.AttestationService()
        states = attestation_service.do_attestation([self.service['host']])
        trusted = 0

        if states:
            for state in states:
                if (state['host_name'] == self.service['host'] and
                        state['trust_lvl'] == 'trusted'):
                    trusted = 1
                    break

        self.attr_str.append_res_str('trust_lvl', trusted)

    def get_rpi_resource_str(self, resource_state=ego_constants.HOST_RES_OK):
        return ("dynamic 1 1 %s host %s %s %s" % (
            str(self.ego_host),
            str(resource_state),
            str(self.attr_str.get_attr_number()),
            self.attr_str.get_attr_string()))


class EGOHost(object):
    """Construct EGO RPI resource which is mapped to nova host."""

    def __init__(self):
        self.context = nova_context.get_admin_context()
        self.filters = CONF.scheduler_default_filters

    def _get_aggregate_ratio(self, ego_host, metadatas, ratio_type):
        """
        get aggregate ratio by ego_host and ratio_type

        :param ego_host: ego host name
        :param metadatas: host meta data
        :param ratio_type: cpu_allocation_ratio or mem_allocation_ratio
        :return aggregate ratio of ratio_type, or 0.0 if not exist or invalid
        """
        aggregate_vals = metadatas.get(ratio_type, set())
        num_values = len(aggregate_vals)

        if num_values == 0:
            LOG.warning(_LW("No %(ratio_type)s values found in aggregate "
                          "metadata for host %(host)s.") %
                        {'host': ego_host, 'ratio_type': ratio_type})
            return 0.0
        elif num_values > 1:
            LOG.warning(_LW("%(num_values)d %(ratio_type)s ratio values found "
                          "in aggregate metadata for host %(host)s, "
                          "of which the minimum value will be used.") %
                        {'num_values': num_values,
                         'host': ego_host,
                         'ratio_type': ratio_type})

        try:
            ratio = float(min(aggregate_vals))
        except ValueError as e:
            LOG.warning(_LW("aggregate value (%(aggregate_val)s) could not be "
                          "decoded with ratio type (%(ratio_type)s): "
                          "'%(e)s'") %
                        {'aggregate_val': min(aggregate_vals),
                         'ratio_type': ratio_type,
                         'e': six.text_type(e)})
            ratio = 0.0

        return ratio

    def get_ego_host(self, ego_manager, ego_host, host_metrics=None,
                     compute=None, metadatas=None, flavors=None,
                     node_stats='{}', host_aggregate_ids=None,
                     extra_stats=None, service=None):
        """
        Construct EGO RPI resource string based on filters.

        :param host: The host name.
        :param host_metrics: The host's metrics cache.
        :param compute: The compute node info.
        :param metadatas: Metadata info.
        :param flavors: Flavor info.
        :param node_stat: Node state info.
        :param host_aggregate_ids: Host aggregate id cache.
        :param extra_stats: Extra state info.
        :prama service: Compute service reference.

        :return: A ego host info.
        """
        host_metrics = host_metrics or {}
        compute = compute or {}
        metadatas = metadatas or {}
        flavors = flavors or {}
        host_aggregate_ids = host_aggregate_ids or []
        extra_stats = extra_stats or {}
        service = service or {}
        if (not compute) or (not service):
            LOG.debug('Compute host %s could not be found' % ego_host)
            cstr = ego_resreq_utils.generate_res_state_and_attr_str(ego_host,
                ego_constants.HOST_RES_REMOVED)
            return {'cstr': cstr}
        # node_stats is CoercedDict type which coerces its value as str
        # type, so we can convert it to dict to avoid the constraint
        node_stats = dict(node_stats)
        node_stats.update(extra_stats)
        metrics = compute.get('metrics')
        if metrics:
            try:
                if isinstance(metrics, basestring):
                    metrics = jsonutils.loads(metrics)
                    compute['metrics'] = ego_resreq_utils.list2map(metrics,
                                                                   'name')
            except Exception:
                LOG.warning(_LW("Bad json format metrics: (%s)") % metrics)
        ego_manager._iterate_extension_managers('handle_for_compute_info',
                                                compute)
        host_attrs = EGOHostAttrs(ego_host, compute,
                                  metadatas, flavors, node_stats,
                                  host_aggregate_ids, service,
                                  self.context)
        # get cpu/mem allocation ratio,
        # from (default value/conf/aggregate metadata) by priority
        cpu_allocation_ratio = ego_constants.DEFAULT_CPU_OVER_COMMIT_RATIO
        ram_allocation_ratio = ego_constants.DEFAULT_MEM_OVER_COMMIT_RATIO
        if 'CoreFilter' in self.filters:
            cpu_allocation_ratio = compute[ego_constants.CPU_ALLOCATION_RATIO]
        if 'RamFilter' in self.filters:
            ram_allocation_ratio = compute[ego_constants.RAM_ALLOCATION_RATIO]
        if 'AggregateCoreFilter' in self.filters:
            aggregate_cpu_ratio = self._get_aggregate_ratio(ego_host,
                        metadatas, ego_constants.CPU_ALLOCATION_RATIO)
            if aggregate_cpu_ratio > 0.0:
                # if CoreFilter and AggregateCoreFilter are both enabled
                # cpu_allocation_ratio should be the minimum of them
                if 'CoreFilter' in self.filters:
                    cpu_allocation_ratio = min(aggregate_cpu_ratio,
                                compute[ego_constants.CPU_ALLOCATION_RATIO])
                else:
                    cpu_allocation_ratio = aggregate_cpu_ratio
        if 'AggregateRamFilter' in self.filters:
            aggregate_ram_ratio = self._get_aggregate_ratio(ego_host,
                        metadatas, ego_constants.RAM_ALLOCATION_RATIO)
            if aggregate_ram_ratio > 0.0:
                # if RamFilter and AggregateRamFilter are both enabled
                # ram_allocation_ratio should be the minimum of them
                if 'RamFilter' in self.filters:
                    ram_allocation_ratio = min(aggregate_ram_ratio,
                            compute[ego_constants.RAM_ALLOCATION_RATIO])
                else:
                    ram_allocation_ratio = aggregate_ram_ratio
        host_attrs.get_ego_fixed(cpu_allocation_ratio, ram_allocation_ratio)
        host_attrs.get_overcommit(cpu_allocation_ratio, ram_allocation_ratio)

        host_attrs.get_compute()
        host_attrs.get_compute_node_stats()
        host_attrs.get_compute_node(host_metrics)
        host_attrs.get_supported_instances()
        host_attrs.get_service()
        host_attrs.get_powervc_driver_host()

        # get DiskFilter/AggregateDiskFilter
        disk_allocation_ratio = ego_constants.DEFAULT_DISK_OVER_COMMIT_RATIO
        if 'DiskFilter' in self.filters:
            disk_allocation_ratio = CONF.disk_allocation_ratio
        if 'AggregateDiskFilter' in self.filters:
            aggregate_disk_ratio = self._get_aggregate_ratio(ego_host,
                        metadatas, ego_constants.DISK_ALLOCATION_RATIO)
            if aggregate_disk_ratio > 0.0:
                if 'DiskFilter' in self.filters:
                    disk_allocation_ratio = min(aggregate_disk_ratio,
                                                CONF.disk_allocation_ratio)
                else:
                    disk_allocation_ratio = aggregate_disk_ratio
        if disk_allocation_ratio > 0.0:
            host_attrs.get_disk(disk_allocation_ratio)

        for filter_name in self.filters:
            if filter_name == 'TypeAffinityFilter':
                host_attrs.get_type_affinity()
            elif filter_name == 'TrustedFilter':
                host_attrs.get_trusted()
        host_attrs.get_aggregate_metadata()
        host_attrs.get_availability_zone(self.context, service.get('host'))
        host_attrs.get_host_aggregate_ids()
        cstr = host_attrs.get_rpi_resource_str(resource_state=(
            host_attrs.resource_state))
        LOG.debug("The resource metrics expression string "
                  "for EGO is (%s)." % cstr)
        host_attrs = {}
        host_attrs['compute'] = compute
        host_attrs['metadatas'] = metadatas
        host_attrs['statmap'] = node_stats
        host_attrs['service'] = service
        host_attrs['metrics'] = host_metrics
        host_attrs['cstr'] = cstr
        return host_attrs

    def get_all_ego_host(self, ego_manager, all_flavors, host_aggregates,
                         host_metadatas, ego_hosts_list=None, extra_stats=None,
                         all_host_attrs=None, is_ha_status_valid=False,
                         hv_status=None):
        """
        get all host attributes including EGO RPI cstrs according to DB

        :param all_flavors: map from ego host name to flavor id list
        :param host_aggregates: map from host name to aggregate id list
        :param host_metadatas: map from host name to aggregate metadata
        :param ego_hosts_list: host name list in EGO
        :param extra_stats: map from host name to extra compute node status
        :param all_host_attrs: map from host name to attributes in cache
        :param is_ha_status_valid: True if ha service is up and enabled
                                   False if ha service is down or disabled
        :param hv_status: HA and maintenance status of hypervisors
        """
        ego_hosts_list = ego_hosts_list or []
        extra_stats = extra_stats or {}
        all_host_attrs = all_host_attrs or {}
        host_attr_dict = {}
        ego_hosts_list_db = []
        compute_nodes = objects.ComputeNodeList.get_all(self.context)

        for compute in compute_nodes:
            try:
                service = objects.Service.get_by_compute_host(
                    self.context, compute.host)
            except exception.ComputeHostNotFound:
                LOG.warning(_LW("No compute service record found for host "
                                "%(host)s") % {'host': compute.host})
                continue
            svc_host = compute.host
            host_aggregate_ids = host_aggregates.get(svc_host, [])
            metadatas = host_metadatas.get(svc_host, {})

            node_name = compute.hypervisor_hostname
            ego_host = ego_resreq_utils.get_rpi_res_name(node_name)
            if not extra_stats.get(ego_host):
                extra_stats[ego_host] = {}
            extra_stats[ego_host]['maintenance_status'] = (
                hv_status.get_maintenance(node_name))
            if is_ha_status_valid:
                extra_stats[ego_host]['ha_status'] = (
                        hv_status.get_ha(node_name))
            ego_hosts_list_db.append(ego_host)
            flavors = all_flavors.get(ego_host, {})
            node_stats = compute.stats or {}

            host_metrics = all_host_attrs.get(ego_host, {}).get('metrics', {})
            host_attr_dict[ego_host] = self.get_ego_host(
                   ego_manager,
                   ego_host, host_metrics,
                   compute=dict(six.iteritems(compute)),
                   metadatas=metadatas,
                   flavors=flavors,
                   node_stats=node_stats,
                   host_aggregate_ids=host_aggregate_ids,
                   extra_stats=extra_stats[ego_host],
                   service=dict(six.iteritems(service)))

        # remove the host from EGO resource list if it does not exist in DB
        for ego_host in ego_hosts_list:
            if ego_host not in ego_hosts_list_db:
                cstr = ego_resreq_utils.generate_res_state_and_attr_str(
                                            ego_host,
                                            ego_constants.HOST_RES_REMOVED)
                host_attr_dict[ego_host] = {'cstr': cstr}
        return host_attr_dict
