# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""
Construct an EGO Resource Group.  All info of a Nova host aggregate can be
mapped to EGO resource group.
"""

import collections
import six

from nova import context as nova_context
from nova import db
from nova import exception
from nova.ibm.ego import ego_constants
from nova.ibm.ego import ego_gettextutils
from nova import objects

from oslo_log import log as logging

_ = ego_gettextutils._
_LI = ego_gettextutils._LI
_LW = ego_gettextutils._LW

LOG = logging.getLogger(__name__)


class EGOHostAggregate(object):
    """
    Construct host aggregate which is mapped to EGO resource group.
    """

    initial_prefix_len = (len(ego_constants.INITIAL_POLICY) +
                          len(ego_constants.POLICY_DELIMITER))
    runtime_prefix_len = (len(ego_constants.RUNTIME_POLICY) +
                          len(ego_constants.POLICY_DELIMITER))

    def __init__(self, policy_actions=ego_constants.POLICY_ACTION_LIST):
        self.context = nova_context.get_admin_context()
        self.aggregate_id_policy_dict = {}
        self.aggid_policy_no_check = {}
        self.host_metadatas = {}
        self.host_aggregates = {}
        self.aggregateid_metadatas = {}
        self.policy_actions = policy_actions

    def __filter_metadata_for_policy_show(self, metadata, agg_id):
        """
        Filter metadata of aggregate with policy-related attributes.
        Do not check the validity of parameters.
        e.g., {'initialpolicy':{'id':'9527'},
               'runtimepolicy':{'id':'jd', 'run_interval':'ncmm',
               'max_parallel':'-3', 'threshold':'3x', 'action':'migrate_vm'}}
        """

        policy = {}
        # initial policy
        initial_policy = {}
        if ego_constants.INITIAL_POLICY_ID in metadata:
            key_id = ego_constants.INITIAL_POLICY_ID[self.initial_prefix_len:]
            text_value = metadata[ego_constants.INITIAL_POLICY_ID]
            initial_policy[key_id] = text_value
            policy[ego_constants.INITIAL_POLICY] = initial_policy

        # runtime policy
        runtime_policy = {}
        runtime_policy_attribute = [
            ego_constants.RUNTIME_POLICY_ID,
            ego_constants.RUNTIME_POLICY_RUN_INTERVAL,
            ego_constants.RUNTIME_POLICY_STABILIZATION,
            ego_constants.RUNTIME_POLICY_MAX_PARALLEL,
            ego_constants.RUNTIME_POLICY_THRESHOLD,
            ego_constants.RUNTIME_POLICY_ACTION
        ]

        for attribute in runtime_policy_attribute:
            if attribute in metadata:
                text_value = metadata[attribute]
                attr_key = attribute[self.runtime_prefix_len:]
                runtime_policy[attr_key] = text_value

        if runtime_policy:
            policy[ego_constants.RUNTIME_POLICY] = runtime_policy
        return policy

    def __parse_metadata(self, metadata, agg_id, valid_policy_id_set):
        """
        Parse metadata of aggregate, check and retrieve policy-related content.
        If the policy does not have all of the correct attributes defined, then
        return empty dict.
        e.g., {'initialpolicy': {'id': 3},
               'runtimepolicy': {'id': 2, 'run_interval': 3, 'stabilization':3,
               'max_parallel': 3, 'threshold': 3, 'action': 'migrate_vm'}}
        """

        policy = {}
        # initial policy
        initial_prefix_len = (
                len(ego_constants.INITIAL_POLICY) +
                len(ego_constants.POLICY_DELIMITER))
        initial_policy = {}
        if ego_constants.INITIAL_POLICY_ID in metadata:
            try:
                text_value = metadata[ego_constants.INITIAL_POLICY_ID]
                key_id = ego_constants.INITIAL_POLICY_ID[initial_prefix_len:]
                if int(text_value) not in valid_policy_id_set:
                    valid_ids_str = ','.join(
                            [str(pid) for pid in valid_policy_id_set])
                    msg = (_LW('The specified initial policy ID (%(value)s) '
                             'for attribute (%(attr)s) is invalid for '
                             'aggregate (%(agg_id)s). '
                             'Specify one of the following policy IDs '
                             'instead: (%(valid_ids)s).') %
                           {'value': text_value,
                            'attr': ego_constants.INITIAL_POLICY_ID,
                            'valid_ids': valid_ids_str,
                            'agg_id': agg_id})
                    LOG.warning(msg)
                else:
                    initial_policy[key_id] = int(text_value)
                    policy[ego_constants.INITIAL_POLICY] = initial_policy
            except Exception:
                msg = (_LW('The value (%(value)s) of attribute (%(attr)s) '
                        'in metadata is not the expected type (%(type)s) '
                        'for aggregate (%(agg_id)s).'
                        ) % {'value': text_value,
                             'attr': ego_constants.INITIAL_POLICY,
                             'type': ego_constants.ATTRIBUTE_TYPE_INT,
                             'agg_id': agg_id})
                LOG.warning(msg)

        # runtime policy
        runtime_policy = {}
        runtime_prefix_len = (
            len(ego_constants.RUNTIME_POLICY) +
            len(ego_constants.POLICY_DELIMITER))
        runtime_policy_attribute = [
            (ego_constants.RUNTIME_POLICY_ID,
             ego_constants.ATTRIBUTE_TYPE_INT),
            (ego_constants.RUNTIME_POLICY_RUN_INTERVAL,
             ego_constants.ATTRIBUTE_TYPE_INT),
            (ego_constants.RUNTIME_POLICY_STABILIZATION,
             ego_constants.ATTRIBUTE_TYPE_INT),
            (ego_constants.RUNTIME_POLICY_MAX_PARALLEL,
             ego_constants.ATTRIBUTE_TYPE_INT),
            (ego_constants.RUNTIME_POLICY_THRESHOLD,
             ego_constants.ATTRIBUTE_TYPE_INT),
            (ego_constants.RUNTIME_POLICY_ACTION,
             ego_constants.ATTRIBUTE_TYPE_STR),
        ]
        log_flag = False
        for attr in runtime_policy_attribute:
            if attr[0] in metadata:
                log_flag = True
                break
        try:
            for attribute in runtime_policy_attribute:
                if attribute[0] not in metadata:
                    msg = (_LW('Host aggregate ID (%(agg_id)s) does '
                            'not have attribute (%(attr)s) in metadata.'
                            ) % {'attr': attribute[0], 'agg_id': agg_id})
                    if log_flag:
                        LOG.warning(msg)
                    raise exception.InvalidMetadata(reason=msg)
                else:
                    text_value = metadata[attribute[0]]
                    attr_key = attribute[0][runtime_prefix_len:]
                    if attribute[1] == ego_constants.ATTRIBUTE_TYPE_STR:
                        runtime_policy[attr_key] = text_value
                    elif attribute[1] == ego_constants.ATTRIBUTE_TYPE_INT:
                        try:
                            runtime_policy[attr_key] = int(text_value)
                        except Exception:
                            if log_flag:
                                msg = (_LW('The value (%(value)s) of '
                                        'attribute (%(attr)s) in metadata '
                                        'is not the expected type (%(type)s) '
                                        'for aggregate (%(agg_id)s).'
                                        ) % {'value': text_value,
                                             'attr': attribute[0],
                                             'type': attribute[1],
                                             'agg_id': agg_id})
                                LOG.warning(msg)
                            raise exception.InvalidMetadata(reason=msg)
                    else:
                        if log_flag:
                            msg = (_LW('The type (%(type)s) of attribute '
                                    '(%(attr)s) in metadata is invalid for '
                                    'host aggregate (%(agg_id)s). It should '
                                    'be "%(type_int)s" or "%(type_str)s".'
                                       ) % {'attr': attribute[0],
                                            'type': attribute[1],
                                            'type_int':
                                            ego_constants.ATTRIBUTE_TYPE_INT,
                                            'type_str':
                                            ego_constants.ATTRIBUTE_TYPE_STR,
                                            'agg_id': agg_id})
                            LOG.warning(msg)
                        raise exception.Invalid()
                    # check if the runtime policy id is valid
                    if (attr_key == ego_constants.POLICY_ID
                            and int(text_value) not in valid_policy_id_set):
                        valid_ids_str = ','.join(
                            [str(pid) for pid in valid_policy_id_set])
                        msg = (_LW('The specified runtime policy ID '
                                 '(%(value)s) for attribute (%(attr)s) is '
                                 'invalid for aggregate (%(agg_id)s). '
                                 'Specify one of the following policy IDs '
                                 'instead: (%(valid_ids)s).') %
                               {'value': text_value,
                                'attr': ego_constants.RUNTIME_POLICY_ID,
                                'valid_ids': valid_ids_str,
                                'agg_id': agg_id})
                        LOG.warning(msg)
                        raise exception.Invalid()
            # runtime policy value check
            self.__runtime_policy_check(
                    runtime_policy[ego_constants.POLICY_RUN_INTERVAL],
                    ego_constants.POLICY_RUN_INTERVAL, agg_id,
                    ego_constants.MIN_POSITIVE_INT,
                    ego_constants.MAX_INT_32,
                    log_flag)
            self.__runtime_policy_check(
                    runtime_policy[ego_constants.POLICY_STABILIZATION],
                    ego_constants.POLICY_STABILIZATION, agg_id,
                    ego_constants.MIN_POSITIVE_INT,
                    ego_constants.MAX_INT_32,
                    log_flag)
            self.__runtime_policy_check(
                    runtime_policy[ego_constants.POLICY_MAX_PARALLEL],
                    ego_constants.POLICY_MAX_PARALLEL, agg_id,
                    ego_constants.MIN_POSITIVE_INT,
                    ego_constants.MAX_INT_32,
                    log_flag)
            self.__runtime_policy_check(
                    runtime_policy[ego_constants.POLICY_THRESHOLD],
                    ego_constants.POLICY_THRESHOLD, agg_id,
                    ego_constants.MIN_INT_32,
                    ego_constants.MAX_INT_32,
                    log_flag)
            if (runtime_policy[ego_constants.POLICY_ACTION] not in
                self.policy_actions):
                if log_flag:
                    expect_val = ', '.join(self.policy_actions)
                    msg = (_LW('The value (%(value)s) of attribute (%(attr)s) '
                            'in metadata is invalid for host aggregate '
                            '(%(agg_id)s). It should be one of: %(val_exp)s.'
                            ) % {'value':
                                 runtime_policy[ego_constants.POLICY_ACTION],
                                'attr': self.__build_runtime_string(
                                             ego_constants.POLICY_ACTION),
                                'val_exp': expect_val,
                                'agg_id': agg_id})
                    LOG.warning(msg)
                raise exception.Invalid()
        except Exception:
            return policy
        policy[ego_constants.RUNTIME_POLICY] = runtime_policy
        return policy

    def __build_runtime_string(self, attr):
        return '%s-%s' % (ego_constants.RUNTIME_POLICY, attr)

    def __runtime_policy_check(self, value, attr, agg_id,
                               val_min, val_max, log_flag):
        if value < val_min or value > val_max:
            if log_flag:
                msg = (_LW('The value (%(val)s) of attribute (%(attr)s) '
                        'in metadata is invalid for host aggregate '
                        '(%(agg_id)s). It should be within range '
                        '(%(val_min)s - %(val_max)s).'
                        ) % {'attr': self.__build_runtime_string(attr),
                             'val': value,
                             'val_min': val_min,
                             'val_max': val_max,
                             'agg_id': agg_id})
                LOG.warning(msg)
            raise exception.Invalid()

    def get_host_aggregate_metadata(self):
        """Get host aggregate's metadata from DB."""
        aggregate_id_metadata_dict = {}
        try:
            aggregates = objects.AggregateList.get_all(self.context)
            for aggregate in aggregates:
                metadata = db.aggregate_metadata_get(self.context,
                                                     aggregate.id)
                aggregate_id_metadata_dict[aggregate.id] = (
                        metadata)
        except Exception as ex:
            LOG.warning(_LW('Failed to retrieve host aggregate metadata from '
                            'the database due to: %s') % six.text_type(ex))
        return aggregate_id_metadata_dict

    def sync_up_aggregate_policy(self, valid_policy_id_set, aggregate_id=None):
        """
        Get host aggregate policy map.

        If an aggregate_id is provided, only that aggregate's policy will be
        synced. Otherwise all aggregates' policies will be synced.
        """
        aggregate_id_metadata_dict = {}
        if aggregate_id:
            agg = objects.Aggregate.get_by_id(self.context, aggregate_id)
            aggregate_id_metadata_dict[aggregate_id] = agg.metadata
        else:
            aggregate_id_metadata_dict = self.get_host_aggregate_metadata()

        aggregate_id_policy_dict = {}
        aggid_policy_nc = {}
        for agg_id, metadata in six.iteritems(aggregate_id_metadata_dict):
            aggregate_id_policy_dict[agg_id] = self.__parse_metadata(
                                metadata, agg_id, valid_policy_id_set)
            aggid_policy_nc[agg_id] = (
                self.__filter_metadata_for_policy_show(metadata, agg_id))

        if aggregate_id:
            self.aggregate_id_policy_dict.update(aggregate_id_policy_dict)
            self.aggid_policy_no_check.update(aggid_policy_nc)
        else:
            self.aggregate_id_policy_dict = aggregate_id_policy_dict
            self.aggid_policy_no_check = aggid_policy_nc

    def aggregate_create(self, payload):
        # update cache 'aggregateid_metadatas' after aggregate was created
        aggregate_id = payload['aggregate_id']
        LOG.info(_LI('Creating aggregate with ID %(agg_id)s in cache.') %
                 {'agg_id': aggregate_id})
        self.aggregateid_metadatas[aggregate_id] = {}

    def aggregate_delete(self, payload):
        # update cache 'aggregateid_metadatas' after aggregate was deleted
        aggregate_id = payload['aggregate_id']
        if aggregate_id in self.aggregateid_metadatas:
            LOG.info(_LI('Deleting aggregate with ID %(agg_id)s from cache.') %
                     {'agg_id': aggregate_id})
            self.aggregateid_metadatas.pop(aggregate_id)

    def aggregate_host_add(self, host, aggregate_id, need_metadata=False):
        """
        Update aggregate ID for specified host and change allocation of all
        VMs running on this host.
        """
        LOG.info(_LI('Adding host %(host)s to aggregate with ID %(agg_id)s in '
                     'cache.') % {'host': host, 'agg_id': aggregate_id})
        # Change host aggregate id firstly
        if (host in self.host_aggregates and
               aggregate_id not in self.host_aggregates[host]):
            self.host_aggregates[host].append(aggregate_id)
        else:
            self.host_aggregates[host] = [aggregate_id]
        if need_metadata:
            self.update_host_metadata()

    def aggregate_host_remove(self, host, aggregate_id, need_metadata=False):
        """
        Update aggregate ID for specified host and change allocation of all
        VMs running on this host.
        """
        LOG.info(_LI('Removing host %(host)s from aggregate with ID '
                     '%(agg_id)s in cache.') %
                 {'host': host, 'agg_id': aggregate_id})
        # Change host aggregate id firstly
        if host in self.host_aggregates:
            if aggregate_id in self.host_aggregates[host]:
                self.host_aggregates[host].remove(aggregate_id)

        if need_metadata:
            self.update_host_metadata()

    def aggregate_set_metadata(self, payload):
        # update cache 'aggregateid_metadatas' and RPI for host who belong
        # to this aggregate after aggregate metadata was changed
        aggregate_id = payload['aggregate_id']
        meta_data = payload.get('meta_data')
        cache_meta_data = self.aggregateid_metadatas.get(aggregate_id, {})
        for (meta_k, meta_v) in meta_data.items():
            # this is for the unspecified metadata case
            if not meta_v:
                if meta_k in cache_meta_data:
                    cache_meta_data.pop(meta_k)
            else:
                cache_meta_data[meta_k] = meta_v
        self.aggregateid_metadatas[aggregate_id] = cache_meta_data
        self.update_host_metadata()

    def sync_aggregates_and_metadatas(self, need_metadata=False):
        """
        Retrieve aggregates and related metadata from the database.
        :param need_metadata: If True, need to load metadata for each aggregate
        :returns: return the relationship of host-aggregates-metadata
        """
        aggregateid_hosts = {}
        aggregateid_metadata = {}
        host_aggregates = {}

        def update_aggregate_metadatas(self, aggregates):
            for aggregate in aggregates:
                metadata = db.aggregate_metadata_get(self.context,
                                                     aggregate.id)
                aggregateid_metadata[aggregate.id] = metadata

        aggregates = objects.AggregateList.get_all(self.context)
        for aggregate in aggregates:
            hosts = db.aggregate_host_get_all(self.context,
                                              aggregate.id)
            aggregateid_hosts[aggregate.id] = hosts
            aggregateid_metadata[aggregate.id] = {}
        for aggregateid, hosts in aggregateid_hosts.items():
            for host in hosts:
                if host in host_aggregates:
                    host_aggregates[host].append(aggregateid)
                else:
                    host_aggregates[host] = [aggregateid]
        if need_metadata:
            update_aggregate_metadatas(self, aggregates)
            self.update_host_metadata(host_aggregates, aggregateid_metadata)

        # update cache
        self.host_aggregates = host_aggregates
        self.aggregateid_metadatas = aggregateid_metadata

    def update_host_metadata(self, host_aggregates=None,
                             aggregateid_metadata=None):
        """Update the metadata for host in cache."""
        host_metadatas = {}
        if host_aggregates is None:
            host_aggregates = self.host_aggregates
        if aggregateid_metadata is None:
            aggregateid_metadata = self.aggregateid_metadatas
        for host, aggregateids in host_aggregates.items():
            host_metadata = collections.defaultdict(set)
            for aggregateid in aggregateids:
                metadata = aggregateid_metadata.get(aggregateid)
                if metadata:
                    for k, v in metadata.items():
                        host_metadata[k].add(v)
            host_metadatas[host] = host_metadata
        self.host_metadatas = host_metadatas
