# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import collections
import copy
import re

import eventlet
import six

from nova import compute as nova_compute
from nova.compute import flavors
from nova.compute import task_states
from nova.compute import vm_states
from nova import context
from nova import db
from nova import exception
from nova.ibm.ego import ego_constants
from nova.ibm.ego import ego_gettextutils
from nova.ibm.ego import ego_utils
from nova.ibm.ego import hapolicy
from nova.ibm.ego import hv_status
from nova import objects
from nova.objects import base as base_obj
from nova.scheduler.ibm.ego import ego_exception
from nova.scheduler.ibm.ego import ego_filter_driver
from nova.scheduler.ibm.ego import ego_handler
from nova.scheduler.ibm.ego import ego_host
from nova.scheduler.ibm.ego import ego_host_aggregate
from nova.scheduler.ibm.ego import ego_resreq_utils

from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_messaging import exceptions as oslo_exceptions
from oslo_utils import importutils

_ = ego_gettextutils._
_LI = ego_gettextutils._LI
_LW = ego_gettextutils._LW
_LE = ego_gettextutils._LE

LOG = logging.getLogger(__name__)

CONF = cfg.CONF
CONF.import_opt('policy_per_host_aggregate', 'nova.ibm.ego.ego_flags')
CONF.import_opt('scheduler_default_filters', 'nova.scheduler.host_manager')
CONF.import_opt('prs_sync_aggregate_reserveon', 'nova.ibm.ego.ego_flags')
CONF.import_opt('prs_allocation_per_server_group', 'nova.ibm.ego.ego_flags')
CONF.import_opt('ego_extension_managers', 'nova.ibm.ego.ego_flags')

# Hypervisor types that need to release EGO allocations if the state of the VMs
# on the respective hypervisors are in any of the RELEASE_ALLOC_VM_STATES.
HYPERVISORS_RELEASE_ALLOC = [ego_constants.HYPERVISOR_TYPE_QEMU,
                             ego_constants.HYPERVISOR_TYPE_KVM,
                             ego_constants.HYPERV_HYPERVISOR_TYPE,
                             ego_constants.DOCKER_HYPERVISOR_TYPE,
                             ego_constants.HYPERVISOR_TYPE_VMWARE_DB,
                             ego_constants.HYPERVISOR_TYPE_ESX_DB,
                             ego_constants.ZVM_HYPERVISOR_TYPE]
RELEASE_ALLOC_VM_STATES = [vm_states.SUSPENDED, vm_states.STOPPED,
                           vm_states.ERROR, vm_states.SHELVED,
                           vm_states.SHELVED_OFFLOADED,
                           vm_states.SOFT_DELETED]


class EGOManager(object):
    """EGO C API Python Wrapper."""
    def __init__(self, client_name=None):
        """
        Construct function for EGO Manager
        """
        self.ego_hdl = ego_handler.EGOHandler(client_name)
        # instance_uuid:alloc_id
        self.inst_allocid_map = {}
        # instance_uuid:allocname
        self.inst_allocname_map = {}
        # instance_uuid:deci_id
        self.inst_dec_map = {}
        # instance_uuid:deci_id
        self.dec_inst_map = {}
        self.inst_egohost_map = {}
        self.inst_resreq_map = {}
        self.inst_maxslots_map = {}
        self.inst_rg_map = {}
        self.inst_props_map = {}
        # {key: alloc_name, value: alloc_id}
        self.allocname_id_map = {}
        self.decid_list = []
        # {key: dec_id, value: alloc_id}
        self.dec_allocid_map = {}
        # orphan decision cache
        self.orphan_decid_list = []
        # Cache for removed policy group
        self.cleared_policy_group = {}

        self.extension_managers = [importutils.import_object(ext, self)
                                     for ext in CONF.ego_extension_managers]

        self.ha_rpc_api = hapolicy.API()
        # host aggregates related
        self.aggregate_manager = ego_host_aggregate.EGOHostAggregate()

        if client_name:
            self.__init_for_registered_client(client_name)

        eventlet.spawn_after(1, self.__conn_recover_in_thread, client_name)

        LOG.debug("Platform Resource Scheduler successfully initialized "
                  "the EGO manager.")

    def __init_for_registered_client(self, client_name):
        self.need_metadata = False
        self.need_aggregate = CONF.policy_per_host_aggregate
        # data format is {host name: set([inst_uuid1, inst_uuid2, inst_uuidx])}
        self.egohost_inst = collections.defaultdict(set)
        # map from EGO host(RPI RES_NAME) to EGO host attributes
        self.all_egohost_attrs = {}
        # map from host(service['host']) to set of EGO host(RPI RES_NAME)
        self.host_egohost = collections.defaultdict(set)
        # data format is {instance_uuid: flavor_id}
        self.inst_flv_idmap = {}
        self.all_flavors = collections.defaultdict(list)
        # data format is {instacne_tpye_id: flavor_id}
        self.typeid_flavorid_map = {}
        # Set vcpu & RAM overuse ratio to EGO resource plan
        self.ego_filter_driver = ego_filter_driver.EGOFilterDriver()
        self.ctxt = context.get_admin_context()
        self.ego_host_obj = ego_host.EGOHost()
        self.egohost_type_map = {}
        self.host_api = nova_compute.HostAPI()
        self.__construct_egohost_type_map()
        self.__update_metadata_flag()
        if not self.ego_hdl.is_conn_lost():
            self.ego_sync_up()

    def __ego_construct_alloc_map(self, instances=None):
        """
        Used for recovery, sync up allocations and OpenStack instances.
        May need to sync up consumer and resource groups later.
        alloc[0]: EGO allocation name
        alloc[1]: EGO allocation ID
        alloc[2]: allocation decision list
                  alloc[2][0]: allocation unit placement host
                  alloc[2][1]: decision id
                  alloc[2][2]: sub demand name
                  alloc[2][3]: allocation unit resource, 1:consume,
                                                         2:not consume
                  alloc[2][4]: decision resource requirement sting
                  alloc[2][5]: resource group
        alloc[3]: allocation client name
        """
        allocList = self.ego_hdl.ego_get_all_alloc()
        if not instances:
            instances = objects.InstanceList.get_all(self.ctxt,
                            expected_attrs=['metadata', 'system_metadata'])
        self.inst_allocid_map.clear()
        self.inst_dec_map.clear()
        self.inst_egohost_map.clear()
        self.inst_resreq_map.clear()
        self.inst_maxslots_map.clear()
        self.inst_rg_map.clear()
        self.decid_list = []
        self.dec_inst_map.clear()
        self.dec_allocid_map.clear()
        self.allocname_id_map.clear()
        insts_without_decid = []
        for inst in instances:
            inst_uuid = inst['uuid']
            try:
                decision_id = ego_resreq_utils.get_decision_id(inst)
            except exception.InstanceNotFound:
                # instance has been deleted since the list of instances has
                # been fetched, just ignore it
                continue
            if decision_id:
                self.inst_dec_map[inst_uuid] = decision_id
                self.dec_inst_map[decision_id] = inst_uuid
            else:
                insts_without_decid.append(inst_uuid)

        for alloc in allocList:
            if alloc[3] != ego_constants.EGO_ALLOC_CLIENT_NAME:
                continue
            self.allocname_id_map[alloc[0]] = alloc[1]
            decisions = alloc[2]
            for decision in decisions:
                decid = decision[1]
                self.decid_list.append(decid)
                self.dec_allocid_map[decid] = alloc[1]
                inst_uuid = self.dec_inst_map.get(decid)

                if not inst_uuid:
                    # for standalone instances, if there is no decision_id info
                    # in the DB; its corresponding decision_id should be reset
                    # into its metadata
                    if alloc[0] in insts_without_decid:
                        inst_uuid = alloc[0]
                        metadata = {'decision_id': decid}
                        ego_resreq_utils.update_instance_system_metadata(
                                                        inst_uuid, metadata)
                        self.inst_dec_map[inst_uuid] = decid
                        self.dec_inst_map[decid] = inst_uuid
                    else:
                        continue
                self.inst_allocname_map[inst_uuid] = alloc[0]
                self.inst_allocid_map[inst_uuid] = alloc[1]
                self.inst_egohost_map[inst_uuid] = decision[0]
                self.inst_maxslots_map[inst_uuid] = decision[3]
                self.inst_resreq_map[inst_uuid] = decision[4]
                self.inst_rg_map[inst_uuid] = decision[5]

    def __construct_egohost_type_map(self):
        compute_nodes = self.host_api.compute_node_get_all(self.ctxt)
        for compute in compute_nodes:
            hypervisor_hostname = compute.get('hypervisor_hostname')
            if not hypervisor_hostname:
                continue
            ego_host_name = ego_resreq_utils.get_rpi_res_name(
                                                hypervisor_hostname)
            self.egohost_type_map[ego_host_name] = (compute['hypervisor_type'])

    def get_htype_from_hostname(self, ego_host_name):
        hypervisor_type = self.egohost_type_map.get(ego_host_name)
        # Reconstruct the egohost_type_map in case there are newly added hosts
        if not hypervisor_type:
            self.egohost_type_map = {}
            self.__construct_egohost_type_map()
            hypervisor_type = self.egohost_type_map.get(ego_host_name)
        return hypervisor_type

    def support_modify_allocation_quota(self, hypervisor_type, instanceid):
        if instanceid:
            # heat allocation's quota should not be modified
            allocname = self.inst_allocname_map.get(instanceid)
            alloc_type = (
                ego_resreq_utils.get_alloc_type_from_allocname(allocname))
            if alloc_type == ego_constants.HEAT:
                return False
        return hypervisor_type and hypervisor_type in HYPERVISORS_RELEASE_ALLOC

    def __get_allocid_from_instid(self, instanceid):
        allocid = self.inst_allocid_map.get(instanceid)
        if not allocid:
            # instance id is not in cache, update from EGO persistent file
            self.__ego_construct_alloc_map()
            allocid = self.inst_allocid_map.get(instanceid)
        if not allocid:
            try:
                # If it doesn't exist, see if the instance is out there and
                # was either missed or the UUID was changed at some point
                expected_attrs = ['metadata', 'system_metadata']
                inst = objects.Instance.get_by_uuid(
                    self.ctxt, instanceid, expected_attrs=expected_attrs)
                self.create_alloc_for_existing_inst(inst)
                allocid = self.inst_allocid_map.get(instanceid)
            # If the instance isn't found, then can't add the allocations
            except Exception as e:
                LOG.exception(e)
        return allocid

    def __get_decid_from_instid(self, instanceid):
        decid = self.inst_dec_map.get(instanceid)
        if not decid:
            # instance id is not in cache, update from EGO persistent file
            self.__ego_construct_alloc_map()
            decid = self.inst_dec_map.get(instanceid)
        return decid

    def __get_instance_list(self):
        try:
            return objects.InstanceList.get_all(self.ctxt,
                       expected_attrs=['metadata', 'system_metadata'])
        except Exception as e:
            LOG.exception(e)
            return []

    @lockutils.synchronized('ego_sync_up', 'prs-ego-sync-up-')
    def ego_sync_up(self):
        """Sync up EGO with info in DB.

        This method will be called by periodic task, too
        """
        LOG.info(_LI('Started syncing EGO with info from the database.'))
        # Get all instances from DB
        instance_list = self.__get_instance_list()
        self.__ego_construct_alloc_map(instances=instance_list)

        # Get all flavors from DB, the return data doesn't include the flavors
        # that have been deleted before.
        flavor_refs = {
            f.flavorid: f for f in objects.FlavorList.get_all(self.ctxt)}
        for flavor in flavor_refs.values():
            self.typeid_flavorid_map[flavor['id']] = flavor['flavorid']
        for inst in instance_list:
            instance_type_id = inst.get('instance_type_id')
            if instance_type_id:
                flavorid = self.typeid_flavorid_map.get(instance_type_id)
                if not flavorid:
                    # Visit the DB to get deleted flavors
                    flavor = objects.Flavor.get_by_id(self.ctxt,
                                                      instance_type_id)
                    flavorid = flavor['flavorid']
                self.inst_flv_idmap[inst['uuid']] = flavorid

        all_flavors = collections.defaultdict(list)
        for instid, egohost in self.inst_egohost_map.items():
            self.egohost_inst[egohost].add(instid)
            if self.inst_flv_idmap and self.inst_flv_idmap.get(instid):
                all_flavors[egohost].append(self.inst_flv_idmap[instid])
        self.all_flavors = all_flavors

        self.ego_sync_aggregates_metadatas()
        self.sync_up_resource_groups()
        self._iterate_extension_managers("sync_up")
        self.ego_sync_up_hosts()
        if CONF.prs_allocation_per_server_group:
            self.ego_sync_up_server_group()

        # There is time window for setting RPI resource when sync up hosts,
        # we need make sure RPI resource is ready before sync up instances,
        # so sleep 1s.
        eventlet.sleep(1)
        self.__ego_sync_up_instances(instance_list)
        LOG.info(_LI('Finished syncing EGO with info from the database.'))

    def __ego_check_connection(self, client_name=None):
        """
        Check EGO connections periodically, this is used
        to check if EGO Resource Manager was restarted

        :returns: True if connection lost
                  False if connection is ok
        """
        if self.ego_hdl.is_conn_lost():
            # try to recover connection
            self.ego_hdl.conn_and_logon(client_name)

            # sync up after reset EGO connection
            if not self.ego_hdl.is_conn_lost() and client_name:
                # Clear this cache so that RPI resources can be set
                # instantly once reconnecting EGO.
                self.all_egohost_attrs.clear()

                self.ego_sync_up()
                LOG.info(_LI('Platform Resource Scheduler successfully '
                       'reconnected to EGO.'))
        else:
            self.ego_hdl.ego_uname()

    def __get_instances_by_host(self, ego_host, is_active=False):
        inst_uuids = []
        for inst_uuid in self.egohost_inst[ego_host]:
            if is_active:
                max_slots = self.inst_maxslots_map.get(inst_uuid, 0)
                if max_slots > 0:
                    inst_uuids.append(inst_uuid)
            else:
                inst_uuids.append(inst_uuid)
        return inst_uuids

    def __gen_filter_props(self, payload, ignore_filter=True, stand_alone=True,
                           scope=ego_constants.FILTER_WORKING_SCOPE_CREATE):
        instance_id = payload['uuid']
        if not isinstance(payload, objects.Instance):
            payload = objects.Instance.get_by_uuid(self.ctxt,
                                instance_id)
            # Force the loading of the metadata attributes so that they are
            # put into the dictionary via the `obj_to_primitive` call.
            payload.metadata
            payload.system_metadata
        # Switch back to a dict for compatibility
        LOG.debug("Converting instance object (%s) to a dict."
                  % payload.uuid)
        instance_type = base_obj.obj_to_primitive(payload.get_flavor())
        payload = base_obj.obj_to_primitive(payload)
        htype = self.get_htype_from_hostname(payload['node'])
        image = {'properties': {'hypervisor_type': htype}}

        return {"request_spec": {"instance_id": instance_id,
                                 "instance_type": instance_type,
                                 "image": image,
                                 "instance_properties": payload,
                                 "operation_type": scope,
                                 ego_constants.STAND_ALONE: stand_alone},
                "ignore_filters": ignore_filter,
                "force_hosts": [ego_resreq_utils.get_rpi_res_name(
                                                     payload['node'])],
                "context": self.ctxt}

    def __ego_sync_up_instances(self, instance_list):
        """
        This function is used to make instances and allocations consistent.
        """
        LOG.info(_LI('Started syncing EGO allocations with instances from '
                     'the database.'))
        alloc_for_sg = CONF.prs_allocation_per_server_group

        inst_group_map = {}
        # make a map { key: inst_uuid, value: server_group}
        server_groups = objects.InstanceGroupList.get_all(self.ctxt)
        for group in server_groups:
            if 'legacy' in group.policies:
                continue
            for inst_uuid in group.members:
                # In the case of one instance belonging to multiple groups,
                # log warning message.
                if inst_group_map.get(inst_uuid):
                    gp_uuid = inst_group_map[inst_uuid].uuid
                    LOG.warning(_LW('Platform Resource Scheduler already '
                                    'found instance (%(inst_uuid)s) in server '
                                    'group %(gp_uuid)s; additional groups '
                                    'will be ignored.')
                                % {"inst_uuid": inst_uuid,
                                   "gp_uuid": gp_uuid})
                    continue
                inst_group_map[inst_uuid] = group

        # Loop through all instances in the DB
        for inst in instance_list:
            vm_state = inst.get('vm_state')
            task_state = inst.get('task_state')

            instid = inst['uuid']
            # clean the scheduler cache
            if (vm_state == vm_states.ACTIVE) and (not task_state) and (
                    instid in self.ego_filter_driver.schedule_cache):
                self.ego_filter_driver.schedule_cache.pop(instid)

            # 1. if instance does not have its own allocation, create it
            try:
                if alloc_for_sg and instid in inst_group_map:
                    # This branch handles an allocation per server group case
                    if instid not in self.inst_allocid_map:
                        group = inst_group_map.get(instid)
                        self.resize_alloc_for_exist_inst(inst,
                                             group_uuid=group.uuid)
                    else:
                        # Handle the upgrade case (Icehouse --> Juno) in which
                        # a VM in the server group with standalone alloc;
                        # upgrade to decisions in server group allocation
                        allocname = self.inst_allocname_map.get(instid)
                        if allocname:
                            alloc_type = (ego_resreq_utils.
                                get_alloc_type_from_allocname(allocname))
                            if alloc_type == ego_constants.STAND_ALONE:
                                old_resreq = self.inst_resreq_map.get(instid)
                                self.ego_free_alloc(allocname)
                                group = inst_group_map.get(instid)
                                self.resize_alloc_for_exist_inst(inst,
                                                     group_uuid=group.uuid,
                                                     old_resreq=old_resreq)
                else:
                    # This branch handles allocation per VM cases.
                    if instid in inst_group_map:
                        if instid not in self.inst_allocid_map:
                            # In server group and no allocation, not mode
                            # switch from per server group to per VM case,
                            # can only add standalone allocation.
                            self.create_alloc_for_existing_inst(inst)
                        else:
                            # In server group and have allocation, mode switch
                            # from per server group to per VM case, create
                            # the allocation using the old resreq.
                            allocname = self.inst_allocname_map.get(instid)
                            alloc_type = None
                            if allocname:
                                alloc_type = (ego_resreq_utils.
                                    get_alloc_type_from_allocname(allocname))
                            if alloc_type == ego_constants.SERVER_GROUP:
                                # Switch for server group alloc to VM alloc.
                                # Create decision here for instance and
                                # server group decision will be deleted later.
                                old_resreq = self.inst_resreq_map.get(instid)
                                self.create_alloc_for_existing_inst(inst,
                                              alloc_for_sg=alloc_for_sg,
                                              old_resreq=old_resreq)
                    else:
                        # Standalone instance without an alloc; create one.
                        if instid not in self.inst_allocid_map:
                            self.create_alloc_for_existing_inst(inst)
            except (ego_exception.EGOExpCreateAllocFailed,
                    ego_exception.EGOExpResizeAllocFailed,
                    ego_exception.EGOExpConnLost) as e:
                LOG.warn(six.text_type(e))
                LOG.exception(e)
                continue
            except Exception as e:
                LOG.warning(_LW('Platform Resource Scheduler could not create '
                           'EGO resource allocation for instance '
                           '(%(instanceid)s) due to: (%(reason)s)')
                         % {"instanceid": instid,
                            "reason": six.text_type(e)})
                LOG.exception(e)
                continue

            # 2. if the instance is running a task or in the building state
            #    skip it (i.e., we don't want to sync BUILDING VMs since their
            #    metadata is still fluctuating and unreliable)
            if task_state or vm_state == vm_states.BUILDING:
                continue

            # 3. if the instance doesn't have an allocation, skip it
            #    (e.g., for instances in the shelved-offloaded state)
            if instid not in self.inst_allocid_map:
                continue

            # 4. if the instance doesn't have both a host and a launched_at
            #    timestamp, then we should delete its allocation.
            if not (inst['host'] and inst['launched_at']):
                decid = self.__get_decid_from_instid(instid)
                try:
                    if decid:
                        self.release_decision(instid)
                except (ego_exception.EGOExpConnLost,
                        ego_exception.EGOExpReleaseDecisionFailed) as e:
                    LOG.error(_LE('Platform Resource Scheduler could not '
                                'release decision %(decid)s due to: '
                                '(%(reason)s)')
                             % {'decid': decid, 'reason': six.text_type(e)})
                continue

            # 5. migrate allocation if VM and EGO allocation isn't on the
            #    same host for one VM
            ego_host = ego_resreq_utils.get_rpi_res_name(inst['node'])
            htype = self.get_htype_from_hostname(ego_host)
            is_mod_alloc_supported = (
                self.support_modify_allocation_quota(htype, instid))
            is_mod_alloc_needed = False
            inst_host = self.inst_egohost_map.get(instid)
            if ego_host and ego_host != inst_host:
                is_mod_alloc_needed = True

            # 6. resize allocation if VM and EGO allocation differs
            max_slots = self.inst_maxslots_map.get(instid, 0)
            # set max slots to 0 (doesn't consume resources) for the VM
            # allocation if the VM doesn't consume resources while it's off.
            if vm_state in RELEASE_ALLOC_VM_STATES:
                if max_slots == 1 and is_mod_alloc_supported:
                    max_slots = 0
                    is_mod_alloc_needed = True
            # set max slots to 1 (consumes resources) for the VM allocation
            # if the VM consumes resources while it's off.
            elif max_slots == 0 and is_mod_alloc_supported:
                max_slots = 1
                is_mod_alloc_needed = True

            # Sync up instance specification and allocation consumption
            # if the allocation consumption which did not match the
            # instance specification, just resize the allocation based
            # on the instance specification
            ego_resreq = self.inst_resreq_map.get(instid)
            filter_properties = self.__gen_filter_props(inst, scope=None)
            self._iterate_extension_managers("handle_for_schedule_info",
                                              filter_properties)
            try:
                new_consume = ego_resreq_utils.get_consume_str(
                                                   filter_properties)
            except exception.InstanceNotFound:
                # instance has been deleted since the list of instances has
                # been fetched, just ignore it
                continue
            old_consume = ego_resreq_utils.get_criteria_str_by_resreq(
                                               ego_resreq, 'consume')
            if old_consume != new_consume:
                ego_resreq = self.__get_new_resreq(
                    ego_resreq,
                    new_consume,
                    prefix=ego_constants.RES_REQ_CONSUME_PREFIX,
                    postfix=ego_constants.RES_REQ_CONSUME_POSTFIX)
                is_mod_alloc_needed = True

            # in aggregate mode, if CONF.prs_sync_aggregate_reserveon is True,
            # it's required to sync up reserveon info with resource group.
            new_reserveon_str = None
            if self.need_aggregate and CONF.prs_sync_aggregate_reserveon:
                new_reserveon_str = self.__get_reserveon_based_on_host(
                                                          ego_host)
            elif not self.need_aggregate:
                new_reserveon_str = (ego_constants.RES_REQ_RESERVEON_PREFIX
                                     + ego_constants.RESOURCE_GROUP_ALL +
                                     ego_constants.RES_REQ_RESERVEON_POSTFIX)
            old_reserveon_str = (
                         re.match(r"reserveon\(.*?\)", ego_resreq).group(0))
            if new_reserveon_str and new_reserveon_str != old_reserveon_str:
                is_mod_alloc_needed = True
                new_reserveon_str = (ego_constants.RES_REQ_RESERVEON_PREFIX
                                     + new_reserveon_str +
                                     ego_constants.RES_REQ_RESERVEON_POSTFIX)
                ego_resreq = self.__get_new_resreq(
                        ego_resreq,
                        new_reserveon_str,
                        prefix=ego_constants.RES_REQ_RESERVEON_PREFIX,
                        postfix=ego_constants.RES_REQ_RESERVEON_POSTFIX)

            if ego_resreq and is_mod_alloc_needed:
                try:
                    extra_resreq = self.__get_extra_resreq(
                                    force_node_list=[ego_host])
                    (allocId, decid, egohost, res_group) = (
                        self.__ego_update_decision(inst['uuid'],
                                                ego_resreq,
                                                extra_resreq,
                                                prefer_same_host=1,
                                                force_realloc_migrate=True,
                                                max_slot=max_slots))
                    cache_dict = {"max_slot": max_slots, "resreq": ego_resreq,
                            "res_group": res_group, "ego_host": ego_host}
                    self.__cache_update_inst(inst['uuid'], cache_dict)
                    self._iterate_extension_managers("instance_update",
                                                  filter_properties)
                except (ego_exception.EGOExpUpdateDecFailed,
                        ego_exception.EGOExpConnLost) as e:
                    LOG.warn(six.text_type(e))
                    continue
                except Exception as e:
                    LOG.warning(_LW('Platform Resource Scheduler could not '
                               'update EGO resource decision for instance '
                               '(%(instanceid)s) due to: (%(reason)s)')
                              % {"instanceid": instid,
                                 "reason": six.text_type(e)})
                    continue
                if egohost:
                    flavor_id = self.inst_flv_idmap.get(instid)
                    if not flavor_id:
                        LOG.warning(_LW('Platform Resource Scheduler could '
                                   'not update resource info for host '
                                   '(%(host)s) due to: the instance '
                                   '(%(instid)s) does not have an '
                                   'allocation in the cache')
                                 % {"instid": instid,
                                    "host": egohost})
                        continue
                    src_inst_type_obj = flavors.get_flavor_by_flavor_id(
                            flavor_id)
                    src_inst_type = base_obj.obj_to_primitive(
                            src_inst_type_obj)
                    src_inst_props = (
                    ego_resreq_utils.generate_inst_property_from_inst_type(
                            inst, src_inst_type))
                    self.consume_inst_from_egohost(egohost,
                                                    src_inst_props, False)
                    self.consume_inst_from_egohost(egohost, inst, True)

        # 7. remove if the VM has a decision, but doesn't have
        #    (instance, decision)
        inst_decid_list = [dec_id for vm,
                            dec_id in self.inst_dec_map.items()]
        for decid in (set(self.decid_list) - set(inst_decid_list)):
            # Did not find related instance for this decision,
            # 1. check if it is an unqueued orphan decision from Heat
            #    stack, queue it, and bypass the release process
            # 2. release the decision which did not related to any
            #    instance. empty allocation remove later.
            # 3. remove alloc mode switch redundant server group allocation
            if decid not in self.orphan_decid_list:
                alloc_id = self.dec_allocid_map[decid]
                if (ego_constants.HEAT ==
                        ego_resreq_utils.get_alloc_type_from_allocid(
                            alloc_id, self.allocname_id_map)):
                    self.orphan_decid_list.append(decid)
                    continue
            try:
                self.ego_release_decision(decid)
            except (ego_exception.EGOExpConnLost,
                    ego_exception.EGOExpReleaseDecisionFailed) as e:
                LOG.error(_LE('Platform Resource Scheduler could not release '
                            'decision %(decid)s due to: (%(reason)s)')
                            % {'decid': decid, 'reason': six.text_type(e)})

        prefix_len = len(ego_constants.SERVER_GROUP + ego_constants.UNDERLINE)
        group_uuids = [str(group_obj.uuid) for group_obj in server_groups]
        for allocname, allocid in self.allocname_id_map.items():
            # 8. free an alloc who has no decision for two cases:
            #    1. this alloc is for standalone
            #    2. this alloc is for server group, but the server group has
            #       been deleted in the database
            if allocid not in self.dec_allocid_map.values():
                alloc_type = ego_resreq_utils.get_alloc_type_from_allocname(
                                                  allocname)
                if (alloc_type == ego_constants.STAND_ALONE or
                        (alloc_type == ego_constants.SERVER_GROUP and
                             allocname[prefix_len:] not in group_uuids) or
                             (not alloc_for_sg)):
                    self.ego_free_alloc(allocname)

        try:
            self._iterate_extension_managers("instance_sync_up", instance_list)
        except Exception as e:
            LOG.error(_LE('Platform Resource Scheduler failed to sync up '
                          'instances for EGO extension managers error: %s.')
                      % six.text_type(e))

        LOG.info(_LI('Finished syncing EGO allocations with instances from '
                     'the database.'))

    def sync_up_resource_groups(self):
        """
        Sync up EGO resource groups with all Nova host aggregates.
        Create the new resource group if there is a new host aggregate; delete
        resource group if there is no corresponding host aggregate.
        """

        # Retrieve all host aggregates and all resource groups to get their
        # difference
        try:
            res_groups_dict = self.ego_get_all_resource_groups()

            # Transfer the db aggregate id to string
            db_keys = [str(key) for key in
                self.aggregate_manager.aggregateid_metadatas.keys()]
            ego_keys = res_groups_dict.keys()
            # If only in the ego, it mean the aggregate is deleted, so
            # delete resource group from ego
            only_in_ego = set(ego_keys) - set(db_keys)
            # If only in db, it mean the aggregate is created, so create
            # resource group in ego
            only_in_db = set(db_keys) - set(ego_keys)

            for aggregate_id in only_in_db:
                res_req_str = ("select(%s)" %
                    ego_resreq_utils.generate_group_resreq(aggregate_id))
                try:
                    self.ego_create_resource_group(aggregate_id, res_req_str)
                except (ego_exception.EGOExpCreateResourceGroupFailure,
                       ego_exception.EGOExpConnLost) as exc:
                    LOG.error(_LE("Platform Resource Scheduler could not "
                                  "create new resource group %(group_name)r "
                                  "due to: %(reason)s")
                              % {'group_name': aggregate_id,
                                 'reason': six.text_type(exc)})
            for rg_name in only_in_ego:
                if rg_name in ego_constants.EGO_BUILTIN_RESGROUP:
                    continue
                try:
                    self.ego_delete_resource_group(rg_name)
                    LOG.debug("Deleted resource group '%s' since it no "
                              "longer exists in the OpenStack "
                              "infrastructure." % rg_name)
                except ego_exception.EGOExpDeleteResourceGroupFailure as exc:
                    LOG.error(_LE("Platform Resource Scheduler could not "
                                  "delete the resource group %(group_name)r "
                                  "due to: %(reason)s")
                              % {'group_name': rg_name,
                                 'reason': six.text_type(exc)})
        except (ego_exception.EGOExpGetResourceGroupFailure,
                ego_exception.EGOExpConnLost) as exc:
            LOG.error(_LE('Platform Resource Scheduler could not '
                          'retrieve all resource groups due to: %(reason)s')
                      % {'reason': six.text_type(exc)})
            return

    def ego_sync_aggregates_metadatas(self):
        """
        This function is used to make the host aggregates, host metadata
        and host RPI resources consistent
        """
        if self.need_aggregate or self.need_metadata:
            self.aggregate_manager.sync_aggregates_and_metadatas(
                self.need_metadata)

    def ego_sync_up_hosts(self):
        """
        This function is used to make hosts and RPI resources consistent
        """
        # 1. get all hosts in EGO resource list
        ego_hosts_list = []
        try:
            # only get RPI resources managed by OpenStack
            ego_hosts_list = [node['name'] for node in
                self.ego_get_resource_list(
                    'select(activeCompute==0 || activeCompute==1)')]
        except (ego_exception.EGOExpGetResourceInfoFailed,
                ego_exception.EGOExpConnLost) as exc:
            LOG.error(_LE('Platform Resource Scheduler could not retrieve '
                          'all RPI resources from EGO due to: '
                          '(%(reason)s)') % {'reason': six.text_type(exc)})

        # 2. get all latest RPI cstr according to DB and EGO
        extra_stats = {}
        # if HA the service is up, get hosts' HA and maintenance status
        # from ha service.  Otherwise, get hosts' maintenance status from
        # the persistent file store and discard HA status.
        is_ha_status_valid = False
        service_status = ego_resreq_utils.get_service_status_by_topic(
                                                  self.ctxt, 'hapolicy')
        if not service_status:
            # if HA related packages are not installed, i.e., the HA feature
            # is not used, the EGO scheduler should treat all compute nodes' HA
            # status and maintenance status as OK.
            status = hv_status.HypervisorStatus({'ha': {}, 'maintenance': {}})
        else:
            if service_status['is_up']:
                if not service_status['disabled']:
                    is_ha_status_valid = True
                try:
                    with eventlet.timeout.Timeout(
                      ego_constants.PRS_RPC_TIMEOUT, True):
                        status = hv_status.HypervisorStatus(
                            self.ha_rpc_api.get_hv_status(self.ctxt))
                except (oslo_exceptions.MessagingTimeout,
                        eventlet.Timeout) as e:
                    # if HA service is down, but its state has not been updated
                    # in the DB, get hosts' maintenance status from the
                    # persistent file and discard HA status.
                    LOG.warn(_LW("Could not retrieve hypervisor status within "
                                 "allotted time; caught error: %s") %
                             six.text_type(e))
                    status = ego_utils.get_compute_status()
                    is_ha_status_valid = False
            else:
                status = ego_utils.get_compute_status()

        host_aggregates = self.aggregate_manager.host_aggregates
        host_aggregates = host_aggregates if self.need_aggregate else {}
        for egohost in ego_hosts_list:
            extra_stat = {'running_vms':
                len(self.__get_instances_by_host(egohost, True))}
            extra_stats[egohost] = extra_stat
        host_attr_dict = self.ego_host_obj.get_all_ego_host(
                    self,
                    self.all_flavors,
                    host_aggregates,
                    self.aggregate_manager.host_metadatas,
                    ego_hosts_list,
                    extra_stats,
                    self.all_egohost_attrs,
                    is_ha_status_valid,
                    status)

        # 3. compare cstr between latest and stored in cache
        #    if not consistent, set latest cstr to EGO
        try:
            hosts_str = ''
            fin_hosts_str = ''
            hosts_count = 0
            for (egohost, attr) in host_attr_dict.items():
                cstr_old = (
                    self.all_egohost_attrs.get(egohost, {}).get('cstr', ''))
                if attr['cstr'] != cstr_old:
                    cstr_list = attr['cstr'].split(' ')
                    plugin_name = cstr_list[0]
                    version = cstr_list[1]
                    hosts_count += 1
                    hosts_str += ' ' + ' '.join(cstr_list[3:])
                    self.all_egohost_attrs[egohost] = attr

                # remove dirty attributes of the EGO host which is removed
                cstr_del = ego_resreq_utils.generate_res_state_and_attr_str(
                                egohost, ego_constants.HOST_RES_REMOVED)
                if attr['cstr'] == cstr_del:
                    self.all_egohost_attrs.pop(egohost)
            if hosts_count > 0:
                fin_hosts_str = (plugin_name + ' ' + version + ' '
                                 + str(hosts_count) + ' ' + hosts_str)
                self.ego_set_resource_info_str(fin_hosts_str)
        except (ego_exception.EGOExpSetResourceInfoFailed,
                ego_exception.EGOExpConnLost) as exc:
            LOG.error(_LE('Platform Resource Scheduler could not update the '
                        'EGO resource using XML string (%(cstr)s) '
                        'due to: (%(reason)s)')
                      % {"cstr": fin_hosts_str, "reason": six.text_type(exc)})

        # 4. update the map relationship from host to ego hosts
        self.host_egohost.clear()
        for egohost in self.all_egohost_attrs.keys():
            svc_host = self.all_egohost_attrs[egohost]['service']['host']
            self.host_egohost[svc_host].add(egohost)

        self._iterate_extension_managers("host_sync_up")

    def __ego_set_resource_info(self, ego_host, compute, service,
                               aggregate_ids, aggregate_metadata):
        # The extra_stat will be saved into node_stat to be sent as an
        # RPI resource.
        #
        # NOTE(jwcroppe): As temp fix for now, here "running_vms" is used
        # to overwrite the same name value within the compute_nodes table.
        extra_stat = {'running_vms':
                    len(self.__get_instances_by_host(ego_host, True))}
        metrics = self.all_egohost_attrs.get(ego_host, {}).get('metrics', {})
        stats = compute.get('stats', '{}')
        host_attrs = self.ego_host_obj.get_ego_host(
                        self,
                        ego_host,
                        host_metrics=metrics,
                        compute=compute,
                        metadatas=aggregate_metadata,
                        flavors=self.all_flavors[ego_host],
                        node_stats=stats,
                        host_aggregate_ids=aggregate_ids,
                        extra_stats=extra_stat,
                        service=service)
        try:
            self.ego_set_resource_info_str(host_attrs.get('cstr', ''))
            self.all_egohost_attrs[ego_host] = host_attrs
            if service and service.get('host'):
                self.host_egohost[service['host']].add(ego_host)
        except (ego_exception.EGOExpSetResourceInfoFailed,
                ego_exception.EGOExpConnLost) as exc:
            LOG.error(_LE('Platform Resource Scheduler could not update the '
                        'EGO resource using XML string (%(cstr)s) '
                        'due to: (%(reason)s)') %
                      {"cstr": host_attrs['cstr'],
                       "reason": six.text_type(exc)})

    def ego_set_resource_info(self, svc_host='', ego_host='',
                              compute=None, service=None):
        """
        Set nova-compute metrics as EGO RPI resource
        :param svc_host: host name of compute service, service['host']
        :param ego_host: name of ego RPI resource
        :param compute: compute node info
        :param service: service which the node belong to

        :returns: None
                  Throw exception if failed
        """
        compute = compute or {}
        service = service or {}
        if not svc_host:
            return

        host_aggregates = self.aggregate_manager.host_aggregates
        host_aggregates = host_aggregates if self.need_aggregate else {}
        aggids = host_aggregates.get(svc_host, [])
        metadatas = self.aggregate_manager.host_metadatas.get(svc_host, {})
        # consider multiple nodes in one host
        if svc_host and ego_host and compute and service:
            # consume instance from compute
            self.__ego_set_resource_info(ego_host, compute, service,
                                         aggids, metadatas)
        else:
            # two situations will step here:
            #    1) receive notification of aggregate change
            #    2) receive notification of service host add
            for ego_host in self.host_egohost.get(svc_host, set()):
                if ego_host in self.all_egohost_attrs:
                    host_attrs = self.all_egohost_attrs[ego_host]
                    self.__ego_set_resource_info(ego_host,
                                                 host_attrs['compute'],
                                                 host_attrs['service'],
                                                 aggids, metadatas)

    def ego_set_resource_info_str(self, cstr):
        self.ego_hdl.ego_set_resource_info_str(cstr)

    def ego_set_resource_plan(self, cstr):
        self.ego_hdl.ego_set_resource_plan(cstr)

    def ego_get_resource_list(self, cstr, state=None, is_attr_required=False):
        """
        Get EGO RPI resources list with resource requirement

        :param cstr: EGO resource requirement
        :param state: EGO resource state requirement
        :param is_attr_required: EGO resource's attributes are required to be
                                 included in the resource info if True
        :returns: [{'name': nodename1, 'state': 1, "attributes": {...}},
                   {'name': nodename2, 'state': 0, "attributes": {...}}, ...]
                  Throw exception if failed
        """
        # nodes format: [{'name': nodename1, 'state': 1, "attributes": {..}},
        #                {'name': nodename2, 'state': 0, "attributes": {}}, ..]
        nodes = self.ego_hdl.ego_get_resource_info(cstr, is_attr_required)
        if state is None:
            return nodes
        else:
            return [node for node in nodes if node['state'] == state]

    def ego_get_release_candidates(self, name, sub_demand_name, num_unit):
        pass

    def create_alloc_for_sg(self, name, sub_demands):
        """
        Create allocation for a server group

        :params name: Allocation name
        :params sub_demands: Allocation sub_demands
        """
        try:
            # the name of allocation for server group should start with "sg"
            name = ego_constants.UNDERLINE.join(
                    [ego_constants.SERVER_GROUP, name])
            # If sync first, then create notification comes. The allocation for
            # the server group has been created. Don't create twice for the SG.
            if name in self.allocname_id_map:
                return
            (allocId, consumer, hostlist) = (
                 self.ego_hdl.ego_create_alloc(name, sub_demands))
            self.__cache_update_alloc(allocId, {"alloc_name": name})
            LOG.debug('The system successfully allocated resources for '
                      'server group %(name)s.' % {'name': name})
        except ego_exception.EGOExpConnLost as e:
            LOG.error(_LE('Platform Resource Scheduler could not allocate '
                        'resource for server group: %(name)s with these '
                        'sub-demands: %(sub_demands)s due to (%(reason)s).')
                      % {'name': name, 'sub_demands': sub_demands,
                         'reason': six.text_type(e)})

    def ego_create_alloc_for_heat(self, context, name, inter_sub_policies,
                                  sub_demands):

        ego_sub_demands = []
        subdemands = ego_resreq_utils.convert_sub_demands_for_heat(
                                                    context, sub_demands)
        self._iterate_extension_managers('check_subdemands_for_heat', context,
                                         'create', subdemands)
        total_slots = 0
        for subdemand in subdemands:
            total_slots += subdemand['maxslots']
            subdemand['filter_properties']['request_spec'].update(
                {"operation_type": ego_constants.FILTER_WORKING_SCOPE_CREATE})
            resreq, extra_resreq = self.__go_through_filters(context,
                                subdemand.pop('filter_properties'))
            subdemand['resreq'] = resreq
            subdemand['extra_resreq'] = extra_resreq
            ego_sub_demands.append(subdemand)
        self.schedule_request_completed(context)
        inter_sub_policies = ego_resreq_utils.dict_key_value_to_str(
                                                    inter_sub_policies)
        # Request allocation; the name of allocation for Heat should start
        # with 'heat'
        (allocId, consumer, decision_list) = (
             self.ego_hdl.ego_create_alloc(name,
                                           ego_sub_demands,
                                           inter_sub_policies))
        if len(decision_list) != total_slots:
            self.ego_hdl.ego_free_alloc(allocId)
            errmsg = (_LE('Platform Resource Scheduler detected '
                        'insufficient resources for allocation %(name)s '
                        'with sub_demands %(sub_demands)s and '
                        'inter_sub_policies %(inter_sub_policies)s') %
                       {'name': name, 'sub_demands': ego_sub_demands,
                        'inter_sub_policies': inter_sub_policies})
            LOG.error(errmsg)
            raise ego_exception.EGOExpCreateAllocFailed(reason=errmsg)
        self.__cache_update_alloc(allocId, {"alloc_name": name})
        placement_decisions = []
        for decision in decision_list:
            self.__cache_update_dec(decision[1], {"alloc_id": allocId})
            dict = {}
            dict['resource_name'] = decision[0]
            dict['decision_id'] = decision[1]
            dict['demand_name'] = decision[2]
            dict['max_slots'] = decision[3]
            dict['resreq'] = decision[4]
            dict['resource_group'] = decision[5]
            placement_decisions.append(dict)
        return placement_decisions

    def __go_through_filters(self, context, filter_properties, index=0):
        if not filter_properties.get('context'):
            filter_properties.update({'context': context})
        self._iterate_extension_managers("handle_for_schedule_info",
                                       filter_properties)
        # Convert to EGO resreq
        ignore_filters = filter_properties.get('ignore_filters', False)

        # Convert to EGO resreq
        aggregateids = self.aggregate_manager.aggregateid_metadatas.keys()
        aggids = aggregateids if self.need_aggregate else []
        resreq_str = ''
        extra_select_str = ''
        spec = filter_properties.get('request_spec', {})
        scope = spec.get('operation_type')
        if not ignore_filters:
            resreq_str, extra_select_str = (
               self.ego_filter_driver.get_ego_filter_res_req(context,
                   filter_properties, self.all_egohost_attrs,
                   self.host_egohost,
                   scope,
                   aggids, index))
        else:
            if scope == ego_constants.FILTER_WORKING_SCOPE_CREATE:
                reserveon_str = ego_resreq_utils.get_reserveon_str(
                    filter_properties, aggids)
                consume_str = ego_resreq_utils.get_consume_str(
                        filter_properties)
                if filter_properties.get('alloc_for_sg') is False:
                    resreq_str = filter_properties['old_resreq']
                else:
                    resreq_str = "".join((reserveon_str, consume_str))

        force_host_list = filter_properties.get('force_hosts')
        force_node_list = filter_properties.get('force_nodes')
        ignore_host_list = filter_properties.get('ignore_hosts')
        extra_resreq = self.__get_extra_resreq(force_host_list,
                                               force_node_list,
                                               ignore_host_list,
                                               extra_select_str)
        return resreq_str, extra_resreq

    def get_placement_decision_for_heat(self, filter_properties):
        """
        Get one EGO unused decision to placement instance for Heat

        :param filter_properties: same with nova filters used
        :returns: return (allocId, decid, egohost) if succeed
                  Throw exception if failed
        """
        spec = filter_properties.get('request_spec', {})
        instance_uuid = str(spec.get('instance_id'))
        allocation_name = spec.get('allocation_name')
        sub_demand_name = spec.get('sub_demand_name')
        allocation_id = self.allocname_id_map.get(allocation_name)
        # get decision list from ego
        filter = {'allocation_name': allocation_name,
                  'sub_demand_name': sub_demand_name}
        filter = ego_resreq_utils.convert_filter(filter)
        decision_list = self.ego_hdl.ego_get_decisions_by_filter(filter)

        # get unused decision
        unused_decision = None
        if decision_list:
            for decision in decision_list:
                # self.decid_list: decisions from metadata
                if decision[1] not in self.inst_dec_map.values():
                    unused_decision = decision
                    break

        # update cache
        if not unused_decision:
            errmsg = (_LE('Platform Resource Scheduler could not find '
                        'an unused decision for instance %(instance_uuid)s '
                        'in stack %(name)s') %
                        {'instance_uuid': instance_uuid,
                         'name': allocation_name})
            LOG.error(errmsg)
            raise ego_exception.EGOExpGetDecisionsByFilterFailed(
                    filter=str(filter), reason=errmsg)
        flavorid = filter_properties.get('request_spec', {}).get(
                                             'instance_type')['flavorid']
        cache_dict = {"decision_id": unused_decision[1],
                      "alloc_name": allocation_name,
                      "ego_host": unused_decision[0],
                      "max_slot": unused_decision[3],
                      "resreq": unused_decision[4],
                      "flavorid": flavorid,
                      "alloc_id": allocation_id,
                      "res_group": unused_decision[5]}
        self.__cache_update_inst(instance_uuid, cache_dict)
        ego_utils.safe_remove(self.orphan_decid_list, unused_decision[1])
        # update decision_id/instance_uuid to database
        metadata = {'decision_id': unused_decision[1]}
        ego_resreq_utils.update_instance_system_metadata(instance_uuid,
                                                            metadata)

        return allocation_id, unused_decision[1], unused_decision[0]

    def ego_create_alloc(self, context, filter_properties, index=0):
        """
        Create EGO allocation when creating instance.
        For re-schedule, the EGOManager will free the old allocation
        and create a new one. The allocation id has changed, but the
        allocation name is unchanged.

        :param filter_properties: same with nova filters used
        :param index: The index of the instance when boot multi instances
                      using '--num-instances' option

        :returns: return (allocId, decid, egohost) if succeed
                  Throw exception if failed
        """
        # Get instance id, this will be set as EGO allocation name
        spec = filter_properties.get('request_spec', {})
        inst_prop = spec.get('instance_properties')
        instanceid = str(spec.get('instance_id'))
        flavorid = spec.get('instance_type')['flavorid']
        force_realloc_migrate = filter_properties.get("force_realloc_migrate",
                                                      False)

        resreq_str, extra_resreq = self.__go_through_filters(context,
            filter_properties, index)

        ts_resreq_str = (_('(%(resreq)s) and extra '
                           'resource requirement (%(extra_resreq)s).')
                        % {'resreq': resreq_str,
                           'extra_resreq': extra_resreq})
        self.ego_filter_driver.schedule_cache_update(filter_properties,
                                   {'resreq': ts_resreq_str})
        LOG.debug('The resource requirements for creation are: '
                  'creation resreq string is (%(resreq)s), '
                  'extra resreq string is (%(extra_resreq)s).'
                  % {'resreq': resreq_str,
                     'extra_resreq': extra_resreq})
        sub_demands = []
        inter_sub_policies = []
        sub_demand = {}
        sub_demand['name'] = instanceid
        sub_demand['resreq'] = resreq_str
        sub_demand['extra_resreq'] = extra_resreq
        sub_demand['maxslots'] = 1
        sub_demand['force_realloc_migrate'] = force_realloc_migrate
        sub_demands.append(sub_demand)
        # Ensure that only one single thread can create EGO allocation
        # and refersh EGO RPI resource at once
        with lockutils.lock('ego_create_alloc',
                            lock_file_prefix='prs-ego-create-alloc-'):
            # Request allocation
            (allocId, consumer, decisionlist) = (
                 self.ego_hdl.ego_create_alloc(instanceid,
                                               sub_demands,
                                               inter_sub_policies))
            if len(decisionlist) == 0:
                self.ego_hdl.ego_free_alloc(allocId)
                errmsg = (_LE('Platform Resource Scheduler detected '
                           'insufficient resources for resource '
                           'requirement (%(resreq)s) and extra resource '
                           'requirement (%(extra_resreq)s); refer '
                           'to /var/log/nova/nova-scheduler.log for '
                           'troubleshooting report details.')
                          % {'resreq': resreq_str,
                             'extra_resreq': extra_resreq})
                LOG.error(errmsg)
                raise ego_exception.EGOExpCreateAllocFailed(
                                                    reason=errmsg)
            self.__cache_update_alloc(allocId, {"alloc_name": instanceid})

            (egohost, decid, sub_demand_name, max_slots,
                    resreq, resource_group) = decisionlist.pop()
            LOG.debug("Platform Resource Scheduler successfully created "
                      "allocation using ID (%(allocId)s) and name "
                      "(%(name)s) on host (%(egohost)s)." %
                        {'allocId': allocId,
                         'name': instanceid,
                         'egohost': egohost})
            self.__cache_update_dec(decid, {"alloc_id": allocId})
            cache_dict = {"decision_id": decid, "alloc_name": instanceid,
                          "ego_host": egohost, "max_slot": max_slots,
                          "resreq": resreq_str, "flavorid": flavorid,
                          "alloc_id": allocId, "res_group": resource_group}
            self.__cache_update_inst(instanceid, cache_dict)
            metadata = {'decision_id': decid}
            ego_resreq_utils.update_instance_system_metadata(instanceid,
                                                                metadata)
            # update resource in cache and set into RPI to make sure
            # EGO can always get the latest resource service_status
            self.consume_inst_from_egohost(egohost, inst_prop, True)
            try:
                self._iterate_extension_managers("instance_create",
                                                 filter_properties)
            except Exception as ex:
                LOG.error(_LE("Platform Resource Scheduler Extension "
                              "operation failed."))
                raise ego_exception.EGOExtenOperationFailed(
                    reason=six.text_type(ex))

        return (allocId, decid, egohost)

    def __get_new_resreq(self, old_resreq_str, new_sub_resreq_str,
                         prefix, postfix):
        resreq_arr = old_resreq_str.split(prefix)
        resreq_after_consume = (
                old_resreq_str[old_resreq_str.find(prefix) + len(prefix):])
        resreq_post_fix = (
                resreq_after_consume[resreq_after_consume.find(postfix)
                                     + len(postfix):])
        return resreq_arr[0] + new_sub_resreq_str + resreq_post_fix

    def ego_update_decision(self, context, filter_properties):
        """
        Resize EGO allocation when creating instance.

        :param filter_properties: same with nova filters used
        :returns: return (allocId, decid, egohost) if succeed
                  Throw exception if failed
        """
        new_resreq_str = ''
        source_host = None
        old_flavorid = None
        inst_props = None
        src_inst_type = None
        dst_inst_type = None
        prefer_same_host = 1
        if (filter_properties.get('force_hosts') or
                filter_properties.get('ignore_hosts')):
            prefer_same_host = 0
        # check is revert-resize or resize
        revert_resize_flag = filter_properties.get("revert_resize", False)
        force_realloc_migrate = filter_properties.get("force_realloc_migrate",
                                                      False)
        retry_resize_flag = filter_properties.get('retry4create', False)
        resreq_str, extra_resreq = self.__go_through_filters(context,
            filter_properties)
        spec = filter_properties.get('request_spec', {})
        force_hosts = filter_properties.get('force_hosts')
        instance_id = str(spec.get('instance_id'))
        old_resreq_str = self.inst_resreq_map.get(instance_id)
        inst_props = spec.get('instance_properties', {})
        if not revert_resize_flag:
            source_host = self.inst_egohost_map[instance_id]
            dst_inst_type = spec.get('instance_type')
            flavorid = dst_inst_type['flavorid']
            if retry_resize_flag:
                old_flavorid = flavorid
                new_resreq_str = self.inst_resreq_map[instance_id]
            else:
                instance_type_id = inst_props.get('instance_type_id')
                old_flavorid = None
                if instance_type_id:
                    instance_type = (
                        objects.Flavor.get_by_id(self.ctxt, instance_type_id))
                    old_flavorid = instance_type['flavorid']
                operation_type = spec.get('operation_type')
                prefer_flavor = ((operation_type ==
                                  ego_constants.FILTER_WORKING_SCOPE_CREATE) or
                                 (operation_type ==
                                  ego_constants.FILTER_WORKING_SCOPE_RESIZE))
                new_consume_str = (
                    ego_resreq_utils.get_consume_str(
                        filter_properties, prefer_instance_type=prefer_flavor))
                new_resreq_str = self.__get_new_resreq(
                    old_resreq_str,
                    new_consume_str,
                    prefix=ego_constants.RES_REQ_CONSUME_PREFIX,
                    postfix=ego_constants.RES_REQ_CONSUME_POSTFIX)
        else:
            # this force host was set by notification
            source_host = self.inst_egohost_map[instance_id]
            old_consume_str = filter_properties.get('old_consume_str')
            flavorid = filter_properties.get('flavorid')
            old_flavorid = self.inst_flv_idmap[instance_id]
            if old_flavorid:
                src_inst_type_obj = flavors.get_flavor_by_flavor_id(
                        old_flavorid)
                src_inst_type = base_obj.obj_to_primitive(src_inst_type_obj)
            new_resreq_str = self.__get_new_resreq(
                old_resreq_str,
                old_consume_str,
                prefix=ego_constants.RES_REQ_CONSUME_PREFIX,
                postfix=ego_constants.RES_REQ_CONSUME_POSTFIX)
        if force_hosts and force_realloc_migrate:
            new_reserveon_str = None
            if self.need_aggregate and CONF.prs_sync_aggregate_reserveon:
                new_reserveon_str = self.__get_reserveon_based_on_host(
                                                          force_hosts[0])
            elif not self.need_aggregate:
                new_reserveon_str = (ego_constants.RES_REQ_RESERVEON_PREFIX
                                     + ego_constants.RESOURCE_GROUP_ALL +
                                     ego_constants.RES_REQ_RESERVEON_POSTFIX)
            if new_reserveon_str:
                new_resreq_str = self.__get_new_resreq(
                            new_resreq_str,
                            new_reserveon_str,
                            prefix=ego_constants.RES_REQ_RESERVEON_PREFIX,
                            postfix=ego_constants.RES_REQ_RESERVEON_POSTFIX)

        ts_resreq_str = (_('(%(resreq)s) and extra '
                           'resource requirement (%(extra_resreq)s).')
                        % {'resreq': new_resreq_str,
                           'extra_resreq': extra_resreq})
        self.ego_filter_driver.schedule_cache_update(filter_properties,
                                   {'resreq': ts_resreq_str})
        LOG.debug('The resource requirements for resize are: '
                  'resize resreq string is (%(resreq)s), '
                  'extra resreq string is (%(extra_resreq)s).'
                  % {'resreq': new_resreq_str,
                     'extra_resreq': extra_resreq})

        # set max_slot to 1 for PowerVM
        max_slot = filter_properties.get("max_slot", 1)
        htype = self.get_htype_from_hostname(source_host)
        if not self.support_modify_allocation_quota(htype, instance_id):
            max_slot = 1

        # Ensure that only one single thread can update EGO decision
        # and refersh EGO RPI resource at once
        with lockutils.lock('ego_update_decision',
                            lock_file_prefix='prs-ego-update-decision-'):
            (allocId, decid, dst_host, res_group) = (
                 self.__ego_update_decision(instance_id, new_resreq_str,
                                         extra_resreq,
                                         prefer_same_host,
                                         force_realloc_migrate,
                                         max_slot,
                                         retry_resize_flag))
            try:
                self._iterate_extension_managers("instance_update",
                                                  filter_properties)
            except Exception as e:
                LOG.error(_LE('Could not update allocation for %(vm_name)r '
                              '(%(uuid)s) due to: (%(reason)s).') %
                              {'uuid': instance_id,
                               'vm_name': inst_props.get('hostname'),
                               'reason': six.text_type(e)})
                if not force_realloc_migrate:
                    if instance_id in self.inst_allocid_map:
                        recover_extra_resreq = self.__get_extra_resreq(
                            force_node_list=[source_host, ])
                        self.__ego_update_decision(instance_id, old_resreq_str,
                                               recover_extra_resreq,
                                               0,
                                               True,
                                               max_slot,
                                               retry_resize_flag)
                    raise ego_exception.EGOExtenOperationFailed(
                        reason=six.text_type(e))
            # record scheduled host for the correct schedule next time
            if dst_host:
                cache_dict = {"ego_host": dst_host, "max_slot": max_slot,
                              "resreq": new_resreq_str, "flavorid": flavorid,
                              "res_group": res_group}
                filter_properties['old_resreq_str'] = old_resreq_str
                self.__cache_update_inst(instance_id, cache_dict)
                # update resource in cache and set into RPI to make sure
                # EGO can always get the latest resource status
                if not inst_props:
                    try:
                        inst_props = objects.Instance.get_by_uuid(
                                            self.ctxt, instance_id)
                    except Exception as ex:
                        LOG.warning(_LW("Instance with UUID %s cannot be "
                                        "found in the database; "
                                        "cannot update EGO RPI.")
                                 % instance_id)
                        LOG.exception(ex)
                        return (allocId, decid, dst_host)
                if not revert_resize_flag:
                    if source_host:
                        self.consume_inst_from_egohost(source_host, inst_props,
                                                       False,
                                                       filter_properties)
                    dst_inst_props = (
                        ego_resreq_utils.generate_inst_property_from_inst_type(
                                                    inst_props, dst_inst_type))
                    self.consume_inst_from_egohost(dst_host, dst_inst_props,
                                                   True, filter_properties)
                else:
                    # for revert resize case: inst_props denotes the original
                    # instance properties before resize operation.
                    src_inst_props = (
                        ego_resreq_utils.generate_inst_property_from_inst_type(
                                                    inst_props, src_inst_type))
                    if source_host:
                        self.consume_inst_from_egohost(source_host,
                                                       src_inst_props, False,
                                                       filter_properties)
                    self.consume_inst_from_egohost(dst_host, inst_props, True,
                                                   filter_properties)
        return (allocId, decid, dst_host)

    def __ego_update_decision(self, instanceid, resreq, extra_resreq,
                              prefer_same_host=0, force_realloc_migrate=False,
                              max_slot=1, retry_resize_flag=False,
                              warning_if_no_resource_group=False):
        """Resize EGO allocation when resize instance."""
        decid = self.inst_dec_map.get(instanceid)
        if not decid:
            LOG.error(_LE("Cannot update decision due to: decision for "
                          "instance (%s) cannot be found")
                      % instanceid)
            raise ego_exception.DecNotFoundByInst(uuid=instanceid)
        allocid = self.__get_allocid_from_instid(instanceid)
        if not allocid:
            errmsg = (_LW('Instance (%s) is not allocated as a resource '
                       'to a cluster. This instance may be a '
                       'stand-alone instance.') % instanceid)
            LOG.warning(errmsg)
            raise ego_exception.EGOExpNoAllocID(inst_id=instanceid)
        return self.ego_hdl.ego_update_decision(decid, allocid, resreq,
                                                extra_resreq, prefer_same_host,
                                                force_realloc_migrate,
                                                max_slot, retry_resize_flag,
                                                warning_if_no_resource_group)

    def ego_resize_alloc_for_heat(self, context, name, sub_demands):
        ego_sub_demands = []
        subdemands = ego_resreq_utils.convert_sub_demands_for_heat(
                                                    context, sub_demands)
        self._iterate_extension_managers('check_subdemands_for_heat', context,
                                         'resize', subdemands)
        for subdemand in subdemands:
            is_first_create = True
            filters = {'allocation_name': name,
                       'sub_demand_name': subdemand['name']}
            filters = ego_resreq_utils.convert_filter(filters)
            decisionlist = self.ego_hdl.ego_get_decisions_by_filter(filters)
            for decision in decisionlist:
                if decision[1] in self.inst_dec_map.values():
                    is_first_create = False
                    break
            if decisionlist and is_first_create:
                continue
            subdemand['filter_properties']['request_spec'].update(
                {"operation_type": ego_constants.FILTER_WORKING_SCOPE_CREATE})
            resreq, extra_resreq = self.__go_through_filters(context,
                                subdemand.pop('filter_properties'))
            subdemand['resreq'] = resreq
            subdemand['extra_resreq'] = extra_resreq
            ego_sub_demands.append(subdemand)
        self.schedule_request_completed(context)

        # skip if called from create nested stack in Heat-created process
        if not ego_sub_demands:
            return []
        allocid = self.allocname_id_map.get(name)
        (retval, allocId, consumer, decision_list, clientname) = (
                self.ego_hdl.ego_resize_alloc(allocid, ego_sub_demands))

        if len(decision_list) == 0:
            errmsg = (_LE('Platform Resource Scheduler detected '
                        'insufficient resources for allocation %(name)s '
                        'with sub_demands %(sub_demands)s') %
                       {'name': name, 'sub_demands': ego_sub_demands})
            LOG.error(errmsg)
            raise ego_exception.EGOExpResizeAllocFailed(reason=errmsg)
        placement_decisions = []
        for decision in decision_list:
            self.__cache_update_dec(decision[1], {"alloc_id": allocId})
            dict = {}
            dict['resource_name'] = decision[0]
            dict['decision_id'] = decision[1]
            dict['demand_name'] = decision[2]
            dict['max_slots'] = decision[3]
            dict['resreq'] = decision[4]
            dict['resource_group'] = decision[5]
            placement_decisions.append(dict)
        return placement_decisions

    def resize_alloc_for_sg(self, context, filter_properties):
        group_uuid = filter_properties.get('group_uuid')
        alloc_name = ego_constants.UNDERLINE.join(
                            [ego_constants.SERVER_GROUP, group_uuid])
        alloc_id = self.allocname_id_map[alloc_name]
        clear_policy = filter_properties.get('clear_policy', False)
        if clear_policy:
            sub_demand = {'name': str(group_uuid),
                          'resreq': '',
                          'extra_resreq': '',
                          'maxslots': 0,
                          'force_realloc_migrate': False,
                          'clear_policy': True}
            self.ego_hdl.ego_resize_alloc(alloc_id, [sub_demand])
            return

        inst_uuid = filter_properties.get('uuid')
        old_resreq = filter_properties.get('old_resreq')
        if not inst_uuid:
            inst_uuid = filter_properties.get('request_spec', {}).get(
                                                        'instance_id')
        if old_resreq:
            ego_resreq = old_resreq
            extra_resreq = ''
        else:
            (ego_resreq,
            extra_resreq) = self.__go_through_filters(context,
                                                      filter_properties)
        force_realloc_migrate = filter_properties.get("force_realloc_migrate",
                                                      False)
        sub_demand = {'name': str(group_uuid),
                      'resreq': ego_resreq,
                      'extra_resreq': extra_resreq,
                      'maxslots': 1,
                      'force_realloc_migrate': force_realloc_migrate}
        errmsg = (_LE('Platform Resource Scheduler could not resize resource '
                      'allocation %(alloc_id)s with sub_demands: '
                      '(%(sub_demands)s); refer '
                      'to /var/log/nova/nova-scheduler.log for '
                      'troubleshooting report details.')
                  % {'alloc_id': alloc_id,
                     'sub_demands': [sub_demand]})
        # Ensure that only one single thread can resize EGO allocation
        # and refersh EGO RPI resource at once
        with lockutils.lock('resize_alloc_for_sg',
                            lock_file_prefix='prs-resize_alloc_for_sg-'):
            retinfo = self.ego_hdl.ego_resize_alloc(alloc_id, [sub_demand])
            (retval, allocId, consumer, decision_list,
                     clientname) = retinfo
            if len(decision_list) == 0:
                LOG.error(errmsg)
                raise ego_exception.EGOExpResizeAllocFailed(reason=errmsg)

            (egohost, decid, sub_demand, max_slots, resreq,
                    resource_group) = decision_list.pop()
            metadata = {'decision_id': decid}
            ego_resreq_utils.update_instance_system_metadata(inst_uuid,
                                                                metadata)
            self.__cache_update_dec(decid, {"alloc_id": allocId})
            flavorid = filter_properties.get('request_spec', {}).get(
                                                 'instance_type')['flavorid']
            cache_dict = {"decision_id": decid, "alloc_name": alloc_name,
                          "ego_host": egohost, "max_slot": max_slots,
                          "resreq": ego_resreq, "flavorid": flavorid,
                          "alloc_id": allocId, "res_group": resource_group}
            self.__cache_update_inst(inst_uuid, cache_dict)
            # instance_type is either in request_spec or filter_properties
            inst_type = filter_properties['request_spec']['instance_type']
            inst_props = (
                filter_properties['request_spec']['instance_properties'])
            inst_pro = ego_resreq_utils.generate_inst_property_from_inst_type(
                                             inst_props, inst_type)
            self.consume_inst_from_egohost(egohost, inst_pro, True)
            self.decid_list.append(decid)
        return (alloc_id, decid, egohost)

    @lockutils.synchronized('ego_free_alloc', 'prs-ego-free-alloc-')
    def ego_free_alloc(self, alloc_name, inst_props=None, standalone=False):
        """
        Free an allocation.

        :param alloc_name: allocation name
        :param inst_props: instance properties
        :param standalone: allocation for standalone instance if True
        :returns: None
                  Throw exception if failed
        """
        inst_props = inst_props or {}
        allocid = self.allocname_id_map.get(alloc_name)
        if not allocid:
            return
        instanceid = None
        if standalone:
            instanceid = alloc_name
        try:
            self.ego_hdl.ego_free_alloc(allocid)
            self.__cache_pop_alloc(alloc_name)
            # for standalone case, the instance and decision cache should be
            # updated too
            if instanceid:
                dec_id = self.inst_dec_map[instanceid]
                self.__cache_pop_dec(dec_id)
                egohost = self.inst_egohost_map[instanceid]
                LOG.debug("Platform Resource Scheduler successfully freed "
                          "allocation %(allocid)s for instance "
                          "%(instanceid)s." %
                            {'allocid': allocid,
                            'instanceid': instanceid})
                # update resource in cache and set send to RPI to make sure
                # EGO always has the latest resource status
                if not inst_props:
                    try:
                        inst = objects.Instance.get_by_uuid(self.ctxt,
                                                            instanceid)
                        inst_props = base_obj.obj_to_primitive(inst)
                    except Exception as ex:
                        LOG.warning(_LW("Instance with UUID %s cannot be "
                                   "found in the database; cannot update "
                                   "EGO RPI.") % instanceid)
                        LOG.exception(ex)
                        return
                self.consume_inst_from_egohost(egohost, inst_props, False)
                self.__cache_pop_inst(instanceid)
                self._iterate_extension_managers("instance_delete", instanceid)
        except Exception as e:
            if standalone:
                msg = (_LE('Platform Resource Scheduler could not release EGO '
                           'resource allocation for instance (%(instanceid)s) '
                           'due to: (%(reason)s)')
                       % {'instanceid': instanceid,
                          'reason': six.text_type(e)})
            else:
                msg = (_LE('Platform Resource Scheduler could not release EGO '
                         'resource allocation %(id)r (%(name)s) '
                         'due to: (%(reason)s)')
                        % {"id": allocid,
                           "name": alloc_name,
                           "reason": six.text_type(e)})
            LOG.error(msg)

    def ego_migrate_decision(self, context, instanceid, targethost,
                             force_realloc_migrate=False,
                             filter_properties=None):
        """
        Interface for Live Migration and HA.

        :param instanceid: allocation name, for OpenStack Resource
                           Optimization, it is the id of the instance
        :param targethost: target host where to migarate the host, if
                           set as '', EGO will select a host as target
                           host, the target host must satisfy the original
                           instance resource requirement.
        :param force_realloc_migrate: force realloc or migration without
                                      considering resource requirement
                                      and resource quota
        :param filter_properties: Instance relate specification, it includes
                                  the folowing informations:
                                  1. instance properties
                                  2. flavor
                                  3. operation type
        :returns: return target host of the migration, if target host
                  was specified, return target host as specified; if
                  target host was not specified, EGO select a target
                  host and return, the return value is a string value
                  Throw exception if failed
        """
        extra_resreq = ''
        resreq = ''
        filter_properties = filter_properties or {}
        allocid = self.__get_allocid_from_instid(instanceid)
        if allocid is None:
            LOG.warning(_LW('Instance (%s) is not allocated '
                        'as a resource to a cluster. This '
                        'instance may be a stand-alone '
                        'instance, and the EGO manager cannot '
                        'find a suitable migration host '
                        'for it.') % instanceid)
            raise ego_exception.EGOExpNoAllocID(inst_id=instanceid)

        source_host = self.inst_egohost_map.get(instanceid)
        resreq = self.inst_resreq_map[instanceid]
        # If an instance has allocid, then it must have decision id
        decid = self.inst_dec_map.get(instanceid)

        if targethost:
            force_host_list = filter_properties.get('force_hosts', [])
            force_host_list.append(targethost)
            filter_properties['force_hosts'] = force_host_list
        resreq_str, extra_resreq = self.__go_through_filters(context,
            filter_properties)
        ts_resreq_str = (_('(%(resreq)s) and extra '
                           'resource requirement (%(extra_resreq)s).')
                        % {'resreq': resreq,
                           'extra_resreq': extra_resreq})
        self.ego_filter_driver.schedule_cache_update(filter_properties,
                                   {'resreq': ts_resreq_str})
        LOG.debug('The resource requirements for migration are: '
                  'migration resreq string is (%(resreq)s), '
                  'extra resreq string is (%(extra_resreq)s).'
                  % {'resreq': resreq,
                     'extra_resreq': extra_resreq})
        # Ensure that only a single thread can migrate EGO decision
        # and refersh EGO RPI resource at once
        with lockutils.lock('ego_migrate_decision',
                            lock_file_prefix='prs-ego-migrate-decision-'):
            self.ego_hdl.ego_migrate_decision(decid, allocid,
                                              force_realloc_migrate,
                                              extra_resreq)
            # Return alloc host after migration
            filters = {'decision_id': decid}
            filters = ego_resreq_utils.convert_filter(filters)
            decisionlist = self.ego_hdl.ego_get_decisions_by_filter(filters)
            (new_target_host, decid, sub_demand, max_slots, resreq,
                    resource_group) = decisionlist.pop()

            cache_dict = {"ego_host": new_target_host}
            self.__cache_update_inst(instanceid, cache_dict)

            LOG.debug("Platform Resource Scheduler successfully migrated "
                      "allocation ID (%(allocid)s) to host "
                      "(%(new_target_host)s)." %
                        {'allocid': allocid,
                         'new_target_host': new_target_host})
            # update resource in cache and set into RPI which ensures that
            # EGO can always get the latest resource status
            inst_props = filter_properties.get('request_spec', {}).get(
                                                    'instance_properties')
            if not inst_props:
                try:
                    inst_props = objects.Instance.get_by_uuid(
                                        self.ctxt, instanceid)
                except Exception as e:
                    errmsg = (_LE('Failed to get instance properties from the '
                                  'database due to: (%(reason)s)') %
                                 {'reason': six.text_type(e)})
                    LOG.error(errmsg)
                    raise
            if inst_props:
                self.consume_inst_from_egohost(source_host,
                                               inst_props, False,
                                               filter_properties)
                self.consume_inst_from_egohost(new_target_host,
                                               inst_props, True,
                                               filter_properties)
        return new_target_host

    def ego_get_possible_hosts(self, context, decid, group_list, order_list,
                               filter_properties):
        """
        Get possible hosts for an instance.

        :param decid: decision id, for OpenStack Resource Optimization,
                           it is the decision id of the instance
        :param group_list: multiple groups which should be reserveon
                               for the instance
        :param order_list: the order list for output possible hosts
        :param filter_properties: filter properties for filters
        :returns: return possible host list which have the same
                  resource requirement with current instance, the
                  return value is in the fomant of ['host1', 'host2', ...]
                  Throw exception if failed
        """
        if not decid:
            LOG.warning(_LW('Instance decision: (%s) is not allocated '
                          'as a resource to a cluster. This '
                          'instance may be a stand-alone '
                          'instance and the EGO manager cannot '
                          'find a suitable migration host for it.')
                        % decid)
            # Resource optimization will not control standalone
            # instances which did not have an EGO allocation
            raise ego_exception.EGOExpGetPossibleHostsFailed(decid=decid)
        else:
            instance_id = filter_properties["request_spec"]["instance_id"]
            source_host = self.inst_egohost_map.get(instance_id)
            resreq = self.inst_resreq_map[instance_id]
            reserveon_str = ego_resreq_utils.get_resreq_str(group_list,
                                                            'reserveon')
            order_str = ego_resreq_utils.get_resreq_str(order_list,
                                                        'order')
            if CONF.policy_per_host_aggregate:
                resreq = re.sub('reserveon\(.*?\)',
                            'reserveon(' + str(group_list[0]) + ')', resreq)

            resreq_str, extra_resreq = self.__go_through_filters(context,
                filter_properties)

            select_str = ego_resreq_utils.get_criteria_str_by_resreq(
                                               extra_resreq, 'select')

            ts_resreq_str = (_('(%(resreq)s) and extra '
                               'resource requirement (%(extra_resreq)s).')
                            % {'resreq': resreq,
                               'extra_resreq': extra_resreq})

            self.ego_filter_driver.schedule_cache_update(filter_properties,
                                       {'resreq': ts_resreq_str,
                                        'source_host': source_host})

            return self.ego_hdl.ego_get_possible_hosts(
                decid, reserveon_str + order_str + select_str)

    def __update_vm_alloc_of_egohost(self, ego_host, host_aggregate_ids):
        """
        Update reserveon info of VMs which belong to those hosts

        :param ego_host: ego host name equals RES_NAME
        :param host_aggregate_ids: id of all host aggregates which
                                  contain the host
        """
        new_reserveon_str = ego_constants.RESOURCE_GROUP_ALL
        if host_aggregate_ids:
            new_reserveon_str = ' '.join(
                    [str(ha_id) for ha_id in host_aggregate_ids])
        new_reserveon_str = (ego_constants.RES_REQ_RESERVEON_PREFIX +
                             new_reserveon_str +
                             ego_constants.RES_REQ_RESERVEON_POSTFIX)

        inst_uuids = self.__get_instances_by_host(ego_host)
        for uuid in inst_uuids:
            # only process instance which has an allocation
            if uuid not in self.inst_allocid_map:
                continue
            max_slot = self.inst_maxslots_map.get(uuid, 0)
            old_res_req = self.inst_resreq_map.get(uuid)
            if old_res_req:
                new_res_req = self.__get_new_resreq(
                    old_res_req,
                    new_reserveon_str,
                    prefix=ego_constants.RES_REQ_RESERVEON_PREFIX,
                    postfix=ego_constants.RES_REQ_RESERVEON_POSTFIX)
                # Do resize alloc regardless of whether new_res_req equals
                # old_res_req -- in case allocation holds reserveon(group_all),
                # but the 'Resource Group' has changed due to aggregates
                # of its host changed.
                extra_resreq = self.__get_extra_resreq(
                                    force_node_list=[ego_host])
                (allocId, decid, egohost, res_group) = (
                    self.__ego_update_decision(uuid,
                                               new_res_req,
                                               extra_resreq,
                                               prefer_same_host=1,
                                               force_realloc_migrate=True,
                                               max_slot=max_slot
                                               ))
                cache_dict = {"resreq": new_res_req,
                              "res_group": res_group}
                self.__cache_update_inst(uuid, cache_dict)

    def aggregate_create(self, payload):
        # update cache in aggregate manager after aggregate was created
        if not self.need_aggregate and not self.need_metadata:
            return
        self.aggregate_manager.aggregate_create(payload)

    def aggregate_delete(self, payload):
        # update cache in aggregate manager after aggregate was deleted
        if not self.need_aggregate and not self.need_metadata:
            return
        self.aggregate_manager.aggregate_delete(payload)

    def aggregate_set_metadata(self, payload):
        # update cache in aggregate manager and RPI for host who belong
        # to this aggregate after aggregate metadata was changed
        if not self.need_metadata:
            return
        aggregate_id = payload['aggregate_id']
        self.aggregate_manager.aggregate_set_metadata(payload)
        for svc_host, aggids in self.aggregate_manager.host_aggregates.items():
            if aggregate_id in aggids:
                self.ego_set_resource_info(svc_host)

    def aggregate_host_add(self, svc_host, aggregate_id):
        """
        Update aggregate id for specified host and change allocation of all
        VMs running on this host.
        """
        if not self.need_aggregate and not self.need_metadata:
            return
        self.aggregate_manager.aggregate_host_add(svc_host, aggregate_id,
                                                  self.need_metadata)
        self.ego_set_resource_info(svc_host)

        if self.need_aggregate and CONF.prs_sync_aggregate_reserveon:
            self.__update_change_host_with_vms(svc_host)

    def aggregate_host_remove(self, svc_host, aggregate_id):
        """
        Update aggregate id for specified host and change allocation of all
        VMs running on this host.
        """
        if not self.need_aggregate and not self.need_metadata:
            return
        self.aggregate_manager.aggregate_host_remove(svc_host, aggregate_id,
                                                   self.need_metadata)
        self.ego_set_resource_info(svc_host)

        if self.need_aggregate and CONF.prs_sync_aggregate_reserveon:
            self.__update_change_host_with_vms(svc_host)

    def __update_change_host_with_vms(self, svc_host='', ego_host=''):
        """
        :param svc_host: host of service, service['host']
        :param ego_host: EGO host name equals RES_NAME
        """
        if not svc_host:
            return
        agg_ids = self.aggregate_manager.host_aggregates.get(svc_host)
        if ego_host:
            self.__update_vm_alloc_of_egohost(ego_host, agg_ids)
        else:
            for ego_host in self.host_egohost.get(svc_host, set()):
                self.__update_vm_alloc_of_egohost(ego_host, agg_ids)

    def sync_aggregate_reserveon_for_inst(self, instance_id):
        """
        Update reserveon info based on the aggregates which the specified
        instance's host belongs to.

        :param instance_id: nova instance uuid
        """
        if not (self.need_aggregate and CONF.prs_sync_aggregate_reserveon):
            return
        if not instance_id or instance_id not in self.inst_resreq_map:
            return
        ego_host = self.inst_egohost_map.get(instance_id)
        new_reserveon_str = self.__get_reserveon_based_on_host(ego_host)
        if not new_reserveon_str:
            return
        old_res_req = self.inst_resreq_map[instance_id]
        old_reserveon_str = (
           re.match(r"reserveon\(.*?\)", old_res_req).group(0))
        if new_reserveon_str != old_reserveon_str:
            max_slot = self.inst_maxslots_map.get(instance_id, 0)
            new_res_req = self.__get_new_resreq(
                    old_res_req,
                    new_reserveon_str,
                    prefix=ego_constants.RES_REQ_RESERVEON_PREFIX,
                    postfix=ego_constants.RES_REQ_RESERVEON_POSTFIX)
            extra_resreq = self.__get_extra_resreq(
                                    force_node_list=[ego_host])

            (allocId, decid, egohost, res_group) = (
                self.__ego_update_decision(instance_id,
                                           new_res_req,
                                           extra_resreq,
                                           prefer_same_host=1,
                                           force_realloc_migrate=True,
                                           max_slot=max_slot))
            cache_dict = {"resreq": new_res_req,
                          "res_group": res_group}
            self.__cache_update_inst(instance_id, cache_dict)

    def __get_reserveon_based_on_host(self, ego_host):
        for host, egohosts in six.iteritems(self.host_egohost):
            if ego_host in egohosts:
                svc_host = host
                break
        else:
            return
        new_reserveon = ego_constants.RESOURCE_GROUP_ALL
        agg_ids = self.aggregate_manager.host_aggregates.get(svc_host)
        if agg_ids:
            new_reserveon = ' '.join(
                          [str(agg_id) for agg_id in agg_ids])
        return (ego_constants.RES_REQ_RESERVEON_PREFIX
                + new_reserveon
                + ego_constants.RES_REQ_RESERVEON_POSTFIX)

    def __conn_recover_in_thread(self, client_name):
        while True:
            eventlet.sleep(1)
            self.__ego_check_connection(client_name)

    def ego_create_resource_group(self, group_name, resreq):
        self.ego_hdl.ego_create_resource_group(group_name, resreq)

    def ego_delete_resource_group(self, group_name):
        self.ego_hdl.ego_delete_resource_group(group_name)

    def ego_get_all_resource_groups(self):
        return self.ego_hdl.ego_get_all_resource_groups()

    def __update_metadata_flag(self):
        """Update the metadata flag"""
        for filter_name in CONF.scheduler_default_filters:
            if filter_name in ['AggregateInstanceExtraSpecsFilter',
                               'AggregateTypeAffinityFilter',
                               'AggregateMultiTenancyIsolation',
                               'AggregateCoreFilter',
                               'AggregateRamFilter',
                               'AvailabilityZoneFilter',
                               'AggregateDiskFilter']:
                self.need_metadata = True

    def create_alloc_for_existing_inst(self, payload, **kw):
        """
        Create an allocation for an existing instance (e.g., used for PowerVC's
        "manage existing VM" functionality).

        :param payload: The payload from which the allocation will be built
        """
        instance_id = (payload.get('uuid') or payload.get('instance_id') or
                       payload.get('args', {}).get('instance', {}).get('uuid'))

        vm_state = payload.get('vm_state') or payload.get('state')
        # if there is no state or the state is shelved-offloaded (which means
        # the resources are no longer on the hypervisor), then don't create
        # an allocation
        if not vm_state or vm_state == vm_states.SHELVED_OFFLOADED:
            return
        svc_host = payload.get('host')
        # if there is no host specified or the VM was never launched, then
        # we won't create an allocation (since we don't know what resource
        # to associate it with)
        if not (svc_host and payload.get('launched_at')):
            return

        alloc_for_sg = kw.get('alloc_for_sg')
        if instance_id not in self.inst_allocid_map or alloc_for_sg is False:
            filter_properties = self.__gen_filter_props(payload)
            filter_properties["force_realloc_migrate"] = True
            filter_properties["old_resreq"] = kw.get('old_resreq')
            filter_properties["alloc_for_sg"] = alloc_for_sg
            (alloc_id, alloc_decid, alloc_host) = (
                self.ego_create_alloc(self.ctxt, filter_properties))
            self.sync_aggregate_reserveon_for_inst(instance_id)

    def consume_inst_from_egohost(self, ego_host, inst,
                                  consume_free_flag,
                                  filter_properties=None):
        """
        Incrementally update compute stats from an instance and then
        send the latest information to EGO's RPI attributes.

        :param ego_host: the name of ego host
        :param inst: instance property from resource request
        :param consume_free_flag:
                    True: consume an instance from host
                    False: free an instance from host
        """
        filter_properties = filter_properties or {}
        if (ego_host not in self.all_egohost_attrs):
            LOG.warning(_LW('Compute host (%s) could not be found in cache; '
                            'it may have been deleted.') % ego_host)
            return

        if ('root_gb' not in inst or 'ephemeral_gb' not in inst or
                'memory_mb' not in inst or 'vcpus' not in inst):
            LOG.warning(_LW('Instance properties (%(inst)s) with uuid '
                       '(%(uuid)s) is invalid for consuming from host '
                       '(%(host)s).') %
                     {'host': ego_host, 'inst': inst,
                      'uuid': inst.get('uuid', '')})
            return

        disk_gb = int(inst['root_gb']) + int(inst['ephemeral_gb'])
        ram_mb = int(inst['memory_mb'])
        vcpus = int(inst['vcpus'])

        # 1.update compute node
        compute = self.all_egohost_attrs[ego_host]['compute']
        if consume_free_flag:
            compute['vcpus_used'] += vcpus
            compute['memory_mb_used'] += ram_mb
            compute['local_gb_used'] += disk_gb
            if compute.get('disk_available_least'):
                compute['disk_available_least'] -= disk_gb
        else:
            compute['vcpus_used'] = max(compute['vcpus_used'] - vcpus, 0)
            compute['memory_mb_used'] = max(
                compute['memory_mb_used'] - ram_mb, 0)
            compute['local_gb_used'] = max(
                compute['local_gb_used'] - disk_gb, 0)
            if compute.get('disk_available_least'):
                compute['disk_available_least'] = min(
                    compute['disk_available_least'] + disk_gb,
                    compute['local_gb'])
        compute['free_ram_mb'] = (compute['memory_mb'] -
                                  compute['memory_mb_used'])
        compute['free_disk_gb'] = (compute['local_gb'] -
                                   compute['local_gb_used'])

        # 2.update compute node stats
        consume_io_flag = False
        vm_state = inst.get('vm_state')
        task_state = inst.get('task_state')
        # NOTE(zhgaoxa): io_workload is not correct for 'nova delete' case;
        # can't judge if should decrise io workload if delete an instance
        if vm_state == vm_states.BUILDING or task_state in [
                task_states.RESIZE_MIGRATING, task_states.REBUILDING,
                task_states.RESIZE_PREP, task_states.IMAGE_SNAPSHOT,
                task_states.IMAGE_BACKUP]:
            consume_io_flag = True

        vm_error_flag = False
        if vm_state == vm_states.ERROR:
            vm_error_flag = True

        stats_map = compute.get('stats') or {}
        # hypervisor ha/maintenance status reported by HA service
        if filter_properties.get(ego_constants.HA_MAINTENANCE_STATUS):
            hm_status = hv_status.HypervisorStatus(
                    filter_properties[ego_constants.HA_MAINTENANCE_STATUS])
            nodename = compute['hypervisor_hostname']
            ha_status = hm_status.get_ha(nodename)
            mt_status = hm_status.get_maintenance(nodename)
            stats_map.update({'ha_status': ha_status})
            stats_map.update({'maintenance_status': mt_status})

        if 'num_instances' not in stats_map:
            stats_map['num_instances'] = 0
        if 'io_workload' not in stats_map:
            stats_map['io_workload'] = 0
        if 'num_vm_error' not in stats_map:
            stats_map['num_vm_error'] = 0

        num_instances = int(stats_map['num_instances'])
        if consume_free_flag:
            stats_map['num_instances'] = num_instances + 1
        else:
            stats_map['num_instances'] = num_instances - 1

        if consume_io_flag:
            io_workload = int(stats_map['io_workload'])
            if consume_free_flag:
                stats_map['io_workload'] = io_workload + 1
            else:
                stats_map['io_workload'] = io_workload - 1

        if vm_error_flag:
            num_vm_error = int(stats_map['num_vm_error'])
            if consume_free_flag:
                stats_map['num_vm_error'] = num_vm_error + 1
            else:
                stats_map['num_vm_error'] = num_vm_error - 1

        self.all_egohost_attrs[ego_host]['compute']['stats'] = stats_map
        # 3.set latest compute info into the respective EGO RPI attributes
        compute = self.all_egohost_attrs[ego_host]['compute']
        service = self.all_egohost_attrs[ego_host]['service']
        inst_uuid = inst.get('uuid')
        resreq_str = self.inst_resreq_map[inst_uuid]
        old_resreq_str = filter_properties.get('old_resreq_str')
        self._iterate_extension_managers("consume_inst_from_egohost",
                                         inst_uuid, ego_host,
                                         resreq_str, old_resreq_str,
                                         compute, consume_free_flag)
        self.ego_set_resource_info(svc_host=service['host'], ego_host=ego_host,
                                   compute=compute, service=service)

    def ego_update_resource_attr(self, node_name, attr_dict=None):
        """
        Update EGO resource's attribute value.

        :param node_name: the name of compute node
        :param attr_dict: the attribute and value map would be updated to
        """
        attr_dict = attr_dict or {}
        ego_host = ego_resreq_utils.get_rpi_res_name(node_name)
        state = None
        try:
            slct_str = "select('%s')" % ego_host
            ego_hosts_info = self.ego_hdl.ego_get_resource_info(slct_str)
            for ego_host_info in ego_hosts_info:
                if ego_host_info['name'] == ego_host:
                    state = ego_host_info['state']
                    break
            cstr = ego_resreq_utils.generate_res_state_and_attr_str(ego_host,
                                                    state, attr_dict)
            self.ego_set_resource_info_str(cstr)
        except (ego_exception.EGOExpGetResourceInfoFailed,
                ego_exception.EGOExpSetResourceInfoFailed) as exc:
            LOG.error(_LE('Platform Resource Scheduler could not update the '
                          'EGO resource %(ego_host)s due to: (%(reason)s)')
                      % {'reason': six.text_type(exc), 'ego_host': ego_host})
            raise

    def ego_sync_up_server_group(self):
        server_groups = objects.InstanceGroupList.get_all(self.ctxt)
        for group in server_groups:
            alloc_name = ego_constants.UNDERLINE.join(
                        [ego_constants.SERVER_GROUP, group.uuid])
            if alloc_name not in self.allocname_id_map:
                # legacy policy in group, filters is skiped,
                # treat as standalone
                if 'legacy' in group.policies:
                    continue
                policy = ego_resreq_utils.build_policy_from_group(group)
                sub_demands = [{'name': group.uuid,
                                'resreq': '',
                                'extra_resreq': '',
                                'maxslots': 0,
                                'force_realloc_migrate': False,
                                'policy': policy}]
                self.create_alloc_for_sg(group.uuid, sub_demands)

        # here we need to retrive the deleted group, modify temp context
        ctx = copy.copy(self.ctxt)
        ctx._read_deleted = 'only'
        removed_groups = objects.InstanceGroupList.get_all(ctx)
        for group in removed_groups:
            grp_allocname = ego_constants.UNDERLINE.join(
                            [ego_constants.SERVER_GROUP, group.uuid])
            if grp_allocname not in self.allocname_id_map:
                continue
            # resize it each time
            members = db.instance_group_members_get(ctx, group.uuid)
            group_insts = [inst for inst, allocname in
                           self.inst_allocname_map.items()
                           if grp_allocname == allocname]
            if members and group_insts:
                cleared = self.cleared_policy_group.get(group.uuid, False)
                if cleared:
                    continue
                filter_properties = {'group_uuid': group.uuid,
                                     'clear_policy': True}
                try:
                    self.resize_alloc_for_sg(self.ctxt, filter_properties)
                    self.cleared_policy_group[group.uuid] = True
                except (ego_exception.EGOExpConnLost,
                        ego_exception.EGOExpResizeAllocFailed) as egoex:
                    LOG.error(six.text_type(egoex))
                except Exception as e:
                    LOG.error(_LE('Failed to delete server group due to: '
                                  '%(reason)s') % {'reason': six.text_type(e)})
            else:
                self.ego_free_alloc(grp_allocname)
                ego_utils.safe_remove(self.cleared_policy_group, group.uuid)

    def __cache_update_inst(self, inst_uuid, cache_dict):
        """
        This method is to update inst related cache.
        :param cache_dict: the inst_uuid related attribute {key: value} dict;
                           its keys can include (decision_id, max_slot, resreq,
                           alloc_name, ego_host, alloc_id, flavorid)
        """
        decision_id = cache_dict.get("decision_id")
        if decision_id:
            self.inst_dec_map[inst_uuid] = decision_id
            self.dec_inst_map[decision_id] = inst_uuid

        max_slot = cache_dict.get("max_slot")
        if max_slot is not None:
            self.inst_maxslots_map[inst_uuid] = max_slot

        resreq = cache_dict.get("resreq")
        if resreq:
            self.inst_resreq_map[inst_uuid] = resreq

        alloc_name = cache_dict.get("alloc_name")
        if alloc_name:
            self.inst_allocname_map[inst_uuid] = alloc_name

        alloc_id = cache_dict.get("alloc_id")
        if alloc_id:
            self.inst_allocid_map[inst_uuid] = alloc_id

        res_group = cache_dict.get("res_group")
        if res_group:
            self.inst_rg_map[inst_uuid] = res_group
        # for instance rebuild, resize and migrate scenarios, the source EGO
        # host related cache should also be updated
        ego_host = cache_dict.get("ego_host")
        flavorid = cache_dict.get("flavorid")
        if ego_host:
            old_egohost = self.inst_egohost_map.get(inst_uuid)
            if old_egohost:
                ego_utils.safe_remove(self.egohost_inst[old_egohost],
                                      inst_uuid)
                old_flavorid = self.inst_flv_idmap.get(inst_uuid)
                if old_flavorid in self.all_flavors[old_egohost]:
                    self.all_flavors[old_egohost].remove(old_flavorid)
                if not flavorid:
                    self.all_flavors[ego_host].append(old_flavorid)
            self.inst_egohost_map[inst_uuid] = ego_host
            self.egohost_inst[ego_host].add(inst_uuid)

        if flavorid:
            self.inst_flv_idmap[inst_uuid] = flavorid
            if ego_host:
                self.all_flavors[ego_host].append(flavorid)

    def __cache_pop_inst(self, inst_uuid):
        decid = self.inst_dec_map.get(inst_uuid)
        ego_utils.safe_remove(self.dec_inst_map, decid)

        egohost = self.inst_egohost_map.get(inst_uuid)
        flavor_id = self.inst_flv_idmap.get(inst_uuid)
        if egohost in self.egohost_inst:
            ego_utils.safe_remove(self.egohost_inst[egohost],
                                  inst_uuid)
        ego_utils.safe_remove(self.inst_flv_idmap, inst_uuid)
        if egohost in self.all_flavors:
            if flavor_id in self.all_flavors[egohost]:
                self.all_flavors[egohost].remove(flavor_id)

        ego_utils.safe_remove(self.inst_allocid_map, inst_uuid)
        ego_utils.safe_remove(self.inst_dec_map, inst_uuid)
        ego_utils.safe_remove(self.inst_egohost_map, inst_uuid)
        ego_utils.safe_remove(self.inst_flv_idmap, inst_uuid)
        ego_utils.safe_remove(self.inst_maxslots_map, inst_uuid)
        ego_utils.safe_remove(self.inst_resreq_map, inst_uuid)
        ego_utils.safe_remove(self.inst_rg_map, inst_uuid)
        ego_utils.safe_remove(self.inst_allocname_map, inst_uuid)

    def __cache_update_alloc(self, alloc_id, cache_dict):
        alloc_name = cache_dict.get("alloc_name")
        if alloc_name:
            self.allocname_id_map[alloc_name] = alloc_id

    def __cache_pop_alloc(self, alloc_name):
        ego_utils.safe_remove(self.allocname_id_map, alloc_name)

    def __cache_update_dec(self, dec_id, cache_dict):
        alloc_id = cache_dict.get("alloc_id")
        self.decid_list.append(dec_id)
        if alloc_id:
            self.dec_allocid_map[dec_id] = alloc_id

    def __cache_pop_dec(self, dec_id):
        ego_utils.safe_remove(self.decid_list, dec_id)
        ego_utils.safe_remove(self.dec_allocid_map, dec_id)

    def ego_release_decision(self, decid):
        allocid = self.dec_allocid_map.get(decid)
        if allocid and allocid in self.allocname_id_map.values():
            self.ego_hdl.ego_release_decision(allocid, decid)
        self.__cache_pop_dec(decid)
        ego_utils.safe_remove(self.orphan_decid_list, decid)

    def release_decision(self, instanceid, inst_props=None):
        inst_props = inst_props or {}
        decid = self.inst_dec_map[instanceid]
        self.ego_release_decision(decid)
        ego_host = self.inst_egohost_map.get(instanceid)
        self.consume_inst_from_egohost(ego_host, inst_props, False)
        self.__cache_pop_inst(instanceid)
        self._iterate_extension_managers("instance_delete", instanceid)

    def resize_alloc_for_exist_inst(self, inst, **kw):
        """Resizes an alloc for an inst."""
        instance_id = inst['uuid']
        vm_state = inst['vm_state']
        if not vm_state or vm_state == vm_states.SHELVED_OFFLOADED:
            return
        if not (inst['host'] and inst['launched_at']):
            return

        filter_properties = self.__gen_filter_props(inst)
        filter_properties["force_realloc_migrate"] = True
        filter_properties['group_uuid'] = kw.get('group_uuid')
        filter_properties['old_resreq'] = kw.get('old_resreq')
        (alloc_id, alloc_decid, alloc_host) = (
                self.resize_alloc_for_sg(self.ctxt, filter_properties))
        self.sync_aggregate_reserveon_for_inst(instance_id)

    def __get_extra_resreq(self, force_host_list=None, force_node_list=None,
                           ignore_host_list=None, extra_select_str=''):
        # get affinity host string
        affinity_hosts = (
            ego_resreq_utils.get_affinity_hosts(force_host_list,
                                                force_node_list,
                                                ignore_host_list,
                                                self.host_egohost))
        extra_resreq = ego_resreq_utils.get_extra_resreq(extra_select_str,
                                                         affinity_hosts)
        return extra_resreq

    def ego_get_resource_plan(self, name, scheduler, flag, with_version):
        return self.ego_hdl.ego_get_resource_plan(
            name, scheduler, flag, with_version)

    def ego_set_resource_plan_md(self, name, cstr, flag):
        return self.ego_hdl.ego_set_resource_plan_md(name, cstr, flag)

    def ego_create_consumer(self, consumer_name, exec_user, parent_path,
                            rebalance, reclaim_time, users, plans):
        return self.ego_hdl.ego_create_consumer(
            consumer_name, exec_user, parent_path,
            rebalance, reclaim_time, users, plans)

    def ego_add_remove_consumer_for_md(self, consumer_path, consumer_name,
                                       md_plan_name, operation):
        return self.ego_hdl.ego_add_remove_consumer_for_md(
            consumer_path, consumer_name, md_plan_name, operation)

    def ego_set_consumer(self, consumer_path, consumer_name,
                         rebalance, reclaim_timeout):
        return self.ego_hdl.ego_set_consumer(
            consumer_path, consumer_name, rebalance, reclaim_timeout)

    def ego_get_consumer_tree(self, role):
        return self.ego_hdl.ego_get_consumer_tree(role)

    def ego_delete_consumer(self, consumer_path):
        return self.ego_hdl.ego_delete_consumer(consumer_path)

    def get_service_instance_info(self, resource_plan):
        return self.ego_hdl.get_service_instance_info(resource_plan)

    def schedule_request_completed(self, ctxt):
        """Called when all scheduling tasks of a request have finished
        """
        # Clear instance properties cache
        self.inst_props_map.clear()
        self.ego_filter_driver.schedule_request_completed(ctxt)

    def _iterate_extension_managers(self, func_name, *args, **kwargs):
        for ext in self.extension_managers:
            try:
                func = getattr(ext, func_name)
                func(*args, **kwargs)
            except AttributeError as e:
                LOG.warning(_LW('Failed to run %(func)s for %(class)s due to: '
                                '(%(reason)s)')
                               % {'func': func_name,
                                  'class': ext.__class__,
                                  'reason': six.text_type(e)})

    def get_instance_info(self, context, instance_uuid):
        """
        Get ego manager cache info and construct the cache based on the
        instance uuid.
        :param context: nova context object
        :param instance_uuid: instance uuid
        :return: return a dict include the specified instance info from
                 ego manager
        """
        host = self.inst_egohost_map.get(instance_uuid)
        resreq = self.inst_resreq_map.get(instance_uuid)
        inst_props = self.inst_props_map.get(instance_uuid)
        cache_dict = {instance_uuid: {'resreq': resreq, 'host': host,
                                      'inst_props': inst_props}}
        return cache_dict
