# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import collections
import re

import prettytable
import six

from nova.compute import flavors
from nova import context
from nova import db
from nova import exception
from nova.ibm.ego import ego_constants
from nova.ibm.ego import ego_gettextutils
from nova.ibm.ego.wrg import constants as wrg_consts
from nova import image
from nova import objects
from nova.objects import base as base_obj
from nova.scheduler.ibm.ego import ego_exception
from nova import servicegroup
from nova import utils

from oslo_config import cfg
from oslo_log import log as logging

_ = ego_gettextutils._
_LW = ego_gettextutils._LW
_LE = ego_gettextutils._LE

CONF = cfg.CONF

LOG = logging.getLogger(__name__)

# Constant for the minimum value of powervm:proc_units in a flavor.
POWERVM_MIN_PROC_UNITS = 0.05
# Constant for the maximum length of row in the prettytable.
MAX_COLS = 100

# Map of velocity goal's key-value pair
VELOCITY_GOAL_MAP = {wrg_consts.SLOWEST_VELOCITY: 40,
                     wrg_consts.SLOW_VELOCITY: 50,
                     wrg_consts.MODERATE_VELOCITY: 60,
                     wrg_consts.FAST_VELOCITY: 75,
                     wrg_consts.FASTEST_VELOCITY: 90}


def list2map(lists, key):
    """Change a list to map with the key we care about.
    [{"name":"cpu.percent","value":0.123,"source":"mysource",
      "timestamp":"20131212"}, {"name":"memory","value":23434,
      "source":"mysource","timestamp":"20131212"}] will be changed
    to {"cpu.percent":0.123,"memory":23434}

    :param lists: The list to be change to map
    :param key: The key we care about.
    :return: A map with key we care and the value.
    """
    return dict((item[key], item['value']) for item in lists)


def __get_instance_system_metadata(instance):
    meta = utils.instance_sys_meta(instance)
    if not meta:
        # Didn't have system metadata, so try to load it from the DB.
        inst = objects.Instance.get_by_uuid(context.get_admin_context(),
                                            instance.get('uuid'))
        meta = utils.instance_sys_meta(inst)
    return meta


def get_hypervisor_data(filter_properties):
    """Returns a dict with 'hypervisor_type' and 'architecture' as keys."""
    request_spec = filter_properties.get('request_spec', {})

    image = request_spec.get('image', {})
    image_props = image.get('properties', {})
    # NOTE(jwcroppe): Handle key changes introduced in Mitaka
    hyper_type = (image_props.get('img_hv_type') or
                  image_props.get('hypervisor_type'))
    arch = (image_props.get('hw_architecture') or
            image_props.get('architecture'))

    if hyper_type is None or arch is None:
        # e.g., we go down this path for evacuate scenarios
        instance_props = request_spec.get('instance_properties', {})
        system_meta = utils.instance_sys_meta(instance_props) or {}
        if hyper_type is None:
            hyper_type = system_meta.get('image_hypervisor_type')
        if arch is None:
            arch = instance_props.get('architecture')

    return dict(hypervisor_type=hyper_type, architecture=arch)


def __get_vcpu_mem_based_on_hypervisor(instance_type, filter_properties,
                                       prefer_instance_type):
    """Get vcpus and memory based on different hypervisor types."""

    # To start with, get the vcpus and memory information from the flavor.
    flavor_vcpus = instance_type.get('vcpus')
    flavor_memory_mb = instance_type.get('memory_mb')
    memory_mb = flavor_memory_mb

    image_hvtype = get_hypervisor_data(filter_properties)['hypervisor_type']
    if image_hvtype in ego_constants.POWERVM_HYPERVISOR_TYPES:
        # For PowerVM, retrieve the related requirements from extra_specs.
        extra_specs = instance_type.get('extra_specs') or {}
        dedicated_mode = extra_specs.get('powervm:dedicated_proc', 'false')
        proc_units = extra_specs.get('powervm:proc_units')

        if dedicated_mode.lower() == 'true':
            # They are in dedicated mode, just use the instance vcpus.
            vcpus = flavor_vcpus
        elif proc_units:
            # Was not dedicated mode, but proc units were specified.
            proc_units = float(proc_units)
            if proc_units < POWERVM_MIN_PROC_UNITS:
                proc_units = POWERVM_MIN_PROC_UNITS
                LOG.warning(_LW("Invalid PowerVM value for processing units. "
                                "The Platform Resource Scheduler will use the "
                                "default value of '%s' instead.")
                            % str(POWERVM_MIN_PROC_UNITS))
            LOG.debug("Image hypervisor type '%(image_hvtype)s' "
                      "and processor units '%(proc_units)s' specified." %
                      {'image_hvtype': image_hvtype,
                       'proc_units': proc_units})
            vcpus = proc_units
        else:
            # No proc units set on flavor, so let's assume the same default
            # behavior that PowerVC performs.
            try:
                from powervc_nova.virt.ibmpowervm.common import proc_utils
                LOG.warning(_LW("The flavor's powervm:proc_units extra spec "
                                "was not specified; assuming default."))
                # Expecting float type for 'proc_units' to be returned
                proc_units = proc_utils.get_num_proc_units(flavor_vcpus)
                vcpus = proc_units
            except Exception:
                # i.e., fall back case for any non-PowerVC-based user.
                vcpus = flavor_vcpus
    else:
        # not PowerVM
        wrg_def = filter_properties.get('request_spec', {}).get('wrg_def')
        if wrg_def:
            # For zHPM cases, Get instance request CPU capacity based on
            # WRG definition
            spec = filter_properties.get('request_spec', {})
            inst_name = spec['instance_properties']['hostname']
            vm_capacity = get_cpu_capacity_for_instance(flavor_vcpus,
                                                    wrg_def, inst_name)
            return (float(vm_capacity), float(memory_mb))
        else:
            vcpus = flavor_vcpus

    if not prefer_instance_type:
        # For situations in which out of band operations can occur (i.e., where
        # a user changes the VM specifications outside of OpenStack), PRS
        # supports the user specifying the 'vcpus' and 'memory_mb' in the
        # VM's system metadata; so let's load that if preferred AND specified.
        request_spec = filter_properties.get('request_spec') or {}
        instance = request_spec.get('instance_properties') or {}

        # Set vcpus
        try:
            vm_vcpus = get_instance_vcpus(instance, image_hvtype)
        except KeyError:
            vm_vcpus = None
        if vm_vcpus is not None:
            vcpus = vm_vcpus

        # Set memory_mb
        try:
            vm_memory_mb = get_instance_memory_mb(instance, image_hvtype)
        except KeyError:
            vm_memory_mb = None
        if vm_memory_mb is not None:
            memory_mb = vm_memory_mb

    return (float(vcpus), float(memory_mb))


def check_metadata_reserved_ratio_value(value):
    ratio = 1.0
    try:
        ratio = float(value)
    except Exception as e:
        LOG.error(_LE('Failed to transform user specified ratio to'
                      ' float due to %s') % six.text_type(e))
        msg = (_("Invalid ratio specified in instance metadata "
                 "when booting the instance. Please specify a valid "
                 "float value."))
        raise exception.InvalidInput(reason=msg)
    if ratio <= 0.0 or ratio > 1.0:
        msg = (_("Invalid reserve ratio specified in instance metadata "
                 "when booting the instance. Please specify a value in "
                 "(0,1]."))
        raise exception.InvalidInput(reason=msg)


def get_service_class_for_instance(wrg_def, inst_name):
    """Return classfied service class"""
    wrg_performance_policy = wrg_def[wrg_consts.PERF_POLICY]
    wrg_service_classes = wrg_performance_policy[wrg_consts.SVC_CLASSES]
    for sc in wrg_service_classes:
        sc_vs_name_filters = sc[wrg_consts.SC_VS_NAME_FILTERS]
        for sc_vs_name_filter in sc_vs_name_filters:
            if re.match(sc_vs_name_filter, inst_name):
                return sc
    return {}


def get_importance_level_for_instance(wrg_def, inst_name):
    """Return importance level for instance"""
    wrg_performance_policy = wrg_def[wrg_consts.PERF_POLICY]
    wrg_policy_importance = wrg_performance_policy[
            wrg_consts.PERF_POLICY_INFO][wrg_consts.POLICY_BUSNS_IMPTC]
    wrg_service_class = get_service_class_for_instance(wrg_def, inst_name)
    sc_importance_level = wrg_service_class[wrg_consts.SC_BUSNS_IMPTC]
    return (wrg_policy_importance + sc_importance_level.capitalize())


def get_cpu_capacity_for_instance(vcpus, wrg_def, inst_name):
    """
    Calculate virtual server's capacity based on vcpus and velocity goal
    """
    wrg_service_class = get_service_class_for_instance(wrg_def, inst_name)
    velocity_goal = wrg_service_class[wrg_consts.SC_VEL_GOAL]
    ratio = CONF.zhpm_vm_cpu_capacity_ratio
    velocity_goal = VELOCITY_GOAL_MAP[velocity_goal]
    return (float(vcpus) / (100 / float(velocity_goal) + 1 / ratio - 1))


def get_consume_str_by_instance_type(instance_type, filter_properties,
                                     metadata=None, prefer_instance_type=True):
    """Get consume string based on given instance type.

    :param instance_type: instance type properties
    :param filter_properties: the scheduler filter properties
    :param prefer_instance_type: If true, will use the vcpu and memory data
    from the flavor, otherwise will fallback to the instance system metadata
    or the instance data
    :returns: 1) instance_type have not wrg_def: consume(vcpu=xxx:memSize=xxx).
              2) instance_type have wrg_def: consume(highLow=2:memSize=2048)
    """
    metadata = metadata or {}
    (vcpus, memory_mb) = __get_vcpu_mem_based_on_hypervisor(
        instance_type, filter_properties, prefer_instance_type)
    # if user specified ratio in metadata, apply the ratio to consume str
    # (For zHPM cases, only apply memory ratio to consume str.)
    wrg_def = filter_properties['request_spec'].get('wrg_def')
    if metadata:
        if wrg_def is None:
            cvalue = metadata.get(ego_constants.INSTANCE_METADATA_CPU_KEY, 1.0)
            check_metadata_reserved_ratio_value(cvalue)
            if vcpus:
                vcpus = vcpus * float(cvalue)
        mvalue = metadata.get(ego_constants.INSTANCE_METADATA_MEMORY_KEY, 1.0)
        check_metadata_reserved_ratio_value(mvalue)
        memory_mb = memory_mb * float(mvalue)

    if wrg_def:
        # Construct consume string for zHPM case
        spec = filter_properties.get('request_spec', {})
        inst_name = spec['instance_properties']['hostname']
        vm_importance = get_importance_level_for_instance(wrg_def, inst_name)
        consume_str = 'consume(%s=%s:memSize=%s)' % (
            str(vm_importance), str(float(vcpus)), str(float(memory_mb)))
    else:
        consume_str = 'consume(vcpu=%s:memSize=%s)' % (
            str(float(vcpus)), str(float(memory_mb)))
    LOG.debug("EGO's consume expression string is (%s)." % consume_str)
    return consume_str


def parse_zhpm_importance_level(consume_str):
    """
    Parse the zHPM instance consume resreq to get the importance level.
    :Param consume_str: zhpm instance consume resreq
    :return string: importance level
            float: importance level capacity
    For example, consume_str is 'consume(highHigh=1.0:memSize=512)',
    function will return (highHigh, 1.0).
    """
    pattern = re.compile(r'consume\((.+):')
    importance_level_str = pattern.search(consume_str).groups()[0]
    importance_level_list = importance_level_str.split("=")
    level_name = importance_level_list[0]
    level_value = importance_level_list[1]
    return (level_name, float(level_value))


def adjust_zhpm_placement_metrics(metrics_map, inst_schedule_info_map):
    """
    Adjust CPU available capacity reported from zHPM monitor:
    1. Reduce available CPU capacity by instance's requested CPU capacity,
        if EGO allocation include some instances that not know by zHPM
    2. Increase available CPU capacity by instance's real CPU capacity,
        if zHPM include some instances that not know by EGO
    3. Calculate total CPU capacity for each importance level based on
        the adjusted available CPU capacity
    :Param metrics_map: The metrics map of compute node
    :Param inst_schedule_info_map: The map about the relationship between
     instance uuid and its res req str
    """
    # Inner function to adjust available CPU capacity from $impl_index
    # importance level to lowest importance level
    def adjust_available_capacity(metrics_map, impl_index, capacity):
        for impl in ego_constants.IMPORTANCE_LEVELS[impl_index:]:
            metrics_map[impl] += capacity
    # Adjust metrics
    for key in ego_constants.IMPORTANCE_LEVELS:
        vs_real_caps = metrics_map[key + ego_constants.VS_CPU_CAPACITY_SUFFIX]
        # Construct map about instance uuid and its real CPU capacity
        # on current importance level: {'1111': '1.5', '2222': 0.6, ...}
        vs_caps_in_zhpm = {}
        if vs_real_caps:
            vs_caps_in_zhpm = convert_vs_caps_to_dict(vs_real_caps)
        vs_caps_in_alloc = inst_schedule_info_map.get(key, {})
        inst_only_in_alloc = (set(vs_caps_in_alloc) - set(vs_caps_in_zhpm))
        inst_only_in_zhpm = (set(vs_caps_in_zhpm) - set(vs_caps_in_alloc))
        key_index = ego_constants.IMPORTANCE_LEVELS.index(key)

        # EGO allocation include some instance that not know by zHPM
        for inst in inst_only_in_alloc:
            adjust_available_capacity(metrics_map, key_index,
                                    0 - vs_caps_in_alloc[inst])
        # zHPM include some instance that not know by EGO
        for inst in inst_only_in_zhpm:
            adjust_available_capacity(metrics_map, key_index,
                                    float(vs_caps_in_zhpm[inst]))
        # Calculate total CPU capacity for each importance level by increase
        # VM request CPU capacity from current importance level to lowest
        # importance level
        if vs_caps_in_alloc:
            index = ego_constants.IMPORTANCE_LEVELS.index(key)
            for vs_cap in six.itervalues(vs_caps_in_alloc):
                for impt_level in ego_constants.IMPORTANCE_LEVELS[index:]:
                    metrics_map[impt_level] += vs_cap


def convert_vs_caps_to_dict(vs_real_caps):
    """
    Convert vs_caps string to dict:
    input = "11111=2.22222:2=3.3:33333=4.4"
    return: {'11111': '2.2', '33333': '4.4', '22222': '3.3'}
    """
    return dict(
        vs_real_cap.split('=') for vs_real_cap in vs_real_caps.split(':'))


def convert_zhpm_placement_metrics(metrics_map):
    """
    Convert zHPM placement metrics:
    1. Replace metrics key:
       # 'zhpm.host.cpu.management' -> 'cpu_mgmt'
       # 'zhpm.host.useable.cpu.capacity' -> 'zhpm_cooperative_mode'
       # 'zhpm.host.migrate.vsuuids' -> 'vs_uuids'
    2. Replace all instance uuids in '($importance)_vs_cpu_capacity'
        by EGO decision id
    3. Merge wrg_business_importance, service_class_business_importance
        and host_available_cpu_capacity to one metrics attribute:
        # key is
            $wrg_business_importance + $service_class_business_importance
        # value is $host_available_cpu_capacity
    4. Add new attribute *_vs_cpu_capacity for each
        importance level to represent all instance and its real capacity
        on this importance level
    :Param metrics_map: The metrics map of compute node
    """
    recommended_vms_str = ego_constants.NORECOMMENDEDVMS
    for key in sorted(metrics_map):
        value = metrics_map[key]
        if key == ego_constants.ZHPM_CPU_MGMT_METRICS_NAME:
            metrics_map[ego_constants.ZHPM_CPU_MGMT] = value.get(
                ego_constants.ZHPM_CPU_MGMT)
        elif key == ego_constants.ZHPM_COOPERATIVE_MODE_METRICS_NAME:
            metrics_map[ego_constants.ZHPM_COOPERATIVE_MODE] = value.get(
                ego_constants.ZHPM_COOPERATIVE_MODE)
        elif key == ego_constants.ZHPM_VS_UUIDS_METRICS_NAME:
            if value != ego_constants.NORECOMMENDEDVMS:
                vs_uuids_list = value.get(ego_constants.ZHPM_VS_UUIDS)
                recommended_vms_str = ','.join(
                    (str(vs_uuid[ego_constants.ZHPM_VS_UUID]) + ';'
                     + str(vs_uuid[ego_constants.OVERALL_BUSINESS_IMPORTANCE])
                     + ';' + str(vs_uuid[ego_constants.ZHPM_SC_PI]))
                    for vs_uuid in vs_uuids_list)
        elif key == ego_constants.ZHPM_USEABLE_CPU_CAPACITY_METRICS_NAME:
            for imp_lel_cap in value:
                overall_imp_lel_index = (
                    imp_lel_cap[ego_constants.OVERALL_BUSINESS_IMPORTANCE])
                overall_imp_lel = (
                    ego_constants.IMPORTANCE_LEVELS[overall_imp_lel_index])
                metrics_map[overall_imp_lel] = (
                    imp_lel_cap[ego_constants.HOST_AVAILABLE_CPU_CAPACITY])
                vs_real_caps = ':'.join((vs.get(ego_constants.ZHPM_VS_UUID)
                + '=' + str(vs.get(ego_constants.CPU_CAPACITY_FOR_GOAL)))
                for vs in imp_lel_cap[ego_constants.VS_CPU_CAPACITY_FOR_GOAL])
                vs_real_caps_key = (
                    overall_imp_lel + ego_constants.VS_CPU_CAPACITY_SUFFIX)
                metrics_map[vs_real_caps_key] = vs_real_caps
    metrics_map[ego_constants.ZHPM_VS_UUIDS] = recommended_vms_str
    metrics_map.pop(ego_constants.ZHPM_CPU_MGMT_METRICS_NAME)
    metrics_map.pop(ego_constants.ZHPM_COOPERATIVE_MODE_METRICS_NAME)
    metrics_map.pop(ego_constants.ZHPM_VS_UUIDS_METRICS_NAME)
    metrics_map.pop(ego_constants.ZHPM_USEABLE_CPU_CAPACITY_METRICS_NAME)


def get_consume_str(filter_properties, prefer_instance_type=True):
    """
    Get consume string: consume(vcpu=xxx:memSize=xxx).

    :param filter_properties: the scheduler filter properties
    :param prefer_instance_type: If true, will use the vcpu and memory data
    from the flavor, otherwise will fallback to the instance system metadata
    or the instance data
    """
    spec = filter_properties.get('request_spec', {})
    instance_type = spec.get('instance_type')
    hypervisor_data = get_hypervisor_data(filter_properties)
    image_hvtype = hypervisor_data['hypervisor_type']
    instance_properties = spec.get('instance_properties', {})
    stand_alone_vm = spec.get(ego_constants.STAND_ALONE, False)
    if stand_alone_vm:
        consume_str = get_consume_str_by_instance(instance_properties,
                                                  image_hvtype)
        return consume_str
    else:
        if not instance_type.get('extra_specs'):
            # Let's load the flavor's extra_specs for potential use later
            # (originally was for PowerVM only, but let's do it consistently).
            instance_type_id = instance_type['id']
            instance_type = objects.Flavor.get_by_id(
                filter_properties['context'], instance_type_id)
        metadata = instance_properties.get('metadata') or {}
        consume_str = get_consume_str_by_instance_type(instance_type,
                                    filter_properties, metadata=metadata,
                                    prefer_instance_type=prefer_instance_type)
    return consume_str


def is_positive_integer(str):
    try:
        return int(str) > 0
    except ValueError:
        return False


def get_reserveon_str(filter_properties, aggregateids=None):
    """Get reserveon string: reserveon(aggregate_id_list); e.g.,
    reserveon(1 2 3) or if there is no aggregate specified, then return
    reserveon(group_all).
    """
    aggregateids = aggregateids or []
    scheduler_hints = filter_properties.get('scheduler_hints', {})
    if scheduler_hints:
        host_aggregate_ids = scheduler_hints.get('host_aggregate_id', [])
    else:
        host_aggregate_ids = []

    if not CONF.policy_per_host_aggregate and host_aggregate_ids:
        msg = (_LW("Platform Resource Scheduler could not boot the instance "
                   "to specify host aggregates because the "
                   "policy_per_host_aggregate value is not set to true in the "
                   "nova.conf file."))
        LOG.warning(msg)
        raise exception.InvalidInput(reason=msg)
    if not isinstance(host_aggregate_ids, list):
        host_aggregate_ids = [host_aggregate_ids]
    host_aggregate_id_set = set(host_aggregate_ids)

    invalid_set = set()
    not_exist_set = set()
    for host_aggregate_id in host_aggregate_id_set:
        if not is_positive_integer(host_aggregate_id):
            invalid_set.add(host_aggregate_id)
        elif int(host_aggregate_id) not in aggregateids:
            not_exist_set.add(host_aggregate_id)
    if invalid_set:
        msg = (_LW("The host aggregate IDs (%s) are invalid. Specify"
                   " a positive integer for each host aggregate ID.") %
               ', '.join([str(ha_id) for ha_id in invalid_set]))
        LOG.warning(msg)
        raise exception.InvalidInput(reason=msg)
    if not_exist_set:
        msg = (_LW("The host aggregates with IDs (%s) do not exist. "
                   "Verify the host aggregate ID.") %
               ', '.join([str(ha_id) for ha_id in not_exist_set]))
        LOG.warning(msg)
        raise exception.InvalidInput(reason=msg)

    host_agg_str = ego_constants.RESOURCE_GROUP_ALL
    valid_set = host_aggregate_id_set - invalid_set
    if valid_set:
        host_agg_str = ' '.join([str(ha_id) for ha_id in valid_set])

    reserveon_str = 'reserveon(%s)' % host_agg_str
    LOG.debug("EGO's reserveon expression string is (%s)."
              % reserveon_str)
    return (reserveon_str)


def get_resreq_str(resreq_list, keyword):
    """Get select string, such as:
        select_list=[vcpu>4, disk1>1024 || disk2<2048, location==Xian]
        return "select((vcpu>4)&&(disk1>1024||disk2<2048)&&(location==Xian))"
    Get samehost string, such as:
       samehost_list=[vm1, vm2]
       return "samehost('vm1' 'vm2')"
    Get samehost string, such as:
       samehost_list=[vm1, vm2]
       return "samehost('vm1' 'vm2')"
    Get order string, such as:
        order_list=[0-vcpu, memSize]
        return "order((0-vcpu)(memSize))"
        Note: the former item in order_list takes higher priority than
        the latter one when ordering
    """
    if not any(resreq_list):
        return ''

    index = 0
    resreq_list.sort()
    for resreq in resreq_list:
        if index == 0:
            resreqstr = "( " + str(resreq) + ")"
        else:
            if keyword == "select":
                resreqstr = resreqstr + "&&" + "( " + str(resreq) + ")"
            else:
                resreqstr = resreqstr + "( " + str(resreq) + ")"
        index = index + 1
    resreq_str = keyword + "(" + resreqstr + ")"

    LOG.debug("EGO's resource requirement expression string "
              "is (%s)." % resreq_str)
    return (resreq_str)


def convert_str(str):
    """Convert the string to int or float if the string can be converted.

    :param str: The string that will be converted
    :returns: converted value
    """
    if not str:
        return ''
    try:
        return float(str) if '.' in str else int(str)
    except ValueError:
        return str


def is_numerical(str):
    """Check if a string is numerical number or not.

    :param str: The string that will be checked
    :returns: True if it can be parsed as a numerical value
              False if a base string
    """
    try:
        float(str)
        return True
    except ValueError:
        return False


def extra_specs_to_select(extra_specs_list):
    """Transfer extra_specs to EGO resource requirement.

    :param extra_specs_list: extra_specs for flavors
    :returns: EGO select list
    """
    select_list = []
    for key, req in six.iteritems(extra_specs_list):
        words = req.split()
        op = None
        if words:
            op = words.pop(0)

        if op != '<or>' and not words:
            if is_numerical(str(req)):
                req_str = "'$" + str(key) + "'" + ' == ' + "'" + str(req) + "'"
            else:
                req_str = (
                    "'$" + str(key) + "'" + ' s== ' + "'" + str(req) + "'")
            select_list.append(req_str)
            continue

        if op == '<or>':  # e.g., <or> v1 <or> v2 <or> v3
            index = 0
            while True:
                value = words.pop(0)
                if index == 0:
                    if is_numerical(str(value)):
                        req_str = "('$%s' == '%s')" % (str(key), str(value))
                    else:
                        req_str = "('$%s' s== '%s')" % (str(key), str(value))
                else:
                    if is_numerical(str(value)):
                        req_str += "||('$%s' == '%s')" % (str(key), str(value))
                    else:
                        req_str += "||('$%s' s== '%s')" % (
                            str(key), str(value))
                if not words:
                    break

                op = words.pop(0)  # remove a keyword <or>
                index = index + 1
                if not words:
                    break
            select_list.append(req_str)
            continue
        if op == '<in>':
            try:
                val = eval(words[0])
            except Exception:
                val = words[0]

            _, res_str = generate_attr_str(val, True)
            req_str = "%s %s '$%s'" % (res_str, str(op), str(key))
        else:
            req_str = "'$%s' %s '%s'" % (str(key), str(op), str(words[0]))
        select_list.append(req_str)

    # the string was constructed by the values from the table
    # of aggregate_metadata
    return select_list


def get_affinity_hosts(force_host_list=None, force_node_list=None,
                       ignore_host_list=None, host_egohost=None):
    """Convert force host list and ignore host list to EGO resource
    requirements.

    :param force_host_list: force host list
    :param force_node_list: force node list
    :param ignore_host_list: ignore host list
    :param host_egohost: map from service host(service['host'])
            to EGO host(RPI RES_NAME)
    :returns: force hosts string, ignore hosts string or all of them
              eg: 1, ignore hosts string: (! 'host1') && (! 'host2')
                  2, force hosts string: ('host3') || ('host4')
                  3, ignore hosts and force hosts string:
                    (((! 'host1') && (! 'host2')) && (('host3') || ('host4')))
    """
    affinity_hosts_str = ""

    def get_host_str(host_type, ego_host_list):
        if host_type == 'force':
            return " || ".join([str("('" + h + "')") for h in ego_host_list])
        if host_type == 'ignore':
            return " && ".join([str("(! '" + h + "')") for h in ego_host_list])

    force_ego_hosts = get_force_ego_hosts_by_hosts_nodes(force_host_list,
                                                force_node_list, host_egohost)
    ignore_ego_hosts = convert_hosts_to_egohosts(ignore_host_list,
                                                 host_egohost)
    ignore_host_str = get_host_str('ignore', ignore_ego_hosts)
    force_host_str = get_host_str('force', force_ego_hosts)

    if ignore_host_str and force_host_str:
        affinity_hosts_str = (
            "((%s) && (%s))" % (ignore_host_str, force_host_str))
    elif ignore_host_str and not force_host_str:
        affinity_hosts_str = "(%s)" % ignore_host_str
    elif not ignore_host_str and force_host_str:
        affinity_hosts_str = "(%s)" % force_host_str
    return affinity_hosts_str


def get_extra_resreq(base_resreq, affinity_hosts):
    """Combine the affinity hosts string with the given EGO resreq string.

    :param base_resreq: EGO resreq string
    :param affinity_hosts: affinity hosts string
    :returns: new resreq string
              example: base_resreq = select((vcpu==4))
                       affinity_hosts = (! 'host1')
                       return: select((vcpu==4) && (! 'host1'))
    """
    extra_resreq = ""
    if base_resreq and not affinity_hosts:
        extra_resreq = base_resreq
    elif not base_resreq and affinity_hosts:
        extra_resreq = "select(%s)" % affinity_hosts
    elif base_resreq and affinity_hosts:
        select = base_resreq[len('select'):len(base_resreq)]
        extra_resreq = "select((%s) && (%s))" % (select, affinity_hosts)
    return extra_resreq


def get_instance_memory_mb(instance, hypervisor_type):
    """Get instance memory_mb based on given instance.

    :param instance: Instance object as returned by DB layer
    :param hypervisor_type: Hypervisor type
    :returns: instance memory_mb
    """
    meta = __get_instance_system_metadata(instance)
    if not meta.get('memory_mb'):
        if hypervisor_type in ego_constants.POWERVM_HYPERVISOR_TYPES:
            msg = (_LW("Instance %s has no system metadata with key "
                       "'memory_mb'; it will instead use "
                       "instance['memory_mb'] as the default.")
                   % instance['uuid'])
            LOG.warning(msg)
        memory_mb = instance.get('memory_mb')
    else:
        memory_mb = meta['memory_mb']
    return float(memory_mb)


def get_instance_vcpus(instance, hypervisor_type):
    """Get instance vcpus based on given instance.

    :param instance: Instance object as returned by DB layer
    :param hypervisor_type: Hypervisor type
    :returns: instance vcpus
    """
    meta = __get_instance_system_metadata(instance)
    if not meta.get('vcpus'):
        if hypervisor_type in ego_constants.POWERVM_HYPERVISOR_TYPES:
            msg = (_LW("Instance %s has no system metadata with key 'vcpus'; "
                       "it will instead use instance['vcpus'] as the default.")
                   % instance['uuid'])
            LOG.warning(msg)
        vcpus = instance.get('vcpus')
    else:
        vcpus = meta['vcpus']
    return float(vcpus)


def get_consume_str_by_instance(instance, hypervisor_type):
    """Get consume string based on given instance object.

    :param instance: Instance object as returned by DB layer.
    :param hypervisor_type: Hypervisor type
    :returns: consume(vcpu=xxx:memSize=xxx) or raise exception if failed
    """
    vcpus = get_instance_vcpus(instance, hypervisor_type)
    memory_mb = get_instance_memory_mb(instance, hypervisor_type)
    # if user specified the reserve ratio in instance metadata, then go ahead
    # and reserve the ratio.
    metadata = utils.instance_meta(instance)

    vcpu_val = metadata.get(ego_constants.INSTANCE_METADATA_CPU_KEY)
    if vcpu_val:
        check_metadata_reserved_ratio_value(vcpu_val)
        if vcpus:
            vcpus = vcpus * float(vcpu_val)

    mem_val = metadata.get(ego_constants.INSTANCE_METADATA_MEMORY_KEY)
    if mem_val:
        check_metadata_reserved_ratio_value(mem_val)
        memory_mb = memory_mb * float(mem_val)

    consume_str = "consume(vcpu=" + (
        str(vcpus) + ":memSize=" + str(memory_mb) + ")")
    return consume_str


def generate_res_state_and_attr_str(ego_host, state, attr_dict=None):
    """Generate EGO resource state and attributes string for specified host.

    :param ego_host: hostname of compute node
    :param state: EGO resource state
    :param attr_dict: attributes and values map of EGO resource to be updated
    :return: EGO resource expression string
    """
    attr_dict = attr_dict or {}
    cstr = "dynamic 1 1 %s host %s" % (ego_host, str(state))
    if attr_dict:
        attr_str_list = []
        for attr_name, attr_val in six.iteritems(attr_dict):
            attr_type, res_str = generate_attr_str(attr_val)
            attr_str_list.append(
                str(attr_name) + ' ' + attr_type + ' ' + res_str)
        attr_str = " ".join(attr_str_list)
        cstr += " %s %s" % (str(len(attr_dict)), attr_str)
        LOG.debug('Set compute node %(node)s state to %(state)s'
                  ' with attributes %(attr_dict)s'
                  % {'node': ego_host, 'state': str(state),
                     'attr_dict': str(attr_dict)})
    else:
        cstr += " 0"
        LOG.debug('Set compute node %(node)s state to %(state)s'
                  % {'node': ego_host, 'state': str(state)})
    return cstr


def generate_inst_property_from_inst_type(inst_property, inst_type):
    """Generate new instance property from instance type w/ fields we need."""
    new_inst_property = dict(uuid=inst_property['uuid'],
                             root_gb=inst_type['root_gb'],
                             ephemeral_gb=inst_type['ephemeral_gb'],
                             memory_mb=inst_type['memory_mb'],
                             vcpus=inst_type['vcpus'],
                             vm_state=inst_property.get('vm_state'),
                             task_state=inst_property.get('task_state'))
    return new_inst_property


def get_rpistat_resreq(resreq_str, keyword, old_consume=None):
    """Generate the rpistat counterpart of ego reserveon/consume requirement.
    Get rpistat reserveon string, such as:
        resreq_str = 'reserveon(1 2)'
        return '((1 <in> host_aggregate_ids) || (2 <in> host_aggregate_ids))'
    Get rpistat consume string, such as:
        resreq_str = 'consume(vcpu=2:memSize=64)'
        return '(((vcpu-vcpuConsumed)>=2) && ((memSize-memSizeConsumed)>= 64))'

    :param resreq_str: ego reserveon/consume resource requirement string
    :keyword: 'reserveon' or 'consume'
    :old_consume: for resize case and keyword is 'consume', it will include
             'source_host' and 'old_instance_type' value of the corresponding
             instance
    :returns: the reserveon/consume rpistat string
    """
    if keyword == 'reserveon':
        resgroups = re.match(r'reserveon\((.*)\)', resreq_str)
        resgrouplist = resgroups.groups()[0].split()
        if ego_constants.RESOURCE_GROUP_ALL in resgrouplist:
            return ''
        rpistat_resgroups = ['(' + str(resgroup) + ' <in> ' + (
            ego_constants.RPI_KEY_HOST_AGGREGATE_ID + ')')
            for resgroup in resgrouplist]
        rpistat_reserveon = ' || '.join(rpistat_resgroups)
        return '(' + rpistat_reserveon + ')'
    if keyword == 'consume':
        consume = re.match(r'consume\((.+)=(.+):(.+)=(.+)\)', resreq_str)
        cpu_key, cpu, mem_key, mem = consume.groups()
        rpistat_mem = '((memSize-memSizeConsumed)>=%s)' % str(mem)
        # In zHPM context, only memory resource is hard constraints
        if cpu_key in ego_constants.IMPORTANCE_LEVELS:
            rpistat_consume = rpistat_mem
        else:
            rpistat_cpu = '((vcpu-vcpuConsumed)>=%s)' % str(cpu)
            rpistat_consume = '(%s && %s)' % (rpistat_cpu, rpistat_mem)
        if old_consume:
            # for resize case, resource requirement is different for same host
            # and other hosts
            source_host = old_consume['source_host']

            old_instance_properties = old_consume['old_instance_type']
            old_system_metadata = old_instance_properties.get(
                'system_metadata') or {}
            hypervisor_type = old_system_metadata.get('image_hypervisor_type')
            # default to the vcpus value from the old instance properties
            old_cpu = old_instance_properties['vcpus']

            if hypervisor_type in ego_constants.POWERVM_HYPERVISOR_TYPES:
                # if PowerVM, then the system metadata vcpus value corresponds
                # to PowerVM's proc units value for the instance and is the
                # correct value to use vs. instance['vcpus']
                old_proc_units = old_system_metadata.get('vcpus')
                if old_proc_units:
                    old_cpu = float(old_proc_units)
                else:
                    LOG.warning(_LW("The scheduler error report will likely "
                                    "have an incorrect resource request "
                                    "string for the 'Resources Required' "
                                    "section because the instance's original "
                                    "processor units could not be "
                                    "determined."))

            old_mem = old_instance_properties['memory_mb']
            delta_mem = float(mem) - old_mem
            if delta_mem >= 0:
                delta_rpistat_mem = ('((memSize-memSizeConsumed)>=%s)' %
                                     str(delta_mem))
            else:
                delta_rpistat_mem = ('((memSize-memSizeConsumed+%s)>=0)' %
                                     str(-delta_mem))
            # In zHPM context, only memory resource is hard constraints
            if cpu_key not in ego_constants.IMPORTANCE_LEVELS:
                delta_cpu = float(cpu) - old_cpu
                if delta_cpu >= 0:
                    delta_rpistat_cpu = ('((vcpu-vcpuConsumed)>=%s)' %
                                         str(delta_cpu))
                else:
                    delta_rpistat_cpu = ('((vcpu-vcpuConsumed+%s)>=0)' %
                                         str(-delta_cpu))
                delta_rpistat_consume = ('(%s && %s)' %
                                    (delta_rpistat_cpu, delta_rpistat_mem))
            else:
                delta_rpistat_consume = delta_rpistat_mem
            rpistat_consume = ('((%(abs_consume)s && %(diffhosts)s) ||'
                               ' (%(delta_consume)s && %(samehost)s))' %
                            {'abs_consume': rpistat_consume,
                            'diffhosts': get_affinity_hosts(
                              ignore_host_list=[source_host], host_egohost={}),
                            'delta_consume': delta_rpistat_consume,
                            'samehost': get_affinity_hosts(
                              force_node_list=[source_host])})
        return rpistat_consume


def get_rpistat_filter_str(filter_req, host_egohost):
    """Get the rpistat counterpart string of filter requirement.

    :param filter_req: {'select': [], 'ignore_hosts': [],
                        'force_hosts': [], 'force_nodes': []}
    :param host_egohost: map from service host(service['host'])
            to EGO host(RPI RES_NAME)
    :returns: the filter rpistat requirement
              eg: filter_req = {'select': [$numinstances <2],
                                'ignore_hosts': [host1],
                                'force_hosts': [host2]}
                  return '(select($numinstances<2) && (! host1) && (host2))'
    """
    select = filter_req.get('select')
    ignore_hosts = filter_req.get('ignore_hosts')
    force_hosts = filter_req.get('force_hosts')
    force_nodes = filter_req.get('force_nodes')
    filter_select = ''
    if select:
        filter_select = get_resreq_str(select, 'select')
    # filter's force_hosts default value is None
    # if it is [], it indicates third party wants to ignore all hosts
    if force_hosts is None:
        # set it [], because get_affinity_hosts only accepts a list param and
        # iterate this list to generate force_hosts resreq string
        force_hosts = []
    elif force_hosts == []:
        # when force_hosts is [], all hosts will be ignored
        ignore_hosts = set(ignore_hosts)
        for key in sorted(host_egohost):
            egohost = host_egohost[key]
            ignore_hosts |= egohost
        ignore_hosts = list(ignore_hosts)
    affinity_hosts_str = get_affinity_hosts(force_hosts, force_nodes,
                                            ignore_hosts, host_egohost)
    if filter_select and affinity_hosts_str:
        return '(' + filter_select + ' && ' + affinity_hosts_str + ')'
    elif filter_select:
        return filter_select
    elif affinity_hosts_str:
        return affinity_hosts_str
    else:
        return ''


def get_criteria_str_by_resreq(resreq, keyword):
    """
    Get substring with format 'keyword(str)' based on given resreq string and
    keyword. Counting the bracket instead of using regex to handle the nested
    bracket in the select statement.

    :param resreq: EGO resource requirement
    :param keyword: the topic of the unit string
    :returns: keyword(str)
              eg: resreq = 'reserveon(1 2 3)consume(vcpu=1:memSize=128)
              select(( $maintenance_status==ok)&&( $activeCompute==1))'
                  1, keyword = consume: return 'consume(vcpu=1:memSize=128)'
                  2, keyword = reserveon: return 'reserveon(1 2 3)'
                  3, keyword = select: return 'select((
                    $maintenance_status==ok)&&( $activeCompute==1))'
    """
    try:
        keyword_str = ""
        proc = resreq[resreq.index(
            keyword):]
        bracket_cnt = 0
        end_pos = 0
        for idx, val in enumerate(proc):
            if val == '(':
                bracket_cnt += 1
            elif val == ')':
                bracket_cnt -= 1
                if bracket_cnt == 0:
                    end_pos = idx + 1
                    break

        keyword_str = proc[:end_pos]
    except Exception:
        pass

    LOG.debug("EGO's %(keyword)s expression string is (%(keyword_str)s)."
              % {'keyword_str': keyword_str, 'keyword': keyword})
    return keyword_str


def get_report_table(row_list, t_property='Property', t_value='Value'):
    """
    Make a prettytable whose rows are filled with row_list info.
    :param row_list: the table's row list, like:
                     [[rowname1, rowvalue1], [rowname2, rowvalues2], ...]
    :param t_property: tha table's property column name
    :param t_value: the table's value column name
    :returns: prettytable instance with row_list filled
    """
    table = prettytable.PrettyTable([t_property, t_value], caching=False)
    table.align = 'l'
    for row in row_list:
        lines = six.text_type(row[1]).strip()
        lines += '\n '
        # if value has a newline, add in multiple rows
        lines = lines.split(r'\n')
        col1 = row[0]
        for line in lines:
            count = 0
            while ((MAX_COLS * (count + 1)) < len(line)):
                table.add_row([col1,
                        line[(count * MAX_COLS):((count + 1) * MAX_COLS)]])
                count += 1
                col1 = ''
            table.add_row([col1, line[(count * MAX_COLS):]])
            col1 = ''
    return table


def get_rpi_res_name(node_name):
    """
    Generate RPI RES_NAME from node_name. Replace all none word letters which
    are not valid resource names for EGO to the '-' character.

    :param node_name: like domain-c206(cluster2)
    :returns: a valid RES_NAME like domain-c206-cluster2
    """
    if node_name:
        tmp = re.sub(r'[^\.\w]+\Z', '', node_name)
        return re.sub(r'[^\.\w]+', '-', tmp)


def convert_hosts_to_egohosts(svc_host_list, host_egohosts):
    """
    Convert service host list to EGO host list(RPI RES_NAME).

    :param svc_host_list: list of service host(service['host'])
    :param host_egohosts: map from service host(service['host'])
            to set of EGO host(RPI RES_NAME)
    :returns: a list of EGO host(RPI RES_NAME)
    """
    svc_host_list = svc_host_list or []
    ego_hosts = set()
    for svc_host in svc_host_list:
        # if the service host is valid, then convert it to its corresponding
        # EGO hosts, else consider itself as its corresponding EGO host.
        ego_hosts |= host_egohosts.get(svc_host, set([svc_host]))
    return sorted(list(ego_hosts))


def get_rg_list(aggids):
    """
    Get resource group list from aggregates ids, return 'OpenStackHosts'
    as default if there is no aggregate id.
    """
    if not aggids:
        return [ego_constants.RESOURCE_GROUP_DEFAULT]
    else:
        return [str(aid) for aid in aggids]


def is_powervc_driver_host(htype, node_stats):
    """
    Judge if the host is sync'd from PowerVC by PowerVC driver.
    The PowerVC driver will NOT have the hypervisor state compute node stat
    for PowerVM hypervisors.
    TODO(jwcroppe): We should have a better way to determine this.

    :param htype: hypervisor type
    :param node_stats: node stats from compute node.
    :return: True, if host is synced from PowerVC by PowerVC driver
             False, if not host synced from PowerVC by PowerVC driver.
    """
    flag = False
    if htype in ego_constants.POWERVM_HYPERVISOR_TYPES:
        if "hypervisor_state" not in node_stats:
            flag = True
    return flag


def generate_attr_str(val, withQuotes=False):
    """
    Convert python data structures to EGO attribute string.
    Notice that no ',' or ';' should exist in the val, since
    these two are delimiters of EGO.

    :param val: inpute value.
    :param withQuotes: the output string should be embraced by quotes
                or not. Sameple: If false: i686,hvm. If true:
                'i686', 'hvm'. True mostly used by filters since
                the resreq expression need quotes.
    :returns: The attribute type and string.
    """
    attr_type = ''
    attr_str = ''

    if isinstance(val, (float, bool)):
        attr_type = '1'
        attr_str = str(val)
        if isinstance(val, bool):
            if val is True:
                attr_str = '1.0'
            else:
                attr_str = '0.0'
    elif isinstance(val, (int, long)):
        attr_type = '0'
        attr_str = str(val)
    else:
        if not val:
            return ('', '')

        attr_type = '2'
        if isinstance(val, six.string_types):
            try:
                val = eval(val)
            except Exception:
                # any exception in eval() and dict() will be ignored
                # such as ValueError, NameError
                pass

        if isinstance(val, dict):
            attr_str = ','.join(sorted(val.keys()))
        elif isinstance(val, (list, set, tuple)):
            # This part of code will try to convert 2 level
            # structures to a string. The 2 level structure
            # could be a hybrid structure. e.g. [('a','b'), ['c', 'd']]
            # will be transfered to a,b;c,d.
            try:
                attr_str = ';'.join(
                    ','.join(sorted(list(instance)))
                    if isinstance(instance, (list, set, tuple, dict))
                    else str(instance)
                    for instance in val)

                if ',' not in attr_str:
                    attr_str = attr_str.replace(';', ',')
            except Exception:
                attr_str = str(val)  # more than 2 level nested structures.
        else:
            attr_str = str(val)

        if attr_str and ' ' in attr_str:
            attr_str = '"' + attr_str.replace('"', "'") + '"'

        if withQuotes \
           and not attr_str.startswith('"') \
           and not attr_str.startswith("'"):
            attrs = []
            attrs = attr_str.split(',')
            attr_str = ','.join("'" + attr + "'" for attr in attrs)

    return (attr_type, attr_str)


def convert_filter(filter):
    """
    Convert filter condition to EGO filter syntax.

    :param filter: filter condition dict
    :return: EGO filter
        example: filter = {'allocation_id': 'xx',
                           'sub_demand_name': 'yy'}
        return {2:'xx', 8:'yy'}
    """
    converted_filer = {}
    ego_filter = {'all_allocation': 0,
                  'resourceplan_name': 1,
                  'allocation_id': 2,
                  'consumer_name': 3,
                  'client_name': 4,
                  'resource_name': 5,
                  'dimension': 6,
                  'sub_demand_name': 7,
                  'activity_id': 8,
                  'decision_id': 9,
                  'topo_namespace': 10,
                  'allocation_name': 11}
    for k, v in filter.items():
        if k in ego_filter.keys():
            converted_filer[ego_filter[k]] = v
    return converted_filer


def build_policy_from_group(group):
    """
    Build group policy from group.

    :param group: Build group policy from the server group
    :return: The group policy
    """
    group_policy = []
    policies = group.policies
    for po in policies:
        policy = {}
        policy_type = None
        if po == 'affinity':
            policy_type = 1
        elif po == 'anti-affinity':
            policy_type = 2
        policy.update({'policy_type': policy_type})
        group_policy.append(policy)
    return group_policy


def get_allocation_type(instance_uuid, inst_alloc_map, inst_subdemand_map):
    """
    Get the allocaiton type by the instance uuid.
        stand alone allocation: allocation_id=sub_demand_id=instance_id
        server group allocation:allocation_id = sub_demand_id=server_group_uuid
        heat allocation: others

    :param instance_uuid: The instance uuid
    :param inst_alloc_map: The instance allocation map cache in ego manager
    :param inst_subdemand_map: The instance allocation subdemand map cache in
                               the EGO manager
    :return: Allocation type
    """
    allocation_name = inst_subdemand_map[instance_uuid]
    sub_demand_name = inst_subdemand_map[instance_uuid]
    if ((instance_uuid == allocation_name) and
            (instance_uuid == sub_demand_name)):
        return ego_constants.STAND_ALONE
    elif allocation_name == sub_demand_name:
        return ego_constants.SERVER_GROUP
    else:
        return ego_constants.HEAT


def get_alloc_type_from_allocname(allocname):
    if allocname.startswith(ego_constants.HEAT):
        return ego_constants.HEAT
    elif allocname.startswith(ego_constants.SERVER_GROUP):
        return ego_constants.SERVER_GROUP
    else:
        return ego_constants.STAND_ALONE


def get_alloc_type_from_allocid(alloc_id, allocname_id_map):
    for allocName, allocId in allocname_id_map.items():
        if allocId == alloc_id:
            return get_alloc_type_from_allocname(allocName)
    return None


def update_instance_system_metadata(instance_uuid, metadata):
    """Update instance system metadata in nova database.

    :param instance_uuid: The instance uuid
    :param metadata: The key/value pair of instance system metadata
    :returns: The updated instance system metadata
    """
    ctxt = context.get_admin_context()
    return db.instance_system_metadata_update(ctxt, instance_uuid, metadata,
                                              False)


def get_image_metadata_by_image_id(context, image_id):
    """Get image metadata by image_id.

    :param context: This context is build from request object
    :param image_id: The uuid of this image
    :returns: The image object
    """
    image_api = image.API()
    return image_api.get(context, image_id)


def convert_sub_demands_for_heat(context, sub_demands):
    """Convert subdemands from heat request to EGO allocation.

    :param context: context from heat request object
    :param sub_demands: subdemands from heat request
    :returns: converted subdemands with a key `filter_properties`,
              and it's value is filter_properties.
    """
    # check if sub_demands is valid
    keys = ['name', 'policy', 'maxslots', 'sub_spec']
    ego_sub_demands = []
    for sub_demand in sub_demands:
        ret = check_keys_in_dict(keys, sub_demand)
        if ret is not True:
            msg = (_('key: %(key)s is not in sub_demand: %(subdemand)s') %
                   {'key': ret, 'subdemand': six.text_type(sub_demand)})
            raise ego_exception.InvalidSubdemandInput(reason=msg)
        sub_spec = sub_demand.get('sub_spec')
        image_id = sub_spec.get('image_id')
        flavor_id = sub_spec.get('flavor_id')
        scheduler_hints = sub_spec.get('scheduler_hints')
        availability_zone = sub_spec.get('availability_zone')
        filter_properties = {}

        image = get_image_metadata_by_image_id(context, image_id)
        instance_type_obj = flavors.get_flavor_by_flavor_id(flavor_id)
        instance_type = base_obj.obj_to_primitive(instance_type_obj)
        request_spec = {'image': image or {},
                        'instance_type': instance_type,
                        'instance_properties': {'availability_zone':
                                                availability_zone}
                        }

        filter_properties['scheduler_hints'] = scheduler_hints
        filter_properties['request_spec'] = request_spec

        ego_sub_demand = {}
        ego_sub_demand['name'] = sub_demand['name']
        ego_sub_demand['filter_properties'] = filter_properties
        ego_sub_demand['maxslots'] = sub_demand['maxslots']
        ego_sub_demand['policy'] = sub_demand['policy']
        ego_sub_demand['force_realloc_migrate'] = False
        ego_sub_demands.append(ego_sub_demand)
    return dict_key_value_to_str(ego_sub_demands)


def check_keys_in_dict(keys, dict):
    """Check if the dict has the key in keys list.

    :param keys: A list of key
    :param dict: A python dict
    :returns: True or dict key
    """
    for key in keys:
        if key not in dict.keys():
            return key
    return True


def dict_key_value_to_str(data):
    """Convert a dict whose keys or values are unicode to string."""
    if isinstance(data, basestring):
        return str(data)
    elif isinstance(data, collections.Mapping):
        return dict(map(dict_key_value_to_str, six.iteritems(data)))
    elif isinstance(data, collections.Iterable):
        return type(data)(map(dict_key_value_to_str, data))
    else:
        return data


def get_force_ego_hosts_by_hosts_nodes(force_svc_hosts, force_nodes,
                                       host_egohost):
    """
    Convert force_hosts and force_nodes from filter_properties to force
    EGO hosts.

    :param force_sve_hosts: a list of service hosts
    :param force_nodes:     a list of nodes
    :param host_egohost:    map from service host to EGO hosts
    :return:
        []:    don't force EGO hosts; can place any EGO host
        None:  no EGO host can meet this requirement
        force_ego_hosts: force EGO hosts
    """
    force_svc_hosts = force_svc_hosts or []
    force_nodes = force_nodes or []
    host_egohost = host_egohost or {}
    if not (force_svc_hosts + force_nodes):
        return []

    ego_hosts = set()
    ego_hosts1 = set([get_rpi_res_name(n) for n in force_nodes])
    ego_hosts2 = set()
    for h in force_svc_hosts:
        ego_hosts2 |= (host_egohost.get(h, set([h])))

    if ego_hosts1 and ego_hosts2:
        ego_hosts = ego_hosts1 & ego_hosts2
    else:
        ego_hosts = ego_hosts1 | ego_hosts2

    return sorted(list(ego_hosts)) if len(ego_hosts) else None


def get_service_status_by_topic(ctxt, service_topic):
    """
    Get the status of the service whose topic is service_topic.
    """
    services = objects.ServiceList.get_all(ctxt)
    service = None
    for svc in services:
        if svc.topic == service_topic:
            service = svc
            break
    svc_status = {}
    if service:
        svc_status = {'is_up': servicegroup.API().service_is_up(service),
                      'disabled': service.disabled}
    return svc_status


def get_decision_id(instance):
    """
    To deal with the upgrade scenario, if we can't find the decision id in the
    system metadata, then fallback to the metadata.

    :param instance: an instance object
    :return string: decision id
    """
    system_metadata = utils.instance_sys_meta(instance)
    decision_id = system_metadata.get('decision_id')
    if decision_id is None:
        metadata = utils.instance_meta(instance)
        decision_id = metadata.get('decision_id')
    return decision_id


def generate_group_resreq(aggregate_id):
    """
    Generate resource group resreq.
    :param aggregate_id: aggregate id
    :return: return resource group resreq
    """

    return "%s <in> $%s" % (str(aggregate_id),
                           ego_constants.RPI_KEY_HOST_AGGREGATE_ID)


def is_zHPM_resreq(resreq):
    """
    Return True if resreq is for a zHPM instance
    """
    (imp_level, imp_value) = parse_zhpm_importance_level(resreq)
    return imp_level in ego_constants.IMPORTANCE_LEVELS
