# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""EGO Scheduler implementation."""

import collections
import six

from nova.compute import rpcapi as compute_rpcapi
from nova.compute import task_states
from nova.compute import vm_states
from nova import conductor
from nova import exception
from nova.ibm.ego import ego_constants
from nova.ibm.ego import ego_gettextutils
from nova.ibm.ego import ego_utils
from nova import objects
from nova.objects import base as base_obj
from nova.objects import instance_group as instance_group_obj
from nova import quota
from nova import rpc
from nova.scheduler import driver
from nova.scheduler import host_manager
from nova.scheduler.ibm.ego import ego_exception
from nova.scheduler.ibm.ego import ego_manager
from nova.scheduler.ibm.ego import ego_resreq_utils as res_utils
from nova.scheduler.ibm.ego.filters import numa_topology_filter
from nova.scheduler import scheduler_options
from nova import utils

from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import timeutils

_ = ego_gettextutils._
_LE = ego_gettextutils._LE
_LI = ego_gettextutils._LI
_LW = ego_gettextutils._LW

CONF = cfg.CONF
CONF.import_opt('policy_per_host_aggregate',
                'nova.ibm.ego.ego_flags')
CONF.import_opt('prs_sync_aggregate_reserveon',
                'nova.ibm.ego.ego_flags')
CONF.import_opt('prs_allocation_per_server_group',
                'nova.ibm.ego.ego_flags')

LOG = logging.getLogger(__name__)


class EGOScheduler(driver.Scheduler):
    """
    Scheduler that talk to EGO to select out the best fitting host.
    """

    def __init__(self, *args, **kwargs):
        self.notifier = rpc.get_notifier('scheduler')
        self.options = scheduler_options.SchedulerOptions()
        self.compute_rpcapi = compute_rpcapi.ComputeAPI()
        self.conductor_api = conductor.API()

        # Init EGO manager
        self.ego_manager = ego_manager.EGOManager(
                            client_name=ego_constants.EGO_ALLOC_CLIENT_NAME)

    def __schedule_prepare(self, context, request_spec, filter_properties,
                           instance_uuids=None):
        LOG.debug("The properties for filtering are %s." % filter_properties)
        LOG.debug("The request specification is %s." % request_spec)

        instance_properties = request_spec['instance_properties']
        instance_type = request_spec.get("instance_type")

        config_options = self._get_configuration_options()

        request_flavorid = instance_type['id']
        current_flavorid = instance_properties['instance_type_id']
        task_state = instance_properties['task_state']
        vm_state = instance_properties['vm_state']

        scheduler_hints = filter_properties.get('scheduler_hints') or {}
        stack_id = scheduler_hints.get('heat_root_stack_id')
        stack_name = scheduler_hints.get('heat_stack_name')
        inst_group_policies = filter_properties.get('group_policies', [])
        group_plicy = scheduler_hints.get('group_policy')
        heat_path_in_stack = scheduler_hints.get('heat_path_in_stack')
        if stack_id and stack_name and group_plicy:
            if stack_id == scheduler_hints.get('heat_stack_id'):
                # for stand alone AWS::EC2::Instance or OS::Nova::Server
                stack_name = scheduler_hints.get('heat_resource_name')
            else:
                # heat_path_in_stack is a list of tuples,
                # (stackresourcename, stackname) with list[0]
                # being (None, rootstackname), so the sub demand
                # should be the second element in this tuples
                stack_name = heat_path_in_stack[1][1].rsplit('-', 1)[0]
            stack_id = '_'.join([ego_constants.HEAT, stack_id])
            request_type = ego_constants.HEAT
            request_spec['allocation_name'] = stack_id
            request_spec['sub_demand_name'] = stack_name
        elif inst_group_policies and CONF.prs_allocation_per_server_group:
            if 'legacy' not in inst_group_policies:
                request_type = ego_constants.SERVER_GROUP
                sg = objects.InstanceGroup.get_by_instance_uuid(context,
                                                   instance_properties['uuid'])
                filter_properties['group_uuid'] = sg.uuid
            else:
                request_type = ego_constants.STAND_ALONE
        else:
            request_type = ego_constants.STAND_ALONE

        max_slot = 1
        # set up max_slot according to vm_state for resize
        if vm_state in [vm_states.STOPPED, vm_states.SUSPENDED,
                        vm_states.ERROR, vm_states.SHELVED]:
            max_slot = 0

        # set operation type if not already set

        # the value of filter working scope from two cases:
        # 1.set by HA service which will trigger instance action later
        # 2.set by rtpolicy service when call ego_get_possible_hosts before
        #   trigger instance action
        # 3.the rest of calculation according to instance task and instance
        #   state
        scope = (request_spec.get('operation_type') or
            filter_properties.get('request_spec', {}).get('operation_type'))
        if not scope:
            if vm_state == vm_states.BUILDING:
                scope = ego_constants.FILTER_WORKING_SCOPE_CREATE
            elif vm_state == vm_states.SHELVED_OFFLOADED:
                scope = ego_constants.FILTER_WORKING_SCOPE_UNSHELVE
            else:
                if task_state == task_states.MIGRATING:
                    scope = ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION
                elif task_state == task_states.RESIZE_PREP:
                    if request_flavorid and current_flavorid and (
                        request_flavorid == current_flavorid):
                        scope = (
                            ego_constants.FILTER_WORKING_SCOPE_COLD_MIGRATION)
                    else:
                        scope = ego_constants.FILTER_WORKING_SCOPE_RESIZE
                elif task_state == task_states.REBUILDING:
                    scope = ego_constants.FILTER_WORKING_SCOPE_EVACUATE
                elif task_state == task_states.SCHEDULING:
                    scope = ego_constants.FILTER_WORKING_SCOPE_CREATE
                else:
                    scope = ego_constants.FILTER_WORKING_SCOPE_RESIZE

        request_spec['operation_type'] = scope
        request_spec['request_type'] = request_type
        filter_properties.update({'context': context,
                                  'request_spec': request_spec,
                                  'config_options': config_options,
                                  'instance_type': instance_type,
                                  'revert_resize': False,
                                  'max_slot': max_slot})
        # check retry policy
        self._populate_retry(filter_properties, instance_properties)
        self.populate_filter_properties(request_spec, filter_properties)

    def select_destinations(self, context, spec_obj):
        """Selects a filtered set of hosts and nodes."""
        inst_info = self.ego_manager.get_instance_info(context,
                                                       spec_obj.instance_uuid)
        (request_spec, filter_properties) = \
            ego_utils.convert_spec_obj_to_legacy_dict(context,
                                        spec_obj, inst_info.get('inst_props'))
        # ensure only one allocation request for same instance could
        # entry select_destination function
        with lockutils.lock('select_destinations-%s' % context.request_id):
            num_instances = request_spec['num_instances']
            instance_id_name = collections.OrderedDict()
            # even if num_instances > 1, only one instance property is included
            # in request_spec, we should add other instances' uuid into
            # request_spec to create related alloc and instances' cache.
            if num_instances > 1:
                # only in instance build case, num_instances can greater
                # than 1.
                filters = {'vm_state': vm_states.BUILDING}
                insts = objects.InstanceList.get_by_filters(context, filters,
                                                            sort_dir='asc')
                for inst in insts:
                    # we can only associate these instances in same boot
                    # request with request id and the id is stored in instance
                    # action table.
                    instAction = objects.InstanceAction.get_by_request_id(
                                    context, inst.uuid, context.request_id)
                    if instAction:
                        instance_id_name[inst.uuid] = [inst.hostname,
                                                       inst.display_name]
                request_spec['instance_uuids'] = instance_id_name.keys()
            else:
                request_spec['instance_uuids'] = \
                    [request_spec['instance_properties']['uuid']]
            request_spec['instance_id_name'] = instance_id_name
            try:
                selected_hosts = self._schedule(context, request_spec,
                                        filter_properties,
                                        request_spec['instance_uuids'])
            except (ego_exception.EGOExpConnLost,
                    ego_exception.EGOExpNoAllocID,
                    ego_exception.EGOExpResizeAllocFailed,
                    ego_exception.EGOExpUpdateDecFailed,
                    exception.NoValidHost,
                    exception.InvalidRequest,
                    exception.NovaException) as ex:
                LOG.error(_LE('Select destinations failed due to: (%s)')
                          % six.text_type(ex))
                if isinstance(ex, exception.NoValidHost):
                    raise
                else:
                    err_code = ego_constants.NOT_ENOUGH_RESOURCES_CODE
                    if isinstance(ex, ego_exception.EGOExpConnLost):
                        err_code = ego_constants.LOST_CONN_CODE
                    raise exception.NoValidHost(reason=six.text_type(ex),
                                                code=err_code)

            # Couldn't fulfill the request_spec
            if len(selected_hosts) < num_instances:
                raise exception.NoValidHost(reason='')

            dests = [dict(host=host.host, nodename=host.nodename,
                          limits=host.limits) for host in selected_hosts]
            return dests

    def ego_get_possible_hosts(self, context, decid, group_list, order_list,
                               filter_properties):
        self.__schedule_prepare(context,
                                filter_properties['request_spec'],
                                filter_properties)
        try:
            possible_hosts = self.ego_manager.ego_get_possible_hosts(
                 context, decid, group_list, order_list, filter_properties)
        except ego_exception.EGOExpGetPossibleHostsFailed as exc:
            msg = (_('Platform Resource Scheduler could not retrieve '
                    'the host for instance %(name)s (%(uuid)s) due to: '
                    '%(reason)s')
                    % {'name': filter_properties["request_spec"]
                       ["instance_properties"]["display_name"],
                       'uuid': filter_properties["request_spec"]
                       ["instance_id"],
                       'reason': six.text_type(exc)})
            self.ego_manager.ego_filter_driver \
                .schedule_cache_update(filter_properties,
                                       {'error_msg': msg, 'select_host': []})
            self._schedule_error_report(context, filter_properties, True)
            raise
        finally:
            self.ego_manager.schedule_request_completed(context)

        if not possible_hosts:
            msg = (_('Platform Resource Scheduler could not find a suitable '
                     'host for instance %(name)s (%(uuid)s) '
                     'because the cluster does not have enough '
                     'resources or there is no suitable host.')
                   % {'name': filter_properties["request_spec"]
                      ["instance_properties"]["display_name"],
                      'uuid': filter_properties["request_spec"]
                      ["instance_id"]})

            self.ego_manager.ego_filter_driver \
                .schedule_cache_update(filter_properties,
                                       {'error_msg': msg, 'select_host': []})
            self._schedule_error_report(context, filter_properties, True)

        return possible_hosts

    def _schedule(self, context, request_spec, filter_properties,
                  instance_uuids=None):
        """
        :param context: OpenStack context transferred from client
        :param request_spec: VM instance request spec
        :param filter_properties: VM instance filter properties
        :param instance_uuids: VM instance uuids

        returns: Returns a list of hosts that meet the required specs
        """

        selected_hosts = []
        try:
            self.__schedule_prepare(context, request_spec, filter_properties,
                                instance_uuids)
        except exception.NoValidHost as ex:
            # Only NoValidHost can be throw here, if there is an exception,
            # its retry reached the max_attempts, just ignore the exception.
            # Make sure to not mask the compute exception.
            instance_uuid = request_spec['instance_id']
            self._schedule_error_report(context, filter_properties)
            return selected_hosts

        # Get the group
        update_group_hosts = filter_properties.get('group_updated', False)

        force_hosts = filter_properties.get('force_hosts')
        force_nodes = filter_properties.get('force_nodes')
        force_ego_hosts = res_utils.get_force_ego_hosts_by_hosts_nodes(
                                force_hosts,
                                force_nodes,
                                self.ego_manager.host_egohost)
        ego_filter_driver = self.ego_manager.ego_filter_driver
        for num, instance_uuid in enumerate(instance_uuids):
            # Instance id will be set as EGO allocation name
            request_spec['instance_id'] = instance_uuid
            request_spec['instance_list'] = instance_uuids
            if request_spec['instance_id_name']:
                request_spec['instance_properties']['hostname'] = \
                    request_spec['instance_id_name'][instance_uuid][0]
                request_spec['instance_properties']['display_name'] = \
                    request_spec['instance_id_name'][instance_uuid][1]
            operation_type = request_spec.get('operation_type')
            request_type = request_spec.get('request_type')

            # Check if need to resize the old allocation if it has ever been
            # scheduled before if RetryFilter enabled.
            retry = filter_properties.get('retry', {})
            if retry and retry['num_attempts'] > 1:
                filter_properties.update({'retry4create': True})
            if not filter_properties.get('retry4create'):
                # for the first schedule attempt, store 'force_hosts' which
                # maybe can't be imported by filter into troubleshooting cache
                if force_ego_hosts:
                    self.ego_manager.ego_filter_driver.schedule_cache_update(
                        filter_properties, {'force_hosts': force_ego_hosts})
                elif force_ego_hosts is None:
                    msg = ("the_force_node_%(nodes)s_does_not_belong_to_"
                           "the_force_host_%(hosts)s" %
                           {'hosts': '_'.join(force_hosts),
                            'nodes': '_'.join(force_nodes)})
                    self.ego_manager.ego_filter_driver.schedule_cache_update(
                        filter_properties, {'force_hosts':
                                            [msg]})

                # for the first schedule attempt, store 'source_host' if the
                # VM has source host
                source_node = filter_properties['request_spec'][
                                                'instance_properties']['node']
                if source_node:
                    self.ego_manager.ego_filter_driver.schedule_cache_update(
                      filter_properties,
                      {'source_host': res_utils.get_rpi_res_name(source_node)})

            try:
                # filter by force hosts and nodes
                if force_ego_hosts is None:
                    msg = (_LE("Scheduling instance (%(inst)s) failed because "
                             "the force node (%(nodes)s) does not belong to "
                             "the force host (%(hosts)s).") %
                             {'inst': instance_uuid,
                              'hosts': force_hosts,
                              'nodes': force_nodes})
                    LOG.error(msg)
                    raise exception.NoValidHost(reason=msg)

                if (operation_type ==
                        ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION):
                    # Migrate allocation
                    LOG.debug("Migrating EGO allocation with filter "
                              "properties %s ..." % (filter_properties))
                    alloc_host = self.ego_manager.ego_migrate_decision(context,
                        instance_uuid, '', False, filter_properties)
                    LOG.debug("Migrate EGO allocation for instance "
                              "%(instance_uuid)s to host %(alloc_host)s" %
                                {'instance_uuid': instance_uuid,
                                 'alloc_host': alloc_host})
                elif (operation_type in
                        [ego_constants.FILTER_WORKING_SCOPE_CREATE,
                        ego_constants.FILTER_WORKING_SCOPE_UNSHELVE] and not
                        filter_properties.get('retry4create', False) and not
                        self.ego_manager.inst_egohost_map.get(instance_uuid)):
                    if request_type == ego_constants.HEAT:
                        (alloc_id, alloc_decid, alloc_host) = (
                             self.ego_manager.get_placement_decision_for_heat(
                                 filter_properties))
                        LOG.debug("Created EGO allocation: %(alloc_id)s, "
                                  "%(alloc_decid)s, %(alloc_host)s" %
                                    {'alloc_id': alloc_id,
                                     'alloc_decid': alloc_decid,
                                     'alloc_host': alloc_host})
                    elif request_type == ego_constants.SERVER_GROUP:
                        # resize allocation for server group
                        (alloc_id, alloc_decid, alloc_host) = (
                             self.ego_manager.resize_alloc_for_sg(context,
                                                   filter_properties))
                        LOG.debug("Created EGO allocation: %(alloc_id)s, "
                                  "%(alloc_decid)s, %(alloc_host)s" %
                                    {'alloc_id': alloc_id,
                                     'alloc_decid': alloc_decid,
                                     'alloc_host': alloc_host})
                    else:
                        # Request allocation for stand alone instance
                        LOG.debug("Creating EGO allocation with filter "
                                  "properties %s ..." % (filter_properties))
                        (alloc_id, alloc_decid, alloc_host) = (
                             self.ego_manager.ego_create_alloc(context,
                                            filter_properties, num))
                        LOG.debug("Created EGO allocation: %(alloc_id)s, "
                                  "%(alloc_decid)s, %(alloc_host)s" %
                                    {'alloc_id': alloc_id,
                                     'alloc_decid': alloc_decid,
                                     'alloc_host': alloc_host})
                else:
                    # Resize allocation
                    LOG.debug("Resizing EGO allocation with filter "
                              "properties %s ..." % (filter_properties))
                    (alloc_id, alloc_decid, alloc_host) = (
                        self.ego_manager.ego_update_decision(context,
                                            filter_properties))
                    LOG.debug("Resized EGO allocation: %(alloc_id)s, "
                              "%(alloc_decid)s, %(alloc_host)s." %
                                {'alloc_id': alloc_id,
                                 'alloc_decid': alloc_decid,
                                 'alloc_host': alloc_host})
                ego_filter_driver.schedule_cache_update(filter_properties,
                                      {'select_host': [alloc_host]})
            except Exception as ex:
                # NOTE: we don't reraise the exception here to make sure
                #       that all instances in the request get set to
                #       error properly. For retry case, nova scheduler
                #       will not overwrite the default but keep the latest
                #       default thrown by nova compute
                if isinstance(ex, (ego_exception.EGOExpCreateAllocFailed,
                                   ego_exception.EGOExpMigrateDecisionFailure,
                                   ego_exception.EGOExpResizeAllocFailed,
                                   ego_exception.EGOExpUpdateDecFailed,
                                   ego_exception.InvalidWorkloadResourceGroup,
                                   exception.NoValidHost)):
                    ego_filter_driver.schedule_cache_update(
                                          filter_properties,
                                          {'error_msg': six.text_type(ex),
                                           'select_host': []})
                    self._schedule_error_report(context, filter_properties,
                                                True)
                # if this is the last instance which need be scheduled, raise
                # exception instantly, otherwise continue to do schedule for
                # other VM instances to get all exact scheduler errors for
                # those VM instances.
                if num == len(instance_uuids) - 1:
                    self.ego_manager.schedule_request_completed(context)
                    raise
                continue

            compute = self.ego_manager.all_egohost_attrs[alloc_host]['compute']

            if update_group_hosts is True:
                if isinstance(filter_properties['group_hosts'], list):
                    filter_properties['group_hosts'] = set(
                        filter_properties['group_hosts'])
                filter_properties['group_hosts'].add(compute['host'])

            # generate HostState object to reuse the formatted data
            host_obj = host_manager.HostState(compute['host'],
                                              compute['hypervisor_hostname'])
            flt = numa_topology_filter.NUMATopologyFilter
            host_obj.limits['numa_topology'] = flt.get_numa_limits(
                                                           compute['host'])
            selected_hosts.append(host_obj)
        self.ego_manager.schedule_request_completed(context)
        return selected_hosts

    def populate_filter_properties(self, request_spec, filter_properties):
        """Stuff things into filter_properties.  Can be overridden in a
        subclass to add more data.
        """
        # Save useful information from the request spec for filter processing:
        inst_prop = request_spec['instance_properties']
        filter_properties['project_id'] = inst_prop['project_id']
        filter_properties['os_type'] = inst_prop['os_type']

        # populate ego_host which instance belong to
        inst_id = inst_prop['uuid']
        ego_host = self.ego_manager.inst_egohost_map.get(inst_id)
        if ego_host:
            filter_properties['request_spec']['instance_properties'].update(
                                                    {'ego_host': ego_host})

    def _get_configuration_options(self):
        """
        Fetch options dictionary. Broken out for testing.
        """
        return self.options.get_configuration()

    def _max_attempts(self):
        max_attempts = CONF.scheduler_max_attempts
        if max_attempts < 1:
            raise exception.NovaException(_(
                        "Invalid value for the schedule_max_attempts "
                        "property. The value must be greater than or "
                        "equal to one."))
        return max_attempts

    def _log_compute_error(self, filter_properties, instance_uuid):
        """
        If the request contained an exception from a previous compute
        build/resize operation, log it to aid debugging
        """
        retry = filter_properties.get('retry', {})
        hosts = retry.get('hosts')
        if not hosts:
            return  # no previously attempted hosts, skip
        last_host, last_node = hosts[-1]
        msg = (_LE('An error occurred spawning the instance on host'
                 ': %(host)s (node %(node)s). '
                 'For detailed information about the failure, see the '
                 'nova-compute log file (usually '
                 'in /var/log/nova/nova-compute-%(node)s.log) '
                 'for hypervisor %(node)s.')
                % {'host': last_host,
                   'node': last_node})
        LOG.error(msg, instance_uuid=instance_uuid)
        self.ego_manager.ego_filter_driver.schedule_cache_update(
                                               filter_properties,
                                               {'host_error': msg})

    def _populate_retry(self, filter_properties, instance_properties):
        """
        Populate filter properties with history of retries for this
        request.
        """
        retry = filter_properties.setdefault(
                'retry', {
                    'num_attempts': 1,
                    'hosts': []})
        instance_uuid = instance_properties.get('uuid')
        filter_properties['request_spec']['instance_id'] = instance_uuid
        self._log_compute_error(filter_properties, instance_uuid)

        force_hosts = filter_properties.get('force_hosts', [])
        retry_hosts_len = len(retry['hosts'])
        force_hosts_len = len(force_hosts)
        force_hosts_list = ','.join(force_hosts)
        if (retry_hosts_len > 0 and force_hosts_len > 0
                     and retry_hosts_len == force_hosts_len):
            # re-scheduling is disabled.
            msg = (_("Instance creation failed on all of the 'force_hosts': "
                     "%s") % force_hosts_list)
            self.ego_manager.ego_filter_driver.schedule_cache_update(
                                                   filter_properties,
                                                   {'error_msg': msg})
            raise exception.NoValidHost(reason=msg)

    def __notify_aggregate_host_add(self, payload):
        # the host has been added into aggregate, should update related
        # cache and RPI resource accordingly
        self.ego_manager.aggregate_host_add(
            svc_host=payload['host_name'],
            aggregate_id=long(payload['aggregate_id']))

    def __notify_aggregate_host_remove(self, payload):
        # the host has been removed from aggregate, should related
        # cache and RPI resource accordingly
        self.ego_manager.aggregate_host_remove(
            svc_host=payload['host_name'],
            aggregate_id=long(payload['aggregate_id']))

    def __create_ego_resource_group(self, aggregate_id):
        try:
            res_groups_dict = self.ego_manager.ego_get_all_resource_groups()
            if aggregate_id not in res_groups_dict:
                res_req_str = ("select(%s)" %
                    res_utils.generate_group_resreq(aggregate_id))
                self.ego_manager.ego_create_resource_group(
                    aggregate_id, res_req_str)
        except ego_exception.EGOExpConnLost as exc:
            LOG.error(_LE('Platform Resource Scheduler could not '
                          'connect to EGO due to: %(reason)s')
                      % {'reason': six.text_type(exc)})
            return
        except ego_exception.EGOExpGetResourceGroupFailure as exc:
            LOG.error(_LE('Platform Resource Scheduler could not '
                          'retrieve all resource groups due to: %(reason)s')
                      % {'reason': six.text_type(exc)})
            return
        except ego_exception.EGOExpCreateResourceGroupFailure as exc:
            LOG.error(_LE("Platform Resource Scheduler could not create "
                          "new resource group '%(group_name)r "
                          "due to: %(reason)s")
                      % {'group_name': aggregate_id,
                         'reason': six.text_type(exc)})

    def __notify_aggregate_create(self, payload):
        # the aggrate has been created, should update related cache
        self.ego_manager.aggregate_create(payload)
        self.__create_ego_resource_group(str(payload['aggregate_id']))

    def __delete_ego_resource_group(self, aggregate_id):
        try:
            res_groups_dict = self.ego_manager.ego_get_all_resource_groups()
            if aggregate_id in res_groups_dict:
                self.ego_manager.ego_delete_resource_group(aggregate_id)
        except ego_exception.EGOExpConnLost as exc:
            LOG.error(_LE('Platform Resource Scheduler could not '
                          'connect to EGO due to: %(reason)s')
                      % {'reason': six.text_type(exc)})
            return
        except ego_exception.EGOExpGetResourceGroupFailure as exc:
            LOG.error(_LE('Platform Resource Scheduler could not '
                          'retrieve all resource groups due to: %(reason)s')
                      % {'reason': six.text_type(exc)})
            return
        except ego_exception.EGOExpDeleteResourceGroupFailure as exc:
            LOG.error(_LE("Platform Resource Scheduler could not delete "
                          "the resource group %(group_name)r due to: "
                          "%(reason)s")
                      % {'group_name': str(aggregate_id),
                         'reason': six.text_type(exc)})

    def __notify_aggregate_delete(self, payload):
        # the aggrate has been deleted, should update related cache
        self.ego_manager.aggregate_delete(payload)
        self.__delete_ego_resource_group(str(payload['aggregate_id']))

    def __notify_aggregate_set_metadata(self, payload):
        # the aggregate metadata has beend changed, should update
        # related cache and RPI resource for the host who belong to
        # this aggregate
        self.ego_manager.aggregate_set_metadata(payload)

    def notify_scheduler_resource_status(self, context, msg):
        """
        Notify nova scheduler take related actions based on
        notification message
        """
        LOG.debug("Resource update message %s." % (msg))

        event_type = msg['event_type']
        payload = msg['payload']
        vm_state = payload.get('state')
        task_state = payload.get('state_description')
        old_task_state = payload.get('old_task_state')
        instance_id = payload.get('instance_id')
        if not instance_id:
            instance_id = payload.get('args',
                    {}).get('instance', {}).get('uuid')
        if instance_id:
            payload['uuid'] = instance_id
        svc_host = payload.get('host')
        node = payload.get('node')
        ego_host = res_utils.get_rpi_res_name(node)
        instance_properties = payload.get('instance_properties')
        # for delete notification, there is no 'instance_properties'
        # attribute, must construct it for ego_free_alloc function
        if not instance_properties:
            instance_properties = payload
            instance_properties['vm_state'] = vm_state
            instance_properties['task_state'] = task_state
        # Construct instance properties cache
        self.ego_manager.inst_props_map[instance_id] = instance_properties
        if vm_state and vm_state == vm_states.ERROR:
            sch_cache = (
                self.ego_manager.ego_filter_driver.schedule_cache.get(
                         instance_id, {}).get(
                             ego_constants.FILTER_WORKING_SCOPE_CREATE))
            if sch_cache:
                self._schedule_error_report_for_create(context, sch_cache,
                                                       payload, instance_id)

            LOG.warning(_LW('Instance moved to error state with message '
                            'block: %s') % msg)
            if (event_type == (
                ego_constants.EVENT_TYPE_COMPUTE_INST_CREATE_START)
                  or event_type == (
                      ego_constants.EVENT_TYPE_COMPUTE_TASK_BUILD_INSTANCES)):
                self.__ego_free_instance(context, instance_id,
                                         instance_properties)
            elif (event_type == ego_constants.EVENT_TYPE_SCHEDULER_LIVE_MIG
                    and instance_properties):
                src_node = instance_properties.get('node')
                if src_node:
                    src_ego_host = res_utils.get_rpi_res_name(src_node)
                    self.__ego_migrate_decision(
                                    context, instance_id,
                                    src_ego_host,
                                    inst_props=instance_properties)
            elif event_type == (
                    ego_constants.EVENT_TYPE_COMPUTE_INST_UPDATE):
                if ((task_state == task_states.RESIZE_MIGRATING
                    and old_task_state == task_states.RESIZE_MIGRATING)
                    or (task_state == task_states.RESIZE_REVERTING
                      and old_task_state == task_states.RESIZE_REVERTING)
                    or (task_state == task_states.REBUILD_BLOCK_DEVICE_MAPPING
                      or task_state == task_states.REBUILDING)):
                    self.__ego_update_decision(context, instance_id,
                                            ego_host,
                                            inst_props=instance_properties)
                elif (not task_state and old_task_state in
                      [task_states.SPAWNING,
                       task_states.SCHEDULING,
                       task_states.BLOCK_DEVICE_MAPPING,
                       task_states.NETWORKING]):
                    self.__ego_free_instance(context, instance_id,
                                             instance_properties)
        elif vm_state is None and task_state is None:
            # Exception is thrown during action won't set state into payload
            # and the event_type will be set as the function name which the
            # exception is thrown from
            if event_type == ego_constants.EVENT_TYPE_BUILD_AND_RUN_INST:
                inst_props = payload.get(
                    'request_spec', {}).get('instance_properties', {})
                self.__ego_free_instance(context, instance_id, inst_props)
            elif event_type in [ego_constants.EVENT_TYPE_PREP_RESIZE,
                    ego_constants.EVENT_TYPE_RESIZE_INST,
                    ego_constants.EVENT_TYPE_FINISH_RESIZE,
                    ego_constants.EVENT_TYPE_FINISH_REVERT_RESIZE]:
                inst_props = payload.get('args', {}).get('instance', {})
                if event_type == ego_constants.EVENT_TYPE_PREP_RESIZE:
                    home_node = inst_props.get('node')
                    home_host = res_utils.get_rpi_res_name(home_node)
                else:
                    migration = payload.get('args', {}).get('migration', {})
                    home_host = migration.get('source_compute')
                    if event_type == ego_constants.EVENT_TYPE_FINISH_RESIZE:
                        home_host = migration.get('dest_compute')
                self.__ego_update_decision(context, instance_id,
                                        home_host,
                                        inst_props=inst_props)
            elif event_type == ego_constants.EVENT_TYPE_UNSHELVE_INST:
                LOG.debug('Platform Resource Scheduler successfully freed '
                          'the EGO allocation for the VM instance '
                          '%s.' % (instance_id))
                self.__ego_free_instance(context, instance_id,
                                         instance_properties,
                                         inst_delete=True)
            # for '_rollback_live_migration' and '_post_live_migration'
            # event_type, it cannot work now since an OSCE issue
            # (https://bugs.launchpad.net/nova/+bug/1527092)
            elif event_type in [ego_constants.EVENT_TYPE_CHECK_MIGRATE_DEST,
                   ego_constants.EVENT_TYPE_REBUILD_INSTANCE,
                   ego_constants.EVENT_TYPE_ROCKBACK_LIVE_MIG_END,
                   ego_constants.EVENT_TYPE_POST_LIVE_MIGRATION_AT_DESTINATION,
                   ego_constants.EVENT_TYPE_POST_LIVE_MIGRATION,
                   ego_constants.EVENT_TYPE_PRE_LIVE_MIGRATION]:
                inst_props = payload.get('args', {}).get('instance', {})
                home_node = inst_props.get('node')
                ego_host = res_utils.get_rpi_res_name(home_node)
                alloc_host = (
                    self.ego_manager.inst_egohost_map.get(instance_id))
                if alloc_host and ego_host and alloc_host != ego_host:
                    self.__ego_migrate_decision(context, instance_id,
                                                ego_host,
                                                inst_props=inst_props,
                                                event_type=event_type)

        # migrate EGO allocation back if get rollback notification
        elif vm_state and vm_state == vm_states.ACTIVE:
            if event_type in (
                [ego_constants.EVENT_TYPE_COMPUTE_INST_ROCKBACK_LIVE_MIG_END,
                 ego_constants.EVENT_TYPE_COMPUTE_INST_LIVE_MIG_ROLLBACK_END]):
                self.__ego_migrate_decision(context, instance_id,
                                            ego_host,
                                            inst_props=instance_properties,
                                            event_type=event_type)
            elif event_type in (
                [ego_constants.EVENT_TYPE_COMPUTE_INST_POWERON_END,
                 ego_constants.EVENT_TYPE_COMPUTE_INST_RESUME,
                 ego_constants.EVENT_TYPE_RESTORE_END]):
                # if there is an allocation for the instance, resize it
                # with max_slot=1
                # otherwise, create an allocation for the instance
                if instance_id in self.ego_manager.inst_allocid_map:
                    self.__ego_modify_alloc(context, payload, ego_host,
                                            instance_id, max_slot=1)
                else:
                    self.ego_manager.create_alloc_for_existing_inst(payload)

        elif vm_state and (vm_state in (
                               [vm_states.STOPPED,
                                vm_states.SUSPENDED,
                                vm_states.SOFT_DELETED,
                                vm_states.SHELVED])):
            if (event_type in (
                    [ego_constants.EVENT_TYPE_COMPUTE_INST_POWEROFF_END,
                     ego_constants.EVENT_TYPE_COMPUTE_INST_SUSPEND,
                     ego_constants.EVENT_TYPE_SOFT_DELETE_END,
                     ego_constants.EVENT_TYPE_COMPUTE_INST_SHELVE_END])):
                self.__ego_modify_alloc(context, payload, ego_host,
                                        instance_id, max_slot=0)

        elif vm_state and vm_state == vm_states.SHELVED_OFFLOADED:
            if (event_type == (
                    ego_constants.EVENT_TYPE_COMPUTE_INST_SHELVE_OFFLOAD_END)):
                self.__ego_free_instance(context, instance_id,
                                         instance_properties,
                                         inst_delete=True)

        # create allocation for instance
        if event_type == ego_constants.EVENT_TYPE_COMPUTE_INST_IMPORT_END:
            self.ego_manager.create_alloc_for_existing_inst(payload)
        # Delete instance finished
        elif event_type == ego_constants.EVENT_TYPE_COMPUTE_INST_DELETE_END:
            LOG.debug('Platform Resource Scheduler successfully freed '
                      'the EGO allocation for the VM instance '
                      '%s.' % (instance_id))
            self.__ego_free_instance(context, instance_id, instance_properties,
                                     inst_delete=True)

        # revert VM finished
        elif event_type == (
                ego_constants.EVENT_TYPE_COMPUTE_INST_RESIZE_REVERT_END):
            LOG.debug('Revert resize with message as %s.' % msg)
            self.__ego_update_decision(context, instance_id,
                                    ego_host,
                                    inst_props=instance_properties)
        # Only PowerVC will send this notification, which indicates that
        # the instance specification (e.g., vcpus/memory) has been changed.
        elif event_type == ego_constants.EVENT_TYPE_COMPUTE_INST_UPDATE_SPEC:
            LOG.debug('Update instance spec with message as %s.' % msg)
            htype = self.ego_manager.get_htype_from_hostname(ego_host)
            # Keep get_consume_str_by_instance consistent with other places.
            # Visit the DB to get the instance, instance metadata and
            # system_metadata.
            instance = objects.Instance.get_by_uuid(context, instance_id,
                                                    ['metadata',
                                                     'system_metadata'])
            metadata = instance.get("metadata") or {}
            payload.update({'metadata': metadata})
            consume_str = res_utils.get_consume_str_by_instance(payload, htype)
            self.__ego_update_decision(context, instance_id, ego_host,
                                       updated_consume_str=consume_str,
                                       inst_props=instance_properties)
        elif event_type in (
            [ego_constants.EVENT_TYPE_COMPUTE_INST_LIVE_MIG_PRE_START,
             ego_constants.EVENT_TYPE_COMPUTE_INST_REBUILD_END]):
            publish_id = msg['publisher_id']
            alloc_ego_host = self.ego_manager.inst_egohost_map.get(instance_id)
            host_attr = self.ego_manager.all_egohost_attrs[alloc_ego_host]
            alloc_svc_host = host_attr['compute']['host']
            dst_svc_host = publish_id[publish_id.find('.') + 1:]
            if (alloc_svc_host and dst_svc_host and
                    alloc_svc_host != dst_svc_host):
                dst_ego_hosts = self.ego_manager.host_egohost.get(dst_svc_host)
                if dst_ego_hosts:
                    self.__ego_migrate_decision(context, instance_id,
                                                list(dst_ego_hosts)[0],
                                                inst_props=instance_properties,
                                                event_type=event_type)
            if (event_type ==
                    ego_constants.EVENT_TYPE_COMPUTE_INST_REBUILD_END):
                if instance_id in (
                    self.ego_manager.ego_filter_driver.schedule_cache):
                    self.ego_manager.ego_filter_driver.schedule_cache.pop(
                                                                  instance_id)
                self.ego_manager.sync_aggregate_reserveon_for_inst(instance_id)
        elif event_type == ego_constants.EVENT_TYPE_COMPUTE_NODE_CREATE_END:
            LOG.debug('A new host %(host)s was added to OpenStack '
                      'cluster.' % {'host': svc_host})
            try:
                self.ego_manager.ego_set_resource_info(svc_host)
            except ego_exception.EGOExpConnLost as egoex:
                LOG.error(six.text_type(egoex))
            except Exception as ex:
                LOG.error(_LE('Failed to add host RPI resource to EGO from DB '
                              'due to: %(reason)s')
                          % {'reason': six.text_type(ex)})
        elif event_type == ego_constants.EVENT_TYPE_COMPUTE_NODE_UPDATE_END:
            LOG.debug('Host %(host)s was updated.' % {'host': svc_host})
            try:
                compute_node_id = payload.get('compute_node_id')
                node = objects.ComputeNode.get_by_id(context, compute_node_id)
                service = objects.Service.get_by_host_and_binary(
                                          context, node.host, 'nova-compute')
                ego_host = res_utils.get_rpi_res_name(node.hypervisor_hostname)

                self.ego_manager.ego_set_resource_info(
                    svc_host=svc_host,
                    ego_host=ego_host,
                    compute=dict(six.iteritems(node)),
                    service=dict(six.iteritems(service)))
            except ego_exception.EGOExpConnLost as egoex:
                LOG.exception(egoex)
            except Exception as ex:
                LOG.exception(ex)
                LOG.error(_LE('Failed to update host RPI resource to EGO '
                              'from the database due to: %(reason)s')
                          % {'reason': six.text_type(ex)})
        elif event_type == ego_constants.EVENT_TYPE_COMPUTE_NODE_DELETE_END:
            LOG.debug('Host %(host)s was removed from OpenStack cluster.'
                       % {'host': six.text_type(svc_host)})
            try:
                # When we boot an instance on a host and then delete the host
                # immediately, the new VM selects the host and tries to boot
                # the instance to nova compute. At this time, nova scheduler
                # receives the notification of compute.node.delete.end, which
                # will cause the nova scheduler failed to cast the boot
                # instance to nova compute. Then the VM become an orphan VM
                # and the scheduler needs to delete those orphan VMs.
                self.__remove_orphan_vms_on_host(context, ego_host)
                cstr = res_utils.generate_res_state_and_attr_str(
                                   ego_host, ego_constants.HOST_RES_REMOVED)
                self.ego_manager.ego_set_resource_info_str(cstr)
            except (ego_exception.EGOExpConnLost,
                    ego_exception.EGOExpSetResourceInfoFailed) as egoex:
                LOG.error(six.text_type(egoex))
            except Exception as ex:
                LOG.error(_LE('Failed to delete host RPI resource from EGO due'
                            ' to: %(reason)s') % {'reason': six.text_type(ex)})
        elif event_type == ego_constants.EVENT_TYPE_AGGREGATE_HOST_ADD_END:
            LOG.info(_LI('Host %(host)s was added into aggregate %(id)s.') %
                     {'host': six.text_type(payload.get('host_name')),
                      'id': six.text_type(payload.get('aggregate_id'))})
            self.__notify_aggregate_host_add(payload)
        elif event_type == ego_constants.EVENT_TYPE_AGGREGATE_HOST_REMOVE_END:
            LOG.info(_LI('Host %(host)s was removed from aggregate %(id)s.') %
                     {'host': six.text_type(payload.get('host_name')),
                      'id': six.text_type(payload.get('aggregate_id'))})
            self.__notify_aggregate_host_remove(payload)
        elif event_type == ego_constants.EVENT_TYPE_AGGREGATE_METADATA_SET_END:
            LOG.debug("Aggregate metadata update end message: %s" % msg)
            self.__notify_aggregate_set_metadata(payload)
        elif event_type == ego_constants.EVENT_TYPE_AGGREGATE_CREATE_END:
            LOG.info(_LI("Aggregate create end message: %s") % msg)
            self.__notify_aggregate_create(payload)
        elif event_type == ego_constants.EVENT_TYPE_AGGREGATE_DELETE_END:
            LOG.info(_LI("Aggregate delete end message: %s") % msg)
            self.__notify_aggregate_delete(payload)
        elif event_type == ego_constants.EVENT_TYPE_HOST_ENABLE_END:
            try:
                cstr = ''
                host_name = payload['host_name']
                ego_host_set = self.ego_manager.host_egohost.get(host_name)
                if payload.get('enabled') and ego_host_set:
                    for ego_host in ego_host_set:
                        cstr = (
                            res_utils.generate_res_state_and_attr_str(
                                                 ego_host,
                                                 ego_constants.HOST_RES_OK,
                                                 {'activeCompute': 1}))
                        self.ego_manager.ego_set_resource_info_str(cstr)
                elif ego_host_set:
                    for ego_host in ego_host_set:
                        cstr = (
                            res_utils.generate_res_state_and_attr_str(
                                                 ego_host,
                                                 ego_constants.HOST_RES_CLOSED,
                                                 {'activeCompute': 0}))
                        self.ego_manager.ego_set_resource_info_str(cstr)
            except (ego_exception.EGOExpConnLost,
                    ego_exception.EGOExpSetResourceInfoFailed) as egoex:
                LOG.error(six.text_type(egoex))
            except Exception as ex:
                LOG.error(_LE('Host status update error due to: %{reason}s')
                          % {'reason': six.text_type(ex)})
        elif event_type in (
            [ego_constants.EVENT_TYPE_COMPUTE_INST_CREATE_END,
             ego_constants.EVENT_TYPE_COMPUTE_INST_UNSHELVE_END,
             ego_constants.EVENT_TYPE_COMPUTE_INST_RESIZE_END,
             ego_constants.EVENT_TYPE_COMPUTE_INST_LIVE_MIGRATION_POST_END]):
            # When schedule operation succeeds, the schedule cache should be
            # cleaned
            if instance_id in (
                    self.ego_manager.ego_filter_driver.schedule_cache):
                self.ego_manager.ego_filter_driver.schedule_cache.pop(
                                                                  instance_id)
            self.ego_manager.sync_aggregate_reserveon_for_inst(instance_id)
            # Temp fix for defect 84046 in liberty release:
            # Re-sync up decision id between PRS cache and nova db after
            # unshelve instance.
            # (For long term, this defect will be resolved by community defect:
            # https://bugs.launchpad.net/nova/+bug/1548209)
            if (event_type ==
                    ego_constants.EVENT_TYPE_COMPUTE_INST_UNSHELVE_END):
                decid = self.ego_manager.inst_dec_map.get(instance_id)
                metadata = {'decision_id': decid}
                res_utils.update_instance_system_metadata(instance_id,
                                                        metadata)
        elif event_type == ego_constants.EVENT_TYPE_SERVERGROUP_CREATE:
            if not CONF.prs_allocation_per_server_group:
                return
            group_uuid = payload['server_group_id']
            LOG.debug('The system successfully created server group'
                      ' %(group_uuid)s.' % {'group_uuid': group_uuid})
            group = instance_group_obj.InstanceGroup.get_by_uuid(context,
                                                            group_uuid)
            if not group:
                LOG.warning(_LW('The system successfully created server group'
                           ' %(group_uuid)s not found in database.')
                         % {'group_uuid': group_uuid})
                return
            # legacy policy in group, filters is skiped, treat as standalone
            if 'legacy' in group.policies:
                return
            self.__ego_create_alloc_for_server_group(group)
        elif event_type == (
                 ego_constants.EVENT_TYPE_PRS_HA_MT_INST_AUTO_MOVE_ABORT):
            ego_host = res_utils.get_rpi_res_name(payload['node'])
            alloc_host = (
                    self.ego_manager.inst_egohost_map.get(instance_id))
            if alloc_host != ego_host:
                self.__ego_migrate_decision(context, instance_id, ego_host)
        elif event_type == ego_constants.EVENT_TYPE_SERVERGROUP_DELETE:
            if not CONF.prs_allocation_per_server_group:
                return
            group_uuid = str(payload['server_group_id'])
            alloc_name = '_'.join(['sg', group_uuid])
            LOG.debug('The system successfully deleted server group'
                      ' %(group_uuid)s.' % {'group_uuid': group_uuid})
            alloname_list = self.ego_manager.inst_allocname_map.values()
            if alloname_list.count(alloc_name) > 0:
                filter_properties = {'group_uuid': group_uuid,
                                     'clear_policy': True}
                try:
                    self.ego_manager.resize_alloc_for_sg(context,
                                                         filter_properties)
                    self.ego_manager.cleared_policy_group[group_uuid] = True
                except (ego_exception.EGOExpConnLost,
                        ego_exception.EGOExpResizeAllocFailed) as egoex:
                    LOG.error(six.text_type(egoex))
                except Exception as e:
                    LOG.error(_LE('The system could not delete the server '
                                'group due to: %(reason)s')
                              % {'reason': six.text_type(e)})
            else:
                self.ego_manager.ego_free_alloc(alloc_name)
                ego_utils.safe_remove(self.ego_manager.cleared_policy_group,
                                      group_uuid)
        self.ego_manager._iterate_extension_managers('handle_notification',
                                                     msg)

    def __ego_modify_alloc(self, context, payload, ego_host, instance_id,
                           max_slot):
        # for an OSCE bug, when delete a VERIFY_RESIZE instance,
        # "compute.instance.delete.end" notification is accepted first, so
        # allocation of the instance is deleted. But then
        # "compute.instance.power_off.end" notification of the instance is
        # also accepted, if we try to modify the instance's allocation,
        # exception would raise. For this case, we should ignore it.
        if instance_id not in self.ego_manager.inst_allocname_map:
            return
        hypervisor_type = (self.ego_manager.get_htype_from_hostname(ego_host))
        if (self.ego_manager.support_modify_allocation_quota(hypervisor_type,
                                                             instance_id)):
            self.__ego_update_decision(context, instance_id,
                                    ego_host,
                                    max_slot, inst_props=payload)

    def __ego_free_instance(self, context, instance_id,
                            instance_properties=None, inst_delete=False):
        """
        Invoke EGO manager method to free the allocated resources of the
        instance.

        :param context: nova context
        :param instanceid: instance uuid
        :param instance_properties: instance properties dict
        :param inst_delete: True if accepting instance delete end notification;
                            False otherwise.
        """
        instance_properties = instance_properties or {}
        allocname = self.ego_manager.inst_allocname_map.get(instance_id)
        # if the instance_id has no related alloc name, that means EGO has
        # free its decision (alloc for standalone instance) by
        # __ego_sync_up_instances.
        if not allocname:
            return
        alloc_type = (
            res_utils.get_alloc_type_from_allocname(allocname))
        if alloc_type == ego_constants.STAND_ALONE:
            self.ego_manager.ego_free_alloc(instance_id,
                                            instance_properties,
                                            True)
        else:
            # for heat instance, its decision should be deleted only when
            # the instance has been deleted
            if not alloc_type == ego_constants.HEAT or inst_delete:
                self.ego_manager.release_decision(instance_id,
                                                  instance_properties)

    def __ego_migrate_decision(self, context, instance_id, dest,
                               ignore_filters=True,
                               inst_props=None,
                               event_type=None):
        """
        Invoke EGO manager method to rollback the allocated resources
        when an error occurred in live-migration.
        """
        inst_props = inst_props or {}
        max_slot = self.ego_manager.inst_maxslots_map.get(instance_id, 1)
        instance_type_id = inst_props.get('instance_type_id')
        instance_type = objects.Flavor.get_by_id(context, instance_type_id)

        image_hvtype = None
        if event_type == ego_constants.EVENT_TYPE_CHECK_MIGRATE_DEST:
            sys_meta = inst_props.get('system_metadata') or {}
            image_hvtype = sys_meta.get('image_hypervisor_type')
        if (event_type in [
                ego_constants.EVENT_TYPE_COMPUTE_INST_LIVE_MIG_PRE_START,
                ego_constants.EVENT_TYPE_COMPUTE_INST_LIVE_MIG_ROLLBACK_END]):
            image_meta = inst_props.get('image_meta') or {}
            image_hvtype = image_meta.get('hypervisor_type')

        filter_properties = {'revert_resize': False,
                             'force_hosts': [dest],
                             'max_slot': max_slot,
                             'force_realloc_migrate': True,
                             'ignore_filters': ignore_filters,
                             'request_spec':
                             {'instance_id': instance_id,
                              'instance_type': instance_type,
                              'instance_properties': inst_props,
                              'image': {'properties':
                                        {'hypervisor_type': image_hvtype}}}}
        self.ego_manager.ego_update_decision(context, filter_properties)

    def __ego_update_decision(self, context, instance_id, ego_host,
                           max_slot=None,
                           ignore_filters=True,
                           updated_consume_str="",
                           inst_props=None):
        """
        Invoke EGO manager method to rollback the allocated resources
        when an error occurred in resize VM.
        :param context: nova context
        :param instance: instance uuid
        :param ego_host: the home ego host of the instance
        :param max_slot: max slot for the allocation, 0 means released
        :param ignore_filters: without go through filters if true
        :param updated_consume_str: the EGO consume string for resize
        """
        inst_props = inst_props or {}
        if max_slot is None:
            max_slot = self.ego_manager.inst_maxslots_map.get(instance_id, 1)
        LOG.debug("Platform Resource Scheduler will resize the EGO "
                  "allocation for instance %(instance_id)s to host "
                  "%(host)s with slots as %(max_slot)s" %
                    {'instance_id': instance_id,
                     'host': ego_host,
                     'max_slot': max_slot})
        instance = objects.Instance.get_by_uuid(context,
                                               instance_id,
                                               ['metadata'])
        instance_type = base_obj.obj_to_primitive(
                            instance.get_flavor())

        if not updated_consume_str:
            metadata = utils.instance_meta(instance)
            # generate consume str based on instance type
            htype = self.ego_manager.get_htype_from_hostname(ego_host)
            image = {'properties': {'hypervisor_type': htype}}
            filter_properties = {"force_hosts": [ego_host],
                                 'request_spec':
                                     {'image': image,
                                      'instance_id': instance.uuid,
                                      'instance_type': instance_type,
                                      'instance_properties': inst_props}}
            self.ego_manager._iterate_extension_managers(
                          "handle_for_schedule_info", filter_properties)
            consume_str = (
                res_utils.get_consume_str_by_instance_type(
                    instance_type, filter_properties, metadata))
        else:
            old_resreq = self.ego_manager.inst_resreq_map.get(instance_id)
            # if the allocation doesn't exist in EGO, ignore this notification
            if not old_resreq:
                return
            old_consume_str = (
                res_utils.get_criteria_str_by_resreq(old_resreq,
                                                     'consume'))
            if old_consume_str == updated_consume_str:
                # EGO allocation consumption didn't change, ignore
                # this notification.
                return
            else:
                consume_str = updated_consume_str
        LOG.debug("The old resource requirement %s." % consume_str)

        flavorid = instance_type['flavorid']

        filter_properties = {"revert_resize": True,
                             "instance_id": instance_id,
                             "old_consume_str": consume_str,
                             "force_hosts": [ego_host],
                             "flavorid": flavorid,
                             "max_slot": max_slot,
                             "force_realloc_migrate": True,
                             "ignore_filters": ignore_filters,
                             'request_spec':
                             {'instance_id': instance_id,
                              'instance_type': instance_type,
                              'instance_properties': inst_props}}
        self.ego_manager.ego_update_decision(context, filter_properties)

    def __remove_orphan_vms_on_host(self, context, ego_host):
        instance_list = []
        for inst_id, hostname in self.ego_manager.inst_egohost_map.items():
            if hostname == ego_host:
                inst = objects.Instance.get_by_uuid(context, inst_id)
                instance_list.append(inst)

        for instance in instance_list:
            reserv = quota.QUOTAS.reserve(context,
                                          project_id=instance['project_id'],
                                          instances=-1,
                                          cores=-instance['vcpus'],
                                          ram=-instance['memory_mb'])
            try:
                now = timeutils.utcnow()
                self.conductor_api.instance_update(context,
                                                   instance['uuid'],
                                                   vm_state='deleted',
                                                   task_state=None,
                                                   terminated_at=now)
                self.conductor_api.instance_destroy(context, instance)
                quota.QUOTAS.commit(context, reserv, instance['project_id'])
                self.__ego_free_instance(context, instance['uuid'])
            except exception.InstanceNotFound:
                quota.QUOTAS.rollback(context, reserv, instance['project_id'])
            except Exception as ex:
                LOG.error(_LE('Removing orphan VMs on %(host)s failed due '
                              'to: %(reason)s') % {'host': ego_host,
                                                   'reason':
                                                        six.text_type(ex)})

    def _schedule_error_report(self, context, filter_properties,
                                 is_schedule_error=False):
        """
        Log the conrresponding schedule cache info into log file
        with OpenStack output style

        :param is_schedule_error: if True, the instance scheduling operation
                                  failed for schedule error, otherwise maximun
                                  scheduling attempt exceeding
        """
        spec = filter_properties.get('request_spec', {})
        instanceid = spec.get('instance_id', '')
        scope = spec.get('operation_type', '')
        tshoot = self.ego_manager.ego_filter_driver.schedule_cache.get(
                                                     instanceid, {}).get(scope)
        if not (instanceid and scope and tshoot):
            return

        report = ''
        source_host = tshoot.get('source_host')
        force_hosts = tshoot.get('force_hosts')
        for attempt in range(len(tshoot), 0, -1):
            tshoot_current = tshoot.get(attempt, {})
            if not tshoot_current:
                continue
            tshoot_entry = []
            tshoot_entry.append(['Instance ID', instanceid])
            tshoot_entry.append(['Scheduled Operation', scope])
            tshoot_entry.append(['Scheduling Attempt', attempt])
            error_msg = tshoot_current.get('error_msg', '')
            tshoot_entry.append(['Error Message', error_msg.replace("$", "")])
            if source_host:
                tshoot_entry.append(['Source Host', source_host])
            if is_schedule_error:
                tshoot_entry.append(['Available Hosts',
                    self.__ego_get_resource_list('',
                                             ego_constants.HOST_RES_OK)])

            consume_req = tshoot_current.get('consume')
            if consume_req:
                old_consume = None
                if scope == ego_constants.FILTER_WORKING_SCOPE_RESIZE:
                    old_consume = {'source_host': source_host,
                                   'old_instance_type':
                      filter_properties['request_spec']['instance_properties']}
                rpistat_consume = res_utils.get_rpistat_resreq(
                                           consume_req, 'consume', old_consume)
                if is_schedule_error:
                    host_list = self.__ego_get_resource_list(rpistat_consume)
                    rpistat_consume = '\n'.join([rpistat_consume,
                                                 six.text_type(host_list)])
                tshoot_entry.append(['Resources Required', rpistat_consume])

            reserveon_req = tshoot_current.get('reserveon')
            if reserveon_req:
                rpistat_reserveon = res_utils.get_rpistat_resreq(
                                                reserveon_req, 'reserveon')
                host_list = ''
                if is_schedule_error:
                    host_list = self.__ego_get_resource_list(rpistat_reserveon)
                if not rpistat_reserveon:
                    rpistat_reserveon = '()'
                rpistat_reserveon = (
                    '\n'.join([rpistat_reserveon, six.text_type(host_list)]))
                tshoot_entry.append(['Resource Groups Required',
                                    rpistat_reserveon])

            if force_hosts:
                rpistat_force_hosts = (
                    res_utils.get_affinity_hosts(
                            force_node_list=force_hosts,
                            host_egohost=self.ego_manager.host_egohost))
                host_list = ''
                if is_schedule_error:
                    host_list = self.__ego_get_resource_list(
                                             rpistat_force_hosts)
                rpistat_force_hosts = (
                    '\n'.join([rpistat_force_hosts, six.text_type(host_list)]))
                tshoot_entry.append(['Force Hosts Required',
                                    rpistat_force_hosts])

            if 'filter_reqs' in tshoot_current:
                for filter_name, filter_req in (
                        six.iteritems(tshoot_current['filter_reqs'])):
                    rpistat_filter = res_utils.get_rpistat_filter_str(
                                          filter_req,
                                          self.ego_manager.host_egohost)
                    host_list = ''
                    if is_schedule_error:
                        host_list = self.__ego_get_resource_list(
                                             rpistat_filter)
                    if not rpistat_filter:
                        rpistat_filter = '()'
                    rpistat_filter = (
                        '\n'.join([rpistat_filter, six.text_type(host_list)]))
                    tshoot_entry.append([filter_name,
                                         rpistat_filter.replace("$", "")])

            resreq = tshoot_current.get('resreq')
            if resreq:
                tshoot_entry.append(['Total Instance Requirements',
                                     resreq.replace("$", "")])

            if 'select_host' in tshoot_current:
                select_host = tshoot_current.get('select_host')
                tshoot_entry.append(['Deployed Host',
                                    six.text_type(select_host)])

            is_schedule_error = False

            vm_name = filter_properties.get('request_spec', {}).get(
                          'instance_properties', {}).get('display_name')
            keyword = (_('Troubleshooting instance %(instanceid)s '
                         '(%(vm_name)s) for attempt %(attempt)s:')
                       % {'instanceid': instanceid,
                          'vm_name': vm_name,
                          'attempt': str(attempt)})
            report_table = (six.text_type(
                       res_utils.get_report_table(tshoot_entry)))
            report = '\n'.join([six.text_type(report), keyword, report_table])
        # clean the instanceid schedule cache
        if instanceid in (
               self.ego_manager.ego_filter_driver.schedule_cache):
            self.ego_manager.ego_filter_driver.schedule_cache.pop(
                                                              instanceid)

        LOG.error(report)

    def __ego_get_resource_list(self, cstr, state=None):
        try:
            return [node['name'] for node in
                        self.ego_manager.ego_get_resource_list(cstr, state)]
        except (ego_exception.EGOExpConnLost,
                ego_exception.EGOExpGetResourceInfoFailed):
            return _('Error: EGO not available')

    def __ego_create_alloc_for_server_group(self, group):
        policy = res_utils.build_policy_from_group(group)
        sub_demands = [{'name': str(group.uuid),
                        'resreq': '',
                        'extra_resreq': '',
                        'maxslots': 0,
                        'force_realloc_migrate': False,
                        'policy': policy}]
        self.ego_manager.create_alloc_for_sg(str(group.uuid), sub_demands)

    def _schedule_error_report_for_create(self, context, sch_cache, payload,
                                          instance_id):
        """
        Show troubleshooting report when VM encounters a boot error due to:
        1. compute error.
        2. schedule attempt exceeds CONF.scheduler_max_attempts.
        3. scheduler extension operation timeout error.
        """

        if ('force_hosts' in sch_cache) or (CONF.scheduler_max_attempts == 1):
            # For VM boot case when use --availability-zone to specify
            # force_hosts/force_nodes to or CONF.scheduler_max_attempts == 1,
            # scheduler will only schedule once and sch_cache has one schedule
            # history data.
            select_host = sch_cache[1].get('select_host', [])
            if select_host:
                alloc_host = select_host[0]
                compute = self.ego_manager.\
                    all_egohost_attrs[alloc_host]['compute']
                host = compute['host']
                node = compute['hypervisor_hostname']
                msg = (_('An error occurred spawning the instance on host: '
                         '%(host)s (node %(node)s). '
                         'For detailed information about the failure, '
                         'see the nova-compute log file (usually '
                         'in /var/log/nova/nova-compute-%(node)s.log) for '
                         'hypervisor %(node)s.')
                      % {'host': host,
                         'node': node})
            else:
                # For scheduler extension timeout error,select_host will not
                # be set into cache.
                msg = (_('Platform Resource Scheduler Extension '
                         'operation failed, please check related service.'))
            filter_properties = {
                'request_spec': {
                    'operation_type':
                    ego_constants.FILTER_WORKING_SCOPE_CREATE,
                    'instance_properties': payload,
                    'instance_id': instance_id},
                'retry': {'num_attempts': 1}}
            self.ego_manager.ego_filter_driver.schedule_cache_update(
                                                        filter_properties,
                                                        {'error_msg': msg})
            self._schedule_error_report(context, filter_properties)

        else:
            num_att = len(sch_cache)
            if num_att == CONF.scheduler_max_attempts:
                err_reason = payload.get('reason', {}).get(
                    'kwargs', {}).get('reason', {})
                if err_reason:
                    ego_host = sch_cache[num_att]['select_host'][0]
                    last_host = None
                    for host, ego_hosts in six.iteritems(
                                             self.ego_manager.host_egohost):
                        if ego_host in ego_hosts:
                            last_host = host
                            break
                    # In most cases except vmware case, ego host name is same
                    # as hypervisor node name, so use ego host name as
                    # hypervisor node name here is a low priority issue.
                    msg = (_('Error from last host %(last_host)s '
                             '(node %(last_node)s): %(err_ex)s')
                           % {'last_host': last_host,
                              'last_node': ego_host,
                              'err_ex': err_reason})
                    filter_properties = {
                        'request_spec': {
                            'operation_type':
                                ego_constants.FILTER_WORKING_SCOPE_CREATE,
                            'instance_properties':
                                payload['instance_properties'],
                            'instance_id': instance_id},
                        'retry': {'num_attempts':
                                      CONF.scheduler_max_attempts + 1}}
                    self.ego_manager.ego_filter_driver.schedule_cache_update(
                                               filter_properties,
                                               {'host_error': msg})
                    msg = (_("Platform Resource Scheduler has exceeded the "
                             "maximum scheduling attempts of %(max_attempts)d "
                             "for instance %(instance_uuid)s.")
                           % {'max_attempts':
                                  CONF.scheduler_max_attempts,
                              'instance_uuid': instance_id})
                    self.ego_manager.ego_filter_driver.schedule_cache_update(
                                                           filter_properties,
                                                           {'error_msg': msg})
                    self._schedule_error_report(context, filter_properties)
