# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""EGO scheduler base filter definition."""

import six

from nova import filters
from nova.ibm.ego import ego_constants
from nova.scheduler import filters as nova_filter
from nova.scheduler.ibm.ego import ego_resreq_utils as res_utils


class EGOBaseFilter(nova_filter.BaseHostFilter):
    """Base class from which all filters should extend."""

    def __init__(self, *args, **kwargs):
        super(EGOBaseFilter, self).__init__(*args, **kwargs)
        self.filter_scope = ego_constants.FILTER_WORKING_SCOPE_ALL
        self.filter_resreq = {'select': [], 'samehost': [], 'diffhost': []}
        self.filter_hosts = {'force_hosts': None, 'ignore_hosts': []}
        self.filter_error_message = ego_constants.FILTER_ERROR_MESSAGE
        self.extra_resreq_for_create = True

    def _set_filter_scope(self, filter_scope):
        """
        Specifies the operations (e.g., create, live migrate, etc.) to which
        this filter applies.
        """
        self.filter_scope = filter_scope

    def parse(self, filter_properties, all_egohost_attrs):
        raise NotImplementedError()

    def _full_name(self):
        """
        module.classname of the filter.
        """
        return "%s.%s" % (self.__module__, self.__class__.__name__)

    def _include_hosts(self, host_list):
        """
        define force host list for VM placement/replacement.
        """
        self.filter_hosts['force_hosts'] = host_list

    def _exclude_hosts(self, host_list):
        """
        define ignore host list for VM placement/replacement.
        """
        self.filter_hosts['ignore_hosts'] = host_list

    def _select(self, select_list):
        """
        define 'select' statements of EGO resreq, it can be JSON
        resource querying syntax or EGO resource querying syntax
        """
        if isinstance(select_list, six.string_types):
            self.filter_resreq['select'] = [select_list]
        else:
            for select in select_list:
                self.filter_resreq['select'].append(select)

    def _is_verified_hypervisor(self, filter_properties, hy_type):
        """
        Helper utility to determine if the operation is allowed for the
        specified hypervisor(s).
        """
        if isinstance(hy_type, six.string_types):
            hy_type = [hy_type]

        # For VMware ESXi case
        request_spec = filter_properties.get('request_spec', {})
        image = request_spec.get('image', {})
        prs_hyper_type = image.get('properties', {}).get('prs_hypervisor_type')
        if prs_hyper_type is not None:
            return prs_hyper_type in hy_type
        else:
            hypervisor_data = res_utils.get_hypervisor_data(filter_properties)
            return hypervisor_data['hypervisor_type'] in hy_type

    def _set_error_message(self, message):
        """
        Define a message for the case in which no valid host can be found for
        VM placement/replacement.
        """
        self.filter_error_message = message

    def _set_filter_extra_resreq_for_create(self, filter_persist):
        """
        Define whether to persist a filter in create scope.  If it is desired
        to *not* persist a filter as part of the resreq, set this to False.
        """
        self.extra_resreq_for_create = filter_persist


class HostFilterHandler(filters.BaseFilterHandler):
    def __init__(self):
        super(HostFilterHandler, self).__init__(EGOBaseFilter)


def all_filters():

    """Return a list of filter classes found in this directory.

    This method is used as the default for available scheduler filters
    and should return a list of all filter classes available.
    """
    return HostFilterHandler().get_all_classes()
