# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2014 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
import netaddr
import six

from nova import context as nova_context
from nova import exception
from nova.ibm.ego import ego_gettextutils
from nova import objects
from nova.scheduler.ibm.ego import ego_resreq_utils
from nova.scheduler.ibm.ego import filters

from oslo_log import log as logging

_ = ego_gettextutils._
_LW = ego_gettextutils._LW

LOG = logging.getLogger(__name__)


class SameHostFilter(filters.EGOBaseFilter):
    # The hosts the instances are running on doesn't change within a request
    run_filter_once_per_request = True

    def __init__(self):
        super(SameHostFilter, self).__init__()
        self.context = nova_context.get_admin_context()

    def parse(self, filter_properties, all_egohost_attrs):
        """
        Convert SameHostFilter requirement to EGO res req
        """
        samehost_list = []
        scheduler_hints = filter_properties.get('scheduler_hints') or {}
        affinity_uuids = scheduler_hints.get('same_host', [])
        if isinstance(affinity_uuids, six.string_types):
            affinity_uuids = [affinity_uuids]

        for uuid in affinity_uuids:
            try:
                instance = objects.Instance.get_by_uuid(self.context, uuid)
            except exception.InstanceNotFound:
                LOG.warning(_LW("VM instance with UUID %s cannot be found, "
                           "same host filter will ignore this instance.")
                           % uuid)
                continue
            decid = ego_resreq_utils.get_decision_id(instance)
            req_str = str("'" + decid + "'")
            samehost_list.append(req_str)
        self.filter_resreq['samehost'] = samehost_list
        LOG.debug("After passing the SameHostFilter filter, "
                  "the resource requirement expression is (%s)."
                    % self.filter_resreq['samehost'])


class DifferentHostFilter(filters.EGOBaseFilter):
    # The hosts the instances are running on doesn't change within a request
    run_filter_once_per_request = True

    def __init__(self):
        super(DifferentHostFilter, self).__init__()
        self.context = nova_context.get_admin_context()

    def parse(self, filter_properties, all_egohost_attrs):
        """
        Convert DifferentHostFilter requirement to EGO res req
        """
        diffhost_list = []
        scheduler_hints = filter_properties.get('scheduler_hints') or {}
        affinity_uuids = scheduler_hints.get('different_host', [])
        if isinstance(affinity_uuids, six.string_types):
            affinity_uuids = [affinity_uuids]

        for uuid in affinity_uuids:
            try:
                instance = objects.Instance.get_by_uuid(self.context, uuid)
            except exception.InstanceNotFound:
                LOG.warning(_LW("VM instance with UUID %s cannot be found, "
                           "different host filter will ignore this instance.")
                           % uuid)
                continue
            decid = ego_resreq_utils.get_decision_id(instance)
            req_str = str("'" + decid + "'")
            diffhost_list.append(req_str)
        self.filter_resreq['diffhost'] = diffhost_list
        LOG.debug("After passing the DifferentHostFilter filter, "
                  "the resource requirement expression is (%s)."
                    % self.filter_resreq['diffhost'])


class SimpleCIDRAffinityFilter(filters.EGOBaseFilter):
    # The address of a host doesn't change within a request
    run_filter_once_per_request = True

    def __init__(self):
        super(SimpleCIDRAffinityFilter, self).__init__()

    def parse(self, filter_properties, all_egohost_attrs):
        """
        Convert SimpleCIDRAffinityFilter requirement to EGO res req
        """
        scheduler_hints = filter_properties.get('scheduler_hints') or {}
        affinity_cidr = scheduler_hints.get('cidr', '/24')
        affinity_host_addr = scheduler_hints.get('build_near_host_ip')
        if not affinity_host_addr:
            return

        affinity_net = netaddr.IPNetwork(str.join('', (affinity_host_addr,
                                                       affinity_cidr)))
        req_str = "$host_ip" + affinity_cidr + "==" + str(affinity_net.network)
        self._select(req_str)
        LOG.debug("After passing the SimpleCIDRAffinityFilter filter, "
                  "the resource requirement expression is (%s)."
                    % req_str)


class _GroupAntiAffinityFilter(filters.EGOBaseFilter):
    """Schedule the instance on a different host from a set of group
    hosts.
    """
    def __init__(self):
        super(_GroupAntiAffinityFilter, self).__init__()

    def parse(self, filter_properties, all_egohost_attrs):
        # Only invoke the filter is 'anti-affinity' is configured
        policies = filter_properties.get('group_policies', [])
        if self.policy_name not in policies:
            return

        group_hosts = filter_properties.get('group_hosts', [])
        # exclude instance's host in group_hosts for resize
        instance_host = filter_properties.get(
                'request_spec', {}).get(
                        'instance_properties', {}).get('host', None)
        if instance_host:
            group_hosts = set(group_hosts) - set([instance_host])
        if group_hosts:
            self._exclude_hosts(list(group_hosts))


class _GroupAffinityFilter(filters.EGOBaseFilter):
    """Schedule the instance on a different host from a set of group
    hosts.
    """
    def __init__(self):
        super(_GroupAffinityFilter, self).__init__()

    def parse(self, filter_properties, all_egohost_attrs):
        # Only invoke the filter is 'affinity' is configured
        policies = filter_properties.get('group_policies', [])
        if self.policy_name not in policies:
            return

        group_hosts = filter_properties.get('group_hosts', [])
        if group_hosts:
            self._include_hosts(list(group_hosts))


class GroupAffinityFilter(_GroupAffinityFilter):
    def __init__(self):
        super(GroupAffinityFilter, self).__init__()
        self.policy_name = 'legacy'


class ServerGroupAffinityFilter(_GroupAffinityFilter):
    def __init__(self):
        super(ServerGroupAffinityFilter, self).__init__()
        self.policy_name = 'affinity'


class GroupAntiAffinityFilter(_GroupAntiAffinityFilter):
    def __init__(self):
        super(GroupAntiAffinityFilter, self).__init__()
        self.policy_name = 'legacy'


class ServerGroupAntiAffinityFilter(_GroupAntiAffinityFilter):
    def __init__(self):
        super(ServerGroupAntiAffinityFilter, self).__init__()
        self.policy_name = 'anti-affinity'
