# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from nova.ibm.ego import ego_gettextutils
from nova.scheduler.ibm.ego import filters

from oslo_config import cfg
from oslo_log import log as logging

_ = ego_gettextutils._

LOG = logging.getLogger(__name__)

CONF = cfg.CONF

CONF.import_opt('isolated_images',
                'nova.scheduler.filters.isolated_hosts_filter')
CONF.import_opt('isolated_hosts',
                'nova.scheduler.filters.isolated_hosts_filter')
CONF.import_opt('restrict_isolated_hosts_to_isolated_images',
                'nova.scheduler.filters.isolated_hosts_filter')


class IsolatedHostsFilter(filters.EGOBaseFilter):
    """Keep specified images to selected hosts."""

    # The configurtaion values do not change within a request
    run_filter_once_per_request = True

    def __init__(self):
        super(IsolatedHostsFilter, self).__init__()

    def parse(self, filter_properties, all_egohost_attrs):
        """
        Convert IsolatedHostsFilter requirement to EGO res req

        Reference from openstack.org:

        Result Matrix with 'restrict_isolated_hosts_to_isolated_images' set
        to True:
                     | isolated_image | non_isolated_image
        -------------+----------------+-------------------
        iso_host     |    True        |     False
        non_iso_host |    False       |      True

        Result Matrix with 'restrict_isolated_hosts_to_isolated_images' set
        to False:
                     | isolated_image | non_isolated_image
        -------------+----------------+-------------------
        iso_host     |    True        |      True
        non_iso_host |    False       |      True
        """
        # If the configuration does not list any hosts, the filter will always
        # return True, assuming a configuration error, so letting all hosts
        # through.
        restrict_flag = CONF.restrict_isolated_hosts_to_isolated_images

        if not any(CONF.isolated_images) or not any(CONF.isolated_hosts):
            return

        spec = filter_properties.get('request_spec', {})
        props = spec.get('instance_properties', {})
        image_ref = props.get('image_ref')
        image_isolated = image_ref in CONF.isolated_images

        if restrict_flag:
            if image_isolated:
                reqstr = self.get_reqstr(CONF.isolated_hosts)
            else:
                reqstr = self.get_reqstr(CONF.isolated_hosts, False)
        else:
            if image_isolated:
                reqstr = self.get_reqstr(CONF.isolated_hosts)
            else:
                return

        self._select(reqstr)
        LOG.debug("After passing the IsolatedHostsFilter filter, "
                  "the resource requirement expression is (%s)."
                  % reqstr)

    def get_reqstr(self, host_list, in_range=True):
        reqstr = ''
        if host_list is None or len(host_list) == 0:
            return reqstr

        index = 0
        if in_range:
            for host in host_list:
                if index == 0:
                    reqstr = "'" + host + "'"
                else:
                    reqstr = reqstr + "||" + "'" + host + "'"
                index = index + 1
        else:
            for host in CONF.isolated_hosts:
                if index == 0:
                    reqstr = "!" + "'" + host + "'"
                else:
                    reqstr = reqstr + "&& !" + "'" + host + "'"
                index = index + 1

        return reqstr
