# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from nova.pci import stats
from nova.scheduler.ibm.ego import filters

from oslo_log import log as logging

LOG = logging.getLogger(__name__)


class PciPassthroughFilter(filters.EGOBaseFilter):
    """Pci Passthrough Filter based on PCI request

    Filter that schedules instances on a host if the host has devices
    to meet the device requests in the 'extra_specs' for the flavor.

    PCI resource tracker provides updated summary information about the
    PCI devices for each host, like::

        | [{"count": 5, "vendor_id": "8086", "product_id": "1520",
        |   "extra_info":'{}'}],

    and VM requests PCI devices via PCI requests, like::

        | [{"count": 1, "vendor_id": "8086", "product_id": "1520",}].

    The filter checks if the host passes or not based on this information.
    """

    def __init__(self):
        super(PciPassthroughFilter, self).__init__()

    def parse(self, filter_properties, all_egohost_attrs):
        """Return true if the host has the required PCI devices."""
        pci_requests = filter_properties.get('pci_requests')

        includes = []
        excludes = []

        for (ego_host, state) in all_egohost_attrs.items():
            if not pci_requests:
                includes.append(ego_host)
                continue

            pci_stats = stats.PciDeviceStats(state['pci_stats'])

            if not pci_stats.support_requests(pci_requests.requests):
                excludes.append(ego_host)
                LOG.debug("%(host_state)s doesn't have the required PCI "
                          "devices (%(requests)s)",
                          {'host_state': ego_host, 'requests': pci_requests})
                continue
            includes.append(ego_host)

        if includes:
            self._include_hosts(includes)
        if excludes:
            self._exclude_hosts(excludes)
