# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
import six

from nova import exception
from nova.ibm.ego import ego_constants
from nova.ibm.ego import ego_gettextutils
from nova import objects
from nova.scheduler.ibm.ego import filters

from oslo_log import log as logging

_ = ego_gettextutils._
_LW = ego_gettextutils._LW

LOG = logging.getLogger(__name__)


class PowerLMBSizeFilter(filters.EGOBaseFilter):
    """
    This filter ensures that the logical memory block (LMB) size on the source
    equals to destination host as this is a requirement for live migration
    """

    def __init__(self):
        super(PowerLMBSizeFilter, self).__init__()
        self._set_filter_scope(
            [ego_constants.FILTER_WORKING_SCOPE_CREATE,
             ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_EVACUATE])
        self._set_filter_extra_resreq_for_create(False)

    def parse(self, filter_properties, all_egohost_attrs):
        def generate_expr(mem):
            return '(%s %% $lmb_size) == 0' % str(mem)

        if not self._is_verified_hypervisor(filter_properties,
                            list(ego_constants.POWERVM_HYPERVISOR_TYPES)):
            LOG.debug("Ignoring PowerLMBSizeFilter...")
            return
        request_spec = filter_properties.get('request_spec', {})
        operation = request_spec.get('operation_type')

        if operation == ego_constants.FILTER_WORKING_SCOPE_CREATE:
            # min/desired/max memory should be a multiple of host's LMB size
            inst_type = request_spec['instance_type']
            inst_type_id = inst_type['id']
            flavor_data = objects.Flavor.get_by_id(
                                filter_properties['context'], inst_type_id)
            desired_mem = flavor_data.get('memory_mb')
            extra_specs = flavor_data.get('extra_specs', {})
            min_mem = extra_specs.get('powervm:min_memory_mb', desired_mem)
            max_mem = extra_specs.get('powervm:max_memory_mb', desired_mem)

            expressions = [generate_expr(mem) for mem in
                           sorted(list(set([min_mem, desired_mem, max_mem])))]
            req_str = '&&'.join(expressions)
            self._select(req_str)
            LOG.debug("After passing the PowerLMBSizeFilter filter, "
                      "the resource requirement expression "
                      "is (%s)." % req_str)
        else:
            instance_properties = request_spec.get('instance_properties', {})
            ego_host = instance_properties.get('ego_host')
            # source host and destination host must have the same LMB
            for hostname, host_data in six.iteritems(all_egohost_attrs):
                if ego_host == hostname:
                    lmb_size = host_data.get('lmb_size')
                    if not lmb_size:
                        if (operation ==
                                ego_constants.FILTER_WORKING_SCOPE_EVACUATE):
                            # If this is evacuate and could not get LMB size,
                            # host data could be corrupt, so continue on with
                            # evacuate and let the driver catch any errors.
                            LOG.warning(_LW("lmb_size not found for host (%s) "
                                            "during evacuate operation.") %
                                        ego_host)
                            return
                        else:
                            msg = (_('lmb_size not found for host (%s)') %
                                ego_host)
                            raise exception.MigrationPreCheckError(msg)
                    req_str = "$lmb_size==" + str(lmb_size)
                    self._select(req_str)
                    LOG.debug("After passing the PowerLMBSizeFilter filter, "
                              "the resource requirement expression "
                              "is (%s)." % req_str)
                    break
