# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from powervc_oslo.config import data_mgr as mgr
from powervc_oslo.config import data_utils as utils
from powervc_oslo.config.data_utils import V120, V121, V122, V123,\
    V130, V131, V1312
from powervc_ceilometer.config import ceilometer_config
from powervc_ceilometer import _
from oslo_db.sqlalchemy import session as db_session
from oslo_config import cfg
from ceilometer.storage.sqlalchemy import models
import yaml

NOVA_CONF = '/etc/nova/nova.conf'
CA_CERTS = '/etc/pki/tls/certs/powervc.crt'
PASTE_CONF = '/etc/ceilometer/api_paste.ini'
CEILO_CONF = '/etc/ceilometer/ceilometer.conf'
CEILOMETER_EVENT_PIPELINE = '/etc/ceilometer/event_pipeline.yaml'
SECTION_DB = 'database'
SECTION_DEFAULT = 'DEFAULT'
DB_STOP_COMMAND = '/opt/ibm/powervc/bin/powervc-services db2 stop'
DB_START_COMMAND = '/opt/ibm/powervc/bin/powervc-services db2 start'
AUDIT_EVENT_TYPES = ["audit.http.request", "audit.http.response"]


######################################################
#        Ceilometer Data Manager Definition          #
######################################################
class CeilometerDataManager(mgr.BaseDataManager):

    def get_component_name(self):
        """Provides the name of the Component providing the Data Manager"""
        return 'ceilometer'

    def get_external_name(self):
        """Provides the external name of the component """
        return 'metering'

    def get_config_desc(self):
        """Provides the description of the component configuration"""
        return _('Manage metering-related settings')

    def is_configurable(self):
        """Provides info on whether config changes are allowed"""
        return True

    def is_auditable(self):
        """Is audit support is available for this component"""
        return True

    def is_debuggable(self):
        """Is debug config support available for this component"""
        return True

    def is_supported_environment(self):
        """Specifies whether the Component is supported in this Environment"""
        return True

    def get_dependent_managers(self):
        """Specifies the Component Name of the Managers it is dependent on"""
        return ['db', 'rabbitmq', 'httpd', 'keystone']

    ######################################################
    #  Data Manager - powervc-config cmd parser  ######
    ######################################################
    def invoke_config_parser(self, subparser):
        """Invokes the parser for each component for powervc-config cmd"""
        metering_parser = super(
            CeilometerDataManager,
            self).invoke_config_parser(subparser)
        config = self._get_added_conf_properties()
        ceilometer_config.setup_parser(metering_parser, config)

    ###############################################
    #     Data Manager - Configure Interface      #
    ###############################################
    def create_database(self):
        """Creates the Database for the given Component"""
        super(CeilometerDataManager, self).create_database()
        return

    def set_user_passwords(self):
        """Sets/Generates the Passwords for the Component's Users"""
        super(CeilometerDataManager, self).set_user_passwords()
        # Add the Telemetry password to those that are set in the Config
        cfgdata = utils.get_config_data()
        if not cfgdata.get('telemetry_password'):
            cfgdata['telemetry_password'] = utils.generate_password()

    def configure_component(self):
        """Performs Configuration specific to the given Component"""
        super(CeilometerDataManager, self).configure_component()
        # Do the KVM Host-specific Nova Configuration
        if utils.is_management_system():
            offering = utils.get_offering()
            pipeline_file = '/etc/ceilometer/pipeline.yaml'
            otheroff = 'powerkvm' if offering == 'powervm' else 'powervm'
            components = dict(nova='compute', cinder='volume',
                              glance='image', neutron='network',
                              ceilometer='metering')
            # Copy over the PowerVC YAML file for the given Offering
            utils.remove_file('/etc/ceilometer/pipeline.yaml.' + otheroff)
            utils.move_file('/etc/ceilometer/pipeline.yaml.' + offering,
                            '/etc/ceilometer/pipeline.yaml')
            utils.chmod(pipeline_file, 0o640)
            utils.chown(pipeline_file, 'root', 'ceilometer')
            # Create a specific Audit Mapping File for each component
            for comp in components.keys():
                audit_file = '/etc/%s/api_audit_map.conf' % comp
                base_file = '/etc/pycadf/%s_api_audit_map.conf' % comp
                # Create a Component-specific file tailed to the Component
                utils.copy_file(base_file, audit_file)
                utils.set_openstack_config_value(
                    audit_file, 'target_endpoint_type', components[comp])
                # Replace 'telemetry = service/telemetry' with
                # 'metering = service/metering'
                # Should remove once we get this updated in the community
                if comp == 'ceilometer':
                    utils.del_openstack_config_value(audit_file, 'telemetry',
                                                     'service_endpoints')
                    utils.set_openstack_config_value(audit_file,
                                                     components[comp],
                                                     'service/metering',
                                                     'service_endpoints')
                # Set the Owner and the Permissions on the Audit Map File
                utils.chmod(audit_file, 0o640)
                utils.chown(audit_file, 'root', comp)

    ###############################################
    # ####  Data Manager - Upgrade Interface  ######
    ###############################################
    def upgrade_conf_files(self, version):
        """Performs the upgrade of the Component's CONF Files"""
        # Update all host nova.conf to disable instance.usage.audit
        # notification
        if utils.compare_version(version, V123) >= 0:
            files = utils.get_matching_files('/etc/nova', 'nova-', '.conf')
            for afile in (files + ['/etc/nova/nova.conf']):
                utils.set_openstack_config_value(
                    afile, 'instance_usage_audit', 'False', SECTION_DEFAULT)
        # Now call the Super-class to Update the Configuration Files
        super(CeilometerDataManager, self).upgrade_conf_files(version)

    def upgrade_component(self, version):
        """Performs Upgrade specific to the given Component"""
        try:
            utils.LOG.info(_("Configuring Ceilometer to disable "
                             "and clean non-audit events"))
            # If coming from before 1.3.1, we are not supporting
            # cleaning of events. 1.3.0 has iFix.
            if utils.compare_version(version, V131) < 0:
                return
            if utils.is_management_system():
                conn_url = utils.get_openstack_config_value(
                    CEILO_CONF, 'connection', SECTION_DB)
                # DB should already be running.
                utils.wait_for_db_to_start()
                cleaner = EventCleaner(conn_url)
                # Remove events
                cleaner.purge_events(exclude_event_types=AUDIT_EVENT_TYPES)
        except Exception as e:
            utils.LOG.warning(_("There was a problem configuring "
                                "Ceilometer: %s") % str(e))

    ###############################################
    #   Data Manager - Subclass Implementation    #
    ###############################################
    def _get_service_user(self):
        """Provides the name of the Service User for the Component"""
        return 'ceilometer'

    def _get_enabled_services(self):
        """Provides a list of the Services for the Component"""
        # If this is the Management System, we need different services
        if utils.is_management_system():
            services = ['api', 'collector', 'notification', 'polling']
            return ['openstack-ceilometer-%s' % svc for svc in services]
        return []

    def _get_service_log_files(self):
        """Provides a list of Service Log Files for the Component"""
        logs = ['api', 'collector', 'polling',
                'agent-notification', 'httpd-error', 'ceilometer-dbsync']
        return ['/var/log/ceilometer/%s.log' % log for log in logs]

    def _get_database_info(self):
        """Provides the Database Name/User Tuple for the Component"""
        return ('ceilometer', 'ceilometerdb', V130)

    def _get_dbsync_commands(self):
        """Provides a command list for the Component's DB Synchronization"""
        return ['/usr/bin/ceilometer-dbsync']

    def _get_added_conf_properties(self):
        """Provides a dictionary of configuration properties to add"""
        config = {CEILO_CONF: dict(), NOVA_CONF: dict()}
        # Add in the KeystoneAuth and Messaging Configuration Properties
        self._added_auth_conf_properties(config[CEILO_CONF])
        self._added_messaging_conf_properties(config[CEILO_CONF])
        self._added_ras_conf_properties(config[CEILO_CONF])
        self.__added_ceilometer_svccreds_conf_properties(config[CEILO_CONF])
        self.__added_nova_audit_conf_properties(config[NOVA_CONF])
        # If this is the Mgmt System, add in its Configuration Properties
        if utils.is_management_system():
            self.__added_ceilometer_mgmtsys_conf_properties(config[CEILO_CONF])
            # Add in the Configuration Properties for the Paste Files
            self.__added_ceilometer_paste_properties(config)
            # Add in the Nova Configuration Properties for Auditing
            self.__update_ceilometer_event_pipeline(AUDIT_EVENT_TYPES)
        return config

    def _added_policy_dir(self, config, cat):
        pass

    def _get_deleted_conf_properties(self):
        """Provides a dictionary of configuration properties to remove"""
        config = {CEILO_CONF: dict(service_credentials=dict())}
        # Remove any of the deleted Properties from ceilometer.conf
        self._deleted_auth_conf_properties(config[CEILO_CONF])
        self._deleted_messaging_conf_properties(config[CEILO_CONF])
        config[CEILO_CONF]['DEFAULT']['sql_connection'] = V120
        config[CEILO_CONF]['DEFAULT']['collector_workers'] = V123
        config[CEILO_CONF]['DEFAULT']['notification_workers'] = V123
        config[CEILO_CONF]['service_credentials']['os_auth_url'] = V131
        config[CEILO_CONF]['service_credentials']['os_password'] = V131
        config[CEILO_CONF]['service_credentials']['os_tenant_name'] = V131
        config[CEILO_CONF]['service_credentials']['os_username'] = V131
        if utils.is_management_system():
            paste_cfg = config.setdefault(PASTE_CONF, dict())
            paste_cfg['pipeline:main'] = {'paste.filter_factory': V122}
        return config

    def _get_audit_conf_properties(self, version=None):
        """Provides a tuple of configuration properties for auditing"""
        return (PASTE_CONF, ['pipeline:main'], ['pipeline'])

    def _get_included_backup_files(self):
        """Provides a list of Component files to include in a backup"""
        return ['/etc/ceilometer/ceilometer.conf']

    def _get_excluded_backup_files(self):
        """Provides a list of Component files to exclude from a backup"""
        return ['/etc/ceilometer/pipeline.yaml',
                '/etc/ceilometer/event_definitions.yaml']

    def _get_service_config_file(self):
        """Provides the configuration file for the given component"""
        return CEILO_CONF

    def _get_service_password_properties(self):
        """Provides a list of the configuration properties for passwords"""
        return [('ceilometer', CEILO_CONF, 'keystone_authtoken', 'password'),
                ('ceilometerdb', CEILO_CONF, 'database', 'connection'),
                ('telemetry', CEILO_CONF, 'publisher', 'telemetry_secret')]

    ###############################################
    #    Internal Data Manager Helper Methods     #
    ###############################################
    def __added_ceilometer_mgmtsys_conf_properties(self, config):
        """Add in the Management System ceilo.conf Configuration Properties"""
        components = ['nova', 'cinder', 'glance',
                      'neutron', 'ceilometer', 'validator', 'clerk']
        # Set a default dictionary for each of the non-default Sections
        for sect in ['api', 'collector', 'database', 'publisher']:
            config.setdefault(sect, dict())
        # Add in the properties for the API Section
        config['api']['port'] = ('8777', V121)
        config['api']['host'] = ('%(local_ip)s', V121)
        # Add in the properties for the Collector Section
        config['collector']['udp_port'] = ('4952', V121)
        config['collector']['udp_address'] = ('%(local_ip)s', V121)
        config['collector']['workers'] = ('1', V130)
        # Add in the properties for the Database Section
        config['database']['connection'] = (
            'aes:mysql://ceilometerdb:%(ceilometerdb_'
            'password)s@%(local_ip)s:50110/ceilometer', V130)
        # Add meter-ttl for 2 weeks
        config['database']['metering_time_to_live'] = (
            str(60 * 60 * 24 * 14), V130)
        # Add event-ttl for 270 days
        config['database']['event_time_to_live'] = (
            str(60 * 60 * 24 * 270), V130)
        config['notification'] = dict(store_events=('True', V130),
                                      workers=('1', V130))
        # Add in the properties for the Default Section
        config['DEFAULT'].update(dict(
            verbose=('True', V120),
            http_control_exchanges=(components, V122),
            sql_expire_samples_only=('True', V130)))

    def __added_ceilometer_svccreds_conf_properties(self, config):
        """Add in the Service Creds ceilometer.conf Configuration Properties"""
        uri = '%(local_ip)s' if utils.is_management_system() else '%(mgmt_ip)s'
        config.setdefault('service_credentials', dict())
        config.setdefault('publisher', dict())
        # Add in the properties for the Service Credentials Section
        config['service_credentials'].update(dict(
            timeout=('60', V131),
            auth_plugin=('v3password', V131),
            auth_url=('https://%s:5000/v3/' % uri, V131),
            user_domain_name=('Service', V131),
            username=('ceilometer', V131),
            password=('aes:%(ceilometer_password)s', V131),
            project_domain_name=('Service', V131),
            project_name=('service', V131)))
        # Add in the properties for the publisher section
        config['publisher'].update(dict(
            telemetry_secret=('aes:%(telemetry_password)s', V123)))
        # Add the properties that are specific to Mgmt System vs. Remote Host
        if utils.is_management_system():
            config['service_credentials'].update(dict(
                insecure=('True', V121),
                os_endpoint_type=('internalURL', V121)))
        else:
            config['service_credentials'].update(dict(
                os_cacert=(CA_CERTS, V121),
                os_endpoint_type=('publicURL', V121)))

    def __added_nova_audit_conf_properties(self, config):
        """Add in the Auditing nova.conf Configuration Properties"""
        config.setdefault('DEFAULT', dict())
        # Add in the Auditing properties for the Default Section
        config['DEFAULT'].update(dict(
            notify_on_any_change=('True', V120),
            instance_usage_audit=('False', V1312),
            instance_usage_audit_period=('hour', V120),
            notify_on_state_change=('vm_and_task_state', V120)))

    def __added_ceilometer_paste_properties(self, config):
        """Add in the Ceilometer Properties to the Paste Configuration Files"""
        comps = dict(nova='8774', cinder='9000', glance='9292',
                     neutron='9696', validator='8428', ceilometer='8777')
        # Loop through each of the Components adding the Paste Properties
        for comp in comps.keys():
            paste_ini = '/etc/%s/api-paste.ini'
            audit_map = '/etc/%s/api_audit_map.conf' % comp
            # If this is Ceilometer, it has an underscore in the name
            if comp == 'ceilometer':
                paste_ini = paste_ini.replace('-', '_')
            # If this is Glance, it has glance in front of the file name
            if comp == 'glance':
                paste_ini = paste_ini.replace('api', 'glance-api')
            paste_ini = paste_ini % comp
            # If this is TTV, it has a different configuration file name
            if comp == 'validator':
                paste_ini = '/etc/ttv-validation/api-paste.ini'
                audit_map = '/etc/ttv-validation/api_audit_map.conf'
            paste_cfg = config[paste_ini] = dict()
            paste_cfg.setdefault('filter:audit', dict())
            # Add in the properties for the Filter Audit Section
            # Note: All the versions are 1.2.1, since don't restore paste.ini
            filter_sect = paste_cfg['filter:audit']
            filter_sect['ignore_req_list'] = ('GET', V121)
            filter_sect['service_name'] = ('%(mgmt_ip)s:' + comps[comp], V121)
            filter_sect['audit_map_file'] = (audit_map, V131)
            filter_sect['paste.filter_factory'] = (
                'keystonemiddleware.audit:filter_factory', V130)
            # If this is the Ceilometer File, need to add a few more
            if comp == 'ceilometer':
                paste_cfg.setdefault('pipeline:main', dict())
                paste_cfg.setdefault('filter:request_id', dict())
                paste_cfg['pipeline:main']['pipeline'] = (
                    'authtoken api-server', V130)
                paste_cfg['filter:request_id']['paste.filter_factory'] = (
                    'oslo_middleware:RequestId.factory', V121)

    def __update_ceilometer_event_pipeline(self, include_events):
        try:
            pipeline_config = yaml.load(open(CEILOMETER_EVENT_PIPELINE, 'r'))
            sources = pipeline_config.get('sources')
            for source in sources:
                if source.get('name') == 'event_source':
                    events = source.get('events')
                    # Remove entry for all events
                    if "*" in events:
                        events.remove("*")
                    # Add entry for audit events only
                    for event_type in include_events:
                        if event_type not in events:
                            events.append(event_type)
            with open(CEILOMETER_EVENT_PIPELINE, 'w') as pipeline:
                pipeline.write(yaml.dump(pipeline_config,
                                         default_flow_style=False))
        except Exception as e:
            utils.LOG.warning(_("Error occurred while configuring ceilometer "
                                "event pipeline : %s") % str(e))


class EventCleaner(object):
    """Class that cleans non audit events from Ceilometer database.
    """
    def __init__(self, url):
        # No need to pass database options. We just need url.
        # This also suppresses unsupported warnings that comes from
        # sqlalchemy for ceilometer specific options.
        options = dict()
        options['max_retries'] = 0
        # Get DB engine facade from oslo
        self._engine_facade = db_session.EngineFacade(url, **options)

    def purge_events(self, exclude_event_types):
        """Remove events from database

        :param: exclude_event_types: List of event types that should not be
                deleted.
        """
        # Get session from engine facade
        session = self._engine_facade.get_session()
        with session.begin():
            # Step 1: Get the event_type ID for events which need to excluded.
            get_exclude_event_type_query = session.query(
                models.EventType.id).filter(
                    models.EventType.desc.in_(exclude_event_types))
            exclude_event_id_rows = get_exclude_event_type_query.all()
            if (exclude_event_id_rows is not None and
                    len(exclude_event_id_rows) == 0):
                # Nothing to do.
                return
            # Step 2: Create query to select all event IDs excluding the IDs
            # of event_types we want to keep
            purge_event_q = session.query(models.Event.id).filter(
                ~models.Event.event_type_id.in_(
                    e_id.id for e_id in exclude_event_id_rows))
            # This can be used as subquery to identify correct traits
            # corresponding to these events.
            purge_event_subq = purge_event_q.subquery()
            # Step 3: delete all traits for the events
            # print "-- Removing Event traits"
            for trait_model in [models.TraitText, models.TraitInt,
                                models.TraitFloat, models.TraitDatetime]:
                trait_q = (
                    session.query(trait_model)
                    .filter(trait_model.event_id.in_(purge_event_subq)))
                trait_q.delete(synchronize_session="fetch")
            # Step 4:  delete the events now.
            event_rows = purge_event_q.delete(synchronize_session="fetch")
            # Step 5: remove EventType with no corresponding matching events
            event_type_del_q = (session.query(models.EventType)
                                .filter(~models.EventType.events.any()))
            event_type_del_q.delete(synchronize_session="fetch")
            utils.LOG.info(_("%d events are removed from "
                             "database") % event_rows)
