# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import sys
import iniparse
from sqlalchemy import MetaData, Table
from sqlalchemy.sql import bindparam
from powervc_oslo.config import data_mgr as mgr
from powervc_oslo.config import data_utils as utils
from powervc_oslo.config.data_utils import V120, V121, V122, V123
from powervc_oslo.config.data_utils import V130, V1302, V131, V1312
from powervc_neutron import _


PASTE_CONF = '/etc/neutron/api-paste.ini'
NEUTRON_CONF = '/etc/neutron/neutron.conf'
ML2_CONF = '/etc/neutron/plugins/ml2/ml2_conf.ini'
OVS_CONF = '/etc/neutron/plugins/ml2/openvswitch_agent.ini'


######################################################
#         Neutron Data Manager Definition            #
######################################################
class NeutronDataManager(mgr.BaseDataManager):

    def get_component_name(self):
        """Provides the name of the Component providing the Data Manager"""
        return 'neutron'

    def get_external_name(self):
        """Provides the external name of the component """
        return 'network'

    def get_config_desc(self):
        """Provides the description of the component configuration"""
        return _('Manage network-related settings')

    def is_configurable(self):
        """Provides info on whether config changes are allowed"""
        return True

    def is_auditable(self):
        """Is audit support is available for this component"""
        return True

    def is_debuggable(self):
        """Is debug config support available for this component"""
        return True

    def is_supported_environment(self):
        """Specifies whether the Component is supported in this Environment"""
        return True

    def get_dependent_managers(self):
        """Specifies the Component Name of the Managers it is dependent on"""
        return ['db', 'rabbitmq', 'httpd', 'keystone']

    ###############################################
    #     Data Manager - Configure Interface     ##
    ###############################################
    def configure_conf_files(self):
        """Performs the configuration of the Component's CONF Files"""
        super(NeutronDataManager, self).configure_conf_files()
        # Link the ML2 Configuration File as the Main Plug-in File
        utils.run_command('/bin/ln -sf %s /etc/neutron/plugin.ini' % ML2_CONF)
        utils.chmod(ML2_CONF, 0o660)
        utils.chmod(NEUTRON_CONF, 0o660)

    def configure_component(self):
        """Performs Configuration specific to the given Component"""
        super(NeutronDataManager, self).configure_component()
        # Do the Management System-specific Neutron Configuration
        if utils.is_management_system():
            config = utils.get_config_data()
            # Link back the ML2 Plug-in to the Base Neutron Plug-in
            cmd = '/bin/ln -sf %s /etc/neutron/plugin.ini'
            utils.run_command(cmd % ML2_CONF)
        # Configure the Neutron OVS Agent if it is a KVM Host
        elif utils.get_compute_node_type() == 'powerkvm':
            self.__configure_openvswitch_agent()
        # Configure the Neutron PowerVM SEA Agent for NovaLink
        elif utils.get_compute_node_type() == 'powervm':
            self.__configure_powervm_sea_agent()

    def unconfigure_component(self, upgrading):
        """Performs UnConfiguration specific to the given Component"""
        super(NeutronDataManager, self).unconfigure_component(upgrading)
        # UnConfigure the Neutron OVS Agent if it is a KVM Host
        if utils.get_compute_node_type() == 'powerkvm':
            self.__unconfigure_openvswitch_agent()
        # UnConfigure the Neutron PowerVM SEA Agent for NovaLink
        elif utils.get_compute_node_type() == 'powervm':
            self.__unconfigure_powervm_sea_agent()

    ###############################################
    #      Data Manager - Upgrade Interface       #
    ###############################################
    def upgrade_conf_files(self, version):
        """Performs the upgrade of the Component's CONF Files"""
        super(NeutronDataManager, self).upgrade_conf_files(version)
        utils.chmod(ML2_CONF, 0o660)
        utils.chmod(NEUTRON_CONF, 0o660)
        # Even though not really related to Configuration Files, we need
        # to stamp the Alembic Version with Icehouse on a 1.2.1 upgrade
        if utils.is_management_system():
            if utils.compare_version(version, V122) < 0:
                cmd = '/usr/bin/neutron-db-manage --config-file /etc/neutron/'\
                      'neutron.conf --config-file /etc/neutron/plugin.ini '
                utils.run_command(cmd + 'stamp icehouse', True, True)
        # Need to put our startup script back in place since rpm overwrote
        elif utils.get_compute_node_type() == 'powerkvm':
            self.__configure_openvswitch_agent()
        # Cleanup a few  extra property files that were deleted recently
        for afile in ['fwaas_driver.ini', 'lbaas_agent.ini', 'vpn_agent.ini']:
            utils.remove_file('/etc/neutron/' + afile)

    ###############################################
    #      Data Manager - Synchronize UUID's      #
    ###############################################
    def synchronize_uuids(self, host, uuid_map):
        """Synchronizes the UUID's that are needed for the given Host"""
        self.__synchronize_instances(host, uuid_map.get('instances', {}))

    ###############################################
    #   Data Manager - Subclass Implementation    #
    ###############################################
    def _get_service_user(self):
        """Provides the name of the Service User for the Component"""
        return 'neutron'

    def _get_enabled_services(self):
        """Provides a list of the Services for the Component"""
        if utils.is_management_system():
            conf = iniparse.ConfigParser()
            driver = utils.get_openstack_config_value(NEUTRON_CONF,
                                                      'ipam_driver')
            if driver == 'infoblox':
                return ['neutron-server', 'infoblox-ipam-agent']
            return ['neutron-server']
        if utils.get_compute_node_type() == 'powerkvm':
            return ['neutron-openvswitch-agent']
        if utils.get_compute_node_type() == 'powervm':
            return ['neutron-powervc-pvm-sea-agent']
        return []

    def _get_service_log_files(self):
        """Provides a list of Service Log Files for the Component"""
        logs = ['server', 'httpd-error']
        return ['/var/log/neutron/%s.log' % log for log in logs]

    def _get_database_info(self):
        """Provides the Database Name/User Tuple for the Component"""
        return ('neutron', 'neutrondb')

    def _get_dbsync_commands(self):
        """Provides a command list for the Component's DB Synchronization"""
        return ['/usr/bin/neutron-db-manage --config-file /etc/neutron/neutron'
                '.conf --config-file /etc/neutron/plugin.ini upgrade head']

    def _get_added_conf_properties(self):
        """Provides a dictionary of configuration properties to add"""
        config = {NEUTRON_CONF: dict(), ML2_CONF: dict(), OVS_CONF: dict()}
        # Add any of the KeystoneAuth and Messaging added properties
        self._added_auth_conf_properties(config[NEUTRON_CONF])
        self._added_messaging_conf_properties(config[NEUTRON_CONF])
        self._added_ras_conf_properties(config[NEUTRON_CONF])
        # Add the Neutron-specific common properties to the Neutron CONF File
        self.__added_neutron_common_conf_properties(config[NEUTRON_CONF])
        # If this is the Management System, need to set specific properties
        if utils.is_management_system():
            self.__added_neutron_mgmtsys_conf_properties(config[NEUTRON_CONF])
            self.__added_neutron_ml2_plugin_properties(config[ML2_CONF])
        # Otherwise add the OVS Plug-in properties if this is the KVM Host
        elif utils.get_compute_node_type() == 'powerkvm':
            self.__added_neutron_ovs_plugin_properties(config[OVS_CONF])
        elif utils.get_compute_node_type() == 'powervm':
            self.__added_neutron_sea_agent_properties(config[NEUTRON_CONF])
        return config

    def _get_deleted_conf_properties(self):
        """Provides a dictionary of configuration properties to remove"""
        config = {NEUTRON_CONF: dict(), ML2_CONF: dict()}
        # Remove any of the KeystoneAuth and Messaging deleted properties
        self._deleted_auth_conf_properties(config[NEUTRON_CONF])
        self._deleted_messaging_conf_properties(config[NEUTRON_CONF])
        config[NEUTRON_CONF]['service_providers'] = dict(service_provider=V123)
        config[NEUTRON_CONF]['DATABASE'] = dict(sql_connection=V123)
        config[ML2_CONF]['DATABASE'] = dict(sql_connection=V123)

        """ Newly added for 1.3.0 """
        config[NEUTRON_CONF]['DEFAULT']['nova_admin_username'] = V130
        config[NEUTRON_CONF]['DEFAULT']['nova_url'] = V130
        config[NEUTRON_CONF]['DEFAULT']['nova_admin_password'] = V130
        config[NEUTRON_CONF]['DEFAULT']['nova_admin_auth_url'] = V130
        config[NEUTRON_CONF]['DEFAULT']['nova_admin_username'] = V130
        config[NEUTRON_CONF]['DEFAULT']['nova_region_name'] = V130
        config[NEUTRON_CONF]['DEFAULT']['nova_api_insecure'] = V130
        config[NEUTRON_CONF]['DEFAULT']['nova_admin_tenant_id'] = V130

        return config

    def _get_audit_conf_properties(self, version=None):
        """Provides a tuple of configuration properties for auditing"""
        return (PASTE_CONF, ['composite:neutronapi_v2_0'], ['keystone'])

    def _get_included_backup_files(self):
        """Provides a list of Component files to include in a backup"""
        return ['/etc/neutron']

    def _get_service_config_file(self):
        """Provides the configuration file for the given component"""
        return NEUTRON_CONF

    ###############################################
    #    Internal Data Manager Helper Methods     #
    ###############################################
    def __added_neutron_common_conf_properties(self, config):
        """Add in the neutron.conf Common Configuration Properties"""
        # Add in the properties for the Default Section
        config['DEFAULT'].update(dict(
            bind_port=('9696', V121), bind_host=('%(local_ip)s', V121),
            report_interval=('30', V122), agent_down_time=('120', V122)))
        # Add in the properties for the Agent Section
        root_helper = 'sudo neutron-rootwrap /etc/neutron/rootwrap.conf'
        config['AGENT'] = dict(root_helper=(root_helper, V130))

    def __added_neutron_mgmtsys_conf_properties(self, config):
        """Add in the neutron.conf MgmtSys Configuration Properties"""
        config.setdefault('QUOTAS', dict())
        config.setdefault('database', dict())
        config.setdefault('nova', dict())
        # Add in the properties for the Database Section
        config['database']['connection'] =\
            ('aes:mysql://neutrondb:%(neutrondb_'
             'password)s@%(local_ip)s:50110/neutron', V131)
        # Add in the properties for the Quotas Section
        config['QUOTAS'].update(dict(
            quota_network=('-1', V122),
            default_quota=('-1', V122), quota_subnet=('-1', V122),
            quota_floatingip=('-1', V122), quota_port=('-1', V122)))
        # Add in the properties for the Default Section
        config['DEFAULT'].update(dict(
            allow_overlapping_ips=('True', V120),
            core_plugin=('neutron.plugins.ml2.plugin.Ml2Plugin', V120),
            api_workers=('8', V130),
            rpc_workers=('8', V1302),
            dhcp_lease_duration=('1', V120),
            verbose=('True', V120)),
            api_extensions_path=(
                '/usr/lib/python2.7/site-packages/'
                'powervc_discovery/registration/network', V1312))
        # Add in the properties for the nova Section
        config['nova'].update(dict(
            auth_url=('https://%(local_ip)s:5000/v3', V130),
            username=('nova', V130),
            password=('aes:%(nova_password)s', V130),
            project_name=('service', V130),
            project_domain_name=('Service', V130),
            user_domain_name=('Service', V130),
            region_name=('RegionOne', V130),
            auth_plugin=('v3password', V130),
            insecure=('True', V130)))

    def __added_neutron_ml2_plugin_properties(self, config):
        """Add in the ml2_conf.ini Configuration Properties"""
        for sect in ['securitygroup', 'database', 'ml2',
                     'ml2_type_vlan', 'ml2_type_flat']:
            config.setdefault(sect, dict())
        # Add in the properties for the Security Group and Database Sections
        config['securitygroup']['enable_security_group'] = ('False', V121)
        config['database']['connection'] =\
            ('aes:mysql://neutrondb:%(neutrondb_'
             'password)s@%(local_ip)s:50110/neutron', V131)
        config['ml2']['extension_drivers'] = ('dns_authorities', V131)
        # Add the Properties specific to whether it is PowerVM or PowerKVM
        if utils.get_offering() == 'powervm':
            config['ml2']['tenant_network_types'] = ('vlan', V120)
            config['ml2']['type_drivers'] = ('vlan_ibmpowervm', V120)
            config['ml2']['mechanism_drivers'] = ('mech_pvc_sea', V130)
            config['ml2_type_vlan']['allow_vlan_reuse'] = ('True', V120)
            config['ml2_type_vlan']['network_vlan_ranges'] =\
                ('default:1:4094,unknown:1:4094', V120)
        else:
            config['ml2_type_flat']['flat_networks'] = ('*', V120)
            config['ml2']['tenant_network_types'] = ('flat,vlan', V120)
            config['ml2']['type_drivers'] = ('flat,vlan_powerkvm', V120)
            config['ml2']['mechanism_drivers'] = ('mech_ibmpower', V120)
            config['ml2_type_vlan']['network_vlan_ranges'] =\
                ('default:1:4094', V120)

    def __added_neutron_ovs_plugin_properties(self, config):
        """Add in the openvswitch_agent.ini Configuration Properties"""
        config.setdefault('OVS', dict())
        config.setdefault('SECURITYGROUP', dict())
        # Add in the properties for the Security Group Section
        config['SECURITYGROUP']['firewall_driver'] =\
            ('neutron.agent.firewall.NoopFirewallDriver', V121)
        config['SECURITYGROUP']['enable_security_group'] = ('False', V121)
        # Add in the properties for the OVS Section
        config['OVS'].update(dict(
            tunnel_id_ranges=('<>', V121),
            integration_bridge=('br-int', V121),
            bridge_mappings=('default:default', V121),
            network_vlan_ranges=('default:1:4094', V121)))

    def __added_neutron_sea_agent_properties(self, config):
        '''Add config needed for novalink hosts'''
        config.setdefault('AGENT', dict())
        # Add in the properties for the AGENT Section
        config['AGENT']['pvid_update_loops'] = ('7200', V130)

    def __configure_openvswitch_agent(self):
        """Helper method run the Neutron Agent Configuration script"""
        modstr = 'powervc_neutron.plugins.powerkvm.agent.powerkvm_agent_config'
        __import__(modstr)
        # Run the configure method on the PowerVC Neutron OVS Agent
        getattr(sys.modules[modstr], 'config')()

    def __unconfigure_openvswitch_agent(self):
        """Helper method run the Neutron Agent Configuration script"""
        modstr = 'powervc_neutron.plugins.powerkvm.agent.powerkvm_agent_config'
        __import__(modstr)
        # Run the unconfigure method on the Configuration Python Module
        getattr(sys.modules[modstr], 'unconfig')()

    def __configure_powervm_sea_agent(self):
        """Helper method to perform the PowerVM SEA Agent Configuration"""
        modstr = 'powervc_neutron.plugins.pvm.agent.powervc_pvm_sea_agent'\
            '_config'
        __import__(modstr)
        # Run the configure method on the PowerVC SEA Agent
        getattr(sys.modules[modstr], 'config')()

    def __unconfigure_powervm_sea_agent(self):
        """Helper method to perform the PowerVM SEA Agent Un-Configuration"""
        modstr = 'powervc_neutron.plugins.pvm.agent.powervc_pvm_sea_agent'\
            '_config'
        __import__(modstr)
        # Run the configure method on the PowerVC SEA Agent
        getattr(sys.modules[modstr], 'unconfig')()

    def __synchronize_instances(self, host, uuid_map):
        """Helper method to synchronize the VM UUID's that are given"""
        # Get a connection to the Database and load the needed Tables
        engine = utils.get_database_engine(
            NEUTRON_CONF, 'connection', 'database')
        metadata = MetaData(bind=engine, schema='neutron')
        # Setup the Update Statement to update the UUID in the Ports Table
        port_table = Table('ports', metadata, autoload=True)
        port_update = port_table.update().where(
            port_table.c.device_id == bindparam('olduuid')).values(
                device_id=bindparam('newuuid'))
        # Loop through each of the UUID's updating the Ports Table
        for old_uuid, new_uuid in uuid_map.iteritems():
            engine.execute(port_update, olduuid=old_uuid, newuuid=new_uuid)
