#
#
# Copyright 2014 IBM Corp.
# All Rights Reserved.
#
#    Copyright (c) 2013 OpenStack Foundation
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr

from neutron.common import exceptions as exc
from neutron.extensions import netmtu
from neutron.extensions import portbindings
from oslo_log import log
from neutron.plugins.ml2.drivers.\
    openvswitch.mech_driver import mech_openvswitch

from powervc_neutron import _

LOG = log.getLogger(__name__)


class IBMPowerMechanismDriver(mech_openvswitch.OpenvswitchMechanismDriver):
    '''
    Overrides the base ML2 vlan type driver to allow physical networks that
    don't already exist in the CONF file.
    '''

    def __init__(self):
        LOG.debug('IBM Power mechanism driver loaded')
        super(IBMPowerMechanismDriver, self).__init__()

        # Do not support hybrid plugs
        self.vif_details[portbindings.OVS_HYBRID_PLUG] = False

    def create_network_precommit(self, context):
        """Allocate resources for a new network.

        :param context: NetworkContext instance describing the new
        network.

        Our override is so we can validate the mtu attribute, if present.
        """
        LOG.debug('Enter with network: %s' % context.current)

        # Verify mtu, if present.  Valid values are 0 (let VM default), or
        # between 1000 and 9000
        mtu = context.current.get(netmtu.MTU, None)
        mtu = int(mtu) if mtu is not None else mtu
        if mtu is not None and mtu < 0 or (mtu > 0 and
                                           mtu < 1000) or mtu > 9000:
            raise exc.InvalidInput(error_message=_('Please provide a MTU value'
                                                   ' between 1000 and 9000'))

        super(IBMPowerMechanismDriver, self).create_network_precommit(context)

    def create_subnet_precommit(self, context):
        '''
        Verifies the subnet provided has more than two IPs in it.  Since the
        high IP in a subnet is used for broadcast, a 2 IP subnet only allows
        for one IP which could either be used for a gateway and no client VMs,
        or a client VM and no gateway.  Either way, it's not very useful.

        :param context: SubnetContext
        '''
        LOG.debug('Enter with subnet: %s' % context.current)

        # Verify existence of the CIDR of the subnet
        cidr = None
        if context.current:
            cidr = context.current.get('cidr')
        if cidr is None:
            raise exc.InvalidInput(error_message=_('Please provide a CIDR '
                                                   'value on the subnet'))

        # Now verify there are more than 2 IPs in the subnet
        if netaddr.IPNetwork(cidr).size < 3:
            err_msg = (_("The subnet specified by %(cidr)s for network "
                         "%(network_id)s is too small. Please specify a "
                         "larger subnet") %
                       {'cidr': cidr,
                        'network_id': context.current.get('network_id')})
            LOG.debug("Validation for subnet %s failed - prefix length is "
                      "too large.", cidr)
            raise exc.InvalidInput(error_message=err_msg)

        super(IBMPowerMechanismDriver, self).create_subnet_precommit(context)
