#
#
# Copyright 2013, 2014 IBM Corp.
# All Rights Reserved.
#
#   Copyright 2012 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.


import sqlalchemy as sa
from oslo_config import cfg

from neutron.common import exceptions as exc
from neutron.plugins.common import constants as p_const
from neutron.plugins.ml2 import driver_api as api
from neutron.plugins.ml2.drivers import type_vlan
from oslo_log import log

from powervc_neutron import _

LOG = log.getLogger(__name__)


vlan_opts = [
    cfg.BoolOpt('allow_vlan_reuse',
                default=False,
                help='Allows multiple reservations for a specific VLAN')
]

cfg.CONF.register_opts(vlan_opts, "ml2_type_vlan")
cfg.CONF.import_opt('network_vlan_ranges',
                    'neutron.plugins.ml2.drivers.type_vlan',
                    'ml2_type_vlan')


class IBMPowerVMVlanTypeDriver(type_vlan.VlanTypeDriver):

    """
    Overrides the vlan driver for allowing networks with duplicate vlan:
    """

    def __init__(self):
        LOG.info("IBM PowerVM vlan driver loaded")
        super(IBMPowerVMVlanTypeDriver, self).__init__()

    def reserve_provider_segment(self, session, segment):
        '''
        validates the segment id, if allow vlan reuse is
        enabled then duplicate segment id allowed otherwise
        VlanIdInUse exception is thrown.
        :param session: session
        :param segment: segment
        '''
        physical_network = segment[api.PHYSICAL_NETWORK]
        vlan_id = segment[api.SEGMENTATION_ID]
        with session.begin(subtransactions=True):
            try:
                alloc = (session.query(type_vlan.VlanAllocation).
                         filter_by(physical_network=physical_network,
                                   vlan_id=vlan_id).
                         with_lockmode('update').
                         one())
                if (alloc.allocated and
                        not cfg.CONF.ml2_type_vlan.allow_vlan_reuse):
                    raise exc.VlanIdInUse(vlan_id=vlan_id,
                                          physical_network=physical_network)
                LOG.debug(_("Reserving specific vlan %(vlan_id)s on physical "
                            "network %(physical_network)s from pool"),
                          {'vlan_id': vlan_id,
                           'physical_network': physical_network})
                alloc.allocated = True
            except sa.orm.exc.NoResultFound:
                LOG.debug(_("Reserving specific vlan %(vlan_id)s on physical "
                            "network %(physical_network)s outside pool"),
                          {'vlan_id': vlan_id,
                           'physical_network': physical_network})
                alloc = type_vlan.VlanAllocation(
                    physical_network=physical_network,
                    vlan_id=vlan_id,
                    allocated=True)
                session.add(alloc)
        return {api.NETWORK_TYPE: p_const.TYPE_VLAN,
                api.PHYSICAL_NETWORK: alloc.physical_network,
                api.SEGMENTATION_ID: alloc.vlan_id,
                api.MTU: self.get_mtu(alloc.physical_network)}
