#
#
# Copyright 2014 IBM Corp.
# All Rights Reserved.
#
#    Copyright (c) 2013 OpenStack Foundation
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import iniparse
from oslo_log import log
from neutron.plugins.ml2 import driver_api as api
from neutron.plugins.ml2.drivers import type_vlan
from oslo_config import cfg

from powervc_neutron import _

LOG = log.getLogger(__name__)


class PowerKVMVlanTypeDriver(type_vlan.VlanTypeDriver):
    '''
    Overrides the base ML2 vlan type driver to allow physical networks that
    don't already exist in the CONF file.
    '''

    def __init__(self):
        LOG.debug('PowerKVM vlan type driver loaded')
        super(PowerKVMVlanTypeDriver, self).__init__()

    def validate_provider_segment(self, segment):
        '''
        Override this method so we don't raise an exception when the passed
        in physical network does not already exist in the CONF file.  Instead,
        add the physical network to the CONF file, load it in the in-memory
        config, then call up to the super class.
        '''
        LOG.debug('Enter with network segment: %s' % segment)

        # Ensure physnet is already in the CONF file
        self._update_network_vlan_ranges(segment.get(api.PHYSICAL_NETWORK))

        # Now call up to the base class to finish the validation.  Note, this
        # could still fail if there are invalid attributes to the segment.
        super(PowerKVMVlanTypeDriver, self).validate_provider_segment(segment)

        LOG.debug('Exit')

    def _update_network_vlan_ranges(self, physnet):
        '''
        This method will check to be sure the passed in physical network name
        already exists in the plugin.ini config file.  If it does not, the
        config file will be updated to include it, the in-memory oslo config
        will be refreshed, and self.network_vlan_ranges will be updated with
        the new list.

        :param physnet: Physical network name
        '''
        LOG.debug('Enter with physical network name: %s' %
                  physnet)

        # Check validity
        if not physnet:
            # Nothing passed in, just return.  The parent class's method
            # will bomb because of this.
            LOG.warn('No physical network name provided!')
            return

        # Check for the existence of the physical network by looping through
        # all the network_vlan_ranges.  Each range has the format:
        #   <physical network name>:<lower vlanid bound>:<upper vlanid bound>
        found_physnet = False
        for vlan_range in cfg.CONF.ml2_type_vlan.network_vlan_ranges:
            # If the current range is for our physnet
            if physnet == vlan_range.split(':')[0]:
                # Set the found flag and exit
                LOG.debug('Physical network name already in plugin.ini')
                found_physnet = True
                break

        # If we didn't find the physical network...
        if not found_physnet:
            # ...we need to add it to the config file, update the oslo
            # configuration in memory, and update self.network_vlan_ranges
            LOG.info(_('Adding physical network: %s') % physnet)
            new_vlan_range = '%s:1:4094' % physnet

            # First, update plugin.ini.  Start by reading the file into mem.
            plugin_ini = cfg.CONF.find_file('plugin.ini')
            conf = iniparse.ConfigParser()
            conf.readfp(open(plugin_ini))

            # Now get the network_vlan_ranges option from the ml2_type_vlan
            # section.  Be sure the section exists first.
            ini_network_vlan_ranges = None
            if conf.has_section('ml2_type_vlan'):
                # If the section already has the option, just grab it.
                if conf.has_option('ml2_type_vlan', 'network_vlan_ranges'):
                    ini_network_vlan_ranges = conf.get('ml2_type_vlan',
                                                       'network_vlan_ranges')
            else:
                # Section doesn't exist.  We must add it.
                conf.add_section('ml2_type_vlan')

            # If there were already values there, just append ours.
            if ini_network_vlan_ranges:
                new_vlan_ranges = '%s,%s' % (ini_network_vlan_ranges,
                                             new_vlan_range)
                LOG.info(_('Appending new physical network.  New ranges: %s') %
                         new_vlan_ranges)
                conf.set('ml2_type_vlan',
                         'network_vlan_ranges',
                         new_vlan_ranges)
            else:
                # No values existed already.  Just set our value.
                LOG.info('Setting network_vlan_ranges to single physical '
                         'network')
                conf.set('ml2_type_vlan',
                         'network_vlan_ranges',
                         new_vlan_range)

            # Write out the new plugin.ini file
            with open(plugin_ini, 'w') as f:
                conf.write(f)

            # Done with plugin.ini, move on to in-memory oslo configuration
            cfg.CONF.ml2_type_vlan.network_vlan_ranges.append(new_vlan_range)

            # Now refresh self.network_vlan_ranges
            self._parse_network_vlan_ranges()

        # All config should be up to date now.
        LOG.debug('Exit')
