#
#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from eventlet.green import subprocess
from eventlet import greenthread
from neutron.common import utils
from oslo_log import log as logging


LOG = logging.getLogger(__name__)


def check_and_start_ovs_service(roothelper=None):
    '''
    Will check if open vswitch service is running
    If not, then start it
    '''

    # First check to see if the service is enabled.
    full_args = _full_args("systemctl is-enabled openvswitch.service",
                           roothelper)
    response = _execute(full_args, check_exit_code=False)
    lines = response.split('\n')
    is_ovs_enabled = True
    for line in lines:
        if 'disabled' in line:
            is_ovs_enabled = False

    # If not enabled, then we need to enable it
    if not is_ovs_enabled:
        full_args = _full_args("systemctl enable openvswitch.service",
                               roothelper)
        _execute(full_args)

    # Check to see if the openvswitch service is running
    full_args = _full_args("systemctl status openvswitch.service",
                           roothelper)
    response = _execute(full_args, check_exit_code=False)
    lines = response.split('\n')
    is_ovs_running = True
    for line in lines:
        if 'inactive' in line:
            is_ovs_running = False
            break

    # If OpenVSwitch isn't running, then it needs the init
    if not is_ovs_running:
        full_args = _full_args("systemctl start openvswitch.service",
                               roothelper)
        _execute(full_args)


def _execute(cmd, check_exit_code=True):
    '''This is a substitute for neutron.agent.linux.utils.execute since that
    one blows up on Japanese characters when running systemctl'''
    try:
        LOG.debug("Running command: %s", cmd)
        obj = utils.subprocess_popen(cmd, shell=False,
                                     stdin=subprocess.PIPE,
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE)
        _stdout, _stderr = obj.communicate()
        returncode = obj.returncode
        obj.stdin.close()

        # Log a debug message of what we ran and what we got.
        m = ("\nCommand: {cmd}\nExit code: {code}\nStdout: {stdout}\n"
             "Stderr: {stderr}").format(cmd=cmd,
                                        code=returncode,
                                        stdout=_stdout,
                                        stderr=_stderr)
        LOG.debug(m)

        if returncode and check_exit_code:
            raise RuntimeError(m)
    finally:
        # Do this to make sure subprocess cleans up properly
        greenthread.sleep(0)

    return _stdout


def _full_args(command, roothelper=None):
    '''
    Gets the array for the 'full arguments'

    :param command: The command string
    :param roothelper: The root helper, may be none
    :returns: Array of commands.
    '''
    full_args = []
    if roothelper is not None:
        full_args.extend(roothelper.split())
    full_args.extend(command.split())
    return full_args
