#
#
# Copyright 2014 IBM Corp.
# All Rights Reserved.
#
#   Copyright 2011 Nicira Networks, Inc.
#   All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from datetime import datetime
import grp
import json
import os
import os.path
import pwd
import sys
import shutil

from neutron.agent.linux import utils
from powervc_neutron.plugins.powerkvm import agent


def config():
    '''
    Performs the agent installation configuration.
    '''
    # Check if ovs service is running, else start it
    agent.check_and_start_ovs_service()

    # Determine which vswitch to use.  KVM 3.1 cannot use 'default' as we did
    # on KVM 2.1.1
    full_args = ["uname", "-r"]
    pkvm_ver = utils.execute(full_args)
    vswitch_to_use = (
        'default' if pkvm_ver.split('pkvm')[1].startswith('2') else 'default0')

    # Check for the default vswitch
    if not has_vswitch(vswitch_to_use):
        # Create it if not on the system
        full_args = ["ovs-vsctl", "--timeout=5", "add-br", vswitch_to_use]
        utils.execute(full_args)

    # Check for the 'internal' vswitch
    if not has_vswitch('br-int'):
        # Create it if not on the system
        full_args = ["ovs-vsctl", "--timeout=5", "add-br", "br-int"]
        utils.execute(full_args)

    # Turn the neutron agent on for system restart
    full_args = ["systemctl", "enable", "neutron-openvswitch-agent.service"]
    utils.execute(full_args)

    # Copy the file /usr/bin/neutron-openvswitch-agent to a backup
    if not os.path.isfile("/usr/bin/.neutron-openvswitch-agent"):
        shutil.copy("/usr/bin/neutron-openvswitch-agent",
                    "/usr/bin/.neutron-openvswitch-agent")

    # Copy our agent startup script into place
    script_dir = os.path.dirname(os.path.realpath(__file__))
    shutil.copy(script_dir + '/powerkvm_agent_startup.py',
                '/usr/bin/neutron-openvswitch-agent')

    # Set the appropriate permissions on our startup script
    full_args = ["chmod", "755", "/usr/bin/neutron-openvswitch-agent"]
    utils.execute(full_args)

    # Since we're updating the OVS neutron ini file, we need to update the
    # permissions to allow us to modify it
    full_args = ["chmod", "660",
                 "/etc/neutron/plugins/ml2/openvswitch_agent.ini"]
    utils.execute(full_args)

    # Create and initialise vm flow mapping file
    _create_vm_flow_mapping_file()


def unconfig():
    '''
    Will unconfigure the system.  Stops the processes, moves the original
    neutron code back into place.
    '''
    # Stop the agent
    full_args = ["systemctl", "stop", "neutron-openvswitch-agent.service"]
    utils.execute(full_args)

    # Assuming the original neutron-openvswitch-agent script exists, move it
    # back into the location
    if os.path.isfile("/usr/bin/.neutron-openvswitch-agent"):
        shutil.move("/usr/bin/.neutron-openvswitch-agent",
                    "/usr/bin/neutron-openvswitch-agent")


def has_vswitch(br_name):
    '''
    Determines whether or not the system has a vswitch

    :param br_name: The name of the switch to find.
    :returns: True if the vswitch is already on the system.  False otherwise.
    '''
    full_args = ["ovs-vsctl", "--timeout=5", "list-br"]
    bridges = utils.execute(full_args)
    bridge_list = bridges.split()

    found_default_br = False
    for bridge in bridge_list:
        if bridge == br_name:
            found_default_br = True
            break

    return found_default_br


def _create_vm_flow_mapping_file():
    """Method to create and initialise the pvc_flow_mapping.json file"""
    VM_FLOW_MAP_FILE = "/etc/sysconfig/network-scripts/pvc_flow_mapping.json"
    VIRTUAL_MACHINES_KEY = "virtual_machines"

    # First, see if a current file is readable.  If it's corrupted, move and
    # recreate
    if os.path.exists(VM_FLOW_MAP_FILE):
        infile = None
        try:
            infile = open(VM_FLOW_MAP_FILE)
            # Read it in to be sure the json is valid
            json.load(infile)
        except Exception:
            # Something is wrong with the file.  Back it up so we create a
            # fresh one below.
            if infile:
                shutil.move(infile.name, '%s.%s' %
                            (infile.name, datetime.today()
                             .strftime('%Y%m%d_%H%M%S')))
        finally:
            if infile:
                infile.close()

    # Check if file exists
    if not os.path.exists(VM_FLOW_MAP_FILE):
        # create and write initial contents to it
        try:
            outfile = open(VM_FLOW_MAP_FILE, 'w')
            json.dump({VIRTUAL_MACHINES_KEY: {}}, outfile)
        except Exception:
            raise
        finally:
            if outfile:
                outfile.close()
    # change owner and group to neutron
    ownerid = pwd.getpwnam("neutron").pw_uid
    grpid = grp.getgrnam("neutron").gr_gid
    os.chown(VM_FLOW_MAP_FILE, ownerid, grpid)
    # set permissions to 644
    os.chmod(VM_FLOW_MAP_FILE, 0o644)


if __name__ == '__main__':
    if len(sys.argv) <= 1:
        print ("Must specify 'configure' or 'unconfigure' as input")
    else:
        option = sys.argv[1]
        if option == 'configure':
            config()
        elif option == 'unconfigure':
            unconfig()
        else:
            print ("Must specify 'configure' or 'unconfigure' as input")
    sys.exit(0)
