#
#
# Copyright 2014 IBM Corp.
# All Rights Reserved.
#
#   Copyright 2011 Nicira Networks, Inc.
#   All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import os
import copy
from lxml import etree
import sys
import iniparse
import eventlet
from neutron.common import config as logging_config
from neutron.plugins.ml2.drivers.openvswitch.agent import ovs_neutron_agent
from neutron.plugins.ml2.drivers.openvswitch.agent.common import constants
from neutron.plugins.common import constants as p_const
from oslo_log import log as logging
from neutron.agent.ovsdb import impl_vsctl as ovsdb_vsctl
from neutron.agent.common import ovs_lib
from neutron.agent.linux import utils
from neutron.common import rpc as n_rpc
from neutron.common import utils as n_utils
from oslo_config import cfg
from powervc_neutron.plugins.powerkvm import agent
from powervc_oslo.network.powerkvm_flow_mapping import \
    VMFlowMapping
from neutron.plugins.ml2.drivers.openvswitch.\
    agent.openflow.ovs_ofctl import br_int
from neutron.plugins.ml2.drivers.openvswitch.\
    agent.openflow.ovs_ofctl import br_phys
from neutron.plugins.ml2.drivers.openvswitch.\
    agent.openflow.ovs_ofctl import br_tun

from powervc_neutron import _

LOG = logging.getLogger(__name__)
MAX_BRIDGE_CHARS = 8
MAX_BRIDGE_MISSING = 5

MIN_UNMANAGED_VLAN_TAG = 3500
MAX_UNMANAGED_VLAN_TAG = 4000

BRIDGE_CLASSES = {
    'br_int': br_int.OVSIntegrationBridge,
    'br_phys': br_phys.OVSPhysicalBridge,
    'br_tun': br_tun.OVSTunnelBridge,
}


class PowerKVMNeutronAgent(ovs_neutron_agent.OVSNeutronAgent):
    '''
    Implements OVS-based VLANs and flat networks.

    Two local bridges are created: an integration bridge (defaults to
    'br-int') and a tunneling bridge (defaults to 'br-tun'). An
    additional bridge is created for each physical network interface
    used for VLANs and/or flat networks.

    All VM VIFs are plugged into the integration bridge. VM VIFs on a
    given virtual network share a common "local" VLAN (i.e. not
    propagated externally). The VLAN id of this local VLAN is mapped
    to the physical networking details realizing that virtual network.

    For each virtual networks realized as a VLANs or flat network, a
    veth is used to connect the local VLAN on the integration bridge
    with the physical network bridge, with flow rules adding,
    modifying, or stripping VLAN tags as necessary.
    '''

    def __init__(self, bridge_classes, conf=None):
        '''
        Constructor.

        :param bridge_classes: a dict for bridge classes.
        :param conf: an instance of ConfigOpts
        '''
        # Sridhar. added next line
        self._local_vlan_hints = {}
        # Keep a list of all the vSwitches that we have sent a message to
        # bumblebee about
        self.bad_vswitches = []

        self.conf = conf or cfg.CONF
        self.ovs = ovs_lib.BaseOVS()
        agent_conf = self.conf.AGENT
        ovs_conf = self.conf.OVS

        # A counter for how long a vswitch has been removed for.  A network
        # restart can cause the vswitch to appear removed for short windows
        # when the vswitch is in fact still present.
        self.vswitch_missing_counter = {}
        # Initialize unmanaged_vm_local_vlan_map before calling super
        self.unmanaged_vm_local_vlan_map = []
        # First call the base constructor
        super(PowerKVMNeutronAgent, self).__init__(
            bridge_classes)

        LOG.info(_('Starting PowerKVM Neutron agent'))

        # maintain the in memory copy of bridge mapping for updates
        self.orig_bridge_mappings = self.bridge_mappings
        #self.integ_br = self.br_int_cls(ovs_conf.integration_bridge)
        self.integ_br = ovs_conf.integration_bridge
        self.setup_integration_br()
        # Now ensure tunneling is not enabled
        self.tunnel_types = []
        self.agent_state['configurations']['tunnel_types'] = self.tunnel_types
        self.enable_tunneling = False

        # initialize flow mapping data in the object, if needed
        VMFlowMapping.get_vm_flow_map()

        self.unmanaged_vm_available_local_vlans = range(MIN_UNMANAGED_VLAN_TAG,
                                                        MAX_UNMANAGED_VLAN_TAG)

        # The Host OVS checker shouldn't run in every loop.  Count the number
        # of times we run and only run once per X times through the RPC loop.
        self.host_ovs_check_counter = 0
        self._notifier = n_rpc.get_notifier('network')

        # TODO Need to tolerate customer changed mappings

    def _add_vswitch_to_config(self, ovs_bridge_info):
        '''
        This method should create the updated bridge mappings and set
        them in in-memory config and the config file : OVS
        :param ovs_bridge_info : map of bridges added/removed
        '''
        LOG.debug('Enter _add_vswitch_to_config with: %s' % ovs_bridge_info)

        # update in memory bridge mapping with new mapping
        # if new vswitch added
        #    Create a new mapping of physical network:device which
        #    would be default:default (if vswitch is named default).
        for bridge in ovs_bridge_info['added']:
            LOG.debug('Adding bridge %s to config' % bridge)
            self.orig_bridge_mappings[bridge] = bridge

        # update bridge mapping in config file
        PowerKVMNeutronAgent.\
            set_bridge_mapping_in_conf(self.orig_bridge_mappings)

        LOG.debug('Exit _add_vswitch_to_config')

    def _remove_vswitch_from_config(self, ovs_bridge_info):
        '''
        This method should delete removed bridge_mappings from the
        config file : OVS
        :param ovs_bridge_info : map of bridges added/removed
        '''
        LOG.debug('Enter _remove_vswitch_from_config with: %s' %
                  ovs_bridge_info)

        # If we detect that a vswitch was removed, but it does NOT have
        # any VMs using it -> remove that vswitch and update/reload the config

        # The system maintains a 'local_vlan_map' - a dictionary
        # (id is the network UUID, the value is a LocalVLANMapping)
        # The LocalVLANMapping contains the physical network and a set
        # of vif ports.
        for bridge in ovs_bridge_info['removed']:
            has_vms = False
            for network_uuid, vlan_mapping in self.local_vlan_map.iteritems():
                # compare to bridge name as physical network name = bridge name
                if vlan_mapping.physical_network == bridge:
                    if len(vlan_mapping.vif_ports) > 0:
                        has_vms = True

            # If we detect that a vswitch was removed,and it has VMs on it
            # -> we send a bumblebee warning
            if has_vms:

                # A network restart can cause the vswitch to appear removed for
                # short timing windows when the vswitch is still present.
                count = self.vswitch_missing_counter.get(bridge, 0) + 1
                if count >= MAX_BRIDGE_MISSING:
                    self.vswitch_missing_counter[bridge] = MAX_BRIDGE_MISSING
                    LOG.debug('Vswitch missing count reached for %s' % bridge)
                else:
                    self.vswitch_missing_counter[bridge] = count
                    LOG.debug('Vswitch missing count is %s for %s' %
                              (str(count), bridge))
                    continue

                LOG.error(_("vswitch %s was removed from system, but is used "
                            "by VMs.  Removing the vswitch from config") %
                          bridge)
                msgb = _('Open vSwitch %(vswitch_name)s was removed from host'
                         ' %(host_name)s. The system detected that there are '
                         'virtual machines that are still attempting to route '
                         'traffic through this virtual switch.  The '
                         'administrator should manually add the virtual switch'
                         ' back to the system to regain network connectivity '
                         'to the virtual machines.') % \
                    {'vswitch_name': bridge, 'host_name': cfg.CONF.host}
                info = {'msg': msgb}
                LOG.error(msgb)
                self._notifier.error(self.context, 'ovs.delete.log', info)

            # if a vswitch was removed update in-memory mapping
            # Be sure the bridge passed in even exists first.  When the
            # user can update config out of band, you never know what'll
            # happen here.
            if bridge in self.orig_bridge_mappings:
                LOG.debug('Removing bridge %s from config.' % bridge)
                del self.orig_bridge_mappings[bridge]
            else:
                LOG.warn(_('Requested bridge %s does not exist for '
                           'removal.') % bridge)

        # update bridge mapping in config file
        PowerKVMNeutronAgent.\
            set_bridge_mapping_in_conf(self.orig_bridge_mappings)

        LOG.debug('Exit _remove_vswitch_from_config')

    @staticmethod
    def set_bridge_mapping_in_conf(bridge_mappings):
        '''
        This method should update the bridge_mappings in the config file
        '''
        LOG.debug('Enter set_bridge_mapping_in_conf')

        # build the mapping to update in config file
        updated_bridge_mapping = ''
        for bridge_key in bridge_mappings:
            updated_bridge_mapping += bridge_key + ":"\
                + bridge_mappings[bridge_key]
            updated_bridge_mapping += ','

        updated_bridge_mapping = \
            updated_bridge_mapping[:len(updated_bridge_mapping) - 1]

        plugin_ini = cfg.CONF.find_file('openvswitch_agent.ini')

        # Build the arguments for the command
        full_args = ['openstack-config']
        if updated_bridge_mapping:
            LOG.debug('Writing bridge mapping \'%s\' to config' %
                      updated_bridge_mapping)
            full_args.append('--set')
        else:
            LOG.info(_('oslo config contained no bridge_mappings'))
            full_args.append('--del')
        full_args.extend([plugin_ini, 'OVS', 'bridge_mappings'])
        if updated_bridge_mapping:
            full_args.append(updated_bridge_mapping)
        utils.execute(full_args, run_as_root=True, check_exit_code=[0])

        LOG.debug('Exit set_bridge_mapping_in_conf')

    def _list_host_ovs_bridges(self):
        '''
        This method will filter host level vswitches ie host ovs bridges
        It will get all the ovs bridges, remove integration bridges,
        and bridges with external id same as bridge
        :return: host_ovs_bridges: updated list of host level vswitches
        except br-int, and those that have external bridge-ids same as
        bridge name
        '''

        LOG.debug('Enter _list_host_ovs_bridges')

        # call the ovs-vsctl list-br command
        ovs = ovs_lib.BaseOVS()
        vsctl = ovsdb_vsctl.OvsdbVsctl(ovs)
        host_ovs_bridges = set(vsctl.list_br().execute())

        LOG.debug('Found host bridges: %s' % host_ovs_bridges)

        # Need to remove all of the known bridges from the bridge list.
        # Start with removing the integration bridge.
        LOG.debug('Removing integration bridge %s from set' % self.integ_br)
        host_ovs_bridges.remove(self.integ_br)

        # As tunneling is disabled no need to remove tunnel_bridges
        # Filter list of bridges that have external
        # bridge-id's configured.Remove those that have external
        # id same as bridge (these are ancillary bridges)
        br_names = []
        for bridge in host_ovs_bridges:
            ex_id = vsctl.br_get_external_id(bridge, 'bridge').execute()
            if ex_id == bridge:
                br_names.append(bridge)

        LOG.debug('Ancillary bridges found: %s' % br_names)

        # Remove the ancillary bridges from the returned list
        host_ovs_bridges.difference_update(br_names)

        # Remove bridges with name > 8 characters
        to_del = []
        for bridge in host_ovs_bridges:
            if len(bridge) > MAX_BRIDGE_CHARS:
                # We only log if we haven't yet...
                if bridge not in self.bad_vswitches:
                    self._notify_user_of_bad_vswitch(bridge)
                    self.bad_vswitches.append(bridge)

                # We always delete
                to_del.append(bridge)
        for del_item in to_del:
            host_ovs_bridges.remove(del_item)

        LOG.debug('Exit _list_host_ovs_bridges with bridges: %s' %
                  host_ovs_bridges)
        return host_ovs_bridges

    def _check_and_handle_signal(self):
        '''
        The parent method determines if we're to exit or not. This method
        overrides the parent's method.  It will allow us to hook into the
        polling loop and add logic to add/remove host level vswitches
        '''
        # If we're going to exit, no need to process host vswitches
        if not super(PowerKVMNeutronAgent, self)._check_and_handle_signal():
            return False

        if self.host_ovs_check_counter % 15 == 0:
            LOG.debug('Starting PowerKVM _check_and_handle_signal')

            # Check to see if we need to update the vswitches
            ovs_bridge_info = self._update_host_ovs_bridges()

            # If any vswitches were previously missing but have re-appeared,
            # clear their missing counter.
            for key in self.vswitch_missing_counter.keys():
                if key not in ovs_bridge_info['removed']:
                    LOG.debug('Clearing vswitch missing counter for %s.' % key)
                    del self.vswitch_missing_counter[key]

            # notify plugin about ovs_bridge deltas
            if ovs_bridge_info:
                LOG.debug("Updated ovs vswitch info i.e. vswitches "
                          "added/removed is %s" % ovs_bridge_info)
                sync2 = self._process_host_ovs_bridges(ovs_bridge_info)
                # If any vswitch was added/deleted, need to setup corresponding
                # physical bridge in mapping and update/reload configuration
                # Note that this will bring the neutron agent down
                if sync2:
                    LOG.debug('Found vswitches to configure')
                    self.setup_physical_bridges(self.orig_bridge_mappings)

                # verify that all integration bridge and physical ports
                # exist; if not, setup the bridges
                else:
                    new_mappings = \
                        self._check_bridge_port_mappings(
                            self.orig_bridge_mappings)

                    if new_mappings:
                        # save off current data to replace after super call to
                        # setup_physical_bridges since we may not be processing
                        # all bridges
                        curr_phys_brs = copy.copy(self.phys_brs)
                        curr_int_ofports = copy.copy(self.int_ofports)
                        curr_phys_ofports = copy.copy(self.phys_ofports)

                        for phys_network in new_mappings:
                            del curr_phys_brs[phys_network]
                            del curr_int_ofports[phys_network]
                            del curr_phys_ofports[phys_network]

                            # clean up integration bridge flows for the
                            # physical network since these will be
                            # recreated later with an updated int-
                            # port with a new id
                            self.int_br.delete_flows(
                                in_port=self.int_ofports[phys_network])

                        # setup bridges that had missing phy- port
                        self.setup_physical_bridges(new_mappings)

                        # reconfigure vlans belonging to the bridges
                        # being operated on here
                        for bridge_name in new_mappings:
                            self._reconstruct_bridge_vlans(bridge_name)

                        # replace entries for non-affected bridges
                        # that may have been removed during
                        # prior processing
                        for phys_network in curr_phys_brs.iterkeys():
                            self.phys_brs[phys_network] = \
                                curr_phys_brs[phys_network]

                        for phys_network in curr_int_ofports.iterkeys():
                            self.int_ofports[phys_network] = \
                                curr_int_ofports[phys_network]

                        for phys_network in curr_phys_ofports.iterkeys():
                            self.phys_ofports[phys_network] = \
                                curr_phys_ofports[phys_network]

            # check for mappings from newly imported VMs that might
            # not be in the flow mapping file since they were in a shutdown
            # state when imported
            VMFlowMapping.process_imported_vm_mappings()

            if self.host_ovs_check_counter % 150 == 0:
                # Reset the counter so we don't eventually overflow the counter
                self.host_ovs_check_counter = 0

                # clean the flow mapping file of entries for deleted VMs
                self._clean_flow_mapping_file()

            LOG.debug('Finished PowerKVM _check_and_handle_signal')

        self.host_ovs_check_counter += 1

        return True

    def check_changed_vlans(self):
        '''Override community code to be more fault tolerant.

        There is a race condition within the base class's call to
        self.int_br.get_port_tag_dict() where a delete of a VM could be
        occurring as this method is running and it results in a RuntimeError
        being raised because an OVS port for that VM being deleted is present
        when we read it in, however, it is deleted by the time we try to
        get information about it.  This override is to allow at least one
        retry so we don't go into a full sync and possibly cause timeouts
        if there are hundreds of ports to scan through.
        '''
        try:
            changed_ports = super(PowerKVMNeutronAgent, self)\
                .check_changed_vlans()
        except RuntimeError as e:
            LOG.warn(_("Exception checking changed vlans, retrying."))
            LOG.exception(e)

            # Retry the operation, this time letting the exception bubble up
            # if it recurs
            changed_ports = super(PowerKVMNeutronAgent, self)\
                .check_changed_vlans()

        # If we get here, we succeeded
        return changed_ports

    def _clean_flow_mapping_file(self):
        LOG.debug('Starting mapping file cleanup')

        # check for data in flow mapping file for VMs that no longer
        # exist
        for tap_name in VMFlowMapping.get_tap_device_names():
            # check to see if VM still exists
            vm_name = \
                VMFlowMapping.get_vm_name_for_tap_device(tap_name)

            if not vm_name:
                LOG.info(_("No virtual machine found for tap device %s "
                           "in the mapping file.") % tap_name)
                return

            virshout = None
            try:
                LOG.debug("Current contents of %s: %s" %
                          (VMFlowMapping.VM_FLOW_MAP_FILE,
                           VMFlowMapping.current_flow_map))
                # need to check both active and inactive domain xml
                # definitions for interface data
                full_args = ['virsh', 'domiflist', vm_name]
                virshout = utils.execute(full_args, run_as_root=True)

                LOG.debug("Current interface list for %s: %s" %
                          (vm_name, virshout))

                # if we get here, we know the VM still exists, now
                # we just need to see if the tap device is actually
                # assigned to the VM still; if not, remove it.
                if tap_name not in virshout:
                    LOG.warning(_("Removing %(tap_name)s from flow "
                                  "mapping for %(vm_name)s.") %
                                {'tap_name': tap_name,
                                 'vm_name': vm_name})
                    kwargs = {}
                    kwargs[tap_name] = {}
                    VMFlowMapping.update_vm_flow_data(vm_name,
                                                      False,
                                                      **kwargs)

                # check to see if we have entries created from
                # imported vm mapping files (these use the
                # mac address in place of the tap device since
                # the tap device name isn't necessarily available
                # at the time of file creation
                if ":" in tap_name:
                    # parse the virsh domiflist output
                    real_tap_name = None
                    split_virshout = virshout.splitlines()
                    for line in split_virshout:
                        if tap_name in line:
                            split_line = line.split()
                            if len(split_line) > 0:
                                real_tap_name = split_line[0]
                                break

                    # there is a match with an actual tap device for
                    # the mac address; updated mapping file to use that
                    if real_tap_name and real_tap_name != "-":
                        tap_data = VMFlowMapping.get_tap_device_data(tap_name)
                        updated_flow_data = {real_tap_name: tap_data}
                        VMFlowMapping.update_vm_flow_data(vm_name,
                                                          False,
                                                          **updated_flow_data)

                        # delete the data that is using the mac address as
                        # tap device name
                        kwargs = {}
                        kwargs[tap_name] = {}
                        VMFlowMapping.update_vm_flow_data(vm_name,
                                                          False,
                                                          **kwargs)

            except Exception as exc:
                # if we get an exception, assume that it is because the
                # VM does not exist and therefore we need to remove
                # the vm data from the flow mapping file
                LOG.warning(_("Virtual machine %s was not found to exist "
                              "on the host system.  The entry from the "
                              "flow mapping file will be removed.") %
                            vm_name)
                LOG.exception(exc)
                if vm_name:
                    VMFlowMapping.delete_vm_flow_data(vm_name, False)

    def _check_bridge_port_mappings(self, bridge_mappings):
        '''
        check bridge to make sure that int- and phy- ports exist

        :param bridge_mappings: map of physical network names to bridge names
        :returns bridge mapping of bridges that need port configuration
        '''
        new_mapping = {}
        for ovs_bridge in bridge_mappings:
            phy_port = ''.join(["phy-",
                                ovs_bridge])
            bridge_obj = \
                ovs_lib.OVSBridge(ovs_bridge)
            if phy_port not in bridge_obj.get_port_name_list():
                new_mapping[ovs_bridge] = ovs_bridge

        return new_mapping

    def _reconstruct_bridge_vlans(self, bridge_name):
        '''recreate vlans for the specified bridge.  Note that
        this code only handles vlan and flat networks.  Any other
        types will cause an error to be generated.

        :param bridge_name: name of bridge for which vlans will be
            reconstructed
        '''
        for net_uuid, lvl_map in self.local_vlan_map.iteritems():
            physical_network = lvl_map.physical_network

            if physical_network == bridge_name:
                network_type = lvl_map.network_type
                segmentation_id = lvl_map.segmentation_id
                lvid = lvl_map.vlan

                if network_type == p_const.TYPE_FLAT:
                    if physical_network in self.phys_brs:
                        # outbound
                        br = self.phys_brs[physical_network]
                        br.add_flow(
                            priority=4,
                            in_port=self.phys_ofports[physical_network],
                            dl_vlan=lvid,
                            actions="strip_vlan,normal")
                        # inbound
                        self.int_br.add_flow(
                            priority=3,
                            in_port=self.int_ofports[physical_network],
                            dl_vlan=0xffff,
                            actions="mod_vlan_vid:%s,normal" % lvid)
                    else:
                        LOG.error(_("Cannot provision flat network for "
                                    "net-id=%(net_uuid)s - no bridge for "
                                    "physical_network %(physical_network)s"),
                                  {'net_uuid': net_uuid,
                                   'physical_network': physical_network})
                elif network_type == p_const.TYPE_VLAN:
                    if physical_network in self.phys_brs:
                        # outbound
                        br = self.phys_brs[physical_network]
                        br.add_flow(
                            priority=4,
                            in_port=self.phys_ofports[physical_network],
                            dl_vlan=lvid,
                            actions="mod_vlan_vid:%s,normal" % segmentation_id)
                        # inbound
                        self.int_br.add_flow(
                            priority=3,
                            in_port=self.
                            int_ofports[physical_network],
                            dl_vlan=segmentation_id,
                            actions="mod_vlan_vid:%s,normal" % lvid)
                    else:
                        LOG.error(_("Cannot provision VLAN network for "
                                    "net-id=%(net_uuid)s - no bridge for "
                                    "physical_network %(physical_network)s"),
                                  {'net_uuid': net_uuid,
                                   'physical_network': physical_network})
                else:
                    LOG.error(_("Cannot provision unknown network type "
                                "%(network_type)s for net-id=%(net_uuid)s"),
                              {'network_type': network_type,
                               'net_uuid': net_uuid})

    def setup_physical_bridges(self, bridge_mappings):
        '''Setup the physical network bridges.

        Creates physical network bridges and links them to the
        integration bridge using veths.

        :param bridge_mappings: map physical network names to bridge names.
        '''
        new_mappings = {}
        for physical_network, bridge in bridge_mappings.iteritems():
            if len(bridge) > MAX_BRIDGE_CHARS:
                # log error
                self._notify_user_of_bad_vswitch(bridge)
            elif(self.vswitch_missing_counter.get(bridge,
                                                  MAX_BRIDGE_MISSING) ==
                 MAX_BRIDGE_MISSING):
                new_mappings[physical_network] = bridge
        super(PowerKVMNeutronAgent, self).setup_physical_bridges(new_mappings)

    def _notify_user_of_bad_vswitch(self, vswitch_name):
        '''
        Will notify the user that the vSwitch has an invalid name
        '''
        # send error to bumblebee/log file
        msgb = _('Virtual Switch %(vswitch_name)s on Host %(host_name)s has a '
                 'name that exceeds %(chars)s characters.  The system '
                 'ignores any virtual switch with a name longer than '
                 '%(chars)s characters.') % \
            {'vswitch_name': vswitch_name, 'host_name': cfg.CONF.host,
             'chars': str(MAX_BRIDGE_CHARS)}
        info = {'msg': msgb}
        LOG.error(msgb)
        self._notifier.error(self.context, 'ovs.delete.log', info)

    def _update_host_ovs_bridges(self):
        '''
        This method will find the host ovs bridges delta

        :returns: a dict. of host level vswitches added/deleted
        '''
        LOG.debug('Enter _update_host_ovs_bridges')

        # use ovs-vsctl list-br to get the current vswitches on host
        host_ovs_bridges = self._list_host_ovs_bridges()
        bridges_added = []
        bridges_removed = []

        for ovs_bridge in host_ovs_bridges:
            exists = False
            for existing_bridge in self.orig_bridge_mappings:
                # ovs_bridge exists in bridge_mapping hence no
                # processing required
                if ovs_bridge == existing_bridge:
                    exists = True
                    break

            # ovs_bridge does not exist in bridge_mapping implies
            # its new hence append to add list
            if not exists:
                bridges_added.append(ovs_bridge)

        for existing_bridge in self.orig_bridge_mappings:
            exists = False
            for ovs_bridge in host_ovs_bridges:
                # ovs_bridge exists in bridge_mapping hence
                # no processing required
                if existing_bridge == ovs_bridge:
                    exists = True
                    break
            # ovs_bridge does not exist in bridge_mapping implies its old
            # hence append to remove list
            if not exists:
                bridges_removed.append(existing_bridge)

        ret_dict = {'added': bridges_added, 'removed': bridges_removed}
        LOG.debug('Exit _update_host_ovs_bridges with: %s' % ret_dict)
        return ret_dict

    def _process_host_ovs_bridges(self, ovs_bridge_info):
        '''
        This method will process the ovs bridges delta

        :param ovs_bridge_info: dictionary of bridges added/removed
        :returns: boolean flag [true if processed host level vswitches]
        '''
        LOG.debug('Enter _process_host_ovs_bridges with %s' % ovs_bridge_info)

        added_bridges = False
        removed_bridges = False
        if len(ovs_bridge_info['added']) > 0:
            added_bridges = True
            self._add_vswitch_to_config(ovs_bridge_info)

        if len(ovs_bridge_info['removed']) > 0:
            removed_bridges = True
            self._remove_vswitch_from_config(ovs_bridge_info)

            # Remove the ports from the integration bridge manually
            for pt_name in ovs_bridge_info['removed']:
                int_veth_name = constants.PEER_INTEGRATION_PREFIX + pt_name
                self.int_br.delete_port(int_veth_name)

        # If one of the above operations succeeds => need to setup physical
        # bridges
        ret_bool = added_bridges | removed_bridges
        LOG.debug('Exit _process_host_ovs_bridges with %s' % ret_bool)
        return ret_bool

    def port_bound(self, port, net_uuid,
                   network_type, physical_network, segmentation_id,
                   fixed_ips, device_owner, ovs_restarted):
        '''
        This agent overrides the port binding method so that the corresponding
        domain of the compute node can be updated in the scenario of an
        out of band restart of the virtual machine.

        :param port: a ovslib.VifPort object.
        :param net_uuid: the net_uuid this port is to be associated with.
        :param network_type: the network type ('gre', 'vlan', 'flat', 'local')
        :param physical_network: the physical network for 'vlan' or 'flat'
        :param segmentation_id: the VID for 'vlan' or tunnel ID for 'tunnel'
        :param ovs_restarted: indicates if this is called for an OVS restart.
        '''

        # We need to update the XML of the device to persist the internal
        # flow on the device post update.  The first thing we need is the
        # VM name.
        vm_name = self._find_vm_name(port)

        # call neutron agent super class
        super(PowerKVMNeutronAgent, self).port_bound(port, net_uuid,
                                                     network_type,
                                                     physical_network,
                                                     segmentation_id,
                                                     fixed_ips,
                                                     device_owner,
                                                     ovs_restarted)
        if network_type == 'flat':
            segmentation_id = VMFlowMapping.FLAT_NETWORK_VALUE
        # update our internal mapping json to indicate that this interface is
        # using this physical network/segmentation id
        dev_mapping = {}
        dev_mapping[VMFlowMapping.VM_EXTERNAL_VSWITCH_KEY] = physical_network
        dev_mapping[VMFlowMapping.VM_EXTERNAL_VLAN_KEY] = segmentation_id
        kwargs = {port.port_name: dev_mapping}
        LOG.debug("Adding flow data for %s.  External vswitch = %s, "
                  "external vlan = %s" % (vm_name,
                                          physical_network,
                                          segmentation_id))
        VMFlowMapping.\
            add_vm_flow_data(vm_name, False, **kwargs)
        tag = self.int_br.db_get_val("Port", port.port_name, "tag")
        self._update_interface_vlans(vm_name, tag, port.vif_id)
        return True

    def reclaim_local_vlan(self, net_uuid):
        '''
        Reclaim a local VLAN.

        :param net_uuid: the network uuid associated with this vlan.
        '''

        lvm = self.local_vlan_map.get(net_uuid)
        if lvm:
            # scan flow mapping to see if anything else is using
            # this lvm
            external_bridge = lvm.physical_network
            external_vlan = lvm.segmentation_id

            if external_vlan == -1:
                external_vlan = 'flat'

            if VMFlowMapping.is_vlan_in_use(external_bridge,
                                            external_vlan):
                LOG.info(_("Local VLAN mapping for virtual switch %s "
                           "with an external VLAN id of %s is still "
                           "in use.  The local vlan will not be "
                           "reclaimed.") % (external_bridge,
                                            external_vlan))

                # since the only VMs that could be using this
                # vlan are unmanaged VMs, add the vlan to
                # the unmanaged local vlan mapping list
                self.unmanaged_vm_local_vlan_map.append(
                    self.local_vlan_map.pop(net_uuid, None))
            else:
                super(PowerKVMNeutronAgent, self).\
                    reclaim_local_vlan(net_uuid)

    def port_dead(self, port, log_errors=True):
        '''Overriding superclass port_dead to assign vlan tag for
        the current flow that matches for the vm/tap device combo
        in the pvc flow mappings.

        :param port: a ovs_lib.VifPort object.
        '''

        cur_tag = self.int_br.db_get_val("Port", port.port_name, "tag")
        flow_dict = VMFlowMapping.get_tap_device_data(port.port_name)
        LOG.debug("tap device %s flow metadata = %s" % (port.port_name,
                                                        flow_dict))

        if not flow_dict.get(VMFlowMapping.VM_EXTERNAL_VSWITCH_KEY) or \
                flow_dict.get(flow_dict[VMFlowMapping.VM_EXTERNAL_VLAN_KEY]):
            LOG.error(_("Data missing for tap device %s metadata.  Each tap "
                        "device in the metadata must have both the "
                        "external_vswitch and external_vlan parameters "
                        "specified.") % port.port_name)
        if flow_dict:
            flow_tag = \
                self._find_internal_tag_for_external_vswitch(
                    flow_dict[VMFlowMapping.VM_EXTERNAL_VSWITCH_KEY],
                    flow_dict[VMFlowMapping.VM_EXTERNAL_VLAN_KEY])

            if flow_tag == str(constants.DEAD_VLAN_TAG):
                # provision local vlan for this unmanaged vm
                network_type = p_const.TYPE_VLAN
                if flow_dict[VMFlowMapping.VM_EXTERNAL_VLAN_KEY] == \
                        VMFlowMapping.FLAT_NETWORK_VALUE:
                    network_type = p_const.TYPE_FLAT
                flow_tag = self._provision_unmanaged_vm_local_vlan(
                    network_type,
                    flow_dict[VMFlowMapping.VM_EXTERNAL_VSWITCH_KEY],
                    flow_dict[VMFlowMapping.VM_EXTERNAL_VLAN_KEY])

            if cur_tag != flow_tag:
                LOG.debug("updating vlan tagcurrent vlan tag on port = %s, "
                          "vlan tag on flow = %s" % (cur_tag,
                                                     flow_tag))
                self.int_br.set_db_attribute("Port", port.port_name, "tag",
                                             flow_tag)

                # update the VM xml
                vm_name = self._find_vm_name(port)
                self._update_interface_vlans(vm_name, flow_tag, port.vif_id)

    def _bind_devices(self, need_binding_ports):
        '''This method will ensure the proper vlan exists in the local_vlan_map
        so the flows are maintained properly for unmanaged vms.'''
        # Find any vlans we have in our unmanaged list that aren't in our
        # local_vlan_map (this can happen on import).
        for port_detail in need_binding_ports:
            uvlan = self._get_unmanaged_vlan(
                port_detail['physical_network'],
                port_detail['segmentation_id'])
            if uvlan:
                # Put the vlan object into the local_vlan_map so the super
                # class properly sets the tag on the OVS port to match the
                # mod_vlan_vid of the flow.
                self.local_vlan_map[port_detail['network_id']] = uvlan
        # Now let the super class do its thing
        return super(PowerKVMNeutronAgent, self)._bind_devices(
            need_binding_ports)

    def provision_local_vlan(self,
                             net_uuid,
                             network_type,
                             physical_network,
                             segmentation_id,
                             local_vlan=None):
        '''Provisions a local VLAN (overrides super class method).

        :param net_uuid: the uuid of the network associated with this vlan.
        :param network_type: the network type ('gre', 'vxlan', 'vlan', 'flat',
        'local')
        :param physical_network: the physical network for 'vlan' or 'flat'
        :param segmentation_id: the VID for 'vlan' or tunnel ID for 'tunnel'
        '''
        unmanaged_vlan = \
            self._get_unmanaged_vlan(physical_network, segmentation_id)
        if unmanaged_vlan:
            self.local_vlan_map[net_uuid] = unmanaged_vlan

        else:
            super(PowerKVMNeutronAgent, self).\
                provision_local_vlan(net_uuid,
                                     network_type,
                                     physical_network,
                                     segmentation_id)

    def _find_vm_name(self, port):
        '''
        Finds the corresponding virtual machine name for the Port.

        :param port: The internal vif Port object.
        '''
        # Safety first.
        if not port or not port.port_name:
            return None
        ovs = ovs_lib.BaseOVS()
        vsctl = ovsdb_vsctl.OvsdbVsctl(ovs)
        dev_name = ('name', '=', port.port_name)
        out = vsctl.db_find(
            'interface', dev_name, columns=['external_ids']).execute()

        # if the output is bad...return something invalid
        if not out:
            LOG.warn(_('Unable to find VM ID for port %s, no output from '
                       'ovs-vsctl for the interface') % port.port_name)
            return None
        if out[0].get('external_ids', {}).get('vm-id') is None:
            LOG.warn(_('Virtual Machine identifier not returned for port '
                       '%s from ovs-vsctl') % port.port_name)
            return None
        vm_id = out[0]['external_ids']['vm-id']

        virshout = None
        try:
            full_args = ['virsh', 'domname', vm_id]
            virshout = utils.execute(full_args, run_as_root=True,
                                     check_exit_code=[0])
        except OSError as ex:
            LOG.error(ex)
        except Exception as e:
            LOG.error(e)
            raise
        LOG.debug("virsh domname output = %s", virshout)
        return virshout.rstrip()

    def _get_virsh_xml(self, vm_name):
        '''
        This method will send the virsh dumpxml command and return the
        xml output.

        :param vm_name: Name of the domain to return xml data for
        :returns: xml of domain specified by vm_name.
        '''

        if not vm_name:
            return None

        virshout = None
        try:
            full_args = ['virsh', 'dumpxml', vm_name]
            virshout = utils.execute(full_args, run_as_root=True,
                                     check_exit_code=[0])
        except OSError as ex:
            LOG.error(ex)
        except Exception as e:
            LOG.error(e)
            raise
        LOG.debug("virsh dumpxml output = %s", virshout)
        return virshout

    def _attach_virsh_interface(self, vm_name, xml_filename):
        '''
        This method will send the virsh attach command to attach
        the interface to the VM on the next reboot of the vm

        :param vm_name: Name of the domain to return xml data for
        :param xml_filename: Name of xml file containing interface data
        '''

        if not vm_name:
            return None

        try:
            full_args = ['virsh', 'attach-device', vm_name, xml_filename,
                         '--config']
            utils.execute(full_args, run_as_root=True, check_exit_code=[0])
        except OSError as ex:
            LOG.error(ex)
        except Exception as e:
            LOG.error(e)
            raise

    def _detach_virsh_interface(self, vm_name, xml_filename):
        '''
        This method will send the virsh detach command to detach
        the interface from the VM on the next reboot of the vm

        :param vm_name: Name of the domain to return xml data for
        :param xml_filename: Name of xml file containing interface data
        '''

        if not vm_name:
            return None

        try:
            full_args = ['virsh', 'detach-device', vm_name, xml_filename,
                         '--config']
            utils.execute(full_args, run_as_root=True, check_exit_code=[0])
        except OSError as ex:
            LOG.error(ex)
        except Exception as e:
            LOG.error(e)
            raise

    def _update_interface_vlans(self, vm_name, tag, port_id):
        '''
        Invokes two virsh calls for a give VM/port.  The first asks the virsh
        to detach the network device on next reboot.  The second asks the
        virsh to attach the device.

        These calls are made to rebuild the config in the scenario where
        PowerVC is no longer managing the device.  In that scenario it updates
        the config so that the Open vSwitch allows the communication to
        maintain continuity.

        :param vm_name: The virsh name for the VM.
        :param tag: The 'internal vid' that should be tagged onto the port
        :param port_id: The unique identifier for the port.
        '''

        try:
            # make sure tag is in string format
            if isinstance(tag, list) and not len(tag):
                tag = ""
            if not isinstance(tag, basestring):
                tag = str(tag)

            xml_attach_filename = "/tmp/pvc_vm_interface.xml"
            xml_detach_filename = "/tmp/pvc_vmd_interface.xml"
            domxml = self._get_virsh_xml(vm_name)

            doc = etree.fromstring(domxml)

            # get the interface elements from the domain
            ret = doc.findall('./devices/interface')

            for node in ret:
                virtport = node.findall('./virtualport/parameters')[0]

                # only update the interface vlan for the current port
                # NOTE: interface id on the virtual port in the interface
                # XML should match the neutron port id
                if virtport.attrib['interfaceid'] != port_id:
                    continue

                # remove pci address to lessen the chance of a
                # pci conflict during the detact/attach sequence
                for child in node:
                    if child.tag == 'address':
                        node.remove(child)

                    elif child.tag == 'target':
                        # the the tap device name as the xml file name for
                        # attach/detach device calls
                        xml_attach_filename = \
                            "/tmp/%s_vm_interface.xml" % child.attrib['dev']
                        xml_detach_filename = \
                            "/tmp/%s_vm_d_interface.xml" % child.attrib['dev']

                # Save off the detach XML prior to updating with the XML
                etree.ElementTree(node).write(xml_detach_filename)

                # Update the XML so that it includes the Tag
                vlan_el = doc.find('./vlan')
                if vlan_el is not None and \
                        vlan_el.find('./tag').get('id') == tag:
                    vlan_el.find('./tag').set('id', tag)
                elif tag:
                    vlan_el = etree.Element("vlan")
                    vlan_el.append(etree.Element("tag", id=tag))
                    node.append(vlan_el)

                etree.ElementTree(node).write(xml_attach_filename)

                # detach the interface
                self._detach_virsh_interface(vm_name,
                                             xml_detach_filename)

                # attach interface
                self._attach_virsh_interface(vm_name,
                                             xml_attach_filename)

                # get rid of the xml file since we no longer need it
                os.remove(xml_attach_filename)
                os.remove(xml_detach_filename)

        except Exception as ex:
            LOG.exception(ex)

    def _find_internal_tag_for_external_vswitch(self,
                                                external_bridge,
                                                external_vlan='flat'):
        """
        Given the name of an external bridge and vlan, trace back through to
        the internal bridge (this will usually be br-int) to find the complete
        flow and determine the proper tag to assign to ovs ports using the
        given external bridge.

        :param vswitch_name: Name of external bridge to trace flow back from
        :param vlan_id:  String value of the vlan id for specified bridge.  If
            set to 'flat' it is assumed this is a flat network
        :returns: internal vlan tag
        """

        # first check internal vlan mapping for flow data
        for lvl in self.local_vlan_map.itervalues():
            if (lvl.segmentation_id == str(external_vlan) or
                    (external_vlan == 'flat' and lvl.network_type == 'flat')) \
                    and lvl.physical_network == external_bridge:
                LOG.debug("found match for external vswitch/external vlan "
                          "combo %s/%s in local vlan map." %
                          (external_bridge, external_vlan))
                return lvl.vlan

        # retrieve necessary ovs-ofctl data needed for flow computation
        ex_br_flows = self._get_flow(external_bridge, external_vlan)
        ex_br_flow_ports = self._get_bridge_ports(external_bridge)

        # the naming convention for the port responsible on the external
        # bridge for controlling communication with the internal bridge
        # is phy-<external_vswitch_name>
        ex_br_in_port = ex_br_flow_ports.get("phy-%s" % external_bridge)

        if ex_br_flows and ex_br_flows['in_port'] == ex_br_in_port:
            # If vlan_id is specified, we must match on in_port and
            # vlan id.  If it is not specified, just need to match on
            # in_port
            if external_vlan == 'flat' or (ex_br_flows['mod_vlan_vid'] ==
                                           str(external_vlan)):
                LOG.debug("Flow for external vswitch/vlan id "
                          "combination found.  Tag value = "
                          "%s" % ex_br_flows['dl_vlan'])
                return ex_br_flows['dl_vlan']

        # no matching flow found, return 4095
        return str(constants.DEAD_VLAN_TAG)

    def _get_interfaces_for_vm(self, vm_name):
        '''
        This method will list all the interfaces on the passed in vm and
        return a list of dictionaries that look like the following:

        [ {"interface": "tapxxx", "type": "bridge", "source": "br-int",
           "model": "virtio", "mac": "fa:16:3e:49:80:4a",
           "segmentation_id": "5", "physical_network": "default0"},
          {"interface": "tapyyy", "type": "bridge", "source": "br-int",
           "model": "virtio", "mac": "fa:16:3e:49:80:4b",
           "segmentation_id": "-1", "physical_network": "second"},
          ...
        ]

        NOTE: A segmentation_id of -1 indicates a flat network.

        :param vm_id: name of the vm/domain
        :returns: list of interfaces
        '''
        ret_list = list()

        if vm_name:
            virshout = None

            try:
                full_args = ['virsh', 'domiflist', vm_name]
                virshout = utils.execute(full_args, run_as_root=True,
                                         check_exit_code=[0])
            except OSError as ex:
                LOG.error(ex)
            except Exception as e:
                LOG.error(e)
                raise
            LOG.debug("virsh domiflist output = %s", virshout)

            if virshout:
                lines = virshout.splitlines()
                for line in lines:
                    # Skip the header lines
                    if(not line.strip() or 'Interface' in line or
                       '---------' in line):
                        continue
                    items = line.split()
                    interface = dict()
                    interface['interface'] = items[0]
                    interface['type'] = items[1]
                    interface['source'] = items[2]
                    interface['model'] = items[3]
                    interface['mac'] = items[4]
                    ret_list.append(interface)

        return ret_list

    def _get_ovs_vsctl_show(self):
        """
        This method is used to send the ovs-vsctl command and get the parsed
        output

        :return: parsed dictionary of the object of the vsctl command
        {'br0': {'br0': {'interface': ['eth0'], 'type': 'internal'},
                '"9"': {'interface': ['"9"']},
                '"br1"': {'interface': ['"br1"', '"arbit"'],
                                   'type': 'internal'}}}
        """
        vsctlout = ""
        try:
            vsctlout = utils.execute(['ovs-vsctl', 'show'],
                                     run_as_root=True,
                                     check_exit_code=[0])
        except OSError as ex:
            LOG.error(ex)
        except Exception as e:
            LOG.error(e)
            raise
        LOG.debug("ovs-vsctl show output = %s", vsctlout)

        bridge_dict = {}

        if vsctlout is None or vsctlout == "":
            return bridge_dict

        i = 0

        # Formulate the bridge dictionary to contain the bridge status for
        # vsctl
        for bridge_chunk in vsctlout.split('\n    Bridge '):
            if bridge_chunk.strip() == "":
                continue
            if i == 0:
                i = i + 1
                continue
            j = 0
            # Let'stripe off the escape characters if used
            bridge_name = bridge_chunk.split("\n")[0].strip().strip('""')
            port_dict = {}
            for port_chunk in bridge_chunk.split('\n        Port '):
                # In that sub-chunk we would want to split it using the
                # port.
                if j != 0:
                    split_port_list = port_chunk.split('\n')
                    # get the port name out from the first element of the
                    # list
                    port_name = split_port_list[0].strip().strip('""')
                    interface_list = []
                    interface_dict = {}
                    # Next we get the items off the split port chunk
                    for item in split_port_list:
                        if "            Interface " in item:
                            iface = item.replace('            Interface ',
                                                 '', 1).strip().strip('""')
                            interface_list.append(iface)
                        if "type" in item:
                            interface_dict['type'] = item.split()[1]
                        if "tag:" in item:
                            interface_dict['tag'] = item.split()[-1].strip()
                    interface_dict['interface'] = interface_list
                    port_dict[port_name] = interface_dict
                j = j + 1
            bridge_dict[bridge_name] = port_dict
            i = i + 1
        LOG.debug("ovs-vsctl show returning = %s", bridge_dict)
        return bridge_dict

    def _find_vswitch_vlan(self, interface):
        """
        For the specified interface, compute and return the flow from
        the internal vswitch through to the external vswitch and
        its connection with the outside world

        :param interface: interface dictionary
        :returns: flow as dictionary
        """
        if not interface:
            LOG.error(_('No interface passed in to _find_vswitch_vlan'))
            return None, None

        # get integration bridge data
        intbr = self._get_ovs_vsctl_show().get(self.integ_br)
        if not intbr:
            LOG.error(_("The ovs-vsctl information is missing "
                        "for the integration vswitch. Ensure "
                        "that the integration bridge exists "
                        "and try again.%(integbr)s" %
                        {'integbr': self.integ_br}))
            return None, None

        # Get all the flow information as well
        intbr_flows = self._get_ofctl_dump_flows(self.integ_br)

        intbr_flow_ports = self._get_ofctl_show_by_port_id(self.integ_br)

        LOG.debug('%(intbr)s\nintbr_flows: %(flows)s' %
                  {'intbr': intbr,
                   'flows': intbr_flows})

        # See if the interface is on the integration bridge (note, it could
        # be on a different vswitch or on linuxbridge)
        tag = 0
        for port in intbr:
            if interface.get('interface') in intbr[port]['interface']:
                tag = intbr[port].get('tag')
                break

        if tag == 0:
            # no matching flow found on the integration bridge, so no
            # flow data available
            return None, None

        # We find the interface's flow by matching the flow's internal vid
        # to the interface's tag
        correct_flow = None
        for flow in intbr_flows:
            if flow.get('mod_vlan_vid') == tag:
                correct_flow = flow

        if correct_flow:
            # Grab the vlan (use 'segmentation_id' to match vm_importer).
            # If it doesn't exist, it's a flat network, which we'll
            # signify with a -1.
            vlan = (correct_flow.get('dl_vlan')
                    if correct_flow.get('dl_vlan')
                    else '-1')

            # Now find the external vswitch this interface is potentially
            # on.
            in_port = correct_flow.get('in_port')
            if not in_port:
                return None, None

            host_ovs = intbr_flow_ports.get(in_port)
            return host_ovs, vlan

        # Didn't find a matching flow.
        LOG.warn('No flow found for interface %s', interface.get('interface'))
        return None, None

    def _get_vm_list(self):
        '''
        This method will send the virsh list command and return the parsed
        output.

        :returns: Dictionary mapping vm_name to vm_id.
        '''
        # get virsh list of all VMs
        virshout = ''
        try:
            virshout = utils.execute(['virsh', 'list'],
                                     run_as_root=True,
                                     check_exit_code=[0])
        except OSError as ex:
            LOG.error(ex)
        except Exception as e:
            LOG.error(e)
            raise
        LOG.debug("virsh list output = %s", virshout)

        # parse the output data
        ret_dict = dict()
        if virshout:
            lines = virshout.splitlines()
            for line in lines:
                # Skip the header lines
                if not line.strip() or 'Name' in line or '---------' in line:
                    continue
                items = line.split()
                ret_dict[items[1]] = items[0]

        return ret_dict

    def _get_flow(self, external_bridge, external_vlan):
        '''
        Find a flow that matches the specified external bridge and
        vlan.

        :param external_bridge: Name of the external bridge/vswitch
        :param external_vlan: external vlan id

        :returns: Information about a single flow in a dict, dict is
            empty if no match is found
        '''
        flows = self._get_ofctl_dump_flows(external_bridge)

        for flow in flows:
            # check for flat network if specified by external
            # vlan value
            if external_vlan == 'flat' \
                    and not flow.get("mod_vlan_vid"):
                return flow

            # check for flows with external vlan specified
            elif flow.get("mod_vlan_vid") == str(external_vlan):
                return flow

        return {}

    def _get_ofctl_show_by_port_id(self, bridge):
        '''
        This method dumps the flow ports for the input bridge name returns
        the parsed output.

        :param bridge: Name of the bridge we are getting data for
        :returns: Dictionary representing the flow ports on the bridge.
                  Like so (port id acts as key):
            { '108': 'default0', '109': 'second', ... }
        '''
        if not bridge:
            return list()

        ofctlout = ''

        br_obj = ovs_lib.OVSBridge(bridge)
        ofctlout = br_obj.run_ofctl("show", [])

        """
        ofctlout = ''
        try:
            ofctlout, stderr = utils.execute('ovs-ofctl', 'show',
                                             bridge, run_as_root=True)
        except OSError as ex:
            LOG.error(ex)
        except Exception as e:
            LOG.error(e)
            raise
        LOG.debug("ovs-ofctl show output = %s", ofctlout)
        """
        ret_dict = dict()

        if ofctlout:
            # Loop through each line
            lines = ofctlout.split('\n')
            for line in lines:
                if not line.strip() or "addr:" not in line:
                    # Skip everything but the port(vswitch) lines
                    continue
                # Parse the line.  Note, strip off the "int-" to get the
                # real vswitch name
                parts = line.split('(')
                parts[0] = parts[0].strip()
                parts[1] = parts[1].split(')')[0]
                if 'int-' in parts[1]:
                    parts[1] = parts[1].split('-')[1]
                ret_dict[parts[0]] = parts[1]

        # Return what we parsed
        return ret_dict

    def _get_ofctl_dump_flows(self, br_name):
        br_obj = ovs_lib.OVSBridge(br_name)
        output = br_obj.run_ofctl("dump-flows", [])

        ret_list = list()

        if output:
            # Loop through each line
            for line in output.splitlines():
                if not line.strip() or "NSXT_FLOW reply" in line:
                    # Skip header line
                    continue
                # Parse the line
                this_dict = dict()
                for item in line.split(','):
                    if 'in_port' in item:
                        this_dict['in_port'] = item.split('=')[-1]
                    elif 'dl_vlan' in item:
                        item = item.split(' ')
                        this_dict['dl_vlan'] = item[0].split('=')[-1]
                        if len(item) > 1 and 'mod_vlan_vid' in item[1]:
                            this_dict['mod_vlan_vid'] = \
                                item[1].split(':')[-1]
                    elif 'mod_vlan_vid' in item:
                        # This is for the 'flat' network scenario
                        item = item.split(' ')
                        this_dict['dl_vlan'] = -1
                        if 'mod_vlan_vid' in item[1]:
                            this_dict['mod_vlan_vid'] = \
                                item[1].split(':')[-1]

                # If we parsed info that's useful to us,
                # append it to the list.
                if this_dict:
                    ret_list.append(this_dict)

        # Return what we parsed
        return ret_list

    def _get_bridge_ports(self, bridge):
        '''
        Return a list of all ovs ports on the specified bridge

        :param bridge: Name of bridge
        :returns: List of information for all ports on the bridge,
            with each element in the list being a dict of the
            following format:

            { 'default0': '108', 'second': '109', ... }
        '''
        ext_br_obj = ovs_lib.OVSBridge(bridge)
        output = ext_br_obj.run_ofctl("show", [])

        def parse_ofctl_show(show_output):
            ret_dict = dict()

            if show_output:
                # Loop through each line
                lines = show_output.split('\n')
                for line in lines:
                    if not line.strip() or "addr:" not in line:
                        # Skip everything but the port(vswitch) lines
                        continue
                    # Parse the line.  Note, strip off the "int-" to get the
                    # real vswitch name
                    parts = line.split('(')
                    parts[0] = parts[0].strip()
                    parts[1] = parts[1].split(')')[0]
                    if 'int-' in parts[1]:
                        parts[1] = parts[1].split('-')[1]
                    ret_dict[parts[1]] = parts[0]

            # Return what we parsed
            return ret_dict

        return parse_ofctl_show(output)

    def _provision_unmanaged_vm_local_vlan(self,
                                           network_type,
                                           physical_network,
                                           segmentation_id):
        '''
        Create flow needed by an unmanaged vm based on vlan_mapping

        :param network_type: the network type ('vlan', 'flat')
        :param physical_network: the physical network for 'vlan' or 'flat'
        :param segmentation_id: the VID for 'vlan'

        :returns: local vlan tag assigned
        '''
        lvid = self.unmanaged_vm_available_local_vlans.pop()

        while lvid not in self.available_local_vlans:
            lvid = self.unmanaged_vm_available_local_vlans.pop()

        # remove this vlan tag from the main available vlan tag list
        self.available_local_vlans.remove(lvid)
        self.unmanaged_vm_local_vlan_map.append(
            ovs_neutron_agent.LocalVLANMapping(lvid,
                                               network_type,
                                               physical_network,
                                               segmentation_id))

        if network_type == p_const.TYPE_FLAT:
            if physical_network in self.phys_brs:
                # outbound
                br = self.phys_brs[physical_network]
                br.add_flow(priority=4,
                            in_port=self.phys_ofports[physical_network],
                            dl_vlan=lvid,
                            actions="strip_vlan,normal")
                # inbound
                self.int_br.add_flow(
                    priority=3,
                    in_port=self.int_ofports[physical_network],
                    dl_vlan=0xffff,
                    actions="mod_vlan_vid:%s,normal" % lvid)
            else:
                LOG.error(_("Cannot provision flat network - no bridge for "
                            "physical_network %(physical_network)s"),
                          {'physical_network': physical_network})
        elif network_type == p_const.TYPE_VLAN:
            if physical_network in self.phys_brs:
                # outbound
                br = self.phys_brs[physical_network]
                br.add_flow(priority=4,
                            in_port=self.phys_ofports[physical_network],
                            dl_vlan=lvid,
                            actions="mod_vlan_vid:%s,normal" % segmentation_id)
                # inbound
                self.int_br.add_flow(priority=3,
                                     in_port=self.
                                     int_ofports[physical_network],
                                     dl_vlan=segmentation_id,
                                     actions="mod_vlan_vid:%s,normal" % lvid)
            else:
                LOG.error(_("Cannot provision VLAN network - no bridge for "
                            "physical_network %(physical_network)s"),
                          {'physical_network': physical_network})

        return lvid

    def _get_unmanaged_vlan(self, external_vswitch, external_vlan):
        '''
        See if an unmanaged vlan for the specified vswitch/vlan combo
        exists, if so, return the LocalVlanMapping object

        :param external_vswitch: name of external vswitch
        :param external_vlan: value of external_vlan (integer for vlan
            networks, 'flat' for flat network
        :returns: LocalVlanMapping object if exists, otherwise None
        '''

        for vlan in self.unmanaged_vm_local_vlan_map:
            if vlan.physical_network == external_vswitch and \
                    (vlan.segmentation_id == 'flat' or
                     vlan.segmentation_id == external_vlan):
                return vlan

        return None


def main():
    from neutron.agent.linux import ip_lib
    eventlet.monkey_patch()
    cfg.CONF.register_opts(ip_lib.OPTS)
    # cfg.CONF(project='neutron')
    logging_config.init(sys.argv[1:])
    logging_config.setup_logging()

    roothelper = cfg.CONF.AGENT.get('root_helper', None)
    # check if ovs service is running, else start it
    agent.check_and_start_ovs_service(roothelper)

    # check if integration bridge exists
    plugin_ini = cfg.CONF.find_file('openvswitch_agent.ini')
    conf = iniparse.ConfigParser()
    conf.readfp(open(plugin_ini))

    # ensure that integration bridge br-int is present
    # file has the integration bridge name
    int_br_name = conf.get('OVS', 'integration_bridge')

    # create the integration bridge,if it does not exist
    if not exists_bridge(int_br_name):
        LOG.info(_("Creating integration bridge %s") % int_br_name)
        create_integration_bridge(int_br_name)

    config_bridge_map_list = cfg.CONF.OVS.get('bridge_mappings', [])
    config_bridge_map_dict = n_utils.parse_mappings(config_bridge_map_list)
    for k, v in config_bridge_map_dict.items():
        if not exists_bridge(k):
            # remove bridge from in memory config and the file
            config_bridge_map_list.remove(k + ':' + v)
    # update conf file
    PowerKVMNeutronAgent.set_bridge_mapping_in_conf(config_bridge_map_dict)
    # update bridge_mappings in agent_config
    cfg.CONF.OVS.bridge_mappings = config_bridge_map_list

    plugin = PowerKVMNeutronAgent(BRIDGE_CLASSES, cfg.CONF)

    # Start polling
    LOG.info(_("Agent initialized successfully, now running... "))
    plugin.daemon_loop()
    sys.exit(0)


def create_integration_bridge(bridge_name):
    '''
    Will create an Open vSwitch on the system with the specified
    name.
    :param bridge_name: The vSwitch name to create.
    '''
    ovs_lib.BaseOVS().add_bridge(bridge_name)


def exists_bridge(bridge_name):
    '''
    Will check if an Open vSwitch with the specified name exists
    :param bridge_name: The vSwitch name to check.
    '''
    if len(bridge_name) > MAX_BRIDGE_CHARS:
        LOG.error(_('Invalid vSwitch referenced in CONF file named %(brdg)s. '
                    'Virtual Switches must not exceed %(chars)s characters in '
                    'their name.') % {'brdg': bridge_name,
                                      'chars': str(MAX_BRIDGE_CHARS)})
        return False
    return ovs_lib.BaseOVS().bridge_exists(bridge_name)


if __name__ == '_main_':
    main()
