#
#
# Copyright 2015 IBM Corp.
# All Rights Reserved.
#
#   Copyright 2011 Nicira Networks, Inc.
#   All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import copy
import eventlet
from neutron.common import config as n_config
from networking_powervm.plugins.ibm.agent.powervm import agent_base
from networking_powervm.plugins.ibm.agent.powervm import sea_agent
from networking_powervm.plugins.ibm.agent.powervm import utils as pvm_utils
from oslo_config import cfg
from oslo_log import log as logging
from powervc_neutron.plugins.pvm.agent import constants
from powervc_neutron.plugins.pvm.agent import utils
from pypowervm import adapter as pvm_adpt
from pypowervm.helpers import log_helper as log_hlp
from pypowervm.tasks import network_bridger as net_br
import sys
import time

from powervc_neutron import _

keystone_opts = [
    cfg.StrOpt('auth_uri', default=None),
    cfg.StrOpt('username', default=None),
    cfg.StrOpt('password', default=None, secret=True),
    cfg.StrOpt('project_name', default=None)
]

pvm_opts = [
    cfg.StrOpt('ip', default=None,
               help='The IP address of the Novalink partition.'),
    cfg.StrOpt('username', default=None,
               help='The username of the Novalink partition.'),
    cfg.StrOpt('password', default=None, secret=True,
               help='The password for the Novalink partition.'),
    cfg.IntOpt('host_available_timeout', default=600,
               help='The amount of time, in seconds, to wait for this host '
               'to become available in nova.  Exceeding this timeout value '
               'will prevent the Shared Ethernet Agent from starting.')
]

CONF = cfg.CONF
CONF.register_opts(keystone_opts, group='keystone_authtoken')
CONF.register_opts(pvm_opts, group='pvm')
ACONF = CONF.AGENT

LOG = logging.getLogger(__name__)


class PVMSharedEthernetNeutronAgent(sea_agent.SharedEthernetNeutronAgent):
    '''
    Neutron agent to listen for changes and apply them to SEAs via pypowervm.

    The base class takes care of configuring the SEAs.  This override is
    necessary so we can dynamically update the conf file and in-memory
    configuration whenever a new vlan/vswitch combo is being configured.  This
    is because PowerVC uses only 1 physical network name (called "default"),
    but we need physical network name to map to a specific SEA... and if there
    are multiple SEAs, then having all networks called "default" does not work.
    '''

    def __init__(self):
        super(PVMSharedEthernetNeutronAgent, self).__init__()

        LOG.info(_('PowerVC SEA Agent config complete'))

    def setup_adapter(self):
        '''
        Set up adapter based on conf options, if present.  Otherwise, default.
        '''
        if CONF.pvm.ip and CONF.pvm.username and CONF.pvm.password:
            self.adapter = pvm_adpt.Adapter(
                pvm_adpt.Session(host=CONF.pvm.ip, username=CONF.pvm.username,
                                 password=CONF.pvm.password, certpath=None),
                helpers=log_hlp.log_helper)
            self.host_uuid = pvm_utils.get_host_uuid(self.adapter)
            evt_listener = self.adapter.session.get_event_listener()
            self._cna_event_handler = agent_base.CNAEventHandler(self)
            evt_listener.subscribe(self._cna_event_handler)
        else:
            super(PVMSharedEthernetNeutronAgent, self).setup_adapter()

        # Start with a fresh set of bridge mappings in conf file
        utils.initialize_bridge_mappings(self.adapter, self.host_uuid)

        # We must loop and wait for this host to be available in nova before
        # proceeding otherwise novaclient calls can fail due to the database
        # not being populated yet.
        retry = True
        novaclient = utils.create_novaclient(CONF.keystone_authtoken)
        start = time.time()
        while retry:
            # See if this host is available in nova...
            if self._host_available(novaclient):
                retry = False
            elif time.time() - start > CONF.pvm.host_available_timeout:
                raise Exception('Host failed to become available in nova')
            else:
                # Host not available but we haven't timed out.  Sleep 5 seconds
                time.sleep(5)

        # Now set up the caches to be sure they're available right away.  Do
        # this in setup_adapter because we know this is before any processing
        # in the sea agent has started.
        self._refresh_ongoing_caches()
        self._refresh_provision_caches()

    def _host_available(self, novaclient):
        '''This method will see if this host is available in nova.

        :param novalclient: novaclient to make os-hypervisors call with
        :return: boolean indicating host available
        '''
        response = novaclient.client.get('/os-hypervisors')
        hosts = [x.get('hypervisor_hostname').encode('utf-8') for x in
                 response[1]['hypervisors']] if response[0].status_code == 200\
            else []
        return CONF.host in hosts

    def setup_rpc(self):
        '''Override simply so we can register our own agent/vif type'''
        self.agent_state['binary'] = constants.AGENT_BINARY_PVC_SEA
        self.agent_state['agent_type'] = constants.AGENT_TYPE_PVC_SEA
        super(PVMSharedEthernetNeutronAgent, self).setup_rpc()

    def _refresh_ongoing_caches(self):
        '''Refresh caches that rarely change.

        These caches are those that rarely change, so we only refresh them
        during heal_and_optimize rather than every provision.
        '''
        self.vios_map = utils.get_vios_map(self.adapter, self.host_uuid)

    def _refresh_provision_caches(self):
        '''Refresh caches used on provision.

        These caches need to be absolutely fresh for each provision batch.
        '''
        novaclient = utils.create_novaclient(CONF.keystone_authtoken)

        LOG.debug('Caching host-seas')
        self.host_sea_resp = novaclient.client.get('/host-seas/%s' % CONF.host)
        LOG.debug('Finished caching host-seas')

        # This call to network mapping will actually create all mappings
        # required so we don't have to worry about creating default mappings
        # later on.
        LOG.debug('Caching network mappings')
        self.net_map_resp = novaclient.client.get(
            '/host-network-mapping/%s?skip_check=true' % CONF.host)
        LOG.debug('Finished caching network mappings')

    def heal_and_optimize(self, is_boot, prov_reqs, lpar_uuids, overall_cnas):
        """Override so we can refresh cache items that are too expensive to
        do for every CNA"""
        LOG.debug('Beginning heal_and_optimize')
        # Refresh the caches
        self._refresh_ongoing_caches()
        self._refresh_provision_caches()

        super(PVMSharedEthernetNeutronAgent, self).\
            heal_and_optimize(is_boot,
                              prov_reqs,
                              lpar_uuids,
                              overall_cnas)

        LOG.debug('Finished heal_and_optimize')

    def provision_devices(self, requests):
        '''Will ensure that the VLANs are on the NBs for the edge devices.

        This override is done so we can cache the host-seas and network
        mapping responses so each _get_nb_and_vlan doesn't need to pay the
        price.'''
        self._refresh_provision_caches()
        super(PVMSharedEthernetNeutronAgent, self).provision_devices(requests)

    def _get_nb_and_vlan(self, dev, emit_warnings=False):
        """Parses bridge mappings to find a match for the device passed in.
        PowerVC uses 'default' for physical_network on all networks.  We must
        look up the appropriate bridge based on vlan/vswitch and return that.
        If one does not exist, then we must look up the appropriate bridge via
        network mapping.  Note that None, None can be returned for CNAs that
        are not in neutron.

        :param dev: Neutron device to find a match for
        :param emit_warnings: (Optional) Defaults to False.  If true, will emit
                              a warning if the configuration is off.

        :return: UUID of the NetBridge
        :return: vlan for the neutron device
        """
        vlan = dev.get('segmentation_id')
        netid = dev.get('network_id')

        # If this is not a neutron network, skip it
        if netid is None:
            LOG.debug('No netid for %s' % dev)
            return None, None

        sea_name, vios_name = utils.find_network_mapping(
            CONF.keystone_authtoken, CONF.host, netid, self.vios_map,
            self.net_map_resp)

        # Validate we got good data
        if sea_name is None:
            if emit_warnings:
                LOG.warn(_('Unable to find Shared Ethernet Agent for '
                           'network %s'), netid)
            return None, None
        if vios_name is None:
            if emit_warnings:
                LOG.warn(_('Unable to find Virtual I/O Server for '
                           'network %s'), netid)
            return None, None

        vswitch = utils.find_vswitch_for_sea(
            CONF.keystone_authtoken, CONF.host, sea_name, vios_name,
            self.host_sea_resp)

        # Validate we got good data
        if vlan is None or vswitch is None:
            if emit_warnings:
                LOG.warn(_('Unable to find vlan or vswitch for device %s'),
                         dev)
            return None, None

        # Quick check to see if we need to create the mapping or if current
        # mapping is correct.
        physnet = utils.build_physnet_string(vswitch, vlan)
        self._validate_or_update_bridge_mapping(physnet, sea_name, vios_name)
        return self.br_map.get(physnet), vlan

    def _validate_or_update_bridge_mapping(self, physnet, sea_name, vios_name):
        """Ensure mappings on disk and in memory are up to date.

        :param physnet: Physical network name to use
        :param sea_name: Name of SEA for this physnet
        :param vios_name: Name of VIOS for this physnet
        """
        # First, update mappings on disk
        mappings = utils.write_bridge_mappings(physnet, sea_name, vios_name)

        # If something was returned, mappings were updated
        if mappings is not None:
            ACONF.bridge_mappings = mappings
            self.br_map = pvm_utils.parse_sea_mappings(
                self.adapter, self.host_uuid, mappings)


def main():
    # Read in the command line args
    n_config.init(sys.argv[1:])
    n_config.setup_logging()

    # Build then run the agent
    agent = PVMSharedEthernetNeutronAgent()
    LOG.info(_('PowerVC Shared Ethernet Agent initialized and running'))
    agent.rpc_loop()


if __name__ == '_main_':
    main()
