# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from keystoneclient.auth.identity import v3 as auth_v3
from keystoneclient import session
from networking_powervm.plugins.ibm.agent.powervm import utils as pvm_utils
from novaclient.client import Client
from oslo_concurrency import lockutils
from oslo_log import log as logging
from powervc_neutron import _
from powervc_neutron.plugins.pvm.agent import constants
from powervc_oslo.config import data_utils as oslo_utils
from pypowervm.utils import retry as pvm_retry
from pypowervm.utils import uuid as pvm_uuid
from pypowervm.wrappers import managed_system as pvm_ms
from pypowervm.wrappers import network as pvm_net
import pypowervm.wrappers.virtual_io_server as pvm_vios
from pypowervm import const as pvm_const

LOG = logging.getLogger(__name__)
_AUTH, _SESSION = None, None


def build_physnet_string(vswitch, vlan):
    """Build the physical network string to be used in conf bridge_mappings

    Example: 'ETHERNET0-VLAN1'

    :param vswitch: vswitch name
    :param vlan: vlanid
    :return: String for the physical network
    """
    return '%(vswitch)s-VLAN%(vlan)d' % ({'vswitch': vswitch, 'vlan': vlan})


def build_bridge_mapping(physnet, sea_name, vios_name):
    """Build the full bridge mapping string for conf bridge_mappings

    Example: 'ETHERNET0-VLAN1:ent9:vios1'

    :param physnet: Physical network name
    :param sea_name: SEA to map to
    :param vios_name: VIOS the SEA resides on
    :return: bridge mapping string
    """
    return '%(physnet)s:%(sea)s:%(vios)s' % {
        'physnet': physnet, 'sea': sea_name, 'vios': vios_name}


def find_network_mapping(kconf, host_name, netid, vios_map, response=None):
    """Find the network mapping for this host based on vswitch/vlan

    :param kconf: keystone_authtoken CONF for creating novaclient
    :param host_name: Name of host we're running on
    :param netid: Neutron network ID to look for mapping
    :param vios_map: Dictionary mapping lpar_id to lpar_name of host's vioses
    :param response: Cached response to use, if any
    :return: SEA name
    :return: VIOS name
    """
    # If we weren't given a response, make the call
    if response is None:
        response = create_novaclient(kconf).client.get(
            '/host-network-mapping/%s?network_id=%s&skip_check=true' %
            (host_name, netid))

    # Get the network mapping entry for just our host
    networks = []
    if response[0].status_code == 200:
        net_maps = [x for x in response[1].get('host-network-mapping', [])
                    if x.get('host_name') == host_name]
        networks.extend(net_maps[0].get('networks', []) if net_maps else [])

    sea_name = None
    vios_name = None
    # Find the sea/vios for our particular network
    for network in networks:
        if netid == network.get('network_id'):
            primary_sea = network.get('primary_sea')
            if primary_sea:
                if primary_sea == 'not-applicable':
                    continue
                sea_name = primary_sea['name']
                if sea_name:
                    sea_name = sea_name.encode('utf-8')
                lpar_id = primary_sea['lpar_id']
                if lpar_id:
                    vios_name = vios_map.get(str(lpar_id))
                return sea_name, vios_name

    # If we get here, something went wrong above
    LOG.warn(_('Failed to find SEA for network %(netid)s from %(nets)s'),
             {'netid': netid, 'nets': networks})
    return sea_name, vios_name


def find_vswitch_for_sea(kconf, host_name, sea_name, vios_name,
                         response=None):
    '''Find vswitch name for given SEA/VIOS

    :param kconf: keystone_authtoken CONF for creating novaclient
    :param host_name: Name of host we're running on
    :param sea_name: Name of SEA to find
    :param vios_name: Name of VIOS containing SEA
    :param response: Cached response to use, if any
    :return: vswitch name
    '''
    # If we weren't given a response, make the call
    if response is None:
        response = create_novaclient(kconf).client.get(
            '/host-seas/%s' % host_name)

    # host-seas query should have only been for this host, so no need to check
    adapters = []
    if response[0].status_code == 200:
        host_seas = [x for x in response[1].get('host-seas', [])
                     if x.get('host_name') == host_name]
        adapters.extend(host_seas[0].get('adapters', []) if host_seas else [])

    # Now find the right one!
    for adapter in adapters:
        s_name = adapter['sea_name'].encode('utf-8')
        v_name = adapter['lpar_name'].encode('utf-8') if adapter['lpar_name']\
            else None
        if s_name == sea_name and v_name == vios_name:
            return adapter['vswitch'].encode('utf-8')

    # If we got here, didn't find it.  Boo.
    return None


@pvm_retry.retry()
def get_vios_map(adapter, host_uuid):
    """Returns a dictionary mapping lpar_id to lpar_name for host's vioses.

    Example: {'1': 'vios1', '2': 'vios2'}

    :param adapter: pypowervm adapter
    :param host_uuid: UUID for the host system
    """
    vioses = pvm_vios.VIOS.wrap(adapter.read(
        root_type=pvm_ms.System.schema_type, root_id=host_uuid,
        child_type=pvm_vios.VIOS.schema_type,
        xag=[pvm_const.XAG.VIO_NET]))
    vios_map = dict()
    for vios in vioses:
        vios_map[str(vios.id)] = vios.name
    return vios_map


def get_vios_related_href_to_name_map(adapter, host_uuid):
    """ Map vios related_href to vios name.

    :param adapter: pypowervm adapter
    :param host_uuid: UUID for the host system
    """
    vioses = pvm_vios.VIOS.wrap(adapter.read(
        root_type=pvm_ms.System.schema_type, root_id=host_uuid,
        child_type=pvm_vios.VIOS.schema_type,
        xag=[pvm_const.XAG.VIO_NET]))
    vios_map = dict()
    for vios in vioses:
        vios_map[vios.related_href] = vios.name
    return vios_map


def create_novaclient(kconf):
    '''Create a novaclient based on the creds in CONF passed in.

    :param kconf: keystone_authtoken config
    :return: novaclient
    '''
    global _AUTH
    global _SESSION

    # Reuse the authentication module to not get a token every time
    if _AUTH is None:
        _AUTH = auth_v3.Password(auth_url=kconf.auth_uri,
                                 user_domain_name='Service',
                                 username=kconf.username,
                                 password=kconf.password,
                                 project_domain_name='Service',
                                 project_name=kconf.project_name)
    # Reuse the same session so we don't connect every time
    if _SESSION is None:
        _SESSION = session.Session(auth=_AUTH,
                                   verify='/etc/pki/tls/certs/powervc.crt')
    # Now we can return a new client using the same session and token
    return Client(version='2', auth=_AUTH, session=_SESSION)


@lockutils.synchronized('update_conf', 'powervc-sea-agent-')
def write_bridge_mappings(physnet, sea_name, vios_name):
    """Writes the mappings to disk, if they changed.  NOTE: this method was
    created so update_bridge_mapping_conf can make the novaclient call back
    to controller without holding the lock.

    :param physnet: Physical network name to write
    :param sea_name: Name of SEA for this physnet
    :param vios_name: Name of VIOS for this physnet
    :return: Raw mappings written to conf file.  NOTE: this will return None
             if something failed (like sea or vios not passed in) or if
             mappings weren't changed.
    """
    if not sea_name:
        LOG.warn(_('Bridge mapping not created, missing Shared Ethernet '
                   'Adapter'))
        return None

    if not vios_name:
        LOG.warn(_('Bridge mapping not created, missing Virtual IO Server'))
        return None

    # Calculate what the bridge mapping string should look like
    new_map = build_bridge_mapping(physnet, sea_name, vios_name)
    mappings = oslo_utils.get_openstack_config_value(
        constants.NEUTRON_CONF, constants.BRIDGE_MAPPING_CONF_ATTR,
        constants.BRIDGE_MAPPING_CONF_SECT)
    if mappings:
        # If not already in the mappings, just append
        if (physnet + ':') not in mappings:
            mappings = '%(orig_mappings)s,%(new_mapping)s' % {
                'orig_mappings': mappings, 'new_mapping': new_map}
        elif new_map not in mappings:
            # Already in the mappings but wrong sea/vios.  Update it.
            sea_vios = mappings.split(physnet + ':')[1].split(',')[0]
            mappings = mappings.replace(physnet + ':' + sea_vios, new_map)
        else:
            # The correct mapping was already there.  Set to None so invoker
            # knows nothing changed.
            mappings = None
    else:
        # No mappings, this is the first!
        mappings = new_map

    # If we updated the mapping, set it in conf file.
    if mappings:
        oslo_utils.set_openstack_config_value(
            constants.NEUTRON_CONF, constants.BRIDGE_MAPPING_CONF_ATTR,
            mappings, constants.BRIDGE_MAPPING_CONF_SECT)

    return mappings


def initialize_bridge_mappings(adapter, host_uuid):
    '''This method will toss current bridge mappings on disk and regenerate'''
    bridges = pvm_net.NetBridge.wrap(
        adapter.read(
            root_type=pvm_ms.System.schema_type, root_id=host_uuid,
            child_type=pvm_net.NetBridge.schema_type))
    if len(bridges) == 0:
        print _("No NetworkBridges found")
        return

    vios_map = get_vios_related_href_to_name_map(adapter, host_uuid)

    # Read vswitches so we can put the name in the physnet, easier to read.
    vswitches = pvm_net.VSwitch.wrap(adapter.read(
        root_type=pvm_ms.System.schema_type, root_id=host_uuid,
        child_type=pvm_net.VSwitch.schema_type))
    if len(vswitches) == 0:
        print _("No VirtualSwitches found")
        return
    vswitch_map = dict()
    for vswitch in vswitches:
        vswitch_map[vswitch.switch_id] = vswitch.name

    # Delete what's there so we start fresh
    oslo_utils.del_openstack_config_value(
        constants.NEUTRON_CONF, constants.BRIDGE_MAPPING_CONF_ATTR,
        constants.BRIDGE_MAPPING_CONF_SECT)

    # Build all the mappings.  The physical network format is vlanXX-vswitch
    mappings = list()
    for bridge in bridges:
        for sea in bridge.seas:
            if sea.is_primary:
                vios_name = vios_map.get(sea.vio_uri)
                vswitch = vswitch_map.get(bridge.vswitch_id)

                # Build a mapping for each routable vlan
                for vlan in bridge.list_vlans():
                    physnet = build_physnet_string(vswitch, vlan)
                    write_bridge_mappings(
                        physnet, sea.dev_name, vios_name)
