#
#
# Copyright 2014, 2015 IBM Corp.
# All Rights Reserved.
# Copyright 2012 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import threading
from logging import DEBUG

from powervc_oslo import _ as oslo_
from powervc_oslo import populate_languages
from powervc_oslo import populate_locale as oslo_populate_locale
from powervc_oslo import translate_message as oslo_translate_message

strong_store = threading.local()
_ = functools.partial(oslo_, domain='powervc_nova')
populate_locale = oslo_populate_locale
translate_message = oslo_translate_message
populate_languages('nova')


def _instance_filter(*args, **kwds):
    keys_to_check = ['name', 'power_state', 'vm_state', 'task_state', 'host']
    keys_to_copy = keys_to_check + ['default_ephemeral_device']

    def check_keys(arg, inst_keys):
        for key in inst_keys:
            if key not in arg:
                return False
        # Found them all!
        return True

    def check_inst(arg):
        type_ = type(arg)
        # check if this is a Nova Instance object
        from nova.objects.instance import Instance
        if type_ == Instance:
            return True
        # check for a dict with all the right keys to match instance
        if type_ == dict and check_keys(arg, keys_to_check):
            return True
        return False

    def partial_inst(arg):
        # Creates a partial instance from a real instance
        fake_inst = {}
        for key in keys_to_copy:
            try:
                fake_inst[key] = arg[key]
            except KeyError:
                # Don't create attributes that are not present
                pass

        fake_inst_wrapper = dict(PARTIAL_INSTANCE=fake_inst)
        return fake_inst_wrapper

    filtered_args = []
    for arg in args:
        filtered_args.append(arg if not check_inst(arg) else partial_inst(arg))
    return (filtered_args, kwds)


def _logcall(filter_=None, dump_parms=False):
    def func_parms(f):
        @functools.wraps(f)
        def wrapper(*args, **kwds):
            from oslo_log import log as logging
            LOG = logging.getLogger(f.__module__)
            # isEnabledFor not introduced until 2.7
            # logging_dbg = LOG.isEnabledFor(DEBUG)
            logging_dbg = LOG.logger.isEnabledFor(DEBUG)
            # logging_dbg = True
            if logging_dbg:
                if dump_parms:
                    d_args, d_kwds = ((args, kwds)
                                      if filter_ is None else filter_(*args,
                                                                      **kwds))
                    LOG.debug("Entering args:%s kwds:%s  '%s' %s" %
                              (d_args, d_kwds, f.__name__, f.__module__))
                else:
                    LOG.debug("Entering '%s' %s" % (f.__name__, f.__module__))

            r = f(*args, **kwds)
            if logging_dbg:
                if dump_parms:
                    LOG.debug("Exiting: return '%s'  '%s' %s" %
                              (r, f.__name__, f.__module__))
                else:
                    LOG.debug("Exiting: return '%s' %s" %
                              (f.__name__, f.__module__))
            return r
        return wrapper
    return func_parms


def cache_driver_data(list_of_data):
    """
    Decorator to save cache driver data for a method.
    This is useful when driver methods are called repeatedly directly or
    indirectly with a method. It's prerequisite that we should only use
    this decorator if we expected the same results from those methods
    being called multiple times. So, it's important that we exam the methods
    carefully before using this decorator.
    On the driver method side, it needs to use @return_cached_driver_data
    for this function to be effective.
    E.g. update_available_resource calls get_available_resource and
    get_host_stats multiple time directly and indirectly.

    :param    list_of_data: list of methods name that we want to cache
                            the data.
    """
    def func_parms(f):
        @functools.wraps(f)
        def wrapper(*args, **kwds):
            if list_of_data:
                for driver_data in list_of_data:
                    # initialize the strong_store with data we want to
                    # cache
                    setattr(strong_store, driver_data, None)
            try:
                return f(*args, **kwds)
            finally:
                # remove the parms from strong_store
                if list_of_data:
                    for driver_data in list_of_data:
                        if hasattr(strong_store, driver_data):
                            delattr(strong_store, driver_data)

        return wrapper
    return func_parms


def return_cached_driver_data():
    """
    Decorator to return cache driver data.
    This is used in conjunction with cache_driver_data. The caller will use
    the cache_driver_data to indicate what driver method's return data will
    be cached. If the driver method is called repeatedly, it will return
    cached data if it's available.

    """
    def func_parms(f):
        @functools.wraps(f)
        def wrapper(*args, **kwds):
            func_name = f.__name__
            if getattr(strong_store, func_name, None) is not None:
                return getattr(strong_store, func_name)

            data = f(*args, **kwds)
            if hasattr(strong_store, func_name):
                setattr(strong_store, func_name, data)
            return data

        return wrapper
    return func_parms


def cache_param_data(list_of_params):
    """
    :param list_of_params
        [
            {<name of param using cached data as string>:
                [<name of driver method to cache as string>,
                 <driver method arg 1>,
                 <driver method arg 2>,
                 ...
                ]
            },
            ...
        ]
    Decorator to call and cache return value for a driver method, cached
    data is then used as a parameter value to the decorated function. This is
    used in conjunction with cache_driver_data. cache_driver_data sets
    (on entry) and unsets (on exit) the name of a driver class method as a
    string in the cache. cache_param_data checks to see if the driver methods
    in list_of_params are set in the cache. If method is set, then the
    decorator invokes the method and caches the return value, using this value
    as the value of the parameter then passed in to the decorated function. If
    the value of the parameter was passed in to the call, then this value will
    be used and the caching is ignored.

    """
    def func_parms(f):
        @functools.wraps(f)
        def wrapper(self, *args, **kwds):
            for param_cache in list_of_params:
                param = param_cache.keys()[0]
                if param not in kwds:
                    method_data = param_cache[param]
                    if hasattr(strong_store, method_data[0]):
                        attr_val = getattr(strong_store, method_data[0], None)
                        if attr_val is not None:
                            data = attr_val
                        else:
                            driver_method = getattr(self, method_data[0])
                            data = driver_method(*method_data[1:])
                            setattr(strong_store, method_data[0], data)
                        kwds[param] = data

            return f(self, *args, **kwds)
        return wrapper
    return func_parms

# Convenience method to log all parameters
logcall_all_parms = _logcall()
# Convenience method to ensure the Instance parameter is logged
logcall_log_instance = _logcall()
# Filter out large Instance dumps.
logcall = _logcall(filter_=_instance_filter)
# Dump parameters and return values - DO NOT ENABLE IN PRODUCTION
# This may dump sensitive information to the log files.
# logcall = _logcall(filter_=_instance_filter, dump_parms=True)
