# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""An API Extension for Host Aggregates."""

import six

import webob.exc

from oslo_config import cfg

from nova.api.openstack import extensions
from nova.api.openstack import wsgi
from nova.compute import api as compute_api
from oslo_log import log as logging
from nova.ibm.ego.hapolicy import constants as ego_ha_consts
from nova.ibm.ego import ego_constants as ego_consts
from nova import exception
from nova import objects as nova_obj

from powervc_nova.common import aggregate_utils as agg_utils
from powervc_nova.prs import constants as prs_consts
from powervc_nova.prs import dro_schedules
from powervc_nova.db import api as db_api
from powervc_nova import _

ALIAS = 'os-aggregates'
AUTHORIZE_V21 = extensions.os_compute_authorizer(ALIAS)
AUTHORIZE_V2 = extensions.extension_authorizer('compute', 'aggregates')
LOG = logging.getLogger(__name__)


class HostAggregatesControllerExtension(wsgi.Controller):
    def __init__(self):
        self.api = compute_api.AggregateAPI()

    @wsgi.extends
    def index(self, req, resp_obj):
        """Check specific action permissions since OpenStack doesn't."""
        ctxt = req.environ['nova.context']
        self._validate_authorization(ctxt, req, action='index')
        dro_schedules = db_api.dro_schedules_read_all_schedules(ctxt)
        for agg in resp_obj.obj['aggregates']:
            if self._is_default_agg(agg):
                agg['powervc_default'] = True
            else:
                agg['powervc_default'] = False
            agg_dro_scheds = dro_schedules.get(agg.get('id')) or {}
            agg[prs_consts.DRO_SCHEDULES] = agg_dro_scheds

    @wsgi.extends
    def create(self, req, body):
        """Check specific action permissions since OpenStack doesn't."""
        ctxt = req.environ['nova.context']
        self._validate_authorization(ctxt, req, action='create')
        yield

    @wsgi.extends
    def show(self, req, resp_obj, id):
        """Check specific action permissions since OpenStack doesn't."""
        ctxt = req.environ['nova.context']
        self._validate_authorization(ctxt, req, action='index')
        if self._is_default_agg(resp_obj.obj['aggregate']):
            resp_obj.obj['aggregate']['powervc_default'] = True
        else:
            resp_obj.obj['aggregate']['powervc_default'] = False

        # Append the DRO Schedules for the aggregate
        schedules = db_api.dro_schedules_read_aggregate_schedules(ctxt, id)
        resp_obj.obj['aggregate'][prs_consts.DRO_SCHEDULES] = schedules

    @wsgi.extends
    def update(self, req, id, body):
        """Check specific action permissions since OpenStack doesn't."""
        ctxt = req.environ['nova.context']
        self._validate_authorization(ctxt, req, action='update')

        # Get incoming updates
        if len(body) != 1:
            raise webob.exc.HTTPBadRequest()
        try:
            updates = body['aggregate']
        except KeyError:
            raise webob.exc.HTTPBadRequest()
        if len(updates) < 1:
            raise webob.exc.HTTPBadRequest()

        try:
            agg = self.api.get_aggregate(ctxt, id)
            if self._is_default_agg(agg):
                # Make sure they aren't trying to change the default
                # aggregate's name.
                agg_name = (updates.get('name') or
                            cfg.CONF.default_host_aggregate_name)
                if agg_name != cfg.CONF.default_host_aggregate_name:
                    msg = _("The default host aggregate's name cannot be "
                            "updated.")
                    raise webob.exc.HTTPBadRequest(explanation=msg)
            # Update DRO schedules if needed
            if prs_consts.DRO_SCHEDULES in updates:
                # Need to pop the extra key or Nova's AggregateController
                # will throw an HTTP error because it's not a key that it is
                # expecting. Because of this the DRO schedules will not appear
                # in the PUT's response but it will show in future GETs
                schedules = updates.pop(prs_consts.DRO_SCHEDULES)
                if schedules:
                    try:
                        # Make sure the DRO schedules are valid
                        dro_schedules.load_dro_schedules(schedules)
                    except Exception as e:
                        raise webob.exc.HTTPBadRequest(
                            explanation=six.text_type(e))
                    values = {prs_consts.DRO_SCHEDULES: schedules}
                    db_api.dro_schedules_update_aggregate_schedules(ctxt, id,
                                                                    values)
                else:
                    db_api.dro_schedules_delete_aggregate_schedules(ctxt, id)
        except exception.AggregateNotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.format_message())
        yield

    @wsgi.extends
    def delete(self, req, id):
        """Prevent the default aggregate from being deleted."""
        ctxt = req.environ['nova.context']
        self._validate_authorization(ctxt, req, action='delete')
        try:
            agg = self.api.get_aggregate(ctxt, id)
            if self._is_default_agg(agg):
                msg = _("The default host aggregate cannot be deleted.")
                raise webob.exc.HTTPBadRequest(explanation=msg)
            db_api.dro_schedules_delete_aggregate_schedules(ctxt, id)
        except exception.AggregateNotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.format_message())

        yield

    def _validate_add_host(self, context, agg_id, host):
        """Prevent the host from belonging to multiple host aggregates."""
        if cfg.CONF.allow_host_in_multiple_host_aggregates or host is None:
            # Just let the call proceed since we don't have to check.
            return
        if len(agg_utils.get_aggregate_ids_for_host(context, host)) >= 1:
            msg = _("The host cannot be added to more than one aggregate.")
            raise webob.exc.HTTPBadRequest(explanation=msg)

    @wsgi.extends(action='add_host')
    def _add_host(self, req, id, body):
        ctxt = req.environ['nova.context']
        self._validate_authorization(ctxt, req, action='add_host')
        # Prevent a single host from being in multiple host aggregates.
        self._validate_add_host(ctxt, id, body['add_host'].get('host'))
        yield

    @wsgi.extends(action='remove_host')
    def _remove_host(self, req, id, body):
        ctxt = req.environ['nova.context']
        self._validate_authorization(ctxt, req, action='remove_host')
        yield

    @wsgi.extends(action='set_metadata')
    def _set_metadata(self, req, id, body):
        # Because PRS uses aggregate metadata to control various
        # behaviors, we need to traverse the keys to determine
        # which authorizations to check.
        ctxt = req.environ['nova.context']
        metadata = body['set_metadata'].get('metadata', {})
        for key in metadata.keys():
            if key.startswith(ego_consts.INITIAL_POLICY):
                self._validate_authorization(
                    ctxt, req, 'initial_placement')
            elif key.startswith(ego_consts.RUNTIME_POLICY):
                self._validate_authorization(
                    ctxt, req, 'runtime_optimization')
            elif key.startswith(ego_ha_consts.KEY_HAPOLICY):
                self._validate_authorization(
                    ctxt, req, 'high_availability')
            else:
                self._validate_authorization(ctxt, req, action='set_metadata')
        yield

    @wsgi.extends
    def action(self, req, id, body):
        # TODO (cjvolzka) Remove this method once compute v2 is deprecated
        # from PowerVC. As of 1.3.1 the GUI and CLI use compute v2.1 which
        # calls the discrete methods above and does not enter this code.
        # Old curl scripts may still use the v2 API which would call this.
        actions = ('add_host', 'remove_host', 'set_metadata')
        ctxt = req.environ['nova.context']
        for action, data in six.iteritems(body):
            if action in actions:
                if action == 'set_metadata':
                    # Because PRS uses aggregate metadata to control various
                    # behaviors, we need to traverse the keys to determine
                    # which authorizations to check.
                    for key in data.get('metadata', {}).keys():
                        if key.startswith(ego_consts.INITIAL_POLICY):
                            self._validate_authorization(
                                ctxt, req, 'initial_placement')
                        elif key.startswith(ego_consts.RUNTIME_POLICY):
                            self._validate_authorization(
                                ctxt, req, 'runtime_optimization')
                        elif key.startswith(ego_ha_consts.KEY_HAPOLICY):
                            self._validate_authorization(
                                ctxt, req, 'high_availability')
                        else:
                            self._validate_authorization(ctxt, req,
                                                         action=action)
                else:
                    self._validate_authorization(ctxt, req, action=action)
                    if action == 'add_host':
                        # Prevent a single host from being in multiple
                        # host aggregates.
                        self._validate_add_host(ctxt, id, data.get('host'))
        yield

    @staticmethod
    def _validate_authorization(context, req, action):
        """Checks the user's access to invoke the action."""
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action=action)
        else:
            AUTHORIZE_V2(context, action=action)

    @staticmethod
    def _is_default_agg(agg):
        """Checks whether the aggregate is the default aggregate."""
        agg_name = (agg.name if isinstance(agg, nova_obj.Aggregate)
                    else agg['name'])
        return agg_name == cfg.CONF.default_host_aggregate_name


class Host_aggregates(extensions.ExtensionDescriptor):
    """Extend the Host Aggregates API."""

    name = "IBM PowerVC Host Aggregates Extension"
    alias = "ibm-powervc-os-aggregates"
    namespace = "http://docs.openstack.org/compute/ext/" + \
                "ibm_aggregates/api/v1.1"
    updated = "2014-09-17T21:00:00-06:00"

    def get_controller_extensions(self):
        controller = HostAggregatesControllerExtension()
        extension = extensions.ControllerExtension(self, 'os-aggregates',
                                                   controller)
        return [extension]


class Host_aggregatesV21(extensions.V21APIExtensionBase):
    """Extend the Host Aggregates API."""

    name = "IBM PowerVC Host Aggregates Extension"
    alias = "ibm-powervc-os-aggregates"
    version = 1

    def get_controller_extensions(self):
        controller = HostAggregatesControllerExtension()
        extension = extensions.ControllerExtension(self, 'os-aggregates',
                                                   controller)
        return [extension]

    def get_resources(self):
        return []
