#
#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
'''
@author: dnip

An API Extension for being able to Report Additional Hypervisor State of a
OS-Hypervisor.
'''

import os.path
from os import listdir
from nova import compute
from nova import objects
from nova.api.openstack import extensions
from nova.api.openstack import wsgi
from oslo_log import log as logging
from powervc_nova import _

LOG = logging.getLogger(__name__)
ALIAS = 'hypervisor-state'
AUTHORIZE_V21 = extensions.os_compute_authorizer(ALIAS)
AUTHORIZE_V2 = extensions.extension_authorizer('compute', 'hypervisor_state')


class HypervisorStateControllerExtension(wsgi.Controller):

    def __init__(self):
        self.host_api = compute.HostAPI()
        self.latest_rmt_version = None

    @wsgi.extends
    def show(self, req, resp_obj, id):
        LOG.debug("Enter: show")
        # Retrieve the Hypervisor Instance specified from the DB and populate
        # the State
        context = req.environ['nova.context']
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context)
        else:
            AUTHORIZE_V2(context)
        dbHypInst = None
        try:
            dbHypInst = objects.ComputeNode.get_by_id(context, int(id))
        except Exception as e:
            LOG.error(_("Unable to retrieve the hypervisor data. The "
                        "hypervisor identifier is '%(id)s'") %
                      {'id': id})
            LOG.exception(e)
        respHypInst = resp_obj.obj['hypervisor']
        if(dbHypInst):
            self._populateStatusAttributes(respHypInst, dbHypInst)
            self._populateVersionAttributes(respHypInst, dbHypInst)
        else:
            respHypInst['hypervisor_state'] = 'UNKNOWN'

    @wsgi.extends
    def detail(self, req, resp_obj):
        LOG.debug("Enter: detail")
        # Loop through all the Hypervisors that the main Hypervisor API is
        # returning
        context = req.environ['nova.context']
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context)
        else:
            AUTHORIZE_V2(context)
        dbHypInst = None
        for hypervisor in list(resp_obj.obj['hypervisors']):
            # Retrieve the Hypervisor Instance specified from the DB and
            # populate the State
            try:
                dbHypInst = objects.ComputeNode.get_by_id(context,
                                                          hypervisor['id'])
            except Exception as e:
                LOG.error(_("Unable to retrieve the hypervisor data. The "
                            "hypervisor identifier is '%(id)s'") %
                          {'id': hypervisor['id']})
                LOG.exception(e)
            if(dbHypInst):
                self._populateStatusAttributes(hypervisor, dbHypInst)
                self._populateVersionAttributes(hypervisor, dbHypInst)
            else:
                hypervisor['hypervisor_state'] = 'UNKNOWN'

    def _populateStatusAttributes(self, respHypInst, dbHypInst):
        LOG.debug("Enter: _populateStatusAttributes")
        # Set the Hypervisor State on the Response to the value retrieved from
        # the Database
        hypState = self._getHypervisorStatsData(dbHypInst, 'hypervisor_state')
        if(hypState):
            respHypInst['hypervisor_state'] = hypState
        else:
            respHypInst['hypervisor_state'] = 'UNKNOWN'
            LOG.warning(_("Unable to get hypervisor state from database."))

    def _populateVersionAttributes(self, rspHypInst, dbHypInst):
        """Helper method to populate the Remote Host Version attributes"""
        mgmt_version = None
        # If there are any stats available, parse the Management Version
        if dbHypInst.stats:
            statsEntries = dbHypInst.stats
            managed_by = statsEntries.get('managed_by')
            mgmt_version = statsEntries.get('mgmt_version')
        # Add the latest PowerVC version available for a Remote Host Host
        if mgmt_version is not None and managed_by != 'HMC':
            rspHypInst['latest_mgmt_version'] = self._get_latest_mgmt_version()
            rspHypInst['current_mgmt_version'] = mgmt_version

    def _getHypervisorStatsData(self, dbHypInst, stats_key):
        """
        get the 'stats' data value from the hypervisor DB instance based on the
        specified stats_key
        """
        hypStatsData = None
        statsEntries = dbHypInst.stats
        if(statsEntries is None):
            LOG.warning(_("Unable to get hypervisor information "
                          "from database."))
            return hypStatsData

        hypStatsData = statsEntries.get(stats_key)
        if(hypStatsData):
            LOG.debug("HypervisorStats key: %s = %s" %
                      (str(stats_key), str(hypStatsData)))
        else:
            LOG.debug("_getHypervisorStatsData unable to get "
                      "HypervisorStats for key: %s" % str(stats_key))
            for key, val in statsEntries.iteritems():
                entry = dict(key=key, value=val)
                LOG.debug("_getHypervisorStatsData stats entry= %s" %
                          str(entry))

        return hypStatsData

    def _get_latest_mgmt_version(self):
        """Helper method to get latest available PowerVC version for Host"""
        # Only retrieve the version once for the process and cache it
        if self.latest_rmt_version is None:
            latest_version = None
            for rmttype in ['kvm', 'powervm']:
                # Only one of the 2 will exists depending on the type
                if not os.path.exists('/opt/ibm/powervc/images/' + rmttype):
                    continue
                # Loop through the possible images, determining the latest one
                for filename in listdir('/opt/ibm/powervc/images/' + rmttype):
                    if ((filename.startswith('powervc-powerkvm') or
                         filename.startswith('powervc-powervm'))):
                        if filename.endswith('.tgz') is False:
                            continue
                        # Parse the version out of the name of the file
                        version = filename[
                            filename.rfind('-') + 1:len(filename) - 4]
                        # If this version is greater, use this one instead
                        if self._compare_version(version, latest_version) > 0:
                            latest_version = version
            self.latest_rmt_version = latest_version
        return self.latest_rmt_version

    @staticmethod
    def _compare_version(new_version, old_version):
        """Utility to compare 2 versions and see if the first is newer"""
        # If the old version isn't specified, the new version is greater if set
        if not old_version:
            return 1 if new_version else 0
        # If no new version is provided, the old is newer since it must be set
        if not new_version:
            return -1
        # We need to break apart the version to single digits for comparison
        old_split = old_version.split('.')
        new_split = new_version.split('.')
        max_digits = max(len(new_split), len(old_split))
        # Loop through each of the digits and do the comparison
        for index in range(max_digits):
            old_num = int(old_split[index]) if len(old_split) > index else 0
            new_num = int(new_split[index]) if len(new_split) > index else 0
            # If a digit is different, then return since 1 since greater
            if old_num != new_num:
                return 1 if new_num > old_num else -1
        # If all digits were the same, the 2 versions must be equal
        return 0


class Hypervisor_state(extensions.ExtensionDescriptor):
    """Extended State/Status Information about the Hypervisor."""
    name = "PowerVC Hypervisor State"
    alias = "powervc-hypervisors-state"
    namespace = "http://www.ibm.com/openstack/compute/contrib/powervc/v1.0"
    updated = "2013-02-12T21:00:00-06:00"

    def get_controller_extensions(self):
        """Provide a Controller Extension to the os-hypervisors Resource"""
        extension_list = []
        extension_list.append(
            extensions.ControllerExtension(
                self,
                'os-hypervisors',
                HypervisorStateControllerExtension()
            )
        )
        return extension_list


class Hypervisor_stateV21(extensions.V21APIExtensionBase):
    """Extended State/Status Information about the Hypervisor."""
    name = "PowerVC Hypervisor State"
    alias = "powervc-hypervisors-state"
    version = 1

    def get_controller_extensions(self):
        """Provide a Controller Extension to the os-hypervisors Resource"""
        extension_list = []
        extension_list.append(
            extensions.ControllerExtension(
                self,
                'os-hypervisors',
                HypervisorStateControllerExtension()
            )
        )
        return extension_list

    def get_resources(self):
        return []
