#
#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import httplib
import six
from webob import exc
from webob import Response

from nova.api.openstack import extensions
from nova.api.openstack import wsgi
from oslo_serialization import jsonutils
from oslo_log import log as logging
from powervc_nova import _
from nova.image import glance
from nova import exception
from powervc_nova.storage import storage_utils as su
from powervc_nova.storage.storage_mapping_api import StorageMapping
from storage_connectivity_groups import DISPLAY_NAME
from storage_connectivity_groups import ALIAS

EXTENSION_ALIAS = "image-stg-conn-groups"
AUTHORIZE_V21 = extensions.os_compute_authorizer(EXTENSION_ALIAS)
AUTHORIZE_V2 = extensions.extension_authorizer('compute',
                                               'image_stg_conn_groups')
LOG = logging.getLogger(__name__)


class StorageConnGroupsByImageController(wsgi.Controller):

    """
    This Controller handles the API entry implementation of the REST
    function to retrieve the applicable SCGs given an Image.
    PowerVC requires this to populate the list of SCG choices on the deploy
    panel for a particular image chosen for deploy.
    The returned list will be in priority order for proper defaulting
    and it will contain only "enabled" SCGs for deploy.
    """

    def __init__(self):
        self.image_service = glance.get_default_image_service()

    def _get_image(self, context, image_id):
        try:
            return self.image_service.show(context, image_id)
        except exception.NotFound:
            msg = _("Image not found for id '%s'.") % image_id
            raise exc.HTTPNotFound(explanation=six.text_type(msg))

    def index(self, req, image_id):
        """
        REST API entry point to return a dictionary containing a list of
        storage connectivity groups (SCGs) that are valid choices to use
        with the deployment of the passed image. If there are problems
        getting the volume information or attributes, the returned list will
        be empty and the log will contain the appropriate warning as this
        means there are no SCGs with effective access.
        Storage connectivity will be checked for the candidate SCGs, such
        that if a SCG does not contain any host with VIOSes that have
        NPIV-capable and active ports, then the SCG will not be returned
        (as an example).

        Options:
            host_name=<host_name>        Constrain the SCGs returned to be
                                         those that contain the specified
                                         host as a member.
            details=<true|false>         'true' for detailed SCG listing,
                                         'false' or default will return the
                                         short listing shown below.

        :param req: The REST request. No special options are supported on this
        :param image_id: The image uuid to base the request on.
        :returns: {
                    'storage_connnectivity_groups': [
                        {
                            'display_name': <scg-name>,
                            'id': <scg-uuid>
                        }
                    ]
                   }

        Flow:
            1. Look up the image from glance.
            2. Find source volume id(s) in the image info. Loop over each.
            3. Use cinder_client to return the volume info
            4. Get cinder host value from the volume info
            5. Lookup storage provider info for the cinder host
            6. Request the SCGs that are applicable to the cinder host.
               If the cinder host is an SSP cluster driver, then specify a
               sort_type of 'cluster'. Otherwise if an 'fc' type of cinder
               host, then specify a sort_type of 'external'.
            7. Return the filtered list in the assumed priority order.
        """
        context = req.environ["nova.context"]
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action='list')
        else:
            AUTHORIZE_V2(context, action='list')
        LOG.debug("**ENTRY**   image_id=%s" % str(image_id))
        su.http_error_if_ivm()  # HMC-managed installs only

        # **lookup image**
        # Note that we go through the image_service rather than the
        # database call [db.s3_image_get(context, image_id)] since the nova
        # db may not have record for this image.
        image = self._get_image(context, image_id)
        LOG.debug("retrieved image: %s" % str(image))

        # Get optional host name if passed, which will further filter
        # the list of applicable SCGs to those that have the host.
        if 'host_name' in req.GET:
            host_name = req.GET['host_name']
        else:
            host_name = None
        details = False
        if 'details' in req.GET:
            details = jsonutils.loads(req.GET['details'].lower())

        # Call an internal storage mapping API to return the list of SCGs
        # applicable to this image
        msg_dict = {'messages': []}
        try:
            volume_refs = su.get_volumes_from_image(image,
                                                    msg_dict=msg_dict)
            scg_list = StorageMapping().get_scgs_for_volumes(context,
                                                             volume_refs,
                                                             host=host_name,
                                                             msg_dict=msg_dict,
                                                             image=image)
        except Exception as ex:
            # It was decided that a 500 response made sense for any
            # exception case here. We use the HTTP internal server error
            # as that lets us pass the exception message on the response.
            # But in general we need to be careful in passing any exception
            # message on as they may not be appropriate for REST consumers.
            LOG.exception(ex)
            raise exc.HTTPInternalServerError(explanation=six.text_type(ex))

        if details:
            # Build the detailed listing of SCG props
            json_list = [scg.to_dict(context=context) for scg in scg_list]
        else:
            # Just build the short (index) list (ignore other properties)
            json_list = map(lambda x: {'id': x.id,
                                       DISPLAY_NAME: x.display_name},
                            scg_list)
        LOG.debug(str(json_list))
        # Return results in JSON body response
        scgs = {"storage_connectivity_groups": json_list}
        if msg_dict['messages']:
            # jsonutils does not do well with unicode strings which have now
            # converted to Message types in openstack, so convert them back
            msgs = [six.text_type(x) for x in msg_dict['messages']]
            scgs['messages'] = msgs

        return Response(request=req,
                        status=httplib.OK,
                        content_type='application/json',
                        body=jsonutils.dumps(scgs))


class Image_scgs(extensions.ExtensionDescriptor):

    """ Support to retrieve Storage Connectivity Groups by Image. """
    name = "Compute Image Storage Connectivity Groups"
    alias = "images-" + ALIAS
    namespace = ("http://docs.openstack.org/compute/ext/images/" +
                 "image_stg_conn_groups" + "/api/v1.1")
    updated = "2013-08-09T11:00:00-06:00"

    def get_resources(self):
        """
        Provide a Resource Extension for getting candidate SCGs
        provided an image.
        """
        parent_def = {'member_name': 'image', 'collection_name': 'images'}
        return [
            extensions.ResourceExtension(
                ALIAS,
                controller=StorageConnGroupsByImageController(),
                parent=parent_def)]


class Image_scgsV21(extensions.V21APIExtensionBase):

    """ Support to retrieve Storage Connectivity Groups by Image. """
    name = "Compute Image Storage Connectivity Groups"
    alias = "images-" + ALIAS
    version = 1

    def get_resources(self):
        """
        Provide a Resource Extension for getting candidate SCGs
        provided an image.
        """
        parent_def = {'member_name': 'image', 'collection_name': 'images'}
        return [
            extensions.ResourceExtension(
                ALIAS,
                controller=StorageConnGroupsByImageController(),
                parent=parent_def)]

    def get_controller_extensions(self):
        return []
