#
#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import httplib
import webob
import six
from webob import Response
from nova import exception
from oslo_serialization import jsonutils
from nova.api.openstack import extensions
from nova.api.openstack import wsgi
from powervc_nova.storage.storage_mapping_api import StorageMapping

from oslo_log import log as logging
from powervc_nova import _
from powervc_nova.storage import storage_utils as su
from nova.image import glance

# ------------------------------------
# Extension: images
# ------------------------------------
LOG = logging.getLogger(__name__)
EXTENSION = 'images'
ALIAS = "scg-to-images"
AUTHORIZE_V21 = extensions.os_compute_authorizer(ALIAS)
AUTHORIZE_V2 = extensions.extension_authorizer('compute', "scg_to_images")


class SCGImageListController(wsgi.Controller):

    """
    This Controller handles the API entry implementation of the REST
    function to retrieve the applicable images tied to storage connectivity
    group. Every image related to a given list of SCGs by the storage providers
    that can access the image's storage volume. This list can be looked at to
    see if a specific SCG is related to that image and the image returned
    in a list if the given SCG is related to it.
    """

    def __init__(self):
        # Controller fields for API services
        self.image_service = glance.get_default_image_service()
        self.stg_api = StorageMapping()

    def _get_image_list(self, context):
        return self.image_service.detail(context)

    def _get_image_scgs(self, context, image, msg_dict):
        """ Helper method to call an internal storage mapping API to return
            the list of SCGs applicable to this image. Errors are logged
            and ignored.
        """
        scg_list = []
        try:
            volume_refs = su.get_volumes_from_image(image,
                                                    msg_dict=msg_dict)
            scg_list = self.stg_api.get_scgs_for_volumes(context, volume_refs,
                                                         image=image,
                                                         msg_dict=msg_dict)
        except exception.VolumeNotFound:
            msg = _("The image '%s' will be skipped.") % image['name']
            msg_dict['messages'].append(msg)
        except Exception as ex:
            msg = _("There was a problem processing image '%(name)s' with "
                    "ID '%(image_id)s', so it will be skipped. See the nova "
                    "api.log for details.") % dict(name=image['name'],
                                                   image_id=image['id'])
            msg_dict['messages'].append(msg)
            LOG.warn(msg)
            LOG.exception(ex)

        return scg_list

    def index(self, req, scg_id):
        """
        REST API entry point to return a dictionary containing a list of
        images that are associated with storage connectivity group (SCG).
        If there are no images or none of the images found reference an SCG
        or if none of the images found reference the SCG passed in by ID
        then will return empty list and the log will contain the appropriate
        warning as this means there are no images that reference the SCG.

        :param req: The REST request. No special options are supported on this
        :param scg_id: The scg uuid to base the request on.
        :returns: {
                   "images": [
                       {
                        "id": <image-uuid>
                        "name": <image-name>,
                       }
                      ]
                     }

        Flow:
            1. Look up all images from glance. Loop over each image.
            2. Find list of SCGs each image references. Loop over that list.
            3. If any image references the SCG by ID that is the SCG looking
               for then add that image to list of referencing images.
            4. Return the filtered list of images that referenced the SCG.

        """
        context = req.environ["nova.context"]
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action='list')
        else:
            AUTHORIZE_V2(context, action='list')
        LOG.debug("index() **ENTRY**   scg_id=%s" % str(scg_id))

        # validate to make sure SCG id is good
        su.http_error_if_ivm()  # HMC-managed installs only
        # Get the storage connectivity group and make sure SCG is real
        try:
            scg = su.get_scg_by_id(context, scg_id)
        except Exception as ex:
            LOG.exception(ex)
            raise webob.exc.HTTPNotFound(explanation=six.text_type(ex))
        LOG.debug("SCG found for scg_id is: %s" % scg.to_dict())

        try:
            LOG.debug("Get a list of all the known images")
            images = self._get_image_list(context)
        except Exception as ex:
            LOG.exception(ex)
            raise webob.exc.HTTPNotFound(explanation=six.text_type(ex))

        msg_dict = {'messages': []}
        image_ref_scg = []
        # if there is at least one image...
        if images:
            LOG.debug(">>>> %s known images found." % len(images))

            # check scg first to see if has any VIOS. If not then
            # cannot be sure is related to any existing images
            # (or not related) so api will fail here with message that
            # scg not related to any VIOS
            vios_list = scg.vios_ids
            if vios_list is None or len(vios_list) == 0:
                ex = (_("No Virtual I/O Servers are members of"
                        " storage connectivity group '%s'."
                        " Unable to determine if it is related to any existing"
                        " images.") % scg.display_name)
                LOG.exception(ex)
                raise webob.exc.HTTPServerError(explanation=six.text_type(ex))
            else:
                LOG.debug("SCG %s has vios nodes." % scg_id)

            # for every image get scgs referenced and see if scg in list
            for image in images:
                if scg_id in map(lambda x: x.id,
                                 self._get_image_scgs(context,
                                                      image, msg_dict)):
                    image_ref_scg.append(image)

        if image_ref_scg:
            # There are images in list so they ref'd SCG
            LOG.debug("%s referencing images found for scg id = %s"
                      % (len(image_ref_scg), scg_id))
        else:
            # Since no images at all then empty return
            LOG.info(_("No images were found referencing volumes that "
                       "storage connectivity group '%s' provides connectivity "
                       "for.") % scg.display_name)

        # Return results as list of images in JSON body response
        # even if empty list
        index_list = map(lambda x: {'id': x['id'],
                                    'name': x['name']},
                         image_ref_scg)
        LOG.debug(str(index_list))
        # Return results in JSON body response
        image_list = {"images": index_list}

        # Add info messages if any to output using the messages key.
        if msg_dict['messages']:
            match = "connectivity group '"
            # Do to the impl of this API, if the info message is referring
            # to a SCG that was not passed in, we remove it here.
            msgs = [six.text_type(x)
                    for x in msg_dict['messages'] if match not in x or
                    scg.display_name in x]
            LOG.debug(msgs)
            image_list['messages'] = msgs

        return Response(request=req,
                        status=httplib.OK,
                        content_type='application/json',
                        body=jsonutils.dumps(image_list))


class Scg_to_images(extensions.ExtensionDescriptor):

    """ Support to list images associated with Storage Connectivity Group. """
    name = "Get SCG Image List"
    alias = "scg-" + EXTENSION
    # The namespace just follows the convention of other APIs - is there
    # separate work that needs to be done to register this?
    namespace = ("http://docs.openstack.org/compute/ext/storage/"
                 "scg_images/api/v1.1")
    updated = "2014-02-14T11:00:00-06:00"

    def get_resources(self):
        '''
        Provide a Resource Extension for listing all images
        associated with a storage connectivity group.
        '''
        parent_def = {'member_name': 'scg',
                      'collection_name': 'storage-connectivity-groups'}
        return [
            extensions.ResourceExtension(
                EXTENSION,
                controller=SCGImageListController(),
                parent=parent_def)]


class Scg_to_imagesV21(extensions.V21APIExtensionBase):

    """ Support to list images associated with Storage Connectivity Group. """
    name = "Get SCG Image List"
    alias = ALIAS
    version = 1

    def get_resources(self):
        '''
        Provide a Resource Extension for listing all images
        associated with a storage connectivity group.
        '''
        parent_def = {'member_name': 'scg',
                      'collection_name': 'storage-connectivity-groups'}
        return [
            extensions.ResourceExtension(
                EXTENSION,
                controller=SCGImageListController(),
                parent=parent_def)]

    def get_controller_extensions(self):
        return []
