#
#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import httplib
import webob
import six
from webob import Response

from oslo_serialization import jsonutils
from nova.api.openstack import extensions
from nova.api.openstack import wsgi
from powervc_nova.storage.storage_mapping_api import StorageMapping

from oslo_log import log as logging
from powervc_nova import _
from powervc_nova.storage import storage_utils as su

# ------------------------------------
# Extension: images
# ------------------------------------
LOG = logging.getLogger(__name__)
EXTENSION = 'volume-types'
ALIAS = "scg-to-volume-types"
AUTHORIZE_V21 = extensions.os_compute_authorizer(ALIAS)
AUTHORIZE_V2 = extensions.extension_authorizer('compute',
                                               "scg_to_volume-types")


class SCGVolumeTypesListController(wsgi.Controller):

    """
    This Controller handles the API entry implementation of the REST
    function to retrieve the applicable volume-types for an given SCG.
    The StorageMapping get_scg_template_references is used to get volume-types
    that reference the SCG id passed in.
    """

    def __init__(self):
        # class statics for the internal API module
        self.stg_api = StorageMapping()

    def index(self, req, scg_id):
        """
        REST API entry point to return a dictionary containing a list of
        volume-types that are associated with storage connectivity group (SCG).
        If there are no volume-types found that reference the SCG
        then will return empty list and the log will contain the appropriate
        messages as this means there are no volume-types related to the SCG.

        :param req: The REST request. No special options are supported on this
        :param scg_id: The scg uuid to base the request on.
        :returns: {
                   "volume-types": [
                       {
                        "name": <volume-type name>
                        "id": <template-uuid>,
                       }
                      ]
                     }

        Flow:
            1. Validate input SCG id.
            2. Use StorageMapping to get list of applicable volume-types.
            3. Return the list of volume-types that are implied by the SCG.

        """
        context = req.environ["nova.context"]
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action='list')
        else:
            AUTHORIZE_V2(context, action='list')
        LOG.debug("index() **ENTRY**   scg_id=%s" % str(scg_id))

        # valid to make sure SCG is good
        su.http_error_if_ivm()  # HMC-managed installs only
        # Get the storage connectivity group and make sure SCG with id is real
        try:
            scg = su.get_scg_by_id(context, scg_id)
        except Exception as ex:
            LOG.exception(ex)
            raise webob.exc.HTTPNotFound(explanation=six.text_type(ex))
        LOG.debug("SCG found for id '%s' is: %s" % (scg_id, scg.to_dict()))

        msg_dict = {'messages': []}
        try:
            # get volume-types associated with this scg
            volume_types = self.stg_api.get_scg_volume_types(context, scg,
                                                             msg_dict=msg_dict)
        except Exception as ex:
            msg = _("There was an error trying to obtain volume types "
                    "(storage templates): %s") % ex
            LOG.exception(ex)
            raise webob.exc.HTTPServerError(explanation=six.text_type(msg))
        # if any template was found related the scg that was passed in
        # Return results as list of volume-types in JSON body response
        volume_types_list = []
        if volume_types:
            # there are volume-types in list - format them
            volume_types_list = map(
                lambda x: {'id': x['id'], 'name': x['name'],
                           'provider': x['extra_specs']
                           ['capabilities:volume_backend_name']}, volume_types)
            LOG.debug(str(volume_types_list))
        else:
            # No volume-types referenced the SCG so empty return
            LOG.debug("NO referencing volume-types found for scg '%s'."
                      % scg.display_name)

        # Return results in JSON body response
        output = {"volume-types": volume_types_list}

        if msg_dict['messages']:
            # jsonutils does not do well with unicode strings which have now
            # converted to Message types in openstack, so convert them back
            msgs = [six.text_type(x) for x in msg_dict['messages']]
            output['messages'] = msgs

        return Response(request=req,
                        status=httplib.OK,
                        content_type='application/json',
                        body=jsonutils.dumps(output))


class Scg_to_volume_types(extensions.ExtensionDescriptor):

    """
    Support to list volume-types is associated with Storage Connectivity Group.
    """
    name = "Get SCG Volume-types List"
    alias = "scg-" + EXTENSION
    # The namespace just follows the convention of other APIs - is there
    # separate work that needs to be done to register this?
    namespace = ("http://docs.openstack.org/compute/ext/storage/"
                 "scg_volume-types/api/v1.1")
    updated = "2013-09-25T11:00:00-06:00"

    def get_resources(self):
        '''
        Provide a Resource Extension for listing all volume-types
        associated with a storage connectivity group.
        '''
        parent_def = {'member_name': 'scg',
                      'collection_name': 'storage-connectivity-groups'}
        return [
            extensions.ResourceExtension(
                EXTENSION,
                controller=SCGVolumeTypesListController(),
                parent=parent_def)]


class Scg_to_volume_typesV21(extensions.V21APIExtensionBase):

    """
    Support to list volume-types is associated with Storage Connectivity Group.
    """
    name = "Get SCG Volume-types List"
    alias = ALIAS
    version = 1

    def get_resources(self):
        '''
        Provide a Resource Extension for listing all volume-types
        associated with a storage connectivity group.
        '''
        parent_def = {'member_name': 'scg',
                      'collection_name': 'storage-connectivity-groups'}
        return [
            extensions.ResourceExtension(
                EXTENSION,
                controller=SCGVolumeTypesListController(),
                parent=parent_def)]

    def get_controller_extensions(self):
        return []
