# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""An API Extension for Server Groups."""

import six
import webob.exc

from nova import db
from nova import exception
from nova import objects
from nova import os
from nova import quota
from nova import utils
from nova.api.openstack import common
from nova.api.openstack import extensions
from nova.api.openstack import wsgi
from nova.api.openstack.compute.views import servers as views_servers
from nova.compute import utils as compute_utils
from oslo_log import log as logging

from oslo_config import cfg

from powervc_nova import _
from powervc_nova.common import server_group_utils as sg_utils
from powervc_nova.objects.compute import dom as compute_dom

CONF = cfg.CONF

AUTHORIZE_V21 = extensions.os_compute_authorizer("os-server-groups")
MEMBER_AUTHORIZE_V21 = extensions.os_compute_authorizer("server-group-members")
AUTHORIZE_V2 = extensions.extension_authorizer('compute', 'server_groups')
MEMBER_AUTHORIZE_V2 = extensions.extension_authorizer('compute',
                                                      'server_group_members')
LOCK_PATH = os.path.join(CONF.instances_path, 'locks')
QUOTAS = quota.QUOTAS
LOG = logging.getLogger(__name__)


class ServerGroupMemberViewBuilder(common.ViewBuilder):
    """Model a server group member response as a Python dictionary."""

    def __init__(self, *args, **kwargs):
        """Initialize view builder."""
        super(ServerGroupMemberViewBuilder, self).__init__(*args, **kwargs)
        self._server_view_builder = views_servers.ViewBuilder()
        self._collection_name_list = ['os-server-groups', 'members']
        self._sep = '/'

    def index(self, request, ServerGroup_id, instances):
        response = []
        for instance in instances:
            identifier_list = [ServerGroup_id, instance['uuid']]
            # Append the basic view of a server. i.e. id, name and links
            instance_response = self._server_view_builder.basic(
                request, instance)
            # Append server group member links for each instance. _get_links
            # calls _get_href_link and _get_bookmark_link
            instance_response['links'] = self._get_links(
                request, identifier_list, self._collection_name_list)
            response.append(instance_response)
        return {'members': response}

    def _get_href_link(self, request, identifier_list, collection_name_list):
        """Return an href string pointing to this object. Overrides parent's
           method to allow for more than one identifier in the link. Server
           group member links include a server group and a member."""
        prefix = self._update_compute_link_prefix(request.application_url)
        link = self._sep.join([prefix, self._get_project_id(request)])
        # Append each collection name and identifier pair to the link
        for i in range(len(identifier_list)):
            link += (self._sep + collection_name_list[i] +
                     self._sep + str(identifier_list[i]))
        return link

    def _get_bookmark_link(self, request, identifier_list,
                           collection_name_list):
        """Create a URL that refers to a specific resource. Overrides parent's
           method to allow for more than one identifier in the link. Server
           group member links include a server group and a member."""
        base_url = common.remove_trailing_version_from_href(
            request.application_url)
        base_url = self._update_compute_link_prefix(base_url)
        link = self._sep.join([base_url, self._get_project_id(request)])
        # Append each collection name and identifier pair to the link
        for i in range(len(identifier_list)):
            link += (self._sep + collection_name_list[i] +
                     self._sep + str(identifier_list[i]))
        return link


class ServerGroupControllerExtension(wsgi.Controller):

    @staticmethod
    def _validate_authorization(context, req, action):
        """Checks the user's access to invoke the action."""
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action=action)
        else:
            AUTHORIZE_V2(context, action=action)

    def _convert_hosts_to_display_names(self, context, host_names):
        """Takes a list of host names and converts them to a list of
           host display names."""
        # Get a list of PowerVC ManagedHost objects
        hostfact = compute_dom.ManagedHostFactory.get_factory()
        hosts = hostfact.find_all_hosts(context)
        # Create a mapping of host names to host display names
        display_name_mapping = {}
        for host in hosts:
            display_name_mapping[host.host_name] = host.host_display_name
        # Extract out the host names we care about from the full mapping
        display_names = [display_name_mapping.get(host_name)
                         for host_name in host_names if
                         display_name_mapping.get(host_name)]
        return display_names

    def _check_violated(self, context, group):
        """Checks if the policy of the group is violated. Returns None, if not
           violated. If policy is violated, returns message string with details
           of the violation."""
        msg = None
        policy = group['policies'][0]
        group_uuid = group['id']

        # Get a list of instances in the group and the host each is on
        sg = objects.instance_group.InstanceGroup()
        try:
            sg = objects.instance_group.InstanceGroup.get_by_uuid(
                context, group_uuid)
            instances = sg_utils.get_hosts_for_each_member(context, sg)
        except exception.InstanceGroupNotFound:
            msg = _('Collocation rule %(id)s not found') % {'id': group_uuid}
            raise webob.exc.HTTPNotFound(explanation=msg)

        # Create a list of all the host names in the instances tuple list
        host_list = [instance[1] for instance in instances]
        if policy == sg_utils.AFFINITY:
            different_hosts = set(host_list)
            if len(different_hosts) > 1:
                different_hosts = self._convert_hosts_to_display_names(
                    context, different_hosts)
                host_str = sg_utils.get_str_from_list(list(different_hosts))
                msg = (_("The virtual machines in this affinity collocation"
                         " rule are not on the same host; they are spread"
                         " across these hosts: %(hosts)s. Take appropriate"
                         " actions to ensure that all the virtual machines in"
                         " this rule are on the same host.") %
                       {'hosts': host_str})
                return msg
        # Policy is anti-affinity
        else:
            # Create a list of each host that is in host_list more than once
            iterated = set()
            duplicate_hosts = set(
                [host for host in host_list if host in iterated or
                 iterated.add(host)])
            if duplicate_hosts:
                duplicate_hosts = self._convert_hosts_to_display_names(
                    context, duplicate_hosts)
                host_str = sg_utils.get_str_from_list(list(duplicate_hosts))
                msg = (_("There are multiple virtual machines in this"
                         " anti-affinity collocation rule on each of the"
                         " following hosts: %(hosts)s. Take appropriate"
                         " actions to ensure that each virtual machine in this"
                         " rule is on a different host.") %
                       {'hosts': host_str})
                return msg
        return msg

    @wsgi.extends
    def index(self, req, resp_obj):
        """Check specific action permissions since OpenStack doesn't."""
        ctxt = req.environ['nova.context']
        self._validate_authorization(ctxt, req, action='index')
        for group in resp_obj.obj['server_groups']:
            msg = self._check_violated(ctxt, group)
            if msg:
                group['policy_violated'] = True
                group['policy_violated_reason'] = msg
            else:
                group['policy_violated'] = False

    @wsgi.extends
    def create(self, req, body):
        """Check specific action permissions since OpenStack doesn't."""
        ctxt = req.environ['nova.context']
        self._validate_authorization(ctxt, req, action='create')
        yield

    @wsgi.extends
    def show(self, req, resp_obj, id):
        """Check specific action permissions since OpenStack doesn't."""
        ctxt = req.environ['nova.context']
        self._validate_authorization(ctxt, req, action='show')
        group = resp_obj.obj['server_group']
        msg = self._check_violated(ctxt, group)
        if msg:
            group['policy_violated'] = True
            group['policy_violated_reason'] = msg
        else:
            group['policy_violated'] = False

    @wsgi.extends
    def delete(self, req, id):
        """Check specific action permissions since OpenStack doesn't."""
        ctxt = req.environ['nova.context']
        self._validate_authorization(ctxt, req, action='delete')
        yield


class ServerGroupMemberController(wsgi.Controller):
    """Server group member API controller for the OpenStack API."""

    def __init__(self, ext_mgr):
        self.ext_mgr = ext_mgr
        self._collection_name = 'members'
        self._entity = 'add_server'

    def _authorize_context(self, req, action=None):
        context = req.environ['nova.context']
        if 'v2.1' in req.url:
            MEMBER_AUTHORIZE_V21(context, action=action)
        else:
            MEMBER_AUTHORIZE_V2(context, action=action)
        return context

    def _validate_body(self, body, entity_name):
        if not self.is_valid_body(body, entity_name):
            msg = _("Malformed request body.")
            raise exception.InvalidInput(reason=msg)

        subbody = dict(body[entity_name])
        server_id = subbody.pop('server_id', None)
        if not server_id:
            msg = _("The value for server_id is either missing or empty.")
            raise exception.InvalidInput(reason=msg)
        elif subbody:
            msg = _("These fields are unsupported: %s") % subbody.keys()
            raise exception.InvalidInput(reason=msg)

    def _delete_member(self, context, group_uuid, instance_uuid):
        """Call the database method to do a soft delete of the
           member in the server group table."""
        payload = {'server_group_id': group_uuid,
                   'instance_uuids': instance_uuid}
        db.instance_group_member_delete(context, group_uuid,
                                        instance_uuid)
        compute_utils.notify_about_server_group_update(context,
                                                       "deletemember", payload)

    def _check_anti_affinity(self, sg_hosts, instance_host, sg, instance):
        """Check that adding the new instance does not violate the
           anti-affinity policy of the server group."""
        if instance_host in sg_hosts:
            # There is already an instance in this server group that
            # is on the same host as the requested instance
            msg = (_("Collocation rule '%(group_uuid)s' contains a virtual"
                     " machine that is on the same host as virtual machine"
                     " '%(instance_name)s' (%(instance_uuid)s). Adding this"
                     " virtual machine would violate the rule's anti-affinity"
                     " policy.") % {'group_uuid': sg.uuid,
                                    'instance_name': instance.display_name,
                                    'instance_uuid': instance.uuid})
            raise webob.exc.HTTPConflict(explanation=msg)

    def _check_affinity(self, sg_hosts, instance_host, sg, instance):
        """Check that the host of the instance is in server group hosts."""
        if sg_hosts and instance_host not in sg_hosts:
            # The instances in this server group are on different hosts
            # than the requested instance is on.
            msg = (_("The virtual machines in collocation rule"
                     " '%(group_uuid)s' are not on the same host that virtual"
                     " machine '%(instance_name)s' (%(instance_uuid)s) is on."
                     " Adding this virtual machine would violate the rule's"
                     " affinity policy.") % {
                   'group_uuid': sg.uuid,
                   'instance_name': instance.display_name,
                   'instance_uuid': instance.uuid})
            raise webob.exc.HTTPConflict(explanation=msg)

    def _check_quota(self, context, sg):
        """Checks that adding the instances does not violate
           the server group members quota."""
        count = QUOTAS.count(context, 'server_group_members',
                             sg, context.user_id)
        try:
            QUOTAS.limit_check(context, server_group_members=count + 1)
        except exception.OverQuota:
            msg = (_("Quota of %(quota)s virtual machine members per"
                     " collocation rule exceeded. There are already the"
                     " maximum number of virtual machines allowed in the"
                     " rule.") % {'quota': CONF.quota_server_group_members})
            raise webob.exc.HTTPForbidden(explanation=msg)

    def _check_instances_states(self, context, group_uuid, instance,
                                is_affinity):
        """Checks the vm_state and task_state for the requested instance
           If it is in a state that involves the instance moving from one
           host to another, raise an exception. If this is an anti-affinity
           policy, do the same check for every instance in the server group.
           This check prevents instances from landing on a new host in the
           middle of adding a new member to the group, which could result in
           the group's policy being violated."""

        # Check states of requested instance
        task_state = instance.task_state
        vm_state = instance.vm_state
        if (task_state in sg_utils.PROHIBITED_TASK_STATES or
                vm_state in sg_utils.PROHIBITED_VM_STATES):
            msg = (_("Virtual machine '%(instance_name)s' (%(instance_uuid)s)"
                     " has a state of '%(vm_state)s' and a task state of"
                     " '%(task_state)s'. There are pending tasks that could"
                     " cause the virtual machine to change hosts. Wait for the"
                     " machine to finish all pending operations then try to"
                     " add it to the collocation rule again.") %
                   {'instance_name': instance.display_name,
                    'instance_uuid': instance.uuid, 'vm_state': vm_state,
                    'task_state': task_state})
            raise webob.exc.HTTPConflict(explanation=msg)

        if is_affinity:
            return

        # If this is an anti-affinity policy, get non deleted instances in the
        # group and do the same check for each.
        try:
            instances = sg_utils.get_members(context, group_uuid)
        except exception.InstanceGroupNotFound:
            msg = (_('Collocation rule %(id)s was not found.') %
                   {'id': group_uuid})
            raise webob.exc.HTTPNotFound(explanation=msg)
        for instance in instances:
            task_state = instance.task_state
            vm_state = instance.vm_state
            if (task_state in sg_utils.PROHIBITED_TASK_STATES or
                    vm_state in sg_utils.PROHIBITED_VM_STATES):
                msg = (_("A member of this collocation rule has a pending task"
                         " that could cause its host to change. Virtual"
                         "  machine '%(instance_name)s'(%(instance_uuid)s) has"
                         " a state of '%(vm_state)s' and a task state of"
                         " '%(task_state)s'. Wait for all members of the"
                         " collocation rule to finish all pending operations"
                         " then try to add the virtual machine again.") %
                       {'instance_name': instance.display_name,
                        'instance_uuid': instance.uuid, 'vm_state': vm_state,
                        'task_state': task_state})
                raise webob.exc.HTTPConflict(explanation=msg)

    def _validate_add(self, context, req, group_uuid, instance):
        """Before adding the instance as a member, validate that it
           can be added to the server group."""

        # Get host of requested instance
        instance_host = instance.host
        if not instance_host:
            msg = (_("Virtual machine '%(instance_name)s' (%(instance_uuid)s)"
                     " is not currently assigned to a host, so it cannot be"
                     " added to a collocation rule.") %
                   {'instance_name': instance.display_name,
                    'instance_uuid': instance.uuid})
            raise webob.exc.HTTPConflict(explanation=msg)

        # Get the instance group object
        sg = objects.instance_group.InstanceGroup()
        try:
            sg = objects.instance_group.InstanceGroup.get_by_uuid(
                context, group_uuid)
        except exception.InstanceGroupNotFound:
            msg = (_('Collocation rule %(id)s was not found.') %
                   {'id': group_uuid})
            raise webob.exc.HTTPNotFound(explanation=msg)

        # Check if instance already member of group
        if instance.uuid in sg.members:
            msg = (_("Virtual machine '%(instance_name)s' (%(instance_uuid)s)"
                     " is already a member of collocation rule '%(id)s'.") %
                   {'instance_name': instance.display_name,
                    'instance_uuid': instance.uuid, 'id': group_uuid})
            raise webob.exc.HTTPConflict(explanation=msg)

        # Check quota
        if 'v2.1' in req.url:
            server_group_quota_extension = self.ext_mgr.get_extensions()
            if 'os-server-group-quotas' in server_group_quota_extension:
                self._check_quota(context, sg)
        else:
            check_server_group_quota = self.ext_mgr.is_loaded(
                'os-server-group-quotas')
            if check_server_group_quota:
                self._check_quota(context, sg)

        # Get a list of hosts for non-deleted instances in the server group
        sg_hosts = sg.get_hosts()

        # Policies can either contain affinity or anti-affinity, not both
        is_affinity = sg_utils.AFFINITY in sg.policies

        # Check state of requested instance
        self._check_instances_states(context, group_uuid, instance,
                                     is_affinity)

        # Validate that adding this instance will not violate the
        # server group's policy
        if is_affinity:
            self._check_affinity(sg_hosts, instance_host, sg, instance)
        else:
            self._check_anti_affinity(sg_hosts, instance_host, sg, instance)

    def _validate_post_add(self, context, group_uuid, instance):
        """ Checks if the addition of the new instance violated the affinity
            or anti-affinty policy of the server group. Reverts the add
            if the policy was violated.

            Due to the prior checks in place, a server group's policy should
            never be violated at this point. It is only here to serve as one
            final check to prevent against a race condition violating the
            policy of the group."""

        # Get the instance group object again, now that a new member
        # has been added.
        sg = objects.instance_group.InstanceGroup()
        try:
            sg = objects.instance_group.InstanceGroup.get_by_uuid(
                context, group_uuid)
            # Get a list of tuples of instance uuids and the host each
            # instance is on for each instance in the server group.
            sg_hosts = sg_utils.get_hosts_for_each_member(context, sg)

        except exception.InstanceGroupNotFound:
            # The server group has been deleted, return
            return

        # Get the host of the instance that was added. Get it from the list of
        # member tuples instead of the instance object passed in, in case the
        # host has changed during the operation.
        for host_tuple in sg_hosts:
            uuid, host = host_tuple
            if uuid == instance.uuid:
                instance_host = host
                break

        # Flatten the list of tuples to a list of only the hosts
        # (including duplicates), and pop off the host of the instance that
        # was added so it can be compared against the list of other hosts in
        # the group.
        sg_hosts = [host2 for uuid2, host2 in sg_hosts if
                    uuid2 != instance.uuid]

        # Check the policy
        try:
            if sg_utils.AFFINITY in sg.policies:
                self._check_affinity(sg_hosts, instance_host, sg, instance)
            else:
                self._check_anti_affinity(sg_hosts, instance_host,
                                          sg, instance)
        except webob.exc.HTTPConflict:
            self._revert_add_member(context, group_uuid, instance)
            raise

    def _revert_add_member(self, context, group_uuid, instance):
        """Revert the addition of the instance into the server group."""
        try:
            self._delete_member(context, group_uuid, instance.uuid)
        except Exception as e:
            LOG.warn(_("Failed to revert the addition of virtual machine"
                       " '%(instance_name)s' (%(instance_uuid)s) to"
                       " collocation rule '%(group_uuid)s' with message"
                       " '%(msg)s'. The policy of the collocation rule may"
                       " have become violated.") %
                     {'instance_name': instance.display_name,
                      'instance_uuid': instance.uuid, 'msg': six.text_type(e),
                      'group_uuid': group_uuid})
            return

    def index(self, req, ServerGroup_id):
        """Returns a list of virtual machines belonging to a server groups."""
        context = self._authorize_context(req, action='index')
        response = {'members': []}
        try:
            instances = sg_utils.get_members(context, ServerGroup_id)
        except exception.InstanceGroupNotFound:
            msg = (_('Collocation rule %(id)s was not found.') %
                   {'id': ServerGroup_id})
            raise webob.exc.HTTPNotFound(explanation=msg)
        if instances:
            # Generate a server view for the members
            view_builder = ServerGroupMemberViewBuilder()
            response = view_builder.index(req, ServerGroup_id, instances)
        return response

    def delete(self, req, ServerGroup_id, id):
        """Removes the specified server from the server group."""
        context = self._authorize_context(req, action='delete')
        try:
            self._delete_member(context, ServerGroup_id, id)
        except exception.InstanceGroupNotFound:
            msg = (_('Collocation rule %(id)s was not found.') %
                   {'id': ServerGroup_id})
            raise webob.exc.HTTPNotFound(explanation=msg)
        except exception.InstanceGroupMemberNotFound:
            msg = _('Virtual machine %(server)s is not a member of collocation'
                    ' rule %(id)s.') % {'server': id, 'id': ServerGroup_id}
            raise webob.exc.HTTPNotFound(explanation=msg)
        return webob.Response(status_int=204)

    def create(self, req, ServerGroup_id, body):
        """Adds a server to the server group."""

        context = self._authorize_context(req, action='create')
        try:
            self._validate_body(body, self._entity)
        except exception.InvalidInput as e:
            raise webob.exc.HTTPBadRequest(explanation=e.format_message())

        # Get uuid of instance to add as a member of the server group
        instance_id = body['add_server']['server_id']
        # Get the instance object
        try:
            instance = objects.instance.Instance.get_by_uuid(
                context, instance_id)
        except exception.InstanceNotFound:
            msg = (_('Virtual machine %(id)s was not found.') %
                   {'id': instance_id})
            raise webob.exc.HTTPNotFound(explanation=msg)

        @utils.synchronized(ServerGroup_id, external=True, lock_path=LOCK_PATH)
        def _do_validate_and_add():
            # Validate this instance can be added as a member
            self._validate_add(context, req, ServerGroup_id, instance)

            # Add the instance as a member
            sg = objects.instance_group.InstanceGroup()
            try:
                sg.add_members(context, ServerGroup_id, [instance_id])
            except exception.InstanceGroupNotFound:
                msg = (_('Collocation rule %(id)s was not found.') %
                       {'id': ServerGroup_id})
                raise webob.exc.HTTPNotFound(explanation=msg)

        _do_validate_and_add()
        self._validate_post_add(context, ServerGroup_id, instance)

        # Get back all the members of the server group
        response = {'members': []}
        try:
            instances = sg_utils.get_members(context, ServerGroup_id)
        except exception.InstanceGroupNotFound:
            msg = (_('Collocation rule %(id)s was not found.') %
                   {'id': ServerGroup_id})
            raise webob.exc.HTTPNotFound(explanation=msg)
        if instances:
            view_builder = ServerGroupMemberViewBuilder()
            response = view_builder.index(req, ServerGroup_id, instances)
        return response


class Server_groups(extensions.ExtensionDescriptor):
    """Extend the Server Groups API."""

    name = "IBM PowerVC Server Groups Extension"
    alias = "ibm-powervc-os-server-groups"
    namespace = "http://docs.openstack.org/compute/ext/" + \
                "ibm_servergroups/api/v1.1"
    updated = "2015-01-09T21:00:00-06:00"

    def get_controller_extensions(self):
        controller = ServerGroupControllerExtension()
        extension = extensions.ControllerExtension(self, 'os-server-groups',
                                                   controller)
        return [extension]

    def get_resources(self):
        res = extensions.ResourceExtension(
            'members',
            controller=ServerGroupMemberController(self.ext_mgr),
            collection_actions={'index': 'GET'},
            parent={'member_name': 'ServerGroup',
                    'collection_name': 'os-server-groups'})

        return [res]


class Server_groupsV21(extensions.V21APIExtensionBase):
    """Extend the Server Groups API."""

    name = "IBM PowerVC Server Groups Extension"
    alias = "ibm-powervc-os-server-groups"
    version = 1

    def get_controller_extensions(self):
        controller = ServerGroupControllerExtension()
        extension = extensions.ControllerExtension(self, 'os-server-groups',
                                                   controller)
        return [extension]

    def get_resources(self):
        res = extensions.ResourceExtension(
            'members',
            controller=ServerGroupMemberController(self.extension_info),
            collection_actions={'index': 'GET'},
            parent={'member_name': 'ServerGroup',
                    'collection_name': 'os-server-groups'})

        return [res]
