#
#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import httplib
import webob
import six
from webob import Response

from oslo_serialization import jsonutils
from nova.api.openstack import extensions
from nova.api.openstack import wsgi
from nova import compute
from oslo_log import log as logging
from powervc_nova import _
from powervc_nova.storage.storage_mapping_api import StorageMapping
from powervc_nova.storage import storage_utils as su

# ------------------------------------
# Extension: storage-hosts
# ------------------------------------
LOG = logging.getLogger(__name__)
EXTENSION = 'storage-hosts'
ALIAS = "server-storage-hosts"
AUTHORIZE_V21 = extensions.os_compute_authorizer(ALIAS)
AUTHORIZE_V2 = extensions.extension_authorizer('compute',
                                               "server_storage_hosts")


class SCGStorageHostsController(wsgi.Controller):

    """
    This Controller handles the API entry implementation of the REST
    function to retrieve the applicable storage-hosts (storage providers)
    for a given instance. The instance may already be associated with a
    storage connectivity group (SCG), but if not, an SCG will be chosen
    for it and used to determine the applicable providers.
    """

    def __init__(self):
        # class statics for the internal API module
        self.stg_api = StorageMapping()
        self.compute_api = compute.API()

    def index(self, req, server_id):
        """
        REST API entry point to call internal API to do work.
        Returns dictionary with the entry 'storage-hosts', containting a list
        of storage host (provider) dictionaries that can be accessed for
        storage allocation, given the VM instance, constrained by the SCG
        associated with the instance.
        For example:
            {
                "storage-hosts": [
                    {
                        "host": "My_V7000_1"
                    }
                ]
            }
        URI e.g.: http://<ip>:8774/v2/714c9fed60be4d4bb53144581c7156d1/
                  servers/d5e89100-65d1-4a6c-b05c-b0b42e0cf346/storage-hosts
        """
        context = req.environ["nova.context"]
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action='list')
        else:
            AUTHORIZE_V2(context, action='list')
        LOG.debug("index() **ENTRY**   server_id=%s" % str(server_id))
        su.http_error_if_ivm()  # HMC-managed installs only

        try:
            instance = self.compute_api.get(context, server_id)
            instance_uuid = instance['uuid']
        except Exception as e:
            LOG.exception(e)
            raise webob.exc.HTTPNotFound(explanation=six.text_type(e))

        # Call the internal nova API
        LOG.debug("Looked up instance %s. Calling "
                  "get_cinder_hosts_by_instance()..." % instance_uuid)
        try:
            resp = self.stg_api.get_providers_by_instance(context,
                                                          instance_uuid,
                                                          instance)
        except Exception as e:
            msg = _("A problem occurred while trying to look up storage "
                    "providers by server instance. Error type is '%(type)s'. "
                    "Error message: %(error)s") %\
                dict(type=e.__class__.__name__, error=e)
            LOG.exception(msg)
            raise webob.exc.HTTPServerError(explanation=six.text_type(msg))

        # Return results in JSON body response
        return Response(request=req,
                        status=httplib.OK,
                        content_type='application/json',
                        body=jsonutils.dumps(resp))


class Server_storage_hosts(extensions.ExtensionDescriptor):
    ''' Support for REST API - SCG to candidate storage-host mapping. '''
    name = "Compute Server StorageHosts"
    alias = "server-" + EXTENSION
    # The namespace just follows the convention of other APIs - is there
    # separate work that needs to be done to register this?
    namespace = ("http://docs.openstack.org/compute/ext/storage/"
                 "server_storage_hosts/api/v1.1")
    updated = "2013-07-12T11:00:00-06:00"

    def get_resources(self):
        '''
        Provide a Resource Extension for getting candidate deploy hosts
        provided an instance associated with a storage connectivity group.
        '''
        parent_def = {'member_name': 'server', 'collection_name': 'servers'}
        return [
            extensions.ResourceExtension(
                EXTENSION,
                controller=SCGStorageHostsController(),
                parent=parent_def)]


class Server_storage_hostsV21(extensions.V21APIExtensionBase):
    ''' Support for REST API - SCG to candidate storage-host mapping. '''
    name = "Compute Server StorageHosts"
    alias = ALIAS
    version = 1

    def get_resources(self):
        '''
        Provide a Resource Extension for getting candidate deploy hosts
        provided an instance associated with a storage connectivity group.
        '''
        parent_def = {'member_name': 'server', 'collection_name': 'servers'}
        return [
            extensions.ResourceExtension(
                EXTENSION,
                controller=SCGStorageHostsController(),
                parent=parent_def)]

    def get_controller_extensions(self):
        return []
