# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import httplib
import six
from webob import Response
import webob.exc as wex
from webob.exc import HTTPError

from oslo_log import log as logging
from oslo_serialization import jsonutils
from nova.api.openstack import extensions, wsgi
from powervc_nova import _
from powervc_nova.volume import cinder as pvc_cinder
from powervc_nova.storage.volume_instance_mapping_api \
    import VolumeInstanceMapping

ALIAS = 'volume_instance_mapping'
IDENTITY = "volume-instance-mapping"

AUTHORIZE_V21 = extensions.os_compute_authorizer(IDENTITY)
AUTHORIZE_V2 = extensions.extension_authorizer('compute',
                                               ALIAS)

LOG = logging.getLogger(__name__)


class VolumeInstanceMappingController(wsgi.Controller):

    """
    This Controller handles GET, POST and PUT requests to query the database
    and return JSON data manage storage connectivity groups.
    """

    @wsgi.extends
    def show(self, req, id):
        """
        REST GET method to return details on one volume.
        """

        context = self._validate_authorization(req.environ['nova.context'],
                                               req,
                                               'show')
        LOG.debug("show() **ENTER**  req = %s, id=%s, context=%s"
                  % (req, id, six.text_type(context)))
        try:
            vol_map_api = VolumeInstanceMapping()
            if 'instance' in req.GET:
                volume_detail = \
                    vol_map_api.get_instance_info(context, id)
            else:
                volume_detail = \
                    vol_map_api.get_volume_info(context, id)

        except HTTPError as http_ex:
            raise http_ex  # rethrow
        except Exception as ex:
            LOG.exception(ex)
            msg = _("Problem occurred while trying to obtain detailed volumes "
                    "and connection information. Error: %s") % ex
            raise wex.HTTPServerError(explanation=six.text_type(msg))

        return Response(request=req,
                        status=httplib.OK,
                        content_type='application/json',
                        body=jsonutils.dumps(volume_detail))

    @staticmethod
    def _validate_authorization(context, req, action):
        """
        This method authorizes the request against the service's policy.
        :param context: API request context.
        :param authorize_method: Which method from this class to authorize.
        :returns: An authorized request context.
        """
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action=action)
        else:
            AUTHORIZE_V2(context, action=action)
        return context.elevated()


class Volume_instance_mapping(extensions.ExtensionDescriptor):

    """ Support to obtain detailed volume and connection info. """
    name = "Volume_Instance_Mapping"
    alias = ALIAS
    # The namespace just follows the convention of other APIs - is there
    # separate work that needs to be done to register this?
    namespace = ("http://docs.openstack.org/compute/ext/" +
                 "host_storage_mapping/api/v1.1")
    updated = "2015-07-27T11:00:00-06:00"

    def get_resources(self):
        """ Provide a Resource Extension for Volume_Instance_Mapping """
        return [extensions.ResourceExtension(
            ALIAS,
            controller=VolumeInstanceMappingController())]


class Volume_instance_mappingV21(extensions.V21APIExtensionBase):

    """ Support to obtain detailed volume and connection info. """
    name = "Volume_Instance_Mapping"
    alias = ALIAS
    version = 1

    def get_resources(self):
        """ Provide a Resource Extension for Volume_Instance_Mapping """
        return [extensions.ResourceExtension(
            ALIAS,
            controller=VolumeInstanceMappingController())]

    def get_controller_extensions(self):
        return []
