# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Shared Processor Pools related utilities."""

from oslo_log import log as logging

from powervc_nova.objects.compute.dom import SharedProcessorPoolFactory

LOG = logging.getLogger(__name__)


def _construct_proc_pool_info(proc_pool):
    """
    Construct a dictionary with shared processor pool info from
    SharedProcessorPool dom object
    :return dict contains id, name, max_proc_units, reserved_proc_units,
            available_proc_units, hypervisor_hostname
    """
    return dict(id=proc_pool.pool_id, name=proc_pool.pool_name,
                max_proc_units=proc_pool.maximum_proc_units,
                reserved_proc_units=proc_pool.reserved_proc_units,
                available_proc_units=proc_pool.available_proc_units,
                hypervisor_hostname=proc_pool.host_name)


def _construct_all_proc_pool_dict(spps, list_name_only):
    """
    Construct a dictionary with all shared processor pool info from
    SharedProcessorPool dom object and host name
    :return dict contains id, name, max_proc_units, reserved_proc_units,
            available_proc_units
    """
    spp_set = set()
    spp_list = []
    for proc_pool in spps:
        if list_name_only == 'true':
            spp_set.add(proc_pool.pool_name)
        else:
            spp_list.append(dict(id=proc_pool.pool_id,
                                 name=proc_pool.pool_name,
                                 max_proc_units=
                                 proc_pool.maximum_proc_units,
                                 reserved_proc_units=
                                 proc_pool.reserved_proc_units,
                                 available_proc_units=
                                 proc_pool.available_proc_units,
                                 host=proc_pool.host_name))

    if list_name_only == 'true':
        return spp_set
    else:
        return spp_list


def get_shared_proc_pool(context, hypervisor_hostname, pool_name):
    """
    Call the dom methods to get information of a shared processor pool
    for a given hypervisor host name and a given pool name.
    """
    shared_processor_pool = None
    # find all the shared processor pools on the host
    spp_fact = SharedProcessorPoolFactory.get_factory()
    spps = spp_fact.find_all_spps_by_host(context, hypervisor_hostname)

    if spps:
        # find the pool matching the name
        for proc_pool in spps:
            if proc_pool.pool_name == pool_name:
                shared_processor_pool = _construct_proc_pool_info(proc_pool)
                break

    return shared_processor_pool


def get_all_shared_proc_pools(context, host_name, list_name_only):
    """
    Call the dom methods to get information for all shared processor pools
    on a given hypervisor host name.
    """
    # find all the shared processor pools on the host
    spp_factory = SharedProcessorPoolFactory.get_factory()
    spps = spp_factory.find_all_spps_by_host(context, host_name)

    # Create the shared processor pools dictionary
    shared_proc_pool = _construct_all_proc_pool_dict(spps, list_name_only)

    return shared_proc_pool
