# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from oslo_log import log as logging
from nova.objects import flavor as flavor_obj
from powervc_nova import logcall
from powervc_nova.volume import cinder as volume

from powervc_nova import _

LOG = logging.getLogger(__name__)

IMAGE_VOLUME_TYPE_PREFIX = 'powervm:image_volume_type_'


def _check_for_invalid_image_volumes_type_keys(volume_ids, extra_specs):
    prefix_len = len(IMAGE_VOLUME_TYPE_PREFIX)
    for extra_spec in extra_specs.keys():
        if extra_spec.startswith(IMAGE_VOLUME_TYPE_PREFIX):
            volume_id = extra_spec[prefix_len:]
            if volume_id not in volume_ids:
                LOG.warn(_('The volume %(volume_id)s from the instance flavor '
                           'was not found in the image.') %
                         {'volume_id': volume_id})


@logcall
def get_volume_types_from_flavor(context, volume_ids, flavor=None,
                                 instance=None):
    """
        This method will return a dictionary of source volume ID to the
        Cinder volume type ID that should be used when the source volume is
        cloned.
    """
    volume_api = volume.PowerVCVolumeAPI()
    available_volume_types = volume_api.list_volume_types(context)
    source_vol_to_vol_type = {}
    if instance and not flavor:
        flavor = flavor_obj.Flavor.get_by_id(
            context, instance['instance_type_id'])

    if not flavor:
        LOG.debug("No flavor passed in. Instance %(inst)s volumes:%(vols)s"
                  % dict(inst=instance['id'], vols=volume_ids))
        return source_vol_to_vol_type
    extra_specs = flavor.get('extra_specs')
    if not extra_specs:
        LOG.debug("No extra_spec in flavor %(flv)s" % dict(flv=flavor))
        return source_vol_to_vol_type

    _check_for_invalid_image_volumes_type_keys(volume_ids, extra_specs)
    # Handle the backward compatibilty case first.
    # If both powervm:boot_volume_type and
    # powervm:image_volume_type_<sourcevolUUID> are set on a deploy with
    # one BDM, the image_volume_type extra spec will take precedence.
    if len(volume_ids) == 1:
        boot_volume_type = extra_specs.get('powervm:'
                                           'boot_volume_type')

        if available_volume_types and boot_volume_type:
            for item in available_volume_types:
                if (item['id'] == boot_volume_type or
                        item['name'] == boot_volume_type):
                    # It is not clear whether boot_volume_type passed in is
                    # a volume type name or id.
                    LOG.debug("boot volume type: "
                              "%(voltype)s, available volume type: "
                              "%(voltypelist)s." %
                              {'voltype': boot_volume_type,
                               'voltypelist': available_volume_types})
                    source_vol_to_vol_type[volume_ids[0]] = item
                    break

    volume_type_dict = dict([(item['id'], item) for item in
                             available_volume_types])
    extra_spec_template = IMAGE_VOLUME_TYPE_PREFIX + '%s'
    for volume_id in volume_ids:
        extra_spec = extra_spec_template % volume_id
        vol_type = extra_specs.get(extra_spec)
        if not vol_type:
            LOG.debug("No image_volume_type in flavor for volume"
                      " %(vol_id)s." % dict(vol_id=volume_id))
        elif vol_type in volume_type_dict:
            if volume_id in source_vol_to_vol_type:
                LOG.debug("Overriding boot volume type with image volume "
                          "type: %s." % volume_type_dict[vol_type])
            source_vol_to_vol_type[volume_id] = volume_type_dict[vol_type]
        else:
            LOG.warn(_('The volume type requested, %(requested_id)s for '
                       'the clone of image volume %(volume_id)s '
                       'does not exist.')
                     % {'requested_id': vol_type,
                        'volume_id': volume_id})

    return source_vol_to_vol_type
