#
#
# Copyright 2013, 2016 IBM Corp.
# All Rights Reserved.
# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2011 Piston Cloud Computing, Inc.
# Copyright 2012-2013 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import eventlet
from oslo_concurrency import lockutils
from oslo_config import cfg
import oslo_messaging as messaging
from nova import objects
from nova import rpc
from nova import servicegroup
from nova.compute import api as base_api
from nova.compute import rpcapi as base_rpcapi
from nova.compute import utils as compute_utils
from nova.objects import base as objects_base
from oslo_log import log as logging
from powervc_nova.objects.network import dom_kvm
from powervc_nova import _
from powervc_nova import logcall
from powervc_nova.volume import cinder as pvc_volume_api


cfg.CONF.register_opts([
    cfg.IntOpt('onboard_discover_timeout', default=1200,
               help='Timeout for RPC Discover Resources for On-boarding'),
    cfg.IntOpt('onboard_query_timeout', default=3600,
               help='Timeout for RPC Query Resource Details for On-boarding'),
    cfg.IntOpt('verify_connection_timeout', default=10,
               help='Timeout for RPC Verify Connection Timeout'),
    cfg.IntOpt('is_instance_on_host_timeout', default=60,
               help='Timeout for is_instance_on_host'),
    cfg.IntOpt('sync_enterprise_pool_for_host_timeout', default=120,
               help='Timeout for RPC sync enterprise for host')
])

LOG = logging.getLogger(__name__)


class PowerVCComputeAPI(base_api.API):

    """PowerVC Extension to Compute API for Compute Manager communication"""

    def __init__(self, *args, **kwargs):
        super(PowerVCComputeAPI, self).__init__(*args, **kwargs)
        self.compute_rpcapi = PowerVCComputeRPCAPI()
        self.volume_api = pvc_volume_api.PowerVCVolumeAPI()


class PowerVCHostAPI(base_api.HostAPI):

    """PowerVC Extension to Host API for Compute Manager communication"""

    def __init__(self, *args, **kwargs):
        super(PowerVCHostAPI, self).__init__(*args, **kwargs)
        self.rpcapi = PowerVCComputeRPCAPI()

    def change_vif_sea(self, context, data, host_name):
        """Call to make the SEA change for a VIF"""
        return self.rpcapi.change_vif_sea(context, data,
                                          host_name)

    def discover_instances(self, context, host):
        """Returns a list of all of the VM's that exist on the Host"""
        return self.rpcapi.discover_instances(context, host)

    def query_instances(self, context, host, instance_ids):
        """Returns details about the VM's that were requested"""
        return self.rpcapi.query_instances(context, host, instance_ids)

    def get_next_instances_chunk(self, context, host, identifier):
        """Provides Chunking of VM's Lists to avoid QPID Limits"""
        return self.rpcapi.get_next_instances_chunk(context, host, identifier)

    def verify_host_running(self, context, host, max_wait=0):
        """Verifies the nova-compute service for the Host is running"""
        # We will default to checking once, but allow the caller to check more
        for index in range(max_wait / cfg.CONF.verify_connection_timeout + 1):
            if self.rpcapi.verify_host_running(context, host):
                index = index + 1
                return True
        return False

    def unmanage_host(self, context, host):
        """Allows the Driver to do any necessary cleanup on Host Removal"""
        # Only call unmanage if the Nova Compute process is running
        if self.verify_host_running(context, host):
            return self.rpcapi.unmanage_host(context, host)

    def remanage_host(self, context, new_mgmt_ip):
        """Tells the old Mgmt System that a new Mgmt System is managing"""
        return self.rpcapi.remanage_host(context, new_mgmt_ip)

    def inventory_instances(self, context, host, instances=None):
        """Call the inventory_instances periodic task on the host"""
        self.rpcapi.inventory_instances(context, host, instances)

    def prepare_instance(self, context, host, instance, preparation):
        """Prepare a VM for onboarding by changing the VM configuration"""
        return self.rpcapi.prepare_instance(context, host, instance,
                                            preparation)

    def finish_onboard(self, context, host, instance, preparation=None):
        """Finish the VM onboarding process"""
        return self.rpcapi.finish_onboard(
            context, host, instance, preparation)

    @logcall
    def get_host_ovs(self, context, host):
        """Retrieve ovs host data from a single kvm host"""
        ovs_dom = None
        try:
            retVal = self.rpcapi.get_host_ovs(context, host)
            ovs_dom = dom_kvm.HostOVSNetworkConfig(host)
            ovs_dom.from_dict(retVal)
        except Exception as e:
            LOG.exception(e)
            if ovs_dom is None:
                ovs_dom = dom_kvm.HostOVSNetworkConfig(host)

            error_obj = dom_kvm.Error("%s" % e)
            ovs_dom.error_list.append(error_obj)

        return ovs_dom

    @logcall
    def update_host_ovs(self, context, host, dom, force_flag, rollback):
        """Verify and/or update kvm host ovs data from dom"""
        ERRORS_KEY = "errors"
        retVal = {}
        try:
            retVal = self.rpcapi.update_host_ovs(context,
                                                 host,
                                                 dom,
                                                 force_flag,
                                                 rollback)
        except Exception as exc:
            LOG.error(exc)
            if ERRORS_KEY not in retVal:
                retVal[ERRORS_KEY] = \
                    [{"%s" % exc.__class__.__name__: "%s" % exc}]
            else:
                retVal[ERRORS_KEY].append(
                    {"%s" % exc.__class__.__name__: "%s" % exc})

        return retVal

    @logcall
    def sync_enterprise_pool_for_host(self, context, host):
        """Sync enterprise pool info for the host"""
        return self.rpcapi.sync_enterprise_pool_for_host(context, host)

    @lockutils.synchronized('enterprise_pools',
                            'powervc-',
                            external=True)
    @logcall
    def set_mobile_cores(self, context, host_name, desired_mobile_cores):
        """
        Set the number of mobile cores on a host by activating/deactivating
        cores

        :param context: nova context object
        :param host_name: the host name of the host to change the number of
                          mobile cores it uses
        :param desired_mobile_cores: an integer representing the number of
                                     mobile cores desired for the host to use
        :returns: a list of [result, reason]. For example, if the mobile cores
                  were successfully updated, then a list of [True] will be
                  returned. Otherwise, it would return a list of
                  [False, 'Reason for failure'].
        """
        host_name = self._assert_host_exists(context, host_name)
        payload = {'host_name': host_name,
                   'desired_mobile_cores': desired_mobile_cores}
        compute_utils.notify_about_host_update(context,
                                               'set_mobile_cores.start',
                                               payload)
        result = self.rpcapi.set_mobile_cores(
            context,
            desired_mobile_cores=desired_mobile_cores,
            host=host_name)
        compute_utils.notify_about_host_update(context,
                                               'set_mobile_cores.end',
                                               payload)
        return result

    @logcall
    def create_host_ovs(self, context, host, dom, force_flag, rollback):
        """Verify and/or create kvm host ovs from dom"""
        ERRORS_KEY = "errors"
        retVal = {}
        try:
            retVal = self.rpcapi.create_host_ovs(context,
                                                 host,
                                                 dom,
                                                 force_flag,
                                                 rollback)
        except Exception as exc:
            LOG.error(exc)
            if ERRORS_KEY not in retVal:
                retVal[ERRORS_KEY] = \
                    [{"%s" % exc.__class__.__name__: "%s" % exc}]
            else:
                retVal[ERRORS_KEY].append(
                    {"%s" % exc.__class__.__name__: "%s" % exc})

        return retVal

    @logcall
    def delete_host_ovs(self, context, host, ovs_list, force_flag, rollback):
        """Verify and/or delete kvm host ovs from dom"""
        ERRORS_KEY = "errors"
        retVal = {}
        try:
            retVal = self.rpcapi.delete_host_ovs(context,
                                                 host,
                                                 ovs_list,
                                                 force_flag,
                                                 rollback)
        except Exception as exc:
            LOG.error(exc)
            if ERRORS_KEY not in retVal:
                retVal[ERRORS_KEY] = \
                    [{"%s" % exc.__class__.__name__: "%s" % exc}]
            else:
                retVal[ERRORS_KEY].append(
                    {"%s" % exc.__class__.__name__: "%s" % exc})

        return retVal

    def set_host_dead(self, context, host, dead=True, target_host=None):
        return self.rpcapi.set_host_dead(context,
                                         host, dead, target_host)


class PowerVCComputeRPCAPI(base_rpcapi.ComputeAPI):

    """PowerVC Extension to RPC API for Compute Manager communication"""

    def __init__(self):
        super(PowerVCComputeRPCAPI, self).__init__()
        self.servicegroup_api = servicegroup.API()
        self.topic = cfg.CONF.compute_topic

    def change_vif_sea(self, context, data, host):
        """Makes a Unplug/Plug call into the VIF Driver"""
        msg = self.make_msg('change_vif_sea', data=data)
        topic = '%s.%s' % (self.topic, host)
        return self.call(context, msg, topic=topic)

    def discover_instances(self, context, host):
        """Returns a list of all of the VM's that exist on the Host"""
        msg = self.make_msg('discover_instances')
        topic = '%s.%s' % (self.topic, host)
        return self.call(context, msg, topic=topic,
                         timeout=cfg.CONF.onboard_discover_timeout)

    def query_instances(self, context, host,
                        instance_ids, allow_unsupported=False):
        """Returns details about the VM's that were requested"""
        msg = self.make_msg('query_instances', instance_ids=instance_ids,
                            allow_unsupported=allow_unsupported)
        topic = '%s.%s' % (self.topic, host)
        return self.call(context, msg, topic=topic,
                         timeout=cfg.CONF.onboard_query_timeout)

    def get_next_instances_chunk(self, context, host, identifier):
        """Provides Chunking of VM's Lists to avoid QPID Limits"""
        msg = self.make_msg('get_next_instances_chunk', identifier=identifier)
        topic = '%s.%s' % (self.topic, host)
        return self.call(context, msg, topic=topic)

    def verify_host_running(self, context, host):
        """Verifies the nova-compute service for the Host is running"""
        msg = self.make_msg('verify_host_running')
        topic = '%s.%s' % (self.topic, host)
        try:
            return self.call(context, msg, topic=topic,
                             timeout=cfg.CONF.verify_connection_timeout)
        # If we got a timeout, just log for debug that the process is down
        except Exception:
            LOG.debug('RPC verify_host_running timed-out for ' + host)
            return False

    def unmanage_host(self, context, host):
        """Allows the Driver to do any necessary cleanup on Host Removal"""
        msg = self.make_msg('unmanage_host')
        topic = '%s.%s' % (self.topic, host)
        return self.call(context, msg, topic=topic)

    def remanage_host(self, context, new_mgmt_ip):
        """Tells the old Mgmt System that a new Mgmt System is managing"""
        # We will do a no-op now until the Compute Mgr is overriden on KVM
        pass

    def inventory_instances(self, context, host, instances=None):
        """Calls the inventory_instances periodic task on the host"""
        cctxt = self.client.prepare(server=host, version='4.0')
        cctxt.cast(context, 'inventory_instances', instances=instances)

    def prepare_instance(self, context, host, instance, preparation):
        """Prepare a VM for onboarding by changing the VM configuration"""
        msg = self.make_msg('prepare_instance', instance=instance,
                            preparation=preparation)
        topic = '%s.%s' % (self.topic, host)
        try:
            return self.call(context, msg, topic=topic,
                             timeout=cfg.CONF.onboard_query_timeout)
        except Exception as exc:
            # uuid is not guaranteed during prepare
            # Changing instance['uuid'] to instance['name']
            LOG.warn(_('RPC prepare_instance failed for %s') %
                     instance['name'])
            LOG.exception(exc)
            return None

    def finish_onboard(self, context, host, instance, preparation):
        """Finish onboarding a VM"""
        msg = self.make_msg(
            'finish_onboard', instance=instance, preparation=preparation)
        topic = '%s.%s' % (self.topic, host)
        try:
            return self.call(context, msg, topic=topic,
                             timeout=cfg.CONF.onboard_discover_timeout)
        except Exception as exc:
            LOG.warn(_('RPC finish_onboard failed for %s') % instance['uuid'])
            LOG.exception(exc)
            return False

    def get_host_ovs(self, context, host):
        """ Make rpc call to retrieve ovs host data from kvm host"""

        retval = {}

        # First, be sure the compute process is running
        context = context.elevated()
        service = objects.Service.get_by_compute_host(context, host)
        if (self.servicegroup_api.service_is_up(service)):
            LOG.info(_("initiated get_host_ovs rpc call to host %s") % host)

            msg = self.make_msg('get_host_ovs')
            topic = '%s.%s' % (self.topic, host)

            tries = 12
            for i in range(tries):
                try:
                    retval = self.call(context, msg, topic=topic, timeout=10)
                    break

                except messaging.MessagingTimeout as exc:
                    if i < (tries - 1):
                        LOG.debug('RPC GET timeout, retrying time %s' % str(i))
                        continue

                    LOG.exception(_("Exception from RPC call for host-ovs GET "
                                    "for host %s." % host))
                    """
                    If we get here, we know that we have no return data from
                    the call and therefore must create a new dict representing
                    the dom that contains the error info for the timeout.
                    """
                    error_dom = dom_kvm.HostOVSNetworkConfig(host)
                    error_obj = dom_kvm.Error("%s" % exc)
                    error_dom.error_list = [error_obj]

                    retval = error_dom.to_dict()
        else:
            LOG.info(_("compute process not running on host %s") % host)

        return retval

    def update_host_ovs(self, context, host, dom, force_flag, rollback):
        """ Make rpc call to update ovs host data on kvm host"""
        LOG.info(_("initiated update_host_ovs rpc call to host %s") % host)
        ERRORS_KEY = "errors"

        msg = self.make_msg('update_host_ovs',
                            dom=dom,
                            force_flag=force_flag,
                            rollback=rollback)
        topic = '%s.%s' % (self.topic, host)
        retVal = {}

        try:
            retVal = self.call(context, msg, topic=topic, timeout=900)
        except messaging.MessagingTimeout as exc:
            LOG.exception(_("Exception from RPC call for host-ovs PUT for "
                            "host %s." % host))
            """
            If we get here, we know that we have no return data from
            the call and therefore must create a new dict containing
            the error info for the timeout.
            """
            LOG.error(exc)
            if ERRORS_KEY not in retVal:
                retVal[ERRORS_KEY] = \
                    [{"%s" % exc.__class__.__name__: "%s" % exc}]
            else:
                retVal[ERRORS_KEY].append(
                    {"%s" % exc.__class__.__name__: "%s" % exc})

        return retVal

    @logcall
    def sync_enterprise_pool_for_host(self, context, host):
        """Make RPC call to sync enterprise pool info for the host"""
        msg = self.make_msg('sync_enterprise_pool_for_host')
        topic = '%s.%s' % (self.topic, host)
        return self.call(
            context, msg, topic=topic,
            timeout=cfg.CONF.sync_enterprise_pool_for_host_timeout)

    @logcall
    def set_mobile_cores(self, context, desired_mobile_cores, host):
        """ Make RPC call to activate/deactivate cores on PowerVM host"""
        msg = self.make_msg('set_mobile_cores',
                            desired_mobile_cores=desired_mobile_cores)
        topic = '%s.%s' % (self.topic, host)
        return self.call(context, msg, topic=topic)

    @logcall
    def create_host_ovs(self, context, host, dom, force_flag, rollback):
        """ Make rpc call to create ovs on kvm host"""
        ERRORS_KEY = "errors"

        msg = self.make_msg('create_host_ovs',
                            dom=dom,
                            force_flag=force_flag,
                            rollback=rollback)
        topic = '%s.%s' % (self.topic, host)
        retVal = {}

        try:
            retVal = self.call(context, msg, topic=topic, timeout=900)
        except messaging.MessagingTimeout as exc:
            LOG.exception(_("Exception from RPC call for host-ovs POST "
                            "for host %s." % host))
            """
            If we get here, we know that we have no return data from
            the call and therefore must create a new dict containing
            the error info for the timeout.
            """
            LOG.error(exc)
            if ERRORS_KEY not in retVal:
                retVal[ERRORS_KEY] = \
                    [{"%s" % exc.__class__.__name__: "%s" % exc}]
            else:
                retVal[ERRORS_KEY].append(
                    {"%s" % exc.__class__.__name__: "%s" % exc})

        return retVal

    @logcall
    def delete_host_ovs(self, context, host, ovs_list, force_flag, rollback):
        """ Make rpc call to delete ovs on kvm host"""
        ERRORS_KEY = "errors"

        msg = self.make_msg('delete_host_ovs',
                            ovs_list=ovs_list,
                            force_flag=force_flag,
                            rollback=rollback)
        topic = '%s.%s' % (self.topic, host)
        retVal = {}

        try:
            retVal = self.call(context, msg, topic=topic, timeout=900)
        except messaging.MessagingTimeout as exc:
            LOG.exception(_("Exception from RPC call for host-ovs DELETE "
                            "for host %s." % host))
            """
            If we get here, we know that we have no return data from
            the call and therefore must create a new dict containing
            the error info for the timeout.
            """
            LOG.error(exc)
            if ERRORS_KEY not in retVal:
                retVal[ERRORS_KEY] = \
                    [{"%s" % exc.__class__.__name__: "%s" % exc}]
            else:
                retVal[ERRORS_KEY].append(
                    {"%s" % exc.__class__.__name__: "%s" % exc})

        return retVal

    def post_live_migration_at_destination(self, ctxt, instance,
                                           block_migration, host,
                                           migrate_data=None):
        """
        Override this method to pass in migrate_data
        """
        version = '4.0'
        cctxt = self.client.prepare(server=host, version=version)
        cctxt.cast(ctxt, 'post_live_migration_at_destination',
                   instance=instance, block_migration=block_migration,
                   migrate_data=migrate_data)

    def is_instance_on_host(self, ctxt, instance_uuid, instance_name, host):
        """
        Verify if an instance is on a remote host
        """
        answer = True
        topic = '%s.%s' % (self.topic, host)
        mesg = self.make_msg('is_instance_on_host',
                             instance_uuid=instance_uuid,
                             instance_name=instance_name)
        try:
            answer = self.call(ctxt,
                               mesg,
                               topic,
                               timeout=(cfg.CONF.
                                        is_instance_on_host_timeout),
                               serializer=objects_base.NovaObjectSerializer())
        except Exception as e:
            LOG.exception(_('Exception encountered when checking if an '
                            'instance %(name)s (%(uuid)s) is being managed '
                            'by remote host %(host)s. Exception: %(exp)s') %
                          {'uuid': instance_uuid, 'name': instance_name,
                           'host': host, 'exp': e})
        finally:
            return answer

    def set_host_dead(self, ctxt, host, dead=True, target_host=None):
        """
        Verify if an instance is on a remote host
        """
        topic = '%s.%s' % (self.topic, host)
        mesg = self.make_msg('set_host_dead', dead=dead,
                             target_host=target_host)
        return self.call(ctxt,
                         mesg,
                         topic,
                         timeout=900)

    def end_compute_with_hmc(self, context, data):
        """
        End nova compute process if it's communicating via the
        specified HMC.
        """
        LOG.debug('Making end_compute_with_hmc cast %s' % data)
        msg = self.make_msg('end_compute_with_hmc', data=data)
        return self.fanout_cast(context, msg, topic=self.topic)

    def make_msg(self, method, **kwargs):
        """
        This method used to exist in base OpenStack, but was removed during
        Icehouse.  It is included here to preserve compatibility with existing
        PowerVC code.
        """
        return {'method': method, 'namespace': None, 'args': kwargs}

    def call(self, ctxt, rpc_message, topic, version='4.0', timeout=None,
             serializer=None):
        """
        This method used to exist in base OpenStack, but was removed during
        Icehouse.  It is included here to preserve compatibility with existing
        PowerVC code.
        """
        topic, _sep, server = topic.partition('.')
        cctxt = rpc.get_client(messaging.Target(topic=topic,
                                                server=server or None,
                                                version=version),
                               serializer=serializer)
        method = rpc_message['method']
        kwargs = rpc_message['args']

        # Since we could get into a deadlock-like situation if the AMQP
        # reconnect hangs, we will set a wrapper timeout slightly larger
        # than original timeout that breaks out rather than hanging forever
        hang_timeout = (timeout + 30) if timeout else 7200
        try:
            # Use the EventLet Timeout to make sure we don't hang
            with eventlet.timeout.Timeout(hang_timeout):
                cctxt = cctxt.prepare(timeout=timeout, version=version)
                return cctxt.call(ctxt, method, **kwargs)
        # Raise an exception so that we can parse appropriately
        except eventlet.timeout.Timeout:
            cstr = topic + ' - ' + method
            raise Exception('Hit wrapper deadlock timeout on %s call' % cstr)
