# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""PowerVC Extensions to the Compute Manager for KVM-specific features"""

import os
import re

from nova.compute import power_state, task_states
from nova.compute import utils as compute_utils
from nova.compute.manager import ComputeManager as base_manager
from nova.compute import vm_states
from nova.network import model as network_model
from nova import exception
from nova.objects import block_device as blkdev_obj
from nova import rpc
from nova.virt.libvirt import utils as libvirt_utils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils


from powervc_nova.compute import kvmhmc_manager
from powervc_nova.compute import manager
from powervc_nova.compute import notify_messages as notify_msg
from powervc_nova.compute import send_ui_notification
from powervc_nova import _
from powervc_nova.virt import discovery_driver as discovery_driver

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class PowerVCKvmComputeManager(kvmhmc_manager.PowerVCKvmHmcComputeManager):
    """
    Extends the Compute Manager to provide the additional KVM-specific
    capabilities that are beyond what the OpenStack Community provides.
    """

    def __init__(self, *args, **kwargs):
        """Performs the Additional HMC-specific Manager Initialization"""
        super(PowerVCKvmComputeManager, self).__init__(*args, **kwargs)

    def _shutdown_instance(self, context, instance,
                           bdms, requested_networks=None, notify=True,
                           try_deallocate_networks=True):
        """ override openstack _shutdown_instance in order to
            1. Add exception handling for getting network info
            2. get latest bdms
            3. check vm state before calling destroy
            4. After shutdown, clean up imported networks
               which are created during onboarding
        """

        """Shutdown an instance on this host."""
        context = context.elevated()
        # get network info before tearing down
        try:
            network_info = compute_utils.get_nw_info_for_instance(instance)
        except (exception.NetworkNotFound, exception.NoMoreFixedIps):
            network_info = network_model.NetworkInfo()

        # refresh bdms from database to sync up the in memory copy
        bdms = blkdev_obj.BlockDeviceMappingList.get_by_instance_uuid(
            context, instance['uuid'])

        if not self._check_vm_state(context, instance):
            return
        if notify:
            self._notify_about_instance_usage(context, instance,
                                              "shutdown.start")

        try:
            # sending try_deallocate_networks as False
            # we will handle after super call
            base_manager._shutdown_instance(
                self, context, instance, bdms, requested_networks=None,
                notify=False, try_deallocate_networks=False)

        except exception.InstancePowerOffFailure:
            # if the instance can't power off, don't release the ip
            with excutils.save_and_reraise_exception():
                pass
        except Exception:
            with excutils.save_and_reraise_exception():
                # deallocate ip and fail without proceeding to
                # volume api calls, preserving current behavior
                try:
                    if try_deallocate_networks:
                        self._try_deallocate_network(context, instance,
                                                     requested_networks)
                        # If this is an imported network, we want to clean it
                        # up
                        if requested_networks is None:
                            self._cleanup_imported_networks(context,
                                                            network_info)
                except Exception as ex:
                    # vm may have been destroyed and there is no way
                    # back. Try to clean up as much as we can.
                    # The error has been logged in _try_deallocate_network.
                    reason = (_("Failed to remove the VLAN associated with "
                                "the vm virtual ethernet. Please check "
                                "host log for further details. Error: %(ex)s")
                              % dict(ex=ex))
                    self._error_notify(instance, context, 'warn', reason)
                    LOG.warn(_("Failed to remove the VLAN associated with "
                               "vm virtual ethernet. Instance: %(inst)s. "
                               "Error:%(error)s") %
                             {'lpar_name': instance['name'],
                              'inst': instance['uuid'],
                              'error': ex})
        try:
            if try_deallocate_networks:
                self._try_deallocate_network(context, instance,
                                             requested_networks)
                # If this is an imported network, we want to clean it up
                if requested_networks is None:
                    self._cleanup_imported_networks(context, network_info)
        except Exception as ex:
            # vm may have been destroyed and there is no way
            # back. Try to clean up as much as we can.
            # The error has been logged in _try_deallocate_network.
            reason = (_("Deleted vm successfully. But failed to remove "
                        "the VLAN associated with the vm virtual ethernet. "
                        "Please check host log for further "
                        "details. Error: %(ex)s") % dict(ex=ex))
            self._error_notify(instance, context, 'warn', reason)
            LOG.warn(_("Deleted vm %(lpar_name)s successfully. But failed "
                       "to remove the VLAN associated with vm virtual "
                       "ethernet. Instance: %(inst)s. Error:%(error)s") %
                     {'lpar_name': instance['name'],
                      'inst': instance['uuid'],
                      'error': ex})

        if notify:
            self._notify_about_instance_usage(context, instance,
                                              "shutdown.end")

    def _check_vm_state(self, context, instance):
        # vm get_info
        # Handling the HypervisorUnavailable exception forPowerKVM
        # If VM is Error: Send notification and delete it from DB
        # else send notification and mark it to Error.

        try:
            lpar_state = self.driver.get_info(instance).state
        except exception.InstanceNotFound:
            lpar_state = power_state.NOSTATE
        except exception.HypervisorUnavailable as e:
            LOG.exception(_("Failed to retrieve vm, Delete VM Failed"
                            "%s") % e)
            vm_state = instance['vm_state']
            if (vm_state != vm_states.ERROR):
                reason = (_("Unable to Delete the virtual machine,"
                            "Virtual machine will be changed to Error "
                            "State. Reason: %s. For more information "
                            "see the host log files.") % e)
                LOG.exception(_("Unable to Delete the virtual machine, "
                                "Virtual machine will be changed to Error "
                                "State. Reason: %s. For more information,"
                                "see the host log files.") % e)
                self._error_notify(instance, context, 'error', reason)
                raise e
            reason = (_("Unable to delete the virtual machine. The virtual "
                        "machine will be removed from PowerVC management. "
                        "Reason: %s. For more information, see the host log"
                        "files.") % e)
            LOG.exception(_("Unable to delete the virtual machine. The "
                            "virtual machine will be removed from PowerVC "
                            "management. Reason: %s. For more information,"
                            "see the host log files.") % e)
            self._error_notify(instance, context, 'error', reason)
            return False

        # If vm has been deleted out-of-band, the attachment information
        # exists while connector is empty. Need to be able to destroy
        # the instance for cleanup purpose.
        LOG.info(_("Destroy Virtual Server %(lpar_name)s, current vm state:"
                   " %(lpar_state)s") %
                 {'lpar_name': instance['name'],
                  'lpar_state': lpar_state})
        return True

    def _cleanup_volumes(self, context, instance_uuid, bdms, raise_exc=True):
        """ override to call community manager
        """
        base_manager._cleanup_volumes(
            self, context, instance_uuid, bdms, raise_exc=True)

    def _destroy_evacuated_instances(self, context):
        '''
            Overriden method as of the compute manager code as we needed
            to handle the condition in which we want to delete the VM.
            The method is same as of the OpenStack implemented method and
            we are changing the condition to handle out of band migration
            case.
        '''
        our_host = self.host
        filters = {'deleted': False}
        local_instances = self._get_instances_on_driver(context, filters)
        for instance in local_instances:
            # Get the system metadata for instance to see if the instance is
            # actually evacuated or not.
            sys_meta = instance.system_metadata
            evacuated_to = None
            if sys_meta and 'evacuated_to' in sys_meta:
                evacuated_to = sys_meta['evacuated_to']
            # Delete only when the instance has the metadata for dest host
            # and the current host is different than of evacuated to host.
            if evacuated_to:
                if our_host != evacuated_to and instance.host != our_host:
                    if (instance.task_state in [task_states.MIGRATING,
                                                task_states.RESIZE_MIGRATING,
                                                task_states.RESIZE_MIGRATED,
                                                task_states.RESIZE_FINISH] or
                            instance.vm_state in [vm_states.RESIZED]):
                        LOG.debug('Will not delete instance as its host ('
                                  '%(instance_host)s) is not equal to our '
                                  'host (%(our_host)s) but its task state is '
                                  '(%(task_state)s) and vm state is '
                                  '(%(vm_state)s)',
                                  {'instance_host': instance.host,
                                   'our_host': our_host,
                                   'task_state': instance.task_state,
                                   'vm_state': instance.vm_state},
                                  instance=instance)
                        continue
                    LOG.info(_("Deleting instance %(instance_name)s as its "
                               "host %(instance_host)s is not equal "
                               "to source host %(our_host)s") %
                             {'instance_name': instance.name,
                              'instance_host': instance.host,
                              'our_host': our_host},
                             instance=instance)
                    try:
                        network_info = self.network_api.get_instance_nw_info(
                            context, instance)
                        bdi = self._get_instance_block_device_info(context,
                                                                   instance)
                        destroy_disks = not (self._is_instance_storage_shared(
                                             context, instance))
                    except exception.InstanceNotFound:
                        LOG.info(_("Instance has been marked deleted, "
                                   "skipping the deletion of instance "
                                   "%(instance_name)s") %
                                 {'instance_name': instance.name})
                        destroy_disks = True
                    if not destroy_disks:
                        self.driver.destroy(context, instance,
                                            network_info,
                                            bdi, destroy_disks)
                        info = {'msg': _("Instance {instance_name} "
                                         "evacuated to host {dest_host}, "
                                         "Deleted the instance from host "
                                         "{source_host}"),
                                'instance_name': instance.name,
                                'dest_host': evacuated_to,
                                'source_host': our_host}
                        notifier = rpc.get_notifier(service='compute',
                                                    host=self.host)
                        notifier.info(context, 'compute.instance.log', info)
        driver = self.driver
        try:
            LOG.debug('Update instead of delete evacuated instances.')
            # Get all LPARs on this host that are in PowerVC DB
            drv_instances = self.driver.list_instances()
            db_instances = self._query_db_instances(context)
            # If the Driver implemented, ask it to handle Evacuated
            # Instances
            if isinstance(driver, discovery_driver.ComputeDiscoveryDriver):
                driver.handle_evacuated_instances(
                    context, db_instances, drv_instances)
        # We don't want to take down the process just because cleanup
        # failed
        except Exception:
            LOG.exception(_('exception in Evacuation Cleanup, moving on'))

    def _rebuild_default_impl(self, context, instance, image_meta,
                              injected_files, admin_password, bdms,
                              detach_block_devices, attach_block_devices,
                              network_info=None,
                              recreate=False, block_device_info=None,
                              preserve_ephemeral=False):
        if preserve_ephemeral:
            # The default code path does not support preserving ephemeral
            # partitions.
            raise exception.PreserveEphemeralNotSupported()

        if recreate:
            detach_block_devices(context, bdms)
        else:
            self._power_off_instance(context, instance, clean_shutdown=True)
            detach_block_devices(context, bdms)
            self.driver.destroy(context, instance,
                                network_info=network_info,
                                block_device_info=block_device_info)
            # Update instance to the new name given in request body of
            # Rebuild API
            metadata = instance.metadata
            new_name = metadata.pop('new_display_name', None)
            if new_name is not None:
                updates = dict(display_name=new_name, metadata=metadata)
                instance.update(updates)

        instance.task_state = task_states.REBUILD_BLOCK_DEVICE_MAPPING
        instance.save(expected_task_state=[task_states.REBUILDING])

        new_block_device_info = attach_block_devices(context, instance, bdms)

        instance.task_state = task_states.REBUILD_SPAWNING
        instance.save(
            expected_task_state=[task_states.REBUILD_BLOCK_DEVICE_MAPPING])

        with instance.mutated_migration_context():
            self.driver.spawn(context, instance, image_meta, injected_files,
                              admin_password, network_info=network_info,
                              block_device_info=new_block_device_info)

    @manager.pvc_comp_mgr_reverts_task_state
    def rebuild_instance(self, context, instance, orig_image_ref, image_ref,
                         injected_files, new_pass, orig_sys_metadata,
                         bdms, recreate, on_shared_storage=None,
                         preserve_ephemeral=False, migration=None,
                         scheduled_node=None, limits=None):
        """Destroy and re-make this instance.

        A 'rebuild' effectively purges all existing data from the system and
        remakes the VM with given 'metadata' and 'personalities'.

        :param context: `nova.RequestContext` object
        :param instance: Instance object
        :param orig_image_ref: Original image_ref before rebuild
        :param image_ref: New image_ref for rebuild
        :param injected_files: Files to inject
        :param new_pass: password to set on rebuilt instance
        :param orig_sys_metadata: instance system metadata from pre-rebuild
        :param bdms: block-device-mappings to use for rebuild
        :param recreate: True if the instance is being recreated (e.g. the
            hypervisor it was on failed) - cleanup of old state will be
            skipped.
        :param on_shared_storage: True if instance files on shared storage
        :param preserve_ephemeral: True if the default ephemeral storage
                                   partition must be preserved on rebuild
        """
        @send_ui_notification(None, notify_msg.EVACUATE_ERROR,
                              notify_msg.EVACUATE_SUCCESS, log_exception=True)
        def _recreate_rebuild(context, instance, orig_image_ref, image_ref,
                              injected_files, new_pass, orig_sys_metadata,
                              bdms, recreate, on_shared_storage,
                              preserve_ephemeral):
            if not on_shared_storage:
                msg = (_("Evacuation is not supported if storage "
                         "is not shared."))
                LOG.error(msg)
                raise Exception(msg)
            if self._instance_on_disk(instance):
                # check if image-backing file path meets
                # the storage requirements.
                instance_base_path = libvirt_utils.get_instance_path(instance)
                disk_file = instance_base_path + '/disk'
                backing_file_path = libvirt_utils.\
                    get_disk_backing_file(disk_file, basename=False)
                LOG.debug("Backing file path: %s" % backing_file_path)
                # there isn't always a backing file with PowerKVM VM.
                # We need to handle cases where there is no backing file
                if (backing_file_path and not
                        re.match("^%s" % os.path.abspath(CONF.instances_path),
                                 os.path.abspath(backing_file_path))):
                    msg = (_("Backing file %(backing_file)s of image "
                             "%(disk_file)s is missing. "
                             "Ensure that the source and target hosts "
                             "have the same mount path.") %
                           {'backing_file': backing_file_path,
                            'disk_file': disk_file})
                    LOG.error(msg)
                    raise Exception(msg)

            base_manager.\
                rebuild_instance(self, context, instance=instance,
                                 orig_image_ref=orig_image_ref,
                                 image_ref=image_ref,
                                 injected_files=injected_files,
                                 new_pass=new_pass,
                                 orig_sys_metadata=orig_sys_metadata,
                                 bdms=bdms, recreate=recreate,
                                 on_shared_storage=on_shared_storage,
                                 preserve_ephemeral=preserve_ephemeral)

            # Store the dest host for the instance to which the instance
            # evacuate
            dest_host = instance.host
            # Update the instance system metadata so that the flag is
            # used for the destroy purpose of the VM's which are evacuated.
            sys_meta = instance.system_metadata
            sys_meta['evacuated_to'] = dest_host
            self._instance_update(context, instance,
                                  system_metadata=sys_meta)
            # Log in info about the instance which is evacuated.
            LOG.info(_("Instance %(instance_name)s was remote restarted to"
                       " host %(dest_host)s") %
                     {'instance_name': instance.name,
                      'dest_host': dest_host})
            # end _recreate_rebuild

        # On a recreate (evacuate), we need to perform some additional checks
        # to ensure evacuation of the VM is possible
        if recreate:
            _recreate_rebuild(context, instance, orig_image_ref, image_ref,
                              injected_files, new_pass, orig_sys_metadata,
                              bdms, recreate, on_shared_storage,
                              preserve_ephemeral)
        else:
            # Perform rebuild action
            base_manager.\
                rebuild_instance(self, context, instance=instance,
                                 orig_image_ref=orig_image_ref,
                                 image_ref=image_ref,
                                 injected_files=injected_files,
                                 new_pass=new_pass,
                                 orig_sys_metadata=orig_sys_metadata,
                                 bdms=bdms, recreate=recreate,
                                 on_shared_storage=on_shared_storage,
                                 preserve_ephemeral=preserve_ephemeral)

    def _instance_on_disk(self, instance):
        # ensure directories exist and are writable
        instance_path = libvirt_utils.get_instance_path(instance)
        LOG.debug('Checking instance files accessibility %s', instance_path)
        shared_instance_path = os.access(instance_path, os.W_OK)
        return shared_instance_path
