#
#
# Copyright 2013, 2016 IBM Corp.
# All Rights Reserved.
# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2011 Justin Santa Barbara
# Copyright 2011 Piston Cloud Computing, Inc.
# Copyright 2012-2013 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Extends the Compute Manager to provide additional Capabilities for PowerVC
   and to override any necessary behavioral changes from the Community code.
   This code is organized into the following categories:
    1. Overriding Compute Manager Initialization
    2. Compute-related Manager Extensions
    3. Discovery-related Manager Extensions
    4. Image-related Manager Extensions
    5. Migration-related Manager Extensions
    6. Monitor-related Manager Extensions
    7. Network-related Manager Extensions
    8. Storage-related Manager Extensions
    9. Overriding the ResourceTracker and VirtAPI
"""

import copy
import functools
from eventlet import greenthread
from oslo_config import cfg
import six
import types

from cinderclient import exceptions as cinder_exception

import nova.context
from nova import exception
from nova import network
from nova import rpc
from nova import utils
from nova.compute.manager import reverts_task_state, wrap_instance_event,\
    wrap_instance_fault, errors_out_migration, object_compat
from nova.compute import power_state
from nova.compute import task_states
from nova.compute import utils as compute_utils
from nova.compute import vm_states
from nova.i18n import _LE
from nova.network import model as network_model
from nova import objects
from nova.objects import base as obj_base
from nova.objects import block_device as blkdev_obj
from nova.objects import instance as instance_obj
from nova.objects import quotas as quotas_obj
from oslo_utils import excutils
from oslo_utils import importutils
from oslo_serialization import jsonutils
from oslo_concurrency import lockutils
from oslo_log import log as logging
from oslo_service import periodic_task
from oslo_utils import timeutils
from oslo_utils import encodeutils

from powervc_nova.common import block_device_utils
from powervc_nova.compute import send_ui_notification
from powervc_nova.compute import check_host_maintenance_mode
from powervc_nova.compute import notify_messages as notify_msg
from powervc_nova.compute import manager
from powervc_nova.compute.metric_cache import MetricCache

from powervc_nova import cache_driver_data
from powervc_nova import logcall
from powervc_nova.compute import exception as pwrvc_excep
from powervc_nova.compute import common_constants
from powervc_nova.objects.network import dom as network_dom
from powervc_nova.virt import discovery_driver
from powervc_nova.virt.ibmpowervm.common import exception as pvm_exc
from powervc_nova.virt.ibmpowervm.common import migrate_utils
from powervc_nova.virt.ibmpowervm.common.constants \
    import CONNECTION_TYPE_NPIV, CONNECTION_TYPE_SSP, CONNECTION_TYPE_VSCSI
from powervc_nova.virt.ibmpowervm.hmc import event as hmcevent
from powervc_nova.objects.network import dom_kvm
from powervc_nova.network.powerkvm import agent
from powervc_nova.objects import migrate_data as mig_obj

from powervc_nova import _


LOG = logging.getLogger(__name__)


ibmpowervm_interval_opts = [
    cfg.IntOpt('ibmpowervm_activating_timeout', default=1200,
               help='Timeout (seconds) for VM in activating after deploy.'),
    cfg.IntOpt('ibmpowervm_iso_detach_timeout',
               default=3600,
               help='Timeout in seconds for removal of optical post-deploy.'),
    cfg.IntOpt('ibmpowervm_optical_cleanup_interval',
               default=1800,
               help='Interval to check for virtual optical needing removal '
               'post-deploy.'),
    cfg.IntOpt('monitor_instances_interval', default=300,
               help='Interval to Refresh the Metrics for Instances'),
    cfg.IntOpt('metrics_message_size', default=20,
               help='The size limit of a compute.instance.metrics message. '
                    'For each message, the number of instances that is '
                    'reported.  For example if there are 40 instances on a '
                    'host, there will be 2 messages sent each time the '
                    'utilization data is collected (40 / 20 = 2)'),
    cfg.IntOpt('volume_status_check_max_retries', default=300,
               help='The maximum number of volume status checks'
               ' during volume creation.'),
    cfg.IntOpt('volume_status_check_interval', default=6,
               help='Interval between status checks'
               ' during volume creation.'),
    cfg.IntOpt('cleanup_net_assn_interval', default=1200,
               help='Timeout (seconds) for cleanup network associations.'),
    cfg.BoolOpt('ibmpowervm_handle_out_of_band_migration', default=True,
                help='Handle out of band migration of instances'),
    cfg.BoolOpt('has_first_boot_occurred', default=False,
                help='Indicates if the agent has run through the first'
                     ' boot setup'),
    cfg.IntOpt('verify_mgmt_sys_max_retry_count', default=5,
               help='Retry count for verify management system'),
    cfg.IntOpt('verify_mgmt_sys_sleep_time_during_retry', default=20,
               help='Sleep time while retrying verify'
               'management system')
]

CONF = cfg.CONF
CONF.register_opts(ibmpowervm_interval_opts)
CONF.import_opt('powervm_mgr_type', 'powervc_nova.virt.ibmpowervm')
CONF.import_opt('host_storage_type', 'powervc_nova.virt.ibmpowervm')

get_notifier = functools.partial(rpc.get_notifier, service='compute')
wrap_exception = functools.partial(exception.wrap_exception,
                                   get_notifier=get_notifier)

# waiting on host task state
WAITING_ON_HOST = 'waiting_on_host'

# order of decorators should be same order  in which they
# appear if applied statically
CUSTOM_DECORATORS = {
    'attach_volume_check_VM_state':
    [manager.declare_decor(
        'powervc_nova.compute.check_host_maintenance_mode',
        common_constants.OPER_AT_VOL)],
    '_destroy_evacuated_instances':
    [manager.declare_decor('powervc_nova.compute.send_ui_notification',
                           notify_msg.SOURCE_EVACUATED_INST_CLEANUP_ERROR,
                           None, log_exception=True)]
}


class PowerVCKvmHmcComputeManager(manager.PowerVCComputeManager):

    """Extends the base Compute Manager class with PowerVC capabilities"""

    @classmethod
    def get_decorators(cls, decorators):
        new_decorators = copy.deepcopy(CUSTOM_DECORATORS)
        new_decorators.update(decorators)
        return super(PowerVCKvmHmcComputeManager, cls).\
            get_decorators(new_decorators)

    def __init__(self, compute_driver=None, *args, **kwargs):
        """Constructor for the PowerVC extension to the Compute Manager"""
        super(PowerVCKvmHmcComputeManager, self).__init__(
            compute_driver=compute_driver, *args, **kwargs)
        self.vm_cpu_util_cache = MetricCache()

    #######################################################
    #           Override Initialization Methods           #
    #######################################################
    @cache_driver_data(['get_lpars_feed'])
    def init_host(self):
        """Override the Init Host to fix Base Compute Manager issues"""
        super(PowerVCKvmHmcComputeManager, self).init_host()

    #######################################################
    #         Compute-related Overrides/Extensions        #
    #######################################################

    def _build_and_run_instance(self, context, instance, image, injected_files,
                                admin_password, requested_networks,
                                security_groups, block_device_mapping, node,
                                limits, filter_properties):

        # Monkey patch method to intercept instance.save() calls
        def proxy_save(self, *args, **kwargs):
            # If super() is marking it active, let's change that!
            if self.vm_state == vm_states.ACTIVE and self.task_state is None:
                is_kvm = False
                if CONF.powervm_mgr_type is None:
                    is_kvm = (CONF.libvirt.virt_type == 'kvm')

                if not is_kvm:
                    # For PowerVM, set task states differently
                    # ISO deploys will go to stopped with no task state
                    # Deploys with do not boot option set to true will go
                    # to stopped with no task state
                    if instance['metadata'].get(
                            'do_not_boot') is not None and (
                            instance['metadata'].get(
                            'do_not_boot').upper() == 'TRUE'):
                        instance.vm_state = vm_states.STOPPED
                        instance.task_state = None

            # Call the original instance method
            self.orig_save(*args, **kwargs)

        # Monkey patch the object to call our version of instance.save()
        # Copy the Class save() and bind it to the instance object
        instance.orig_save = types.MethodType(instance_obj.Instance.save,
                                              instance)

        # Bind our version of the save() to the instance object
        instance.save = types.MethodType(proxy_save, instance)
        """Spawn an instance with error logging and update its power state.
           Override to send notification.
        """
        super(PowerVCKvmHmcComputeManager, self)._build_and_run_instance(
            context, instance, image,
            injected_files, admin_password,
            requested_networks, security_groups,
            block_device_mapping, node, limits,
            filter_properties)

        # Undo the monkey patch.  The methods are already bound to the instance
        instance.save = instance.orig_save

    def _do_build_and_run_instance(self, context, instance, image,
                                   request_spec, filter_properties,
                                   admin_password, injected_files,
                                   requested_networks, security_groups,
                                   block_device_mapping,
                                   node=None, limits=None):
        # OSEE 192018
        # Until it's fixed in Juno, we will override this method to store
        # context in strong store.
        context.update_store()
        super(PowerVCKvmHmcComputeManager, self)._do_build_and_run_instance(
            context, instance, image,
            request_spec, filter_properties, admin_password, injected_files,
            requested_networks, security_groups, block_device_mapping,
            node=node, limits=limits)

        LOG.info("Instance Metadata : %s" % instance.system_metadata)
        try:
            LOG.info(_("Refreshing statistics for the  %s instance.")
                     % instance.name)
            db_inst = self._query_db_instances(context, [instance])
            self.driver.inventory_instances(context, db_inst)
        except Exception as exc:
            LOG.warn(_('Failed to refresh instance inventory '))
            LOG.exception(exc)

    def _destroy_evacuated_instances(self, context):
        """
            Override Base Compute Manager to update evacuated instances.
            If the hypervisor is KVM, read the instance metadata and
            delete the instance from the source if the instance host
            is different than the source host and instance system
            metadata contains the 'evacuated_to'.

            In the case of the HMC driver, call back the driver method to
            handle evacuated instance.

            The method is overriden for sending notifications as well to make
            sure user is notified with the proper messages in case of cleanup
            failures from the source host or in case of successful cleanup.
        """
        driver = self.driver
        try:
            LOG.debug('Update instead of delete evacuated instances.')
            # Get all LPARs on this host that are in PowerVC DB
            drv_instances = self.driver.list_instances()
            db_instances = self._query_db_instances(context)
            # If the Driver implemented, ask it to handle Evacuated
            # Instances
            if isinstance(driver, discovery_driver.ComputeDiscoveryDriver):
                driver.handle_evacuated_instances(
                    context, db_instances, drv_instances)
        # We don't want to take down the process just because cleanup
        # failed
        except Exception:
            LOG.exception(_('exception in Evacuation Cleanup, moving on'))

    def handle_lifecycle_event(self, event):
        '''
        Override to handle new events used in HMC environments.
        For PowerKVM, since state of the VM is overridden in the
        case of SUSPEND, we need to call get_info() to get
        the real state rather than relying on the event data.
        '''
        context = nova.context.get_admin_context()
        instance = instance_obj.Instance.get_by_uuid(
            context, event.get_instance_uuid())

        # Handle the event issued when the host of an instance goes down.
        if event.get_transition() == hmcevent.EVENT_HOST_DOWN:
            host_power_state = event.get_host_state()
            # Update states only if there is not a current task on the instance
            if instance.task_state is None:
                # If the host was powered down, then the instance will have
                # been powered off, so update the vm_state in the database
                if ((host_power_state in ['power off']) and
                        instance.vm_state == vm_states.ACTIVE):
                    LOG.info(_("Host down event on VM %(uuid)s, because state"
                               " of the host is not operational. Updating VM"
                               " state to %(vm_state)s and task state to"
                               " %(task_state)s.") %
                             {'uuid': event.get_instance_uuid(),
                              'vm_state': vm_states.STOPPED,
                              'task_state': WAITING_ON_HOST})
                    self._instance_update(context, instance,
                                          vm_state=vm_states.STOPPED,
                                          task_state=WAITING_ON_HOST)
                elif host_power_state not in ['no connection']:
                    LOG.info(_("Host down event on VM %(uuid)s, because state"
                               "of the host is not operational. Updating task"
                               " state of VM to %(task_state)s.") %
                             {'uuid': event.get_instance_uuid(),
                              'task_state': WAITING_ON_HOST})
                    self._instance_update(context, instance,
                                          task_state=WAITING_ON_HOST)
                else:
                    LOG.info("Host is in power state %s and cannot "
                             "be reached" % host_power_state)
            return
        # Handle the event to clear the waiting on host task state
        elif event.get_transition() == hmcevent.EVENT_CLEAR_TASK:
            if instance.task_state == WAITING_ON_HOST:
                self._instance_update(context, instance,
                                      task_state=None)
            return

        inst_info = self.driver.get_info(instance)
        vm_power_state = inst_info.state

        LOG.info(_("Life cycle event %(state)d on VM %(uuid)s. State is"
                   " %(vm_state)s.") %
                 {'state': event.get_transition(),
                  'uuid': event.get_instance_uuid(),
                  'vm_state': vm_power_state})

        self._sync_instance_power_state(context,
                                        instance,
                                        vm_power_state)

    def _running_deleted_instances(self, context):
        """Override the running deleted to filter unmanaged VM's"""
        insts = super(PowerVCKvmHmcComputeManager,
                      self)._running_deleted_instances(context)
        # VM's that are unmanaged by PowerVC are being marked as soft deleted
        return [i for i in insts if i['vm_state'] != vm_states.SOFT_DELETED]

    @logcall
    def get_host_ovs(self, context):
        retVal = None
        try:
            if hasattr(self.driver, 'get_host_ovs'):
                LOG.debug('Retrieving host ovs data using Driver...')
                retVal = self.driver.get_host_ovs(context)
        except Exception as exc:
            """
            NOTE: Since this is the interface boundary between the agent
            and server we will not raise the exception and will just
            log it on the host side
            """
            LOG.exception(exc)

            error_obj = dom_kvm.Error("%s" % exc)
            ovs_dom = dom_kvm.HostOVSNetworkConfig(CONF.get("host"))

            if retVal is None:
                ovs_dom.error_list = [error_obj]

            else:
                ovs_dom.from_dict(retVal)
                ovs_dom.error_list.append(error_obj)

            retVal = ovs_dom.to_dict()

        return retVal

    @logcall
    def update_host_ovs(self, context, dom, force_flag, rollback):
        retVal = {}
        try:
            if hasattr(self.driver, 'update_host_ovs'):
                LOG.debug('Updating host ovs data using Driver...')
                retVal = self.driver.update_host_ovs(context,
                                                     dom,
                                                     force_flag,
                                                     rollback)
        except Exception as exc:
            """
            NOTE: Since this is the interface boundary between the agent
            and server we will not raise the exception and will just
            log it on the host side
            """
            LOG.error(exc)

            if agent.ERRORS_KEY not in retVal:
                retVal[agent.ERRORS_KEY] = [{'message': six.text_type(exc)}]
            else:
                retVal[agent.ERRORS_KEY].append(
                    {'message': six.text_type(exc)})

        return retVal

    @logcall
    def create_host_ovs(self, context, dom, force_flag, rollback):
        retVal = {}
        try:
            if hasattr(self.driver, 'create_host_ovs'):
                LOG.debug('Creating host ovs using Driver...')
                retVal = self.driver.create_host_ovs(context,
                                                     dom,
                                                     force_flag,
                                                     rollback)
        except Exception as exc:
            """
            NOTE: Since this is the interface boundary between the agent
            and server we will not raise the exception and will just
            log it on the host side
            """
            LOG.error(exc)

            if agent.ERRORS_KEY not in retVal:
                retVal[agent.ERRORS_KEY] = [{'message': six.text_type(exc)}]
            else:
                retVal[agent.ERRORS_KEY].append(
                    {'message': six.text_type(exc)})

        return retVal

    @logcall
    def delete_host_ovs(self, context, ovs_list, force_flag, rollback):
        retVal = {}
        try:
            if hasattr(self.driver, 'delete_host_ovs'):
                LOG.debug('Deleting host ovs using Driver...')
                retVal = self.driver.delete_host_ovs(context,
                                                     ovs_list,
                                                     force_flag,
                                                     rollback)
        except Exception as exc:
            """
            NOTE: Since this is the interface boundary between the agent
            and server we will not raise the exception and will just
            log it on the host side
            """
            LOG.error(exc)

            if agent.ERRORS_KEY not in retVal:
                retVal[agent.ERRORS_KEY] = [{'message': six.text_type(exc)}]
            else:
                retVal[agent.ERRORS_KEY].append(
                    {'message': six.text_type(exc)})

        return retVal

    #######################################################
    #          Image-related Overrides/Extensions         #
    #######################################################
    @periodic_task.periodic_task(
        spacing=CONF.ibmpowervm_optical_cleanup_interval)
    def _check_instance_vopt_removal(self, context):
        """Clean up virtual optical devices after 1 hour."""
        if CONF.powervm_mgr_type and CONF.powervm_mgr_type.lower() == 'hmc':
            # only done for HMC
            timeout = CONF.ibmpowervm_iso_detach_timeout
            if timeout < 0:
                return

            insts = instance_obj.InstanceList.get_by_host(context, self.host)
            # Loop through all the Instances to find ones older than 1 hour
            old_instances = []
            for instance in insts:
                if timeutils.is_older_than(instance['created_at'],
                                           timeout):
                    old_instances.append(instance['uuid'])
            self.driver.vopt_cleanup(old_instances)

    def _await_volume_create(self, context, vol_id):
        """
        Wait for volume create to complete. This routine needs to
        wait until the driver moves the volume out of the creating
        state. Eventually it will move to available or error.
        For SSP Thick or EMC clones this can be double digit mins.
        :param context: operation context
        :param vol_id: volume id
        """
        while True:
            volume = self.volume_api.get(context, vol_id)
            volume_status = volume['status']
            # Would we every need downloading? Or other status here?
            if volume_status not in ['creating']:
                if volume_status != 'available':
                    LOG.warn(_("Volume id: %(vol)s finished being created"
                               " with status: '%(status)s'") %
                             {'vol': vol_id,
                              'status': volume_status})
                return
            # for now sleep based on the bd retry interval
            greenthread.sleep(CONF.block_device_allocate_retries_interval)

    def _clean_bdm(self, bdm):
        # sanitize the password out of an iscsi bdm
        c_info = bdm.get("connection_info")
        if c_info:
            if isinstance(c_info, basestring):
                c_info_d = jsonutils.loads(c_info)
            elif isinstance(c_info, dict):
                c_info_d = c_info
            else:
                return bdm
            sanitized = self._clean_conn_info(c_info_d)
            if c_info_d == sanitized:
                return bdm
            bdm_copy = copy.deepcopy(bdm)
            bdm_copy["connection_info"] = sanitized
            return bdm_copy
        return bdm

    def _clean_conn_info(self, connection_info):
        # cleanup the password out of an iscsi connector
        if (connection_info['driver_volume_type'] != 'iscsi'):
            return connection_info
        if 'data' not in connection_info:
            return connection_info
        if 'auth_password' not in connection_info['data']:
            return connection_info
        conn_info = copy.deepcopy(connection_info)
        conn_info['data']['auth_password'] = "**********"
        return conn_info

    #######################################################
    #        Migration-related Overrides/Extensions       #
    #######################################################
    @wrap_exception()
    @reverts_task_state
    @wrap_instance_event
    @errors_out_migration
    @wrap_instance_fault
    @send_ui_notification(None,
                          notify_msg.RESIZE_ERROR, None,
                          log_exception=True)
    @check_host_maintenance_mode(None, common_constants.OPER_RESIZE)
    def resize_instance(self, context, instance, image,
                        reservations, migration, instance_type,
                        clean_shutdown):
        """Starts the migration of a running instance to another host."""

        quotas = quotas_obj.Quotas.from_reservations(context,
                                                     reservations,
                                                     instance=instance)
        with self._error_out_instance_on_exception(context, instance,
                                                   quotas=quotas):
            if not instance_type:
                instance_type = objects.Flavor.get_by_id(
                    context, migration['new_instance_type_id'])

            network_info = self.network_api.get_instance_nw_info(
                context, instance)

            migration.status = 'migrating'
            with migration.obj_as_admin():
                migration.save()
            instance.task_state = task_states.RESIZE_MIGRATING
            instance.save(expected_task_state=task_states.RESIZE_PREP)

            self._notify_about_instance_usage(
                context, instance, "resize.start", network_info=network_info)

            bdms = objects.BlockDeviceMappingList.get_by_instance_uuid(
                context, instance.uuid)
            block_device_info = self._get_instance_block_device_info(
                context, instance, bdms=bdms)

            timeout, retry_interval = self.\
                _get_power_off_values(context, instance, clean_shutdown)
            disk_info = self.driver.migrate_disk_and_power_off(
                context, instance, migration.dest_host,
                instance_type, network_info,
                block_device_info,
                timeout, retry_interval)

            migration_p = obj_base.obj_to_primitive(migration)
            instance_p = obj_base.obj_to_primitive(instance)
            self.network_api.migrate_instance_start(context,
                                                    instance_p,
                                                    migration_p)

            migration.status = 'post-migrating'
            with migration.obj_as_admin():
                migration.save()

            instance.host = migration.dest_compute
            instance.node = migration.dest_node
            instance.task_state = task_states.RESIZE_MIGRATED
            instance.save(expected_task_state=task_states.RESIZE_MIGRATING)

            self.compute_rpcapi.finish_resize(context, instance,
                                              migration, image, disk_info,
                                              migration.dest_compute,
                                              reservations=quotas.reservations)

            self._notify_about_instance_usage(context, instance, "resize.end",
                                              network_info=network_info)
            self.instance_events.clear_events_for_instance(instance)

    @wrap_exception()
    @send_ui_notification(None,
                          notify_msg.RESIZE_ERROR, notify_msg.RESIZE_SUCCESS,
                          log_exception=True)
    def confirm_resize(self, context, instance, reservations, migration):

        super(PowerVCKvmHmcComputeManager, self). \
            confirm_resize(context, instance, reservations, migration)
        try:
            LOG.info(_("Refreshing statistics for the  %s instance.")
                     % instance.name)
            db_inst = self._query_db_instances(context, [instance])
            self.driver.inventory_instances(context, db_inst)
        except Exception as exc:
            LOG.warn(_('Failed to refresh instance invetory '))
            LOG.exception(exc)

    @wrap_exception()
    def pre_live_migration(self, context, instance, block_migration=False,
                           disk=None, migrate_data=None):

        # Issue 5649: Onboarded lpars do not have a fixed IP assigned causing
        # pre_live_migration to fail.  Rather than wait for a community fix
        # we'll reimplement the manager level pre_live_migration without the
        # fixed IP check.
        try:
            LOG.info('pre_live_migration data is %s', migrate_data)

            if hasattr(migrate_data, 'refresh_conn_info'):
                refresh_conn_info = migrate_data.refresh_conn_info
            else:
                refresh_conn_info = False

            block_device_info = self._get_instance_block_device_info(
                context, instance, refresh_conn_info=refresh_conn_info)
            network_info = self.network_api.get_instance_nw_info(
                context, instance)
        except Exception:
            with excutils.save_and_reraise_exception():
                message = _("An internal error occurred. For more information"
                            ", see log file for host %s." % CONF.host)
                tgt_name = migrate_data.dest_sys_name
                migrate_utils.send_migration_failure_notification(context,
                                                                  instance,
                                                                  tgt_name,
                                                                  message)
        self._notify_about_instance_usage(context, instance,
                                          "live_migration.pre.start",
                                          network_info=network_info)

        pre_live_migration_data = self.driver.pre_live_migration(
            context, instance, block_device_info,
            network_info, disk, migrate_data)

        # NOTE(tr3buchet): setup networks on destination host
        self.network_api.setup_networks_on_host(context, instance,
                                                self.host)

        self._notify_about_instance_usage(context, instance,
                                          "live_migration.pre.end",
                                          network_info=network_info)
        return pre_live_migration_data

    @wrap_exception()
    def live_migration(self, context, dest, instance,
                       block_migration, migration, migrate_data):
        # BDMS will be updated by pre_live_migration() if there are
        # attached cinder volumes. The info in the BDMS may be
        # changed. Save it for roll-back during exception.

        old_bdms_src = blkdev_obj.BlockDeviceMappingList.get_by_instance_uuid(
            context, instance['uuid'])
        migrate_data.pre_migrate_bdms = copy.deepcopy(old_bdms_src)
        super(PowerVCKvmHmcComputeManager, self).live_migration(
            context, dest, instance, block_migration, migration, migrate_data)

    @wrap_exception()
    @reverts_task_state
    def post_live_migration_at_destination(self, context, instance,
                                           block_migration=False,
                                           migrate_data=None):
        """Post operations for live migration .

        :param context: security context
        :param instance: Instance dict
        :param block_migration: if true, prepare for block migration
        :param migrate_data: implementation specific data dictionary
        """
        # Notify post live at destination has started
        LOG.info(_('Post operation of migration started on destination host'),
                 instance=instance)
        # clean up the migration information where we track whether
        # there is a migration from host A to B, so we can prevent
        # migration from B to A.
        try:
            network_info = self.network_api.get_instance_nw_info(context,
                                                                 instance)
            self._notify_about_instance_usage(context, instance,
                                              "live_migration.post.dest.start",
                                              network_info=network_info)

            # NOTE(tr3buchet): setup networks on destination host
            #                  this is called a second time because
            #                  multi_host does not create the bridge in
            #                  plug_vifs
            self.network_api.setup_networks_on_host(context, instance,
                                                    self.host)

            migration = {'source_compute': instance['host'],
                         'dest_compute': self.host, }
            self.network_api.migrate_instance_finish(context, instance,
                                                     migration)
        except Exception:
            with excutils.save_and_reraise_exception():
                self.driver.migration_cleanup(context, instance)

        try:
            # Call driver
            block_device_info = self._get_instance_block_device_info(
                context, instance)
        except Exception:
            with excutils.save_and_reraise_exception():
                self.driver.migration_cleanup(context, instance)
                message = _("An internal error occurred. For more information"
                            ", see log file for host %s." % CONF.host)
                tgt_name = migrate_data.dest_sys_name
                migrate_utils.send_migration_failure_notification(context,
                                                                  instance,
                                                                  tgt_name,
                                                                  message)

        self.driver.post_live_migration_at_destination(context, instance,
                                                       network_info,
                                                       block_migration,
                                                       block_device_info,
                                                       migrate_data)

        # Restore instance state
        current_power_state = self._get_power_state(context, instance)
        node_name = None
        try:
            compute_node = self._get_compute_info(context, self.host)
            node_name = compute_node['hypervisor_hostname']
        except exception.NotFound:
            LOG.exception(_('Failed to get compute_info for %s') % self.host)
        finally:
            # Do not check for expected task state following a successful
            # migration.  This leads to the health state following to go
            # critical when something other than migration causes the task
            # state to clear. Also, don't clear the task state here, as
            # it will be cleared by a thread kicked off by
            # driver.post_live_migration_at_destination when RMC becomes active
            # use instance.save instead since _instance_update return dict,
            # it will cause error in _notify_about_instance_usage
            instance.host = self.host
            instance.power_state = current_power_state
            instance.vm_state = vm_states.ACTIVE
            instance.task_state = None
            instance.node = node_name
            instance.save()

        # NOTE(vish): this is necessary to update dhcp
        self.network_api.setup_networks_on_host(context, instance, self.host)
        self._notify_about_instance_usage(context, instance,
                                          "live_migration.post.dest.end",
                                          network_info=network_info)

    @wrap_exception()
    @reverts_task_state
    def rollback_live_migration_at_destination(self, context, instance,
                                               destroy_disks, migrate_data):

        # We're not calling 'super' so we still need to send notifications
        network_info = self.network_api.get_instance_nw_info(context, instance)
        self._notify_about_instance_usage(
            context, instance, "live_migration.rollback.dest.start",
            network_info=network_info)

        # NOTE(tr3buchet): tear down networks on destination host
        self.network_api.setup_networks_on_host(context, instance,
                                                self.host, teardown=True)

        # NOTE(vish): The mapping is passed in so the driver can disconnect
        #             from remote volumes if necessary
        block_device_info = self._get_instance_block_device_info(
            context, instance)

        # Don't call the parent roll-back, we define our own driver level
        self.driver.rollback_live_migration_at_destination(context, instance,
                                                           network_info,
                                                           block_device_info,
                                                           destroy_disks,
                                                           migrate_data)

        # unplug VLAN at Destination during rollback
        self.driver.unplug_vifs(instance, network_info)

        self._notify_about_instance_usage(
            context, instance, "live_migration.rollback.dest.end",
            network_info=network_info)

    @wrap_exception()
    @reverts_task_state
    @wrap_instance_fault
    def _post_live_migration(self, ctxt, instance,
                             dest, block_migration=False, migrate_data=None):
        """Post operations for live migration on source host.
        We override the manager level code as the behavior for
        IVM and HMC needs to be different

        This method is called from live_migration
        and mainly updating database record.

        :param ctxt: security context
        :param instance: Instance object
        :param dest: destination host
        :param block_migration: if true, prepare for block migration
        :param migrate_data: if not None, it is a dict which has data
        required for live migration without shared storage

        """
        try:
            # Notify the post migration has started
            network_info = self.network_api.get_instance_nw_info(
                ctxt, instance)
            self._notify_about_instance_usage(ctxt, instance,
                                              "live_migration._post.start",
                                              network_info=network_info)
            LOG.info(_('_post_live_migration() is started..'),
                     instance=instance)
            LOG.debug('Migration data = %s' % migrate_data)

            # Get the block device info
            bdms = blkdev_obj.BlockDeviceMappingList.get_by_instance_uuid(
                ctxt, instance['uuid'])
            block_device_info = self._get_instance_block_device_info(
                ctxt, instance, bdms=bdms)
        except Exception:
            with excutils.save_and_reraise_exception():
                self.driver.migration_cleanup(ctxt, instance, source=True)
                message = _("An internal error occurred. For more information"
                            ", see log file for host %s." % CONF.host)
                tgt_name = migrate_data.dest_sys_name
                migrate_utils.send_migration_failure_notification(ctxt,
                                                                  instance,
                                                                  tgt_name,
                                                                  message)

        self.driver.post_live_migration(ctxt, instance,
                                        block_device_info, migrate_data)
        # Detach volumes for IVM only.
        # For NPIV/SSP storage, supported only by HMC managed systems,
        # the disk attachment is maintained by powervm during LPM.
        # From openstack's point of view, the attachment is the same no matter
        # which host it is running on. For example, in NPIV scenario, the disk
        # is attached to a virtualized fibre channel adapter assigned to the
        # instance. The attached disk is only visible to the instance.
        # This level of virtualization hides all the implementation details for
        # LPM, and it stays unchanged during the life cycle of the instance
        # regardless which host it is running on.

        if (hasattr(migrate_data, 'refresh_conn_info') and
                migrate_data.refresh_conn_info):
            LOG.debug("Terminating the storage connections for migration.")
            connector = self.driver.get_volume_connector(instance)
            for bdm in bdms:
                # NOTE(vish): We don't want to actually mark the volume
                #             detached, or delete the bdm, just remove the
                #             connection from this host.

                # remove the volume connection without detaching
                # from hypervisor because the instance is not
                # running anymore on the current host
                if bdm.is_volume:
                    self.volume_api.terminate_connection(ctxt, bdm.volume_id,
                                                         connector)
        else:
            LOG.debug("Skipping storage operations for migration.")

        # Releasing vlan.
        # (not necessary in current implementation?)
        migration = {'source_compute': self.host,
                     'dest_compute': dest, }
        self.network_api.migrate_instance_start(ctxt, instance, migration)

        # Define domain at destination host, without doing it,
        # pause/suspend/terminate do not work.
        self.compute_rpcapi.post_live_migration_at_destination(
            ctxt, instance, block_migration, dest, migrate_data)

        # NOTE(tr3buchet): tear down networks on source host
        self.driver.unplug_vifs(instance, network_info)
        self.network_api.setup_networks_on_host(ctxt, instance,
                                                self.host, teardown=True)
        self.instance_events.clear_events_for_instance(instance)
        self._update_scheduler_instance_info(ctxt, instance)

        self._notify_about_instance_usage(ctxt, instance,
                                          "live_migration._post.end",
                                          network_info=network_info)
        LOG.info(_('Migrating instance to %s finished successfully.'),
                 dest, instance=instance)

    @wrap_exception()
    @reverts_task_state
    def _rollback_live_migration(self, context, instance, dest,
                                 block_migration, migrate_data=None,
                                 migration_status='error'):
        """Rollback operations for live migration on source host.
        We override the manager level code as the behavior for
        IVM and HMC needs to be different
        """

        # If pre_live_migration or live_migration fails
        # we need to clean up the migration dict on the source.
        self.driver.migration_cleanup(context, instance, source=True)

        # Notify rollback has started
        LOG.debug('Rollback starting on source host for instance %s. '
                  'Migration data = %s' %
                  (encodeutils.safe_decode(instance['display_name']),
                   migrate_data.to_legacy_dict()))
        self._notify_about_instance_usage(context, instance,
                                          "live_migration._rollback.start")
        # Set the state back to Active
        instance.vm_state = vm_states.ACTIVE
        instance.task_state = None
        instance.save(expected_task_state=[task_states.MIGRATING])

        # NOTE(tr3buchet): setup networks on source host (really it's re-setup)
        self.network_api.setup_networks_on_host(context, instance, self.host)

        # Detach volumes for IVM only.

        if (hasattr(migrate_data, 'refresh_conn_info') and
                migrate_data.refresh_conn_info):
            LOG.debug("Removing volume connections for migration.")
            bdms = blkdev_obj.BlockDeviceMappingList.get_by_instance_uuid(
                context, instance['uuid'])
            for bdm in bdms:
                if bdm.is_volume:
                    volume_id = bdm.volume_id
                    try:
                        self.compute_rpcapi.remove_volume_connection(context,
                                                                     instance,
                                                                     volume_id,
                                                                     dest)
                    except Exception as ex:
                        LOG.error(_("Failed to remove volume %(volume_id)s "
                                    "connection  due to error: %(error)s") %
                                  {'volume_id': volume_id, 'error': ex})
        else:
            LOG.debug("Skipping removing volume connection for migration.")

        # Rollback on destination
        self.compute_rpcapi.rollback_live_migration_at_destination(
            context, instance, dest, migrate_data=migrate_data)

        # Block device mapping for attached volumes may have been changed
        # by pre_live_migration(). Need to restore it after roll-back.
        self._restore_pre_migrate_bdms(context, instance['uuid'], migrate_data)

        # Notify we're done
        LOG.debug('Rollback ending on source host for instance %s.' %
                  instance['display_name'])
        self._notify_about_instance_usage(context, instance,
                                          "live_migration._rollback.end")

    #######################################################
    #         Monitor-related Overrides/Extensions        #
    #######################################################
    @periodic_task.periodic_task(spacing=CONF.monitor_instances_interval)
    def monitor_instances(self, context, instances=None):
        """Periodic Task to Refresh Metrics for the Instances"""
        driver, context = (self.driver, context.elevated())
        LOG.debug('Invoking Method to Refresh Instance Metrics - Begin')
        try:
            drv_insts = self.driver.list_instances()
            name_map = dict([(inst, '') for inst in drv_insts])
            # Retrieve All of the Instances from the Nova DB for the Given Host
            db_insts = self._query_db_instances(context, instances)
            db_insts = [inst for inst in db_insts if inst['name'] in name_map]
            # If this is a Discovery Driver, then let it Gather Metrics
            if isinstance(driver, discovery_driver.ComputeDiscoveryDriver):
                # It is only worth Gathering Metrics if there are Instances
                if len(db_insts) > 0:
                    metric_data = driver.monitor_instances(context, db_insts)
                    # Update metrics in CPU utilization metrics cache.
                    self.vm_cpu_util_cache.add_value(metric_data,
                                                     'cpu_utilization')
                    util_dict = self.vm_cpu_util_cache.get_average()
                    for instance in db_insts:
                        cpu_util = util_dict.get(instance['uuid'])
                        power_specs = instance.get('power_specs', dict())
                        # Update only if cpu_util is different.
                        if cpu_util != power_specs.get('cpu_util'):
                            instance.power_specs = dict(cpu_util=cpu_util)
                            instance.save()
        except Exception as exc:
            LOG.warn(_('Error refreshing Instance Metrics'))
            LOG.exception(exc)
        LOG.debug('Invoking Method to Refresh Instance Metrics - End')

    def _send_metric_notifications_host(self, context, metric_data):
        """Collects with host monitor data (i.e. CPU utilization)"""
        payload, payloads = (list(), list())
        # If there wasn't any Metric data, then log a debug statement
        if len(metric_data) <= 0:
            LOG.debug('No CPU data found: ' + str(metric_data))
        # Loop through the Metric Data on a per host level
        for i in range(len(metric_data)):
            data = metric_data[i]
            payload.append({'host_id': data['uuid'],
                            'cpu_utilization': data['cpu_utilization']})
            # If we have hit the Message Size limit, add a new chunk
            if (i > 1) and (i % CONF.metrics_message_size == 0):
                payloads.append(payload)
                payload = []
            # If this is the last one, then we want to add the last chunk
            elif i == (len(metric_data) - 1):
                payloads.append(payload)
        # Loop through each of the Payload Chunks, sending a Notification
        for p in payloads:
            notifier = rpc.get_notifier(service='compute', host=self.host)
            notifier.info(context, 'compute.host.metrics', {'data': p})

    def _send_metric_notifications(self, context, metric_data):
        """Collects with LPAR monitor data (i.e. CPU utilization)"""
        payload, payloads = (list(), list())
        # If there wasn't any Metric data, then log a debug statement
        if len(metric_data) <= 0:
            LOG.debug('No CPU data found: ' + str(metric_data))
        # Loop through the Metric Data on a per-Instance level
        for i in range(len(metric_data)):
            inst_data = metric_data[i]
            payload.append({'instance_id': inst_data['uuid'],
                            'cpu_utilization': inst_data['cpu_utilization']})
            # If we have hit the Message Size limit, add a new chunk
            if (i > 1) and (i % CONF.metrics_message_size == 0):
                payloads.append(payload)
                payload = []
            # If this is the last one, then we want to add the last chunk
            elif i == (len(metric_data) - 1):
                payloads.append(payload)
        # Loop through each of the Payload Chunks, sending a Notification
        for p in payloads:
            notifier = rpc.get_notifier(service='compute', host=self.host)
            notifier.info(context, 'compute.instance.metrics', {'data': p})

    #######################################################
    #         Network-related Overrides/Extensions        #
    #######################################################

    def _get_super(self):
        """
        Helper for UT since the mock library can't properly return the super
        class.
        """
        return super(PowerVCKvmHmcComputeManager, self)

    @wrap_exception()
    def external_instance_event(self, context, instances, events):
        """
        Wrapper until 29618 can be updated in the community
        """
        # Do the original...
        self._get_super().external_instance_event(context, instances, events)

        # Fire off an instance update for the appropriate objects
        for event in events:
            instance = [inst for inst in instances
                        if inst.uuid == event.instance_uuid][0]
            if event.name == 'network-changed':
                self._instance_update(context, instance,
                                      updated_at=timeutils.utcnow())

    def attach_interface(self, context, instance, network_id, port_id,
                         requested_ip):
        """
        Subclassing simply to throw the instance update for Bumblebee
        """
        # call the parent
        ret = self._get_super().attach_interface(context, instance,
                                                 network_id, port_id,
                                                 requested_ip)

        self._instance_update(context, instance,
                              updated_at=timeutils.utcnow())
        return ret

    def detach_interface(self, context, instance, port_id):
        """
        Subclassing simply to throw the instance update for Bumblebee
        """
        # call the parent
        ret = self._get_super().detach_interface(context, instance, port_id)

        self._instance_update(context, instance,
                              updated_at=timeutils.utcnow())
        return ret

    def change_vif_sea(self, context, data):
        """
        This API calls the plug method on the compute driver to plug a VLAN
        on a VEA.
        :param context: Context Object
        :param data: The dictionary object containing information to call
        the plug  operation, sent by the API service.
        :return: Calls the compute driver and returns the plug call back.
        """
        greenthread.spawn(self._change_vif_sea, context, data)
        return True

    @logcall
    @lockutils.synchronized('change_powervm_vifs', 'nova-change-vif-powervm-')
    def _change_vif_sea(self, context, data):
        """
        This API calls the plug method on the compute driver to plug a VLAN
        on a VEA.  This is the internal call that should be done on the
        greenthread.

        :param context: Context Object
        :param data: The dictionary object containing information to call
        the plug  operation, sent by the API service.
        :return: Calls the compute driver and returns the plug call back.
        """
        # Obtain data from the input.  Start with common data.
        vlan = data['vlan']

        # Move on to the 'new' data
        # Note that the force=True is only used in the rollback, but
        # we set it here so that we don't have to rebuild.
        new_sea = data['new_sea']
        new_lpar = data['new_lpar_id']
        net_id = {'id': data['net_id']}
        new_vif = network_model.VIF('fake_id', 'fake_address', net_id,
                                    'fake_dev_name', sea_lpar_id=new_lpar,
                                    sea_name=new_sea, force=True,
                                    vlan=vlan)

        # Build a list of network ids that will be updated when complete
        net_ids_to_update = [data['net_id']]
        net_ids_to_update.extend(data.get('peers', []))

        # Build the 'old' data...what is 'currently' plugged in and find all
        # other vifs that will be affected by this change.
        old_sea = ''
        old_lpar = ''
        old_vif_list = []
        na_and_peers_tuple_list = []
        netafact = network_dom.NetworkAssociationFactory.get_factory()
        for net_id_to_update in net_ids_to_update:
            net_assn = netafact.find_network_assoc(
                context, self.host, net_id_to_update)
            if(not net_assn or net_assn is None or not net_assn.sea or
               net_assn.sea is None):
                continue
            if net_id['id'] == net_id_to_update:
                old_sea = net_assn.sea.name
                old_lpar = net_assn.sea.vioserver.lpar_id
            na_and_peers_tuple_list.append((net_assn.sea.name,
                                            [net_id_to_update]))

            # It's possible to find duplicates, so make sure we aren't already
            # replugging this vif.
            net_assn_lpar = net_assn.sea.vioserver.lpar_id
            already_listed = False
            for old_vif_entry in old_vif_list:
                if(old_vif_entry['meta']['sea_name'] == net_assn.sea.name or
                   old_vif_entry['network']['id'] == net_id_to_update):
                    # We've already got this vif on the list, skip it
                    already_listed = True
            if not already_listed:
                cur_vif = network_model.VIF('fake_id', 'fake_address',
                                            {'id': net_id_to_update},
                                            'fake_dev_name',
                                            sea_lpar_id=net_assn_lpar,
                                            sea_name=net_assn.sea.name,
                                            force=True,
                                            vlan=vlan)
                old_vif_list.append(cur_vif)

        # Build the driver (IVM or HMC) that will perform the
        # plug/unplug for us
        vif_conf = CONF.powervm_vif_driver
        vif_drv = importutils.import_object(vif_conf, host_name=self.host)

        # For logging purposes, get the network name.
        net_api = network.API()
        neutron_obj = net_api.get(context, data['net_id'])
        if neutron_obj is not None and hasattr(neutron_obj, 'name'):
            net_name = neutron_obj.name
        else:
            net_name = data['net_id']

        try:
            # Now start to plug/unplug
            msg = (_("Network %(net)s is being modified to use Shared "
                     "Ethernet Adapter (SEA) %(new_sea)s instead of SEA "
                     "%(old_sea)s on host %(host)s.") %
                   {'host': self.host, 'net': net_name,
                    'old_sea': old_sea, 'new_sea': new_sea})
            LOG.info(msg)

            # Happy path should just be 'unplug and plug'
            vif_drv.unplug(None, old_vif_list)

            # We update the data base now with the new SEA and network_ids.
            # This is because the corresponding plug call will check the
            # database for the network association to find the SEA to work
            # against.
            self.__update_na_and_peers_in_db(context, new_lpar,
                                             [(new_sea, net_ids_to_update)])
            vif_drv.plug(None, [new_vif])

            # Lastly notify the user that the change operation has completed
            msg = (_("Network %(net)s was successfully modified to use Shared "
                     "Ethernet Adapter (SEA) %(new_sea)s instead of SEA "
                     "%(old_sea)s on host %(host)s. Connectivity was restored "
                     "on all virtual machines that use the network") %
                   {'host': self.host, 'net': net_name,
                    'old_sea': old_sea, 'new_sea': new_sea})
            info = {'msg': msg, 'sea_name': new_sea, 'host_name': self.host}
            LOG.info(msg)
            notifier = rpc.get_notifier(service='compute', host=self.host)
            notifier.info(context, 'compute.instance.log', info)

        except Exception as ex:
            # Something went wrong...we need to log and then attempt to replug
            # to the original state.
            LOG.exception(ex)

            # Update the VIFs, note that we now need to set the force to true
            try:
                vif_drv.unplug(None, [new_vif])

                # Update the database to set the Network Mapping back to
                # the original value. This would mean reverting the peers to
                # the older SEA.
                self.__update_na_and_peers_in_db(context, old_lpar,
                                                 na_and_peers_tuple_list)

                vif_drv.plug(None, old_vif_list)
            except Exception as ex2:
                LOG.exception(ex2)

            # Lastly notify the user that something has gone wrong.
            msg = (_("The adapter change on Host %(host)s for Network "
                     "%(net)s failed to finish successfully.  The system has "
                     "returned the following error message: %(msg)s.  "
                     "Please review the issue and retry the command.") %
                   {'host': self.host, 'msg': ex.message,
                    'net': net_name})
            info = {'msg': msg, 'sea_name': new_sea, 'host_name': self.host}
            LOG.warn(msg)
            notifier = rpc.get_notifier(service='compute', host=self.host)
            notifier.warn(context, 'compute.instance.log', info)

    @periodic_task.periodic_task(spacing=CONF.cleanup_net_assn_interval)
    def _cleanup_network_associations(self, context):
        """
        This task periodically cleans up network associations
        of deleted networks
        """
        LOG.info(_('pvc_nova.compute.manager.PowerVCKvmHmcComputeManager '
                   '_cleanup_network_associations: updating network'
                   ' associations...'))
        if isinstance(self.driver, discovery_driver.ComputeDiscoveryDriver):
            self.driver.cleanup_network_associations(
                context, self.host)

    def __update_na_and_peers_in_db(self, context, lpar_id, sea_na_tuple_list):
        """
        Updates the network associations for the net_ids with a new SEA.

        :param context: The user context making the request
        :param lpar_id: The lpar housing the sea
        :param sea_na_tuple_list: A list of SEA/NA tuples to update.
        """
        netafact = network_dom.NetworkAssociationFactory.get_factory()
        hostfact = network_dom.ManagedHostFactory.get_factory()
        transaction = netafact.get_resource_transaction()
        with transaction.begin(subtransactions=True):
            host_dom = hostfact.find_host_by_name(
                context, self.host, transaction)

            for ttuple in sea_na_tuple_list:
                sea = ttuple[0]
                net_ids = ttuple[1]
                sea_dom = host_dom.find_adapter(lpar_id, sea)
                for net_id in net_ids:
                    netafact.update_network_assoc(
                        context, self.host, net_id, sea_dom, transaction)

    #######################################################
    # Moved check_host_maintance_mode here so we can catch
    # the exeption in attach_volume() and reset the state
    # of the volume to available when this check fails.
    #######################################################
    def attach_volume_check_VM_state(self, context):
        """
        checks that the VM is not in maint mode.
        :param context: operation context
        :return: None
        """
        LOG.debug("attach_volume_check")

    #######################################################
    #         Storage-related Overrides/Extensions        #
    #######################################################
    @object_compat
    @wrap_exception()
    @manager.pvc_comp_mgr_reverts_task_state
    @wrap_instance_fault
    def reserve_block_device_name(self, context, instance, device,
                                  volume_id, disk_bus, device_type):
        # copied from openstack nova manager.py to restrict attaching same
        # volume to same instance twice

        @utils.synchronized(instance.uuid)
        def do_reserve():
            if volume_id:
                bdm = None
                try:
                    bdm = objects.BlockDeviceMapping.\
                        get_by_volume_and_instance(
                            context, volume_id, instance.uuid)
                except Exception as e:
                    LOG.debug("Volume is not attached to bdm %s" %
                              six.text_type(e))
                if bdm:
                    reason = "Volume %s already attached to instance %s"\
                        % (volume_id, instance.uuid)
                    raise exception.InvalidVolume(reason=reason)
            bdms = (
                objects.BlockDeviceMappingList.get_by_instance_uuid(
                    context, instance.uuid))

            device_name = compute_utils.get_device_name_for_instance(
                instance, bdms, device)

            # NOTE(vish): create bdm here to avoid race condition
            bdm = objects.BlockDeviceMapping(
                context=context,
                source_type='volume', destination_type='volume',
                instance_uuid=instance.uuid,
                volume_id=volume_id or 'reserved',
                device_name=device_name,
                disk_bus=disk_bus, device_type=device_type)
            bdm.create()
            return bdm

        return do_reserve()

    @wrap_exception()
    @manager.pvc_comp_mgr_reverts_task_state
    @wrap_instance_fault
    @send_ui_notification(None, notify_msg.ATTACH_ERROR,
                          notify_msg.ATTACH_SUCCESS,
                          log_exception=True)
    def attach_volume(self, context, instance, bdm):
        """
        attach a volume to an instance, prevent any image volume
        from being attached.
        :param context: operation context
        :param instance: compute instance for attaching.
        :param bdm: block device mapping object.
        :return: None
        """
        volume_id = bdm.volume_id
        if not volume_id and bdm:
            volume_id = bdm.volume_id
        # check that the vm is in a state that allows attach
        try:
            self.attach_volume_check_VM_state(context)
        except Exception:
            # reset the state of the volume to available
            # we don't want to leave the state as attaching
            with excutils.save_and_reraise_exception():
                self.volume_api.unreserve_volume(context, volume_id)
                LOG.info("Cannot attach image because of the state of the VM")

        volume = self.volume_api.get(context, volume_id)
        metadata = volume.get("volume_metadata")

        # For image-volume, is_image_volume will be set
        # in the volume metadata.
        if metadata and ('is_image_volume' in metadata and
                         bool(metadata['is_image_volume'])):
            # roll back attaching state and log a warning message
            # if it is an image volume
            LOG.warn(_("Cannot attach image volume: %(volume_id)s") % locals())
            self.volume_api.unreserve_volume(context, volume_id)
            bdm.destroy(context)
            raise exception.InvalidVolume(reason='Cannot attach image volume')
        else:
            # Cleanup bdms for this specific volume_id before
            # attachment. Retreive a list of all bdms, run through the list
            # and destroy only bdms that match given volume_id

            s = super(PowerVCKvmHmcComputeManager, self)
            s.attach_volume(context, instance, bdm)

    def _get_all_bdms_for_instance_and_vol(self, context, instance, volume_id):
        """ Helper function to retrieve all bdms for
        instance"""
        # go get the bdm for this instance and volume id
        bdms = blkdev_obj.BlockDeviceMappingList.get_by_instance_uuid(
            context, instance['uuid'])
        return bdms

    def remove_volume_connection(self, context, volume_id, instance):
        """Override Remove Volume Connection to Clean-up Stale Disks"""
        bdm = objects.BlockDeviceMapping.get_by_volume_and_instance(
            context, volume_id, instance['uuid'])
        cinfo = jsonutils.loads(bdm['connection_info'])
        super(PowerVCKvmHmcComputeManager,
              self).remove_volume_connection(context, volume_id, instance)
        self.driver.remove_volume_device_at_destination(
            context, instance, cinfo)

    def _delete_instance(self, context, instance, bdms,
                         reservations=None):
        """Overridden to handle the HypervisorUnavailable exception
        in the case of PowerKVM. When the libvirt service is down Delete
        task hang indefinitely."""
        s = super(PowerVCKvmHmcComputeManager, self)
        try:
            s._delete_instance(context, instance, bdms,
                               reservations)
        except exception.HypervisorUnavailable:
            with excutils.save_and_reraise_exception():
                if (instance.task_state == task_states.DELETING):
                    instance.task_state = None
                    instance.expected_task_state = None
                    instance.save()

    def _shutdown_instance(self, context, instance,
                           bdms, requested_networks=None, notify=True,
                           try_deallocate_networks=True):
        """ override openstack _shutdown_instance in order to
        save the volume connector before destroy the partition.
        For NPIV/SSP attached volumes, they are not assigned to host.
        The connector doesn't exist after destroy the instance.
        For NPIV, this connector is important to remove the
        hostmap and delete the fabric zone. For SSP, termiante_connection
        is a no-op. For IVM, it is the same as openstack."""

        """Shutdown an instance on this host."""
        context = context.elevated()
        LOG.info(_('%(action_str)s instance: %(instuuid)s') %
                 {'action_str': 'Terminating', 'instuuid': instance['uuid']},
                 context=context, instance=instance)

        if notify:
            self._notify_about_instance_usage(context, instance,
                                              "shutdown.start")

        # get network info before tearing down
        try:
            network_info = compute_utils.get_nw_info_for_instance(instance)
        except (exception.NetworkNotFound, exception.NoMoreFixedIps):
            network_info = network_model.NetworkInfo()

        # refresh bdms from database to sync up the in memory copy
        bdms = blkdev_obj.BlockDeviceMappingList.get_by_instance_uuid(
            context, instance['uuid'])

        # get bdms and volume connector before destroying the instance
        vol_bdms = [bdm for bdm in bdms if bdm.is_volume]

        block_device_info = self._get_instance_block_device_info(
            context, instance)
        LOG.debug("block_device_info: %s" %
                  str(self._clean_bdm_info(block_device_info)))
        # Attempt driver destroy before releasing ip, may
        # want to keep ip allocated for certain failures
        # The volume mapping for the instance will be removed
        # by driver as well. _shutdown_instance should handle
        # the storage side hostmap and SAN fabric zoning removal.
        try:
            self.driver.destroy(context, instance, network_info,
                                block_device_info)

        except exception.InstancePowerOffFailure:
            # if the instance can't power off, don't release the ip
            with excutils.save_and_reraise_exception():
                pass
        except Exception:
            with excutils.save_and_reraise_exception():
                self.__cleanup_instance_bdm(context, instance, vol_bdms)
                # deallocate ip and fail without proceeding to
                # volume api calls, preserving current behavior
                try:
                    if try_deallocate_networks:
                        self._try_deallocate_network(context, instance,
                                                     requested_networks)
                        # If this is an imported network, we want to clean it
                        # up
                        if requested_networks is None:
                            self._cleanup_imported_networks(context,
                                                            network_info)
                except Exception as ex:
                    # LPAR may have been destroyed and there is no way
                    # back. Try to clean up as much as we can.
                    # The error has been logged in _try_deallocate_network.
                    reason = (_("Failed to remove the VLAN associated with "
                                "the LPAR virtual ethernet. Please check "
                                "host log for further details. Error: %(ex)s")
                              % dict(ex=ex))
                    self._error_notify(instance, context, 'warn', reason)
                    LOG.warn(_("Failed to remove the VLAN associated with "
                               "LPAR virtual ethernet. Instance: %(inst)s. "
                               "Error:%(error)s") %
                             {'lpar_name': instance['name'],
                              'inst': instance['uuid'],
                              'error': ex})
        try:
            if try_deallocate_networks:
                self._try_deallocate_network(context, instance,
                                             requested_networks)
                # If this is an imported network, we want to clean it up
                if requested_networks is None:
                    self._cleanup_imported_networks(context, network_info)
        except Exception as ex:
            # LPAR may have been destroyed and there is no way
            # back. Try to clean up as much as we can.
            # The error has been logged in _try_deallocate_network.
            reason = (_("Deleted LPAR successfully. But failed to remove "
                        "the VLAN associated with the LPAR virtual ethernet. "
                        "Please check host log for further "
                        "details. Error: %(ex)s") % dict(ex=ex))
            self._error_notify(instance, context, 'warn', reason)
            LOG.warn(_("Deleted LPAR %(lpar_name)s successfully. But failed "
                       "to remove the VLAN associated with LPAR virtual "
                       "ethernet. Instance: %(inst)s. Error:%(error)s") %
                     {'lpar_name': instance['name'],
                      'inst': instance['uuid'],
                      'error': ex})

        self.__cleanup_instance_bdm(context, instance, vol_bdms)
        # cleanup instance after block device cleanup.
        instance.root_gb = 0
        instance.save()

        if notify:
            self._notify_about_instance_usage(context, instance,
                                              "shutdown.end")

    def __cleanup_instance_bdm(self, context, instance, vol_bdms):
        try:
            # Get the volumes from the VM's image to ensure we do not delete
            # image volumes when deploys fail before the image volume in the
            # BDM is cloned and BDM updated.
            img_vols_to_clone = []
            try:
                image_ref = instance['image_ref']
                if image_ref != 'ONBOARDED_NO_IMAGE':
                    img_vols_to_clone = self.\
                        _get_image_bdm_volumes_to_clone(context, image_ref)
            except Exception:
                msg_args = {'image': image_ref,
                            'instance': instance['uuid']}
                LOG.warn(_('The image %(image)s for instance %(instance)s '
                           'could not be found.') % msg_args)
            last_connector = None
            vol_connector = None
            # Cleanup the zones on the 'first' disk rather than the last disk
            # We keep trying to cleanup until we get a valid connector back
            # from the cleanup helper with clean_zones property.
            # Handles the case where we have an invalid or partial connector.
            # Random order of SSP VSCSI and NPIV is handled because
            # clean_zones only shows up in the connector if the volume type
            # is npiv.
            clean_zones = True
            for bdm in vol_bdms:
                last_connector = self._bdm_cleanup_helper(
                    context, instance, bdm, vol_connector, clean_zones,
                    img_vols_to_clone)
                if last_connector is not None and \
                   'clean_zones' in last_connector:
                    clean_zones = False
        except Exception as ex:
            # LPAR may have been destroyed and there is no way
            # back. Try to clean up as much as we can.
            msg = (_("Failed to cleanup the vol bdms "
                     "Error: %(ex)s") % dict(ex=ex))
            LOG.warn(msg)

    def _is_volume_boot_set(self, context, volume_id):
        volume = self.volume_api.get(context, volume_id)
        if 'is_boot_volume' in volume['volume_metadata'] and \
                volume['volume_metadata']['is_boot_volume'] == 'True':
            return True
        return False

    @logcall
    def _attach_volume(self, context, instance, bdm):
        """
        Extend this method to make sure instance has the required virtual
        adapter and maps to physical fibre channel adapter prior to
        cinder's initialize_connection.
        """
        volume_id = bdm.volume_id
        is_boot_set = self._is_volume_boot_set(context, volume_id)
        mountpoint = bdm['mount_device']

        try:
            stg_type = self.volume_api.get_storage_type_for_volume(
                context, volume_id)
            # since vSCSI and NPIV both has the storage type as 'fc'
            # we need to identify if the volume attachment to be done
            # for vSCSI or NPIV before we pass to prep_instance_for_attach
            # in order to create the proper SCSI(vSCSI) or FC (NPIV) Adapters
            if stg_type == 'fc':
                connector = self.driver.get_volume_connector(instance)
                if not connector:
                    raise Exception(_('Cannot get volume connector to '
                                      'virtual machine %s') % instance['name'])
                if is_boot_set:
                    connector['connection_type'] = \
                        connector.pop('boot_conn_types', None)
                else:
                    connector['connection_type'] = \
                        connector.pop('data_conn_types', None)

                if ((connector['connection_type'] and
                     CONNECTION_TYPE_VSCSI in connector['connection_type'])):
                    stg_type = CONNECTION_TYPE_VSCSI

            self.driver.prep_instance_for_attach(context, instance, stg_type)
        except Exception:
            with excutils.save_and_reraise_exception() as ex:
                self.volume_api.unreserve_volume(context, volume_id)

                # We send an asynchronous notification that attachment failed,
                # but do not attempt to gather more useful info for the error
                # report (e.g. storage provider metainfo since that is likely
                # how we got here (volume_api.get_storage_type_for_volume())
                notify_info = {'instance_id': instance['uuid'],
                               'instance_name': instance['name'],
                               'volume_id': volume_id,
                               'host_name': self.host}
                if (isinstance(ex.value, dict) and
                        'kwargs' in ex.value and
                        'reason' in ex.value['kwargs']):
                    error = six.text_type(ex.value['kwargs']['reason'])
                else:
                    error = six.text_type(ex.value)
                log_msg = (_("Failed to attach volume %(volume_id)s to "
                             "%(name)s, likely due to lack of connectivity "
                             "to storage provider(s). Error: %(error)s") %
                           {'name': instance['name'],
                            'volume_id': volume_id,
                            'error': error})
                LOG.error(log_msg)

                msg = _("Could not attach volume {volume_id} to virtual "
                        "server {instance_name}.  Check that host "
                        "{host_name} has connectivity to all storage "
                        "providers.  Check the host's log files for "
                        "further details of this error.")

                notify_info['msg'] = msg
                notifier = rpc.get_notifier(service='volume.%s' % volume_id)
                notifier.error(context, 'volume.update.errors', notify_info)
                notifier = rpc.get_notifier(service='compute', host=self.host)
                notifier.error(context, 'compute.instance.log', notify_info)

        try:
            info = self._attach_volume_osee(context, volume_id, mountpoint,
                                            instance, is_boot_set)
        except Exception:
            with excutils.save_and_reraise_exception() as ex:
                # We send an asynchronous notification that attachment failed.
                notify_info = {'instance_id': instance['uuid'],
                               'instance_name': instance['name'],
                               'volume_id': volume_id,
                               'host_name': self.host}
                error = ex.value

                try:
                    # Attempt to gather some more useful information for the
                    # error report.

                    # Volume details
                    vol = self.volume_api.get(context, volume_id)
                    notify_info['volume_display_name'] = vol['display_name']

                    # Storage provider details
                    provider_info = self.volume_api.get_stg_provider_info(
                        context, volume_id=volume_id)
                    update = {'storage_provider_display_name':
                              provider_info['host_display_name'],
                              'storage_provider_name': provider_info['host'],
                              'storage_provider_id': provider_info['id'],
                              'error': error}
                    notify_info.update(update)

                    # UI replaces volume_id with volume_display_name hyperlink.
                    msg = _("Could not attach volume {volume_id} to "
                            "virtual server {instance_name}.  Check that "
                            "host {host_name} has connectivity to storage "
                            "provider {storage_provider_id}. If the provider "
                            "is a Fibre Channel SAN controller, then check "
                            "that the host has connectivity to the FC switch "
                            "for each registered fabric (ensure the FC "
                            "switches are operational). Also, check the "
                            "host's log files for further details of this "
                            "error: %(error)s") % {'error': error}

                except Exception as e2:
                    LOG.exception(e2)
                    # If anything goes wrong, fall back to just reporting the
                    # information that we have to hand.  We'd end up here if we
                    # failed due to Cinder being down, for example.
                    msg = _("Could not attach volume {volume_id} to virtual "
                            "server {instance_name}.  Check that host "
                            "{host_name} has connectivity to all storage "
                            "providers.  Check the host's log files for "
                            "further details of this error.")

                notify_info['msg'] = msg
                notifier = rpc.get_notifier(service='volume.%s' % volume_id)
                notifier.error(context, 'volume.update.errors', notify_info)
                notifier = rpc.get_notifier(service='compute', host=self.host)
                notifier.error(context, 'compute.instance.log', notify_info)

        return info

    def _driver_detach_volume(self, context, instance, bdm):
        """Do the actual driver detach using block device mapping."""
        volume_id = bdm['volume_id']

        s = super(PowerVCKvmHmcComputeManager, self)
        try:
            s._driver_detach_volume(context, instance, bdm)
        except Exception:
            with excutils.save_and_reraise_exception() as ex:
                if (isinstance(ex.value, dict) and
                        'kwargs' in ex.value and
                        'reason' in ex.value['kwargs']):
                    error = six.text_type(ex.value['kwargs']['reason'])
                else:
                    error = six.text_type(ex.value)
                # Attempt to gather some more useful information for the
                # error report.
                notify_info = {'instance_id': instance['uuid'],
                               'instance_name': instance['name'],
                               'volume_id': volume_id,
                               'host_name': self.host}
                try:
                    # Volume details
                    vol = self.volume_api.get(context, volume_id)
                    notify_info['volume_display_name'] = vol['display_name']

                    # Storage provider details
                    provider_info = self.volume_api.get_stg_provider_info(
                        context, volume_id=volume_id)
                    update = {'storage_provider_display_name':
                              provider_info['host_display_name'],
                              'storage_provider_name': provider_info['host'],
                              'storage_provider_id': provider_info['id'],
                              'error': error}
                    notify_info.update(update)

                    # UI replaces volume_id with volume_display_name hyperlink.
                    msg = _("Could not detach volume {volume_id} from "
                            "virtual server {instance_name}.  Check that "
                            "host {host_name} has connectivity to storage "
                            "provider {storage_provider_id}. If the provider "
                            "is a Fibre Channel SAN controller, then check "
                            "that the host has connectivity to the FC switch "
                            "for each registered fabric (ensure the FC "
                            "switches are operational). Also check the "
                            "host's log files for further details of this "
                            "error: {error}")

                except Exception as e2:
                    LOG.exception(e2)
                    # If anything goes wrong, fall back to just reporting the
                    # information that we have to hand.  We'd end up here if we
                    # failed due to Cinder being down, for example.
                    msg = _("Could not detach volume {volume_id} from virtual "
                            "server {instance_name}.  Check that host "
                            "{host_name} has connectivity to all storage "
                            "providers.  Check the host's log files for "
                            "further details of this error.")

                notify_info['msg'] = msg
                notifier = rpc.get_notifier(service='compute', host=self.host)
                notifier.error(context, 'compute.instance.log', notify_info)

    def _cleanup_volumes(self, context, instance_uuid, bdms, raise_exc=True):
        """ Override nova compute manager's method. All the volumes
        that have been marked with 'delete_on_termination' are
        handled by _shutdown_instance already.
        NO-OP the parent's _cleanup_volumes otherwise
        it will generate an exception in the compute log.
        """
        pass

    def _restore_pre_migrate_bdms(self, context, inst_uuid, migrate_data):
        """Override restore pre-migrate BDMS data for attached volumes"""
        org_bdms = None
        if hasattr(migrate_data, 'pre_migrate_bdms'):
            org_bdms = migrate_data.pre_migrate_bdms
        if not org_bdms:
            return
        # First get the Block Device Mappings for the Instance to Update them
        bdms = blkdev_obj.BlockDeviceMappingList.\
            get_by_instance_uuid(context, inst_uuid)
        bdms = dict([(bdm['id'], bdm) for bdm in bdms])
        # Loop through the Original BDM's, updating the queried ones
        for bdm in org_bdms:
            if bdm['id'] in bdms:
                bdms[bdm['id']].connection_info = bdm['connection_info']
                bdms[bdm['id']].save()

    def _get_instance_block_device_info(self, context, instance,
                                        refresh_conn_info=False, bdms=None):
        """Override Nova compute manager's method in order to control
        refresh_conn_info. For NPIV VFC attached volume(HMC only), the
        connection stays unchanged during the instance's life cycle.
        There is no need to refresh.
        """
        s = super(PowerVCKvmHmcComputeManager, self)
        if CONF.powervm_mgr_type and CONF.powervm_mgr_type.lower() == 'hmc' \
                and refresh_conn_info:
            # refresh connection is true under resize. There is no need
            # to refresh for NPIV/SSP attached volume during resize.
            return s._get_instance_block_device_info(
                context, instance, refresh_conn_info=False, bdms=bdms)
        else:
            return s._get_instance_block_device_info(
                context, instance, refresh_conn_info=refresh_conn_info,
                bdms=bdms)

    def _attach_volumes_and_boot(self, context, instance):
        """
        This is callback function to attach volumes before boot
        up lpar.
        :param context: operation context
        :param instance: compute instance
        :return: block_device_info
        """
        LOG.debug("Begin _attach_volumes_and_boot(), %s" % instance['uuid'])

        self._validate_state_before_attach_and_boot(
            context, instance['uuid'],
            method=self._attach_volumes_and_boot.func_name
        )

        bdms = blkdev_obj.BlockDeviceMappingList.get_by_instance_uuid(
            context, instance['uuid'])
        original_volume_bdms = [bdm for bdm in bdms if bdm.is_volume]

        attached_bdms = []

        # loop through all the volumes ready to attach to the instance
        # during boot and handle the volumes attach.
        # Sort the volumes by 'boot_index' so they get attached
        # in the proper boot order.
        sorted_bdms = block_device_utils.sort_instance_bdms(bdms)
        for bdm in sorted_bdms:
            LOG.debug("Processing BDM %s" % bdm)
            volume_id = bdm['volume_id']

            mountpoint = bdm['device_name']
            # save delete_on_termination since nova compute manage
            # _attach_volume will set delete_on_termination to False
            # regardless.

            delete_on_termination = bdm['delete_on_termination']

            if not mountpoint:
                # now generate a unique mountpoint. Since the instance
                # is not in the state the volume can be attached to,
                # no serialization is needed here. The volume has been
                #  reserved. It cannot be attached to other instance
                # at the same time.

                # bdms is updated every time a volume is attached. Get new copy
                bdms = blkdev_obj.BlockDeviceMappingList.get_by_instance_uuid(
                    context, instance['uuid'])

                # Take the bdms out of the list that don't have a device_name,
                # because nova/compute/utils/get_next_device_name() expects all
                # device_names to have values.
                bdms = [b for b in bdms if b.get('device_name')]

                mountpoint = compute_utils.get_device_name_for_instance(
                    instance, bdms, None)

                # Set the device_name in the bdm so attach doesn't try to
                # create a new bdm.
                self._update_bdm_device_name(context, bdm, mountpoint)
                # set the device name in the bdm we already have.
                bdm['device_name'] = mountpoint

            try:
                driver_bdm_list = self._convert_block_devices([bdm])
                self._attach_volume(context, instance, driver_bdm_list[0])
                # _attach_volume is from parent class, which overwrite the
                # delete_on_termination to False regardless. Restore the proper
                # delete_on_termination after attach_volume.
                bdm.delete_on_termination = delete_on_termination
                bdm.save()

            except Exception:
                with excutils.save_and_reraise_exception() as ex:
                    ex_info = ex.value

                    reason = (_("Could not deploy %(name)s due to "
                                "volume: %(volume_id)s attach failure. "
                                "Error: %(error)s") %
                              {'name': instance['name'],
                               'volume_id': volume_id,
                               'error': ex_info})
                    # The notification has been sent by _attach_volume
                    # don't send it again. Just log the error.
                    LOG.error(reason)
                    # also need to terminate and detach all the attached
                    # volumes if we fail to attach boot volume
                    # For the 'delete_on_termination' volumes,
                    # _delete_instance will clean them up during "delete"
                    # Just clean up attachment here. _delete_instnace will
                    # also clean up the instance BDMS.
                    attached_vol_ids = []
                    for tmpbdm in attached_bdms:
                        volid = tmpbdm['volume_id']
                        attached_vol_ids.append(volid)
                        self._detach_volume(context, volid, instance,
                                            destroy_bdm=False)
                        connector = \
                            self.driver.get_volume_connector(instance)
                        self.driver.call_terminate_connection(context,
                                                              instance,
                                                              volid,
                                                              connector)
                        self.volume_api.detach(context, volid,
                                               instance_uuid=instance['uuid'])

                    # All volumes have their state set to 'attaching' by the
                    # _validate_volumes callback before this method is called.
                    # We need to unreserve any volumes that were not
                    # successfully attached to free-up any volumes that the
                    # user added to the deploy as 'attach existing volume'.
                    for bdm in original_volume_bdms:
                        if bdm['volume_id'] not in attached_vol_ids:
                            self.volume_api.unreserve_volume(context,
                                                             bdm['volume_id'])

            attached_bdms.append(bdm)

        # return block_device_info
        return self._get_instance_block_device_info(context, instance)

    def _clean_bdm_info(self, bdm_info):
        # sanitize the passwords out of a bdm_info
        bdm_list = []
        found_diff = False
        if bdm_info and 'block_device_mapping' in bdm_info:
            for bdm in bdm_info['block_device_mapping']:
                bdm_sanitized = self._clean_bdm(bdm)
                bdm_list.append(bdm_sanitized)
                if bdm != bdm_sanitized:
                    found_diff = True
            if found_diff:
                bdm_info_copy = copy.deepcopy(bdm_info)
                bdm_info_copy['block_device_mapping'] = bdm_list
                return bdm_info_copy
        return bdm_info

    def _rollback_initialize_connection(self, context, instance, volume_id,
                                        connector, conninfo=None):
        """
        Calls the terminate connection to remove the mapping from the
        SAN, It handles the case of Dual VIOS mapping in case of
        vSCSI attachments
        """
        LOG.info(_("_rollback_initialize_connection %(con)s"
                   "and %(con_info)s"), {'con': connector,
                                         'con_info': conninfo})
        self.driver.call_terminate_connection(context, instance, volume_id,
                                              connector, conninfo)

    def _get_connection_type(self, connector):
        if (connector is not None and 'connection_type' in connector and
                (connector['connection_type'] is not None)):
            return connector['connection_type'][0]

    @logcall
    def _attach_volume_osee(self, context, volume_id, mountpoint,
                            instance, is_boot_set):
        context = context.elevated()
        LOG.info(_('Attaching volume %(volume_id)s to %(mountpoint)s'),
                 {'volume_id': volume_id, 'mountpoint': mountpoint},
                 context=context, instance=instance)
        try:
            connector = self.driver.get_volume_connector(instance)
            msg = (_("Attaching volume %(volume_id)s to instance %(inst)s with"
                     " connector: %(connector)s") %
                   {'volume_id': volume_id,
                    'inst': instance['uuid'],
                    'connector': connector})
            LOG.info(msg)
            # for boot volumes the boot_connectivity value needs to be passed
            # in connector info
            c_info = None
            if connector:
                if is_boot_set:
                    connector['connection_type'] = connector.pop(
                        'boot_conn_types')\
                        if 'boot_conn_types' in connector else None
                    connector.pop('data_conn_types', None)
                else:
                    connector['connection_type'] = connector.pop(
                        'data_conn_types')\
                        if 'data_conn_types' in connector else None
                    connector.pop('boot_conn_types', None)

                if connector['connection_type'] and\
                   CONNECTION_TYPE_VSCSI in connector['connection_type']:
                    # for vSCSI we do not need this property so pop it
                    if 'phy_to_virt_initiators' in connector.keys():
                        connector.pop('phy_to_virt_initiators')

                    wwpn_vscsi = connector.pop('wwpn_vscsi')
                    vscsi_lun_list = []
                    for wwpns in wwpn_vscsi:
                        connector['wwpns'] = wwpns['wwpns']
                        connector['host'] = wwpns['vios_host']
                        connector['ip'] = wwpns['vios_ip']
                        c_info = self.volume_api.\
                            initialize_connection(context,
                                                  volume_id, connector)
                        vios_target_lun = {
                            'vios_name': wwpns['vios_name'],
                            'target_lun': c_info['data']['target_lun'],
                            'vios_wwpn': wwpns['wwpns'],
                            'target_wwpn': c_info['data']['target_wwn'],
                            'vios_ip': wwpns['vios_ip'],
                            'vios_host': wwpns['vios_host']
                        }
                        vscsi_lun_list.append(vios_target_lun)
                    if c_info:
                        c_info['vscsi_target_lun'] = vscsi_lun_list
                    else:
                        raise Exception("Failed to intialize vSCSI Volume "
                                        "connection")
                else:
                    # for NPIV we do not need this property so pop it
                    if 'wwpn_vscsi' in connector.keys():
                        connector.pop('wwpn_vscsi')
            if not c_info:
                c_info = self.volume_api.\
                    initialize_connection(context, volume_id, connector)
            # [shyama]connection_info should now have
            # {'driver_volume_type':'ssp'/'fibre_channel',
            #  'data': {..'connector_type':'npiv'/'pv_vscsi'...} }
            msg = (_("Successfully established volume %(volid)s "
                     "connectivity, connection_info: %(conn_info)s") %
                   dict(volid=volume_id,
                        conn_info=self._clean_conn_info(c_info)))
            LOG.info(msg)
        except Exception:  # pylint: disable=W0702
            with excutils.save_and_reraise_exception():
                LOG.exception(_("Failed to connect to volume %(volume_id)s "
                                "while attaching at %(mountpoint)s"),
                              {'volume_id': volume_id,
                               'mountpoint': mountpoint},
                              context=context, instance=instance)
                self.volume_api.unreserve_volume(context, volume_id)
        if 'serial' not in c_info:
            c_info['serial'] = volume_id

        try:
            self.driver.attach_volume(context,
                                      c_info,
                                      instance,
                                      mountpoint,
                                      encryption=None)
            msg = (_("Successfully attached volume: %(volid)s to instance:"
                     "%(inst)s.") % dict(volid=volume_id,
                                         inst=instance['uuid']))
            LOG.info(msg)
        except Exception:  # pylint: disable=W0702
            with excutils.save_and_reraise_exception():
                LOG.exception(_("Failed to attach volume %(volume_id)s "
                                "at %(mountpoint)s") %
                              {'volume_id': volume_id,
                               'mountpoint': mountpoint},
                              context=context, instance=instance)
                LOG.info(_("Clean up failed volume %(volid)s attachment "
                           "to instance %(inst)s with connector: "
                           "%(connector)s") %
                         {'volid': volume_id,
                          'inst': instance['uuid'],
                          'connector': connector})
                self._rollback_initialize_connection(context, instance,
                                                     volume_id,
                                                     connector,
                                                     c_info)

        self.volume_api.attach(context,
                               volume_id,
                               instance['uuid'],
                               mountpoint)
        # validate connector and connection_info and merge.
        # the updated connection info will have connector saved it
        # if a valid NPIV or SSP connector. Each volume attached
        # will its own connector saved.
        conn_info = self._update_conninfo_with_connector(
            context, instance, volume_id, c_info, connector)
        bdm_obj = blkdev_obj.BlockDeviceMapping(
            context=context, volume_id=volume_id or 'reserved',
            instance_uuid=instance.uuid, device_name=mountpoint,
            connection_info=jsonutils.dumps(conn_info))
        bdm_obj.update_or_create()

        self._instance_update(context, instance,
                              updated_at=timeutils.utcnow())
        info = dict(volume_id=volume_id)
        self._notify_about_instance_usage(
            context, instance, "volume.attach", extra_usage_info=info)

    def _validate_state_before_attach_and_boot(self, context,
                                               instance_uuid, method=None):
        """
        validate instance state before attach volumes and boot

        :param context: operation context
        :param instance_uuid: compute instance
        :param method: compute manager method that calls validate
        :return: instance if validate or None
        """

        instance = instance_obj.Instance.get_by_uuid(context, instance_uuid)
        vm_state = instance['vm_state']
        task_state = instance['task_state']

        if (vm_state != vm_states.BUILDING or
                task_state != task_states.SPAWNING):
            msg = (_("_validate_volumes() is called when instance is "
                     "in the wrong state: vm_state=%(vm_state)s "
                     "task_state=%(task_state)s") % locals())
            LOG.error(msg)
            exc_args = {'instance_uuid': instance_uuid,
                        'attr': vm_state,
                        'state': task_state,
                        'method': method}
            raise exception.InstanceInvalidState(**exc_args)

        return instance

    def _validate_attachment(self, context, instance_uuid):
        pass

    def _validate_volumes(self, context, instance_uuid):
        """
        callback function used by virtual driver during spawn to
        validate all the volumes state before attach those
        volumes and boot.
        :param context: operation context
        :param instance_uuid: compute instance uuid
        :return: None
        """
        LOG.debug("Begin _validate_volumes():")
        instance = self._validate_state_before_attach_and_boot(
            context, instance_uuid, method=self._validate_volumes.func_name
        )
        bdms = blkdev_obj.BlockDeviceMappingList.get_by_instance_uuid(
            context, instance['uuid'])
        bdms = [bdm for bdm in bdms if bdm.is_volume]
        for bdm in bdms:
            volume = self.volume_api.get(context, bdm.volume_id)
            blank_volume = (bdm.source_type == 'blank')
            self._validate_for_error_status(context, instance, volume,
                                            blank_volume)
            if volume['status'] not in ['available', 'in-use']:
                stg_type = 'fc'
                try:
                    stg_type = self.volume_api.get_storage_type_for_volume(
                        context, bdm.volume_id)
                except Exception:
                    stg_type = 'fc'

                try:
                    if stg_type != 'cluster':
                        self._await_block_device_map_created(context,
                                                             bdm.volume_id)
                    else:
                        # Wait forever for SSP cluster volume clone creations
                        self._await_volume_create(context, bdm.volume_id)
                except exception.VolumeNotCreated:
                    with excutils.save_and_reraise_exception():
                        volume = self.volume_api.get(context, bdm.volume_id)
                        self._notify_deploy_volume_create_failure(context,
                                                                  instance,
                                                                  volume)
                volume = self.volume_api.get(context, bdm.volume_id)

            # make sure the volume is in the right state to attach
            self.volume_api.check_attach(context, volume, instance=instance)
            # change the volume to "attaching" state
            self.volume_api.reserve_volume(context, volume['id'])

    def _reset_volume_bdm(self, bdm, source_volume_id, img_vols_to_clone):
        if not bdm:
            return
        if bdm['delete_on_termination']:
            if (bdm['source_type'] == 'volume' and
                    source_volume_id in img_vols_to_clone):
                bdm['volume_id'] = source_volume_id
                bdm['delete_on_termination'] = False
                LOG.info(_('Resetting BDM %(bdm_id)s volume_id to %(vol_id)s'
                           'and delete_on_termination value to False.')
                         % {'bdm_id': bdm['id'],
                            'vol_id': source_volume_id})
            elif bdm['source_type'] == 'blank':
                bdm['volume_id'] = None
                LOG.info(_('Resetting BDM %(bdm_id)s volume_id to %(vol_id)s')
                         % {'bdm_id': bdm['id'],
                            'vol_id': 'None'})
        bdm['connection_info'] = None
        bdm.save()

    def _bdm_cleanup_helper(self, context, instance, bdm, connector=None,
                            clean_zones=False, img_vols_to_clone=[]):
        """
        The helper function to be used by _shutdown_instance(). The
        purpose is to consolidate all the volume detach and terminate
        connection function needed during VM destroy.
        """
        if not bdm:
            return

        volume_id = None
        source_volume_id = None
        try:
            volume_id = bdm['volume_id']
            # BDMs with source type of blank may not have a volume_id
            # if the deploy failed very early.
            if not volume_id:
                LOG.debug("BDM has the volume_id set to None")
                return

            # We do not process volumes that are in the image and will be
            # cloned as all operations would be against a clone and not
            # the image volume.
            if volume_id in img_vols_to_clone:
                LOG.debug("Volume %(vol)s is a part of the image %(img)s" %
                          {'vol': volume_id, 'img': img_vols_to_clone})
                return

            volume = self.volume_api.get(context, volume_id)
            source_volume_id = volume['source_volid']
            stg_type = self.volume_api.get_storage_type_for_volume(context,
                                                                   volume_id)
            # For the volume that is not fully attached yet.
            if not self._bdm_has_attachment(bdm):
                # Volumes are 'reserved' and placed into 'attaching' state
                # early in the deploy process before the LPAR is created and
                # before attachment really begins.
                # We need to unreserve the volumes so that volumes marked with
                # delete_on_termination=true can be deleted and volumes
                # marked with delete_on_termination=false can be freed up
                # so they are available for VM attach to other VMs on deploy
                # failure or this same VM on deploy-fail-reschedule.
                if volume['status'] == 'attaching':
                    self.volume_api.unreserve_volume(
                        context, bdm['volume_id'])
                if bdm['delete_on_termination']:
                    LOG.debug("Terminating bdm: %(bdm)s" % dict(bdm=bdm))
                    # if the volume is creating we are deleting a failed
                    # deploy. So we need to wait for the volume to
                    # finish creating before we delete it.
                    # If we leave it as 'creating' delete will fail.
                    # The volume create will eventually finish or fail
                    # so wait here.
                    # If the volume is attached we would not get here
                    # and we would
                    if volume['status'] == 'creating':
                        self._await_volume_create(context,
                                                  volume_id)
                    self.volume_api.delete(context, bdm['volume_id'])
                return

            if not connector:
                # retrieve the connector from bdm connection_info
                connector = self._get_bdm_volume_connector(
                    context, instance, bdm, bdm['volume_id'])
                # for SSP connector will be None
                if not connector and stg_type == 'fc':
                    # the connector is lost, but there is an
                    # attachment.
                    if volume['attach_status'] == 'attached':
                        connector = self._construct_connector_from_conninfo(
                            instance, bdm)

                        # for onboarded vm's vscsi volumes
                        # wwpns are missing so get connector again
                        if not connector:
                            try:
                                c_info = bdm.get("connection_info")
                                connection_info = jsonutils.loads(c_info)
                                vol_conn = connection_info.get(
                                    "volume_connector")
                                connection_type = vol_conn.get(
                                    'connection-type')
                                if CONNECTION_TYPE_VSCSI == connection_type:
                                    connector = self.driver. \
                                        get_volume_connector(instance,
                                                             isVscsi=True)
                                    LOG.info(_("Retrieved connector from"
                                               " driver : %(connector)s" %
                                               dict(connector=connector)))
                            except Exception as ex:
                                LOG.error(_("Failed to retrieve connector for "
                                            "instance %(inst)s due to ,"
                                            "%(error)s") %
                                          {'inst': instance['uuid'],
                                           'error': ex})
                        if not connector:
                            # This is critical condition. Send UI notification.
                            message = (_("Could not terminate storage host "
                                         "mapping and SAN zoning for volume: "
                                         "%(volid)s when deleting "
                                         "the virtual server. Please check the"
                                         " host logs for further details.") %
                                       dict(volid=volume_id))
                            self._error_notify(instance, context,
                                               'critical',
                                               message)
                            # No need to try terminate_connection which will
                            # raise exception. The nova "detached" volume
                            # will be left in 'in-use' state. This needs
                            # some out of band clean up.
                            # TODO: Need to have cinder volume driver to
                            # rebuild the connector and send it back.
                            return
                        # Built a connnector based on the connection_info.
                        LOG.debug("connector created from connection_info: "
                                  "%(connector)s" % dict(connector=connector))

            LOG.info(_("Detach volume %(volume_id)s for instance %(inst)s "
                       "with connector: %(connector)s") %
                     {'volume_id': volume_id,
                      'inst': instance['uuid'],
                      'connector': connector})

            # detach then terminate
            self.volume_api.detach(context, volume_id,
                                   instance_uuid=instance['uuid'])

            if (connector is not None and 'connection_type' in connector and
                CONNECTION_TYPE_VSCSI in connector['connection_type']) or\
               (connector is not None and 'wwpn_vscsi' in connector):
                if 'phy_to_virt_initiators' in connector and\
                   not connector['phy_to_virt_initiators']:
                    connector.pop('phy_to_virt_initiators', None)
                conninfo = jsonutils.loads(bdm['connection_info'])
                self.driver.call_terminate_connection(context, instance,
                                                      bdm['volume_id'],
                                                      connector, conninfo)
            else:
                # clean_zones only applies to NPIV (stg_type fc)
                # no zoning for iscsi or vscsi
                if stg_type == 'fc' and clean_zones:
                    # It is the last attached volume for VM. Tell cinder that
                    # it is time to remove all the zones.
                    connector['clean_zones'] = 'all'
                self.volume_api.terminate_connection(context,
                                                     bdm['volume_id'],
                                                     connector)
            # make sure the volume in attaching state changes back to
            # available, so it could be handled by cleanup_volumes().
            self.volume_api.unreserve_volume(context, bdm['volume_id'])
            if volume_id and bdm['delete_on_termination']:
                LOG.debug("Terminating bdm: %(bdm)s" % dict(bdm=bdm))
                self.volume_api.delete(context, bdm['volume_id'])
        except cinder_exception.EndpointNotFound as exc:
            # if the failure is due to endpoint not found,
            # it's likely that context is out of date.
            # We should fail the operation and ask for retry
            with excutils.save_and_reraise_exception():
                raise Exception(_('Failed to access cinder service to '
                                  'remove volumes. Try the operation again.'))
        except exception.DiskNotFound as exc:
            LOG.warn(_('Ignoring DiskNotFound: %(exc)s, instance: %(inst)s,'
                       'volume id: %(volid)s') %
                     dict(exc=exc, inst=instance['uuid'], volid=volume_id))
        except exception.VolumeNotFound as exc:
            LOG.warn(_('Ignoring VolumeNotFound: %(exc)s, instance: %(inst)s,'
                       'volume id: %(volid)s') %
                     dict(exc=exc, inst=instance['uuid'], volid=volume_id))
        except Exception as exc:
            LOG.exception(exc)
            # differentiate based on SSP or not
            if not connector or ('connection_type' in connector and
                                 CONNECTION_TYPE_SSP in
                                 connector['connection_type']):
                reason = (_("The virtual server was deleted. "
                            "However, a virtual server cleanup "
                            "problem occurred on one of the "
                            "following operations: "
                            "The host mapping was unable to be "
                            "removed for one or more volumes on the "
                            "storage controllers and the host "
                            "mapping may require manual "
                            "cleanup, or one or more volumes on the "
                            "storage controllers failed removal and "
                            "may require manual cleanup. "
                            "Error: %(exc)s")
                          % dict(exc=exc))
            else:
                reason = (_("The virtual server was deleted. "
                            "However, a virtual server cleanup "
                            "problem occurred on one of the "
                            "following operations: "
                            "The host mapping was unable to be "
                            "removed for one or more volumes on the "
                            "storage controllers and the host "
                            "mapping may require manual "
                            "cleanup, or one or more volumes on the "
                            "storage controllers failed removal and "
                            "may require manual cleanup, "
                            "or the NPIV zone information was "
                            "unable to be removed from the switch "
                            "fabric and the zone may require manual "
                            "cleanup. Error: %(exc)s")
                          % dict(exc=exc))
            self._error_notify(instance, context, 'info', reason)
            LOG.debug("Unable to terminate and detach volume for instance "
                      "%(inst)s %(volid)s with connector: %(conn)s, BDM: "
                      "%(bdm)s" %
                      {'volid': bdm['volume_id'],
                       'inst': instance['uuid'],
                       'conn': connector,
                       'bdm': self._clean_bdm(bdm)})
        finally:
            try:
                context.read_deleted = 'yes'
                self._reset_volume_bdm(bdm, source_volume_id,
                                       img_vols_to_clone)
            finally:
                context.read_deleted = 'no'
        return connector

    @logcall
    def _get_bdm_volume_connector(self, context, instance, bdm, volume_id):
        """ For each volume that has been attached, it will be attached
        with a certain connector which will not change during the life
        cycle of the attachment. Save the connector to the connection_info.
        And it will be retrieved during terminate_connection. For NPIV,
        get_volume_connector could change since the SCG could be different.
        By saving the connector, the terminate_connection will be successful.
        Otherwise, the terminate_connector may not work if a different
        connector is given that is different from the one that used
        to initialize_connection().
        """

        if not bdm:
            bdm = objects.BlockDeviceMapping.get_by_volume_and_instance(
                context, volume_id, instance['uuid'])
        if not bdm or bdm['volume_id'] != volume_id:
            return None

        connector = None
        c_info = bdm.get("connection_info")

        if c_info and isinstance(c_info, basestring):
            try:
                connection_info = jsonutils.loads(c_info)
                sanitized_bdm = self._clean_bdm(bdm)
                LOG.debug("BDM: %(bdm)s" % {'bdm': sanitized_bdm})
                vol_conn = connection_info.get("volume_connector")
                connector = self._volume_connector_validator(vol_conn)
            except:
                # invalid connector
                connector = None
        if not connector:
            LOG.info(_("Volume_connector is missing from block device "
                       "mapping for instance %(vm_id)s and volume "
                       "%(vol_id)s. A connector may be generated.") %
                     dict(vm_id=instance['id'], vol_id=volume_id))
        return connector

    @logcall
    def _get_SSP_volume_connector(self, instance, vol_id, context):
        """ Check to see if the storage type for this volume is SSP
        """
        try:
            stg_type = self.volume_api.get_storage_type_for_volume(context,
                                                                   vol_id)
        except Exception:
            stg_type = 'fc'
        if stg_type == 'cluster':
            volume_connector = {'connection-type': CONNECTION_TYPE_SSP,
                                'connector': None}
            return volume_connector
        return None

    def _construct_connector_from_conninfo(self, instance, bdm):
        """ The connection_info for NPIV attached volume has an
        attribute called "initiator_target_map". It contains the
        vfc wwpn to target wwpn mapping. This could also be used
        to build a potential "good enough" connector for
        terminate_connection to use.
        """
        connector = None
        if not bdm or not bdm.get("connection_info"):
            return connector
        try:
            conninfo = jsonutils.loads(bdm['connection_info'])
            if not conninfo.get('driver_volume_type'):
                return connector
            conn_type = conninfo['driver_volume_type']
            if conn_type == 'fibre_channel' and conninfo.get('data'):
                conn_data = conninfo['data']
                if conn_data.get('initiator_target_map'):
                    # this is the NPIV connection. Let's construct the
                    # connector from the map.
                    # initiator_target_map looks like this:
                    # {"c0507606529b01a3":
                    #         ["500507680510041a", "500507680510041b"],
                    #  "c0507606529b01a2":
                    #         ["500507680510041a", "500507680510041b"],
                    #  ....}
                    i_t_map = conn_data['initiator_target_map']
                    if i_t_map and isinstance(i_t_map, dict):
                        i_wwpns = set(i_t_map.keys())
                        connector = {'connection_type': [CONNECTION_TYPE_NPIV],
                                     'host': instance['name'],
                                     'wwpns': list(i_wwpns),
                                     'phy_to_virt_initiators': None}
                if(conn_data.get('wwpns')):
                    connector = {'connection_type': [CONNECTION_TYPE_VSCSI],
                                 'host': instance['name'],
                                 'wwpns': conn_data.get('wwpns')}
            elif conn_type == CONNECTION_TYPE_SSP:
                connector = {'connection_type': [CONNECTION_TYPE_SSP],
                             'connector': None}
        except Exception as ex:
            LOG.exception(ex)
            connector = None
        LOG.debug("_construct_connector_from_conninfo: %(connector)s" %
                  locals())
        return connector

    @logcall
    def _update_conninfo_with_connector(self, context, instance, volume_id,
                                        connection_info, connector):
        """
        This is helper function to update the connection_info with a new
        volume_connector field if the attached volume connector is VFC(NPIV)
        or SSP based. For HMC managed PowerVC, the volume connector will
        not change during the life cycle of the volume attachment even
        after migration. Save the connector can help speed up the
        detach_volume patch as well destroy/cleanup path with may not
        have a valid volume connector since LPAR may be in some invalid
        state.
        """
        if not volume_id or not connection_info or not connector:
            return connection_info
        volume_connector = None

        try:
            volume = self.volume_api.get(context, volume_id)

            if 'error' in volume['status']:
                # The volume is not healthy after initialize_connection.
                # something went wrong and we don't want to update
                # the BDM with the connection info.
                raise ValueError
            if connection_info['driver_volume_type'] == 'ibm_ssp':
                volume_connector = {'connection-type': CONNECTION_TYPE_SSP,
                                    'connector': None}
            elif (connection_info['driver_volume_type'] == 'fibre_channel' and
                  connection_info['data']['connector_type'] ==
                  CONNECTION_TYPE_NPIV and
                  connector['phy_to_virt_initiators'] and
                  connector['host'] and
                  connector['wwpns'] and
                  connection_info['data']['target_wwn'] and
                  int(connection_info['data']['target_lun']) >= 0 and
                  connection_info['data']['volume_id'] == volume_id):
                # if there is anything missing, it will cause exception.
                volume_connector = {'connection-type': CONNECTION_TYPE_NPIV,
                                    'connector': connector}
            elif (connection_info['driver_volume_type'] == 'fibre_channel' and
                  connection_info['data']['connector_type'] ==
                  CONNECTION_TYPE_VSCSI and
                  connector['host'] and
                  connector['wwpns'] and
                  connection_info['data']['target_wwn'] and
                  int(connection_info['data']['target_lun']) >= 0 and
                  connection_info['data']['volume_id'] == volume_id):
                # if there is anything missing, it will cause exception.
                volume_connector = {'connection-type': CONNECTION_TYPE_VSCSI,
                                    'connector': connector}
            else:
                volume_connector = None
        except:
            volume_connector = None
        # for IVM san FC connection, it will not be updated with the
        # connector
        if volume_connector:
            connection_info.update({'volume_connector': volume_connector})
        return connection_info

    @logcall
    def _volume_connector_validator(self, volume_connector):
        """
        Validate the volume connector format saved into the. Not the content.
        This is the volume_connector saved in the BDM connection_info.
        Will return the real volume connector if it is valid.
        For IVM FC connector, we don't save it since it could be
        attached to both source/target host during migration.
        volume_connector:
            { 'connection-type': 'npiv'| 'ibm_ssp',
              'connector': {...} actual connector returned from
              get_volume_connector
            }
            fc is the legacy FibreChannel connector
            npiv is the NPIV VFC connector
            ibm_ssp is the ssp vscsi connector
        """

        if not volume_connector:
            return None

        conn_info = None
        try:
            c_type = volume_connector['connection-type']
            conn_info = volume_connector['connector']
            if c_type == CONNECTION_TYPE_NPIV:
                if (not conn_info['host'] or
                    not conn_info['wwpns'] or
                        not conn_info['phy_to_virt_initiators']):
                    conn_info = None
            elif c_type == CONNECTION_TYPE_VSCSI:
                if (not conn_info['host'] or
                        not conn_info['wwpns']):
                    conn_info = None
            elif c_type == CONNECTION_TYPE_SSP:
                conn_info = None
            else:
                conn_info = None
        except:
            # validation failure
            conn_info = None
        return conn_info

    def _update_bdm_device_name(self, context, bdm, device_name):
        bdm.device_name = device_name
        bdm.save()

    def _valid_volume_connector(self, connector):
        if not connector:
            return False
        if connector.get('host') and connector.get('wwpns'):
            return True
        else:
            return False

    @periodic_task.periodic_task
    @cache_driver_data(['get_available_resource'])
    @lockutils.synchronized('set_mobile_cores', 'powervc-')
    def pvc_update_available_resource(self, context):
        return super(PowerVCKvmHmcComputeManager, self).\
            pvc_update_available_resource(context)

    @periodic_task.periodic_task(spacing=CONF.sync_power_state_interval,
                                 run_immediately=True)
    def _pvc_sync_power_states(self, context):
        """
        Override _sync_power_states for performance purpose.
        With HMC, each get_info would result a k2operator call of
        LogicalPartition/<uuid> GET. The data is cached in k2operator as
        feed or single entry unmarshalled. In most cases, we have to search
        via lpars feed. When there are hundreds of LPARs, the unmarshalling
        process can take a while.
        To improve performance, we will get lpars feed unmarshalled first
        and pass it for each get_info.
        """
        is_hmc = False
        if CONF.powervm_mgr_type and CONF.powervm_mgr_type.lower() == 'hmc':
            is_hmc = True
        if not is_hmc:
            return super(PowerVCKvmHmcComputeManager, self)._sync_power_states(
                context)

        db_instances = objects.InstanceList.get_by_host(context,
                                                        self.host,
                                                        use_slave=True)

        num_vm_instances = self.driver.get_num_instances()
        num_db_instances = len(db_instances)

        if num_vm_instances != num_db_instances:
            LOG.warn(_("While synchronizing instance power states, found "
                       "%(num_db_instances)s instances in the database and "
                       "%(num_vm_instances)s instances on the hypervisor."),
                     {'num_db_instances': num_db_instances,
                      'num_vm_instances': num_vm_instances})

        def _sync(db_instance):
            # NOTE(melwitt): This must be synchronized as we query state from
            #                two separate sources, the driver and the database.
            #                They are set (in stop_instance) and read, in sync.
            @utils.synchronized(db_instance.uuid)
            def query_driver_power_state_and_sync():
                self._pvc_query_driver_power_state_and_sync(context,
                                                            db_instance,
                                                            lpars_feed=lpars)

            try:
                # this if stanza is moved from
                # _query_driver_power_state_and_sync for performance purpose
                if db_instance.task_state is not None:
                    LOG.info(_("During sync_power_state the instance has a "
                               "pending task (%(task)s). Skip."),
                             {'task': db_instance.task_state},
                             instance=db_instance)
                else:
                    query_driver_power_state_and_sync()
            except Exception:
                LOG.exception(_LE("Periodic sync_power_state task had an "
                                  "error while processing an instance."),
                              instance=db_instance)

            self._syncs_in_progress.pop(db_instance.uuid)

        lpars = self.driver.get_lpars_feed()
        for db_instance in db_instances:
            # process syncs asynchronously - don't want instance locking to
            # block entire periodic task thread
            uuid = db_instance.uuid
            if uuid in self._syncs_in_progress:
                LOG.debug('Sync already in progress for %s' % uuid)
            else:
                LOG.debug('Triggering sync for uuid %s' % uuid)
                self._syncs_in_progress[uuid] = True
                self._sync_power_pool.spawn_n(_sync, db_instance)

    def _pvc_query_driver_power_state_and_sync(self, context, db_instance,
                                               lpars_feed=None):
        # No pending tasks. Now try to figure out the real vm_power_state.
        try:
            vm_instance = self.driver.get_info(db_instance,
                                               lpars_feed=lpars_feed)
            vm_power_state = vm_instance.state
        except exception.InstanceNotFound:
            vm_power_state = power_state.NOSTATE
        # Note(maoy): the above get_info call might take a long time,
        # for example, because of a broken libvirt driver.
        try:
            self._sync_instance_power_state(context,
                                            db_instance,
                                            vm_power_state,
                                            use_slave=True)
        except exception.InstanceNotFound:
            # NOTE(hanlind): If the instance gets deleted during sync,
            # silently ignore.
            pass

    def set_host_dead(self, context, dead=True, target_host=None):
        if (self._is_managing_PowerKVM()):
            # for KVM, we can only enable remote restart if the host
            # service is dead
            raise exception.ComputeServiceInUse(host=self.host)
        else:
            self.driver.set_host_dead(context, self.host,
                                      dead, target_host)
            self.update_available_resource(context)
