# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from oslo_utils import timeutils

# Indicates how many previous times to store. Update this value
# to cache last 'n' utilization metric values.
CACHE_SIZE = 3


class MetricCache(object):
    '''
    This class implements the caching logic to store last CACHE_SIZE
    metrics for all instances.
    '''

    def __init__(self, cache_size=CACHE_SIZE):
        self._metric_cache = dict()
        self._cache_size = cache_size
        # This list stores key names for last CACHE_SIZE values.
        # The cache for each resource will store the utilization metric against
        # these time keys (t0, t1, t2 etc).
        self._time_cache_keys = (
            list(map((lambda x: 't' + str(x)),
                     [i for i in xrange(0, self._cache_size)])))

    def add_value(self, metric_data, metric_key):
        '''
        Add a new value to the list of values, pushing out the
        oldest if needed--i.e., it's full
        :param: metric_data: This is a dictionary with following attributes
        1. uuid: The ID of the resource. 2. metric_key.
        :param: metric_key: The name of the metric being cached.
        '''
        if metric_data:
            # Iterate over the metric list. Each item is a dict of uuid, name,
            # and metric key values.
            for metric_dict in metric_data:
                if metric_key not in metric_dict:
                    continue
                metric_val = metric_dict.get(metric_key)
                # Do not take None value. Manager sends a complete list of
                # dicts for all vms, some of which might not have util values.
                if metric_val is None:
                    continue
                uuid = metric_dict.get('uuid')
                resource_metric = dict(t0=dict(
                    metric=metric_dict.get(metric_key),
                    timestamp=timeutils.utcnow()))
                # Validate that we have values for this resource
                #  in our CPU metric cache.
                if uuid in self._metric_cache:
                    # Check if we have the dict w.r.t. metric_key
                    uuid_cache = self._metric_cache[uuid]
                    if metric_key in uuid_cache:
                        # Update the cache for the uuid.
                        self._update_metric_for_resource(
                            metric_dict, metric_key)
                    else:
                        # Cache had entry for uuid, but not for this metric.
                        uuid_metric_dict = {metric_key: resource_metric}
                        self._metric_cache[uuid].update(uuid_metric_dict)
                # If resource not present in cache, then create
                # new entry with t0.
                else:
                    uuid_metric_dict = {metric_key: resource_metric}
                    self._metric_cache.update({uuid: uuid_metric_dict})

    def _get_cached_time_dict(self, uuid, metric_key, t_x):
        '''
        This method returns cached metric as t_x.
        '''
        if uuid in self._metric_cache:
            uuid_metric_dict = self._metric_cache[uuid]
            if metric_key in uuid_metric_dict:
                return uuid_metric_dict.get(t_x)

    def _update_metric_for_resource(self, metric_dict, metric_key):
        '''
        This method updates the cache for specific resource.
        '''
        uuid = metric_dict.get('uuid')
        t_now = dict(metric=metric_dict.get(metric_key),
                     timestamp=timeutils.utcnow())
        # At this point, we know that cache has entry for uuid and metric for
        # the uuid.
        uuid_metric_dict = self._metric_cache[uuid][metric_key]
        for t_x in self._time_cache_keys:
            time_x_dict = uuid_metric_dict.get(t_x)
            if not time_x_dict:
                resource_metric = {t_x: t_now}
                uuid_metric_dict.update(resource_metric)
                return
        # If we are here, then it means the cache already had
        # previous time values. Shift the time values and update
        # the current time utilization.
        for i in xrange(0, self._cache_size - 1):
            uuid_metric_dict['t' + str(i)] = uuid_metric_dict['t' + str(i + 1)]
        uuid_metric_dict['t' + str(self._cache_size - 1)] = t_now

    def get_average_for_instance(self, uuid):
        '''
        Computes the average of the current list of cached values
        for an instance.
        '''
        util = None
        uuid_avg_metrics = dict()
        if uuid not in self._metric_cache:
            return util
        uuid_cache = self._metric_cache[uuid]
        for metric_key in uuid_cache:
            cache_len = len(uuid_cache[metric_key])
            util = 0.0
            for time_key in self._time_cache_keys[:cache_len]:
                util += float(uuid_cache[metric_key][time_key]['metric'])
            util = util / cache_len
            # Return averages accurate to two decimal points
            avg_val = float("%.2f" % util)
            uuid_avg_metrics.update({metric_key: avg_val})
        return uuid_avg_metrics

    def get_average(self):
        '''
        Computes the average of the current list of cached values.
        '''
        util_dict = dict()
        for uuid in self._metric_cache:
            util = self.get_average_for_instance(uuid)
            util_dict.update({uuid: util})
        return util_dict
