# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from oslo_log import log as logging

from nova.compute.monitors.cpu.virt_driver import Monitor

from powervc_nova import _

LOG = logging.getLogger(__name__)


class PowerVCMonitor(Monitor):
    """
    PowerVC's version of the virt monitor.  This is needed to ensure that the
    cpu.percent metric values can't exceed 100 or be less than 0 for PowerVM,
    which can occur during sample periods that span a live migration operation.
    """

    def __init__(self, *args, **kwargs):
        super(PowerVCMonitor, self).__init__(*args, **kwargs)
        self._last_good_cpu_perc_value = 0

    def get_metrics(self):
        metrics = super(PowerVCMonitor, self).get_metrics()
        new_metrics = []

        for (name, value, timestamp) in metrics:
            value_prime = value
            if name == "cpu.percent":
                # On PowerVM, it is possible that the underlying cycle counts
                # lead to erroneous data points during mobility operations, so
                # we just "throw out" these data points and use the last known
                # good data point (since this is updated every minute, we can
                # tolerate these infrequent cases...).
                if not (0 <= value <= 100):
                    LOG.warning(_("Host's CPU utilization was evaluated to be "
                                  "%(new_val)d; this could be due to a VM "
                                  "migration to or from this host; returning "
                                  "last known valid CPU utilization value of "
                                  "%(last_val)d instead.")
                                % {'new_val': value,
                                   'last_val': self._last_good_cpu_perc_value})
                    # Adjust the value back to the last known good value.
                    value_prime = self._last_good_cpu_perc_value
                self._last_good_cpu_perc_value = value_prime
            # Build up our new set of metrics to return.
            new_metrics.append((name, value_prime, timestamp))

        return new_metrics
