# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
import copy
from eventlet import greenthread
import functools
import six
import sys

from cinderclient import exceptions as cinder_exception
from keystoneclient import exceptions as keystone_exception
from nova import exception
from nova import rpc
from nova import objects
from nova.objects import volume_usage
from nova.compute import utils as compute_utils
from nova.compute import task_states
from nova.compute.manager import ComputeManager as base_manager
from nova.compute.manager import reverts_task_state, wrap_instance_event,\
    wrap_instance_fault, errors_out_migration
from powervc_nova.compute import common_constants
from powervc_nova.compute import send_ui_notification
from powervc_nova.compute import check_host_maintenance_mode
from powervc_nova.compute import notify_messages as notify_msg

from nova.virt import driver
from nova.i18n import _LW, _LE
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import periodic_task
from oslo_utils import excutils
from powervc_nova.common import block_device_utils
from powervc_nova.compute import manager
from powervc_nova.objects import migrate_data as mig_obj
from powervc_nova import _

"""PowerVC Extensions to the Compute Manager for NovaLink-specific features"""

CONF = cfg.CONF
LOG = logging.getLogger(__name__)

get_notifier = functools.partial(rpc.get_notifier, service='compute')
wrap_exception = functools.partial(exception.wrap_exception,
                                   get_notifier=get_notifier)

# order of decorators should be same order  in which they
# appear if applied statically
CUSTOM_DECORATORS = {
    'check_can_live_migrate_destination':
    [manager.declare_decor('powervc_nova.compute.send_ui_notification',
                           notify_msg.LPM_ERROR,
                           migrate_op=True,
                           log_exception=True)],
    '_destroy_evacuated_instances':
    [manager.declare_decor('powervc_nova.compute.send_ui_notification',
                           notify_msg.SOURCE_EVACUATED_INST_CLEANUP_ERROR,
                           None, log_exception=True)],
    '_do_live_migration':
    [manager.declare_decor('powervc_nova.compute.send_ui_notification',
                           notify_msg.LPM_ERROR,
                           migrate_op=True,
                           log_exception=True)]
}


class PowerVCPvmComputeManager(manager.PowerVCComputeManager):
    """
    Extends the Compute Manager to provide the additional NovaLink-specific
    capabilities that are beyond what the OpenStack Community provides.
    """

    @classmethod
    def get_decorators(cls, decorators):
        new_decorators = copy.deepcopy(CUSTOM_DECORATORS)
        new_decorators.update(decorators)
        return super(PowerVCPvmComputeManager, cls).\
            get_decorators(new_decorators)

    def __init__(self, *args, **kwargs):
        """Performs the Additional NovaLink-specific Manager Initialization"""
        super(PowerVCPvmComputeManager, self).__init__(*args, **kwargs)

    @periodic_task.periodic_task(spacing=CONF.sync_power_state_interval,
                                 run_immediately=True)
    def _pvc_sync_power_states(self, context):
        super(PowerVCPvmComputeManager, self)._sync_power_states(context)

    @periodic_task.periodic_task
    @lockutils.synchronized('set_mobile_cores', 'powervc-')
    def pvc_update_available_resource(self, context):
        """
        We cannot monkey patch the periodic task decorator. So, we have to
        override here.
        """
        return super(PowerVCPvmComputeManager, self).\
            pvc_update_available_resource(context)

    def _shutdown_instance(self, context, instance,
                           bdms, requested_networks=None, notify=True,
                           try_deallocate_networks=True):
        # We are overriding this method to remove VM imported networks
        if requested_networks is None:
            network_info = compute_utils.get_nw_info_for_instance(instance)

        cleanup_imported = True
        try:
            self._pvm_shutdown_instance(
                context, instance, bdms, requested_networks=requested_networks,
                notify=notify, try_deallocate_networks=try_deallocate_networks)

        except exception.InstancePowerOffFailure:
            # if the instance can't power off, don't release the ip
            with excutils.save_and_reraise_exception():
                cleanup_imported = False
        finally:
            if cleanup_imported and requested_networks is None:
                self._cleanup_imported_networks(context,
                                                network_info)

    def _pvm_shutdown_instance(self, context, instance,
                               bdms, requested_networks=None, notify=True,
                               try_deallocate_networks=True):
        """Shutdown an instance on this host.

        :param:context: security context
        :param:instance: a nova.objects.Instance object
        :param:bdms: the block devices for the instance to be torn
                     down
        :param:requested_networks: the networks on which the instance
                                   has ports
        :param:notify: true if a final usage notification should be
                       emitted
        :param:try_deallocate_networks: false if we should avoid
                                        trying to teardown networking
        """
        # This method is a copy of the commnunity shutdown method and should
        # only exist until the community support for shared volume attach
        # arrives.
        context = context.elevated()
        LOG.info(_('%(action_str)s instance') %
                 {'action_str': 'Terminating'},
                 context=context, instance=instance)

        if notify:
            self._notify_about_instance_usage(context, instance,
                                              "shutdown.start")

        network_info = compute_utils.get_nw_info_for_instance(instance)

        # NOTE(vish) get bdms before destroying the instance
        vol_bdms = [bdm for bdm in bdms if bdm.is_volume]
        block_device_info = self._get_instance_block_device_info(
            context, instance, bdms=bdms)

        # NOTE(melwitt): attempt driver destroy before releasing ip, may
        #                want to keep ip allocated for certain failures
        try:
            self.driver.destroy(context, instance, network_info,
                                block_device_info)
        except exception.InstancePowerOffFailure:
            # if the instance can't power off, don't release the ip
            with excutils.save_and_reraise_exception():
                pass
        except Exception:
            with excutils.save_and_reraise_exception():
                # deallocate ip and fail without proceeding to
                # volume api calls, preserving current behavior
                if try_deallocate_networks:
                    self._try_deallocate_network(context, instance,
                                                 requested_networks)

        if try_deallocate_networks:
            self._try_deallocate_network(context, instance, requested_networks)

        for bdm in vol_bdms:
            try:
                # NOTE(vish): actual driver detach done in driver.destroy, so
                #             just tell cinder that we are done with it.
                connector = self.driver.get_volume_connector(instance)
                self.volume_api.terminate_connection(context,
                                                     bdm.volume_id,
                                                     connector)
                self.volume_api.detach(context.elevated(),
                                       bdm.volume_id,
                                       instance.uuid)
            except exception.DiskNotFound as exc:
                LOG.debug('Ignoring DiskNotFound: %s', exc,
                          instance=instance)
            except exception.VolumeNotFound as exc:
                LOG.debug('Ignoring VolumeNotFound: %s', exc,
                          instance=instance)
            except exception.InvalidVolume as exc:
                LOG.debug('Ignoring Invalid volume: %s', exc,
                          instance=instance)
            except (cinder_exception.EndpointNotFound,
                    keystone_exception.EndpointNotFound) as exc:
                LOG.warning(_LW('Ignoring EndpointNotFound: %s'), exc,
                            instance=instance)

        if notify:
            self._notify_about_instance_usage(context, instance,
                                              "shutdown.end")

    def _cleanup_volumes(self, context, instance_uuid, bdms, raise_exc=True):
        exc_info = None
        for bdm in bdms:
            if not self._bdm_has_attachment(bdm):
                if bdm.delete_on_termination:
                    volume_id = bdm.volume_id
                    try:
                        volume = self.volume_api.get(context, volume_id)
                        # if the volume is creating we are deleting a failed
                        # deploy. So we need to wait for the volume to
                        # finish creating before we delete it.
                        # If we leave it as 'creating' delete will fail.
                        # The volume create will eventually finish or fail
                        # so wait here.
                        # This is would not happen in the community but
                        # happens in Novalink because in Novalink we let
                        # volume creates run in parallel and the community
                        # does them serially.
                        if volume['status'] == 'creating':
                            self._await_block_device_map_created(context,
                                                                 volume_id)
                    except exception.VolumeNotFound as exc:
                        LOG.debug('Ignoring VolumeNotFound: %s', exc,
                                  volume=volume_id)
                    except Exception as exc:
                        exc_info = sys.exc_info()
                        LOG.warning(_LW('Failure when awating volume creation '
                                        'to complete before deleting : '
                                        '%(volume_id)s due to %(exc)s'),
                                    {'volume_id': bdm.volume_id, 'exc': exc})
        #if there was an exception when we waited for volume creation to
        # complete and the caller asks us to raise exception do so.
        if exc_info is not None:
            if raise_exc:
                six.reraise(exc_info[0], exc_info[1], exc_info[2])
            else:
                LOG.debug("raise_exc value is %s cleanup_volumes will not"
                          " proceed" % raise_exc)
        else:
            base_manager._cleanup_volumes(
                self, context, instance_uuid, bdms, raise_exc)

    def _destroy_evacuated_instances(self, context):
        """Override to send UI notification."""
        super(PowerVCPvmComputeManager, self)._destroy_evacuated_instances(
            context)

    def _try_deallocate_network(self, context, instance,
                                requested_networks=None):
        try:
            super(PowerVCPvmComputeManager, self).\
                _try_deallocate_network(context, instance,
                                        requested_networks=requested_networks)
        except Exception as ex:
            # Catch the exception and don't re-throw. This allows the PowerVC
            # behavior of allowing shutdown_instance
            # to continue to BDM cleanup and possibly reschedule
            # when network deallocation fails.
            # It is debateable whether the current functionality is a bug
            # in the community but we will discuss with them.  Until that
            # discussion is complete we will keep historical PowerVC
            # behavior here to allow the cleanup of the Cinder volumes.
            # The error has been logged in the parent _try_deallocate_network.
            reason = (_("Failed to remove the VLAN associated with "
                        "the LPAR virtual ethernet. Please check "
                        "host log for further details. Error: %(ex)s")
                      % dict(ex=ex))
            self._error_notify(instance, context, 'warn', reason)
            LOG.warn(_("Failed to remove the VLAN associated with "
                       "LPAR virtual ethernet. Instance: %(inst)s. "
                       "Error:%(error)s") %
                     {'lpar_name': instance['name'],
                      'inst': instance['uuid'],
                      'error': ex})

    def _prep_block_device_volume_attach(self, context, instance, virt_bdms,
                                         do_check_attach):
        # The prep_block device flow has now cloned the image volumes and
        # created any blank volumes requested.  We wait for them to move
        # to the available state and then attach them.
        # Wait on any virt_bdms that aren't already in-use (shared volumes)
        # or available (existing volumes that are being attached or volumes
        # that already finished their creation.
        bdms_to_wait_on = []
        for bdm in virt_bdms:
            volume = self.volume_api.get(context, bdm.volume_id)
            if volume['status'] not in ['available', 'in-use']:
                bdms_to_wait_on.append(bdm)

        for bdm in bdms_to_wait_on:
            self._await_block_device_map_created(context, bdm.volume_id)

        # The following sort and attach is commented out until we get
        # community bug https://bugs.launchpad.net/nova/+bug/1488111 fixed.
        # This bug prevents deploy reschedule from happening when
        # initialize_connection on volumes fails.
        # Until that is fixed this attach is being done in the virt driver
        # spawn method.
        # sort the BDMs before attach since some Linux configurations are
        # impacted by disk order
        # sorted_virt_bdms = block_device_utils.sort_virt_bdms(virt_bdms)
        # blk_dev.attach_block_devices(sorted_virt_bdms, context, instance,
        #                             self.volume_api, self.driver,
        #                             do_check_attach=do_check_attach)

        # blkdev_info = self._get_instance_block_device_info(context,
        #                                                   instance)

        # The following way of setting up blk_dev info should be removed
        # when the above code is commented out.
        bdms = objects.BlockDeviceMappingList.\
            get_by_instance_uuid(context, instance.uuid)
        blkdev_info = driver.get_block_device_info(instance, bdms)

        # sort the BDMs in the returned list since it will be used downstream
        # for attach in the virt driver.
        blkdev_info['block_device_mapping'] = block_device_utils.\
            sort_virt_bdms(blkdev_info['block_device_mapping'])
        return blkdev_info

    @wrap_exception()
    @manager.pvc_comp_mgr_reverts_task_state
    @send_ui_notification(None,
                          notify_msg.ATTACH_ERROR,
                          notify_msg.ATTACH_SUCCESS,
                          log_exception=True)
    @check_host_maintenance_mode(None, common_constants.OPER_AT_VOL)
    def attach_volume(self, context, instance, bdm):
        """Attach a volume to an instance"""
        # Instance must be passed as a keyword argument in order for
        # revert_task_state decorator to function properly
        super(PowerVCPvmComputeManager, self).attach_volume(context,
                                                            instance=instance,
                                                            bdm=bdm)
        # for a vscsi-npiv vm we need to call instance save explicitly
        # to allow the instance meta to be saved with the fabric mappings
        instance.save()

    def _get_host_name(self, vios_name, connection_type):
        """ Returns the vios name minus the vios number
        Eg. for vios_name '828642A_10D6C7T_2',
        the host name returned will be '828642A_10D6C7T'
        """
        if connection_type == 'pv_vscsi':
            if vios_name is not None:
                last_index = vios_name.rfind("_")
                host_name = vios_name[:last_index]
        else:
            host_name = vios_name
        return host_name

    def _detach_volume(self, context, volume_id, instance, destroy_bdm=True,
                       attachment_id=None):
        """override community version to take care of vscsi remote restart
         rtc #51826
        """
        bdm = objects.BlockDeviceMapping.get_by_volume_and_instance(
            context, volume_id, instance.uuid)
        if CONF.volume_usage_poll_interval > 0:
            vol_stats = []
            mp = bdm.device_name
            # Handle bootable volumes which will not contain /dev/
            if '/dev/' in mp:
                mp = mp[5:]
            try:
                vol_stats = self.driver.block_stats(instance, mp)
            except NotImplementedError:
                pass

            if vol_stats:
                LOG.debug("Updating volume usage cache with totals",
                          instance=instance)
                rd_req, rd_bytes, wr_req, wr_bytes, flush_ops = vol_stats
                vol_usage = objects.VolumeUsage(context)
                vol_usage.volume_id = volume_id
                vol_usage.instance_uuid = instance.uuid
                vol_usage.project_id = instance.project_id
                vol_usage.user_id = instance.user_id
                vol_usage.availability_zone = instance.availability_zone
                vol_usage.curr_reads = rd_req
                vol_usage.curr_read_bytes = rd_bytes
                vol_usage.curr_writes = wr_req
                vol_usage.curr_write_bytes = wr_bytes
                vol_usage.save(update_totals=True)
                self.notifier.info(context, 'volume.usage',
                                   compute_utils.usage_volume_info(vol_usage))

        connection_info = self._driver_detach_volume(context, instance, bdm)
        connector = self.driver.get_volume_connector(instance)

        if connection_info and not destroy_bdm and (
           connector.get('host') != instance.host):
            # If the volume is attached to another host (evacuate) then
            # this connector is for the wrong host. Use the connector that
            # was stored in connection_info instead (if we have one, and it
            # is for the expected host).
            stashed_connector = connection_info.get('connector')
            stashed_connector_host = None
            #for onboard vscsi case look for host name in vscsi_target_lun dict
            if not stashed_connector:
                stashed_connector = self.__build_conn_for_onboard_vscsi(
                    connection_info)
            # set stashed_connector host to None as default
            if stashed_connector and\
                    'host' in stashed_connector.keys() and\
                    stashed_connector.get('host'):
                stashed_connector_host = stashed_connector.get('host')
            if not stashed_connector:
                # Volume was attached before we began stashing connectors
                LOG.warning(_LW("Host mismatch detected, but stashed "
                                "volume connector not found. Instance host is "
                                "%(ihost)s, but volume connector host is "
                                "%(chost)s."),
                            {'ihost': instance.host,
                             'chost': connector.get('host')})
            elif stashed_connector_host is None or\
                    (instance.host != self._get_host_name(
                        stashed_connector_host,
                        stashed_connector.get('connection_type'))
                        and not self._is_onboard_vscsi(
                            connection_info, instance)):
                # Unexpected error. The stashed connector is also not matching
                # the needed instance host.
                LOG.error(_LE("Host mismatch detected in stashed volume "
                              "connector. Will use local volume connector. "
                              "Instance host is %(ihost)s. Local volume "
                              "connector host is %(chost)s. Stashed volume "
                              "connector host is %(schost)s."),
                          {'ihost': instance.host,
                           'chost': connector.get('host'),
                           'schost': stashed_connector_host})
            else:
                # Fix found. Use stashed connector.
                LOG.debug("Host mismatch detected. Found usable stashed "
                          "volume connector. Instance host is %(ihost)s. "
                          "Local volume connector host was %(chost)s. "
                          "Stashed volume connector host is %(schost)s.",
                          {'ihost': instance.host,
                           'chost': connector.get('host'),
                           'schost': stashed_connector_host})
                data = connection_info.get('data')
                vscsi_data = data.get('vios_data') if data else None

                if not vscsi_data:
                    #for onboard vm vscsi_target_lun is used
                    vscsi_data = connection_info.get('vscsi_target_lun')
                    data['vscsi_target_lun'] = vscsi_data

                if vscsi_data:
                    stashed_connector = data
                LOG.debug("stashed_connector %s " % stashed_connector)
                connector = stashed_connector

        self.volume_api.terminate_connection(context, volume_id, connector)

        if destroy_bdm:
            bdm.destroy()

        info = dict(volume_id=volume_id)
        self._notify_about_instance_usage(
            context, instance, "volume.detach", extra_usage_info=info)
        self.volume_api.detach(context.elevated(), volume_id, instance.uuid,
                               attachment_id)

    def __build_conn_for_onboard_vscsi(self, connection_info):
        """
        the connection info for onboard vm looks like
        {u'driver_volume_type': u'fibre_channel', u'data':
        {u'vendor': u'', u'target_discovered': False, u'qos_spec': None,...
        <truncating>..
         u'vscsi_target_lun': [{u'source_wwn': [u'10000090FA454137'],
         u'target_lun': u'7', u'vios_host': u'P8_55_vios2',
         'target_wwn': [u'5005076802264658', u'5005076802264657']},
         ], u'volume_connector': {u'connector': {u'connection_type':
         [u'pv_vscsi'], u'host': u'RR_vn1-25c15e4f-00000016'},
         u'connection-type': u'pv_vscsi'}}
        """
        connector = dict()
        vscsi_target_lun = connection_info.get('vscsi_target_lun', None)
        if vscsi_target_lun:
            connector['host'] = vscsi_target_lun[0].get('vios_host')
            connector['connection-type'] = 'pv_vscsi'
        return connector

    def _is_onboard_vscsi(self, connection_info, instance):
        """
        If this is onboard VM and operation is remote restart
        """
        vscsi_target_lun = connection_info.get('vscsi_target_lun', None)
        if vscsi_target_lun and instance.task_state == task_states.REBUILDING:
            return True
        else:
            return False

    def rebuild_instance(self, context, instance, orig_image_ref, image_ref,
                         injected_files, new_pass, orig_sys_metadata,
                         bdms, recreate, on_shared_storage=None,
                         preserve_ephemeral=False, migration=None,
                         scheduled_node=None, limits=None):
        """Destroy and re-make this instance.

        Handles sending an UI notification during evacuation. No UI
        notification is sent during a rebuild server call.
        """
        @send_ui_notification(None, notify_msg.EVACUATE_ERROR,
                              notify_msg.EVACUATE_SUCCESS, log_exception=True)
        def _recreate_rebuild(context, instance, orig_image_ref, image_ref,
                              injected_files, new_pass, orig_sys_metadata,
                              bdms, recreate, on_shared_storage,
                              preserve_ephemeral):
            # Perform evacuate action
            super(PowerVCPvmComputeManager, self).\
                rebuild_instance(context, instance=instance,
                                 orig_image_ref=orig_image_ref,
                                 image_ref=image_ref,
                                 injected_files=injected_files,
                                 new_pass=new_pass,
                                 orig_sys_metadata=orig_sys_metadata,
                                 bdms=bdms, recreate=recreate,
                                 on_shared_storage=on_shared_storage,
                                 preserve_ephemeral=preserve_ephemeral)

        if recreate:
            # Perform evacuate action
            _recreate_rebuild(context, instance, orig_image_ref, image_ref,
                              injected_files, new_pass, orig_sys_metadata,
                              bdms, recreate, on_shared_storage,
                              preserve_ephemeral)
        else:
            # Perform rebuild action
            super(PowerVCPvmComputeManager, self).\
                rebuild_instance(context, instance=instance,
                                 orig_image_ref=orig_image_ref,
                                 image_ref=image_ref,
                                 injected_files=injected_files,
                                 new_pass=new_pass,
                                 orig_sys_metadata=orig_sys_metadata,
                                 bdms=bdms, recreate=recreate,
                                 on_shared_storage=on_shared_storage,
                                 preserve_ephemeral=preserve_ephemeral)

    #######################################################
    #        Resize-related Overrides/Extensions          #
    #######################################################
    @wrap_exception()
    @reverts_task_state
    @wrap_instance_event
    @errors_out_migration
    @wrap_instance_fault
    @send_ui_notification(None,
                          notify_msg.RESIZE_ERROR, None,
                          log_exception=True)
    @check_host_maintenance_mode(None, common_constants.OPER_RESIZE)
    def resize_instance(self, context, instance, image,
                        reservations, migration, instance_type,
                        clean_shutdown):
        """Starts the migration of a running instance to another host."""
        super(PowerVCPvmComputeManager, self).\
            resize_instance(context, instance, image, reservations, migration,
                            instance_type, clean_shutdown)

    @wrap_exception()
    @send_ui_notification(None,
                          notify_msg.RESIZE_ERROR, notify_msg.RESIZE_SUCCESS,
                          log_exception=True)
    def confirm_resize(self, context, instance, reservations, migration):

        super(PowerVCPvmComputeManager, self).\
            confirm_resize(context, instance, reservations, migration)
        try:
            LOG.info(_("Refreshing statistics for the  %s instance.")
                     % instance.name)
            db_inst = self._query_db_instances(context, [instance])
            self.driver.inventory_instances(context, db_inst)
        except Exception as exc:
            LOG.warn(_('Failed to refresh instance inventory '))
            LOG.exception(exc)

    #######################################################
    #        Migrate-related Overrides/Extensions          #
    #######################################################
    # Overriding to revert task state
    @reverts_task_state
    def _post_live_migration(self, ctxt, instance,
                             dest, block_migration=False, migrate_data=None):
        """Post operations for live migration.

        This method is called from live_migration
        and mainly updating database record.

        :param ctxt: security context
        :param instance: instance dict
        :param dest: destination host
        :param block_migration: if true, prepare for block migration
        :param migrate_data: if not None, it is a dict which has data
        required for live migration without shared storage

        """
        LOG.info(_('_post_live_migration() is started..'),
                 instance=instance)
        super(PowerVCPvmComputeManager, self).\
            _post_live_migration(ctxt, instance,
                                 dest, block_migration, migrate_data)

    @reverts_task_state
    @send_ui_notification(None,
                          notify_msg.LPM_ERROR,
                          notify_msg.LPM_SUCCESS,
                          migrate_op=True,
                          log_exception=True)
    def post_live_migration_at_destination(self, context, instance,
                                           block_migration, migrate_data):
        """Post operations for live migration .

        :param context: security context
        :param instance: Instance dict
        :param block_migration: if true, prepare for block migration

        """
        # Overriding for the sake of migrate_data
        # Utilize this variable for further assignment in future
        super(PowerVCPvmComputeManager, self).\
            post_live_migration_at_destination(context, instance,
                                               block_migration)

    @reverts_task_state
    def rollback_live_migration_at_destination(self, context, instance,
                                               destroy_disks, migrate_data):
        """Cleaning up image directory that is created pre_live_migration.

        :param context: security context
        :param instance: a nova.objects.instance.Instance object sent over rpc
        """
        # Overriding for the sake of reverting the task state if it
        # fails during roll back live migration
        super(PowerVCPvmComputeManager, self).\
            rollback_live_migration_at_destination(context, instance,
                                                   destroy_disks,
                                                   migrate_data)

    def change_vif_sea(self, context, data):
        """
        This API passes the request to move a vif to another sea on to the
        driver.  This is done in a greenthread so we can return to the api
        invoker immediately.

        :param context: Context Object
        :param data: The dictionary object containing information to call
        the plug operation, sent by the API service.
        :return: Calls the compute driver and returns the plug call back.
        """
        greenthread.spawn(self.driver.change_vif_sea, context, data)
        return True

    def sync_enterprise_pool_for_host(self, context):
        try:
            return self.driver.sync_enterprise_pool_for_host(context)
        except Exception as exc:
            with excutils.save_and_reraise_exception():
                LOG.error(_('Failed to sync enterprise pool for host.'))
                LOG.exception(exc)

    @lockutils.synchronized('set_mobile_cores', 'powervc-')
    def set_mobile_cores(self, context, desired_mobile_cores):
        try:
            return self.driver.set_mobile_cores(context, desired_mobile_cores)
        except Exception as exc:
            with excutils.save_and_reraise_exception():
                LOG.error(_('Error setting mobile cores for host to %s.') %
                          desired_mobile_cores)
                LOG.exception(exc)
