# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import ConfigParser
import os
from powervc_nova import _
from powervc_oslo.config import config_utils as cutils
from powervc_oslo.config import data_mgr as mgr
from powervc_oslo.config import data_utils as oslo_utils
import re
import sys


def _find_matching_hosts(host):
    """Finds the name of the hosts that match the value specified on input

    :param host: An optional host name or host_display_name to find.  If not
                 specified, or specified as '[global]', all hosts will be
                 returned.
    :returns hosts: A list of dictionaries with host info from the conf files.
    """
    hosts = list()
    # If they didn't specify a host, use a keyword for the global value
    if host is None or host == '[global]':
        hosts.append({'host_name': '[global]'})
    # Loop through each of the nova-xxx configuration files in the directory
    for filename in os.listdir('/etc/nova'):
        if filename == 'nova-health.conf':
            continue
        # Parse out any of the nova-xxx configuration files from the directory
        if filename.startswith('nova-') and filename.endswith('.conf'):
            hostname = filename[5: len(filename) - 5]
            # Load and parse this host conf file
            host_dict = _parse_conf_to_dict(filename, hostname)
            host_display_name = host_dict.get('host_display_name')
            # If they asked for all or matched the input, return this one
            if(host is None or host == hostname or host == host_display_name):
                hosts.append(host_dict)
    # If there weren't any matching hosts, then exit return an error
    if len(hosts) <= 0:
        msg = _('The operation failed because host \'%s\' was not found. '
                'Specify a different host and retry the operation.') %\
            str(host)
        cutils.safe_print(msg)
        sys.exit(11)
    return hosts


def _parse_conf_to_dict(filename, host_name):
    """Helper method to parse data from a conf file and put it in a dictionary.

    :param filename: The filename of the conf file to parse.
    :param host_name: The nova name of the host (not to be confused with the
                      networking concept of hostname).
    :returns host_dict: A dictionary with information from the specified conf
                        file.
    """
    # Load and parse the conf file
    parser = ConfigParser.RawConfigParser()
    parser.read('/etc/nova/' + filename)

    # Create the return dictionary and add the specified host_name
    host_dict = {'host_name': host_name}

    # If host_display_name is in the conf file, put it in the dictionary
    if 'host_display_name' in parser.defaults():
        host_display_name = parser.get('DEFAULT', 'host_display_name')
        host_dict['host_display_name'] = host_display_name

    # Load dictionary attributes based on the host type
    if 'host_type' in parser.defaults():
        host_type = parser.get('DEFAULT', 'host_type')
        if host_type:
            host_dict['host_type'] = host_type
        if host_type == 'kvm':
            # For KVM hosts, load the IP, user, password, and priavte key
            if 'kvm_host_ip' in parser.defaults():
                host_ip = parser.get('DEFAULT', 'kvm_host_ip')
                if host_ip:
                    host_dict['ip'] = host_ip
            if 'kvm_host_user' in parser.defaults():
                host_user = parser.get('DEFAULT', 'kvm_host_user')
                if host_user:
                    host_dict['user'] = host_user
            if 'kvm_host_passwd' in parser.defaults():
                host_password = parser.get('DEFAULT', 'kvm_host_passwd')
                if host_password:
                    host_password = _decrypt(host_password)
                    host_dict['passwd'] = host_password
            if 'kvm_host_private_key' in parser.defaults():
                host_key = parser.get('DEFAULT', 'kvm_host_private_key')
                if host_key:
                    host_dict['private_key'] = host_key
        elif host_type == 'powervm':
            # For Novalink hosts, load the IP, user, password, and priavte key
            if 'powervm_mgr_type' in parser.defaults():
                powervm_mgr_type = parser.get('DEFAULT', 'powervm_mgr_type')
                if powervm_mgr_type:
                    host_dict['powervm_mgr_type'] = powervm_mgr_type
                if powervm_mgr_type == 'pvm':
                    if 'powervm_host_ip' in parser.defaults():
                        host_ip = parser.get('DEFAULT', 'powervm_host_ip')
                        if host_ip:
                            host_dict['ip'] = host_ip
                    if 'powervm_host_user' in parser.defaults():
                        host_user = parser.get('DEFAULT', 'powervm_host_user')
                        if host_user:
                            host_dict['user'] = host_user
                    if 'powervm_host_passwd' in parser.defaults():
                        host_password = parser.get('DEFAULT',
                                                   'powervm_host_passwd')
                        if host_password:
                            host_password = _decrypt(host_password)
                            host_dict['passwd'] = host_password
                    if 'powervm_host_private_key' in parser.defaults():
                        host_key = parser.get('DEFAULT',
                                              'powervm_host_private_key')
                        if host_key:
                            host_dict['private_key'] = host_key
    return host_dict


def _decrypt(text):
    """Convert an encrypted string to a clear text one.

    :param text: Encrypted text, usually from a conf file.
    :returns text: Un-encrypted text.
    """
    if text is None:
        return None
    encrypt_handler = _get_encrypt_handler()
    if encrypt_handler:
        return encrypt_handler.decode(text)
    return text


def _get_encrypt_handler():
    """
    Retrieve the EncryptHandler in a safe fashion that protects against it
    not being available.

    :returns encrypt_handler: The EncryptHandler() or None if it is not
                              available.  This will return None when run on
                              hosts that do not have EncryptHandler available.
    """
    try:
        __import__('powervc_keystone.encrypthandler')
    except ImportError:
        return None
    else:
        from powervc_keystone.encrypthandler import EncryptHandler
        return EncryptHandler()


def _print_domain_names(host_dicts):
    """Prints the dhcp_domain attribute from the nova.conf files"""
    # Loop through each of the matching hosts, find the domain name
    dhcp_domain_list = []
    for host_dict in host_dicts:
        host = host_dict['host_name']
        dhcp_domain = None
        # If it was the global keyword, just use nova.conf for the file name
        prefix = 'nova' if host == '[global]' else ('nova-' + host)
        parser = ConfigParser.RawConfigParser()
        parser.read('/etc/nova/' + prefix + '.conf')
        # Parse the dhcp_domain attribute out of the configuration file
        if 'dhcp_domain' in parser.defaults():
            dhcp_domain = parser.get('DEFAULT', 'dhcp_domain')
        # We will follow the domainname command and do (none) for not set
        dhcp_domain = '(none)' if dhcp_domain is None else dhcp_domain
        dhcp_domain_list.append('%-22s %s' % (host, dhcp_domain))
    # Print out each domain name
    for dhcp_domain in dhcp_domain_list:
        print dhcp_domain
    return dhcp_domain_list


def _update_domain_names(host_dicts, dhcp_domain, restart):
    """Updates the dhcp_domain attribute in the nova.conf files

    :param host_dicts: A list of dictionaries with conf file attributes from
                       each host.
    :param dhcp_domain: The new dhcp_domain value to set in the conf files.
    :param restart: A boolean indicating whether to restart the nova services
                    after updating the conf files.
    """
    expr, dhcp_parts = re.compile('^[A-Za-z0-9_-]+$'), dhcp_domain.split(".")
    # Verify that the Domain Name has only valid characters in it
    # We will allow it to be set to an empty string which will make it None
    if len(dhcp_parts[0]) > 0 and not all(expr.match(x) for x in dhcp_parts):
        msg = _('The operation failed because invalid characters were '
                'specified for the domain name. Valid characters are a-z, '
                'A-Z, 0-9, periods (.), dashes (-), and underscores (_). '
                'Specify a domain name with only valid characters and '
                'retry the operation.')
        cutils.safe_print(msg)
        sys.exit(12)
    # Loop through each of the matching hosts, updating the domain name
    for host_dict in host_dicts:
        host = host_dict['host_name']
        # If it was the global keyword, just use nova.conf for the file name
        prefix = 'nova' if host == '[global]' else ('nova-' + host)
        conf_file = '/etc/nova/' + prefix + '.conf'
        # Execute the openstack-config command to update the attribute
        oslo_utils.set_openstack_config_value(conf_file, 'dhcp_domain',
                                              dhcp_domain, 'DEFAULT')
        # If the host is PowerKVM or Novalink, update the remote conf files too
        if (host_dict.get('host_type') == 'kvm' or
            (host_dict.get('host_type') == 'powervm' and
             host_dict.get('powervm_mgr_type') == 'pvm')):
            cmd = '/usr/bin/openstack-config --set --inplace ' \
                  '/etc/nova/nova.conf DEFAULT dhcp_domain %s' % dhcp_domain
            requires_root = host_dict.get('powervm_mgr_type') == 'pvm'
            oslo_utils.run_ssh_command(ipcreds=host_dict, command=cmd,
                                       requires_root=requires_root)
            if restart:
                msg = _('Restarting the nova service on %s') % host
                cmd = '/opt/ibm/powervc/bin/powervc-services nova restart'
                oslo_utils.run_ssh_command(ipcreds=host_dict, command=cmd,
                                           requires_root=requires_root)


def server_domain(args):
    '''
    Display or set, in the nova.conf files, the default virtual machine
    domain name
    '''
    # Figure out which of the hosts on the system match the one provided
    host_dicts = _find_matching_hosts(args.host_name)
    # If they provided a domain name, then update the configuration files
    if args.domain is not None:
        _update_domain_names(host_dicts, args.domain, args.restart)
        if not args.restart:
            msg = _('The default domain name was updated for the specified '
                    'hosts. Verify that no tasks are in-progress, then run '
                    'the \'/opt/ibm/powervc/bin/powervc-services nova '
                    'restart\' command to restart the nova services and '
                    'complete the operation.')
            cutils.safe_print(msg)
        else:
            service = 'nova'
            msg = _('Restarting the %s service') % service
            cutils.safe_print(msg)
            mgr.restart_services(service)
    # If they didn't specify a domain name, then just print out the values
    else:
        _print_domain_names(host_dicts)


def setup_parser(nova_subp, service_name):
    parser = nova_subp.add_parser(
        'server-domain',
        help=_('Display or set the default virtual machine domain name'),
        description=_('The default domain name is set on virtual machines '
                      'that PowerVC deploys either on individual managed '
                      'hosts or on a global level. If no domain name is '
                      'specified, the command displays the currently '
                      'configured domain name values.'))
    parser.add_argument('--set',
                        dest='domain',
                        help=_('Specifies the domain name to use as the '
                               'default for deployed virtual machines'),
                        action='store')
    parser.add_argument('--host',
                        dest='host_name',
                        help=_('Specifies the name of the specific host to '
                               'which the default domain name should apply'),
                        action='store')
    parser.add_argument('--restart',
                        dest='restart',
                        help=_('Restart the nova compute processes after '
                               'updating the config files'),
                        action='store_true')
    parser.set_defaults(func=server_domain)
