#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
"""Table Creation/Deletion for Schema changes in the PowerVC 1.2.2 Release"""

import uuid
from sqlalchemy import String, Table, MetaData, Boolean, Column, Integer, Text
from sqlalchemy.sql import bindparam, expression
from oslo_log import log as logging
from powervc_nova.db.sqlalchemy import models

LOG = logging.getLogger(__name__)

NETWORK_ID_COLUMNS = [Column('id', String(36))]

STORAGE_FC_COLUMNS = [
    Column('fc_protocol', String(31)),
    Column('max_vfcs', Integer),
    Column('mapped_vfcs', Integer),
    Column('reserved_vfcs', Integer),
    Column('avail_active_vfcs', Integer)
]

COMPUTE_VIOS_COLUMNS = [
    Column('id', String(36)),
    Column('backend_uuid', String(36))
]

STORAGE_SCG_COLUMNS = [
    Column('vios_redundancy', Integer),
    Column('exact_redundancy', Boolean),
    Column('npiv_fabric', String(255)),
    Column('boot_connectivity', String(255)),
    Column('data_connectivity', String(255)),
    Column('backend_host_group', String(255))
]

TABLE_COLUMN_UPDATES = [
    ('storage_connectivity_group', STORAGE_SCG_COLUMNS),
    ('host_network_association', NETWORK_ID_COLUMNS),
    ('adapter_ethernet_virtual', NETWORK_ID_COLUMNS),
    ('adapter_ethernet_shared', NETWORK_ID_COLUMNS),
    ('server_vio', COMPUTE_VIOS_COLUMNS), ('port_fc', STORAGE_FC_COLUMNS),
    ('instance_power_specs', [Column('vnc_info', Text)])
]


def upgrade(migration_engine):
    """ Upgrades the Nova DB Tables to include those for PowerVC 1.2.2 """
    metadata = MetaData(migration_engine)
    metadata.reflect(migration_engine)
    metadata.bind = migration_engine
    # Loop through all of the DTO's that we added new in PowerVC 1.2.2
    # creating the tables off of the definition defined in the DTO
    for dto in models.POWERVC_V1R2M2_DTOS:
        table = dto.__table__.tometadata(metadata, None)
        table.create(checkfirst=True)
    # Loop through each of the tables we need to add columns to
    for column_update in TABLE_COLUMN_UPDATES:
        table_name, table_columns = column_update
        # We need to add the additional columns to the given table
        atable = Table(table_name, metadata, autoload=True)
        for column in table_columns:
            # Check first to see if the Column exists on the Table already
            if atable._columns.get(column.name) is None:
                atable.create_column(column.copy())
    # Now we will generate ID values for each entries in several tables
    _generate_uuid_values(migration_engine, metadata)
    # Next we can go through and add the default SCG column values
    _add_default_scg_values(migration_engine, metadata)


def downgrade(migration_engine):
    """ Downgrades the Nova DB Tables to remove those for PowerVC 1.2.2 """
    metadata = MetaData(migration_engine)
    metadata.reflect(migration_engine)
    metadata.bind = migration_engine
    # Loop through all of the DTO's that we added new in PowerVC 1.2.2
    # dropping the tables off of the definition defined in the DTO
    for dto in reversed(models.POWERVC_V1R2M2_DTOS):
        dto.__table__.metadata = metadata
        dto.__table__.drop(checkfirst=True)


def _generate_uuid_values(migration_engine, metadata):
    """ Helper method to generate the UUID values for several tables """
    namespc = uuid.UUID('a0dd4880-6115-39d6-b26b-77df18fe749f')
    table_names = ['server_vio', 'host_network_association',
                   'adapter_ethernet_shared', 'adapter_ethernet_virtual']
    # Load the Compute Node and Service table so that we can use later
    svc_table = Table('services', metadata, autoload=True)
    node_table = Table('compute_nodes', metadata, autoload=True)
    svc_query = expression.select(
        [svc_table.c.host]).where(svc_table.c.id == bindparam('svcid'))
    node_query = expression.select(
        [node_table.c.service_id]).where(node_table.c.id == bindparam('cnid'))
    # Loop through each of the tables generating the UUID for the ID Column
    for table_name in table_names:
        db_table = Table(table_name, metadata, autoload=True)
        db_query = expression.select([db_table])
        db_rows = migration_engine.execute(db_query).fetchall()
        db_update = db_table.update().where(
            db_table.c.pk_id == bindparam('pkid')).values(id=bindparam('uuid'))
        # Loop through each of the Rows updating the ID to a generated Value
        for row in db_rows:
            # If the ID column is NULL, we need to generate an ID value to use
            if not row['id']:
                auuid = str(uuid.uuid4())
                # If this is the NetworkAssoc table then can generate correctly
                if table_name == 'host_network_association':
                    # First query the host_name from the Service table
                    svcid = migration_engine.execute(
                        node_query, cnid=row['compute_node_id']).first()
                    host = migration_engine.execute(
                        svc_query, svcid=svcid[0]).first()
                    # Then Generate the UUID for the Network Association
                    namestr = "ibm-powervc://NetworkAssociationFactory/"
                    namestr += "host_name='" + str(host[0]) + "',neutron_net_"
                    namestr += "id='" + str(row['neutron_network_id']) + "'"
                    auuid = str(uuid.uuid3(namespc, str(namestr)))
                # Now execute the update to the given Database Table Row
                migration_engine.execute(
                    db_update, pkid=row['pk_id'], uuid=auuid)


def _add_default_scg_values(migration_engine, metadata):
    """ Helper method to add the default values for some SCG columns """
    db_table = Table('storage_connectivity_group', metadata, autoload=True)
    db_query = expression.select([db_table])
    db_rows = migration_engine.execute(db_query).fetchall()
    db_update = db_table.update().where(
        db_table.c.pk_id == bindparam('pkid')).values(
            vios_redundancy=bindparam('c1'), exact_redundancy=bindparam('c2'),
            npiv_fabric=bindparam('c5'), boot_connectivity=bindparam('c3'),
            data_connectivity=bindparam('c4'))
    # Loop through each of the Rows updating the ID to a generated Value
    for row in db_rows:
        values = dict(c1=1, c2=0, c3=None, c4=None, c5=None)
        # If the fc_storage_access is True, then it is an NPIV-based SCG
        if row['fc_storage_access'] == 1:
            if row['cluster_provider_name']:
                values['c3'] = values['c4'] = 'npiv,ibm_ssp'
            # If the Cluster Provider Name isn't set then it is only NPIV
            else:
                values['c3'] = values['c4'] = 'npiv'
            values['c5'] = 'any'
        # Otherwise if the Cluster Provider Name is set, the SCG is SSP-based
        elif row['cluster_provider_name']:
            values['c3'] = values['c4'] = 'ibm_ssp'
        migration_engine.execute(db_update, pkid=row['pk_id'], **values)
