# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
"""Table Creation/Deletion for Schema changes in the PowerVC 1.2.3 Release"""

import ConfigParser
from datetime import datetime
from sqlalchemy import Table, MetaData, Boolean, Column, Text
from sqlalchemy.sql import bindparam, expression, and_
from oslo_log import log as logging
from powervc_nova import _

TABLE_COLUMN_UPDATES = [
    ('storage_connectivity_group', [Column('include_untagged', Boolean)]),
    ('instance_power_specs',
     [Column('srr_capability', Boolean), Column('srr_state', Text)])
]

LOG = logging.getLogger(__name__)


def upgrade(migration_engine):
    """Upgrades the Nova DB Tables to include those for PowerVC 1.2.3"""
    metadata = MetaData(migration_engine)
    metadata.reflect(migration_engine)
    metadata.bind = migration_engine
    # Loop through each of the tables we need to add columns to
    for column_update in TABLE_COLUMN_UPDATES:
        table_name, table_columns = column_update
        # We need to add the additional columns to the given table
        atable = Table(table_name, metadata, autoload=True)
        for column in table_columns:
            # Check first to see if the Column exists on the Table already
            if atable._columns.get(column.name) is None:
                atable.create_column(column.copy())
    # Make sure we mark all deleted VM's as soft-delete to cover unmanaged ones
    _mark_unmanaged_vms_soft_deleted(migration_engine, metadata)
    # Only want to do the updates related to image if it is PowrVM
    if _is_powervm():
        # Need to update the Instance & BlockDeviceMapping tables for
        # Multi-disk
        _updates_for_multi_disk_support(migration_engine, metadata)
        # Need to set the Endianness attribute in the MetaData for each
        # Instance
        _set_endianness_on_instances(migration_engine, metadata)


def downgrade(migration_engine):
    """Downgrades the Nova DB Tables to remove those for PowerVC 1.2.3"""
    pass


def _updates_for_multi_disk_support(migration_engine, metadata):
    """Update Instance and BlockDeviceMapping tables for Multi-disk"""
    boot_bdm_list, inst_root_map = list(), dict()
    image_bdm_list = list()
    vm_table = Table('instances', metadata, autoload=True)
    bdm_table = Table('block_device_mapping', metadata, autoload=True)
    vm_system_metadata_table = Table('instance_system_metadata',
                                     metadata, autoload=True)
    # Query all of the non-deleted BlockDeviceMappings and Instances
    vm_query = expression.select([vm_table]).where(vm_table.c.deleted == 0)
    bdm_query = expression.select([bdm_table]).where(bdm_table.c.deleted == 0)
    vm_metadata_query = expression.select([vm_system_metadata_table]).\
        where(vm_system_metadata_table.c.key == 'image_os_distro')
    vm_rows = migration_engine.execute(vm_query).fetchall()
    bdm_rows = migration_engine.execute(bdm_query).fetchall()
    vm_rows = dict([(row['uuid'], row) for row in vm_rows])
    vm_metadata_rows = migration_engine.execute(vm_metadata_query).fetchall()
    vm_metadata_rows = dict([(row['instance_uuid'], row)
                             for row in vm_metadata_rows])
    # If there weren't any Instances or BlockDeviceMappings, then nothing to do
    if len(vm_rows) <= 0 or len(bdm_rows) <= 0:
        return
    # Loop through each BDM, finding a match with the Instance on Boot Device
    for bdm_row in bdm_rows:
        inst_uuid = bdm_row['instance_uuid']
        vm_row = vm_rows.get(inst_uuid)
        # If the VM for the BlockDeviceMapping wasn't found, then continue on
        if vm_row is None or not vm_row['default_ephemeral_device']:
            LOG.warning(_("Could not get Instance %s for BDM") % inst_uuid)
            continue
        # Check if we have the bdm is Image BDM. This need to be deleted.
        if bdm_row['image_id'] and bdm_row['boot_index'] == 0:
            image_bdm_list.append(bdm_row['id'])
        # If this BDM is the Default Ephemeral Device, then tag as such
        if vm_row['default_ephemeral_device']:
            if bdm_row['volume_id'] == vm_row['default_ephemeral_device']:
                boot_bdm_list.append(bdm_row['id'])
                inst_root_map[inst_uuid] = bdm_row['device_name']
    # Delete Image BDMs
    if image_bdm_list:
        delete_image_bdm_query = bdm_table.delete().where(
            bdm_table.c.id.in_(image_bdm_list))
        migration_engine.execute(delete_image_bdm_query)
    # Now that we have the list, we can take Boot Index 0 on the Boot BDM's
    boot_index_update = bdm_table.update().where(
        bdm_table.c.id.in_(boot_bdm_list)).values(boot_index=0)
    migration_engine.execute(boot_index_update)
    # Finally we can swap from the default_ephemeral_device to root_device_name
    vm_update = vm_table.update().where(
        vm_table.c.uuid == bindparam('uid')).values(
            default_ephemeral_device=None, root_device_name=bindparam('dev'))
    # Loop through each of the Instances, executing the update defined above
    for uuid, device_name in inst_root_map.iteritems():
        migration_engine.execute(vm_update, uid=uuid, dev=device_name)
    # Update configuration strategy for the instance which was both
    # onboarded and 'finialize onboarded'.
    for inst_uuid in vm_rows.keys():
        vm_row = vm_rows.get(inst_uuid)
        inst_prop_image_os_distro = vm_metadata_rows.get(inst_uuid)
        image_ref = vm_row['image_ref']
        if image_ref and image_ref == 'ONBOARDED_NO_IMAGE' and \
                inst_prop_image_os_distro:
            vm_system_metadata_table.insert().values(
                instance_uuid=inst_uuid,
                key='image_configuration_strategy',
                value='NEEDED',
                deleted=0,
                created_at=datetime.utcnow()).execute()


def _set_endianness_on_instances(migration_engine, metadata):
    """Set the Endianness attribute in the MetaData for each Instance"""
    vm_table = Table('instances', metadata, autoload=True)
    meta_table = Table('instance_system_metadata', metadata, autoload=True)
    # Query all of the non-deleted VM's and Endianness Meta from the Database
    vm_query = expression.select(
        [vm_table.c.uuid]).where(vm_table.c.deleted == 0)
    meta_query = expression.select([meta_table.c.instance_uuid]).where(
        and_(meta_table.c.deleted == 0,
             meta_table.c.key == 'image_endianness'))
    vm_rows = migration_engine.execute(vm_query).fetchall()
    meta_rows = migration_engine.execute(meta_query).fetchall()
    meta_rows = dict([(inst_uuid[0], '') for inst_uuid in meta_rows])
    # Loop through each of the VM's seeing if we need to insert the Endianness
    for vm_uuid in vm_rows:
        vmid = vm_uuid[0]
        # Skip it if the Endianness attribute is already there for the Instance
        if vmid in meta_rows:
            continue
        meta_table.insert().values(
            created_at=datetime.utcnow(), deleted=0,
            instance_uuid=vmid,
            key='image_endianness',
            value='big-endian').execute()


def _mark_unmanaged_vms_soft_deleted(migration_engine, metadata):
    """Marks an Un-managed VM's as Soft-deleted to fix APAR IT06171"""
    vm_table = Table('instances', metadata, autoload=True)
    vm_update = vm_table.update().values(vm_state='soft-delete')
    vm_update = vm_update.where(and_(
        vm_table.c.deleted != 0, vm_table.c.vm_state == 'deleted'))
    migration_engine.execute(vm_update)


def _is_powervm():
    """Helper method to check if the Management System is PowerVM"""
    parser = ConfigParser.RawConfigParser()
    parser.read('/etc/nova/nova.conf')
    if parser.has_option('DEFAULT', 'host_type'):
        host_type = parser.get('DEFAULT', 'host_type')
        return host_type == 'powervm'
    return False
