# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
"""Table Creation/Deletion for Schema changes in the PowerVC 1.3.0 Release"""

import ConfigParser
from nova.compute import arch
from nova.compute import vm_mode
from oslo_log import log as logging
from oslo_serialization import jsonutils
from powervc_nova.db.sqlalchemy import models
from sqlalchemy import Float, Table, Column, String, Integer, MetaData
from sqlalchemy.sql import bindparam, expression, and_

POWER_SPEC_COLUMNS = [Column('cpu_util', Float)]
STORAGE_SCG_COLUMNS = [
    Column('fabric_set_npiv', String(255)),
    Column('fabric_set_req', String(31)),
    Column('ports_per_fabric_npiv', Integer)
]
TABLE_COLUMN_UPDATES = [
    ('storage_connectivity_group', STORAGE_SCG_COLUMNS),
    ('instance_power_specs', POWER_SPEC_COLUMNS),
    ('ibm_hmcs', [Column('mtmsn', String(255))])
]
SYSMETA_DELETED_KEYS = [
    'cpus', 'min_cpus', 'min_vcpus', 'min_memory_mb', 'max_memory_mb',
    'max_cpus', 'max_vcpus', 'current_compatibility_mode', 'rmc_state',
    'desired_compatibility_mode', 'srr_capability', 'srr_state']

TYPES = jsonutils.dumps([(arch.PPC64, "phyp", vm_mode.HVM),
                         (arch.PPC64LE, "phyp", vm_mode.HVM)])

LOG = logging.getLogger(__name__)


def upgrade(migration_engine):
    """Upgrades the Nova DB Tables to include those for PowerVC 1.3.0"""
    metadata = MetaData(migration_engine)
    metadata.reflect(migration_engine)
    metadata.bind = migration_engine

    # Loop through each of the tables we need to add columns to
    for column_update in TABLE_COLUMN_UPDATES:
        table_name, table_columns = column_update
        # We need to add the additional columns to the given table
        atable = Table(table_name, metadata, autoload=True)
        for column in table_columns:
            # Check first to see if the Column exists on the Table already
            if atable._columns.get(column.name) is None:
                atable.create_column(column.copy())

    # Loop through all of the DTO's that we added new in PowerVC 1.3.0
    # creating the tables off of the definition defined in the DTO
    for dto in models.POWERVC_V1R3M0_DTOS:
        table = dto.__table__.tometadata(metadata, None)
        table.create(checkfirst=True)

    if _is_powervm():
        update_instance_meta_props_to_phyp(migration_engine, metadata)
        update_host_props_to_phyp(migration_engine, metadata)
        update_storage_connectivity_group_conversion(
            migration_engine, metadata)
    # Cleanup the System MetaData Keys that are no longer being used
    _cleanup_sysmeta_keys(migration_engine, metadata)
    # Populate the Host property for the Compute Nodes before 1.2.3
    _populate_host_on_compute_nodes(migration_engine, metadata)


def downgrade(migration_engine):
    """Downgrades the Nova DB Tables to remove those for PowerVC 1.3.0"""
    metadata = MetaData(migration_engine)
    metadata.reflect(migration_engine)
    metadata.bind = migration_engine
    # Loop through all of the DTO's that we added new in PowerVC 1.3.0
    # dropping the tables off of the definition defined in the DTO
    for dto in reversed(models.POWERVC_V1R3M0_DTOS):
        dto.__table__.metadata = metadata
        dto.__table__.drop(checkfirst=True)


def update_instance_meta_props_to_phyp(migration_engine, metadata):
    """Updates the VM image_hypervisor_type property to phyp from powervm"""
    inst_table = Table('instances', metadata, autoload=True)
    meta_table = Table('instance_system_metadata', metadata, autoload=True)
    # Setup the queries that we need to use for updating the DB Tables
    inst_query = expression.select(
        [inst_table.c.uuid]).where(inst_table.c.deleted == 0)

    ac = and_(
        meta_table.c.deleted == 0,
        meta_table.c.key == 'image_hypervisor_type')
    meta_query = expression.select(
        [meta_table.c.instance_uuid, meta_table.c.value]).where(ac)

    meta_update = meta_table.update().where(and_(
        meta_table.c.instance_uuid == bindparam('uid'),
        meta_table.c.key == 'image_hypervisor_type')).values(value='phyp')

    img_rows = migration_engine.execute(inst_query).fetchall()
    meta_rows = migration_engine.execute(meta_query).fetchall()
    meta_rows = dict([(row[0], row[1]) for row in meta_rows])

    # Loop through each of the instance in the DB, updating the necessary
    # fields
    for row in img_rows:
        image_id = row[0]
        hypervisor_type = meta_rows.get(image_id)
        if hypervisor_type is not None:
            migration_engine.execute(meta_update, uid=image_id)


def update_host_props_to_phyp(migration_engine, metadata):
    """Updates the Host hypervisor_type property to phyp from powervm"""
    inst_table = Table('compute_nodes', metadata, autoload=True)
    # Setup the queries that we need to use for updating the DB Tables
    inst_update = inst_table.update().where(
        inst_table.c.deleted == 0).values(hypervisor_type='phyp',
                                          supported_instances=TYPES)
    migration_engine.execute(inst_update)


def update_storage_connectivity_group_conversion(migration_engine, metadata):
    """Updates/converts the old Storage Connectivity Group properties to the
    new fabric_set values"""
    """ Helper method to add the default values for some SCG columns """
    db_table = Table('storage_connectivity_group', metadata, autoload=True)
    db_query = expression.select([db_table])
    db_rows = migration_engine.execute(db_query).fetchall()
    db_update = db_table.update().where(
        db_table.c.pk_id == bindparam('pkid')).values(
            fabric_set_npiv=bindparam('c1'), fabric_set_req=bindparam('c2'),
            ports_per_fabric_npiv=bindparam('c3'), npiv_fabric=bindparam('c4'))
    # Loop through each of the Rows updating the ID to a generated Value
    for row in db_rows:
        values = dict(c1=None, c2=None, c3=1, c4='any')
        # If the fc_storage_access is True, then it is an NPIV-based SCG
        if row['fc_storage_access'] == 1:
            if row['npiv_fabric'] == 'A':
                values['c1'] = 'A'
                values['c2'] = 'at_least_one'
            elif row['npiv_fabric'] == 'B':
                values['c1'] = 'B'
                values['c2'] = 'at_least_one'
            elif row['npiv_fabric'] == 'single':
                values['c1'] = 'A,B'
                values['c2'] = 'exactly_one'
            elif row['npiv_fabric'] == 'dual':
                values['c1'] = 'A,B'
                values['c2'] = 'per_host'
            elif row['npiv_fabric'] == 'dual_per_vios':
                values['c1'] = 'A,B'
                values['c2'] = 'per_vios'
            else:
                values['c1'] = 'A,B'
                values['c2'] = 'at_least_one'
        migration_engine.execute(db_update, pkid=row['pk_id'], **values)


def _cleanup_sysmeta_keys(migration_engine, metadata):
    """Cleanup the System MetaData Keys that are no longer used"""
    meta_table = Table('instance_system_metadata', metadata, autoload=True)
    migration_engine.execute(
        meta_table.delete().where(meta_table.c.key.in_(SYSMETA_DELETED_KEYS)))


def _populate_host_on_compute_nodes(migration_engine, metadata):
    """Populate the Host property on Compute Nodes older than 1.2.3"""
    srv_table = Table('services', metadata, autoload=True)
    cpt_table = Table('compute_nodes', metadata, autoload=True)
    # Setup the necessary query for Compute Node rows with Host == NULL
    node_query = expression.select(
        [cpt_table.c.service_id]).where(cpt_table.c.host.is_(None))
    node_rows = migration_engine.execute(node_query).fetchall()
    node_update = cpt_table.update().where(
        cpt_table.c.service_id == bindparam('s1')).values(host=bindparam('h1'))
    # Loop through the Compute Nodes finding and updating the Host property
    for node_row in node_rows:
        service_id = node_row[0]
        if service_id is None:
            continue
        # Setup the query for the Host property from the Service table
        svc_query = expression.select(
            [srv_table.c.host]).where(srv_table.c.id == service_id)
        svc_row = migration_engine.execute(svc_query).first()
        # If the Host property was actually set, then update the Compute Node
        if svc_row is not None and svc_row[0]:
            migration_engine.execute(node_update, s1=service_id, h1=svc_row[0])


def _is_powervm():
    """Helper method to check if the Management System is PowerVM"""
    parser = ConfigParser.RawConfigParser()
    parser.read('/etc/nova/nova.conf')
    if parser.has_option('DEFAULT', 'host_type'):
        host_type = parser.get('DEFAULT', 'host_type')
        return host_type == 'powervm'
    return False
