# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
"""Table Creation/Deletion for Schema changes in the PowerVC 1.3.1 Release"""

from oslo_log import log as logging
from powervc_nova.db.sqlalchemy import models
from sqlalchemy import Table, Column, Integer, String
from sqlalchemy import Float, Index, ForeignKey, MetaData
from sqlalchemy.sql import bindparam, expression, and_

DRO_EVENTS_COLUMNS = [
    Column('aggregate_id', Integer, ForeignKey('aggregates.id'),
           nullable=True)
]
POWER_SPEC_COLUMNS = [Column('mem_util', Float),
                      Column('pg_in_rate', Integer),
                      Column('pg_out_rate', Integer),
                      Column('swap_in_rate', Integer),
                      Column('swap_out_rate', Integer)]

TABLE_RENAMES = [('ibm_cod_pools', 'ibm_enterprise_pools'),
                 ('ibm_cod_pool_hosts', 'ibm_enterprise_pool_members')]

TABLE_COLUMN_ADDITIONS = [
    ('dro_events', DRO_EVENTS_COLUMNS),
    ('instance_power_specs', POWER_SPEC_COLUMNS),
    ('ibm_console_messages', [Column('project_id', String(255))])]
TABLE_INDEX_ADDITIONS = [('dro_events', 'aggregate_id',
                          'dro_events_aggregate_id_idx')]

TABLE_COLUMN_ALTERS = [
    ('dro_events', 'hypervisor_hostname', {'nullable': True})
]

LOG = logging.getLogger(__name__)


def upgrade(migration_engine):
    """Upgrades the Nova DB Tables to include those for PowerVC 1.3.1"""
    metadata = MetaData(migration_engine)
    metadata.reflect(migration_engine)
    metadata.bind = migration_engine

    # Loop through each of the tables we need to rename
    for table_rename in TABLE_RENAMES:
        old_table_name, new_table_name = table_rename
        # Check first to see if a table with the old name already exists
        existing_table = Table(old_table_name, metadata)
        if existing_table.exists():
            # do the rename if the table already exists with the old name
            existing_table.rename(new_table_name)

    # Loop through each of the tables we need to add columns to
    for column_addition in TABLE_COLUMN_ADDITIONS:
        table_name, table_columns = column_addition
        # We need to add the additional columns to the given table
        atable = Table(table_name, metadata, autoload=True)
        for column in table_columns:
            # Check first to see if the Column exists on the Table already
            if atable._columns.get(column.name) is None:
                atable.create_column(column.copy())

    # Loop through each of the tables for which we need to add an index
    for index_addition in TABLE_INDEX_ADDITIONS:
        table_name, column_name, index_name = index_addition
        atable = Table(table_name, metadata, autoload=True)
        # Check that an index doesn't already exist for the column
        for idx in atable.indexes:
            if idx.columns.get(column_name) is not None:
                break
        else:
            # Create the index
            acolumn = atable._columns.get(column_name)
            Index(index_name, acolumn).create(migration_engine)

    # Loop through each of the columns we need to alter
    for column_alter in TABLE_COLUMN_ALTERS:
        table_name, column_name, alterations = column_alter
        # Get the column to be altered
        atable = Table(table_name, metadata, autoload=True)
        acolumn = atable._columns.get(column_name)
        # Alter the column if it exists
        if acolumn is not None:
            acolumn.alter(**alterations)

    # Loop through all of the DTO's that we added new in PowerVC 1.3.1
    # creating the tables off of the definition defined in the DTO
    for dto in models.POWERVC_V1R3M1_DTOS:
        table = dto.__table__.tometadata(metadata, None)
        table.create(checkfirst=True)

    # Update the behavior for deleting attached volumes for On-boarded VM's
    _update_onboarded_vm_volume_deletion(migration_engine, metadata)


def downgrade(migration_engine):
    """Downgrades the Nova DB Tables to remove those for PowerVC 1.3.0"""
    metadata = MetaData(migration_engine)
    metadata.reflect(migration_engine)
    metadata.bind = migration_engine
    # Loop through all of the DTO's that we added new in PowerVC 1.3.1
    # dropping the tables off of the definition defined in the DTO
    for dto in reversed(models.POWERVC_V1R3M1_DTOS):
        dto.__table__.metadata = metadata
        dto.__table__.drop(checkfirst=True)

    # Undo all of the table renames that we did in PowerVC 1.3.1
    for table_rename in reversed(TABLE_RENAMES):
        old_table_name, new_table_name = table_rename
        # Check first to see if a table with the new name exists
        existing_table = Table(new_table_name, metadata, autoload=True)
        if existing_table.exists():
            # Rename the table back to the old name
            existing_table.rename(old_table_name)


def _update_onboarded_vm_volume_deletion(migration_engine, metadata):
    """Update the behavior for deleting attached volumes for On-boarded VM's"""
    bdm_table = Table('block_device_mapping', metadata, autoload=True)
    spec_table = Table('instance_power_specs', metadata, autoload=True)
    # Query all of the Instances that have a "delete" in the pending actions,
    # which means that they have multi-volumes and aren't finished on-boarding
    spec_query = expression.select([spec_table]).where(
        and_(spec_table.c.deleted == 0,
             spec_table.c.pending_onboard_actions.like('%delete%')))
    bdm_update = bdm_table.update().where(
        bdm_table.c.instance_uuid == bindparam('uid')).values(
            delete_on_termination=False)
    spec_update = spec_table.update().where(
        spec_table.c.instance_uuid == bindparam('uid')).values(
            pending_onboard_actions=bindparam('actions'))
    # Loop through each row, removing "delete" from the pending actions and
    # setting all the delete_on_termination flags to false for the BDM's
    for row in migration_engine.execute(spec_query).fetchall():
        actions = row['pending_onboard_actions'].split(',')
        actions.remove('delete')
        actions = ','.join(actions) if len(actions) > 0 else None
        # Now we can execute the update operations on the PowerSpecs and BDM's
        migration_engine.execute(
            spec_update, uid=row['instance_uuid'], actions=actions)
        migration_engine.execute(bdm_update, uid=row['instance_uuid'])
