# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
"""Table Creation/Deletion for Schema changes in the PowerVC 1.3.1.1 Release"""

import json
from datetime import datetime
from oslo_log import log as logging
from sqlalchemy import Table, MetaData, Text
from sqlalchemy.sql import bindparam, expression
from sqlalchemy.dialects.mysql import MEDIUMTEXT

LOG = logging.getLogger(__name__)


def upgrade(migration_engine):
    """Upgrades the Nova DB Tables to include those for PowerVC 1.3.1.1"""
    metadata = MetaData(migration_engine)
    metadata.reflect(migration_engine)
    metadata.bind = migration_engine
    # Clean up any Console Messages that got messed up on the upgrade
    _cleanup_misformed_console_messages(migration_engine, metadata)


def downgrade(migration_engine):
    """Downgrades the Nova DB Tables to remove those for PowerVC 1.3.1.1"""
    pass


def _cleanup_misformed_console_messages(migration_engine, metadata):
    """Clean up any Console Messages that got messed up on the upgrade"""
    msg_table = Table('ibm_console_messages', metadata, autoload=True)
    # First increase the length of the column so don't hit in the future
    if str(msg_table.c.payload.type) == 'TEXT':
        newtype = Text().with_variant(MEDIUMTEXT(), 'mysql')
        msg_table.c.payload.alter(type=newtype)
    # Query all of the non-deleted Messages and make sure there are valid
    # and if not then we will mark the as deleted so they don't cause issues
    msg_query = expression.select([msg_table]).where(msg_table.c.deleted == 0)
    msg_update = msg_table.update().where(
        msg_table.c.id == bindparam('oid')).values(
            deleted=bindparam('deleted'), deleted_at=datetime.utcnow())
    # Loop through each row, marking any misformed ones as deleted
    for row in migration_engine.execute(msg_query).fetchall():
        try:
            json.loads(row['payload'])
        # If we got an exception loading the payload, mark it ias deleted
        except:
            migration_engine.execute(
                msg_update, oid=row['id'], deleted=row['id'])
