# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from nova import exception as novaexception
from nova import objects
from oslo_concurrency import lockutils
from oslo_log import log as logging
from oslo_config import cfg

from powervc_nova.network.common.api_suppress_mixin import NetworkAPISuppressor
from powervc_nova.network.common import exception as net_excp

import powervc_oslo

LOG = logging.getLogger(__name__)
CONF = cfg.CONF

MOCK_DATA_INDEX = {
    "host-sriov": [
        {
            "host_name": "host1",
            "adapters": [
                {
                    "name": "sriov_eth0",
                    "pf_name": "sriov_eth0_pf",
                    "max_vfs": 32,
                    "speed_mb": 10000,
                    "state": "OK",
                    "vfs": [
                        {
                            "name": "vf0",
                            "strong_low_pct": 10,
                            "state": "OK",
                            "vios": 1
                        },
                        {
                            "name": "vf1",
                            "strong_low_pct": 10,
                            "state": "OK",
                            "vios": 1
                        }
                    ]
                }
            ]

        },
        {
            "host_name": "host2",
            "adapters": [
                {
                    "name": "sriov_eth3",
                    "pf_name": "sriov_eth3_pf",
                    "max_vfs": 32,
                    "speed_mb": 10000,
                    "state": "OK",
                    "vfs": [
                        {
                            "name": "vf0",
                            "strong_low_pct": 25,
                            "state": "OK",
                            "vios": 2
                        }
                    ]
                }
            ]
        },
        {
            "host_name": "host3",
            "adapters": [],
            "errors": [
                {
                    "message": "Communication error."
                }
            ]

        }

    ]
}

MOCK_DATA_SHOW = {
    "host-sriov": [
        {
            "host_name": "host1",
            "adapters": [
                {
                    "name": "sriov_eth0",
                    "pf_name": "sriov_eth0_pf",
                    "speed_mb": 10000,
                    "max_vfs": 32,
                    "state": "OK",
                    "vfs": [
                        {
                            "name": "vf0",
                            "strong_low_pct": 25,
                            "state": "OK",
                            "vios": 1
                        },
                        {
                            "name": "vf1",
                            "strong_low_pct": 25,
                            "state": "OK",
                            "vios": 1
                        },
                        {
                            "name": "vf2",
                            "strong_low_pct": 30,
                            "state": "OK",
                            "vios": 1
                        },
                        {
                            "name": "vf3",
                            "strong_low_pct": 20,
                            "state": "OK",
                            "vios": 1
                        }
                    ]
                },
                {
                    "name": "sriov_eth1",
                    "pf_name": "sriov_eth1_pf",
                    "speed_mb": 10000,
                    "max_vfs": 32,
                    "state": "OK",
                    "vfs": [
                        {
                            "name": "vf0",
                            "strong_low_pct": 3,
                            "state": "OK",
                            "vios": 2
                        },
                        {
                            "name": "vf1",
                            "strong_low_pct": 3,
                            "state": "OK",
                            "vios": 2
                        }
                    ]
                }
            ]
        }
    ]
}


class HostSriovTask(NetworkAPISuppressor):
    """
    This class handles processing for the host-sriov REST API.
    """

    def __init__(self):
        # Invoke superclass constructor
        super(HostSriovTask, self).__init__()

    @lockutils.synchronized('host_sriov', 'host-sriov-')
    def get_host_sriov(self, context, req, host_name=None):
        """
        Read SRIOV configuration.
        :param context: The HTTP request context.
        :param req: The HTTP request object.
        :param host_name: An optional host name to filter data with.
        :returns: A dictionary of host-sriov data.
        """
        # This class should only be used in PowerVM environments
        self.raise_if_not_powervm()

        # Add locale to the context
        context = powervc_oslo.populate_locale(context, req)

        # If a host was specified, verify it exists
        host_list = self._get_all_host_names(context)
        if host_name is not None:
            if host_name not in host_list:
                raise novaexception.ComputeHostNotFound(host=host_name)
            host_list = [host_name]

        LOG.debug("host-sriov GET for host(s) %s" % host_list)

        # Parse the vlanid from the request, if present
        vlanid = None
        if 'vlanid' in req.GET:
            vlanid = req.GET['vlanid']
        if vlanid == '':
            raise net_excp.IBMPowerVMInvalidVlan(vlanid='')

        # Make the remote call to the endpoint(s) to fetch sriov data
        # TODO - for now just mock
        dom_list = MOCK_DATA_INDEX
        if host_name:
            dom_list = MOCK_DATA_SHOW

        # Convert DOM objects into REST API dictionary format
        host_dict = self._dom_to_dict(dom_list)
        return host_dict

    def _get_all_host_names(self, context):
        """
        Finds a list of compute node names.
        :param context: DB context object.
        :returns: A list of compute node names.
        """
        context = context.elevated()
        return [compute.host
                for compute in objects.ComputeNodeList.get_all(context)]

    def _validate_body(self, verb, body, current_dom, context):
        """
        TODO
        """
        pass

    def _dict_to_dom(self, host_dict, host_dom_list):
        """
        TODO
        """
        return host_dom_list

    def _dom_to_dict(self, host_list):
        """
        TODO
        """
        return host_list
