#
#
# Copyright 2014, 2015 IBM Corp.
# All Rights Reserved.
# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
# Copyright (c) 2012 NEC Corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
import nova
from nova.compute import vm_states
from nova.network.neutronv2 import api as netapi
from nova.objects import compute_node as comp_node_obj
from nova.objects import instance as inst_obj
from nova.objects import network as net_obj
from oslo_log import log as logging
from nova.network import neutronv2
from oslo_config import cfg
from nova import exception
from nova.objects import network_request
from oslo_utils import excutils
from neutronclient.common import exceptions as neutron_client_exc
from powervc_nova.objects.network import dom as network_dom
from powervc_nova import _
from nova.i18n import _LW, _LE

LOG = logging.getLogger(__name__)
CONF = cfg.CONF

multiple_nics_same_network_opt = [
    cfg.BoolOpt('multiple_nics_same_network',
                default=False,
                help='Whether multiple NICs from same network be allowed'),
]

CONF.register_opts(multiple_nics_same_network_opt)


class PowerBaseAPI(netapi.API):

    """
    Base class for API wrapper to abstract out function shared by Power VM
    and Power KVM.
    """

    def allocate_for_instance(self, context, instance, **kwargs):
        '''
        This is a resilience add on method for the standard allocate.
        '''
        vifs = []
        try:
            vifs = super(PowerBaseAPI, self).allocate_for_instance(context,
                                                                   instance,
                                                                   **kwargs)
        except exception.FixedIpAlreadyInUse:
            # If this occurs, it could be a re-schedule and the neutron port
            # is left over from a previous bad allocate
            vifs = self._possible_port_clean_and_reallocate(context, instance,
                                                            **kwargs)

        return vifs

    def get_all(self, context):
        """Get all networks for client."""
        network_objs = list()
        client = netapi.get_client(context)
        networks = client.list_networks().get('networks')
        for network in networks:
            network_obj = self._construct_network_object(context, network)
            network_objs.append(network_obj)
        return net_obj.NetworkList(context=context, objects=network_objs)

    def get(self, context, network_uuid):
        """Get specific network for client."""
        client = netapi.get_client(context)
        try:
            network = client.show_network(network_uuid).get('network') or {}
        except neutron_client_exc.NetworkNotFoundClient:
            raise exception.NetworkNotFound(network_id=network_uuid)
        return self._construct_network_object(context, network)

    def _construct_network_object(self, context, network):
        """Helper method to construct a Network DOM Object"""
        net_obj = network_dom.Network(context=context, name=network['name'],
                                      label=network['name'],
                                      uuid=network['id'])
        net_obj.segmentation_id = network.get('provider:segmentation_id')
        net_obj.physical_network = network.get('provider:physical_network')
        return net_obj

    def _possible_port_clean_and_reallocate(self, context, instance,
                                            **kwargs):
        '''
        If a nova allocate fails during a neutron port allocation, it could
        leave a neutron port hanging around that is allocated against a
        nova instance that does not exist.

        This method will detect if a neutron port exists against an invalid
        nova instance.  If so, it will clean out that neutron port and then
        return the allocate.
        '''
        requested_networks = kwargs.get('requested_networks')
        neutron = neutronv2.api.get_client(context)

        LOG.warn(_("An error occurred during the allocation of the network.  "
                   "This could be due to a stale port.  Attempting a clean "
                   "up operation and then another allocation."))

        for req_net in requested_networks:
            # Convert to a Tuple if the incoming is a NetworkRequest
            if isinstance(req_net, network_request.NetworkRequest):
                req_net = req_net.to_tuple()
            network_id, fixed_ip, port_id, _pci_id = req_net
            # If they gave us a port id, then this doesn't apply
            if port_id:
                continue

            # If it is a network and a fixed ip...then we need to investigate
            if network_id and fixed_ip:
                # See if this exists in neutron
                search_opts = {'network_id': network_id,
                               'fixed_ips': 'ip_address=%s' % fixed_ip}
                existing_ports = neutron.list_ports(**search_opts)

                # If there were no existing ports, just move on.
                if not existing_ports or\
                        len(existing_ports.get('ports')) == 0:
                    continue

                # Check to see if the owner of that port is nova.
                port = existing_ports.get('ports')[0]
                if port.get('device_owner') != 'compute:None':
                    continue

                # Port is owned by compute...lets get the UUID
                owning_device = port.get('device_id')

                # If the owner actually is a VM, then this is a valid error.
                # It will show up on the next allocate...
                try:
                    owner = inst_obj.Instance.get_by_uuid(context,
                                                          owning_device)
                    # If we get here, then the InstanceNotFound exception was
                    # not raised.  Check to see if WE own it and are still in
                    # building state.  This means we're in reschedule and the
                    # first deploy attempt failed, leaving the port out there.
                    # We DON'T want to do this after the VM is up as we could
                    # inadvertently destroy a neutron port during attach vnic
                    # if they were to erroneously give us an IP in use already.
                    if owner is not None and owner.uuid == instance.uuid and\
                            instance.vm_state == vm_states.BUILDING:
                        # We do own it.  Delete it so the below
                        # allocate_for_instance works.
                        neutron.delete_port(port.get('id'))

                except exception.InstanceNotFound:
                    # At this point, we know if has no owning instance.
                    # It was incorrectly created.  Lets wipe it out and
                    # then a subsequent allocate will rebuild it with
                    # the proper device.
                    msg = _("Found a port that was claimed by Nova, on "
                            "an instance with UUID of %(inst_uuid)s.  "
                            "That instance does not exist in Nova.  "
                            "Therefore, the port is stale and is going "
                            "to be removed.  The port that will be "
                            "automatically removed has a UUID in Neutron "
                            "of %(port_uuid)s.") % {'inst_uuid':
                                                    owning_device,
                                                    'port_uuid':
                                                    port.get('id')}
                    LOG.warn(msg)
                    neutron.delete_port(port.get('id'))

        # At this point, we want to retry the allocation.  Notice that we call
        # the super class so that we don't recursively end up back here.  :-)
        LOG.warn(_("Retrying the allocation of the port post clean up."))
        return super(PowerBaseAPI, self).allocate_for_instance(context,
                                                               instance,
                                                               **kwargs)

    def get_viable_hosts(self, context, instance_uuid,
                         requested_networks=None, source_host=None):
        """
        Find the viable hosts that match the network requirements.

        :param context: A context object that is used to authorize
                        the DB access.
        :param instance_uuid: The instance UUID that will be used to
                        get the network requirements of an instance.
        :param requested_networks: An optional list of requested networks for
                        the deployment.  If set, will only place on hosts that
                        have valid targets for that host.  If not set, then
                        all hosts will be allowed.

                        Format of this should be:
                        [['1d4dff20-ed6d-47ff-a7b0-5f62bbf5b635',
                          '9.114.181.199', None]]
                        Note that this is a list of list.  The first element
                        of the embedded list is the UUID of the network.
        :param source_host: When this API is invoked for LPM, the host that
                            is being migrated from.  This is needed to check
                            vswitch config on the source host.
        :returns: A dictionary of hosts with the hostid and the hostname
        """
        LOG.debug('instance uuid: %s\nrequested networks: %s\n'
                  'source_host: %s' %
                  (instance_uuid, requested_networks, source_host))
        # Parse the requested networks to a UUID list
        network_uuids = []
        if requested_networks:
            for req_net in requested_networks:
                # Convert to a Tuple if the incoming is a NetworkRequest
                if isinstance(req_net, network_request.NetworkRequest):
                    req_net = req_net.to_tuple()
                if len(req_net) > 1:
                    if len(req_net) > 2 and req_net[0] is None and req_net[2]:
                        # This means it's possibly a boot with the port-id
                        # Let's fetch the network and set things right.
                        search_opts = {'id': req_net[2]}
                        data = neutronv2.api.get_client(context).\
                            list_ports(**search_opts)
                        # Extra debug added for 34261 - ports should never be
                        # empty, but for some reason it is
                        if not data.get('ports', None):
                            ports = neutronv2.api.get_client(context).\
                                list_ports()
                            networks = self._get_network_all(context)
                            LOG.warn(_('Missing port data for req_net: '
                                       '%(req_net)s and data: %(data)s') %
                                     {'req_net': req_net,
                                      'data': data})
                            LOG.warn(_('All ports: %(ports)s Nets: %(nets)s') %
                                     {'ports': ports,
                                      'nets': networks})
                            LOG.warn(_('Port not found. No eligible hosts'))
                            return []
                        else:
                            net_id = data['ports'][0]['network_id']
                            network_uuids.append(net_id)
                    else:
                        network_uuids.append(req_net[0])
        else:
            instance = inst_obj.Instance.get_by_uuid(context, instance_uuid)
            nw_infos = []
            if instance is not None:
                nw_infos = self.get_instance_nw_info(context, instance)
            for nw_info in nw_infos:
                network_uuids.append(nw_info['network']['id'])

        # Get all compute hosts from the database
        hosts_from_db = self._get_compute_nodes_from_DB(context)
        if not hosts_from_db:
            LOG.warn(_('No hosts found in the database'))

        # Determine which hosts are viable using a hypervisor-specific method
        candidate_hosts = self._get_viable_hosts(context, instance_uuid,
                                                 hosts_from_db, network_uuids,
                                                 source_host)

        if not candidate_hosts:
            LOG.warn(_('No viable hosts found for instance %(inst)s, source '
                       'host %(host)s, and requested networks %(nets)s'),
                     {'inst': instance_uuid, 'host': source_host,
                      'nets': requested_networks})

        # Convert the viable hosts to the format expected by the scheduler
        viable_hosts = {}
        for candidate_host in candidate_hosts:
            host_name = {'host': candidate_host}
            viable_hosts[candidate_host] = host_name
        return viable_hosts

    def _get_compute_nodes_from_DB(self, context):
        """
        This returns a list of compute nodes from Nova ComputeNode objects

        :param context: A context object that is used to authorize
                        the DB access.
        :returns: A set of compute nodes.
        """
        context = context.elevated()
        return [compute.host
                for compute in comp_node_obj.ComputeNodeList.get_all(context)]

    def _create_port(self, port_client, instance, network_id, port_req_body,
                     fixed_ip=None, security_group_ids=None,
                     available_macs=None, dhcp_opts=None):
        """Attempts to create a port for the instance on the given network.

        :param port_client: The client to use to create the port.
        :param instance: Create the port for the given instance.
        :param network_id: Create the port on the given network.
        :param port_req_body: Pre-populated port request. Should have the
            device_id, device_owner, and any required neutron extension values.
        :param fixed_ip: Optional fixed IP to use from the given network.
        :param security_group_ids: Optional list of security group IDs to
            apply to the port.
        :param available_macs: Optional set of available MAC addresses to use.
        :param dhcp_opts: Optional DHCP options.
        :returns: ID of the created port.
        :raises PortLimitExceeded: If neutron fails with an OverQuota error.
        """
        port_client.retries = 5
        # Check to see if the port is in use already, if there is a fixed ip
        if fixed_ip is not None and network_id is not None:
            search_opts = {'network_id': network_id,
                           'fixed_ips': 'ip_address=%s' % fixed_ip}
            existing_ports = port_client.list_ports(**search_opts)

            # If there are any ports, then the IP is already in use on this
            # network.
            if existing_ports is not None and\
                    len(existing_ports.get('ports')) > 0:
                i_uuid = existing_ports.get('ports')[0].get('device_id')
                raise nova.exception.FixedIpAlreadyInUse(address=fixed_ip,
                                                         instance_uuid=i_uuid)

        # Need to change the available macs to be an ordered list, with the
        # lowest MAC as the last element.  This is needed because most AE's
        # will assume lowest mac to eth0, and highest mac to the last adapter.
        # The set of macs passed in pops in no determinate order...so we flip
        # to a list, reorder and then pass in the new list.
        ordered_macs = None
        if available_macs:
            ordered_macs = list(available_macs)
            ordered_macs.sort(reverse=True)

        # Temporary workaround until this is fixed in the community.  Where
        # when an IPAddress type is passed in, the Neutron Client will now
        # blow up trying to convert that type to a string, so need to convert
        if fixed_ip is not None and isinstance(fixed_ip, netaddr.IPAddress):
            fixed_ip = str(fixed_ip)

        # Create the actual port.
        #port = super(PowerBaseAPI, self)._create_port(port_client, instance,
        #                                              network_id,
        #                                              port_req_body, fixed_ip,
        #                                              security_group_ids,
        #                                              ordered_macs, dhcp_opts)
        try:
            if fixed_ip:
                port_req_body['port']['fixed_ips'] = [
                    {'ip_address': str(fixed_ip)}]
            port_req_body['port']['network_id'] = network_id
            port_req_body['port']['admin_state_up'] = True
            port_req_body['port']['tenant_id'] = instance.project_id
            if security_group_ids:
                port_req_body['port']['security_groups'] = security_group_ids
            if ordered_macs is not None:
                if not ordered_macs:
                    raise exception.PortNotFree(
                        instance=instance.uuid)
                mac_address = ordered_macs.pop()
                port_req_body['port']['mac_address'] = mac_address
            if dhcp_opts is not None:
                port_req_body['port']['extra_dhcp_opts'] = dhcp_opts
            port = port_client.create_port(port_req_body)
            port_id = port['port']['id']
            LOG.debug('Successfully created port: %s', port_id,
                      instance=instance)
            port = port_id
        except neutron_client_exc.InvalidIpForNetworkClient:
            LOG.warning(_LW('Neutron error: %(ip)s is not a valid ip address '
                            'for network %(network_id)s.'),
                        {'ip': fixed_ip, 'network_id': network_id})
            msg = (_('Fixed IP %(ip)s is not a valid ip address for '
                     'network %(network_id)s.') %
                   {'ip': fixed_ip, 'network_id': network_id})
            raise exception.InvalidInput(reason=msg)
        except neutron_client_exc.IpAddressInUseClient:
            LOG.warning(_LW('Neutron error: Fixed IP %s is '
                            'already in use.'), fixed_ip)
            msg = _("Fixed IP %s is already in use.") % fixed_ip
            raise exception.FixedIpAlreadyInUse(message=msg)
        except neutron_client_exc.OverQuotaClient:
            LOG.warning(_LW(
                'Neutron error: Port quota exceeded in tenant: %s'),
                port_req_body['port']['tenant_id'], instance=instance)
            raise exception.PortLimitExceeded()
        except neutron_client_exc.IpAddressGenerationFailureClient:
            LOG.warning(_LW('Neutron error: No more fixed IPs in network: %s'),
                        network_id, instance=instance)
            raise exception.NoMoreFixedIps(net=network_id)
        except neutron_client_exc.MacAddressInUseClient:
            LOG.warning(_LW('Neutron error: MAC address %(mac)s is already '
                            'in use on network %(network)s.') %
                        {'mac': mac_address, 'network': network_id},
                        instance=instance)
            raise exception.PortInUse(port_id=mac_address)
        except neutron_client_exc.NeutronClientException:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE('Neutron error creating port on network %s'),
                              network_id, instance=instance)

        # Find the used mac...and remove it from the original set
        if available_macs:
            mac_to_remove = None
            for mac in available_macs:
                if mac not in ordered_macs:
                    mac_to_remove = mac
                    break
            available_macs.remove(mac_to_remove)

        return port

    def _get_network_all(self, context):
        """
        This returns a list of networks after querying the Neutron DB
        wrote a getter here for the super class wrapper, we will remove
        this method once, we find a better way of mocking a super class
        object in pymox.

        :param context: A context object that is used to authorize
                        the DB access.
        :returns: A set of networks.
        """
        context = context.elevated()
        return super(PowerBaseAPI, self).get_all(context)

    def _get_viable_hosts(self, context, instance_uuid, hosts_from_db,
                          network_uuids, source_host=None):
        """
        Find the viable hosts that match the network requirements.

        :param context: A context object used to authorize the DB access.
        :param instance_uuid: The instance UUID that will be used to get the
                              network requirements of an instance.
        :param hosts_from_db: A list of compute nodes
        :param network_uuids: An list of requested networks for the deployment.
        :param source_host: When this API is invoked for LPM, the host that
                            is being migrated from.  This is needed to check
                            vswitch config on the source host.
        :returns: A dictionary of hosts with the hostid and the hostname
        """
        raise NotImplementedError('_get_viable_hosts not implemented')

    def validate_networks(self, context, requested_networks, num_instances):
        """Validate that the tenant can use the requested networks.

        Return the number of instances than can be successfully allocated
        with the requested network configuration.
        """
        LOG.debug(_('validate_networks() for %s'),
                  requested_networks)
        LOG.debug("This function is added as a temporary work around to "
                  " fix PowerVC STG Defect 19419 and this OpenStack community "
                  "bug: https://bugs.launchpad.net/nova/+bug/1304409. "
                  "This function should be removed when the community bug "
                  "is fixed and picked up by PowerVC.")
        neutron = neutronv2.api.get_client(context)
        neutron.retries = 5
        ports_needed_per_instance = 0

        if not requested_networks:
            nets = self._get_available_networks(context, context.project_id,
                                                neutron=neutron)
            if len(nets) > 1:
                # Attaching to more than one network by default doesn't
                # make sense, as the order will be arbitrary and the guest OS
                # won't know which to configure
                msg = _("Multiple possible networks found, use a Network "
                        "ID to be more specific.")
                raise exception.NetworkAmbiguous(msg)
            else:
                ports_needed_per_instance = 1

        else:
            net_ids = []

            for req_net in requested_networks:
                # Convert to a Tuple if the incoming is a NetworkRequest
                if isinstance(req_net, network_request.NetworkRequest):
                    req_net = req_net.to_tuple()
                net_id, _i, port_id, _pci_id = req_net
                if port_id:
                    try:
                        port = neutron.show_port(port_id).get('port')
                    except neutronv2.api.neutron_client_exc.\
                        NeutronClientException \
                            as e:
                        if e.status_code == 404:
                            port = None
                        else:
                            with excutils.save_and_reraise_exception():
                                LOG.exception(_("Failed to access port %s"),
                                              port_id)
                    if not port:
                        raise exception.PortNotFound(port_id=port_id)
                    if port.get('device_id', None):
                        raise exception.PortInUse(port_id=port_id)
                    net_id = port['network_id']
                else:
                    ports_needed_per_instance += 1
                # If multiple NICs from same network is not allowed, if
                # net_id is already in net_ids, raise exception
                if not CONF.multiple_nics_same_network:
                    if net_id in net_ids:
                        raise exception.NetworkDuplicated(network_id=net_id)

                # Add net_id to net_ids only if not already in it
                # Because, _get_avialable_networks returns unique list
                if net_id not in net_ids:
                    net_ids.append(net_id)
            LOG.debug(_('net_ids %s'), net_ids)
            # Now check to see if all requested networks exist
            nets = self._get_available_networks(context,
                                                context.project_id, net_ids,
                                                neutron=neutron)

            LOG.debug(_('nets %s'), nets)
            if len(nets) != len(net_ids):
                requsted_netid_set = set(net_ids)
                returned_netid_set = set([net['id'] for net in nets])
                lostid_set = requsted_netid_set - returned_netid_set
                id_str = ''
                for _id in lostid_set:
                    id_str = id_str and id_str + ', ' + _id or _id
                raise exception.NetworkNotFound(network_id=id_str)

        # Note(PhilD): Ideally Nova would create all required ports as part of
        # network validation, but port creation requires some details
        # from the hypervisor.  So we just check the quota and return
        # how many of the requested number of instances can be created
        quotas = neutron.show_quota(tenant_id=context.project_id)['quota']
        if quotas.get('port') == -1:
            # Unlimited Port Quota
            return num_instances
        else:
            ports = neutron.list_ports(tenant_id=context.project_id)['ports']
            free_ports = quotas.get('port') - len(ports)
            ports_needed = ports_needed_per_instance * num_instances
            if free_ports >= ports_needed:
                return num_instances
            else:
                return free_ports // ports_needed_per_instance

    def _build_network_info_model(self, context, instance, networks=None,
                                  port_ids=None, admin_client=None,
                                  preexisting_port_ids=None):
        """Return list of ordered VIFs attached to instance.

        :param context - request context.
        :param instance - instance we are returning network info for.
        :param networks - List of networks being attached to an instance.
                          If value is None this value will be populated
                          from the existing cached value.
        :param port_ids - List of port_ids that are being attached to an
                          instance in order of attachment. If value is None
                          this value will be populated from the existing
                          cached value.
        :param admin_client - a neutron client for the admin context.
        :param preexisting_port_ids - List of port_ids that nova didn't
                                      allocate and therefore shouldn't be
                                      deleted when an instance is deallocated.
                                      If value is None or empty the value will
                                      be populated from existing cached value.
        """
        net_infos = super(PowerBaseAPI, self)._build_network_info_model(
            context, instance, networks, port_ids, admin_client,
            preexisting_port_ids)
        if networks is None:
            return net_infos
        for network in networks:
            nets = self._find_net_infos_for_network(net_infos, network)
            if nets and network.get('mtu'):
                for net_info in nets:
                    net_info['network']['mtu'] = network.get('mtu')

        return net_infos

    def _find_net_infos_for_network(self, net_infos, network):
        """
        Finds the matching net_infos for the network

        :param net_infos: A set of Neutron Networks (the values returned from
                          the Neutron REST API)
        :param network: The specific NOVA network object to pair to a specific
                        Neutron Network (net_infos)
        :returns: The corresponding set of Neutron Network (net_info) for the
                  NOVA network.  Will return None if there is not an
                  appropriate Neutron Network for the Nova network.
        """
        resp = []
        for net_info in net_infos:
            if net_info['network']['id'] == network['id']:
                resp.append(net_info)
        return resp
