#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import ConfigParser
import StringIO
import commands
import copy
import json
from nova import utils
from oslo_concurrency import lockutils
from oslo_log import log as logging
import socket
import ast
from os import SEEK_SET
from oslo_config import cfg
import re
import tempfile
import time
from lxml import etree

from powervc_nova import _
from powervc_nova.network.powerkvm import agent
from powervc_nova.network.powerkvm.agent.common import exception as exp
from powervc_nova.objects.network import dom_kvm
from powervc_oslo.config import data_utils


LOG = logging.getLogger(__name__)

CONF = cfg.CONF
CONF.import_opt('integration_bridge', 'powervc_nova.network.powerkvm.agent')
CONF.import_opt('management_ip', 'powervc_nova.network.powerkvm.agent')
ADAPTER_REGEX = '([a-zA-Z0-9!@#$%^&\*_\-\[\]\{\}?\(\)]+)'

# Neutron only supports vSwitches with a certain number of characters.  Should
# ignore any of longer length
MAX_VSWITCH_CHARS = 8

CONF.register_opts([
    cfg.IntOpt('max_mgmt_adpt_ping_tries', default=20,
               help='Max attempts for ping request to management adapter')
])


class CommandExecutor():

    """
    cached value for lshw data
    """
    lshwCachedData = None

    @staticmethod
    @lockutils.synchronized('powerkvm_ovslshw', 'nova-ovslshw-powerkvm-')
    def send_lshw_command():
        """
        This method is used to send the lshw command. The command
        is sent to return a POSIX output. Call the parser method to
        parse the output and eventually send it back to the caller.
        No filtering is done at this level of the code.

        :return : [{'name': 'eth0', 'speed': '10Gbit/s'},
                {'name': 'eth3', 'speed': '10Gbit/s'},
                {'name': 'virbr1-nic', 'speed': '10Mbit/s'},
                {'name': 'virbr0-nic', 'speed': '10Mbit/s'}]

        """
        lshwout = ""
        try:
            lshwout, stderr = utils.execute('lshw', '-class', 'network',
                                            run_as_root=True)
            if stderr and stderr != "":
                raise exp.IBMPowerKVMCommandExecError(command="lshw",
                                                      exp=stderr)
        except OSError as ex:
            LOG.error(ex)
        except Exception as e:
            LOG.error(e)
            raise
        # TODO: If an stderr is encoutered throw an exception.
        return CommandParser.parse_lshw(lshwout)

    @staticmethod
    def send_bond_info_command():
        """
        This method is used to retrieve the bond information. We don't have a
        direct of determining this and hence a set of commands are sent.

        :return : A parsed list of bond data
        """
        bond_data_dict = {}
        bonds = ""
        stderr = 'Unknown'
        try:
            directory = "/proc/net/bonding"
            file_exists = _check_file_exists(directory)
            if file_exists:
                bonds = _list_directory(directory)
            LOG.debug("The bonds are: %s", bonds)
        except Exception as e:
            bonds = None
            LOG.error(e)
        try:
            if bonds:
                # It means there are bonds.
                for bond in bonds:
                    if bond != "":
                        comm = "/proc/net/bonding/" + bond
                        bondinfo, stderr = utils.execute('cat', comm,
                                                         run_as_root=True)
                        # parse the command output but not here. Call the
                        # parser
                        slave_list = CommandParser.parse_bond_data(bondinfo)
                        bond_data_dict[bond] = slave_list
                        if stderr and stderr != "":
                            raise exp.IBMPowerKVMCommandExecError(cmd="bond",
                                                                  exp=stderr)
                        LOG.debug("Bond info for %s is: %s", bond, bondinfo)
        except OSError as ex:
            LOG.error(ex)
            bond_data_dict = ""
        except Exception as e:
            # LOG and move on
            LOG.error(_(
                "There was a problem in executing cat %(command)s "
                " The system returned the following error: %(error)s") %
                dict(command=comm, error=stderr))
            LOG.error(e)
            raise
        return bond_data_dict

    @staticmethod
    def get_lshw_results():
        """
        return cached lshw data

        :return: lshwCachedData
        """
        if CommandExecutor.lshwCachedData is None:
            # If its none, we should run it to gather the data.
            data = CommandExecutor.send_lshw_command()
            CommandExecutor.lshwCachedData = data
        return CommandExecutor.lshwCachedData

    @staticmethod
    def set_lshw_results(value):
        """
        set lshw cached data
        """
        CommandExecutor.lshwCachedData = value

    @staticmethod
    def send_ofctl_dump_flows(bridge):
        '''
        This method dumps the flows for the input bridge name and returns
        the parsed output.

        :returns: List of dictionaries representing the flows on the passed in
                  bridge.  See CommandParser.parse_ofctl_dump_flows for the
                  format.
        '''
        if not bridge:
            return list()

        ofctlout = ''
        try:
            ofctlout, stderr = utils.execute('ovs-ofctl', 'dump-flows',
                                             bridge, run_as_root=True)
            if stderr and stderr != "":
                raise exp.IBMPowerKVMCommandExecError(cmd="ovs-ofctl dump-"
                                                      "flows", exp=stderr)
        except OSError as ex:
            LOG.error(ex)
        except Exception as e:
            LOG.error(e)
            raise
        LOG.debug("ovs-ofctl dump-flows output = %s", ofctlout)
        return CommandParser.parse_ofctl_dump_flows(ofctlout)

    @staticmethod
    def send_ofctl_show(bridge):
        '''
        This method dumps the flow ports for the input bridge name returns
        the parsed output.

        :returns: List of dictionaries representing the flow ports on the
                  passed in bridge.  See CommandParser.parse_ofctl_show
                  for the format.
        '''
        if not bridge:
            return list()

        ofctlout = ''
        try:
            ofctlout, stderr = utils.execute('ovs-ofctl', 'show',
                                             bridge, run_as_root=True)
            if stderr and stderr != "":
                raise exp.IBMPowerKVMCommandExecError(cmd="ovs-ofctl show",
                                                      exp=stderr)
        except OSError as ex:
            LOG.error(ex)
        except Exception as e:
            LOG.error(e)
            raise
        LOG.debug("ovs-ofctl show output = %s", ofctlout)
        return ofctlout

    @staticmethod
    def send_ofctl_show_order_by_port_id(bridge):
        ofctlout = CommandExecutor.send_ofctl_show(bridge)
        return CommandParser.parse_ofctl_show_order_by_port_id(ofctlout)

    @staticmethod
    def send_ofctl_show_order_by_port_name(bridge):
        ofctlout = CommandExecutor.send_ofctl_show(bridge)
        return CommandParser.parse_ofctl_show_order_by_port_name(ofctlout)

    @staticmethod
    def send_virsh_list(libvirt_conn, list_all=False):
        '''
        This method will return a dictionary containing the name of each
        domain and its id, similar to what is returned by the virsh list
        command.

        :param libvirt_conn: connection to libvirt
        :param list_all: If set to True, this will return all domains
            whether they are active or inactive.  When set to False,
            only active domains will be returned.
        :returns: dictionary containing domain names and ids in the
            format of {"vm_name1": "1", "vm_name2": "2", ...}
        '''
        domain_dict = {}

        try:
            for dom in libvirt_conn.listAllDomains():
                if (not list_all and dom.isActive()) or list_all:
                    domain_dict[dom.name()] = dom.ID()
        except Exception as ex:
            LOG.exception(ex)

        return domain_dict

    @staticmethod
    def send_virsh_domiflist(libvirt_conn, dom_name):
        '''
        This method will return a dictionary containing information on
        each interface for the specified domain, similar to what is
        returned by the virsh <domain id> domiflist command.

        :param libvirt_conn: connection to libvirt
        :param dom_name: name of the domain
        :returns: dictionary containing a list of interfaces in the
            format of
            [ {"interface": "tapxxx", "type": "bridge", "source": "br-int",
               "model": "virtio", "mac_address": "fa:16:3e:49:80:4a"},
              {"interface": "tapyyy", "type": "bridge", "source": "br-int",
               "model": "virtio", "mac_address": "fa:16:3e:49:80:4b"},
              ...
            ]
        '''
        interfaces = []

        try:
            # get the virDomain based on specified dom id
            dom = libvirt_conn.lookupByName(dom_name)
            domxml = dom.XMLDesc()
            doc = etree.fromstring(domxml)

            # get the interface elements from the domain
            ret = doc.findall('./devices/interface')

            for node in ret:
                devdst = {}

                for child in list(node):
                    if child.tag == 'target':
                        devdst['interface'] = child.attrib['dev']
                    elif child.tag == 'source':
                        # Set the source and known_vlan based on child type
                        if child.attrib.get('bridge'):
                            # Type bridge.
                            devdst['source'] = child.attrib.get('bridge')
                            devdst['known_vlan'] = \
                                CommandExecutor._find_vlan(
                                    child.attrib.get('bridge'))
                        elif child.attrib.get('network'):
                            # Type network.  This could be a virbrX linuxbridge
                            # or it could be a kimchi bridge on a guest that
                            # hasn't started up before.  Either way, we can
                            # handle it.
                            vlan, src = CommandExecutor.find_network_info(
                                libvirt_conn, child.attrib.get('network'))
                            devdst['source'] = src
                            devdst['known_vlan'] = vlan
                        elif node.attrib.get('type') == 'direct' and \
                                child.attrib.get('mode') == 'bridge':
                            # This is a kimchi bridge.
                            devdst['source'] = child.attrib.get('dev')
                            devdst['known_vlan'] = \
                                CommandExecutor._find_vlan(
                                    child.attrib.get('dev'))
                        else:
                            # Not one of the known bridge types, error.
                            raise exp.IBMPowerKVMNonBridgeInterfaces
                    elif child.tag == 'model':
                        devdst['model'] = child.attrib['type']
                    elif child.tag == 'mac':
                        devdst['mac_address'] = child.attrib['address']

                devdst['type'] = node.attrib.get('type')

                # If the type is network or direct, flip that to bridge,
                # because it is a bridge under the covers
                if devdst['type'] == 'network' or devdst['type'] == 'direct':
                    devdst['orig_type'] = devdst['type']
                    devdst['type'] = 'bridge'

                # If the type is not bridge, then we need to raise an error
                if devdst['type'] != 'bridge':
                    raise exp.IBMPowerKVMNonBridgeInterfaces

                if devdst == {}:
                    continue

                interfaces.append(devdst)

        except exp.IBMPowerKVMNonBridgeInterfaces as nbi_ex:
            LOG.error(nbi_ex)
            raise
        except Exception as ex:
            LOG.exception(ex)

        return interfaces

    @staticmethod
    def find_network_info(libvirt_conn, network_name):
        '''
        Some devices are created with 'networks' rather than straight attach
        to bridges.  This is usually done via Kimchi or other VM managers.

        This method will return the network info when not direct attached
        to a VM
        '''
        virsh_net = libvirt_conn.networkLookupByName(network_name)
        net_xml = virsh_net.XMLDesc()
        doc = etree.fromstring(net_xml)

        # First check is to see if it has a bridge.
        ret = doc.findall('./bridge')
        if len(ret) > 0:
            name = ret[0].attrib['name']
            vlan = CommandExecutor._find_vlan(name)
            # TODO FIX
            return vlan, name

        # The next approach is to accept 'forwards' direct to devices.  So
        # they are guaranteed to be flat.
        ret = doc.findall('./forward')
        if len(ret) > 0:
            mode = ret[0].attrib['mode']
            if mode == 'bridge':
                return -1, ret[0].attrib['dev']

        # If we're here...then the network wasn't found as a supported type.
        # Need to throw an error
        raise exp.IBMPowerKVMNonBridgeInterfaces

    @staticmethod
    def send_vsctl_command():
        """
        This method is used to send the ovs-vsctl command and get the parsed
        output

        :return: parsed dictionary of the object of the vsctl command
        {'br0': {'br0': {'interface': ['eth0'], 'type': 'internal'},
                '"9"': {'interface': ['"9"']},
                '"br1"': {'interface': ['"br1"', '"arbit"'],
                                   'type': 'internal'}}}
        """
        vsctlout = ""
        try:
            vsctlout, stderr = utils.execute('ovs-vsctl', 'show',
                                             run_as_root=True)
            if stderr and stderr != "":
                raise exp.IBMPowerKVMCommandExecError(cmd="ovs-vsctl",
                                                      exp=stderr)
        except OSError as ex:
            LOG.error(ex)
        except Exception as e:
            LOG.error(e)
            raise
        LOG.debug("ovs-vsctl show output = %s", vsctlout)
        return CommandParser.parse_ovs_vsctl_show(vsctlout)

    @staticmethod
    def send_vsctl_list_port_command():
        """
        This method is used to send the ovs-vsctl list port command and get the
        parsed output.

        :return: parsed dictionary of the object of the vsctl command
        """
        vsctlout = ""
        try:
            vsctlout, stderr = utils.execute('ovs-vsctl', '--timeout=60',
                                             'list', 'port', run_as_root=True)
            if stderr and stderr != "":
                raise exp.IBMPowerKVMCommandExecError(cmd='ovs-vsctl',
                                                      exp=stderr)
        except OSError as ex:
            LOG.error(ex)
        except Exception as e:
            LOG.error(e)
            raise
        LOG.debug('ovs-vsctl list port output = %s', vsctlout)
        return CommandParser.parse_ovs_vsctl_list_port(vsctlout)

    @staticmethod
    def get_powerkvm_version():
        '''
        This method will send the uname -r command to get the PowerKVM version.

        :return: String representing PowerKVM version.  eg: 2_1_1 or 3_1_0
        '''
        uname = ''
        try:
            uname, stderr = utils.execute('uname', '-r', run_as_root=True)
            if stderr and stderr != '':
                raise exp.IBMPowerKVMCommandExecError(cmd='uname', exp=stderr)
        except OSError as ex:
            LOG.error(ex)
        except Exception as ex:
            LOG.error(ex)
            raise
        LOG.debug('uname -r output = %s', uname)
        return uname.split('pkvm')[1].split('.')[0] if uname else ""

    @staticmethod
    def send_ifconfig_command():
        """
        This method is used to send the ifconfig -a command.

        :return: parsed list of dictionary of the ifconfig output
                    [{'name': 'eth0', 'state': 'UP',
                     'ipv4': '9.79.214.43', 'ipv6': ''},
                    {'name': 'eth3', 'state': 'UP',
                     'ipv4': '127.0.0.1', 'ipv6': ''},
                    {'name': 'wc0', 'state': 'UP',
                     'ipv4': '9.79.214.43', 'ipv6': ''}]
        """
        ifcfgout = ""
        try:
            ifcfgout, stderr = utils.execute('ifconfig', '-a',
                                             run_as_root=True)
            if stderr and stderr != "":
                raise exp.IBMPowerKVMCommandExecError(cmd="ifconfig",
                                                      exp=stderr)
        except OSError as ex:
            LOG.error(ex)
        except Exception as e:
            LOG.error(e)
            raise
        LOG.debug("ifconfig output = %s", ifcfgout)
        return CommandParser.parse_ifconfig(ifcfgout)

    @staticmethod
    def send_ip_address_command():
        """
        This method is used to send the ifconfig -a command.

        :return: parsed list of dictionary of the ifconfig output
                    [{'name': 'eth0', 'state': 'UP',
                     'ipv4': '9.79.214.43', 'ipv6': ''},
                    {'name': 'eth3', 'state': 'UP',
                     'ipv4': '127.0.0.1', 'ipv6': ''},
                    {'name': 'wc0', 'state': 'UP',
                     'ipv4': '9.79.214.43', 'ipv6': ''}]
        """
        ip_address_out = ""
        try:
            ip_address_out, stderr = utils.execute('ip', 'address',
                                                   run_as_root=True)
            if stderr and stderr != "":
                raise exp.IBMPowerKVMCommandExecError(cmd="ip address",
                                                      exp=stderr)
        except OSError as ex:
            LOG.error(ex)
        except Exception as e:
            LOG.error(e)
            raise
        LOG.debug("ip address command output = %s", ip_address_out)
        return CommandParser.parse_ip_address_command(ip_address_out)

    @staticmethod
    def send_brctl_command():
        """
        This method is used to send the brctl command

        :return: a dict with the brctl output
          {
            'bridges': [
                {
                    'interfaces': [
                        {
                            'name': 'virbr0-nic'
                        },
                        {
                            'name': 'test-nic'
                        },
                        {
                            'name': 'test-nic2'
                        }
                    ],
                    'bridge_name': 'virbr0'
                },
                {
                    'interfaces': [
                        {
                            'name': 'virbr1-nic'
                        }
                    ],
                    'bridge_name': 'virbr1'
                },
                {
                    'interfaces': [],
                    'bridge_name': 'virbr2'
                }
            ]
        }
        """
        try:
            brctlout, stderr = utils.execute('brctl', 'show',
                                             run_as_root=True)
            if stderr and stderr != "":
                raise exp.IBMPowerKVMCommandExecError(cmd="brctl",
                                                      exp=stderr)
        except Exception as e:
            LOG.error(e)
            raise
        LOG.debug("brctl output = %s" % brctlout)
        return CommandParser.parse_brctl(brctlout)

    @staticmethod
    def get_ovs_bond_port_details(bond_name):
        """
        This method is used to get the bond port details primarily to detect
        LACP specific configuration. Even though we don't care much about
        how the information returned is used.
        :param bond_name: The name of the bond for which details are needed.
        :return bond_info: A dictionary with the details of the bond.
        """
        args = ["list", "port", bond_name]
        try:
            bond_info_raw, stderr = CommandExecutor.run_vsctl(args, True)
            if stderr and stderr != "":
                raise exp.IBMPowerKVMCommandExecError(cmd="list port",
                                                      exp=stderr)
        except Exception as e:
            LOG.error(e)
            raise
        LOG.debug("The bond_info output = %s", bond_info_raw)
        return CommandParser.parse_ovs_bond_info(bond_name, bond_info_raw)

    @staticmethod
    def run_vsctl(args, check_error=False):
        """
        This is a utility method to run ovs-vsctl commands. This could be used
        to do various operations that involve ovs-vsctl. Note we can also pass
        a json format for outputs - so use it per your discretion
        :param args: A list of arguments that can be sent to the ovs-vsctl com
        for execution.
        :param check_error: This should be set to True, if you would like the
        exception to be propagated.
        :return stdout: A standard output after running the command
        """
        full_args = ["ovs-vsctl", "--timeout=2"] + args
        stdout = ''
        stderr = ''
        try:
            stdout, stderr = utils.execute(*full_args,
                                           run_as_root=True)
            if stderr and stderr != "":
                raise exp.IBMPowerKVMCommandExecError(cmd="ovs-vsctl",
                                                      exp=stderr)
        except Exception as e:
            LOG.error(_("Unable to execute %(cmd)s. Exception: %(exception)s"),
                      {'cmd': full_args, 'exception': e})
            if check_error:
                raise
        return stdout, stderr

    @staticmethod
    def run_ping(target, adapter=None):
        """
        This is a utility method to run ping command. Will run exactly one
        ping to the target and return whether or not it was successful.
        :param target: The host to run the ping against
        :param adapter: Defaults to None (optional).  If specified, the
                        adapter to run the ping through.
        :return success: True if ping worked.  False if it is not routable.
        """
        result = None
        try:
            result = socket.getaddrinfo(target, 80, 0,
                                        socket.SOCK_STREAM)
        except Exception as e:
            LOG.error(_("The specified IP address %s is not valid") % target)

        if result:
            socket_family = result[0][0]
            if socket_family == socket.AF_INET6:
                full_args = "ping6"
            else:
                full_args = "ping"
        else:
            # We default to ping
            full_args = "ping"
        if adapter is not None:
            full_args = full_args + " -I " + adapter
        full_args = full_args + " -c 1 -W 1 " + target
        try:
            status, output, stderr = data_utils.run_command_output(full_args)
            if status != 0:
                LOG.error(_("Ping failed with return code %(rc)s and stderr: "
                            "%(stderr)s"), {'rc': status, 'stderr': stderr})
            return status == 0
        except Exception as e:
            # Some other fundamental error occurred
            LOG.error(_("Unable to execute %(cmd)s. Exception: %(exception)s"),
                      {'cmd': full_args, 'exception': e})
            return False

    @staticmethod
    def get_ifcfg(device_name):
        """
        This method is used to get the contents of a networking ifcfg file
        for a specified device.

        :param device_name: The networking device to get the ifcfg for.
        :return: The fully qualified path of the ifcfg file.
        :return: a dict with the key/value pairs from the ifcfg file.
        """
        LOG.debug('searching for ifcfg file for %s' % device_name)
        try:
            fpath, ifcfg_dict = CommandExecutor._find_ifcfg_data(device_name)
            return fpath, ifcfg_dict

        except Exception as e:
            LOG.exception(e)
            raise
        LOG.debug('ifcfg file not found for %s' % device_name)
        return None, None

    @staticmethod
    def send_ifcfg(device_name, device_dict):
        """
        This method is used to update the contents of a networking ifcfg file
        for a specified device.

        :param device_name: The networking device to update the ifcfg for.
        :param device_dict: a dict with the key/value pairs for the ifcfg file.
        :returns fnamepath: The path and name of the file data was written to.
        """
        try:
            fname = CommandExecutor._find_ifcfg_data(device_name)[0]
            if not fname:
                fname = CommandExecutor._generate_ifcfg_name(device_name)
            lines = []
            for key in device_dict.keys():
                lines.append(key + '=' + device_dict[key] + '\n')

            # Write the output as root
            _write_to_file_as_root(fname, lines)

            return fname
        except Exception as e:
            LOG.error(e)
            raise
        return None

    @staticmethod
    def _find_vlan(bridge_name):
        """
        find and return the vlan associated with the specified linux bridge

        :param bridge_name: name of the linux bridge
        """
        if_name = None
        brctl_dict = CommandExecutor.send_brctl_command()
        if brctl_dict and brctl_dict['bridges']:
            for bridge in brctl_dict['bridges']:
                if bridge.get('bridge_name') == bridge_name:
                    interfaces = bridge.get('interfaces')
                    if interfaces and len(interfaces) > 0:
                        if len(interfaces) > 1:
                            LOG.warning(_("More than one interface found on "
                                          "bridge %s.  The first interface "
                                          "will be used.") % bridge_name)
                        if_name = interfaces[0]['name']
                        break

        if if_name:
            # parse the vlan file
            vlan_filename = '/proc/net/vlan/config'
            if _check_file_exists(vlan_filename):
                try:
                    comm = vlan_filename
                    vlaninfo, stderr = utils.execute('cat', comm,
                                                     run_as_root=True)

                    if stderr and stderr != "":
                        raise exp.IBMPowerKVMCommandExecError(cmd="vlan_conf",
                                                              exp=stderr)

                    if vlaninfo is None or vlaninfo == "":
                        # no vlan info, return -1 to indicate flat network
                        LOG.debug("No vlan info available in file, returning "
                                  "as flat network")
                        return -1

                    # parse the file data
                    outputList = vlaninfo.splitlines()
                    # skip first two header lines
                    for i in range(2, len(outputList)):
                        LOG.debug("current vlan entry = %s" % outputList[i])
                        splitStr = outputList[i].split('|')
                        if len(splitStr) > 1 and \
                                splitStr[0].strip() == if_name:
                            return int(splitStr[1].strip())

                    LOG.debug("vlan conf info: %s" % vlaninfo)
                except OSError as ex:
                    LOG.error(ex)
                except Exception as e:
                    # LOG and move on
                    LOG.error(_("There was a problem in executing cat "
                                "%(command)s The system returned the "
                                "following error: %(error)s") %
                              dict(command=comm, error=stderr))
                    LOG.error(e)
                    raise

        # no matching vlan found, assume this is a flat network
        return -1

    @staticmethod
    def _find_ifcfg_data(device_name):
        """
        Look up the ifcfg file name and contents for a specified device_name.

        :param device_name: A device name to look up such as eth0.
        :return: The fully qualified path of the ifcfg file.
        :return: A dictionary with the parsed file contents.
        """
        ifcfg_dir = '/etc/sysconfig/network-scripts'

        # The device_name is usually in the ifcfg file name, so check for this
        # case first.
        fname = '%s/ifcfg-%s' % (ifcfg_dir, device_name)
        if _check_file_exists(fname):
            stdout = utils.execute('cat', fname, run_as_root=True)[0]
            ifcfg_dict = CommandParser.parse_ifcfg(device_name, stdout)

            # make sure there are no quote around the device name we
            # get from the ifcfg file
            # ifcfg_dev_name = \
            #    re.sub(r'^"|"$', '', ifcfg_dict.get(agent.IFCFG_DEVICE, ''))
            ifcfg_dev_name = ifcfg_dict.get(agent.IFCFG_DEVICE, '')
            if ifcfg_dev_name == device_name:
                LOG.debug('found ifcfg file in expected location.')
                return fname, ifcfg_dict

        # The ifcfg with the desired device name either does not exist, or
        # contains a different internal device name.  Search all ifcfg files
        # looking for the correct device.
        ifcfg_files = [f for f in _list_directory(ifcfg_dir)
                       if _check_file_exists('%s/%s' % (ifcfg_dir, f))]
        for ifcfg_file in ifcfg_files:
            fname = '%s/%s' % (ifcfg_dir, ifcfg_file)
            stdout = utils.execute('cat', fname, run_as_root=True)[0]
            ifcfg_dict = CommandParser.parse_ifcfg(device_name, stdout)

            # make sure there are no quote around the device name we
            # get from the ifcfg file
            # ifcfg_dev_name = \
            #    re.sub(r'^"|"$', '', ifcfg_dict.get(agent.IFCFG_DEVICE, ''))
            ifcfg_dev_name = ifcfg_dict.get(agent.IFCFG_DEVICE, '')
            if ifcfg_dev_name == device_name:
                LOG.debug('found ifcfg data in file %s.' % ifcfg_file)
                return fname, ifcfg_dict
        return None, {}

    @staticmethod
    def _generate_ifcfg_name(device_name):
        """
        Generate an ifcfg name for the specified device.  This method addresses
        the following case:

        -> host-ovs API caller wants to update eth0 config.
        -> ifcfg-eth0 exists, but has Device set to eth1.
        -> This method will generate ifcfg-eth0-1, or the next available file
           name.

        :param device_name: A device name to generate a file name for.
        :return: The fully qualified path of the ifcfg file.
        """
        # Check for simple case
        ifcfg_dir = '/etc/sysconfig/network-scripts'
        desired_fname = '%s/ifcfg-%s' % (ifcfg_dir, device_name)
        if not _check_file_exists(desired_fname):
            return desired_fname

        # Begin generating names, find first one that's not in use
        for i in range(0, 100000):
            fname = desired_fname + '-' + str(i)
            if not _check_file_exists(fname):
                return fname

        # This should not happen ?
        return None

    @staticmethod
    def get_all_ifcfg_files_for_logging():
        """
        Get the contents of all ifcfg files.  These contents are generally
        needed to debug KVM networking reconfiguration problems.

        :returns: All files as a string list.
        """
        ifcfg_dir = '/etc/sysconfig/network-scripts'
        ifcfg_files = [f for f in _list_directory(ifcfg_dir)
                       if _check_file_exists('%s/%s' % (ifcfg_dir, f))]
        return_ifcfgs = list()
        for ifcfg_file in ifcfg_files:
            fname = '%s/%s' % (ifcfg_dir, ifcfg_file)
            stdout = utils.execute('cat', fname, run_as_root=True)[0]
            ifcfg_dict = CommandParser.parse_ifcfg('arbitrary_string', stdout)
            if ifcfg_dict and len(ifcfg_dict):
                return_ifcfgs.append(_('Contents of %s') % (ifcfg_file))
                return_ifcfgs.append(json.dumps(ifcfg_dict, sort_keys=True,
                                                indent=4))
        return return_ifcfgs

    @staticmethod
    @lockutils.synchronized('powerkvm_ovslshw', 'nova-ovslshw-powerkvm-')
    def send_network_restart():
        """
        This method is used to restart the networking on the host.
        """
        try:
            stdout = utils.execute('systemctl', 'restart', 'network.service',
                                   run_as_root=True)[0]
            LOG.debug("service network restart output = %s" % stdout)
        except Exception as e:
            LOG.exception(e)

    @staticmethod
    def restart_neutron_agent():
        """
        This method is used to restart the neutron agent on the host.
        """
        try:
            stdout = utils.execute('systemctl', 'restart',
                                   'neutron-openvswitch-agent.service',
                                   run_as_root=True)[0]
            LOG.debug("systemctl restart neutron-openvswitch-agent "
                      "output = %s" % stdout)
        except Exception as e:
            LOG.exception(e)


class CommandParser():

    """
    This class is used to parse various commands in dictionary or list formats
    This class could be made a private class, if we don't want this to be used
    from outside this module.
    """

    @staticmethod
    def parse_ovs_vsctl_show(ovs_vsctl_output):
        """
        Parses OpenVSwitch's ovs-vsctl command output to find network bridges.

        :param ovs_vsctl_output: A string with the raw ovs-vsctl command
        output.
        :return dict_list: {'"br0"':
                            {' "br0"':
                                {'interface': ['"br0"'], 'type': 'internal'}
                            }
                        '}
        """
        if ovs_vsctl_output is None or ovs_vsctl_output == "":
            raise exp.IBMPowerKVMInvalidVSCTLOutput()
        i = 0
        bridge_dict = {}
        # Formulate the bridge dictionary to contain the bridge status for
        # vsctl
        for bridge_chunk in ovs_vsctl_output.split('\n    Bridge '):
            if bridge_chunk.strip() == "":
                continue
            if i == 0:
                i = i + 1
                continue
            j = 0
            # Let'stripe off the escape characters if used
            bridge_name = bridge_chunk.split("\n")[0].strip().strip('""')
            port_dict = {}
            for port_chunk in bridge_chunk.split('\n        Port '):
                # In that sub-chunk we would want to split it using the
                # port.
                if j != 0:
                    split_port_list = port_chunk.split('\n')
                    # get the port name out from the first element of the
                    # list
                    port_name = split_port_list[0].strip().strip('""')
                    interface_list = []
                    interface_dict = {}
                    # Next we get the items off the split port chunk
                    for item in split_port_list:
                        if "            Interface " in item:
                            iface = item.replace('            Interface ',
                                                 '', 1).strip().strip('""')
                            interface_list.append(iface)
                        if "type" in item:
                            interface_dict['type'] = item.split()[1]
                        if "tag:" in item:
                            interface_dict['tag'] = item.split()[-1].strip()
                    interface_dict['interface'] = interface_list
                    port_dict[port_name] = interface_dict
                j = j + 1
            bridge_dict[bridge_name] = port_dict
            i = i + 1
        return bridge_dict

    @staticmethod
    def parse_ovs_vsctl_list_port(ovs_vsctl_output):
        """
        Parses OpenVSwitch's ovs-vsctl list port command output to find ovs
        port and bond information.

        :param ovs_vsctl_output: A string with the raw ovs-vsctl command
                                 output.
        :return output_dict: The parsed command output.
        """
        if not ovs_vsctl_output:
            raise exp.IBMPowerKVMInvalidVSCTLPortOutput()

        # Split the command output into multi-line chunks, with one chunk
        # for each OVS port
        all_ovs_port_dict = {}
        for port_chunk in ovs_vsctl_output.split('\n_uuid'):
            if not port_chunk or not port_chunk.strip():
                continue

            # Each line of data about this port contains an attribute,
            # represented as a key/value pair separated by a colon (:)
            attr_dict = {}
            for port_attr in port_chunk.split('\n'):

                # Split up the key/value pairs
                attr_parts = port_attr.rsplit(':')
                if len(attr_parts) < 2:
                    continue
                key = attr_parts[0].strip()
                value = attr_parts[1].strip()

                # Special case for the first line - this will be the UUID
                if not key:
                    key = '_uuid'

                # Handle various attribute parsing cases
                value = value.replace('"', '')
                if value == '{}' or value == '[]':
                    value = ''
                elif value.startswith('['):
                    value = value.replace('[', '', 1).replace(']', '')
                elif value.startswith('{'):
                    value = value.replace('{', '', 1).replace('}', '')

                # Store the resulting attribute for this OVS port
                attr_dict[key] = value

            # Store all attributes for this OVS port
            all_ovs_port_dict[attr_dict['name']] = attr_dict

        return all_ovs_port_dict

    @staticmethod
    def parse_bond_data(bonddata):
        """
        Parse the bond related data for each bond found in the system.

        :return: slave_list: [{'state': 'up', 'speed': 10000, 'name': 'eth0'},
        {'state': 'up', 'speed': 10000, 'name': 'eth1'}]

        """
        # split the bond data into paragraphs
        if bonddata is None or bonddata == "":
            raise exp.IBMPowerKVMInvalidBondOutput()
        slave_list = []
        # Similar explanation of the regex construct is described in the
        # ifconfig parser.
        reg_slv = 'Slave Interface:[ ]*' + ADAPTER_REGEX
        slave_int_re = re.compile(reg_slv)
        slave_speed_re = re.compile('Speed:[ ]*([a-zA-Z0-9]+)')
        slave_state_re = re.compile('MII Status:[ ]*([a-z]+)')
        for para in bonddata.split("\n\n"):
            if "Slave Interface" in para:
                slave_name = ""
                slave_speed = 0
                match_slave_name = slave_int_re.search(para)
                match_slave_speed = slave_speed_re.search(para)
                match_slave_state = slave_state_re.search(para)
                if match_slave_name:
                    slave_name = match_slave_name.group(1)
                if match_slave_speed:
                    slave_speed = match_slave_speed.group(1)
                if match_slave_state:
                    slave_state = match_slave_state.group(1)
                # We don't want to create objects here.
                if slave_name and slave_state and slave_speed:
                    slave_info_dict = {}
                    slave_info_dict['name'] = slave_name
                    try:
                        slave_info_dict['speed'] = int(slave_speed)
                    except Exception as e:
                        LOG.debug(e)
                        slave_info_dict['speed'] = 0
                    slave_info_dict['state'] = slave_state
                    slave_list.append(slave_info_dict)
        return slave_list

    @staticmethod
    def parse_ip_address_command(ip_address_output):
        """
        This method replaces the ifconfig parser and parses the data out of the
        "ip address" command instead.

        :param ip_address_output: The output from the ip address command.
        :return dict_list: A list of dictionary object:
        """
        if ip_address_output is None or ip_address_output == "":
            raise exp.IBMPowerKVMInvalidIPAddressOutput()
        dict_list = []
        ipv4_value = None
        ipv6_value = []
        regex_dev = '[0-9]+:[ ]+' + ADAPTER_REGEX
        device_re = re.compile(regex_dev)
        ipv4re = re.compile('inet [ ]*([0-9\.]+)')
        ipv6re = re.compile('inet6[ ]*([0-9a-f:]+)')
        interface_name = None

        for line in ip_address_output.split('\n'):
            match_interface = device_re.search(line)
            if match_interface:
                old_interface_name = interface_name
                interface_name = match_interface.group(1)
            match_ipv4 = ipv4re.search(line)
            if match_ipv4:
                ipv4_value = match_ipv4.group(1)
            match_ipv6 = ipv6re.search(line)
            if match_ipv6:
                if "scope global" in line:
                    ipv6_value.append(match_ipv6.group(1))

            # If we match on interface, then we are wrapped around
            if match_interface and old_interface_name is not None:
                dict_list.append({'name': old_interface_name,
                                  'ipv4': ipv4_value,
                                  'ipv6': ipv6_value})
                ipv4_value = None
                ipv6_value = []

        # last time through assume an add
        dict_list.append({'name': interface_name, 'ipv4': ipv4_value,
                          'ipv6': ipv6_value})

        return dict_list

    @staticmethod
    def parse_ovs_bond_info(bond_name, raw_bond_info):
        """
        This method is used to parse the ovs bond information.

        :param raw_bond_info: The raw bond output obtained from ovs-vsctl
        :param bond_name: Name of the bond port
        :return bond_info_dict: Contains the information of the bond.
        """
        if raw_bond_info is None or raw_bond_info == "":
            raise exp.IBMPowerKVMInvalidBondInfoOutput(bond=bond_name)
        bond_info_dict = {}
        for line in raw_bond_info.split('\n'):
            key_value_list = line.split(":")
            if len(key_value_list) == 2:
                key = key_value_list[0].strip()
                value = key_value_list[1].strip()
                value = CommandParser._convert_to_primitive(value)
                bond_info_dict[key] = value
        return bond_info_dict

    @staticmethod
    def _convert_to_primitive(value):
        """
        This is a internal parser utility method that take a value in string
        format and returns the converted primitive format
        :param value: The value to convert
        :return prim_val: The primitive converted value
        """
        # Treat the booleans in the string
        if value == "false":
            value = False
        elif value == "true":
            value = True
        # Treat the lists in the string
        elif value.startswith('[') and value.endswith(']'):
            # A direct eval will not work due to the uuids pattern
            value = value[1:-1]
            if value == '':
                value = []
            else:
                value = value.split(',')
        # Treat the dictionary in the string
        elif value.startswith('{') and value.endswith('}'):
            value = ast.literal_eval(value)
        # Treat the integers
        elif value.isdigit():
            value = int(value)
        return value

    @staticmethod
    def parse_ifconfig(ifconfig_output):
        """
        This method is used to parse the ifconfig output.

        :param ifconfig_output: A string containing the inconfig shell output
        :return dict_list: A list of dictionary object:
            {'interface': 'eth0', 'state': 'UP', 'ipv4': '', 'ipv6': ''}
        """
        # TBD: This could be better done with regular expression, so can be mod
        if ifconfig_output is None or ifconfig_output == "":
            raise exp.IBMPowerKVMInvalidIFCONFIGOutput()
        dict_list = []
        # This is how the regular expr can be decoded. Search for a pattern
        # in the string that starts with 'inet addr' followed by 0 or more " "
        # this is denoted by [ ]*. Followed by an evaluation statement that
        # starts with ( and says this could be any number between 0 to 9
        # followed by a . (\. - escape character used) that can repeat for one
        # or more time (denoted by +). The value inside the "()" denotes the
        # expression whose value would be found by the search command.
        ipv4re = re.compile('inet [ ]*([0-9\.]+)')
        # Same as above, only that changes here is a use of albhabets with the
        # integers. So [0-9a-z:]+ means, the value can be any digit combination
        # with letter combination, followed by a : and this can repeat for 1 or
        # more times.
        ipv6re = re.compile('inet6 addr:[ ]*([0-9a-z:]+)')
        # Take the interface chunk from the ifconfig output and compile the
        # ipv4 and ipv6 addresses out of it.
        for interface_chunk in ifconfig_output.split('\n\n'):
            list_int_prop = interface_chunk.split('\n')
            ipv4_value = ""
            ipv6_value = ""
            for line in list_int_prop:
                match_ipv4 = ipv4re.search(line)
                if match_ipv4:
                    ipv4_value = match_ipv4.group(1)
                match_ipv6 = ipv6re.search(line)
                if match_ipv6:
                    ipv6_value = match_ipv6.group(1)
            interface_name = list_int_prop[0].split()[0]
            # We formulate a dictionay that fits into our scheme of things
            # going forward
            if interface_name:
                dict_list.append({'name': interface_name, 'ipv4': ipv4_value,
                                  'ipv6': ipv6_value})
        return dict_list

    @staticmethod
    def parse_lshw(lshwout):
        """
        This method parses the lshw output and only returns a list of those
        devices which are of type 'Ethernet device'.

        :param lshwout: Raw output after running the lshw command
        :return list_lshw:
            [{'speed': '10Gbit/s', 'interface': ' eth2'}]
        """
        if lshwout is None or lshwout == "":
            raise exp.IBMPowerKVMInvalidLSHWOutput()
        # Similar explanation for the regex is provided in the ifconfig parser
        regex = 'logical name:[ ]*' + ADAPTER_REGEX
        ethcardre = re.compile(regex)
        speedre = re.compile('speed=([a-z0-9A-Z/]+)')
        linkre = re.compile('link=([a-z/]+)')
        capacityre = re.compile('capacity:[ ]*([a-z0-9A-Z]+)')
        list_lshw = []
        # Let's devide the output into paragraphs. These are decided by
        # parsing the output based of the network pattern as below.
        for network_chunk in lshwout.split('*-network'):
            # Not all the devices which are part of lshw are physical ethernet
            # cards, hence we filter them out on the basis of the tag below.
            if network_chunk.strip(" ") == "":
                continue
            if "Ethernet" in network_chunk and "pci" in network_chunk:
                speed_mb = None
                capacity_mb = None
                name = ""
                state = ""
                for line in network_chunk.split("\n"):
                    match_eth = ethcardre.search(line)
                    if match_eth:
                        name = match_eth.group(1)

                    # link is the state of the adapter
                    match_link = linkre.search(line)
                    if match_link:
                        state_data = match_link.group(1)
                        if state_data == "yes":
                            state = "Available"
                        else:
                            state = "Unavailable"

                    # Speed is the current negotiated line rate.  Only
                    # available when plugged in.
                    match_speed = speedre.search(line)
                    if match_speed:
                        speed = match_speed.group(1)
                        if "G" in speed:
                            speed_mb = int(speed.split('G')[0]) * 1000
                        elif "M" in speed:
                            speed_mb = speed.split('M')[0]
                        else:
                            speed_mb = None

                    # Capacity is the speed that this port is capable of.
                    # Should only be used when the link is not up (ex. No
                    # Speed).
                    match_capacity = capacityre.search(line)
                    if match_capacity:
                        capacity = match_capacity.group(1)
                        if "G" in capacity:
                            capacity_mb = int(capacity.split('G')[0]) * 1000
                        elif "M" in capacity:
                            capacity_mb = capacity.split('M')[0]
                        else:
                            capacity_mb = None

                # It's not good enough to report a ethernet device without the
                # speed or name. A bridge can also be called "Ethernet
                # Interface which has a eth device added to that. But that
                # is not necessarily what we would want to report here.
                if name:
                    # The overall speed is determined by a mix of the capacity
                    # and speed.  Speed should be preferred.
                    overall_speed = speed_mb
                    if not speed_mb:
                        overall_speed = capacity_mb
                    if not overall_speed:
                        overall_speed = 0

                    list_lshw.append({'name': name.strip(),
                                      'speed': overall_speed,
                                      'state': state})
        return list_lshw

    @staticmethod
    def parse_brctl(brctlout):
        """
        This method parses the brctl output.

        :return bridgeDict:
        {
            'bridges': [
                {
                    'interfaces': [
                        {
                            'name': 'virbr0-nic'
                        },
                        {
                            'name': 'test-nic'
                        },
                        {
                            'name': 'test-nic2'
                        }
                    ],
                    'bridge_name': 'virbr0'
                },
                {
                    'interfaces': [
                        {
                            'name': 'virbr1-nic'
                        }
                    ],
                    'bridge_name': 'virbr1'
                },
                {
                    'interfaces': [],
                    'bridge_name': 'virbr2'
                }
            ]
        }
        """
        if brctlout is None or brctlout == "":
            raise exp.IBMPowerKVMInvalidBRCTLOutput()
        ROOT_NAME = "bridges"
        BRIDGE_NAME = "bridge_name"
        INTERFACES = "interfaces"
        INTERFACE_NAME = "name"

        outputList = brctlout.splitlines()
        bridgeDict = {"bridges": []}
        currBridgeDict = None

        # skip first line since it is just column headings
        for i in range(1, len(outputList)):
            LOG.debug("current brctl entry = %s" % outputList[i])
            splitStr = outputList[i].split()

            # determine if this is an entry for a bridge
            # or an entry containing only an interface for
            # the previously listed bridge entry
            if len(splitStr) > 1:
                if currBridgeDict is not None:
                    bridgeDict[ROOT_NAME].append(currBridgeDict)

                # reinitialize current bridge dictionary
                currBridgeDict = dict()

                # set bridge name, need to preserve until next
                # iteration since next line of output could be
                # another interface for this bridge
                currBridgeDict[BRIDGE_NAME] = splitStr[0]
                if len(splitStr) == 4:
                    currBridgeDict[INTERFACES] = \
                        [{INTERFACE_NAME: splitStr[3]}]
                else:
                    currBridgeDict[INTERFACES] = []

            # need to process interface for bridge found
            # in a previous line of output
            else:
                if(currBridgeDict is not None and
                   currBridgeDict.get(BRIDGE_NAME) is not None):
                    currBridgeDict[INTERFACES].\
                        append({INTERFACE_NAME: splitStr[0]})

        # add final bridge to dictionary
        bridgeDict[ROOT_NAME].append(currBridgeDict)

        LOG.debug("brctl data dictionary = %s" % bridgeDict)

        return bridgeDict

    @staticmethod
    def parse_ifcfg(device_name, ifcfg_contents):
        """
        This method parses the ifcfg file contents.  A dictionary of key/value
        pairs from the ifcfg file is returned.

        :param device_name: The ifcfg device, for example "eth0".
        :param ifcfg_contents: The contents of the ifcfg file.
        :return ifcfg_dict: The ifcfg file parsed into a dictionary.
        """
        # ConfigParser (mostly) handles the properties file format, which is
        # (mostly) what ifcfg files are, so ConfigParser is used here.
        # ConfigParser requires a file, so StringIO is used to convert a string
        # into a file.

        # This code creates a config StringIO that looks something like:
        # [eth0]
        # TYPE=Ethernet
        # DEVICE='eth0'
        # BOOTPROTO='dhcp'
        attribute_list = []
        try:
            config = StringIO.StringIO()
            config.write('[%s]\n' % device_name)
            config.write(ifcfg_contents)
            config.seek(0, SEEK_SET)

            # Use ConfigParser to parse our StringIO-posing-as-a-file
            cp = ConfigParser.ConfigParser()
            cp.optionxform = str
            cp.readfp(config)
            attribute_list = cp.items(device_name)
        except Exception:
            # Many files in network-scripts do not have ifcfg format, this is
            # fine, continue on.
            pass

        # Convert parsed data to a dictionary
        ifcfg_dict = IfcfgDict(attribute_list)

        # Remove leading/trailing single quotes from the dictionary entries
        ifcfg_dict = IfcfgDict(map(lambda (k, v): (k, v.strip("'")),
                                   ifcfg_dict.iteritems()))

        return ifcfg_dict

    @staticmethod
    def parse_ofctl_dump_flows(ofctl_output):
        '''
        This method parses the output passed in and generates a list of
        dictionaries showing the flows on the bridge.

        :param ofctl_output: Raw output from "ovs-ofctl dump-flows <bridge>"
        :returns: List of dictionaries representing the flows on the bridge.
                  Like so:
            [ {"in_port": "108", "mod_vlan_vid": "3", "dl_vlan": "1"},
              {"in_port": "109", "mod_vlan_vid": "4", "dl_vlan": "2"} ]
        '''
        ret_list = list()

        if ofctl_output:
            # Loop through each line
            for line in ofctl_output.splitlines():
                if not line.strip() or "NSXT_FLOW reply" in line:
                    # Skip header line
                    continue
                # Parse the line
                this_dict = dict()
                for item in line.split(','):
                    if 'in_port' in item:
                        this_dict['in_port'] = item.split('=')[-1]
                    elif 'dl_vlan' in item:
                        item = item.split(' ')
                        this_dict['dl_vlan'] = item[0].split('=')[-1]
                        if 'mod_vlan_vid' in item[1]:
                            this_dict['mod_vlan_vid'] = item[1].split(':')[-1]
                    elif 'mod_vlan_vid' in item:
                        # This is for the 'flat' network scenario
                        item = item.split(' ')
                        this_dict['dl_vlan'] = -1
                        if 'mod_vlan_vid' in item[1]:
                            this_dict['mod_vlan_vid'] = item[1].split(':')[-1]

                # If we parsed info that's useful to us, append it to the list.
                if this_dict:
                    ret_list.append(this_dict)

        # Return what we parsed
        return ret_list

    @staticmethod
    def parse_ofctl_show_order_by_port_id(ofctl_output):
        '''
        This method parses the output passed in and generates a dictionary
        showing the flow ports on the bridge.

        :param ofctl_output: Raw output from "ovs-ofctl show <bridge>"
        :returns: Dictionary representing the flow ports on the bridge.
                  Like so (port id acts as key):
            { '108': 'default', '109': 'second', ... }
        '''
        ret_dict = dict()

        if ofctl_output:
            # Loop through each line
            lines = ofctl_output.split('\n')
            for line in lines:
                if not line.strip() or "addr:" not in line:
                    # Skip everything but the port(vswitch) lines
                    continue
                # Parse the line.  Note, strip off the "int-" to get the
                # real vswitch name
                parts = line.split('(')
                parts[0] = parts[0].strip()
                parts[1] = parts[1].split(')')[0]
                if 'int-' in parts[1]:
                    parts[1] = parts[1].split('-')[1]
                ret_dict[parts[0]] = parts[1]

        # Return what we parsed
        return ret_dict

    @staticmethod
    def parse_ofctl_show_order_by_port_name(ofctl_output):
        '''
        This method parses the output passed in and generates a dictionary
        showing the flow ports on the bridge.

        :param ofctl_output: Raw output from "ovs-ofctl show <bridge>"
        :returns: Dictionary representing the flow ports on the bridge.
                  Like so (port name acts as key):
            { 'default': '108', 'second': '109', ... }
        '''
        ret_dict = dict()

        if ofctl_output:
            # Loop through each line
            lines = ofctl_output.split('\n')
            for line in lines:
                if not line.strip() or "addr:" not in line:
                    # Skip everything but the port(vswitch) lines
                    continue
                # Parse the line.  Note, strip off the "int-" to get the
                # real vswitch name
                parts = line.split('(')
                parts[0] = parts[0].strip()
                parts[1] = parts[1].split(')')[0]
                if 'int-' in parts[1]:
                    parts[1] = parts[1].split('-')[1]
                ret_dict[parts[1]] = parts[0]

        # Return what we parsed
        return ret_dict


class DOMObjectConverter():

    """
    The reason for this abstraction is to segregate functionality of what the
    parser does vs what is expected vi the DOM objects. If at any certain point
    we decide to not use the DOM objects, only this layer will get affected
    """

    def get_dom(self, host_name):
        """
        This method is used to get the DOM object representation of th devices
        on the system.
        :param: host_name: The name of the host
        :return: host_ovs_config: The DOM object with the host-ovs config info.
        """
        # Get the physical ports
        # Get the bonds and their children
        # Get the bridges and their children
        # Ignore others
        LOG.debug("Entering get_dom()...")
        lshw = CommandExecutor.get_lshw_results()
        if lshw is not None and lshw == "":
            raise exp.IBMPowerKVMInvalidLSHWOutput()
        ip_info = CommandExecutor.send_ip_address_command()
        if ip_info is not None and ip_info == "":
            raise exp.IBMPowerKVMInvalidIFCONFIGOutput()
        bonds = CommandExecutor.send_bond_info_command()
        ovs_ports = CommandExecutor.send_vsctl_list_port_command()
        bridges = CommandExecutor.send_brctl_command()
        if bridges is not None and bridges == "":
            raise exp.IBMPowerKVMInvalidBRCTLOutput()
        # Lists for the respective elements
        physical_devices = []
        bonded_devices = {}
        bridge_devices = {}

        # A converged list of all DOMs
        all_doms = []

        # First we check for physical ports
        for phy_int in lshw:
            phy_dev_name = phy_int['name']
            state = phy_int['state']
            speed = phy_int['speed']
            eth_obj = dom_kvm.PhysicalPort(phy_dev_name, state)
            eth_obj.speed_mb = speed

            # In some scenarios, speed may not be reported.  Set to 0 in that
            # case
            if eth_obj.speed_mb is None or eth_obj.speed_mb == '':
                eth_obj.speed_mb = 0
            LOG.debug("Formed the dom for the physical device: %s",
                      eth_obj.name)
            physical_devices.append(eth_obj)
            all_doms.append(eth_obj)

        # Next we check for the bonded adapters
        if bonds and bonds != {}:
            for bond_name, slave_list in bonds.items():
                bond_obj = dom_kvm.EthernetBond(bond_name)
                bonded_devices[bond_name] = {'obj': bond_obj,
                                             'slaves': slave_list}
                LOG.debug("Formed the dom for the bond: %s", bond_obj.name)
                all_doms.append(bond_obj)

        # Last, we check for bridges
        for bridge in bridges.get('bridges'):
            bridge_name = bridge['bridge_name']
            bridge_obj = dom_kvm.LinuxBridge(bridge_name)
            bridge_devices[bridge_name] = {'obj': bridge_obj,
                                           'slaves': bridge['interfaces']}
            LOG.debug("Formed the dom for the linuxbridge: %s",
                      bridge_obj.name)
            all_doms.append(bridge_obj)

        # Build a list of unused DOMs that is based off of all the DOMs
        unused_doms = copy.copy(all_doms)

        # Now that we've looped, we start to add the children to the bonds
        # and then the bridges.
        for bond in bonded_devices.values():
            slave_devs = [o['name'] for o in bond['slaves']]
            bond['obj'].port_list = self.get_dom_objects_by_name(all_doms,
                                                                 slave_devs)

        for bridge in bridge_devices.values():
            slave_devs = [o['name'] for o in bridge['slaves']]
            bridge['obj'].port_list = self.get_dom_objects_by_name(all_doms,
                                                                   slave_devs)

        # Now that we've gone through the trouble of adding the children
        # to the bridges/bonds (note because a bond could have a bridge as
        # a child and vice versa), we can remove the children of those elements
        # from the return list.
        for bond in bonded_devices.values():
            for b_port in bond['obj'].port_list:
                if b_port in unused_doms:
                    unused_doms.remove(b_port)
        for bridge in bridge_devices.values():
            for b_port in bridge['obj'].port_list:
                if b_port in unused_doms:
                    unused_doms.remove(b_port)

        # Next up are the OVS Ports.  We need to find those and add them
        # to the list of elements.
        vsctl = CommandExecutor.send_vsctl_command()
        if vsctl is None or vsctl == "":
            raise exp.IBMPowerKVMInvalidVSCTLOutput()
        ovs_doms = []
        for ovs_name in vsctl.keys():
            # Ignore the integration bridge...not use managed
            if ovs_name == CONF.integration_bridge:
                # Remove all the children interfaces of br-int
                ovs = vsctl.get(ovs_name)
                for ovs_port_name in ovs.keys():
                    backing_ports = ovs[ovs_port_name]['interface']
                    for linux_port in backing_ports:
                        # All the qvo devices have a linux bridge device which
                        # are detected by the lshw/ip command and hence we
                        # would have to translate that and remove.
                        if "qvo" in linux_port:
                            linux_port = linux_port.replace("qvo", "qbr")
                            linux_dom = self.get_dom_object_by_name(all_doms,
                                                                    linux_port)
                            if linux_dom in unused_doms:
                                LOG.debug("Removing the device: %s",
                                          linux_dom.name)
                                unused_doms.remove(linux_dom)
                        if "tap" in linux_port:
                            tap_dom = self.get_dom_object_by_name(all_doms,
                                                                  linux_port)
                            if tap_dom in unused_doms:
                                LOG.debug("Removing the device: %s",
                                          tap_dom.name)
                                unused_doms.remove(tap_dom)
                continue

            # Build our DOM objects
            ovs = vsctl.get(ovs_name)
            ovs_dom = dom_kvm.OpenVSwitch(ovs_name)
            ovs_doms.append(ovs_dom)
            all_doms.append(ovs_dom)

            # Loop through the ports and link them in
            for ovs_port_name in ovs.keys():
                # If the name matches, just ignore
                if ovs_port_name == ovs_name:
                    continue

                # If the name is phy-<ovsname>, then ignore
                if 'phy-' + ovs_name == ovs_port_name:
                    continue

                # Find the DOM objects
                backing_ports = ovs[ovs_port_name]['interface']
                b_dom_ports = self.get_dom_objects_by_name(all_doms,
                                                           backing_ports)

                # Build the OVS port
                ovs_port = dom_kvm.OVSPort(ovs_port_name)
                ovs_port.bond = dom_kvm.BondConfiguration(None, None, None)

                ovs_port.port_list = b_dom_ports

                # Add OVS bonding info if applicable
                if len(b_dom_ports) > 1:
                    bond_mode = ovs_ports[ovs_port_name]['bond_mode']
                    bond_fail_detc = None
                    other_config = ovs_ports[ovs_port_name].get('other_config',
                                                                '')
                    if 'bond-detect-mode=' in other_config:
                        bond_fail_detc = other_config.split(
                            'bond-detect-mode=')[1].split()[0]
                    bond_lacp = ovs_ports[ovs_port_name]['lacp']
                    if not bond_lacp:
                        bond_lacp = None  # Prevents return of empty string
                    ovs_port.bond.mode = bond_mode
                    ovs_port.bond.detect_mode = bond_fail_detc
                    ovs_port.bond.lacp = bond_lacp

                # Remove the DOMs from the unused doms list, as they're now
                # children and no longer independent.
                for dom_port in b_dom_ports:
                    if dom_port in unused_doms:
                        LOG.debug("Removing the device: %s", dom_port.name)
                        unused_doms.remove(dom_port)

                # Add this OVS Port DOM to the OVS
                ovs_dom.ovs_port_list.append(ovs_port)

        # Now we need to splice in the IP information on each of the DOM
        # objects.  We utilize the all_doms list we had originally.
        for ip in ip_info:
            ip_addr = ip.get('ipv4')
            ipv6_addr_list = ip.get('ipv6')
            dev_name = ip['name']
            dom = self.get_dom_object_by_name(all_doms, dev_name)
            if dom is None or ip_addr is None:
                continue
            dom.ip_addresses.append(dom_kvm.IPAddress(ip_addr))
            if ipv6_addr_list:
                for ipv6 in ipv6_addr_list:
                    if ipv6 != "":
                        dom.ip_addresses.append(dom_kvm.IPAddress(ipv6))
            LOG.debug("Added IP address to the device: %s", dom.name)

        # We now have a list of all the DOMs.  The OVS doms and at this point,
        # all doms are just use unused components
        host_ovs_config = dom_kvm.HostOVSNetworkConfig(
            host_name, version=CommandExecutor.get_powerkvm_version())
        host_ovs_config.ovs_list = ovs_doms
        host_ovs_config.unused_component_list = unused_doms

        # We wait until the end of this method to parse out Open vSwitches
        # that are too long in length.  If we did it earlier in the process
        # there is a chance that we would have accidentally shown it as an
        # adapter.
        to_del = []
        for ovs in host_ovs_config.ovs_list:
            if len(ovs.name) > MAX_VSWITCH_CHARS:
                LOG.debug("Removing OVS %s due to too long of a name",
                          ovs.name)
                to_del.append(ovs)
        for del_item in to_del:
            host_ovs_config.ovs_list.remove(del_item)

        LOG.debug("Returning data to the API from get_dom: %s",
                  host_ovs_config.to_dict())
        return host_ovs_config

    def get_dom_objects_by_name(self, all_doms, name_list):
        """
        This method is used to get a list of dom objects by referring them
        through a list of names
        :param all_doms: All the dom objects in the system.
        :param name_list: List of names for which we need the dom objects.
        :return: dom_obj_list: A list of dom objects to be returned.
        """
        dom_obj_list = []
        for dom in all_doms:
            for name in name_list:
                if dom.name == name:
                    dom_obj_list.append(dom)
                    continue
        return dom_obj_list

    def get_dom_object_by_name(self, all_doms, name):
        """
        This is used to retrieve a DOM object for a device referred with a name
        :param all_doms: Dom representation of all the devices.
        :param name: The name of the device for which we need the DOM object.
        :return: A dom object or the given device name
        """
        for dom in all_doms:
            if dom.name == name:
                return dom
        return None

    def find_management_adapter(self):
        """
        This is used to find out which adapter is used for communicating with
        the PowerVC management node.if no management adapter found then host
        cant be reachable from PowerVC node and vice versa
        :returns: Name of adapter used for communication if found, else None
        """
        # Query the Management server ip, if its None then return None
        if CONF.management_ip is None:
            LOG.debug("The management ip for the server is not set")
            return None

        ATTEMPT_COUNT = CONF.max_mgmt_adpt_ping_tries

        min_loops = 4
        max_time_in_seconds = 120
        start_time = time.time()

        for i in range(0, ATTEMPT_COUNT):
            # Query the DOM to find all adapters/vswitches on the system
            #
            # Must be done IN the for loop, because outside of the for loop
            # will just get the adapters available up front.  There is
            # an oddity in neutron that will reconfigure the 'default' vSwitch
            # so the default vswitch may be in, then out, then in...so we
            # have to re-query
            host_dom = DOMObjectConverter().get_dom(CONF.host)

            # Combine all the OVS together with all the unused components
            dom_top_level_components = host_dom.ovs_list
            dom_top_level_components.extend(host_dom.unused_component_list)

            # For each element from above list
            #  If the state is up
            #    Check to see if the adapter has an IP Address.
            #    If not, continue to next element
            for component in dom_top_level_components:
                if component.get_state() != 'Unavailable':
                    if len(component.ip_addresses) > 0:
                        # Try to ping that qpid host with this adapter.
                        # using 'ping -I <adapter> -c 1 -W 1 <mgmt address>'
                        # If true returned, then save the adapter
                        # If it fails, continue to next adapter.
                        if CommandExecutor.run_ping(CONF.management_ip,
                                                    component.name):
                            return component.name

            if i > min_loops and \
                    time.time() - start_time > max_time_in_seconds:
                break

            # take a brief pause between each loop
            time.sleep(1)

        # If we have an adapter that pings, return it. Otherwise return none.
        return None


def _write_to_file_as_root(output_file, lines):
    """
    Writes a file to the file system as root.  This method is NOT efficient.
    It was created to work around rootwrap issues.

    :param output_file: The file to write to
    :param lines: A list of strings.  Each is a new line within the file.
    """
    try:
        # Write the new contents to a temp file
        f = tempfile.NamedTemporaryFile(delete=False)
        for line in lines:
            f.write(line)
        f.flush()
        f.close()

        # Change the permissions to root / root
        utils.execute('chown', 'root', f.name, run_as_root=True)
        utils.execute('chgrp', 'root', f.name, run_as_root=True)

        # Move the file to the correct location
        utils.execute('mv', '-f', f.name, output_file, run_as_root=True)
    except Exception as e:
        LOG.error(e)
        raise


def _check_file_exists(full_path):
    """
    Will check to see if a file exists.

    :param full_path: The path to the file.
    :returns: True if it does, false otherwise
    """
    try:
        # If the file does not exist, a non-zero exit code will be returned
        utils.execute('ls', full_path, run_as_root=True, check_exit_code=[0])
        return True
    except Exception:
        return False
    return False


def _list_directory(full_path, run_as_root=False):
    """
    Lists out the full file contents of a directory.

    :param full_path: The path to the directory
    :returns: List of paths.
    """
    if not _check_file_exists(full_path):
        return []
    files = []
    try:
        stdout, stderr = utils.execute('ls', '-m',
                                       full_path,
                                       run_as_root=run_as_root)
        if stderr and stderr != '':
            raise exp.IBMPowerKVMCommandExecError(cmd='ls -m', exp=stderr)
        for line in stdout.splitlines():
            for elem in line.split(','):
                if len(elem.strip()) != 0:
                    files.append(elem.strip())
    except Exception as e:
        LOG.exception(e)
        raise
    return files


def remove_file_as_root(file_name_and_path):
    """
    Removes a file from the file system as root.  This method is NOT efficient.
    It was created to work around rootwrap issues.

    :param fname: The file to remove.
    """
    try:
        # If the file does not exist, a non-zero exit code will be returned
        utils.execute('rm', '-f',
                      file_name_and_path,
                      run_as_root=True,
                      check_exit_code=[0])
    except Exception as e:
        LOG.exception(e)
        raise


class IfcfgDict(dict):

    def __init__(self, value=None):
        if isinstance(value, dict):
            for k, v in value.iteritems():
                super(IfcfgDict, self).__setitem__(k, IfcfgString(v))
        elif isinstance(value, list):
            for item in value:
                if isinstance(item, tuple):
                    super(IfcfgDict, self).__setitem__(item[0],
                                                       IfcfgString(item[1]))
        else:
            super(IfcfgDict, self).__init__()

    def __setitem__(self, key, value):
        super(IfcfgDict, self).__setitem__(key, IfcfgString(value))


class IfcfgString(str):

    def __eq__(self, other):
        return re.sub(r'^"|"$', '', self) == \
            re.sub(r'^"|"$', '', other)

    def __ne__(self, other):
        return re.sub(r'^"|"$', '', self) != \
            re.sub(r'^"|"$', '', other)

    def upper(self):
        return IfcfgString(super(IfcfgString, self).upper())

    def lower(self):
        return IfcfgString(super(IfcfgString, self).lower())
