#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
from powervc_nova.network.powerkvm.agent.common import micro_op
from powervc_nova.network.powerkvm.agent.common import exception
from powervc_nova.network.powerkvm.agent.common import warning
from powervc_nova.network.powerkvm.agent import commandlet
from oslo_log import log as logging
from nova import utils
from nova import network
from powervc_nova import logcall
from oslo_config import cfg

CONF = cfg.CONF
CONF.import_opt('integration_bridge', 'powervc_nova.network.powerkvm.agent')

LOG = logging.getLogger(__name__)


class DeletevSwitch(micro_op.MicroOperation):
    """
    This micro-op deletes a Open vSwitch
    """

    def __init__(self, context, ovs_name):
        """
        Constructor

        :param ovs_name: The name of the Open vSwitch
        """
        self.context = context
        self.ovs_name = ovs_name
        self.current_dom = None
        self.source_ovs_restore = None

    @logcall
    def validate(self, curr_dom):
        """
        Performs a set of validations and throws errors accordingly
        :param current_dom: This is the current DOM representation of the
                            system before the micro op is performed.
        :returns updated_dom: An updated version of the current_dom that
                              has been updated 'as if' the operation had been
                              executed.
        :returns warning_list: A list of warnings (but not errors) that were
                               generated based off the state of the DOM and
                               the operation that had been run.
        """
        self.current_dom = curr_dom
        warning_list = []

        # warning if attempting to delete default ovs
        default_vswitch = (
            'default' if curr_dom.version.startswith('2') else 'default0')
        if self.ovs_name == default_vswitch:
            warning_list.append(warning.OVSDeletingDefaultOVSWarning())

        # Determine whether the selected vswitch has any VMs with adapters
        # associated with it.  If so, generate a WARNING
        else:
            net_api = network.API()
            net_list = net_api.get_all(self.context)
            LOG.debug("net list = %s" % str(net_list))

            net_found = False

            for net in net_list:
                if net['physical_network'] == self.ovs_name:
                    LOG.debug("vSwitch %s assigned to a "
                              "network" % self.ovs_name)
                    net_found = True
                    break

            if net_found:
                warning_list.append(
                    warning.OVSDeletingOVSWithAttachedNetworkWarning())
            else:
                warning_list.append(warning.OVSDeletingOVSWarning())

        # If the vSwitch does not exist, it should throw an ERROR.
        ovs_obj = self.current_dom.find_ovs_obj(self.ovs_name)
        if ovs_obj is None:
            raise exception.IBMPowerKVMOVSDoesNotExistError(
                switch_name=self.ovs_name)
        # Move components from under OVS to unused component list
        if len(ovs_obj.ovs_port_list) > 0 and ovs_obj.ovs_port_list != []:
            self.current_dom.unused_component_list.extend(
                ovs_obj.ovs_port_list)
        # Remove itself to the DOM that is passed in, do a return
        self.current_dom.ovs_list.remove(ovs_obj)
        return self.current_dom, warning_list

    @logcall
    def execute(self):
        """
        Overrides parent method.

        Execute command line call to delete a OVS and backup/delete ifcfg file
        """
        stdout_del_br = None
        stdout_del_port = None

        try:
            # run 'ovs-vsctl del-br <bridge>'
            stdout_del_br, stderr = \
                utils.execute('ovs-vsctl', 'del-br',
                              self.ovs_name, run_as_root=True)
            if stderr and stderr != '':
                raise exception.IBMPowerKVMCommandExecError(
                    cmd='ovs-vsctl del-br', exp=stderr)

        except Exception as e:
            LOG.error(e)
            raise
        LOG.debug('ovs-vsctl del-br output = %s' % stdout_del_br)

        try:
            # Remove int- port for deleted vswitch from integration bridge.
            # Don't fail the overall delete operation if this fails since
            # the impact of leaving the port on the system should not
            # have any affect on the user and shouldn't be visible to the
            # user unless they are running ovs CLIs.
            stdout_del_port, stderr = \
                utils.execute('ovs-vsctl', 'del-port',
                              CONF.integration_bridge,
                              'int-%s' % self.ovs_name,
                              run_as_root=True)
            if stderr and stderr != '':
                LOG.debug("Integration bridge port for %s failed to "
                          "delete." % self.ovs_name)
        except Exception as e:
            LOG.warning(e)

        LOG.debug('ovs-vsctl del-port output = %s' % stdout_del_port)

        # Save contents and location of current ifcfg file (if it exists!)
        fnamepath, self.source_ovs_restore = \
            commandlet.CommandExecutor.get_ifcfg(self.ovs_name)
        # Delete ifcfg file
        if fnamepath and len(fnamepath):
            LOG.debug('Removing ifcfg: %s', fnamepath)
            commandlet.remove_file_as_root(fnamepath)

    @logcall
    def undo(self):
        """
        Overrides parent method.

        Execute command line call to add an OVS and restore its ifcfg file.
        """
        stdout = None

        # run 'ovs-vsctl add-br <bridge>'
        try:
            stdout, stderr = utils.execute('ovs-vsctl', 'add-br',
                                           self.ovs_name,
                                           run_as_root=True)
            if stderr and stderr != '':
                raise exception.IBMPowerKVMCommandExecError(cmd='ovs-vsctl',
                                                            exp=stderr)
        except Exception as e:
            LOG.error(e)
            raise
        LOG.debug('ovs-vsctl add-br output = %s' % stdout)

        try:
            # Readd int- port for deleted vswitch from integration bridge.
            # Don't fail if stderr is set since if the port was not deleted
            # during the execute phase (i.e. failure occurred with vswitch
            # delete so port delete didn't happen) stderr will be set
            # since we tried to create a port that is already on the
            # integration bridge.  This is ok and so we should ignore.
            # However, we should raise an exception here if one is
            # received.
            stdout, stderr = \
                utils.execute('ovs-vsctl', 'add-port',
                              CONF.integration_bridge,
                              'int-%s' % self.ovs_name,
                              run_as_root=True)
            if stderr and stderr != '':
                LOG.debug("Integration bridge port for %(name)s failed to "
                          "add: %(err_msg)s" % {'name': self.ovs_name,
                                                'err_msg': stderr})
        except Exception as e:
            LOG.error(e)
            raise
        LOG.debug('ovs-vsctl add-port output = %s' % stdout)

        # Should restore the ifcfg file back to its original contents.
        commandlet.CommandExecutor.\
            send_ifcfg(self.ovs_name, self.source_ovs_restore)
