#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import copy
import uuid

from nova import network
from nova.network import neutronv2
from nova.objects import base
from nova.objects import fields
from nova.objects import network as network_obj
from oslo_log import log as logging

from oslo_config import cfg

from powervc_nova import _
from powervc_nova.db import api as db_api
import powervc_nova.objects.dom as dom
import powervc_nova.objects.compute.dom as compute_dom
from powervc_nova.virt.ibmpowervm.vif.common import exception as excp

LOG = logging.getLogger(__name__)

cfg.CONF.register_opts([
    cfg.IntOpt('used_port_threshold', default=100,
               help='the setting drives if port cleanup should be '
               'invoked')
])

USED_PORT_COUNT = 0


######################################################
#           Managed Host DOM Implementation          #
######################################################
@base.NovaObjectRegistry.register
class ManagedHostFactory(compute_dom.ManagedHostFactory):
    """Production DOM Factory implementation for Hosts"""

    ###############################################
    #   Implementing Resource Factory Definition  #
    ###############################################
    def _get_resource_classes(self):
        """Returns a List of DOM Classes or a Dictionary Mapping DOM Classes"""
        return {compute_dom.ManagedHost: ManagedHost}

    def _get_vios_factory(self):
        """Provides the Factory to use for the VIO Server Implementation"""
        return VioServerFactory.get_factory()

    ###############################################
    #       Defining Host Factory Interface       #
    ###############################################
    def find_all_hosts(self, context, filters=None,
                       transaction=None, load_all=False):
        """Queries all Host that are registered on the Management System"""
        hosts = super(ManagedHostFactory, self).\
            find_all_hosts(context, filters, transaction)
        # If they asked us to load all the info, do it now
        if load_all and hosts:
            self.__load_host_info(context, hosts, transaction)
        return hosts

    def find_host_by_name(self, context, host_name,
                          transaction=None, load_all=False):
        """Queries at most one Host matching the Host Name specified"""
        host = super(ManagedHostFactory, self).\
            find_host_by_name(context, host_name, transaction)
        # If they asked us to load all the info, do it now
        if load_all and host is not None:
            self.__load_host_info(context, [host], transaction)
        return host

    ###############################################
    #      Internal Host DOM Helper Methods       #
    ###############################################
    def __load_host_info(self, context, hosts, transaction):
        """Helper method to load lazy-loaded host attributes"""
        vios_host_map = dict()
        host_name = hosts[0].host_name if len(hosts) == 1 else None
        filters = dict(host_name=host_name) if host_name else None
        # Load all the VIOS/SEA/VEA instances to do the population
        viosfact = VioServerFactory.get_factory()
        vioses = viosfact.find_all_vioses(
            context, filters, transaction, load_all=True)
        # Loop through the VIOS's populating the necessary maps
        for vios in vioses:
            vios_host_map.setdefault(vios.host_name, list())
            vios_host_map[vios.host_name].append(vios)
        # Now we can loop through the hosts fully populating the info
        for host in hosts:
            host._vioservers = vios_host_map.get(host.host_name, [])
            for vios in host._vioservers:
                vios._host = host


@base.NovaObjectRegistry.register
class ManagedHost(compute_dom.ManagedHost):
    """Production DOM Resource implementation for Managed Hosts"""

    obj_extra_fields = compute_dom.ManagedHost.\
        obj_extra_fields + ['seas', 'veas', 'available_vid_pool']

    def __init__(self, context=None,
                 factory_class=None, transaction=None, **kwargs):
        "Constructs an instance of the given DOM Resource Class"
        super(ManagedHost, self).__init__(
            context, factory_class, transaction, **kwargs)

    ###############################################
    #     Implementing DOM Resource Definition    #
    ###############################################
    def _initialize_resource_attributes(self):
        """Sets any default values that are defined for any attributes"""
        super(ManagedHost, self)._initialize_resource_attributes()
        self._available_vid_pool = dict()
        self._seas, self._veas, self._primary_sea_list = (None, None, None)

    def _reset_cached_attributes(self):
        """Wipes out any Cached Resource Attributes that are no longer valid"""
        super(ManagedHost, self)._reset_cached_attributes()
        self._seas, self._veas, self._primary_sea_list = (None, None, None)

    ###############################################
    #       Defining Managed Host Interface       #
    ###############################################
    @property
    def available_vid_pool(self):
        """Getter Method for the Available VID Pool for the Host"""
        return self._available_vid_pool

    @property
    def seas(self):
        """Getter Method for the list of SEA's for this Host"""
        # We always rebuild...in case in memory has changed
        temp = list()
        # Load the VIO Servers for the Host and get the SEA's there
        for vios in self.vioservers:
            temp.extend(vios.seas)
        self._seas = temp
        return self._seas

    @property
    def veas(self):
        """Getter Method for the list of VEA's for this Host"""
        # We always rebuild...in case in memory has changed
        temp = list()
        # Load the VIO Servers for the Host and get the VEA's there
        for vios in self.vioservers:
            self.temp.extend(vios.veas)
        self._veas = temp
        return self._veas

    def maintain_available_vid_pool(self):
        """Looks up all the active VLAN ID's on the Host and removes them"""
        """from the available VID Pool to avoid conflicts during Plug time"""
        # List storing the available VLAN ID's used for the PVID VEA's Port VID
        self._available_vid_pool = dict()

        # Find all of the VLAN ID's in use for each vSwitch including
        # port VLAN ID and additional 8021Q VLAN ID's from the PowerVM.
        # Those VIDS will be removed from the available VID pool.
        for vios in self.vioservers:
            for vea in vios.veas:
                vswitch_name = vea.vswitch_name
                # If we haven't encountered this vswitch yet, init it's list
                if vswitch_name not in self._available_vid_pool.keys():
                    self._available_vid_pool[vswitch_name] = range(1, 4095)

                # Now remove used vids from this vswitch's pool
                if self._available_vid_pool[vswitch_name].count(vea.pvid) > 0:
                    self._available_vid_pool[vswitch_name].remove(vea.pvid)
                for vid in vea.addl_vlan_ids:
                    if self._available_vid_pool[vswitch_name].count(vid) > 0:
                        self._available_vid_pool[vswitch_name].remove(vid)

    def find_all_primary_seas(self, rebuild=False):
        """Retrieves a list of all of the Primary SEA's on the Host"""
        primary_sea_list = list()
        for sea_chain in self.find_sea_chains(rebuild):
            primary_sea_list.append(sea_chain[0])
        return primary_sea_list

    def find_sea_chain_for_sea(self, sea, rebuild=False):
        """Finds the SEA chain (ordered list) for a given SEA"""
        for sea_chain in self.find_sea_chains(rebuild):
            if sea_chain.count(sea) > 0:
                return sea_chain
        # If no chain was found, build itself as its own chain
        return [sea]

    def find_sea_chain_for_sea_info(self, sea_name, lpar_id, rebuild=False):
        """Finds the SEA chain (ordered list for the given SEA Name/LPAR """
        for sea_chain in self.find_sea_chains(rebuild):
            for sea in sea_chain:
                if sea_name == sea.name and lpar_id == sea.vioserver.lpar_id:
                    return sea_chain
        return list()

    def find_sea_chain_by_pvea_info(self, sea, rebuild=False):
        """Finds an SEA chain (ordered list) by inspecting the Primary VEA"""
        # If we weren't given an SEA then there isn't anything to do
        if sea is None:
            return list()
        # Look through the chains and find one with the same Primary VEA info
        for chain in self.find_sea_chains(rebuild):
            # If there isn't any chain, then there is nothing to match on
            if chain is None or len(chain) == 0:
                continue
            # Look at the first SEA as the candidate for a match
            cand_sea, cand_pvea = (chain[0], chain[0].primary_vea)
            # Check to see if the Primary VEA Attributes match the candidate
            if ((cand_pvea.pvid == sea.primary_vea.pvid and
                 cand_pvea.vswitch_name == sea.primary_vea.vswitch_name and
                 len(cand_sea.additional_veas) == len(sea.additional_veas))):
                return chain
        return []

    def find_sea_chains(self, rebuild=False):
        """Returns a list of lists representing the SEA's on the Host"""
        if self._primary_sea_list is None or rebuild:
            self._primary_sea_list = self._find_sea_chains(self.seas)
        return self._primary_sea_list

    def find_adapter(self, lpar_id, adapter_name):
        """Returns an adapter on a given LPAR for the Host, if found"""
        return self._find_adapter(self.vioservers, lpar_id, adapter_name)

    def is_vlan_vswitch_on_sea(self, vlan_id, vswitch_name):
        """Determines if a VLAN/vSwitch combination is on an SEA"""
        sea = self.find_sea_for_vlan_vswitch(vlan_id, vswitch_name)
        return sea is not None

    def find_sea_for_vlan_vswitch(self, vlan_id, vswitch_name):
        """Retrieves the SEA that is hosting the VLAN/vSwitch combination"""
        for sea in self.find_all_primary_seas():
            # See if it is on the SEA's Primary VEA
            if ((sea.primary_vea.is_vlan_configured(
                    vlan_id, vswitch_name, False))):
                return sea
            # Loop through all the additional VEAs and see if it exists there
            for vea in sea.additional_veas:
                if vea.is_vlan_configured(vlan_id, vswitch_name, False):
                    return sea
        return None

    def find_vio_server(self, lpar_id=0, lpar_name=None):
        """Retrieves the VIOS on this Host which matches the LPAR ID/Name"""
        return self._find_vio_server(self.vioservers, lpar_id, lpar_name)

    def get_unusable_vlanids(self, vswitch):
        """Returns all VLAN's on this Host that cannot be used.  This"""
        """includes orphan VEA's and all Control Channel PVID's."""
        existing_vlans, unusable_cands = (list(), list())
        # First gather all of the VLANs that are used - across ALL VIOSes
        for vioserver in self.vioservers:
            for sea in vioserver.seas:
                # If given a vSwitch name, make sure it matches the SEA
                if vswitch and sea.primary_vea.vswitch_name != vswitch:
                    continue
                # Add the Primary VEA PVID and the Additional VLAN ID's
                existing_vlans.append(sea.primary_vea.pvid)
                if sea.primary_vea.addl_vlan_ids:
                    existing_vlans.extend(sea.primary_vea.addl_vlan_ids)
                # Walk through the additional VEA's, adding those VLAN ID's
                for vea in sea.additional_veas:
                    existing_vlans.append(vea.pvid)
                    if vea.addl_vlan_ids:
                        existing_vlans.extend(vea.addl_vlan_ids)
        # Now walk through each VIO Server to see the unusable candidates
        for vioserver in self.vioservers:
            # First get all the orphan VLAN ID's and mark them unusable
            unusable_cands.extend(vioserver.get_vlans_orphan_vea(vswitch))
            # Now loop through and look at Control Channel VEA's
            for sea in vioserver.seas:
                # If given a vSwitch name, make sure it matches the SEA
                if vswitch and sea.primary_vea.vswitch_name != vswitch:
                    continue
                # If the Control Channel is set, mark it as unusable
                if sea.control_channel_vea is not None:
                    unusable_cands.append(sea.control_channel_vea.pvid)
        # Return the list of VLANs that are in the unusable
        # candidate list but are not in the existing VLAN list
        return [cand for cand in unusable_cands
                if unusable_cands not in existing_vlans]

    ###############################################
    #        Shared Host DOM Helper Methods       #
    ###############################################
    @staticmethod
    def _find_vio_server(vioservers, lpar_id=0, lpar_name=None):
        """Retrieves the VIOS on this Host which matches the LPAR ID/Name"""
        if not lpar_id and not lpar_name:
            return None
        # If they asked us to search for a specific LPAR ID
        if lpar_id:
            # Loop through all VIO Servers on the given Host
            for vios in vioservers:
                # If we found an LPAR ID match, may need to check the name
                if vios.lpar_id == lpar_id:
                    # If the name matches, or they didn't provide on
                    if not lpar_name or vios.lpar_name == lpar_name:
                        return vios
                    # Not going to find another ID match so return None
                    return None
            return None
        # No LPAR ID was provided, so just match on the LPAR Name
        for vios in vioservers:
            if vios.lpar_name == lpar_name:
                return vios
        return None

    @staticmethod
    def _find_adapter(vioservers, lpar_id, adapter_name):
        """Returns an adapter on a given LPAR for the Host, if found"""
        for vioserver in vioservers:
            if vioserver.lpar_id != int(lpar_id):
                continue
            # Loop through the adapters on the VIOS for a match
            for adapter in vioserver.get_all_adapters():
                if adapter.name == adapter_name:
                    return adapter
        return None

    @staticmethod
    def _find_sea_chains(seas):
        """Returns a list of lists representing the SEA's on the Host"""
        # Maintain a list of lists of all the primary SEAs. Each list will be
        # a sorted 'chain' of SEAs.  The second element within that list would
        # be the fail-over SEA.  The third would be the third fail-over (rare).
        primary_sea_list, seas_already_listed = (list(), list())
        # Loop through all of the SEAs.
        for sea in seas:
            # See if this SEA has already made it into a list
            if sea.id in seas_already_listed:
                continue
            sea_chain = [sea]
            # Append all fail-over SEAs associated with this one
            sea_chain.extend(ManagedHost._find_failover_seas(sea, seas))
            # Sort the list such that the available SEAs are first (sorted
            # by trunk priority), then the unavailable SEAs are after that.
            sea_chain = sorted(sea_chain, key=lambda sea: sea.trunk_pri)
            # Now add everything we just processed to seas_already_listed
            seas_already_listed.extend([sea.id for sea in sea_chain])
            # And add the chain to the main list of chains
            primary_sea_list.append(sea_chain)
        return primary_sea_list

    @staticmethod
    def _find_failover_seas(sea, all_seas):
        """Finds all SEA's that are considered fail-overs for the SEA given"""
        ret_list = list()
        # If we weren't given an SEA, then there is nothing to do
        if sea is None:
            return list()
        # Look through all of the SEA's to check each candidate
        for cand_sea in all_seas:
            cand_pvea = cand_sea.primary_vea
            # Be sure not to include the SEA that was passed in
            if cand_sea.id == sea.id:
                continue
            # If we have the same VID, vSwitch, and number of trunk adapters
            if ((cand_pvea.pvid == sea.primary_vea.pvid and
                 cand_pvea.vswitch_name == sea.primary_vea.vswitch_name and
                 len(cand_sea.additional_veas) == len(sea.additional_veas))):
                ret_list.append(cand_sea)
        return ret_list


######################################################
#         Managed Instance DOM Implementation        #
######################################################
@base.NovaObjectRegistry.register
class ManagedInstanceFactory(compute_dom.ManagedInstanceFactory):
    """Production DOM Resource Factory class implementation for Instances"""

    ###############################################
    #     Defining Instance Factory Interface     #
    ###############################################
    def find_all_instances_by_network(self, context, host_name, neutron_net_id,
                                      neutron_ports=None, transaction=None):
        """Queries all Instances on a Host with the Network specified"""
        host_instances = self.find_all_instances_by_host(
            context, host_name, transaction)
        # If they gave us the Ports, lets filter here rather than the DB
        # If they asked to filter on the Neutron Network ID, do so now
        if not neutron_ports:
            search_opts = dict(network_id=neutron_net_id)
            network_data = network.API().list_ports(context, **search_opts)
            neutron_ports = network_data.get('ports', [])
        port_instances = list()
        # Loop through each of the Ports and Instances looking for a match
        for instance in host_instances:
            for port in neutron_ports:
                # If we found a match, then add this to the list returned
                if port['device_id'] == instance['uuid']:
                    port_instances.append(instance)
                    break
        return port_instances


######################################################
#            VIO Server DOM Implementation           #
######################################################
@base.NovaObjectRegistry.register
class VioServerFactory(compute_dom.VioServerFactory):
    """Production DOM Factory implementation for VIOS's"""

    ###############################################
    #   Implementing Resource Factory Definition  #
    ###############################################
    def construct_vios(self, context, host_name, lpar_id, lpar_name,
                       state, rmc_state, cluster_provider_name=None,
                       auto_create=False, transaction=None):
        """Constructs (and optionally creates) the VIOS specified"""
        # First, create the vios from the parent
        vios = super(VioServerFactory, self).construct_vios(
            context, host_name, lpar_id, lpar_name, state, rmc_state,
            cluster_provider_name, auto_create, transaction)

        vios.seas = []
        vios.veas = []

        return vios

    def _get_resource_classes(self):
        """Returns a List of DOM Classes or a Dictionary Mapping DOM Classes"""
        return {compute_dom.VioServer: VioServer}

    def _get_host_factory(self):
        """Provides the Factory to use for the Managed Host Implementation"""
        return ManagedHostFactory.get_factory()

    ###############################################
    #       Defining VIOS Factory Interface       #
    ###############################################
    def find_all_vioses(self, context, filters=None,
                        transaction=None, load_all=False):
        """Queries all VIOS's matching the filters specified"""
        vioses = super(VioServerFactory, self).\
            find_all_vioses(context, filters, transaction)
        # If they asked us to load all the info, do it now
        if load_all and vioses:
            self.__load_vios_info(context, vioses, transaction)
        return vioses

    def find_all_vioses_by_host(self, context, host_name,
                                transaction=None, load_all=False):
        """Queries all VIOS's that are part of the host specified"""
        vioses = super(VioServerFactory, self).\
            find_all_vioses_by_host(context, host_name, transaction)
        # If they asked us to load all the info, do it now
        if load_all and vioses:
            self.__load_vios_info(context, vioses, transaction)
        return vioses

    def find_vios_by_id(self, context, vios_id,
                        transaction=None, load_all=False):
        """Returns VIOS matching the given ID if found, else None."""
        vios = super(VioServerFactory, self).\
            find_vios_by_id(context, vios_id, transaction)
        # If they asked us to load all the info, do it now
        if load_all and vios is not None:
            self.__load_vios_info(context, [vios], transaction)
        return vios

    ###############################################
    #      Internal Host DOM Helper Methods       #
    ###############################################
    def __load_vios_info(self, context, vioses, transaction):
        """Helper method to load lazy-loaded VIOS attributes"""
        vios_id = vioses[0].id if len(vioses) == 1 else None
        filters = dict(vios_id=vios_id) if vios_id else None
        # Load all the SEA/VEA instances to do the population
        seafact = SharedEthernetAdapterFactory.get_factory()
        veafact = VirtualEthernetAdapterFactory.get_factory()
        seas = seafact.find_all_seas(context, filters, transaction)
        veas = veafact.find_all_veas(context, filters, transaction)
        # Build the maps needed for efficient population
        sea_id_map, sea_vios_map = dict(), dict()
        vea_id_map, vea_vios_map, vea_sea_map = dict(), dict(), dict()
        # Loop through the SEA's populating the necessary maps
        for sea in seas:
            sea_vios_map.setdefault(sea.vios_id, list())
            sea_vios_map[sea.vios_id].append(sea)
            sea_id_map[sea.id] = sea
        # Loop through the VEA's populating the necessary maps
        for vea in veas:
            vea_vios_map.setdefault(vea.vios_id, list())
            vea_sea_map.setdefault(vea.sea_id, list())
            vea_vios_map[vea.vios_id].append(vea)
            vea_sea_map[vea.sea_id].append(vea)
            vea_id_map[vea.id] = vea
        # Now we can loop through the hosts fully populating the info
        for vios in vioses:
            vios._seas = sea_vios_map.get(vios.id, [])
            vios._veas = vea_vios_map.get(vios.id, [])
            # Loop through the SEA's fully populating the info
            for sea in vios._seas:
                sea._veas, sea._vioserver = vea_sea_map.get(sea.id, []), vios
                sea._primary_vea = vea_id_map.get(sea.primary_vea_id)
                sea._control_channel_vea = \
                    vea_id_map.get(sea.control_channel_vea_id)
            # Loop through the VEA's fully populating the info
            for vea in vios._veas:
                vea._sea, vea._vioserver = sea_id_map.get(vea.sea_id), vios


@base.NovaObjectRegistry.register
class VioServer(compute_dom.VioServer):
    """Production DOM Resource implementation for VIO Servers"""

    obj_extra_fields = compute_dom.VioServer.\
        obj_extra_fields + ['seas', 'veas']

    def __init__(self, context=None,
                 factory_class=None, transaction=None, **kwargs):
        "Constructs an instance of the given DOM Resource Class"
        super(VioServer, self).__init__(
            context, factory_class, transaction, **kwargs)

    ###############################################
    #     Implementing DOM Resource Definition    #
    ###############################################
    def _initialize_resource_attributes(self):
        """Sets any default values that are defined for any attributes"""
        super(VioServer, self)._initialize_resource_attributes()
        self._seas, self._veas = (None, None)

    def _reset_cached_attributes(self):
        """Wipes out any Cached Resource Attributes that are no longer valid"""
        super(VioServer, self)._reset_cached_attributes()
        self._seas, self._veas = (None, None)

    ###############################################
    #        Defining VIO Server Interface        #
    ###############################################
    @property
    def seas(self):
        """Getter Method for the list of SEA's for this VIOS"""
        # If the SEA's haven't been loaded yet, load them and cache them
        if self._seas is None:
            sea_fact = SharedEthernetAdapterFactory.get_factory()
            self._seas = sea_fact.find_all_seas_by_vios(
                self._context, self.id, self._transaction)
        return self._seas

    @seas.setter
    def seas(self, seas):
        """Setter Method for updating the caching of SEA's for this VIOS"""
        # Note:  This method only updates the Cache, it isn't DB update-able
        self._seas = seas if seas else list()

    @property
    def veas(self):
        """Getter Method for the list of VEA's for this VIOS"""
        # If the VEA's haven't been loaded yet, load them and cache them
        if self._veas is None:
            vea_fact = VirtualEthernetAdapterFactory.get_factory()
            self._veas = vea_fact.find_all_veas_by_vios(
                self._context, self.id, self._transaction)
        return self._veas

    @veas.setter
    def veas(self, veas):
        """Setter Method for updating the caching of VEA's for this VIOS"""
        # Note:  This method only updates the Cache, it isn't DB update-able
        self._veas = veas if veas else list()

    def add_adapter(self, adapter):
        """Setter Method to add an SEA or VEA to the cache on the VIOS.
        Note, if an adapter already exists, we'll remove the old adapter and
        add the new."""
        # Do nothing if the adapter is already on the system...
        existing_adapter = self.find_adapter(adapter)
        if existing_adapter:
            self.remove_adapter(existing_adapter)

        if isinstance(adapter, SharedEthernetAdapter):
            self.seas.append(adapter)
        if isinstance(adapter, VirtualEthernetAdapter):
            self.veas.append(adapter)

    def find_adapter(self, adapter):
        """Returns the existing adapter, if there is one, this vioserver
        already contains matching the passed in adapter."""
        if isinstance(adapter, SharedEthernetAdapter):
            for sea in self.seas:
                if sea.name == adapter.name:
                    return sea
        if isinstance(adapter, VirtualEthernetAdapter):
            for vea in self.veas:
                if vea.name == adapter.name:
                    return vea
        return None

    def remove_adapter(self, adapter):
        """Setter Method to remove an SEA or VEA to the cache on the VIOS"""
        # Do nothing if the adapter is already on the system...
        if not self.find_adapter(adapter):
            return

        if isinstance(adapter, SharedEthernetAdapter):
            self._seas.remove(adapter)
        if isinstance(adapter, VirtualEthernetAdapter):
            for temp_sea in self._seas:
                if temp_sea.primary_vea.name == adapter.name:
                    temp_sea.primary_vea = None
                elif temp_sea.control_channel_vea == adapter.name:
                    temp_sea.control_channel_vea = None
                else:
                    for temp_vea in temp_sea.veas:
                        if temp_vea.name == adapter.name:
                            temp_sea.veas.remove(adapter)
            self._veas.remove(adapter)

    def get_all_adapters(self):
        """Provides an aggregate list of all the adapters on the VioServer."""
        return self.seas + self.veas

    def get_orphan_virtual_ethernet_adapters(self, vSwitch=None):
        """
        Provides a list of all of the Orphan Virtual Ethernet Adapters present
        on the system.

        An Orphan VEA is one that is not 'owned' by a Shared Ethernet Adapter.
        """
        vea_list = []
        for vea in self.veas:
            # Check to make sure that the vSwitches match
            if vSwitch is not None and vSwitch != vea.vswitch_name:
                continue

            # Check to make sure that the adapter is a VEA, is a trunk
            # adapter, and that there is not an owning SEA.
            if vea.is_trunk and vea.sea is None:
                vea_list.append(vea)

        return vea_list

    def is_vlan_configured(self, vlan_id, vswitch=None):
        """Checks whether a VLAN ID has been configured on this VIOS"""
        for vea in self.veas:
            if vea.is_vlan_configured(vlan_id, vswitch):
                return True
        return False

    def get_reserved_vlanids_dict(self, vswitch=None):
        """Returns a dictionary of all the reserved VLAN ID's on the VIOS"""
        reserved_vlan_dict, reserved_adapters = (dict(), list())
        # A reserved adapter is an Orphan VEA, Primary VEA, or a Control
        # Channel
        orphan_vlans = self.get_vlans_orphan_vea(vswitch)
        reserved_vlan_dict['orphan'] = orphan_vlans
        # Loop through the list of SEA's, finding the Reserved VLAN's
        for sea in self.seas:
            reserved_vlan_dict[sea.name] = list()
            # The SEAs should check both the Primary VEA and the Control
            # Channel
            if vswitch is not None:
                if sea.primary_vea.vswitch_name == vswitch:
                    reserved_adapters.append(sea.primary_vea)
            else:
                reserved_adapters.append(sea.primary_vea)
            # If the Control Channel is set, check to see if it is a match
            if sea.control_channel_vea is not None:
                if ((vswitch is None or vswitch ==
                     sea.control_channel_vea.vswitch_name)):
                    reserved_adapters.append(sea.control_channel_vea)
        # Loop through all the Reserved Adapters, adding their VLAN ID's
        for adapter in reserved_adapters:
            reserved_vlanids = list()
            reserved_vlanids.append(adapter.pvid)
            # Reserve addl_vlan_ids, if there are any
            if adapter.addl_vlan_ids:
                reserved_vlanids.extend(adapter.addl_vlan_ids)
            # We need to find out the SEA for the given VEA to add the VLAN's
            reserved_vlan_dict[adapter.sea.name].extend(reserved_vlanids)
        return reserved_vlan_dict

    def get_reserved_vlanids(self, vswitch=None):
        """Returns a list of all the Reserved VLAN ID's on the VIOS"""
        reserved_vlanids = list()
        # The Dictionary of Reserved VLAN ID's is parsed into a list
        reserved_vlan_dict = self.get_reserved_vlanids_dict(vswitch)
        for adpt in reserved_vlan_dict:
            reserved_vlanids.extend(reserved_vlan_dict.get(adpt, []))
        # Sort the list of VLAN ID's in the proper order to be returned
        reserved_vlanids.sort()
        return reserved_vlanids

    def get_vlans_orphan_vea(self, vswitch=None):
        """Retrieves the list of VLAN ID's for Orphaned VEA's"""
        used_vlans, orphan_vlans = (list(), list())
        # A VLAN may be 'orphaned' and also used simultaneously.  We don't
        # want an orphan to overlap with a used VLAN, so we gather a list
        # of all used and only add the orphan if it is not in this list.
        for sea in self.seas:
            # Skip if the vSwitches don't match with the VEA for the SEA
            if vswitch and vswitch != sea.primary_vea.vswitch_name:
                continue
            # Add the PVID for the Primary VEA to the list of used VLAN ID's
            used_vlans.append(sea.primary_vea.pvid)
            # If there are additional VLAN ID's, add those to the list
            if sea.primary_vea.addl_vlan_ids:
                used_vlans.extend(sea.primary_vea.addl_vlan_ids)
            # Walk through the Additional VEA's, adding their VLAN ID's
            for vea in sea.additional_veas:
                used_vlans.append(vea.pvid)
                if vea.addl_vlan_ids:
                    used_vlans.extend(vea.addl_vlan_ids)
        # Now walk through the standard orphan, adding those to the list
        orphan_adapters = self.get_orphan_virtual_ethernet_adapters()
        for adapter in orphan_adapters:
            if vswitch and adapter.vswitch_name != vswitch:
                continue
            # These are potential orphans, but need to be cross-checked first
            potential_adds = [adapter.pvid]
            if adapter.addl_vlan_ids:
                potential_adds.extend(adapter.addl_vlan_ids)
            # Now for each potential add, add it to the list of orphaned
            for potential_add in potential_adds:
                if potential_add not in used_vlans:
                    orphan_vlans.append(potential_add)
        return orphan_vlans


######################################################
#     Shared Ethernet Adapter DOM Implementation     #
######################################################
@base.NovaObjectRegistry.register
class SharedEthernetAdapterFactory(compute_dom.VioServerComponentFactory):
    """Production DOM Factory implementation for SEA's"""

    ###############################################
    #       Defining SEA Factory Interface        #
    ###############################################
    def construct_sea(self, context, name, vios, slot, state,
                      primary_vea, control_channel_vea=None,
                      additional_veas=None, auto_create=False,
                      transaction=None):
        """Constructs (and optionally creates) the SEA's specified"""
        if not name:
            raise ValueError(_("Shared Ethernet adapter name must be "
                               "specified"))
        if not state:
            raise ValueError(_("State must be specified on the shared "
                               "Ethernet adapter creation."))
        kwargs = dict(name=name, vios_id=vios.id, slot=slot, state=state)
        # Call the Parent Class to actually Construct the given Resource
        sea = self.construct_resource(SharedEthernetAdapter, context,
                                      False, transaction, **kwargs)

        # Before we add the SEA to the vios, verify it has a valid primary vea
        if primary_vea is None:
            raise ValueError(_("The shared Ethernet adapter must have a "
                               "primary virtual Ethernet adapter"))

        # Add the SEA to the cached list of SEA's on the VIOS.
        vios.add_adapter(sea)
        sea.vioserver = vios

        # Update the primary vea to point to us.
        sea.primary_vea, primary_vea.sea = (primary_vea, sea)
        sea.add_vea_to_sea(primary_vea)

        # If they gave us a Primary VEA, update ourselves on it
        if control_channel_vea is not None:
            sea.control_channel_vea = control_channel_vea
            control_channel_vea.sea = sea
            sea.add_vea_to_sea(control_channel_vea)
        # If they gave us additional VEA's, we need to add them to the cache
        if additional_veas:
            for vea in additional_veas:
                vea.sea = sea
                sea.add_vea_to_sea(vea)
        # If they asked us to auto-create it, we need to save it first
        if auto_create is True:
            sea.save()
            for vea in sea.veas:
                vea.save()
        return sea

    def find_all_seas(self, context, filters=None, transaction=None):
        """Queries all of the SEA's matching the filter specified"""
        return self.find_all_resources(
            SharedEthernetAdapter, context, filters, transaction)

    def find_all_seas_by_vios(self, context, vios_id, transaction=None):
        """Queries all of the SEA's matching the filter specified"""
        filters = {self._get_vios_id_attr_name(): vios_id}
        return self.find_all_resources(
            SharedEthernetAdapter, context, filters, transaction)

    def find_all_seas_by_host(self, context, host_name, transaction=None):
        """Queries all of the SEA's matching the filter specified"""
        filters = dict(host_name=host_name)
        return self.find_all_resources(
            SharedEthernetAdapter, context, filters, transaction)

    def find_sea_by_id(self, context, sea_id, transaction=None):
        """Queries at most one SEA's matching the SEA ID specified"""
        filters = {self._get_resource_id_attr_name(): sea_id}
        return self.find_resource(
            SharedEthernetAdapter, context, filters, transaction)

    def delete_sea(self, context, sea_id, transaction=None):
        """Deletes the SEA with the SEA ID specified"""
        return self.delete_resource(
            SharedEthernetAdapter, context, sea_id, transaction)

    ###############################################
    #   Implementing Resource Factory Definition  #
    ###############################################
    def _get_resource_classes(self):
        """Returns a List of DOM Classes or a Dictionary Mapping DOM Classes"""
        return [SharedEthernetAdapter]

    def _auto_generate_resource_id(self, attribs):
        """Auto-generates the Resource Identifier for this Resource"""
        return self._generate_resource_uuid(attribs, ['vios_id', 'name'])

    def _get_vios_factory(self):
        """Provides the Factory to use for the VIO Server Implementation"""
        return VioServerFactory.get_factory()

    @base.remotable
    def _find_all_resources(self, resource_class,
                            filters=None, transaction=None):
        """Factory implementation to query all resources matching the filter"""
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        seas = db_api.sea_find_all(self._context, filters, dbtxn)
        return self._construct_resources(
            resource_class, self._context, seas, transaction)

    @base.remotable
    def _delete_multiple_resources(self, resource_class,
                                   resource_ids, transaction=None):
        """Factory implementation to delete the resources specified"""
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        for resource_id in resource_ids:
            db_api.sea_delete(self._context, resource_id, dbtxn)


@base.NovaObjectRegistry.register
class SharedEthernetAdapter(compute_dom.VioServerComponent):
    """Production DOM Resource implementation for Shared Ethernet Adapter"""
    fields = dict(compute_dom.VioServerComponent.fields,
                  name=fields.StringField(),
                  state=fields.StringField(nullable=True),
                  slot=fields.IntegerField(nullable=True),
                  primary_vea_id=fields.StringField(nullable=True),
                  control_channel_vea_id=fields.StringField(nullable=True))

    obj_extra_fields = compute_dom.VioServerComponent.\
        obj_extra_fields + ['veas', 'additional_veas',
                            'primary_vea', 'control_channel_vea']

    def __init__(self, context=None,
                 factory_class=None, transaction=None, **kwargs):
        "Constructs an instance of the given DOM Resource Class"
        super(SharedEthernetAdapter, self).__init__(
            context, factory_class, transaction, **kwargs)

    ###############################################
    #     Implementing DOM Resource Definition    #
    ###############################################
    def _initialize_resource_attributes(self):
        """Sets any default values that are defined for any attributes"""
        super(SharedEthernetAdapter, self)._initialize_resource_attributes()
        self._veas, self._primary_vea = (None, None)
        self._control_channel_vea = None

    def _reset_cached_attributes(self):
        """Wipes out any Cached Resource Attributes that are no longer valid"""
        super(SharedEthernetAdapter, self)._reset_cached_attributes()
        self._veas, self._primary_vea = (None, None)
        self._control_channel_vea = None

    @base.remotable
    def _create_resource(self, transaction):
        """Actual Implementation that will Create the given Resource"""
        values = self._get_resource_values()
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        sea = db_api.sea_create(self._context, values, dbtxn)
        self._hydrate_all_resource_attributes(sea)

    @base.remotable
    def _update_resource(self, transaction):
        """Actual Implementation that will Update the given Resource"""
        sea_id = self._get_resource_identifer()
        values = self._get_resource_values(changes_only=True)
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        sea = db_api.sea_update(self._context, sea_id, values, dbtxn)
        self._hydrate_all_resource_attributes(sea)

    ###############################################
    #           Defining SEA Interface            #
    ###############################################
    @property
    def veas(self):
        """Getter Method for the list of VEA's for this SEA"""
        # If the Primary VEA hasn't been loaded yet, load and cache it
        if self._veas is None:
            veafact = VirtualEthernetAdapterFactory.get_factory()
            self._veas = veafact.find_all_veas_by_sea(
                self._context, self.id, self._transaction)
        return self._veas

    @veas.setter
    def veas(self, veas):
        """Setter Method for updating the caching of VEA's for this VIOS"""
        # Note:  This method only updates the Cache, it isn't DB update-able
        self._veas = veas if veas else list()

    @property
    def additional_veas(self):
        """Getter Method for the Additional VEA's for this SEA"""
        additional_veas = list()
        # Only return it if it isn't the Primary or the Control Channel
        for vea in self.veas:
            if ((vea.id != self.primary_vea_id and
                 vea.id != self.control_channel_vea_id)):
                additional_veas.append(vea)
        return additional_veas

    @property
    def primary_vea(self):
        """Getter Method for the Primary VEA for this SEA"""
        # If the Primary VEA hasn't been loaded yet, load and cache it
        if self._primary_vea is None and self.primary_vea_id is not None:
            veafact = VirtualEthernetAdapterFactory.get_factory()
            self._primary_vea = veafact.find_vea_by_id(
                self._context, self.primary_vea_id, self._transaction)
        return self._primary_vea

    @primary_vea.setter
    def primary_vea(self, primary_vea):
        """Setter Method for the Primary VEA for this SEA"""
        self.primary_vea_id = None
        self._primary_vea = primary_vea
        # If the Primary VEA wasn't wiped out, update the ID to match
        if primary_vea is not None:
            self.primary_vea_id = primary_vea.id

    @property
    def control_channel_vea(self):
        """Getter Method for the Control Channel for this SEA"""
        # If the Control Channel hasn't been loaded yet, load and cache it
        if self._control_channel_vea is None and self.control_channel_vea_id:
            veafact = VirtualEthernetAdapterFactory.get_factory()
            self._control_channel_vea = veafact.find_vea_by_id(
                self._context, self.control_channel_vea_id, self._transaction)
        return self._control_channel_vea

    @control_channel_vea.setter
    def control_channel_vea(self, control_channel_vea):
        """Setter Method for the Control Channel for this SEA"""
        self.control_channel_vea_id = None
        self._control_channel_vea = control_channel_vea
        # If the Control Channel wasn't wiped out, update the ID to match
        if control_channel_vea is not None:
            self.control_channel_vea_id = control_channel_vea.id

    @property
    def pvid(self):
        """Getter Method for the PVID for the SEA's Primary VEA"""
        return self.primary_vea.pvid

    @property
    def trunk_pri(self):
        """Getter Method for the Trunk Priority for the SEA's Primary VEA"""
        return self.primary_vea.trunk_pri

    @property
    def addl_vlan_ids(self):
        """Getter Method for the VLAN ID's for the SEA's Primary VEA"""
        return self.primary_vea.addl_vlan_ids

    def all_addl_vlan_ids(self):
        """
        Returns all of the VEAs additional VLAN IDs.  Does not include the
        control channel
        """
        vlans = []
        vlans.extend(self.primary_vea.addl_vlan_ids)
        for vea in self.additional_veas:
            vlans.extend(vea.addl_vlan_ids)
        return vlans

    def add_vea_to_sea(self, vea):
        """Setter Method to add a VEA to the cached list for the SEA"""
        if vea is None:
            return

        if not isinstance(vea, VirtualEthernetAdapter):
            raise ValueError(_("Object must be a virtual Ethernet adapter for "
                               "addition to a shared Ethernet adapter"))

        self._veas = self._veas if self._veas else list()
        in_list = [avea for avea in self._veas if avea.id == vea.id]
        # If the VEA wasn't in the existing SEA list, add it to it
        if len(in_list) <= 0:
            self._veas.append(vea)

    def remove_vea_from_sea(self, vea):
        """Setter Method to remove a VEA to the cached list for the SEA"""
        if vea is None:
            return

        if not isinstance(vea, VirtualEthernetAdapter):
            raise ValueError(_("Object must be a virtual Ethernet adapter for "
                               "addition to a shared Ethernet adapter"))

        self._veas = self._veas if self._veas else list()
        in_list = [avea for avea in self._veas if avea.id == vea.id]
        # If the VEA wasn't in the existing SEA list, add it to it
        if len(in_list) > 0:
            self._veas.remove(vea)

    def is_available(self):
        """Aggregates the various states (SEA/VIOS/RMC) into availableness"""
        if self.vioserver is not None:
            return self.vioserver.state.lower() == 'running' and \
                self.vioserver.rmc_state.lower() == 'active' and \
                self.state.lower() == 'available'
        else:
            return self.state.lower() == 'available'

    def is_managing_vea(self, vea):
        """Inspects all VEAs on this SEA for a specific VEA."""
        # If they didn't give us a VEA, then there is nothing to check
        if vea is not None:
            # Loop through all of the VEA's for this SEA, seeing if a match
            for avea in self.veas:
                if avea.id == vea.id:
                    return True
        return False

    def get_vea_by_vlan_id(self, vlan_id, throw_error=True):
        """Retrieve the VEA based on the given VLAN ID"""
        # If the VLAN ID is configured on the Primary VEA, return that
        if self.primary_vea.is_vlan_configured(vlan_id,
                                               throw_error=throw_error):
            return self.primary_vea
        # If the VLAN ID is configured on another VEA, return that one
        for vea in self.additional_veas:
            if vea.is_vlan_configured(vlan_id, throw_error=throw_error):
                return vea
        return None

    def get_least_used_vea(self):
        """Scan the VEA's to find the one with the fews VLAN's bridged"""
        # If there are no additional VEA's then none to choose from
        if not self.additional_veas or len(self.additional_veas) <= 0:
            return None
        # Start with the first one, and see if any better suited
        least_used_vea = self.additional_veas[0]
        vea_8021Q_vlan_counts = len(least_used_vea.addl_vlan_ids)
        # Loop through each of the VEA's, checking their VLAN counts
        for vea in self.additional_veas:
            if len(vea.addl_vlan_ids) < vea_8021Q_vlan_counts:
                vea_8021Q_vlan_counts = len(vea.addl_vlan_ids)
                least_used_vea = vea
        # Return the VEA with the fewest number of VLAN's bridged
        return least_used_vea


######################################################
#     Virtual Ethernet Adapter DOM Implementation    #
######################################################
@base.NovaObjectRegistry.register
class VirtualEthernetAdapterFactory(compute_dom.VioServerComponentFactory):
    """Production DOM Factory implementation for VEA's"""

    ###############################################
    #       Defining VEA Factory Interface        #
    ###############################################
    def construct_vea(self, context, name, vios, slot, pvid, is_trunk,
                      trunk_pri, state, ieee_eth, vswitch_name, addl_vlan_ids,
                      sea_id=None, auto_create=False, transaction=None):
        """Constructs (and optionally creates) the VEA's specified"""
        kwargs = dict(name=name, vios_id=vios.id, slot=slot, is_trunk=is_trunk,
                      state=state, trunk_pri=trunk_pri, ieee_eth=ieee_eth,
                      pvid=pvid, vswitch_name=vswitch_name,
                      addl_vlan_ids=copy.copy(addl_vlan_ids), sea_id=sea_id)
        # Some value validation
        if name == '':
            raise ValueError(_("Name of the virtual Ethernet adapter must "
                               "not be empty"))
        if trunk_pri < 0:
            raise ValueError(_("Trunk Priority on the virtual Ethernet "
                               "adapter must not be less than 0."))
        if not state:
            raise ValueError(_("The state must be set on the virtual Ethernet "
                               "adapter."))
        if pvid <= 0 or pvid > 4095:
            raise ValueError(_("The PVID of the virtual Ethernet adapter "
                               "must be between 1 and 4095"))
        if slot <= 0:
            raise ValueError(_("The slot must be set to a positive value."))
        if not vswitch_name:
            raise ValueError(_("The virtual switch for the virtual Ethernet "
                               "adapter must be specified."))
        # Call the Parent Class to actually Construct the given Resource
        vea = self.construct_resource(VirtualEthernetAdapter, context,
                                      auto_create, transaction, **kwargs)
        # Add the VEA to the cached list of VEA's on the VIOS.
        vios.add_adapter(vea)
        vea.vioserver = vios
        return vea

    def find_all_veas(self, context, filters=None, transaction=None):
        """Queries all of the VEA's matching the filter specified"""
        return self.find_all_resources(
            VirtualEthernetAdapter, context, filters, transaction)

    def find_all_veas_by_vios(self, context, vios_id, transaction=None):
        """Queries all of the VEA's matching the filter specified"""
        filters = {self._get_vios_id_attr_name(): vios_id}
        return self.find_all_resources(
            VirtualEthernetAdapter, context, filters, transaction)

    def find_all_veas_by_host(self, context, host_name, transaction=None):
        """Queries all of the VEA's matching the filter specified"""
        filters = dict(host_name=host_name)
        return self.find_all_resources(
            VirtualEthernetAdapter, context, filters, transaction)

    def find_all_veas_by_sea(self, context, sea_id, transaction=None):
        """Queries all of the VEA's matching the filter specified"""
        filters = dict(sea_id=sea_id)
        return self.find_all_resources(
            VirtualEthernetAdapter, context, filters, transaction)

    def find_vea_by_id(self, context, vea_id, transaction=None):
        """Queries at most one VEA's matching the VEA ID specified"""
        filters = {self._get_resource_id_attr_name(): vea_id}
        return self.find_resource(
            VirtualEthernetAdapter, context, filters, transaction)

    def delete_vea(self, context, vea_id, transaction=None):
        """Deletes the VEA with the VEA ID specified"""
        return self.delete_resource(
            VirtualEthernetAdapter, context, vea_id, transaction)

    ###############################################
    #   Implementing Resource Factory Definition  #
    ###############################################
    def _get_resource_classes(self):
        """Returns a List of DOM Classes or a Dictionary Mapping DOM Classes"""
        return [VirtualEthernetAdapter]

    def _auto_generate_resource_id(self, attribs):
        """Auto-generates the Resource Identifier for this Resource"""
        return self._generate_resource_uuid(attribs, ['vios_id', 'name'])

    def _get_vios_factory(self):
        """Provides the Factory to use for the VIO Server Implementation"""
        return VioServerFactory.get_factory()

    @base.remotable
    def _find_all_resources(self, resource_class,
                            filters=None, transaction=None):
        """Factory implementation to query all resources matching the filter"""
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        veas = db_api.vea_find_all(self._context, filters, dbtxn)
        return self._construct_resources(
            resource_class, self._context, veas, transaction)

    @base.remotable
    def _delete_multiple_resources(self, resource_class,
                                   resource_ids, transaction=None):
        """Factory implementation to delete the resources specified"""
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        for resource_id in resource_ids:
            db_api.vea_delete(self._context, resource_id, dbtxn)


@base.NovaObjectRegistry.register
class VirtualEthernetAdapter(compute_dom.VioServerComponent):
    """Production DOM Resource implementation for Virtual Ethernet Adapter"""

    fields = dict(compute_dom.VioServerComponent.fields,
                  name=fields.StringField(),
                  state=fields.StringField(nullable=True),
                  slot=fields.IntegerField(nullable=True),
                  pvid=fields.IntegerField(nullable=True),
                  is_trunk=fields.BooleanField(nullable=True),
                  trunk_pri=fields.IntegerField(nullable=True),
                  ieee_eth=fields.BooleanField(nullable=True),
                  vswitch_name=fields.StringField(nullable=True),
                  addl_vlan_ids=dom.ListOfIntegersField(nullable=True),
                  sea_id=fields.StringField(nullable=True))

    obj_extra_fields = compute_dom.\
        VioServerComponent.obj_extra_fields + ['sea']

    def __init__(self, context=None,
                 factory_class=None, transaction=None, **kwargs):
        "Constructs an instance of the given DOM Resource Class"
        super(VirtualEthernetAdapter, self).__init__(
            context, factory_class, transaction, **kwargs)

    ###############################################
    #     Implementing DOM Resource Definition    #
    ###############################################
    def _initialize_resource_attributes(self):
        """Sets any default values that are defined for any attributes"""
        super(VirtualEthernetAdapter, self)._initialize_resource_attributes()
        self._sea = None

    def _reset_cached_attributes(self):
        """Wipes out any Cached Resource Attributes that are no longer valid"""
        super(VirtualEthernetAdapter, self)._reset_cached_attributes()
        self._sea = None

    @base.remotable
    def _create_resource(self, transaction):
        """Actual Implementation that will Create the given Resource"""
        values = self._get_resource_values()
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        vea = db_api.vea_create(self._context, values, dbtxn)
        self._hydrate_all_resource_attributes(vea)

    @base.remotable
    def _update_resource(self, transaction):
        """Actual Implementation that will Update the given Resource"""
        vea_id = self._get_resource_identifer()
        values = self._get_resource_values(changes_only=True)
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        vea = db_api.vea_update(self._context, vea_id, values, dbtxn)
        self._hydrate_all_resource_attributes(vea)

    ###############################################
    #           Defining VEA Interface            #
    ###############################################
    @property
    def sea(self):
        """Getter Method for the SEA for this VEA"""
        # If the SEA hasn't been loaded yet, load and cache it
        if self._sea is None and self.sea_id is not None:
            fact = SharedEthernetAdapterFactory.get_factory()
            self._sea = fact.find_sea_by_id(
                self._context, self.sea_id, self._transaction)
        return self._sea

    @sea.setter
    def sea(self, sea):
        """Setter Method for the SEA for this VEA"""
        self.sea_id = None
        self._sea = sea
        # If the SEA wasn't wiped out, update the ID to match
        if sea is not None:
            self.sea_id = sea.id

    def is_pvid_adapter(self):
        """Determines whether this VEA is the Primary for the SEA"""
        if self.sea is None:
            return False
        return self.sea.primary_vea.id == self.id

    def is_ctl_channel_adapter(self):
        """Determines whether this VEA is the Control Channel for the SEA"""
        if self.sea is None:
            return False
        return self.sea.control_channel_vea and\
            self.sea.control_channel_vea.id == self.id

    def is_vlan_configured(self, vlan_id, vswitch_name=None, throw_error=True):
        """Determines whether a given VLAN ID has been configured on the VEA"""
        # Not the same domain if the vSwitches on the system do not match
        if vswitch_name and vswitch_name != self.vswitch_name:
            return False
        # Runtime checking to ensure data integrity of the IEEE VLANs
        if not self.ieee_eth and len(self.addl_vlan_ids) > 0:
            # If they asked to throw an exception, do so, else return True
            if throw_error is True:
                raise excp.IBMPowerVMInvalidVETHConfig(attr=self.name)
            return True
        # The VLAN may be attached to either the primary or additional VEA
        if self.is_pvid_adapter():
            return vlan_id == self.pvid or vlan_id in self.addl_vlan_ids
        # Check to see if this is the Control Channel VEA Adapter
        elif self.is_ctl_channel_adapter():
            # The control channel adapter can't be used at all, check if a
            # match
            if vlan_id in self.addl_vlan_ids or vlan_id == self.pvid:
                # If they asked to throw an exception, do so, else return True
                if throw_error is True:
                    raise excp.IBMPowerVMVIDisCtlChannel()
                return True
        # Non-PVID Adapter VEA's port_vlan_id is picked randomly above 4000
        elif vlan_id == self.pvid:
            # If they asked to throw an exception, do so, else return True
            if throw_error is True:
                raise excp.IBMPowerVMVIDPVIDConflict(vea=self)
            return True
        else:
            return vlan_id in self.addl_vlan_ids
        return False


######################################################
#    NetworkAssociation DOM Implementation           #
######################################################
@base.NovaObjectRegistry.register
class NetworkAssociationFactory(dom.ResourceFactory):
    """Production DOM Factory implementation for NetworkAssociations"""

    ###############################################
    #       Defining VEA Factory Interface        #
    ###############################################
    def construct_network_assoc(self, context, host_name,
                                neutron_net_id, sea=None,
                                auto_create=False, transaction=None):
        """Constructs (and optionally creates) a Network Assoc as specified"""
        kwargs = dict(host_name=host_name, neutron_net_id=neutron_net_id)
        # Call the Parent Class to actually Construct the given Resource
        netassoc = self.construct_resource(NetworkAssociation, context,
                                           transaction, **kwargs)
        netassoc.sea = sea
        # If they asked us to auto-create it, we need to save it first
        if auto_create is True:
            netassoc.save()
        return netassoc

    def update_network_assoc(self, context, host_name, neutron_net_id,
                             sea=None, transaction=None):
        """Creates or Updates (if exists) the Network Assoc as specified"""
        netw_assoc = self.find_network_assoc(
            context, host_name, neutron_net_id, transaction)

        # If the Network Association doesn't exist, we need to create it
        if netw_assoc is None:
            return self.construct_network_assoc(
                context, host_name, neutron_net_id, sea, True, transaction)
        # Otherwise just update the SEA on the existing Network Association
        else:
            netw_assoc.sea = sea
            netw_assoc.save()
        return netw_assoc

    def find_network_assoc(self, context, host_name, neutron_net_id,
                           transaction=None, load_all=False):
        """Queries at most one Network Assoc matching the filter specified"""
        filters = dict(host_name=host_name, neutron_net_id=neutron_net_id)
        network_assoc = self.find_resource(
            NetworkAssociation, context, filters, transaction)
        # If they asked us to load all the info, do it now
        if load_all and network_assoc is not None:
            self.__load_netasc_info(context, [network_assoc], transaction)
        return network_assoc

    def find_all_network_assocs(self, context, filters=None,
                                transaction=None, load_all=False):
        """Queries all of the Network Assocs matching the filter specified"""
        network_assocs = self.find_all_resources(
            NetworkAssociation, context, filters, transaction)
        # If they asked us to load all the info, do it now
        if load_all and network_assocs:
            self.__load_netasc_info(context, network_assocs, transaction)
        return network_assocs

    def find_all_network_assocs_by_host(self, context, host_name,
                                        transaction=None, load_all=False):
        """Queries all of the Network Assocs for the given Host"""
        filters = dict(host_name=host_name)
        network_assocs = self.find_all_resources(
            NetworkAssociation, context, filters, transaction)
        # If they asked us to load all the info, do it now
        if load_all and network_assocs:
            self.__load_netasc_info(context, network_assocs, transaction)
        return network_assocs

    def find_all_network_assocs_by_network(self, context, neutron_net_id,
                                           transaction=None, load_all=False):
        """Queries all of the Network Assocs for the Network ID"""
        filters = dict(neutron_net_id=neutron_net_id)
        network_assocs = self.find_all_resources(
            NetworkAssociation, context, filters, transaction)
        # If they asked us to load all the info, do it now
        if load_all and network_assocs:
            self.__load_netasc_info(context, network_assocs, transaction)
        return network_assocs

    def delete_network_assoc(self, context, netwassoc_id, transaction=None):
        """Deletes the Network Association with the NetAssoc ID specified"""
        return self.delete_resource(
            NetworkAssociation, context, netwassoc_id, transaction)

    def cleanup_unused_ports(self, context):
        """Cleans up any unused Network Ports from the Neutron Database"""
        try:
            self._context = context
            self._cleanup_unused_ports()
        # Just log an exception we get from the DB or REST or PRC
        except Exception as exc:
            LOG.exception(exc)

    ###############################################
    #   Implementing Resource Factory Definition  #
    ###############################################
    def _get_resource_classes(self):
        """Returns a List of DOM Classes or a Dictionary Mapping DOM Classes"""
        return [NetworkAssociation]

    def _auto_generate_resource_id(self, attribs):
        """Auto-generates the Resource Identifier for this Resource"""
        return self._generate_resource_uuid(
            attribs, ['host_name', 'neutron_net_id'])

    @base.remotable
    def _find_all_resources(self, resource_class,
                            filters=None, transaction=None):
        """Factory implementation to query all resources matching the filter"""
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        nwassns = db_api.network_assoc_find_all(self._context, filters, dbtxn)
        return self._construct_resources(
            resource_class, self._context, nwassns, transaction)

    @base.remotable
    def _delete_multiple_resources(self, resource_class,
                                   resource_ids, transaction=None):
        """Factory implementation to delete the resources specified"""
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        for resource_id in resource_ids:
            db_api.network_assoc_delete(self._context, resource_id, dbtxn)

    ###############################################
    #        Shared Host DOM Helper Methods       #
    ###############################################
    @base.remotable
    def _cleanup_unused_ports(self):
        """Cleans up any unused Network Ports from the Neutron Database"""
        unused_ports = list()
        neutron_client = neutronv2.api.get_client(self._context)
        # Call through the DOM to the Nova DB to get the list of Instances
        instfact = ManagedInstanceFactory.get_factory()
        instances = instfact.find_all_instances(self._context)
        # Call thru the REST API to Neutron to retrieve all of the Ports
        network_ports = neutron_client.list_ports().get('ports', [])
        # Loop through the Ports/Instances finding which ones are unused
        for port in network_ports:
            port_used = False
            device_id = port.get('device_id')
            device_owner = port.get('device_owner')
            # Loop through each of the Instances, checking for a match
            for instance in instances:
                if device_id and instance['uuid'] == device_id:
                    port_used = True
                    break
            # Only delete Ports owned by Compute, not owned by SCE, etc
            if not port_used and device_owner == 'compute:None':
                unused_ports.append(port['id'])
        # Loop through the Unused Ports, try to delete each one from Neutron
        for unused_port in unused_ports:
            try:
                neutron_client.delete_port(unused_port)
            # Just log it if an error occurred, so we can delete the rest
            except Exception as exc:
                LOG.exception(exc)

    ###############################################
    #      Internal Host DOM Helper Methods       #
    ###############################################
    def __load_netasc_info(self, context, netascs, transaction):
        """Helper method to load lazy-loaded host attributes"""
        host_map, sea_map = dict(), dict()
        host_name = netascs[0].host_name if len(netascs) == 1 else None
        filters = dict(host_name=host_name) if host_name else None
        # Load all the VIOS/SEA/VEA instances to do the population
        hostfact = ManagedHostFactory.get_factory()
        hosts = hostfact.find_all_hosts(
            context, filters, transaction, load_all=True)
        # Loop through the Hosts populating the necessary maps
        for host in hosts:
            for sea in host.seas:
                sea_map[sea.id] = sea
            host_map[host.host_name] = host
        # Now we can loop through the NetAssocs fully populating the info
        for netasc in netascs:
            netasc._host = host_map.get(netasc.host_name)
            netasc._sea = sea_map.get(netasc.sea_id)


@base.NovaObjectRegistry.register
class NetworkAssociation(dom.Resource):
    """Production DOM Resource implementation for Virtual Ethernet Adapter"""

    fields = dict(dom.Resource.fields,
                  id=fields.StringField(),
                  host_name=fields.StringField(),
                  neutron_net_id=fields.StringField(),
                  sea_id=fields.StringField(nullable=True))

    obj_extra_fields = ['host', 'sea']

    def __init__(self, context=None,
                 factory_class=None, transaction=None, **kwargs):
        "Constructs an instance of the given DOM Resource Class"
        super(NetworkAssociation, self).__init__(
            context, factory_class, transaction, **kwargs)

    ###############################################
    #     Implementing DOM Resource Definition    #
    ###############################################
    def _initialize_resource_attributes(self):
        """Sets any default values that are defined for any attributes"""
        super(NetworkAssociation, self)._initialize_resource_attributes()
        self._host, self._sea = (None, None)

    def _reset_cached_attributes(self):
        """Wipes out any Cached Resource Attributes that are no longer valid"""
        super(NetworkAssociation, self)._reset_cached_attributes()
        self._host, self._sea = (None, None)

    @base.remotable
    def _create_resource(self, transaction):
        """Actual Implementation that will Create the given Resource"""
        values = self._get_resource_values()
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        nwassn = db_api.network_assoc_create(self._context, values, dbtxn)
        self._hydrate_all_resource_attributes(nwassn)

    @base.remotable
    def _update_resource(self, transaction):
        """Actual Implementation that will Update the given Resource"""
        nwassn_id = self._get_resource_identifer()
        values = self._get_resource_values(changes_only=True)
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        nwassn = db_api.network_assoc_update(
            self._context, nwassn_id, values, dbtxn)
        self._hydrate_all_resource_attributes(nwassn)

    @property
    def host(self):
        """Getter Method for the Host for this Network Association"""
        # If the Host hasn't been loaded yet, load and cache it
        if self._host is None and self.host_name is not None:
            hostfact = ManagedHost.get_factory()
            self._host = hostfact.find_host_by_name(
                self._context, self.host_name, self._transaction)
        return self._host

    @host.setter
    def host(self, host):
        """Setter Method for the Host for this Network Association"""
        self.host_name = None
        self._host = host
        # If the Host wasn't wiped out, update the Name to match
        if host is not None:
            self.host_name = host.id

    @property
    def sea(self):
        """Getter Method for the SEA for this Network Association"""
        # If the SEA hasn't been loaded yet, load and cache it
        if self._sea is None and self.sea_id is not None:
            seafact = SharedEthernetAdapterFactory.get_factory()
            self._sea = seafact.find_sea_by_id(
                self._context, self.sea_id, self._transaction)
        return self._sea

    @sea.setter
    def sea(self, sea):
        """Setter Method for the SEA for this Network Association"""
        self.sea_id = None
        self._sea = sea
        # If the SEA wasn't wiped out, update the ID to match
        if sea is not None:
            self.sea_id = sea.id


######################################################
#      Host Network Topology DOM Implementation      #
######################################################
@base.NovaObjectRegistry.register
class HostNetworkTopologyFactory(dom.ResourceFactory):
    """Production DOM Factory implementation for Host Network Topology"""

    ###############################################
    #       Defining HNT Factory Interface        #
    ###############################################
    def construct_network_topology(self, context, host_name,
                                   vioservers=[], transaction=None):
        """Constructs the Network Topology for the Host specified"""
        # Call the Parent Class to actually Construct the given Resource
        topo = self.construct_resource(HostNetworkTopology, context, False,
                                       transaction, host_name=host_name)
        # We need to do something special to traverse the VIOS for SEA/VEA's
        for vios in vioservers:
            if not len(vios.seas):
                LOG.warning(_('No SEA in vios list'))
            topo.vioservers.append(vios)
            topo.seas.extend(vios.seas)
            topo.veas.extend(vios.veas)
        return topo

    def construct_network_topology_from_host(self, context,
                                             host_dom, transaction=None):
        """Constructs the Network Topology for the Host specified"""
        return self.construct_network_topology(
            context, host_dom.host_name, host_dom.vioservers, transaction)

    def find_network_topology(self, context, host_name, transaction=None):
        """Queries the Network Topology for the Host specified"""
        return self.find_resource(HostNetworkTopology, context,
                                  dict(host_name=host_name), transaction)

    ###############################################
    #   Implementing Resource Factory Definition  #
    ###############################################
    def _get_resource_id_attr_name(self):
        """Provides the Name of the Attribute for the Resource Identifier"""
        return 'host_name'

    def _get_resource_classes(self):
        """Returns a List of DOM Classes or a Dictionary Mapping DOM Classes"""
        return [HostNetworkTopology]

    @base.remotable
    def _find_all_resources(self, resource_class,
                            filters=None, transaction=None):
        """Factory implementation to query all resources matching the filter"""
        return list()


@base.NovaObjectRegistry.register
class HostNetworkTopology(dom.Resource):
    """Production DOM Resource implementation for Host Network Topology"""

    fields = dict(dom.Resource.fields,
                  host_name=fields.StringField(),
                  vioservers=dom.DomListField(VioServer),
                  seas=dom.DomListField(SharedEthernetAdapter),
                  veas=dom.DomListField(VirtualEthernetAdapter))

    def __init__(self, context=None,
                 factory_class=None, transaction=None, **kwargs):
        "Constructs an instance of the given DOM Resource Class"
        super(HostNetworkTopology, self).__init__(
            context, factory_class, transaction, **kwargs)

    ###############################################
    #     Implementing DOM Resource Definition    #
    ###############################################
    def _initialize_resource_attributes(self):
        """Sets any default values that are defined for any attributes"""
        super(HostNetworkTopology, self)._initialize_resource_attributes()
        setattr(self, 'vioservers', list())
        setattr(self, 'seas', list())
        setattr(self, 'veas', list())

    def _create_resource(self, transaction):
        """Actual Implementation that will Create the given Resource"""
        # Since the Topology isn't actually persisted we will treat
        # someone who queried the Topology first the same as construct it
        self._update_resource(transaction)

    @base.remotable
    def _update_resource(self, transaction):
        """Actual Implementation that will Update the given Resource"""
        netfact = NetworkAssociationFactory.get_factory()
        # If the Transaction isn't set, then we need to create one to use
        if transaction is None:
            topofact = self._get_resource_factory()
            transaction = topofact.get_resource_transaction()
        # Begin the Transaction to make sure everything is committed together
        with transaction.begin(subtransactions=True):
            dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
            # If there weren't any VIOS's returned from the Host, mark as error
            if len(self.vioservers) <= 0:
                self._mark_system_unavailable(self._context, dbtxn)
            else:
                # Query the Database to get the list of existing NA's
                net_assocs = netfact.find_all_network_assocs_by_host(
                    self._context, self.host_name, transaction)
                # Need to force the retrieval of the Primary VEA's for the SEA
                for net_assoc in net_assocs:
                    if net_assoc.sea is not None:
                        net_assoc.sea.primary_vea
                # Merge in the latest VIOS Information into the Database
                active_vioses, deleted_vioses =\
                    self._update_vioses(self._context, dbtxn)
                # Merge in the latest VEA Information into the Database
                vea_seas = self._update_veas(
                    self._context, active_vioses, dbtxn)
                # Merge in the latest SEA Information into the Database
                self._update_seas(self._context, transaction, active_vioses,
                                  deleted_vioses, vea_seas, net_assocs)
                # Update the SEA's for any Network Association in the DB
                self._update_network_associations(
                    self._context, transaction, net_assocs)
                # Now we want to clean up any VIOS's that no longer exist
                for vios_id in deleted_vioses:
                    db_api.vios_delete(self._context, vios_id, dbtxn)
        # Invoke the Hook at the end to Cleanup any Unused Network Ports
        self._cleanup_unused_ports(self._context)

    ###############################################
    #        Defining Topology Interface          #
    ###############################################
    def find_vio_server(self, lpar_id=0, lpar_name=None):
        """Retrieves the VIOS on this Host which matches the LPAR ID/Name"""
        return ManagedHost._find_vio_server(
            self.vioservers, lpar_id, lpar_name)

    def find_all_primary_seas(self):
        """Retrieves a list of all of the Primary SEA's on the Host"""
        return [sea_chain[0] for sea_chain in
                ManagedHost._find_sea_chains(self.seas)]

    def find_adapter(self, lpar_id, adapter_name):
        """Returns an adapter on a given LPAR for the Host, if found"""
        return ManagedHost._find_adapter(
            self.vioservers, lpar_id, adapter_name)

    def maintain_available_vid_pool(self):
        """Looks up all the active VLAN ID's on the Host and removes them"""
        """from the available VID Pool to avoid conflicts during Plug time"""
        # List storing the available VLAN ID's used for the PVID VEA's Port VID
        self._available_vid_pool = dict()

        # Find all of the VLAN ID's in use for each vSwitch including
        # port VLAN ID and additional 8021Q VLAN ID's from the PowerVM.
        # Those VIDS will be removed from the available VID pool.
        for vios in self.vioservers:
            for vea in vios.veas:
                vswitch_name = vea.vswitch_name
                # If we haven't encountered this vswitch yet, init it's list
                if vswitch_name not in self._available_vid_pool.keys():
                    self._available_vid_pool[vswitch_name] = range(1, 4095)

                # Now remove used vids from this vswitch's pool
                if self._available_vid_pool[vswitch_name].count(vea.pvid) > 0:
                    self._available_vid_pool[vswitch_name].remove(vea.pvid)
                for vid in vea.addl_vlan_ids:
                    if self._available_vid_pool[vswitch_name].count(vid) > 0:
                        self._available_vid_pool[vswitch_name].remove(vid)

    ###############################################
    #      Shared Topology DOM Helper Methods     #
    ###############################################
    def _update_vioses(self, context, dbtxn):
        """Helper method to update the VIOS's in the DB to match Topology"""
        deleted_vioses = list()
        viosfmt = "%(host_name)s##%(lpar_id)s"
        filters = dict(host_name=self.host_name)
        # Retrieve the existing VIOS's for the Host to figure out differences
        dbvioses = db_api.vios_find_all(context, filters, dbtxn)
        exist_vioses = dict([(vios['id'], vios) for vios in dbvioses])
        exist_vioses2 = dict([(viosfmt % vios, vios) for vios in dbvioses])
        new_vioses = dict([(vios['id'], vios) for vios in self.vioservers])
        active_vioses = [vios['id'] for vios in self.vioservers
                         if vios['rmc_state'] == 'active']
        # Determine whether the VIOS needs to be Created or updated
        for vios in new_vioses.values():
            values = vios._get_resource_values(changes_only=False)
            # We aren't the ones updating the VIOS Cluster, so remove that
            values.pop('cluster_provider_name', None)
            # If the VIOS already exists, we just want to update what changed
            if exist_vioses.pop(vios['id'], None) is not None:
                db_api.vios_update(context, vios['id'], values, dbtxn)
            # Since the K2 ID was added new to the VIOS table in 1.2.2, we
            # need to fill it in for existing VIOS's and since the upgrade
            # doesn't have access to it, need to update the random one here
            elif exist_vioses2.get(viosfmt % vios) is not None:
                orig_id = exist_vioses2[viosfmt % vios]['id']
                self._correct_vios_uuid(context, orig_id, vios['id'], dbtxn)
                db_api.vios_update(context, vios['id'], values, dbtxn)
                exist_vioses.pop(orig_id, None)
            # If the VIOS is new, then we want to create it in the database
            else:
                db_api.vios_create(context, values, dbtxn)
        # Now we want to tag any VIOS's that no longer exist in topology
        for vios in exist_vioses.values():
            if new_vioses.get(vios['id']) is None:
                deleted_vioses.append(vios['id'])
        return active_vioses, deleted_vioses

    def _update_seas(self, context, transaction, active_vioses,
                     deleted_vioses, vea_seas, net_assocs):
        """Helper method to update the SEA's in the DB to match Topology"""
        filters = dict(host_name=self.host_name)
        dbtxn, deleted_seas = (transaction.dbtxn, list())
        # Retrieve the existing SEA's for the Host to figure out differences
        dbseas = db_api.sea_find_all(context, filters, dbtxn)
        exist_seas = dict([(sea['id'], sea) for sea in dbseas])
        new_seas = dict([(sea['id'], sea) for sea in self.seas])
        # Determine whether the SEA needs to be Created or updated
        for sea in new_seas.values():
            values = sea._get_resource_values(changes_only=False)
            # We won't do any updates to VIOS's that we can't contact RMC
            if values['vios_id'] not in active_vioses:
                continue
            # If the SEA already exists, we just want to update what changed
            if exist_seas.pop(sea['id'], None) is not None:
                db_api.sea_update(context, sea['id'], values, dbtxn)
            # If the SEA is new, then we want to create it in the database
            else:
                db_api.sea_create(context, values, dbtxn)
        # Now we want to clean up any SEA's that no longer exist in the
        # topology
        for sea in exist_seas.values():
            if not new_seas.get(sea['id']):
                if sea['vios_id'] in (active_vioses + deleted_vioses):
                    deleted_seas.append(sea['id'])
        # If there were any deleted SEA's, update any NA that is referencing it
        if len(deleted_seas) > 0 and len(net_assocs) > 0:
            self._handle_deleted_na_seas(
                context, transaction, net_assocs, deleted_seas)
        # Now we can actually go through and clean up the deleted SEA's
        for sea_id in deleted_seas:
            # Since the VEA's have constraints to the SEA, need to null that
            # out
            for v in db_api.vea_find_all(context, dict(sea_id=sea_id), dbtxn):
                LOG.warning(_('Removing sea %(sea_id)s from vea %(vea_id)s'),
                            {'sea_id': sea_id, 'vea_id': v['id']})
                db_api.vea_update(context, v['id'], dict(sea_id=None), dbtxn)
            db_api.sea_delete(context, sea_id, dbtxn)
        # Finally need to add the SEA ID's to the VEA's now that the SEA's
        # exist
        for vea_id, sea_id in vea_seas.iteritems():
            db_api.vea_update(context, vea_id, dict(sea_id=sea_id), dbtxn)

    def _update_veas(self, context, active_vioses, dbtxn):
        """Helper method to update the VEA's in the DB to match Topology"""
        filters, vea_seas = (dict(host_name=self.host_name), dict())
        # Retrieve the existing VEA's for the Host to figure out differences
        dbveas = db_api.vea_find_all(context, filters, dbtxn)
        exist_veas = dict([(vea['id'], vea) for vea in dbveas])
        new_veas = dict([(vea['id'], vea) for vea in self.veas])
        # Determine whether the VEA needs to be Created or updated
        for vea in new_veas.values():
            values = vea._get_resource_values(changes_only=False)
            # We won't do any updates to VIOS's that we can't contact RMC
            if values['vios_id'] not in active_vioses:
                continue
            # Need to take out the SEA ID, since the SEA doesn't exist yet
            if values.get('sea_id') is not None:
                vea_seas[vea['id']] = values.pop('sea_id')
            # If the VEA already exists, we just want to update what changed
            if exist_veas.pop(vea['id'], None) is not None:
                # Need to take out the SEA ID, since the SEA doesn't exist yet
                if values.get('sea_id') is not None:
                    vea_seas[vea['id']] = values.pop('sea_id')
                db_api.vea_update(context, vea['id'], values, dbtxn)
            # If the VEA is new, then we want to create it in the database
            else:
                # Need to take out the SEA ID, since the SEA doesn't exist yet
                if values.get('sea_id') is not None:
                    vea_seas[vea['id']] = values.pop('sea_id')
                db_api.vea_create(context, values, dbtxn)
        # Now we want to clean up any VEA's that no longer exist in the
        # topology
        for vea in exist_veas.values():
            if not new_veas.get(vea['id']) and vea['vios_id'] in active_vioses:
                db_api.vea_delete(context, vea['id'], dbtxn)
        return vea_seas

    def _correct_vios_uuid(self, context, orig_id, new_id, dbtxn):
        """Helper method to correct the VIOS ID to the K2 UUID"""
        namespc = uuid.UUID('a0dd4880-6115-39d6-b26b-77df18fe749f')
        # First update the VIOS ID from the original value to the new one
        db_api.vios_update(context, orig_id, dict(id=new_id), dbtxn)
        # Next update the SEA ID's based on the updated VIOS ID
        dbseas = db_api.sea_find_all(context, dict(vios_id=new_id), dbtxn)
        for dbsea in dbseas:
            namestr = "ibm-powervc://SharedEthernetAdapterFactory/"
            namestr += "vios_id='" + new_id + "',name='" + dbsea['name'] + "'"
            sea_id = str(uuid.uuid3(namespc, str(namestr)))
            db_api.sea_update(context, dbsea['id'], dict(id=sea_id), dbtxn)
        # Next update the VEA ID's based on the updated VIOS ID
        dbveas = db_api.vea_find_all(context, dict(vios_id=new_id), dbtxn)
        for dbvea in dbveas:
            namestr = 'ibm-powervc://VirtualEthernetAdapterFactory/'
            namestr += 'vios_id=' + new_id + ',name=' + dbvea['name']
            vea_id = str(uuid.uuid3(namespc, str(namestr)))
            db_api.vea_update(context, dbvea['id'], dict(id=vea_id), dbtxn)

    def _mark_system_unavailable(self, context, dbtxn):
        """Helper method to update the VIOS's to Error in the Database"""
        filters = dict(host_name=self.host_name)
        vios_values = dict(rmc_state='inactive')
        sea_values = dict(state='inactive')
        # Retrieve the existing SEA's for the Host to figure out differences
        dbvioses = db_api.vios_find_all(context, filters, dbtxn)
        # Loop through each of the existing VIOS's, marking them as inactive
        for dbvios in dbvioses:
            sea_filters = dict(vios_id=dbvios['id'])
            # Update the RMC State to inactive in the DB for the VIOS
            db_api.vios_update(context, dbvios['id'], vios_values, dbtxn)
            # Update the State to inactive in the DB for each SEA on the VIOS
            dbseas = db_api.sea_find_all(context, sea_filters, dbtxn)
            for dbsea in dbseas:
                db_api.sea_update(context, dbsea['id'], sea_values, dbtxn)

    def _update_network_associations(self, context, transaction, net_assocs):
        """Helper method to update the Network Association SEA's in the DB"""
        hostfact = ManagedHostFactory.get_factory()
        # If there weren't any Network Associations, then nothing to do
        if not net_assocs:
            return
        # Query the Host so we can figure out the SEA Chain for it
        host = hostfact.find_host_by_name(context, self.host_name, transaction)
        # Force a building of the SEA Chain once, so we don't each time
        host.find_sea_chains(rebuild=True)
        # Loop through each Network Association, updating the SEA if needed
        for net_assoc in net_assocs:
            # Continue on if there wasn't an SEA set on the Network Association
            if net_assoc.sea is None:
                continue
            # Find the highest priority SEA that should be the one used
            sea_chain = host.find_sea_chain_by_pvea_info(net_assoc.sea)
            sea_to_use = (sea_chain[0] if sea_chain else None)
            # Update the SEA on the Network Association if it is different
            if sea_to_use is not None and sea_to_use.id != net_assoc.sea.id:
                net_assoc.sea = sea_to_use
                net_assoc.save()

    def _handle_deleted_na_seas(self, context, transaction,
                                net_assocs, deleted_seas):
        """Helper method to update Network Association when an SEA's deleted"""
        net_assoc_map = dict()
        hostfact = ManagedHostFactory.get_factory()
        # Query the Host so we can figure out the SEA Chain for it
        host = hostfact.find_host_by_name(
            context, self.host_name, transaction)
        # Update the SEA list on the host to remove the deleted ones
        for vios in host.vioservers:
            vios.seas[:] = [s for s in vios.seas if s.id not in deleted_seas]
        host.find_sea_chains(rebuild=True)
        # Build a Map of the Network Associations based on the SEA for them
        for net_assoc in net_assocs:
            if net_assoc.sea_id is not None:
                na_list = net_assoc_map.get(net_assoc.sea_id, list())
                na_list.append(net_assoc)
                net_assoc_map[net_assoc.sea_id] = na_list
        # Now that we have the Deleted List, loop through and do the delete
        for sea_id in deleted_seas:
            # If there are any Network Associations for this SEA, update them
            if net_assoc_map.get(sea_id):
                old_sea, new_sea = (net_assoc_map[sea_id][0].sea, None)
                sea_chain = host.find_sea_chain_by_pvea_info(old_sea)
                # Find the first SEA in the Chain that isn't being deleted
                for sea in sea_chain:
                    if sea.id not in deleted_seas:
                        new_sea = sea
                        break
                for net_assoc in net_assoc_map[sea_id]:
                    net_assoc.sea = new_sea
                    net_assoc.save()

    def _cleanup_unused_ports(self, context):
        """Cleans up any unused Network Ports from the Neutron Database"""
        global USED_PORT_COUNT
        USED_PORT_COUNT = USED_PORT_COUNT + 1
        # We only want to run the used port clean up rarely, as it is expensive
        if USED_PORT_COUNT >= cfg.CONF.used_port_threshold:
            USED_PORT_COUNT = 0
            # Call the Network Factory to cleanup the Ports from Neutron
            netfact = NetworkAssociationFactory.get_factory()
            netfact.cleanup_unused_ports(context)


@base.NovaObjectRegistry.register
class Network(network_obj.Network):
    """Production DOM Resource implementation for Managed Hosts"""

    obj_extra_fields = ['segmentation_id', 'physical_network']

    def __init__(self, *args, **kwargs):
        "Constructs an instance of the given DOM Resource Class"
        super(Network, self).__init__(*args, **kwargs)
        self._segmentation_id, self._physical_network = None, None

    ###############################################
    #         Defining Network Interface          #
    ###############################################
    @property
    def segmentation_id(self):
        """Getter Method for the Segmentation ID for this Network"""
        return self._segmentation_id

    @segmentation_id.setter
    def segmentation_id(self, segmentation_id):
        """Setter Method for the Segmentation ID for this Network"""
        self._segmentation_id = segmentation_id

    @property
    def physical_network(self):
        """Getter Method for the Physical Network for this Network"""
        return self._physical_network

    @physical_network.setter
    def physical_network(self, physical_network):
        """Setter Method for the Physical Network for this Network"""
        self._physical_network = physical_network


######################################################
#    Network DOM Resource Factory Registration       #
######################################################
dom.ResourceFactoryLocator.register_factory(VioServerFactory())
dom.ResourceFactoryLocator.register_factory(ManagedHostFactory())
dom.ResourceFactoryLocator.register_factory(ManagedInstanceFactory())
dom.ResourceFactoryLocator.register_factory(NetworkAssociationFactory())
dom.ResourceFactoryLocator.register_factory(HostNetworkTopologyFactory())
dom.ResourceFactoryLocator.register_factory(SharedEthernetAdapterFactory())
dom.ResourceFactoryLocator.register_factory(VirtualEthernetAdapterFactory())
