# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import copy
import json
import six

from collections import namedtuple
from datetime import datetime

from nova.ibm.ego import ego_constants as ego_consts
from powervc_nova.common import aggregate_utils as agg_utils
from powervc_nova.db import api as db_api
from powervc_nova.prs import constants as consts
from powervc_nova import _

from oslo_log import log as logging
LOG = logging.getLogger(__name__)

RULES_KEY = consts.DRO_SCHEDULES_RULES_KEY
SCHED_ENABLED_KEY = consts.DRO_SCHEDULES_ENABLED_KEY


def _load_rules_as_datetimes(dro_function_key, rules):
    """
    Helper method to validate the rules from raw DRO schedules DB values. Also
    returns a new list that has the time values converted to datetime.time
    objects instead of UTC strings. The method issues an exception if the
    contents of the rules are malformed. (out of range, bad time strings, etc).

    :param dro_function_key: Key name for the DRO schedule function that the
                             rules came from. Used to make error messages.
    :param rules: Raw 'rules' list from the DRO schedules DB
    :return new list with datetime.time objects instead of strings OR exception
    """
    # Check if each rule is valid
    valid_rules = []
    for rule in rules:
        rule_error_prepend = (_("For DRO schedule function '%(func)s', "
                                "rule '%(rule)s' is not valid") % dict(
                                    func=dro_function_key, rule=rule))

        # Load and verify data for the day integer keys
        for day_key in ['start_day', 'end_day']:
            try:
                # Confirm value is an integer between 1 and 7
                if not (isinstance(rule[day_key], int) and
                        1 <= rule[day_key] <= 7):
                    msg = (_("%(prepend)s. Key '%(key)s' must be an "
                             "integer between [1, 7] but found "
                             "'%(value)s'") % dict(
                                 key=day_key, value=rule[day_key],
                                 prepend=rule_error_prepend))
                    raise ValueError(msg)
            except KeyError as e:
                msg = (_("%(prepend)s. Failed to find key '%(key)s'") %
                       dict(prepend=rule_error_prepend, key=day_key))
                raise KeyError(msg)

        # Load and verify data for the time keys
        for time_key in ['start_time', 'end_time']:
            try:
                time_str = datetime.strptime(
                    rule[time_key],
                    consts.DRO_SCHEDULES_TIME_MASK)
                rule[time_key] = time_str.time()

            except KeyError as e:
                msg = (_("%(prepend)s. Failed to find key '%(key)s'") %
                       dict(prepend=rule_error_prepend, key=time_key))
                raise KeyError(msg)
            except ValueError as e:
                msg = (_("%(prepend)s. Key '%(key)s' %(e)s or is an "
                         "invalid time") % dict(prepend=rule_error_prepend,
                                                key=time_key,
                                                e=six.text_type(e)))
                raise ValueError(msg)

        # Rules with the same start and end would never be enabled so
        # we don't allow it
        if (rule.get('start_day') == rule.get('end_day') and
                rule.get('start_time') == rule.get('end_time')):
            msg = (_("%(prepend)s. Rule has the same start and end which "
                     "means the rule would have no effect")
                   % dict(prepend=rule_error_prepend))
            raise ValueError(msg)

        valid_rules.append(rule)
    return valid_rules


def _send_dro_scheduler_event(context, policy_engine_manager, host_group,
                              event_type, rule):
    """
    Sends a DRO event notification for DRO scheduler events. The DRO
    schedule rule that triggered the event is also passed. This will be None
    if no rule applies. This occurs when DRO scheduling is disabling DRO.

    :param context: nova context object
    :param policy_engine_manager: manager to store the events
    :param host_group: nova aggregate to which this event applies
    :param event_type: type of event to be generated
    :param rule (optional) DRO schedule rule that triggered this event
    :return: a tuple comprised of (the number of cores that should be taken
             from the pool, the number of cores that need to be reclaimed
             from other hosts)
    """
    serializable_rule = None

    # rule will be empty when scheduler disables DRO
    if rule:
        # datetime objects are not serializable which breaks
        # store_event(). The following makes a new copy of the
        # line and converts the datetimes back into strings
        serializable_rule = rule.copy()
        for key in ['start_time', 'end_time']:
            time_str = serializable_rule[key].strftime(
                consts.DRO_SCHEDULES_TIME_MASK)
            serializable_rule[key] = time_str

    meta = host_group.metadata
    payload = {
        'aggregate_id': host_group.id,
        'schedule_rule': serializable_rule,
        consts.DRO_ENABLED_KEY: meta.get(consts.DRO_ENABLED_KEY),
        'action_name': meta.get(ego_consts.RUNTIME_POLICY_ACTION),
        'rtpolicy_id': meta.get(ego_consts.RUNTIME_POLICY_ID),
        'threshold': meta.get(ego_consts.RUNTIME_POLICY_THRESHOLD),
        'stabilization': meta.get(ego_consts.RUNTIME_POLICY_STABILIZATION),
        'run_interval': meta.get(ego_consts.RUNTIME_POLICY_RUN_INTERVAL)}

    policy_engine_manager.store_event(context, event_type, payload)


def is_currently_scheduled(hg_name, schedule):
    """
    Checks the DRO schedule dictionary against the current time to
    determine whether or not DRO should be enabled. The passed in DRO schedule
    must already have its times converted from strings to datetime objects.
    The method returns True based on the first rule that contains the
    current time. If no rule matches, False is returned. The matching rule is
    also returned or None on False.

    :param hg_name: host group's name that the schedule came from
    :param schedule: host group's schedule as a dictionary

    :return: boolean, rule
             True and the first matching rule OR
             False and None
    """
    if not schedule:
        return False, None

    # Get current time information in UTC
    now = datetime.utcnow()
    now_day = now.isoweekday()
    now_time = now.time()

    # Checks each time_range and return True if contains the current time
    for time_range in schedule.get(RULES_KEY):
        start_day = time_range.get('start_day')
        end_day = time_range.get('end_day')
        start_time = time_range.get('start_time')
        end_time = time_range.get('end_time')

        # Special check when start/stop wraps around same day
        # Effectively the whole week is scheduled except for a few hours
        if start_day == end_day and start_time > end_time:
            # Continue if we are in the small excluded time window
            if now_day == start_day and start_time > now_time >= end_time:
                continue

            return True, time_range

        # Checks if now_day falls within this time_range's start/stop day
        # Elif logic handles start/stop day wrapping around the end of the week
        elif ((start_day <= end_day and start_day <= now_day <= end_day) or
              (start_day > end_day and (
                now_day >= start_day or now_day <= end_day))):

            # Stop checking time_range if we are before the start_time
            if now_day == start_day and now_time < start_time:
                continue
            # Stop checking time_range if we are at or after the end_time
            if now_day == end_day and now_time >= end_time:
                continue

            return True, time_range

    # None of the time ranges contained the current time
    return False, None


def load_dro_schedules(db_dro_schedules):
    """
    Loads the DRO scheduler dict from the given DB results. The string
    representations of times are converted back into datetime objects.
    The method issues an exception if the contents are malformed
    (int out of range, bad time strings, etc).

    :param db_dro_schedules: Raw results from the DRO schedules DB
    :return new DRO schedules dict with objects instead of strings OR exception
    """
    loaded_scheds = copy.deepcopy(db_dro_schedules)
    for dro_function_key in loaded_scheds:
        if dro_function_key not in consts.DRO_FUNCTIONS:
            msg = (_("Could not find function '%(func)s' in list of DRO "
                     "functions %(dro_funcs)s") % dict(
                         func=dro_function_key,
                         dro_funcs=consts.DRO_FUNCTIONS))
            raise ValueError(msg)
        func_dict = loaded_scheds[dro_function_key]

        # Check if schedule enablement key is valid
        try:
            schedule_enabled = func_dict[SCHED_ENABLED_KEY]
            if not isinstance(schedule_enabled, bool):
                msg = (_("For DRO schedule function '%(func)s' key '%(key)s' "
                         "must be a boolean") % dict(
                             func=dro_function_key, key=SCHED_ENABLED_KEY))
                raise ValueError(msg)
        except KeyError:
            msg = (_("For DRO schedule function '%(func)s' key '%(key)s' "
                     "could not be found") % dict(func=dro_function_key,
                                                  key=SCHED_ENABLED_KEY))
            raise KeyError(msg)

        # Check if schedule rules key is present and contains a list
        try:
            rules = func_dict[RULES_KEY]
            if not isinstance(func_dict[RULES_KEY], list):
                msg = (_("For DRO schedule function '%(func)s' key '%(key)s' "
                         "must be a list") % dict(func=dro_function_key,
                                                  key=RULES_KEY))
                raise ValueError(msg)
        except KeyError:
            msg = (_("For DRO schedule function '%(func)s' key '%(key)s' "
                     "could not be found") % dict(func=dro_function_key,
                                                  key=RULES_KEY))
            raise KeyError(msg)

        # Check that when the schedule is enabled, there is at least one rule
        if schedule_enabled and not rules:
            msg = (_("For DRO schedule function '%(func)s' list '%(rules)s' "
                     "must contain at least one rule when setting "
                     "'%(enabled)s' to 'True'") % dict(
                         func=dro_function_key, rules=RULES_KEY,
                         enabled=SCHED_ENABLED_KEY))
            raise ValueError(msg)

        rules = _load_rules_as_datetimes(dro_function_key, rules)
    return loaded_scheds


def _check_and_update_host_group(context, policy_engine_manager, host_group,
                                 hg_dro_schedules):
    """
    Helper method that checks whether or not the host group needs to be
    updated based on the provided DRO schedules. If an update is necessary,
    the host group is updated and the appropriate DRO event is sent.

    :param context: nova context object
    :param policy_engine_manager: manager to send the DRO event if necessary
    :param host_group: nova aggregate to check
    :param hg_dro_schedules: schedules for the host group to check against.
           These schedules must already have their times converted to datetime
           objects.
    """
    # Tuples to map each DRO function to the correct DRO schedule key and
    # DRO Event types
    KeyMap = namedtuple(
        'KeyMap', 'func_enabled_key sched_key start_event end_event')
    host_opt_map = KeyMap(consts.DRO_ENABLED_KEY,
                          consts.DRO_FUNCTION_HOST_OPTIMIZATION,
                          consts.DRO_SCHED_HOST_OPTIMIZATION_START,
                          consts.DRO_SCHED_HOST_OPTIMIZATION_END)
    key_maps = [host_opt_map]

    # Checks each DRO function's state against the scheduled state and
    # updates it if needed
    for key_map in key_maps:
        function_state = host_group.metadata.get(key_map.func_enabled_key)
        if not function_state:
            LOG.warn(_("Host group %(hg)s is missing metadata key "
                       "'%(func_key)s'; skipping DRO schedule check "
                       "for '%(sched_key)s'.") %
                     dict(hg=host_group.name,
                          func_key=key_map.func_enabled_key,
                          sched_key=key_map.sched_key))
            continue
        function_state = function_state.lower() == 'true'
        schedule = hg_dro_schedules[key_map.sched_key]
        schedule_enabled = schedule[SCHED_ENABLED_KEY]

        if schedule.get(RULES_KEY) and not schedule_enabled:
            LOG.debug("Found DRO scheduling rules for function "
                      "'%(func)s' on host group %(hg)s but DRO "
                      "scheduling is not enabled" %
                      dict(func=key_map.func_enabled_key, hg=host_group.name))

        if schedule_enabled:
            scheduled_state, rule = is_currently_scheduled(host_group.name,
                                                           schedule)
            # Update DRO function enablement based on scheduler results
            if function_state != scheduled_state:
                # Logging info
                now = datetime.utcnow()
                now_str = (_('day "%(day)s" "%(time)s"') %
                           dict(day=now.isoweekday(),
                                time=now.strftime(
                                    consts.DRO_SCHEDULES_TIME_MASK)))
                LOG.info((_('For host group %(id)s at %(now)s DRO schedule '
                            'for %(sched_key)s said key %(key)s should be '
                            '"%(sched)s" based on schedule rule %(rule)s. '
                            '%(key)s will be updated to "%(sched)s"') %
                          dict(id=host_group.id,
                               now=now_str,
                               key=key_map.func_enabled_key,
                               sched_key=key_map.sched_key,
                               sched=scheduled_state,
                               rule=(rule or '<no rules matched>'))))

                host_group.update_metadata(
                    {key_map.func_enabled_key: str(scheduled_state)})

                # Set the event type
                if scheduled_state:
                    event_type = key_map.start_event
                else:
                    event_type = key_map.end_event

                _send_dro_scheduler_event(context, policy_engine_manager,
                                          host_group, event_type, rule)


def update_dro_based_on_schedules(context, policy_engine_manager):
    """
    Method checks the current DRO schedules for every host group. If needed,
    a host group's DRO settings will be updated based on the DRO schedules
    and a DRO event is sent. In the event of a bad DRO schedule an error is
    sent to the log and that host group's DRO settings will not be altered.
    Other host group's schedules will continue to be checked. This method is
    meant to be called as part of a periodic task but may be called directly
    as well.

    :param context: nova context object
    :param policy_engine_manager: manager to send the DRO event if necessary
    """
    LOG.debug("Checking all host groups' DRO schedules")
    try:
        all_hgs_scheds = db_api.dro_schedules_read_all_schedules(context)
        if not isinstance(all_hgs_scheds, dict):
            msg = _("Expected a dict of DRO schedules")
            raise TypeError(msg)
    except Exception as e:
        LOG.error(_("Failed to read DRO schedules: %s" % six.text_type(e)))
        return

    if not all_hgs_scheds:
        LOG.debug("Skipped checking DRO schedules because there were "
                  "no host groups with schedules")
        return

    all_hgs_by_id = {agg.id: agg
                     for agg in agg_utils.get_aggregate_list(context)}

    # Check the DRO schedules for each host group that has schedules
    for hg_id in all_hgs_scheds:
        hg = all_hgs_by_id.get(hg_id)
        hg_dro_scheds = all_hgs_scheds.get(hg_id)

        if not hg or not hg_dro_scheds:
            continue

        # Load and verify the host group's schedules
        try:
            schedules = load_dro_schedules(hg_dro_scheds)
        except Exception as e:
            LOG.error((_('Skipping check of DRO schedules for host group '
                         '"%(name)s" because loading the schedules failed. '
                         '%(e)s.') % dict(name=hg.name, e=six.text_type(e))))
            continue

        # If needed, update the host group and send DRO event
        _check_and_update_host_group(
            context, policy_engine_manager, hg, schedules)
