# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from nova import db
from nova import objects
from nova import servicegroup

from nova.compute import task_states
from nova.compute import vm_states
from oslo_config import cfg
from oslo_log import log as logging

from nova.ibm.ego.hapolicy import constants as ha_const
from nova.ibm.ego.hapolicy.hypervisor_driver import driver

from powervc_nova.common import aggregate_utils as agg_utils
from powervc_nova.prs import utils as prs_utils
from powervc_nova.virt.ibmpowervm.common import constants as pvmc
from powervc_nova import _

LOG = logging.getLogger(__name__)

cfg.CONF.register_opts([
    cfg.BoolOpt('allow_maintenance_while_vms_relocating',
                default=False,
                help='Determines whether to allow a host to enter maintenance '
                'mode if there are virtual machines currently being migrated '
                'or rebuilt that are located on hosts in any of the same '
                'host groups that the requested host belongs to. Specifying '
                'True to allow this could cause a virtual machine being '
                'migrated or rebuilt at the time of the maintenance request '
                'to get placed on the host after maintenance mode is enabled '
                'for it.')])

VM_RELOCATION_TASKS = [task_states.REBUILDING,
                       task_states.REBUILD_BLOCK_DEVICE_MAPPING,
                       task_states.REBUILD_SPAWNING,
                       task_states.MIGRATING]


class PowerVCHAHypervisorDriver(driver.HAHypervisorDriver):
    """PowerVC's HA service driver class to extend PRS base logic."""

    def hypervisor_is_up(self, context, hv_name, **kwargs):
        """
        Determine if a hypervisor is up. If the hypervisor is deemed down, the
        HA service orchestrator will begin its process of evacuating the host.

        :param context: nova context
        :param hv_name: name of a hypervisor
        :return: True if the hypervisor is deemed up and False otherwise
        """
        node = prs_utils.get_compute_node(context, hv_name)

        if node is None:
            LOG.error(_("Could not locate compute node for hypervisor %s; "
                        "assuming the hypervisor is down.")
                      % hv_name)
            return False

        # This block simply determines whether Nova and the hypervisor stats
        # assert that the host is active.
        hypervisor_type = node.hypervisor_type
        service = objects.Service.get_by_compute_host(context, node.host)

        src = servicegroup.API().service_is_up(service)
        stats = node.stats or {}
        if hypervisor_type in pvmc.IBM_POWERVM_HYPERVISOR_TYPES:
            hyper_state = stats.get(pvmc.HYPERVISOR_STATE)
            # We need to say the hypervisor is up when the state is set to
            # "not master - " in order to put the host in maintenance mode
            not_master = hyper_state and hyper_state.startswith('not master')
            ret = src and (hyper_state == 'operating' or not_master)
        else:
            ret = src

        # Now we need to reconcile with what the user has explicitly told us.
        host_dead = str(stats.get(pvmc.HOST_DEAD, False)).lower() == 'true'
        if not kwargs.get('for_maintenance', False):
            # If we're called in the context of the PRS HA engine to
            # determine whether HA actions (e.g., rebuilds) should be
            # taken, then we need to ensure the host has been explicitly
            # marked as down before we start rebuilding the VMs.
            if not host_dead:
                ret = True
        else:
            # The 'for_maintenance' bit will be set for cases in which we
            # determining whether maintenance mode can be enabled.  We need
            # to ensure the services are operating properly *and* that the
            # user hasn't explicitly told us the host is dead.
            ret = ret and not host_dead
        return ret

    def get_target_service_host(self, context, ha_action, hv_name):
        """
        Provides an extension point for exploiters to specify a target host
        for the specified HA action. The return value should be a valid
        service host that will be ultimately passed to the 'force_hosts'
        scheduling hint.

        :param context: nova context
        :param ha_action: one of live-migration|cold-migration|rebuild
        :param hv_name: hypervisor hostname of the source
        :return: the service host to which the operation should be targeted or
                 None if the scheduler should select
        """
        node = prs_utils.get_compute_node(context, hv_name)

        if node is None:
            LOG.error(_("Error retrieving the target service host for %s "
                        "because its compute node couldn't be found.")
                      % hv_name)
            return None
        elif ha_action != ha_const.HA_ACTION_REBUILD:
            LOG.error(_("Error retrieving the target service host for %(hv)s "
                        "because the HA action (%(act)s) is not '%(exp)s'.")
                      % {'hv': hv_name, 'act': ha_action,
                         'exp': ha_const.HA_ACTION_REBUILD})
            return None

        stats = node.stats or {}
        target_host = stats.get(pvmc.REMOTE_RESTART_TARGET_HOST)
        if target_host is None:
            LOG.debug("No target host was specified for the automated "
                      "rebuild operation.")
        else:
            LOG.info(_("Target host '%s' was specified for the automated "
                       "rebuild operation.") % target_host)
        return target_host

    def _include_node(self, context, node):
        """Determines whether the hypervisor is a candidate for HA tasks."""
        hyper_is_up = self.hypervisor_is_up(context,
                                            node.hypervisor_hostname)

        if hyper_is_up:
            # We don't support HA logic on hypervisors that are active; this is
            # reserved for when predictive failure alerts are available.
            return False

        if node.hypervisor_type in pvmc.IBM_POWERVM_HYPERVISOR_TYPES:
            # If the hypervisor is PowerVM, then we need to make sure it
            # supports the "simplified remote restart" capability, otherwise
            # it is not eligible to participate in remote restart operations.
            stats = node.stats or {}
            if not stats.get(pvmc.PVM_SIMP_REMOTE_RESTART_CAPABLE, False):
                LOG.warning(_("PowerVM hypervisor '%s' was deemed down, but "
                              "will not be considered for automated rebuild "
                              "operations because it does not support the "
                              "simplified remote restart capability.")
                            % node.hypervisor_hostname)
                return False

        # At this point, the hypervisor is going to be selected by the HA
        # orchestrator and its VMs will start to be evacuated.
        return True

    def select_hypervisors(self, context, aggregate_id=None):
        """
        Determine which hosts are deemed inactive and potential candidates
        for HA (i.e., to be rebuilt on "healthy" hosts).

        :param context: nova context
        :param aggregate_id: ID of aggregate; None if no aggregate
        :return: [hypervisor_hostname]
        """
        if not aggregate_id:
            LOG.error(_("The select_hypervisors routine was erroneously "
                        "called without an aggregate identifier!"))
            # If this happens, assume that no hosts are down to avoid any
            # erroneous rebuild operations.
            return []

        svc_hosts = db.aggregate_host_get_all(context, aggregate_id)
        hv_hostnames = [node.hypervisor_hostname
                        for node in objects.ComputeNodeList.get_all(context)
                        if node.host in svc_hosts and
                        self._include_node(context, node)]
        return hv_hostnames

    def can_enter_maintenance(self, context, hv_name, migrate_action):
        """
        Determines whether this host can enter maintenance mode.  This logic is
        above and beyond the basic checks that PRS already does (e.g., ensuring
        that you can't put a host into maintenance that's already in
        maintenance mode).

        :param context: A context object that is used to authorize DB access
        :param hv_name: The hypervisor hostname of the host that is trying to
                        be put into maintenance mode
        :param migrate_action: migrate option used when maintenance mode is
                               enabled; it can be one of 'none', 'all' or
                               'active-only'
        :returns: (True, ) if maintenance is allowed or (False, <string>) where
                  <string> represents an error message to be sent back to the
                  operator with some hints as to why the operation is not
                  allowed
        """
        if migrate_action == 'all':
            return (False, _("Inactive migrations are not supported; as "
                             "such, the 'all' migration action is not "
                             "allowed at this time."))
        elif migrate_action == 'none':
            # The remaining logic below checks the hypervisor's preparedness to
            # perform live migrations; if the user isn't requesting any
            # migration logic, we can simply allow the operation immediately.
            return (True, )

        compute_node = prs_utils.get_compute_node(context, hv_name)
        if compute_node is None:
            LOG.error(_("The compute node for hypervisor (%s) could not be "
                        "found!") % hv_name)
            return (False, _("An unexpected error occurred; please try "
                             "restarting the PowerVC services and retry "
                             "your request or contact your service "
                             "representative if this problem persists."))

        stats = compute_node.stats or {}
        # Check to ensure the hypervisor is operating normally.
        hypervisor_state = stats.get(pvmc.HYPERVISOR_STATE)
        if hypervisor_state != 'operating':
            return (False, _("The hypervisor is not operating normally right "
                             "now. Its current state is '%s'; remedy this "
                             "problem and try again.") % hypervisor_state)

        lpm_capable = prs_utils.can_host_live_migrate(context, compute_node)
        if not lpm_capable['allowed']:
            # If not allowed, get the reason code.
            reason_code = lpm_capable['reason_code']
            if reason_code == prs_utils.NO_LPM_LICENSE:
                return (False, _("The hypervisor is not licensed for live "
                                 "partition mobility (LPM) operations."))
            elif reason_code == prs_utils.KVM_LOCAL_STORAGE:
                return (False, _("The hypervisor is not configured with "
                                 "shared storage, so its virtual machines "
                                 "cannot be migrated."))
            else:
                # Generic error message.
                return (False, _("The hypervisor is not capable of "
                                 "performing live migration operations."))

        host = compute_node.host
        instances = objects.InstanceList.get_by_host(
            context, host, expected_attrs=['uuid', 'vm_state',
                                           'display_name', 'power_specs,'
                                           'system_metadata'])

        for inst in instances:
            vm_state = inst.vm_state
            uuid = inst.uuid
            disp_name = inst.display_name
            if vm_state == vm_states.ERROR:
                # If any instances are in ERROR state, don't allow maintenance
                # mode as it's highly likely the migration(s) will not work.
                return (False, _("There is at least one virtual machine, "
                                 "%(disp_name)s (%(uuid)s), in the error "
                                 "state; remedy this problem and try "
                                 "again.") % {'uuid': uuid,
                                              'disp_name': disp_name})
            elif vm_state == vm_states.PAUSED:
                # Since a paused VM is still retained in memory, we want to
                # avoid maintenance mode as it's likely the administrator
                # needs to power down the host.
                return (False, _("There is at least one virtual machine, "
                                 "%(disp_name)s (%(uuid)s), in the paused "
                                 "state; resume the virtual machine so that "
                                 "it can live migrated and try again.") %
                        {'uuid': uuid, 'disp_name': disp_name})
            elif vm_state == vm_states.ACTIVE:
                lpm_capable = prs_utils.can_instance_live_migrate(context,
                                                                  inst,
                                                                  compute_node)
                if not lpm_capable['allowed']:
                    # If not allowed, get reason.
                    reason_code = lpm_capable['reason_code']
                    extra_data = lpm_capable['extra_data']
                    if reason_code == prs_utils.RMC_NOT_ACTIVE:
                        # If any instances' RMC state is inactive, don't allow
                        # maintenance mode as live migrations will definitely
                        # fail. This check is only applicable to the PowerVM
                        # hypervisor.
                        return (False, _("There is at least one active "
                                         "virtual machine, %(disp_name)s "
                                         "(%(uuid)s), whose RMC state is "
                                         "inactive and any live migration "
                                         "operation would fail. Its current "
                                         "RMC state is %(rmc_state)s; remedy "
                                         "this problem and try again.") %
                                {'uuid': uuid, 'disp_name': disp_name,
                                 'rmc_state': extra_data['rmc_state']})
                    elif reason_code == prs_utils.COLLOCATION_AFFINITY:
                        r = extra_data['unsupported_affinity_groups']
                        return (False, _("The host has at least one active "
                                         "virtual machine, %(disp_name)s "
                                         "(%(uuid)s), that belongs to the "
                                         "following affinity collocation "
                                         "rules: %(rules)s. Therefore, a "
                                         "migration would fail for this "
                                         "virtual machine. To continue with "
                                         "the migration, remove this virtual "
                                         "machine from the affinity "
                                         "collocation rules and try again. "
                                         "Alternatively, place the host into "
                                         "maintenance mode without migrating "
                                         "the virtual machines.")
                                % {'uuid': uuid, 'disp_name': disp_name,
                                   'rules': r})
                    elif reason_code == prs_utils.COLLOCATION_ONGOING_OPS:
                        r = extra_data['unsupported_groups_with_parallel_ops']
                        return (False, _("The host has at least one active "
                                         "virtual machine, %(disp_name)s "
                                         "(%(uuid)s), that belongs to the "
                                         "following collocation rules: "
                                         "%(rules)s. These rules contain "
                                         "another virtual machine that is "
                                         "undergoing an operation that might "
                                         "cause it to change hosts and "
                                         "violate the rule. Wait until "
                                         "ongoing operations finish for "
                                         "virtual machines in these "
                                         "collocation rules and try again.")
                                % {'uuid': uuid, 'disp_name': disp_name,
                                   'rules': r})

        if not cfg.CONF.allow_maintenance_while_vms_relocating:
            # Do not allow maintenance if any virtual machine belonging to
            # a host in any of the same host aggregates as the requested host
            # are currently migrating or relocating; doing so could cause
            # the in-flight VM to end up on this host after maintenance
            # mode is enabled.
            agg_ids = agg_utils.get_aggregate_ids_for_host(context, host)

            # Get hosts in all the aggregates this host belongs to.
            all_aggs_hosts = set()
            for agg_id in agg_ids:
                hosts = agg_utils.get_hosts_for_aggregate(context, agg_id)
                all_aggs_hosts.update(hosts)

            # Remove the requested host since we don't care about VMs
            # migrating away from it, only VMs that could view it as a
            # target host.
            all_aggs_hosts.remove(host)
            for host_name in all_aggs_hosts:
                instances = objects.InstanceList.get_by_host(
                    context, host_name,
                    expected_attrs=['uuid', 'task_state', 'display_name'])
                for instance in instances:
                    if instance.task_state in VM_RELOCATION_TASKS:
                        return (False, _("There is at least one migrating or "
                                         "rebuilding virtual machine, "
                                         "%(disp_name)s (%(uuid)s), in the "
                                         "same host group as the requested "
                                         "maintenance host. Wait for this "
                                         "operation to finish and try "
                                         "again.") %
                                {'uuid': instance.uuid,
                                 'disp_name': instance.display_name})

        return (True, )
