# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from oslo_log import log as logging
from oslo_serialization import jsonutils
from datetime import datetime, timedelta
from eventlet import greenthread
from time import sleep

from nova import context as ctxt
from nova import exception
from nova import objects
from nova.ibm.ego.rtpolicy import rtpolicy_driver as driver
from nova.scheduler.ibm.ego import ego_rpcapi

from powervc_nova.db import api as db_api
from powervc_nova.prs import constants as pvc_prs_const
from powervc_nova.prs import utils as prs_utils
from powervc_nova.prs import rtpolicy_sort
from powervc_nova.virt.ibmpowervm.common import constants as pvmc
from powervc_nova import _

from oslo_config import cfg
from oslo_utils import timeutils

LOG = logging.getLogger(__name__)

cfg.CONF.register_opts([
    cfg.BoolOpt('dro_bypass_non_default_spp',
                default=True,
                help='Determines whether DRO will automatically bypass '
                'virtual machines in a non-default shared processor pool. '
                'This only applies to virtual machines on the PowerVM '
                'hypervisor.'),
    cfg.IntOpt('dro_events_ttl',
               default=14,
               help='The maximum number of days that DRO events '
               'should be persisted.'),
    cfg.IntOpt('dro_events_prune_interval',
               default=300,
               help='The interval (in seconds) at which DRO events are '
               'pruned.')])

PRUNING_ENABLED = False


class PowerVCRTPolicyDriver(driver.RTPolicyDriver):
    """PowerVC's RTPolicy driver class to extend PRS base logic."""

    def __init__(self, *args, **kwargs):
        super(PowerVCRTPolicyDriver, self).__init__(*args, **kwargs)
        self._scheduler_api = ego_rpcapi.EGOSchedulerAPI()
        self._enable_pruning_task()

    def get_instance_cpu_util(self, instance_obj):
        """
        Retrieves the effective CPU utilization percentage for the specified
        instance object.

        :param instance_obj: instance object (i.e., objects.Instance)
        :return: a floating point representing the instance's CPU utilization
                 percentage (e.g., 20.4% would be returned as 20.4)
        """
        cpu_util = instance_obj.power_specs.get('cpu_util')
        if cpu_util is None:
            LOG.error(_("Virtual machine %(name)s (%(uuid)s) does not have "
                        "its 'cpu_util' power_specs entry defined; "
                        "assuming 0.0 for now.")
                      % {'name': instance_obj.display_name,
                         'uuid': instance_obj.uuid})
            cpu_util = 0.0
        return float(cpu_util)

    def get_instance_memory_util(self, instance_obj):
        """
        Retrieves the effective memory utilization percentage for the specified
        instance object.

        :param instance_obj: instance object (i.e., objects.Instance)
        :return: a floating point representing the instance's memory
        utilization percentage (e.g., 20.4% would be returned as 20.4)
        """
        # NOTE(jwcroppe): Needed to support memory-based optimization.
        raise NotImplementedError()

    def is_aggregate_optimization_enabled(self, context, agg_id, policy_info):
        """
        Determines whether optimization is enabled for the host aggregate.  If
        optimization is disabled, no optimization-related notifications will be
        generated for the respective aggregate or its child hosts or VMs.

        :param context: nova context object
        :param agg_id: host aggregate id
        :param policy_info: a dict containing all of the policy information,
                            including the aggregate ID, policy id, policy name,
                            policy goal, policy threshold, policy run interval,
                            policy stabilization and policy max migrations.
        :return: True if enabled and False otherwise
        """
        agg_obj = objects.Aggregate()
        try:
            agg = agg_obj.get_by_id(context, agg_id)
        except exception.AggregateNotFound:
            LOG.error(_("Could not locate aggregate (%s).") % agg_id)
            return False
        agg_meta = agg.metadata
        # Determine if optimization is enabled for aggregate; default to False.
        is_enabled = (agg_meta.get(pvc_prs_const.DRO_ENABLED_KEY) or
                      pvc_prs_const.DEFAULT_DRO_ENABLED)
        return is_enabled.lower() != 'false'

    def is_host_optimization_enabled(self, context, hv_name, policy_info):
        """
        Determines whether optimization is enabled for the host.  If
        optimization is disabled, no optimization-related notifications will be
        generated for the respective host or its child VMs.

        :param context: nova context object
        :param hv_name: name of a hypervisor
        :param policy_info: a dict containing all of the policy information,
                            including the aggregate ID, policy id, policy name,
                            policy goal, policy threshold, policy run interval,
                            policy stabilization and policy max migrations.
        :return: True if enabled and False otherwise
        """
        compute_node = prs_utils.get_compute_node(context, hv_name)
        if not compute_node:
            LOG.error(_("Could not retrieve compute node structure (%s).")
                      % hv_name)
            return False
        return prs_utils.get_dro_enabled_for_host(context, compute_node)

    def can_perform_host_optimization(self, context, hv_name, policy_info):
        """
        Determines whether optimization will be performed on the hypervisor.
        If optimization is not permitted, the return value's tuple will include
        a reason (string) with further explanation.

        :param context: nova context object
        :param hv_name: name of a hypervisor
        :param policy_info: a dict containing all of the policy information,
                            including the aggregate ID, policy id, policy name,
                            policy goal, policy threshold, policy run interval,
                            policy stabilization and policy max migrations.
        :return: (boolean, reason_string_if_false)
        """
        compute_node = prs_utils.get_compute_node(context, hv_name)
        if not compute_node:
            LOG.error(_("Could not retrieve compute node structure (%s).")
                      % hv_name)
            return (False, )

        # If we're NOT going to be performing live migration, then just allow
        # the operation to proceed at this point.
        policy_action = policy_info['policy_action']
        if policy_action in [pvc_prs_const.CPU_COD_RT_ACTION,
                             pvc_prs_const.CPU_COD_ADVISE_RT_ACTION]:
            return (True, )

        # Check if host is capable of live migration.
        lpm_capable = prs_utils.can_host_live_migrate(context, compute_node)
        if not lpm_capable['allowed']:
            # If not allowed, get the reason code.
            reason_code = lpm_capable['reason_code']
            if reason_code == prs_utils.NO_LPM_LICENSE:
                return (False, _("Hypervisor %(hv_name)s is not licensed for "
                                 "live partition mobility (LPM) operations.")
                        % {'hv_name': hv_name})
            elif reason_code == prs_utils.KVM_LOCAL_STORAGE:
                return (False, _("Hypervisor %(hv_name)s is not configured "
                                 "with shared storage, so its virtual "
                                 "machines cannot be migrated.") %
                        {'hv_name': hv_name})
            else:
                # Generic error message.
                return (False, _("Hypervisor %(hv_name)s is not capable of "
                                 "performing live migration operations.") %
                        {'hv_name': hv_name})

        return (True, )

    def is_instance_optimization_enabled(self, context, instance_obj,
                                         policy_info):
        """
        Determines whether optimization is enabled for the instance.  If
        optimization is disabled, no optimization-related notifications will be
        generated for the respective VM.

        :param context: nova context object
        :param instance_obj: instance object (i.e., objects.Instance)
        :param policy_info: a dict containing all of the policy information,
                            including the aggregate ID, policy id, policy name,
                            policy goal, policy threshold, policy run interval,
                            policy stabilization and policy max migrations.
        :return: True if enabled and False otherwise
        """
        inst_meta = instance_obj.metadata or {}
        is_enabled = inst_meta.get(pvc_prs_const.DRO_ENABLED_KEY) or 'true'
        return is_enabled.lower() != 'false'

    def can_perform_instance_optimization(self, context, instance_obj,
                                          policy_info):
        """
        Determines whether optimization will be performed on the instance.  If
        optimization is not permitted, the return value's tuple will include a
        reason (string) with further explanation.

        :param context: nova context object
        :param instance_obj: instance object (i.e., objects.Instance)
        :param policy_info: a dict containing all of the policy information,
                            including the aggregate ID, policy id, policy name,
                            policy goal, policy threshold, policy run interval,
                            policy stabilization and policy max migrations.
        :return: (boolean, reason_string_if_false)
        """
        uuid = instance_obj.uuid
        disp_name = instance_obj.display_name

        # Get compute node object for this instance
        compute_node = prs_utils.get_compute_node_by_service_host(
            context, instance_obj.host)
        if not compute_node:
            LOG.error(_("Could not retrieve compute node structure (%s).")
                      % instance_obj.host)
            return (False, )

        # Check if instance is capable of live migration
        lpm_capable = prs_utils.can_instance_live_migrate(context,
                                                          instance_obj,
                                                          compute_node)
        if not lpm_capable['allowed']:
            # If not allowed, get reason.
            reason_code = lpm_capable['reason_code']
            extra_data = lpm_capable['extra_data']
            if reason_code == prs_utils.RMC_NOT_ACTIVE:
                return (False, _("Virtual machine, %(disp_name)s (%(uuid)s), "
                                 "has a RMC state of %(rmc_state)s and any "
                                 "live migration operation would fail.") %
                        {'uuid': uuid, 'disp_name': disp_name,
                         'rmc_state': lpm_capable['extra_data']['rmc_state']})
            elif reason_code == prs_utils.COLLOCATION_AFFINITY:
                r = extra_data['unsupported_affinity_groups']
                return (False, _("Virtual machine, %(disp_name)s (%(uuid)s), "
                                 "belongs to the following affinity "
                                 "collocation rules: %(rules)s. Therefore, a "
                                 "migration would fail for this virtual "
                                 "machine.")
                        % {'uuid': uuid, 'disp_name': disp_name, 'rules': r})
            elif reason_code == prs_utils.COLLOCATION_ONGOING_OPS:
                r = extra_data['unsupported_groups_with_parallel_ops']
                return (False, _("Virtual machine, %(disp_name)s (%(uuid)s), "
                                 "belongs to the following collocation rules: "
                                 "%(rules)s. These rules contain another "
                                 "virtual machine that is undergoing an "
                                 "operation that might cause it to change "
                                 "hosts and violate the rule.")
                        % {'uuid': uuid, 'disp_name': disp_name, 'rules': r})

        if compute_node.hypervisor_type in pvmc.IBM_POWERVM_HYPERVISOR_TYPES:
            # If the instance resides in a non-default shared processor pool,
            # we will not perform optimization on it due to the likeliness of
            # impacting software licensing, unless the user has explicitly
            # enabled this capability (applies to PowerVM only).
            power_specs = instance_obj.power_specs or {}
            proc_pool_name = (power_specs.get('shared_proc_pool_name')
                              or pvmc.DEFAULT_PROC_POOL)
            if (proc_pool_name != pvmc.DEFAULT_PROC_POOL and
                    cfg.CONF.dro_bypass_non_default_spp):
                # A non-default pool is being used; bypass this instance.
                proc_pool_id = int(power_specs.get('shared_proc_pool_id') or 0)
                return (False, _("Virtual machine, %(disp_name)s (%(uuid)s), "
                                 "is not in the default shared processor pool "
                                 "and therefore DRO will bypass it so as to "
                                 "not potentially violate software licensing; "
                                 "the virtual machine currently resides in "
                                 "pool %(pool_name)s:%(pool_id)d.") %
                        {'uuid': uuid, 'disp_name': disp_name,
                         'pool_name': proc_pool_name, 'pool_id': proc_pool_id})

        return (True, )

    def should_perform_instance_migration(self, context, instance_obj,
                                          instance_rank, num_instances,
                                          policy_info):
        """
        Determines whether an instance should be migrated during an
        optimization cycle.  By the time this is called, the instance has been
        deemed "OK" to be subjected to migration-based optimization, but this
        provides the ability to determine whether the migration is allowed.
        For example, if an instance isn't even contributing 1% CPU utilization
        to the overall host utilization, it is probably wise to bypass the
        migration.

        It is also implied that if this function returns False for an instance
        whose rank is x, then any instance whose rank is > x must also return
        False (i.e., this means that the migration optimizer will stop trying
        to migrate VMs once this returns False).

        :param context: nova context object
        :param instance_obj: instance object (i.e., objects.Instance)
        :param instance_rank: an integer between [1, n] where `n` represents
                              the number of instances on the host that are
                              subject to migration-based optimizations (note
                              that the instance whose rank is 1 is attempted
                              to be migrated first)
        :param num_instances: the number of instances on the host that are
                              subject to migration-based optimizations
        :param policy_info: a dict containing all of the policy information,
                            including the aggregate ID, policy id, policy name,
                            policy goal, policy threshold, policy run interval,
                            policy stabilization and policy max migrations.
        :returns: (boolean, reason_string_if_false)
        """
        goal = policy_info['policy_goal']
        sort_obj = None
        try:
            sort_obj = rtpolicy_sort.RTPolicyInstanceSort.factory(goal)
        except ValueError:
            raise NotImplementedError()
        return (sort_obj.should_perform_instance_migration(context,
                                                           instance_obj,
                                                           policy_info), )

    def sort_instances_for_optimization(self, context, instances, policy_info):
        """
        Performs an in-place sort of the instances once a violated host is
        selected for optimization.  The default order is based on the ongoing
        optimization policy (e.g., for cpu.util, the instances with the lowest
        CPU utilization will be migrated first).

        :param context: nova context object
        :param instances: a list of instances on the hypervisor that are
                          eligible for optimization (i.e.,
                          can_perform_instance_optimization(...) is True)
        :param policy_info: a dict containing all of the policy information,
                            including the aggregate ID, policy id, policy name,
                            policy goal, policy threshold, policy run interval,
                            policy stabilization and policy max migrations.
        """
        goal = policy_info['policy_goal']
        sort_obj = None
        try:
            sort_obj = rtpolicy_sort.RTPolicyInstanceSort.factory(goal)
        except ValueError:
            raise NotImplementedError()
        sort_obj.sort_instances_for_optimization(context, instances,
                                                 policy_info)

    def handle_notification(self, context, event_type, payload):
        """
        Serves as a notification handler for all notifications that are
        generated by the optimization service.

        :param context: nova context object
        :param event_type: the notification's event type
        :param payload: the notification's payload
        :return: a dict of additional information to add to the payload before
                 it is sent out via the message bus (or `None` if nothing)
        """
        if not (event_type and payload):
            LOG.warning(_("The event type (%(event_type)s) or message payload "
                          "(%(payload)s) were none; ignoring.")
                        % {'event_type': event_type, 'payload': payload})
            return None
        if event_type in pvc_prs_const.DRO_DB_EVENT_TYPES:
            cpu_util = None
            values = dict()
            hypervisor_hostname = payload.get('hypervisor_hostname')
            aggregate_id = payload.get('aggregate_id')

            if event_type in pvc_prs_const.DRO_DB_HYPERVISOR_EVENT_TYPES:
                if not hypervisor_hostname:
                    LOG.warning(_("Ignoring the request because the "
                                  "hypervisor hostname was not specified in  "
                                  "the payload (%(payload)s).")
                                % {'payload': payload})
                    return None

            if event_type in pvc_prs_const.DRO_DB_AGGREGATE_EVENT_TYPES:
                if not aggregate_id:
                    LOG.warning(_("Ignoring the request because the "
                                  "aggregate id was not specified in  "
                                  "the payload (%(payload)s).")
                                % {'payload': payload})
                    return None

            if event_type in pvc_prs_const.DRO_ADD_CPU_UTIL_EVENT_TYPES:
                # Add the CPU utilization to the payload for these events
                compute_node = prs_utils.get_compute_node(context,
                                                          hypervisor_hostname)
                cpu_util = prs_utils.get_cpu_utilization(context,
                                                         compute_node,
                                                         self._scheduler_api)
                # Create a copy of the payload prior to adding the CPU
                # utilization to it (which will be inserted into the database)
                payload = payload.copy()
                payload['dro_hypervisor_cpu_util'] = cpu_util
            tstamp = timeutils.utcnow()
            values.update({
                pvc_prs_const.DRO_EVENT_TABLE_EVENT_TYPE: event_type,
                pvc_prs_const.DRO_EVENT_TABLE_EVENT_TIME: tstamp,
                pvc_prs_const.DRO_EVENT_TABLE_HYPERVISOR_HOSTNAME:
                hypervisor_hostname,
                pvc_prs_const.DRO_EVENT_TABLE_AGGREGATE_ID: aggregate_id,
                pvc_prs_const.DRO_EVENT_TABLE_PAYLOAD: payload})
            result = db_api.dro_events_create(context, values)
            ret = dict(dro_event_id=result['id'],
                       dro_event_time=jsonutils.to_primitive(tstamp))
            if cpu_util is not None:
                # Add the CPU utilization to the notification payload
                ret['dro_hypervisor_cpu_util'] = cpu_util
            return ret
        else:
            return None

    def get_policy_actions(self):
        """
        Returns a list of policy actions that this driver supports.  If a user
        tries to set a policy action to a value not contained within this list,
        the resource optimization process will not run.

        :return: a list of policy actions supported by this driver
        """
        return (list(pvc_prs_const.COD_POLICY_ACTIONS) +
                super(PowerVCRTPolicyDriver, self).get_policy_actions())

    def get_advise_policy_actions(self):
        """
        Returns a list of policy actions that are considered to be "advise
        only", which means the optimizations will not actually be performed.
        The optimizers use this information to determine what action(s) to
        take during their execution.

        :return: a list of policy actions that are considered to be "advise
                 only"
        """
        return (list(pvc_prs_const.COD_ADVISE_POLICY_ACTIONS) +
                super(PowerVCRTPolicyDriver, self).get_advise_policy_actions())

    def _enable_pruning_task(self):
        """Helper method to enable the pruning task for old events."""
        global PRUNING_ENABLED
        # Startup a periodic interval thread to prune old events
        if not PRUNING_ENABLED:
            PRUNING_ENABLED = True
            greenthread.spawn(self._pruning_task)

    def _pruning_task(self):
        """Helper method to prune old events from the database."""
        while cfg.CONF.dro_events_ttl > 0:
            try:
                sleep(cfg.CONF.dro_events_prune_interval)
                self._prune_db_events()
            except Exception as exc:
                LOG.error(_("Error pruning message from the database."))
                LOG.exception(exc)

    def _prune_db_events(self):
        """Helper method to prune DRO events from the database."""
        max_event_time = (datetime.utcnow() -
                          timedelta(days=cfg.CONF.dro_events_ttl))
        db_api.dro_events_prune(ctxt.get_admin_context(),
                                max_event_time)
