# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""PowerVC RTPolicyDriver sort routines."""

from nova import utils as nova_utils

from nova.ibm.ego import ego_constants
from nova.scheduler.ibm.ego import ego_resreq_utils
from nova.scheduler.ibm.ego import ego_rpcapi

from oslo_log import log as logging

from powervc_nova.prs import utils as prs_utils
from powervc_nova.virt.ibmpowervm.common import constants as pvmc
from powervc_nova import _

LOG = logging.getLogger(__name__)


class RTPolicyInstanceSort(object):
    """Base interface for all optimization sorting routines."""

    def __init__(self, *args, **kwargs):
        super(RTPolicyInstanceSort, self).__init__(*args, **kwargs)

    @staticmethod
    def factory(policy_goal):
        if policy_goal == ego_constants.CPU_UTILIZATION_METRICS_NAME:
            return RTPolicyInstanceCPUUtilSort()
        else:
            msg = _("Unexpected policy_goal passed to the factory.")
            raise ValueError(msg)

    def should_perform_instance_migration(self, context, instance_obj,
                                          policy_info):
        """
        Determines whether an instance should be migrated during an
        optimization cycle.

        :param context: nova context object
        :param instance_obj: instance object (i.e., objects.Instance)
        :param policy_info: a dict containing all of the policy information,
                            including the aggregate ID, policy id, policy name,
                            policy goal, policy threshold, policy run interval,
                            policy stabilization and policy max migrations.
        :returns: True if a migration should be performed and False otherwise
        """
        raise NotImplementedError()

    def sort_instances_for_optimization(self, context, instances, policy_info):
        """
        Performs an in-place sort of the instances once a violated host is
        selected for optimization.  The default order is based on the ongoing
        optimization policy (e.g., for cpu.util, the instances with the lowest
        CPU utilization will be migrated first).

        :param context: nova context object
        :param instances: a list of instances on the hypervisor that are
                          eligible for optimization (i.e.,
                          can_perform_instance_optimization(...) is True)
        :param policy_info: a dict containing all of the policy information,
                            including the aggregate ID, policy id, policy name,
                            policy goal, policy threshold, policy run interval,
                            policy stabilization and policy max migrations.
        """
        raise NotImplementedError()


class RTPolicyInstanceCPUUtilSort(RTPolicyInstanceSort):
    """Implements the interface for CPU utilization-based optimizations."""

    def __init__(self, *args, **kwargs):
        super(RTPolicyInstanceCPUUtilSort, self).__init__(*args, **kwargs)
        self._scheduler_api = ego_rpcapi.EGOSchedulerAPI()

    def __get_instance_load(self, instance):
        """Return the CPU utilization for the specified VM."""
        cpu_util = instance.power_specs.get('cpu_util')
        if cpu_util is None:
            LOG.error(_("Virtual machine %(name)s (%(uuid)s) does not have "
                        "its 'cpu_util' power_specs entry defined; "
                        "assuming 0.0 for now.")
                      % {'name': instance.display_name,
                         'uuid': instance.uuid})
            cpu_util = 0.0
        return float(cpu_util)

    def __get_predicted_host_load(self, context, instance):
        """
        Retrieves the predicted CPU utilization of the source host after
        the respective instance migrates off of the source host.  The returned
        result will be a floating point value between [0.0, 1.0]; i.e.,
        a value of 0.50 means the host is predicted to be 50% utilized.

        predicted_load = src_host.cpu_ut -
                             (vm.cpu_ut * vm.vcpus / src_host.vcpus)

        where:
          src_host.cpu_ut is the source host CPU utilization per the metrics,
          vm.cpu_ut is the VM's CPU utilization (1.0 = 100% utilized),
          vm.vcpus is the # of vcpus (proc units on PowerVM) used by the VM,
          src_host.vcpus is the total # of vcpus ("configured" proc units on
                         PowerVM) on the source host.
        """
        def get_hypervisor_type(instance):
            """Retrieve the hypervisor type."""
            sys_meta = nova_utils.instance_sys_meta(instance) or {}
            return sys_meta.get('image_hypervisor_type')

        def get_instance_vcpus(instance):
            """Retrieve the instance's effective vcpus (float)."""
            return ego_resreq_utils.get_instance_vcpus(
                instance, get_hypervisor_type(instance))

        def get_hypervisor_vcpus(node):
            """Retrieves the hypervisor's total vcpus capacity."""
            if node.hypervisor_type in pvmc.IBM_POWERVM_HYPERVISOR_TYPES:
                stats = node.stats or {}
                vcpus = stats.get('proc_units') or node.vcpus
            else:
                vcpus = node.vcpus
            return float(vcpus)

        node = prs_utils.get_compute_node_by_service_host(context,
                                                          instance.host)
        src_host_cpu_ut = prs_utils.get_cpu_utilization(
            context, node, self._scheduler_api) / 100.0
        vm_cpu_ut = self.__get_instance_load(instance) / 100.0
        vm_vcpus = get_instance_vcpus(instance)
        src_host_vcpus = get_hypervisor_vcpus(node)
        return float("%.2f" % (src_host_cpu_ut -
                               (vm_cpu_ut * vm_vcpus / src_host_vcpus)))

    def should_perform_instance_migration(self, context, instance_obj,
                                          policy_info):
        """Determine if we should migrate the instance."""
        node = prs_utils.get_compute_node_by_service_host(
            context, instance_obj.host)
        curr_cpu_util = int(prs_utils.get_cpu_utilization(
            context, node, self._scheduler_api))
        predicted_cpu_util = int(100 * self.__get_predicted_host_load(
            context, instance_obj))

        if predicted_cpu_util > curr_cpu_util:
            # NOTE(jwcroppe): Handle the case where we get some rounding
            #                 oddities and avoid confusing messages.
            predicted_cpu_util = curr_cpu_util

        # Only perform the migration if the predicted CPU utilization post-
        # migration will decrease the current CPU utilization to the next
        # lowest percentage point (this avoids us performing migrations
        # that are effectively useless).
        do_migrate = True if predicted_cpu_util < curr_cpu_util else False

        if not do_migrate:
            LOG.info(_("Migration of virtual machine %(name)s (%(uuid)s) "
                       "should be skipped because its predicted CPU "
                       "utilization post-migration is %(predicted)d "
                       "percent and its current CPU utilization is %(curr)s "
                       "percent; no further virtual machines need to be "
                       "examined for host %(host)s.")
                     % {'name': instance_obj.display_name,
                        'uuid': instance_obj.uuid,
                        'predicted': predicted_cpu_util,
                        'curr': curr_cpu_util,
                        'host': instance_obj.host})

        return do_migrate

    def sort_instances_for_optimization(self, context, instances, policy_info):
        """Performs an in-place sort of the instances."""
        def satisfies_threshold(instance):
            cpu_util = self.__get_predicted_host_load(context, instance)
            return cpu_util * 100.0 <= int(policy_info['policy_threshold'])

        # Generate a list of VMs that will predictably satisfy the threshold.
        insts_satisfied = [vm for vm in instances if satisfies_threshold(vm)]
        # Sort them from lowest to highest memory footprint.
        insts_satisfied = sorted(insts_satisfied, key=lambda vm: vm.memory_mb)

        # Generate a list of VMs that do *not* satisfy the threshold.
        insts_satisfied_uuids = [vm.uuid for vm in insts_satisfied]
        insts_not_satisfied = [vm for vm in instances
                               if vm.uuid not in insts_satisfied_uuids]
        # Sort them from lowest to highest predicted CPU utilization post
        # migration (i.e., move the highest impact VMs first) and then
        # by memory (i.e., move the lowest memory VMs first).
        insts_not_satisfied = sorted(
            insts_not_satisfied,
            key=lambda vm: (self.__get_predicted_host_load(context, vm),
                            vm.memory_mb))

        # Perform the in-place update.
        instances[:] = insts_satisfied + insts_not_satisfied
