# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""PRS-related utilities."""

import time

from nova import objects
from nova import servicegroup
from nova.ibm.ego import ego_constants
from nova.ibm.ego import ego_utils
from nova.ibm.ego.hapolicy import constants as ha_constants

from oslo_log import log as logging

from powervc_nova import _
from powervc_nova.common import aggregate_utils as agg_utils
from powervc_nova.common import server_group_utils as sg_utils
from powervc_nova.db import api as db_api
from powervc_nova.objects.compute import dom as compute_dom
# This import is needed for PowerVCMigrateData object registry
# required for migration Please do NOT remove it,.
from powervc_nova.objects import migrate_data as mig_obj
from powervc_nova.prs import constants as pvc_prs_const
from powervc_nova.virt.ibmpowervm.common import constants as pvmc

POWERKVM_HYPERVISOR_TYPE = 'qemu'
NFS = 'nfs'

# Error codes for can_host_live_migrate
NO_LPM_LICENSE = 1
KVM_LOCAL_STORAGE = 2

# Error codes for can_instance_live_migrate
RMC_NOT_ACTIVE = 1
COLLOCATION_AFFINITY = 2
COLLOCATION_ONGOING_OPS = 3

# Defines the time-to-live for utilization cache entries (in seconds)
UTIL_CACHE_TTL = 30

# A cache for the `get_cpu_utilization` utility. It caches CPU utilization
# values for a compute node for up to `UTIL_CACHE_TTL` seconds. The purpose
# of this cache is to avoid several RPC calls within a short period. The cache
# format is cache[hypervisor_name] = {util: value, expires_at: UTC-time}.
_cpu_util_cache = dict()
# Similar cache for memory utilization.
_mem_util_cache = dict()

LOG = logging.getLogger(__name__)


def get_compute_node(context, hv_name):
    """
    This returns the respective compute node given a hypervisor hostname.

    :param context: A context object that is used to authorize DB access
    :param hv_name: The hypervisor hostname whose compute node to locate
    :returns: The corresponding compute node or None if not found
    """
    context = context.elevated()
    compute_nodes = (objects.ComputeNodeList.get_by_hypervisor(
                     context, hv_name) or [])
    return None if not len(compute_nodes) else compute_nodes[0]


def get_compute_node_by_service_host(context, service_host):
    """
    This returns the respective compute node given a service host.

    :param context: A context object that is used to authorize DB access
    :param hv_name: The service host whose compute node to locate
    :returns: The corresponding compute node or None if not found
    """
    context = context.elevated()
    compute_nodes = (objects.ComputeNodeList.get_all_by_host(
                     context, service_host) or [])
    return None if not len(compute_nodes) else compute_nodes[0]


def update_dro_enabled_in_compute_node_metadata(context, compute_node_id,
                                                dro_enabled):
    """
    This updates dro_enabled in the compute node metadata database table.

    :param context: A context object that is used to authorize DB access
    :param hv_name: The compute node ID
    :param dro_enabled: A string of either 'true' or 'false'
    """
    metadata = {pvc_prs_const.DRO_ENABLED_KEY: dro_enabled}
    db_api.compute_node_metadata_update(context, compute_node_id, metadata)


def get_dro_enabled_for_host(context, compute_node, check_host_groups=False):
    """
    This queries the compute node metadata DB API to obtain the dro_enabled
    property. If not set, it is defaulted to True.

    :param context: A context object that is used to authorize DB access
    :param hv_name: The compute node ID
    :param check_host_groups: If True, this will also check the parent host
                              groups' DRO status as part of the final
                              determination as to whether the host has
                              DRO enabled
    :returns: True if DRO is enabled for the host and False otherwise
    """
    dro_enabled = True
    metadata = db_api.compute_node_metadata_get(context, compute_node.id)
    dro_enabled_value = metadata.get(pvc_prs_const.DRO_ENABLED_KEY)
    if dro_enabled_value is not None and dro_enabled_value.lower() == 'false':
        dro_enabled = False
    elif check_host_groups:
        aggregate_ids = agg_utils.get_aggregate_ids_for_host(context,
                                                             compute_node.host)
        for agg_id in aggregate_ids:
            if agg_utils.get_dro_enabled_for_aggregate(context, agg_id):
                break
        else:
            # NOTE(tpeoples): Yes, the else is supposed to be indented at the
            # same level as the for loop.  See https://docs.python.org/2/
            # tutorial/controlflow.html#break-and-continue-statements-
            # and-else-clauses-on-loops for more info.
            #
            # All aggregates that the host belongs to have DRO disabled, so DRO
            # must be disabled for the host as well
            dro_enabled = False

    return dro_enabled


def is_host_up(context, compute_node):
    """
    Determines whether or not a host is considered up.

    :param context: Nova context object
    :param compute_node: Nova compute node object
    :returns: True if the host is deemed up, False otherwise
    """
    service = objects.Service.get_by_compute_host(context, compute_node.host)

    servicegroup_api = servicegroup.API()
    return servicegroup_api.service_is_up(service)


def is_in_maintenance_mode(compute_node):
    """
    Determines whether or not a host is in maintenance mode.

    :param compute_node: Nova compute node object
    :returns: True if the host is in maintenance mode, False otherwise
    """
    ego_states = ego_utils.get_compute_status()
    return ego_states.get_maintenance(
        compute_node.hypervisor_hostname) != ha_constants.STATE_OK


def can_host_live_migrate(context, compute_node):
    """
    Determines whether or not a host is capable of live migrations.

    :param context: Nova context object
    :param compute_node: Nova compute node object
    :returns: dict of a boolean value for 'allowed' and an integer
              reason code for 'reason_code'.
    """
    result = {'allowed': True, 'reason_code': None}
    hypervisor_type = compute_node.hypervisor_type
    if not hypervisor_type:
        return result
    elif hypervisor_type in pvmc.IBM_POWERVM_HYPERVISOR_TYPES:
        stats = compute_node.stats or {}
        # Check to ensure the PowerVM hypervisor is LPM capable.
        if not stats.get(pvmc.ACTIVE_LPAR_MOBILITY_CAPABLE):
            result['allowed'] = False
            result['reason_code'] = NO_LPM_LICENSE
    elif hypervisor_type.lower() == POWERKVM_HYPERVISOR_TYPE:
        # Make sure the storage is not local storage.
        host_fact = compute_dom.ManagedHostFactory.get_factory()
        managed_host = host_fact.find_host_by_name(context, compute_node.host)
        if managed_host and managed_host.host_storage_type.lower() != NFS:
            result['allowed'] = False
            result['reason_code'] = KVM_LOCAL_STORAGE
    return result


def can_instance_live_migrate(context, instance, compute_node):
    """
    Determines whether an instance can perform live migration.

    :param context: Nova context object
    :param instance: Nova instance object
    :param compute_node: Nova compute node object
    :returns: a dict where key 'allowed' is True/False, key 'reason_code'
              represents the error code and key 'extra_data' is a dict
              representing ancillary information about the reason code.
    """

    result = dict(allowed=True, reason_code=None, extra_data=None)

    # If hypervisor type is PowerVM, the RMC state of the instance
    # must be active.
    if compute_node.hypervisor_type in pvmc.IBM_POWERVM_HYPERVISOR_TYPES:
        inst_sys_meta = instance.system_metadata
        inst_power_specs = instance.power_specs
        rmc_state = str(inst_power_specs.get('rmc_state', 'unknown'))
        os_distro = str(inst_sys_meta.get('image_os_distro'))
        # For IBM i VMs, rmc_state is not needed for migration
        if rmc_state != 'active' and pvmc.IBMI_OS_DISTRO != os_distro:
            result['allowed'] = False
            result['reason_code'] = RMC_NOT_ACTIVE
            result['extra_data'] = dict(rmc_state=rmc_state)

    # Live migration is not supported for this instance if it is a member of
    # an affinity rule and is not the only instance in the rule. It is also
    # not supported if this instance is a member of a collocation rule
    # where another instance is undergoing an operation that could cause it
    # to change hosts. This is because, in both of these cases, a live
    # migration of this instance would fail.
    unsupported_groups = get_unsupported_collocation_rules(context,
                                                           instance.uuid)

    unsupported_affinity_groups = (
        unsupported_groups.get('unsupported_affinity_groups'))
    if unsupported_affinity_groups:
        result['allowed'] = False
        result['reason_code'] = COLLOCATION_AFFINITY
        result['extra_data'] = dict(
            unsupported_affinity_groups=unsupported_affinity_groups)

    unsupported_groups_w_prl_ops = (
        unsupported_groups.get(
            'unsupported_groups_with_parallel_ops'))
    if unsupported_groups_w_prl_ops:
        result['allowed'] = False
        result['reason_code'] = COLLOCATION_ONGOING_OPS
        result['extra_data'] = dict(
            unsupported_groups_with_parallel_ops=unsupported_groups_w_prl_ops)

    return result


def get_unsupported_collocation_rules(context, uuid):
    """
    Determines which collocation rules an instance belongs to and if:

    1) Any of its collocation rules are 'affinity' and there is more than one
       instance in the collocation rule and they're involved in ongoing
       operations; or
    2) Any of its collocation rules are 'anti-affinity' and they're involved in
       ongoing operations.

    If either of these are true, the instance does not support being involved
    in a live migration operation.

    :param context: Nova context object
    :param uuid: Instance uuid
    :returns: a dict of unsupported affinity rules and unsupported rules due to
              ongoing instance operations--where each of these is a list of
              names of the unsupported collocation rules.
    """
    unsupported_groups = dict()
    # Ignore these checks if the nova conf option is set.
    if sg_utils.ignore_collocation_rules():
        return unsupported_groups

    # Check if the instance is in any server groups.
    groups = sg_utils.get_all_groups_for_instance(context, uuid)
    unsupported_affinity_groups = []
    unsupported_groups_with_parallel_ops = []
    for group in groups:
        # If any instances are a member of an affinity collocation rule and
        # they are not the only instance in the in the rule, live migrations
        # will fail.
        members = sg_utils.get_members(context, group.uuid)
        if sg_utils.AFFINITY in group.policies and len(members) > 1:
            unsupported_affinity_groups.append(group.name)
            continue

        # If any other instance in the group is undergoing an operation that
        # would cause that instance to change hosts, live migrations will fail.
        has_ongoing = sg_utils.has_ongoing_operations_in_group(
            group.uuid, uuid, members,
            ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION)
        if has_ongoing:
            unsupported_groups_with_parallel_ops.append(group.name)

    if unsupported_affinity_groups:
        groups_str = sg_utils.get_str_from_list(unsupported_affinity_groups)
        unsupported_groups['unsupported_affinity_groups'] = groups_str

    if unsupported_groups_with_parallel_ops:
        groups_str = sg_utils.get_str_from_list(
            unsupported_groups_with_parallel_ops)
        unsupported_groups['unsupported_groups_with_parallel_ops'] = groups_str

    return unsupported_groups


def _get_utilization(context, compute_node, scheduler_api, cache, cache_ttl,
                     rpi_name):
    hypervisor_name = compute_node.hypervisor_hostname
    cached_data = cache.get(hypervisor_name)
    curr_time = int(time.time())

    if cached_data is not None:
        expires_at = cached_data['expires_at']
        if (expires_at - curr_time) > cache_ttl:
            # This handles the case in which the system clock was set back
            # (e.g., for daylight savings adjustments); we don't want the
            # cache to be 'stuck' while time catches up, so just force the
            # cached entry to expire and cache a new value.
            expires_at = curr_time
        if expires_at > curr_time:
            # i.e., the cached entry's TTL has not expired.
            return cached_data['util']

    # i.e., either we had a cache miss or the TTL expired.
    load = 0
    try:
        select = "select(hypervisor_hostname=='%s')" % hypervisor_name
        perc = [n['attributes'][rpi_name] for n in
                scheduler_api.ego_get_resource_list(context, select)]
        load = perc[0]
    except (IndexError, KeyError):
        LOG.warning(_("Hypervisor's (%(hv_name)s) %(rpi_name)s RPI attribute "
                      "could not be retrieved from EGO because EGO could not "
                      "find a resource associated with the hypervisor, likely "
                      "due to a recent (re)start of services; assuming 0.0.")
                    % dict(hv_name=hypervisor_name, rpi_name=rpi_name))
    except Exception as e:
        LOG.warning(_("Error retrieving hypervisor's (%(hv_name)s) "
                      "%(rpi_name)s RPI attribute from EGO; assuming 0.0.")
                    % dict(hv_name=hypervisor_name, rpi_name=rpi_name))
        LOG.exception(e)

    # Calculate the result and cache the value.
    load = float("%.2f" % (float(load) * 100))
    cache[hypervisor_name] = dict(util=load, expires_at=curr_time + cache_ttl)

    return load


def get_cpu_utilization(context, compute_node, scheduler_api):
    """
    Returns the CPU utilization of the host corresponding to the compute node.
    The return value will be a percentage instead of a decimal. For example,
    if the utilization is 90%, the return value will be 90.0.
    """
    return _get_utilization(context, compute_node, scheduler_api,
                            _cpu_util_cache, UTIL_CACHE_TTL,
                            'cpu.percent')


def get_mem_utilization(context, compute_node, scheduler_api):
    """
    Returns the memory utilization of the host corresponding to the compute
    node.  The return value will be a percentage instead of a decimal.
    For example, if the utilization is 90%, the return value will be 90.0.
    """
    return _get_utilization(context, compute_node, scheduler_api,
                            _mem_util_cache, UTIL_CACHE_TTL,
                            'memory.percent')


def get_assigned_proc_units(context, compute_node, scheduler_api):
    """
    Finds the total number of assigned proc units to a host, which is
    defined by the proc units that are assigned to VMs managed by this
    PowerVC instance plus the proc units that are reserved (assigned to VMs
    outside of this PowerVC instance)
    """
    hypervisor_name = compute_node.hypervisor_hostname
    assigned_proc_units = 0
    try:
        select = "select(hypervisor_hostname=='%s')" % hypervisor_name

        vcpu_consumed = [n['attributes']['vcpuConsumed'] for n in
                         scheduler_api.ego_get_resource_list(context, select)]
        vcpu_consumed = float(vcpu_consumed[0])

        proc_units_reserved = [
            n['attributes']['proc_units_reserved'] for n in
            scheduler_api.ego_get_resource_list(context, select)]
        proc_units_reserved = float(proc_units_reserved[0])

        assigned_proc_units = vcpu_consumed + proc_units_reserved
    except Exception as e:
        LOG.error(_("Error retrieving hypervisor's (%s) vcpuConsumed and "
                    "proc_units_reserved RPI attributes from EGO; "
                    "assuming 0.0 for both.")
                  % hypervisor_name)
        LOG.exception(e)
    return float(assigned_proc_units)
