# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import six

from oslo_log import log as logging
from nova.scheduler.ibm.ego import ego_resreq_utils
from nova.scheduler.ibm.ego import filters
from nova import objects
from nova import utils

from powervc_nova.prs import utils as prs_utils
from powervc_nova.virt.ibmpowervm.common import constants as pvmc
from powervc_nova import logcall
from powervc_nova import _

LOG = logging.getLogger(__name__)


class PowerVCBaseFilter(filters.EGOBaseFilter):
    """The base filter class for all PowerVC EGO filters."""

    def __init__(self, *args, **kwargs):
        super(PowerVCBaseFilter, self).__init__(*args, **kwargs)

    @staticmethod
    def _get_hypervisor_data(filter_properties):
        """Returns a dict of hypervisor data."""
        request_spec = filter_properties.get('request_spec', {})

        image = request_spec.get('image', {})
        image_props = image.get('properties', {})

        # NOTE (cjvolzka): Handle key changes introduced in Mitaka
        hyper_type = (image_props.get('img_hv_type') or
                      image_props.get('hypervisor_type'))
        arch = (image_props.get('hw_architecture') or
                image_props.get('architecture'))

        if hyper_type is None or arch is None:
            # e.g., we go down this path for evacuate (a.k.a. remote restart)
            instance_props = request_spec.get('instance_properties', {})
            sys_meta = utils.instance_sys_meta(instance_props) or {}
            if hyper_type is None:
                if sys_meta:
                    hyper_type = sys_meta.get('image_hypervisor_type')
                else:
                    node = prs_utils.get_compute_node_by_service_host(
                        filter_properties.get('context'),
                        instance_props.get('host'))
                    if node:
                        hyper_type = node.get('hypervisor_type')
                    else:
                        LOG.warn(_("Could not determine hypervisor_type of "
                                   "current host (%(host)s) for instance "
                                   "%(inst_uuid)s for filtering.")
                                 % dict(host=instance_props.get('host'),
                                        inst_uuid=instance_props['uuid']))
            if arch is None:
                arch = instance_props.get('architecture')

        return dict(hypervisor_type=hyper_type, architecture=arch)

    @logcall
    def _allow_oper_for_powerkvm(self, filter_properties, all_host_attrs):
        """
        Helper utility to determine if the operation is allowed for PowerKVM.
        We need to ensure the hypervisor type is QEMU and that the architecture
        is ppc64; these two checks ensure we're dealing with PowerKVM vs. xKVM.
        """
        hyper_data = PowerVCBaseFilter._get_hypervisor_data(filter_properties)
        hyper_type = hyper_data.get('hypervisor_type') or ''
        arch = hyper_data.get('architecture') or ''
        return hyper_type.lower() == 'qemu' and arch.lower() == 'ppc64'

    @logcall
    def _allow_oper_for_powervm(self, filter_properties, all_host_attrs):
        """
        Helper utility to determine if the operation is allowed for PowerVM.
        """
        hyper_data = PowerVCBaseFilter._get_hypervisor_data(filter_properties)
        hyper_type = hyper_data.get('hypervisor_type') or ''
        return hyper_type.lower() in pvmc.IBM_POWERVM_HYPERVISOR_TYPES

    @logcall
    def _get_hosts_to_hypervisors_map(self, context):
        """
        This returns a mapping of service hosts to the corresponding compute
        service hypervisor hostnames. This is mainly used in situations in
        which EGO was previously keying their RPI resources off of the
        service['host'] value instead of compute['hypervisor_hostname']; this
        conversion allows the other components of PowerVC to remain unaffected.

        :param context: A context object that is used to authorize DB access
        :returns: A map of service hosts to hypervisor hostnames
        """
        context = context.elevated()
        compute_nodes = objects.ComputeNodeList.get_all(context)
        host_hyper_map = {}

        for compute in compute_nodes:
            host_hyper_map[compute.host] = ego_resreq_utils.get_rpi_res_name(
                compute.hypervisor_hostname)

        LOG.debug("The service-host-to-compute-hypervisor-hostname mapping "
                  "is: %s" % str(host_hyper_map))
        return host_hyper_map

    @logcall
    def _get_source_host_data(self, instance_properties, all_host_attrs):
        """Helper utility to retrieve the source host's data."""
        source_host = instance_properties.get('ego_host')
        source_host_data = None

        if source_host is None:
            LOG.error(_("[%s] The key 'ego_host' was not set on the "
                        "instance_properties!") % self.__class__.__name__)
            return None

        # Iterate through the hosts until we find the source
        for hostname, host_data in six.iteritems(all_host_attrs):
            if source_host == hostname:
                # Found it; quit looping
                source_host_data = host_data
                break

        if source_host_data is None:
            LOG.error(_("[%(cls)s] Could not find source host (%(src_host)s) "
                        "in all_host_attrs!") %
                      dict(cls=self.__class__.__name__, src_host=source_host))

        return source_host_data
