# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from oslo_log import log as logging

import nova.context
from nova.ibm.ego import ego_constants
from nova.objects import flavor as flavor_obj

from powervc_nova.scheduler.filters import PowerVCBaseFilter
from powervc_nova.virt.powerkvm import proc_utils
from powervc_nova import logcall
from powervc_nova import _

LOG = logging.getLogger(__name__)


class PowerKVMSMTFilter(PowerVCBaseFilter):
    """
    This EGO scheduler filter eliminates candidate hosts whose SMT split-core
    mode is not compatible with the respective VM. The filter is applicable for
    deploy, resize and migration operations.
    """

    run_filter_once_per_request = True

    def __init__(self):
        super(PowerKVMSMTFilter, self).__init__()
        self._set_filter_scope(
            [ego_constants.FILTER_WORKING_SCOPE_CREATE,
             ego_constants.FILTER_WORKING_SCOPE_RESIZE,
             ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_COLD_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_EVACUATE,
             ego_constants.FILTER_WORKING_SCOPE_UNSHELVE])

    @logcall
    def _get_smt_value(self, inst_type):
        inst_type_id = inst_type['id']
        flavor_data = flavor_obj.Flavor.get_by_id(
            nova.context.get_admin_context(), inst_type_id)

        if not flavor_data or 'extra_specs' not in flavor_data:
            LOG.warning((_("Could not locate flavor (%(inst_type)s); its "
                           "flavor data is: %(flavor_data)s")
                         % {'inst_type': str(inst_type_id),
                            'flavor_data': str(flavor_data)}))
            # In this (unexpected) case, let's just assume they want SMT off.
            return proc_utils.IBM_POWERKVM_SMT_OFF

        smt_val = proc_utils.get_smt_value(flavor_data['extra_specs'])
        LOG.debug("The instance's (%s) flavor (%s) is requesting an SMT "
                  "value of: %s" % (str(inst_type_id), str(flavor_data),
                                    str(smt_val)))

        return smt_val

    @logcall
    def parse(self, filter_properties, all_host_attrs):
        if not self._allow_oper_for_powerkvm(filter_properties,
                                             all_host_attrs):
            LOG.debug("Ignoring %s..." % self.__class__.__name__)
            return

        request_spec = filter_properties.get('request_spec', {})
        inst_type = request_spec.get('instance_type', {})
        smt_val = self._get_smt_value(inst_type)

        if smt_val == proc_utils.IBM_POWERKVM_SMT_OFF:
            # In this case, the flavor is requesting that SMT is off, so this
            # guest can run anywhere as the threads will default to 1.
            LOG.debug("Ignoring %s as the flavor's requested "
                      "SMT value is 'off' and no filtering is required." %
                      self.__class__.__name__)
            return
        else:
            # Otherwise, locate any hosts capable of minimally supporting the
            # requested guest SMT setting.
            req_str = "$max_smt_per_guest >= %s" % str(smt_val)

        self._select(req_str)
        LOG.debug("After passing the %(filter)s, the resource requirement "
                  "expression is (%(exp)s)." % dict(
                      filter=self.__class__.__name__, exp=req_str))
