# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from nova.ibm.ego import ego_constants
from nova.scheduler.ibm.ego import ego_resreq_utils as ego_utils

from oslo_log import log as logging

from powervc_nova import logcall
from powervc_nova.objects.compute import dom as compute_dom
from powervc_nova.scheduler.filters import PowerVCBaseFilter
from powervc_nova.virt.ibmpowervm.common import constants as pvmc

LOG = logging.getLogger(__name__)

POWER_SPEC_AME_KEY = 'ame_expansion_factor'


class PowerVMActiveMemoryExpansionFilter(PowerVCBaseFilter):
    """
    This EGO scheduler filter ensures that the PowerVM host supports Active
    Memory Expansion (AME) if AME has been set for the virtual machine.
    """

    run_filter_once_per_request = True

    def __init__(self):
        super(PowerVMActiveMemoryExpansionFilter, self).__init__()
        self._set_filter_scope(
            [ego_constants.FILTER_WORKING_SCOPE_CREATE,
             ego_constants.FILTER_WORKING_SCOPE_RESIZE,
             ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_COLD_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_EVACUATE,
             ego_constants.FILTER_WORKING_SCOPE_UNSHELVE])

    @logcall
    def parse(self, filter_properties, all_host_attrs):
        if not self._allow_oper_for_powervm(filter_properties, all_host_attrs):
            LOG.debug("Ignoring %s..." % self.__class__.__name__)
            return

        apply_filter = False

        request_spec = filter_properties.get('request_spec', {})
        operation_type = request_spec.get('operation_type')

        ame_factor = None
        if operation_type == ego_constants.FILTER_WORKING_SCOPE_CREATE:
            # For deploys, check the flavor's extra_specs.
            instance_type = filter_properties.get('instance_type', {})
            extra_spec = instance_type.get('extra_specs', {})
            ame_factor = extra_spec.get('powervm:' +
                                        POWER_SPEC_AME_KEY, 0)
        else:
            # For any non-deploy operation, we need to grab the AME capability
            # name from the power_specs because of potential out-of-band
            # changes.
            context = filter_properties.get('context')
            instance_props = request_spec.get('instance_properties', {})
            inst_uuid = instance_props.get('uuid')
            inst_fact = compute_dom.ManagedInstanceFactory.get_factory()
            inst = inst_fact.find_instance_by_uuid(context, inst_uuid) or {}

            power_specs = inst.get('power_specs') or {}
            ame_factor = power_specs.get(POWER_SPEC_AME_KEY, 0)

        if ame_factor:
            try:
                if float(ame_factor) >= 1:
                    apply_filter = True
            # Converting to a float fails
            except ValueError as e:
                LOG.exception(e)
                LOG.error("%(val)s is an invalid %(key)s; "
                          "%(name)s will not be applied." %
                          dict(val=ame_factor, key=POWER_SPEC_AME_KEY,
                               name=self.__class__.__name__))

        if not apply_filter:
            LOG.debug("Ignoring %s because the virtual machine does not "
                      "require PowerVM's AME capabilities." %
                      self.__class__.__name__)
            return

        req_str = ("$%(ame_cap_name)s == True" %
                   dict(ame_cap_name=pvmc.ACTIVE_MEMORY_EXPANSION_CAPABLE))
        self._select(req_str)
        LOG.debug("After passing the %(name)s, the "
                  "resource requirement expression is (%(req)s)." %
                  dict(name=self.__class__.__name__, req=req_str))
