# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from oslo_log import log as logging
from nova import utils

from nova.ibm.ego import ego_constants

from powervc_nova.scheduler.filters import PowerVCBaseFilter
from powervc_nova.virt.ibmpowervm.common import constants as pvmc
from powervc_nova.virt.ibmpowervm.common import proc_utils
from powervc_nova import logcall
from powervc_nova import _

LOG = logging.getLogger(__name__)


class PowerVMEndianFilter(PowerVCBaseFilter):
    """
    This EGO scheduler filter ensures that if a Little-Endian image is
    being used that only Little-Endian capable hosts are selected.
    """

    run_filter_once_per_request = True

    def __init__(self):
        super(PowerVMEndianFilter, self).__init__()
        self._set_filter_scope(
            [ego_constants.FILTER_WORKING_SCOPE_CREATE,
             ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_COLD_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_EVACUATE,
             ego_constants.FILTER_WORKING_SCOPE_UNSHELVE])
        self._set_error_message(_("A compute template with a processor "
                                  "compatibility mode less than POWER8 cannot "
                                  "be used when deploying a Little Endian "
                                  "image."))

    @logcall
    def parse(self, filter_properties, all_host_attrs):
        if not self._allow_oper_for_powervm(filter_properties, all_host_attrs):
            LOG.debug("Ignoring %s..." % self.__class__.__name__)
            return

        request_spec = filter_properties.get('request_spec', {})
        image = request_spec.get('image', {})
        image_name = image.get('name', '<?>')
        image_props = image.get('properties', {})
        operation_type = request_spec.get('operation_type')
        endianness = image_props.get('endianness')

        if endianness is None:
            LOG.debug("There was no 'endianness' image property set for "
                      "image '(%(image_name)s)'; checking system metadata."
                      % dict(image_name=image_name))
            instance_props = request_spec.get('instance_properties', {})
            sys_meta = utils.instance_sys_meta(instance_props) or {}
            # Retrieve the Endianness of the image; default to Big-Endian
            endianness = sys_meta.get('image_endianness', 'big-endian')

        if endianness == 'little-endian':
            # If deploy, check processor compatibility mode
            if operation_type == ego_constants.FILTER_WORKING_SCOPE_CREATE:
                instance_type = filter_properties.get('instance_type', {})
                extra_spec = instance_type.get('extra_specs', {})
                flavor_compat_mode = extra_spec.get(
                    'powervm:processor_compatibility')
                if (flavor_compat_mode and
                        flavor_compat_mode.lower() != 'default'):
                    target_mode = proc_utils.extract_highest_compat_mode(
                        [flavor_compat_mode])
                    # The processor compatibility mode in the flavor cannot
                    # be lower than Power8
                    if target_mode < 8:
                        # Set include hosts to an empty list so the deploy
                        # is not scheduled to any hosts
                        self._include_hosts([])

            req_str = ("$%(cap_name)s == True"
                       % dict(cap_name=pvmc.LE_GUEST_CAPABLE))
            self._select(req_str)
            LOG.debug("After passing the %(filter)s, the resource requirement "
                      "expression is (%(exp)s)." % dict(
                          filter=self.__class__.__name__, exp=req_str))
