# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import six

from oslo_log import log as logging

from nova.ibm.ego import ego_constants

from powervc_nova.scheduler.filters import PowerVCBaseFilter
from powervc_nova.virt.ibmpowervm.common import constants as pvmc
from powervc_nova.virt.ibmpowervm.common import proc_utils
from powervc_nova import logcall
from powervc_nova import _

LOG = logging.getLogger(__name__)


class PowerVMNovaLinkMobilityFilter(PowerVCBaseFilter):
    """
    This EGO scheduler filter eliminates candidate hosts who are not eligible
    to serve as the destination host of a mobility operation where the source
    host of the instance is running on NovaLink. For example, if the candidate
    host is on less than POWER7, the candidate host will be excluded. This is
    because mobility operations from NovaLink hosts to HMC-managed P6,P5,etc.
    hosts are not supported by the NovaLink platform (P6 does not allow setting
    MAC addresses).

    This filter is applicable for all mobility operations, including active and
    inactive migrations, evacuations, and unshelves.
    """

    def __init__(self):
        super(PowerVMNovaLinkMobilityFilter, self).__init__()
        self._set_filter_scope(
            [ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_COLD_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_EVACUATE,
             ego_constants.FILTER_WORKING_SCOPE_UNSHELVE])
        self._set_error_message(_('No suitable host could be found for the '
                                  'operation because the source host is '
                                  'managed by NovaLink which does not support '
                                  'mobility operations to hosts on less than '
                                  'POWER7. All remaining hosts are on less '
                                  'than POWER7.'))

    @logcall
    def parse(self, filter_properties, all_host_attrs):
        if not self._allow_oper_for_powervm(filter_properties, all_host_attrs):
            LOG.debug("Ignoring %s..." % self.__class__.__name__)
            return

        request_spec = filter_properties.get('request_spec', {})
        instance_props = request_spec.get('instance_properties', {})

        host_data = self._get_source_host_data(
            instance_props, all_host_attrs) or {}
        if host_data.get(pvmc.MANAGED_BY) != pvmc.NOVALINK_PLATFORM:
            # If the source host of the instance is not managed by NovaLink,
            # then we don't need to check anything else.
            return

        exclude_hosts = []

        # Iterate over all hosts and check if they are not valid destinations
        for hostname, host_data in six.iteritems(all_host_attrs):
            # POWER6 (and lower) hosts will not show compatibility for POWER7
            # or above hosts, so we can use that to determine if a host is P6
            # or lower.
            compat_modes_str = host_data.get('compatibility_modes', '')
            compat_modes_list = compat_modes_str.split(',')

            highest_compat_mode = proc_utils.extract_highest_compat_mode(
                compat_modes_list)

            if highest_compat_mode < 7:
                inst_uuid = instance_props.get('uuid')
                LOG.debug('Filtered host (%(target)s) for instance (%(inst)s) '
                          'because the source host of the instance is managed '
                          'by NovaLink and %(target)s is not at least a '
                          'POWER7 machine. Mobility operations from NovaLink '
                          'to a host on less than POWER7 are not supported.' %
                          {'target': hostname,
                           'inst': inst_uuid})
                exclude_hosts.append(hostname)

        self._exclude_hosts(exclude_hosts)
