# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import six

from nova import utils
from oslo_log import log as logging

from nova.ibm.ego import ego_constants

from powervc_nova.objects.compute import dom as compute_dom
from powervc_nova.scheduler.filters import PowerVCBaseFilter
from powervc_nova.virt.ibmpowervm.common import proc_utils
from powervc_nova.virt.ibmpowervm.common.constants import IBMI_OS_DISTRO
from powervc_nova import logcall

LOG = logging.getLogger(__name__)


class PowerVMProcCompatModeFilter(PowerVCBaseFilter):
    """
    This EGO scheduler filter eliminates candidate hosts whose processor compat
    modes don't meet the to-be-relocated virtual machine's compat modes. The
    filter is only applicable for active and inactive migration operations.
    """

    run_filter_once_per_request = True

    def __init__(self):
        super(PowerVMProcCompatModeFilter, self).__init__()
        self._set_filter_scope(
            [ego_constants.FILTER_WORKING_SCOPE_CREATE,
             ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_COLD_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_EVACUATE])

    @logcall
    def parse(self, filter_properties, all_host_attrs):
        if not self._allow_oper_for_powervm(filter_properties, all_host_attrs):
            LOG.debug("Ignoring %s..." % self.__class__.__name__)
            return

        request_spec = filter_properties.get('request_spec', {})
        instance_props = request_spec.get('instance_properties', {})
        operation_type = request_spec.get('operation_type')
        is_live_mig = (operation_type ==
                       ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION)
        is_deploy = (operation_type ==
                     ego_constants.FILTER_WORKING_SCOPE_CREATE)
        # Treat rebuilds as synonymous with cold migrations
        is_cold_mig = ((operation_type ==
                        ego_constants.FILTER_WORKING_SCOPE_COLD_MIGRATION) or
                       (operation_type ==
                        ego_constants.FILTER_WORKING_SCOPE_EVACUATE))

        context = filter_properties.get('context')
        inst_uuid = instance_props.get('uuid')
        inst_fact = compute_dom.ManagedInstanceFactory.get_factory()
        inst = inst_fact.find_instance_by_uuid(context, inst_uuid) or {}

        power_specs = inst.get('power_specs') or {}
        desired = power_specs.get('desired_compatibility_mode')
        curr = power_specs.get('current_compatibility_mode')

        system_metadata = utils.instance_sys_meta(instance_props)

        # Compat_mode filter for IBM i
        os_type = system_metadata.get('image_os_distro')
        flavor_compat_mode = ""
        if is_deploy:
            instance_type = filter_properties.get('instance_type')
            extra_spec = instance_type.get('extra_specs')
            flavor_compat_mode = extra_spec.get(
                'powervm:processor_compatibility')

        exclude_hosts = []

        # Iterate over all hosts and check their compatibility modes
        for hostname, host_data in six.iteritems(all_host_attrs):
            compat_modes_str = host_data.get('compatibility_modes', '')
            compat_modes_list = compat_modes_str.split(',')

            if is_live_mig:
                if not proc_utils.check_proc_compat_mode_active_migration(
                        desired, curr, compat_modes_list):
                    LOG.debug(("Filtered host (%s) for instance (%s) "
                               "because of its processor compatibility "
                               "modes (current = %s; desired = %s; "
                               "compat modes = %s; active migration).") %
                              (hostname, inst_uuid, curr, desired,
                               compat_modes_str))
                    exclude_hosts.append(hostname)
                    continue
                if IBMI_OS_DISTRO == os_type:
                    # LPM host filter for IBM i
                    target_model = proc_utils.extract_highest_compat_mode(
                        compat_modes_list)
                    if target_model < 7:
                        LOG.debug(("Filtered host (%s) for instance (%s) "
                                   "because live migration of an IBM i "
                                   "instance is only supported "
                                   "on POWER7 and above.") %
                                  (hostname, inst_uuid))
                        exclude_hosts.append(hostname)
                        continue
            elif is_deploy and IBMI_OS_DISTRO == os_type:
                if (flavor_compat_mode and
                        flavor_compat_mode.lower() != 'default'):
                    target_mode = proc_utils.extract_highest_compat_mode(
                        compat_modes_list)
                    if target_mode == 6:
                        LOG.debug(("Filtered host (%s) for instance (%s) "
                                   "because for IBM i instances, "
                                   "its processor compatibility mode can "
                                   "only be set as 'default' on POWER6 "
                                   "hosts.") % (hostname, inst_uuid))
                        exclude_hosts.append(hostname)
            elif is_cold_mig:
                if not proc_utils.check_proc_compat_mode_inactive_migration(
                        desired, compat_modes_list):
                    LOG.debug(("Filtered host (%s) for instance (%s) "
                               "because of its processor compatibility "
                               "modes (desired = %s; compat modes = %s; "
                               "inactive migration).") %
                              (hostname, inst_uuid, desired, compat_modes_str))
                    exclude_hosts.append(hostname)
                    continue
                if IBMI_OS_DISTRO == os_type:
                    # For IBM i pre-requisite check
                    if desired and desired.lower() != 'default':
                        target_mode = proc_utils.extract_highest_compat_mode(
                            compat_modes_list)
                        if target_mode <= 6:
                            LOG.debug(("Filtered host (%s) for instance "
                                       "(%s) because for an IBM i instance, "
                                       "if its processor compatibility modes "
                                       "= POWER(%s), it only can be inactive "
                                       "migrated to POWER7 or above.") %
                                      (hostname, inst_uuid, target_mode))
                            exclude_hosts.append(hostname)

        self._exclude_hosts(exclude_hosts)
