# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from nova import objects
from oslo_log import log as logging

from nova.ibm.ego import ego_constants
from nova.ibm.ego.hapolicy import constants as ha_const

from powervc_nova.db import api as pvcdb_api
from powervc_nova.objects.compute import dom as compute_dom
from powervc_nova.prs import utils as prs_utils
from powervc_nova.scheduler.filters import PowerVCBaseFilter
from powervc_nova.virt.ibmpowervm.common import constants as pvmc
from powervc_nova import logcall
from powervc_nova import _

LOG = logging.getLogger(__name__)


class PowerVMRemoteRestartFilter(PowerVCBaseFilter):
    """
    This EGO scheduler filter ensures that the PowerVM host supports remote
    restart operations.  The simplified remote restart operations are only
    supported on POWER8 or newer hosts.
    """

    run_filter_once_per_request = True

    def __init__(self):
        super(PowerVMRemoteRestartFilter, self).__init__()
        self._set_filter_scope(
            [ego_constants.FILTER_WORKING_SCOPE_CREATE,
             ego_constants.FILTER_WORKING_SCOPE_RESIZE,
             ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_COLD_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_EVACUATE,
             ego_constants.FILTER_WORKING_SCOPE_UNSHELVE])

    @logcall
    def parse(self, filter_properties, all_host_attrs):
        if not self._allow_oper_for_powervm(filter_properties, all_host_attrs):
            LOG.debug("Ignoring %s..." % self.__class__.__name__)
            return

        request_spec = filter_properties.get('request_spec', {})
        operation_type = request_spec.get('operation_type')
        is_deploy = (operation_type ==
                     ego_constants.FILTER_WORKING_SCOPE_CREATE)
        is_other = (operation_type in
                    [ego_constants.FILTER_WORKING_SCOPE_RESIZE,
                     ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION,
                     ego_constants.FILTER_WORKING_SCOPE_COLD_MIGRATION,
                     ego_constants.FILTER_WORKING_SCOPE_UNSHELVE])
        apply_filter = False

        if is_deploy:
            # For deploys, check the flavor's extra_specs.
            instance_type = filter_properties.get('instance_type', {})
            extra_spec = instance_type.get('extra_specs', {})
            flavor_srr_cap = str(extra_spec.get('powervm:srr_capability',
                                                False)).lower()
            if flavor_srr_cap == 'true':
                apply_filter = True
        elif is_other:
            # For any non-deploy operation, we need to grab the SRR capability
            # name from the power_specs because of potential out-of-band
            # changes.
            context = filter_properties.get('context')
            instance_props = request_spec.get('instance_properties', {})
            inst_uuid = instance_props.get('uuid')
            inst_fact = compute_dom.ManagedInstanceFactory.get_factory()
            inst = inst_fact.find_instance_by_uuid(context, inst_uuid) or {}

            power_specs = inst.get('power_specs') or {}
            srr_cap = str(power_specs.get('srr_capability', False)).lower()

            if srr_cap == 'true':
                apply_filter = True
        else:
            # For evacuates, always apply the filter.
            apply_filter = True
            context = filter_properties.get('context')
            instance_props = request_spec.get('instance_properties', {})

        if not apply_filter:
            LOG.debug("Ignoring %s because the virtual machine does not "
                      "require PowerVM's simplified remote restart "
                      "capabilities." % self.__class__.__name__)
            return

        # In non-evacuate cases, don't apply the check for the number of remote
        # restart operations in progress.
        if (operation_type in
                [ego_constants.FILTER_WORKING_SCOPE_CREATE,
                 ego_constants.FILTER_WORKING_SCOPE_RESIZE,
                 ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION,
                 ego_constants.FILTER_WORKING_SCOPE_COLD_MIGRATION,
                 ego_constants.FILTER_WORKING_SCOPE_UNSHELVE]):
            req_str = ("$%(rr_cap_name)s == True" %
                       dict(rr_cap_name=pvmc.PVM_SIMP_REMOTE_RESTART_CAPABLE))
        # Note the operation type is implicitly 'evacuate' in the below cases
        else:
            req_str = ("($%(rr_cap_name)s == True)" % dict(
                       rr_cap_name=pvmc.PVM_SIMP_REMOTE_RESTART_CAPABLE))

            service_host = instance_props.get('host')
            node = prs_utils.get_compute_node_by_service_host(context,
                                                              service_host)
            if node:
                req_str = ("%(base_req)s && "
                           "($%(mb_key)s == '%(host_mb)s')" % dict(
                               base_req=req_str,
                               mb_key=pvmc.MANAGED_BY,
                               host_mb=node['stats'].get(pvmc.MANAGED_BY)))
            else:
                LOG.warn(_("Could not get node for host %(host)s. Skipping "
                           "'managed_by' check in %(filter)s.") % dict(
                         host=service_host, filter=self.__class__.__name__))

            # If the HA engine invoked us, then we shouldn't filter based on
            # count because the HA engine needs to 'throttle' the requests
            # appropriately (and if we filtered, the HA engine would get
            # NoValidHost, which would send it down an error path).
            if ego_constants.HA_MAINTENANCE_STATUS in filter_properties:
                LOG.info(_("%(filter)s invoked via a rebuild operation from "
                           "the HA engine; ignoring progress counts as the HA "
                           "engine will address those.") % dict(
                         filter=self.__class__.__name__))
            else:
                req_str = ("%(base_req)s && "
                           "($%(rrip_cap_name)s < $'%(rrs_cap_name)s')" % dict(
                               base_req=req_str,
                               rrip_cap_name=pvmc.NUM_RR_IN_PROGRESS,
                               rrs_cap_name=ha_const.PARALLEL_DEST_REBUILD))
        self._select(req_str)
        LOG.debug("After passing the %(filter)s, the resource requirement "
                  "expression is (%(exp)s)." % dict(
                      filter=self.__class__.__name__, exp=req_str))


class PowerVMRemoteRestartSameHMCFilter(PowerVCBaseFilter):
    """
    This EGO scheduler filter ensures that the remote restart operations stay
    within the domain of a single HMC as cross-HMC remote restart operations
    are not allowed at this time.
    """

    def __init__(self):
        super(PowerVMRemoteRestartSameHMCFilter, self).__init__()
        self._set_filter_scope(
            [ego_constants.FILTER_WORKING_SCOPE_EVACUATE])
        self._set_error_message(_("No suitable host could be found for the "
                                  "remote restart operation because the "
                                  "source host uses an HMC that is not shared "
                                  "with other hosts in the environment."))

    @staticmethod
    def _get_hmc_uuids(context, host_name):
        """Retrieve a list of HMC UUIDs managing the host."""
        return [str(hmc['hmc_uuid']) for hmc in
                pvcdb_api.hmc_find_all(context, dict(host_name=host_name))]

    @logcall
    def parse(self, filter_properties, all_host_attrs):
        context = filter_properties.get('context')
        context = context.elevated()
        request_spec = filter_properties.get('request_spec', {})
        instance_props = request_spec.get('instance_properties', {})
        host = instance_props.get('host')
        node = prs_utils.get_compute_node_by_service_host(context, host)

        if node:
            # If the source host is not managed by HMC, skip this filter.
            if node['stats'].get(pvmc.MANAGED_BY) != pvmc.HMC_PLATFORM:
                LOG.debug("Skipping %(filter)s because %(host)s is "
                          "not managed by HMC." % dict(
                          filter=self.__class__.__name__,
                          host=host))
                return
        else:
            LOG.warn(_("Could not get node for host %(host)s; %(filter)s "
                       "will be enforced.") % dict(
                     host=host, filter=self.__class__.__name__))

        src_hmc_uuids = PowerVMRemoteRestartSameHMCFilter._get_hmc_uuids(
            context, host)
        valid_hosts = []

        for node in objects.ComputeNodeList.get_all(context):
            dest_host = node.host
            dst_hmc_uuids = PowerVMRemoteRestartSameHMCFilter._get_hmc_uuids(
                context, dest_host)
            # Perform the set intersection to find common management HMCs.
            intersection = list(set(src_hmc_uuids) & set(dst_hmc_uuids))

            if not intersection:
                # No common management HMCs; not a valid destination host.
                continue
            valid_hosts.append(dest_host)

        host_to_hyper_map = self._get_hosts_to_hypervisors_map(context)
        ret = [host_to_hyper_map[host2] for host2 in valid_hosts
               if host2 in host_to_hyper_map]
        LOG.debug("Setting included hosts in %(filter)s (%(ret)s)." %
                  dict(filter=self.__class__.__name__, ret=ret))
        self._include_hosts(ret)
