# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from oslo_log import log as logging

from nova.ibm.ego import ego_constants

from powervc_nova.scheduler.filters import PowerVCBaseFilter
from powervc_nova.storage import storage_mapping_api
from powervc_nova import logcall
from powervc_nova import _

LOG = logging.getLogger(__name__)


class PowerVMStorageFilter(PowerVCBaseFilter):
    """
    This EGO scheduler filter eliminates candidate hosts whose storage
    is not compatible with the respective image/LPAR. The filter is
    applicable for all operations (create, relocate, resize, etc.).
    """

    run_filter_once_per_request = True

    def __init__(self):
        super(PowerVMStorageFilter, self).__init__()
        self._set_filter_scope(
            [ego_constants.FILTER_WORKING_SCOPE_CREATE,
             ego_constants.FILTER_WORKING_SCOPE_RESIZE,
             ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_COLD_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_EVACUATE,
             ego_constants.FILTER_WORKING_SCOPE_UNSHELVE])
        self._set_error_message(_("No suitable host could be found due to the "
                                  "storage configuration. Please ensure "
                                  "there are available hosts in the Storage "
                                  "Connectivity Group and that there is "
                                  "sufficient storage capacity for this "
                                  "operation and retry your request."))
        self._storage_mapping_api = storage_mapping_api.StorageMapping()

    @logcall
    def parse(self, filter_properties, all_host_attrs):
        if not self._allow_oper_for_powervm(filter_properties, all_host_attrs):
            LOG.debug("Ignoring %s..." % self.__class__.__name__)
            return

        request_spec = filter_properties.get('request_spec', {})
        context = filter_properties.get('context')
        instance_props = request_spec.get('instance_properties', {})
        uuid = instance_props.get('uuid')
        operation_type = request_spec.get('operation_type')
        flavor = None

        if operation_type == ego_constants.FILTER_WORKING_SCOPE_RESIZE:
            flavor = request_spec.get('instance_type')
        # Query storage for list of acceptable hosts
        host_list = self._get_viable_hosts(context, uuid, operation_type,
                                           flavor)
        xform_viable_hosts = []
        hosts = []

        # Transform the dictionary of dictionaries into a list of hosts
        for host_data in host_list.values():
            xform_viable_hosts.append(host_data.get('host'))

        # Build a list of hosts that satisfy the network requirements
        for host in all_host_attrs.keys():
            if host in xform_viable_hosts:
                hosts.append(host)
            else:
                LOG.debug(("Filtered host (%s) for instance (%s) because of "
                           "storage deficiencies.") % (host, uuid))

        LOG.debug("Setting included hosts in %(filter)s (%(hosts)s)." %
                  dict(filter=self.__class__.__name__, hosts=hosts))
        self._include_hosts(hosts)

    def _get_viable_hosts(self, context, uuid, operation_type, flavor):
        ret = {}
        res = self._storage_mapping_api.get_viable_hosts(
            context, instance_uuid=uuid, flavor=flavor,
            operation_type=operation_type)
        host_to_hyper_map = self._get_hosts_to_hypervisors_map(context)

        # Convert the host names to hypervisor hostnames
        for host in res.keys():
            hyper_hostname = host_to_hyper_map[host]
            ret[hyper_hostname] = {'host': hyper_hostname}

        return ret
