#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import re
import time
from oslo_log import log as logging
from oslo_config import cfg

from powervc_nova import _
import powervc_nova.db.api as db
import powervc_nova.objects.compute.dom as compute_dom
import powervc_nova.volume.cinder as pvc_cinder


LOG = logging.getLogger(__name__)
_NPIV_, _SSP_, _VSCSI_ = ('npiv', 'ibm_ssp', 'pv_vscsi')
CONF = cfg.CONF
CONF.import_opt('ibmpowervm_use_mapped_vfc_scheduling',
                'powervc_nova.objects.storage.dom')

# Define port status strings for PowerVC use since K2 does not have
# a port status entry for physical FC ports at this time.
PORT_STATUS_OK = 'OK'
PORT_STATUS_OFFLINE = 'Offline'
PORT_STATUS_UNKNOWN = 'Unknown'
PORT_STATUS_FULL = 'Full'
PORT_NPIV_STATUS_UNK_FABRIC = 'Unknown_fabric'
PORT_NPIV_STATUS_UNS_ADAPTER = 'Unsupported_adapter'
PORT_NPIV_STATUS_UNS_FABRIC = 'Unsupported_fabric'
PORT_BUFFER = 0


class StorageTopologyCommon(object):
    """
    This class alleviates having too much code in objects/storage/dom
    by compartmentalizing topology collection and reconciliation helpers.
    It also provides an abstraction layer for the hypervisor technology
    such that there should be sublclasses for HMC, PVM, etc.

    This class can me instantiated to provide database-topology, or
    a child class can be instantiated to get live topology from compute host
    """
    # compiled reg-exps
    RE_TYPE1 = re.compile('\d+ Giga')
    RE_TYPE2 = re.compile('\d+Gb ')

    def __init__(self, host_name, context):
        """ Base class constructor """
        self.topology = None
        self.context = context
        self.host_name = host_name
        self.host_type = 'none'
        # A reference to the storage DOM module is made at the class instance
        # level rather than as a class variable or global import to avoid
        # circular dependencies when this module is loaded.
        from powervc_nova.objects.storage import dom
        self.dom = dom
        self.use_mapped_vfc_scheduling =\
            CONF.ibmpowervm_use_mapped_vfc_scheduling

    def get_live_topology(self, skip_cluster=False, max_cache_age=-1):
        """ Subclass needs to implement. Put common work here. """
        LOG.error(_("Coding error: Base get_live_topology not implemented."))
        return {}

    def reconcile_topology(self):
        """ Main flow to get live topology and reconcile with DB """
        t1 = time.time()
        self.topology = self.get_live_topology()
        self.topology['cluster_membership_change'] = False  # init
        t2 = time.time()
        LOG.debug("Reconcile storage topology with DB...")
        db_vio_servers = self.update_db_resources_with_topo(self.topology)
        t3 = time.time()
        ################################################################
        # Prior DB transaction(s) are now closed.
        # Call a method to reconcile SCG membership with DB vios data.
        # It will establish its own new transaction to the db,
        # and determine if extra reconciliation is needed for SCG
        # membership.
        ################################################################
        if db_vio_servers is not None:
            LOG.debug("Run SCG membership reconciliation...")
            self.update_scgs_vios_membership(self.context, self.host_name,
                                             db_vio_servers, self.topology)
        t4 = time.time()
        LOG.debug("Completed storage reconciliation[%d]. Get live topo, Sync "
                  "FCPorts and clusters, Sync SCGs, Total seconds= "
                  "(%.2f, %.2f, %.2f, %.2f)." %
                  (self.topology['static_topo'].sequence_num, t2 - t1,
                   t3 - t2, t4 - t3, t4 - t1))
        return self.topology

    def update_db_resources_with_topo(self, topology):
        """
        Common storage method for synchronizing the DB with information
        retrieved during hypervisor topo collection. All collectors
        populate.
        :param topology: Dictionary structure populated in a common way
                         by the collectors.
        :return:         Topology dictionary passed in.
        """
        db_vio_servers = None
        if 'error' in topology:
            LOG.warn(_("Host storage topology returned with an error "
                       "condition set: %s") % topology['error'])
            LOG.debug("Skip topology reconciliation")
            return db_vio_servers
        LOG.debug("Storage Topology data retrieved for host_type: %s" %
                  self.host_type)

        # Get the ResourceTransaction (cf. DB2 session) for the nova db
        vios_factory = self.dom.VioServerFactory.get_factory()
        txn = vios_factory.get_resource_transaction()
        with txn.begin():
            # Get the list of VIOSes from the nova DB.
            # With the addition of cached in-memory VIOS objects, the
            # reason why they are still loaded from the DB here is that
            # FC port properties like port_tag and fc_protocol may have been
            # updated by the administrator user. So we get the most recent
            # snapshot of VIOSes here, and they are cached at the end of
            # the vios loop in update_vios_fc_ports_in_db().
            # It might be tempting use 'static_topo'.get_vios() cached vios
            # cached vios entries here, and only if a DB update was needed,
            # then query for the current port in the DB at that point and
            # update it as needed. but that would violate the caching design
            # put in place for deploy and other operations. If a user added a
            # port tag, for example, the cached port in would not reflect that
            # port tag and remain stale until some environment change happened
            # to it like AvailablePorts changing. The deploy flow assumes that
            # the cached ports are current within the timeframe of the periodic
            # task interval (e.g. 5 mins). This is the tradeoff to allow
            # scale-ability - we always query for the host vios & ports every
            # topo reconciliation interval, whether we need to or not AND the
            # deploy processing never queries for the DB ports on the host
            # (it uses the cached entries). So the DB query load back to the
            # PowerVC management server conductor process remains constant as
            # the concurrent number of deploy requests increase.
            #
            # TODO: There could be significant savings (would not require the
            # trade-off) if we could be notified of an admin update of FC port
            # user data like this. The in-memory cache could be updated
            # on-demand and this find_all_vioses_by_host could be skipped.
            # Similarly for the get_live_topology(), events for ports changing
            # on the host could cause the topo to update on-demand and always
            # be current.
            db_vio_servers = vios_factory.find_all_vioses_by_host(
                self.context, self.host_name, transaction=txn)

            if 'error' in topology or\
                    topology['skip-update-clusters']:
                LOG.debug(
                    "Skip cluster provider reconciliation since either "
                    "there was a failure, clusters are configured to "
                    "be skipped, no VIOSes are in a trusted state, "
                    "not reconciling topology for a single compute "
                    "host, or HMC level is below 8.1.10")
            else:
                # Update cluster related db tables if needed.
                self.update_cluster_storage_providers(
                    self.context, topology, db_vio_servers, self.host_name)

            # Call the helper method to save the VIOS ports in the DB and
            # update vios cluster membership if needed. It is important
            # that the update_cluster_storage_providers() call is made
            # beforehand, so that the cluster can exist in the DB for vios
            # entries to link to.
            LOG.debug("Reconcile VIOS cluster membership and FC Ports in "
                      "DB with Topology data...")
            self.update_vios_fc_ports_in_db(
                self.context, topology, db_vio_servers, transaction=txn)
            LOG.info(_("Host storage topology reconciled against Virtual "
                       "I/O Server IDs %s in collected topology data.") %
                     topology['vios_by_id'].keys())

        return db_vio_servers

    def update_cluster_storage_providers(self, context, topology, db_vios_list,
                                         host_name):
        """Update Clusters in the DB only if HMC Version > 8.1.0"""
        k2_clusters = dict()
        k2host_clusters = list()
        prvfact = self.dom.StorageProviderFactory.get_factory()
        hostfact = compute_dom.ManagedHostFactory.get_factory()
        hmc_uuids = None
        old_cluster_id = None
        if not topology['vios_by_id']:
            # Is this the behavior we want?
            LOG.debug("No VIOS partitions were retrieved from K2 for the host."
                      " Prior cluster data will likely be removed from the "
                      "DB.")

        # Get the list of Cluster Provider Names off of each VIOS for the given
        # Host
        for k2vios in [v['vios_cluster'] for v in
                       topology['vios_by_id'].values() if 'vios_cluster' in v]:
            if k2vios['cluster_id'] not in k2host_clusters:
                k2host_clusters.append(k2vios['cluster_id'])
                k2_clusters[k2vios['cluster_id']] = \
                    dict(cluster_name=k2vios['cluster_name'],
                         backend_id=k2vios['backend_id'])

        # Build a mapping of what all hosts are registered for each cluster
        db_clusters, dbhost_clusters = self.get_db_host_cluster_maps(context,
                                                                     host_name)
        LOG.debug("k2_clusters=%s, k2host_clusters=%s, db_clusters=%s, "
                  "dbhost_clusters=%s" % (k2_clusters, k2host_clusters,
                                          db_clusters, dbhost_clusters))

        #######################
        # Add processing      #
        #######################
        #   Note that we can get away with doing add processing before remove
        #   processing because:
        #   (1) Logic was added to the StorageConnectivityGroupFactory to first
        #       remove any SCG with the same default name before inserting the
        #       new SCG for a provider with the same display name. AND
        #   (2) We assume that any new cluster will have a unique cluster_id.
        #       Although this is true, default volume types adhere to a cluster
        #       by the cluster's display name AND it is awkward to see two
        #       providers with the same display name in a table (and template
        #       names). So we check for the case where a provider with the same
        #       name is in the DB already, but not coming back from K2 and we
        #       wait until it is removed and add it on the next run.
        #       It could be special cased and added after the remove, or the
        #       removes could happen first, but the amount of restructuring
        #       needed is too risky - I'd rather wait until post 1.2.1.
        #
        # Loop through the Clusters and see if we need to add any Storage
        # Providers
        for cluster_id in k2_clusters:
            backend_id = k2_clusters[cluster_id]['backend_id']
            cluster_name = k2_clusters[cluster_id]['cluster_name']
            if cluster_id not in dbhost_clusters:
                LOG.debug("K2 cluster ID %s is not in DB cluster list." %
                          cluster_id)
                # Get the UUID used for the HMC for the given Host
                if not hmc_uuids:
                    hmc_uuids = hostfact.find_host_by_name(context,
                                                           host_name).hmc_uuids
                try:
                    # Check if there are registered clusters with the same
                    # display name.
                    matches = prvfact.find_providers_by_display_name(
                        context, display_name=cluster_name)
                    LOG.debug("find_providers_by_display_name [%s]: %s." %
                              (cluster_name, matches))
                    for clst in matches:
                        if clst['host_name'] not in k2_clusters:
                            LOG.info(_("Cluster ID %(old)s is not in "
                                       "K2 data for this collection, so not "
                                       "adding new cluster for ID %(new)s "
                                       "until the old one is removed.") %
                                     dict(old=clst['host_name'],
                                          new=cluster_id))
                            old_cluster_id = clst['host_name']
                            break
                    if old_cluster_id:
                        continue
                    # Add the Cluster as a Storage Provider if it isn't
                    # already one
                    if prvfact.find_provider_by_name(context, cluster_id) is\
                            None:
                        LOG.info(_("Register new VIOS cluster provider "
                                   "%(cluster_id)s with name '%(name)s'.") %
                                 dict(cluster_id=cluster_id,
                                      name=cluster_name))
                        prvfact.construct_vios_cluster(context, cluster_id,
                                                       hmc_uuids, cluster_name,
                                                       backend_id,
                                                       auto_create=True)
                    # Add Cluster to this Host in the Host Clusters Database
                    # Table
                    values = dict(hmc_uuid=hmc_uuids[0], host_name=host_name,
                                  cluster_id=cluster_id)
                    db.host_cluster_create(context, values)
                    # If we executed the above block, then it could have been
                    # long running to create the SSP provider. So we need to
                    # re-acquire current DB data again, as other hosts may have
                    # updated it.
                    db_clusters, dbhost_clusters =\
                        self.get_db_host_cluster_maps(context, host_name)
                except Exception as ex:
                    LOG.warn('Error adding VIOS Cluster ' + cluster_id +
                             ' for host ' + host_name + '.')
                    LOG.exception(ex)

        #######################
        # Removal processing  #
        #######################
        static_topo = topology['static_topo']
        questionable_clusters = static_topo.get_clusters_in_ques(
            dbhost_clusters)
        # Loop through the Clusters with a check to delete any not returned by
        # k2
        for cluster_id in dbhost_clusters:
            if cluster_id not in k2host_clusters:
                LOG.debug("DB cluster ID %s is not in k2 cluster list." %
                          cluster_id)
                ##############################################################
                # The cluster id may not be in the k2host topology data because
                # the VIOSs had trouble reporting cluster membership to HMC.
                # In this case, we choose not to remove the DB association
                # and provider entry just yet.
                #
                # The flip side of this is what if the cluster membership
                # remains in question indefinitely? For example: K2 repeatedly
                # errors out, or the VIOSs remain in 'busy' state, or VIOS
                # partitions are deleted & different ones created to join or
                # not to join the cluster.
                # For the latter case, we probably would want to eventually
                # perform the cleanup in this code block if there was not going
                # to be new VIOSes from the host joining the cluster, but
                # ferreting out that particular case was not attempted as part
                # of 19372.
                ##############################################################
                # Additional note: We special case old_cluster_id that may
                # have been looked up earlier. We don't care if it is still
                # not "trusted". We implicitly trust that it needs to be
                # removed because a new cluster is coming in with the same
                # name.
                ##############################################################
                if cluster_id in questionable_clusters and\
                        cluster_id != old_cluster_id:
                    LOG.info(_("Skipping cluster '%(cluster)s' removal from "
                               "host %(host)s as cluster membership for the "
                               "host VIOSs may have dropped to 0 for temporary"
                               " environmental reasons.") %
                             dict(cluster=cluster_id, host=host_name))
                    continue
                try:
                    # Remove Cluster from Host in the Host Clusters Database
                    # Table
                    LOG.info(_("Remove cluster %(cluster)s from host %(host)s "
                               "in database.") % dict(cluster=cluster_id,
                                                      host=host_name))
                    db.host_cluster_delete(context, cluster_id, host_name)
                    # Remove Host from Storage Provider if it has no other
                    # hosts.
                    if len(db_clusters[cluster_id]) <= 0:
                        LOG.debug("There are no other hosts claiming "
                                  "cluster_id %s." % cluster_id)
                        if prvfact.\
                           find_provider_by_name(
                                context, cluster_id) is not None:
                            LOG.info(_("Remove VIOS cluster %(cluster)s from "
                                       "cinder volume providers.") %
                                     dict(cluster=cluster_id))
                            prvfact.delete_provider(context, cluster_id)
                        # If successful, remove our cache entry for it
                        if cluster_id in static_topo.cluster_keyed:
                            LOG.debug("Remove cached entry: %s." %
                                      static_topo.cluster_keyed[cluster_id])
                            del static_topo.cluster_keyed[cluster_id]
                    else:
                        LOG.debug("There is still a claim on cluster_id %s "
                                  "from: %s" % (cluster_id,
                                                db_clusters[cluster_id]))
                except Exception as ex:
                    LOG.warn('Error removing VIOS Cluster ' + cluster_id +
                             ' for host ' + host_name + '.')
                    LOG.exception(ex)
        return

    def get_db_host_cluster_maps(self, context, host_name):
        """ Return a mapping of clusters to other hosts. Also return a list of
            clusters for this host """
        db_clusters = dict()
        dbhost_clusters = list()
        for dbclust in db.host_cluster_find_all(context):
            if dbclust['cluster_id'] not in db_clusters:
                db_clusters[dbclust['cluster_id']] = list()
            # If the Cluster Mapping is to this Host, add it to this Hosts list
            if dbclust['host_name'] == host_name:
                dbhost_clusters.append(dbclust['cluster_id'])
            else:
                # Add to the db host list if the Cluster Mapping is to a
                # different host.
                db_clusters[dbclust['cluster_id']].append(dbclust['host_name'])
        return (db_clusters, dbhost_clusters)

    def update_vios_fc_ports_in_db(self, context, topology, db_vios_list,
                                   transaction):
        """
        Given the topology (built from data in the K2 responses) and the list
        of vioses from the database, format the data appropriately to persist
        the FC port information with their related VIOSes.
        Also, if the VIOS cluster membership has changed, update that as well.
        """
        for vios_id in topology['vios_by_id']:
            found = False
            for db_vios in db_vios_list:
                if vios_id == db_vios.id:
                    found = True

            if not found:
                # This is unexpected since the network topology collection
                # should run first and populated the VIOS entries.
                viosfact = compute_dom.VioServerFactory.get_factory()
                vio_dict = topology['vios_by_id'][vios_id]
                LOG.warn(_("The VIOS with ID=%(vid)s collected from the "
                           "hypervisior was not found in the database. This "
                           "is unexpected, but an entry is being added for "
                           "it. LPAR name is '%(name)s'.") %
                         {'vid': vios_id, 'name': vio_dict['name']})
                vios = viosfact.construct_vios(context,
                                               host_name=self.host_name,
                                               lpar_id=vio_dict['lpar_id'],
                                               lpar_name=vio_dict['name'],
                                               state=vio_dict['state'],
                                               rmc_state=vio_dict['rmc_state'])
                vios.id = vios_id
                vios.save(context=context, transaction=transaction)

        # dict to host mapping from wwpn to fcport DOM without fabric set
        wwpn_to_dom = {}
        fcport_factory = self.dom.FcPortFactory.get_factory()
        # Loop through the VIOSes from the DB. The assumption is that storage
        # will not be adding or removing VIOS entries in the DB as the network
        # reconciliation flow handles that. So we just key off of what is
        # already there.
        for db_vios in db_vios_list:
            db_vios_id = db_vios.id
            LOG.debug("LOOP entry: db_vios='%s', db_vios.id=%s."
                      % (db_vios.lpar_name, db_vios.id))

            if db_vios_id not in topology['vios_by_id']:
                LOG.warn(_("The Virtual I/O Server with id '%s' was"
                           " not retrieved in the "
                           "HMC reconciliation data. Skipping.") % db_vios_id)
                continue
            vios = topology['vios_by_id'][db_vios_id]
            LOG.debug("DB VIOS '%s' was matched to K2 vios: %s." %
                      (db_vios_id, vios))

            # Update the db_vios object with the cluster if needed.
            update_vios = False
            static_topo = topology['static_topo']
            if not topology['skip-update-clusters']:
                if static_topo.is_trust_criteria_met(db_vios_id):
                    if 'vios_cluster' in vios:
                        k2_clust = vios['vios_cluster']['cluster_id']
                    else:
                        k2_clust = None
                        # This is a trusted value, so clear out its tracked
                        # entry if any. The static topology can be used to
                        # remove or add vioses to SCGs later.
                        static_topo.remove_vios_from_cluster(db_vios_id)

                    if db_vios.cluster_provider_name != k2_clust:
                        update_vios = True
                        msg = _("Updating cluster_provider_name from "
                                "'%(old_id)s' to '%(new_id)s' on VIOS DB entry"
                                " with ID: '%(vios)s'.") % \
                            dict(old_id=db_vios.cluster_provider_name,
                                 new_id=k2_clust, vios=db_vios_id)
                        if(db_vios.cluster_provider_name and not k2_clust):
                            LOG.warn(msg)  # cluster going away from vios
                        else:
                            LOG.info(msg)
                        db_vios.cluster_provider_name = k2_clust
                        topology['cluster_membership_change'] = True
                    else:
                        LOG.debug("VIOS id '%s' has cluster already set "
                                  "correctly." % db_vios_id)
                else:
                    LOG.debug("Trust criteria not met for VIOS %s." %
                              db_vios_id)
                    # The next if check is bootstrap logic for the static topo
                    # where the cluster for a VIOS never came back from K2, but
                    # but it is still in the DB. We can set this as the initial
                    # cluster it was a member of if it was never cached in the
                    # static topology before, and on subsequent topology
                    # collections, this value will go through the normal trust/
                    # not-trusted determinations.
                    if db_vios.cluster_provider_name and\
                            'vios_cluster' not in vios and\
                            static_topo.vios_keyed[db_vios_id]['last_cluster']\
                            is None:
                        static_topo.set_cluster_for_vios(
                            db_vios_id, db_vios.cluster_provider_name, None)
            else:
                LOG.debug("Cluster information not collected for VIOS %s." %
                          db_vios_id)

            # I see that the rmc state can thrash between busy and not busy.
            # Since network and potentially compute topology is already setting
            # these, we won't here as well, but we'll output a log message to
            # correlate to a similar message network puts out.
            if 'state' in vios.keys():
                if (vios['state'] is not None and
                        vios['state'] != db_vios.state):
                    LOG.debug("discrepancy: lpar %s VIOS '%s' live state=%s, "
                              "but DB state=%s." %
                              (vios['lpar_id'], vios['name'],
                               vios['state'], db_vios.state))

            # Update the db_vios object with the rmc state
            if 'rmc_state' in vios.keys():
                if (vios['rmc_state'] is not None and
                        vios['rmc_state'] != db_vios.rmc_state):
                    LOG.debug("discrepancy: lpar %s VIOS '%s' live rmc_"
                              "state=%s, but DB rmc_state=%s." %
                              (vios['lpar_id'], vios['name'],
                               vios['rmc_state'],
                               db_vios.rmc_state))

            # Update the VIOS in the db if needed.
            if update_vios:
                db_vios.save(context=context, transaction=transaction)

            self._update_db_fc_ports(context, transaction, db_vios, vios,
                                     wwpn_to_dom)
            # VIOS has been reconciled. Cache it.
            topology['static_topo'].cache_vios(db_vios)
            # End VIOS loop
        if wwpn_to_dom.keys():
            # There are potentially ports to sync the fabric on - call helper.
            fcport_factory.sync_fcports_to_fabric(context, transaction,
                                                  wwpn_to_dom)
        LOG.debug("Completed reconciliation to DB for FC Ports.")

    def _update_db_fc_ports(self, context, transaction, db_vios, vios,
                            wwpn_to_dom):
        """ Helper method to sync ports for a VIOS as needed. """
        fcport_factory = self.dom.FcPortFactory.get_factory()
        # Get VIOS fc port DOM resources - this will trigger another DB
        # query if fcports have not been preloaded or touched yet for this
        # vios.
        db_vios_fcports = db_vios.fcports

        k2fcports = vios['fcports'] if vios['fcports'] is not None else {}
        # Loop through the FC Port database entries
        # and delete entries that no longer exist on the HMC
        for db_fcport in db_vios_fcports:
            LOG.debug("Processing db_fcport: %s." %
                      self.dom.fcport_to_dict(db_fcport))
            if db_fcport.id not in k2fcports:
                # The existing DB port entry was not found in the
                # dictionary of FC ports obtained through K2.
                xml = None  # for log messages
                if 'k2element' in vios:
                    xml = vios['k2element'].element.toxmlstring()
                if 'IOAdapters' in vios and (db_fcport.adapter_id in
                                             vios['IOAdapters']):
                    # Work around the problem where the K2 VIOS request
                    # does not return PhysicalFibreChannelAdapters
                    # periodically. But there was an IOAdapter
                    # element with the same DeviceName (udid) returned
                    # as for this port's adapter. We do not want to
                    # remove the existing port from the DB
                    # for this type of hiccup and it does not necessarily
                    # happen when the VIOS rmc_state is busy. Log the
                    # condition to allow future debug gathering.
                    LOG.info(_(
                        "PhysicalFibreChannelAdapter element is "
                        "missing from K2 operator feed for an existing "
                        "FC port db entry [%(port)s] where its adapter"
                        "_id matches a returned IOAdapter entry "
                        "[%(adapter)s]. Not removing the database port "
                        "entry for this case because the HMC may not "
                        "have been able to correlate FC adapters to a "
                        "temporarily unavailable VIOS. The VIOS "
                        "'%(viosn)s' has state '%(state)s' and RMC "
                        "state '%(rmc)s'.") %
                        dict(port=db_fcport.id,
                             adapter=db_fcport.adapter_id,
                             viosn=vios['name'], state=vios['state'],
                             rmc=vios['rmc_state']))
                    if 'trusted' in vios and vios['trusted']:
                        # Don't both dumping if VIOS not trusted.
                        LOG.debug("VIOS topology=%s, K2 element=%s" %
                                  (vios, xml))
                    # need to continue instead of breaking here so another
                    # port is not left to be created in the next loop
                    continue
                ########################
                # Delete FC Port case  #
                ########################
                # K2 did not report this FCport even though it did provide
                # vios information. The normal case here is to remove it
                # from the database.
                LOG.warn(_("Removing FC port from DB: %s") %
                         self.dom.fcport_to_dict(db_fcport))
                db_fcport.delete(context, transaction)
                if 'k2element' in vios:
                    LOG.info(_("K2 feed element used for VIOS when ports "
                               "found to be gone: %s") % xml)
            else:
                ########################
                # Update FC port case  #
                ########################
                update_needed = False
                k2fcport = k2fcports[db_fcport.id]
                status_combo = self.get_status_combo(
                    k2fcport, is_update=True, port_factory=fcport_factory,
                    db_fabric=db_fcport.fabric)
                LOG.debug("Found an existing port with "
                          "id '%s', dict=%s, status_combo=%s." %
                          (db_fcport.id, k2fcport, status_combo))
                if db_fcport.name != k2fcport['name']:
                    db_fcport.name = k2fcport['name']
                    update_needed = True
                    LOG.info(_("FC Port DB update needed per 'name' prop: "
                               " %s.") % k2fcport)

                if db_fcport.enabled is None:
                    db_fcport.enabled = True
                    update_needed = True
                if db_fcport.wwpn != k2fcport['wwpn']:
                    db_fcport.wwpn = k2fcport['wwpn']
                    update_needed = True
                    LOG.info(_("FC Port DB update needed per 'wwpn' prop: "
                               " %s.") % k2fcport)
                if db_fcport.adapter_id != k2fcport['adapter_id']:
                    db_fcport.adapter_id = k2fcport['adapter_id']
                    update_needed = True
                    LOG.info(_("FC Port DB update needed per 'adapter_id' "
                               "prop: %s.") % k2fcport)
                if db_fcport.fabric is None:
                    db_fcport.fabric = "None"
                    update_needed = True
                if db_fcport.status != status_combo:
                    db_fcport.status = status_combo
                    update_needed = True
                if 'mapped_vfcs' in k2fcport and\
                        db_fcport.mapped_vfcs != k2fcport.get('mapped_vfcs'):
                    db_fcport.mapped_vfcs = k2fcport['mapped_vfcs']
                    update_needed = True
                if 'available_vports' in k2fcport and\
                    (db_fcport.avail_active_vfcs !=
                     k2fcport.get('available_vports')):
                    db_fcport.avail_active_vfcs = \
                        k2fcport['available_vports']
                    update_needed = True
                if 'total_vports' in k2fcport and\
                        db_fcport.max_vfcs != k2fcport.get('total_vports'):
                    db_fcport.max_vfcs = k2fcport['total_vports']
                    update_needed = True

                if update_needed:
                    LOG.debug("Updating DB FC Port '%s' with K2 data: %s."
                              % (self.dom.fcport_to_dict(db_fcport),
                                 k2fcport))
                    db_fcport.save(context=context,
                                   transaction=transaction)

                # Add port to tracking dict even if a fabric is already set
                wwpn_to_dom[db_fcport.wwpn] = db_fcport
                # Remove port from k2 dictionary to indicate it's
                # 'processed'
                del k2fcports[db_fcport.id]
        # End for each DB FC port for the cur VIOS

        # Any ports left in k2fcports are new ones to insert in the db.
        for new_k2port in k2fcports.values():
            #########################
            # Create FC Port case   #
            #########################
            # Format the FC Port entry to match the criteria
            # of the Factory and provide defaults.
            fabric_default = "None"
            fcport_data = {'id': new_k2port['id'],
                           'vios_id': db_vios.id,
                           'adapter_id': new_k2port['adapter_id'],
                           'wwpn': new_k2port['wwpn'],
                           'name': new_k2port['name'],
                           'enabled': True,
                           'status': self.get_status_combo(new_k2port),
                           'fabric': fabric_default}
            if 'mapped_vfcs' in new_k2port:
                fcport_data['mapped_vfcs'] = new_k2port['mapped_vfcs']
            if 'available_vports' in new_k2port:
                fcport_data['avail_active_vfcs'] = \
                    new_k2port['available_vports']
            if 'total_vports' in new_k2port:
                fcport_data['max_vfcs'] = new_k2port['total_vports']

            LOG.info(_("Insert new FC Port in DB: %s.") % fcport_data)
            dom_fcport = fcport_factory.construct_fcport(
                context, auto_create=True, transaction=transaction,
                **fcport_data)
            # Save the FC port in the DB
            dom_fcport.save(context=context, transaction=transaction)
            # OK, port is persisted back to the DB, but it is not
            # necessarily hooked into the db_vios in-memory DOM object that
            # is going to get cached for this topo-reconcilation run.
            # We need to reach into the internal port list variable and
            # append this port.
            if db_vios._fcports is None:
                db_vios._fcports = []  # should not be needed, but just in case
            db_vios._fcports.append(dom_fcport)
            # And add it to our tracking dictionary for port-to-fabirc mapping
            wwpn_to_dom[new_k2port['wwpn']] = dom_fcport
        # End for new FC Port creation loop

    def get_status_combo(self, port_dict, is_update=False, port_factory=None,
                         db_fabric=None):
        """ Return port status and npiv_status in a combo string """
        if not is_update:
            # New port case
            return port_dict['status'] + '|' + port_dict['npiv_status']
        # For a port status being updated, handle a couple special cases.
        # To determine these cases, we are looking at the current DB port
        # status value, which is from the prior topology collection, so it
        # can take two topology cycles for the port status to update with
        # these new values. This is OK because they do not cause 'product'
        # function to be restricted. For case A, the fabric setting of None
        # already restricts the port, so the status does not add filtering
        # behavior - it only helps in the UI visibility for the user.
        # For case B, it adds function for vscsi and logs a warning.
        #   Case A: If there are fabrics discovered, but this port does not
        #           have a fabric associated, then we say the npiv status is
        #           unknown fabric.
        #   Case B: If a port is associated with a registered fabric but
        #           hypervisor reports it as offline, then there is an issue
        #           with the hypervisor status determination and we change
        #           the status to be OK so it does not get filtered out for
        #           this reason. i.e. we say the switch is authoritative for
        #           this case.
        if (port_factory.fabrics and port_dict['npiv_status'] ==
            PORT_STATUS_OK and
                (db_fabric is None or db_fabric == "None")):
            LOG.debug("Change NPIV OK status to unknown fabric for port: %s"
                      % port_dict)
            return port_dict['status'] + '|' + PORT_NPIV_STATUS_UNK_FABRIC
        if (db_fabric is not None and db_fabric != "None" and
                port_dict['status'] == PORT_STATUS_OFFLINE):
            LOG.warn(_("Change port offline status to OK as it is found logged"
                       " into a switch, but NPIV status is unknown: %(port)s")
                     % {'port': port_dict})
            return PORT_STATUS_OK + '|' + PORT_STATUS_UNKNOWN
        # Default
        return port_dict['status'] + '|' + port_dict['npiv_status']

    def is_npiv_adapter_description(self, adapter_desc):
        """
        Return whether the adapter might support NPIV or not.
        For 1.2.3, we say that an adapter that contains '4 Giga' or '4Gb'
        or less Gbs speed is not NPIV-capable.
        In a future release, we need a property from K2/? to tell us.
        """
        is_NPIV_supported = True  # default to expected NPIV support
        if not adapter_desc:
            LOG.debug("Empty adapter description.")
            return is_NPIV_supported
        adapter_desc = adapter_desc.replace('&amp;', '&')
        speeds = StorageTopologyCommon.RE_TYPE1.findall(adapter_desc)
        if not speeds:
            speeds = StorageTopologyCommon.RE_TYPE2.findall(adapter_desc)
        if not speeds:
            LOG.debug("FC speed pattern not found in: %s" % adapter_desc)
            return is_NPIV_supported
        gigabits = int(re.search(r'\d+', speeds[0]).group())
        if gigabits <= 4:
            # We are comfortable with assuming NPIV is not supported in this
            # case. The reason is that we only go through this code if the
            # port looks offline. But we error-conservatively here and say
            # that NPIV may not be supported, otherwise it is left in offline
            # state.
            is_NPIV_supported = False
        LOG.debug("Description: %s,   FC speed pattern match(es): %s,   "
                  "Gigabits=%d,   is_NPIV_capable=%s. " %
                  (adapter_desc, speeds, gigabits, is_NPIV_supported))
        # TODO: Power9 may have non-NPIV adapters likely greater than
        #       4 gigabit, but Timothy Damron says No to this.
        return is_NPIV_supported

    def update_scgs_vios_membership(self, context, host_name, db_vioses,
                                    topology):
        """
        Like most of the other methods in this module, update_scgs_vios_
        membership gets run during the periodic task to collect topology
        information and reconcile the PVC DB with it. There are four cases:

        (1) A VIOS exists in an SCG, but does not exist in the DB. The VIOS
            will be removed from the SCG, regardless of the SCG type or other
            properties since VIOSes should always be automatically removed.
            This case is HANDLED FOR US because of the PVC DB rules! The
            membership of a SCG is implemented as a special table association
            (many-to-many) such that if a VIOS member goes away, it's link
            in the SCG definition will also ago away.

        (2) An SCG is defined for Cluster-SSP access, and a VIOS is a member
            of the SCG, the VIOS exists in the DB, but that VIOS is not
            a member of the Cluster the SCG has access to. The VIOS will be
            removed from the SCG.
            TODO: Need to revist this. It does not seem safe to remove a VIOS
                  member if the SCG is associated with one or more VMs that
                  have a NPIV disk attached. The SCG membership may not match
                  what is in use and the VM may no longer migrate because NPIV
                  VIOS connectivity is removed for target hosts when the intent
                  is to remove SSP connectivity only.

        (3) An SCG is defined for Registered controller access only and has
            'auto_add_vios' set to True. A VIOS exists in the DB, but is
            not a member of the SCG. The VIOS's 'created_at' timestamp
            is compared to the SCG's 'updated_at' timestamp (or it's
            'created_at' timestamp if 'update_at' is not set). If the VIOS
            was registered after the last update of the SCG, then the
            VIOS is added to the SCG membership. This logic is not just an
            impl decision, but needs to be tied to the basic definiton of
            'auto_add_vios'. Timestamp comparisons are used because there
            is no SCG membership history being preserved.

        (4) An SCG is defined for Cluster-SSP access (it may also have
            registered controller access) and it has 'auto_add_vios' set
            to True. At some point earlier in the topology reconcilation flow
            we recorded the fact that a VIOS changed its cluster membership -
            either it was added to the SCG's cluster, or it was both removed
            from a differnt cluster and added to the SCG's cluster.
            This VIOS is not currently a member of the SCG. It will be added.
            Note that timestamp comparisons cannot be used for this case.

        Notes:
          * If no VIOSes remain as members of the SCG, the SCG does not get
            deleted in this flow. It just has an empty membership list. The
            DB lifecycle constraints could be set to delete the SCG when the
            members drop to 0, but currently there are no plans to enforce such
            a constraint since it seems better to allow the UI to present
            the situation to the admin and the admin can choose to remove the
            SCG or conclude that the situation is expected because they plan
            to add additional VIOSes later and the SCG is set to automatically
            include them.
          * Individual DB transactions will be created for each SCG
            so that problems with updating once SCG will not abort updates for
            all SCGs. It could be that an SCG no longer exists by the time the
            transaction commits for it and this needs to be silently
            handled.
          * We want to sequence intelligent checks that decide on the four
            cases intelligently so as to avoid unnecessary processing.
        """
        # Create a dictionary of vios id to vios DTO mappings.
        vios_map = {vios.id: vios for vios in db_vioses}

        # Construct a set of vios ids from the map
        cur_set = set(vios_map.keys())

        # Get a static var that we can use as a cache while the process runs.
        # It contains the last processed set of VIOSes for host when that
        # processing resulted in no SCG membership changes.
        cache_set = topology['static_topo'].host_vios_set

        # The '^' operator will yield the VIOSes that are in either set but
        # not both.
        exclusion = cur_set ^ cache_set
        t_since_eval = time.time() - topology['static_topo'].scg_eval_time
        LOG.debug("Exclusion=%s, cur_set=%s, cache_set=%s, skip_clust=%s, "
                  "time_since_eval_mins=%.1f" %
                  (exclusion, cur_set, cache_set,
                   topology['skip-update-clusters'], t_since_eval / 60.0))
        # Now if both the exclusion set of vioses is empty and there have been
        # no cluster member changes, then there is nothing that needs to be
        # done as none of the 4 SCG change cases are hit.
        #
        # Note that the 'cache' comparison is not telling us that for sure
        # something changed, but just that those VIOSes need to be looked at.
        # This could be the first execution of this code after a PVC restart,
        # in which case the cache will be empty and all the vioses need to be
        # looked at. Also, if it has been over 6 hours, do the SCG eval
        # as well - just a as a safeguard.
        if (not exclusion and not topology['cluster_membership_change'] and
                (t_since_eval < 3600 * 6)):
            LOG.debug("No SCG updates needed.")
            return

        topology['static_topo'].scg_eval_time = time.time()
        # The new_since_cache_set set computation represents vioses that are
        # being looked at for the first time, which will be all vioses the
        # first time this code is run after PVC startup.
        new_since_cache_set = (cur_set - cache_set)
        cluster_ids = [vios_map[x].cluster_provider_name for x in
                       cur_set if vios_map[x].cluster_provider_name]
        cluster_set = set(cluster_ids)
        LOG.debug("new_since_cache_set=%s, cluster_set=%s." %
                  (new_since_cache_set, cluster_set))

        # Get all the SCGs from the DB
        scg_factory = self.dom.StorageConnectivityGroupFactory.get_factory()
        scgs_found = scg_factory.find_all_scgs(context)
        num_changed = 0
        was_error = False
        #######################
        #   Main SCG Loop     #
        #######################
        for scg_found in scgs_found:
            txn = scg_factory.get_resource_transaction()
            with txn.begin():
                # Lookup the current SCG in the context of transaction for
                # potential updating.
                scg = scg_factory.find_scg_by_id(context, scg_found.id,
                                                 transaction=txn)
                if scg is None:
                    LOG.info(_("The storage connectivity group has been "
                               "removed since the time it was previously "
                               "looked up: '%s'." % scg_found.display_name))
                    # Drop out of transaction and continue to next SCG
                    continue
                LOG.debug("***Examining SCG***: %s." %
                          scg.to_dict(context=context, transaction=txn,
                                      skip_extended_lookups=True))
                scg_name = scg.display_name
                none_before = False
                if scg.cluster_provider_name and\
                        scg.cluster_provider_name in cluster_set:
                    cluster_set.remove(scg.cluster_provider_name)
                ##################################
                #   Exception Handling per SCG   #
                ##################################
                try:
                    if not scg.vios_ids and scg.cluster_provider_name is\
                            not None:
                        none_before = True
                    # call the helper
                    is_changed = self._scg_case_helper(
                        context, txn, scg, new_since_cache_set,
                        host_name, vios_map, topology['static_topo'])
                    if is_changed:
                        num_changed += 1
                    if not scg.vios_ids and none_before and scg.auto_defined:
                        LOG.info(_("Attempt removal of auto_defined storage "
                                   "connectivity group '%(scg)s' with no "
                                   "members for previous cluster provider ID "
                                   "%(cluster)s.") %
                                 dict(scg=scg_name,
                                      cluster=scg.cluster_provider_name))
                        scg_factory.delete_scg(context, scg.id)
                        num_changed += 1
                except Exception as ex:
                    LOG.exception(ex)
                    LOG.warn(_("There was an error attempting to check or "
                               "update the membership of storage connectivity "
                               "group (SCG) '%s'. Continue "
                               "to the next SCG..." % scg_name))
                    was_error = True
            LOG.debug("Transaction finished for SCG '%s'." % scg_name)

        # There is a timing window where host A process can overwrite
        # the VIOS members of an SCG that host B inserted. In case that
        # happened, we want to evaluate again on the next topology
        # collection cycle. But we should not need to eval again if there
        # were no changes to any SCGs. For that situation, we update
        # the cached set here.
        if num_changed == 0 and not was_error:
            topology['static_topo'].host_vios_set = cur_set
        else:
            # Reset to the empty set. This is just a safeguard. This set
            # is only used in evaluating the case where a VIOS needs to be
            # added to an SCG.
            topology['static_topo'].host_vios_set = set()
        LOG.debug("Finished reconciling SCGs. Num_changed=%d" % num_changed)

    def _scg_case_helper(self, context, transaction, scg, vios_ids_for_eval,
                         host_name, vios_map, static_topo):
        """
        Helper to update the SCG with a new VIOS membership list if needed.
        The different cases are evaluated here.
        """
        LOG.debug("_scg_case_helper: vios_ids_for_eval=%s, vios_to_clstr=%s." %
                  (vios_ids_for_eval, static_topo.vios_keyed))
        to_add = []
        to_remove = []
        scg_name = scg.display_name
        if hasattr(scg, 'updated_at') and scg.updated_at is not None:
            scg_time = scg.updated_at
        else:
            scg_time = scg.created_at

        ############################################################
        #   SCG VIOS Loop - only look at members for this host     #
        ############################################################
        if scg.cluster_provider_name is not None:
            for vios in static_topo.get_vioses(scg.vios_ids):
                vid = vios.id
                vlog_ref = "'" + vios.lpar_name + "'(" + vid + ")"
                LOG.debug("Case 2 check vios: %s" % vlog_ref)
                ###################
                #  CASE 2 check   #
                ###################
                if (scg.cluster_provider_name != vios.cluster_provider_name):
                    # Note: The trusted criteria would already have been met
                    # for the vios cluster membership to be changed in the DB,
                    # so that need not be checked again.
                    LOG.info(_("VIOS %(vios)s is no longer a member of "
                               "cluster '%(cluster)s'. Removing the member "
                               "from storage connectivity group '%(scg)s'.") %
                             dict(vios=vlog_ref,
                                  cluster=scg.cluster_provider_name,
                                  scg=scg_name))
                    to_remove.append(vios.id)
        ###################
        #  CASE 3 check   #
        ###################
        scg_vios_id_set = set(scg.vios_ids)
        if ('npiv' in scg.boot_connectivity or
                'npiv' in scg.data_connectivity or
                'pv_vscsi' in scg.boot_connectivity or
                'pv_vscsi' in scg.data_connectivity):
            is_external_scg = True
        else:
            is_external_scg = False

        if (is_external_scg and not scg.cluster_provider_name and
                scg.auto_add_vios):
            for vios_id in (vios_ids_for_eval - scg_vios_id_set):
                if scg_time < vios_map[vios_id].created_at:
                    LOG.info(_("Newly registered VIOS with ID %(vios)s will be"
                               " added to FC/NPIV-based storage connectivity "
                               "group '%(scg)s'.") %
                             dict(vios=vios_id, scg=scg_name))
                    to_add.append(vios_id)
                elif scg.auto_defined:
                    # If the SCG is auto-defined at start-up, then timestamps
                    # don't matter because the VIOS member always needs to
                    # be added. The SCG may have been created after VIOS
                    # registration due to a timing window.
                    LOG.info(_("New VIOS with ID '%(vios)s' will be added to "
                               "auto-defined FC/NPIV-based storage "
                               "connectivity group '%(scg)s'.") %
                             dict(vios=vios_id, scg=scg_name))
                    to_add.append(vios_id)
        ###################
        #  CASE 4 check   #
        ###################
        if scg.cluster_provider_name is not None and scg.auto_add_vios:
            scg_prov = scg.cluster_provider_name
            for db_vios_id in static_topo.vios_keyed.keys():
                if (db_vios_id not in scg_vios_id_set and
                        static_topo.is_trust_criteria_met(db_vios_id) and
                        static_topo.get_cluster_for_vios(db_vios_id) ==
                        scg_prov):
                    LOG.info(_("The VIOS with ID '%(vios)s' is a member of "
                               "cluster '%(cluster)s' and will be added to "
                               "storage connectivity group '%(scg)s'.") %
                             dict(vios=db_vios_id, cluster=scg_prov,
                                  scg=scg_name))
                    to_add.append(db_vios_id)
                else:
                    LOG.debug("Current SCG vios membership OK (or still "
                              "waiting on VIOS trust criteria to be met). "
                              "db_vios_id=%s, vios_cluster=%s, scg_cluster=%s."
                              % (db_vios_id,
                                 static_topo.get_cluster_for_vios(db_vios_id),
                                 scg.cluster_provider_name))
        num_to_add = 0  # init
        if to_remove or to_add:
            # Use the vioserver_members() call to ensure we get all vioses,
            # not just the members for this compute host.
            scg_vioses = scg.vioserver_members(port_preload=False)

            # Do the removes
            scg_vioses = [v for v in scg_vioses if v.id not in to_remove]
            LOG.debug("Num vioses after removes is %d. Num to add is %d." %
                      (len(scg_vioses), len(to_add)))

            # Check and do adds
            if to_add:
                # For cases 3 and 4, we don't have VIOS DOMs from the current
                # txn so we need to look each up.
                vios_factory = self.dom.VioServerFactory.get_factory()
                txn_vioses = vios_factory.find_all_vioses_by_host(
                    context, host_name, transaction=transaction)
                LOG.debug("txn_vioses=%s." % str(txn_vioses))
                for id_to_add in to_add:
                    txn_vios = [v for v in txn_vioses if v.id == id_to_add]
                    if not txn_vios:
                        LOG.warn(_("Virtual I/O Server ID '%(vios)s' marked "
                                   "for storage connectivity group '%(scg)s' "
                                   "inclusion, but is no longer found in the "
                                   "database for this host.") %
                                 dict(vios=id_to_add, scg=scg_name))
                    else:
                        num_to_add += 1
                        scg_vioses.append(txn_vios[0])

        ################
        # SAVE the SCG #
        ################
        if num_to_add > 0 or to_remove:
            # The big thing to note here is that the vios_list is updated
            # directly by this method prior to the save below. The
            # scg.replace_vios_list() method is not called to handle this
            # and do validation which might reject the change and raise
            # an exception. This means that auto_add_vios==True can result
            # in SCGs with non-applicable VIOS members unless extra logic
            # is added to this flow to individually validate. For the
            # auto_defined SCGs at least, this is the intention - show the user
            # every VIOS possible with state/health info so the user can
            # get a handle on what is 'going_on'. The question is whether this
            # is needed/desired for the custom SCGs with auto_add_vios True?
            scg.vioservers = scg_vioses
            scg.save(context=context, transaction=transaction)
            LOG.info(_("Performed save on updated storage connectivity group "
                       "'%(scg)s' with VIOS list: %(vioses)s.") %
                     dict(scg=scg_name, vioses=scg.vios_ids))
            return True
        else:
            LOG.debug("SCG '%s' not changed." % scg_name)
            return False

    def get_host_topology_from_db(self, port_tag=None,
                                  provider_check_type=None):
        """
        Build a list of host topology entries based on the PVC database only.
        The logic here must align partially with that of the SCG DOM in
        to_dict_with_conn() for optional parameters passed in to this method.
        e.g. If a port tag is passed, then computed connectivity information
        that is added to the topology should be handled in the same fashion
        as for an SCG with that same port tag. If this class instance has
        a host_name attribute that is None, then the topology is built for
        all registered hosts.
        :param port_tag:  If passed, then use the port_tag to further filter
                    the counts of ready ports, and thus storage-ready vioses.
        :param provider_check_type: If not None, then VIOSes will NOT be
                    considered
                    'storage-ready' if the backend provider is in error
                    state. A matching option has been added to the REST API
                    (host-storage-topologies) so that the console can get
                    'storage-ready' values that are consistent with values
                    obtained from the storage-connectivity-groups API.
                    Allowable values for this option are:

                        None - don't make VIOS storage readiness depend on
                               provider readiness. This is backwards compatible
                               with how the API worked before.
                        'cluster' - If a VIOS is a member of a cluster, then
                               that cluster must be not in error state for
                               the VIOS to be storage-ready.
                        'external' - There must be at least one registered SAN
                               provider not in error state for the VIOS to be
                               considered storage ready. This aligns with
                               NPIV-only connectivity.
                        'any' - If either the cluster provider or any SAN
                               provider is not in error state, then the VIOS
                               can be storage-ready, provided it passes the
                               other checks.
        :returns:   A python data structure that can be converted to a json
                    REST API response for storage topology information.
        """
        # Initialize empty list
        host_list = []

        # Get storage provider information to correlate cluster providers
        # and to evaluate storage provider readiness.
        cinder_api = pvc_cinder.PowerVCVolumeAPI()
        stg_providers = {}
        try:
            stg_providers = cinder_api.get_stg_provider_info(
                self.context, exclude_in_error=True)
        except Exception as ex:
            # log and continue
            LOG.exception(ex)

        host_set = set()  # set of hosts already seen
        hostfact = compute_dom.ManagedHostFactory.get_factory()
        # Get the ResourceTransaction (cf. DB2 session) for the nova db
        txn = hostfact.get_resource_transaction()

        with txn.begin():

            managed_hosts = hostfact.find_all_hosts(self.context)
            # Get the list of hosts from the nova database
            for host in managed_hosts:
                managed_hostname = host.host_name
                if self.host_name and managed_hostname != self.host_name:
                    LOG.debug("Skip filtered host %s." % host)
                    continue
                if managed_hostname in host_set:
                    LOG.debug("Already processed host '%s'" % host)
                    continue

                host_dict = {}
                host_set.add(managed_hostname)
                host_dict['hmc_uuid'] = host.hmc_uuids
                host_dict['name'] = managed_hostname
                LOG.debug('host name = %s' % managed_hostname)
                host_dict['port_ready_count'] = 0

                # Initialize the vios list to empty.
                vios_list = []
                host_dict['vios_ready_count'] = 0
                vio_servers = host.vioservers
                # Loop over known vioses
                for vio_server in vio_servers:
                    fcport_factory = self.dom.FcPortFactory.get_factory()
                    # get vios sever object from storage dom
                    vios_id = vio_server.id
                    vios_fact = self.dom.VioServerFactory.get_factory()
                    vios = vios_fact.find_vios_by_id(self.context,
                                                     vios_id,
                                                     txn)
                    vios_dict = self._get_dbvios_dict(
                        vios, stg_providers, txn, fcport_factory,
                        provider_check_type, port_tag, cinder_api)
                    if vios_dict['storage_ready']:
                        host_dict['vios_ready_count'] += 1

                    host_dict['port_ready_count'] += \
                        vios_dict['port_ready_count']
                    # Add the vios object to the vios list in all cases
                    # since it was decided that users would prefer to see
                    # the VIOS and address the issues it may have.
                    vios_list.append(vios_dict)
                # Add the vios list to the host object
                host_dict['vios_list'] = vios_list
                LOG.debug("host_dict = %s" % host_dict)
                # Add the host object to the host list.
                host_list.append(host_dict)
        return host_list

    def _get_dbvios_dict(self, vio_server, stg_providers, transaction,
                         fcport_factory, provider_check_type, port_tag,
                         cinder_api, fab_req='any'):
        """
        Helper method to build a vios dictionary for topology data from
        the DB. vio_server is the DB VIOS DTO and stg_providers is a dictionary
        of providers that comes from the volume/cinder.py helpers for use
        when SSPs are supported. Here is an example VIOS dict that is built:
            {
             "id": "789522X_067A34B##1"
             "name": "acme109",
             "storage_ready": true,
             "lpar_id": 1,
             "total_fcport_count": 1,
             "rmc_state": "active",
             "port_ready_count": 1,
             "state": "running",
             "fcport_list": [
              {
               "status": "OK",
               "fabric": "None",
               "vio_server": "acme109",
               "available_connections": 64,
               "enabled": true,
               "adapter_id": "U78AE.001.WZS02BR-P1-C19-L1",
               "wwpn": "21000024FF4182D8",
               "id": "1aU78AE.001.WZS02BR-P1-C19-L1-T1",
               "name": "fcs0"
              },...
             ],
            }
        The most recent adds are for storage_ready, total_fcport_count, and
        port_ready_count properties which are aids to consumers that are in
        a flow where they don't need to traverse the ports and port properties.
        We do the aggregate computations for them. The 'storage_ready' value
        of true does not predict a successful deployment using this VIOS as
        there are potentially many other factors (even VIOS related) that could
        impact deployment and/or relocation.
        """
        # Build the vios object with elements.
        vios_dict = {'lpar_id': vio_server.lpar_id,
                     'name': vio_server.lpar_name, 'id': vio_server.id,
                     'state': vio_server.state,
                     'rmc_state': vio_server.rmc_state,
                     'port_ready_counts': {_NPIV_: 0, _VSCSI_: 0},
                     'connectivity_types': []
                     }
        conn_types_set = set()
        applicable_providers = []
        provider_ready = True  # default to true
        # Cluster provider props if applicable
        # Don't do any cluster provider checking if 'skip' flag is passed.
        if ((provider_check_type != 'skip') and
                (vio_server.cluster_provider_name is not None) and
                (vio_server.cluster_provider_name != '')):
            prov_name = vio_server.cluster_provider_name
            vios_dict['cluster_provider_name'] = prov_name
            if prov_name in stg_providers:
                p = stg_providers[prov_name]
                vios_dict['cluster_provider_display_name'] =\
                    p['host_display_name']
                applicable_providers.append(prov_name)
                vios_dict['cluster_provider_state'] = p['backend_state']
                vios_dict['port_ready_counts'][_SSP_] = 0
            # Note that it's not really possible for the vios to be related
            # to an unregistered cluster provider since the association is
            # to a table of registered providers.
            else:
                try:
                    # The list we originally asked for excluded those in error
                    # state. Get the in-error cluster provider record now from
                    # the cache.
                    p = cinder_api.get_stg_provider_info(
                        self.context, storage_hostname=prov_name,
                        use_cache=True)
                    vios_dict['cluster_provider_state'] = p['backend_state']
                    vios_dict['cluster_provider_display_name'] =\
                        p['host_display_name']
                except Exception as ex:
                    # OK, maybe it wasn't in-error. It could have been deleted
                    # in a timing window and we don't want to fail the API in
                    # this case.
                    LOG.warn(_("Could not retrieve VIOS associated cluster "
                               "provider '%(name)s': %(ex)s") %
                             dict(name=prov_name, ex=ex))
                    vios_dict['cluster_provider_state'] =\
                        'unregistered/unknown'
                if provider_check_type == 'cluster':
                    provider_ready = False

        # Get the FC ports related to this vios from the DB.
        vios_fcports = fcport_factory.find_all_fcports_by_vios(
            self.context, vio_server.id, transaction=transaction)

        LOG.debug("len(vios_fcports) = %s " % len(vios_fcports))
        # Initialize the fcport list to empty.
        fcport_list = []
        # ok_ports will track the number of ports that are deemed 'ready' for
        # use within an SCG that may contain this VIOS. The optional passed
        # port_tag can further limit the 'ready' ports.
        ok_ports = 0
        if vios_fcports:
            for fcport in vios_fcports:
                # Build the fc port object with elements.
                port = self.dom.fcport_to_dict(fcport)
                LOG.debug("fcport_dict = %s" % port)
                port['connectivity'] = set()
                fcport_list.append(port)
                # Track number of 'good' ports
                if (port['status'].startswith("OK") and
                        port['enabled'] and port['fabric'] and
                        port['fabric'] != "None" and
                        port['wwpn'] and len(port['wwpn']) and
                        (port_tag is None or  # 'port_tag' not in port or
                         ('port_tag' in port and
                          port_tag == port['port_tag']))):
                    # Port satisfies requirements. Note that we do not check
                    # any include_untagged setting here. Untagged ports are
                    # always included so as not to cause unnecessary warnings
                    # on console screens during SCG creation. To honor
                    # include_untagged, the host-storage-topologies API would
                    # need to be enhanced to take that as a query option.
                    # The complexity is probably not warranted.
                    ok_ports = ok_ports + 1
                    fab_ok = False
                    if fab_req == 'any' or fab_req == port['fabric'] or\
                            fab_req.startswith('dual'):
                        fab_ok = True
                    if (port['fc_protocol'] == 'any' or
                            port['fc_protocol'] == 'npiv') and fab_ok:
                        port['connectivity'].add(_NPIV_)
                        vios_dict['port_ready_counts'][_NPIV_] += 1
                        conn_types_set.add(_NPIV_)
                # The vscsi ports only need to be enabled for vscsi. At this
                # time they are not filtered out by status or port tag.
                if port['enabled'] and (port['fc_protocol'] == 'any' or
                                        port['fc_protocol'] == 'vscsi'):
                    port['connectivity'].add(_VSCSI_)
                    vios_dict['port_ready_counts'][_VSCSI_] += 1
                    conn_types_set.add(_VSCSI_)
                if _SSP_ in vios_dict['port_ready_counts']:
                    port['connectivity'].add(_SSP_)
                    vios_dict['port_ready_counts'][_SSP_] += 1
                    conn_types_set.add(_SSP_)
        else:
            LOG.debug("This VIOS SCG entry is not reporting any "
                      "Fibre Channel ports: %s." % vios_dict)
        vios_dict['port_ready_count'] = ok_ports
        vios_dict['total_fcport_count'] = len(vios_fcports)

        if conn_types_set:
            fc_providers = [x for x, y in stg_providers.iteritems()
                            if y['stg_type'] == 'fc']
            if not fc_providers and provider_check_type == 'external':
                provider_ready = False
            else:
                applicable_providers.extend(fc_providers)
                if provider_check_type == 'any' and not applicable_providers:
                    provider_ready = False
            LOG.debug("vios_id=%s, applicable_providers=%s" %
                      (vios_dict['id'], applicable_providers))

        # Track whether this VIOS is storage-ready. This needs to be
        # clearly defined for a customer, yet the implementation needs to
        # retain the responsibility of making this true/false determination.
        # I.e. it is not up to consumers/exploiters to assume 'readiness' based
        # on documented property values - the checks here may change and it
        # is not a defect if they do, though docs may need to be updated.
        if vio_server.state == 'running'\
                and vio_server.is_rmc_state_ok()\
                and ok_ports > 0 and provider_ready:
            vios_dict['storage_ready'] = True
        else:
            vios_dict['storage_ready'] = False

        # Add the fcport list to the vios object.
        vios_dict['fcport_list'] = fcport_list
        vios_dict['connectivity_types'] = list(conn_types_set)
        LOG.debug("VIOS %s connectivity_types: %s" %
                  (vio_server.lpar_name, vios_dict['connectivity_types']))
        return vios_dict
