# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
from oslo_log import log as logging

import six

from powervc_nova import _
from powervc_nova.storage import exception as stgex
import powervc_nova.storage.topo_common as topo_cmn
import pypowervm.wrappers.base_partition as pvm_base
import pypowervm.wrappers.managed_system as pvm_ms
import pypowervm.wrappers.virtual_io_server as pvm_vios
from pypowervm import const as pvm_const


LOG = logging.getLogger(__name__)


class StorageTopologyPvm(topo_cmn.StorageTopologyCommon):
    """ NovaLink-managed compute node specific collection """

    def __init__(self, host_name=None, context=None, adapter=None,
                 host_uuid=None):
        """
        StorageTopologyPvm constructor
        :param host_name: The server host_name
        :param context: nova context object
        :param adapter: pypowervm.adapter.Adapter object to be used to
                interact with PVM partition
        :param host_uuid: The uuid of the server
        """
        super(StorageTopologyPvm, self).__init__(host_name, context)
        self.host_type = 'pvm'
        self.adapter = adapter
        if not host_uuid:
            m = (_("A valid host_uuid was not provided."))
            LOG.error(m)
            raise stgex.IBMPowerVCStorageError(error=m)
        self.host_uuid = host_uuid
        LOG.debug("Initialized PVM topology collector: %s" % host_name)

    def get_live_topology(self, skip_cluster=True, max_cache_age=-1):
        """
        Builds a host topology for the server by getting information
        using pypowervm adapter.
        """
        LOG.debug("Begin get_live_topology...")
        # skip-cluster is true till spp support for pvm comes in
        topo_factory = self.dom.HostStorageTopologyFactory.get_factory()
        static_topo = topo_factory.get_static_topology()
        topology = {'hosts': {}, 'skip-update-clusters': skip_cluster,
                    'vios_by_id': {},
                    'vios_by_uuid': {},
                    'vios_num_trusted': 0,
                    'seq': static_topo.bump_topology_sequence(),
                    'static_topo': static_topo,
                    'max_cache_age': max_cache_age}
        host_dict = {'name': self.host_name,
                     'uuid': self.host_uuid}
        topology['hosts'][self.host_name] = host_dict

        self.collect_vios_info(topology)
        LOG.debug("End get_live_topology()")
        return topology

    def collect_vios_info(self, topology):
        """
        Collect vios and storage information of all vios on this
        sever.
        """
        vios_wrap = None
        # get list of vios
        try:
            # When use_mapped_vfc_scheduling is set the number of available
            # mappings for a npiv port is based on vfc_mappings. The default
            # case uses the number of aports which is the number of logged
            # in clients
            if self.use_mapped_vfc_scheduling:
                xag = [pvm_const.XAG.VIO_STOR,
                       pvm_const.XAG.VIO_FMAP]
            else:
                xag = [pvm_const.XAG.VIO_STOR]

            vios_resp = self.adapter.read(pvm_ms.System.schema_type,
                                          root_id=self.host_uuid,
                                          child_type=pvm_vios.VIOS.schema_type,
                                          xag=xag)
            vios_wrap = pvm_vios.VIOS.wrap(vios_resp)
        except Exception as ex:
            LOG.error(_("Error getting vios information during topology "
                        "collection: %(ex)s") % {'ex': six.text_type(ex)})
            raise ex

        for vio_entry in vios_wrap:
            vios_info = self.get_vios_info(vio_entry)
            vios_info['id'] = vio_entry.uuid
            topology['vios_by_id'][vios_info['id']] = vios_info
            topology['vios_by_uuid'][vios_info['uuid']] = vios_info
            if 'state' in vios_info and vios_info['state'] == "running" and\
                    'rmc_state' in vios_info and\
                    vios_info['rmc_state'] == "active":
                topology['vios_num_trusted'] += 1
                vios_info['trusted'] = True
            else:
                # Some VIOS information like cluster membership and FC ports
                # should not be trusted if state info is not all OK.
                vios_info['trusted'] = False
            LOG.debug("Added VIOS dict entry to topology: %s" % vios_info)

        LOG.debug("Topology returned: %(topo)s" % {'topo': topology})
        return topology

    def get_vios_info(self, vio_entry):
        """
        Parse single vios information.
        """
        vios_info = {}
        if not vio_entry:
            LOG.warn(_("No Virtual I/O Servers returned"))
            return None

        vios_info['name'] = vio_entry.name
        vios_info['uuid'] = vio_entry.uuid
        vios_info['state'] = vio_entry.state
        vios_info['rmc_state'] = vio_entry.rmc_state
        vios_info['lpar_id'] = vio_entry.id

        vios_info['fcports'] = self.parse_fcports(vio_entry)
        return vios_info

    def parse_fcports(self, vio_entry):
        """
        Parse the fc port information for each port.
        """
        fcports = {}

        if not vio_entry.pfc_ports:
            LOG.warn(_("No fc ports found for VIOS %(vios)s") %
                     dict(vios=vio_entry.uuid))
            return None

        adapter_id = None
        # get the adapter id
        io_slots = vio_entry.io_config.io_slots
        for io_slot in io_slots:
            desc = io_slot.description
            if not io_slot.io_adapter:
                LOG.debug("The io_slot has no io_adapter: %s, description=%s"
                          % (io_slot, desc))
                continue
            if not isinstance(io_slot.io_adapter, pvm_base.PhysFCAdapter):
                LOG.debug("%s not an fc adapter" % io_slot.io_adapter.dev_name)
                continue
            LOG.debug("Processing FC io_adapter: %s" %
                      io_slot.io_adapter.dev_name)
            is_npiv_adapter = self.is_npiv_adapter_description(desc)
            fc_adapter = io_slot.io_adapter
            adapter_id = fc_adapter.dev_name if fc_adapter.dev_name\
                else "[unknown]"
            for fcport in fc_adapter.fc_ports:
                fcport_dict = dict(id=fcport.udid,
                                   name=fcport.name,
                                   adapter_id=adapter_id,
                                   wwpn=fcport.wwpn,
                                   status=topo_cmn.PORT_STATUS_UNKNOWN,
                                   npiv_status=topo_cmn.PORT_STATUS_UNKNOWN
                                   )
                if fcport.wwpn:
                    if fcport.npiv_total_ports:
                        fcport_dict['total_vports'] = fcport.npiv_total_ports
                        fcport_dict['available_vports'] =\
                            fcport.npiv_available_ports
                        fcport_dict['status'] = topo_cmn.PORT_STATUS_OK
                        if fcport_dict['available_vports'] > 0:
                            # put number of VFCs left on the port for OK status
                            # this info can be split out and used for API info.
                            fcport_dict['npiv_status'] =\
                                topo_cmn.PORT_STATUS_OK
                        else:
                            fcport_dict['npiv_status'] =\
                                topo_cmn.PORT_STATUS_FULL
                    else:
                        LOG.debug("Not a currently supported FC Port since "
                                  "TotalPorts are missing from the element.")
                        if is_npiv_adapter:
                            fcport_dict['npiv_status'] =\
                                topo_cmn.PORT_STATUS_OFFLINE
                            fcport_dict['status'] =\
                                topo_cmn.PORT_STATUS_OFFLINE
                        else:
                            fcport_dict['npiv_status'] =\
                                topo_cmn.PORT_NPIV_STATUS_UNS_ADAPTER
                            fcport_dict['status'] = topo_cmn.PORT_STATUS_OK

                fcports[fcport_dict['id']] = fcport_dict

        # check the number of vfc mappings if configured
        if self.use_mapped_vfc_scheduling:
            # get the vfc mappings only if the use_mapped_vfc_scheduling is
            # set.
            for vfc_map in vio_entry.vfc_mappings:
                port_id = vfc_map.backing_port.udid if vfc_map.backing_port\
                    else None
                if not port_id:
                    continue
                if 'mapped_vfcs' in fcports[port_id]:
                    fcports[port_id]['mapped_vfcs'] += 1
                else:
                    fcports[port_id]['mapped_vfcs'] = 1

            for port_id in fcports:
                port_dict = fcports[port_id]
                if 'total_vports' in port_dict:
                    num_vports = port_dict['total_vports'] -\
                        topo_cmn.PORT_BUFFER
                    if 'mapped_vfcs' in port_dict and\
                            port_dict['mapped_vfcs'] >= num_vports:
                        LOG.debug("For port %s, using mapped vFC scheduling, "
                                  "so set npiv_status to Full." % port_dict)
                        port_dict['npiv_status'] = topo_cmn.PORT_STATUS_FULL
                else:
                    LOG.debug("There are no total_vports for %s" % port_id)

        LOG.debug("FCport info for vios %(vios_id)s is %(fcports)s" %
                  dict(vios_id=vio_entry.uuid, fcports=fcports))
        return fcports
