# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import six
import copy
import webob
from nova import db
from nova.volume import cinder
from powervc_nova.objects import block_device as pvc_blkdev
from powervc_nova.compute import populate_admin_context
from oslo_serialization import jsonutils as json
from oslo_log import log as logging

LOG = logging.getLogger(__name__)


class VolumeInstanceMapping():
    """
    Internal helper method for Volume_Instance_Mapping API.
    Used to retrieve relevant volume connection information and to filter
    the data to be returned.
    """
    def _dict_key_filter(self, obj, obj_filter):
        """
        Input: obj - A dictionary. Or list of dictionaries.
               obj_filter - A list of key values to keep
        Returns - Will retain all values from the given key obj_filter.
                  Deletes the rest. returns a new dict obj.
        """
        if isinstance(obj, dict):
            retdict = {}
            for key, value in obj.iteritems():
                if key in obj_filter and isinstance(value, (dict, list)):
                    child = self._dict_key_filter(value, obj_filter)
                    retdict[key] = child if child else value
                elif key in obj_filter:
                    retdict[key] = copy.deepcopy(value)
            return retdict if retdict else None
        elif isinstance(obj, list):
            retlist = []
            for value in obj:
                if isinstance(value, (dict, list)):
                    child = self._dict_key_filter(value, obj_filter)
                    if child:
                        retlist.append(child)
                else:
                    retlist.append(value)
            return retlist if retlist else None
        else:
            return None

    def _dict_key_removal(self, obj, obj_filter, check=True):
        """
        Input: obj - A dictionary. Or list of dictionaries.
               obj_filter - A list of key values to remove.
        Returns - Will remove all values from the given key obj_filter.
                 returns the same dict obj passed in.
        """
        if isinstance(obj, dict):
            key_found = False
            for key, value in list(obj.iteritems()):
                if key in obj_filter and isinstance(value, (dict, list)):
                    child = self._dict_key_removal(value, obj_filter, False)
                    key_found = True
                    if not child:
                        obj.pop(key, None)
                elif key in obj_filter:
                    key_found = True
                    obj.pop(key, None)
            return obj if key_found or check else None
        elif isinstance(obj, list):
            for value in list(obj):
                if isinstance(value, (dict, list)):
                    child = self._dict_key_removal(value, obj_filter, check)
                    if not child:
                        obj.remove(value)
            return obj if obj else None
        else:
            return obj

    def _filter_volume(self, context, vol_dict, vol_filter, cinderclient):
        volume_id = vol_dict.get('id')
        urlpath = "/volumes/%s/restricted-metadata" % volume_id
        r_metadata = cinderclient.client.get(urlpath)
        r_metadata = r_metadata[1]['metadata'] \
            if len(r_metadata) == 2 else None
        for data in r_metadata:
            if data in ['lu_udid', 'lu_name', 'ssp_udid',
                        'vdisk_id', 'vdisk_name']:
                vol_dict['metadata'][data] = r_metadata[data]
        filtered_vol = self._dict_key_filter(vol_dict, vol_filter)
        return filtered_vol

    def _filter_fcs(self, context, txn, fcpport_factory,
                    fcs_dict, conn_info, fcs_filter):
        fcs_dict_temp = {}
        connector = conn_info.get('volume_connector', {})\
            .get('connector', {})
        connection_type = connector.get('connection_type', []) \
            if connector else []
        if 'npiv' in connection_type:
            phy_wwpns = connector.get('phy_to_virt_initiators', {}).keys()
        elif 'pv_vscsi' in connection_type:
            phy_wwpns = []
            vscsi_target_lun = conn_info.get('vscsi_target_lun', {})
            [phy_wwpns.extend(x.get('vios_wwpn')) for x in vscsi_target_lun]
        else:
            # ssp does not have a connector, no fcs found
            return {}
        if len(phy_wwpns) == 0:
            LOG.debug(_("Unable to retrieve Fiber Channel Ports, no phy_wwpns "
                        "found! Skipping retrieval of Virtual I/O Server"))
            return {}
        for p_wwpn in phy_wwpns:
            fcs = fcpport_factory.find_all_fcports(context,
                                                   filters={'wwpn': p_wwpn},
                                                   transaction=txn)
            fcs = dict((key, value) for key, value in fcs[0].iteritems())
            fcs_dict_temp[p_wwpn] = self._dict_key_filter(fcs, fcs_filter)
        fcs_dict.update(fcs_dict_temp)
        return fcs_dict

    def _filter_vios(self, context, txn, vios_factory, vios_dict,
                     fcs_dict, vios_filter, cluster_name=None):
        vios_dict_temp = {'ssp_cluster': {}}
        # find vios info by ssp cluster
        if cluster_name and cluster_name \
                not in vios_dict.get('ssp_cluster', {}):

            vioses = vios_factory.\
                find_all_vioses_by_cluster(context, cluster_name, txn)
            vios_list = [dict((key, value) for key, value in vios.iteritems())
                         for vios in vioses]

            for vios in vios_list:
                cluster_provider = vios.pop('cluster_provider', {})
                # converting object to dict
                cluster_provider = dict((key, value) for key, value in
                                        cluster_provider.iteritems())
                cluster_provider = self._dict_key_filter(cluster_provider,
                                                         vios_filter)
                cluster_prov_name = cluster_provider.get('host_name')
                if cluster_prov_name not in vios_dict.get('ssp_cluster', {}):
                    vios_dict['ssp_cluster'][cluster_prov_name] = \
                        cluster_provider
            vios_fil_list = self._dict_key_filter(vios_list, vios_filter)
            vios_dict_temp = dict((vios['id'], vios) for vios in vios_fil_list)
            vios_dict.update(vios_dict_temp)
            return vios_dict

        # find vios info by fcs
        for fcs_key, fcs_value in fcs_dict.iteritems():
            if fcs_value.get('vios_id') and \
                    fcs_value.get('vios_id') not in vios_dict.keys():
                vios = vios_factory.find_vios_by_id(context,
                                                    fcs_value.get('vios_id'),
                                                    txn)
                vios = dict((key, value) for key, value in vios.iteritems())
                vios_dict[fcs_value.get('vios_id')] = \
                    self._dict_key_filter(vios, vios_filter)
        return vios_dict

    def get_volume_info(self, context, volume_id):
        cclient = cinder.cinderclient(context)
        # Avoiding some circular import
        from powervc_nova.volume import cinder as pvc_cinder
        from powervc_nova.objects.storage import dom as storage_dom
        fcpport_factory = storage_dom.FcPortFactory.get_factory()
        vios_factory = storage_dom.VioServerFactory.get_factory()
        txn = fcpport_factory.get_resource_transaction()
        fabricapi = pvc_cinder.PowerVCFabricAPI()
        volumeapi = pvc_cinder.PowerVCVolumeAPI()
        volume_info = {}
        # Obtaining detailed volume information, so as to be selective
        # in the filtering process
        volume = cclient.volumes.get(volume_id)
        vol_dict = dict(volume.__dict__.iteritems())
        vol_filter = ['attachments', 'attachment_id', 'name', 'id',
                      'os-vol-host-attr:host',
                      'os-vol-tenant-attr:tenant_id', 'server_id',
                      'lu_udid', 'lu_name', 'ssp_udid', 'vdisk_id',
                      'vdisk_name', 'backend_id', 'backend_type',
                      'default_volume_type', 'hmc_uuids', 'host',
                      'host_display_name', 'storage_hostname',
                      'stg_provider_info', 'metadata', 'stg_type',
                      'volume_wwn']
        provider = volumeapi.get_stg_provider_info(context,
                                                   volume_id=volume_id)
        if provider:
            vol_dict['stg_provider_info'] = provider
        volume_info['volume_info'] = self._filter_volume(context, vol_dict,
                                                         vol_filter, cclient)
        # Obtaining detailed volume connections information
        # PowerVC currently overrides community to retrieve
        # all bdms of shared volumes
        bdms_filter = ['connection_info', 'data', 'vscsi_target_lun',
                       'boot_index', 'volume_connector', 'connector',
                       'connection_type', 'target_lun', 'target_wwn', 'vendor',
                       'volume_id', 'wwpns', 'host', 'initiator_target_map',
                       'ip', 'instance_uuid', 'target_UDID', 'target_UID',
                       'vios_name', 'vios_wwpn', 'vdisk_name', 'vdisk_id',
                       'volume_wwn', 'target_lun', 'phy_to_virt_initiators',
                       'driver_volume_type', 'UniqueDeviceID',
                       'connection-type']
        bdms = pvc_blkdev.PowerVCBlockDeviceMappingList.\
            get_by_volume_id(context, volume_id)
        if len(bdms) == 0:
            msg = (_("Volume_instance_mapping api is only applicable to"
                     "attached volumes"))
            raise webob.exc.HTTPBadRequest(explanation=six.text_type(msg))
        bdms = [dict((key, value) for key, value in bdm.iteritems())
                for bdm in bdms]
        for bdm in bdms:
            bdm['connection_info'] = json.loads(bdm.get('connection_info'))
        bdms = self._dict_key_filter(bdms, bdms_filter)
        volume_info['block_device_mapping'] = bdms
        vios_dict = {}
        fcs_dict = {}
        zoning_info = {}
        vios_filter = ['host_name', 'id',
                       'lpar_id', 'lpar_name']
        metadata = volume_info.get('volume_info', {}).get('metadata', {})
        cluster_provider_name = None
        if 'ssp_udid' in metadata:
            volume_info['ssp_info'] = {}
            vios_dict['ssp_cluster'] = {}
            vios_filter.extend(['cluster_provider', 'cluster_provider_name',
                                'hmc_uuids', 'host_display_name', 'host_name',
                                'host_type', 'volume_pool_name'])
            cluster_provider_name = \
                (volume_info.get('volume_info', {})
                 .get('stg_provider_info', {}).get('host'))

        fcs_filter = ['adapter_id', 'id', 'name', 'vios_id']
        for bdm in bdms:
            conn_info = bdm.get('connection_info', {})
            vol_connector = bdm.get('connection_info', {}).\
                get('volume_connector', {})
            with txn.begin():
                # filter to fcs and vios to only relevant data needed
                try:
                    fcs_dict = self._filter_fcs(context, txn, fcpport_factory,
                                                fcs_dict, conn_info,
                                                fcs_filter)
                    # perform a vios lookup only if there are fiber channel
                    # ports or if it is a ssp.
                    if len(fcs_dict) > 0 or 'ssp_udid' in metadata:
                        vios_dict = \
                            self._filter_vios(
                                context, txn, vios_factory,
                                vios_dict, fcs_dict, vios_filter,
                                cluster_name=cluster_provider_name)
                except Exception as e:
                    msg = (_("Unable to obtain fiber channel port or Virtual "
                             "I/O server. Exception raised %s"
                             % six.text_type(e)))
                    LOG.warning(msg)
                # Moving cluster info into ssp_info

            # obtain detailed zoning information
            initiators = []
            connector = vol_connector.get('connector', {})
            # ssp connector is None setting to a empty list
            connection_type = connector.get('connection_type', []) \
                if connector else []
            # NEO does not have connection_type or volume_connector stored
            connection_type = \
                ([conn_info.get('data', {}).get('connection-type', [])] if
                 not connection_type else connection_type)
            if 'npiv' in connection_type:
                itl_map = (conn_info.get('data', {}).
                           get('initiator_target_map', {}))
                initiators = itl_map.keys()
            # else pv_vscsi
            elif 'pv_vscsi' in connection_type:
                vscsi_target_lun = conn_info.get('vscsi_target_lun', {})
                [initiators.extend(x.get('vios_wwpn'))
                 for x in vscsi_target_lun]

            instance_uuid = bdm.get('instance_uuid')
            # Zoning info is gathered from a query to the fabrics registered
            if initiators:
                zones = fabricapi.map_zones_to_fabrics(context,
                                                       initiators=initiators)
                (zones.pop(x) for x in zones.keys() if zones.get(x) is None)
                attachments = volume_info.get('volume_info').get('attachments')
                for attach_inst_id in attachments:
                    if instance_uuid == attach_inst_id and zones:
                        # [Not sure what the following code is accomplishing
                        #  since 'attachements' are removed from volume_info
                        #  that is returned later]
                        zoning_info['zoning_info'] = zones
                        attachments[attach_inst_id].update(zoning_info)
        if 'ssp_udid' in metadata:
            volume_info['ssp_info'].\
                update(vios_dict.pop('ssp_cluster'))
        volume_info['fcs_info'] = fcs_dict
        volume_info['vios_info'] = vios_dict
        remove_keys = ['volume_info', 'attachments', 'id']
        volume_info = self._dict_key_removal(volume_info, remove_keys)
        return volume_info

    def get_instance_info(self, context, instance_uuid):
        instance = db.instance_get_by_uuid(context, instance_uuid)
        instance = dict((key, value) for key, value in instance.iteritems())
        instance['system_metadata'] = \
            [dict((key, value) for key, value in mdata.iteritems())
             for mdata in instance['system_metadata']]
        bdms = pvc_blkdev.PowerVCBlockDeviceMappingList.\
            get_by_instance_uuid(context, instance_uuid)
        bdms = [dict((key, value) for key, value in bdm.iteritems())
                for bdm in bdms]
        vol_ids = [x.get('volume_id') for x in bdms]
        vols = {}
        for volume_id in vol_ids:
            volume_info = self.get_volume_info(context, volume_id)
            vols[volume_id] = volume_info
        return vols

    def _fix_context(self, context):
        """
        In case the context is None or is the bare admin context, we use
        a static conductor facility here to generate a new context with
        the service catalog filled in.
        """
        try:
            if not context or not context.service_catalog:
                new_context = populate_admin_context()
                if new_context:
                    LOG.debug("Return enhanced context: %s" %
                              new_context.to_dict())
                    return new_context
        except Exception as ex:
            LOG.exception(ex)
            LOG.info(_("Ignore context gen exception and continue with "
                       "existing context."))
        LOG.debug('Context unchanged.')
        return context
