# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import copy
import six

from nova import objects
from nova import rpc
from oslo_config import cfg
from oslo_log import log as logging

from powervc_nova import _
from powervc_nova.objects.compute import dom as compute_dom
from powervc_nova.virt.ibmpowervm.common import constants
from powervc_nova.virt.ibmpowervm.common import exception

CONF = cfg.CONF

LOG = logging.getLogger(__name__)


def get_pool_for_host(context, hostname):
    """Gets the enterprise pool that the host belongs to.

    :param context: The security context for the query
    :param hostname: The string hostname corresponding to the host of interest
    :returns: powervc_nova.objects.compute.dom.EnterprisePool object
              representing the pool that the host belongs to. If the host does
              not belong to any pools, then None is returned.
    """
    pool_fact = compute_dom.EnterprisePoolFactory.get_factory()
    pools = pool_fact.find_all_enterprise_pools_by_host(context, hostname)

    pool = None
    if not pools:
        LOG.debug('Host %s is not a member of any enterprise pools.' %
                  hostname)
        return None
    elif len(pools) > 1:
        LOG.error(_("More than one enterprise pool found for host %(host)s: "
                    "%(pools)s.") % {'host': hostname, 'pools': pools})
        pool_ids = [pool.uuid for pool in pools]
        raise exception.IBMPowerVMHostInMultiplePools(host=hostname,
                                                      pools=pool_ids)
    elif len(pools) == 1:
        pool = pools[0]
        LOG.debug('Host %s is a member of an enterprise pool: %s.' %
                  (hostname, pool.uuid))

    return pool


def get_pool_member_uuids(context, hostname):
    """Gets the enterprise pool and member UUIDs for the hostname.

    :param context: The security context for the query
    :param hostname: The string hostname corresponding to the host of interest
    :returns: Tuple of strings: (pool_uuid, pool_member_uuid), where pool_uuid
              is the UUID of the pool that hostname belongs to and
              pool_member_uuid is the UUID of the pool member corresponding to
              the hostname. Returns None if no pool or pool member is found.
    """
    pool = get_pool_for_host(context, hostname)
    if not pool:
        return None

    member_uuid = pool.host_names.get(hostname)
    if member_uuid is None:
        LOG.error(_("Host %(host)s exists in pool %(pool)s with no member "
                    "UUID.") % {'host': hostname, 'pool': pool.uuid})
        return None

    return pool.uuid, member_uuid


def create_pool(context, pool_dict, members):
    """Creates enterprise pool in the database.

    :param context: The security context for the query
    :param pool_dict: Dictionary representing the updated pool. Example:
        {'uuid': <pool_uuid> (str),
         'name': <pool_name> (str),
         'master_hmc_mtmsn': <master_hmc_mtmsn> (str),
         'available_mobile_cores': <available_mobile_cores> (int),
         'total_mobile_cores': <total_mobile_cores> (int),
         'compliance_state': <compliance_state> (int)}
    :param members: Dictionary with keys of string hostnames and values of
                    string member UUIDs representing the pool members.
    """
    dbcfg = copy.deepcopy(pool_dict)
    dbcfg['host_names'] = members

    pool_fact = compute_dom.EnterprisePoolFactory.get_factory()
    pool_fact.construct_enterprise_pool(context, auto_create=True, **dbcfg)


def update_pool(pool, pool_dict, members_to_add):
    """Updates an enterprise pool in the database.

    :param context: The security context for the query
    :param pool: powervc_nova.objects.compute.dom.EnterprisePool object
    :param pool_dict: Dictionary representing the updated pool. Example:
        {'uuid': <pool_uuid> (str),
         'name': <pool_name> (str),
         'master_hmc_mtmsn': <master_hmc_mtmsn> (str),
         'available_mobile_cores': <available_mobile_cores> (int),
         'total_mobile_cores': <total_mobile_cores> (int),
         'compliance_state': <compliance_state> (int)}
    :param members_to_add: Dictionary with keys of string hostnames and values
                           of string member UUIDs representing the pool members
                           that should be added as part of the update.
    """
    pool.name = pool_dict['name']
    pool.master_hmc_mtmsn = pool_dict['master_hmc_mtmsn']
    pool.available_mobile_cores = pool_dict['available_mobile_cores']
    pool.total_mobile_cores = pool_dict['total_mobile_cores']
    pool.compliance_state = pool_dict['compliance_state']

    if members_to_add:
        new_members = pool.host_names
        new_members.update(members_to_add)
        pool.host_names = new_members

    pool.save()


def remove_member(context, pool, hostname):
    """Removes member from an enterprise pool in the database.

    :param context: The security context for the query
    :param pool: powervc_nova.objects.compute.dom.EnterprisePool object
    :param hostname: The string hostname of the member to remove from the pool
    """
    if hostname not in pool.host_names:
        return
    new_hosts = pool.host_names
    del new_hosts[hostname]
    pool.host_names = new_hosts
    if pool.host_names:
        pool.save()
    else:
        # There are no more hosts managed by PowerVC that are in the pool, so
        # delete it from our database
        pool_fact = compute_dom.EnterprisePoolFactory.get_factory()
        pool_fact.delete_enterprise_pool(context, pool.uuid)


def sync_pool_for_host(context, pool_dict, pool_member_uuid, hostname):
    """Syncs the pool info corresponding to a specific host.

    Since this code is ran from each individual compute driver process, it will
    only sync membership for this specific host (that corresponds to hostname).
    It will also sync the overall pool properties, such as the available mobile
    cores, total mobile cores, etc., that aren't tied to a specific member.

    :param context: The security context for the query
    :param pool_dict: Dictionary representing the updated pool. Example:
        {'uuid': <pool_uuid> (str),
         'name': <pool_name> (str),
         'master_hmc_mtmsn': <master_hmc_mtmsn> (str),
         'available_mobile_cores': <available_mobile_cores> (int),
         'total_mobile_cores': <total_mobile_cores> (int),
         'compliance_state': <compliance_state> (int)}
    :param pool_member_uuid: The member UUID of the member that belongs to the
                             pool represented by pool_dict, which corresponds
                             to the host represented by hostname
    :param hostname: The hostname of interest
    """
    pool_from_hostname = get_pool_for_host(context, hostname)

    # Handle the case when PowerVM says no pool is found for this host
    if not pool_dict:
        # If there previously was a pool associated with this host, then
        # remove the host from the pool
        if pool_from_hostname:
            remove_member(context, pool_from_hostname, hostname)
        return

    # Find the pool in the DB corresponding to the pool UUID from PowerVM
    pool_fact = compute_dom.EnterprisePoolFactory.get_factory()
    pool_from_powervm_uuid = pool_fact.find_enterprise_pool_by_uuid(
        context,
        pool_dict['uuid'])

    pool_members = {hostname: pool_member_uuid}

    if pool_from_powervm_uuid is None:
        # PowerVC does not know about this pool
        if pool_from_hostname is None:
            # PowerVC does not have this host belonging to any pools
            LOG.info(_('There is no pool corresponding to the pool UUID '
                       '(%(pool_uuid)s) from PowerVM or corresponding to the '
                       'hostname (%(hostname)s).') %
                     {'pool_uuid': pool_dict['uuid'],
                      'hostname': hostname})
            # Therefore, create the new pool with this host as a member of it.
            create_pool(context, pool_dict, pool_members)
        else:
            # PowerVC has this host belonging to a pool
            LOG.info(_('There is no pool corresponding to the pool UUID '
                       '(%(pool_uuid)s) from PowerVM, but there is a pool '
                       '(%(other_pool_uuid)s) corresponding to the hostname '
                       '(%(hostname)s).') %
                     {'pool_uuid': pool_dict['uuid'],
                      'other_pool_uuid': pool_from_hostname.uuid,
                      'hostname': hostname})
            # Therefore, remove the host from the other pool and create the new
            # pool with this host as a member of it.
            remove_member(context, pool_from_hostname, hostname)
            create_pool(context, pool_dict, pool_members)
    else:
        # PowerVC knows about this pool
        if pool_from_hostname is None:
            LOG.info(_('There is a pool corresponding to the pool UUID '
                       '(%(pool_uuid)s) from PowerVM, but there is no pool '
                       'corresponding to the hostname (%(hostname)s).') %
                     {'pool_uuid': pool_from_powervm_uuid.uuid,
                      'hostname': hostname})
            # PowerVC does not have this host belonging to any pools
            # Therefore, this host will be added as a member of this pool via
            # the call to _update_pool below.
        else:
            # PowerVC has this host belonging to a pool
            # Check if the pool that PowerVC thinks the host belongs to is the
            # same one as this pool.  If not, remove the host from the other
            # pool.  The host will be added as a member of this pool via the
            # call to _update_pool below.
            if pool_from_powervm_uuid.uuid != pool_from_hostname.uuid:
                LOG.info(_('There is a pool corresponding to the pool UUID '
                           '(%(pool_uuid)s) from PowerVM and there is a '
                           'different pool (%(other_pool_uuid)s) '
                           'corresponding to the hostname (%(hostname)s). '
                           'This is most likely due to the host switching '
                           'pools.') %
                         {'pool_uuid': pool_from_powervm_uuid.uuid,
                          'other_pool_uuid': pool_from_hostname.uuid,
                          'hostname': hostname})
                remove_member(context, pool_from_hostname, hostname)

        update_pool(pool_from_powervm_uuid, pool_dict, pool_members)


def validate_setting_mobile_resources(
        resource_type, hostname, desired_mobile_resources,
        unreturned_mobile_resources, inactive_resources,
        current_mobile_resources, pool_uuid, pool_name,
        pool_available_mobile_resources, pool_compliance_state):
    """Validates the desired mobile resources value.

    :param resource_type: Type of mobile resources. Possible values:
        constants.COD_MOBILE_CORES_RESOURCE_TYPE
        constants.COD_MOBILE_MEM_RESOURCE_TYPE
    :param hostname: The string hostname of the host
    :param desired_mobile_resources: Integer
    :param unreturned_mobile_resources: Integer
    :param inactive_resources: Integer
    :param current_mobile_resources: Integer
    :param pool_uuid: String UUID of the pool
    :param pool_name: String name of the pool
    :param pool_available_mobile_resources: Integer
    :param pool_compliance_state: String
    :returns: Nothing if validation succeeds, otherwise raises an exception of
              the following types:
                  IBMPowerVMInvalidInputMobileResourcesValue
                  IBMPowerVMPoolNotEnoughMobileResources
                  IBMPowerVMPoolNotCompliantTooManyResources
                  IBMPowerVMHostNotEnoughResources
    """
    if desired_mobile_resources < 0:
        raise exception.IBMPowerVMInvalidInputMobileResourcesValue(
            resource_type=resource_type,
            desired_mobile_resources=desired_mobile_resources)

    mobile_resources_delta = (desired_mobile_resources -
                              current_mobile_resources)
    if (mobile_resources_delta > 0 and
            mobile_resources_delta > pool_available_mobile_resources):
        raise exception.IBMPowerVMPoolNotEnoughMobileResources(
            resource_type=resource_type,
            pool_uuid=pool_uuid,
            pool_name=pool_name,
            host=hostname,
            increase_amount=mobile_resources_delta,
            available=pool_available_mobile_resources)
    elif (pool_compliance_state == constants.EPOOL_OUT_OF_COMPLIANCE and
            mobile_resources_delta > unreturned_mobile_resources):
        raise exception.IBMPowerVMPoolNotCompliantTooManyResources(
            resource_type=resource_type,
            host=hostname, unreturned_resources=unreturned_mobile_resources)
    elif (mobile_resources_delta > 0 and
            mobile_resources_delta > (inactive_resources +
                                      unreturned_mobile_resources)):
        raise exception.IBMPowerVMHostNotEnoughResources(
            resource_type=resource_type,
            increase_amount=mobile_resources_delta,
            unreturned=unreturned_mobile_resources,
            available=inactive_resources,
            host=hostname)


def set_mobile_cores(context, desired_mobile_cores, hostname,
                     update_mobile_cores_partial,
                     extra_validation_partial=None):
    """Sets the mobile cores that a host is using to the desired value.

    :param context: The security context for the query
    :param desired_mobile_cores: the number of mobile cores that the host
                                 should use
    :param hostname: The string hostname of the host
    :param update_mobile_cores_partial: functools.partial object, which
                                        represents the method that should be
                                        called to do the actual update of the
                                        pool member, which is different
                                        depending on HMC vs. NovaLink. The
                                        method the partial represents should
                                        take in the pool UUID, member UUID, and
                                        the desired_mobile_cores value. It
                                        should return a tuple of the form:
                                            (unreturned mobile cores,
                                             mobile cores, inactive cores)
                                        corresponding to the updated pool
                                        member.
    :param extra_validation_partial: Optional functools.partial object, which
                                     represents an extra validation method that
                                     should be called in addition to the
                                     validation already being done in this
                                     method. The method the partial represents
                                     should take in a
                                     powervc_nova.objects.compute.dom.
                                     EnterprisePool object.
    :returns: Tuple of (result, reason). For example, if the mobile cores
              were successfully updated, then a tuple of (True,) will be
              returned. Otherwise, it would return a tuple of
              (False, 'Reason for failure').
    """
    try:
        desired_mobile_cores = int(desired_mobile_cores)
    except ValueError as e:
        msg = _('The number of mobile cores that the host should use '
                'specified as desired_mobile_cores (%s) is not an integer '
                'value.') % desired_mobile_cores
        LOG.error(msg)
        return (False, msg)

    compute_nodes = (objects.ComputeNodeList.get_all_by_host(
        context, hostname) or [])

    if not compute_nodes:
        msg = _('Compute node not found for host %s.') % hostname
        LOG.error(msg)
        return (False, msg)

    node = compute_nodes[0]

    pool = get_pool_for_host(context, hostname)
    if not pool:
        msg = _('Member not found in enterprise pool for host: %s.' % hostname)
        LOG.error(msg)
        return (False, msg)

    management_platform = node.stats[constants.MANAGED_BY]
    if (management_platform not in
            constants.PLATFORMS_SUPPORTING_EPOOLS):
        msg = (_('Host %(host)s is managed by a platform (%(platform)s) that '
                 'does not support enterprise pools.') %
               {'host': hostname,
                'platform': management_platform})
        return (False, msg)

    if extra_validation_partial:
        extra_validation_res = extra_validation_partial(pool=pool)

        if not extra_validation_res[0]:
            return extra_validation_res

    member_uuid = pool.host_names[hostname]

    mobile_cores = int(node.stats[constants.COD_MOBILE_CORES])
    if mobile_cores == desired_mobile_cores:
        # Nothing to do, so just return
        return (True, )

    unreturned_cores = int(node.stats[constants.COD_UNRETURNED_MOBILE_CORES])
    inactive_cores = int(node.stats[constants.COD_INACTIVE_CORES])

    try:
        validate_setting_mobile_resources(
            constants.COD_MOBILE_CORES_RESOURCE_TYPE,
            hostname, desired_mobile_cores, unreturned_cores, inactive_cores,
            mobile_cores, pool.uuid, pool.name, pool.available_mobile_cores,
            pool.compliance_state)
    except (exception.IBMPowerVMInvalidInputMobileResourcesValue,
            exception.IBMPowerVMPoolNotEnoughMobileResources,
            exception.IBMPowerVMPoolNotCompliantTooManyResources,
            exception.IBMPowerVMHostNotEnoughResources) as e:
        LOG.info(_('Did not attempt to update mobile cores for host %(host)s '
                   'because a prerequisite failed: %(error)s') %
                 {'host': hostname,
                  'error': e.format_message()})
        return (False, six.text_type(e))

    # Save off the existing node stats prior to doing the update so we can
    # calculate the deltas easily
    stats = copy.deepcopy(node.stats)

    try:
        (unreturned_mobile_cores,
         mobile_cores,
         inactive_cores) = update_mobile_cores_partial(
             pool_uuid=pool.uuid, member_uuid=member_uuid,
             desired_mobile_cores=desired_mobile_cores)
    except Exception as e:
        LOG.error(_('Error updating mobile cores for host: %(name)s; '
                    'member_id: %(member)s; Exception: %(exc)s') %
                  {'name': hostname,
                   'member': member_uuid,
                   'exc': six.text_type(e)})
        return (False, six.text_type(e))

    mobile_core_delta = mobile_cores - int(stats[constants.COD_MOBILE_CORES])
    unreturned_core_delta = unreturned_mobile_cores - unreturned_cores

    if unreturned_core_delta > 0:
        LOG.warn(_('The number of unreturned mobile cores for host '
                   '%(host)s increased from %(original)s to %(new)s.')
                 % {'host': hostname,
                    'original': unreturned_cores,
                    'new': unreturned_mobile_cores})

    LOG.info(_('The number of mobile cores for host %(host)s changed '
               'from %(orig)s to %(new)s.') %
             {'host': hostname,
              'orig': stats[constants.COD_MOBILE_CORES],
              'new': mobile_cores})

    # Go ahead and update the node host stats with the new number of mobile
    # cores, inactive cores, and unreturned cores from the response that was
    # was returned from the POST operation. This response will have the latest
    # values from PowerVM's point of view.
    _update_node_stats(context, node, stats, hostname, mobile_cores,
                       unreturned_mobile_cores, inactive_cores,
                       unreturned_core_delta, mobile_core_delta)

    # Go ahead and update the enterprise pool DB with the new number of
    # available cores
    new_available_mobile_cores = (pool.available_mobile_cores -
                                  mobile_core_delta)
    pool.available_mobile_cores = new_available_mobile_cores
    pool.save()

    return (True, )


def _update_node_stats(context, node, stats, hostname, mobile_cores,
                       unreturned_mobile_cores,
                       inactive_cores, unreturned_core_delta,
                       mobile_core_delta):
    """Update the node stats for the compute node and send a notification.

    :param context: The security context for the query
    :param node: objects.ComputeNode object
    :param stats: the older node stats dictionary
    :param hostname: String hostname of the host
    :param mobile_cores: New integer value of num of mobile cores
    :param unreturned_mobile_cores: New int value of num of unreturned cores
    :param inactive_cores: New integer value of the number of inactive cores
    :param unreturned_core_delta: Difference between new unreturned cores and
                                  old unreturned cores.
    :param mobile_core_delta: Difference between new mobile cores and old
                              mobile cores.
    """

    stats[constants.COD_MOBILE_CORES] = mobile_cores
    stats[constants.COD_UNRETURNED_MOBILE_CORES] = unreturned_mobile_cores
    stats[constants.COD_INACTIVE_CORES] = inactive_cores

    new_proc_units = (int(float(stats[constants.PROC_UNITS])) +
                      mobile_core_delta + unreturned_core_delta)
    stats[constants.PROC_UNITS] = six.text_type(new_proc_units)

    node.stats = stats

    # Update the node vcpus value
    new_vcpus = node.vcpus + mobile_core_delta + unreturned_core_delta
    node.vcpus = new_vcpus

    node.save()

    # Send a notification to let listeners (scheduler and UI) know that
    # the node stats have been updated
    notifier = rpc.get_notifier('compute', hostname)
    event_type = 'compute.node.update.end'

    service = objects.Service.get_by_compute_host(context, node.host)

    info = {'host': hostname,
            'service_id': service.id,
            'compute_node_id': node.id,
            'hypervisor_hostname': node.hypervisor_hostname,
            'stats': node.stats}

    # Send the notification that we are updating the Compute Node
    notifier.info(context, event_type, info)


def convert_mtms_to_hostname(machine_type, model, serial_num):
    """Converts the MTMS (machine type, model, serial num) to PowerVC hostname.

    :param machine_type: String representing machine type
    :param model: String representing machine type
    :param serial_num: String representing machine type
    """
    return '{0}{1}_{2}'.format(machine_type, model, serial_num)
