#
#
# Copyright 2013, 2015 IBM Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import json
import os
import time
import requests

from eventlet import greenthread
from nova.compute.ibm import configuration_strategy_common as config_strategy
from nova.compute.ibm import update_configuration_strategy as update_strategy
from nova.compute import task_states

from nova import block_device
from nova import exception
from nova import rpc
from nova.image import glance

from oslo_utils import excutils
from oslo_serialization import jsonutils
from oslo_service import sslutils
from oslo_log import log as logging

from nova.compute import power_state
from nova.compute import vm_states
from powervc_nova import logcall
from powervc_nova.virt.ibmpowervm.common import exception as pvcex
import six
from powervc_nova.virt.ibmpowervm.common.volume_utils \
    import PowerVCNovaVirtDriverVolumeAPI
from powervc_nova.virt.ibmpowervm.common.constants import IBMI_OS_DISTRO
from oslo_config import cfg

from powervc_nova import _

CONF = cfg.CONF
ibmpowervm_opts = [
    cfg.BoolOpt('ibmpowervm_allow_live_capture',
                default=False,
                help='Allow live captures')
]
CONF.register_opts(ibmpowervm_opts)

LOG = logging.getLogger(__name__)


def get_volume_ids_from_image_bdms(image_bdms):
    bdm_objs = jsonutils.loads(image_bdms)
    image_defined_bdms = map(block_device.BlockDeviceDict,
                             bdm_objs)
    volume_ids = []
    for bdm in image_defined_bdms:
        vol_id = bdm.get('volume_id')
        if vol_id:
            volume_ids.append(vol_id)
    return volume_ids


@logcall
def get_default_ovf_descriptor():
    # load ibmpowervm OVF template
    t_path = os.path.dirname(__file__)
    new_ovf_file = os.path.join(t_path, '../templates/ibmpowervm.ovf')
    with open(new_ovf_file, 'r') as f:
        ovf_descriptor = f.read()
    return ovf_descriptor


class ImageEntry (object):
    """
    An class representing an image entry in Glance
    """
    def __init__(self, context, image_id):
        self._image_id = image_id
        self._glance_service, img_id = glance.get_remote_image_service(
            context,
            image_id)
        self._metadata = self._glance_service.show(context, image_id)
        self._props = self._metadata['properties']

        self._additional_vols = {}
        for prop in self._props:
            if prop.startswith("powervc_capture_disk"):
                # save volume IDs in local list
                self._additional_vols[prop] = self._props[prop]

    @logcall
    def add_config_strategy(self, context):
        """
        Adds a configuration strategy to the image entry in Glance.
        No-ops if the image already has a configuration strategy.

        :param context: security context
        """
        # If the image does not have a configuration strategy yet,
        # add the default configuration strategy.
        # If the base image that this VM was deployed from
        # contains a configuration strategy, it will be copied
        # forward into this image in the compute manager.
        if 'configuration_strategy' not in self._props:
            LOG.debug("The new image metadata does not contain"
                      " a configuration strategy. Adding one now.")

            mappings = []
            ovf_descriptor = get_default_ovf_descriptor()
            cs_data = config_strategy.calc_common_metadata('ovf',
                                                           ovf_descriptor,
                                                           mappings)
            cs_json = json.dumps(cs_data)

            # Send the request to Glance to add the configuration strategy
            update_strategy.send_request(self._image_id, context.auth_token,
                                         'add',
                                         config_strategy_data=cs_json)

        self._metadata = self._glance_service.show(context, self._image_id)
        self._props = self._metadata['properties']

    @logcall
    def get_properties(self):
        """
        Get a dictionary of the image properties
        """
        props = copy.deepcopy(self._props)

        return props

    @logcall
    def get_metadata(self):
        """"
        Get a dictionary of the image metadata.
        """
        metadata = copy.deepcopy(self._metadata)
        return metadata

    @logcall
    def get_capture_updated_metadata(self, instance):
        """
        Get a dict containing the initial updated metadata
        that will be used to 'complete' the image when capture
        is complete.
        """
        # Initialize the metadata that will be sent as an update
        # to the image when the capture is complete.
        new_snapshot_meta = {'is_public': False,
                             'name': self._metadata['name'],
                             'status': 'active',
                             'properties': {'image_location': 'snapshot',
                                            'image_state': 'available',
                                            'owner_id': instance['project_id'],
                                            'hypervisor_type': 'phyp',
                                            'architecture': 'ppc64'
                                            },
                             'disk_format': 'raw',
                             'container_format': 'bare',
                             }
        return new_snapshot_meta

    @logcall
    def complete_capture(self, context, instance, update_task_state,
                         new_snapshot_meta, purge_ae_props):
        """
        Update the image entry in Glance for a successful capture.

        For use during capture after the volume has been cloned.
        It will update the image and make it 'active'.  This method will also
        be responsible for taking the volume_data,
        and updating the image accordingly.

        :param context: security context
        :param instance: Instance object as returned by DB layer.
        :param update_task_state: Function reference that allows for updates
                                  to the instance task state.
        :param new_snapshot_meta: Updated metadata for the image
        """

        # Updating instance task state before uploading image
        # Snapshot will complete but instance state will not change
        # to none in compute manager if expected state is not correct
        update_task_state(task_state=task_states.IMAGE_UPLOADING,
                          expected_state=task_states.IMAGE_PENDING_UPLOAD)

        # Update the image metadata in Glance.

        # Glance v1 APIs, and likely v2 only support
        # creating an image with no location or no data during the create
        # operation while specifying image size = 0.  There is no way to
        # move an image from queued to active state without giving it an
        # external location or image data.  Using glanceclient directly you
        # can pass a zero length data.  Using Nova's glance service you
        # cannot, and you have to pass at least 1 byte of data.

        bdms = new_snapshot_meta['properties'].pop('block_device_mapping')
        ae_props_to_purge = []
        if purge_ae_props:
            if 'configuration_strategy' in self._props:
                ae_props_to_purge.append('configuration_strategy')
            if 'powervc_user_ovf' in self._props:
                ae_props_to_purge.append('powervc_user_ovf')

        update_image(self._image_id, context.auth_token,
                     self._additional_vols.keys(), bdms, ae_props_to_purge)
        self._glance_service.update(context,
                                    self._image_id,
                                    new_snapshot_meta, purge_props=False,
                                    data='1')

        self._metadata = self._glance_service.show(context, self._image_id)
        self._props = self._metadata['properties']

    @logcall
    def get_id(self):
        """
        Return the image id
        """
        return self._image_id

    @logcall
    def get_additional_vols(self):
        """
        Return dict of additional volumes to capture
        """
        return self._additional_vols


class Image(object):
    """
    An class representing an image
    """

    @logcall
    def capture(self, context, instance, image_entry, update_task_state,
                get_partition_state, remove_opticals):
        """
        Captures an image of the instance and stores information about
        the image in the image_entry.

        :param context: security context
        :param instance: Instance object as returned by DB layer.
        :param image_entry: ImageEntry object for a pre-created image that will
                         hold the captured image.
        :param update_task_state: Function reference that allows for updates
                                  to the instance task state.
        :param get_partition_state: Function reference that returns the
                partition state from the hypervisor.
        :param remove_opticals: Function reference that removes virtual
            opticals from the VM.
        """
        try:
            properties = image_entry.get_properties()
            distro = properties.get('os_distro')
            if not distro:
                raise pvcex.IBMPowerVMNoOSForVM(instance_name=instance['name'])

            os_options = ['aix', 'rhel', 'sles', 'ubuntu', IBMI_OS_DISTRO]
            i_name = instance['name']
            if distro not in os_options:
                raise pvcex.IBMPowerVMUnsupportedOSForVM(os_distro=distro,
                                                         instance_name=i_name,
                                                         options=os_options)
            # Copy config strategy metadata from parent image to snapshot
            # image.
            update_strategy.update_metadata(context, image_entry.get_id(),
                                            instance)

            # Ensure the image has a configuration strategy if the instance
            # uses AE.
            system_metadata = instance.get('system_metadata', {})
            uses_ae = system_metadata.get('image_configuration_strategy')
            purge_ae_props = False
            # refresh image_entry after update_strategy.update_metadata
            image_id = image_entry.get_id()
            image_entry = ImageEntry(context, image_id)
            if uses_ae:
                self._check_for_lost_custom_ovf(context, instance, image_entry)
                image_entry.add_config_strategy(context)
            else:
                properties = image_entry.get_properties()
                if ('configuration_strategy' in properties or
                        'powervc_user_ovf' in properties):
                    purge_ae_props = True

            # ensure partition is stopped
            self._wait_for_instance_to_stop(instance, get_partition_state)
            # remove virtual optical
            remove_opticals()

            meta_data = image_entry.get_capture_updated_metadata(instance)
            meta_data['id'] = image_entry.get_id()
            # capture storage, pass metadata so storage capture can
            # update the volume related information
            # This call requires the id to be in the metadata so we set it
            # from the image entry and pull it out after the call, before we
            # send the properties up to Glance
            update_task_state(task_state=task_states.IMAGE_PENDING_UPLOAD)
            volume_api = PowerVCNovaVirtDriverVolumeAPI()
            additional_vols = image_entry.get_additional_vols().values()
            meta_data = volume_api.instance_volume_capture(context,
                                                           instance,
                                                           meta_data,
                                                           additional_vols)
            del meta_data['id']
        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.exception(_("Snapshot failed with reason %s ") %
                              six.text_type(e))

        image_bdms = meta_data['properties']['block_device_mapping']
        # We put the complete capture call in a separate block to remove
        # the volume if it fails.  The above try-expect block does not need
        # the remove call since instance_boot_volume_capture should
        # self-clean-up on failure.
        try:
            image_entry.complete_capture(context, instance,
                                         update_task_state, meta_data,
                                         purge_ae_props)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.exception(_("Snapshot failed with reason %s ") %
                              six.text_type(e))
                volume_ids = get_volume_ids_from_image_bdms(image_bdms)
                # Deleting volumes in case of error.
                volume_api.delete_image_volumes_on_capture_fail(context,
                                                                volume_ids)

    def _check_for_lost_custom_ovf(self, context, instance, image_entry):
        properties = image_entry.get_properties()
        user_ovf = properties.get('powervc_user_ovf')
        config_strat = properties.get('configuration_strategy')
        if user_ovf and not config_strat:
            msg_args = {'instance_id': instance.uuid,
                        'image_id': image_entry.get_id()}
            msg = _('Image %(image_id)s had a custom OVF associated with it '
                    'and was'
                    ' used to deploy %(instance_id)s.  However, the image and'
                    ' its OVF cannot be found.  Therefore, if you capture an '
                    'image from %(instance_id)s and then deploy an image from '
                    'that captured image, the deployed image will not run any '
                    'custom activation using extensions that were associated '
                    'with the OVF.') % msg_args
            LOG.warning(msg)
            notify_msg = _('Image %(image_id)s had a custom OVF associated '
                           'with it and was used to deploy {instance_name}.  '
                           'However, the image and its OVF cannot be found.  '
                           'Therefore, if you capture an image from '
                           '{instance_name} and then deploy an image from '
                           'that captured image, the deployed image will not '
                           'run any custom activation using extensions that '
                           'were associated with the OVF.') % msg_args
            try:
                payload = {}
                payload['msg'] = notify_msg
                payload['instance_name'] = instance.display_name
                notifier = rpc.get_notifier(service='compute',
                                            host=CONF.host)
                notifier.warn(context,
                              'compute.instance.log',
                              payload)

            except Exception as exc:
                LOG.debug('Notify failed with exception: %s' % exc)

    @logcall
    def _wait_for_instance_to_stop(self, instance, get_partition_state):
        """
        Wait for the instance to stop and throw an exception if
        the instance does not reach the stopped state within a certain
        time threshold.
        :param get_partition_state: Function reference for retrieving the
            partition state from the hypervisor
        :param instance: Instance object as returned by DB layer.
        """

        # Check to see if we should allow live capture
        if CONF.ibmpowervm_allow_live_capture:
            return

        # Wait up to two minutes for the server to reach the
        # Not Activated state to cover the case of the capture being called
        # immediately after the AE initiated shut down started.
        total_wait_time = 120
        wait_time_between_attempts = 5
        loop_iterations = total_wait_time / wait_time_between_attempts
        LOG.debug("Checking to see if the instance is stopped.")
        timeout_count = range(loop_iterations)
        start_time = time.time()
        while timeout_count:
            state = get_partition_state()
            if state == power_state.SHUTDOWN:
                LOG.info(_("Virtual machine is stopped."),
                         instance=instance)
                if instance['vm_state'] is not None and \
                        instance['vm_state'] != vm_states.STOPPED:
                    instance['vm_state'] = vm_states.STOPPED
                    instance.save()
                break
            timeout_count.pop()
            current_time = time.time()
            loop_time = current_time - start_time
            if len(timeout_count) == 0 or total_wait_time < loop_time:
                LOG.error(_("Virtual machine '%s' failed to "
                            "stop during capture.") %
                          instance['name'])
                raise pvcex.IBMPowerVMLPARIsRunningDuringCapture(
                    instance_name=instance['name'])

            LOG.debug("The instance is not stopped. Waiting to check again.")
            greenthread.sleep(wait_time_between_attempts)


def update_image(image_id, auth_token, additional_volume_props, bdms,
                 ae_props_to_purge):
    """This method will build a request to be sent to Glance
       to remove powervc_capture_disk keys from the image and set the
       block_device_mapping property.

    :param image_id: id of the new image or snapshot
    :param auth_token: authentication token for update
    :param additional_volume_props: Additional volume image properties to
        remove.
    :param bdms:  The value of the block_device_mapping property.
    :param ae_props_to_purge: AE related properties to remove from the image.
    """

    resource_path = '/v2/images/%s' % image_id

    headers = {'X-Auth-Token': auth_token,
               'Content-Type': 'application/openstack-images-v2.0-json-patch'
               }

    operations = []
    for capture_vol in additional_volume_props:
        data = {}
        data['remove'] = '/%s' % capture_vol
        operations.append(data)

    if ae_props_to_purge:
        for ae_prop in ae_props_to_purge:
            data = {}
            data['remove'] = '/%s' % ae_prop
            operations.append(data)

    data = {}
    data['add'] = '/block_device_mapping'
    data['value'] = bdms
    operations.append(data)
    data_str = jsonutils.dumps(operations)

    session = requests.Session()
    session.headers["X-Auth-Token"] = auth_token
    endpoint = glance.generate_glance_url() + resource_path
    sslutils.is_enabled(CONF)
    session.verify = CONF.ssl.ca_file

    # Build the request
    response = session.request(
        'PATCH', endpoint, data=data_str, headers=headers)

    if 200 <= response.status_code < 300:
        pass
    else:
        # We need to inform the user the transfer failed
        raise exception.InvalidRequest()
