# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from eventlet import greenthread

import six
import powervc_nova.virt.ibmpowervm.hmc.k2.constants as k2const

from nova import exception as nova_exception
from nova import network as net_api
from nova import rpc
from nova.compute import power_state
from nova.compute import task_states
from nova.compute import vm_states
from nova.ibm.ego import ego_constants
from oslo_log import log as logging

from powervc_nova.common import aggregate_utils as agg_utils
from powervc_nova.common import server_group_utils as sg_utils
from powervc_nova.virt.ibmpowervm.common import exception
from powervc_nova.virt.ibmpowervm.common import proc_utils
from powervc_nova.virt.ibmpowervm.common.constants import RR_HYP_STATES
from powervc_nova import logcall
from powervc_nova.compute import api
from powervc_nova.virt.ibmpowervm.hmc.k2.lparentry_wrapper \
    import LPAREntryWrapper

from oslo_config import cfg

from powervc_nova import _

LOG = logging.getLogger(__name__)
CONF = cfg.CONF

MANAGED_BY_STR = {'hmc': _('HMC'), 'novalink': _('NovaLink')}
VIOS_KEY = "vios_"


@logcall
def verify_aggregate_domain(context, instance_name, src_host, dst_host):
    """
    Checks whether the source and destination hosts belong to the same host
    aggregate(s).

    :param context: The nova context
    :param instance_name: The name of the instance
    :param src_host: The source host
    :param dst_host: The destination host
    """
    src_aggs = agg_utils.get_aggregate_ids_for_host(context, src_host)
    dst_aggs = agg_utils.get_aggregate_ids_for_host(context, dst_host)
    agg_intersect = set(src_aggs) & set(dst_aggs)
    # If the source host is in an aggregate and the intersection of its
    # aggregates and the destination's aggregates is the empty set, then
    # the migration should be denied.
    if src_aggs and not agg_intersect:
        error = (_("Cannot migrate %(instance)s because the source host "
                   "and destination host are not members of the same "
                   "host aggregate(s).") %
                 {'instance': instance_name})
        LOG.exception(error)
        raise nova_exception.MigrationPreCheckError(reason=error)


@logcall
def verify_collocation_rules(context, instance, dst_host):
    """
    Checks the collocation rules the instance belongs to and
    makes sure the migration to the destination host would not
    violate any of the rules.

    :param context: The nova context
    :param instance: The instance object
    :param dst_host: The destination host
    """
    if sg_utils.ignore_collocation_rules():
        # Nothing to verify if ignore_collocation_rules conf option is True
        return
    groups = sg_utils.get_all_groups_for_instance(context, instance.uuid)
    # If instance is not a member of any collocation rules, return
    if not groups:
        return

    affinity_rules = [group for group in groups if
                      sg_utils.AFFINITY in group.policies]
    anti_affinity_rules = [group for group in groups if
                           sg_utils.ANTI_AFFINITY in group.policies]
    # If there is more than one instance in any of this instance's affinity
    # rules, it cannot be migrated.
    unsupported_rules = []
    for group in affinity_rules:
        num_instances = len(sg_utils.get_members(context, group.uuid))
        if num_instances > 1:
            unsupported_rules.append(group.name)
    if unsupported_rules:
        rules_list = sg_utils.get_str_from_list(unsupported_rules)
        error = (_("Cannot migrate %(instance)s because it is a member "
                   "of one or more affinity collocation rules. To migrate "
                   "this virtual machine, first remove it from the following "
                   "affinity collocation rules: %(rules)s.") %
                 {'instance': instance.display_name, 'rules': rules_list})
        LOG.exception(error)
        raise nova_exception.MigrationPreCheckError(reason=error)

    for group in anti_affinity_rules:
        # Get instance objects for each member in the rule
        members = sg_utils.get_members(context, group.uuid)
        # If any other migrations, rebuilds, or deploys are ongoing for any
        # other instances in the same rule, fail the migration to ensure no
        # race conditions could occur that would violate the rule.
        has_ongoing_operation = sg_utils.has_ongoing_operations_in_group(
            group.uuid, instance.uuid, members,
            ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION)
        if has_ongoing_operation:
            error = (_("Cannot migrate %(instance)s to host %(host)s "
                       "because it belongs to anti-affinity collocation rule "
                       "%(rule)s. That collocation rule contains another "
                       "virtual machine that is undergoing an operation that "
                       "might cause it to move to a new host and violate "
                       "the collocation rule. Wait until ongoing operations "
                       "for all virtual machines in this group have finished "
                       "and try the migration again.") %
                     {'instance': instance.display_name,
                      'host': dst_host, 'rule': group.name})
            LOG.exception(error)
            raise nova_exception.MigrationPreCheckError(reason=error)
        # If there is already an instance in the anti-affinity rule
        # on the destination host, the instance cannot be relocated
        # to this host.
        for member in members:
            if dst_host == member.host:
                error = (_("Cannot migrate %(instance)s to host %(host)s "
                           "because it belongs to anti-affinity collocation "
                           "rule %(rule)s. That collocation rule contains "
                           "another virtual machine %(conflict)s "
                           "which is located on this host.") %
                         {'instance': instance.display_name,
                          'host': dst_host, 'rule': group.name,
                          'conflict': member.display_name})
                LOG.exception(error)
                raise nova_exception.MigrationPreCheckError(reason=error)


@logcall
def verify_reserve_policy(reserve_policy, instance_name, disk_name):
    """
    The reserve policy of the hdisk attached to the lpar;
    it must be set to 'no_reserve' for migrations to succeed.

    :param reserve_policy: The reserve policy to check
    :param intsance_name: The name of the instance (for logging)
    :param disk_name: The name of the disk (for logging)
    """
    if reserve_policy not in ('no_reserve', 'NoReserve'):
        error = (_("Cannot migrate %(instance)s because "
                   "the reserve policy on disk %(hdisk)s is set to "
                   "%(policy)s. It must be set to no_reserve.") %
                 {'instance': instance_name,
                  'hdisk': disk_name,
                  'policy': reserve_policy})
        LOG.exception(error)
        raise nova_exception.MigrationPreCheckError(reason=error)


@logcall
def verify_rmc_state(rmc_state, instance_name):
    """
    The rmc state of the instance must be 'active'

    :param rmc_state: The rmc_state to check
    :param intsance_name: The name of the instance (for logging)
    """
    if rmc_state != 'active':
        error = (_("Cannot live migrate %(inst)s because its RMC "
                   "state is %(rmc)s. The RMC state must be active.") %
                 {'inst': instance_name,
                  'rmc': rmc_state})
        LOG.exception(error)
        raise nova_exception.MigrationPreCheckError(reason=error)


@logcall
def verify_dlpar_enabled(dlpar, instance_name):
    """
    Dlpar must be enabled on the instance

    :param dlpar: Boolean value indicating dlpar is enabled
    :param intsance_name: The name of the instance (for logging)
    """
    if not dlpar:
        error = (_("Cannot live migrate %s because DLPAR "
                   "is not enabled.") % instance_name)
        LOG.exception(error)
        raise nova_exception.MigrationPreCheckError(reason=error)


@logcall
def verify_proc_compat_mode(lpar_compat_mode, host_compat_modes,
                            instance_name):
    """
    The processor compatibility mode of the lpar must be in the list of
    supported processor compatibility modes on the target host

    :param lpar_compat_mode: The proc compat mode of the lpar
    :param host_compat_modes: The proc compat modes of the target host
    :param instance_name: The name of the instance (for logging)
    """
    mode_list = host_compat_modes.split(',')
    if not lpar_compat_mode in mode_list:
        error = (_("Cannot migrate %(inst)s because its "
                   "processor compatibility mode %(mode)s "
                   "is not in the list of modes %(modes)s "
                   "supported by the target host.") %
                 {'inst': instance_name,
                  'mode': lpar_compat_mode,
                  'modes': mode_list})
        LOG.exception(error)
        raise nova_exception.MigrationPreCheckError(reason=error)


@logcall
def verify_shared_proc_pool(lpar_spp, host_spp_id, instance_name):
    """
    The shared processor pool of the lpar must be in the list of
    shared processor pools on the target host.

    :param lpar_spp: The shared proc pool name of the lpar
    :param host_spp_id: The shared proc pool id on the target host
    :param instance_name: The name of the instance (for logging)
    """
    if host_spp_id is None:
        error = (_("Cannot migrate %(inst)s because the target host "
                   "does not have the shared processor pool named %(spp)s.") %
                 {'inst': instance_name,
                  'spp': lpar_spp})
        LOG.exception(error)
        raise nova_exception.MigrationPreCheckError(reason=error)


@logcall
def verify_host_capacity_for_migration(migrations_allowed, migrations_running,
                                       instance_name, host_name):
    """
    The host can only support a limited number of concurrent migrations.
    Verify there is room for another.

    :param migrations_allowed: The number of migrations allowed
    :param migrations_running: The number of migrations in progress
    :param instance_name: The name of the instance (for logging)
    :param host_name: The name of the host (for logging)
    """
    if migrations_running >= migrations_allowed:
        error = (_("Cannot migrate %(inst)s because host %(host)s "
                   "only allows %(allowed)s concurrent migrations and "
                   "%(running)s migrations are currently running.") %
                 {'inst': instance_name,
                  'host': host_name,
                  'allowed': migrations_allowed,
                  'running': migrations_running})
        LOG.exception(error)
        raise nova_exception.MigrationPreCheckError(reason=error)


@logcall
def verify_logical_memory_block_size(source_lmb, target_lmb, instance_name):
    """
    The logical memory block size of the source and target host
    in a live migration operation must be the same.

    :param source_lmb: The block size of the source host
    :param target_lmb: The block size of the target host
    :param instance_name: The name of the instance (for logging)
    """
    if source_lmb != target_lmb:
        error = (_("Cannot migrate %(inst)s because the logical "
                   "memory block size of the source (%(source_lmb)s MB) "
                   "does not match the logical memory block size of the "
                   "target (%(target_lmb)s MB).") % {'inst': instance_name,
                                                     'source_lmb': source_lmb,
                                                     'target_lmb': target_lmb})
        LOG.exception(error)
        raise nova_exception.MigrationPreCheckError(reason=error)


@logcall
def send_migration_failure_notification(context, instance,
                                        tgt_host, exception):
    """
    Sends a notification of live migration failure to the GUI

    :param context: security context
    :param instance: The instance that was migrating
    :param tgt_host: The target host name
    :param exception: The exception that was thrown
    """

    if hasattr(exception, 'message'):
        err_msg = six.text_type(exception.message)
    else:
        err_msg = six.text_type(exception)

    # Send error notification
    info = {'msg': _('Migration of virtual machine {instance_name} '
                     'to host {host_name} failed. '
                     '{error}'),
            'instance_id': instance['uuid'],
            'instance_name': instance['display_name'],
            'host_name': tgt_host,
            'error': err_msg}
    notifier = rpc.get_notifier(service='compute', host=CONF.host)
    notifier.error(context, 'compute.instance.log', info)


@logcall
def send_migration_success_notification(context, instance, src_host, tgt_host):
    """
    Sends a notification of live migration failure to the GUI

    :param context: security context
    :param instance: The instance that was migrating
    :param src_host: The source host name
    :param tgt_host: The target host name
    """

    # Send error notification
    info = {'msg': _('Migration of virtual machine {instance_name} from '
                     '{source_host} to {target_host} was successful.'),
            'instance_id': instance['uuid'],
            'instance_name': instance['display_name'],
            'source_host': src_host,
            'target_host': tgt_host}
    notifier = rpc.get_notifier(service='compute', host=CONF.host)
    notifier.info(context, 'compute.instance.log', info)


@logcall
def send_migration_notification(context, instance, priority, message):
    """
    Sends a general live migration notification to the GUI

    :param context: security context
    :param instance: The instance that was migrating
    :param priority: The level of message (DEBUG, WARN, INFO, ERROR, CRITICAL)
    :param message: The message to print
    """

    # Send event notification
    info = {'msg': '{message}',
            'instance_id': instance['uuid'],
            'instance_name': instance['display_name'],
            'message': message}
    notifier = rpc.get_notifier(service='compute', host=CONF.host)
    if priority == 'error':
        notifier.error(context, 'compute.instance.log', info)
    elif priority == 'info':
        notifier.info(context, 'compute.instance.log', info)
    elif priority == 'critical':
        notifier.critical(context, 'compute.instance.log', info)
    elif priority == 'warn':
        notifier.warn(context, 'compute.instance.log', info)
    elif priority == 'debug':
        notifier.debug(context, 'compute.instance.log', info)


def verify_out_of_band_evacuation(context, instance, uuid, uuid_in_db,
                                  host_in_db, host_name, vm_info, node,
                                  src_host_state=None, src_host_uuid=None,
                                  k2oper=None):
    # Verify if an out of band migration or remote restart has moved an
    # instance to current host. Spawn a thread to avoid blocking main function.
    greenthread.spawn(_verify_out_of_band_evacuation, context, instance, uuid,
                      uuid_in_db, host_in_db, host_name, vm_info, node,
                      src_host_state=src_host_state,
                      src_host_uuid=src_host_uuid, k2oper=k2oper)


def _verify_out_of_band_evacuation(context, instance, uuid, uuid_in_db,
                                   host_in_db, host_name, vm_info, node,
                                   src_host_state=None, src_host_uuid=None,
                                   k2oper=None):
    """
    Verify if an out of band migration or remote restart
    has moved an instance to current host

    :param context: Security context
    :param instance: The instance that was migrating
    :param uuid: uuid of the instance
    :param uuid_in_db: uuid of the instance in database
    :param host_in_db: The source host of the instance
    :param src_host_state: State of the source host
    :param src_host_uuid: UUID of source host
    :param host_name: Current host name
    :param vm_info: Result of driver.get_info()
    :param node: Current node name
    :param k2oper: k2operator
    """

    # If LPAR appears on different host, check to see
    # if it is migrating or remote restarting
    # Check task_state, skip the rest if the instance
    # is deploying, deleting, migrating, or remote restarting.
    instance_task_state = instance.task_state
    instance_vm_state = instance.vm_state
    instance_name = instance.display_name
    if (instance_task_state in (task_states.SCHEDULING,
                                task_states.
                                BLOCK_DEVICE_MAPPING,
                                task_states.NETWORKING,
                                task_states.SPAWNING,
                                task_states.MIGRATING,
                                task_states.DELETING,
                                task_states.REBUILDING,
                                task_states.REBUILD_BLOCK_DEVICE_MAPPING,
                                task_states.REBUILD_SPAWNING) or
            instance_vm_state in (vm_states.BUILDING)):
        LOG.info(_('Virtual machine %(name)s is in '
                   'task state: %(task_state)s and '
                   'virtual machine state: '
                   '%(vm_state)s. The operation to '
                   'determine whether a migration or '
                   'remote restart was performed outside '
                   'of PowerVC will be skipped. '
                   'UUID on host: %(uuid_on_host)s. '
                   'UUID in DB: %(uuid_in_db)s. '
                   'Source host: %(src_host)s. Target '
                   'host: %(tgt_host)s')
                 % {'name': instance_name,
                    'task_state': instance_task_state,
                    'vm_state': instance_vm_state,
                    'uuid_on_host': uuid,
                    'uuid_in_db': uuid_in_db,
                    'src_host': host_in_db,
                    'tgt_host': host_name})
        return
    # Check 'instance_migrating' in case migration is out of band
    if hasattr(vm_info, 'instance_migrating') and vm_info.instance_migrating:
        LOG.debug('Instance %s is migrating. Skipped.' % instance_name)
        return
    # Check SRR state in case remote restart is out of band
    if (hasattr(vm_info, 'srr_state') and vm_info.srr_state ==
            'Destination_Remote_Restarting'):
        LOG.debug('Instance %s is remote restarting. Skipped.' % instance_name)
        return
    action = _('migrated')
    # If source host is in any of these 4 states then VM was remote restarted
    if src_host_state in RR_HYP_STATES:
        # check if instance on the source is in Remote_Restarted state.
        # if it's not, then LPARs are in bad states between 2 hosts.
        # if LPAR doesn't exist on the source host continue
        if src_host_uuid:
            try:
                src_lpar_resp = k2oper.read(k2const.MGT_SYS,
                                            rootId=src_host_uuid,
                                            childType=k2const.LPAR,
                                            childId=uuid)
                lpar_wrapper = LPAREntryWrapper(src_lpar_resp.entry)
                if lpar_wrapper.get_srr_state() not in 'Remote_Restarted':
                    LOG.info(_('Virtual machine %(inst)s is being managed by '
                               'remote host %(host)s and is not Remote '
                               'Restarted. This could mean that the virtual '
                               'machine is on two hosts simultaneously '
                               'after migration outside of PowerVC') %
                             {'inst': instance_name, 'host': host_in_db})
                    return
            except Exception:
                # Likely exception is that the LPAR is not on the host
                # which is ok
                pass
        action = _('remote restarted')
    # If instance is still on source host (host_in_db) for migration, skip.
    elif _is_instance_on_source_host(context, host_in_db, uuid, instance_name):
        return
    # Sync up host and node
    instance.host = host_name
    instance.node = node
    # Check the instance state in DB, compare it with instance state known to
    # hypervisor. Update DB if needed
    vm_power_state = vm_info.state
    if instance_vm_state == vm_states.ERROR:
        if vm_power_state == power_state.RUNNING:
            instance.vm_state = vm_states.ACTIVE
        elif vm_power_state == power_state.SHUTDOWN:
            instance.vm_state = vm_states.STOPPED
    elif instance_vm_state == vm_states.STOPPED:
        if vm_power_state == power_state.RUNNING:
            instance.vm_state = vm_states.ACTIVE
    inst_output = instance.obj_to_primitive()
    LOG.info(_('Virtual machine information: %s') % inst_output)
    instance.save()
    napi = net_api.API()
    napi.setup_instance_network_on_host(context, instance, host_name)
    LOG.info(_('Virtual machine %(name)s '
               '%(action)s from %(src)s to %(tgt)s. '
               'UUID on host: %(uuid)s. UUID in DB: '
               '%(uuid_in_db)s') %
             {'name': instance_name, 'action': action,
              'src': host_in_db, 'tgt': host_name, 'uuid': uuid,
              'uuid_in_db': uuid_in_db},
             instance=instance)
    # Send event notification
    info = {'msg': (_('Virtual machine {instance_name} '
                      'has been %(action)s from host '
                      '{source_host_name} to '
                      'host {destination_host_name}.') %
                    {'action': action}),
            'instance_id': instance.uuid,
            'instance_name': instance_name,
            'source_host_name': host_in_db,
            'destination_host_name': host_name}
    notifier = rpc.get_notifier(service='compute', host=host_name)
    notifier.info(context, 'compute.instance.log', info)


@logcall
def verify_ibmi_mobility_capable_for_migration(capability, instance_name):
    """
    The host must have property ibmi_lpar_mobility_capable= true if IBM i lpar
    on it want to do LPM.
    Check the capability.

    :param capability: The capability of a host
    :param instance_name: The name of the instance (for logging)
    """
    mobility_cap = capability.get('ibmi_lpar_mobility_capable')
    if not mobility_cap:
        error = (_("Cannot migrate %(lpar)s because the source and target "
                   "hosts are not capable of migrating IBM i virtual "
                   "machines. To migrate an IBM i virtual machine, "
                   "'IBM i Partition Mobility Capable' must be set to true on "
                   "the source and target hosts.") %
                 {'lpar': instance_name})
        LOG.exception(error)
        raise nova_exception.MigrationPreCheckError(reason=error)


def verify_ibmi_iorestricted_for_migration(iorestricted, instance_name):
    """
    The lpar must have property IsRestrictedIOPartition= true for LPM
    Check the capability.

    :param iorestricted: The iorestricted state of the  lpar
    :param instance_name: The name of the instance (for logging)
    """
    if not iorestricted:
        error = (_("Cannot migrate %(lpar)s because 'IBM i restricted I/O "
                   "mode' is not enabled. To migrate an IBM i virtual "
                   "machine, IBM i 'restricted I/O mode' property must be "
                   "enabled.")
                 % {'lpar': instance_name})
        LOG.exception(error)
        raise nova_exception.MigrationPreCheckError(reason=error)


def verify_ibmi_proc_compat_mode(host_compat_modes, instance_name):
    """
    The IBM i lpar migration only support on Power7 and above.

    :param host_compat_modes: All compat mode a host supports
    :param instance_name: The name of the instance (for logging)
    """
    mode_list = host_compat_modes.split(',')
    target_model = proc_utils.extract_highest_compat_mode(mode_list)
    if target_model < 7:
        error = (_("Cannot migrate %(lpar)s because live migration of an IBM "
                   "i virtual machine is only supported on POWER7"
                   " and above.") %
                 {'lpar': instance_name})
        LOG.exception(error)
        raise nova_exception.MigrationPreCheckError(reason=error)


def _is_instance_on_source_host(context, source_host, instance_uuid,
                                instance_name):
    """
    While handling possible out of band migration, we send a message to
    source host, asking if instance is still on the source host.
    """
    # Cooperative yield
    greenthread.sleep(0)
    answer = api.PowerVCComputeRPCAPI().is_instance_on_host(context,
                                                            instance_uuid,
                                                            instance_name,
                                                            source_host)
    if answer:
        LOG.info(_('Virtual machine %(inst)s is being managed by remote host '
                   '%(host)s. This could indicate the virtual machine is on '
                   'two hosts simultaneously after migration '
                   'outside of PowerVC') %
                 {'inst': instance_name, 'host': source_host})
    else:
        LOG.info(_('Instance %(inst)s is not being managed by remote '
                   'host %(host)s.') %
                 {'inst': instance_name, 'host': source_host})
    return answer


def get_dest_managed_by(dst_comp_info):
    """This is called by driver.check_can_live_migrate_destination and
    kvmhmc_manager._post_live_migration to determine the management console
    that the destination host is managed by.
    :param dst_comp_info destination compute node information passed to
           check_can_live_migrate_destination or _post_live_migration
    """
    dst_managed_by = dst_comp_info.get('stats', {}).get('managed_by')
    if dst_managed_by:
        dst_str = MANAGED_BY_STR.get(dst_managed_by.lower())
    else:
        dst_str = None
    return dst_str


def set_lpar_slot_info_from_map(instance, vios_uuid, wwpn_slot_map):
    """ Save wwpn and lpar_slot_info into instance_system_metadata.
    vios_uuid: [wwpn_lpar-slot-no1, wwpn2_lpar-slot-no2]
    vios_uuid: [wwpn_lpar-slot-no3]

    :param instance: The instance object.
    :param vios_uuid: uuid of vios
    :param wwpn_slot_map: The wwpn slot map in the below format
        {'wwpn1': ['slot1', 'slot2']}
    """
    LOG.debug("Enter: set_lpar_slot_info_from_map vios %(vios)s wwpn_slot_map"
              " %(wwpn_slot_map)s " % dict(vios=vios_uuid,
                                           wwpn_slot_map=wwpn_slot_map))
    for wwpn in wwpn_slot_map:
        for lpar_slot in wwpn_slot_map[wwpn]:
            _add_slot_data_to_sys_meta(instance, vios_uuid, wwpn, lpar_slot)


def _add_slot_data_to_sys_meta(instance, vios_uuid, wwpn, slot):
    """Add the slot info into the system metadata
    """
    LOG.debug("Enter: _add_slot_data_to_sys_meta vios %(vios)s wwpn %(wwpn)s "
              "slot %(slot)s" % dict(vios=vios_uuid, wwpn=wwpn, slot=slot))
    smd = instance.get("system_metadata")
    key = __get_vios_slot_key(vios_uuid)
    wwpn_slot_str = ""
    if key in smd:
        wwpn_slot_str = smd[key] + ","
    wwpn_slot_str += __build_wwpn_slot_str(wwpn, slot)

    smd[key] = wwpn_slot_str
    LOG.debug("Append wwpn_slot_str %s" % wwpn_slot_str)


def set_lpar_slot_info_from_fc_data(instance, fc_data):
    """Save the wwpn and slot info into sys metadata
    :param instance: instance object
    :param fc_data: A list of
             {'phys_wwpn': port,
              'client_slot': slot,
              'vios_id': vios_id}
    """
    LOG.debug("Enter: set_lpar_slot_info_from_fc_data %(fc_data)s" %
              dict(fc_data=fc_data))

    for data in fc_data:
        _add_slot_data_to_sys_meta(instance, data['vios_id'],
                                   data['phys_wwpn'], data["client_slot"])


def __build_wwpn_slot_str(wwpn, slot):
    """Return wwpn_slot string.
    """
    return wwpn + "_" + slot


def __get_vios_slot_key(vios_uuid):
    """Returns vios key to be used in the system metadata to save the slot
       info.
    """
    return VIOS_KEY + vios_uuid


def get_lpar_slot_info(instance):
    """Gets the lpar slot info from the instance's system metadata.

    :param instance: The instance object.
    :return: A list of
             {'phys_wwpn': port,
              'client_slot': slot,
              'vios_id': vios_id}

    """
    fc_data = []
    sys_metadata = instance.get("system_metadata")
    if not sys_metadata:
        LOG.info("No sysmetadata for instance %s" % instance)
        return None
    LOG.debug("inside get_lpar_slot_info sys_meta %s" % sys_metadata)
    for key in sys_metadata:
        if VIOS_KEY in key:
            npiv_port_str = sys_metadata[key]
            vios_id = key.split("_")[1]
            for wwpn_slot in npiv_port_str.split(","):
                port, slot = parse_wwpn_slot_str(wwpn_slot)
                if port and slot:
                    adapter_data = {'phys_wwpn': port,
                                    'client_slot': slot,
                                    'vios_id': vios_id}
                    fc_data.append(adapter_data)
    LOG.debug("Return fc_data %s" % fc_data)
    return fc_data


def clear_lpar_slot_info(instance):
    """Clears the slot info from system metadata

    :param instance: the instance dict containing system metadata
    """
    sys_metadata = instance.get("system_metadata")
    if not sys_metadata:
        LOG.info("No sysmetadata to clear for instance %s" % instance)
        return

    LOG.debug("Enter: clear_lpar_slot_info %s" % sys_metadata)
    for key in list(sys_metadata):
        if VIOS_KEY in key:
            value = sys_metadata.pop(key)
            LOG.debug("Removed %(value)s from metadata for key %(key)s" %
                      dict(value=value, key=key))


def parse_wwpn_slot_str(wwpn_slot):
    """Parse the wwpn slot string and retuns wwpn, slot
    """
    port_slot_list = wwpn_slot.split('_')
    if len(port_slot_list) != 2:
        return None, None

    port = port_slot_list[0]
    slot = port_slot_list[1]
    return port, slot
