# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Processor-related utilities."""

import re

# Multiplier for determining # of processor units given # of vcpus
IBM_POWERVM_VCPU_MULTIPLIER = 0.5


def check_proc_compat_mode_active_migration(preferred_mode=None,
                                            current_mode=None,
                                            supported_modes=[]):
    """Checks whether an LPAR can live migrate to the dest host based upon its
       supported compatibility modes

       :param preferred_mode: preferred compat mode of the LPAR
       :param current_mode: current compat mode of the LPAR
       :param supported_modes: proc compat modes supported by the dest host
       :returns: True if the LPAR can migrate and False otherwise
    """

    def _check_current_compat_mode(current_mode, supported_modes):
        return current_mode in supported_modes

    return (_check_current_compat_mode(current_mode, supported_modes) and
            _check_preferred_compat_mode(preferred_mode, supported_modes))


def check_proc_compat_mode_inactive_migration(preferred_mode=None,
                                              supported_modes=[]):
    """Checks whether an LPAR can cold migrate to the dest host based upon its
       supported compatibility modes

       :param preferred_mode: preferred compat mode of the LPAR
       :param supported_modes: proc compat modes supported by the dest host
       :returns: True if the LPAR can migrate and False otherwise
    """

    return _check_preferred_compat_mode(preferred_mode, supported_modes)


def _check_preferred_compat_mode(preferred_mode, supported_modes):
    """Checks whether the LPAR's preferred mode is supported

       :param preferred_mode: preferred compat mode of the LPAR
       :param supported_modes: proc compat modes supported by the dest host
       :returns: True if the preferred mode is supported and False otherwise
    """

    return (True if preferred_mode == 'default' else
            preferred_mode in supported_modes)


def get_num_proc_units(vcpus):
    """Get the default # of proc units given the # of vcpus.

       :param vcpus: Number of virtual cpus (int)
       :returns: number of processor units (float)
    """

    return IBM_POWERVM_VCPU_MULTIPLIER * vcpus


def extract_highest_compat_mode(compat_modes_list):
    # This method extract the highest compatibility mode to determine the model
    # of the Power host
    modes = []
    pattern = "^(power)([0-9]+)(\+?)$"
    for mode in compat_modes_list:
        match = re.search(pattern, mode.lower())
        if match:
            modes.append(int(match.group(2)))
    modes = sorted(modes)
    if modes:
        return modes[-1]
    else:
        return 0


def le_guest_support_from_proc_compat(compat_modes):
    """
    Determine whether a host support little endian vm based on the processor
    compatibility modes.
    If a host support Power8 or above, it supports litten endian VM.
    :param compat_modes: It's the compatibility_modes from host_stats.
    """
    compat_modes_list = compat_modes.split(',')
    target_mode = extract_highest_compat_mode(compat_modes_list)
    if target_mode >= 8:
        return True
    else:
        return False
