# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""
Implements common cinder volume helper functions needed for ibmpowervm
virt driver(both IVM and HMC)
"""
from urllib3.exceptions import ReadTimeoutError
from nova import objects
from nova import block_device
from oslo_utils import excutils
from oslo_serialization import jsonutils
from oslo_log import log as logging
from oslo_service import loopingcall as loopingcall

from oslo_config import cfg

from powervc_nova import logcall
from powervc_nova.common import block_device_utils
from powervc_nova.compute import common_constants
from powervc_nova.conductor import api as conductor
from powervc_nova.virt.ibmpowervm.common import exception as pvcex
from powervc_nova.volume import cinder as volume

from powervc_nova import _

LOG = logging.getLogger(__name__)
MAX_VOLUME_CHECK_RETRY = 10


CONF = cfg.CONF
CONF.import_opt('volume_status_check_max_retries',
                'powervc_nova.compute.kvmhmc_manager')
CONF.import_opt('volume_status_check_interval',
                'powervc_nova.compute.kvmhmc_manager')
CONF.import_opt('host_storage_type',
                'powervc_nova.virt.ibmpowervm')


def _populate_image_metadata(meta, root_device_name, cinder_volume_ids):
    meta['min_disk'] = 1
    meta['properties']['root_device_name'] = root_device_name

    bdm = _generate_block_device_mapping(cinder_volume_ids)
    meta['properties']['block_device_mapping'] = jsonutils.dumps(bdm)
    meta['properties']['bdm_v2'] = jsonutils.dumps(True)

    return meta


def is_boot_volume(volume):
    return ('is_boot_volume' in volume['volume_metadata']
            and volume['volume_metadata']['is_boot_volume'] == 'True')


@logcall
def _generate_block_device_mapping(volume_ids):
    '''
    :param volumes: list of volume ids, assumed to be in the correct order
    :returns bdm_list: list of block device mapping dictionaries
    '''
    bdm_list = []
    for boot_index, volume1 in enumerate(volume_ids):
        d = {'source_type': 'volume',
             'destination_type': 'volume',
             'volume_id': volume1,
             'device_type': 'disk',
             'boot_index': boot_index,
             'delete_on_termination': True}
        bdm_dict = block_device.BlockDeviceDict(d)

        bdm_list.append(bdm_dict.get_image_mapping())

    return bdm_list


class PowerVCNovaVirtDriverVolumeAPI(object):

    """
    PowerVC virtual driver volume helper functions.
    """

    def __init__(self):
        self.volume_api = volume.PowerVCVolumeAPI()
        self.conductor_api = conductor.PowerVCConductorAPI()

    def _valid_image(self, image):
        """
        double check the image meta data passed in.
        """
        if image.get('id') and image.get('properties') \
                and image.get('name'):
            return True
        else:
            return False

    @logcall
    def _wait_for_clone(self, context, image_volume, count,
                        wait_forever=False):
        count['count'] += 1
        tries = count['count']
        vol = None
        try:
            LOG.info(_("Call to get volume status %s") % image_volume['id'])
            vol = self.volume_api.get(context, image_volume['id'])
        except ReadTimeoutError as ex:
            # in case of a read timeout alone continue
            LOG.error(ex)
        if vol:
            if vol['status'] == 'available' or vol['status'] == 'in-use':
                msg = (_("The allocation of the image volume was successful: "
                         "image volume %(vol)s after tries=%(tries)s") %
                       locals())
                LOG.info(msg)
                raise loopingcall.LoopingCallDone()
            elif vol['status'] == 'error':
                msg = (_("The allocation of the image volume either failed or "
                         "timed out: "
                         "image volume %(vol)s tries=%(tries)s") %
                       locals())
                LOG.warn(msg)
                raise loopingcall.LoopingCallDone()
            elif (not wait_forever and
                    tries > CONF.volume_status_check_max_retries):
                msg = (_("The allocation of the image volume either failed or "
                         "timed out: "
                         "image volume %(vol)s tries=%(tries)s") %
                       locals())
                LOG.warn(msg)
                raise loopingcall.LoopingCallDone()

    def _check_volume_status(self, context, volume, method, interval=1,
                             initial_delay=None, wait_forever=False):
        tries = {'count': 0}
        timer = loopingcall.FixedIntervalLoopingCall(method,
                                                     context, volume,
                                                     tries, wait_forever)
        timer.start(interval, initial_delay).wait()
        timer.stop()
        vol = self.volume_api.get(context, volume['id'])
        return vol['status']

    def _wait_for_extend(self, context, volume, count):
        count['count'] += 1
        tries = count['count']
        vol = self.volume_api.get(context, volume['id'])
        if vol['status'] == 'in-use':
            raise loopingcall.LoopingCallDone()
        elif (vol['status'] == 'error_extending' or vol['status'] == 'error'):
            msg = (_("Error increasing size of volume: "
                     "volume %(volume)s tries=%(tries)s") %
                   locals())
            LOG.warn(msg)
            raise loopingcall.LoopingCallDone()

    def extend_volume(self, context, volume_id, size):
        volume = {'id': volume_id}
        self.volume_api.extend_volume(context, volume_id, size)
        # _check_volume_status is green thread safe blocking call
        status = self._check_volume_status(context, volume,
                                           self._wait_for_extend, 200, 60)
        # Only error state concerns us here.
        if status == "error" or status == 'error_extending':
            ex_args = {'volume_id': volume_id}
            raise pvcex.IBMPowerVMErrorExtendingVolume(**ex_args)

    def delete_volume(self, context, volume_id):
        """
        Deletes volume by given volume id.
        :param context: Auth Context
        :param volume_id: Volume-Id to be resized
        """
        self.volume_api.delete(context, volume_id)

    @logcall
    def _create_volume(self, context, src_vol, instance, display_name,
                       description, vol_meta):
        """
         Create the volume from the source volume and start it copying
         The capture uses the scheduler
         volume_type will be replicated from source volume
         """
        LOG.info(_("Creating an image volume '%(name)s' with "
                   "metadata:'%(image)s'")
                 % dict(name=display_name, image=vol_meta))
        image_volume = self.volume_api.create(context, src_vol['size'],
                                              display_name, description,
                                              source_vol=src_vol,
                                              metadata=vol_meta)
        status = image_volume['status']
        error_states = ['error']
        self._validate_volume_capture_state(instance, status, image_volume,
                                            error_states)
        return image_volume

    @logcall
    def _clone_volume(self, context, image, instance, volume,
                      boot_vol, target_unique_id):
        # set image volume meta data
        image_volume_meta = {'image_id': image['id'],
                             'instance_uuid': instance['uuid'],
                             'is_image_volume': 'True',
                             }
        if boot_vol:
            image_volume_meta['is_boot_volume'] = 'True'
        display_name = "Image %s volume %s" % (image['name'], target_unique_id)
        description = "Volume for image %s" % image['name']

        image_volume = self._create_volume(context, volume,
                                           instance,
                                           display_name,
                                           description,
                                           image_volume_meta)

        return image_volume

    @logcall
    def instance_volume_capture(self, context, instance, image,
                                additional_volumes):
        """
        Helper function to capture the instance's boot volumes by cloning
        them into image volumes and setting volume meta for the image volumes.
        Return image object with the updated meta data based on the
        captured information.

        ASSUMPTIONS:
        ============
        1. virtual optical device has been removed when this function
           is called.
        2. image passed in has been created.

        :param context: The security context.
        :param instance: The instance object
        :param image: The glance image object
        :return: updated image object otherwise exception raised
        """

        context = context.elevated()

        if not image or not self._valid_image(image):
            # image is empty or malformed, give up.
            raise pvcex.IBMPowerVMInvalidImageObject(
                instance_name=instance['name'])

        # Get the list of BDMs for this VM
        bdms = objects.BlockDeviceMappingList.\
            get_by_instance_uuid(context, instance['uuid'])

        self._validate_bdms(context, instance, bdms, additional_volumes)

        sorted_bdms = block_device_utils.sort_instance_bdms(bdms)
        vm_volumes_captured = []
        vm_volumes_not_captured = []
        captured_volume_ids = []
        root_device_name = None
        try:
            image_volumes = []
            for bdm in sorted_bdms:
                if bdm.no_device:
                    continue

                if bdm.is_volume:
                    # get the volume
                    volume = self.volume_api.get(context, bdm.volume_id)
                    boot_vol = is_boot_volume(volume)

                    # Only capture boot volumes and additional volumes.
                    if(not boot_vol and
                       bdm.volume_id not in additional_volumes):
                        vm_volumes_not_captured.append(volume)
                        continue
                    # We set the root device name to the first volume we
                    # capture.  This volume will have boot_index = 0 in the
                    # image's BDMs.
                    if not root_device_name:
                        root_device_name = bdm.device_name
                    vm_volumes_captured.append(volume)
                    id_for_target_name = len(captured_volume_ids) + 1
                    image_volume = self._clone_volume(context, image,
                                                      instance, volume,
                                                      boot_vol,
                                                      id_for_target_name)
                    captured_volume_ids.append(image_volume['id'])
                    image_volumes.append(image_volume)
            # The image volumes are ready. Need to update the image metadata
            # accordingly for the image volumes.
            image = _populate_image_metadata(image, root_device_name,
                                             captured_volume_ids)
            self._wait_on_clones(context, instance, image_volumes)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception(_('An error occurred while cloning the '
                                'instance\'s volumes.'))
                self.delete_image_volumes_on_capture_fail(context,
                                                          captured_volume_ids)
        # Per our project architect we must log which volumes of a VM are
        # captured and which are not at a log level that is always logged.
        # Do not remove these logging statements even if someone thinks
        # they are cluttering the logs in the future.
        for volume in vm_volumes_captured:
            LOG.info(_('Captured instance '
                       '"%(instance_name)s, %(instance_uuid)s" '
                       'volume "%(volume_name)s, %(volume_uuid)s" '
                       'for image "%(image_name)s, %(image_id)s".') %
                     {'instance_name': instance['display_name'],
                      'instance_uuid': instance['uuid'],
                      'volume_name': volume['display_name'],
                      'volume_uuid': volume['id'],
                      'image_name': image['name'],
                      'image_id': image['id']}
                     )
        for volume in vm_volumes_not_captured:
            LOG.info(_('Instance "%(instance_name)s, %(instance_uuid)s" '
                       'volume "%(volume_name)s, %(volume_uuid)s" '
                       'not captured for '
                       'image "%(image_name)s, %(image_id)s".') %
                     {'instance_name': instance['display_name'],
                      'instance_uuid': instance['uuid'],
                      'volume_name': volume['display_name'],
                      'volume_uuid': volume['id'],
                      'image_name': image['name'],
                      'image_id': image['id']}
                     )
        return image

    def _validate_bdms(self, ctxt, instance, bdms, additional_vols):

        # Validate the root bdm
        root_dev_name = instance['root_device_name']
        root_bdm = block_device.get_root_bdm(bdms)
        if not root_bdm:
            ex_args = {'instance_name': instance['name']}
            raise pvcex.IBMPowerVMMissingRootBDM(**ex_args)

        rootdev = root_bdm.device_name
        if root_dev_name != rootdev:
            ex_args = {'instance_root': root_dev_name,
                       'instance_name': instance['name'],
                       'bdm_root_dev': rootdev}
            raise pvcex.IBMPowerVMRootDeviceMismatch(**ex_args)

        if not root_bdm.is_volume:
            ex_args = {'instance_name': instance['name']}
            raise pvcex.IBMPowerVMInvalidRootDevice(**ex_args)

        # Check all BDM statuses and ensure all additional volumes specified
        # are attached to the VM.
        vols_not_found = list(additional_vols)
        for bdm in bdms:
            if bdm.no_device or bdm['connection_info'] is None:
                LOG.debug("BDM had no device or connection info. Skip it %s"
                          % bdm)
                continue

            if bdm.is_volume:
                if bdm.volume_id in vols_not_found:
                    vols_not_found.remove(bdm.volume_id)
                # get the volume
                volume = self.volume_api.get(ctxt, bdm.volume_id)

                # Only validate volume status for the volumes that will be
                # captured.  This will allow capture of a subset of VM volumes
                # when some volumes are in error state.
                if is_boot_volume(volume) or bdm.volume_id in additional_vols:
                    if volume['status'].lower() != 'in-use':
                        ex_args = {'instance_name': instance['name'],
                                   'volume': volume['display_name'],
                                   'state': volume['status']}
                        raise pvcex.IBMPowerVMInvalidCaptureVolume(**ex_args)

        # Validate that all additional volume IDs are attached to the VM
        if vols_not_found:
            ex_args = {'instance_name': instance['name'],
                       'volumes': vols_not_found}
            raise pvcex.IBMPowerVMVolumesNotAttached(**ex_args)

    @logcall
    def delete_image_volumes_on_capture_fail(self, ctxt, volume_ids):
        LOG.info(_('Removing image volumes due to failed capture: '
                   '%(volume_ids)s') % {'volume_ids': volume_ids})
        for volume1 in volume_ids:
            try:
                LOG.debug('Unsetting is_image_volume on volume %s'
                          % volume1)
                metadata = {'is_image_volume': "False"}
                self.volume_api.update_volume_metadata(ctxt,
                                                       volume1,
                                                       metadata)
                LOG.debug('Unsetting image_id on volume %s' % volume1)
                metadata = {'image_id': ""}
                self.volume_api.update_volume_metadata(ctxt,
                                                       volume1,
                                                       metadata)
                LOG.debug('Deleting volume %s' % volume1)
                self.volume_api.delete_volume(ctxt, volume1)
            except Exception:
                exc_args = {'volume_id': volume1}
                LOG.exception(_('An error occurred while removing '
                                'image volume (volume_id)%s')
                              % exc_args)

    @logcall
    def _validate_volume_capture_state(self, instance, status,
                                       volume, error_states):
        if status in error_states:
            create_fail_msg = volume['volume_metadata'].get(common_constants.
                                                            CREATE_FAILURE_KEY)
            if not create_fail_msg:
                create_fail_msg = _('No detailed volume error message.')
            ex_args = {'hstatus': volume['health_status'],
                       'volume_error': create_fail_msg,
                       'instance_name': instance['name']}
            raise pvcex.IBMPowerVMImageVolumeCloneFailure(**ex_args)

    @logcall
    def _wait_on_clones(self, context, instance, image_volumes):
        # check every 6 seconds and we are going to loop by default 300
        # times this will allow us wait for 30 min. Plenty of time for
        # a misbehaving clone to finish.
        wait_interval = CONF.volume_status_check_interval
        error_states = ['error', 'creating']
        for image_volume in image_volumes:
            # if the type is "cluster" then this is ssp
            stg_type = 'fc'
            try:
                stg_type = self.volume_api.get_storage_type_for_volume(
                    context, image_volume['id'])
            except Exception:
                stg_type = 'fc'

            wait_forever = (stg_type == 'cluster')
            # _check_volume_clone_status is green thread safe blocking call
            status = self._check_volume_status(context, image_volume,
                                               self._wait_for_clone,
                                               wait_interval,
                                               wait_forever=wait_forever)
            self._validate_volume_capture_state(instance, status,
                                                image_volume, error_states)

    @logcall
    def get_instance_volumes(self, context, instance, volume_category='all'):
        volume_list = []
        # Get the list of BDMs for this VM
        bdms = objects.BlockDeviceMappingList.\
            get_by_instance_uuid(context, instance['uuid'])
        volume = {}
        for bdm in bdms:
            if 'volume_id' not in bdm:
                continue
            try:
                volume = self.volume_api.get(context, bdm['volume_id'])
            except Exception:
                LOG.exception(_("Unable to retrieve volume %s"
                                % bdm['volume_id']))
            if not volume:
                continue
            if is_boot_volume(volume):
                if volume_category in ['boot', 'all']:
                    volume_list.append(bdm['volume_id'])
            else:
                if volume_category in ['data', 'all']:
                    volume_list.append(bdm['volume_id'])
        return volume_list
