# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from functools import partial
import six

from nova import objects
from oslo_config import cfg
from oslo_log import log as logging

from powervc_nova import _
from powervc_nova.objects.compute import dom as compute_dom
from powervc_nova.virt.ibmpowervm.common import constants
from powervc_nova.virt.ibmpowervm.common import enterprise_pool as common_ep
from powervc_nova.virt.ibmpowervm.hmc.k2 import enterprise_pool_wrapper as\
    ep_wrap
from powervc_nova.virt.ibmpowervm.hmc.k2 import enterprise_pool_member_wrapper\
    as ep_member_wrap
from powervc_nova.virt.ibmpowervm.hmc.k2oprguard import K2OperGuard
from powervc_nova.virt.ibmpowervm.hmc import exception
from powervc_oslo.k2.k2operator import K2Error
from powervc_nova.virt.ibmpowervm.hmc.k2 import constants as k2const
from powervc_nova.virt.ibmpowervm.hmc.k2 import mgmt_console_wrapper

LOG = logging.getLogger(__name__)

CONF = cfg.CONF
POST_RETRIES = CONF.ibmpowervm_update_collision_retries

# MS = ManagedSystem, MT = MachineType, MSN = ModelSerialNumber
MS_MTMSN = 'ManagedSystemMachineTypeModelSerialNumber'
MS_MACHINE_TYPE = 'MachineType'
MS_MODEL = 'Model'
MS_SERIAL_NUMBER = 'SerialNumber'
# K2 resources
POWER_ENTERPRISE_POOL = 'PowerEnterprisePool'
POWER_ENTERPRISE_POOL_MEMBER = 'PowerEnterprisePoolMember'


def sync_pool_for_host(context, host):
    """Syncs the pool info corresponding to a specific host managed by a HMC.

    Since this code is ran from each individual compute driver process, it will
    only sync membership for this specific host (that corresponds to hostname).
    It will also sync the overall pool properties, such as the available mobile
    cores, total mobile cores, etc., that aren't tied to a specific member.

    :param context: The security context for the query
    :param host: The host object
    """
    # Fetch all the pools from the HMC
    pools = get_pools(host)

    pool_wrapper = None
    pool_member_uuid = None

    if pools:
        # Determine if this host belongs to any of the pools
        for entry in pools.feed.entries:
            pool_uuid = entry.properties['id']

            pool_member_uuid = _find_pool_member_uuid(host, pool_uuid)
            if pool_member_uuid:
                # We found a pool member corresponding to the host, so break
                pool_wrapper = ep_wrap.EnterprisePoolEntryWrapper(entry)
                break

    pool_dict = {}
    if pool_wrapper:
        pool_dict = {
            'uuid': pool_wrapper.get_uuid(),
            'name': pool_wrapper.pool_name,
            'master_hmc_mtmsn': pool_wrapper.master_console_mtmsn,
            'available_mobile_cores': pool_wrapper.available_mobile_cores,
            'total_mobile_cores': pool_wrapper.total_mobile_cores,
            'compliance_state': pool_wrapper.compliance_state}

    common_ep.sync_pool_for_host(context, pool_dict, pool_member_uuid,
                                 host.host_id)


def _find_pool_member_uuid(host, pool_uuid):
    """Returns the pool member UUID of the host if it is a member of the pool.

    :param host: Host object
    :param pool_uuid: The UUID of the pool to search for a relevant member.
    :returns: The pool member UUID of the host if it is a member of the pool
              corresponding to pool_uuid, None otherwise.
    """
    # Fetch all the pool members from the HMC
    pool_members = get_pool_members(host, pool_uuid)

    if not pool_members:
        return None

    for entry in pool_members.feed.entries:
        mtmsn = _get_pool_member_mtms(entry)
        if mtmsn == host.host_id:
            # Found a member that corresponds to the host
            return entry.properties['id']

    # No member was found corresponding to the host
    return None


def _get_pool_member_mtms(member_entry):
    """Returns the MTMS (machine type, model, serial num) of the pool member.

    :param member_entry: K2 member entry object
    :returns: A string representation of the MTMS in the form: "MTM_SN". For
              example: "8231E2D_109EFBT".
    """
    k2_elem = member_entry.element
    mtmsn_elem = k2_elem.find(MS_MTMSN)
    mt = mtmsn_elem.findtext(MS_MACHINE_TYPE)
    mm = mtmsn_elem.findtext(MS_MODEL)
    msn = mtmsn_elem.findtext(MS_SERIAL_NUMBER)

    return common_ep.convert_mtms_to_hostname(mt, mm, msn)


def get_pool_members(host, pool_uuid):
    """Get all members of the enterprise pool with the specified UUID.

    :param host: Host object
    :param pool_uuid: UUID of the pool of interest
    :returns: K2 response object
    """
    members = None
    with K2OperGuard(host=host) as operator:
        try:
            LOG.debug('CoD K2 read call to %s/%s/%s' %
                      (POWER_ENTERPRISE_POOL, pool_uuid,
                       POWER_ENTERPRISE_POOL_MEMBER))
            members = operator.read(POWER_ENTERPRISE_POOL, pool_uuid,
                                    POWER_ENTERPRISE_POOL_MEMBER,
                                    xag=[None])
        except Exception as e:
            LOG.error(_("Could not GET pool members for pool: %(pool)s. K2 "
                        "returned with an error: %(error)s") %
                      {'pool': pool_uuid,
                       'error': six.text_type(e)})
    return members


def get_pools(host):
    """Get all enterprise pools.

    :param host: Host object
    :returns: K2 response object
    """
    pools = None
    try:
        with K2OperGuard(host=host) as operator:
            LOG.debug('CoD K2 read call to %s' % POWER_ENTERPRISE_POOL)
            pools = operator.read(POWER_ENTERPRISE_POOL, xag=[None])
    except Exception as e:
        LOG.error(_("Could not GET enterprise pools. K2 returned with an "
                    "error: %(error)s") %
                  {'error': six.text_type(e)})
    return pools


def get_member(host, pool_uuid, member_uuid):
    """Get pool member with specified pool UUID and member UUID.

    :param host: Host object
    :param pool_uuid: UUID of the pool that member UUID is in
    :param member_uuid: UUID of the member of interest
    :returns: K2 response object
    """
    member = None
    try:
        with K2OperGuard(host=host) as k2opr:
            LOG.debug('CoD K2 read call to %s/%s/%s/%s' %
                      (POWER_ENTERPRISE_POOL, pool_uuid,
                       POWER_ENTERPRISE_POOL_MEMBER, member_uuid))
            member = k2opr.read(POWER_ENTERPRISE_POOL,
                                rootId=pool_uuid,
                                childType=POWER_ENTERPRISE_POOL_MEMBER,
                                childId=member_uuid,
                                xag=[None])
    except Exception as e:
        LOG.error(_("Could not get the corresponding pool member data for "
                    "host %(host)s in pool %(pool)s. K2 returned with an "
                    "error: %(error)s") %
                  {'host': host.host_id,
                   'pool': pool_uuid,
                   'error': six.text_type(e)})

    return member


def validate_set_mobile_cores_prereqs(context, host, pool):
    """Validate prereqs for setting the mobile cores for this host.

    :param context: The security context for the query
    :param host: Host object
    :param pool: powervc_nova.objects.compute.dom.EnterprisePool object
    :returns: Tuple of form (result, [reason]), where result is a boolean
              representing whether or not the validation was successful, and
              the reason is specified in the case it was not successful
    """
    compute_nodes = objects.ComputeNodeList.get_all_by_host(
        context, host.host_id) or []
    compute_node = None if not len(compute_nodes) else compute_nodes[0]

    if not compute_node:
        msg = (_('Compute node corresponding to %s could not be found.') %
               host.host_id)
        LOG.error(msg)
        return (False, msg)

    # Enterprise pool POST operations to non-master HMCs will fail if the HMC
    # version is older than the 850 release.
    hmc_version = compute_node.get('stats', {}).get(constants.HMC_VERSION)

    if not hmc_version:
        LOG.warning(_('Version of HMC managing host %s could not be '
                      'determined.') % compute_node.host)

    hmc_version_wrap = mgmt_console_wrapper.HMCVersion(hmc_version)
    hmc_version_850 = mgmt_console_wrapper.HMCVersion('V8R8.5.0')

    # If the HMC version is not known, then just assume it is >= 850. A warning
    # is logged above in the unlikely case this happens. Otherwise, actually
    # do the check.
    if hmc_version is not None and hmc_version_wrap < hmc_version_850:
        hmc_fact = compute_dom.ManagementConsoleFactory.get_factory()
        master_hmc = hmc_fact.find_hmc_by_mtmsn(context,
                                                pool.master_hmc_mtmsn)
        if not master_hmc:
            msg = (_('HMC with MTMSN (machine type, model, and serial '
                     'number) %s could not be found in the database.') %
                   pool.master_hmc_mtmsn)
            LOG.error(msg)
            return (False, msg)

        active_hmc = host.select_active_hmc()
        if not active_hmc:
            msg = _('There is not an active HMC managing host %s.' %
                    host.host_id)
            LOG.error(msg)
            return (False, msg)

        if master_hmc.access_ip != active_hmc.ipaddr:
            msg = (_('The HMC (%(active_hmc_ip)s) currently managing host '
                     '%(host)s is not the same as the master HMC of the '
                     'enterprise pool: %(master_hmc_ip)s. Make sure that '
                     'the host is managed by the master HMC of the '
                     'pool.') %
                   {'active_hmc_ip': active_hmc.ipaddr,
                    'host': host.host_id,
                    'master_hmc_ip': master_hmc.access_ip})
            LOG.error(msg)
            return (False, msg)

    return (True,)


def update_mobile_cores(host, pool_uuid, member_uuid, desired_mobile_cores):
    """Update the pool member's num of mobile cores via K2 (retrying helper).

    :param host: Host object
    :param pool_uuid: UUID of the enterprise pool
    :param member_uuid: UUID of the enterprise pool member
    :param desired_mobile_cores: Int number of desired value for mobile cores
    :returns: Tuple of (unreturned mobile cores, mobile cores, inactive cores)
              corresponding to the updated pool member.
    """
    # Retry the K2 update if it fails based on etag mismatch.
    for retries in xrange(POST_RETRIES + 1):
        try:
            member_wrapper = _update_mobile_cores(host, pool_uuid,
                                                  member_uuid,
                                                  desired_mobile_cores,
                                                  retries == POST_RETRIES)

            return (member_wrapper.unreturned_mobile_cores,
                    member_wrapper.mobile_cores,
                    member_wrapper.inactive_cores)
        except Exception as e:
            if isinstance(e, exception.IBMPowerVMUpdateFailed):
                LOG.debug('Retry #%s for _update_mobile_cores' % retries)
                # Retry for POST_RETRIES
                if retries < POST_RETRIES:
                    continue
            raise


def _update_mobile_cores(host, pool_uuid, member_uuid,
                         desired_mobile_cores, last_retry=False):
    """Update the pool member's number of mobile cores via K2.

    :param host: Host object
    :param pool_uuid: UUID of the enterprise pool
    :param member_uuid: UUID of the enterprise pool member
    :param desired_mobile_cores: Int number of desired value for mobile cores
    :param last_retry: Boolean value for deciding to stop retrying the update
                       request if we receive an ETAG mismatch back from K2
    :returns: updated CodPoolMemberEntryWrapper if the request was successful
    """
    member = get_member(host, pool_uuid, member_uuid)

    member_wrapper = ep_member_wrap.EnterprisePoolMemberEntryWrapper(
        member.entry)

    member_wrapper.mobile_cores = desired_mobile_cores

    etag = member.headers.get('etag')
    with K2OperGuard(host=host) as oper:
        try:
            LOG.debug('CoD K2 update call to %s/%s/%s/%s' %
                      (POWER_ENTERPRISE_POOL, pool_uuid,
                       POWER_ENTERPRISE_POOL_MEMBER, member_uuid))
            resp = oper.update(member.entry.element, etag,
                               POWER_ENTERPRISE_POOL, pool_uuid,
                               POWER_ENTERPRISE_POOL_MEMBER, member_uuid,
                               xag=[None])
            member_wrapper = ep_member_wrap.EnterprisePoolMemberEntryWrapper(
                resp.entry)
        except K2Error as e:
            if (hasattr(e, 'k2response') and e.k2response is not None and
                    e.k2response.status ==
                    k2const.HTTP_STATUS_ETAG_MISMATCH):
                if last_retry:
                    # Dump all the k2 requests & responses accumulated
                    # since the wrapper would not log single
                    # HTTP_STATUS_ETAG_MISMATCH failure
                    oper.dump_logs()
                raise exception.IBMPowerVMUpdateFailed(uuid=member_uuid)
            raise

        try:
            # Try to invalidate the cached K2 response for the managed system,
            # overall pool, and pool member in order to avoid stale values
            oper.invalidate_cache_elem(k2const.MGT_SYS, rootId=host.uuid,
                                       invalidate_feeds=True)
            oper.invalidate_cache_elem(POWER_ENTERPRISE_POOL, rootId=pool_uuid,
                                       invalidate_feeds=True)
            oper.invalidate_cache_elem(POWER_ENTERPRISE_POOL_MEMBER,
                                       rootId=member_uuid,
                                       invalidate_feeds=True)
        except Exception as e:
            LOG.warn(_('Error invalidating k2operator cache: %(exc)s') %
                     {'exc': six.text_type(e)})

        return member_wrapper


def set_mobile_cores(context, desired_mobile_cores, host):
    """Sets the mobile cores that a host is using to the desired value.

    :param context: The security context for the query
    :param desired_mobile_cores: the number of mobile cores that the host
                                 should use
    :param host: host object
    :returns: Tuple of (result, reason). For example, if the mobile cores
              were successfully updated, then a tuple of (True,) will be
              returned. Otherwise, it would return a tuple of
              (False, 'Reason for failure').
    """
    extra_validation_partial = partial(validate_set_mobile_cores_prereqs,
                                       context=context, host=host)
    update_mobile_cores_partial = partial(update_mobile_cores, host=host)

    return common_ep.set_mobile_cores(
        context, desired_mobile_cores, host.host_id,
        update_mobile_cores_partial,
        extra_validation_partial=extra_validation_partial)
