# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
"""IBMPowerVM Instance Type

IBMPowerVM Instance Type.
"""

from nova import rpc

from nova.compute import power_state
from nova.context import get_admin_context

from oslo_log import log as logging

from powervc_nova import _
from powervc_nova import logcall
from powervc_nova.virt.ibmpowervm.common import exception
from powervc_nova.virt.ibmpowervm.common import proc_utils
from powervc_nova.virt.ibmpowervm.common.constants import IBMI_OS_TYPE,\
    DEFAULT_PROC_POOL, SHR_PROC_POOL_CAPABLE
from powervc_nova.virt.ibmpowervm.hmc import exception as hmcexception

LOG = logging.getLogger(__name__)

DED_SHARING_MODES = ('share_idle_procs', 'keep_idle_procs',
                     'share_idle_procs_active', 'share_idle_procs_always')
DEFAULT_SHARING_MODE = DED_SHARING_MODES[0]
UNCAP_WEIGHT_UP_BOUND = 255
UNCAP_WEIGHT_DEFAULT = 128
AVAIL_PRIOR_UP_BOUND = 255
AVAIL_PRIOR_DEFAULT = 127
MEM_LOW_BOUND = 128
VCPUS_LOW_BOUND = 1
PROC_UNITS_LOW_BOUND = .1
MIN_PROC_UNIT_PER_VCPU = .1

# Strings used for exception messages
MIN_MEM_STR = _('minimum memory')
MIN_VCPU_STR = _('minimum virtual CPUs')
MIN_PROC_UNIT_STR = _('minimum processing units')
MAX_MEM_STR = _('maximum memory')
MAX_VCPU_STR = _('maximum virtual CPUs')
MAX_PROC_UNIT_STR = _('maximum processing units')

MIN_STR = _('minimum')
DESIRED_STR = _('desired')
MAX_STR = _('maximum')

# Flavor extra_specs for memory and processor properties
EXPECTED_RESOURCE_SPECS = ['powervm:min_mem',
                           'powervm:max_mem',
                           'powervm:min_vcpu',
                           'powervm:max_vcpu',
                           'powervm:proc_units',
                           'powervm:min_proc_units',
                           'powervm:max_proc_units',
                           'powervm:dedicated_proc',
                           'powervm:uncapped',
                           'powervm:dedicated_sharing_mode',
                           'powervm:processor_compatibility',
                           'powervm:srr_capability',
                           'powervm:shared_weight',
                           'powervm:availability_priority',
                           'powervm:ame_expansion_factor',
                           'powervm:shared_proc_pool_name']

# Other acceptable extra_specs attributes
EXPECTED_OTHER_SPECS = ['powervm:boot_volume_type',
                        'powervm:storage_connectivity_group']


class InstanceType(object):
    """
    Class representing the instance's flavor
    """

    @logcall
    def __init__(self, instance, instance_type):
        self._inst_name = instance['name']
        self._inst_id = instance['uuid']
        self.instance_type = instance_type
        self.extra_specs = dict(instance_type['extra_specs'])
        self._resize_and_running = False
        self._host_id = None
        self._inst_os_type = instance['os_type']

        # Validate the extra spec keys
        for key in self.extra_specs.keys():
            # If the extra spec does not begin with powervm, ignore it
            if not key.startswith('powervm:'):
                continue
            # If the key starts with 'powervm:image_volume_type_', let it pass
            # as this is the prefix for volume type keys to be used during
            # multi-volume image deploy.
            if key.startswith('powervm:image_volume_type_'):
                continue

            # If the extra spec is a powervm extra spec, make sure it is
            # an attribute we support
            if key not in EXPECTED_RESOURCE_SPECS + EXPECTED_OTHER_SPECS:
                ex_args = {'key': key,
                           'instance_name': self._inst_name}
                LOG.error(_("Invalid attribute name '%(key)s' was passed"
                            " in as part of flavor extra specs for"
                            " virtual machine '%(instance_name)s'.") % ex_args)
                raise hmcexception.IBMPowerVMInvalidExtraSpec(**ex_args)
        # Populate the instance type attrs dictionary
        self._attrs = {}
        for attr in EXPECTED_RESOURCE_SPECS:
            self._attrs[attr.split(':')[1]] = self.extra_specs.get(attr)

    @logcall
    def validate(self, host, num_eth_adapters, curr_flavor=None):
        """Validate instance type attributes for create
        and resize, and fills out default values. Saves
        updated extra_specs to its public extra_specs
        dictionary.

        :param host: Host Object.
        :param num_eth_adapters: Number of virtual ethernet adapters. This is
                used when calculationg PHYP memory overhead on the host
        :param curr_flavor: Current flavor of the instance being resized.
            Passed in only during a resize operation.
        """

        self._host_id = host.host_id
        host_stats = host.get_host_stats_internal(refresh=True)
        update_needed = True
        # If curr_flavor is set, this is a resize.
        if curr_flavor:
            # Compare current flavor with requested flavor to determine if
            # any LPAR attributes are changing
            update_needed = self._is_changing(curr_flavor)
            if not update_needed:
                return update_needed

            # set self._resize_and_running to true if state is Running
            self._resize_and_running = (curr_flavor.get('state') ==
                                        power_state.RUNNING)
            # If the instance is running, make sure it has RMC connectivity,
            # otherwise the resize is not supported
            if (self._resize_and_running and
                    IBMI_OS_TYPE != self._inst_os_type):
                if curr_flavor['has_rmc_connectivity'] is False:
                    LOG.error(_("Resizing instance '%s' in"
                                " Active state is not supported without RMC"
                                " connectivity.") % self._inst_name)
                    raise hmcexception.IBMPowerVMNoRMCConnectivity(
                        instance_name=self._inst_name)

        # Validate memory
        self._validate_memory(host_stats, curr_flavor)

        # Validate cpu options
        self._validate_cpu(host_stats, curr_flavor)

        curr_srr_cap = None
        if curr_flavor and 'powervm:srr_capability' in curr_flavor:
            curr_srr_cap = curr_flavor.get('powervm:srr_capability').lower()
        # Validate simplified remote restart capability of host
        req_srr_cap = self._attrs['srr_capability']
        if req_srr_cap is not None:
            self._validate_srr_capability(host_stats, req_srr_cap.lower(),
                                          curr_srr_cap)

        curr_exp_factor = None
        if curr_flavor:
            if 'powervm:ame_expansion_factor' in curr_flavor:
                curr_exp_factor = curr_flavor.\
                    get('powervm:ame_expansion_factor')
        req_exp_factor = self._attrs['ame_expansion_factor']
        # If AME is requested to be false then the expansion factor must be 0
        # or else K2 won't disable the capability and still return 200 OK
        if req_exp_factor is not None:
            self._validate_ame_capability(host_stats, req_exp_factor,
                                          curr_exp_factor)

        # Validate that there are enough processor
        # and memory resources on the host
        if curr_flavor is not None:
            # If this is a resize operation, calculate deltas
            deltas = self._calculate_resize_deltas(curr_flavor)
            self._check_host_resources(host, host_stats, num_eth_adapters,
                                       resource_deltas=deltas,
                                       curr_flavor=curr_flavor)
        else:
            self._check_host_resources(host, host_stats, num_eth_adapters)

        # Format and save updated extra_specs
        for key, value in self._attrs.items():
            # Format floats
            if value is None:
                self.extra_specs['powervm:' + key] = value
            elif type(value) is float:
                if key == 'ame_expansion_factor':
                    self.extra_specs['powervm:' + key] = '%.2f' % value
                else:
                    self.extra_specs['powervm:' + key] = '%.1f' % value
            elif type(value) is bool:
                self.extra_specs['powervm:' + key] = '%s' % str(value).lower()
            else:
                self.extra_specs['powervm:' + key] = '%s' % value

        return update_needed

    def _is_changing(self, curr_flavor):
        # Helper method to compare current flavor with requested flavor
        # to determine if any LPAR attributes are changing
        is_changing = False
        for attr in curr_flavor.keys():
            if attr == 'memory_mb' or attr == 'vcpus':
                if int(curr_flavor[attr]) != self.instance_type.get(attr):
                    is_changing = True
                    break
            elif attr.startswith('powervm'):
                extra_specs = self.instance_type['extra_specs']
                # Check for the case where there is an attribute present in
                # the new flavor, that does not exist in the old flavor
                for new_extra_spec in extra_specs.keys():
                    if new_extra_spec not in curr_flavor.keys():
                        is_changing = True
                        break
                if attr == 'powervm:processor_compatibility':
                    # Processor compatibility is a special case, since
                    # 'default' is a possible value for pending
                    new_proc_compat = extra_specs.get(attr)
                    if new_proc_compat:
                        new_proc_compat = new_proc_compat.lower()
                        curr_des_proc_compat = curr_flavor[attr][0].lower()
                        curr_pend_proc_compat = curr_flavor[attr][1].lower()
                        if curr_pend_proc_compat == 'default':
                            if (new_proc_compat != 'default' and
                                    curr_des_proc_compat != new_proc_compat):
                                is_changing = True
                                break
                        # Compare pending value with new proc compat mode
                        elif curr_pend_proc_compat != new_proc_compat:
                            is_changing = True
                            break
                else:
                    new_attr = extra_specs.get(attr)
                    if (new_attr and
                            (curr_flavor[attr].lower() != new_attr.lower())):
                        is_changing = True
                        break

        return is_changing

    def _validate_memory(self, host_stats, curr_flavor=None):
        """Validate the user specified values for memory.
        Set defaults for min/max memory if not set.

        :param host_stats: host attributes
        :param curr_flavor: If this is a resize, current flavor of
            instance, otherwise None.
        """
        def validate_with_lmb(mem_attr, mem_value):
            lmb = int(host_stats['lmb_size'])

            """Checks that memory attribute is a multiple of LMB size
            and raises exception if not.

            :param mem_attr: string to represent min/desired/max memory
            :param mem_value: requested memory value to check
            """
            if (mem_value % lmb) != 0:
                ex_args = {'attr': mem_attr,
                           'value': mem_value,
                           'lmb': lmb,
                           'instance_name': self._inst_name}
                LOG.error(_("%(attr)s (%(value)d MB) is not a multiple of"
                            " memory region size (%(lmb)d MB) for"
                            " virtual machine '%(instance_name)s'.") % ex_args)
                raise hmcexception.IBMPowerVMNotLMBMultiple(**ex_args)

        # Deisred memory is required. Check that it is set,
        # it is a valid number, and it is at least 128 MB.
        try:
            desired_mem = int(self.instance_type['memory_mb'])
        except KeyError:
            ex_args = {'attr': _('desired memory'),
                       'instance_name': self._inst_name}
            LOG.error(_("A value for %(attr)s is required for virtual machine"
                        " '%(instance_name)s'.") % ex_args)
            raise hmcexception.IBMPowerVMAttrRequired(**ex_args)
        except ValueError:
            ex_args = {'attr': _('Desired memory'),
                       'value': '%s' % self.instance_type['memory_mb'],
                       'instance_name': self._inst_name}
            LOG.error(_("'%(value)s' is not a valid value for %(attr)s"
                        " for virtual machine '%(instance_name)s'.") % ex_args)
            raise hmcexception.IBMPowerVMAttrValueInvalid(**ex_args)
        if desired_mem < MEM_LOW_BOUND:
            ex_args = {'mem_requested': desired_mem,
                       'mem_min': MEM_LOW_BOUND,
                       'instance_name': self._inst_name}
            LOG.error(_("The requested memory (%(mem_requested)d MB) is less "
                        "than the minimum required value (%(mem_min)d MB) for"
                        " virtual machine '%(instance_name)s'.") % ex_args)
            raise exception.IBMPowerVMMemoryBelowMin(**ex_args)

        # If this is a resize, get current min/max values.
        if curr_flavor:
            curr_min_mem = int(curr_flavor.get('powervm:min_mem'))
            curr_max_mem = int(curr_flavor.get('powervm:max_mem'))
            if self._resize_and_running:
                # This is a resize and current state is running. Desired memory
                # cannot be outside of the current range of minimum and maximum
                # memory.
                if desired_mem < curr_min_mem or desired_mem > curr_max_mem:
                    ex_args = {'attr': _('memory'),
                               'desired': '%s' % desired_mem,
                               'min': '%s' % curr_min_mem,
                               'max': '%s' % curr_max_mem,
                               'instance_name': self._inst_name}
                    LOG.error(_("Desired %(attr)s requested (%(desired)s) is "
                                " outside the currently assigned minimum "
                                " (%(min)s) and maximum (%(max)s) values for "
                                " virtual machine"
                                " '%(instance_name)s'.") % ex_args)
                    raise hmcexception.IBMPowerVMDesiredOutsideCurrentRange(
                        **ex_args)

        # desired_mem must be a multiple of LMB size
        validate_with_lmb(_('Desired memory'), desired_mem)

        # Validate min/max memory if user provided, otherwise set default
        min_mem = self._attrs.get('min_mem')
        max_mem = self._attrs.get('max_mem')

        # If one is set the other must be set
        self._validate_both_min_max_set(min_mem, max_mem, _('memory'))

        # If min_mem is specified, max_mem will be too.
        if min_mem is not None:
            # Validate minimum memory
            min_mem = self._validate_min_attr(min_mem, desired_mem,
                                              MIN_MEM_STR)
            # min_mem must be a multiple of LMB size
            validate_with_lmb(_('Minimum memory'), min_mem)

            # Validate maximum memory
            max_mem = self._validate_max_attr(max_mem, desired_mem,
                                              MAX_MEM_STR)
            # max_mem must be a multiple of LMB size
            validate_with_lmb(_('Maximum memory'), max_mem)

            # If resizing a running instance, min/max memory cannot be changed
            if (self._resize_and_running and
                    (max_mem != curr_max_mem or min_mem != curr_min_mem)):
                ex_args = {'attr': _('minimum or maximum memory'),
                           'instance_name': self._inst_name}
                LOG.error(_("Changing %(attr)s is not supported while "
                            "virtual machine '%(instance_name)s' is running. "
                            "It must be powered off first.") % ex_args)
                raise hmcexception.IBMPowerVMUnsupportedInDLPAR(**ex_args)

        # min_mem and max_mem were not specified. Set defaults if this is
        # a spawn or if resizing a non-running instance and new desired memory
        # is outside the current range
        else:
            # Creating a new instance.
            if curr_flavor is None:
                # Set default values for min and max memory to desired memory
                min_mem = desired_mem
                LOG.info(_("Defaulting minimum memory to desired "
                           "memory '%(mem)d MB' for virtual machine %(inst)s.")
                         % {'mem': desired_mem, 'inst': self._inst_name})
                max_mem = desired_mem
                LOG.info(_("Defaulting maximum memory to desired "
                           "memory '%(mem)d MB' for virtual machine %(inst)s.")
                         % {'mem': desired_mem, 'inst': self._inst_name})
            # Resizing
            else:
                # Resizing a non-running instance
                if self._resize_and_running is False:
                    # If new desired memory is below current min, reset
                    # min mem
                    if curr_min_mem > desired_mem:
                        min_mem = desired_mem
                        LOG.info(_("Defaulting minimum memory to desired "
                                   "memory '%(mem)d MB' "
                                   "for virtual machine %(inst)s.") %
                                 {'mem': desired_mem, 'inst': self._inst_name})
                    else:
                        min_mem = curr_min_mem
                    # If new desired memory is above current max, reset
                    # max mem
                    if curr_max_mem < desired_mem:
                        max_mem = desired_mem
                        LOG.info(_("Defaulting maximum memory to desired "
                                   "memory '%(mem)d MB' "
                                   "for virtual machine %(inst)s.") %
                                 {'mem': desired_mem, 'inst': self._inst_name})
                    else:
                        max_mem = curr_max_mem
                # Resizing a running instance. Set min/max to current min/max
                else:
                    min_mem = curr_min_mem
                    max_mem = curr_max_mem

        self._attrs['min_mem'] = min_mem
        self.instance_type['memory_mb'] = desired_mem
        self._attrs['max_mem'] = max_mem

    def _validate_cpu(self, host_stats, curr_flavor=None):
        """Validate the the user specified values for cpu options
        and values. Set defaults for values that are  not specified.

        :param host_stats: host attributes
        :param curr_flavor: If this is a resize, current flavor of
            instance, otherwise None.
        """

        # Check for valid processor compatibility mode if specified, and set
        # default if needed
        self._validate_proc_compat_mode(host_stats, curr_flavor)

        # Check for valid availability priority if specified, and set default
        # if needed
        self._validate_avail_priority(curr_flavor)

        # Check and set defaults for processor mode. (Dedicated or Shared).
        # Default to shared processor mode.
        self._validate_proc_mode(curr_flavor)

        # If processor mode is dedicated
        if self._attrs['dedicated_proc']:
            self._validate_ded_proc_options(curr_flavor)

        # This is shared processor mode
        else:
            self._validate_shared_proc_options(curr_flavor)

        # Validate min, desired, and max for vcpus
        # and processing units and set defaults
        self._validate_cpu_values(curr_flavor)

    def _validate_ded_proc_options(self, curr_flavor=None):
        """Validate dedicated processor options and set defaults.
        Ensure no shared processor related options are set.
        :param curr_flavor: If this is a resize, current flavor of
            instance, otherwise None.
        """
        # Check for invalid options with dedicated processors
        inval_ded_options = {'uncapped': _('Sharing mode of uncapped'),
                             'shared_weight': _('Shared processor weight'),
                             'proc_units': _('Processing units'),
                             'min_proc_units': _('Minimum Processing units'),
                             'max_proc_units': _('Maximum Processing units'),
                             'shared_proc_pool_name':
                             _('Shared processor pool name')}

        for option, msg in inval_ded_options.items():
            if self._attrs[option] is not None:
                ex_args = {'instance_name': self._inst_name,
                           'option': msg}
                LOG.error(_("%(option)s is not a valid option"
                            " for dedicated processor mode for virtual machine"
                            " '%(instance_name)s'.") % ex_args)
                raise hmcexception.IBMPowerVMInvalidAttrWithDed(**ex_args)

        # Validate dedicated sharing mode if set
        if self._attrs['dedicated_sharing_mode'] is not None:
            ded_share_mode = self._attrs['dedicated_sharing_mode'].lower()
            if ded_share_mode not in DED_SHARING_MODES:
                ex_args = {'instance_name': self._inst_name,
                           'ded_share_mode':
                           self._attrs['dedicated_sharing_mode'],
                           'valid_values': ', '.join(DED_SHARING_MODES)}
                LOG.error(_("'%(ded_share_mode)s' is not a valid option for"
                            " sharing mode for dedicated processors. Valid"
                            " options are '%(valid_values)s'.") % ex_args)
                raise hmcexception.IBMPowerVMInvalidDedShareMode(**ex_args)
            self._attrs['dedicated_sharing_mode'] = ded_share_mode
            return

        # Dedicated sharing mode is not set. Only assign default if this is a
        # spawn or if this is a resize of an instance that is changing from
        # shared proc mode to dedicated proc mode. The current value should be
        # kept for stopped instances already in dedicated processor mode.
        if (curr_flavor is None or (
                curr_flavor and curr_flavor.get('powervm:dedicated_proc') ==
                'false')):
            self._attrs['dedicated_sharing_mode'] = DEFAULT_SHARING_MODE
            LOG.info(_("Defaulting dedicated sharing mode to '%(mode)s' for "
                       "virtual machine %(inst)s.")
                     % {'mode': DEFAULT_SHARING_MODE, 'inst': self._inst_name})
        else:
            self._attrs['dedicated_sharing_mode'] = (
                curr_flavor.get('powervm:dedicated_sharing_mode'))

    def _validate_shared_proc_options(self, curr_flavor=None):
        """Validate shared processor options and set defaults.
        Ensure no dedicated processor related options are set.
        :param curr_flavor: If this is a resize, current flavor of
            instance, otherwise None.
        """
        # If resize, get current processor mode, current sharing mode,
        # and current shared weight
        if curr_flavor:
            curr_ded_proc = (
                curr_flavor.get('powervm:dedicated_proc') == 'true')
            curr_uncapped = (
                curr_flavor.get('powervm:uncapped') == 'true')
            curr_shared_weight = (
                curr_flavor.get('powervm:shared_weight'))

        # dedicated_sharing_mode is not a valid option
        if self._attrs['dedicated_sharing_mode'] is not None:
            ex_args = {'instance_name': self._inst_name,
                       'option': _('Dedicated sharing mode')}
            LOG.error(_("%(option)s is not a valid option"
                        " for shared processor mode for virtual machine"
                        " '%(instance_name)s'.") % ex_args)
            raise hmcexception.IBMPowerVMInvalidAttrWithShared(**ex_args)

        # add check for the shared_proc_pool_name
        # we allow change from dedicated processors to non-default shared
        # processor pool
        proc_pool_name = self._attrs.get('shared_proc_pool_name')
        # set the default to DefaultPool if it's not a resize operation.
        # Otherwise, it's the same as the current pool
        if proc_pool_name is None and (curr_flavor is None or curr_ded_proc):
            self._attrs['shared_proc_pool_name'] = DEFAULT_PROC_POOL
            proc_pool_name = DEFAULT_PROC_POOL
        if curr_flavor and not curr_ded_proc:
            # make sure the shared_proc_pool_name is the same
            curr_proc_pool_name = curr_flavor.\
                get('powervm:shared_proc_pool_name')
            # this is resize operation. If proc pool is not provided,
            # assume it's the same as the current pool
            if proc_pool_name is None:
                self._attrs['shared_proc_pool_name'] = curr_proc_pool_name
                proc_pool_name = curr_proc_pool_name
            if curr_proc_pool_name != proc_pool_name:
                ex_args = {'instance_name': self._inst_name,
                           'curr_proc_pool': curr_proc_pool_name,
                           'proc_pool': proc_pool_name}
                LOG.error(_("Changing shared processor pool name is not "
                            "supported for virtual machine %(instance_name)s. "
                            "The current processor pool is: "
                            "%(curr_proc_pool)s. The requested processor pool "
                            "is: %(proc_pool)s.") % ex_args)
                raise hmcexception.IBMPowerVMProcPoolNameModNotAllowed(
                    **ex_args)

        # Validate sharing mode (uncapped/capped)
        # Set default for uncapped if not set
        if self._attrs['uncapped'] is not None:
            uncapped = self._attrs['uncapped'].lower()
            if uncapped != 'true' and uncapped != 'false':
                ex_args = {'option': _('Uncapped sharing mode'),
                           'value': self._attrs['uncapped'],
                           'instance_name': self._inst_name}
                LOG.error(_(" %(option)s can either be true or false."
                            " '%(value)s' is not valid for virtual machine"
                            " '%(instance_name)s'.") % ex_args)
                raise hmcexception.IBMPowerVMInvalidBoolean(**ex_args)
            self._attrs['uncapped'] = uncapped == 'true'

        # Uncapped is not set. Only assign default if this is a spawn,
        # or if this is a resize of an instance that is changing from
        # dedicated proc mode to shared proc mode. The current value should be
        # kept for stopped instances already in shared processor mode.
        elif (curr_flavor is None or (curr_flavor and curr_ded_proc is True)):
            self._attrs['uncapped'] = True
            LOG.info(_("Defaulting sharing mode to uncapped "
                       "for virtual machine"
                       " %s.") % self._inst_name)
        else:
            self._attrs['uncapped'] = curr_uncapped

        # capped sharing mode
        if self._attrs['uncapped'] is False:
            # Shared weight cannot be set or must be 0
            # if sharing mode is capped
            if (self._attrs['shared_weight'] is not None and
                    self._attrs['shared_weight'] != '0'):
                ex_args = {'instance_name': self._inst_name}
                LOG.error(_("Sharing weight is not a valid option"
                            " for capped sharing mode for virtual machine"
                            " '%(instance_name)s'.") % ex_args)
                raise hmcexception.IBMPowerVMInvalidAttrWithCapped(
                    **ex_args)
            # The rest of this method deals with uncapped options. Return.
            return

        # uncapped sharing mode
        # if shared_weight is specified, check if valid
        if self._attrs['shared_weight'] is not None:
            invalid = False
            try:
                shared_weight = int(self._attrs['shared_weight'])
            except ValueError:
                invalid = True
            if (invalid is True or
                    shared_weight not in xrange(0, UNCAP_WEIGHT_UP_BOUND + 1)):
                ex_args = {'instance_name': self._inst_name,
                           'shared_weight':
                           self._attrs['shared_weight']}
                LOG.exception(_("'%(shared_weight)s' is not a valid"
                                " value for shared processor weight for"
                                " virtual machine '%(instance_name)s'. Specify"
                                " a value between 0 and 255.") % ex_args)
                raise hmcexception.IBMPowerVMInvalidSharedWeight(**ex_args)

        # Shared weight is not set. Only assign default if this is a spawn,
        # or if this is a resize of an instance that is changing from capped
        # to uncapped or dedicated to shared. The current value should be
        # kept for instances already in uncapped shared processor mode.
        elif (curr_flavor is None or
              (curr_flavor and curr_uncapped is False)):
            shared_weight = UNCAP_WEIGHT_DEFAULT
            LOG.info(_("Defaulting shared weight to '%(weight)d' "
                       "for virtual machine %(inst)s.") %
                     {'weight': UNCAP_WEIGHT_DEFAULT, 'inst': self._inst_name})
        else:
            shared_weight = curr_shared_weight

        self._attrs['shared_weight'] = shared_weight

    def _validate_cpu_values(self, curr_flavor=None):
        """Validate min/desired/max values for vcpus and
        processing units. Set defaults when not specified.
        :param curr_flavor: If this is a resize, current flavor of
            instance, otherwise None.
        """

        # Validate dependencies
        self._validate_cpu_dependencies()

        # Deisred vcpus is required. Check that it is set,
        # it is a valid number, and it is at least 1.
        try:
            desired_vcpus = int(self.instance_type['vcpus'])
        except KeyError:
            ex_args = {'attr': _('desired vcpus'),
                       'instance_name': self._inst_name}
            LOG.error(_("A value for %(attr)s is required for virtual machine"
                        " '%(instance_name)s'.") % ex_args)
            raise hmcexception.IBMPowerVMAttrRequired(**ex_args)
        except ValueError:
            ex_args = {'attr': _('Desired vcpus'),
                       'value': '%s' % self.instance_type['vcpus'],
                       'instance_name': self._inst_name}
            LOG.error(_("'%(value)s' is not a valid value for %(attr)s"
                        " for instance '%(instance_name)s'.") % ex_args)
            raise hmcexception.IBMPowerVMAttrValueInvalid(**ex_args)
        if desired_vcpus < VCPUS_LOW_BOUND:
            ex_args = {'cpus_requested': desired_vcpus,
                       'cpus_min': VCPUS_LOW_BOUND,
                       'instance_name': self._inst_name}
            LOG.error(_("CPUs requested (%(cpus_requested)d) is less than"
                        " minimum required value (%(cpus_min)d) for virtual "
                        "machine '%(instance_name)s'.") % ex_args)
            raise exception.IBMPowerVMCPUsBelowMin(**ex_args)

        # If this is a resize, get current min/max values.
        if curr_flavor:
            curr_min_vcpu = int(curr_flavor.get('powervm:min_vcpu'))
            curr_max_vcpu = int(curr_flavor.get('powervm:max_vcpu'))
        # If this is a resize and current state is running, desired vcpus
        # cannot be outside of the current range of minimum and maximum
        # vcpus.
        if (self._resize_and_running and (
            desired_vcpus < curr_min_vcpu or
                desired_vcpus > curr_max_vcpu)):
            ex_args = {'attr': _('processors'),
                       'desired': '%s' % desired_vcpus,
                       'min': '%s' % curr_min_vcpu,
                       'max': '%s' % curr_max_vcpu,
                       'instance_name': self._inst_name}
            LOG.error(_("Desired %(attr)s requested (%(desired)s) is "
                        " outside the currently assigned minimum "
                        " (%(min)s) and maximum (%(max)s) values for "
                        " virtual machine '%(instance_name)s'.") % ex_args)
            raise hmcexception.IBMPowerVMDesiredOutsideCurrentRange(**ex_args)

        # If min_vcpu is specified, max_vcpu will be too.
        if self._attrs['min_vcpu'] is not None:
            min_vcpu = self._validate_min_attr(
                self._attrs['min_vcpu'], desired_vcpus, MIN_VCPU_STR)

            max_vcpu = self._validate_max_attr(
                self._attrs['max_vcpu'], desired_vcpus, MAX_VCPU_STR)

            # If resizing a running instance, min/max vcpus cannot be changed
            if (self._resize_and_running and
                    (max_vcpu != curr_max_vcpu or
                     min_vcpu != curr_min_vcpu)):
                ex_args = {'attr': _('minimum or maximum processors'),
                           'instance_name': self._inst_name}
                LOG.error(_("Changing %(attr)s is not supported while "
                            "virtual machine '%(instance_name)s' is running. "
                            "It must be powered off first.") % ex_args)
                raise hmcexception.IBMPowerVMUnsupportedInDLPAR(**ex_args)

        # min_vcpu and max_vcpu were not specified. Set defaults if this is
        # a spawn or if resizing a non-running instance and new desired vcpus
        # is outside the current range
        else:
            # This is a spawn
            if curr_flavor is None:
                # Set default values for min and max vcpus to desired vcpus
                min_vcpu = desired_vcpus
                LOG.info(_("Setting default minimum processors to %(vcpus)d"
                           " for virtual machine %(inst)s")
                         % {'vcpus': desired_vcpus, 'inst': self._inst_name})
                max_vcpu = desired_vcpus
                LOG.info(_("Setting default maximum processors to %(vcpus)d"
                           " for virtual machine %(inst)s")
                         % {'vcpus': desired_vcpus, 'inst': self._inst_name})
            # This is a resize
            else:
                # non-running instance
                if self._resize_and_running is False:
                    # If new desired vcpus is below current min, reset
                    # min vcpus
                    if curr_min_vcpu > desired_vcpus:
                        min_vcpu = desired_vcpus
                        LOG.info(_("Setting default minimum processors to "
                                   "%(vcpus)d for virtual machine %(inst)s") %
                                 {'vcpus': desired_vcpus,
                                  'inst': self._inst_name})
                    else:
                        min_vcpu = curr_min_vcpu
                    # If new desired vcpus is above current max, reset
                    # max vcpus
                    if curr_max_vcpu < desired_vcpus:
                        max_vcpu = desired_vcpus
                        LOG.info(_("Setting default maximum processors to "
                                   "%(vcpus)d for virtual machine %(inst)s") %
                                 {'vcpus': desired_vcpus,
                                  'inst': self._inst_name})
                    else:
                        max_vcpu = curr_max_vcpu
                # Running instance. Set min/max to current min/max
                else:
                    min_vcpu = curr_min_vcpu
                    max_vcpu = curr_max_vcpu

        # Repopulate the vcpus attributes with the int values and defaults
        self._attrs['min_vcpu'] = min_vcpu
        self.instance_type['vcpus'] = desired_vcpus
        self._attrs['max_vcpu'] = max_vcpu

        # The rest of this method deals with processing units, return
        # if processor mode is dedicated
        if self._attrs['dedicated_proc']:
            return

        # If this  is a resize and not resizing from dedicated,
        # get current processor unit values
        if curr_flavor and (
                curr_flavor.get('powervm:dedicated_proc') != 'true'):
            curr_desired_proc_units = (
                round(float(curr_flavor.get('powervm:proc_units')), 2))
            curr_min_proc_units = (
                round(float(curr_flavor.get('powervm:min_proc_units')), 2))
            curr_max_proc_units = (
                round(float(curr_flavor.get('powervm:max_proc_units')), 2))
        # Validate proc_units if set
        if self._attrs['proc_units'] is not None:
            try:
                desired_proc_units = float(self._attrs['proc_units'])
            except ValueError:
                ex_args = {'attr': _('Desired processing units'),
                           'value': self._attrs['proc_units'],
                           'instance_name': self._inst_name}
                LOG.error(_("'%(value)s' is not a valid value for %(attr)s"
                            " for virtual machine '%(instance_name)s'.")
                          % ex_args)
                raise hmcexception.IBMPowerVMAttrValueInvalid(**ex_args)
            if desired_proc_units < round(PROC_UNITS_LOW_BOUND, 2):
                ex_args = {'units_requested': '%.2f' % desired_proc_units,
                           'units_min': '%.2f' % PROC_UNITS_LOW_BOUND,
                           'instance_name': self._inst_name}
                LOG.error(_("Processing units requested (%(units_requested)s"
                            " are below minimum required (%(units_min)s) for"
                            " virtual machine '%(instance_name)s'.") % ex_args)
                raise exception.IBMPowerVMProcUnitsBelowMin(**ex_args)

            # If this is a resize and current state is running. Desired proc
            # units cannot be outside of the current range of minimum and
            # maximum proc units.
            if (self._resize_and_running and (
                desired_proc_units < curr_min_proc_units or
                    desired_proc_units > curr_max_proc_units)):
                ex_args = {'attr': _('processor units'),
                           'desired': '%s' % desired_proc_units,
                           'min': '%s' % curr_min_proc_units,
                           'max': '%s' % curr_max_proc_units,
                           'instance_name': self._inst_name}
                LOG.error(_("Desired %(attr)s requested (%(desired)s) is "
                            " outside the currently assigned minimum "
                            " (%(min)s) and maximum (%(max)s) values for "
                            " virtual machine '%(instance_name)s'.") % ex_args)
                raise hmcexception.IBMPowerVMDesiredOutsideCurrentRange(
                    **ex_args)

        # Desired processor units was not set.
        else:
            # Set a default if this is a spawn, or this is a resize and the
            # instance is changing from dedicated proc mode to shared proc mode
            if(curr_flavor is None or (
                curr_flavor and curr_flavor.get('powervm:dedicated_proc') ==
                    'true')):
                desired_proc_units = (
                    proc_utils.get_num_proc_units(desired_vcpus))
                LOG.info(_("Setting default processing units to %(proc)s "
                           "for virtual machine %(inst)s") %
                         {'proc': '%.2f' % desired_proc_units,
                          'inst': self._inst_name})
            # Resizing
            else:
                # Set default proc units to current proc units
                desired_proc_units = curr_desired_proc_units
                # is current desired proc units still valid with new
                # desired VCPUs?
                is_valid_min_per_vcpu = (
                    self._validate_min_proc_unit_per_vcpu(
                        desired_proc_units, desired_vcpus, '', False))
                is_valid_vcpu_ge_proc_unit = (
                    self._validate_vcpus_ge_proc_units(
                        desired_proc_units, desired_vcpus, '', False))
                # new desired VCPUs * MIN_PROC_UNIT_PER_VCPU is greater
                # than current desired proc units, so change desired proc units
                # to as close as they can be to the current desired proc unit
                # value without being invalid.
                if is_valid_min_per_vcpu is False:
                    desired_proc_units = desired_vcpus * MIN_PROC_UNIT_PER_VCPU
                # new desired VCPUs is less than current desired proc units,
                # so change desired proc units to as close as they can be to
                # the current desired proc unit value without being invalid.
                if is_valid_vcpu_ge_proc_unit is False:
                    desired_proc_units = desired_vcpus

        self._attrs['proc_units'] = desired_proc_units

        # If min_proc_units is specified, max_proc_units will be too.
        if self._attrs['min_proc_units'] is not None:
            min_proc_units = self._validate_min_attr(
                self._attrs['min_proc_units'], desired_proc_units,
                MIN_PROC_UNIT_STR)

            max_proc_units = self._validate_max_attr(
                self._attrs['max_proc_units'], desired_proc_units,
                MAX_PROC_UNIT_STR)

            # If resizing a running instance, min/max proc_units cannot
            # be changed
            if (self._resize_and_running and (
                round(max_proc_units, 2) != curr_max_proc_units or
                    round(min_proc_units, 2) != curr_min_proc_units)):
                ex_args = {'attr': _('minimum or maximum processor units'),
                           'instance_name': self._inst_name}
                LOG.error(_("Changing %(attr)s is not supported while "
                            "virtual machine '%(instance_name)s' is running. "
                            "It must be powered off first.") % ex_args)
                raise hmcexception.IBMPowerVMUnsupportedInDLPAR(**ex_args)

        # max_proc_units and max_proc_units were not specified. Calculate
        # default values.
        else:
            min_proc_units, max_proc_units = (
                self._calc_min_max_proc_units(curr_flavor))

        # Validate relationship between vcpus and proc_units
        # min
        self._validate_min_proc_unit_per_vcpu(min_proc_units, min_vcpu,
                                              MIN_STR)
        self._validate_vcpus_ge_proc_units(min_proc_units, min_vcpu, MIN_STR)
        # desired
        self._validate_min_proc_unit_per_vcpu(desired_proc_units,
                                              desired_vcpus, DESIRED_STR)
        self._validate_vcpus_ge_proc_units(desired_proc_units, desired_vcpus,
                                           DESIRED_STR)
        # max
        self._validate_min_proc_unit_per_vcpu(max_proc_units, max_vcpu,
                                              MAX_STR)
        self._validate_vcpus_ge_proc_units(max_proc_units, max_vcpu, MAX_STR)

        # Repopulate the min/max proc unit attributes with the float values
        # and defaults
        self._attrs['min_proc_units'] = min_proc_units
        self._attrs['max_proc_units'] = max_proc_units

    def _calculate_resize_deltas(self, curr_flavor):
        """Calculate the resource deltas before
        a resize by comparing the current instance type
        to the attributes of this instance type.
        :param curr_flavor: Current flavor dictionary

        :return dict of memory, processor, and
            processing units deltas
        """
        deltas = {}

        # Calculate memory deltas
        deltas['delta_mem'] = (
            self.instance_type.get('memory_mb') -
            int(curr_flavor.get('memory_mb')))
        deltas['delta_max_mem'] = (
            int(self._attrs.get('max_mem')) -
            int(curr_flavor.get('powervm:max_mem')))

        # Calculate VCPU deltas
        deltas['delta_vcpu'] = (
            self.instance_type.get('vcpus') -
            int(curr_flavor.get('vcpus')))

        curr_proc_units = curr_flavor.get('powervm:proc_units')
        proc_mode = self._attrs.get('dedicated_proc')
        # If this is dedicated processor mode, there are no proc_units.
        if proc_mode:
            # if current flavor has proc_units set, this is a resize
            # from shared to dedicated processor mode, so set delta
            # to requested minus current proc units
            if curr_proc_units:
                deltas['delta_vcpu'] = round((self.instance_type.get('vcpus') -
                                              float(curr_proc_units)), 2)
            return deltas

        # Calculate proc unit deltas. If current proc unit values are none,
        # this is a resize from dedicated to shared processor mode, so
        # set delta to requested minus current vcpus.
        if curr_proc_units:
            deltas['delta_proc_units'] = (
                round(float(self._attrs.get('proc_units')) -
                      float(curr_proc_units), 2))
        else:
            deltas['delta_proc_units'] = (
                round(float(self._attrs.get('proc_units')) -
                      float(curr_flavor.get('vcpus')), 2))

        return deltas

    def _calculate_memory_overhead(self, host, host_stats, num_eth_adapters,
                                   desired_mem, max_mem, force=False):
        """Calls calculate_memory_overhead_on_host on the host object to
        invoke K2 job to get PHYP memory overhead for the deploy/resize.
        :parm host: Host object
        :parm num_eth_adapters: number of virtual ethernet adapters
        :parm desired_mem: Desired memory
        :parm max_mem: Maximum memory requested
        """
        # lower bound constant is at least twice of max mem requested
        mem_lower_bound = 2 * max_mem

        # Retrieve currently known Host stats
        host_mem_free = host_stats['host_memory_free']

        # Call K2 memory overhead check only when
        # host memory free is less than 2 times of max mem requested
        if host_mem_free < mem_lower_bound or force is True:
            info = {'desired_mem': desired_mem,
                    'max_mem': max_mem,
                    'num_virt_eth_adapters': num_eth_adapters}
            overhead, free_mem = (
                host.calculate_memory_overhead_on_host
                (reserved_mem_data=info))
        else:
            overhead = 0
            free_mem = host_mem_free

        return overhead, free_mem

    def _check_host_resources(self, host, host_stats, num_eth_adapters,
                              resource_deltas=None, curr_flavor=None):
        """Checks memory and processor resources on host for resize
        create. For creations operation, the self._attrs dict is
        compared against host resources. For resize operation, the
        resource_deltas dict is compared against host resources.

        :param host_stats: dict of host stats
        :param resource_delats: dict of resource deltas. Populated
            only during a resize operation.
        NOTE: _attrs['shared_proc_pool_name'] is set to DefaultPool
              in the _validate_shared_proc_options.
              If this method is ever called without calling
              _validate_shared_proc_options first, there would be a problem.
        """
        is_resize = resource_deltas is not None
        has_dedicated_proc = self._attrs['dedicated_proc']

        if is_resize:
            delta_desired_mem = resource_deltas['delta_mem']
            delta_desired_vcpus = resource_deltas['delta_vcpu']
        # Need to check full values for desired_mem, desired_vcpus, and
        # max_vcpu, even if resize.
        desired_memory = self.instance_type['memory_mb']
        desired_vcpus = self.instance_type['vcpus']
        max_vcpu = self._attrs['max_vcpu']

        # Memory
        free_memory = int(host_stats['host_memory_free'])
        mem_overhead = 0

        # If this is a spawn, or this is a resize where max memory is
        # increasing, get memory overhead.
        if is_resize is False or (
                is_resize and resource_deltas['delta_max_mem'] > 0):
            mem_overhead, avail_memory = self._calculate_memory_overhead(
                host, host_stats, num_eth_adapters,
                desired_memory, self._attrs['max_mem'])

            # This is a resize, so we need old overhead to compare it to.
            if is_resize:
                force = True if mem_overhead > 0 else False
                old_overhead, avail_memory = self._calculate_memory_overhead(
                    host, host_stats, num_eth_adapters,
                    int(curr_flavor['memory_mb']),
                    int(curr_flavor['powervm:max_mem']), force=force)
                mem_overhead = mem_overhead - old_overhead
            # Use the most current value for free memory on host
            if avail_memory is not None:
                free_memory = avail_memory

        # Check if there is enough free memory on the host for desired memory
        mem_to_check = (delta_desired_mem if is_resize else desired_memory)
        if (mem_to_check + mem_overhead) > free_memory:
            ex_args = {'mem_requested': mem_to_check,
                       'mem_avail': free_memory - mem_overhead,
                       'instance_name': self._inst_name}
            LOG.error(_("Insufficient free memory on host for virtual machine"
                        " '%(instance_name)s' (%(mem_requested)d MB"
                        " requested, %(mem_avail)d MB free)") % ex_args)
            raise exception.IBMPowerVMInsufficientFreeMemory(**ex_args)

        # Check CPU
        # Get maximum allowed processor info from host stats
        if has_dedicated_proc:
            max_procs_per_aix_linux_lpar = (
                int(host_stats['max_procs_per_aix_linux_lpar']))
            max_sys_proc_limit = int(host_stats['max_sys_procs_limit'])
        else:
            max_procs_per_aix_linux_lpar = (
                int(host_stats['max_vcpus_per_aix_linux_lpar']))
            max_sys_proc_limit = int(host_stats['max_sys_vcpus_limit'])
        # Desired processors cannot be above maximum allowed processors per
        # AIX/Linux LPAR.
        if desired_vcpus > max_procs_per_aix_linux_lpar:
            ex_args = {'vcpus': desired_vcpus,
                       'max_allowed': max_procs_per_aix_linux_lpar,
                       'instance_name': self._inst_name}
            LOG.error(_("The desired processors (%(vcpus)d) cannot be above"
                        " the maximum allowed processors per partition"
                        " (%(max_allowed)d) for virtual machine"
                        " '%(instance_name)s'.") % ex_args)
            raise hmcexception.IBMPowerVMVCPUsAboveMaxAllowed(**ex_args)
        # Maximum processors cannot be above maximum system capacity
        # processor limit.
        if max_vcpu > max_sys_proc_limit:
            ex_args = {'vcpus': max_vcpu,
                       'max_allowed': max_sys_proc_limit,
                       'instance_name': self._inst_name}
            LOG.error(_("The maximum processors (%(vcpus)d) cannot be above"
                        " the maximum system capacity processor limit"
                        " (%(max_allowed)d) for virtual machine"
                        " '%(instance_name)s'.") % ex_args)
            raise hmcexception.IBMPowerVMMaxVCPUsAboveMaxAllowed(**ex_args)

        total_free_procs = (float(host_stats['proc_units']) -
                            float(host_stats['proc_units_used']) -
                            float(host_stats['proc_units_reserved']))

        # Dedicated Processing Mode
        if has_dedicated_proc:
            procs_to_check = (
                delta_desired_vcpus if is_resize else desired_vcpus)
            # Check if there are enough free processors on the host
            # for desired processors
            if round(procs_to_check, 2) > round(total_free_procs, 2):
                ex_args = {'cpus_requested': '%.2f' % procs_to_check,
                           'cpus_avail': '%.2f' % total_free_procs,
                           'instance_name': self._inst_name}
                LOG.error(_("Insufficient available CPUs on host for virtual "
                            "machine '%(instance_name)s' (%(cpus_requested)s "
                            "requested, %(cpus_avail)s available)") % ex_args)
                raise hmcexception.IBMPowerVMInsufficientCPU(**ex_args)

        # In shared mode
        else:
            # extract the shared processor pool name
            proc_pool_name = self._attrs.get('shared_proc_pool_name')
            # NOTE: _attrs['shared_proc_pool_name'] is set to DefaultPool
            # in the _validate_shared_proc_options.
            # if this method is ever call without calling
            # _validate_shared_proc_options first. There would be a problem.
            if proc_pool_name != DEFAULT_PROC_POOL:
                # make sure the sysem support shared processor pool
                if not host_stats['stats'].get(SHR_PROC_POOL_CAPABLE,
                                               False):
                    raise hmcexception.IBMPowerVMShrProcPoolNotSupported(
                        host=host.host_display_name)
                host_proc_pools = host.get_shared_proc_pools()
                # make sure the shared processor pool exists on the host
                if not host_proc_pools.get(proc_pool_name):
                    ex_args = {'pool_name': proc_pool_name,
                               'host': host.host_display_name}
                    LOG.error(_("Unable to locate the shared processor pool "
                                "%(pool_name)s on managed host %(host)s."))
                    raise hmcexception.IBMPowerVMProcPoolNameNotFound(
                        **ex_args)
                pool_free_procs = host_proc_pools.get(proc_pool_name).get(
                    'available_proc_units', float(0))
                # take whichever value is the lowest
                if pool_free_procs < total_free_procs:
                    total_free_procs = pool_free_procs
                LOG.info(_('The current available processing units '
                           'on pool %(pool_name)s is %(proc_units)s.') %
                         {'pool_name': proc_pool_name,
                          'proc_units': total_free_procs})
                if curr_flavor:
                    # extract the current processor pool used by the instance
                    curr_proc_pool_id = curr_flavor.get(
                        'powervm:shared_proc_pool_id')
                    # pool id might be different
                    new_pool_id = host_proc_pools.get(proc_pool_name).get('id')
                    if curr_proc_pool_id != new_pool_id:
                        LOG.warn(_('Shared processor pool name is the same '
                                   'however the pool id is different. '
                                   'Pool name: %(name)s, '
                                   'previous pool id: %(old_id)s, and '
                                   'new pool id: %(new_id)s.') %
                                 {'name': proc_pool_name,
                                  'old_id': curr_proc_pool_id,
                                  'new_id': new_pool_id})
                        if self._resize_and_running:
                            # we cannot change the pool id at runtime
                            ex_args = {'proc_pool': proc_pool_name,
                                       'curr_proc_pool_id': curr_proc_pool_id,
                                       'proc_pool_id': new_pool_id}
                            raise hmcexception.IBMPowerVMProcPoolIDChanged(
                                **ex_args)

            proc_units_to_check = (
                resource_deltas['delta_proc_units'] if is_resize else
                self._attrs['proc_units'])
            # Check if there are enough processing units on the host
            # for desired processing units
            if round(proc_units_to_check, 2) > round(total_free_procs, 2):
                ex_args = {'units_requested': '%.2f' % proc_units_to_check,
                           'units_avail': '%.2f' % total_free_procs,
                           'instance_name': self._inst_name}
                LOG.error(_("Insufficient available processing units on host"
                            " for virtual machine '%(instance_name)s'"
                            " (%(units_requested)s requested, %(units_avail)s"
                            " available)") % ex_args)
                raise exception.IBMPowerVMInsufficientProcUnits(**ex_args)

    def _validate_both_min_max_set(self, min_val, max_val, attr_str):
        """Helper method. For the min/max ranges for memory, vcpus, and
        proc_units, if min is set, max must be set and vice versa.
        :param min_val: minimum value to check
        :param max_val: maximum value to check
        :param attr_str: attribute getting checked
        """
        if ((min_val is not None and max_val is None) or
                (max_val is not None and min_val is None)):
            ex_args = {'attr': attr_str,
                       'instance_name': self._inst_name}
            LOG.error(_("If specifying a minimum or maximum value for %(attr)s"
                        ", both minimum and maximum %(attr)s must be specified"
                        " for virtual machine '%(instance_name)s'.") % ex_args)
            raise hmcexception.IBMPowerVMBothMinMax(**ex_args)

    def _validate_min_attr(self, min_value, desired_value, attr_str):
        """Helper method to validate that the user specified minimum
        memory/vcpus/proc_units is a valid number and less than
        or equal to desired memory/vcpus/proc_units
        :param min_value: minimum memory/vcpus/proc_units to validate
        :param desired_value: desired memory/vcpus/proc_units to check against
        :param attr_str: attribute being validated

        :return min value converted to int or float
        """
        if attr_str == MIN_PROC_UNIT_STR:
            is_float = True
        else:
            is_float = False

        try:
            if is_float is True:
                min_value = float(min_value)
            else:
                min_value = int(min_value)
        # min value was not a valid number
        except ValueError:
            ex_args = {'attr': attr_str,
                       'instance_name': self._inst_name,
                       'value': min_value}
            if is_float is True:
                ex_args['des_value'] = '%.2f' % desired_value
            else:
                ex_args['des_value'] = '%s' % desired_value
            LOG.error(_("'%(value)s' is not a valid value for %(attr)s"
                        " for virtual machine '%(instance_name)s'.") % ex_args)
            raise hmcexception.IBMPowerVMAttrValueInvalid(**ex_args)
        # min value is greater than desired value
        if round(min_value, 2) > round(desired_value, 2):
            ex_args = {'attr': attr_str,
                       'instance_name': self._inst_name}
            if is_float is True:
                ex_args['value'] = '%.2f' % min_value
                ex_args['des_value'] = '%.2f' % desired_value
            else:
                ex_args['value'] = '%s' % min_value
                ex_args['des_value'] = '%s' % desired_value
            LOG.error(_("Value for %(attr)s (%(value)s) cannot be greater than"
                        " desired %(attr)s (%(des_value)s) for virtual machine"
                        " '%(instance_name)s'.") % ex_args)
            raise hmcexception.IBMPowerVMDesiredLTMin(**ex_args)
        return min_value

    def _validate_max_attr(self, max_value, desired_value, attr_str):
        """Helper method to validate that the user specified maximum
        memory/vcpus/proc_units is a valid number and greater than
        or equal to desired memory/vcpus/proc_units
        :param max_value: minimum memory/vcpus/proc_units to validate
        :param desired_value: desired memory/vcpus/proc_units to check against
        :param attr_str: attribute being validated

        :return max value converted to int or float
        """
        if attr_str == MAX_PROC_UNIT_STR:
            is_float = True
        else:
            is_float = False

        try:
            if is_float is True:
                max_value = float(max_value)
            else:
                max_value = int(max_value)
        # max value was not a valid number
        except ValueError:
            ex_args = {'attr': attr_str,
                       'instance_name': self._inst_name,
                       'value': max_value}
            if is_float is True:
                ex_args['des_value'] = '%.2f' % desired_value
            else:
                ex_args['des_value'] = '%s' % desired_value
            LOG.error(_("'%(value)s' is not a valid value for %(attr)s"
                        " for virtual machine '%(instance_name)s'.") % ex_args)
            raise hmcexception.IBMPowerVMAttrValueInvalid(**ex_args)
        # max value is greater than desired value
        if round(max_value, 2) < round(desired_value, 2):
            ex_args = {'attr': attr_str,
                       'instance_name': self._inst_name}
            if is_float is True:
                ex_args['value'] = '%.2f' % max_value
                ex_args['des_value'] = '%.2f' % desired_value
            else:
                ex_args['value'] = '%s' % max_value
                ex_args['des_value'] = '%s' % desired_value
            LOG.error(_("Value for %(attr)s (%(value)s) cannot be less than"
                        " desired %(attr)s (%(des_value)s) for virtual machine"
                        " '%(instance_name)s'.") % ex_args)
            raise hmcexception.IBMPowerVMDesiredGTMax(**ex_args)
        return max_value

    def _validate_cpu_dependencies(self):
        """Helper function to make sure dependencies are met.
        """
        min_vcpu = self._attrs['min_vcpu']
        max_vcpu = self._attrs['max_vcpu']
        min_proc_units = self._attrs['min_proc_units']
        desired_proc_units = self._attrs['proc_units']
        max_proc_units = self._attrs['max_proc_units']

        # If either min or max is set, the other must be set
        self._validate_both_min_max_set(min_vcpu, max_vcpu,
                                        _('virtual CPUs'))

        # The rest of the dependency validation method deals with
        # processing units. Return if processor mode is dedicated
        if self._attrs['dedicated_proc']:
            return

        self._validate_both_min_max_set(min_proc_units, max_proc_units,
                                        _('processing units'))
        # If min/max vcpus are not set, min/max proc_units cannot be set.
        if ((min_vcpu is None) and (min_proc_units is not None)):
            ex_args = {'instance_name': self._inst_name}
            LOG.error(_("If specifying a minimum and maximum for"
                        " processing units, you must also specify a"
                        " minimum and maximum for virtual CPUs for"
                        " virtual machine '%(instance_name)s'.") % ex_args)
            raise hmcexception.IBMPowerVMMustSpecifyVCPURange(**ex_args)

        # If min/max proc units set, desired proc_units must be set
        if ((min_proc_units is not None) and (desired_proc_units is None)):
            ex_args = {'instance_name': self._inst_name}
            LOG.error(_("If specifying a minimum and maximum for"
                        " processing units, you must also specify desired"
                        " processing units for virtual machine"
                        " '%(instance_name)s'.") % ex_args)
            raise hmcexception.IBMPowerVMMustSpecifyProcUnits(**ex_args)

    def _validate_vcpus_ge_proc_units(self, proc_units, vcpus, level_str,
                                      raise_exception=True):
        """Helper method to validate that processing units are not greater
        than vcpus.
        :param proc_units: requested processing units
        :param vcpus: requested vcpus
        :param level_str: string representing minimum, desired or maximum
        """
        is_valid = True
        if round(proc_units, 2) > vcpus:
            is_valid = False
            if raise_exception is False:
                return is_valid
            ex_args = {'level': level_str,
                       'proc_units': '%.2f' % proc_units,
                       'vcpus': vcpus,
                       'instance_name': self._inst_name}
            LOG.error(_("Processing units cannot be a larger value than"
                        " virtual CPUs. The requested %(level)s processing"
                        " units were '%(proc_units)s' and the requested"
                        " %(level)s virtual CPUs were '%(vcpus)d' for virtual "
                        "machine '%(instance_name)s'.") % ex_args)
            raise hmcexception.IBMPowerVMProcUnitsGTVCPUs(**ex_args)
        return is_valid

    def _validate_min_proc_unit_per_vcpu(self, proc_units, vcpus, level_str,
                                         raise_exception=True):
        """Helper method to validate that processing units meet the
        minimum processing units per vcpu requirement.
        :param proc_units: requested processing units
        :param vcpus: requested vcpus
        :param level_str: string representing minimum, desired or maximum
        :return boolean: True if valid, False if invalid
        """
        is_valid = True
        if round(vcpus * MIN_PROC_UNIT_PER_VCPU, 2) > round(proc_units, 2):
            is_valid = False
            if raise_exception is False:
                return is_valid
            ex_args = {'level': level_str,
                       'proc_units': '%.2f' % proc_units,
                       'vcpus': vcpus,
                       'proc_units_per_vcpu': '%.2f' % MIN_PROC_UNIT_PER_VCPU,
                       'instance_name': self._inst_name}
            LOG.error(_("There must be a minimum of %(proc_units_per_vcpu)s"
                        " processing units per virtual CPU. The requested"
                        " %(level)s processing units were '%(proc_units)s' and"
                        " the requested %(level)s virtual CPUs were "
                        "'%(vcpus)d' for virtual machine '%(instance_name)s'.")
                      % ex_args)
            raise hmcexception.IBMPowerVMInvalidProcUnitsPerVCPU(**ex_args)
        return is_valid

    def _validate_srr_capability(self, host_stats, req_srr_cap,
                                 curr_srr_cap=None):
        """ Validate simplified remote restart capability for lpar. Throw
        error if host simplified remote restart capability is not enabled.
        """

        # Validate simplified remote restart capability
        if req_srr_cap != 'true' and req_srr_cap != 'false':
            ex_args = {'option': _('Simplified remote restart capability'),
                       'value': req_srr_cap,
                       'instance_name': self._inst_name}
            LOG.error(_(" %(option)s can either be true or false."
                        " '%(value)s' is not valid for virtual machine"
                        " '%(instance_name)s'.") % ex_args)
            raise hmcexception.IBMPowerVMInvalidBoolean(**ex_args)

        if curr_srr_cap is not None and curr_srr_cap != req_srr_cap:
            if self._resize_and_running:
                ex_args = {'attr': _('simplified remote restart capability'),
                           'instance_name': self._inst_name}
                LOG.error(_("Changing %(attr)s is not supported while "
                            "virtual machine '%(instance_name)s' is"
                            " running. It must be powered off first.")
                          % ex_args)
                raise hmcexception.IBMPowerVMUnsupportedInDLPAR(**ex_args)
        stats = host_stats['stats']
        host_srr_cap = stats['powervm_lpar_simplified_remote_restart_capable']
        if not host_srr_cap and req_srr_cap == 'true':
            ex_args = {'attr': _('simplified remote restart capability'),
                       'instance_name': self._inst_name}
            LOG.error(_("Operation on the virtual machine %(instance_name)s"
                        " failed. Managed system does not support %(attr)s ")
                      % ex_args)
            raise hmcexception.IBMPowerVMHostUnsupportedCapability(**ex_args)

    def _validate_ame_capability(self, host_stats, req_exp_factor,
                                 curr_exp_factor=None):
        """ Validate the Active Memory Expansion capability for the LPAR

        From the perspective of PowerVC, there won't be an AME capability
        key value boolean for LPARs, unlike K2. Instead we say if the
        expansion factor is 0 then the capability is false. If it is
        1.0 <= x <= 10.0 then the capability is true. Any other value is
        invalid. This method validates the expansion factor under this
        assumption.

        This method also validates the LPAR's AME capability against the
        host's supported capability.
        """
        if req_exp_factor is None:
            return
        # Convert expansion factor to float for validation
        exp_fact_float = round(float(req_exp_factor), 2)
        if exp_fact_float != 0 and exp_fact_float < 1 or exp_fact_float > 10:
            ex_args = {'option': _('active memory expansion factor'),
                       'value': req_exp_factor,
                       'instance_name': self._inst_name}
            LOG.error(_("%(option)s must be greater than or equal to 1.0 and "
                        "less than or equal to 10.0. '%(value)s' is not valid "
                        "for virtual machine '%(instance_name)s'.") % ex_args)
            raise hmcexception.IBMPowerVMInvalidExpFactor(**ex_args)
        if curr_exp_factor is not None and curr_exp_factor != req_exp_factor:
            if self._resize_and_running:
                ex_args = {'attr': _('active memory expansion factor'),
                           'instance_name': self._inst_name}
                LOG.error(_("Changing %(attr)s is not supported while "
                            "virtual mahcine '%(instance_name)s' is "
                            "running. It must be powered off first.")
                          % ex_args)
                raise hmcexception.IBMPowerVMUnsupportedInDLPAR(**ex_args)
        stats = host_stats['stats']
        host_ame_cap = stats['active_memory_expansion_capable']
        if not host_ame_cap and exp_fact_float != 0:
            ex_args = {'attr': _('active memory expansion capability'),
                       'instance_name': self._inst_name}
            LOG.error(_("Operation on the virtual machine %(instance_name)s "
                        "failed. Managed system does not support %(attr)s ")
                      % ex_args)
            raise hmcexception.IBMPowerVMHostUnsupportedCapability(**ex_args)

    def _validate_proc_compat_mode(self, host_stats, curr_flavor=None):
        """ Validate processor compatibility mode. If not set and
        this is a spawn, set default to 'default'. This property can not be
        changed on a running instance.
        :param curr_flavor: None if spawn. Current flavor of instance if resize
        """
        proc_compat = self._attrs['processor_compatibility']
        # If this is a resize, get current values:
        if curr_flavor:
            curr_proc_compat, pend_proc_compat = (
                curr_flavor.get('powervm:processor_compatibility'))
        if proc_compat is not None:
            proc_compat = proc_compat.lower()
            # This is a resize
            if curr_flavor:
                # Changing proc_compat mode is not supported while instance is
                # running.
                if self._resize_and_running:
                    # Requested value is same as current value, return
                    if (proc_compat == curr_proc_compat.lower() or (
                            proc_compat == pend_proc_compat.lower())):
                        self._attrs['processor_compatibility'] = (
                            pend_proc_compat)
                        return
                    # If requested was not the same as current, this is
                    # not supported while instance is running exception
                    ex_args = {'attr': _('processor compatibility mode'),
                               'instance_name': self._inst_name}
                    LOG.error(_("Changing %(attr)s is not supported while "
                                "virtual machine '%(instance_name)s' is "
                                "running. It must be powered off first.")
                              % ex_args)
                    raise hmcexception.IBMPowerVMUnsupportedInDLPAR(**ex_args)

            # Obtain supported processor compatibility modes on system,
            # and verify that the specified mode is supported
            proc_compat_modes = host_stats['compatibility_modes'].split(',')

            # for IBM i, if the host is Power6, the lpar running on it
            # can only have proc_compat_mode of 'default'
            if (curr_flavor and IBMI_OS_TYPE == self._inst_os_type and
                    proc_compat != 'default'):
                target_mode = proc_utils.extract_highest_compat_mode(
                    proc_compat_modes)
                if target_mode == 6:
                    ex_args = {'instance_name': self._inst_name,
                               'processor_compatibility':
                               self._attrs['processor_compatibility']}
                    LOG.error(_("Changing '%(processor_compatibility)s' "
                                "is not supported because for an IBM i "
                                "virtual machine '%(instance_name)s', its "
                                "processor compatibility mode can only be "
                                "set as 'default' on Power6 host.") %
                              ex_args)
                    raise hmcexception.IBMPowerVMInvalidProcCompatIBMi(
                        **ex_args)

            if proc_compat not in [x.lower() for x in proc_compat_modes]:
                ex_args = {'instance_name': self._inst_name,
                           'processor_compatibility':
                           self._attrs['processor_compatibility'],
                           'valid_values':
                           ', '.join(proc_compat_modes)}
                LOG.error(_("'%(processor_compatibility)s' is not a"
                            " valid option for processor compatibility for"
                            " virtual machine '%(instance_name)s'. Valid"
                            " options are '%(valid_values)s'.") % ex_args)
                raise hmcexception.IBMPowerVMInvalidProcCompat(**ex_args)
            # If this is a resize, and mode is changing from POWER7 to
            # a POWER6 derivative, send a warning about migration. POWER7
            # to POWER6 migrations are only allowed when the VM was
            # initally deployed with a processor compatibility mode of
            # a POWER6 derivative so the MAC address will have been
            # created on the HMC.
            if curr_flavor:
                if (curr_proc_compat in ['POWER7', 'POWER8'] and
                        'power6' in proc_compat):
                    proc_compat_str = (
                        proc_compat.title().replace('Power', 'POWER'))
                    LOG.warn(_('The processor compatibility mode for virtual'
                               ' machine %(instance_name)s is currently'
                               ' %(cur_proc_compat)s and the desired value'
                               ' will be set to %(proc_compat)s. If'
                               ' this virtual machine was deployed with a'
                               ' processor compatibility mode of POWER7 or'
                               ' POWER8, you cannot migrate the virtual'
                               ' machine to a POWER6 host.') %
                             {'instance_name': self._inst_name,
                              'cur_proc_compat': curr_proc_compat,
                              'proc_compat': proc_compat_str})
                    info = {'msg': _(
                        'The processor compatibility mode for'
                        ' virtual machine {instance_name} is currently'
                        ' {cur_proc_compat} and the desired value will be'
                        ' set to {proc_compat}. If you deployed this'
                        ' virtual machine with a processor'
                        ' compatibility mode of POWER7 or POWER8, you'
                        ' cannot migrate the virtual machine to a POWER6'
                        ' host. To migrate from a POWER7 or POWER8 to a'
                        ' POWER6 host, you must initially deploy to a'
                        ' POWER7 or POWER8 system with the processor'
                        ' compatibility mode set to a POWER6 derivative or'
                        ' initially deploy to a POWER6 host.'),
                        'instance_name': self._inst_name,
                        'instance_id': self._inst_id,
                        'cur_proc_compat': curr_proc_compat,
                        'proc_compat': proc_compat_str}
                    notifier = rpc.get_notifier(service='compute',
                                                host=self._host_id)
                    notifier.warn(get_admin_context(), 'compute.instance.log',
                                  info)

            # Format processor compatability mode string
            if proc_compat != 'default':
                proc_compat = proc_compat.title().replace('Power', 'POWER')

            self._attrs['processor_compatibility'] = proc_compat
            return

        # Processor compatibility is not set. Set default if this is a spawn.
        if curr_flavor is None:
            self._attrs['processor_compatibility'] = 'default'
            LOG.info(_("Defaulting processor compatibility to 'default'"
                       " for virtual machine %s.") % self._inst_name)
        # This is a resize, keep current value for pending processor
        # compatibility.
        else:
            self._attrs['processor_compatibility'] = pend_proc_compat

    def _validate_avail_priority(self, curr_flavor=None):
        """ Validate availability priority. If not set and this is a spawn,
        default to AVAIL_PRIOR_DEFAULT constant.
        :param curr_flavor: None if spawn. Current flavor of instance if resize
        """
        # Check for valid availability priority if specified
        avail_priority = self._attrs['availability_priority']
        if avail_priority is not None:
            invalid = False
            try:
                avail_priority = int(self._attrs['availability_priority'])
            except ValueError:
                invalid = True
            if (avail_priority not in xrange(0, AVAIL_PRIOR_UP_BOUND + 1) or
                    invalid is True):
                ex_args = {'instance_name': self._inst_name,
                           'availability_priority':
                           self._attrs['availability_priority']}
                LOG.error(_("'%(availability_priority)s' is not a"
                            " valid value for availability priority"
                            " for virtual machine '%(instance_name)s'. Specify"
                            " a value between 0 and 255.") % ex_args)
                raise hmcexception.IBMPowerVMInvalidAvailPriority(**ex_args)

            self._attrs['availability_priority'] = avail_priority
            return

        # Availibility priority is not set. Set default if this is a spawn.
        if curr_flavor is None:
            self._attrs['availability_priority'] = AVAIL_PRIOR_DEFAULT
            LOG.info(_("Defaulting availability priority to '%(default)d' "
                       "for virtual machine %(inst)s.") %
                     {'default': AVAIL_PRIOR_DEFAULT, 'inst': self._inst_name})
        else:
            self._attrs['availability_priority'] = (
                int(curr_flavor.get('powervm:availability_priority')))

    def _validate_proc_mode(self, curr_flavor=None):
        """ Validate processor mode, shared or dedicated, if not set and
        this is a spawn, set default to False for shared processors. This
        property can not be changed on a running instance.
        :param curr_flavor: None if spawn. Current flavor of instance if resize
        """
        ded_proc = self._attrs['dedicated_proc']
        # If this is a resize, get current flavor
        if curr_flavor:
            curr_ded_proc = curr_flavor.get('powervm:dedicated_proc')
        if ded_proc is not None:
            ded_proc = self._attrs['dedicated_proc'].lower()
            if curr_flavor:
                # Changing processing mode is not supported while instance is
                # running.
                if ((ded_proc != curr_ded_proc) and self._resize_and_running):
                    ex_args = {'attr': _('Processing mode'),
                               'instance_name': self._inst_name}
                    LOG.error(_("Changing %(attr)s is not supported while "
                                "virtual machine '%(instance_name)s' is "
                                "running. It must be powered off first.")
                              % ex_args)
                    raise hmcexception.IBMPowerVMUnsupportedInDLPAR(**ex_args)

            if ded_proc != 'true' and ded_proc != 'false':
                ex_args = {'option': _('Dedicated processor mode'),
                           'value': self._attrs['dedicated_proc'],
                           'instance_name': self._inst_name}
                LOG.error(_("%(option)s can either be true or false."
                            " '%(value)s' is not valid for virtual machine"
                            " '%(instance_name)s'.") % ex_args)
                raise hmcexception.IBMPowerVMInvalidBoolean(**ex_args)
            self._attrs['dedicated_proc'] = ded_proc == 'true'
            return
        # dedicated proc not set. Set defaults if this is a spawn
        if curr_flavor is None:
            self._attrs['dedicated_proc'] = False
            LOG.info(_("Default processing mode is shared. Defaulting"
                       " dedicated_proc to 'false' for virtual machine %s.") %
                     self._inst_name)
        else:
            self._attrs['dedicated_proc'] = curr_ded_proc == 'true'

    def _calc_min_max_proc_units(self, curr_flavor=None):
        """If this is a spawn, or if this is a resize of an instance changing
        from dedicated processor mode to shared processor mode, or if
        resizing a non-running instance and new desired proc_units is outside
        the current range calculate proc units. Otherwise, set proc units
        to current proc units values.
        :param curr_flavor: None if spawn. Current flavor of instance if resize
        :return min_proc_units and max_proc_units
        """
        desired_proc_units = round(self._attrs['proc_units'], 2)
        min_vcpu = self._attrs['min_vcpu']
        max_vcpu = self._attrs['max_vcpu']

        # Creating a new instance or changing an existing instance to shared
        # processor mode
        if (curr_flavor is None or (
            curr_flavor and curr_flavor.get('powervm:dedicated_proc') ==
                'true')):
            # Use either desired_proc_units or calculated proc_units for
            # min_vcpu. Use the lesser of the two, so min_proc_units is
            # never greater than desired_proc_units
            calculated_proc_units = (
                proc_utils.get_num_proc_units(min_vcpu))
            min_proc_units = min(calculated_proc_units, desired_proc_units)
            LOG.info(_("Setting default minimum processing units to %(proc)s"
                       " for virtual machine %(inst)s") %
                     {'proc': '%.2f' % min_proc_units,
                      'inst': self._inst_name})

            # Use either desired_proc_units or calculated proc_units for
            # max_vcpu. Use the greater of the two, so max_proc_units is
            # never less than desired_proc_units
            calculated_proc_units = (
                proc_utils.get_num_proc_units(max_vcpu))
            max_proc_units = max(calculated_proc_units, desired_proc_units)
            LOG.info(_("Setting default maximum processing units to %(proc)s "
                       "for virtual machine %(inst)s")
                     % {'proc': max_proc_units, 'inst': self._inst_name})
            return min_proc_units, max_proc_units

        # This is a resize that was already in shared processor mode
        # Only change min/max if new desired is outside current range
        curr_min_proc_units = (
            float(curr_flavor.get('powervm:min_proc_units')))
        curr_max_proc_units = (
            float(curr_flavor.get('powervm:max_proc_units')))
        # Instance is not running
        if self._resize_and_running is False:
            if desired_proc_units < round(curr_min_proc_units, 2):
                # Use either desired_proc_units or calculated
                # proc_units for min_vcpu. Use the lesser of the two,
                # so min_proc_units is  never greater than
                # desired_proc_units
                calculated_proc_units = (
                    proc_utils.get_num_proc_units(min_vcpu))
                min_proc_units = min(calculated_proc_units,
                                     desired_proc_units)
                LOG.info(_("Setting default minimum processing units to"
                           "%(proc)s for virtual machine %(inst)s") %
                         {'proc': '%.2f' % min_proc_units,
                          'inst': self._inst_name})
            else:
                min_proc_units = curr_min_proc_units
            if desired_proc_units > round(curr_max_proc_units, 2):
                # Use either desired_proc_units or calculated
                # proc_units for max_vcpu. Use the greater of the two,
                # so max_proc_units is never less than
                # desired_proc_units
                calculated_proc_units = (
                    proc_utils.get_num_proc_units(max_vcpu))
                max_proc_units = max(calculated_proc_units,
                                     desired_proc_units)
                LOG.info(_("Setting default maximum processing units to "
                           "%(proc)s for virtual machine %(inst)s") %
                         {'proc': max_proc_units, 'inst': self._inst_name})
            else:
                max_proc_units = curr_max_proc_units
        # Running instance. Set min/max to current min/max
        else:
            min_proc_units = curr_min_proc_units
            max_proc_units = curr_max_proc_units

        return min_proc_units, max_proc_units
