# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import time
import six

from eventlet import greenthread

from oslo_log import log as logging

from oslo_config import cfg

from powervc_oslo.k2.k2operator import K2Element
from powervc_oslo.k2.k2operator import K2Error

from powervc_nova.virt.ibmpowervm.common import exception
from powervc_nova.virt.ibmpowervm.hmc.k2 import constants as k2_const
from powervc_nova.virt.ibmpowervm.hmc.k2.entry_wrapper import EntryWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2.entry_wrapper import DELIM

from powervc_nova import _

LOG = logging.getLogger(__name__)


ibmpowervm_opts = [
    cfg.IntOpt('ibmpowervm_job_request_timeout',
               default=1800,
               help='Default timeout in seconds for K2 job requests.')
]

CONF = cfg.CONF
CONF.register_opts(ibmpowervm_opts)

ROOT = './'

# Job Constants
JOB_ID = ROOT + 'JobID'
JOB_STATUS = ROOT + 'Status'
JOB_MESSAGE = ROOT + 'ResponseException' + DELIM + 'Message'
JOB_STACKTRACE = ROOT + 'ResponseException' + DELIM + 'StackTrace'
JOB_RESULTS_NAME = \
    ROOT + 'Results' + DELIM + 'JobParameter' + DELIM + 'ParameterName'
JOB_RESULTS_VALUE = \
    ROOT + 'Results' + DELIM + 'JobParameter' + DELIM + 'ParameterValue'
JOB_GROUP_NAME = \
    ROOT + 'RequestedOperation' + DELIM + 'GroupName'
JOB_OPERATION_NAME = \
    ROOT + 'RequestedOperation' + DELIM + 'OperationName'


class JobEntryWrapper(EntryWrapper):
    """
    Wrapper object for k2 job response schema
    """

    def create_job_parameter(self, name, value):
        """Creates a JobParameter K2Element
           :param name: ParameterName text value
           :param value: ParametervValue text value
           :returns: JobParameter K2Element
        """
        job_parm = K2Element('JobParameter',
                             attrib={'schemaVersion': 'V1_0'},
                             ns=k2_const.WEB_NS)
        job_parm.append(K2Element('ParameterName',
                                  text=name, ns=k2_const.WEB_NS))
        job_parm.append(K2Element('ParameterValue',
                                  text=value, ns=k2_const.WEB_NS))
        return job_parm

    def add_job_parameters_to_existing(self, *add_parms):
        """Adds JoParameter K2Elements to existing JobParameters xml.
           Must be a job response entry.
           :param add_parms: list of JobParamters to add
        """
        job_parms = self._k2entry.element.find('JobParameters')
        for parm in add_parms:
            job_parms.append(parm)

    def get_job_id(self):
        """Gets the job ID string

        :returns: String containing the job ID
        """
        return self.get_parm_value(JOB_ID)

    def get_job_status(self):
        """Gets the job status string

        :returns: String containing the job status
        """
        return self.get_parm_value(JOB_STATUS)

    def get_job_response_exception_message(self, default=''):
        """Gets the job message string from the ResponseException

        :returns: String containing the job message or
                  default (defaults to empty string) if not found
        """
        job_message = self.get_parm_value(JOB_MESSAGE, default)
        if job_message:
            # See if there is a stack trace to log
            stack_trace = self.get_parm_value(JOB_STACKTRACE, default)
            if stack_trace:
                exc = exception.IBMPowerVMJobRequestFailed(
                    operation_name=self.op, error=stack_trace)
                LOG.error(six.text_type(exc))
        return job_message

    def get_job_results_message(self, default=''):
        """Gets the job result message string

        :returns: String containing the job result message or
                  default (defaults to empty string) if not found
        """
        message = default
        err_data = default
        parm_names = self.get_parm_values(JOB_RESULTS_NAME)
        parm_values = self.get_parm_values(JOB_RESULTS_VALUE)
        for i in range(len(parm_names)):
            if parm_names[i] == 'result':
                message = parm_values[i]
                break
            elif parm_names[i] == 'ErrorData':
                err_data = parm_values[i]
        message = message if message else err_data
        return message

    def get_job_results_as_dict(self, default={}):
        """Gets the job results as a dictionary

        :returns: Dictionary with result parm names and parm
                  values as key, value pairs.
        """
        results = default
        parm_names = self.get_parm_values(JOB_RESULTS_NAME)
        parm_values = self.get_parm_values(JOB_RESULTS_VALUE)
        for i in range(len(parm_names)):
            results[parm_names[i]] = parm_values[i]
        return results

    def get_job_message(self, default=''):
        """Gets the job message string. It checks job results
           message first, if results message is not found, it
           checks for a ResponseException message. If neither
           are found, it returns the default.

        :returns: String containing the job message or
                  default (defaults to empty string) if not found
        """
        message = self.get_job_results_message(default=None)
        if not message:
            message = self.get_job_response_exception_message(default=default)
        return message

    def run_job(self, k2oper, uuid, job_parms=None,
                timeout=CONF.ibmpowervm_job_request_timeout,
                sensitive=False, host_uuid=None):
        """Adds job parameters to the job element if specified and calls
           the k2 createjob method. It then monitors the job for
           completion and sends a IBMPowerJobRequestFailed exception
           if it did not complete successfully.

        :param k2oper: K2 operator
        :param uuid: uuid of the target
        :param job_parms: list of JobParamters to add
        :param timeout: maximum number of seconds for job to complete
        :param host_uuid: UUID of the host
        """
        job = self._k2entry.element
        entry_type = self.get_parm_value(JOB_GROUP_NAME)
        self.op = self.get_parm_value(JOB_OPERATION_NAME)
        if job_parms:
            self.add_job_parameters_to_existing(*job_parms)
        try:
            # With the remote restart feature, the ManagedSystem UUID needs to
            # be specified to K2 requests on an LPAR
            if host_uuid:
                k2resp = k2oper.createjob(job, 'ManagedSystem', host_uuid,
                                          entry_type, uuid,
                                          sensitive=sensitive)
            else:
                k2resp = k2oper.createjob(job, entry_type, uuid,
                                          sensitive=sensitive)
            self._k2entry = k2resp.entry
        except K2Error as exc:
            LOG.exception(exc)
            raise exception.IBMPowerVMJobRequestFailed(
                operation_name=self.op, error=exc)
        status, message, timed_out = self.monitor_job(k2oper, timeout=timeout,
                                                      sensitive=sensitive)
        if timed_out:
            try:
                self.cancel_job(k2oper)
            except exception.IBMPowerVMJobRequestFailed as e:
                LOG.warn(six.text_type(e))
            exc = exception.IBMPowerVMJobRequestTimedOut(
                operation_name=self.op, seconds=str(timeout))
            LOG.exception(exc)
            raise exc
        if status != k2_const.K2_JOB_STATUS_COMPLETED_OK:
            self.delete_job(k2oper)
            exc = exception.IBMPowerVMJobRequestFailed(
                operation_name=self.op, error=message)
            LOG.exception(exc)
            raise exc
        self.delete_job(k2oper)

    def monitor_job(self, k2oper, job_id=None,
                    timeout=CONF.ibmpowervm_job_request_timeout,
                    sensitive=False):
        """Waits on a k2 job until the is no longer running or not started or
           if a timeout value is passed, it times out in the given amount of
           time.

        :param k2oper: K2 operator
        :param job_id: k2 job id
        :param timeout: maximum number of seconds to keep checking job status
        :returns status: String containing the job status
        :returns message: String containing the job results message
        :returns timed_out: boolean true if timed out waiting for job
                            completion
        """
        if not job_id:
            job_id = self.get_job_id()
        status = self.get_job_status()
        start_time = time.time()
        timed_out = False
        while (status == k2_const.K2_JOB_STATUS_RUNNING or
               status == k2_const.K2_JOB_STATUS_NOT_ACTIVE):
            if timeout:
                curr_time = time.time()
                # wait up to timeout seconds
                if (curr_time - start_time) > timeout:
                    timed_out = True
                    break
            greenthread.sleep(1)
            k2resp = k2oper.readjob(job_id, sensitive=sensitive)
            self._k2entry = k2resp.entry
            status = self.get_job_status()

        message = ''
        if (not timed_out and status != k2_const.K2_JOB_STATUS_COMPLETED_OK and
            status != k2_const.K2_JOB_STATUS_CANCELED_WHILE_RUNNING and
                status != k2_const.K2_JOB_STATUS_CANCELED_BEFORE_START):
            message = self.get_job_message()
            if not message:
                # if we cannot extract the message, dump the response
                resp = k2resp.body if k2resp else ''
                LOG.error(six.text_type(resp))
        return status, message, timed_out

    def cancel_job(self, k2oper,
                   timeout=CONF.ibmpowervm_job_request_timeout,
                   sensitive=False):
        """Cancels and deletes incomplete/running jobs.

        :param k2oper: K2 operator
        :param job_id: k2 job id
        :param status: k2 job status
        :param timeout: maximum number of seconds to keep checking job status
        """

        job_id = self.get_job_id()
        try:
            k2oper.update(None,
                          None,
                          rootType=k2_const.JOBS,
                          rootId=job_id,
                          suffixType='cancel')
        except K2Error as exc:
            LOG.exception(exc)
        status, message, timed_out = self.monitor_job(k2oper, timeout=timeout,
                                                      sensitive=sensitive)
        if timed_out:
            error = (_("Job %(job_id)s failed to cancel after %(timeout)s "
                       "seconds.") % dict(job_id=job_id,
                                          timeout=six.text_type(timeout)))
            exc = exception.IBMPowerVMJobRequestFailed(error)
            raise exc
        self.delete_job(k2oper, job_id, status)

    def delete_job(self, k2oper, job_id=None, status=None):
        """Cleans up completed jobs. IBMPowerJobRequestFailed
           exception sent if job is in running status

        :param k2oper: K2 operator
        :param job_id: k2 job id
        :param status: k2 job status
        """
        if not job_id:
            job_id = self.get_job_id()
        if not status:
            status = self.get_job_status()
        if status == k2_const.K2_JOB_STATUS_RUNNING:
            error = (_("Job %s not deleted. Job is in running status."
                       % job_id))
            exc = exception.IBMPowerVMJobRequestFailed(error)
            LOG.exception(exc)
            raise exc
        try:
            k2oper.delete(k2_const.JOBS, job_id)
        except K2Error as exc:
            LOG.exception(exc)
