# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import copy
import six

from eventlet import greenthread

from oslo_log import log as logging
from oslo_concurrency.lockutils import synchronized

from oslo_config import cfg

from powervc_oslo.k2.k2operator import K2Element
from powervc_oslo.k2.k2operator import K2Error

from powervc_nova import _
from powervc_nova import logcall

from powervc_nova.virt.ibmpowervm.hmc.k2 import contains_msg_in_k2_err
from powervc_nova.virt.ibmpowervm.hmc.k2 import contains_msgs_in_k2_err
from powervc_nova.virt.ibmpowervm.hmc.k2 import constants as k2const
from powervc_nova.virt.ibmpowervm.hmc.k2.profentry_wrapper \
    import LPARProfileEntryWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2.viosentry_wrapper \
    import VIOSEntryWrapper

from powervc_nova.virt.ibmpowervm.hmc.k2owrappers import VIOS_BUSY_ERR_MSG
from powervc_nova.virt.ibmpowervm.hmc import exception as hmc_ex

CONF = cfg.CONF

LOG = logging.getLogger(__name__)

DELIM = '/'
ROOT = './'
ETAG = 'etag'
PARTITION_ID = 'PartitionID'

# indicating whether we are dealing VirtualSCSIMappings or
# VirtualFibreChannelMappings
SCSI = 'scsi'
VFC = 'vfc'

CUR_PROF_SYNC_PATH = ROOT + 'CurrentProfileSync'
# Storage mapping names and path
SCSI_MAPPINGS = 'VirtualSCSIMappings'
SCSI_MAPPING_ELEM = 'VirtualSCSIMapping'
SCSI_MAPPING_PATH = ROOT + SCSI_MAPPINGS + DELIM + SCSI_MAPPING_ELEM
FC_MAPPINGS = 'VirtualFibreChannelMappings'
FC_MAPPING_ELEM = 'VirtualFibreChannelMapping'
FC_MAPPING_PATH = ROOT + FC_MAPPINGS + DELIM + FC_MAPPING_ELEM

# Adapter elements, values, and path
SERVER_ADAPTER = 'ServerAdapter'
CLIENT_ADAPTER = 'ClientAdapter'
ADAPTER_TYPE = 'AdapterType'
SERVER = 'Server'
CLIENT = 'Client'
USE_NEXT_AVAIL_SLOT = 'UseNextAvailableSlotID'
LOCAL_LPAR_ID = 'LocalPartitionID'
REMOTE_LPAR_ID = 'RemoteLogicalPartitionID'
CONN_LPAR_ID = 'ConnectingPartitionID'
VIR_SLOT_NUM = 'VirtualSlotNumber'
REMOTE_SLOT_NUM = 'RemoteSlotNumber'
CONN_SLOT_NUM = 'ConnectingVirtualSlotNumber'
ASSOC_LPAR = 'AssociatedLogicalPartition'
BACKING_DEV = 'BackingDeviceName'
MAP_PORT = 'MapPort'
WWPNS = 'WWPNs'
# like vhost0
ADAPTER_NAME = 'AdapterName'
SERVER_ADAPTER_PATH = ROOT + SERVER_ADAPTER
CLIENT_ADAPTER_PATH = ROOT + CLIENT_ADAPTER
CLIENT_VIR_SLOT_NUM_REL_PATH = CLIENT_ADAPTER_PATH + DELIM + VIR_SLOT_NUM
CLIENT_CONN_LPAR_REL_PATH = CLIENT_ADAPTER_PATH + DELIM + CONN_LPAR_ID
CLIENT_REMOTE_LPAR_REL_PATH = CLIENT_ADAPTER_PATH + DELIM + REMOTE_LPAR_ID
SERVER_VIR_SLOT_NUM_REL_PATH = SERVER_ADAPTER_PATH + DELIM + VIR_SLOT_NUM
SERVER_REMOTE_LPAR_REL_PATH = SERVER_ADAPTER_PATH + DELIM + REMOTE_LPAR_ID
SERVER_REMOTE_SLOT_REL_PATH = SERVER_ADAPTER_PATH + DELIM + REMOTE_SLOT_NUM
SERVER_CONN_LPAR_REL_PATH = SERVER_ADAPTER_PATH + DELIM + CONN_LPAR_ID
SERVER_CONN_SLOT_REL_PATH = SERVER_ADAPTER_PATH + DELIM + CONN_SLOT_NUM
SERVER_BACKING_DEV_REL_PATH = SERVER_ADAPTER_PATH + DELIM + BACKING_DEV
SERVER_MAP_PORT_REL_PATH = SERVER_ADAPTER_PATH + DELIM + MAP_PORT
CLIENT_LPAR_REL_PATH = CLIENT_ADAPTER_PATH + DELIM + LOCAL_LPAR_ID
CLIENT_WWPNS_REL_PATH = CLIENT_ADAPTER_PATH + DELIM + WWPNS

# Storage element
STORAGE = 'Storage'
# Storage schema names
PV = 'PhysicalVolume'
VDISK = 'VirtualDisk'
LU = 'LogicalUnit'
VOPT = 'VirtualOpticalMedia'
NPIV = 'npiv'
# vdisk attribute
DISK_NAME = 'DiskName'
# physical volume attribute
RESERVE_POLICY = 'ReservePolicy'
VOL_NAME = 'VolumeName'
VOL_SIZE = 'VolumeCapacity'
VOL_UID = 'VolumeUniqueID'
# LU attribute
UDID = 'UniqueDeviceID'
UNIT_NAME = 'UnitName'
UNIT_SIZE = 'UnitCapacity'
# virtual optical attribute
MEDIA_NAME = 'MediaName'

# Target Device
TARGET_DEV = 'TargetDevice'
VIRT_TARGET_DEV = 'VirtualTargetDevice'
LU_ADDR = 'LogicalUnitAddress'
LU_ADDR_REL_PATH = ROOT + VIRT_TARGET_DEV + DELIM + LU_ADDR
# like vtscsi1
TARGET_NAME = 'TargetName'
TARGET_NAME_REL_PATH = ROOT + VIRT_TARGET_DEV + DELIM + TARGET_NAME

# Storage type element
STORAGE_TYPE = 'StorageType'
# Storage type acceepted by K2
VDISK_TYPE = 'VIRTUAL_DISK'
PV_TYPE = 'PHYSICAL_VOLUME'
LU_TYPE = 'LOGICAL_UNIT'
VOPT_TYPE = 'VIRTUAL_OPTICAL_MEDIA'

PORT = 'Port'
PORT_NAME = 'PortName'

POST_RETRIES = CONF.ibmpowervm_update_collision_retries

# K2 extended properties
EXT_SCSI_MAP = k2const.VIOS_VSCSI_MAP_EXT_PROP
EXT_FC_MAP = k2const.VIOS_VFC_MAP_EXT_PROP

# constants to help initialize variables
SCSI_MAP_ATTRS = dict(map_type=SCSI_MAPPING_ELEM,
                      map_collection=SCSI_MAPPINGS,
                      serv_remote_lpar_path=SERVER_REMOTE_LPAR_REL_PATH,
                      serv_remote_slot_path=SERVER_REMOTE_SLOT_REL_PATH,
                      serv_backing_dev_path=SERVER_BACKING_DEV_REL_PATH,
                      k2_ext_prop=EXT_SCSI_MAP)
FC_MAP_ATTRS = dict(map_type=FC_MAPPING_ELEM,
                    map_collection=FC_MAPPINGS,
                    serv_remote_lpar_path=SERVER_CONN_LPAR_REL_PATH,
                    serv_remote_slot_path=SERVER_CONN_SLOT_REL_PATH,
                    serv_backing_dev_path=SERVER_MAP_PORT_REL_PATH,
                    k2_ext_prop=EXT_FC_MAP)
STG_ATTRS = dict(scsi=SCSI_MAP_ATTRS, vfc=FC_MAP_ATTRS)

REQ_RES_HMC_ERR_PREFIX = 'HSCL294D'

PROF_SYNC_ERR_MSG = ("This profile is synchronized with the partition's"
                     " current configuration")

# Updates failed indications
# Add workaround for defect 42465
VIOS_UPDATE_FAILED_HSC_ERRS = ['HSCL2523']
VIOS_UPDATE_FAILED = [k2const.HTTP_STATUS_ETAG_MISMATCH,
                      k2const.HTTP_STATUS_CONNECTION_RESET,
                      k2const.HTTP_STATUS_INVALID_URL]


def vios_is_valid_for_update(vios_resp, raise_nonfatal=False):

    def _get_msg_dict(vios_wrapper):
        msg_dict = {'vios_name': vios_wrapper.get_vios_name(),
                    'vios_state': vios_wrapper.get_vios_state()}
        return msg_dict

    vios_wrapper = VIOSEntryWrapper(vios_resp.entry)
    # Ensure the VIOS is running
    vios_state = vios_wrapper.get_vios_state()
    if not vios_wrapper.is_running() and vios_state != 'not activated':
        msg_dict = _get_msg_dict(vios_wrapper)
        raise hmc_ex.IBMPowerVIOSIsNotRunning(**msg_dict)

    # Check if the rmc state was busy, which means the data is not complete
    # We can just get the data again, so it's not a fatal error
    if vios_wrapper.is_running() and vios_wrapper.get_rmc_state() == 'busy':
        msg_dict = _get_msg_dict(vios_wrapper)
        if not raise_nonfatal:
            LOG.warning(_('Virtual I/O Server %(vios_name)s is currently '
                          'too busy to process update operations.') % msg_dict)
            return False
        else:
            vios_name = vios_wrapper.get_vios_name()
            raise hmc_ex.IBMPowerVIOSIsTooBusy(vios_name=vios_name)

    # Ensure the VIOS has RMC connection
    if vios_wrapper.is_running() and not vios_wrapper.is_rmc_active():
        msg_dict = _get_msg_dict(vios_wrapper)
        raise hmc_ex.IBMPowerVIOSRMCNotAcitve(**msg_dict)

    # Ensure the VIOS license has been accepted
    if not vios_wrapper.license_accepted():
        msg_dict = _get_msg_dict(vios_wrapper)
        raise hmc_ex.IBMPowerVIOSLicenseNotAccepted(**msg_dict)

    return True


def create_storage_element(backing_dev, storage_type, add_info=None):
    """
    Create a storage element that can be used for attach volume or
    part of create and attach in 1 POST
    :param backing_dev: volume name
    :param storage_type: LogicalUnit, PhysicalVolume, etc
    :param add_info: logical unit's UDID
    """
    schema_attr = k2const.DEFAULT_SCHEMA_ATTR
    # building client adapter
    if storage_type == VDISK:
        disk_elem = K2Element(VDISK,
                              attrib=schema_attr,
                              children=[K2Element(DISK_NAME,
                                                  text=backing_dev)])
    elif storage_type == PV:
        disk_elem = K2Element(PV,
                              attrib=schema_attr,
                              children=[K2Element(VOL_NAME,
                                                  text=backing_dev)])
    elif storage_type == LU:
        disk_elem = K2Element(LU,
                              attrib=schema_attr,
                              children=[K2Element(UDID,
                                                  text=add_info),
                                        K2Element(UNIT_NAME,
                                                  text=backing_dev)])
    elif storage_type == VOPT:
        disk_elem = K2Element(VOPT,
                              attrib=schema_attr,
                              children=[K2Element(MEDIA_NAME,
                                                  text=backing_dev)])
    else:
        raise hmc_ex.IBMPowerVMInvalidVIOConfig()

    return K2Element(STORAGE, children=[disk_elem])


class StorageMapProcessor(object):
    """
    This class is the parent class of SCSIMapProcessor and
    FibreChannelMapProcessor.
    It provides method to parse, extract, and manipulate K2
    VirtualSCSIMappings and VirtualFibreChannelMapping.
    """
    def __init__(self, storage_type=SCSI):
        """
        Initialize the storage map attributes.
        :param vios_entry: VirtualIOServer K2 entry
        """
        if storage_type not in STG_ATTRS.keys():
            raise hmc_ex.IBMPowerVMInvalidVIOConfig()

        # remember the type of storage & initialize the storage mapping
        # information
        self.storage_type = storage_type
        stg_attrs = STG_ATTRS[self.storage_type]
        self.map_type = stg_attrs['map_type']
        self.map_collection = stg_attrs['map_collection']
        self.serv_remote_lpar_path = stg_attrs['serv_remote_lpar_path']
        self.serv_remote_slot_path = stg_attrs['serv_remote_slot_path']
        self.serv_backing_dev_path = stg_attrs['serv_backing_dev_path']
        self.k2_ext_prop = stg_attrs['k2_ext_prop']

    def read_vios(self, oper, cec_uuid, vios_uuid, xag=None, age=-1):
        ''' This method is a raw read of the VIOS entry
        '''
        ext_props = xag if xag is not None else [self.k2_ext_prop]
        if cec_uuid is None:
            vios_resp = oper.vios_read(k2const.VIOS, rootId=vios_uuid,
                                       xag=ext_props, age=age,
                                       vios_name=vios_uuid)
        else:
            vios_resp = oper.vios_read(k2const.MGT_SYS,
                                       rootId=cec_uuid,
                                       childType=k2const.VIOS,
                                       childId=vios_uuid,
                                       xag=ext_props,
                                       age=age,
                                       vios_name=vios_uuid)
        if vios_resp is None:
            # raise exception
            raise hmc_ex.IBMPowerVMInvalidVIOConfig()

        # return the response, etag, and xag list used
        return vios_resp, vios_resp.headers.get(ETAG), ext_props

    def load_vios(self, oper, cec_uuid, vios_uuid,
                  xag=None, age=-1, retries=5, delay=5):
        '''
        This method load the VIOS entry. If the VIOS entry is not
        in the proper state it raises an exception or retries the
        read operation
        '''

        tries = 0
        while True:
            vios_resp, etag, ext_props = self.read_vios(oper, cec_uuid,
                                                        vios_uuid,
                                                        xag=xag, age=age)
            tries += 1
            raise_exc = tries > retries
            if vios_is_valid_for_update(vios_resp, raise_nonfatal=raise_exc):
                break
            else:
                greenthread.sleep(delay)

        # return the response, etag, and xag list used
        return vios_resp, etag, ext_props

    def initial_vios_load(self, k2resp, k2resp_xag, oper,
                          cec_uuid, vios_uuid, xag=None):

        # start, check if we need to get the vios info
        if k2resp is None:
            # Get the latest VIOS server properties
            vios_resp, etag, ext_props = self.load_vios(oper, cec_uuid,
                                                        vios_uuid,
                                                        xag=xag)

        else:
            # We were sent a vios resp.
            ext_props = k2resp_xag
            vios_resp = k2resp
            etag = vios_resp.headers.get(ETAG)

            # Validate the vios_resp
            valid = vios_is_valid_for_update(
                vios_resp, raise_nonfatal=False)

            if not etag or not valid:
                # Get the latest VIOS server properties
                vios_wrapper = VIOSEntryWrapper(vios_resp.entry)
                LOG.info(_("Retrieving Virtual I/O server information for"
                           " %s") % vios_wrapper.get_vios_name())

                vios_resp, etag, ext_props = (
                    self.load_vios(oper, cec_uuid, vios_uuid, xag=xag))

        return vios_resp, etag, ext_props

    def _check_update_failed(self, k2resp, vios_uuid):
        if (k2resp is not None and
                (k2resp.status in VIOS_UPDATE_FAILED or
                 contains_msgs_in_k2_err(None,
                                         VIOS_UPDATE_FAILED_HSC_ERRS,
                                         k2resp=k2resp))):
            raise hmc_ex.IBMPowerVMUpdateFailed(uuid=vios_uuid)

    def _find_new_slot_pair_created_for_lpar(self, client_lpar_id,
                                             new_storage_map,
                                             old_storage_map=None):
        """
        Find the new adapter pair created for a given client lpar id.
        This method is mainly used right after create_adapter_pair is called
        to figure out the slot numbers used for server & client adapter.
        """
        # find existing slot numbers for the client lpar
        old_client_slots = set()
        new_client_slots = set()
        client_server_lookup = {}
        if old_storage_map is not None:
            old_matched_entries = self.find_map_entries(old_storage_map,
                                                        client_lpar_id)
            for entry in old_matched_entries:
                client_adapter = (
                    entry.findtext(CLIENT_VIR_SLOT_NUM_REL_PATH))
                if client_adapter is not None:
                    # record the slot number
                    old_client_slots.add(client_adapter)

        new_matched_entries = self.find_map_entries(new_storage_map,
                                                    client_lpar_id)
        wwpns_list = []
        wwpn_slot_map = dict()
        for entry in new_matched_entries:
            client_adapter = (
                entry.findtext(CLIENT_VIR_SLOT_NUM_REL_PATH))
            # only keep track of slot number that's different from existing
            if (client_adapter is not None and
                    client_adapter not in old_client_slots):
                new_client_slots.add(client_adapter)
                # find the corresponding server adapter
                server_adapter = (
                    entry.findtext(SERVER_VIR_SLOT_NUM_REL_PATH))
                # extract wwpn if it's NPIV
                if self.storage_type == VFC:
                    wwpns = entry.findtext(CLIENT_WWPNS_REL_PATH)
                    wwpns_list.append(wwpns)
                else:
                    wwpns = None

                if server_adapter is not None:
                    client_server_lookup[client_adapter] = (server_adapter,
                                                            wwpns)
                # Fetch wwpn_slot if npiv
                phy_wwpn = entry.findtext(ROOT + PORT + DELIM + "WWPN")
                if phy_wwpn:
                    if phy_wwpn not in wwpn_slot_map:
                        wwpn_slot_map[phy_wwpn] = []
                    wwpn_slot_map[phy_wwpn].append(client_adapter)

        # if there are more than 1 slot only return 1 and log a warning
        if len(new_client_slots) > 1:
            LOG.warning(_('More than one new virtual slot pair is found %s')
                        % client_server_lookup)

        return {'client_slot_num': new_client_slots,
                'server_slot_num': client_server_lookup,
                'wwpn_slot_map': wwpn_slot_map}

    @logcall
    def find_map_entries(self, vios_entry, lpar_id=None,
                         slot_num=None, backing_dev=None,
                         with_valid_client_adapter=False,
                         without_port_name=False):
        """
        Iterate through VirtualSCSIMappings or VirtualFibreChannelMappings
        to find entries matching to the specified client lpar id, slot num,
        and backing device.
        It's required that at least lpar_id or backing_dev is not None.
        :param vios_entry: K2Element pertains to the either VirtualSCSIMappings
                           or VirtualFibreChannelMappings.
        :param lpar_id: partition ID of the client lpar
        :param slot_num: virtual slot number of the client adapter
        :param backing_dev: physical port name or volume name(i.e fcs2 or
                            volume-8d075364-59ad-40aa-8f19-4f2d2867582e)
        :param with_valid_client_adapter: mapping with ClientAdapter element
        :param without_port_name: ServerAdapter element without MapPort attr
        :return list VirtualSCSIMapping or VirtualFibreChannelMapping
        """
        # extract either VirtualSCSIMapping or VirtualFibreChannelMapping
        # based on the element being passed in
        stg_type = SCSI if vios_entry.gettag() == SCSI_MAPPINGS else VFC
        stg_map_attrs = STG_ATTRS.get(stg_type)
        map_type = stg_map_attrs['map_type']
        serv_remote_lpar_path = stg_map_attrs['serv_remote_lpar_path']
        serv_remote_slot_path = stg_map_attrs['serv_remote_slot_path']
        map_entries = vios_entry.findall(ROOT + map_type)

        entries = []
        # handle whether input is int or string
        if isinstance(lpar_id, int):
            client_lpar_id = str(lpar_id)
        else:
            client_lpar_id = lpar_id

        if isinstance(slot_num, int):
            client_slot_num = str(slot_num)
        else:
            client_slot_num = slot_num

        if map_entries:
            for entry in map_entries:
                # since there can only be one entry matching to backing device
                # specified, we will check for backing device first.
                # if virtual target device name is specified, find
                # a match
                if backing_dev:
                    if self.storage_type == SCSI:
                        curr_backing_dev = entry.findtext(
                            SERVER_BACKING_DEV_REL_PATH)
                    else:
                        curr_backing_dev = entry.findtext(
                            SERVER_MAP_PORT_REL_PATH)
                    # if backing_dev is not a list, conver to list
                    backing_dev_list = (
                        backing_dev if isinstance(backing_dev, list) else
                        [backing_dev])
                    if (curr_backing_dev and
                            curr_backing_dev in backing_dev_list):
                        LOG.debug('Found matching backing device %s' %
                                  curr_backing_dev)
                    else:
                        continue
                else:
                    if without_port_name:
                        # make sure there is no PortName element
                        curr_backing_dev = entry.findtext(
                            SERVER_MAP_PORT_REL_PATH)
                        if curr_backing_dev is not None:
                            continue

                found_match_lpar = False
                if client_lpar_id is not None:
                    if client_slot_num is not None:
                        client_adapter = (
                            entry.findtext(CLIENT_VIR_SLOT_NUM_REL_PATH))
                    client_lpar = (entry.findtext(CLIENT_LPAR_REL_PATH))

                    if (client_lpar is not None and
                            client_lpar == client_lpar_id and
                            (client_slot_num is None or
                             (client_adapter is not None and
                              client_slot_num == client_adapter))):
                        LOG.debug('found matching client adapter %s' %
                                  entry.toxmlstring())
                        found_match_lpar = True
                    elif client_lpar is None and not with_valid_client_adapter:
                        # this would occurred if the client lpar has been
                        # deleted or only server adapter is created.
                        # we need to check the server adapter to find a match
                        server_remote_lpar_id = (
                            entry.findtext(serv_remote_lpar_path))
                        if client_slot_num is not None:
                            server_remote_slot = (
                                entry.findtext(serv_remote_slot_path))
                        # if we found the ServerAdapter's remote lpar ID,
                        # make sure it matches to client_lpar_id.
                        # if user specify the client slot num, make sure it
                        # matches as well
                        if (server_remote_lpar_id is not None and
                                server_remote_lpar_id == client_lpar_id and
                                (client_slot_num is None or
                                 (server_remote_slot is not None and
                                  client_slot_num == server_remote_slot))):
                            LOG.debug('found matching client adapter %s'
                                      % entry.toxmlstring())
                            found_match_lpar = True
                else:
                    found_match_lpar = True
                if found_match_lpar:
                    entries.append(entry)
        # reduce logging, only log if backing_dev is specified.
        if not entries and backing_dev is not None:
            LOG.debug(_('No matching entries with LPAR ID: %(lpar_id)s, Slot'
                        ' Number: %(slot_num)s, Backing Device: '
                        '%(backing_dev)s. Dumping entries:\n%(vios_entry)s')
                      % {'lpar_id': client_lpar_id,
                         'slot_num': client_slot_num,
                         'backing_dev': six.text_type(backing_dev),
                         'vios_entry': vios_entry.toxmlstring()})
        return entries

    @logcall
    def detach_and_delete_adapter(self, oper, cec_uuid, vios_uuid,
                                  client_lpar_id=None, client_slot_num=None,
                                  backing_dev=None, resp=None,
                                  detach_only=False, ignore_type=False):
        """
        To detach storage and remove adapter via k2.
        If backing_dev is provided, it will try to find an entry matching
        with the backing device specified. If it's the last storage attached
        for the virtual slot. It will delete the virtual server slot.
        Otherwise, it will find all the entries matching with the client lpar
        and slot number, and detach & delete.
        If response is not passed in, we will extract the latest via
        k2operator.
        :param oper: K2Operator that will be used for GET and POST
        :param cec_uuid: k2 uuid of the managed system
        :param vios_uuid: k2 uuid of the virtual io server
        :param client_lpar_id: partition ID of the client lpar
        :param client_slot_num: virtual slot number of the client adapter
        :param backing_dev: physical port name or volume name(i.e fcs2,
                            volume-8d075364-59ad-40aa-8f19-4f2d2867582e)
        :param resp: VirtualIOServer K2 entry. Provided for performance
                     purpose. If it's not provided, we will issue a GET to
                     extract the most up-to-date VirtualIOServer output
        :return VirtualIOServer entry from the K2 POST response
        """
        def _get_k2_ext_props():
            props = []
            if ignore_type:
                props = [EXT_SCSI_MAP, EXT_FC_MAP]
            else:
                props.append(self.k2_ext_prop)
            return props

        def _detach_and_delete_adapter(k2resp):
            # start, check if we need to get the vios info
            xag = _get_k2_ext_props()
            vios_resp, etag, ext_props = (
                self.initial_vios_load(k2resp, [], oper, cec_uuid, vios_uuid,
                                       xag=xag))

            # if user input ignore_type, we will go through both scsi and fc
            # mappings to delete adapters
            if ignore_type:
                map_collections = [SCSI_MAPPINGS, FC_MAPPINGS]
            else:
                map_collections = [self.map_collection]

            for map_collection in map_collections:
                storage_map = vios_resp.entry.element.find(ROOT +
                                                           map_collection)

                if storage_map:
                    # find the entries with matching info
                    entries = self.find_map_entries(
                        storage_map, client_lpar_id, client_slot_num,
                        backing_dev)

                    if entries:
                        new_entries = None
                        if detach_only:
                            new_entries = self.remove_storage_elem(entries)
                        for elem in entries:
                            storage_map.remove(elem)
                        if new_entries:
                            for new_elem in new_entries:
                                storage_map.append(new_elem)

            try:
                return self.k2oper_update_vios(oper, cec_uuid, vios_uuid,
                                               etag, vios_resp,
                                               ext_props=ext_props)
            except K2Error as e:
                self._check_update_failed(e.k2response, vios_uuid)
                # if we get here then it's an unexpected error
                self._log_k2err(vios_resp.entry.element, e)
                raise e

        @synchronized('stg-mapping-lock' + vios_uuid)
        def _call_detach_and_delete_adapter():
            """
            With built-in retry logics if we get back etag mismatch error.
            """
            k2resp = resp
            for retries in xrange(POST_RETRIES + 1):
                try:
                    return _detach_and_delete_adapter(k2resp)
                except (hmc_ex.IBMPowerVMUpdateFailed,
                        hmc_ex.IBMPowerVIOSIsTooBusy) as e:
                    LOG.debug('Retry _call_detach_and_delete_adapter %s'
                              % retries)
                    if retries == POST_RETRIES:
                        raise e
                    else:
                        k2resp = None
                except Exception as e:
                    LOG.exception(e)
                    raise e

        _call_detach_and_delete_adapter()

    @logcall
    def create_adapter_pair(self, oper, cec_uuid, vios_uuid, client_lpar_href,
                            client_lpar_id, resp=None, num_adapter=1,
                            ext_slot_info=False, ext_props=[]):
        """
        Create a virtual scsi or fibre channel server/client adapter pair.
        If the storage mapping is not provided, we will extract from the
        latest via k2operator.
        :param oper: K2Operator that will be used for GET and POST
        :param cec_uuid: k2 uuid of the managed system
        :param vios_uuid: k2 uuid of the virtual io server
        :param client_href: href of the client logical partition.
        :param client_lpar_id: partition ID of the client lpar
        :param resp: VirtualIOServer K2 entry. Provided for performance
                     purpose. If it's not provided, we will issue a GET to
                     extract the most up-to-date VirtualIOServer output
        :param num_adapter: number of new adapter pair to create
        :param ext_slot_info: extract newly created slot information
        :return dictionary containg the newly created adapter pair and
                VirtualIOServer entry from the K2 POST response
        """
        num_scsi = 0
        num_fc = 0
        if self.storage_type == SCSI:
            num_scsi = num_adapter
        else:
            num_fc = num_adapter
        return self.create_mixed_adapter_pairs(oper, cec_uuid, vios_uuid,
                                               client_lpar_href,
                                               client_lpar_id,
                                               num_scsi, num_fc,
                                               delete_old_adapters=False,
                                               ext_slot_info=ext_slot_info,
                                               resp=resp,
                                               k2_ext_props=ext_props)

    def attach_storage(self, oper, cec_uuid, vios_uuid,
                       client_lpar_id, client_slot_num,
                       backing_dev, storage_type, add_info=None, resp=None):
        """
        To attach storage to the specified client lpar and slot via k2.
        It will try to find an entry matching
        If response is not passed in, we will extract the latest via
        k2operator.
        """
        raise NotImplementedError()

    def populate_client_adapter(self, lpar_id, slot_num, remote_lpar,
                                remote_slot):
        raise NotImplementedError()

    def populate_server_adapter(self, lpar_id, slot_num, remote_lpar,
                                remote_slot):
        raise NotImplementedError()

    def check_and_create_adapter_pair(self, oper, cec_uuid, vios_uuid,
                                      client_lpar_href,
                                      client_lpar_id, resp=None,
                                      num_adapter=1):
        """
        Make sure there are sufficient number of adapters created for the lpar.
        Otherwise, create the number of adapters as specified.
        :param oper: K2Operator that will be used for GET and POST
        :param cec_uuid: k2 uuid of the managed system
        :param vios_uuid: k2 uuid of the virtual io server
        :param client_href: href of the client logical partition.
        :param client_lpar_id: partition ID of the client lpar
        :param resp: VirtualIOServer K2 entry. Provided for performance
                     purpose. If it's not provided, we will issue a GET to
                     extract the most up-to-date VirtualIOServer output
        :param num_adapter: number of new adapter pair to create
        :return dictionary containg the newly created adapter pair and
                VirtualIOServer entry from the K2 POST response
        """
        # start, check if we need to get the vios info
        vios_resp, etag, ext_props = self.initial_vios_load(resp, [], oper,
                                                            cec_uuid,
                                                            vios_uuid)

        storage_map = vios_resp.entry.element.find(ROOT + self.map_collection)

        # find the entries with matching info
        entries = self.find_map_entries(storage_map, client_lpar_id,
                                        with_valid_client_adapter=True,
                                        without_port_name=True)
        num_entries = 0 if entries is None else len(entries)
        # if there sufficient number of adapters already, do nothing.
        if num_entries >= num_adapter:
            LOG.debug("No need to create adapter pairs. Num_entries=%d" %
                      num_entries)
            return None
        return self.create_adapter_pair(oper, cec_uuid, vios_uuid,
                                        client_lpar_href,
                                        client_lpar_id, vios_resp,
                                        num_adapter - num_entries,
                                        ext_props=ext_props)

    @logcall
    def remove_storage_elem(self, storage_map_entries):
        new_map_entries = []
        for map_entry in storage_map_entries:
            # make copy of server & client adapter and port
            server_adapter = copy.deepcopy(map_entry.find(SERVER_ADAPTER_PATH))
            client_adapter = copy.deepcopy(map_entry.find(CLIENT_ADAPTER_PATH))
            associated_lpar = copy.deepcopy(map_entry.find(ROOT + ASSOC_LPAR))

            # Create the list of children
            children = []
            for child in [associated_lpar, client_adapter, server_adapter]:
                if child is not None:
                    children.append(child)

            schema_attr = k2const.DEFAULT_SCHEMA_ATTR
            new_map_entry = K2Element(self.map_type,
                                      attrib=schema_attr,
                                      children=children)
            new_map_entries.append(new_map_entry)
        return new_map_entries

    def _log_k2err(self, req, k2err):
        LOG.error(_('Failed to modify object with request %s.')
                  % req.toxmlstring() if req else '')
        resp = (k2err.k2response.body
                if k2err and k2err.k2response and k2err.k2response.body
                else '')
        LOG.error(_('Failed to modify object with response %s.') % resp)

    @logcall
    def _create_adapter_pair_entry(self, cec_uuid, vios_uuid,
                                   client_lpar_href,
                                   adapter_type=None, backing_dev=None,
                                   dev_type=None, add_info=None):
        """
        This method should only be called after we obtain the lock via
        synchronized.
        Create a new adapter pair entry based on adapter_type specified.
        If adapter_type is not specified, we would construct the entry
        based on self.map_type
        """
        schema_attr = k2const.DEFAULT_SCHEMA_ATTR

        # building associated lpar
        if not adapter_type:
            map_type = self.map_type
        else:
            stg_attrs = STG_ATTRS[adapter_type]
            map_type = stg_attrs['map_type']

        assoc_lpar_elem_name = ASSOC_LPAR

        associated_lpar = K2Element(assoc_lpar_elem_name,
                                    attrib={'href': client_lpar_href,
                                            'rel': 'related'})

        new_serv_adpt = K2Element(SERVER_ADAPTER,
                                  attrib=k2const.DEFAULT_SCHEMA_ATTR,
                                  children=[K2Element(ADAPTER_TYPE,
                                                      text=SERVER),
                                            K2Element(USE_NEXT_AVAIL_SLOT,
                                                      text='true')])

        new_client_adpt = K2Element(CLIENT_ADAPTER,
                                    attrib=k2const.DEFAULT_SCHEMA_ATTR,
                                    children=[K2Element(ADAPTER_TYPE,
                                                        text=CLIENT),
                                              K2Element(USE_NEXT_AVAIL_SLOT,
                                                        text='true')])
        # construct VirtualSCSIMapping or VirtualFibreChannelMapping
        children_list = [associated_lpar, new_client_adpt,
                         new_serv_adpt]
        if adapter_type == VFC and backing_dev:
            port = K2Element(PORT, attrib=schema_attr,
                             children=[K2Element(PORT_NAME, text=backing_dev)])
            children_list = [associated_lpar, new_client_adpt,
                             port,
                             new_serv_adpt]
        elif backing_dev and dev_type:
            # it's scsi type, create the storge element based on the type
            storage_elem = create_storage_element(backing_dev, dev_type,
                                                  add_info)
            children_list = [associated_lpar, new_client_adpt,
                             new_serv_adpt, storage_elem]
        new_map_entry = K2Element(map_type,
                                  attrib=schema_attr,
                                  children=children_list)

        return new_map_entry

    @logcall
    def create_mixed_adapter_pairs(self, oper, cec_uuid, vios_uuid,
                                   client_lpar_href,
                                   client_lpar_id,
                                   num_scsi=0, num_fc=0,
                                   delete_old_adapters=True,
                                   ext_slot_info=False,
                                   resp=None, fc_backing_devs=None,
                                   k2_ext_props=[],
                                   scsi_backing_devs=None):
        """
        Create a virtual scsi and fibre channel server/client adapter pairs.
        If the storage mapping is not provided, we will extract from the
        latest via k2operator.
        This method typically is used as part of deploy. To guarantee the
        adapter configuration is clean, we will remove old adapters associated
        with the client_lpar_id.
        :param oper: K2Operator that will be used for GET and POST
        :param cec_uuid: k2 uuid of the managed system
        :param vios_uuid: k2 uuid of the virtual io server
        :param client_href: href of the client logical partition.
        :param client_lpar_id: partition ID of the client lpar
        :param num_scsi: number of new scsi adapter pair to create
        :param num_fc: number of new fc adapter pair to create
        :param delete_old_adapters: Whether to delete old adapters associated
                                    with the client_lpar_id.
        :param ext_slot_info: Extract the newly created adapter pair info
        :param resp: VirtualIOServer K2 entry. Provided for performance
                     purpose. If it's not provided, we will issue a GET to
                     extract the most up-to-date VirtualIOServer output
        :param fc_backing_devs: physical port map for vfc
        :param k2_ext_props: K2 extended properties for GET & POST
        :param scsi_backing_devs: volumes info needed for attach.
        :return dictionary containg the VirtualIOServer entry from the
                K2 POST response
        """
        def _get_k2_ext_props():
            props = []
            if delete_old_adapters:
                props = [EXT_SCSI_MAP, EXT_FC_MAP]
            else:
                if num_scsi > 0:
                    props.append(EXT_SCSI_MAP)
                if num_fc > 0:
                    props.append(EXT_FC_MAP)
            return props

        def _create_mixed_adapter_pairs(k2resp):
            # check if we need to get the vios info
            vios_resp, etag, ext_props = (
                self.initial_vios_load(k2resp, k2_ext_props, oper,
                                       cec_uuid, vios_uuid,
                                       xag=_get_k2_ext_props()))

            if ext_slot_info:
                old_stg_maps = {}

            # make copy of fc_backing_devs
            fc_devs = copy.copy(fc_backing_devs)
            scsi_devs = copy.copy(scsi_backing_devs)
            adapter_num_info = {SCSI: num_scsi, VFC: num_fc}
            for adapter_type, num_adapter in adapter_num_info.items():
                stg_attrs = STG_ATTRS[adapter_type]
                elem_name = stg_attrs['map_collection']
                storage_maps = vios_resp.entry.element.find(ROOT + elem_name)
                # delete old adapter entries
                if delete_old_adapters:
                    entries_to_delete = self.find_map_entries(
                        storage_maps, client_lpar_id)
                    if entries_to_delete:
                        for entry in entries_to_delete:
                            storage_maps.remove(entry)

                # save copy of storage map before new adapters is created
                if ext_slot_info:
                    old_stg_maps[adapter_type] = copy.deepcopy(storage_maps)
                for num in xrange(num_adapter):
                    backing_dev = None
                    dev_type = None
                    add_info = None
                    if adapter_type == VFC and fc_devs:
                        backing_dev = fc_devs.pop()
                    elif adapter_type == SCSI and scsi_devs:
                        backing_dev, dev_type, add_info = scsi_devs.pop()
                    new_map_entry = self._create_adapter_pair_entry(
                        cec_uuid, vios_uuid,
                        client_lpar_href, adapter_type, backing_dev,
                        dev_type=dev_type, add_info=add_info)
                    LOG.debug("%d: Create %s adapter entry for device %s "
                              "on VIO %s." %
                              (num, adapter_type, backing_dev, vios_uuid))
                    storage_maps.append(new_map_entry)

            try:
                ret_resp = self.k2oper_update_vios(oper, cec_uuid, vios_uuid,
                                                   etag, vios_resp,
                                                   ext_props=ext_props)
            except K2Error as e:
                self._check_update_failed(e.k2response, vios_uuid)
                # if we get here then it's an unexpected error
                self._log_k2err(vios_resp.entry.element, e)
                raise e

            ret_data = {}
            # extract newly created slot info
            if ext_slot_info:
                LOG.debug('dump k2 response:\n%s' % ret_resp.body)
                for adapter_type in adapter_num_info.keys():
                    elem_name = STG_ATTRS[adapter_type]['map_collection']
                    new_stg_maps = ret_resp.entry.element.find(ROOT +
                                                               elem_name)
                    slot_data = self._find_new_slot_pair_created_for_lpar(
                        client_lpar_id, new_stg_maps,
                        old_stg_maps[adapter_type])
                    if not ret_data.get('client_slot_num'):
                        ret_data['client_slot_num'] = []
                    ret_data['client_slot_num'] += slot_data['client_slot_num']
                    if not ret_data.get('server_slot_num'):
                        ret_data['server_slot_num'] = []
                    ret_data['server_slot_num'] += slot_data['server_slot_num']
                    if not ret_data.get('slot_info'):
                        ret_data['slot_info'] = {}
                    ret_data['slot_info'][adapter_type] = \
                        slot_data['client_slot_num']
                    ret_data['wwpn_slot_map'] = slot_data['wwpn_slot_map']

            ret_data['new_vios_entry'] = ret_resp
            return ret_data

        @synchronized('stg-mapping-lock' + vios_uuid)
        def _call_create_mixed_adapter_pairs():
            """
            With built-in retry logics if we get back etag mismatch error.
            """
            k2resp = resp
            for retries in xrange(POST_RETRIES + 1):
                try:
                    return _create_mixed_adapter_pairs(k2resp)
                except (hmc_ex.IBMPowerVMUpdateFailed,
                        hmc_ex.IBMPowerVIOSIsTooBusy) as e:
                    LOG.debug('Retry  _call_create_adapter_pair %s' % retries)
                    if retries == POST_RETRIES:
                        # dump all the k2 requests & responses accumulated
                        # since the wrapper would not log single
                        # HTTP_STATUS_ETAG_MISMATCH failure
                        oper.dump_logs()
                        raise e
                    else:
                        k2resp = None
                except Exception as e:
                    LOG.exception(e)
                    if isinstance(e, hmc_ex.IBMPowerVMInvalidVIOConfig):
                        raise e
                    raise hmc_ex.IBMPowerVMCreateAdapter(uuid=vios_uuid)

        return _call_create_mixed_adapter_pairs()

    def k2oper_update_vios(self, oper, sys_uuid, vios_uuid, etag,
                           vios_resp, ext_props=[]):
        """
        Make sure CurrentProfileSync is set to On whenever we perform a
        POST to vios.
        """
        syncToProf = vios_resp.entry.element.find(CUR_PROF_SYNC_PATH)
        ex_k2_rr_vopt_errs = ['HSCLA9D6', 'HSCLA9D7']
        if syncToProf is not None:
            syncToProf.settext('On')
        try:
            ret_resp = oper.update(vios_resp.entry.element,
                                   etag,
                                   k2const.MGT_SYS,
                                   rootId=sys_uuid,
                                   childType=k2const.VIOS,
                                   childId=vios_uuid, xag=ext_props)
        except Exception as e:
            if contains_msg_in_k2_err(e, VIOS_BUSY_ERR_MSG):
                raise hmc_ex.IBMPowerVIOSIsTooBusy(vios_name=vios_uuid)
            elif contains_msgs_in_k2_err(e, ex_k2_rr_vopt_errs):
                self._log_k2err(vios_resp.entry.element, e)
                # If the simplified remote restart capability is not
                # updated to the system
                LOG.warning(_("The configuration change was successful, "
                              "however the simplified remote restart"
                              " cabaility was not applied. Please run"
                              " refdev command after removing the"
                              " attached vopt device."))
                ret_resp, etag, ext_props = self.read_vios(oper,
                                                           sys_uuid,
                                                           vios_uuid,
                                                           xag=ext_props,
                                                           age=0)
            else:
                raise e

        return ret_resp

    @logcall
    def cleanup_orphaned_adapters(self, oper, cec_uuid, vios_uuid,
                                  client_lpar_id=None, resp=None,
                                  ignore_type=False):
        """
        To clean up orphaned adapters via k2.
        It will find all the entries matching with the client lpar but without
        ClientAdapter and delete them.
        If response is not passed in, we will extract the latest via
        k2operator.
        :param oper: K2Operator that will be used for GET and POST
        :param cec_uuid: k2 uuid of the managed system
        :param vios_uuid: k2 uuid of the virtual io server
        :param client_lpar_id: partition ID of the client lpar
        :param resp: VirtualIOServer K2 entry. Provided for performance
                     purpose. If it's not provided, we will issue a GET to
                     extract the most up-to-date VirtualIOServer output
        :return VirtualIOServer entry from the K2 POST response
        """
        def _get_k2_ext_props():
            props = []
            if ignore_type:
                props = [EXT_SCSI_MAP, EXT_FC_MAP]
            else:
                props.append(self.k2_ext_prop)
            return props

        def _cleanup_orphaned_adapters(k2resp, disable_sync_to_curr):
            """
            :param k2resp: VirtualIOServer's K2Response
            :param disable_sync_to_curr: If VIOS is activated with required
                adapter, we must suspend the sync in order to modify the
                profile.
            """

            vios_resp, etag, ext_props = (
                self.initial_vios_load(k2resp, [], oper, cec_uuid, vios_uuid,
                                       xag=_get_k2_ext_props()))

            # keep copy of the original vios_resp, so if we failed
            # removed adapter due to the required setting. We need
            # the original response to disable sync first.
            orig_vios_resp = copy.deepcopy(vios_resp)
            orphaned_adapters = self.check_for_orphaned_adapters(
                vios_resp.entry, client_lpar_id, ignore_type)

            # if there are any orphaned adapters, remove them
            if orphaned_adapters:
                for map_collection, elems in orphaned_adapters.iteritems():
                    storage_map = vios_resp.entry.element.find(ROOT +
                                                               map_collection)
                    for elem in elems:
                        storage_map.remove(elem)

                try:
                    if not disable_sync_to_curr:
                        return self.k2oper_update_vios(oper, cec_uuid,
                                                       vios_uuid,
                                                       etag, vios_resp,
                                                       ext_props=ext_props)
                    else:
                        # we need to break up the remove & sync modification
                        # otherwise, HMC will try perform sync first whinc
                        # in turn modified the profile and undo the changes
                        # we have done in the previous iteration
                        updated_vios_resp = oper.update(
                            vios_resp.entry.element,
                            etag,
                            k2const.MGT_SYS,
                            rootId=cec_uuid,
                            childType=k2const.VIOS,
                            childId=vios_uuid, xag=ext_props)
                        # enable the sync back on
                        etag = updated_vios_resp.headers.get(ETAG)
                        return self.k2oper_update_vios(oper, cec_uuid,
                                                       vios_uuid,
                                                       etag,
                                                       updated_vios_resp,
                                                       ext_props=ext_props)
                except K2Error as e:
                    self._check_update_failed(e.k2response, vios_uuid)
                    # if we get here then it's an unexpected error

                    # if the operation failed due to required setting,
                    # modify the profile and failed this request.
                    # The retry will kick in with disable_sync_to_curr
                    # to True
                    if contains_msg_in_k2_err(e, REQ_RES_HMC_ERR_PREFIX):
                        self._undo_req_vio_slots_from_prof(
                            oper, orig_vios_resp)
                        raise hmc_ex.IBMPowerVMRmReqAdapter(uuid=vios_uuid)
                    else:
                        self._log_k2err(vios_resp.entry.element, e)
                        raise e

        @synchronized('stg-mapping-lock' + vios_uuid)
        def _call_cleanup_orphaned_adapters():
            """
            With built-in retry logics if we get back etag mismatch error.
            """
            k2resp = resp
            disable_sync = False
            for retries in xrange(POST_RETRIES + 1):
                try:
                    return _cleanup_orphaned_adapters(k2resp, disable_sync)
                except (hmc_ex.IBMPowerVMUpdateFailed,
                        hmc_ex.IBMPowerVMRmReqAdapter,
                        hmc_ex.IBMPowerVIOSIsTooBusy) as e:
                    LOG.debug('Retry _call_cleanup_orphaned_adapters %s'
                              % retries)
                    # set disable_sync to True for the next iteration
                    if isinstance(e, hmc_ex.IBMPowerVMRmReqAdapter):
                        disable_sync = True
                    if retries == POST_RETRIES:
                        raise e
                    else:
                        k2resp = None
                except Exception as e:
                    LOG.exception(e)
                    raise e

        _call_cleanup_orphaned_adapters()

    @logcall
    def check_for_orphaned_adapters(self, vios_entry, client_lpar_id,
                                    ignore_type=True):
        """
        Check whether the vios entry contains any orphaned adapters for a
        client lpar id.
        If it's a VFC, it also find entries that doesn't map to a physical port
        :return if there are any orphaned adapters, it would be placed in a
        dict categorized by VirtualFibreChannelMappings or VirtualSCSIMappings.
        """
        orphaned_adapters = {}
        if ignore_type:
            map_collections = [SCSI_MAPPINGS, FC_MAPPINGS]
        else:
            map_collections = [self.map_collection]

        for map_collection in map_collections:
            storage_map = vios_entry.element.find(ROOT + map_collection)

            if storage_map is not None:
                # find the entries with matching info
                entries = self.find_map_entries(storage_map, client_lpar_id)
                if entries:
                    for elem in entries:
                        # check whether there is ClientAdapter
                        client_adapter = elem.find(ROOT + CLIENT_ADAPTER)
                        if client_adapter is None:
                            if orphaned_adapters.get(map_collection) is None:
                                orphaned_adapters[map_collection] = []
                            orphaned_adapters[map_collection].append(elem)
                # find the entries without mapping to physical port
                if map_collection == FC_MAPPINGS:
                    # only find vfc with client/server pair but without
                    # physical port
                    entries = self.find_map_entries(
                        storage_map, client_lpar_id,
                        with_valid_client_adapter=True, without_port_name=True)
                    if entries:
                        for elem in entries:
                            if orphaned_adapters.get(map_collection) is None:
                                orphaned_adapters[map_collection] = []
                            orphaned_adapters[map_collection].append(elem)
        return orphaned_adapters

    @logcall
    def _undo_req_vio_slots_from_prof(self, oper, vios_resp, lpar_id=None):
        """
        Mark required vio slot as desired for a given vios and lpar id.
        :param oper: K2Operator that will be used for GET and POST
        :param vios_entry: VirtualIOServer K2 response
        :param lpar_id: Only modify the virtual slot config if it's for the
                        lpar_id. If it's not provided, we will mark all vio
                        slot as desired.
        """
        def _update_vios_prof(vios_uuid, prof_uuid, prof_resp, prof_href):
            def _k2_update_vios_prof(k2resp):
                if k2resp is None:
                    k2resp = oper.readbyhref(prof_href)
                prof_entry = k2resp.entry
                prof_wrapper = LPARProfileEntryWrapper(prof_entry)
                mod = prof_wrapper.undo_req_vfc_vscsi_slots(lpar_id)
                if mod:
                    try:
                        oper.update(prof_wrapper._k2entry.element,
                                    k2resp.headers.get(ETAG),
                                    k2const.VIOS, vios_uuid,
                                    k2const.LPAR_PROF, prof_uuid)
                    except K2Error as e:
                        self._check_update_failed(e.k2response, prof_uuid)
                        # if we get here then it's an unexpected error
                        if contains_msg_in_k2_err(e, PROF_SYNC_ERR_MSG):
                            # temporarily turned off the sync
                            self._turned_off_vios_sync(oper, vios_uuid,
                                                       vios_resp)
                            raise hmc_ex.IBMPowerVMUpdateFailed(
                                uuid=prof_uuid)
                        raise e

            @synchronized('vios-prof-lock' + prof_uuid)
            def _call_k2_update_vios_prof():
                k2resp = prof_resp
                for retries in xrange(POST_RETRIES + 1):
                    try:
                        return _k2_update_vios_prof(k2resp)
                    except (hmc_ex.IBMPowerVMUpdateFailed,
                            hmc_ex.IBMPowerVIOSIsTooBusy) as e:
                        LOG.debug('Retry _call_k2_update_vios_prof %s'
                                  % retries)
                        if retries == POST_RETRIES:
                            raise hmc_ex.IBMPowerVMDelReqProfFailed(
                                uuid=prof_uuid)
                        else:
                            k2resp = None
                    except Exception as e:
                        LOG.exception(e)
                        raise e

            _call_k2_update_vios_prof()

        vios_wrapper = VIOSEntryWrapper(vios_resp.entry)
        prof_href = vios_wrapper.get_lpar_prof_path()
        prof_resp = oper.readbyhref(prof_href)
        prof_entry = prof_resp.entry
        prof_wrapper = LPARProfileEntryWrapper(prof_entry)
        _update_vios_prof(vios_wrapper.get_uuid(), prof_wrapper.get_uuid(),
                          prof_resp, prof_href)

    def _turned_off_vios_sync(self, oper, vios_uuid, resp, ext_props=[]):
        """
        Make sure CurrentProfileSync is set to On whenever we perform a
        POST to vios.
        """
        def _do_turned_off_vios_sync(k2resp):
            vios_resp, etag, new_ext_props = (
                self.initial_vios_load(k2resp, ext_props, oper,
                                       None, vios_uuid, xag=[]))

            try:
                syncToProf = vios_resp.entry.element.find(CUR_PROF_SYNC_PATH)
                if syncToProf is not None:
                    syncToProf.settext('Suspended')
                    return oper.update(vios_resp.entry.element,
                                       etag,
                                       k2const.VIOS,
                                       rootId=vios_uuid, xag=new_ext_props)
            except K2Error as e:
                self._check_update_failed(e.k2response, vios_uuid)
                # if we get here then it's an unexpected error
                self._log_k2err(vios_resp.entry.element, e)
                raise e

        def _call_turned_off_vios_sync():
            """
            With built-in retry logics if we get back etag mismatch error.
            """
            k2resp = resp
            for retries in xrange(POST_RETRIES + 1):
                try:
                    return _do_turned_off_vios_sync(k2resp)
                except hmc_ex.IBMPowerVMUpdateFailed as e:
                    LOG.debug('Retry _call_turned_off_vios_sync %s'
                              % retries)
                    if retries == POST_RETRIES:
                        raise e
                    else:
                        k2resp = None
                except Exception as e:
                    LOG.exception(e)
                    raise e

        _call_turned_off_vios_sync()
