# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from nova import context as ctx
from nova.network import base_api as base_net_api
from nova.network import model as network_model
from powervc_nova import _
import six

from oslo_config import cfg

from powervc_oslo.k2.k2operator import K2Element

from powervc_nova import logcall
from powervc_nova.objects.network import dom as network_dom
from powervc_nova.virt.ibmpowervm.hmc.exception import IBMPowerMACNotSupported
from powervc_nova.virt.ibmpowervm.hmc.k2 import constants as k2const
from powervc_nova.virt.ibmpowervm.hmc.k2.entry_wrapper import EntryWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2.entry_wrapper import DELIM
from powervc_nova.virt.ibmpowervm.hmc.k2.entry_wrapper import ROOT
from powervc_nova.virt.ibmpowervm.hmc.k2.lparentry_wrapper \
    import LPAREntryWrapper
from powervc_nova.virt.ibmpowervm.vif.hmc import utils
from powervc_nova.network.neutronv2.ibmpowervm_api import PowerVMAPI as API

from oslo_log import log as logging
from nova import objects

LOG = logging.getLogger(__name__)
MAC_ADDRESS = ROOT + DELIM + 'MACAddress'
CONF = cfg.CONF
SLOT_ERROR = _("Unable to proceed without a slot number")


@logcall
def create_virtual_ethernet_adapter(oper, lpar, network_vif,
                                    custom_mac_addr_capable,
                                    managed_system_uuid,
                                    precreated_port,
                                    vswitch_map=None):
    """
    Creates a VEA on the LPAR (VIOS VEA creation is not handled here).

    :param oper: The K2 HMC operator to execute REST API calls against.
    :param lpar: LPAREntryWrapper for the LPAR to create the VEA on.
    :param network_vif: A dict containing vif data with the following keys:
                        address, meta (for VLAN), network (a dict with key id).
    :param custom_mac_addr_capable: Boolean value that indicates if the MAC
                                    address can be set on the adapter or not.
    :param managed_system_uuid: The unique identifer for the POWER host
                                managed by the HMC.
    :param precreated_port: Boolean to indicate if the neutron port for this
                            vnic was precreated.
    :param vswitch_map: Dictionary mapping vswitch id to its corresponding K2
                        entry.  If None or empty dict passed in, it will be
                        built and then returned at the end so the invoker can
                        use it to cache the response, negating the need to
                        recalculate it on the next call down.
    :return: K2 response object
    :return: vswitch_map used during operation
    """
    if not network_vif:
        LOG.info(_('No vif passed in'))
        return

    # If vswitch_map is None or empty, build it by querying K2
    if not vswitch_map:
        LOG.debug('Getting vswitch map')
        vswitch_map = utils.find_vswitch_map(oper, managed_system_uuid)

    # Find the slot for this network_vif.  We must always get it from the vif
    # if it's there, despite use_slot_from_mac, because that is what will be
    # passed to AE for activation.
    mac = network_vif['address']
    slot_id = network_vif.get('slot_id')
    if slot_id:
        # If slot_id was set because we're using a precreated neutron port,
        # we can't allow that on P6 because we won't be able to honor the mac
        # address set in there.  If we don't have a precreated port, that just
        # means we're here for attach vnic, in which case we always calculate
        # the slot_id to maintain order.
        if precreated_port and not custom_mac_addr_capable:
            raise IBMPowerMACNotSupported(name=lpar.get_lpar_name(), mac=mac)
    else:
        slot_id = int(mac[-2:], 16)

    # If we don't have a slot_id at this point, it's likely a coding error.
    if not slot_id:
        raise Exception(SLOT_ERROR)

    LOG.debug('mac = %s, slot_id = %s' % (mac, slot_id))

    # The VLAN should be in the network VIF
    meta = network_vif['meta']
    pvid = meta['vlan']

    # To find the SEA, we must query for the network mapping.  At this point,
    # a network mapping must exist.  It should be tied to a SEA (any that
    # are not-applicable would have failed before here).  The SEA has the
    # vswitch name
    net_fact = network_dom.NetworkAssociationFactory.get_factory()
    net_assn = net_fact.find_network_assoc(
        context=_get_ctx(),
        host_name=_get_host(),
        neutron_net_id=network_vif.get('network').get('id'))
    vswitch_name = net_assn.sea.primary_vea.vswitch_name

    # Find the vswitch for this vif
    vswitch_id = utils.find_vswitch_id_for_name(vswitch_name, vswitch_map)
    vswitch_uri = utils.find_vswitch_uri_for_id(vswitch_id, vswitch_map)

    # The order of the following is determined by the XSD schema from the
    # K2 SDK that is embedded in each HMC build.  To see the ordering:
    #
    # 1) Get the SDK from your HMC install:
    #    https://hostname:12443/rest/api/web/sdk
    # 2) Unzip the SDK and get this file:
    #    lib/rest/pmc.schema.uom-7.7.8.0.jar
    # 3) Look inside for the schema files, and follow the order of
    #    attributes defined for the object you're working with.

    vswitch_child = K2Element('link', attrib={'href': vswitch_uri,
                                              'rel': 'related'})

    # Build the attributes of the new VEA.
    attrs = list()
    attrs.append(K2Element('VirtualSlotNumber', text=str(slot_id)))

    if custom_mac_addr_capable:
        attrs.extend([K2Element('AllowedOperatingSystemMACAddresses',
                                text='ALL'),
                      K2Element('MACAddress', text=mac)
                      ])
    attrs.extend([K2Element('PortVLANID', text=str(pvid)),
                  K2Element('QualityOfServicePriorityEnabled', text='false'),
                  K2Element('TaggedVLANSupported', text='false'),
                  K2Element('AssociatedVirtualSwitch',
                            children=[vswitch_child]),
                  K2Element('VirtualSwitchID', text=str(vswitch_id))
                  ])

    # Create the K2 object representing the VEA we want to add
    LOG.debug('Creating the K2 object representing the '
              'ClientNetworkAdapter (VEA) to add.  attrs=%s' % attrs)
    vea = K2Element('ClientNetworkAdapter', attrib=k2const.DEFAULT_SCHEMA_ATTR,
                    children=attrs)

    # Invoke the K2 REST API to create the VEA
    return oper.create(vea,
                       'LogicalPartition',
                       lpar.get_uuid(),
                       'ClientNetworkAdapter'), vswitch_map


@logcall
def delete_virtual_ethernet_adapter(oper, lpar_entry, network_vif):
    """
    Deletes a VEA on the LPAR (VIOS VEA deletion is not handled here).

    :param oper: The K2 HMC operator to execute REST API calls against.
    :param lpar: LPAREntryWrapper to delete the VEA from.
    :param network_vif: A dict containing vif data with the following keys:
                        address, meta (for VLAN), network (a dict with key id).
    :return: K2 response object
    """
    if not network_vif:
        return

    # Get the mac address.  This helps us find the right CNA
    lpar = LPAREntryWrapper(lpar_entry)
    mac = network_vif['address']
    cna_uuid = None

    # Loop through all CNAs in the lpar K2 entry
    for link in lpar_entry.element.findall('./ClientNetworkAdapters/link'):
        # Parse out the URI
        href = link.getattrib()['href']
        # Read it in
        cna = oper.readbyhref(href=href)
        # Get the mac from the response
        cna_mac = cna.entry.element.findtext('MACAddress')
        # Compare to what our vif addresses.  Note, K2 response does not have
        # colons, so strip those and convert to lowercase for comparison.
        if cna_mac.lower() == mac.replace(':', '').lower():
            # This was a match.  Remember the CNA uuid
            cna_uuid = href.split('/')[-1]
            break

    # Be sure we found the CNA in question
    if not cna_uuid:
        return

    # Invoke the K2 REST API to delete the VEA
    return oper.delete('LogicalPartition',
                       lpar.get_uuid(),
                       'ClientNetworkAdapter',
                       cna_uuid)


def _get_host():
    """
    Returns the host that this process is running on.
    """
    return CONF.host


def _get_ctx():
    """
    Returns the context that should be used for database calls.
    """
    return ctx.get_admin_context()


def update_mac_addresses(context, instance, respList, nw_info, deploy):
    """
    Method used to update neutron with the MAC Addresses provided by
    the hypervisor used for the cases when host does not support the
    capability to assign the MAC Addresses.
    :param context invocation context
    :param instance The instance being modified
    :param respList The list of k2 responses from the adapters creation
    :param nw_info The network info list
    """
    net_api = API()

    # If we're here for deploy, we must deallocate all at once.
    if deploy:
        LOG.info(_("Deallocating existing MAC addresses"))
        API.deallocate_for_instance(net_api, context, instance)

    if not respList:
        LOG.info(_('No virtual Ethernet adapters found'))
        return
    new_nw_info = []
    for i in range(len(respList)):
        # get the k2resp and the associated VIF
        k2resp = respList[i]
        vif = nw_info[i]
        # create a requested_networks list of a single network
        network_id = vif['network']['id']
        fixed_ips = vif.fixed_ips()
        if fixed_ips:
            rn = [[network_id, fixed_ips[0]['address'], None, None]]
        else:
            rn = [[network_id, None, None, None]]

        rn = objects.NetworkRequestList(
            objects=[objects.NetworkRequest.from_tuple(t)
                     for t in rn])

        if k2resp and k2resp.entry:
            # First, deallocate the old neutron port if this is attach vnic.
            # If we're here for deploy, all were deallocated up front.
            if not deploy:
                LOG.info(_('Deallocating neutron port %s') % vif.get('id'))
                API.deallocate_port_for_instance(net_api, context, instance,
                                                 vif.get('id'))
            entry = VeaEntryWrapper(k2resp.entry)
            mac_address = entry.get_mac_address()
            LOG.info(_('Allocating system provided MAC address=%s') %
                     six.text_type(mac_address))
            macset = set([mac_address])

            # Now allocate the new port.
            new_nw_info.extend(API.allocate_for_instance(net_api,
                                                         context,
                                                         instance,
                                                         requested_networks=rn,
                                                         macs=macset))
            mac_addresses = []
            for new_vif in new_nw_info:
                mac_addresses.append(new_vif['address'])
                LOG.debug('MAC Addresses in updated network info=%s' %
                          str(mac_addresses))
        else:
            LOG.error(_('Unexpected K2 response, not a feed or an entry'))
            break

    # Getting the network_info here with the full set of networks will cause
    # the instance's info cache to get updated with all the correct networks.
    LOG.debug('network_info allocated=%s' % str(new_nw_info))
    net_info = network_model.NetworkInfo.hydrate(new_nw_info)

    # We want to update the nw_info that was passed in to contain all of our
    # new net infos
    for i in range(0, len(net_info)):
        nw_info[i] = new_nw_info[i]

    # Update the instance cache
    base_net_api.update_instance_cache_with_nw_info(net_api,
                                                    context,
                                                    instance,
                                                    nw_info)


class VeaEntryWrapper(EntryWrapper):
    """
    Provides a wrapper for K2 Entry for a Client Network Adapter
    """
    @logcall
    def get_mac_address(self):
        """
        Method to get the MAC Address for this Client Network Adapter
        from the K2 entry
        :return: A Hex String of the MAC Address
        """
        address = self.get_parm_value(MAC_ADDRESS)
        if address:
            address = address.strip()
            if len(address) == 12:
                # insert a colon between every two hex digits
                mac_address = ':'.join(['%02s' % address[i:i + 2]
                                        for i in range(0, len(address), 2)])
        else:
            LOG.error(_('MAC address is not valid:%s') %
                      six.text_type(address))
            mac_address = None
        LOG.debug('MACAddress=%s' % mac_address)
        return mac_address
