# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import copy
import pprint
import six
import logging as python_logging
from eventlet import greenthread
from nova import block_device
from nova.objects import block_device as blk_dev_obj
from nova import context
from nova import conductor
from nova import network
from nova import exception as nova_exception
from nova import rpc
from nova.compute import power_state
from nova.compute import vm_states
from nova.context import get_admin_context
from nova.virt import hardware
from oslo_utils import excutils
from powervc_nova.virt.ibmpowervm.common.constants import IBMI_OS_TYPE,\
    DEFAULT_PROC_POOL, MIGRATING_STATES
from oslo_log import log as logging
from oslo_concurrency.lockutils import synchronized

from oslo_config import cfg
import xml.etree.ElementTree as ElementTree
from xml.etree.ElementTree import Element
from powervc_oslo.k2.k2operator import K2Element
from powervc_oslo.k2.k2operator import K2Error
import base64

from powervc_nova import logcall
from powervc_nova.storage.storage_mapping_api import StorageMapping
from powervc_nova.virt.ibmpowervm.common import exception as common_exception
from powervc_nova.virt.ibmpowervm.common import power_spec_utils
from powervc_nova.virt.ibmpowervm.common import volume_utils, migrate_utils
from powervc_nova.virt.ibmpowervm.common.constants \
    import CONNECTION_TYPE_NPIV, CONNECTION_TYPE_SSP, CUSTOM_MAC_ADDR_CAPABLE,\
    CONNECTION_TYPE_VSCSI, CHK_BOOT, CHK_ALL, PVM_SIMP_REMOTE_RESTART_CAPABLE,\
    ACTIVE_MEMORY_EXPANSION_CAPABLE
from powervc_nova.virt.ibmpowervm.common.constants import IBMI_OS_DISTRO
from powervc_nova.virt.ibmpowervm.hmc import exception
from powervc_nova.virt.ibmpowervm.hmc.instance_type import InstanceType
from powervc_nova.virt.ibmpowervm.hmc.k2oprguard import K2OperGuard
from powervc_nova.virt.ibmpowervm.hmc.k2 import contains_msgs_in_k2_err
import powervc_nova.virt.ibmpowervm.hmc.k2.constants as k2const
from powervc_nova.virt.ibmpowervm.hmc.k2.jobentry_wrapper \
    import JobEntryWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2.cna_wrapper \
    import ClientNetworkAdapterWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2.entry_wrapper \
    import hmc_sharing_mode_to_nova
from powervc_nova.virt.ibmpowervm.hmc.k2.fc_map_processor \
    import FCMapProcessor
from powervc_nova.virt.ibmpowervm.hmc.k2.lparentry_wrapper \
    import LPAREntryWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2.lparentry_wrapper \
    import PROC_CONFIG
from powervc_nova.virt.ibmpowervm.hmc.k2.lparentry_wrapper \
    import SHARED_PROC_CONFIG
from powervc_nova.virt.ibmpowervm.hmc.k2.lparentry_wrapper \
    import PARTITION_IO
from powervc_nova.virt.ibmpowervm.hmc.k2.lparentry_wrapper \
    import IPL_SOURCE
from powervc_nova.virt.ibmpowervm.hmc.k2.lparentry_wrapper \
    import KEY_LOCK
from powervc_nova.virt.ibmpowervm.hmc.k2.profentry_wrapper \
    import LPARProfileEntryWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2.scsi_map_processor \
    import SCSIMapProcessor
from powervc_nova.virt.ibmpowervm.hmc.k2.viosentry_wrapper \
    import VIOSEntryWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2 import storage_map_processor
from powervc_nova.virt.ibmpowervm.hmc.k2 import vea_wrapper
from powervc_nova.virt.ibmpowervm.hmc import exception as hmc_ex

from powervc_nova.virt.ibmpowervm.hmc.virtualoptical import VirtualOptical
from powervc_nova.volume import cinder as volume

from powervc_nova import _


LOG = logging.getLogger(__name__)

attrib = {'schemaVersion': 'V1_0'}

CONF = cfg.CONF
CONF.import_opt('ibmpowervm_job_request_timeout',
                'powervc_nova.virt.ibmpowervm.hmc.k2.jobentry_wrapper')
CONF.import_opt('ibmpowervm_rmc_active_timeout',
                'powervc_nova.virt.ibmpowervm.common.lpar_state_checker')
POST_RETRIES = CONF.ibmpowervm_update_collision_retries
CONNECTOR_HOST_LEN = 55


@logcall
def get_info(k2entry):

    wrapper = LPAREntryWrapper(k2entry)

    current_proc_mode = wrapper.get_current_proc_mode()
    if current_proc_mode == 'true':
        isSharedMode = False
    else:
        isSharedMode = True

    cpus = \
        wrapper.get_current_vcpus() if isSharedMode \
        else \
        wrapper.get_current_procs()
    instance_migrating = wrapper.get_partition_state() in MIGRATING_STATES
    srr_state = wrapper.get_srr_state()

    info = hardware.InstanceInfo(
        id=wrapper.get_lpar_name(),
        state=wrapper.get_lpar_state(),
        mem_kb=wrapper.get_current_mem(),
        max_mem_kb=wrapper.get_current_max_mem(),
        num_cpu=cpus)
    info.instance_migrating = instance_migrating
    info.srr_state = srr_state

    if LOG.logger.isEnabledFor(python_logging.DEBUG):
            LOG.debug('lpar info=%s' %
                      pprint.PrettyPrinter(indent=1).pformat(info.__dict__))

    return info


@logcall
def get_power_specs(k2entry):

    wrapper = LPAREntryWrapper(k2entry)

    current_proc_mode = wrapper.get_current_proc_mode()
    if current_proc_mode == 'true':
        isSharedMode = False
    else:
        isSharedMode = True

    isDedicatedMode = not isSharedMode

    power_specs = {}

    power_specs['vm_id'] = wrapper.get_lpar_id()
    power_specs['vm_name'] = wrapper.get_lpar_name()
    power_specs['vm_uuid'] = wrapper.get_uuid()
    power_specs['power_state'] = wrapper.get_lpar_state()

    power_specs['vcpus'] = \
        wrapper.get_current_vcpus() if isSharedMode \
        else \
        wrapper.get_current_procs()

    power_specs['proc_units'] = \
        wrapper.get_current_proc_units() if isSharedMode \
        else \
        wrapper.get_current_procs()

    power_specs['max_proc_units'] = \
        wrapper.get_current_max_proc_units() if isSharedMode \
        else \
        wrapper.get_current_max_procs()

    power_specs['max_vcpus'] =  \
        wrapper.get_current_max_vcpus() if isSharedMode \
        else \
        wrapper.get_current_max_procs()

    power_specs['min_proc_units'] = \
        wrapper.get_current_min_proc_units() if isSharedMode \
        else \
        wrapper.get_current_min_procs()

    power_specs['min_vcpus'] = \
        wrapper.get_current_min_vcpus() if isSharedMode \
        else \
        wrapper.get_current_min_procs()

    power_specs['pending_proc_units'] = \
        wrapper.get_desired_proc_units() if isSharedMode \
        else \
        wrapper.get_desired_procs()

    power_specs['pending_vcpus'] = \
        wrapper.get_desired_vcpus() if isSharedMode \
        else \
        wrapper.get_desired_procs()

    power_specs['pending_max_proc_units'] = \
        wrapper.get_max_proc_units() if isSharedMode \
        else \
        wrapper.get_max_procs()

    power_specs['pending_max_vcpus'] = \
        wrapper.get_max_vcpus() if isSharedMode \
        else \
        wrapper.get_max_procs()

    power_specs['pending_min_vcpus'] = \
        wrapper.get_min_vcpus() if isSharedMode \
        else \
        wrapper.get_min_procs()

    power_specs['pending_min_proc_units'] = \
        wrapper.get_min_proc_units() if isSharedMode \
        else \
        wrapper.get_min_procs()

    power_specs['run_proc_units'] = \
        wrapper.get_current_proc_units() if isSharedMode \
        else wrapper.get_run_procs()

    power_specs['run_vcpus'] = \
        wrapper.get_run_vcpus() if isSharedMode \
        else wrapper.get_run_procs()

    power_specs['vcpu_mode'] = 'shared' if isSharedMode else 'dedicated'

    power_specs['dedicated_proc'] = current_proc_mode

    if isDedicatedMode:
        power_specs['dedicated_sharing_mode'] = \
            wrapper.get_current_sharing_mode()
        power_specs['uncapped'] = 'false'
        power_specs['shared_weight'] = '0'

    power_specs['max_memory_mb'] = wrapper.get_current_max_mem()
    power_specs['memory_mb'] = wrapper.get_current_mem()

    power_specs['min_memory_mb'] = wrapper.get_current_min_mem()
    power_specs['pending_max_memory_mb'] = wrapper.get_max_mem()
    power_specs['pending_memory_mb'] = wrapper.get_desired_mem()
    power_specs['pending_min_memory_mb'] = wrapper.get_min_mem()
    power_specs['run_memory_mb'] = wrapper.get_run_mem()
    power_specs['ame_expansion_factor'] = wrapper.get_exp_factor()

    shr_mem_enabled = wrapper.get_current_mem_share_enabled()
    # If the shr_mem_enabled is None, don't return it
    if shr_mem_enabled is True:
        power_specs['memory_mode'] = 'shared'
    elif shr_mem_enabled is False:
        power_specs['memory_mode'] = 'dedicated'

    power_specs['rmc_state'] = wrapper.get_rmc_state()

    if isSharedMode:
        power_specs['shared_proc_pool_id'] = wrapper.get_shared_proc_pool_id()
        power_specs['dedicated_sharing_mode'] = ''
        sharing_mode = wrapper.get_current_sharing_mode()
        if sharing_mode is 'uncapped':
            power_specs['uncapped'] = 'true'
            power_specs['shared_weight'] = \
                wrapper.get_curr_uncapped_weight()
        elif sharing_mode is 'capped':
            power_specs['uncapped'] = 'false'
            power_specs['shared_weight'] = '0'

    power_specs['avail_priority'] = wrapper.get_avail_priority()

    # Get processor compatibility modes
    curr_mode, pend_mode = wrapper.get_proc_compat_modes()

    power_specs['processor_compatibility'] = \
        power_specs['current_compatibility_mode'] = curr_mode
    power_specs['desired_compatibility_mode'] = pend_mode

    # Get operating system and version on the virtual machine
    power_specs['operating_system'] = wrapper.get_operating_system()
    power_specs['srr_capability'] = wrapper.get_srr_capability()
    power_specs['srr_state'] = wrapper.get_srr_state()

    if LOG.logger.isEnabledFor(python_logging.DEBUG):
            LOG.debug('lpar power specs=%s' %
                      pprint.PrettyPrinter(indent=1).pformat(power_specs))

    return power_specs


def nova_sharing_mode_to_hmc(sharing_mode):
    for key, value in hmc_sharing_mode_to_nova.items():
        if value == sharing_mode:
            return key


def build_lpar_template(host_stats, instance, instance_type, proc_pools):

    partitionname = instance['name']
    min_mem = instance_type.extra_specs['powervm:min_mem']
    mem = str(instance_type.instance_type['memory_mb'])
    max_mem = instance_type.extra_specs['powervm:max_mem']

    min_proc = instance_type.extra_specs['powervm:min_vcpu']
    proc = str(instance_type.instance_type['vcpus'])
    max_proc = instance_type.extra_specs['powervm:max_vcpu']
    dedicated_proc = instance_type.extra_specs['powervm:dedicated_proc']

    avail_priority = instance_type.extra_specs['powervm:availability_priority']
    proc_compat = instance_type.extra_specs['powervm:processor_compatibility']
    srr_cap = instance_type.extra_specs['powervm:srr_capability']
    ame_expansion_factor = instance_type.extra_specs[
        'powervm:ame_expansion_factor']
    # Format PowerVC proc compat mode string to K2 string
    proc_compat = proc_compat.replace('+', '_Plus')

    # Dedicated Processor Configuration
    if dedicated_proc == 'true':
        sharing_mode = nova_sharing_mode_to_hmc(
            instance_type.extra_specs['powervm:dedicated_sharing_mode'])

        children = [K2Element('DesiredProcessors', text=proc),
                    K2Element('MaximumProcessors', text=max_proc),
                    K2Element('MinimumProcessors', text=min_proc)]
        proc_cfg_children = [K2Element('DedicatedProcessorConfiguration',
                                       attrib=attrib, children=children),
                             K2Element('HasDedicatedProcessors', text='true'),
                             K2Element('SharingMode', text=sharing_mode)]
    # Shared Processor Configuration
    else:
        min_proc_unit = instance_type.extra_specs['powervm:min_proc_units']
        proc_unit = instance_type.extra_specs['powervm:proc_units']
        max_proc_unit = instance_type.extra_specs['powervm:max_proc_units']
        uncapped = instance_type.extra_specs['powervm:uncapped']
        proc_pool_name = instance_type.extra_specs.get(
            'powervm:shared_proc_pool_name')
        proc_pool_name = proc_pool_name if proc_pool_name else \
            DEFAULT_PROC_POOL
        # extract the pool id from the name
        proc_pool_info = proc_pools.get(proc_pool_name)
        proc_pool_id = 0
        if proc_pool_info is not None:
            proc_pool_id = proc_pool_info.get('id', 0)

        LOG.info(_('Shared processor pool ID: %d. Name: %s') %
                 (proc_pool_id, proc_pool_name))
        children = [K2Element('DesiredProcessingUnits', text=proc_unit),
                    K2Element('DesiredVirtualProcessors', text=proc),
                    K2Element('MaximumProcessingUnits', text=max_proc_unit),
                    K2Element('MaximumVirtualProcessors', text=max_proc),
                    K2Element('MinimumProcessingUnits', text=min_proc_unit),
                    K2Element('MinimumVirtualProcessors', text=min_proc),
                    K2Element('SharedProcessorPoolID',
                              text=six.text_type(proc_pool_id))]

        # uncapped sharing mode
        if uncapped == 'true':
            uncap_weight = instance_type.extra_specs['powervm:shared_weight']
            children.append(K2Element('UncappedWeight', text=uncap_weight))
            sharing_mode = 'uncapped'
        # capped sharing mode
        else:
            sharing_mode = 'capped'

        proc_cfg_children = [K2Element('HasDedicatedProcessors', text='false'),
                             K2Element('SharedProcessorConfiguration',
                                       attrib=attrib, children=children),
                             K2Element('SharingMode', text=sharing_mode)]
    partitionio_cfg_child = [K2Element('MaximumVirtualIOSlots', text='64')]
    partition_type = 'AIX/Linux'
    addIORestricted = False

    capabilities = host_stats.get('capabilities')
    if IBMI_OS_DISTRO == instance['system_metadata']['image_os_distro']:
        children = [K2Element('AlternateLoadSource', text='NONE'),
                    K2Element('Console', text='HMC'),
                    K2Element('LoadSource', text='0')]
        io_cfg_children = [K2Element('MaximumVirtualIOSlots', text='64'),
                           K2Element('TaggedIO', attrib=attrib,
                                     children=children)]
        partitionio_cfg_child = io_cfg_children
        partition_type = 'OS400'
        # To determine whether a host has os400_restrcited_io_mode_capable
        restrcited_io_mode_capable = (capabilities.
                                      get('ibmi_restrictedio_capable'))
        if restrcited_io_mode_capable:
            addIORestricted = True
        # To determine whether the host support SRR capability

    partition_mem_config = []
    # PowerVC does not track a boolean value for AME capability of an LPAR.
    # The value is instead inferred from the expansion factor value here and
    # used by K2.
    ame_cap = None
    if ame_expansion_factor is not None:
        ameef_float = round(float(ame_expansion_factor), 2)
        if ameef_float == 0:
            ame_cap = 'false'
        elif ameef_float >= 1 and ameef_float <= 10:
            ame_cap = 'true'
    # NOTE (mdrabe): perhaps just always include ActiveMemoryExpansionEnabled
    # and ExpansionFactor elements.
    # Ordering of K2 Elements is important.
    host_ame_cap = capabilities.get(ACTIVE_MEMORY_EXPANSION_CAPABLE)
    if (host_ame_cap and ame_cap is not None):
        partition_mem_config.extend([K2Element('ActiveMemoryExpansionEnabled',
                                               text=ame_cap),
                                     K2Element('DesiredMemory', text=mem),
                                     K2Element('ExpansionFactor',
                                               text=ame_expansion_factor)])
    else:
        partition_mem_config.extend([K2Element('DesiredMemory', text=mem)])
    partition_mem_config.extend([K2Element('MaximumMemory', text=max_mem),
                                 K2Element('MinimumMemory', text=min_mem)])

    attrs = [K2Element('AvailabilityPriority', text=avail_priority),
             K2Element('CurrentProfileSync', text='On'),
             K2Element('PartitionIOConfiguration', attrib=attrib,
                       children=partitionio_cfg_child),
             K2Element('PartitionMemoryConfiguration', attrib=attrib,
                       children=partition_mem_config),
             K2Element('PartitionName', text=partitionname),
             K2Element('PartitionProcessorConfiguration',
                       attrib=attrib,
                       children=proc_cfg_children),
             K2Element('PartitionType', text=partition_type),
             K2Element('PendingProcessorCompatibilityMode',
                       text=proc_compat)]

    simp_rr_capable = (capabilities.get(PVM_SIMP_REMOTE_RESTART_CAPABLE))
    if simp_rr_capable and srr_cap is not None:
        attrs.append(K2Element('SimplifiedRemoteRestartCapable',
                               attrib={'ksv': 'V1_2_0'},
                               text=srr_cap.lower()))
    if addIORestricted:
        attrs.append(K2Element('IsRestrictedIOPartition', text='true'))

    lpar = K2Element('LogicalPartition', attrib=attrib, children=attrs)

    return lpar


@logcall
def update_load_source(instance, host, lpar_uuid,
                       load_source, alternated_loadsource):
    """
    After adapter created,update load source slot number for IBM i vm
    :param host: the host
    :param lpar_uuid: uuid of the instance
    :param load_source: the load source slot number
    :param alternated_loadsource: the alternated load source slot number if
    there is or NONE
    """
    def _update_k2_element(k2resp, element_name, element_find, new_element):
        index = -1
        # Find PartitionIOConfiguration element's index
        for i in range(len(k2resp.entry.element)):
            if (k2resp.entry.element[i].gettag() ==
                    element_name):
                index = i
                break

        curr_element_children = k2resp.entry.element.find(element_find)
        k2resp.entry.element.remove(curr_element_children)
        k2resp.entry.element.insert(index, new_element)

    def _update_load_source(lpar_uuid, load_source, alternated_loadsource,
                            last_retry=False):
        # Read in existing LPAR to be used for the update and
        # It is important that this is the most
        # up to date data for the update command.
        with K2OperGuard(host=host) as oper:
            try:
                k2resp = oper.read(k2const.MGT_SYS, host.uuid, k2const.LPAR,
                                   lpar_uuid)
            except Exception as e:
                LOG.error(_('Cannot read virtual machine: %s') % name)
                LOG.exception(_('Error reading virtual machine %(name)s.'
                                ' Exception: %(exc)s') %
                              {'name': name, 'exc': six.text_type(e)})
                raise
            etag = k2resp.headers['etag']

        children = [K2Element('AlternateLoadSource',
                              text=alternated_loadsource),
                    K2Element('Console', text='HMC'),
                    K2Element('LoadSource', text=load_source)]
        new_taggedIO = [K2Element('TaggedIO', attrib=attrib,
                                  children=children)]

        # Find PartitionIOConfiguration element's index
        new_partitionIO = K2Element('PartitionIOConfiguration', attrib=attrib,
                                    children=new_taggedIO)
        _update_k2_element(k2resp, 'PartitionIOConfiguration',
                           PARTITION_IO, new_partitionIO)

        # Find designatedIPLSource
        IPLSource = K2Element('DesignatedIPLSource', text='b')
        _update_k2_element(k2resp, 'DesignatedIPLSource',
                           IPL_SOURCE, IPLSource)

        # Find KeylockPosition
        Keylock = K2Element('KeylockPosition', text='normal')
        _update_k2_element(k2resp, 'KeylockPosition', KEY_LOCK, Keylock)

        with K2OperGuard(host=host) as oper:
            try:
                oper.update(k2resp.entry.element, etag, k2const.MGT_SYS,
                            host.uuid, k2const.LPAR, lpar_uuid, xag=[None])
            except K2Error as e:
                if (hasattr(e, 'k2response') and e.k2response is not None and
                        e.k2response.status ==
                        k2const.HTTP_STATUS_ETAG_MISMATCH):
                    if last_retry:
                        # dump all the k2 requests & responses accumulated
                        # since the wrapper would not log single
                        # HTTP_STATUS_ETAG_MISMATCH failure
                        oper.dump_logs()
                    raise exception.IBMPowerVMUpdateFailed(uuid=lpar_uuid)
                raise

    name = instance['name']
    if load_source is None:
        load_source = 'NONE'
    if alternated_loadsource is None:
        alternated_loadsource = 'NONE'

    # Retry the K2 update if it fails based on etag mismatch.
    for retries in xrange(POST_RETRIES + 1):
        try:
            _update_load_source(lpar_uuid, load_source,
                                alternated_loadsource,
                                retries == POST_RETRIES)
            break
        except Exception as e:
            if isinstance(e, exception.IBMPowerVMUpdateFailed):
                LOG.debug('Retry _update_load_source %s' % retries)
                # Retry for POST_RETRIES
                if (retries < POST_RETRIES):
                    continue
            LOG.error(_('Cannot update load source for virtual machine: %s')
                      % name)
            LOG.exception(_('Error update load source for '
                            'virtual machine %(name)s. Exception: %(exc)s') %
                          {'name': name, 'exc': six.text_type(e)})
            raise


@logcall
def spawn(context, virtapi, host, instance, image_meta,
          injected_files, admin_password, network_info,
          block_device_info, validate_volumes=None, attach_volumes=None,
          validate_attachment=None):

    from powervc_nova.virt.ibmpowervm.hmc.orchestrator import\
        DeployOrchestrator
    o = DeployOrchestrator(context, virtapi, host, instance, image_meta,
                           injected_files, admin_password, network_info,
                           block_device_info, validate_volumes,
                           attach_volumes, validate_attachment,
                           flavor=instance.flavor)
    o.start_deploy_simple()


def update_lpar_for_resize(lpar_entry, curr_flavor, instance_type, proc_pools):
    """Takes the current LogicalPartition K2Element, and updates it with
    the new flavor.
    :param lpar_entry: the LogicalPartition K2 entry
    :param curr_flavor: the current flavor extracted from the K2 element
    :param instance_type: the new validated/defaulted instance type obj
    :param proc_pools: the dictionary containing shared processor pools
                       information
    """
    lpar_wrapper = LPAREntryWrapper(lpar_entry)
    extra_specs = instance_type.extra_specs

    # memory
    memory_mb = str(instance_type.instance_type.get('memory_mb'))
    if curr_flavor.get('memory_mb') != memory_mb:
        lpar_wrapper.set_desired_mem(memory_mb)
    # max mem
    max_mem = extra_specs.get('powervm:max_mem')
    if curr_flavor.get('powervm:max_mem') != max_mem:
        lpar_wrapper.set_max_mem(max_mem)
    # min mem
    min_mem = extra_specs.get('powervm:min_mem')
    if curr_flavor.get('powervm:min_mem') != min_mem:
        lpar_wrapper.set_min_mem(min_mem)

    # active memory expansion
    ame_ef = extra_specs.get('powervm:ame_expansion_factor')
    if ame_ef is not None:
        if curr_flavor.get('powervm:ame_expansion_factor') != ame_ef:
            ameef_float = round(float(ame_ef), 2)
            if ameef_float == 0:
                lpar_wrapper.set_ame_enabled('false')
            elif ameef_float >= 1 and ameef_float <= 10:
                lpar_wrapper.set_ame_enabled('true')
            lpar_wrapper.set_exp_factor(ame_ef)

    # availability priority
    avail_priority = extra_specs.get('powervm:availability_priority')
    if curr_flavor.get('powervm:availability_priority') != avail_priority:
        lpar_wrapper.set_avail_priority(avail_priority)
    # processor compatibility
    proc_compat = extra_specs.get('powervm:processor_compatibility')
    if curr_flavor.get('powervm:processor_compatibility')[0] != proc_compat:
        lpar_wrapper.set_proc_compat_mode(proc_compat.replace('+', '_Plus'))

    # simplified remote restart capability
    srr_cap = extra_specs.get('powervm:srr_capability')
    if srr_cap is not None:
        if curr_flavor.get('powervm:srr_capability') != srr_cap:
            lpar_wrapper.set_srr_capability(srr_cap.lower())

    is_dedicated = extra_specs.get('powervm:dedicated_proc')
    vcpus = str(instance_type.instance_type.get('vcpus'))
    max_vcpu = extra_specs.get('powervm:max_vcpu')
    min_vcpu = extra_specs.get('powervm:min_vcpu')
    # If processor mode is not changing, continue with setting individual
    # properties
    if curr_flavor.get('powervm:dedicated_proc') == is_dedicated:
        # vcpus
        if curr_flavor.get('vcpus') != vcpus:
            if is_dedicated == 'true':
                lpar_wrapper.set_desired_procs(vcpus)
            else:
                lpar_wrapper.set_desired_vcpus(vcpus)
        # max vcpus
        if curr_flavor.get('powervm:max_vcpu') != max_vcpu:
            if is_dedicated == 'true':
                lpar_wrapper.set_max_procs(max_vcpu)
            else:
                lpar_wrapper.set_max_vcpus(max_vcpu)
        # min vcpus
        if curr_flavor.get('powervm:min_vcpu') != min_vcpu:
            if is_dedicated == 'true':
                lpar_wrapper.set_min_procs(min_vcpu)
            else:
                lpar_wrapper.set_min_vcpus(min_vcpu)

        # if shared, check for changes to proc units
        if is_dedicated == 'false':
            # shared processor pool id
            curr_shared_proc_pool_id = lpar_wrapper.get_shared_proc_pool_id()
            proc_pool_name = instance_type.extra_specs.\
                get('powervm:shared_proc_pool_name')
            if proc_pool_name is not None:
                # Extract the pool id from the name. If unable to locate the
                # shared processor pool on the managed host we would not reach
                # until here. So, we can certainly get the pool id by
                # pool name.
                proc_pool_info = proc_pools.get(proc_pool_name)
                shared_proc_pool_id = proc_pool_info.get('id')
                if (int(curr_shared_proc_pool_id) != int(shared_proc_pool_id)):
                    new_pool_id = six.text_type(shared_proc_pool_id)
                    lpar_wrapper.set_shared_proc_pool_id(new_pool_id)
            # proc units
            desired_proc_units = extra_specs.get('powervm:proc_units')
            if (float(curr_flavor.get('powervm:proc_units')) !=
                    float(desired_proc_units)):
                lpar_wrapper.set_desired_proc_units(desired_proc_units)
            # max proc units
            max_proc_units = extra_specs.get('powervm:max_proc_units')
            if (float(curr_flavor.get('powervm:max_proc_units')) !=
                    float(max_proc_units)):
                lpar_wrapper.set_max_proc_units(max_proc_units)
            # min proc units
            min_proc_units = extra_specs.get('powervm:min_proc_units')
            if (float(curr_flavor.get('powervm:min_proc_units')) !=
                    float(min_proc_units)):
                lpar_wrapper.set_min_proc_units(min_proc_units)
            # uncapped or capped
            is_uncapped = extra_specs.get('powervm:uncapped')
            uncap_weight = extra_specs.get('powervm:shared_weight')
            if curr_flavor.get('powervm:uncapped') != is_uncapped:
                # Going from capped to uncapped. Append uncapped weight
                if is_uncapped == 'true':
                    sharing_mode = 'uncapped'
                    weight = K2Element('UncappedWeight', text=uncap_weight)
                    shared_cfg = (
                        lpar_entry.element.find(SHARED_PROC_CONFIG))
                    shared_cfg.append(weight)
                # Going from uncapped to capped.
                else:
                    sharing_mode = 'capped'
                lpar_wrapper.set_sharing_mode(sharing_mode)
            # Sharing mode is not changing, and is uncapped. Change uncapped
            # weight if different
            elif (is_uncapped == 'true' and
                  (curr_flavor.get('powervm:shared_weight') != uncap_weight)):
                lpar_wrapper.set_uncapped_weight(uncap_weight)
        # Dedicated
        else:
            # Is dedicated or shared processor mode being requested
            sharing_mode = extra_specs.get('powervm:dedicated_sharing_mode')
            if (curr_flavor.get('powervm:dedicated_sharing_mode') !=
                    sharing_mode):
                sharing_mode = nova_sharing_mode_to_hmc(sharing_mode)
                lpar_wrapper.set_sharing_mode(sharing_mode)
        return

    # This resize is changing the processor mode
    lpar_wrapper.set_proc_mode(is_dedicated)

    # Changing from shared to dedicated processors.
    # Remove SharedProcessorConfiguration, and add
    # DedicatedProcessorConfiguration.
    if is_dedicated == 'true':
        update_to_dedicated_proc(lpar_wrapper, lpar_entry, instance_type)

    # Changing from dedicated to shared processors.
    # Remove DedicatedProcessorConfiguration, and add
    # SharedProcessorConfiguration.
    else:
        update_to_shared_proc(lpar_wrapper, lpar_entry, instance_type,
                              proc_pools)


@logcall
def update_to_dedicated_proc(lpar_wrapper, lpar_entry, instance_type):
        extra_specs = instance_type.extra_specs
        desired_procs = str(instance_type.instance_type.get('vcpus'))
        max_procs = extra_specs.get('powervm:max_vcpu')
        min_procs = extra_specs.get('powervm:min_vcpu')
        sharing_mode = nova_sharing_mode_to_hmc(
            extra_specs.get('powervm:dedicated_sharing_mode'))

        # Create new PartitionProcessorConfiguration
        children = [K2Element('DesiredProcessors', text=desired_procs),
                    K2Element('MaximumProcessors', text=max_procs),
                    K2Element('MinimumProcessors', text=min_procs)]

        proccfg_children = [K2Element('DedicatedProcessorConfiguration',
                                      attrib=attrib, children=children),
                            K2Element('HasDedicatedProcessors', text='true'),
                            K2Element('SharingMode', text=sharing_mode)]

        new_proccfg = K2Element(
            'PartitionProcessorConfiguration', attrib=attrib,
            children=proccfg_children)

        # Find PartitionProcessorConfiguration element's index
        for i in range(len(lpar_entry.element)):
            if (lpar_entry.element[i].gettag() ==
               'PartitionProcessorConfiguration'):
                index = i
                break

        # Remove old PartitionProcessorConfiguration
        curr_proccfg = lpar_entry.element.find(PROC_CONFIG)
        lpar_entry.element.remove(curr_proccfg)
        # Replace with new PartitionProcessorConfiguration
        lpar_entry.element.insert(index, new_proccfg)


@logcall
def update_to_shared_proc(lpar_wrapper, lpar_entry, instance_type, proc_pools):
        extra_specs = instance_type.extra_specs
        desired_vcpus = str(instance_type.instance_type.get('vcpus'))
        max_vcpu = extra_specs.get('powervm:max_vcpu')
        min_vcpu = extra_specs.get('powervm:min_vcpu')
        desired_proc_units = extra_specs.get('powervm:proc_units')
        max_proc_units = extra_specs.get('powervm:max_proc_units')
        min_proc_units = extra_specs.get('powervm:min_proc_units')
        uncap_weight = extra_specs.get('powervm:shared_weight')
        is_uncapped = extra_specs.get('powervm:uncapped')
        proc_pool_name = extra_specs.get('powervm:shared_proc_pool_name')
        if proc_pool_name is not None:
            # Extract the pool id from the name. If unable to locate the
            # shared processor pool on the managed host we would not reach
            # until here. So, we can certainly get the pool id by pool name.
            proc_pool_info = proc_pools.get(proc_pool_name)
            shared_proc_pool_id = proc_pool_info.get('id')
        else:
            # If no pool name is given while changing from dedicated to shared
            # processor pool, use default pool (id=0)
            LOG.warn(_('The shared processor pool name is not provided.'
                       ' Defaulting to shared processor pool DefaultPool.'))
            shared_proc_pool_id = 0

        # Create new SharedProcessorConfiguration
        children = [
            K2Element('DesiredProcessingUnits', text=desired_proc_units),
            K2Element('DesiredVirtualProcessors', text=desired_vcpus),
            K2Element('MaximumProcessingUnits', text=max_proc_units),
            K2Element('MaximumVirtualProcessors', text=max_vcpu),
            K2Element('MinimumProcessingUnits', text=min_proc_units),
            K2Element('MinimumVirtualProcessors', text=min_vcpu),
            K2Element('SharedProcessorPoolID',
                      text=six.text_type(shared_proc_pool_id))]
        if is_uncapped == 'true':
            sharing_mode = 'uncapped'
            children.append(K2Element('UncappedWeight', text=uncap_weight))
        else:
            sharing_mode = 'capped'

        proccfg_children = [K2Element('HasDedicatedProcessors', text='false'),
                            K2Element('SharedProcessorConfiguration',
                                      attrib=attrib, children=children),
                            K2Element('SharingMode', text=sharing_mode)]

        new_proccfg = K2Element('PartitionProcessorConfiguration',
                                attrib=attrib, children=proccfg_children)

        # Find PartitionProcessorConfiguration element's index
        for i in range(len(lpar_entry.element)):
            if (lpar_entry.element[i].gettag() ==
                    'PartitionProcessorConfiguration'):
                index = i
                break

        # Remove old PartitionProcessorConfiguration
        curr_proccfg = lpar_entry.element.find(PROC_CONFIG)
        lpar_entry.element.remove(curr_proccfg)
        # Replace with new PartitionProcessorConfiguration
        lpar_entry.element.insert(index, new_proccfg)


@logcall
def resize(context, host, uuid, instance, instance_type, network_info,
           block_device_info):
    """Resizes the VM to new flavor

    :param host: the host
    :param uuid: uuid of the instance
    :param instance: the current instance
    :param instance_type: new flavor specification
    :param network_info: network specifications of VM
    :param block_device_info: instance block device information
    """
    def _get_lpar():
        # Read in existing LPAR to be used for the update and
        # flavor validation.
        with K2OperGuard(host=host) as oper:
            try:
                k2resp = oper.read(k2const.MGT_SYS, host.uuid, k2const.LPAR,
                                   uuid, age=0)
            except Exception as e:
                LOG.error(_('Cannot read virtual machine: %s') % name)
                LOG.exception(_('Error reading virtual machine %(name)s.'
                                ' Exception: %(exc)s') %
                              {'name': name, 'exc': six.text_type(e)})
                raise

        return k2resp

    def _resize_lpar(k2resp, last_retry=False):

        etag = k2resp.headers['etag']

        # Extract the nova flavor from the lpar
        proc_pools = host.get_shared_proc_pools()
        proc_pools_id_names = host.get_shared_proc_pools_id_name()
        curr_flavor = extract_flavor_from_lpar(k2resp.entry,
                                               proc_pools_id_names)
        LOG.info(_('Current properties of virtual machine %(name)s are'
                   ' %(flavor)r') % {'name': name, 'flavor': curr_flavor})
        LOG.info(_('Resizing virtual machine %(name)s to new flavor'
                   ' %(flavor)r') % {'name': name, 'flavor': instance_type})

        # Check AME against the OS
        sys_meta = instance.get('system_metadata', {})
        inst_distro = sys_meta.get('image_os_distro', None)
        new_flav = power_spec_utils._check_os_distro_is_aix_for_ame_resize(
            context, instance_type, inst_distro, instance['name'],
            curr_flavor.get('powervm:ame_expansion_factor'))
        if instance.new_flavor.flavorid != new_flav.flavorid:
            instance.new_flavor = new_flav
            instance.save()

        # Create a driver instance type object from the db instance type
        instance_type_obj = InstanceType(instance, new_flav)

        # Validate the requested instance_type. It's extra_specs will be
        # filled out with new requested properties or current properties
        # if the property is not changing.

        update_needed = (instance_type_obj.validate(
                         host, len(network_info), curr_flavor))

        if not update_needed:
            LOG.info(_('No changes to memory or CPU related attributes have'
                     ' been requested for virtual machine %s.') % name)
            return

        # update Logical partitions sub elements
        update_lpar_for_resize(k2resp.entry, curr_flavor, instance_type_obj,
                               proc_pools)
        exp_k2_rem_mem_errs = ['HSCL2932', 'HSCL2935']
        ex_k2_rr_vopt_errs = ['HSCLA9D6', 'HSCLA9D7']
        ex_k2_shr_proc_pool_errs = ['HSCL3601']
        with K2OperGuard(host=host) as oper:
            try:
                # in the k2owrappers, we override the LogicalPartition GET
                # with xag=[None]. We need to do the same for the POST here.
                # pass in expected errors HSCL2932 and HSCL2935
                oper.update(k2resp.entry.element, etag, 'ManagedSystem',
                            rootId=host.uuid, childType='LogicalPartition',
                            childId=uuid, xag=[None],
                            exp_errs=exp_k2_rem_mem_errs)
            except K2Error as e:
                if (hasattr(e, 'k2response') and e.k2response is not None and
                        e.k2response.status ==
                        k2const.HTTP_STATUS_ETAG_MISMATCH):
                    if last_retry:
                        # dump all the k2 requests & responses accumulated
                        # since the wrapper would not log single
                        # HTTP_STATUS_ETAG_MISMATCH failure
                        oper.dump_logs()
                    raise exception.IBMPowerVMUpdateFailed(uuid=uuid)
                elif contains_msgs_in_k2_err(e, exp_k2_rem_mem_errs):
                    LOG.error(_('Error resizing virtual machine %(name)s.'
                                ' Exception: %(exc)s') %
                              {'name': name, 'exc': six.text_type(e)})
                    ex_args = {'instance_name': name,
                               'file_name': (
                                   '/var/log/nova/nova-compute-%s.log' %
                                   host.host_id)}
                    raise exception.IBMPowerVMOSRmMemFailed(**ex_args)
                elif contains_msgs_in_k2_err(e, ex_k2_shr_proc_pool_errs):
                    ex_args = {'pool_name':
                               instance_type['extra_specs'].
                               get('powervm:shared_proc_pool_name')}
                    LOG.warning(_("The shared processor pool %(pool_name)s is "
                                  "not configured. Configure the shared "
                                  "processor pool and retry the operation.") %
                                ex_args)
                    raise exception.\
                        IBMPowerVMShrProcPoolNotConfigured(**ex_args)
                elif contains_msgs_in_k2_err(e, ex_k2_rr_vopt_errs):
                    # If the simplified remote restart capability is not
                    # updated to the system
                    LOG.warning(_("The configuration change was successful, "
                                  "however the simplified remote restart"
                                  " capability was not applied. Please run"
                                  " refdev command after removing the"
                                  " attached vopt device."))
                else:
                    raise

    name = instance['name']
    k2resp = _get_lpar()
    lpar_wrapper = LPAREntryWrapper(k2resp.entry)
    os_type = lpar_wrapper.get_lpar_type()
    instance['os_type'] = os_type

    new_size = instance_type['root_gb']
    size = instance['root_gb']
    if new_size != size:
        # Resize disk
        # For IBM i, resize disk is not supported, raise exception for IBM i
        if IBMI_OS_TYPE == os_type:
            LOG.error(_('Cannot resize virtual machine: %s') % name)
            LOG.exception(_('Unable to change disk size for virtual machine '
                            '%(name)s because disk size change is not '
                            'supported for IBM i.') %
                          {'name': name})
            ex_args = {'instance_name': instance['name']}
            raise (common_exception.
                   IBMPowerVMIBMiDiskResizeNotSupported(**ex_args))
        else:
            # If disk resize fails, we won't continue on with resize.
            try:
                bdms = blk_dev_obj.BlockDeviceMappingList.get_by_instance_uuid(
                    context, instance['uuid'])
                root_bdm = block_device.get_root_bdm(bdms)
                volume_id = root_bdm['volume_id']
                LOG.info(_('Extending volume %(disk)s of virtual machine '
                           '%(name)s from %(curr_size)dGB to '
                           '%(new_size)dGB') %
                         {'disk': volume_id, 'name': name,
                          'curr_size': size, 'new_size': new_size})
                volume_api = volume.PowerVCVolumeAPI()
                volume_api.extend_volume(context, volume_id, new_size)
            except Exception as e:
                LOG.error(_('Cannot resize virtual machine: %s') % name)
                LOG.exception(_('Error resizing virtual machine %(name)s.'
                                ' Extending the volume failed with the '
                                'following exception: %(exc)s') %
                              {'name': name, 'exc': six.text_type(e)})
                raise

    # Retry the K2 update if it fails based on etag mismatch.
    for retries in xrange(POST_RETRIES + 1):
        try:
            _resize_lpar(k2resp, retries == POST_RETRIES)
            break
        except Exception as e:
            if isinstance(e, exception.IBMPowerVMUpdateFailed):
                LOG.debug('Retry _resize_lpar %s' % retries)
                # Retry for POST_RETRIES
                if (retries < POST_RETRIES):
                    # Read in existing LPAR to be used for the update and
                    # flavor validation. It is important that this is the most
                    # up to date data for the update command.
                    k2resp = _get_lpar()
                    continue
            # If new disk size is different than old disk size and
            # we made it to this point, it means resizing the logical
            # partition failed, but resizing the disk succeeded.
            # Update the DB with the new disk size.
            if new_size != size:
                instance.root_gb = new_size
                instance.save()
            LOG.error(_('Cannot resize virtual machine: %s') % name)
            LOG.exception(_('Error resizing virtual machine %(name)s.'
                            ' Exception: %(exc)s') %
                          {'name': name, 'exc': six.text_type(e)})
            raise


@logcall
def extract_flavor_from_lpar(entry, proc_pools_id_names):
    """
    Extract flavor information needed for resizing an instance from HMC LPAR
    """
    flavor = {}
    lpar_wrapper = LPAREntryWrapper(entry)
    # check whether it's using dedicated or shared proc
    proc_mode = lpar_wrapper.get_proc_mode()
    sharing_mode = lpar_wrapper.get_sharing_mode()
    flavor['powervm:dedicated_proc'] = proc_mode
    if proc_mode == 'true':
        flavor['vcpus'] = lpar_wrapper.get_desired_procs()
        flavor['powervm:max_vcpu'] = lpar_wrapper.get_max_procs()
        flavor['powervm:min_vcpu'] = lpar_wrapper.get_min_procs()
        flavor['powervm:dedicated_sharing_mode'] = sharing_mode
    else:
        flavor['vcpus'] = lpar_wrapper.get_desired_vcpus()
        flavor['powervm:max_vcpu'] = lpar_wrapper.get_max_vcpus()
        flavor['powervm:min_vcpu'] = lpar_wrapper.get_min_vcpus()
        flavor['powervm:proc_units'] = lpar_wrapper.get_desired_proc_units()
        flavor['powervm:max_proc_units'] = lpar_wrapper.get_max_proc_units()
        flavor['powervm:min_proc_units'] = lpar_wrapper.get_min_proc_units()
        proc_pool_id = lpar_wrapper.get_shared_proc_pool_id()
        if proc_pool_id is not None:
            proc_pool_name = proc_pools_id_names.get(proc_pool_id)
            flavor['powervm:shared_proc_pool_name'] = proc_pool_name
            flavor['powervm:shared_proc_pool_id'] = proc_pool_id
        uncapped = 'true'
        if sharing_mode != 'uncapped':
            uncapped = 'false'
        else:
            flavor['powervm:shared_weight'] = (
                lpar_wrapper.get_uncapped_weight())
        flavor['powervm:uncapped'] = uncapped

    flavor['memory_mb'] = lpar_wrapper.get_desired_mem()
    flavor['powervm:min_mem'] = lpar_wrapper.get_min_mem()
    flavor['powervm:max_mem'] = lpar_wrapper.get_max_mem()
    # Tuple with current and pending processor compatibility mode
    flavor['powervm:processor_compatibility'] = (
        lpar_wrapper.get_proc_compat_modes())
    flavor['powervm:availability_priority'] = lpar_wrapper.get_avail_priority()
    flavor['powervm:srr_capability'] = lpar_wrapper.get_srr_capability()
    flavor['powervm:ame_expansion_factor'] = lpar_wrapper.get_exp_factor()
    # Return information about instance state and RMC connectivity
    flavor['state'] = lpar_wrapper.get_lpar_state()
    flavor['has_rmc_connectivity'] = lpar_wrapper.get_rmc_state() == 'active'
    return flavor


def get_vios_lpar_id_from_vadapters(vscsi_resp):
    vios_ids = set()
    for entry in vscsi_resp.feed.entries:
        p = entry.element.findtext('./ServerAdapter/LocalPartitionID')
        if p is not None:
            vios_ids.add(int(p))
    return vios_ids


def _update_onboarding_reasons(self, onboard_data, lpar_id, reason):
    """
    Update the onboarding information based on supported reason from
    either network or storage components.
    """
    if lpar_id in onboard_data:
        current_supported = onboard_data[lpar_id]['support']
        if current_supported['status'] == 'supported':
            current_supported['status'] = 'not_supported'
            current_supported['reasons'] = []
        current_supported['reasons'].append(reason)


@logcall
def destroy(host, lpar_name, lpar_uuid, entry,
            destroy_disks=True):
    """Destroy (shutdown and delete) the specified instance.

    If the instance is not found (for example if networking failed), this
    function should still succeed.  It's probably a good idea to log a
    warning in that case.

    :param host: Host object representing the host where LPAR resides
    :param lpar_name: HMC name of the LPAR
    :param lpar_uuid: HMC UUID of the LPAR
    :param entry: k2 entry for the LPAR
    :param destroy_disks: Indicates if disks should be destroyed. Ignored
           for NPIV disks since they are invisible to VIOS.
           TODO: Honor this boolean. Currently only ephemeral disk is deleted
    """
    def _delete_lpar():
        """
        Delete LPAR from HMC and close vterm if necessary
        """
        for count in range(2):
            # Try to delete an LPAR. If deletion fails because of an opened
            # vterm, try to close the vterm and delete again; otherwise,
            # raise exception.
            try:
                LOG.info(_('Deleting virtual machine. Attempt %s') % count)
                k2resp = oper.delete(k2const.MGT_SYS, host.uuid,
                                     k2const.LPAR, lpar_uuid)
                # Possible status values: 200 (OK), 202 (Accepted),
                # 304 (Document not modified)
                LOG.info(_('Virtual machine delete status: %s') %
                         k2resp.status)
                return
            except Exception as e:
                try_again = False
                if isinstance(e, K2Error):
                    k2e_resp = e.k2response
                    if (k2e_resp is not None and k2e_resp.body is not None and
                            ('HSCL151B' in str(k2e_resp.body) or 'HSCL05E6'
                             in str(k2e_resp.body)) and count == 0):
                        # If the error is about a vterm that is still open,
                        # try to close it.
                        try:
                            LOG.info(_('Closing virtual terminal'))
                            _close_vterm(oper, host.uuid, lpar_uuid)
                            # If we can close vterm, give delete another try
                            try_again = True
                        except Exception as e:
                            # Fall through and raise exception in the following
                            # if-block
                            pass
                        if 'HSCL05E6' in str(k2e_resp.body):
                            # If the VM has not been powered off then we need
                            # to retry
                            power_on_off(oper,
                                         lpar_uuid,
                                         k2const.SUFFIX_PARM_POWER_OFF,
                                         host,
                                         force_immediate=True)
                if not try_again:
                    # WI 47393: If we get HSCL151B again raise a clearer
                    # exception message other than a 500 error.
                    if isinstance(e, K2Error):
                        k2e_resp = e.k2response
                        if (k2e_resp is not None and k2e_resp.body is not None
                           and ('HSCL151B' in str(k2e_resp.body))):
                            raise exception.IBMPowerVMCannotCloseVTerm(
                                name=lpar_name)
                    raise e

    LOG.info(_('Deleting virtual machine %(lpar_name)s. UUID = %(lpar_uuid)s')
             % {'lpar_name': lpar_name, 'lpar_uuid': lpar_uuid})
    if lpar_uuid is None or entry is None:
        entry_details = _entry_to_string(entry)
        LOG.warning(_('Virtual machine cannot be found. Cannot proceed. '
                      'LPAR UUID: %(lpar_uuid)s. Entry: %(entry_details)s.') %
                    {'lpar_uuid': lpar_uuid, 'entry_details': entry_details})
        return
    with K2OperGuard(host=host) as oper:
        try:
            state = get_lpar_state(oper, host, lpar_uuid)
            LOG.info(_('Virtual machine state: %s') % state)
            # Power off LPAR if needed
            if state != power_state.SHUTDOWN:
                power_on_off(oper,
                             lpar_uuid,
                             k2const.SUFFIX_PARM_POWER_OFF,
                             host,
                             force_immediate=True)
            # Issue 2360. Delete LPAR before removing adapter mappings
            # Delete LPAR
            _delete_lpar()
            # Detach all volumes associated with this LPAR
            detach_all_volumes(oper, host, lpar_name, lpar_uuid, entry)
            # Remove instance from UUID cache
            host.pvm_uuid_cache.remove(lpar_name)
            # Invalidate the cache so the next call to
            # conditional_refresh forces a refresh of the data
#            host.lpar_cache.invalidate()
            LOG.info(_('Finished deleting virtual machine %(lpar_name)s.'
                       ' UUID = %(lpar_uuid)s') %
                     {'lpar_name': lpar_name, 'lpar_uuid': lpar_uuid})
        except Exception as e:
            LOG.exception(_('Error deleting virtual machine %(lpar_name)s. '
                            ' Exception:'
                            ' %(exc)s') % {'lpar_name': lpar_name,
                                           'exc': six.text_type(e)})
            raise common_exception.\
                IBMPowerVMLPARInstanceCleanupFailed(instance_name=lpar_name,
                                                    reason=six.text_type(e))


def get_lpar_state(oper, host, lpar_uuid):
    """
    Get state of LPAR directly from HMC
    """
    k2resp = oper.read(k2const.MGT_SYS,
                       rootId=host.uuid,
                       childType=k2const.LPAR,
                       childId=lpar_uuid)
    lpar_entry = k2resp.entry
    lpar_wrapper = LPAREntryWrapper(lpar_entry)
    lpar_state = lpar_wrapper.get_lpar_state()
    return lpar_state


def extract_flavor_from_prof(entry):
    """
    Extract flavor information needed by onboarding a LPAR via profile
    """
    flavor = {}
    extra_specs = {}
    prof_wrapper = LPARProfileEntryWrapper(entry)
    # check whether it's using dedicated or shared proc
    proc_mode = prof_wrapper.get_proc_mode()
    sharing_mode = prof_wrapper.get_sharing_mode()
    extra_specs['powervm:dedicated_proc'] = proc_mode
    if proc_mode == 'true':
        flavor['vcpus'] = prof_wrapper.get_procs()
        extra_specs['powervm:max_vcpu'] = prof_wrapper.get_max_procs()
        extra_specs['powervm:min_vcpu'] = prof_wrapper.get_min_procs()
        extra_specs['powervm:dedicated_sharing_mode'] = sharing_mode
    else:
        flavor['vcpus'] = prof_wrapper.get_vcpus()
        extra_specs['powervm:max_vcpu'] = prof_wrapper.get_max_vcpus()
        extra_specs['powervm:min_vcpu'] = prof_wrapper.get_min_vcpus()
        extra_specs['powervm:proc_units'] = prof_wrapper.get_proc_units()
        extra_specs['powervm:_max_proc_units'] = \
            prof_wrapper.get_max_proc_units()
        extra_specs['powervm:_min_proc_units'] = \
            prof_wrapper.get_min_proc_units()
        uncapped = 'true'
        if sharing_mode != 'uncapped':
            uncapped = 'false'
        else:
            extra_specs['powervm:shared_weight'] = (
                prof_wrapper.get_uncapped_weight())
        extra_specs['powervm:uncapped'] = uncapped

    flavor['memory_mb'] = prof_wrapper.get_mem()
    extra_specs['powervm:min_mem'] = prof_wrapper.get_max_mem()
    extra_specs['powervm:max_mem'] = prof_wrapper.get_min_mem()
    extra_specs['powervm:processor_compatibility'] = (
        prof_wrapper.get_proc_compatibility_mode())
    flavor['extra_specs'] = extra_specs
    # availability need to extract from LPAR
    return flavor


@logcall
def power_on_off(opr, uuid, suffix, host, entry=None,
                 force_immediate=False, restart=False,
                 timeout=CONF.ibmpowervm_job_request_timeout):
    """
    Power on or off an instance (based on suffix passed)
    :param opr: K2 operator
    :param uuid: uuid of the instance to power off or power on
    :param suffix: power option - 'PowerOn' or 'PowerOff'
    :param host: lpar's host
    :param entry: k2 entry for the lpar (for PowerOff suffix only)
    :param force_immediate: do immediate shutdown
                            (for PowerOff suffix only)
    :param restart: do a restart after power off
                            (for PowerOff suffix only)
    :param timeout: value in seconds for specifying
                    how long to wait for the LPAR to stop
                    (for PowerOff suffix only)
    """
    complete = False
    try:
        while not complete:
            k2resp = opr.read(k2const.LPAR,
                              uuid, suffixType=k2const.SUFFIX_TYPE_DO,
                              suffixParm=suffix)
            job_wrapper = JobEntryWrapper(k2resp.entry)
            job_parms = None
            if suffix == k2const.SUFFIX_PARM_POWER_OFF:
                operation = 'shutdown'
                add_immediate = False
                if force_immediate:
                    add_immediate = True
                elif entry is not None:
                    lpar_wrapper = LPAREntryWrapper(entry)
                    rmc_state = lpar_wrapper.check_dlpar_connectivity()[1]
                    # Check IBMi VM's reference code
                    ref_code = lpar_wrapper.get_ref_code()
                    lpar_type = lpar_wrapper.get_lpar_type()
                    if lpar_type == IBMI_OS_TYPE and ref_code == '00000000':
                        timeout = CONF.ibmpowervm_job_request_timeout
                        operation = 'osshutdown'
                    elif (lpar_type != IBMI_OS_TYPE and rmc_state == 'active'
                          and lpar_wrapper.get_rmc_os_shutdown_capability()):
                        operation = 'osshutdown'
                    else:
                        add_immediate = True
                job_parms = [job_wrapper.create_job_parameter('operation',
                                                              operation)]
                if add_immediate:
                    job_parms.append(
                        job_wrapper.create_job_parameter('immediate', 'true'))
                if restart:
                    job_parms.append(
                        job_wrapper.create_job_parameter('restart', 'true'))
            try:
                job_wrapper.run_job(opr, uuid, job_parms=job_parms,
                                    timeout=timeout, host_uuid=host.uuid)
                complete = True
            except common_exception.IBMPowerVMJobRequestTimedOut as error:
                if (suffix == k2const.SUFFIX_PARM_POWER_OFF and
                        operation == 'osshutdown'):
                    timeout = CONF.ibmpowervm_job_request_timeout
                    force_immediate = True
                else:
                    LOG.exception(_('Error: %s') % error)
                    if suffix == k2const.SUFFIX_PARM_POWER_OFF:
                        raise nova_exception.InstancePowerOffFailure(
                            six.text_type(error))
                    else:
                        raise nova_exception.InstancePowerOnFailure(
                            six.text_type(error))
            except common_exception.IBMPowerVMJobRequestFailed as error:
                LOG.exception(_('Error: %s') % error)
                error_msg = six.text_type(error)
                if suffix == k2const.SUFFIX_PARM_POWER_OFF:
                    # If already powered off and not a reboot,
                    # don't send exception
                    if 'HSCL1558' in error_msg and not restart:
                        complete = True
                    # If failed because RMC is now down, retry with force
                    elif ('HSCL0DB4' in error_msg and
                          operation == 'osshutdown'):
                        timeout = CONF.ibmpowervm_job_request_timeout
                        force_immediate = True
                    else:
                        raise nova_exception.InstancePowerOffFailure(
                            six.text_type(error))
                else:
                    # If already powered on, don't send exception
                    if 'HSCL3681' in error_msg:
                        complete = True
                    else:
                        raise nova_exception.InstancePowerOnFailure(
                            six.text_type(error))
    # Invalidate the LPARentry in the k2operator cache so the
    # compute manager gets the current LPAR state at the end of the
    # start/stop operation. Feeds must be invalidated too, since the
    # k2 operator will use them if an entry is not in the cache
    # before it will request the entry from the HMC.
    finally:
        try:
            opr.invalidate_cache_elem(k2const.MGT_SYS, rootId=host.uuid,
                                      childType=k2const.LPAR, childId=uuid,
                                      invalidate_feeds=True)
        except Exception as e:
            LOG.exception(_('Error invalidating k2operator cache for LPAR'
                            ' UUID %(lpar_uuid)s: %(exc)s') %
                          {'lpar_uuid': uuid, 'exc': six.text_type(e)})


@logcall
@synchronized('lpar-attach-interfaces-lock')
def attach_interfaces(context, host, host_stats, lpar_entry, instance,
                      network_info, deploy):
    # Determine custom mac capability
    if host_stats is None:
        host_stats = host.get_host_stats_internal()
    p6_compat = False
    lpar_wrapper = LPAREntryWrapper(lpar_entry)
    curr_proc_mode, pend_proc_mode = lpar_wrapper.get_proc_compat_modes()
    LOG.debug('Retrieved Processor Compatibility Modes: '
              'current = %s, pending/desired = %s'
              % (curr_proc_mode, pend_proc_mode))
    if 'power6' in pend_proc_mode.lower():
        p6_compat = True

    # Determine if we can set the MAC address on the VM
    custom_mac_addr_capable = host_stats['stats'][CUSTOM_MAC_ADDR_CAPABLE]\
        and not p6_compat

    if not custom_mac_addr_capable:
        LOG.info(_('Server %s is not custom MAC capable') % instance['name'])

    with K2OperGuard(host=host) as oper:
        try:
            # If we're not here for deploy, we need to calculate the next
            # highest slot to use so subsequent boots of the lpar do not
            # reorder the network adapters and accidentally move IPs on to
            # the wrong adapters.
            slot = 0
            if not deploy:
                for cna_link in lpar_wrapper.get_cna_links():
                    cna_resp = oper.readbyhref(cna_link)
                    cna = ClientNetworkAdapterWrapper(cna_resp.entry)
                    slot = cna.get_slot() if cna.get_slot() > slot else slot

            respList = []
            # Cache the vswitch_map so we don't have to build it for every
            # vif passed to create_virtual_ethernet_adapter
            vswitch_map = None
            for vif in network_info:
                # Only precreated neutron ports have slot_id set (by
                # ibmpowervm_api.allocate_for_instance) prior to now.
                precreated_port = vif.get('slot_id') is not None

                # If we calculated the slot above, then we're here for an
                # attach vnic.  Set vif info accordingly.
                if slot > 0:
                    # Slot is the highest currently used, bump it first
                    slot += 1
                    vif['slot_id'] = slot

                k2resp, vswitch_map = vea_wrapper.\
                    create_virtual_ethernet_adapter(
                        oper=oper,
                        lpar=lpar_wrapper,
                        network_vif=vif,
                        custom_mac_addr_capable=custom_mac_addr_capable,
                        managed_system_uuid=host.uuid,
                        precreated_port=precreated_port,
                        vswitch_map=vswitch_map)
                if not custom_mac_addr_capable:
                    if k2resp:
                        respList.append(k2resp)
                    else:
                        LOG.error(_("create_virtual_ethernet_adapter for "
                                    "server %s, returned an empty "
                                    "k2resp") % instance['name'])
            if respList:
                LOG.info(_("Updating the network information for server "
                           "%s with the configured MAC addresses "
                           "because the server is not custom "
                           "MAC capable") %
                         instance['name'])
                vea_wrapper.update_mac_addresses(context,
                                                 instance, respList,
                                                 network_info, deploy)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.exception(_('Error creating network adapters: %s')
                              % six.text_type(e))


@logcall
def detach_interfaces(context, host, lpar_entry, instance, network_info):
    LOG.info(_('Server %s is ') % instance['name'])

    with K2OperGuard(host=host) as oper:
        try:
            for vif in network_info:
                k2resp = vea_wrapper.\
                    delete_virtual_ethernet_adapter(
                        oper=oper,
                        lpar_entry=lpar_entry,
                        network_vif=vif)

                # k2 response can be empty for various reasons.
                # if there is no CNA to detach, k2 response will be
                # empty. This happens during deploy failure scenarios as well
                # So, moved below logging statement to debug from error.
                if not k2resp:
                    LOG.warning(_("delete_virtual_ethernet_adapter for "
                                "server %s, returned an empty "
                                "k2resp") % instance['name'])
        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.exception(_('Error deleting network adapters: %s')
                              % six.text_type(e))


@logcall
def _create_luarecovery_xml(device_list,
                            cmd_version,
                            ireliableITL=True,
                            iversion="2.0"):
    """Create an luarecovery Input XML

        :param devlice_list: a list of (ivendor, ideviceTag, itls)
        :param ivendor: This string must be upper case.
                        The following are expected vendor strings:
                        all other cases the attribute should be
                        set to "OTHER":
                        1. IBM
                        2. EMC
                        3. NETAPP
                        4. HDS
                        5. HP

        :param ideviceTag: a valid deviceTag of the storage
        :param device_list: a list of (iwwpn, twwpn, lua)
            :param iwwpn: world wide port name of a VIOS initiator
            :param twwpn: world wide port name of a target port on
                          storage appliance WWPN
            :param lua: Logical unit address of a particular piece of
                        storage (device) that can be accessed from the
                        target port.
        :param ireliableITL: False if the ITL nexus is based on a current
                             VIOS adapter inventory and True if its based
                             on an ITL nexus at the time the logical unit
                             was created.
        :param version: string containing requested version of the xml doc,
                        must be "1.1" for Capability version 3
    """

    root = Element("XML_LIST")
    # general
    general = Element("general")
    root.append(general)
    lcmd_version = Element("cmd_version")
    lcmd_version.text = cmd_version
    general.append(lcmd_version)
    version = Element("version")
    version.text = iversion
    general.append(version)

    # reliableITL is ignored for Lua recovery v3 by VIOS
    reliableITL = Element("reliableITL")
    reliableITL.text = "true" if ireliableITL else "false"
    root.append(reliableITL)

    # deviceList
    deviceList = Element("deviceList")
    root.append(deviceList)

    for idevice in device_list:
        (ivendor, ideviceid, itls) = idevice

        device = Element("device")
        deviceList.append(device)

        # vendor
        if ivendor is not None:
            vendor = Element("vendor")
            vendor.text = ivendor
            device.append(vendor)

        # deviceid
        if ideviceid is not None:
            deviceid = Element("deviceID")
            deviceid.text = ideviceid
            device.append(deviceid)

        # deviceTag
        deviceTag = Element("deviceTag")
        deviceTag.text = "1"
        device.append(deviceTag)

        itlList = Element("itlList")
        device.append(itlList)

        first = True
        for itl in itls:
            if first:
                number = Element("number")
                number.text = "%d" % (len(itls))
                itlList.append(number)
                first = False

            (iIwwpn, iTwwpn, ilua) = itl

            itl = Element("itl")
            itlList.append(itl)

            # Iwwpn
            Iwwpn = Element("Iwwpn")
            Iwwpn.text = iIwwpn
            itl.append(Iwwpn)

            # Twwpn
            Twwpn = Element("Twwpn")
            Twwpn.text = iTwwpn
            itl.append(Twwpn)

            # Lua
            lua = Element("lua")
            lua.text = ilua
            itl.append(lua)

    return ElementTree.tostring(root)


@logcall
def _validate_lua_connection(connection):
    """ Validates the LUA Connection for Target LUN """
    try:
        # validate connection
        if (connection is None
                or 'vscsi_target_lun' not in connection
                or len(connection['vscsi_target_lun']) == 0):
            msg = (_("lua_recovery Failed: bad connection: >%s<"
                     "no target") % connection)
            raise Exception(msg)

    except Exception as e:
        with excutils.save_and_reraise_exception():
            LOG.exception(_('lua_recovery: Invalid Connection: %s')
                          % six.text_type(e))


@logcall
def _process_lua_result(result):
    """ Processes the Output XML returned by LUARecovery """

    device_list = []
    if result is not None:
        root = ElementTree.fromstring(result)
        for child in root:
            if child.tag == "deviceList":
                for gchild in child:
                    status = None
                    pv_name = None
                    unique_id = None
                    udid = None
                    message = None
                    for ggchild in gchild:
                        if ggchild.tag == "status":
                            status = ggchild.text
                        elif ggchild.tag == "pvName":
                            pv_name = ggchild.text
                        elif ggchild.tag == "uniqueID":
                            unique_id = ggchild.text
                        elif ggchild.tag == "udid":
                            udid = ggchild.text
                        elif ggchild.tag == "msg":
                            for mchild in ggchild:
                                if mchild.tag == "msgText":
                                    message = mchild.text
                    lua_rslt = {'status': status, 'pv_name': pv_name,
                                'unique_id': unique_id, 'udid': udid,
                                'message': message}
                    device_list.append(lua_rslt)

        LOG.debug('RESPONSE: %s' % device_list)
        return device_list


@logcall
def _get_lua_status(host, result):
    status = result[0]['status']
    dev_name = result[0]['pv_name']
    message = result[0]['message']
    udid = result[0]['udid']
    if (status == k2const.LUA_STATUS_DEVICE_AVIALABLE):
        LOG.info(_("LUA Discovery Successful Device Found: %s"),
                 dev_name)
    elif (status == k2const.LUA_STATUS_FOUND_ITL_ERR):
        message = result[0]['message']
    elif (status == k2const.LUA_STATUS_DEVICE_IN_USE):
        message = dev_name + " Device is currently in Use"
        LOG.warn(six.text_type(message))
    elif (status == k2const.LUA_STATUS_FOUND_DEVICE_UNKNOWN_UDID):
        message = dev_name + " Device discovered with unknown uuid"
        LOG.warn(six.text_type(message))
    elif (status == k2const.LUA_STATUS_INCORRECT_ITL):
        message = "Failed to Discover the Device Status : " + status
        LOG.warn(six.text_type(message))
    return status, dev_name, message, udid


@logcall
def _update_connection_itls(connection, vol_itl_list, vios_name, vios_wwpns):
    """ For onboarded volumes we do not have the vios names
    in the connection info so connection info needs to be updated
    to have the same which can be used for lua discovery
    """
    # validate connection
    if (connection is None
            or 'vscsi_target_lun' not in connection):
        LOG.warn(_("connection info does not have the ITL List"))

    vscsi_itl_list = copy.deepcopy(vol_itl_list)
    connection['vscsi_target_lun'] = []
    for itl in vscsi_itl_list:
        match = list(set(vios_wwpns) & set(itl['vios_wwpn']))
        LOG.debug("Vios wwpn %(vios)s itl wwpn %(itl)s and "
                  "matched WWPN %(wwpn)s Updating vios_name %(name)s: " %
                  {'vios': vios_wwpns, 'itl': itl['vios_wwpn'],
                   'wwpn': match, 'name': vios_name})
        if match:
            itl['vios_name'] = vios_name
            itl['vios_wwpn'] = match
        else:
            itl['vios_name'] = None
        connection['vscsi_target_lun'].append(itl)

    return


@logcall
def cleanup_disks(host, instance, lpar_name, connection_info, context):
    """ cleanup routine to remove the left hdisk devices on the
        VIOS. It performs the below steps
        1. Iterate through the VIOS
        2. Invoke LUA discovery using the ITLs from the connection info
        3. If LUA discovery returns with the status 6 that means
        hdisk is properly removed from the VIOS
        4. For any other status code it means device still exists
        and might be used by some vhost
        5. No exception is thrown in case cleanup fails.
    """
    LOG.debug('cleanup_disks on host %(host)s. '
              'connection_info = %(conn_info)s' %
              {'host': host, 'conn_info': connection_info})

    scg_map = StorageMapping()
    # use default chk_conn_type of 'either' since we don't want exceptions
    # raised unnecessarily.
    scg_info = scg_map.get_connectivity_info(get_admin_context(),
                                             instance['uuid'], host.host_id,
                                             instance,
                                             live_data=False)
    # extract scg class to six.text_type to avoid UnicodeEncodeError
    scg_class = scg_info.pop('scg')
    LOG.debug("Cleanup SCG info :%s, SCG class %s" %
              (scg_info, six.text_type(scg_class)))
    scg_info['scg'] = scg_class
    vios_names = [vio['name'] for vio in scg_info.get('vios_list')
                  if CONNECTION_TYPE_VSCSI in
                  vio.get('connectivity_types')]
    vol_itl_list = connection_info['vscsi_target_lun'] if 'vscsi_target_lun'\
        in connection_info else None

    volume_id = connection_info.get('data').get('volume_id')
    with K2OperGuard(host=host) as oper:
        try:
            # extract vios info
            # Get vios data - need to populate the vios cache first
            # by refreshing
            vioses_resp = oper.vios_read(k2const.MGT_SYS, rootId=host.uuid,
                                         childType=k2const.VIOS, age=0,
                                         xag=k2const.ALL_VIOS_EXT_PROPS)
            vios_entries = vioses_resp.feed.entries
            for vios_name in vios_names:
                # find the entry with the matching name
                feed = vioses_resp.feed
                vios_entries = feed.findentries(k2const.XP_PARTITION_NAME,
                                                vios_name)
                if len(vios_entries) == 0:
                    LOG.warning(_('No VIOS %s') % vios_name)
                    continue

                vios_uuid = vios_entries[0].properties['id']
                vios_wwpns = None
                for v in scg_info.get('vios_list'):
                    if (v['id'] == vios_uuid and CONNECTION_TYPE_VSCSI in
                            v.get('connectivity_types')):
                            vios_wwpns = [i['wwpn'] for i in v['vscsi:ports']]

                if ('vscsi_target_lun' not in connection_info or
                    'vios_name' not in connection_info['vscsi_target_lun'] or
                    [True for i in connection_info['vscsi_target_lun']
                     if 'vios_name' not in i or i['vios_name'] != vios_name]):
                        _update_connection_itls(connection_info, vol_itl_list,
                                                vios_name, vios_wwpns)
                try:
                    status, dev_name = _invoke_lua_recovery(
                        host, connection_info, vios_uuid,
                        lpar_name, vios_name)[:2]
                    if status == k2const.LUA_STATUS_INCORRECT_ITL:
                        msg = (_("Successfully Cleaned volume %(vol)s "
                                 "for %(lpar)s on %(vios)s") %
                               {'vol': volume_id, 'lpar': lpar_name,
                                'vios': vios_name})
                        LOG.info(msg)
                    else:
                        LOG.info(_("Cannot cleanup since device %(dev)s for"
                                   "volume %(volume)s on %(vios)s "
                                   "is discovered with status: %(status)s."),
                                 {'dev': dev_name, 'volume': volume_id,
                                  'vios': vios_name, 'status': status})
                except Exception as e:
                    msg = (_("Failed Cleanup volume %(vol)s for %(lpar)s "
                             "on %(vios)s reason %(msg)s") %
                           {'vol': volume_id, 'lpar': lpar_name,
                            'vios': vios_name, 'msg': six.text_type(e)})
                    LOG.warn(msg)

        except Exception as e:
            LOG.warn(_('Cannot cleanup volume to virtual machine %s') %
                     six.text_type(e))


@logcall
def _invoke_lua_recovery(host, connection, vios, lpar, vios_name):
    """
    invoke the LUARecovery for discovering the volume before attachment.
    but take the same serializer as we use for storage mapping
    """
    @synchronized('stg-mapping-lock' + vios)
    def _call_lua_recovery_locked(host, connection, vios, lpar, vios_name):
        for retries in xrange(POST_RETRIES + 1):
            try:
                return _call_lua_recovery(host, connection, vios,
                                          lpar, vios_name)
            except hmc_ex.IBMPowerVIOSIsTooBusy:
                with excutils.save_and_reraise_exception() as ctxt:
                    if (retries < POST_RETRIES):
                        LOG.debug(_("VIOS busy retry lua_recovery try %s") %
                                  retries)
                        greenthread.sleep(15)
                        # Do not re-raise the original exception right now
                        ctxt.reraise = False

    def _call_lua_recovery(host, connection, vios, lpar, vios_name):
        """
        K2 interface to invoke the LUARecovery for discovering the
        volume before attachment.
        """
        # validate the connection info
        try:
            _validate_lua_connection(connection)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.exception(_('lua_recovery: Invalid Connection: %s')
                              % six.text_type(e))
        lua = None
        itl_list = connection['vscsi_target_lun']
        for lun in itl_list:
            if vios_name == lun['vios_name']:
                lua = hex(int(lun['target_lun']))[2:]
                if len(lua) == 8:
                    lua += "00000000"
                else:
                    lua += "000000000000"
                ports = list(set(lun['vios_wwpn']))
                if isinstance(lun['target_wwpn'], list):
                    wwpns = list(set(lun['target_wwpn']))
                else:
                    wwpns = [lun['target_wwpn']]

        LOG.info(_("CONNECTION INFO : %s"), connection)
        # extract wwpnsvios
#        wwpns = connection['data']['target_wwn']
        vol_id = connection['data']['volume_id']
        itls = []
        for t_wwpn in list(set(wwpns)):
            for i_wwpn in ports:
                itls.append((i_wwpn, t_wwpn, lua))

        vendor = connection.get('data').get('vendor')
        device_enc = None
        if 'pg83NAA' in connection['data']:
            device_id = connection.get('data').get('pg83NAA')
            device_enc = base64.b64encode(device_id)

        cmd_version = "3"
        device_list = [(vendor, device_enc, itls)]
        inputXML = _create_luarecovery_xml(device_list, cmd_version)
        dev_name = None
        status = None
        message = None
        udid = None
        with K2OperGuard(host=host) as oper:
            try:
                # Get the K2 response for the LuaRecovery command
                k2resp = oper.vios_read(
                    k2const.VIOS, vios,
                    suffixType=k2const.SUFFIX_TYPE_DO,
                    suffixParm=k2const.SUFFIX_PARM_LUA_RECOVERY,
                    vios_name=vios_name)
                job_wrapper = JobEntryWrapper(k2resp.entry)
                job_parms = None
                job_parms = [job_wrapper.create_job_parameter(
                    'inputXML', inputXML)]
                LOG.info(_("INPUT XML:%s"), inputXML)
                job_wrapper.run_job(oper, vios, job_parms=job_parms)
                result = job_wrapper.get_job_results_as_dict()
                luaresult = _process_lua_result(result['OutputXML'])
                status, dev_name, message, udid = _get_lua_status(host,
                                                                  luaresult)
                if message and status == k2const.LUA_STATUS_FOUND_ITL_ERR:
                    info = (_("Volume %(vol)s to be attach on virtual "
                              "machine %(vm)s on host %(host)s is discovered"
                              " as %(dev)s with few ITL Errors."
                              "Reason: %(reason)s") %
                            {'vol': vol_id, 'vm': lpar, 'host': host.host_id,
                             'dev': dev_name, 'reason': message})
                    LOG.info(info)
            except common_exception.IBMPowerVMJobRequestFailed as error:
                LOG.warn(_('Error: %s') % error)

            return status, dev_name, udid

    return _call_lua_recovery_locked(host, connection, vios, lpar, vios_name)


@logcall
def _invoke_rmdev(host, cec_name, vios_name, dev_name, vios_uuid):
    """ K2 command to remove the device from the VIOS"""

    @synchronized('stg-mapping-lock' + vios_uuid)
    def _call_rmdev(host, cec_name, vios_name, dev_name):
        with K2OperGuard(host=host) as k2oper:
            try:
                rm_cmd = ('viosvrcmd -m ' + cec_name + ' -p ' + vios_name +
                          ' -c \"rmdev -dev ' + dev_name + '\"')
                LOG.debug('RMDEV Command Input: %s' % rm_cmd)
                # Get the K2 response for the CLIRunner command
                k2resp = k2oper.read(k2const.MGT_CONSOLE, None,
                                     suffixType=k2const.SUFFIX_TYPE_DO,
                                     suffixParm=k2const.SUFFIX_PARM_CLI_RUNNER)

                # Create the job parameters
                job_wrapper = JobEntryWrapper(k2resp.entry)
                ack_parm = 'acknowledgeThisAPIMayGoAwayInTheFuture'
                job_parms = [job_wrapper.create_job_parameter('cmd', rm_cmd),
                             job_wrapper.create_job_parameter(ack_parm,
                                                              'true')]
                job_wrapper.run_job(k2oper, None, job_parms=job_parms)

            except common_exception.IBMPowerVMJobRequestFailed as error:
                LOG.info(_('CLIRunner Error: %s') % error)
                pass

        result = job_wrapper.get_job_status()
        return result

    return _call_rmdev(host, cec_name, vios_name, dev_name)


@logcall
def attach_volume(host, instance, lpar_name, lpar_uuid, entry,
                  connection_info, context):
    """
    Attached the storage specified in the connection info to VIOS(es)
    for the instance.
    connection_info should look like:
    {
        'driver_volume_type': 'ibm_ssp'
        'data': {'target_discovered': True,
                 'UniqueDeviceID: <LU udid>
                 'UnitName': <LU name>
                }
    }
    """
    LOG.info(_('Attach volume to  LPAR %(lpar_name)s. UUID = %(lpar_uuid)s')
             % {'lpar_name': lpar_name, 'lpar_uuid': lpar_uuid})
    if lpar_uuid is None or entry is None:
        entry_details = _entry_to_string(entry)
        LOG.error(_('Virtual machine cannot be found. Cannot proceed. '
                    'LPAR UUID: %(lpar_uuid)s. Entry: %(entry_details)s.') %
                  {'lpar_uuid': lpar_uuid, 'entry_details': entry_details})
        raise common_exception.IBMPowerVMLPARInstanceNotFound(
            instance_name=lpar_name)

    lpar_wrapper = LPAREntryWrapper(entry)

    # extract the partition id
    lpar_id = lpar_wrapper.get_lpar_id()
    LOG.debug("lpar_id %s" % lpar_id)
    # extract the scg info
    scg_map = StorageMapping()
    # IMPORTANT: attach_volume is always called after prep_instance_for_attach.
    # Live data should have already been collected.
    # If that ever change, we would need to start collecting live data
    # powervc_nova.computer.manager override _attach_volume to call
    # prep_instance_for_attach first, it then calls super._attach_volume
    # which in turn calls driver.attach_volume
    # Since the prep_instance_for_attach has already called
    # get_connectivity_info asking for the proper validation, it is safe
    # for us to pass in CHK_EITHER since we need to figure out whether the
    # attachment is for a boot or data disk again.
    scg_info = scg_map.get_connectivity_info(context,
                                             instance['uuid'], host.host_id,
                                             instance, live_data=False)
    # extract scg class to six.text_type to avoid UnicodeEncodeError
    scg_class = scg_info.pop('scg')
    LOG.debug('scg_info: %s, scg_class: %s' %
              (scg_info, six.text_type(scg_class)))
    scg_info['scg'] = scg_class

    # extract connection info
    vios_names = []
    volume_type = connection_info.get('driver_volume_type')
    backing_dev = None
    add_info = None
    if volume_type == 'ibm_ssp':
        storage_type = storage_map_processor.LU
        backing_dev = connection_info.get('data').get('UnitName')
        add_info = connection_info.get('data').get('UniqueDeviceID')
        storage_map = SCSIMapProcessor()

        if CONNECTION_TYPE_SSP in scg_info.get('connectivity_types'):
            vios_names = [vio['name'] for vio in scg_info.get('vios_list')
                          if CONNECTION_TYPE_SSP in
                          vio.get('connectivity_types')]
        else:
            raise exception.IBMPowerVMInvalidStorageConnectivityInfo(
                instance_name=lpar_name)
    elif volume_type == 'fibre_channel':
        connection_type = _get_volume_connection_type(connection_info)
        if connection_type == CONNECTION_TYPE_VSCSI:
            storage_type = storage_map_processor.PV
            vios_names = [vio['name'] for vio in scg_info.get('vios_list')
                          if CONNECTION_TYPE_VSCSI in
                          vio.get('connectivity_types')]
            storage_map = SCSIMapProcessor()
        elif connection_type == CONNECTION_TYPE_NPIV:
            # no-op for npiv since physical to virtual map is already done.
            return
        else:
            raise exception.IBMPowerVMInvalidStorageConnectivityInfo(
                instance_name=lpar_name)
    else:
        LOG.error(_('volume type is not supported %s') % volume_type)
        raise exception.\
            IBMPowerVMStorageTypeNotSupported(storage_type=volume_type)

    # extract the minimum vios required per vol
    # default to all vios since that's the default in previous release
    min_vios_per_vol = scg_info.get('min_vioses_per_vol', len(vios_names))

    num_vios_met_req = 0
    num_vios_failed = 0
    with K2OperGuard(host=host) as oper:
        devs_attached = {}
        try:
            # extract vios info
            # Get vios data - need to populate the vios cache first
            # by refreshing
            vioses_resp = oper.vios_read(k2const.MGT_SYS, rootId=host.uuid,
                                         childType=k2const.VIOS, age=0,
                                         xag=k2const.ALL_VIOS_EXT_PROPS)
            err_msgs = []
            for vios_name in vios_names:
                # find the entry with the matching name
                feed = vioses_resp.feed
                vios_entries = feed.findentries(k2const.XP_PARTITION_NAME,
                                                vios_name)

                if len(vios_entries) == 0:
                    LOG.warning(_('No VIOS %s') % vios_name)
                    continue

                if (len(vios_names) - num_vios_failed < min_vios_per_vol):
                    LOG.error(_("number of failed vioses exceeded the "
                                "minimum vios requirement %s")
                              % num_vios_failed)
                    err_msgs = ','.join([six.text_type(err_msgs.index(x))
                                         + ") " + six.text_type(x)
                                         for x in err_msgs])
                    raise common_exception.IBMPowerVMAllocStgForMinVIOSFailed(
                        **dict(instance_name=lpar_name,
                               min_vios=min_vios_per_vol,
                               reasons=err_msgs))

                vios_uuid = vios_entries[0].properties['id']
                LOG.debug("vios uuid %s" % vios_uuid)
                dev_name = None
                udid = None
                if storage_type == storage_map_processor.PV:
                    try:
                        volume_id = connection_info.get('data').\
                            get('volume_id')
                        status, dev_name, udid = _invoke_lua_recovery(
                            host, connection_info, vios_uuid,
                            lpar_name, vios_name)
                        if (dev_name is not None and
                                status == k2const.LUA_STATUS_DEVICE_IN_USE):
                            msg = (_("Discovered device %(dev)s for"
                                     "volume %(volume)s on %(vios)s "
                                     "is in use Shared Volume: %(status)s.") %
                                   {'dev': dev_name, 'volume': volume_id,
                                    'vios': vios_name, 'status': status})
                            LOG.info(msg)
                            backing_dev = dev_name
                        elif status in [k2const.LUA_STATUS_DEVICE_AVIALABLE,
                                        k2const.LUA_STATUS_FOUND_ITL_ERR]:
                            LOG.info(_("Discovered %(hdisk)s on vios %(vios)s "
                                       "for volume %(volume_id)s. "
                                       "Status code: %(status)s.") %
                                     {'hdisk': dev_name, 'vios': vios_name,
                                      'volume_id': volume_id,
                                      'status': status})
                            backing_dev = dev_name
                        else:
                            msg = (_("Failed to discover hdisk on %(vios)s "
                                     "for volume %(volume_id)s. status: "
                                     "%(status)s.") %
                                   {'vios': vios_name, 'volume_id': volume_id,
                                    'status': status})
                            LOG.warn(msg)
                            err_msgs.append(six.text_type(msg))
                            num_vios_failed += 1
                            continue

                    except Exception as e:
                        LOG.exception(_('LUA Recovery failure Error: %s')
                                      % six.text_type(e))
                        msg = (_("Failed to discover hdisk on %(vios)s for "
                                 "volume %(volume_id)s. message: "
                                 "%(status)s.") %
                               {'vios': vios_name, 'volume_id': volume_id,
                                'status': 'LUA Recovery Job Failed'})
                        LOG.warn(msg)
                        err_msgs.append(six.text_type(msg))
                        num_vios_failed += 1
                        continue

                # keep track of the storage we have already attached within
                # the method.
                # this information is used later to rollback the operation
                devs_attached[vios_uuid] = []

                vios_resp = None
                # for scsi, we just attach the volume specified
                try:
                    vios_resp = storage_map.attach_storage(oper, host.uuid,
                                                           vios_uuid,
                                                           lpar_id, None,
                                                           backing_dev,
                                                           storage_type,
                                                           add_info,
                                                           resp=vios_resp,
                                                           lpar_uuid=lpar_uuid)
                    devs_attached[vios_uuid].append(backing_dev)
                    LOG.info(_("Device Attached : %s"), backing_dev)
                    connection_info['data']['target_UDID'] = udid
                    num_vios_met_req += 1
                except Exception as attach_err:
                    LOG.warning(six.text_type(attach_err))
                    LOG.warn(_("Failed to attach storage on lpar %s") %
                             lpar_id)
                    err_msgs.append(six.text_type(attach_err))
            if num_vios_met_req < min_vios_per_vol:
                raise common_exception.IBMPowerVMAllocStgForMinVIOSFailed(
                    **dict(instance_name=lpar_name,
                           min_vios=min_vios_per_vol,
                           reasons=','.join([six.text_type(err_msgs.index(x))
                                             + ") " + six.text_type(x) for x in
                                             err_msgs])))
        except Exception as e:
            LOG.error(_('Cannot attach volume to virtual machine: %s')
                      % lpar_name)
            # rollback storage attached within the call
            try:
                for vios_uuid, devs in devs_attached.items():
                    if devs:
                        storage_map.detach_and_delete_adapter(
                            oper, host.uuid, vios_uuid,
                            backing_dev=devs, detach_only=True)
            except Exception as rollbackErr:
                LOG.debug('rollback attach failed %s' % rollbackErr)
            LOG.exception(_('Error: %s') % six.text_type(e))
            ex_args = {'backing_dev': six.text_type(backing_dev),
                       'instance_name': six.text_type(lpar_name),
                       'reason': six.text_type(e)}
            raise exception.IBMPowerVMVolumeAttachFailed(**ex_args)


@logcall
def detach_volume(host, lpar_name, lpar_uuid, entry, connection_info):
    """
    Detached the storage specified in the connection info to VIOS(es)
    for the instance.
    connection_info should look like:
    {
        'driver_volume_type': 'ibm_ssp'
        'data': {'target_discovered': True,
                 'UniqueDeviceID: <LU udid>
                 'UnitName': <LU name>
                }
    }
    """
    LOG.info(_('Detach volumes from LPAR %(lpar_name)s. UUID = %(lpar_uuid)s')
             % {'lpar_name': lpar_name, 'lpar_uuid': lpar_uuid})
    LOG.info(_('Detach Volume Connection Info: %s'), connection_info)
    if lpar_uuid is None or entry is None:
        entry_details = _entry_to_string(entry)
        # detach_volume is called during destroy. We need to handle cases
        # where lpar is already deleted and allow destroy to continue
        LOG.warning(_('Virtual machine cannot be found. Cannot proceed. '
                      'LPAR UUID: %(lpar_uuid)s. Entry: %(entry_details)s.') %
                    {'lpar_uuid': lpar_uuid, 'entry_details': entry_details})
        return

    # extract connection info
    lpar_wrapper = LPAREntryWrapper(entry)
    lpar_id = lpar_wrapper.get_lpar_id()
    backing_dev = None
    volume_type = connection_info.get('driver_volume_type')
    if volume_type == 'ibm_ssp':
        connector_type = CONNECTION_TYPE_SSP
        backing_dev = connection_info.get('data').get('UnitName')
        storage_map = SCSIMapProcessor()
    elif volume_type == 'fibre_channel':
        connector_type = _get_volume_connection_type(connection_info)
        if connector_type == CONNECTION_TYPE_VSCSI:
            volume_uid = connection_info.get('data').get('target_UDID')
            volume_id = connection_info.get('data').get('volume_id')
            storage_map = SCSIMapProcessor()
        else:
            # no-op for npiv since physical to virtual map is already done.
            return

    else:
        LOG.error(_('volume type is not supported %s') % volume_type)
        raise exception.\
            IBMPowerVMStorageTypeNotSupported(storage_type=volume_type)

    with K2OperGuard(host=host) as oper:
        try:
            # extract vios info
            # Get vios data - need to populate the vios cache first
            # by refreshing
            vioses_resp = oper.vios_read(k2const.MGT_SYS, rootId=host.uuid,
                                         childType=k2const.VIOS, age=0,
                                         xag=k2const.ALL_VIOS_EXT_PROPS)

            vios_entries = vioses_resp.feed.entries
            for entry in vios_entries:
                vios_uuid = entry.properties['id']
                vios_wrapper = VIOSEntryWrapper(entry)
                vios_name = vios_wrapper.get_vios_name()
                LOG.debug("vios uuid %s" % vios_uuid)
                if connector_type == CONNECTION_TYPE_VSCSI:
                    try:
                        backing_dev = vios_wrapper.\
                            get_hdisk_from_uid(volume_uid)

                        LOG.info(_("%(hdisk)s discovered device on vios "
                                   "%(vios_name)s for volume %(volume_id)s. "
                                   "volume_uid: %(volume_uid)s.") %
                                 {'hdisk': backing_dev,
                                  'vios_name': vios_name,
                                  'volume_id': volume_id,
                                  'volume_uid': volume_uid})

                        if not backing_dev:
                            LOG.info(_("No mapped device found on vios "
                                       "%(vios)s for volume %(volume_id)s."
                                       "volume_uid: %(volume_uid)s ")
                                     % {'volume_uid': volume_uid,
                                        'volume_id': volume_id,
                                        'vios': vios_name})
                            continue

                    except Exception as e:
                        LOG.error(_("Failed to get associated disk on vios "
                                    "%(vios_name)s for volume %(volume_id)s. "
                                    "volume_uid: %(volume_uid)s. "
                                    "Error: %(error)s")
                                  % {'error': six.text_type(e),
                                     'volume_uid': volume_uid,
                                     'volume_id': volume_id,
                                     'vios_name': vios_name})
                        continue
                if backing_dev:
                    storage_map.\
                        detach_and_delete_adapter(oper, host.uuid,
                                                  vios_uuid, lpar_id,
                                                  backing_dev=backing_dev)
                    if (CONF.enable_rmdev and
                            connector_type == CONNECTION_TYPE_VSCSI and
                            "hdiskpower" not in backing_dev):
                        _invoke_rmdev(host, CONF.host_display_name,
                                      vios_name, backing_dev,
                                      vios_uuid)
        except Exception as e:
            LOG.error(_('Cannot detach volumes from virtual machine: %s') %
                      lpar_name)
            LOG.exception(_('Error: %s') % six.text_type(e))
            ex_args = {'backing_dev': backing_dev,
                       'instance_name': lpar_name,
                       'reason': six.text_type(e)}
            raise exception.IBMPowerVMVolumeDetachFailed(**ex_args)


@logcall
def get_volume_mapped_entries(host, volume_id, vios_name):
    """
    Returns the number of VSCSI mappings for the specified volume
    :param lpar_uuid uuid of the lpar
    :param volume_id K2udid of the volume
    :param vios_name
    """
    LOG.info(_('get_volume_mapping: %(vol)s on %(vios)s'), {'vol': volume_id,
                                                            'vios': vios_name})

    backing_dev = None
    num_map_found = 0
    with K2OperGuard(host=host) as oper:
        try:
            # extract vios info
            # Get vios data - need to populate the vios cache first
            # by refreshing
            vioses_resp = oper.vios_read(k2const.MGT_SYS, rootId=host.uuid,
                                         childType=k2const.VIOS, age=0,
                                         xag=k2const.ALL_VIOS_EXT_PROPS)
            # find the entry with the matching name
            feed = vioses_resp.feed
            vios_entries = feed.findentries(k2const.XP_PARTITION_NAME,
                                            vios_name)

            if len(vios_entries) == 0:
                LOG.warning(_('No VIOS Entries found %s') % vios_name)
                return num_map_found

            vios_wrapper = VIOSEntryWrapper(vios_entries[0])
            try:
                backing_dev = vios_wrapper.\
                    get_hdisk_from_uid(volume_id)

                LOG.info(_("%(hdisk)s found device on vios "
                           "%(vios_name)s for volume_id %(volume_id)s. ") %
                         {'hdisk': backing_dev,
                          'vios_name': vios_name,
                          'volume_id': volume_id})
                storage_map = SCSIMapProcessor()
                scsi_map_entries = storage_map.find_map_entries(
                    vios_wrapper.get_virtual_scsi_mappings(), None, None,
                    backing_dev)
                if scsi_map_entries:
                    LOG.info("Number of mapped entries found %s" %
                             len(scsi_map_entries))
                    num_map_found = len(scsi_map_entries)

                if not backing_dev:
                    LOG.info(_("No mapped device found on vios "
                               "%(vios)s for volume %(volume_id)s.") %
                             {'vios': vios_name,
                              'volume_id': volume_id})
            except Exception as e:
                LOG.error(_("Failed to get associated disk on vios "
                            "%(vios_name)s for volume %(volume_id)s. "
                            "Error: %(error)s")
                          % {'error': six.text_type(e),
                             'volume_id': volume_id})
        except Exception as e:
            LOG.error(_('Cannot find the map entries for volumes %(vol)s'
                        'on vios %(vios)s') % {'vol': volume_id,
                                               'vios': vios_name})
            LOG.info(six.text_type(e))
    return num_map_found


@logcall
def detach_all_volumes(oper, host, lpar_name, lpar_uuid, entry):
    """
    Detached all the the storage mapped to all VIOSes for the instance.
    """
    LOG.info(_('Detach volumes from LPAR %(lpar_name)s. UUID = %(lpar_uuid)s')
             % {'lpar_name': lpar_name, 'lpar_uuid': lpar_uuid})

    lpar_wrapper = LPAREntryWrapper(entry)
    # extract the partition id
    lpar_id = lpar_wrapper.get_lpar_id()
    LOG.debug("lpar_id %s" % lpar_id)

    try:
        # Get vios data. Need to populate the vios cache first by refreshing
        with K2OperGuard(host=host) as oper:
            vioses_resp = oper.vios_read(k2const.MGT_SYS, rootId=host.uuid,
                                         childType=k2const.VIOS,
                                         xag=[k2const.VIOS_VSCSI_MAP_EXT_PROP,
                                              k2const.VIOS_VFC_MAP_EXT_PROP])
            vios_entries = vioses_resp.feed.entries
            if len(vios_entries) <= 0:
                LOG.error(_('No VIOS found on host %s.')
                          % host.host_display_name)
                oper.dump_logs()
                raise exception.IBMPowerVMVIOSNotFound()

        for vios_entry in vios_entries:
            vios_uuid = vios_entry.properties['id']
            stg_map_processor = SCSIMapProcessor()
            # Find out media names
            media_names = stg_map_processor.\
                find_attached_medias(vios_entry.element,
                                     lpar_id)
            LOG.debug("Media names %s " % media_names)
            # Detach and delete adapters
            LOG.debug("Detaching storage adapters on vios uuid %s"
                      % vios_uuid)
            notify_del_adapters_msg = False
            try:
                stg_map_processor.\
                    detach_and_delete_adapter(oper,
                                              host.uuid,
                                              vios_uuid,
                                              client_lpar_id=lpar_id,
                                              ignore_type=True)
            except exception.IBMPowerVIOSIsNotRunning as vios_err:
                # ignore the error if the VIOS is not up
                # virtual server adapters will be cleaned up
                notify_del_adapters_msg = True
                LOG.warning(_('Unable to remove virtual storage devices: '
                              'Exception %s.')
                            % six.text_type(vios_err))
            except Exception as e:
                LOG.warning(_("Detachment of vSCSI adapters failed."
                              " Try again. Error: %s") % six.text_type(e))
                greenthread.sleep(5)
                try:
                    stg_map_processor.\
                        detach_and_delete_adapter(oper,
                                                  host.uuid,
                                                  vios_uuid,
                                                  client_lpar_id=lpar_id,
                                                  ignore_type=True)
                except exception.IBMPowerVIOSIsNotRunning as vios_err:
                    notify_del_adapters_msg = True
                    LOG.warning(_('Unable to remove virtual storage devices: '
                                  'Exception %s.')
                                % six.text_type(vios_err))

            if notify_del_adapters_msg:
                vios_wrapper = VIOSEntryWrapper(vios_entry)
                info = {'msg': _('The virtual storage devices '
                                 'for virtual machine {lpar_name} cannot '
                                 'be removed because Virtual I/O Server '
                                 '{vios_name} is not running. '
                                 'After the virtual machine is deleted, you '
                                 'must manually remove the devices from the '
                                 'Virtual I/O Server.'),
                        'lpar_name': lpar_name,
                        'vios_name': vios_wrapper.get_vios_name()}
                notifier = rpc.get_notifier(service='compute',
                                            host=host.host_id)
                notifier.warn(context.get_admin_context(),
                              'compute.instance.log',
                              info)
            # Remove media from media repository
            for media_name in media_names:
                LOG.debug("Removing media %s " % media_name)
                virtual_optical = VirtualOptical()
                try:
                    virtual_optical.remove_vopt_media(host,
                                                      media_name,
                                                      vios_uuid)
                except Exception as exp:
                    LOG.warn(_('Unable to remove virtual optical '
                               'device %(media)s from repository with '
                               'exception %(exp)s. User needs to '
                               'manually remove it.') %
                             {'media': media_name, 'exp': exp})
                    info = {'msg': _('Virtual optical device {media_name} '
                                     'for virtual machine {lpar_name} cannot '
                                     'be removed from the media repository. '
                                     'After the virtual machine is deleted, '
                                     'you must manually remove the device '
                                     'from the media repository'),
                            'lpar_name': lpar_name,
                            'media_name': media_name}
                    notifier = rpc.get_notifier(service='compute',
                                                host=host.host_id)
                    notifier.warn(context.get_admin_context(),
                                  'compute.instance.log',
                                  info)

    except Exception as e:
        LOG.error(_('Cannot detach volumes from virtual machine: %s') %
                  lpar_name)
        LOG.exception(_('Error: %s') % six.text_type(e))
        ex_args = {'backing_dev': None,
                   'instance_name': lpar_name,
                   'reason': six.text_type(e)}
        raise exception.IBMPowerVMVolumeDetachFailed(**ex_args)


def rr_cleanup(ctx, source_host, lpar_uuid, instance, lpar_name):
    """
    Creates a remote restart cleanup k2 request job, removes
    VM UUID from cache and unplugs vifs.

    :param ctx: security context
    :param source_host: Source host
    :param lpar_uuid: The UUID of the LPAR
    :param instance: instance object of the VM to be cleaned up
    :param lpar_name: The pvm name of the LPAR
    """

    inst_name = instance['display_name']
    with K2OperGuard(host=source_host) as k2oper:
        try:
            _rr_cleanup_job(k2oper, source_host, lpar_uuid)
        except common_exception.IBMPowerVMJobRequestTimedOut as exc:
            LOG.exception(exc)
            LOG.info(_("Remote restart cleanup for virtual machine "
                       "%(lpar)s timed out. ") %
                     {'lpar': inst_name})
            return
        except Exception as e:
            LOG.error(_("Remote restart cleanup for virtual machine "
                        "%(lpar)s failed with error '%(error)s'") %
                      {'lpar': inst_name,
                       'error': six.text_type(e)})
            return
    # Remove instance from UUID cache
    source_host.pvm_uuid_cache.remove(lpar_name)
    # unplug VLAN at source after cleanup
    network_api = network.API()
    network_info = network_api.get_instance_nw_info(ctx, instance)
    source_host.unplug_vifs(instance, network_info)

    host_name = source_host.host_display_name
    LOG.info(_('Remote restart cleanup for virtual machine '
               '%(lpar)s on host %(src)s was successful.') %
             {'lpar': inst_name, 'src': host_name})
    # Send event notification
    info = {'msg': _('Remote restart cleanup of virtual machine '
                     '{instance_name} on host {source_host_name} '
                     'was successful.'),
            'instance_name': inst_name,
            'source_host_name': host_name}
    notifier = rpc.get_notifier(service='compute', host=host_name)
    notifier.info(ctx, 'compute.instance.log', info)


def _rr_cleanup_job(k2oper, source_host, lpar_uuid):
    """
    Creates a remote restart cleanup k2 job.
    """
    # Get the K2 response for the remote restart cleanup command
    # SW390979: remote restart job with ManagedSystem UUID in the URI
    # is not supported
    k2resp = k2oper.read(k2const.LPAR, lpar_uuid,
                         suffixType=k2const.SUFFIX_TYPE_DO,
                         suffixParm=k2const.SUFFIX_PARM_REMOTE_RESTART)

    # Create the job parameters
    job_wrapper = JobEntryWrapper(k2resp.entry)

    job_parms = [job_wrapper.create_job_parameter('managedSystem',
                                                  source_host.uuid)]
    job_parms.append(job_wrapper.create_job_parameter('operation',
                                                      'cleanup'))

    LOG.debug('Submitting K2 remote restart cleanup job '
              'for VM: %s' % lpar_uuid)
    # Submit the job
    job_wrapper.run_job(k2oper, lpar_uuid,
                        job_parms=job_parms)


def _entry_to_string(entry):
    if entry is None:
        string = 'None'
    else:
        string = 'properties:' + entry.properties + ". element: " + \
            entry.element.toxmlstring()
    return string


@logcall
def _get_volume_connection_type(connection_info):
    """
    Gets the volume connection type.

    Volume connection type is available in connection_info (one/two/all
    these three properties may be available.
        - connection-type, connection_type, connector_type).

    Either of those properties is fine. However, we would prefer in this
    order - connection-type, connection_type, connector_type

    :param connection_info: connection info for the volume.
    """
    if connection_info is None:
        return None

    vol_connection_type = None

    if 'volume_connector' not in connection_info:
        vol_connection_type = \
            connection_info.get('data').get('connection-type')
        if vol_connection_type is None:
            vol_connection_type = \
                connection_info.get('data').get('connector_type')
    else:
        volume_connector = connection_info.get('volume_connector')
        # Get the connection-type from volume_connector
        if 'connection-type' in volume_connector:
            vol_connection_type = volume_connector.get('connection-type')

        # Still could not get connection type? Get it from connector
        if vol_connection_type is None:
            connector = volume_connector.get('connector')
            if connector is not None:
                vol_connection_type = connector.get('connection_type')

    return vol_connection_type


@logcall
def get_volume_connector(host, instance, lpar_name, lpar_uuid, entry):
    """
    Extract client adapter wwpns specifiec in connection info to VIOS(es)
    for the instance.
    connection_info should look like:
    {
     "connectivity_types": [
         "npiv", "ibm_ssp"
         ],
     "vios_list": [
         {
          "lpar_id": 1,
          "name": "hv4lpar05a",
          "id": "8203E4A_10007CA##1",
          "ports": {
              "A": [
                  {
                   "udid": "1aU789C.001.AAA0140-P1-C1-T2",
                   "wwpn": "10000000C9BB327B",
                   "name": "fcs1"
                  }
              ],
              "B": [
                  {
                   "udid": "1aU789C.001.AAA0140-P1-C1-T4",
                   "wwpn": "10000000C9BB327D",
                   "name": "fcs3"
                  }
              ]
          }
         }
     ]
    }
    """
    def _extract_wwpn_mappings(vios_entries, lpar_id):
        if vios_entries is None:
            LOG.error(_('No VIOS'))
            raise exception.IBMPowerVMVIOSNotFound()
        storage_map = FCMapProcessor()
        return storage_map.extract_wwpn_mappings(vios_entries[0],
                                                 lpar_id)

    # extract the scg info
    scg_map = StorageMapping()
    collect_live_data = True
    # Using 'either' returns the same data as 'all' but won't blow up
    # if one of boot/data are not OK. We don't want it to validate a
    # specific case by default because in this scope we cannot assume
    # which of boot or data connectivity is needed. Other points in the
    # attach flow will verify boot and/or data. The same is true for other
    # operations that need to validate the SCG's connectivity.
    chk_conn_type = 'either'
    # if this is done via spawn, we have already collected live data.
    if instance['vm_state'] == vm_states.BUILDING:
        collect_live_data = False
        bdms = blk_dev_obj.BlockDeviceMappingList.get_by_instance_uuid(
            get_admin_context(), instance['uuid'])
        # Just keep checking only boot for this single volume deploy case.
        if not bdms or len(bdms) == 1:
            chk_conn_type = CHK_BOOT

    scg_info = scg_map.get_connectivity_info(get_admin_context(),
                                             instance['uuid'], host.host_id,
                                             instance,
                                             live_data=collect_live_data,
                                             chk_conn_type=chk_conn_type)
    # extract scg class to six.text_type to avoid UnicodeEncodeError
    scg_class = scg_info.pop('scg')
    LOG.debug('scg_info: %s, scg_class: %s' %
              (scg_info, six.text_type(scg_class)))
    scg_info['scg'] = scg_class

    lpar_id = None
    lpar_ip = None
    os_type = None
    if lpar_uuid is None or entry is None:
        if scg_info and CONNECTION_TYPE_VSCSI not in \
                scg_info.get('connectivity_types'):
            # It must be NPIV only. In case of NPIV, we cannot proceed without
            # lpar_uuid and entry.
            LOG.warning(_('Missing volume connector because the virtual '
                          'machine %s cannot be found.') % instance['name'])
            return None
    else:
        lpar_wrapper = LPAREntryWrapper(entry)
        # extract the partition id
        if lpar_wrapper is not None:
            try:
                lpar_id = lpar_wrapper.get_lpar_id()
            except Exception as e:
                LOG.error(_('Error in getting lpar_id: %s') %
                          (six.text_type(e)))
            # extract the lpar ipaddress
            try:
                lpar_ip = lpar_wrapper.get_rmc_ipaddr()
                LOG.debug("Get LPAR IP: %s" % lpar_ip)
            except Exception as e:
                LOG.error(_('Error in getting lpar_ip: %s') %
                          (six.text_type(e)))
    # the os_distro is currently used by ds8k to determine the type of host
    # that will be created on the back end server. The ds8k treats
    # ibmi/aix/linux differently.
    if ((instance.get('system_metadata') and
         (instance['system_metadata'].
          get('image_os_distro')))):
        os_type = (instance['system_metadata'].
                   get('image_os_distro'))

    LOG.debug("lpar_id %(lpar_id)s os_type %(os_type)s" %
              dict(lpar_id=lpar_id, os_type=os_type))

    if scg_info and (CONNECTION_TYPE_NPIV in scg_info.get('connectivity_types')
                     or CONNECTION_TYPE_VSCSI in
                     scg_info.get('connectivity_types')):
        vios_t = []
        for vio in scg_info.get('vios_list'):
            if CONNECTION_TYPE_NPIV in vio.get('connectivity_types'):
                vios_t.append(vio['name'])
            if CONNECTION_TYPE_VSCSI in vio.get('connectivity_types'):
                vios_t.append(vio['name'])
        vios_names = set(vios_t)
    else:
        LOG.warning(_('Missing volume connector because the storage '
                      'connectivity group is None or does not contain '
                      'supported connectivity type. %s') % scg_info)
        return None

    wwpns = []
    wwpn_vscsi = []
    phy_port_map = {}
    wwpn_port_list = []
    backend_host_group = None
    with K2OperGuard(host=host) as oper:
        dumped_vioses_resp = False
        try:
            # extract vios info
            # this is called while instance is being deployed
            vioses_resp = oper.vios_read(k2const.MGT_SYS, rootId=host.uuid,
                                         childType=k2const.VIOS,
                                         xag=[k2const.VIOS_VFC_MAP_EXT_PROP,
                                              k2const.VIOS_VSCSI_MAP_EXT_PROP])
            # loop through each vios
            for vios_name in vios_names:
                # find the entry with the matching name
                feed = vioses_resp.feed
                vios_entries = feed.findentries(k2const.XP_PARTITION_NAME,
                                                vios_name)

                if vios_entries is None:
                    LOG.error(_('No VIOS'))
                    raise exception.IBMPowerVMVIOSNotFound()
                if CONNECTION_TYPE_NPIV in scg_info.get('connectivity_types') \
                        and lpar_id is not None:
                    # variable to keep track whether we retried to refresh
                    # vios feed response
                    retried = False
                    # whether we found the wwpn
                    found = False
                    # TODO: workaround for defect 42218
                    # VIOS feed is out of data, performance vios
                    # GET to obtain the latest data
                    while not found:
                        ret_data = _extract_wwpn_mappings(vios_entries,
                                                          lpar_id)
                        if retried:
                            LOG.debug('wwpns after retry %s' % ret_data)
                            break
                        if not (ret_data and
                                ret_data['phy_to_virt_initiators']):
                            # additional debugging for the vios_resp
                            # out of date issue
                            if not dumped_vioses_resp:
                                LOG.debug('Possible cache out-of-date issue:\n'
                                          '%s' % vioses_resp.body)
                                dumped_vioses_resp = True
                            vioses_resp = oper.vios_read(
                                k2const.MGT_SYS, rootId=host.uuid,
                                childType=k2const.VIOS,
                                xag=[k2const.VIOS_VFC_MAP_EXT_PROP,
                                     k2const.VIOS_VSCSI_MAP_EXT_PROP],
                                age=0)
                            feed = vioses_resp.feed
                            vios_entries = feed.findentries(
                                k2const.XP_PARTITION_NAME, vios_name)
                            retried = True
                        else:
                            found = True

                    if ret_data:
                        wwpns += ret_data['wwpns']
                        phy_port_map.update(ret_data['phy_to_virt_initiators'])

                if CONNECTION_TYPE_VSCSI in scg_info.get('connectivity_types'):
                    vios_uuid = vios_entries[0].properties['id']
                    vscsi_ports = None
                    # check for vios not found
                    vios_lpar_id = None
                    for v in scg_info.get('vios_list'):
                        if v['id'] == vios_uuid:
                            vios_lpar_id = v['lpar_id']
                            vscsi_ports = v['vscsi:ports']
                    # It could be the case that vios was not found
                    if not vios_lpar_id:
                        LOG.error(_("No Matching VIOS %(uuid)s found for"
                                    "%(vios)s via K2"), vios_uuid, vios_name)
                        continue
                    # or that there are no valid vscsi ports for that vios
                    if not vscsi_ports:
                        LOG.error(_("No valid vscsi ports found for VIOS"
                                    " %(uuid)s."), vios_uuid)

                    wwpn_vscsi += [p['wwpn'] for p in vscsi_ports]

                    vios_host = host.host_id + "_" + str(vios_lpar_id)
                    vios_host = vios_host[:CONNECTOR_HOST_LEN]
                    vios_wrapper = VIOSEntryWrapper(vios_entries[0])
                    vios_ip = None
                    vios_ip = vios_wrapper.get_rmc_ipaddr()

                    vscsi_port_map = {'wwpns': copy.deepcopy(wwpn_vscsi),
                                      'vios_name': vios_name,
                                      'vios_uuid': vios_uuid,
                                      'vios_host': vios_host,
                                      'vios_ip': vios_ip}
                    wwpn_port_list.append(vscsi_port_map)
                    wwpn_vscsi = []
            # end of loop
            # get the backend_host_group if present
            backend_host_group = scg_info.get('backend_host_group')
            LOG.info("backend_host_group from scg %s" % backend_host_group)
        except Exception as e:
            LOG.warning(_('Cannot get volume connector to virtual machine: %s')
                        % lpar_name)
            LOG.exception(_('Error: %s') % six.text_type(e))
            return None

    # host is used for storage host name definition as well as
    # zone name definition. Keep it short and meaningful.

    if CONNECTION_TYPE_NPIV in scg_info.get('connectivity_types'):
        # set npiv hostname format
        fchost = lpar_name if lpar_name else host.host_id + "_" + str(lpar_id)
    else:
        # it must be VSCSI only
        # set vscsi hostname format
        fchost = host.host_id

    connector = {'host': fchost[:CONNECTOR_HOST_LEN],
                 'boot_conn_types': scg_info.get('boot_conn_types'),
                 'data_conn_types': scg_info.get('data_conn_types')}

    if CONNECTION_TYPE_NPIV in scg_info.get('connectivity_types'):
        connector['wwpns'] = wwpns
        connector['wwnns'] = wwpns
        connector['phy_to_virt_initiators'] = phy_port_map
        if lpar_ip:
            connector['ip'] = lpar_ip
        # add the OS type to the npiv connector this is used by some
        # volume drivers to determine support features for npiv
        # connections
        if os_type:
            connector['os_type'] = os_type
    if CONNECTION_TYPE_VSCSI in scg_info.get('connectivity_types'):
        connector['wwpn_vscsi'] = wwpn_port_list
        if backend_host_group:
            connector['backend_host_group'] = backend_host_group

    return connector


@logcall
def prep_instance_for_attach(host, instance, lpar_name, lpar_uuid, entry,
                             backend_type, context):
    """
    This method is to make sure there are sufficient number virtual adapters
    created to allow a successful storage attach.
    """
    @synchronized('prep_inst' + instance['uuid'])
    def _call_prep_instance_for_attach():
        if lpar_uuid is None or entry is None:
            entry_details = _entry_to_string(entry)
            LOG.error(_('Virtual machine cannot be found. Cannot proceed. '
                        'LPAR UUID: %(lpar_uuid)s. Entry: %(entry_details)s.')
                      % {'lpar_uuid': lpar_uuid,
                         'entry_details': entry_details})
            raise common_exception.IBMPowerVMLPARInstanceNotFound(
                instance_name=lpar_name)

        lpar_wrapper = LPAREntryWrapper(entry)

        dlpar_adapter_cap = _config_adapter_capable(lpar_wrapper)

        # extract the partition id
        lpar_id = lpar_wrapper.get_lpar_id()
        LOG.debug("lpar_id %s" % lpar_id)

        # extract the scg info
        scg_map = StorageMapping()

        # The following check should not fall under the vm state of BUILDING.
        # It is true for the general case since we support attaching
        # extra volumes to the boot set now.
        if volume_utils.PowerVCNovaVirtDriverVolumeAPI().\
                get_instance_volumes(context,
                                     instance, 'data'):
            chk_conn_type = CHK_ALL
        else:
            chk_conn_type = CHK_BOOT
        # get_connectivity_info() will raise exceptions for failures or
        # if there is no connectivity available.
        scg_info = scg_map.get_connectivity_info(context,
                                                 instance['uuid'],
                                                 host.host_id,
                                                 instance,
                                                 chk_conn_type=chk_conn_type)

        # extract scg class to six.text_type to avoid UnicodeEncodeError
        scg_class = scg_info.pop('scg')
        LOG.debug('scg_info: %s, scg_class: %s' %
                  (scg_info, six.text_type(scg_class)))
        scg_info['scg'] = scg_class

        # list of vios names we would need to update
        vios_uuids = []

        # keep track number of fc we need to creat for the VIOS
        ports_lookup = {}
        fab_order = {}
        devs_attached = {}
        npiv_fab_req = None
        ports_per_fab = None
        fabrics_needed = None

        # since we can only attach 1 storage at a time,
        # we would either need to create vscsi or vfc. Not both.
        prep_for_vscsi = False if backend_type == 'fc' else True
        if prep_for_vscsi:
            if CONNECTION_TYPE_SSP in scg_info.get('connectivity_types'):
                storage_map = SCSIMapProcessor()
                vios_uuids = (
                    [vio['id'] for vio in scg_info.get('vios_list')
                        if CONNECTION_TYPE_SSP in
                        vio.get('connectivity_types')])
            elif CONNECTION_TYPE_VSCSI in scg_info.get('connectivity_types'):
                storage_map = SCSIMapProcessor()
                vios_uuids = (
                    [vio['id'] for vio in scg_info.get('vios_list')
                        if CONNECTION_TYPE_VSCSI in
                        vio.get('connectivity_types')])
            else:
                raise exception.IBMPowerVMInvalidStorageConnectivityInfo(
                    instance_name=lpar_name)
        else:
            if CONNECTION_TYPE_NPIV in scg_info.get('connectivity_types'):
                storage_type = storage_map_processor.NPIV
                storage_map = FCMapProcessor()
                # create 1 vfc pair per vios per farbric
                for vio in scg_info.get('vios_list'):
                    if CONNECTION_TYPE_NPIV in vio.get('connectivity_types'):
                        vios_uuids.append(vio['id'])
                        ports_lookup[vio['id']] = vio['ports']
                        fab_order[vio['id']] = vio['npiv_fab_order']
                # extract fibric channel requirement
                npiv_fab_req = scg_info['scg'].fabric_set_req
                ports_per_fab = scg_info['scg'].ports_per_fabric_npiv
                fabrics_needed = scg_info['fabric_num_required']
            else:
                raise exception.IBMPowerVMInvalidStorageConnectivityInfo(
                    instance_name=lpar_name)

        # extract the minimum vios required per vol
        # default to all vios since that's the default in previous release
        min_vios_per_vol = scg_info.get('min_vioses_per_vol', len(vios_uuids))
        # if there are no vios in the scg info, do nothing
        if len(vios_uuids) == 0:
            return

        num_vios_met_req = 0
        err_msgs = []
        # keep track fabric mapped from all vioses
        host_fabric_mapped = set([])
        fabric_mapped_err_msgs = []
        with K2OperGuard(host=host) as oper:
            try:
                lpar_href = oper.buildhref(rootType=k2const.LPAR,
                                           rootId=lpar_wrapper.get_uuid())

                for vios_uuid in vios_uuids:
                    vios_fabric_mapped = set()
                    LOG.debug("vios uuid %s" % vios_uuid)
                    vios_name = ([vio['name'] for vio in
                                  scg_info.get('vios_list') if vio['id'] ==
                                  vios_uuid][0])
                    # extract vios info
                    vios_resp = None
                    try:
                        vios_resp = oper.vios_read(
                            k2const.MGT_SYS, rootId=host.uuid,
                            childType=k2const.VIOS,
                            childId=vios_uuid,
                            xag=[k2const.VIOS_VFC_MAP_EXT_PROP,
                                 k2const.VIOS_VSCSI_MAP_EXT_PROP],
                            vios_name=vios_name)
                    except Exception as k2err:
                        LOG.exception(k2err)

                    if vios_resp is None or vios_resp.entry is None:
                        LOG.error(_('Unable to find VIOS: %s.') % vios_uuid)
                        oper.dump_logs()
                        raise exception.IBMPowerVMVIOSNotFound()

                    # determine how many adapter pair we need to create
                    if prep_for_vscsi:
                        if (CONNECTION_TYPE_VSCSI in
                           scg_info.get('connectivity_types')):
                            num_adapters = 1
                            # For vscsi attach check to see if a valid adapter
                            # is already present. For eg., in case of npiv
                            # boot volume the fist vscsi data attach might not
                            # have an adapter created already.
                            vios_wrapper = VIOSEntryWrapper(vios_resp.entry)
                            scsi_map_entries = storage_map.find_map_entries(
                                vios_wrapper.get_virtual_scsi_mappings(),
                                lpar_id, with_valid_client_adapter=True)
                            scsi_storage = storage_map.parse_and_populate(
                                vios_resp.entry.element,
                                scsi_map_entries)
                            if scsi_storage is not None:
                                disk_info = scsi_storage.disk_info_by_lpar.get(
                                    lpar_id)
                                if disk_info is not None:
                                    for disk in disk_info:
                                        storage_type = disk.storage_type
                                        if storage_type == "PhysicalVolume":
                                            num_adapters = 0
                                            LOG.info(_("The RMC connection is "
                                                       "not needed because "
                                                       "the virtual adapter "
                                                       "has already been "
                                                       "created for the "
                                                       "virtual machine %s.")
                                                     % lpar_name)

                        else:
                            num_adapters = 1
                    else:
                        # for npiv , we need to check whether the physical port
                        # is already mapped,
                        # if so, no need to create fc adapter pair
                        mapped = {}
                        num_adapters = 0
                        fabric_map = ports_lookup[vios_uuid]
                        # loop over fabrics in a pre-determined order
                        for fabric in fab_order[vios_uuid]:
                            if fabric not in fabric_map.keys():
                                LOG.debug("Skip fabric %s for VIOS %s." %
                                          (fabric, vios_name))
                                continue
                            # extract the port names for the fabric
                            ports = [p['name'] for p in fabric_map[fabric]]
                            LOG.debug('ports %s' % ports)
                            # see whether we can find any of the ports already
                            # mapped
                            fc_map = vios_resp.entry.element.find(
                                storage_map_processor.ROOT +
                                storage_map_processor.FC_MAPPINGS)
                            ret_entries = storage_map.find_map_entries(
                                fc_map, lpar_id, backing_dev=ports,
                                with_valid_client_adapter=True,
                                without_port_name=False)
                            # We don't need to check if ret_entries is at least
                            # ports_per_fabric_npiv cardinality. See later
                            # comment were this mapped setting is checked.
                            if ret_entries:
                                mapped[fabric] = True
                                host_fabric_mapped.add(fabric)
                                vios_fabric_mapped.add(fabric)
                            else:
                                mapped[fabric] = False
                                num_adapters +=\
                                    scg_info['scg'].ports_per_fabric_npiv
                            if npiv_fab_req == "exactly_one":
                                LOG.debug("Only need fabric %s for fabric req"
                                          " %s." % (fabric, npiv_fab_req))
                                break  # Stop counting num_adapters needed

                    ret_data = None
                    create_adapter_err = False
                    if num_adapters > 0:
                        if not dlpar_adapter_cap:
                            rmc_err = exception.\
                                IBMPowerVMNoRMCConnectivityForStorage(
                                    instance_name=lpar_name)
                            err_msgs.append(six.text_type(rmc_err))
                            create_adapter_err = True
                            continue
                        try:
                            ret_data = (
                                storage_map.check_and_create_adapter_pair(
                                    oper, host.uuid, vios_uuid, lpar_href,
                                    lpar_id, num_adapter=num_adapters))

                            if ret_data:
                                # save wwpn and slot no into
                                # instance_system_metadata
                                wwpn_slot_map = ret_data.get('wwpn_slot_map')
                                if wwpn_slot_map:
                                    migrate_utils.set_lpar_slot_info_from_map(
                                        instance, vios_uuid, wwpn_slot_map)

                        except Exception as create_err:
                            # fail to create adapter
                            err_msgs.append(six.text_type(create_err))
                            create_adapter_err = True
                            continue
                    if ret_data:
                            vios_resp = ret_data.get('new_vios_entry')
                    else:
                        vios_resp = None

                    if not (create_adapter_err or prep_for_vscsi):
                        # keep track of the storage we have already attached
                        # within the method.
                        # this information is used later to rollback the
                        # operation
                        devs_attached[vios_uuid] = []

                        # map physical fc
                        # perform map on each fabric
                        fabric_map = ports_lookup[vios_uuid]
                        attach_err = False
                        # loop over fabrics in the same pre-determined order
                        for fabric in fab_order[vios_uuid]:
                            if fabric not in fabric_map.keys():
                                LOG.debug("Skip fabric %s for VIOS %s." %
                                          (fabric, vios_name))
                                continue

                            # if 1 of the ports in the fabric is already mapped
                            # then no-op.
                            # Note that this behavior should still be OK for
                            # 1.3 since you cannot onboard a VM into a custom
                            # SCG. So the default SCG setting will only
                            # require 1 port per fabric. For all other VMs,
                            # if one port is mapped, the required number is.
                            if mapped[fabric]:
                                if npiv_fab_req == "exactly_one":
                                    LOG.debug("exactly_one already mapped.")
                                    break
                                LOG.debug("Fabric %s already mapped. On to "
                                          "next." % fabric)
                                continue

                            # We need to pick ports_per_fab physical ports
                            # from the fabric.
                            # If we fail to attach, move on to the next port in
                            # the same fabric.
                            err_backing_devs = []
                            attached_fabric = False
                            for phy_port in fabric_map[fabric]:
                                port_name = phy_port['name']
                                LOG.debug('port_name to map %s' % port_name)
                                try:
                                    vios_resp = storage_map.attach_storage(
                                        oper, host.uuid, vios_uuid, lpar_id,
                                        None, port_name, storage_type, None,
                                        resp=vios_resp, lpar_uuid=lpar_uuid)

                                    if vios_resp:
                                        devs_attached[vios_uuid].append(
                                            port_name)
                                        if len(devs_attached[vios_uuid]) >=\
                                                ports_per_fab:
                                            LOG.debug("Met fabric attach req.")
                                            attached_fabric = True
                                            host_fabric_mapped.add(fabric)
                                            vios_fabric_mapped.add(fabric)
                                            break
                                except Exception as attach_e:
                                    LOG.debug('Failed to map to port %s: %s'
                                              % (port_name,
                                                 six.text_type(attach_e)))
                                    err_backing_devs.append(port_name)
                            # make sure we attached to the fabric
                            if not attached_fabric:
                                ex_args = {'backing_dev': err_backing_devs,
                                           'instance_name': lpar_name,
                                           'reason': ''}
                                err_msgs.append(
                                    six.text_type(
                                        exception.IBMPowerVMVolumeAttachFailed(
                                            **ex_args)))
                                fabric_mapped_err_msgs.append(
                                    six.text_type(
                                        exception.IBMPowerVMVolumeAttachFailed(
                                            **ex_args)))
                                attach_err = True
                            elif npiv_fab_req == "exactly_one":
                                LOG.debug("exactly_one new vFC to phy mapping "
                                          "satisfied for fabric %s" % fabric)
                                break  # don't go to the next fabric.
                        if (not attach_err and (npiv_fab_req != 'per_vios' or
                                                (len(vios_fabric_mapped) >=
                                                 fabrics_needed))):
                            num_vios_met_req += 1
                    elif not create_adapter_err and prep_for_vscsi:
                        num_vios_met_req += 1
                if num_vios_met_req < min_vios_per_vol:
                    raise common_exception.IBMPowerVMAllocStgForMinVIOSFailed(
                        **dict(instance_name=lpar_name,
                               min_vios=min_vios_per_vol,
                               reasons=','.join
                               ([six.text_type(err_msgs.index(x))
                                 + ") " + six.text_type(x) for x in
                                 err_msgs])))
                # For the per_host case, we need to ensure that the number
                # of fabrics mapped across the host meets the requirement.
                if npiv_fab_req == 'per_host' and len(host_fabric_mapped) <\
                        fabrics_needed:
                    fabric_mapped_err_msgs = ','.join(
                        [six.text_type(fabric_mapped_err_msgs.index(x))
                         + ") " + six.text_type(x) for x in
                         fabric_mapped_err_msgs])
                    raise exception.IBMPowerVMAllocStgForDualFabricFailed(
                        **dict(instance_name=lpar_name,
                               reasons=fabric_mapped_err_msgs))
            except Exception as e:
                LOG.error(_('Could not create adapter to virtual machine: %s')
                          % lpar_name)
                # rollback storage attached within the call
                try:
                    for vios_uuid, devs in devs_attached.items():
                        if devs:
                            storage_map.detach_and_delete_adapter(
                                oper, host.uuid, vios_uuid,
                                backing_dev=devs, detach_only=True)
                except Exception as rollbackErr:
                    LOG.debug('rollback attach failed %s' %
                              six.text_type(rollbackErr))
                LOG.exception(_('Error: %s') % six.text_type(e))
                raise exception.IBMPowerVMAdapterCreateFailed(
                    instance_name=lpar_name, reason=six.text_type(e))

    _call_prep_instance_for_attach()


def _config_adapter_capable(lpar_entry):
    """
    Check whether we can add or remove adapter for the LPAR.
    :param lpar_entry: lparentry_wrapper
    :return boolean
    """
    # in order to the create or remove adapter, LPAR either need to be
    # powered off or running with active RMC connection
    lpar_state = lpar_entry.get_lpar_state()
    if lpar_state == power_state.SHUTDOWN:
        return True
    else:
        lpar_type = lpar_entry.get_lpar_type()
        # For IBM i ,there is no need to check rmc state
        if IBMI_OS_TYPE == lpar_type:
            return True
        # extract rmc state
        lpar_rmc_state = lpar_entry.get_rmc_state()
        if lpar_rmc_state == 'active':
            return True
    return False


@logcall
def _close_vterm(k2oper, host_uuid, lpar_uuid):
    """
    Close vterm on the lpar
    :param k2oper: K2 operator
    :param host_uuid: host uuid
    :param lpar_uuid: partition uuid
    """
    # SW390979: GET LogicalPartition job with ManagedSystem UUID in the URI
    # is not supported
    k2resp = k2oper.read(k2const.LPAR,
                         lpar_uuid, suffixType=k2const.SUFFIX_TYPE_DO,
                         suffixParm=k2const.SUFFIX_PARM_CLOSE_VTERM)
    job_wrapper = JobEntryWrapper(k2resp.entry)

    try:
        job_wrapper.run_job(k2oper, lpar_uuid, host_uuid=host_uuid)
    except Exception as e:
        # just log the error.
        # delete_lpar will handle error from HMC if we failed to delete lpar
        LOG.debug('Failed to close vterm %s' % six.text_type(e))
