# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""
Implements a multi-threaded orchestrator to handle deploy operations.

Based on threading model.

"""

from eventlet import greenthread
from powervc_nova import logcall
from nova.context import get_admin_context
from nova.objects import flavor as flavor_obj
from powervc_nova.virt.ibmpowervm.hmc.k2oprguard import K2OperGuard
import lpar
from virtualoptical import VirtualOptical
from nova import context
from nova import objects
from oslo_utils import excutils
from oslo_log import log as logging
from oslo_config import cfg
from powervc_nova.conductor import api as conductor
from powervc_nova.objects.compute import dom as compute_dom
from powervc_nova.virt.ibmpowervm.common.constants \
    import CONNECTION_TYPE_NPIV, CONNECTION_TYPE_SSP, CONNECTION_TYPE_VSCSI
import powervc_nova.virt.ibmpowervm.hmc.k2.constants as k2const
from powervc_nova.storage.storage_mapping_api import StorageMapping
from powervc_nova.virt.ibmpowervm.common import progress_updater, migrate_utils
from powervc_nova.virt.ibmpowervm.common.lpar_state_checker \
    import check_lpar_state
from powervc_nova.virt.ibmpowervm.common import volume_utils
from powervc_nova.virt.ibmpowervm.common.constants import IBMI_OS_TYPE
from powervc_nova.virt.ibmpowervm.hmc.instance_type import InstanceType

from powervc_nova.virt.ibmpowervm.hmc.k2.viosentry_wrapper \
    import VIOSEntryWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2.lparentry_wrapper \
    import LPAREntryWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2 import storage_map_processor as smp
from powervc_nova.virt.ibmpowervm.hmc.k2 import scsi_map_processor
from powervc_nova.virt.ibmpowervm.hmc.mediarepository import MediaRepository
from powervc_nova.volume import cinder as volume
from powervc_nova.common.threadgroup import GreenThreadGroup
from powervc_nova import _

import random
import six
from exception import IBMPowerVMSNoStorageConnectivityFound as NoConn
import exception

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


# A a dict of points in the deploy flow and the associated percent complete
# that is reported to the progress_updater
progress_markers = {'enter': 10,
                    'vm_created': 30,
                    'scg_conn_info': 40,
                    'veas_created': 50,
                    'stg_adapters_created': 70,
                    'load_source_updated': 80,
                    'volume_attach_complete': 95
                    }


@logcall
def start_lpar(conductor_api, host, instance, lpar_uuid):

    with K2OperGuard(host=host) as opr:
        lpar.power_on_off(opr, lpar_uuid,
                          k2const.SUFFIX_PARM_POWER_ON,
                          host)
        # At this point the VM power state is either RUNNING or CRASHED
        # Call to get the actual value
        current_power_state = lpar.get_lpar_state(opr,
                                                  host,
                                                  lpar_uuid)
    LOG.debug('Current power state is %s' % str(current_power_state))
    # Update the instance with the correct power state before
    # proceeding to check the DLPAR / RMC connectivity
    try:
        instance.power_state = current_power_state
        instance.save()
    except Exception as e:
        LOG.exception(_('Error updating Power State: %s') % six.text_type(e))

    LOG.debug('Current power state updated.')

    check_lpar_state(context.get_admin_context(), instance, host, lpar_uuid)


class DeployOrchestrator():

    def __init__(self, context, virtapi, host, instance, image_meta,
                 injected_files, admin_password, network_info,
                 block_device_info, validate_volumes=None,
                 attach_volumes=None, validate_attachment=None, flavor=None):

        self.context = context
        self.conductor_api = conductor.PowerVCConductorAPI()
        self.host = host
        self.instance = instance
        self.image_meta = image_meta
        self.injected_files = injected_files
        self.admin_password = admin_password
        self.network_info = network_info
        self.block_device_info = block_device_info
        self.validate_volumes = validate_volumes
        self.attach_volumes = attach_volumes
        self.validate_attachment = validate_attachment
        self.flavor = flavor
        self.optical_vios_uuid = None
        self.vios_uuids = []
        self.boot_connectivity_type = None
        # Dict of vios_uuid to physical fibre channel to map
        self.fc_ports = {}
        self.progress = progress_updater.ProgressUpdater(context,
                                                         instance['uuid'])
        self.host_stats = None
        self.lpar_entry = None
        self.iso_attach_success = False
        self.vopt_media = None
        self.slot_mapping = {smp.SCSI: set(), smp.VFC: set()}
        self.npiv_fab_req = None
        self.port_num_req = None
        self.ports_per_fab = None

        self.storage_conn_adp_info = {}

        self.CONN_TYPE_MAP = {CONNECTION_TYPE_SSP: smp.SCSI,
                              CONNECTION_TYPE_NPIV: smp.VFC,
                              CONNECTION_TYPE_VSCSI: smp.SCSI}
        self.srr_enabled = False

    @logcall
    def vif_plug(self):
        """
        Create network on VIOS.
        """
        self.host.plug_vifs(self.instance, self.network_info)

    @logcall
    def lpar_create(self):
        """
        Create LPAR
        """
        LOG.debug('Orchestrator: Starting LPAR create thread')

        admin_context = get_admin_context(read_deleted='yes')
        if self.flavor is None:
            self.flavor = \
                flavor_obj.Flavor.get_by_id(admin_context,
                                            self.instance['instance_type_id'])
        # Create a driver instance type object from the db instance type
        instance_type = InstanceType(self.instance, self.flavor)
        # Validate the requested instance_type
        instance_type.validate(self.host, len(self.network_info))
        if self.host_stats is None:
            self.host_stats = self.host.get_host_stats_internal()
        # extract processor pool info
        proc_pools = self.host.get_shared_proc_pools()
        new_lpar = lpar.build_lpar_template(self.host_stats,
                                            self.instance, instance_type,
                                            proc_pools)
        # extract srr capability from instance_type
        srr_cap = instance_type.extra_specs['powervm:srr_capability']
        if srr_cap and srr_cap.lower() == 'true':
            self.srr_enabled = True
        name = self.instance['name']
        LOG.debug('Orchestrator: creating LPAR %s' % name)
        lpar_resp = None
        with K2OperGuard(host=self.host) as oper:
            try:
                lpar_resp = oper.create(new_lpar, 'ManagedSystem',
                                        self.host.uuid, 'LogicalPartition')
            except Exception as e:
                with excutils.save_and_reraise_exception():
                    LOG.error(_('Cannot create virtual machine: %s') % name)
                    LOG.exception(_('Error creating instance: %s')
                                  % six.text_type(e))

        # Get the PVM UUID and add it to the UUID cache
        # use K2 level caching
        uuid = lpar_resp.entry.properties['id']
        # issue 8181
        # Update the db first in case the periodic task inventory_instances
        # ran in between the cache & db update
        # We need to set the K2 UUID on the Server instance in the DB
        # lpar_values = dict(power_specs=dict(vm_uuid=uuid))
        inst_fact = compute_dom.ManagedInstanceFactory.get_factory()
        managed_inst = inst_fact.find_instance_by_uuid(
            admin_context, self.instance['uuid'])
        managed_inst.power_specs = dict(vm_uuid=uuid)
        managed_inst.save()

        self.host.pvm_uuid_cache.add(name, uuid)

        # Save lpar info for use later.
        self.lpar_uuid = uuid
        self.lpar_entry = lpar_resp.entry

    @logcall
    def network_adapters(self):
        """
        Create Network adapters.
        """

        # RTC 6437 - the VEA on the VIOS must be created before the VEA on the
        # LPAR can be created.  This next call waits for the VIOS VEA.
        self.tg.wait(self.vif_t)

        # retrieve host stats if not available
        if self.host_stats is None:
            self.host_stats = self.host.get_host_stats_internal()

        lpar.attach_interfaces(self.context,
                               self.host,
                               self.host_stats,
                               self.lpar_entry,
                               self.instance,
                               self.network_info,
                               True)

    @logcall
    def _stg_adapter_single_vios(self, vios_uuid, lpar_href):
        storage_map = smp.StorageMapProcessor()
        with K2OperGuard(host=self.host) as oper:
            try:
                num_scsi = 0
                num_fc = 0

                for connection_type in self.storage_conn_adp_info:
                    if 'adapter_pair_counts' in \
                        self.storage_conn_adp_info[connection_type] and\
                        vios_uuid in self.storage_conn_adp_info[
                            connection_type]['adapter_pair_counts']:
                        if self.CONN_TYPE_MAP[connection_type] == smp.SCSI:
                            num_scsi = self.storage_conn_adp_info[
                                connection_type]['adapter_pair_counts'][
                                    vios_uuid]
                        else:
                            num_fc = self.storage_conn_adp_info[
                                connection_type]['adapter_pair_counts'][
                                    vios_uuid]

                # if this is the optical VIOS and SCSI isn't in the
                # connectivity types, then
                # we need to create a SCSI pair here for the
                # optical drive since it wouldn't have been created above.
                if vios_uuid == self.optical_vios_uuid and\
                        CONNECTION_TYPE_SSP not in \
                        self.storage_conn_adp_info and\
                        CONNECTION_TYPE_VSCSI not in \
                        self.storage_conn_adp_info:
                    num_scsi = 1

                replacement_text = {'num_scsi': num_scsi,
                                    'num_fc': num_fc,
                                    'lpar': lpar_href,
                                    'vios': vios_uuid}
                ret_data = dict(vios_uuid=vios_uuid,
                                num_scsi=num_scsi,
                                num_fc=num_fc,
                                met_req_num=True)

                if num_scsi > 0 or num_fc > 0:
                    try:
                        ret_slot = storage_map.create_mixed_adapter_pairs(
                            oper, self.host.uuid, vios_uuid,
                            lpar_href, self.lpar_id, num_scsi, num_fc,
                            ext_slot_info=True,
                            fc_backing_devs=self.fc_ports.get(vios_uuid))

                        # save wwpn and slot no into instance_system_metadata
                        if ret_slot and 'wwpn_slot_map' in ret_slot:
                            wwpn_slot_map = ret_slot.get('wwpn_slot_map')
                            migrate_utils.set_lpar_slot_info_from_map(
                                self.instance, vios_uuid, wwpn_slot_map)

                        # retrieve slot number info for IBM i and then used for
                        # load source and alternated load source
                        for conn_type in ret_slot['slot_info'].keys():
                            slot_number = ret_slot['slot_info'][conn_type]
                            self.slot_mapping[conn_type].update(slot_number)

                    except Exception as create_err:
                        # check whether adapters are created
                        ext_props = [k2const.VIOS_VSCSI_MAP_EXT_PROP,
                                     k2const.VIOS_VFC_MAP_EXT_PROP]
                        vios_resp = oper.read(k2const.VIOS, vios_uuid,
                                              xag=ext_props)
                        stg_map = {smp.SCSI_MAPPINGS: num_scsi,
                                   smp.FC_MAPPINGS: num_fc}
                        if vios_resp is not None:
                            for map_type, num_adapters in stg_map.items():
                                mappings = vios_resp.entry.element.find(
                                    smp.ROOT + map_type)
                                if mappings is not None:
                                    adapters = storage_map.find_map_entries(
                                        mappings, self.lpar_id,
                                        with_valid_client_adapter=True)
                                    num = len(adapters) if mappings else 0
                                    # if the number of adapter is insufficient
                                    # check mapping type for handling.
                                    if num < num_adapters:
                                        LOG.debug('Created %(num)d %(type)s '
                                                  'adapter pairs.' %
                                                  dict(num=num, type=map_type))
                                        # raise create_err
                                        if map_type == smp.SCSI_MAPPINGS:
                                            # if not able to create adapter
                                            # for iso attach
                                            if (vios_uuid ==
                                                    self.optical_vios_uuid and
                                                    num < 1):
                                                raise create_err
                                            ret_data['num_scsi'] = num
                                        else:
                                            ret_data['num_fc'] = num
                                        ret_data['met_req_num'] = False
                        else:
                            # if not able to create adapter
                            # for iso attach
                            if (vios_uuid == self.optical_vios_uuid):
                                raise create_err
                            ret_data['num_scsi'] = 0
                            ret_data['num_fc'] = 0
                            ret_data['met_req_num'] = False
                        if not ret_data['met_req_num']:
                            ret_data['msg'] = six.text_type(create_err)

                # Create adapter pairs
                LOG.debug('Created %(num_scsi)d SCSI pairs and '
                          '%(num_fc)s FC pairs for LPAR '
                          'href: %(lpar)s and VIOS uuid %(vios)s' %
                          replacement_text)
                return ret_data
            except Exception:
                with excutils.save_and_reraise_exception():
                    LOG.error(_('Cannot create adapters: %s') % lpar_href)

    @logcall
    def stg_adapters(self, vea_t):

        # Get the lpar id
        lpar_wrapper = LPAREntryWrapper(self.lpar_entry)
        self.lpar_id = lpar_wrapper.get_lpar_id()
        LOG.debug("Adapter pair creation starting on lpar %s" % self.lpar_id)

        # Wait until VIOS info is populated.
        self.tg.wait(self.vios_t)
        self.progress.set_progress(progress_markers['scg_conn_info'])

        # Check to see if there were any errors in our thread group...
        if GreenThreadGroup.thread_should_bailout():
            return

        # Wait for network adapter to complete.
        self.tg.wait(vea_t)
        self.progress.set_progress(progress_markers['veas_created'])

        # Check to see if there were any errors in our thread group...
        if GreenThreadGroup.thread_should_bailout():
            return

        with K2OperGuard(host=self.host) as oper:
            try:
                lpar_href = oper.buildhref(rootType='LogicalPartition',
                                           rootId=self.lpar_uuid)

                LOG.debug('LPAR href: %s' % lpar_href)

            except Exception:
                with excutils.save_and_reraise_exception():
                    LOG.error(_('Cannot get virtual machine href: %s') %
                              self.lpar_uuid)

        # Create adapters on each VIOS
        # TODO: use greenpool?
        # Due to HMC defect SW225231, add ibmpowervm_stg_adapters_threaded
        # configure option to allow adapter pair creation single threaded
        # or multithreaded. The default for ibmpowervm_stg_adapters_threaded
        # is currently set to True as SW225231 has been integrated
        if CONF.ibmpowervm_stg_adapters_threaded:
            # Create our own thread group so we're the only ones waiting
            # on our children and their exceptions are passed to us.
            stg_adapter_tg = GreenThreadGroup()
            try:
                threads = []
                for vios_uuid in self.vios_uuids:
                    t = stg_adapter_tg.spawn(self._stg_adapter_single_vios,
                                             vios_uuid, lpar_href)
                    threads.append(t)
                    LOG.debug("Spawned adapter thread for vios_uuid %s"
                              % vios_uuid)
                reasons = []
                is_scsi = True if (CONNECTION_TYPE_SSP
                                   in self.storage_conn_adp_info or
                                   CONNECTION_TYPE_VSCSI in
                                   self.storage_conn_adp_info) else False
                is_vfc = True if(CONNECTION_TYPE_NPIV
                                 in self.storage_conn_adp_info) else False
                success_num_vios_scsi = 0
                success_num_vios_vfc = 0
                min_vios_per_vol_scsi = 0
                min_vios_per_vol_vfc = 0

                for thread in threads:
                    adapter_info = stg_adapter_tg.wait(thread)
                    LOG.debug('adapter_info %s' % adapter_info)
                    if adapter_info['met_req_num']:
                        if is_scsi and adapter_info['num_scsi'] > 0:
                            success_num_vios_scsi += 1
                        if is_vfc and self.npiv_fab_req == 'per_vios':
                            if adapter_info['num_fc'] >= self.port_num_req:
                                success_num_vios_vfc += 1
                        elif is_vfc and self.npiv_fab_req != 'per_vios':
                            if adapter_info['num_fc'] >= self.ports_per_fab:
                                success_num_vios_vfc += 1
                        LOG.debug("Processed thread. VFC vios count: %d" %
                                  success_num_vios_vfc)
                    elif adapter_info.get('msg'):
                        reasons.append(adapter_info.get('msg'))
                # retrieve min_vios_per_vol for
                # each connectivity type
                for connection_type in self.storage_conn_adp_info.keys():
                    if self.CONN_TYPE_MAP[connection_type] == smp.SCSI:
                        min_vios_per_vol_scsi = self.storage_conn_adp_info[
                            connection_type]['min_vios_per_vol']
                    elif self.CONN_TYPE_MAP[connection_type] == smp.VFC:
                        min_vios_per_vol_vfc = self.storage_conn_adp_info[
                            connection_type]['min_vios_per_vol']

                if success_num_vios_scsi < min_vios_per_vol_scsi:
                    raise exception.IBMPowerVMInsufficientVIOSForStg(
                        **dict(instance_name=self.instance['name'],
                               min_vios=min_vios_per_vol_scsi,
                               avail_vios=success_num_vios_scsi,
                               reasons=reasons))

                if success_num_vios_vfc < min_vios_per_vol_vfc:
                    raise exception.IBMPowerVMInsufficientVIOSForStg(
                        **dict(instance_name=self.instance['name'],
                               min_vios=min_vios_per_vol_vfc,
                               avail_vios=success_num_vios_vfc,
                               reasons=reasons))

            except Exception as e:
                with excutils.save_and_reraise_exception():
                    LOG.exception(_('Error creating storage adapters: %s') %
                                  six.text_type(e))
                    # Kill all threads we started
                    stg_adapter_tg.kill()
        else:
            for vios_uuid in self.vios_uuids:
                LOG.debug("START: adapter creation: vios_uuid %s" % vios_uuid)
                self._stg_adapter_single_vios(vios_uuid, lpar_href)
                LOG.debug("END: adapter creation: vios_uuid %s" % vios_uuid)

        LOG.debug("Adapter pair creation completed.")

    @logcall
    def update_load_source(self):
        """
        Update load source info for IBM i
        """
        load_source = None
        alternated_load_source = None
        slot_numbers = set()
        if (self.CONN_TYPE_MAP[self.boot_connectivity_type] == smp.SCSI):
            slot_numbers = self.slot_mapping[smp.SCSI]
        else:
            slot_numbers = self.slot_mapping[smp.VFC]

        if len(slot_numbers) > 0:
            load_source = slot_numbers.pop()
        if len(slot_numbers) > 0:
            alternated_load_source = slot_numbers.pop()

        # self.host_stats has been retrieved in network_adapters
        lpar.update_load_source(self.instance, self.host,
                                self.lpar_uuid, load_source,
                                alternated_load_source)

    @logcall
    def stg_attach(self, adapters_t):
        """
        Attach volumes to VM prior to boot.
        """

        # Wait until the storage adapters are created.
        self.tg.wait(adapters_t)

        self.attach_volumes(self.context, self.instance)
        self.validate_attachment(self.context, self.instance['uuid'])

        return

    @logcall
    def verify_media_repository(self):
        # Get VIOS info
        self.tg.wait(self.vios_t)

        rep = MediaRepository(self.host, self.optical_vios_uuid)

        if not rep.exists():
            rep.create()
            LOG.debug("orchestrator: repository doesn't exist; creating.")
        else:
            LOG.debug("orchestrator: existing media repository found.")

        return rep

    @logcall
    def iso_create_transfer(self, media_rep_t, vea_t):
        """
        Verify media repository availability (create if necessary)
        and transfer the ISO once it's created.
        """

        # VIOS and media repository are available
        self.tg.wait(self.vios_t)
        rep = self.tg.wait(media_rep_t)
        self.tg.wait(vea_t)

        # TODO verify free space on repository
        self.vo = VirtualOptical()
        self.vopt_media = \
            self.vo.transmit_configuration_on_optical(self.host, self.instance,
                                                      self.injected_files,
                                                      self.admin_password,
                                                      self.network_info,
                                                      rep, self.image_meta)

    @logcall
    def attach_iso(self, iso_transfer_t):
        # Wait for storage adapters to be created
        self.tg.wait(iso_transfer_t)
        if self.vopt_media:
            self.vo.attach_to_vhost(self.host, self.vopt_media, self.lpar_id,
                                    self.optical_vios_uuid,
                                    lpar_uuid=self.lpar_uuid)
        self.iso_attach_success = True

    def _determine_scg_conn_types(
            self, boot_conn_types, data_conn_types, data_volumes):
        '''
        This method is called only when length of boot/data connection types
        is more than 1.In this case we will
        loop through all bdms and
            determine storage_type for each volume,
            if strage_type is 'cluster'
                 if volume is data volume
                     if data_conn_type has ssp
                     then add it to list else raise
                 if volume is boot volume
                     if boot_conn_type has ssp
                     then add it to list else raise
            else
                if volume is data volume
                    add non cluster connection type(NPIV/VSCSI) in
                    data_conn_types to the list
                if volume is boot volume
                    add non cluster connection type(NPIV/VSCSI) in
                    boot_conn_types to the list
        Note:Here the assumption is VSCSI and NPIV can not exist together.
        so possible combinations are [SSP,VSCSI][SSP,NPIV]
        '''
        LOG.info(_("Boot connection types : %(boot_conn_types)s "
                   "Data connection types : %(data_conn_types)s "
                   "Data volumes are :  %(data_volumes)s")
                 % {'boot_conn_types': boot_conn_types,
                    'data_conn_types': data_conn_types,
                    'data_volumes': data_volumes})
        vol_api = volume.PowerVCVolumeAPI()
        # Get the list of BDMs for this VM
        bdms = objects.BlockDeviceMappingList.\
            get_by_instance_uuid(self.context, self.instance['uuid'])
        for bdm in bdms:
            stg_type = ''
            volume_id = ''
            if 'volume_id' not in bdm:
                continue
            try:
                volume_id = bdm['volume_id']
                stg_type = vol_api.get_storage_type_for_volume(
                    self.context, volume_id)
            except Exception:
                LOG.exception(_("Unable to retrieve volume %s"
                                % bdm['volume_id']))
            if stg_type == 'cluster':
                if volume_id in data_volumes:
                    if CONNECTION_TYPE_SSP in data_conn_types:
                        self.storage_conn_adp_info[CONNECTION_TYPE_SSP] \
                            = {}
                    else:
                        LOG.error(_("The deploy contained data volumes but "
                                    "no data connectivity was found"))
                        raise exception.\
                            IBMPowerVMInvalidStorageConnectivityInfo(
                                instance_name=self.instance['name'])
                else:
                    if CONNECTION_TYPE_SSP in boot_conn_types:
                        self.storage_conn_adp_info[CONNECTION_TYPE_SSP] \
                            = {}
                        # store boot connectivity for IBMi
                        self.boot_connectivity_type = CONNECTION_TYPE_SSP
                    else:
                        LOG.error(_("There were no valid boot connectivity"
                                    " types found for this deploy"))
                        raise exception.\
                            IBMPowerVMInvalidStorageConnectivityInfo(
                                instance_name=self.instance['name'])
            elif stg_type == 'fc':
                if volume_id in data_volumes:
                    for conn_type in data_conn_types:
                        if conn_type != CONNECTION_TYPE_SSP:
                            self.storage_conn_adp_info[conn_type] = {}
                else:
                    for conn_type in boot_conn_types:
                        if conn_type != CONNECTION_TYPE_SSP:
                            self.storage_conn_adp_info[conn_type] = {}
                            # store boot connectivity for IBMi
                            self.boot_connectivity_type = conn_type

    @logcall
    def populate_stg_conn_info(self):
        """
        Get VIOS and VIOS connection information for this VM
        """

        stg_api = StorageMapping()
        # The default is CHK_EITHER type here since later checks will ask
        # for boot or data specifically.
        conn_info = \
            stg_api.get_connectivity_info(self.context,
                                          self.instance['uuid'],
                                          CONF.host,
                                          self.instance)
        # extract scg class to six.text_type to avoid UnicodeEncodeError
        scg_class = conn_info.pop('scg')
        LOG.debug('scg_info: %s, scg_class: %s' %
                  (conn_info, six.text_type(scg_class)))
        conn_info['scg'] = scg_class
        scg_conn_types = []

        boot_conn_types = conn_info.get('boot_conn_types')
        data_conn_types = conn_info.get('data_conn_types')

        # retrieve data volumes if any attached
        data_volumes = volume_utils.PowerVCNovaVirtDriverVolumeAPI().\
            get_instance_volumes(self.context, self.instance,
                                 volume_category='data')

        if len(boot_conn_types) == 1 and len(data_conn_types) == 1:

            # if instance has data volumes then take effective
            # connectivity types
            if len(data_volumes) > 0:
                scg_conn_types.extend(data_conn_types)
                scg_conn_types.extend(boot_conn_types)
            else:
                scg_conn_types = boot_conn_types

            LOG.debug("Connection types from SCG %s" % scg_conn_types)
            for conn_type in scg_conn_types:
                self.storage_conn_adp_info[conn_type] = {}

            # store boot connectivity for IBMi
            self.boot_connectivity_type = conn_info.get('boot_conn_types')[0]

        elif len(data_volumes) > 0 and len(data_conn_types) == 0:
            LOG.error(_("The deploy contained data volumes but "
                        " no data connectivity was found"))
            raise exception.IBMPowerVMInvalidStorageConnectivityInfo(
                instance_name=self.instance['name'])
        elif len(boot_conn_types) == 0:
            LOG.error(_("There were no valid boot connectivity types"
                        " found for this deploy "))
            raise exception.IBMPowerVMInvalidStorageConnectivityInfo(
                instance_name=self.instance['name'])
        else:
            self._determine_scg_conn_types(
                boot_conn_types, data_conn_types, data_volumes)

        if not self.storage_conn_adp_info:
            raise NoConn(scg_conn_types=scg_conn_types)

        LOG.debug("Chosen connectivity type is %s" %
                  self.storage_conn_adp_info)

        # for each required connection type
        # get all vios names which supports
        # the connection type
        for connection_type in self.storage_conn_adp_info.keys():
            vios_names = [vio['name'] for vio in conn_info.get('vios_list')
                          if connection_type in vio.get('connectivity_types')]
            # extract the minimum vios required per vol
            # default to all vios since that's the default in previous release
            self.storage_conn_adp_info[connection_type] =\
                {'vios_names': vios_names,
                 'min_vios_per_vol': conn_info.get('min_vioses_per_vol',
                                                   len(vios_names))}
        all_conn_vios_uuids = []
        # use K2 level caching
        # for collected vios names
        # collect vios uuids from K2
        with K2OperGuard(host=self.host) as k2oper:
            vioses_resp = k2oper.read(k2const.MGT_SYS, rootId=self.host.uuid,
                                      childType=k2const.VIOS, age=0,
                                      xag=k2const.ALL_VIOS_EXT_PROPS)
            vios_entries = vioses_resp.feed.entries

            # Create list of UUIDs of VIOSes in the SCG
            name_to_uuid = {}
            for connection_type in self.storage_conn_adp_info.keys():
                vios_uuids = []
                name_to_uuid_temp = {}
                vios_names = self.storage_conn_adp_info[
                    connection_type]['vios_names']
                for v in vios_entries:
                    vew = VIOSEntryWrapper(v)
                    if vew.get_vios_name() in vios_names:
                        vios_uuids.append(v.properties['id'])
                        name_to_uuid_temp[
                            vew.get_vios_name()] = v.properties['id']

                # the vios_list returned from scg can contain smaller set of
                # vioses
                if len(vios_uuids) < len(vios_names):
                    missing_vioses = list(set(vios_names) -
                                          set(name_to_uuid_temp.keys()))
                    LOG.error(_('Unable to find VIOSes %s on HMC.')
                              % missing_vioses)
                    k2oper.dump_logs()
                    raise exception.IBMPowerVMVIOSNotFound()
                self.storage_conn_adp_info[connection_type]\
                    .update({'vios_uuids': vios_uuids})
                all_conn_vios_uuids = list(
                    set(all_conn_vios_uuids) | set(vios_uuids))
                name_to_uuid.update(name_to_uuid_temp)
            self.vios_uuids = all_conn_vios_uuids

        # collect adapter_pair_counts for
        # each connection type
        for connection_type in self.storage_conn_adp_info.keys():
            adapter_pair_counts = {}
            if connection_type == CONNECTION_TYPE_NPIV:
                # Extract fabric channel requirement we need to use.
                self.npiv_fab_req = conn_info['scg'].fabric_set_req

                # For NPIV we need to get the count of fabrics
                # for each of the VIOSes.  We use this count to create
                # one NPIV adapter per fabric.
                for vio in conn_info.get('vios_list'):
                    if CONNECTION_TYPE_NPIV not in vio.get(
                            'connectivity_types', []):
                        LOG.debug("VIOS not applicable for NPIV conn: %s" %
                                  vio['name'])
                        continue
                    if 'ports' not in vio or len(vio['ports']) == 0:
                        raise exception.IBMPowerVIOSMissingPhysicalPort(
                            dict(vios_name=vio['name']))
                    # The 'ports' dict is indexed by fabric, so sum the fabrics
                    fabric_count = len(vio['ports'])
                    vios_uuid = name_to_uuid[vio['name']]

                    # Check for the 'exactly_one' fabric requirement
                    if self.npiv_fab_req == 'exactly_one':
                        fabric_count = 1
                        LOG.debug("Force fabrics connected to by vios to 1 "
                                  "due to fabric_set_req setting.")

                    # ports_per_fab gets the 'multiplier' from the SCG
                    self.ports_per_fab = conn_info['scg'].ports_per_fabric_npiv
                    ports_per = self.ports_per_fab  # local var to modify
                    adapter_pair_counts[vios_uuid] = fabric_count * ports_per
                    self.port_num_req = adapter_pair_counts[vios_uuid]
                    LOG.debug("Number of ports required: %d" %
                              self.port_num_req)

                    # Pick the required number of ports from each fabric
                    # that we need to connect to. The SCG conn_info gives
                    # us potentially more fabrics and ports than we want to
                    # connect to. This allows the client adapter creation
                    # impl to tolerate a certain type of failure where not all
                    # vfcs we asked for were created, but enough to meet the
                    # SCG requirement. Currently though, we will ask for only
                    # what we expect to get, and appropriate failures will
                    # cause a different host to be tried. And we set up two
                    # nested loops below that we can break out of if we only
                    # want one one fabric and/or one port per fabric.
                    if fabric_count > 0:
                        self.fc_ports[vios_uuid] = []
                        for fab in vio['npiv_fab_order']:
                            if fab not in vio['ports']:
                                LOG.debug("Skip fabric %s for VIOS %s." %
                                          (fab, vio['name']))
                                continue
                            p_list = vio['ports'][fab]
                            # Reset ports_per count for each fabric
                            ports_per = self.ports_per_fab
                            for port in p_list:
                                self.fc_ports[vios_uuid].append(port['name'])
                                LOG.debug("Select port %s on adapter %s and "
                                          "fabric %s." %
                                          (port['name'], port['adapter_id'],
                                           fab))
                                ports_per -= 1
                                if ports_per == 0:
                                    LOG.debug("Satisfied the ports_per_fab.")
                                    break
                            if fabric_count == 1:
                                LOG.debug("Satisfied single fab requirement.")
                                break
                    else:
                        LOG.info(_("SCG allows VIOS %(vios)s to be skipped "
                                   "for NPIV connectivity. Fabrics and/or "
                                   "ports not available.") %
                                 {'vios': vio['name']})

                self.storage_conn_adp_info[connection_type].\
                    update({'adapter_pair_counts': adapter_pair_counts})

            else:
                vios_uuids = self.storage_conn_adp_info[
                    connection_type]['vios_uuids']
                adapter_pair_counts = {}
                for uuid in vios_uuids:
                    adapter_pair_counts[uuid] = 1
                self.storage_conn_adp_info[connection_type].\
                    update({'adapter_pair_counts': adapter_pair_counts})

        # Choose one VIOS at random to use for virtual optical
        # during this deploy operation
        common_vios_uuids = []
        for connection_type in self.storage_conn_adp_info.keys():
            vios_uuids = self.storage_conn_adp_info[
                connection_type]['vios_uuids']
            if common_vios_uuids:
                common_vios_uuids = list(
                    set(vios_uuids) & set(common_vios_uuids))
            else:
                common_vios_uuids = vios_uuids

        if self.optical_vios_uuid is None:
            self.optical_vios_uuid = random.choice(common_vios_uuids)

        LOG.debug("VIOS optical UUID %s" % self.optical_vios_uuid)
        LOG.debug("storage_conn_adp_info %s" % self.storage_conn_adp_info)

    @logcall
    def cleanup(self):
        """
        Additional cleanup for orchestrator threads
        """
        # sleep 15 seconds to allow any inprocess K2 calls to complete
        greenthread.sleep(15)
        try:
            if (not self.iso_attach_success) and (self.vopt_media):
                attached = False
                # transfer was successful, but attach was not,
                # need to double-check for possible in-flight attach
                with K2OperGuard(host=self.host) as oper:
                    ext_props = [k2const.VIOS_VSCSI_MAP_EXT_PROP]
                    vios_resp = oper.read(k2const.MGT_SYS,
                                          rootId=self.host.uuid,
                                          childType=k2const.VIOS,
                                          childId=self.optical_vios_uuid,
                                          xag=ext_props)
                    scsi_map = scsi_map_processor.SCSIMapProcessor()
                    media_list = scsi_map.find_attached_medias(
                        vios_resp.entry.element, self.lpar_id)
                    for media_name in media_list:
                        if self.vopt_media in media_name:
                            attached = True
                            break
                if (not attached):
                    self.vo.remove_vopt_media(self.host, self.vopt_media,
                                              self.optical_vios_uuid)
        except Exception:
            # catch any exceptions so other cleanup continues
            LOG.info(_('Failure during cleanup of virtual optical %s.  Manual'
                       'cleanup may be required') % self.vopt_media)

    @logcall
    def start_deploy_simple(self):
        """
        Create threads needed for deploy to complete.
        """

        for d in self.network_info:
            LOG.debug("Network info: %s" % d)

        LOG.debug("Deploy: Entering")
        self.progress.set_progress(progress_markers['enter'])
        try:
            self.tg = GreenThreadGroup()

            # Thread to monitor volume creations that were kicked off
            # in the compute manager.
            volume_create_mon = self.tg.spawn(self.validate_volumes,
                                              self.context,
                                              self.instance['uuid'])
            lpar_create = self.tg.spawn(self.lpar_create)
            self.vif_t = self.tg.spawn(self.vif_plug)

            # Need to get VIOS info from host
            self.vios_t = self.tg.spawn(self.populate_stg_conn_info)

            # - Check for media repository existence / create if needed.
            media_rep_t = self.tg.spawn(self.verify_media_repository)

            LOG.debug("orchestrator: lpar create spawned..waiting")

            # Nothing can be done here until  LPAR creation to completes.
            self.tg.wait(lpar_create)
            self.progress.set_progress(progress_markers['vm_created'])

            LOG.debug("Orchestrator: LPAR created..continuing")

            # Create network adapter.
            LOG.debug("orchestrator: starting network adapter creation")
            vea_t = self.tg.spawn(self.network_adapters)

            # ISO with environment file:
            # - Create file and ISO.
            # - Transfer to VIOS and attach to vhost.
            iso_transfer_t = self.tg.spawn(self.iso_create_transfer,
                                           media_rep_t, vea_t)

            # Create storage adapters
            LOG.debug("orchestrator: waiting for storage adapter creation")
            self.stg_t = self.tg.spawn(self.stg_adapters, vea_t)

            self.tg.wait(self.stg_t)
            LOG.debug("orchestrator: storage adapter creation done.")
            self.progress.\
                set_progress(progress_markers['stg_adapters_created'])

            # Update load source for IBM i
            lpar_wrapper = LPAREntryWrapper(self.lpar_entry)
            lpar_type = lpar_wrapper.get_lpar_type()
            if IBMI_OS_TYPE == lpar_type:
                LOG.debug("orchestrator: staring update load source.")
                load_source = self.tg.spawn(self.update_load_source)
                self.tg.wait(load_source)
                LOG.debug("orchestrator: update load source done.")
                self.progress.\
                    set_progress(progress_markers['load_source_updated'])

            # Wait for volumes to be created and moved to the 'attaching'
            # state by the create monitor thread.
            self.tg.wait(volume_create_mon)
            # Volume attachment
            LOG.debug("orchestrator: attaching volumes")
            attach_t = self.tg.spawn(self.stg_attach, self.stg_t)
            # we only will enforce iso attach to be the last step if
            # VM is deployed with SRR enabled
            if self.srr_enabled:
                self.tg.wait(attach_t)

            # Attach ISO to lpar
            LOG.debug("orchestrator: attaching ISO")
            attach_iso_t = self.tg.spawn(self.attach_iso, iso_transfer_t)
            # if SRR is enabled, we only need to wait for iso attach
            if self.srr_enabled:
                self.tg.wait(attach_iso_t)
            else:
                # otherwise, we need to wait for both attach volume and iso
                self.tg.wait([attach_t, attach_iso_t])

            self.progress.\
                set_progress(progress_markers['volume_attach_complete'])

            # Note: Any post-deploy activity that is run asynch in a separate
            # thread cannot use instance methods of DeployOrchestrator or a
            # memory leak will result since doing so keeps DeployOrchestrator
            # and all deploy threads in scope/memory along with the objects the
            # threads have in their thread local stores.

            # Do not start the LPAR if appropriate server metadata exists
            if (self.instance['metadata'].get('do_not_boot') is not None and (
                    self.instance['metadata'].get('do_not_boot').upper(
                    ) == 'TRUE')):
                LOG.info("orchestrator: not booting VM due to metadata")
            else:
                # Kickoff thread to start LPAR
                LOG.debug("orchestrator: booting VM")
                greenthread.spawn(start_lpar, self.conductor_api, self.host,
                                  self.instance, self.lpar_uuid)

            self.progress.complete_progress()

        # Cleanup after deploy failure.
        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.debug("Deploy error.")
                LOG.exception(_('Error creating instance: %s') %
                              six.text_type(e))

                # Kill all threads we started
                self.tg.kill()
                self.cleanup()
