# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from oslo_log import log as logging

from powervc_nova import logcall
from powervc_nova.virt.ibmpowervm.hmc.k2 import storage_map_processor

LOG = logging.getLogger(__name__)

LU_TYPE = storage_map_processor.LU
PV_TYPE = storage_map_processor.PV
VDISK_TYPE = storage_map_processor.VDISK
VOPT_TYPE = storage_map_processor.VOPT

key_storage_info_attrs = ['backing_device', 'storage_type', 'lu_addr',
                          'storage_uid', 'storage_udid']


@logcall
def contains_storage_types(disks, storage_types):
    """
    Check the list of disks containing storage type listed in the storage_types
    parameter.
    """
    if disks is None or len(disks) == 0:
        return False
    for d in disks:
        if d.storage_type in storage_types:
            return True
    return False


@logcall
def guess_boot_order(disks):
    """
    This method is run after we have combined the disk info from
    VirtualSCSIMappings and VirtualFibreChannelMappings.
    We use combination of virtual slot number and LUAddress to perform
    educated guess on boot disk ordering.
    It will not include VIRTUAL_OPTICAL_MEDIA storage type.
    """
    # perform educated guest on the boot disk ordering
    if disks is None:
        return None
    ordered_disks = sorted(disks,
                           key=lambda x: (x.client_slot_num, x.lu_addr))
    return [d for d in ordered_disks if d.storage_type != VOPT_TYPE]


class VIOSDiskInfo(object):
    """
    The object represents a valid backing device entry with VirtualSCSIMappings
    or VirtualFibreChannelMappings. It's used to construct VIOSStorageInfo
    object.
    """
    def __init__(self, storage_info, server_adapter, client_adapter):
        """
        Initialize disks information hosted by VIOS.
        storage_info should look like:
        {'adapter_name': ...,
         'backing_device': ...,
         'disk_size': ...,
         'lu_addr': ...,
         'target_device_name': ...,
         'storage_type': ...,
         'storage_uid':...}
        """
        # extract disk info
        # default key attributes to None if we cannot find them
        # in dict, self.backing_device, self.storage_type,
        # self.lu_addr, self.storage_uid = None
        for attr in key_storage_info_attrs:
            try:
                setattr(self, attr, storage_info.pop(attr))
            except KeyError:
                setattr(self, attr, None)
        self.storage_info = storage_info
        self.server_adapter = server_adapter
        self.client_adapter = client_adapter
        # virtual slot number will be used to determine boot disk ordering
        if self.client_adapter is not None:
            self.client_slot_num = client_adapter.slot_num

    def __eq__(self, other):
        """
        Determine whether disks are backed by same physical volume or
        logical unit.
        """
        # make sure storage_type matches
        if (self.storage_type == other.storage_type and
                self.storage_type != VDISK_TYPE and
                self.storage_type != VOPT_TYPE and
                self.storage_uid == other.storage_uid):
            return True
        return False

    def __ne__(self, other):
        """
        It's not equal if it's not backed by same physical volume or logical
        unit. It's also not equal if it's using local storage.
        """
        if (self.storage_type != other.storage_type or
                self.storage_type == VDISK_TYPE or
                self.storage_type == VOPT_TYPE or
                self.storage_uid != other.storage_uid):
            return True
        return False

    def __repr__(self):
        return ('storage_info: %s , server_adapter: %s, client_adapter: %s '
                'backing_dev: %s, storage_type: %s, lu_addr: %s, uid: %s'
                'udid: %s' %
                (self.storage_info, self.server_adapter, self.client_adapter,
                 self.backing_device, self.storage_type, self.lu_addr,
                 self.storage_uid, self.storage_udid))


class VIOSStorageInfo(object):
    """
    This object contains storage information parsed from VirtualSCSIMappings
    or VirtualFibreChannelMappings. It categorizes the disk information
    based on client partition ID. This object is particular useful for
    onboarding support.
    """
    def __init__(self, vios_lpar_id):
        """
        Initialize storage information.
        :param vios_lpar_id (int): id of VIOS hosting the storage. This
                                   attribute is not required for case where
                                   we are using VirtualStorageMap to create
                                   adapters.
        """
        self.vios_lpar_id = vios_lpar_id
        self.disk_info_by_lpar = {}

    def add(self, storage_info, server_adapter, client_adapter):
        # extract the client partition id
        new_entry = VIOSDiskInfo(storage_info, server_adapter,
                                 client_adapter)
        if client_adapter is not None:
            client_partition_id = client_adapter.lpar_id
            disks = self.disk_info_by_lpar.get(client_partition_id)
            if disks is None:
                disks = [new_entry]
                self.disk_info_by_lpar[client_partition_id] = disks
            else:
                disks.append(new_entry)

    @logcall
    def combine(self, other):
        """
        This method combine the data parse from VirtualSCSIMappings and
        VirtualFibreChannelMappings to generate all the disks info.
        """
        new_output = VIOSStorageInfo(self.vios_lpar_id)
        # iterate through the disk info and combines the list for each lpar
        for lpar_id, disks in self.disk_info_by_lpar.iteritems():
            new_output.disk_info_by_lpar[lpar_id] = disks
            if lpar_id in other.disk_info_by_lpar:
                new_output.disk_info_by_lpar[lpar_id] += (
                    other.disk_info_by_lpar[lpar_id])
        for lpar_id, disks in other.disk_info_by_lpar.iteritems():
            if lpar_id not in new_output.disk_info_by_lpar:
                new_output.disk_info_by_lpar[lpar_id] = disks
        return new_output

    @logcall
    def guess_boot_order_for_lpar(self, lpar_id):
        """
        This method is run after we have combined the disk info from
        VirtualSCSIMappings and VirtualFibreChannelMappings.
        We use combination of virtual slot number and LUAddress to perform
        educated guess on boot disk ordering.
        It will not include VIRTUAL_OPTICAL_MEDIA storage type.
        """
        # perform educated guest on the boot disk ordering
        disks = self.disk_info_by_lpar.get(lpar_id)
        return guess_boot_order(disks)

    @logcall
    def contains_storage_types_for_lpar(self, lpar_id, storage_types):
        """
        Check a partition is using storage type listed in the storage_types
        parameter.
        """
        disks = self.disk_info_by_lpar.get(lpar_id)
        return contains_storage_types(disks, storage_types)

    def __repr__(self):
        return 'disks: %s' % self.disk_info_by_lpar
