# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import os
import hashlib
import greenlet

from oslo_config import cfg

from oslo_log import log as logging
from nova.virt import configdrive
from powervc_nova import logcall

from powervc_nova.virt.ibmpowervm.common import config_iso
from powervc_nova.virt.ibmpowervm.common import exception as pvcex

import powervc_oslo.k2.k2operator as k2o
from powervc_oslo.k2.k2operator import K2Element as K2Elem
from powervc_oslo.k2.k2operator import K2Error

from powervc_nova.virt.ibmpowervm.hmc.k2 import storage_map_processor
from powervc_nova.virt.ibmpowervm.hmc.k2 import scsi_map_processor
from powervc_nova.virt.ibmpowervm.hmc.k2 import contains_msg_in_k2_err
from powervc_nova.virt.ibmpowervm.hmc import exception as hmc_ex
from powervc_nova.virt.ibmpowervm.hmc.k2 import constants as k2const
from powervc_nova.virt.ibmpowervm.hmc.k2oprguard import K2OperGuard
from powervc_nova.virt.ibmpowervm.hmc.k2.viosentry_wrapper \
    import VIOSEntryWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2.media_repository_processor \
    import MediaRepositoryProcessor

from powervc_nova import _

LOG = logging.getLogger(__name__)

CONF = cfg.CONF

DELIM = storage_map_processor.DELIM
POST_RETRIES = CONF.ibmpowervm_update_collision_retries
RMC_STATE = 'ResourceMonitoringControlState'


class VirtualOptical(object):
    '''
    classdocs
    '''
    def __init__(self):
        '''
        Constructor
        '''
        self.vopt_name = None
        self.fd = None

    @logcall
    def transmit_configuration_on_optical(self, host, instance,
                                          injected_files, admin_password,
                                          network_info,
                                          repository, image_meta):
        """
        Configure a virtual optical containing the configuration data
        for the given instance on the given VIOS.
        :param host the compute host object
        :param instance: Instance object as returned by DB layer.
        :param vios_wrapper: vios entry from K2
        """
        vopt_media = None
        try:
            if configdrive.required_by(instance):
                iso_file = config_iso.create_config_drive_iso(instance,
                                                              injected_files,
                                                              admin_password,
                                                              network_info,
                                                              image_meta)
            if iso_file:
                LOG.info(_('iso created %s') % iso_file)
                try:
                    vopt_media = self._create_vopt_media(
                        iso_file, repository, instance, host)
                    LOG.info(_('vopt created in repository'))
                finally:
                    if iso_file:
                        os.remove(iso_file)
        except greenlet.GreenletExit:
            if self.vopt_name:
                self.remove_vopt_media(host, self.vopt_name,
                                       repository.vios_uuid)
        return vopt_media

    @logcall
    def _create_vopt_media(self, iso_file, repository, instance, host):
        """
        Create a vopt media file in a media repository on the given
        VIOS for the given local ISO file
        This will handle the transmission of the ISO file to the VIOS.

        :param iso_file the local path to the ISO
        :param repository class representing the media repository to use
        :param instance Instance object as returned by DB layer.
        :param host the compute host object
        :returns the vopt name in the media repository
        """

        # TODO, use fractional MB size for local file
        local_iso_size_mb = os.path.getsize(iso_file) / 1048576.0
        rep_free_mb = float(repository.get_free_space()) * 1024

        if local_iso_size_mb > rep_free_mb:
            raise pvcex.IBMPowerVMMediaRepOutOfSpace(
                size=local_iso_size_mb, free=rep_free_mb)

        vopt_img_name = None
        try:
            for retries in xrange(POST_RETRIES + 1):
                try:
                    vopt_img_name = self._transfer_iso(
                        iso_file, repository.vios_uuid, instance, host)
                    if vopt_img_name:
                        break
                except hmc_ex.IBMPowerVIOSIsTooBusy as e:
                    LOG.debug('Retry  _transfer_iso %s' % retries)
                    if retries == POST_RETRIES:
                        raise e
                except Exception as e:
                    raise e
        finally:
            if self.fd:
                self._remove_iso_file(self.fd.findtext('FileUUID'), host)

        return vopt_img_name

    @logcall
    def _transfer_iso(self, iso_file, vios_uuid, instance, host):
        """
        This method transfers the ISO to the VIOS and returns the remote path
        to the file.
        :param iso_file the local path to the ISO
        :param vios_uuid the VIOS to transfer the ISO to
        :param instance Instance object as returned by DB layer.
        :param host the compute host object
        :returns the remote (VIOS) path to the ISO
        """
        def _sha256sum(filename):
            sha256 = hashlib.sha256()
            with open(filename, 'rb') as f:
                for chunk in iter(lambda: f.read(
                                  128 * sha256.block_size), b''):
                    sha256.update(chunk)
            return sha256.hexdigest()

        def _create_file_metadata(vopt_name):
            # Create file descriptor
            filesize = os.path.getsize(iso_file)
            fd = K2Elem('File',
                        ns=k2o.web_ns,
                        attrib={'schemaVersion': 'V1_0'},
                        children=[K2Elem('Filename',
                                         ns=k2o.web_ns,
                                         text=vopt_name),
                                  K2Elem('InternetMediaType',
                                         ns=k2o.web_ns,
                                         text='application/octet-stream'),
                                  K2Elem('SHA256',
                                         ns=k2o.web_ns,
                                         text=_sha256sum(iso_file)),
                                  K2Elem('ExpectedFileSizeInBytes',
                                         ns=k2o.web_ns,
                                         text=str(filesize)),
                                  K2Elem('FileEnumType',
                                         ns=k2o.web_ns,
                                         text='BROKERED_MEDIA_ISO'),
                                  K2Elem('TargetVirtualIOServerUUID',
                                         ns=k2o.web_ns,
                                         text=vios_uuid)])
            return fd

        vopt_name = self._get_optical_name(instance['uuid'])
        LOG.info(_('Uploading vopt: %s') % vopt_name)
        with K2OperGuard(host=host) as oper:
            meta_resp = None
            try:
                if not self.fd:
                    fmeta = _create_file_metadata(vopt_name)
                    meta_resp = oper.create(fmeta, 'File', service='web')
                    self.fd = meta_resp.entry.element
                self.vopt_name = vopt_name
                with open(iso_file, 'rb') as f:
                    oper.uploadfile(self.fd, f)
            except K2Error as e:
                exp = e
                if e.k2response and \
                        e.k2response.status == \
                        k2const.HTTP_STATUS_INVALID_URL \
                        and contains_msg_in_k2_err(e, RMC_STATE):
                    exp = hmc_ex.IBMPowerVIOSIsTooBusy(vios_name=vios_uuid)
                self._log_k2err(self.fd if self.fd else None, e)
                raise exp

        return vopt_name

    @logcall
    def _remove_iso_file(self, remote_iso_file, host):
        """
        This method removes an ISO file from the VIOS.
        :param remote_iso_file the local path to the ISO
        :param host the compute host object
        """
        LOG.info(_('Deleting file metadata %s') % remote_iso_file)
        with K2OperGuard(host=host) as oper:
            oper.delete('File', remote_iso_file, service='web')

    @logcall
    def attach_to_vhost(self, host, optical_media_name,
                        lpar_id, vios_uuid, vios_resp=None,
                        lpar_uuid=None):
        """
        This method attaches a virtual optical to the VSCSI vhost
        :param host the compute host object
        :param optical_media_name the name of the virtual optical media
            object in the repository to attach
        :param lpar_id the id of the lpar that this virtual optical
            will be attached to
        :param vios_uuid the UUID of the vios that this virtual optical
            will be attached to
        :param vios_resp: If response is not passed in, we will extract
            the latest via k2operator.
        """
        LOG.info(_("Attaching ISO %s ") % optical_media_name)
        with K2OperGuard(host=host) as oper:
            storage_map = scsi_map_processor.SCSIMapProcessor()
            storage_map.attach_storage(oper, host.uuid, vios_uuid,
                                       lpar_id, None,
                                       optical_media_name,
                                       scsi_map_processor.VOPT,
                                       resp=vios_resp, lpar_uuid=lpar_uuid)

    @logcall
    def detach_from_vhost(self, host, optical_media_name,
                          lpar_id, vios_uuid,
                          remove_optical_media=False, vios_resp=None,
                          detach_only=True):
        """
        This method detaches a virtual optical from the VSCSI vhost
        :param host the compute host object
        :param optical_media_name the name of the virtual optical media
            object in the repository to attach
        :param lpar_id the id of the lpar that this virtual optical
            will be detached from
        :param vios_uuid the UUID of the vios that this virtual optical
            will be detached from
        :param remove_optical_media: boolean to remove the virtual optical
            media from the media repository
        :param vios_resp: If response is not passed in, we will extract
            the latest via k2operator.
        """
        LOG.info(_("Detaching ISO %s ") % optical_media_name)
        with K2OperGuard(host=host) as oper:
            storage_map = scsi_map_processor.SCSIMapProcessor()
            storage_map.detach_and_delete_adapter(oper, host.uuid, vios_uuid,
                                                  lpar_id, None,
                                                  optical_media_name,
                                                  resp=vios_resp,
                                                  detach_only=detach_only)

        if remove_optical_media:
            self.remove_vopt_media(host, optical_media_name, vios_uuid)

    @logcall
    def remove_vopt_media(self, host, optical_media_name, vios_uuid):
        """
        This method deletes the virtual optical media from the repository
        :param host the compute host object
        :param optical_media_name the name of the virtual optical media
            object in the repository to delete
        :param vios_uuid the UUID of the vios that this virtual optical
            will be deleted from
        """
        with K2OperGuard(host=host) as oper:
            media_repo_processor = MediaRepositoryProcessor()
            media_repo_processor.delete_optical_media(
                oper, host.uuid, vios_uuid, optical_media_name)

    @logcall
    def timed_cleanup(self, host, instance_list):
        """
        This method detaches and deletes the virtual optical media from the
        list of instance_uuid
        :param host the compute host object
        :param instance_list the list of instance_uuid(s) that the activation
        virtualoptical will be removed from
        """
        # use K2 level caching
        with K2OperGuard(host=host) as k2oper:
            vioses_resp = k2oper.read(k2const.MGT_SYS, rootId=host.uuid,
                                      childType=k2const.VIOS,
                                      xag=k2const.ALL_VIOS_EXT_PROPS)
        backing_dev_path = storage_map_processor.SCSI_MAPPING_PATH +\
            DELIM + storage_map_processor.SERVER_ADAPTER +\
            DELIM + storage_map_processor.BACKING_DEV
        vios_map = {}
        for instance_uuid in instance_list:
            optical_name = self._get_optical_name(instance_uuid)
            vios_backing_entries = vioses_resp.feed.findentries(
                backing_dev_path, optical_name)
            if vios_backing_entries:
                vios_entry = vios_backing_entries[0]
                vios_wrapper = VIOSEntryWrapper(vios_entry)
                uuid = vios_wrapper.get_uuid()
                if uuid in vios_map:
                    vopt_list = vios_map.get(uuid)
                    vopt_list.append(optical_name)
                else:
                    vios_map.update({uuid: [optical_name]})

        for key in vios_map.keys():
            optical_list = vios_map.get(key)
            self.detach_from_vhost(host, optical_list,
                                   None, key, remove_optical_media=True)

    def detach_all_vopt_from_vios(self, host, vios_uuid,
                                  lpar_id, instance_uuid, detach_only=True):
        LOG.info(_("Detaching all vopt for virtual machine ID:  %s ")
                 % lpar_id)
        with K2OperGuard(host=host) as oper:
            vios_resp = oper.read(k2const.MGT_SYS,
                                  rootId=host.uuid,
                                  childType=k2const.VIOS,
                                  childId=vios_uuid,
                                  xag=k2const.ALL_VIOS_EXT_PROPS)
            scsi_map = scsi_map_processor.SCSIMapProcessor()
            media_list = scsi_map.find_attached_medias(vios_resp.entry.element,
                                                       lpar_id)
            for media_name in media_list:
                delete_media = False
                if instance_uuid.replace('-', '') in media_name:
                    delete_media = True
                self.detach_from_vhost(host, media_name, lpar_id,
                                       vios_uuid, delete_media, vios_resp,
                                       detach_only=detach_only)

    def _get_optical_name(self, instance_uuid):
        optical_name = 'vopt_%s' % instance_uuid.replace('-', '')
        return optical_name

    def _log_k2err(self, req, k2err):
        LOG.error(_('Failed to modify object with request %s.')
                  % req.toxmlstring() if req else '')
        resp = (k2err.k2response.body
                if k2err and k2err.k2response and k2err.k2response.body
                else '')
        LOG.error(_('Failed to modify object with response %s.') % resp)
