# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import math
import six
import powervc_nova.virt.ibmpowervm.common.constants as commonconsts

from nova import context
from oslo_config import cfg
from powervc_nova import _
from powervc_nova.virt.ibmpowervm.common import enterprise_pool as common_ep
from powervc_nova.virt.ibmpowervm.common import proc_utils
from powervc_nova.virt.ibmpowervm.pvm import convert_pypowervm_proc_compat
from powervc_nova.virt.ibmpowervm.pvm import enterprise_pool as pvm_ep

from nova_powervm.virt.powervm import host as pvm_host
from oslo_log import log as logging
from powervc_nova.objects.compute import dom as compute_dom
from pypowervm.utils import uuid as pvm_uuid
from pypowervm.wrappers import logical_partition as pvm_lpar
from pypowervm.wrappers import managed_system as pvm_ms
from pypowervm.wrappers import shared_proc_pool as pvm_spp
from pypowervm.wrappers import virtual_io_server as pvm_vios

LOG = logging.getLogger(__name__)

# TODO: ensure option gets migrated from ibmpowervm_common_max_mem_mb
# once data_mgr supports it
host_opts = [
    cfg.IntOpt('common_max_mem_mb',
               default=32768,
               help='A commonly used value for max memory MB of an instance '
               'during deploy to help calculate the reserved memory on the '
               'host')
]

CONF = cfg.CONF
CONF.register_opts(host_opts, 'powervm')

# stats list specific to this driver. It will be added to the
# COMPUTE_STATS common list for values that should be returned in
# the 'stats' value of the Host.get_host_stats table
EXTRA_COMPUTE_STATS = [
    'max_migration_ops_supported',
    'mover_service_partitions']

FAKE_STG_SIZE = (1 << 21)

# During events such as LPAR creation or DLPAR memory resize phyp will
# require additional memory for operation. As part of calculating the
# reserved memory used by the system we factor in this buffer to account
# for the additional memory needed by the firmware.
FW_RESERVED_MEM_BUFFER = 512


def get_host_stats(adapter):
    """
    Return currently known host stats.
    """
    syswraps = pvm_ms.System.wrap(adapter.read(pvm_ms.System.schema_type))
    instances_res_totals = _calculate_used_values_stats_from_insts(adapter)
    data = _build_host_stats_from_entry(adapter, syswraps[0],
                                        instances_res_totals)

    return data


def _build_host_stats_from_entry(adapter, entry, instances_res_totals):
    """
    Build the host_stats dictionary from the System entry.
    """

    # Invoke community resource builder to get a dictionary
    # with basic statistics.
    data = pvm_host.build_host_resource_from_ms(entry)

    # Additional resources needed by PowerVC
    data['system_name'] = entry.system_name

    data['hypervisor_hostname'] = CONF.host
    capabilities = entry.get_capabilities()
    if capabilities is not None:
        # PowerVC expects 'powervm_lpar_simplified_remote_restart_capable'
        # in capabilities whereas pypowervm returns
        # 'simplified_remote_restart_capable'.
        # To-Do: try to remove these inconsistencies.
        if 'simplified_remote_restart_capable' in capabilities:
            capabilities[commonconsts.PVM_SIMP_REMOTE_RESTART_CAPABLE] = \
                capabilities['simplified_remote_restart_capable']
    data['max_sys_procs_limit'] = entry.max_sys_procs_limit
    data['max_procs_per_aix_linux_lpar'] =\
        entry.max_procs_per_aix_linux_lpar
    data['max_sys_vcpus_limit'] = entry.max_sys_vcpus_limit
    data['max_vcpus_per_aix_linux_lpar'] =\
        entry.max_vcpus_per_aix_linux_lpar

    data['lmb_size'] = entry.memory_region_size
    data['host_memory_total'] = entry.memory_total
    data['host_memory_free'] = entry.memory_free

    data['local_gb'] = FAKE_STG_SIZE
    data['local_gb_used'] = 0

    # The database needs key value pairs where the value is a string.
    # Compatibility modes is a list of strings - covert to string.
    # To be consistent with HMC driver implementation and usability
    # need to convert _Plus with +
    # e.g Power6_Plus to Power6+
    modes = entry.proc_compat_modes
    if modes:
        modes = [convert_pypowervm_proc_compat(mode) for mode in modes]
    data['compatibility_modes'] = ",".join(modes)

    # determine whether host suport little endian guest based on
    # proc compatibility mode
    le_guest_cap = proc_utils.le_guest_support_from_proc_compat(
        data['compatibility_modes'])
    capabilities[commonconsts.LE_GUEST_CAPABLE] = le_guest_cap

    data.update(entry.migration_data)

    data['prs:max_parallel_migrations'] = (
        entry.migration_data['preferred_active_migrations_supported'])
    # TODO(tpeoples): This needs to somehow be populated with the actual value
    # when migration for NovaLink is implemented in S5
    data['source_host_for_migration'] = False

    # CW: Override the hypervisor_version for now until community code
    # is updated
    # https://bugs.launchpad.net/nova-powervm/+bug/1498100
    data['hypervisor_version'] = commonconsts.IBM_POWERVM_HYPERVISOR_VERSION

    stats = {}
    stats[commonconsts.HYPERVISOR_STATE] = entry.system_state
    stats[commonconsts.MANAGED_BY] = commonconsts.NOVALINK_PLATFORM
    stats[commonconsts.HOST_DEAD] = False
    stats[commonconsts.REMOTE_RESTART_TARGET_HOST] = None

    if capabilities:
        capable_list = [commonconsts.ACTIVE_LPAR_MOBILITY_CAPABLE,
                        commonconsts.INACTIVE_LPAR_MOBILITY_CAPABLE,
                        commonconsts.PVM_SIMP_REMOTE_RESTART_CAPABLE,
                        commonconsts.ACTIVE_MEMORY_EXPANSION_CAPABLE,
                        commonconsts.CUSTOM_MAC_ADDR_CAPABLE,
                        commonconsts.IBMi_LPAR_MOBILITY_CAPABLE,
                        commonconsts.IBMi_RESTRICTED_CAPABLE,
                        commonconsts.LE_GUEST_CAPABLE,
                        commonconsts.AIX_CAPABLE,
                        commonconsts.IBMI_CAPABLE,
                        commonconsts.LINUX_CAPABLE]
        for key in capable_list:
            if key in capabilities:
                stats[key] = capabilities[key]
    data['capabilities'] = capabilities

    data['proc_units_reserved'] = '%.2f' % (
        float(entry.proc_units_configurable) -
        float(entry.proc_units_avail) -
        instances_res_totals['managed_proc_units'])
    data['vcpus_reserved'] = int(
        math.ceil(float(data['proc_units_reserved'])))

    mem_data = {'max_mem': CONF.powervm.common_max_mem_mb}
    # host_memory_reserved = memory allocated to instances on the host
    # not managed by PowerVC, a small buffer for phyp, and memory allocated
    # to firmware on the host
    data['host_memory_reserved'] = (
        instances_res_totals['unmanaged_memory'] +
        FW_RESERVED_MEM_BUFFER +
        entry.firmware_memory)

    # Replace the nova-powervm 'used' values which includes the 'reserved'
    # values in the 'used' values because OpenStack doesn't consider the
    # concept of 'reserved'. The full equation is as follows for PowerVC:
    # reserved (VIOSes, firmware, phyp, etc.) +
    # used (managed by PowerVC/OpenStack) + available (what's left) =
    # configurable (total)
    stats['proc_units_used'] = '%.2f' % (
        instances_res_totals['managed_proc_units'])
    data['vcpus_used'] = int(math.ceil(float(stats['proc_units_used'])))
    data['memory_mb_used'] = instances_res_totals['managed_memory']

    # enterprise pool (CoD) statistics
    ep_uuids = common_ep.get_pool_member_uuids(
        context.get_admin_context(), CONF.host)
    if ep_uuids is not None:
        pool_uuid, member_uuid = ep_uuids
        pool_member = None
        try:
            pool_member = pvm_ep.get_member(adapter, pool_uuid, member_uuid)
        except Exception as e:
            # Very unlikely, but in the event the PowerVC DB has invalid UUIDs
            # for the pool / member, then just log an error and move on. We
            # don't want this method to raise an exception due to this.
            LOG.error(_('Failed to retrieve enterprise pool member with UUID '
                        '%(member_uuid)s from pool with UUID %(pool_uuid)s: '
                        '%(error)s') %
                      {'member_uuid': member_uuid,
                       'pool_uuid': pool_uuid,
                       'error': six.text_type(e)})

        if pool_member:
            stats[commonconsts.COD_ENABLED] = True
            stats[commonconsts.COD_INACTIVE_CORES] = pool_member.inactive_procs
            stats[commonconsts.COD_MOBILE_CORES] = pool_member.mobile_procs
            stats[commonconsts.COD_UNRETURNED_MOBILE_CORES] = (
                pool_member.unret_mobile_procs)
        else:
            stats[commonconsts.COD_ENABLED] = False
            stats[commonconsts.COD_INACTIVE_CORES] = 0
            stats[commonconsts.COD_MOBILE_CORES] = 0
            stats[commonconsts.COD_UNRETURNED_MOBILE_CORES] = 0

    stats[commonconsts.PRS_MAX_PARA_REBUILDS] = \
        CONF.prs_max_parallel_rebuilds_novalink
    stats[commonconsts.PRS_MAX_PARA_DEST_REBUILDS] = \
        CONF.prs_max_parallel_dest_rebuilds_novalink

    stats[commonconsts.NUM_RR_IN_PROGRESS] = \
        instances_res_totals[commonconsts.NUM_RR_IN_PROGRESS]

    # If the current Manager isn't the Master update the stats accordingly
    _update_stats_for_non_master(entry, stats)

    # Stats should only contain a specific list
    # of data attributes as defined in COMPUTE_STATS
    # If additional properties should be in 'stats',
    # then those properties must be added to the
    # COMPUTE_STATS list
    missing_stats = []
    for key in commonconsts.COMPUTE_STATS + EXTRA_COMPUTE_STATS:
        if key in data:
            stats[key] = data[key]
        else:
            missing_stats.append(key)

    data['stats'].update(stats)

    return data


def _calculate_used_values_stats_from_insts(adapter):
    """
    Finds instances on host that are not present in the database,
    and calculates totals for unmanaged memory and unmanaged proc units.
    Calculates totals for managed memory and managed proc units
    for instances in database.
    Also keep track of number VM with remote restart status.
    :returns: dictionary with the following keys:
              unmanaged_memory
              unmanaged_proc_units
              managed_memory
              managed_proc_units
              remote_restarts_in_progress
    """
    res_totals = {'unmanaged_memory': 0,
                  'unmanaged_proc_units': 0,
                  'managed_memory': 0,
                  'managed_proc_units': 0,
                  commonconsts.NUM_RR_IN_PROGRESS: 0}

    vioswraps = None
    lparwraps = None
    try:
        vioswraps = pvm_vios.VIOS.wrap(adapter.read(pvm_vios.VIOS.schema_type))
        lparwraps = pvm_lpar.LPAR.wrap(adapter.read(pvm_lpar.LPAR.schema_type))
    except Exception:
        LOG.debug("Cannot extract NovaLink feed.")

    if lparwraps is None:
        LOG.warning(_("Cannot determine reserved values without "
                      "the LogicalPartition response."))
        return res_totals

    # Create a list of instance uuids from the DB
    managed_instance_uuids = []
    db_instances = _get_instances_from_db()
    for db_instance in db_instances:
        vm_uuid = db_instance.get('uuid')
        if vm_uuid is not None:
            # Convert Open Stack UUID to PowerVM UUID
            managed_instance_uuids.append(pvm_uuid.
                                          convert_uuid_to_pvm(vm_uuid).upper())

    # Iterate through LPARs on the host and check if they are
    # in the DB or not.
    for lpar in lparwraps:
        memory = lpar.mem_config.current
        if lpar.proc_config.has_dedicated:
            proc_units = lpar.proc_config.dedicated_proc_cfg.desired
        else:
            proc_units = lpar.proc_config.shared_proc_cfg.desired_units
        # If managed add memory and proc units to managed totals
        if lpar.uuid in managed_instance_uuids:
            res_totals['managed_memory'] += memory
            res_totals['managed_proc_units'] += proc_units
        # If unmanaged add memory and proc units to unmanaged totals
        else:
            res_totals['unmanaged_memory'] += memory
            res_totals['unmanaged_proc_units'] += proc_units
        # (TODO) mdrabe: add remote restart when neo-rest and pypowervm
        # supports it
        """
        if (lpar.srr_state in
            ['Destination_Remote_Restarting',
             'Destination_Remote_Restarted']):
            res_totals[commonconsts.NUM_RR_IN_PROGRESS] += 1
        """

    if vioswraps is None:
        LOG.warning(_("Cannot determine reserved values without "
                      "the VirtualIOServer response."))
        return res_totals

    # Add mem and procs for VIOS Partitions to unmanaged total
    for vios in vioswraps:
        memory = vios.mem_config.current
        if vios.proc_config.has_dedicated:
            proc_units = vios.proc_config.dedicated_proc_cfg.desired
        else:
            proc_units = vios.proc_config.shared_proc_cfg.desired_units
        res_totals['unmanaged_memory'] += memory
        res_totals['unmanaged_proc_units'] += proc_units

    return res_totals


def _update_stats_for_non_master(entry, stats):
    """Helper method to update the stats if the Manager isn't the Master"""
    stats['maintenance_mode_needed'] = 'False'
    is_master = entry.element.find('IsPowerVMManagementMaster')
    # If HMC is the master, change the State and put in Maintenance Mode
    if is_master is not None and is_master.text == 'false':
        stats['hypervisor_state'] = \
            'not master - acquire master or use hmc to manage'
        stats['maintenance_mode_needed'] = 'True'


def _get_instances_from_db():
    """Returns a list of db instances for this host."""
    instfact = compute_dom.ManagedInstanceFactory.get_factory()
    db_instances = instfact.find_all_instances_by_host(
        context.get_admin_context(), CONF.host)
    return db_instances


def get_shared_proc_pools_id_name(adapter, host_uuid):
    """
    Extract all shared processor pools info (id and name) for this host.
    :return dictionary containing id and names for all shared processor pools.
    e.g
    {1: 'SharedPool01',
     2: 'SharedPool02'
    }
    """
    proc_pools = {}
    # Extract all shared processor pools info.
    pool_wraps = pvm_spp.SharedProcPool.get(adapter,
                                            parent_type=pvm_ms.System,
                                            parent_uuid=host_uuid)
    for pool in pool_wraps:
        proc_pools[pool.id] = pool.name
    return proc_pools


def get_spp_name(adapter, host_uuid, pool_id):
    """Returns the shared proc pool name for a given pool id.

    :param pool_id: The shared proc pool id.
    :return: The name of the shared proc pool.
    """
    if pool_id == 0:
        return pvm_spp.DEFAULT_POOL_DISPLAY_NAME

    # Search for the pool with this id
    pool_wraps = pvm_spp.SharedProcPool.search(
        adapter, id=pool_id,
        parent_type=pvm_ms.System.schema_type,
        parent_uuid=host_uuid)

    if len(pool_wraps) == 0:
        msg = (_('Unable to find Shared Processor Pool with ID '
                 '%(pool_id)s.') % {'pool_id': pool_id})
        LOG.error(msg)
        raise Exception(msg)

    return pool_wraps[0].name


def get_spp_id(adapter, host_uuid, pool_name):
    """Returns the shared proc pool id for a given pool name.

    :param pool_name: The shared proc pool name.
    :return: The shared proc pool id if the proc pool exists, None otherwise.
    """
    if (pool_name is None or
            pool_name == pvm_spp.DEFAULT_POOL_DISPLAY_NAME):
        # The default pool is 0
        return 0

    # Search for the pool with this name
    pool_wraps = pvm_spp.SharedProcPool.search(
        adapter, name=pool_name,
        parent_type=pvm_ms.System.schema_type,
        parent_uuid=host_uuid)

    if len(pool_wraps) == 0:
        return None

    return pool_wraps[0].id
