# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
import re
import six

from nova import exception
from oslo_log import log as logging
from oslo_config import cfg
from powervc_nova import logcall
from powervc_nova.virt.ibmpowervm.pvm import host as pvm_host
from powervc_nova.virt.ibmpowervm.pvm import lpar as pvc_lpar
from nova_powervm.virt.powervm import vm
from pypowervm.wrappers import base_partition as pvm_bp
from pypowervm.wrappers import logical_partition as pvm_lpar
from pypowervm.wrappers import managed_system as pvm_ms
from pypowervm.wrappers import shared_proc_pool as pvm_spp
from pypowervm.wrappers import virtual_io_server as pvm_vios
from pypowervm.wrappers import storage as pvm_stor
from pypowervm.tasks import scsi_mapper
from pypowervm.tasks import vfc_mapper
from pypowervm import util as pyp_util
from pypowervm import const as pvm_const
from powervc_nova import _

from powervc_nova.virt.ibmpowervm.vif.pvm.vm_import_pvm \
    import ImportVMNetworkPVM
from nova.exception import NovaException
from powervc_nova.virt.ibmpowervm.common.power_spec_utils \
    import POWER_SPECS_SYS_META_MAP

LOG = logging.getLogger(__name__)

REASONS = 'reasons'
STATUS = 'status'
ADAPTERS = 'adapters'

SUPPORTED = 'supported'
NOT_SUPPORTED = 'not_supported'

CONF = cfg.CONF

NON_PV_ATTACHED = _('This virtual machine cannot be managed by '
                    'PowerVC because it has a non SAN-backed '
                    'physical volume attached to the storage '
                    'controller by virtual SCSI. '
                    'PowerVC manages NPIV-attached volumes and '
                    'SAN-backed physical volumes that are attached '
                    'by virtual SCSI. To manage this virtual machine '
                    'with PowerVC, remove all non SAN-backed '
                    'physical volumes that are attached by virtual '
                    'SCSI.')

INCOMP_STG_FOR_IMPORT = _('This virtual machine is not a candidate'
                          ' for management, either because it is not'
                          ' enabled for NPIV or because it has no'
                          ' volumes attached. PowerVC can only manage'
                          ' existing virtual machines that are using'
                          ' NPIV or virtual SCSI'
                          ' and have at least a boot volume'
                          ' attached.')

MULTI_VFC_ADAPTERS = _('This virtual machine uses more than one VFC client'
                       ' adapter for the same server adapter. PowerVC does'
                       ' not support multiple client adapter creation for'
                       ' VFC server adapter.')

PORT_NOT_ATTACHED = _('This virtual machine is not a candidate'
                      ' for management because the virtual'
                      ' machine has not been enabled for NPIV.'
                      ' It means that the virtual Fibre'
                      ' Channel client adapters on the virtual'
                      ' machine and virtual Fibre Channel'
                      ' server adapters on the Virtual I/O'
                      ' Server are mapped to a physical Fibre'
                      ' Channel port.')

SSP_ATTACHED = _('This virtual machine cannot be managed by '
                 'PowerVC because it has a SSP volume attached '
                 'to the storage controller by virtual SCSI. '
                 'PowerVC can only manage existing virtual '
                 'machines that are using NPIV or virtual SCSI '
                 'and have at least a boot volume '
                 'attached.')


class ImportVM(object):
    """
    Object to handle import vm functions for a given host
    """
    def __init__(self, host_uuid, adapter):
        self.host_uuid = host_uuid
        self.adapter = adapter

    @logcall
    def list_lpar_for_onboard(self, context, instances):
        """
        Generate the data needed by the discovery_driver.discover_instances
        for onboarding
        """
        lpar_list = []

        vif_driver = ImportVMNetworkPVM(CONF.host,
                                        self.adapter,
                                        self.host_uuid)

        # get list of VIOS wrappers
        xag = [pvm_const.XAG.VIO_STOR, pvm_const.XAG.VIO_SMAP,
               pvm_const.XAG.VIO_FMAP]
        vios_resp = self.adapter.read(pvm_ms.System.schema_type,
                                      root_id=self.host_uuid,
                                      child_type=pvm_vios.VIOS.schema_type,
                                      xag=xag)
        vios_wraps = pvm_vios.VIOS.wrap(vios_resp)

        for instance in instances:
            # get LPAR entry wrapper
            lpar_entry = vm.get_instance_wrapper(self.adapter,
                                                 instance,
                                                 self.host_uuid)

            lpar_result = self._is_lpar_supported(lpar_entry,
                                                  instance,
                                                  vif_driver,
                                                  vios_wraps)

            support = {'status': lpar_result.get(STATUS)}
            if lpar_result.get(STATUS) == NOT_SUPPORTED:
                support['reasons'] = lpar_result.get(REASONS)

            power_specs = {'vm_id': lpar_entry.id,
                           'vm_uuid': lpar_entry.uuid,
                           'vm_name': lpar_entry.name}

            lpar_info = {'uuid': lpar_entry.uuid.lower(),
                         'name': lpar_entry.name,
                         'state': vm.POWERVM_TO_NOVA_STATE[lpar_entry.state],
                         'power_specs': power_specs,
                         'support': support,
                         'adapters': lpar_result.get(ADAPTERS)}

            lpar_list.append(lpar_info)

        return lpar_list

    @logcall
    def query_onboarding_instances(self, context, instances):
        """
        Generate the data needed by the discovery_driver.query_instances
        for onboarding.
        """
        lpar_list = []

        vif_driver = ImportVMNetworkPVM(CONF.host,
                                        self.adapter,
                                        self.host_uuid)

        # get list of VIOS wrappers
        xag = [pvm_const.XAG.VIO_STOR, pvm_const.XAG.VIO_SMAP,
               pvm_const.XAG.VIO_FMAP]
        vios_resp = self.adapter.read(pvm_ms.System.schema_type,
                                      root_id=self.host_uuid,
                                      child_type=pvm_vios.VIOS.schema_type,
                                      xag=xag)
        vios_wraps = pvm_vios.VIOS.wrap(vios_resp)

        # get shared processor pool ids and names
        proc_pools_id_names = pvm_host.get_shared_proc_pools_id_name(
            self.adapter, self.host_uuid)
        # get list of VIOS IDs
        # get wwpns info from VIOS
        vios_wwpns = {}
        vios_ids_list = []
        for vios_entry in vios_wraps:
            vios_ids_list.append(str(vios_entry.id))
            wwpn_lst = vios_entry.get_pfc_wwpns()
            vios_wwpns[vios_entry.id] = wwpn_lst
        vios_ids = ','.join(vios_ids_list)

        for instance in instances:
            # get LPAR entry wrapper
            lpar_uuid = instance['uuid'].upper()
            lpar_resp = self.adapter.read(
                pvm_ms.System.schema_type, root_id=self.host_uuid,
                child_type=pvm_lpar.LPAR.schema_type, child_id=lpar_uuid)
            lpar_entry = pvm_lpar.LPAR.wrap(lpar_resp)

            # get power specs data and flavor data
            lpar_flavor = pvc_lpar.extract_flavor_from_lpar(
                self.adapter, self.host_uuid, lpar_entry)
            lpar_flavor = self._convert_flavor_to_include_extra_specs(
                lpar_flavor)
            power_specs = pvc_lpar.get_lpar_power_stats(lpar_entry)
            power_specs['lpar_type'] = lpar_entry.env
            power_specs['vios_ids'] = vios_ids
            # Add the shared processor pool name if applicable
            proc_pool_id = power_specs.get('shared_proc_pool_id')
            if proc_pool_id is not None:
                if proc_pool_id == 0:
                    proc_pool_name = pvm_spp.DEFAULT_POOL_DISPLAY_NAME
                else:
                    proc_pool_name = proc_pools_id_names.get(proc_pool_id)
                power_specs['shared_proc_pool_name'] = proc_pool_name
            # Need to update to include volume info
            ports = []
            volumes = []
            disks = None
            vol_connector = None
            # call network component to  update the ports field
            vif_ports = vif_driver.list_ports(lpar_entry)
            for port in vif_ports:
                # extract the info we needed
                ports.append({'mac_address': port.get('mac_address'),
                              'status': port.get('status'),
                              'vswitch': port.get('vswitch'),
                              'segmentation_id':
                              port.get('provider:segmentation_id'),
                              'physical_network':
                              port.get('physical_network')})

            # validate & extract storage info
            storage_result = self.interrogate_lpar_storage(vios_wraps,
                                                           lpar_entry.uuid)
            if storage_result.get(STATUS) == SUPPORTED:
                # extract the disk info
                disks = storage_result.get('data')
                vol_connector = storage_result.get('volume_connector')
                vol_connector['host'] = lpar_entry.name

            # parse through the disk information
            root_gb, ephemeral_gb = 0, 0
            if disks is not None and len(disks) > 0:
                root_gb = getattr(disks[0], 'capacity', None)
                root_gb = 0 if root_gb is None else root_gb
                for i in xrange(len(disks)):
                    unique_id = None
                    if isinstance(disks[i], pvm_vios.VFCMapping):
                        k2udid = None
                        name = None
                        if disks[i].backing_port:
                            unique_id = disks[i].backing_port.udid
                        client_adpt = getattr(disks[i], 'client_adapter', None)
                        if client_adpt:
                            wwpns = disks[i].client_adapter.wwpns
                        else:
                            wwpns = None
                    else:
                        k2udid = disks[i].udid
                        name = disks[i].name
                        unique_id = disks[i]._get_val_str('VolumeUniqueID')
                        wwpns = None
                    if unique_id:
                        match = re.search(r'^[0-9]{5}([0-9A-F]{32}).+$',
                                          unique_id)
                        # only update unique_id if found match
                        if match and match.lastindex >= 1:
                            unique_id = match.group(1)
                    vol = {'name': name,
                           'unique_device_id': unique_id,
                           'k2udid': k2udid,
                           }
                    if k2udid is not None:
                        # get the wwpns of vios
                        vol['vios_wwpns'] = vios_wwpns
                        # get the page 83 NAA
                        pg83 = disks[i].pg83
                        if pg83 is not None:
                            vol['pg83NAA'] = pg83
                    if wwpns is not None:
                        vol['wwpns'] = wwpns
                    volumes.append(vol)

            # duplicate the system_metadata
            sys_meta = {}
            for key, val in six.iteritems(power_specs):
                if key in POWER_SPECS_SYS_META_MAP:
                    sys_meta[POWER_SPECS_SYS_META_MAP[key]] = str(val)

            lpar_info = {'uuid': power_specs['vm_uuid'].lower(),
                         'name': power_specs['vm_name'],
                         'state': power_specs['power_state'],
                         'architecture': 'ppc64',
                         'power_specs': power_specs,
                         'system_metadata': sys_meta,
                         'vcpus': power_specs['vcpus'],
                         'memory_mb': power_specs['memory_mb'],
                         'root_gb': root_gb,
                         'ephemeral_gb': ephemeral_gb,
                         'volumes': volumes,
                         'ports': ports,
                         'volume_connector': vol_connector}

            lpar_info['flavor'] = lpar_flavor
            lpar_list.append(lpar_info)

        return lpar_list

    def _is_lpar_supported(self, lpar_entry, instance, vif_driver, vios_wraps):
        status = SUPPORTED
        reasons = []

        lpar_type = lpar_entry.env
        if (lpar_type != pvm_bp.LPARType.AIXLINUX and
                lpar_type != pvm_bp.LPARType.OS400):
            status = NOT_SUPPORTED
            reasons.append(_('This virtual machine is not a candidate for '
                             'management because you can manage only '
                             'virtual machines that are configured for AIX,'
                             ' Linux, or IBM i operating systems.'))

        shared_mem_enabled = lpar_entry.mem_config.shared_enabled
        if shared_mem_enabled is not None and shared_mem_enabled:
            status = NOT_SUPPORTED
            reasons.append(_('This virtual machine is not a candidate for '
                             'management because it is configured to use a '
                             'shared processor memory pool other than the '
                             'default.  You can only manage virtual machines '
                             'that use the default shared pool.'))

        net_prep_steps, vif_reasons, adapters = \
            vif_driver.interrogate_vm(lpar_entry)
        if vif_reasons is not None and len(vif_reasons) > 0:
            status = NOT_SUPPORTED
            reasons.extend(vif_reasons)

        # check for storage
        storage_result = self.interrogate_lpar_storage(
            vios_wraps, lpar_entry.uuid)
        if storage_result.get(STATUS) == NOT_SUPPORTED:
            status = NOT_SUPPORTED
            reasons += storage_result.get(REASONS)

        return {STATUS: status,
                REASONS: reasons,
                ADAPTERS: adapters}

    @logcall
    def interrogate_lpar_storage(self, vios_wraps, lpar_uuid):
        """
        Iterate through the storages allocated for the partition to verify
        it's eligible for onboarding.
        """
        status = SUPPORTED
        reasons = []
        valid_scsi_cfg = True
        ssp_attached_volumes = False
        vscsi_connector_data = None
        npiv_connector_data = None

        # find SCSI and FC storage
        # return error if no PV SCSI Mapping or FC Mapping
        fc_list = []
        scsi_list = []
        for vios_entry in vios_wraps:
            # get list of SCSI mappings
            scsi_mappings = scsi_mapper.find_maps(vios_entry.scsi_mappings,
                                                  lpar_uuid)
            # get list of VFC mappings
            vfc_mappings = vfc_mapper.find_maps(vios_entry.vfc_mappings,
                                                lpar_uuid)
            fc_list.extend(vfc_mappings)

            for scsi_map in scsi_mappings:
                backing_storage_elem = scsi_map.backing_storage
                # we do not support SSP attached volumes
                if isinstance(backing_storage_elem, pvm_stor.LU):
                    ssp_attached_volumes = True
                elif isinstance(backing_storage_elem, pvm_stor.PV):
                    scsi_list.append(scsi_map)
                elif isinstance(backing_storage_elem, pvm_stor.VDisk):
                    valid_scsi_cfg = False

        if not valid_scsi_cfg:
            status = NOT_SUPPORTED
            reasons.append(NON_PV_ATTACHED)
        elif ssp_attached_volumes:
            status = NOT_SUPPORTED
            reasons.append(SSP_ATTACHED)
        else:
            combined_data = []
            npiv_connector_data = \
                self._convert_to_volume_connector_info(fc_list, 'npiv')
            vscsi_connector_data = \
                self._convert_to_volume_connector_info(scsi_list, 'pv_vscsi')

            if (npiv_connector_data and
                    npiv_connector_data.get('multi_npiv_adapters')):
                status = NOT_SUPPORTED
                reasons.append(MULTI_VFC_ADAPTERS)

            # return error if no PV SCSI Mapping and no fc mapping
            if not scsi_list and not fc_list:
                status = NOT_SUPPORTED
                reasons.append(INCOMP_STG_FOR_IMPORT)
            elif not scsi_list:
                # make sure it's mapped to physical port
                if (npiv_connector_data and
                        npiv_connector_data.get('phy_to_virt_initiators')):
                    combined_data = fc_list
                else:
                    status = NOT_SUPPORTED
                    reasons.append(PORT_NOT_ATTACHED)
            elif not fc_list:
                # check for local and pv storage
                combined_data = scsi_list
            else:
                # check for local and pv storage
                combined_data += (scsi_list + fc_list)

            # extract disks from combined_data
            if status == SUPPORTED:
                disks = self._guess_boot_order(combined_data)
                # the output from guess_boot_order will exclude
                # virtual media
                # so, we will make sure that disks or connector_data
                # is not empty
                if ((npiv_connector_data and fc_list) or
                        (disks and vscsi_connector_data)):
                    connector_data = {}
                    connector_data['vscsi_connector_data'] = \
                        vscsi_connector_data
                    connector_data['npiv_connector_data'] = npiv_connector_data
                    return {STATUS: SUPPORTED,
                            'data': disks,
                            'volume_connector': connector_data}
                else:
                    status = NOT_SUPPORTED
                    reasons.append(INCOMP_STG_FOR_IMPORT)

        return {STATUS: status,
                REASONS: reasons}

    def _convert_to_volume_connector_info(self, data, connection_type):
        """
        Convert storage data to volume connector information that can be
        consumed by cinder to clean up zoning setup.
        """
        vol_conn = {}
        vol_conn['connection_type'] = [connection_type]
        # Virtual Fibre Channel Client Adapter dictionary
        vfccas = {}

        if data and connection_type == 'npiv':
            # iterate through the storage info
            for stg_entry in data:
                # NPIV PhysFCPort data
                if isinstance(stg_entry.backing_port, pvm_bp.PhysFCPort):
                    # check client adapter exists
                    wwpns = None
                    client_adpt = getattr(stg_entry, 'client_adapter', None)
                    if client_adpt:
                        # The wwpns in the client adapter from the VFCMapping
                        # will not always put the logged in wwpn first, we need
                        # to read the VFCClientAdapter to get the proper wwpn
                        # order.
                        if not vfccas:
                            # here we read all the vfc for given instance
                            lpar_uuid = pyp_util.get_req_path_uuid(
                                stg_entry.client_lpar_href, preserve_case=True)
                            vfc_resp = (
                                self.adapter.read(
                                    pvm_lpar.LPAR.schema_type,
                                    root_id=lpar_uuid,
                                    child_type=(pvm_stor.VFCClientAdapter.
                                                schema_type)))
                            vfcca_wraps = pvm_stor.VFCClientAdapter.wrap(
                                vfc_resp)

                            npiv_adapters = []
                            for vfcca in vfcca_wraps:
                                # check if the vfc are for the same port
                                vfc_port = vfcca.loc_code
                                if vfc_port in npiv_adapters:
                                    # this vm has multiple client adapters to
                                    # same server port
                                    vol_conn['multi_npiv_adapters'] = True
                                else:
                                    npiv_adapters.append(vfc_port)
                                # Key on slot_number since we have access to
                                # slot_number in the VFCMapping as well
                                vfccas[vfcca.lpar_slot_num] = vfcca

                        c_adpt = vfccas[stg_entry.client_adapter.lpar_slot_num]
                        wwpns = c_adpt.wwpns
                    # extract client wwpns
                    if wwpns:
                        if vol_conn.get('wwpns') is None:
                            vol_conn['wwpns'] = []
                        vol_conn['wwpns'] += wwpns
                        if vol_conn.get('phy_to_virt_initiators') is None:
                            vol_conn['phy_to_virt_initiators'] = {}
                        # extract port wwpn
                        port_wwpn = stg_entry.backing_port.wwpn
                        if port_wwpn:
                            vol_conn['phy_to_virt_initiators'][port_wwpn] = (
                                wwpns)
        return vol_conn

    def _guess_boot_order(self, disks):
        """
        This method is run after we have combined the disk info from
        VirtualSCSIMappings and VirtualFibreChannelMappings.
        We use combination of virtual slot number and LUAddress to perform
        educated guess on boot disk ordering.
        """
        # perform educated guest on the boot disk ordering
        ordered_disks = []
        if disks is None:
            return None
        sorted_disks = sorted(disks,
                              key=lambda x: x.client_adapter.lpar_slot_num
                              if (x and x.client_adapter)
                              else None)

        for d in sorted_disks:
            if getattr(d, 'backing_storage', None) is not None:
                ordered_disks.append(d.backing_storage)
            else:
                ordered_disks.append(d)
        return ordered_disks

    def _convert_flavor_to_include_extra_specs(self, lpar_flavor):
        """extract_flavor_from_lpar return a flat dict including the
        extra_specs. We need to extract extra_specs attributes and put them
        in extra_specs dict."""
        new_flavor = {}
        extra_specs = {}
        for k, v in six.iteritems(lpar_flavor):
            if k.startswith('powervm:'):
                extra_specs[k] = v
            else:
                new_flavor[k] = v
        new_flavor['extra_specs'] = extra_specs
        return new_flavor
