# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
import copy
import six
from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils

from nova.objects import block_device as blk_dev_obj
from nova import block_device
from nova.compute import power_state
from nova_powervm.virt.powervm import vm as vm_helper
from nova_powervm.virt.powervm import slot

from powervc_nova import _
from powervc_nova.virt.ibmpowervm.common import exception as common_exception
from powervc_nova.volume import cinder as volume
from powervc_nova.virt.ibmpowervm.common.constants import IBMI_OS_TYPE
from powervc_nova.virt.ibmpowervm.common import proc_utils
from powervc_nova.virt.ibmpowervm.pvm import convert_pypowervm_proc_compat
from powervc_nova.virt.ibmpowervm.pvm import host as pvm_host

from pypowervm import util as pvm_util
from pypowervm.tasks import client_storage as stg_tsk
from pypowervm.tasks import scsi_mapper
from pypowervm.utils import validation as vldn
from pypowervm.wrappers import shared_proc_pool as pvm_spp
from pypowervm.wrappers import storage as pvm_stor

LOG = logging.getLogger(__name__)
CONF = cfg.CONF

MIN_PROC_UNIT_PER_VCPU = .1
IBM_POWERVM_VCPU_MULTIPLIER = 0.5
POWERVM_TO_NOVA_SHARING_MODE = {
    'sre idle proces': 'share_idle_procs',
    'sre idle procs active': 'share_idle_procs_active',
    'sre idle procs always': 'share_idle_procs_always',
    'keep idle procs': 'keep_idle_procs',
    'uncapped': 'uncapped', 'capped': 'capped'}


def get_lpar_power_stats(lpar_entry):
    """
    Return the vm stats for the lpar.
    """
    #     The logic should be -
    #     - Find if the VM is running in shared or dedicated mode
    #     - Based on that set the VM attributes
    #     - Dedicated mode needs few additional attributes
    #     - For shared mode get whether its capped or uncapped
    #     - Get other details like OS and SRR and update the specs

    power_specs = {}
    power_specs['vm_id'] = lpar_entry.id
    power_specs['vm_name'] = lpar_entry.name
    power_specs['vm_uuid'] = lpar_entry.uuid

    power_specs['power_state'] = \
        vm_helper.POWERVM_TO_NOVA_STATE[lpar_entry.state]
    power_specs['rmc_state'] = lpar_entry.rmc_state
    power_specs['avail_priority'] = lpar_entry.avail_priority

    is_dedicated = lpar_entry.proc_config.has_dedicated

    power_specs['dedicated_proc'] = is_dedicated

    LOG.debug('Getting data for %s, is_dedicated, sharing_mode - %s, %s',
              lpar_entry.name,
              is_dedicated,
              lpar_entry.proc_config.sharing_mode)

    if is_dedicated:
        lpar_proc_config = lpar_entry.proc_config.dedicated_proc_cfg
        power_specs['vcpus'] = lpar_proc_config.desired
        power_specs['proc_units'] = lpar_proc_config.desired
        power_specs['max_proc_units'] = lpar_proc_config.max
        power_specs['max_vcpus'] = lpar_proc_config.max
        power_specs['min_proc_units'] = lpar_proc_config.min
        power_specs['min_vcpus'] = lpar_proc_config.min
        power_specs['pending_proc_units'] = lpar_proc_config.desired
        power_specs['pending_vcpus'] = lpar_proc_config.desired
        power_specs['pending_max_proc_units'] = lpar_proc_config.max
        power_specs['pending_max_vcpus'] = lpar_proc_config.max
        power_specs['pending_min_vcpus'] = lpar_proc_config.min
        power_specs['pending_min_proc_units'] = lpar_proc_config.min
        power_specs['run_proc_units'] = lpar_proc_config.desired
        power_specs['run_vcpus'] = lpar_proc_config.desired
        power_specs['dedicated_sharing_mode'] = (
            POWERVM_TO_NOVA_SHARING_MODE.get(
                lpar_entry.proc_config.sharing_mode))
        power_specs['uncapped'] = 'false'
        power_specs['shared_weight'] = '0'
    else:
        lpar_proc_config = lpar_entry.proc_config.shared_proc_cfg
        power_specs['vcpus'] = lpar_proc_config.desired_virtual
        power_specs['proc_units'] = lpar_proc_config.desired_units
        power_specs['max_proc_units'] = lpar_proc_config.max_units
        power_specs['max_vcpus'] = lpar_proc_config.max_virtual
        power_specs['min_proc_units'] = lpar_proc_config.min_units
        power_specs['min_vcpus'] = lpar_proc_config.min_virtual
        power_specs['pending_proc_units'] = lpar_proc_config.desired_units
        power_specs['pending_vcpus'] = lpar_proc_config.desired_virtual
        power_specs['pending_max_proc_units'] = lpar_proc_config.max_units
        power_specs['pending_max_vcpus'] = lpar_proc_config.max_virtual
        power_specs['pending_min_vcpus'] = lpar_proc_config.min_virtual
        power_specs['pending_min_proc_units'] = lpar_proc_config.min_units
        power_specs['run_proc_units'] = lpar_proc_config.desired_units
        power_specs['run_vcpus'] = lpar_proc_config.desired_virtual
        power_specs['shared_proc_pool_id'] = lpar_proc_config.pool_id
        power_specs['dedicated_sharing_mode'] = ''
        sharing_mode = lpar_entry.proc_config.sharing_mode
        if sharing_mode == 'uncapped':
            power_specs['uncapped'] = 'true'
            power_specs['shared_weight'] = \
                lpar_proc_config.uncapped_weight
        else:
            power_specs['uncapped'] = 'false'
            power_specs['shared_weight'] = '0'

    power_specs['vcpu_mode'] = 'dedicated' if is_dedicated else 'shared'
    power_specs['max_memory_mb'] = lpar_entry.mem_config.max
    power_specs['min_memory_mb'] = lpar_entry.mem_config.min
    power_specs['memory_mb'] = lpar_entry.mem_config.current

    power_specs['pending_max_memory_mb'] = lpar_entry.mem_config.max
    power_specs['pending_memory_mb'] = lpar_entry.mem_config.desired
    power_specs['pending_min_memory_mb'] = lpar_entry.mem_config.min
    power_specs['run_memory_mb'] = lpar_entry.mem_config.desired
    power_specs['ame_expansion_factor'] = lpar_entry.mem_config.exp_factor

    shr_mem_enabled = lpar_entry.mem_config.shared_enabled
    # If the shr_mem_enabled is None, don't return it
    if shr_mem_enabled:
        power_specs['memory_mode'] = 'shared'
    else:
        power_specs['memory_mode'] = 'dedicated'

    power_specs['processor_compatibility'] = (
        convert_pypowervm_proc_compat(lpar_entry.proc_compat_mode))
    power_specs['current_compatibility_mode'] = (
        convert_pypowervm_proc_compat(lpar_entry.proc_compat_mode))
    power_specs['desired_compatibility_mode'] = (
        convert_pypowervm_proc_compat(lpar_entry.pending_proc_compat_mode))

    # Get operating system and version on the virtual machine
    power_specs['operating_system'] = lpar_entry.operating_system
    power_specs['srr_capability'] = lpar_entry.srr_enabled

    return power_specs


def extract_flavor_from_lpar(adapter, host_uuid, lpar_entry):
    """
    Extract flavor information needed by onboarding a LPAR via profile
    """
    flavor = {}
    # check whether it's using dedicated or shared proc
    is_dedicated = lpar_entry.proc_config.has_dedicated
    sharing_mode = POWERVM_TO_NOVA_SHARING_MODE.get(
        lpar_entry.proc_config.sharing_mode)
    flavor['powervm:dedicated_proc'] = is_dedicated

    if is_dedicated:
        lpar_proc_config = lpar_entry.proc_config.dedicated_proc_cfg
        flavor['vcpus'] = lpar_proc_config.desired
        flavor['powervm:max_vcpu'] = lpar_proc_config.max
        flavor['powervm:min_vcpu'] = lpar_proc_config.min
        flavor['powervm:dedicated_sharing_mode'] = sharing_mode
    else:
        lpar_proc_config = lpar_entry.proc_config.shared_proc_cfg
        flavor['vcpus'] = lpar_proc_config.desired_virtual
        flavor['powervm:max_vcpu'] = lpar_proc_config.max_virtual
        flavor['powervm:min_vcpu'] = lpar_proc_config.min_virtual
        flavor['powervm:proc_units'] = lpar_proc_config.desired_units
        flavor['powervm:max_proc_units'] = lpar_proc_config.max_units
        flavor['powervm:min_proc_units'] = lpar_proc_config.min_units
        proc_pool_id = lpar_proc_config.pool_id
        if proc_pool_id is not None:
            proc_pool_name = pvm_host.get_spp_name(adapter, host_uuid,
                                                   proc_pool_id)
            flavor['powervm:shared_proc_pool_name'] = proc_pool_name
        uncapped = 'true'
        if sharing_mode != 'uncapped':
            uncapped = 'false'
        else:
            flavor['powervm:shared_weight'] = lpar_proc_config.uncapped_weight
        flavor['powervm:uncapped'] = uncapped

    flavor['memory_mb'] = lpar_entry.mem_config.current
    flavor['powervm:max_mem'] = lpar_entry.mem_config.max
    flavor['powervm:min_mem'] = lpar_entry.mem_config.min
    flavor['powervm:processor_compatibility'] = (
        convert_pypowervm_proc_compat(lpar_entry.pending_proc_compat_mode))
    flavor['powervm:availability_priority'] = lpar_entry.avail_priority
    flavor['powervm:ame_expansion_factor'] = lpar_entry.mem_config.exp_factor
    flavor['powervm:srr_capability'] = lpar_entry.srr_enabled
    # availability need to extract from LPAR
    return flavor


def extend_disk(context, instance, new_size_in_gb):
    """
    Extends the disk to a new size.
    If multiple volumes are attached to the instance,
    only the first volume in the boot order will be extended.

    :param context: security context
    :param instance: nova.objects.instance.Instance
    :param new_size_in_gb: new size in GB to which the volume
           has to be extended.
    """
    name = instance['name']
    size = instance['root_gb']
    os_type = instance['os_type']
    new_size = new_size_in_gb
    if new_size != size:
        # Resize disk
        # For IBM i, resize disk is not supported, raise exception for IBM i
        if IBMI_OS_TYPE == os_type:
            LOG.error(_('Cannot resize virtual machine: %s') % name)
            LOG.exception(_('Unable to change disk size for virtual machine '
                            '%(name)s because disk size change is not '
                            'supported for IBM i.') %
                          {'name': name})
            ex_args = {'instance_name': instance['name']}
            raise (common_exception.
                   IBMPowerVMIBMiDiskResizeNotSupported(**ex_args))
        else:
            try:
                bdms = blk_dev_obj.BlockDeviceMappingList.get_by_instance_uuid(
                    context, instance['uuid'])
                root_bdm = block_device.get_root_bdm(bdms)
                volume_id = root_bdm['volume_id']
                LOG.info(_('Extending volume %(disk)s of virtual machine '
                           '%(name)s from %(curr_size)dGB to '
                           '%(new_size)dGB') %
                         {'disk': volume_id, 'name': name,
                          'curr_size': size, 'new_size': new_size})
                volume_api = volume.PowerVCVolumeAPI()
                volume_api.extend_volume(context, volume_id, new_size)
            except Exception as e:
                LOG.error(_('Cannot resize virtual machine: %s') % name)
                LOG.exception(_('Error resizing virtual machine %(name)s.'
                                ' Extending the volume failed with the '
                                'following exception: %(exc)s') %
                              {'name': name, 'exc': six.text_type(e)})
                raise


def _rebuild_resize_flavor(lpar_entry, flav_obj, curr_flavor):
    """
    Returns the rebuilt flavor by setting the missing min/max values.
    This is intended to be used by resize to deal with the basic
    flavor where min/max values are missing in resize request.
    For active resize, missing min/max values would default to current
    configuration values. For inactive resize, missing min/max values
    would be decided based on the desired values. For missing extra_spec
    attributes in resize flavor the lpar's current values are retained.

    :param lpar_entry: lpar entry
    :param flav_obj: resize flavor
    :param curr_flavor: Current flavor of the instance being resized.
    """
    running = (vm_helper.POWERVM_TO_NOVA_STATE[lpar_entry.state] ==
               power_state.RUNNING)
    new_flavor_obj = copy.deepcopy(flav_obj)
    extra_specs = new_flavor_obj['extra_specs']
    inst_name = lpar_entry.name
    desired_vcpus = flav_obj['vcpus']
    curr_proc_mode_dedicated = curr_flavor.get('powervm:dedicated_proc')
    if extra_specs.get('powervm:dedicated_proc') is not None:
        is_dedicated = extra_specs.get('powervm:dedicated_proc').lower()
    else:
        is_dedicated = 'true' if curr_proc_mode_dedicated else 'false'

    if running:
        # vcpus
        if ('powervm:min_vcpu' not in extra_specs):
            curr_min_vcpu = int(curr_flavor.get('powervm:min_vcpu'))
            extra_specs['powervm:min_vcpu'] = curr_min_vcpu
        if ('powervm:max_vcpu' not in extra_specs):
            curr_max_vcpu = int(curr_flavor.get('powervm:max_vcpu'))
            extra_specs['powervm:max_vcpu'] = curr_max_vcpu
        # memory
        if ('powervm:min_mem' not in extra_specs):
            curr_min_mem = int(curr_flavor.get('powervm:min_mem'))
            extra_specs['powervm:min_mem'] = curr_min_mem
        if ('powervm:max_mem' not in extra_specs):
            curr_max_mem = int(curr_flavor.get('powervm:max_mem'))
            extra_specs['powervm:max_mem'] = curr_max_mem
        # Resizing to shared and min/max proc units not set? set them.
        if (not curr_proc_mode_dedicated and
            (extra_specs.get('powervm:dedicated_proc') is None or
             extra_specs.get('powervm:dedicated_proc').lower() != 'true')):
            if('powervm:min_proc_units' not in extra_specs):
                curr_min_proc_units = \
                    float(curr_flavor.get('powervm:min_proc_units'))
                curr_min_proc_units = round(curr_min_proc_units, 2)
                extra_specs['powervm:min_proc_units'] = \
                    curr_min_proc_units
            if ('powervm:max_proc_units' not in extra_specs):
                curr_max_proc_units = \
                    float(curr_flavor.get('powervm:max_proc_units'))
                curr_max_proc_units = round(curr_max_proc_units, 2)
                extra_specs['powervm:max_proc_units'] = \
                    curr_max_proc_units
    # Inactive resize
    else:
        # vcpus
        if ('powervm:min_vcpu' not in extra_specs):
            curr_min_vcpu = int(curr_flavor.get('powervm:min_vcpu'))
            # If new desired vcpus is below current min, reset
            # min vcpus
            if curr_min_vcpu > desired_vcpus:
                extra_specs['powervm:min_vcpu'] = desired_vcpus
                LOG.info(_("Setting default minimum processors to "
                           "%(vcpus)d for virtual machine %(inst)s") %
                         {'vcpus': desired_vcpus,
                          'inst': inst_name})
            else:
                extra_specs['powervm:min_vcpu'] = curr_min_vcpu
        if ('powervm:max_vcpu' not in extra_specs):
            curr_max_vcpu = int(curr_flavor.get('powervm:max_vcpu'))
            # If new desired vcpus is above current max, reset
            # max vcpus
            if curr_max_vcpu < desired_vcpus:
                extra_specs['powervm:max_vcpu'] = desired_vcpus
                LOG.info(_("Setting default maximum processors to "
                           "%(vcpus)d for virtual machine %(inst)s") %
                         {'vcpus': desired_vcpus,
                          'inst': inst_name})
            else:
                extra_specs['powervm:max_vcpu'] = curr_max_vcpu
        # memory
        if ('powervm:min_mem' not in extra_specs):
            curr_min_mem = int(curr_flavor.get('powervm:min_mem'))
            desired_mem = flav_obj['memory_mb']
            # If new desired memory is below current min, reset
            # min mem
            if curr_min_mem > desired_mem:
                extra_specs['powervm:min_mem'] = desired_mem
                LOG.info(_("Defaulting minimum memory to desired "
                           "memory '%(mem)d MB' "
                           "for virtual machine %(inst)s.") %
                         {'mem': desired_mem, 'inst': inst_name})
            else:
                extra_specs['powervm:min_mem'] = curr_min_mem
        if ('powervm:max_mem' not in extra_specs):
            curr_max_mem = int(curr_flavor.get('powervm:max_mem'))
            desired_mem = flav_obj['memory_mb']
            # If new desired mem is above current max, reset
            # max mem
            if curr_max_mem < desired_mem:
                extra_specs['powervm:max_mem'] = desired_mem
                LOG.info(_("Defaulting maximum memory to desired "
                           "memory '%(mem)d MB' "
                           "for virtual machine %(inst)s.") %
                         {'mem': desired_mem, 'inst': inst_name})
            else:
                extra_specs['powervm:max_mem'] = curr_max_mem

        # Resizing to shared and des/min/max proc units not set? set them.
        if is_dedicated != 'true':
            # Resizing to shared
            if('powervm:proc_units' not in extra_specs):
                if curr_proc_mode_dedicated:
                    # Resizing from dedicated to shared
                    des_proc_units = proc_utils.\
                        get_num_proc_units(desired_vcpus)
                    LOG.info(_("Setting default processing units to %(proc)s "
                               "for virtual machine %(inst)s") %
                             {'proc': '%.2f' % des_proc_units,
                              'inst': inst_name})
                    extra_specs['powervm:proc_units'] = des_proc_units
                else:
                    # Resizing from shared
                    # Set default proc units to current proc units
                    des_proc_units = curr_flavor.get('powervm:proc_units')
                    des_proc_units = float(des_proc_units)
                    # is current desired proc units still valid with new
                    # desired VCPUs?
                    if (round(desired_vcpus * MIN_PROC_UNIT_PER_VCPU, 2) >
                            round(des_proc_units, 2)):
                        des_proc_units = (desired_vcpus *
                                          MIN_PROC_UNIT_PER_VCPU)
                    if round(des_proc_units, 2) > desired_vcpus:
                        des_proc_units = desired_vcpus
                    extra_specs['powervm:proc_units'] = des_proc_units

            if('powervm:min_proc_units' not in extra_specs and
               'powervm:max_proc_units' not in extra_specs):
                min_proc_units, max_proc_units = \
                    _calc_min_max_proc_units(lpar_entry,
                                             new_flavor_obj, curr_flavor)
                extra_specs['powervm:min_proc_units'] = min_proc_units
                extra_specs['powervm:max_proc_units'] = max_proc_units

    # For missing extra_spec attributes in resize flavor retain the
    # lpar current values.
    expected_extra_specs_attrs = {
        'powervm:processor_compatibility', 'powervm:availability_priority',
        'powervm:srr_capability', 'powervm:dedicated_proc', 'powervm:uncapped',
        'powervm:shared_weight', 'powervm:dedicated_sharing_mode',
        'powervm:ame_expansion_factor', 'powervm:shared_proc_pool_name'}

    for attr in expected_extra_specs_attrs:
        if attr not in extra_specs and curr_flavor.get(attr) is not None:
            if (attr == 'powervm:uncapped' or attr == 'powervm:shared_weight'
                    or attr == 'powervm:shared_proc_pool_name'):
                if is_dedicated == 'true':
                    continue
            if attr == 'powervm:dedicated_sharing_mode':
                if is_dedicated == 'false':
                    continue
            # Retain the current lpar value
            extra_specs[attr] = curr_flavor.get(attr)

    return new_flavor_obj


def _calc_min_max_proc_units(lpar_entry, resize_flavor, curr_flavor):
    """If this is a resize of an instance changing
    from dedicated processor mode to shared processor mode, or if
    resizing a non-running instance and new desired proc_units is outside
    the current range calculate proc units. Otherwise, set proc units
    to current proc units values.

    :param lpar_entry: lpar entry
    :param resize_flavor: resize flavor
    :param curr_flavor: None if spawn. Current flavor of instance if resize
    :return min_proc_units and max_proc_units
    """
    extra_specs = resize_flavor['extra_specs']
    desired_proc_units = round(float(extra_specs['powervm:proc_units']), 2)
    curr_proc_mode_dedicated = curr_flavor.get('powervm:dedicated_proc')
    min_vcpu = int(extra_specs['powervm:min_vcpu'])
    max_vcpu = int(extra_specs['powervm:max_vcpu'])
    inst_name = lpar_entry.name

    # Changing an existing instance to shared processor mode
    if (curr_flavor and curr_proc_mode_dedicated is True):
        # Use either desired_proc_units or calculated proc_units for
        # min_vcpu. Use the lesser of the two, so min_proc_units is
        # never greater than desired_proc_units
        calculated_proc_units = (
            proc_utils.get_num_proc_units(min_vcpu))
        min_proc_units = min(calculated_proc_units, desired_proc_units)
        LOG.info(_("Setting default minimum processing units to %(proc)s"
                   " for virtual machine %(inst)s") %
                 {'proc': '%.2f' % min_proc_units,
                  'inst': inst_name})

        # Use either desired_proc_units or calculated proc_units for
        # max_vcpu. Use the greater of the two, so max_proc_units is
        # never less than desired_proc_units
        calculated_proc_units = (
            proc_utils.get_num_proc_units(max_vcpu))
        max_proc_units = max(calculated_proc_units, desired_proc_units)
        LOG.info(_("Setting default maximum processing units to %(proc)s "
                   "for virtual machine %(inst)s")
                 % {'proc': max_proc_units, 'inst': inst_name})
        return min_proc_units, max_proc_units

    # This is a resize that was already in shared processor mode
    # Only change min/max if new desired is outside current range
    curr_min_proc_units = (
        float(curr_flavor.get('powervm:min_proc_units')))
    curr_max_proc_units = (
        float(curr_flavor.get('powervm:max_proc_units')))

    running = (vm_helper.POWERVM_TO_NOVA_STATE[lpar_entry.state] ==
               power_state.RUNNING)
    if running is False:
        if desired_proc_units < round(curr_min_proc_units, 2):
            # Use either desired_proc_units or calculated
            # proc_units for min_vcpu. Use the lesser of the two,
            # so min_proc_units is  never greater than
            # desired_proc_units
            calculated_proc_units = (
                proc_utils.get_num_proc_units(min_vcpu))
            min_proc_units = min(calculated_proc_units,
                                 desired_proc_units)
            LOG.info(_("Setting default minimum processing units to"
                       "%(proc)s for virtual machine %(inst)s") %
                     {'proc': '%.2f' % min_proc_units,
                      'inst': inst_name})
        else:
            min_proc_units = curr_min_proc_units
        if desired_proc_units > round(curr_max_proc_units, 2):
            # Use either desired_proc_units or calculated
            # proc_units for max_vcpu. Use the greater of the two,
            # so max_proc_units is never less than
            # desired_proc_units
            calculated_proc_units = (
                proc_utils.get_num_proc_units(max_vcpu))
            max_proc_units = max(calculated_proc_units,
                                 desired_proc_units)
            LOG.info(_("Setting default maximum processing units to "
                       "%(proc)s for virtual machine %(inst)s") %
                     {'proc': max_proc_units, 'inst': inst_name})
        else:
            max_proc_units = curr_max_proc_units
    # Running instance. Set min/max to current min/max
    else:
        min_proc_units = curr_min_proc_units
        max_proc_units = curr_max_proc_units

    return min_proc_units, max_proc_units


def update(adapter, host_wrapper, instance, flavor, entry=None, name=None):
    """Update an LPAR based on the host based on the instance

    :param adapter: The adapter for the pypowervm API
    :param host_wrapper: The host wrapper
    :param instance: The nova instance.
    :param flavor: The nova flavor.
    :param entry: The instance pvm entry, if available, otherwise it will
        be fetched.
    :param name: VM name to use for the update.  Used on resize when we want
        to rename it but not use the instance name.
    :returns: The updated LPAR wrapper.
    """
    if not entry:
        entry = vm_helper.get_instance_wrapper(adapter, instance,
                                               host_wrapper.uuid)
    cur_lpar_w = copy.deepcopy(entry)
    lpar_b = vm_helper.VMBuilder(host_wrapper, adapter).lpar_builder(instance,
                                                                     flavor)
    lpar_b.rebuild(entry)
    vldn.LPARWrapperValidator(entry, host_wrapper, cur_lpar_w).validate_all()

    # Set the new name if the instance name is not desired.
    if name:
        entry.name = pvm_util.sanitize_partition_name_for_api(name)
    # Write out the new specs, return the updated version
    return entry.update()


def _add_vfc_slot_info(vios_wrap, instance, inst_slot_map):
    """Helper method to add vfc slot information.

    :param vios_wrap: single vios wrapper object
    :param instance: instance for which info needs to be updated
    :param inst_slot_map : Instance slot store which needs to be updated
    """

    npiv_port_map_keys = [key for key in instance.system_metadata.keys()
                          if key.startswith('npiv_adpt_wwpns')]
    for key in npiv_port_map_keys:
        c_wwpns = instance.system_metadata[key].split(',')[1:]
        fab_name = key.split('npiv_adpt_wwpns_')[1]
        for wwpn in c_wwpns:
            vfc_map = stg_tsk.c_wwpn_to_vfc_mapping(
                vios_wrap, six.text_type(wwpn))
            if vfc_map:
                inst_slot_map.register_vfc_mapping(vfc_map, fab_name)


def _add_scsi_slot_info(vios_wrap, instance, inst_slot_map):
    """Helper method to add scsi slot information.

    :param vios_wrap: single vios wrapper object
    :param instance: instance for which info needs to be updated
    :param bdms: block device mapping objects to check connectivity type
                and volume udid and volume id
    :param inst_slot_map : Instance slot store which needs to be updated
    """
    lpar_uuid = str(instance.get('power_specs', {}).get('vm_uuid'))
    scsi_mappings = scsi_mapper.find_maps(
        vios_wrap.scsi_mappings, client_lpar_id=lpar_uuid)
    # Loop though each of the scsi mappings adding them to the map
    for scsi_mapping in scsi_mappings:
        # We are skipping vopt at this time until we support it
        if not isinstance(scsi_mapping.backing_storage, pvm_stor.VOptMedia):
            inst_slot_map.register_vscsi_mapping(scsi_mapping)


def _add_vea_slot_info(cna_list, inst_slot_map):
    """Helper method to add vea slot information.

    :param cna_list: client network adapters of the instance
    :param instance: instance for which info needs to be updated
    :param inst_slot_map : Instance slot store which needs to be updated
    """
    for cna in cna_list:
        inst_slot_map.register_cna(cna)


def update_instance_slot_map(context, instance, vios_wraps,
                             cna_list, inst_slot_map):
    """Helper method to add slot information of the instance
    for vea/scsi/vfc in swift object store

    :param context: lpar context
    :param instance: resize instance
    :param vios_wraps: vios wrapper objects of this host
    :param cna_list: client network adapters of the instance
    :param inst_slot_map : Instance slot store which needs to be
    updated and saved
    """

    if not vios_wraps:
        return
    for vios_wrap in vios_wraps:
        _add_vfc_slot_info(vios_wrap, instance, inst_slot_map)
        _add_scsi_slot_info(vios_wrap, instance, inst_slot_map)
    _add_vea_slot_info(cna_list, inst_slot_map)
    inst_slot_map.save()
