# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
from oslo_log import log as logging
from pypowervm.tasks import power
from pypowervm.tasks import storage as pvm_stg
import six
from taskflow import task
from taskflow.types import failure as task_fail

from nova_powervm.virt.powervm.i18n import _LE
from nova_powervm.virt.powervm.i18n import _LI
from nova_powervm.virt.powervm import vm

from powervc_nova.virt.ibmpowervm.common import power_spec_utils
from powervc_nova.virt.ibmpowervm.pvm import lpar as pvc_lpar

LOG = logging.getLogger(__name__)


class Resize(task.Task):

    """The task for resizing an existing VM."""

    def __init__(self, context, adapter, host_wrapper, instance, flavor,
                 name=None):
        """Creates the Task to resize a VM.

        Provides the 'lpar_wrap' for other tasks.

        :param adapter: The adapter for the pypowervm API
        :param host_wrapper: The managed system wrapper
        :param instance: The nova instance.
        :param flavor: The nova flavor.
        :param name: VM name to use for the update.  Used on resize when we
            want to rename it but not use the instance name.
        """
        super(Resize, self).__init__(name='resize_lpar',
                                     provides='lpar_wrap')
        self.context = context
        self.adapter = adapter
        self.host_wrapper = host_wrapper
        self.host_uuid = host_wrapper.uuid
        self.instance = instance
        self.flavor = flavor
        self.vm_name = name

    def _is_changing(self, flav_obj, curr_flavor):
        # Helper method to compare current flavor with requested flavor
        # to determine if any LPAR attributes are changing
        is_changing = False
        extra_specs = flav_obj['extra_specs']

        for attr in curr_flavor.keys():
            if attr == 'memory_mb' or attr == 'vcpus':
                if int(curr_flavor[attr]) != int(flav_obj.get(attr)):
                    is_changing = True
                    break
            elif attr.startswith('powervm'):
                if attr == 'powervm:processor_compatibility':
                    # Processor compatibilty is a special case, since
                    # 'default' is a possible value for pending
                    new_proc_compat = extra_specs.get(attr)
                    if new_proc_compat:
                        new_proc_compat = new_proc_compat.lower()
                        curr_des_proc_compat = curr_flavor[attr][0].lower()
                        curr_pend_proc_compat = curr_flavor[attr][1].lower()
                        if curr_pend_proc_compat == 'default':
                            if (new_proc_compat != 'default' and
                                    curr_des_proc_compat != new_proc_compat):
                                is_changing = True
                                break
                        # Compare pending value with new proc compat mode
                        elif curr_pend_proc_compat != new_proc_compat:
                            is_changing = True
                            break
                else:
                    new_attr = extra_specs.get(attr)
                    if new_attr is not None:
                        if(str(curr_flavor[attr]).lower() !=
                           str(new_attr).lower()):
                            is_changing = True
                            break

        if not is_changing:
            # Check for the case where there is an attribute present in
            # the new flavor, that does not exist in the old flavor
            for new_extra_spec in extra_specs.keys():
                if new_extra_spec not in curr_flavor.keys():
                    is_changing = True
                    break

        return is_changing

    def _forbid_spp_name_change(self, instance_name, curr_flavor, new_flavor):
        # Helper method to forbid changing the shared processor pool name
        # when there is no change in the processor mode.
        curr_ded_proc = curr_flavor.get('powervm:dedicated_proc')

        extra_specs = new_flavor['extra_specs']
        proc_mode_dedicated = extra_specs.get('powervm:dedicated_proc')
        if proc_mode_dedicated.lower() == "true":
            is_dedicated = True
        else:
            is_dedicated = False

        if curr_ded_proc is False and is_dedicated is False:
            curr_spp_name = curr_flavor.get('powervm:shared_proc_pool_name')
            spp_name = extra_specs.get('powervm:shared_proc_pool_name')
            if curr_spp_name != spp_name:
                ex_args = {'instance_name': instance_name,
                           'curr_spp_name': curr_spp_name,
                           'spp_name': spp_name}
                msg = _("Changing the specified shared processor pool "
                        "for virtual machine %(instance_name)s is not "
                        "supported. The current processor pool is: "
                        "%(curr_spp_name)s. The requested processor pool "
                        "is: %(spp_name)s.")
                LOG.error(msg % ex_args)
                raise Exception(msg % ex_args)

    def execute(self):
        LOG.info(_('Resizing instance: %s'), self.instance.name,
                 instance=self.instance)

        def _update_vm():
            LOG.debug('Resizing instance %s.', self.instance.name,
                      instance=self.instance)
            lpar_entry = vm.get_instance_wrapper(self.adapter, self.instance,
                                                 self.host_uuid)
            curr_flavor = pvc_lpar.\
                extract_flavor_from_lpar(self.adapter, self.host_uuid,
                                         lpar_entry)
            LOG.info(_('Current properties of virtual machine %(name)s are'
                       ' %(flavor)r') % {'name': self.instance.name,
                                         'flavor': curr_flavor})
            LOG.info(_('Resize flavor is '
                       ' %(flavor)r') % {'flavor': self.flavor})

            # Check AME against the OS
            sys_meta = self.instance.get('system_metadata', {})
            inst_distro = sys_meta.get('image_os_distro', None)
            new_flav = power_spec_utils._check_os_distro_is_aix_for_ame_resize(
                self.context, self.flavor, inst_distro, self.instance.name,
                curr_flavor.get('powervm:ame_expansion_factor'))
            if self.instance.new_flavor.flavorid != new_flav.flavorid:
                self.instance.new_flavor = new_flav
                self.instance.save()

            # Compare current flavor with requested flavor to determine if
            # any LPAR attributes are changing
            update_needed = self._is_changing(new_flav, curr_flavor)
            if not update_needed:
                LOG.debug("Resize is not required.")
                return lpar_entry

            new_flavor_obj = pvc_lpar.\
                _rebuild_resize_flavor(lpar_entry, new_flav, curr_flavor)
            LOG.info(_('Resizing virtual machine %(name)s to new flavor'
                       ' %(flavor)r') % {'name': self.instance.name,
                                         'flavor': new_flavor_obj}
                     )
            self._forbid_spp_name_change(self.instance.name, curr_flavor,
                                         new_flavor_obj)
            wrap = pvc_lpar.update(self.adapter, self.host_wrapper,
                                   self.instance, new_flavor_obj,
                                   entry=lpar_entry,
                                   name=self.vm_name)
            return wrap

        # Update the VM
        _update_vm()

    def revert(self, *args, **kwargs):
        # revert to VM's original name if resize task fails
        LOG.info(_('Renaming instance to original name: %s'),
                 self.instance.name, instance=self.instance)
        return vm.rename(self.adapter, self.instance, self.instance.name)
