# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import logging
import time

from nova import context as ctx
from nova import rpc

from pypowervm.wrappers import base_partition
import pypowervm.wrappers.managed_system as pvm_system
import pypowervm.wrappers.network as pvm_net
import pypowervm.wrappers.virtual_io_server as pvm_vios
from pypowervm import const as pvm_const

from powervc_nova.objects.network import dom as network_dom
from powervc_nova.virt.ibmpowervm.vif import topo
from powervc_nova.virt.ibmpowervm.vif.pvm import utils

from powervc_nova import _

from oslo_config import cfg

LOG = logging.getLogger(__name__)

CONF = cfg.CONF
CONF.import_opt('ibmpowervm_comm_timeout', 'powervc_nova.virt.ibmpowervm.hmc')
CONF.import_opt('send_rmc_busy_message', 'powervc_nova.virt.ibmpowervm.hmc')

_XAGS = [pvm_const.XAG.VIO_NET]


class IBMPowerVMNetworkTopoPVM(topo.IBMPowerVMNetworkTopo):

    """
    IBM PowerVM Shared Ethernet Adapter Topology PVM class.
    """

    last_rmc_busy_notification = 0

    def __init__(self, host_name, host_uuid, adapter):
        """
        Set the session we'll use to interface with PVM.

        :param host_name: The name for this compute process' host.
        :param host_uuid: The UUID for this compute process' host.
        :param adapter: PVM adapter used to interface with the host.
        """
        super(IBMPowerVMNetworkTopoPVM, self).__init__(host_name=host_name,
                                                       operator=adapter)
        self.host_uuid = host_uuid

        # Cache the pypowervm VIOS objects so they can be used by both the
        # VIOS helper methods and the adapter helper methods.
        self.vios_wrap = None

    def _populate_adapters_into_vios(self, vios):
        """
        This method should put all the adapters, both SEA and VEA, into the
        given VIOS.  Each VEA should be attached to it's owning SEA, if it
        has one.

        :param vios: VioServer to fetch adapters from
        :returns boolean: True if at least one adapter was found, False
                          otherwise.
        """
        # If the VIOS RMC state is down, we can't use adapters on it.
        curr_state = vios.rmc_state.lower()
        if curr_state != base_partition.RMCState.ACTIVE:
            # Send notification to UI if RMC is in a busy state and
            # notification hasn't been sent during the interval specified.
            if(curr_state == base_partition.RMCState.BUSY and
               CONF.send_rmc_busy_message):
                curr_time = time.time()
                send_interval = 3600  # 60 minutes

                if((curr_time -
                    IBMPowerVMNetworkTopoPVM.last_rmc_busy_notification) >
                   send_interval):

                    IBMPowerVMNetworkTopoPVM.last_rmc_busy_notification = \
                        curr_time
                    payload = {}
                    payload['msg'] = _('The RMC for %(vio)s on %(host)s has '
                                       'a status code of busy.  One reason '
                                       'for this can be the incorrect sizing '
                                       'of the VIOS. Refer to the IBM PowerVM '
                                       'Best Practices red book to ensure '
                                       'that the VIOS has the appropriate set '
                                       'of resources to service the '
                                       'requests.') \
                        % {'vio': vios.lpar_name,
                           'host': CONF.host_display_name}
                    notifier = rpc.get_notifier(service='compute',
                                                host=CONF.host)
                    notifier.warn(ctx.get_admin_context(),
                                  'compute.instance.log',
                                  payload)

            LOG.warning(_('RMC state is not active on VIOS lpar %s') %
                        vios.lpar_name)

            return False

        # Get network bridges from pypowervm
        LOG.debug('VIF topology PVM - populating adapters into VIOS')
        response_feed = self.operator.read(
            root_type=pvm_system.System.schema_type,
            root_id=self.host_uuid,
            child_type=pvm_net.NetBridge.schema_type)
        bridge_wrap = pvm_net.NetBridge.wrap(response_feed)

        # Return if no NetBridge objects were found by pypowervm
        if not bridge_wrap:
            LOG.debug('VIF topology PVM - no NetBridges found')
            return False

        # Create factories that are used to build DOM objects
        veafact = network_dom.VirtualEthernetAdapterFactory.get_factory()
        seafact = network_dom.SharedEthernetAdapterFactory.get_factory()

        # Build orphan VEA DOM objects
        for vios_w in self.vios_wrap:
            if vios.id == vios_w.uuid:
                for vea_w in vios_w.derive_orphan_trunk_adapters():
                    utils._pvm_vea_to_dom_vea(self.operator, self.host_uuid,
                                              veafact, vios, vea_w)

        # Build SEA DOM objects which will cause VEA DOM objects to be built
        sea_found = False
        for bridge_w in bridge_wrap:
            for sea_w in bridge_w.seas:
                if vios.id in sea_w.vio_uri:
                    sea_found = True
                    utils._pvm_sea_to_dom_sea(self.operator, self.host_uuid,
                                              seafact, veafact, vios, sea_w,
                                              bridge_w)
        return sea_found

    def _get_all_vios(self):
        """
        This method should return all VIOS available on the current host (as
        identified in the operator).  For IVM, this should only be one VIOS,
        for HMC it could be one or more.

        :returns vioses: A list of VioServer objects
        """
        # Create a factory to build VIOS DOM objects
        viosfact = network_dom.VioServerFactory.get_factory()

        # Get all the VIOSes from pypowervm
        LOG.debug('VIF topology PVM - Getting VIOS')
        response_feed = self.operator.read(pvm_vios.VIOS.schema_type,
                                           xag=_XAGS)
        self.vios_wrap = pvm_vios.VIOS.wrap(response_feed)

        # If we find no VIOS, just return
        if not self.vios_wrap:
            LOG.debug('VIF topology PVM - no VIOS found')
            return []

        # Loop through each pypowervm VIOS object returned from the host
        return [utils._pvm_vios_to_dom_vios(self.host_name, viosfact, x)
                for x in self.vios_wrap]

    def _is_cec_running(self):
        """
        Will return if the CEC is up.
        """
        # If this code is getting called, it means the nova compute process
        # on the host is running, and thus the host itself is running.
        return True
