# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import logging
import nova.context

import pypowervm.wrappers.managed_system as pvm_system
import pypowervm.wrappers.network as pvm_net
import pypowervm.wrappers.virtual_io_server as pvm_vios

from powervc_nova import _

LOG = logging.getLogger(__name__)


def _pvm_vios_to_dom_vios(host_name, viosfact, vios_w):
    """
    Helper method to convert from a pypowervm VIOS wrapper to a DOM VIOS
    object.

    :param host_name: The name of the host the vios is on.
    :param viosfact: The factory object used to create DOM objects.
    :param vios_w: The VIOS object wrapper retrieved from pypowervm.
    :returns: A DOM VIOS object.
    """
    # Get the necessary attributes
    lpar_id = vios_w.id
    lpar_name = vios_w.name

    # Create a VIOS DOM object
    vios = viosfact.construct_vios(nova.context.get_admin_context(),
                                   host_name=host_name,
                                   lpar_id=lpar_id,
                                   lpar_name=lpar_name,
                                   state=vios_w.state,
                                   rmc_state=vios_w.rmc_state)

    # To sync with Storage, use the pypowervm UUID for the ID
    vios.id = vios_w.uuid

    LOG.debug('VIF topology PVM - Found VIOS, lpar_id: %(lpar_id)s, '
              'lpar_name: %(lpar_name)s' % {'lpar_id': lpar_id,
                                            'lpar_name': lpar_name})
    return vios


def _pvm_sea_to_dom_sea(adapter, host_uuid, seafact, veafact, vios, sea_w,
                        bridge_w):
    """
    Helper method to convert from a pypowervm SEA wrapper to a DOM SEA object.

    :param adapter: The PVM adapter object to communicate with the host.
    :param host_uuid: The UUID of the host the vswitch is on.
    :param seafact: The factory object used to create DOM objects.
    :param veafact: The factory object used to create DOM objects.
    :param vios: The VIOS this SEA is within.
    :param sea_w: The SEA wrapper object retrieved from pypowervm.
    :param bridge_w: The pypowervm NetBridge wrapper object.
    :returns: A DOM SEA object.
    """
    # Get primary VEA (pvea) from this SEA
    pvea_w = sea_w.primary_adpt
    pvea = _pvm_vea_to_dom_vea(adapter, host_uuid, veafact, vios, pvea_w)

    # Get additional trunk adapters (VEAs) for this SEA
    additional_veas = []
    for vea_w in sea_w.addl_adpts:
        vea = _pvm_vea_to_dom_vea(adapter, host_uuid, veafact, vios, vea_w)
        if vea:
            additional_veas.append(vea)

    # Get control channel VEA
    control_channel = None
    control_channel_ifname = sea_w.control_channel
    ctrl_chan_id = bridge_w.control_channel_id
    if control_channel_ifname:
        # A VIOS can have 'Client Network Adapters' - which are adapters with
        # a single VLAN that are not yet part of a SEA.  We need to take these
        # into account as they are also VEAs, just with a special situation.
        # These can be the control channel adapters.
        response_feed = adapter.read(
            root_type=pvm_vios.VIOS.schema_type,
            root_id=vios.id,
            child_type=pvm_net.CNA.schema_type)
        cna_wrap = pvm_net.CNA.wrap(response_feed)

        # We can't blindly create DOM VEA objects for all VEAs, or we will
        # create incorrect SEA/VEA mappings in the database.  We want to only
        # create VEA objects for the CNA that represents the control channel
        # adapter we are looking for.
        for cna_w in cna_wrap:
            # Look up the vswitch name for this CNA
            vswitch_id = cna_w.vswitch_id
            cna_vswitch_name = _get_vswitch_name_for_id(adapter, host_uuid,
                                                        vswitch_id)
            if not cna_vswitch_name:
                # Use a default value
                LOG.debug('Using default virtual switch name. '
                          'Found %s.' % cna_vswitch_name)
                cna_vswitch_name = 'ETHERNET0'
            # If this is a CNA (non-trunk adapter) and the pvid matches the
            # control channel id and it's on the same vswitch as this SEA.
            if(cna_w.pvid == ctrl_chan_id and
               cna_vswitch_name == pvea.vswitch_name):
                # We've got our match!
                control_channel = _pvm_vea_to_dom_vea(adapter, host_uuid,
                                                      veafact, vios, cna_w)
                break

    # Log if a control channel VEA was expected but not found
    if control_channel_ifname and not control_channel:
        LOG.warning(_("Didn't find control channel for device %(dev)s, vswitch"
                      " %(vswitch)s, and id %(c_id)d)") %
                    {'dev': control_channel_ifname,
                     'vswitch': pvea.vswitch_name,
                     'c_id': ctrl_chan_id})

    # Determine the SEA state
    state = 'Available'
    configured = sea_w.configuration_state
    if configured != pvm_net.SEAState.CONFIGURED:
        LOG.warn(_('Shared Ethernet Adapter %(name)s '
                   'ConfigurationState: %(state)s') %
                 {'name': sea_w.dev_name, 'state': configured})
        state = 'Unavailable'

    # Build the SEA DOM object and attach it to the VIOS DOM object
    sea = seafact.construct_sea(
        nova.context.get_admin_context(),
        name=sea_w.dev_name,
        vios=vios,
        slot=pvea.slot,
        state=state,
        primary_vea=pvea,
        control_channel_vea=control_channel,
        additional_veas=additional_veas)
    return sea


def _pvm_vea_to_dom_vea(adapter, host_uuid, veafact, vios, vea_w):
    """
    Helper method to convert from a pypowervm VEA wrapper to a DOM VEA object.

    :param adapter: The PVM adapter object to communicate with the host.
    :param host_uuid: The UUID of the host the vswitch is on.
    :param veafact: The factory object used to create DOM objects.
    :param vios: The VIOS this VEA is within.
    :param vea_w: The VEA wrapper object retrieved from pypowervm.
    :returns: A DOM VEA object.
    """
    # If this is a CNA that is actually a trunk adapter, ignore it.  This is a
    # new concept that pypowervm added that just creates duplicates in our
    # topology.
    if isinstance(vea_w, pvm_net.CNA) and vea_w.is_trunk:
        return None

    # Device name
    if isinstance(vea_w, pvm_net.TrunkAdapter):
        name = vea_w.dev_name
    else:
        # CNA case
        name = 'CNA_' + vea_w.loc_code
    if not name:
        LOG.warning(_('Missing name on VEA.'))

    # Slot number
    if isinstance(vea_w, pvm_net.TrunkAdapter):
        slot = vea_w.virtual_slot_number
    else:
        # CNA case
        slot = vea_w.slot
    if slot:
        slot = int(slot)
    else:
        # Set to an invalid value so constructor blows up
        slot = 0
        LOG.warning(_('Missing slot on VEA, using slot 0 instead.'))

    # Port VLAN ID
    pvid = vea_w.pvid
    if pvid:
        pvid = int(pvid)
    else:
        # Set to an invalid value so constructor blows up
        pvid = 0
        LOG.warning(_('Missing pvid on VEA, using pvid 0 instead.'))

    if isinstance(vea_w, pvm_net.TrunkAdapter):
        # This is a TrunkAdapter..so by definition is_trunk is true
        is_trunk = True
        trunk_priority = vea_w.trunk_pri
        if trunk_priority:
            trunk_priority = int(trunk_priority)
        else:
            # Set to an invalid value so constructor blows up
            trunk_priority = 0
            LOG.warning(_('Missing trunk_priority on VEA, using 0 instead.'))
    else:
        # This adapter is a CNA
        is_trunk = False
        trunk_priority = 1

    # State
    if isinstance(vea_w, pvm_net.TrunkAdapter) and vea_w.varied_on:
        state = 'Available'
    else:
        state = 'Defined'

    # 8021Q enabled
    if isinstance(vea_w, pvm_net.TrunkAdapter):
        ieee_eth = vea_w.has_tag_support
    else:
        # CNA case
        ieee_eth = vea_w.is_tagged_vlan_supported

    # Addl vlans
    addl_vlan_ids = vea_w.tagged_vlans
    if not addl_vlan_ids:
        # VirtualEthernetAdapter requires empty list, not None
        addl_vlan_ids = []

    # vswitch name
    vswitch_id = vea_w.vswitch_id
    vswitch_name = _get_vswitch_name_for_id(adapter, host_uuid, vswitch_id)
    if not vswitch_name:
        # Use a default value
        LOG.debug('Using default virtual switch name. '
                  'Found %s.' % vswitch_name)
        vswitch_name = 'ETHERNET0'

    # Build and return the DOM object.  This should raise an exception if
    # there is something wrong with the configuration of this VEA.
    try:
        vea = veafact.construct_vea(
            nova.context.get_admin_context(),
            name=name,
            vios=vios,
            slot=slot,
            pvid=pvid,
            is_trunk=is_trunk,
            trunk_pri=trunk_priority,
            state=state,
            ieee_eth=ieee_eth,
            vswitch_name=vswitch_name,
            addl_vlan_ids=addl_vlan_ids)
        return vea
    except Exception as e:
        # Exception creating the VEA.  Log the failing VEA info and proceed
        LOG.warning(_('Failed to create VEA DOM. (%s)') % e)
        LOG.exception(e)
        LOG.warning(_('Invalid VEA attributes - vswitch: %(vs)s, name: '
                      '%(name)s, slot: %(slot)d, pvid: %(pvid)d, is_trunk:'
                      ' %(is_trunk)s, trunk_priority: %(trunk_priority)d, '
                      'state: %(state)s, ieee_eth: %(ieee_eth)s, '
                      'addl_vlan_ids: %(addl_vlan_ids)s') %
                    {'vs': vswitch_name,
                     'name': name,
                     'slot': slot,
                     'pvid': pvid,
                     'is_trunk': is_trunk,
                     'trunk_priority': trunk_priority,
                     'state': state,
                     'ieee_eth': ieee_eth,
                     'addl_vlan_ids': addl_vlan_ids})
        return None


def _get_vswitch_name_for_id(adapter, host_uuid, vswitch_id):
    """
    Helper method to find the name of a vswitch given its ID.

    :param adapter: The PVM adapter object to communicate with the host.
    :param host_uuid: The UUID of the host the vswitch is on.
    :param vswitch_id: The vswitch ID to look up.
    :returns: vswitch name.
    """
    # Get vswitches
    response_feed = adapter.read(root_type=pvm_system.System.schema_type,
                                 root_id=host_uuid,
                                 child_type=pvm_net.VSwitch.schema_type)
    vswitches_w = pvm_net.VSwitch.wrap(response_feed)

    # Find vswitch with specified ID
    for vswitch_w in vswitches_w:
        if vswitch_w.switch_id == vswitch_id:
            return vswitch_w.name
    return None
