# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2014 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import logging
import six

import nova.context
from powervc_oslo.k2.k2operator import K2Error
from powervc_nova.objects.network import dom as network_dom
from powervc_nova.virt.ibmpowervm.vif.common import exception as excp
from powervc_nova.virt.ibmpowervm.vif.common import ras
from powervc_nova import logcall

from powervc_nova import _

LOG = logging.getLogger(__name__)


class IBMPowerVMNetworkTopo(object):
    """
    IBM PowerVM Shared Ethernet Adapter Topology base class, version 2.
    This class provides all the logic for collecting the topology of a VIOS.
    Subclasses need only implement the methods that actually retrieve
    information from the hardware itself.  These method stubs are located at
    the bottom of this class and describe what they are to return.
    """

    def __init__(self, host_name, operator=None):
        """
        Initialize the DOM objects to None and set the operator we'll use
        to interface with the system.  NOTE: Subclasses' __init__ should call
        this __init__() and then set up their specific operator.  For IVM, this
        will be an IVMOperator object.  For HMC, this will be a K2 operator.
        For Neo/Novalink, this will be an Adapter object.

        :param host_name: Host_name for this compute process's node
        :param operator: Operator used to interface with the system.
        """
        self.host_topo = None
        if not host_name:
            LOG.error(ras.vif_get_msg('error',
                                      'DEVNAME_INVALID') %
                      {'devname': host_name})
            raise excp.IBMPowerVMInvalidHostConfig(attr='host_name')
        self.host_name = host_name
        self.operator = operator

    def get_host_network_topology(self, context):
        """Retrieves the current Host Network Config in a Topology object"""
        host_dom = self.get_current_config(False)
        topofact = network_dom.HostNetworkTopologyFactory.get_factory()
        return topofact.construct_network_topology_from_host(context, host_dom)

    @logcall
    def get_current_config(self, must_discover_sea=True):
        """
        This method will retrieve the current configuration of the SEAs on the
        PowerVM host and return information about them in a Host object.

        :param must_discover_sea: If True, we will raise an exception if no
                                  SEAs are discovered.
        :returns host: A Host object (from the DOM) containing info on
                       the VIOS(es) and their adapters.
        """
        # Refresh the topology.  Force the discovery so that any errors
        # may be thrown upwards.
        self.refresh(must_discover_sea)

        # If the Topology wasn't retrieved, send back an empty one
        if self.host_topo is None:
            ctxt = nova.context.get_admin_context()
            netfact = network_dom.HostNetworkTopologyFactory.get_factory()
            return netfact.construct_network_topology(ctxt, self.host_name)
        return self.host_topo

    def refresh(self, must_discover_sea=True):
        """
        This method will refresh the topology such that it will reset its
        internal data model and then run a discovery on the system.  This
        ensures that any changes done on the system itself will be picked up
        and added to the data model.
        """
        # Reset the data model
        self.host_topo = None

        # If the CEC is not running, do not discover the host.
        if self._is_cec_running():
            # Discover the VIOSes and remember if discovery succeeds
            self._discover_vios_config()

        # If we were told we must discover at least one SEA, verify we did.
        if must_discover_sea:
            if self.host_topo is None:
                msg = ras.vif_get_msg('error', 'NET_INVALIDCFG')
                LOG.error(msg)
                raise excp.IBMPowerVMInvalidMgmtNetworkCfg(msg=msg)
            elif not self.host_topo.find_all_primary_seas():
                # If any of the VIOSes have RMC down, they could have been the
                # one with an SEA on it.  Raise an exception pointing to RMC
                # as the problem.
                at_least_one_rmc_active = False
                at_least_one_rmc_inactive = False
                for vios in self.host_topo.vioservers:
                    if vios.rmc_state.lower() != 'active':
                        LOG.error(ras.vif_get_msg('error',
                                                  'VIOS_NORMC') %
                                  {'lpar': vios.lpar_name,
                                   'state': vios.rmc_state})
                        at_least_one_rmc_inactive = True
                    else:
                        at_least_one_rmc_active = True

                if at_least_one_rmc_active and at_least_one_rmc_inactive:
                    # We found a mixture of both active and inactive RMC
                    # connections.  We can't definitively point to RMC as the
                    # problem, so put out a message that it COULD be the
                    # reason we found no SEAs.
                    raise excp.IBMPowerVMRMCDownNoSEA(host=self.host_name)
                elif at_least_one_rmc_active:
                    # We found an active RMC connection but no inactive ones,
                    # so we really have a situation where no SEAs were found.
                    raise excp.IBMPowerVMValidSEANotFound()
                elif at_least_one_rmc_inactive:
                    # We must've found NO active RMC connections.  Put out
                    # the very specific message about this.
                    raise excp.IBMPowerVMAllRMCDown(host=self.host_name)
                else:
                    # This can only be reached if neither active nor inactive
                    # RMC connections were found.  In other words, if no
                    # VIOSes were found.  I doubt this will ever happen, but
                    # better safe than sorry.
                    msg = ras.vif_get_msg('error', 'NET_INVALIDCFG')
                    raise excp.IBMPowerVMInvalidHostConfig(attr='No VIOS')
        else:
            # Even if an SEA isn't required, raise an exception if any
            # RMC was found to be busy.  This prevents bad data from getting
            # stored as a valid topology by the caller.
            for vios in self.host_topo.vioservers:
                if vios.rmc_state.lower() == 'busy':
                    raise excp.IBMPowerVMRMCDown(lpar=vios.lpar_name,
                                                 state=vios.rmc_state)

    @logcall
    def _discover_vios_config(self):
        """
        This function will discover the SEA configuration on the managed
        VIOSes. If it detects any faulty configuration, an exception will
        be thrown.  The exception should include data on what the issue was.
        """
        try:
            # Get all the VIOS under this host, and verify we have at least one
            vio_servers = self._get_all_vios()
            if not vio_servers:
                LOG.error(ras.vif_get_msg('error',
                                          'VIOS_NONE') %
                          self.host_name)
                raise excp.IBMPowerVMInvalidHostConfig(attr='vios')

            # Loop through every VIOS on the host.
            for vios in vio_servers:
                # See if we find some adapters
                # Look for some adapters
                if not self._populate_adapters_into_vios(vios):
                    # Found no adapters... this could be fine, but log it.
                    LOG.warning(_('%s: %s') % (vios.lpar_name,
                                               ras.vif_get_msg(
                                                   'error',
                                                   'VIOS_NOADAPTERS')))

            # If we got here found all Adapters and added them to their VIOS's
            topfact = network_dom.HostNetworkTopologyFactory.get_factory()
            self.host_topo = topfact.construct_network_topology(
                nova.context.get_admin_context(), self.host_name, vio_servers)

            # Note, we are no longer maintaining the available_vid_pool here.
            # With the new dom model, the available_vid_pool is on the
            # ManagedHost object, not the HostNetworkTopology object we're
            # returning here.  Currently, the only user that cares about the
            # vid pool is the network vif driver since it's doing a plug/unplug
            # based on vids.  Other users of this method are just collecting
            # topology in order to save/update the db.  vid pools aren't saved
            # in the db, so no need for it on the HostNetworkTopology object.

        except K2Error as e:
            # If this was a K2Error, we want to reraise it so we don't put
            # out a message about an invalid configuration, which is misleading
            if e.k2response is not None:
                LOG.error(_('Request headers:'))
                LOG.error(e.k2response.reqheaders)
                LOG.error(_('Request body:'))
                LOG.error(e.k2response.reqbody)
                LOG.error(_('Response headers:'))
                LOG.error(e.k2response.headers)
                LOG.error(_('Response body:'))
                LOG.error(e.k2response.body)
            raise

        except Exception as e:
            msg = (ras.vif_get_msg('error', 'VIOS_UNKNOWN'))
            LOG.error(msg)
            LOG.exception(six.text_type(e))

    """
    ALL METHODS BELOW MUST BE IMPLEMENTED BY SUBCLASSES
    """

    def _populate_adapters_into_vios(self, vios):
        """
        This method should put all the adapters, both SEA and VEA, into the
        given VIOS.  Each VEA should be attached to it's owning SEA, if it
        has one.

        :param vios: VioServer to fetch adapters from
        :returns boolean: True if at least one adapter was found, False
                          otherwise.
        """
        raise NotImplementedError("_get_adapters_for_vios not implemented")

    def _get_all_vios(self):
        """
        This method should return all VIOS available on the current host (as
        identified in the operator).  For IVM, this should only be one VIOS,
        for HMC it could be one or more.

        :returns vioses: A list of VioServer objects
        """
        raise NotImplementedError("_get_all_vios not implemented")

    def _is_cec_running(self):
        """
        Will return if the CEC is up.
        """
        # Default implementation is that the CEC is alive.
        return True
