# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from ceilometer.transformer import conversions
from ceilometer.sample import Sample

# The below scale factor is defined in ceilometer's pipeline.yaml
# configuration for KVM hypervisor.
SCALE_FACTOR = '100.0 / (10**9 * (resource_metadata.cpu_number or 1))'


class CpuUtilProcessor(object):
    '''
    This class utilizes Ceilometer's RateOfChangeTransformer to calculate
    CPU utilization metric for PowerKVM instance.
    Ceilometer's RateOfChangeTransformer has the logic of caching
    previous values to calculate in percentage units.
    '''
    def __init__(self):
        # For KVM, domain information returns the number of nanoseconds VM has
        # utilized on host. So the source metric is in nanoseconds.
        # We will utilize Ceilometer's RateOfChangeTransformer to convert
        # this nanoseconds to percentage based on the scaling factor.
        source = dict(name='cpu_ns',
                      unit='ns',
                      type='cumalative')
        target = dict(name='cpu_util',
                      unit='%',
                      type='gauge',
                      scale=SCALE_FACTOR)
        kwargs = dict(source=source, target=target)
        self._converter = conversions.RateOfChangeTransformer(**kwargs)

    def evaluate_cpu_util(self, instance, cpu_stats):
        '''
        This method provides to converter the Sample data that Ceilometer
        understands.

        :return: CPU utilization of instance in percentage unit.
        '''
        sample = Sample(
            name=instance.name,
            type='cumalative',
            unit='ns',
            volume=cpu_stats.get('time'),
            user_id=None,
            project_id=None,
            resource_id=instance.uuid,
            timestamp=cpu_stats.get('ts'),
            resource_metadata=dict(cpu_number=cpu_stats.get('number')))
        cpu_util = self._converter.handle_sample(None, sample)
        return cpu_util
