#
#
# Copyright 2014, 2016 IBM Corp.
# All Rights Reserved.
# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
# Copyright (c) 2010 Citrix Systems, Inc.
# Copyright (c) 2011 Piston Cloud Computing, Inc
# Copyright (c) 2012 University Of Minho
# (c) Copyright 2013 Hewlett-Packard Development Company, L.P.
# Copyright 2012 Grid Dynamics
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import errno
import os
import re
import copy
import six

from datetime import datetime
from eventlet import greenthread
from lxml import etree
from nova import context as ctx
from nova import exception
from nova import network as net_api
from nova import rpc
from nova import utils
from nova.compute import power_state
from nova.compute import task_states
from nova.compute import vm_states
from nova import context as nova_context
from nova.network import base_api
from nova.network import model as net_model
from nova.objects import instance as instance_obj
from nova.objects import flavor as flavor_obj
from nova.objects import image_meta as img_meta
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_service import loopingcall
from oslo_serialization import jsonutils
from oslo_concurrency import processutils
from oslo_utils import units, encodeutils
from oslo_utils import fileutils
from nova.objects.migrate_data import LibvirtLiveMigrateBDMInfo
from nova.virt.disk import api as disk
from nova.virt import images
from nova.virt import libvirt
from nova.virt.libvirt import config as vconfig
from nova.virt.libvirt import designer
from nova.virt.libvirt import utils as libvirt_utils
from nova.virt.libvirt.imagebackend import Qcow2
from oslo_config import cfg
from powervc_oslo.config import data_utils
from powervc_nova import logcall
from powervc_nova.common import EnhancedDict
from powervc_nova.compute import notify_messages as notify_msg
from powervc_nova.virt import powerkvm as pkvm_net
from powervc_nova.virt import instance_error_seen_multiple_time
from powervc_nova.virt import instance_found_remove_from_cache
from powervc_nova.virt import set_instance_error_state_and_notify
from powervc_nova.compute import send_ui_notification, log_operation
from powervc_nova.objects.network import dom_kvm
from powervc_nova.objects.migrate_data import PowerVCLibvirtLiveMigrateData
from nova.virt.libvirt.guest import LIBVIRT_POWER_STATE
from powervc_nova.network.powerkvm.agent \
    import ovs_operation_runner as op_runner
from powervc_nova.network.powerkvm.agent \
    import beginning_vswitch_validator
from powervc_nova.network.powerkvm import agent
from powervc_nova.virt import discovery_driver as disc_driver
import powervc_nova.virt.ibmpowervm.common.constants as commonconsts
from powervc_nova.virt.ibmpowervm.common import migrate_utils
from powervc_nova.virt.ibmpowervm.common import power_spec_utils
from powervc_nova.virt.powerkvm import config as pvc_config
from powervc_nova.virt.powerkvm import constants as pkvmc
from powervc_nova.virt.powerkvm.cpu_util_processor import CpuUtilProcessor
from powervc_nova.virt.powerkvm import proc_utils
from powervc_nova.virt.powerkvm import resource_utils
from powervc_nova.virt.powerkvm import exception as pkvmexception
from powervc_nova.virt.powerkvm import glanceclient_patch
from powervc_nova.virt.powerkvm import vm_import
from powervc_nova.virt.powerkvm import config as pkvm_config
from powervc_nova.volume import cinder as volume_cinder
from powervc_nova.network.powerkvm.agent.commandlet \
    import CommandExecutor
from powervc_nova.network.powerkvm.agent.commandlet \
    import DOMObjectConverter

from powervc_nova.network.powerkvm.agent \
    import host_port_logic as hpl
from powervc_nova import _

# temp to supprot pre_live_migration override
# from nova.virt import configdrive
from nova.virt import driver
from nova.virt.libvirt import blockinfo
# end temp

LOG = logging.getLogger(__name__)
CONF = cfg.CONF

CONF.import_opt('enable_kvm_ovs_auto_config',
                'powervc_nova.virt.powerkvm')
CONF.host_display_name = encodeutils.safe_decode(CONF.host_display_name,
                                                 'utf-8')
pvc_opts =\
    [
        cfg.ListOpt('volume_drivers',
                    default=['iscsi=powervc_nova.virt.powerkvm.volume.'
                             'PowerKVMSCSIVolumeDriver'],
                    help='Libvirt handlers for remote volumes.')
    ]
CONF.register_opts(pvc_opts, 'libvirt')


TOPOLOGY_XML = '/cpu/topology'

# virDomainMigrateFlags
VIR_MIGRATE_LIVE = 1
VIR_MIGRATE_PEER2PEER = 2
VIR_MIGRATE_TUNNELLED = 4
VIR_MIGRATE_PERSIST_DEST = 8
VIR_MIGRATE_UNDEFINE_SOURCE = 16
VIR_MIGRATE_PAUSED = 32
VIR_MIGRATE_NON_SHARED_DISK = 64
VIR_MIGRATE_NON_SHARED_INC = 128
VIR_MIGRATE_CHANGE_PROTECTION = 256
VIR_MIGRATE_UNSAFE = 512
VIR_MIGRATE_OFFLINE = 1024
VIR_MIGRATE_COMPRESSED = 2048
VIR_MIGRATE_ABORT_ON_ERROR = 4096


class PowerKVMDriver(libvirt.LibvirtDriver,
                     disc_driver.ComputeDiscoveryDriver):

    """PowerKVM Implementation of Compute Driver."""

    def __init__(self, virtapi):
        glanceclient_patch.monkey_patch_glanceclient()
        self.monkey_patch_fetch_image()
        super(PowerKVMDriver, self).__init__(virtapi)
        # keep copy of get_available_resource output
        # in case we have error communicating with hypervisor
        # and we want to keep the old stats info
        self._data = {}
        # override how to determine image backend
        # for import vm support
        orig_default = self.image_backend.BACKEND['default']
        self.image_backend.BACKEND['default'] = (
            PowerKVMQcow2 if orig_default == Qcow2 else orig_default)
        self._cpu_util_processor = CpuUtilProcessor()

    # Override OpenStack quality warnings about ppc64
    def _do_quality_warnings(self):
        pass

    def _get_volume_drivers(self):
        """ Override the parent class method returning hardcoded driver list"""
        LOG.debug("Return PowerVC's conf setting for volume_drivers: %s" %
                  CONF.libvirt.volume_drivers)
        return CONF.libvirt.volume_drivers

    def inventory_instances(self, context, instances):
        """
        Provides a mechanism for the Driver to gather Inventory-related
        information for the Instances provided off of the Hypervisor at
        periodic intervals.  The Driver is free from there to populate
        the information directly in the Database rather than return it.

        :param context: The security context for the query
        :param instances: A list of dictionary objects for each VM containing:
            uuid:      The UUID of the VM when created thru OS
            name:      The Name of the VM defined on the Hypervisor
            hostuuid:  The Hypervisor-specific UUID of the VM
        """
        for instance in instances:
            try:
                virt_dom = self._host._get_domain_by_name(instance['name'])
            except exception.NovaException:
                # If this instance cannot be found, continue with the next
                continue

            power_specs = {}
            # Get domain info
            info = virt_dom.info()
            power_specs['memory_mb'] = info[2] / units.Ki
            power_specs['vcpus'] = info[3]

            # Get domain XML
            xml = virt_dom.XMLDesc(0)
            doc = etree.fromstring(xml)
            dom_info = pvc_config.ConfigGuest()
            dom_info.parse_dom(doc)

            # Create topology dict
            topology = {}
            if dom_info.cpu is None:
                topology['sockets'] = 1
                topology['cores'] = power_specs['vcpus']
                topology['threads'] = 1
            else:
                topology['sockets'] = dom_info.cpu.sockets
                topology['cores'] = dom_info.cpu.cores
                topology['threads'] = dom_info.cpu.threads
            power_specs['cpu_topology'] = jsonutils.dumps(topology)

            # Adding tolerance for VNC device
            vnc_dict = {}

            for device in dom_info.devices:
                if (device.root_name == 'graphics' and device.type == 'vnc'):
                    vnc_dict['vnc_port'] = device.port
                    vnc_dict['vnc_autoport'] = device.autoport
                    vnc_dict['vnc_keymap'] = device.keymap
                    vnc_dict['vnc_listen'] = device.listen
                    vnc_dict['vnc_listen_type'] = device.listen_type
                    if device.listen_type == 'address':
                        vnc_dict['vnc_listen_address'] = device.listen_address
                    if device.listen_type == 'network':
                        vnc_dict['vnc_listen_network'] = device.listen_network

            power_specs['vnc_info'] = jsonutils.dumps(vnc_dict)

            power_spec_utils.update_instance_power_specs(
                context, self.virtapi, instance, power_specs)

    def spawn(self, context, instance, image_meta, *args, **kwargs):
        """Spawn a new instance.
        Extended only to include the send_ui_notification decorator.
        """
        log_operation(_('Deploying'), instance)

        resource_utils.check_resource(None, instance)
        # Once LP bug is resolved, we can remove this workaround
        # LP https://bugs.launchpad.net/nova/+bug/1511539
        # this force the architecture to be ppc64, so remote restart
        # will default bus_type to scsi instead of IDE
        if 'properties' not in image_meta:
            image_meta['properties'] = dict(architecture='ppc64')
        super(PowerKVMDriver, self).spawn(context, instance, image_meta,
                                          *args, **kwargs)

    def inventory_host(self, context):
        """
        Provides a mechanism for the Driver to gather Inventory-related
        information for the Host at periodic intervals.  The Driver is
        free from there to populate the information directly in the
        Database and should not return any information from the method.

        :param context: The security context for the query
        """
        # Update cached value for lshw data
        try:
            LOG.info(_('Collecting Hardware Inventory...'))
            pkvm_net.update_hw_inv()

            self._cleanup_neutron_ports(context)

        # Log the exception and continue on if lshw collection fails
        except Exception as e:
            LOG.error(_("Collection of hardware inventory "
                        "from lshw failed."))
            LOG.exception(e)

    def post_registration(self, context):
        """
        This method is run after the system is first registered.  It should
        only be run once.  Subsequent restarts of the nova agent will not
        invoke this.

        :param context: The nova context.
        """
        if CONF.enable_kvm_ovs_auto_config:
            if not self._validate_host_vswitch(context):
                notifier = rpc.get_notifier(service='compute',
                                            host=CONF.host)
                msg = _("The management adapter could not be attached to "
                        "the default virtual switch.  Manually add the "
                        "management adapter to the default virtual "
                        "switch before proceeding.")
                LOG.error(msg)
                info = {'msg': msg, 'host_name': CONF.host_display_name}
                admin_context = ctx.get_admin_context()
                notifier.error(admin_context, 'compute.instance.log', info)

    @logcall
    def _validate_host_vswitch(self, context):
        """
         Validate and configure default host vswitch.

         :return: if management adapter is properly
             configured, dict of the host dom,
             otherwise None
         """
        try:
            aggregator = hpl.HostPortAggregator()
            curr_dom = aggregator.get_host_dom()
            new_dom = copy.deepcopy(curr_dom)

            # check if default vswitch exists; if not, exit
            default_vswitch = (
                'default' if new_dom.version.startswith('2') else 'default0')
            default_ovs_dom = new_dom.find_ovs_obj(default_vswitch)
            if not default_ovs_dom:
                LOG.error(_("Default virtual switch not found."))
                return None

            # check to see if default vswitch is already
            # configured; if so, return true, if not continue
            # processing to configure vswitch
            if ((default_ovs_dom.ip_addresses is not None and
                 len(default_ovs_dom.ip_addresses) > 0) or
                (default_ovs_dom.ovs_port_list is not None and
                 len(default_ovs_dom.ovs_port_list) > 0)):
                return curr_dom.to_dict()

            # configure default vswitch
            dom_converter = DOMObjectConverter()
            mgmt_adapter_name = dom_converter.find_management_adapter()

            if mgmt_adapter_name is None:
                LOG.error(_("Management adapter not found."))
                return None

            # check for dhcp settings on the management adapter
            vs_validator = beginning_vswitch_validator.\
                BeginningVswitchValidator(default_vswitch, new_dom)

            ifcfg_dict = vs_validator._has_ifcfg_data(mgmt_adapter_name)
            LOG.debug('ifcfg data for %s: %s'
                      % (mgmt_adapter_name, str(ifcfg_dict)))
            if ifcfg_dict and agent.IFCFG_BOOTPROTO in ifcfg_dict and \
                    "DHCP" in ifcfg_dict[agent.IFCFG_BOOTPROTO].upper():
                LOG.error(_("The management adapter is configured for DHCP "
                            "and cannot be added to the virtual switch."))
                return None

            # make sure management adapter is unused
            if not new_dom.contains_unused_port_by_name(mgmt_adapter_name):
                LOG.error(_("Management adapter in use by another "
                            "component."))
                return None

            mgmt_adapter_port = new_dom.find_port_or_ovs(mgmt_adapter_name)

            # cannot add bridge to default as management adapter
            if isinstance(mgmt_adapter_port, dom_kvm.LinuxBridge):
                LOG.error(_("Management adapter is a bridge and cannot"
                            "be added to the virtual switch."))
                return None

            # set the management adapter on the default vswitch
            new_dom.unused_component_list.remove(mgmt_adapter_port)

            # wrapper mgmt adapter in ovs port
            ovs_port = dom_kvm.OVSPort(mgmt_adapter_name)
            ovs_port.port_list.append(mgmt_adapter_port)

            default_ovs_dom.ovs_port_list.append(ovs_port)

            runner = op_runner.OVSOperationRunner(curr_dom, True,
                                                  False)
            runner.update_host_ovs(context, new_dom)

            return new_dom.to_dict()

        except Exception as e:
            LOG.error(_("Host virtual switch validation "
                        "has failed."))
            LOG.exception(e)
            return None

    def get_host_ovs(self, context):
        """Get host ovs data from current host"""
        LOG.debug('Retrieving ovs host data...')

        aggregator = hpl.HostPortAggregator()

        ovs_data = None

        try:
            ovs_dom = aggregator.get_host_dom()
            ovs_data = ovs_dom.to_dict()
        except Exception as e:
            if ovs_data is None:
                ovs_dom = dom_kvm.HostOVSNetworkConfig(CONF.get("host"))
                error_obj = dom_kvm.Error(six.text_type(e).encode('utf-8'))
                ovs_dom.error_list = [error_obj]
                ovs_data = ovs_dom.to_dict()

        return ovs_data

    def update_host_ovs(self, context, desired_dom, force_flag, rollback):
        """update host ovs data on current host"""
        LOG.debug('Updating ovs host data...')

        ERRORS_KEY = "errors"
        retVal = {}

        try:
            aggregator = hpl.HostPortAggregator()
            current_dom_obj = aggregator.get_host_dom()

            desired_dom_obj = dom_kvm.HostOVSNetworkConfig(None)
            desired_dom_obj.from_dict(desired_dom)

            update_runner = op_runner.OVSOperationRunner(current_dom_obj,
                                                         force_flag,
                                                         rollback)
            retVal = update_runner.update_host_ovs(context, desired_dom_obj)
        except Exception as exc:
            LOG.exception(exc)
            cls_name = exc.__class__.__name__
            if ERRORS_KEY not in retVal:
                retVal[ERRORS_KEY] = \
                    [{cls_name: six.text_type(exc).encode('utf-8')}]
            else:
                retVal[ERRORS_KEY].append(
                    {cls_name: six.text_type(exc).encode('utf-8')})

        return retVal

    @logcall
    def create_host_ovs(self, context, desired_dom, force_flag, rollback):
        """create host ovs on current host"""

        ERRORS_KEY = "errors"
        retVal = {}

        try:
            aggregator = hpl.HostPortAggregator()
            current_dom_obj = aggregator.get_host_dom()

            desired_dom_obj = dom_kvm.HostOVSNetworkConfig(None)
            desired_dom_obj.from_dict(desired_dom)

            create_runner = op_runner.OVSOperationRunner(current_dom_obj,
                                                         force_flag,
                                                         rollback)
            retVal = create_runner.create_host_ovs(context, desired_dom_obj)
        except Exception as exc:
            LOG.exception(exc)
            cls_name = exc.__class__.__name__
            if ERRORS_KEY not in retVal:
                retVal[ERRORS_KEY] = \
                    [{cls_name: six.text_type(exc).encode('utf-8')}]
            else:
                retVal[ERRORS_KEY].append(
                    {cls_name: six.text_type(exc).encode('utf-8')})

        return retVal

    @logcall
    def delete_host_ovs(self, context, ovs_list, force_flag, rollback):
        """delete host ovs on current host"""

        ERRORS_KEY = "errors"
        retVal = {}

        try:
            aggregator = hpl.HostPortAggregator()
            current_dom_obj = aggregator.get_host_dom()

            delete_runner = op_runner.OVSOperationRunner(current_dom_obj,
                                                         force_flag,
                                                         rollback)
            retVal = delete_runner.delete_host_ovs(context, ovs_list)
        except Exception as exc:
            LOG.exception(exc)
            cls_name = exc.__class__.__name__
            if ERRORS_KEY not in retVal:
                retVal[ERRORS_KEY] = \
                    [{cls_name: six.text_type(exc).encode('utf-8')}]
            else:
                retVal[ERRORS_KEY].append(
                    {cls_name: six.text_type(exc).encode('utf-8')})

        return retVal

    def _get_hypervisor_type(self):
        """Get hypervisor type (ex. qemu)"""
        return self._conn.getType()

    def _skip_smt(self, extra_specs, guest):
        if extra_specs is None:
            return False
        else:
            # returning self._data since it's refresh every min via
            # periodic task and values don't really change.
            # See stg defect 35366
            return proc_utils.skip_smt(extra_specs,
                                       guest.vcpus, guest.name,
                                       self._data)

    def _get_smt(self, extra_specs, guest):
        """Extract SMT from flavor and validate it with
           vcpus and host stats.
        :param extra_specs: extra_specs from the flavor (dict)
        :param guest: LibvirtConfigGuest object
        :returns smt: Value for Guest SMT (int)
        """
        if extra_specs is None:
            smt = proc_utils.IBM_POWERKVM_SMT_OFF
        else:
            smt = proc_utils.get_smt_value(
                extra_specs, raise_exception=True,
                instance_name=guest.name)
        # Validate smt is valid with vcpus and is supported by host
        if smt != proc_utils.IBM_POWERKVM_SMT_OFF:
            proc_utils.validate_smt(
                smt, guest.vcpus, guest.name,
                self.get_available_resource())
        return smt

    def _get_guest_cpu_topology_config(self, smt, guest):
        """Add cpu topology on to the guest cpu config.
           At this point, the smt value has been validated.
        :param smt: SMT from flavor (int)
        :param guest: LibvirtConfigGuestCPU object
        """
        cpu = guest.cpu
        cpu.sockets = 1
        if smt == proc_utils.IBM_POWERKVM_SMT_OFF:
            # If no SMT is requested on the guest virtual machine,
            # threads should be set to 1, and cores should be set
            # to vCPUs from the flavor.
            cpu.threads = 1
            cpu.cores = guest.vcpus
        else:
            # If SMT is to be enabled on the guest virtual machine,
            # threads should be set to the smt value, and cores should
            # be calculated by vCPUs in the flavor divided by threads.
            cpu.threads = smt
            cpu.cores = guest.vcpus / cpu.threads
        return cpu

    #
    # Overriding the method for following reasons.
    # 1. Community code added hpet inside the clock for
    #    guest config which is not supported case for libvirt
    #    in power. We will revert the code once the community
    #    provide the fix for it.
    #
    # 2. Still redirection of console is by default in a log
    #    file which will make the serial console un accessible
    #    Changed the code to have the redirection only when its
    #    configured inside the property serial_console_log_enabled
    #    else the default behavior will be to have the pty configured
    #    so that we can access the console using virsh console.
    #
    # 3. Add CPU topology to the config based on SMT from flavor
    #
    # 4. The fourth reason for change is we do not have to merge the logic
    #    for generating guest config back from community as now we are
    #    modifying the piece of object which needs to be supported by power.

    def _get_guest_config(self, instance, network_info, image_meta,
                          disk_info, rescue=None, block_device_info=None,
                          context=None):
        guest = super(PowerKVMDriver, self)._get_guest_config(
            instance, network_info, image_meta, disk_info,
            rescue=rescue, block_device_info=block_device_info,
            context=context)

        inst_type = flavor_obj.Flavor.get_by_id(
            nova_context.get_admin_context(read_deleted='yes'),
            instance['instance_type_id'])
        extra_specs = inst_type.get('extra_specs')

        skip_smt = self._skip_smt(extra_specs, guest)

        if not skip_smt:
            # Extract and validate SMT from extra_specs
            smt = self._get_smt(extra_specs, guest)
            # Update guest cpu config with cpu topology from smt.
            self._get_guest_cpu_topology_config(smt, guest)

        # Loop through the devices and check their configuration
        for device in guest.devices[:]:
            # Delete the configured device in the guest config for
            # serial console.
            if (isinstance(device, vconfig.LibvirtConfigGuestSerial) or
                    isinstance(device, vconfig.LibvirtConfigGuestSerial)):
                guest.devices.remove(device)

            # Find the cdrom device
            if (isinstance(device, vconfig.LibvirtConfigGuestDisk) and
                    device.source_device == 'cdrom'):
                source_path = device.source_path
                # If the cdrom device does not exist, remove it from the
                # XML definition
                if not os.path.exists(source_path):
                    guest.devices.remove(device)
                # Override the driver format (eg. qcow2, raw) with
                # the actual format of the file
                source_format = images.qemu_img_info(source_path).file_format
                device.driver_format = source_format

        #
        # Check whether the redirection inside the console log
        # is enabled or not. If the redirection in console log
        # is enabled then only configure the redirection of the
        # console output inside the log file.
        #
        if CONF.serial_console_log_enabled:
            consolelog = vconfig.LibvirtConfigGuestSerial()
            consolelog.type = "file"
            consolelog.source_path = self._get_console_log_path(instance)
            guest.add_device(consolelog)

        consolepty = vconfig.LibvirtConfigGuestSerial()
        consolepty.type = "pty"
        guest.add_device(consolepty)

        # Adding vnc device if only vnc device exists in old xml
        vnc_dev = pkvm_config.ConfigGuestVNC()
        if instance.get('power_specs'):
            vnc_dict_str = instance['power_specs']['vnc_info']
            vnc_dict = jsonutils.loads(vnc_dict_str) if vnc_dict_str else {}
            if vnc_dict:
                vnc_dev.port = vnc_dict['vnc_port']
                vnc_dev.autoport = vnc_dict['vnc_autoport']
                vnc_dev.keymap = vnc_dict['vnc_keymap']
                vnc_dev.listen = vnc_dict['vnc_listen']
                vnc_dev.listen_type = vnc_dict['vnc_listen_type']
                if vnc_dev.listen_type == 'address':
                    vnc_dev.listen_address = vnc_dict['vnc_listen_address']
                if vnc_dev.listen_type == 'network':
                    vnc_dev.listen_network = vnc_dict['vnc_listen_network']

                guest.add_device(vnc_dev)

        # Set boot order for ISO deploys to allow the VM to
        # boot into the local OS after installation.
        if (image_meta.obj_attr_is_set('disk_format') and
                image_meta.disk_format == 'iso'):
            guest.os_boot_dev = ['hd', 'cdrom']

        return guest

    def snapshot(self, context, instance, image_href, update_task_state):
        """Capture the specified instance.
        Extended only to include the send_ui_notification decorator.

        :param context: security context
        :param instance: Instance object as returned by DB layer.
        :param image_id: Reference to a pre-created image that will
                         hold the snapshot.
        """

        log_operation(_('Snapshot'), instance)
        # Fix for RTC-17457
        # the following try block has been added to only type cast the error
        # libvirt driver is throwing, in case if root disk file (disk or
        # disk.local) is missing from instances folder
        # E.g. /var/lib/libvirt/images/nova/instances/<VM-ID>/
        disk_path = None
        try:
            # read virt dom
            virt_dom = self._host._get_domain_by_name(instance['name'])
            # get disk path
            disk_path, source_format = libvirt_utils.find_disk(virt_dom)
            # following statement throws TypeError if file is not
            # -available.
            libvirt_utils.get_disk_size(disk_path, source_format)
        except exception.InstanceNotFound:
            raise exception.InstanceNotRunning(instance_id=instance['uuid'])
        except TypeError:
            ex_args = {'inst': instance['name'],
                       'disk_path': disk_path}
            raise pkvmexception.PowerKVMBaseDiskFileDoesnotExist(**ex_args)

        super(PowerKVMDriver, self).snapshot(context, instance,
                                             image_href, update_task_state)

    def _create_snapshot_metadata(self, image_meta, instance,
                                  img_fmt, snp_name):
        """
        Creates snapshot's metadata.
        Extended only to set min_disk which read from instance's
        ephemeral_gb value.
        """
        # Call super's _create_snapshot_metadata
        metadata = super(PowerKVMDriver, self).\
            _create_snapshot_metadata(image_meta, instance,
                                      img_fmt, snp_name)
        # set min_disk property with instance's ephemeral_gb, if its an 'iso'
        if (image_meta.obj_attr_is_set('disk_format') and
                image_meta.disk_format == 'iso'):
            metadata['min_disk'] = instance['ephemeral_gb']

        return metadata

    def reboot(self, context, instance, network_info, reboot_type='SOFT',
               block_device_info=None, bad_volumes_callback=None):
        """Reboot the specified instance.
        Extended only to include the send_ui_notification decorator.

        :param context: security context
        :param instance: Instance object as returned by DB layer.
        :param network_info: network info for creating the domain
        :reboot_type: either SOFT or HARD reboot flag
        :block_device_info: Block device information
        :bad_volumes_callback: Not used
        """

        log_operation(_('Reboot'), instance)
        dom = self._host._get_domain_by_name(instance["name"])
        (state, _max_mem, _mem, _cpus, _t) = dom.info()
        state = LIBVIRT_POWER_STATE[state]
        if state is not power_state.RUNNING:
            super(PowerKVMDriver, self).power_on(context, instance,
                                                 network_info,
                                                 block_device_info)
        else:
            super(PowerKVMDriver, self).reboot(context, instance,
                                               network_info, reboot_type,
                                               block_device_info,
                                               bad_volumes_callback)

    def destroy(self, context, instance, network_info, block_device_info,
                destroy_disks=True, migrate_data=None):
        """ Delete the specified instance.
        Extended only to include the send_ui_notification decorator.
        :param instance: Instance object as returned by DB Layer.
        :param network_info: Network associated with the instance.
        :param block_device_info: Volume block device info attached with
        the instance.
        :param destroy_disks: disk attached to the instance.
        """

        log_operation(_('Deleting'), instance)
        try:
            super(PowerKVMDriver, self).destroy(context, instance,
                                                network_info,
                                                block_device_info,
                                                destroy_disks)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.error(_("Destroy of the instance failure: %s,"
                            "Perform Cleanup.. "), six.text_type(e),
                          instance=instance)
                # Perform the cleanup routine in case files are left behind
                # when destroy fails for any reason.
                self.delete_instance_files(instance)
        finally:
            # Perform the cleanup of the resize files if they exists.
            # This is to handle the case when resize fails and temporary
            try:
                self._cleanup_resize(instance, network_info)
            except Exception as e:
                LOG.warn(_("Cleanup resize failed: %s"), six.text_type(e))

    def _undefine_domain(self, instance):
        """ Handles the instanceNotFound exception by searching for libvirt
        guest by instance uuid.  If the guest cannot be found by instance name,
        we will search for the guest given the instance uuid.
        """
        try:
            guest = self._host.get_guest(instance)
        except exception.InstanceNotFound:
            guest = None

        if guest:
            # if the instance was found by name, continue with driver method
            super(PowerKVMDriver, self)._undefine_domain(instance)
        else:
            # if the instance was NOT found by name, search for guest given
            # instance uuid
            guests = self._host.list_guests(only_running=False)
            for guest in guests:
                if instance.uuid == guest.uuid:
                    try:
                        guest.delete_configuration()
                    except libvirt.libvirtError as e:
                        with excutils.save_and_reraise_exception():
                            errcode = e.get_error_code()
                            LOG.error(_('Error from libvirt during undefine.'
                                        ' Code=%(errcode)s Error=%(e)s'),
                                      {'errcode': errcode, 'e': e},
                                      instance=instance)

    def get_info(self, instance):
        """ Handles the instanceNotFound exception gracefully
        by setting the power_state.SHUTDOWN in order to delete
        the Error VM,
        """
        suspend_states = [power_state.SUSPENDED,
                          power_state.SHUTDOWN,
                          power_state.PAUSED]
        try:
            info = super(PowerKVMDriver, self).get_info(instance)
            # Fix suspend state since base code sets it to shutdown
            # Should be only done when there are no on going task_state.
            # task_state SUSPENDING check is needed because during concurrent
            # scale test periodic task hangs so to avoid changing the VM state
            # to Shutoff by the Life Cycle Events.
            if (instance['task_state'] in [None, task_states.SUSPENDING] and
                    info.state in suspend_states):
                virt_dom = self._host._get_domain_by_name(instance['name'])
                if virt_dom.hasManagedSaveImage(0) == 1:
                    info.state = power_state.SUSPENDED
            # Remove the Instance from the NotFound Cache when it is reported
            instance_found_remove_from_cache(instance)
            return info
        except exception.InstanceNotFound as exp:
            # Retrieve the current vm_state and task_state from instance.
            # Instance could be a dictionary when this method is called during
            # virtual machine delete process. If instance is in the middle of
            # a task, don't set to Error state. This will occur next time
            # get_info is called after the task completes.
            inst_vm_state = instance['vm_state']
            inst_task_state = instance['task_state']
            vm_state_ignore_list = [vm_states.BUILDING,
                                    vm_states.ERROR]
            repeats = CONF.number_of_exceptions_needed_for_missing_inst
            # Ignore setting to error if in certain states or a hiccup
            if ((inst_vm_state not in vm_state_ignore_list) and
                    inst_task_state is None and
                    instance_error_seen_multiple_time(instance, repeats)):
                set_instance_error_state_and_notify(instance)
            raise exp

    @logcall
    def cleanup_disks(self, context, connection_info, instance):
        """ pass through function implemented for PowerVM"
        to cleanup stale disks
        """
        pass

    @logcall
    def attach_interface(self, instance, image_meta, vif):
        """
        Attach an interface to the instance.
        """
        # First need to check to see whether or not we support the attachment
        # of this on the system, or if the vswitch does not exist on host
        # then we should give them the correct message.
        net_id = vif.get('network').get('id')
        admin_context = ctx.get_admin_context()
        napi = net_api.API()

        net_info = napi.get(admin_context, net_id)
        physical_network = net_info.get('physical_network')

        host_ovs = self.get_host_ovs(admin_context)
        vswitches = host_ovs['host-ovs-network-config']['vswitches']

        local_host_viable = False
        for vswitch in vswitches:
            if physical_network == vswitch['ovs']['name']:
                local_host_viable = True
                break

        # If PowerVC created this port, do not preserve on delete
        port = napi.show_port(admin_context, vif.get('id'))
        name = port['port'].get('name', '')
        if name.startswith('powervc_attached'):
            # Clear the preserve flag
            vif['preserve_on_delete'] = False

            # Check to see if any other VIFs are already on this instance
            cur_vifs = napi.get_instance_nw_info(admin_context, instance)
            if not isinstance(cur_vifs, list):
                cur_vifs = []

            # Remove / readd the VIF we're updating
            cur_vifs[:] = [d for d in cur_vifs if d.get('id') != vif.get('id')]
            cur_vifs.append(vif)

            # Convert VIF list to NetworkInfos
            net_infos = cur_vifs
            if not isinstance(net_infos, net_model.NetworkInfo):
                net_infos = net_model.NetworkInfo(cur_vifs)

            # Update VIF list
            base_api.update_instance_cache_with_nw_info(
                napi, admin_context, instance, net_infos)

        if not local_host_viable:
            msg = _("Cannot attach the network %(network)s to the "
                    "host %(host)s because the virtual switch used "
                    "is either invalid or unavailable.  To resolve "
                    "this, use the appropriate virtual switch on "
                    "the network and try again.") % \
                {'network': vif.get('network').get('label'),
                 'host': CONF.host}
            LOG.error(msg)
            raise exception.InterfaceAttachFailed(instance)

        # Continue on to the original attach code supported by the community.
        super(PowerKVMDriver, self).attach_interface(instance, image_meta,
                                                     vif)

    @send_ui_notification(error_message=notify_msg.LPM_ERROR,
                          success_message=None,
                          migrate_op=True,
                          log_exception=True)
    @logcall
    def check_can_live_migrate_destination(self, context, instance,
                                           src_compute_info, dst_compute_info,
                                           block_migration=False,
                                           disk_over_commit=False):
        """
        Validate the destination host is capable of live partition migration.

        :param ctxt: security context
        :param instance_ref: instance to be migrated
        :param src_compute_info: source host information
        :param dst_compute_info: destination host information
        :param block_migration: if true, prepare for block migration
        :param disk_over_commit: if true, allow disk over commit
        :returns dest_check_data: dictionary containing destination data
        """
        # Print an informational trace message to the logs
        tgt_sys_name = CONF.host_display_name
        tgt_hyp_name = dst_compute_info.get('hypervisor_hostname')

        msg = (_('Validating destination host %(host)s for live migration')
               % {'host': tgt_sys_name})
        log_operation(msg=msg)

        # Check if the tls port is open
        if not self.check_tls_firewall_port():
            if CONF.open_libvirt_tls_firewall_port:
                LOG.info(_("Opening libvirt TLS port 16514 on destination "
                           "host %(host)s for migration.") %
                         {'host': tgt_sys_name})
                self.open_tls_firewall_port()
            else:
                error = (_("Cannot migrate %(lpar)s because libvirt TLS port "
                           "16514 on destination host %(host)s is closed. "
                           "To resolve the problem, open the port and try "
                           "to migrate the virtual machine again.") %
                         {'lpar': instance['display_name'],
                          'host': tgt_sys_name})
                raise exception.MigrationPreCheckError(reason=error)

        # extract the host capability to fill in the cpu.model
        dest_check_data = super(PowerKVMDriver, self). \
            check_can_live_migrate_destination(context, instance,
                                               src_compute_info,
                                               dst_compute_info,
                                               block_migration,
                                               disk_over_commit)
        pvc_dest_check_data = PowerVCLibvirtLiveMigrateData()

        pvc_dest_check_data.dest_sys_name = tgt_sys_name
        pvc_dest_check_data.dest_hyp_name = tgt_hyp_name
        pvc_dest_check_data._convert_libvirtobj_to_pvckvm(dest_check_data)

        return pvc_dest_check_data

    @send_ui_notification(error_message=notify_msg.LPM_ERROR,
                          success_message=None, migrate_op=True,
                          log_exception=True)
    @logcall
    def check_can_live_migrate_source(self, ctxt, instance, dest_check_data,
                                      block_device_info=None):
        """
        Validate the source host is capable of live partition migration.

        :param context: security context
        :param instance: instance to be migrated
        :param dest_check_data: results from check_can_live_migrate_destination
        :returns migrate_data: dictionary containing source and
            destination data for migration
        """
        # Print an informational trace message to the logs
        src_sys_name = CONF.host_display_name
        msg = (_('Validating source host %(host)s for live migration')
               % {'host': src_sys_name})
        log_operation(msg=msg)

        migrate_data = super(PowerKVMDriver, self). \
            check_can_live_migrate_source(ctxt, instance, dest_check_data,
                                          block_device_info=block_device_info)
        pvc_mig_data = PowerVCLibvirtLiveMigrateData()
        # Update the return dictionary with PowerVC specific entries
        pvc_mig_data.refresh_conn_info = True
        pvc_mig_data.src_sys_name = src_sys_name
        pvc_mig_data._convert_libvirtobj_to_pvckvm(migrate_data)
        return pvc_mig_data

    @send_ui_notification(error_message=notify_msg.LPM_ERROR,
                          success_message=None, migrate_op=True,
                          log_exception=True)
    @logcall
    def pre_live_migration(self, context, instance, block_device_info,
                           network_info, disk_info, migrate_data=None):
        """
        Perfoms any required prerequisites on the destination
        host prior to live partition migration.

        :param context: security context
        :param instance: instance to be migrated
        :param block_device_info: instance block device information
        :param network_info: instance network information
        :param migrate_data: implementation specific data dictionary
        """
        # Print an informational trace message to the logs
        msg = (_('Preparing destination host %(host)s for live migration')
               % {'host': migrate_data.dest_sys_name})
        log_operation(msg=msg)

#        super(PowerKVMDriver, self).pre_live_migration(context, instance,
#                                                       block_device_info,
#                                                       network_info,
#                                                       disk_info,
#                                                       migrate_data)
#
# Overriding to fix indentation bug in libvirt.driver
        """Preparation live migration."""
        # Steps for volume backed instance live migration w/o shared storage.
        is_shared_block_storage = True
        is_shared_instance_path = True
        is_block_migration = True
        instance_relative_path = None
        if migrate_data:
            is_shared_block_storage = migrate_data.is_shared_block_storage
            is_shared_instance_path = migrate_data.is_shared_instance_path
            is_block_migration = migrate_data.block_migration
            instance_relative_path = migrate_data.instance_relative_path

        if not (is_shared_instance_path and is_shared_block_storage):
            # NOTE(mikal): live migration of instances using config drive is
            # not supported because of a bug in libvirt (read only devices
            # are not copied by libvirt). See bug/1246201
            # if configdrive.required_by(instance):
            #    raise exception.NoLiveMigrationForConfigDriveInLibVirt()
            pass

        if not is_shared_instance_path:
            # NOTE(mikal): this doesn't use libvirt_utils.get_instance_path
            # because we are ensuring that the same instance directory name
            # is used as was at the source
            if instance_relative_path:
                instance_dir = os.path.join(CONF.instances_path,
                                            instance_relative_path)
            else:
                instance_dir = libvirt_utils.get_instance_path(instance)

            if os.path.exists(instance_dir):
                raise exception.DestinationDiskExists(path=instance_dir)
            os.mkdir(instance_dir)

            if not is_shared_block_storage:
                # Ensure images and backing files are present.
                self._create_images_and_backing(context, instance,
                                                instance_dir, disk_info)

        if not (is_block_migration or is_shared_instance_path):
            # NOTE(angdraug): when block storage is shared between source and
            # destination and instance path isn't (e.g. volume backed or rbd
            # backed instance), instance path on destination has to be prepared

            # Touch the console.log file, required by libvirt.
            console_file = self._get_console_log_path(instance)
            libvirt_utils.file_open(console_file, 'a').close()

            # if image has kernel and ramdisk, just download
            # following normal way.
            self._fetch_instance_kernel_ramdisk(context, instance)

        # Establishing connection to volume server.
        block_device_mapping = driver.block_device_info_get_mapping(
            block_device_info)
        image_meta = utils.get_image_from_system_metadata(
            instance['system_metadata'])
        for vol in block_device_mapping:
            connection_info = vol['connection_info']
            disk_info = blockinfo.get_info_from_bdm(
                instance, CONF.libvirt.virt_type, image_meta, vol)
            self._connect_volume(connection_info, disk_info)

        # We call plug_vifs before the compute manager calls
        # ensure_filtering_rules_for_instance, to ensure bridge is set up
        # Retry operation is necessary because continuously request comes,
        # concurrent request occurs to iptables, then it complains.
        max_retry = CONF.live_migration_retry_count
        for cnt in range(max_retry):
            try:
                self.plug_vifs(instance, network_info)
                break
            except processutils.ProcessExecutionError:
                if cnt == max_retry - 1:
                    raise
                else:
                    LOG.warn(('plug_vifs() failed %(cnt)d. Retry up to '
                              '%(max_retry)d.'),
                             {'cnt': cnt,
                              'max_retry': max_retry},
                             instance=instance)
                    greenthread.sleep(1)

        if not migrate_data:
            migrate_data = PowerVCLibvirtLiveMigrateData(bdms=[])
        else:
            migrate_data.bdms = []
        migrate_data.graphics_listen_addr_vnc = CONF.vnc.vncserver_listen
        migrate_data.graphics_listen_addr_spice = CONF.spice.server_listen

        migrate_data.serial_listen_addr = \
            CONF.serial_console.proxyclient_address
        # Store live_migration_inbound_addr
        migrate_data.target_connect_addr = \
            CONF.libvirt.live_migration_inbound_addr

        for vol in block_device_mapping:
            connection_info = vol['connection_info']
            if connection_info.get('serial'):
                disk_info = blockinfo.get_info_from_bdm(
                    instance, CONF.libvirt.virt_type,
                    instance.image_meta, vol)

                bdmi = LibvirtLiveMigrateBDMInfo()
                bdmi.serial = connection_info['serial']
                bdmi.connection_info = connection_info
                bdmi.bus = disk_info['bus']
                bdmi.dev = disk_info['dev']
                bdmi.type = disk_info['type']
                bdmi.format = disk_info.get('format')
                bdmi.boot_index = disk_info.get('boot_index')
                migrate_data.bdms.append(bdmi)

        return migrate_data

    @send_ui_notification(error_message=notify_msg.LPM_ERROR,
                          success_message=None, migrate_op=True,
                          log_exception=True)
    @logcall
    def live_migration(self, ctxt, instance_ref, dest,
                       post_method, recover_method,
                       block_migration=False, migrate_data=None):
        """
        Live migrates a partition from one host to another.

        :param ctxt: security context
        :param instance_ref: instance to be migrated.
        :param dest: destination host
        :param post_method: post operation method.
            nova.compute.manager.post_live_migration.
        :param recover_method: recovery method when any exception occurs.
            nova.compute.manager.recover_live_migration.
        :param block_migration: if true, migrate VM disk.
        :param migrate_data: implementation specific data dictionary.
        """
        src_sys_name = migrate_data.src_sys_name
        tgt_sys_name = migrate_data.dest_sys_name
        inst_name = instance_ref['display_name']

        # Print an informational trace message to the logs
        msg = (_('Migrating virtual machine %(inst)s from host %(src)s '
                 'to host %(tgt)s.')
               % {'inst': inst_name,
                  'src': src_sys_name,
                  'tgt': tgt_sys_name})
        log_operation(msg=msg)

        super(PowerKVMDriver, self).live_migration(ctxt, instance_ref,
                                                   dest, post_method,
                                                   recover_method,
                                                   block_migration,
                                                   migrate_data)

    @send_ui_notification(error_message=notify_msg.LPM_ERROR,
                          success_message=None, migrate_op=True,
                          log_exception=True)
    @logcall
    def _live_migration(self, context, instance, dest, post_method,
                        recover_method, block_migration,
                        migrate_data):
        """Do live migration.

        :param context: security context
        :param instance:
            nova.db.sqlalchemy.models.Instance object
            instance object that is migrated.
        :param dest: destination host
        :param post_method:
            post operation method.
            expected nova.compute.manager.post_live_migration.
        :param recover_method:
            recovery method when any exception occurs.
            expected nova.compute.manager.recover_live_migration.
        :param migrate_data: implementation specific params
        """

        # Do live migration.
        try:
            if block_migration:
                flaglist = CONF.libvirt.block_migration_flag.split(',')
            else:
                flaglist = CONF.libvirt.live_migration_flag.split(',')
            # flagvals = [getattr(libvirt, x.strip()) for x in flaglist]
            # Pull the flag values from the global variable list
            # instead of the libvirt module
            flagvals = [globals().get(x.strip()) for x in flaglist]
            LOG.debug("flagvals = %s" % flagvals)
            logical_sum = reduce(lambda x, y: x | y, flagvals)

            # Use the hypervisor hostname for the migration (defect 17144)
            hyp_hostname = migrate_data.dest_hyp_name
            migration_uri = CONF.libvirt.live_migration_uri % hyp_hostname
            LOG.debug("Migration URI = %s" % migration_uri)

            dom = self._host._get_domain_by_name(instance["name"])

            listen_addrs = {}
            if 'graphics_listen_addr_vnc' in migrate_data:
                listen_addrs['vnc'] = str(
                    migrate_data.graphics_listen_addr_vnc)
            if 'graphics_listen_addr_spice' in migrate_data:
                listen_addrs['spice'] = str(
                    migrate_data.graphics_listen_addr_spice)
            volume = migrate_data.bdms

            migratable_flag = getattr(libvirt, 'VIR_DOMAIN_XML_MIGRATABLE',
                                      None)

            # copy from libvirt driver
            if (migratable_flag is None or
                    (listen_addrs is None and not volume)):
                self._check_graphics_addresses_can_live_migrate(listen_addrs)
                dom.migrateToURI(migration_uri, logical_sum, None,
                                 CONF.libvirt.live_migration_bandwidth)
            else:
                old_xml_str = dom.XMLDesc(migratable_flag)
                new_xml_str = self._update_xml(old_xml_str,
                                               volume,
                                               listen_addrs, None)
                try:
                    dom.migrateToURI2(migration_uri,
                                      None,
                                      new_xml_str,
                                      logical_sum,
                                      None,
                                      CONF.libvirt.live_migration_bandwidth)
                except libvirt.libvirtError as ex:
                    # NOTE(mriedem): There is a bug in older versions of
                    # libvirt where the VIR_DOMAIN_XML_MIGRATABLE flag causes
                    # virDomainDefCheckABIStability to not compare the source
                    # and target domain xml's correctly for the CPU model.
                    # We try to handle that error here and attempt the legacy
                    # migrateToURI path, which could fail if the console
                    # addresses are not correct, but in that case we have the
                    # _check_graphics_addresses_can_live_migrate check in place
                    # to catch it.
                    # TODO(mriedem): Remove this workaround when
                    # Red Hat BZ #1141838 is closed.
                    error_code = ex.get_error_code()
                    if error_code == libvirt.VIR_ERR_CONFIG_UNSUPPORTED:
                        LOG.warn(_('An error occurred trying to live '
                                   'migrate. Falling back to legacy live '
                                   'migrate flow. Error: %s'),
                                 six.text_type(ex),
                                 instance=instance)
                        self._check_graphics_addresses_can_live_migrate(
                            listen_addrs)
                        dom.migrateToURI(
                            migration_uri,
                            logical_sum,
                            None,
                            CONF.libvirt.live_migration_bandwidth)
                    else:
                        raise
        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.error(_("Live Migration failure: %s"), six.text_type(e),
                          instance=instance)
                # Add the migrate_data parameter to the recover method
                # This is a bug in the openstack libvirt driver IMO
                recover_method(context, instance, dest, block_migration,
                               migrate_data)

        # Waiting for completion of live_migration.
        timer = loopingcall.FixedIntervalLoopingCall(f=None)

        def wait_for_live_migration():
            """waiting for live migration completion."""
            try:
                super(PowerKVMDriver, self).get_info(instance).state
            except exception.InstanceNotFound:
                timer.stop()
                post_method(context, instance, dest, block_migration,
                            migrate_data)

        timer.f = wait_for_live_migration
        timer.start(interval=0.5).wait()

    @send_ui_notification(error_message=notify_msg.LPM_ERROR,
                          success_message=None, migrate_op=True,
                          log_exception=True)
    @logcall
    def post_live_migration(self, ctxt, instance_ref, block_device_info,
                            migrate_data=None):
        """
        Performs post operations on the source host
        following a successful live migration.

        :param ctxt: security context
        :param instance_ref: migrated instance
        :param block_device_info: instance volume block device info
        :param migrate_data: implementation specific data dictionary.
        """
        # Print an informational trace message to the logs
        msg = _("Performing post-live migration activities on the source "
                "host after the successful live migration of '%(vm)s'") %\
            dict(vm=instance_ref['name'])
        log_operation(msg=msg)

        super(PowerKVMDriver, self).post_live_migration(ctxt, instance_ref,
                                                        block_device_info,
                                                        migrate_data)

    @send_ui_notification(error_message=notify_msg.LPM_ERROR,
                          success_message=notify_msg.LPM_SUCCESS,
                          migrate_op=True, log_exception=True)
    @logcall
    def post_live_migration_at_destination(self, ctxt, instance_ref,
                                           network_info,
                                           block_migration=False,
                                           block_device_info=None,
                                           migrate_data=None):
        """
        Performs post operations on the destination host
        following a successful live migration.

        :param ctxt: security context
        :param instance_ref: migrated instance
        :param network_info: dictionary of network info for instance
        :param block_migration: boolean for block migration
        :param block_device_info: instance volume block device info
        :param migrate_data: implementation specific data dictionary.
        """

        # Print an informational trace message to the logs
        msg = _("Performing post-live migration activities, including "
                "persisting the virtual machine definition on the destination "
                "host after the successful live migration of '%(vm)s'") %\
            dict(vm=instance_ref['name'])
        log_operation(msg=msg)

        super(PowerKVMDriver, self).post_live_migration_at_destination(
            ctxt, instance_ref, network_info,
            block_migration, block_device_info)

    @send_ui_notification(error_message=notify_msg.LPM_ERROR,
                          success_message=None, migrate_op=True,
                          log_exception=True)
    @logcall
    def rollback_live_migration_at_destination(self, context, instance,
                                               network_info,
                                               block_device_info,
                                               destroy_disks=True,
                                               migrate_data=None):
        """
        Performs post operations on the destination host
        following a failing live migration.

        :param context: security context
        :param instance: failed migrated instance
        :param network_info: dictionary of network info for instance
        :param block_device_info: instance volume block device info
        """
        # Print an informational trace message to the logs
        msg = (_('Cleaning up destination host after failed live migration'))
        log_operation(msg=msg)

        # Override the libvirt method so that we do not delete any disks.
        # Also, only attempt the destroy if the instance exists.
        instance_exists = True
        try:
            self._host._get_domain_by_name(instance['name'])
        except exception.NovaException:
            instance_exists = False

        if instance_exists:
            self.destroy(context, instance, network_info, block_device_info,
                         destroy_disks=False)

    def migration_cleanup(self, context, instance, source=False):
        """
        This is called from the powervc compute manager, and currently
        only used for HMC/IVM migration.
        """
        pass

    def remove_volume_device_at_destination(self, context, instance, cinfo):
        """ clean up volume device at destination during Rollback

        :param context: security context
        :param instance: instance dictionary of failed migration
        :param cinfo: block_disk_map['connection_info']

        Called from manager.py, and needed only for IVM migration.
        """
        pass

    @logcall
    def check_tls_firewall_port(self):
        """
        This method issues the 'firewall-cmd --list-services'
        command on the PowerKVM host checking that the libvirt-tls
        service is listed.  If a service is listed that port is open.

        :return: True if the libvirt-tls port is open, otherwise False.
        """
        try:
            stdout, stderr = utils.execute('firewall-cmd', '--list-services',
                                           run_as_root=True,
                                           check_exit_code=True)
            if stdout:
                LOG.debug('List of firewall services: %s' % stdout)

                # Check for libvirt-tls
                if 'libvirt-tls' in stdout.split(' '):
                    return True

            return False
        except processutils.ProcessExecutionError as error:
            LOG.info(_("Unable to list the firewall services: %(error)s")
                     % locals())

            # Pull the exit code from the error message. If the firewall
            # service is off, the exit code is 252. We can allow the migration
            if error.exit_code == 252:
                LOG.debug("Firewall service is not started. "
                          "Will allow migration.")
                return True
            else:
                raise

    @logcall
    def open_tls_firewall_port(self):
        """
        This method issues the 'firewall-cmd --add-service=libvirt-tls'
        command on the PowerKVM host temporarily opening the libvirt-tls port.
        """
        try:
            utils.execute('firewall-cmd', '--add-service=libvirt-tls',
                          run_as_root=True, check_exit_code=True)
        except processutils.ProcessExecutionError as error:
            LOG.error(_("Unable to open firewall to allow live "
                        "migration: %(error)s") % locals())
            raise

    def prep_instance_for_attach(self, context, instance, storage_type):
        pass

    def get_available_resource(self, nodename):
        """Retrieve resource information.

        This method is called when nova-compute launches, and
        as part of a periodic task that records the results in the DB.

        This method is being pulled in from libvirt driver to do a None
        check on 'supported_instances', since this field may not
        be set if libvirtd connection is down.

        :param nodename: will be put in PCI device
        :returns: dictionary containing resource info
        """

        try:
            data = super(PowerKVMDriver, self).get_available_resource(nodename)
        except exception.HypervisorUnavailable as e:
            # If there was an error getting stats, return with
            # hypervisor state set to error
            LOG.error(_("HypervisorUnavailable: %s") % six.text_type(e))
            self._data['stats'] = dict()
            self._data['stats']['hypervisor_state'] = 'error'
            return copy.deepcopy(self._data)

        # Add PowerKVM specific stats here
        if data.get('stats') is None:
            data['stats'] = dict()
        stats = dict()
        stats['hypervisor_state'] = 'operating'
        stats[pkvmc.IBM_POWERKVM_VERSION] = proc_utils.get_powerkvm_version()
        # reset host-level remote restart
        stats[commonconsts.HOST_DEAD] = False
        cpu_info = jsonutils.loads(data['cpu_info'])
        topology = cpu_info['topology']
        stats['threads_per_core'] = topology['threads']
        stats['num_cores'] = topology['cores']
        stats['num_sockets'] = topology['sockets']
        stats['split_core'] = proc_utils.run_ppc64_cpu_subcores_command()
        stats['max_smt_per_guest'] = (
            stats['threads_per_core'] / stats['split_core'])
        stats['mgmt_version'] = data_utils.get_powervc_version()
        # add stats for host evacuate
        stats[commonconsts.PRS_MAX_PARA_REBUILDS] = \
            CONF.prs_max_parallel_rebuilds
        stats[commonconsts.PRS_MAX_PARA_DEST_REBUILDS] = \
            CONF.prs_max_parallel_dest_rebuilds
        data['stats'].update(stats)
        # Override vcpus with vcpus multiplied by subcores per core
        data['vcpus'] = data['vcpus'] * stats['split_core']

        self._data = copy.deepcopy(data)
        LOG.debug('Info for host %s is %s' %
                  (CONF.host_display_name, self._data))
        return data

    @logcall
    def migrate_disk_and_power_off(self, context, instance, dest,
                                   flavor, network_info,
                                   block_device_info=None,
                                   timeout=0, retry_interval=0):
        """
        This method is called during resize to
        migrate disk and power off virtual machine.

        :param ctxt: security context
        :param instance: instance to resize
        :param dest: destination host only populated in case of cold migration
                    (which is currently not supported in powervc)
        :param flavor: flavor with new values to resize
        :param network_info: dictionary of network info for instance
        :param block_device_info: instance volume block device info
        """
        # Print an informational trace message to the logs

        log_operation(_('Resizing'), instance)

        resource_utils.check_resource(flavor, instance)
        return super(PowerKVMDriver, self).migrate_disk_and_power_off(
            context, instance, dest,
            flavor, network_info,
            block_device_info=block_device_info,
            timeout=timeout, retry_interval=retry_interval)

    @logcall
    def finish_migration(self, context, migration, instance, disk_info,
                         network_info, image_meta, resize_instance,
                         block_device_info=None, power_on=True):
        """
        This method is called during resize to
        create a new virtual machine with resized values

        :param ctxt: security context
        :param migration: migration data
        :param instance: instance to resize
        :param disk_info: disk info attached to instances
        :param network_info: dictionary of network info for instance
        :param image_meta: image meta data
        :param block_device_info: instance volume block device info
        :param power_on: flag, using which instances state is determined
                        defaulted to 'True'
        """

        # Print an informational trace message to the logs
        msg = (_('Creating new virtual machine %(inst)s '
                 'with resized values on host %(host)s '
                 % {'host': CONF.host_display_name,
                    'inst': instance['display_name']}))
        log_operation(msg=msg)

        super(PowerKVMDriver, self).finish_migration(
            context, migration, instance, disk_info,
            network_info, image_meta, resize_instance,
            block_device_info=block_device_info, power_on=power_on)

    @logcall
    def confirm_migration(self, migration, instance, network_info):
        """
        This method is called during resize to
        clean up old virtual machine and confirms resize

        :param migration: migration data
        :param instance: instance to resize
        :param disk_info: disk info attached to instances
        :param network_info: dictionary of network info for instance
        """
        # Print an informational trace message to the logs
        msg = (_('Cleaning up original virtual machine resources '
                 'after virtual machine %(inst)s is resized on host %(host)s '
                 % {'host': CONF.host_display_name,
                    'inst': instance['display_name']}))
        log_operation(msg=msg)

        super(PowerKVMDriver, self).confirm_migration(
            migration, instance, network_info)

    # Override in order to avoid running mkfs on ephemeral volume,
    # which causes RHEL6.5 to hang during disk detection.
    @staticmethod
    def _create_ephemeral(target, ephemeral_size,
                          fs_label, os_type, is_block_dev=False,
                          max_size=None, context=None, specified_fs=None):
        if not is_block_dev:
            libvirt_utils.create_image('raw', target, '%dG' % ephemeral_size)

    @logcall
    def handle_evacuated_instances(self, context, db_instances, drv_instances):
        """Handle the case when an instance is migrated out of band
           on PowerKVM hosts
        """
        # If we were asked not to handle the out-of-band changes, just return
        if not CONF.ibmpowervm_handle_out_of_band_migration:
            return
        host_name = CONF.host
        all_instances = None

        # Create a list of instance uuids from the DB
        managed_instance_uuids = []
        for db_instance in db_instances:
            instance_uuid = db_instance.uuid
            if instance_uuid is not None:
                managed_instance_uuids.append(instance_uuid.lower())
        for uuid in self.list_instance_uuids():
            try:
                if uuid.lower() not in managed_instance_uuids:
                    if all_instances is None:
                        filters = {'deleted': False}
                        all_instances = instance_obj.InstanceList.\
                            get_by_filters(context,
                                           filters)
                    # Go through the list to find the one with matching UUID
                    for instance in all_instances:
                        uuid_in_db = instance.uuid
                        host_in_db = instance.host
                        if (uuid_in_db is not None and
                                uuid.lower() == uuid_in_db.lower() and
                                host_in_db is not None and
                                host_in_db != host_name):
                            vm_info = self.get_info(instance)
                            node = self.get_available_nodes()[0]
                            migrate_utils.\
                                verify_out_of_band_evacuation(context,
                                                              instance,
                                                              uuid,
                                                              uuid_in_db,
                                                              host_in_db,
                                                              host_name,
                                                              vm_info,
                                                              node)
            except Exception as e:
                LOG.error(_('Exception encountered while processing migrated '
                            'instances. Error: %s') % six.text_type(e))

    def _get_disk_over_committed_size_total(self):
        """Return total over committed disk size for all instances.

           Override to only total up over committed size for disks
           in the filesystem of the storage path defined during
           host registration, and to ignore exceptions raised by
           qemu-img command, such as when the disk does not
           have correct permissions set.
        """
        domains = self._host.list_instance_domains(only_running=False)
        # Directory defined during host registration
        storage_dir = re.sub('/nova/instances', '', CONF.instances_path)
        # Disk size that all instance uses : virtual_size - disk_size
        disk_over_committed_size = 0
        for dom in domains:
            try:
                xml = dom.XMLDesc(0)
                disk_infos = self._get_instance_disk_info(dom.name(), xml)
                for info in disk_infos:
                    if info['path'].startswith(storage_dir):
                        disk_over_committed_size += int(
                            info['over_committed_disk_size'])
            except OSError as e:
                if e.errno == errno.ENOENT:
                    LOG.warning(_('Periodic task is updating the host stat, '
                                  'it is trying to get disk %(i_name)s, '
                                  'but disk file was removed by concurrent '
                                  'operations such as resize.'),
                                {'i_name': dom.name()})
                else:
                    raise

            except processutils.ProcessExecutionError as error:
                # This exception will be raised if there was an error running
                # the qemu-img command. For example, incorrect permissions set
                # on the disk file. Log a warning and ignore this instance in
                # the calculation.
                LOG.warning(_('Unable to obtain allocation size '
                              'of disk for %(instance)s. %(error)s.'),
                            {'instance': dom.name(),
                             'error': error})

            except exception.InstanceNotFound:
                # Instance was deleted during the check so ignore it
                pass
            except exception.InvalidDiskInfo as disk_err:
                # It's likely that the instance is being deleted,
                # so log the exception
                LOG.warning(_('Periodic task is updating the host stat, '
                              'retrieving disk information for instance '
                              '%(i_name)s, but disk file was likely removed '
                              'by concurrent operations such as resize. '
                              'Error: %(err)s'),
                            {'i_name': dom.name(), 'err': disk_err})
            except exception.VolumeBDMPathNotFound as disk_err:
                # Volume file link is no longer available.
                LOG.warning(_('A periodic task was trying to '
                              'collect disk information for virtual machine '
                              '%(i_name)s, but the disk file was likely '
                              'removed by concurrent operations. Error: '
                              '%(err)s'), {'i_name': dom.name(),
                                           'err': disk_err})
            except Exception as ex:
                LOG.warning(_("A general error occurred when trying to "
                              "collect disk information for virtual machine "
                              "instance %(i_name)s. Error: %(err)s"),
                            {'i_name': dom.name(), 'err': ex})
            # NOTE(gtt116): give change to do other task.
            greenthread.sleep(0)

        # Add disk_over_committed_size to disk_over_committed size for
        # each of the hosts that share the same storage to get total
        # committed disks space in the storage directory.
        disk_over_committed_size = resource_utils.get_all_disk_committed(
            disk_over_committed_size)
        return disk_over_committed_size

    # Override to set a CD-ROM and local disk
    # order in XML that allows capture of ISO-deployed VMs.
    def _get_guest_storage_config(self, instance, image_meta,
                                  disk_info,
                                  rescue, block_device_info,
                                  inst_type, os_type):

        devices = super(PowerKVMDriver, self)._get_guest_storage_config(
            instance, image_meta, disk_info, rescue, block_device_info,
            inst_type, os_type)

        # If this is an ISO deploy, swap order of disks so that
        # the local OS disk is first. The base class function
        # puts the CD first, while snapshot captures the first
        # device. Verify that the expected two devices are there
        # in the expected order before doing the swap.
        if (image_meta.obj_attr_is_set('disk_format') and
                image_meta.disk_format == 'iso'):
            if (len(devices) == 2 and devices[0].source_device == 'cdrom' and
                    devices[1].source_device == 'disk'):
                LOG.info(_("Swapping order of disk devices for ISO deploy."))
                devices[0], devices[1] = devices[1], devices[0]
            else:
                LOG.info(_("Leaving device order intact for ISO deploy."))

        return devices

    def is_instance_on_host(self, instance_uuid, instance_name):
        """
        Check if an instance exists on this host
        """
        ret = False
        if instance_uuid in self.list_instance_uuids():
            ret = True
        return ret

    def _destroy(self, instance):

        libvirt = __import__('libvirt')

        def _wait_for_device(instance, attempt=1):
            # Destroy can fail over NFS, because Device is busy.
            # So retry the destroy again till max attempts
            try:
                guest = self._host.get_guest(instance)
            except exception.InstanceNotFound:
                guest = None

            if guest:
                try:
                    super(PowerKVMDriver, self)._destroy(instance, attempt)
                    raise loopingcall.LoopingCallDone()
                except libvirt.libvirtError as e:
                    error_code = e.get_error_code()
                    if error_code == libvirt.VIR_FROM_STREAMS:
                        LOG.warning(_("Device busy error received"
                                      "while shutting down VM"),
                                    instance=instance)
                        max_retry_attempts = kwargs['max_retry_attempts']
                        if max_retry_attempts == 0:
                            raise loopingcall.LoopingCallDone()
                        else:
                            max_retry_attempts = max_retry_attempts - 1
                            kwargs['max_retry_attempts'] = max_retry_attempts
                    else:
                        with excutils.save_and_reraise_exception():
                            LOG.error(_('Error from libvirt during destroy. '
                                        'Code=%(errcode)s Error=%(e)s'),
                                      {'errcode': error_code,
                                       'e': six.text_type(e)},
                                      instance=instance)
            else:
                guests = self._host.list_guests()
                for guest in guests:
                    if instance.uuid == guest.uuid:
                        try:
                            old_domid = guest.id
                            guest.poweroff()
                            raise loopingcall.LoopingCallDone()
                        except libvirt.libvirtError as e:
                            error_code = e.get_error_code()
                            if error_code == libvirt.VIR_FROM_STREAMS:
                                LOG.warning(_("Device busy error received"
                                              "while shutting down VM"),
                                            instance=instance)
                                max_retry_attempts = \
                                    kwargs['max_retry_attempts']
                                if max_retry_attempts == 0:
                                    raise loopingcall.LoopingCallDone()
                                else:
                                    max_retry_attempts = max_retry_attempts - 1
                                    kwargs['max_retry_attempts'] = \
                                        max_retry_attempts
                            else:
                                with excutils.save_and_reraise_exception():
                                    LOG.error(_('Error from libvirt during '
                                                'destroy. Code=%(errcode)s '
                                                'Error=%(e)s'),
                                              {'errcode': error_code,
                                               'e': six.text_type(e)},
                                              instance=instance)
                # if no guest exists with matching UUID, end call
                raise loopingcall.LoopingCallDone()
        # print 'Error from libvirt during destroy.Error=%s ' % e
        kwargs = {'max_retry_attempts': 5}

        timer = loopingcall.FixedIntervalLoopingCall(_wait_for_device,
                                                     instance)
        timer.start(interval=0.5).wait()

    def _get_hypervisor_hostname(self):
        """Returns the hostname of the hypervisor. Override to send
           notification if hostname has changed.
        """
        hostname = self._conn.getHostname()
        if not hasattr(self, '_hypervisor_hostname'):
            self._hypervisor_hostname = hostname
        elif hostname != self._hypervisor_hostname:
            # If hostname was changed out of band and is different
            # than the original host name, log an error.
            error_args = {'old_hostname': self._hypervisor_hostname,
                          'new_hostname': hostname}
            LOG.error(_('Hostname has changed from %(old_hostname)s '
                        'to %(new_hostname)s. A restart is required to '
                        'take effect.') % error_args)
            if hostname != self._prev_hostname:
                # If this hostname is different than the previous hostname,
                # send a notification
                payload = {'msg': notify_msg.HOSTNAME_CHANGED_ERROR}
                payload.update(error_args)
                notifier = rpc.get_notifier(service='compute',
                                            host=CONF.host)
                context = nova_context.get_admin_context(read_deleted='yes')
                notifier.error(context, 'compute.instance.log', payload)
        # Keep track of what hostname was for the next time through this method
        self._prev_hostname = hostname
        return self._hypervisor_hostname

    def unmanage_host(self, context):
        """Called when unmanaging the host. Cleans up the host
           from the compute_node_usage file in the storage directory.
        """
        resource_utils.remove_from_usage_file()

    @logcall
    def discover_instances(self, context):
        """
        Returns a list of all of the VM's that exist on the given Host.
        For each VM the driver needs to return a dictionary containing
        the following attributes:
            name:      The Name of the VM defined on the Hypervisor
            state:     The State of the VM, matching the power_state definition
            vm_uuid:   The UUID of the VM from the domain XML
            host_id:   The libvirt ID of the VM
            supported: Dictionary stating whether the VM can be managed
              status:  Whether or not it is supported or if preparation
                       is required.
              reasons: List of dictionaries as to why it isn't supported
                id:      Message ID to key off of for the reason
                message: Message Text as to why it isn't supported
              steps: List of dictionaries of steps required for preparation

        :param context:  The security context for the query
        """
        disc_insts = []
        instances = self.list_instances()
        importvm = vm_import.ImportVM(self._conn)
        info = importvm.list_instances_for_onboard(instances)
        disc_insts += info
        return disc_insts

    def get_instance_path(self, instance, forceold=False, relative=False):
        """
        Defect 42007
        Replaces get_instance_path in nova/virt/libvirt/utils.
        This is done because get_instance_path was changed to
        reference attributes in an instance object, but PowerVC cannot
        construct the object in prepare_instance. Dictionary access
        is used here instead.
        """
        pre_grizzly_name = os.path.join(CONF.instances_path, instance['name'])
        if forceold or os.path.exists(pre_grizzly_name):
            if relative:
                return instance['name']
            return pre_grizzly_name

        if relative:
            return instance['uuid']
        return os.path.join(CONF.instances_path, instance['uuid'])

    @logcall
    def prepare_instance(self, context, instance, preparation):
        """
        This is the driver entry point for the prepare phase of VM import. It
        takes in a dictionary of preparation steps for a single instance and
        performs the necassary actions to prepare this VM for import into
        PowerVC.

        :param instance: A dictionary of information about the instance to
                         prepare.
        :param preparation: A dictionary of information to use in preparation
                            of the instance.
        """

        def _convert_inst_dict():
            """Convert the instance dictionary to EnhancedDict in order
            to run libvirt driver method for instance that has not
            yet imported.
            """
            updated_inst = EnhancedDict()
            for k, v in six.iteritems(instance):
                updated_inst[k] = v
            return updated_inst

        name = instance['name']

        try:
            domain = self._conn.lookupByName(name)
        except Exception as ex:
            LOG.exception(_("Error from libvirt while looking up"
                            " %(instance_name)s: %(ex)s")
                          % {'instance_name': name,
                             'ex': ex})
            raise

        # Set instance attrs required for upcoming method calls
        curr_state = domain.info()[0]
        curr_state = LIBVIRT_POWER_STATE[curr_state]
        instance['vm_state'] = curr_state
        instance['uuid'] = domain.UUIDString()
        instance['task_state'] = None

        # Save off current XML to file in instane storage directory
        xml = domain.XMLDesc(0)
        instance_dir = self.get_instance_path(instance)
        if not os.path.exists(instance_dir):
            os.mkdir(instance_dir, 0777)
        xml_path = os.path.join(instance_dir, 'pre_import.xml')
        libvirt_utils.write_to_file(xml_path, xml)

        # Power off the VM
        orig_state = curr_state
        if curr_state != power_state.SHUTDOWN:
            LOG.info(_("Powering off virtual machine %s as part of the"
                       " virtual machine management process.") % name)
            try:
                new_inst = _convert_inst_dict()
                self.power_off(new_inst)
            except Exception as e:
                LOG.error(_("An error occurred while trying to shutdown"
                            " domain %(name)s: %(error)s")
                          % {'name': name,
                             'error': six.text_type(e)})
                raise

        # Do the actual preparation for the instance and update its config
        importvm = vm_import.ImportVM(self._conn)
        try:
            vm_info = importvm.prepare_instance_for_onboard(
                domain, preparation, orig_state)
            return vm_info
        except Exception as e:
            # If an error occured during prepare, try to put the VM back in its
            # original state.
            LOG.info(_("Rolling back the state of virtual machine %s since"
                       " onboard preparation failed.") % name)
            if orig_state != power_state.SHUTDOWN:
                try:
                    LOG.info(_("Powering on virtual machine %s as part of the"
                               " virtual machine management process.") % name)
                    domain.create()

                except Exception:
                    LOG.warn(_("An error occurred while trying to launch"
                               " defined domain %(name)s with xml:"
                               " %(xml)s") % {'name': instance['name'],
                                              'xml': domain.XMLDesc(0)})
                    raise e

            # If the VM was paused when we began onboarding, pause the VM now
            if orig_state == power_state.PAUSED:
                try:
                    LOG.info(_("Pausing virtual machine %s as part of the"
                               " virtual machine management process.") % name)
                    domain.suspend()

                except Exception:
                    LOG.warn(_("An error occurred while trying to pause"
                               " defined domain %(name)s with xml:"
                               " %(xml)s") % {'name': instance['name'],
                                              'xml': domain.XMLDesc(0)})
            raise e

    @logcall
    def finish_onboard(self, context,
                       instance,
                       preparation,
                       network_info,
                       block_device_info):

        # If this isn't a prepare, the VM was never shutdown, so return
        if not preparation or instance['uuid'] not in preparation['servers']:
            for iface in network_info:
                try:
                    stdout, stderr = CommandExecutor.run_vsctl(
                        ['--format=json',
                         '--column=name',
                         'find',
                         'interface',
                         'external_ids:attached-mac=\"%s\"' %
                         iface['address']])

                    if not stdout:
                        LOG.error(_("Unable to find an Open vSwitch interface "
                                    "for the virtual machine %s.") %
                                  instance['name'])
                        break

                    iface_devname = jsonutils.loads(stdout)['data'][0][0]

                    iface_id = iface['ovs_interfaceid']

                    CommandExecutor.run_vsctl(['set', 'interface',
                                               iface_devname,
                                               'external_ids:iface-id=%s' %
                                               iface_id], True)
                except Exception:
                    LOG.error(_("An error occurred while attempting to "
                                "update the interface id for "
                                "openvswitch port %(iface_name)s.") %
                              {"iface_name": iface_devname})
            return

        try:
            name = instance['name']
            image_meta = utils.get_image_from_system_metadata(
                instance['system_metadata'])
            image_meta = img_meta.ImageMeta.from_dict(image_meta)
            disk_info = blockinfo.get_disk_info(
                CONF.libvirt.virt_type, instance, image_meta,
                block_device_info=block_device_info)
            new_xml = super(PowerKVMDriver, self)._get_guest_xml(
                context, instance, network_info, disk_info, image_meta,
                block_device_info=block_device_info,
                write_to_disk=True)
            LOG.info(_("Defining virtual machine %(name)s to be managed"
                       " using xml: %(xml)s") % {'name': name,
                                                 'xml': new_xml})
            self._undefine_domain(instance)
            domain = self._conn.defineXML(new_xml)

        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.error(_("An error occurred while trying to "
                            "define domain %s" % name))

        # Power on the VM if it was running before onboarding began
        if instance['vm_state'] != vm_states.STOPPED:
            try:
                LOG.info(_("Powering on virtual machine %s as part of the"
                           " virtual machine management process.") % name)
                domain.create()

            except Exception:
                with excutils.save_and_reraise_exception():
                    LOG.error(_("An error occurred while trying to launch "
                                "defined domain %(name)s with xml: %(xml)s") %
                              {'name': instance['name'],
                               'xml': domain.XMLDesc(0)})

        # If the VM was paused when we began onboarding, pause the VM now
        if instance['vm_state'] == vm_states.PAUSED:
            try:
                LOG.info(_("Pausing virtual machine %s as part of the virtual "
                           "machine management process.") % name)
                domain.suspend()

            except Exception:
                with excutils.save_and_reraise_exception():
                    LOG.error(_("An error occurred while trying to pause "
                                "defined domain %(name)s with xml: %(xml)s") %
                              {'name': instance['name'],
                               'xml': domain.XMLDesc(0)})

    @logcall
    def query_instances(self, context, instances):
        """
        Returns a list of VM's (matching those specified on input) with
        enough additional details about each VM to be able to import the
        VM's into the Nova Database such as OpenStack can start managing.
        For each VM the driver needs to return a dictionary containing
        the following attributes:
            uuid:      The UUID of the VM from the domain XML
            host_id:   The libvirt ID of the VM
            name:      The Name of the VM defined on the hypervisor
            state:     The State of the VM, matching the power_state definition
            vcpus:     The number of virtual CPU's currently allocated
            architecture:  The processor architecture
            memory_mb: The amount of memory currently allocated
            root_gb:   The amount of storage currently allocated
            ephemeral_gb:  The amount of ephemeral storage currently allocated
            sytem_metadata:  Dictionary containing system metadata
              vcpus:     The number of virtual CPU's defined for allocation
              image_hw_vif_model: The vif model
            flavor:      Dictionary containing the desired allocations
              vcpus:     The number of virtual CPU's defined for allocation
              memory_mb: The amount of memory defined for allocation
              powerkvm:smt:  The SMT value for the VM. -1 if no SMT
            volumes:   List of dictionary objects with the attached volumes:
              uuid:      The UUID of the Volume when created thru OS (Optional)
              name:      The Name of the Volume defined on the Backend
              host:      The registered name of the Storage Provider
              provider_location:  The Volume ID known on the Backend (Optional)
            network_ports: List of dictionary objects with the network ports:
              mac_address:  The MAC Address of the Port allocated to the VM
              status:       The Status of the Port, matching the definition
              segmentation_id: The VLAN ID for the Network the Port is part of
              physical_network: The Physical Network the Port is part of

        :param context:   The security context for the query
        :param instances: A list of dictionary objects for each VM containing:
            uuid:      The UUID of the VM when created thru OS
            name:      The Name of the VM defined on the Hypervisor
            hostuuid:  The Hypervisor-specific UUID of the VM (Optional)
        """
        disc_insts = []
        importvm = vm_import.ImportVM(self._conn)
        info = importvm.query_onboarding_instances(instances)
        disc_insts += info
        return disc_insts

    def monkey_patch_fetch_image(self):
        original_fetch_method = libvirt_utils.fetch_image
        LOG.debug("Monkey patching libvirt utils fetch_image")

        def patched_fetch(context, target, image_id, user_id, project_id,
                          max_size=0):
            if 'ONBOARDED_NO_IMAGE' in image_id:
                LOG.warning("Attempted to fetch image %s, image not retrieved."
                            % image_id)
                return
            else:
                return original_fetch_method(context, target, image_id,
                                             user_id, project_id, max_size)

        libvirt_utils.fetch_image = patched_fetch

    def monitor_instances(self, context, instances):
        """
        Returns details about the VM's specified that can be considered
        an accurate inventory of the VM that is periodically collected.
        For each VM the driver needs to return a dictionary containing
        the following attributes:
            uuid:      The UUID of the Instance that was passed in
            name:      The Name of the Instance that was passed in
            cpu_utilization: The percent of CPU that is being used

        :param context: The security context for the query
        :param instances: A list of dictionary objects for each VM containing:
            uuid:      The UUID of the VM when created thru OpenStack
            name:      The Name of the VM when created thru OpenStack
            vm_uuid:   The Hypervisor-specific UUID of the VM (Optional)
        """
        DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S"
        data = []
        for instance in instances:
            name = instance["name"]
            dom = self._host._get_domain_by_name(name)
            dom_info = dom.info()
            ts_val = datetime.utcnow()
            cpu_stats = dict(number=dom_info[3],
                             time=dom_info[4],
                             ts=ts_val.strftime(DATETIME_FORMAT))
            cpu_util = self._cpu_util_processor.evaluate_cpu_util(
                instance, cpu_stats)
            if cpu_util:
                data.append(dict(uuid=instance["uuid"],
                                 name=name,
                                 cpu_util=cpu_util.volume))
        return data

    def _neutron_failed_callback(self, event_name, instance):
        """
        Over-riding this method as we want the VIF callback to fail, but not
        be an exception that aborts the build.

        By using a different exception type, we should fall into the reschedule
        path rather than the abort path.  This is code to remove once we get
        a fix in the community.
        """
        LOG.error(_('PowerKVM Neutron agent callback failed for event '
                    '%(event)s for instance %(uuid)s'),
                  {'event': event_name, 'uuid': instance.uuid})
        if CONF.vif_plugging_is_fatal:
            raise pkvmexception.PowerKVMVifCallbackFailed()

    def _create_domain_and_network(self, context, xml, instance, network_info,
                                   disk_info, block_device_info=None,
                                   power_on=True, reboot=False,
                                   vifs_already_plugged=False):
        """
        Override method to allow for a different exception to be raised,
        thereby allowing for a reschedule to occur.
        """
        try:
            super(PowerKVMDriver, self)._create_domain_and_network(
                context, xml, instance, network_info,
                block_device_info=block_device_info, power_on=power_on,
                reboot=reboot, vifs_already_plugged=vifs_already_plugged,
                disk_info=disk_info)
        except exception.VirtualInterfaceCreateException:
            LOG.error(_('Domain and network creation failed for '
                        'instance %(uuid)s'),
                      {'uuid': instance.uuid}, instance=instance)
            raise pkvmexception.PowerKVMVifCallbackFailed()

    def _get_cpu_info(self):
        """
        Override _get_cpu_info to include additional information
        needed by PowerVC.
        """
        cpu_info = super(PowerKVMDriver, self)._get_cpu_info()

        # PowerVC specific modification
        socket, cores = proc_utils.get_sockets_and_cores_on_host()

        cpu_info['topology']['sockets'] = socket
        cpu_info['topology']['cores'] = cores
        cpu_info['topology']['threads'] = proc_utils.THREADS_PER_CORE
        # END - PowerVC specific modification
        return cpu_info

    def get_volume_mapped_entries(self, volume_udid, vios_name):
        """ Not needed for PowerKVM so pass through"""
        pass

    def call_terminate_connection(self, context, instance, volume_id,
                                  connector, conninfo=None, clean=True):
        """ cleanup the hdisk left after vscsi mapping"
        """
        volume_api = volume_cinder.PowerVCVolumeAPI()
        volume_api.terminate_connection(context, volume_id, connector)


class PowerKVMQcow2(Qcow2):
    """
    This class is created to workaround the import vm using ONBOARDED_NO_IMAGE
    for image reference.
    OpenStack code will try to verify size using that info and result
    operation errors.
    See stg defect 42007 for detail
    """
    def create_image(self, prepare_template, base, size, *args, **kwargs):
        """
        This method should be exact match to imagebackend.py with exception
        of the self.verify_base_size statement within the else clause
        """
        filename = self._get_lock_name(base)

        @utils.synchronized(filename, external=True, lock_path=self.lock_path)
        def copy_qcow2_image(base, target, size):
            # TODO(pbrady): Consider copying the cow image here
            # with preallocation=metadata set for performance reasons.
            # This would be keyed on a 'preallocate_images' setting.
            libvirt_utils.create_cow_image(base, target)
            if size:
                disk.extend(target, size)

        # Download the unmodified base image unless we already have a copy.
        if not os.path.exists(base):
            prepare_template(target=base, max_size=size, *args, **kwargs)
        else:
            self.verify_base_size(base, size)

        legacy_backing_size = None
        legacy_base = base

        # Determine whether an existing qcow2 disk uses a legacy backing by
        # actually looking at the image itself and parsing the output of the
        # backing file it expects to be using.
        if os.path.exists(self.path):
            backing_path = libvirt_utils.get_disk_backing_file(self.path)
            if backing_path is not None:
                backing_file = os.path.basename(backing_path)
                backing_parts = backing_file.rpartition('_')
                if backing_file != backing_parts[-1] and \
                        backing_parts[-1].isdigit():
                    legacy_backing_size = int(backing_parts[-1])
                    legacy_base += '_%d' % legacy_backing_size
                    legacy_backing_size *= units.Gi

        # Create the legacy backing file if necessary.
        if legacy_backing_size:
            if not os.path.exists(legacy_base):
                with fileutils.remove_path_on_error(legacy_base):
                    libvirt_utils.copy_image(base, legacy_base)
                    disk.extend(legacy_base, legacy_backing_size)

        if not os.path.exists(self.path):
            with fileutils.remove_path_on_error(self.path):
                copy_qcow2_image(base, self.path, size)


# ###########################################
# ###  Temporary Libvirt Designer Patch  ####
# ###########################################
def _set_vif_host_backend_ovs_config(conf, brname, interfaceid, tapname):
    """Override the Config Script to None until Community Fixed"""
    _designer = designer
    _designer._orig_set_vif_host_backend_ovs_config(
        conf, brname, interfaceid, tapname)
    conf.script = None

designer._orig_set_vif_host_backend_ovs_config = \
    designer.set_vif_host_backend_ovs_config
designer.set_vif_host_backend_ovs_config = _set_vif_host_backend_ovs_config
