# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Resources related utilities"""

import json
import os
import time

from nova import utils
from oslo_log import log as logging
from oslo_config import cfg

from powervc_nova import _
from powervc_nova.virt.ibmpowervm.common import exception as common_ex
from powervc_nova.virt.powerkvm import constants

CONF = cfg.CONF
LOG = logging.getLogger(__name__)

TWENTY_FOUR_HOURS = 3600 * 24


def get_all_disk_committed(disk_over_committed):
    """Writes the disk_over_committed value for this host to the
    compute nodes usage file saved in the storage path, and sums up
    disk_over_committed values for every node using this storage.

    :param disk_over_committed: total disk space committed to all disks
       allocated to all VMs on this host
    :return: total disk space committed to all hosts sharing this storage
    """
    storage_path = CONF.instances_path
    hostname = CONF.host
    LOCK_PATH = os.path.join(CONF.instances_path, 'locks')

    @utils.synchronized('storage-registry-lock', external=True,
                        lock_path=LOCK_PATH)
    def do_get_all_disk_committed():
        def _add_up_all_usage():
            """Add up each disk committed values for each node sharing
            this storage path, and clean up any expired nodes
            """
            total_usage = 0
            for node, info in usage_dict.items():
                # Check for expired nodes and remove
                if (time.time() - info['time']) > TWENTY_FOUR_HOURS:
                    usage_dict.pop(node)
                else:
                    total_usage += int(info.get('disk_gb_committed', 0))
            return total_usage

        usage_dict = {}
        id_path = os.path.join(storage_path, 'compute_nodes_usage')
        if os.path.exists(id_path):
            with open(id_path) as f:
                try:
                    usage_dict = json.loads(f.read())
                except ValueError:
                    LOG.warning(_("Cannot decode JSON from %(id_path)s"),
                                {"id_path": id_path})
                    return disk_over_committed

        # Update the file with current time and disk usage for the host
        usage_dict[hostname] = {'time': time.time(),
                                'disk_gb_committed': disk_over_committed}

        total_usage = _add_up_all_usage()

        with open(id_path, 'w') as f:
            f.write(json.dumps(usage_dict))

        return total_usage

    return do_get_all_disk_committed()


def remove_from_usage_file():
    """This method is called when unmanaging a host. It removes
    this host from the compute nodes usage file.
    """

    storage_path = CONF.instances_path
    hostname = CONF.host
    LOCK_PATH = os.path.join(CONF.instances_path, 'locks')

    @utils.synchronized('storage-registry-lock', external=True,
                        lock_path=LOCK_PATH)
    def do_remove_from_usage_file():
        usage_dict = {}
        id_path = os.path.join(storage_path, 'compute_nodes_usage')
        if not os.path.exists(id_path):
            return
        with open(id_path) as f:
            try:
                usage_dict = json.loads(f.read())
            except ValueError:
                LOG.warning(_("Cannot decode JSON from %(id_path)s"),
                            {"id_path": id_path})
                return

        if usage_dict[hostname]:
            usage_dict.pop(hostname)

        with open(id_path, 'w') as f:
            f.write(json.dumps(usage_dict))

    return do_remove_from_usage_file()


def check_resource(newFlavor, instance):
    """
    This method checks for resources of the instance
    for following

    1.requested memory should be greater than IBM_POWERKVM_MIN_MEM
    2.request disk size should be greater than existing disk
      size(in case of resize)

    :param newFlavor: new values to resize
    :param instance: instance to resize
    """

    new_disk_size = -1
    disk_size = instance['root_gb']

    # new flavor is populated in case of resize only
    if newFlavor is not None:
        memory_mb = newFlavor.get('memory_mb', None)
        new_disk_size = newFlavor.get('root_gb', -1)
    else:
        memory_mb = instance['memory_mb']

    mem_min = constants.IBM_POWERKVM_MIN_MEM

    # Minimum memory check
    if memory_mb < mem_min:
        ex_args = {'instance_name': instance['display_name'],
                   'mem_requested': memory_mb,
                   'mem_min': mem_min}
        LOG.error(_("Memory requested (%(mem_requested)d MB) is less than "
                    "minimum required value (%(mem_min)d MB) for instance "
                    "'%(instance_name)s'") % ex_args)
        raise common_ex.IBMPowerVMMemoryBelowMin(**ex_args)

    # Disk less than actual check
    if new_disk_size > -1 and new_disk_size < disk_size:
        ex_args = {'instance_name': instance['display_name'],
                   'new_disk_size': new_disk_size,
                   'disk_size': disk_size}
        LOG.error(_("Disk requested (%(new_disk_size)d GB) is less than "
                    "existing disk (%(disk_size)d GB) for instance "
                    "'%(instance_name)s'") % ex_args)
        raise common_ex.IBMPowerVMDiskBelowActual(**ex_args)
