# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
"""
Handles all requests relating to volumes + cinder.
"""

import inspect
from cinderclient import exceptions as cinder_exception
import copy
from datetime import datetime, timedelta
from keystoneclient import exceptions as keystone_exception
from nova import exception
from oslo_log import log as logging
from powervc_nova import _
from powervc_nova.compute import populate_admin_context
from powervc_nova.compute import exception as pvc_exc
from nova.volume import cinder
from nova import availability_zones as az
from nova.objects import instance as instance_obj
from nova.compute import task_states
from oslo_config import cfg
import sys
import six
from nova.compute import api
from powervc_nova.storage import exception as stgex

CONF = cfg.CONF
PROVIDER_KEY = 'host'
CINDER_HOST_KEY = 'os-vol-host-attr:host'
LOG = logging.getLogger(__name__)
NOVA_COMPUTE_API_FILE = inspect.getsourcefile(api.API)


class PowerVCCinderConflict(cinder_exception.ClientException):

    """
    HTTP 409 - This indicates that the request could not be completed due
    to a conflict with the current state of the resource.
    """
    http_status = 409
    message = "Conflict"


class PowerVCCinderInternalServerError(cinder_exception.ClientException):

    """
    HTTP 500 -  This indicates that the cinder service is currently in
    error state.
    """
    http_status = 500
    message = "Internal Server Error"


class PowerVCCinderInsufficientStorage(cinder_exception.ClientException):

    """
    HTTP 507 - This indicates that the server does not have enough space to
    save the resource.
    """
    http_status = 507
    message = "Insufficient Storage"


class PowerVCExpendvdiskCapacityException(exception.NovaException):

    """
    The resize request exceeds the storage capacity
    """
    msg_fmt = _("%(error)s")


class PowerVCSVCVdiskNotFoundException(exception.NovaException):

    """
    The back end SVC volume has been deleted
    """
    msg_fmt = _("%(error)s")


class PowerVCExpendvdiskFCMapException(exception.NovaException):

    """
    The SVC volume can not be extended due to existing flashcopy mapping.
    """
    msg_fmt = _("%(errors)s")


class PowerVCExtendVolumeNotImplementedError(exception.NovaException):

    """
    The extend volume REST API is not implemented for certain volume
    driver.
    """
    msg_fmt = _("%(error)s")


class PowerVCVolumeServiceInternalError(exception.NovaException):

    """
    The extend-volume service had internal error when sending request
    to volume service.
    """
    msg_fmt = _("%(error)s")


def _untranslate_volume_summary_view(context, vol):
    """Maps keys for volumes summary view."""
    # Since if v2 they switched the 'display_name' to 'name', the we
    # need to work around the OpenStack bug that is assuming display_name
    if not hasattr(vol, 'display_name') and hasattr(vol, 'name'):
        vol.display_name = vol.name
    if not hasattr(vol, 'display_description') and hasattr(vol, 'description'):
        vol.display_description = vol.description
    # Call the base OpenStack method to do the actual translation
    volume = cinder._untranslate_volume_summary_view(context, vol)
    # OSEE is broken on volume_type_id field. Fix it here.
    if "volume_type_id" in volume.keys():
        volume['volume_type'] = volume['volume_type_id']
        volume['volume_type_id'] = None
    if hasattr(vol, CINDER_HOST_KEY):
        volume[PROVIDER_KEY] = getattr(vol, CINDER_HOST_KEY)
    else:
        volume[PROVIDER_KEY] = None
    if hasattr(vol, 'source_volid'):
        volume['source_volid'] = vol.source_volid
    else:
        volume['source_volid'] = None
    # report volume health status too
    if hasattr(vol, 'health_status'):
        volume['health_status'] = copy.deepcopy(vol.health_status)
    else:
        volume['health_status'] = None

    if hasattr(vol, 'multiattach'):
        volume['multiattach'] = vol.multiattach
    else:
        volume['multiattach'] = False
    return volume


def _untranslate_volume_type_view(volume_type):
    """ Maps keys for volume type summary view """
    t = {}
    t['id'] = volume_type.id
    t['name'] = volume_type.name
    t['extra_specs'] = volume_type.extra_specs
    return t


def translate_volume_exception(method):
    """Transforms the exception for the volume but keeps its traceback intact.
    """

    def wrapper(self, ctx, volume_id, *args, **kwargs):
        try:
            res = method(self, ctx, volume_id, *args, **kwargs)
        except (Exception, cinder_exception.ClientException) as excp:
            __, exc_value, exc_trace = sys.exc_info()
            vol_error = exc_value.message
            exc_reasons = vol_error.split(':')
            vol_exc_value = (exc_reasons[1] if len(exc_reasons) > 1
                             else vol_error)
            if isinstance(exc_value, (keystone_exception.NotFound,
                                      cinder_exception.NotFound)):
                if 'SVCVdiskNotFoundException' in vol_error:
                    exc_value = PowerVCSVCVdiskNotFoundException(
                        error=vol_exc_value)
                else:
                    exc_value = exception.VolumeNotFound(volume_id=volume_id)
            elif isinstance(exc_value, cinder_exception.BadRequest):
                # If there are no attachments for the volume, then we can
                # map to an exception that will be ignored by the community
                # compute manager. Currently no msg id for this from cinder -
                # needed to support ignoring beyond english NLS.
                if 'no active attachments' in exc_value.message:
                    msg = ("VM attachments for volume %(vol)s") %\
                        {'vol': volume_id}
                    exc_value = exception.DiskNotFound(location=msg)
                    LOG.warn(_("Not found: %s") % msg)
                else:
                    exc_value = exception.InvalidInput(
                        reason=exc_value.message)
            elif isinstance(exc_value, cinder_exception.HTTPNotImplemented):
                exc_value = pvc_exc.IBMPowerVCNotAllowed()
            elif isinstance(exc_value, PowerVCCinderInsufficientStorage):
                if "PVCExpendvdiskCapacityException" in vol_error:
                    exc_value = PowerVCExpendvdiskCapacityException(
                        error=vol_exc_value)
            elif isinstance(exc_value, PowerVCCinderConflict):
                if "NotImplementedError" in vol_error:
                    exc_value = PowerVCExtendVolumeNotImplementedError(
                        error=vol_exc_value)
                elif "PVCExpendvdiskFCMapException" in vol_error:
                    exc_value = PowerVCExpendvdiskFCMapException(
                        error=vol_exc_value)
                else:
                    # any other HTTPConflit error we don't know about
                    raise
            elif isinstance(exc_value, PowerVCCinderInternalServerError):
                exc_value = PowerVCVolumeServiceInternalError(
                    error=vol_exc_value)
            else:
                # unknown error - re raise exception
                raise excp
            raise exc_value, None, exc_trace
        return res
    return wrapper


class PowerVCVolumeAPI(cinder.API):

    """ Extend base Nova volume API with PowerVC capabilities"""

    provider_info_cached = None
    cache_time = None
    """ provider_info_cached gets set to the output of get_stg_provider_info().
        It is not to be used for anything that requires up-to-date information.
        Nor should it be depended on for existing. There is no locking around
        this static variable.
        cache_time is the time that the cache info was populated. It times
        out after 1 min.
    """

    def __init__(self):
        """ Constructor for the PowerVC volume API extension"""
        super(PowerVCVolumeAPI, self).__init__()

    @translate_volume_exception
    def get(self, context, volume_id):
        # fill-out missing context
        context = self._fix_context(context)
        item = cinder.cinderclient(context).volumes.get(volume_id)
        volume = _untranslate_volume_summary_view(context, item)
        return volume

    def get_all(self, context, search_opts=None):
        # override to provide PowerVC specific _untranslate_volume_summary_view
        search_opts = search_opts or {}
        items =\
            cinder.cinderclient(context).volumes.list(detailed=True,
                                                      search_opts=search_opts)
        rval = []

        for item in items:
            rval.append(_untranslate_volume_summary_view(context, item))

        return rval

    def detach(self, context, volume_id, instance_uuid=None,
               attachment_id=None):
        """
        Override community code to pass in attachment id
        irrespective of multiattach flag. This is because PowerVC
        allows volume multiattach property to be changed even
        after creation of volume.
        """
        if attachment_id is None and instance_uuid:
            volume = self.get(context, volume_id)
            attachments = volume.get('attachments', {})
            attachment_id = attachments.get(instance_uuid, {}).\
                get('attachment_id')

        super(PowerVCVolumeAPI, self).detach(context, volume_id, instance_uuid,
                                             attachment_id)

    def check_attach(self, context, volume, instance=None):
        if (volume['status'] == 'creating' and
                self._is_called_from('_validate_bdm',
                                     calling_filename=NOVA_COMPUTE_API_FILE)):
            # Volumes in creating status are OK when check_attach is called
            # from nova.compute.api.API._validate_bdm.  This is only called
            # during the API validation of a create server/deploy request.
            # We allow volumes in the 'creating' state so that the GUI and
            # end users can create volumes and pass them directly into the
            # deploy call without waiting for volume creation to finish.
            # Volume creation success or failure is validated later in the
            # deploy flow.
            pass
        elif 'multiattach' in volume and\
                volume.get('multiattach'):
            # override community check attach to allow shared volume in nova
            instance_list = []
            if instance:
                instance_list.append(instance)
            if 'attachments' in volume:
                for instance_uuid in volume['attachments']:
                    instance_list.append(instance_obj.Instance.
                                         get_by_uuid(context, instance_uuid))
            for instance_key in instance_list:
                if ((instance_key.get('system_metadata') and
                        (instance_key['system_metadata'].
                            get('image_os_distro')))):
                    distro = (instance_key['system_metadata'].
                              get('image_os_distro'))
                    if distro == 'ibmi' and\
                            volume['status'] in ('in-use', 'attaching'):
                        msg = _("Shared volumes are not supported for IBM i "
                                "workloads. The volume with ID %(volume_id)s "
                                "has a status of '%(status)s'.") %\
                            dict(volume_id=volume['id'],
                                 status=volume['status'])
                        LOG.warn(msg)
                        raise exception.InvalidVolume(reason=msg)
            if(volume['status'] not in ('available', 'in-use', 'attaching')):
                msg = _("volume '%(vol)s status must be available or"
                        " 'in-use' or 'attaching'") % {'vol': volume['id']}
                LOG.error(msg)
                raise exception.InvalidVolume(reason=msg)
        else:
            if volume['status'] != "available":
                msg = _("volume '%(vol)s' status must be 'available'."
                        "And multi-attach attribute is not set."
                        "Currently in '%(status)s'") %\
                    dict(vol=volume['id'], status=volume['status'])
                raise exception.InvalidVolume(reason=msg)
            if volume['attach_status'] == "attached":
                msg = _("volume %s already attached") % volume['id']
                raise exception.InvalidVolume(reason=msg)
        if instance and not CONF.cinder.cross_az_attach:
            if instance.get('host'):
                instance_az = az.get_instance_availability_zone(
                    context, instance)
            else:
                instance_az = instance['availability_zone']
            if instance_az != volume['availability_zone']:
                msg = (_("Instance %(instance)s and volume %(vol)s are not in "
                         "the same availability_zone. Instance is in "
                         "%(ins_zone)s. Volume is in %(vol_zone)s") %
                       {"instance": instance['id'],
                        "vol": volume['id'],
                        'ins_zone': instance_az,
                        'vol_zone': volume['availability_zone']})
                raise exception.InvalidVolume(reason=msg)

    def _is_called_from(self, function, calling_filename=None,
                        calling_short_fn=None, levels=5):
        called_from_function = False
        stack = inspect.stack()
        try:
            for x in range(min(levels, len(stack))):
                frame_record = stack[x]
                if calling_filename:
                    if (function == frame_record[3] and
                            calling_filename == frame_record[1]):
                        called_from_function = True
                elif calling_short_fn:
                    if (function == frame_record[3] and
                            calling_short_fn in frame_record[1]):
                        called_from_function = True
                else:
                    reason = _("No filename matching criteria")
                    raise exception.InvalidInput(reason=reason)
        finally:
            del stack
        return called_from_function

    def create(self, context, size, name, description, snapshot=None,
               source_vol=None, image_id=None, volume_type=None, metadata=None,
               availability_zone=None, multiattach=False):
        # override community create till the cinder client changes for
        # multiattach is in
        if source_vol is not None:
            source_volid = source_vol['id']
        else:
            source_volid = None

        if snapshot is not None:
            snapshot_id = snapshot['id']
        else:
            snapshot_id = None

        body = {'volume': {'size': size,
                           'snapshot_id': snapshot_id,
                           'name': name,
                           'description': description,
                           'volume_type': volume_type,
                           'user_id': context.user_id,
                           'project_id': context.project_id,
                           'availability_zone': availability_zone,
                           'status': "creating",
                           'attach_status': "detached",
                           'metadata': metadata,
                           'imageRef': image_id,
                           'source_volid': source_volid,
                           'multiattach': False
                           }}

        # use the _create method to include multiattach parameter
        # item = cinder.cinderclient(context).volumes.create(size, **kwargs)
        item = cinder.cinderclient(context).volumes._create('/volumes', body,
                                                            'volume')
        return _untranslate_volume_summary_view(context, item)

    @translate_volume_exception
    def delete_volume(self, context, volume_id):
        """
        """
        cinder.cinderclient(context).volumes.delete(volume_id)

    @translate_volume_exception
    def extend_volume(self, context, volume_id, new_size):
        """
         increase volume size to the new_size(gigabytes).
         :param volume_id: The volume to extend
         :param new_size: The new volume size to extend to.
         :return: None
        """
        if not volume_id:
            return None

        client = cinder.cinderclient(context)
        client.volumes._action('ibm-extend',
                               volume_id,
                               {'new_size': new_size})

    def get_volume_type(self, context, voltype_id):
        """ Get volume type details based on volume type id
        :param volume_type_id: volume type id
        :param context: operation context
        :return: volume type name or None
        """
        if not voltype_id:
            return None

        volume_type =\
            cinder.cinderclient(context).volume_types.get(voltype_id)
        return _untranslate_volume_type_view(volume_type)

    def _fix_context(self, context):
        """
        In case the context is None or is the bare admin context, we use
        a static conductor facility here to generate a new context with
        the service catalog filled in.
        """
        try:
            if not context or not context.service_catalog:
                new_context = populate_admin_context()
                if new_context:
                    LOG.debug("Return enhanced context: %s" %
                              new_context.to_dict())
                    return new_context
        except Exception as ex:
            LOG.exception(ex)
            LOG.info(_("Ignore context gen exception and continue with "
                       "existing context."))
        LOG.debug('Context unchanged.')
        return context

    def _call_storage_provider_api(self, context, cinder_c, use_cache):
        info = None
        # check cache
        if use_cache and PowerVCVolumeAPI.cache_time:
            d = datetime.now() - PowerVCVolumeAPI.cache_time
            # It seems reasonable to make the cache time here 5 minutes
            # to correspond with the topology collection interval for
            # the periodic task. We don't have reason to go to cinder
            # more often than going to the host.
            if d < timedelta(0, 300, 0):
                LOG.debug("Using cached provider info from: %s"
                          % str(PowerVCVolumeAPI.cache_time))
                info = PowerVCVolumeAPI.provider_info_cached
        if not info:
            if cinder_c is None:
                if not getattr(context, 'service_catalog', None):
                    LOG.info(_("No service_catalog available in this request "
                               "context. Skipping cinder client call."))
                    return None
                cinder_c = cinder.cinderclient(context)
            try:
                urlpath = "/storage-providers/detail"
                info = cinder_c.client.get(urlpath)
            except Exception as e:
                LOG.exception(e)
                response = \
                    six.text_type(e.__class__.__name__) + \
                    six.text_type(" - ") + \
                    six.text_type(e)
                msg = stgex.IBMPowerVCProviderError.msg_fmt % locals()
                ex = stgex.IBMPowerVCProviderError(msg)
                raise ex
            # info will be returned as a tuple and we expect the second element
            # to contain the provider dictionary.
            if (len(info) != 2) or ("storage_providers" not in info[1]):
                LOG.warn(str(info))
                response = str(info[0])
                exc_args = {'response': response}
                ex = stgex.IBMPowerVCProviderError(**exc_args)
                LOG.exception(ex)
                raise ex
            LOG.debug("GET cinder 'storage-providers/detail' returns: %s"
                      % str(info))
            # update cache
            PowerVCVolumeAPI.provider_info_cached = info
            PowerVCVolumeAPI.cache_time = datetime.now()

        return info

    def _get_properties_from_vol(self, cinder_c, volume_id, prop_list):
        storage_hostname = None
        vol = cinder_c.volumes.get(volume_id)
        LOG.debug("Volume looked up for id '%s': %s" % (volume_id, vol))
        ret_val = {}
        for prop in prop_list:
            try:
                # The volume case now folds into the cinder host case
                value = vol.__getattr__(prop)
                ret_val[prop] = value
            except Exception as ex:
                # Treat as an exception case now to avoid empty drop down
                # in GUI, with no explanation.
                LOG.exception(ex)
                error = \
                    _("Volume with id '%(vol_id)s' does not contain "
                      "the property '%(prop)s'. Ensure that the "
                      "volume has been onboarded as a PowerVC "
                      "cinder volume.") \
                    % dict(vol_id=volume_id, prop=prop)
                msg = stgex.IBMPowerVCStorageError.msg_fmt % locals()
                raise stgex.IBMPowerVCStorageError(msg)
        return ret_val

    def _parse_stg_prov_resp(self, info, exclude_in_error=False):
        provider_info = {}
        for provider in info[1]["storage_providers"]:
            if "service" in provider:
                for key in provider["service"]:
                    # We don't want to replace id, since it is the service id
                    if key != 'id':
                        provider[key] = provider["service"][key]
                del provider["service"]
            if "backend_type" in provider and\
                    provider["backend_type"] == "ssp":
                provider["stg_type"] = "cluster"
            else:
                provider["stg_type"] = "fc"
            if exclude_in_error and provider['backend_state'] != 'error':
                provider_info[provider["storage_hostname"]] = provider
            elif not exclude_in_error:
                provider_info[provider["storage_hostname"]] = provider
            else:
                LOG.debug("Provider excluded...")
            LOG.debug("Provider-INFO: %s." % provider)
        return provider_info

    def get_stg_provider_lst(self, context, storage_hostname_lst=None,
                             cinder_c=None, volume_id_lst=None,
                             use_cache=False,
                             exclude_in_error=False):
        """
        Get the storage provider information, including the stg_type for a
        list of providers, if specified, or for all registered storage
        providers. In this context, the type is one of two possibilities:
            cluster or fc.
        :param context:  A context object that is used to authorize
                         any DB or API access.
        :param storage_hostname_lst: The storage providers to get the storage
                         info for. If None and volume_id_lst is None, then
                         return info for all providers.
        :param cinder_c: If provided, this is the cinder client service
                         to issue GET calls against. If this is not
                         specified, then a new client will be created.
        :param volume_id_lst: If provided, then return the list of storage
                         providers that hosts the volumes. Each storage
                         provider will be marked as boot or data provider or
                         both depending on whether the volumes from this
                         provider are marked as is_boot_volume
        :param use_cache: If a provider cache is available and is not timed
                         out, then pull info from the cache if this param
                         is True.
        :param exclude_in_error: If True, then provider(s) with a
                         backend_state of 'error' are excluded. This option
                         is ignored if storage_hostname is specified or if
                         volume_id is specified.
        :returns: A dict of provider information. This is the same info
                  as from the "storage-providers" REST API, but in a flat
                  dictionary. If volume_id_lst is specified, then the provider
                  is marked as boot or data provider or both depending on
                  whether the volumes hosted by this provider are
                  _is_boot_volume or not. If storage_hostname_lst is not
                  specified, then the provider dictionary is the value of
                  the parent dictionary keyed by provider name. Some of the
                  interesting properties are shown in the following example
                  return format:
            {<provider-name>:
                {
                    "stg_type": <"fc"|"cluster">,
                    "host_display_name": <display-name>,
                    "backend_id": <k2-id>,
                    "backend_state": <running/stopped/...>,
                    "boot": <True/False>,
                    "data": <True/False>,
                    ...
                },
             ...
            }
        Note that the stg_type is a mapping from the "backend_type" property
        (also available), where "ssp" maps to "cluster" and everything else
        maps to "fc". This may be expanded in the future, but today,
        storage connectivity groups only distinguish between these two types.
        """
        context = self._fix_context(context)
        info = self._call_storage_provider_api(context, cinder_c, use_cache)
        if info is None:
            return None
        ################################
        # Handle specific volume case
        ################################
        # Mark a provider as boot, data or both to verify connectivity
        boot_providers = []
        data_providers = []
        if volume_id_lst:
            if not cinder_c:
                cinder_c = cinder.cinderclient(context)
            for volume_id in volume_id_lst:
                prop_list = [CINDER_HOST_KEY, 'metadata']
                props = self._get_properties_from_vol(cinder_c,
                                                      volume_id,
                                                      prop_list)
                if not storage_hostname_lst:
                    storage_hostname_lst = []
                storage_hostname = props[CINDER_HOST_KEY]
                storage_hostname_lst.append(storage_hostname)
                vol_meta = props['metadata']
                is_boot = bool(vol_meta['is_boot_volume']) if vol_meta\
                    and 'is_boot_volume' in vol_meta else False
                if is_boot:
                    boot_providers.append(storage_hostname)
                else:
                    data_providers.append(storage_hostname)

        provider_info = self._parse_stg_prov_resp(info, exclude_in_error)
        provider_lst = []
        if storage_hostname_lst is not None:
            for provider in provider_info:
                if provider in storage_hostname_lst:
                    # set a boot and data flag to know if the provider need
                    # to be checked for which type of connectivity
                    if provider in boot_providers:
                        provider_info[provider]["boot"] = True
                    if provider in data_providers:
                        provider_info[provider]["data"] = True
                    provider_lst.append(provider_info[provider])

        if provider_lst:
            return provider_lst

        if storage_hostname_lst is not None:
            # Since the specified provider was not found, raise an exception
            if volume_id_lst:
                error = _("The volume with ids '%(vol_id)s' references a "
                          "storage provider (%(prov)s) in property "
                          "'%(prop)s' that is not currently registered.") \
                    % dict(vol_id=volume_id_lst, prov=storage_hostname_lst,
                           prop=CINDER_HOST_KEY)
            else:
                error = _("The requested provider name, '%s', is not "
                          "currently registered.") % storage_hostname_lst
            msg = stgex.IBMPowerVCStorageError.msg_fmt % locals()
            raise stgex.IBMPowerVCStorageError(msg)
        return provider_info

    def get_stg_provider_info(self, context, storage_hostname=None,
                              cinder_c=None, volume_id=None, use_cache=False,
                              exclude_in_error=False):
        """
        Get the storage provider information, including the stg_type for a
        provider, if specified, or for all registered storage providers.
        In this context, the type is one of two possibilities:
            cluster or fc.
        :param context:  A context object that is used to authorize
                         any DB or API access.
        :param storage_hostname: The storage provider to get the storage info
                         for. If None and volume_id is None, then return info
                         for all providers.
        :param cinder_c: If provided, this is the cinder client service
                         to issue GET calls against. If this is not
                         specified, then a new client will be created.
        :param volume_id: If provided, then return the single storage
                         provider that hosts the volume.
        :param use_cache: If a provider cache is available and is not timed
                         out, then pull info from the cache if this param
                         is True.
        :param exclude_in_error: If True, then provider(s) with a
                         backend_state of 'error' are excluded. This option
                         is ignored if storage_hostname is specified or if
                         volume_id is specified.
        :returns: A dict of provider information. This is the same info
                  as from the "storage-providers" REST API, but in a flat
                  dictionary. If storage_hostname is not specified, then the
                  provider dictionary is the value of the parent dictionary
                  keyed by provider name. Some of the interesting properties
                  are shown in the following example return format:
            {<provider-name>:
                {
                    "stg_type": <"fc"|"cluster">,
                    "host_display_name": <display-name>,
                    "backend_id": <k2-id>,
                    "backend_state": <running/stopped/...>,
                    ...
                },
             ...
            }
        Note that the stg_type is a mapping from the "backend_type" property
        (also available), where "ssp" maps to "cluster" and everything else
        maps to "fc". This may be expanded in the future, but today,
        storage connectivity groups only distinguish between these two types.
        """
        context = self._fix_context(context)
        info = self._call_storage_provider_api(context, cinder_c, use_cache)
        if info is None:
            return None
        ################################
        # Handle specific volume case
        ################################
        if volume_id is not None:
            if not cinder_c:
                cinder_c = cinder.cinderclient(context)
            prop_list = [CINDER_HOST_KEY]
            prop_val = self._get_properties_from_vol(cinder_c,
                                                     volume_id,
                                                     prop_list)
            storage_hostname = prop_val[CINDER_HOST_KEY]

        provider_info = self._parse_stg_prov_resp(info, exclude_in_error)
        for provider in provider_info:
            if storage_hostname is not None and\
                    storage_hostname == provider:
                return provider_info[provider]

        if storage_hostname is not None:
            # Since the specified provider was not found, raise an exception
            if volume_id:
                error = _("The volume with id '%(vol_id)s' references a "
                          "storage provider (%(prov)s) in property "
                          "'%(prop)s' that is not currently registered.") \
                    % dict(vol_id=volume_id, prov=storage_hostname,
                           prop=CINDER_HOST_KEY)
            else:
                error = _("The requested provider name, '%s', is not "
                          "currently registered.") % storage_hostname
            msg = stgex.IBMPowerVCStorageError.msg_fmt % locals()
            raise stgex.IBMPowerVCStorageError(msg)

        return provider_info

    def get_storage_type_for_volume(self, context, volume_id):
        """
        Return the 'stg-type' value from the provider_info structure
        given by get_stg_provider_info(). Note that the 'stg-type' is
        a folding of the 'backend_type" property:
            'ssp'      --> 'cluster'
            all others --> 'fc'

        :param context: The authorization context for a cinder REST call.
        :param volume_id: ID for the volume to look up the owning provider
                          for.
        """
        try:
            provider = self.get_stg_provider_info(context, volume_id=volume_id,
                                                  use_cache=True)
        except stgex.IBMPowerVCStorageError:
            # This specific error can occur if the cache did not have the
            # provider, so we just call again without the cache option.
            provider = self.get_stg_provider_info(context, volume_id=volume_id)

        return provider['stg_type']

    def list_volume_types(self, context):
        """ List volumes types """
        context = self._fix_context(context)
        voltype_list =\
            cinder.cinderclient(context).volume_types.list()
        if not voltype_list:
            return []
        else:
            return [_untranslate_volume_type_view(voltype) for
                    voltype in voltype_list]

    @translate_volume_exception
    def update_volume_metadata(self, context, volume_id,
                               metadata, delete=False):
        if delete:
            # Completely replace volume metadata with one given
            return cinder.cinderclient(context).volumes.update_all_metadata(
                volume_id, metadata)
        else:
            return cinder.cinderclient(context).volumes.set_metadata(
                volume_id, metadata)


class PowerVCFabricAPI(cinder.API):

    """ Extend base cinder API with PowerVC wrapper for fabric calls"""

    def __init__(self):
        """ Constructor for the PowerVC volume API extension"""
        super(PowerVCFabricAPI, self).__init__()

    def _fix_context(self, context):
        """
        In case the context is None or is the bare admin context, we use
        a static conductor facility here to generate a new context with
        the service catalog filled in.
        """
        try:
            if not context or not context.service_catalog:
                new_context = populate_admin_context()
                if new_context:
                    LOG.debug("Return enhanced context: %s" %
                              new_context.to_dict())
                    return new_context
        except Exception as ex:
            LOG.exception(ex)
            LOG.info(_("Ignore context gen exception and continue with "
                       "existing context."))
        LOG.debug('Context unchanged.')
        return context

    def get_fabrics(self, context):
        """
        Retrieve the list of fabric details from cinder. Return a
        dictionary keyed by fabric name.
        Example return:
            {
              'A': {
                       "fabric_name": "A",
                       "fabric_display_name": "fab184",
                       "access_ip": "9.114.181.184",
                       "user_id": "admin"
                   },
              'B': {
                       "fabric_name": "B",
                       "fabric_display_name": "fab185",
                       "access_ip": "9.114.181.185",
                       "user_id": "admin"
                   }
            {
        """
        context = self._fix_context(context)
        cinder_c = cinder.cinderclient(context)
        # Adjust the retries down to 0 so as not to waste time on nova client
        # side retries. The cinder zonemanager has it's own set of
        # configurable retry logic. In addition, the host storage topology
        # reconciliation compute task will be retrying this on its periodic
        # interval.
        cinder_c.client.retries = 0
        # Also adjust the timeout so that we don't wait forever on this
        # special fabric API call. Normal response time is within a couple
        # seconds.
        cinder_c.client.timeout = 45
        try:
            urlpath = "/san-fabrics/detail"
            info = cinder_c.client.get(urlpath)
        except Exception as e:
            LOG.exception(e)
            msg = \
                six.text_type("Unable to get fabrics "
                              "from cinder service: ") + \
                six.text_type(e.__class__.__name__) + \
                six.text_type(" - ") + \
                six.text_type(e)
            LOG.warn(msg)
            return None
        # info will be returned as a tuple and we expect the second element
        # to contain the fabrics dictionary.
        if (len(info) != 2) or ("fabrics" not in info[1]):
            LOG.warn(str(info))
            msg = _("Unable to get fabrics from cinder service. Unexpected "
                    "data: %s") % str(info)
            LOG.warn(msg)
            return None
        LOG.debug("GET cinder '%s' returns: %s" % (urlpath, str(info)))

        name_to_fab = {}
        for fabric in info[1]['fabrics']:
            name_to_fab[fabric['fabric_name']] = fabric
        return name_to_fab

    def map_ports_to_fabrics(self, context, wwpn_list, give_message=False):
        """
        Given a list of FC Port WWPNs, this method calls to cinder to
        return a mapping of those wwpns to a registered fabric designation
        that each port is logged into. If a port is not found logged into
        a registered switch, its mapping will be None.
        Example:

            Input:
                [
                    "10000090FA2A5866",
                    "10000090FA2A8923",
                    "c0507606d56e03af"
                ]
            Returns:
                {
                    "10000090FA2A8923": "B",
                    "10000090FA2A5866": "A",
                    "c0507606d56e03af": None
                }
        """
        context = self._fix_context(context)
        cinder_c = cinder.cinderclient(context)
        # Adjust the retries down to 0 so as not to waste time on nova client
        # side retries. The cinder zonemanager has it's own set of
        # configurable retry logic. In addition, the host storage topology
        # reconciliation compute task will be retrying this on its periodic
        # interval.
        cinder_c.client.retries = 0
        # Also adjust the timeout so that we don't wait forever on this
        # special fabric API call. Normal response time is within a couple
        # seconds.
        cinder_c.client.timeout = 45
        try:
            urlpath = "/san-fabrics/map_ports"
            args = {'body': {'wwpn_list': wwpn_list}}
            info = cinder_c.client.post(urlpath, **args)
        except Exception as e:
            LOG.exception(e)
            msg = \
                six.text_type("Unable to get port-to-fabric "
                              "map from cinder service: ") + \
                six.text_type(e.__class__.__name__) + \
                six.text_type(" - ") + \
                six.text_type(e)
            LOG.warn(msg)
            if give_message:
                return msg
            return None
        # info will be returned as a tuple and we expect the second element
        # to contain the fabrics dictionary.
        if (len(info) != 2) or ("wwpn_mapping" not in info[1]):
            LOG.warn(str(info))
            msg = _("Unable to get port-to-fabric map from cinder service. "
                    "Unexpected data: %s") % str(info)
            LOG.warn(msg)
            if give_message:
                return msg
            return None

        LOG.debug("GET cinder '%s' returns: %s" % (urlpath, str(info)))
        return info[1]['wwpn_mapping']

    def map_zones_to_fabrics(self, context, initiators=[], targets=[]):
        """
        Given a filter list of initiators (and or) targets wwpns,
        returns zones with matching members. If a single list is provided,
        then filter against only that list. If both list provided,
        then filter against both lists. If no filter is supplied
        then a list of all active zones will be returned
        Example:
        active_zones-
            { "zone1": ["init_wwpn1", "target_wwpn1"],
              "zone2": ["init_wwpn2", "target_wwpn2"]
            }
        initiators: ["init_wwpn1", "init_wwpn2"]
        targets   : ["target_wwpn1"]
        Returns   : ["zone1"]

        initiators: ["init_wwpn1", "init_wwpn2"]
        targets   : []
        Returns   : ["zone1", "zone2"]
        """
        context = self._fix_context(context)
        cinder_c = cinder.cinderclient(context)
        try:
            urlpath = "/san-fabrics/map_zones"
            args = {'body': {'initiators': initiators,
                             'targets': targets}}
            zones = cinder_c.client.post(urlpath, **args)
            if len(zones) != 2:
                msg = _("Unable to get zones-to-fabric map from cinder "
                        "service. Unexpected return: %s") \
                    % six.text_type(zones)
                LOG.warn(msg)
        except Exception as ex:
            LOG.error(ex)
            raise ex
        return zones[1]


class PowerVCPVMVolumeAPI(PowerVCVolumeAPI):
    """ Extend base PowerVCVolumeAPI volume API with Novalink
    specific overrides"""

    def __init__(self):
        """ Constructor for the PowerVC volume API extension"""
        super(PowerVCPVMVolumeAPI, self).__init__()

    def initialize_connection(self, context, volume_id, connector):
        """
        Override the initialize_connection to be called multiple
        times if needed for vscsi connections
        example return for vscsi:
            {
                u'driver_volume_type': u'fibre_channel',
                u'serial': u'72f409ed-81fc-417d-9a0d-701bfa4ab0d8',
                u'data':
                    {u'vendor': u'IBM',
                    u'target_discovered': False,
                    u'target_UID': u'600507680282861D88000000000001E7',
                    u'connection-type': u'pv_vscsi',
                    u'qos_specs': None,
                    u'volume_id': u'72f409ed-81fc-417d-9a0d-701bfa4ab0d8',
                    u'access_mode': u'rw',
                    u'target_wwn': u'500507680220E522',
                    u'vios_data':
                    {u'09846EF0-0292-4BE9-B5A9-3229408D5B83':
                        {u'target_lun': u'9',
                        u'initiator_target_map':
                        {u'21000024ff64e8e0': [u'500507680220E522']}},
                     u'42768453-1292-4AE9-B5A9-3229408D5B83':
                        {u'target_lun': u'10',
                        u'initiator_target_map':
                        {u'21000024ff64e8d0': [u'500507680220E522']}}
                    }
                    }
                }

        """
        conn_info_dict = {}
        s = super(PowerVCVolumeAPI, self)
        conn_lst, onb_pend = self._get_vol_connector(context, volume_id,
                                                     connector)
        LOG.debug("initialize conn conn_lst %(conn_lst)s %(onb_pend)s" %
                  dict(conn_lst=conn_lst, onb_pend=onb_pend))
        if not conn_lst:
            # Putting this check in during development. Likely APAR indicator.
            msg = _("No connector information was determined for handling "
                    "initialize_connection for volume with ID "
                    "%s.") % volume_id
            # raise exception.InvalidMetadata(reason=msg)
            LOG.warn(msg)
            return {'data': {'volume_id': volume_id}}

        for conn in conn_lst:
            LOG.info(_("initialize_connection called with: "
                       "built_connector=%(conn)s, vol_id=%(vol)s, "
                       "connector_in=%(conn_in)s") %
                     {'conn': conn, 'vol': volume_id, 'conn_in': connector})
            if 'active_wwpns' in conn:
                # Pop this data off since it is only used during
                # terminate connection for now.
                conn.pop('active_wwpns')
            conn_info = s.initialize_connection(context, volume_id, conn)
            conn_info_data = conn_info['data']
            ctype = conn_info_data.get('connector_type')
            # the volume can be attached using only one of npiv or
            # vscsi at one time.
            if ctype == 'pv_vscsi':
                vios_id = conn['host']
                vios_data = self._create_vios_conn_info_dict(conn_info,
                                                             vios_id)
                if not conn_info_dict:
                    conn_info_dict = conn_info
                    conn_info_dict['data']['vios_data'] = {}
                conn_info_dict['data']['vios_data'][vios_id] = vios_data
            else:
                # Npiv case
                if len(conn_lst) > 1:
                    msg = (_("More than one connector found for NPIV "
                             "attachment of volume %(vol_id)s %(conn_lst)s")
                           % dict(vol_id=volume_id, conn_lst=conn_lst))
                    LOG.error(msg)
                    raise cinder_exception.BadRequest(six.text_type(msg))
                conn_info_dict = conn_info
                conn_info_dict['data']['connection-type'] = 'npiv'

        return conn_info_dict

    def terminate_connection(self, context, volume_id, connector):
        """
        override the terminate connection to be called multiple
        times if needed per vios for vscsi
        """
        s = super(PowerVCVolumeAPI, self)
        conn_lst, onboard_pend = self._get_vol_connector(context, volume_id,
                                                         connector,
                                                         is_initializing=False)
        if not conn_lst:
            LOG.info(_("Connection list is empty. Nothing to do."))
            return

        pvm_shutdown = self._is_called_from('_shutdown_instance',
                                            calling_short_fn='pvm_manager.py')
        # if it's calling from resize_instance, there is no need to
        # terminate connection
        resize_inst = self._is_called_from('resize_instance',
                                           calling_short_fn='manager.py')
        if resize_inst:
            LOG.info(_("Calling from resize_instance. Nothing to do."))
            return
        exceptions = []
        for conn in conn_lst:
            # If this is a NPIV based migration, then no clean up is needed
            # on the Cinder side.  It is simply re-using the full connector.
            if conn.get('is_npiv_migration', False):
                LOG.info(_("NPIV Volume %(vol_id)s found for migration.  "
                           "Will not terminate connection on source host."),
                         {'vol_id': volume_id})
                continue

            if 'active_wwpns' in conn:
                # This is a remote restart + NPIV case. Handle 2 subcases.
                if conn['active_wwpns']:
                    msg = (_("In terminate NPIV volume connection for a "
                             "remote restarting virtual machine, one or more "
                             "WWPNs were found active on managed fabrics: "
                             "%(wwpns)s. A running instance cannot be remote "
                             "restarted. The volume ID is %(vol_id)s.") %
                           {'wwpns': conn['active_wwpns'],
                            'vol_id': volume_id})
                    LOG.error(msg)
                    raise stgex.IBMPowerVCStorageError(error=msg)
                else:
                    LOG.info(_("NPIV instance ports not found active on "
                               "managed fabrics. Instance will restart with "
                               "unchanged WWPNs, so connection will not be "
                               "terminated for volume with ID %(vol_id)s.") %
                             {'vol_id': volume_id})
                    continue

            if (pvm_shutdown and 'connection_type' in conn and
                    conn['connection_type'] == 'npiv'):
                # We need to modify the connector to contain a
                # clean_zones=all key-value pair when deleting VMs and the
                # connector is NPIV. PowerVC's overridden Cinder volume
                # drivers require this to remove the VM's NPIV zone on
                # deletion.  This check will be removed when the PowerVC
                # Cinder volume drivers are modified to
                # be more in line with how the community removes zones.
                conn['clean_zones'] = 'all'

            LOG.info(_("terminate_connection called with: "
                       "built_connector=%(conn)s, vol_id=%(vol)s, "
                       "connector_in=%(conn_in)s") %
                     {'conn': conn, 'vol': volume_id, 'conn_in': connector})

            try:
                s.terminate_connection(context, volume_id, conn)
            except Exception as e:
                LOG.exception(_("Error during terminate connection %s") %
                              e)
                if onboard_pend and next(conn_lst.__iter__()):
                    #If there are more connectors, allow the
                    #"for conn in conn_lst" loop to continue before
                    #failing (if all fail)
                    LOG.info("Exception during terminate conn for onboard"
                             " pending. Ignore the above exception log.")
                    exceptions.append(e)
                else:
                    raise e
        #if all the connectors failed on cinder re raise the exception
        if len(exceptions) == len(conn_lst):
            raise exceptions[0]

    def _get_vol_connector(self, context, volume_id, connector,
                           is_initializing=True):
        """
        Generated the connector with wwpns and host from the
        volume adapter object passed in.
        """
        connector_lst = []
        is_boot = self._is_volume_boot_set(context, volume_id)

        #onboard pending case for vscsi npiv scg
        onboard_pending = False
        if 'vios_data' in connector or 'vscsi_target_lun' in connector:
            LOG.debug("Found exisiting vios data %s " % connector)
            return self._parse_vios_data_vscsi_conn(connector), onboard_pending

        if 'pv_vscsi' in connector.keys() and 'npiv' in connector.keys() and\
                not is_initializing:
            vscsi_adap = connector['pv_vscsi']['vol_adap']
            npiv_adap = connector['npiv']['vol_adap']
            onboard_pending = vscsi_adap._get_finish_onboard_status(context)
            if onboard_pending:
                LOG.debug("onboard_pending %s" % onboard_pending)
                connector_lst = self._get_vscsi_connector(vscsi_adap,
                                                          volume_id,
                                                          context,
                                                          is_initializing)
                npiv_conn_lst = self._get_npiv_connector(npiv_adap,
                                                         is_initializing)
                if npiv_conn_lst and len(npiv_conn_lst) > 0:
                    connector_lst.append(npiv_conn_lst[0])
                return connector_lst, onboard_pending

        for conn_type in connector:
            if conn_type == 'npiv' or conn_type == 'pv_vscsi':
                cinfo = connector[conn_type]
                LOG.debug("conn_type=%s, cinfo=%s" % (conn_type, cinfo))
                vol_adap = cinfo['vol_adap']
                if (is_boot == cinfo.get("is_boot_type", False)) or\
                        ((not is_boot) == cinfo.get("is_data_type", False)):
                    # build the connector using the adapter
                    if conn_type == 'pv_vscsi':
                        connector_lst = self._get_vscsi_connector(
                            vol_adap, volume_id, context, is_initializing)
                    elif conn_type == 'npiv':
                        connector_lst = self._get_npiv_connector(
                            vol_adap, is_initializing)
                        LOG.debug("Using NPIV volume adapter. Task_state=%s" %
                                  str(vol_adap.instance.task_state))
                        if (connector_lst and vol_adap.instance.task_state ==
                                task_states.REBUILDING):
                            # The instance wants to rebuild on a target host
                            # and this could be a detach request for the dead
                            # host. We will check weather the wwpns are logged
                            # into any managed fabric. If they are, then the
                            # Rebuilding state for the VM is likely not valid
                            # (instance is running somewhere). If there is any
                            # problem in getting the active wwpn fabric logins,
                            # then it is treated as not being found and will
                            # not generate an error condition for RR.
                            fabAPI = PowerVCFabricAPI()
                            active_map = fabAPI.map_ports_to_fabrics(
                                context, connector_lst[0]['wwpns'])
                            # p = port wwpn and f = fabric identifier in
                            # composition.
                            connector_lst[0]['active_wwpns'] = (
                                [p for p, f in active_map.items()
                                 if f is not None])
                            LOG.debug("Setting active_wwpns: %s" %
                                      connector_lst[0]['active_wwpns'])

        return connector_lst, onboard_pending

    def _parse_vios_data_vscsi_conn(self, connector):
        conn_lst = []
        vios_data = connector.get('vios_data')
        vscsi_target_lun = connector.get('vscsi_target_lun')
        host = None
        if vios_data:
            for host in vios_data:
                LOG.debug("vios info %s" % vios_data[host])
                vios_wwpn = vios_data[host].get('initiator_target_map').keys()
                new_connector = dict(host=host,
                                     wwpns=vios_wwpn,
                                     connection_type='pv_vscsi',
                                     is_npiv_migration=False)
                conn_lst.append(new_connector)
        elif vscsi_target_lun:
            for itl in vscsi_target_lun:
                vios_wwpn = itl.get('source_wwn')
                if not vios_wwpn:
                    vios_wwpn = itl.get('vios_wwpn')
                host = itl.get('vios_host', None)
                new_connector = dict(host=host,
                                     wwpns=vios_wwpn,
                                     connection_type='pv_vscsi',
                                     is_npiv_migration=False)
                conn_lst.append(new_connector)
        return conn_lst

    def _get_vscsi_connector(self, vol_adap, volume_id, context,
                             is_initializing=True):
        """
        Generate vscsi connector.
        """
        conn_lst = []
        wwpn_dict = vol_adap.wwpns()
        for vios_host_name in wwpn_dict:
            if not is_initializing:
                # this is teminate connection flow
                multiattach = vol_adap._get_shared_vol_key(volume_id,
                                                           vios_host_name)
                if multiattach:
                    LOG.info(_("The backend connection for volume %(vol)s will"
                               " not be terminated because the volume is "
                               "shared on VIOS %(vios)s.") %
                             dict(vol=volume_id, vios=vios_host_name))
                    continue
            vios_wwpn = wwpn_dict[vios_host_name]
            if vios_wwpn:
                new_connector = dict(host=vios_host_name,
                                     wwpns=vios_wwpn,
                                     connection_type='pv_vscsi',
                                     is_npiv_migration=False)
                conn_lst.append(new_connector)
        return conn_lst

    def _get_npiv_connector(self, vol_adap, is_initializing=True):
        """
        Generated npiv connector.
        """
        # May not need is_initializing param anymore.
        LOG.debug("is_initializing=%s" % is_initializing)
        host = vol_adap.host_name()
        wwpns = vol_adap.wwpns()

        if not wwpns:
            return []

        is_migration = False

        # Get the phys_to_virt mappings
        phys_to_virt = dict()
        scg_fab_port_map = vol_adap._get_scg_info()['fab_port_map']
        for fabric in scg_fab_port_map:

            # We won't be adding/removing fabrics during migration.  During a
            # migration, all the fabric states should be the same...so we can
            # just query the first (or last...doesn't really matter) for the
            # migration status. Check task state and host instance host as
            # well to cover the broadest possibilities. We don't want to
            # terminate on failure rollback either.
            fab_state = vol_adap._get_fabric_state(fabric)
            is_migration = (vol_adap.instance.task_state !=
                            task_states.REBUILDING) and\
                           (vol_adap.instance.task_state ==
                            task_states.MIGRATING or
                            fab_state == 'migrating' or
                            vol_adap.instance.host != CONF.host)

            # Build the phy_to_virt map
            maps = vol_adap._get_fabric_meta(fabric)
            for p_wwpn, v_wwpns in maps:
                if phys_to_virt.get(p_wwpn) is None:
                    phys_to_virt[p_wwpn] = []

                phys_to_virt[p_wwpn].extend(v_wwpns.split())
        new_connector = dict(
            host=host, wwpns=wwpns, phy_to_virt_initiators=phys_to_virt,
            wwnns=wwpns, is_npiv_migration=is_migration,
            connection_type='npiv')
        # the os_type ibmi aix ... are passed down through the connector
        # and ds8k will create a host that is appropriate for OS400, AIX, or
        # Linux.
        if ((vol_adap.instance.get('system_metadata') and
             (vol_adap.instance['system_metadata'].
              get('image_os_distro')))):
            new_connector['os_type'] = (vol_adap.instance['system_metadata'].
                                        get('image_os_distro'))
        return [new_connector]

    def _is_volume_boot_set(self, context, volume_id):
        volume = self.get(context, volume_id)
        if 'is_boot_volume' in volume['volume_metadata'] and \
                volume['volume_metadata']['is_boot_volume'] == 'True':
            return True
        return False

    def _create_vios_conn_info_dict(self, conn_info, vios_id):
        """
        Get the itmap and target_lun and create a vios_data
        dict to be added to the connection info
        """
        conn_info_data = conn_info['data']
        i_t_map = conn_info_data.pop('initiator_target_map')
        target_lun = conn_info_data.pop('target_lun')
        ctype = conn_info_data.pop('connector_type')
        conn_info_data['connection-type'] = ctype

        vios_data = dict(initiator_target_map=i_t_map,
                         target_lun=target_lun)
        return vios_data
