# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import gettext  # noqa
import locale
import os  # noqa
import six  # noqa
import textwrap as _textwrap
import functools

from argparse import ArgumentParser
from argparse import HelpFormatter
from argparse import RawDescriptionHelpFormatter
from oslo_config import cfg
from oslo_i18n import _gettextutils as i18n
from oslo_i18n._message import Message
from oslo_utils import encodeutils
from oslo_log import log as logging

_PVC_LANGS = ['en_US', 'de', 'es', 'fr', 'it', 'ja', 'ko',
              'nl', 'pt_BR', 'ru', 'zh_CN', 'zh_TW', 'zh']

cfg.CONF.register_opts([
    cfg.BoolOpt('translation_lazy_load', default=True,
                help='Specifies whether to Lazy-Load Translation'),
    cfg.ListOpt('translated_languages', default=_PVC_LANGS,
                help='Specifies the languages that have been translated'),
    cfg.BoolOpt('refstack_interoperability', default=False,
                help="Filters PowerVC injection to Base Openstack API")
])

LOG = logging.getLogger(__name__)
_COMMAND_TRANSLATION = False


def _(msg, domain='powervc_oslo'):
    """Shortcut method to return a translated message"""
    global _COMMAND_TRANSLATION
    msg = Message(msg, domain=domain)
    # If this is a command that wants str-type strings, give them that
    if _COMMAND_TRANSLATION:
        return encodeutils.safe_encode(msg.translate())
    # If we are supposed to lazy-load, then return the message object
    return msg if cfg.CONF.translation_lazy_load else msg.translate()


def enable_command_translation():
    """Utility to say this is a command so return str-type strings"""
    global _COMMAND_TRANSLATION
    _COMMAND_TRANSLATION = True


def populate_locale(context, req):
    """Utility to populate the request locale on the context"""
    context.locale = req.best_match_language()
    return context


def translate_message(context, msg):
    """Utility to populate the context locale on the message"""
    if isinstance(msg, Message) and hasattr(context, 'locale'):
        return msg.translate(context.locale)
    return msg


def translate_message_as_locale(msg, desired_locale):
    """Utility to populate the message based on user locale"""
    if isinstance(msg, Message):
        return msg.translate(desired_locale)
    return msg


def populate_languages(domain):
    """Utility method to use the PowerVC languages for the given domain"""
    langs = i18n.get_available_languages(domain)
    i18n._AVAILABLE_LANGUAGES[domain] =\
        [l for l in cfg.CONF.translated_languages if l in langs]


def override_context_serialize(ctxt_class):
    """Utility to add the locale to the Request Context serialization"""
    if not hasattr(ctxt_class, '_old_to_dict'):
        setattr(ctxt_class, '_old_to_dict', getattr(ctxt_class, 'to_dict'))
        setattr(ctxt_class, '_old_from_dict', getattr(ctxt_class, 'from_dict'))
        setattr(ctxt_class, 'to_dict', _powervc_context_to_dict)
        setattr(ctxt_class, 'from_dict', _powervc_context_from_dict)


class DoubleByteHelpFormatter(RawDescriptionHelpFormatter):
    """Decodes the byte string to unicode so proper
    formatting can be performed"""
    def _split_lines(self, text, width):
        text = self._whitespace_matcher.sub(' ', text).strip()
        _width = width

        if len(text) > _width:
            _width /= 2

        unicode_text = text.decode('utf-8')
        unicode_lines = _textwrap.wrap(unicode_text, _width)

        byte_lines = []

        for line in unicode_lines:
            # line by line, re-encode the unicode string
            # to a str-type byte string
            byte_lines.append(line.encode('utf-8'))

        return byte_lines


def get_formatter_class():
    """If the locale involves a DBCS supported by PowerVC,
    use the DoubleByteHelpFormatter"""
    dbcs_locales = ["zh", "zh_CN", "zh_TW", "ko_KR", "ja_JP"]

    current_locale = locale.getdefaultlocale()[0]

    if current_locale in dbcs_locales:
        return DoubleByteHelpFormatter
    else:
        return HelpFormatter


class DoubleByteArgumentParser(ArgumentParser):
    """The formatter class set on the parser is not propagated to
    its sub=parsers. As a result, the formatter class for sub-parsers
    always defaults to formatter class set in the class definition for
    the ArgumentParser class it is associated with. By default, it is
    the HelpFormatter class. We need to subclass the ArgumentParser
    with a new class that sets its default formatter
    class to the DoubleByteHelpFormatter"""
    def __init__(self, **kwargs):
        super(DoubleByteArgumentParser, self).__init__(**kwargs)
        self.formatter_class = DoubleByteHelpFormatter


def get_parser_class():
    """Returns the parser class which is ultimately determined
    by the current locale"""
    _formatter_class = get_formatter_class()
    if _formatter_class is DoubleByteHelpFormatter:
        return DoubleByteArgumentParser
    else:
        return ArgumentParser


class KeyedMessage(Message):

    """Extension Class for adding a message id/key prefix"""

    def __init__(self, msgid, msg):
        self.msgid = msgid
        self.msg = msg

    @property
    def data(self):
        return self.msgid + ' ' + self.msg.data

    @property
    def locale(self):
        return self.msg.locale

    @locale.setter
    def locale(self, value):
        self.msg.locale = value

    def translate(self, desired_locale=None):
        msgtxt = self.msg.translate(desired_locale=desired_locale)
        return self.msgid + ' ' + msgtxt


###################################################
#        Override Serialization on Context        #
###################################################
# The way OpenStack has it setup is that they convert the Context to and
# from a Dictionary to serialized across the wire for RPC.  However they
# only serialize certain attributes from the context, so even though we
# have added the Locale to the Context, it isn't serialized across, so
# we need to override to add the Locale so we can handle for Translation.
def _powervc_context_to_dict(self):
    """Helper method to add the Locale to the Context Dictionary"""
    adict = self._old_to_dict()
    # If the Locale was present in the Context, then add to Dictionary
    if hasattr(self, 'locale'):
        adict['locale'] = getattr(self, 'locale')
    return adict


@classmethod
def _powervc_context_from_dict(cls, values):
    """Helper method to parse the Locale from the Context Dictionary"""
    locale = values.pop('locale', None)
    aobj = cls._old_from_dict(values)
    # If the Locale was present in the Dictionary, then add to Context
    if locale is not None:
        aobj.locale = locale
    return aobj


###################################################
#        Override RabbitMQ Connection Close       #
###################################################
# There is an issue in the OpenStack RabbitMQ driver when it tries
# to close the existing channel when a new one is created on a new
# connection it will blow up, so keeps using the old channel instead.
# We will override this to handle the exception and use new connection.
def _powervc_set_current_channel(self, new_channel):
    """Override set_current_channel to handle exceptions on close"""
    if self.channel is not None and new_channel != self.channel:
        try:
            self.connection.maybe_close_channel(self.channel)
        # If an exception on closing, just ignore it and continue on
        except:
            pass
    self.channel = new_channel


###################################################
#        Common Decorator                         #
###################################################
def powervc_api_filter(api_func):
    """Decorator for API calls to check/gate PowerVC injections
       into API returns.
       IMPORTANT - Must be placed as the first decorator to function!

       #example
           @wsgi.extends
           @nova_api_filter
           def index(self, req, server)
               ....
    """
    LOG.debug(_("PowerVC injection to API returns is disabled, "
                "'nova_api_injection': %(filter)s. "
                "Module: %(func_module)s "
                "Function:%(func_name)s"
              % dict(filter=cfg.CONF.refstack_interoperability,
                     func_module=api_func.__module__,
                     func_name=api_func.__name__)))

    @functools.wraps(api_func)
    def api_filter_wrapper(self, *args, **kwargs):
        req = kwargs.get('req', {})
        if ((req and 'include_extra_props' in req.GET and
             req.GET['include_extra_props'] == "true") or
                not cfg.CONF.refstack_interoperability):
            return api_func(self, *args, **kwargs)

        def nop(self, *args, **kwargs):
            pass
        return nop(self, *args, **kwargs)

    return api_filter_wrapper

from oslo_messaging._drivers.impl_rabbit import Connection
Connection._set_current_channel = _powervc_set_current_channel
