# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
from __future__ import print_function
import argparse
import csv
import json
import os
import requests
import signal
import sys
import urllib
from powervc_oslo import _
from powervc_oslo import enable_command_translation
from powervc_oslo import get_formatter_class
from powervc_oslo.cmd import auth
enable_command_translation()


BASE_URL = 'http://localhost:8777'

REQUEST_FILTER = [{'q.field': 'event_type',
                   'q.op': 'eq',
                   'q.value': 'audit.http.request'}]
RESPONSE_FILTER = [{'q.field': 'event_type',
                    'q.op': 'eq',
                    'q.value': 'audit.http.response'}]


def signal_handler(signal, frame):
    print
    sys.exit(0)


def create_parser():
    parser = argparse.ArgumentParser(formatter_class=get_formatter_class(),
                                     description=_('Export audit records to '
                                                   'a specified file'))
    parser.add_argument('-u', '--user_name',
                        help=_('The name of the user with which to '
                               'authenticate and request audit data. This is '
                               'an optional flag. The default is the logged '
                               'in user.'),
                        metavar='<user name>')
    parser.add_argument('-p', '--project_name',
                        help=_('The name of the project for which to '
                               'authenticate and request audit data. This is '
                               'an optional flag. The default is '
                               'ibm-default.'),
                        metavar='<project name>',
                        default='ibm-default')
    parser.add_argument('-n', '--top_n',
                        help=_('Upper bound for number of audit records to '
                               'return. This is an optional flag. '
                               'Default is 1000.'),
                        metavar='<number of records>',
                        default=1000)
    parser.add_argument('-o', '--output',
                        help=_('The file to which audit data should be '
                               'exported. This is an optional flag. The '
                               'default is export_audit.json or '
                               'export_audit.csv depending on the output '
                               'format.'),
                        metavar='<output file>')
    parser.add_argument('-f', '--filter',
                        help=_('The file containing filter parameters with '
                               'which to further control what audit data will '
                               'be exported, e.g. to get data for a single '
                               'user. Refer to the PowerVC Knowledge Center '
                               'for more information on filter structure and '
                               'usage. This is an optional flag.'),
                        metavar='<filter file>')
    parser.add_argument('-x', '--output_format',
                        help=_('The format of the exported audit data. This '
                               'is an optional flag. The supported formats '
                               'are json and csv. The default is json.'),
                        choices=['json', 'csv'], default='json')
    return parser


def build_filter_query_string(filters):
    '''Builds a query string given the filters.

    :param filters: List of dictionary objects
    :returns: a url compatible query string
    :raise Exception when filter q.op filter is invalid
    '''

    query_filter = ''
    separator = ''
    for filter_dict in filters:

        if 'q.op' in filter_dict and filter_dict['q.op'] not in ['eq', 'ne',
                                                                 'lt', 'gt',
                                                                 'le', 'ge']:
            raise ValueError('q.op filter field must be one of "eq", "ne", '
                             '"lt", "gt", "le", or "ge"')
        query_filter += separator + urllib.urlencode(filter_dict)
        separator = '&'
    return query_filter


def combine_query_filters(query_filter_filename=None, extra_filter_obj=[],
                          limit=None):
    '''Combine query filter from file, extra default filter,
    and 'limit' parameter.

    :param query_filter_filename: Qualified path to the filter file
    :param extra_filter_obj: Extra query filter
    :param limit: Number of audit records to return
    :returns: query filter string
    '''
    query_filter = ''
    filter_obj = extra_filter_obj
    if query_filter_filename:
        try:
            with open(query_filter_filename, 'r') as f:
                filter_json = f.read().strip()
        except IOError as e:
            print(_('Error opening filter file. %s.') % e,
                  file=sys.stderr)
            sys.exit(1)
        try:
            if filter_json:
                filter_obj += json.loads(filter_json)
        except ValueError as e:
            print(_('Error: Filter file contents are not valid JSON.'
                    'Please try with the right filter format.'),
                  file=sys.stderr)
            sys.exit(1)
    try:
        query_filter = build_filter_query_string(filter_obj)
    except ValueError as e:
        print(_('Error: Check file contents. %s.') % e,
              file=sys.stderr)
        sys.exit(1)
    if limit:
        query_filter += '&limit=%s' % (limit)

    return query_filter


def do_request(auth_token, request_url, query_filter):
    '''Builds and performs the ceilometer events request.

    :param auth_token: The authentication token
    :param request_url: The url of request
    :param query_filter: Query filter for the request
    :returns: HTTP Response in json format
    '''

    headers = {'X-Auth-Token': auth_token,
               'Accept': 'application/json',
               'User-Agent': 'powervc-audit-export'}

    response = requests.get(request_url,
                            headers=headers,
                            params=query_filter)

    raw_response = response.text
    if response.status_code != 200:
        print(_('REST API call failed with error %s.') % response.status_code,
              file=sys.stderr)
        print(raw_response, file=sys.stderr)
        sys.exit(1)

    return json.loads(raw_response)


def query_and_combine(auth_token, top_n, query_filter_filename=None):
    '''Combines the result of two http requests.

    Makes two queries to ceilometer events resource -
    one for audit.http.request and one for audit.http.response.
    It then combines the two into single list.

    :param auth_token_filename: Authentication token
    :param top_n: Number of records to return
    :param query_filter_filename: Query filter file name
    :returns: Combined list of dictionaries
    '''

    # get more records because we might be missing some responses
    limit = int(top_n * 0.6)

    query_filter_requests = combine_query_filters(query_filter_filename,
                                                  REQUEST_FILTER, limit)
    query_filter_responses = combine_query_filters(query_filter_filename,
                                                   RESPONSE_FILTER, limit)
    request_url = '%s/v2/events' % (BASE_URL)
    reqs = do_request(auth_token, request_url, query_filter_requests)
    resps = do_request(auth_token, request_url, query_filter_responses)

    combined = reqs + resps
    # sort records by date/time
    combined.sort(key=lambda item: item['generated'])
    # trim to top_n
    combined = combined[:top_n]

    return combined


def export_to_file(combined, output_filename, output_format):
    '''Exports results to a file taking into account the output format.

    :param combined: Combined list of requests and responses
    :param output_filename: Output file name
    :param output_format: Output format - json or csv
    '''

    if combined:

        if output_format == 'json':
            out_json = None
            out_json = json.dumps(combined,
                                  sort_keys=True,
                                  indent=4,
                                  separators=(',', ': '))
            with os.fdopen(os.open(output_filename, os.O_CREAT | os.O_WRONLY,
                                   0o600), 'w') as f:
                f.write(out_json)
                # truncate old content
                f.truncate()

        elif output_format == 'csv':
            keys = combined[0].keys()
            headers = []

            # Only use desired column headers
            for header in keys:
                if header != "raw" and header != "traits":
                    headers.append(header)

            # Reformat each event to split out the traits as columns
            for event in combined:
                for trait in event["traits"]:
                    event[trait["name"]] = trait["value"]
                    # Check if this is a new header
                    new = True
                    for header in headers:
                        if header == trait["name"]:
                            new = False
                            break
                    # if new header, add to header list
                    if new:
                        headers.append(trait["name"])
                # Remove undesirable columns
                del event["traits"]
                del event["raw"]

            # Order headers
            headers.sort()

            # Output to CSV
            with os.fdopen(os.open(output_filename, os.O_CREAT | os.O_WRONLY,
                                   0o600), 'w') as f:
                dict_writer = csv.DictWriter(f, headers)
                dict_writer.writeheader()
                dict_writer.writerows(combined)
                # truncate old content
                f.truncate()

        print(_('Audit records written to file: %s') % output_filename)

    else:
        print(_('No audit records available. Nothing to do.'))


def export_audit(options):
    '''Main function.

    :param options: Command line arguments
    '''
    try:
        sess = auth.get_session(options.user_name, options.project_name,
                                user_agent='powervc-audit-export')
    except:
        print(_('Login failed.'), file=sys.stderr)
        sys.exit(1)
    is_admin = False
    roles = sess.auth.auth_ref.role_names
    for role in roles:
        if role == 'admin':
            is_admin = True
    if not is_admin:
        print(_('Only users with the admin role are permitted to view ' +
                'audit records.'),
              file=sys.stderr)
        sys.exit(1)

    user_path = os.path.expanduser('~')
    if options.output:
        if os.path.dirname(options.output) != '':
            output_file = options.output
        else:
            output_file = user_path + '/' + options.output
    else:
        output_file = user_path + '/export_audit.' + options.output_format

    try:
        top_n = int(options.top_n)
        if top_n < 1:
            print(_('Number of records must be larger than 0'),
                  file=sys.stderr)
            sys.exit(1)
    except ValueError:
        print(_('Number of records must be an integer'), file=sys.stderr)
        sys.exit(1)

    token = sess.get_token()
    combined = query_and_combine(token, top_n, options.filter)
    export_to_file(combined, output_file, options.output_format)


def main():
    argv = sys.argv[1:]
    signal.signal(signal.SIGINT, signal_handler)
    options = create_parser().parse_args()

    # Delete arguments before reading ceilometer config file.
    # This is done to prevent this script's CLI arguments from being flagged
    # by the ceilometer config
    if argv:
        del argv[0:]
        del sys.argv[1:]

    export_audit(options)
