# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
import getpass
import os
import requests
import sys

from keystoneauth1.identity import v3 as auth_v3
from keystoneauth1 import session

from powervc_oslo import _

from requests.packages.urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)


def get_session(user=None, project=None, user_agent='powervc'):
    """
    Helper method to get a keystone session for PowerVC CLI commands.

    The session object will be created and authentication with the provided
    credentials will be tested.  The credentials are a combination of user,
    password and project name.  The user and project name can optionally be
    passed in as parameters.

    If a user is passed in, a prompt is provided for the password and the
    passed project will be used, defaulting to 'ibm-default'.

    If a user is not passed in, the OS_USERNAME environment variable will be
    checked.  If it is set, the corresponding OS_PASSWORD and OS_PROJECT_NAME
    environment variables will also be used.

    If no username is passed in and the OS_USERNAME environment variable is not
    set, the Python internals will be used to get the current user and a prompt
    will be shown for the password, and the passed project will be used,
    defaulting to 'ibm-default'.

    If the credentials are invalid, an exception will be raised.

    :param user: The optional user name to authenticate the session with.
    :param project: The optional project name for the associated user.
    :param user_agent: The user-agent string to pass on requests using this
                       session.
    :returns session: A session object.
    """
    password = None
    if user:
        # first choice, input from caller
        username = user
        project_name = project or 'ibm-default'
    else:
        # second choice, env vars
        username = os.getenv('OS_USERNAME')
        if username:
            password = os.getenv('OS_PASSWORD', '')
            project_name = os.getenv('OS_PROJECT_NAME', 'ibm-default')
    if not username:
        # third choice, OS user
        username = getpass.getuser()
        project_name = project or 'ibm-default'
    if not password:
        sys.stderr.write(_("\nEnter password for %s:\n") % username)
        password = getpass.getpass('')

    # Create auth and session objects
    auth = auth_v3.Password(auth_url='https://localhost:5000/v3',
                            user_domain_name='Default',
                            username=username,
                            password=password,
                            project_domain_name='Default',
                            project_name=project_name)
    sess = session.Session(auth=auth, verify=False,  # nosec localhost
                           user_agent=user_agent)

    # Attempt to get a token to test the credentials
    sess.get_token()

    return sess
