# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import os
import sys
import argparse

from powervc_oslo.config import data_mgr as mgr
from powervc_oslo.config import data_utils as utils
from powervc_oslo import _


#################################################
#            Common Utility Methods             #
#################################################
def _parse_arguments():
    """Parses the arguments from the command line that was input"""
    # Introduce the MRI to get translated until we expose the active parameter
    _active = _("Perform backup while the PowerVC services are running.")
    parser = argparse.ArgumentParser(
        description=_("The powervc-backup command is used to save off the "
                      "PowerVC data to an archive .tgz file so that it can be "
                      "used at a later point to recover to the saved state."))
    parser.add_argument('--noprompt', dest='noprompt', action='store_true',
                        help=_("If specified, no user intervention is required"
                               " during execution of the backup process."))
    parser.add_argument('--targetdir', dest='location', action='store',
                        help=_("Target location in which to save the backup "
                               "archive.  Default value is /var/opt/ibm/"
                               "powervc/backups."), default='')
    parser.add_argument('--active', dest='active',
                        action='store_true', help=argparse.SUPPRESS)
    parser.add_argument('--usestdout', dest='usestdout',
                        action='store_true', help=argparse.SUPPRESS)
    parser.add_argument('--noPrompt', dest='nopromptold',
                        action='store_true', help=argparse.SUPPRESS)
    parser.add_argument('--targetDir', dest='locationold', default='',
                        action='store', help=argparse.SUPPRESS)
    return parser.parse_args()


#################################################
#               Main Entry Point                #
#################################################
def main():
    """Main invocation for the powervc-backup command"""
    # Parse the arguments that were provided on the command line
    args = _parse_arguments()
    location = args.location if args.location else args.locationold
    # If they gave us a relative path for the location, correct it
    if location and not location.startswith('/'):
        location = os.path.abspath(location)
    # If they didn't ask to suppress prompting, then prompt them now
    if not args.noprompt and not args.nopromptold:
        warning = _("Continuing with this operation will stop all "
                    "PowerVC services.  Do you want to continue? (y/N):")
        if not args.active and not utils.prompt_for_confirmation(warning):
            sys.exit(0)
        args.usestdout = True
    # Enable Logging to the Log File and Standard Out if they asked
    log_file = utils.setup_logging('powervc_backup', args.usestdout)
    # Now we can actually perform the backup that was requested
    try:
        mgr.backup_data(location, args.active)
    # If they gave us a user error, just log the message and exit
    except SystemExit as exc:
        utils.LOG.log(utils.CONSOLE, exc.message)
        sys.exit(1)
    # If we got an unexpected exception, print the generic error
    except Exception as exc:
        msg = _("Errors occurred during the backup operation.  See log file "
                "%s for details about any potential solutions for any errors.")
        utils.LOG.log(utils.CONSOLE, unicode(msg % log_file))
        utils.LOG.exception(unicode(exc))
        sys.exit(2)
