# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import os
import sys
import argparse
import traceback
import pbr.version

from oslo_config import cfg
from oslo_i18n import _gettextutils
from oslo_db.sqlalchemy import migration
from powervc_oslo.config import data_utils
from powervc_oslo import _

# Initially setup a no-op logger, so we don't get the config messages
data_utils.setup_logging()


#################################################
#            Common Utility Methods             #
#################################################
def _parse_arguments():
    """Parses the arguments from the command line that was input"""
    parser = argparse.ArgumentParser(
        description=_("The powervc-dbsync command is used to update the "
                      "component's database tables to the latest version."))
    parser.add_argument('component', action='store',
                        help=_("The component for which to perform the table u"
                               "pdate.  Valid values are cinder|nova|glance."))
    parser.add_argument('version', action='store', nargs='?',
                        help=_("The version to which to update the tables.  "
                               "The default will be the latest version."))
    return parser, parser.parse_args()


def _initialize_component(component):
    """Initializes the Configuration and Logging for the Component"""
    dbstr = '%s.db.sqlalchemy.api' if (
        component != 'ceilometer' and component != 'clerk')else (
            'powervc_%s.db.sqlalchemy.api')

    migstr = 'powervc_%s.db.sqlalchemy.migrate_repo'
    from oslo_log import log as logging
    # Do the Installation Setup for the NLS Bundle
    _gettextutils.install(component)
    # Load the Imports that we need for the specified Component
    __import__(migstr % component)
    __import__(dbstr % component)
    # Setup the Configuration properties for the Component
    cfile = ['/etc/glance/glance-api.conf'] if component == 'glance' else None
    # Setup the Logging for the user-specified Component
    logging.register_options(cfg.CONF)
    cfg.CONF([], project=component, default_config_files=cfile,
             version=pbr.version.VersionInfo(component).version_string())
    logging.setup(cfg.CONF, component)
    # Return the Path of the Migration Repository for the Component
    return sys.modules[dbstr % component], sys.modules[migstr % component]


#################################################
#               Main Entry Point                #
#################################################
def main():
    """Main invocation for the powervc-dbsync command"""
    # Parse the arguments that were provided on the command line
    parser, args = _parse_arguments()
    # Only Cinder, Nova & Glance are supported at this time for DB Synch'ing
    if args.component not in ['cinder', 'nova', 'glance', 'ceilometer',
                              'clerk']:
        parser.print_usage()
        sys.stderr.write(_("powervc-dbsync: error: unsupported component"))
        sys.exit(1)
    try:
        # Perform any initialization that is needed for the Component
        db_api, migr_repo = _initialize_component(args.component)
        # Now we want to update the Repository to match our PowerVC Repository
        path = os.path.abspath(os.path.dirname(migr_repo.__file__))
        engine = db_api.get_engine()
        # We need to create our repository in the DB if it doesn't exist yet
        try:
            migration.db_version(engine, path, 000)
        except:
            migration.db_version_control(engine, path, 000)
        # Now we can run the DB Synchronization on the PowerVC Repository
        migration.db_sync(engine, path, args.version, 0, False)
    except Exception as exc:
        # If we got an exception, lets log the CallStack, so we know the
        # failure
        traceback.print_exc(exc)
        sys.exit(2)
