# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import sys
import argparse
from datetime import datetime, timedelta

from powervc_oslo.config import data_mgr
from powervc_oslo.config import data_utils
from powervc_oslo import _

SYSINIT_DIR = '/etc/rc.d/init.d/'
SYSTEMD_DIR = '/usr/lib/systemd/system/'


#################################################
#            Main Operation Methods             #
#################################################
def _start_managing(input_file):
    """Configures the KVM Host to be managed by the new Mgmt System"""
    mgmt_cfg = data_utils.read_config_file(input_file)['DEFAULT']
    data_utils.get_config_data().update(mgmt_cfg)
    # Verify that the Host Name is properly set on the KVM Host
    _verify_local_hostname_set()
    # Verify there isn't too much of a time skew for us to manage
    _verify_mgmt_system_time_skew(mgmt_cfg)
    # Copy the certificates to their actual location since running as root
    _copy_certificates_to_pki(input_file)
    try:
        # Call the Data Manager API to configure the data
        data_mgr.configure_data()
        # Finally we can restart the PowerVC services when done
        data_mgr.restart_services()
    except:
        # If we got an exception, we want to change everything back
        data_utils.safe_method_call(_stop_managing)
        raise


def _stop_managing(new_mgmt_ip=None):
    """Configures the KVM Host to no longer be managed by the Mgmt System"""
    # Stop the Services before resetting config, since it isn't managed anymore
    data_mgr.stop_services()
    # If they asked use to notify the old Management System, do that now
    if new_mgmt_ip:
        notifycmd = '/usr/bin/python -c "from powervc_nova.compute import '\
                    'remanage; remanage.notify_unmanage(\'%s\')"' % new_mgmt_ip
        data_utils.run_command(notifycmd, False)
    # Now we can loop through the Data Managers an unconfigure any data set
    data_mgr.unconfigure_data()


def _dump_info(output_file):
    """Dumps Config from the Mgmt System to use to configure the KVM Host"""
    mgrs = data_mgr.get_registered_managers()
    # Get the Management Information and Service Passwords for the Data
    # Managers
    mgmt_cfg = dict(mgmt_ip=data_utils.get_config_data()['mgmt_ip'])
    mgmt_cfg.update(data_mgr._get_service_passwords(mgrs))
    # Write out the Management Configuration to the provided File
    data_utils.write_config_file(output_file, mgmt_cfg)


def _update_mgmt_ip(new_mgmt_ip):
    """Performs any Configuration changes on the KVM Host on a MgmtIP Change"""
    data_utils.get_config_data()['mgmt_ip'] = new_mgmt_ip
    # As the Data Manager API to do the updates for the Management IP changing
    data_mgr.management_ip_changed()


def _generate_certificates(output_dir):
    """Generates the Certificates to be used on the KVM Host"""
    mgrs = data_mgr.get_registered_managers()
    # Loop through each of the Managers, getting the KVM Host Certificate Info
    for mgr in mgrs:
        ca_info = mgr._get_certificate_authority_info()
        cert_info = mgr._get_remotehost_certificate_info()
        # If the Manager has provided KVM Certificate Info, then create CSR's
        if cert_info and ca_info and cert_info.get('csrs'):
            cert_dir = cert_info.get('certdir')
            # Setup the Directories now for the Certificate and Private Keys
            data_utils.makedirs(cert_dir, 0o755)
            data_utils.makedirs(cert_dir + '/private', 0o750)
            data_utils.chown(cert_dir, 'root', cert_info['group'])
            data_utils.chown(cert_dir + '/private', 'root', cert_info['group'])
            # Loop through each of the CSR's, creating the Private Key and CSR
            for csr in cert_info['csrs']:
                crt_file = ca_info['component'] + '-' + csr['certfile']
                csr_file = output_dir + '/' + crt_file.replace('.pem', '.csr')
                # Create the Signed Certificate based on the CSR given
                data_utils.create_signed_certificate(
                    ca_info, output_dir, crt_file, csr_file, csr['exttype'])
                # Update the certificate start time in case KVM host is ahead
                _update_cert_notbefore(
                    output_dir + '/' + crt_file, ca_info['privkey'])
                # Cleanup the temporary CSR from the Output Directory
                data_utils.remove_file(csr_file)
                # Make sure Nova has authority to transfer the Certificates
                data_utils.chown(output_dir + '/' + crt_file, 'nova', 'nova')


def _generate_signing_requests(output_dir):
    """Generates the Certificate Signing Requests to be used on the KVM Host"""
    data_utils.makedirs(output_dir, 0o777)
    mgrs = data_mgr.get_registered_managers()
    # Loop through each of the Managers, getting the KVM Host Certificate Info
    for mgr in mgrs:
        ca_info = mgr._get_certificate_authority_info()
        cert_info = mgr._get_remotehost_certificate_info()
        # If the Manager has provided KVM Certificate Info, then create CSR's
        if cert_info and ca_info and cert_info.get('csrs'):
            comp, cert_dir = ca_info.get('component'), cert_info.get('certdir')
            # Setup the Directories now for the Certificate and Private Keys
            data_utils.makedirs(cert_dir, 0o755)
            data_utils.makedirs(cert_dir + '/private', 0o750)
            data_utils.chown(cert_dir, 'root', cert_info['group'])
            data_utils.chown(cert_dir + '/private', 'root', cert_info['group'])
            # Loop through each of the CSR's, creating the Private Key and CSR
            for csr in cert_info['csrs']:
                key_file = cert_dir + '/private/' + csr['keyfile']
                csr_file = comp + '-' + csr['certfile'].replace('.pem', '.csr')
                # Create the Private Key in its final resting location
                data_utils.create_private_key(
                    cert_dir + '/private', csr['keyfile'], cert_info['group'])
                # Create the CSR in the temporary directory to be transferred
                data_utils.create_cert_signing_request(
                    ca_info, output_dir, csr_file, key_file, csr['exttype'])


def _add_host(service, file_name):
    """Adds the Service and Configuration File for the Host/Storage"""
    parts = service.split('-')
    comp, svcname, host = parts[1], parts[2], '-'.join(parts[3:])
    config = data_utils.read_config_file(file_name)
    config.setdefault('DEFAULT', dict())['host'] = host
    base_svc = 'openstack-%s-%s' % (comp, svcname)
    base_file = '/etc/%(comp)s/%(comp)s.conf' % locals()
    conf_file = '/etc/%(comp)s/%(comp)s-%(host)s.conf' % locals()
    # Special case for the backup/restore to just copy the backup
    if file_name.endswith('_backup'):
        base_file, config = file_name, dict()
    # First we will copy over the Template File to the new Name
    data_utils.copy_file(base_file, conf_file)
    data_utils.chown(conf_file, 'root', comp)
    data_utils.chmod(conf_file, 0o660)
    # Next create the log file in order to get proper permissions
    prefix = 'nova-' if comp == 'nova' else ''
    log_file = '/var/log/%s/%s%s-%s.log' % (comp, prefix, svcname, host)
    if not data_utils.file_exists(log_file):
        data_utils.create_file(log_file)
    data_utils.chown(log_file, 'root', comp)
    data_utils.chmod(log_file, 0o660)
    # Next we will write out any of the new properties to the file
    _update_host_config(conf_file, config)
    # Create the new service file for the Systemd Service
    temp_service = '/usr/share/%s/services/%s.service' % (comp, base_svc)
    # Temporary fall-back to the old location for the service template
    if not data_utils.file_exists(temp_service):
        temp_service = '/etc/%s/%s.service' % (comp, base_svc)
    data_utils.copy_file(temp_service, SYSTEMD_DIR + service + '.service')
    _update_host_service_file(
        SYSTEMD_DIR + service + '.service', comp, svcname, host)
    data_utils.run_command('/bin/systemctl daemon-reload')
    # Finally we will enable the service so that we can start it up
    data_utils.run_command('/sbin/chkconfig ' + service + ' on')


def _edit_host(service, file_name):
    """Adds the Service and Configuration File for the Host/Storage"""
    parts = service.split('-')
    comp, svcname, host = parts[1], parts[2], '-'.join(parts[3:])
    config = data_utils.read_config_file(file_name)
    conf_file = '/etc/%(comp)s/%(comp)s-%(host)s.conf' % locals()
    # Update the Configuration File and don't restart the Service
    _update_host_config(conf_file, config)


def _delete_host(service):
    """Adds the Service and Configuration File for the Host/Storage"""
    parts = service.split('-')
    comp, svcname, host = parts[1], parts[2], '-'.join(parts[3:])
    conf_file = '/etc/%(comp)s/%(comp)s-%(host)s.conf' % locals()
    # First we need to stop the Service if it is still running
    data_utils.run_command('/bin/systemctl stop %s' % service)
    # Next delete the Configuration File for the given Host
    data_utils.remove_file(conf_file)
    # If this is the new SystemD Service, clean that up
    if data_utils.file_exists(SYSTEMD_DIR + service + '.service'):
        data_utils.remove_file(SYSTEMD_DIR + service + '.service')
        data_utils.run_command('/bin/systemctl daemon-reload')


def _synchronize_uuids(host, vm_uuid):
    """Synchronizes the OpenStack VM UUID's with the Hypervisor"""
    uuid_map = dict()
    mgrs = data_mgr.get_registered_managers()
    # First loop through the managers building the UUID Mappings
    for mgr in mgrs:
        uuids = mgr.uuids_to_synchronize(host, vm_uuid)
        if uuids:
            uuid_map.update(uuids)
    # Now we can go through the managers synchronizing the UUID's
    for mgr in mgrs:
        if uuid_map:
            mgr.synchronize_uuids(host, uuid_map)


def _purge_deleted_database_records(days, hours, noprompt, chunksize):
    """Purges deleted records from the OpenStack Databases"""
    # If they gave us days, verify and convert to hours
    if days is not None:
        # They must give us a positive number of days for deletion
        if not days.isdigit() or int(days) <= 0:
            sys.stderr.write(
                _("The number of days must be a positive integer."))
            sys.stderr.write('\n')
            sys.exit(3)
        hours = str(int(days) * 24)
    # If we didn't get days or hours, then say days is required
    if not hours:
        sys.stderr.write(_("The number of days parameter must be specified."))
        sys.stderr.write('\n')
        sys.exit(4)
    # They must give us a positive number of hours for deletion
    if not hours.isdigit() or int(hours) <= 0:
        sys.stderr.write(
            _("The number of hours must be a positive integer."))
        sys.stderr.write('\n')
        sys.exit(5)
    # If they didn't ask to suppress the prompt, then confirm with the user
    if not noprompt:
        warning = _("Continuing with this operation will purge deleted records"
                    " from the databases.  Do you want to continue? (y/N):")
        if not data_utils.prompt_for_confirmation(warning):
            sys.exit(0)
    # Get a time-stamp that subtracts the number of days/hours provided
    timestamp = datetime.utcnow() - timedelta(hours=int(hours))
    # Loop through the Data Managers, asking to purge the deleted records
    for mgr in data_mgr.get_registered_managers():
        mgr.purge_deleted_database_records(timestamp, chunksize)


#################################################
#            Common Utility Methods             #
#################################################
def _parse_arguments():
    """Parses the arguments from the command line that was input"""
    parser = argparse.ArgumentParser(
        description='The powervc-manage command is used to configure '
                    'OpenStack on a Compute Node to be managed by a '
                    'given Management System.  This command allows '
                    'the caller to either specify a new system to '
                    'manage this node or specify to stop managing this '
                    'node.  This command will also provide the ability '
                    'to purge deleted database records older than the '
                    'number of days given.'
    )
    parser.add_argument("-o", '--operation', dest="operation",
                        help="specifies the operation to be perfomed "
                             "by the command <manage|unmanage|dbpurge>")
    parser.add_argument("-f", '--file', dest="file_name",
                        help="specifies the path of the file that "
                             "is used as part of the operation")
    parser.add_argument("-n", '--notify', dest="new_mgmt_ip",
                        help="specifies to notify the existing management "
                             "system that the management system with the "
                             "ip specified will be taking over management")
    parser.add_argument("-d", '--older-than-days', dest="days",
                        help="specifies the number of days that database "
                             "records have been deleted for before being "
                             "purged (recommended 7 days or greater)")
    parser.add_argument("-y", '--noprompt',
                        dest="noprompt", action='store_true',
                        help="specifies whether to suppress the warning "
                             "prompt message when purging deleted records")
    parser.add_argument("-c", '--commonname',
                        dest="hostname", help=argparse.SUPPRESS)
    parser.add_argument("-s", '--service',
                        dest="service", help=argparse.SUPPRESS)
    parser.add_argument("-u", '--uuid',
                        dest="uuid", help=argparse.SUPPRESS)
    parser.add_argument("-r", '--older-than-hours',
                        dest="hours", help=argparse.SUPPRESS)
    parser.add_argument("-k", '--chunk-size', default='500',
                        dest="chunksize", help=argparse.SUPPRESS)
    return parser.parse_args()


def _verify_local_hostname_set():
    """Helper method to verify that the local host name is set"""
    output = data_utils.run_command_output('/bin/hostname', False)[1]
    if output.strip() == 'localhost' or output.startswith('localhost.'):
        sys.stderr.write("The hostname is set to localhost, can't configure "
                         "KVM Host to be managed without the hostname set.")
        sys.exit(32)


def _verify_mgmt_system_time_skew(mgmt_cfg):
    """Helper method to verify the Management System time skew is in-check"""
    # If we weren't given the Management System's time, we can't do anything
    if not mgmt_cfg.get('time_now'):
        return
    # Get the current time on the KVM Host and determine the difference
    date_secs = datetime.utcnow() - datetime(1970, 1, 1)
    time_secs = (date_secs.days * 86400L) + date_secs.seconds
    time_diff = time_secs - long(mgmt_cfg['time_now'])
    # If the time is off by more than 30 minutes, then we won't allow managing
    if time_diff > 1800 or time_diff < -1800:
        sys.stderr.write("Time difference from Management System greater than "
                         "30 minutes, can't configure KVM Host to be managed.")
        sys.exit(30)


def _update_host_config(conf_file, config):
    """Helper method to update the configuration properties in the file"""
    for sect in config:
        # Loop through each of the Key/Value Pairs in the Configuration File
        for key, val in config[sect].iteritems():
            # If they provided an empty value, this means we should delete it
            if val is None or len(val) <= 0:
                data_utils.del_openstack_config_value(conf_file, key, sect)
                continue
            # Otherwise we will set the Key/Value in the Configuration File
            data_utils.set_openstack_config_value(conf_file, key, val, sect)


def _update_host_service_file(file_name, comp, svcname, host):
    """Helper method to update the template service file for this host"""
    data1, data2 = data_utils.read_from_file(file_name), ''
    add = ' --config-file /etc/cinder/fabrics.conf' if comp == 'cinder' else ''
    # Loop through each of the lines in the file doing the substitution
    for line in data1.split('\n'):
        # We don't want to right out the config file check line
        if line.find('[ -f $config ]') >= 0:
            continue
        line = line.replace('/etc/$proj/$proj.conf',
                            '/etc/$proj/$proj-' + host + '.conf' + add)
        line = line.replace('/etc/%s/%s.conf' % (comp, comp),
                            '/etc/%s/%s-%s.conf%s' % (comp, comp, host, add))
        line = line.replace('/var/log/%s/%s.log' % (comp, svcname),
                            '/var/log/%s/%s-%s.log' % (comp, svcname, host))
        line = line.replace('/var/log/$proj/$suffix.log',
                            '/var/log/$proj/$suffix-' + host + '.log')
        line = line.replace('/var/run/$proj/$proj-$suffix.pid',
                            '/var/run/$proj/$proj-$suffix-' + host + '.pid')
        # If they didn't specify the log file, we need to add the log file
        if line.startswith('ExecStart=') and line.find('--logfile') < 0:
            replvars = (comp, comp, svcname, host)
            line += ' --log-file /var/log/%s/%s-%s-%s.log' % replvars
        # If they didn't specify the CONF file, we need to add that file
        if line.startswith('ExecStart=') and line.find('--config-file') < 0:
            line += ' --config-file /etc/%s/%s-%s.conf' % (comp, comp, host)
        data2 += line + '\n'
    data_utils.write_to_file(file_name, data2)


def _copy_certificates_to_pki(input_file):
    """Helper method copy the certificates to their actual location"""
    temp_dir, certs = input_file[:input_file.rfind('/')], list()
    # Loop through each of the managers getting the certificate info
    for manager in data_mgr.get_registered_managers():
        ca_info = manager._get_certificate_authority_info()
        cert_info = manager._get_remotehost_certificate_info()
        comp = ca_info.get('component') if ca_info else None
        # If there are any CA Certificates to transfer, add them now
        if cert_info and cert_info.get('cacert'):
            prefix = (comp + '-') if comp else ''
            cert_file = cert_info['cacert'][0]
            cert_file = cert_file[cert_file.rfind('/') + 1:]
            cert_file = temp_dir + '/' + prefix + cert_file
            certs.append((cert_file, cert_info['cacert'][1]))
        # Loop through the CSR's, adding any of those to the transfer list
        if cert_info and cert_info.get('csrs'):
            cert_dir = cert_info['certdir']
            for csr in cert_info['csrs']:
                cert_file = comp + '-' + csr['certfile']
                certs.append((temp_dir + '/' + cert_file,
                              cert_dir + '/' + csr['certfile']))
    # Loop through the certificate files, copying over the existing ones
    for fromfile, tofile in certs:
        if data_utils.file_exists(fromfile):
            todir = tofile[:tofile.rfind('/')]
            # If the Directory doesn't exist, then we need to create it
            if not data_utils.file_exists(todir):
                data_utils.makedirs(todir, 0o755)
                data_utils.chown(todir, 'root', 'root')
            # Copy the file over and make sure the permissions are correct
            data_utils.copy_file(fromfile, tofile)
            data_utils.chown(tofile, 'root', 'root')
            data_utils.chmod(tofile, 0o444)
    return certs


def _update_cert_notbefore(cert_file, key_file):
    """Helper method to update the Certificate start date if too new"""
    from OpenSSL import crypto
    # Read the Certificate into Memory to be able to update it
    data = data_utils.read_from_file(cert_file)
    cert = crypto.load_certificate(crypto.FILETYPE_PEM, data)
    # Since the KVM Host needs to be within 30 minutes, go an hour back
    starttime = datetime.utcnow() - timedelta(minutes=60)
    cert.set_notBefore(starttime.strftime('%Y%m%d%H%M%SZ'))
    # Load the CA's Private Key to be able to sign the updated Certificate
    data = data_utils.read_from_file(key_file)
    pkey = crypto.load_privatekey(crypto.FILETYPE_PEM, data)
    cert.sign(pkey, 'sha256')
    # Write out the updated Certificate to the original file name
    data = crypto.dump_certificate(crypto.FILETYPE_PEM, cert)
    data_utils.write_to_file(cert_file, data)


#################################################
#               Main Entry Point                #
#################################################
def main():
    """Main invocation for the powervc-manage command"""
    # Parse the arguments that were provided on the command line
    args = _parse_arguments()
    # Disable Logging completely for the powervc-manage command
    data_utils.setup_logging()
    # If they asked us to start managing with a new Management System
    if args.operation and args.operation.lower() == 'manage':
        _start_managing(args.file_name)
    # If they asked us to stop managing from existing Management System
    elif args.operation and args.operation.lower() == 'unmanage':
        _stop_managing(args.new_mgmt_ip)
    # If they asked us to dump the configuration for the Management System
    elif args.operation and args.operation.lower() == 'dump':
        _dump_info(args.file_name)
    # If they asked us to generate the Certificates for the KVM Host
    elif args.operation and args.operation.lower() == 'gencert':
        _generate_certificates(args.file_name)
    # If they asked us to generate the Signing Requests for the KVM Host
    elif args.operation and args.operation.lower() == 'gencsr':
        _generate_signing_requests(args.file_name)
    # If they asked us to upgrade the configuration for the KVM Host
    elif args.operation and args.operation.lower() == 'updateip':
        _update_mgmt_ip(args.new_mgmt_ip)
    # If they asked us to add the OpenStack Service for the Host/Storage
    elif args.operation and args.operation.lower() == 'addhost':
        _add_host(args.service, args.file_name)
    # If they asked us to update the OpenStack Service for the Host/Storage
    elif args.operation and args.operation.lower() == 'edithost':
        _edit_host(args.service, args.file_name)
    # If they asked us to delete the OpenStack Service for the Host/Storage
    elif args.operation and args.operation.lower() == 'removehost':
        _delete_host(args.service)
    # If they asked us to delete the OpenStack Service for the Host/Storage
    elif args.operation and args.operation.lower() == 'syncuuids':
        _synchronize_uuids(args.service, args.uuid)
    # If they asked us to purge deleted database records for the services
    elif args.operation and args.operation.lower() == 'dbpurge':
        _purge_deleted_database_records(
            args.days, args.hours, args.noprompt, int(args.chunksize))
    # If they didn't give an operation or it was invalid, print error message
    else:
        sys.stderr.write(
            'The operation specified must be <manage|unmanage|dbpurge>.\n')
        sys.exit(2)
