# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import sys
import argparse

from powervc_oslo.config import data_mgr
from powervc_oslo.config import data_utils
from powervc_oslo import _


#################################################
#            Common Utility Methods             #
#################################################
def _parse_arguments():
    """Parses the arguments from the command line that was input"""
    subset = _get_service_control_list()
    parser = argparse.ArgumentParser(
        description=_("The powervc-services command is used to control "
                      "the lifecycle of the PowerVC services."))
    parser.add_argument('--skiplogcomp', dest='skiplogcomp',
                        action='store_true', help=argparse.SUPPRESS)
    parser.add_argument('services_subset', action='store', nargs='?',
                        help=_("The subset of services to control if "
                               "specified, otherwise all will be.  The "
                               "valid values are: %s") % subset)
    parser.add_argument('operation', action='store',
                        help=_("The operation to take on the services.  "
                               "Valid values are start|stop|restart|status"))
    parser.add_argument('single_op', action='store',
                        nargs='?', help=argparse.SUPPRESS)
    return parser, parser.parse_args()


def _get_service_control_list():
    """Returns a comma-separated list of the components with services"""
    mgrs = data_mgr.get_registered_managers()
    comps = [m.get_component_name() for m in mgrs if m._get_enabled_services()]
    return ', '.join(comps)


#################################################
#               Main Entry Point                #
#################################################
def main():
    """Main invocation for the powervc-services command"""
    # Parse the arguments that were provided on the command line
    parser, args = _parse_arguments()
    component = args.services_subset
    # Enable Logging to the of everything to standard out only
    data_utils.setup_logging(use_console=True)
    # If they provided a 3rd parameter, verify it is the special single option
    if args.single_op:
        # If the 1st parameter isn't single, then invalid syntax
        if args.services_subset != 'single':
            parser.print_usage()
            sys.stderr.write(
                _("powervc-services: error: too many arguments\n"))
            sys.exit(1)
        # Otherwise run a command to perform the operation on the service
        rc = data_utils.run_command(
            '/bin/systemctl %s %s' % (args.single_op, args.operation))
        sys.exit(rc)
    try:
        # For legacy reasons we will support keystone to restart httpd
        component = 'httpd' if component == 'keystone' else component
        component = 'db' if component == 'db2' else component
        # If they asked to start the services, ask the manager to start them
        if args.operation == 'start':
            rc = data_mgr.start_services(component, not args.skiplogcomp)
        # If they asked to stop the services, ask the manager to stop them
        elif args.operation == 'stop':
            rc = data_mgr.stop_services(component, not args.skiplogcomp)
        # If they asked to restart the services, ask the manager to restart
        # them
        elif args.operation == 'restart':
            rc = data_mgr.restart_services(component, not args.skiplogcomp)
        # If they asked for the service status, ask the manager for it
        elif args.operation == 'status':
            rc = data_mgr.services_status(component)
        # If they didn't give an expected operation, then invalid usage
        else:
            _x, rc = parser.print_usage(), 1
            sys.stderr.write(_("powervc-services: error: unrecognized "
                               "arguments: %s\n") % args.operation)
        sys.exit(rc)
    # If we got an exception, print the main reason for the user
    except Exception as exc:
        sys.stderr.write(unicode(exc) + '\n')
        sys.exit(2)
