# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
from __future__ import print_function
import sys
from powervc_oslo.config import data_utils
from powervc_oslo import _


NOVA_API_CONF = '/etc/nova/api-paste.ini'
NEUTRON_API_CONF = '/etc/neutron/api-paste.ini'
CINDER_API_CONF = '/etc/cinder/api-paste.ini'
GLANCE_API_CONF = '/etc/glance/glance-api-paste.ini'
VALIDATOR_API_CONF = '/etc/ttv-validation/api-paste.ini'
CEILOMETER_API_CONF = '/etc/ceilometer/api_paste.ini'
CLERK_CONF = '/etc/clerk/api-paste.ini'

SERVICE_CONF = {'nova': NOVA_API_CONF,
                'glance': GLANCE_API_CONF,
                'ceilometer': CEILOMETER_API_CONF,
                'neutron': NEUTRON_API_CONF,
                'cinder': CINDER_API_CONF,
                'validator': VALIDATOR_API_CONF,
                'clerk': CLERK_CONF}

SERVICE_NAMES = {'compute': 'nova',
                 'image': 'glance',
                 'metering': 'ceilometer',
                 'network': 'neutron',
                 'storage': 'cinder',
                 'validation': 'validator',
                 'self-service': 'clerk'}

service_map = {'glance': 'image',
               'nova': 'compute',
               'neutron': 'network',
               'cinder': 'storage',
               'ceilometer': 'metering',
               'validator': 'validation',
               'clerk': 'self-service'}

SECTIONS = {'nova': ['composite:openstack_compute_api_legacy_v2',
                     'composite:openstack_compute_api_v21',
                     ('composite:openstack_compute_api_v21_legacy_v2_' +
                      'compatible')],
            'glance': ['pipeline:glance-api-keystone',
                       'pipeline:glance-api-keystone+caching',
                       'pipeline:glance-api-keystone+cachemanagement'],
            'ceilometer': ['pipeline:main'],
            'neutron': ['composite:neutronapi_v2_0'],
            'cinder': ['composite:openstack_volume_api_v1',
                       'composite:openstack_volume_api_v2'],
            'validator': ['composite:ttv_validation_api_v1'],
            'clerk': ['composite:clerkapiv1']}

SECTION_KEYS = {'nova': ['keystone', 'keystone_nolimit'],
                'glance': ['pipeline'],
                'ceilometer': ['pipeline'],
                'cinder': ['keystone', 'keystone_nolimit'],
                'neutron': ['keystone'],
                'validator': ['keystone'],
                'clerk': ['keystone']}

KEYSTONE_CXT = 'keystonecontext'
AUTH_TOKEN = 'authtoken'
AUDIT = 'audit'

AUDIT_SECTION = 'filter:audit'
IGNORE_KEY = 'ignore_req_list'

audit_services = ['nova', 'glance', 'ceilometer', 'neutron', 'cinder',
                  'validator']

ignore_ops = ['GET', 'PUT', 'POST', 'DELETE', 'HEAD', 'TRACE',
              'OPTIONS', 'PATCH', 'CONNECT', 'NONE']


def validate_ignore(ignore):
    ignore_error = False
    if ignore is None:
        return
    try:
        # ignore can have multiple values
        ignore_list = ignore.split(',')
        for ig in ignore_list:
            if ig not in ignore_ops:
                ignore_error = True
                break
            if 'NONE' in ignore_list and len(ignore_list) > 1:
                print(_('NONE must be used alone'),
                      file=sys.stderr)
                sys.exit(1)
    except Exception:
        ignore_error = True
    if ignore_error:
        print(_('Invalid value for -i | --ignore option'),
              file=sys.stderr)
        sys.exit(1)


def get_ignorelist(service):
    ret_val = data_utils.get_openstack_config_value(
        SERVICE_CONF[service],
        IGNORE_KEY,
        AUDIT_SECTION)
    return 'NONE' if ret_val is None else ret_val


def insert_audit(pipeline_vals, conf_file, section, key):
    try:
        pipeline_vals.insert(pipeline_vals.index(KEYSTONE_CXT) + 1, AUDIT)
    except ValueError:
        pipeline_vals.insert(pipeline_vals.index(AUTH_TOKEN) + 1, AUDIT)
    data_utils.set_openstack_config_value(conf_file, key,
                                          ' '.join(pipeline_vals),
                                          section)
    return True


def remove_audit(pipeline_vals, conf_file, section, key):
    pipeline_vals.remove(AUDIT)
    data_utils.set_openstack_config_value(conf_file, key,
                                          ' '.join(pipeline_vals),
                                          section)
    return True


def print_status(status, service_name):
    error = False
    value = None
    section_keys = SECTION_KEYS[service_name]
    for s in SECTIONS[service_name]:
        for k in section_keys:
            temp_value = status.get(s + ':' + k, None)
            if temp_value is not None:
                if value is None:
                    # init with first status we find
                    value = temp_value
                elif value != temp_value:
                    # all values must be in synch
                    error = True
    if value is None:
        error = True

    if error:
        print(_('Error: %s configuration is in a bad state')
              % service_map[service_name],
              file=sys.stderr)
        return

    if value:
        state = 'ENABLED'
    else:
        state = 'DISABLED'

    ignore_list = get_ignorelist(service_name)

    print(_('%(component)s auditing is %(enabled_or_disabled)s' +
            ' with ignore list "%(ignore_list)s"') %
          {'component': service_map[service_name],
           'enabled_or_disabled': state, 'ignore_list': ignore_list})
    print


def handle_audit(args):
    opt_set = None
    if args.enable:
        opt_set = 'on'
    elif args.disable:
        opt_set = 'off'
    elif ((args.ignore is None) and (args.restart is False)):
        opt_set = 'status'
    elif ((args.ignore is None) and (args.restart)):
        print(_('--restart option cannot be used alone'),
              file=sys.stderr)
        sys.exit(1)
    if data_utils.get_edition() == "cloud" and ("clerk" not in audit_services):
        audit_services.append('clerk')
    tgt_services = [args.service]
    if args.service is 'general':
        tgt_services = audit_services
    no_op = True
    for t_service in tgt_services:
        no_op = not (handle_config_for_service(opt_set, args.ignore,
                                               args.restart, t_service)
                     or not no_op)
    if no_op:
        print(_('Nothing to do'))


def handle_config_for_service(opt_set, ignore, restart, service_name):
    validate_ignore(ignore)
    config = False
    conf_file = SERVICE_CONF[service_name]
    sections = SECTIONS[service_name]
    section_keys = SECTION_KEYS[service_name]
    status = {}
    for section in sections:
        for key in section_keys:
            pipeline = data_utils.get_openstack_config_value(
                conf_file,
                key,
                section)
            pipeline_vals = [] if pipeline is None else pipeline.split(' ')
            audit_status = 'audit' in pipeline_vals
            if opt_set == 'on':
                if pipeline is not None and not audit_status:
                    config = insert_audit(pipeline_vals,
                                          conf_file, section, key)
            elif opt_set == 'off':
                if pipeline is not None and audit_status:
                    config = remove_audit(pipeline_vals,
                                          conf_file, section, key)
            elif opt_set == 'status':
                if pipeline is not None:
                    status[section + ':' + key] = audit_status

    if opt_set == 'status':
        print_status(status, service_name)
        return True
    config = set_ignore(ignore, conf_file) or config
    if config and restart:
        # Haven't used data_mgr utility method for service restart as it
        # causes circular dependencies
        print(_('Restarting %s services') % service_map[service_name])
        restart_cmd = ('/opt/ibm/powervc/bin/powervc-services %s restart' %
                       service_name)
        data_utils.run_command(restart_cmd)
    return config


def set_ignore(ignore, conf_file):
    config = False
    # if user has specified the ignore list, update if auditing is on
    if ignore is not None:
        curr_ignore = data_utils.get_openstack_config_value(
            conf_file,
            IGNORE_KEY,
            AUDIT_SECTION)
        if ignore == 'NONE':
            if curr_ignore is not None:
                config = True
                data_utils.del_openstack_config_value(
                    conf_file,
                    IGNORE_KEY,
                    AUDIT_SECTION)
        else:
            if curr_ignore is None or curr_ignore != ignore:
                config = True
                data_utils.set_openstack_config_value(conf_file, IGNORE_KEY,
                                                      ignore, AUDIT_SECTION)

    return config


def build_audit_parser(comp_parser, service_name):
    if service_name is 'general':
        help_msg = _('Configure the audit settings for all PowerVC services')
        desc_msg = _('Run this sub-command without any option' +
                     ' to view the current audit setting for ' +
                     'all PowerVC services')
        enable_msg = _('Used to enable auditing for all PowerVC services')
        disable_msg = _('Used to disable auditing for all PowerVC services')
        restart_msg = _('Use this flag to restart all PowerVC services. ' +
                        'A restart is required for changes to take effect.')
    else:
        component = service_map[service_name]
        help_msg = _('Configure the audit settings for the %s ' +
                     'service. ') % component
        desc_msg = _('Run this sub-command without any option' +
                     ' to view the current audit setting for ' +
                     'this service')
        enable_msg = (_('Used to enable auditing for the %s service')
                      % component)
        disable_msg = (_('Used to disable auditing for the %s service')
                       % component)
        restart_msg = (_('Use this flag to restart the %s service. ' +
                         'A restart is required for changes to take effect.')
                       % component)

    audit_parser = comp_parser.add_parser(
        'audit',
        help=help_msg,
        description=desc_msg)

    group = audit_parser.add_mutually_exclusive_group()
    group.add_argument('--enable',
                       help=(enable_msg),
                       action="store_true")
    group.add_argument('--disable',
                       help=(disable_msg),
                       action="store_true")
    audit_parser.add_argument(
        '-i',
        '--ignore',
        help=_('Will not audit the specified operations. ' +
               'Supported operations are GET, PUT, POST, DELETE, ' +
               'HEAD, TRACE, OPTIONS, PATCH, CONNECT. Separate ' +
               'multiple selections with commas. Specify NONE ' +
               'to ignore no operations.'),
        metavar='<operations>')
    audit_parser.add_argument('--restart',
                              help=(restart_msg),
                              action="store_true")
    audit_parser.set_defaults(service=service_name)
    audit_parser.set_defaults(func=handle_audit)
    return comp_parser
