# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import math
import datetime
import re
import socket
import sys
from oslo_utils import encodeutils
from ipaddress import ip_address as ipaddr_utils
from powervc_oslo import _
from powervc_oslo.config import data_utils as utils
from powervc_oslo.config import audit_config
from powervc_oslo.config import debug_config

CURRENT = 'current'
DEFAULT = 'default'
SEC = 'sec'
MINUTES = 'min'
HOUR = 'hr'
DISPLAY = 'DISPLAY'
SAVE = 'SAVE'


def strip_trailing_zero(val):
    """Remove the trailing zero from the given value and return it.
    :param val: the value to strip the trailing zero from."""
    new_val = str(val)
    if new_val.endswith('.0'):
        new_val = new_val[:-2]
    return new_val


def change_openstack_timeout(args, fname, pattern, name,
                             default_timeout,
                             default_unit='min'):
    """Change the timeout values in the openstack config file provided.
    :params args: the command line argparse.Namespace object.
    :params fnam: filename of the file to read/modify.
    :params: pattern: the regex pattern to match against
    :name: the name of the timeout warning to change.
    :default_timeout: the default amount of time before timeout
    :default_unit: unit of time (sec, hr, min) default of 'min'"""
    unit = default_unit
    if args.unit:
        unit = args.unit
    if args.set:
        stp_val = strip_trailing_zero(str(args.set))
        input_val = convert_time(stp_val, unit, default_unit, SAVE)
        if not check_max_time(input_val, default_unit):
            safe_print(_('Error : Value is too large.'
                         ' Enter a value less than %s'
                         % datetime.timedelta.max))
            sys.exit(1)
        new_pattern = pattern.replace('.*?', str(input_val))
        contents = ''
        r = re.compile(pattern)
        with open(fname) as f:
            contents = f.read()
        contents = r.sub(new_pattern, contents)
        with open(fname, 'w') as f:
            f.write(contents)
            safe_print(_('Setting %s to %s %s') % (name, stp_val, unit))
    else:
        with open(fname) as f:
            contents = f.read()
            r = re.search(pattern, contents)
            if r:
                value = re.split(': ', r.group())
                dis_val = re.split(',', value[1])
                val = convert_time(dis_val[0], default_unit, SEC)
                print_time_output(args, unit, val,
                                  CURRENT, SEC)
                print_time_output(args, unit, default_timeout,
                                  DEFAULT, default_unit)


##############################
# Input Validation Utilities #
##############################


def check_is_negative(val):
    """ Check if the provided value is negative. Return True if negative
    else, return False, or raise a ValueError on invalid input.
    :param val: the value to check if negative."""
    try:
        if int(val) < 0:
            safe_print(_("Error: new value is not a positive number"))
            return True
    except (ValueError, TypeError) as exc:
        raise exc
    return False


def validate_restart_args(args):
    """ Check if args contains the key value --set. If not, return False,
    else return True
    :param args: the arguments to check."""
    if args.set is None:
        if args.restart:
            safe_print(_('Error: --restart only allowed with --set'))
            sys.exit(1)


def check_max_time(val, unit=None):
    """Check that the max time is not large than the float cast value.
    If it is larger than the maximum value, return False, else return True
    :param: val: The time value to check.
    :param unit: The unit type of the time value."""
    try:
        convert_to_iso(unit, float(val))
    except OverflowError:
        return False
    return True

###################
# Print Utilities #
###################


def print_readable_time(val, unit=None, scope=CURRENT):
    new_val = convert_to_correct_type(val)
    display_val = convert_to_iso(unit, new_val)
    print_current_default(display_val, scope)


def print_time_output(args, unit, value, scope=CURRENT, default_unit=SEC):
    if args.unit is None:
        print_readable_time(value, default_unit, scope)
    else:
        cur_val = convert_time(value, default_unit, unit)
        stp_val = strip_trailing_zero(cur_val)
        display_val = '%s %s' % (stp_val, unit)
        print_current_default(display_val, scope)


def print_current_default(value, scope=CURRENT):
    if scope is CURRENT:
        safe_print(_('Current value: %s') % value)
    elif scope is DEFAULT:
        safe_print(_('Default value: %s') % value)


def match_and_print(pattern, value, unit, scope=CURRENT):
    m = re.match(pattern, str(value))
    print_byte_output(m.group(1), m.group(2), unit, scope)


def print_byte_output(val, curr_unit, new_unit, scope=CURRENT):
    new_val = convert_bytes(val, curr_unit, new_unit)
    dis_val = '%s %s' % (new_val, new_unit)
    print_current_default(dis_val, scope)


def safe_print(value):
    print(encodeutils.safe_encode(value))


######################
# Conversion Methods #
######################


def convert_to_correct_type(val):
    val = str(val)
    try:
        val = int(val)
    except ValueError:
        val = float(val)
    return val


def convert_to_iso(unit, val):
    iso_time = None
    if not unit or unit == SEC:
        iso_time = datetime.timedelta(seconds=val)
    elif unit is MINUTES:
        iso_time = datetime.timedelta(minutes=val)
    return iso_time


def convert_time(val, cur_unit, new_unit, tag=DISPLAY):
    units = [SEC, MINUTES, HOUR]
    val = convert_to_correct_type(val)
    return convert_units(units, val, cur_unit, new_unit, 60, tag)


def convert_bytes(val, cur_unit, new_unit, tag=DISPLAY):
    units = ['B', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB']
    # TODO: we should support float vals
    val = int(val)
    return convert_units(units, val, cur_unit, new_unit, 1024, tag)


def convert_units(units, val, cur_unit, new_unit, power, tag=DISPLAY):
    cur_idx = units.index(cur_unit)
    new_idx = units.index(new_unit)
    decimal = None
    if tag is DISPLAY:
        decimal = '%.2f'
    else:
        decimal = '%.6f'

    if cur_idx == new_idx:
        ret_val = val
    elif cur_idx < new_idx:
        # calculate to 2 decimal places
        ret_val = decimal % (val / math.pow(power, (new_idx - cur_idx)))
        # trim off decimal places if they are all zeroes
        if ret_val.endswith('.00'):
            ret_val = ret_val[:-3]
    else:
        cap_val = val * math.pow(power, (cur_idx - new_idx))
        ret_val = (str('%.2f' % cap_val)).split('.')[0]
    return ret_val


######################
# IP Methods #
######################


def get_management_ip_version():
    """Return the management version stored in the config file."""
    return utils.get_openstack_config_value(utils.REF_CONF,
                                            'install_ip_version')


def validate_ip(ip_address):
    utils.LOG.log(utils.CONSOLE,
                  safe_print(_('ipaddress passed in: %s') % ip_address))
    # check if entered ip is the system ip
    system_ip = utils.get_local_ipaddress()
    if system_ip != ip_address:
        utils.LOG.info(_('System ip %s') % system_ip)
        safe_print(_('System ip does not match the entered ip value.'))
        sys.exit(3)
    try:
        ip_version = _get_ip_version(ip_address)
        utils.LOG.info(_('%s is ipv%s') % (ip_address, ip_version))
    except ValueError as e:
        utils.LOG.exception('Invalid address: %s' % e)
        safe_print(_('Invalid address: %s') % e)
        sys.exit(3)
    if str(get_management_ip_version()) != str(ip_version):
        print(safe_print(_('Current ip version does not match the'
                           ' ip address version that was entered.')))
        sys.exit(3)
    if not is_ipaddress_pingable(ip_address):
        print(int(_('The entered ip address is not pingable.'
                    ' Ensure the ip address is pingable and run '
                    'the command again.')))
        sys.exit(3)


def is_valid_address(ipaddress):
    """see if it is an IP Address"""
    try:
        ip_version = _get_ip_version(ipaddress)
        utils.LOG.info(_('%s is ipv%s') % (ipaddress, ip_version))
    except ValueError as e:
        utils.LOG.warning(e)
        return False
    return True


def _get_ip_version(ip_address):
    """Helper method to get the version of the IP address"""
    if _is_valid_ipv4_address(ip_address):
        return 4
    if _is_valid_ipv6_address(ip_address):
        return 6
    raise ValueError(
        "'%s' does not appear to be an IPv4 or IPv6 address" % ip_address)


def _is_valid_ipv4_address(ip_address):
    """Helper method to see if the IP Address is IPv4"""
    try:
        socket.inet_aton(ip_address)
        return True
    except socket.error:
        return False


def _is_valid_ipv6_address(ip_address):
    """Helper method to see if the IP Address is IPv6"""
    try:
        socket.inet_pton(socket.AF_INET6, ip_address)
        return True
    except socket.error:
        return False


def is_valid_hostname(hostname):
    """Return True if the hostname is pingable and DNS resolvable."""
    if not is_hostname_pingable(hostname):
        safe_print((_('The hostname is not pingable.'
                      ' Ensure the hostname is pingable and run'
                      ' the command again.')))
        sys.exit(3)

    if is_resolvable(hostname):
        return True
    else:
        return False


def validate_hostname(hostname):
    """Verify whether the supplied hostname is pingable and DNS resolvable"""
    if not is_hostname_pingable(hostname):
        safe_print(_('The entered hostname is not pingable.'
                     ' Ensure the hostname is pingable and run'
                     ' the command again.'))
        sys.exit(3)
    current_mgmt_ip = utils.get_management_ip().strip('[]')
    # if mgmt_ip is not a ip address must be hostname
    # so it needs to be dns resolvable
    if not is_valid_address(current_mgmt_ip):
        if not is_resolvable(hostname):
            safe_print(_('The hostname is not DNS resolvable.'
                         ' Ensure the hostname is resolvable and'
                         ' run the command again.'))
            sys.exit(3)


def get_management_config_type():
    """Helper method to return whether the management system uses the IP
    address or hostname."""
    config = utils.get_config_data()
    mgmtip = config.get('mgmt_ip')
    hostname = config.get('host_name')
    if mgmtip:
        if utils._is_ipaddress(mgmtip):
            return mgmtip
        else:
            return hostname


def sed_replace_global(afile, before, after):
    """Wrapper method to sed to replace a substitution pattern"""
    before = before.replace('[', '\[').replace(']', '\]')
    after = after.replace('[', '\[').replace(']', '\]')
    utils.run_command("/bin/sed -i -e 's/%s/%s/g' %s" % (before, after, afile))


def is_ipaddress_pingable(ip_address):
    """Ping the provided IP address. Return False if the IP address is
     pingable, False otherwise.
    :param ip_address: the IP address to be pinged."""
    ping_cmd = '/bin/ping'
    try:
        ip_addr_obj = ipaddr_utils(u'%s' % ip_address)
        if ip_addr_obj.version == 6:
            ping_cmd = '/bin/ping6'
        status, output, stderr = utils.\
            run_command_output('%s -c 4 %s' %
                               (ping_cmd, ip_address), False)
        if status != 0:
            utils.LOG.error(_('Unable to ping %s.') % ip_address)
            return False
    except Exception as exc:
        utils.LOG.error(exc)
    return True


def is_hostname_pingable(hostname):
    """Determine if the hostname is pingable. If not, return False, otherwise
    return True.
    :param hostname: the hostname to check."""
    try:
        ping_cmd = '/bin/ping'
        output = utils.run_command_output('%s -c 4 %s' %
                                          (ping_cmd, hostname), False)
        if 'unknown host' in output[2].strip():
            ping_cmd = '/bin/ping6'
            output = utils.run_command_output('%s -c 4 %s' %
                                              (ping_cmd, hostname), False)
        if 'received' not in output[1].strip() or \
           'errors' in output[1].strip():
            utils.LOG.error(_('Unable to ping %s.') % hostname)
            return False
    except Exception as exc:
        utils.LOG.error(exc)
    return True


def is_resolvable(hostname):
    """Check if hostname is dns resolvable"""
    nslookup_cmd = '/usr/bin/nslookup'
    if str(get_management_ip_version()) == '6':
        nslookup_cmd = nslookup_cmd + ' -query=AAAA'
    output = utils.run_command_output('%s %s' %
                                      (nslookup_cmd, hostname), False)
    # if not resolved, return false
    if output[0] != 0:
        return False
    return True


def hostname_resolves_to_ipaddress(ip_address):
    """Check that the current hostname resolves to the given IP address
    :params ip_address: IP address to verify current hostname resolves to
    """
    try:
        hostname_cmd = 'host %s' % utils.get_local_hostname()
        output = utils.run_command_output(hostname_cmd, True)
        output = output[1].split()[-1]
        if ip_address == output:
            return True
        else:
            utils.LOG.warning('Hostname does not resolve to provided IP'
                              ' address: %s' % ip_address)
            return False
    except Exception as exc:
        utils.LOG.exception(exc)


def create_common_parser(subparser, service, component, desc,
                         auditable=True, debuggable=True):
    if component:
        comp_parser = subparser.add_parser(
            component,
            help=desc)
        comp_subp = comp_parser.add_subparsers(
            title='%s sub-commands' %
            component.title(),
            help=(_('Run powervc-config %s <sub-command> -h for help '
                    'on a specific setting') % component))

        # audit settings
        if auditable:
            audit_config.build_audit_parser(comp_subp, service)
        # debug settings
        if debuggable:
            debug_config.build_debug_parser(comp_subp, service)
    return comp_subp
