# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import logging
import collections
from datetime import datetime
from powervc_oslo.config import audit_config
from powervc_oslo.config import data_utils as utils
from powervc_oslo.config import config_utils as cutils
from powervc_oslo.config.data_utils import V120, V121, V122, V123
from powervc_oslo.config.data_utils import V130, V1301, V131, V1311
from powervc_oslo.config.data_utils import CONSOLE
from powervc_oslo import _

_ALL_MGRS = collections.OrderedDict()
BACKUP_FILE = 'powervc_backup.tar.gz'
PLATFORM_FILE = '/tmp/powervc_platform_info.dat'
CHECKSUM_FILE = '/tmp/powervc_cert_checksum.dat'
BACKUP_LOCATION = '/var/opt/ibm/powervc/backups'
ENCRYPT_FILE = '/opt/ibm/powervc/data/powervc-keystone.conf'
RECOVERY_FILE = '/opt/ibm/powervc/data/restore-recovery.dat'

SERVICE_PORTS = dict(nova='8774', cinder='9000', glance='9292', neutron='9696',
                     keystone='5000', validator='8428', ceilometer='8777',
                     swift='8080')
SERVICE_TYPES = dict(keystone='identity', glance='image', nova='compute',
                     cinder='volume', neutron='network', validator='ttv',
                     ceilometer='metering', swift='object-store')

LOG = logging.getLogger(__name__)


######################################################
#          Base Data Manager Definition              #
######################################################
class BaseDataManager(object):

    def get_component_name(self):
        """Provides the name of the Component providing the Data Manager"""
        raise NotImplementedError()

    def get_external_name(self):
        """Provides the external name of the component """
        raise NotImplementedError()

    def get_config_desc(self):
        """Provides the description of the component configuration"""
        raise NotImplementedError()

    def is_configurable(self):
        """Provides info on whether config changes are allowed"""
        return False

    def is_auditable(self):
        """Is audit support is available for this component"""
        return False

    def is_debuggable(self):
        """Is debug config support available for this component"""
        return False

    def is_supported_environment(self):
        """Specifies whether the Component is supported in this Environment"""
        raise NotImplementedError()

    def get_dependent_managers(self):
        """Specifies the Component Name of the Managers it is dependent on"""
        return []

    ######################################################
    #     Data Manager - powervc-config cmd parser       #
    ######################################################
    def invoke_config_parser(self, subparser):
        """Invokes the parser for each component for powervc-config cmd"""
        try:
            if self.is_configurable():
                return cutils.create_common_parser(
                    subparser,
                    self.get_component_name(),
                    self.get_external_name(),
                    self.get_config_desc(),
                    self.is_auditable(),
                    self.is_debuggable())
        except:
            pass

    ###############################################
    #     Data Manager - Configure Interface      #
    ###############################################
    def configure_conf_files(self):
        """Performs the configuration of the Component's CONF Files"""
        for cfile, cdata in self._get_added_conf_properties().iteritems():
            updated_values = dict()
            for sect, data in cdata.iteritems():
                updated_values.setdefault(sect, dict())
                for key, info in data.iteritems():
                    val = self.__config_value_substitution(key, info[0])
                    updated_values[sect][key] = val
            # To improve performance we will do a bulk-update of key/values
            utils.set_openstack_config_values(cfile, updated_values)
        for cfile, cdata in self._get_deleted_conf_properties().iteritems():
            deleted_values = dict()
            for sect, data in cdata.iteritems():
                deleted_values.setdefault(sect, list())
                for key in data.keys():
                    deleted_values[sect].append(key)
            # To improve performance we will do a bulk-delete of keys
            utils.del_openstack_config_values(cfile, deleted_values)

    def unconfigure_conf_files(self):
        """Performs the un-configuration of the Component's CONF Files"""
        for conffile in self._get_added_conf_properties():
            # If we have a saved off original copy, we want to copy that back
            if utils.file_exists(conffile + '.powervc'):
                utils.copy_file(conffile + '.powervc', conffile)

    def set_log_permissions(self):
        """Sets the appropriate permissions on the Component's Log Files"""
        svcuser = self._get_service_user()
        # For now we will only set permissions if it is the Management System
        if utils.is_management_system():
            # Loop through each of the Log Files setting the permissions
            for logfile in self._get_service_log_files():
                logdir = utils.split_path(logfile)[0]
                # If the log directory doesn't exist, create it first
                if not utils.file_exists(logdir):
                    utils.makedirs(logdir, 0o755)
                    utils.chown(logdir, svcuser, 'root')
                if not utils.file_exists(logfile):
                    utils.create_file(logfile)
                utils.chown(logfile, 'root', svcuser)
                utils.chmod(logfile, 0o660)

    def set_user_passwords(self):
        """Sets/Generates the Passwords for the Component's Users"""
        cfgdata = utils.get_config_data()
        dbinfo, svcuser = self._get_database_info(), self._get_service_user()
        # For now we will only set passwords if it is the Management System
        if utils.is_management_system():
            # If they gave us a service user, then set its password
            if svcuser:
                svcatt = '%s_password' % svcuser
                svcpwd = cfgdata.get(svcatt)
                cfgdata[svcatt] = utils.set_password(svcuser, svcpwd)
            # If there is a database user and different than the service user
            if dbinfo and dbinfo[1] != svcuser:
                dbatt = '%s_password' % dbinfo[1]
                # The DB User isn't an OS User, so don't try to set on the OS
                if cfgdata.get(dbatt) is None:
                    cfgdata[dbatt] = utils.generate_password()

    def configure_component(self):
        """Performs Configuration specific to the given Component"""
        self._temporary_openstack_patches()

    def unconfigure_component(self, upgrading):
        """Performs Un-Configuration specific to the given Component"""
        pass

    ###############################################
    #      Data Manager - Upgrade Interface       #
    ###############################################
    def upgrade_conf_files(self, version):
        """Performs the upgrade of the Component's CONF Files"""
        for cfile, cdata in self._get_added_conf_properties().iteritems():
            for sect, data in cdata.iteritems():
                for key, info in data.iteritems():
                    if utils.compare_version(info[1], version) > 0:
                        val = self.__config_value_substitution(key, info[0])
                        utils.set_openstack_config_value(
                            cfile, key, val, sect)
        for cfile, cdata in self._get_deleted_conf_properties().iteritems():
            for sect, data in cdata.iteritems():
                for key, ver in data.iteritems():
                    if utils.compare_version(ver, version) > 0:
                        utils.del_openstack_config_value(cfile, key, sect)
        # If this is the KVM Host, then wipe out all the old orig_ values
        if not utils.is_management_system():
            for cfil in self._get_added_conf_properties():
                utils.run_command("/bin/sed -i -e '/orig_/d' %s" % cfil, False)

    def upgrade_component(self, version):
        """Performs Upgrade specific to the given Component"""
        pass

    ###############################################
    #       Data Manager - Backup Interface       #
    ###############################################
    def get_service_passwords(self):
        """Retrieves the current Service Passwords for the Component"""
        pwds = dict()
        # Loop through each of the Service Password Properties
        for cfgprops in self._get_service_password_properties():
            val = None
            # The section might have moved, so they gave us two
            for sect in cfgprops[2].split('|'):
                if val is None:
                    val = utils.get_openstack_config_value(
                        cfgprops[1], cfgprops[3], sect)
                    # If this a DB URL, we need to parse it out of the URL
                    if val and val.startswith('mysql://'):
                        start, end = val[8:].find(':'), val[8:].find('@')
                        val = val[9 + start: 8 + end]
                    pwds[cfgprops[0] + '_password'] = val
        return pwds

    def create_database(self):
        """Creates the Database for the given Component"""
        if self._get_database_info() is not None:
            dbname = self._get_database_info()[0]
            dbuser = self._get_database_info()[1]
            dbpwd = utils.get_config_data()[dbuser + '_password']
            LOG.log(CONSOLE, _("Creating the %s database...") % dbname)
            return utils.create_database(dbname, dbuser, dbpwd)

    def run_database_sync(self, log=False):
        """Performs the Database Synchronization for the given Component"""
        if log and self._get_dbsync_commands():
            msg = _("Creating tables for the %s database...")
            LOG.log(CONSOLE, msg % self._get_database_info()[0])
        for cmd in self._get_dbsync_commands():
            utils.run_command(cmd, throw=True)

    def prepare_component_restoration(self):
        """Performs any Component-specific preparation needed for restore"""

    def restore_component(self, upgrading):
        """Performs any Component-specific restoration that is needed"""

    ###############################################
    #   Data Manager - System Changed Interface   #
    ###############################################
    def service_passwords_changed(self):
        """Updates the Service Passwords in the CONF files when they change"""
        for cfile, cdata in self._get_added_conf_properties().iteritems():
            for sect, data in cdata.iteritems():
                for key, info in data.iteritems():
                    islist = isinstance(info[0], list)
                    if not islist and info[0].find('_password)s') > 0:
                        val = self.__config_value_substitution(key, info[0])
                        utils.set_openstack_config_value(
                            cfile, key, val, sect)

    def management_ip_changed(self, upgrading):
        """Updates the Management IP in the CONF files when they change"""
        for cfile, cdata in self._get_added_conf_properties().iteritems():
            for sect, data in cdata.iteritems():
                for key, info in data.iteritems():
                    if (not isinstance(info[0], list) and
                       (info[0].find('%(mgmt_ip)s') >= 0 or
                       info[0].find('%(local_ip)s') >= 0)):
                        val = self.__config_value_substitution(key, info[0])
                        utils.set_openstack_config_value(
                            cfile, key, val, sect)

    def management_system_changed(self, upgrading):
        """Performs Component-specific deltas when Management System changes"""

    ###############################################
    #      Data Manager - Auditing Interface      #
    ###############################################
    def merge_audit_values(self, version):
        """Merges in the existing auditing values on a restore"""
        old_audit_ini = self._get_audit_conf_properties(version=version)
        restore_dir = utils.get_config_data()['db_restore_dir'] + '/pwrvctmp'
        # If this component didn't provide auditing values, nothing to do
        if old_audit_ini is None:
            return
        # Determine whether auditing was enabled in the old file
        # Note: We don't support enabling auditing for some pipelines and not
        # for others, so finding one place that auditing is enabled is enough
        set = 'off'  # default
        for sect in old_audit_ini[1]:
            for key in old_audit_ini[2]:
                old_val = utils.get_openstack_config_value(
                    restore_dir + old_audit_ini[0], key, sect)
                # If Auditing was on in the old file, turn it on in the new
                if old_val and old_val.find('audit') >= 0:
                    set = 'on'
                    break
            if set == 'on':
                break
        # Get the ignore list for audit records based on the old file
        ignore_val = utils.get_openstack_config_value(
            restore_dir + old_audit_ini[0], 'ignore_req_list', 'filter:audit')
        # Update the new file accordingly
        if self.get_component_name() in audit_config.SERVICE_CONF:
            audit_config.handle_config_for_service(set, ignore_val, False,
                                                   self.get_component_name())

    ###############################################
    #      Data Manager - Synchronize UUID's      #
    ###############################################
    def uuids_to_synchronize(self, host, vm_uuid):
        """Retrieves the UUID's that need to be Synchronized for the Host"""
        return None

    def synchronize_uuids(self, host, uuid_map):
        """Synchronizes the UUID's that are needed for the given Host"""
        pass

    ###############################################
    #    Data Manager - Purge Database Records    #
    ###############################################
    def purge_deleted_database_records(self, timestamp, chunksize):
        """Purge deleted records from the Component's Database"""
        pass

    ###############################################
    #      Data Manager - Services Interface      #
    ###############################################
    def start_services(self, log_components=False):
        """Starts the Services for the given Component"""
        services = self._get_enabled_services()
        # If they asked us to log each component, do so here
        if log_components and len(services) > 0:
            component = self.get_component_name()
            LOG.log(CONSOLE, _("Starting %s services...") % component)
        # Run the command to start each of the services for the component
        commands = ['/bin/systemctl start %s' % s for s in services]
        return utils.multi_run_command(commands)

    def stop_services(self, log_components=False):
        """Stops the Services for the given Component"""
        services = self._get_enabled_services()[::-1]
        # If they asked us to log each component, do so here
        if log_components and len(services) > 0:
            component = self.get_component_name()
            LOG.log(CONSOLE, _("Stopping %s services...") % component)
        # Run the command to stop each of the services for the component
        commands = ['/bin/systemctl stop %s' % s for s in services]
        return utils.multi_run_command(commands)

    def services_status(self):
        """Status of the Services for the given Component"""
        overall_rc = 0
        # Loop through each of the services, checking to see the status
        for service in self._get_enabled_services():
            command = '/bin/systemctl status %s' % service
            # Call the command first to see whether should log call stack
            rc = utils.run_command(command, False)
            # Run it the second time logging the call stack on failure
            rc = utils.run_command(command + (' -n 0' if rc == 0 else ''))
            overall_rc = rc if rc != 0 else overall_rc
        return overall_rc

    def enable_services(self):
        """Enables the Component's Services for auto-start on reboot"""
        # If they asked to disable any services, move them now
        if self._get_disabled_services():
            fromdir = '/usr/lib/systemd/system/'
            todir = '/usr/share/%s/services/' % self._get_service_user()
            utils.makedirs(todir)
            # Move each of the disabled services to /usr/share
            for svc in self._get_disabled_services():
                utils.move_file(fromdir + svc + '.service',
                                todir + '/' + svc + '.service')
        # Now we can enable any of the services that they asked us to
        services = self._get_enabled_services()
        _x = services.remove('http') if 'http' in services else None
        commands = ['/sbin/chkconfig %s on' % s for s in services]
        return utils.multi_run_command(commands)

    def unenable_services(self):
        """Disables the Component's Services that were enabled"""
        # If they asked to disable any services, move them back now
        if self._get_disabled_services():
            todir = '/usr/lib/systemd/system/'
            fromdir = '/usr/share/%s/services/' % self._get_service_user()
            # Move each of the disabled services back
            for svc in self._get_disabled_services():
                if utils.file_exists(fromdir + '/' + svc + '.service'):
                    utils.move_file(fromdir + '/' + svc + '.service',
                                    todir + svc + '.service')
            utils.remove_directory(fromdir)
        # Now we can disable any of the services that they asked to enable
        services = self._get_enabled_services()
        _x = services.remove('http') if 'http' in services else None
        commands = ['/sbin/chkconfig %s off' % s for s in services]
        return utils.multi_run_command(commands)

    ###############################################
    #     Data Manager - Subclass Definition      #
    ###############################################
    def _get_service_user(self):
        """Provides the name of the Service User for the Component"""
        return None

    def _get_enabled_services(self):
        """Provides a list of the Services for the Component"""
        return []

    def _get_disabled_services(self):
        """Provides a list of the Disabled Services for the Component"""
        return []

    def _get_service_log_files(self):
        """Provides a list of Service Log Files for the Component"""
        return []

    def _get_database_info(self):
        """Provides the Database Name/User Tuple for the Component"""
        return None

    def _get_dbsync_commands(self):
        """Provides a command list for the Component's DB Synchronization"""
        return []

    def _get_added_conf_properties(self):
        """Provides a dictionary of configuration properties to add"""
        return dict()

    def _get_deleted_conf_properties(self):
        """Provides a dictionary of configuration properties to remove"""
        return dict()

    def _get_audit_conf_properties(self, version=None):
        """Provides a tuple of configuration properties for auditing"""
        return None

    def _get_included_backup_files(self):
        """Provides a list of Component files to include in a backup"""
        return []

    def _get_excluded_backup_files(self):
        """Provides a list of Component files to exclude from a backup"""
        return []

    def _create_certificates(self):
        """Creates the Private Keys and Certificates for the Component"""
        ca_info = self._get_certificate_authority_info()
        # Only want to do anything if they provided us a CA to start with
        if ca_info:
            cadir, cafile = utils.split_path(ca_info['cacert'])
            keydir, keyfile = utils.split_path(ca_info['privkey'])
            # Create the Private Key and CA Certificate for the Component
            utils.create_private_key(keydir, keyfile)
            utils.create_ca_certificate(
                ca_info, cadir, cafile, ca_info['privkey'])
            cert_info = self._get_mgmtsys_certificate_info()
            # If there weren't an CSR's provided, nothing more to do
            if not cert_info or not cert_info.get('csrs'):
                return
            # Loop through the signing requests, creating the certificates
            for csr in cert_info['csrs']:
                keydir, keyfile = utils.split_path(csr['keyfile'])
                crtdir, crtfile = utils.split_path(csr['certfile'])
                csrfile = crtfile.replace('pem', 'csr')
                # First we need to create the key/certificate directory
                for adir in [keydir, crtdir]:
                    utils.makedirs(adir, 0o750)
                    utils.chown(adir, 'root', csr.get('group'))
                # Create the Private Key, CSR, Certificate for the Component
                utils.create_private_key(keydir, keyfile, csr.get('group'))
                utils.create_cert_signing_request(
                    ca_info, crtdir, csrfile, csr['keyfile'], csr['exttype'])
                utils.create_signed_certificate(
                    ca_info, crtdir, crtfile, csrfile, csr['exttype'])

    def _get_certificate_authority_info(self):
        """Provides the info to create the Component's CA Certificate"""
        return None

    def _get_mgmtsys_certificate_info(self):
        """Provides the info to create the MgmtSys Signed Certificates"""
        return None

    def _get_remotehost_certificate_info(self):
        """Provides the info to create the Remote Host Signed Certificates"""
        return None

    def _get_service_config_file(self):
        """Provides the configuration file for the given component"""
        return None

    def _get_service_password_properties(self):
        """Provides a list of the configuration properties for passwords"""
        # Note: This only represents the passwords for the users controlled by
        # this service, not all password properties in the components CONF file
        props, conf_file = list(), self._get_service_config_file()
        svcusr, dbinfo = self._get_service_user(), self._get_database_info()
        # If they gave us a configuration file and service user, use them
        if conf_file is not None and svcusr is not None:
            props.append((svcusr, conf_file, 'keystone_authtoken', 'password'))
            # If the database user is different, add that one to the list also
            if dbinfo and dbinfo[1] != svcusr:
                props.append((dbinfo[1], conf_file, 'database', 'connection'))
        return props

    def _temporary_openstack_patches(self):
        """Perform any temporary openstack patching for the Component"""
        pass

    ###############################################
    #     Shared Data Manager Helper Methods      #
    ###############################################
    def _added_auth_conf_properties(self, config):
        """Shared helper method to add the Keystone Auth Config Settings"""
        cat, user = self.get_component_name(), self._get_service_user()
        self.__added_auth_properties(config, cat, user)
        self._added_policy_file(config, cat)

    def __added_auth_properties(self, config, cat, user):
        """Shared helper method to add the Keystone Auth Config Settings"""
        config.setdefault('DEFAULT', dict())
        config.setdefault('keystone_authtoken', dict())
        # The Identity URL uses Mgmt IP on KVM Host and Loopback on Mgmt System
        uri = '%(local_ip)s' if utils.is_management_system() else '%(mgmt_ip)s'
        # Set the Authentication Strategy to be Keystone
        config['DEFAULT']['auth_strategy'] = ('keystone', V120)

        # Add the properties to the Keystone AuthToken Configuration Section
        config['keystone_authtoken'].update(dict(
            check_revocations_for_cached=('True', V122),
            revocation_cache_time=('30', V1311),
            auth_plugin=('v3password', V130),
            auth_uri=('https://%(mgmt_ip)s:5000/v3/', V123),
            auth_url=('https://%s:5000/v3/' % uri, V130),
            username=(user, V130),
            password=('aes:%(' + user + '_password)s', V130),
            project_name=('service', V130),
            project_domain_name=('Service', V130),
            user_domain_name=('Service', V130)))
        # If this is the Management System, we want to bypass Cert Verification
        if utils.is_management_system():
            config['keystone_authtoken']['insecure'] = ('True', V122)

    def _added_old_policy_file(self, config, cat):
        if utils.is_management_system():
            policy_file = '/opt/ibm/powervc/policy/%s/policy.json' % cat
            config['DEFAULT']['policy_file'] = (policy_file, V122)

    def _added_policy_file(self, config, cat):
        if utils.is_management_system():
            policy_file = '/opt/ibm/powervc/policy/%s/policy.json' % cat
            config.setdefault('oslo_policy', dict())
            config['oslo_policy']['policy_file'] = (policy_file, V130)

    def _added_ras_conf_properties(self, config):
        config.setdefault('DEFAULT', dict())
        config['DEFAULT'].update(dict(
            use_stderr=('False', V1301)))

    def _added_messaging_conf_properties(self, config,
                                         extra_thread_pool_size=0):
        """Shared helper method to add the Messaging Config Settings"""
        CERT_PATH, rabbit_sect = '/etc/pki/messages/', 'oslo_messaging_rabbit'
        config.setdefault('DEFAULT', dict())
        config.setdefault(rabbit_sect, dict())
        psize = str(128 + extra_thread_pool_size if
                    utils.is_management_system() else '48')
        # We only want to enable the heartbeat if this is a remote host
        heartbeat = '0' if utils.is_management_system() else '60'
        # Add e Rabbit MQ properties to the Default Configuration Section
        config['DEFAULT'].update(dict(
            rpc_conn_pool_size=('15', V123),
            rpc_response_timeout=('1800', V130),
            notification_driver=('messaging', V122),
            notification_topics=('notifications', V123),
            control_exchange=(self._get_service_user(), V120),
            executor_thread_pool_size=(psize, V130),
            rpc_backend=('rabbit', V122)))
        # Add the Rabbit MQ properties to the oslo_messaging_rabbit Section
        config[rabbit_sect].update(dict(
            rpc_backend=('rabbit', V123),
            rabbit_login_method=('AMQPLAIN', V123),
            heartbeat_timeout_threshold=(heartbeat, V131),
            rabbit_password=('aes:%(mq_password)s', V123),
            rpc_conn_pool_size=('15', V123), rabbit_use_ssl=('True', V123),
            rabbit_userid=('powervc_mq', V123), rabbit_port=('5671', V123)))
        kvm = not utils.is_management_system()
        cacert = '%scacert.pem' % ('' if kvm else 'ca/')
        cert = '%scert.pem' % ('client' if kvm else 'pvcservices/')
        key = '%skey.pem' % \
            ('private/client' if kvm else 'pvcservices/private/')
        config[rabbit_sect]['kombu_ssl_ca_certs'] = (CERT_PATH + cacert, V123)
        config[rabbit_sect]['kombu_ssl_certfile'] = (CERT_PATH + cert, V123)
        config[rabbit_sect]['kombu_ssl_keyfile'] = (CERT_PATH + key, V123)
        # The Identity URL uses Mgmt IP on KVM Host and Loopback on Mgmt System
        uri = '%(local_ip)s' if utils.is_management_system() else '%(mgmt_ip)s'
        config[rabbit_sect]['rabbit_host'] = (uri, V123)
        # If this is a Remote Host we need to set the host property
        if not utils.is_management_system():
            config['DEFAULT']['host'] = ('%(host)s', V121)
            # Since later 2.7 versions of Python need the SSL Version set,
            # we need to make sure that we set it for Remote PowerVM Hosts
            if utils.get_compute_node_type() == 'powervm':
                config[rabbit_sect].update(dict(
                    kombu_ssl_version=('TLSv1_2', V130)))

    def _deleted_auth_conf_properties(self, config):
        self._deleted_auth_properties(config)
        self._deleted_policy_properties(config)

    def _deleted_auth_properties(self, config):
        """Shared helper method to remove the Keystone Auth Config Settings"""
        config.setdefault('DEFAULT', dict())
        config.setdefault('keystone_authtoken', dict())
        config['DEFAULT']['auth_uri'] = V123
        # Delete the properties from Keystone AuthToken Configuration Section
        config['keystone_authtoken'].update(dict(
            admin_uri=V123, auth_port=V123, auth_protocol=V123, auth_host=V123,
            auth_version=V123, auth_admin_prefix=V122, signing_dir=V122,
            identity_uri=V123, admin_user=V123, admin_password=V123,
            admin_tenant_name=V123))

    def _deleted_policy_properties(self, config):
        """Helper method to delete policy file path"""
        config['DEFAULT'].update(dict(policy_file=V130))

    def _deleted_messaging_conf_properties(self, config):
        """Shared helper method to remove the Messaging Config Settings"""
        config.setdefault('DEFAULT', dict())
        # Delete the properties from the Default Configuration Section
        config['DEFAULT'].update(dict(
            rpc_thread_pool_size=V130,
            rabbit_password=V123, rabbit_use_ssl=V123,
            kombu_ssl_certfile=V123, rabbit_host=V123,
            kombu_ssl_keyfile=V123, rabbit_userid=V123,
            kombu_ssl_ca_certs=V123, rabbit_hosts=V123,
            rabbit_login_method=V123, rabbit_port=V123,
            notifier_strategy=V122, qpid_reconnect_limit=V122,
            qpid_sasl_mechanisms=V122, qpid_tcp_nodelay=V122,
            qpid_reconnect_timeout=V122, qpid_heartbeat=V122,
            qpid_notification_topic=V122, qpid_protocol=V122,
            qpid_reconnect_interval=V122, qpid_password=V122,
            qpid_reconnect_interval_max=V122, qpid_port=V122,
            qpid_notification_exchange=V122, qpid_username=V122,
            qpid_reconnect_interval_min=V122, qpid_hostname=V122))

    def __config_value_substitution(self, key, value):
        """Internal Helper method to replace any substitution variables"""
        cfg, encrypt = utils.get_config_data().copy(), False
        # If this is a list property, we need to do substitution on each one
        if isinstance(value, list):
            return [self.__config_value_substitution(key, v) for v in value]
        # If the value is in brackets, they want us to skip substitution
        if len(value) > 2 and value.startswith('<') and value.endswith('>'):
            return value[1:-1]
        # If the value starts with aes: they want use to encrypt the value
        if value.startswith('aes:'):
            value, encrypt = value[4:], True
        # If the Management IP is IPv6 without brackets, add them in
        if cfg['mgmt_ip'].find(':') >= 0 and cfg['mgmt_ip'].find('[') < 0:
            cfg['mgmt_ip'] = '[' + cfg['mgmt_ip'] + ']'
        # If the Local Host IP is IPv6, add the brackets only in certain cases
        if cfg['local_ip'].find(':') >= 0:
            bk = value.find('/') > 0 or key in ['rabbit_host', 'registry_host']
            # If this should have brackets and doesn't, add them in
            if bk and cfg['local_ip'].find('[') < 0:
                cfg['local_ip'] = '[' + cfg['local_ip'] + ']'
            # If this does have brackets and shouldn't, remove them
            elif not bk and cfg['local_ip'].find('[') >= 0:
                cfg['local_ip'] = cfg['local_ip'][1:-1]
        # Use the String Formatter operation to do the substitution
        value = value % cfg
        # If they asked ust to encrypt it then encrypt the value to return
        if encrypt:
            value = utils.encrypt_data(value)
        return value


######################################################
#        Data Management Public Operations           #
######################################################
def get_registered_managers(component=None):
    """Retrieves the registered managers in a dictionary by component name"""
    global _ALL_MGRS
    # We only want to retrieve the Registered Data Managers one time
    if not _ALL_MGRS:
        mgrs = utils.get_registered_extensions('powervc_oslo.config.data_mgr')
        _ALL_MGRS = _order_managers_for_dependencies(mgrs)
    mgrs = collections.OrderedDict(_ALL_MGRS)
    # If they provided a Component Name, we need to filter on that Name
    if component:
        # Special case if it starts with "non" it means all but the given one
        if component == 'noncore':
            mgrs.pop('httpd', mgrs.pop('db', mgrs.pop('rabbitmq', None)))
            mgrs.pop('bumblebee', None)
            return mgrs.values()[:]
        # If the Component isn't registered, then we want to throw an exception
        if component not in mgrs:
            raise SystemExit(_("Invalid component specified: %s") % component)
        return [mgrs.get(component)]
    return mgrs.values()[:]


def configure_data(component=None):
    """Performs that data configuration done as part of install"""
    mgrs = get_registered_managers(component)
    # If a single component, add the Service Passwords for the other components
    if component is not None:
        all_mgrs = get_registered_managers()
        utils.get_config_data().update(_get_service_passwords(all_mgrs))
    # Need to Create the Certificates first before restarting services
    if utils.is_management_system():
        if component is None:
            LOG.log(CONSOLE, _("Creating the PowerVC certificates..."))
        _multi_manager_execute(mgrs, '_create_certificates')
    _multi_manager_execute(mgrs, 'set_user_passwords')
    _multi_manager_execute(mgrs, 'set_log_permissions')
    LOG.log(CONSOLE, _("Updating the configuration files..."))
    _multi_manager_execute(mgrs, 'configure_conf_files')
    # There are certain operations that only apply to the Mgmt System
    if utils.is_management_system():
        _multi_manager_execute(mgrs, 'create_database')
        _multi_manager_execute(mgrs, 'run_database_sync', True)
        _multi_manager_execute(mgrs, 'enable_services')
    _multi_manager_execute(mgrs, 'configure_component')


def unconfigure_data(upgrading=False):
    """Performs that data un-configuration done as part of uninstall"""
    mgrs = get_registered_managers()
    mgrs.reverse()
    _multi_manager_execute(mgrs, 'unconfigure_conf_files')
    # There are certain operations that only apply to the Mgmt System
    if utils.is_management_system():
        _multi_manager_execute(mgrs, 'unenable_services')
    _multi_manager_execute(mgrs, 'unconfigure_component', upgrading)


def upgrade_config(version):
    """Performs the upgrade of the config files to the current version"""
    mgrs = get_registered_managers()
    LOG.log(CONSOLE, _("Updating configuration files from version %s...")
            % version)
    # Add the Service Passwords into the Global Configuration Data
    utils.get_config_data().update(_get_service_passwords(mgrs))
    _multi_manager_execute(mgrs, 'upgrade_conf_files', version)
    return mgrs


def upgrade_data(version):
    """Performs the upgrade of the data needed to the current version"""
    mgrs = upgrade_config(version)
    # There are certain operations that only apply to the Mgmt System
    if utils.is_management_system():
        _multi_manager_execute(mgrs, 'run_database_sync')
    _multi_manager_execute(mgrs, 'upgrade_component', version)


def backup_data(location=None, active=False):
    """Performs the backup of the PowerVC data to the directory provided"""
    mgrs = get_registered_managers()
    mgrs.reverse()
    # Determine the Backup Location and Files/Databases
    location = _get_backup_location(location)
    tar_file = location + '/' + BACKUP_FILE
    include_files, exclude_files = _get_backup_data_files(mgrs)
    # Make sure we have enough Disk Space for the Backup
    utils.check_backup_space(location, include_files)
    # If not backing up active, Stop Services and Enter Maintenance Mode
    if not active:
        utils.enter_maintenance_mode()
        stop_services('noncore')
    try:
        # Now we can actually backup the Database and Data Files
        _backup_data_files(location, include_files, exclude_files)
        LOG.log(CONSOLE, _("Database and file backup completed. "
                           "Backup data is in archive %s") % tar_file)
    finally:
        utils.remove_directory(location + '/pwrvctmp')
        # If not backing up active, Start Services and Exit Maintenance Mode
        if not active:
            start_services('noncore')
            utils.exit_maintenance_mode()
    LOG.log(CONSOLE, _("PowerVC backup completed successfully."))


def restore_data(location=None, upgrading=False):
    """Performs a restore of the PowerVC data from the directory provided"""
    mgrs = get_registered_managers()
    # Determine the Backup file that we should use for the Restore
    tardir, tarname = _get_restore_location(location)
    bencr, binfo, bcheck = utils.extract_tar_file_content(
        tardir, tarname, [ENCRYPT_FILE, PLATFORM_FILE, CHECKSUM_FILE])
    oldver = binfo.split(',')[0]
    # Make sure that the Backup and Restore Systems are Compatible
    _check_platform_compatibility(
        tardir, tarname, binfo, utils.get_platform_info())
    # Make sure we have enough Disk Space for the Restore
    utils.check_restore_space(tardir, tardir + '/' + tarname)
    # Create a temporary directory for the data to be restored to
    timestamp = datetime.now().strftime('%Y%m%d%H%M%S%f')
    temp_dbdir = tardir + '/' + timestamp
    utils.makedirs(temp_dbdir, 0o700)
    # Add the Tar Directory/File to the Global Data for Component's to use
    utils.get_config_data().update(dict(
        restore_dir=tardir, db_restore_dir=temp_dbdir, restore_file=tarname))
    utils.get_config_data().update(dict(restored_version=oldver))
    # If the Upgrade didn't already, Stop Services and Enter Maintenance Mode
    if not upgrading:
        utils.enter_maintenance_mode()
        stop_services('noncore')
    try:
        # Query System Information to determine if restore from other system
        hash1, vers = utils.get_powervc_checksum(), utils.get_powervc_version()
        myip1, pwds1, enckey1 = _get_restore_current_values(mgrs)
        # We also need to let the components do any restore preparation needed
        _multi_manager_execute(mgrs, 'prepare_component_restoration')
        # Now we can actually do the restoration of the Database and Files
        include_files, exclude_files = _get_backup_data_files(mgrs)
        # Parse out the Encryption Key from the Backup since not restoring
        enckey2 = [l.split('=')[1].strip() for l in
                   bencr.split('\n') if l.startswith('pass')][0]
        # Now we can actually to the restore of the Data File and Database
        _restore_data_files(tardir, tarname, include_files, exclude_files)
        # We need to reset the Encryption Key now that we did the restore
        utils.reset_encryption()
        # Update the Global Configuration Data to contain Original IP/Passwords
        utils.get_config_data().update(dict(pwds1, mgmt_ip=myip1))
        # If the Service Passwords are different, need to update in CONF files
        if enckey1 != enckey2 or pwds1 != _get_service_passwords(mgrs):
            service_passwords_changed()
        # Determine first if we are doing an upgrade, so Extensions can defer
        is_upgrade = utils.compare_version(vers, oldver)
        # We also need to let the components do any additional restoration
        _multi_manager_execute(mgrs, 'restore_component', is_upgrade)
        _multi_manager_execute(mgrs, 'merge_audit_values', oldver)
        # If the Management IP is different, need to update in config files
        if myip1 != utils.get_management_ip():
            management_ip_changed(is_upgrade)
        # If the Checksum is different, we need to inform different System
        if hash1 != bcheck:
            management_system_changed(is_upgrade)
        # If the version is different (even a FP), then perform the upgrades
        if is_upgrade:
            upgrade_data(oldver)
        # Clean up the temporary recovery file that we created
        utils.remove_file(RECOVERY_FILE)
    finally:
        utils.remove_directory(temp_dbdir)
        # If not an Upgrade, Start Services and Exit Maintenance Mode
        if not upgrading:
            start_services('noncore')
            utils.exit_maintenance_mode()
    LOG.log(CONSOLE, _("PowerVC restore completed successfully."))


def start_services(component=None, log_components=False):
    """Starts the PowerVC services for the component specified"""
    mgrs = get_registered_managers(component)
    # We want to put the Component Name in the message if a single one
    component = component if component != 'noncore' else None
    msg = _("Starting PowerVC %s services...") % str(component)
    msg = _("Starting PowerVC services...") if not component else msg
    # Only log here if we aren't going to log for each component separately
    if not log_components:
        LOG.log(CONSOLE, msg)
    return _multi_manager_execute(mgrs, 'start_services', log_components)


def stop_services(component=None, log_components=False):
    """Stops the PowerVC services for the component specified"""
    mgrs = get_registered_managers(component)
    mgrs.reverse()
    # We want to put the Component Name in the message if a single one
    component = component if component != 'noncore' else None
    msg = _("Stopping PowerVC %s services...") % str(component)
    msg = _("Stopping PowerVC services...") if not component else msg
    # Only log here if we aren't going to log for each component separately
    if not log_components:
        LOG.log(CONSOLE, msg)
    return _multi_manager_execute(mgrs, 'stop_services', log_components)


def restart_services(component=None, log_components=False):
    """Restarts the PowerVC services for the component specified"""
    stop_services(component, log_components)
    return start_services(component, log_components)


def services_status(component=None):
    """Displays Status of the PowerVC services for the component specified"""
    mgrs = get_registered_managers(component)
    mgrs.reverse()
    return _multi_manager_execute(mgrs, 'services_status')


def service_passwords_changed():
    """Updates the Service Passwords in the Configuration files on change"""
    return _multi_manager_execute(
        get_registered_managers(), 'service_passwords_changed')


def management_ip_changed(upgrading=False):
    """Updates the Management IP in the Configuration files on change"""
    config, mgrs = utils.get_config_data(), get_registered_managers()
    mgmt_ip, mq_pwd = config.get('mgmt_ip'), config.get('mq_password')
    # If the passwords haven't been loaded yet, load them in the config
    if mq_pwd is None:
        config.update(_get_service_passwords(mgrs))
    LOG.log(CONSOLE, _("Updating data to use management "
                       "address of %s...") % mgmt_ip)
    return _multi_manager_execute(
        get_registered_managers(), 'management_ip_changed', upgrading)


def management_system_changed(upgrading=False):
    """Performs Component-specific deltas when Management System changes"""
    LOG.log(CONSOLE, _("Updating data to reflect the "
                       "management system has changed..."))
    return _multi_manager_execute(
        get_registered_managers(), 'management_system_changed', upgrading)


######################################################
#          Data Management Helper Methods            #
######################################################
def _get_service_passwords(mgrs):
    """Retrieves System Information used to determine system changes"""
    pwds = dict()
    # Retrieve the Service Passwords from each of the Data Managers
    for mgr in mgrs:
        pwds.update(mgr.get_service_passwords())
    return pwds


def _order_managers_for_dependencies(mgrs):
    """Orders the Managers based on which are dependent on each other"""
    ordered_map = collections.OrderedDict()
    mgr_map = dict([(mgr.get_component_name(),
                     mgr) for mgr in mgrs if mgr.is_supported_environment()])
    # Loop through each of the managers, adding it and its dependecies
    for mgr in mgr_map.values():
        _add_manager_and_dependencies(mgr_map, ordered_map, mgr)
    return ordered_map


def _add_manager_and_dependencies(mgr_map, ordered_map, manager):
    """Adds the manager and its dependencies to the Ordered Map"""
    # First try to add any of its dependencies to the ordered map first
    for comp in manager.get_dependent_managers():
        # Add the Dependent Manager if it is registered on this system
        if mgr_map.get(comp) is not None:
            _add_manager_and_dependencies(mgr_map, ordered_map, mgr_map[comp])
    # Then we can add this manager to the ordered map if it isn't there already
    if manager.get_component_name() not in ordered_map:
        ordered_map[manager.get_component_name()] = manager


def _multi_manager_execute(managers, method, *parms):
    """Helper method to execute a method on multiple managers"""
    multi_rc = 0
    # Loop through each Manager executing the method on each of them
    for mgr in managers:
        rc = getattr(mgr, method)(*parms)
        # If the return code is not zero, return this calls return code
        multi_rc = rc if rc else multi_rc
    return multi_rc


def _get_backup_location(location):
    """Helper method to generate and create the backup location"""
    location = location if location else BACKUP_LOCATION
    timestamp = datetime.now().strftime('%Y%m%d%H%M%S%f')
    temp_location = location + '/' + timestamp + '/pwrvctmp'
    # Need to create the directories/ownership for the backup location
    utils.makedirs(temp_location)
    return location + '/' + timestamp


def _get_restore_location(location):
    """Helper method to generate and create the backup location"""
    location = location if location else BACKUP_LOCATION
    tar_file = utils.find_latest_file(location, BACKUP_FILE)
    # If we couldn't find a tar file to use on the restore, error out
    if not tar_file:
        msg = _("Could not find a backup archive in %s.")
        raise SystemExit(msg % location)
    index = tar_file.rfind('/')
    # Print out for the user what archive is being used for the restore
    LOG.log(CONSOLE, _("Using archive %s for the restore.") % tar_file)
    # Break the directory and the file name apart to return
    return tar_file[:index], tar_file[index + 1:]


def _get_restore_current_values(mgrs):
    """Helper method to get the current IP/Password Values for the Restore"""
    # If we had a fail restore, then use the values from the recovery file
    if utils.file_exists(RECOVERY_FILE):
        cfg = utils.read_config_file(RECOVERY_FILE)['DEFAULT']
        myip, enckey = cfg.pop('management_ip'), cfg.pop('encryption_key')
        # We need to use the file's encryption key to decode the passwords
        utils.reset_encryption(enckey)
        pwds = dict([(k, utils.decrypt_data(v)) for k, v in cfg.iteritems()])
        utils.reset_encryption()
        return myip, pwds, enckey
    # Otherwise get the values live off the system since a normal restore
    myip, pwds = utils.get_management_ip(), _get_service_passwords(mgrs)
    enckey = utils.get_openstack_config_value(ENCRYPT_FILE, 'passphrase')
    config = dict([(k, utils.encrypt_data(v)) for k, v in pwds.iteritems()])
    config['management_ip'], config['encryption_key'] = myip, enckey
    utils.write_config_file(RECOVERY_FILE, config)
    utils.chmod(RECOVERY_FILE, 0o600)
    return myip, pwds, enckey


def _get_backup_data_files(managers):
    """Helper method to get a list of the data files to backup"""
    include_files, exclude_files = list(), list()
    # Loop through each of the managers, returning a list of data files
    for mgr in managers:
        include_files.extend(mgr._get_included_backup_files())
    # Loop through each of the managers, returning a list of data files
    for mgr in managers:
        exclude_files.extend(mgr._get_excluded_backup_files())
    # We want to also add a few core files to the overall exclude list
    exclude_files.extend(
        ['policy.json', '.yaml', 'sample.conf', '.template', 'audit_map.conf',
         'openstack-', 'rootwrap.conf', 'rootwrap.d', 'release', '.sample'])
    exclude_files.append(RECOVERY_FILE)
    return include_files, exclude_files


def _backup_data_files(location, include_files, exclude_files):
    """Helper method to backup the data files into the given location"""
    LOG.log(CONSOLE, _("Backing up the databases and data files..."))
    # Write out both the Platform info and Certificate Checksum files
    utils.write_to_file(PLATFORM_FILE, utils.get_platform_info())
    utils.write_to_file(CHECKSUM_FILE, utils.get_powervc_checksum())
    try:
        # We need to turn the lists into an expanded include list
        files = _build_backup_file_list(include_files, exclude_files)
        files.extend([CHECKSUM_FILE, PLATFORM_FILE])
        # Create a tar file with the files (including DB/Checksum/Platform)
        utils.create_tar_file(location, BACKUP_FILE, files)
    finally:
        # Need to do cleanup no matter if the backup worked or failed
        utils.remove_file(CHECKSUM_FILE)
        utils.remove_file(PLATFORM_FILE)


def _restore_data_files(location, tar_name, include_files, exclude_files):
    """Helper method to restore the data files from the given location"""
    LOG.log(CONSOLE, _("Restoring the databases and data files..."))
    # Now we can do the actual restore of the file contents
    utils.extract_tar_file_contents(
        location, tar_name, include_files, exclude_files)
    # Need to restart the database to pickup the new files restored
    for amgr in get_registered_managers():
        if amgr.get_component_name() == 'db':
            amgr.stop_services()
            amgr.start_services()


def _build_backup_file_list(include_files, exclude_files):
    """Helper method to build an expanded list of backup files to use"""
    expanded_files, final_list = list(), dict()
    # First go through and expand the include files to a full list
    for include_file in include_files:
        expanded_files.extend(utils.expand_file_list(include_file))
    # Convert the files to a dictionary to form a unique list to use
    for include_file in expanded_files:
        # We want to skip the file if it is in the overall exclude list
        if not [f for f in exclude_files if include_file.find(f) >= 0]:
            final_list[include_file] = ''
    return final_list.keys()


def _check_platform_compatibility(tardir, tarname, backup_info, restore_info):
    """Verifies that the Backup and Restore System are Compatible"""
    infob, infor = backup_info.split(','), restore_info.split(',')
    versb, buildb, archb = infob[0], infob[1], infob[2]
    versr, buildr, archr = infor[0], infor[1], infor[2]
    offb, offr = infob[4] if len(infob) >= 5 else None, infor[4]
    edib, edir = infob[5] if len(infob) >= 6 else 'standard', infor[5]
    # Print out in the log file what the backup platform info is
    LOG.info(_("Backup Platform Info: %s") % backup_info)
    # Make sure the architecture is the same on the 2 systems
    if archb != archr:
        msg = _("The backup architecture is %s.  The architecture "
                "must be the same on the backup and restore system.")
        raise SystemExit(msg % archb)
    # Make sure we aren't restoring to an older version system
    if utils.compare_version(versr, versb) < 0:
        msg = _("The backup version %s is newer than this system's "
                "version.  The backup can only be restored to a system "
                "with a version equal or greater than the backup.")
        raise SystemExit(msg % versb)
    # We don't support build-to-build upgrades on a given version
    if versb == versr and buildb != buildr:
        msg = _("The backup build number %s is different than this system's "
                "build number.  The backup can only be restored to a system "
                "with the same build number for a given release.")
        raise SystemExit(msg % buildb)
    # We don't support upgrading from prior to 1.2.3.x in the 1.3.1 release
    if utils.compare_version(versb, '1.2.3.1') < 0:
        msg = _("The backup version %s is older than supported for "
                "restoration.  Upgrade the backup system to at least version "
                "1.2.3.1, take a new backup, and then try the restore again.")
        raise SystemExit(msg % versb)
    # If the backup didn't contain the offering, we need to figure it out
    if not offb:
        kvm_str, nv_file = 'host_type = kvm', '/etc/nova/nova.conf'
        nv_cfg = utils.extract_tar_file_content(tardir, tarname, [nv_file])[0]
        offb = 'powerkvm' if str(nv_cfg).find(kvm_str) > 0 else 'powervm'
    # We don't support restoring to a different offering than the backup
    if offb != offr:
        msg = _("The backup offering is %s.  The offering must "
                "be the same on the backup and restore system.")
        raise SystemExit(msg % offb)
    # We don't support restoring to a different offering than the backup
    if edib != edir:
        msg = _("The backup edition is %s.  The edition must "
                "be the same on the backup and restore system.")
        raise SystemExit(msg % edib)
