# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import os
import re
import grp
import pwd
import sys
import time
import shlex
import random
import shutil
import signal
import socket
import string
import hashlib
import logging
import statvfs
import tarfile
import tempfile
import binascii
import eventlet
import iniparse
import paramiko
import platform
import subprocess
import ConfigParser
from datetime import datetime
from time import sleep
from stevedore import extension
from sqlalchemy import engine as sqlengine

import oslo_messaging as messaging
from oslo_config import cfg
from powervc_oslo import _

REF_CONF = '/etc/nova/nova.conf'
CONSOLE, _CONFIG_DATA, V999 = 21, None, '9.9.9.9'
V120, V121, V122, V123 = '1.2.0.0', '1.2.1.0', '1.2.2.0', '1.2.3.0'
V1232, V130, V1301, V1302 = '1.2.3.2', '1.3.0.0', '1.3.0.1', '1.3.0.2'
V131, V1311, V1312 = '1.3.1.0', '1.3.1.1', '1.3.1.2'

logging.addLevelName(CONSOLE, 'CONSOLE')
LOG = logging.getLogger(__name__)


######################################################
#              Command Utility Methods               #
######################################################
def run_command(command, log=True, throw=False, inputd=None):
    """Runs the command specified on the local system, outputs result
     to stdout
    :param command: Command to be run
    :param log: boolean if command should be logged
    :param throw: boolean if exceptions should be thrown
    :param inputd: input to be encoded to utf8
    """
    return run_command_output(command, log, throw, inputd)[0]


def run_command_output(command, log=True, throw=False, inputd=None):
    """Runs the command specified on the local system, output result to
     stdout
    :param command: Command to be run
    :param log: boolean if command should be logged
    :param throw: boolean if exceptions should be thrown
    :param inputd: input to be encoded to utf8"""
    def _subprocess_signal():
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)  # @UndefinedVariable
    # If they provided us a list of parameters, make sure they are in UTF-8
    if isinstance(command, list):
        for index in range(len(command)):
            if isinstance(command[index], unicode):
                command[index] = command[index].encode('utf8')
    # If they provided a string, then we need to convert to parameters
    else:
        command = shlex.split(command.encode('utf8'))
    # Make sure input is in UTF-8
    if isinstance(inputd, unicode):
        inputd = inputd.encode('utf8')
    # Execute the Command using the Sub-Process utility method
    proc = subprocess.Popen(
        command, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
        stderr=subprocess.PIPE, close_fds=True, preexec_fn=_subprocess_signal)
    result = proc.communicate(inputd)
    # If they asked us to log the output, then right out the output
    if log:
        if len(result[0].strip()) > 0:
            output = result[0][:-1] if result[0].endswith('\n') else result[0]
            for line in output.split('\n'):
                LOG.info(line.strip())
        if len(result[1].strip()) > 0:
            output = result[1][:-1] if result[1].endswith('\n') else result[1]
            for line in output.split('\n'):
                LOG.warn(line.strip())
    # If they asked us to throw an exception, throw one if not 0 return code
    if throw and proc.returncode != 0:
        raise Exception(result[1])
    # Return the Process Return Code, Standard Output, and Standard Error
    return (proc.returncode, result[0], result[1])


def multi_run_command(commands, log=True, throw=False):
    """Runs multiple commands and returns the overall return code
    :param commands: List of commands to be run sequentially
    :param log: boolean if command should be logged
    :param throw: boolean if exceptions should be thrown"""
    multi_return_code = 0
    # For each command that is in the list, run each one sequentially
    for command in commands:
        return_code = run_command(command, log, throw)
        # If the return code is not 0, then we will return that one
        if return_code != 0:
            multi_return_code = return_code
    return multi_return_code


def run_ssh_command(ipcreds, command, log=True, throw=False,
                    requires_root=False):
    """Run a command over ssh to the system.
    :param ipcreds: the credentials of the system to run the command on
    :param command: the command to be run
    :param log: boolean if command should be logged
    :param throw: boolean if exceptions should be thrown"""
    try:
        # Get the Connection to the SSH Server on the Host provided
        ssh, proc_input = _get_ssh_connection(ipcreds), None
        # If this command requires SUDO, then tack it onto the front
        # We will pass in the password to STDIN on SUDO if they gave us one
        # If using a Private Key, do non-interactive so no password prompt
        # Also, only do this if it isn't root, since root doesn't need
        if requires_root and ipcreds.get('user') != 'root':
            option = '-S ' if ipcreds.get('passwd') else '-n '
            proc_input = ipcreds['passwd'] if ipcreds.get('passwd') else None
            command = '/usr/bin/sudo ' + option + command
        # Now we can actually execute the SSH Command on the Remote Host
        # Verified the inputs on the Paramiko issue found by Bandit scan
        # Checked inputs for injection
        streams = ssh.exec_command(command)  # nosec
        # If they gave us something for standard input, then write it out
        if proc_input is not None:
            streams[0].write(proc_input + '\n')
            streams[0].flush()
        exit_code = streams[1].channel.recv_exit_status()
        stdout, stderr = streams[1].read(), streams[2].read()
        streams[0].close()
        ssh.close()
        # If they asked us to log the output, then right out the output
        if log:
            if len(stdout.strip()) > 0:
                output = stdout[:-1] if stdout.endswith('\n') else stdout
                for line in output.split('\n'):
                    LOG.info(line.strip())
            if len(stderr.strip()) > 0:
                output = stderr[:-1] if stderr.endswith('\n') else stderr
                for line in output.split('\n'):
                    LOG.warn(line.strip())
        # If they asked us to throw an exception, throw one if not 0 return
        # code
        if throw and exit_code != 0:
            raise Exception(stderr)
        # Return the Process Return Code, Standard Output, and Standard Error
        return (exit_code, stdout, exit_code)
    except Exception as exc:
        if log is True:
            LOG.exception(exc)
        # If they asked to throw an exception, throw the one we got
        if throw is True:
            raise
        return (1, '', unicode(exc))


def sftp_transfer_file(ipcreds, local_file, remote_file, get=False):
    """Transfer a file using SFTP to the specified Remote System
    :param ipcreds: the credentials of the system to connect to
    :param local_file: the name of file to be sent/received by scp
    :param remote_file: the name of the file to be received/sent by scp
    :param get: boolean if file should be sent or fetched
    """
    try:
        # Get the Connection to the SSH Server on the Host provided
        ssh = _get_ssh_connection(ipcreds)
        sftp = ssh.open_sftp()
        # Either do a PUT or GET depending on what they asked
        if get:
            sftp.get(remote_file, local_file)
        else:
            sftp.put(local_file, remote_file)
        # Make sure to close the SSH connection when we are done with it
        (sftp.close(), ssh.close())
        return 0
    except Exception as exc:
        LOG.exception(exc)
        return 1


def sftp_transfer_directory(ipcreds, local_dir, remote_dir, get=False):
    """Transfer a Directory thru SFTP to/from the specified Remote System"""
    try:
        # Get the Connection to the SSH Server on the Host provided
        ssh = _get_ssh_connection(ipcreds)
        sftp = ssh.open_sftp()
        # List the local or remote directory files depending which way
        files = sftp.listdir(remote_dir) if get else os.listdir(local_dir)
        # Loop through the files in the directory that was queried
        for afile in files:
            local_file = local_dir + '/' + afile
            remote_file = remote_dir + '/' + afile
            # Either do a PUT or GET depending on what they asked
            if get:
                sftp.get(remote_file, local_file)
            else:
                sftp.put(local_file, remote_file)
        # Make sure to close the SSH connection when we are done with it
        (sftp.close(), ssh.close())
        return 0
    except Exception as exc:
        LOG.exception(exc)
        return 1


def get_remote_host_credentials(conf_file):
    """Helper method to provide a list of KVM Hosts managed"""
    creds = dict(ip=None, user=None, passwd=None, private_key=None)
    # Read the configuration properties from the configuration file
    config = read_config_file(conf_file).get('DEFAULT', dict())
    # Loop through each of the Keys, seeing if there is a match
    for key1 in config:
        for key2 in creds:
            # The remote host properties are yyy_host_<key>
            if key1.endswith('_host_' + key2):
                val = config[key1]
                creds[key2] = decrypt_data(val) if is_encrypted(val) else val
    return creds


def _get_ssh_connection(ipcreds):
    """Internal helper method to establish an SSH connection to a System
    :param ipcreds: credentials of the system to connect to"""
    ssh = paramiko.SSHClient()
    ssh.load_host_keys('/opt/ibm/powervc/data/known_hosts')
    ssh.set_missing_host_key_policy(paramiko.RejectPolicy())
    # If they didn't give us a file name instead, parse from that file
    ipcreds = ipcreds if isinstance(ipcreds, dict) \
        else get_remote_host_credentials(ipcreds)
    # If they specified a Password, use that to try to connect to the SVC
    if ipcreds.get('passwd'):
        ssh.connect(ipcreds['ip'], port=22, timeout=30,
                    username=ipcreds['user'], password=ipcreds['passwd'])
    # If they specified a Private Key, use that to try to connect to the SVC
    elif ipcreds.get('private_key'):
        privkey = paramiko.RSAKey.from_private_key_file(ipcreds['private_key'])
        ssh.connect(ipcreds['ip'], port=22, timeout=30,
                    username=ipcreds['user'], pkey=privkey)
    return ssh


######################################################
#          Openstack Config Utility Methods          #
######################################################
def get_openstack_config_value(conf_file, key, sect='DEFAULT'):
    """Helper method to get a property from an openstack configuration file"""
    parser = ConfigParser.RawConfigParser()
    parser.read(conf_file)
    # If the configuration file contains the section and key specified return
    # it
    sct = parser._defaults if sect == 'DEFAULT' else parser._sections.get(sect)
    val = sct.get(key) if sct is not None else None
    val = decrypt_data(val) if is_encrypted(val) else val
    # If they asked for the database connection, try each of the old ones
    if val is None and sect == 'database' and key == 'connection':
        val1 = get_openstack_config_value(conf_file, 'sql_connection')
        val2 = get_openstack_config_value(conf_file, 'connection', 'sql')
        return val1 if val1 else val2
    return val


def set_openstack_config_value(conf_file, key, val, sect='DEFAULT'):
    """Helper method to set a property in an openstack configuration file"""
    sedbefore, idx = '%s_[[:digit:]][[:digit:]]*', conf_file.rfind('/')
    # If this is a list property, need to do something special to get all of
    # them in the configuration file, since openstack-config oesn't handle
    if isinstance(val, list):
        # First delete the old values, since we are writing out new ones
        del_openstack_config_value(conf_file, key, sect)
        # Next set temporary indexed properties for each one to make unique
        for index in range(len(val)):
            set_openstack_config_value(
                conf_file, key + '_' + str(index), val[index], sect)
        # Finally we are going to run a sed command to change the keys back
        return sed_replace(conf_file, sedbefore % key, key)
    # If the lock file was erroneously left over, clean it up now
    lock_file = conf_file[:idx + 1] + '.' + conf_file[idx + 1:] + \
        '.crudini.lck'
    _x = remove_file(lock_file) if file_exists(lock_file) else None
    # If this is a normal string property than we can set it as usual
    cmd = ['/usr/bin/openstack-config', '--set', '--inplace']
    return run_command(cmd + [conf_file, sect, key, val])


def del_openstack_config_value(conf_file, key, sect='DEFAULT'):
    """Helper method to delete property from an openstack configuration file"""
    idx = conf_file.rfind('/')
    # If the property isn't in the file then nothing to delete
    if get_openstack_config_value(conf_file, key, sect) is None:
        return
    # If the lock file was erroneously left over, clean it up now
    lock_file = conf_file[:idx + 1] + '.' + conf_file[idx + 1:] + \
        '.crudini.lck'
    _x = remove_file(lock_file) if file_exists(lock_file) else None
    # Now we can actually run the command to delete the configuration value
    cmd = ['/usr/bin/openstack-config', '--del', '--inplace']
    return run_command(cmd + [conf_file, sect, key])


def set_openstack_config_values(conf_file, updated_values):
    """Helper method to set properties in an openstack configuration file"""
    # Note: This method exists for performance reasons to do a bulk update
    parser, list_values = iniparse.ConfigParser(), list()
    # If the file doesn't exist then nothing for us to do here
    if not updated_values or not file_exists(conf_file):
        return
    # Now we can read and parse the configuration file to get current values
    with open(conf_file, 'r') as f:
        parser.readfp(f)
    # Loop through the deleted values, removing them from the configuration
    for sect, data in updated_values.iteritems():
        # If the section doesn't exist yet, we need to add it
        if not parser.has_section(sect):
            parser.add_section(sect)
        # Loop through the key/value pairs, adding them to the section
        for key, val in data.iteritems():
            # If this is a list value, then we can't handle the same way
            if isinstance(val, list):
                list_values.append((sect, key, val))
            else:
                parser.set(sect, key, val)
    # Now we can write out configuration file with the updated values for it
    with open(conf_file, 'w') as f:
        parser.write(f)
    # Now we can deal with any of the list values that are done special
    for val in list_values:
        set_openstack_config_value(conf_file, val[1], val[2], val[0])


def del_openstack_config_values(conf_file, deleted_values):
    """Helper method to delete properties from openstack configuration file"""
    # Note: This method exists for performance reasons to do a bulk update
    parser = iniparse.ConfigParser()
    # If the file doesn't exist then nothing for us to do here
    if not deleted_values or not file_exists(conf_file):
        return
    # Now we can read and parse the configuration file to get current values
    with open(conf_file, 'r') as f:
        parser.readfp(f)
    # Loop through the deleted values, removing them from the configuration
    for sect, keys in deleted_values.iteritems():
        for key in keys:
            if parser.has_section(sect) and key in parser.data[sect]._options:
                parser.remove_option(sect, key)
    # Now we can write out configuration file with the removed keys from it
    with open(conf_file, 'w') as f:
        parser.write(f)


def del_openstack_config_section(conf_file, section):
    """Helper method to delete an entire section from openstack config file"""
    # If the file doesn't exist then nothing for us to do here
    if not section or not file_exists(conf_file):
        return
    # Now we can read and parse the configuration file to get current sections
    parser = iniparse.ConfigParser()
    with open(conf_file, 'r') as f:
        parser.readfp(f)
    # Delete the section
    if parser.remove_section(section):
        # Now we can write out configuration file with the removed section
        with open(conf_file, 'w') as f:
            parser.write(f)


def add_conf_list_value(conf_file, sect, key, newval, beforeval=None):
    """Helper method to insert a value in a list configuration property"""
    val = get_openstack_config_value(conf_file, key, sect)
    # Only can update if the property is in the Configuration File
    if val is not None:
        vals = val.split(' ')
        # If the value already is in the list then nothing to do
        if newval in vals:
            return
        # If they didn't provide a before value, insert at the end
        if beforeval is None:
            vals.append(newval)
        # Otherwise insert before the value that was given to us
        else:
            idx = vals.index(beforeval)
            if idx < 0:
                raise ValueError(_('%s was not found in %s') %
                                 (beforeval, val))
            vals.insert(idx, newval)
        # Now we can update the property in the configuration file
        set_openstack_config_value(conf_file, key, ' '.join(vals), sect)


def del_conf_list_value(conf_file, sect, key, delval):
    """Helper method to remove a value from a list configuration property"""
    val = get_openstack_config_value(conf_file, key, sect)
    # Only can update if the property is in the Configuration File
    if val is not None:
        vals = val.split(' ')
        # To make it easier for the caller, check if the value exists
        if delval in vals:
            vals.remove(delval)
            set_openstack_config_value(conf_file, key, ' '.join(vals), sect)


def del_conf_list_values(conf_file, sect, key, delvals):
    """Helper method to remove multiple values from a list
    configuration property"""
    val = get_openstack_config_value(conf_file, key, sect)
    # Only can update if the property is in the Configuration File
    if val is not None:
        vals = val.split(' ')
        deleted = False
        # To make it easier for the caller, check if the value exists
        for delval in delvals:
            if delval in vals:
                vals.remove(delval)
                deleted = True
        if deleted:
            set_openstack_config_value(conf_file, key, ' '.join(vals), sect)


######################################################
#               File Utility Methods                 #
######################################################
def makedirs(path, mode=0o755):
    """Wrapper method to make a directory to help with automation"""
    if not os.path.exists(path):
        os.makedirs(path, mode)


def chown(path, username=None, groupname=None):
    """Wrapper method to os.chown to help with automation"""
    uid = pwd.getpwnam(username).pw_uid if username else -1
    gid = grp.getgrnam(groupname).gr_gid if groupname else -1
    os.chown(path, uid, gid)


def chmod(path, mode):
    """Wrapper method to os.chmod to help with automation"""
    os.chmod(path, mode)


def copy_file(fromfile, tofile):
    """Wrapper method to copy a file to help with automation"""
    shutil.copy(fromfile, tofile)


def move_file(fromfile, tofile):
    """Wrapper method to copy a file to help with automation"""
    remove_file(tofile)
    shutil.move(fromfile, tofile)


def remove_file(afile):
    """Wrapper method to remove a file to help with automation"""
    if os.path.exists(afile):
        os.remove(afile)


def remove_directory(adir):
    """Wrapper method to remove a directory to help with automation"""
    if os.path.exists(adir):
        shutil.rmtree(adir)


def create_file(filename, overwrite=False):
    """Wrapper method to create a file to help with automation"""
    if overwrite and os.path.isfile(filename):
        os.remove(filename)
    os.close(os.open(filename, os.O_CREAT | os.O_EXCL, 0o600))


def file_exists(afile):
    """Wrapper method to os.path.exists to help with automation"""
    return os.path.exists(afile)


def read_from_file(filename):
    """Shared Helper Method to read the data from a given file"""
    with open(filename, 'r') as f:
        return f.read()


def write_to_file(filename, data):
    """Shared Helper Method to write the data to the given file"""
    with open(filename, 'w') as f:
        f.write(data)


def append_to_file(filename, data):
    """Shared Helper Method to write the data to the given file"""
    with open(filename, 'a') as f:
        f.write(data)


def read_config_file(filename):
    """Shared Helper Method to read from a Configuration file"""
    config = dict(DEFAULT=dict())
    parser = ConfigParser.RawConfigParser()
    parser.optionxform = str
    parser.read(filename)
    # Read each of the values from the default section first
    for key in parser._defaults:
        if not key.startswith('__'):
            config['DEFAULT'][key] = parser._defaults.get(key)
    # Loop through the values of each of the non-default sections
    for sect in parser._sections:
        config.setdefault(sect, dict())
        for key in parser._sections[sect]:
            if not key.startswith('__'):
                config[sect][key] = parser._sections[sect].get(key)
    return config


def write_config_file(filename, config_dict):
    """Shared Helper Method to write to a Configuration file"""
    parser = ConfigParser.RawConfigParser()
    parser.optionxform = str
    # If they didn't provide the sections, we need to convert
    if not isinstance(config_dict.itervalues().next(), dict):
        config_dict = dict(DEFAULT=config_dict)
    # Loop through the Dictionary adding the Configuration Values
    for sect in config_dict.keys():
        if sect != 'DEFAULT' and not parser.has_section(sect):
            parser.add_section(sect)
        for key, val in config_dict[sect].iteritems():
            parser.set(sect, key, val)
    # Now we can write the values to the Configuration File
    with open(filename, 'w') as f:
        parser.write(f)


def sed_replace(afile, before, after):
    """Wrapper method to sed to replace a substitution pattern"""
    filelist = find_matching_files(afile)
    for afile in filelist:
        run_command("/bin/sed -i -e 's/%s/%s/' %s" % (before, after, afile))


def create_unique_temp_directory(adir='/tmp'):
    """Helper method to generate a uniquely-named temporary directory"""
    return tempfile.mkdtemp(prefix='powervc_temp_', dir=adir)


def create_remote_unique_temp_directory(ipcreds):
    """Helper method to generate a uniquely-named remote temporary directory"""
    cmd = "/usr/bin/python -c \"import tempfile; print "\
          "tempfile.mkdtemp(prefix='powervc_temp_', dir='/tmp')\""
    return run_ssh_command(ipcreds, cmd, True, True)[1].strip()


def split_path(path):
    """Helper method to split a path into the path and file name"""
    return os.path.split(path)


def get_matching_files(adir, prefix, suffix):
    """Helper Method to return the list of files matching input"""
    filenames = []
    for filename in os.listdir(adir):
        # We need to exclude the config files for the health services
        if filename == (prefix + 'health' + suffix):
            continue
        # If this file name matches the criteria given then return it
        if filename.startswith(prefix) and filename.endswith(suffix):
            filenames.append(filename)
    return filenames


def find_matching_files(wildcard_path):
    """Helper Method to find the files that match the wildcard path"""
    matching_files = list()
    # If the path isn't a wild card, just return the path as-is
    if wildcard_path.find('*') < 0:
        return [wildcard_path]
    # Next we can break up the path at the wildcard location
    before, after = wildcard_path.split('*')
    before_idx, after_idx = before.rfind('/'), after.find('/')
    path, prefix = before[:before_idx + 1], before[before_idx + 1:]
    suffix = after[:(after_idx if after_idx > 0 else len(after))]
    postpend = after[(after_idx if after_idx > 0 else len(after)):]
    # Now we can loop through the files in the path checking for matches
    for filename in (os.listdir(path) if os.path.exists(path) else []):
        if filename.startswith(prefix) and filename.endswith(suffix):
            # If a sub-directory is involved, see if the file exists there
            if not postpend or os.path.exists(path + filename + postpend):
                matching_files.append(path + filename + postpend)
    return matching_files


def expand_file_list(path):
    """Helper method to expand all the files in a directory"""
    file_list = list()
    # If the directory given doesn't exist then there is nothing to expand
    if not os.path.exists(path):
        return file_list
    # We want to skip lock files because they cause lock contention for backups
    if path.find('#') >= 0 or path.find('.lock') >= 0:
        return file_list
    # If the path is a file rather than directory then just return it
    if not os.path.isdir(path):
        file_list.append(path)
        return file_list
    # Go through each of the files/directories contained traversing down
    for afile in os.listdir(path):
        file_list.extend(expand_file_list(path + '/' + afile))
    return file_list


def find_latest_file(location, filename):
    """Helper method to find the latest file matching file name or posfix"""
    # If the directory specified doesn't exist, nothing to find
    if not os.path.exists(location):
        return None
    # If this is a directory then we need to search it, otherwise return file
    if os.path.isdir(location):
        # If the file name exists directly in the directory then return it
        if os.path.exists(location + '/' + filename):
            return location + '/' + filename
        latest_time, latest_file = datetime(1970, 1, 1), None
        # Loop through each of the sub-directories trying to find a match
        for adir in os.listdir(location):
            fulldir = location + '/' + adir
            try:
                # If this isn't a sub-directory then we want to skip it
                if not os.path.isdir(fulldir):
                    continue
                # If the given file isn't in the sub-directory then skip it
                if not os.path.exists(fulldir + '/' + filename):
                    continue
                # See if this is the latest created sub-directory that matches
                if datetime.strptime(adir, '%Y%m%d%H%M%S%f') > latest_time:
                    latest_file = fulldir + '/' + filename
            # If the sub-directory doesn't match the format, then skip it
            except:
                pass
        return latest_file
    return location


######################################################
#             Database Utility Methods               #
######################################################
def create_database(dbname, dbuser, dbpwd):
    """Wrapper method to execute the needed database creation statements"""
    stmts = list()
    # Add the necessary statements to create the database and user
    stmts.append("CREATE DATABASE %s CHARACTER SET utf8" % dbname)
    for ip in ['localhost', '127.0.0.1', '::1']:
        fulluser = "'%s'@'%s'" % (dbuser, ip)
        stmts.append("CREATE USER %s IDENTIFIED BY '%s'" % (fulluser, dbpwd))
        stmts.append("GRANT ALL PRIVILEGES ON %s.* TO %s" % (dbname, fulluser))
    stmts.append("FLUSH PRIVILEGES")
    # Now we can do the batch execute of the statements provided
    return batch_database_execute(stmts)


def batch_database_execute(statements, database=None):
    """Wrapper method to run multiple SQL statements to the database"""
    import pexpect
    # Since the more statements there are the longer the time it will take and
    # we don't want to wait forever on a timeout and also pexpect slows down
    # with larger buffers we will break the statements up into chunks of 10,000
    if len(statements) > 10000:
        for index in xrange(0, len(statements), 10000):
            batch_database_execute(statements[index: index + 10000], database)
        return 0
    # Create the temporary file with the statements to execute first
    with tempfile.NamedTemporaryFile(prefix='powervc_temp_', dir='/tmp') as f:
        # First we need to write out all the statements to the file
        _x = [f.file.write(stmt + ';\n') for stmt in statements]
        f.file.flush()
        # Use the MySQL command to execute the provided SQL statements
        sqlcmd = '/usr/bin/mysql -u root -p --silent '\
                 '--default-character-set=utf8 --prompt="mariadb => " '
        sqlcmd += database if database else ''
        p = pexpect.spawn(sqlcmd, env={'TERM': 'dumb'})
        # We need to specify the root password to be able to execute statements
        p.expect(['Enter password:'], timeout=20)
        p.sendline(get_config_data()['db_password'])
        # Next at the prompt need to specify the file with the SQL statements
        p.expect(['.*mariadb =\> '], timeout=30)
        p.sendline('source ' + f.name)
        # Send the quit statement at the end to exit out
        p.expect(['.*mariadb =\> '], timeout=600, searchwindowsize=20)
        p.sendline('quit;')
        p.expect(pexpect.EOF, timeout=20)
    return 0


def export_database_data(outfile, database, exclude_tables=None):
    """Helper method to dump out the data from the database for import"""
    import pexpect
    # Create the file first to restrict permissions (export will overwrite)
    create_file(outfile, True)
    # Use the MySQL Dump command to dump out the given database data
    sqlcmd = '/usr/bin/mysqldump -u root -p -t --compact '\
             '--default-character-set=utf8 -r ' + outfile
    # If there are any tables to exclude, add the option in for them
    for table in exclude_tables or []:
        sqlcmd += ' --ignore-table=' + database + '.' + table
    p = pexpect.spawn(sqlcmd + ' ' + database, env={'TERM': 'dumb'})
    # We need to specify the root password to be able to execute statements
    p.expect(['Enter password:'], timeout=10)
    p.sendline(get_config_data()['db_password'])
    p.expect(pexpect.EOF, timeout=30)
    return 0


def drop_database_tables(database):
    """Helper method to dump out the data from the database for import"""
    import pexpect
    stmts = ['SET FOREIGN_KEY_CHECKS = 0']
    # Create the file first to restrict permissions (export will overwrite)
    with tempfile.NamedTemporaryFile(prefix='powervc_temp_', dir='/tmp') as f:
        # Use the MySQL Dump command to dump out the given database data
        sqlcmd = '/usr/bin/mysqldump -u root -p --add-drop-table '
        sqlcmd += '--no-data -r ' + f.name
        p = pexpect.spawn(sqlcmd + ' ' + database, env={'TERM': 'dumb'})
        # We need to specify the root password to be able to execute statements
        p.expect(['Enter password:'], timeout=10)
        p.sendline(get_config_data()['db_password'])
        p.expect(pexpect.EOF, timeout=30)
        # Now we can read the file and get the drop statements
        for line in read_from_file(f.name).split('\n'):
            if line.strip().startswith('DROP TABLE '):
                stmts.append(line.strip().strip(';'))
    # Now we can execute the statements to drop all the tables in the DB
    return batch_database_execute(stmts, database)


def check_backup_space(location, files):
    """Helper method to determine if backup location has enough free space"""
    total_size = 0
    # Loop through each of the data files, calculating its backup size
    for afile in files:
        total_size += int(os.path.getsize(afile))
    # Figure out what the current size is available in the file system
    stats = os.statvfs(location)
    avail_space = stats[statvfs.F_FRSIZE] * stats[statvfs.F_BAVAIL]
    # If there is less available than the total size, then need to fail
    if avail_space < total_size:
        avail_mb = str(avail_space / (1024 ** 2))
        total_mb = str(total_size / (1024 ** 2))
        msg = _("There is not enough disk space available to perform "
                "the backup.  %(total)s MB is required and only "
                "%(avail)s MB is available.")
        raise SystemExit(msg % dict(total=total_mb, avail=avail_mb))


def check_restore_space(location, tar_name):
    """Helper method to determine if enough free space for the restore"""
    total_size = 0
    # Read the members from the tar file to determine the real size of it
    tar = tarfile.open(tar_name, "r:gz")
    # Loop through each of the members, calculating the total file sizes
    for member in tar.getmembers():
        total_size += member.size
    # Figure out what the current size is available in the file system
    stats = os.statvfs(location)
    avail_space = stats[statvfs.F_FRSIZE] * stats[statvfs.F_BAVAIL]
    # If there is less available than the total size, then need to fail
    if avail_space < total_size:
        avail_mb = str(avail_space / (1024 ** 2))
        total_mb = str(total_size / (1024 ** 2))
        msg = _("There is not enough disk space available to perform "
                "the restore.  %(total)s MB is required and only "
                "%(avail)s MB is available.")
        raise SystemExit(msg % dict(total=total_mb, avail=avail_mb))


def get_database_engine(conf_file, key, sect='DEFAULT'):
    """Helper method to get an engine to connect to the database"""
    sql_connection = get_openstack_config_value(conf_file, key, sect)
    return _get_database_engine(sql_connection)


def _get_database_engine(sql_connection):
    """Helper method to get an engine to connect to the database"""
    # Create the SQLAlchemy Engine that wrappers the SQL Connection
    db_args = {"pool_recycle": 3600, "echo": False, 'convert_unicode': True}
    engine = sqlengine.create_engine(sql_connection, **db_args)
    eventlet.greenthread.sleep(0)
    return engine


def wait_for_db_to_start(local_ip=None):
    """Helper method to make sure the database is fully started"""
    sql_connection = get_openstack_config_value(
        '/etc/nova/nova.conf', 'connection', 'database')
    # If this is IPv4 now, change the IPv6 address to IPv4
    if local_ip == '127.0.0.1':
        sql_connection = sql_connection.replace('[::1]', '127.0.0.1')
    # If this is IPv6 now, change the IPv4 address to IPv6
    elif local_ip == '::1':
        sql_connection = sql_connection.replace('127.0.0.1', '[::1]')
    engine = _get_database_engine(sql_connection)
    # Keep trying to connect to the database until it as fully started
    for _idx in range(30):
        try:
            connection = engine.connect()
            connection.close()
            return True
        except:
            LOG.info(_("Trying to establish a database connection..."))
            sleep(1)
    return False


######################################################
#           Miscellaneous Utility Methods            #
######################################################
def get_config_data(initial_config=None):
    """Helper method to get the Configuration Data Properties"""
    global _CONFIG_DATA
    # If it isn't initialized yet, then read it from the file
    if _CONFIG_DATA is None:
        _CONFIG_DATA = initial_config or dict()
        _CONFIG_DATA = _CONFIG_DATA.copy()
        # Now we need to add in any of the default properties
        props = dict(host='host',
                     ip_version='install_ip_version',
                     local_ip='rabbit_host',
                     ip_option='install_ip_option',
                     ip_interface='install_ip_interface')
        _CONFIG_DATA.setdefault('mgmt_ip', get_management_ip())
        # Loop through each of the properties defined
        # and set the vaule in _CONFIG_DATA if the value exists, else
        # set the retrieved default
        for k1, k2 in props.items():
            _CONFIG_DATA.setdefault(
                k1, get_openstack_config_value(REF_CONF, k2))
        # Get the Host Name and IP Address for the Local System
        is_ipv6 = _CONFIG_DATA['ip_version'] == '6'
        if not _CONFIG_DATA.get('local_ip'):
            _CONFIG_DATA['local_ip'] = '::1' if is_ipv6 else '127.0.0.1'
        _CONFIG_DATA.setdefault('host_ip', get_local_ipaddress())
        _CONFIG_DATA.setdefault('host_name', get_local_hostname())
        # If they didn't give us a DB location to use, query it
        if not _CONFIG_DATA.get('db_location'):
            _CONFIG_DATA['db_location'] = get_openstack_config_value(
                '/opt/ibm/powervc/data/powervc-db.conf', 'db_location')
            _CONFIG_DATA['db_password'] = get_openstack_config_value(
                '/opt/ibm/powervc/data/powervc-db.conf', 'db_password')
    return _CONFIG_DATA


def get_offering():
    """Helper method to determine what type of offering this is"""
    # If we haven't cached it yet, query the offering now
    if not get_config_data().get('offering'):
        info = get_platform_info().split(',')
        get_config_data()['offering'] = info[4] if info[4] != 'None' else None
    return get_config_data().get('offering')


def get_edition():
    """Helper method to determine what type of edition this is"""
    # If we haven't cached it yet, query the edition now
    if not get_config_data().get('edition'):
        info = get_platform_info().split(',')
        get_config_data()['edition'] = info[5] if info[5] != 'None' else None
    return get_config_data().get('edition')


def is_management_system():
    """Helper method to determine if this is the Management System"""
    # If we haven't cached it yet, query the offering now
    if not get_config_data().get('is_mgmt_system'):
        parser = ConfigParser.RawConfigParser()
        parser.read('/opt/ibm/powervc/version.properties')
        sections = parser.sections()
        # If the Name contains PowerKVM we know it isn't the MgmtSystem
        if parser.has_option(sections[0], 'name'):
            notkvm = parser.get(sections[0], 'name').find('PowerKVM') < 0
            notneo = parser.get(sections[0], 'name').find('Novalink') < 0
            get_config_data()['is_mgmt_system'] = notkvm and notneo
    return get_config_data().get('is_mgmt_system', True)


def get_compute_node_type():
    """Helper method to determine what type of Compute Node this is"""
    # If this is the Management System then it isn't a Compute Node
    if is_management_system():
        return None
    # If we haven't cached it yet, query the Compute Node type now
    if not get_config_data().get('compute_node_type'):
        etc_issue, compute_type = read_from_file('/etc/issue'), None
        # If this has the pvmctl command we will assume NovaLink
        if file_exists('/usr/bin/pvmctl'):
            compute_type = 'powervm'
        # Otherwise see if the /etc/issue contains PowerKVM
        elif etc_issue.find('PowerKVM') >= 0:
            compute_type = 'powerkvm'
        get_config_data()['compute_node_type'] = compute_type
    return get_config_data().get('compute_node_type')


def get_install_options(option_name=None):
     # If we haven't cached it yet, query the Platform Info now
    if not get_config_data().get('install_options'):
        parser = ConfigParser.RawConfigParser()
        parser.optionxform = str
        parser.read('/opt/ibm/powervc/install/settings.conf')
        sections = parser.sections()
        _x, _y = sections.sort(), sections.reverse()
        firewall_config = parser.get(sections[0], 'firewall_config')
        ip_version = parser.get(sections[0], 'ip_version')
        ip_interface = parser.get(sections[0], 'ip_interface')
        info = {}
        info['firewall_config'] = firewall_config
        info['ip_version'] = ip_version
        info['ip_interface'] = ip_interface
        get_config_data()['install_options'] = info
    if option_name:
        return get_config_data().get('install_options')[option_name]
    return get_config_data().get('install_options')


def get_platform_info():
    """Helper method to get Platform Info for the Backup comparison"""
    offering, edition = None, 'standard'
    # If we haven't cached it yet, query the Platform Info now
    if not get_config_data().get('platform_info'):
        parser = ConfigParser.RawConfigParser()
        parser.optionxform = str
        parser.read('/opt/ibm/powervc/version.properties')
        sections = parser.sections()
        # We will get the version and build from the latest section
        _x, _y = sections.sort(), sections.reverse()
        version = parser.get(sections[0], 'version')
        build = parser.get(sections[0], 'build')
        # Loop through all the sections looking for one with the offering
        for sect in sections:
            if parser.has_option(sect, 'offering'):
                offering = parser.get(sect, 'offering')
                offering = 'powervm' if offering == 'standard' else offering
            if parser.has_option(sect, 'cloud_enabled'):
                enabled = parser.get(sect, 'cloud_enabled')
                edition = 'cloud' if enabled == 'yes' else edition
        # Get the Architecture and Kernel Type from the Platform Utility
        machine, system = platform.machine(), platform.system()
        info = '%s,%s,%s,%s,%s,%s' % \
            (version, build, machine, system, offering, edition)
        get_config_data()['platform_info'] = info
    return get_config_data().get('platform_info')


def get_total_memory():
    """Helper method to get the total memory for the current system"""
    import psutil
    return psutil.virtual_memory().total


def get_management_ip():
    """Helper method to get the PowerVC IP Address on the system"""
    mgmt_ip = get_openstack_config_value(REF_CONF, 'management_ip')
    # If the management_ip property isn't set, fall-back to the auth_uri
    if not mgmt_ip:
        uri = get_openstack_config_value(
            REF_CONF, 'osapi_compute_link_prefix')
        mgmt_ip = uri[8:-5] if uri else None
    return mgmt_ip if mgmt_ip else get_local_hostname()


def get_local_hostname():
    """Helper method to get the Hostname on the local system"""
    return socket.getfqdn()


def get_local_ipaddress():
    """Helper method to get the IP Address on the local system"""
    config = get_config_data()
    ipversion = config.get('ip_version') or '4'
    ipinterface = config.get('ip_interface') or 'eth0'
    # Run the ifconfig command to get the full IP Address information
    output = run_command_output('/sbin/ip addr show ' + ipinterface, False)
    # Depending on if it is IPv4 or IPv6, use a different regular expression
    iprex = re.compile('inet6 [ ]*([0-9a-z:]+)')
    iprex = re.compile('inet [ ]*([0-9\.]+)') if ipversion == '4' else iprex
    # Loop through the ifconfig output to see if there is a matching line
    for line in output[1].split('\n\n'):
        match = iprex.search(line)
        if match:
            return match.group(1).strip()
    return None


def get_management_hostname():
    """Helper method to get the PowerVC hostname"""
    mgmthn = get_openstack_config_value(REF_CONF, 'ego_master_list')
    # If the ego_master_list property isn't set, fall-back to hostname command
    if mgmthn is None:
        mgmthn = get_local_hostname()
    return mgmthn


def get_powervc_version():
    """Helper method to get the PowerVC version on the system"""
    return get_platform_info().split(',')[0]


def get_powervc_checksum():
    """Helper method to generate a checksum of the PowerVC Certificate"""
    try:
        # Read the PowerVC.crt file and generate a checksum hash of it
        with open('/etc/pki/tls/certs/powervc.crt', 'rb') as f:
            return str(hashlib.sha256(f.read()).hexdigest())
    # If we can't query the certificate, nothing to return data for
    except:
        return str(None)


def compare_version(new_version, old_version):
    """Utility to compare 2 versions and see if the first is newer"""
    # If the old system isn't installed, the new system is greater if installed
    if not old_version:
        return 1 if new_version else 0
    # If the new system isn't installed, the old is newer since it must be set
    if not new_version:
        return -1
    # We need to break apart the version to individual digits for comparison
    old_split = old_version.split('.')
    new_split = new_version.split('.')
    max_digits = max(len(new_split), len(old_split))
    # Loop through each of the digits and do the comparison
    for index in range(max_digits):
        old_num = int(old_split[index]) if len(old_split) > index else 0
        new_num = int(new_split[index]) if len(new_split) > index else 0
        # If a digit is different, then return since 1 is greater than the
        # other
        if old_num != new_num:
            return 1 if new_num > old_num else -1
    # If all digits were the same, the 2 versions must be equal
    return 0


def get_property_value(output, key, default=None):
    """Utility to get the property value for a key from a command output"""
    val = default
    # Since the output might have multiple lines, look at each line separate
    for line in output.strip().split('\n'):
        line = line.strip()
        # If this line is commented out than we want to skip it
        if len(line) > 0 and not line.startswith('#'):
            parts = line.split('=')
            # If this is a property line and the value is set, use it
            if len(parts) > 1 and len(parts[1].strip()) > 0:
                if parts[0].strip() == key:
                    val = parts[1].strip()
    return val


def prompt_for_confirmation(warning):
    """Utility to prompt the user on whether they wish to continue"""
    invalid = _("Invalid Response. Please enter y or N:")
    response = raw_input(warning.encode('UTF-8')).lower()
    while True:
        # If they gave a valid response, return True if they yes
        if response in ['y', 'n', '']:
            return response == 'y'
        # If they gave an invalid response, ask them again
        else:
            response = raw_input(invalid.encode('UTF-8')).lower()


def safe_method_call(method, *args, **kwargs):
    """Helper method to make a safe method call without throwing exception"""
    try:
        method(*args, **kwargs)
    # Since this is a safe method call, eat any exception that is thrown
    except:
        pass


def get_registered_extensions(namespace):
    """Utility for querying a list of extensions registered for a name-space"""
    extmgr = extension.ExtensionManager(namespace)
    return [ext.plugin() for ext in extmgr.extensions]


def setup_logging(file_prefix=None, use_console=False):
    """Utility method to enable logging to a file and/or to the console
    :param file_prefix: specify the log filename prefix
    :param use_console: set logging level to INFO if True"""
    class ConsoleFilter(logging.Filter):

        def filter(self, record):
            return record.levelno == CONSOLE
    log_file = file_prefix
    # If they asked us to disable logging, then set it up so no logging
    if file_prefix is None and not use_console:
        logging.root.addHandler(logging.NullHandler())
    # If they specified a Log File, set up a logger for that File
    if file_prefix is not None:
        datefmt = '%Y-%m-%d %H:%M:%S'
        timestamp = datetime.now().strftime('%Y-%m-%d-%H%M%S')
        fmtstr = '%(asctime)s %(levelname)s %(name)s [-] %(message)s'
        # If they just gave us the file prefix, then tack on the timestamp
        if not log_file.endswith('.log'):
            log_file = '%s_%s.log' % (file_prefix, timestamp)
        # If they didn't give us the directory, then use the default one
        if log_file.find('/') < 0:
            log_file = '/opt/ibm/powervc/log/' + log_file
        log_hdlr = logging.FileHandler(log_file)
        log_hdlr.setFormatter(logging.Formatter(fmtstr, datefmt))
        logging.root.addHandler(log_hdlr)
        logging.root.setLevel(logging.INFO)
    # If they asked us to log to the Console, setup that Log Handler
    if use_console:
        log_hdlr = logging.StreamHandler(sys.stdout)
        log_hdlr.setFormatter(logging.Formatter('%(message)s'))
        logging.root.addHandler(log_hdlr)
        logging.root.setLevel(logging.INFO)
        # If they specified a file, then only log CONSOLE level
        if file_prefix is not None:
            log_hdlr.addFilter(ConsoleFilter())
    return log_file


######################################################
#             Tar File Utility Methods               #
######################################################
def create_tar_file(location, tar_name, files):
    """Helper method to create a tar file with the files specified"""
    cmd = '/bin/tar -czf %s -C %s -T %s'
    tar_name = location + '/' + tar_name
    # Write out each file name/patch to the temporary list file
    with tempfile.NamedTemporaryFile(prefix='powervc_temp_', dir='/tmp') as f:
        for afile in files:
            f.file.write(afile + '\n')
        f.file.flush()
        # Run the tar command to create the tar, since the performance of
        # the python tarfile is really reall bad, and can't take that hit
        run_command(cmd % (tar_name, location, f.name), throw=True)


def extract_tar_file_content(location, tar_name, file_names):
    """Helper method to read a single file from the given tar file"""
    tar_name, lines = location + '/' + tar_name, list()
    tar = tarfile.open(tar_name, "r:gz")
    avail_names = tar.getnames()
    # Extract the individual file and read its line from the file
    for fname in file_names:
        fname, line = fname[1:] if fname.startswith('/') else fname, ''
        # We can only extract the contents, if the file is in the tar
        if fname in avail_names:
            file_handle = tar.extractfile(fname)
            line = file_handle.read().strip()
        lines.append(line)
    tar.close()
    return lines


def extract_tar_file_contents(location, tar_name,
                              include_files, exclude_files):
    """Helper method to extract all of the files from a tar file"""
    db_restoredir = get_config_data().get('db_restore_dir', location)
    temp_location = db_restoredir + '/pwrvctmp'
    tar_name = location + '/' + tar_name
    tar = tarfile.open(tar_name, "r:gz")
    # Loop through each file in the tar extracting each individual one
    for afile in tar.getnames():
        path = temp_location if afile.endswith('paste.ini') else '/'
        afile2 = afile if afile.startswith('/') else ('/' + afile)
        # Only extract the file if it is in our new include list
        if [f for f in include_files if afile2.find(f) >= 0]:
            # Only extract the file if it isn't in our new exclude list
            if not [f for f in exclude_files if afile2.find(f) >= 0]:
                tar.extract(afile, path=path)
    tar.close()


def tar_file_contains_file(location, tar_name, file_name):
    """Helper method to read a single file from the given tar file"""
    tar = tarfile.open(location + '/' + tar_name, "r:gz")
    # Check to see if the file name is one of those in the tar file
    match = file_name in tar.getnames()
    tar.close()
    return match


######################################################
#           Certificate Utility Methods              #
######################################################
def create_private_key(keydir, keyfile, group='root'):
    """Generate a Private Key in the directory specified"""
    if keydir.endswith('/private'):
        keydir = keydir[:-8]
    keypath = keydir + '/private/' + keyfile
    if not file_exists(keypath):
        makedirs(keydir, 0o755)
        makedirs(keydir + '/private', 0o750)
        chown(keydir + '/private', None, group)
        run_command('/usr/bin/openssl genrsa -out '
                    '%s 2048' % keypath, False, True)
    chown(keypath, None, group)
    chmod(keypath, 0o440)


def create_ca_certificate(ca_info, out_dir, csr_file, key_file):
    """Helper method to a create an X509 CA Certificate"""
    sub_name = '/CN=PowerVC %s CA' % ca_info['component']
    tem_file = _generate_certificate_template(ca_info)
    # If they didn't provide a Component Name, then need to use Hostname
    if ca_info['component'] == '':
        sub_name = '/CN=%s' % get_config_data()['mgmt_ip'].strip('[]')
        #If name too long, switch to hostname instead of FQDN
        if len(sub_name) > 67:
            sub_name = sub_name[0:sub_name.find('.')]
    # Run the Command to generate the Certificate Authority Certificate
    _create_certificate_database(ca_info, tem_file)
    run_command('/usr/bin/openssl req -new -x509 -days 3652 '
                '-key %s -config %s -subj "%s" -out %s/%s'
                % (key_file, tem_file, sub_name, out_dir, csr_file))
    remove_file(tem_file)


def create_signed_certificate(ca_info, out_dir, crt_file, csr_file, ext_type):
    """Helper method to create a signed certificate for the CA"""
    ext_type = ext_type if ext_type != 'signer' else ''
    ext_type = ' -extensions %s_ext' % ext_type if ext_type else ''
    tem_file = _generate_certificate_template(ca_info)
    # If they didn't path-qualify the CSR file, assume in the output directory
    has_path = csr_file.find('/') >= 0
    csr_file = csr_file if has_path else (out_dir + '/' + csr_file)
    # Run the Command to generate the Certificate Signing Request
    run_command('/usr/bin/openssl ca -batch -notext -config %s%s -in %s -out '
                '%s/%s' % (tem_file, ext_type, csr_file, out_dir, crt_file))
    remove_file(tem_file)
    remove_file(csr_file)


def create_cert_signing_request(ca_info, out_dir,
                                csr_file, key_file, ext_type):
    """Create a certificate to sign a request.
    :param ca_info: the certificate_info object
    :param out_dir: the directory to place the certificate
    :param csr_file: used to generate certificates
    :param key_file: the key file used to generate signing request
    :param ext_type: the extension type of the component"""
    sub_name = '/CN=PowerVC %s %s'\
        % (ca_info['component'].title(), ext_type.title())
    tem_file = _generate_certificate_template(ca_info)
    # If this is Libvirt then we need to put the hostname in the Certificate
    if ca_info['component'] == 'libvirt':
        sub_name = '/CN=' + get_local_hostname()
    # Run the Command to generate the Certificate Signing Request
    run_command('/usr/bin/openssl req -new -key %s -config %s -subj "%s" -out '
                '%s/%s' % (key_file, tem_file, sub_name, out_dir, csr_file))
    remove_file(tem_file)


def _generate_certificate_template(ca_info):
    """Helper method to generate the template file based on the info"""
    config = dict()
    component = ca_info.get('component')
    extensions = ca_info.get('extensions', [])
    altnameexts = ca_info.get('altnamexts', [])
    cacert, privkey = ca_info.get('cacert'), ca_info.get('privkey')
    cname, opt = 'Common Name (hostname, IP, or your name)', 'optional'
    is_ca = 'ca' in extensions
    # Setup the Common Request Attributes before we do the extensons
    config['req'] = dict(default_bits=2048, default_md='sha256',
                         distinguished_name='req_dn', string_mask='utf8only')
    config['req']['x509_extensions'] = 'ca_ext' if is_ca else 'x509_ext'
    config['req_dn'] = dict(commonName=cname, commonName_max=64)
    if altnameexts:
        config['alt_names'] = _get_certificate_alternate_names()
    config['policy_any'] = dict(
        countryName=opt, stateOrProvinceName=opt, organizationName=opt,
        organizationalUnitName=opt, emailAddress=opt, commonName='supplied')
    # If this is a Certificate Authority, add in some special properties
    if is_ca:
        compsect = '%s_ca' % component
        config['ca'] = dict(default_ca=compsect)
        config[compsect] = dict(
            default_days='3652', default_md='sha256', unique_subject='no',
            preserve='no', email_in_dn='no', nameopt='ca_default',
            certopt='ca_default', policy='policy_any')
        # If this has a cert extension, then we need to add it to the list
        if 'cert' in extensions:
            config[compsect]['x509_extensions'] = 'cert_ext'
        # Add in the file-specfic properties to the CA-specific section
        certdir = cacert[:cacert.rfind('/')]
        config[compsect]['certificate'] = cacert
        config[compsect]['private_key'] = privkey
        config[compsect]['new_certs_dir'] = \
            certdir if certdir.endswith('certs') else (certdir + '/certs')
        # Temporary hack until all the certificates are consistent
        if certdir.find('libvirt') > 0:
            certdir = certdir[:-6]
        config[compsect]['serial'] = certdir + '/serial'
        config[compsect]['database'] = certdir + '/index.txt'
    # Loop through all of the extensions, adding in their sections
    for ext in extensions:
        extsect = config.setdefault(
            '%s_ext' % ext, dict(subjectKeyIdentifier='hash'))
        const = 'TRUE, pathlen:0' if ext == 'ca' else 'FALSE'
        authid = dict(x509='keyid,issuer', ca='keyid:always,issuer')
        extsect['basicConstraints'] = 'critical, CA:' + const
        extsect['authorityKeyIdentifier'] = authid.get(ext, 'keyid:always')
        # If this is a Client/Server section, then add the extended usage
        if ext in ['server', 'client']:
            extsect['extendedKeyUsage'] = '%sAuth' % ext
        if ext == 'x509':
            extsect['extendedKeyUsage'] = 'serverAuth'
        # Add in the Key Usage if it is appropriate for the section
        keyuse = dict(ca='keyCertSign, cRLSign',
                      client='keyEncipherment, digitalSignature',
                      server='keyEncipherment, digitalSignature',
                      cert='nonRepudiation, digitalSignature, keyEncipherment',
                      x509='keyEncipherment, digitalSignature, keyCertSign')
        if ext in keyuse:
            extsect['keyUsage'] = 'critical, ' + keyuse[ext]
    # Loop through the extension that they want us to add the Alternate names
    for ext in altnameexts:
        config['%s_ext' % ext]['subjectAltName'] = '@alt_names'
    # Now we can actually create the template file on the system
    fh, fname = tempfile.mkstemp(
        suffix='.info', prefix='powervc_temp_', dir='/tmp')
    os.close(fh)
    # Now we can write out the configuration data to the temporary file
    write_config_file(fname, config)
    return fname


def _create_certificate_database(ca_info, info_file):
    """Helper method to create the CA Serial and Database"""
    cacert = ca_info.get('cacert')
    certdir = cacert[:cacert.rfind('/')]
    makedirs(
        certdir if certdir.endswith('certs') else (certdir + '/certs'), 0o755)
    # Temporary hack until all the certificates are consistent
    if certdir.find('libvirt') > 0:
        certdir = certdir[:-6]
    serial = '01\n' if certdir.find('keystone') > 0 else '00\n'
    # Write out the index.txt and serial files for the CA Certificate
    write_to_file(certdir + '/serial', serial)
    create_file(certdir + '/index.txt', overwrite=True)
    # Copy the info file over to the certificate directory
    copy_file(info_file, certdir + '/openssl.conf')
    chmod(certdir + '/openssl.conf', 0o644)


def _get_certificate_alternate_names():
    """Helper method to generate the Certificate Alternate Names"""
    config = get_config_data()
    mgmtip = config.get('mgmt_ip')
    hostname, hostip = config.get('host_name'), config.get('host_ip')
    # Can't do anything if they didn't give us a Management IP override
    if mgmtip:
        if _is_ipaddress(mgmtip):
            hostip = mgmtip
        # If it is a hostname, then use the Management IP as the hostname
        else:
            hostname = mgmtip
    hostip = hostip.replace('[', '').replace(']', '') if hostip else hostip
    altnames = {'IP.1': hostip}
    # For some reason we can't add the DNS entries if this is IPv6
    # We also don't want to put in the DNS entries if the primary is an IP
    if config.get('ip_version') != '6' and not _is_ipaddress(mgmtip):
        altnames['DNS.1'] = hostname
        # If this isn't already a short name, parse out the short name
        if hostname.find('.') > 0:
            altnames['DNS.2'] = hostname[0:hostname.find('.')]
    return altnames


def _is_ipaddress(ipaddress):
    """Helper method to see if it is an IP Address or Host Name"""
    iprex = re.compile("^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\."
                       "){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])$")
    return ipaddress.find(':') >= 0 or iprex.match(ipaddress) is not None


######################################################
#             Security Utility Methods               #
######################################################
def is_encrypted(data):
    """Helper method to encrypt the data to help with automation"""
    if data and is_management_system():
        return data and data.startswith('aes-ctr:')
    return data and __is_obfuscated(data)


def encrypt_data(data):
    """Helper method to encrypt the data to help with automation"""
    if is_management_system():
        return __get_encrypt_handler()().encode(data)
    return __obfuscate_data(data)


def decrypt_data(data):
    """Helper method to encrypt the data to help with automation"""
    if is_management_system():
        return __get_encrypt_handler()().decode(data)
    return __unobfuscate_data(data)


def reset_encryption(passphrase=None):
    """Helper method to reset the cache with the encryption key changes"""
    if is_management_system():
        adict = dict(passphrase=passphrase) if passphrase else dict()
        __get_encrypt_handler()._cliopts = adict


def generate_password(length=20, special_chars='.-_'):
    """Helper method to generate a password for the user"""
    chars = string.ascii_letters + string.digits + special_chars
    rng = random.SystemRandom()
    password = [rng.choice(chars) for _i in range(length)]
    # Can't have a dash start the password since linux commands don't handle
    while password[0] == '-':
        password[0] = rng.choice(chars)
    return ''.join(password)


def set_password(username, password=None):
    """Helper method to set (and generate) a password for the user"""
    # If the password wasn't provided then we are going to generate one
    if not password:
        password = generate_password()
    # Run the passwd command piping in the password for the user
    run_command(['/usr/bin/passwd', '--stdin', username], inputd=password)
    # We also want to change it so these service users don't expire
    run_command(['/usr/bin/passwd', '-x', '-1', username])
    return password


def is_ldap_configured():
    """Helper method to determine if LDAP is configured on the system"""
    ldap_file = '/etc/keystone/domains/keystone.Default.conf'
    # Check to see if LDAP is configured in the domain-specific file
    val = get_openstack_config_value(ldap_file, 'driver', 'identity')
    if val:
        return 'ldap' in val
    return False


def __get_encrypt_handler():
    """Helper method to get the Encryption Handler to help with automation"""
    if sys.modules.get('powervc_keystone.encrypthandler') is None:
        __import__('powervc_keystone.encrypthandler')
    module = sys.modules['powervc_keystone.encrypthandler']
    return getattr(module, 'EncryptHandler')


def __obfuscate_data(data):
    """Helper method to obfuscate the data with a base64 encoding"""
    if not __is_obfuscated(data):
        encoded = '[IBM:v1]' + data.encode('rot13')
        data = encoded.encode('base64').replace('\n', '')
    return data


def __unobfuscate_data(data):
    """Helper method to decode the obfuscated data with base64 encoding"""
    if __is_obfuscated(data):
        decoded = data.decode('base64')
        data = decoded[8:].decode('rot13')
    return data


def __is_obfuscated(data):
    """Helper method to see if the data has been base64 encoded"""
    try:
        return data.decode('base64').startswith('[IBM:v1]')
    except binascii.Error:
        return False
    except AttributeError:
        return False


######################################################
#          Maintenance Mode Utility Methods          #
######################################################
def enter_maintenance_mode():
    """Update Apache so that the UI says PowerVC is in maintenance mode"""
    eventlet.monkey_patch(os=False)
    # The contents that should be written out to the .htaccess Apache file
    contents = 'RewriteEngine on\nRewriteCond %{REQUEST_URI} !/maintenance.ht'\
        'ml$\nRewriteCond %{REQUEST_URI} !/*.js$\nRewriteCond %{REQUEST_URI} '\
        '!/*.css$\nRewriteCond %{REQUEST_URI} !/*.ico$\nRewriteCond %{REQUEST'\
        '_URI} !/*.png$\nRewriteCond %{REQUEST_URI} !/*.gif\nRewriteCond %{RE'\
        'QUEST_URI} !/*.json\nRewriteCond %{REQUEST_URI} !/*idx/form/*\nRewri'\
        'teRule $ /powervc/maintenance.html [R=302,L]\nHeader set Maintenance'\
        ' "on"\n'
    try:
        # Write out the .htaccess to put to the UI in Maintenance Mode
        write_to_file('/var/www/html/powervc/.htaccess', contents)
        # Add a timeout on sending the notification so don't hang forever
        with eventlet.timeout.Timeout(30):
            cfg.CONF([], project='nova')
            cfg.CONF.notification_driver = ['messaging']
            from oslo_context import context
            ctxt = context.get_admin_context()
            transport = messaging.get_transport(cfg.CONF)
            notifier = messaging.Notifier(transport, 'compute')
            # Send a notification to the UI saying entering Maintenance Mode
            notifier.info(ctxt.to_dict(), "system.maintenance", {})
    except eventlet.timeout.Timeout:
        LOG.warn(_("Timed out trying to notify UI of maintenance mode"))
    except Exception as ex:
        LOG.exception(_("Exception entering maintenance mode: %s") % ex)


def exit_maintenance_mode():
    """Updates Apache so the UI no longer says in maintenance mode"""
    try:
        # Remove .htaccess file to remove the UI from Maintenance Mode
        if file_exists('/var/www/html/powervc/.htaccess'):
            remove_file('/var/www/html/powervc/.htaccess')
    except Exception as ex:
        LOG.exception(_("Exception exiting maintenance mode: %s") % ex)
