#!/usr/bin/python
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import os
import argparse
import logging
from sqlalchemy import MetaData
from sqlalchemy.sql import expression
from sqlalchemy.sql.schema import UniqueConstraint
from oslo_utils import encodeutils

from powervc_oslo.config import data_utils
from powervc_oslo.config.data_utils import CONSOLE, V130
from powervc_oslo import _

DATA_FILE = '/etc/nova/exported_db2_data.dat'
DB_NAMES = ['cinder', 'ceilometer', 'glance', 'keystone', 'neutron', 'nova']
TYPE_MAP = dict(smallint='TINYINT', integer='INT', bigint='BIGINT',
                double='FLOAT', decimal='DECIMAL', timestamp='DATETIME',
                var='VARCHAR', clob='TEXT')
LOG = logging.getLogger(__name__)


#################################################
# ############   Export DB2 Data   ##############
#################################################
def export_db2_data():
    """Exports the DB2 data to be imported into MariaDB"""
    msg = _("Exporting database data from DB2 to restore to MariaDB...")
    LOG.log(CONSOLE, msg)
    # Get the version to determine which databases to export from
    version = data_utils.get_powervc_version()
    isv12 = data_utils.compare_version(version, V130) < 0
    # First create the file we are going to export the data to
    if os.path.isfile(DATA_FILE):
        os.remove(DATA_FILE)
    os.close(os.open(DATA_FILE, os.O_CREAT | os.O_EXCL, 0o600))
    # Next loop through each of the databases, dumping out the data
    for dbname in DB_NAMES:
        # Need to skip Ceilometer if it used the NoSQL support in 1.2.x
        if dbname != 'ceilometer' or not isv12:
            _export_database_data(dbname)


def _export_database_data(dbname):
    """Helper method to export the data for a given database"""
    engine = _get_database_engine(dbname)
    metadata = MetaData(bind=engine, schema=dbname, reflect=True)
    # Loop through each of the database tables in the given database
    for table in metadata.tables.values():
        _export_database_table(engine, dbname, table)
    cascade_deletes = _get_cascading_deletes(engine)
    # Add in the Foreign Keys afterwards to make sure tables exist
    for table in metadata.tables.values():
        _export_table_foreign_keys(dbname, table, cascade_deletes)


def _export_database_table(engine, dbname, table):
    """Helper method to export the data for an individual table"""
    cols = ', '.join(_get_table_columns(table))
    # With Ceilometer, they use decimal in MySQL instead of timestamp
    if dbname == 'ceilometer':
        cols = cols.replace('DATETIME', 'DECIMAL(20,6)')
    # First we will write out the create table statement to the file
    _write_statement('CREATE TABLE %s.`%s` (%s)' % (dbname, table.name, cols))
    # Next loop through the Indexes writing out the statements for those
    for index in table.indexes:
        idxname = _get_index_name(index)
        addon = 'UNIQUE ' if index.unique else ''
        cnames = ', '.join([_escape_name(c) for c in index.columns])
        # Need to create a dummy index to allow glance to delete
        if index.name == 'ix_metadef_ns_res_types_namespace_id':
            cnames = cnames.replace('namespace_id', 'prefix')
        # Now we can write out the index for the given constraint
        _write_statement('CREATE %sINDEX %s on %s.`%s` (%s)' %
                         (addon, idxname, dbname, table.name, cnames))
    # Next loop through the Unique Constraints writing out those Indexes
    for cst in table.constraints:
        if isinstance(cst, UniqueConstraint):
            cstname = _get_index_name(cst)
            unique = 'UNIQUE ' if _is_unique_constraint(cst) else ''
            cnames = ', '.join([_escape_name(c) for c in cst.columns])
            _write_statement('CREATE %sINDEX %s on %s.`%s` (%s)' %
                             (unique, cstname, dbname, table.name, cnames))
    # Now we can write out the data that is in the given table
    _export_table_rows(engine, dbname, table)


def _export_table_rows(engine, dbname, table):
    """Helper method to export the row data for the individual table"""
    # Since we switched to use the community migrate_repo in 1.3.1 and
    # this needs to be done up front, we need to fix the value right now
    if dbname == 'ceilometer' and table.name == 'migrate_version':
        path = '/usr/lib/python2.7/site-packages/'\
               'ceilometer/storage/sqlalchemy/migrate_repo'
        _write_statement("INSERT INTO ceilometer.migrate_version "
                         "VALUES('ceilometer', '%s', 45)" % path)
        return
    # Loop through each of the rows inserting the values for that row
    for row in engine.execute(expression.select([table])).fetchall():
        values = [_get_column_value(row, c) for c in table._columns.values()]
        _write_statement('INSERT INTO %s.`%s` VALUES(%s)' %
                         (dbname, table.name, ', '.join(values)))


def _export_table_foreign_keys(dbname, table, cascade_deletes):
    """Helper method to export the foreign key alter statements"""
    stmt = 'ALTER TABLE %s.`%s` ADD FOREIGN KEY '\
           '%s (`%s`) REFERENCES %s.`%s`(`%s`)'
    # Loop through the columns, writing out the foreign key alter statements
    for col in table.columns:
        for fkey in col.foreign_keys:
            vals = (dbname, table.name, _get_fkey_name(table, col, fkey),
                    col.name, dbname, fkey.column.table.name, fkey.column.name)
            fstmt = stmt % vals
            # If this is a cascading delete foreign key, tag it as such
            if _is_cascading_delete(cascade_deletes, table, col, fkey):
                fstmt += ' ON DELETE CASCADE'
            _write_statement(fstmt)


def _get_cascading_deletes(engine):
    """Helper method to get the cascading deletes for the given database"""
    stmt, rows = "SELECT TABNAME, FK_COLNAMES, REFTABNAME, PK_COLNAMES "\
        "FROM SYSCAT.REFERENCES WHERE DELETERULE='C'", dict()
    # Loop through each of the rows inserting the values for that row
    for row in engine.execute(stmt).fetchall():
        rows['.'.join([r.strip().lower() for r in row])] = ''
    return rows


def _is_cascading_delete(cascade_deletes, table, col, fkey):
    """Helper method to determine if this is a cascading delete"""
    return '%s.%s.%s.%s' % (table.name, col.name, fkey.column.table.name,
                            fkey.column.name) in cascade_deletes


def _get_table_columns(table):
    """Helper method to get the column information for a table"""
    columns = [_get_column_data(c) for c in table._columns.values()]
    keys = [_escape_name(c) for c in table._columns.values() if c.primary_key]
    # Next add in the primary keys to the column portion of the create
    if keys:
        columns.append('PRIMARY KEY (%s)' % ', '.join(keys))
    return columns


def _get_fkey_name(table, column, fkey):
    """Helper method to generate the name to use for the foreign key"""
    if fkey.name.startswith('sql'):
        return '%s_%s_fkey' % (table.name, column.name)
    return fkey.name


def _get_index_name(index):
    """Helper method to return the manpulated index/constraint name"""
    if index.name == 'ix_tags_namespace_id_name':
        return 'ix_metadef_tags_namespace_id'
    if index.name == 'ix_namespaces_namespace':
        return 'ix_metadef_namespaces_namespace'
    if index.name == 'ix_metadef_ns_res_types_namespace_id':
        return 'ix_metadef_ns_res_types_res_type_id_ns_id'
    if index.name == 'user_dom_name_unique':
        return 'ixu_user_name_domain_id'
    return index.name


def _is_unique_constraint(constraint):
    """Helper method to return whether the constraint should be unique"""
    names = ['ix_objects_namespace_id_name', 'ix_tags_namespace_id_name',
             'ix_namespaces_namespace', 'ix_metadef_resource_types_name',
             'ix_metadef_properties_namespace_id_name']
    return constraint.name not in names


def _escape_name(column):
    """Helper method to return the column name to use in the statement"""
    # If it is a reserved keyword, we need to escape the name
    return '`' + column.name + '`'


def _get_column_value(row, column):
    """Helper method to get the column value in the given row"""
    typestr, valstr = _get_column_type(column, False), row[column.name]
    # Manipulate the value to match what is needed in the insert
    valstr = 'NULL' if valstr is None else valstr
    valstr = valstr if isinstance(valstr, basestring) else unicode(valstr)
    valstr = valstr.replace('\\', '\\\\').replace("'", "\\'")
    # If this is a text value we need to put this into quotes
    if valstr != 'NULL' and typestr in ['VARCHAR', 'TEXT', 'DATETIME']:
        valstr = "'" + valstr + "'"
    return valstr


def _get_column_data(column):
    """Helper method to get the information for the column"""
    data = '%s %s' % (_escape_name(column), _get_column_type(column))
    # Check if we need to add the not-null tag to the string
    data += '' if column.nullable else ' NOT NULL'
    # Check if we need to add the auto-increment tag to the string
    data += ' AUTO_INCREMENT' if column.autoincrement else ''
    # Add in the default value if there is one that is specified
    if column.server_default is not None:
        data += " DEFAULT " + unicode(column.server_default.arg)
    return data


def _get_column_type(column, include_length=True):
    """Helper method to get the type information for the column"""
    typestr = str(type(column.type)).lower()
    # Loop through the mapping of types checking which one this is
    for oldtype, newtype in TYPE_MAP.iteritems():
        if typestr.find(oldtype) > 0:
            # For varchar we need to tack on the length to use
            if include_length and newtype == 'VARCHAR':
                return 'VARCHAR(%s)' % column.type.length
            return newtype
    raise Exception('Unexpected ColumnType Found')


def _get_database_engine(dbname):
    """Helper method to get an engine for the connection to a database"""
    key, sect = 'connection', 'database'
    conf_file = '/etc/%s/%s.conf' % (dbname, dbname)
    # Need to use a different configuration file if this is glance
    if dbname == 'glance':
        conf_file = '/etc/glance/glance-api.conf'
    return data_utils.get_database_engine(conf_file, key, sect)


def _write_statement(stmt):
    """Helper method to write a statement line to the file"""
    stmt = encodeutils.safe_encode(stmt)
    data_utils.append_to_file(DATA_FILE, stmt + ';\n')


#################################################
# ############   Import DB2 Data   ##############
#################################################
def import_db2_data():
    """Imports the exported DB2 data into MariaDB"""
    msg = _("Importing database data from DB2 to restore to MariaDB...")
    LOG.log(CONSOLE, msg)
    # Get the version to determine which databases to export from
    version = data_utils.get_config_data().get('restored_version')
    isv12 = version and data_utils.compare_version(version, V130) < 0
    # First we will drop the tables in each of the databases in the backup
    for dbname in DB_NAMES:
        # Need to skip Ceilometer if it used the NoSQL support in 1.2.x
        if dbname != 'ceilometer' or not isv12:
            data_utils.drop_database_tables(dbname)
    # Need to swap one of the index names for Keystone to handle 1.2.1
    before = 'CREATE UNIQUE INDEX user_dom_name_unique'
    after = 'CREATE UNIQUE INDEX ixu_user_name_domain_id'
    data_utils.sed_replace(DATA_FILE, before, after)
    # Next we want to run all the statements in the exported data file
    statements = data_utils.read_from_file(DATA_FILE).split(';\n')
    data_utils.batch_database_execute(statements[:-1])
    # When we are done cleanup the exported database data file
    data_utils.remove_file(DATA_FILE)


#################################################
# ############   Main Entry Point   #############
#################################################
def main():
    """Main invocation for the DB conversion script"""
    parser = argparse.ArgumentParser()
    parser.add_argument('--logfile', dest='logfile',
                        action='store', help=argparse.SUPPRESS)
    args = parser.parse_args()
    # Enable Logging to the Log File if they asked
    data_utils.setup_logging(args.logfile)
    # The main program is used for exporting the DB2 data
    export_db2_data()

if __name__ == '__main__':
    main()
