# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import logging
import six

from powervc_oslo.config import data_mgr as mgr
from powervc_oslo.config import data_utils as utils
from powervc_oslo.config import db_conversion
from powervc_oslo.config.data_utils import CONSOLE, V131
from powervc_oslo import _

PWRVCDB_CONF = '/opt/ibm/powervc/data/powervc-db.conf'
MARIADB_CONF = '/etc/my.cnf.d/server.cnf'
DB2_EXPORT_DATA = db_conversion.DATA_FILE
PORT = '50110'

LOG = logging.getLogger(__name__)


######################################################
#             DB Data Manager Definition             #
######################################################
class DbDataManager(mgr.BaseDataManager):

    def get_component_name(self):
        """Provides the name of the Component providing the Data Manager"""
        return 'db'

    def is_supported_environment(self):
        """Specifies whether the Component is supported in this Environment"""
        return utils.is_management_system()

    ###############################################
    #     Data Manager - Configure Interface      #
    ###############################################
    def unconfigure_component(self, upgrading):
        """Performs Un-Configuration specific to the given Component"""
        if not upgrading:
            self.__cleanup_mariadb()

    ###############################################
    #       Data Manager - Backup Interface       #
    ###############################################
    def create_database(self):
        """Creates the Database for the given Component"""
        self.__initialize_mariadb()

    def prepare_component_restoration(self):
        """Performs any Component-specific preparation needed for restore"""
        config, dbfile = utils.get_config_data(), DB2_EXPORT_DATA[1:]
        # If before 1.3.1, we only support if db2 converted data is available
        if utils.compare_version(config['restored_version'], V131) < 0:
            tloc, tfile = config['restore_dir'], config['restore_file']
            if not utils.tar_file_contains_file(tloc, tfile, dbfile):
                msg = _("Only data backed up on a 1.3.1 system can be restored"
                        " to this system.  Upgrade that system to 1.3.1 before"
                        " trying to restore its data to another system.")
                raise SystemExit(msg)
        else:
            # Since we are restoring the database over top, wipe out the old DB
            utils.remove_directory(utils.get_config_data()['db_location'])

    def restore_component(self, upgrading):
        """Performs any Component-specific restoration that is needed"""
        config = utils.get_config_data()
        dbloc, version = config['db_location'], config['restored_version']
        # Make sure that the MySQL user has permission to the entire database
        utils.run_command('/usr/bin/chown -R mysql:mysql ' + dbloc)
        # If we are coming from before 1.3.1, then we need to import DB2 data
        if utils.compare_version(version, V131) < 0:
            db_conversion.import_db2_data()

    ###############################################
    # ## Data Manager - System Changed Interface ##
    ###############################################
    def service_passwords_changed(self):
        """Updates the Service Passwords in the CONF files when they change"""
        newpwd, stmts = utils.get_config_data()['db_password'], list()
        # If we are coming form before 1.3.1, then no old password to change
        version = utils.get_config_data().get('restored_version')
        if version and utils.compare_version(version, V131) < 0:
            utils.set_openstack_config_value(
                PWRVCDB_CONF, 'db_password', utils.encrypt_data(newpwd))
            return
        # Get the old password from the configuration file to use on to change
        oldpwd = utils.get_openstack_config_value(PWRVCDB_CONF, 'db_password')
        # Call into MariaDB to change the password to the value for this system
        self.__change_mariadb_password(newpwd, oldpwd)
        # Update the configuration file with the new password that was set
        utils.set_openstack_config_value(
            PWRVCDB_CONF, 'db_password', utils.encrypt_data(newpwd))
        # Now loop through all the database users, updating their passwords
        for key, val in utils.get_config_data().iteritems():
            if key != 'db_password' and key.endswith('db_password'):
                for ip in ['localhost', '127.0.0.1', '::1']:
                    dbpwd = "PASSWORD('%s')" % val
                    dbuser = "'%s'@'%s'" % (key[:-9], ip)
                    stmts.append("SET PASSWORD FOR %s = %s" % (dbuser, dbpwd))
        utils.batch_database_execute(stmts)

    def local_ip_changed(self):
        """Updates the Management IP in the CONF files when they change"""
        old_ip = utils.get_openstack_config_value(
            MARIADB_CONF, 'bind-address', 'mysqld')
        new_ip = utils.get_config_data()['local_ip']
        # If the old and new local ip address are different, change the binding
        if old_ip != new_ip:
            utils.set_openstack_config_value(
                MARIADB_CONF, 'bind-address', new_ip, 'mysqld')
            utils.run_command('/bin/systemctl restart mariadb')

    ###############################################
    #   Data Manager - Subclass Implementation    #
    ###############################################
    def _get_enabled_services(self):
        """Provides a list of the Services for the Component"""
        return ['mariadb']

    def _get_included_backup_files(self, upgrading=False):
        """Provides a list of Component files to exclude from a backup"""
        return [PWRVCDB_CONF, utils.get_config_data()['db_location']]

    def _get_service_password_properties(self):
        """Provides a list of the configuration properties for passwords"""
        return [('db', PWRVCDB_CONF, 'DEFAULT', 'db_password')]

    ###############################################
    #    Internal Data Manager Helper Methods     #
    ###############################################
    def __initialize_mariadb(self):
        """Helper method to do the initial setup for MariaDB"""
        config = utils.get_config_data()
        dbpwd = config['db_password'] = utils.generate_password()
        dbloc, dbpwd = config['db_location'], utils.encrypt_data(dbpwd)
        # First we want to log that we are configuring MariaDB
        LOG.log(CONSOLE, _("Configuring PowerVC to use the MariaDB server..."))
        # Update the PowerVC DB Configuration file with the location/password
        utils.set_openstack_config_value(PWRVCDB_CONF, 'db_location', dbloc)
        utils.set_openstack_config_value(PWRVCDB_CONF, 'db_password', dbpwd)
        # Update the MariaDB Configuration file with the location/port/address
        utils.set_openstack_config_value(MARIADB_CONF, 'data', dbloc, 'mysqld')
        utils.set_openstack_config_value(MARIADB_CONF, 'port', PORT, 'server')
        utils.set_openstack_config_value(
            MARIADB_CONF, 'bind-address', config['local_ip'], 'mysqld')
        # Add the performance/scale related Configuration properties
        utils.set_openstack_config_value(
            MARIADB_CONF, 'open_files_limit', '8192', 'mysqld')
        utils.set_openstack_config_value(
            MARIADB_CONF, 'max_connections', '1024', 'mysqld')
        utils.remove_directory(dbloc)
        # Run the command to do the initial database setup needed
        utils.run_command('/usr/bin/mysql_install_db', False)
        # Make sure that the MySQL user has permission to the entire database
        utils.run_command('/usr/bin/chown -R mysql:mysql ' + dbloc)
        # Next we need to startup the MariaDB server before we do the rest
        utils.run_command('/bin/systemctl restart mariadb')
        # Next we will change the root password for the the MariaDB server
        self.__change_mariadb_password(config['db_password'])
        # Next we can run the configuration changes to make MariaDB secure
        self.__make_mariadb_secure()

    def __cleanup_mariadb(self):
        """Helper method to cleanup anything configured for MariaDB"""
        # Remove the location/password from the PowerVC DB Configuration file
        utils.del_openstack_config_value(PWRVCDB_CONF, 'db_location')
        utils.del_openstack_config_value(PWRVCDB_CONF, 'db_password')
        # Remove the location/port from the MariaDB Configuration file
        utils.del_openstack_config_value(MARIADB_CONF, 'data', 'mysqld')
        utils.del_openstack_config_value(MARIADB_CONF, 'port', 'server')
        utils.del_openstack_config_value(
            MARIADB_CONF, 'bind-address', 'mysqld')
        # Remove the performance/scale related Configuration properties
        utils.del_openstack_config_value(
            MARIADB_CONF, 'open_files_limit', 'mysqld')
        utils.del_openstack_config_value(
            MARIADB_CONF, 'max_connections', 'mysqld')
        # Remove the directory that we used to store the database data
        utils.remove_directory(utils.get_config_data()['db_location'])

    def __change_mariadb_password(self, newpwd, oldpwd=None):
        """Helper method to change the MariaDB password to the value"""
        import pexpect
        # Use the MySQLAdmin command to change the root password
        pwdcmd = '/usr/bin/mysqladmin password'
        pwdcmd = (pwdcmd + ' -p') if oldpwd else pwdcmd
        p = pexpect.spawn(pwdcmd)
        # If the password was previously set, we need to enter it
        if oldpwd:
            p.expect(['Enter password:'], timeout=10)
            p.sendline(oldpwd)
        # Send in the new password and the confirmation for it
        p.expect(['New password:'], timeout=10)
        p.sendline(newpwd)
        p.expect(['Confirm new password:'], timeout=10)
        p.sendline(newpwd)
        p.expect(pexpect.EOF, timeout=10)

    def __make_mariadb_secure(self):
        """Helper method to make the MariaDB Server secure for use"""
        stmts, hosts = list(), six.text_type(('localhost', '127.0.0.1', '::1'))
        # Run the various SQL statements needed to make the server secure
        stmts.append("DELETE FROM mysql.db WHERE Db='test' OR Db='test\\_%'")
        stmts.append("DELETE FROM mysql.user WHERE Host NOT IN %s" % hosts)
        stmts.append("DELETE FROM mysql.user WHERE User=''")
        stmts.append("DROP DATABASE test")
        # Now we can do the batch execute of the statements provided
        utils.batch_database_execute(stmts)
