# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
from __future__ import print_function
import glob
import re
import sys
from powervc_oslo.config import data_utils
from powervc_oslo import _


service_map = {'glance': 'image',
               'nova': 'compute',
               'neutron': 'network',
               'cinder': 'storage',
               'ceilometer': 'metering',
               'validator': 'validation',
               'keystone': 'identity',
               'clerk': 'self-service'}

CONF_EXCLUDES = ['/etc/nova/nova-health.conf',
                 '/etc/glance/glance-cache.conf',
                 '/etc/glance/glance-glare.conf',
                 '/etc/glance/glance-scrubber.conf',
                 '/etc/cinder/cinder-health.conf']


def handle_debug(args):
    tgt_services = [args.service]
    if args.service is 'general':
        tgt_services = service_map.keys()

    config = False
    for srv in tgt_services:
        view = True
        if args.enable:
            view = False
            enable_debug(srv)
            config = True
        elif args.disable:
            view = False
            if args.modules:
                print(_('--module option cannot be used with --disable'),
                      file=sys.stderr)
                sys.exit(1)
            disable_debug(srv)
            config = True
        if args.modules:
            view = False
            modules = args.modules.split(',')
            if srv in modules:
                # The main service module is special
                modules.remove(srv)
                enable_debug(srv)
            add_modules(srv, modules)
            config = True

        if args.restart:
            view = False
            if config:
                print(_('Restarting %s services') % service_map[srv])
                restart_cmd = (
                    '/opt/ibm/powervc/bin/powervc-services %s restart' %
                    args.service)
                data_utils.run_command(restart_cmd)
            else:
                print(_('--restart option cannot be used alone'),
                      file=sys.stderr)
                sys.exit(1)
        if view:
            show_debug_status(srv)


def show_debug_status(service_name):
    r_list = file_list(service_name)
    debug_level = {}
    modules = {}
    for conf in sorted(r_list):
        debug_level[conf] = data_utils.get_openstack_config_value(
            conf, "debug")
        if debug_level[conf] is None:
            debug_level[conf] = False  # default is False if not set
        else:
            # convert strings to booleans
            debug_level[conf] = debug_level[conf].lower() == 'true'

        modules[conf] = []
        val = data_utils.get_openstack_config_value(conf, "default_log_levels")
        if val is None:
            val = ''  # default is empty string if not set
        if val != '':
            for s in val.split(','):
                (m, v) = s.split('=')
                if v == 'DEBUG':
                    modules[conf].append(m)

    base_prefix = '%s' % service_map[service_name]
    prefix = base_prefix
    for conf in sorted(r_list):
        if len(r_list) > 1:
            short_conf = re.match('.*/(.*)\.conf', conf).group(1)
            prefix = base_prefix + ': %s' % short_conf
        if debug_level[conf] is False and modules[conf] == []:
            print(_('%(prefix)s: Debug logging is disabled') %
                  {'prefix': prefix})
        else:
            sorted_modules = []
            if debug_level[conf] is True:
                sorted_modules.append(service_name)
            sorted_modules.extend(sorted(modules[conf]))
            modules_str = ','.join(sorted_modules)
            print(_('%(prefix)s: Debug logging is enabled for the '
                    'following modules: %(modules)s') %
                  {'prefix': prefix, 'modules': modules_str})


def update_module_val(conf_file, modules):
    log_level = data_utils.get_openstack_config_value(
        conf_file,
        "default_log_levels")
    if log_level is None:
        log_level = ''

    # Backup the default_log_level values against a
    # new attribute named default_log_levels_orig.
    # Only done the first time.
    log_lvl_orig = data_utils.get_openstack_config_value(
        conf_file,
        "default_log_levels_orig")
    if (log_lvl_orig is None):
        log_lvl_orig = log_level
        data_utils.set_openstack_config_value(
            conf_file,
            "default_log_levels_orig",
            log_lvl_orig)

    # merge original default_log_levels and user-specified settings
    default_modules = []
    if log_lvl_orig != '':
        default_modules = [s.split('=') for s in log_lvl_orig.split(',')]
    if len(modules) > 0:
        val = '=DEBUG,'.join(modules) + '=DEBUG'
    for d in default_modules:
        if d[0] not in modules:
            val += ',%s=%s' % (d[0], d[1])

    # set the new value
    data_utils.set_openstack_config_value(
        conf_file,
        "default_log_levels",
        val)


def file_list(service_name):
    if service_name is "validator":
        service_name = "ttv-validation"
    f_list = list()
    select_files = "/etc/{0}/{1}*.conf".format(service_name, service_name)
    for conf in glob.glob(select_files):
        if conf not in CONF_EXCLUDES:
            f_list.append(conf)
    return f_list


def add_modules(service_name, modules):
    r_list = file_list(service_name)
    for conf in r_list:
        update_module_val(conf, modules)


def enable_debug(service_name):
    r_list = file_list(service_name)
    for conf in r_list:
        data_utils.set_openstack_config_value(conf, "debug", "True")


def disable_debug(service_name):
    r_list = file_list(service_name)
    for conf in r_list:
        data_utils.set_openstack_config_value(conf, "debug", "False")
        # Revert default_log_level to the original state.
        log_lvl_orig = data_utils.get_openstack_config_value(
            conf,
            "default_log_levels_orig")
        if log_lvl_orig is not None:
            if log_lvl_orig == '':
                data_utils.del_openstack_config_value(
                    conf,
                    "default_log_levels")
            else:
                data_utils.set_openstack_config_value(
                    conf,
                    "default_log_levels",
                    log_lvl_orig)


def build_debug_parser(comp_parser, service_name):
    if service_name is 'general':
        help_msg = _('Configure the debug settings for all PowerVC services')
        desc_msg = _('Run this sub-command without any options to view the '
                     'current debug settings for all PowerVC services. '
                     'NOTE: This command should only be used under direction '
                     'from IBM service.')
        enable_msg = _('Used to enable debug for all PowerVC services')
        disable_msg = _('Used to disable debug for all PowerVC services')
        restart_msg = _('Use this flag to restart all PowerVC services. '
                        'A restart is required for changes to take effect.')
    else:
        component = service_map[service_name]
        help_msg = _('Configure the debug settings for the %s '
                     'service. ') % component
        desc_msg = _('Run this sub-command without any options to view the '
                     'current debug settings for this service. '
                     'NOTE: This command should only be used under direction '
                     'from IBM service.')
        enable_msg = _('Used to enable debug for the %s service') % component
        disable_msg = _('Used to disable debug for the %s service') % component
        restart_msg = (_('Use this flag to restart the %s service. '
                         'A restart is required for changes to take effect.')
                       % component)
    modules_msg = _('Specify external modules that need to be debug-enabled')

    debug_parser = comp_parser.add_parser('debug',
                                          help=help_msg,
                                          description=desc_msg)

    group = debug_parser.add_mutually_exclusive_group()
    group.add_argument('--enable',
                       help=(enable_msg),
                       action="store_true")
    group.add_argument('--disable',
                       help=(disable_msg),
                       action="store_true")
    debug_parser.add_argument('-m', '--modules',
                              help=modules_msg)
    debug_parser.add_argument('--restart',
                              help=(restart_msg),
                              action="store_true")

    debug_parser.set_defaults(service=service_name)
    debug_parser.set_defaults(func=handle_debug)
    return comp_parser
