# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import glob
import sys
from oslo_utils import encodeutils
from powervc_oslo.config import config_utils as cutils
from powervc_oslo.config import data_utils as utils
from powervc_oslo.config.data_utils import CONSOLE
from powervc_oslo.config import data_mgr as mgr
from powervc_oslo.config import db_data_mgr
from powervc_oslo.config import http_data_mgr
from powervc_oslo import _
from signal import signal, SIGINT

HTTPD_CONF_FILES = ['powervc_bumblebee',
                    'powervc_ceilometer',
                    'powervc_cinder',
                    'powervc_glance',
                    'powervc_neutron',
                    'powervc_nova',
                    'powervc_swift',
                    'ttv-validation']


def _restore_values_on_interrupt():
    """Restore the previous management values on a signal interrupt such as
    a keyboard interrupt. Use a closure to pass values to the internal
    _signal_handler function.
    """
    ip_version = cutils.get_management_ip_version()
    ip_address = utils.get_management_ip()
    hostname = utils.get_management_hostname()
    config = utils.get_config_data()

    # Use a closure to pass stored values to signal.signal
    def _signal_handler(signal, frame):
        utils.LOG.log(CONSOLE, _('Interrupt Detected: Reverting Changes'))
        utils.LOG.log(CONSOLE, _('Reverting IP version to: %s' % ip_version))
        utils.LOG.log(CONSOLE, _('Reverting IP address to: %s' % ip_address))
        utils.LOG.log(CONSOLE, _('Reverting hostname to : %s' % hostname))
        mgmt_address = utils.get_management_ip()
        mgmt_hostname = utils.get_management_hostname()
        validation_code = _validate_ifconfig_params()
        _change_powervc_management_address(ip_version, ip_address, hostname,
                                           mgmt_address, mgmt_hostname,
                                           config, validation_code)
        mgr.restart_services()
        sys.exit(1)

    signal(SIGINT, _signal_handler)


def change_ip_address(args, restart_services=True, log=None):
    """Change the PowerVC IP address configuration.
    :params args: the argparse.Namespace object with command line args
    :params: restart_services: Boolean to restart services, default True
    :params: log: log filename, default powervc_config
    """
    # if log is placeholder /dev/null then setup logging
    if not log:
        log = utils.setup_logging('powervc_config', False)
    _restore_values_on_interrupt()
    current_ip = utils.get_local_ipaddress().strip('[]')
    utils.LOG.info(_("Current ip: %s") % current_ip)
    if not args.set:
        cutils.print_current_default(current_ip)
    else:
        try:
            ip_to_set = "".join(args.set)
            utils.LOG.info(_("ip to set = %s") % ip_to_set)
            cutils.validate_ip(ip_to_set)
            # check if current setting is ip based or hostname
            # if hostname is current mgmtip then do not overwrite it
            current_mgmt_ip = utils.get_management_ip().strip('[]')
            utils.LOG.info(_("current mgmt_ip setting = %s") % current_mgmt_ip)
            if cutils.is_valid_address(current_mgmt_ip):
                utils.LOG.info(_("Updating mgmt ip"))
                # Set the global _CONFIG_DATA['mgmt_ip'] value
                utils.get_config_data()['mgmt_ip'] = ip_to_set
                # Detect the changes in the global _CONFIG_DATA
                mgr.management_ip_changed()
                # Update the ssl.conf file for ipv4 systems only
                if str(cutils.get_management_ip_version()) == '4':
                    _update_ssl_conf_file(ip_to_set)
                # Update the httpd conf files
                _update_httpd_conf_files(current_mgmt_ip, ip_to_set)
            _reconfigure_cert_credentials()
            if utils.get_management_hostname() == utils.get_local_hostname():
                if restart_services:
                    cutils.safe_print(_('Restarting services'))
                    mgr.restart_services()
                else:
                    utils.LOG.info(_("Services were not set to restart"))
            else:
                raise ValueError('Local hostname did not match management '
                                 'hostname')
            cutils.safe_print(_('Configuration complete.'
                                'New IP: %s') % utils.get_management_ip())
        except Exception as exc:
            cutils.safe_print(_('There were errors during '
                                'configuration.'
                                ' Check the log file %s') % log)
            utils.LOG.exception(exc)
            sys.exit(2)


def change_hostname(args, restart_services=True, log=None):
    """Change the PowerVC hostname configuration.
    :params args: the argparse.Namespace object with command line args
    :params restart_services: Boolean to restart services, default True
    :params log: log filename, default powervc_config
    """
    # if log is placeholder /dev/null then setup logging
    if not log:
        log = utils.setup_logging('powervc_config', False)
    _restore_values_on_interrupt()
    current_hostname = utils.get_local_hostname()
    utils.LOG.info(_("Current hostname: %s") % current_hostname)
    if not args.set:
        cutils.print_current_default(current_hostname)
    else:
        try:
            hostname_to_set = "".join(args.set)
            utils.LOG.info(_("Hostname to set %s") % hostname_to_set)
            cutils.validate_hostname(encodeutils.safe_decode(str(
                hostname_to_set)))
            mgmt_sys = utils.get_management_ip().strip('[]')
            if not cutils.is_valid_address(mgmt_sys):
                # If the management IP is not a valid address
                # get the value of the mgmt_ip property from
                # the global _CONFIG_DATA variable or the nova.conf file
                utils.get_config_data()['mgmt_ip'] = hostname_to_set
                mgr.management_ip_changed()
                _update_httpd_conf_files(mgmt_sys, hostname_to_set)
            # update ssl conf hostname setting
            _update_ssl_conf_file(hostname_to_set)
            _reconfigure_cert_credentials()
            # need to set new hostname for ego props
            _update_ego_settings(mgmt_sys, hostname_to_set)
            # setup rabbitmq again under new hostname
            # rabbitmq only needs updating if hostname changes
            _update_rabbitmq()
            if restart_services:
                cutils.safe_print(_('Restarting services'))
                mgr.restart_services()
            else:
                utils.LOG.info(_('Services were not set to restart'))
            cutils.safe_print(_('Configuration complete.'))
        except Exception as exc:
            cutils.safe_print(_('There were errors during configuration.'
                                ' Check the log file %s') % log)
            utils.LOG.exception(exc)
            sys.exit(2)


def _change_interface(interface):
    """Set the network interface to be used by the PowerVC instance.
    :param interface: The network interface to use for PowerVC"""
    try:
        # If the return code is 0, interface exists
        if utils.run_command_output('ifconfig %s' % interface,
                                    throw=True)[0] == 0:
            utils.get_config_data()['ip_interface'] = interface
            # Change ['host_ip'] value
            utils.get_config_data()['host_ip'] = \
                utils.get_local_ipaddress().strip('[]')
            utils.set_openstack_config_value(utils.REF_CONF,
                                             'install_ip_interface',
                                             interface)
        else:
            raise ValueError('Invalid specified network interface: %s'
                             % interface)
    except Exception, exc:
        utils.LOG.exception(exc)
        sys.exit(3)


def _change_ipversion(version):
    """Select the IP version from the network interface to use in PowerVC
    :param version: the IP version to use (4/6)"""
    try:
        if str(version) == '4':
            utils.get_config_data()['ip_version'] = '4'
            utils.get_config_data()['local_ip'] = '127.0.0.1'
            utils.set_openstack_config_value(utils.REF_CONF,
                                             'install_ip_version',
                                             '4')
            utils.set_openstack_config_value(utils.REF_CONF,
                                             'use_ipv6',
                                             'False')
        elif str(version) == '6':
            utils.get_config_data()['ip_version'] = '6'
            utils.get_config_data()['local_ip'] = '::1'
            utils.set_openstack_config_value(utils.REF_CONF,
                                             'install_ip_version',
                                             '6')
            utils.set_openstack_config_value(utils.REF_CONF,
                                             'use_ipv6',
                                             'True')
        else:
            raise ValueError('Invalid IP Version %s' % version)
        # rebind the DB to the IP values
        mgr.get_registered_managers('db')[0].local_ip_changed()
    except Exception, exc:
        utils.LOG.exception(exc)
        sys.exit(3)


def _reconfigure_cert_credentials():
    """Refresh the certification credentials for the PowerVC system."""
    mgr_httpd = mgr.get_registered_managers('httpd')
    if mgr_httpd:
        ca_info = getattr(mgr_httpd[0], '_get_certificate_authority_info')()
    else:
        raise ValueError('Httpd certificate info not found.')
    utils.LOG.log(CONSOLE, _('Creating PowerVC certificate files'))
    utils.create_ca_certificate(ca_info,
                                '/etc/pki/tls/certs',
                                'powervc.crt',
                                '/etc/pki/tls/private/powervc.key')
    # Run mgt sys changed to recreate remote host certs
    # on pkvm or novalink hosts
    mgr.management_system_changed()


def _update_ssl_conf_file(new_server):
    """Update the /etc/httpd/conf.d/ssl.conf ServerName attribute with
     the new server info.
    :param new_server: The server hostname to set as ssl ServerName"""
    utils.LOG.log(CONSOLE, _('Updating ssl conf file'))
    utils.LOG.log(CONSOLE, _('ssl.conf ServerName: %s') % new_server)
    cutils.sed_replace_global('/etc/httpd/conf.d/ssl.conf',
                              'ServerName.*',
                              'ServerName %s' % new_server)


def _update_httpd_conf_files(original_server, new_server):
    """Update each of the /etc/httpd/conf.d/*.conf files, replacing each
    instance of original_server with new_server
    :param original_server: The string representing the server to be replaced
    :param new_server: The string representing the new server to replace with
    """
    utils.LOG.log(CONSOLE, _('Updating httpd conf files'))
    utils.LOG.log(CONSOLE, _('Original name: %s') % original_server)
    utils.LOG.log(CONSOLE, _('New name: %s') % new_server)
    for conf_file in HTTPD_CONF_FILES:
        cutils.sed_replace_global('/etc/httpd/conf.d/%s.conf' % conf_file,
                                  original_server,
                                  new_server)
    # if cloud edition then update clerk conf file
    if utils.get_edition() == 'cloud':
        cutils.sed_replace_global('/etc/httpd/conf.d/powervc_clerk.conf',
                                  original_server,
                                  new_server)


def _update_ego_settings(original_server, new_server):
    """Update the ego settings in /etc/nova/nova.conf and
    /opt/ibm/ego/kernel/conf/ego.conf files, replacing each
     instance of original_server with new_server
    :param original_server: The string representing the server to be replaced
    :param new_server: The string representing the new server to replace with
    """
    utils.LOG.info(_('Updating ego content'))
    utils.set_openstack_config_value('/etc/nova/nova.conf',
                                     'ego_master_list',
                                     new_server,
                                     'DEFAULT')

    conf_files = glob.glob("{0}/nova-*.conf".format('/etc/nova'))
    for conf_file in conf_files:
        if 'health' in conf_file:
            continue
        utils.set_openstack_config_value(conf_file,
                                         'ego_master_list',
                                         new_server,
                                         'DEFAULT')
    # update kernel conf files
    cutils.sed_replace_global('/opt/ibm/ego/kernel/conf/ego.conf',
                              original_server,
                              new_server)
    cluster_files = glob.glob('/opt/ibm/ego/kernel/conf/ego.cluster.*')
    for cluster_file in cluster_files:
        cutils.sed_replace_global(cluster_file,
                                  original_server,
                                  new_server)


def _update_rabbitmq():
    """ Restart the rabbitmg-server and update the users."""
    try:
        # Restart the rabbitmq-server so we can update the Users
        utils.run_command('/bin/systemctl restart rabbitmq-server', throw=True)
        utils.LOG.info(_('Restart rabbitmq service for updating users'))
        # get rabbitmq_password from nova.conf
        mq_password = utils.get_openstack_config_value(
            '/etc/nova/nova.conf', 'rabbit_password', 'oslo_messaging_rabbit')
        mqctl = '/bin/su - rabbitmq -c "/usr/lib/rabbitmq/bin/rabbitmqctl %s"'
        # Next try to add the powervc_mq user or update its password
        utils.LOG.info(_('Adding powervc_mq user'))
        if utils.run_command(mqctl % 'add_user powervc_mq %s'
                             % mq_password) != 0:
            utils.run_command(mqctl % 'change_password powervc_mq %s'
                              % mq_password, throw=True)
        # set file/dir permissions
        cmd = '/usr/lib/rabbitmq/bin/rabbitmqctl '
        cmd += 'set_permissions powervc_mq ".*" ".*" ".*"'
        utils.run_command(['/bin/su', '-', 'rabbitmq', '-c', cmd], throw=True)
        # remove guest user
        utils.run_command(mqctl % 'delete_user guest')
        utils.LOG.info(_('Deleting guest user'))
    except Exception as exc:
        utils.LOG.exception(exc)
        utils.LOG.log(CONSOLE, _('Error encountered during '
                                 'reconfigure'
                                 ' of rabbitmq service. Check the '
                                 'powervc-config log'
                                 ' for details on the error.'))


def _reconfigure_firewall(config):
    """Use the http data manager to reconfigure the
    firewall with the new changes.
    :param config: The config settings for the firewall"""
    try:
        if config.get('configure_firewall') != 'no':
            http_data_manager = http_data_mgr.HttpDataManager()
            http_data_manager._configure_firewall(config)
        else:
            utils.LOG.log(CONSOLE, _('Firewall was not reconfigured'))
    except Exception as exc:
        utils.LOG.exception(exc)
        raise exc


def _ifconfig_unchanged(ip_addr, hostname, mgmt_addr, mgmt_hostname):
    """If the IP address and the hostname are unchanged, return True."""
    if hostname == mgmt_hostname:
            if ip_addr == mgmt_addr.strip('[]'):
                utils.LOG.log(CONSOLE, _('System hostname and ip match'
                                         ' the current configuration.'))
                return True
    return False


def get_current_ifconfig_values():
    """Get the current values for hostname and IP address from both the local
    system and the PowerVC installation. Return them as a tuple in the form
    (local_ipaddress, local_hostname, management_ipaddress,
      management_hostname, ip_version)
    """
    ip_version = str(cutils.get_management_ip_version())
    utils.LOG.info(_('IP version: %s' % ip_version))
    ip_address = utils.get_local_ipaddress()
    utils.LOG.info(_('Current local address: %s' % ip_address))
    hostname = utils.get_local_hostname()
    utils.LOG.info(_('Current local hostname: %s' % hostname))
    # management_ip could be ip or hostname depending on how
    # the initial install went
    mgmt_addr = utils.get_management_ip()
    utils.LOG.info(_('Current management address: %s' % mgmt_addr))
    mgmt_hostname = utils.get_management_hostname()
    utils.LOG.info(_('Current management hostname: %s'
                     % mgmt_hostname))
    return (ip_version, ip_address, hostname, mgmt_addr, mgmt_hostname)


def _change_powervc_hostname(hostname, mgmt_addr, config):
    """Update the management IP with the local hostanme.
    :param hostname: the local hostname
    :param mgmt_addr: the current PowerVC management address
    :param config: the global config object
    """
    try:
        # Use pingable and DNS resovable hostname for config settings
        utils.LOG.log(CONSOLE, _('Setting PowerVC management address to: '
                                 '%s') % hostname)
        config['mgmt_ip'] = hostname
        mgr.management_ip_changed()
        # the props in httpd conf files could be hostname or ip
        # need to search/replace the old management_ip
        _update_httpd_conf_files(mgmt_addr, hostname)
        # update ssl conf hostname setting
        _update_ssl_conf_file(hostname)
        _reconfigure_cert_credentials()
    except Exception as exc:
        utils.LOG.exception(exc)
        raise exc


def _change_powervc_ip_address(ip_version, ip_address, hostname, mgmt_addr,
                               config):
    """Update the management IP with the local ipaddress.
    :param ip_version: IP version 4 or 6
    :param ip_address: the local system IP address
    :param hostname: the local hostname
    :param mgmt_addr: the current management address
    :param config: the global config object.
    """
    try:
        utils.LOG.log(CONSOLE, _('Setting PowerVC management address'
                                 ' to: %s') % ip_address)
        config['mgmt_ip'] = ip_address
        mgr.management_ip_changed()
        # Update the conf files with the IP address
        if str(ip_version) == '6':
            ip_address = '[' + ip_address + ']'
        _update_httpd_conf_files(mgmt_addr, ip_address)
        if cutils.is_hostname_pingable(hostname):
            _update_ssl_conf_file(hostname)
        # Update the ssl.conf file ServerName for ipv4 systems only
        # as '[' ']' will cause an httpd error
        else:
            if str(ip_version) == '4':
                _update_ssl_conf_file(ip_address)
            else:
                _update_ssl_conf_file('%s-old-hostname' % hostname)
        _reconfigure_cert_credentials()
    except Exception as exc:
        utils.LOG.exception(exc)
        raise exc


def _change_powervc_management_address(ip_version, ip_address, hostname,
                                       mgmt_addr, mgmt_hostname, config,
                                       validation_code):
    """Change the PowerVC management address
    :params ip_version: IP version 4 or 6
    :params ip_address: local machine IP address
    :params hostname: local machine hostname
    :params mgmt_addr: current PowerVC management address
    :params config: config object
    :params validation_code: the code returned by _validate_ifconfig_params,
    0 for valid hostname, 1 for valid IP address
    """
    try:
        # Need to update the local_ip to make sure it matches the ip version
        # that is being configured. And needs to be set before calling
        # management_ip_changed
        localip = '127.0.0.1' if ip_version == '4' else '::1'
        config['local_ip'] = localip
        if not utils.wait_for_db_to_start(localip):
            raise Exception('DB failed to start on IP Version change')
        if validation_code == 0:
            _change_powervc_hostname(hostname, mgmt_addr, config)
        # If the hostname is not pingable AND DNS resolvable
        # Use the IP address
        elif validation_code == 1:
            _change_powervc_ip_address(ip_version, ip_address, hostname,
                                       mgmt_addr, config)
        # need to update ego and rabbitmq with new hostname regardless
        if hostname != mgmt_hostname:
            utils.LOG.log(CONSOLE, _('Updating ego and rabbitmq with: %s' %
                                     hostname))
            _update_ego_settings(mgmt_hostname, hostname)
            _update_rabbitmq()
        _reconfigure_firewall(config)
    except Exception as exc:
        utils.LOG.exception(exc)
        raise exc


def log_and_print_current_values():
    """Log the current values of the IP version, IP address and hostname
    of the PowerVC system.
    Display values to the console.
    """
    utils.LOG.log(CONSOLE, _('PowerVC IP version : %s' %
                             cutils.get_management_ip_version()))
    utils.LOG.log(CONSOLE, _('PowerVC management address: %s' %
                             utils.get_management_ip()))
    utils.LOG.log(CONSOLE, _('PowerVC hostname: %s' %
                             utils.get_management_hostname()))
    sys.exit(0)


def _validate_ifconfig_params():
    """Validate the parameters for the change_ifconfig function,
    and return 0 if all the parameters are valid. Return 1 if hostname
    is not valid, but local IP address is pingable and valid. Otherwise
     raise an exception, and raise and exception if an invalid parameter
     is passed.
    """
    try:
        local_ip_address = utils.get_local_ipaddress()
        local_hostname = utils.get_local_hostname()
        if (cutils.is_valid_hostname(local_hostname) and
           cutils.hostname_resolves_to_ipaddress(local_ip_address)):
            return 0
        elif (cutils.is_valid_address(local_ip_address) and
              cutils.is_ipaddress_pingable(local_ip_address)):
                return 1
        else:
            raise ValueError('Neither the hostname, nor the IP address are'
                             'valid. Check the hostname and IP address of the'
                             ' system. \n'
                             'Attempted hostname: %s\n'
                             'Attempted IP address: %s \n'
                             % (local_hostname, local_ip_address))
    except Exception, exc:
        utils.LOG.log(CONSOLE, _(exc))
        raise exc


def _format_and_validate_ipversion_and_interface(args):
    """ Helper function to format the ip_version and interface.
    Return as a tuple of (ipversion, interface).
    :params args: argparse.Namespace object with stored command line arguments
    """
    try:
        ipv, intf = None, None
        if args.ipversion:
            ipv = ''.join(args.ipversion)
            if ipv:
                if ipv != '4' and ipv != '6':
                    raise ValueError('Invalid IP version: %s, please'
                                     'use either: \n'
                                     '--ipversion 4 \n'
                                     'or \n'
                                     '--ipversion 6 \n'
                                     % ipv)
        if args.interface:
            intf = ''.join(args.interface)
            if utils.run_command_output('ifconfig %s' % intf,
                                        throw=True)[0] != 0:
                raise ValueError('Invalid network interface: %s, please '
                                 'select a valid interface.' % intf)
    except Exception as exc:
        utils.LOG.exception(exc)
        raise exc
    return (ipv, intf)


def change_ifconfig(args):
    """Set the hostname and IP address to the local hostname and IP address,
    if the hostname is resolvable and pingable and the IP address is pingable.
    :param args: The arguments to the command line powervc-config general
    ifconfig command"""
    log_file = utils.setup_logging('powervc_config', True)
    _restore_values_on_interrupt()
    try:
        if not args.set:
            log_and_print_current_values()
        ip_v, ip_int = _format_and_validate_ipversion_and_interface(args)
        # If the --ipversion flag is specified, update the ip_version
        if ip_v:
            _change_ipversion(ip_v)
        # If the --interface flag is specified, update the interface
        if ip_int:
            _change_interface(ip_int)
        validation_code = _validate_ifconfig_params()
        # Get the current management address variable values
        ip_version, ip_address, hostname, mgmt_addr, mgmt_hostname =\
            get_current_ifconfig_values()
        if _ifconfig_unchanged(ip_address, hostname,
                               mgmt_addr, mgmt_hostname):
            return
        config = utils.get_config_data()
    except Exception as exc:
        utils.LOG.log(CONSOLE, _('There were errors in configuration inputs.'
                                 ' Check the log file %s') % log_file)
        utils.LOG.exception(exc)
        sys.exit(1)
    try:
        _change_powervc_management_address(ip_version, ip_address, hostname,
                                           mgmt_addr, mgmt_hostname, config,
                                           validation_code)
    except Exception as exc:
        utils.LOG.log(CONSOLE, _('There were errors during configuration.'
                                 ' Check the log file %s') % log_file)
        utils.LOG.exception(exc)
        sys.exit(2)
    if args.no_restart:
        utils.LOG.log(CONSOLE, _('Stopping services'))
        utils.LOG.log(CONSOLE, 'Warning, you are stopping '
                      'PowerVC services.'
                      ' You will need to manually restart them'
                      ' using powervc-services restart')
        mgr.stop_services()
    # If --restart argument isn't explicitly called, stop all services
    else:
        utils.LOG.log(CONSOLE, _('Restarting services'))
        mgr.restart_services()
    utils.LOG.log(CONSOLE, _('powervc-config ifconfig Configuration '
                             'Complete'))


def setup_parser(subparser):
    """Create the parser to accept general command options."""
    desc = _('Manage general settings')
    gen_subparser = cutils.create_common_parser(subparser,
                                                'general',
                                                'general', desc)
    # Management Address
    ifconfig_parser = \
        gen_subparser.add_parser('ifconfig',
                                 help=_('Change the hostname or IP'
                                        ' address of the PowerVC controller'),
                                 description=_('Run this sub-command without'
                                               ' any options to display '
                                               ' current PowerVC IP address '
                                               'and hostname. \n'
                                               'To change the current IP '
                                               'address or hostname, run this'
                                               ' command with the --set '
                                               'option.'),
                                 usage=_('powervc-config general ifconfig\n'
                                         'powervc-config general ifconfig'
                                         ' --set\n'
                                         'powervc-config general ifconfig'
                                         ' --set --ipversion 6\n'
                                         'powervc-config general ifconfig'
                                         ' --set --interface eth1\n'
                                         'powervc-config general ifconfig'
                                         ' --set --no-restart'))
    ifconfig_parser.add_argument('--set',
                                 help=_('Set the ip address and hostname'
                                        ' on the PowerVC controller'),
                                 action="store_true",
                                 default=None)
    ifconfig_parser.add_argument('--interface',
                                 help=_('Set the interface on the '
                                        'PowerVC controller'),
                                 nargs='?')
    ifconfig_parser.add_argument('--ipversion',
                                 help=_('IP version to use on the '
                                        'PowerVC controller (4 or 6)'),
                                 nargs='?')
    ifconfig_parser.add_argument('--no-restart',
                                 help=_('Do not restart powervc-services.\n'
                                        'Warning: PowerVC will not be '
                                        'operational until services are '
                                        'restarted.'),
                                 action="store_true")
    ifconfig_parser.set_defaults(func=change_ifconfig)
    # IP Address
    gen_ip_parser = gen_subparser.add_parser('ipaddress',
                                             help=_('The IP address of the '
                                                    'PowerVC controller\n'
                                                    'Warning: This option '
                                                    'will be deprecated in '
                                                    'future releases.'),
                                             usage=_('powervc-config general '
                                                     'ipaddress --set '
                                                     '123.123.123.123'))
    gen_ip_parser.add_argument('--set',
                               help=_('Provide the new ipaddress that will '
                                      'be updated across the PowerVC '
                                      'controller'),
                               metavar='<ipaddress>',
                               default=None)
    gen_ip_parser.set_defaults(func=change_ip_address)

    # Hostname
    gen_hn_parser = gen_subparser.add_parser('hostname',
                                             help=_('The hostname of the '
                                                    'PowerVC controller\n'
                                                    'Warning: This option '
                                                    'will be deprecated in '
                                                    'future releases.'),
                                             usage=_('powervc-config general'
                                                     ' hostname --set '
                                                     'powervc1.ibm.com'))
    gen_hn_parser.add_argument('--set',
                               help=_('Provide the new hostname that '
                                      'will be updated across the '
                                      'PowerVC controller\n'
                                      'Warning: This option will be '
                                      'deprecated in future releases.'),
                               metavar='<hostname>',
                               default=None)
    gen_hn_parser.set_defaults(func=change_hostname)
