# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from datetime import datetime
from powervc_oslo.config import data_mgr as mgr
from powervc_oslo.config import data_utils as utils

HTTP_CONF = '/etc/httpd/conf/httpd.conf'
HTTP_SSL_CONF = '/etc/httpd/conf.d/ssl.conf'


######################################################
#            HTTP Data Manager Definition            #
######################################################
class HttpDataManager(mgr.BaseDataManager):

    def get_component_name(self):
        """Provides the name of the Component providing the Data Manager"""
        return 'httpd'

    def is_supported_environment(self):
        """Specifies whether the Component is supported in this Environment"""
        return True

    def get_dependent_managers(self):
        """Specifies the Component Name of the Managers it is dependent on"""
        return ['db']

    ###############################################
    #     Data Manager - Configure Interface      #
    ###############################################
    def configure_component(self):
        """Performs Configuration specific to the given Component"""
        super(HttpDataManager, self).configure_component()
        config = utils.get_config_data()
        # We only want to do this configuration if on the Management System
        if not utils.is_management_system():
            return
        # If they asked us to configure the Firewall, do so now
        if config.get('configure_firewall') != 'no':
            self._configure_firewall(config)
        # Enable SELinux for the HTTP Connection
        utils.run_command(
            '/usr/sbin/setsebool -P httpd_can_network_connect 1')
        # Update the httpd.conf file with the PowerVC settings
        self.__update_httpd_conf_file()
        # Update the httpd ssl.conf file with the PowerVC settings
        self.__update_httpd_ssl_conf_file()
        # Update the PowerVC httpd conf file with the ports
        self.__update_powervc_httpd_conf_file()
        # Move PHP and PHP-ODBC to resolve defect 24159/39836 for Keystone
        phpfiles = ['/etc/httpd/conf.d/php.conf',
                    '/etc/httpd/conf.modules.d/10-php.conf']
        postfix = '.powervc_' + datetime.now().strftime('%Y-%m-%d-%H%M%S')
        for afile in phpfiles:
            if utils.file_exists(afile):
                utils.move_file(afile, afile + postfix)

    ###############################################
    #   Data Manager - Subclass Implementation    #
    ###############################################
    def _get_enabled_services(self):
        """Provides a list of the Services for the Component"""
        return ['httpd'] if utils.is_management_system() else []

    def _get_certificate_authority_info(self):
        """Provides the info to create the Component's CA Certificate"""
        ca_info = dict(component='', extensions=['x509'], altnamexts=['x509'])
        ca_info['privkey'] = '/etc/pki/tls/private/powervc.key'
        ca_info['cacert'] = '/etc/pki/tls/certs/powervc.crt'
        return ca_info

    def _get_remotehost_certificate_info(self):
        """Provides the info to create the Remote Host Signed Certificates"""
        cacert = ('/etc/pki/tls/certs/powervc.crt',
                  '/etc/pki/tls/certs/powervc.crt')
        return dict(cacert=cacert)

    def _configure_firewall(self, config):
        """Helper method to configure the Firewall for PowerVC"""
        icmp = '-m icmp -p icmp --icmp-type'
        iptb = 'iptables'
        inbound = mgr.SERVICE_PORTS.values() + [80,  # HTTP
                                                443,  # HTTPS
                                                5470,  # bumblebee
                                                5671,  # rabbitmq
                                                35357]  # keystone-admin
        outbound = [22,  # SSH
                    389,  # LDAP
                    636]  # LDAPS
        # Need to create the IP Tables files in /etc/sysconfig
        for afile in ['/etc/sysconfig/iptables', '/etc/sysconfig/ip6tables']:
            if not utils.file_exists(afile):
                utils.create_file(afile)
        # Disable the FirewallD services and enable IP Tables
        utils.run_command('/usr/bin/systemctl disable firewalld')
        utils.run_command('/usr/bin/systemctl stop firewalld')
        utils.run_command('/usr/bin/systemctl start iptables')
        utils.run_command('/usr/bin/systemctl start ip6tables')
        utils.run_command('/usr/bin/systemctl enable iptables')
        utils.run_command('/usr/bin/systemctl enable ip6tables')
        # If this is IPv6, then we need to run different commands
        if config.get('ip_version') == '6':
            icmp = '-m icmpv6 -p ipv6-icmp --icmpv6-type'
            iptb = 'ip6tables'
        # We need to open up ping and syntax is a little different than others
        utils.run_command(
            '/sbin/%s -I INPUT %s echo-request -j ACCEPT' % (iptb, icmp))
        utils.run_command(
            '/sbin/%s -I OUTPUT %s echo-reply -j ACCEPT' % (iptb, icmp))
        if utils.get_offering() == 'powervm':
            outbound.append(443)  # EMC VNX
            outbound.append(5989)  # EMC VMAX
            outbound.append(7778)  # XIV
            outbound.append(12443)  # HMC
        # Loop through each of the ports opening up incoming/outgoing traffic
        for port in inbound:
            utils.run_command('/sbin/%s -I INPUT'
                              ' -m state --state NEW,ESTABLISHED'
                              ' -p tcp --dport %s -j ACCEPT' % (iptb, port))
            utils.run_command('/sbin/%s -I OUTPUT'
                              ' -m state --state ESTABLISHED'
                              ' -p tcp --sport %s -j ACCEPT' % (iptb, port))
        for port in outbound:
            utils.run_command('/sbin/%s -I INPUT'
                              ' -m state --state ESTABLISHED'
                              ' -p tcp --sport %s -j ACCEPT' % (iptb, port))
            utils.run_command('/sbin/%s -I OUTPUT'
                              ' -m state --state NEW,ESTABLISHED'
                              ' -p tcp --dport %s -j ACCEPT' % (iptb, port))
        # Save the IP Tables changes and restart the service
        utils.run_command('/sbin/service %s save' % iptb)
        utils.run_command('/sbin/service %s restart' % iptb)

    ###############################################
    #    Internal Data Manager Helper Methods     #
    ###############################################
    def __update_httpd_conf_file(self):
        """Helper method to update the httpd.conf file"""
        data = '<Directory "/var/www/html/powervc">\n'
        data += '  Options FollowSymLinks\n  AllowOverride FileInfo\n'
        data += '  Order allow,deny\n  Allow from all\n</Directory>\n'
        grepdata = utils.read_from_file(HTTP_CONF)
        # If this is PowerKVM, then set the Timeout and KeepAlive values
        if utils.get_offering() == 'powerkvm':
            # If the Timeout exists, replace it, otherwise add it to the file
            if grepdata.find('Timeout ') > 0:
                utils.sed_replace(HTTP_CONF, '^Timeout.*', 'Timeout 300')
            else:
                utils.append_to_file(HTTP_CONF, 'Timeout 300\n')
            # If the Timeout exists, replace it, otherwise add it to the file
            if grepdata.find('KeepAlive ') > 0:
                utils.sed_replace(HTTP_CONF, '^KeepAlive.*', 'KeepAlive Off')
            else:
                utils.append_to_file(HTTP_CONF, 'KeepAlive Off\n')
        # Append the info to the httpd.conf file to enable .htaccess
        if grepdata.find('/var/www/html/powervc') < 0:
            utils.append_to_file(HTTP_CONF, data)
        # set ServerTokens Prod to limit the information
        # emanating from the server in its response headers
        utils.append_to_file(HTTP_CONF, 'ServerTokens Prod\n')
        # Comment out the CustomLog/AccessLog from the combined section
        utils.run_command("/bin/sed -i -e '/CustomLog.*combined/s/ Custom"
                          "Log/ #CustomLog/' " + HTTP_CONF)
        # Enable symbolic links so Keystone can point to the Python packages
        utils.run_command("/bin/sed -i -e 's/Options None/"
                          "Options FollowSymLinks/' " + HTTP_CONF)
        utils.run_command("/bin/sed -i -e 's/Options Indexes MultiViews "
                          "FollowSymLinks/Options MultiViews FollowSym"
                          "Links/' " + HTTP_CONF)

    def __update_httpd_ssl_conf_file(self):
        """Helper method to update the httpd ssl.conf file"""
        cfg = utils.get_config_data()
        host = cfg['host_name'] if cfg['ip_version'] == '6' else cfg['mgmt_ip']
        # Override the Default SSL Settings for the Apache HTTP Server
        utils.run_command("/bin/sed -i -e 's/^#SSLProtocol/##SSLProtocol/' -e "
                          "'s/^SSLProtocol/#SSLProtocol/' -e '/^#SSLProtocol/a"
                          "\SSLProtocol +TLSv1.2' " + HTTP_SSL_CONF)
        utils.run_command("/bin/sed -i -e 's/^#SSLCipherSuite/##SSLCipherSuite"
                          "/' -e 's/^SSLCipherSuite/#SSLCipherSuite/' -e '/^#"
                          "SSLCipherSuite/a\SSLCipherSuite "
                          "ECDHE-RSA-AES128-GCM-SHA256:"
                          "ECDHE-ECDSA-AES128-GCM-SHA256:"
                          "ECDHE-RSA-AES256-GCM-SHA384:"
                          "ECDHE-ECDSA-AES256-GCM-SHA384:"
                          "ECDHE-RSA-AES128-SHA256:"
                          "ECDHE-ECDSA-AES128-SHA256:"
                          "ECDHE-RSA-AES256-SHA384:"
                          "ECDHE-ECDSA-AES256-SHA384:"
                          "DHE-RSA-AES128-GCM-SHA256:"
                          "DHE-RSA-AES256-GCM-SHA384:"
                          "DHE-RSA-AES128-SHA256:"
                          "DHE-RSA-AES256-SHA256' " + HTTP_SSL_CONF)
        utils.run_command("/bin/sed -i -e 's/^#SSLHonorCipherOrder "
                          "/##SSLHonorCipherOrder/' -e "
                          "'s/^SSLHonorCipherOrder/#SSLHonorCipherOrder/'"
                          " -e '/^#SSLHonorCipherOrder/a"
                          "\SSLHonorCipherOrder on' " + HTTP_SSL_CONF)
        utils.run_command("/bin/sed -i -e '/ssl_access_log/s/^"
                          "TransferLog/#TransferLog/' " + HTTP_SSL_CONF)
        utils.run_command("/bin/sed -i -e '/ssl_request_log/s/^"
                          "CustomLog/#CustomLog/' " + HTTP_SSL_CONF)
        utils.run_command("/bin/sed -i -e '/SSLCertificateFile/s"
                          "/localhost.crt/powervc.crt/' " + HTTP_SSL_CONF)
        utils.run_command("/bin/sed -i -e '/SSLCertificateKeyFile/s"
                          "/localhost.key/powervc.key/' " + HTTP_SSL_CONF)
        # New Defaults needed for Apache HTTP Server 2.4
        utils.run_command("/bin/sed -i -e '/^SSLEngine/a SSLProxyCheckPeerName"
                          " off' -e '/^SSLEngine/a SSLProxyCheckPeerCN off' "
                          "-e '/^SSLEngine/a SSLProxyCheckPeerExpire off' " +
                          HTTP_SSL_CONF)
        utils.run_command("/bin/sed -i -e '/ServerName/s/.*ServerName"
                          ".*/ServerName %s/' " % host + HTTP_SSL_CONF)

    def __update_powervc_httpd_conf_file(self):
        """Helper method to update the ports in the PowerVC httpd conf files"""
        config, cfile = utils.get_config_data(), '/etc/httpd/conf.d/*.conf'
        # Update the Mgmt/Local IP in the component configuration files
        utils.sed_replace(cfile, 'POWERVC_SERVER_IP', config['mgmt_ip'])
        utils.sed_replace(cfile, 'POWERVC_LOCALHOST', 'localhost')
        # Now put all of port numbers into the component configuration files
        for comp, port in dict(mgr.SERVICE_PORTS, bumblebee=5470).iteritems():
            comp = 'ttv_validation' if comp == 'validator' else comp
            prefix = ('powervc_' + comp) if comp != 'ttv_validation' else comp
            utils.sed_replace(cfile, '%s_PORT' % prefix.upper(), port)
