# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from powervc_oslo.config import data_mgr as mgr
from powervc_oslo.config import data_utils as utils

MQ_CONF = '/etc/rabbitmq/rabbitmq.config'


######################################################
#            MQ Data Manager Definition              #
######################################################
class MqDataManager(mgr.BaseDataManager):

    def get_component_name(self):
        """Provides the name of the Component providing the Data Manager"""
        return 'rabbitmq'

    def is_supported_environment(self):
        """Specifies whether the Component is supported in this Environment"""
        return True

    ###############################################
    #     Data Manager - Configure Interface      #
    ###############################################
    def configure_component(self):
        """Performs Configuration specific to the given Component"""
        super(MqDataManager, self).configure_component()
        mqpwd = utils.get_config_data().get('mq_password')
        mqctl = '/bin/su - rabbitmq -c "/usr/lib/rabbitmq/bin/rabbitmqctl %s"'
        # We only want to do this configuration if on the Management System
        if not utils.is_management_system():
            return
        # Write out the rabbitmq.conf File for our SSL Settings
        self.__write_rabbitmq_conf_file()
        # Restart the rabbitmq-server so we can update the Users
        utils.run_command('/bin/systemctl restart rabbitmq-server')
        # Try to delete the Guest user that RabbitMQ sets up by default
        utils.run_command(mqctl % 'delete_user guest')
        # Next try to add the powervc_mq user or update its password
        if utils.run_command(mqctl % 'add_user powervc_mq %s' % mqpwd) != 0:
            utils.run_command(
                mqctl % 'change_password powervc_mq %s' % mqpwd, throw=True)
        # Next try to update the permissions on the create powervc_mq user
        cmd = '/usr/lib/rabbitmq/bin/rabbitmqctl '
        cmd += 'set_permissions powervc_mq ".*" ".*" ".*"'
        utils.run_command(['/bin/su', '-', 'rabbitmq', '-c', cmd], throw=True)
        # Restart the rabbitmq-server again for changes to take effect
        utils.run_command('/bin/systemctl restart rabbitmq-server')

    def unconfigure_component(self, upgrading):
        """Performs Configuration specific to the given Component"""
        super(MqDataManager, self).unconfigure_component(upgrading)
        # If this is the KVM Host, need to cleanup the certificates
        if not utils.is_management_system():
            utils.remove_directory('/etc/pki/messages')

    def set_user_passwords(self):
        """Sets/Generates the Passwords for the Component's Users"""
        # Need to generate the powervc_mq Password even though not a User
        utils.get_config_data().setdefault(
            'mq_password', utils.generate_password())

    ###############################################
    #   Data Manager - Subclass Implementation    #
    ###############################################
    def _get_enabled_services(self):
        """Provides a list of the Services for the Component"""
        return ['rabbitmq-server'] if utils.is_management_system() else []

    def _get_certificate_authority_info(self):
        """Provides the info to create the Component's CA Certificate"""
        ca_info = dict(component='messaging', altnamexts=[],
                       extensions=['ca', 'client', 'server'])
        ca_info['privkey'] = '/etc/pki/messages/ca/private/cakey.pem'
        ca_info['cacert'] = '/etc/pki/messages/ca/cacert.pem'
        return ca_info

    def _get_mgmtsys_certificate_info(self):
        """Provides the info to create the MgmtSys Signed Certificates"""
        csr1 = dict(keyfile='/etc/pki/messages/server/private/key.pem',
                    certfile='/etc/pki/messages/server/cert.pem',
                    group='rabbitmq', exttype='server')
        csr2 = dict(keyfile='/etc/pki/messages/pvcservices/private/key.pem',
                    certfile='/etc/pki/messages/pvcservices/cert.pem',
                    group='pvcservices', exttype='client')
        return dict(csrs=[csr1, csr2])

    def _get_remotehost_certificate_info(self):
        """Provides the info to create the Remote Host Signed Certificates"""
        cacert = ('/etc/pki/messages/ca/cacert.pem',
                  '/etc/pki/messages/cacert.pem')
        csr = dict(certfile='clientcert.pem',
                   keyfile='clientkey.pem', exttype='client')
        return dict(cacert=cacert, certdir='/etc/pki/messages',
                    group='pvcservices', csrs=[csr])

    def _get_service_password_properties(self):
        """Provides a list of the configuration properties for passwords"""
        return [('mq', '/etc/nova/nova.conf',
                 'oslo_messaging_rabbit|DEFAULT', 'rabbit_password')]

    def _temporary_openstack_patches(self):
        """Perform any temporary openstack patching for the Component"""
        pass

    ###############################################
    #    Internal Data Manager Helper Methods     #
    ###############################################
    def __write_rabbitmq_conf_file(self):
        """Helper method to write the rabbitmq.conf file"""
        data = '[\n {rabbit, [\n  {tcp_listeners, [] },\n'
        data += '  {ssl_handshake_timeout, 20000},\n'
        data += '  {ssl_listeners, [5671]},\n  {ssl_options, [\n'
        data += '   {versions, [\'tlsv1.2\']},\n'
        data += '   {cacertfile, "/etc/pki/messages/ca/cacert.pem"},\n'
        data += '   {certfile, "/etc/pki/messages/server/cert.pem"},\n'
        data += '   {keyfile, "/etc/pki/messages/server/private/key.pem"},\n'
        data += '   {verify, verify_peer},   {fail_if_no_peer_cert, true},\n'
        data += '   {ciphers,[{ecdhe_rsa,aes_256_cbc,sha384}]}]}\n ]}\n].'
        # Write out the rabbitmq.conf file for our SSL Settings
        utils.write_to_file(MQ_CONF, data)
        # Change the owner of the file to the RabbitMQ user
        utils.chown(MQ_CONF, 'rabbitmq', 'rabbitmq')
