# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from abc import ABCMeta
from abc import abstractmethod
import copy
from datetime import datetime
from datetime import timedelta
try:
    from eventlet.greenthread import sleep
except:
    from time import sleep
from hashlib import md5
import logging
import os
try:
    from pwd import getpwuid
except ImportError:
    pass
import re
import requests
import requests.exceptions as exceptions
import six
import threading
from urlparse import urlparse
from lxml import etree
from xml.sax import saxutils

from powervc_oslo.k2 import synchronized

__import__('powervc_oslo.k2.k2_verification')


cachesize = 1024
# TODO: is anything else not cacheable?
uncacheable = ['/rest/api/web/', '/rest/api/pcm/',
               '/quick', '/search/', '?detail', '/jobs']
# TODO: invalidate SSP cache based on ClusterLULinkedClone jobs (see 17528)

k2basepath = '/rest/api/'
_logon_path = '/rest/api/web/Logon'
type_template = 'application/vnd.ibm.powervm.%s+xml; type=%s'
logonrequest_template = (
    '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n' +
    '<LogonRequest xmlns="http://www.ibm.com/xmlns/systems/power/firmware/' +
    'web/mc/2012_10/" schemaVersion="V1_0">\n' +
    '    <UserID>%(userid)s</UserID>\n' +
    '    <Password>%(passwd)s</Password>\n' +
    '</LogonRequest>').decode('UTF-8')

atom_ns = 'http://www.w3.org/2005/Atom'
xsi_ns = 'http://www.w3.org/2001/XMLSchema-instance'
web_ns = 'http://www.ibm.com/xmlns/systems/power/firmware/web/mc/2012_10/'
uom_ns = 'http://www.ibm.com/xmlns/systems/power/firmware/uom/mc/2012_10/'
uom_base_ns = 'http://www.ibm.com/xmlns/systems/power/firmware/uom/mc'

uuid_search_str = ('[A-Fa-f0-9]{8}-[A-Fa-f0-9]{4}-[A-Fa-f0-9]{4}-' +
                   '[A-Fa-f0-9]{4}-[A-Fa-f0-9]{12}')

K2_SCHEMA_VERSION = 'V1'

# setup logging
LOG = logging.getLogger(__name__)

# Register the namespaces we'll use
etree.register_namespace('atom', atom_ns)
etree.register_namespace('xsi', xsi_ns)
etree.register_namespace('web', web_ns)
etree.register_namespace('uom', uom_ns)

QName = etree.QName


class K2Error(Exception):

    """Error on K2 Operator method invocation"""

    def __init__(self, msg, k2response=None):
        """ Initializes K2Error with optional `k2response` object. """
        self.k2response = k2response
        self.orig_k2response = None
        super(K2Error, self).__init__(msg)


class K2ConnectionError(K2Error):

    """Connection Error on K2 Operator method invocation"""


class K2SSLError(K2Error):

    """SSL Error on K2 Operator method invocation"""


class K2TimeoutError(K2Error):

    """Timeout Error on K2 Operator method invocation"""


class K2HttpError(K2Error):

    """HTTP Error on K2 Operator method invocation"""

    def __init__(self, msg, k2response):
        """ Initializes K2HttpError with required `k2response` object. """
        super(K2HttpError, self).__init__(msg, k2response=k2response)


class K2AtomError(K2Error):

    """Atom Error on K2 Operator method invocation"""

    def __init__(self, msg, k2response):
        """ Initializes K2AtomError with required `k2response` object. """
        super(K2AtomError, self).__init__(msg, k2response=k2response)


class K2Session(object):

    """Responsible for K2 session management"""

    def __init__(self, host, username, password, auditmemento=None,
                 protocol='https', port=12443, timeout=60,
                 certpath='/opt/ibm/powervc/data/certs/', certext='.crt'):
        self.username = username
        self.password = password

        audmem = auditmemento
        if not audmem:
            try:
                audmem = getpwuid(os.getuid())[0]
            except Exception:
                LOG.warn("Calculating default audit memento failed, using "
                         "'default'.")
                audmem = 'default'
        self.auditmemento = audmem

        self.protocol = protocol
        if protocol == 'http':
            LOG.warn('Unencrypted communication with HMC! ' +
                     'Revert configuration to https.')
        if self.protocol not in ['https', 'http']:
            raise ValueError('Invalid protocol "%s"' % self.protocol)
        self.host = host
        self.port = port
        # support IPv6 addresses
        if self.host[0] != '[' and ':' in self.host:
            self.dest = '%s://[%s]:%i' % (self.protocol, self.host, self.port)
        else:
            self.dest = '%s://%s:%i' % (self.protocol, self.host, self.port)

        self.timeout = timeout
        self.certpath = certpath
        self.certext = certext
        if self.certpath is None:
            self.verify = False
        else:
            certfile = certpath + re.sub('[:.]', '_', host) + certext
            if os.path.exists(certfile):
                self.verify = certfile  # will use specific cert file
            else:
                self.verify = True  # will use system CA bundle

        self._lock = threading.RLock()
        self._logged_in = False
        self._relogin_unsafe = False
        self._sessToken = None
        self.schema_version = None
        self._eventlistener = None
        self._logon()

    def __del__(self):
        try:
            if self._eventlistener:
                self._eventlistener.shutdown()
        finally:
            self._logoff()

    def get_event_listener(self):
        if not self._eventlistener:
            # spawn a separate session for listening to events. If we used the
            # same session, K2Session.__del__() would never be called
            # TODO: investigate whether we can find a way to share the session
            LOG.info("setting up event listener for %s" % self.host)
            listen_session = K2Session(self.host,
                                       self.username,
                                       self.password,
                                       auditmemento=self.auditmemento,
                                       protocol=self.protocol,
                                       port=self.port,
                                       timeout=self.timeout,
                                       certpath=self.certpath,
                                       certext=self.certext)
            self._eventlistener = K2EventListener(listen_session)
        return self._eventlistener

    def request(self, method, path, headers={}, body='', sensitive=False,
                timeout=-1, auditmemento=None, relogin=True,
                login=False, filehandle=None, chunksize=65536):
        """Send an HTTP/HTTPS request to a K2 interface"""

        url = self.dest + path

        session = requests.Session()
        session.verify = self.verify
        if isinstance(session.verify, six.string_types):
            # Override the Protocol so we use our own HTTPSConnectionPool
            # See k2_verification
            if url.startswith('https://'):
                url = 'httpsk2://' + url[8:]
            # Since we overrode the protocol for the Pool,
            # we need to register the HTTP adapter
            session.mount("httpsk2://", requests.adapters.HTTPAdapter())

        # if timeout isn't specified, use session default
        if timeout == -1:
            timeout = self.timeout

        if auditmemento:
            headers['X-Audit-Memento'] = auditmemento
        else:
            headers['X-Audit-Memento'] = self.auditmemento

        isupload = False
        isdownload = False
        if filehandle:
            if method in ['PUT', 'POST']:
                isupload = True
            elif method in ['GET']:
                isdownload = True
            else:
                raise ValueError('unexpected filehandle on %s request'
                                 % method)

        if isupload:
            LOG.debug('sending %s %s headers=%s body=<file contents>' %
                      (method, url,
                       headers if not sensitive else "<sensitive>"))
        else:
            LOG.debug('sending %s %s headers=%s body=%s' %
                      (method, url,
                       headers if not sensitive else "<sensitive>",
                       body if not sensitive else "<sensitive>"))

        # Add X-API-Session header after above so it's not printed in log
        sessTokenTry = None
        if not login:
            with self._lock:
                assert(self._sessToken), "missing session token"
                headers['X-API-Session'] = self._sessToken
                sessTokenTry = self._sessToken

        response = None
        try:
            if isupload:

                def chunkreader():
                    while True:
                        d = filehandle.read(chunksize)
                        if not d:
                            break
                        yield d

                response = session.request(method, url, data=chunkreader(),
                                           headers=headers, timeout=timeout)
            elif isdownload:
                response = session.request(method, url, stream=True,
                                           headers=headers, timeout=timeout)
            else:
                response = session.request(method, url, data=body,
                                           headers=headers, timeout=timeout)
        except exceptions.SSLError as e:
            msg = '%s for %s %s: %s' % (e.__class__.__name__, method, url, e)
            LOG.warn(msg)
            raise K2SSLError(msg)
        except exceptions.ConnectionError as e:
            msg = '%s for %s %s: %s' % (e.__class__.__name__, method, url, e)
            LOG.warn(msg)
            raise K2ConnectionError(msg)
        except exceptions.Timeout as e:
            msg = '%s for %s %s: %s' % (e.__class__.__name__, method, url, e)
            LOG.warn(msg)
            raise K2TimeoutError(msg)
        except Exception as e:
            LOG.exception('Unexpected error for %s %s' % (method, url))
            raise K2Error('Unexpected error: %s for %s %s: %s' %
                          (e.__class__.__name__, method, url, e))
        finally:
            session.close()

        # remove X-API-Session header so it won't get printed
        if not login:
            try:
                del headers['X-API-Session']
            except KeyError:
                # something modifying the submitted request headers??
                # TODO: why does this happen and what else may result?
                pass

        LOG.debug('result: %s (%s) for %s %s' %
                  (response.status_code, response.reason, method, url))
        LOG.debug('response headers: %s' %
                  (response.headers if not sensitive else "<sensitive>"))

        if response.status_code in [204, 304]:
            return K2Response(method, path, response.status_code,
                              response.reason, response.headers,
                              reqheaders=headers, reqbody=body)
        else:
            LOG.debug('response body:\n%s' %
                      (response.text.encode('utf8') if not sensitive
                       else "<sensitive>"))

        # re-login processing
        if response.status_code == 401:
            LOG.debug('processing HTTP 401')

            with self._lock:
                if not relogin:
                    LOG.debug('requester specified no re-login')
                elif self._relogin_unsafe:
                    LOG.warn('Re-login has been deemed unsafe. ' +
                             'This instance should no longer be used.')
                else:
                    if self._sessToken != sessTokenTry:
                        LOG.debug('someone else handled re-login for %s'
                                  % self.host)
                    else:
                        self._logged_in = False

                        LOG.info('will re-login %s' % self.host)
                        try:
                            self._logon()
                        except K2Error as e:
                            if e.k2response:
                                if e.k2response.status == 401:
                                    # can't continue re-login attempts lest we
                                    # lock the account
                                    self._relogin_unsafe = True
                                    LOG.warn('Re-login 401, response body:\n%s'
                                             % e.k2response.body)
                                else:
                                    # safe to try re-login again in this case
                                    LOG.warn('Re-login failed, resp body:\n%s'
                                             % e.k2response.body)
                            else:
                                # safe to try re-login again in this case
                                LOG.warn('Re-login failed:\n%s' % str(e))
                            e.orig_k2response = K2Response(
                                method, path, response.status_code,
                                response.reason, response.headers,
                                reqheaders=headers, reqbody=body,
                                body=response.text)
                            _unmarshal_k2httperror(e.orig_k2response)
                            raise e

                    # Retry the original request
                    try:
                        return self.request(method, path, headers, body,
                                            sensitive=sensitive,
                                            timeout=timeout, relogin=False)
                    except K2HttpError as e:
                        if e.k2response.status == 401:
                            # This is a special case... normally on a 401 we
                            # would retry login, but we won't here because
                            # we just did that... Handle it specially.
                            LOG.warn('Re-attempt failed with another 401, ' +
                                     'response body:\n%s' % e.k2response.body)
                            raise K2Error('suspicious HTTP 401 response ' +
                                          'for %s %s: token is brand new'
                                          % (method, path))
                        raise

        if not isdownload:
            k2resp = K2Response(method, path, response.status_code,
                                response.reason, response.headers,
                                reqheaders=headers, reqbody=body,
                                body=response.text)

        if 200 <= response.status_code < 300:
            if isdownload:
                for chunk in response.iter_content(chunksize):
                    filehandle.write(chunk)
                k2resp = K2Response(method, path, response.status_code,
                                    response.reason, response.headers,
                                    reqheaders=headers, reqbody=body)
            return k2resp
        else:
            if isdownload:
                errtext = ''
                for chunk in response.iter_content(chunksize):
                    errtext += chunk
                k2resp = K2Response(method, path, response.status_code,
                                    response.reason, response.headers,
                                    reqheaders=headers, reqbody=body,
                                    body=errtext)
            errmsg = 'HTTP error for %s %s: %s (%s)' % (method, path,
                                                        response.status_code,
                                                        response.reason)
            _unmarshal_k2httperror(k2resp)
            raise K2HttpError(errmsg, k2resp)

    def _logon(self):
        LOG.info("logging on %s" % self.host)
        headers = {
            'Accept': type_template % ('web', 'LogonResponse'),
            'Content-Type': type_template % ('web', 'LogonRequest')
        }
        passwd = saxutils.escape(self.password)
        body = logonrequest_template % {'userid': self.username,
                                        'passwd': passwd}
        # Convert body to UTF-8 encoded binary string from unicode string
        # Without this, the socket code will implicitly convert as ASCII
        body = body.encode('utf8')

        try:
            # relogin=False to prevent mult. attempts / account lockout
            k2resp = self.request('PUT', _logon_path, headers=headers,
                                  body=body, sensitive=True,
                                  relogin=False, login=True)
        except K2Error as e:
            if e.k2response:
                # strip out sensitive data
                e.k2response.reqbody = "<sensitive>"
            if isinstance(e, K2HttpError):
                # clarify the error message
                raise K2HttpError('HTTP error on Logon: %s (%s)' %
                                  (e.k2response.status, e.k2response.reason),
                                  e.k2response)
            raise

        # parse out X-API-Session value
        root = etree.fromstring(k2resp.body.encode('utf8'),
                                parser=etree.XMLParser(strip_cdata=False,
                                                       encoding='utf-8'))
        with self._lock:
            tok = root.findtext('./{%s}X-API-Session' % web_ns)
            if not tok:
                k2resp.reqbody = "<sensitive>"
                msg = "failed to parse session token from K2 response"
                LOG.error((msg + ' body= %s') % k2resp.body)
                raise K2Error(msg, k2response=k2resp)
            self._sessToken = tok
            self._logged_in = True
            # Currently only support schema version starting with 'V1'
            if root.get('schemaVersion').startswith(K2_SCHEMA_VERSION):
                self.schema_version = root.get('schemaVersion')
            else:
                self.schema_version = self.get_schema_version()

    def get_schema_version(self):
        path = k2basepath + 'uom/RESTSchemaVersion'
        headers = {}
        headers['Accept'] = 'application/atom+xml'

        k2resp = self.request('GET', path, headers=headers)
        k2resp._unmarshalAtom()

        # find all the other supported schema version for this HMC
        results = k2resp.entry.element.findall('./SupportedRESTSchemaVersion')
        # sort the returned list of all supported schema version
        # from newest schema version to oldest schema version
        results.sort(reverse=True)

        # return latest supported schema version
        for version in results:
            if version.gettext().startswith(K2_SCHEMA_VERSION):
                return version.gettext()

    def _logoff(self):
        with self._lock:
            if not self._logged_in:
                return
            LOG.info("logging off %s" % self.host)
            try:
                # relogin=False to prevent mult. attempts / account lockout
                self.request('DELETE', _logon_path, relogin=False)
            except Exception:
                LOG.exception('problem logging off')
                pass
            self._logged_in = False
            # this should only ever be called when K2Session has gone out of
            # scope, but just in case someone calls it directly while requests
            # are in flight, set _relogin_unsafe so that those requests won't
            # enter relogin processing when they get an HTTP 401.
            self._relogin_unsafe = True


class K2Operator(object):

    """K2 Operator for HMC remote management"""

    # TODO: way to retrieve cache timestamp along with / instead of data?

    def __init__(self, session, use_cache=True):
        self.session = session
        self._cache = None
        self._eventlistener = None
        self._refreshtime4path = {}
        if use_cache:
            self._cache = _K2Cache(session.host)
            # use events even with pre-810 HMC builds... events wouldn't
            # always be sent when they should, but they should be
            # trustworthy when they are sent
            try:
                self._eventlistener = session.get_event_listener()
            except:
                LOG.exception('failed to register for events')
            if self._eventlistener is not None:
                self._evthandler = _EventHandler(self._cache)
                self._eventlistener.subscribe(self._evthandler)

    def __del__(self):
        if self._eventlistener is not None:
            self._eventlistener.unsubscribe(self._evthandler)

    def create(self, element, rootType, rootId=None, childType=None,
               childId=None, suffixType=None, suffixParm=None, detail=None,
               service='uom', contentService=None, timeout=-1,
               auditmemento=None, xag=[], sensitive=False):
        """Create a new resource.
        Will build the URI path using the provided arguments."""
        K2Operator._validate('create', rootType, rootId, childType, childId,
                             suffixType, suffixParm, detail)
        path = K2Operator.buildpath(service, rootType, rootId, childType,
                                    childId, suffixType, suffixParm, detail,
                                    xag=xag)
        return self.createbypath(element, path, contentService=contentService,
                                 timeout=timeout, auditmemento=auditmemento,
                                 sensitive=sensitive)

    def createjob(self, job, rootType, rootId=None, childType=None,
                  childId=None, timeout=-1, auditmemento=None,
                  sensitive=False):
        if not job.gettag() == 'JobRequest':
            raise ValueError('job must be a JobRequest element')
        op = job.findtext('./RequestedOperation/OperationName')
        if not op:
            raise ValueError('JobRequest is missing OperationName')
        return self.create(job, rootType, rootId, childType, childId,
                           suffixType='do', suffixParm=op,
                           contentService='web', timeout=timeout,
                           auditmemento=auditmemento, sensitive=sensitive)

    def createbypath(self, element, path, contentService=None, timeout=-1,
                     auditmemento=None, sensitive=False):
        """Create a new resource where the URI path is already known"""
        path = _sanitize_path(path)
        m = re.search('%s(\w+)/(\w+)' % k2basepath, path)
        if not m:
            raise ValueError('path=%s is not a K2 api reference' % path)
        if not contentService:
            contentService = m.group(1)
        headers = {}
        headers['Accept'] = 'application/atom+xml'
        if (re.search('/do/', path)):
            headers['Content-Type'] = type_template % (contentService,
                                                       'JobRequest')
            headers['X-HMC-Schema-Version'] = 'V1_0'
        else:
            p = path.split('\?', 1)[0]  # strip off details, if present
            headers['Content-Type'] = type_template % (contentService,
                                                       p.rsplit('/', 1)[1])
            headers['X-HMC-Schema-Version'] = self.session.schema_version
        # Encode body as UTF-8 for unicode support
        # Without this, the socket code will implicitly convert as ASCII
        body = element.toxmlstring().encode('utf8')
        k2resp = self.session.request('PUT', path, headers=headers,
                                      body=body,
                                      timeout=timeout,
                                      auditmemento=auditmemento,
                                      sensitive=sensitive)
        if k2resp.body is not None:
            k2resp.body = k2resp.body.encode('utf8')
        k2resp_to_cache = None
        is_cacheable = self._cache and not any(p in path for p in uncacheable)
        if is_cacheable:
            # TODO: are there cases where PUT response differs from GET?
            # TODO: only cache this when we don't have event support?
            # If events are supported for this element, they should quickly
            # invalidate this cache entry. But if events aren't available, or
            # could be missed, caching the response may be useful.
            # We will cache unmarshaled to reduce cache overhead, but we do
            # need to unmarshal to determine all paths to use as cache key, so
            # save this off (unmarshaled) to cache a little later.
            k2resp_to_cache = copy.deepcopy(k2resp)
        k2resp._unmarshalAtom()
        if k2resp_to_cache:
            paths = _determine_paths(k2resp)
            # we're creating, so what we send is the full entity
            # so it should be safe to strip off xag (no xag = all)
            new_path = (k2resp.reqpath.split('?', 1)[0] + '/' +
                        k2resp.entry.properties['id'])
            self._cache.set(new_path, paths, k2resp_to_cache)
            # need to invalidate the feeds containing this entry
            feed_paths = self._cache.get_feed_paths(new_path)
            for feed_path in feed_paths:
                self._cache.remove(feed_path)
        return k2resp

    def read(self, rootType, rootId=None, childType=None, childId=None,
             suffixType=None, suffixParm=None, detail=None, service='uom',
             etag=None, timeout=-1, auditmemento=None, age=-1, xag=[],
             sensitive=False):
        """Retrieve an existing resource.
        Will build the URI path using the provided arguments."""
        K2Operator._validate('read', rootType, rootId, childType, childId,
                             suffixType, suffixParm, detail)
        path = K2Operator.buildpath(service, rootType, rootId, childType,
                                    childId, suffixType, suffixParm, detail,
                                    xag=xag)
        return self.readbypath(path, etag, timeout=timeout,
                               auditmemento=auditmemento, age=age,
                               sensitive=sensitive)

    def readjob(self, jobId, etag=None, timeout=-1, auditmemento=None,
                sensitive=False):
        return self.read('jobs', jobId, etag=etag, timeout=timeout,
                         auditmemento=auditmemento, sensitive=sensitive)

    def readjobs(self, rootType=None, rootId=None, childType=None,
                 childId=None, detail=None, etag=None, timeout=-1,
                 auditmemento=None, sensitive=False):
        return self.read(rootType, rootId, childType, childId,
                         suffixType='jobs', detail=detail, etag=etag,
                         timeout=timeout, auditmemento=auditmemento,
                         sensitive=sensitive)

    def readbyhref(self, href, suffixType=None, suffixParm=None, detail=None,
                   etag=None, timeout=-1, auditmemento=None, age=-1,
                   sensitive=False):
        """Retrieve an existing resource based on a link's href"""
        o = urlparse(href)
        hostname_mismatch = (o.hostname.lower() != self.session.host.lower())
        if hostname_mismatch or o.port != self.session.port:
            LOG.debug('href=%s will be modified to use %s:%s' %
                      (href, self.session.host, self.session.port))
        path = K2Operator.extendpath(o.path, suffixType, suffixParm, detail)
        return self.readbypath(path, etag=etag, timeout=timeout,
                               auditmemento=auditmemento, age=age,
                               sensitive=sensitive)

    def readbypath(self, path, etag=None, timeout=-1, auditmemento=None,
                   age=-1, sensitive=False):
        """Retrieve an existing resource where the URI path is already known"""

        @synchronized(path)
        def _locked_refresh(refresh_time, cached_k2resp, etag,
                            etag_from_cache):
            # If the request time is after the last time we started this,
            # then for all intents and purposes the data is current
            _refresh_time = self._refreshtime4path.get(path)
            if (_refresh_time is not None and _refresh_time >= refresh_time):
                # The callers request was queued behind the lock before
                # we started updating, so we know any operations they
                # were expecting should be reflected in the latest data.
                # Re-retrieve from cache
                if etag_from_cache:
                    # this optimization won't work in this case because the
                    # cache has already been refreshed since getting this etag
                    # value from the cache
                    etag = None
                    etag_from_cache = False
                # don't need to check age because we know we just refreshed
                k2resp = self._get_resp_from_cache(path, etag=etag)
                if k2resp:
                    if etag and etag == k2resp.headers.get('etag'):
                        # ETag matches what caller specified, so return an
                        # HTTP 304 (Not Modified) response
                        k2resp.status = 304
                        k2resp.body = ''
                    elif 'atom' in k2resp.reqheaders['Accept']:
                        k2resp._unmarshalAtom()
                    return k2resp

            # Either first to get the lock or found the cache to be empty
            # First, note the time so it can be checked by the next guy
            self._refreshtime4path[path] = datetime.now()
            # Now read from HMC
            k2resp = self._readbypath(path, etag, timeout, auditmemento,
                                      sensitive)
            k2resp_to_cache = None
            if is_cacheable:
                if k2resp.status == 304:
                    # don't want to cache the 304, which has no body
                    # instead, just update the cache ordering
                    self._cache.touch(path)
                    if etag_from_cache:
                        # the caller didn't specify that ETag, so they won't
                        # be expecting a 304 response... return from the cache
                        k2resp = cached_k2resp
                else:
                    # cache unmarshaled to reduce cache overhead
                    # But we need to unmarshal to determine all paths, so
                    # save this off (unmarshaled) to cache a little later
                    k2resp_to_cache = copy.deepcopy(k2resp)
            if 'atom' in k2resp.reqheaders['Accept']:
                k2resp._unmarshalAtom()
            if k2resp_to_cache:
                self._cache.set(path, _determine_paths(k2resp),
                                k2resp_to_cache)
            return k2resp
        # end _locked_refresh

        path = _sanitize_path(path)
        # First, test whether we should be pulling from cache, determined
        # by asking a) is there a cache? and b) is this path cacheable?
        is_cacheable = self._cache and not any(p in path for p in uncacheable)
        k2resp = None
        cached_k2resp = None
        etag_from_cache = False

        if is_cacheable:
            # Attempt to retrieve from cache
            max_age = _get_max_age(path, self._eventlistener is not None,
                                   self.session.schema_version)
            if age == -1 or age > max_age:
                age = max_age

            # if requested entry is not in cache, check for feed in
            # cache and build entry response from that. Requires feed entry
            # etag support, new in 810, unless we're ok with returning a
            # response that doesn't include the etag, and making them request
            # again with age=0 if they really need the etag.
            k2resp = self._get_resp_from_cache(path, age, etag)
            if k2resp:
                if etag and etag == k2resp.headers.get('etag'):
                    # ETag matches what caller specified, so return an
                    # HTTP 304 (Not Modified) response
                    k2resp.status = 304
                    k2resp.body = ''
                elif 'atom' in k2resp.reqheaders['Accept']:
                    k2resp._unmarshalAtom()
            elif not etag:
                # we'll bypass the cache, but if there is a cache entry that
                # doesn't meet the age requirement, we can still optimize
                # our GET request by using its ETag to see if it has changed
                cached_k2resp = self._get_resp_from_cache(path)
                if cached_k2resp:
                    etag = cached_k2resp.headers.get('etag')
                    etag_from_cache = True
        if not k2resp:
            k2resp = _locked_refresh(datetime.now(), cached_k2resp, etag,
                                     etag_from_cache)
        return k2resp

    def _readbypath(self, path, etag, timeout, auditmemento, sensitive):
        m = re.search('%s(\w+)/(\w+)' % k2basepath, path)
        if not m:
            raise ValueError('path=%s is not a K2 api reference' % path)
        headers = {}
        json_search_str = (uuid_search_str + '/quick$' +
                           '|/quick/' +
                           '|.json$')
        if re.search(json_search_str, path):
            headers['Accept'] = 'application/json'
        else:
            headers['Accept'] = 'application/atom+xml'
        if etag:
            headers['If-None-Match'] = etag
        headers['X-HMC-Schema-Version'] = self.session.schema_version
        k2resp = self.session.request('GET', path, headers=headers,
                                      timeout=timeout,
                                      auditmemento=auditmemento,
                                      sensitive=sensitive)
        if k2resp.body is not None:
            k2resp.body = k2resp.body.encode('utf8')
        return k2resp

    def update(self, data, etag, rootType, rootId=None, childType=None,
               childId=None, suffixType=None, service='uom', timeout=-1,
               auditmemento=None, xag=[], sensitive=False):
        """Update an existing resource.
        Will build the URI path using the provided arguments."""
        K2Operator._validate('update', rootType, rootId, childType, childId,
                             suffixType)
        path = K2Operator.buildpath(service, rootType, rootId, childType,
                                    childId, suffixType, xag=xag)
        return self.updatebypath(data, etag, path, timeout=timeout,
                                 auditmemento=auditmemento,
                                 sensitive=sensitive)

    def updatebypath(self, data, etag, path, timeout=-1, auditmemento=None,
                     sensitive=False):
        """Update an existing resource where the URI path is already known"""
        path = _sanitize_path(path)
        try:
            k2resp = self._updatebypath(data, etag, path, timeout,
                                        auditmemento, sensitive)
        except K2HttpError as e:
            if self._cache and e.k2response.status == 412:
                # ETag didn't match
                # see if we need to invalidate entry in cache
                k2resp = self._cache.get(path)
                if k2resp and etag == k2resp.headers.get('etag'):
                    # need to invalidate this in the cache
                    self._cache.remove(path)
                # see if we need to invalidate feed in cache
                # extract the entry uuid
                uuid = _get_req_path_uuid(path)
                # extract feed paths pertaining to the entry
                feed_paths = self._cache.get_feed_paths(path)
                for feed_path in feed_paths:
                    k2resp = self._build_entry_resp(feed_path, uuid)
                    if not k2resp or etag == k2resp.headers.get('etag'):
                        # need to invalidate this in the cache
                        self._cache.remove(feed_path)
                        LOG.debug('Invalidate feed %s for uuid %s' %
                                  (feed_path, uuid))
            raise
        k2resp_to_cache = None
        is_cacheable = self._cache and not any(p in path for p in uncacheable)
        if is_cacheable:
            # TODO: are there cases where POST response differs from GET?
            # TODO: only cache this when we don't have event support?
            # If events are supported for this element, they should quickly
            # invalidate this cache entry. But if events aren't available, or
            # could be missed, caching the response may be useful.
            # We will cache unmarshaled to reduce cache overhead, but we do
            # need to unmarshal to determine all paths to use as cache key, so
            # save this off (unmarshaled) to cache a little later.
            k2resp_to_cache = copy.deepcopy(k2resp)
        k2resp._unmarshalAtom()
        if k2resp_to_cache:
            paths = _determine_paths(k2resp)
            self._cache.set(path, paths, k2resp_to_cache)
            # need to invalidate the feeds containing this entry
            feed_paths = self._cache.get_feed_paths(path)
            for feed_path in feed_paths:
                self._cache.remove(feed_path)
        return k2resp

    def _updatebypath(self, data, etag, path, timeout, auditmemento,
                      sensitive):
        m = re.match('%s(\w+)/(\w+)' % k2basepath, path)
        if not m:
            raise ValueError('path=%s is not a K2 api reference' % path)
        headers = {}
        headers['Accept'] = 'application/atom+xml'
        headers['X-HMC-Schema-Version'] = self.session.schema_version
        if m.group(1) == 'pcm':
            headers['Content-Type'] = 'application/xml'
        else:
            t = path.rsplit('/', 2)[1]
            headers['Content-Type'] = type_template % (m.group(1), t)
        if etag:
            headers['If-Match'] = etag
        if isinstance(data, K2Element):
            body = data.toxmlstring()
        else:
            body = data
        # Convert body to UTF-8 encoded binary string from unicode string
        # Without this, the socket code will implicitly convert as ASCII
        if body is not None:
            body = body.encode('utf8')
        k2resp = self.session.request('POST', path, headers=headers,
                                      body=body, timeout=timeout,
                                      auditmemento=auditmemento,
                                      sensitive=sensitive)
        if k2resp.body is not None:
            k2resp.body = k2resp.body.encode('utf8')
        return k2resp

    def delete(self, rootType, rootId=None, childType=None, childId=None,
               suffixType=None, suffixParm=None, service='uom', etag=None,
               timeout=-1, auditmemento=None):
        """Delete an existing resource.
        Will build the URI path using the provided arguments."""
        K2Operator._validate('delete', rootType, rootId, childType, childId,
                             suffixType, suffixParm)
        path = K2Operator.buildpath(service, rootType, rootId, childType,
                                    childId, suffixType, suffixParm)
        return self.deletebypath(path, etag, timeout=timeout,
                                 auditmemento=auditmemento)

    def deletebyhref(self, href, etag=None, timeout=-1, auditmemento=None):
        """Delete an existing resource based on a link's href"""
        o = urlparse(href)
        hostname_mismatch = (o.hostname.lower() != self.session.host.lower())
        if hostname_mismatch or o.port != self.session.port:
            LOG.debug('href=%s will be modified to use %s:%s' %
                      (href, self.session.host, self.session.port))
        return self.deletebypath(o.path, etag=etag, timeout=timeout,
                                 auditmemento=auditmemento)

    def deletebypath(self, path, etag=None, timeout=-1, auditmemento=None):
        """Delete an existing resource where the URI path is already known"""
        path = _sanitize_path(path)
        try:
            k2resp = self._deletebypath(path, etag, timeout, auditmemento)
        except K2HttpError as e:
            if self._cache and e.k2response.status == 412:
                # ETag didn't match
                # see if we need to invalidate entry in cache
                k2resp = self._cache.get(path)
                if k2resp and etag == k2resp.headers.get('etag'):
                    # need to invalidate this in the cache
                    self._cache.remove(path)
                # see if we need to invalidate feed in cache
                # extract entry uuid
                uuid = _get_req_path_uuid(path)
                # extract feed paths pertaining to the entry
                feed_paths = self._cache.get_feed_paths(path)
                for feed_path in feed_paths:
                    k2resp = self._build_entry_resp(feed_path, uuid)
                    if not k2resp or etag == k2resp.headers.get('etag'):
                        # need to invalidate this in the cache
                        self._cache.remove(feed_path)
                        LOG.debug('Invalidate feed %s for uuid %s' %
                                  (feed_path, uuid))
            raise
        if self._cache is not None:
            # get feed_paths before removing the entry (won't work after)
            feed_paths = self._cache.get_feed_paths(path)
            # need to invalidate this in the cache
            self._cache.remove(path, delete=True)
            # also need to invalidate the feeds containing this entry
            for feed_path in feed_paths:
                self._cache.remove(feed_path)

        return k2resp

    def _deletebypath(self, path, etag, timeout, auditmemento):
        m = re.search('%s(\w+)/(\w+)' % k2basepath, path)
        if not m:
            raise ValueError('path=%s is not a K2 api reference' % path)
        headers = {}
        if etag:
            headers['If-Match'] = etag
        headers['X-HMC-Schema-Version'] = self.session.schema_version
        k2resp = self.session.request('DELETE', path, headers=headers,
                                      timeout=timeout,
                                      auditmemento=auditmemento)
        if k2resp.body is not None:
            k2resp.body = k2resp.body.encode('utf8')
        return k2resp

    def uploadfile(self, k2filedescr, filehandle, chunksize=65536,
                   timeout=-1, auditmemento=None, replacing=False):
        try:
            fileid = k2filedescr.findtext('FileUUID')
            mediatype = k2filedescr.findtext('InternetMediaType')
        except:
            raise ValueError('invalid file descriptor')

        path = k2basepath + 'web/File/contents/' + fileid
        headers = {'Accept': 'application/vnd.ibm.powervm.web+xml',
                   'Content-Type': mediatype}
        headers['X-HMC-Schema-Version'] = self.session.schema_version

        return self.session.request('POST' if replacing else 'PUT',
                                    path, headers=headers, timeout=timeout,
                                    auditmemento=auditmemento,
                                    filehandle=filehandle,
                                    chunksize=chunksize)

    def downloadfile(self, k2filedescr, filehandle, chunksize=65536,
                     timeout=-1, auditmemento=None):
        try:
            fileid = k2filedescr.findtext('FileUUID')
            mediatype = k2filedescr.findtext('InternetMediaType')
        except:
            raise ValueError('invalid file descriptor')

        path = k2basepath + 'web/File/contents/' + fileid
        headers = {'Accept': mediatype}
        headers['X-HMC-Schema-Version'] = self.session.schema_version

        return self.session.request('GET', path, headers=headers,
                                    timeout=timeout,
                                    auditmemento=auditmemento,
                                    filehandle=filehandle,
                                    chunksize=chunksize)

    def buildhref(self, rootType, rootId=None, childType=None, childId=None,
                  suffixType=None, suffixParm=None, detail=None, xag=[],
                  service='uom'):
        p = K2Operator.buildpath(service, rootType, rootId, childType, childId,
                                 suffixType, suffixParm, detail, xag=xag)
        return self.session.dest + p

    def _build_entry_resp(self, feed_path, uuid, etag=None, age=-1):
        """build K2Response based on entry within a cached feed"""
        feed_k2resp = self._cache.get(feed_path, age)
        if not feed_k2resp or not feed_k2resp.body:
            return

        resp = None
        root = etree.fromstring(feed_k2resp.body,
                                parser=etree.XMLParser(strip_cdata=False,
                                                       encoding='utf-8'))
        entry_body, entry_etag = _getEntryAtomFeed(root, uuid)
        if entry_body:
            # generate reqpath based on the feed and including uuid
            reqpath = feed_k2resp.reqpath
            parms_str = ''
            if '?' in feed_k2resp.reqpath:
                reqpath, parms_str = feed_k2resp.reqpath.rsplit('?', 1)

            reqpath = reqpath + '/' + uuid
            if parms_str:
                reqpath = reqpath + '?' + parms_str
            if entry_etag and etag == entry_etag:
                resp = K2Response(feed_k2resp.reqmethod,
                                  reqpath,
                                  304,
                                  feed_k2resp.reason,
                                  feed_k2resp.headers,
                                  body='',
                                  orig_reqpath=feed_k2resp.reqpath)
                LOG.debug('built HTTP 304 resp from cached feed')
            else:
                # should we upate the parameters?
                resp = K2Response(feed_k2resp.reqmethod,
                                  reqpath,
                                  feed_k2resp.status,
                                  feed_k2resp.reason,
                                  feed_k2resp.headers,
                                  feed_k2resp.reqheaders,
                                  body=entry_body,
                                  orig_reqpath=feed_k2resp.reqpath)
                # override the etag in the header
                if entry_etag:
                    resp.headers['etag'] = entry_etag
                elif etag:
                    # if there is no entry etag and etag is specified,
                    # just return
                    return
                LOG.debug('built entry from cached feed etag=%s body=%s' %
                          (entry_etag, entry_body))

        return resp

    @staticmethod
    def buildjobparm(name, value):
        p = K2Element('JobParameter', attrib={'schemaVersion': 'V1_0'},
                      ns=web_ns)
        p.append(K2Element('ParameterName', text=name, ns=web_ns))
        p.append(K2Element('ParameterValue', text=value, ns=web_ns))
        return p

    @staticmethod
    def buildpath(service, rootType, rootId=None, childType=None,
                  childId=None, suffixType=None, suffixParm=None,
                  detail=None, xag=[]):
        path = k2basepath + service + '/' + rootType
        if rootId:
            path += '/' + rootId
            if childType:
                path += '/' + childType
                if childId:
                    path += '/' + childId
        return K2Operator.extendpath(path, suffixType, suffixParm, detail, xag)

    @staticmethod
    def extendpath(basepath, suffixType=None, suffixParm=None, detail=None,
                   xag=[]):
        path = basepath
        if suffixType:
            # operations, do, jobs, cancel, quick, search, ${search-string}
            path += '/' + suffixType
            if suffixParm:
                path += '/' + suffixParm
        if detail:
            # TODO: is ?detail={detail} supported in any HMC build?
            path += '?detail=' + detail
        if xag:
            # sort xag in order
            xag.sort()
            path += '?group=%s' % xag[0]
            if len(xag) > 1:
                for g in xag[1:len(xag)]:
                    path += ',%s' % g
        return path

    @staticmethod
    def _validate(reqMethod, rootType, rootId=None, childType=None,
                  childId=None, suffixType=None, suffixParm=None, detail=None):
        if childType and not rootId:
            raise ValueError('expected rootId')
        if childId and not childType:
            raise ValueError('expected childType')
        if reqMethod == 'create':
            if suffixType:
                if suffixType != 'do':
                    raise ValueError('unexpected suffixType=%s' % suffixType)
                if not suffixParm:
                    raise ValueError('expected suffixParm')
                if childType and not childId:
                    raise ValueError('expected childId')
            else:
                if childId:
                    raise ValueError('unexpected childId')
                if rootId and not childType:
                    raise ValueError('unexpected rootId')
        elif reqMethod == 'read':
            pass  # no read-specific validation at this time
        elif reqMethod == 'update':
            if 'preferences' in [rootType, childType]:
                if childId:
                    raise ValueError('unexpected childId')
                if rootId and not childType:
                    raise ValueError('unexpected rootId')
            else:
                if not rootId:
                    raise ValueError('expected rootId')
                if childType and not childId:
                    raise ValueError('expected childId')
                if suffixType is not None and suffixType != 'cancel':
                    raise ValueError('unexpected suffixType=%s' % suffixType)
        elif reqMethod == 'delete':
            if suffixType:
                if suffixType != 'jobs':
                    raise ValueError('unexpected suffixType=%s' % suffixType)
                if not suffixParm:
                    raise ValueError('expected suffixParm')
            else:
                if not rootId:
                    raise ValueError('expected rootId')
                if childType and not childId:
                    raise ValueError('expected childId')
        else:
            raise ValueError('unexpected reqMethod=%s' % reqMethod)

    def _get_resp_from_cache(self, path, age=-1, etag=None):
        """extract K2Response from cached data either from cached entry
        or feed."""
        cached_resp = self._cache.get(path, age=age)
        if cached_resp is None:
            # check the feed
            uuid, xag_str = _get_uuid_xag_from_path(path)
            if uuid:
                feed_paths = self._cache.get_feed_paths(path)
                LOG.debug('Checking cached feeds %s for uuid %s %s' %
                          (feed_paths, uuid, xag_str))
                for f_path in feed_paths:
                    cached_resp = self._build_entry_resp(f_path, uuid, etag,
                                                         age)
                    if cached_resp is not None:
                        break
        else:
            LOG.debug('Got cached entry for path %s' % path)

        return cached_resp

    def invalidate_cache_elem_by_path(self, path, invalidate_feeds=False):
        """Invalidates a cache entry where the URI path is already known"""
        path = _sanitize_path(path)

        # TODO: return K2Response object? what shoud we return?
        if self._cache is not None:
            # need to invalidate this path in the cache
            self._cache.remove(path)
            # if user wants to invalidate feeds as well,
            # invalidate_feeds will equal True
            if invalidate_feeds:
                feed_paths = self._cache.get_feed_paths(path)
                for feed_path in feed_paths:
                    self._cache.remove(feed_path)

    # Invalidates the cache entry apart from CRUD operations
    def invalidate_cache_elem(self, rootType, rootId=None, childType=None,
                              childId=None, service='uom',
                              invalidate_feeds=False):
        """Invalidates a cache entry.
        Will build the URI path using the provided arguments."""
        K2Operator._validate('read', rootType, rootId, childType, childId)
        path = K2Operator.buildpath(service, rootType, rootId,
                                    childType, childId)
        self.invalidate_cache_elem_by_path(path, invalidate_feeds)


class K2Response(object):

    """Response to K2 Operator method invocation"""

    def __init__(self, reqmethod, reqpath, status, reason, headers,
                 reqheaders={}, reqbody='', body='', orig_reqpath=''):
        self.reqmethod = reqmethod
        self.reqpath = reqpath
        self.reqheaders = reqheaders
        self.reqbody = reqbody
        self.status = status
        self.reason = reason
        self.headers = headers
        self.body = body
        self.feed = None
        self.entry = None
        # to keep track of original reqpath
        # if the K2Response is built from cached feed
        self.orig_reqpath = orig_reqpath

    def _unmarshalAtom(self):
        err_reason = None
        if self.body:
            root = None
            try:
                root = etree.fromstring(self.body,
                                        parser=etree.XMLParser(
                                            strip_cdata=False,
                                            encoding='utf-8'))
            except Exception as e:
                err_reason = 'error parsing XML response from K2: %s' % str(e)
            if root is not None and root.tag == str(QName(atom_ns, 'feed')):
                self.feed = _unmarshalAtomFeed(root)
            elif root is not None and root.tag == str(QName(atom_ns, 'entry')):
                self.entry = _unmarshalAtomEntry(root)
            elif err_reason is None:
                err_reason = 'response is not an Atom feed/entry'
        elif self.reqmethod == 'GET':
            if self.status == 204:
                if re.match(uuid_search_str,
                            self.reqpath.split('?')[0].rsplit('/', 1)[1]):
                    err_reason = 'unexpected HTTP 204 for request'
                else:
                    # This is a hack to workaround SW221445, where HMC returns
                    # HTTP 204 (No Content) when you ask for a feed that has no
                    # entries (contrary to the behavior of earlier HMC builds)
                    self.feed = K2Feed({}, [])
            elif self.status == 304:
                pass
            else:
                err_reason = 'unexpectedly empty response body'

        if err_reason is not None:
            LOG.error(('%(err_reason)s:\n'
                       'request headers: %(reqheaders)s\n\n'
                       'request body: %(reqbody)s\n\n'
                       'response headers: %(respheaders)s\n\n'
                       'response body: %(respbody)s')
                      % {'err_reason': err_reason,
                         'reqheaders': self.reqheaders,
                         'reqbody': self.reqbody,
                         'respheaders': self.headers,
                         'respbody': self.body})
            raise K2AtomError('Atom error for %s %s: %s'
                              % (self.reqmethod, self.reqpath, err_reason),
                              self)


class K2Feed(object):

    def __init__(self, properties, entries):
        self.properties = properties
        self.entries = entries

    def findentries(self, subelem, text):
        entries = []
        for entry in self.entries:
            subs = entry.element.findall(subelem)
            for s in subs:
                if s.gettext() == text:
                    entries.append(entry)
                    break
        return entries


class K2Entry(object):

    def __init__(self, properties, element):
        self.properties = properties
        self.element = K2Element.wrapelement(element)


class K2Element(object):

    def __init__(self, tag, ns=uom_ns, attrib={}, text='', children=[]):
        if ns:
            self._element = etree.Element(str(QName(ns, tag)), attrib=attrib)
        else:
            self._element = etree.Element(tag, attrib=attrib)
        if text:
            self._element.text = text
        for c in children:
            self.append(c)

    def __len__(self):
        return len(self._element)

    def __getitem__(self, index):
        return K2Element.wrapelement(self._element[index])

    def __setitem__(self, index, value):
        if not isinstance(value, K2Element):
            raise ValueError('value must be of type K2Element')
        self._element[index] = value._element

    def __delitem__(self, index):
        del self._element[index]

    @classmethod
    def wrapelement(cls, element):
        if element is None:
            return None
        e = cls('element')  # create with minimum inputs
        e._element = element  # assign element over the one __init__ creates
        return e

    def toxmlstring(self):
        # TODO: in Python 2.7+, use ElementTree's default_namespace support
        return etree.tostring(self._element)

    def gettag(self):
        return K2Element._gettag(self._element.tag)

    @staticmethod
    def _gettag(tag):
        return QName(tag).localname

    def settag(self, tag):
        ns = self.getnamespace()
        if ns:
            self._element.tag = QName(ns, tag).text
        else:
            self._element.tag = tag
        return self

    tag = property(gettag, settag)

    def getnamespace(self):
        return K2Element._getnamespace(self._element.tag)

    @staticmethod
    def _getnamespace(qtag):
        ns = QName(qtag).namespace
        return '' if ns is None else ns

    def setnamespace(self, ns):
        self._element.tag = QName(ns, self.gettag()).text
        return self

    namespace = property(getnamespace, setnamespace)

    def gettext(self):
        return self._element.text

    def settext(self, text):
        self._element.text = text
        return self

    text = property(gettext, settext)

    def getattrib(self):
        return self._element.attrib

    def setattrib(self, attrib):
        self._element.attrib = attrib
        return self

    attrib = property(getattrib, setattrib)

    def get(self, key, default=None):
        """Gets the element attribute named key.

        Returns the attribute value, or default if the attribute was not found.

        """
        return self._element.get(key, default)

    def items(self):
        """Returns the element attributes as a sequence of (name, value) pairs.

        The attributes are returned in an arbitrary order.

        """
        return self._element.items()

    def keys(self):
        """Returns the element attribute names as a list.

        The names are returned in an arbitrary order.

        """
        return self._element.keys()

    def set(self, key, value):
        """Set the attribute key on the element to value."""
        self._element.set(key, value)

    def append(self, subelement, deepcopy=True):
        """Adds subelement to the end of this element's list of subelements."""
        new_elem = subelement._element
        # lxml doesn't handle reusing elements well, so we'll make sure that
        # isn't being done here unless explicitly told that is unnecessary
        if deepcopy:
            new_elem = copy.deepcopy(new_elem)
        self._element.append(new_elem)

    def find(self, match):
        """Finds the first subelement matching match.

        match may be a tag name or path.
        Returns an element instance or None.

        """
        qpath = K2Element._qualifypath(match, self.getnamespace())
        e = self._element.find(qpath)
        if e is not None:  # must specify "is not None" here to work
            return K2Element.wrapelement(e)
        else:
            return None

    def findall(self, match):
        """Finds all matching subelements.

        match may be a tag name or path.
        Returns a list containing all matching elements in document order.

        """
        qpath = K2Element._qualifypath(match, self.getnamespace())
        e_iter = self._element.findall(qpath)
        k2elems = []
        for e in e_iter:
            k2elems.append(K2Element.wrapelement(e))
        return k2elems

    def findtext(self, match, default=None):
        """Finds text for the first subelement matching match.

        match may be a tag name or path.
        Returns the text content of the first matching element, or default
        if no element was found. Note that if the matching element has no text
        content an empty string is returned.

        """
        qpath = K2Element._qualifypath(match, self.getnamespace())
        text = self._element.findtext(qpath, default)
        return text if text else default

    def insert(self, index, subelement):
        """Inserts subelement at the given position in this element.

        Raises TypeError if subelement is not an Element.

        """
        self._element.insert(index, subelement._element)

    def iter(self, tag=None):
        """Creates a tree iterator with the current element as the root.

        The iterator iterates over this element and all elements below it, in
        document (depth first) order. If tag is not None or '*', only elements
        whose tag equals tag are returned from the iterator. If the tree
        structure is modified during iteration, the result is undefined.

        """
        # TODO: this method doesn't seem to be working...
        if not tag or tag == '*':
            qtag = None
        else:
            qtag = str(QName(self.getnamespace(), tag))
        for e in self._element.iter(tag=qtag):
            yield K2Element.wrapelement(e)

    def remove(self, subelement):
        """Removes subelement from the element.

        Unlike the find* methods this method compares elements based on the
        instance identity, not on tag value or contents.

        """
        self._element.remove(subelement._element)

    @staticmethod
    def _qualifypath(path, ns):
        if not ns:
            return path
        parts = path.split('/')
        for i in range(len(parts)):
            if parts[i] and not re.match('[\.\*\[\{]', parts[i]):
                parts[i] = str(QName(ns, parts[i]))
        return '/'.join(parts)


class K2ElementIterator(object):

    def __init__(self, it):
        self.it = it

    def __next__(self):
        return K2Element.wrapelement(next(self.it))


class K2EventListener(object):

    def __init__(self, k2session, timeout=-1, interval=15):
        if k2session is None:
            raise ValueError('k2session must not be None')
        # Verified MD5 issue from Bandit scan
        # MD5 is only generating a unique name
        self.appid = md5(k2session._sessToken).hexdigest()  # nosec
        self.timeout = timeout if timeout != -1 else k2session.timeout
        self.interval = interval
        self._lock = threading.RLock()
        self.handlers = []
        self._pthread = None
        try:
            self.oper = K2Operator(k2session, use_cache=False)
            allevents = self.getevents()  # initialize
        except K2Error as e:
            raise K2Error('failed to initialize event feed listener: %s' % e)
        if not allevents.get('general') == 'init':
            # Something else is sharing this feed!
            # TODO: improve K2 design so this doesn't happen
            raise ValueError('application id "%s" is not unique' % self.appid)

    def subscribe(self, handler):
        if not isinstance(handler, K2EventHandler):
            raise ValueError('handler must be an EventHandler')
        if self.oper is None:
            raise Exception('shutting down')
        with self._lock:
            if handler in self.handlers:
                raise ValueError('this handler is already subscribed')
            self.handlers.append(handler)
            if not self._pthread:
                self._pthread = _EventPollThread(self, self.interval)
                self._pthread.start()

    def unsubscribe(self, handler):
        if not isinstance(handler, K2EventHandler):
            raise ValueError('handler must be an EventHandler')
        with self._lock:
            if handler not in self.handlers:
                raise ValueError('this handler not found in subscriber list')
            self.handlers.remove(handler)
            if not self.handlers:
                self._pthread.stop()
                self._pthread = None

    def shutdown(self):
        host = self.oper.session.host
        LOG.info('shutting down K2EventListener for %s' % host)
        with self._lock:
            for handler in self.handlers:
                self.unsubscribe(handler)
            self.oper = None
        LOG.info('K2EventListener shutdown complete for %s' % host)

    def getevents(self):

        events = {}

        # Read event feed
        try:
            k2resp = self.oper.read('Event?QUEUE_CLIENTKEY_METHOD='
                                    'USE_APPLICATIONID&QUEUE_APPLICATIONID=%s'
                                    % self.appid, timeout=self.timeout)
        except K2Error:
            # TODO: improve error handling
            LOG.exception('error while getting K2 events')
            return events

        # Parse event feed
        for entry in k2resp.feed.entries:
            etype = entry.element.findtext('EventType')
            href = entry.element.findtext('EventData')
            if etype == 'NEW_CLIENT':
                events['general'] = 'init'
            elif etype in ['CACHE_CLEARED', 'MISSING_EVENTS']:
                events = {'general': 'invalidate'}
                break
            elif etype == 'ADD_URI':
                events[href] = 'add'
            elif etype == 'DELETE_URI':
                events[href] = 'delete'
            elif etype in ['MODIFY_URI', 'INVALID_URI', 'HIDDEN_URI']:
                if href not in events:
                    events[href] = 'invalidate'
            elif etype not in ['VISIBLE_URI']:
                LOG.error('unexpected EventType=%s' % etype)

        # Notify subscribers
        with self._lock:
            for h in self.handlers:
                try:
                    h.process(events)
                except Exception:
                    LOG.exception('error while processing K2 events')

        return events


class K2EventHandler(object):

    __metaclass__ = ABCMeta

    @abstractmethod
    def process(self, events):
        pass


class _EventPollThread(threading.Thread):

    def __init__(self, eventlistener, interval):
        threading.Thread.__init__(self)
        self.eventlistener = eventlistener
        self.interval = interval
        self.done = False
#        self.daemon = True

    def run(self):
        while not self.done:
            self.eventlistener.getevents()
            interval = self.interval
            while interval > 0 and not self.done:
                sleep(1)
                interval -= 1

    def stop(self):
        self.done = True


class _Cache(object):

    def __init__(self, host, order='byaccess'):
        if order not in ['byaccess', 'byadd']:
            raise ValueError('invalid order=%s', order)
        self._order = order
        self._keys = []
        self._data = {}
        self._lock = threading.RLock()
        self.host = host

    def __iter__(self):
        # use deepcopy so that the contents can't be changed
        return copy.deepcopy(self._data).__iter__()

    def clear(self):
        with self._lock:
            LOG.debug("%s cache cleared" % (self.host))
            self._keys = []
            self._data = {}

    def get(self, key, age=-1):
        if not age:
            return
        with self._lock:
            if key in self._keys:
                if self._order == 'byaccess':
                    # keep keys in least-accessed-first order
                    self._keys.append(self._keys.pop(self._keys.index(key)))
                (t, d) = self._data.get(key, (0, None))
                if d is not None:
                    if(age < 0 or (datetime.now() - t <
                                   timedelta(seconds=age))):
                        LOG.debug("%s cache get for %s" % (self.host, key))
                        return copy.deepcopy(d)

    def set(self, key, value):
        if not key:
            raise ValueError('key must not be empty')
        if not value:
            raise ValueError('value must not be empty')
        with self._lock:
            LOG.debug("%s cache set for %s" % (self.host, key))

            if key not in self._keys:
                # new add, nothing already there
                self._keys.append(key)
                if len(self._keys) > cachesize:
                    # limit the size of the cache
                    rem_key = self._keys.pop(0)
                    LOG.debug("%s cache overflow for %s" %
                              (self.host, rem_key))
                    del self._data[rem_key]
            elif self._order == 'byaccess':
                # keep paths in least-accessed-first order
                self._keys.append(self._keys.pop(self._keys.index(key)))

            # deepcopy of value is the responsibility of the caller
            self._data[key] = (datetime.now(), value)

    def touch(self, key):
        with self._lock:
            if key not in self._keys:
                LOG.debug("%s cache touch did not find %s" % (self.host, key))
            else:
                LOG.debug("%s cache touch for %s" % (self.host, key))
                self._keys.append(self._keys.pop(self._keys.index(key)))
                d = self._data.get(key, (0, None))[1]
                self._data[key] = (datetime.now(), d)

    def remove(self, key):
        with self._lock:
            if key in self._keys:
                LOG.debug("%s cache remove for %s" % (self.host, key))
                self._keys.remove(key)
                del self._data[key]


class _K2Cache(_Cache):

    def __init__(self, host, order='byaccess'):
        super(_K2Cache, self).__init__(host, order)
        # keep track of uuid & its links
        # this is used to figure out all the feed path
        # entry should only be remove if entry is deleted
        self._uuid_feeds_map = {}

    def clear(self):
        with self._lock:
            self._uuid_feeds_map = {}
            super(_K2Cache, self).clear()

    def get(self, key, age=-1):
        return super(_K2Cache, self).get(self._get_internal_key(key), age)

    def set(self, key, paths, value):
        if not key:
            raise ValueError('key must not be empty')
        if not paths:
            raise ValueError('paths must not be empty')
        if not value:
            raise ValueError('value must not be empty')
        with self._lock:
            i_key = self._get_internal_key(key)
            LOG.debug("%s cache set for %s" % (self.host, i_key))
            if i_key not in self._keys:
                # new add, nothing already there
                self._keys.append(i_key)
                if len(self._keys) > cachesize:
                    # limit the size of the cache
                    rem_key = self._keys.pop(0)
                    LOG.debug("%s cache overflow for %s" %
                              (self.host, rem_key))
                    del self._data[rem_key]
            elif self._order == 'byaccess':
                # keep paths in least-accessed-first order
                self._keys.append(self._keys.pop(self._keys.index(i_key)))
            entry_uuid = self._get_entry_uuid(key)

            if entry_uuid:
                feedpaths = []
                for path in paths:
                    feedpaths.append(path.rsplit('/', 1)[0])
                self._update_feed_paths(entry_uuid.lower(), feedpaths)
            # deepcopy of value is the responsibility of the caller
            self._data[i_key] = (datetime.now(), value)

    def touch(self, key):
        super(_K2Cache, self).touch(self._get_internal_key(key))

    def remove(self, key, delete=False):
        with self._lock:
            i_key = self._get_internal_key(key.split('?', 1)[0])
            keys_to_remove = []
            for k in self._keys:
                if k == i_key or k.startswith(i_key + '?'):
                    keys_to_remove.append(k)
            for k in keys_to_remove:
                LOG.debug("%s cache remove for %s via %s" % (self.host,
                                                             k, key))
                self._keys.remove(k)
                del self._data[k]
                entry_uuid = self._get_entry_uuid(key)
                if (delete and entry_uuid):
                    # change this part
                    del self._uuid_feeds_map[entry_uuid]

    def _get_internal_key(self, key):
        # extract the internal key from the path. If it's a feed, leave the
        # key as is. Otherwise, it's uuid?group=xag
        uuid, xag_str = _get_uuid_xag_from_path(key)
        if not uuid:
            return key
        return uuid + '?group=' + xag_str if xag_str else uuid

    def _get_entry_uuid(self, key):
        uuid = _get_req_path_uuid(key)
        if uuid:
            uuid = uuid.lower()
        return uuid

    def get_feed_paths(self, key):
        with self._lock:
            feed_paths = []
            entry_uuid, xag_str = _get_uuid_xag_from_path(key)
            if entry_uuid:
                default_feed_path = key.rsplit('/', 1)[0]
                keys = self._uuid_feeds_map.get(entry_uuid.lower(), [])
                if default_feed_path not in keys:
                    keys = self._update_feed_paths(entry_uuid.lower(),
                                                   [default_feed_path])
                for k in keys:
                    if xag_str:
                        feed_paths.append(k + '?group=' + xag_str)
                    else:
                        feed_paths.append(k)
            LOG.debug('key %s feed_paths = %s' % (key, feed_paths))
            return feed_paths

    def _update_feed_paths(self, entry_uuid, new_paths):
        with self._lock:
            paths = self._uuid_feeds_map.get(entry_uuid, [])
            if not paths:
                self._uuid_feeds_map[entry_uuid] = new_paths
            else:
                for n in new_paths:
                    if n not in paths:
                        paths.append(n)
                self._uuid_feeds_map[entry_uuid] = paths
            return paths


class _EventHandler(K2EventHandler):

    def __init__(self, cache):
        self.cache = cache

    def process(self, events):
        for k, v in six.iteritems(events):
            if k == 'general':
                if v == 'invalidate':
                    self.cache.clear()
            elif v in ['delete', 'invalidate']:
                path = _sanitize_path(urlparse(k).path)
                # remove from cache
                self.cache.remove(path)
                # if entry, remove corresponding feeds from cache
                feed_paths = self.cache.get_feed_paths(path)
                for feed_path in feed_paths:
                    self.cache.remove(feed_path)


def _sanitize_path(path):
    # trim trailing '/' from path, if present
    if len(path) > 1 and path[len(path) - 1] == '/':
        path = path[0:len(path) - 1]
    return path


def _determine_paths(k2resp):
    paths = []
    if k2resp.feed:
        links = k2resp.feed.properties.get('links')
    else:
        links = k2resp.entry.properties.get('links')
    if links:
        self_links = links.get('SELF')
        if self_links:
            for lnk in self_links:
                paths.append(urlparse(lnk).path)
    if k2resp.reqmethod == 'PUT':
        # BEGIN HACK for defect SW246342
        # TODO: remove when offending HMC drivers are no longer supported
        for idx in range(len(paths)):
            path = _sanitize_path(paths[idx])
            pieces = path.rsplit('/', 1)
            if not re.match(uuid_search_str, pieces[1]):
                paths[idx] = path + '/' + k2resp.entry.properties['id']
        # END HACK for defect SW246342
    if not paths:
        if k2resp.reqmethod == 'PUT':
            # a PUT's reqpath will be the feed, to which we need to add
            # the new entry id (which didn't exist before the PUT)
            paths = [k2resp.reqpath + '/' + k2resp.entry.properties['id']]
        else:
            paths = [k2resp.reqpath]
        LOG.debug('Unable to determine full set of paths from %s response' %
                  k2resp.reqmethod + ' ' + k2resp.reqpath)
    return paths


def _get_max_age(path, use_events, schema_version):
    if any(p in path for p in ['Cluster', 'SharedStoragePool']):
        # no event support
        # TODO: update when event support is added
        return 15
    if not use_events or schema_version.startswith('V1_0'):
        # bad event support
        # attempt to return the same values used by PowerVC 1.2.0 feed caches
        # TODO: allow these values to be customizable?
        if re.search('/LogicalPartition$', path):
            return 30
        if re.search('/VirtualIOServer$', path):
            return 90
        if re.search('/SharedProcessorPool$', path):
            return 600
        if re.search('/ManagedSystem/%s$' % uuid_search_str, path):
            return 30
        else:
            # TODO: can we trust the cache longer than 0 for anything else?
            return 0
    else:
        # TODO: consider extending as we grow more confident in events
        return 600


def _unmarshalAtomFeed(feedelem):
    feedprops = {}
    entries = []
    for child in list(feedelem):
        if child.tag == str(QName(atom_ns, 'entry')):
            entries.append(_unmarshalAtomEntry(child))
        elif not list(child):
            if re.match('{%s}' % atom_ns, child.tag):
                # strip off atom namespace qualification for easier access
                paramName = child.tag[child.tag.index('}') + 1:len(child.tag)]
            else:
                # leave qualified anything that is not in the atom namespace
                paramName = child.tag
            # TODO: handle links?
            feedprops[paramName] = child.text
    return K2Feed(feedprops, entries)


def _unmarshalAtomEntry(entryelem):
    entryprops = {}
    element = None
    for child in list(entryelem):
        if child.tag == str(QName(atom_ns, 'content')):
            element = child[0]  # K2 only has one element per entry
        elif not list(child):
            if re.match('{%s}' % atom_ns, child.tag):
                # strip off atom namespace qualification for easier access
                paramName = child.tag[child.tag.index('}') + 1:len(child.tag)]
            else:
                # leave qualified anything that is not in the atom namespace
                paramName = child.tag
            if paramName == 'link':
                entryprops[paramName] = child.get('href')
                rel = child.get('rel')
                if rel:
                    if 'links' not in entryprops:
                        entryprops['links'] = {}
                    if rel not in entryprops['links']:
                        entryprops['links'][rel] = []
                    entryprops['links'][rel].append(child.get('href'))
            elif paramName == 'category':
                entryprops[paramName] = child.get('term')
            elif paramName == '{%s}etag' % uom_ns:
                entryprops['etag'] = child.text
            elif child.text:
                entryprops[paramName] = child.text
    return K2Entry(entryprops, element)


def _unmarshal_k2httperror(k2resp):
    # attempt to extract K2's HttpErrorResponse object
    try:
        root = etree.fromstring(k2resp.body,
                                parser=etree.XMLParser(strip_cdata=False,
                                                       encoding='utf-8'))
        if root is not None and root.tag == str(QName(atom_ns, 'entry')):
            k2resp.k2err = _unmarshalAtomEntry(root).element
    except:
        pass


def _getEntryAtomFeed(feedelem, uuid):
    """Parse atom feed to extract the entry matching to the uuid"""
    if not feedelem:
        return None, None
    uuid = uuid.lower()
    entry = None
    for f_elem in list(feedelem):
        if f_elem.tag == str(QName(atom_ns, 'entry')):
            etag = None
            for e_elem in list(f_elem):
                if not list(e_elem):
                    if re.match('{%s}' % atom_ns, e_elem.tag):
                        # strip off atom namespace qualification for easier
                        # access
                        paramName = (e_elem.tag[e_elem.tag.index('}')
                                                + 1:len(e_elem.tag)])
                    else:
                        # leave qualified anything that is not in the atom
                        # namespace
                        paramName = e_elem.tag
                    if paramName == '{%s}etag' % uom_ns:
                        etag = e_elem.text
                    elif paramName == 'id':
                        if e_elem.text.lower() == uuid:
                            entry = etree.tostring(f_elem)

        if entry is not None:
            return entry, etag
    return None, None


def _get_req_path_uuid(path):
    """extract request target uuid of sanitized path"""
    if '/' in path:
        target_id = path.rsplit('/', 1)[1]
        uuid_match = re.match(uuid_search_str, target_id)
        if uuid_match:
            return uuid_match.group(0).lower()
    return None


def _get_uuid_xag_from_path(path):
    if '/' in path:
        uuid = path.rsplit('/', 1)[1]
        match_str = '(' + uuid_search_str + ')' + '(.*)'
        matched_uuid = re.match(match_str, uuid)
        if matched_uuid:
            (uuid, parms) = matched_uuid.group(1, 2)
            xag_search = re.search('([&|\?]group=)(.*)', parms)
            if xag_search:
                xag_str = xag_search.group(2)
                if '&' in xag_str:
                    xag_str = xag_str.split('&')[0]
            else:
                xag_str = ''
            return (uuid.lower(), xag_str)
    return None, None
