#
#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import json
import copy
import logging
import os
from powervc_oslo import _


LOG = logging.getLogger(__name__)


class VMFlowMapping:

    '''
    Class to represent a json formatted collection of
    the flows used by the host's virtual machines.  The
    following format is used:

    {"virtual_machines":
        {
            "vm_name1":
                {
                 "tap_device_name1":
                    {
                      "external_vswitch": "vswitch_name",
                      "external_vlan": vlan_id_as_integer
                    },
                 "tap_device_name1-2":
                    {
                      "external_vswitch": "vswitch_name",
                      "external_vlan": vlan_id_as_integer
                    }
                },
            "vm_name2":
                {
                 "tap_device_name2":
                    {
                      "external_vswitch": "vswitch_name",
                      "external_vlan": vlan_id_as_integer
                    }
                }
        }
    }

    external_vlan should be set to "flat" if it is a flat network
    '''

    VM_FLOW_MAP_DIR = "/etc/sysconfig/network-scripts"
    VM_FLOW_MAP_FILE = "%s/pvc_flow_mapping.json" % VM_FLOW_MAP_DIR
    IMPORT_VM_FLOW_MAP_FILE_SUFFIX = \
        "_pvc_flow_mapping.json"
    IMPORT_VM_FLOW_MAP_DIR = "/tmp"
    VIRTUAL_MACHINES_KEY = "virtual_machines"
    VM_EXTERNAL_VSWITCH_KEY = "external_vswitch"
    VM_EXTERNAL_VLAN_KEY = "external_vlan"
    FLAT_NETWORK_VALUE = "flat"
    MIN_VLAN_VALUE = 1
    MAX_VLAN_VALUE = 4094
    UNDEFINED_VLAN_VALUE = 4095

    valid_vlan_list = []
    valid_vlan_list.append(FLAT_NETWORK_VALUE)
    valid_vlan_list.append(UNDEFINED_VLAN_VALUE)
    valid_vlan_list.extend(range(MIN_VLAN_VALUE,
                                 MAX_VLAN_VALUE + 1))

    # current flow map should be static
    current_flow_map = {}

    @classmethod
    def get_vm_flow_map(cls):
        '''
        Return the current flow map.  If the in-memory
        map is set, return that; otherwise read data
        from the file, set the in-memory map, and
        return the data

        :returns: current flow map
        '''
        if not cls.current_flow_map:
            flow_file = None
            try:
                flow_file = open(cls.VM_FLOW_MAP_FILE)
                VMFlowMapping.current_flow_map = json.load(flow_file)
            except Exception as ex:
                LOG.exception(ex)
            finally:
                if flow_file:
                    flow_file.close()

        return VMFlowMapping.current_flow_map

    @classmethod
    def refresh_vm_flow_map(cls):
        '''
        Refresh the in-memory current flow map from the
        file on the system
        '''
        flow_file = None
        try:
            flow_file = open(cls.VM_FLOW_MAP_FILE)
            VMFlowMapping.current_flow_map = json.load(flow_file)
        except Exception as ex:
            LOG.exception(ex)
        finally:
            if flow_file:
                flow_file.close()

    @classmethod
    def set_vm_flow_map(cls, flow_data):
        '''
        Set new data into both the file containing
        the persistent flow data and the
        in-memory data

        :param flow_data: flow data to be used to
            set the persistent and in-memory flow
            data
        :returns: True if successful, False otherwise
        '''

        if not cls.validate_vm_flow_map(flow_data):
            return False

        outfile = None
        ret_val = True

        try:
            VMFlowMapping.current_flow_map = flow_data
            outfile = open(cls.VM_FLOW_MAP_FILE, 'w')
            json.dump(flow_data, outfile)
        except Exception as ex:
            LOG.exception(ex)
            ret_val = False
        finally:
            if outfile:
                outfile.close()

        return ret_val

    @classmethod
    def validate_vm_flow_map(cls, flow_data):
        '''
        Validate that the flow data json corresponds to the
        expected formatting.

        :param flow_data: flow mapping in json format
        :return True if data is valid, False otherwise
        '''
        if cls.VIRTUAL_MACHINES_KEY not in flow_data:
            LOG.error(_("Flow data is incomplete, missing %s parameter."
                        % cls.VIRTUAL_MACHINES_KEY))
            return False

        for vm_name, tap_devices in \
                flow_data[cls.VIRTUAL_MACHINES_KEY].iteritems():
            for tap_name, tap_data in \
                    tap_devices.iteritems():
                if cls.VM_EXTERNAL_VSWITCH_KEY not in tap_data:
                    LOG.error(_("Missing %s parameter for virtual "
                                "machine %s on device %s." %
                                (cls.VM_EXTERNAL_VSWITCH_KEY,
                                 vm_name, tap_name)))
                    return False
                if cls.VM_EXTERNAL_VLAN_KEY not in tap_data:
                    LOG.error(_("Missing %s parameter for virtual "
                                "machine %s on device %s." %
                                (cls.VM_EXTERNAL_VLAN_KEY,
                                 vm_name, tap_name)))
                    return False

                if tap_data[cls.VM_EXTERNAL_VLAN_KEY] not in \
                        cls.valid_vlan_list:
                    LOG.error(_("Specified external vlan value %s for tap "
                                "device %s on virtual machine %s "
                                "is not valid.  Valid values "
                                "are 'flat' or a numeric value between 1 "
                                "and 4094." %
                                (cls.VM_EXTERNAL_VLAN_KEY,
                                 tap_name, vm_name)))
                    return False

        return True

    @classmethod
    def get_vm_flow_data(cls, vm_name):
        '''
        Get the flow data for a specific vm by name

        :param vm_name: Name of the vm to retrieve flow data
            for
        :returns: Dict with the flow data for the specified vm
        '''

        return VMFlowMapping.current_flow_map[cls.VIRTUAL_MACHINES_KEY].\
            get(vm_name)

    @classmethod
    def get_tap_device_data(cls, tap_name):
        '''
        Get the flow data for a specific tap device by name

        :param tap_name: Name of the tap device to retrieve flow data
            for
        :returns: Dict with the flow data for the specified tap device
        '''
        if not VMFlowMapping.current_flow_map:
            if not cls.get_vm_flow_map():
                return {}

        for vm in VMFlowMapping.current_flow_map[cls.VIRTUAL_MACHINES_KEY].\
                itervalues():
            for tap, flow_data in vm.iteritems():
                if tap == tap_name:
                    return flow_data

        return {}

    @classmethod
    def get_tap_device_names(cls):
        '''
        Get the names of all tap devices in the file

        :returns: List of all tap device names
        '''

        tap_name_list = []

        if not VMFlowMapping.current_flow_map:
            if not cls.get_vm_flow_map():
                return {}

        for vm in VMFlowMapping.current_flow_map[cls.VIRTUAL_MACHINES_KEY].\
                itervalues():
            for tap in vm.iterkeys():
                tap_name_list.append(tap)

        return tap_name_list

    @classmethod
    def get_vm_name_for_tap_device(cls, tap_name):
        '''
        Get the name of vm associated with the tap device

        :param tap_name: Name of the tap device to retrieve vm data
            for
        :returns: name of the vm
        '''
        if not VMFlowMapping.current_flow_map:
            if not cls.get_vm_flow_map():
                return {}

        for vm_name, vm in VMFlowMapping.\
                current_flow_map[cls.VIRTUAL_MACHINES_KEY].iteritems():
            for tap in vm.iterkeys():
                if tap == tap_name:
                    return vm_name

        return None

    @classmethod
    def update_vm_flow_data(cls,
                            vm_name,
                            update_copy=True,
                            **kwargs):
        '''
        Update the data for a vm that is already listed
        in the flow mapping

        :param vm_name: Name of the vm to update the data for
        :param update_copy:  If True, the in-memory and file-based
            flow data will not be updated and a copy of the updated
            data will be returned; if False, the persistent data
            will be updated automatically and will not require a
            set_vm_flow_map to be called explicitly
        :param kwargs: keywords should correspond to tap device
            names and value should be a dict with external vswitch
            and external vlan data.  Value of None indicates
            to delete the tap data from the vm entry
                ex. - tap1={"external_vswitch": "vs1",
                            "external_vlan": 99}
        :returns: Updated flow data if update successful,
            otherwise {}
        '''
        flow_data = copy.deepcopy(cls.get_vm_flow_map())

        if vm_name in flow_data[cls.VIRTUAL_MACHINES_KEY]:
            curr_vm_data = \
                flow_data[cls.VIRTUAL_MACHINES_KEY][vm_name]

            for key in kwargs:
                if key not in curr_vm_data:
                    curr_vm_data[key] = {}

                # no data means to delete the tap device specified
                # by the key
                if not kwargs[key]:
                    del curr_vm_data[key]
                else:
                    for subkey in kwargs[key].iterkeys():
                        curr_vm_data[key][subkey] = kwargs[key][subkey]

        if cls.validate_vm_flow_map(flow_data):
            if not update_copy:
                cls.set_vm_flow_map(flow_data)
            return flow_data
        else:
            # error logging handled in validation method
            return {}

    @classmethod
    def add_vm_flow_data(cls, vm_name,
                         update_copy=True,
                         update_if_vm_exists=True,
                         **kwargs):
        '''
        Add a new set of flow data for the specified VM

        :param vm_name: Name of the vm to add the data for
        :param update_copy:  If True, the in-memory and file-based
            flow data will not be updated and a copy of the updated
            data will be returned; if False, the persistent data
            will be updated automatically and will not require a
            set_vm_flow_map to be called explicitly
        :param kwargs: keywords should correspond to tap device
            names and value should be a dict with external vswitch
            and external vlan data.  Value of None indicates
            to delete the tap data from the vm entry
                ex. - tap1={"external_vswitch": "vs1",
                            "external_vlan": 99}
        :returns: Updated flow data if update successful,
            otherwise {}
        '''
        flow_data = copy.deepcopy(cls.get_vm_flow_map())

        if vm_name in flow_data[cls.VIRTUAL_MACHINES_KEY]:
            if update_if_vm_exists:
                return cls.update_vm_flow_data(vm_name,
                                               update_copy,
                                               **kwargs)
            else:
                LOG.error(_("Flow data for %s already exists.  Use "
                            "update method to update the data for "
                            "an existing virtual machine." % vm_name))
                return {}

        flow_data[cls.VIRTUAL_MACHINES_KEY][vm_name] = {}
        for key in kwargs:
            flow_data[cls.VIRTUAL_MACHINES_KEY][vm_name][key] = \
                kwargs[key]

        if cls.validate_vm_flow_map(flow_data):
            if not update_copy:
                cls.set_vm_flow_map(flow_data)
            return flow_data
        else:
            # error logging handled in validation method
            return {}

    @classmethod
    def delete_vm_flow_data(cls, vm_name, update_copy=True):
        '''
        Delete the flow data for the specified VM

        :param vm_name: Name of the vm to delete the data for
        :param update_copy:  If True, the in-memory and file-based
            flow data will not be updated and a copy of the updated
            data will be returned; if False, the persistent data
            will be updated automatically and will not require a
            set_vm_flow_map to be called explicitly
        :returns: Updated flow data if update successful,
            otherwise {}
        '''
        flow_data = copy.deepcopy(cls.get_vm_flow_map())

        del flow_data[cls.VIRTUAL_MACHINES_KEY][vm_name]

        LOG.debug("flow data = %s" % flow_data)

        if cls.validate_vm_flow_map(flow_data):
            if not update_copy:
                LOG.debug("updating the mapping file")
                cls.set_vm_flow_map(flow_data)
            return flow_data
        else:
            # error logging handled in validation method
            return {}

    @classmethod
    def is_vlan_in_use(cls, external_bridge, external_vlan):
        '''
        Determine if the external_bridge/external_vlan combo
        appears in the flow mapping.

        :param external_bridge: bridge name
        :param external_vlan: vlan id as int or 'flat'
            if flat network
        :returns: True if match found, otherwise False
        '''

        if not VMFlowMapping.current_flow_map:
            if not cls.get_vm_flow_map():
                return False

        for vm in VMFlowMapping.current_flow_map[cls.VIRTUAL_MACHINES_KEY].\
                itervalues():
            for tap in vm.itervalues():
                if tap[cls.VM_EXTERNAL_VSWITCH_KEY] == external_bridge \
                        and tap[cls.VM_EXTERNAL_VLAN_KEY] == external_vlan:
                    return True

        return False

    @classmethod
    def save_flow_mapping_for_imported_vm(cls,
                                          vm_name,
                                          **kwargs):
        """
        Record the bridge and vlan information in a file so that
        the neutron agent can later add the data to the master
        flow mapping file.

        :param vm_name: name of the VM being imported
        :param kwargs: parameters to associated with the VM in
            the mapping dictionary
        """
        flow_data = {}
        flow_data[cls.VIRTUAL_MACHINES_KEY] = {}
        flow_data[cls.VIRTUAL_MACHINES_KEY][vm_name] = {}
        for key in kwargs:
            flow_data[cls.VIRTUAL_MACHINES_KEY][vm_name][key] = \
                kwargs[key]

        # write the flow data out to a file standalone file
        # containing data only for this VM.  Later processing will
        # read the contents of this file and add to the master
        # flow mapping file
        outfile = None

        filename = "%s/%s%s" % (cls.IMPORT_VM_FLOW_MAP_DIR, vm_name,
                                cls.IMPORT_VM_FLOW_MAP_FILE_SUFFIX)

        try:
            outfile = open(filename, 'w')
            json.dump(flow_data, outfile)
        except Exception as ex:
            LOG.exception(ex)
        finally:
            if outfile:
                outfile.close()

    @classmethod
    def process_imported_vm_mappings(cls):
        if not os.path.exists(cls.IMPORT_VM_FLOW_MAP_DIR):
            return

        file_list = os.listdir(cls.IMPORT_VM_FLOW_MAP_DIR)
        for filename in file_list:
            vm_flow_map = {}
            if cls.IMPORT_VM_FLOW_MAP_FILE_SUFFIX in \
                    filename:

                file_with_path = "%s/%s" % \
                    (cls.IMPORT_VM_FLOW_MAP_DIR, filename)
                try:
                    flow_file = open(file_with_path, 'r')
                    vm_flow_map = json.load(flow_file)
                except Exception as ex:
                    LOG.exception(ex)
                finally:
                    if flow_file:
                        flow_file.close()
                        os.remove(file_with_path)

                vm_data = vm_flow_map[cls.VIRTUAL_MACHINES_KEY]

                # There should only be one VM per file.  Note that
                # if the VM was powered on as part of the import process
                # there will already be an entry in the mapping file.  If
                # that is the case we don't want to overwrite what is
                # already there.
                for vm_name, mapping_data in vm_data.iteritems():
                    cls.add_vm_flow_data(vm_name,
                                         False,
                                         False,
                                         **mapping_data)
