# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import datetime
import ssl
import urllib3
import warnings
from hashlib import sha256
from pyasn1.codec.der import decoder as der_decoder
from pyasn1_modules import rfc2459


class StubClass(object):
    pass


try:
    from requests.packages import urllib3 as req_urllib3
    from requests.packages.urllib3.connection import VerifiedHTTPSConnection
    from requests.packages.urllib3.connectionpool import HTTPSConnectionPool
    requests_has_urllib3 = True
except:
    requests_has_urllib3 = False
    VerifiedHTTPSConnection = StubClass
    HTTPSConnectionPool = StubClass


RECENT_DATE = datetime.date(2014, 1, 1)


def verifycert(sock, ca_certs):
    # If they asked us to not verify the Certificate then nothing to do
    if not ca_certs:
        return

    # Retrieve the Existing Certificate from the File in Binary Form
    peercert = sock.getpeercert(True)
    try:
        with open(ca_certs, 'r') as f:
            existcert = ssl.PEM_cert_to_DER_cert(f.read())
    except:
        raise ssl.SSLError("Failed to read certificate from %(file)s" %
                           {'file': ca_certs})

    # First check to make sure the 2 certificates are the same ones
    if not sha256(existcert).digest() == sha256(peercert).digest():
        raise ssl.SSLError("The certificate doesn't match the trusted one")

    # Next check if the certificate has passed its expiration
    cert = der_decoder.decode(peercert, asn1Spec=rfc2459.Certificate())[0]
    tbs = cert.getComponentByName('tbsCertificate')
    validity = tbs.getComponentByName('validity')
    not_after = validity.getComponentByName('notAfter').getComponent()
    not_after = datetime.datetime.strptime(str(not_after), '%y%m%d%H%M%SZ')

    # Throw an exception if the certificate given to us has expired
    if datetime.datetime.utcnow() >= not_after:
        raise ssl.SSLError("The certificate expired: %s" % str(not_after))


##########################################################
#   Override the HTTP Connection for Cert Verification   #
##########################################################
class DS8KHTTPSConnection(urllib3.connection.VerifiedHTTPSConnection):

    """Extend the HTTPS Connection to do our own Certificate Verification"""

    def connect(self):
        """Override the Connect Method to fix the Certificate Verification"""
        # Add certificate verification
        conn = self._new_conn()

        if getattr(self, '_tunnel_host', None):
            # _tunnel_host was added in Python 2.6.3
            # (See: http://hg.python.org/cpython/rev/0f57b30a152f)

            self.sock = conn
            # Calls self._set_hostport(), so self.host is
            # self._tunnel_host below.
            self._tunnel()
            # Mark this connection as not reusable
            self.auto_open = 0

        # Wrap socket using verification with the root certs in
        # trusted_root_certs
        self.sock = ssl.wrap_socket(conn)

        verifycert(self.sock, self.ca_certs)
        self.is_verified = True

    def putrequest(self, method, url, **kwargs):
        """Override the Put Request method take the DS8K off of the URL"""
        if url and url.startswith('httpsds8k://'):
            url = 'https://' + url[12:]
        return super(DS8KHTTPSConnection,
                     self).putrequest(method, url, **kwargs)

    def request(self, method, url, **kwargs):
        """Override the Request method take the DS8K off of the URL"""
        if url and url.startswith('httpsds8k://'):
            url = 'https://' + url[12:]
        return super(DS8KHTTPSConnection, self).request(method, url, **kwargs)


class DS8KConnectionPool(urllib3.connectionpool.HTTPSConnectionPool):

    """Extend the HTTPS Connection Pool to our own Certificate verification"""

    scheme = 'httpsds8k'
    ConnectionCls = DS8KHTTPSConnection

    def urlopen(self, method, url, **kwargs):
        """Override URL Open method to take the DS8K out of the URL protocol"""
        if url and url.startswith('httpsds8k://'):
            url = 'https://' + url[12:]
        return super(DS8KConnectionPool, self).urlopen(method, url, **kwargs)


class DS8KHTTPSConnection2(VerifiedHTTPSConnection):

    """Extend the HTTPS Connection to do our own Certificate Verification"""

    def connect(self):
        """Override the Connect Method to fix the Certificate Verification"""
        # Add certificate verification
        conn = self._new_conn()

        if getattr(self, '_tunnel_host', None):
            # _tunnel_host was added in Python 2.6.3
            # (See: http://hg.python.org/cpython/rev/0f57b30a152f)

            self.sock = conn
            # Calls self._set_hostport(), so self.host is
            # self._tunnel_host below.
            self._tunnel()
            # Mark this connection as not reusable
            self.auto_open = 0

        is_time_off = datetime.date.today() < RECENT_DATE
        if is_time_off:
            warnings.warn((
                'System time is way off (before {0}). This will probably '
                'lead to SSL verification errors').format(RECENT_DATE),
                req_urllib3.exceptions.SystemTimeWarning
            )

        # Wrap socket using verification with the root certs in
        # trusted_root_certs
        self.sock = ssl.wrap_socket(conn)

        verifycert(self.sock, self.ca_certs)
        self.is_verified = True

    def putrequest(self, method, url, **kwargs):
        """Override the Put Request method take the DS8K off of the URL"""
        if url and url.startswith('httpsds8k://'):
            url = 'https://' + url[12:]
        return super(DS8KHTTPSConnection2,
                     self).putrequest(method, url, **kwargs)

    def request(self, method, url, **kwargs):
        """Override the Request method take the DS8K off of the URL"""
        if url and url.startswith('httpsds8k://'):
            url = 'https://' + url[12:]
        return super(DS8KHTTPSConnection2, self).request(method, url, **kwargs)


class DS8KConnectionPool2(HTTPSConnectionPool):

    """Extend the HTTPS Connection Pool to our own Certificate verification"""

    scheme = 'httpsds8k'
    ConnectionCls = DS8KHTTPSConnection2

    def urlopen(self, method, url, **kwargs):
        """Override URL Open method to take the DS8K out of the URL protocol"""
        if url and url.startswith('httpsds8k://'):
            url = 'https://' + url[12:]
        return super(DS8KConnectionPool2, self).urlopen(method, url, **kwargs)


# Override the Connection Pool we use for DS8K for DS8K HTTPS communication
urllib3.poolmanager.pool_classes_by_scheme["httpsds8k"] = DS8KConnectionPool
if hasattr(urllib3.poolmanager, 'key_fn_by_scheme'):
    key_dict = getattr(urllib3.poolmanager, 'key_fn_by_scheme')
    key_dict['httpsds8k'] = key_dict['https']

if requests_has_urllib3:
    # We can't be sure which urllib3 will be used in this case, so we'll also
    # override the Connection Pool found under requests
    pool_cls = DS8KConnectionPool2
    req_urllib3.poolmanager.pool_classes_by_scheme["httpsds8k"] = pool_cls
    if hasattr(req_urllib3.poolmanager, 'key_fn_by_scheme'):
        key_dict = getattr(req_urllib3.poolmanager, 'key_fn_by_scheme')
        key_dict['httpsds8k'] = key_dict['https']
