# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import base64
import os
import requests
import pywbem
import threading
from requests.packages import urllib3
from oslo_log import log as logging

LOG = logging.getLogger(__name__)


def powervc_wbem_request(url, data, creds, headers=[], debug=0,
                         x509=None, verify_callback=None,
                         ca_certs=None, no_verification=False):
    """Override the WBEM Request to use the Requests Library"""
    session = getattr(_local_store, 'wbem_session', None)
    # Since we use WBEM as the protocol, we need to register the HTTP adapter
    if session is None:
        session = requests.Session()
        session.mount("httpswbem://", requests.adapters.HTTPAdapter())
        setattr(_local_store, 'wbem_session', session)
        LOG.debug("@PVC: Acquired new wbem session: %s" % session)
    else:
        LOG.debug("@PVC: Reuse existing wbem session on thread: %s" % session)
    # If they asks us to not skip verification, then use the certificate
    if no_verification:
        session.verify = False
    elif os.path.exists(ca_certs):
        session.verify = ca_certs
    else:
        session.verify = True
    # Add in the Header information for the COntent and Authorization
    headers2 = {'Content-length': len(data)}
    headers2['Content-type'] = 'application/xml; charset="utf-8"'
    headers2['Authorization'] = 'Basic %s' %\
        base64.encodestring('%s:%s' % (creds[0], creds[1])).replace('\n', '')
    # Loop through each of the Header entries passed in, splitting apart
    for hdr in headers:
        headers2[hdr[:hdr.find(':')]] = hdr[hdr.find(':') + 1:].strip()
    # We are specifying WBEM as the protocol to turn off host-name verification
    url = url.replace('https://', 'httpswbem://')
    # Now we can actually execute the POST Request to the WBEM Server
    resp = session.request(
        'POST', url + '/cimom', data=data, headers=headers2)
    # If we didn't get an OK status code, then throw an exception
    if resp.status_code != 200:
        raise Exception('HTTP error: %s' % resp.reason)
    LOG.debug("@PVC: Response OK. URL=%s, len=%d." % (url, len(resp.text)))
    return resp.text


class WbemConnectionPool(urllib3.connectionpool.HTTPSConnectionPool):

    """Extend the HTTPS Connection Pool to disable host-name verification"""

    scheme = 'httpswbem'

    def __init__(self, *args, **kwargs):
        """Override the Constructor to disable host-name verification"""
        super(WbemConnectionPool, self).__init__(*args, **kwargs)
        self.assert_fingerprint = False
        self.assert_hostname = False
        LOG.debug("@PVC: New WbemConnectionPool with: assert_hostname = False")

# Override the Connection Pool we use for WBEM for WBEM HTTPS communication
urllib3.poolmanager.pool_classes_by_scheme["httpswbem"] = WbemConnectionPool
if hasattr(urllib3.poolmanager, 'key_fn_by_scheme'):
    key_dict = getattr(urllib3.poolmanager, 'key_fn_by_scheme')
    key_dict['httpswbem'] = key_dict['https']
# Override the WBEM Request to use the Request Librarys instead
pywbem.cim_http.wbem_request = powervc_wbem_request
_local_store = threading.local()
