# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import re
import sys
import ssl
import socket
import logging
from backports.ssl_match_hostname import match_hostname, CertificateError
from datetime import datetime
from hashlib import sha256
from pyasn1.codec.der import decoder as der_decoder
from pyasn1_modules import rfc2459

__import__('xsf.xiv.client.xcli.transports')
transports = getattr(sys.modules['xsf.xiv.client.xcli'], 'transports')
LOG = logging.getLogger(transports.__name__)


@classmethod
def _powervc_connect_ssl(cls, hostname, port=7778, timeout=5.0,
                         ca_certs=None, trust_type1=False, trust_type2=True):
    """Override the Connect SSL method to handle bad certificates"""
    LOG.debug("CONNECT SSL %s:%s", hostname, port)
    existcert = _get_existing_certificate(hostname)
    # Create the SSL Socket Connection to the given Hostname and Port
    sock = socket.create_connection((hostname, port), timeout)
    sock.setblocking(True)
    if existcert is None:
        # use the local OS's CA bundle
        sock = ssl.wrap_socket(sock, cert_reqs=ssl.CERT_REQUIRED,
                               ca_certs='/etc/pki/tls/certs/ca-bundle.crt')
        try:
            peercert = sock.getpeercert()
            try:
                match_hostname(peercert, hostname)
            except CertificateError as e:
                _throw_ssl_error(e.message)
        except:
            # need to close the connection before failing
            sock.close()
            raise
    else:
        # We have an existing certificate to validate against
        sock = ssl.wrap_socket(sock)
        try:
            # make sure that the 2 certificates match and are valid
            peercert = sock.getpeercert(True)
            # First check to make sure the 2 certificates are the same
            if not sha256(existcert).digest() == sha256(peercert).digest():
                _throw_ssl_error("The certificate doesn't match the "
                                 "trusted one")
            # Next check if the certificate has passed its expiration
            cert = der_decoder.decode(peercert,
                                      asn1Spec=rfc2459.Certificate())[0]
            tbs = cert.getComponentByName('tbsCertificate')
            validity = tbs.getComponentByName('validity')
            not_after = validity.getComponentByName('notAfter').getComponent()
            not_after = datetime.strptime(str(not_after), '%y%m%d%H%M%SZ')
            # Throw an exception if the certificate given to us has expired
            if datetime.utcnow() >= not_after:
                _throw_ssl_error("The certificate expired at %s" %
                                 str(not_after))
        except:
            # need to close the connection before failing
            sock.close()
            raise
    return cls(sock)


def _get_existing_certificate(hostname):
    """Helper method to get the existing certificate on the Mgmt System"""
    hostname = re.sub('[:.]', '_', hostname)
    certfile = "/opt/ibm/powervc/data/certs/%s.crt" % hostname
    try:
        # If we found the certificate, read it and return in binary
        with open(certfile, 'r') as f:
            return ssl.PEM_cert_to_DER_cert(f.read())
    except:
        return None


def _throw_ssl_error(excstr):
    """Helper method to log and throw an SSL error with the given string"""
    LOG.warn('SSL Error: ' + excstr)
    raise ssl.SSLError(excstr)

# Override the Socket Transport's connect_ssl method to use our version
transports.SocketTransport.connect_ssl = _powervc_connect_ssl
