# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
import time
import commands
import glob
import httplib
import json
import os
import logging
from powervc_ras.cli.diag_utils import utils
from powervc_ras.cli.diag_utils import logs
from powervc_ras.cli.diag_utils import security
from powervc_ras.cli.diag_utils import constants
from powervc_ras import _

debug = False
if os.getenv('POWERVCDIAG_DEBUG') is not None:
    debug = True

LOG = logging.getLogger(__name__)


def start_validator(session):
    validation_id = None
    print _('INFO: Invoking validators...')
    LOG.debug(_('INFO: Invoking validator api...'))
    try:
        resp = session.get('/v1/validation',
                           endpoint_filter={'service_type': 'ttv',
                                            'interface': 'internal'})
        if not resp or resp.status_code != httplib.ACCEPTED:
            print _("ERROR: Validator's execution failed.")
        else:
            result = json.loads(resp.text)
            validation_id = result['ValidationId']
    except Exception as ex:
        print _("ERROR: Caught exception while invoking validator, "
                "check logs for more information")
        LOG.error(_("Caught an unexpected exception while "
                    "invoking validator. %s") % unicode(ex))
    return validation_id


def _get_validator_response(session, url):
    response = None
    execution_in_progress = []
    resp = session.get(url, endpoint_filter={'service_type': 'ttv',
                                             'interface': 'internal'})
    valid_rcs = [httplib.ACCEPTED, httplib.PARTIAL_CONTENT, httplib.OK]
    if not resp or (resp.status_code not in valid_rcs):
            print _("ERROR: Not collecting validator output, "
                    "as execution failed.")
    else:
        response = json.loads(resp.text)
        execution_in_progress = response['executing-tasks']
    return response, execution_in_progress


def collect_validator_output(session, validator_id, dinfo):
    try:
        if validator_id:
            url = 'v1/validate/result/{0}'.format(validator_id)

            response, exec_in_progress = _get_validator_response(session, url)
            while len(exec_in_progress) != 0:
                print _("INFO: Waiting for validator's execution to finish...")
                time.sleep(10)
                response, exec_in_progress = _get_validator_response(session,
                                                                     url)

            if response:
                # collect validator information only if successful
                print _("INFO: Collecting validator information...")
                dinfo_powervcvalidator = "{0}/ttv-validation".format(dinfo)
                utils.createdirs(dinfo_powervcvalidator)
                fname = "{0}/powervc-diag-" \
                    "validate.info".format(dinfo_powervcvalidator)
                with open(fname, "w") as finfo:
                    finfo.write(str(response))
        else:
            print _("ERROR: Not collecting validator information, "
                    "as execution failed.")

    except Exception as ex:
        print _("ERROR: Caught exception while collecting validator "
                "output, check logs for more information")
        LOG.error(_("Caught an unexpected exception while "
                    "collecting validator output. %s") % unicode(ex))


# Capture generic diagnostic data from the PowerVC controller
def capture_generic_data(dinfo, maxwaittime, dtmp):
    collect_product_version(dtmp)
    collect_platform_info(dinfo)
    collect_service_status(dinfo)
    security.collect_powervc_user(dinfo)
    collect_python_info(dinfo, maxwaittime)
    collect_system_info(dinfo)
    collect_db_info(dtmp)
    collect_rabbit_info(dinfo, maxwaittime, dtmp)
    logs.collect_http_logs(dtmp)


def collect_error_log_summary(dtmp, maxwaittime):
    print _("INFO: Generating error summary...")
    LOG.debug(_("INFO: Generating error summary..."))
    if debug:
        stime = time.time()
    fname = dtmp + "/var/log/error-summary.log"
    cmd = "for x in {0}; do /bin/find {1}/var/log/$x -type f -name "\
          "'*.log' | xargs grep -H ERROR | "\
          "/usr/bin/uniq -c;done>{3}".format(" ".
          join(constants.vservices),
          dtmp, "{}", fname)
    stsOK, cout, cerr = utils.runcmd(cmd, wtime=maxwaittime)
    if debug:
        print _("DEBUG: Generated error summary "
                "in {0} seconds.").format(str(round(time.time() - stime,
                                                    3)))


def collect_rabbit_info(dinfo, maxwaittime, dtmp):
    # Collect conf
    conffiles_rabbit = "/etc/rabbitmq/*"
    utils.filteredfcopy(conffiles_rabbit, "password", debug, dtmp)

    # Collectlogs
    logs_rabbit = "/var/log/rabbitmq/*"
    utils.fcopy(logs_rabbit, dtmp)

    # Collect live queueinfo
    dinfo_rabbit = "{0}/rabbit".format(dinfo)
    utils.createdirs(dinfo_rabbit)
    fname_rabbit = dinfo_rabbit + "/rabbit_report.info"
    if len(glob.glob('/etc/rabbitmq/*')):
        cmd = "/usr/sbin/rabbitmqctl report &> {0}".format(fname_rabbit)
        utils.runcmd(cmd, wtime=maxwaittime)


def collect_db_info(dtmp):
        # collect database server information (log files, info)
    print _("INFO: Collecting database server log files and information...")
    if debug:
        stime = time.time()
    db_logfile = "/var/log/mariadb/mariadb.log*"
    utils.fcopy(db_logfile, dtmp)
    if debug:
        print _("DEBUG: Collected database server log files and information "
                "in {0} seconds.").format(str(round(time.time() - stime, 3)))


def collect_system_info(dinfo):
        # collect detailed system information...
    print _("INFO: Collecting detailed system information...")
    if debug:
        stime = time.time()
    fname = dinfo + '/systemdetail.info'
    with open(fname, "w") as finfo:
        finfo.write("Detailed System Information\n")
        finfo.write("===========================\n\n")
    utils.sysdetail(fname, "/bin/date")
    utils.sysdetail(fname, "/usr/bin/id")
    utils.sysdetail(fname, "/bin/ps auxw")
    utils.sysdetail(fname, "/bin/df -l")
    utils.sysdetail(fname, "/bin/mount")
    utils.sysdetail(fname, "/bin/cat /etc/fstab")
    utils.sysdetail(fname, "/bin/ls -ld /home /opt /var")
    utils.sysdetail(fname, "/bin/netstat -r")
    utils.sysdetail(fname, "/usr/bin/free")
    utils.sysdetail(fname, "/usr/bin/systemctl list-unit-files")
    utils.sysdetail(fname, "/sbin/arp -a")
    utils.sysdetail(fname, "/bin/cat /etc/resolv.conf")
    utils.sysdetail(fname, "/bin/cat /etc/hosts")
    utils.sysdetail(fname, "/bin/cat /proc/version")
    utils.sysdetail(fname, "/bin/cat /etc/*-release")
    utils.sysdetail(fname, "/bin/cat /proc/meminfo")
    utils.sysdetail(fname, "/bin/cat /proc/cpuinfo")
    utils.sysdetail(fname, "/bin/uname -a")
    utils.sysdetail(fname, "/sbin/ifconfig -a")
    utils.sysdetail(fname, "/usr/bin/locale")
    utils.sysdetail(fname, "set | /bin/egrep -v  \"OS_PASSWORD | OS_TOKEN |"
                    "OS_AUTH_TOKEN |OS_SERVICE_TOKEN \"|/bin/sort")
    utils.sysdetail(fname, "/usr/sbin/sestatus")
    utils.sysdetail(fname, "/usr/sbin/semodule -l")
    # Fix for defect 42667
    utils.sysdetail_getnetstat(fname, "/bin/netstat -na")
    fname_lsof = dinfo + '/lsof_output.info'
    utils.sysdetail(fname_lsof, "/usr/sbin/lsof > {0}".format(fname_lsof))
    if debug:
        print _("DEBUG: Collected detailed system information "
                "in {0} seconds.").format(str(round(time.time() - stime, 3)))


def collect_python_info(dinfo, maxwaittime):
    cmd = "/bin/ps ax | /bin/grep python"
    stsOK, cout, cerr = utils.runcmd(cmd, wtime=maxwaittime)
    if stsOK and cout is not None:
        fname = dinfo + '/pythonprocesses.info'
        with open(fname, "w") as finfo:
            finfo.write(cout)
    if debug:
        print _("DEBUG: Collected python information ")


def collect_service_status(dinfo):
        # collect runtime information
    print _("INFO: Collecting runtime information...")
    if debug:
        stime = time.time()
    cmd = "/opt/ibm/powervc/bin/powervc-services status"
    LOG.debug(_("Executing {0}").format(cmd))
    cout = commands.getoutput(cmd)
    if cout is not None:
        fname = dinfo + '/powervcservices.info'
        with open(fname, "w") as finfo:
            finfo.write(cout)
    else:
        print _("ERROR: The command \"{0}\" failed to run.").format(cmd)


def collect_product_version(dtmp):
        # copy version information
    print _("INFO: Collecting product version information...")
    if debug:
        stime = time.time()
    dinst = "/opt/ibm/powervc"
    instinfo_osee = "{0}/osee.properties".format(dinst)
    instinfo_powervc = "{0}/version.properties".format(dinst)
    utils.fcopy(instinfo_osee, dtmp)
    utils.fcopy(instinfo_powervc, dtmp)
    if debug:
        print _("DEBUG: Collected product version information "
                "in {0} seconds.").format(str(round(time.time() - stime, 3)))
    print _("INFO: Collecting product install, powervc-config,"
            " backup and restore logs...")
    if debug:
        stime = time.time()
    logfiles_powervc = "/opt/ibm/powervc/log/*"
    utils.fcopy(logfiles_powervc, dtmp)
    if debug:
        print _("DEBUG: Collected product install, powervc-config,"
                " backup and restore logs "
                "in {0} seconds.").format(str(round(time.time() - stime, 3)))


def collect_platform_info(dinfo):
        # copy installation information
    print _("INFO: Collecting installation information...")
    if debug:
        stime = time.time()
    utils.createdirs(dinfo)
    utils.getinstallinfo(dinfo)
    if debug:
        print _("DEBUG: Collected installation information "
                "in {0} seconds.").format(str(round(time.time() - stime, 3)))
