# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
import logging
import os
import time
from powervc_oslo.config import data_utils
from powervc_ras.cli.diag_utils import utils
from sqlalchemy import create_engine
from powervc_ras import _

debug = False
if os.getenv('POWERVCDIAG_DEBUG') is not None:
    debug = True

LOG = logging.getLogger(__name__)


# DB utility
def dbexport(dname, sname, dcur, exclude_tables=None):
    """Export the specified service (i.e. nova) database to the specified
    directory.
    @param: dname - directory name to export database to
    @param: sname - name of service (i.e. nova)
    @param: dcur - the current directory that the command is run from
    @param: exclude_tables - list of table names to exclude from export
    """
    # Dump out the data from the given database to a file in the directory
    if utils.createdirs(dname):
        outfile = dname + '/' + sname + '_export.sql'
        data_utils.export_database_data(outfile, sname, exclude_tables)
    else:
        print _("ERROR: Unable to create the archive output directory "
                "{0} or the user invoking the command does not have write "
                "authority to the directory. Unable to export and collect "
                "database.").format(dname)


def executeOnDB(sqlconnction, query):
    """Executes a sql query on DB and returns the result"""
    engine = create_engine(sqlconnction, echo=False)
    global connection
    connection = engine.connect()
    result = connection.execute(query)
    # connection.close()
    return result


def exporttables(dbname, schema, tables, directory, dcur):
    """
    Exports database tables to CSV files
    @param dbname the database name connect to
    @param schema the schema name to set
    @param tables a list of table names to export
    @param directory the directory to write the CSV files into
    """
    dbdir = data_utils.get_config_data()['db_location']
    if debug:
        print (_("Exporting tables %s") % tables)
        LOG.debug(_("Exporting tables {0}").format(tables))
    # Since the Systemd service for MariaDB uses a Private /tmp,
    # we need to write the data elsewhere initially where the
    # mysql user has authority and transfer back to the output dir
    tmpdir = data_utils.create_unique_temp_directory(dbdir)
    data_utils.chown(tmpdir, 'mysql', 'mysql')
    try:
        statements = list()
        # Loop through each of the tables, executing the export statements
        for table in tables:
            stmt = "SELECT * INTO OUTFILE '%s/%s.csv' FIELDS TERMINATED BY "\
                "',' OPTIONALLY ENCLOSED BY '\"' LINES TERMINATED BY '\n' "\
                "FROM %s.%s" % (tmpdir, table.lower(), dbname, table.lower())
            statements.append(stmt)
        data_utils.batch_database_execute(statements)
        # Now we can move back the files to the real output directory
        for fname in data_utils.get_matching_files(tmpdir, '', '.csv'):
            data_utils.move_file(tmpdir + '/' + fname, directory + '/' + fname)
    finally:
        data_utils.remove_directory(tmpdir)


def novadb(ddb, dcur):
    """Export and collect nova database.
    """

    print _("INFO: Exporting and collecting nova (compute) database...")
    if debug:
        stime = time.time()
    dname_novadb = "{0}/nova".format(ddb)
    dbexport(dname_novadb, "nova", dcur)

    tables = ['INSTANCES', 'INSTANCE_SYSTEM_METADATA', 'BLOCK_DEVICE_MAPPING',
              'STORAGE_CONNECTIVITY_GROUP', 'IBM_HMC_HOSTS',
              'IBM_HMC_HOST_CLUSTERS', 'PORT_FC', 'SCG_VIOS_ASSOCIATION',
              'ADAPTER_ETHERNET_SHARED', 'ADAPTER_ETHERNET_VIRTUAL',
              'SERVER_VIO', 'HOST_NETWORK_ASSOCIATION', 'INSTANCE_FAULTS',
              'IBM_CONSOLE_MESSAGES', 'AGGREGATES', 'AGGREGATE_HOSTS',
              'AGGREGATE_METADATA', 'INSTANCE_GROUPS', 'INSTANCE_GROUP_MEMBER',
              'INSTANCE_GROUP_POLICY']
    exporttables('nova', 'nova', tables, dname_novadb, dcur)
    if debug:
        print _("DEBUG: Exported and collected nova (compute) database "
                "in {0} seconds.").format(str(round(time.time() - stime, 3)))


def cinderdb(ddb, dcur):
    """Export and collect cinder database.
    """

    print _("INFO: Exporting and collecting cinder (storage) database...")
    if debug:
        stime = time.time()
    dname_cinderdb = "{0}/cinder".format(ddb)
    dbexport(dname_cinderdb, "cinder", dcur)
    tables = ['IBM_HMC_CLUSTERS', 'ONBOARD_TASKS', 'ONBOARD_TASK_VOLUMES',
              'SERVICES', 'STORAGE_NODES', 'STORAGE_NODE_HEALTH_STATUS',
              'VOLUME_ADMIN_METADATA', 'VOLUME_ADMIN_METADATA',
              'VOLUME_HEALTH_STATUS', 'VOLUME_METADATA',
              'VOLUME_RESTRICTED_METADATA', 'VOLUME_TYPES',
              'VOLUME_TYPE_EXTRA_SPECS', 'volumes']

    exporttables('cinder', 'cinder', tables, dname_cinderdb, dcur)
    if debug:
        print _("DEBUG: Exported and collected cinder (storage) database "
                "in {0} seconds.").format(str(round(time.time() - stime, 3)))


def glancedb(ddb, dcur):
    """Export and collect glance database.
    """

    print _("INFO: Exporting and collecting glance (image) database...")
    if debug:
        stime = time.time()
    dname_glancedb = "{0}/glance".format(ddb)
    dbexport(dname_glancedb, "glance", dcur)
    tables = ['IMAGES', 'IMAGE_PROPERTIES']

    exporttables('glance', 'glance', tables, dname_glancedb, dcur)
    if debug:
        print _("DEBUG: Exported and collected glance (image) database "
                "in {0} seconds.").format(str(round(time.time() - stime, 3)))


def neutrondb(ddb, dcur):
    """Export and collect neutron database.
    """

    print _("INFO: Exporting and collecting neutron (networking) database...")
    if debug:
        stime = time.time()
    dname_neutrondb = "{0}/neutron".format(ddb)
    # powervc neutron database is named qtm_ibm
    dbexport(dname_neutrondb, "qtm_ibm", dcur)
    tables = ['networks', 'ports', 'subnets']
    exporttables('qtm_ibm', 'neutron', tables, dname_neutrondb, dcur)
    if debug:
        print _("DEBUG: Exported and collected neutron (networking) database "
                "in {0} seconds.").format(str(round(time.time() - stime, 3)))


def keystonedb(ddb, dcur):
    """Export and collect keystone database.
    """

    print _("INFO: Exporting and collecting keystone (security) database...")
    if debug:
        stime = time.time()
    dname_keystonedb = "{0}/keystone".format(ddb)
    # Export the entire DB except for the Token table for security reasons
    dbexport(dname_keystonedb, "keystone", dcur, ['token'])
    if debug:
        print _("DEBUG: Exported and collected keystone (security) database "
                "in {0} seconds.").format(str(round(time.time() - stime, 3)))


def clerkdb(ddb, dcur):
    """Export and collect clerk database.
    """

    print _("INFO: Exporting and collecting clerk (powervc-clerk) database...")
    if debug:
        stime = time.time()
    dname_clerkdb = "{0}/clerk".format(ddb)
    dbexport(dname_clerkdb, "clerk", dcur)
    tables = ['POLICIES', 'IMAGE_TEMPLATES', 'REQUESTS']

    exporttables('clerk', 'clerk', tables, dname_clerkdb, dcur)
    if debug:
        print _("DEBUG: Exported and collected clerk (powervc-clerk) database "
                "in {0} seconds.").format(str(round(time.time() - stime, 3)))


def capture_db(service, ddb, dcur):
    if service in ("bumblebee", "ceilometer", "ttv-validation", "swift"):
        return

    service_data = service + 'db'
    globals()[service_data](ddb, dcur)
