# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
import commands
import ConfigParser
import datetime
import os
import platform
import rpm
import time
import sys
import pwd
import shutil
import subprocess
import socket
import logging
import atexit
import locale
import codecs
import pexpect
import paramiko
from powervc_keystone.encrypthandler import EncryptHandler
from powervc_ras.cli.diag_utils import constants
from powervc_ras import _
LOG = logging.getLogger(__name__)
debug = False
if os.getenv('POWERVCDIAG_DEBUG') is not None:
    debug = True

maxwaittime = 300
powerkvm = False
offeringHMC = False
dcur = None
dtmp = None
vservices = constants.vservices


def ispowerKVM():
    global powerkvm
    if not powerkvm:
        powerkvm = checkOffering('powerkvm', dtmp)
    return powerkvm


def isofferingHMC():
    global offeringHMC
    if not offeringHMC:
        offeringHMC = checkOffering('standard', dtmp)
    return offeringHMC


def checkOffering(offering, dtmp):
    """ Helper method to get offering """
    try:
        with open("/opt/ibm/powervc/version.properties", "r") as f:
            lines = f.read()
        value = getValue('offering', lines)
        if value is not None and value == offering:
            powerkvm = True
        else:
            powerkvm = False
    except:
        # attempt to cleanup archive output directory
        if dtmp is not None:
            shutil.rmtree(dtmp, ignore_errors=True)
        LOG.error(_('Error: The file /opt/ibm/powervc/version.properties'
                  ' either does not exist or has been deleted.'))
        cmdexit(_('Error: The file /opt/ibm/powervc/version.properties'
                ' either does not exist or has been deleted.'), 1)

    return powerkvm


def getValue(key, buffer):
    """ helper method to fnd the value of key in buffer"""
    match = None
    for line in buffer.splitlines():
        if key in line:
            match = line
            break
    if match is not None:
        return match[match.find("=") + 2:].strip()
    else:
        return None


def cmdexit(msg, status):
    """Exit command with specified message and exit code.
    @param: msg - Exit message
    @param: status - Exit code (0 is successful, non-zero indicates failure)
    """
    print (msg)
    sys.exit(status)


# Security related utilities
def decrypt(text):
    """Convert an encrypted string to a clear text one. Clear text will be
    unchanged"""
    if text is None:
        return None
    encrypt_handler = EncryptHandler()
    return encrypt_handler.decode(text)


# Command execution Utlities
def runrcmd(ssh, cmd, timeout):
    chan = ssh.get_transport().open_session()
    chan.exec_command(cmd)
    stdout = chan.recv(1024)
    status = 1
    maxwait = timeout
    while maxwait > 0:
        if chan.exit_status_ready():
            status = chan.recv_exit_status()
            break
        else:
            time.sleep(10)
            maxwait = maxwait - 10
    if maxwait <= 0:
        status = -99
    return status, stdout


def runcmd(cmd, ignoreerrors=False, wtime=maxwaittime,
           sout=subprocess.PIPE, serr=subprocess.PIPE):
    LOG.debug(_("Executing {0}").format(cmd))
    """
    Start and wait for a command to complete for a specified amount of time.
    The subprocess will be killed if it does not complete within
    the specified amount of time.
    @param: cmd - the command to run
    @param: wtime - total amount of time to wait (default: 300 seconds)
    @param: sout - standard output target (default: pipe)
    @param: serr - standard error target (default: pipe)
    @return: a tuple
        stsOK - command execution boolean status
        cout - standard output from the command execution
        cerr - standard error from the command execution
    """
    cout = None
    cerr = None
    stsOK, sproc = startcmd(cmd, ignoreerrors, sout, serr)
    if stsOK and sproc:
        stsOK, cout, cerr = waitforcmd(sproc, time.time(), wtime, ignoreerrors)
    else:
        stsOK = False
    if ((not stsOK) and (not ignoreerrors)):
        print _("ERROR: The command \"{0}\" failed. "
                "Retry with increased value for -w option").format(cmd)
    LOG.debug(_("Command execution boolean status is - {0}").format(stsOK))
    if cout:
        cout = cout.decode('ascii', 'ignore')
    LOG.debug(_("Standard output from the command execution - {0}").
              format(cout))
    if cerr:
        cerr = cerr.decode('ascii', 'ignore')
    LOG.debug(_("Standard error from the command execution - {0}").
              format(cerr))
    return (stsOK, cout, cerr)


def waitforcmd(sproc, stime=time.time(), wtime=maxwaittime,
               ignoreerrors=False, pint=1):
    """
    Wait for a subprocess to complete for a specified amount of time.
    The subprocess will be killed if it does not complete within
    the specified amount of time.
    @param: sproc - subprocess instance
    @param: stime - time the subprocess was started
    @param: wtime - total amount of time to wait (after specified start time)
    @param: ignoreerrors - if True, will suppress error messages
    @param: pint - polling interval
    @return: a tuple
        stsOK - command execution boolean status
        cout - standard output from the command execution
        cerr - standard error from the command execution
    """

    stsOK = True
    if wtime is not None:
        ftime = stime + wtime
        while sproc.poll() is None and ftime > time.time():
            time.sleep(pint)
        # if the reason we exited while is because we exceeded
        # timeout, kill the process and print a message
        if sproc.poll() is None and ftime <= time.time():
            stsOK = False
            sproc.kill()
            if not ignoreerrors:
                print _("ERROR: The command did not complete within the "
                        "specified timeout of {0} seconds.").format(wtime)
            stsOK = False
            cout = None
            cerr = None
        else:
            cout, cerr = sproc.communicate()
            if sproc.returncode != 0:
                stsOK = False
    return (stsOK, cout, cerr)


def startcmd(cmd, ignoreerrors=False,
             sout=subprocess.PIPE, serr=subprocess.PIPE):
    """
    Start a command using subprocess.Popen and return
    the subprocess instance.
    @param: cmd - the command to run
    @param: ignoreerrors - if True, will suppress error messages
    @param: sout - standard output target (default: pipe)
    @param: serr - standard error target (default: pipe)
    @return: a tuple
        stsOK - command execution boolean status
        sproc - subprocess instance
    """

    stsOK = True
    sproc = None
    try:
        current_env = os.environ.copy()
        sproc = subprocess.Popen(cmd, shell=True, env=current_env,
                                 stdout=sout, stderr=serr)
    except Exception:
        stsOK = False
    if sproc is not None:
        if sproc.poll() is not None and sproc.returncode != 0:
            stsOK = False
    else:
        stsOK = False
    if not stsOK:
        if not ignoreerrors:
            print _("ERROR: The command \"{0}\" failed.").format(cmd)

    return (stsOK, sproc)


# General utlity files
def collect_remote_files(host, remotepath, localpath, maxwaittime, ip):
    try:
        if str(host['passwd']).strip() != '':
            result = socket.getaddrinfo(ip, 80, 0,
                                        socket.SOCK_STREAM)
            socket_family = result[0][0]
            if socket_family == socket.AF_INET6:
                scp_command = '/usr/bin/scp -r -6'
                ip = '[' + ip + ']'
            else:
                scp_command = '/usr/bin/scp -r'
            p = pexpect.spawn('{0} -o StrictHostKeyChecking=yes -o \
                             UserKnownHostsFile={1} {2}@{3}:{4}  {5} '.format(
                scp_command,
                constants.known_hosts_file,
                host['user'],
                ip,
                remotepath,
                localpath))
            result = p.expect(['[pP]assword:', pexpect.EOF],
                              timeout=maxwaittime)
            if result == 0:
                p.sendline(host['passwd'])
                p.expect(pexpect.EOF, timeout=maxwaittime)
            elif result == 1:
                print _("Error:Connection to {0}  "
                        "host is timed out").format(host['ip'])
            else:
                print _("Error:Failed to collect remote data from host"
                        " {0}").format(host['ip'])
        else:
            # Fix for defect 46751
            try:
                port = 22
                transport = paramiko.Transport((host['ip'], port))
                key = paramiko.RSAKey.from_private_key_file(host['key_file'])
                transport.connect(username=host['user'], pkey=key)
                sftp = paramiko.SFTPClient.from_transport(transport)
                if remotepath.endswith('*'):
                    remotepath = remotepath.rstrip('*')
                    for fileattr in sftp.listdir_attr(remotepath):
                        filepath = remotepath + fileattr.filename
                        if os.path.isdir(filepath):
                            continue
                        else:
                            # If we are here, then its a file
                            destpath = localpath + '/' + fileattr.filename
                            sftp.get(filepath, destpath)
                elif remotepath.endswith('syslog'):
                    destpath = localpath + '/syslog'
                    sftp.get(remotepath, destpath)
                elif remotepath.endswith('messages'):
                    destpath = localpath + '/messages'
                    sftp.get(remotepath, destpath)
                else:
                    pass
            except Exception as e:
                print _("Error:Failed to collect remote data from host"
                        " {0}, {1}").format(host['ip'], e)

    except pexpect.TIMEOUT:
        p.close()
        print _("Error: Remote {0} logs could not be collected from {1} in"
                " specified time {2}").format(remotepath,
                                              host[
                                              'ip'],
                                              maxwaittime)

    except Exception as e:
        print _("Error:Failed to collect data from {0} on {1} "
                "host").format(remotepath, host['ip'])


def rfcopy(source, destination):
    """ helper method to copy agent logs file """
    createdirs(destination)
    cmd = '/bin/cp --preserve  {0} {1}'.format(source, destination)
    runcmd(cmd, ignoreerrors=True, wtime=maxwaittime)


def createdirs(dname):
    """Create the specified directory along with any parent
    directories.
    """

    dirOK = os.path.isdir(dname) and os.access(dname, os.W_OK)
    if not os.path.isdir(dname):
        try:
            os.makedirs(dname)
            dirOK = True
        except IOError as err:
            dirOK = False
            sys.stderr.write(_('ERROR: %s\n') % str(err))
    return dirOK


# Signal Handler
def sighdlr(snum, sfrm, dtmp):
    """Signal handler which will cleanly handle ctrl-c.
    @param: snum - signal number
    @param: sfrm - stack frame
    """

    # attempt to cleanup archive output directory
    if dtmp is not None:
        shutil.rmtree(dtmp, ignore_errors=True)

    # signal handler which will cleanly handle ctrl-c
    LOG.error(_("INFO: User pressed ctrl-c to stop the command."))
    cmdexit(_("INFO: User pressed ctrl-c to stop the command."), 0)


def filteredfcopy(fspec, fpat, debug, dtmp, frpat=None):
    """Copy files matching the file specification and filter
        out any lines from files which match the specified pattern.
    @param: fspec - file specification that will be used by find command
    @param: fpat - filter pattern that will be used by grep command to
        exclude lines from files being copied
    """

    cmd = "/bin/find {0}".format(fspec)
    stsOK, cout, cerr = runcmd(cmd, ignoreerrors=True, wtime=maxwaittime)
    if stsOK and cout is not None:
        for fname in cout.rstrip().split('\n'):
            if frpat:
                with open(fname, "r") as f:
                    lines = f.read()
                found = getValue(frpat, lines)
                if found:
                    continue
            dname_out = dtmp + os.path.dirname(fname)
            createdirs(dname_out)
            fname_out = "{0}/{1}".format(dname_out, os.path.basename(fname))
            cmd = '/bin/grep -iv "{0}" {1}'.format(fpat, fname)
            if debug:
                cmdstime = time.time()
                print _("DEBUG: In filteredfcopy(), running command "
                        "\'{0}\'").format(cmd)
            # stsOK, cout, cerr = runcmd(cmd, ignoreerrors=True)
            # if stsOK and cout is not None:
            with open(fname_out, "w") as fconf:
                stsOK, cout, cerr = runcmd(cmd, ignoreerrors=True, sout=fconf,
                                           wtime=maxwaittime)
                # fconf.write(cout)
            if debug:
                msg = "DEBUG: In filteredfcopy(), command \'{0}\' ran " \
                    "in {1} seconds."
                print _(msg.format(cmd,
                                   str(round(time.time() - cmdstime, 3))))


def fcopy(fspec, dtmp):
    """Copy files matching the file specification.
    @param: fspec - file specification that will be used by find command
    """

    cmd = "/bin/find {0}".format(fspec)
    stsOK, cout, cerr = runcmd(cmd, ignoreerrors=True, wtime=maxwaittime)
    if stsOK and cout is not None:
        for fname in cout.rstrip().split('\n'):
            cmd = '/bin/cp --preserve --parents {0} {1}'.format(fname, dtmp)
            runcmd(cmd, ignoreerrors=True, wtime=maxwaittime)


def sysdetail(fname, cmd):
    stsOK, cout, cerr = runcmd(cmd, wtime=maxwaittime)
    if stsOK and cout is not None:
        with open(fname, "a") as finfo:
            finfo.write("Output from: {0}\n".format(cmd))
            finfo.write(cout)
            finfo.write("\n\n")


def sysdetail_getnetstat(fname, cmd):
    cout = commands.getoutput(cmd)
    if cout is not None:
        with open(fname, "a") as finfo:
            finfo.write("Output from: {0}\n".format(cmd))
            finfo.write(cout)
            finfo.write("\n\n")


# Process lock related utlity
def remove_lock(spid):
    """Remove the powervc-diag process lock file.
    directory.
    @param: spid - if specified, will only delete lock file if it contains
        matching pid.
    """

    lfname = "/tmp/powervc-diag.lck"
    pid = None
    rmok = True

    if spid is not None:
        with open(lfname) as lf:
            pid = lf.readline().strip()

    if str(pid) != str(spid):
        rmok = False

    if rmok:
        lfname = "/tmp/powervc-diag.lck"
        cmd = '/bin/rm -f {0}'.format(lfname)
        runcmd(cmd, ignoreerrors=True, wtime=maxwaittime)


# OS related Utlity
def getinstallinfo(dinfo):
    """Get installation information.  This is OS-specific.  Therefore
    will need to make OS check and call OS-specific command to get
    information.
    """
    plt = platform.platform()
    if "redhat" in str(plt).lower():
        # rpm -qa
        tset = rpm.TransactionSet()
        miter = tset.dbMatch()
        fname = dinfo + '/rpmqalist.info'
        with open(fname, "w") as finfo:
            pkglist = []
            for hobj in miter:
                pkglist.append("%s-%s-%s\n" % (hobj['name'],
                                               hobj['version'],
                                               hobj['release']))
            for pkg in sorted(pkglist):
                finfo.write(pkg)
    else:
        print _("WARNING: Unrecognized operating system ({0}) detected. "
                "Unable to collect operating system specific install "
                "information.").format(plt)


# Config relate Utlity
def get_value(conf, section, option):
    try:
        config = ConfigParser.ConfigParser()
        config.optionxform = str
        config.read(conf)
        return config.get(section, option)
    except ConfigParser.NoOptionError:
        return None


# Logger related utility
def initialize_logger():
    """Initialize the log object.
    Log filename and the entries will have timestamp
    """
    global timestamp
    dt = datetime.datetime.now()
    # Configure Logging
    if not os.path.exists(constants.LOG_FILE_BASENAME):
        os.makedirs(constants.LOG_FILE_BASENAME)
    global log_filename
    timestamp = "%s-%s-%s-%s-%s-%s-%s" % (dt.year,
                                          dt.month,
                                          dt.day,
                                          dt.hour,
                                          dt.minute,
                                          dt.second,
                                          dt.microsecond)
    log_filename = "%s/powervc-diag_%s.log" % (constants.LOG_FILE_BASENAME,
                                               timestamp)

    loglevel_number = getattr(logging, "DEBUG", None)

    logging.basicConfig(level=loglevel_number,
                        format='%(asctime)s %(name)-12s '
                        '%(levelname)-8s %(message)s',
                        datefmt='%m-%d %H:%M',
                        filename=log_filename,
                        filemode='w')
    # define a Handler which writes INFO messages or higher to sys.stderr
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    # set a format which is simpler for console use
    formatter = logging.Formatter('%(message)s')
    # tell the handler to use this format
    console.setFormatter(formatter)
    # add the handler to the root logger
    logging.getLogger(__name__).addHandler(console)
    LOG.debug("log filename = %s" % log_filename)


# DNS Utlity
def resolveHostname(nameOrIP):
    """Resolve the passed hostname to an IP address. If the passed name is
    already an IP address it returns that back.

    Returns None if the name is unresolvable."""
    try:
        result = socket.getaddrinfo(nameOrIP, 80, 0,
                                    socket.SOCK_STREAM)
        if result:
            ip_address = result[0][4][0]
            return ip_address
    except:
        return None


def exit_running_proc():
    output_captured = False
    check_lock_file = True
    sys_args = sys.argv
    if sys_args is not None:
        if "-h" in sys_args or "--help" in sys_args:
            check_lock_file = False
    if check_lock_file:
        # check for process lock file
        lfname = "/tmp/powervc-diag.lck"
        lfexists = None
        pidstale = None

        try:
            with open(lfname):
                lfexists = True
        except IOError:
            lfexists = False

        # if lock file exists, ensure it's not stale
        if lfexists:
            # open the file and see if the pid is valid
            # note: it is possible that the pid is in use for
            # a process *not* running powervc-diag; a
            # console message will be output with
            # necessary information/instructions so that
            # user can deal with this rare situation
            with open(lfname) as lf:
                pid = lf.readline().strip()
                if not pid:
                    pidstale = True
                else:
                    try:
                        os.kill(int(pid), 0)
                        pidstale = False
                    except OSError:
                        pidstale = True
        else:
            pidstale = True

        if not pidstale:
            # if not stale, exit with message
            msg = _('INFO: The powervc-diag command is already running '
                    '(process id:{0}).  Wait for that process to finish \
                    before '
                    'invoking powervc-diag.  If the indicated process is not '
                    'running powervc-diag, you can remove the file {1} and '
                    'run powervc-diag again.').format(pid, lfname)
            LOG.error(_('INFO: The powervc-diag command is already running '
                        '(process id:{0}).  Wait for that process to \
                      finish before '
                        'invoking powervc-diag.  If the indicated \
                      process is not '
                        'running powervc-diag, you can remove the file {1} \
                        and '
                        'run powervc-diag again.').format(pid, lfname))
            cmdexit(msg, 1)

        else:
            # if stale, remove lock file and continue
            remove_lock(None)

        # create lock file to prevent concurrent execution
        pid = os.getpid()
        with open(lfname, 'w') as lf:
            lf.write("{0}\n".format(pid))

        # register exit handler to clean up lock file
        atexit.register(remove_lock, pid)

        # redirect output to text file
        # Unbuffer output
        sys.stdout = os.fdopen(sys.stdout.fileno(), 'w', 0)
        # set encoding
        sys.stdout = codecs.getwriter(
            locale.getpreferredencoding())(sys.stdout)
        try:
            outputfile = "/tmp/output.txt"
            tee = subprocess.Popen(["tee", outputfile], stdin=subprocess.PIPE)
            os.dup2(tee.stdin.fileno(), sys.stdout.fileno())
            os.dup2(tee.stdin.fileno(), sys.stderr.fileno())
            output_captured = True
        except:
            print _("WARNING: The output file:\"output.txt\" will not be "
                    "created, tee command to redirect stdout/stderr failed")
            output_captured = False
    return output_captured


# Archive related Utilities
def create_final_archive(output_captured, archive_file, dtmp, maxwaittime,
                         dcur,
                         dout):
    # copy output file
    if output_captured:
        shutil.move(constants.outputfile, "{0}".format(dtmp))
    # Changing file permission
    cmd = '/bin/chmod -R 500 {0}/*'.format(dtmp)
    runcmd(cmd, ignoreerrors=True, wtime=maxwaittime)
    cmd = '/bin/chmod 400 `/bin/find {0} -type f`'.format(dtmp)
    runcmd(cmd, ignoreerrors=True, wtime=maxwaittime)

    # generate archive
    if debug:
        stime = time.time()
    os.chdir(dout)
    cmd = "/bin/tar --ignore-failed-read --atime-preserve -czf {0}.tgz " \
        "{1}".format(archive_file, archive_file)
    stsOK, cout, cerr = runcmd(cmd, wtime=maxwaittime)
    if debug:
        print _("DEBUG: Generated archive file "
                "in {0} seconds.").format(str(round(time.time() - stime, 3)))

    # Changing file permission
    cmd = "/bin/chmod -R 600 {0}.tgz".format(archive_file)
    runcmd(cmd, ignoreerrors=True, wtime=maxwaittime)

    # cleanup
    if debug:
        stime = time.time()
    shutil.rmtree(dtmp, ignore_errors=True)
    os.chdir(dcur)
    if debug:
        print _("DEBUG: Completed cleanup "
                "in {0} seconds.").format(str(round(time.time() - stime, 3)))

    # print message and exit
    if stsOK:
        cmdexit(_("\nINFO: Diagnostic data collection is complete.\nINFO: "
                "Archive file {0}/{1}.tgz was created."
                "\n").format(dout, archive_file), 0)
    else:
        cmdexit(_("\nERROR: Archive file {0}/{1}.tgz could not be "
                "created.").format(dout, archive_file), 1)


def delete_old_archives(maxwaittime, dout, noprompt):
    # check for existing archives and present option to delete
    afspec = "{0}/powervc-diag*.tgz".format(dout)
    cmd = "/bin/find {0}".format(afspec)
    LOG.debug(_("Executing {0}").format(cmd))
    stsOK, cout, cerr = runcmd(cmd, ignoreerrors=True, wtime=maxwaittime)
    afiles_num = 0
    if stsOK and cout is not None and not str(cout) == "":
        afiles = cout.rstrip().split('\n')
        afiles_num = len(afiles)
    if afiles_num > 0:
        print _("\nWARNING: There are {0} existing archive files ({1}) "
                "in output directory {2}.").format(afiles_num, afspec, dout)

        if noprompt:
            print _("\nWARNING: Existing archive files will not be deleted. "
                    "You should delete or move existing archives to avoid "
                    "problems due to lack of available disk space.")
        else:
            choice = "N"
            while choice != "y" and choice != "n":
                sys.stdout.write(_("\nDo you want to delete existing archive "
                                   "files now to increase available disk "
                                   "space? [y/n] "))
                choice = raw_input().lower()
                if choice != "y" and choice != "n":
                    print _("'{0}' is not a valid response.").format(choice)

            if choice == "y":
                cmd = '/bin/rm -f {0}'.format(afspec)
                runcmd(cmd, wtime=maxwaittime)
            else:
                print _("\nWARNING: Existing archive files will not be "
                        "deleted.  You should delete or move existing archives"
                        " to avoid problems due to lack of available "
                        "disk space.")


def create_archive_file_dir(args):
    # setup temporary directory and archive file name
    # check if any overrides were specified: output directory, archive filename
    tcurrent = time.localtime()
    tsuffix = time.strftime('%y%m%d-%H%M%S', tcurrent)
    archive_file = "powervc-diag_{0}".format(tsuffix)
    dout = os.getenv('TMPDIR', '/tmp')
    if args.archive_fname:
        archive_file = str(args.archive_fname[0])
    if args.output_dir:
        dout = str(args.output_dir[0])

    if not createdirs(dout):
        LOG.error(_("ERROR: Unable to create the archive output directory "
                    "{0} or the user invoking the command does not have write "
                    "authority to the directory.").format(dout))
        cmdexit(_("ERROR: Unable to create the archive output directory "
                  "{0} or the user invoking the command does not have write "
                  "authority to the directory.").format(dout), 1)
    if args.runparallel:
        print _("DEPRECRATED: -p option is valid only with -c/-novalink"
                " option")
    return archive_file, dout


# Argument validations
def validate_arguments(args):
    validate_max_time(args)
    validate_service(args)


def validate_max_time(args):
    global maxwaittime
    # check if maxwait time was specified
    if args.maxwait:
        try:
            maxwaittime = int(args.maxwait[0])
        except ValueError:
            print _("WARNING: The value {0} specified for the maxwait "
                    "argument is not a positive integer value. "
                    "The maxwait value "
                    "will be set to the default value.").format(
                        args.maxwait[0])


def validate_service(args):
    global vservices
    # determine what services/data will be processed/collected
    '''if args.services:
        vservices = args.services.split(',')'''
    tmpvservices = list(vservices)
    for service in tmpvservices:
        if service not in constants.vldservices:
            print _("WARNING: {0} is not a valid service name and will be "
                    "ignored.").format(service)
            vservices.remove(service)
    if not vservices:
        LOG.error(_("ERROR: No valid service names were specified."))
        cmdexit(_("ERROR: No valid service names were specified."), 1)


def serviceconf(service, dtmp):
    """Collect service configuration files.
    """
    if service is "bumblebee":
        return

    print _("INFO: Collecting %s (%s) configuration files...") \
        % (service, constants.service_map[service])
    if debug:
        stime = time.time()

    conffiles_service = "/etc/%s/*.conf /etc/%s/*.ini \
        /opt/ibm/powervc/policy/%s/policy.json" % (service, service, service)

    if service is "swift":
        conffiles_service = "/etc/%s/*.conf /etc/%s/*.gz" % (service, service)

    if service is "keystone":
        conffiles_service = conffiles_service + " /etc/keystone/domains/*"

    if service is "ttv-validation":
        conffiles_service = conffiles_service + " /opt/ibm/powervc/"\
            "ttv-validation/conf/*.json"

    filteredfcopy(conffiles_service, "password", debug, dtmp)
    # use filtered copy to strip out any password data that might be present
    if service is "nova":
        filteredfcopy(conffiles_service, "kvm_host_ip", debug, dtmp)
        filteredfcopy(conffiles_service, "powervm_host_ip", debug, dtmp)

    if debug:
        print _("DEBUG: Collected %s (%s) configuration files "
                "in {0} seconds.").format(str(round(time.time() - stime, 3))
                                          % (service, service))


def checkCloudEnabled():
    """ Helper method to check if cloud enabled """
    cloud_enabled = False
    try:
        with open("/opt/ibm/powervc/version.properties", "r") as f:
            lines = f.read()
            if 'cloud_enabled' in lines:
                cloud_enabled = True
            else:
                cloud_enabled = False
    except:
        # attempt to cleanup archive output directory
        if dtmp is not None:
            shutil.rmtree(dtmp, ignore_errors=True)
        LOG.error(_('Error: The file /opt/ibm/powervc/version.properties'
                  ' either does not exist or has been deleted.'))
        cmdexit(_('Error: The file /opt/ibm/powervc/version.properties'
                ' either does not exist or has been deleted.'), 1)

    return cloud_enabled
