# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
import ConfigParser
from threading import Thread
import glob
import os
import paramiko
from oslo_utils import encodeutils
from powervc_ras.cli.diag_utils import utils
from powervc_ras.cli.diag_utils import logs
from powervc_ras.cli.diag_utils import parser as diag_parser
from powervc_ras.cli.diag_utils import security
from powervc_ras.cli.diag_utils import common
from powervc_ras.cli.diag_utils import db
from powervc_ras.cli.diag_utils import info
from powervc_ras.cli.diag_utils import constants
from powervc_ras import _
import logging
import shutil
import signal
import sys
import getpass
import commands
import requests
dcur = None
dtmp = None
dinfo = None
ddb = None
dkvminfo = None
dhmcinfo = None
maxwaittime = 300
debug = False
POWERVC_INSTALL_DIR = '/opt/ibm/powervc'
LOG_FILE_BASENAME = '%s/log/powervc-diag' % POWERVC_INSTALL_DIR
LOG = logging.getLogger(__name__)

if os.getenv('POWERVCDIAG_DEBUG'):
    debug = True
from requests.packages.urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)


def load_remote_host_details(host_type, remote_hosts):
    """ helper method to load novalink/kvm hosts
    information in to a mapping table"""
    # valid_hosts = None
    valid_hosts = {}
    confile_path = "/etc/nova"
    conffiles = glob.glob("{0}/nova-*.conf".format(confile_path))
    for conffile in conffiles:
        with open(conffile, "r") as f:
            lines = f.read()
            remote_host_ip = utils.getValue(
                constants.version_details_map[
                    host_type]['ip'], lines)
            if remote_host_ip is not None:
                hostdetails = {}
                key_file = None
                hostdetails['host_display_name'] = utils.getValue(
                    'host_display_name', lines)

                host_crit = ((remote_host_ip in remote_hosts) or (
                    hostdetails['host_display_name'] in remote_hosts))
                if ('all' not in remote_hosts) and (not host_crit):
                    continue
                user = constants.version_details_map[host_type]['user']
                passwd = constants.version_details_map[host_type]['passwd']
                key_file = constants.version_details_map[host_type]['key_file']
                hostdetails[user] = utils.getValue(user, lines)
                encrypt_passwd = utils.getValue(passwd, lines)
                hostdetails[passwd] = utils.decrypt(encrypt_passwd)
                hostdetails[key_file] = utils.getValue(key_file, lines)
                valid_hosts[remote_host_ip] = remote_host_ip
                valid_hosts[remote_host_ip] = {
                    'user': hostdetails[user],
                    'passwd': hostdetails[passwd],
                    'host_display_name': hostdetails['host_display_name'],
                    'ip': remote_host_ip,
                    'key_file': hostdetails[key_file]}
    return valid_hosts


def collectrfiles(host, remotepath, localpath):
    utils.collect_remote_files(
        host,
        remotepath,
        localpath,
        utils.maxwaittime,
        host['ip'])


def collect_pkvm_novalink_info(host_type, hosts, runparallel):
    """ collect info from all kvm hosts registered with powerVC """
    """ Need to create threads for each collection and run parallely"""
    threads = []
    for host_ip in hosts:
        host = hosts[host_ip]
        if not runparallel:
            collect_remote_diagnostics(host, host_type)
        else:
            host.update(
                {constants.version_details_map[host_type]['ip']: host_ip})
            collection_thread = Thread(
                target=collect_remote_diagnostics,
                args=(host, host_type))
            threads.append(collection_thread)
            collection_thread.start()
    for thread in threads:
        thread.join()


def collectHMCInfo(hmc):
    print _("INFO: Collecting log files from HMC '{0}'").format(
        hmc['hmc_display_name'])
    LOG.debug(_("INFO: Collecting log files from HMC '{0}'").format(
        hmc['hmc_display_name']))
    client_hmc = hmc['client_hmc']
    try:
        localpath = dhmcinfo + '/' + hmc['hmc_display_name']
        collection_cmd = "/usr/hmcrbin/pedbg -c -q 4".format(file)
        LOG.debug(_("Executing {0}").format(collection_cmd))
        status, sout = utils.runrcmd(
            client_hmc, collection_cmd, utils.maxwaittime)
        collection_status = status
        if collection_status == 0:

            # get filename
            ls_cmd = 'cd /dump; ls | /bin/grep HMClogs'
            LOG.debug(_("Executing {0}").format(ls_cmd))
            stdin, stdout, stderr = client_hmc.exec_command(ls_cmd)
            LOG.debug("{0}".format(stdin))
            LOG.debug("{0}".format(stdout))
            LOG.debug("{0}".format(stderr))
            hmcfile = stdout.readlines()[0].rstrip()
            print _("INFO: Copying file {0}"
                    " from HMC {1}").format(hmcfile, hmc['hmc_display_name'])
            utils.createdirs(localpath)
            try:
                sftp = client_hmc.open_sftp()
                sftp.get('/dump/' + hmcfile, localpath + '/' + hmcfile)
            except Exception as e:
                print _('ERROR: Failed to copy file {0} from HMC'
                        ' {1} {2}').format(hmcfile, hmc['hmc_display_name'], e)
            finally:
                sftp.close()
        else:
            print _("Error: Failed to collect remote logs from {0} HMC."
                    " Retry by increasing the maximum amount of time to"
                    " wait using --maxwait \
                    option").format(hmc['hmc_display_name'])
    except Exception as e:
        print _("Error:Failed to collect data from HMC \'{0}\'."
                " {1}").format(hmc['hmc_display_name'], e)

    finally:
        client_hmc.close()


def gethscpepass(hmctarget):
    print _("Enter Password for the \'hscpe\' "
            "account on HMC {0}: \n").format(hmctarget)
    # password = ''
    # while password == '':
    password = getpass.getpass()
    return password


def loadHMC(target, hmc_host_list):
    conf = glob.glob(constants.base_nova_conf_file)
    base_nova_conf = ConfigParser.ConfigParser()
    base_nova_conf.read(conf)
    text = base_nova_conf.get("database", "connection")
    sql_connection = utils.decrypt(text)
    hmc_query = "select * from ibm_hmcs where deleted = 0"
    result = db.executeOnDB(sql_connection, hmc_query)
    HMC = []
    hmc_dict = {}
    for row in result:
        hmc = {}
        hmc["id"] = row['id']
        hmc["uuid"] = row['uuid']
        hmc["hmc_display_name"] = row['display_name']
        ip = utils.resolveHostname(row['access_ip'])
        if not ip:
            continue
        hmc["ip"] = ip
        hmc["hostname"] = row['access_ip']
        HMC.append(hmc)

    if target is "all":
        # callect all data
        for hmc in HMC:
            password = str(gethscpepass(hmc['ip']))
            hmc['password'] = password
            hmc_dict[hmc['hmc_display_name']] = hmc
    else:
        for name in hmc_host_list:
            found = False
            for hmc in HMC:
                if hmc['hmc_display_name'] == name or hmc['ip'] == name:
                    password = str(gethscpepass(name))
                    hmc['password'] = password
                    hmc_dict[name] = hmc
                    found = True
            if not found:
                print (_("HMC: %s not found") % name)
    return hmc_dict


def load_and_capture_diag(input_host, host_type, parallel):
    hosts_valid = []
    remote_hosts = str(input_host)
    remote_hosts = remote_hosts.split(',')
    print _("INFO: Validating remote host information")
    valid_hosts = load_remote_host_details(host_type, remote_hosts)
    for host_ip in valid_hosts:
        host_name = valid_hosts[host_ip]['host_display_name']
        hosts_valid.append(host_name)
        host_ip = valid_hosts[host_ip]['ip']
        hosts_valid.append(host_ip)
    for host in remote_hosts:
        if host == 'all':
            break
        if host not in hosts_valid:
            print (_("WARNING: Host %s not found") % host)
    collect_pkvm_novalink_info(host_type, valid_hosts, parallel)


def collect_remote_diagnostics(host, host_type):
    """ collect host information of specified host """
    print _("INFO: Collecting data from registered hosts ...")
    client = None
    try:
        client = paramiko.SSHClient()
        client.load_host_keys(constants.known_hosts_file)
        client.set_missing_host_key_policy(paramiko.RejectPolicy())
        # Fix for defect 46751
        if str(host['passwd']).strip() != '':
            client.connect(host['ip'],
                           username=host['user'],
                           password=host['passwd'])
        else:
            client.connect(host['ip'],
                           username=host['user'],
                           key_filename=host['key_file'])
        if host_type is 'kvm':
            localpath = dkvminfo + "/" + host['ip']
        if host_type is 'novalink':
            localpath = dneoinfo + "/" + host['ip']

        # Diagnostic collection for PowerKVM host
        if host_type is 'kvm':
            _collect_remote_data_from_pkvm_host(client, localpath, host)

        # Diagnostic collection for NovaLink host
        if host_type is 'novalink':
            _collect_remote_data_from_novalink_host(client, localpath, host)

    except Exception as e:
        print _("Error:Failed to collect data from {0} "
                "host. {1}").format(host['ip'], e)

    finally:
        if client is not None:
            client.close()


def _collect_remote_data_from_pkvm_host(client, localpath, host):
    # Collect OpenVSwitch logs
    try:
        remoteOpenVswitchPath = "/var/log/openvswitch/*"
        openvswitchLocalPath = localpath + '/openvswitch'
        utils.createdirs(openvswitchLocalPath)
        collectrfiles(host, remoteOpenVswitchPath,
                      openvswitchLocalPath)
    except Exception as e:
        print _("ERROR: Failed to collect OpenVSwitch logs. {0}").format(e)

    # Capture kvm GMR
    """ Trigger and collect GMR report for nova-compute process in  host"""
    try:
        print _("INFO: Collecting GMR report for nova-compute "
                "process from  host ...")
        getnovapid_cmd = '/usr/bin/pgrep nova-compute'
        status, nova_pid = utils.runrcmd(client, getnovapid_cmd,
                                         utils.maxwaittime)
        if nova_pid is not None:
            gmrcmd = '/usr/bin/kill -USR1 {0}'.format(nova_pid)
            utils.runrcmd(client, gmrcmd, utils.maxwaittime)
        else:
            print _("ERROR:Unable to get process id of nova-compute "
                    "process")
    except Exception as e:
        print _("ERROR:Failed to collect GMR report for "
                "nova-compute. {0}").format(e)

    # Collect general messages
    try:
        remotepath = "/var/log/messages"
        print _("INFO: Collecting /var/log/messages from {0} "
                "host...").format(host['ip'])
        agentlocalpath = localpath + '/messages'
        utils.createdirs(agentlocalpath)
        collectrfiles(host, remotepath, agentlocalpath)
    except Exception as e:
        print _("ERROR:Failed to collect /var/log/messages. {0}").format(e)

    try:
        for service in ['nova', 'neutron', 'ceilometer']:
            # Collect openstack service agent logs from remote host
            remotepath = ("/var/log/%s/*") % service
            agentlocalpath = localpath + '/var/log/' + service
            print _("INFO: Collecting {0} logs from {1} host...")\
                .format(service, host['ip'])
            utils.createdirs(agentlocalpath)
            collectrfiles(host, remotepath, agentlocalpath)

            # Collect OpenStack service conf files
            remote_conf_path = ("/etc/%s/*") % service
            print _("INFO: Collecting {0} conf from {1} "
                    "host...").format(service, host['ip'])
            agentlocal_cpath = localpath + '/etc/' + service
            utils.createdirs(agentlocal_cpath)
            collectrfiles(host, remote_conf_path, agentlocal_cpath)

    except Exception as e:
        print _("ERROR:Failed to collect agent logs/conf file "
                "information. {0}").format(e)

    try:
        # collect runtime information
        print _("INFO: Collecting runtime information...")
        status_cmd = "/opt/ibm/powervc/bin/powervc-services status"
        status, stdout = utils.runrcmd(client, status_cmd,
                                       utils.maxwaittime)
        if stdout is not None:
            fname = localpath + '/powervcservices.info'
            with open(fname, "w") as finfo:
                finfo.write(stdout)
        else:
            print _("ERROR: The command \"{0}\" failed to"
                    " run.").format(status_cmd)
    except Exception as e:
        print _("Error:Failed to collect runtime information from {0} "
                "host. {1}").format(host['ip'], e)

    try:
        print _("INFO: Collecting SOS report from {0} ...").format(host['ip'])
        diag_file = '/var/tmp/*DIAG-{0}*'.format(host['ip'])
        cleanup_cmd = "/bin/rm -rf {0}".format(diag_file)
        cleanup_status, stdout = utils.\
            runrcmd(client, cleanup_cmd, utils.maxwaittime)
        collection_cmd = "/usr/sbin/sosreport --batch --name={0}".\
            format(diag_file)
        collection_status, stdout = utils.runrcmd(client,
                                                  collection_cmd,
                                                  utils.maxwaittime)
        if cleanup_status == 0 and collection_status == 0:
            collectrfiles(host, diag_file + ".xz", localpath)
        elif collection_status == -99:
            print _("Error: Timed out, Consider increasing the max "
                    "timeout value")
            print _("Error: Failed to collect system data from {0} \
                     host").format(host['ip'])
    except Exception as e:
        print _("Error:Failed to collect SOS report from {0} "
                "host. {1}").format(host['ip'], e)


def _collect_remote_data_from_novalink_host(client, localpath, host):
    # Copy all the service files to a temp directory on NovaLink host
    try:
        # Clean up old data before copying service files to temp directory
        temp_file = '/var/tmp/powervc-diag_data'
        cleanup_cmd = "/bin/rm -rf {0}".format(temp_file)
        if str(host['passwd']).strip() != '':
            stdin, stdout, stderr = client.exec_command(
                'sudo -S {0}'.format(cleanup_cmd))
            password = host['passwd']
            stdin.write(password + '\n')
            stdin.flush()
        else:
            stdin, stdout, stderr = client.exec_command(
                'sudo -n {0}'.format(cleanup_cmd))

        if stdout.channel.recv_exit_status() == 0:
            # Copy files only if clean-up is successful
            mkdir_cmd = '/bin/mkdir {0}'.format(temp_file)
            status, stdout = utils.runrcmd(client, mkdir_cmd,
                                           utils.maxwaittime)
            if status == 0:
                for service in ['nova', 'neutron', 'ceilometer']:
                    # Copy log files to temp location
                    remote_log_path = ("/var/log/%s/*") % service
                    tmppath_cmd = '/bin/mkdir -p {0}/var/log/{1}/'.format(
                                  temp_file, service)
                    status, stdout = utils.runrcmd(client, tmppath_cmd,
                                                   utils.maxwaittime)
                    tmppath = '/var/tmp/powervc-diag_data/var/log/'\
                              '{0}/'.format(service)
                    if str(host['passwd']).strip() != '':
                        stdin, stdout, stderr = client.exec_command(
                            'sudo -S /bin/cp {0} {1}'.format(remote_log_path,
                                                             tmppath))
                        password = host['passwd']
                        stdin.write(password + '\n')
                        stdin.flush()
                    else:
                        stdin, stdout, stderr = client.exec_command(
                            'sudo -n /bin/cp {0} {1}'.format(remote_log_path,
                                                             tmppath))

                    # Copy conf files to temp location
                    remote_conf_path = ("/etc/%s/*") % service
                    tmppath_cmd = '/bin/mkdir -p {0}/etc/{1}/'.\
                                  format(temp_file, service)
                    status, stdout = utils.runrcmd(client, tmppath_cmd,
                                                   utils.maxwaittime)
                    tmppath = '/var/tmp/powervc-diag_data/etc/{0}/'.\
                              format(service)
                    if str(host['passwd']).strip() != '':
                        stdin, stdout, stderr = client.exec_command(
                            'sudo -S /bin/cp {0} {1}'.format(remote_conf_path,
                                                             tmppath))
                        password = host['passwd']
                        stdin.write(password + '\n')
                        stdin.flush()
                    else:
                        stdin, stdout, stderr = client.exec_command(
                            'sudo -n /bin/cp {0} {1}'.format(remote_conf_path,
                                                             tmppath))

                remote_pvm_path = '/var/log/pvm/pvm_ap.dbg'
                tmp_pvm_cmd = '/bin/mkdir {0}/var/log/pvm/'.format(temp_file)
                status, stdout = utils.runrcmd(client, tmp_pvm_cmd,
                                               utils.maxwaittime)
                tmp_pvm_path = '/var/tmp/powervc-diag_data/var/log/pvm/'
                if str(host['passwd']).strip() != '':
                    stdin, stdout, stderr = client.exec_command(
                        'sudo -S /bin/cp {0} {1}'.format(remote_pvm_path,
                                                         tmp_pvm_path))
                    password = host['passwd']
                    stdin.write(password + '\n')
                    stdin.flush()
                else:
                    stdin, stdout, stderr = client.exec_command(
                        'sudo -n /bin/cp {0} {1}'.format(remote_pvm_path,
                                                         tmp_pvm_path))

                remote_syslog_path = '/var/log/syslog'
                tmp_syslog_path = '/var/tmp/powervc-diag_data/var/log'
                if str(host['passwd']).strip() != '':
                    stdin, stdout, stderr = client.exec_command(
                        'sudo -S /bin/cp {0} {1}'.format(remote_syslog_path,
                                                         tmp_syslog_path))
                    password = host['passwd']
                    stdin.write(password + '\n')
                    stdin.flush()
                else:
                    stdin, stdout, stderr = client.exec_command(
                        'sudo -n /bin/cp {0} {1}'.format(remote_syslog_path,
                                                         tmp_syslog_path))

                change_permissions_cmd = 'chown root:' + host['user'] + \
                                         ' /var/tmp/powervc-diag_data/* -R'
                if str(host['passwd']).strip() != '':
                    stdin, stdout, stderr = client.exec_command(
                        'sudo -S {0}'.format(change_permissions_cmd))
                    password = host['passwd']
                    stdin.write(password + '\n')
                    stdin.flush()
                else:
                    stdin, stdout, stderr = client.exec_command(
                        'sudo -n {0}'.format(change_permissions_cmd))
        else:
            print _("Error: Unable to clean up old files on Novalink host "
                    "{0}").format(host['ip'])

    except Exception as e:
        print _("Error:Failed to copy service files to a temp directory on "
                "{0} host. {1}").format(host['ip'], e)

    # Collect openstack service agent logs from remote NovaLink host
    try:
        for service in ['nova', 'neutron', 'ceilometer']:
            remote_log_path = ('/var/tmp/powervc-diag_data/var/'
                               'log/%s/*') % service
            agentlocalpath = localpath + '/var/log/' + service
            print _("INFO: Collecting {0} logs from {1} host...")\
                .format(service, host['ip'])
            utils.createdirs(agentlocalpath)
            collectrfiles(host, remote_log_path, agentlocalpath)

            # Collect OpenStack service conf files
            remote_conf_path = ('/var/tmp/powervc-diag_data/etc/'
                                '%s/*') % service
            print _("INFO: Collecting {0} conf from {1} "
                    "host...").format(service, host['ip'])
            agentlocal_cpath = localpath + '/etc/' + service
            utils.createdirs(agentlocal_cpath)
            collectrfiles(host, remote_conf_path, agentlocal_cpath)

    except Exception as e:
        print _("ERROR:Failed to collect agent logs/conf file "
                "information. {0}").format(e)

    try:
        remotepath = "/var/tmp/powervc-diag_data/var/log/pvm/*"
        print _("INFO: Collecting /var/log/pvm from {0} "
                "host...").format(host['ip'])
        agentlocalpath = localpath + '/var/log/pvm'
        utils.createdirs(agentlocalpath)
        collectrfiles(host, remotepath, agentlocalpath)
    except Exception as e:
        print _("ERROR:Failed to collect pvm log file from host,{0} "
                "{1}").format(host['ip'], e)

    try:
        remotepath = "/var/tmp/powervc-diag_data/var/log/syslog"
        print _("INFO: Collecting /var/log/syslog from {0} "
                "host...").format(host['ip'])
        agentlocalpath = localpath + '/var/log/syslog'
        utils.createdirs(agentlocalpath)
        collectrfiles(host, remotepath, agentlocalpath)
    except Exception as e:
        print _("ERROR:Failed to collect syslog information from host,{0} "
                "{1}").format(host['ip'], e)

    try:
        # collect powervc-services status
        print _("INFO: Collecting powervc-services status on NovaLink {0} "
                "host...").format(host['ip'])
        cmd = "sudo /opt/ibm/powervc/bin/powervc-services status"
        LOG.debug(_("Executing {0}").format(cmd))
        if str(host['passwd']).strip() != '':
            stdin, stdout, stderr = client.exec_command(
                'sudo -S {0}'.format(cmd))
            password = host['passwd']
            stdin.write(password + '\n')
            stdin.flush()
        else:
            stdin, stdout, stderr = client.exec_command(
                'sudo -n {0}'.format(cmd))

        if stdout is not None:
            fname = localpath + '/powervcservices.info'
            with open(fname, "w") as finfo:
                finfo.write(str(stdout.readlines()))
        else:
            print _("ERROR: The command \"{0}\" failed to run.").format(cmd)
    except Exception as e:
        print _("ERROR:Failed to collect runtime information from host,{0} "
                "{1}").format(host['ip'], e)

    try:
        # collect journalctl information
        print _("INFO: Collecting journalctl information...")
        journalctl_cmd = "journalctl"
        LOG.debug(_("Executing {0}").format(cmd))
        if str(host['passwd']).strip() != '':
            stdin, stdout, stderr = client.exec_command(
                'sudo -S {0}'.format(journalctl_cmd))
            password = host['passwd']
            stdin.write(password + '\n')
            stdin.flush()
        else:
            stdin, stdout, stderr = client.exec_command(
                'sudo -n {0}'.format(journalctl_cmd))

        if stdout is not None:
            fname = localpath + '/journalctl.info'
            with open(fname, "w") as finfo:
                finfo.write(str(stdout.readlines()))
        else:
            print _("ERROR: The command \"{0}\" failed to run.").format(cmd)
    except Exception as e:
        print _("ERROR:Failed to collect journalctl information from host,{0} "
                "{1}").format(host['ip'], e)


def collect_remote_host_data(args):
    # Connect to remote hosts and validate them
    # Collect HMC specifc diagmostic data from HMC
    if utils.ispowerKVM():
        if args.kvminfo:
            load_and_capture_diag(args.kvminfo, 'kvm', args.runparallel)

    if utils.isofferingHMC():
        # Collect remote diag data from novalink hosts
        if args.novalinkinfo:
            load_and_capture_diag(args.novalinkinfo,
                                  'novalink', args.runparallel)
        # Collect remote diag data from HMC
        if args.hmcinfo:
            hmcname = str(args.hmcinfo)
            hmc_host_list = []
            if hmcname is not None:
                if hmcname is not "all":
                    hmc_host_list = hmcname.split(',')
                hmc_hosts = loadHMC(hmcname, hmc_host_list)
            security.ssh_hmc(hmc_hosts)
            threads = []
            for hmc in hmc_hosts.iterkeys():
                if not args.runparallel:
                    collectHMCInfo(hmc_hosts[hmc])
                else:
                    collection_thread = Thread(target=collectHMCInfo,
                                               args=(hmc_hosts[hmc],))
                    threads.append(collection_thread)
                    collection_thread.start()
            for thread in threads:
                thread.join()


def printSummary(args, archive_file, dout, username):
    # print out summary of invocation and provide prompt to continue or abort
    print _("\n+--------------------------------------+")
    print _("| Summary of the opts you specified |")
    print _("+--------------------------------------+")
    # using this unconventional structure to avoid changing the msgid and
    # affecting all languages
    print _("Services to process            :"), ",".join(constants.vservices)
    print _("Output directory               : {0}").format(dout)
    print _("Archive filename               : {0}").format(archive_file)
    print _("Interactive mode               : {0}").format(not args.noprompt)
    print _("Maximum wait time (secs)       : {0}").format(utils.maxwaittime)
    print _("Run validators                 : {0}").format(str(args.validate))

    if utils.ispowerKVM():
        print _("Collect KVM Host Information       : {0}").format(
            encodeutils.safe_decode(str(args.kvminfo)))
        if args.kvminfo:
            print _("Collect HOST\'s Info in Parallel : {0}").format(str(
                args.runparallel))
    if utils.isofferingHMC():
        if args.novalinkinfo:
            print _("Collect Novalink Information   : {0}")\
                .format(encodeutils.safe_decode(str(args.novalinkinfo)))
            print _("Collect HOST\'s Info in Parallel: {0}").format(
                str(args.runparallel))
        if args.hmcinfo:
            print _("Collect HMC Information        : {0}")\
                .format(encodeutils.safe_decode(str(args.hmcinfo)))
            print _("Collect HMC\'s Info in Parallel : {0}").format(
                str(args.runparallel))
    print _("User                           : {0}").format(username)
    print _("\n")
    if args.runparallel:
        print _("DEPRECATED: -p/-runparallel option is deprecated")
    if args.validate:
        print _("INFO: The validators may take several minutes to run.\n")
    if utils.ispowerKVM() and args.kvminfo and not args.runparallel:
        print _('INFO: The KVM host info collection may '
                'take several minutes '
                'to run.\n')
    if utils.isofferingHMC() and args.hmcinfo and not args.runparallel:
        print _('INFO: The HMC log collection may '
                'take several minutes '
                'to run.\n')
    if utils.isofferingHMC() and args.novalinkinfo and not args.runparallel:
        print _('INFO: The Novalink host info collection may '
                'take several minutes '
                'to run.\n')
    if not args.noprompt:
        choice = "Y"
        while choice != "y" and choice != "n":
            sys.stdout.write(_("Do you want to continue? [y/n] "))
            choice = raw_input().lower()
            if choice != "y" and choice != "n":
                print _("'{0}' is not a valid response.").format(choice)
        if choice == "n":
            utils.cmdexit(_("INFO: User chose to stop the command."), 0)

    return True


def setupdirs(archive_file, dout):
    """Create archive base directories.
    """
    global dcur
    global dtmp
    global dinfo
    global ddb
    global dkvminfo
    global dhmcinfo
    global dneoinfo
    # directory names
    dcur = os.getcwd()
    dtmp = "{0}/{1}".format(dout, archive_file)
    dinfo = "{0}/info".format(dtmp)
    ddb = "{0}/db".format(dtmp)
    dkvminfo = "{0}/kvminfo".format(dtmp)
    dhmcinfo = "{0}/hmcinfo".format(dtmp)
    dneoinfo = "{0}/novalinkinfo".format(dtmp)
    # cleanup existing archive output directory
    shutil.rmtree(dtmp, ignore_errors=True)

    for dir in constants.parent_dir:
        dir_path = "{0}/{1}".format(dtmp, dir)
        # create requisite archive output directories
        if not utils.createdirs(dir_path):
            utils.cmdexit(
                _(
                    "ERROR: Unable to create the archive output directory "
                    "{0} or the user invoking the command does not "
                    "have write authority to the directory.").format(dir_path),
                1)


def main():
    # register ctrl-c signal handler
    signal.signal(signal.SIGINT, utils.sighdlr)
    security.validate_user()
    utils.initialize_logger()
    global client_hmc
    client_hmc = None
    try:
        begin()
    finally:
        if client_hmc is not None:
            client_hmc.close()


def begin():
    output_captured = utils.exit_running_proc()
    # parse the command invocation
    parser = diag_parser.parsecmd()
    args = parser.parse_args()
    archive_file, dout = utils.create_archive_file_dir(args)
    utils.validate_arguments(args)
    username = security.get_user(args)
    project = security.get_project(args)
    session = security.generate_token(username, project)
    if session:
        username = session.auth.auth_methods[0].username
    cloud_enabled = utils.checkCloudEnabled()
    if cloud_enabled:
        constants.vldservices.append('clerk')
        constants.service_map.update({'clerk': 'self-service'})
    printSummary(args, archive_file, dout, username)
    utils.delete_old_archives(utils.maxwaittime, dout, args.noprompt)

    # insert a blank line after invocation summary
    sys.stdout.write("\n")
    # setup requisite directories
    setupdirs(archive_file, dout)
    if args.validate:
        validator_id = common.start_validator(session)

    common.capture_generic_data(dinfo, utils.maxwaittime, dtmp)
    # collect requested diagnostic data for each service
    for service in (utils.vservices or constants.vservices):
        utils.serviceconf(service, dtmp)
        logs.collect_log(service, dtmp)
        db.capture_db(service, ddb, dcur)
        info.serviceinfo(service, dinfo, session)

    # Collect remote diagnostic data from KVM/novalink/HMC
    collect_remote_host_data(args)

    common.collect_error_log_summary(dtmp, utils.maxwaittime)
    if args.validate:
        common.collect_validator_output(
            session,
            validator_id,
            dinfo)
    utils.create_final_archive(
        output_captured,
        archive_file,
        dtmp,
        utils.maxwaittime,
        dcur,
        dout)

if __name__ == "__main__":
    main()
