#!/usr/bin/python
# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import ConfigParser
import grp
import pwd
from eventlet import greenthread
from sqlalchemy import engine as sqlengine
from sqlalchemy import MetaData, Table
from sqlalchemy.sql import expression

from powervc_keystone.encrypthandler import EncryptHandler


class ValidateUsersRole(object):

    def _get_db_engine(self):
        """Return a new SQLAlchemy Database Engine for the given DB"""
        parser = ConfigParser.RawConfigParser()
        parser.read('/etc/keystone/keystone.conf')
        sql_connection = parser.get('database', 'connection')
        sql_connection = EncryptHandler().decode(sql_connection)
        db_args = \
            {"pool_recycle": 3600, "echo": False, 'convert_unicode': True}
        engine = sqlengine.create_engine(sql_connection, **db_args)
        greenthread.sleep(0)
        return engine

    def isLdapEnabled(self):
        """ Return true if LDAP is enabled, else false"""
        parser = ConfigParser.RawConfigParser()
        parser.read('/etc/keystone/domains/keystone.Default.conf')
        driver = parser.get('identity', 'driver')
        if "ldap" in driver:
            return True
        else:
            return False

    def _get_table(self, name, db_connection, db_metadata):
        """Returns a list of rows for the specified table"""
        table = Table(name, db_metadata, autoload=True)
        query = expression.select([table])
        return db_connection.execute(query).fetchall()

    def _get_id_to_name_map(self, entity, db_connection, db_metadata):
        """Returns an id/name mapping for each \
            instance of the specified entity"""
        names = dict()
        rows = self._get_table(entity, db_connection, db_metadata)
        for row in rows:
            names[row['id']] = row['name']
        return names

    def _get_user_id_map(self, entity, db_connection, db_metadata):
        """Returns an id/group mapping for each \
            user/actor instance of the specified entity"""
        idmap = dict()
        rows = self._get_table(entity, db_connection, db_metadata)
        for row in rows:
            idmap[row['public_id']] = row['local_id']
        return idmap

    def _get_user_assignments(self):
        assign = list()
        db_engine = self._get_db_engine()
        db_connection = db_engine.connect()
        db_metadata = MetaData(bind=db_engine, schema='keystone')

        roles = self._get_id_to_name_map('role', db_connection,
                                         db_metadata)
        projects = self._get_id_to_name_map('project', db_connection,
                                            db_metadata)
        assignment_table = self._get_table('assignment', db_connection,
                                           db_metadata)
        id_mapping = self._get_user_id_map('id_mapping', db_connection,
                                           db_metadata)

        for row in assignment_table:
            assignment = dict()
            for col_name in row.keys():
                ref_name = col_name
                col_value = row[col_name]

                if col_name == 'target_id':
                    if row['type'].endswith('Project'):
                        assignment['project_name'] = projects[row[col_name]]
                        ref_name = 'project_id'
                elif col_name == 'actor_id':
                    col_value = str(id_mapping.get(row[col_name]))

                    if row['type'].startswith('User'):
                        ref_name = 'user_name'
                        try:
                            assignment['user_id'] = pwd.getpwnam(
                                col_value).pw_uid
                        except:
                            assignment['user_id'] = '<missing>'
                    elif row['type'].startswith('Group'):
                        ref_name = 'group_name'
                        try:
                            grp_info = grp.getgrnam(col_value)
                            assignment['group_id'] = grp_info.gr_gid
                            grp_mem = self._get_users_for_group(
                                grp_info.gr_gid)
                            grp_mem[1:1] = grp_info.gr_mem
                            assignment['group_members'] = grp_mem
                        except:
                            assignment['group_id'] = '<missing>'
                            assignment['group_members'] = list()

                assignment[ref_name] = col_value
            assignment['role_name'] = roles[row['role_id']]
            assign.append(assignment)
        db_connection.close()
        return assign

    def _get_users_for_group(self, grp_id):
        grp_users = list()
        all_user_data = pwd.getpwall()
        for user in all_user_data:
            if user.pw_gid == grp_id:
                grp_users.append(user.pw_name)
        return grp_users

    def _map_users_and_groups(self, assignments):
        """
        Create a dictionary mapping of roles to users.

        :param assignments: A dictionary of augmented assignment data.
        :returns: Three objects.  The first is a list of missing users.  The
                  second object is a list of missing groups.  The third object
                  is a dictionary where each key is the name of a project, and
                  each value is a dictionary.  For the inner dictionary, each
                  key is the name of a role with " Users" concatenated as the
                  suffix, and each value is a set of user names with that role.
        """
        # Init objects to hold return data
        missing_users = set()
        missing_groups = set()
        user_dict = {}

        # Inspect assignments dictionary to build mapping
        for a in assignments:
            # Check for missing users
            if a.get('user_id', None) == '<missing>':
                missing_users.add(a['user_name'])
            # Check for missing groups
            if a.get('group_id', None) == '<missing>':
                missing_groups.add(a['group_name'])
            # Find the project used for this assignment
            prj = a.get('project_name', None)
            if not prj:
                # This should never happen, but best to protect ourselves
                continue
            # Initialize the inner dictionary to return
            if prj not in user_dict.keys():
                user_dict[prj] = {}
            if not a.get('role_name', None):
                # This should never happen, but best to protect ourselves
                continue
            # Build the inner dictionary to return
            role_key = str(a['role_name'] + ' Users')
            if role_key not in user_dict[prj].keys():
                user_dict[prj][role_key] = set()
            if a.get('user_name', None):
                user_dict[prj][role_key].add(a['user_name'])
            elif a.get('group_name', None):
                for u in a.get('group_members', []):
                    user_dict[prj][role_key].add(u)

        # Convert sets to lists for pretty printing, remove any empty sets, and
        # remove the service project
        missing_users = list(missing_users)
        missing_groups = list(missing_groups)
        if 'service' in user_dict:
            del user_dict['service']
        for project_dict in user_dict.values():
            for role, role_set in project_dict.items():
                if role_set:
                    project_dict[role] = list(role_set)
                else:
                    del project_dict[role]

        return missing_users, missing_groups, user_dict

    def get_user_map(self):
        user_dict = dict()

        assignments = self._get_user_assignments()
        assign_dict = {'Assignments': assignments}
        is_ldap = self.isLdapEnabled()
        if is_ldap:
            return [], [], assign_dict
        else:
            missing_users, missing_groups, user_dict = \
                self._map_users_and_groups(assignments)

        user_dict.update(assign_dict)
        return missing_users, missing_groups, user_dict
