# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from powervc_oslo.config import data_mgr as mgr
from powervc_oslo.config import data_utils as utils
from powervc_oslo.config.data_utils import V130
from powervc_oslo.config.data_utils import CONSOLE

######################################################
#            RAS Data Manager Definition             #
######################################################


class RasDataManager(mgr.BaseDataManager):

    def get_component_name(self):
        """Provides the name of the Component providing the Data Manager"""
        return 'ras'

    def is_supported_environment(self):
        """Specifies whether the Component is supported in this Environment"""
        return True

    ###############################################
    #     Data Manager - Configure Interface      #
    ###############################################
    def configure_component(self):
        """Performs Configuration specific to the given Component"""
        super(RasDataManager, self).configure_component()
        # Do the Configuration to the log rotation Cron Job
        self.__configure_log_rotation()

    def upgrade_component(self, version):
        # Do the Configuration to the log rotation Cron Job
        self.__configure_log_rotation()

    ###############################################
    #   Data Manager - Subclass Implementation    #
    ###############################################
    def _get_added_conf_properties(self):
        """Provides a dictionary of configuration properties to add"""
        return dict()

    def _get_deleted_conf_properties(self):
        """Provides a dictionary of configuration properties to remove"""
        config = dict()
        # No properties to set if this isn't the KVM Host
        if utils.get_compute_node_type() != 'powerkvm':
            return config
        # For Nova, Neutron and Ceilometer, need to disable rsyslog
        for comp in ['nova', 'neutron', 'ceilometer']:
            conf_file = '/etc/%(comp)s/%(comp)s.conf' % dict(comp=comp)
            config.setdefault(conf_file, dict())
            # Remove the properties from the DEFAULT Section
            config[conf_file]['DEFAULT'] = dict(
                use_syslog=V130, syslog_log_facility=V130)
        return config

    ###############################################
    #    Internal Data Manager Helper Methods     #
    ###############################################
    def __configure_log_rotation(self):
        """Helper method configurat the log rotation cron jobs"""
        old = '\/var\/log\/cinder\/\*.log'
        new = '\/var\/log\/cinder\/\*.log \ \/var\/log\/cinder\/*\/*\/*\/*.k2r'
        if utils.is_management_system():
            # If this is PowerVM then we need to add
            # log rotation for .k2r files
            if utils.get_offering() == 'powervm':
                if ".k2r" not in \
                        open('/etc/logrotate.d/openstack-cinder').read():
                    utils.sed_replace('/etc/logrotate.d/openstack-cinder',
                                      old, new)
                else:
                    pass
            # Build a list of the log files for all the service components
            comps = [c for c in mgr.SERVICE_TYPES if (c != 'validator' and
                                                      c != 'clerk')]
            logs = [('openstack-' + c) for c in comps] + \
                ['powervc-bumblebee']
        else:
            logs = [('openstack-' + c) for c in ['nova',
                                                 'neutron',
                                                 'ceilometer']]
        # Loop through each of the log files, updating the log rotation file
        for log in logs:
            log = '/etc/logrotate.d/%s' % log
            is_size = True
            is_rotate = True
            with open(log, 'r') as fin:
                file_content = fin.read()
                if 'size 10M' not in file_content:
                    is_size = False
                if ('rotate' not in file_content) or\
                        ('post' in file_content):
                    is_rotate = False
            utils.run_command(
                "/bin/sed -i '/\(weekly\|minsize\|monthly\)/d' %s" % log)
            if not is_size:
                # Run the sed commands to update the component's log rotation
                # file
                utils.run_command("/bin/sed -i '/{/a \    size 10M\' %s" % log)
            if not is_rotate:
                # If no rotate setting is found, set a default value to 4
                utils.run_command("/bin/sed -i '/{/a \    rotate 4\' %s" % log)
