# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""An API Extension for being able to Persist/Query Console Messages"""

import webob
import ConfigParser
from time import sleep
from datetime import datetime
from eventlet import greenthread
from oslo_config import cfg
from oslo_concurrency import lockutils
from oslo_log import log as logging

from nova import context as ctxt
from nova.api.openstack import extensions
from nova.api.openstack import wsgi

from powervc_oslo.config import data_utils as oslo_utils

from powervc_nova.db import api as db_api
from powervc_nova import _

cfg.CONF.register_opts([
    cfg.IntOpt('console_messages_maximum', default=5000,
               help='The maximum number of Console Messages to keep'),
    cfg.IntOpt('console_messages_prune_interval', default=300,
               help='The interval (in seconds) at which to Prune Messages'),
    cfg.BoolOpt('console_messages_notify', default=True,
                help='Whether to Send Notifications for Console Messages')
])

LOG = logging.getLogger(__name__)
ALIAS = 'ibm-console-messages'
AUTHORIZE_V21 = extensions.os_compute_authorizer(ALIAS)
AUTHORIZE_V2 = extensions.extension_authorizer('compute', 'console_messages')
PRUNING_ENABLED = False

# Define the list of Possible Attributes to return to the User from the DB
QUERY_PARAMETERS = [
    'id', 'msg_time', 'msg_type', 'priority',
    'publisher', 'msg_start', 'msg_end']

MESSAGE_ATTRIBUTES = [
    'id', 'msg_time', 'msg_type', 'msg_user',
    'project_id', 'priority', 'publisher', 'payload']


class ConsoleMessagesController(wsgi.Controller):
    """Main Controller Class to provide the ibm-console-messages API"""

    def index(self, req):
        """Implement Rest GET Method to return a list of Console Messages"""
        messages = list()
        context = req.environ['nova.context']
        context = self._validate_authorization(context, req, 'query')
        # Call the Database API to Query the Messages from the DB
        filters, maximum = self._create_query_filter(context, req)
        dbmsgs = db_api.console_messages_find_all(context, filters, maximum)
        # Loop through the Messages only returning the external attributes
        for dbmsg in dbmsgs:
            messages.append(self._extract_message_contents(dbmsg))
        return dict(messages=messages)

    def create(self, req, body):
        """Implement Rest POST Method to persist a new Console Message"""
        messages = list()
        context = req.environ['nova.context']
        context = self._validate_authorization(context, req, 'create')
        # Call the Database API to Create the Message in the DB
        for message in body['messages']:
            value = self._extract_message_contents(message)
            msg = db_api.console_messages_create(context, value)
            messages.append(self._extract_message_contents(msg))
        # Enable the Pruning task here since we can't on startup
        self._enable_pruning_task()
        # Return all the created messages
        return dict(messages=messages)

    @wsgi.response(204)
    def update(self, req, id, body):
        """Implement Rest PUT Method to update an existing Console Message"""
        context = req.environ['nova.context']
        context = self._validate_authorization(context, req, 'update')
        # Call the Database API to Update the Message in the DB
        values = self._extract_message_contents(body['message'])
        # Call the Database API to Query the Message from the DB
        if not db_api.console_messages_find(context, dict(id=id)):
            text = _("Message %s not found.  Specify a "
                     "message id that exists and try again.") % id
            raise webob.exc.HTTPNotFound(explanation=text)
        # Call the Database API to Update the Message in the DB
        db_api.console_messages_update(context, id, values)

    @wsgi.response(204)
    def delete(self, req, id):
        """Implement Rest PUT Method to remove an existing Console Message"""
        context, authexc = req.environ['nova.context'], None
        try:
            if 'v2.1' in req.url:
                AUTHORIZE_V21(context, action='delete')
            else:
                AUTHORIZE_V2(context, action='delete')
        # If the Authorization failed, means they aren't ADMIN, so use later
        except Exception as exc:
            authexc = exc
        # If they gave us a special id of "all", then delete all for that user
        if id.lower() == 'all':
            for msg in db_api.console_messages_find_all(context):
                # If this is either an ADMIN or a Message for this User delete
                if not authexc or msg['msg_user'] is not None:
                    db_api.console_messages_delete(context, msg['id'])
            return
        # Call the Database API to Query the Message from the DB
        message = db_api.console_messages_find(
            ctxt.get_admin_context(), dict(id=id))
        # If the Message wasn't found in the database, throw an exception
        if not message:
            text = _("Message %s not found.  Specify a "
                     "message id that exists and try again.") % id
            raise webob.exc.HTTPNotFound(explanation=text)
        # If they aren't an ADMIN and not the Message User, throw exception
        if authexc and message['msg_user'] != context.user_name:
            raise authexc
        # Call the Database API to Delete the Message from the DB
        db_api.console_messages_delete(context, id)

    @staticmethod
    def _validate_authorization(context, req, action):
        """Internal Helper Method to Confirm the Requester is Authorized"""
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action=action)
        else:
            AUTHORIZE_V2(context, action=action)
        return context.elevated()

    def _create_query_filter(self, context, req):
        """Helper method to create a query filter from the Request"""
        filters = dict()
        for param in QUERY_PARAMETERS:
            if param in req.GET:
                filters[param] = req.GET[param]
                # If this is a time parameter, make sure the right format
                if param in ['msg_time', 'msg_start', 'msg_end']:
                    self._validate_time_parameter(filters, param)
        maximum = req.GET.get('maximum')
        maximum = int(maximum) if maximum is not None else None
        return filters, maximum

    def _validate_time_parameter(self, filters, param):
        """Helper method to validate the format of the time parameter"""
        try:
            try:
                val = datetime.strptime(filters[param], '%Y-%m-%dT%H:%M:%SZ')
                filters[param] = val.strftime('%Y-%m-%d-%H.%M.%S.%f')
            except ValueError:
                datetime.strptime(filters[param], '%Y-%m-%d-%H.%M.%S.%f')
        except:
            msg = _("The '%s' query parameter must be in the "
                    "YYYY-mm-ddTHH:MM:SSZ string format") % param
            raise webob.exc.HTTPBadRequest(explanation=msg)

    def _extract_message_contents(self, message):
        """Helper method to extract the applicable Message attributes"""
        new_message = dict()
        for attr in MESSAGE_ATTRIBUTES:
            if attr in message:
                new_message[attr] = message[attr]
        return new_message

    def _enable_pruning_task(self):
        """Helper method to enable the pruning task for old messages"""
        global PRUNING_ENABLED
        # Startup a periodic interval thread to Prune Old Messages
        if not PRUNING_ENABLED:
            PRUNING_ENABLED = True
            greenthread.spawn(self._pruning_task)

    def _pruning_task(self):
        """Helper method to Prune Old Messages from the Database"""
        while cfg.CONF.console_messages_maximum > 0:
            try:
                sleep(cfg.CONF.console_messages_prune_interval)
                self._prune_db_messages()
            except Exception as exc:
                LOG.warn(_("Error pruning message from the database"))
                LOG.exception(exc)

    @lockutils.synchronized('prune_messages', 'powervc-')
    def _prune_db_messages(self):
        """Helper method to prune messages from the database"""
        db_api.console_messages_prune(
            ctxt.get_admin_context(), cfg.CONF.console_messages_maximum)


class ConsoleMessagePreferencesController(wsgi.Controller):

    """Controller Class to provide the ibm-console-message-preferences API"""

    def index(self, req):
        """Implement Rest GET Method to return the current Preferences"""
        prefs = dict()
        parser = ConfigParser.RawConfigParser()
        context = req.environ['nova.context']
        context = ConsoleMessagesController.\
            _validate_authorization(context, req, 'query')
        # Default the values to what is set in the ConfigOpts loaded
        prefs['console_messages_notify'] = cfg.CONF.console_messages_notify
        prefs['console_messages_maximum'] = cfg.CONF.console_messages_maximum
        prefs['console_messages_prune_interval'] =\
            cfg.CONF.console_messages_prune_interval
        # Get and return the current Preferences from the Configuration File
        parser.read('/etc/nova/nova.conf')
        for key in ['console_messages_prune_interval',
                    'console_messages_maximum', 'console_messages_notify']:
            # Only set the value if it is actually in the configuration file
            if parser.has_option('DEFAULT', key):
                prefs[key] = parser.getboolean('DEFAULT', key) if isinstance(
                    prefs[key], bool) else parser.getint('DEFAULT', key)
        return dict(preferences=prefs)

    @wsgi.response(204)
    def create(self, req, body):
        """Implement Rest POST Method to update the existing Preferences"""
        updated_props = dict()
        context = req.environ['nova.context']
        context = ConsoleMessagesController.\
            _validate_authorization(context, req, 'create')
        prefs = body.get('preferences', dict())
        # If they specified the Maximum Messages, update that property
        if prefs.get('console_messages_maximum') is not None:
            new_val = int(prefs['console_messages_maximum'])
            if new_val < 0:
                msg = "value of console_messages_maximum "\
                    "should not be negative"
                raise webob.exc.HTTPBadRequest(explanation=msg)
            else:
                cfg.CONF.console_messages_maximum = new_val
            updated_props['console_messages_maximum'] = str(new_val)
        # If they specified the Pruning Interval, update that property
        if prefs.get('console_messages_prune_interval') is not None:
            new_val = int(prefs['console_messages_prune_interval'])
            if new_val < 0:
                msg = "value of console_messages_prune_interval "\
                    "should not be negative"
                raise webob.exc.HTTPBadRequest(explanation=msg)
            else:
                cfg.CONF.console_messages_prune_interval = new_val
            updated_props['console_messages_prune_interval'] = str(new_val)
        # If they specified the Maximum Messages, update that property
        if prefs.get('console_messages_notify') is not None:
            new_val = bool(prefs['console_messages_notify'])
            cfg.CONF.console_messages_notify = new_val
            updated_props['console_messages_notify'] = str(new_val)
        # Now for each changed property, update the nova.conf file
        for key, val in updated_props.iteritems():
            oslo_utils.set_openstack_config_value('/etc/nova/nova.conf', key,
                                                  val, 'DEFAULT')


class ConsoleMessages(extensions.ExtensionDescriptor):

    """Capability to Persist and Query Console Messages."""
    name = "IBM Querying and Storing of Console Messages"
    alias = "ibm-console-messages"
    namespace = "http://docs.openstack.org/compute/ext/" + \
                "ibm_console_messages/api/v1.1"
    updated = "2014-07-02T21:00:00-06:00"

    def get_resources(self):
        """Provide a Resource Extension to the ibm-console-messages Resource"""
        resource1 = extensions.ResourceExtension(
            'ibm-console-messages', controller=ConsoleMessagesController())
        resource2 = extensions.ResourceExtension(
            'ibm-console-message-preferences',
            controller=ConsoleMessagePreferencesController())
        return [resource1, resource2]


class ConsoleMessagesV21(extensions.V21APIExtensionBase):
    """Capability to Persist and Query Console Messages."""
    name = "IBM Querying and Storing of Console Messages"
    alias = ALIAS
    version = 1

    def get_resources(self):
        """Provide a Resource Extension to the ibm-console-messages Resource"""
        resource1 = extensions.ResourceExtension(
            'ibm-console-messages', controller=ConsoleMessagesController())
        resource2 = extensions.ResourceExtension(
            'ibm-console-message-preferences',
            controller=ConsoleMessagePreferencesController())
        return [resource1, resource2]

    def get_controller_extensions(self):
        return []
