# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""An API Extension for being able to Persist/Query DRO Events"""

import webob
from datetime import datetime
from oslo_log import log as logging

from nova.api.openstack import extensions
from nova.api.openstack import wsgi

from powervc_nova.db import api as db_api
from powervc_nova import _

LOG = logging.getLogger(__name__)
ALIAS = 'dro-events'
AUTHORIZE_V21 = extensions.os_compute_authorizer(ALIAS)
AUTHORIZE_V2 = extensions.extension_authorizer('compute', 'dro_events')

# Define the list of Possible Attributes to return to the User from the DB
QUERY_PARAMETERS = ['id', 'event_time', 'event_type', 'exclude_event_type',
                    'hypervisor_hostname', 'aggregate_id',
                    'event_start_tstamp', 'event_end_tstamp']

EVENT_ATTRIBUTES = ['id', 'event_time', 'event_type',
                    'hypervisor_hostname', 'aggregate_id', 'payload']


class DroEventsController(wsgi.Controller):
    """Main Controller Class to provide the dro-events API"""

    @staticmethod
    def _validate_authorization(context, req, action):
        """Internal Helper Method to Confirm the Requester is Authorized"""
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action=action)
        else:
            AUTHORIZE_V2(context, action=action)
        return context.elevated()

    def _create_query_filter(self, context, req):
        """Helper method to create a query filter from the Request"""
        filters = dict()
        for param in QUERY_PARAMETERS:
            if param in req.GET:
                if param == 'exclude_event_type':
                    # Support a comma-delimited set of event types to exclude
                    event_types = req.GET[param].split(',')
                    filters[param] = [event.strip() for event in event_types]
                else:
                    filters[param] = req.GET[param]
                # If this is a time parameter, make sure the right format
                if param in ['event_time', 'event_start_tstamp',
                             'event_end_tstamp']:
                    self._validate_time_parameter(filters, param)
        # Validate that end time is greater than start time.
        if 'event_start_tstamp' in req.GET and 'event_end_tstamp' in req.GET:
            self._validate_query_time_limits(
                req.GET['event_start_tstamp'], req.GET['event_end_tstamp'])
        max_events = req.GET.get('max_events')
        # The max_events is not None check is to ensure that empty string
        # value also goes down the _is_valid_int() validation path.
        if max_events is not None and not self._is_valid_int(max_events):
            msg = (_("The '%(value)s' query parameter must be "
                     "a positive number less than or equal to %(max)s.")
                   % dict(value=max_events,
                          max=db_api.MAX_QUERY_RESULTS))
            raise webob.exc.HTTPBadRequest(explanation=msg)
        max_events = int(max_events) if max_events else None
        return filters, max_events

    def _is_valid_int(self, str_val):
        '''
        This method validates that the str_val is positive integer.
        '''
        try:
            # If value is None or empty, return false.
            if not str_val:
                return False
            val = int(str_val)
            if not 1 <= val <= db_api.MAX_QUERY_RESULTS:
                raise ValueError()
            return True
        except ValueError:
            return False

    def _validate_time_parameter(self, filters, param):
        """Helper method to validate the format of the time parameter"""
        try:
            try:
                val = datetime.strptime(filters[param], '%Y-%m-%dT%H:%M:%SZ')
                filters[param] = val.strftime('%Y-%m-%d-%H.%M.%S.%f')
            except ValueError:
                datetime.strptime(filters[param], '%Y-%m-%d-%H.%M.%S.%f')
        except:
            msg = _("The '%s' query parameter must be in the "
                    "YYYY-mm-ddTHH:MM:SSZ string format") % param
            raise webob.exc.HTTPBadRequest(explanation=msg)

    def _validate_query_time_limits(self, start_time, end_time):
        start_dt_time = datetime.strptime(start_time, '%Y-%m-%dT%H:%M:%SZ')
        end_dt_time = datetime.strptime(end_time, '%Y-%m-%dT%H:%M:%SZ')
        diff = end_dt_time - start_dt_time
        if diff.total_seconds() < 0:
            msg = _("The event_end_tstamp query parameter must be greater "
                    "than or equal to the event_start_tstamp query parameter")
            raise webob.exc.HTTPBadRequest(explanation=msg)

    def _extract_event_contents(self, event):
        """Helper method to extract the applicable Event attributes"""
        new_event = dict()
        for attr in EVENT_ATTRIBUTES:
            if attr in event:
                new_event[attr] = event[attr]
        return new_event

    def index(self, req):
        """Implement Rest GET Method to return a list of DRO Events"""
        events = list()
        context = req.environ['nova.context']
        context = self._validate_authorization(context, req, 'index')
        # Call the Database API to Query the Events from the DB
        filters, maximum = self._create_query_filter(context, req)
        db_events = db_api.dro_events_find_all(context, filters, maximum)
        # Loop through the Events only returning the external attributes
        for db_event in db_events:
            events.append(self._extract_event_contents(db_event))
        return dict(events=events)


class DroEvents(extensions.ExtensionDescriptor):
    """Capability to Query DRO Events."""
    name = "IBM DRO Events"
    alias = "dro-events"
    namespace = "http://docs.openstack.org/compute/ext/" + \
                "dro_events/api/v1.1"
    updated = "2015-07-10T00:00:00-00:00"

    def get_resources(self):
        """Provide a Resource Extension to the dro-events Resource"""
        resource = extensions.ResourceExtension(
            'dro-events', controller=DroEventsController())
        return [resource]


class DroEventsV21(extensions.V21APIExtensionBase):
    """Capability to Query DRO Events."""
    name = "IBM DRO Events"
    alias = ALIAS
    version = 1

    def get_resources(self):
        """Provide a Resource Extension to the dro-events Resource"""
        resource = extensions.ResourceExtension(
            'dro-events', controller=DroEventsController())
        return [resource]

    def get_controller_extensions(self):
        return []
