# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

'''
Extend server create and resize action with resource allocation.
'''
import hashlib

from oslo_log import log as logging
from webob import exc

from nova.api.openstack import extensions
from nova.api.openstack import wsgi
from nova import compute
from nova import exception
from nova.i18n import _, _LE
from nova import objects


LOG = logging.getLogger(__name__)
ALIAS = 'flavor-dynamic'
AUTHORIZE = extensions.os_compute_authorizer(ALIAS)


class Flavor_dynamic(extensions.V21APIExtensionBase):
    """Support to create a dynamic flavor."""

    name = "FlavorDynamic"
    alias = "ibm-flavor-dynamic"
    version = 1

    def get_controller_extensions(self):
        controller = FlavorDynamicController()
        extension = extensions.ControllerExtension(self, 'servers', controller)
        return [extension]

    def get_resources(self):
        return []


class FlavorDynamicController(wsgi.Controller):
    """Dynamic flavor API controller."""

    def __init__(self, **kwargs):
        super(FlavorDynamicController, self).__init__(**kwargs)
        self.compute_api = compute.API()

    def _get_server(self, context, req, instance_uuid):
        """Utility function for looking up an instance by uuid."""
        try:
            instance = self.compute_api.get(context, instance_uuid)
        except exception.NotFound:
            msg = _("Instance could not be found")
            raise exc.HTTPNotFound(explanation=msg)
        req.cache_db_instance(instance)
        return instance

    def _generate_unique_name(self, flavor_dict, project_id):
        memory_mb = flavor_dict.get('ram')
        vcpus = flavor_dict.get('vcpus')
        root_gb = flavor_dict.get('disk')
        ephemeral_gb = flavor_dict.get('OS-FLV-EXT-DATA:ephemeral')
        # Add project_id to the text_name so that a unique flavor
        # name is generated for each flavor.
        text_name = ("auto-flavor-%(vcpus)s-%(memory_mb)s-"
                     "%(root_gb)s-%(ephemeral_gb)s-%(project_id)s" %
                     {'vcpus': vcpus, 'memory_mb': memory_mb,
                      'root_gb': root_gb, 'ephemeral_gb': ephemeral_gb,
                      'project_id': project_id})

        extra_specs_dict = flavor_dict.get('extra_specs')

        if extra_specs_dict:
            sorted_keys = sorted(extra_specs_dict.keys())
            for key in sorted_keys:
                text_name = '-'.join([text_name, key,
                                      str(extra_specs_dict[key])])

        # Verifed MD5 issue from Bandit scan
        # MD5 is creating a unique name for flavor
        return hashlib.md5(text_name).hexdigest()  # nosec

    def _create(self, context, req, flavor_dict):
        """Create and cache the flavor in the request if it does not already
        exist.
        """
        memory_mb = flavor_dict.get('ram')
        vcpus = flavor_dict.get('vcpus')
        root_gb = flavor_dict.get('disk')
        ephemeral_gb = flavor_dict.get('OS-FLV-EXT-DATA:ephemeral', 0)
        admin_context = (context.is_admin and
                         context or context.elevated())

        # query the flavor
        flavor_name = self._generate_unique_name(flavor_dict,
                                                 admin_context.project_id)
        flavor = None
        try:
            flavor = objects.Flavor.get_by_name(context, flavor_name)
        except exception.FlavorNotFoundByName:
            LOG.debug("Flavor '%s' not found. Creating...", flavor_name)

        # create flavor if no matched flavor
        if not flavor:
            try:
                flavor = objects.Flavor(admin_context, name=flavor_name,
                                        memory_mb=memory_mb, vcpus=vcpus,
                                        root_gb=root_gb,
                                        ephemeral_gb=ephemeral_gb,
                                        flavorid=flavor_name, swap=0,
                                        is_public=False,
                                        projects=[admin_context.project_id],
                                        extra_specs=
                                        flavor_dict.get('extra_specs', {}))
                flavor.create()
                req.cache_db_flavor(flavor)
            except (exception.FlavorExists,
                    exception.FlavorIdExists):
                try:
                    flavor = objects.Flavor.get_by_name(context, flavor_name)
                except exception.FlavorNotFoundByName:
                    raise exception.FlavorCreateFailed()
            except exception.FlavorCreateFailed as e:
                LOG.exception(_LE('DB error: %s'), e)
                msg = _('An unknown db error has occurred. ')
                raise exc.HTTPInternalServerError(explanation=msg)
        return flavor

    @wsgi.extends
    def create(self, req, body):
        """Extend Server create action with resource allocation.
        """
        if not self.is_valid_body(body, 'server'):
            raise exc.HTTPUnprocessableEntity()
        context = req.environ['nova.context']

        server_dict = body['server']
        flavor_dict = server_dict.get('flavor')

        if flavor_dict:
            # only authorize if they have the flavor_dict otherwise the policy
            # is always applied whether it is present or not
            AUTHORIZE(context)

            # verify input parameters
            if server_dict.get('flavorRef'):
                msg = _("only one of flavorRef or flavor can be specified.")
                raise exc.HTTPBadRequest(explanation=msg)
            for opt in ['ram', 'vcpus', 'disk']:
                try:
                    val = int(flavor_dict.get(opt))
                    if opt == 'disk':
                        assert val >= 0
                    else:
                        assert val > 0
                except (ValueError, TypeError, AssertionError):
                    if opt == 'disk':
                        msg = _("%s argument must be an integer") % opt
                    else:
                        msg = _("%s argument must be a positive integer") % opt
                    raise exc.HTTPBadRequest(explanation=msg)

            flavor = self._create(context, req, flavor_dict)

            # update flavorRef parameters
            server_dict['flavorRef'] = flavor.flavorid
            del server_dict['flavor']
        elif not server_dict.get('flavorRef'):
            msg = _("Missing flavorRef or flavor attribute.")
            raise exc.HTTPBadRequest(explanation=msg)
        yield

    @wsgi.extends(action='resize')
    def _action_resize(self, req, id, body):
        """Extend Server resize action with resource allocation.
        """

        if not self.is_valid_body(body, 'resize'):
            raise exc.HTTPUnprocessableEntity()
        context = req.environ['nova.context']

        resize_server_dict = body['resize']
        flavor_dict = resize_server_dict.get('flavor')

        if flavor_dict:
            # only authorize if they have the flavor_dict otherwise the policy
            # is always applied whether it is present or not
            AUTHORIZE(context)

            # verify input parameters
            if resize_server_dict.get('flavorRef'):
                msg = _("only one of flavorRef or flavor can be specified.")
                raise exc.HTTPBadRequest(explanation=msg)
            memory_mb = flavor_dict.get('ram')
            vcpus = flavor_dict.get('vcpus')
            root_gb = flavor_dict.get('disk')

            if memory_mb is None:
                msg = _("The ram parameter in flavor{} must be specified")
                raise exc.HTTPBadRequest(explanation=msg)
            if vcpus is None:
                msg = _("The vcpus parameter in flavor{} must be specified")
                raise exc.HTTPBadRequest(explanation=msg)
            if root_gb is None:
                msg = _("The disk parameter in flavor{} must be specified")
                raise exc.HTTPBadRequest(explanation=msg)

            for opt in ['ram', 'vcpus', 'disk']:
                try:
                    val = int(flavor_dict.get(opt))
                    if opt == 'disk':
                        assert val >= 0
                    else:
                        assert val > 0
                except (ValueError, TypeError, AssertionError):
                    if opt == 'disk':
                        msg = _("%s argument must be an integer") % opt
                    else:
                        msg = _("%s argument must be a positive integer") % opt
                    raise exc.HTTPBadRequest(explanation=msg)

            server = self._get_server(context, req, id)
            if not server:
                msg = _("Can not get target server by server id %s") % id
                raise exc.HTTPBadRequest(explanation=msg)
            instance_type = server.get('instance_type', None)

            if instance_type:
                currentVcpus = int(instance_type['vcpus'])
                currentRam = int(instance_type['memory_mb'])
                currentDisk = int(instance_type['root_gb'])
                if int(vcpus) == currentVcpus:
                    if int(memory_mb) == currentRam:
                        if int(root_gb) == currentDisk:
                            msg = _("ram, vcpus, or disk allocation in %s has"
                                    " not changed.") % instance_type['name']
                            raise exc.HTTPBadRequest(explanation=msg)

            flavor = self._create(context, req, flavor_dict)

            # update flavorRef parameters
            resize_server_dict['flavorRef'] = flavor.flavorid
            del resize_server_dict['flavor']
        elif not resize_server_dict.get('flavorRef'):
            msg = _("Resize requests require flavorRef or flavor "
                    "attribute.")
            raise exc.HTTPBadRequest(explanation=msg)
        yield
