#
#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""An API Extension for host level evacuation"""

import six
import webob
import webob.exc as exc

from nova.api.openstack import extensions
from nova.api.openstack import wsgi
from oslo_log import log as logging
from nova import compute
from nova import exception

from powervc_nova.common import host_evacuate_utils as evac_utils
from powervc_nova.compute import exception as pvc_exc
from powervc_nova.virt.ibmpowervm.common import constants as pvmc
from powervc_nova import _

LOG = logging.getLogger(__name__)
ALIAS = 'ibm-evacuate'
AUTHORIZE_V21 = extensions.os_compute_authorizer(ALIAS)
AUTHORIZE_V2 = extensions.extension_authorizer('compute', 'ibm_evacuate')


class HostEvacuateControllerExtension(wsgi.Controller):
    """
    Controller Class to extend the os-hosts API for evacuating VMs
    """

    def __init__(self, *args, **kwargs):
        super(HostEvacuateControllerExtension, self).__init__(*args, **kwargs)
        self.compute_api = compute.API()
        self.host_api = compute.HostAPI()

    @extensions.expected_errors((400, 404, 409, 412))
    @wsgi.action("create")
    def ibm_evacuate(self, req, host_name_id, body):
        """
        Add Rest POST Method to signal evacuation of VMs
        for a given Host
        """
        context = req.environ['nova.context']
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context)
        else:
            AUTHORIZE_V2(context)
        # Parse out the key attributes from the request body
        target_host = body.get("host") if body else None

        try:
            self.host_api.service_get_by_compute_host(context, host_name_id)
        except exception.NotFound:
            msg = _("Compute host %s is not found.") % host_name_id
            raise exc.HTTPNotFound(explanation=msg)

        if host_name_id == target_host:
            msg = _("The target host can't be the same one.")
            raise exc.HTTPBadRequest(explanation=msg)

        if target_host is not None:
            try:
                self.host_api.service_get_by_compute_host(context, target_host)
            except exception.NotFound:
                msg = _("Target compute host %s is not found.") % target_host
                raise exc.HTTPNotFound(explanation=msg)

        try:
            evac_utils.set_host_dead(context, host_name_id,
                                     dead=True, target_node_name=target_host)
        except (exception.ConstraintNotMet, pvc_exc.IBMPowerVCNotAllowed) as e:
            msg = six.text_type(e)
            raise exc.HTTPPreconditionFailed(explanation=msg)

        return webob.Response(status_int=202)


class StopHostEvacuateControllerExtension(wsgi.Controller):
    """
    Controller Class to extend the os-hosts API for stopping evacuation of VMs
    """

    def __init__(self, *args, **kwargs):
        super(StopHostEvacuateControllerExtension,
              self).__init__(*args, **kwargs)
        self.compute_api = compute.API()
        self.host_api = compute.HostAPI()

    @extensions.expected_errors((400, 404))
    @wsgi.action("create")
    def ibm_stop_evacuate(self, req, host_name_id, body):
        """
        Add Rest POST Method to signal stopping evacuation of VMs
        for a given Host
        """
        context = req.environ['nova.context']
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context)
        else:
            AUTHORIZE_V2(context)

        if host_name_id is not None:
            try:
                self.host_api.service_get_by_compute_host(context,
                                                          host_name_id)
            except exception.NotFound:
                msg = _("Compute host %s is not found.") % host_name_id
                raise exc.HTTPNotFound(explanation=msg)
            evac_utils.set_host_dead(context, host_name_id, dead=False)
        return webob.Response(status_int=202)


class HostEvacuate(extensions.ExtensionDescriptor):
    """
    Extension of evacuation support for a Compute Host.
    """
    name = "Compute Host Evacuation"
    alias = "ibm-evacuate"
    namespace = "http://docs.openstack.org/compute/ext/" + \
                "ibm_evacuate/api/v1.2"
    updated = "2013-07-21T21:00:00-06:00"

    def get_resources(self):
        """
        Provide a Sub-Resource Extension to the os-hosts Resource
        """
        resource1 = extensions.ResourceExtension(
            'ibm-evacuate', controller=HostEvacuateControllerExtension(),
            parent=dict(collection_name='os-hosts', member_name='host_name'))
        resource2 = extensions.ResourceExtension(
            'ibm-stop-evacuate',
            controller=StopHostEvacuateControllerExtension(),
            parent=dict(collection_name='os-hosts', member_name='host_name'))
        return [resource1, resource2]


class HostEvacuateV21(extensions.V21APIExtensionBase):
    """
    Extension of evacuation support for a Compute Host.
    """
    name = "Compute Host Evacuation"
    alias = ALIAS
    version = 1

    def get_resources(self):
        """
        Provide a Sub-Resource Extension to the os-hosts Resource
        """
        resource1 = extensions.ResourceExtension(
            'ibm-evacuate', controller=HostEvacuateControllerExtension(),
            parent=dict(collection_name='os-hosts', member_name='host_name'))
        resource2 = extensions.ResourceExtension(
            'ibm-stop-evacuate',
            controller=StopHostEvacuateControllerExtension(),
            parent=dict(collection_name='os-hosts', member_name='host_name'))
        return [resource1, resource2]

    def get_controller_extensions(self):
        return []
