#
#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import httplib
import json
import webob
from webob import Response
from nova import exception as novaexception
from nova import network as novanetwork
from nova import objects
from nova.api.openstack import extensions
from nova.api.openstack import wsgi
from oslo_log import log as logging
from oslo_config import cfg
from powervc_nova import network as powervc_nova_network
from powervc_nova.virt.ibmpowervm.vif.common import ras

ALIAS = 'host-network-placement'
AUTHORIZE_V21 = extensions.os_compute_authorizer(ALIAS)
AUTHORIZE_V2 = extensions.extension_authorizer('compute',
                                               'host_network_placement')

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class HostNetworkPlacementController(wsgi.Controller):
    """
    This Controller handles GET requests to query the database and return JSON
    data to map hosts to networks for use by deploy/placement.
    """

    def __init__(self):
        self.placement_module = powervc_nova_network.placement_module()

    @wsgi.extends
    def index(self, req):
        """
        This method handles GET requests to fetch placement data.
        :param req: HTTP Request
        :returns: HTTP Response with host network placement data.
        """
        context = self._validate_authorization(req.environ['nova.context'],
                                               req, action='index')

        # Parse the host aggregate id from the URI, if present
        agg = None
        if 'host_aggregate' in req.GET:
            agg = req.GET['host_aggregate']

        # Parse the network_id from the URI, if present
        network_id = None
        if 'network_id' in req.GET:
            network_id = req.GET['network_id']

        # Parse the list_only param from the URI, if present
        list_only = None
        if 'return_only_id_list' in req.GET:
            list_only = req.GET['return_only_id_list'].lower()

        # If a network was provided, validate that the network exists
        if network_id or network_id == '':
            try:
                novanetwork.API().get(context, network_id)
            except:
                return Response(request=req,
                                status=httplib.NOT_FOUND,
                                content_type='application/json',
                                body='')

        if list_only == 'true' and not network_id and not agg:
            msg = ras.vif_get_msg('error', 'PLACEMENT_PARAM_MISSING')
            return Response(request=req,
                            status=httplib.BAD_REQUEST,
                            content_type='application/json',
                            body=json.dumps({'message': msg}))

        # Get the placement dictionary
        placement_dict = self.placement_module.get_placement(context,
                                                             None,
                                                             network_id,
                                                             agg,
                                                             list_only)

        # Return the results
        return webob.Response(request=req,
                              status=httplib.OK,
                              content_type='application/json',
                              body=json.dumps(placement_dict))

    @wsgi.extends
    def show(self, req, id):
        """
        This method handles GET requests to fetch placement data for a specific
        host.
        :param req: HTTP Request
        :param id: host
        :returns: HTTP Response with placement data.
        """
        context = self._validate_authorization(req.environ['nova.context'],
                                               req, action='show')

        # A host aggregate is not allowed, verify it was not provided
        if 'host_aggregate' in req.GET:
            # Host and host aggregate cannot both be specified together
            return Response(request=req,
                            status=httplib.BAD_REQUEST,
                            content_type='application/json',
                            body='')

        # Validate the specified host exists
        hosts = self._find_all_host_names(context)
        if id not in hosts:
            raise novaexception.ComputeHostNotFound(host=id)

        # Parse the network_id from the URI, if present
        network_id = None
        if 'network_id' in req.GET:
            network_id = req.GET['network_id']

        # If a network was provided, throw an error
        if network_id:
            msg = ras.vif_get_msg('error', 'PLACEMENT_PARAM_ERROR_BOTH')
            return Response(request=req,
                            status=httplib.BAD_REQUEST,
                            content_type='application/json',
                            body=json.dumps({'message': msg}))

        # Parse the list_only param from the URI, if present
        list_only = None
        if 'return_only_id_list' in req.GET:
            list_only = req.GET['return_only_id_list'].lower()

        # Get the placement dictionary
        placement_dict = self.placement_module.get_placement(context,
                                                             id,
                                                             None,
                                                             None,
                                                             list_only)

        # Return the results
        return webob.Response(request=req,
                              status=httplib.OK,
                              content_type='application/json',
                              body=json.dumps(placement_dict))

    @staticmethod
    def _validate_authorization(context, req, action):
        """
        This method authorizes the request against the service's policy.
        :param context: API request context.
        :param authorize_method: Which method from this class to authorize.
        :returns: An authorized request context.
        """
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action=action)
        else:
            AUTHORIZE_V2(context, action=action)
        return context.elevated()

    def _find_all_host_names(self, context):
        """
        Finds a list of compute node names.
        :param context: DB context object.
        :returns: A list of compute node names.
        """
        context = context.elevated()
        return [node.host for node in objects.ComputeNodeList.get_all(context)]


class Host_network_placement(extensions.ExtensionDescriptor):
    """
    Capability to list Network Placement on all hosts.
    """
    name = "Compute Host Network Placement"
    alias = "host-network-placement"
    namespace = "http://docs.openstack.org/compute/ext/ibm_network_placement\
                 /api/v1.1"
    updated = "2013-12-09T11:00:00-06:00"

    def get_resources(self):
        """
        Provide a Resource Extension for host-network-mapping
        """
        return [
            extensions.ResourceExtension(
                'host-network-placement',
                controller=HostNetworkPlacementController())]


class Host_network_placementV21(extensions.V21APIExtensionBase):
    """
    Capability to list Network Placement on all hosts.
    """
    name = "Compute Host Network Placement"
    alias = ALIAS
    version = 1

    def get_resources(self):
        """
        Provide a Resource Extension for host-network-mapping
        """
        return [
            extensions.ResourceExtension(
                'host-network-placement',
                controller=HostNetworkPlacementController())]

    def get_controller_extensions(self):
        return []
