#
#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import httplib
import json
import webob
from nova.api.openstack import extensions
from nova.api.openstack import wsgi
from oslo_log import log as logging
from powervc_nova.network.common import exception as net_excp
from powervc_nova.network.powerkvm import host_ovs
from oslo_config import cfg

ALIAS = 'host-ovs'
AUTHORIZE_V21 = extensions.os_compute_authorizer(ALIAS)
AUTHORIZE_V2 = extensions.extension_authorizer(
    'compute', 'host_ovs')

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class HostOvsController(wsgi.Controller):
    """
    This Controller handles GET, POST, PUT and DELETE requests for OpenVSwitch
    config on hosts.
    """

    def __init__(self):
        self.ovs_task = host_ovs.HostOvsTask()

    @wsgi.action("create")
    def create(self, req, body):
        """
        This method handles POST requests to create OpenVSwitch configs.
        :param req: The HTTP request object.
        :param body: The HTTP request body.
        :returns: An HTTP response object.
        """
        context = self._validate_authorization(
            req.environ['nova.context'], req, action='create')

        # Check for the force parameter on the URL
        force = False
        if 'force' in req.GET and req.GET['force'] == 'true':
            force = True

        # Parameter for testing to force a rollback
        rollback = False
        if 'rollback' in req.GET and req.GET['rollback'] == 'true':
            rollback = True

        # Run the task to generate the HTTP response
        host_ovs_dict = self.ovs_task.create_host_ovs(context, req, body,
                                                      force, rollback)[0]
        http_code = self._post_put_delete_http_code(host_ovs_dict)
        return webob.Response(request=req,
                              status=http_code,
                              content_type='application/json',
                              body=json.dumps(host_ovs_dict))

    @wsgi.extends
    def index(self, req):
        """
        This method handles GET requests to query OpenVSwitch config.
        :param req: The HTTP request object.
        :returns: An HTTP response object.
        """
        context = self._validate_authorization(
            req.environ['nova.context'], req, action='index')

        # Run the task and generate the HTTP response
        host_ovs_dict = self.ovs_task.get_host_ovs(context, req)
        http_code = self._get_http_code(host_ovs_dict)
        return webob.Response(request=req,
                              status=http_code,
                              content_type='application/json',
                              body=json.dumps(host_ovs_dict))

    @wsgi.extends
    def show(self, req, id):
        """
        This method handles GET requests to query OpenVSwitch config for a
        specific host.
        :param req: The HTTP request object.
        :param id: The host to query from.
        :returns: An HTTP response object.
        """
        context = self._validate_authorization(req.environ['nova.context'],
                                               req, action='show')

        # Run the task and generate the HTTP response
        host_ovs_dict = self.ovs_task.get_host_ovs(context, req, id)
        http_code = self._get_http_code(host_ovs_dict)
        return webob.Response(request=req,
                              status=http_code,
                              content_type='application/json',
                              body=json.dumps(host_ovs_dict))

    @wsgi.action("update")
    def update(self, req, id, body):
        """
        This method handles PUT requests to update OpenVSwitch config.
        :param req: The HTTP request object.
        :param id: The host to update.
        :param body: The HTTP request body contents.
        :returns: An HTTP response object.
        """
        context = self._validate_authorization(
            req.environ['nova.context'], req, action='update')

        # Check for the force parameter on the URL
        force = False
        if 'force' in req.GET and req.GET['force'] == 'true':
            force = True

        # Parameter for testing to force a rollback
        rollback = False
        if 'rollback' in req.GET and req.GET['rollback'] == 'true':
            rollback = True

        # Run the task and generate the HTTP response
        host_ovs_dct = self.ovs_task.update_host_ovs(context, req, id, body,
                                                     force, rollback)[0]
        http_code = self._post_put_delete_http_code(host_ovs_dct)
        return webob.Response(request=req,
                              status=http_code,
                              content_type='application/json',
                              body=json.dumps(host_ovs_dct))

    @wsgi.action("delete")
    def delete(self, req, id):
        """
        This method handles DELETE requests to remove OpenVSwitch config.
        :param req: The HTTP request object.
        :param id: The host to delete the vswitch from.
        :returns: An HTTP response object.
        """
        context = self._validate_authorization(
            req.environ['nova.context'], req, action='delete')

        # Check for the force parameter on the URL
        force = False
        if 'force' in req.GET and req.GET['force'] == 'true':
            force = True

        # Parameter for testing to force a rollback
        rollback = False
        if 'rollback' in req.GET and req.GET['rollback'] == 'true':
            rollback = True

        # Parse the vswitch_name from the request, if present
        vswitch_name = None
        if 'vswitch_name' in req.GET:
            vswitch_name = req.GET['vswitch_name']
        if not vswitch_name or vswitch_name == '':
            raise net_excp.IBMPowerKVMVswitchNotFound(ovs_name=vswitch_name)

        # Run the task to generate the HTTP response
        host_ovs_dct = self.ovs_task.delete_host_ovs(context, req, id,
                                                     vswitch_name, force,
                                                     rollback)[0]
        http_code = self._post_put_delete_http_code(host_ovs_dct)
        return webob.Response(request=req,
                              status=http_code,
                              content_type='application/json',
                              body=json.dumps(host_ovs_dct))

    @staticmethod
    def _validate_authorization(context, req, action):
        """
        This method authorizes the request.
        :param context: The HTTP request context.
        :returns: The authorized HTTP request context.
        """
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action=action)
        else:
            AUTHORIZE_V2(context, action=action)
        return context.elevated()

    def _get_http_code(self, response_dict):
        """
        Inspect the response dictionary to determine if all hosts had an error.
        If all hosts had an error, return 404.  If at least one host was
        successful, return 200.

        :param response_dict: A JSON response body.
        :returns: An httplib return code.
        """
        if not response_dict or 'host-ovs' not in response_dict:
            # Some sort of test or other problem
            return httplib.OK

        successHostFound = False
        for host_dict in response_dict['host-ovs']:
            if('errors' not in host_dict or
               ('errors' in host_dict and len(host_dict['errors']) == 0)):
                successHostFound = True
                break
        if len(response_dict['host-ovs']) == 0:
            successHostFound = True

        if not successHostFound:
            return httplib.NOT_FOUND
        return httplib.OK

    def _post_put_delete_http_code(self, response_dict):
        """
        Inspect the response dictionary to determine the HTTP response code.

        :param response_dict: A JSON response body.
        :returns: An httplib return code.
        """
        if not response_dict:
            # Some sort of test or other problem
            return httplib.OK

        # If any errors occurred, send back a 400 BAD REQUEST
        if 'errors' in response_dict and len(response_dict['errors']) > 0:
            return httplib.BAD_REQUEST

        # If no errors occurred, but warnings did, send back a 409 CONFLICT
        if 'warnings' in response_dict and len(response_dict['warnings']) > 0:
            return httplib.CONFLICT

        # No errors or warnings, send back a 200 OK
        return httplib.OK


class Host_ovs(extensions.ExtensionDescriptor):
    """
    Capability to manager OpenVSwitch config on all hosts.
    """
    name = "Compute Host OVS"
    alias = "host-ovs"
    namespace = "http://docs.openstack.org/compute/ext/ibm_host_ovs/api/v1.1"
    updated = "2013-05-30T11:00:00-06:00"

    def get_resources(self):
        """
        Provide a Resource Extension for host-ovs
        """
        return [
            extensions.ResourceExtension(
                'host-ovs',
                controller=HostOvsController(),
                collection_actions={'update': 'PUT'})]


class Host_ovsV21(extensions.V21APIExtensionBase):
    """
    Capability to manager OpenVSwitch config on all hosts.
    """
    name = "Compute Host OVS"
    alias = ALIAS
    version = 1

    def get_resources(self):
        """
        Provide a Resource Extension for host-ovs
        """
        return [
            extensions.ResourceExtension(
                'host-ovs',
                controller=HostOvsController(),
                collection_actions={'update': 'PUT'})]

    def get_controller_extensions(self):
        return []
