# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import httplib
from webob import Response
import webob
import six
from nova import exception
from oslo_serialization import jsonutils
from nova.api.openstack import extensions
from nova.api.openstack import wsgi
from oslo_log import log as logging
from powervc_nova import _
import powervc_nova.objects.storage.dom as storage_dom
from powervc_nova.storage import topo_common
from powervc_nova.storage import storage_utils as su

# authorizations
ALIAS = "host-storage-topologies"
AUTHORIZE_V21 = extensions.os_compute_authorizer(ALIAS)
AUTHORIZE_V2 = extensions.extension_authorizer('compute',
                                               'host_storage_topologies')
# Note that the 'metadata' extension is really just for authorization
# purposes. It does not show up in the URI for my attempts.

LOG = logging.getLogger(__name__)


class HostStorageTopoController(wsgi.Controller):
    """
    This Controller handles APIs for retrieving topology information
    related to the configuration of storage connectivity groups and
    FC ports.
    """

    @wsgi.extends
    def index(self, req):
        """
        REST GET method to return full host storage topology info.
        Options:
            host_name: If passed, then the topology is limited to a list of
                       this single host. This is the hypervisor host name
                       (MTMS name), not the host's display name.
            port_tag:  If passed, then use the port_tag to further filter
                       the counts of ready ports, and thus storage-ready
                       vioses.
            provider_check_type: If given, then VIOSes will NOT be
                considered 'storage-ready' if the backend provider is in
                error state. If the option is given, allowable values are:

                    'cluster' - If a VIOS is a member of a cluster, then that
                           cluster must be not in error state for the VIOS
                           to be storage-ready.
                    'external' - There must be at least one registered SAN
                           provider not in error state for the VIOS to be
                           considered storage ready. This aligns with
                           NPIV-only connectivity.
                    'any' - If either the cluster provider or any SAN provider
                           is not in error state, then the VIOS can be
                           storage-ready, provided it passes the other
                           checks.
                If the option is not given, then don't make VIOS storage
                readiness depend on provider readiness. This is backwards
                compatible with how the API worked before.
        :returns: A json structure of storage topology that can be used to
                  create SCGs do FC port configuration and other related
                  APIs. EXAMPLE:
            {
             "host_list": [
              {
               "vios_ready_count": 2,
               "name": "8231E2D_109EFET",
               "port_ready_count": 8,
               "hmc_uuid": "230cd59c-fa61-3c1c-b621-e62645120bf3",
               "vios_list": [
                {
                 "id": "8231E2D_109EFET##2",
                 "cluster_provider_name": "22d22e9ac88d0111e3b6e900006cae8b02",
                 "cluster_provider_display_name": "p730_SSP_B",
                 "name": "vios2230_2",
                 "storage_ready": true,
                 "lpar_id": 2,
                 "total_fcport_count": 2,
                 "state": "running",
                 "rmc_state": "active",
                 "port_ready_count": 2,
                 "cluster_provider_state": "running",
                 "fcport_list": [
                  {
                   "status": "OK",
                   "fabric": "A",
                   "vio_server": "vios2230_2",
                   "available_connections": 64,
                   "enabled": true,
                   "adapter_id": "U78AB.001.WZSJMPP-P1-C4",
                   "wwpn": "10000090FA2A4E00",
                   "id": "1aU78AB.001.WZSJMPP-P1-C4-T1",
                   "name": "fcs0"
                  },
                  {
                   "status": "OK",
                   "fabric": "B",
                   "vio_server": "vios2230_2",
                   "available_connections": 64,
                   "enabled": true,
                   "adapter_id": "U78AB.001.WZSJMPP-P1-C4",
                   "wwpn": "10000090FA2A4E01",
                   "id": "1aU78AB.001.WZSJMPP-P1-C4-T2",
                   "name": "fcs1"
                  }
                 ]
                },
                {
                 "id": "8231E2D_109EFET##1",
                 "cluster_provider_name": "226da740207fa311e391ee00006cae8b02",
                 "cluster_provider_display_name": "PVC_p730_A",
                 "name": "vios2230_1",
                 "storage_ready": true,
                 "lpar_id": 1,
                 "total_fcport_count": 2,
                 "state": "running",
                 "rmc_state": "active",
                 "port_ready_count": 2,
                 "cluster_provider_state": "running",
                 "fcport_list": [
                  {
                   "status": "OK",
                   "fabric": "B",
                   "vio_server": "vios2230_1",
                   "available_connections": 64,
                   "enabled": true,
                   "adapter_id": "U78AB.001.WZSJMPP-P1-C2",
                   "wwpn": "10000090FA2A550D",
                   "id": "1aU78AB.001.WZSJMPP-P1-C2-T2",
                   "name": "fcs1"
                  },
                  {
                   "status": "OK",
                   "fabric": "A",
                   "vio_server": "vios2230_1",
                   "available_connections": 64,
                   "enabled": true,
                   "adapter_id": "U78AB.001.WZSJMPP-P1-C2",
                   "wwpn": "10000090FA2A550C",
                   "id": "1aU78AB.001.WZSJMPP-P1-C2-T1",
                   "name": "fcs0"
                  }
                 ]
                }
               ]
              }
             ]
            }
        """
        context = req.environ['nova.context']
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action='list')
        else:
            AUTHORIZE_V2(context, action='list')
        LOG.debug("index() **ENTER**  context=%s" % str(context))
        su.http_error_if_ivm()  # HMC-managed installs only
        port_tag = req.GET['port_tag'] if 'port_tag' in req.GET else None
        provider_check_type = req.GET['provider_check_type'] if\
            'provider_check_type' in req.GET else None
        if provider_check_type not in [None, 'cluster', 'external', 'any']:
            msg = _("The provider_check_type REST URI option must specify "
                    "a value of 'external', 'cluster', or 'any'.")
            LOG.error(msg)
            raise webob.exc.HTTPBadRequest(explanation=six.text_type(msg))
        host_name = req.GET['host_name'] if 'host_name' in req.GET else None

        try:
            topoBase = topo_common.StorageTopologyCommon(host_name, context)
            topo = topoBase.get_host_topology_from_db(port_tag,
                                                      provider_check_type)
        except exception.ComputeHostNotFound:
            msg = _("The host_name filter option provided with value '%s' "
                    "does not exist as a registered hypervisor host name.")
            msg = msg % host_name
            LOG.error(msg)
            raise webob.exc.HTTPNotFound(explanation=six.text_type(msg))
        except Exception as ex:
            LOG.exception(ex)
            raise webob.exc.HTTPServerError(explanation=six.text_type(ex))

        topo_dict = {'host_list': topo}

        # Return results in JSON body response
        return Response(request=req,
                        status=httplib.OK,
                        content_type='application/json',
                        body=jsonutils.dumps(topo_dict))

    @wsgi.action("create")
    def create(self, req, body):
        """
        REST POST method to update FC port metadata.

        Persists PowerVC database info that will contain metadata for VIOS
        owned FC Ports. Currently, the metadata that can be updated includes:
            1. 'enabled' - if a port is marked disabled,
               then PowerVC should not use it to create virtual ports for
               LPAR deployment/relocation.
            2. 'port_tag' - Admin supplied string tag for a FC port. This tag
               can be specified in the SCG definition to filter the ports
               available for deployment/relocation.
            3. 'fc_protocol' - This can contain one of ['npiv', 'vscsi',
                'any', None]. This specifies the type of connectivity
                the port can be used in. Default is 'any'.
            4. 'fabric' - The fabric designation that the port is
               "logged into". In PowerVC 1.2.1 and later releases, the
               fabric designation is automatically determined for registered
               fabric switches. However, the administrator can still
               override the automatic setting - for example, in cases where
               the port is cabled differently later or different fabrics
               become registered or where fabrics are not managed.
               Possible values are "A", "B", or "None".

        Example request body:
            {
             "fcport_list": [
                {
                   "enabled": true,
                   "port_tag": "newtag",
                   "id": "1aU789D.001.DQD50G3-P1-C1-T1",
                   "fc_protocol": "any",
                   "fabric": "A"
                },
                {
                   "enabled": false,
                   "id": "1aU789D.001.DQD50G3-P1-C1-T2",
                   "fabric": "B"
                }
             ]
            }
        """
        context = req.environ['nova.context']
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action='metadata')
        else:
            AUTHORIZE_V2(context, action='metadata')

        LOG.debug("create() **ENTER**  context=%s, body=%s"
                  % (str(context), str(body)))
        su.http_error_if_ivm()  # HMC-managed installs only

        self._set_fc_port_metadata(context, req, body)
        # Return HTTP NO_CONTENT
        return Response(status=httplib.NO_CONTENT)

    def _set_fc_port_metadata(self, context, req, body):
        # gets the list if fcports being updated
        if not isinstance(body, dict):
            msg = _("The request body does not specify a dictionary.")
            LOG.error(msg)
            raise webob.exc.HTTPBadRequest(explanation=six.text_type(msg))

        if 'fcport_list' not in body:
            msg = _("The request body must specify an outer dictionary "
                    "with a key of 'fcport_list'. "
                    "The value is the list of fibre channel port IDs to set.")
            LOG.error(msg)
            raise webob.exc.HTTPBadRequest(explanation=six.text_type(msg))

        # get list of fcports to update
        port_list = body['fcport_list']
        LOG.debug("There are %d fcports in input list to update." %
                  len(port_list))
        scg_factory = storage_dom.StorageConnectivityGroupFactory.get_factory()
        txn = scg_factory.get_resource_transaction()
        with txn.begin():
            # validate every port to be sure exists and update input is good
            for fcport_entry in port_list:
                LOG.debug("Process fcport update %s in list." % fcport_entry)
                if 'id' not in fcport_entry:
                    msg = _("The request body fcport_list has an entry "
                            "that does not specify an id for a "
                            "fibre channel port to update.")
                    LOG.error(msg)
                    raise webob.exc.HTTPBadRequest(
                        explanation=six.text_type(msg))
                fc_id = fcport_entry["id"]
                LOG.debug("See if a fcport with id = %s exists." % fc_id)
                # get port for id NOTE will throw exception
                # if no port found with that id
                try:
                    fcport = su.get_fcport_by_id(context, txn, fc_id)
                except Exception as ex:
                    LOG.exception(ex)
                    raise webob.exc.HTTPNotFound(explanation=six.text_type(ex))
                # port with id is there so validate updates
                LOG.debug("Real fcport with id = %s exists. "
                          "Validate update values." % fc_id)
                self._validate_fcport_update(fcport_entry, fcport)

                # make sure update port
                for key in fcport_entry:
                    if key == "port_tag":
                        fcport.port_tag = fcport_entry[key]
                    elif key == "enabled":
                        fcport.enabled = fcport_entry[key]
                    elif key == "fc_protocol":
                        fcport.fc_protocol = fcport_entry[key]
                    elif key == "fabric":
                        fcport.fabric = fcport_entry[key]
                    elif key == "id":
                        # ignore since id already processed
                        pass
                    else:
                        msg = _("The request body fcport_list entry for ID "
                                "'%(fc_id)s' specified a Fibre Channel port "
                                "property name '%(property)s' that is not "
                                "valid.") % dict(fc_id=fc_id, property=key)
                        LOG.error(msg)
                        raise webob.exc.HTTPBadRequest(
                            explanation=six.text_type(msg))
                LOG.debug("Updated fcport with id = %s." % fc_id)
                # Do the save of this port update
                fcport.save()
        # all of the fcport updates written out here when With ends

    def _validate_fcport_update(self, fcport_entry, curFcport):
        """
        Check fcport update rules.
        :fcport: port being updated
        :fcport_entry: dictionary of the port resource to check
        :returns: Nothing. Exception thrown if not valid
        """
        allowedFabrics = set(map(chr, range(ord('A'), ord('Z')+1)))
        allowedFabrics.add('None')
        allowedFcProtocol = set(['npiv', 'vscsi', 'any'])
        key_set = set(fcport_entry.keys())

        if "fabric" in key_set:
            LOG.debug("Validate fabric value %s." % fcport_entry["fabric"])
            fabric = fcport_entry["fabric"]
            if fabric not in allowedFabrics:
                # fabric not valid input
                msg = _("The fabric property value '%(fabric)s' specified"
                        " for fcport_list entry id '%(fcid)s' is not valid."
                        " The allowed values are: %(allowed)s") %\
                    dict(fabric=fabric, fcid=fcport_entry["id"],
                         allowed=str(allowedFabrics))
                LOG.error(msg)
                raise webob.exc.HTTPBadRequest(explanation=six.text_type(msg))
        if "enabled" in key_set:
            LOG.debug("Validate enabled value %s." % fcport_entry["enabled"])
            enabled = fcport_entry["enabled"]
            not_bool_str = True
            if isinstance(enabled, str) or isinstance(enabled, unicode):
                if enabled.lower() in ['true', 'false']:
                    fcport_entry["enabled"] = (enabled.lower() == 'true')
                    not_bool_str = False
            if not_bool_str and not isinstance(enabled, bool):
                # enabled not valid input
                msg = _("The 'enabled' property value '%(enabled)s' specified"
                        " for fcport_port list entry id '%(fcid)s' is not "
                        "valid. The allowed values are: true, false") %\
                    dict(enabled=enabled, fcid=fcport_entry["id"])
                LOG.error(msg)
                raise webob.exc.HTTPBadRequest(explanation=six.text_type(msg))
        enabled = fcport_entry["enabled"] if "enabled" in fcport_entry\
            else curFcport.enabled
        if "fc_protocol" in key_set:
            LOG.debug("Validate fc_protocol %s" % fcport_entry["fc_protocol"])
            fc_protocol = fcport_entry["fc_protocol"]
            is_str = isinstance(fc_protocol, str) or\
                isinstance(fc_protocol, unicode)
            if fc_protocol is not None and (not is_str or
                                            len(fc_protocol) > 31):
                msg = _("The fc_protocol property value '%(fc_protocol)s' "
                        "specified for fc_port list entry id '%(fcid)s' is not"
                        " valid. The value must be string and up to 31 "
                        "characters long.") % dict(fc_protocol=fc_protocol,
                                                   fcid=fcport_entry["id"])
                LOG.error(msg)
                raise webob.exc.HTTPBadRequest(explanation=six.text_type(msg))
            if fc_protocol is not None and\
                    fc_protocol not in allowedFcProtocol:
                msg = _("The fc_protocol property value '%(fc_protocol)s' "
                        "specified for fc_port list entry id '%(fcid)s' is not"
                        " valid. The allowed values are:'%(allowed)s' ") %\
                    dict(fc_protocol=fc_protocol, fcid=fcport_entry["id"],
                         allowed=str(allowedFcProtocol))
                LOG.error(msg)
                raise webob.exc.HTTPBadRequest(explanation=six.text_type(msg))

        if not enabled and "fc_protocol" in key_set:
            LOG.debug("Cannot give fc_protocol when port is disabled")
            msg = _("The port '%(fcid)s' is not enabled and cannot be used for"
                    " given protocol '%(fcprotocol)s'")
            msg = msg % dict(fcid=fcport_entry["id"],
                             fcprotocol=fcport_entry["fc_protocol"])
            LOG.error(msg)
            raise webob.exc.HTTPBadRequest(explanation=six.text_type(msg))

        if "port_tag" in key_set:
            LOG.debug("Validate port_tag %s." % fcport_entry["port_tag"])
            port_tag = fcport_entry["port_tag"]
            is_str = isinstance(port_tag, str) or isinstance(port_tag, unicode)
            if port_tag is not None and (not is_str or len(port_tag) > 255):
                # port tag too long
                msg = _("The port_tag property value '%(tag)s' specified for "
                        "the fcport_list entry id '%(fcid)s' is not valid. "
                        "The value must be a string up to 255 characters "
                        "long.") % dict(tag=str(port_tag),
                                        fcid=fcport_entry["id"])
                LOG.error(msg)
                raise webob.exc.HTTPBadRequest(explanation=six.text_type(msg))
            elif port_tag is not None and len(port_tag) == 0:
                LOG.debug("The port_tag property value for FC port id '%s'"
                          "is of length 0. Converting it to None." %
                          fcport_entry["id"])
                fcport_entry['port_tag'] = None
        LOG.debug("Parameter values for fcport are valid: %s." % fcport_entry)
        return


class Host_storage_topologies(extensions.ExtensionDescriptor):

    """ Support for host to storage topology mapping. """
    name = "Compute Host Storage Topologies"
    alias = "host-storage-topologies"
    # The namespace just follows the convention of other APIs - is there
    # separate work that needs to be done to register this?
    namespace = ("http://docs.openstack.org/compute/ext/"
                 "ibm_host_storage_topologies/api/v1.1")
    updated = "2014-02-12T11:00:00-06:00"

    def get_resources(self):
        """ Provide a Resource Extension for host-storage-topologies """
        return [
            extensions.ResourceExtension(
                'host-storage-topologies',
                controller=HostStorageTopoController(),
                collection_actions={'metadata': 'POST'})]


class Host_storage_topologiesV21(extensions.V21APIExtensionBase):

    """ Support for host to storage topology mapping. """
    name = "Compute Host Storage Topologies"
    alias = ALIAS
    version = 1

    def get_resources(self):
        """ Provide a Resource Extension for host-storage-topologies """
        return [
            extensions.ResourceExtension(
                'host-storage-topologies',
                controller=HostStorageTopoController(),
                collection_actions={'metadata': 'POST'})]

    def get_controller_extensions(self):
        return []
