#
#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import httplib
import webob
import six
from webob import Response

from oslo_serialization import jsonutils
from nova.api.openstack import extensions
from nova.api.openstack import wsgi
from powervc_nova.storage.storage_mapping_api import StorageMapping

from oslo_log import log as logging
from powervc_nova.storage import storage_utils as su

# ------------------------------------
# Extension: images
# ------------------------------------
LOG = logging.getLogger(__name__)
EXTENSION = 'serverlist'
ALIAS = "scg-to-servers"
AUTHORIZE_V21 = extensions.os_compute_authorizer(ALIAS)
AUTHORIZE_V2 = extensions.extension_authorizer('compute', "scg_to_servers")


class SCGServerListController(wsgi.Controller):

    """
    This Controller handles the API entry implementation of the REST
    function to retrieve servers that have been deployed, onboarded,
    or otherwise associated to a given SCG.
    The StorageMapping get_instances_for_scg() is used to get SCG
    references for each server and those are checked to see if they
    match SCG id passed in.
    """

    def __init__(self):
        # class statics for the internal API module
        self.stg_api = StorageMapping()

    def index(self, req, scg_id):
        """
        REST API entry point to return a dictionary containing a list of
        servers that are associated with storage connectivity group (SCG).
        If there are no servers found that reference an SCG
        then will return empty list.

        :param req: The REST request. No special options are supported on this
        :param scg_id: The scg uuid to base the request on.
        :returns: {
                   "servers": [
                       {
                        "display name": <server-name>
                        "id": <server-uuid>,
                       }
                      ]
                     }

        Flow:
            1. Validate input SCG
            2. Use StorageMapping to get list of servers related to SCG
            3. Return the list of servers that referenced the SCG.

        """
        context = req.environ["nova.context"]
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action='list')
        else:
            AUTHORIZE_V2(context, action='list')
        LOG.debug("index() **ENTRY** scg_id=%s" % str(scg_id))

        # NOTE servers = instances in this internal implementation.
        # "Server" seemed more appropriate for external name.

        # valid to make sure SCG is good
        su.http_error_if_ivm()  # HMC-managed installs only
        # Get the storage connectivity group and make sure SCG with id is real
        try:
            scg = su.get_scg_by_id(context, scg_id)
        except Exception as ex:
            LOG.exception(ex)
            raise webob.exc.HTTPNotFound(explanation=six.text_type(ex))
        LOG.debug(" Scg found for scg_id is: %s" % str(scg))

        msg_dict = {'messages': []}
        # get servers associated with this scg
        servers = self.stg_api.get_instances_for_scg(context,
                                                     scg, msg_dict)

        if not servers:
            # No instances referenced the SCG so give warning
            LOG.debug("NO referencing servers found for scg id = %s"
                      % (scg_id))
        else:
            LOG.debug("%s referencing instances found for scg id = %s"
                      % (len(servers), scg_id))

        # Return results as list of servers in JSON body response
        # format and return list even if empty list
        server_list = map(lambda x: {'id': x['uuid'] if
                                     'uuid' in x else x['id'],
                                     'display_name': x['display_name']},
                          servers)
        LOG.debug(str(server_list))
        # Return results in JSON body response
        output = {"servers": server_list}

        if msg_dict['messages']:
            # jsonutils does not do well with unicode strings which have now
            # converted to Message types in openstack, so convert them back
            msgs = [six.text_type(x) for x in msg_dict['messages']]
            output['messages'] = msgs

        return Response(request=req,
                        status=httplib.OK,
                        content_type='application/json',
                        body=jsonutils.dumps(output))


class Scg_to_servers(extensions.ExtensionDescriptor):

    """
    Support to list servers is associated with Storage Connectivity Group.
    """
    name = "Get SCG Servers List"
    alias = "scg-" + EXTENSION
    # The namespace just follows the convention of other APIs - is there
    # separate work that needs to be done to register this?
    namespace = ("http://docs.openstack.org/compute/ext/storage/"
                 "scg_to_servers/api/v1.1")
    updated = "2013-09-20T11:00:00-06:00"

    def get_resources(self):
        '''
        Provide a Resource Extension for listing all servers
        associated with a storage connectivity group.
        '''
        parent_def = {'member_name': 'scg',
                      'collection_name': 'storage-connectivity-groups'}
        return [
            extensions.ResourceExtension(
                EXTENSION,
                controller=SCGServerListController(),
                parent=parent_def)]


class Scg_to_serversV21(extensions.V21APIExtensionBase):

    """
    Support to list servers is associated with Storage Connectivity Group.
    """
    name = "Get SCG Servers List"
    alias = ALIAS
    version = 1

    def get_resources(self):
        '''
        Provide a Resource Extension for listing all servers
        associated with a storage connectivity group.
        '''
        parent_def = {'member_name': 'scg',
                      'collection_name': 'storage-connectivity-groups'}
        return [
            extensions.ResourceExtension(
                EXTENSION,
                controller=SCGServerListController(),
                parent=parent_def)]

    def get_controller_extensions(self):
        return []
