#
#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import httplib
import webob
import six
from webob import Response

from oslo_serialization import jsonutils
from nova.api.openstack import extensions
from nova.api.openstack import wsgi
from powervc_nova.storage.storage_mapping_api import StorageMapping

from oslo_log import log as logging
from powervc_nova import _
from powervc_nova.storage import storage_utils as su
from powervc_nova.volume.cinder import PROVIDER_KEY

# ------------------------------------
# Extension: images
# ------------------------------------
LOG = logging.getLogger(__name__)
EXTENSION = 'volumes'
ALIAS = "scg-to-volumes"
AUTHORIZE_V21 = extensions.os_compute_authorizer(ALIAS)
AUTHORIZE_V2 = extensions.extension_authorizer('compute', "scg_to_volumes")


class SCGVolumeListController(wsgi.Controller):

    """
    This Controller handles the API entry implementation of the REST
    function to retrieve the applicable volumes for an given SCG.
    The StorageMapping stg_api.get_scg_volume_refs is used to get SCG
    references for each volume and those are checked to see if they
    match SCG id passed in.
    """

    def __init__(self):
        # class statics for the internal API module
        self.stg_api = StorageMapping()

    def index(self, req, scg_id):
        """
        REST API entry point to return a dictionary containing a list of
        volumes that are associated with storage connectivity group (SCG).
        If there are no volumes that SCG can provide connectivity to,
        then returns empty list.

        :param req: The REST request. No special options are supported on this
        :param scg_id: The scg uuid to base the request on.
        :returns: {
                   "volumes": [
                       {
                        "display_name": <volume-name>,
                        "id": <volume-uuid>,
                        "status": <volume-state>,
                        "volume_type": <vol-type>,
                        "provider": <storage provider id that owns volume>
                       }
                      ]
                     }

        Flow:
            1. Validate input SCG
            2. Call an internal API to do the work of obtaining volumes
               that the SCG implies access to, whether they are attached
               as disks, or not.
            3. Return the list of volumes.

        """
        context = req.environ["nova.context"]
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action='list')
        else:
            AUTHORIZE_V2(context, action='list')
        su.http_error_if_ivm()  # HMC-managed installs only
        # Get the storage connectivity group and make sure SCG with id is real
        try:
            scg = su.get_scg_by_id(context, scg_id)
        except Exception as ex:
            LOG.exception(ex)
            raise webob.exc.HTTPNotFound(explanation=six.text_type(ex))
        LOG.debug("Scg found for scg_id is: %s" % scg.to_dict())
        msg_dict = {'messages': []}

        # Handle query parameters passed
        include_boot = True
        if 'include_boot' in req.GET:
            if req.GET['include_boot'].lower() == "true":
                include_boot = True
            elif req.GET['include_boot'].lower() == "false":
                include_boot = False
            else:
                msg = _("The value given for the 'include_boot' option did "
                        "not parse to 'true' or 'false'.")
                LOG.error(msg)
                raise webob.exc.HTTPBadRequest(explanation=six.text_type(msg))
        include_unavailable = True
        if 'include_unavailable' in req.GET:
            if req.GET['include_unavailable'].lower() == "true":
                include_unavailable = True
            elif req.GET['include_unavailable'].lower() == "false":
                include_unavailable = False
            else:
                msg = _("The value given for the 'include_unavailable' option "
                        "did not parse to 'true' or 'false'.")
                LOG.error(msg)
                raise webob.exc.HTTPBadRequest(explanation=six.text_type(msg))
        details = False
        if 'details' in req.GET:
            if req.GET['details'].lower() == "true":
                details = True
            elif req.GET['details'].lower() == "false":
                details = False
            else:
                msg = _("The value given for the 'details' option "
                        "did not parse to 'true' or 'false'.")
                LOG.error(msg)
                raise webob.exc.HTTPBadRequest(explanation=six.text_type(msg))
        LOG.debug("index() **ENTRY** args:  scg_id=%s, details=%s, "
                  "include_unavailable=%s, include_boot=%s."
                  % (scg_id, details, include_unavailable, include_boot))
        try:
            # get volumes associated with this scg
            volumes = self.stg_api.get_scg_volume_refs(context, scg,
                                                       msg_dict=msg_dict)
        except webob.exc.HTTPError as httpe:
            LOG.exception(httpe)
            # We can re-raise this type
            LOG.error(_("Unable to obtain volumes for storage connectivity "
                        "group '%(name)s'. Error is: %(error)s") %
                      dict(name=scg.display_name, error=httpe))
            raise httpe
        except Exception as ex:
            LOG.exception(ex)
            msg = _("There was a problem attempting to obtain volumes for "
                    "storage connectivity group '%(name)s'. Error is: "
                    "%(error)s") % dict(name=scg.display_name, error=ex)
            LOG.error(msg)
            raise webob.exc.HTTPServerError(explanation=six.text_type(msg))

        volume_list = []
        if volumes:
            # there are volumes in the obtained list
            LOG.debug("%s referencing volumes found for scg id = %s"
                      % (len(volumes), scg_id))
            if not include_unavailable:
                volumes = [v for v in volumes if v['status'] == 'available']
            if not include_boot:
                volumes = [v for v in volumes if 'volume_metadata' not in v or
                           'is_boot_volume' not in v['volume_metadata'] or
                           v['volume_metadata']['is_boot_volume'].lower() ==
                           'false']
            # Return results as list of volumes in JSON body response.
            if details:
                for vol in volumes:
                    vol['provider'] = vol[PROVIDER_KEY]
                    del vol[PROVIDER_KEY]
                volume_list = volumes
            else:
                volume_list = map(lambda x:
                                  {'id': x['id'],
                                   'display_name': x['display_name'],
                                   'status': x['status'],
                                   'volume_type': x['volume_type'],
                                   'provider': x[PROVIDER_KEY]},
                                  volumes)
            LOG.debug(str(volume_list))
        else:
            LOG.debug("NO associated volumes found for scg id = %s" % (scg_id))

        # Return results in JSON body response
        output = {"volumes": volume_list}
        if msg_dict['messages']:
            # jsonutils does not do well with unicode strings which have now
            # converted to Message types in openstack, so convert them back
            msgs = [six.text_type(x) for x in msg_dict['messages']]
            output['messages'] = msgs

        return Response(request=req,
                        status=httplib.OK,
                        content_type='application/json',
                        body=jsonutils.dumps(output))


class Scg_to_volumes(extensions.ExtensionDescriptor):

    """
    Support to list volumes associated with Storage Connectivity Group.
    """
    name = "Get SCG Volumes List"
    alias = "scg-" + EXTENSION
    # The namespace just follows the convention of other APIs - is there
    # separate work that needs to be done to register this?
    namespace = ("http://docs.openstack.org/compute/ext/storage/"
                 "scg_volumes/api/v1.1")
    updated = "2014-01-24T11:00:00-06:00"

    def get_resources(self):
        '''
        Provide a Resource Extension for listing all volumes
        associated with a storage connectivity group.
        '''
        parent_def = {'member_name': 'scg',
                      'collection_name': 'storage-connectivity-groups'}
        return [
            extensions.ResourceExtension(
                EXTENSION,
                controller=SCGVolumeListController(),
                parent=parent_def)]


class Scg_to_volumesV21(extensions.V21APIExtensionBase):

    """
    Support to list volumes associated with Storage Connectivity Group.
    """
    name = "Get SCG Volumes List"
    alias = ALIAS
    version = 1

    def get_resources(self):
        '''
        Provide a Resource Extension for listing all volumes
        associated with a storage connectivity group.
        '''
        parent_def = {'member_name': 'scg',
                      'collection_name': 'storage-connectivity-groups'}
        return [
            extensions.ResourceExtension(
                EXTENSION,
                controller=SCGVolumeListController(),
                parent=parent_def)]

    def get_controller_extensions(self):
        return []
