#
#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from nova.api.openstack import extensions
from nova.api.openstack import wsgi
from oslo_log import log as logging
from oslo_config import cfg

from powervc_oslo.__init__ import powervc_api_filter

LOG = logging.getLogger(__name__)
SCG_KEY = 'storage_connectivity_group_id'
HYPERVISOR_TYPE_POWERVM = 'phyp'
HYPERVISOR_TYPE_POWERKVM = 'QEMU'
HOST_TYPE_KEY_NAME = 'host_type'
HOST_TYPE_KEY_KVM_VALUE = 'kvm'


class ServerDetailSCGController(wsgi.Controller):
    """Controller Class to extend the servers API with the SCG"""

    @wsgi.extends
    @powervc_api_filter
    def show(self, req, resp_obj, id):
        """Rest GET Method to add the Pending On-board Actions attribute to"""
        ctxt = req.environ['nova.context']
        # Add the Pending On-board Actions attribute to the given Server
        self._add_scg_attribute(ctxt, req, resp_obj.obj['server'])

    @wsgi.extends
    @powervc_api_filter
    def detail(self, req, resp_obj):
        """Rest GET Method to add the Pending On-board Actions attributes to"""
        context = req.environ['nova.context']
        # Add the Pending On-board Actions attribute to each of the Servers
        for server in resp_obj.obj['servers']:
            self._add_scg_attribute(context, req, server)

    def _add_scg_attribute(self, context, req, server):
        """Helper method to add the SCG Identifier to the Server"""
        try:
            # Use the Server cache that the main extension created from the DB
            db_inst = req.get_db_instance(server['id'])
            hypervisor_type = self._get_platform_type()
            if hypervisor_type == HYPERVISOR_TYPE_POWERVM:
                LOG.debug("Adding SCG for platform " + HYPERVISOR_TYPE_POWERVM)
                metadata = db_inst.get('system_metadata')
                metadata = metadata if metadata is not None else dict()
                # Set the SCG Identifier on the Server from the System MetaData
                server[SCG_KEY] = metadata.get(SCG_KEY)
        # If we go an exception getting the Server from DB, just ignore it
        # here since the main extension will fail, and we want to play nice
        except:
            pass

    def _get_platform_type(self):
        auth_host_type = cfg.CONF.get(HOST_TYPE_KEY_NAME, None)
        if auth_host_type:
            if auth_host_type == HOST_TYPE_KEY_KVM_VALUE:
                return HYPERVISOR_TYPE_POWERKVM
        return HYPERVISOR_TYPE_POWERVM


class Server_detail_scg(extensions.ExtensionDescriptor):
    """Provides the Storage Connectivity Group for a Server"""
    name = "Server Storage Connectivity Group Attribute Extension"
    alias = "server-storage-connectivity-group"
    namespace = "http://docs.openstack.org/compute/ext/" + \
                "server/server_detail_scg/api/v1.2"
    updated = "2013-09-23T11:00:00-06:00"

    def get_controller_extensions(self):
        controller = ServerDetailSCGController()
        extension = extensions.ControllerExtension(self, 'servers', controller)
        return [extension]


class Server_detail_scgV21(extensions.V21APIExtensionBase):
    """Provides the Storage Connectivity Group for a Server"""
    name = "Server Storage Connectivity Group Attribute Extension"
    alias = "server-storage-connectivity-group"
    version = 1

    def get_controller_extensions(self):
        controller = ServerDetailSCGController()
        extension = extensions.ControllerExtension(self, 'servers', controller)
        return [extension]

    def get_resources(self):
        return []
