#
#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import httplib
import webob
import six
from webob import Response

from oslo_serialization import jsonutils
from nova.api.openstack import extensions
from nova.api.openstack import wsgi
from nova import compute
from powervc_nova.storage.storage_mapping_api import StorageMapping

from oslo_log import log as logging
from powervc_nova import _
from powervc_nova.storage import storage_utils as su
from storage_connectivity_groups import DISPLAY_NAME
from storage_connectivity_groups import ALIAS

# ------------------------------------
# Extension: images
# ------------------------------------
LOG = logging.getLogger(__name__)
EXTENSION_ALIAS = "server-to-scg"
AUTHORIZE_V21 = extensions.os_compute_authorizer(EXTENSION_ALIAS)
AUTHORIZE_V2 = extensions.extension_authorizer('compute', 'server_to_scg')


class ServerToSCGController(wsgi.Controller):

    """
    server-->flavor-->extra_specs-->"storage_connectivity_group"

    This Controller handles the API entry implementation of the REST
    function to retrieve the applicable SCG (Storage Connectivity Group)
    for a given instance (virtual server).
    Return the SCG associated with the instance in the instance's
    system_metadata properties.
    """

    def __init__(self):
        # class statics for the internal API module
        self.stg_api = StorageMapping()
        self.compute_api = compute.API()

    def _get_image_list(self, context):
        return self.image_service.detail(context)

    def index(self, req, server_id):
        """
        REST API entry point to return a dictionary containing a list of
        storage connectivity groups (SCG) that are associated with a server.
        Returns empty list if server meta data does not reference an SCG.

        :param req: The REST request. No special options are supported.
        :param server_id: The instance (server) id to base the request on.
        :returns: {
                   "storage_connectivity_groups": [
                       {
                        "display_name": <scg-name>
                        "id": <scg-uuid>,
                       }
                      ]
                     }
        URI e.g.: GET http://localhost:8774/v2/
        8b930f024ec74da699c75bd6baa5fd8a/servers/
        <server-id>/storage-connectivity-groups HTTP/1.1

        Flow:
        server-->server-metadata-->"storage_connectivity_group"
            1. Look up the server(instance) from nova.
            2. Find the SCG using the server's metadata
            storage_connectivity_group_id
            3. Return the SCG referenced by the server(instance).

        """
        context = req.environ["nova.context"]
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action='list')
        else:
            AUTHORIZE_V2(context, action='list')
        LOG.debug("index() **ENTRY**   server_id=%s" % str(server_id))

        # NOTE servers = instances in this internal implementation.
        # "Server" seemed more appropriate for external name.

        # validation - SCG-based requests are not supported for IVM/express
        su.http_error_if_ivm()  # HMC-managed installs only

        # Get the server(instance) associated with the provided instance_id
        try:
            server = self.compute_api.get(context, server_id)
        except Exception as e:
            LOG.exception(e)
            raise webob.exc.HTTPNotFound(explanation=six.text_type(e))
        LOG.debug('Server(instance) found for server_id is: %s' % str(server))

        # Get the SCG associated with the server
        msg_dict = {'scg': None, 'messages': []}
        try:
            scg = self.stg_api.get_scg_from_instance_meta(context,
                                                          server,
                                                          msg_dict)
        except Exception as ex:
            msg = _("Failed trying to get scg using server meta information "
                    "for server id %s.") % (server_id)
            LOG.error(msg)
            LOG.exception(ex)
            raise webob.exc.HTTPBadRequest(explanation=six.text_type(ex))

        scg_map = []
        if scg:
            LOG.debug(" SCG found for server_id is: %s" % str(scg))
            # PowerVC only supports a single scg associated with an server,
            # however for future extensibility, treat it as a list.
            # Returns list even if empty list
            scg_list = [scg]
            scg_map = map(lambda x: {'id': x.id,
                                     DISPLAY_NAME: x.display_name}, scg_list)
            LOG.debug(str(scg_map))

        else:
            # Server did not reference SCG so give warning
            LOG.debug("NO Storage Connectivity Group found "
                      "for server id = %s" % server_id)

        # Return results in JSON body response
        output = {"storage_connectivity_groups": scg_map}

        if msg_dict['messages']:
            # jsonutils doesn't do well with unicode strings which have now
            # converted to Message types in openstack, so convert them back
            msgs = [six.text_type(x) for x in msg_dict['messages']]
            output['messages'] = msgs

        return Response(request=req,
                        status=httplib.OK,
                        content_type='application/json',
                        body=jsonutils.dumps(output))


class Server_to_scg(extensions.ExtensionDescriptor):
    """
    Support to list Storage Connectivity Group associated with a Server.
    """
    name = "Get the SCG for a Server"
    alias = "server-" + ALIAS
    # The namespace just follows the convention of other APIs - is there
    # separate work that needs to be done to register this?
    namespace = ("http://docs.openstack.org/compute/ext/server/"
                 "server_to_scg/api/v1.1")
    updated = "2013-09-23T11:00:00-06:00"

    def get_resources(self):
        '''
        Provide a Resource Extension for listing the SCG
        associated with a server.
        '''
        parent_def = {'member_name': 'server',
                      'collection_name': 'servers'}
        return [
            extensions.ResourceExtension(
                ALIAS,
                controller=ServerToSCGController(),
                parent=parent_def)]


class Server_to_scgV21(extensions.V21APIExtensionBase):
    """
    Support to list Storage Connectivity Group associated with a Server.
    """
    name = "Get the SCG for a Server"
    alias = "server-" + ALIAS
    version = 1

    def get_resources(self):
        '''
        Provide a Resource Extension for listing the SCG
        associated with a server.
        '''
        parent_def = {'member_name': 'server',
                      'collection_name': 'servers'}
        return [
            extensions.ResourceExtension(
                ALIAS,
                controller=ServerToSCGController(),
                parent=parent_def)]

    def get_controller_extensions(self):
        return []
