#
#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import httplib
import json
import webob
from webob import Response
from nova import exception as novaexception
from nova import objects
from nova.api.openstack import extensions
from nova.api.openstack import wsgi
from oslo_log import log as logging
from oslo_config import cfg
from powervc_nova.common import aggregate_utils
from powervc_nova.common import shared_processor_pools_utils
from powervc_nova import _

ALIAS = 'shared-processor-pools'
AUTHORIZE_V21 = extensions.os_compute_authorizer(ALIAS)
AUTHORIZE_V2 = extensions.extension_authorizer('compute',
                                               'shared_processor_pools')

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class SharedProcessorPoolsController(wsgi.Controller):
    """
    This Controller handles GET requests to query the database and return JSON
    data of shared processor pools for use by deploy.
    """

    @wsgi.extends
    def index(self, req):
        """
        This method handles GET requests to fetch shared processor pool data..
        :param req: HTTP Request
        :returns: HTTP Response with shared processor pool data.
        """
        context = self._validate_authorization(req.environ['nova.context'],
                                               req, action='index')

        # Parse the host aggregate id from the URI, if present
        agg_id = None
        if 'aggregate' in req.GET:
            agg_id = req.GET['aggregate']
            # Validate the specified host aggregate exists
            host_aggs = aggregate_utils.get_aggregate_list(context)
            host_agg_ids = [str(x.id) for x in host_aggs]
            if not str(agg_id) in host_agg_ids:
                raise novaexception.AggregateNotFound(aggregate_id=agg_id)

        # Parse the host name from the URI, if present
        host_name = None
        if 'host' in req.GET:
            host_name = req.GET['host']
            # Validate the specified host exists
            hosts = self._find_all_host_names(context)
            if not host_name in hosts:
                raise novaexception.ComputeHostNotFound(host=host_name)

        # Both a host name and a host aggregate id cannot be accepted in
        # the URI. Only one parameter can be specified in the URI
        if agg_id is not None and host_name is not None:
            reason = _("Specify only 1 parameter in the URI, "
                       "either a host name OR a host aggregate id.")
            return Response(request=req,
                            status=httplib.BAD_REQUEST,
                            content_type='application/json',
                            body=reason)

        # Parse the name_only param from the URI, if present
        list_name_only = None
        if 'name_only' in req.GET:
            list_name_only = req.GET['name_only'].lower()

        hosts = []
        if not host_name and not agg_id:
            # Find all hosts on compute node
            hosts = self._find_all_host_names(context)
        elif agg_id:
            # Find all hosts in the host aggregate
            hosts = aggregate_utils.get_hosts_for_aggregate(context, agg_id)
        else:
            hosts.append(host_name)

        # Get the shared proc pools dictionary for a given list of host(s)
        response = []
        response_name = set()
        for host in hosts:
            host_response = shared_processor_pools_utils.\
                get_all_shared_proc_pools(context, host, list_name_only)
            if list_name_only == 'true':
                response_name = response_name.union(host_response)
            else:
                response.extend(host_response)

        # Return the results
        if list_name_only == 'true':
            shared_proc_pool = {'pools': list(response_name)}
        else:
            shared_proc_pool = {'pools': response}
        return webob.Response(request=req,
                              status=httplib.OK,
                              content_type='application/json',
                              body=json.dumps(shared_proc_pool))

    @staticmethod
    def _validate_authorization(context, req, action):
        """
        This method authorizes the request against the service's policy.
        :param context: API request context.
        :param authorize_method: Which method from this class to authorize.
        :returns: An authorized request context.
        """
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action=action)
        else:
            AUTHORIZE_V2(context, action=action)
        return context.elevated()

    def _find_all_host_names(self, context):
        """
        Finds a list of compute node names.
        :param context: DB context object.
        :returns: A list of compute node names.
        """
        context = context.elevated()
        return [node.host for node in objects.ComputeNodeList.get_all(context)]


class Shared_processor_pools(extensions.ExtensionDescriptor):
    """
    Capability to list shared processor pools on all hosts.
    """
    name = "Compute Shared Processor Pools"
    alias = "shared-processor-pools"
    namespace = "http://docs.openstack.org/compute/ext\
                /ibm_host_shared_processor_pools/api/v1.1"
    updated = "2013-12-09T11:00:00-06:00"

    def get_resources(self):
        """
        Provide a Resource Extension for shared-processor-pools
        """
        return [extensions.ResourceExtension('shared-proc-pools',
                controller=SharedProcessorPoolsController())]


class Shared_processor_poolsV21(extensions.V21APIExtensionBase):
    """
    Capability to list shared processor pools on all hosts.
    """
    name = "Compute Shared Processor Pools"
    alias = ALIAS
    version = 1

    def get_resources(self):
        """
        Provide a Resource Extension for shared-processor-pools
        """
        return [extensions.ResourceExtension('shared-proc-pools',
                controller=SharedProcessorPoolsController())]

    def get_controller_extensions(self):
        return []
