# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import httplib
import six
from nova import db
from nova.api.openstack import extensions, wsgi
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_db.exception import DBError
from powervc_nova import _, populate_locale
from webob import Response
import webob.exc as wex
from webob.exc import HTTPError

from powervc_nova.storage import exception as stgex, storage_utils as su
from powervc_nova.storage.storage_mapping_api import StorageMapping
from powervc_nova.volume import cinder as pvc_cinder
from powervc_nova.objects.storage.dom import StorageConnectivityGroupFactory

from powervc_nova.virt.ibmpowervm.common.constants \
    import CONNECTION_TYPE_NPIV as _NPIV_, CONNECTION_TYPE_SSP as _SSP_,\
    CONNECTION_TYPE_VSCSI as _VSCSI_

# Common keys that may need to change
DISPLAY_NAME = 'display_name'
ALIAS = 'storage-connectivity-groups'
IDENTITY = "storage_connectivity_groups"
SINGLE_SCG = "storage_connectivity_group"

# authorizations
AUTHORIZE_V21 = extensions.os_compute_authorizer(ALIAS)
AUTHORIZE_V2 = extensions.extension_authorizer('compute',
                                               "storage_connectivity_groups")

LOG = logging.getLogger(__name__)
EDITABLE_KEYS = {DISPLAY_NAME: 'string', "enabled": 'bool',
                 "fc_storage_access": 'bool', "auto_add_vios": 'bool',
                 "port_tag": 'string', "cluster_provider_name": 'string',
                 "vios_ids": 'list', 'vios_redundancy': 'integer',
                 'exact_redundancy': 'bool', 'npiv_fabric': 'string',
                 'boot_connectivity': 'list', 'data_connectivity': 'list',
                 'include_untagged': 'bool', 'fabric_set_npiv': 'list',
                 'fabric_set_req': 'string',
                 'ports_per_fabric_npiv': 'integer'}

NON_EDITABLE_KEYS = ["id", "auto-defined", "vios_cluster", "host_list",
                     "updated_at", "created_at", "priority_external",
                     "priority_cluster", "port_ready_count", "vios_cluster",
                     "vios_ready_count", "applicable_providers",
                     "dynamic_fabric_set"]
MAX_FABS = 25


class StorageConnGroupController(wsgi.Controller):

    """
    This Controller handles GET, POST and PUT requests to query the database
    and return JSON data manage storage connectivity groups.
    """

    def __init__(self):
        # class statics for the internal API module
        self.stg_api = StorageMapping()

    @wsgi.extends
    def index(self, req):
        """
        REST GET method to return a short listing (non-detailed)
        of all applicable storage connectivity groups (SCGs).

        Applicability is determined by supported options:
            sort_type=external|cluster
                    The returned list will be in priority order. An SCG
                    could have both a cluster priority and an external
                    priority. If a sort type is specified, only SCGs that
                    containing a priority matching the sort type will be
                    returned. If a sort_type of cluster is specified and
                    the cluster_provider option is given, then the priority
                    order for only those groups applicable to the
                    particular vios cluster are returned.
            enabled=True|False
                    Only return groups that are not disabled.
            cluster_provider=<vios-cluster-provider>
                    Filter returned SCGs to those defined for the given
                    vios cluster provider. This is a property of SCGs -
                    the provider_id within the vios_cluster substructure.
            deploy_ready_only=True|False
                    Default is False. If True, return only those SCGs
                    that match other options and are valid deploy candidates
                    according to what information is currently available
                    in the nova database.
        """
        context = req.environ['nova.context']
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action='list')
        else:
            AUTHORIZE_V2(context, action='list')
        context = self._populate_locale(context, req)
        LOG.debug("index() **ENTER**  context=%s" % str(context))
        su.http_error_if_ivm()  # HMC-managed installs only
        # Get the filtered list of SCGs
        data = self._lookup_scgs(context, req, details=False)
        data_resp = {IDENTITY: data}

        # Return results in JSON body response
        return Response(request=req,
                        status=httplib.OK,
                        content_type='application/json',
                        body=jsonutils.dumps(data_resp))

    @wsgi.extends
    def show(self, req, id):
        """
        REST GET method to return details on one storage connectivity group.
        """
        context = req.environ['nova.context']
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action='show')
        else:
            AUTHORIZE_V2(context, action='show')
        context = self._populate_locale(context, req)

        LOG.debug("show() **ENTER**  id=%s, context=%s"
                  % (id, str(context)))
        su.http_error_if_ivm()  # HMC-managed installs only
        include_ports = False
        if 'include_ports' in req.GET:
            if req.GET['include_ports'].lower() == "true":
                include_ports = True
            elif req.GET['include_ports'].lower() == "false":
                include_ports = False
            else:
                msg = _("The value given for the 'include_ports' option did "
                        "not parse to 'true' or 'false'.")
                LOG.error(msg)
                raise wex.HTTPBadRequest(explanation=six.text_type(msg))
        include_conn = False
        if 'include_connectivity' in req.GET:
            if req.GET['include_connectivity'].lower() == "true":
                include_conn = True
            elif req.GET['include_connectivity'].lower() == "false":
                include_conn = False
            else:
                msg = _("The value given for the 'include_connectivity' "
                        "option did not parse to 'true' or 'false'.")
                LOG.error(msg)
                raise wex.HTTPBadRequest(explanation=six.text_type(msg))
        # Get the dictionary of storage connectivity group
        try:
            scg = su.get_scg_by_id(context, id)
        except Exception as ex:
            LOG.exception(ex)
            raise wex.HTTPNotFound(explanation=six.text_type(ex))

        if include_ports or include_conn:
            scg_dict = scg.to_dict_with_conn(context,
                                             include_ports=include_ports)
            # Remove properties that don't provide additional information
            # since only storage_ready ports are included.
            if include_ports:
                for host in scg_dict['host_list']:
                    for vios in host['vios_list']:
                        for port in vios['fcport_list']:
                            if 'port_tag' in port:
                                del port['port_tag']
                            if 'vio_server' in port:
                                del port['vio_server']
                            del port['enabled']
        else:
            scg_dict = scg.to_dict(context=context)
        data_resp = {SINGLE_SCG: scg_dict}
        # Return results in JSON body response
        return Response(request=req,
                        status=httplib.OK,
                        content_type='application/json',
                        body=jsonutils.dumps(data_resp))

    @wsgi.extends
    def detail(self, req):
        """
        REST GET method to return full details on applicable
        storage connectivity groups. There is an additional details option:
            include_ports=<True|False>
                    Default is False. If True, include FC port information
                    in the 'host-list' hierarchy of member resources for
                    the SCG. FC ports are optional since they apply only
                    to particular storage connectivity technology, whereas
                    the product may support others.

        Applicability is determined by other supported options:
            sort_type=external|cluster
                    The returned list will be in priority order. An SCG
                    could have both a cluster priority and an external
                    priority. If a sort type is specified, only SCGs that
                    containing a priority matching the sort type will be
                    returned. If a sort_type of cluster is specified and
                    the cluster_provider option is given, then the priority
                    order for only those groups applicable to the
                    particular vios cluster are returned.
            enabled_only=True|False
                    Only return groups that are not disabled. Default is
                    False.
            cluster_provider=<vios-cluster-provider>
                    Filter returned SCGs to those defined for the given
                    vios cluster provider. This is a property of SCGs -
                    the provider_id within the vios_cluster substructure.
            deploy_ready_only=True|False
                    Default is False. If True, return only those SCGs
                    that match other options and are valid deploy candidates
                    according to what information is currently available
                    in the nova database.
        """
        context = req.environ['nova.context']
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action='detail')
        else:
            AUTHORIZE_V2(context, action='detail')
        context = self._populate_locale(context, req)

        LOG.debug("detail() **ENTER**  context=%s" % str(context))
        su.http_error_if_ivm()  # HMC-managed installs only

        # Get the filtered list of SCGs
        data = self._lookup_scgs(context, req, details=True)
        data_resp = {IDENTITY: data}
        # Return results in JSON body response
        return Response(request=req,
                        status=httplib.OK,
                        content_type='application/json',
                        body=jsonutils.dumps(data_resp))

    def quickdetail(self, req):
        """
        REST GET method to return quick details on applicable
        storage connectivity groups. This detail variation does not do
        extra nova and cinder database lookups to build-out the current
        connectivity and topology information of the SCG. So it will return
        just the raw properties.
        Applicability is determined by other supported options:
            sort_type=external|cluster
                    The returned list will be in priority order. An SCG
                    could have both a cluster priority and an external
                    priority. If a sort type is specified, only SCGs that
                    contain a priority matching the sort type will be
                    returned. If a sort_type of cluster is specified and
                    the cluster_provider option is given, then the priority
                    order for only those groups applicable to the
                    particular vios cluster are returned.
            enabled_only=True|False
                    Only return groups that are not disabled. Default is
                    False.
            cluster_provider=<vios-cluster-provider>
                    Filter returned SCGs to those defined for the given
                    vios cluster provider. This is a property of SCGs -
                    the provider_id within the vios_cluster substructure.
        """
        context = req.environ['nova.context']
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action='detail')
        else:
            AUTHORIZE_V2(context, action='detail')
        context = self._populate_locale(context, req)

        LOG.debug("quickdetail() **ENTER**  ")
        su.http_error_if_ivm()  # HMC-managed installs only

        # Get the filtered list of SCGs
        data = self._lookup_scgs(context, req, details=True, quickdetail=True)
        data_resp = {IDENTITY: data}
        # Return results in JSON body response
        return Response(request=req,
                        status=httplib.OK,
                        content_type='application/json',
                        body=jsonutils.dumps(data_resp))

    @wsgi.action("create")
    def create(self, req, body):
        """
        REST POST method to create a storage connectivity group.
        The request body takes a nested JSON dictionary structure with the
        properties that must be set for the new storage connectivity group.
        display_name, fc_storage_access, and vios_ids are required properties.
        The releases starting from IBM PowerVC 1.2.1 support the
        cluster_provider_name property to specify that the storage
        connectivity group allows connectivity to the shared storage pool
        for the provider.

        The response header Location contains the URL to access the new
        storage connectivity group resource. An error is returned for the
        following conditions:
          1. No properties are specified in the request body dictionary.
          2. A dictionary is not given in the request body.
          3. One or more required keys are missing in the request.
          4. A non user-editable property is specified in the request body.
          5. A badly formatted request structure as determined by the web
             server.
          6. Property values that have an improper type. For example, an
             integer is given when a string type is expected.
          7. A display_name is given that is a duplicate for another storage
             connectivity group (409 Conflict). They must be unique.
        """
        context = req.environ['nova.context']
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action='create')
        else:
            AUTHORIZE_V2(context, action='create')
        context = self._populate_locale(context, req)

        su.http_error_if_ivm()  # HMC-managed installs only

        if not isinstance(body, dict):
            msg = _("The request body does not specify a dictionary.")
            LOG.error(msg)
            raise wex.HTTPBadRequest(explanation=six.text_type(msg))

        if SINGLE_SCG not in body and "priority_lists" not in body:
            msg = _("The request body must specify an outer dictionary "
                    "with a key of 'storage_connectivity_group' or "
                    "'priority_lists'. The value is the data to set.")
            LOG.error(msg)
            raise wex.HTTPBadRequest(explanation=six.text_type(msg))

        for key in body.keys():
            if key == "priority_lists":
                # if priority_lists is given, then the value for this key
                # will be a dictionary of priority lists. The key for this
                # dictionary is either "external" or the vios_cluster_id.
                LOG.debug("create() priority_lists data=%s, context=%s"
                          % (str(body[key]), str(context)))
                self._set_priority_lists(body[key], context)
                # Return HTTP NO CONTENT
                return Response(status=httplib.NO_CONTENT)
            elif key == SINGLE_SCG:
                # Normal create of an SCG. The value of this key is a
                # dictionary containing the properties of the SCG.
                LOG.debug("create() SSG scg=%s, context=%s"
                          % (str(body[key]), str(context)))
                scg_dict = self._validate_create(body[key], context)
                LOG.debug("validated new scg: %s" % str(scg_dict))
                scg_name = scg_dict["display_name"]
                try:
                    scg = su.scg_insert(context, scg_dict)
                except DBError as ex:
                    LOG.exception(ex)
                    msg = _("A Database problem was encountered trying "
                            "to save the storage connectivity group (SCG) "
                            "'%s'. Check that the display_name is not already "
                            "in use by another SCG.") % scg_name
                    raise wex.HTTPConflict(explanation=six.text_type(msg))
                except stgex.IBMPowerVCViosNotFound as ex:
                    LOG.exception(ex)
                    # convert to HHTPNotFound
                    ex_msg = six.text_type(ex)
                    msg = _(" %(ex)s Check that the vios_ids field "
                            "passed in on the request contains "
                            "valid IDs from the 'host-"
                            "storage-topologies' API.") % dict(ex=ex_msg)
                    raise wex.HTTPNotFound(explanation=six.text_type(msg))
                except stgex.IBMPowerVCSCGConflict as ex:
                    LOG.exception(ex)
                    msg = _("There was a problem trying to create the "
                            "storage connectivity group '%(name)s'. Error: "
                            "%(error)s") % dict(name=scg_name, error=ex)
                    raise wex.HTTPConflict(explanation=six.text_type(msg))
                except Exception as ex:
                    LOG.exception(ex)
                    msg = _("There was a problem trying to create the "
                            "storage connectivity group '%(name)s'. Error: "
                            "%(error)s") % dict(name=scg_name, error=ex)
                    raise wex.HTTPServerError(explanation=six.text_type(msg))

                resp_body = {"created_id": scg.id}
                # Return HTTP OK
                resp = Response(request=req,
                                status=httplib.CREATED,
                                content_type='application/json',
                                body=jsonutils.dumps(resp_body))
                # add the URL to get SCG details to the response headers
                resp.location = req.url + ('/%s' % scg.id)
                return resp
            else:
                msg = _("The POST action type '%s' in the request body is "
                        "not supported.") % key
                LOG.error(msg)
                raise wex.HTTPBadRequest(explanation=six.text_type(msg))

    @wsgi.action("update")
    def update(self, req, id, body):
        """ REST PUT method to update/edit a storage connectivity group. """
        context = req.environ['nova.context']
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action='update')
        else:
            AUTHORIZE_V2(context, action='update')
        context = self._populate_locale(context, req)

        LOG.debug("update() **ENTER**  id=%s, body=%s, context=%s"
                  % (id, str(body), str(context)))
        su.http_error_if_ivm()  # HMC-managed installs only

        if not isinstance(body, dict):
            msg = _("The request body does not specify a dictionary.")
            LOG.error(msg)
            raise wex.HTTPBadRequest(explanation=six.text_type(msg))

        force = False
        if 'force' in body:
            force = body['force']
            if not isinstance(force, bool) and (
                    isinstance(force, str) or isinstance(force, unicode)) and\
                    force.lower() in ['true', 'false']:
                force = (force.lower() == 'true')
            if force:
                LOG.warn(_("The 'force' option was passed to the storage "
                           "connectivity group update operation. "
                           "This could result in virtual machines being "
                           "associated with an inconsistent storage "
                           "connectivity group. Request body=%s") % body)
            else:
                LOG.debug("body 'force' option set to default of 'false'.")
            del body['force']

        # if 'server_id' is the key in the body dict, then change the SCG
        # for instance. This allows an override to the default one chosen.
        if 'server_id' in body:
            try:
                LOG.warn(_("Warning: Attempting to change the storage "
                           "connectivity group ID set on the server with ID "
                           "'%(server)s' to '%(scg_id)s'.") %
                         dict(server=body['server_id'], scg_id=id))
                instance = db.instance_get_by_uuid(context, body['server_id'])
                scg = su.get_scg_by_id(context, id) if\
                    id.lower() != "none" else None
                self._set_scg_to_server(context, scg, instance, force)
                return Response(status=httplib.NO_CONTENT)
            except HTTPError as http_ex:
                raise http_ex  # rethrow
            except Exception as ex:
                LOG.exception(ex)
                msg = _("Problem occurred attempting to set a storage "
                        "connectivity group association with a virtual "
                        "machine. Error: %s") % ex
                raise wex.HTTPServerError(explanation=six.text_type(msg))

        # Else normal update SCG path
        if not (len(body) == 1 and SINGLE_SCG in body.keys()):
            msg = _("The request body dictionary must contain a single "
                    "entry with the key of '%s'." % SINGLE_SCG)
            LOG.error(msg)
            raise wex.HTTPBadRequest(explanation=six.text_type(msg))
        scg_req = body[SINGLE_SCG]
        try:
            self._db_update(context, id, scg_req, force)
        except HTTPError as http_ex:
            raise http_ex  # rethrow
        except DBError as ex:
            LOG.exception(ex)
            dbmsg = _("A Database problem was encountered trying "
                      "to update the storage connectivity group (SCG) with ID "
                      "'%s'.") % id
            if 'display_name' in scg_req:
                msg = (_("%s Check that the passed display name of "
                         " '%s' is not a duplicate of another SCG."
                         " The SCG names must be unique.") %
                       (dbmsg, scg_req['display_name']))
                raise wex.HTTPConflict(explanation=six.text_type(msg))
            else:
                msg = (_("%s Error is: %s") % (dbmsg, ex))
                raise wex.HTTPServerError(explanation=six.text_type(msg))
        except stgex.IBMPowerVCSCGConflict as ex:
            LOG.exception(ex)
            msg = _("There was a problem trying to update the "
                    "storage connectivity group with ID %(scgid)s. Error: "
                    "%(error)s") % dict(scgid=id, error=ex)
            raise wex.HTTPConflict(explanation=six.text_type(msg))
        except Exception as ex:
            LOG.exception(ex)
            raise wex.HTTPServerError(explanation=six.text_type(ex))
        # Return HTTP NO_CONTENT
        return Response(status=httplib.NO_CONTENT)

    @wsgi.action("delete")
    def delete(self, req, id):
        """ REST POST method to create a storage connectivity group. """
        context = req.environ['nova.context']
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action='delete')
        else:
            AUTHORIZE_V2(context, action='delete')
        context = self._populate_locale(context, req)

        su.http_error_if_ivm()  # HMC-managed installs only

        force = False
        if 'force' in req.GET and req.GET["force"] == "true":
            force = True
        LOG.debug("delete() **ENTER**  context=%s, id=%s, force=%s" %
                  (str(context), id, force))
        # db interface method will bubble up any errors through http
        self._db_remove(context, id, force)
        LOG.debug("delete() **EXIT**  OK  id=%s" % id)
        # No Content success return
        return Response(status=httplib.NO_CONTENT)

    @staticmethod
    def _populate_locale(context, req):
        """Internal Helper Method to populate locale from request"""
        return populate_locale(context.elevated(), req)

    def _db_remove(self, context, scg_id, force=False):
        """
        This method interfaces with the database and removes an SCG record
        and any associations it has.
        :params context: The authorization context
        :params scg_id: The SCG id string to reference the object to remove.
        :returns: Nothing returned. If error, an exception is raised.
        """
        err_msg = _("Cannot delete a Storage Connectivity Group that is "
                    "defined by the product ('auto-defined' = true)")
        try:
            scg = su.get_scg_by_id(context, scg_id, ignore_dict_err=True)
        except Exception as ex:
            LOG.exception(ex)
            raise wex.HTTPNotFound(explanation=six.text_type(ex))

        if scg.auto_defined is True and not force:
            raise wex.HTTPBadRequest(explanation=six.text_type(err_msg))

        # validate SCG delete
        if not force:
            self._validate_delete(context, scg, scg_id)
        try:
            scg_name = scg.display_name
            scg.delete()
            LOG.info(_("Storage connectivity group '%(scg_name)s' with ID "
                       "'%(scg_id)s' was removed.") % locals())
        except Exception as ex:
            LOG.exception(ex)
            raise wex.HTTPConflict(explanation=six.text_type(ex))
        # Note: for added robustness, we could do the retrieval
        #       above and the delete under the same database transaction
        #       object that we construct in this method. However,
        #       I don't think the REST response would look much
        #       different if the not-found error was detected during
        #       lookup or delete.
        return

    def _db_update(self, context, scg_id, new_scg, force):
        """
        This method interfaces with the database and updates a SCG.
        :param context: authorization context
        :param scg_id: The SCG id string to reference the object to update.
        :param new_scg: The SCG dictionary of keys/values to update
        :param force: Ignore some checks if true
        :param returns: The SCG dictionary after the update.
        """
        try:
            scg = su.get_scg_by_id(context, scg_id)
        except Exception as ex:
            LOG.exception(ex)
            raise wex.HTTPNotFound(explanation=six.text_type(ex))

        scg_upd = self._validate_update(new_scg, scg, context, force)
        # We construct a new transaction here so that the
        # association linkages to other resources are "active" and we
        # can update them on the scg object. So we first re-acquire the
        # SCG under the transaction.
        scg_factory = StorageConnectivityGroupFactory.get_factory()
        txn = scg_factory.get_resource_transaction()
        with txn.begin(subtransactions=True):
            scg = scg_factory.find_scg_by_id(context, scg_id, transaction=txn)
            if scg is None:
                msg = stgex.IBMPowerVCSCGNotFound.msg_fmt % locals()
                raise wex.HTTPNotFound(explanation=six.text_type(msg))
            # Loop over keys and update the object properties.
            upd_vios_list = None

            for key in scg_upd:
                LOG.debug("Prop key/val for to update: %s=%s)" %
                          (key, scg_upd[key]))
                # vios_ids does not need to be set in the scg yet. It will be
                # changed after the update is performed by replace_vios_list
                if key == 'vios_ids':
                    upd_vios_list = scg_upd["vios_ids"]
                    key = None
                if key in EDITABLE_KEYS:
                    setattr(scg, key, scg_upd[key])
                    # specific key unique updates
                    if key == 'vios_redundancy':
                        if upd_vios_list is None:
                            # use current list for validation purposes
                            upd_vios_list = scg.vios_ids
                    elif key == 'npiv_fabric' or key == 'fabric_set_req':
                        if (upd_vios_list is None and
                                scg.npiv_fabric is not None):
                            # use current list for validation purposes
                            upd_vios_list = scg.vios_ids
                else:
                    LOG.warn(_("Skipping unknown update key: %s.") % key)

            if upd_vios_list is not None:
                scg.replace_vios_list(context, txn, upd_vios_list, force=force)
            # Do the save
            scg.save(context=context, transaction=txn)
            updated_scg = scg.to_dict(context=context, transaction=txn,
                                      skip_extended_lookups=True)
            LOG.debug("Updated SCG dictionary: %s" % updated_scg)
        return updated_scg

    def _validate_delete(self, context, scg, scg_id):
        """
        Check SCG deletion rules.
        :param context: authorization context
        :param scg: SCG resource to check
        :param scg_id: SCG identifier related to SCG
        :returns: Nothing returned. If error, an exception is raised.
        """
        LOG.debug("_validate_delete() **ENTER** Validating scg = %s" %
                  scg)
        # note that this assumes the the meta information in an instance
        # is the final authority as far as intances that reference the
        # SCG being deleted
        msg_dict = {'messages': []}
        instances = self.stg_api.get_instances_for_scg(context, scg, msg_dict)
        if not instances:
            LOG.debug(">>> NO instances were found related to scg id = %s" %
                      scg_id)
        else:
            instance_names = map(lambda x: str(x['display_name']), instances)
            LOG.error(_("Server instances '%(server_list)s' were found related"
                        " to storage connectivity group ID '%(scg_id)s', "
                        "so the storage connectivity group cannot be deleted.")
                      % dict(server_list=str(instance_names), scg_id=scg_id))
            scg_name = scg.display_name
            msg = stgex.IBMPowerVCSCGRefByInstance.msg_fmt % locals()
            raise wex.HTTPBadRequest(explanation=six.text_type(msg))
        LOG.debug("_validate_delete() **EXIT** scg id%s" % scg_id)

    def _validate_create(self, scg, context):
        """
        Check SCG creation rules.
        :param scg: dictionary of the SCG resource to check
        :returns: The validated dictionary.
        """
        scg = self._validate_common_pre(scg, False)
        key_set = set(scg.keys())
        required_keys = [DISPLAY_NAME, "vios_ids", "fc_storage_access"]
        if not (set(required_keys).issubset(key_set)):
            msg = _("A required property is missing for create. The required "
                    "properties are: %s" % str(required_keys))
            LOG.error(msg)
            raise wex.HTTPBadRequest(explanation=six.text_type(msg))

        scg["auto_defined"] = False
        # Fill in defaults
        if "auto_add_vios" not in key_set:
            scg["auto_add_vios"] = False
        if "enabled" not in key_set:
            scg["enabled"] = True

        # fill in defaults for boot and data connectivity
        if 'boot_connectivity' not in scg or\
                not scg['boot_connectivity']:
            if 'cluster_provider_name' in scg and\
                    scg['cluster_provider_name'] is not None:
                if scg['fc_storage_access']:
                    scg['boot_connectivity'] = [_SSP_, _NPIV_]
                else:
                    scg['boot_connectivity'] = [_SSP_]
            elif scg['fc_storage_access']:
                scg['boot_connectivity'] = [_NPIV_]
            else:
                msg = _("The intended connectivity is ambiguous- please use"
                        " specific values for data connectivity and "
                        "boot connectivity")
                LOG.error(msg)
                raise wex.HTTPBadRequest(explanation=six.text_type(msg))

        if 'data_connectivity' not in scg or\
                not scg['data_connectivity']:

            if 'cluster_provider_name' in scg and\
                    scg['cluster_provider_name'] is not None:
                if scg['fc_storage_access']:
                    scg['data_connectivity'] = [_SSP_, _NPIV_]
                else:
                    scg['data_connectivity'] = [_SSP_]

            elif scg['fc_storage_access']:
                scg['data_connectivity'] = [_NPIV_]
            else:
                msg = _("The intended connectivity is ambiguous- please use"
                        " specific values for data connectivity and "
                        "boot connectivity")
                LOG.error(msg)
                raise wex.HTTPBadRequest(explanation=six.text_type(msg))

        if 'include_untagged' not in scg:
            scg['include_untagged'] = False  # default for new SCGs.
        # NOTE: the storage DOM controls the defaults for vios_redundancy,
        #       exact_redundancy, and npiv_fabric.
        scg = self._validate_common_post(scg, context)
        return scg

    def _validate_update(self, new_props, scg_dom, context, force):
        """
        Do validation checks on input for an SCG update operation.
        :param new_props: dictionary of property key/values to update.
        :param scg_dom: DOM of the existing SCG resource
        :param force: If true, then certain checks are skipped
        :param context: The authorization context
        :returns: The validated dictionary with items to update.
        """
        cur_scg_dict = scg_dom.to_dict()
        new_props = self._validate_common_pre(new_props, force)
        if not force:
            self._forcable_update_checks(new_props, scg_dom, context)

        ###################################################################
        # Update cur_scg dictionary with new props for 'post' validation  #
        ###################################################################
        cur_scg_dict.update(new_props)
        # Do unique validation/tweaking here.
        if 'npiv_fabric' not in new_props and\
                not cur_scg_dict['fc_storage_access'] and\
                'npiv_fabric' in cur_scg_dict:
            cur_scg_dict['npiv_fabric'] = new_props['npiv_fabric'] = None
        elif ('npiv_fabric' not in new_props and
              cur_scg_dict['fc_storage_access'] and
              'npiv_fabric' not in cur_scg_dict):
            cur_scg_dict['npiv_fabric'] = new_props['npiv_fabric'] = 'any'

        if (('fabric_set_req' not in new_props or
             'fabric_set_npiv' not in new_props) and
            not cur_scg_dict['fc_storage_access'] and
                'fabric_set_req' in cur_scg_dict):
            cur_scg_dict['ports_per_fabric_npiv'] = \
                new_props['ports_per_fabric_npiv'] = None
            cur_scg_dict['fabric_set_npiv'] = \
                new_props['fabric_set_npiv'] = None
            cur_scg_dict['fabric_set_req'] = new_props['fabric_set_req'] = None
        elif 'fabric_set_req' in new_props and \
            cur_scg_dict['fc_storage_access'] and 'npiv_fabric' in \
                cur_scg_dict and cur_scg_dict['npiv_fabric'] != 'any':
            cur_scg_dict['npiv_fabric'] = new_props['npiv_fabric'] = 'any'
            LOG.info("Updating fabric_set_req and forcing 'npiv_fabric' to "
                     "'any', because npiv_fabric is obsolete. Storage "
                     "connectivity groups update properties : %s" % new_props)
#        if (_NPIV_ in new_props['boot_connectivity'] or
#            _NPIV_ in new_props['data_connectivity']) and\
#                not cur_scg_dict['fc_storage_access']:
#            cur_scg_dict['fc_storage_access'] = \
#                new_props['fc_storage_access'] = True

        self._validate_common_post(cur_scg_dict, context)
        # We don't return the combined dictionary put in cur_scg_dict as that
        # is just used for validation. We still only update the fields
        # provided in new_props
        return new_props

    def _forcable_update_checks(self, new_props, scg_dom, context,
                                vm=None):
        """
        This validation method is used for two different update cases since
        it is useful to do the same checks for both updates, however the
        error messages are different for each case. If a 'vm' param is
        passed in, then the API request is to change the SCG set on a
        virtual machine. The messages in this case are less verbose since
        the PowerVC console does not support this path. The other case is
        when the 'vm' param is not passed, in which case we are validating
        the update of a specific SCG. That SCG may be associated with
        one or more existing virtual machines, and that is the case where the
        same sort of checks are done.

        :param new_props: dictionary of property key/values to update.
        :param scg_dom: DOM of the existing SCG resource
        :param context: The authorization context
        :param vm: The server instance object if the scg_dom passed is the
                   SCG for the instance that is being replaced/updated.
                   If None, then new_props are taken to be REST properties
                   requested to be updated on the scg_dom resource.
        """
        LOG.debug("Updating/replacing SCG '%s'. The new_props=%s" %
                  (scg_dom.display_name, new_props))
        if scg_dom.auto_defined and not vm:
            for key in new_props:
                if key not in ["enabled"]:
                    # We do not let the default SCG display names to be
                    # updated either here to avoid confusion when this
                    # failure is given.
                    msg = _("Cannot update '%(key)s' value of an auto-defined "
                            "storage connectivity group.") % {'key': key}
                    LOG.error(msg)
                    raise wex.HTTPBadRequest(explanation=msg)
                ########################################################
                # Else should we let them disable an auto-defined group
                # if there are no other groups that provide connectivity
                # to the storage accessible by this group?
                # It was decided the PowerVC console code does that check.
                # The API will not get into that level of complexity.
                #######################################################
        check_props = set(['vios_ids', 'cluster_provider_name',
                           'fc_storage_access', 'boot_connectivity',
                           'data_connectivity', 'vios_redundancy',
                           'exact_redundancy', 'npiv_fabric',
                           'fabric_set_req'])
        if any(item in check_props for item in new_props):
            #####################################################
            # Check if the SCG is in-use by any managed VMs.    #
            # If so, then there are additional restrictions.    #
            #####################################################
            msg_dict = {'messages': []}
            recovery = _(" The virtual machines must be removed first or you "
                         "may create a new storage connectivity group with "
                         "the desired properties.")

            if vm:
                vms = [vm]
            else:
                vms = self.stg_api.get_instances_for_scg(context,
                                                         scg_dom, msg_dict,
                                                         fast=True)
            vm_ids = [s['uuid'] for s in vms]
            msgs = [six.text_type(x) for x in msg_dict['messages']]
            vios_ids = scg_dom.vios_ids
            LOG.debug("scg_dom.vios_ids = %s, messages=%s, Associated "
                      "VMs=%s." % (vios_ids, msgs, vm_ids))
            if 'vios_ids' in new_props and vms:
                removing = [v for v in vios_ids if
                            v not in new_props['vios_ids']]
                if removing:
                    if vm:
                        # server-->scg association update
                        msg = _("Changing the storage connectivity group "
                                "associated with virtual machine '%(vm)s' "
                                "to one without at least the same set of "
                                "Virtual I/O Server members is not allowed.")\
                            % dict(vm=vm['name'])
                    else:
                        # scg property update
                        msg = _("Removing Virtual I/O Server (VIOS) members "
                                "from storage connectivity group "
                                "'%(scg)s' is not allowed because there are "
                                " existing virtual machines associated "
                                "with the group that may be using the VIOS "
                                "members for storage connectivity."
                                "%(recovery)s")\
                            % dict(scg=scg_dom.display_name,
                                   recovery=recovery)
                    # We could be less conservative in the VIOS removal
                    # check we did above - Are the VMs being removed on a host
                    # that one or more of the using VMs is on?, are attached
                    # disks to the VMs even using the VIOSs?, etc. The
                    # problem is there isn't a simple way to make these checks
                    # ironclad without introducing a complex locking structure
                    # into powervc because the VM may be migrating to a new
                    # host, or changing in some other way after the check is
                    # made that would invalidate the SCG props.
                    LOG.error(msg)
                    raise wex.HTTPConflict(explanation=six.text_type(msg))
            if 'vios_ids' in new_props and\
                    set(vios_ids) != set(new_props['vios_ids']):
                LOG.debug("The VIOS membership change is allowed. "
                          "New_list=%s." % new_props['vios_ids'])

            if 'cluster_provider_name' in new_props and vms and\
                    scg_dom.cluster_provider_name and\
                    new_props['cluster_provider_name'] !=\
                    scg_dom.cluster_provider_name:
                if vm:
                    # server-->scg association update
                    msg = _("Changing the storage connectivity group "
                            "associated with virtual machine '%(vm)s' "
                            "to one that does not specify the same shared "
                            "storage pool provider '%(provider)s' is not "
                            "allowed.") % dict(vm=vm['name'], provider=scg_dom
                                               .cluster_provider_name)
                else:
                    # scg property update
                    msg = _("Changing or removing the connectivity to shared "
                            "storage pool provider '%(provider)s' from storage"
                            " connectivity group '%(scg)s' is not allowed "
                            "because there are existing virtual "
                            "machines associated with the group."
                            "%(recovery)s") %\
                        dict(provider=scg_dom.cluster_provider_name,
                             scg=scg_dom.display_name, recovery=recovery)
                LOG.error(msg)
                raise wex.HTTPConflict(explanation=six.text_type(msg))

            if (('boot_connectivity' in new_props and
                 set(scg_dom.boot_connectivity) !=
                 set(new_props['boot_connectivity'])) or
                ('data_connectivity' in new_props and
                 set(scg_dom.data_connectivity) !=
                 set(new_props['data_connectivity']))) and\
                    vms:
                if vm:
                    msg = _("Changing the boot or data connectivity type "
                            "of the storage connectivity group associated"
                            " with virtual machine '%(vm)s' is not"
                            " allowed") % dict(vm=vm['name'])
                else:
                    msg = _("Updating the boot or data connectivity type "
                            "for storage connectivity group '%(scg)s' is not "
                            "allowed because there are "
                            "existing virtual machines "
                            "associated with the group."
                            "%(recovery)s") %\
                        dict(scg=scg_dom.display_name, recovery=recovery)
                LOG.error(msg)
                raise wex.HTTPConflict(explanation=six.text_type(msg))

            if 'fc_storage_access' in new_props and vms and\
                    scg_dom.fc_storage_access and not\
                    new_props['fc_storage_access']:
                if vm:
                    # server-->scg association update
                    msg = _("Changing the storage connectivity group "
                            "associated with virtual machine '%(vm)s' "
                            "to one that does not also have connectivity to "
                            "registered SAN storage providers is not "
                            "allowed.") % dict(vm=vm['name'])
                else:
                    # scg property update
                    msg = _("Updating the storage connectivity group '%(scg)s'"
                            " to remove connectivity to registered SAN "
                            "storage providers is not allowed because "
                            "there are existing virtual machines "
                            "associated with the group."
                            "%(recovery)s") %\
                        dict(scg=scg_dom.display_name, recovery=recovery)
                LOG.error(msg)
                raise wex.HTTPConflict(explanation=six.text_type(msg))
            if 'vios_redundancy' in new_props and vms and\
                    scg_dom.vios_redundancy != new_props['vios_redundancy']:
                if vm and (scg_dom.vios_redundancy <
                           new_props['vios_redundancy']):
                    # server-->scg association update
                    msg = _("Changing the storage connectivity group "
                            "associated with virtual machine '%(vm)s' to one "
                            "that has a larger VIOS redundancy requirement "
                            "is not allowed.") % dict(vm=vm['name'])
                else:
                    # scg property update
                    msg = _("Updating the storage connectivity group '%(scg)s'"
                            " with a different value for vios_redundancy is "
                            "not allowed because "
                            "there are existing virtual machines "
                            "associated with the group."
                            "%(recovery)s") %\
                        dict(scg=scg_dom.display_name, recovery=recovery)
                LOG.error(msg)
                raise wex.HTTPConflict(explanation=six.text_type(msg))
            if 'npiv_fabric' in new_props and vms and\
                    scg_dom.npiv_fabric != new_props['npiv_fabric']:
                if vm:
                    # server-->scg association update
                    msg = _("Changing the storage connectivity group "
                            "associated with virtual machine '%(vm)s' to one "
                            "that has a different npiv_fabric requirement "
                            "is not allowed.") % dict(vm=vm['name'])
                else:
                    # scg property update
                    msg = _("Updating the storage connectivity group '%(scg)s'"
                            " with a different value for npiv_fabric is "
                            "not allowed because "
                            "there are existing virtual machines "
                            "associated with the group."
                            "%(recovery)s") %\
                        dict(scg=scg_dom.display_name, recovery=recovery)
                LOG.error(msg)
                raise wex.HTTPConflict(explanation=six.text_type(msg))

            if (('fabric_set_req ' in new_props or
                 'fabric_set_npiv' in new_props or
                 'ports_per_fabric_npiv' in new_props) and vms and
                (scg_dom.fabric_set_req != new_props['fabric_set_req'] or
                 set(scg_dom.fabric_set_npiv) !=
                 set(new_props['fabric_set_npiv']) or
                 scg_dom.ports_per_fabric_npiv !=
                 new_props['ports_per_fabric_npiv'])):
                if vm:
                    # server-->scg association update
                    msg = _("Changing the storage connectivity group "
                            "associated with virtual machine '%(vm)s' to one "
                            "that has a different fabric_set_req or "
                            "fabric_set_npiv or ports_per_fabric_npiv "
                            "requirement is not allowed.") \
                        % dict(vm=vm['name'])
                else:
                    # scg property update
                    msg = _("Updating the storage connectivity group '%(scg)s'"
                            " with a different value for fabric_set_req or "
                            "fabric_set_npiv or ports_per_fabric_npiv is "
                            "not allowed because "
                            "there are existing virtual machines "
                            "associated with the group."
                            "%(recovery)s") %\
                        dict(scg=scg_dom.display_name, recovery=recovery)
                LOG.error(msg)
                raise wex.HTTPConflict(explanation=six.text_type(msg))

    def _validate_common_pre(self, scg, force):
        """
        Do validation checks prior to operation for create/update
        :param scg: dictionary of the SCG resource to check
        :returns: The validated dictionary.
        """
        if not isinstance(scg, dict):
            msg = _("The request body is not a dictionary of "
                    "property/value pairs")
            LOG.error(msg)
            raise wex.HTTPBadRequest(explanation=six.text_type(msg))
        if len(scg) == 0:
            msg = _("The body of the request is empty.")
            LOG.error(msg)
            raise wex.HTTPBadRequest(explanation=six.text_type(msg))

        for k in scg:
            if k not in EDITABLE_KEYS.keys() and k not in NON_EDITABLE_KEYS:
                msg = _("The specified property name '%s' is not supported "
                        "for a storage connectivity group.") % (k)
                LOG.error(msg)
                raise wex.HTTPBadRequest(explanation=six.text_type(msg))
            elif k not in EDITABLE_KEYS.keys():
                msg = _("The property name '%s' is not editable. "
                        "It cannot be specified on a storage connectivity "
                        "group create or update operation.") % (k)
                LOG.error(msg)
                raise wex.HTTPBadRequest(explanation=six.text_type(msg))

            # Now check the input property types
            key_type = EDITABLE_KEYS[k]
            is_str = (isinstance(scg[k], str) or isinstance(scg[k], unicode))
            if key_type == 'bool' and not isinstance(scg[k], bool):
                if not is_str or not (scg[k].lower() in ['true', 'false']):
                    msg = _("The property name '%s' must specify a boolean "
                            "value.") % (k)
                    LOG.error(msg)
                    raise wex.HTTPBadRequest(explanation=six.text_type(msg))
                # Do conversion of string to boolean.
                scg[k] = (scg[k].lower() == 'true')
            elif key_type == 'list' and not isinstance(scg[k], list):
                msg = _("The property name '%s' must specify a list for the "
                        "value.") % (k)
                LOG.error(msg)
                raise wex.HTTPBadRequest(explanation=six.text_type(msg))
            elif (key_type == 'string' and not is_str):
                if (k == 'port_tag' or k == 'cluster_provider_name' or
                    k == 'npiv_fabric' or k == 'fabric_set_req') and \
                        scg[k] is None:
                    LOG.debug("None type allowed for port_tag or "
                              "cluster_provider_name values.")
                else:
                    msg = _("The property name '%(prop)s' must specify a "
                            "string for the value. The value given was: "
                            "%(value)s") % dict(prop=k, value=str(scg[k]))
                    LOG.error(msg)
                    raise wex.HTTPBadRequest(explanation=six.text_type(msg))
            elif (key_type == 'integer' and (not isinstance(scg[k], int) or
                                             isinstance(scg[k], bool))):
                # We have to do the bool checks as well because booleans also
                # evaluate to integers in python. i.e. True + True == 2
                if not isinstance(scg[k], bool) and scg[k].isdigit():
                    # We can convert a numeric string to its numeric value.
                    scg[k] = int(scg[k])
                else:
                    msg = _("The property name '%s' must specify an integer "
                            "for a value.") % (k)
                    LOG.error(msg)
                    raise wex.HTTPBadRequest(explanation=six.text_type(msg))
            if key_type == 'list':
                # make lists have min uniqe values.
                scg[k] = list(set(scg[k]))

        if 'display_name' in scg and len(scg['display_name'].strip()) == 0:
            msg = _("The name of the SCG being created or modified must not "
                    "be blank.")
            LOG.error(msg)
            raise wex.HTTPBadRequest(explanation=six.text_type(msg))
        elif 'display_name' in scg and\
                scg['display_name'].startswith('Any host in '):
            msg = _("The name of the SCG being created or modified must not "
                    "begin with the reserved string 'Any host in '.")
            LOG.error(msg)
            raise wex.HTTPBadRequest(explanation=msg)
        if 'vios_ids' in scg:
            if len(scg['vios_ids']) == 0 and not force:
                msg = _("At least one Virtual I/O Server must be specified "
                        "on a storage connectivity group create request. If "
                        "updating a storage connectivity group, then all "
                        "Virtual I/O Server members cannot "
                        "be removed from membership.")
                LOG.error(msg)
                raise wex.HTTPBadRequest(explanation=six.text_type(msg))
            if [x for x in scg['vios_ids'] if not isinstance(x, str) and
                    not isinstance(x, unicode)]:
                msg = _("Each vios_ids list element must be a string.")
                LOG.error(msg)
                raise wex.HTTPBadRequest(explanation=six.text_type(msg))
        fabric_vals = StorageConnectivityGroupFactory.ALOWED_NPIV_FABRIC_VALS
        if 'npiv_fabric' in scg and scg['npiv_fabric'] not in fabric_vals:
            msg = _("The 'npiv_fabric' property is restricted to one of the "
                    "the following values: %s") % fabric_vals
            LOG.error(msg)
            raise wex.HTTPBadRequest(explanation=six.text_type(msg))
        fab_set_vals = StorageConnectivityGroupFactory.ALLOWED_FABRIC_SET_REQ
        if ('fabric_set_req' in scg and
                scg['fabric_set_req'] not in fab_set_vals):
            msg = _("The 'fabric_set_req' property is restricted to one of "
                    "the the following values: %s") % fab_set_vals
            LOG.error(msg)
            raise wex.HTTPBadRequest(explanation=six.text_type(msg))
        if ('fabric_set_req' in scg and 'npiv_fabric' in scg):
            msg = _("Creation or Update of a storage connectivity group with "
                    "both 'fabric_set_req' and 'npiv_fabric' is not valid, "
                    "please specify one or the other")
            LOG.error(msg)
            raise wex.HTTPBadRequest(explanation=six.text_type(msg))
        if 'ports_per_fabric_npiv' in scg:
            if scg['ports_per_fabric_npiv'] != 1 and 'npiv_fabric' in scg:
                msg = _("The 'ports_per_fabric_npiv' property is restricted "
                        "to value 1 if npiv_fabric is specified.")
                raise wex.HTTPBadRequest(explanation=six.text_type(msg))
            elif scg['ports_per_fabric_npiv'] < 1 or\
                    scg['ports_per_fabric_npiv'] > 64:
                msg = _("The 'ports_per_fabric_npiv' property is restricted "
                        "to values between 1 and 64.")
                raise wex.HTTPBadRequest(explanation=six.text_type(msg))
        if 'vios_redundancy' in scg and scg['vios_redundancy'] < 1:
            msg = _("The 'vios_redundancy' property value must be an integer "
                    "greater than or equal to 1.")
            LOG.error(msg)
            raise wex.HTTPBadRequest(explanation=six.text_type(msg))
        # convert any new scg creation/update requests using 'npiv_fabric' to
        # the new fabric_set values.
        if 'npiv_fabric' in scg and scg['npiv_fabric'] is not None:
            if scg['npiv_fabric'] == 'A':
                scg['fabric_set_npiv'] = ['A']
                scg['fabric_set_req'] = 'at_least_one'
            elif scg['npiv_fabric'] == 'B':
                scg['fabric_set_npiv'] = ['B']
                scg['fabric_set_req'] = 'at_least_one'
            elif scg['npiv_fabric'] == 'single':
                scg['fabric_set_npiv'] = ['A', 'B']
                scg['fabric_set_req'] = 'exactly_one'
            elif scg['npiv_fabric'] == 'dual':
                scg['fabric_set_npiv'] = ['A', 'B']
                scg['fabric_set_req'] = 'per_host'
            elif scg['npiv_fabric'] == 'dual_per_vios':
                scg['fabric_set_npiv'] = ['A', 'B']
                scg['fabric_set_req'] = 'per_vios'
            else:
                scg['fabric_set_npiv'] = ['A', 'B']
                scg['fabric_set_req'] = 'at_least_one'
            scg['npiv_fabric'] = None
            msg = _("The 'npiv_fabric' property is deprecated and will no "
                    "longer be supported in the future. The new settings:"
                    "'fabric_set_req':%(req)s, 'fabric_set_npiv':"
                    "%(set_npiv)s, will be utilized instead") % \
                dict(req=scg['fabric_set_req'],
                     set_npiv=scg['fabric_set_npiv'])
            LOG.warn(msg)

        if 'fabric_set_npiv' in scg:
            fk = 'fabric_set_npiv'
            scg[fk] = [x.upper() for x in scg[fk]]
            if '*' in scg[fk] and len(scg[fk]) > 1:
                msg = _("The list value for 'fabric_set_npiv' cannot contain "
                        "fabrics in addition to the wildcard '*'.")
                raise wex.HTTPBadRequest(explanation=six.text_type(msg))
            allowed_fabs = [chr(ord('A') + x) for x in range(MAX_FABS)]
            if '*' not in scg[fk] and\
                    not set(scg[fk]).issubset(set(allowed_fabs)):
                LOG.debug("allowed_fabs=%s, passed fab_set=%s" %
                          (allowed_fabs, scg[fk]))
                msg = _("'fabric_set_npiv' list value can only contain "
                        "fabric names in the range of 'A' to '%(last)s'.")
                raise wex.HTTPBadRequest(explanation=six.text_type(
                    msg % {'last': allowed_fabs[-1:][0]}))

        conn_vals = [_NPIV_, _SSP_, _VSCSI_]
        if (('boot_connectivity' in scg and
             scg['boot_connectivity'] is not None and
             [x for x in scg['boot_connectivity'] if x not in conn_vals]) or
            ('data_connectivity' in scg and
             scg['data_connectivity'] is not None and
             [y for y in scg['data_connectivity'] if y not in conn_vals])):
            msg = _("Connectivity values can only be one of: %s") % conn_vals
            LOG.error(msg)
            raise wex.HTTPBadRequest(explanation=six.text_type(msg))

        return scg

    def _validate_common_post(self, scg, context):
        """
        Do validation checks after filling defaults for create/update
        :param scg: dictionary of the SCG resource to check
        :returns: The validated dictionary.
        """
        if (('cluster_provider_name' not in scg) and
                ("vios_cluster" not in scg)):
            if _SSP_ in scg['data_connectivity'] or\
                    _SSP_ in scg['boot_connectivity']:
                msg = _("No cluster details given for SSP connectivity.")
                LOG.error(msg)
                raise wex.HTTPBadRequest(explanation=six.text_type(msg))
        elif 'cluster_provider_name' in scg:
            # Check that the cluster reference is valid
            cin = pvc_cinder.PowerVCVolumeAPI()
            try:
                prov = cin.get_stg_provider_info(
                    context,
                    scg['cluster_provider_name'])
            except:
                prov = None
                LOG.warn(_("Could not retrieve provider data."))
            if prov is None:
                msg = _("The Virtual I/O Server cluster storage provider "
                        "name specified '%(provider)s' on the "
                        "storage connectivity group request was not found.")\
                    % dict(provider=scg['cluster_provider_name'])
                LOG.error(msg)
                raise wex.HTTPNotFound(explanation=six.text_type(msg))

        if ((_NPIV_ not in scg['data_connectivity'] and
             _NPIV_ not in scg['boot_connectivity']) and
            scg['fc_storage_access']) or\
            ((_NPIV_ in scg['boot_connectivity'] or
              _NPIV_ in scg['data_connectivity']) and
             not scg['fc_storage_access']):
            msg = _("fc_storage_access value is not consistent with "
                    "either boot_connectivity or data_connectivity. "
                    "fc_storage_access must be true if 'npiv' is "
                    "specified for boot or data connectivity. Otherwise it "
                    "must be false. fc_storage_access is deprecated.")
            LOG.error(msg)
            raise wex.HTTPConflict(explanation=six.text_type(msg))

        if not scg['fc_storage_access'] and 'npiv_fabric' in scg and\
                scg['npiv_fabric'] is not None:
            msg = _("The 'npiv_fabric' property cannot be specified if the "
                    "storage connectivity group does not allow NPIV "
                    "connectivity.")
            LOG.error(msg)
            raise wex.HTTPBadRequest(explanation=six.text_type(msg))

        if not scg['fc_storage_access'] and 'fabric_set_req' in scg and\
                scg['fabric_set_req'] is not None:
            msg = _("The 'fabric_set_req' property cannot be specified if the "
                    "storage connectivity group does not allow NPIV "
                    "connectivity.")
            LOG.error(msg)
            raise wex.HTTPBadRequest(explanation=six.text_type(msg))
        if not scg['fc_storage_access'] and 'ports_per_fabric_npiv' in scg and\
                scg['ports_per_fabric_npiv'] is not None:
            msg = _("The 'ports_per_fabric_npiv' property cannot be specified "
                    "if the storage connectivity group does not allow NPIV "
                    "connectivity.")
            LOG.error(msg)
            raise wex.HTTPBadRequest(explanation=six.text_type(msg))

        if ((_NPIV_ in scg['boot_connectivity'] and
             _VSCSI_ in scg['boot_connectivity']) or
            (_NPIV_ in scg['data_connectivity'] and
             _VSCSI_ in scg['data_connectivity'])):
            msg = _("Only one of npiv and pv-vscsi allowed")
            LOG.error(msg)
            raise wex.HTTPBadRequest(explanation=six.text_type(msg))

        return scg

    def _set_priority_lists(self, priority_dict, context):
        """
        This method takes the dictionary of one or more priority lists
        and persists the ordering.
        """
        if not isinstance(priority_dict, dict):
            msg = _("The request body does not contain a dictionary.")
            LOG.error(msg)
            raise wex.HTTPBadRequest(explanation=six.text_type(msg))
        # We want to keep all the priority setting under a single transaction
        scg_factory = StorageConnectivityGroupFactory.get_factory()
        txn = scg_factory.get_resource_transaction()
        priority = 0
        with txn.begin():
            for key, plist in priority_dict.iteritems():
                if not isinstance(plist, list):
                    msg = _("The request body does not contain a priority "
                            "list to set.")
                    LOG.error(msg)
                    raise wex.HTTPBadRequest(explanation=six.text_type(msg))
                ####################
                # DB facing impl
                ####################
                priority = 1
                for scg_id in plist:
                    self._set_priority(scg_id, priority, key, context, txn)
                    priority = priority + 1
        LOG.debug("**EXIT** last priority=%d." % priority)

    def _set_priority(self, scg_id, priority, key, context, transaction):
        """
        Set the priority value for a particular SCG.
        """
        scg_factory = StorageConnectivityGroupFactory.get_factory()
        scg = scg_factory.find_scg_by_id(context, scg_id,
                                         transaction=transaction)
        if scg is None:
            msg = stgex.IBMPowerVCSCGNotFound.msg_fmt % locals()
            LOG.exception(msg)
            raise wex.HTTPNotFound(explanation=six.text_type(msg))
        if key == "external":
            LOG.debug("Setting %s external priority from '%s' to '%d'."
                      % (scg_id, str(scg.priority_external), priority))
            scg.priority_external = priority
        else:
            LOG.debug("Setting %s cluster priority from '%s' to '%d'."
                      % (scg_id, str(scg.priority_cluster), priority))
            if key != scg.cluster_provider_name:
                msg = _("Storage connectivity group with ID '%(scg_id)s' is "
                        "not associated with shared storage pool cluster "
                        "'%(provider)s'. Setting cluster priority "
                        "failed.") % dict(scg_id=scg_id, provider=key)
                LOG.exception(msg)
                raise wex.HTTPBadRequest(explanation=six.text_type(msg))
            scg.priority_cluster = priority
        # Queue up the commit within the transaction
        scg.save(context=context, transaction=transaction)

    def _lookup_scgs(self, context, req, details=False, quickdetail=False):
        """
        This method validates REST URI options and then retrieves
        SCGs from the database.
        """
        sort_type = None
        if 'sort_type' in req.GET:
            sort_type = req.GET['sort_type']
            if sort_type != "external" and sort_type != "cluster":
                msg = _("Unsupported sort_type option value passed: "
                        "%(sort_type)s. Valid sort_type values are: "
                        "external, cluster.") % (locals())
                LOG.error(msg)
                raise wex.HTTPBadRequest(explanation=six.text_type(msg))
        enabled_only = False
        if 'enabled_only' in req.GET:
            if req.GET['enabled_only'].lower() == "true":
                enabled_only = True
            elif req.GET['enabled_only'].lower() == "false":
                enabled_only = False
            else:
                msg = _("The value given for the 'enabled_only' option did "
                        "not parse to 'true' or 'false'.")
                LOG.error(msg)
                raise wex.HTTPBadRequest(explanation=six.text_type(msg))
        cluster_provider = None
        if 'cluster_provider' in req.GET:
            cluster_provider = req.GET['cluster_provider']
            if sort_type == "external":
                msg = _("'cluster_provider=' option is not supported when "
                        "sort_type is external.")
                LOG.error(msg)
                raise wex.HTTPBadRequest(explanation=six.text_type(msg))
        include_ports = False
        if 'include_ports' in req.GET and not quickdetail:
            if req.GET['include_ports'].lower() == "true":
                include_ports = True
            elif req.GET['include_ports'].lower() == "false":
                include_ports = False
            else:
                msg = _("The value given for the 'include_ports' option did "
                        "not parse to 'true' or 'false'.")
                LOG.error(msg)
                raise wex.HTTPBadRequest(explanation=six.text_type(msg))
        deploy_ready_only = False
        if 'deploy_ready_only' in req.GET and not quickdetail:
            if req.GET['deploy_ready_only'].lower() == "true":
                deploy_ready_only = True
            elif req.GET['deploy_ready_only'].lower() == "false":
                deploy_ready_only = False
            else:
                msg = _("The value given for the 'deploy_ready_only' option "
                        "did not parse to 'true' or 'false'.")
                LOG.error(msg)
                raise wex.HTTPBadRequest(explanation=six.text_type(msg))

        LOG.debug("*SCG* details=%(details)s, sort_type=%(sort_type)s, "
                  "enabled_only=%(enabled_only)s, cluster_provider=%(cluster"
                  "_provider)s, include_ports=%(include_ports)s, "
                  "deploy_ready_only=%(deploy_ready_only)s" % locals())
        scg_list = su.query_scgs(context, sort_type=sort_type,
                                 enabled_only=enabled_only,
                                 cluster_provider=cluster_provider,
                                 include_ports=include_ports,
                                 deploy_ready_only=deploy_ready_only,
                                 quickdetail=quickdetail)
        if details:
            for scg in scg_list:
                for h in scg.get('host_list', []):
                    for v in h.get('vios_list', []):
                        if 'backend_uuid' in v:
                            del v['backend_uuid']
            return scg_list

        # Now just build the short (index) list (ignore other properties)
        index_list = map(lambda x: {'id': x['id'],
                                    DISPLAY_NAME: x[DISPLAY_NAME]}, scg_list)
        return index_list

    def _set_scg_to_server(self, context, scg, instance, force):
        """
        Helper method to store the passed SCG onto the passed VM's
        instance's metadata. If force is not True, then checks are first
        done for SCG update rules.
        """
        msgd = {'messages': []}
        if not force:
            # Do checks
            cur_scg = self.stg_api.get_scg_from_instance_meta(context,
                                                              instance, msgd)
            if cur_scg:
                # Share the same forcable update checks as for updating an
                # SCG. To prepare for the helper call, we generate the
                # new_props dictionary from the new SCG to set, and then
                # replace the vios_cluster property dictionary with vios_ids.
                new_props = scg.to_dict(context=context,
                                        skip_extended_lookups=True)
                self._forcable_update_checks(new_props, cur_scg, context,
                                             instance)
                # At some point it would be nice to only restrict the SCG
                # setting if the VM was actually using the VIOS or storage
                # being lost, but we don't do that today because of the
                # complexity.
        self.stg_api.set_scg_to_instance(context, instance, scg, msgd)
        LOG.debug("Messages = %s." % msgd)


class Storage_connectivity_groups(extensions.ExtensionDescriptor):

    """ Support to manage Storage Connectivity Groups. """
    name = "Compute Host Storage Connectivity Groups"
    alias = ALIAS
    # The namespace just follows the convention of other APIs - is there
    # separate work that needs to be done to register this?
    namespace = ("http://docs.openstack.org/compute/ext/ibm_" +
                 "storage_connectivity_groups/api/v1.1")
    updated = "2013-07-25T11:00:00-06:00"

    def get_resources(self):
        """ Provide a Resource Extension for storage-connectivity-groups """
        return [extensions.ResourceExtension(
            ALIAS,
            controller=StorageConnGroupController(),
            collection_actions={'update': 'PUT', 'detail': 'GET',
                                'quickdetail': 'GET'})]


class Storage_connectivity_groupsV21(extensions.V21APIExtensionBase):

    """ Support to manage Storage Connectivity Groups. """
    name = "Compute Host Storage Connectivity Groups"
    alias = ALIAS
    version = 1

    def get_resources(self):
        """ Provide a Resource Extension for storage-connectivity-groups """
        return [extensions.ResourceExtension(
            ALIAS,
            controller=StorageConnGroupController(),
            collection_actions={'update': 'PUT', 'detail': 'GET',
                                'quickdetail': 'GET'})]

    def get_controller_extensions(self):
        return []
