# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import httplib
import json
import six
import types
from ast import literal_eval
from webob import Response
from webob import exc

from nova.api.openstack import extensions
from nova.api.openstack import wsgi
from nova import db
from nova import compute
from nova.image import glance
from nova import exception
from oslo_log import log as logging
from nova.ibm.ego import ego_constants as op_types
from powervc_nova import _
from powervc_nova.storage import storage_mapping_api, kvm_storage_mapping
from powervc_nova.storage import storage_utils as su
from powervc_nova.volume import cinder as pvc_cinder


# ------------------------------------
# Extension: deploy-hosts
# ------------------------------------
LOG = logging.getLogger(__name__)
ALIAS = 'storage-viable-hosts'
AUTHORIZE_V21 = extensions.os_compute_authorizer(ALIAS)
AUTHORIZE_V2 = extensions.extension_authorizer('compute',
                                               "storage_viable_hosts")


class StorageViableHostsController(wsgi.Controller):
    """
    This Controller handles the API entry implementation of the REST
    function to retrieve the applicable deploy-host targets and
    relocation-host targets for either a given existing instance or the
    flavor and image to be used for a future instance.
    This is a multi-purpose API designed for PowerVC and not intended to be
    driven into OpenStack at some point, and there is a corollary API
    for network-based filtering. See the PowerVC scheduler team for details
    on this approach. The following URI options are recognized:

        instance_uuid=<uuid>
        server_id=<server_id>
        imageRef=<image_id>
        flavorRef=<flavor_id>

    See the index() list method for details on these options.
    """

    def __init__(self):
        self.image_service = glance.get_default_image_service()
        # Set the storage api class according to the host type
        if su.is_kvm_supported():
            self.stg_api = kvm_storage_mapping.KVMStorageMapping()
        else:
            self.stg_api = storage_mapping_api.StorageMapping()

    def _get_image(self, context, image_id):
        try:
            return self.image_service.show(context, image_id)
        except exception.NotFound:
            raise exception.ImageNotFound(image_id=image_id)

    def index(self, req):
        """
        REST API entry point to call internal API to do work. At least
        a 'viewer' role is required to be authorized. Here are the URI
        query options that are recognized:

        instance_uuid=<uuid>: The instance UUID that will be used to
                        get the storage requirements of an instance.
                        Providing this option will result in a database
                        lookup of the instance.
        server_id=<server_id> Similar to the instance_uuid option except that
                        the value will be looked up through the compute API
                        rather than through the database.
        flavorRef=<flavor_id>: Optional flavor object that can be used to
                        get the SCG to do the filtering.
                        NOT currently supported for KVM hosts.
        imageRef=<image_id>: Optional image. We can get to the Volume type
                        from the image to filter based on the provider if
                        needed. The size of the image can be used to filter
                        out all hosts if the target storage does not have
                        enough room for deployment of the image.
                        NOT currently supported for KVM hosts.
        volume_typeRef=<{vol_id:'vol_type_id'}>: Optional volume type. This
                        overrides any volume type on either the passed
                        image or instance. If the volume type specifies a
                        storage provider that is in error state, then all
                        hosts will be filtered out (for example).
                        NOT currently supported for KVM hosts.
        messages=<any-val>: Optional. If passed, then info and warning
                        messages that are acrrued during API processing
                        will be returned in the output json dictionary
                        under the key 'messages'.
        operation_type=<EGO-defined-ops>: Optional. Defaults to a 'create',
                        which is a deploy. Other valid values are:
                        live-migration, cold-migration, resize, evacuate,
                        unshelve.

        Returns: A dictionary of child host dictionaries keyed by
                 host (registered name):
                    {
                     "8231E2D_222EFET": {
                      "host": "8231E2D_222EFET"
                     },
                     "8231E2D_222EFCT": {
                      "host": "8231E2D_222EFCT"
                     }
                    }

                This format matches that of the network scheduler filter
                for powervc.
        """
        context = req.environ['nova.context']
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action='list')
        else:
            AUTHORIZE_V2(context, action='list')
        LOG.debug("** ENTRY ** context=%s" % str(context))
        su.http_error_if_ivm()  # HMC-managed installs only

        # initialize options
        instance_uuid = None
        imageRef = None
        flavorRef = None
        vol_type_id = None
        instance = None
        image = None
        flavor = None
        vol_type_map = {}
        # for backward compatibility
        old_vol_type = None
        operation_type = None
        try:
            if 'instance_uuid' in req.GET:
                instance_uuid = req.GET['instance_uuid']
                instance = db.instance_get_by_uuid(context, instance_uuid)
            elif 'server_id' in req.GET:
                compute_api = compute.API()
                instance = compute_api.get(context, req.GET['server_id'])
                instance_uuid = instance['uuid']
            if 'flavorRef' in req.GET:
                flavorRef = req.GET['flavorRef']
                flavor = db.flavor_get_by_flavor_id(context, flavorRef)
            if 'imageRef' in req.GET:
                imageRef = req.GET['imageRef']
                image = self._get_image(context, imageRef)
            if 'operation_type' in req.GET:
                operation_type = req.GET['operation_type']
            if 'volume_typeRef' in req.GET:
                voltype_ref = req.GET['volume_typeRef']
                if isinstance(voltype_ref, basestring):
                    try:
                        voltype_ref = literal_eval(voltype_ref)
                    except (ValueError, SyntaxError):
                        # If cannot convert use the value as it is
                        pass

                if isinstance(voltype_ref, types.DictType):
                    for vol in voltype_ref:
                        vol_type = pvc_cinder.PowerVCVolumeAPI().\
                            get_volume_type(context, voltype_ref[vol])
                        if vol_type:
                            vol_type_map[vol] = vol_type
                else:
                    old_vol_type = pvc_cinder.PowerVCVolumeAPI().\
                        get_volume_type(context, voltype_ref)

        except Exception as e:
            msg = _("Unable to find resource given as an option. "
                    "Error is: %s" % six.text_type(e))
            LOG.exception(e)
            raise exc.HTTPNotFound(explanation=six.text_type(msg))

        LOG.debug("Options: instance_uuid=%s, flavorRef=%s, imageRef=%s, "
                  "vol_typeRef=%s, operation_type=%s."
                  % (instance_uuid, flavorRef, imageRef, vol_type_id,
                     operation_type))
        if operation_type and \
                operation_type not in op_types.FILTER_WORKING_SCOPE_ALL:
            msg = _("Not a valid operation type: %(op_type)s.")
            msg = msg % {'op_type': operation_type}
            raise exc.HTTPServerError(explanation=six.text_type(msg))

        img_name = 'n/a' if not image else image['name']
        LOG.debug("After lookups: instance=%s, flavor=%s, image_name=%s, "
                  "vol_type_map=%s, old_vol_type=%s"
                  % (str(instance), flavor, img_name, str(vol_type_map),
                     old_vol_type))

        msg_dict = {'messages': []}
        # Call the internal nova API
        LOG.debug("Calling get_viable_hosts()...")
        try:
            # Assume stg_api has been set in the constructor to the correct
            # host_type target.
            input_vol_type = vol_type_map if vol_type_map else old_vol_type
            data = self.stg_api.get_viable_hosts(context, instance=instance,
                                                 flavor=flavor, image=image,
                                                 vol_type=input_vol_type,
                                                 msg_dict=msg_dict,
                                                 operation_type=operation_type)
        except Exception as ex:
            LOG.exception(ex)
            raise exc.HTTPServerError(explanation=six.text_type(ex))

        if msg_dict['messages']:
            # Convert Message types or unicode strings to plain strings
            msgs = [six.text_type(x) for x in msg_dict['messages']]
            # The storage-viable-hosts API return does not lend itself to
            # adding messages in the response, so just log them here as
            # debug since the individual messages should have been logged as
            # INFO during the flow.
            LOG.debug("Accumulated messages during API processing: %s" % msgs)
            if 'messages' in req.GET:  # can add to response with special flag
                data['messages'] = msgs
        # Return results in JSON body response
        return Response(request=req,
                        status=httplib.OK,
                        content_type='application/json',
                        body=json.dumps(data))


class Storage_viable_hosts(extensions.ExtensionDescriptor):
    """
        Support for REST API(s) returning lists of target hosts
        constrained by storage related to the input options.
    """
    name = "Compute Viable Deploy Hosts by Storage"
    alias = ALIAS
    # The namespace just follows the convention of other APIs - is there
    # separate work that needs to be done to register this?
    namespace = ("http://docs.openstack.org/compute/ext/storage/"
                 "storage_viable_hosts/api/v1.1")
    updated = "2014-02-03T11:00:00-06:00"

    def get_resources(self):
        """
        Provide a Resource Extension for getting candidate deploy hosts
        provided one or more options..
        """
        return [
            extensions.ResourceExtension(
                ALIAS,
                controller=StorageViableHostsController())]


class Storage_viable_hostsV21(extensions.V21APIExtensionBase):
    """
        Support for REST API(s) returning lists of target hosts
        constrained by storage related to the input options.
    """
    name = "Compute Viable Deploy Hosts by Storage"
    alias = ALIAS
    version = 1

    def get_resources(self):
        """
        Provide a Resource Extension for getting candidate deploy hosts
        provided one or more options..
        """
        return [extensions.ResourceExtension(ALIAS,
                controller=StorageViableHostsController())]

    def get_controller_extensions(self):
        return []
