# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import os
import six
import sys
import time
import errno
import greenlet
import traceback
from eventlet import greenthread
from fasteners import process_lock

from oslo_concurrency import lockutils
from oslo_log import log as logging

from nova import exception
from oslo_service import eventlet_backdoor as backdoor

from powervc_nova import _

if os.name == 'posix':
    import fcntl
else:
    # This leg is only for UT runs on Windows
    import mock
    sys.modules['fcntl'] = mock.MagicMock()

LOG = logging.getLogger(__name__)
CANARY_RUN = None


######################################################
#               Dump Thread Stacks                   #
######################################################
def dump_thread_stacks():
    for i, t in enumerate(backdoor._find_objects(greenlet.greenlet)):
        LOG.error('Greenthread %s %s' % (i, t))
        LOG.error(traceback.format_stack(t.gr_frame))


######################################################
#             Periodic Task Watchdog                 #
######################################################
def periodic_task_watchdog():
    """
    This works by inserting a "canary" periodic task into the list of tasks
    to run, which records a timestamp each time it runs.  Then this method
    is run as a regular (non periodic) task to checkup on the canary and
    see if it's still alive.  If it hasn't reported a timestamp since the
    timeout then we know the periodic tasks are hung or starved.
    """
    LOG.info(_('Starting periodic task watch dog.'))
    delay_interval = 600
    last_print = None
    last_check = time.time()
    global CANARY_RUN
    while True:
        try:
            greenthread.sleep(delay_interval)
            curr_time = time.time()
            # Check if it appears the time was changed and if so
            # just skip this check and we'll get it next interval
            if curr_time - last_check > 2 * delay_interval:
                continue
            # Don't do anything unless the canary has run at least once
            if CANARY_RUN:
                # Check if the canary ran since the last time we printed
                # out a message saying they stopped and if so, they must
                # have resumed.
                if last_print and last_print != CANARY_RUN:
                    last_print = None
                    LOG.info(_('Detected periodic tasks have resumed.'))

                # if canary has run but earlier than our interval
                if (last_print != CANARY_RUN and
                        (CANARY_RUN + delay_interval < curr_time)):
                    # Seems the periodic tasks are stuck, dump thread stacks
                    LOG.error(_('Detected periodic tasks have stopped'
                                ' running.'))
                    last_print = CANARY_RUN
                    dump_thread_stacks()
        except Exception as ex:
            LOG.exception(six.text_type(ex))
        finally:
            last_check = time.time()


def update_canary_run_time():
    """Helper method to update the last Canary Run time"""
    global CANARY_RUN
    CANARY_RUN = time.time()


######################################################
#             Monkey Patch File Lock                 #
######################################################
class Monkey_FcntlLock(process_lock._FcntlLock):
    """
    Monkey patch file locking because it does spin locks
    and gives no indication of troubles.  The result is
    that periodic tasks can hang, if NFS lockd continues
    """
    def __init__(self, name):
        super(Monkey_FcntlLock, self).__init__(name)
        self._max_retries = 4096
        self._max_time = 3600
        self._msg_interval = 2048

    def trylock(self):
        tries = 0
        start_time = time.time()
        err_msg = 'Unable to obtain lock for file: %s'
        while True:
            try:
                fcntl.lockf(self.lockfile, fcntl.LOCK_EX | fcntl.LOCK_NB)
                return self
            except IOError as e:
                tries += 1
                if e.errno in (errno.EACCES, errno.EAGAIN):
                    # Check if we've been trying for quite a while.  Use
                    # both time and tries, in order to avoid time change
                    # issues and yet give it enough tries.
                    if (time.time() - start_time > self._max_time and
                            tries > self._max_retries):
                        LOG.error(_(err_msg) % self.path)
                        raise exception.NovaException(message=(_(err_msg)
                                                               % self.path))
                    # If we think we might have an issue, send a warning
                    if tries % self._msg_interval == 0:
                        LOG.warn(_(err_msg) % self.path)
                    time.sleep(0.05)
                else:
                    raise


class EnhancedDict(dict):
    """Extends dictionary to handle dot-notation also"""

    def __getattr__(self, attr):
        return self.get(attr)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)
