# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Host aggregate-related utilities."""

import six

from oslo_config import cfg
from oslo_concurrency import lockutils
from oslo_log import log as logging

from nova.compute import api as compute_api
from nova import context as nova_context
from nova import exception
from nova import objects

from nova.ibm.ego.hapolicy import constants as ego_ha_const
from nova.ibm.ego import ego_constants as ego_const

from powervc_nova.prs import constants as pvc_prs_const
from powervc_nova import _

cfg.CONF.register_opts([
    cfg.StrOpt('default_host_aggregate_name',
               default='Default Group',
               help='Name of the PowerVC default host aggregate'),
    cfg.IntOpt('reconcile_host_aggregates_interval', default=3600,
               help='The interval (in seconds) at which to reconcile host '
               'aggregates (e.g., add aggregate-less hosts to the default '
               'host aggregate)'),
    cfg.BoolOpt('allow_host_in_multiple_host_aggregates', default=False,
                help='Determines whether a host can belong to multiple '
                'host aggregates (not officially supported); setting this '
                'to True is to be used at your own risk'),
    cfg.BoolOpt('ensure_aggregate_metadata_exists', default=True,
                help='Determines whether expected host aggregate metadata '
                'keys will, as part of the reconciliation process, be '
                'automatically added to the aggregate if they are not set')
])

LOG = logging.getLogger(__name__)


def _ensure_aggregate_exists(context, agg_name, metadata=None):
    """Creates the specified host aggregate if necessary."""
    agg_id = _get_aggregate_id_by_name(context, agg_name)
    if not agg_id:
        api = compute_api.AggregateAPI()
        agg = api.create_aggregate(context, agg_name, None)
        if metadata:
            api.update_aggregate_metadata(context, agg.id, metadata)


def _ensure_aggregate_metadata(context):
    """Ensures the aggregates have the proper metadata defined."""
    def _check_key(context, agg, key, default_val):
        if key not in agg.metadata:
            LOG.info(_("Detected that host aggregate '%(host_agg)s' didn't "
                       "have '%(key)s' metadata defined; defaulting its "
                       "value to '%(default)s' immediately.")
                     % {'host_agg': agg.name, 'key': key,
                        'default': default_val})
            api.update_aggregate_metadata(
                context, agg.id, metadata={key: default_val})

    if not cfg.CONF.ensure_aggregate_metadata_exists:
        LOG.info(_("Skipping validation of aggregate metadata because the "
                   "'ensure_aggregate_metadata_exists' nova.conf option "
                   "is set to false."))
        return

    api = compute_api.AggregateAPI()
    for agg in objects.AggregateList.get_all(context):
        # Make sure the aggregate's initial placement policy ID is set.
        _check_key(context, agg, ego_const.INITIAL_POLICY_ID,
                   pvc_prs_const.DEFAULT_INITIAL_POLICY_ID)
        # Make sure the aggregate's HA policy fields are set.
        _check_key(context, agg, ego_ha_const.KEY_HA_ID,
                   pvc_prs_const.DEFAULT_HA_ID)
        _check_key(context, agg, ego_ha_const.KEY_HA_RUN_INTERVAL,
                   pvc_prs_const.DEFAULT_HA_RUN_INTERVAL)
        _check_key(context, agg, ego_ha_const.KEY_HA_STABILIZATION,
                   pvc_prs_const.DEFAULT_HA_STABILIZATION)
        # Make sure the aggregate's DRO policy fields are set.
        _check_key(context, agg, pvc_prs_const.DRO_ENABLED_KEY,
                   pvc_prs_const.DEFAULT_DRO_ENABLED)
        _check_key(context, agg, ego_const.RUNTIME_POLICY_ID,
                   pvc_prs_const.DEFAULT_RT_POLICY_ID)
        _check_key(context, agg, ego_const.RUNTIME_POLICY_RUN_INTERVAL,
                   pvc_prs_const.DEFAULT_RT_RUN_INTERVAL)
        _check_key(context, agg, ego_const.RUNTIME_POLICY_STABILIZATION,
                   pvc_prs_const.DEFAULT_RT_STABILIZATION)
        _check_key(context, agg, ego_const.RUNTIME_POLICY_THRESHOLD,
                   pvc_prs_const.DEFAULT_RT_THRESHOLD)
        _check_key(context, agg, ego_const.RUNTIME_POLICY_MAX_PARALLEL,
                   pvc_prs_const.DEFAULT_RT_MAX_PARALLEL)
        _check_key(context, agg, ego_const.RUNTIME_POLICY_ACTION,
                   pvc_prs_const.DEFAULT_RT_ACTION)


def _get_aggregate_id_by_name(context, agg_name):
    """Retrieves the ID of the host aggregate given its name."""
    aggregates = [agg.id for agg in objects.AggregateList.get_all(context)
                  if agg.name == agg_name]
    return None if not aggregates else aggregates[0]


def add_host_to_aggregate(context, host_name, agg_name):
    """
    Adds a host to the specified host aggregate.

    :param context: the nova context object for the operation
    :param host_name: the host name to add to the aggregate
    :param agg_name: the name of the aggregate in which to place the host
    :returns: a dictionary of the aggregate's properties, metadata and hosts
    """
    agg_id = _get_aggregate_id_by_name(context, agg_name)
    if not agg_id:
        raise exception.AggregateNotFound(aggregate_id=agg_name)
    api = compute_api.AggregateAPI()
    return api.add_host_to_aggregate(context, agg_id, host_name)


def remove_host_from_aggregate(context, host_name, agg_id):
    """
    Removes a host from the specified host aggregate.

    :param context: the nova context object for the operation
    :param host_name: the host name to remove from the aggregate
    :param agg_id: the identifier of the aggregate from which the host should
                   be removed
    :returns: a dictionary of the aggregate's properties, metadata and hosts
    """
    api = compute_api.AggregateAPI()
    return api.remove_host_from_aggregate(context, agg_id, host_name)


def get_dro_enabled_for_aggregate(context, agg_id):
    """
    Returns whether or not DRO is enabled for the given aggregate.

    :param context: the nova context object for the operation
    :param agg_id: the identifier of the aggregate of interest
    :returns: True if DRO is enabled for the aggregate, False otherwise
    """
    api = compute_api.AggregateAPI()
    aggregate = api.get_aggregate(context, agg_id)
    return str(aggregate.metadata.get(
        pvc_prs_const.DRO_ENABLED_KEY)).lower() == 'true'


def get_aggregate_ids_for_host(context, host_name):
    """
    Retrieves a list of host aggregate IDs to which a host belongs.

    :param context: the nova context object for the operation
    :param host_name: the host name whose host aggregate IDs to retrieve
    :returns: a list of host aggregate IDs to which the host belongs
    """
    aggregates = objects.AggregateList.get_by_host(context, host_name)
    return [agg.id for agg in aggregates]


def get_aggregate_list(context):
    """
    Get a list of all host aggregates

    :param context: The nova context object for the operation.
    :returns: A list of dictionaries of the aggregate's properties, metadata
              and hosts.
    """
    api = compute_api.AggregateAPI()
    return api.get_aggregate_list(context)


def get_hosts_for_aggregate(context, agg_id):
    """
    Get a list of all hosts in the specific aggregate

    :param context: The nova context object for the operation.
    :param agg_id: The identifier of the aggregate for which the list of
                   member hosts is desired.
    :returns: A list of host names that are part of the specified aggregate.
    """
    api = compute_api.AggregateAPI()
    return api.get_aggregate(context, agg_id).hosts


@lockutils.synchronized('reconcile_aggregates', 'nova-reconcile-aggregates-',
                        external=True)
def reconcile_aggregates(default_agg_name):
    """
    This method reconciles various aspects of host aggregates.  This method is
    synchronized with an external process lock to prevent multiple API worker
    processes from trying to create the default aggregate several times.
    Namely, it:

    1. Ensures the default host aggregate exists;
    2. Ensures the host aggregates have the expected metadata defined;
    3. Ensures that there aren't any aggregate-less hosts by adding them to
       the default host aggregate if necessary.

    :param default_agg_name: the name of the default host aggregate
    """
    context = nova_context.get_admin_context()

    # 1. Make sure the default host aggregate exists.
    try:
        _ensure_aggregate_exists(
            context, default_agg_name,
            metadata={ego_const.INITIAL_POLICY_ID:
                      pvc_prs_const.DEFAULT_INITIAL_POLICY_ID,
                      ego_ha_const.KEY_HA_ID:
                      pvc_prs_const.DEFAULT_HA_ID,
                      ego_ha_const.KEY_HA_RUN_INTERVAL:
                      pvc_prs_const.DEFAULT_HA_RUN_INTERVAL,
                      ego_ha_const.KEY_HA_STABILIZATION:
                      pvc_prs_const.DEFAULT_HA_STABILIZATION,
                      pvc_prs_const.DRO_ENABLED_KEY:
                      pvc_prs_const.DEFAULT_DRO_ENABLED,
                      ego_const.RUNTIME_POLICY_ID:
                      pvc_prs_const.DEFAULT_RT_POLICY_ID,
                      ego_const.RUNTIME_POLICY_RUN_INTERVAL:
                      pvc_prs_const.DEFAULT_RT_RUN_INTERVAL,
                      ego_const.RUNTIME_POLICY_STABILIZATION:
                      pvc_prs_const.DEFAULT_RT_STABILIZATION,
                      ego_const.RUNTIME_POLICY_THRESHOLD:
                      pvc_prs_const.DEFAULT_RT_THRESHOLD,
                      ego_const.RUNTIME_POLICY_MAX_PARALLEL:
                      pvc_prs_const.DEFAULT_RT_MAX_PARALLEL,
                      ego_const.RUNTIME_POLICY_ACTION:
                      pvc_prs_const.DEFAULT_RT_ACTION})
    except Exception as exc:
        LOG.error(_("Error ensuring that the default host aggregate exists: "
                    "%s") % six.text_type(exc))
        LOG.exception(exc)

    # 2. Make sure the aggregates have the expected metadata defined.
    try:
        _ensure_aggregate_metadata(context)
    except Exception as exc:
        LOG.error(_("Error ensuring that all host aggregates have the "
                    "expected metadata defined: %s") % six.text_type(exc))
        LOG.exception(exc)

    # 3. Put aggregate-less hosts in the default host aggregate.
    context = context.elevated()
    host_names = [node.host
                  for node in objects.ComputeNodeList.get_all(context)]
    for host_name in host_names:
        agg_ids = get_aggregate_ids_for_host(context, host_name)
        if not agg_ids:
            # The host wasn't a member of any aggregates; add to the default.
            LOG.info(_("Detected that host '%(host)s' was not a member of any "
                       "host aggregate; adding it to the default host "
                       "aggregate.") % {'host': host_name})
            try:
                add_host_to_aggregate(context, host_name, default_agg_name)
            except Exception as exc:
                LOG.error(_("Error automatically adding host to default host "
                            "aggregate: %s") % six.text_type(exc))
                LOG.exception(exc)
