# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Host evacuation related utilities."""

import functools
import six

from oslo_utils import encodeutils
from oslo_utils import excutils
from oslo_log import log as logging
from nova import db
from nova import rpc
from nova import servicegroup
from nova import exception as nova_exc
from nova import objects
from nova.compute import vm_states
from nova.compute import task_states
from nova.ibm.ego import ego_utils
from nova.ibm.ego.hapolicy import constants as ha_constants

from powervc_nova import _
from powervc_nova import compute
from powervc_nova.compute import exception as pvc_exc
from powervc_nova.virt.ibmpowervm.common import constants as consts

LOG = logging.getLogger(__name__)

get_notifier = functools.partial(rpc.get_notifier, service='compute')

SUCCESS_MSG = (_("Remote restart for all virtual machines under the host "
                 "{host_name} was {enablement} successfully."))

ERROR_MSG = (_("Change to {enablement} remote restart all virtual machines "
               "under host {host_name} failed with exception: {error}"))

ENABLED = _('enabled')

ENABLE = _('enabled')

DISABLED = _('disabled')

DISABLE = _('disable')

HOST_MAINTENANCE_STATUSES = [ha_constants.STATE_MT_ERROR,
                             ha_constants.STATE_MT_ON,
                             ha_constants.STATE_MT_ENTERING]


def _get_compute_node(context, service_host):
    """
    This returns the respective compute node given a service host string.

    :param context: A context object that is used to authorize DB access
    :param service_host: The service host whose compute node to locate
    :returns: The corresponding compute node or None if not found
    """
    context = context.elevated()
    compute_nodes = (objects.ComputeNodeList.get_all_by_host(
        context, service_host) or [])

    return None if not len(compute_nodes) else compute_nodes[0]


def reset_all_instances(context, host):
    instances = objects.InstanceList.get_by_host(context, host)
    if instances is None:
        return
    for inst in instances:
        LOG.debug('instance - %s, task state - %s' % (inst.name,
                                                      inst.task_state))
        if (inst.vm_state not in [vm_states.ACTIVE, vm_states.STOPPED,
                                  vm_states.ERROR]):
            if (inst.task_state not in [None, task_states.REBUILDING,
                                        task_states.REBUILD_SPAWNING]):
                inst.vm_state = vm_states.ACTIVE
                inst.task_state = None
                inst.save()
                LOG.info(_('Reset states of instance %s for remote restart '
                           'operation.') % inst.name)


def _notify_compute_node_update(context, node, stats, postfix):
    notifier = rpc.get_notifier('compute', node.host)
    event_type = 'compute.node.update.'
    event_suffix = event_type + postfix

    service = objects.Service.get_by_compute_host(context, node.host)

    info = {'host': node.host,
            'hypervisor_hostname': node.hypervisor_hostname,
            'compute_node_id': node.id,
            'service_id': service.id,
            'stats': stats}

    # Send the notification that we are Creating/Updating the Compute Node
    notifier.info(context, event_suffix, info)


def send_ui_notification():
    def func_parms(f):
        @functools.wraps(f)
        def wrapper(*args, **kwds):
            def _notify(exception_msg=None):
                payload = {}
                payload['msg'] = ''
                payload['host_name'] = host_name
                payload['enablement'] = enablement
                notifier = rpc.get_notifier('compute', host_name)
                try:
                    if exception_msg is not None:
                        if error_message is not None:
                            payload['msg'] = error_message
                            payload['error'] = exception_msg
                            try:
                                # Convert the brackets to replacement
                                # variables
                                log_exception_msg = error_message.replace(
                                    "{", "%(").replace("}", ")s")
                                LOG.exception(log_exception_msg % payload)
                            except Exception as exc:
                                LOG.debug('LOG.exception failed with '
                                          'exception: %s' %
                                          six.text_type(exc))
                            notifier.error(context,
                                           'compute.instance.log',
                                           payload)

                    else:
                        if success_message is not None:
                            payload['msg'] = success_message
                            notifier.info(context,
                                          'compute.instance.log',
                                          payload)
                except Exception as exc:
                    LOG.debug('Notify failed with exception: %s' %
                              six.text_type(exc))

            context = args[0]
            host_name = args[1]
            error_message = ERROR_MSG
            success_message = SUCCESS_MSG
            host_dead = kwds['dead'] if 'dead' in kwds else True
            enablement = ENABLED if host_dead else DISABLED
            try:
                r = f(*args, **kwds)
                # notify success
                _notify()
                return r
            except Exception as exc:
                with excutils.save_and_reraise_exception():
                    # notify failure
                    exception_msg = six.text_type(exc)
                    enablement = ENABLE if host_dead else DISABLE
                    if not exception_msg:
                        exception_msg = exc.__class__.__name__
                    _notify(exception_msg)

        return wrapper
    return func_parms


def _check_host_maintenance(node_name, target_node_name=None):
    def in_host_maintenance(status):
        return status in HOST_MAINTENANCE_STATUSES

    ego_states = ego_utils.get_compute_status()
    maint_status = ego_states.get_maintenance(node_name)
    if in_host_maintenance(maint_status):
        exc_args = {'host': encodeutils.safe_decode(node_name),
                    'state': maint_status}
        raise pvc_exc.IBMPowerVCNotAllowed(
            _('Remote restart cannot be performed on source host %(host)s '
              'because its maintenance state is %(state)s.') % exc_args)
    if target_node_name:
        maint_status = ego_states.get_maintenance(target_node_name)
        if in_host_maintenance(maint_status):
            exc_args = {'host': target_node_name,
                        'state': maint_status}
            raise pvc_exc.IBMPowerVCNotAllowed(
                _('Host %(host)s cannot be a remote restart destination '
                  'because its maintenance state is %(state)s.') % exc_args)


@send_ui_notification()
def set_host_dead(context, node_name, dead=True,
                  target_node_name=None):
    node = _get_compute_node(context, node_name)
    if node is None:
        LOG.error(_('Compute host %s is not found.')
                  % encodeutils.safe_decode(node_name))
        raise nova_exc.NotFound(_('Compute host %s is not found.')
                                % encodeutils.safe_decode(node_name))
    # extract the current stats info
    service = objects.Service.get_by_compute_host(context, node.host)
    src = servicegroup.API().service_is_up(service)
    if not src:
        LOG.debug('service is not up')
    # make sure there are at least 2 hosts
    all_nodes = db.compute_node_get_all(context)
    stats = node.stats or {}
    host_dead = str(stats.get(consts.HOST_DEAD, False)).lower() == 'true'
    if host_dead != dead:
        if dead:
            if not all_nodes or len(all_nodes) < 2:
                raise nova_exc.ConstraintNotMet(
                    _('PowerVC must be managing at least two hosts to '
                      'remote restart all virtual machines under the failed '
                      'host.'))
            _check_host_maintenance(node_name, target_node_name)

        # need to change host_dead value
        # if the service is not up, modify the db value directly
        if not src:
            if dead:
                reset_all_instances(context, node_name)
            stats[consts.HOST_DEAD] = dead
            if dead and target_node_name:
                stats[consts.REMOTE_RESTART_TARGET_HOST] = (
                    target_node_name)
            else:
                stats.pop(consts.REMOTE_RESTART_TARGET_HOST, None)
            _notify_compute_node_update(context, node, stats, 'start')
            node.stats = stats
            node.save()
            _notify_compute_node_update(context, node, stats, 'end')
            LOG.info(_('Remote restart for all virtual machines under host '
                       '%s is set to %s.')
                     % (encodeutils.safe_decode(node_name), dead))
        else:
            # check whether it's powerkvm env or not
            hyp_type = node.hypervisor_type
            if not hyp_type or hyp_type.lower() != 'qemu':
                # get the HostAPI
                host_api = compute.HostAPI()
                host_api.set_host_dead(context, node_name, dead,
                                       target_node_name)
            else:
                raise pvc_exc.IBMPowerVCNotAllowed(
                    _('Remote restart is allowed only '
                      'when the source host is down.'))
    else:
        raise pvc_exc.IBMPowerVCNotAllowed(
            _('Remote restart for all virtual machines under host %s '
              'is already set to %s.') % (encodeutils.safe_decode(node_name),
                                          dead))
