# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Server group related utilities."""

from oslo_config import cfg
from nova import objects
from nova.ibm.ego import ego_constants
from nova.compute import task_states
from nova.compute import vm_states
from oslo_log import log as logging

from powervc_nova import _

LOG = logging.getLogger(__name__)

cfg.CONF.register_opts([
    cfg.BoolOpt('allow_parallel_operations_per_collocation_rule',
                default=False,
                help='Determines whether to allow a deploy, rebuild, or '
                'migration operation for a virtual machine in a collocation '
                'rule, if another virtual machine in the same collocation '
                'rule is currently undergoing one of these operations. '
                'If True, then a race condition could occur that would cause '
                'the collocation rule to become violated.'),
    cfg.BoolOpt('ignore_collocation_rules',
                default=False,
                help='Determines whether to ignore all collocation rules '
                'during all placement decisions.')
])

AFFINITY = 'affinity'
ANTI_AFFINITY = 'anti-affinity'

# States instances cannot be in when adding a member.
PROHIBITED_VM_STATES = [vm_states.BUILDING]
PROHIBITED_TASK_STATES = [task_states.REBUILDING,
                          task_states.REBUILD_BLOCK_DEVICE_MAPPING,
                          task_states.REBUILD_SPAWNING,
                          task_states.MIGRATING]
BUILDING_TASKS = [task_states.SCHEDULING,
                  task_states.BLOCK_DEVICE_MAPPING,
                  task_states.NETWORKING,
                  task_states.SPAWNING]
PROHIBITED_TASK_STATES = PROHIBITED_TASK_STATES + BUILDING_TASKS


def get_members(context, group_id):
    """
    Call the database method to get back a list of instances
    in a server group. Then return back a list of instance
    objects for each non-deleted instance in the list.
    """
    instances = []
    sg = objects.InstanceGroup.get_by_uuid(context, group_id)
    if sg.members:
        # Only return instances that have not been deleted
        filters = {'uuid': sg.members, 'deleted': False}
        instances = objects.instance.InstanceList.get_by_filters(
            context, filters=filters)
    return instances


def get_hosts_for_each_member(context, sg, members=None, has_host_only=True):
    """
    Returns back a list of tuples of instance uuids and instance host
    for each instance in the server group. If has_host_only is True,
    then only members with hosts assigned will be returned. Otherwise,
    all non deleted members will be returned.
    """
    if not members:
        members = get_members(context, sg.uuid)
    if has_host_only:
        return [(instance.uuid, instance.host) for instance in members
                if instance.host]
    else:
        return [(instance.uuid, instance.host) for instance in members]


def get_all_groups_for_instance(context, instance_uuid):
    """
    Returns instance group objects that the instance is a member of.
    """
    sgs = objects.InstanceGroupList.get_all(context)
    return [group for group in sgs if instance_uuid in group.members]


def get_groups_for_instance_by_policy(context, instance_uuid, policy):
    """
    Returns instance group objects that the instance is a member of
    for each group with either an affinity or anti-affinity policy.
    """
    sgs = get_all_groups_for_instance(context, instance_uuid)
    return [group for group in sgs if policy in group.policies]


def get_str_from_list(items):
    """Converts a list of items into a comma separated string."""
    num_items = len(items)
    if num_items == 1:
        return items[0]
    else:
        items[-1] = 'and ' + items[-1]
    if num_items == 2:
        return ' '.join(items)
    else:
        return ', '.join(items)


def has_ongoing_operations_in_group(group_uuid, instance_uuid, group_instances,
                                    operation, request_spec=None):
    """
    Returns True if there is an ongoing create, rebuild, or migrate operation
    for other instances in this group. request_spec must be passed in if
    operation is create.

    :param group_uuid: uuid of the server group
    :param instance_uuid: uuid of the instance undergoing the operation
    :param group_instances: list of instance objects for all instances in the
                            group.
    :param operation operation the instance is undergoing in the form
                           of ego_constances.FILTER_WORKING_SCOPE_...
    :param request_spec: request spec from the filter properties of the
                         request. Only required when operation is create.
    :returns: True/False
    """
    allow_ongoing = cfg.CONF.allow_parallel_operations_per_collocation_rule
    has_ongoing_operation = False
    # Loop through each instance in the group, and check its vm_state and
    # task_state.
    for instance in group_instances:
        if (instance.uuid != instance_uuid and (
            instance.task_state in PROHIBITED_TASK_STATES or
                instance.vm_state in PROHIBITED_VM_STATES)):
            # If this is a create operation, and another instance is
            # also being created in the same group
            if (operation == ego_constants.FILTER_WORKING_SCOPE_CREATE and
                (instance.vm_state == vm_states.BUILDING or
                    instance.task_state in BUILDING_TASKS)):
                # Get a list of instances for the request
                instance_list = request_spec.get('instance_list')
                # If there is more than one instance being created as par of
                # this same request, the scheduler and compute manager will
                # enforce the rule to make sure no race conditions will
                # occur, so this is not counted as a separate parallel
                # operation in the same group.
                if len(instance_list) > 1 and instance.uuid in instance_list:
                    continue

            # There is an ongoing operation for one of the other instances
            # in the same group. Log a warning, and if allow_ongoing is set
            # to False, set has_ongoing_operation to True.
            LOG.warn(_("Collocation rule %(group)s contains virtual machine "
                       "'%(instance_name)s' (%(instance_uuid)s) with a "
                       "vm_state of %(vm_state)s and a task_state of "
                       "%(task_state)s.") %
                     {'group': group_uuid,
                      'instance_name': instance.display_name,
                      'instance_uuid': instance.uuid,
                      'vm_state': instance.vm_state,
                      'task_state': instance.task_state})
            if not allow_ongoing:
                has_ongoing_operation = True
                break
    return has_ongoing_operation


def ignore_collocation_rules():
    return cfg.CONF.ignore_collocation_rules
