#
#
# Copyright 2013, 2015 IBM Corp.
# All Rights Reserved.
# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2011 Justin Santa Barbara
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import time
import threading
import traceback
import weakref

from eventlet import greenthread
from eventlet import timeout
import greenlet
from oslo_utils import excutils
from oslo_log import log as logging
from oslo_concurrency.lockutils import synchronized
from oslo_context import context as common_context
from powervc_nova.compute import exception as pvc_exception

from powervc_nova import _

LOG = logging.getLogger(__name__)


class WeakLocal(threading.local):
    def __getattribute__(self, attr):
        rval = super(WeakLocal, self).__getattribute__(attr)
        if rval:
            rval = rval()
        return rval

    def __setattr__(self, attr, value):
        value = weakref.ref(value)
        return super(WeakLocal, self).__setattr__(attr, value)

_LOCAL_STORE = WeakLocal()


class GreenThreadGroupCB(object):
    def __init__(self, tg):
        self.tg = tg
        self.bailout = False
        self.killing = False
        self.exception = None
        self.except_thread = None
        self.except_thread_name = None


class GreenThreadGroup(object):
    """
    This class adds the concept of thread groups to the greenthread library
    Principles:
      -Fail early and often.
      -Allow parent to wait on any thread or group of threads
      -Allow parent to get the return value of any individual thread
      -Allow child threads to know when other threads have bailed out
      -Allow the parent to tell the children to bail out
      -Set the context in children if it's available at spawn
      -Allow new threads to be added to the group at any time
    """

    def __init__(self, uid=None, context=None, bail_check_spawn=True):
        self.thread_start_count = 0
        self.thread_end_count = 0
        self.run_thread_map = {}
        self.ended_thread_map = {}
        self.claim_thread_map = {}
        self.uid = uid if uid is not None else time.time()
        self.bail_check_spawn = bail_check_spawn
        self.control_block = GreenThreadGroupCB(self)
        self.control_block.send_except_to_waiters = False

    @staticmethod
    def get_control_block():
        return None if not hasattr(_LOCAL_STORE,
                                   'thread_group_control_block') else \
            _LOCAL_STORE.thread_group_control_block

    @staticmethod
    def thread_should_bailout():
        cb = GreenThreadGroup.get_control_block()
        return False if cb is None else cb.bailout

    @staticmethod
    def get_thread_group():
        cb = GreenThreadGroup.get_control_block()
        return None if cb is None else cb.tg

    def spawn(self, f, *args, **kwds):
        """
        Keep track of threads that are started so they can be cleaned up.

        It will also grab the context from the threadlocal store and add it to
        the store on the new thread. This allows for continuity in logging the
        context when using this method to spawn a new thread.
        See LP bug: https://bugs.launchpad.net/nova/+bug/1404268
        """
        _context = common_context.get_current()

        # This method will be the entry point of our new thread
        def green_thread(control_block, go, f, *args, **kwds):
            # NOTE: If update_store is not called after spawn_n it won't be
            # available for the logger to pull from threadlocal storage.
            if _context is not None:
                _context.update_store()
            _LOCAL_STORE.thread_group_control_block = control_block

            # Wait until we're completely setup before running the real func
            while not go[0]:
                greenthread.sleep(0.1)
            try:
                rtnval = f(*args, **kwds)
            finally:
                LOG.debug("Returned from spawned thread.")

                # Clean up after our thread
                if hasattr(_LOCAL_STORE, 'thread_group_control_block'):
                    del _LOCAL_STORE.thread_group_control_block

            return rtnval

        @synchronized(self.uid, 'thread-group')
        def _spawn():
            go = [False]
            t = greenthread.spawn(green_thread, self.control_block, go,
                                  f, *args, **kwds)
            self.run_thread_map[t] = f.__name__
            t.link(self._thread_end)
            self.thread_start_count += 1
            # The thread is completely setup, set it free
            go[0] = True
            LOG.debug("Created thread %s for %s: %s" %
                      (self.thread_start_count, f.__name__, t))
            greenthread.sleep(0)
            return t

        # Before we spawn more collect up any ended threads to see
        # if they might have thrown an exception.
        self._process_ended_threads()
        if ((self.bail_check_spawn and self.control_block.bailout) or
                self.control_block.killing):
            return None
        return _spawn()

    def _thread_end(self, t):
        """
        Remove completed threads from map of threads.
        Sync modify access to the dict
        """
        @synchronized(self.uid, 'thread-group')
        def __thread_end():
            # try in case it was already removed.
            self.thread_end_count += 1
            try:
                # Put the thread in the ending thread map
                self.ended_thread_map[t] = self.run_thread_map[t]
                del self.run_thread_map[t]
            except Exception as e:
                LOG.debug("Error removing thread from map.")
                LOG.exception(e)

        LOG.debug("Removing %s from thread_map: %s" % (self.run_thread_map[t],
                                                       t))
        # print 'Thread ending... count is %s' % self.thread_end_count
        __thread_end()

    def _wait_on_thread(self, t, t_name):
        """
        Wait on a thread. Remove it from the running dict and
        record its return value to be claimed
        """
        @synchronized(self.uid, 'thread-group')
        def __wait_on_thread():
            try:
                rtnval = t.wait()
                # Put this value in the dict.
                self.claim_thread_map[t] = rtnval
            except (object, Exception) as e:
                with excutils.save_and_reraise_exception():
                    LOG.debug("Caught exception in wait on thread: %s" % t)
                    self.control_block.bailout = True
                    # Record the first exception of the group
                    if not self.control_block.exception:
                        self.control_block.exception = e
                        self.control_block.except_thread = t
                        self.control_block.except_thread_name = t_name
                    self.claim_thread_map[t] = e
            except greenlet.GreenletExit:
                # don't allow an exit exception to kill the caller
                # pretend they didn't return anything.
                self.claim_thread_map[t] = None
                pass
            finally:
                try:
                    del self.ended_thread_map[t]
                except (object, Exception) as e:
                    LOG.debug("Error removing thread from map.")
                    LOG.exception(e)

        __wait_on_thread()

    def _process_ended_threads(self):
        """
        Process the ended dict.
        """
        for t in self.ended_thread_map.keys():
            LOG.debug("Thread is in ended: %s" % t)
            # wait on and save the return value
            self._wait_on_thread(t, self.ended_thread_map[t])

    def wait(self, threads=None, sleep_time=1):
        """
        Wait on the group of threads
        Our job is to:
          - process all ended threads, we need their results and exceptions
          - wait on a list of threads if given or all threads if no list
        """

        def _process_claims(thread_list):
            """
            This routine will take the list of threads that we're
            waiting on and see if we have a return code for them already.
            If so just remove them from the list we're waiting on.
            """
            for t in thread_list:
                if t in self.claim_thread_map:
                    thread_list.remove(t)

        LOG.debug("Waiting on thread group.")
        # Build the list of all threads to wait on.
        # This could be one, a list or all
        if isinstance(threads, list):
            wait_for_threads = copy.copy(threads)
        elif isinstance(threads, greenthread.GreenThread):
            wait_for_threads = [threads]
        else:
            wait_for_threads = [t for t in self.run_thread_map.keys()]

        # Check if we're killing threads and never let a waiting thread
        # out until we're done.  Only the parent should be left.
        # Always process ended threads
        # Then consider if we have running threads we may be waiting for
        while (self.control_block.killing or
               len(self.ended_thread_map) or
               (len(self.run_thread_map) and len(wait_for_threads))):

            self._process_ended_threads()
            _process_claims(wait_for_threads)
            # See if we found all our threads
            if len(wait_for_threads) > 0 or self.control_block.killing:
                greenthread.sleep(sleep_time)
            if self.control_block.send_except_to_waiters:
                break

        if self.control_block.send_except_to_waiters:
            LOG.info(_("All threads in the thread group have been ended"))
            raise pvc_exception.IBMPowerVCThreadGroupEnding()
        # See if we need a return value or exception
        if isinstance(threads, greenthread.GreenThread):
            if isinstance(self.claim_thread_map[threads], Exception):
                raise self.claim_thread_map[threads]
            else:
                return self.claim_thread_map[threads]

    def kill(self, delay=None, order=None, time_out=1200, forceful=False):
        """
        Kill all remaining threads

        Note:  There is a bug in the interaction between threading.py,
        eventlet threading and @synchronized such that if you kill a running
        thread, it will leave a lock causing all threads to lockup.
        """

        def _kill(t, forceful=False):
            if forceful:
                t.kill()
            # else:
            # Remove this cancel since a cancelled thread may not call the
            # thread end method and then we'd lose track of it.
                # t.cancel()

        def _kill_all(forceful=False):
            # The dict changes while we kill so make a copy
            running = copy.copy(self.run_thread_map)
            for t in running:
                try:
                    LOG.info(_("Terminating thread. %s") % running[t])
                    LOG.debug("Thread counts: start: %s ended: %s"
                              % (self.thread_start_count,
                                 self.thread_end_count))
                    _kill(t, forceful=forceful)
                except (object, Exception):
                    pass

        def _wait_for_all():
            while len(self.run_thread_map):
                try:
                    LOG.info(_("Waiting for all threads to end."))
                    LOG.debug("Thread counts: start: %s ended: %s"
                              % (self.thread_start_count,
                                 self.thread_end_count))
                    self.wait()
                except (object, Exception):
                    pass

            return True

        LOG.debug("Starting to terminate all threads in the group.")
        self.control_block.bailout = True
        self.control_block.killing = True
        if delay is not None:
            greenthread.sleep(delay)

        # if an order is specified, kill in sequence
        if order is not None:
            for t in order:
                _kill(t, forceful=forceful)

        # kill all others
        _kill_all(forceful=forceful)

        # wait for them all...
        self.control_block.killing = False
        try:
            with timeout.Timeout(time_out):
                _wait_for_all()
        except timeout.Timeout:
            LOG.error(_('Error ending all child threads within the timeout.'))
            running = copy.copy(self.run_thread_map)
            for t in running:
                LOG.error(_('Thread %s is still running.') % running[t])
                # log the stack trace
                LOG.error(traceback.format_stack(t.gr_frame))
            # REMOVE this forceful kill of the remaining threads.  It
            # causes deadlocks when threads are holding @synchronized
            # locks on methods
            # _kill_all(forceful=True)
            # Instead, wait for them all to end.

            # in case of threads still waiting
            # interrupt them
            self.control_block.send_except_to_waiters = True
            _wait_for_all()
            LOG.info(_("All threads in the group have now ended."))

        finally:
            LOG.debug("After termination thread counts: start: %s ended: %s"
                      % (self.thread_start_count, self.thread_end_count))
