# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from eventlet import greenthread
from nova.objects import service as service_obj
from oslo_context import context as oslo_ctxt
from oslo_messaging import RemoteError
from oslo_utils import importutils, encodeutils
from nova import objects
from nova.compute import api as base_api
from nova.compute import instance_actions
from nova.compute import task_states
from nova.compute import vm_states
from nova.virt import block_device
import six
from eventlet import timeout
from functools import wraps
from keystoneclient.v3 import client as keystone_client
from keystonemiddleware.auth_token._request import _v3_to_v2_catalog
from oslo_config import cfg
from nova import context as ctx
from nova import rpc
from nova import exception as nova_exception
from nova.ibm.ego import hapolicy
from nova.ibm.ego.hapolicy import constants as ha_constants
from nova.network.neutronv2 import api as neutronv2_api
from nova.objects import instance as instance_obj
from nova.objects import compute_node as comp_node_obj
from oslo_utils import excutils, timeutils
from oslo_serialization import jsonutils
from oslo_log import log as logging
from powervc_nova.compute import common_constants
from powervc_nova.compute import exception as pvc_exception
from powervc_nova.virt.ibmpowervm.common import exception as pvc_virt_exception
from powervc_nova.virt.ibmpowervm.common import constants as pvmc
from powervc_oslo import override_context_serialize
from powervc_nova import _
from powervc_nova.virt.ibmpowervm.common import constants
from powervc_nova.objects import migrate_data as mig_obj
from powervc_oslo.config import config_utils

LOG = logging.getLogger(__name__)

CONF = cfg.CONF

EVENT_TYPE_ERROR = 'ERROR'
EVENT_TYPE_WARN = 'WARN'
datacache = dict()

MIG_OPER_TYPES = [common_constants.OPER_MIGRATE_AS_SRC,
                  common_constants.OPER_MIGRATE_AS_TGT]

# Override the Context Serialization to add Locale
override_context_serialize(ctx.RequestContext)


def API(*args, **kwargs):
    class_name = 'powervc_nova.compute.api.PowerVCComputeAPI'
    return importutils.import_object(class_name, *args, **kwargs)


def HostAPI(*args, **kwargs):
    class_name = 'powervc_nova.compute.api.PowerVCHostAPI'
    return importutils.import_object(class_name, *args, **kwargs)


def monkey_patch_evacuate():
    """
    We need to monkey patch the nova.compute.api.API class'
    evacuate method to avoid checking the source compute service state.
    """
    base_api.API.powervc_mod_evacuate = \
        base_api.API.evacuate

    @base_api.check_instance_state(vm_state=[vm_states.ACTIVE,
                                             vm_states.STOPPED,
                                             vm_states.ERROR])
    def powervc_mod_evacuate(self, context, instance, host, on_shared_storage,
                             admin_password=None):
        service = service_obj.Service.get_by_compute_host(context.elevated(),
                                                          instance['host'])

        if service and service['compute_node']:
            hypervisor_type = service['compute_node']['hypervisor_type']

        LOG.debug('hypervisor_type - %s ' % hypervisor_type)

        if hypervisor_type and hypervisor_type == 'phyp':
            if self.servicegroup_api.service_is_up(service):
                context = context.elevated()
                nodes = comp_node_obj.ComputeNodeList.get_by_hypervisor(
                    context, instance['host'])
                node = nodes[0]
                hypervisor_type = node.hypervisor_type
                stats = node.stats or {}
                hyp_state = stats.get(pvmc.HYPERVISOR_STATE)
                if (hyp_state not in {'error', 'error - dump in progress',
                                      'power off', 'no connection'}):
                    LOG.error(_('Cannot remote restart %s because the source '
                                'host %s is expected to be down, but it was '
                                'up.'),
                               (instance['name'], instance['host']))
                    msg_fmt = _('Remote restart operation cannot be performed '
                                'when source host state is %s') % hyp_state
                    raise nova_exception.ComputeServiceInUse(message=msg_fmt)
            else:
                # NovaLink scenario, the service seems to be down but need
                # to ensure that the NovaLink partition is really down before
                # proceeding with Remote Restart
                context = context.elevated()
                nodes = comp_node_obj.ComputeNodeList.get_by_hypervisor(
                    context, instance['host'])
                node = nodes[0]
                source_novalink_ip = node.host_ip
                if config_utils.is_ipaddress_pingable(source_novalink_ip):
                    LOG.error(_('Cannot remote restart %s because the source '
                                'host %s is reachable via ping.'),
                               (instance['name'], instance['host']))
                    msg_fmt = _('Remote restart operation cannot be performed '
                                'when the source host is up and running.')
                    raise nova_exception.ComputeServiceInUse(message=msg_fmt)

            LOG.debug('vm evacuation scheduled', instance=instance)
            instance.task_state = task_states.REBUILDING
            instance.save(expected_task_state=[None])
            self._record_action_start(context, instance,
                                      instance_actions.EVACUATE)

            # NOTE(mdrabe): Create this as a tombstone for the source compute
            # to find and cleanup. No need to pass it anywhere else. This is
            # for NovaLink.
            migration = objects.Migration(context,
                                          source_compute=instance.host,
                                          source_node=instance.node,
                                          instance_uuid=instance.uuid,
                                          status='accepted',
                                          migration_type='evacuation')
            if host:
                migration.dest_compute = host
            migration.create()

            return self.compute_task_api.\
                rebuild_instance(context,
                                 instance=instance,
                                 new_pass=admin_password,
                                 injected_files=None,
                                 image_ref=None,
                                 orig_image_ref=None,
                                 orig_sys_metadata=None,
                                 bdms=None,
                                 recreate=True,
                                 on_shared_storage=on_shared_storage,
                                 host=host)

        else:
            return self.powervc_mod_evacuate(context, instance, host,
                                             on_shared_storage,
                                             admin_password=None)

    base_api.API.evacuate = powervc_mod_evacuate

monkey_patch_evacuate()


def monkey_patch_rebuild():
    """
    We need to monkey patch the nova.compute.api.API class'
    rebuild method to prevent the renaming of the instance before the
    original instance with it's original name is destroyed.
    This is only an issue for KVM.
    """
    base_api.API.powervc_mod_rebuild = \
        base_api.API.rebuild

    @base_api.check_instance_state(vm_state=[vm_states.ACTIVE,
                                             vm_states.STOPPED,
                                             vm_states.ERROR])
    def powervc_mod_rebuild(self, context, instance, image_href,
                            admin_password, files_to_inject=None, **kwargs):
        """Rebuild the given instance with the provided attributes."""
        service = service_obj.Service.get_by_compute_host(context.elevated(),
                                                          instance['host'])

        if service and service['compute_node']:
            hypervisor_type = service['compute_node']['hypervisor_type']
        if hypervisor_type and hypervisor_type.lower() == 'qemu':
            # Grab new display name and temporarily store
            # the display name in the instance's metadata
            new_name = kwargs.get('display_name')
            metadata = kwargs.get('metadata', {})
            if instance.display_name != new_name and new_name is not None:
                kwargs['display_name'] = instance.display_name
                metadata['new_display_name'] = new_name
                kwargs['metadata'] = metadata

        return self.powervc_mod_rebuild(context, instance, image_href,
                                        admin_password, files_to_inject,
                                        **kwargs)

    base_api.API.rebuild = powervc_mod_rebuild

monkey_patch_rebuild()


def monkey_patch_is_volume_backed_instance():
    """
    We need to monkey patch the nova.compute.api.API class'
    is_volume_backed_instance method to allow the PowerVM capture path to
    go through the compute manager and the virt-drivers for IVM/HMC.  The
    existing method in OpenStack drives capture of our PowerVM VMs down
    a volume backed / Cinder snapshot path that is not supported on our VMs.
    """
    base_api.API.powervc_mod_is_volume_backed_instance = \
        base_api.API.is_volume_backed_instance

    def powervc_mod_is_volume_backed_instance(self, context, instance,
                                              bdms=None):
        service = service_obj.Service.get_by_compute_host(context.elevated(),
                                                          instance['host'])

        if service and service['compute_node']:
            hypervisor_type = service['compute_node']['hypervisor_type']
        if hypervisor_type and hypervisor_type in constants.\
                IBM_POWERVM_HYPERVISOR_TYPES:
            base_api.LOG.debug("Instance %s is on a PowerVM host, "
                               "returning False" % instance['uuid'])
            return False
        return self.powervc_mod_is_volume_backed_instance(context,
                                                          instance, bdms)

    base_api.API.is_volume_backed_instance = \
        powervc_mod_is_volume_backed_instance

monkey_patch_is_volume_backed_instance()


def monkey_patch__delete():
    """
    We need to monkey patch the nova.compute.api.API class'
    _delete method to prevent delete of VM when compute service is down.
    See LP bug for detail: https://bugs.launchpad.net/bugs/1389899
    """
    base_api.API.powervc_mod__delete = \
        base_api.API._delete

    def powervc_mod__delete(self, context, instance, delete_type, cb,
                            **instance_attrs):
        host = instance['host']
        is_up = False
        if host:
            try:
                service = service_obj.Service.get_by_compute_host(
                    context.elevated(), instance.host)
                if self.servicegroup_api.service_is_up(service):
                    is_up = True
            except Exception as e:
                LOG.exception(e)
        else:
            # if host is None, it's result from failed deploy.
            # We should allow the delete to go through
            is_up = True
        if not is_up:
            raise nova_exception.ComputeServiceUnavailable(host=host)
        else:
            return self.powervc_mod__delete(context, instance, delete_type,
                                            cb, **instance_attrs)

    base_api.API._delete = \
        powervc_mod__delete

monkey_patch__delete()


def monkey_patch__check_and_transform_bdm():
    """
    We need to monkey patch the nova.compute.api.API class'
    _check_and_transform_bdm method to allow the multi-deploy of PowerVC
    volume backed images.
    """
    base_api.API.powervc_mod__check_and_transform_bdm = \
        base_api.API._check_and_transform_bdm

    def override_check_and_transform_bdm(self, context, base_options,
                                         instance_type,
                                         image_meta, min_count, max_count,
                                         block_device_mapping, legacy_bdm):
        return self.powervc_mod__check_and_transform_bdm(context, base_options,
                                                         instance_type,
                                                         image_meta, 1, 1,
                                                         block_device_mapping,
                                                         legacy_bdm)
    base_api.API._check_and_transform_bdm = \
        override_check_and_transform_bdm

monkey_patch__check_and_transform_bdm()


def monkey_vol_block_device_attach():
    block_device.DriverVolumeBlockDevice._vol_block_device_attach = \
        block_device.DriverVolumeBlockDevice.attach

    def _pvc_vol_block_device_attach(self, context, instance, volume_api,
                                     virt_driver, do_check_attach=True,
                                     do_driver_attach=False, **kwargs):
        """
        Save the bdm and update volume even if it was in-use to support
        multi attach
        """
        # This should be removed once shared volume support is
        # available in community. https://review.openstack.org/#/c/193134/
        volume_before = volume_api.get(context, self.volume_id)
        self._vol_block_device_attach(context, instance, volume_api,
                                      virt_driver,
                                      do_check_attach=do_check_attach,
                                      do_driver_attach=do_driver_attach,
                                      **kwargs)

        connection_info = self['connection_info']
        volume_id = volume_before['id']
        mode = 'rw'
        if 'data' in connection_info:
            mode = connection_info['data'].get('access_mode', 'rw')
        if volume_before['attach_status'] == "attached":
            # A save was not done in osee code , so doing in powervc
            # monkey patch to support multiattach
            self.save()
            volume_api.attach(context, volume_id, instance.uuid,
                              self['mount_device'], mode=mode)

    block_device.DriverVolumeBlockDevice.attach = _pvc_vol_block_device_attach

monkey_vol_block_device_attach()


def log_operation(op=None, instance=None, msg=None):
    """
    Log entry point of user initiated operation to aid debugging
    """
    if op and instance:
        LOG.info(_('Operation: %(op)s. Virtual machine display name: '
                   '%(display_name)s, name: %(name)s, UUID: %(uuid)s') %
                 {'op': op, 'display_name': instance['display_name'],
                  'name': instance['name'], 'uuid': instance['uuid']})
    elif msg:
        LOG.info(msg)


def is_host_in_maintenance_mode():
    """Helper method to query EGO to see if the host is in Maintenance Mode"""
    maint_stat = get_host_maintenance_mode()
    return maint_stat in (ha_constants.STATE_MT_ERROR,
                          ha_constants.STATE_MT_ON,
                          ha_constants.STATE_MT_ENTERING)


def get_host_maintenance_mode():
    """Helper method to query EGO to get the Maintenance Mode State"""
    host_state = {'ego_stat': None}
    time_out = CONF.rpc_response_timeout

    def _get_host_status(context):
        try:
            with timeout.Timeout(time_out, False):
                host_state['ego_stat'] = hapolicy.API().\
                    get_hv_status(context)
        except Exception as e:
            # If for some reason HA service is down
            # or not able to retrieve host state
            # then allow the operation on VM
            LOG.info(_("Exception while retrieving host state "
                       "from HA service %s") % six.text_type(e))

    def _get_node_name(host):
        if datacache is not None and host in datacache.keys():
            return datacache[host]

        try:
            nodes = comp_node_obj.ComputeNodeList.get_all_by_host(
                context.elevated(), host)
            node_name = nodes[0].hypervisor_hostname
            datacache[host] = node_name
        except nova_exception.NotFound:
            raise nova_exception.NotFound(_("Host %s not found")
                                          % host)
        return datacache[host]

    host = CONF.host
    context = ctx.get_admin_context()
    node_name = _get_node_name(host)
    # note (ymadhavi):
    # To retrieve host status from HA service
    # green thread is spawned separately to avoid
    # API call hanging if HA service is down
    gt = greenthread.spawn(_get_host_status, context)
    gt.wait()

    mainttag = None
    maint_stat = ""

    if host_state['ego_stat']:
        if node_name in host_state['ego_stat']['maintenance'].keys():
            mainttag = host_state['ego_stat']['maintenance'][node_name]
            maint_stat = mainttag['state']

    return maint_stat


def check_host_maintenance_mode(func_name, func_type):
    """ Decorator to check if the target host of operation
        is in maintenance mode.
        If host status is OK
            then all operations are allowed
        If host status is Maintenance error/on/entering
            then delete/stop
        live_migrate as source host is the only operation allowed
        all the time,so this decorator is not used for live_migration.
        as source.
    """
    def func_parms(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            maint_stat = get_host_maintenance_mode()
            block_operation = maint_stat in (ha_constants.STATE_MT_ERROR,
                                             ha_constants.STATE_MT_ON,
                                             ha_constants.STATE_MT_ENTERING)
            if block_operation:
                exc_args = {'state': maint_stat}
                exp = pvc_exception.IBMPowerVCInvalidOperation(**exc_args)
                # for migration operations, we needs to raise
                # MigrationPreCheckError in order to revert states
                # back to active
                if func_type in MIG_OPER_TYPES:
                    reason = six.text_type(exp)
                    exp = nova_exception.MigrationPreCheckError(reason=reason)
                    ui_notifier(reason, None, except_to_raise=exp,
                                error_event_type=EVENT_TYPE_ERROR,
                                log_exception=False)
                raise exp
            else:
                return function(*args, **kwargs)

        return wrapper
    if func_name:
        return func_parms(func_name)
    else:
        return func_parms


def put_host_in_maintenance_mode():
    """Helper method to tell EGO to put the host into Maintenance Mode"""
    # Nothing to do if it is already in Maintenance Mode
    if is_host_in_maintenance_mode():
        return
    try:
        # Call the HA Service to try to put the Host in Maintenance Mode
        hapolicy.API().set_maintenance(
            ctx.get_admin_context(), datacache[CONF.host], 'enable')
    except Exception as e:
        LOG.warn(_("Exception calling the HA Service to "
                   "enter maintenance mode: %s") % six.text_type(e))


def remove_host_from_maintenance_mode():
    """Helper method to tell EGO to remove the host from Maintenance Mode"""
    # Nothing to do if it is already in Maintenance Mode
    if not is_host_in_maintenance_mode():
        return
    try:
        # Call the HA Service to try to put the Host in Maintenance Mode
        hapolicy.API().set_maintenance(
            ctx.get_admin_context(), datacache[CONF.host], 'disable')
    except Exception as e:
        LOG.warn(_("Exception calling the HA Service to "
                   "exit maintenance mode: %s") % six.text_type(e))


def send_ui_notification(func_name=None, error_message=None,
                         success_message=None,
                         error_event_type=EVENT_TYPE_ERROR,
                         migrate_op=False, log_exception=False):
    """
    Send a notification to the GUI. If the decorated method throws an
    exception, an error notification will be sent, else a successful
    info notification will be sent. lpar name and uuid will be pulled from
    the instance object or dict parameter passed to the decorated method.

    Note: message string parameters (error_message and success_message) must be
          translated (wrappered with _() ). Replacement text variables
          must be surrounded with brackets.

          Replacement variables must be from the following list:'
                        {instance_id}
                        {instance_name}
                        {host_name}
                        {volume_id}
                        {source_host_name}
                        {target_host_name}
                        {error}

          Example: Stop of virtual machine {instance_name} on host {host_name}
                   failed with exception: {error}

    :param error_message: message string for exception notifications
                          (see note above)
    :param success_message: message string for success notifications
                            (see note above)
    :param error_event_type: either ERROR (EVENT_TYPE_ERROR)
                             or WARN (EVENT_TYPE_WARN)
    :param migrate_op: set to true for migation operations. Tells the decorator
                       to look for the migrate data dict. Only use when you
                       have replacement variables for source and target host.
    :param log_exception: if true in and error case, the event message is
                          also logged (LOG.exception)
    """
    def func_parms(f):
        @wraps(f)
        def wrapper(*args, **kwds):

            instance_dict_keys = ['uuid', 'display_name',
                                  'vm_state', 'task_state']
            migrate_dict = ['migrate_data']
            migrate_source_key = ['src_sys_name']
            migrate_target_key = ['dest_sys_name']

            def check_keys(arg, inst_keys):
                for key in inst_keys:
                    if key not in arg:
                        return False
                # Found them all!
                return True

            def check_inst(arg):
                # check if this is a Nova Instance object
                if isinstance(arg, instance_obj.Instance):
                    return True
                # check for a dict with all the right keys to match instance
                if isinstance(arg, dict) and check_keys(arg,
                                                        instance_dict_keys):
                    return True
                return False

            def check_migrate_source(arg):
                # check for a PowerVCLiveMigrateData object
                # and it has the source migration host
                if isinstance(arg, mig_obj.PowerVCLiveMigrateData) and (
                        check_keys(arg, migrate_source_key)):
                    return True
                return False

            def check_migrate_target(arg):
                # check for a PowerVCLiveMigrateData object
                # and it has the target migration host
                if isinstance(arg, mig_obj.PowerVCLiveMigrateData) and (
                        check_keys(arg, migrate_target_key)):
                    return True
                return False

            def check_migrate_target_embedded(arg):
                # check for an embedded migrate dict
                if isinstance(arg, dict) and check_keys(arg,
                                                        migrate_dict):
                    if isinstance(arg, dict) and (
                            check_keys(arg['migrate_data'],
                                       migrate_target_key)):
                        return True
                return False

            def _notify(exception_msg=None):
                payload = {}
                payload['msg'] = ''
                payload['instance_id'] = lpar_uuid
                payload['instance_name'] = lpar_name
                payload['host_name'] = encodeutils.safe_decode(
                    CONF.host_display_name, 'utf-8')
                if volume_id is not None:
                    payload['volume_id'] = volume_id
                notifier = rpc.get_notifier(service='compute',
                                            host=CONF.host)
                try:
                    if migrate_op:
                        payload['source_host_name'] = source_host
                        payload['target_host_name'] = target_host
                    if exception_msg is not None:
                        if error_message is not None:
                            payload['msg'] = error_message
                            payload['error'] = exception_msg
                            if log_exception:
                                try:
                                    # Convert the brackets to replacement
                                    # variables
                                    log_exception_msg = error_message.replace(
                                        "{", "%(").replace("}", ")s")
                                    LOG.exception(log_exception_msg % payload)
                                except Exception as exc:
                                    LOG.debug('LOG.exception failed with '
                                              'exception: %s' %
                                              six.text_type(exc))
                            if error_event_type is EVENT_TYPE_ERROR:
                                notifier.error(context,
                                               'compute.instance.log',
                                               payload)
                            elif error_event_type is EVENT_TYPE_WARN:
                                notifier.warn(context,
                                              'compute.instance.log',
                                              payload)
                            else:
                                LOG.debug('Exception was handled, but the '
                                          'error_event_type parameter was not '
                                          'set to a supported value. No GUI '
                                          'notification sent.')
                        else:
                            LOG.debug('Operation failed, but no error_message '
                                      'was provided. No GUI notification '
                                      'sent.')
                    else:
                        if success_message is not None:
                            payload['msg'] = success_message
                            notifier.info(context,
                                          'compute.instance.log',
                                          payload)
                        else:
                            LOG.debug('Operation completed successfully, but '
                                      'no success_message was provided. No GUI'
                                      ' notification sent.')
                except Exception as exc:
                    LOG.debug('Notify failed with exception: %s' %
                              six.text_type(exc))

            lpar_uuid = None
            lpar_name = None
            source_host = None
            target_host = None
            volume_id = None
            context = None
            for arg in args:
                if check_inst(arg):
                    lpar_uuid = arg['uuid']
                    lpar_name = arg['display_name']
                    source_host = arg['host']
                if migrate_op:
                    if check_migrate_source(arg):
                        source_host = arg.src_sys_name
                    if check_migrate_target(arg):
                        target_host = arg.dest_sys_name
                    elif check_migrate_target_embedded(arg):
                        migrate_data = arg['migrate_data']
                        target_host = migrate_data['dest_sys_name']
                # We want to use the user's context if given one
                if isinstance(arg, oslo_ctxt.RequestContext):
                    # Only do it on a VM delete for now to reduce the risk,
                    # since this is the only scenario where the instance
                    # might not exist at the time the notification is received
                    if six.text_type(func_name).find('terminate_inst') >= 0:
                        context = arg
                if (context is not None and
                    lpar_uuid is not None and
                        (not migrate_op or target_host is not None)):
                    break

            if (migrate_op and 'migrate_data' in kwds and
               kwds['migrate_data']):
                migrate_data = kwds['migrate_data']
                if check_migrate_source(migrate_data):
                    source_host = migrate_data.src_sys_name
                if check_migrate_target(migrate_data):
                    target_host = migrate_data.dest_sys_name

            if 'volume_id' in kwds:
                volume_id = kwds['volume_id']
            if 'bdm' in kwds and volume_id is None:
                if kwds['bdm'] is not None:
                    bdm = kwds['bdm']
                    volume_id = bdm['volume_id']
            if 'instance' in kwds:
                if check_inst(kwds['instance']):
                    if lpar_uuid is None:
                        lpar_uuid = kwds['instance']['uuid']
                    if lpar_name is None:
                        lpar_name = kwds['instance']['display_name']

            if lpar_uuid is None:
                LOG.debug('Did not find the Instance object in '
                          'the decorated method arguments')
                lpar_uuid = ''
                lpar_name = ''
            if migrate_op:
                if source_host is None:
                    source_host = ''
                    if 'instance' in kwds:
                        if check_inst(kwds['instance']):
                            source_host = kwds['instance']['host']
                if target_host is None:
                    LOG.debug('Did not find the migrate dict in '
                              'the decorated method arguments')
                    target_host = CONF.host
            if not context:
                context = ctx.get_admin_context()
            try:
                r = f(*args, **kwds)
                # notify success
                _notify()
                return r
            except Exception as exc:
                with excutils.save_and_reraise_exception():
                    # notify failure
                    # for exception raise from rpc call,
                    # it includes stack trace when converted
                    # via six.text_type
                    # So, we will extract message first
                    if hasattr(exc, 'message') and exc.message:
                        exception_msg = exc.message
                    else:
                        exception_msg = six.text_type(exc)
                    # if it's RemoteErrror, the message is updated
                    # with stack trace.
                    # extract from the value attribute instead
                    if isinstance(exc, RemoteError):
                        exception_msg = exc.value
                    if not exception_msg:
                        exception_msg = exc.__class__.__name__
                    _notify(exception_msg)

        return wrapper
    if func_name:
        return func_parms(func_name)
    else:
        return func_parms


def ui_notifier(notify_message, instance, except_to_raise=None,
                error_event_type=EVENT_TYPE_ERROR,
                log_exception=False):

    error_msg = None
    success_msg = None
    if except_to_raise is not None:
        error_msg = notify_message
    else:
        success_msg = notify_message

    @send_ui_notification(None, error_msg, success_msg,
                          error_event_type=error_event_type,
                          log_exception=log_exception)
    def _ui_notifier(instance, except_to_raise=None):
        if except_to_raise is not None:
            raise except_to_raise

    _ui_notifier(instance, except_to_raise=except_to_raise)


def _get_token_info(httpclient):
    """Helper method to check if the token we are going to use is expiring"""
    is_expiring, catalog = True, None
    kclient = keystone_client.Client(auth=httpclient.auth,
                                     session=httpclient.session,
                                     endpoint_type='internal')
    try:
        tok_data = kclient.tokens.get_token_data(httpclient.auth_token)
    except Exception as exc:
        LOG.exception(_("Error trying to get token data"))
    else:
        expiration = timeutils.parse_isotime(tok_data['token']['expires_at'])
        # Determine if the Token is within 15 minutes from Expiring
        is_expiring = not timeutils.is_newer_than(expiration, 900)
        if is_expiring:
            LOG.info(_("The periodic task token is about to expire"))
        # context expects catalog to be in v2 format
        catalog = _v3_to_v2_catalog(tok_data['token']['catalog'])
    return is_expiring, catalog


def populate_admin_context(context=None, context_service_catalog=None):
    """Helper method to populate the Token/ServiceCatalog on the Context"""
    if context is None:
        context = ctx.get_admin_context()
    # We don't need to do anything if the Token/ServiceCatalog are populated
    if context.auth_token and context.service_catalog:
        return context
    # We are using the Neutron Client since they cache the auth plugin, so the
    # token may already be cached, saving us from having to talk to keystone
    nclient = neutronv2_api.get_client(context, admin=True)
    # Determine whether the token is expiring and get the service catalog
    expiring, service_catalog = _get_token_info(nclient.httpclient)
    # Make sure the token isn't going to expire while executing
    if expiring:
        nclient.httpclient.authenticate()
        nclient.list_extensions()
    context.auth_token = neutronv2_api.\
        _ADMIN_AUTH.get_token(neutronv2_api._SESSION)
    # Cache the Service Catalog so that we can use it later
    if context_service_catalog is None:
        context_service_catalog = service_catalog
    context.service_catalog = context_service_catalog
    if context_service_catalog is None:
        LOG.info(_('No service_catalog found on network client: %s') %
                 nclient)
        return None
    return context


# To try to get new notification types to use the new version'ing topic,
# they started blocking all but a hardcoded list of event types to the
# the old topic (even though all notifications are sent still through
# the old topic), so we need to make a change to allow our types through
def _powervc_legacy_notify(self, priority, ctxt, event_type, payload):
    getattr(self.notifier, priority)(ctxt, event_type, payload)

rpc.LegacyValidatingNotifier._notify = _powervc_legacy_notify
