# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""PowerVC Extensions to the Compute Manager for On-boarding/Inventory"""

import uuid
from oslo_config import cfg
from oslo_log import log as logging
from oslo_concurrency import lockutils

from nova.objects import instance as instance_obj
from oslo_service import periodic_task

from powervc_nova import translate_message
from powervc_nova.virt import discovery_driver
from powervc_nova import _
from powervc_nova.virt.ibmpowervm.vif.common import ras

cfg.CONF.register_opts([
    cfg.IntOpt('inventory_host_interval', default=300,
               help='Interval to Refresh the Inventory for the Host'),
    cfg.IntOpt('quick_inventory_host_interval', default=60,
               help='Interval to do a quick check of host\'s VIOSes\' status'),
    cfg.IntOpt('inventory_instances_interval', default=120,
               help='Interval to Refresh the Inventory for Instances'),
])

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class PowerVCDiscoveryComputeManager(periodic_task.PeriodicTasks):
    """
    Extends the Compute Manager to provide the additional discovery-related
    capabilities that are beyond what the OpenStack Community provides.

    This extension provides the following capabilities/operations:
        1) inventory_host - collect host-level periodic inventory info
        2) inventory_instances - collect instance periodic inventory info
        3) discovery_instances - query the list of VM's for on-boarding
        4) query_instances - query detailed info on the VM's for on-boarding
        5) prepare_instances - prepare the VM's to be on-boarded
        6) finish_onboard - finish on-boarding for the VM after the import
        7) verify_host_running - verify that the host/service is running
        8) unmanage_host - hook to do driver cleanup on host removal
    """

    def __init__(self, *args, **kwargs):
        """Performs the Additional Discovery Initialization"""
        super(PowerVCDiscoveryComputeManager, self).__init__(CONF)
        # Add each of the Discovery Periodic Tasks to the sub-class
        for task in PowerVCDiscoveryComputeManager._periodic_tasks:
            self._periodic_tasks.append(task)
            self._periodic_spacing[task[0]] = task[1]._periodic_spacing
            self._periodic_last_run[task[0]] = task[1]._periodic_last_run

    ###############################################
    #        Inventory-related Extensions         #
    ###############################################
    @periodic_task.periodic_task(spacing=CONF.inventory_instances_interval)
    def inventory_instances(self, context, instances=None):
        """Periodic Task to Refresh the Inventory for the Instances"""
        driver, context = (self.driver, context.elevated())
        LOG.debug('Invoking Method to Refresh Instance Inventory - Begin')
        try:
            drv_insts = self.driver.list_instances()
            # Retrieve All of the Instances from the Nova DB for the Given Host
            db_insts = self._query_db_instances(context, instances)
            # If this is a Discovery Driver, then let it Gather Inventory
            if isinstance(driver, discovery_driver.ComputeDiscoveryDriver):
                # We will also allow the Driver to handle Evacuated Instances
                driver.handle_evacuated_instances(context, db_insts, drv_insts)
                # Now we can actually collect Inventory on the Instances
                driver.inventory_instances(context, db_insts)
        except Exception as exc:
            LOG.warn(_('Error refreshing Instance Inventory'))
            LOG.exception(exc)
        LOG.debug('Invoking Method to Refresh Instance Inventory - End')

    @periodic_task.periodic_task(spacing=CONF.inventory_host_interval)
    @lockutils.synchronized('powervc-inventory', 'host-inventory-')
    def inventory_host(self, context):
        """Periodic Task to Refresh the Inventory for the Host"""
        driver, context = (self.driver, context.elevated())
        try:
            # The Periodic Task Context may not have the Token/ServiceCatalog
            self._populate_admin_context(context)
            # If this is a Discovery Driver, then let it Gather Inventory
            if isinstance(driver, discovery_driver.ComputeDiscoveryDriver):
                driver.inventory_host(context)
        except Exception as exc:
            LOG.warn(_('Error refreshing Host Inventory'))
            LOG.exception(exc)

    @periodic_task.periodic_task(spacing=CONF.quick_inventory_host_interval)
    @lockutils.synchronized('powervc-inventory', 'host-inventory-')
    def quick_inventory_host(self, context):
        """Periodic Task to do a Quick Inventory Refresh for the Host"""
        driver, context = (self.driver, context.elevated())
        try:
            # The Periodic Task Context may not have the Token/ServiceCatalog
            self._populate_admin_context(context)
            # If this is a Discovery Driver, then let it Gather Inventory
            if isinstance(driver, discovery_driver.ComputeDiscoveryDriver):
                if driver.is_inventory_needed(context):
                    driver.inventory_host(context)
        except Exception as exc:
            LOG.exception(exc)

    ###############################################
    #         OnBoard-related Extensions          #
    ###############################################
    def discover_instances(self, context):
        """Returns a list of all of the VM's that exist on the Host"""
        driver = self.driver
        # Currently we won't throw an exception if it isn't a Discover Driver
        if not isinstance(driver, discovery_driver.ComputeDiscoveryDriver):
            drvclass = driver.__class__.__name__
            LOG.warn(_('Driver %s does not implement Discover Driver')
                     % drvclass)
            return {'identifier': None, 'servers': [], 'chunking': False}
        # Call the Compute Discovery Driver to get a list of existing VM's
        all_instances = self._discover_instances(context)
        # We need to modify the Instances returned from the Driver slightly
        self._manipulate_driver_instances(context, all_instances, False)
        # Break up the list of VM's to a set of Chunks to be returned
        identifier = self._persist_instances_chunks(all_instances)
        # Return the first chunk of persisted volumes to the caller
        return self.get_next_instances_chunk(context, identifier)

    def query_instances(self, context, instance_ids, allow_unsupported=False):
        """Returns details about the VM's that were requested"""
        return_all = '*all' in instance_ids
        match_instances, query_instances = (list(), list())
        # First we need to figure out what VM's really exist on the Host
        all_instances = self._discover_instances(context)
        uuid_map = dict([(inst_id, '') for inst_id in instance_ids])
        # Loop through each VM's on the Host, seeing if we should query
        for instance in all_instances:
            # See if they requested all instances or this specific one
            if return_all or instance['uuid'] in uuid_map:
                support = instance.get('support', {})
                supported = support.get('status', 'supported')
                # If this is managed, then no need to do a query, just return
                if instance.get('managed') is True:
                    match_instances.append(instance)
                # If this isn't supported and no override, just return it
                elif not allow_unsupported and supported == 'not_supported':
                    match_instances.append(instance)
                # Otherwise it is supported/un-managed and a match so query
                else:
                    instance.pop('support', instance.pop('managed', None))
                    query_instances.append(instance)
        # Only worth calling the Driver if some VM's exist on the System
        if len(query_instances) > 0:
            instances = self.driver.query_instances(context, query_instances)
            match_instances.extend(instances)
        # We need to modify the Instances returned from the Driver slightly
        self._manipulate_driver_instances(
            context, match_instances, True, all_instances)
        # Break up the list of VM's to a set of Chunks to be returned
        identifier = self._persist_instances_chunks(match_instances)
        # Return the first chunk of persisted VM's to the caller
        return self.get_next_instances_chunk(context, identifier)

    def prepare_instance(self, context, instance, preparation):
        """Prepare a VM for on-boarding by changing the VM configuration"""
        # We only want to call prepare on the Driver if it implements it
        if isinstance(self.driver, discovery_driver.ComputeDiscoveryDriver):
            LOG.info(_('Prepare on-board for instance %s') % instance['name'])
            instance = self.driver.prepare_instance(
                context, instance, preparation)
            # If the driver returned a valid instance, correct the values
            if instance:
                self._manipulate_driver_instances(context, [instance], True)
            return instance

    def finish_onboard(self, context, instance, preparation=None):
        """Finish on-boarding a VM after the VM is imported into PowerVC"""
        # We only want to call prepare on the Driver if it implements it
        if isinstance(self.driver, discovery_driver.ComputeDiscoveryDriver):
            LOG.info(_('Finish on-board for instance %s') % instance['name'])
            instance = instance_obj.Instance.get_by_uuid(
                context, instance['uuid'])
            # Get the Network Information from the Database for the VM
            network_info = self.network_api.\
                get_instance_nw_info(context, instance)
            # Get the Block Device Mapping from the Database for the VM
            bdm_info = self._get_instance_block_device_info(context, instance)
            # Invoke the Driver to finish on-boarding the VM
            self.driver.finish_onboard(
                context, instance, preparation, network_info, bdm_info)

    ###############################################
    #     Service LifeCycle-related Extensions    #
    ###############################################
    def verify_host_running(self, context):
        """Verifies the nova-compute service for the Host is running"""
        return True

    def unmanage_host(self, context):
        """Allows the Driver to do any necessary cleanup on Host Removal"""
        # We only want to call unManage on the Driver if it implements it
        if isinstance(self.driver, discovery_driver.ComputeDiscoveryDriver):
            LOG.info(_('Unmanaging Host, allowing Driver to do cleanup...'))
            self.driver.unmanage_host(context)

    ###############################################
    #       Shared Discovery Helper Methods       #
    ###############################################
    @lockutils.synchronized('instance_chunking', 'nova-')
    def get_next_instances_chunk(self, context, identifier):
        """Provides Chunking of VM's Lists to avoid QPID Limits"""
        instance_chunks = self.instance_chunk_map.get(identifier)
        # If the Identifier doesn't exist, we will just return an empty list
        if instance_chunks is None:
            return dict(identifier=identifier, servers=[], chunking=False)
        # If this is the last chunk (or no chunks), just return that list
        if len(instance_chunks) == 1:
            self.instance_chunk_map.pop(identifier, None)
            return dict(identifier=identifier,
                        servers=instance_chunks[0], chunking=False)
        # Otherwise return the first chunk and say that there are more left
        self.instance_chunk_map[identifier] = instance_chunks[1:]
        return dict(identifier=identifier,
                    servers=instance_chunks[0], chunking=True)

    def _discover_instances(self, context):
        """Internal Method to list of all of the VM's that exist on the Host"""
        # Call the Compute Discovery Driver to get a list of existing VM's
        drv_instances = self.driver.discover_instances(context)
        # Collect a list of all neutron ports
        ports = self.network_api.list_ports(context)
        ports = ports.pop('ports', [])
        ports = dict([(p['mac_address'].lower(), p) for p in ports])
        for instance in drv_instances:
            adapters = instance['adapters']
            for adapter in adapters:
                maca = adapter['mac_address']
                # Update the mac address format
                if len(maca) == 12:
                    maca = ':'.join(a + b for a, b in zip(maca[::2],
                                                          maca[1::2]))
                port = ports.get(maca.lower())
                if port:
                    tenant = port['tenant_id']
                    if not tenant == context.project_id:
                        instance['support']['reasons'] = \
                            [ras.msg('info', 'IMPORT_TENANT_MISMATCH')]
                        instance['support']['status'] = 'not_supported'
        # Generate the UUID's for the VM's and determine which are Managed
        self._generate_instance_uuids(context, drv_instances)
        return drv_instances

    @lockutils.synchronized('instance_chunking', 'nova-')
    def _persist_instances_chunks(self, instances):
        """Internal Helper method to generate the UUID's for the VM's"""
        current_len = 0
        current_list, instance_chunks = (list(), list())
        # First get an identifier to be used to reference this chunking
        identifier = str(self.instance_chunk_counter)
        self.instance_chunk_counter = self.instance_chunk_counter + 1
        # Loop through each instance, breaking it up into chunks based on size
        while len(instances) > 0:
            instance = instances.pop(0)
            instance_len = len(str(instance))
            # If we could possibly go over the 64K Message size, break up
            if len(current_list) > 0 and (current_len + instance_len) > 40000:
                instance_chunks.append(current_list)
                current_list = []
                current_len = 0
            # Add this instance to the current chunk that is being returned
            current_len = current_len + instance_len
            current_list.append(instance)
        # Add the final chunk to the overall set of chunks for the instance
        instance_chunks.append(current_list)
        self.instance_chunk_map[identifier] = instance_chunks
        return identifier

    def _manipulate_driver_instances(self, context,
                                     drv_instances, query, discinsts=[]):
        """Internal Helper method to modify attributes on the Instances"""
        state_map = {'1': 'active', '3': 'paused', '4': 'stopped',
                     '6': 'error', '7': 'suspended', '9': 'building'}
        name_map = dict([(inst['name'], inst) for inst in discinsts])
        # Loop through each of the Driver Instances, modifying the data
        for inst in drv_instances:
            # Get some things off the discover call if query didn't return them
            disc_inst = name_map.get(inst['name'], {})
            state = inst.pop('state', disc_inst.get('state', 0))
            inst['id'] = inst.pop('uuid', disc_inst.get('uuid'))
            # If this is the result of a Driver Query Instances call
            if query:
                inst['power_state'] = state
                inst['vm_state'] = state_map.get(str(state), 'error')
                inst['memory_mb'] = int(inst.get('memory_mb', 0))
                inst['root_gb'] = int(inst.get('root_gb', 0))
                inst['ephemeral_gb'] = int(inst.get('ephemeral_gb', 0))
                vcpus = str(inst.get('vcpus', 0))
                if vcpus.find('.') > 0:
                    vcpus = vcpus[:vcpus.find('.')]
                inst['vcpus'] = int(vcpus)
            # If this is the result of a Driver Discover Instances call
            else:
                inst['status'] = state_map.get(str(state), 'error')
                inst.pop('power_specs', None)
            # Need to go through the reasons as convert to the Browser Locale
            if inst.get('support') is not None:
                if inst['support'].get('reasons') is not None:
                    reasons = inst['support']['reasons']
                    reasons = [translate_message(context, r) for r in reasons]
                    inst['support']['reasons'] = reasons
                # Also need to convert the Preparation Actions over
                if inst['support'].get('steps') is not None:
                    for step in inst['support']['steps']:
                        msgs = step.get('actions', [])
                        msgs = [translate_message(context, m) for m in msgs]
                        if step.get('actions') is not None:
                            step['actions'] = msgs

    def _generate_instance_uuids(self, context, drv_instances):
        """Internal Helper method to generate the UUID's for the VM's"""
        name_map, uuid_map = (dict(), dict())
        vm_name_map, vm_uuid_map = (dict(), dict())
        # Query the Database to get the Instance info for this Host
        db_instances = self._query_db_instances(
            context, project_deletion_cleanup=True)
        # Loop through each of the Instances, creating the Maps for Lookup
        for db_instance in db_instances:
            power_specs = db_instance.get('power_specs', {})
            vm_uuid = power_specs.get('vm_uuid')
            vm_name = power_specs.get('vm_name')
            vm_uuid = '!' if vm_uuid is None else vm_uuid
            vm_name = '!' if vm_name is None else vm_name
            uuid_map[db_instance['uuid']] = db_instance
            name_map[db_instance.get('name', '!')] = db_instance
            vm_uuid_map[vm_uuid.lower()] = db_instance
            vm_name_map[vm_name] = db_instance
        # Cleanup any default invalid entries we added to the Maps
        name_map.pop('!', None)
        vm_name_map.pop('!', None)
        vm_uuid_map.pop('!', None)
        # Loop through the instances generating UUID's for any that need one
        for drv_vm in drv_instances:
            managed = False
            db_instance = uuid_map.get(drv_vm.get('uuid', '!'))
            power_specs = drv_vm.get('power_specs', {})
            vm_uuid = power_specs.get('vm_uuid')
            vm_name = power_specs.get('vm_name')
            # See if there was a match on the Name with what is in the DB
            if db_instance is None and vm_name is not None:
                db_instance = name_map.get(vm_name)
                if db_instance is None:
                    db_instance = vm_name_map.get(vm_name)
            # See if there was a match on the UUID with what is in the DB
            if db_instance is None and vm_uuid is not None:
                db_instance = uuid_map.get(vm_uuid.lower())
                if db_instance is None:
                    db_instance = vm_uuid_map.get(vm_uuid.lower())
            # If the UUID isn't set, do a secondary lookup to see if it exists
            if db_instance is not None:
                drv_vm['uuid'] = db_instance['uuid']
                drv_vm['name'] = db_instance['display_name']
                managed = True
            # Otherwise it isn't managed, so we need to generate one
            elif not drv_vm.get('uuid'):
                namesp = uuid.UUID('a0dd4880-6115-39d6-b26b-77df18fe749f')
                namestr = self._get_unique_instance_str(drv_vm)
                drv_vm['uuid'] = str(uuid.uuid3(namesp, namestr))
            # Set the Managed attribute based on what we determined earlier
            drv_vm['managed'] = managed

    def _get_unique_instance_str(self, instance):
        """Internal Helper method to create a unique string for the instance"""
        instance_name = instance.get('name', '')
        instance_id = instance.get('hostuuid', '')
        return "ibm-powervc://server/host='%s',id='%s',name='%s'" % \
            (self.host, str(instance_name), str(instance_id))
