# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""PowerVC Extensions to the Compute Manager for HMC-specific features"""

import copy
import six
import sys
from nova import objects
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import periodic_task
from oslo_serialization import jsonutils
from oslo_utils import excutils

from powervc_nova.compute import kvmhmc_manager
from powervc_nova.compute import manager
from powervc_nova.compute import notify_messages as notify_msg
from powervc_nova.compute import exception as pwrvc_excep
from powervc_nova.objects.compute import dom as compute_dom
from powervc_nova import _

LOG = logging.getLogger(__name__)
cfg.CONF.import_opt('powervm_mgr', 'powervc_nova.virt.ibmpowervm')

# order of decorators should be same order  in which they
# appear if applied statically
CUSTOM_DECORATORS = {
    'rebuild_instance':
    [manager.declare_decor(
        'powervc_nova.compute.send_ui_notification',
        notify_msg.EVACUATE_ERROR, notify_msg.EVACUATE_SUCCESS,
        log_exception=True)]
}


class PowerVCHmcComputeManager(kvmhmc_manager.PowerVCKvmHmcComputeManager):
    """
    Extends the Compute Manager to provide the additional HMC-specific
    capabilities that are beyond what the OpenStack Community provides.
    """
    @classmethod
    def get_decorators(cls, decorators):
        new_decorators = copy.deepcopy(CUSTOM_DECORATORS)
        new_decorators.update(decorators)
        return super(PowerVCHmcComputeManager, cls).\
            get_decorators(new_decorators)

    def __init__(self, *args, **kwargs):
        """Performs the Additional HMC-specific Manager Initialization"""
        super(PowerVCHmcComputeManager, self).__init__(*args, **kwargs)

    #######################################################
    # **** HMC-specific Manager Overrides/Extensions **** #
    #######################################################
    def end_compute_with_hmc(self, context, data):
        LOG.debug('Received end_compute_with_hmc cast %s' % data)
        hmc_ip = data['hmc_ip']
        if hmc_ip == cfg.CONF.powervm_mgr:
            # this is needed to trigger the health monitoring
            LOG.error(_('Received 401 unauthorized error from HMC %s')
                      % hmc_ip)
            sys.exit()

    def sync_enterprise_pool_for_host(self, context):
        try:
            return self.driver.sync_enterprise_pool_for_host(context)
        except Exception as exc:
            with excutils.save_and_reraise_exception():
                LOG.error(_('Failed to sync enterprise pool for host.'))
                LOG.exception(exc)

    @lockutils.synchronized('set_mobile_cores', 'powervc-')
    def set_mobile_cores(self, context, desired_mobile_cores):
        try:
            return self.driver.set_mobile_cores(context, desired_mobile_cores)
        except Exception as exc:
            with excutils.save_and_reraise_exception():
                LOG.error(_('Error setting mobile cores for host to %s.') %
                          desired_mobile_cores)
                LOG.exception(exc)

    def _prep_block_device_volume_attach(self, context, instance, virt_bdms,
                                         do_check_attach):
        # When deploys fail on a Novalink host and are rescheduled to an HMC
        # managed host, the BDMs may have connection info set. This needs to
        # be blanked to avoid issues downstream.
        for bdm in virt_bdms:
            if bdm['connection_info']:
                bdm['connection_info'] = ''
                bdm.save()

        # We override the parent class here because for the HMC virt driver
        # we do not wait for the volumes to finish being created.  The
        # creation happens in parallel with the rest of the deploy and
        # the attach is accomplished at the end of the deploy with
        # compute manager callbacks.
        blkdev_info = self._get_instance_block_device_info(context,
                                                           instance)
        # extend the block_device_info with call method for HMC driver
        # to call back boot volume attach during spawn.
        blkdev_info['powervc_blockdev_callbacks'] = {
            'validate_volumes': self._validate_volumes,
            'attach_volumes': self._attach_volumes_and_boot,
            'validate_attachment': self._validate_attachment}
        return blkdev_info

    @periodic_task.periodic_task(spacing=3600, run_immediately=True)
    def _populate_hmc_mtmsn_values(self, context):
        """
        Fetch the HMC MTMSN (machine type, model, serial number) for any HMCs
        that do not have an MTMSN value in the DB, which is possible if the
        HMC existed in the DB before an upgrade to PowerVC 1.3
        """
        # NOTE(tpeoples): if you refactor this method name, then you must also
        # change the variable below so that the periodic task gets removed
        method_name = '_populate_hmc_mtmsn_values'

        LOG.debug('Running _populate_hmc_mtmsn_values periodic task.')

        hmc_fact = compute_dom.ManagementConsoleFactory.get_factory()
        hmcs = hmc_fact.find_all_hmcs(context)

        hmcs_without_mtmsn = [hmc for hmc in hmcs if not hmc.mtmsn]

        if not hmcs_without_mtmsn:
            LOG.debug('There are no HMCs in the database with a null '
                      'MTMSN (machine type, model, serial number) value. '
                      'Therefore the _populate_hmc_mtmsn_values periodic '
                      'task is being removed.')
            for task_name, task in self._periodic_tasks:
                if task_name == method_name:
                    LOG.debug('Removing defined periodic task: %s' %
                              task_name)
                    self._periodic_tasks.remove((task_name, task))
                    return

        for hmc in hmcs_without_mtmsn:
            try:
                LOG.info(_('Attempting to fetch the MTMSN value for HMC at IP '
                           '%(hmc_ip)s because it is currently unknown.') %
                         {'hmc_ip': hmc.access_ip})
                mtmsn = self.driver.fetch_hmc_mtmsn(hmc.access_ip,
                                                    hmc.user_id,
                                                    hmc.password)
                LOG.info(_('MTMSN value for HMC at IP %(hmc_ip)s found to be: '
                           '%(mtmsn)s.') %
                         {'hmc_ip': hmc.access_ip,
                          'mtmsn': mtmsn})
                if mtmsn:
                    hmc.mtmsn = mtmsn
                    hmc.save()
            except Exception as e:
                LOG.error(_('Error while trying to fetch unknown MTMSN '
                            'value for existing HMC at IP %s or while '
                            'saving HMC object to the database.') %
                          hmc.access_ip)
                LOG.exception(e)

    def _detach_volume(self, context, volume_id, instance, destroy_bdm=True,
                       attachment_id=None):
        """
        detach a volume from an instance, prevent any boot volume
        from being detached.
        :param context: operation context
        :param volume_id: volume id for detaching
        :param instance: compute instance for detaching.
        :return: None
        """
        # get particular bdm for the instance
        bdm = objects.BlockDeviceMapping.get_by_volume_and_instance(
            context, volume_id, instance['uuid'])
        if cfg.CONF.volume_usage_poll_interval > 0:
            vol_stats = []
            mp = bdm.device_name
            # Handle bootable volumes which will not contain /dev/
            if '/dev/' in mp:
                mp = mp[5:]
            try:
                vol_stats = self.driver.block_stats(instance, mp)
            except NotImplementedError:
                pass

            if vol_stats:
                LOG.debug("Updating volume usage cache with totals",
                          instance=instance)
                rd_req, rd_bytes, wr_req, wr_bytes, flush_ops = vol_stats
                self.conductor_api.vol_usage_update(context, volume_id,
                                                    rd_req, rd_bytes,
                                                    wr_req, wr_bytes,
                                                    instance,
                                                    update_totals=True)

        self._driver_detach_volume(context, instance, bdm)
        # get a default SPP connector
        connector = self._get_SSP_volume_connector(instance, volume_id,
                                                   context)

        # Get a real FC connector
        if not connector:
            connector = self._get_bdm_volume_connector(context,
                                                       instance,
                                                       bdm,
                                                       volume_id)
        # HMC managed system and has no connector, try to generate
        # it from the saved connection_info.
        if not connector:
            connector = self._construct_connector_from_conninfo(instance,
                                                                bdm)
        # Last try by called the virt driver. Could be expensive and
        # have exception.
        if not connector:
            try:
                connector = self.driver.get_volume_connector(instance)
            except Exception as ex:
                LOG.error(_("Failed to detach volume %(volid)s from "
                            "instance %(inst)s due to null "
                            "volume connector. Error: %(error)s") %
                          {'volid': volume_id,
                           'inst': instance['uuid'],
                           'error': ex})
                err_msg = {'volid': volume_id,
                           'inst': instance['uuid']}
                raise pwrvc_excep.IBMPowerVCVInvalidVolumeConnector(
                    **err_msg)
        LOG.info(_("Instance %(inst)s detach volume %(volid)s with connector: "
                   "%(conn)s ") %
                 {'inst': instance['uuid'],
                  'volid': volume_id,
                  'conn': connector})
        try:
            # For Onboarded VM: the connector returned from the
            # driver.get_volume_connector does not have connection_type
            # set. Need to check if wwpn_vscsi is present
            self.volume_api.detach(context.elevated(), volume_id,
                                   attachment_id=attachment_id)

            conninfo = jsonutils.loads(bdm['connection_info'])
            self.driver.call_terminate_connection(context, instance,
                                                  bdm['volume_id'],
                                                  connector, conninfo)
        except Exception:
            with excutils.save_and_reraise_exception():
                # Roll back detaching state.
                # Otherwise, volume stucks in detaching state forever.
                self.volume_api.roll_detaching(context, volume_id)

        if destroy_bdm:
            bdm.destroy()

        info = dict(volume_id=volume_id)
        self._notify_about_instance_usage(
            context, instance, "volume.detach", extra_usage_info=info)
