#
#
# Copyright 2015, 2016 IBM Corp.
# All Rights Reserved.
# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2011 Justin Santa Barbara
# Copyright 2011 Piston Cloud Computing, Inc.
# Copyright 2012-2013 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from eventlet import greenthread, GreenPool
import functools
import inspect
import six
import sys
import time
import traceback
import importlib

from keystoneauth1.identity import v3 as auth_v3
from keystoneclient import session as ks_session
from keystoneclient.v3 import client as keystone_client

from nova.compute import manager
from nova import block_device
import nova.context
from nova.compute import power_state
from nova.compute import vm_states
from nova.compute.manager import reverts_task_state
from nova import exception
from nova.image import glance
from nova import rpc
from nova import safe_utils
from nova import utils
from nova.network import neutronv2
from oslo_service import periodic_task
from nova.virt import block_device as blk_dev
from nova.virt import driver
from nova.volume import cinder as cinderv2
from nova import volume
from nova import objects
from nova.objects import block_device as blkdev_obj

from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils

from powervc_nova.common import periodic_task_watchdog
from powervc_nova.common import Monkey_FcntlLock
from powervc_nova.common import update_canary_run_time
from powervc_nova.common import volume_type_utils
from powervc_nova.compute import common_constants
from powervc_nova.compute import discovery_manager
from powervc_nova.compute import exception as pwrvc_excep
from powervc_nova.compute import resource_tracker
from powervc_nova.compute import send_ui_notification
from powervc_nova.compute import populate_admin_context
from powervc_nova.compute import check_host_maintenance_mode
from powervc_nova.compute import notify_messages as notify_msg
from powervc_nova.compute import api as pvc_compute_rpcapi
from powervc_nova.conductor import api as conductor
from powervc_nova.objects.compute import dom as compute_dom
from powervc_nova.virt import discovery_driver
from powervc_nova.virt import netutils
from powervc_nova.virt.ibmpowervm.common import exception as pvm_exc
from powervc_nova.virt.ibmpowervm.common import migrate_utils
from powervc_nova.compute import metric_cache
from powervc_nova import _
from powervc_nova.objects import migrate_data as mig_obj
from powervc_oslo.config import data_utils as oslo_utils

LOG = logging.getLogger(__name__)
NOVA_COMP_MGR_FILE = inspect.getsourcefile(manager.ComputeManager)
_KEYSTONE_SESSION = None
METRICS_ALLOWED_VARIATION = {'util': 0.5, 'page_rate': 10}

get_notifier = functools.partial(rpc.get_notifier, service='compute')
wrap_exception = functools.partial(exception.wrap_exception,
                                   get_notifier=get_notifier)

# List of periodic tasks to remove. This will remove all occurrences
# of the name.  It's possible to have periodic tasks with the same name but
# it should not be done.  If overriding a periodic task, prefix the name with
# 'pvc' or something similar to create a unique name.
# This is not needed at the moment since we got a fix from OpenStack but
# good to keep around incase we need it in the future.
PERIODIC_TASKS_TO_REMOVE = ['update_available_resource', '_sync_power_states']
RUN_PERIODIC_TASK_ASYNC = ['pvc_update_available_resource',
                           'inventory_instances',
                           'monitor_instances',
                           '_check_instance_vopt_removal',
                           '_pvc_sync_power_states'
                           '_poll_unconfirmed_resizes']

ibm_interval_opts = [
    cfg.StrOpt('boot_volume_name_prefix', default='boot-',
               help='Prefix for instance boot set volumes.'),
    cfg.StrOpt('data_volume_name_prefix', default='data-',
               help='Prefix for instance data volumes.'),
    cfg.IntOpt('periodic_task_runtime_msg_threshold', default=300,
               help='The threshold (seconds) before messages are'
                    ' logged for periodic tasks running longer than the'
                    ' threshold.'),
    cfg.IntOpt('periodic_task_completion_msg_threshold', default=90,
               help='The threshold (seconds) before messages are'
                    ' logged for periodic tasks that ran longer to complete'
                    ' than the threshold.'),
    cfg.IntOpt('monitor_instances_interval', default=300,
               help='Interval to Refresh the Metrics for Instances'),
    cfg.StrOpt('host_display_name', default=None,
               help='The display name set by the user for the Host'),
    cfg.IntOpt('volume_status_check_max_retries', default=300,
               help='The maximum number of volume status checks'
               ' during volume creation.'),
    cfg.IntOpt('volume_status_check_interval', default=6,
               help='Interval between status checks'
               ' during volume creation.'),
    cfg.IntOpt('verify_mgmt_sys_max_retry_count', default=5,
               help='Retry count for verify management system'),
    cfg.IntOpt('verify_mgmt_sys_sleep_time_during_retry', default=20,
               help='Sleep time while retrying verify'
               'management system'),
    cfg.BoolOpt('has_first_boot_occurred', default=False,
                help='Indicates if the agent has run through the first'
                     ' boot setup'),
]

CONF = cfg.CONF
CONF.register_opts(ibm_interval_opts)
CONF.import_opt(
    'auth_uri', 'keystonemiddleware.auth_token.__init__', 'keystone_authtoken')


@utils.expects_func_args('instance')
def pvc_comp_mgr_reverts_task_state(function):
    """
    IMPORTANT: This can only be used by PowerVCComputeManager.
    It has specific checking for whether an exception has
    raised from super.function to determine task_state has
    been reset already.
    Decorator to revert task_state on failure.
    """

    @functools.wraps(function)
    def decorated_function(self, context, *args, **kwargs):
        try:
            return function(self, context, *args, **kwargs)
        except exception.UnexpectedTaskStateError as e:
            # Note(maoy): unexpected task state means the current
            # task is preempted. Do not clear task state in this
            # case.
            with excutils.save_and_reraise_exception():
                LOG.info(_("Task possibly preempted: %s"),
                         e.format_message())
        except Exception:
            with excutils.save_and_reraise_exception():
                exc_raised_from_super = False
                try:
                    # check whether the exception is raise from super
                    exc_type, exc_value, exc_traceback = sys.exc_info()
                    tb_list = traceback.extract_tb(exc_traceback, 30)
                    for (filename, line, func_name, text) in tb_list:
                        if (function.__name__ == func_name and
                                NOVA_COMP_MGR_FILE == filename):
                            exc_raised_from_super = True
                            break
                except Exception as ex:
                    LOG.exception(ex)
                try:
                    if not exc_raised_from_super:
                        wrapped_func = utils.get_wrapped_function(function)
                        keyed_args = safe_utils.getcallargs(wrapped_func,
                                                            context,
                                                            *args, **kwargs)
                        # NOTE(mriedem): 'instance' must be in keyed_args
                        # because we
                        # have utils.expects_func_args('instance') decorating
                        # this method.
                        self._instance_update(context, keyed_args['instance'],
                                              task_state=None)
                    else:
                        LOG.info(_('The task_state value has been reverted in '
                                   'the parent method %s.')
                                 % function.__name__)
                except Exception as save_exc:
                    LOG.exception(save_exc)
                    pass

    return decorated_function


def declare_decor(dec_name, *args, **kwargs):
    return {dec_name: dict(args=list(args), kwargs=kwargs)}

# Do not add decorators which are taking self as wrapper args
# Do not add wrap_exception decorator as it may not work
# with dynamic decorators
PVC_DECORATORS = {
    'stop_instance':
    [declare_decor('powervc_nova.compute.send_ui_notification',
                   notify_msg.STOP_ERROR, notify_msg.STOP_SUCCESS,
                   log_exception=True)],
    'terminate_instance':
    [declare_decor('powervc_nova.compute.send_ui_notification',
                   notify_msg.DELETE_ERROR, notify_msg.DELETE_SUCCESS,
                   log_exception=True)],
    '_build_and_run_instance':
    [declare_decor('powervc_nova.compute.send_ui_notification',
                   notify_msg.DEPLOY_ERROR,
                   notify_msg.DEPLOY_SUCCESS,
                   log_exception=True)]}


class PowerVCComputeManager(manager.ComputeManager,
                            discovery_manager.PowerVCDiscoveryComputeManager):
    """Extends the base Compute Manager class with PowerVC capabilities"""

    def monkey_patch_decorators(self, decorators):
        LOG.info(_("Adding decorators to manager"))
        for method_name in decorators:
            try:
                method = getattr(self, method_name)
                decorator_appended = None
                for decoratordict in decorators[method_name]:
                    decorator_method = None
                    decorator = decoratordict.keys()[0]
                    params = decoratordict[decoratordict.keys()[0]]
                    decorator_name = None
                    module_decorator = decorator.rsplit('.', 1)
                    decorator_name = module_decorator[1]
                    module_name = module_decorator[0]
                    decorator_method = getattr(
                        importlib.import_module(module_name),
                        decorator_name)
                    if decorator_appended:
                        decorator_appended = decorator_method(
                            decorator_appended, *params['args'],
                            **params['kwargs'])
                    else:
                        decorator_appended = decorator_method(
                            method, *params['args'], **params['kwargs'])
                setattr(self, method_name, decorator_appended)
            except Exception as ex:
                print ex
                LOG.exception(_("Exception occured, while adding "
                                "decorators to manager"))
        LOG.info(_("Decorators added successfully"))

    @classmethod
    def get_decorators(cls, decorators):
        all_decorators = copy.deepcopy(PVC_DECORATORS)
        all_decorators.update(decorators)
        return all_decorators

    def __init__(self, compute_driver=None, *args, **kwargs):
        """Constructor for the PowerVC extension to the Compute Manager"""
        try:
            self.monkey_patch_decorators(self.get_decorators({}))
            # Remove any unwanted periodic tasks before initializing our
            # parent class, since it will start running tasks
            self._undefine_periodic_tasks()
            # Wrap the periodic tasks so we can log and time them
            self._init_ptask_nanny()

            super(PowerVCComputeManager, self).__init__(
                compute_driver='nova.virt.fake.FakeDriver', *args, **kwargs)
            # Quickly replace parents pool size.  Default is 1000 which could
            # overwhelm an HMC.
            self._sync_power_pool = GreenPool(size=10)
            # Since we need to add methods to the VirtAPI so the driver can
            # call other DB methods, we first construct a temporary driver so
            # that we can override the VirtAPI before we construct the
            # real driver
            self.compute_rpcapi = pvc_compute_rpcapi.PowerVCComputeRPCAPI()
            self.virtapi = manager.ComputeVirtAPI(self)
            self.conductor_api = conductor.PowerVCConductorAPI()
            self.volume_api = volume.API()
            self.driver = driver.\
                load_compute_driver(self.virtapi, compute_driver)

            # calculate the following flag
            # after actual driver is loaded
            self.use_legacy_block_device_info = \
                self.driver.need_legacy_block_device_info

            self.instance_chunk_map = dict()
            self.instance_chunk_counter = 0
            self.context_service_catalog = None

            self._periodic_task_pool = GreenPool(size=10)
            self._periodic_tasks_running = {}

            # Monkey patches:
            netutils.monkey_patch_netutils_get_injected_network_template()

            self.vm_util_cache = metric_cache.MetricCache()

        except Exception as exc:
            LOG.exception(_('error in ComputeMgr init, ending nova-compute'))
            raise exc

    #######################################################
    #           Override Initialization Methods           #
    #######################################################
    def pre_start_hook(self, **kwargs):
        """Override the Pre-Start Hook to fix Base Compute Manager issues"""
        # Call the Parent pre-start hook to create the ComputeNode in the DB
        super(PowerVCComputeManager, self).pre_start_hook(**kwargs)
        # After the Compute Node is inserted in the DB, then Inventory the Host
        if isinstance(self.driver, discovery_driver.ComputeDiscoveryDriver):
            self.inventory_host(nova.context.get_admin_context())
            self.inventory_instances(nova.context.get_admin_context())

    def init_host(self):
        """Override the Init Host to fix Base Compute Manager issues"""
        try:
            super(PowerVCComputeManager, self).init_host()
            self._verify_mgmt_system_access(nova.context.get_admin_context())
            self.check_and_run_first_boot(nova.context.get_admin_context())
        except Exception:
            LOG.exception(_('ComputeMgr init_host error, ending nova-compute'))
            raise

    def _get_resource_tracker(self, nodename):
        """Override to return our own Resource Tracker implementation"""
        rt = self._resource_tracker_dict.get(nodename)
        if not rt:
            if nodename not in self.driver.get_available_nodes():
                msg = nodename + ' is not a valid node managed by the host.'
                raise exception.NovaException(msg)
            rt = resource_tracker.PowerVCResourceTracker(
                self.host, self.driver, nodename)
            self._resource_tracker_dict[nodename] = rt
        return rt

    def _query_db_instances(self, context, filter_instances=None,
                            project_deletion_cleanup=False):
        """Helper method to filter out the subset of the DB Instances"""
        filtered_instances = list()
        context = context.elevated()
        instfact = compute_dom.ManagedInstanceFactory.get_factory()
        # Query all of the Instance for the Host from the Database
        db_instances = instfact.find_all_instances_by_host(context, self.host)
        # If they asked us to cleanup volumes for deleted projects, do so now
        if project_deletion_cleanup:
            db_instances = self._cleanup_instances_for_deleted_projects(
                context, db_instances)
        # If they didn't provide any Filter criteria, just return all
        if not filter_instances:
            return db_instances
        # Build a Map of UUID's for the Instances, for quicker lookup
        filter_map = dict([(instance['uuid'],
                            instance) for instance in filter_instances])
        # Loop through each of the DB Instances, seeing if there is a match
        for db_instance in db_instances:
            if db_instance['uuid'] in filter_map:
                filtered_instances.append(db_instance)
        return filtered_instances

    def _cleanup_instances_for_deleted_projects(self, context, instances):
        """Internal Helper method to cleanup instances for deleted projects"""
        proj_dict, ret_insts, del_insts = dict(), list(), list()
        # Group the volumes together by the projects they are a part of
        for inst in instances:
            proj_dict.setdefault(inst['project_id'], []).append(inst)
        # We know the project exists if they are all the context's project
        if not proj_dict or (len(proj_dict) == 1 and
                             proj_dict.items()[0][0] == context.project_id):
            return instances
        try:
            projects = self._get_all_projects(context)
            # Loop through the list of projects seeing which still exist
            for proj, proj_insts in proj_dict.iteritems():
                if proj in projects:
                    ret_insts.extend(proj_insts)
                # If the project doesn't exist, then need to remove them
                else:
                    LOG.info(_("Project %s deleted, removing vms") % proj)
                    del_insts.extend([v['uuid'] for v in proj_insts])
            # If any are for deleted projects, we need to unmanage them now
            if len(del_insts) > 0:
                self._unmanage_instances(context, del_insts)
            return ret_insts
        # If we got an exception, have to just go with the instances as-is
        except Exception as exc:
            LOG.warn(_("Error cleaning up deleted project vms: %s") % exc)
            LOG.exception(exc)
            return instances

    def _get_all_projects(self, context):
        """Internal Helper method to query Keystone for all projects"""
        global _KEYSTONE_SESSION
        # Reuse the same session so we don't connect every time
        if _KEYSTONE_SESSION is None:
            _KEYSTONE_SESSION = ks_session.Session(
                verify='/etc/pki/tls/certs/powervc.crt')
        # Construct a Token Authentication Wrapper for the Auth Token
        auth = auth_v3.Token(cfg.CONF.keystone_authtoken.auth_uri,
                             context.auth_token, project_id=context.project_id)
        # Use the Keystone Client to get a list of Project ID's to return
        client = keystone_client.Client(auth=auth, session=_KEYSTONE_SESSION)
        return [p.id for p in client.projects.list()]

    def _unmanage_instances(self, context, inst_uuids):
        """Internal Helper method to unmanage all of the vm's specified"""
        # Use the conductor to do unmanage since the code is on the Mgmt Server
        self.conductor_api.unmanage_instances(context, self.host, inst_uuids)

    def _populate_admin_context(self, context):
        """Helper method to populate the Token/ServiceCatalog on the Context"""
        tmp_context = populate_admin_context(
            context, self.context_service_catalog)
        if tmp_context:
            self.context_service_catalog = tmp_context.service_catalog

    def _retry_operation(self, function, context,
                         max_retry_count, sleep_time):
        for x in range(max_retry_count):
            try:
                function(context)
                # if call succeeds then
                # break the loop
                break
            # Log the exception and re-throw
            # one we can parse in the Registration
            except Exception as e:
                with excutils.save_and_reraise_exception() as ctxt:
                    if x < max_retry_count - 1:
                        ctxt.reraise = False
                        greenthread.sleep(sleep_time)
                        LOG.info(_("Exception %s occurred. "
                                   "Retry in progress...") %
                                 six.text_type(e))
                    else:
                        LOG.exception(_('Error trying to connect to the '
                                        'Management System %s')
                                      % six.text_type(e))
                        raise Exception(_('Failed to connect to the '
                                        'Management System'))

    @periodic_task.periodic_task(spacing=60, run_immediately=True)
    def canary(self, context):
        """
        A simple periodic task to record when it ran last
        This will give us an indication if periodic tasks
        are running normally or if they may be hung.
        """
        update_canary_run_time()

    @periodic_task.periodic_task(spacing=CONF.monitor_instances_interval)
    def monitor_instances(self, context, instances=None):
        """Periodic Task to Refresh Metrics for the Instances"""
        driver, context = (self.driver, context.elevated())
        metric_keys = ['cpu_util', 'mem_util', 'pg_in_rate',
                       'pg_out_rate', 'swap_in_rate', 'swap_out_rate']
        LOG.debug('Invoking Method to Refresh Instance Metrics - Begin')
        try:
            drv_insts = self.driver.list_instances()
            name_map = dict([(inst, '') for inst in drv_insts])
            # Retrieve All of the Instances from the Nova DB for the Given Host
            db_insts = self._query_db_instances(context, instances)
            db_insts = [inst for inst in db_insts if inst['name'] in name_map]
            # If this is a Discovery Driver, then let it Gather Metrics
            if isinstance(driver, discovery_driver.ComputeDiscoveryDriver):
                # It is only worth Gathering Metrics if there are Instances
                if len(db_insts) > 0:
                    metric_data = driver.monitor_instances(context, db_insts)
                    # Add metric values to cache.
                    for metric_key in metric_keys:
                        self.vm_util_cache.add_value(metric_data, metric_key)
                    util_dict = self.vm_util_cache.get_average()
                    for instance in db_insts:
                        metrics_dict = dict()
                        power_specs = instance.get('power_specs', dict())
                        for metric_key in metric_keys:
                            self.__update_metrics_dict(
                                util_dict, metric_key, instance,
                                power_specs, metrics_dict)
                        # Update only if there is something to update.
                        if metrics_dict:
                            instance.power_specs = metrics_dict
                            instance.save()
        except Exception as exc:
            LOG.warn(_('Error refreshing Instance Metrics'))
            LOG.exception(exc)
        LOG.debug('Invoking Method to Refresh Instance Metrics - End')

    def __update_metrics_dict(self, util_dict, util_key,
                              instance, power_specs, metrics_dict):
        """Updates metrics_dict with value only if the value in power_specs
        is within acceptable range of update
        """
        util_val = util_dict.get(instance.uuid, dict()).get(util_key)
        variation = 0
        if util_key in ('cpu_util', 'mem_util'):
            # Don't update if new cpu_util/mem_util value is within +- 0.5%.
            variation = METRICS_ALLOWED_VARIATION.get('util')
        elif util_key in ('pg_in_rate', 'pg_out_rate',
                          'swap_in_rate', 'swap_out_rate'):
            # Don't update if new page rate values are within +-10 pages/sec.
            variation = METRICS_ALLOWED_VARIATION.get('page_rate')
        if self.__need_update(power_specs.get(util_key), util_val, variation):
            metrics_dict.update({util_key: util_val})

    def __need_update(self, old_val, new_val, variation):
        # Update the utilization only if the delta is more than 0.5 % units;
        # otherwise it's effectively the same value and no update is required.
        if old_val is None and new_val is None:
            # No need to update if both are None.
            return False
        if ((old_val is None and new_val is not None) or
                (old_val is not None and new_val is None)):
            # Ensure that None values does not go for comparison check.
            return True
        return abs(new_val - old_val) > variation

    def check_and_run_first_boot(self, context):
        """
        Method to be run on first boot post registration
        """
        # check if first boot has already occurred, if yes return
        if CONF.has_first_boot_occurred:
            return
        # if not, check that driver is instance of ComputeDiscoveryDriver
        if isinstance(self.driver, discovery_driver.ComputeDiscoveryDriver):
            # run the post registration against the driver
            LOG.debug('Running post_registration against ' + str(self.driver))
            self.driver.post_registration(context)
            # run openstack-config to set has_first_boot_occurred as True
            LOG.debug('Running openstack-config to set '
                      'has_first_boot_occurred')
            oslo_utils.set_openstack_config_value(CONF.config_file[0],
                                                  'has_first_boot_occurred',
                                                  'True', 'DEFAULT')
            # update in-memory has_first_boot_occurred
            LOG.debug('Updating in-memory conf option has_first_boot_occurred')
            CONF.has_first_boot_occurred = True

    def _verify_mgmt_system_access(self, context):
        """Helper method to verify the Management System can be accessed"""
        max_retry_count = CONF.verify_mgmt_sys_max_retry_count
        sleep_time = CONF.verify_mgmt_sys_sleep_time_during_retry
        self._retry_operation(self._populate_admin_context, context,
                              max_retry_count, sleep_time)
        self._retry_operation(self._check_cinder_service, context,
                              max_retry_count, sleep_time)
        self._retry_operation(self._check_glance_service, context,
                              max_retry_count, sleep_time)
        self._retry_operation(self._check_neutron_service, context,
                              max_retry_count, sleep_time)

    def _check_neutron_service(self, context):
        neutronv2.api.get_client(context, True).list_networks()

    def _check_cinder_service(self, context):
        cinderv2.cinderclient(context).client.get('/os-hosts')

    def _check_glance_service(self, context):
        glance.GlanceClientWrapper().call(context, 1, 'list')

    def _undefine_periodic_tasks(self):
        for task_name, task in self._periodic_tasks:
            LOG.debug("Defined periodic task: %s" % task_name)
            if task_name in PERIODIC_TASKS_TO_REMOVE:
                LOG.debug("Removing defined periodic task: %s" % task_name)
                self._periodic_tasks.remove((task_name, task))

    def _init_ptask_nanny(self):
        nanny_wrapped = []
        for task_name, task in self._periodic_tasks:
            LOG.debug("Defined periodic task nanny for: %s" % task_name)
            nanny = functools.partial(self._periodic_task_nanny,
                                      task_=task, task_name=task_name)
            # Copy the settings for the task
            nanny._periodic_spacing = task._periodic_spacing
            nanny._periodic_immediate = task._periodic_immediate
            nanny._periodic_last_run = task._periodic_last_run
            nanny._periodic_external_ok = task._periodic_external_ok
            nanny._periodic_enabled = task._periodic_enabled
            # Add the new task to the list
            nanny_wrapped.append((task_name, nanny))

        # Replace the list of tasks with our wrapped ones
        self._periodic_tasks_orig = self._periodic_tasks
        self._periodic_tasks = nanny_wrapped

    def _periodic_task_nanny(self, *args, **kwargs):
        # Grab parameters
        context = args[1]
        task = kwargs['task_']
        task_name = kwargs['task_name']

        def run_task(task, task_name, context):
            LOG.debug('Starting periodic task: %s' % task_name)
            start_time = time.time()
            try:
                task(self, context)
                end_time = time.time()
                LOG.debug('Completed periodic task: %s' % task_name)
                delta = end_time - start_time
                if delta > CONF.periodic_task_completion_msg_threshold:
                    LOG.info(_('Periodic task %(task_name)s ran '
                               'for %(time)s seconds.')
                             % dict(task_name=task_name, time=int(delta)))
            except Exception:
                LOG.exception(_('Exception during periodic task %(task_name)s '
                                ' execution') % dict(task_name=task_name))
            finally:
                self._periodic_tasks_running.pop(task_name)

        # Only run the task if it's not already running
        if task_name not in self._periodic_tasks_running:
            self._periodic_tasks_running[task_name] = time.time()
            if task_name in RUN_PERIODIC_TASK_ASYNC:
                self._periodic_task_pool.spawn_n(run_task, task,
                                                 task_name, context)
            else:
                run_task(task, task_name, context)
        else:
            # The periodic task is still running a prior invocation
            # check if we should flag it.
            try:
                dispatch_time = self._periodic_tasks_running[task_name]
                delta = time.time() - dispatch_time
                if delta > CONF.periodic_task_runtime_msg_threshold:
                    LOG.warn(_('Periodic task %(task_name)s has been running '
                               'for %(time)s seconds.')
                             % dict(task_name=task_name, time=int(delta)))
            except KeyError:
                # It must have finished...
                pass

    def _init_instance(self, context, instance):
        """Override Init-Instance method to fix Base Compute Manager issues"""
        try:
            super(PowerVCComputeManager,
                  self)._init_instance(context, instance)
        except Exception:
            LOG.exception(_('exception initializing the Instance, moving on'))

    @pvc_comp_mgr_reverts_task_state
    @send_ui_notification(None,
                          notify_msg.START_ERROR,
                          notify_msg.START_SUCCESS,
                          log_exception=True)
    @check_host_maintenance_mode(None, common_constants.OPER_START)
    def start_instance(self, context, instance):
        """Starting an instance on this host."""
        # Instance must be passed as a keyword argument in order for
        # revert_task_state decorator to function properly
        super(PowerVCComputeManager, self).start_instance(context,
                                                          instance=instance)

    @pvc_comp_mgr_reverts_task_state
    @send_ui_notification(None,
                          notify_msg.PAUSE_ERROR,
                          notify_msg.PAUSE_SUCCESS,
                          log_exception=True)
    @check_host_maintenance_mode(None, common_constants.OPER_PAUSE)
    def pause_instance(self, context, instance):
        """Pausing an instance on this host."""
        # Instance must be passed as a keyword argument in order for
        # revert_task_state decorator to function properly
        super(PowerVCComputeManager, self).pause_instance(context,
                                                          instance=instance)

    @pvc_comp_mgr_reverts_task_state
    @send_ui_notification(None,
                          notify_msg.RESUME_ERROR,
                          notify_msg.RESUME_SUCCESS,
                          log_exception=True)
    @check_host_maintenance_mode(None, common_constants.OPER_UNPAUSE)
    def unpause_instance(self, context, instance):
        """Resuming an instance on this host."""
        # Instance must be passed as a keyword argument in order for
        # revert_task_state decorator to function properly
        super(PowerVCComputeManager, self).unpause_instance(context,
                                                            instance=instance)

    @pvc_comp_mgr_reverts_task_state
    @send_ui_notification(None,
                          notify_msg.SUSPEND_ERROR,
                          notify_msg.SUSPEND_SUCCESS,
                          log_exception=True)
    @check_host_maintenance_mode(None, common_constants.OPER_SUSPEND)
    def suspend_instance(self, context, instance):
        """Suspending an instance on this host."""
        # Instance must be passed as a keyword argument in order for
        # revert_task_state decorator to function properly
        super(PowerVCComputeManager, self).suspend_instance(context,
                                                            instance=instance)

    @pvc_comp_mgr_reverts_task_state
    @send_ui_notification(None,
                          notify_msg.RESUME_ERROR,
                          notify_msg.RESUME_SUCCESS,
                          log_exception=True)
    @check_host_maintenance_mode(None, common_constants.OPER_RESUME)
    def resume_instance(self, context, instance):
        """Resuming an instance on this host."""
        # Instance must be passed as a keyword argument in order for
        # revert_task_state decorator to function properly
        super(PowerVCComputeManager, self).resume_instance(context,
                                                           instance=instance)

    @pvc_comp_mgr_reverts_task_state
    @send_ui_notification(None,
                          notify_msg.RESTART_ERROR,
                          notify_msg.RESTART_SUCCESS,
                          log_exception=True)
    @check_host_maintenance_mode(None, common_constants.OPER_REBOOT)
    def reboot_instance(self, context, instance, block_device_info,
                        reboot_type):
        """Rebooting an instance on this host."""
        # Instance must be passed as a keyword argument in order for
        # revert_task_state decorator to function properly
        super(PowerVCComputeManager, self).reboot_instance(
            context, instance=instance, block_device_info=block_device_info,
            reboot_type=reboot_type)

    @manager.delete_image_on_error
    @send_ui_notification(None, notify_msg.CAPTURE_ERROR,
                          notify_msg.CAPTURE_SUCCESS,
                          log_exception=True)
    @check_host_maintenance_mode(None, common_constants.OPER_CAPTURE)
    def snapshot_instance(self, context, image_id, instance):
        # Instance must be passed as a keyword argument in order for
        # revert_task_state decorator to function properly
        super(PowerVCComputeManager, self).snapshot_instance(context,
                                                             image_id,
                                                             instance=instance)

    @wrap_exception()
    def remove_volume_connection(self, context, volume_id, instance):
        """Override this method as we need to fetch the bdm based on volume
           and instance id. For shared volumes there would be more than one bdm
           and in case of failed deploys the bdm are left behind.
        """
        #TODO: Give LP defect # here
        try:
            bdm = objects.BlockDeviceMapping.get_by_volume_and_instance(
                context, volume_id, instance['uuid'])
            self._driver_detach_volume(context, instance, bdm)
            connector = self.driver.get_volume_connector(instance)
            #workaround for LP defect 1535643. Log and continue when
            # no connector is passed in
            if not connector:
                LOG.info(_("No connector to remove vol connection for instance"
                           " %(inst)s and vol %(vol)s") %
                         dict(inst=instance['uuid'], vol=volume_id))
            else:
                self.volume_api.terminate_connection(context, volume_id,
                                                     connector)
        except exception.NotFound:
            pass

    @wrap_exception()
    @pvc_comp_mgr_reverts_task_state
    @send_ui_notification(None,
                          notify_msg.DETACH_ERROR,
                          notify_msg.DETACH_SUCCESS,
                          log_exception=True)
    @check_host_maintenance_mode(None, common_constants.OPER_DT_VOL)
    def detach_volume(self, context, volume_id, instance, attachment_id=None):
        """Detach a volume from an instance"""
        # Instance must be passed as a keyword argument in order for
        # revert_task_state decorator to function properly
        super(PowerVCComputeManager, self).detach_volume(
            context, volume_id=volume_id, instance=instance,
            attachment_id=attachment_id)

    def verify_host_running(self, context):
        """Verifies the nova-compute service for the Host is running"""
        return True

    def pvc_update_available_resource(self, context):
        return super(PowerVCComputeManager, self).\
            update_available_resource(context)

    @wrap_exception()
    @reverts_task_state
    @check_host_maintenance_mode(None, common_constants.OPER_MIGRATE_AS_TGT)
    def check_can_live_migrate_destination(self, ctxt, instance,
                                           block_migration=False,
                                           disk_over_commit=False):
        """
        When source host validation fails we need to clean up the driver's
        migration object on the target system.  We can't do this via the
        dictionaries passed between the methods.
        NOTE: No need to use @pvc_comp_mgr_reverts_task_state since
        super doesn't have @reverts_task_state
        """

        # Dictionary to pass between all the migration methods
        # on both the source and target host threads

        try:
            # Because the scheduler filters aren't invoked for targeted
            # migrations, we need to ensure that we aren't migrating outside
            # of the host aggregate domain or violating a collocation rule.
            migrate_utils.verify_aggregate_domain(ctxt, instance.display_name,
                                                  instance.host, CONF.host)
            migrate_utils.verify_collocation_rules(ctxt, instance, CONF.host)
            migrate_data = super(PowerVCComputeManager, self). \
                check_can_live_migrate_destination(
                    ctxt, instance=instance, block_migration=block_migration,
                    disk_over_commit=disk_over_commit)
        except pvm_exc.IBMPowerVMMigrationInProgress:
            with excutils.save_and_reraise_exception():
                pass
        except Exception:
            with excutils.save_and_reraise_exception():
                self.driver.migration_cleanup(ctxt, instance)

        return migrate_data

    @reverts_task_state
    def _do_live_migration(self, context, dest, instance, block_migration,
                           migration, migrate_data):
        super(PowerVCComputeManager, self)._do_live_migration(
            context, dest, instance, block_migration,
            migration, migrate_data)

    @wrap_exception()
    @reverts_task_state
    def _rollback_live_migration(self, context, instance,
                                 dest, block_migration, migrate_data=None):
        try:
            super(PowerVCComputeManager, self)._rollback_live_migration(
                context, instance, dest, True, migrate_data=migrate_data)
        except Exception as ex:
            LOG.error(_("Rollback live migration failed with error %s") %
                      six.text_type(ex))
        finally:
            # Make sure the migration flags are cleared on the destination
            # else all further migrations to this host will fail
            self.compute_rpcapi.rollback_live_migration_at_destination(
                context, instance, dest, migrate_data=migrate_data)

    def _get_instances_on_driver(self, context, filters=None):
        """Override to include vm_names in filters for instance queries.
        """
        if not filters:
            filters = {}
        try:
            driver_uuids = self.driver.list_instance_uuids()
            if len(driver_uuids) == 0:
                # Short circuit, don't waste a DB call
                return objects.InstanceList()
            filters['uuid'] = driver_uuids
            local_instances = objects.InstanceList.get_by_filters(
                context, filters, use_slave=True)
            return local_instances
        except NotImplementedError:
            pass

        # WI 46938, performance improvement to query instances
        # based off of the vm_name in the instance_power_specs. The
        # community manager employs brute force iteration over all
        # deleted instances which can be expensive at scale.
        filters['vm_names'] = self.driver.list_instances()
        return objects.InstanceList.get_by_filters(context, filters,
                                                   use_slave=True)

    def _prep_block_device(self, context, instance, bdms,
                           do_check_attach=True):
        """
        Override nova ComputeManager's _prep_block_device
        in order to consume the new image meta data for PowerVC to
        prepare the VM's volumes.  Volumes are allocated but not attached here.
        :param context: operation context
        :param instance: compute instance
        :param bdms: The block device mapping based on image mapping
                     and block_device_mapping properties or the
                     block_device_mapping passed as parameter during
                     deploy.
        :return block_device_info:
        """
        if self._is_managing_PowerKVM():
            s = super(PowerVCComputeManager, self)
            return s._prep_block_device(context, instance, bdms)

        # this is the path for PowerVM volume backed VMs
        image_ref = instance['image_ref']
        if not image_ref:
            LOG.warn(_("ComputeMgr _prep_block_device() exception"
                       " since missing image_ref from instance"))
            raise exception.InvalidImageRef(image_href=image_ref)

        if not bdms:
            exc_args = {'instuuid': instance['uuid']}
            raise pwrvc_excep.IBMPowerVCMissingBDMs(**exc_args)

        image_meta = self._get_image_meta(context, image_ref)
        if image_ref != 'ONBOARDED_NO_IMAGE':
            self._validate_image_meta(image_meta)

        virt_bdms = self._prepare_volumes(context, self.volume_api, instance,
                                          bdms, image_meta)
        return self._prep_block_device_volume_attach(context, instance,
                                                     virt_bdms,
                                                     do_check_attach)

    def _bdm_has_attachment(self, bdm):
        if not bdm or not bdm.get("connection_info"):
            return False
        connection_info = bdm.get("connection_info")
        if (not connection_info or 'driver_volume_type' not in
                connection_info or 'data' not in connection_info):
            return False
        else:
            return True

    def _cleanup_imported_networks(self, context, network_info):
        """Cleanup Networks if they were for Imported VM's and not used"""
        network_ids = dict()
        NETNM1, NETNM2 = ('Imported Flat Network', 'Imported Network for VLAN')
        try:
            client = neutronv2.api.get_client(context)
            # Loop through each of the VIF's, check if any are
            for vif in network_info:
                if vif.get('network') and vif['network'].get('label'):
                    name = vif['network']['label']
                    # We only want to delete it if it is an Import created one
                    if name.startswith(NETNM1) or name.startswith(NETNM2):
                        network_ids[vif['network']['id']] = ''
            # Loop through the Network ID's seeing if they have no Ports left
            for net_id in network_ids:
                name = client.show_network(net_id)['network']['name']
                # Check the Name again to make sure user didn't rename later
                if not name.startswith(NETNM1) and not name.startswith(NETNM2):
                    continue
                # If no ports were left, then we want to delete the Network
                if len(client.list_ports(network_id=net_id)['ports']) <= 0:
                    LOG.info(_("Removing unused Imported Network %s") % net_id)
                    client.delete_network(net_id)
        # Since the cleanup isn't essential, just log errors and don't fail
        except Exception as exc:
            LOG.warn(_("Error trying to Cleanup Imported Networks"))
            LOG.exception(exc)

    def _prep_block_device_volume_attach(self, context, instance, virt_bdms,
                                         do_check_attach):
        """Handle volume attachment during prep_block device.
        This method is a PowerVC specific method that exists to allow
        differentiation between HMC and Novalink for how volume attachment
        is handled at the tail end of _prep_block_device.

        :param instance: nova.objects.instance.Instance
        :param context: the context
        :param virt_bdms: the BDMs represented as nova.virt.block_device
            classes
        :param do_check_attach: the value of the do_check_attach parameter
            from the _prep_block_device call
        """
        raise NotImplementedError()

    def _is_managing_PowerKVM(self):
        if hasattr(self.driver, "_get_hypervisor_type"):
            hypervisor_type = self.driver._get_hypervisor_type()
            # For PowerKVM, use openstack block device path directly
            if hypervisor_type and hypervisor_type.lower() == 'qemu':
                return True
        return False

    @staticmethod
    def _get_image_meta(context, image_ref):
        if image_ref == 'ONBOARDED_NO_IMAGE':
            return None
        image_service, image_id = glance.get_remote_image_service(context,
                                                                  image_ref)
        return image_service.show(context, image_id)

    def _validate_image_meta(self, image_meta):
        image_properties = image_meta.get('properties', {})
        required_props = ['bdm_v2', 'block_device_mapping']
        for prop in required_props:
            if not image_properties.get(prop):
                exc_args = {'image_name': image_meta['name'],
                            'property_name': prop}
                raise pwrvc_excep.\
                    IBMPowerVCMissingRequiredImageProperty(**exc_args)

    def _prepare_volumes(self, context, volume_api, instance, bdms,
                         image_meta):
        """
        Prepare the volumes from the block device mappings.
        """
        LOG.debug('Preparing volumes for BDMs %s' % bdms)
        virt_bdms = self._convert_block_devices(bdms)

        self._validate_supported_bdms(virt_bdms)

        image_ref = instance['image_ref']
        volumes_to_clone = self.\
            _get_image_bdm_volumes_to_clone(context, image_ref,
                                            image_metadata=image_meta)
        LOG.debug('Image volumes that will be cloned: %s' % volumes_to_clone)
        self._check_for_duplicate_attach(instance['uuid'],
                                         bdms, volumes_to_clone)
        source_vol_to_volume_type = volume_type_utils.\
            get_volume_types_from_flavor(context, volumes_to_clone,
                                         instance=instance)
        boot_index_to_vol = {}
        for bdm in virt_bdms:
            vol = None
            if isinstance(bdm, blk_dev.DriverBlankBlockDevice):
                if not bdm.volume_id:
                    prefix = CONF.data_volume_name_prefix
                    vol_name = self._generate_instance_volume_name(instance,
                                                                   bdm,
                                                                   prefix)
                    vol = volume_api.create(context, bdm.volume_size,
                                            vol_name, '')
                    LOG.debug('Blank volume created: %s' % vol)
            elif isinstance(bdm, blk_dev.DriverVolumeBlockDevice):
                if bdm.volume_id in volumes_to_clone:
                    flv_vol_type = source_vol_to_volume_type.get(bdm.volume_id)
                    vol_type_id = None
                    if flv_vol_type and 'id' in flv_vol_type:
                        vol_type_id = flv_vol_type['id']
                    vol = self._clone_bdm_source_volume(context, volume_api,
                                                        bdm, instance,
                                                        vol_type_id)
                    # Here we change the delete_on_termination value to True
                    # since we have created a cloned volume from an image
                    # volume and this volume will need to be deleted when the
                    # VM is deleted. We set delete_on_termination on the
                    # BDM object because the nova.virt.block_device class
                    # does not proxy delete_on_termination to the actual BDM
                    # object.
                    bdm._bdm_obj.delete_on_termination = True
                    bdm['delete_on_termination'] = True
                    LOG.debug('Image volume %s cloned to volume %s' %
                              (bdm.volume_id, vol))
            if vol:
                bdm.volume_id = vol['id']
                bdm.save()

        # Check the volumes before leaving to catch volume creation errors
        # early, before the deploy proceeds to LPAR and network creation.
        bdms = [bdm for bdm in bdms if bdm.is_volume]
        for bdm in bdms:
            volume = volume_api.get(context, bdm['volume_id'])
            boot_index_to_vol[bdm['boot_index']] = volume
            blank_volume = (bdm.source_type == 'blank')
            self._validate_for_error_status(context, instance, volume,
                                            blank_volume)

        # Update root_gb on the instance with the size of the volume at
        # boot_index = 0 to allow resize to work as it has in past releases.
        boot_disk_size = boot_index_to_vol[0]['size']
        instance.root_gb = boot_disk_size
        instance.save()

        return virt_bdms

    def _convert_block_devices(self, bdms):
        """
        Convert a list of Nova object block device mappings into
        nova.virt.block_device wrappered instances.
        """
        driver_bdm_list = blk_dev.convert_volumes(bdms) +\
            blk_dev.convert_snapshots(bdms) +\
            blk_dev.convert_images(bdms) +\
            blk_dev.convert_blanks(bdms)
        return driver_bdm_list

    def _validate_supported_bdms(self, bdms):
        for bdm in bdms:
            if isinstance(bdm, blk_dev.DriverBlankBlockDevice):
                continue
            elif isinstance(bdm, blk_dev.DriverSnapshotBlockDevice):
                exc_args = {'bdm_type': 'snapshot'}
                raise pwrvc_excep.IBMPowerVCUnsupportedBDM(**exc_args)
            elif isinstance(bdm, blk_dev.DriverEphemeralBlockDevice):
                exc_args = {'bdm_type': 'ephemeral'}
                raise pwrvc_excep.IBMPowerVCUnsupportedBDM(**exc_args)
            elif isinstance(bdm, blk_dev.DriverSwapBlockDevice):
                exc_args = {'bdm_type': 'swap'}
                raise pwrvc_excep.IBMPowerVCUnsupportedBDM(**exc_args)
            elif isinstance(bdm, blk_dev.DriverImageBlockDevice):
                exc_args = {'bdm_type': 'image'}
                raise pwrvc_excep.IBMPowerVCUnsupportedBDM(**exc_args)
            elif isinstance(bdm, blk_dev.DriverVolumeBlockDevice):
                continue
            else:
                exc_args = {'bdm_type': type(bdm).__name__}
                raise pwrvc_excep.IBMPowerVCUnsupportedBDM(**exc_args)

    def _get_image_bdm_volumes_to_clone(self, context, image_ref,
                                        image_metadata=None):
        """
        Get the volumes from the image that we will clone.  We only clone
        volumes in image BDMs if they have delete_on_termination=true.
        """
        vols_to_clone = []

        if not image_metadata:
            image_metadata = self._get_image_meta(context, image_ref)
            if image_metadata is None:
                return vols_to_clone

        image_properties = image_metadata.get('properties', {})

        # Get the block device mappings defined by the image.
        image_defined_bdms = image_properties.get('block_device_mapping', [])
        image_defined_bdms = map(block_device.BlockDeviceDict,
                                 image_defined_bdms)
        for bdm in image_defined_bdms:
            vol_id = bdm.get('volume_id')
            delete_on_termination = bdm.get('delete_on_termination')
            if vol_id and delete_on_termination:
                vols_to_clone.append(vol_id)
        return vols_to_clone

    def _check_for_duplicate_attach(self, instance_uuid, bdms, image_volumes):
        """Check for duplicate volume attach BDMs and raise an exception if
        they are found. Do not prevent duplicate volume_ids in BDMs if the
        volume ID is from an image since the image volumes will be cloned
        before attach.

        @param bdms: The instance BDMs, expected to be a list of nova object
        BDMs
        @param image_volumes: The list of image volume IDs that will be cloned
        """
        found_volumes = []
        for bdm in bdms:
            if bdm.destination_type == 'volume' and bdm.volume_id:
                if bdm.volume_id in image_volumes:
                    continue
                if bdm.volume_id in found_volumes:
                    ex_ags = {'volume_id': bdm.volume_id,
                              'instuuid': instance_uuid}
                    raise pwrvc_excep.IBMPowerVCDuplicateVolumeAttach(**ex_ags)
                found_volumes.append(bdm.volume_id)

    def _clone_bdm_source_volume(self, context, volume_api, bdm,
                                 instance, instance_flavor_vol_type):
        src_volume = volume_api.get(context, bdm.volume_id)
        LOG.debug('Creating volume from volume %s' % bdm.volume_id)

        source_metadata = src_volume['volume_metadata']

        target_metadata = {'instance_uuid': instance['uuid']}
        prefix = CONF.data_volume_name_prefix
        if 'is_boot_volume' in source_metadata and \
                source_metadata['is_boot_volume'] == 'True':
            target_metadata['is_boot_volume'] = 'True'
            prefix = CONF.boot_volume_name_prefix

        volume_name = self._generate_instance_volume_name(instance,
                                                          bdm,
                                                          prefix)

        return volume_api.create(context, bdm.volume_size, volume_name, '',
                                 source_vol=src_volume,
                                 metadata=target_metadata,
                                 volume_type=instance_flavor_vol_type)

    def _generate_instance_volume_name(self, instance, virtbdm, boot_or_data):
        """
        Generate volume names for the instance's volumes.
        """
        boot_index = virtbdm.get('boot_index')
        bdm_id = virtbdm._bdm_obj.id
        unique_id = (six.text_type(boot_index)
                     if boot_index is not None else six.text_type(bdm_id))
        if instance.get('name'):
            prefix = instance['name']
        elif instance.get('uuid'):
            prefix = instance['uuid']
        else:
            prefix = ''
        return prefix + '-' + boot_or_data + unique_id

    def _notify_deploy_volume_create_failure(self, context, instance, volume):
        create_fail_msg = volume['volume_metadata'].get(common_constants.
                                                        CREATE_FAILURE_KEY)
        if not create_fail_msg:
            create_fail_msg = _('No detailed volume error message.')
        reason = (_("Instance %(instance_name)s volume creation "
                    "failure. Health status: %(hstatus)s. "
                    " Volume error: %(volume_error)s.") %
                  {'instance_name': instance['display_name'],
                   'hstatus': volume.get('health_status'),
                   'volume_error': create_fail_msg})
        LOG.error(reason)
        self._error_notify(instance, context, 'error', reason)

    def _validate_for_error_status(self, context, instance, volume, blank_vol):
        if 'error' in volume['status']:
            vol_msg = self._notify_deploy_volume_create_failure(context,
                                                                instance,
                                                                volume)
            src_vol_id = volume['source_volid']
            exc_args = {'instname': instance['display_name'],
                        'instuuid': instance['uuid'],
                        'vol_error': vol_msg}
            # Throw different exceptions if this is a volume create
            # from volume vs a blank volume create vs an existing
            # volume that is being attached which happens to be in error.
            if src_vol_id:
                exc_args['srcvolid'] = src_vol_id
                raise pwrvc_excep.IBMPowerVCVolumeCloneFailure(**exc_args)
            elif blank_vol:
                raise pwrvc_excep.IBMPowerVCVolumeCreateFailure(**exc_args)
            else:
                exc_args['vol_id'] = volume['id']
                exc_args['vol_name'] = volume['display_name']
                raise pwrvc_excep.IBMPowerVCExistingVolumeInError(**exc_args)

    def _error_notify(self, instance, context, severity, details):
        info = {'msg': _('{message} Virtual Server: instance UUID:'
                         '{instance_id}, instance name: {instance_name} '
                         'on host: {host_name}'),
                'message': details,
                'instance_id': instance['uuid'],
                'instance_name': instance['display_name'],
                'host_name': self.host}
        notifier = rpc.get_notifier(service='compute', host=self.host)
        if severity == 'error':
            notifier.error(context, 'compute.instance.log', info)
        elif severity == 'info':
            notifier.info(context, 'compute.instance.log', info)
        elif severity == 'critical':
            notifier.critical(context, 'compute.instance.log', info)
        elif severity == 'warn':
            notifier.warn(context, 'compute.instance.log', info)
        elif severity == 'debug':
            notifier.debug(context, 'compute.instance.log', info)

    def set_vm_state(self, db_instance, vm_power_state, new_state):
        """
        Set the vm_state of an instance to new_state
        """
        LOG.warn(_('Current instance state of %(name)s is %(old_state)s. '
                   'Instance power state is %(power_state)s. '
                   'Setting instance state to %(new_state)s '
                   'in database.') %
                 {'name': db_instance['display_name'],
                  'old_state': db_instance['vm_state'],
                  'power_state': vm_power_state, 'new_state': new_state})
        db_instance.vm_state = new_state
        db_instance.save()

    def _sync_instance_power_state(self, context, db_instance, vm_power_state,
                                   use_slave=False):
        """Align instance power state between the database and hypervisor.

        If the instance is not found on the hypervisor, but is in the database,
        then a stop() API will be called on the instance.
        """

        # We re-query the DB to get the latest instance info to minimize
        # (not eliminate) race condition.
        db_instance.refresh()
        db_power_state = db_instance.power_state
        vm_state = db_instance.vm_state

        if self.host != db_instance.host:
            # on the sending end of nova-compute _sync_power_state
            # may have yielded to the greenthread performing a live
            # migration; this in turn has changed the resident-host
            # for the VM; However, the instance is still active, it
            # is just in the process of migrating to another host.
            # This implies that the compute source must relinquish
            # control to the compute destination.
            LOG.info(_("During the sync_power process the "
                       "instance has moved from "
                       "host %(src)s to host %(dst)s") %
                     {'src': self.host,
                      'dst': db_instance.host},
                     instance=db_instance)
            return
        elif db_instance.task_state is not None:
            # on the receiving end of nova-compute, it could happen
            # that the DB instance already report the new resident
            # but the actual VM has not showed up on the hypervisor
            # yet. In this case, let's allow the loop to continue
            # and run the state sync in a later round
            LOG.info(_("During sync_power_state the instance has a "
                       "pending task. Skip."), instance=db_instance)
            return

        if vm_power_state != db_power_state:
            # power_state is always updated from hypervisor to db
            db_instance.power_state = vm_power_state
            db_instance.save()
            db_power_state = vm_power_state

        # Note(maoy): Now resolve the discrepancy between vm_state and
        # vm_power_state. We go through all possible vm_states.
        if vm_state in (vm_states.BUILDING,
                        vm_states.RESCUED,
                        vm_states.RESIZED,
                        vm_states.ERROR):
            # TODO(maoy): we ignore these vm_state for now.
            pass
        elif vm_state == vm_states.ACTIVE:
            # The only rational power state should be RUNNING
            if vm_power_state in (power_state.SHUTDOWN,
                                  power_state.CRASHED):
                # double check on vm_power_state before we running the stop api
                # if the vm_power_state is the same, just return and wait
                # for the next sync
                new_vm_pwr_state = self._get_power_state(context, db_instance)
                if vm_power_state != new_vm_pwr_state:
                    LOG.warn(_('During sync_power_state the instance '
                               'state is not up to date. Skip.'),
                             instance=db_instance)
                    return
                LOG.warn(_("Instance shutdown by itself. Calling "
                           "the stop API."), instance=db_instance)
                try:
                    # Note(maoy): here we call the API instead of
                    # brutally updating the vm_state in the database
                    # to allow all the hooks and checks to be performed.
                    self.compute_api.stop(context, db_instance)
                except Exception:
                    # Note(maoy): there is no need to propagate the error
                    # because the same power_state will be retrieved next
                    # time and retried.
                    # For example, there might be another task scheduled.
                    LOG.exception(_("error during stop for shutdown/crashed"),
                                  instance=db_instance)
            elif vm_power_state == power_state.SUSPENDED:
                # Instance is suspended. Setting state of
                # instance to suspended in DB accordingly.
                self.set_vm_state(db_instance, vm_power_state,
                                  vm_states.SUSPENDED)
            elif vm_power_state == power_state.PAUSED:
                # Instance is paused. Setting state of
                # instance to paused in DB accordingly.
                self.set_vm_state(db_instance, vm_power_state,
                                  vm_states.PAUSED)
            elif vm_power_state == power_state.NOSTATE:
                # Occasionally, depending on the status of the hypervisor,
                # which could be restarting for example, an instance may
                # not be found.  Therefore just log the condidtion.
                LOG.warn(_("Instance is unexpectedly not found. Ignore."),
                         instance=db_instance)
        elif vm_state in (vm_states.STOPPED,
                          vm_states.PAUSED,
                          vm_states.SUSPENDED):
            if vm_power_state == power_state.RUNNING:
                # Instance is running. Setting state of
                # instance to active in DB accordingly.
                self.set_vm_state(db_instance, vm_power_state,
                                  vm_states.ACTIVE)
            elif vm_power_state == power_state.SHUTDOWN:
                if vm_state != vm_states.STOPPED:
                    # Instance is stopped. Setting
                    # state of instance to stopped in DB.
                    self.set_vm_state(db_instance, vm_power_state,
                                      vm_states.STOPPED)
            elif vm_power_state == power_state.SUSPENDED:
                if vm_state != vm_states.SUSPENDED:
                    # If an instance is suspended out of band, its vm_state
                    # could become 'stopped' or 'paused' first. Sync up
                    # vm_state with power_state here.
                    self.set_vm_state(db_instance, vm_power_state,
                                      vm_states.SUSPENDED)
            else:
                # Do nothing if instance is in following state
                # as reported by hypervisor
                # NOSTATE
                # CRASHED
                # BUILDING
                LOG.warn(_("Instance state is %(s1)s as reported by "
                           "hypervisor, but its state in PowerVC is "
                           "%(s2)s. Ignore.")
                         % {'s1': vm_power_state, 's2': vm_state})
        elif vm_state in (vm_states.SOFT_DELETED,
                          vm_states.DELETED):
            if vm_power_state not in (power_state.NOSTATE,
                                      power_state.SHUTDOWN):
                # Note(maoy): this should be taken care of periodically in
                # _cleanup_running_deleted_instances().
                LOG.warn(_("Instance is not (soft-)deleted."),
                         instance=db_instance)

    def is_instance_on_host(self, context, instance_uuid, instance_name):
        """ Answer query from another host if instance
            is on this host.
        """
        ret = self.driver.is_instance_on_host(instance_uuid, instance_name)
        if ret:
            LOG.info(_('Virtual machine %(name)s (%(uuid)s) is being '
                       'managed by this host.') %
                     {'uuid': instance_uuid, 'name': instance_name})
        else:
            LOG.info(_('Virtual machine %(name)s (%(uuid)s) is not '
                       'being managed by this host.') %
                     {'uuid': instance_uuid, 'name': instance_name})
        return ret


# Spawn a thread to be the periodic task watchdog
greenthread.spawn(periodic_task_watchdog)

# Inject our own overwritten lock function to avoid spinning forever
setattr(lockutils, 'InterProcessLock', Monkey_FcntlLock)
