# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Extensions to the OpenStack Compute ResourceTracker class"""

import copy
from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import encodeutils

from nova import exception
from nova import rpc
from nova.compute import resources
from nova.compute import resource_tracker
from nova.compute.monitors import MonitorHandler
from nova.objects import base as obj_base
from nova.objects import compute_node as compute_obj
from nova.objects import instance as instance_obj
from nova.objects import service as service_obj

from powervc_nova.compute import put_host_in_maintenance_mode
from powervc_nova.compute import remove_host_from_maintenance_mode
from powervc_nova.conductor import api as conductor

from powervc_nova import _

LOG = logging.getLogger(__name__)


#######################################################
#           Override ResourceTracker Class            #
#######################################################
class PowerVCResourceTracker(resource_tracker.ResourceTracker):
    """Extends the base ResourceTracker class for PowerVC to fix issues"""

    def __init__(self, host, driver, nodename):
        # Monkey patch the resource handler to load our resources
        resources.RESOURCE_NAMESPACE = 'powervc_nova.compute.resources'
        """Constructor for the PowerVC extension to the Resource Tracker."""
        super(PowerVCResourceTracker, self).__init__(host, driver, nodename)
        self.conductor_api = conductor.PowerVCConductorAPI()

    def _get_compute_node(self, context):
        """Override the Compute Node Query to not query by Node Name"""
        try:
            return compute_obj.ComputeNodeList.\
                get_all_by_host(context, self.host)[0]
        except exception.NotFound:
            return None

    def _resource_change(self, old_rscs=None, save=True):
        """Override the Resource Changed method to handle differences"""
        # TODO(tpeoples): remove pci_device_pools from ignore list once the
        # community fixes https://bugs.launchpad.net/nova/+bug/1492342
        ignore_list = ['updated_at', 'pci_device_pools']
        has_changed = not obj_base.obj_equal_prims(
            self.compute_node, self.old_resources, ignore_list)
        # Now we can update the old resources list and return if has changed
        if has_changed:
            self.old_resources = copy.deepcopy(self.compute_node)
        return has_changed

    def _should_notify(self, old_resources):
        """Helper method determine if changes worth notifying or not"""
        # TODO(tpeoples): remove pci_device_pools from ignore list once the
        # community fixes https://bugs.launchpad.net/nova/+bug/1492342
        ignore_list = ['updated_at', 'pci_device_pools', 'metrics']
        has_changed = not obj_base.obj_equal_prims(
            self.compute_node, old_resources, ignore_list)
        return has_changed

    def _verify_resources(self, resources):
        """Override Verify Resources to not wipe out States in Error case"""
        stats = resources.get('stats', {})
        state = stats.get('hypervisor_state')
        # If the only thing in the stats is the state and is error, this means
        # we couldn't connect on startup, so bypass everything and set to error
        if len(stats) <= 4 and (state != 'operating' and state != 'standby'):
            resources = copy.deepcopy(self.old_resources)
            # self.old_resources is init with ComputeNode, so it doesn't
            # have all the attributes loaded.
            # we need to check it before calling .stats
            if resources.obj_attr_is_set('stats'):
                old_stats = resources.stats
                if old_stats and not isinstance(old_stats,
                                                dict):
                    resources.stats = jsonutils.loads(resources.stats)
            else:
                resources.stats = {}
            if resources.stats is None:
                resources.stats = {}
            resources.stats.update(stats)
        else:
            # if the system is powered off, it will not have some of the
            # key values.
            # Now we can do the super class verification of the resource data
            super(PowerVCResourceTracker, self)._verify_resources(resources)

    def _write_ext_resources(self, resources):
        """Override Write Resources so that our Stats aren't wiped out"""
        stats = resources.stats or dict()
        # Convert the stats to a dictionary to be able to merge in
        if stats and not isinstance(stats, dict):
            stats = jsonutils.loads(stats)
        # First call the super class to let it write out the resources
        super(PowerVCResourceTracker, self)._write_ext_resources(resources)
        # Now add back in our stats that were wiped out by the super class
        for key, val in stats.iteritems():
            if key not in resources.stats:
                resources.stats[key] = val
        resources.stats = resources.stats

    def _update(self, context):
        """Override Update Compute Node to send additional notifications"""
        orig_values = copy.deepcopy(self.old_resources)
        old_hostname = self.compute_node.hypervisor_hostname
        info = dict(host=self.host, hypervisor_hostname=self.nodename)
        # Query the Service so that we can fill in the ID in the Notification
        service = service_obj.Service.get_by_compute_host(
            context, self.compute_node.host)
        info['service_id'] = service.id
        info['compute_node_id'] = self.compute_node.id
        info['host_display_name'] = encodeutils.safe_decode(
            cfg.CONF.host_display_name, 'utf-8')
        # Next we need to save off the old maintenance mode toggle to compare
        maint_needed = self._get_current_maintenance_needed(orig_values)
        # Now we can do the super class equivalent to update Compute Node
        self._write_ext_resources(self.compute_node)
        # Send a notification that we are updating the Compute Node
        if self._resource_change():
            # First see if we need to put the host into/out-of Maintenance Mode
            self._update_maintenance_mode_state(maint_needed)
            should_notify = self._should_notify(orig_values)
            # Just because something changed we may not want to notify so
            if should_notify:
                # Next we can send the notification and actually save to the DB
                LOG.info(_(
                    "Sending notification that the Compute Node changed."))
                notifier = rpc.get_notifier(service='compute', host=self.host)
                notifier.info(context, 'compute.node.update.start', info)
            # Call the Scheduler Client that will be the one to do the DB Save
            self.scheduler_client.update_resource_stats(self.compute_node)
            if should_notify:
                notifier.info(context, 'compute.node.update.end', info)
        # See if we need to update the VM's due to the Hostname changing
        new_hostname = self.compute_node.hypervisor_hostname
        if old_hostname and new_hostname and old_hostname != new_hostname:
            self._update_vm_hostnames(context, new_hostname)

    def _report_hypervisor_resource_view(self, resources):
        """Log the hypervisor's view of free resources.

        This is just a snapshot of resource usage recorded by the
        virt driver.

        The following resources are logged:
            - free memory
            - free disk
            - free CPUs
            - assignable PCI devices

        Override, so that the driver can log the appropriate value
        for free disk. Free disk on a host is only applicable in
        local storage environments.
        """
        stats = resources.get('stats', {})
        state = stats.get('hypervisor_state')
        if (len(stats) <= 4 and
                (state != 'operating' and state != 'standby')):
            LOG.debug('Host is not up and cannot update resource view.')
        else:
            if hasattr(self.driver, "log_free_disk"):
                is_audit = False
                self.driver.log_free_disk(is_audit)

                free_ram_mb = (resources['memory_mb'] -
                               resources['memory_mb_used'])
                LOG.debug(_("Hypervisor: free ram (MB): %s") % free_ram_mb)

                vcpus = resources['vcpus']
                if vcpus:
                    free_vcpus = vcpus - resources['vcpus_used']
                    LOG.debug(_("Hypervisor: free VCPUs: %s") % free_vcpus)
                else:
                    LOG.debug(_("Hypervisor: VCPU information unavailable"))

                if 'pci_passthrough_devices' in resources and \
                        resources['pci_passthrough_devices']:
                    LOG.debug(_("Hypervisor: assignable PCI devices: %s") %
                              resources['pci_passthrough_devices'])
                else:
                    LOG.debug(_("Hypervisor: no assignable PCI devices"))
            else:
                super(PowerVCResourceTracker, self). \
                    _report_hypervisor_resource_view(resources)

    def _report_final_resource_view(self):
        """Report final calculate of free memory, disk, CPUs, and PCI devices,
        including instance calculations and in-progress resource claims. These
        values will be exposed via the compute node table to the scheduler.

        Override, so that the driver can log the appropriate value
        for free disk. Free disk on a host is only applicable in
        local storage environments.
        """
        stats = self.compute_node.stats or dict()
        state = stats.get('hypervisor_state')
        if (len(stats) <= 4 and
                (state != 'operating' and state != 'standby')):
            LOG.debug('Host is not up and cannot update resource view.')
        else:
            if hasattr(self.driver, "log_free_disk"):
                is_audit = True
                self.driver.log_free_disk(is_audit)

                LOG.info(
                    _("Free ram (MB): %s") % self.compute_node.free_ram_mb)

                vcpus = self.compute_node.vcpus
                if vcpus:
                    free_vcpus = vcpus - self.compute_node.vcpus_used
                    LOG.info(_("Free VCPUS: %s") % free_vcpus)
                else:
                    LOG.info(_("Free VCPU information unavailable"))

            else:
                super(PowerVCResourceTracker, self).\
                    _report_final_resource_view()

    def _get_current_maintenance_needed(self, orig_values):
        """Helper method to get the current maintenance_mode_needed value"""
        orig_val = orig_values.get('stats', {}).get('maintenance_mode_needed')
        # If the value has already been cached, then use that value
        if orig_val is not None or self.compute_node is None:
            return orig_val
        return self.compute_node.stats.get('maintenance_mode_needed')

    def _update_maintenance_mode_state(self, orig_val):
        """Helper method to put the host into/out-of Maintenance Mode"""
        new_val = self.compute_node.stats.get('maintenance_mode_needed')
        # Either enter or exist Maintenance Mode if the value has changed
        if new_val == 'True' and orig_val != 'True':
            LOG.info(_("Putting Host into Maintenance Mode...."))
            put_host_in_maintenance_mode()
        elif new_val == 'False' and orig_val == 'True':
            LOG.info(_("Taking Host out of Maintenance Mode...."))
            remove_host_from_maintenance_mode()

    def _update_vm_hostnames(self, context, new_hostname):
        """Helper method to update the hostname on VM's to the new hostname"""
        try:
            LOG.info(_("Updating VM's to the new Hostname %s.") % new_hostname)
            # Retrieve all of the Instances for this Host so we can update them
            insts = instance_obj.InstanceList.get_by_host(context, self.host)
            # Loop through each of the Instances, updating the Hostname
            for inst in insts:
                inst.node = new_hostname
                inst.save(context)
        # If we got an exception just log it since don't want to fail
        except Exception as exc:
            LOG.warn(_("Error updating VM's to the new Hostname"))
            LOG.exception(exc)

    def _update_usage_from_instances(self, context, instances):
        """Override the method so we don't try to update instance info
        when the system is down.
        """
        stats = self.compute_node.stats or dict()
        state = stats.get('hypervisor_state')
        if (len(stats) <= 4 and state is not None and
                (state != 'operating' and state != 'standby')):
            LOG.debug('Host is not up and cannot update resource view.')
        else:
            super(PowerVCResourceTracker, self). \
                _update_usage_from_instances(context, instances)


# #########################################
# Override the Compute Node Create/Delete #
# #########################################
def _compute_node_create(self):
    """Override Compute Node create to send additional notifications"""
    info = dict(host=self.host)
    info['hypervisor_hostname'] = self.hypervisor_hostname
    info['host_display_name'] = encodeutils.safe_decode(
        cfg.CONF.host_display_name, 'utf-8')
    # Send a notification that we are creating the Compute Node
    notifier = rpc.get_notifier(service='compute', host=self.host)
    notifier.info(self._context, 'compute.node.create.start', info)
    # Call the super class to actually create the Compute Node
    _orig_compute_node_create(self)
    info['compute_node_id'] = self.id
    # Send a notification that we finished creating the Compute Node
    notifier.info(self._context, 'compute.node.create.end', info)


def _compute_node_destroy(self):
    """Override Compute Node delete to not delete on Hostname change"""
    # Since OpenStack will try to remove Compute Nodes when the Hostname
    # changes and we can't override the code that does this change and
    # the delete call, we will override the delete call to not let them
    # delete it.  We have went with an approach to just update the
    # hypervisor_hostname property on the existing Compute Node instead.
    LOG.info(_("Ignoring Compute Node delete request on hostname change."))

_orig_compute_node_create = compute_obj.ComputeNode.create
compute_obj.ComputeNode.create = _compute_node_create
compute_obj.ComputeNode.destroy = _compute_node_destroy


# #########################################
#  Override the Monitor NameSpace Check   #
# #########################################
def _check_enabled_monitor(self, ext):
    """Override to fix Community NameSpace issues with non-nova prefix"""
    return ext.entry_point_target.startswith('powervc_nova')

MonitorHandler.check_enabled_monitor = _check_enabled_monitor
