#
#
# Copyright 2013, 2016 IBM Corp.
# All Rights Reserved.
#
#    Copyright 2013 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Overrides the Conductor Manager for additional PowerVC DB Access:
    1. Override instance_update to break out the PowerSpecs
    2. Add-in notify_unmanage to allow cleanup of KVM Host when switching
"""

import copy
import datetime
import os
import six
from stevedore import extension
import sys

import nova.context
import oslo_messaging as messaging
from nova import objects
from nova.objects import compute_node
from nova.objects import instance as inst_obj
from oslo_config import cfg
from nova import servicegroup
from nova import notifications
from nova import rpc as rpc_api
from nova import utils
from nova.compute import vm_states
from nova.conductor import manager
from nova.conductor.tasks import live_migrate
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_serialization import jsonutils
from oslo_log import log as logging
from oslo_messaging import exceptions as oslo_messaging_exceptions
from oslo_service import periodic_task
from nova.scheduler import utils as scheduler_utils
from nova.scheduler.filters import affinity_filter

from powervc_nova.common import aggregate_utils as agg_utils
from powervc_nova.compute import populate_admin_context
from powervc_nova.compute import api as compute_api
from powervc_nova.db import api as db_api
# This is needed for PowerVCMigrateData Object registry
from powervc_nova.objects.compute import dom as compute_dom
from powervc_nova.objects.network import dom as network_dom
from powervc_nova.objects.storage import dom as storage_dom
from powervc_nova import _

CONF = cfg.CONF

CONF.register_opts([
    cfg.IntOpt('sync_enterprise_pools_interval', default=300,
               help='Interval to sync enterprise pool info for all hosts')])

CONF.import_opt('powervm_mgr_type', 'powervc_nova.virt.ibmpowervm')

LOG = logging.getLogger(__name__)
HOST_REG = 'powervc_discovery.registration.compute.host_registrar'
AFFINITY_FILTERS = [
    affinity_filter.ServerGroupAffinityFilter().__class__.__name__,
    affinity_filter.ServerGroupAntiAffinityFilter().__class__.__name__]


class PowerVCConductorManager(manager.ConductorManager):
    """Extends the base Conductor Manager class with PowerVC capabilities"""

    def __init__(self, *args, **kwargs):
        """Constructor for the PowerVC extension to the Conductor Manager"""
        super(PowerVCConductorManager, self).__init__(*args, **kwargs)
        self.additional_endpoints.append(_PowerVCConductorManagerProxy(self))
        # Construct a Default Factory from each Module to allow initialization
        self.hmcfac = compute_dom.ManagementConsoleFactory.get_factory()
        self.seafac = network_dom.SharedEthernetAdapterFactory.get_factory()
        self.scgfac = storage_dom.StorageConnectivityGroupFactory.get_factory()

        if CONF.powervm_mgr_type:
            self.servicegroup_api = servicegroup.API()
            self._init_sync_enterprise_pools_file()
        else:
            # If this is KVM, then the sync_enterprise_pools periodic task is
            # irrelevant and shouldn't be wasting any cycles, so remove it.
            for task_name, task in self._periodic_tasks:
                if task_name == self.sync_enterprise_pools.__name__:
                    LOG.debug('Removing defined periodic task: %s' %
                              task_name)
                    self._periodic_tasks.remove((task_name, task))

    def _init_sync_enterprise_pools_file(self):
        path = CONF.oslo_concurrency.lock_path
        if not path:
            raise cfg.RequiredOptError('lock_path')

        self.sync_enterprise_pools_filepath = os.path.join(
            path, 'powervc-sync_enterprise_pools-last-ran-timestamp')

        # This may cause multiple conductor child processes to delete the file
        # when they are instantiated, resulting in the sync enterprise pools
        # logic being ran redundantly. This is better than having to wait
        # potentially sync_enterprise_pools_interval seconds after init for the
        # enterprise pools info to reload, though.
        if os.path.isfile(self.sync_enterprise_pools_filepath):
            try:
                os.remove(self.sync_enterprise_pools_filepath)
            except Exception as e:
                LOG.warn(_('Error encountered when trying to delete file '
                           '%(filepath)s: %(error)s') %
                         {'filepath': self.sync_enterprise_pools_filepath,
                          'error': six.text_type(e)})

    #######################################################
    #           Override Initialization Methods           #
    #######################################################
    def pre_start_hook(self, **kwargs):
        """Override the Pre-Start Hook to perform some initialization"""
        # Call the Parent preStart hook if it ever implements anything
        super(PowerVCConductorManager, self).pre_start_hook(**kwargs)
        try:
            # Make sure we reconcile the Aggregates right away on startup
            self.reconcile_aggregates(nova.context.get_admin_context())
            # We need to create the default SCG the first time Conductor starts
            self.scgfac.create_default_fc_scg(nova.context.get_admin_context())
            # Try to cleanup any on-board tasks stuck when services restarted
            self.cleanup_onboard_tasks(nova.context.get_admin_context())
            # Force the DB API to initialize by doing a random query for HMC's
            self.hmcfac.find_all_hmcs(nova.context.get_admin_context())
        except:
            pass  # We don't care if it fails, since it is just initialization

    @periodic_task.periodic_task(
        spacing=cfg.CONF.reconcile_host_aggregates_interval)
    def reconcile_aggregates(self, context):
        try:
            agg_utils.reconcile_aggregates(
                cfg.CONF.default_host_aggregate_name)
        except Exception as exc:
            LOG.warn(_("Error reconciling host aggregates: %s") % exc)
            LOG.exception(exc)

    @periodic_task.periodic_task(spacing=1800)
    def cleanup_onboard_tasks(self, context):
        """Periodic task to cleanup any stuck on-boarding tasks"""
        try:
            db_api.onboard_task_cleanup(context)
        except Exception as exc:
            LOG.warn(_("Error cleaning up onboard tasks: %s") % exc)
            LOG.exception(exc)

    @lockutils.synchronized('enterprise_pools_periodic_task', 'powervc-',
                            external=True)
    @periodic_task.periodic_task(spacing=CONF.sync_enterprise_pools_interval,
                                 run_immediately=True)
    def sync_enterprise_pools(self, context):
        # The enterprise_pools_periodic_task lock is used to prevent the
        # different conductor child processes from running this at the same
        # time
        try:
            self._sync_enterprise_pools(context)
        except Exception as e:
            LOG.error(_('Failed to sync enterprise pools information.'))
            LOG.exception(e)

    ###################################################
    #      PowerSpec Conductor API Implementation     #
    ###################################################
    def instance_update(self, context, instance_uuid,
                        updates, service):
        """ Update the attributes for the Instance given in the Database """
        # STG Defect 29618 - monkey patch info_from_instance() method
        def patched_info_from_instance(context, instance_ref, network_info,
                                       system_metadata, **kw):
            instance_info = notifications.orig_info_from_instance(
                context, instance_ref, network_info, system_metadata)
            try:
                updated_at = instance_ref.get('updated_at', None)
                if updated_at:
                    instance_info['updated_at'] = updated_at
            except Exception:
                pass
            return instance_info
        if not hasattr(notifications, 'orig_info_from_instance'):
            notifications.orig_info_from_instance = \
                notifications.info_from_instance
            notifications.info_from_instance = patched_info_from_instance

        updates = copy.deepcopy(updates)
        pwr_specs = updates.pop('power_specs', None)
        updates['power_specs'] = dict()
        # Call the Parent's method to actually update the Instance in the DB
        instance_ref = super(PowerVCConductorManager, self).\
            instance_update(context, instance_uuid, updates, service)
        # If there were any Power Specs given, update those in the database
        if pwr_specs is not None:
            pwr_specs = db_api.\
                instance_power_specs_update(context, instance_uuid, pwr_specs)
            instance_ref['power_specs'] = self._convert_power_specs(pwr_specs)
        return instance_ref

    ####################################################
    #           UnManage Host Implementation           #
    ####################################################
    def notify_unmanage(self, context, host, mgmt_ip):
        """Notifies this Management System to remove Host Management"""
        info = dict(hostname=host, ip=mgmt_ip)
        try:
            # Log a message for debug purposes that were are removing the Host
            LOG.info(_("Removing Host %(hostname)s, switching "
                       "to Management System %(ip)s...") % info)
            # Generate a Context with a Token to use for the Requests
            context = populate_admin_context()
            # If the Compute Node doesn't exist, we don't need to notify
            comp_node = self._get_compute_node(context, host)
            if comp_node is None:
                return
            # Notify the old Management System is no longer managing the host
            text = _("The PowerVC management system at %(ip)s is taking over "
                     "management of host %(hostname)s.  The host will be "
                     "removed from this management system.") % info
            anotifier = rpc_api.get_notifier('compute', host)
            anotifier.info(context, 'compute.instance.log', {'msg': text})
            try:
                __import__(HOST_REG)
                # Call the Host Registrar to do the full clean-up of the Host
                get_registrar = getattr(sys.modules[HOST_REG], 'get_registrar')
                registrar = get_registrar(context, host_name=host)
                registrar.skip_remote_commands = True
                registrar.deregister(force=True)
            except Exception as ex:
                LOG.warn(_("Exception trying to fully remove the Host."))
                LOG.exception(ex)
                # Send a notification that we are removing the Compute Node
                anotifier.info(context, 'compute.node.delete.start', comp_node)
                # Fall-back to just cleaning the DB, if the main flow failed
                hostfact = compute_dom.ManagedHostFactory.get_factory()
                hostfact.delete_host(context, host)
                # Send a notification that we removed the Compute Node
                anotifier.info(context, 'compute.node.delete.end', comp_node)
            # Log a message for debug purposes that we removed the Host
            LOG.info(_("Removed Host %(hostname)s, switching "
                       "to Management System %(ip)s.") % info)
        except Exception as exc:
            # Log the Exception that occurred while trying to Remove the Host
            LOG.warn(_("Failed to remove Host %(hostname)s while "
                       "switching to Management System %(ip)s") % info)
            LOG.exception(exc)

    def unmanage_instances(self, context, host, inst_uuids):
        """Utility method to unmanage all of the vm instances specified"""
        extmgr = extension.ExtensionManager('powervc_nova.server.unmanage')
        # Loop through each of the extensions calling the unmanage method
        for ext in extmgr.extensions:
            ext.plugin(context, host).unmanage_resources(inst_uuids, True)

    def _sync_enterprise_pools(self, context):
        # Check the last time a conductor child process started running the
        # enterprise pools sync. If it has been too long since a child process
        # has ran this periodic task, then this child process should run it.

        # Conductor child processes signify that they started running this
        # periodic task by updating the modification time of a file.

        if not os.path.isfile(self.sync_enterprise_pools_filepath):
            # If the file is not yet created, then that means no other
            # conductor child processes have ran this periodic task. This
            # child process should go ahead and run it.
            os.close(os.open(self.sync_enterprise_pools_filepath,
                             os.O_CREAT | os.O_EXCL, 0o600))
        else:
            # File is already created. Check to see the last modification time
            # of the file.
            last_modification_time = datetime.datetime.fromtimestamp(
                os.stat(self.sync_enterprise_pools_filepath).st_mtime)

            seconds_since_last_modification = (datetime.datetime.now() -
                                               last_modification_time).seconds

            if (seconds_since_last_modification <
                    CONF.sync_enterprise_pools_interval):
                LOG.debug('It has only been %s seconds since the last time a '
                          'conductor child process has tried to sync the '
                          'enterprise pool data from PowerVM. Therefore this '
                          'conductor child process will not try to sync with '
                          'PowerVM and will instead just return.' %
                          seconds_since_last_modification)
                return

        # Update the last modification time of the file to the current time to
        # signify that this child process is syncing the enterprise pool data
        # from PowerVM.
        os.utime(self.sync_enterprise_pools_filepath, None)

        LOG.debug('Starting to sync enterprise pools information for hosts.')

        compute_services = objects.ServiceList.get_by_binary(context,
                                                             'nova-compute')
        alive_hosts = [s.host for s in compute_services if
                       self.servicegroup_api.service_is_up(s)]
        LOG.debug('Hosts that have services considered up in '
                  '_sync_enterprise_pools: %s' % alive_hosts)

        # We need to lock against the following:
        #    1) Syncing enterprise pool data, which includes:
        #        a) Fetching the data from PowerVM
        #        b) Writing any updates of the data to the database
        #    2) Setting the mobile cores for a host, which includes:
        #        a) Putting data to PowerVM
        #        b) Writing updates upon success to the database
        #
        # We definitely need to lock against multiple 1s running at the same
        # time due to potential races in between 1a and 1b, which is
        # accomplished here by conducting the syncing serially for each host.
        #
        # We also need to lock against multiple 2s running at the same time due
        # to potential races. This is accomplished by the enterprise_pools lock
        # on the set_mobile_cores method in the PowerVC compute API.
        #
        # We also need to lock against any 1s running at the same time as 2s,
        # because there is a race where 1a is completed, then 2a and 2b finish
        # before 1b is completed for some reason. This would result in stale
        # data from 1a being written to the database in 1b. The lock between
        # 1 and 2 could conceivably be accomplished by placing the same
        # enterprise_pools lock on top of this method. This however would
        # likely cause performance issues if there are hundreds of hosts - a
        # set mobile cores request would have to spin on the lock until every
        # host is synced. Instead we will manually grab and release the lock
        # before and after each host sync so that the set mobile cores method
        # can grab the lock in between if needed.

        compute_rpcapi = compute_api.PowerVCComputeRPCAPI()
        for host in alive_hosts:
            try:
                with lockutils.lock('enterprise_pools', 'powervc-',
                                    external=True):
                    LOG.debug('Starting to sync enterprise pool info for '
                              'host %s.' % host)
                    compute_rpcapi.sync_enterprise_pool_for_host(context, host)
                    LOG.debug('Finished syncing enterprise pool info for '
                              'host %s.' % host)
            except oslo_messaging_exceptions.MessagingTimeout as e:
                LOG.warn(_('Failed to sync enterprise pool information for '
                           'host %(host)s because the RPC call to the '
                           'nova-compute process of the host timed out after '
                           '%(tout)s seconds. Please ensure that there are no '
                           'network connectivity issues or check the host\'s '
                           'nova-compute log for possible details. Timeout '
                           'info: %(exc)s') %
                         {'host': host,
                          'tout': CONF.sync_enterprise_pool_for_host_timeout,
                          'exc': six.text_type(e)})
            except Exception as e:
                LOG.error(_('Failed to sync enterprise pool information for '
                            'host %(host)s.') % {'host': host})
                LOG.exception(e)

        LOG.debug('Finished syncing enterprise pool information for hosts.')

    ####################################################
    #         Internal Conductor Helper Methods        #
    ####################################################
    @staticmethod
    def _get_compute_node(context, host):
        """Helper method to query the Compute Node from the DB"""
        service = objects.Service.get_by_compute_host(context, host)
        # If we weren't able to find the Server or Compute Node, just return
        if service is None or service['compute_node'] is None:
            return None
        # Return the key info from the Compute Node as a dictionary
        compute_node = service['compute_node']
        return dict(compute_node_id=compute_node.id,
                    host=service.host, service_id=service.id,
                    hypervisor_hostname=compute_node.hypervisor_hostname)

    @staticmethod
    def _convert_power_specs(power_specs):
        """Internal Helper Method to Convert PowerSpecs to a Dictionary"""
        lst = ['created_at', 'updated_at', 'deleted_at', 'deleted', 'instance']
        power_specs = jsonutils.to_primitive(power_specs)
        # There are certain properties on the PowerSpecs we don't want returned
        for attr in lst:
            if attr in power_specs:
                power_specs.pop(attr, None)
        return power_specs


def patch_validate_filter(filter_name):
    """
    Monkey patch validate_filter to always return true if called with Nova
    Affinity filters. PowerVC uses its own affinity filters, but still needs
    scheduler_utils::validate_filter to return True during
    scheduler_utils::setup_instance_group.
    """
    if filter_name in AFFINITY_FILTERS:
        return True
    else:
        return filter in cfg.CONF.scheduler_default_filters


setattr(scheduler_utils, 'validate_filter', patch_validate_filter)


###################################################
#    Conductor Manager Adapter Extension Class    #
###################################################
class _PowerVCConductorManagerProxy(object):
    """Adapter Class to extend Remote Methods for the Conductor"""

    target = messaging.Target(version='3.0')

    def __init__(self, manager):
        """Construct a Class to extend Remote Methods for the Conductor"""
        self.manager = manager

    def notify_unmanage(self, context, host, mgmt_ip):
        """Notifies this Management System to remove Host Management"""
        return self.manager.notify_unmanage(context, host, mgmt_ip)

    def unmanage_instances(self, context, host, inst_uuids):
        """Utility method to unmanage all of the vm instances specified"""
        return self.manager.unmanage_instances(context, host, inst_uuids)


###################################################
#           Override Live Migration Task          #
###################################################
"""
When migrating a VM from HMC to NovaLink we need to fix-up the OpenStack UUID
to match the Hypervisor UUID, since NovaLink assumes that they match.  This
was previously done after the migration was finished, but there are issues
with waiting that late, so we will extend the Migration Task to do this up
front before the migration but after the migration task has been established.
"""


class _Extended_LiveMigrationTask(live_migrate.LiveMigrationTask):
    """Extends the default Live Migration Task to add the UUID fix-up"""

    def _execute(self):
        """Override the execute to do the VM UUID fix-ups for NovaLink"""
        self._check_instance_is_active()
        self._check_host_is_up(self.source)
        # Inject our code to fix-up the UUID if going from HMC to NovaLink
        self._fixup_uuid_for_novalink()
        # Let the destination be determined before we inject our UUID fix-up
        if not self.destination:
            self.destination = self._find_destination()
            self.migration.dest_compute = self.destination
            self.migration.save()
        else:
            self._check_requested_destination()
        # Now we can make the RPC call to Compute since we did the fix-up
        return self.compute_rpcapi.live_migration(
            self.context, host=self.source, instance=self.instance,
            dest=self.destination, block_migration=self.block_migration,
            migration=self.migration, migrate_data=self.migrate_data)

    def _fixup_uuid_for_novalink(self):
        """Fixes up the Instance UUID if we are going from HMC to NovaLink"""
        power_specs = self.instance.get('power_specs', dict())
        # Check that vm_uuid exists, as KVM doesn't populate vm_uuid
        if power_specs.get('vm_uuid') is None:
            return
        vm_uuid = power_specs.get('vm_uuid').lower()
        orig_inst_uuid = self.instance.uuid
        # If everything but the first bit matches, already been synch'ed up
        if orig_inst_uuid[1:] == vm_uuid[1:]:
            LOG.debug("Skipping fix-up for %s since UUID matches." % vm_uuid)
            return
        # If we aren't going to NovaLink, then there is no point to convert
        if not self._is_destination_novalink():
            LOG.debug("Skipping fix-up since destination is not NovaLink.")
            return
        # Now we can run the command to do the actual uuid fix-up
        utils.execute('powervc-manage', '-o', 'syncuuids', '-s',
                      self.source, '-u', orig_inst_uuid, run_as_root=True)
        # Refresh the instance so that we have the current uuid info
        self.instance.info_cache.instance_uuid = self.instance.uuid = vm_uuid
        self.migration.instance_uuid = vm_uuid
        self.instance.refresh()
        # Update the UUID in the RequestSpec if the object is set
        if self.request_spec is not None:
            self.request_spec.instance_uuid = vm_uuid
        # Notify the UI that it needs to refresh the Instance Cache
        anotifier = rpc_api.get_notifier('compute', self.instance.host)
        info = notifications.info_from_instance(
            self.context, self.instance, None, None)
        info.update(old_instance_id=orig_inst_uuid, instance_id=vm_uuid)
        anotifier.info(self.context, 'compute.instance.update', info)

    def _is_destination_novalink(self):
        """Helper method to see if the designation is NovaLink or not"""
        # If this is an un-targeted migration we just have to check all hosts
        # to see if any are NovaLink and assume if there are that we are
        # going to that host, since if we wait until the destination is
        # determined it is too late to fix-up the UUID so need to assume.
        # It isn't necessarily harmful to convert even if it goes to HMC.
        if not self.migration.dest_compute:
            nodes = compute_node.ComputeNodeList.get_all(self.context)
        # If we were given a destination, just filter on it instead
        else:
            nodes = compute_node.ComputeNodeList.get_all_by_host(
                self.context, self.migration.dest_compute)
        # Loop through the Compute Nodes seeing if any are NovaLink
        for node in nodes:
            managed_by = node.get('stats', {}).get('managed_by')
            if managed_by and managed_by.lower() == 'novalink':
                return True
        return False


def _build_live_migrate_task(self, context, instance, destination,
                             block_migration, disk_over_commit, migration,
                             request_spec=None):
    "Override the method to build the task to allow our injection"
    return _Extended_LiveMigrationTask(
        context, instance, destination, block_migration,
        disk_over_commit, migration, self.compute_rpcapi,
        self.servicegroup_api, self.scheduler_client, request_spec)

setattr(manager.ComputeTaskManager,
        '_build_live_migrate_task', _build_live_migrate_task)
