# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import json
import os
import pwd
import uuid
import eventlet
from datetime import datetime
from sqlalchemy import MetaData, Table, Column, Text
from sqlalchemy.sql import bindparam, expression, and_

from oslo_config import cfg
import oslo_messaging as messaging
from powervc_nova.config import compute_config
from powervc_oslo.config import data_mgr as mgr
from powervc_oslo.config import data_utils as utils
from powervc_oslo.config.data_utils import V120, V121, V122, V123, V130, V131,\
    V1312
from powervc_nova import _

NOVA_CONF = '/etc/nova/nova.conf'
PASTE_CONF = '/etc/nova/api-paste.ini'
CA_CERTS = '/etc/pki/tls/certs/powervc.crt'
FLAVOR_JSON = '/etc/nova/flavors-config.json'
POWERVC_MANAGE = '/opt/ibm/powervc/bin/powervc-manage'


######################################################
#          Nova Data Manager Definition              #
######################################################
class NovaDataManager(mgr.BaseDataManager):

    def get_component_name(self):
        """Provides the name of the Component providing the Data Manager"""
        return 'nova'

    def get_external_name(self):
        """Provides the external name of the component """
        return 'compute'

    def get_config_desc(self):
        """Provides the description of the component configuration"""
        return _('Manage compute-related settings')

    def is_configurable(self):
        """Provides info on whether config changes are allowed"""
        return True

    def is_auditable(self):
        """Is audit support is available for this component"""
        return True

    def is_debuggable(self):
        """Is debug config support available for this component"""
        return True

    def is_supported_environment(self):
        """Specifies whether the Component is supported in this Environment"""
        return True

    def get_dependent_managers(self):
        """Specifies the Component Name of the Managers it is dependent on"""
        return ['db', 'rabbitmq', 'httpd', 'ego',
                'keystone', 'cinder', 'glance', 'neutron']

    ###############################################
    #     Data Manager - Configure Interface      #
    ###############################################
    def configure_component(self):
        """Performs Configuration specific to the given Component"""
        super(NovaDataManager, self).configure_component()
        # Do the KVM Host-specific Nova Configuration
        if utils.get_compute_node_type() == 'powerkvm':
            config = utils.get_config_data()
            # For the Ephemeral Disk Path we need to also create the directory
            self.__create_ephemeral_disk_path(config)
            # Configure libvirt to be able to work properly for migrations
            self.__configure_libvirt_for_migration(config)
        # On the Management System we need to insert the PowerVM Flavors
        elif utils.get_offering() == 'powervm':
            self.__create_powervm_flavors()
        # Change the permissions for the Nova configuration files
        self.__update_nova_conf_permissions()

    ###############################################
    #      Data Manager - Upgrade Interface       #
    ###############################################
    def upgrade_conf_files(self, version):
        """Performs the upgrade of the Component's CONF Files"""
        # First we need to wipe out all of the extraneous notification_drivers
        if utils.compare_version(version, V123) < 0:
            files = utils.get_matching_files('/etc/nova', 'nova-', '.conf')
            for afile in (files + ['nova.conf']):
                utils.run_command("/bin/sed -i -e '/notification_"
                                  "driver = /d' /etc/nova/%s" % afile)
        # Now call the Super-class to Update the Configuration Files
        super(NovaDataManager, self).upgrade_conf_files(version)
        # Although not related to the CONF files, we need to delete
        # the duplicate Compute Nodes before the Nova DBSync is run
        if utils.is_management_system():
            self.__cleanup_deleted_compute_nodes()

    def upgrade_component(self, version):
        """Performs Upgrade specific to the given Component"""
        # If coming from before 1.2.3, need to insert the new flavors
        if utils.compare_version(version, V123) < 0:
            if utils.get_offering() == 'powervm':
                self.__create_powervm_flavors()
        # Reset disk size of default flavors to zero
        if utils.compare_version(version, V1312) < 0:
            if utils.get_offering() == 'powervm':
                self.__update_powervm_flavors()
        # If coming from before 1.3, need to adjust rsakeys permissions
        if utils.compare_version(version, V130) < 0:
            if os.path.exists('/etc/nova/rsakeys'):
                pw_info = pwd.getpwnam('nova')
                os.chown('/etc/nova/rsakeys', pw_info.pw_uid, pw_info.pw_gid)
                os.chmod('/etc/nova/rsakeys', 0o750)
                for file_name in os.listdir('/etc/nova/rsakeys'):
                    file_path = '/etc/nova/rsakeys/' + file_name
                    os.chown(file_path, pw_info.pw_uid, pw_info.pw_gid)
                    os.chmod(file_path, 0o640)
        # If coming from before 1.3.1, need to default the message project id
        if utils.is_management_system():
            if utils.compare_version(version, V131) < 0:
                self.__default_console_messages_project_id()
        # Change the permissions for the Nova configuration files
        self.__update_nova_conf_permissions()

    ###############################################
    #       Data Manager - Backup Interface       #
    ###############################################
    def prepare_component_restoration(self):
        """Performs any Component-specific preparation needed for restore"""
        config = utils.get_config_data()
        # Generate a unique time-stamp to put in the existing nova.conf files
        timestamp = datetime.now().strftime('%Y%m%d%H%M%S%f')
        config['nova_restore_timestamp'] = timestamp
        # Loop through each CONF file adding a time-stamp to determine changes
        for host in self.__get_compute_hosts():
            conf_file = '/etc/nova/nova-%s.conf' % host
            utils.set_openstack_config_value(
                conf_file, 'restore_timestamp', timestamp)

    def restore_component(self, upgrading):
        """Performs any Component-specific restoration that is needed"""
        config = utils.get_config_data()
        # Loop through each CONF file seeing if the host is new or old
        for host in self.__get_compute_hosts():
            conf_file = '/etc/nova/nova-%s.conf' % host
            val = utils.get_openstack_config_value(
                conf_file, 'restore_timestamp')
            # If file came from the restore, we need to verify service exists
            if val is None or val != config['nova_restore_timestamp']:
                self.__update_new_host(host)
            # If file is left from before the restore, need to delete it
            else:
                self.__remove_old_host(host)

    def run_database_sync(self, log=False):
        """Performs the Database Synchronization for the given Component"""
        oldver = utils.get_config_data().get('restored_version')
        # If we are upgrading from 1.2.2.x then we need to move over the
        # flavor information before the OpenStack DB Sync runs, since they
        # check to make sure it is moved and blow up if it hasn't been
        if oldver and utils.compare_version(oldver, V123) < 0:
            self.__move_instance_flavor_data()
        # Now call the Super-class to Run the DB Synchronization Commands
        super(NovaDataManager, self).run_database_sync(log)
        # Need to initialize the migrate repository first for nova_api
        self.__initialize_api_database_sync()
        utils.run_command('/usr/bin/nova-manage api_db sync', throw=True)

    ###############################################
    #      Data Manager - Synchronize UUID's      #
    ###############################################
    def uuids_to_synchronize(self, host, vm_uuid=None):
        """Retrieves the UUID's that need to be Synchronized for the Host"""
        inst_map = self.__get_instances_to_synchronize(host, vm_uuid)
        return dict(instances=inst_map) if inst_map else None

    def synchronize_uuids(self, host, uuid_map):
        """Synchronizes the UUID's that are needed for the given Host"""
        self.__synchronize_instances(host, uuid_map.get('instances', {}))

    ###############################################
    #    Data Manager - Purge Database Records    #
    ###############################################
    def purge_deleted_database_records(self, timestamp, chunksize):
        """Purge deleted records from the Component's Database"""
        self.__purge_deleted_instances(timestamp, chunksize)

    ######################################################
    ####  Data Manager - powervc-config cmd parser  ######
    ######################################################
    def invoke_config_parser(self, subparser):
        """Invokes the parser for each component for powervc-config cmd"""
        nova_parser = super(NovaDataManager, self)\
            .invoke_config_parser(subparser)
        compute_config.setup_parser(nova_parser,
                                    self.get_component_name())

    ###############################################
    #   Data Manager - Subclass Implementation    #
    ###############################################
    def _added_policy_file(self, config, cat):
        self._added_old_policy_file(config, cat)

    def _deleted_policy_properties(self, config):
        """Helper method to delete policy file path"""
        pass

    def _get_service_user(self):
        """Provides the name of the Service User for the Component"""
        return 'nova'

    def _get_enabled_services(self):
        """Provides a list of the Services for the Component"""
        # If this is the Management System, we need different services
        if utils.is_management_system():
            services = ['api', 'conductor', 'scheduler']
            services.extend(self.__get_compute_services())
            return ['openstack-nova-%s' % svc for svc in services]
        return ['openstack-nova-compute']

    def _get_disabled_services(self):
        """Provides a list of the Disabled Services for the Component"""
        # We want to disable a bunch of services if this is the Mgmt System
        services = ['cells', 'cert', 'compute', 'console', 'consoleauth',
                    'metadata-api', 'network', 'xvpvncproxy']
        return ['openstack-nova-%s' % svc for svc in services]

    def _get_service_log_files(self):
        """Provides a list of Service Log Files for the Component"""
        logs = ['nova-api', 'nova-conductor', 'nova-scheduler',
                'nova-manage', 'httpd-error', 'powervc-dbsync']
        return ['/var/log/nova/%s.log' % log for log in logs]

    def _get_database_info(self):
        """Provides the Database Name/User Tuple for the Component"""
        return ('nova', 'novadb')

    def _get_dbsync_commands(self):
        """Provides a command list for the Component's DB Synchronization"""
        return ['/usr/bin/nova-manage db sync',
                '/opt/ibm/powervc/bin/powervc-dbsync nova']

    def _get_added_conf_properties(self):
        """Provides a dictionary of configuration properties to add"""
        config = {NOVA_CONF: dict()}
        # If this is the Mgmt System, also need to update the Host Conf Files
        if utils.is_management_system():
            for host in self.__get_compute_hosts():
                config['/etc/nova/nova-%s.conf' % host] = dict()
        # Loop through each of the Config Files, adding the Properties
        for key in config.keys():
            self._added_auth_conf_properties(config[key])
            self._added_messaging_conf_properties(config[key])
            self._added_ras_conf_properties(config[key])
            self.__added_nova_common_conf_properties(config[key])
            # Add the Specific Properties for Mgmt System vs KVM Host
            if utils.is_management_system():
                self.__added_nova_mgmtsystem_conf_properties(config[key])
            else:
                self.__added_nova_remotehost_conf_properties(config[key])
        # Need to set back the api-paste to use the legacy v2 extensions
        if utils.is_management_system():
            paste_dict = config.setdefault(PASTE_CONF, dict())
            api_dict = paste_dict.setdefault('composite:osapi_compute', dict())
            api_dict['/v1.1'] = ('openstack_compute_api_legacy_v2', V130)
            api_dict['/v2'] = ('openstack_compute_api_legacy_v2', V130)
        return config

    def _get_deleted_conf_properties(self):
        """Provides a dictionary of configuration properties to remove"""
        config = {NOVA_CONF: dict()}
        # If this is the Mgmt System, also need to update the Host Conf Files
        if utils.is_management_system():
            for host in self.__get_compute_hosts():
                config['/etc/nova/nova-%s.conf' % host] = dict()
        # Loop through each of the Config Files, adding the Properties
        for key in config.keys():
            self._deleted_auth_conf_properties(config[key])
            self._deleted_messaging_conf_properties(config[key])
            self.__deleted_nova_conf_properties(config[key])
        return config

    def _get_audit_conf_properties(self, version=None):
        """Provides a tuple of configuration properties for auditing"""
        if version and utils.compare_version(version, V130) < 0:
            sects = ['composite:openstack_compute_api_v2',
                     'composite:openstack_compute_api_v21',
                     'composite:openstack_compute_api_v3']
        else:
            sects = ['composite:openstack_compute_api_legacy_v2',
                     'composite:openstack_compute_api_v21',
                     ('composite:openstack_compute_api_v21_legacy_v2_' +
                      'compatible')]
        return (PASTE_CONF, sects, ['keystone', 'keystone_nolimit'])

    def _get_included_backup_files(self):
        """Provides a list of Component files to include in a backup"""
        return ['/etc/nova']

    def _get_excluded_backup_files(self):
        """Provides a list of Component files to exclude from a backup"""
        return ['/etc/nova/rtpolicy_conf.xml']

    def _get_certificate_authority_info(self):
        """Provides the info to create the Component's CA Certificate"""
        offer, comptype = utils.get_offering(), utils.get_compute_node_type()
        ca_info = dict(component='libvirt', altnamexts=[],
                       extensions=['ca', 'client', 'server'])
        ca_info['privkey'] = '/etc/pki/libvirt/private/ca.key'
        ca_info['cacert'] = '/etc/pki/libvirt/certs/ca.crt'
        # Only want to return the Libvirt Certificate Info if PowerKVM
        if offer == 'powerkvm' or comptype == 'powerkvm':
            return ca_info

    def _get_remotehost_certificate_info(self):
        """Provides the info to create the Remote Host Signed Certificates"""
        offer, comptype = utils.get_offering(), utils.get_compute_node_type()
        cacert = ('/etc/pki/libvirt/certs/ca.crt', '/etc/pki/CA/cacert.pem')
        ccsr = dict(certfile='clientcert.pem',
                    keyfile='clientkey.pem', exttype='client')
        scsr = dict(certfile='servercert.pem',
                    keyfile='serverkey.pem', exttype='server')
        # Only want to return the Libvirt Certificate Info if PowerKVM
        if offer == 'powerkvm' or comptype == 'powerkvm':
            return dict(cacert=cacert, certdir='/etc/pki/libvirt',
                        group='root', csrs=[ccsr, scsr])

    def _get_service_config_file(self):
        """Provides the configuration file for the given component"""
        return NOVA_CONF

    ###############################################
    #    Internal Data Manager Helper Methods     #
    ###############################################
    def __get_compute_hosts(self):
        """Helper method to get a list of registered compute hosts"""
        files = utils.get_matching_files('/etc/nova', 'nova-', '.conf')
        return [f[5:-5] for f in files]

    def __get_compute_services(self):
        """Helper method to get a list of nova-compute-xxx services"""
        files1 = utils.get_matching_files(
            '/etc/init.d', 'openstack-nova-compute-', '')
        files2 = utils.get_matching_files(
            '/usr/lib/systemd/system', 'openstack-nova-compute-', '')
        if files1:
            return [f[15:] for f in files1]
        return [f[15:-8] for f in files2]

    def __added_nova_common_conf_properties(self, config):
        """Add in the Common nova.conf Configuration Properties"""
        drivers = ['ceilometer.compute.nova_notifier',
                   'messaging', 'nova.ibm.ego.notifier.rpc_notifier']
        for sect in ['cinder', 'glance', 'neutron']:
            config.setdefault(sect, dict())
        # Add in the properties for the Cinder Section
        config['cinder'].update(dict(
            timeout=('1800', V123),
            catalog_info=('volume:cinder:publicURL', V123)))
        # Add in the properties for the Glance Section
        uri = '%(local_ip)s' if utils.is_management_system() else '%(mgmt_ip)s'
        config['glance'].update(dict(
            host=(uri, V122), port=('9292', V122)))
        # Add in the properties for the Neutron Section
        config['neutron'].update(dict(
            timeout=('60', V123),
            auth_plugin=('v3password', V130),
            auth_url=('https://%s:5000/v3/' % uri, V130),
            user_domain_name=('Service', V130),
            username=('neutron', V130),
            password=('aes:%(neutron_password)s', V130),
            project_domain_name=('Service', V130),
            project_name=('service', V130)))
        # Add in the properties for the Default Section
        config['DEFAULT'].update(dict(
            verbose=('True', V130),
            log_dir=('/var/log/nova', V130),
            state_path=('/var/lib/nova', V130),
            notification_driver=(drivers, V123),
            allow_resize_to_same_host=('True', V130),
            instance_build_timeout=('7200', V120),
            default_log_levels=('urllib3=WARN', V120),
            running_deleted_instance_action=('log', V123),
            running_deleted_instance_poll_interval=('86400', V130),
            heal_instance_info_cache_interval=('180', V123),
            ibm_notification_topics=('ibm_notifications', V120),
            use_ipv6=(str(utils.get_config_data()['ip_version'] == '6'), V122),
            flat_injected=('True', V121), instance_delete_interval=('0', V123),
            service_down_time=('120', V122), resize_confirm_window=('1', V120),
            report_interval=('30', V122), management_ip=('%(mgmt_ip)s', V120),
            instance_name_template=(
                '<%(display_name).13s-%(uuid).8s-%(id)08x>', V120),
            injected_network_template=(
                '<$pybasedir/powervc_nova/virt/interfaces.template>', V121)))

    def __added_nova_mgmtsystem_conf_properties(self, config):
        """Add in the Management System nova.conf Configuration Properties"""
        for sect in ['cinder', 'glance', 'neutron', 'database', 'conductor']:
            config.setdefault(sect, dict())
        # Add in the properties for the Cinder/DB/Conductor Sections
        config['cinder']['insecure'] = ('True', V123)
        config['database'].update(dict(
            pool_timeout=('60', V123),
            max_overflow=('100', V123), max_pool_size=('30', V123),
            connection=('aes:mysql://novadb:%(novadb_'
                        'password)s@%(local_ip)s:50110/nova', V131)))
        config['api_database'] = dict(
            connection=config['database']['connection'])
        config['conductor'].update(dict(
            use_local=('True', V120), workers=('4', V121),
            manager=('powervc_nova.conductor.'
                     'manager.PowerVCConductorManager', V120)))
        # Add in the properties for the Glance Section
        config['glance'].update(dict(
            protocol=('http', V122), api_insecure=('True', V122),
            api_servers=('http://%(local_ip)s:9292', V122)))
        # Add in the properties for the Neutron Section
        config['neutron'].update(dict(
            insecure=('True', V123),
            url=('http://%(local_ip)s:9696/', V123)))
        # Add in the properties for the Oslo Currency Section
        config['oslo_concurrency'] = \
            dict(lock_path=('/var/lib/nova/tmp', V123))
        # Add in the properties for the Default Section
        config['DEFAULT'].update(dict(
            osapi_compute_listen_port=('8774', V121),
            quota_server_groups=('25', V130),
            quota_server_group_members=('100', V123),
            install_ip_option=('%(ip_option)s', V122),
            install_ip_version=('%(ip_version)s', V122),
            install_ip_interface=('%(ip_interface)s', V123),
            osapi_compute_listen=('%(local_ip)s', V121),
            api_paste_config=('/etc/nova/api-paste.ini', V120),
            osapi_glance_link_prefix=('https://%(mgmt_ip)s:9292', V121),
            osapi_compute_link_prefix=('https://%(mgmt_ip)s:8774', V121),
            quota_volumes=('10000', V130),
            metadata_workers=('4', V121), osapi_compute_workers=('4', V121),
            enabled_apis=('osapi_compute', V122),
            osapi_compute_extension=(
                ['powervc_nova.api.plugins.api_extensions_v2',
                 'powervc_nova.api.plugins.api_extensions_v3'], V130),
            volume_api_class=(
                'powervc_nova.volume.cinder.PowerVCVolumeAPI', V122),
            non_inheritable_image_properties=(
                'cache_in_nova, bittorrent, block_device_mapping', V123)))
        # Add the Properties specific to whether it is PowerVM or PowerKVM
        if utils.get_offering() == 'powervm':
            self.__added_nova_powervm_conf_properties(config)
        else:
            self.__added_nova_powerkvm_conf_properties(config)

    def __added_nova_remotehost_conf_properties(self, config):
        """Add in the Remote Host nova.conf Configuration Properties"""
        gconfig = utils.get_config_data()
        for sect in ['cinder', 'glance', 'neutron', 'ssl']:
            config.setdefault(sect, dict())
        # Add in the properties for the SSL/Cinder Section
        config['ssl']['ca_file'] = (CA_CERTS, V122)
        config['cinder']['cafile'] = (CA_CERTS, V123)
        # Add in the properties for the Glance Section
        config['glance'].update(dict(
            protocol=('https', V122),
            api_servers=('https://%(mgmt_ip)s:9292', V122)))
        # Add in the properties for the Neutron Section
        config['neutron'].update(dict(
            ca_certificates_file=(CA_CERTS, V123),
            url=('https://%(mgmt_ip)s:9696/', V122)))
        # Add in the properties for the Default Section
        config['DEFAULT'].update(dict(
            dhcp_domain=(gconfig.get('dhcp_domain') or '', V121),
            host_display_name=('%(host_display_name)s', V121),
            vnc_enabled=('False', V121), ca_certificates_file=(CA_CERTS, V121),
            firewall_driver=('nova.virt.firewall.NoopFirewallDriver', V122),
            network_allocate_retries=('5', V123)))
        # Add the Properties specific to whether it is a PVM or KVM Host
        if utils.get_compute_node_type() == 'powerkvm':
            self.__added_nova_kvmhost_conf_properties(config)
        elif utils.get_compute_node_type() == 'powervm':
            self.__added_nova_pvmhost_conf_properties(config)

    def __added_nova_kvmhost_conf_properties(self, config):
        """Add in the KVM Host nova.conf Configuration Properties"""
        config.setdefault('libvirt', dict())
        # Add in the properties for the Libvirt Section
        config['libvirt'].update(dict(
            virt_type=('kvm', V121), libvirt_type=('kvm', V121),
            cpu_mode=('none', V121), inject_partition=('-2', V121),
            live_migration_uri=('<qemu+tls://%s/system>', V121),
            volume_drivers=('iscsi=powervc_nova.virt.powerkvm.'
                            'volume.PowerKVMSCSIVolumeDriver', V122),
            live_migration_flag=('VIR_MIGRATE_UNDEFINE_SOURCE,VIR_MIGRATE_'
                                 'PEER2PEER,VIR_MIGRATE_TUNNELLED', V121)))
        # Add in the properties for the Default Section
        config['DEFAULT'].update(dict(
            compute_monitors=('virt_driver', V130),
            enable_kvm_ovs_auto_config=('True', V121),
            instances_path=('%(instances_path)s', V121),
            state_path=('/var/lib/libvirt/images/nova', V121),
            host_type=('kvm', V121), integration_bridge=('br-int', V121),
            compute_driver=(
                'powervc_nova.virt.powerkvm.driver.PowerKVMDriver', V121),
            compute_manager=(
                'powervc_nova.compute.kvm_manager.PowerVCKvmComputeManager',
                V130),
            volume_api_class=('powervc_nova.volume.cinder.PowerVCVolumeAPI',
                              V130),
            network_api_class=('powervc_nova.network.neutronv2.'
                               'powerkvm_api.PowerKVMAPI', V121),
            max_concurrent_live_migrations=('16', V130)))

    def __added_nova_pvmhost_conf_properties(self, config):
        """Add in the PowerVM Novalink nova.conf Configuration Properties"""
        config.setdefault('powervm', dict())
        # Add in the properties for the Default Section
        config['DEFAULT'].update(dict(
            compute_monitors=('powervc_cpu_virt_driver', V131),
            host_type=('powervm', V130), powervm_mgr_type=('pvm', V130),
            compute_driver=('powervc_nova.virt.ibmpowervm.'
                            'pvm.driver.PowerVMDriver', V130),
            compute_manager=('powervc_nova.compute.pvm_manager.'
                             'PowerVCPvmComputeManager', V130),
            powervm_vif_topo_driver=('powervc_nova.virt.ibmpowervm.vif.pvm.top'
                                     'o_pvm.IBMPowerVMNetworkTopoPVM', V130),
            network_api_class=('powervc_nova.network.neutronv2.'
                               'ibmpowervm_api.PowerVMAPI', V130),
            volume_api_class=('powervc_nova.volume.cinder.'
                              'PowerVCPVMVolumeAPI', V130),
            send_rmc_busy_message=('True', V130),
            automated_powervm_vlan_cleanup=('True', V130),
            force_config_drive=('True', V131),
            vif_plugging_timeout=('1200', V130),
            block_device_allocate_retries_interval=('3', V130),
            block_device_allocate_retries=('600', V130),
            max_concurrent_live_migrations=('16', V130)))
        # If we are switching from HMC to PVM, set that flag
        switch_key, gconfig = 'powervm_switch_type', utils.get_config_data()
        if gconfig.get(switch_key):
            config['DEFAULT'][switch_key] = (gconfig[switch_key], V130)
        # Add the properties in to store the NVRAM information in Swift
        config['powervm'].update(dict(
            swift_username=('nova', V131),
            swift_project_name=('powervm', V131),
            swift_user_domain_name=('Service', V131),
            swift_project_domain_name=('Service', V131),
            swift_password=('aes:%(nova_password)s', V131),
            swift_auth_url=(' https://%(mgmt_ip)s:5000/v3/', V131),
            swift_cacert=(CA_CERTS, V131), nvram_store=('swift', V131)))

    def __added_nova_powervm_conf_properties(self, config):
        """Add in the PowerVM managed by HMC nova.conf
           Configuration Properties"""
        netclass = 'powervc_nova.network.neutronv2.ibmpowervm_api.PowerVMAPI'
        # Add in the properties for the Default Section
        config['DEFAULT'].update(dict(
            compute_monitors=('powervc_cpu_virt_driver', V131),
            network_api_class=(netclass, V121),
            force_config_drive=('True', V131),
            send_rmc_busy_message=('True', V122),
            supported_host_types=('powervm', V122),
            block_device_allocate_retries=('600', V130),
            automated_powervm_vlan_cleanup=('True', V122),
            block_device_allocate_retries_interval=('3', V130),
            powervm_img_local_path=('/var/lib/nova/images', V120),
            powervm_mgr_type=('hmc', V122), host_type=('powervm', V120),
            compute_manager=(
                'powervc_nova.compute.hmc_manager.PowerVCHmcComputeManager',
                V130),
            quota_ram=('40000000', V130),
            quota_cores=('55000', V130),
            quota_instances=('5500', V130),
            osapi_max_limit=('6000', V130),
            max_concurrent_live_migrations=('16', V130)))

    def __added_nova_powerkvm_conf_properties(self, config):
        """Add in the PowerKVM-specific nova.conf Configuration Properties"""
        netclass = 'powervc_nova.network.neutronv2.powerkvm_api.PowerKVMAPI'
        # Add in the properties for the Default Section
        config['DEFAULT'].update(dict(
            network_api_class=(netclass, V121),
            enable_kvm_ovs_auto_config=('True', V122),
            host_type=('kvm', V121), supported_host_types=('kvm', V122),
            quota_ram=('24000000', V123),
            quota_cores=('35000', V123),
            quota_instances=('3500', V123),
            osapi_max_limit=('4000', V123),
            disk_allocation_ratio=('1.0', V131)))

    def __deleted_nova_conf_properties(self, config):
        """Remove the Common nova.conf Configuration Properties"""
        # Add properties to this list to remove them from the conf file.
        # The version value should be the version in which the property
        # is removed in, not the version it was introduced in.
        # Add in the properties for the Default Section
        for sect in ['neutron']:
            config.setdefault(sect, dict())
        config['DEFAULT'].update(dict(
            sql_connection=V123, lock_path=V123,
            neutron_admin_password=V123, neutron_url=V122,
            cinder_api_insecure=V123, cinder_http_timeout=V123,
            glance_api_servers=V122, glance_api_insecure=V122,
            glance_host=V122, glance_port=V122, glance_protocol=V122,
            neutron_admin_username=V122, neutron_api_insecure=V122,
            neutron_admin_auth_url=V122, neutron_auth_strategy=V122,
            neutron_admin_tenant_name=V122, cinder_ca_certificates_file=V123,
            backdoor_port=V123, san_disk_prefix=V130,
            powervm_img_remote_path=V130, compute_available_monitors=V130,
            resize_to_same_host=V131))
        # Add in the properties for the neutron Section
        config['neutron'].update(dict(auth_strategy=V130,
                                      admin_auth_url=V130,
                                      admin_username=V130,
                                      admin_password=V130,
                                      admin_tenant_name=V130))

    def __cleanup_deleted_compute_nodes(self):
        """Helper method to remove the deleted Compute Node entries"""
        # Since for some inexplicable reason, OpenStack decided to put a
        # Unique Constraint on the Compute Nodes DB Table that didn't include
        # the deleted flag, we need to remove the deleted entries to avoid
        # having a unique constraint error occurring when the DBSync is run
        engine = utils.get_database_engine(NOVA_CONF, 'connection', 'database')
        metadata = MetaData(bind=engine, schema='nova')
        table = Table('compute_nodes', metadata, autoload=True)
        table.delete().where(table.c.deleted != 0).execute()

    def __update_new_host(self, host):
        """Helper method to make sure the service exists for the host"""
        service = 'openstack-nova-compute-%s' % host
        conf_file = '/etc/nova/nova-%s.conf' % host
        manager_type = utils.get_openstack_config_value(
            conf_file, 'powervm_mgr_type')
        # We only want to create the service if this is an HMC-managed Host
        if manager_type == 'hmc':
            if not self.__service_file_exists(host):
                utils.move_file(conf_file, conf_file + '_backup')
                utils.run_command(POWERVC_MANAGE + ' -o addhost -s ' +
                                  service + ' -f ' + conf_file + '_backup')
                utils.remove_file(conf_file + '_backup')

    def __remove_old_host(self, host):
        """Helper method to make sure the service is removed for old host"""
        service = 'openstack-nova-compute-%s' % host
        conf_file = '/etc/nova/nova-%s.conf' % host
        manager_type = utils.get_openstack_config_value(
            conf_file, 'powervm_mgr_type')
        # We only want to delete the service if this is an HMC-managed Host
        if manager_type == 'hmc':
            if self.__service_file_exists(host):
                utils.run_command(
                    POWERVC_MANAGE + ' -o removehost -s ' + service)
            utils.remove_file(conf_file)

    def __service_file_exists(self, host):
        """Helper method to see if the service file exists for the host"""
        svc = 'openstack-nova-compute-%s' % host
        svc1 = '/etc/rc.d/init.d/%s' % svc
        svc2 = '/usr/lib/systemd/system/%s.service' % svc
        return utils.file_exists(svc1) or utils.file_exists(svc2)

    def __update_nova_conf_permissions(self):
        """Helper method to update the permissions for the nova conf files"""
        utils.chown('/etc/nova', 'nova', 'nova')
        utils.chown('/etc/nova/nova.conf', 'nova', 'nova')

    def __configure_libvirt_for_migration(self, config):
        """Helper method to configure libvirt to work with migration"""
        # Open up the libvirt-tls service through firewalld
        utils.run_command('firewall-cmd --permanent --add-service=libvirt-tls')
        utils.run_command('firewall-cmd --reload')
        # If a NFS Ephemeral Disk Path, we need to open up SELinux for it
        if config.get('host_storage_type') == 'nfs':
            utils.run_command('/sbin/setsebool -P virt_use_nfs 1')
        # Update the libvirtd config file to listen for incoming connections
        utils.run_command('/bin/sed -i \'/#LIBVIRTD_ARGS="--listen"/c\\'
                          'LIBVIRTD_ARGS="--listen"\' /etc/sysconfig/libvirtd')
        # Restart the libvirtd service when we are done to take affect
        utils.run_command('/bin/systemctl restart libvirtd')

    def __create_ephemeral_disk_path(self, config):
        """Helper method create the State/Instances Path if doesn't exist"""
        state_path = '/var/lib/libvirt/images/nova'
        subdirs = ['buckets', 'images', 'instances', 'keys', 'networks', 'tmp']
        self.__prep_nova_path_directory(state_path, subdirs)
        # If the Instances Path is provided, make sure nova has access to them
        if config.get('instances_path'):
            self.__prep_nova_path_directory(config['instances_path'], [''])

    def __prep_nova_path_directory(self, adir, subdirs):
        """Helper method to make sure nova has access to a directory"""
        for subdir in ([''] + subdirs):
            full_path = adir + '/' + subdir
            # If the Path doesn't exist yet, then we need to create it
            if not utils.file_exists(full_path):
                su = '/usr/bin/su -s /bin/bash nova -c "%s"'
                utils.run_command(su % ('/bin/mkdir -p -m 755 ' + full_path))
            # As a fall-back, try creating as root if nova doesn't have access
            if not utils.file_exists(full_path):
                utils.makedirs(full_path)
            # Make sure that Nova is the Owner and has permission to it
            utils.chown(full_path, 'nova', 'nova')
            # Verified permission issue found by Bandit scan
            utils.chmod(full_path, 0o755)  # nosec
        # We now want to make sure Nova has permission all the way up the chain
        self.__grant_search_to_parents(adir)

    def __grant_search_to_parents(self, adir):
        """Helper method to grant public excute to the directory and parents"""
        utils.run_command('/usr/bin/chmod o+x ' + adir)
        if adir.rfind('/') > 0:
            self.__grant_search_to_parents(adir[:adir.rfind('/')])

    def __create_powervm_flavors(self):
        """Helper method to create the Flavors specific to PowerVM"""
        flavor_data = json.loads(utils.read_from_file(FLAVOR_JSON))
        global_specs = flavor_data.pop('pvm_global_extra_specs')
        # Get a connection to the Database and load the needed Tables
        engine = utils.get_database_engine(NOVA_CONF, 'connection', 'database')
        metadata = MetaData(bind=engine, schema='nova')
        flv_table = Table('instance_types', metadata, autoload=True)
        spc_table = Table('instance_type_extra_specs', metadata, autoload=True)
        flav_insert, spec_insert = flv_table.insert(), spc_table.insert()
        connection = engine.connect()
        # Loop through each of the flavor info, creating the database entries
        for flavor_key in flavor_data:
            # If this is the extra specs, it is just an auxiliary entry
            if flavor_key.endswith('_extra_specs'):
                continue
            # Get the entry and extra specs from the overall json dictionary
            flavor = flavor_data.get(flavor_key)
            flavor_specs = flavor_data.get(flavor_key + '_extra_specs')
            flavor_specs.update(global_specs)
            # Insert the Flavor in the instance_types database table
            rs = connection.execute(
                flav_insert, name=flavor['name'], created_at=datetime.utcnow(),
                deleted=0, disabled=0, is_public=1, memory_mb=flavor['ram'],
                root_gb=0, vcpus=flavor['vcpus'], swap=0,
                ephemeral_gb=0, rxtx_factor=1.0, flavorid=str(uuid.uuid4()))
            flavor_id = rs.inserted_primary_key[0]
            # Insert the Extra Specs into the instance_type_extra_specs table
            for key, val in flavor_specs.iteritems():
                spec_row = dict(instance_type_id=flavor_id, deleted=0, key=key,
                                value=val, created_at=datetime.utcnow())
                connection.execute(spec_insert, spec_row)
        connection.close()

    def __update_powervm_flavors(self):
        """Helper method to update the disk size of Flavors specific
           to PowerVM"""
        # Get a connection to the Database and load the needed Tables
        engine = utils.get_database_engine(NOVA_CONF, 'connection', 'database')
        metadata = MetaData(bind=engine, schema='nova')
        flv_table = Table('instance_types', metadata, autoload=True)
        flav_update = flv_table.update()
        connection = engine.connect()
        # Loop through each of the flavor info, as we want to update
        # the database entries only for flavors created by default
        connection.execute(flav_update, root_gb=0)
        connection.close()

    def __move_instance_flavor_data(self):
        """Moves the Flavor Data from System MetaData to Instance Extras"""
        datestr = datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')
        # Need to force loading of the migrate.changeset to add create_colum
        __import__('migrate.changeset')
        # Get a connection to the Database and load the needed Tables
        engine = utils.get_database_engine(NOVA_CONF, 'connection', 'database')
        flavors, metadata = dict(), MetaData(bind=engine, schema='nova')
        # Setup a template string in the format of the new flavor structure
        fmt = '{"new": null, "old": null, "cur": {"nova_object.version": "1.1'\
              '", "nova_object.name": "Flavor", "nova_object.data": {"disable'\
              'd": false, "root_gb": 0, "name": "%(name)s", "flavorid": "%(fl'\
              'avorid)s", "deleted": false, "created_at": "%(date)s", "epheme'\
              'ral_gb": %(ephemeral_gb)s, "updated_at": null, "memory_mb": %('\
              'memory_mb)s, "vcpus": %(vcpus)s, "extra_specs": {}, "swap": %('\
              'swap)s, "rxtx_factor": %(rxtx_factor)s, "is_public": false, "d'\
              'eleted_at": null, "vcpu_weight": %(vcpu_weight)s, "id": %(id)s'\
              ', "projects": ["%(project)s"]}, "nova_object.namespace": "nova'\
              '"}}'
        # Set up the queries from the tables to get the info needed for moving
        inst_table = Table('instances', metadata, autoload=True)
        extr_table = Table('instance_extra', metadata, autoload=True)
        meta_table = Table('instance_system_metadata', metadata, autoload=True)
        # First we need to make sure that the new flavors column exists
        flavor_column = Column('flavor', Text, nullable=True)
        if extr_table._columns.get(flavor_column.name) is None:
            extr_table.create_column(flavor_column.copy())
        # Next we will setup the query and insert/update statement definitions
        extr_query = expression.select([extr_table.c.instance_uuid])
        meta_query = expression.select([meta_table]).where(and_(
            meta_table.c.deleted == 0, meta_table.c.key.like('inst%_type_%')))
        inst_query = expression.select(
            [inst_table.c.uuid, inst_table.c.project_id])
        extr_insert = extr_table.insert().values(
            deleted=0, created_at=datetime.utcnow(), pci_requests='[]',
            instance_uuid=bindparam('uuid'), flavor=bindparam('flavor'))
        extr_update = extr_table.update().where(
            extr_table.c.instance_uuid == bindparam('uuid')).values(
                flavor=bindparam('flavor'))
        # Execute the Query on the Tables and Format for additional Processing
        extr_rows = engine.execute(extr_query).fetchall()
        inst_rows = engine.execute(inst_query).fetchall()
        meta_rows = engine.execute(meta_query).fetchall()
        inst_rows = dict([(row[0], row[1]) for row in inst_rows])
        extr_rows = dict([(row[0], 'extra') for row in extr_rows])
        for row in meta_rows:
            key, val = row['key'][14:], row['value'] if row['value'] else '0'
            flavors.setdefault(row['instance_uuid'], dict())[key] = val
        # Now we can actually go through insert the new Flavor Rows in the DB
        for inst_uuid in inst_rows:
            # Nothing to do if no System MetaData Flavor Info for the Instance
            if inst_uuid not in flavors:
                continue
            # Execute the Update Operation with the additional Flavor Info
            flavor = flavors[inst_uuid]
            flavor['project'], flavor['date'] = inst_rows[inst_uuid], datestr
            # If the Extra Info entry exists we need to do an update instead
            extr_op = extr_update if inst_uuid in extr_rows else extr_insert
            engine.execute(extr_op, uuid=inst_uuid, flavor=fmt % flavor)
        # Finally we can clean up all of the Flavor System MetaData Properties
        engine.execute(meta_table.delete().where(
            meta_table.c.key.like('instance_type_%')))

    def __default_console_messages_project_id(self):
        """Helper method to fill-in project_id on previous Console Messages"""
        # Get a connection to the Nova and Keystone Databases to update
        neng = utils.get_database_engine(NOVA_CONF, 'connection', 'database')
        keng = utils.get_database_engine(
            '/etc/keystone/keystone.conf', 'connection', 'database')
        kmeta = MetaData(bind=keng, schema='keystone')
        nmeta = MetaData(bind=neng, schema='nova')
        # Load the project and console message tables to query/update
        msg_table = Table('ibm_console_messages', nmeta, autoload=True)
        proj_table = Table('project', kmeta, autoload=True)
        # Query the Project Id for the 'ibm-default' project to default to
        proj_query = expression.select([proj_table.c.id]).where(
            proj_table.c.name == 'ibm-default')
        proj_rows = keng.execute(proj_query).fetchall()
        # Now update any console messages without a project_id to use that one
        if len(proj_rows) == 1 and proj_rows[0][0]:
            neng.execute(msg_table.update().where(
                msg_table.c.project_id.is_(None)).values(
                    project_id=proj_rows[0][0]))

    def __load_association_tables(self, metadata):
        """Helper method to load the Association Tables for the Instance"""
        tables = dict()
        assoc_names = [
            'instance_faults', 'instance_metadata', 'instance_health_status',
            'instance_id_mappings', 'instance_info_caches', 'instance_extra',
            'instance_power_specs', 'migrations', 'instance_system_metadata',
            'block_device_mapping', 'security_group_instance_association',
            'instance_actions']
        for name in assoc_names:
            tables[name] = Table(name, metadata, autoload=True)
        return tables

    def __get_table_uuid_column(self, table, table_name):
        """Helper method to get the UUID column from the given table"""
        if table_name == 'instance_health_status':
            return table.c.id
        elif table_name == 'instance_id_mappings':
            return table.c.uuid
        return table.c.instance_uuid

    def __purge_deleted_instances(self, timestamp, chunksize):
        """Helper method to purge deleted instances from the database"""
        engine = utils.get_database_engine(NOVA_CONF, 'connection', 'database')
        metadata = MetaData(bind=engine, schema='nova')
        inst_table = Table('instances', metadata, autoload=True)
        aevt_table = Table('instance_actions_events', metadata, autoload=True)
        assoc_tables = self.__load_association_tables(metadata)
        # Query the Instances table to get the UUID's of the deleted Instances
        inst_query = expression.select([inst_table.c.uuid]).where(and_(
            inst_table.c.deleted > 0, inst_table.c.deleted_at < timestamp))
        conn = engine.connect()
        # We want to close the ResultSet for the query after construct it
        inst_rs = conn.execute(inst_query)
        inst_uuids = [r[0] for r in inst_rs.fetchall()]
        inst_rs.close()
        # Break the UUID's into Chunks so that DB2 can handle
        for index in xrange(0, len(inst_uuids), chunksize):
            uuid_chunk = inst_uuids[index: index + chunksize]
            with conn.begin():
                # Loop through the Association Tables updating those UUID's
                for name, table in assoc_tables.iteritems():
                    uuidcol = self.__get_table_uuid_column(table, name)
                    # If the Instance Action table, need to delete the Events
                    if name == 'instance_actions':
                        query = expression.select([table.c.id]).where(
                            table.c.instance_uuid.in_(uuid_chunk))
                        actids = [r[0] for r in conn.execute(query).fetchall()]
                        # Loop through the Action ID's, deleting them in chunks
                        for index2 in xrange(0, len(actids), chunksize):
                            id_chunk = actids[index2: index2 + chunksize]
                            conn.execute(aevt_table.delete().where(
                                aevt_table.c.action_id.in_(id_chunk)))
                    # Once we know the UUID column, we can delete the entries
                    conn.execute(table.delete().where(uuidcol.in_(uuid_chunk)))
                # Now we can deleted the actual rows from the instance table
                conn.execute(inst_table.delete().where(
                    inst_table.c.uuid.in_(uuid_chunk)))
        conn.close()

    def __get_instances_to_synchronize(self, host, vm_uuid=None):
        """Helper method to get the VM UUID's that need to be synchronized"""
        # Get a connection to the Database and load the needed Tables
        engine = utils.get_database_engine(NOVA_CONF, 'connection', 'database')
        uuid_map, metadata = dict(), MetaData(bind=engine, schema='nova')
        inst_table = Table('instances', metadata, autoload=True)
        spec_table = Table('instance_power_specs', metadata, autoload=True)
        # Query the PowerSpecs table to get the UUID's that are different
        inst_query = expression.select([inst_table.c.uuid])
        spec_query = expression.select(
            [spec_table.c.instance_uuid, spec_table.c.vm_uuid])
        if vm_uuid is None:
            inst_query = inst_query.where(inst_table.c.host.in_([host]))
            spec_query = spec_query.where(and_(
                spec_table.c.deleted == 0,
                spec_table.c.instance_uuid != spec_table.c.vm_uuid))
        # If they gave us a single UUID, then ignore the host and filter on it
        else:
            inst_query = inst_query.where(inst_table.c.uuid == vm_uuid)
            spec_query = spec_query.where(and_(
                spec_table.c.deleted == 0,
                spec_table.c.instance_uuid == vm_uuid,
                spec_table.c.instance_uuid != spec_table.c.vm_uuid))
        # Perform the actual queries of the Instance and PowerSpecs tables
        inst_rows = engine.execute(inst_query).fetchall()
        spec_rows = engine.execute(spec_query).fetchall()
        host_uuids = dict([(row[0], '') for row in inst_rows])
        # Only add the Instances that are for the given Host
        for row in spec_rows:
            if row[0] in host_uuids:
                if row[0].lower() != row[1].lower():
                    uuid_map[row[0]] = row[1].lower()
        return uuid_map

    def __synchronize_instances(self, host, uuid_map):
        """Helper method to synchronize the VM UUID's that are given"""
        all_uuids = uuid_map.keys() + uuid_map.values()
        # Get a connection to the Database and load the needed Tables
        engine = utils.get_database_engine(NOVA_CONF, 'connection', 'database')
        metadata = MetaData(bind=engine, schema='nova')
        inst_table = Table('instances', metadata, autoload=True)
        assoc_tables = self.__load_association_tables(metadata)
        # Query the Instances to see if we need to copy the old to the new
        inst_query = expression.select([inst_table])
        inst_query = inst_query.where(inst_table.c.uuid.in_(all_uuids))
        inst_delete = inst_table.delete().where(
            inst_table.c.uuid == bindparam('uuid'))
        inst_rows = engine.execute(inst_query).fetchall()
        inst_map = dict([(row['uuid'], row) for row in inst_rows])
        # Loop through each of the UUID's updating the Instance-related Tables
        for old_uuid, new_uuid in uuid_map.iteritems():
            # Only want to dupicate the instance if it wasn't done yet
            if old_uuid in inst_map and new_uuid not in inst_map:
                oldrow = inst_map[old_uuid]
                newrow = dict([(c.name, oldrow[c.name]) for c in inst_table.c])
                newrow['uuid'], _x = new_uuid, newrow.pop('id', None)
                engine.execute(inst_table.insert().values(**newrow))
            # Loop through the Association Tables updating those UUID's
            for name, table in assoc_tables.iteritems():
                if name == 'instance_health_status':
                    engine.execute(table.update().where(
                        table.c.id == old_uuid).values(id=new_uuid))
                elif name == 'instance_id_mappings':
                    engine.execute(table.update().where(
                        table.c.uuid == old_uuid).values(uuid=new_uuid))
                else:
                    engine.execute(table.update().where(
                        table.c.instance_uuid == old_uuid).values(
                            instance_uuid=new_uuid))
                # Need to clean up the decision id from the MetaData tables
                # This is needed so that EGO synchronizes its allocations
                if name in ['instance_metadata', 'instance_system_metadata']:
                    engine.execute(table.delete().where(and_(
                        table.c.key == 'decision_id',
                        table.c.instance_uuid == new_uuid)))
            # Now we can finally clean up the old VM instance from the DB
            engine.execute(inst_delete, uuid=old_uuid)
        # Once we are done updating the Instance UUID's, then notify EGO
        if len(uuid_map.keys()) > 1:
            self.__synchronize_ego()

    def __synchronize_ego(self):
        """Helper method to tell EGO to re-synchronize the instances"""
        eventlet.monkey_patch(os=False)
        try:
            # Add a timeout on making the rpc call so doesn't hang forever
            with eventlet.timeout.Timeout(15):
                cfg.CONF([], project='nova')
                cfg.CONF.notification_driver = ['messaging']
                from oslo_context import context
                ctxt = context.get_admin_context()
                transport = messaging.get_transport(cfg.CONF)
                # Make an RPC call to EGO to do the synchronization
                serializer = messaging.RequestContextSerializer(None)
                target = messaging.Target(topic='scheduler', version='4.0')
                client = messaging.RPCClient(
                    transport, target, serializer=serializer)
                client.prepare().cast(ctxt, "ego_sync_up")
        except eventlet.timeout.Timeout:
            utils.LOG.warn(_("Timed out trying to synchronize EGO"))
        except Exception as ex:
            utils.LOG.exception(_("Exception synchronizing EGO: %s") % ex)

    def __initialize_api_database_sync(self):
        """Helper method to initialize migrate_version table for nova_api"""
        engine = utils.get_database_engine(NOVA_CONF, 'connection', 'database')
        metadata = MetaData(bind=engine, schema='nova')
        vers_table = Table('migrate_version', metadata, autoload=True)
        # Query the Versions to see if we need to insert a new record
        vers_query = expression.select([vers_table])
        vers_rows = [i[0] for i in engine.execute(vers_query).fetchall()]
        # Now we can try to insert the missing nova_api record from the db
        if 'nova_api' not in vers_rows:
            path = '/usr/lib/python2.7/site-packages/nova/'\
                   'db/sqlalchemy/api_migrations/migrate_repo'
            engine.execute(vers_table.insert(), repository_id='nova_api',
                           repository_path=path, version=0)
