# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import os
import re
import sys
from datetime import datetime
from sqlalchemy import MetaData, Table

from powervc_oslo.config import data_mgr as mgr
from powervc_oslo.config import data_utils as utils
from powervc_oslo.config.data_utils import V120, V121, V122, V123, \
    V130, V1302, V1312, V999

from powervc_oslo import _

NOVA_CONF = '/etc/nova/nova.conf'
RTPOLICY_FILE = '/etc/nova/rtpolicy_conf.xml'

EGO_USERS_FILE = '/opt/ibm/ego/kernel/conf/users.xml'


######################################################
#           EGO Data Manager Definition              #
######################################################
class EgoDataManager(mgr.BaseDataManager):

    def get_component_name(self):
        """Provides the name of the Component providing the Data Manager"""
        return 'ego'

    def is_supported_environment(self):
        """Specifies whether the Component is supported in this Environment"""
        return utils.is_management_system()

    def get_dependent_managers(self):
        """Specifies the Component Name of the Managers it is dependent on"""
        return ['db', 'rabbitmq', 'httpd', 'keystone']

    ###############################################
    #   Data Manager - Configure Implementation   #
    ###############################################
    def configure_component(self):
        """Performs Configuration specific to the given Component"""
        super(EgoDataManager, self).configure_component()
        # Need to replace the RTPolicy file with the PowerVC one
        offering = utils.get_offering()
        afile = '/etc/nova/rtpolicy_conf_%s.xml.pvc'
        otheroff = 'powerkvm' if offering == 'powervm' else 'powervm'
        utils.copy_file(afile % offering, '/etc/nova/rtpolicy_conf.xml')
        utils.remove_file(afile % otheroff)
        # Need to remove the SetUID from the EGOSH File
        egosh_path = utils.run_command_output(
            '/usr/bin/find /opt/ibm/ego/ -name egosh', False)[1].strip()
        utils.run_command('/usr/bin/chmod -s ' + egosh_path)

    def set_user_passwords(self):
        """Sets/Generates the Passwords for the Component's Users"""
        password = utils.generate_password()
        utils.get_config_data().setdefault('ego_password', password)
        self.__set_ego_admin_password(password)

    ###############################################
    #      Data Manager - Upgrade Interface       #
    ###############################################
    def upgrade_component(self, version):
        """Performs Upgrade specific to the given Component"""
        super(EgoDataManager, self).upgrade_component(version)
        # Only perform this upgrade logic on the management system
        if utils.is_management_system():
            # The following logic is responsible for upgrading to V123 in which
            # host groups were added; find the currently active placement
            # policy and assign that policy to the default host group
            config = utils.get_config_data()
            (tardir, tarfile) = (config.get('restore_dir'),
                                 config.get('restore_file'))
            if tardir and tarfile:
                data = utils.extract_tar_file_content(tardir, tarfile,
                                                      [RTPOLICY_FILE])
                # Upgrade the Aggregate MetaData if we are coming from pre-V123
                if data and utils.compare_version(version, V123) < 0:
                    self.__upgrade_aggregate_metadata_to_v123(tardir, tarfile,
                                                              data[0])

    ###############################################
    #   Data Manager - System Changed Interface   #
    ###############################################
    def management_ip_changed(self, upgrading):
        """Updates the Management IP in the CONF files when they change"""
        utils.set_openstack_config_value(
            NOVA_CONF, 'ego_master_list', utils.get_local_hostname())

    ###############################################
    #      Data Manager - Services Interface      #
    ###############################################
    def start_services(self, log_components=False):
        """Starts the Services for the given Component"""
        # Need to override to start ego separate until switched to SystemD
        utils.run_command('/usr/sbin/service ego start')
        return super(EgoDataManager, self).start_services(log_components)

    def stop_services(self, log_components=False):
        """Stops the Services for the given Component"""
        super(EgoDataManager, self).stop_services(log_components)
        # Need to override to stop ego separate until switched to SystemD
        return utils.run_command('/usr/sbin/service ego stop')

    def services_status(self):
        """Status of the Services for the given Component"""
        # Need to override to check ego separate until switched to SystemD
        utils.run_command('/usr/sbin/service ego status')
        return super(EgoDataManager, self).services_status()

    ###############################################
    #   Data Manager - Subclass Implementation    #
    ###############################################
    def _get_service_user(self):
        """Provides the name of the Service User for the Component"""
        return 'nova'

    def _get_enabled_services(self):
        """Provides a list of the Services for the Component"""
        services = ['ibm-notification', 'ibm-ego-ha-service',
                    'ibm-ego-resource-optimization']
        return ['openstack-nova-%s' % svc for svc in services]

    def _get_service_log_files(self):
        """Provides a list of Service Log Files for the Component"""
        logs = ['nova-ibm-notification', 'nova-ibm-ego-ha-service',
                'nova-ibm-ego-resource-optimization']
        return ['/var/log/nova/%s.log' % log for log in logs]

    def _get_added_conf_properties(self):
        """Provides a dictionary of configuration properties to add"""
        config = {NOVA_CONF: dict()}
        # Add in all of the Common EGO Configuration Properties
        self.__added_ego_common_conf_properties(config[NOVA_CONF])
        # Add the Properties specific to whether it is PowerVM or PowerKVM
        if utils.get_offering() == 'powervm':
            self.__added_ego_powervm_conf_properties(config[NOVA_CONF])
        else:
            self.__added_ego_powerkvm_conf_properties(config[NOVA_CONF])
        return config

    def _get_deleted_conf_properties(self):
        """Provides a dictionary of configuration properties to remove"""
        config = {NOVA_CONF: dict(ibm_prs_ceilometer=dict())}
        config[NOVA_CONF]['ibm_prs_ceilometer'].update(dict(
            auth_url=V130, insecure=V130, project_domain_name=V130,
            project_name=V130, user_domain_name=V130, username=V130,
            password=V130))
        return config

    def _get_service_password_properties(self):
        """Provides a list of the configuration properties for passwords"""
        return [('ego', '/etc/nova/nova.conf', 'DEFAULT', 'ego_password')]

    ###############################################
    #    Internal Data Manager Helper Methods     #
    ###############################################
    def __added_ego_common_conf_properties(self, config):
        """Add in the Management System nova.conf Configuration Properties"""
        config.setdefault('DEFAULT', dict())
        config.setdefault('ibm_prs_ceilometer', dict())
        ego_path = self.__find_ego_lib_path()
        hostname = utils.get_local_hostname()
        # Add in the Common EGO properties for the Default Section
        config['DEFAULT'].update(dict(
            ego_lib_path=(ego_path, V999),
            ego_user=('Admin', V130),
            ego_password=('aes:%(ego_password)s', V130),
            prs_ut_data_sample_num=('30', V123),
            ego_master_list=('%s' % hostname, V120),
            prs_ha_timeout_seconds=('3600', V123),
            policy_per_host_aggregate=('True', V123),
            prs_sync_aggregate_reserveon=('True', V123),
            scheduler_driver=('nova.scheduler.ibm.ego.ego_'
                              'scheduler.EGOScheduler', V120),
            scheduler_manager=('nova.scheduler.ibm.ego.ego_scheduler_'
                               'manager.SchedulerManager', V120),
            prs_mt_migration_preferred_order=('memory_mb.most', V122),
            ha_policy_instance_driver=('powervc_nova.prs.instance_driver.'
                                       'PowerVCHAInstanceDriver', V123),
            ram_allocation_ratio=('1.0', V120), ego_kd_port=('7870', V120),
            ha_policy_hypervisor_driver=('powervc_nova.prs.hypervisor_driver.'
                                         'PowerVCHAHypervisorDriver', V123),
            prs_driver=('powervc_nova.prs.prs_driver.PowerVCPRSDriver', V123),
            rtpolicy_driver=('powervc_nova.prs.rtpolicy_driver.'
                             'PowerVCRTPolicyDriver', V130)))

    def __find_ego_lib_path(self):
        # Find the Path to the EGO Shared Library on this System
        ego_path = utils.run_command_output(
            '/usr/bin/find /opt/ibm/ego/ -name "ego.so"', False)[1].strip()
        ego_path = ego_path[:ego_path.rfind('/') + 1]
        return ego_path

    def __added_ego_powervm_conf_properties(self, config):
        """Add in the PowerVM-specific nova.conf Configuration Properties"""
        config['DEFAULT'].update(dict(
            cpu_allocation_ratio=('1.0', V120),
            reserved_host_memory_mb=('0', V120),
            scheduler_available_filters=(self.__get_scheduler_filters(),
                                         V1312),
            scheduler_default_filters=(
                'RamFilter,CoreFilter,ComputeFilter,RetryFilter,Availability'
                'ZoneFilter,ImagePropertiesFilter,ComputeCapabilitiesFilter,'
                'MaintenanceFilter,PowerVCServerGroupAffinityFilter,PowerVC'
                'ServerGroupAntiAffinityFilter,PowerVCHostAggregateFilter,'
                'PowerVMNetworkFilter,PowerVMProcCompatModeFilter,PowerLMB'
                'SizeFilter,PowerMigrationLicenseFilter,PowerVMMigrationCount'
                'Filter,PowerVMStorageFilter,PowerVMIBMiMobilityFilter,'
                'PowerVMRemoteRestartFilter,PowerVMRemoteRestartSameHMCFilter,'
                'PowerVMEndianFilter,PowerVMGuestCapableFilter,'
                'PowerVMSharedProcPoolFilter,PowerVCResizeSameHostFilter,'
                'PowerVCDROFilter,PowerVMActiveMemoryExpansionFilter,'
                'PowerVMNovaLinkMobilityFilter', V1312),
            rtpolicy_optimizers=('powervc_nova.prs.rtpolicy_cod_optimizer.'
                                 'PowerVCRTPolicyCodOptimizer', V130)))

    def __added_ego_powerkvm_conf_properties(self, config):
        """Add in the PowerKVM-specific nova.conf Configuration Properties"""
        config['DEFAULT'].update(dict(
            cpu_allocation_ratio=('4.0', V121),
            scheduler_available_filters=(self.__get_scheduler_filters(), V130),
            scheduler_default_filters=(
                'RamFilter,CoreFilter,ComputeFilter,RetryFilter,Availability'
                'ZoneFilter,ImagePropertiesFilter,ComputeCapabilitiesFilter,'
                'MaintenanceFilter,DiskFilter,PowerVCServerGroupAffinityFilter'
                ',PowerVCServerGroupAntiAffinityFilter,PowerVCHostAggregate'
                'Filter,PowerKVMNetworkFilter,PowerKVMSMTFilter,PowerKVM'
                'StorageFilter,PowerKVMMigrationVersionFilter,'
                'PowerVCResizeSameHostFilter,PowerVCDROFilter',
                V130)))

    def __get_scheduler_filters(self):
        """Helper method to build a list of scheduler filters"""
        # Make sure to update the Version number above when changing lists
        common_filters = [
            'host_aggregate_filter.PowerVCHostAggregateFilter',
            'affinity_filter.PowerVCServerGroupAffinityFilter',
            'affinity_filter.PowerVCServerGroupAntiAffinityFilter',
            'resize_same_host_filter.PowerVCResizeSameHostFilter',
            'dro_filter.PowerVCDROFilter']
        # Make sure to update the Version number above when changing lists
        powerkvm_filters = [
            'smt_filter.PowerKVMSMTFilter',
            'network_filter.PowerKVMNetworkFilter',
            'storage_filter.PowerKVMStorageFilter',
            'migration_version_filter.PowerKVMMigrationVersionFilter']
        # Make sure to update the Version number above when changing lists
        powervm_filters = [
            'active_mem_exp_filter.PowerVMActiveMemoryExpansionFilter',
            'endian_filter.PowerVMEndianFilter',
            'network_filter.PowerVMNetworkFilter',
            'storage_filter.PowerVMStorageFilter',
            'guest_capable_filter.PowerVMGuestCapableFilter',
            'ibmi_mobility_filter.PowerVMIBMiMobilityFilter',
            'remote_restart_filter.PowerVMRemoteRestartFilter',
            'migration_count_filter.PowerVMMigrationCountFilter',
            'proc_compat_mode_filter.PowerVMProcCompatModeFilter',
            'remote_restart_filter.PowerVMRemoteRestartSameHMCFilter',
            'shared_proc_pool_filter.PowerVMSharedProcPoolFilter',
            'novalink_mobility_filter.PowerVMNovaLinkMobilityFilter']
        prefix = 'powervc_nova.scheduler.filters.common.'
        filters = [(prefix + filt) for filt in common_filters]
        # Depending on if it is PowerVM or KVM, need to add different filters
        if utils.get_offering() == 'powervm':
            prefix = 'powervc_nova.scheduler.filters.powervm.'
            filters.extend([(prefix + filt) for filt in powervm_filters])
        else:
            prefix = 'powervc_nova.scheduler.filters.kvm.'
            filters.extend([(prefix + filt) for filt in powerkvm_filters])
        return filters

    def __upgrade_aggregate_metadata_to_v123(self, tardir, tarfile,
                                             policy_data):
        """
        Helper method to upgrade the Aggregate MetaData to the DB.  This is
        needed when upgrading from Icehouse to Kilo (i.e., PowerVC 1.2.3).
        """
        initial_policy_id = ''
        state = ''
        for line in policy_data.split():
            line = line.strip()
            if line.startswith('<id>'):
                initial_policy_id = re.compile('<id>(.*?)</id>', re.DOTALL |
                                               re.IGNORECASE).findall(line)
            elif line.startswith('<state>'):
                state = re.compile('<state>(.*?)</state>', re.DOTALL |
                                   re.IGNORECASE).findall(line)
                if str(state) == "['enabled']":
                    break
        initial_policy_id = str(initial_policy_id).strip("['']")
        mgr.LOG.info(_("Restoring initial policy id: %(initial_policy_id)s; "
                       "state: %(state)s") % {'initial_policy_id':
                                              initial_policy_id,
                                              'state': state})
        engine = utils.get_database_engine(NOVA_CONF, 'connection', 'database')
        metadata = MetaData(bind=engine, schema='nova')
        aggr_table = Table('aggregates', metadata, autoload=True)
        meta_table = Table('aggregate_metadata', metadata, autoload=True)
        # Insert the Default Aggregate into the DB
        aggr_table.insert().values(
            id=1, name='Default Group', created_at=datetime.utcnow(),
            deleted=0).execute()
        # Insert each of the Aggregate MetaData Items into the DB
        self.__insert_aggregate_metadata(meta_table, 1,
                                         'hapolicy-id', '1')
        self.__insert_aggregate_metadata(meta_table, 1,
                                         'hapolicy-run_interval', '1')
        self.__insert_aggregate_metadata(meta_table, 1,
                                         'hapolicy-stabilization', '1')

        if initial_policy_id:
            self.__insert_aggregate_metadata(
                meta_table, 1, 'initialpolicy-id', str(initial_policy_id))

    def __insert_aggregate_metadata(self, table, aggr_id, key, value):
        """Helper method to insert an individual Aggregate MetaData Item"""
        table.insert().values(
            aggregate_id=aggr_id, key=key, value=value,
            created_at=datetime.utcnow(), deleted=0).execute()

    def __set_ego_admin_password(self, password):
        """Helper method to start the EGO service set the EGO admin user's
        password, then stop the EGO service."""

        ego_path = self.__find_ego_lib_path()
        os.environ['EGO_LIBDIR'] = ego_path

        ego_encrypted_password = self.__ego_encrypt_password(password)

        users_file_contents = (
            '<?xml version="1.0" encoding="UTF-8"?>\n'
            '<UserDatabase xmlns="http://www.platform.com/ego/2005/05/schema" '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            'xsi:schemaLocation="http://www.platform.com/ego/2005/05/schema '
            '../ego.xsd">\n  <UserList>\n    <UserEntry>\n      <User>Admin'
            '</User>\n      <Password>%s</Password>\n    </UserEntry>\n  '
            '</UserList>\n</UserDatabase>')

        with open(EGO_USERS_FILE, 'w') as f:
            f.write(users_file_contents % ego_encrypted_password)

    def __ego_encrypt_password(self, str_to_encrypt):
        """
        Encrypt a string in EGO user password encryption way. It is required to
        install EGO and run 'source /opt/ibm/ego/profile.platform' before
        calling this method.
        """
        ego_lib_path = os.environ.get('EGO_LIBDIR')
        if not ego_lib_path:
            raise Exception(_('Could not find EGO_LIBDIR in the environment; '
                              'please ensure EGO is installed and then retry '
                              'the request.'))
        sys.path.insert(0, ego_lib_path)
        ego = __import__('ego')
        ego_handler = ego.ego()
        return ego_handler.encryptPassword(str_to_encrypt)
