# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
"""
This is the Database API interface for PowerVC extensions to the
OpenStack Nova Database Schema.  These extensions include new database
tables to represent HMC's, VIOS's, SCG's, Network Adapters, FC Ports,
Health Status, On-boarding Task Progress, etc.
"""

import copy
from oslo_config import cfg
from oslo_db import concurrency
from nova.db import api as db_api

##################################################
#          Database API Initialization           #
##################################################
_BACKEND_MAPPING = {'sqlalchemy': 'powervc_nova.db.sqlalchemy.api'}
IMPL = concurrency.TpoolDbapiWrapper(cfg.CONF, _BACKEND_MAPPING)

# Maximum for the unsigned 8 byte data type MySQL uses
MAX_QUERY_RESULTS = (2 ** (8 * 8)) - 1


def initialize():
    """Forces the initialization for underlying database back-end module"""
    return IMPL.initialize()


def get_dbtransaction():
    """Creates a Database Session/Transaction to be used for commitment"""
    return IMPL.get_dbtransaction()


##################################################
#             HMC DB API Definition              #
##################################################
def hmc_find_all(context, filters=None, dbtxn=None):
    """Retrieves all of the matching HMC's that are in the Database"""
    return IMPL.hmc_find_all(context, filters, dbtxn)


def hmc_create(context, values, dbtxn=None):
    """Creates a new HMC instance in the Database"""
    return IMPL.hmc_create(context, values, dbtxn)


def hmc_update(context, hmc_uuid, values, dbtxn=None):
    """Updates an existing HMC instance in the Database"""
    return IMPL.hmc_update(context, hmc_uuid, values, dbtxn)


def hmc_delete(context, hmc_uuid, dbtxn=None):
    """Removes an existing HMC instance from the Database"""
    return IMPL.hmc_delete(context, hmc_uuid, dbtxn)


def hmc_host_delete(context, host_name, dbtxn=None):
    """Removes an existing HMC Host Mapping from the Database"""
    return IMPL.hmc_host_delete(context, host_name, dbtxn)


##################################################
#            Host DB API Definition              #
##################################################
def host_find_all(context, filters=None, dbtxn=None):
    """Retrieves all of the matching Hosts that are in the Database"""
    return IMPL.host_find_all(context, filters, dbtxn)


def host_delete(context, host_name, dbtxn=None):
    """Removes an existing Host instance from the Database"""
    return IMPL.host_delete(context, host_name, dbtxn)


def host_cluster_find_all(context, filters=None, dbtxn=None):
    """Retrieves all of the matching cluster hosts that are in the Database"""
    return IMPL.host_cluster_find_all(context, filters, dbtxn)


def host_cluster_create(context, values, dbtxn=None):
    """Creates a new cluster host instance in the Database"""
    return IMPL.host_cluster_create(context, values, dbtxn)


def host_cluster_delete(context, cluster_id, host_name, dbtxn=None):
    """Removes an existing cluster host instance from the Database"""
    return IMPL.host_cluster_delete(context, cluster_id, host_name, dbtxn)


##################################################
#           Instance DB API Definition           #
##################################################
def instance_find_all(context, filters=None, dbtxn=None):
    """Retrieves all of the matching VM's that are in the Database"""
    filters = filters if filters is not None else dict()
    # Since OpenStack seems to include Deleted on this call, explicitly exclude
    if filters.get('deleted') is None:
        filters['deleted'] = False
    columns = ['system_metadata', 'info_cache', 'extra', 'extra.flavor']
    return db_api.instance_get_all_by_filters(
        context, filters, columns_to_join=columns)


def instance_create(context, values, dbtxn=None):
    """Creates a new VM instance in the Database"""
    power_specs = values.pop('power_specs', None)
    # In case this is being managed again, do a hard delete
    if values.get('uuid') is not None:
        IMPL.instance_hard_delete(context, values['uuid'], dbtxn)
    # Now we can call the OpenStack method to create the Instance
    inst_ref = db_api.instance_create(context, values)
    # If there were PowerSpecs provided, then insert them now
    if power_specs is not None:
        instance_power_specs_create(
            context, inst_ref['uuid'], power_specs, dbtxn)
        # Query the Instance again to make sure the PowerSpecs is populated
        inst_ref = db_api.instance_get_by_uuid(context, inst_ref['uuid'])
    return inst_ref


def instance_update(context, instance_uuid, values, dbtxn=None):
    """Updates an existing VM instance in the Database"""
    values = copy.deepcopy(values)
    power_specs = values.pop('power_specs', None)
    # Merge in the existing MetaData if they asked for Partial Updates
    _instance_merge_metadata(context, instance_uuid, values)
    values.pop('updates_only', None)
    # Delegate to the OpenStack method to actually update the Instance
    inst_ref = db_api.instance_update(context, instance_uuid, values)
    # If there were PowerSpecs provided, then insert them now
    if power_specs is not None and inst_ref is not None:
        instance_power_specs_update(
            context, instance_uuid, power_specs, dbtxn)
    # Query the Instance again to make sure the PowerSpecs is populated
    columns = ['system_metadata', 'info_cache', 'extra', 'extra.flavor']
    return db_api.instance_get_by_uuid(
        context, inst_ref['uuid'], columns_to_join=columns)


def instance_delete(context, instance_uuid, dbtxn=None):
    """Removes an existing VM instance from the Database"""
    instance_power_specs_delete(context, instance_uuid, dbtxn)
    return db_api.instance_destroy(context, instance_uuid)


def _instance_merge_metadata(context, instance_uuid, values):
    """Helper Method to Merge Partial Metadata into existing"""
    system_metadata = values.get('system_metadata', {})
    partial_metadata = system_metadata.pop('updates_only', False)
    # If they gave us only Partial System MetaData, we need to merge
    if partial_metadata:
        old_inst = db_api.instance_get_by_uuid(context, instance_uuid)
        # Convert the MetaData on the Instance from a List to a Dictionary
        old_metal = old_inst.get('system_metadata', [])
        old_metad = dict([(itm['key'], itm['value']) for itm in old_metal])
        # Add in the new MetaData over top of the all of the old MetaData
        old_metad.update(system_metadata)
        values['system_metadata'] = old_metad


##################################################
#           PowerSpec DB API Definition          #
##################################################
def instance_power_specs_find_all(context, filters=None, dbtxn=None):
    """Retrieves all of the matching PowerSpecs that are in the Database"""
    return IMPL.instance_power_specs_find_all(context, filters, dbtxn)


def instance_power_specs_create(context, instance_uuid, values, dbtxn=None):
    """Creates a new Server PowerSpecs in the Database"""
    return IMPL.instance_power_specs_update(
        context, instance_uuid, values, dbtxn)


def instance_power_specs_update(context, instance_uuid, values, dbtxn=None):
    """Updates an existing Server PowerSpecs in the Database"""
    return IMPL.instance_power_specs_update(
        context, instance_uuid, values, dbtxn)


def instance_power_specs_delete(context, instance_uuid, dbtxn=None):
    """Removes an existing Server PowerSpecs from the Database"""
    return IMPL.instance_power_specs_delete(context, instance_uuid, dbtxn)


##################################################
#             SPP DB API Definition              #
##################################################
def spp_find_all(context, filters=None, dbtxn=None):
    """Retrieves all of the matching SPP's that are in the Database"""
    return IMPL.spp_find_all(context, filters, dbtxn)


def spp_create(context, values, dbtxn=None):
    """Creates a new SPP instance in the Database"""
    return IMPL.spp_create(context, values, dbtxn)


def spp_update(context, pool_uuid, values, dbtxn=None):
    """Updates an existing SPP instance in the Database"""
    return IMPL.spp_update(context, pool_uuid, values, dbtxn)


def spp_delete(context, pool_uuid, dbtxn=None):
    """Removes an existing SPP instance from the Database"""
    return IMPL.spp_delete(context, pool_uuid, dbtxn)


##################################################
#            VIOS DB API Definition              #
##################################################
def vios_find_all(context, filters=None, dbtxn=None):
    """Retrieves all of the matching VIOS's that are in the Database"""
    return IMPL.vios_find_all(context, filters, dbtxn)


def vios_create(context, values, dbtxn=None):
    """Creates a new VIOS instance in the Database"""
    return IMPL.vios_create(context, values, dbtxn)


def vios_update(context, vios_id, values, dbtxn=None):
    """Updates an existing VIOS instance in the Database"""
    return IMPL.vios_update(context, vios_id, values, dbtxn)


def vios_delete(context, vios_id, dbtxn=None):
    """Removes an existing VIOS instance from the Database"""
    return IMPL.vios_delete(context, vios_id, dbtxn)


##################################################
#             SEA DB API Definition              #
##################################################
def sea_find_all(context, filters=None, dbtxn=None):
    """Retrieves all of the matching SEA's that are in the Database"""
    return IMPL.sea_find_all(context, filters, dbtxn)


def sea_create(context, values, dbtxn=None):
    """Creates a new SEA instance in the Database"""
    return IMPL.sea_create(context, values, dbtxn)


def sea_update(context, sea_id, values, dbtxn=None):
    """Updates an existing SEA instance in the Database"""
    return IMPL.sea_update(context, sea_id, values, dbtxn)


def sea_delete(context, sea_id, dbtxn=None):
    """Removes an existing SEA instance from the Database"""
    return IMPL.sea_delete(context, sea_id, dbtxn)


##################################################
#             VEA DB API Definition              #
##################################################
def vea_find_all(context, filters=None, dbtxn=None):
    """Retrieves all of the matching VEA's that are in the Database"""
    return IMPL.vea_find_all(context, filters, dbtxn)


def vea_create(context, values, dbtxn=None):
    """Creates a new VEA instance in the Database"""
    return IMPL.vea_create(context, values, dbtxn)


def vea_update(context, vea_id, values, dbtxn=None):
    """Updates an existing VEA instance in the Database"""
    return IMPL.vea_update(context, vea_id, values, dbtxn)


def vea_delete(context, vea_id, dbtxn=None):
    """Removes an existing VEA instance from the Database"""
    return IMPL.vea_delete(context, vea_id, dbtxn)


##################################################
#     Network Association DB API Definition      #
##################################################
def network_assoc_find_all(context, filters=None, dbtxn=None):
    """Retrieves all of the matching Network Associations that are in the DB"""
    return IMPL.network_assoc_find_all(context, filters, dbtxn)


def network_assoc_create(context, values, dbtxn=None):
    """Creates a new Network Association instance in the Database"""
    return IMPL.network_assoc_create(context, values, dbtxn)


def network_assoc_update(context, netwassoc_id, values, dbtxn=None):
    """Updates an existing Network Association instance in the Database"""
    return IMPL.network_assoc_update(context, netwassoc_id, values, dbtxn)


def network_assoc_delete(context, netwassoc_id, dbtxn=None):
    """Removes an existing Network Association instance from the Database"""
    return IMPL.network_assoc_delete(context, netwassoc_id, dbtxn)


##################################################
#             SCG DB API Definition              #
##################################################
def scg_find_all(context, filters=None, dbtxn=None):
    """Retrieves all of the matching SCG's that are in the Database"""
    return IMPL.scg_find_all(context, filters, dbtxn)


def scg_create(context, values, dbtxn=None):
    """Creates a new SCG instance in the Database"""
    return IMPL.scg_create(context, values, dbtxn)


def scg_update(context, scg_id, values, dbtxn=None):
    """Updates an existing SCG instance in the Database"""
    return IMPL.scg_update(context, scg_id, values, dbtxn)


def scg_delete(context, scg_id, dbtxn=None):
    """Removes an existing SCG instance from the Database"""
    return IMPL.scg_delete(context, scg_id, dbtxn)


##################################################
#           FCPort DB API Definition             #
##################################################
def fcport_find_all(context, filters=None, dbtxn=None):
    """Retrieves all of the matching FCPorts that are in the Database"""
    return IMPL.fcport_find_all(context, filters, dbtxn)


def fcport_create(context, values, dbtxn=None):
    """Creates a new FCPort instance in the Database"""
    return IMPL.fcport_create(context, values, dbtxn)


def fcport_update(context, port_id, values, dbtxn=None):
    """Updates an existing FCPort instance in the Database"""
    return IMPL.fcport_update(context, port_id, values, dbtxn)


def fcport_delete(context, port_id, dbtxn=None):
    """Removes an existing FCPort instance from the Database"""
    return IMPL.fcport_delete(context, port_id, dbtxn)


##################################################
#        On-board Task DB API Definition         #
##################################################
def onboard_task_find_all(context, host, dbtxn=None):
    """Retrieves all of the On-board Tasks from the DB."""
    return IMPL.onboard_task_find_all(context, host, dbtxn)


def onboard_task_find(context, task_id, dbtxn=None):
    """Retrieves one of the On-board Tasks from the DB."""
    return IMPL.onboard_task_find(context, task_id, dbtxn)


def onboard_task_create(context, host, dbtxn=None):
    """Creates the given On-board Task in the DB."""
    return IMPL.onboard_task_create(context, host, dbtxn)


def onboard_task_update(context, task_id, values, dbtxn=None):
    """Updates the given On-board Task in the DB."""
    return IMPL.onboard_task_update(context, task_id, values, dbtxn)


def onboard_task_delete(context, task_id, dbtxn=None):
    """Deletes the given On-board Task from the DB."""
    return IMPL.onboard_task_delete(context, task_id, dbtxn)


def onboard_task_cleanup(context, dbtxn=None):
    """Cleanup any stuck On-board Task in the DB."""
    return IMPL.onboard_task_cleanup(context, dbtxn)


def onboard_task_server_create(context, task_id, svr_uuid, values, dbtxn=None):
    """Create the Server record for the given On-board Task"""
    return IMPL.onboard_task_server_create(
        context, task_id, svr_uuid, values, dbtxn)


def onboard_task_server_update(context, task_id, svr_uuid, values, dbtxn=None):
    """Updates the Server record for the given On-board Task"""
    return IMPL.onboard_task_server_update(
        context, task_id, svr_uuid, values, dbtxn)


##################################################
#       Console Messages DB API Definition       #
##################################################
def console_messages_find_all(context, filters=None, maximum=None):
    """Retrieves all of the matching Messages that are in the Database"""
    return IMPL.console_messages_find_all(
        context, filters=filters, maximum=maximum)


def console_messages_find(context, filters=None, maximum=None):
    """Retrieves zero or one matching Messages that are in the Database"""
    messages = console_messages_find_all(
        context, filters=filters, maximum=maximum)
    return messages[0] if messages else None


def console_messages_create(context, values):
    """Creates a new Console Message instance in the Database"""
    return IMPL.console_messages_create(context, values)


def console_messages_update(context, msg_id, values):
    """Updates an existing Console Message instance in the Database"""
    return IMPL.console_messages_update(context, msg_id, values)


def console_messages_delete(context, msg_id):
    """Removes an existing Console Message instance from the Database"""
    return IMPL.console_messages_delete(context, msg_id)


def console_messages_prune(context, maximum):
    """Prunes older Console Messages from the Database when maximum exceeded"""
    return IMPL.console_messages_prune(context, maximum)


#######################################################
#       Compute Node Metadata DB API Definition       #
#######################################################
def compute_node_metadata_get(context, compute_node_id, dbtxn=None):
    """Get all metadata for a compute node."""
    return IMPL.compute_node_metadata_get(context, compute_node_id,
                                          dbtxn=dbtxn)


def compute_node_metadata_delete(context, compute_node_id, key, dbtxn=None):
    """Delete the given metadata item."""
    IMPL.compute_node_metadata_delete(context, compute_node_id, key,
                                      dbtxn=dbtxn)


def compute_node_metadata_update(context, compute_node_id, metadata,
                                 set_delete=False, dbtxn=None):
    """Update metadata if it exists, otherwise create it."""
    return IMPL.compute_node_metadata_update(context, compute_node_id,
                                             metadata, set_delete=set_delete,
                                             dbtxn=dbtxn)


def block_device_mapping_get_all_by_volume_id(context, volume_id,
                                              columns_to_join=None):
    return IMPL.block_device_mapping_get_all_by_volume_id(context, volume_id,
                                                          columns_to_join=None)


###########################################################
# Enterprise Pools (Capacity on Demand) DB API Definition #
###########################################################
def enterprise_pool_find_all(context, filters=None, dbtxn=None):
    """
    Retrieves all of the matching enterprise pools that are in the Database
    """
    return IMPL.enterprise_pool_find_all(context, filters, dbtxn)


def enterprise_pool_create(context, values, dbtxn=None):
    """Creates a new enterprise pool instance in the Database"""
    return IMPL.enterprise_pool_create(context, values, dbtxn)


def enterprise_pool_update(context, pool_uuid, values, dbtxn=None):
    """Updates an existing enterprise pool instance in the Database"""
    return IMPL.enterprise_pool_update(context, pool_uuid, values, dbtxn)


def enterprise_pool_delete(context, pool_uuid, dbtxn=None):
    """Removes an existing enterprise pool instance from the Database"""
    return IMPL.enterprise_pool_delete(context, pool_uuid, dbtxn)


####################################################
# DRO Events Table DB API Definition               #
####################################################
def dro_events_create(context, values):
    """Creates new DRO events instances in the Database"""
    return IMPL.dro_events_create(context, values)


def dro_events_prune(context, max_event_time):
    """Prunes older DRO events from the Database"""
    return IMPL.dro_events_prune(context, max_event_time)


def dro_events_find_all(context, filters=None, maximum=None):
    """Retrieves zero or one matching DRO Events that are in the Database"""
    return IMPL.dro_events_find_all(context, filters, maximum)


####################################################
# DRO Schedules Table DB API Definition            #
####################################################
def dro_schedules_create_aggregate_schedules(context, values, dbtxn=None):
    """Creates DRO schedules for the aggregate"""
    return IMPL.dro_schedules_create_aggregate_schedules(
        context, values, dbtxn)


def dro_schedules_read_all_schedules(context, dbtxn=None):
    """Retrieves the DRO schedules for all aggregates"""
    return IMPL.dro_schedules_read_all_schedules(context, dbtxn)


def dro_schedules_read_aggregate_schedules(context, aggregate_id, dbtxn=None):
    """Retrieves the DRO schedules for the specified aggregate_id"""
    return IMPL.dro_schedules_read_aggregate_schedules(
        context, aggregate_id, dbtxn)


def dro_schedules_update_aggregate_schedules(context, aggregate_id, values,
                                             dbtxn=None):
    """Replaces the aggregate's current DRO schedules or creates a new record
       if one does not currently exist"""
    return IMPL.dro_schedules_update_aggregate_schedules(
        context, aggregate_id, values, dbtxn)


def dro_schedules_delete_aggregate_schedules(context, aggregate_id,
                                             dbtxn=None):
    """Soft deletes the existing DRO schedules for the specified aggregate"""
    return IMPL.dro_schedules_delete_aggregate_schedules(
        context, aggregate_id, dbtxn)
