# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
"""
This is the Database SQLAlchemy API implementation for PowerVC extensions
to the OpenStack Nova Database Schema.  These extensions include new database
tables to represent HMC's, VIOS's, SCG's, Network Adapters, FC Ports,
Health Status, On-boarding Task Progress, etc.
"""

import sys
import uuid
import os.path
import datetime
from sqlalchemy import or_
from sqlalchemy import and_
from sqlalchemy.orm import joinedload

import nova.db.sqlalchemy.api as base_dbapi
from nova.db.sqlalchemy import models as base_dbmodels
from oslo_db.sqlalchemy import utils as sqlalchemyutils
from oslo_serialization import jsonutils
from oslo_utils import timeutils
from oslo_log import log as logging

from powervc_nova.db.sqlalchemy import models as dbmodels
from powervc_nova import _

LOG = logging.getLogger(__name__)


##################################################
#          Database API Initialization           #
##################################################
def get_backend():
    """The back-end in this particular case is this module itself."""
    return sys.modules[__name__]


def get_dbtransaction():
    """Creates a Database Session/Transaction to be used for commitment"""
    return base_dbapi.main_context_manager.get_legacy_facade().get_session()


def initialize():
    """Used to Forces the initialization of this database back-end module"""
    pass


##################################################
#           HMC DB API Implementation            #
##################################################
def hmc_find_all(context, filters=None, dbtxn=None):
    """Retrieves all of the matching HMC's that are in the Database"""
    # Translated the Filters to account for the Hosts
    filters = _hmc_hosts_filters(context, filters, dbtxn)
    # Query the list of matching entries from the given Database table
    result_set = _find_all_db_entries(
        context, dbmodels.HmcDTO, filters, dbtxn)
    # Translate the Result Values to account for the Hosts
    return _hmc_hosts_result_set(context, result_set, dbtxn)


def hmc_create(context, values, dbtxn=None):
    """Creates a new HMC instance in the Database"""
    hosts = values.pop('managed_hosts', None)
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # Translate the Input Values to account for the HMC Hosts
    values = _hmc_hosts_values(context, values, False, session)
    # Create a Transaction around the create in the Database
    with session.begin(subtransactions=True):
        # Insert the new entry in the specified Database table
        result_set = _create_db_entry(
            context, dbmodels.HmcDTO, values, session)
        # Update the hmc_hosts table with the new list of Hosts
        _hmc_update_hosts(context, values['hmc_uuid'], hosts, session)
    # Translate the Result Values to account for the Hosts
    return _hmc_hosts_result_set(context, result_set, session)


def hmc_update(context, hmc_uuid, values, dbtxn=None):
    """Updates an existing HMC instance in the Database"""
    filters = dict(hmc_uuid=hmc_uuid)
    hosts = values.pop('managed_hosts', None)
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # Translate the Input Values to account for the HMC Hosts
    values = _hmc_hosts_values(context, values, True, session)
    # Create a Transaction around the create in the Database
    with session.begin(subtransactions=True):
        # Update the existing entry in the specified Database table
        result_set = _update_db_entry(
            context, dbmodels.HmcDTO, filters, values, session)
        # Update the hmc_hosts table with the new list of Hosts
        _hmc_update_hosts(context, hmc_uuid, hosts, session)
    # Translate the Result Values to account for the Hosts
    return _hmc_hosts_result_set(context, result_set, session)


def hmc_delete(context, hmc_uuid, dbtxn=None):
    """Removes an existing HMC instance from the Database"""
    filters = [dict(hmc_uuid=hmc_uuid), dict(hmc_uuid=hmc_uuid)]
    model_classes = [dbmodels.HmcHostsDTO, dbmodels.HmcDTO]
    # Delete the existing entries from the specified Database tables
    _delete_db_entries(context, model_classes, filters, dbtxn, True)


def hmc_host_delete(context, host_name, dbtxn=None):
    """Removes an existing HMC Host Mapping from the Database"""
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # Create a Transaction around the delete in the Database
    with session.begin(subtransactions=True):
        query = _model_query(
            context, dbmodels.HmcHostsDTO, session=session)
        query = query.filter_by(host_name=host_name)
        query.soft_delete(synchronize_session=False)


def _hmc_hosts_filters(context, filters, dbtxn=None):
    """Helper method to convert the filters to the DB schema columns"""
    filters = filters.copy() if filters is not None else None
    # If they asked us to filter on the host name, we need something special
    if filters and 'host_name' in filters:
        host_name = filters.pop('host_name', None)
        # If they also filtered on the HMC UUID, that takes precedence
        if 'hmc_uuid' not in filters:
            hmc_uuids = _hmc_get_hmcs(context, host_name, dbtxn)
            # Add a special filter to handle a bunch of HMC UUID's
            filters['special'] = filters.get('special', [])
            filters['special'].append(dbmodels.HmcDTO.hmc_uuid.in_(hmc_uuids))
    return filters


def _hmc_hosts_values(context, values, is_update=False, dbtxn=None):
    """Helper method to convert the values to the DB schema columns"""
    values = values.copy()
    # Make sure to Encrypt the Password before inserting the database
    if values.get('password') is not None:
        from powervc_keystone.encrypthandler import EncryptHandler
        values['password'] = EncryptHandler().encode(values['password'])
    return values


def _hmc_hosts_result_set(context, result_set, dbtxn=None):
    """Helper method to convert the results from the DB schema columns"""
    # If this is a list, then convert each item in that Result Set list
    if isinstance(result_set, list):
        return [_hmc_hosts_result_set(
                context, itm, dbtxn) for itm in result_set]
    # Make sure to DeCrypt the Password after retrieving from the database
    password = result_set.get('password')
    if password is not None:
        from powervc_keystone.encrypthandler import EncryptHandler
        result_set['password'] = EncryptHandler().decode(password)
    # Add the Hosts that are managed by the HMC to the HMC returned
    result_set['managed_hosts'] = \
        _hmc_get_hosts(context, result_set['hmc_uuid'], dbtxn)
    return result_set


def _hmc_get_hosts(context, hmc_uuid, dbtxn=None):
    """Helper method to list the Hosts that are managed by an HMC"""
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # We need to query the Hosts for the HMC based on the HMC UUID
    query = session.query(dbmodels.HmcHostsDTO.host_name)
    host_names = query.filter_by(hmc_uuid=hmc_uuid, deleted=0).all()
    return [host_name[0] for host_name in host_names]


def _hmc_get_hmcs(context, host_name, dbtxn=None):
    """Helper method to list the Hosts that are managed by an HMC"""
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # We need to query the Hosts for the HMC based on the HMC UUID
    query = session.query(dbmodels.HmcHostsDTO.hmc_uuid)
    hmc_uuids = query.filter_by(host_name=host_name, deleted=0).all()
    return [hmc_uuid[0] for hmc_uuid in hmc_uuids]


def _hmc_get_host_map(context, dbtxn=None):
    """Helper method to list the Hosts that are managed by an HMC"""
    host_map = dict()
    # We need to query the all of the HMC Hosts that aren't deleted
    result_set = _find_all_db_entries(
        context, dbmodels.HmcHostsDTO, None, dbtxn)
    # Loop through each of the HMC Host entries, building a map based on host
    for hmchost in result_set:
        host_map[hmchost['host_name']] = host_map.get(hmchost['host_name'], [])
        host_map[hmchost['host_name']].append(hmchost['hmc_uuid'])
    return host_map


def _hmc_update_hosts(context, hmc_uuid, hosts, dbtxn=None):
    """Helper method to update the Hosts that are managed by an HMC"""
    # If we weren't given any Hosts, then there is nothing to do
    if hosts is not None:
        # Query the list of Host that are associated with this HMC
        db_hosts = _hmc_get_hosts(context, hmc_uuid, dbtxn)
        # For each Host provided that doesn't exist, add it to the database
        for host in hosts:
            if host not in db_hosts:
                values = dict(hmc_uuid=hmc_uuid, host_name=host)
                _create_db_entry(
                    context, dbmodels.HmcHostsDTO, values, dbtxn)
        # For each Host in the DB but not provided, remove from database
        for db_host in db_hosts:
            if db_host not in hosts:
                filters = dict(hmc_uuid=hmc_uuid, host_name=db_host)
                _delete_db_entry(
                    context, dbmodels.HmcHostsDTO, filters, dbtxn, True)


##################################################
#           Host DB API Implementation           #
##################################################
def host_find_all(context, filters=None, dbtxn=None):
    """Retrieves all of the matching Hosts that are in the Database"""
    hosts = list()
    hmchosts = _hmc_get_host_map(context, dbtxn)
    # Translated the Filters to account for the Hosts
    filters = _host_filters(context, filters, dbtxn)
    # Query the list of matching entries from the given Database table
    result_set = _find_all_db_entries(
        context, base_dbmodels.Service, filters, dbtxn)
    # Query all of the Compute Nodes that match the Hosts filtered on
    filters = dict(host=filters['host']) if 'host' in filters else None
    compute_nodes = _find_all_db_entries(
        context, base_dbmodels.ComputeNode, filters, dbtxn)
    compute_nodes = dict([(r['host'], r) for r in compute_nodes])
    # Loop through each of the Services, getting the Compute Node
    for service in result_set:
        host = dict(service_id=service['id'], host_name=service['host'])
        host['hmc_uuids'] = hmchosts.get(host['host_name'], [])
        host['registered_at'] = service['created_at']
        compute_node = compute_nodes.get(service['host'])
        # If the Compute Node existed, then we will return it
        if compute_node is not None:
            host.update(compute_node)
        hosts.append(host)
    # Translate the Result Values to account for the Hosts
    return hosts


def host_delete(context, host_name, dbtxn=None):
    """Removes an existing Host instance from the Database"""
    compute_node = None
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # Query the Service and the Compute Node from the Database
    query1 = _model_query(context, base_dbmodels.Service, session=session)
    query2 = _model_query(context, base_dbmodels.ComputeNode, session=session)
    query1 = query1.filter_by(host=host_name, topic='compute')
    service = query1.first()
    if service is not None:
        query2 = query2.filter_by(host=host_name)
        compute_node = query2.first()
    # Query the VIOS's and Network Associations for the Host from the DB
    vioses = vios_find_all(context, dict(host_name=host_name))
    netassocs = network_assoc_find_all(context, dict(host_name=host_name))
    # Create a Transaction around the delete from the Database
    with session.begin(subtransactions=True):
        for netassoc in netassocs:
            network_assoc_delete(context, netassoc['id'], session)
        for vios in vioses:
            vios_delete(context, vios['id'], session)
        # Also need to clean up the entry in the HMC Hosts DB Table
        hmc_host_delete(context, host_name, session)
        # Clean up the Shared Processor Pools associated with the host
        _delete_db_entry(context, dbmodels.SharedProcessorPoolDTO,
                         dict(host_name=host_name), session, False)
        # Clean up the Enterprise Pool entries associated with the host
        enterprise_pools = enterprise_pool_find_all(
            context, dict(host_name=host_name), session)
        _delete_db_entry(context, dbmodels.EnterprisePoolMembersDTO,
                         dict(host_name=host_name), session, False)
        for pool in enterprise_pools:
            if len(pool['host_names']) <= 1:
                enterprise_pool_delete(context, pool['uuid'], session)
        # Need to query the Service based on the Host to know what to delete
        # If the Service exists, Delete the Service/ComputeNode from the DB
        if service is not None:
            # If the Compute Node exists, then delete it from the Database
            if compute_node is not None:
                # Clean up the entries from the Compute Node MetaData table
                _delete_db_entry(context, dbmodels.IBMComputeNodeMetadataDTO,
                                 dict(compute_node_id=compute_node.id),
                                 session, False)
                query2.delete(synchronize_session=False)
            # Now that the Compute Node is gone, we can delete the Service
            query1.soft_delete(synchronize_session=False)


def host_cluster_find_all(context, filters=None, dbtxn=None):
    """Retrieves all of the matching cluster hosts that are in the Database"""
    # Query the list of matching entries from the given Database table
    return _find_all_db_entries(
        context, dbmodels.HmcHostClustersDTO, filters, dbtxn)


def host_cluster_create(context, values, dbtxn=None):
    """Creates a new cluster host instance in the Database"""
    # Insert the new entry in the specified Database table
    return _create_db_entry(
        context, dbmodels.HmcHostClustersDTO, values, dbtxn)


def host_cluster_delete(context, cluster_id, host_name, dbtxn=None):
    """Removes an existing cluster host instance from the Database"""
    filters = dict(cluster_id=cluster_id, host_name=host_name)
    # Delete the existing entry in the specified Database table
    _delete_db_entry(
        context, dbmodels.HmcHostClustersDTO, filters, dbtxn, True)


def _host_filters(context, filters, dbtxn=None):
    """Helper method to convert the filters to the DB schema columns"""
    filters = filters.copy() if filters is not None else dict()
    # If they filtered on the Host Name, then we need to get the Compute Node
    if 'host_name' in filters:
        filters['host'] = filters.pop('host_name', None)
    # If they asked us to filter on the HMC UUID, we need something special
    if filters and 'hmc_uuid' in filters:
        hmc_uuid = filters.pop('hmc_uuid', None)
        # If they also filtered on the Host Name, that takes precedence
        if 'host_name' not in filters:
            hosts = _hmc_get_hosts(context, hmc_uuid, dbtxn)
            # Add a special filter to handle a bunch of Host Names
            filters['special'] = filters.get('special', [])
            filters['special'].append(base_dbmodels.Service.host.in_(hosts))
    filters['topic'] = 'compute'
    return filters


def _host_result_set(context, result_set, dbtxn=None):
    # If this is a list, then convert each item in that Result Set list
    if isinstance(result_set, list):
        return [_host_result_set(
                context, itm, dbtxn) for itm in result_set]
    # Convert the Compute Node ID to a Host Name to be returned
    result_set['host_name'] = _host_get_host_name(
        context, result_set.pop('_compute_node_id', None), dbtxn)
    return result_set


def _host_component_filters(context, filters, dbtxn=None):
    """Helper method to convert the filters to the DB schema columns"""
    filters = filters.copy() if filters is not None else None
    # If they filtered on the Host Name, then we need to get the Compute Node
    if filters and 'host_name' in filters:
        host_name = filters.pop('host_name')
        if host_name is not None:
            filters['_compute_node_id'] = \
                _host_get_compute_node_id(context, host_name, dbtxn)
    return filters


def _host_component_values(context, values, is_update=False, dbtxn=None):
    """Helper method to convert the values to the DB schema columns"""
    values = values.copy()
    # If the Host Name is provided, need to convert to a Compute Node ID
    if 'host_name' in values:
        values['_compute_node_id'] = _host_get_compute_node_id(
            context, values.pop('host_name', None), dbtxn)
    return values


def _host_component_result_set(context, result_set, dbtxn=None):
    # If this is a list, then convert each item in that Result Set list
    if isinstance(result_set, list):
        if len(result_set) >= 5 and 'host_name' not in result_set[0]:
            _host_populate_compute_node_id(context, result_set, dbtxn)
        return [_host_component_result_set(
                context, itm, dbtxn) for itm in result_set]
    # Convert the Compute Node ID to a Host Name to be returned
    if 'host_name' not in result_set:
        result_set['host_name'] = _host_get_host_name(
            context, result_set.pop('_compute_node_id', None), dbtxn)
    return result_set


def _host_get_compute_node_id(context, host_name, dbtxn=None):
    """Helper method to retrieve the Compute Node ID for a Host Name"""
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # We need to query the Service based on the Host to know the Service ID
    query = session.query(base_dbmodels.ComputeNode.id)
    compute_node_id = query.filter_by(host=host_name, deleted=0).first()
    return compute_node_id[0] if compute_node_id is not None else None


def _host_get_host_name(context, compute_node_id, dbtxn=None):
    """Helper method to retrieve the Compute Node ID for a Host Name"""
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # We need to query the Compute Node based on the ID to know it as
    query = session.query(base_dbmodels.ComputeNode.host)
    host_name = query.filter_by(id=compute_node_id, deleted=0).first()
    return host_name[0] if host_name is not None else None


def _host_populate_compute_node_id(context, result_set, dbtxn=None):
    """Helper method to populate the Host Name from the Primary Key ID"""
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # We need to query all the Hosts to get the mapping
    query = session.query(
        base_dbmodels.ComputeNode.id, base_dbmodels.ComputeNode.host)
    host_map = dict([(r[0], r[1]) for r in query.all()])
    # Loop through the Result Set, populating the Host Name
    for row in result_set:
        row['host_name'] = host_map.get(row.pop('_compute_node_id'))


##################################################
#         Instance DB API Implementation         #
##################################################
def instance_hard_delete(context, instance_uuid, dbtxn=None):
    """Performs a Hard Delete of a previously existing Instance"""
    session = get_dbtransaction() if dbtxn is None else dbtxn
    query = _model_query(context, base_dbmodels.Instance, read_deleted='yes')
    query = query.filter(base_dbmodels.Instance.deleted > 0)
    # If there wasn't an existing deleted instance, then nothing to do
    if len(query.filter_by(uuid=instance_uuid).all()) <= 0:
        return
    # Define the tables that we wish to to a hard delete from
    tables = [
        base_dbmodels.InstanceFault, base_dbmodels.Migration,
        dbmodels.InstanceHealthStatusDTO, base_dbmodels.InstanceIdMapping,
        base_dbmodels.InstanceInfoCache, dbmodels.InstancePowerSpecsDTO,
        base_dbmodels.InstanceSystemMetadata, base_dbmodels.BlockDeviceMapping,
        base_dbmodels.InstanceMetadata, base_dbmodels.InstanceExtra,
        base_dbmodels.SecurityGroupInstanceAssociation, base_dbmodels.Instance]
    # Create a Transaction around the delete in the Database
    with session.begin(subtransactions=True):
        # Cleanup the Instance Actions/Events tables first, since are specific
        __delete_instance_actions(context, instance_uuid, session)
        # Loop through the tables to cleanup the entries based on the UUID
        for table in tables:
            query = _model_query(
                context, table, read_deleted='yes', session=session)
            # Depending on the table, we have to filter on different columns
            if table.__name__ == 'InstanceHealthStatusDTO':
                query = query.filter_by(id=instance_uuid)
            elif table.__name__ in ['Instance', 'InstanceIdMapping']:
                query = query.filter_by(uuid=instance_uuid)
            else:
                query = query.filter_by(instance_uuid=instance_uuid)
            # Now we can do the hard delete since we setup our filter
            query.delete(synchronize_session=False)


def __delete_instance_actions(context, instance_uuid, session):
    """Helper method to clean up the Instance Actions/Events tables"""
    query = _model_query(context, base_dbmodels.InstanceAction,
                         read_deleted='yes', session=session)
    query = query.filter_by(instance_uuid=instance_uuid)
    for row in query.all():
        query2 = _model_query(context, base_dbmodels.InstanceActionEvent,
                              read_deleted='yes', session=session)
        # Delete the Events for the given Action id that was found
        query2 = query2.filter_by(action_id=row['id'])
        query2.delete(synchronize_session=False)
    # Finally we can cleanup the Instance Actions entries themselves
    query.delete(synchronize_session=False)


##################################################
#         PowerSpec DB API Implementation        #
##################################################
def instance_power_specs_find_all(context, filters=None, dbtxn=None):
    """Retrieves all of the matching PowerSpecs that are in the Database"""
    # Query the list of matching entries from the given Database table
    return _find_all_db_entries(
        context, dbmodels.InstancePowerSpecsDTO, filters, dbtxn)


def instance_power_specs_update(context, instance_uuid, values, dbtxn=None):
    """Updates an existing Server PowerSpecs in the Database"""
    filters = dict(instance_uuid=instance_uuid)
    result_set = instance_power_specs_find_all(context, filters, dbtxn)
    # Correct the format of the values passed in so they won't blow up
    _instance_power_specs_correct_values(values)
    # If the PowerSpecs already exist, then we need to update the db entry
    if result_set:
        return _update_db_entry(
            context, dbmodels.InstancePowerSpecsDTO, filters, values, dbtxn)
    # If the PowerSpecs doesn't yet exist, we need to create the db entry
    else:
        return _create_db_entry(
            context, dbmodels.InstancePowerSpecsDTO, values, dbtxn)


def instance_power_specs_delete(context, instance_uuid, dbtxn=None):
    """Removes an existing Server PowerSpecs from the Database"""
    # Delete the existing entry in the specified Database table
    _delete_db_entry(context, dbmodels.InstancePowerSpecsDTO,
                     dict(instance_uuid=instance_uuid), dbtxn, True)


def _instance_power_specs_correct_values(values):
    """Helper method to correct values in the wrong format"""
    values = values if values else dict()
    # Loop through the values, correcting srr_capability and uncapped
    for key, val in values.iteritems():
        if key in ['srr_capability', 'uncapped']:
            if val is not None and isinstance(val, basestring):
                values[key] = val.lower() == 'true'


##################################################
#            SPP DB API Implementation           #
##################################################
def spp_find_all(context, filters=None, dbtxn=None):
    """Retrieves all of the matching SPP's that are in the Database"""
    # Query the list of matching entries from the given Database table
    result_set = _find_all_db_entries(
        context, dbmodels.SharedProcessorPoolDTO, filters, dbtxn)
    # Translate the Result Values into a Dictionary
    return _result_set_as_dict(result_set)


def spp_create(context, values, dbtxn=None):
    """Creates a new SPP instance in the Database"""
    # Insert the new entry in the specified Database table
    result_set = _create_db_entry(
        context, dbmodels.SharedProcessorPoolDTO, values, dbtxn)
    # Translate the Result Values into a Dictionary
    return _result_set_as_dict(result_set)


def spp_update(context, pool_uuid, values, dbtxn=None):
    """Updates an existing SPP instance in the Database"""
    filters = dict(pool_uuid=pool_uuid)
    # Update the existing entry in the specified Database table
    result_set = _update_db_entry(
        context, dbmodels.SharedProcessorPoolDTO, filters, values, dbtxn)
    # Translate the Result Values into a Dictionary
    return _result_set_as_dict(result_set)


def spp_delete(context, pool_uuid, dbtxn=None):
    """Removes an existing SPP instance from the Database"""
    filters = dict(pool_uuid=pool_uuid)
    # Delete the existing entry in the specified Database table
    _delete_db_entry(
        context, dbmodels.SharedProcessorPoolDTO, filters, dbtxn, False)


##################################################
#           VIOS DB API Implementation           #
##################################################
def vios_find_all(context, filters=None, dbtxn=None):
    """Retrieves all of the matching VIOS's that are in the Database"""
    # Translated the Filters to account for the Host Name
    filters = _host_component_filters(context, filters, dbtxn)
    filters = _vios_special_filters(context, filters, dbtxn)
    # Query the list of matching entries from the given Database table
    result_set = _find_all_db_entries(
        context, dbmodels.VioServerDTO, filters, dbtxn)
    # Translate the Result Values to account for the Host Name
    return _host_component_result_set(context, result_set, dbtxn)


def vios_create(context, values, dbtxn=None):
    """Creates a new VIOS instance in the Database"""
    # Translate the Input Values to account for the Host Name
    values = _host_component_values(context, values, False, dbtxn)
    # Insert the new entry in the specified Database table
    result_set = _create_db_entry(
        context, dbmodels.VioServerDTO, values, dbtxn)
    # Translate the Result Values to account for the Host Name
    return _host_component_result_set(context, result_set, dbtxn)


def vios_update(context, vios_id, values, dbtxn=None):
    """Updates an existing VIOS instance in the Database"""
    filters = dict(id=vios_id)
    # Translate the Input Values to account for the Host Name
    values = _host_component_values(context, values, True, dbtxn)
    # Update the existing entry in the specified Database table
    result_set = _update_db_entry(
        context, dbmodels.VioServerDTO, filters, values, dbtxn)
    # Translate the Result Values to account for the Host Name
    return _host_component_result_set(context, result_set, dbtxn)


def vios_delete(context, vios_id, dbtxn=None):
    """Removes an existing VIOS instance from the Database"""
    vios_pk_id = _vios_get_vios_pk_id(context, vios_id, dbtxn)
    filters = [dict(_vios_pk_id=vios_pk_id), dict(_vios_pk_id=vios_pk_id),
               dict(_vios_pk_id=vios_pk_id), dict(_vios_pk_id=vios_pk_id),
               dict(id=vios_id)]
    # We need to first null out all the SEA ID's on the VEA's
    veas = vea_find_all(context, filters[0], dbtxn)
    for vea in veas:
        vea_update(context, vea['id'], dict(sea_id=None), dbtxn)
    model_classes = [
        dbmodels.SharedEthernetAdapterDTO, dbmodels.VirtualEthernetAdapterDTO,
        dbmodels.ScgViosAssociationDTO, dbmodels.FcPortDTO,
        dbmodels.VioServerDTO]
    _delete_db_entries(context, model_classes, filters, dbtxn)


def _vios_component_filters(context, filters, dbtxn=None):
    """Helper method to convert the filters to the DB schema columns"""
    filters = filters.copy() if filters is not None else None
    # If they filtered on the VIOS ID, then we need to get the Primary Key ID
    if filters and 'vios_id' in filters:
        filters['_vios_pk_id'] = _vios_get_vios_pk_id(
            context, filters.pop('vios_id', None), dbtxn)
    # If they asked us to filter on the VIOS, we need something special
    if filters and 'host_name' in filters:
        host_name = filters.pop('host_name', None)
        # If they also filtered on the VIOS ID, that takes precedence
        if '_vios_pk_id' not in filters:
            compute_node_id = _host_get_compute_node_id(
                context, host_name, dbtxn)
            vios_pk_ids = _vios_get_vios_pk_ids(
                context, compute_node_id, dbtxn)
            # Add a special filter to handle a bunch of VIOS ID's
            filters['special'] = filters.get('special', [])
            filters['special'].append(
                dbmodels.VioServerDTO._pk_id.in_(vios_pk_ids))
    return filters


def _vios_component_values(context, values, is_update=False, dbtxn=None):
    """Helper method to convert the values to the DB schema columns"""
    values = values.copy()
    # If the VIOS ID is provided, need to convert to the Primary Key ID
    if 'vios_id' in values:
        values['_vios_pk_id'] = _vios_get_vios_pk_id(
            context, values.pop('vios_id', None), dbtxn)
    return values


def _vios_component_result_set(context, result_set, dbtxn=None):
    # If this is a list, then convert each item in that Result Set list
    if isinstance(result_set, list):
        # If there are a bunch of rows, query all VIOS's to map
        if len(result_set) >= 5 and 'vios_id' not in result_set[0]:
            _vios_populate_vios_id(context, result_set, dbtxn)
        return [_vios_component_result_set(
                context, itm, dbtxn) for itm in result_set]
    # Convert the Primary Key ID to a VIOS ID to be returned
    if 'vios_id' not in result_set:
        result_set['vios_id'] = _vios_get_vios_id(
            context, result_set.pop('_vios_pk_id', None), dbtxn)
    return result_set


def _vios_special_filters(context, filters, dbtxn=None):
    """Helper method to convert the filters to the DB schema columns"""
    vios_ids = filters.pop('vios_ids', None) if filters else None
    # If they gave us VIOS ID's, then we need to filter on those
    if vios_ids is not None:
        filters['special'] = filters.get('special', [])
        filters['special'].append(
            dbmodels.VioServerDTO.id.in_(vios_ids))
    return filters


def _vios_get_vios_pk_ids(context, compute_node_id, dbtxn=None):
    """Helper method to get the VIOS Primary Key ID's for a given Host"""
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # We need to query the VIOS's based on the Compute Node to know the ID
    query = session.query(dbmodels.VioServerDTO._pk_id)
    vios_pk_ids = query.filter_by(_compute_node_id=compute_node_id).all()
    return [vios_pk_id[0] for vios_pk_id in vios_pk_ids]


def _vios_get_vios_pk_id(context, vios_id, dbtxn=None):
    """Helper method to get the Primary Key ID from the VIOS ID"""
    return _get_primary_key_from_id(
        context, dbmodels.VioServerDTO, vios_id, dbtxn)


def _vios_get_vios_id(context, vios_pk_id, dbtxn=None):
    """Helper method to get the VIOS ID from the Primary Key ID"""
    return _get_id_from_primary_key(
        context, dbmodels.VioServerDTO, vios_pk_id, dbtxn)


def _vios_populate_vios_id(context, result_set, dbtxn=None):
    """Helper method to populate the VIOS ID from the Primary Key ID"""
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # We need to query all the VIOS's to get the mapping
    query = session.query(
        dbmodels.VioServerDTO._pk_id, dbmodels.VioServerDTO.id)
    vios_map = dict([(r[0], r[1]) for r in query.all()])
    # Loop through the Result Set, populating the VIOS ID
    for row in result_set:
        row['vios_id'] = vios_map.get(row.pop('_vios_pk_id'))


##################################################
#           SEA DB API Implementation            #
##################################################
def sea_find_all(context, filters=None, dbtxn=None):
    """Retrieves all of the matching SEA's that are in the Database"""
    # Translated the Filters to account for the VIOS ID and VEA's
    filters = _sea_vea_filters(context, filters, dbtxn)
    # Query the list of matching entries from the given Database table
    result_set = _find_all_db_entries(
        context, dbmodels.SharedEthernetAdapterDTO, filters, dbtxn)
    # Translate the Result Values to account for the VIOS ID and VEA's
    return _sea_vea_result_set(context, result_set, dbtxn)


def sea_create(context, values, dbtxn=None):
    """Creates a new SEA instance in the Database"""
    # Translate the Input Values to account for the VIOS ID and VEA's
    values = _sea_vea_values(context, values, False, dbtxn)
    # Insert the new entry in the specified Database table
    result_set = _create_db_entry(
        context, dbmodels.SharedEthernetAdapterDTO, values, dbtxn)
    # Translate the Result Values to account for the VIOS ID and VEA's
    return _sea_vea_result_set(context, result_set, dbtxn)


def sea_update(context, sea_id, values, dbtxn=None):
    """Updates an existing SEA instance in the Database"""
    filters = dict(id=sea_id)
    # Translate the Input Values to account for the VIOS ID and VEA's
    values = _sea_vea_values(context, values, True, dbtxn)
    # Update the existing entry in the specified Database table
    result_set = _update_db_entry(
        context, dbmodels.SharedEthernetAdapterDTO, filters, values, dbtxn)
    # Translate the Result Values to account for the VIOS ID and VEA's
    return _sea_vea_result_set(context, result_set, dbtxn)


def sea_delete(context, sea_id, dbtxn=None):
    """Removes an existing SEA instance from the Database"""
    # Delete the existing entry in the specified Database table
    _delete_db_entry(context, dbmodels.SharedEthernetAdapterDTO,
                     dict(id=sea_id), dbtxn)


def _sea_vea_filters(context, filters, dbtxn=None):
    """Helper method to convert the filters to the DB schema columns"""
    # First call the VIOS Component filter method to convert the VIOS ID
    filters = _vios_component_filters(context, filters, dbtxn)
    # If they filtered on Primary SEA, then need to get the Primary Key ID
    if filters and 'primary_vea_id' in filters:
        filters['_primary_vea_pk_id'] = _vea_get_vea_pk_id(
            context, filters.pop('primary_vea_id', None), dbtxn)
    # If they filtered on Control Channel, then need to get the Primary Key ID
    if filters and 'control_channel_vea_id' in filters:
        filters['_control_channel_vea_pk_id'] = _vea_get_vea_pk_id(
            context, filters.pop('control_channel_vea_id', None), dbtxn)
    return filters


def _sea_vea_values(context, values, is_update=False, dbtxn=None):
    """Helper method to convert the values to the DB schema columns"""
    # First call the VIOS Component values method to convert the VIOS ID
    values = _vios_component_values(context, values, is_update, dbtxn)
    # If the Primary VEA is provided, need to convert to the Primary Key ID
    if 'primary_vea_id' in values:
        values['_primary_vea_pk_id'] = _vea_get_vea_pk_id(
            context, values.pop('primary_vea_id', None), dbtxn)
    # If the Control Channel is provided, need to convert to the Primary Key ID
    if 'control_channel_vea_id' in values:
        values['_control_channel_vea_pk_id'] = _vea_get_vea_pk_id(
            context, values.pop('control_channel_vea_id', None), dbtxn)
    return values


def _sea_vea_result_set(context, result_set, dbtxn=None):
    # If this is a list, then convert each item in that Result Set list
    if isinstance(result_set, list):
        result_set = _vios_component_result_set(context, result_set, dbtxn)
        # If there are a bunch of rows, query all VEA's to map
        if len(result_set) >= 5:
            _sea_populate_vea_ids(context, result_set, dbtxn)
        return [_sea_vea_result_set(
                context, itm, dbtxn) for itm in result_set]
    # First call the VIOS Component result set method to convert the VIOS ID
    result_set = _vios_component_result_set(context, result_set, dbtxn)
    # Convert the Primary Key ID to a Primary VEA to be returned
    if 'primary_vea_id' not in result_set:
        result_set['primary_vea_id'] = _vea_get_vea_id(
            context, result_set.pop('_primary_vea_pk_id', None), dbtxn)
    # Convert the Primary Key ID to a Control Channel to be returned
    if 'control_channel_vea_id' not in result_set:
        result_set['control_channel_vea_id'] = _vea_get_vea_id(
            context, result_set.pop('_control_channel_vea_pk_id', None), dbtxn)
    return result_set


def _sea_get_sea_pk_id(context, sea_id, dbtxn=None):
    """Helper method to get the Primary Key ID from the SEA ID"""
    return _get_primary_key_from_id(
        context, dbmodels.SharedEthernetAdapterDTO, sea_id, dbtxn)


def _sea_get_sea_id(context, sea_pk_id, dbtxn=None):
    """Helper method to get the SEA ID from the Primary Key ID"""
    return _get_id_from_primary_key(
        context, dbmodels.SharedEthernetAdapterDTO, sea_pk_id, dbtxn)


def _sea_populate_vea_ids(context, result_set, dbtxn=None):
    """Helper method to populate the VEA ID from the Primary Key ID"""
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # We need to query all the VEA's to get the mapping
    query = session.query(dbmodels.VirtualEthernetAdapterDTO._pk_id,
                          dbmodels.VirtualEthernetAdapterDTO.id)
    vea_map = dict([(r[0], r[1]) for r in query.all()])
    # Loop through the Result Set, populating the VEA ID
    for row in result_set:
        row['primary_vea_id'] = vea_map.get(row.pop('_primary_vea_pk_id'))
        row['control_channel_vea_id'] = \
            vea_map.get(row.pop('_control_channel_vea_pk_id'))


##################################################
#           VEA DB API Implementation            #
##################################################
def vea_find_all(context, filters=None, dbtxn=None):
    """Retrieves all of the matching VEA's that are in the Database"""
    # Translated the Filters to account for the VIOS ID and SEA
    filters = _vea_sea_filters(context, filters, dbtxn)
    # Query the list of matching entries from the given Database table
    result_set = _find_all_db_entries(
        context, dbmodels.VirtualEthernetAdapterDTO, filters, dbtxn)
    # Translate the Result Values to account for the VIOS ID and SEA
    return _vea_sea_result_set(context, result_set, dbtxn)


def vea_create(context, values, dbtxn=None):
    """Creates a new VEA instance in the Database"""
    # Translate the Input Values to account for the VIOS ID and SEA
    values = _vea_sea_values(context, values, False, dbtxn)
    # Insert the new entry in the specified Database table
    result_set = _create_db_entry(
        context, dbmodels.VirtualEthernetAdapterDTO, values, dbtxn)
    # Translate the Result Values to account for the VIOS ID and SEA
    return _vea_sea_result_set(context, result_set, dbtxn)


def vea_update(context, vea_id, values, dbtxn=None):
    """Updates an existing VEA instance in the Database"""
    filters = dict(id=vea_id)
    # Translate the Input Values to account for the VIOS ID and SEA
    values = _vea_sea_values(context, values, True, dbtxn)
    # Update the existing entry in the specified Database table
    result_set = _update_db_entry(
        context, dbmodels.VirtualEthernetAdapterDTO, filters, values, dbtxn)
    # Translate the Result Values to account for the VIOS ID and SEA
    return _vea_sea_result_set(context, result_set, dbtxn)


def vea_delete(context, vea_id, dbtxn=None):
    """Removes an existing VEA instance from the Database"""
    # Delete the existing entry in the specified Database table
    _delete_db_entry(context, dbmodels.VirtualEthernetAdapterDTO,
                     dict(id=vea_id), dbtxn)


def _vea_sea_filters(context, filters, dbtxn=None):
    """Helper method to convert the filters to the DB schema columns"""
    # First call the VIOS Component filter method to convert the VIOS ID
    filters = _vios_component_filters(context, filters, dbtxn)
    # If they filtered on the SEA ID, then we need to get the Primary Key ID
    if filters and 'sea_id' in filters:
        filters['_sea_pk_id'] = _sea_get_sea_pk_id(
            context, filters.pop('sea_id', None), dbtxn)
    return filters


def _vea_sea_values(context, values, is_update=False, dbtxn=None):
    """Helper method to convert the values to the DB schema columns"""
    # First call the VIOS Component values method to convert the VIOS ID
    values = _vios_component_values(context, values, is_update, dbtxn)
    # If the SEA ID is provided, need to convert to the Primary Key ID
    if 'sea_id' in values:
        values['_sea_pk_id'] = _sea_get_sea_pk_id(
            context, values.pop('sea_id', None), dbtxn)
    # If the VLAN ID's are provided, need to convert to a String format
    if 'addl_vlan_ids' in values:
        values['_vlan_ids_string'] = ','.join(map(str,
                                                  values['addl_vlan_ids']))
    return values


def _vea_sea_result_set(context, result_set, dbtxn=None):
    # If this is a list, then convert each item in that Result Set list
    if isinstance(result_set, list):
        # If there are a bunch of rows, query all SEA's to map
        result_set = _vios_component_result_set(context, result_set, dbtxn)
        if len(result_set) >= 5:
            _vea_populate_sea_id(context, result_set, dbtxn)
        return [_vea_sea_result_set(
            context, itm, dbtxn) for itm in result_set]
    # First call the VIOS Component result set method to convert the VIOS ID
    result_set = _vios_component_result_set(context, result_set, dbtxn)
    # Convert the Primary Key ID to a SEA ID to be returned
    if 'sea_id' not in result_set:
        result_set['sea_id'] = _sea_get_sea_id(
            context, result_set.pop('_sea_pk_id', None), dbtxn)
    # Convert the VLAN ID's from the String format to a List
    vlan_ids = result_set.pop('_vlan_ids_string', '').split(',')
    result_set['addl_vlan_ids'] = map(int, vlan_ids) if (vlan_ids and
                                                         vlan_ids[0]) else []
    return result_set


def _vea_get_vea_pk_id(context, vea_id, dbtxn=None):
    """Helper method to get the Primary Key ID from the VEA ID"""
    return _get_primary_key_from_id(
        context, dbmodels.VirtualEthernetAdapterDTO, vea_id, dbtxn)


def _vea_get_vea_id(context, vea_pk_id, dbtxn=None):
    """Helper method to get the VEA ID from the Primary Key ID"""
    return _get_id_from_primary_key(
        context, dbmodels.VirtualEthernetAdapterDTO, vea_pk_id, dbtxn)


def _vea_populate_sea_id(context, result_set, dbtxn=None):
    """Helper method to populate the SEA ID from the Primary Key ID"""
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # We need to query all the SEA's to get the mapping
    query = session.query(dbmodels.SharedEthernetAdapterDTO._pk_id,
                          dbmodels.SharedEthernetAdapterDTO.id)
    sea_map = dict([(r[0], r[1]) for r in query.all()])
    # Loop through the Result Set, populating the SEA ID
    for row in result_set:
        row['sea_id'] = sea_map.get(row.pop('_sea_pk_id'))


##################################################
#   Network Association DB API Implementation    #
##################################################
def network_assoc_find_all(context, filters=None, dbtxn=None):
    """Retrieves all of the matching Network Associations that are in the DB"""
    # Translate the Filters to account for the Host Name and SEA
    filters = _netw_assoc_filters(context, filters, dbtxn)
    # Query the list of matching entries from the given Database table
    result_set = _find_all_db_entries(
        context, dbmodels.NetworkAssociationDTO, filters, dbtxn)
    # Translate the Result Values to account for the Host Name and SEA
    return _netw_assoc_result_set(context, result_set, dbtxn)


def network_assoc_create(context, values, dbtxn=None):
    """Creates a new Network Association instance in the Database"""
    # Translate the Input Values to account for the Host Name and SEA
    values = _netw_assoc_values(context, values, False, dbtxn)
    # Insert the new entry in the specified Database table
    result_set = _create_db_entry(
        context, dbmodels.NetworkAssociationDTO, values, dbtxn)
    # Translate the Result Values to account for the Host Name and SEA
    return _netw_assoc_result_set(context, result_set, dbtxn)


def network_assoc_update(context, netwassoc_id, values, dbtxn=None):
    """Updates an existing Network Association instance in the Database"""
    filters = dict(id=netwassoc_id)
    # Translate the Input Values to account for the Host Name and SEA
    values = _netw_assoc_values(context, values, True, dbtxn)
    # Update the existing entry in the specified Database table
    result_set = _update_db_entry(
        context, dbmodels.NetworkAssociationDTO, filters, values, dbtxn)
    # Translate the Result Values to account for the Host Name and SEA
    return _netw_assoc_result_set(context, result_set, dbtxn)


def network_assoc_delete(context, netwassoc_id, dbtxn=None):
    """Removes an existing Network Association instance from the Database"""
    # Delete the existing entry in the specified Database table
    _delete_db_entry(context, dbmodels.NetworkAssociationDTO,
                     dict(id=netwassoc_id), dbtxn)


def _netw_assoc_filters(context, filters, dbtxn=None):
    """Helper method to convert the filters to the DB schema columns"""
    # First call the Host Component filter method to convert the Host Name
    filters = _host_component_filters(context, filters, dbtxn)
    # If they filtered on the SEA ID, then we need to get the Primary Key ID
    if filters and 'sea_id' in filters:
        filters['_sea_pk_id'] = _sea_get_sea_pk_id(
            context, filters.pop('sea_id', None), dbtxn)
    return filters


def _netw_assoc_values(context, values, is_update=False, dbtxn=None):
    """Helper method to convert the values to the DB schema columns"""
    # First call the Host Component values method to convert the Host Name
    values = _host_component_values(context, values, is_update, dbtxn)
    # If the SEA ID is provided, need to convert to the Primary Key ID
    if 'sea_id' in values:
        values['_sea_pk_id'] = _sea_get_sea_pk_id(
            context, values.pop('sea_id', None), dbtxn)
    return values


def _netw_assoc_result_set(context, result_set, dbtxn=None):
    # If this is a list, then convert each item in that Result Set list
    if isinstance(result_set, list):
        result_set = _host_component_result_set(context, result_set, dbtxn)
        # If there are a bunch of rows, query all SEA's to map
        if len(result_set) >= 5:
            _netw_assoc_populate_sea_id(context, result_set, dbtxn)
        return [_netw_assoc_result_set(
                context, itm, dbtxn) for itm in result_set]
    # First call the Host Component result set method to convert the Host Name
    result_set = _host_component_result_set(context, result_set, dbtxn)
    # Convert the Primary Key ID to a SEA ID to be returned
    if 'sea_id' not in result_set:
        result_set['sea_id'] = _sea_get_sea_id(
            context, result_set.pop('_sea_pk_id', None), dbtxn)
    return result_set


def _netw_assoc_populate_sea_id(context, result_set, dbtxn=None):
    """Helper method to populate the SEA ID from the Primary Key ID"""
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # We need to query all the SEA's to get the mapping
    query = session.query(dbmodels.SharedEthernetAdapterDTO._pk_id,
                          dbmodels.SharedEthernetAdapterDTO.id)
    sea_map = dict([(r[0], r[1]) for r in query.all()])
    # Loop through the Result Set, populating the SEA ID
    for row in result_set:
        row['sea_id'] = sea_map.get(row.pop('_sea_pk_id'))


##################################################
#           SCG DB API Implementation            #
##################################################
def scg_find_all(context, filters=None, dbtxn=None):
    """Retrieves all of the matching SCG's that are in the Database"""
    # Translated the Filters to account for the VIOS IDs
    filters = _scg_vios_filters(context, filters, dbtxn)
    # Query the list of matching entries from the given Database table
    result_set = _find_all_db_entries(
        context, dbmodels.StorageConnectivityGroupDTO, filters, dbtxn)
    # Translate the Result Values to account for the VIOS IDs
    return _scg_vios_result_set(context, result_set, dbtxn)


def scg_create(context, values, dbtxn=None):
    """Creates a new SCG instance in the Database"""
    vioses = values.pop('vios_ids', None)
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # Translate the Input Values to account for the VIOS IDs
    values = _scg_vios_values(context, values, False, session)
    # Create a Transaction around the create in the Database
    with session.begin(subtransactions=True):
        # Insert the new entry in the specified Database table
        result_set = _create_db_entry(
            context, dbmodels.StorageConnectivityGroupDTO, values, session)
        # Update the hmc_hosts table with the new list of Hosts
        _scg_update_vioses(context, values['id'], vioses, session)
    # Translate the Result Values to account for the Hosts
    return _scg_vios_result_set(context, result_set, session)


def scg_update(context, scg_id, values, dbtxn=None):
    """Updates an existing SCG instance in the Database"""
    filters = dict(id=scg_id)
    vioses = values.pop('vios_ids', None)
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # Translate the Input Values to account for the VIOS IDs
    values = _scg_vios_values(context, values, True, session)
    # Create a Transaction around the update in the Database
    with session.begin(subtransactions=True):
        # Update the existing entry in the specified Database table
        result_set = _update_db_entry(
            context, dbmodels.StorageConnectivityGroupDTO,
            filters, values, session)
        # Update the scg_vios_assocation table with the new list of VIOS's
        _scg_update_vioses(context, scg_id, vioses, session)
    # Translate the Result Values to account for the SCG VIOS's
    return _scg_vios_result_set(context, result_set, session)


def scg_delete(context, scg_id, dbtxn=None):
    """Removes an existing SCG instance from the Database"""
    scg_pk_id = _scg_get_scg_pk_id(context, scg_id, dbtxn)
    filters = [dict(_scg_pk_id=scg_pk_id), dict(id=scg_id)]
    model_classes = [dbmodels.ScgViosAssociationDTO,
                     dbmodels.StorageConnectivityGroupDTO]
    # Delete the existing entries from the specified Database tables
    _delete_db_entries(context, model_classes, filters, dbtxn)


def _scg_vios_filters(context, filters, dbtxn=None):
    """Helper method to convert the filters to the DB schema columns"""
    filters = filters.copy() if filters is not None else None
    # If they asked us to filter on the VIOS, we need something special
    if filters and 'vios_id' in filters:
        vios_id = filters.pop('vios_id', None)
        # If they also filtered on the SCG ID, that takes precedence
        if 'id' not in filters:
            scg_ids = _scg_get_scgs(context, vios_id, dbtxn)
            # Add a special filter to handle a bunch of SCG ID's
            filters['special'] = filters.get('special', [])
            filters['special'].append(
                dbmodels.StorageConnectivityGroupDTO.id.in_(scg_ids))
    return filters


def _scg_vios_values(context, values, is_update=False, dbtxn=None):
    """Helper method to convert the values to the DB schema columns"""
    values = values.copy()
    # We only want to generate the ID attribute if this is a Create
    if is_update is False:
        values['id'] = values.get('id', str(uuid.uuid4()))
    # Convert the List properties to the appropriate form
    if values.get('boot_connectivity') is not None:
        values['boot_connectivity'] =\
            ",".join([str(x) for x in values['boot_connectivity']])
    if values.get('data_connectivity') is not None:
        values['data_connectivity'] =\
            ",".join([str(x) for x in values['data_connectivity']])
    if values.get('fabric_set_npiv') is not None:
        values['fabric_set_npiv'] =\
            ",".join([str(x) for x in values['fabric_set_npiv']])
    return values


def _scg_vios_result_set(context, result_set, dbtxn=None):
    # If this is a list, then convert each item in that Result Set list
    if isinstance(result_set, list):
        return [_scg_vios_result_set(
                context, itm, dbtxn) for itm in result_set]
    # Add the VIOS's that are part by the SCG to the SCG returned
    result_set['vios_ids'] = \
        _scg_get_vioses(context, result_set['id'], dbtxn)
    # Convert the List properties to the appropriate form
    if result_set.get('boot_connectivity') is not None:
        result_set['boot_connectivity'] =\
            result_set['boot_connectivity'].split(',')
    if result_set.get('data_connectivity') is not None:
        result_set['data_connectivity'] =\
            result_set['data_connectivity'].split(',')
    if result_set.get('fabric_set_npiv') is not None:
        result_set['fabric_set_npiv'] =\
            result_set['fabric_set_npiv'].split(',')
    return result_set


def _scg_get_vioses(context, scg_id, dbtxn=None):
    """Helper method to list the VIOS's that are part of the SCG"""
    scg_vioses = list()
    session = get_dbtransaction() if dbtxn is None else dbtxn
    scg_pk_id = _scg_get_scg_pk_id(context, scg_id, session)
    # We need to query the VIOS's for the SCG based on the Primary Key
    query = session.query(dbmodels.ScgViosAssociationDTO._vios_pk_id)
    vioses = query.filter_by(_scg_pk_id=scg_pk_id, deleted=0).all()
    for vios in vioses:
        vios_id = _vios_get_vios_id(context, vios[0], session)
        # Only add the VIOS to the List if it really does exist still
        # This is to get around timing windows where the VIOS is deleted
        if vios_id is not None:
            scg_vioses.append(vios_id)
    return scg_vioses


def _scg_get_scgs(context, vios_id, dbtxn=None):
    """Helper method to list the VIOS's that are part of the SCG"""
    session = get_dbtransaction() if dbtxn is None else dbtxn
    vios_pk_id = _vios_get_vios_pk_id(context, vios_id, session)
    # We need to query the VIOS's for the SCG based on the Primary Key
    query = session.query(dbmodels.ScgViosAssociationDTO._scg_pk_id)
    scgs = query.filter_by(_vios_pk_id=vios_pk_id, deleted=0).all()
    return [_scg_get_scg_id(context, scg[0], session) for scg in scgs]


def _scg_update_vioses(context, scg_id, vioses, dbtxn=None):
    """Helper method to update the VIOS's that are part of the SCG"""
    # If we weren't given any VIOS's, then there is nothing to do
    if vioses is not None:
        scg_pk_id = _scg_get_scg_pk_id(context, scg_id, dbtxn)
        # Query the list of VIOS's that are associated with this SCG
        db_vioses = _scg_get_vioses(context, scg_id, dbtxn)
        # For each VIOS provided that doesn't exist, add it to the database
        for vios in vioses:
            if vios not in db_vioses:
                vios_pk_id = _vios_get_vios_pk_id(context, vios, dbtxn)
                values = dict(_scg_pk_id=scg_pk_id, _vios_pk_id=vios_pk_id)
                _create_db_entry(
                    context, dbmodels.ScgViosAssociationDTO, values, dbtxn)
        # For each VIOS in the DB but not provided, remove from database
        for db_vios in db_vioses:
            if db_vios not in vioses:
                vios_pk_id = _vios_get_vios_pk_id(context, db_vios, dbtxn)
                filters = dict(_scg_pk_id=scg_pk_id, _vios_pk_id=vios_pk_id)
                _delete_db_entry(
                    context, dbmodels.ScgViosAssociationDTO, filters, dbtxn)


def _scg_get_scg_pk_id(context, scg_id, dbtxn=None):
    """Helper method to get the Primary Key ID from the SCG ID"""
    return _get_primary_key_from_id(
        context, dbmodels.StorageConnectivityGroupDTO, scg_id, dbtxn)


def _scg_get_scg_id(context, scg_pk_id, dbtxn=None):
    """Helper method to get the SCG ID from the Primary Key ID"""
    return _get_id_from_primary_key(
        context, dbmodels.StorageConnectivityGroupDTO, scg_pk_id, dbtxn)


##################################################
#          FCPort DB API Implementation          #
##################################################
def fcport_find_all(context, filters=None, dbtxn=None):
    """Retrieves all of the matching FCPorts that are in the Database"""
    # Translate the Filters to account for the VIOS ID
    filters = _vios_component_filters(context, filters, dbtxn)
    # Query the list of matching entries from the given Database table
    result_set = _find_all_db_entries(
        context, dbmodels.FcPortDTO, filters, dbtxn)
    # Translate the Result Values to account for the VIOS ID
    return _vios_component_result_set(context, result_set, dbtxn)


def fcport_create(context, values, dbtxn=None):
    """Creates a new FCPort instance in the Database"""
    # Translate the Input Values to account for the VIOS ID
    values = _vios_component_values(context, values, False, dbtxn)
    # Insert the new entry in the specified Database table
    result_set = _create_db_entry(
        context, dbmodels.FcPortDTO, values, dbtxn)
    # Translate the Result Values to account for the VIOS ID
    return _vios_component_result_set(context, result_set, dbtxn)


def fcport_update(context, port_id, values, dbtxn=None):
    """Updates an existing FCPort instance in the Database"""
    filters = dict(id=port_id)
    # Translate the Input Values to account for the VIOS ID
    values = _vios_component_values(context, values, True, dbtxn)
    # Update the existing entry in the specified Database table
    result_set = _update_db_entry(
        context, dbmodels.FcPortDTO, filters, values, dbtxn)
    # Translate the Result Values to account for the VIOS ID
    return _vios_component_result_set(context, result_set, dbtxn)


def fcport_delete(context, port_id, dbtxn=None):
    """Removes an existing FCPort instance from the Database"""
    # Delete the existing entry in the specified Database table
    _delete_db_entry(context, dbmodels.FcPortDTO,
                     dict(id=port_id), dbtxn)


##################################################
#      On-board Task DB API Implementation       #
##################################################
def onboard_task_find_all(context, host, dbtxn=None):
    """Retrieves all of the On-board Tasks from the DB."""
    # Query the list of matching entries from the given Database table
    return _find_all_db_entries(
        context, dbmodels.OnboardTaskDTO, dict(host=host), dbtxn)


def onboard_task_find(context, task_id, dbtxn=None):
    """Retrieves one of the On-board Tasks from the DB."""
    # Query the list of matching entries from the given Database table
    result_set = _find_all_db_entries(
        context, dbmodels.OnboardTaskDTO, dict(id=task_id), dbtxn)
    # If we got a Task back, we need to append the Server records to it
    if result_set:
        filters = dict(task_id=task_id)
        result_set[0]['servers'] = _find_all_db_entries(
            context, dbmodels.OnboardTaskServerDTO, filters, dbtxn)
    return result_set[0] if result_set else None


def onboard_task_create(context, host, dbtxn=None):
    """Creates the given On-board Task in the DB."""
    # Insert the new entry in the specified Database table
    return _create_db_entry(
        context, dbmodels.OnboardTaskDTO, dict(host=host), dbtxn)


def onboard_task_update(context, task_id, values, dbtxn=None):
    """Updates the given On-board Task in the DB."""
    filters = dict(id=task_id)
    values = dict([(k, v) for k, v in values.iteritems() if v is not None])
    status = values.get('status', '')
    # If this is a final status, then set the end date/time
    if status == 'completed' or status == 'failed':
        values['ended'] = timeutils.utcnow()
    # Update the existing entry in the specified Database table
    return _update_db_entry(
        context, dbmodels.OnboardTaskDTO, filters, values, dbtxn)


def onboard_task_delete(context, task_id, dbtxn=None):
    """Deletes the given On-board Task from the DB."""
    filters = [dict(task_id=task_id), dict(id=task_id)]
    model_classes = [dbmodels.OnboardTaskServerDTO, dbmodels.OnboardTaskDTO]
    # Delete the existing entries from the specified Database tables
    _delete_db_entries(context, model_classes, filters, dbtxn, True)


def onboard_task_cleanup(context, dbtxn=None):
    """Cleanup any stuck On-board Task in the DB."""
    session = get_dbtransaction() if dbtxn is None else dbtxn
    checktime = datetime.datetime.utcnow() - datetime.timedelta(seconds=3600)
    # Create a Transaction around the update in the Database
    with session.begin(subtransactions=True):
        query = _model_query(context, dbmodels.OnboardTaskDTO, session=session)
        query = query.filter(dbmodels.OnboardTaskDTO.progress < 100)
        query = query.filter(dbmodels.OnboardTaskDTO.created_at < checktime)
        # For any tasks considered stuck, mark them done and failed
        for atask in query.all():
            atask.update(dict(progress=100, status='failed',
                              ended_at=datetime.datetime.utcnow()))
            atask.save(session=session)


def onboard_task_server_create(context, task_id, svr_uuid, values, dbtxn=None):
    """Create the Server record for the given On-board Task"""
    values = dict(values, task_id=task_id, server_uuid=svr_uuid)
    # Insert the new entry in the specified Database table
    return _create_db_entry(
        context, dbmodels.OnboardTaskServerDTO, values, dbtxn)


def onboard_task_server_update(context, task_id, svr_uuid, values, dbtxn=None):
    """Updates the Server record for the given On-board Task"""
    filters = dict(task_id=task_id, server_uuid=svr_uuid)
    values = dict([(k, v) for k, v in values.iteritems() if v is not None])
    # Update the existing entry in the specified Database table
    return _update_db_entry(
        context, dbmodels.OnboardTaskServerDTO, filters, values, dbtxn)


##################################################
#     Console Messages DB API Implementation     #
##################################################
def console_messages_find_all(context, filters=None, maximum=None):
    """Retrieves all of the matching Messages that are in the Database"""
    filters = _console_messages_filters(context, filters)
    query = _model_query(context, dbmodels.ConsoleMessagesDTO)
    # If there were any filters provided, filter the query with them
    if filters:
        special_filters = filters.pop('special', [])
        query = query.filter_by(**filters)
        # Add any special filters we were given to the query
        for special_filter in special_filters:
            query = query.filter(special_filter)
    # We want to order results based on time in case there is a maximum
    query.order_by(dbmodels.ConsoleMessagesDTO.msg_time.desc())
    # Limit the results to the maximum they ask for if they did
    result_set = query[:maximum] if maximum else query.all()
    return _console_messages_result_set(context, result_set)


def console_messages_create(context, values):
    """Creates a new Console Message instance in the Database"""
    session = get_dbtransaction()
    values = _console_messages_values(context, values)
    # Create a Transaction around the create in the Database
    with session.begin():
        msg_ref = dbmodels.ConsoleMessagesDTO()
        msg_ref.update(values)
        msg_ref.save(session=session)
    return _console_messages_result_set(context, msg_ref)


def console_messages_update(context, msg_id, values):
    """Updates an existing Console Message instance in the Database"""
    session = get_dbtransaction()
    values = _console_messages_values(context, values)
    # Create a Transaction around the update in the Database
    with session.begin():
        query = _model_query(
            context, dbmodels.ConsoleMessagesDTO, session=session)
        msg_ref = query.filter_by(id=msg_id).first()
        msg_ref.update(values)
        msg_ref.save(session=session)
    return _console_messages_result_set(context, msg_ref)


def console_messages_delete(context, msg_id):
    """Removes an existing Console Message instance from the Database"""
    session = get_dbtransaction()
    # Create a Transaction around the delete in the Database
    with session.begin():
        query = _model_query(
            context, dbmodels.ConsoleMessagesDTO, session=session)
        query = query.filter_by(id=msg_id)
        query.soft_delete(synchronize_session=False)


def console_messages_prune(context, maximum):
    """Prunes older Console Messages from the Database when maximum exceeded"""
    session = get_dbtransaction()
    query = session.query(dbmodels.ConsoleMessagesDTO.msg_time)
    query = query.filter_by(deleted=0)
    msg_count = query.count()
    if msg_count > maximum:
        query = query.order_by(dbmodels.ConsoleMessagesDTO.msg_time.desc())
        msg_time = query[maximum - 1].msg_time
        # Create a Transaction around the delete in the Database
        with session.begin():
            # First we want to do a hard delete on any deleted entries
            query = _model_query(context, dbmodels.ConsoleMessagesDTO,
                                 session=session, read_deleted='yes')
            query = query.filter(dbmodels.ConsoleMessagesDTO.deleted > 0)
            query.delete(synchronize_session=False)
            # Now we can go through and delete any older messages than the max
            query = _model_query(
                context, dbmodels.ConsoleMessagesDTO, session=session)
            query = query.filter(
                dbmodels.ConsoleMessagesDTO.msg_time < msg_time)
            query.delete(synchronize_session=False)


def _console_messages_filters(context, filters):
    """Helper method to convert the filters to the DB schema columns"""
    filters = filters.copy() if filters is not None else dict()
    starttime = filters.pop('msg_start', None)
    endtime = filters.pop('msg_end', None)
    filters['special'] = list()
    # If there was a Start Time provided, use as the minimum time
    if starttime is not None:
        filters['special'].append(
            dbmodels.ConsoleMessagesDTO.msg_time >= starttime)
    # If there was a Start Time provided, use as the maximum time
    if endtime is not None:
        filters['special'].append(
            dbmodels.ConsoleMessagesDTO.msg_time <= endtime)
    # If there was a User in the context, filter on that user only
    if context.user_name:
        user_clause = or_(
            dbmodels.ConsoleMessagesDTO.msg_user == context.user_name,
            dbmodels.ConsoleMessagesDTO.msg_user.is_(None))
        filters['special'].append(user_clause)
    # If there was a Project in the context, filter on that project only
    if context.project_id:
        proj_clause = or_(
            dbmodels.ConsoleMessagesDTO.project_id == context.project_id,
            dbmodels.ConsoleMessagesDTO.project_id.is_(None))
        filters['special'].append(proj_clause)
    return filters


def _console_messages_values(context, values):
    """Helper method to convert the values to the DB schema columns"""
    values = values.copy()
    if values.get('payload') is not None:
        if isinstance(values['payload'], dict):
            values['payload'] = jsonutils.dumps(values['payload'])
    return values


def _console_messages_result_set(context, result_set):
    # If this is a list, then convert each item in that Result Set list
    if isinstance(result_set, list):
        return [_console_messages_result_set(
                context, itm) for itm in result_set]
    # Convert the Compute Node ID to a Host Name to be returned
    result_set = jsonutils.to_primitive(result_set)
    result_set['payload'] = jsonutils.loads(result_set['payload'])
    return result_set


###############################################################
#            Compute Node Metadata API Utilities              #
###############################################################
def _compute_node_metadata_get_query(context, compute_node_id, session):
    return _model_query(context, dbmodels.IBMComputeNodeMetadataDTO,
                        session=session, read_deleted="no").\
        filter_by(compute_node_id=compute_node_id)


def compute_node_metadata_get(context, compute_node_id, dbtxn=None):
    session = get_dbtransaction() if dbtxn is None else dbtxn
    rows = _compute_node_metadata_get_query(context, compute_node_id,
                                            session).all()
    return {row['key']: row['value'] for row in rows}


def compute_node_metadata_delete(context, compute_node_id, key, dbtxn=None):
    session = get_dbtransaction() if dbtxn is None else dbtxn
    _compute_node_metadata_get_query(context, compute_node_id, session).\
        filter_by(key=key).\
        soft_delete()


def compute_node_metadata_update(context, compute_node_id, metadata,
                                 set_delete=False, dbtxn=None):
    all_keys = metadata.keys()
    session = get_dbtransaction() if dbtxn is None else dbtxn
    with session.begin(subtransactions=True):
        if set_delete:
            _compute_node_metadata_get_query(context, compute_node_id,
                                             session).\
                filter(~dbmodels.IBMComputeNodeMetadataDTO.key.
                       in_(all_keys)).soft_delete(synchronize_session=False)

        already_existing_keys = []
        meta_refs = _compute_node_metadata_get_query(context, compute_node_id,
                                                     session).\
            filter(dbmodels.IBMComputeNodeMetadataDTO.key.in_(all_keys)).\
            all()

        for meta_ref in meta_refs:
            already_existing_keys.append(meta_ref.key)
            meta_ref.update({"value": metadata[meta_ref.key]})

        new_keys = set(all_keys) - set(already_existing_keys)
        for key in new_keys:
            meta_ref = dbmodels.IBMComputeNodeMetadataDTO()
            meta_ref.update({"key": key, "value": metadata[key],
                             "compute_node_id": compute_node_id})
            session.add(meta_ref)

        return metadata


###############################################################
# Enterprise Pools (Capacity on Demand) DB API Implementation #
###############################################################
def enterprise_pool_find_all(context, filters=None, dbtxn=None):
    """Retrieves all of the matching enterprise pools that are in the DB"""
    # Translate the Filters to account for the Hosts
    filters = _enterprise_pool_hosts_filters(context, filters, dbtxn)
    # Query the list of matching entries from the given Database table
    result_set = _find_all_db_entries(
        context, dbmodels.EnterprisePoolDTO, filters, dbtxn)
    # Translate the Result Values to account for the Hosts
    return _enterprise_pool_hosts_result_set(context, result_set, dbtxn)


def enterprise_pool_create(context, values, dbtxn=None):
    """Creates a new enterprise pool instance in the Database"""
    hosts = values.pop('host_names', None)
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # Create a Transaction around the create in the Database
    with session.begin(subtransactions=True):
        # Insert the new entry in the specified Database table
        result_set = _create_db_entry(
            context, dbmodels.EnterprisePoolDTO, values, session)
        # Update the enterprise pool hosts table with the new list of hosts
        _enterprise_pool_update_hosts(context, values['uuid'], hosts, session)
    # Translate the Result Values to account for the Hosts
    return _enterprise_pool_hosts_result_set(context, result_set, session)


def enterprise_pool_update(context, pool_uuid, values, dbtxn=None):
    """Updates an existing enterprise pool instance in the Database"""
    filters = dict(uuid=pool_uuid)
    hosts = values.pop('host_names', None)
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # Create a Transaction around the create in the Database
    with session.begin(subtransactions=True):
        # Update the existing entry in the specified Database table
        result_set = _update_db_entry(
            context, dbmodels.EnterprisePoolDTO,
            filters, values, session)
        # Update the enterprise pool hosts table with the new list of Hosts
        _enterprise_pool_update_hosts(context, pool_uuid, hosts, session)
    # Translate the Result Values to account for the Hosts
    return _enterprise_pool_hosts_result_set(context, result_set, session)


def enterprise_pool_delete(context, pool_uuid, dbtxn=None):
    """Removes an existing enterprise pool instance from the Database"""
    filters = [dict(uuid=pool_uuid), dict(pool_uuid=pool_uuid)]
    model_classes = [dbmodels.EnterprisePoolDTO,
                     dbmodels.EnterprisePoolMembersDTO]
    # Delete the existing entries from the specified Database tables
    _delete_db_entries(context, model_classes, filters, dbtxn, True)


def _enterprise_pool_hosts_filters(context, filters, dbtxn=None):
    """Helper method to convert the filters to the DB schema columns"""
    filters = filters.copy() if filters is not None else None
    # If they asked us to filter on the host name, we need something special
    if filters and 'host_name' in filters:
        host_name = filters.pop('host_name', None)
        # If they also filtered on the pool UUID, that takes precedence
        if 'uuid' not in filters:
            uuids = _enterprise_pool_get_pools(context, host_name, dbtxn)
            # Add a special filter to handle a bunch of pool UUIDs
            filters['special'] = filters.get('special', [])
            filters['special'].append(
                dbmodels.EnterprisePoolDTO.uuid.in_(uuids))
    return filters


def _enterprise_pool_hosts_result_set(context, result_set, dbtxn=None):
    """Helper method to convert the results from the DB schema columns"""
    # If this is a list, then convert each item in that Result Set list
    if isinstance(result_set, list):
        return [_enterprise_pool_hosts_result_set(
                context, itm, dbtxn) for itm in result_set]
    # Add the Hosts that are managed by the enterprise pool to the enterprise
    # pool returned
    result_set['host_names'] = _enterprise_pool_get_hosts(context,
                                                          result_set['uuid'],
                                                          dbtxn)
    return result_set


def _enterprise_pool_get_hosts(context, pool_uuid, dbtxn=None):
    """Helper method to list the Hosts that are managed by a enterprise pool"""
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # We need to query the Hosts for the enterprise pool based on the pool UUID
    query = session.query(dbmodels.EnterprisePoolMembersDTO.host_name,
                          dbmodels.EnterprisePoolMembersDTO.member_uuid)
    host_names = query.filter_by(pool_uuid=pool_uuid, deleted=0).all()
    return dict(host_names)


def _enterprise_pool_get_pools(context, host_name, dbtxn=None):
    """Helper method to list enterprise pools for a given host"""
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # We need to query the members for the enterprise pool based on pool UUID
    query = session.query(dbmodels.EnterprisePoolMembersDTO.pool_uuid)
    uuids = query.filter_by(host_name=host_name, deleted=0).all()
    return [uuid[0] for uuid in uuids]


def _enterprise_pool_update_hosts(context, pool_uuid, hosts, dbtxn=None):
    """Helper method to update the Hosts that belong to an enterprise pool"""
    # If we weren't given any Hosts, then there is nothing to do
    if hosts is not None:
        # Query the list of Host that are associated with this enterprise pool
        db_hosts = _enterprise_pool_get_hosts(context, pool_uuid, dbtxn)
        # For each Host provided that doesn't exist, add it to the database
        new_hosts = [host for host in hosts.items() if host[0] not in db_hosts]
        for new_host in new_hosts:
            values = dict(pool_uuid=pool_uuid, host_name=new_host[0],
                          member_uuid=new_host[1])
            _create_db_entry(
                context, dbmodels.EnterprisePoolMembersDTO,
                values, dbtxn)
        # For each Host provided that exists in the DB, update its member UUID
        # in the database if the member UUID has changed
        hosts_with_changed_member_uuids = [host for host in hosts.items() if
                                           host[0] in db_hosts and
                                           db_hosts[host[0]] != host[1]]
        for updated_host in hosts_with_changed_member_uuids:
            filters = dict(pool_uuid=pool_uuid, host_name=updated_host[0])
            values = dict(pool_uuid=pool_uuid, host_name=updated_host[0],
                          member_uuid=updated_host[1])
            _update_db_entry(
                context, dbmodels.EnterprisePoolMembersDTO, filters,
                values, dbtxn)
        # For each Host in the DB but not provided, remove from database
        removed_hosts = [db_host for db_host in db_hosts.items()
                         if db_host[0] not in hosts]
        for removed_host in removed_hosts:
            filters = dict(pool_uuid=pool_uuid, host_name=removed_host[0],
                           member_uuid=removed_host[1])
            _delete_db_entry(
                context, dbmodels.EnterprisePoolMembersDTO,
                filters, dbtxn, True)


########################################################
# DRO Events  DB API Implementation                    #
########################################################
def _dro_events_values(context, values):
    """Helper method to convert the values to the DB schema columns"""
    values = values.copy()
    if values.get('payload') is not None:
        if isinstance(values['payload'], dict):
            values['payload'] = jsonutils.dumps(values['payload'])
    return values


def _dro_events_filters(context, filters):
    """Helper method to convert the filters to the DB schema columns"""
    filters = filters.copy() if filters is not None else dict()
    start_time = filters.pop('event_start_tstamp', None)
    end_time = filters.pop('event_end_tstamp', None)
    exclude_event_type = filters.pop('exclude_event_type', None)
    filters['special'] = list()
    # If the start time is provided, use the minimum time after start time
    if start_time is not None:
        filters['special'].append(
            dbmodels.DROEventsDTO.event_time >= start_time)
    # If the end time is provided, use the maximum time before end time
    if end_time is not None:
        filters['special'].append(
            dbmodels.DROEventsDTO.event_time <= end_time)
    # If specified, exclude certain event types
    if exclude_event_type is not None:
        filters['special'].append(
            dbmodels.DROEventsDTO.event_type.notin_(exclude_event_type))
    return filters


def _dro_events_result_set(context, result_set):
    """Helper method to convert the ResultSet List to Python list of dicts"""
    if isinstance(result_set, list):
        return [_dro_events_result_set(context, item) for item in result_set]
    result_set = jsonutils.to_primitive(result_set)
    result_set['payload'] = jsonutils.loads(result_set['payload'])
    return result_set


def dro_events_create(context, values):
    """ Create DRO events in database """
    session = get_dbtransaction()
    values = _dro_events_values(context, values)
    with session.begin():
        dro_events = dbmodels.DROEventsDTO()
        dro_events.update(values)
        dro_events.save(session)
    return _dro_events_result_set(context, dro_events)


def dro_events_prune(context, max_event_time):
    """
    Prunes older DRO events from the Database
    :param max_event_time: Maximum time after which events should be deleted.
    """
    session = get_dbtransaction()
    query = session.query(dbmodels.DROEventsDTO.event_time)
    query = query.filter_by(deleted=0)
    query = query.order_by(dbmodels.DROEventsDTO.event_time.desc())
    # Create a Transaction around the delete in the Database
    with session.begin():
        # Hard delete on any deleted entries
        query = _model_query(context, dbmodels.DROEventsDTO,
                             session=session, read_deleted='yes')
        query = query.filter(dbmodels.DROEventsDTO.deleted > 0)
        query.delete(synchronize_session=False)
        # Delete any older events than the max
        query = _model_query(context, dbmodels.DROEventsDTO,
                             session=session)
        query = query.filter(
            dbmodels.DROEventsDTO.event_time <= max_event_time)
        query.delete(synchronize_session=False)


def dro_events_find_all(context, filters=None, maximum=None):
    """Retrieves zero or one matching DRO Events that are in the Database"""
    filters = _dro_events_filters(context, filters)
    query = _model_query(context, dbmodels.DROEventsDTO)
    # If there were any filters provided, filter the query with them
    if filters:
        special_filters = filters.pop('special', [])
        query = query.filter_by(**filters)
        for special_filter in special_filters:
            query = query.filter(special_filter)
    # Order results based on time
    query.order_by(dbmodels.DROEventsDTO.event_time.desc())
    result_set = query[:maximum] if maximum else query.all()
    return _dro_events_result_set(context, result_set)


########################################################
# DRO Schedules  DB API Implementation                 #
########################################################
def _dro_schedules_result_set(context, result_set, group_by_agg=False):
    """Helper method to convert the ResultSet List to Python dictionaries"""
    if isinstance(result_set, list):
        if len(result_set) == 0:
            return {}
        elif group_by_agg:
            # Dict of results grouped by aggregate_id
            return {result['aggregate_id']:
                    jsonutils.loads(
                        _result_set_as_dict(result)['schedules_json'])
                    for result in result_set}
        elif len(result_set) == 1:
            # Dict of results from a single aggregate
            result_str = _result_set_as_dict(result_set[0])['schedules_json']
            return jsonutils.loads(result_str)
        else:
            # Returns the all the raw DB columns as a list of dictionaries
            return _result_set_as_dict(result_set)
    return _result_set_as_dict(result_set)


def dro_schedules_create_aggregate_schedules(context, values, dbtxn=None):
    """Creates DRO schedules for the aggregate"""
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # Create a Transaction around the create in the Database
    with session.begin(subtransactions=True):
        # Delete any "soft" deleted records for the aggregate
        agg_id = values.get('aggregate_id')
        schedules = values.pop('dro_schedules')
        values['schedules_json'] = jsonutils.dumps(schedules)
        query = _model_query(context, dbmodels.DROSchedulesDTO,
                             read_deleted='yes', session=session)
        query = query.filter(
            and_(dbmodels.DROSchedulesDTO.aggregate_id == agg_id,
                 dbmodels.DROSchedulesDTO.deleted > 0))
        if query.first():
            query.delete(synchronize_session=False)

        # Create a new record in the table
        result_set = _create_db_entry(
            context, dbmodels.DROSchedulesDTO, values, session)
        return _dro_schedules_result_set(context, result_set)


def dro_schedules_read_all_schedules(context, dbtxn=None):
    """Retrieves the DRO schedules for all aggregates."""
    query = _model_query(context, dbmodels.DROSchedulesDTO)
    return _dro_schedules_result_set(context, query.all(), group_by_agg=True)


def dro_schedules_read_aggregate_schedules(context, aggregate_id, dbtxn=None):
    """Retrieves the DRO schedules for the specified aggregate_id"""
    query = _model_query(context, dbmodels.DROSchedulesDTO)
    query = query.filter_by(aggregate_id=aggregate_id)
    return _dro_schedules_result_set(context, query.all())


def dro_schedules_update_aggregate_schedules(context, aggregate_id, values,
                                             dbtxn=None):
    """Replaces the aggregate's current DRO schedules or creates a new record
       if one does not currently exist"""
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # Create a Transaction around the update in the Database
    with session.begin(subtransactions=True):
        filters = dict(aggregate_id=aggregate_id)
        result_set = _find_all_db_entries(
            context, dbmodels.DROSchedulesDTO, filters, dbtxn=session)
        # If aggregate's DRO schedules already exist, update them
        if result_set:
            schedules = values.pop('dro_schedules')
            values['schedules_json'] = jsonutils.dumps(schedules)
            updates = _update_db_entry(context, dbmodels.DROSchedulesDTO,
                                       filters, values, session)
            return _dro_schedules_result_set(context, updates)
        # If aggregate's DRO schedules don't yet exist, create them
        else:
            values['aggregate_id'] = aggregate_id
            return dro_schedules_create_aggregate_schedules(context, values,
                                                            session)


def dro_schedules_delete_aggregate_schedules(context, aggregate_id,
                                             dbtxn=None):
    """Soft deletes the existing DRO schedules for the specified aggregate"""
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # Create a Transaction around the delete in the Database
    with session.begin(subtransactions=True):
        filters = dict(aggregate_id=aggregate_id)
        # Delete the existing entry in the specified Database table
        _delete_db_entry(context, dbmodels.DROSchedulesDTO, filters, session,
                         soft_delete=True)


##################################################
#            Database API Utilities              #
##################################################
def _model_query(context, model_class, *args, **kwargs):
    """ Wrappers the Base Model Query to provide PowerVC-specific logic """
    session, deleted = kwargs.get('session'), kwargs.get('read_deleted')
    session = session if session is not None else get_dbtransaction()
    deleted = context.read_deleted if deleted is None else deleted
    qkwargs = dict(deleted=False) if deleted == 'no' else dict()
    return sqlalchemyutils.model_query(model_class, session, None, **qkwargs)


def _find_all_db_entries(context, model_class, filters=None, dbtxn=None):
    """Queries the db entries from the database table specified"""
    query = _model_query(context, model_class, session=dbtxn)
    # If they provided a filter as input, add it to the query
    if filters:
        special_filters = filters.pop('special', [])
        query = query.filter_by(**filters)
        # Add any special filters we were given to the query
        for special_filter in special_filters:
            query = query.filter(special_filter)
    return _result_set_as_dict(query.all())


def _create_db_entry(context, model_class, values, dbtxn=None):
    """Creates a new entry in the database table specified"""
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # Create a Transaction around the create in the Database
    with session.begin(subtransactions=True):
        model_ref = model_class()
        model_ref.update(values)
        model_ref.save(session=session)
    # Convert the objects in the Result Set to Dictionaries to return
    return _result_set_as_dict(model_ref)


def _update_db_entry(context, model_class, filters, values, dbtxn=None):
    """Updates an existing entry in the database table specified"""
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # Create a Transaction around the update in the Database
    with session.begin(subtransactions=True):
        query = _model_query(context, model_class, session=session)
        # If they provided a filter as input, add it to the query
        if filters:
            special_filters = filters.pop('special', [])
            query = query.filter_by(**filters)
            # Add any special filters we were given to the query
            for special_filter in special_filters:
                query = query.filter(special_filter)
        model_ref = query.first()
        model_ref.update(values)
        model_ref.save(session=session)
    # Convert the objects in the Result Set to Dictionaries to return
    return _result_set_as_dict(model_ref)


def _delete_db_entry(context, model_class, filters,
                     dbtxn=None, soft_delete=False):
    """Removes an existing entry from the database table specified"""
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # Create a Transaction around the delete from the Database
    with session.begin(subtransactions=True):
        query = _model_query(context, model_class, session=session)
        # If they provided a filter as input, add it to the query
        if filters:
            special_filters = filters.pop('special', [])
            query = query.filter_by(**filters)
            # Add any special filters we were given to the query
            for special_filter in special_filters:
                query = query.filter(special_filter)
        # Either do a hard delete or soft delete depending on what they said
        if soft_delete is True:
            query.soft_delete(synchronize_session=False)
        else:
            query.delete(synchronize_session=False)


def _delete_db_entries(context, model_classes, filtersets,
                       dbtxn=None, soft_delete=False):
    """Removes existing entries from the database tables specified"""
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # Create a Transaction around the delete from the Database
    with session.begin(subtransactions=True):
        # Loop through each of the classes given, call delete on that class
        for idx in range(len(model_classes)):
            query = _model_query(context, model_classes[idx], session=session)
            query = query.filter_by(**filtersets[idx])
            # Either do a hard delete or soft delete depending on what they
            # said
            if soft_delete is True:
                query.soft_delete(synchronize_session=False)
            else:
                query.delete(synchronize_session=False)


def _get_primary_key_from_id(context, model_class, main_id, dbtxn=None):
    """Helper method to get the Primary Key ID from the Main ID"""
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # We need to query the Model Class based on the ID to know the Primary Key
    query = session.query(model_class._pk_id)
    pk_id = query.filter_by(id=main_id, deleted=0).first()
    return pk_id[0] if pk_id is not None else None


def _get_id_from_primary_key(context, model_class, pk_id, dbtxn=None):
    """Helper method to get the Main ID from the Primary Key ID"""
    session = get_dbtransaction() if dbtxn is None else dbtxn
    # We need to query the Model Class based on the Primary Key to know the ID
    query = session.query(model_class.id)
    main_id = query.filter_by(_pk_id=pk_id, deleted=0).first()
    return main_id[0] if main_id is not None else None


def _result_set_as_dict(result_set):
    """Utility to Convert the objects in the Result Set to Dictionaries"""
    # If this is a list, then convert each item in that result set list
    if isinstance(result_set, list):
        return [_result_set_as_dict(item) for item in result_set]
    return jsonutils.to_primitive(result_set)


def block_device_mapping_get_all_by_volume_id(context, volume_id,
                                              columns_to_join=None):
    if columns_to_join is None:
        columns_to_join = []

    query = _model_query(context, base_dbmodels.BlockDeviceMapping)

    for column in columns_to_join:
        query = query.options(joinedload(column))

    return query.filter_by(volume_id=volume_id).all()


###################################################
# #####  Override Service/ComputeNode Create  #####
###################################################
def _powervc_service_create(context, values):
    """Override the Service Create to do additional check"""
    if values.get('topic') == 'nova-compute':
        _verify_host_is_registered(values['host'])
    return _orig_service_create(context, values)


def _powervc_compute_node_create(context, values):
    """Override the Compute Node Create to do additional check"""
    _verify_host_is_registered(values['host'])
    return _orig_compute_node_create(context, values)


def _verify_host_is_registered(host):
    """Helper method to make sure the host is registered to allow create"""
    # Make sure that the host is really registered still and the remote
    # host didn't just not get un-managed and is connecting in as an orphan
    if not os.path.exists('/etc/nova/nova-%s.conf' % host):
        text = _("Host %s is no longer registered, won't create") % host
        LOG.info(text)
        raise Exception(text)

_orig_service_create = base_dbapi.service_create
_orig_compute_node_create = base_dbapi.compute_node_create
base_dbapi.service_create = _powervc_service_create
base_dbapi.compute_node_create = _powervc_compute_node_create


###################################################
# ######  Override Instance Get By Filters  #######
###################################################
def _powervc_instance_get_all_by_filters(*args, **kwargs):
    """Override the Instance Get to handle specifying vm_names/uuids"""
    # If the vm_names/vm_uuids is in the filters, then handle special
    if 'vm_names' in args[1] or 'vm_uuids' in args[1]:
        session = get_dbtransaction()
        query2 = session.query(base_dbmodels.Instance)
        query1 = session.query(dbmodels.InstancePowerSpecsDTO.instance_uuid)
        # If they asked to filter on the vm_names, query based on it
        if 'vm_names' in args[1]:
            query1 = query1.filter(dbmodels.InstancePowerSpecsDTO.
                                   vm_name.in_(args[1]['vm_names']))
        # If they asked to filter on the vm_uuids, query based on it
        if 'vm_uuids' in args[1]:
            query1 = query1.filter(dbmodels.InstancePowerSpecsDTO.
                                   vm_name.in_(args[1]['vm_uuids']))
        # Now we can add the instance uuid's to the query filter
        uuids = [row[0] for row in query1.all()]
        query2 = query2.filter(base_dbmodels.Instance.uuid.in_(uuids))
        # If they asked to filter on deleted, do so now
        if args[1].get('deleted'):
            query2 = query2.filter(
                base_dbmodels.Instance.deleted == base_dbmodels.Instance.id)
        else:
            query2 = query2.filter(base_dbmodels.Instance.deleted == 0)
        # Query the instances if we found any in the PowerSpecs table
        return query2.all() if uuids else list()
    # Otherwise fall back to just using the base implementation for it
    return _orig_instance_get_all_by_filters(*args, **kwargs)

_orig_instance_get_all_by_filters = base_dbapi.instance_get_all_by_filters
base_dbapi.instance_get_all_by_filters = _powervc_instance_get_all_by_filters
