# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from oslo_log import log as logging
from nova.network import neutronv2
from nova.objects import instance as instance_obj
from oslo_config import cfg
from powervc_nova import _


LOG = logging.getLogger(__name__)
CONF = cfg.CONF


def _get_unused_ports(context):
    """
    This task periodically runs to check if there are any 'stale' neutron
    ports that need cleanup or not. 'stale' ports are those which are not in
    use by any of the instance and hence should be freed up for deploys if they
    exit.  Note, this will only operate on instances and ports for this compute
    node since this periodic task is run in each compute process.

    :param context: The context object.
    """
    LOG.debug('pvc_nova.compute.manager.PowerVCComputeManager '
              '_get_unused_ports: Cleaning up unused neutron '
              'ports...')
    try:
        # Get all the non-deleted instances on this host.
        db_instances = \
            instance_obj.InstanceList.get_by_filters(context,
                                                     dict(deleted=False,
                                                          host=CONF.host))
        LOG.debug('Found instances: %s' % db_instances)

        # Get all the neutron ports for this host.
        search_opts = {'binding:host_id': CONF.host}
        network_data = neutronv2.api.get_client(context)\
            .list_ports(**search_opts)
        ports = network_data.get('ports', [])
        LOG.debug('Ports found: %s' % ports)

        port_ids_to_delete = _find_ports_for_cleanup(ports, db_instances)
        LOG.debug('Unused ports: %s' % port_ids_to_delete)
        return port_ids_to_delete
    except Exception as exc:
        LOG.exception(exc)
        return []


def _neutron_unused_port_cleanup(context, ports):
    """
    This task periodically runs to clean up identified'stale'
    ports which are not in use by any of the instance and hence
    should be freed up for deploys if they exit

    :param context: The context object.
    :param ports: Ports to be deleted.
    """
    # Now iterate the legit candidates for deletion and delete them.
    for port_id in ports:
        try:
            LOG.warning(_('Cleaning up the unused neutron port with id '
                          '%(port)s') % {'port': port_id})
            neutronv2.api.get_client(context).delete_port(port_id)
        except Exception as exc:
            LOG.exception(exc)
    LOG.debug('Exiting neutron port clean up')


def _find_ports_for_cleanup(ports, instances):
    ports_ids_to_delete = []
    try:
        # Make a list of the instance UUIDs so we have a quick comparison below
        instance_uuids = [instance['uuid'] for instance in instances]
        LOG.debug('Instance UUIDs: %s' % instance_uuids)

        for port in ports:
            dev_id = port.get('device_id', None)
            dev_owner = port.get('device_owner', None)

            # Only delete ports owned by compute (we don't want to delete ports
            # owned by SCE, for example) but NOT in use by an instance
            if dev_owner is not None and dev_owner == 'compute:None' and\
                    dev_id is not None and dev_id not in instance_uuids:
                ports_ids_to_delete.append(port['id'])
                LOG.info(_('Deleting neutron port with id %(id)s and data '
                           '%(data)s') % {'id': port['id'], 'data': str(port)})
            else:
                LOG.debug('Skipping port %s' % port)

    except Exception as exc:
        LOG.exception(exc)

    return ports_ids_to_delete
