# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
import copy

from nova.network.neutronv2 import api as netapi
from oslo_log import log as logging

from oslo_config import cfg

from powervc_nova.virt.ibmpowervm.vif.common import ras
from powervc_nova.virt.ibmpowervm.vif.common import exception as powexc
from powervc_nova.network.ibmpowervm import adapter_mapping as mapping_task
from powervc_nova.network.neutronv2 import ibmpower_api
from powervc_nova.objects.network import dom as network_dom

from powervc_nova import _

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class PowerVMAPI(ibmpower_api.PowerBaseAPI):

    """
    API wrapper for using POWER VM Neutron plugin.  Must be paired with the
    POWER VM Neutron Plugin.  To utilize, update the nova.conf with the
    following.
        - network_api_class = nova.network.neutronv2.ibmpowervm_api.PowerVMAPI
    """

    def __init__(self, skip_policy_check=False):
        netapi.API.__init__(self, skip_policy_check=skip_policy_check)
        self.adapter_mapping_task = mapping_task.AdapterMappingTaskClass()

        self.hostfact = network_dom.ManagedHostFactory.get_factory()
        self.netafact = network_dom.NetworkAssociationFactory.get_factory()

    def allocate_for_instance(self, context, instance, **kwargs):
        LOG.debug('Enter allocate_for_instance with: %s' % kwargs)
        requested_networks = kwargs.get('requested_networks')

        # NOTE: valid slot id values range from 32 to 63
        slot_id_list = range(32, 64)

        # track all mac addresses associated with port id
        # specified networks
        port_id_mac_list = []

        if requested_networks:
            for req_net in requested_networks:
                if req_net.port_id:
                    LOG.debug("port id = %s" % req_net.port_id)

                    # add the port mac address to the mac list
                    port = self.show_port(context,
                                          req_net.port_id)['port']
                    if port:
                        port_mac = port.get('mac_address', None)
                        if port_mac:
                            port_id_mac_list.append(port_mac)
                elif not req_net.network_id:
                    LOG.error(_("A network entry was found that is missing "
                                "both the port id and the network "
                                "id.  This network specification "
                                "will not be included in the "
                                "list of networks available for "
                                "allocation."))

        # set updated mac list
        if 'macs' in kwargs.keys() and kwargs['macs']:
            kwargs['macs'] = kwargs['macs'].union(port_id_mac_list)

        vifs = super(PowerVMAPI, self).allocate_for_instance(context,
                                                             instance,
                                                             **kwargs)

        # first we need to remove from the list of available slot ids
        # all those ids that will be used by the network id defined
        # networks (these use the last 2 digits of the mac address
        # to calculate the slot id to use)
        for vif in vifs:
            slot_id = int(vif['address'][-2:], 16)
            if vif['address'] not in port_id_mac_list and\
                    (slot_id in slot_id_list):
                slot_id_list.remove(slot_id)

        # now that those are gone, we can assign a slot id to the
        # port id defined networks knowing that we won't assign a
        # slot id that will be used by a network id defined network
        # specification
        for vif in vifs:
            if vif['address'] in port_id_mac_list:
                # get the first value from our slot id list
                vif['slot_id'] = slot_id_list.pop(0)

        return vifs

    def _build_network_info_model(self, context, instance, networks=None,
                                  port_ids=None, admin_client=None,
                                  preexisting_port_ids=None):
        LOG.debug('Enter _build_network_info_model with instance %s and '
                  'networks %s' % (instance['project_id'], networks))

        # Get all the nova networks associated with the neutron network passed
        # in.  If no neutron network was passed in, this will return all nova
        # networks.
        net_infos = super(PowerVMAPI, self)._build_network_info_model(
            context, instance, networks, port_ids,
            admin_client, preexisting_port_ids)

        LOG.debug('Found net_infos %s' % net_infos)

        # Get the neutron networks related to our instance if the invoker did
        # not pass them in.
        if not networks:
            networks = self._get_available_networks(context,
                                                    instance['project_id'])
            msg = (ras.vif_get_msg
                   ('info', 'NO_NET_FOUND') % {'host': instance['host']})
            LOG.info(msg)

        # Pare down the nova networks list to only include those that map to
        # the neutron network we're interested in.  This will also add in the
        # vlan id to the nova network object.
        return self._match_network_models(networks, net_infos, context)

    def _match_network_models(self, networks, net_infos, context):
        """
        Will match the neutron network model into the Nova network model and
        will embed the vSwitch and VLAN ID into the bridge name.

        :param networks: A set of Neutron Networks (the values returned from
                         the REST API)
        :param net_infos: The NOVA Network objects used by the compute layer.
                          These objects will be updated to contain the
                          provider:segmentation_id into the bridge attribute
                          in this object
        :param context: This is the context object to access the DB.
        :returns: The original net_infos, however updated with the bridge
                  which will be vSwitch/VLAN_ID
        """
        LOG.debug(('Enter _match_network_models with networks:\n%s\n\n'
                   'net_infos:\n%s' % (networks, net_infos)))
        # For each element, add in the proper VLAN ID into the bridge
        for network in networks:
            nets = self._find_net_infos_for_network(net_infos, network)
            if nets is None:
                LOG.debug('Found no net_info for network %s' % network)
                continue

            # Get the VLAN off the network.  If one doesn't exist, default
            # to a value of 1.
            if network.get('provider:segmentation_id') is not None:
                vlanid = int(network['provider:segmentation_id'])
                LOG.debug(ras.vif_get_msg('info', 'NET_VLAND_ID') %
                          {'vlan': vlanid})
            else:
                # Since there was no vlan id - setting it to 1
                vlanid = 1
                LOG.info(ras.vif_get_msg('info', 'NO_NET_VLAN'))

            # Set the data for this network into its corresponding net_info
            for net_info in nets:
                net_info['vlan'] = ("%(vlan)s" % {"vlan": vlanid})

        LOG.debug('Exiting _match_network_models with net_infos %s' %
                  net_infos)
        return net_infos

    def build_default_network_assn(self, context, host_dom, vlanid, net_id,
                                   vswitch=None):
        """
        Invokes the default network association.  Mostly pass thru, however
        if a SEA can not be found will return None.
        """
        try:
            adpt_mapper = self.adapter_mapping_task
            return adpt_mapper.build_default_network_assn(context, host_dom,
                                                          vlanid, net_id,
                                                          vswitch)
        except powexc.IBMPowerVMValidSEANotFound:
            LOG.info(_("Unable to build default network association.  No valid"
                       " Shared Ethernet Adapters for VLAN ID %(vid)s" %
                       {'vid': str(vlanid)}))
            return None

    def _get_viable_hosts(self, context, instance_uuid, hosts_from_db,
                          network_uuids, source_host=None):
        """
        Find the viable hosts that match the network requirements.

        :param context: A context object used to authorize the DB access.
        :param instance_uuid: The instance UUID that will be used to get the
                              network requirements of an instance.
        :param hosts_from_db: A list of compute nodes
        :param network_uuids: An list of requested networks for the deployment.
        :param source_host: When this API is invoked for LPM, the host that
                            is being migrated from.  This is needed to check
                            vswitch config on the source host.
        :returns: A dictionary of hosts with the hostid and the hostname
        """
        candidate_hosts = copy.copy(hosts_from_db)
        LOG.debug('instance uuid: %s\ncandidate hosts: %s\nnetwork uuids: %s\n'
                  'source_host: %s' %
                  (instance_uuid, candidate_hosts, network_uuids, source_host))

        # Log if we're bound to find nothing because of input
        if not hosts_from_db:
            LOG.warn(_('No hosts passed in'))

        # Do some lookups outside the loop so we don't make multiple calls.
        # First the host doms and the contained sea doms
        host_doms, sea_doms = {}, {}
        for host in self.hostfact.find_all_hosts(context, load_all=True):
            host_doms[host.host_name] = host
            for sea in host.seas:
                sea_doms[sea.id] = sea

        # Now get the network associations
        net_assns = {}
        for net_assn in self.netafact.find_all_network_assocs(context):
            net_assn._host = host_doms.get(net_assn.host_name)
            net_assn._sea = sea_doms.get(net_assn.sea_id)
            net_assns.setdefault(net_assn.host_name, list())
            net_assns[net_assn.host_name].append(net_assn)

        # Determine which vswitches are needed for each network on source_host
        required_vswitches = {}
        if source_host:
            netasc_list = net_assns.get(source_host, [])
            for netasc in netasc_list:
                for net_uuid in network_uuids:
                    if netasc.neutron_net_id == net_uuid and\
                            netasc.sea is not None:
                        required_vswitches[net_uuid] = \
                            netasc.sea.primary_vea.vswitch_name

        # Check if each host contains a valid network
        for host in hosts_from_db:
            # Keep track of the networks that remaining to be processed.
            networks_to_process = copy.copy(network_uuids)
            if not networks_to_process:
                continue

            # Get the DOM so that we can verify the state of the chains.
            host_dom = host_doms[host]

            # Walk through each of the network associations that exist for
            # the host and see if it was set to be used.  A 'use this host'
            # indication is if the network associations SEA is not None
            for net_assn in net_assns.get(host, []):
                if net_assn.neutron_net_id in network_uuids and\
                        net_assn.neutron_net_id in networks_to_process:
                    # No longer a network to process
                    networks_to_process.remove(net_assn.neutron_net_id)

                    # Check the SEA chain.  At least one SEA in the chain
                    # should be available.  If there are no SEAs, then remove
                    # it as a candidate
                    at_least_one_available = False
                    if net_assn.sea:
                        s_nm = net_assn.sea.name
                        s_lp = net_assn.sea.vioserver.lpar_id
                        sea_chain = host_dom.find_sea_chain_for_sea_info(s_nm,
                                                                         s_lp)
                        for sea in sea_chain:
                            if sea and sea.is_available():
                                at_least_one_available = True
                                break

                    # Check the vswitches.  The same vswitch name must be used
                    # on the source and target host.
                    vswitch_matches = True
                    if net_assn.neutron_net_id in required_vswitches and\
                            net_assn.sea is not None:
                        vswitch_name = net_assn.sea.primary_vea.vswitch_name
                        if vswitch_name == "ETHERNET0(Default)":
                            vswitch_name = vswitch_name.strip('(Default)')
                        netid = net_assn.neutron_net_id
                        if (required_vswitches[netid] ==
                                "ETHERNET0(Default)"):
                                    required_vswitches[netid] = "ETHERNET0"
                        if (vswitch_name !=
                                required_vswitches[net_assn.neutron_net_id]):
                                vswitch_matches = False
                        # if(net_assn.sea.primary_vea.vswitch_name !=
                        #   required_vswitches[net_assn.neutron_net_id]):
                        #    vswitch_matches = False

                    # Check the conditions.  If there isn't one available,
                    # yet the host is still here...remove that host as a
                    # candidate.
                    if((not vswitch_matches or not at_least_one_available) and
                       host in candidate_hosts):
                        LOG.info('Removing candidate host %s because %s' %
                                 (host, 'vswitch does not match' if not
                                  vswitch_matches else 'sea not available'))
                        candidate_hosts.remove(host)

            # Skip the next, computationally expensive, step if there are no
            # more networks to process.
            if len(networks_to_process) == 0:
                continue

            # There may be some networks that haven't been processed yet.
            # This is typically for when a host was added after the network
            # was created.  In this scenario, we should call down into
            # the adapter mapping task and find the default.  The default
            # may be do not use.
            #
            # We don't do this by default because it's far more expensive
            # than a database operation, and we want the standard flow to be
            # as fast as possible.
            temp_context = context.elevated()
            for net_id in networks_to_process:
                net_info = self.get(temp_context, net_id)
                if not net_info or\
                        net_info.get('segmentation_id') is None:
                    continue
                vlanid = int(net_info.get('segmentation_id'))
                sea = self.build_default_network_assn(temp_context,
                                                      host_dom,
                                                      vlanid,
                                                      net_id)
                # If the SEA that came back from this is None, we need to
                # remove it as a candidate.  We can be confident at this
                # point though that the network association was created,
                # thus speeding up our flow through next time.
                if sea is None or not sea.is_available():
                    LOG.info(_('Removing host %(host)s due to invalid sea '
                               '%(sea)s'), {'host': host, 'sea': sea})
                    if host in candidate_hosts:
                        candidate_hosts.remove(host)

        return candidate_hosts
