#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
from nova import exception as novaexception
from oslo_config import cfg
from oslo_log import log as logging

from powervc_nova.common import aggregate_utils as agg_utils

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class IBMPowerVMBaseNetworkPlacement(object):
    """
    Serves as a base class for host-agnostic placement/deploy data.
    """

    def get_placement(self, context, host_name, network_id, agg, list_only):
        """
        Get the placement dictionary, optionally constrained by a host or
        network.
        :param context: HTTP request context.
        :param host_name: Nova name of a host to optionally constrain
                          placement.
        :param network_id: Neutron network ID to optionally constrain
                           placement.
        :param agg: The host aggregate ID, if specified on the request.
        :param list_only: If true, only a list of IDs will be returned.
        :returns: Dict of placement data.
        """

        # Validate that host aggregate exists
        if agg:
            host_aggs = agg_utils.get_aggregate_list(context)
            host_agg_ids = [str(x.id) for x in host_aggs]
            if not str(agg) in host_agg_ids:
                raise novaexception.AggregateNotFound(aggregate_id=str(agg))

        # Handle the case where a host_name was provided - return network ids
        if host_name and not network_id and not agg:
            id_list = self._get_host_placement_for_host(context, host_name)
            if list_only == 'true':
                placement_dict = {'host-network-placement': id_list}
            else:
                host_dict = {}
                host_dict['host_name'] = host_name
                host_dict['networks'] = []
                for network_id in id_list:
                    host_dict['networks'].append({'network_id': network_id})
                placement_dict = {'host-network-placement':
                                  [host_dict]}

        # Handle the case where a network was provided - return host names
        elif network_id and not host_name and not agg:
            id_list = self._get_host_placement_for_network(context, network_id)

            if list_only == 'true':
                placement_dict = {'host-network-placement': id_list}
            else:
                placement_dict = {'host-network-placement': []}
                for host_id in id_list:
                    host_dict = {}
                    host_dict['host_name'] = host_id
                    host_dict['networks'] = []
                    host_dict['networks'].append({'network_id': network_id})
                    placement_dict['host-network-placement'].append(host_dict)

        # Handle the case where a host aggregate was provided - return networks
        elif agg and not host_name and not network_id:
            # Get the list of hosts in the specified host aggregate
            host_list = []
            try:
                host_list = agg_utils.get_hosts_for_aggregate(context, agg)
            except Exception as e:
                LOG.exception(e)

            # Get the host-network-placement for each host in the aggregate
            hnp_dict = {}

            # If there's only one host in the agg, call placement for that
            # host.  Otherwise, it's probably faster to make a single call to
            # get all hosts and filter the result.
            if len(host_list) == 1:
                net_id_list = self._get_host_placement_for_host(
                    context, host_list[0])
                hnp_dict[host_list[0]] = net_id_list
            elif len(host_list) > 1:
                host_dicts = self._get_host_placement_all(context)[
                    'host-network-placement']
                for host_dict in host_dicts:
                    if host_dict['host_name'] in host_list:
                        hnp_dict[host_dict['host_name']] = [
                            net['network_id'] for net in host_dict['networks']]

            # Find the union of available networks for the hosts in the
            # specified aggregate.
            placement_dict = self._hnp_union(hnp_dict, agg, list_only)

        # Handle the case where nothing was provided - return all data
        elif not host_name and not network_id and not agg:
            placement_dict = self._get_host_placement_all(context)

        return placement_dict

    def _hnp_union(self, hnp_dict, agg, list_only):
        """
        Accepts a list of host-network-placement dictionaries, and takes the
        union of networks between the hosts.  In other words, a networks is
        returned if that network is connected to any host in the list.

        :param hnp_dict: A dictionary with key/value pairs that are:
                         -key: the host name from nova.
                         -value: a list of network IDs valid for this host.
        :param agg: The identifier for the host aggregate.
        :returns hnp_dict: A dictionary that represents the union of networks
                           across the specified set of hosts.
        """
        # Compute union of networks across all hosts
        net_set = set()
        for host in hnp_dict.keys():
            for net_id in hnp_dict[host]:
                net_set.add(net_id)

        # Build return dictionary
        if list_only:
            # Use short form of return dictionary
            return {'host-network-placement': list(net_set)}
        else:
            # Use long form of return dictionary
            net_dict_list = []
            for net_id in net_set:
                net_dict_list.append({'network_id': net_id})
            return {
                'host-network-placement': [
                    {
                        'host_aggregate': str(agg),
                        'networks': net_dict_list
                    }
                ]
            }

    def _get_host_placement_all(self, context):
        """
        Implemented by child classes.
        :param context: HTTP request context.
        :return: dict of host-network-placement data.
        """
        raise NotImplementedError('_get_host_placement_all not implemented')

    def _get_host_placement_for_host(self, context, host_name):
        """
        Implemented by child classes.
        :param context: HTTP request context.
        :param host_name: Nova name of a host to constrain placement
        :return: list of network ids.
        """
        raise NotImplementedError('_get_host_placement_for_host not '
                                  'implemented')

    def _get_host_placement_for_network(self, context, network_id):
        """
        Implemented by child classes.
        :param context: HTTP request context.
        :param network_id: Neutron network ID to constrain placement.
        :return: list of host names.
        """
        raise NotImplementedError('_get_host_placement_for_network not '
                                  'implemented')
