#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================


class MicroOperation():
    """
    A MicroOperation or "micro op" is an atomic operation that can be run
    for configuring networking.  Examples include creating a vswitch port or
    migrating an IP address.

    MicroOperations are strung together by the micro op builder to resolve
    a request from a user to change the state of the system.

    A Micro Operation consists of three states.
    1) Validate - takes in the state of the system, determines if the operation
                  can be run.  If it can not an exception is thrown.  If it
                  has warnings, the warnings are returned.  The state of the
                  system, as if the MicroOperation had been run, is also
                  returned so that the next MicroOperation (as defined by
                  the micro op builder) can run with that input.
    2) Execute - Perform the micro operation.  Should be as minimal as
                 possible.  If an issue is encountered, throw an error.
    3) Undo - If a previous action had an issue, the undo should roll back the
              state of the system as if it had been run.
    """

    def validate(self, current_dom):
        """
        This method should be overridden in all the micro operations.

        The validate method should assume that the current_dom passed in is the
        current state of the system.  It then looks at the appropriate parts
        of the DOM to determine if the Micro Operation can be run.

        If there are errors, they should be thrown as an exception.

        The validate method should return the updated state of the DOM (as
        if the operation had actually been run) as well as a list of warnings
        generated by the MicroOperation.

        :param current_dom: This is the current DOM representation of the
                            system before the micro op is performed.  The type
                            of this object is HostOVSNetworkConfig.

        :returns updated_dom: An updated version of the current_dom that
                              has been updated 'as if' the operation had been
                              executed.
        :returns warning_list: A list of warnings (but not errors) that were
                               generated based off the state of the DOM and
                               the operation that had been run.
        """
        raise NotImplementedError('This method implemented by child classes.')

    def execute(self):
        """
        Performs the Micro Operation against the system.  Should be interacting
        directly against the system.

        Validation should be assumed to have been done beforehand.

        If an error was encountered, since the error may have occurred mid
        operation, THIS method should roll back whatever changes may have
        occurred.  It should then throw an appropriate exception.
        """
        raise NotImplementedError('This method implemented by child classes.')

    def undo(self):
        """
        Performs an undo of the Micro Operation.  Assumes that the execute
        completed fully.

        Should completely reverse the operation from the finished execute state
        to the state prior to the action being run.

        Should NOT throw any errors.
        """
        raise NotImplementedError('This method implemented by child classes.')
