#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
from powervc_nova.network.powerkvm.agent.common import micro_op
from powervc_nova.network.powerkvm.agent.common import exception
from powervc_nova.objects.network import dom_kvm
from powervc_nova.network.powerkvm.agent import multi_op
from powervc_nova.network.powerkvm.agent import ifcfg_builder
from powervc_nova.network.powerkvm.agent import commandlet
from nova import utils
from oslo_log import log as logging
from powervc_nova.network.powerkvm.agent.ovs_validation_mixin \
    import OVSValidator
from powervc_nova import logcall

LOG = logging.getLogger(__name__)
MAX_VSWITCH_NAME_LENGTH = 8


class CreatevSwitch(micro_op.MicroOperation,
                    OVSValidator):
    """
    This micro-op creates a Open vSwitch
    """

    def __init__(self, ovs_name):
        """
        Constructor

        :param ovs_name: The name of the Open vSwitch
        """
        self.ovs_name = ovs_name
        self.current_dom = None
        # Each operation can spawn child operations.
        self.sub_micro_ops = multi_op.MultiMicroOp()

    @logcall
    def validate(self, curr_dom):
        """
        Performs a set of validations and throws errors accordingly
        :param current_dom: This is the current DOM representation of the
                            system before the micro op is performed.
        :returns updated_dom: An updated version of the current_dom that
                              has been updated 'as if' the operation had been
                              executed.
        :returns warning_list: A list of warnings (but not errors) that were
                               generated based off the state of the DOM and
                               the operation that had been run.
        """
        self.current_dom = curr_dom
        # If the DOM already has a vSwitch with the same name, throw an ERROR
        if self.current_dom.find_ovs_obj(self.ovs_name) is not None:
            raise\
                exception.IBMPowerKVMOVSExistsError(switch_name=self.ovs_name)
        # If the vswitch name is present as any port name - throw an error
        if self.current_dom.find_parent_name(self.ovs_name):
            raise\
                exception.IBMPowerKVMOVSNameInUse(switch_name=self.ovs_name)
        # If the vswitch name is empty - throw an error
        if len(self.ovs_name) == 0:
            raise exception.IBMPowerKVMPortNameEmpty()
        # If the name of the vswitch is more than 8 characters, throw an ERROR
        if len(self.ovs_name) > MAX_VSWITCH_NAME_LENGTH:
            raise exception.IBMPowerKVMPortNameExceedsLimit(
                switch_name=self.ovs_name, chars=str(MAX_VSWITCH_NAME_LENGTH))

        # If the name of the vswitch contains invalid (?) characters,
        # throw an ERROR
        self.validate_ovs_name(self.ovs_name)

        # If any other adapter in the DOM exists with the same name,
        # throw an ERROR (OVS Ports included as well)
        if self.current_dom.contains_port_by_name(self.ovs_name):
            raise exception.IBMPowerKVMOVSAdapterNotAvailableError(
                adapter_name=self.ovs_name)

        # Add itself to the DOM that is passed in and do a return
        self.current_dom.ovs_list.append(dom_kvm.OpenVSwitch(self.ovs_name))
        return self.current_dom, []

    @logcall
    def execute(self):
        """
        Overrides parent method.

        Execute command line call to create a OVS and add a sub micro-op
        """
        # run 'ovs-vsctl add-br <bridge>'
        try:
            stdout, stderr = utils.execute('ovs-vsctl', 'add-br',
                                           self.ovs_name, run_as_root=True)
            if stderr and stderr != '':
                raise exception.IBMPowerKVMCommandExecError(cmd='ovs-vsctl',
                                                            exp=stderr)
            # Add a sub micro operation of IfcfgFileOVSBuilder
            op = ifcfg_builder.IfcfgFileOVSBuilder(self.ovs_name)
            self.sub_micro_ops.micro_ops.append(op)
            self.sub_micro_ops.execute()
        except Exception as e:
            LOG.exception(e)
            raise
        LOG.debug('ovs-vsctl add-br output = %s' % stdout)

    @logcall
    def undo(self):
        """
        Overrides parent method.

        Execute command line call to remove an OVS and related ifcfg file.
        """
        # run 'ovs-vsctl del-br <bridge>'
        try:
            stdout, stderr = utils.execute('ovs-vsctl', 'del-br',
                                           self.ovs_name, run_as_root=True)
            if stderr and stderr != '':
                raise exception.IBMPowerKVMCommandExecError(cmd='ovs-vsctl',
                                                            exp=stderr)
            # Call the sub micro operation
            self.sub_micro_ops.undo()
            # Should remove the ifcfg file that was created for the adapter.
            # Get contents and location of current ifcfg file (if it exists!)
            fnamepath, self.source_ovs_restore = \
                commandlet.CommandExecutor.get_ifcfg(self.ovs_name)
            # Delete ifcfg file
            if fnamepath and len(fnamepath):
                LOG.debug('Removing ifcfg: %s', fnamepath)
                commandlet.remove_file_as_root(fnamepath)
        except Exception as e:
            LOG.exception(e)
            raise
        LOG.debug('ovs-vsctl del-br output = %s' % stdout)
