#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import uuid
import ConfigParser

from oslo_serialization import jsonutils
from nova.db import api as base_db_api
from nova.objects import base
from nova.objects import fields
from nova.objects import flavor
from nova.objects import instance_info_cache
from oslo_log import log as logging

from powervc_nova.db import api as db_api
import powervc_nova.objects.dom as dom

LOG = logging.getLogger(__name__)


######################################################
#        Management Console DOM Implementation       #
######################################################
@base.NovaObjectRegistry.register
class ManagementConsoleFactory(dom.ResourceFactory):

    """Production DOM Resource Factory class implementation for HMC's"""

    ###############################################
    #       Defining HMC Factory Interface        #
    ###############################################
    def construct_hmc(self, context, hmc_uuid, mtmsn, access_ip,
                      user_id, password, hmc_display_name,
                      auto_create=False, transaction=None):
        """Constructs (and optionally creates) the HMC specified"""
        kwargs = dict(hmc_uuid=hmc_uuid, mtmsn=mtmsn, access_ip=access_ip,
                      user_id=user_id, password=password,
                      hmc_display_name=hmc_display_name)
        # Call the Parent Class to actually Construct the given Resource
        return self.construct_resource(
            ManagementConsole, context, auto_create, transaction, **kwargs)

    def find_all_hmcs(self, context, filters=None, transaction=None):
        """Queries at most one HMC matching the HMC UUID specified"""
        return self.find_all_resources(
            ManagementConsole, context, filters, transaction)

    def find_all_hmcs_by_host(self, context, host_name, transaction=None):
        """Queries all HMC's that are managing the Host specified"""
        hmcs = db_api.hmc_find_all(context, dict(host_name=host_name))
        return self._construct_resources(
            ManagementConsole, context, hmcs, transaction)

    def find_hmc_by_uuid(self, context, hmc_uuid, transaction=None):
        """Queries at most one HMC matching the HMC UUID specified"""
        return self.find_resource(ManagementConsole, context,
                                  dict(hmc_uuid=hmc_uuid), transaction)

    def find_hmc_by_mtmsn(self, context, mtmsn, transaction=None):
        """Queries at most one HMC matching the HMC MTMSN (machine type, model,
        and serial number) specified"""
        return self.find_resource(ManagementConsole, context,
                                  dict(mtmsn=mtmsn), transaction)

    def delete_hmc(self, context, hmc_uuid, transaction=None):
        """Deletes the HMC with the HMC UUID specified"""
        return self.delete_resource(
            ManagementConsole, context, hmc_uuid, transaction)

    ###############################################
    #   Implementing Resource Factory Definition  #
    ###############################################
    def _get_resource_id_attr_name(self):
        """Provides the Name of the Attribute for the Resource Identifier"""
        return 'hmc_uuid'

    def _get_resource_classes(self):
        """Returns a List of DOM Classes or a Dictionary Mapping DOM Classes"""
        return [ManagementConsole]

    @base.remotable
    def _find_all_resources(self, resource_class,
                            filters=None, transaction=None):
        """Factory implementation to query all resources matching the filter"""
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        hmcs = db_api.hmc_find_all(self._context, filters, dbtxn)
        return self._construct_resources(
            resource_class, self._context, hmcs, transaction)

    @base.remotable
    def _delete_multiple_resources(self, resource_class,
                                   resource_ids, transaction=None):
        """Factory implementation to delete the resources specified"""
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        for resource_id in resource_ids:
            db_api.hmc_delete(self._context, resource_id, dbtxn)


@base.NovaObjectRegistry.register
class ManagementConsole(dom.Resource):

    """Production DOM Resource class implementation for HMC's"""

    fields = dict(dom.Resource.fields,
                  hmc_uuid=fields.StringField(),
                  mtmsn=fields.StringField(nullable=True),
                  hmc_display_name=fields.StringField(),
                  access_ip=fields.StringField(),
                  user_id=fields.StringField(),
                  password=fields.StringField(),
                  registered_at=fields.DateTimeField(nullable=True),
                  managed_hosts=fields.ListOfStringsField(nullable=True))

    def __init__(self, context=None,
                 factory_class=None, transaction=None, **kwargs):
        "Constructs an instance of the given DOM Resource Class"
        super(ManagementConsole, self).__init__(
            context, factory_class, transaction, **kwargs)

    ###############################################
    #     Implementing DOM Resource Definition    #
    ###############################################
    def _initialize_resource_attributes(self):
        """Sets any default values that are defined for any attributes"""
        super(ManagementConsole, self)._initialize_resource_attributes()
        setattr(self, 'managed_hosts', list())

    @base.remotable
    def _create_resource(self, transaction):
        """Actual Implementation that will Create the given Resource"""
        values = self._get_resource_values()
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        hmc = db_api.hmc_create(self._context, values, dbtxn)
        self._hydrate_all_resource_attributes(hmc)

    @base.remotable
    def _update_resource(self, transaction):
        """Actual Implementation that will Update the given Resource"""
        hmc_uuid = self._get_resource_identifer()
        values = self._get_resource_values(changes_only=True)
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        hmc = db_api.hmc_update(self._context, hmc_uuid, values, dbtxn)
        self._hydrate_all_resource_attributes(hmc)


######################################################
#           Managed Host DOM Implementation          #
######################################################
@base.NovaObjectRegistry.register
class ManagedHostFactory(dom.ResourceFactory):

    """Production DOM Resource Factory class implementation for Hosts"""

    ###############################################
    #       Defining Host Factory Interface       #
    ###############################################
    def find_all_hosts(self, context, filters=None, transaction=None):
        """Queries all Host that are registered on the Management System"""
        return self.find_all_resources(
            ManagedHost, context, filters, transaction)

    def find_all_hosts_by_hmc(self, context, hmc_uuid, transaction=None):
        """Queries all Host that are registered on the Management System"""
        filters = dict(hmc_uuid=hmc_uuid)
        return self.find_all_resources(
            ManagedHost, context, filters, transaction)

    def find_host_by_name(self, context, host_name, transaction=None):
        """Queries at most one Host matching the Host Name specified"""
        return self.find_resource(ManagedHost, context,
                                  dict(host_name=host_name), transaction)

    def delete_host(self, context, host_name, transaction=None):
        """Deletes the Host with the Host Name specified"""
        return self.delete_resource(
            ManagedHost, context, host_name, transaction)

    ###############################################
    #    Defining Host Factory Abstract Methods   #
    ###############################################
    def _get_vios_factory(self):
        """Provides the Factory to use for the VIO Server Implementation"""
        return VioServerFactory.get_factory()

    ###############################################
    #   Implementing Resource Factory Definition  #
    ###############################################
    def _get_resource_id_attr_name(self):
        """Provides the Name of the Attribute for the Resource Identifier"""
        return 'host_name'

    def _get_resource_classes(self):
        """Returns a List of DOM Classes or a Dictionary Mapping DOM Classes"""
        return [ManagedHost]

    @base.remotable
    def _find_all_resources(self, resource_class,
                            filters=None, transaction=None):
        """Factory implementation to query all resources matching the filter"""
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        hosts = db_api.host_find_all(self._context, filters, dbtxn)
        return self._construct_resources(
            resource_class, self._context, hosts, transaction)

    @base.remotable
    def _delete_multiple_resources(self, resource_class,
                                   resource_ids, transaction=None):
        """Factory implementation to delete the resources specified"""
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        for resource_id in resource_ids:
            db_api.host_delete(self._context, resource_id, dbtxn)


@base.NovaObjectRegistry.register
class ManagedHost(dom.Resource):

    """Production DOM Resource class implementation for Hosts"""

    fields = dict(dom.Resource.fields,
                  host_name=fields.StringField(),
                  hypervisor_hostname=fields.StringField(nullable=True),
                  hypervisor_type=fields.StringField(nullable=True),
                  hypervisor_version=fields.IntegerField(nullable=True),
                  hmc_uuids=fields.ListOfStringsField(nullable=True),
                  registered_at=fields.DateTimeField(nullable=True))

    obj_extra_fields = ['config_values', 'vioservers', 'hmcs',
                        'enterprise_pools', 'shared_processor_pools',
                        'host_type', 'host_display_name', 'host_storage_type',
                        'host_uniqueid', 'host_ephemeral_disk_path',
                        'host_network_server', 'host_network_share']
    _supported_host_types = None

    def __init__(self, context=None,
                 factory_class=None, transaction=None, **kwargs):
        "Constructs an instance of the given DOM Resource Class"
        super(ManagedHost, self).__init__(
            context, factory_class, transaction, **kwargs)

    ###############################################
    #     Implementing DOM Resource Definition    #
    ###############################################
    def _initialize_resource_attributes(self):
        """Sets any default values that are defined for any attributes"""
        super(ManagedHost, self)._initialize_resource_attributes()
        self._vioservers, self._hmcs, self._enterprise_pools = (None, None,
                                                                None)
        self._shared_processor_pools, self._config_values = None, None

    def _reset_cached_attributes(self):
        """Wipes out any Cached Resource Attributes that are no longer valid"""
        super(ManagedHost, self)._reset_cached_attributes()
        self._vioservers, self._hmcs, self._enterprise_pools = (None, None,
                                                                None)
        self._shared_processor_pools, self._config_values = None, None

    ###############################################
    #       Defining Managed Host Interface       #
    ###############################################
    @staticmethod
    def get_supported_host_types():
        """Helper method to determine what types of Hosts are supported"""
        # For performance we will only retrieve the Host Type once
        if ManagedHost._supported_host_types is None:
            host_type = None
            parser = ConfigParser.RawConfigParser()
            parser.read('/etc/nova/nova.conf')
            # Use the PowerVM-specific property for now
            if parser.has_option('DEFAULT', 'powervm_mgr_type'):
                host_type = parser.get('DEFAULT', 'powervm_mgr_type')
            # Default to IVM for now if the property is not set
            ManagedHost._supported_host_types =\
                [host_type] if host_type else ['ivm']
        return ManagedHost._supported_host_types

    @property
    def config_values(self):
        """Helper method to retrieve the Compute Node for a given Host Name"""
        if self._config_values is None:
            self._config_values = dict()
            attrs = ['host_type', 'host_display_name', 'host_storage_type',
                     'host_uniqueid', 'kvm_host_ephemeral_disk_path',
                     'kvm_host_network_server', 'kvm_host_network_share']
            try:
                # Query the Nova CONF file for the Host to get the attributes
                parser = ConfigParser.RawConfigParser()
                parser.read('/etc/nova/nova-' + self.host_name + '.conf')
                for attr in attrs:
                    name = attr[4:] if attr.startswith('kvm_') else attr
                    if parser.has_option('DEFAULT', attr):
                        self._config_values[name] = parser.get('DEFAULT', attr)
            # If we got an exception retrieving the Host, it wasn't found
            except Exception:
                pass
        return self._config_values

    @property
    def vios_factory(self):
        """Return the VioServerFactory that is associated to this object."""
        return self._get_resource_factory()._get_vios_factory()

    @property
    def host_type(self):
        """Getter Method for the Host Type for this Host"""
        return self.config_values.get('host_type')

    @property
    def host_display_name(self):
        """Getter Method for the Display Name for this Host"""
        return self.config_values.get('host_display_name')

    @property
    def host_storage_type(self):
        """Getter Method for the Storage Type for this Host"""
        return self.config_values.get('host_storage_type')

    @property
    def host_uniqueid(self):
        """Getter Method for the Unique ID for this Host"""
        return self.config_values.get('host_uniqueid')

    @property
    def host_ephemeral_disk_path(self):
        """Getter Method for the Ephemeral Disk Path for this Host"""
        return self.config_values.get('host_ephemeral_disk_path')

    @property
    def host_network_server(self):
        """Getter Method for the NFS Server IP for this Host"""
        return self.config_values.get('host_network_server')

    @property
    def host_network_share(self):
        """Getter Method for the NFS Share Name for this Host"""
        return self.config_values.get('host_network_share')

    @property
    def vioservers(self):
        """Getter Method for the list of VIOS's for this Host"""
        # If the VIOS's haven't been loaded yet, load them and cache them
        if self._vioservers is None:
            # Call the VioServer Factory to retrieve the list of VIO Servers
            self._vioservers = self.vios_factory.find_all_vioses_by_host(
                self._context, self.host_name, self._transaction)
        return self._vioservers

    @vioservers.setter
    def vioservers(self, vioservers):
        """Setter Method for updating the caching of VIOS's for this Host"""
        # Note:  This method only updates the Cache, it isn't DB update-able
        self._vioservers = vioservers if vioservers else list()

    @property
    def hmcs(self):
        """Getter Method for the list of HMC's for this Host"""
        # If the HMC's haven't been loaded yet, load them and cache them
        if self._hmcs is None and self.hmc_uuids:
            self._hmcs = list()
            hmcfact = ManagementConsoleFactory.get_factory()
            # Loop through each of the HMC UUID's, looking up each one
            for hmc_uuid in self.hmc_uuids:
                # Call the Factory to retrieve the HMC specified
                self._hmcs.append(hmcfact.find_hmc_by_uuid(
                                  self._context, hmc_uuid, self._transaction))
        return self._hmcs if self._hmcs is not None else list()

    @property
    def enterprise_pools(self):
        """Getter Method for the list of enterprise pools for this Host"""
        # If the enterprise pools haven't been loaded yet, load them and cache
        # them
        if self._enterprise_pools is None:
            fact = EnterprisePoolFactory.get_factory()
            self.enterprise_pools = fact.find_all_enterprise_pools_by_host(
                self._context, self.host_name, self._transaction)
        return self._enterprise_pools

    @property
    def shared_processor_pools(self):
        """Getter Method for the list of SPP's for this Host"""
        # If the SPP's haven't been loaded yet, load them and cache them
        if self._shared_processor_pools is None:
            sppfact = SharedProcessorPoolFactory.get_factory()
            self._shared_processor_pools = sppfact.find_all_spps_by_host(
                self._context, self.host_name, self._transaction)
        return self._shared_processor_pools


######################################################
#         Managed Instance DOM Implementation        #
######################################################
@base.NovaObjectRegistry.register
class ManagedInstanceFactory(dom.ResourceFactory):

    """Production DOM Resource Factory class implementation for Instances"""

    ###############################################
    #     Defining Instance Factory Interface     #
    ###############################################
    def find_all_instances(self, context, filters=None, transaction=None):
        """Queries all Instance matching the Filters specified"""
        return self.find_all_resources(
            ManagedInstance, context, filters, transaction)

    def find_all_instances_by_host(self, context, host_name, transaction=None):
        """Queries all Instances that exist on the Host specified"""
        return self.find_all_resources(ManagedInstance, context,
                                       dict(host=host_name), transaction)

    def find_instance_by_uuid(self, context, instance_uuid, transaction=None):
        """Queries at most one Instance matching the UUID specified"""
        return self.find_resource(ManagedInstance, context,
                                  dict(uuid=instance_uuid), transaction)

    def delete_instance(self, context, instance_uuid, transaction=None):
        """Deletes the Instance with the UUID specified"""
        return self.delete_resource(
            ManagedInstance, context, instance_uuid, transaction)

    ###############################################
    #   Implementing Resource Factory Definition  #
    ###############################################
    def _get_resource_id_attr_name(self):
        """Provides the Name of the Attribute for the Resource Identifier"""
        return 'uuid'

    def _auto_generate_resource_id(self, attribs):
        """Auto-generates the Resource Identifier for this Resource"""
        return str(uuid.uuid4())

    def _get_resource_classes(self):
        """Returns a List of DOM Classes or a Dictionary Mapping DOM Classes"""
        return [ManagedInstance]

    @base.remotable
    def _find_all_resources(self, resource_class,
                            filters=None, transaction=None):
        """Factory implementation to query all resources matching the filter"""
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        instances = db_api.instance_find_all(self._context, filters, dbtxn)
        return self._construct_resources(
            resource_class, self._context, instances, transaction)

    @base.remotable
    def _delete_multiple_resources(self, resource_class,
                                   resource_ids, transaction=None):
        """Factory implementation to delete the resources specified"""
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        for resource_id in resource_ids:
            db_api.instance_delete(self._context, resource_id, dbtxn)


@base.NovaObjectRegistry.register
class ManagedInstance(dom.Resource):

    """Production DOM Resource class implementation for Instances"""

    fields = dict(dom.Resource.fields,
                  id=fields.IntegerField(nullable=True),
                  uuid=fields.UUIDField(),
                  name=fields.StringField(nullable=True),
                  display_name=fields.StringField(nullable=True),
                  display_description=fields.StringField(default=''),
                  host=fields.StringField(nullable=True),
                  node=fields.StringField(nullable=True),
                  hostname=fields.StringField(nullable=True),
                  vm_state=fields.StringField(nullable=True),
                  power_state=fields.IntegerField(nullable=True),
                  task_state=fields.StringField(nullable=True),
                  memory_mb=fields.IntegerField(nullable=True),
                  vcpus=fields.IntegerField(nullable=True),
                  root_gb=fields.IntegerField(nullable=True),
                  ephemeral_gb=fields.IntegerField(nullable=True),
                  os_type=fields.StringField(nullable=True),
                  architecture=fields.StringField(nullable=True),
                  vm_mode=fields.StringField(nullable=True),
                  root_device_name=fields.StringField(nullable=True),
                  default_ephemeral_device=fields.StringField(nullable=True),
                  default_swap_device=fields.StringField(nullable=True),
                  config_drive=fields.StringField(default=''),
                  config_drive_id=fields.StringField(default=''),
                  kernel_id=fields.StringField(default=''),
                  ramdisk_id=fields.StringField(default=''),
                  access_ip_v4=fields.IPV4AddressField(nullable=True),
                  access_ip_v6=fields.IPV6AddressField(nullable=True),
                  user_id=fields.StringField(nullable=True),
                  project_id=fields.StringField(nullable=True),
                  availability_zone=fields.StringField(nullable=True),
                  image_ref=fields.StringField(nullable=True),
                  instance_type_id=fields.IntegerField(nullable=True),
                  reservation_id=fields.StringField(nullable=True),
                  key_name=fields.StringField(nullable=True),
                  key_data=fields.StringField(nullable=True),
                  user_data=fields.StringField(nullable=True),
                  created_at=fields.DateTimeField(nullable=True),
                  updated_at=fields.DateTimeField(nullable=True),
                  deleted_at=fields.DateTimeField(nullable=True),
                  scheduled_at=fields.DateTimeField(nullable=True),
                  launched_at=fields.DateTimeField(nullable=True),
                  terminated_at=fields.DateTimeField(nullable=True),
                  progress=fields.IntegerField(nullable=True),
                  cell_name=fields.StringField(nullable=True),
                  locked=fields.BooleanField(nullable=True),
                  cleaned=fields.BooleanField(nullable=True),
                  metadata=dom.MetaDataDictField(nullable=True),
                  system_metadata=dom.MetaDataDictField(nullable=True),
                  info_cache=fields.ObjectField(
                      'InstanceInfoCache', nullable=True),
                  flavor=fields.ObjectField('Flavor', nullable=True),
                  power_specs=dom.DictOfAnyTypeField(nullable=True))

    def __init__(self, context=None,
                 factory_class=None, transaction=None, **kwargs):
        "Constructs an instance of the given DOM Resource Class"
        super(ManagedInstance, self).__init__(
            context, factory_class, transaction, **kwargs)

    ###############################################
    #     Implementing DOM Resource Definition    #
    ###############################################
    def _hydrate_resource_attribute(self, attributes, key, value):
        """Hydrates a given Attribute onto this Resource Instance"""
        if key == 'info_cache' and value is not None:
            info_cache = instance_info_cache.InstanceInfoCache()
            info_cache.instance_uuid = value.get('instance_uuid')
            info_cache.network_info = value.get('network_info')
            info_cache._context = self._context
            value = info_cache
        # Need to parse the Flavor out of the Instance Extra Info
        if key == 'flavor':
            value = value if value else attributes.get('extra')
            # If the Flavor isn't in there then nothing to do
            if not value or not value.get('flavor'):
                return
            flavor_info = jsonutils.loads(value['flavor'])
            value = flavor.Flavor.obj_from_primitive(flavor_info['cur'])
        super(ManagedInstance, self).\
            _hydrate_resource_attribute(attributes, key, value)

    @base.remotable
    def _create_resource(self, transaction):
        """Actual Implementation that will Create the given Resource"""
        net_info, values = (u'[]', self._get_resource_values())
        flavor_info = values.pop('flavor')
        values.pop('name', None)
        # We need to default the Instance Cache's Network Info
        if values.get('info_cache') is not None:
            net_info = values['info_cache'].network_info.json()
        values['info_cache'] = {'network_info': net_info}
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        instance = db_api.instance_create(self._context, values, dbtxn)
        self._hydrate_all_resource_attributes(instance)
        # If they gave us the Flavor, we need to do that separately
        if flavor_info is not None:
            flavor_dict = dict(old=None, new=None)
            flavor_dict['cur'] = flavor_info.obj_to_primitive()
            extra_dict = dict(flavor=jsonutils.dumps(flavor_dict))
            base_db_api.instance_extra_update_by_uuid(
                self._context, instance['uuid'], extra_dict)
            self._hydrate_resource_attribute(dict(), 'extra', extra_dict)
            self.obj_reset_changes()

    @base.remotable
    def _update_resource(self, transaction):
        """Actual Implementation that will Update the given Resource"""
        inst_uuid = self._get_resource_identifer()
        values = self._get_resource_values(changes_only=True)
        values.pop('name', values.pop('info_cache', None))
        values.pop('flavor', None)
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        instance = db_api.instance_update(
            self._context, inst_uuid, values, dbtxn)
        self._hydrate_all_resource_attributes(instance)

    def get_flavor(self, namespace=None):
        prefix = ('%s_' % namespace) if namespace is not None else ''
        try:
            return getattr(self, '%sflavor' % prefix)
        except Exception:
            return None


######################################################
#      Shared Processor Pool DOM Implementation      #
######################################################
@base.NovaObjectRegistry.register
class SharedProcessorPoolFactory(dom.ResourceFactory):

    """Production DOM Resource Factory class implementation for SPP's"""

    ###############################################
    #       Defining SSP Factory Interface       #
    ###############################################
    def construct_spp(self, context, host_name, pool_id,
                      pool_name, maximum_proc_units,
                      reserved_proc_units, available_proc_units,
                      auto_create=False, transaction=None):
        """Constructs (and optionally creates) the VIOS specified"""
        kwargs = dict(host_name=host_name,
                      pool_id=pool_id, pool_name=pool_name,
                      maximum_proc_units=maximum_proc_units,
                      reserved_proc_units=reserved_proc_units,
                      available_proc_units=available_proc_units)
        # Call the Parent Class to actually Construct the given Resource
        return self.construct_resource(
            SharedProcessorPool, context, auto_create, transaction, **kwargs)

    def find_all_spps(self, context, filters=None, transaction=None):
        """Queries all SPP's matching the filters specified"""
        return self.find_all_resources(
            SharedProcessorPool, context, filters, transaction)

    def find_all_spps_by_host(self, context, host_name, transaction=None):
        """Queries all SPP's that are part of the host specified"""
        filters = dict(host_name=host_name)
        return self.find_all_resources(
            SharedProcessorPool, context, filters, transaction)

    def find_spp_by_uuid(self, context, pool_uuid, transaction=None):
        """Returns VIOS matching the given ID if found, else None."""
        filters = {self._get_resource_id_attr_name(): pool_uuid}
        return self.find_resource(
            SharedProcessorPool, context, filters, transaction)

    ###############################################
    #   Implementing Resource Factory Definition  #
    ###############################################
    def _get_resource_classes(self):
        """Returns a List of DOM Classes or a Dictionary Mapping DOM Classes"""
        return [SharedProcessorPool]

    def _get_resource_id_attr_name(self):
        """Provides the Name of the Attribute for the Resource Identifier"""
        return 'pool_uuid'

    def _auto_generate_resource_id(self, attribs):
        """Auto-generates the Resource Identifier for this Resource"""
        return self._generate_resource_uuid(
            attribs, ['host_name', 'pool_id', 'pool_name'])

    @base.remotable
    def _find_all_resources(self, resource_class,
                            filters=None, transaction=None):
        """Factory implementation to query all resources matching the filter"""
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        pools = db_api.spp_find_all(self._context, filters, dbtxn)
        return self._construct_resources(
            resource_class, self._context, pools, transaction)

    @base.remotable
    def _delete_multiple_resources(self, resource_class,
                                   resource_ids, transaction=None):
        """Factory implementation to delete the resources specified"""
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        for resource_id in resource_ids:
            db_api.spp_delete(self._context, resource_id, dbtxn)


@base.NovaObjectRegistry.register
class SharedProcessorPool(dom.Resource):

    """Production DOM Resource class implementation for SPP's"""

    fields = dict(dom.Resource.fields,
                  pool_id=fields.IntegerField(),
                  pool_name=fields.StringField(),
                  host_name=fields.StringField(),
                  pool_uuid=fields.StringField(nullable=True),
                  maximum_proc_units=fields.FloatField(),
                  reserved_proc_units=fields.FloatField(),
                  available_proc_units=fields.FloatField())

    obj_extra_fields = ['host']

    def __init__(self, context=None,
                 factory_class=None, transaction=None, **kwargs):
        "Constructs an instance of the given DOM Resource Class"
        super(SharedProcessorPool, self).__init__(
            context, factory_class, transaction, **kwargs)

    obj_extra_fields = ['host']

    ###############################################
    #     Implementing DOM Resource Definition    #
    ###############################################
    def _initialize_resource_attributes(self):
        """Sets any default values that are defined for any attributes"""
        super(SharedProcessorPool, self)._initialize_resource_attributes()
        self._host = None

    def _reset_cached_attributes(self):
        """Wipes out any Cached Resource Attributes that are no longer valid"""
        super(SharedProcessorPool, self)._reset_cached_attributes()
        self._host = None

    @base.remotable
    def _create_resource(self, transaction):
        """Actual Implementation that will Create the given Resource"""
        values = self._get_resource_values()
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        pool = db_api.spp_create(self._context, values, dbtxn)
        self._hydrate_all_resource_attributes(pool)

    @base.remotable
    def _update_resource(self, transaction):
        """Actual Implementation that will Update the given Resource"""
        pool_uuid = self._get_resource_identifer()
        values = self._get_resource_values(changes_only=True)
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        pool = db_api.spp_update(self._context, pool_uuid, values, dbtxn)
        self._hydrate_all_resource_attributes(pool)

    ###############################################
    #        Defining VioServer Interface         #
    ###############################################
    @property
    def host(self):
        """Getter Method for the Host for this Shared Processor Pool"""
        # If the Host hasn't been loaded yet, load and cache it
        if self._host is None:
            self._host = ManagedHostFactory.get_factory().find_host_by_name(
                self._context, self.host_name, self._transaction)
        return self._host


######################################################
#            VIO Server DOM Implementation           #
######################################################
@base.NovaObjectRegistry.register
class VioServerFactory(dom.ResourceFactory):

    """Production DOM Resource Factory class implementation for VIOS's"""

    ###############################################
    #       Defining VIOS Factory Interface       #
    ###############################################
    def construct_vios(self, context, host_name, lpar_id, lpar_name,
                       state, rmc_state, cluster_provider_name=None,
                       auto_create=False, transaction=None):
        """Constructs (and optionally creates) the VIOS specified"""
        kwargs = dict(host_name=host_name,
                      lpar_id=lpar_id, lpar_name=lpar_name,
                      state=state, rmc_state=rmc_state,
                      cluster_provider_name=cluster_provider_name)
        # Call the Parent Class to actually Construct the given Resource
        return self.construct_resource(
            VioServer, context, auto_create, transaction, **kwargs)

    def find_all_vioses(self, context, filters=None, transaction=None):
        """Queries all VIOS's matching the filters specified"""
        return self.find_all_resources(
            VioServer, context, filters, transaction)

    def find_all_vioses_by_host(self, context, host_name, transaction=None):
        """Queries all VIOS's that are part of the host specified"""
        filters = dict(host_name=host_name)
        return self.find_all_resources(
            VioServer, context, filters, transaction)

    def find_vios_by_id(self, context, vios_id, transaction=None):
        """Returns VIOS matching the given ID if found, else None."""
        filters = {self._get_resource_id_attr_name(): vios_id}
        return self.find_resource(VioServer, context, filters, transaction)

    def delete_vios(self, context, vios_id, transaction=None):
        """Deletes the VIOS with the Host Name + LPAR ID specified"""
        return self.delete_resource(
            VioServer, context, vios_id, transaction)

    ###############################################
    #    Defining VIOS Factory Abstract Methods   #
    ###############################################
    def _get_host_factory(self):
        """Provides the Factory to use for the Managed Host Implementation"""
        return ManagedHostFactory.get_factory()

    ###############################################
    #   Implementing Resource Factory Definition  #
    ###############################################
    def _get_resource_classes(self):
        """Returns a List of DOM Classes or a Dictionary Mapping DOM Classes"""
        return [VioServer]

    def _auto_generate_resource_id(self, attribs):
        """Auto-generates the Resource Identifier for this Resource"""
        return self._generate_resource_uuid(attribs, ['host_name', 'lpar_id'])

    @base.remotable
    def _find_all_resources(self, resource_class,
                            filters=None, transaction=None):
        """Factory implementation to query all resources matching the filter"""
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        vioses = db_api.vios_find_all(self._context, filters, dbtxn)
        return self._construct_resources(
            resource_class, self._context, vioses, transaction)

    @base.remotable
    def _delete_multiple_resources(self, resource_class,
                                   resource_ids, transaction=None):
        """Factory implementation to delete the resources specified"""
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        for resource_id in resource_ids:
            db_api.vios_delete(self._context, resource_id, dbtxn)


@base.NovaObjectRegistry.register
class VioServer(dom.Resource):

    """Production DOM Resource class implementation for VIOS's"""

    fields = dict(dom.Resource.fields,
                  id=fields.StringField(nullable=True),
                  lpar_id=fields.IntegerField(),
                  lpar_name=fields.StringField(),
                  host_name=fields.StringField(),
                  state=fields.StringField(nullable=True),
                  rmc_state=fields.StringField(nullable=True),
                  cluster_provider_name=fields.StringField(nullable=True),
                  backend_uuid=fields.StringField(nullable=True))

    obj_extra_fields = ['host']

    def __init__(self, context=None,
                 factory_class=None, transaction=None, **kwargs):
        "Constructs an instance of the given DOM Resource Class"
        super(VioServer, self).__init__(
            context, factory_class, transaction, **kwargs)

    obj_extra_fields = ['host']

    ###############################################
    #     Implementing DOM Resource Definition    #
    ###############################################
    def _initialize_resource_attributes(self):
        """Sets any default values that are defined for any attributes"""
        super(VioServer, self)._initialize_resource_attributes()
        self._host = None

    def _reset_cached_attributes(self):
        """Wipes out any Cached Resource Attributes that are no longer valid"""
        super(VioServer, self)._reset_cached_attributes()
        self._host = None

    @base.remotable
    def _create_resource(self, transaction):
        """Actual Implementation that will Create the given Resource"""
        values = self._get_resource_values()
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        vios = db_api.vios_create(self._context, values, dbtxn)
        self._hydrate_all_resource_attributes(vios)

    @base.remotable
    def _update_resource(self, transaction):
        """Actual Implementation that will Update the given Resource"""
        vios_id = self._get_resource_identifer()
        values = self._get_resource_values(changes_only=True)
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        vios = db_api.vios_update(self._context, vios_id, values, dbtxn)
        self._hydrate_all_resource_attributes(vios)

    ###############################################
    #        Defining VioServer Interface         #
    ###############################################
    @property
    def host_factory(self):
        """Return the HostFactory that is associated to this object."""
        return self._get_resource_factory()._get_host_factory()

    @property
    def host(self):
        """Getter Method for the Host for this VioServer"""
        # If the Host hasn't been loaded yet, load and cache it
        if self._host is None and self.host_name is not None:
            self._host = self.host_factory.find_host_by_name(
                self._context, self.host_name, self._transaction)
        return self._host

    @host.setter
    def host(self, host):
        """Setting Method for the VIOS for this VioServer"""
        self.host_name = None
        self._host = host
        # If the Host wasn't wiped out, update the Name to match
        if host is not None:
            self.host_name = host.host_name


######################################################
#          VIOS Component DOM Implementation         #
######################################################
@base.NovaObjectRegistry.register
class VioServerComponentFactory(dom.ResourceFactory):

    """Production DOM Factory class implementation for VIOS Components"""

    ###############################################
    #    Defining VIOS Factory Abstract Methods   #
    ###############################################
    def _get_vios_id_attr_name(self):
        """Provides the Name of the Attribute for the VIOS Identifier"""
        return 'vios_id'

    def _get_vios_factory(self):
        """Provides the Factory to use for the VIO Server Implementation"""
        return VioServerFactory.get_factory()


@base.NovaObjectRegistry.register
class VioServerComponent(dom.Resource):

    """Production DOM Resource class implementation for VIOS Components"""
    fields = dict(dom.Resource.fields,
                  id=fields.StringField(nullable=True),
                  vios_id=fields.StringField(nullable=False))

    obj_extra_fields = ['vioserver']

    def __init__(self, context=None,
                 factory_class=None, transaction=None, **kwargs):
        "Constructs an instance of the given DOM Resource Class"
        super(VioServerComponent, self).__init__(
            context, factory_class, transaction, **kwargs)

    ###############################################
    #     Implementing DOM Resource Definition    #
    ###############################################
    def _initialize_resource_attributes(self):
        """Sets any default values that are defined for any attributes"""
        super(VioServerComponent, self)._initialize_resource_attributes()
        self._vioserver = None

    def _reset_cached_attributes(self):
        """Wipes out any Cached Resource Attributes that are no longer valid"""
        super(VioServerComponent, self)._reset_cached_attributes()
        self._vioserver = None

    ###############################################
    #      Defining VIO Component Interface       #
    ###############################################
    @property
    def vios_factory(self):
        """Return the VioServerFactory that is associated to this object."""
        return self._get_resource_factory()._get_vios_factory()

    @property
    def vioserver(self):
        """Getter Method for the VIOS for this VIOS Component"""
        # If the VIOS hasn't been loaded yet, load and cache it
        if self._vioserver is None and self.vios_id is not None:
            self._vioserver = self.vios_factory.find_vios_by_id(
                self._context, self.vios_id, self._transaction)
        return self._vioserver

    @vioserver.setter
    def vioserver(self, vioserver):
        """Setting Method for the VIOS for this VIOS Component"""
        self._vioserver = vioserver
        # If the VIOS wasn't wiped out, update the ID to match
        if vioserver is not None:
            self.vios_id = vioserver.id
        else:
            # Not valid without a vioserver, the following will fail
            self.vios_id = None


#################################################################
#    Enterprise Pool (Capacity on Demand) DOM Implementation    #
#################################################################
@base.NovaObjectRegistry.register
class EnterprisePoolFactory(dom.ResourceFactory):

    """
    Production DOM Resource Factory class implementation for enterprise pools
    """

    ######################################################
    #       Defining Enterprise Pool Factory Interface   #
    ######################################################
    def construct_enterprise_pool(self, context, uuid, name, master_hmc_mtmsn,
                                  available_mobile_cores, total_mobile_cores,
                                  compliance_state, host_names=None,
                                  auto_create=False, transaction=None):
        """Constructs (and optionally creates) the enterprise pool specified"""
        kwargs = dict(uuid=uuid, name=name, master_hmc_mtmsn=master_hmc_mtmsn,
                      available_mobile_cores=available_mobile_cores,
                      total_mobile_cores=total_mobile_cores,
                      compliance_state=compliance_state, host_names=host_names)
        # Call the Parent Class to actually Construct the given Resource
        return self.construct_resource(EnterprisePool, context, auto_create,
                                       transaction, **kwargs)

    def find_all_enterprise_pools(self, context, filters=None,
                                  transaction=None):
        """Queries all enterprise pools that match the filters specified"""
        return self.find_all_resources(EnterprisePool, context, filters,
                                       transaction)

    def find_all_enterprise_pools_by_host(self, context, host_name,
                                          transaction=None):
        """Queries all enterprise pools that contain the Host specified"""
        return self.find_all_resources(EnterprisePool, context,
                                       dict(host_name=host_name), transaction)

    def find_enterprise_pool_by_uuid(self, context, uuid, transaction=None):
        """
        Queries at most one enterprise pool matching the pool UUID specified
        """
        return self.find_resource(EnterprisePool, context,
                                  dict(uuid=uuid),
                                  transaction)

    def delete_enterprise_pool(self, context, uuid, transaction=None):
        """Deletes the enterprise pool with the pool UUID specified"""
        return self.delete_resource(EnterprisePool, context, uuid,
                                    transaction)

    ###############################################
    #   Implementing Resource Factory Definition  #
    ###############################################
    def _get_resource_id_attr_name(self):
        """Provides the Name of the Attribute for the Resource Identifier"""
        return 'uuid'

    def _get_resource_classes(self):
        """Returns a List of DOM Classes or a Dictionary Mapping DOM Classes"""
        return [EnterprisePool]

    @base.remotable
    def _find_all_resources(self, resource_class,
                            filters=None, transaction=None):
        """Factory implementation to query all resources matching the filter"""
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        pools = db_api.enterprise_pool_find_all(self._context, filters, dbtxn)
        return self._construct_resources(
            resource_class, self._context, pools, transaction)

    @base.remotable
    def _delete_multiple_resources(self, resource_class,
                                   resource_ids, transaction=None):
        """Factory implementation to delete the resources specified"""
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        for resource_id in resource_ids:
            db_api.enterprise_pool_delete(self._context, resource_id, dbtxn)


@base.NovaObjectRegistry.register
class EnterprisePool(dom.Resource):

    """Production DOM Resource class implementation for enterprise pools"""

    fields = dict(dom.Resource.fields,
                  uuid=fields.StringField(),
                  name=fields.StringField(),
                  master_hmc_mtmsn=fields.StringField(),
                  available_mobile_cores=fields.IntegerField(),
                  total_mobile_cores=fields.IntegerField(),
                  compliance_state=fields.StringField(),
                  host_names=fields.DictOfStringsField(nullable=True))

    def __init__(self, context=None,
                 factory_class=None, transaction=None, **kwargs):
        "Constructs an instance of the given DOM Resource Class"
        super(EnterprisePool, self).__init__(
            context, factory_class, transaction, **kwargs)

    ###############################################
    #     Implementing DOM Resource Definition    #
    ###############################################
    def _initialize_resource_attributes(self):
        """Sets any default values that are defined for any attributes"""
        super(EnterprisePool, self)._initialize_resource_attributes()
        setattr(self, 'host_names', dict())

    def _reset_cached_attributes(self):
        """Wipes out any Cached Resource Attributes that are no longer valid"""
        super(EnterprisePool, self)._reset_cached_attributes()

    @base.remotable
    def _create_resource(self, transaction):
        """Actual Implementation that will Create the given Resource"""
        values = self._get_resource_values()
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        pool = db_api.enterprise_pool_create(self._context, values, dbtxn)
        self._hydrate_all_resource_attributes(pool)

    @base.remotable
    def _update_resource(self, transaction):
        """Actual Implementation that will Update the given Resource"""
        pool_uuid = self._get_resource_identifer()
        values = self._get_resource_values(changes_only=True)
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        pool = db_api.enterprise_pool_update(self._context, pool_uuid, values,
                                             dbtxn)
        self._hydrate_all_resource_attributes(pool)


######################################################
#      Host Compute Topology DOM Implementation      #
######################################################
@base.NovaObjectRegistry.register
class HostComputeTopologyFactory(dom.ResourceFactory):

    """Production DOM Factory implementation for Host Compute Topology"""

    ###############################################
    #       Defining HCT Factory Interface        #
    ###############################################
    def construct_compute_topology(self, context, host_name,
                                   shared_processor_pools, transaction=None):
        """Constructs the Network Topology for the Host specified"""
        pools = shared_processor_pools if shared_processor_pools else list()
        # Call the Parent Class to actually Construct the given Resource
        topo = self.construct_resource(HostComputeTopology, context, False,
                                       transaction, host_name=host_name)
        topo.shared_processor_pools = [pool for pool in pools]
        return topo

    def find_network_topology(self, context, host_name, transaction=None):
        """Queries the Network Topology for the Host specified"""
        return self.find_resource(HostComputeTopology, context,
                                  dict(host_name=host_name), transaction)

    ###############################################
    #   Implementing Resource Factory Definition  #
    ###############################################
    def _get_resource_id_attr_name(self):
        """Provides the Name of the Attribute for the Resource Identifier"""
        return 'host_name'

    def _get_resource_classes(self):
        """Returns a List of DOM Classes or a Dictionary Mapping DOM Classes"""
        return [HostComputeTopology]

    @base.remotable
    def _find_all_resources(self, resource_class,
                            filters=None, transaction=None):
        """Factory implementation to query all resources matching the filter"""
        return list()


@base.NovaObjectRegistry.register
class HostComputeTopology(dom.Resource):

    """Production DOM Resource implementation for Host Compute Topology"""

    fields = dict(dom.Resource.fields,
                  host_name=fields.StringField(),
                  shared_processor_pools=dom.DomListField(SharedProcessorPool))

    def __init__(self, context=None,
                 factory_class=None, transaction=None, **kwargs):
        "Constructs an instance of the given DOM Resource Class"
        super(HostComputeTopology, self).__init__(
            context, factory_class, transaction, **kwargs)

    ###############################################
    #     Implementing DOM Resource Definition    #
    ###############################################
    def _initialize_resource_attributes(self):
        """Sets any default values that are defined for any attributes"""
        super(HostComputeTopology, self)._initialize_resource_attributes()
        setattr(self, 'shared_processor_pools', list())

    def _create_resource(self, transaction):
        """Actual Implementation that will Create the given Resource"""
        # Since the Topology isn't actually persisted we will treat
        # someone who queried the Topology first the same as construct it
        self._update_resource(transaction)

    @base.remotable
    def _update_resource(self, transaction):
        """Actual Implementation that will Update the given Resource"""
        filters = dict(host_name=self.host_name)
        # If the Transaction isn't set, then we need to create one to use
        if transaction is None:
            topofact = self._get_resource_factory()
            transaction = topofact.get_resource_transaction()
        # Begin the Transaction to make sure everything is committed together
        with transaction.begin(subtransactions=True):
            dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
            # Retrieve the existing VEA's for the Host to figure out
            # differences
            dbpools = db_api.spp_find_all(self._context, filters, dbtxn)
            exist_pools = dict([(p['pool_uuid'], p) for p in dbpools])
            new_pools = dict([(p['pool_uuid'], p)
                              for p in self.shared_processor_pools])
            # Determine whether the Processor Pool needs to be Created/Updated
            for pool in new_pools.values():
                values = pool._get_resource_values(changes_only=False)
                # If the Pool already exists, just want to update what changed
                if exist_pools.pop(pool['pool_uuid'], None) is not None:
                    db_api.spp_update(
                        self._context, pool['pool_uuid'], values, dbtxn)
                # If the Pool is new, then we want to create it in the database
                else:
                    db_api.spp_create(self._context, values, dbtxn)
            # Now we want to clean up any Pools that no longer exist
            for pool in exist_pools.values():
                db_api.spp_delete(self._context, pool['pool_uuid'], dbtxn)


######################################################
#     Compute DOM Resource Factory Registration      #
######################################################
dom.ResourceFactoryLocator.register_factory(VioServerFactory())
dom.ResourceFactoryLocator.register_factory(ManagedHostFactory())
dom.ResourceFactoryLocator.register_factory(ManagedInstanceFactory())
dom.ResourceFactoryLocator.register_factory(ManagementConsoleFactory())
dom.ResourceFactoryLocator.register_factory(EnterprisePoolFactory())
dom.ResourceFactoryLocator.register_factory(SharedProcessorPoolFactory())
dom.ResourceFactoryLocator.register_factory(HostComputeTopologyFactory())
db_api.initialize()
