# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import time
import uuid
from copy import deepcopy
import six

from oslo_config import cfg
from oslo_log import log as logging
from oslo_context.context import _request_store
from nova.objects import base, fields
from nova.volume import cinder
from nova.ibm.ego import ego_constants as op_types
from nova import context as nova_context

from powervc_nova.db import api as db_api
from powervc_nova.storage import exception as stgex
from powervc_nova.storage.static_topology import StaticStgTopo
from powervc_nova.storage.warnings import IBMPowerVCStorageWarning as WarnMsg
from powervc_nova.volume.cinder import PowerVCVolumeAPI, PowerVCFabricAPI

import powervc_nova.objects.dom as dom
import powervc_nova.objects.compute.dom as compute_dom
from powervc_nova import logcall, _


cfg.CONF.register_opts([
    # At least while novalink is in development, we need to sync more often
    # so setting the default interval to be smaller than the periodic task
    # default interval.
    cfg.IntOpt('ibmpowervc_fabric_sync_interval', default=60,
               help='The general synchronization interval in minutes, '
               'for setting the fabric information in the DB that Fibre '
               'Channel ports are logged into. This covers cases where '
               'a fabric change occurs, but a new fabric did not get '
               'registered'),
    ######################################################################
    # For the new vios redundancy and pv-vscsi support, a number of
    # properties have been added to the SCG. We don't plan on adding more
    # auto_defined SCGs, so the following config options are added to
    # give some flexibility in how the auto-defined (default) SCGs are
    # configured.
    ######################################################################
    cfg.IntOpt('ibmpowervm_default_vios_redundancy',
               default=1,
               help='If vios_redundancy is not specified on storage '
               'connectivity group (SCG) creation, then the attribute will '
               'default to this value. The auto_defined SCGs will also '
               'use this as their vios_redundancy value. For example, a '
               'value of 2 means volume attach operations must create '
               'virtual adapters to at least 2 VIOSes for the attach to '
               'be successful. Valid range is 1 - 4.'),
    cfg.BoolOpt('ibmpowervm_default_exact_redundancy',
                default=False,
                help='If exact_redundancy is not specified on storage '
                'connectivity group (SCG) creation, then the attribute will '
                'default to this value. The auto_defined SCGs will also '
                'use this value. For example, if this option is set to '
                'true, SCGs that use it will restrict virtual adapter '
                'creation to reference exactly the number of VIOSs '
                'specified by vios_redundancy.'),
    cfg.BoolOpt('ibmpowervm_prefer_unique_adapter',
                default=True,
                help='If True, and choosing more than one port per fabric per '
                'VIOS, then choose the least utilized port on unique adapters '
                'when possible. Otherwise just choose the least utilized '
                'without respect to physical adapter.'),
    cfg.BoolOpt('ibmpowervm_use_mapped_vfc_scheduling',
                default=False,
                help='If the mapped vfc scheduling is False then the '
                'number of available connections on each fc port will '
                'depend on the number of active npiv logins on that port.'
                'When True it depends on the number of vfc mappings on that '
                'port.')
])

cfg.CONF.import_opt('powervm_mgr_type',
                    'powervc_nova.virt.ibmpowervm')

LOG = logging.getLogger(__name__)
_NPIV_, _SSP_, _VSCSI_ = ('npiv', 'ibm_ssp', 'pv_vscsi')
SCG_DOWN = 'Inactive'
CHK_BOOT, CHK_DATA, CHK_ALL, CHK_EITHER = 'boot', 'data', 'all', 'either'


# #####################################################
# ########   Managed Host DOM Implementation   ########
# #####################################################
@base.NovaObjectRegistry.register
class ManagedHostFactory(compute_dom.ManagedHostFactory):

    """Production DOM Factory implementation for Hosts"""

    # ##############################################
    # ## Implementing Resource Factory Definition ##
    # ##############################################
    def _get_resource_classes(self):
        """Returns a List of DOM Classes or a Dictionary Mapping DOM Classes"""
        return {compute_dom.ManagedHost: ManagedHost}

    def _get_vios_factory(self):
        """Provides the Factory to use for the VIO Server Implementation"""
        return VioServerFactory.get_factory()


@base.NovaObjectRegistry.register
class ManagedHost(compute_dom.ManagedHost):

    """Production DOM Resource implementation for Managed Hosts"""

    obj_extra_fields = ['fcports']

    def __init__(self, context=None,
                 factory_class=None, transaction=None, **kwargs):
        "Constructs an instance of the given DOM Resource Class"
        super(ManagedHost, self).__init__(
            context, factory_class, transaction, **kwargs)

    # ##############################################
    # #### Implementing DOM Resource Definition ####
    # ##############################################
    def _initialize_resource_attributes(self):
        """Sets any default values that are defined for any attributes"""
        super(ManagedHost, self)._initialize_resource_attributes()
        self._fcports = None

    def _reset_cached_attributes(self):
        """Wipes out any Cached Resource Attributes that are no longer valid"""
        super(ManagedHost, self)._reset_cached_attributes()
        self._fcports = None

    # ##############################################
    # #####  Defining Managed Host Interface  ######
    # ##############################################
    @property
    def fcports(self):
        """Getter Method for the list of FCPorts for this Host"""
        # If the Ports haven't been loaded yet, load them and cache them
        if self._fcports is None:
            fact = FcPortFactory.get_factory()
            # Call the FCPort Factory to retrieve the list of Ports for the
            # Host
            self._fcports = fact.find_all_fcports_by_host(
                self._context, self.host_name, self._transaction)
        return self._fcports


# #####################################################
# #########   VIO Server DOM Implementation   #########
# #####################################################
@base.NovaObjectRegistry.register
class VioServerFactory(compute_dom.VioServerFactory):

    """Production DOM Factory implementation for VIOS's"""

    # ##############################################
    # #####  Defining VIOS Factory Interface  ######
    # ##############################################
    def find_all_vioses_by_cluster(self, context,
                                   cluster_provider_name, transaction=None):
        """Queries all VIOS's that are part of the VIOS Cluster specified"""
        filters = dict(cluster_provider_name=cluster_provider_name)
        return self.find_all_resources(
            VioServer, context, filters, transaction)

    # ##############################################
    # ## Implementing Resource Factory Definition ##
    # ##############################################
    def _get_resource_classes(self):
        """Returns a List of DOM Classes or a Dictionary Mapping DOM Classes"""
        return {compute_dom.VioServer: VioServer}

    def _get_host_factory(self):
        """Provides the Factory to use for the Managed Host Implementation"""
        return ManagedHostFactory.get_factory()


@base.NovaObjectRegistry.register
class VioServer(compute_dom.VioServer):

    """Production DOM Resource implementation for VIO Servers"""

    obj_extra_fields = ['fcports', 'cluster_provider']

    def __init__(self, context=None,
                 factory_class=None, transaction=None, **kwargs):
        "Constructs an instance of the given DOM Resource Class"
        super(VioServer, self).__init__(
            context, factory_class, transaction, **kwargs)

    # ##############################################
    # #### Implementing DOM Resource Definition ####
    # ##############################################
    def _initialize_resource_attributes(self):
        """Sets any default values that are defined for any attributes"""
        super(VioServer, self)._initialize_resource_attributes()
        self._fcports, self._cluster_provider = (None, None)

    def _reset_cached_attributes(self):
        """Wipes out any Cached Resource Attributes that are no longer valid"""
        super(VioServer, self)._reset_cached_attributes()
        self._fcports, self._cluster_provider = (None, None)

    # ##############################################
    # ######  Defining VIO Server Interface  #######
    # ##############################################
    @property
    def fcports(self):
        """Getter Method for the list of FCPorts for this VIOS"""
        # If the Ports haven't been loaded yet, load them and cache them
        if self._fcports is None:
            fact = FcPortFactory.get_factory()
            # Call the FCPort Factory to get the FC Ports for the VIOS.
            LOG.debug("Lookup ports in DB for: %s" % self.lpar_name)
            self._fcports = fact.find_all_fcports_by_vios(
                self._context, self.id, self._transaction)
        return self._fcports

    @property
    def cluster_provider(self):
        """Getter Method for the VIOS Cluster for this VIOS"""
        # If they don't have a VIOS Cluster Name set, then just return None
        if self.cluster_provider_name is None:
            return None
        # If the Storage Provider haven't been loaded yet, load and cache it
        if self._cluster_provider is None:
            fact = StorageProviderFactory.get_factory()
            # Call the Provider Factory to retrieve the given Storage Provider
            self._cluster_provider = fact.find_provider_by_name(
                self._context, self.cluster_provider_name, self._transaction)
        return self._cluster_provider

    @cluster_provider.setter
    def cluster_provider(self, provider):
        """Setting Method for the VIOS Cluster for this VIOS"""
        self.cluster_provider_name = None
        self._cluster_provider = provider
        # If the Provider wasn't wiped out, update the Name to match
        if provider is not None:
            self.cluster_provider_name = provider.host_name

    def is_rmc_state_ok(self):
        """ Use this object method to return True if the RMC state of this
            VIOS is one of the allowed states that will not prevent a VIOS
            from being flagged as storage-ready for SCGs.

            Current impl accepts ['active', 'busy']
            A busy state may mean that a VIOS is not ready for storage
            connectivity, but that will be determined at attach time rather
            than when an SCG is evaluated through public APIs.
        """
        if self.rmc_state == 'busy':
            LOG.info(_("The Virtual I/O Server '%(vios)s' on host %(host)s "
                       "has an RMC state of 'busy', which may impact its "
                       "connectivity capability for storage operations.") %
                     dict(vios=self.lpar_name, host=self.host_name))
        return (self.rmc_state in ['active', 'busy'])


# #####################################################
# #######  Storage Provider DOM Implementation  #######
# #####################################################
@base.NovaObjectRegistry.register
class StorageProviderFactory(dom.ResourceFactory):

    """Production DOM Factory implementation for Storage Providers"""

    # ##############################################
    # #### Defining Provider Factory Interface #####
    # ##############################################
    def construct_emc_controller(self, context, access_ip,
                                 volume_pool_name, provider_ip,
                                 provider_user_id, provider_password,
                                 provider_port=None, host_display_name=None,
                                 auto_create=False, transaction=None):
        """Constructs (and optionally creates) the Provider specified"""
        kwargs = dict(host_type='emc', access_ip=access_ip,
                      volume_pool_name=volume_pool_name,
                      provider_ip=provider_ip,
                      provider_port=provider_port,
                      provider_password=provider_password,
                      provider_user_id=provider_user_id,
                      host_display_name=host_display_name)
        # Call the Parent Class to actually Construct the given Resource
        return self.construct_resource(EmcController, context,
                                       auto_create, transaction, **kwargs)

    def construct_emc_vmax_controller(self, context, access_ip,
                                      volume_pool_name, provider_ip,
                                      provider_user_id, provider_password,
                                      provider_port=None,
                                      host_display_name=None,
                                      auto_create=False, transaction=None):
        """Constructs (and optionally creates) the Provider specified"""
        kwargs = dict(host_type='emc_vmax', access_ip=access_ip,
                      volume_pool_name=volume_pool_name,
                      provider_ip=provider_ip,
                      provider_port=provider_port,
                      provider_password=provider_password,
                      provider_user_id=provider_user_id,
                      host_display_name=host_display_name)
        # Call the Parent Class to actually Construct the given Resource
        return self.construct_resource(EmcVMAXController, context,
                                       auto_create, transaction, **kwargs)

    def construct_emc_vnx_controller(self, context, access_ip,
                                     volume_pool_name, provider_ip,
                                     provider_user_id, provider_password,
                                     provider_port=None,
                                     host_display_name=None,
                                     auto_create=False, transaction=None):
        """Constructs (and optionally creates) the Provider specified"""
        kwargs = dict(host_type='emc_vnx', access_ip=access_ip,
                      volume_pool_name=volume_pool_name,
                      provider_ip=provider_ip,
                      provider_port=provider_port,
                      provider_password=provider_password,
                      provider_user_id=provider_user_id,
                      host_display_name=host_display_name)
        # Call the Parent Class to actually Construct the given Resource
        return self.construct_resource(EmcVNXController, context,
                                       auto_create, transaction, **kwargs)

    def construct_svc_controller(self, context, access_ip,
                                 volume_pool_name, user_id,
                                 password=None, private_key=None,
                                 ssh_port=None, host_display_name=None,
                                 auto_create=False, transaction=None):
        """Constructs (and optionally creates) the Provider specified"""
        kwargs = dict(host_type='svc', access_ip=access_ip,
                      volume_pool_name=volume_pool_name, user_id=user_id,
                      password=password, private_key=private_key,
                      ssh_port=ssh_port, host_display_name=host_display_name)
        # Call the Parent Class to actually Construct the given Resource
        return self.construct_resource(SvcController, context,
                                       auto_create, transaction, **kwargs)

    def construct_vios_cluster(self, context, host_name, hmc_uuids,
                               host_display_name=None, unique_id=None,
                               auto_create=False, transaction=None):
        """Constructs (and optionally creates) the Provider specified"""
        kwargs = dict(host_type='ssp', host_name=host_name,
                      hmc_uuids=hmc_uuids, unique_id=unique_id,
                      host_display_name=host_display_name)
        # Call the Parent Class to actually Construct the given Resource
        return self.construct_resource(ViosCluster, context,
                                       auto_create, transaction, **kwargs)

    def find_all_providers(self, context, transaction=None):
        """Queries all Storage Providers through the Cinder REST API"""
        return self.find_all_resources(
            StorageProvider, context, None, transaction)

    def find_all_providers_by_type(self, context, host_type, transaction=None):
        """Queries all Providers matching the Host Type specified"""
        return self.find_all_resources(StorageProvider, context,
                                       dict(host_type=host_type), transaction)

    def find_provider_by_name(self, context, host_name, transaction=None):
        """Queries at most one Provider matching the Host Name specified"""
        return self.find_resource(StorageProvider, context,
                                  dict(host_name=host_name), transaction)

    def find_providers_by_display_name(self, context, display_name,
                                       transaction=None):
        """Queries Providers matching the display name specified"""
        return self.find_all_resources(StorageProvider, context,
                                       dict(host_display_name=display_name),
                                       transaction)

    def delete_provider(self, context, host_name, transaction=None):
        """Deletes the Storage Provider with the given Name from Cinder"""
        return self.delete_resource(
            StorageProvider, context, host_name, transaction)

    # ##############################################
    # ## Implementing Resource Factory Definition ##
    # ##############################################
    def _get_resource_id_attr_name(self):
        """Provides the Name of the Attribute for the Resource Identifier"""
        return 'host_name'

    def _get_resource_classes(self):
        """Returns a List of DOM Classes or a Dictionary Mapping DOM Classes"""
        return [StorageProvider, EmcController, EmcVMAXController,
                EmcVNXController, SvcController, XivController,
                Ds8kController, ViosCluster, GenericController,
                HitachiController]

    @base.remotable
    def _find_all_resources(self, resource_class,
                            filters=None, transaction=None):
        """Factory implementation to query all resources matching the filter"""
        resources = list()
        # Call the Cinder REST API to retrieve the list of Storage Providers
        providers = self._get_providers_from_cinder(self._context, filters)
        for provider in providers:
            provider_class = self._get_provider_class(provider['host_type'])
            # Construct the appropriate Resource Class for the Provider
            aresource = self._construct_resources(
                provider_class, self._context, [provider], transaction)[0]
            resources.append(aresource)
        return resources

    @base.remotable
    def _delete_multiple_resources(self, resource_class,
                                   resource_ids, transaction=None):
        """Factory implementation to delete the resources specified"""
        client = cinder.cinderclient(self._context).client
        scgfact = StorageConnectivityGroupFactory.get_factory()
        # For each Host Name provided delete the Provider through REST
        for resource_id in resource_ids:
            resp = client.delete('/os-hosts/' + resource_id +
                                 '/internal', raise_exc=False)[0]
            # Handle Exceptions ourself, since the Clients don't give info
            if resp.status_code >= 400:
                raise Exception(
                    'HTTP Error %s - %s' % (resp.status_code, resp.text))
            # Delete the default SCG created for the given Cluster
            scgfact.delete_default_cluster_scg(self._context, resource_id)

    # ##############################################
    # ####  Shared Provider DOM Helper Methods  ####
    # ##############################################
    def _get_provider_class(self, backend_type):
        """Helper method to return the right resource class for the Provider"""
        type_map = dict(emc=EmcController, svc=SvcController,
                        xiv=XivController, ssp=ViosCluster,
                        ds8k=Ds8kController, generic=GenericController,
                        hitachi=HitachiController,
                        emc_vmax=EmcVMAXController, emc_vnx=EmcVNXController)
        return type_map.get(backend_type)

    def _get_providers_from_cinder(self, context, filters):
        """Helper method to call the Cinder REST API to retrieve Providers"""
        providers = []
        client = cinder.cinderclient(context).client
        # Call the Cinder REST API to retrieve the list of Providers
        for host in client.get('/os-hosts')[1]['hosts']:
            # If this is a Provider retrieve the details of the Provider
            if host['service'] == 'cinder-volume':
                body = client.get('/os-hosts/' + host['host_name'])[1]
                for resource in body['host']:
                    # We want to try to find the Registration section
                    if 'registration' not in resource:
                        continue
                    provider = resource['registration']
                    # Make sure that the Provider matches the Filter specified
                    if self._provider_matches_filters(provider, filters):
                        providers.append(provider)
        return providers

    def _provider_matches_filters(self, provider, filters):
        """Helper method to determine if the provider matches the filters"""
        # See if any of the provided Filters don't match the Provider
        if filters:
            for key, val in filters.iteritems():
                if provider.get(key) != val:
                    return False
        return provider is not None and len(provider) > 0


@base.NovaObjectRegistry.register
class StorageProvider(dom.Resource):

    """Production DOM Resource implementation for Storage Providers"""

    fields = dict(dom.Resource.fields,
                  host_name=fields.StringField(),
                  host_type=fields.StringField(),
                  host_display_name=fields.StringField(nullable=True),
                  volume_pool_name=fields.StringField(nullable=True),
                  registered_at=fields.DateTimeField(nullable=True))

    def __init__(self, context=None,
                 factory_class=None, transaction=None, **kwargs):
        "Constructs an instance of the given DOM Resource Class"
        super(StorageProvider, self).__init__(
            context, factory_class, transaction, **kwargs)

    # ##############################################
    # #### Implementing DOM Resource Definition ####
    # ##############################################
    @base.remotable
    def _create_resource(self, transaction):
        """Actual Implementation that will Create the given Resource"""
        client = cinder.cinderclient(self._context).client
        values = self._get_resource_values()
        # Make sure that the Display Name is unique for the Storage Provider
        self._make_unique_display_name(client, values)
        registration = dict(registration=values)
        # Call the Cinder REST API to Register the Storage Provider
        client.post('/os-hosts', body=dict(host=registration))
        self._hydrate_all_resource_attributes(values)
        # If this is a VIOS Cluster, then create the default SCG for it
        scgfact = StorageConnectivityGroupFactory.get_factory()
        scgfact.create_default_cluster_scg(
            self._context, values['host_name'], values['host_display_name'])

    @base.remotable
    def _update_resource(self, transaction):
        """Actual Implementation that will Update the given Resource"""
        client = cinder.cinderclient(self._context).client
        host_name = self._get_resource_identifer()
        values = self._get_resource_values(changes_only=True)
        # Call the Cinder REST API to Register the Storage Provider
        url = '/os-hosts/' + host_name + '/update-registration'
        client.put(url, body=dict(registration=values))
        self._hydrate_all_resource_attributes(values)

    def _make_unique_display_name(self, client, values):
        """Helper method to generate a unique display name"""
        try:
            # Only try to update the Display Name if we were given one
            if values.get('host_display_name'):
                orig_name, count = (values['host_display_name'], 2)
                # Get the set of the Display Names for the current Providers
                names = self._get_display_names(client, values['host_name'])
                # Loop through until we can get a unique Display Name
                while values['host_display_name'] in names:
                    adict = dict(origname=orig_name, postfix=count)
                    # Add a numerical post-fix such as: My Provider (2)
                    new_name = _("%(origname)s (%(postfix)d)") % adict
                    values['host_display_name'] = new_name
                    count = count + 1
        # If we couldn't retrieve the Providers, assume unique
        except Exception as exc:
            LOG.warn(_("Error retrieving Storage Providers"))
            LOG.exception(exc)

    def _get_display_names(self, client, host_name):
        """Helper method to get the list of current Display Names"""
        dispnames = list()
        # Call the Cinder REST API to retrieve the list of Providers
        for host in client.get('/os-hosts')[1]['hosts']:
            # If this is a Provider retrieve the details of the Provider
            # If the Provider is registered already, skip since name is fine
            if ((host['service'] != 'cinder-volume' or
                 host['host_name'] == host_name)):
                continue
            body = client.get('/os-hosts/' + host['host_name'])[1]
            for resource in body['host']:
                # We want to try to find the Registration section
                if 'registration' in resource:
                    provider = resource['registration']
                    dispnames.append(provider['host_display_name'])
        return dispnames


@base.NovaObjectRegistry.register
class EmcController(StorageProvider):

    """Production DOM Resource implementation for EMC Storage Controllers"""

    fields = dict(StorageProvider.fields,
                  access_ip=fields.StringField(),
                  user_id=fields.StringField(),
                  provider_ip=fields.StringField(),
                  provider_user_id=fields.StringField(),
                  password=fields.StringField(nullable=True),
                  provider_password=fields.StringField(nullable=True))


@base.NovaObjectRegistry.register
class EmcVMAXController(StorageProvider):

    """Production DOM Resource implementation for XIV Storage Controllers"""

    fields = dict(StorageProvider.fields,
                  access_ip=fields.StringField(),
                  user_id=fields.StringField(),
                  password=fields.StringField(nullable=True))


@base.NovaObjectRegistry.register
class EmcVNXController(StorageProvider):

    """Production DOM Resource implementation for XIV Storage Controllers"""

    fields = dict(StorageProvider.fields,
                  access_ip=fields.StringField(),
                  user_id=fields.StringField(),
                  password=fields.StringField(nullable=True))


@base.NovaObjectRegistry.register
class SvcController(StorageProvider):

    """Production DOM Resource implementation for SVC Storage Controllers"""

    fields = dict(StorageProvider.fields,
                  access_ip=fields.StringField(),
                  user_id=fields.StringField(),
                  password=fields.StringField(nullable=True),
                  ssh_port=fields.StringField(nullable=True),
                  private_key_data=fields.StringField(nullable=True))


@base.NovaObjectRegistry.register
class XivController(StorageProvider):

    """Production DOM Resource implementation for XIV Storage Controllers"""

    fields = dict(StorageProvider.fields,
                  access_ip=fields.StringField(),
                  user_id=fields.StringField(),
                  password=fields.StringField(nullable=True))


@base.NovaObjectRegistry.register
class Ds8kController(StorageProvider):

    """Production DOM Resource implementation for DS8K Storage Controllers"""

    fields = dict(StorageProvider.fields,
                  access_ip=fields.StringField(),
                  user_id=fields.StringField(),
                  password=fields.StringField(nullable=True))


@base.NovaObjectRegistry.register
class HitachiController(StorageProvider):

    """ Production DOM Resource implementation for
        Hitachi Storage Controllers"""

    fields = dict(StorageProvider.fields,
                  access_ip=fields.StringField(),
                  user_id=fields.StringField(),
                  password=fields.StringField(nullable=True))


@base.NovaObjectRegistry.register
class ViosCluster(StorageProvider):

    """Production DOM Resource implementation for VIOS Cluster Providers"""

    fields = dict(StorageProvider.fields,
                  hmc_uuids=fields.ListOfStringsField(nullable=True))


@base.NovaObjectRegistry.register
class GenericController(StorageProvider):

    """Production DOM Resource implementation for Generic Providers"""

    fields = dict(StorageProvider.fields,
                  volume_driver=fields.StringField())


# #####################################################
# ######  Connectivity Group DOM Implementation  ######
# #####################################################
@base.NovaObjectRegistry.register
class StorageConnectivityGroupFactory(dom.ResourceFactory):

    """Production DOM Factory implementation for Storage Connectivity Groups"""

    SORTTYPE_EXTERNAL, SORTTYPE_CLUSTER = ('external', 'cluster')
    # TODO: npiv_fabric and following list is being made obsolete.
    ALOWED_NPIV_FABRIC_VALS = ['A', 'B', 'dual', 'single', 'any',
                               'dual_per_vios', None]
    ALLOWED_FABRIC_SET_REQ = ['at_least_one', 'exactly_one',
                              'per_vios', 'per_host', None]
    # ##############################################
    # #####  Defining SCG Factory Interface  #######
    # ##############################################

    def construct_scg(self, context, display_name, enabled=None,
                      auto_defined=None, auto_add_vios=None,
                      fc_storage_access=None, port_tag=None,
                      cluster_provider_name=None, priority_cluster=None,
                      priority_external=None, vios_ids=None,
                      vios_redundancy=None, exact_redundancy=None,
                      npiv_fabric=None, backend_host_group=None,
                      boot_connectivity=None, data_connectivity=None,
                      include_untagged=None, fabric_set_npiv=None,
                      fabric_set_req=None, ports_per_fabric_npiv=None,
                      auto_create=False, transaction=None):
        """Constructs (and optionally creates) the SCG specified"""
        LOG.debug("Construct SCG '%s'." % display_name)
        # ****************************************#
        #           Set the defaults              #
        # ****************************************#
        # Make sure the VIOS ID List is defaulted to an empty list
        vios_ids = vios_ids if vios_ids else list()
        # Set the default value form the VIOS Redundancy parameter
        if vios_redundancy is None:
            vios_redundancy = cfg.CONF.ibmpowervm_default_vios_redundancy
            if vios_redundancy < 1 or vios_redundancy > 4:
                LOG.warn(_("CONF.ibmpowervm_default_vios_redundancy=%d is"
                           "outside the allowable default range 1-4.") %
                         vios_redundancy)
                vios_redundancy = 1
        # Set the default value from the Exact Redundancy parameter
        if exact_redundancy is None:
            exact_redundancy = cfg.CONF.ibmpowervm_default_exact_redundancy
        # npiv_fabric has been superseded by fabric_set_req
        if npiv_fabric not in self.ALOWED_NPIV_FABRIC_VALS:
            LOG.warn(_("The npiv_fabric value of '%s' is not allowed. "
                       "Converting to 'any' or None."))
            npiv_fabric = 'any' if fc_storage_access else None
        if backend_host_group is not None and len(backend_host_group) == 0:
            backend_host_group = None
        if boot_connectivity is None:
            boot_connectivity = []
            if fc_storage_access:
                boot_connectivity.append(_NPIV_)
            if cluster_provider_name:
                boot_connectivity.append(_SSP_)
        if data_connectivity is None:
            data_connectivity = []
            if fc_storage_access:
                data_connectivity.append(_NPIV_)
            if cluster_provider_name:
                data_connectivity.append(_SSP_)
        if include_untagged is None:
            include_untagged = True  # default for existing SCGs
        if fabric_set_req is None:
            fabric_set_req = 'at_least_one' if fc_storage_access else None
        if ports_per_fabric_npiv is None:
            ports_per_fabric_npiv = 1 if fc_storage_access else None
        if fabric_set_npiv is None:
            fabric_set_npiv = ['*'] if fc_storage_access else None

        # Create a dictionary housing the parameters provided as input
        kwargs = dict(display_name=display_name, enabled=enabled,
                      auto_defined=auto_defined, auto_add_vios=auto_add_vios,
                      fc_storage_access=fc_storage_access, port_tag=port_tag,
                      cluster_provider_name=cluster_provider_name,
                      priority_external=priority_external,
                      priority_cluster=priority_cluster, vios_ids=vios_ids,
                      vios_redundancy=vios_redundancy, npiv_fabric=npiv_fabric,
                      exact_redundancy=exact_redundancy,
                      backend_host_group=backend_host_group,
                      boot_connectivity=boot_connectivity,
                      data_connectivity=data_connectivity,
                      include_untagged=include_untagged,
                      fabric_set_npiv=fabric_set_npiv,
                      fabric_set_req=fabric_set_req,
                      ports_per_fabric_npiv=ports_per_fabric_npiv)
        # Call the Parent Class to actually Construct the given Resource
        return self.construct_resource(StorageConnectivityGroup, context,
                                       auto_create, transaction, **kwargs)

    def create_default_fc_scg(self, context, transaction=None):
        """Creates the overall Default SCG (NPIV) if it doesn't exist yet"""
        # We don't want to even try creating the Default SCG if not HMC
        if 'hmc' not in compute_dom.ManagedHost.get_supported_host_types():
            return None
        filters = dict(fc_storage_access=True, auto_defined=True)
        scgs = self.find_all_scgs(context, filters, transaction)
        # Only want to create the Default FC SCG if it doesn't exist
        found = None
        for scg in scgs:
            if scg.display_name == 'Any host, all VIOS':
                found = scg
            if scg.fc_storage_access and '*' not in scg.fabric_set_npiv:
                LOG.info(_("Automatically updating auto-generated storage "
                           "connectivity group '%(scg)s' to have a dynamic "
                           "fabric set.") % {'scg': scg.display_name})
                scg.fabric_set_npiv = ['*']
                scg.save(context=context, transaction=transaction)
        if not found:
            LOG.info(_("Auto-create the default storage connectivity group."))
            return self.construct_scg(
                context, 'Any host, all VIOS', True, True, True, True,
                priority_external=1, boot_connectivity=[_NPIV_],
                data_connectivity=[_NPIV_], auto_create=True,
                transaction=transaction, include_untagged=False,
                fabric_set_req='at_least_one', fabric_set_npiv=['*'],
                ports_per_fabric_npiv=1)
        return found

    def create_default_cluster_scg(self, context,
                                   provider_id, cluster_display_name):
        """Creates the Default SCG for the Cluster if it doesn't exist yet"""
        name = 'Any host in ' + cluster_display_name
        scgs = self.find_all_scgs_by_cluster(context, provider_id)
        # Only want to create the Default Cluster SCG if it doesn't exit
        for scg in scgs:
            if scg.display_name == name and scg.auto_defined:
                LOG.info(_("Default storage connectivity group for SSP \
                provider %(provider_id)s already exists: %(scg_id)s.") %
                         dict(provider_id=provider_id, scg_id=scg.id))
                return scg

        # We also need to check for default SCGs with the same name because
        # SCG display_name must be unique per DB constraints. This needs
        # to be removed if found.
        filters = {'display_name': name, 'auto_defined': True}
        scgs = self.find_all_resources(
            StorageConnectivityGroup, context, filters=filters)
        if len(scgs) > 0:
            scgids = [scg.id for scg in scgs]
            LOG.info(_("Delete the automatically defined storage connectivity "
                       "group with ID %(scgids)s for the cluster provider "
                       "with ID %(prov_id)s that no longer exists, in "
                       "preparation for registering a provider with the "
                       "same name.") %
                     dict(scgids=scgids,
                          prov_id=scgs[0].cluster_provider_name))
            self.delete_multiple_resources(
                StorageConnectivityGroup, context, scgids)
        LOG.info(_("Now construct a new SSP-based storage connectivity group "
                   "with name: %s.") % name)
        return self.construct_scg(
            context, name, True, True, True, True,
            cluster_provider_name=provider_id, boot_connectivity=[_SSP_],
            data_connectivity=[_SSP_, _NPIV_], auto_create=True)

    def find_all_scgs(self, context, filters=None,
                      sort_type=None, transaction=None):
        """Queries all SCG's matching the Filters specified"""
        filters = dict() if filters is None else filters
        specific_cluster = False
        if filters.get('cluster_provider_name', None) is not None:
            priority_select = lambda scg: scg.priority_cluster
            specific_cluster = True
        if sort_type != self.SORTTYPE_CLUSTER or specific_cluster:
            self._context = context
            # Look for cluster filter to avoid selecting all clusters
            scg_dicts_list = self.find_all_resources(
                StorageConnectivityGroup, context, filters=filters,
                transaction=transaction)
            if sort_type == self.SORTTYPE_EXTERNAL:
                priority_select = lambda scg: scg.priority_external
                LOG.debug("first_scg=%s" % scg_dicts_list[0])
                scg_dicts_list = [x for x in scg_dicts_list if _NPIV_ in
                                  x['boot_connectivity'] or _VSCSI_ in
                                  x['boot_connectivity'] or _NPIV_ in
                                  x['data_connectivity'] or _VSCSI_ in
                                  x['data_connectivity']]
        else:
            priority_select = lambda scg: scg.priority_cluster
            scg_dicts_list = self.find_all_scgs_with_any_cluster(
                context, filters=filters, transaction=transaction)
        scg_list = self._construct_resources(StorageConnectivityGroup,
                                             context, scg_dicts_list,
                                             transaction)
        return sorted(scg_list, key=priority_select) if sort_type else scg_list

    def find_all_scgs_by_cluster(self, context, cluster_provider_name=None,
                                 filters=None, transaction=None):
        """Queries all SCG's matching the Cluster Provider specified"""
        filters = dict() if filters is None else filters
        # If the Cluster Provider is None, we want to query on all of them
        if cluster_provider_name is not None:
            filters['cluster_provider_name'] = cluster_provider_name
        # Now we can run the database query for the matching SCG's
        return self.find_all_resources(
            StorageConnectivityGroup, context, filters, transaction)

    def find_all_scgs_with_any_cluster(self, context,
                                       filters=None, transaction=None):
        """Queries all SCG's with a Cluster Provider set on it"""
        return self.find_all_scgs_by_cluster(
            context, None, filters, transaction)

    def find_scg_by_id(self, context, scg_id, transaction=None):
        """Queries at most one SCG matching the SCG ID specified"""
        filters = {self._get_resource_id_attr_name(): scg_id}
        return self.find_resource(
            StorageConnectivityGroup, context, filters, transaction)

    def delete_scg(self, context, scg_id, transaction=None):
        """Deletes the SCG with the SCG ID specified"""
        LOG.info(_("The storage connectivity group with ID %(scg_id)s is "
                   "requested to be deleted.") % dict(scg_id=scg_id))
        return self.delete_resource(
            StorageConnectivityGroup, context, scg_id, transaction)

    def delete_default_cluster_scg(self, context,
                                   provider_id, transaction=None):
        """Deletes the default SCG for the VIOS Cluster specified"""
        filters = dict(auto_defined=True, cluster_provider_name=provider_id)
        # Query the SCG's that match the given Cluster Provider
        scgs = self.find_all_scgs(context, filters, None, transaction)
        scgids = [scg.id for scg in scgs]
        LOG.info(_("Delete the automatically defined storage connectivity "
                   "group for the cluster provider '%(provider)s' with the "
                   "following connectivity group ID or IDs: %(scgids)s.") %
                 dict(provider=provider_id, scgids=scgids))
        if scgids:
            # Only call if we looked up any matching SCGs since it may
            # have already been removed.
            return self.delete_multiple_resources(
                StorageConnectivityGroup, context, scgids, transaction)

    # ##############################################
    # ## Implementing Resource Factory Definition ##
    # ##############################################
    def _get_resource_classes(self):
        """Returns a List of DOM Classes or a Dictionary Mapping DOM Classes"""
        return [StorageConnectivityGroup]

    def _auto_generate_resource_id(self, attribs):
        """Auto-generates the Resource Identifier for this Resource"""
        return str(uuid.uuid4())

    @base.remotable
    def _find_all_resources(self, resource_class,
                            filters=None, transaction=None):
        """Factory implementation to query all resources matching the filter"""
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        scgs = db_api.scg_find_all(self._context, filters, dbtxn)
        return self._construct_resources(
            resource_class, self._context, scgs, transaction)

    @base.remotable
    def _delete_multiple_resources(self, resource_class,
                                   resource_ids, transaction=None):
        """Factory implementation to delete the resources specified"""
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        for resource_id in resource_ids:
            db_api.scg_delete(self._context, resource_id, dbtxn)


@base.NovaObjectRegistry.register
class StorageConnectivityGroup(dom.Resource):

    """Production DOM Resource implementation for Storage Connectivity Group"""

    fields = dict(dom.Resource.fields,
                  id=fields.StringField(),
                  display_name=fields.StringField(),
                  enabled=fields.BooleanField(nullable=True),
                  auto_defined=fields.BooleanField(nullable=True),
                  auto_add_vios=fields.BooleanField(nullable=True),
                  fc_storage_access=fields.BooleanField(nullable=True),
                  port_tag=fields.StringField(nullable=True),
                  cluster_provider_name=fields.StringField(nullable=True),
                  priority_cluster=fields.IntegerField(nullable=True),
                  priority_external=fields.IntegerField(nullable=True),
                  vios_ids=fields.ListOfStringsField(nullable=True),
                  vios_redundancy=fields.IntegerField(nullable=True),
                  exact_redundancy=fields.BooleanField(nullable=True),
                  npiv_fabric=fields.StringField(nullable=True),
                  boot_connectivity=fields.ListOfStringsField(nullable=True),
                  data_connectivity=fields.ListOfStringsField(nullable=True),
                  backend_host_group=fields.StringField(nullable=True),
                  include_untagged=fields.BooleanField(nullable=True),
                  fabric_set_npiv=fields.ListOfStringsField(nullable=True),
                  fabric_set_req=fields.StringField(nullable=True),
                  ports_per_fabric_npiv=fields.IntegerField(nullable=True))

    obj_extra_fields = ['vioservers']

    def __init__(self, context=None,
                 factory_class=None, transaction=None, **kwargs):
        "Constructs an instance of the given DOM Resource Class"
        super(StorageConnectivityGroup, self).__init__(
            context, factory_class, transaction, **kwargs)
        self.prefer_unique_adapter = cfg.CONF.ibmpowervm_prefer_unique_adapter

    # ##############################################
    # #### Implementing DOM Resource Definition ####
    # ##############################################
    def _initialize_resource_attributes(self):
        """Sets any default values that are defined for any attributes"""
        super(StorageConnectivityGroup, self)._initialize_resource_attributes()
        self._vioservers, self._dict_rep = (None, None)
        self._force_load_all = False
        self.vios_ids = []
        self.vol_api = PowerVCVolumeAPI()

    def _reset_cached_attributes(self):
        """Wipes out any Cached Resource Attributes that are no longer valid"""
        super(StorageConnectivityGroup, self)._reset_cached_attributes()
        self._vioservers = None

    @base.remotable
    def _create_resource(self, transaction):
        """Actual Implementation that will Create the given Resource"""
        values = self._get_resource_values()
        viosfact = VioServerFactory.get_factory()
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        # If we should Auto-Add the VIOS's then we want to add them on creation
        if values.get('auto_add_vios') and not values['vios_ids']:
            provider = values.get('cluster_provider_name')
            filters = None if provider is None \
                else dict(cluster_provider_name=provider)
            vioses = viosfact.find_all_vioses(
                self._context, filters, transaction)
            values['vios_ids'] = [vios.id for vios in vioses]
        # Now call the Database API to actually create the SCG in the DB
        scg = db_api.scg_create(self._context, values, dbtxn)
        self._hydrate_all_resource_attributes(scg)

    @base.remotable
    def _update_resource(self, transaction):
        """Actual Implementation that will Update the given Resource"""
        scg_id = self._get_resource_identifer()
        values = self._get_resource_values(changes_only=True)
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        scg = db_api.scg_update(self._context, scg_id, values, dbtxn)
        self._hydrate_all_resource_attributes(scg)

    # ##############################################
    # ########    Defining SCG Interface   #########
    # ##############################################
    @property
    def vioservers(self):
        """Getter Method for the list of VIOS's for this SCG"""
        # This method contains an added optimization for compute hosts.
        # If the code is running in the compute host process, then in most
        # cases, the compute host only cares about VIOSs that it owns.
        # The exception to this is during topology collection for the
        # compute host - it needs to update the VIOS membership of SCGs
        # when VIOSs are to be automatically added or removed from the
        # SCGs. The full list of VIOSs needs to be stored back through the
        # conductor, not just the delta. So the vioserver_members is
        # added below to handle that case and has the original logic.
        #
        # Compute hosts have a process-wide cache for their topology-
        # reconciled VIOSs. So, the optimization is to check if there are
        # any items in the cache. If so, then we can assume we are running
        # on the compute host and can return the VIOS doms for IDs that
        # match self.vios_ids.
        topo = HostStorageTopologyFactory.theFactory.get_static_topology()
        if topo.db_vioses:
            if self._vioservers is not None:
                return self._vioservers
            self._vioservers = topo.get_vioses(self.vios_ids)
            LOG.debug("***Return only subset of SCG VIOSs - the ones on "
                      "this host: %s" % [v.id for v in self._vioservers])
            return self._vioservers

        # Otherwise, take the normal (old) path.
        return self.vioserver_members()

    def vioserver_members(self, port_preload=True):
        """ Getter for 'vioservers' that always returns all the SCG
            member vios objects. Used by topology reconcilation to be
            able to update the full list appropriately. The logic here
            used to be in vioservers().
        """
        viosfact = VioServerFactory.get_factory()
        # If the VIOS's haven't been loaded yet, load them and cache them
        if self._vioservers is not None:
            id_set = set([v.id for v in self._vioservers])
        # If the VIOS's weren't loaded or the list has changed, load now
        if self._vioservers is None or id_set != set(self.vios_ids):
            LOG.debug("Lookup VIOSs for scg %s." % self.display_name)
            self._vioservers = viosfact.find_all_vioses(
                self._context, dict(vios_ids=self.vios_ids), self._transaction)
            # A DB optimization can be done here under the assumption
            # that if SCG vios entries are requested, then typically the
            # FC ports for those vioses will also be requested later.
            # Rather than load the set of FC ports for an individual VIOS,
            # we call a new factory method to load all the ports of all
            # the SCG member VIOSes here. So we pay a bigger penalty once
            # rather than a smaller penalty multiple times that grows as
            # you scale the number of hosts.
            if port_preload:
                fcportFact = FcPortFactory.get_factory()
                ports = fcportFact.find_all_fcports_by_vios_ids(
                    self._context, self.vios_ids,
                    transaction=self._transaction, vioses=self._vioservers)
                # At this point, the ports should be linked into their VIOS
                # DOMs so just log the number and return.
                LOG.debug("Materialized %d ports across %d VIOS members for "
                          "SCG: %s" % (len(ports), len(self.vios_ids),
                                       self.display_name))
        return self._vioservers

    @vioservers.setter
    def vioservers(self, vioservers):
        self.vios_ids = list()
        self._vioservers = vioservers
        for vios in vioservers:
            self.vios_ids.append(vios.id)

    # #################################
    # Extended DOM methods for SCGs  #
    # #################################
    @logcall
    def to_dict(self, context=None, transaction=None,
                skip_extended_lookups=False,
                force_load_all_members=False):
        """
        Return a dictionary representation of this SCG resource object.
        If a cached dictionary was previously built, then that will be
        returned, else one is built and cached on the DOM, and returned.
        :param context: Optional authorization context
        :param transaction: Optional active transaction scope to pass down
        :param skip_extended_lookup: If True, then processing will be
                optimized to not do extra database lookups or cinder calls
                to get SCG information that is pulled in from those areas.
                This type of call is not mainline and not cached.
        :param force_load_all_members: When running in the compute host
               process, there is a typical optimization that is done to
               only load vios scg members that are located on this host.
               When True, the force_load_all_members bypasses that optim-
               ization and loads all the member vioses. Needed for migrate.
        :returns: A dictionary representation of this SCG instance that
                  has the form defined by callers of the REST
                  storage-connectivity-groups API if not skipping extended
                  lookups.
        """
        if self._dict_rep is not None and not skip_extended_lookups and\
                self._force_load_all == force_load_all_members:
            LOG.debug("SCG dict already computed for id=%s, name=%s." %
                      (self.id, self.display_name))
            return self._dict_rep
        t1 = time.time()
        # Otherwise construct the dictionary
        d = self._get_dict_of_basic_attrs()

        # remove props we don't expose in the DOM fashion
        if "cluster_provider_name" in d.keys():
            del d["cluster_provider_name"]

        # Add in the basic cluster dict if SCG is associated with one
        if self.cluster_provider_name:
            d["vios_cluster"] = {"provider_name": self.cluster_provider_name}

        # If not doing extended lookups, then just return a simplified
        # dictionary, but don't cache it.
        if skip_extended_lookups:
            d['vios_ids'] = self.vios_ids if self.vios_ids is not None else []
            if 'priority_cluster' in d:
                del d['priority_cluster']
            if 'priority_external' in d:
                del d['priority_external']
            return d

        # Mainline path: Convert the associated topology info into nested
        # dictionary data for the SCG.
        hosts = {}

        if force_load_all_members:
            vios_members = self.vioserver_members()
            self._force_load_all = True
        else:
            # Here we grab the VioServer DOMs. If we are running in the compute
            # host, then we only "see" the member VIOSs of this host.
            vios_members = self.vioservers

        h_name = None
        for vios in vios_members:
            h_name = vios.host_name
            if h_name not in hosts.keys():
                hosts[h_name] = {"name": h_name, "vios_list": []}
            v = {"name": vios.lpar_name, "id": vios.id,
                 "lpar_id": vios.lpar_id,
                 "state": vios.state, "rmc_state": vios.rmc_state}
            hosts[h_name]["vios_list"].append(v)
        d["host_list"] = hosts.values()
        if len(vios_members) == len(self.vios_ids):
            LOG.debug("SCG %s has %d hosts and %d VIOSs." %
                      (self.id, len(d['host_list']), len(vios_members)))
        elif h_name:
            LOG.debug("SCG %s has %d VIOSs for this compute host %s." %
                      (self.id, len(vios_members), h_name))

        # If fabric_set_npiv is a wildcard, we want to build a fabric_list
        # containing the vios associated with the scg to be returned.
        # In an update to a scg where the wildcard is removed,
        # than clear fabric_list
        if 'fabric_set_npiv' in d and d['fabric_set_npiv'] \
                and '*' in d['fabric_set_npiv']:
            fabric_list = []
            # Loop over vios_members again, which means if running on
            # compute, then only the set from the host is seen to
            # perform better. This should be OK if we assume that
            # get_connectivity_info() called from compute assumes that
            # the SCG meets the fabric requirement. Otherwise, we may need
            # to persist 'dynamic_fabric_set' as an extra prop.
            for vios in vios_members:
                for db_port in vios.fcports:
                    port = fcport_to_dict(db_port)
                    if port['fabric'] and port['fabric'] != 'None' \
                            and port['fabric'] not in fabric_list:
                        fabric_list.append(port['fabric'])
            d['dynamic_fabric_set'] = self.fabric_list = fabric_list

        # Add provider information to dictionary.
        # 'applicable_providers' is new in 1.2.1 and it contains a list
        # of applicable storage_hostnames for the SCG that are not in
        # 'error' state. This allows for efficiencies in logic that needs
        # the SCG provider info, and if enabled in API output, this could
        # render the server_storage_hosts API obsolete.
        pci = None
        try:
            if self.vol_api:
                pci = self.vol_api.get_stg_provider_info(context,
                                                         use_cache=True)
                d['applicable_providers'] = []
                d['boot_providers'] = []
                d['data_providers'] = []
        except Exception as ex:
            LOG.exception(ex)
            LOG.warn(_("Could not retrieve provider data."))
        if pci and (_NPIV_ in d['boot_connectivity'] or
                    _VSCSI_ in d['boot_connectivity'] or
                    _NPIV_ in d['data_connectivity'] or
                    _VSCSI_ in d['data_connectivity']):
            for p in pci.values():
                if p['stg_type'] == 'fc' and p['backend_state'] != "error":
                    d['applicable_providers'].append(p['storage_hostname'])
                    if _NPIV_ in d['boot_connectivity'] or\
                            _VSCSI_ in d['boot_connectivity']:
                        d['boot_providers'].append(p['storage_hostname'])
                    if _NPIV_ in d['data_connectivity'] or\
                            _VSCSI_ in d['data_connectivity']:
                        d['data_providers'].append(p['storage_hostname'])
        cluster = self.cluster_provider_name
        if cluster is not None:
            if pci and cluster in pci:
                d["vios_cluster"]["provider_display_name"] = \
                    pci[cluster]["host_display_name"]
                d["vios_cluster"]["backend_state"] = \
                    pci[cluster]["backend_state"]
                if pci[cluster]["backend_state"] != "error":
                    d['applicable_providers'].append(cluster)
                    if _SSP_ in d['boot_connectivity']:
                        d['boot_providers'].append(cluster)
                    if _SSP_ in d['data_connectivity']:
                        d['data_providers'].append(cluster)

        self._dict_rep = d
        LOG.debug("EXIT SCG-DOM.to_dict() --> %s." % d)
        LOG.debug("Delta seconds=%.4f " % (time.time() - t1))
        return d

    def _get_dict_of_basic_attrs(self):
        """Returns a dict of simple, non-None attributes on this SCG.

        Attributes that are themselves other Resources are not included here.
        """
        scg_attr_keys = [
            'auto_add_vios', 'auto_defined', 'cluster_provider_name',
            'created_at', 'display_name', 'enabled', 'fc_storage_access',
            'id', 'priority_cluster', 'priority_external',
            'updated_at', 'vios_redundancy', 'exact_redundancy',
            'npiv_fabric', 'backend_host_group',
            'port_tag', 'include_untagged',
            'boot_connectivity', 'data_connectivity', 'fabric_set_npiv',
            'fabric_set_req', 'ports_per_fabric_npiv']
        attr_dict = {}
        for key in scg_attr_keys:
            if self[key] is not None and self[key] != "None":
                attr_dict[key] = self[key]
        # Pre-existing SCGs default to True for untagged inclusion.
        if self.include_untagged is None and self.port_tag is not None:
            self.include_untagged = attr_dict['include_untagged'] = True
        # Remove properties that are not applicable for connectivity types
        if _NPIV_ not in attr_dict['boot_connectivity'] and\
                _NPIV_ not in attr_dict['data_connectivity']:
            attr_dict.pop('fabric_set_npiv', None)
            attr_dict.pop('fabric_set_req', None)
            attr_dict.pop('ports_per_fabric_npiv', None)
            attr_dict.pop('npiv_fabric', None)
        return attr_dict

    @logcall
    def to_dict_with_conn(self, context, include_ports=True, host_name=None,
                          transaction=None, filter_applicable=False,
                          skip_provider_check=False):
        """
        Return the SCG dictionary with connectivity information computed.
        So the nested structure is
        {
            'prop': <value>,...
            'host_list': [{
                'vios_list': [{
                    'fcport_list': [{
                        'port_prop': <value>,...
                    },...]
                },...]
            },...]
        }

        If include_ports is False, then use the port information only in
        calculating if the member VIOS is 'ready' and other health attributes.
        Otherwise, FC Ports are included in the output.

        If host_name is not none, then scope SCG output to only the provided
        host. This is useful for checking whether a particular host has
        sufficient connectivity.

        If filter_applicable is True, then VIOSs that don't have
        connectivity applicable to the SCG will not be returned in the output.
        Likewise for hosts with no applicable VIOSs. However, there will be
        health_status info returned for filtered resources.

        Regardless the filter_applicable value, only Ports that are
        'enabled', have a matching SCG port_tag, have an OK status, and are
        connected to an appropriate fabric - are included in the output.
        (we may choose to include un-applicable ports in the future based on
         the filter_applicable value)

        skip_provider_check - workaround because some processes like compute
            and scheduler may have issues connecting to cinder.
            TODO: add check back in, in future
        """
        scg = self  # init var for use in this method
        if host_name is None or host_name != cfg.CONF.host:
            # The connectivity request could be for the destination host of
            # a migration. In this case, we will pass in a param that says
            # to bypass optimization done for the compute host (source).
            LOG.debug("Calling SCG.to_dict() with force_load_all_members.")
            load_all = True
        else:
            load_all = False
        # First deep copy the base to_dict() data structure.
        # This will look up member VIOSes if necessary.
        scg_dict = deepcopy(self.to_dict(context=context,
                                         transaction=transaction,
                                         force_load_all_members=load_all))
        # Build a map of vios id to vios DOM for use later.
        vios_map = dict(map(lambda x: (x.id, x), scg.vioservers))
        # Get connectivity types allowed by the SCG.
        scg_conn_types = self.get_connectivity_types()
        effective_conn_stg = deepcopy(scg_conn_types)
        health_status = {'value_reasons': [], 'member_status': {},
                         'boot_health': 'OK', 'data_health': 'OK',
                         'vios_boot_redundancy_met': False,
                         'vios_data_redundancy_met': False,
                         'scg_fabric_req_met': False}

        if not self.enabled:
            msg = WarnMsg(WarnMsg.SCGNotEnabled).getMsg(context)
            health_status['value_reasons'].append(msg)
            health_status['boot_health'] = 'SCG_Disabled'
            health_status['data_health'] = 'SCG_Disabled'

        # check if a provider is available for SCG
        if 'applicable_providers' in scg_dict and not skip_provider_check:
            if not scg_dict['applicable_providers']:
                # There is not at least one OK registered provider
                msg = WarnMsg(WarnMsg.NoStorageProviders).getMsg(context)
                health_status['value_reasons'].append(msg)
                effective_conn_stg = set()
            elif 'vios_cluster' in scg_dict and scg_dict['vios_cluster']:
                clst = scg_dict['vios_cluster']
                if 'backend_state' in clst and\
                        clst['backend_state'] == 'error':
                    effective_conn_stg = effective_conn_stg - set([_SSP_])
                elif len(scg_dict['applicable_providers']) == 1:
                    effective_conn_stg = effective_conn_stg - set([_NPIV_,
                                                                   _VSCSI_])
            if not scg_dict.get('boot_providers', None):
                health_status['boot_health'] = 'No_Providers'
            if not scg_dict.get('data_providers', None):
                health_status['data_health'] = 'No_Providers'

            LOG.debug("After provider check: SCG connectivity set=%s, "
                      "effective_conn=%s, boot_health=%s, data_health=%s." %
                      (scg_conn_types, effective_conn_stg,
                       health_status['boot_health'],
                       health_status['data_health']))
        # ##################################################################
        # 'vios_ready_count' tracks the number of VIOSes that have been
        # determined to be 'storage_ready'. Note that The 'storage_ready'
        # value of true does not necessarily predict a successful
        # deployment using this VIOS, and since vios_redundancy may now
        # require more than a single VIOS (per host) for successful
        # attach/deploy, the 'xxx_host_ready_count's are a more
        # useful indicator for SCG health.
        # ##################################################################
        scg_dict['vios_ready_count'] = 0
        scg_dict['boot_vios_ready_count'] = 0
        scg_dict['data_vios_ready_count'] = 0
        scg_dict['boot_host_ready_count'] = 0
        scg_dict['data_host_ready_count'] = 0
        scg_dict['port_ready_count'] = dict([(x, 0) for x in scg_conn_types])
        host_name_verified = False
        effective_vios_conn = set()
        # Trim list to passed host if applicable
        if host_name is not None:
            scg_dict['host_list'] = [h for h in scg_dict['host_list'] if
                                     h['name'] == host_name]
        #######################
        # Host loop           #
        #######################
        # TODO: This is a large method which could be broken up into smaller
        #       parts for readability.
        for host in scg_dict['host_list']:
            host_name_verified = True
            host_vios_ready_count = 0  # per host
            host_vios_boot_ready_count = 0
            host_vios_data_ready_count = 0
            host_boot_conn = set(self.boot_connectivity)
            host_data_conn = set(self.data_connectivity)
            health = {'connectivity': 'none', 'reasons': [],
                      'effective_fabrics': set(), 'port_health': [],
                      'fabric_req_met': False}
            conn_type_map = dict([(c, []) for c in scg_conn_types])
            host['port_ready_count'] = dict([(x, 0) for x in scg_conn_types])
            # TOO -yf- this code portion is to pass unit testing. WILL REMOVE
            # when we convert all old scg to new fabric_set settings
            if self.fabric_set_npiv and '*' in self.fabric_set_npiv:
                fabrics = self.fabric_list
            elif self.fabric_set_npiv:
                fabrics = self.fabric_set_npiv
            elif self.npiv_fabric == "A":
                fabrics = ['A']
            elif self.npiv_fabric == "B":
                fabrics = ['B']
            else:  # Cannot rule out ports by other npiv_fabric values yet
                fabrics = ['A', 'B']
            fabric_to_vios_map = dict((x, []) for x in fabrics)
            #######################
            # VIOS in host loop   #
            #######################
            for vios in host['vios_list']:
                db_vios = vios_map[vios['id']]
                if db_vios.state != 'running' or not\
                        db_vios.is_rmc_state_ok():
                    msg = WarnMsg(WarnMsg.VIOSBadState).getMsg(
                        context, dict(vios=vios['name']))
                    health['reasons'].append(msg)
                # Filter the ports
                vios['connectivity_types'] = deepcopy(scg_conn_types)

                filtered = self._filter_applicable_ports(vios, db_vios.fcports,
                                                         health, context)
                num_ports = len(filtered)
                LOG.debug("Length of db_vios.fcports = %d. The "
                          "filtered list length = %d." %
                          (len(db_vios.fcports), num_ports))
                if include_ports:
                    vios['fcport_list'] = filtered
                if _NPIV_ in vios['connectivity_types'] and \
                        health['effective_fabrics']:
                    for fabric, port_count in \
                            vios['fabrics_to_ports'].iteritems():
                        if port_count >= self.ports_per_fabric_npiv and \
                                vios['id'] not in fabric_to_vios_map[fabric]:
                            fabric_to_vios_map[fabric].append(vios['id'])

                vios['total_fcport_count'] = len(db_vios.fcports)
                vios['port_ready_count'] = num_ports  # remove?

                # Track whether this VIOS is storage-ready.
                if db_vios.state == 'running' and db_vios.is_rmc_state_ok()\
                        and num_ports > 0 and vios['connectivity_types']:
                    vios['storage_ready'] = True
                    host_vios_ready_count += 1
                    vio_conn_set = set(vios['connectivity_types'])
                else:
                    vios['storage_ready'] = False
                    vio_conn_set = set()

                # Update the VIOS ready counts on a boot/data basis.
                if vios['storage_ready'] and vio_conn_set.intersection(
                        host_boot_conn):
                    host_vios_boot_ready_count += 1
                    LOG.debug("Incremented host_vios_boot_ready_count for "
                              "%s: %d." %
                              (vios['id'], host_vios_boot_ready_count))
                if vios['storage_ready'] and vio_conn_set.intersection(
                        host_data_conn):
                    host_vios_data_ready_count += 1
                    LOG.debug("Incremented host_vios_data_ready_count for "
                              "%s: %d." %
                              (vios['id'], host_vios_data_ready_count))
                # Fill in the mapping of conn type to list of vioses.
                # If vios is not storage_ready, then vio_conn_set will
                # be empty so that the vios redundancy check later
                # evaluates correctly.
                for conn_type in vio_conn_set:
                    conn_type_map[conn_type].append(vios)
                # update the port counts
                for ctype, val in vios['port_ready_counts'].iteritems():
                    host['port_ready_count'][ctype] += val
            ###############################################
            # Determine and set computed host attributes  #
            ###############################################
            host_valid = all(len(vioses) > 0 for vioses in
                             fabric_to_vios_map.values())
            if (self.npiv_fabric == 'dual' and
                len(health['effective_fabrics']) < 2) or \
                    (self.fabric_set_req == 'per_host' and not host_valid):
                health['fabric_req_met'] = False
                remove_list = []
                for vios in host['vios_list']:
                    if _NPIV_ in vios['connectivity_types']:
                        host['port_ready_count'][_NPIV_] -= \
                            vios['port_ready_counts'][_NPIV_]
                        vios['port_ready_counts'][_NPIV_] = 0
                        remove_list.append(vios['name'])
                        vios['connectivity_types'].remove(_NPIV_)
                        vios_conn = set(vios['connectivity_types'])
                        if vios in conn_type_map[_NPIV_]:
                            conn_type_map[_NPIV_].remove(vios)
                        if vios['storage_ready']:
                            # Decrement the vios connectivity count for
                            # boot/data only if there is no longer any valid
                            # connectivity ie. VSCSI or SSP
                            if _NPIV_ in host_boot_conn and \
                                    not vios_conn.intersection(host_boot_conn):
                                host_vios_boot_ready_count -= 1
                                LOG.debug("Decremented "
                                          "host_vios_boot_ready_count "
                                          "for %s: %d." %
                                          (vios['id'],
                                           host_vios_boot_ready_count))
                            if _NPIV_ in host_data_conn and \
                                    not vios_conn.intersection(host_data_conn):
                                host_vios_data_ready_count -= 1
                                LOG.debug("Decremented "
                                          "host_vios_data_ready_count "
                                          "for %s: %d." %
                                          (vios['id'],
                                           host_vios_data_ready_count))

                        # Need to re-evaluate storage-readiness since now
                        if not vios['connectivity_types']:
                            vios['storage_ready'] = False
                            host_vios_ready_count -= 1
                if remove_list:
                    if self.npiv_fabric == 'dual':
                        msg = WarnMsg(WarnMsg.FabricsPerHostNotMet).getMsg(
                            context, dict(host=host['name'], num=2,
                                          vioses=remove_list))
                    elif self.fabric_set_req == 'per_host':
                        msg = WarnMsg(WarnMsg.FabricsSetPerHostNotMet).\
                            getMsg(context,
                                   dict(host=host['name'],
                                        fab_list=self.fabric_set_npiv,
                                        req_port=self.ports_per_fabric_npiv,
                                        vios_map=fabric_to_vios_map,
                                        vioses=remove_list))
                    LOG.debug(msg)
                    health['reasons'].append(msg)

            LOG.debug("Conn_type_map for host %s: %s." % (host['name'],
                                                          conn_type_map))
            # Check redundancy (vios multiplicity) requirement.
            # Redundancy needs to be checked per-connectivity-type.
            for conn_type, v_list in conn_type_map.iteritems():
                if len(v_list) < self.vios_redundancy:
                    msg = WarnMsg(WarnMsg.VIOSRedundancyNotMet).getMsg(
                        context, dict(host=host['name'], type=conn_type,
                                      req=self.vios_redundancy))
                    health['reasons'].append(msg)
                    # remove the connectivity type from the host tracked lists
                    if conn_type in host_boot_conn:
                        host_boot_conn.remove(conn_type)
                    if conn_type in host_data_conn:
                        host_data_conn.remove(conn_type)

            # Now check overall vios readiness for the host, for boot and data
            scg_dict['boot_vios_ready_count'] += host_vios_boot_ready_count
            scg_dict['data_vios_ready_count'] += host_vios_data_ready_count
            if host_vios_boot_ready_count >= self.vios_redundancy:
                health_status['vios_boot_redundancy_met'] = True
                health_status['connectivity'] = 'partial'
                host['boot_vios_req_met'] = True
            else:
                host['boot_vios_req_met'] = False
            if host_vios_data_ready_count >= self.vios_redundancy:
                health_status['vios_data_redundancy_met'] = True
                health_status['connectivity'] = 'partial'
                host['data_vios_req_met'] = True
            else:
                host['data_vios_req_met'] = False
            if health['fabric_req_met']:
                health_status['scg_fabric_req_met'] = True
            vol_types = []
            if not host_boot_conn:
                vol_types.append('boot')
            else:
                scg_dict['boot_host_ready_count'] += 1
            if not host_data_conn:
                vol_types.append('data')
            else:
                scg_dict['data_host_ready_count'] += 1
            if vol_types:
                msg = WarnMsg(WarnMsg.NoHostVIOSReady).getMsg(
                    context, dict(host=host['name'], bord="+".join(vol_types),
                                  req=self.vios_redundancy,
                                  boot=host_vios_boot_ready_count,
                                  data=host_vios_data_ready_count))
                health_status['value_reasons'].append(msg)

            # update the remaining host level counts
            host['vios_ready_count'] = host_vios_ready_count
            scg_dict['vios_ready_count'] += host_vios_ready_count
            if host_vios_boot_ready_count == len(host['vios_list']) and\
                    host_vios_data_ready_count == len(host['vios_list']):
                health_status['connectivity'] = 'full'
            health_status['member_status'][host['name']] = health
            # update effective connectivity for vioses
            effective_vios_conn = effective_vios_conn.union(
                *[v['connectivity_types'] for v in host['vios_list']])
            # update scg port counts
            for ctype, val in host['port_ready_count'].iteritems():
                scg_dict['port_ready_count'][ctype] += val
        # END host loop
        ###########################################
        # Set remaining SCG computed attributes   #
        ###########################################
        # Intersect the effective connectivity for the VIOS members with the
        # effective connectivity types implied by the applicable storage
        # providers to arrive at the overall effective connectivity of the
        # SCG. Right now this is just user information and is stuck in
        # health. Nothing else should depend on it.
        health_status['effective_connectivity'] =\
            effective_vios_conn.intersection(effective_conn_stg)
        scg_dict['health_status'] = health_status
        vol_types = []
        if not scg_dict['boot_host_ready_count']:
            health_status['boot_health'] = SCG_DOWN
            vol_types.append('boot')
        if not scg_dict['data_host_ready_count']:
            health_status['data_health'] = SCG_DOWN
            vol_types.append('data')
        if vol_types:
            if host_name:
                msg = WarnMsg(WarnMsg.NoConnectivityForHost).getMsg(
                    context, dict(host=host_name,
                                  vol_types="+".join(vol_types)))
            else:
                msg = WarnMsg(WarnMsg.NoConnectivityForSCG).getMsg(
                    context, dict(vol_types="+".join(vol_types)))
            health_status['value_reasons'].insert(0, msg)  # first message

        # check for usage error
        if host_name is not None and not host_name_verified:
            error = _("The provided hypervisor host name '%(host)s' is not a "
                      "member of the storage connectivity group "
                      "'%(name)s'") % dict(host=host_name,
                                           name=self.display_name)
            LOG.error(error)
            raise stgex.IBMPowerVCStorageError(error=error)
        return scg_dict

    def _filter_applicable_ports(self, vios, fcport_list, health, context):
        """
        Convert each FCPort to a dictionary, then
        filter the ports by enabled and port_tag and an
        existing wwpn, and optionally, by 'status'.
        """
        ports = []
        not_npiv_count = not_vscsi_count = 0
        filt_causes = {'npiv_disabled': [], 'status': [], 'tag': [],
                       'no_fabric': [], 'excluded_fabric': [],
                       'available_connections': [], 'vscsi_disabled': [],
                       'npiv_status': [], 'port_not_met': []}
        fabrics_to_ports = {}
        vios_fabric_req_met = True
        vios['port_ready_counts'] = dict([(x, 0) for x in
                                          vios['connectivity_types']])
        if _NPIV_ in vios['connectivity_types']:
            # fabric_set_npiv takes precedence, npiv_fabric should be 'any'
            if self.fabric_set_npiv and '*' in self.fabric_set_npiv:
                fabrics = self.fabric_list
            elif self.fabric_set_npiv:
                fabrics = self.fabric_set_npiv
            elif self.npiv_fabric == "A":
                fabrics = ['A']
            elif self.npiv_fabric == "B":
                fabrics = ['B']
            else:  # Cannot rule out ports by other npiv_fabric values yet
                fabrics = ['A', 'B']
            # initialize fabric count for ports per fabric seen
            for fabric in fabrics:
                fabrics_to_ports[fabric] = 0
        bad_fab = 0
        for fcport in fcport_list:
            port = fcport_to_dict(fcport)
            # The 'connectivity' attribute says what conn types an individual
            # port is applicable to.
            port['connectivity'] = set()
            npiv_applicable = False
            has_port_tag = True if 'port_tag' in port and port['port_tag']\
                else False
            tag_satisfied = True
            if self.port_tag and ((has_port_tag and self.port_tag !=
                                   port['port_tag']) or
                                  (not has_port_tag and not
                                   self.include_untagged)):
                tag_satisfied = False
                filt_causes['tag'].append(fcport['name'])
            if _NPIV_ in vios['connectivity_types']:
                npiv_applicable = True
                # check npiv filter cases
                if not port['enabled'] or (port['fc_protocol'] != 'any' and
                                           port['fc_protocol'] != 'npiv'):
                    npiv_applicable = False
                    filt_causes['npiv_disabled'].append(fcport['name'])
                if port['npiv_status'] != 'OK':
                    npiv_applicable = False
                    filt_causes['npiv_status'].append(fcport['name'])
                # Could use a 'buffer' variable to be checked against
                # instead of 0 to not max out the number of connections
                if not port['available_connections'] or\
                   port['available_connections'] <= 0:
                    npiv_applicable = False
                    filt_causes['available_connections'].append(fcport['name'])
                if not tag_satisfied:
                    npiv_applicable = False
                if port['fabric'] not in fabrics:
                    npiv_applicable = False
                    bad_fab += 1
                    if port['fabric'] is None or port['fabric'] == 'None':
                        filt_causes['no_fabric'].append(fcport['name'])
                    else:
                        filt_causes['excluded_fabric'].append(fcport['name'])
                # Track fabrics for other checks against npiv_fabric setting
                if npiv_applicable:
                    fabrics_to_ports[port['fabric']] += 1
                    vios['port_ready_counts'][_NPIV_] += 1
                    port['connectivity'].add(_NPIV_)
                else:
                    not_npiv_count += 1
            if _VSCSI_ in vios['connectivity_types']:
                vscsi_applicable = True
                if not port['enabled'] or (port['fc_protocol'] != 'any' and
                                           port['fc_protocol'] != 'vscsi'):
                    vscsi_applicable = False
                    filt_causes['vscsi_disabled'].append(fcport['name'])
                if not tag_satisfied:
                    vscsi_applicable = False
                if port['status'] != 'OK':
                    if port['fc_protocol'] != 'vscsi':
                        vscsi_applicable = False
                        filt_causes['status'].append(fcport['name'])
                    else:
                        LOG.info(_("Designating %(wwpn)s as a candidate vSCSI "
                                   "port since the fc_protocol is set "
                                   "specifically to 'vscsi'. Set to 'any' to "
                                   "have this port skipped.") %
                                 {'wwpn': port['wwpn']})

                if vscsi_applicable:
                    port['connectivity'].add(_VSCSI_)
                    vios['port_ready_counts'][_VSCSI_] += 1
                else:
                    not_vscsi_count += 1
            if _SSP_ in vios['connectivity_types']:
                port['connectivity'].add(_SSP_)
                vios['port_ready_counts'][_SSP_] += 1
            if port['connectivity']:
                ports.append(port)
            # Else the port was filtered out.

        # After filtering, we need to sort the ports by available_connections
        # So that the database retrieval of this list is consistent
        # with the live-data retrieval when determining connectivity
        # information.
        ports.sort(key=(lambda x: x['available_connections']),
                   reverse=True)

        if not ports:
            msg = WarnMsg(WarnMsg.NoUsablePorts).getMsg(
                context, dict(vios=vios['name']))
            health['reasons'].append(msg)
            vios['connectivity_types'] = []
        # If npiv is still in the vios connection types,
        # check for npiv_fabric->dual_fabric, or fabric_set_req->per_vios
        if _NPIV_ in vios['connectivity_types']:
            # return fabrics_to_ports count for npiv case
            vios['fabrics_to_ports'] = fabrics_to_ports
            if (bad_fab == len(fcport_list)):
                vios_fabric_req_met = False
            # The calculation for fab_to_ports_met is only applicable to
            # the requirements of per_vios or dual_per_vios
            fab_to_ports_met = dict((x, False)
                                    for x in fabrics_to_ports.keys())
            for fabric, ports_seen in fabrics_to_ports.iteritems():
                if ports_seen >= self.ports_per_fabric_npiv:
                    fab_to_ports_met[fabric] = True
            valid_fabrics = sum(fab_to_ports_met.values())
            fabrics_seen = \
                set(fabric for fabric, ports_n in
                    fabrics_to_ports.iteritems() if ports_n > 0)
            # overwrite fabrics_seen if 'per_host', since
            # it is undetermined yet if all vios has valid connections
            # to which fabric, so instead we will overwrite and only update the
            # 'effective_fabrics' with valid fabrics.
            if self.fabric_set_req == 'per_host':
                fabrics_seen = set([fabric for fabric, met
                                    in fab_to_ports_met.iteritems() if met])
            # any fabrics seen will be added here, and will be removed later
            # if it is determine the fabric does not meet all requirements
            health['effective_fabrics'] =\
                health['effective_fabrics'].union(fabrics_seen)
            # if the ports from the vios does not meet the
            # 'ports_per_fabric_npiv' from all fabrics then no ports will be
            # returned to the caller, signify that this vios is invalid
            if self.fabric_set_req == 'per_vios' and \
                    valid_fabrics < len(fabrics):
                vios['connectivity_types'].remove(_NPIV_)
                vios_fabric_req_met = False
                msg = WarnMsg(WarnMsg.FabricsSetPerVIOSNotMet).getMsg(
                    context, dict(vios=vios['name'], fab_list=fabrics,
                                  port_num=self.ports_per_fabric_npiv,
                                  port_map=fabrics_to_ports))
                health['reasons'].append(msg)
                LOG.debug("Clearing the ports for vios: %s, reason: %s" %
                          (vios['name'], msg))
                for fabric, met in fab_to_ports_met.iteritems():
                    if not met and fabric in health['effective_fabrics']:
                        health['effective_fabrics'].discard(fabric)
                vios['port_ready_counts'][_NPIV_] = 0
            # 'per_host', 'at_least_one', 'exactly_one' setting validates
            # the vios only if the required
            # 'ports_per_fabric_npiv' is met for at least one fabric.
            elif self.fabric_set_req in \
                    ['per_host', 'at_least_one', 'exactly_one']:
                # clear ports that did not meet
                # the ports_per_fabric_npiv requirement.
                for fabric, met in fab_to_ports_met.iteritems():
                    if not met:
                        vios['port_ready_counts'][_NPIV_] -= \
                            fabrics_to_ports[fabric]
                        ex_ports = self._remove_invalid_ports(ports, fabric)
                        filt_causes['port_not_met'].extend(ex_ports)
                        not_npiv_count += len(ex_ports)
                if valid_fabrics < 1:
                    vios['connectivity_types'].remove(_NPIV_)
                    vios_fabric_req_met = False
                    msg = WarnMsg(WarnMsg.FabricsSetPerVIOSNotMet).getMsg(
                        context, dict(vios=vios['name'], fab_list=fabrics,
                                      port_num=self.ports_per_fabric_npiv,
                                      port_map=fabrics_to_ports))
                    health['reasons'].append(msg)
                    LOG.debug("Clearing the ports for vios: %s, reason: %s" %
                              (vios['name'], msg))
                    vios['port_ready_counts'][_NPIV_] = 0

            if self.npiv_fabric == 'dual_per_vios' and valid_fabrics < 2:
                vios['connectivity_types'].remove(_NPIV_)
                msg = WarnMsg(WarnMsg.FabricsPerVIOSNotMet).getMsg(
                    context, dict(vios=vios['name'], num=2))
                health['reasons'].append(msg)
                vios_fabric_req_met = False
        # Check validity of _VSCSI_ connectivity on vios
        if _VSCSI_ in vios['connectivity_types']:
            if vios['port_ready_counts'][_VSCSI_] == 0:
                vios['connectivity_types'].remove(_VSCSI_)
        # Handle port health message construction
        if not_vscsi_count:
            pmsg_1 = pmsg_2 = pmsg_3 = ""
            if filt_causes['vscsi_disabled']:
                pmsg_1 = ("FC Ports disabled for vSCSI attachment: %s. ") %\
                    filt_causes['vscsi_disabled']
            if filt_causes['tag']:
                pmsg_2 = ("FC Ports without matching port tag: %s. ") %\
                    filt_causes['tag']
            if filt_causes['status']:
                pmsg_3 = ("FC Ports without OK status: %s. ") %\
                    filt_causes['status']
            pmsg = _("%s %s %s") % (pmsg_1, pmsg_2, pmsg_3)
            msg = WarnMsg(WarnMsg.NonConnFCPorts).getMsg(
                context, dict(vios=vios['name'], count=not_vscsi_count,
                              total=len(fcport_list), conn_type=_VSCSI_,
                              reasons=pmsg))
            health['port_health'].append(msg)
            if not_vscsi_count == len(fcport_list):
                if _VSCSI_ in vios['connectivity_types']:
                    vios['connectivity_types'].remove(_VSCSI_)
                LOG.debug("VIOS %s does not have vSCSI connectivity." %
                          vios['name'])
        if not_npiv_count:
            # Create a detailed npiv port exclusion message.
            pmsg_1 = pmsg_2 = pmsg_3 = pmsg_4 = pmsg_5 = pmsg_6 = pmsg_7 = ""
            if filt_causes['npiv_disabled']:
                pmsg_1 = ("FC Ports disabled for NPIV attachment: %s. ") %\
                    filt_causes['npiv_disabled']
            if filt_causes['npiv_status']:
                pmsg_2 = ("FC Ports without OK NPIV status: %s. ") %\
                    filt_causes['npiv_status']
            if filt_causes['available_connections']:
                pmsg_3 = ("FC Ports without available connections: %s. ") %\
                    filt_causes['available_connections']
            if filt_causes['tag']:
                pmsg_4 = ("FC Ports without matching port tag: %s. ") %\
                    filt_causes['tag']
            if filt_causes['no_fabric']:
                pmsg_5 = ("FC Ports not associated with a registered fabric:"
                          " %s. ") % filt_causes['no_fabric']
            if filt_causes['excluded_fabric']:
                msg = ("FC Ports excluded by fabric requirement "
                       "'%(npiv_fabric)s': %(ex_ports)s. ")
                pmsg_6 = (msg % dict(npiv_fabric=self.fabric_set_req,
                                     ex_ports=filt_causes['excluded_fabric']))
            if filt_causes["port_not_met"]:
                pmsg_7 = ("FC Ports excluded by requirement of "
                          "ports_per_fabric_npiv:%(port_req)d not met."
                          "fc_ports:%(ex_ports)s") %\
                    dict(port_req=self.ports_per_fabric_npiv,
                         ex_ports=filt_causes["port_not_met"])
            pmsg = _("%s %s %s %s %s %s %s") %\
                (pmsg_1, pmsg_2, pmsg_3, pmsg_4, pmsg_5, pmsg_6, pmsg_7)
            msg = WarnMsg(WarnMsg.NonConnFCPorts).getMsg(
                context, dict(vios=vios['name'], count=not_npiv_count,
                              total=len(fcport_list), conn_type=_NPIV_,
                              reasons=pmsg))
            health['port_health'].append(msg)
            if not_npiv_count == len(fcport_list):
                if _NPIV_ in vios['connectivity_types']:
                    vios['connectivity_types'].remove(_NPIV_)
                LOG.debug("VIOS %s does not have NPIV connectivity." %
                          vios['name'])

        # Do we need to remove port['connectivity'] and counts too?
        LOG.debug("Vios-conn=%s. bad_fab=%d, Sorted and filtered ports: %s." %
                  (vios['connectivity_types'], bad_fab, ports))
        if vios_fabric_req_met:
            health['fabric_req_met'] = True
        return ports

    def _remove_invalid_ports(self, ports, fabric):
        """
        helper method to remove any ports that maps to the given
        fabric. If it only has NPIV in its connectivity.
        """
        not_npiv_port = []
        for port in ports[:]:
            if port['fabric'] == fabric:
                if _NPIV_ in port['connectivity']:
                    port['connectivity'].remove(_NPIV_)
                    if len(port['connectivity']) == 0:
                        not_npiv_port.append(port['name'])
                        ports.remove(port)
        return not_npiv_port

    def get_connectivity_types(self):
        """
        :returns: A set of connectivity types based on the union of
                  boot_connectivity and data_connectivity values for this
                  SCG.
        """
        types = set(self.boot_connectivity) | set(self.data_connectivity)
        return types

    def replace_vios_list(self, context, transaction, new_viosids,
                          force=False):
        """
        Replace the SCG 'vioservers' field with the VIOSes for the passed
        list of IDs.
        This method also performs validations on the list, however there
        is no longer a need to validate the ID format itself as it has
        switched to be an HMC-based UUID. This
        method can be called with the SCG's current list to validate it.
        """
        host_map, max_len = ({}, 0)
        new_vios_doms = []
        if len(new_viosids) == 0 and not force:
            m = _("At least one VIOS ID must be specified for a storage "
                  "connectivity group.")
            raise stgex.IBMPowerVCStorageError(error=m)

        vios_factory = VioServerFactory.get_factory()
        # Now loop over the map of incoming items to look up VIOS DOMs and
        # do vios.id and vios.fcports validation.
        db_vioses = vios_factory.find_all_vioses(
            context, transaction=transaction)
        db_vios_by_id = dict([(db_v.id, db_v) for db_v in db_vioses])
        LOG.debug("db_vios_by_id=%s" % db_vios_by_id)
        # total_fabrics = set()
        for v_id in new_viosids:
            if v_id not in db_vios_by_id:
                raise stgex.IBMPowerVCViosNotFound(v_id=v_id)
            if self.cluster_provider_name and\
                    db_vios_by_id[v_id].cluster_provider_name !=\
                    self.cluster_provider_name:
                m = _("The member VIOS partition %(vios)s specified in "
                      "the storage connectivity group is not a member of "
                      "the VIOS cluster for storage provider %(prov)s. "
                      "Either choose an appropriate VIOS or the correct "
                      "shared storage pool provider for the operation.")\
                    % dict(vios=v_id, prov=self.cluster_provider_name)
                raise stgex.IBMPowerVCSCGConflict(error=m)
            # If this is the first VIOS for that Host, we need to initialize
            if db_vios_by_id[v_id].host_name not in host_map:
                host_map[db_vios_by_id[v_id].host_name] = list()
            host_map[db_vios_by_id[v_id].host_name].append(v_id)
            # We want to figure which Host has the most VIOS's for this SCG
            if len(host_map[db_vios_by_id[v_id].host_name]) > max_len:
                max_len = len(host_map[db_vios_by_id[v_id].host_name])
            LOG.debug("Appending VIOS dom to list with id %s." %
                      db_vios_by_id[v_id].id)
            new_vios_doms.append(db_vios_by_id[v_id])

        if self.vios_redundancy > max_len and not force:
            m = _("No host has sufficient VIOS members in this connectivity "
                  "group to fulfill the redundancy requirement of %(val)s. "
                  "Change the VIOS membership or lower the redundancy "
                  "requirement and try the request again.") %\
                dict(val=self.vios_redundancy)
            raise stgex.IBMPowerVCSCGConflict(error=m)
        # Update the DOM property if needed
        if set(new_viosids) != set(self.vios_ids):
            LOG.debug("Updating SCG VIOS list for %s." % self.display_name)
            self.vioservers = new_vios_doms
        else:
            LOG.debug("VIOS members the same. Update not needed for "
                      "%s." % self.display_name)

    def get_connectivity(self, context, host_name, live_data=False,
                         chk_conn_type=CHK_EITHER):
        """
        Return deploy/attach-time storage connectivity information, given this
        storage connectivity group and a host name chosen.
        to_dict_with_conn() is used to do a lot of connectivity evaluation
        against the collected topology. This flow does extra processing
        needed for the attach like (1) collecting npiv ports on a per-fabric
        basis and (2) choosing VIOSs for connectivity info inclusion when
        there are more available than will be used for virtual connections
        (exact_redundancy=True in that case).

        :param context:   An authorization context.
        :param host_name: The host to constrain the VIOS candidates to.
                          This is the registered compute host name.
        :param live_data: Starting in 1.3, this param is ignored.
        :returns:         See get_connectivity_info() method in
                          StorageMapping()-->get_connectivity_info() for
                          details on this structure of the returned
                          connectivity information.
                          Raises a IBMPowerVCStorageError exception if none
                          of the SCG-member VIOSes allow storage
                          connectivity with current topology data.
        """
        # We introduce a thread local cache here to avoid SCG related
        # computations multiple times during the same deploy request or
        # other operation. We also need to match on the host_name and
        # chk_conn_type params passed.
        cache_conn_type = chk_conn_type if chk_conn_type else 'none'
        req_id = conn_infos = None
        req_context = getattr(_request_store, 'context', None)
        if req_context:
            req_id = getattr(req_context, 'request_id', None)
            if req_id and getattr(_request_store, 'conn_infos', None):
                conn_infos = getattr(_request_store, 'conn_infos', None)
                if self.id in conn_infos:
                    cache_ci = conn_infos[self.id]
                    if (cache_ci['request_id'] == req_id and
                            cache_ci['target_host'] == host_name and
                            cache_ci['chk_conn_type'] == cache_conn_type):
                        LOG.debug(
                            "Return already computed conn_info for scg_id=%s, "
                            "req_id=%s, host=%s, chk_conn_type=%s." %
                            (self.id, req_id, host_name, cache_conn_type))
                        return cache_ci
                    else:
                        LOG.debug("Cache mismatch. Cached tuple is (%s, %s, "
                                  "%s), incoming request for (%s, %s, %s)." %
                                  (cache_ci['request_id'],
                                   cache_ci['target_host'],
                                   cache_ci['chk_conn_type'],
                                   req_id, host_name, cache_conn_type))
                else:
                    LOG.debug("No conn_info cached for SCG %s on threadlocal."
                              "Num set=%d." % (self.id, len(conn_infos)))
            else:
                LOG.debug("No conn_infos stored yet. In: (%s, %s, %s, %s)" %
                          (self.id, req_id, host_name, cache_conn_type))
        else:
            LOG.debug("No request context on thread. Skip cache lookup.")

        # We do not verify the 'enabled' state of the SCG here. We
        # assume the enabled state is filtered on when choosing an
        # SCG for deployment, but does not restrict operations later.
        scg_name = self.display_name  # for logging
        # Construct a dictionary representation of this SCG with all
        # the connectivity processing & fields added for the provided host.
        scg_dict = self.to_dict_with_conn(context, host_name=host_name,
                                          skip_provider_check=True)
        LOG.debug("chk_conn_type=%s, SCG_DATA: to_dict_with_conn(host=%s) "
                  "returns: %s." % (chk_conn_type, host_name, scg_dict))
        effective_conn = scg_dict['health_status']['effective_connectivity']

        data_conn_types = list(set(self.data_connectivity) & effective_conn)
        boot_conn_types = list(set(self.boot_connectivity) & effective_conn)
        # when CHK_BOOT or CHK_DATA is set we return the connectivity type
        # as the effective boot or data conn
        if chk_conn_type == CHK_BOOT:
            effective_conn = boot_conn_types
        elif chk_conn_type == CHK_DATA:
            effective_conn = data_conn_types
        else:
            # effective_conn is a set, which looks like a list when converted
            # to json for the REST APIs, but here we need to specifically
            # convert to a list to keep our contract with the
            # 'connectivity_types' value.
            effective_conn = list(effective_conn) if\
                effective_conn is not None else []
        fabric_num_required = None
        if _NPIV_ in effective_conn:
            if 'dynamic_fabric_set' in scg_dict:
                fabric_num_required = len(scg_dict['dynamic_fabric_set'])
            elif 'fabric_set_npiv' in scg_dict:
                fabric_num_required = len(scg_dict['fabric_set_npiv'])
            else:
                if _NPIV_ in boot_conn_types or _NPIV_ in data_conn_types:
                    error = _("Required fabric set property is missing from "
                              "the storage connectivity group: %(scg)s") %\
                        {'scg': scg_name}
                    raise stgex.IBMPowerVCStorageError(error=error)
            if self.fabric_set_req == "exactly_one" or\
                    self.fabric_set_req == 'at_least_one':
                LOG.debug("fabric_num_required = 1 (from %d)." %
                          fabric_num_required)
                fabric_num_required = 1

        # Initialize the return data dictionary.
        conn_info = {'connectivity_types': effective_conn,
                     'data_conn_types': data_conn_types,
                     'boot_conn_types': boot_conn_types,
                     'scg_name': scg_name, 'scg': self,
                     'target_host': host_name, 'vios_list': [],
                     'health_status': scg_dict['health_status'],
                     'min_vioses_per_vol': self.vios_redundancy,
                     'backend_host_group': self.backend_host_group,
                     'fabric_num_required': fabric_num_required,
                     'request_id': req_id, 'chk_conn_type': cache_conn_type}

        # Check connectivity health first.
        if chk_conn_type and\
           ((chk_conn_type == CHK_BOOT and
            scg_dict['health_status']['boot_health'] != "OK") or
            (chk_conn_type == CHK_DATA and
            scg_dict['health_status']['data_health'] != "OK") or
            (chk_conn_type == CHK_ALL and (
                scg_dict['health_status']['boot_health'] != "OK" or
                scg_dict['health_status']['data_health'] != "OK")) or
            (chk_conn_type == CHK_EITHER and (
                scg_dict['health_status']['boot_health'] != "OK" and
                scg_dict['health_status']['data_health'] != "OK"))):
            error = _("The storage connectivity group '%(scg)s' associated "
                      "with this request does not evaluate to satisfied "
                      "connectivity for host %(host)s. The health data is: "
                      "%(health)s.") % dict(scg=scg_name, host=host_name,
                                            health=scg_dict['health_status'])
            LOG.error(error)
            error = _("The storage connectivity group '%(scg)s' associated "
                      "with this request does not evaluate to satisfied "
                      "connectivity for host %(host)s. The health data is: "
                      "boot_heath=%(boot)s, data_health=%(data)s.") % \
                dict(scg=scg_name, host=host_name,
                     boot=scg_dict['health_status']['boot_health'],
                     data=scg_dict['health_status']['data_health'])
            raise stgex.IBMPowerVCStorageError(error=error)

        # Retrieve the member VIOSs list within the output. There should
        # be at least one if health check above passed.
        scg_vioses = scg_dict['host_list'][0]['vios_list']
        orig_num_v = len(scg_vioses)
        # For conn_info output, we weed out the vioses that are not
        # storage-ready.
        scg_vioses = [v for v in scg_vioses if v['storage_ready']]
        LOG.debug("There are %d of %d storage-ready VIOSs." % (len(scg_vioses),
                                                               orig_num_v))
        if live_data:
            LOG.debug("In-flow attach host-storage-port topology collection "
                      "and checking will be skipped (ignore live_data).")

        # TODO: If live topology remains out of flow, then some additional
        #       refactoring may be done to consolidate processing below into
        #       to_dict_with_conn.
        # Construct map of connectivity type to list of vios entries to be
        # filled in by _add_vios_conn_info()
        conn_type_map = dict([(c, []) for c in effective_conn])
        # Call a private helper method to add the connectivity information.
        self._add_vios_conn_info(conn_info, scg_vioses, conn_type_map)

        # Check for the redundancy setting. If true, then we
        # need to trim down the number of VIOSes in conn_info to match
        # the 'exact' number requested.
        self._check_redundancy(conn_info, self.vios_redundancy,
                               conn_type_map)

        msg = ("Returning storage connectivity information for host "
               "%(host)s and SCG '%(scg)s' with %(num_vioses)d storage-"
               "ready member VIOSs. Allowed VIOS connectivity:") %\
            dict(host=host_name, scg=self.display_name, num_vioses=orig_num_v)
        # Now loop over the vioses left in conn_info. These are the ones
        # we expect to get used for the attachment operation in progress.
        # We will update the in-memory attach-time-stamp associated with
        # these vioses so they can be sorted correctly SCGs that need an
        # exact redundancy number in the future.
        st = HostStorageTopologyFactory.theFactory.get_static_topology()
        for vios in conn_info['vios_list']:
            st.set_attach_candidate_time(vios['id'])
            v_msg = (" VIOS '%(vios)s' with state '%(state)s' and RMC state "
                     "'%(rmc)s' has connectivity types %(conn)s.") %\
                dict(vios=vios['name'], state=vios['state'],
                     rmc=vios['rmc_state'], conn=vios['connectivity_types'])
            msg = six.text_type(msg) + six.text_type(v_msg)
        LOG.info(msg)
        # have to six.text_type scg class within conn_info to fix encode error
        scg_class = conn_info.pop('scg')
        LOG.debug("connectivity_info for SCG '%s': %s."
                  % (scg_name, conn_info))
        conn_info['scg'] = scg_class
        # cache connectivity info if we have a real request context
        if req_id:
            if conn_infos:
                conn_infos[self.id] = conn_info
            else:
                conn_infos = {self.id: conn_info}
                setattr(_request_store, 'conn_infos', conn_infos)
            LOG.debug("Cached conn_info for (%s, %s, %s, %s)." %
                      (self.id, req_id, host_name, cache_conn_type))
        # return the data
        return conn_info

    def _check_redundancy(self, conn_info, num, conn_type_map):
        """
        Trim the number of vioses in conn_info down (if needed) to 'num'
        on a per connectivity type basis. Let's say VIOSs 1 - 3 have npiv
        connectivity and VIOSs 2 - 4 have SSP. Then we need to remove 'npiv'
        from one VIOS and 'ssp' from another (possibly the same).
        The way VIOSs are sorted (per connection type) is an implementation
        detail of this method. NPIV VIOSs can be sorted according to how
        their ports are sorted - sort at the top those the VIOS with the most
        available connections, summing across its usable ports. Other
        connectivity types could sort VIOS to the top with the least virtual
        connections uses for storage adapters. Or it could just round-robin
        through them by maintaining state. Or it could be random.
        """
        LOG.debug("check_redundancy num=%d. Check list." % num)
        for conn_type, v_list in conn_type_map.iteritems():
            if len(v_list) > num and self.exact_redundancy:
                if conn_type == _NPIV_:
                    v_list.sort(key=(lambda v: v['npiv_sort_field']),
                                reverse=True)
                else:
                    v_list.sort(key=(lambda v: v['other_sort_field']),
                                reverse=False)
                LOG.debug("Trimming for '%s'. Sorted vios_list=%s." %
                          (conn_type, v_list))
                # Now the list is sorted with the first num items to keep
                remove_list = v_list[num:]
                for v in remove_list:
                    if conn_type in v['connectivity_types']:
                        v['connectivity_types'].remove(conn_type)
                        LOG.debug("Remove '%s' from vios %s since excluded "
                                  "by exact_redundancy." %
                                  (conn_type, v['id']))
                    else:
                        LOG.debug("Connectivity type '%s' already gone from "
                                  "vios %s." % (conn_type, v['id']))
                    if not v['connectivity_types'] and 'removed' not in v:
                        LOG.debug("Removed last connection type from "
                                  "VIOS %s. Remove vios entry from "
                                  "conn_info." % v['id'])
                        v['removed'] = True
                        conn_info['vios_list'].remove(v)
                LOG.debug("VIOSes kept for '%s': %s." %
                          (conn_type, [i['id'] for i in v_list[:num]]))
            elif len(v_list) == num or\
                    (len(v_list) > num and not self.exact_redundancy):
                LOG.debug("redundancy=%s for %d already satisfied by "
                          "vios_list for type '%s'." %
                          (self.exact_redundancy, self.vios_redundancy,
                           conn_type))
            else:
                # This condition should have been caught earlier.
                msg = _("The minimum VIOS redundancy value of %(val)d is "
                        "not satisfied by the storage connectivity group "
                        "%(scg)s for the host and connectivity type "
                        "'%(type)s'.") %\
                    dict(val=self.vios_redundancy, scg=self.display_name,
                         type=conn_type)
                LOG.info(msg)
                raise stgex.IBMPowerVCStorageError(error=msg)

    def _add_vios_conn_info(self, conn_info, scg_vioses, conn_type_map):
        """
        Private method to add connection information to conn_info
        structure passed. The VIOSes that are members of the SCG and
        the chosen host are passed in and are assumed to be current so they
        will not be looked up again from the SCG.
        This method deletes the 'fcport_list' dictionary entry from
        scg_vioses and replaces it with lists split out by fabric.
        """
        conn_info['vios_list'] = []  # initialize
        st = HostStorageTopologyFactory.theFactory.get_static_topology()
        for vios_dict in scg_vioses:
            remove_conn_type = []
            for conn_type in vios_dict['connectivity_types']:
                # the vios dict might have the conn_type but
                # if the connectivity_type does not have
                # then we need not add that info
                if conn_type not in conn_type_map:
                    LOG.warn("Connection type %s is not in effective"
                             " connectivity" % conn_type)
                    continue
                conn_type_map[conn_type].append(vios_dict)
                LOG.debug("Added vios %s to conn_type %s." %
                          (vios_dict['id'], conn_type))
                if conn_type == _NPIV_:
                    vios_dict['ports'], vios_dict['npiv_fab_order'] =\
                        self._vios_ports_by_fabric(vios_dict)
                elif conn_type == _VSCSI_:
                    vios_dict["vscsi:ports"] =\
                        self._get_vios_ports_for_vscsi(vios_dict)
                if vios_dict['rmc_state'] == 'busy':
                    # The other_sort_field value is set to a future time
                    # so that it sorts last. Note that this field is not
                    # currently used for NPIV vios sorting.
                    vios_dict['other_sort_field'] = time.time() + 100
                else:
                    vios_dict['other_sort_field'] =\
                        st.get_attach_candidate_time(vios_dict['id'])

                if ("ports" in vios_dict and
                    not vios_dict["ports"]) and _NPIV_ in\
                   vios_dict['connectivity_types']:
                    remove_conn_type.append(_NPIV_)
                    conn_type_map[_NPIV_].remove(vios_dict)
                elif ("vscsi:ports" in vios_dict and not
                      vios_dict["vscsi:ports"]) and _VSCSI_ in\
                        vios_dict['connectivity_types']:
                    remove_conn_type.append(_VSCSI_)
                    conn_type_map[_VSCSI_].remove(vios_dict)

            if remove_conn_type:
                for conn_type in remove_conn_type:
                    vios_dict['connectivity_types'].remove(conn_type)
            if 'fcport_list' in vios_dict:
                del vios_dict['fcport_list']
            if vios_dict['connectivity_types']:
                conn_info['vios_list'].append(vios_dict)

        # Return the connectivity structure
        return conn_info

    def _get_vios_ports_for_vscsi(self, vios_dict):
        """
        Given a vios dictionary containing a fcport_list of Fibre
        Channel port information, return a dictionary of the ports
        for vSCSI
        """
        if not vios_dict['fcport_list']:
            LOG.info(_("Virtual I/O Server '%(name)s' has no applicable FC "
                       "Ports for the storage connectivity group '%(scg)s'. "
                       "Skipping it.") %
                     dict(name=vios_dict['name'], scg=self.display_name))
            return []

        wwpns = set()
        vscsi_ports = []
        for port in vios_dict['fcport_list']:
            if port['wwpn'] in wwpns:
                LOG.warn(_("FC Port with WWPN already seen for Virtual I/O"
                           " Servers. Skipping FC Port: %s.") % port)
                continue
            if _VSCSI_ not in port['connectivity']:
                # Exclude non-vscsi port from output
                LOG.debug(_("Exclude non-vscsi port from output: %s") % port)
                continue
            wwpns.add(port['wwpn'])
            vscsi_ports.append(port)
        # end for each db_port
        return vscsi_ports

    def _vios_ports_by_fabric(self, vios_dict):
        """
        Given a vios dictionary containing a fcport_list of Fibre
        Channel port information, return a dictionary mapping of the ports
        split out by their respective fabrics, as required by the
        get_connectivity_info() API. It is expected that the "fcport_list"
        coming in is already filtered for the specific SCG DOM passed
        (i.e. If the SCG only allows NPIV connectivity, then vios_dict will
        not contain disabled ports or ports with tags not matching the SCG
        port_tag or ports without a registered fabric mapping, or ports
        connected to a fabric not allowed by the SCG), AND already sorted by
        'available_connections', whatever semantic that field is configured
        to represent.

        This method does extra filtering and sorting by unique adapters.
        Prefering unique adapters and choosing more than one port per fabric
        is new for 1.3.

        An assumption of this method is that we will not find ports that
        are not intended to be used in the connectivity solution. For example,
        in the "per_host" requirement case, if a vios does not satisfy the
        fabric/port needs (but another one does), the one that does not
        should have its NPIV ports pruned by the to_dict_with_conn()
        processing done earlier.

        :returns: dictionary indexed by fabric with values being lists of
                  sorted ports. Secondly, returns a list of fabrics ordered
                  by preference of use - i.e. least port utilization. This
                  order list is valid for the VIOS alone.
        """
        ret_dict = {}

        if not vios_dict['fcport_list']:
            LOG.info(_("Virtual I/O Server '%(name)s' has no applicable FC "
                       "Ports for the storage connectivity group '%(scg)s'. "
                       "Skipping it.") %
                     dict(name=vios_dict['name'], scg=self.display_name))
            return {}

        wwpns = set()
        vios_dict['npiv_sort_field'] = 0
        fab_connections = []
        LOG.debug("Sort ports by fabric and sort fabrics. Adapter pref=%s"
                  % self.prefer_unique_adapter)
        # The ports are already sorted by available connections.
        # However we need to do second sort that is by unique adapter.
        # So create a new sorted data structure: by_fabric_by_adapter.
        # Example data:
        #  [('A', [('adpt_2', [{port_dict1}, {port_dict2}]),
        #          ('adpt_1', [{port_dict3}])]),
        #   ('B', [('adpt_0', [{port_dict4}]), ('adpt_1', [{port_dict5}])])]
        by_fab_by_adpt = []
        for port in vios_dict['fcport_list']:
            if _NPIV_ not in port['connectivity']:
                # Port should not be used for NPIV connectivity.
                LOG.debug("Exclude non-npiv port from output: %s" % port)
                continue
            if port['wwpn'] in wwpns:
                LOG.warn(_("FC Port with WWPN already seen for Virtual I/O"
                           " Servers. Skipping FC Port: %s.") % port)
                continue
            wwpns.add(port['wwpn'])

            # Collect the fabrics already seen
            fabs = [x[0] for x in by_fab_by_adpt]
            if port['fabric'] in fabs:
                # extract tuple already set
                f_tuple = by_fab_by_adpt[fabs.index(port['fabric'])]
            else:  # create a new tuple
                f_tuple = (port['fabric'], [])
                by_fab_by_adpt.append(f_tuple)
            if self.prefer_unique_adapter:
                adpt = port['adapter_id']
            else:
                # If non-standard config, then we use a dummy string for the
                # adapter ID so that all fabric ports sort under it.
                adpt = 'single_group'
            # Collect the adapter ids from the list assocated with the fabric
            adapters = [x[0] for x in f_tuple[1]]
            # If this port's adapter is in the list then pull out its tuple
            if adpt in adapters:
                idx = adapters.index(adpt)
                a_tuple = f_tuple[1][idx]
            else:  # otherwise, create a fresh adapter tuple
                a_tuple = (adpt, [])
                f_tuple[1].append(a_tuple)
            a_tuple[1].append(port)  # associate the port with the adapter

        LOG.debug("by_fab_by_adpt=%s" % by_fab_by_adpt)
        # Now ports are split by fabric and adapter and sorted within each.
        # Build the return dictionary by placing the ports into lists
        # split out by fabric. We prefer unique adapters.
        for f_tuple in by_fab_by_adpt:
            fabric = f_tuple[0]
            ports = []
            adpt_list = list(f_tuple[1])
            while adpt_list:
                for a_tuple in f_tuple[1]:
                    if a_tuple[1]:
                        ports.append(a_tuple[1].pop(0))
                # trim the adapter list of any empty adapter tuples
                adpt_list = [a for a in f_tuple[1] if a[1]]
            ret_dict[fabric] = ports

            # For sorting fabrics within the VIOS, we will sum the
            # available_connections across the first ports_per_fabric_npiv
            # (cardinality requirement) list of ports for the fabric.
            sum_avail = sum([p['available_connections'] for p in
                             ports[:self.ports_per_fabric_npiv]])
            fab_connections.append((fabric, sum_avail))
            # For sorting VIOSs, we want to use the computed sum above
            # and add it to the sort field for the VIOS.
            vios_dict['npiv_sort_field'] += sum_avail

        # Sort the fabrics
        sorted_fabs = sorted(fab_connections, key=lambda x: x[1], reverse=True)
        sorted_fabs = [x[0] for x in sorted_fabs]
        # Return the main dict plus the sorted fabric list.
        LOG.debug("Returning %s: sorted_fabs=%s, by-fabric-ports=%s" %
                  (vios_dict['name'], sorted_fabs, ret_dict))
        return ret_dict, sorted_fabs

    def verify_host_connectivity(self, context, host_name=None, msg_dict=None,
                                 conn_type=CHK_EITHER, stg_provider_lst=None,
                                 operation_type=None):
        """
        Check that the passed SCG has any storage connectivity using
        topology and metadata information currently in the PVC DB.
        Do not go to K2 synchronously for this check.
        :param scg: The storage connectivity group to check.
        :param host_name: Optional host to constrain the checking to,
                    i.e. check just this host
        :param msg_dict: Optional dictionary with key 'messages' containing
                         a list of informational or warning messages that
                         can be appended to.
        :param conn_type: Optional. If CHK_BOOT is passed we check for valid
                        boot connectivity and if CHK_DATA is passed we check
                        for valid data connectivity. If CHK_ALL then both
                        boot or data connectivity should be valid. If
                        CHK_EITHER then either should be valid.
        :param stg_provider_lst: Optional.If passed in the validity of the SCG
                        for the given list of stg providers are verified.
        :returns: A list of hosts with at least one VIOS member of the SCG
                  that has at least one connectivity type. An empty list
                  means no connectivity.
        """

        def _check_prov_conn(storage_prov, connectivity):
            if (storage_prov['stg_type'] == 'fc' and
                (_NPIV_ not in connectivity and
                 _VSCSI_ not in connectivity))\
                    or (self.cluster_provider_name and
                        _SSP_ not in connectivity):
                return False
            else:
                return True

        t1 = time.time()
        hosts_with_conn = []
        hosts_wout = []
        scg_dict = self.to_dict_with_conn(context, host_name=host_name,
                                          include_ports=False,
                                          skip_provider_check=True)
        LOG.debug("conn_type=%s, operation_type=%s, stg_provider_lst=%s, "
                  "to_dict_with_conn(host=%s) delta1 seconds=%.4f returns: "
                  "<next_line>\n%s." %
                  (conn_type, operation_type, stg_provider_lst, host_name,
                   (time.time() - t1), scg_dict))

        if operation_type == op_types.FILTER_WORKING_SCOPE_RESIZE:
            # We are not changing hosts for a resize, so skip host filtering
            # except for SCG membership.
            hosts_with_conn = [host['name'] for host in scg_dict['host_list']]
            return hosts_with_conn

        if (conn_type == CHK_BOOT and
            scg_dict['health_status']['boot_health'] != "OK") or\
           (conn_type == CHK_DATA and
            scg_dict['health_status']['data_health'] != "OK") or\
            (conn_type == CHK_ALL and (
                scg_dict['health_status']['boot_health'] != "OK" or
                scg_dict['health_status']['data_health'] != "OK")) or\
           (conn_type == CHK_EITHER and (
                scg_dict['health_status']['boot_health'] != "OK" and
                scg_dict['health_status']['data_health'] != "OK")):
            msg = _("The storage connectivity group '%(scg)s' is not currently"
                    " available to any host for storage connectivity for "
                    "'%(type)s'. The health information is: %(health)s") %\
                dict(scg=self.display_name,
                     type=conn_type,
                     health=scg_dict['health_status'])
            LOG.info(msg)
            if msg_dict:
                msg_dict['messages'].extend(
                    scg_dict['health_status']['value_reasons'])
            LOG.debug("Delta seconds=%.4f " % (time.time() - t1))
            return hosts_with_conn

        if stg_provider_lst:
            for stg_provider in stg_provider_lst:
                eff_conn = scg_dict['health_status']['effective_connectivity']

                # A provider can be used for both boot and data volumes or only
                # one. In the case where they are marked for both boot and data
                # the scg is valid only if it can provide both boot and data.
                type_not_met = None
                if (CHK_BOOT in stg_provider and stg_provider[CHK_BOOT]):
                    connectivity = set(self.boot_connectivity) & eff_conn
                    if not _check_prov_conn(stg_provider, connectivity):
                        type_not_met = 'boot'
                if (CHK_DATA in stg_provider and stg_provider[CHK_DATA]):
                    connectivity = set(self.data_connectivity) & eff_conn
                    if not _check_prov_conn(stg_provider, connectivity):
                        if type_not_met:
                            type_not_met = 'boot+data'
                        else:
                            type_not_met = 'data'
                if type_not_met:
                    msg = _("Storage connectivity group '%(scg)s' does not "
                            "provide '%(type)s' connectivity for the "
                            "provider '%(provider)s' to any host."
                            % dict(scg=self.display_name,
                                   type=type_not_met,
                                   provider=stg_provider['host_display_name']))
                    LOG.info(msg)
                    if msg_dict:
                        msg_dict['messages'].extend(
                            scg_dict['health_status']['value_reasons'])
                    LOG.debug("Delta seconds=%.4f " % (time.time() - t1))
                    return hosts_with_conn

        # loop over list of host names
        for host in scg_dict['host_list']:
            health = scg_dict['health_status']['member_status'][host['name']]
            added_host = False
            if host_name and host_name != host['name']:
                LOG.debug("Host %s not considered since filtered out by "
                          "specific host to check connectivity for: %s." %
                          (host['name'], host_name))
            # There are different values that could be looked at here to
            # verify host connectivity.
            elif (conn_type == CHK_ALL and
                  (host['boot_vios_req_met'] and
                   host['data_vios_req_met'])):
                hosts_with_conn.append(host['name'])
                added_host = True
            elif (conn_type == CHK_EITHER and
                  (host['boot_vios_req_met'] or
                   host['data_vios_req_met'])):
                hosts_with_conn.append(host['name'])
                added_host = True
            elif (conn_type == CHK_BOOT and
                  host['boot_vios_req_met']):
                hosts_with_conn.append(host['name'])
                added_host = True
            elif (conn_type == CHK_DATA and
                  host['data_vios_req_met']):
                hosts_with_conn.append(host['name'])
                added_host = True
            else:
                hosts_wout.append(host['name'])
                msg = _("Host '%(host_name)s' does not meet the connectivity "
                        "criteria per storage connectivity group '%(scg)s', "
                        "and it is filtered out of host candidacy lists. The "
                        "reference connectivity information is: host_data="
                        "%(host)s and health_data=%(health)s.") %\
                    dict(host_name=host['name'], scg=self.display_name,
                         host=host, health=health)
                LOG.info(msg)
            if added_host:
                LOG.debug("Adding host '%s' to connectivity list for SCG "
                          "'%s'. host_conn_info=%s, health=%s." %
                          (host['name'], self.display_name, host, health))

        if hosts_with_conn:
            msg = ("Storage connectivity group '%(scg)s' allows at least one "
                   "type of connectivity from hosts: %(hosts)s.")\
                % dict(scg=self.display_name, hosts=hosts_with_conn)
            if not host_name:
                msg = six.text_type(msg) + \
                    six.text_type(" Hosts without connectivity: %s"
                                  % hosts_wout)
            LOG.info(msg)
        elif hosts_wout and msg_dict:
            msg = _("INFO: One or more hosts do not meet the connectivity "
                    "requirements of storage connectivity group '%(scg)s', "
                    "so the group's use is restricted. The member hosts "
                    "checked for connectivity were: %(hosts)s. The health "
                    "status information of the connectivity group follows: "
                    "%(health)s")
            msg_dict['messages'].append(
                msg % dict(scg=self.display_name, hosts=hosts_wout,
                           health=scg_dict['health_status']))
        LOG.debug("Delta seconds=%.4f " % (time.time() - t1))
        return hosts_with_conn

    def get_host_name_list(self):
        """
        Return a list of host names in the SCG DOM passed. The SCG must have
        been looked up by a utility that populated it's dictionary rep in the
        context of a db session.
        """
        return [host['name'] for host in self.to_dict()['host_list']]


# #####################################################
# ######  Fibre Channel Port DOM Implementation  ######
# #####################################################
@base.NovaObjectRegistry.register
class FcPortFactory(compute_dom.VioServerComponentFactory):

    """Production DOM Factory implementation for FC Ports"""

    def __init__(self):
        """ Call super constructor and set class vars """
        super(FcPortFactory, self).__init__()
        self.fabrics = {}
        self.fabric_sync_time = 0
        self.max_sync_mins = cfg.CONF.ibmpowervc_fabric_sync_interval

    # ##############################################
    # # FC fabric helper methods. Fabric info is  ##
    # # not persisted in the nova DB, but the DOM ##
    # # is extended to support this information   ##
    # ##############################################
    def sync_port_to_fabrics_needed(self):
        """ Returns True if max time has elapsed and a general sync of
            of port entries to fabric entries is needed. Also returns True
            if the ports have not been mapped yet or if this is the first
            or second topology collection (we want to ensure two full cycles
            and then we can back off to the normal interval).
            Else return False.
            Note: This does not control whether we are going to cinder or
                  not. We always go to cinder to get the list of fabrics
                  every collection. But if the fabrics have not changed
                  then this interval is used to sync up port mappings in
                  case cabling gets changed.
        """
        topo = HostStorageTopologyFactory.theFactory.get_static_topology()
        t_delta = time.time() - self.fabric_sync_time
        LOG.debug("delta_sync_time=%.1f, self.max_sync_mins=%d, "
                  "collection_seq=%d." %
                  (t_delta/60.0, self.max_sync_mins, topo.sequence_num))
        if topo.sequence_num <= 2 or t_delta > (self.max_sync_mins * 60):
            return True
        return False

    def sync_fcports_to_fabric(self, context, transaction, wwpn_to_dom):
        """
        Synchronize the fcport.fabric property to a registered fabric.
        :param context: The API and DB authorization context.
        :param transaction: The current DB transaction object.
        :param wwpn_to_dom: Dictionary mapping of wwpn string to FCPort DOM.
        """
        # Get the internal API (cinder client for nova)
        fabricAPI = PowerVCFabricAPI()

        # Always pay the expense of getting the current registered fabrics
        LOG.debug("Get registered fabric information from cinder.")
        cur_fabs = fabricAPI.get_fabrics(context)
        if cur_fabs is None:
            LOG.info(_("Unable to retrieve fabrics from cinder API. Will not "
                       "update any port fabric mappings during this "
                       "topology collection."))
            return
        pri_fabs = self.fabrics  # prior tracked fabrics
        is_sync_needed = self.sync_port_to_fabrics_needed()
        LOG.debug("cur_fabs=%s,\n prior_fabs=%s,\nis_sync_needed=%s" %
                  (cur_fabs, self.fabrics, is_sync_needed))

        # If we got fabrics, but didn't have any before, then this could
        # be the situation for the first topo collection of the process.
        if cur_fabs and not pri_fabs and is_sync_needed:
            # There is a special case to handle here where the user
            # has removed a fabric from powervc, then immediately stopped
            # or restarted the host process. Rather than a fabric being
            # removed, it will look like a fabric was added. So we clear
            # the fabric mapping. For this case,
            # we will fall into the map_ports_to_fabrics() logic later
            # and the correct mapping will be saved if there is one.
            # This costs two saves(), but is not a common occurrance.
            for port in wwpn_to_dom.values():
                if port.fabric != "None":
                    LOG.debug("1st time window. Clear fabric mapping for "
                              "port %s" % port.wwpn)
                    port.fabric = "None"  # the fabric default
                    port.save(context=context, transaction=transaction)

        changed_fabrics = []
        for fab_name, cur_fab in cur_fabs.iteritems():
            if 'access_ip' not in cur_fab:
                LOG.info(_("Generic or unsupported switch type registered. "
                           "Port to fabric mapping is not supported. Continu"
                           "ing with remainder of topology collection."))
                return
            if fab_name in pri_fabs and (cur_fabs[fab_name]['access_ip'] !=
                                         pri_fabs[fab_name]['access_ip']):
                changed_fabrics.append(fab_name)

        # Fabric removal case and changed fabrics also handled up front.
        # We may set a port.fabric to None in this block that will later
        # get mapped properly if fabric is in changed_fabrics.
        if len(cur_fabs) < len(pri_fabs) or changed_fabrics:
            LOG.info(_("A Fabric was removed (or changed) from management. "
                       "Clear out FC port mappings for it."))
            for f_name in pri_fabs:
                if (f_name not in cur_fabs) or (f_name in changed_fabrics):
                    for port in wwpn_to_dom.values():
                        if port.fabric == f_name:
                            LOG.debug("Clear fabric setting for %s" %
                                      port.wwpn)
                            port.fabric = "None"  # the fabric default
                            port.save(context=context, transaction=transaction)

        # We want to do the relatively more expensive work of mapping ports to
        # fabrics in these situations:
        # 1. There are registered fabrics and the last time the port-to-fabric
        #    sychronization was exceeds the configured time interval. This
        #    covers rare cases where a port is cabled up or cables are
        #    switched sometime after fabric registration.
        # 2. One or more new fabrics have been registered since the last check.
        # 3. One of the current fabrics registered was registered before,
        #    but the IP has changed since the last check (effectively a
        #    new registration).
        if ((cur_fabs and is_sync_needed) or (len(cur_fabs) > len(pri_fabs)) or
                changed_fabrics):
            # Now try to map the wwpns to any detected changed fabric setting.
            # The fabric information that comes back on the prior call does
            # not include any fabric health status, so a switch problem
            # would show up in not getting data back from the next call.
            port_mapping = fabricAPI.map_ports_to_fabrics(context,
                                                          wwpn_to_dom.keys())
            # We set the sync time even if the the API call failed
            # because we don't want to waste time on every topo collection
            # cycle hitting the failure (which could be a timeout).
            self.fabric_sync_time = time.time()
            if port_mapping is not None:
                for wwpn, fabric in port_mapping.iteritems():
                    # Since there could be issues getting the fabric data,
                    # we never map a port to "None", only if it is a valid
                    # fabric and if it has changed from the current setting.
                    # TODO: investigate allowing the 'set to None' case here
                    #       for 1.3.1 since a non-reporting fabric may mean
                    #       a larger issue than just managment of the fabric.
                    if fabric and fabric != wwpn_to_dom[wwpn].fabric:
                        LOG.info(_("Setting FC Port with WWPN '%(wwpn)s' to "
                                   "fabric '%(to_fabric)s' from prior value "
                                   "of '%(from_fabric)s'.") %
                                 dict(wwpn=wwpn, to_fabric=fabric,
                                      from_fabric=wwpn_to_dom[wwpn].fabric))
                        wwpn_to_dom[wwpn].fabric = fabric
                        from powervc_nova.storage.topo_common import\
                            PORT_NPIV_STATUS_UNK_FABRIC
                        unk_fab = 'OK|' + PORT_NPIV_STATUS_UNK_FABRIC
                        if wwpn_to_dom[wwpn].status == unk_fab:
                            wwpn_to_dom[wwpn].status = 'OK|OK'
                        wwpn_to_dom[wwpn].save(context=context,
                                               transaction=transaction)
                    else:
                        LOG.debug("WWPN %s fabric is unknown or unchanged: "
                                  "%s." % (wwpn, wwpn_to_dom[wwpn].fabric))
            else:
                LOG.warn(_("Unable to get mapped WWPN information from "
                           "fabric switches."))
        else:
            LOG.debug("FC switch fabric registration criteria not met for "
                      "syncing ports to fabrics.")
        self.fabrics = cur_fabs

    # ##############################################
    # ##### Defining FCPort Factory Interface ######
    # ##############################################
    def construct_fcport(self, context, name, vios_id, id=None,
                         wwpn=None, status=None, enabled=None, port_tag=None,
                         fabric=None, adapter_id=None, avail_active_vfcs=None,
                         mapped_vfcs=None, max_vfcs=None, reserved_vfcs=None,
                         auto_create=False, transaction=None):
        """Constructs (and optionally creates) the SCG specified"""
        kwargs = dict(name=name, vios_id=vios_id, id=id, wwpn=wwpn,
                      status=status, enabled=enabled, port_tag=port_tag,
                      fabric=fabric, adapter_id=adapter_id, max_vfcs=max_vfcs,
                      mapped_vfcs=mapped_vfcs, reserved_vfcs=reserved_vfcs,
                      avail_active_vfcs=avail_active_vfcs)
        # Call the Parent Class to actually Construct the given Resource
        return self.construct_resource(FcPort, context, auto_create,
                                       transaction, **kwargs)

    def find_all_fcports(self, context, filters=None, transaction=None):
        """Queries all of the FCPorts matching the filter specified"""
        return self.find_all_resources(
            FcPort, context, filters, transaction)

    def find_all_fcports_by_vios(self, context, vios_id, transaction=None):
        """Queries all of the FCPorts matching the filter specified"""
        filters = {self._get_vios_id_attr_name(): vios_id}
        return self.find_all_resources(
            FcPort, context, filters, transaction)

    def find_all_fcports_by_vios_ids(self, context, vios_ids,
                                     transaction=None, vioses=None):
        """ Queries all of the FCPorts matching the VIOS filter """
        t1 = time.time()
        # TODO: plumb the vios_ids special filter in SQLalchemy.
        # filters = {'vios_ids': vios_ids}
        ports = self.find_all_resources(
            FcPort, context, None, transaction)
        t2 = time.time()
        LOG.debug("Obtained %d total vios fcports." % len(ports))
        subset_ports = []
        id_to_vio = {}
        if vioses:
            # For a significant number of VIOSes, it will be more efficient
            # if we build a hash table to use below. Exclude VIOSes with
            # ports already cached...we don't want to append in that case
            # and end up with duplicates.
            id_to_vio = dict([(v.id, v) for v in vioses if not v._fcports])

        # Now filter ports and link the ports into the dom cache of vioses
        for port in ports:
            if port.vios_id not in vios_ids:
                continue
            subset_ports.append(port)
            if port.vios_id in id_to_vio:
                vios = id_to_vio[port.vios_id]
                if vios._fcports is None:
                    vios._fcports = []
                vios._fcports.append(port)
            else:
                LOG.debug("Port %s has VIOS %s not found in 'vioses'. "
                          "vioses not provided or ports already cached or "
                          "mismatched data structured provided." %
                          (port.wwpn, port.vios_id))
        # Return the list of ports.
        LOG.debug("Timing delta seconds=%.4f, hashing time=%.4f, hash-size"
                  "=%d." % (t2 - t1, time.time() - t2, len(id_to_vio)))
        return subset_ports

    def find_all_fcports_by_host(self, context, host_name, transaction=None):
        """Queries all of the FCPorts matching the filter specified"""
        return self.find_all_resources(
            FcPort, context, dict(host_name=host_name), transaction)

    def find_fcport_by_id(self, context, port_id, transaction=None):
        """Queries at most one FCPort matching the Port ID specified"""
        filters = {self._get_resource_id_attr_name(): port_id}
        return self.find_resource(
            FcPort, context, filters, transaction)

    def delete_fcport(self, context, port_id, transaction=None):
        """Deletes the FCPort with the Port ID specified"""
        return self.delete_resource(
            FcPort, context, port_id, transaction)

    # ##############################################
    # ## Implementing Resource Factory Definition ##
    # ##############################################
    def _get_resource_classes(self):
        """Returns a List of DOM Classes or a Dictionary Mapping DOM Classes"""
        return [FcPort]

    def _auto_generate_resource_id(self, attribs):
        """Auto-generates the Resource Identifier for this Resource"""
        return self._generate_resource_uuid(attribs, ['vios_id', 'adapter_id'])

    def _get_vios_factory(self):
        """Provides the Factory to use for the VIO Server Implementation"""
        return VioServerFactory.get_factory()

    @base.remotable
    def _find_all_resources(self, resource_class,
                            filters=None, transaction=None):
        """Factory implementation to query all resources matching the filter"""
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        fcports = db_api.fcport_find_all(self._context, filters, dbtxn)
        return self._construct_resources(
            resource_class, self._context, fcports, transaction)

    @base.remotable
    def _delete_multiple_resources(self, resource_class,
                                   resource_ids, transaction=None):
        """Factory implementation to delete the resources specified"""
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        for resource_id in resource_ids:
            db_api.fcport_delete(self._context, resource_id, dbtxn)


@base.NovaObjectRegistry.register
class FcPort(compute_dom.VioServerComponent):

    """Production DOM Resource implementation for FC Ports"""

    fields = dict(compute_dom.VioServerComponent.fields,
                  name=fields.StringField(nullable=True),
                  status=fields.StringField(nullable=True),
                  enabled=fields.BooleanField(nullable=True),
                  wwpn=fields.StringField(nullable=True),
                  adapter_id=fields.StringField(nullable=True),
                  port_tag=fields.StringField(nullable=True),
                  fabric=fields.StringField(nullable=True),
                  fc_protocol=fields.StringField(nullable=True),
                  max_vfcs=fields.IntegerField(nullable=True),
                  mapped_vfcs=fields.IntegerField(nullable=True),
                  reserved_vfcs=fields.IntegerField(nullable=True),
                  avail_active_vfcs=fields.IntegerField(nullable=True))

    def __init__(self, context=None,
                 factory_class=None, transaction=None, **kwargs):
        "Constructs an instance of the given DOM Resource Class"
        super(FcPort, self).__init__(
            context, factory_class, transaction, **kwargs)

    # ##############################################
    # #### Implementing DOM Resource Definition ####
    # ##############################################
    @base.remotable
    def _create_resource(self, transaction):
        """Actual Implementation that will Create the given Resource"""
        values = self._get_resource_values()
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        fcport = db_api.fcport_create(self._context, values, dbtxn)
        self._hydrate_all_resource_attributes(fcport)

    @base.remotable
    def _update_resource(self, transaction):
        """Actual Implementation that will Update the given Resource"""
        fcport_id = self._get_resource_identifer()
        values = self._get_resource_values(changes_only=True)
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        fcport = db_api.fcport_update(
            self._context, fcport_id, values, dbtxn)
        self._hydrate_all_resource_attributes(fcport)


# #####################################################
# ####  Host Storage Topology DOM Implementation  #####
# #####################################################
@base.NovaObjectRegistry.register
class HostStorageTopologyFactory(dom.ResourceFactory):

    """Production DOM Factory implementation for Host Storage Topology"""
    theFactory = None  # static class var

    # ##############################################
    # # Cluster membership tracked info.           #
    # # This will likely move away from the        #
    # # factory once HostStorageTopology instances #
    # # come into use.                             #
    # ##############################################
    def __init__(self):
        super(HostStorageTopologyFactory, self).__init__()
        self.static_topo = StaticStgTopo()
        HostStorageTopologyFactory.theFactory = self

    def get_static_topology(self):
        return self.static_topo

    # ##############################################
    # #####  Defining HST Factory Interface  #######
    # ##############################################
    def construct_storage_topology(self, context, host_name, transaction=None,
                                   **kwargs):
        """Constructs the Storage Topology for the Host specified"""
        # Call the Parent Class to actually Construct the given Resource
        return self.construct_resource(HostStorageTopology, context, False,
                                       transaction, host_name=host_name,
                                       vioservers=[], fcports=[], **kwargs)

    def find_storage_topology(self, context, host_name, transaction=None):
        """Queries the Storage Topology for the Host specified"""
        return self.find_resource(HostStorageTopology, context,
                                  dict(host_name=host_name), transaction)

    # ##############################################
    # ## Implementing Resource Factory Definition ##
    # ##############################################
    def _get_resource_classes(self):
        """Returns a List of DOM Classes or a Dictionary Mapping DOM Classes"""
        return [HostStorageTopology]

    @base.remotable
    def _find_all_resources(self, resource_class,
                            filters=None, transaction=None):
        """Factory implementation to query all resources matching the filter"""
#        topo = dict()
#         host = filters.get('host_name', '')
#         # Construct each of the Factories used to retrieve the info
#         portFact = FcPortFactory.getFactory()
#         viosFact = VioServerFactory.getFactory()
#         # Retrieve the VIOS's, FCPorts, and VIOS Clusters for the given Host
#         topo['vioservers'] = viosFact.\
#             find_all_vioses_by_host(self._context, host, transaction)
#         topo['fcports'] = portFact.\
#             find_all_fcports_by_host(self._context, host, transaction)
#         return self._construct_resources(
#             resource_class, self._context, [topo], transaction)
        return []

    def _get_resource_id_attr_name(self):
        """Provides the Name of the Attribute for the Resource Identifier"""
        return 'host_name'

    def _auto_generate_resource_id(self, attribs):
        """Auto-generates the Resource Identifier for this Resource"""
        return '%s--%07d' % (attribs['host_name'],
                             self.static_topo.sequence_num + 1)


@base.NovaObjectRegistry.register
class HostStorageTopology(dom.Resource):

    """Production DOM Resource implementation for Host Storage Topology"""

    fields = dict(dom.Resource.fields,
                  host_name=fields.StringField(),
                  vioservers=fields.ListOfObjectsField(dom.Resource),
                  fcports=fields.ListOfObjectsField(dom.Resource))

    def __init__(self, context=None,
                 factory_class=None, transaction=None, **kwargs):
        "Constructs an instance of the given DOM Resource Class"
        super(HostStorageTopology, self).__init__(
            context, factory_class, transaction, **kwargs)
        if not self._context:
            self._context = nova_context.get_admin_context()

    def construct_pvm_helper(self, host_name, host_uuid, adapter):
        """ construct and return the pvm topology helper """
        from powervc_nova.storage import topo_pvm
        self.common_topo = topo_pvm.StorageTopologyPvm(
            host_name, self._context, adapter, host_uuid)
        return self.common_topo

    def construct_hmc_helper(self, host_name):
        """ construct and return the hmc topology helper """
        from powervc_nova.storage import topo_hmc
        self.common_topo = topo_hmc.StorageTopologyHmc(
            host_name, self._context)
        return self.common_topo

    # ##############################################
    # #### Implementing DOM Resource Definition ####
    # ##############################################

    def _initialize_resource_attributes(self):
        """Sets any default values that are defined for any attributes"""
        super(HostStorageTopology, self)._initialize_resource_attributes()
        setattr(self, 'vioservers', list())
        setattr(self, 'fcports', list())

    def _create_resource(self, transaction):
        """Actual Implementation that will Create the given Resource"""
        # TODO: This is not yet called since topo.save() has been cut out
        #       of virt/ibmpowervm/.../driver.py
        #       The reason for bypassing this flow is that it blew up when
        #       switching to the Ubuntu proxy configuration.
        #       Brent's comment: Yeah, cause the constructor is thinking it is
        #       running in nova-compute so constructing the HMC one in the
        #       conductor because it is set to "hmc" on the Mgmt System
        #       (eventually it would be None). The conductor log has message:
        #       "NV-F1250C2 Improper call for method. The
        #       dvr_host.compute_host is not set."
        #
        # Since the Topology isn't actually persisted we will treat
        # someone who queried the Topology first, the same as constructing it.
        LOG.debug("Create resource via _update_resource().")
        self._update_resource(transaction)

    @base.remotable
    def _update_resource(self, transaction):
        """Actual Implementation that will Update the given Resource"""
        LOG.debug("_update_resource for HostStorageTopology %s" % self)
        # TODO: Cleanup flow by restructuring topology collection to get
        # topo information from HMC or PowerVM, then persist it in this method
        # which is "remotable" and runs back on the management server.
        # Note that the code here is original generic flow not yet used.
        dbtxn = dom.ResourceTransaction.get_dbtransaction(transaction)
        # Update the Cluster Provider Name on each of the VIO Servers
        for vios in self.vioservers:
            vios_id = vios._get_resource_identifer()
            values = vios._get_resource_values(changes_only=True)
            db_api.vios_update(self._context, vios_id, values, dbtxn)
        # Retrieve the existing FCPorts for the Host to figure out differences
        dbps = db_api.fcport_find_all(
            self._context, dict(host_name=self.host_name), dbtxn)
        exist_ports = dict([(port['id'], port) for port in dbps])
        new_ports = dict([(port['id'], port) for port in self.fcports])
        # Determine whether the Port needs to be Created or updated
        for port in new_ports:
            # If the Port already exists, we just want to update what changed
            if exist_ports.get(port['id']) is not None:
                values = port._get_resource_values(changes_only=True)
                db_api.fcport_update(self._context, port['id'], values, dbtxn)
            # If the Port is new, then we want to create it in the database
            else:
                values = port._get_resource_values(changes_only=False)
                db_api.fcport_create(self._context, values, dbtxn)
        # Now we want to clean up any ports that no longer exist in the
        # topology
        for port in exist_ports:
            if new_ports.get(port['id']) is None:
                db_api.fcport_delete(self._context, port['id'], dbtxn)


# #################################################
# ###  Functions migrated from storage.olddom  ####
# #################################################
def fcport_to_dict(db_fcport):
    """
    Convenience method to take a FC Port entry DTO and return a dict
    representation of the DTO. When FCPorts are created and looked up by
    their own DOM above, then this logic can move to the DOM's to_dict().
    """
    from powervc_nova.storage.topo_common import PORT_STATUS_UNKNOWN,\
        PORT_STATUS_OK
    fcport_dict = {'id': db_fcport.id,
                   'adapter_id': db_fcport.adapter_id,
                   'wwpn': db_fcport.wwpn,
                   'name': db_fcport.name if db_fcport.name else db_fcport.id,
                   'enabled': db_fcport.enabled,
                   'fc_protocol': db_fcport.fc_protocol,
                   'fabric': db_fcport.fabric,
                   'max_vfcs': db_fcport.max_vfcs,
                   'mapped_vfcs': db_fcport.mapped_vfcs,
                   'avail_active_vfcs': db_fcport.avail_active_vfcs,
                   'status': PORT_STATUS_UNKNOWN,
                   'npiv_status': PORT_STATUS_OK}
    # Note: A large performance boost was gained by removing a line here
    # that referenced db_fcport.vioserver. Avoid that reference since it
    # spawns conductor calls back to the pareent VIOS, and we should already
    # know the parent VIOS.
    if db_fcport.port_tag:
        fcport_dict['port_tag'] = db_fcport.port_tag
    if db_fcport.status:
        # In 1.2.3, we began storing both status and npiv_status in the
        # status property string, separated by '|'. Note that npiv_status
        # defaults to OK here since an upgraded PowerVC will initially not
        # have port status with '|' separator.
        status_combo = db_fcport.status.split('|')
        fcport_dict['status'] = status_combo[0]
        if len(status_combo) > 1:
            fcport_dict['npiv_status'] = status_combo[1]

    if cfg.CONF.ibmpowervm_use_mapped_vfc_scheduling:
        if db_fcport.max_vfcs and db_fcport.mapped_vfcs:
            fcport_dict['available_connections'] =\
                db_fcport.max_vfcs - db_fcport.mapped_vfcs
        elif db_fcport.mapped_vfcs is None:
            # it could be that mapped_vfcs are not present
            # or k2 gave an error?
            fcport_dict['available_connections'] =\
                db_fcport.max_vfcs
        else:
            fcport_dict['available_connections'] = 0
    else:
        fcport_dict['available_connections'] = db_fcport.avail_active_vfcs
    # If fc_protocol is not set, it is considered the same as 'any'
    if fcport_dict['fc_protocol'] is None and fcport_dict['enabled']:
        fcport_dict['fc_protocol'] = 'any'
    return fcport_dict

# #####################################################
# #### Storage DOM Resource Factory Registration ######
# #####################################################
dom.ResourceFactoryLocator.register_factory(FcPortFactory())
dom.ResourceFactoryLocator.register_factory(VioServerFactory())
dom.ResourceFactoryLocator.register_factory(ManagedHostFactory())
dom.ResourceFactoryLocator.register_factory(StorageProviderFactory())
dom.ResourceFactoryLocator.register_factory(HostStorageTopologyFactory())
dom.ResourceFactoryLocator.register_factory(StorageConnectivityGroupFactory())
