# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import math

import six

from nova.ibm.ego import ego_constants as ego_consts
from nova.ibm.ego.rtpolicy import rtpolicy_optimizer as optimizers
from nova.scheduler.ibm.ego import ego_rpcapi

from oslo_log import log as logging

from powervc_nova.common import aggregate_utils as agg_utils
from powervc_nova.compute import api as pvc_compute
from powervc_nova.objects.compute import dom as compute_dom
from powervc_nova.virt.ibmpowervm.common import enterprise_pool as common_ep
from powervc_nova.prs import constants as pvc_prs_consts
from powervc_nova.prs import utils as prs_utils
from powervc_nova.prs import dro_schedules
from powervc_nova.virt.ibmpowervm.common import constants
from powervc_nova.virt.ibmpowervm.hmc.k2 import mgmt_console_wrapper
from powervc_nova import _

LOG = logging.getLogger(__name__)


class PowerVCRTPolicyCodOptimizer(optimizers.RTPolicyOptimizer):
    """
    Implementation of the enterprise pool capacity on demand (CoD) optimizer.
    """

    def __init__(self, *args, **kwargs):
        super(PowerVCRTPolicyCodOptimizer, self).__init__(*args, **kwargs)
        self._migration_optimizer = optimizers.RTPolicyMigrationOptimizer(
            *args, **kwargs)
        self._scheduler_api = ego_rpcapi.EGOSchedulerAPI()
        self._compute_api = pvc_compute.PowerVCHostAPI()

    def receive_notification(self, context, msg):
        self._migration_optimizer.receive_notification(context, msg)

    def reset_optimizer(self):
        self._migration_optimizer.reset_optimizer()

    def run_ticker(self, context):
        dro_schedules.update_dro_based_on_schedules(
            context, self._policy_engine_manager)
        self._migration_optimizer.run_ticker(context)

    def run_interval_ticker(self, context, policy_info):
        self._migration_optimizer.run_interval_ticker(context, policy_info)

    def _check_optimization_host_prereqs(self, context, host_name, pool):
        """
        Check any prerequisites required to be satisfied prior to allowing
        ongoing optimization to occur on the hypervisor.

        :param context: nova context object
        :param host_name: name of a service host
        :return: True if optimization is allowed and False if not allowed
        """
        # Check that the host belongs to an enterprise pool
        if not pool:
            LOG.debug('Host %s does not belong to any enterprise pools, '
                      'therefore the optimizer will not try to optimize it '
                      'via enterprise pool CoD operations.' % host_name)
            return False

        compute_node = prs_utils.get_compute_node_by_service_host(context,
                                                                  host_name)

        # Check that the host is managed by a platform that supports enterprise
        # pools
        platform = self._get_host_stat(compute_node, constants.MANAGED_BY)
        if platform not in constants.PLATFORMS_SUPPORTING_EPOOLS:
            LOG.info(_('Host %(host)s is managed by a platform (%(platform)s) '
                       'that does not support enterprise pools.') %
                     {'host': host_name,
                      'platform': platform})
            return False

        if self._need_to_check_managing_hmc(compute_node):
            hmc_fact = compute_dom.ManagementConsoleFactory.get_factory()
            master_hmc = hmc_fact.find_hmc_by_mtmsn(context,
                                                    pool.master_hmc_mtmsn)
            if not master_hmc:
                LOG.info(_('HMC with MTMSN (machine type, model, serial '
                           'number) %(mtmsn)s could not be found in the '
                           'database; this is likely because the master HMC '
                           'of the pool is not managed by PowerVC. The '
                           'optimizer will therefore not try to optimize host '
                           '%(host)s via enterprise pool CoD operations.')
                         % {'mtmsn': pool.master_hmc_mtmsn,
                            'host': host_name})
                return False
            elif host_name not in master_hmc.managed_hosts:
                LOG.info(_('Host %s is not managed by the master HMC of the '
                           'pool; therefore the optimizer will not try to '
                           'optimize it via enterprise pool CoD operations.') %
                         host_name)
                return False

        # Check that the enterprise pool is not out of compliance
        if pool.compliance_state == constants.EPOOL_OUT_OF_COMPLIANCE:
            LOG.info(_('Host %s belongs to an enterprise pool that is not in '
                       'compliance; therefore the optimizer will not try to '
                       'optimize it via enterprise pool CoD operations.') %
                     host_name)
            return False

        # Check if the host has inactive cores available for use
        if self._get_inactive_cores(compute_node) < 1:
            LOG.info(_('Host %s does not have any inactive cores available '
                       'for use; therefore the optimizer will not try to '
                       'optimize it via enterprise pool CoD operations.') %
                     host_name)
            return False

        # Check if the host has any unreturned mobile cores
        unreturned_cores = self._get_unreturned_mobile_cores(compute_node)
        if unreturned_cores > 0:
            LOG.info(_('Host %(host)s has %(num)d unreturned mobile cores; '
                       'therefore the optimizer will not try to optimize it '
                       'via enterprise pool CoD operations.') %
                     {'host': host_name,
                      'num': unreturned_cores})
            return False

        return True

    def _calculate_core_requirements(self, context, host_name, threshold):
        """
        Calculates the number of cores that are needed to satisfy the threshold
        of the host, limits the number of cores that can be used based on what
        is available and pre-defined limits, and then determines the number of
        cores that are available for use in the pool and the number of cores
        that will need to be reclaimed from other hosts to satisfy the
        requirements.

        :param context: nova context object
        :param host_name: name of the service host
        :param threshold: threshold number as defined in the policy
        :return: a tuple comprised of (the number of cores that should be taken
                 from the pool, the number of cores that need to be reclaimed
                 from other hosts)
        """
        cores_to_satisfy_threshold = (
            self._calculate_cores_to_satisfy_threshold(context, host_name,
                                                       threshold))

        if cores_to_satisfy_threshold == 0:
            LOG.info(_('The host %(host)s does not need any mobile cores to '
                       'satisfy the threshold.') %
                     {'host': host_name})

        pool = common_ep.get_pool_for_host(context, host_name)

        cores_needed = self._limit_cores_needed(
            context, host_name, cores_to_satisfy_threshold,
            pool.total_mobile_cores)
        cores_to_reclaim = self._calculate_cores_to_reclaim(
            host_name, cores_needed, pool.available_mobile_cores)

        cores_to_take = cores_needed - cores_to_reclaim

        LOG.info(_("Calculated that %(cores_to_take)d mobile cores should "
                   "be taken from the pool's available mobile cores and "
                   "%(cores_to_reclaim)d mobile cores should be reclaimed "
                   "from other hosts in the same enterprise pool if possible "
                   "to optimize host %(host)s.") %
                 {'cores_to_take': cores_to_take,
                  'cores_to_reclaim': cores_to_reclaim,
                  'host': host_name})

        return cores_to_take, cores_to_reclaim

    def _limit_cores_needed(self, context, host_name, cores_needed,
                            total_mobile_cores_in_pool):
        """
        Limits the number of cores needed to satisfy the threshold with what is
        currently available in the pool. The number of cores activated per host
        per round can also be restricted.

        :param context: nova context object
        :param host_name: name of the service host
        :param cores_needed: number of cores needed to satisfy the threshold as
                             determined by a mathematical equation
        :param total_mobile_cores_in_pool: total number of mobile cores that
                                           exist in the pool
        :return: the limited number of cores that are needed
        """
        # Restrict the number of cores that we try to move to the host
        # in a given round of optimization
        if (pvc_prs_consts.COD_RESTRICT_CORE_ACTIVATIONS_PER_CYCLE and
                cores_needed >=
                pvc_prs_consts.COD_MAX_CORE_ACTIVATIONS_PER_CYCLE):
            LOG.info(_('The number of cores (%(cores_needed)d) that host '
                       '%(host)s needs is greater than the maximum number of '
                       'cores (%(max)d) that can be activated during one '
                       'cycle of optimization. Therefore only %(max)d core(s) '
                       'will be attempted to be activated.') %
                     {'cores_needed': cores_needed,
                      'host': host_name,
                      'max':
                      pvc_prs_consts.COD_MAX_CORE_ACTIVATIONS_PER_CYCLE})
            cores_needed = pvc_prs_consts.COD_MAX_CORE_ACTIVATIONS_PER_CYCLE

        # If more cores are needed than the total number of mobile
        # cores in the pool, then only try and move the total number
        # of cores in the pool
        if cores_needed > total_mobile_cores_in_pool:
            LOG.info(_('The number of cores (%(cores_needed)d) that host '
                       '%(host)s needs is greater than the total number of '
                       'mobile cores in the pool (%(total)d). Therefore only '
                       '%(total)d cores will be attempted to be activated.') %
                     {'cores_needed': cores_needed,
                      'host': host_name,
                      'total': total_mobile_cores_in_pool})
            cores_needed = total_mobile_cores_in_pool

        # If more cores are needed than the number of physical cores
        # available on the host itself, then only try and move the
        # number of available cores left on the host
        compute_node = prs_utils.get_compute_node_by_service_host(context,
                                                                  host_name)
        inactive_cores = self._get_inactive_cores(compute_node)
        if cores_needed > inactive_cores:
            LOG.info(_('The number of cores (%(cores_needed)d) that host '
                       '%(host)s needs is greater than the number of '
                       'inactive cores on the host (%(inactive)d). Therefore '
                       'only %(inactive)d cores will be attempted to be '
                       'activated.') %
                     {'cores_needed': cores_needed,
                      'host': host_name,
                      'inactive': inactive_cores})
            cores_needed = inactive_cores

        return cores_needed

    def _calculate_cores_to_satisfy_threshold(self, context, host_name,
                                              threshold):
        """
        Determine the number of cores that are needed by solving for the
        number of cores that would result in the CPU utilization equaling
        the threshold and then taking the ceiling of that result since mobile
        core assignments must operate in terms of whole numbers.

        :param context: nova context object
        :param host_name: name of the service host
        :param threshold: threshold number as defined in the policy
        :return: number of cores needed to satisfy the threshold for the host
        """
        compute_node = prs_utils.get_compute_node_by_service_host(context,
                                                                  host_name)

        current_active_cores = self._get_active_cores(compute_node)
        current_cpu_utilization = prs_utils.get_cpu_utilization(
            context, compute_node, self._scheduler_api)

        # In order to determine the number of cores needed, we set the
        # threshold equal to the current CPU utilization times the number of
        # currently active cores (which gets us the number of currently
        # utilized proc units) and then divide that by the number of currently
        # active cores plus the cores_needed (which is the total number of
        # cores that are needed to satisfy the threshold). Solving for
        # cores_needed gets us the number of cores that need to be added in
        # order for the CPU utilization to be <= to the threshold. The equation
        # used below is derived here:
        #
        # threshold = (current_cpu_utilization * current_active_cores) /
        #             (current_active_cores + cores_needed)
        # threshold * (current_active_cores + cores_needed) =
        #    (current_cpu_utilization * current_active_cores)
        # cores_needed = ((current_cpu_utilization * current_active_cores) /
        #                 threshold) - current_active_cores
        # cores_needed = current_active_cores * (
        #                    (current_cpu_utilization / threshold) - 1)

        if int(threshold) == 0:
            LOG.debug('Threshold was set to %s, therefore setting it to 0.1 '
                      'to avoid a division by zero error.' % threshold)
            threshold = 0.1

        # Truncate at two decimal places to help eliminate over-compensation
        val = float("%.2f" % (current_cpu_utilization / threshold - 1))
        # Mobile cores can only be activated in whole numbers, not fractions,
        # hence the need for taking the ceiling of the result
        cores_needed = int(math.ceil(current_active_cores * val))

        # Although not likely (or probably possible), account for when the
        # current_cpu_utilization < threshold in the calculation above, which
        # would result in a negative number of cores needed
        cores_needed = max(0, cores_needed)

        LOG.debug('Determined that host %(host)s needs %(cores_needed)d '
                  'additional cores to bring the CPU utilization down from '
                  '%(current_util)s%% to %(threshold_util)s%% or less. The '
                  'current number of active cores is %(current_cores)d.'
                  % {'host': host_name,
                     'cores_needed': cores_needed,
                     'current_util': current_cpu_utilization,
                     'threshold_util': threshold,
                     'current_cores': current_active_cores})

        return cores_needed

    def _calculate_cores_to_reclaim(self, host_name, cores_needed,
                                    available_mobile_cores_in_pool):
        """
        Calculates the number of cores that need to be reclaimed from other
        hosts based on the number of available cores in the pool in order to
        meet the cores_needed requirement. Essentially, if more cores are
        needed than the number of cores available in the pool, then some must
        be reclaimed.

        :param host_name: name of the service host
        :param cores_needed: number of cores that are needed to satisfy the
                             threshold requirement for the host
        :param available_mobile_cores_in_pool: number of mobile cores in the
                                               pool that are not assigned to
                                               any host
        :return: number of cores that need to be reclaimed
        """
        cores_to_reclaim = 0
        if cores_needed > available_mobile_cores_in_pool:
            cores_to_reclaim = cores_needed - available_mobile_cores_in_pool
            LOG.info(_('The host %(host)s needs to have %(cores_needed)d '
                       'cores activated to reduce its CPU utilization to the '
                       'desired threshold, but the enterprise pool it belongs '
                       'to only has %(available_cores)d available mobile '
                       'cores. Therefore %(cores_to_reclaim)d cores need to '
                       'be reclaimed from other hosts using mobile cores '
                       'belonging to the same pool.') %
                     {'host': host_name,
                      'cores_needed': cores_needed,
                      'available_cores': available_mobile_cores_in_pool,
                      'cores_to_reclaim': cores_to_reclaim})

        return cores_to_reclaim

    def _consider_host_for_reclamation(self, context, host, pool):
        '''
        Only consider the host for reclaiming if the following conditions are
        met:
            1) The host's associated compute service is up (in :-) state)
            2) The host is not in maintenance mode
            3) The host has DRO enabled
            4) The host has no unreturned cores
            5) If the host is managed by an HMC of version less than 850, it
               must be managed by the master HMC of the pool

        Only consider hosts that have 0 unreturned cores because reclaiming
        from hosts with unreturned cores could exasperate any enterprise pool
        CoD compliance issues by creating more unreturned cores.
        Therefore we explicitly choose not to touch any hosts in this
        state.

        :param context: nova context object
        :param host: name of the service host
        :param pool: obj for the enterprise pool the violated host belongs to
        :return: True if the host should be considered for reclamation,
                 otherwise False
        '''
        node = prs_utils.get_compute_node_by_service_host(context, host)

        if not prs_utils.is_host_up(context, node):
            LOG.info(_('Not considering host %s for reclamations because the '
                       'associated compute service is deemed down.') % host)
            return False
        elif prs_utils.is_in_maintenance_mode(node):
            LOG.info(_('Not considering host %s for reclamations because it '
                       'is currently in maintenance mode.') % host)
            return False
        elif not prs_utils.get_dro_enabled_for_host(context, node):
            LOG.info(_('Not considering host %s for reclamations because it '
                       'has DRO disabled.') % host)
            return False
        elif self._get_unreturned_mobile_cores(node) != 0:
            LOG.info(_('Not considering host %s for reclamations because it '
                       'has unreturned cores.') % host)
            return False
        elif self._need_to_check_managing_hmc(node):
            hmc_fact = compute_dom.ManagementConsoleFactory.get_factory()
            master_hmc = hmc_fact.find_hmc_by_mtmsn(context,
                                                    pool.master_hmc_mtmsn)
            if not master_hmc:
                LOG.info(_('Not considering host %(host)s for reclamations '
                           'because the HMC with MTMSN (machine type, model, '
                           'serial number) %(mtmsn)s could not be found in '
                           'the database; this is likely because the master '
                           'HMC of the pool is not managed by PowerVC.')
                         % {'mtmsn': pool.master_hmc_mtmsn,
                            'host': host})
                return False
            elif host not in master_hmc.managed_hosts:
                LOG.info(_('Not considering host %s for reclamations '
                           'because it is not managed by the master HMC of '
                           'the pool.') % host)
                return False

        return True

    def _find_hosts_to_reclaim_from(self, context, threshold, aggregate_id,
                                    violated_hosts, already_optimized_hosts,
                                    pool, num_cores_to_reclaim):
        """
        Given a number of cores that need to be reclaimed, determines which
        hosts to reclaim from and how many cores to reclaim from each host.

        The algorithm to determine which hosts to reclaim from and how many
        cores to reclaim from said hosts works in the following way:

        1) While we haven't found enough mobile cores to reclaim, and there
           still exists hosts to try to reclaim more from:
            1a) For every host with potentially more cores to "give up",
                calculate its predicted CPU utilization if it gives up 1
                additional core. If the predicted value is less than or equal
                to the threshold and removing this additional core does not
                result in unreturned cores, increment the number of cores that
                this host can give up. Otherwise, mark this host as having no
                more cores to "give up".
        2) Remove any hosts that can not give up at least 1 core
        3) Sort the list of hosts based on two things: first the number of
           cores that the host can "give up" safely; second the predicted CPU
           utilization with that number of cores reclaimed
        4) Loop through the sorted list of hosts / number pairs and use each
           one to reclaim from until the number of cores that are reclaimed
           matches the number of cores required.

        This should theoretically spread out the impact of reclaiming cores
        across as many hosts as possible to reduce the chance of one host
        getting "taxed" too much.

        :param context: nova context object
        :param threshold: threshold number as defined in the policy
        :param aggregate_id: aggregate ID that the host belongs to
        :param violated_hosts: list of hosts that are deemed violated
        :param already_optimized_hosts: list of hosts that were recently
                                        optimized
        :param pool: object for the enterprise pool the host belongs to
        :param num_cores_to_reclaim: integer value for the number of mobile
                                     cores that need to be reclaimed from the
                                     other members of the pool if possible
        :return: list of "reclaims", where a reclaim is a tuple of the form:
                    (host_name, # of cores to deactivate from said host)
        """
        hosts_in_pool = pool.host_names

        # Only consider reclaiming from hosts that are in the same pool.
        # Also, do not consider hosts that are in the set of violated hosts
        # or have already been optimized since those are not good choices to
        # be reclaiming cores from this round (note that the set of violated
        # hosts includes the host we are trying to find cores for)
        potential_hosts_to_reclaim_from = (
            set(hosts_in_pool) - set(violated_hosts) -
            set(already_optimized_hosts))

        # Only consider hosts that are also in the same aggregate
        hosts_in_aggregate = agg_utils.get_hosts_for_aggregate(context,
                                                               aggregate_id)
        potential_hosts_to_reclaim_from &= set(hosts_in_aggregate)

        potential_hosts_to_reclaim_from = [
            host for host in potential_hosts_to_reclaim_from if
            self._consider_host_for_reclamation(context, host, pool)
        ]

        # Build a list of possible reclaims, where a reclaim is considered a
        # number of mobile cores to take away from a specific host
        possible_reclaims = []
        for host_name in potential_hosts_to_reclaim_from:
            compute_node = prs_utils.get_compute_node_by_service_host(
                context, host_name)
            # Truncate to 2 decimal places to avoid problems with floating
            # point comparisons
            current_cpu_util = float("%.2f" % prs_utils.get_cpu_utilization(
                context, compute_node, self._scheduler_api))
            mobile_cores = self._get_mobile_cores(compute_node)
            assigned_cores = int(math.ceil(prs_utils.get_assigned_proc_units(
                context, compute_node, self._scheduler_api)))
            active_cores = self._get_active_cores(compute_node)

            # Only consider hosts that have mobile cores that could be
            # reclaimed and that are currently < the CPU utilization
            # threshold (probably not needed, but including as a double check)
            if mobile_cores > 0 and current_cpu_util < threshold:
                possible_reclaims.append({'name': host_name,
                                          'cores_to_reclaim': 0,
                                          'predicted_cpu_util': None,
                                          'assigned_cores': assigned_cores,
                                          'active_cores': active_cores,
                                          'try_to_claim_more': True,
                                          'mobile_cores': mobile_cores,
                                          'compute_node': compute_node})

        # While we have not identified enough mobile cores to reclaim and while
        # there are still hosts to try to reclaim more mobile cores from, loop
        # through each host that we have identified as a potential host to
        # reclaim from and determine if we can reclaim an additional core from
        # it without violating the threshold and without creating unreturned
        # cores.
        cores_identified_for_reclamation = 0
        while (cores_identified_for_reclamation < num_cores_to_reclaim and
               len([r for r in possible_reclaims if r['try_to_claim_more']])
               > 0):
            for reclaim in possible_reclaims:
                if not reclaim['try_to_claim_more']:
                    # If we have determined that this host can not have any
                    # more mobile cores claimed without violating the
                    # threshold, or if the host has no more potential mobile
                    # cores to reclaim, then just continue
                    continue
                elif reclaim['cores_to_reclaim'] == reclaim['mobile_cores']:
                    # There are no more potential mobile cores to reclaim from
                    # this host, so stop trying to claim more
                    reclaim['try_to_claim_more'] = False
                    continue

                # Predict the CPU utilization if we reclaim one more core
                predicted_cpu_util = self._get_predicted_cpu_utilization(
                    context,
                    reclaim['compute_node'],
                    reclaim['cores_to_reclaim'] + 1)

                if (predicted_cpu_util <= threshold and
                        ((reclaim['active_cores'] -
                          reclaim['cores_to_reclaim'] - 1) >
                         reclaim['assigned_cores'])):
                    # We can reclaim this additional mobile core without
                    # violating the threshold and without resulting in an
                    # unreturned core, so increment the number of cores
                    # to reclaim and set the predicted CPU utilization
                    reclaim['cores_to_reclaim'] += 1
                    reclaim['predicted_cpu_util'] = predicted_cpu_util
                    cores_identified_for_reclamation += 1
                else:
                    # If we reclaim this mobile core then the threshold will be
                    # violated or it will result in an unreturned core, so
                    # stop trying to claim more
                    reclaim['try_to_claim_more'] = False

        # Remove any possible reclaims if they have cores_to_reclaim set to
        # 0 (i.e., if the host gave up any mobile cores it would violate the
        # threshold)
        possible_reclaims = [r for r in possible_reclaims
                             if r['cores_to_reclaim'] > 0]

        # Sort the reclamations first on the cores_to_reclaim to spread out the
        # CPU utilization impact across hosts, and second on the predicted
        # CPU utilization. This results in a list that conceptually looks like
        # this:     [host_name, num_cores, predicted_cpu_utilization]
        #           [hostA,     1,         0.3]
        #           [hostB,     1,         0.4]
        #           [hostC,     2,         0.2]
        #           [hostD,     2,         0.5]
        possible_reclaims = sorted(
            possible_reclaims,
            key=lambda x: (x['cores_to_reclaim'], x['predicted_cpu_util']))

        cores_reclaimed = 0
        reclaims = []
        for reclaim in possible_reclaims:
            if cores_reclaimed == num_cores_to_reclaim:
                # No more reclaims are needed
                break

            if (cores_reclaimed + reclaim['cores_to_reclaim'] >
                    num_cores_to_reclaim):
                # If reclaiming all of these cores would put us over the needed
                # amount, then only take the needed amount and break
                reclaims.append((reclaim['name'],
                                 num_cores_to_reclaim - cores_reclaimed))
                break
            else:
                reclaims.append((reclaim['name'], reclaim['cores_to_reclaim']))
                cores_reclaimed += reclaim['cores_to_reclaim']

        return reclaims

    def _get_predicted_cpu_utilization(self, context, compute_node,
                                       cores_to_remove):
        """
        Given a compute node, calculates predicted CPU utilization for a given
        host when a certain number of cores are deactivated.

        :param compute_node: compute node object
        :param cores_to_remove: number of cores to theoretically remove
        :return: predicted CPU utilization as a float
        """
        current_active_cores = self._get_active_cores(compute_node)
        current_cpu_utilization = prs_utils.get_cpu_utilization(
            context, compute_node, self._scheduler_api)

        current_proc_units = current_cpu_utilization * current_active_cores
        new_active_cores = current_active_cores - cores_to_remove

        if new_active_cores <= 0:
            LOG.debug('Predicted CPU utilization is being returned as 100%% '
                      'because the new number of active cores is <= 0, likely '
                      'due to an out-of-band operation (current active cores: '
                      '%(curr_active)s; cores to remove: %(remove)s).' %
                      {'curr_active': current_active_cores,
                       'remove': cores_to_remove})
            return 100.0

        # Truncate the result to 2 decimal places to avoid problems with
        # floating point comparisons
        return float("%.2f" % (current_proc_units / new_active_cores))

    def _log_decision(self, policy_action, host_name, pool, cores_to_take,
                      reclaims):
        """
        Given a number of cores to take from a pool and the information for
        what hosts to reclaim from and how many cores to reclaim from each
        host, log messages about what was decided.
        """
        if cores_to_take:
            LOG.info(_('The optimizer is in %(advise)s mode; '
                       '%(cores_to_take)d mobile cores should be taken from '
                       'the available mobile cores in pool %(pool_uuid)s and '
                       'activated on host %(host)s.') %
                     {'advise': policy_action,
                      'cores_to_take': cores_to_take,
                      'pool_uuid': pool.uuid,
                      'host': host_name})

        for host_to_reclaim_from, num in reclaims:
            LOG.info(_('The optimizer is in %(advise)s mode; '
                       '%(num_to_reclaim)d mobile cores should be reclaimed '
                       'from host %(host_to_reclaim_from)s.') %
                     {'advise': policy_action,
                      'num_to_reclaim': num,
                      'host_to_reclaim_from': host_to_reclaim_from})

    def _generate_notification_payload(self, context, policy_action,
                                       policy_info, host_name, agg_id,
                                       cores_to_take, reclaims, pool):
        """
        Generates a notification payload.
        """
        context.request_id = (ego_consts.RTPOLICY_PREFFIX + context.request_id)
        ego_host_dict = self._policy_engine_manager.ego_host_dict
        alert_times = self._policy_engine_manager.get_violation_count(
            host_name, agg_id)
        host_dict = ego_host_dict[host_name]

        # cores_to_take represents the number of cores to simply take from any
        # cores that the pool has available (just laying around, not in use)
        # reclaims will be a list of (host_name, # cores to reclaim)
        # tuples that will be needed
        payload = {'cores_to_take': cores_to_take,
                   'reclaims': reclaims,
                   'pool_uuid': pool.uuid,
                   'action_name': str(policy_action),
                   'hypervisor_id': host_dict['node_id'],
                   'hypervisor_hostname': host_dict['node_name'],
                   'rtpolicy_id': policy_info['policy_id'],
                   'rtpolicy_name': policy_info['policy_name'],
                   'threshold': policy_info['policy_threshold'],
                   'stabilization': policy_info['policy_stabilization'],
                   'run_interval': policy_info['policy_run_interval'],
                   'is_violated': 'Yes',
                   'alert_times': alert_times,
                   'aggregate_id': agg_id}

        return payload

    def _reclaim_from_hosts(self, context, reclaims):
        """
        Reclaims mobile cores from hosts and makes those mobile cores available
        to the rest of the pool.

        :param context: nova context object
        :param reclaims: list of tuples of the form: (host_name,
                                                      # of cores to reclaim)
        :return: updated list of reclaim tuples with the actual number of cores
                 that were reclaimed
        """
        actual_reclaims = []
        for host, num in reclaims:
            returned_cores = self._return_mobile_cores(context, host, num)

            # Update the reclaims dictionary to represent what was actually
            # reclaimed
            actual_reclaims.append((host, returned_cores))

            if returned_cores == 0:
                LOG.warning(_('The optimizer could not successfully reclaim '
                              'any mobile cores from host %(host)s.') %
                            {'host': host})
            elif returned_cores < num:
                LOG.warning(_('The optimizer could only reclaim '
                              '%(returned_cores)s mobile cores from host '
                              '%(host)s.') %
                            {'returned_cores': returned_cores,
                             'host': host})
            else:
                LOG.info(_('The optimizer successfully reclaimed %(num)d '
                           'mobile cores from host %(host)s.') %
                         {'num': num,
                          'host': host})

        return actual_reclaims

    def optimize_hosts(self, context, violated_hosts, already_optimized_hosts,
                       policy_info):
        """Optimize the hosts via enterprise pool CoD operations."""
        LOG.debug("The following hosts have already been optimized this cycle "
                  "and will not undergo enterprise pool CoD-based "
                  "optimization: %s"
                  % six.text_type(already_optimized_hosts))

        # Do not consider the already optimized hosts
        violated_hosts = [host for host in violated_hosts
                          if host[0] not in already_optimized_hosts]

        policy_action = policy_info['policy_action']
        agg_id = policy_info['aggregate_id']
        threshold = policy_info['policy_threshold']
        ret = dict(code=ego_consts.OPTIMIZE_HOSTS_NOT_PERFORMED_CONTINUE,
                   optimized_hosts=[])

        # Query which actions are "advise only" actions
        advise_actions = self._rtpolicy_driver.get_advise_policy_actions()

        for host in violated_hosts:
            host_name = host[0]
            pool = common_ep.get_pool_for_host(context, host_name)
            if (policy_action in list(pvc_prs_consts.COD_POLICY_ACTIONS) and
                    self._check_optimization_host_prereqs(context, host_name,
                                                          pool)):
                LOG.debug('Determining whether host %s can be optimized via '
                          'enterprise pool CoD operations.' % host_name)
                (cores_to_take,
                 cores_to_reclaim) = self._calculate_core_requirements(
                     context, host_name, threshold)

                if cores_to_take == 0 and cores_to_reclaim == 0:
                    # There's nothing to do; the host doesn't need any cores!
                    continue

                able_to_do_cod_operation = True

                reclaims = []
                if cores_to_reclaim:
                    reclaims = self._find_hosts_to_reclaim_from(
                        context, threshold, agg_id, violated_hosts,
                        already_optimized_hosts, pool, cores_to_reclaim)

                    if cores_to_take == 0 and not reclaims:
                        LOG.info(_('The pool does not have any available '
                                   'mobile cores and no hosts within the pool '
                                   'are able to give up their mobile cores '
                                   'without violating the threshold or '
                                   'creating unreturned cores. Therefore '
                                   'there is nothing that the optimizer can '
                                   'do to optimize host %(host)s via '
                                   'enterprise pool CoD operations.') %
                                 {'host': host_name})
                        able_to_do_cod_operation = False

                self._log_decision(policy_action, host_name, pool,
                                   cores_to_take, reclaims)

                if able_to_do_cod_operation:
                    payload = self._generate_notification_payload(
                        context, policy_action, policy_info, host_name, agg_id,
                        cores_to_take, reclaims, pool)

                    # Send a notification marking the start of enterprise pool
                    # CoD optimization for this host
                    self._policy_engine_manager.store_event(
                        context,
                        ego_consts.PRS_RTPOLICY_ACTION_START,
                        payload)

                    if policy_action in advise_actions:
                        # We're only advising on what to do; omit the
                        # enterprise pool CoD actions
                        payload['action_result'] = 'success'
                        ret['code'] = (ego_consts.
                                       OPTIMIZE_HOSTS_PERFORMED_CONTINUE)
                        ret['optimized_hosts'].append(host_name)

                        self._policy_engine_manager.store_event(
                            context,
                            ego_consts.PRS_RTPOLICY_ACTION_END,
                            payload)

                        # Move on to the next violated host
                        continue

                    # Do the enterprise pool CoD actions
                    if reclaims:
                        # If mobile cores need to be reclaimed (i.e. put back
                        # into the pool), then do that
                        reclaims = self._reclaim_from_hosts(context, reclaims)
                        payload['reclaims'] = reclaims

                        total_cores_reclaimed = sum(rec[1] for rec in reclaims)

                        if total_cores_reclaimed <= 0:
                            # If we failed to reclaim a core, then the
                            # optimizer failed
                            payload['action_result'] = 'failed'
                            self._policy_engine_manager.store_event(
                                context,
                                ego_consts.PRS_RTPOLICY_ACTION_END,
                                payload)

                    if not reclaims or total_cores_reclaimed > 0:
                        # Sanity check to make sure the number of available
                        # cores in the pool is what we think it is
                        pool = common_ep.get_pool_for_host(context, host_name)

                        cores_to_add = cores_to_take + cores_to_reclaim
                        if cores_to_add > pool.available_mobile_cores:
                            LOG.warning(_('The pool only has %(avail)s mobile '
                                          'cores, but %(expected)s mobile '
                                          'cores were expected to be '
                                          'available. This is likely due to '
                                          'an out-of-band enterprise pool CoD '
                                          'operation. The optimizer will only '
                                          'try to add the number of available '
                                          'cores in the pool to host '
                                          '%(host)s.') %
                                        {'avail': pool.available_mobile_cores,
                                         'expected': cores_to_add,
                                         'host': host_name})
                            cores_to_add = pool.available_mobile_cores

                        # Activate mobile cores on this specific violated host
                        # to satisfy the threshold
                        activation_succeeded = self._add_mobile_cores(
                            context, host_name,
                            cores_to_add)

                        if activation_succeeded:
                            LOG.info(_('The optimizer successfully added '
                                       '%(num)d mobile core(s) to host '
                                       '%(host)s.') %
                                     {'num': cores_to_add,
                                      'host': host_name})

                            ret['code'] = (
                                ego_consts.OPTIMIZE_HOSTS_PERFORMED_CONTINUE)
                            ret['optimized_hosts'].append(host_name)

                            payload['action_result'] = 'success'
                            self._policy_engine_manager.store_event(
                                context,
                                ego_consts.PRS_RTPOLICY_ACTION_END,
                                payload)

                            # Move on to the next violated host
                            continue
                        else:
                            # Failed to activate mobile cores for some reason
                            payload['action_result'] = 'failed'
                            self._policy_engine_manager.store_event(
                                context,
                                ego_consts.PRS_RTPOLICY_ACTION_END,
                                payload)

            migrate_actions = (
                list(pvc_prs_consts.COD_MIGRATE_POLICY_ACTIONS) +
                [ego_consts.POLICY_ACTION_MIGRATE_VM,
                 ego_consts.POLICY_ACTION_MIGRATE_VM_ADVISE_ONLY])
            if policy_action in migrate_actions:
                # We need to try the optimize this specific host using
                # migration since it was not able to be optimized via
                # enterprise pool CoD actions
                migration_optimizer_ret = (
                    self._migration_optimizer.optimize_hosts(
                        context, [host],
                        already_optimized_hosts + ret['optimized_hosts'],
                        policy_info))

                ret['code'] = migration_optimizer_ret['code']
                ret['optimized_hosts'] += (
                    migration_optimizer_ret['optimized_hosts'])

        return ret

    def _return_mobile_cores(self, context, host, cores_to_return):
        """
        Helper method for wrapping set_mobile_cores for removing (returning)
        mobile cores to the pool.

        :return: number of cores that were removed from the host that did not
                 create unreturned cores
        """
        LOG.info(_('Returning %(num)d mobile cores from host %(host)s back to '
                   'its pool.') % {'num': cores_to_return,
                                   'host': host})
        compute_node = prs_utils.get_compute_node_by_service_host(context,
                                                                  host)
        current_mobile_cores = self._get_mobile_cores(compute_node)
        result = self._set_mobile_cores(context, host,
                                        current_mobile_cores - cores_to_return)

        if not result:
            # Something went wrong while setting the mobile cores, so return 0
            return 0

        # Fetch the compute node again and check if there are any unreturned
        # cores. If so, we should undo the return of mobile cores since the
        # optimizer should never be creating unreturned cores. This can happen
        # if a deploy or OOB operation occurred between the time we decided
        # what reclamations to do. We will only undo up to minimum between the
        # number of unreturned cores and the cores which were returned, which
        # will potentially allow for us to still do an optimization with a
        # smaller number of cores.
        compute_node = prs_utils.get_compute_node_by_service_host(context,
                                                                  host)
        unreturned_cores = self._get_unreturned_mobile_cores(compute_node)

        if unreturned_cores > 0:
            # Since we just removed them, there should be at least
            # "unreturned_cores" available in the pool, so add those back to
            # this host
            result = self._add_mobile_cores(
                context, host, min(unreturned_cores, cores_to_return))

            if not result:
                # Something went wrong while setting the mobile cores. Since we
                # already reclaimed them, we will just go ahead and continue
                # instead of retrying to add them back in a loop
                return cores_to_return

            # Go ahead and return the maximum between 0 and the difference
            # between the original cores_to_return and the number of unreturned
            # cores we just returned, as there could potentially have been an
            # out-of-band operation that resulted in unreturned_cores >
            # cores_to_return
            return max(0, cores_to_return - unreturned_cores)

        # Everything worked as expected
        return cores_to_return

    def _add_mobile_cores(self, context, host, cores_to_add):
        """
        Helper method for wrapping set_mobile_cores for adding mobile cores
        """
        LOG.info(_('Adding %(num)d mobile cores to host %(host)s.') %
                 {'num': cores_to_add,
                  'host': host})
        compute_node = prs_utils.get_compute_node_by_service_host(context,
                                                                  host)
        current_mobile_cores = self._get_mobile_cores(compute_node)
        return self._set_mobile_cores(context, host,
                                      current_mobile_cores + cores_to_add)

    def _set_mobile_cores(self, context, host, desired_mobile_cores):
        """
        Helper method for using the compute API to set the number of mobile
        cores for a given host.
        """
        LOG.info(_('Setting mobile cores for host %(host)s to '
                   '%(desired_mobile_cores)d.') %
                 {'host': host,
                  'desired_mobile_cores': desired_mobile_cores})

        try:
            result = self._compute_api.set_mobile_cores(
                context, host, desired_mobile_cores)
        except Exception as e:
            LOG.error(_('A problem occurred while trying to set the mobile '
                        'cores for host %(host)s to %(num)d.') %
                      {'host': host,
                       'num': desired_mobile_cores})
            LOG.exception(e)
            return False

        if not result[0] and len(result) == 1:
            # Result says that the operation failed but gave no reason
            LOG.error(_('A problem occurred while trying to set the mobile '
                        'cores for host %(host)s to %(num)d.') %
                      {'host': host,
                       'num': desired_mobile_cores})
            return False
        elif not result[0]:
            LOG.error(_('Failed to set the mobile cores for host %(host)s to '
                        '%(num)d due to: %(reason)s.') %
                      {'host': host,
                       'num': desired_mobile_cores,
                       'reason': result[1]})
            return False
        else:
            LOG.info(_('Successfully set the mobile cores for host %(host)s '
                       'to %(num)d.') % {'host': host,
                                         'num': desired_mobile_cores})
            return True

    def _need_to_check_managing_hmc(self, compute_node):
        """Returns a boolean representing whether or not the HMC managing the
        host needs to be checked to see if it is the master HMC of the
        enterprise pool.

        Enterprise pool POST operations to non-master HMCs will fail if the HMC
        version is older than the 850 release.
        """
        platform = self._get_host_stat(compute_node, constants.MANAGED_BY)

        if platform != constants.HMC_PLATFORM:
            return False

        hmc_version = self._get_host_stat(compute_node, constants.HMC_VERSION)

        if not hmc_version:
            LOG.warning(_('Version of HMC managing host %s could not be '
                          'determined.') % compute_node.host)
            return False

        hmc_version = mgmt_console_wrapper.HMCVersion(hmc_version)
        hmc_version_850 = mgmt_console_wrapper.HMCVersion('V8R8.5.0')
        return hmc_version < hmc_version_850

    def _get_mobile_cores(self, compute_node):
        """Returns the CoD mobile cores stat for a given node"""
        return int(self._get_host_stat(
            compute_node,
            constants.COD_MOBILE_CORES,
            default=0))

    def _get_active_cores(self, compute_node):
        """Returns the active cores stat for a given node"""
        return int(float(self._get_host_stat(
            compute_node,
            constants.PROC_UNITS,
            default=0)))

    def _get_inactive_cores(self, compute_node):
        """Returns the inactive cores stat for a given node"""
        return int(self._get_host_stat(
            compute_node,
            constants.COD_INACTIVE_CORES,
            default=0))

    def _get_unreturned_mobile_cores(self, compute_node):
        """Returns the CoD unreturned mobile cores stat for a given node"""
        return int(self._get_host_stat(
            compute_node,
            constants.COD_UNRETURNED_MOBILE_CORES,
            default=0))

    def _get_host_stat(self, compute_node, key, default=None):
        """Helper method to get a value from the node's stats dictionary"""
        return compute_node.get('stats', {}).get(key, default)
