# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from nova.ibm.ego import ego_constants
from oslo_log import log as logging

from powervc_nova.common import server_group_utils as sg_utils
from powervc_nova.scheduler.filters import PowerVCBaseFilter
from powervc_nova import logcall
from powervc_nova import _

LOG = logging.getLogger(__name__)


class PowerVCServerGroupAntiAffinityFilter(PowerVCBaseFilter):
    """Schedule the instance on a different host from a set of group
    hosts. Overrides EGO filter since OpenStack only supports
    one affinity rule per VM.
    """
    def __init__(self):
        super(PowerVCServerGroupAntiAffinityFilter, self).__init__()
        self.policy_name = 'anti-affinity'
        self._set_filter_scope(
            [ego_constants.FILTER_WORKING_SCOPE_CREATE,
             ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION,
             # NOTE(jwcroppe): If cold migration is implemented in PowerVC,
             # cold migration vm_states and task_states will need
             # to be added to the PRHOHIBITED_VM_STATES and
             # PROHIBITED_TASK_STATES in
             # powervc_nova.common.server_group_utils.py. This will
             # prevent requests for a VM in a group from being
             # scheduled if another VM in the same group is currently
             # in the process of a cold migration.
             ego_constants.FILTER_WORKING_SCOPE_COLD_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_EVACUATE,
             ego_constants.FILTER_WORKING_SCOPE_UNSHELVE])
        self._set_error_message(_("The attempted operation would violate an "
                                  "anti-affinity collocation rule or there "
                                  "is an ongoing operation for a "
                                  "different virtual machine in the same "
                                  "anti-affinity collocation rule as the "
                                  "virtual machine in the scheduling request. "
                                  "This ongoing operation could result in the "
                                  "other virtual machine relocating to a new "
                                  "host and violating the collocation rule. "
                                  "Wait until ongoing operations complete for "
                                  "all virtual machines in the same "
                                  "collocation rules as the virtual machine "
                                  "in the request, then try the request "
                                  "again."))

    @logcall
    def parse(self, filter_properties, all_host_attrs):
        # Return if ignore_collocaiton_rules conf option is True
        if sg_utils.ignore_collocation_rules():
            return
        context = filter_properties.get('context')
        request_spec = filter_properties.get('request_spec', {})
        instance_uuid = request_spec.get('instance_id')
        instance_props = request_spec.get('instance_properties', {})
        operation = request_spec.get('operation_type')
        source_host = instance_props.get('host')

        # Get anti-affinity groups this instance belongs to
        groups = sg_utils.get_groups_for_instance_by_policy(
            context, instance_uuid, self.policy_name)

        # Only invoke the filter if instance is a member of any
        # 'anti-affinity' groups
        if not groups:
            return

        all_member_hosts = set()
        for group in groups:
            # Get back instance objects for each member in the group
            instances = sg_utils.get_members(context, group.uuid)

            # Make sure there are no ongoing create, rebuild, or  migrate
            # operations for other instances in this group. Fail the
            # scheduling request if there are any, so potential race
            # conditions that would violate group's policy are blocked.
            has_ongoing_operation = sg_utils.has_ongoing_operations_in_group(
                group.uuid, instance_uuid, instances, operation,
                request_spec=request_spec)
            if has_ongoing_operation:
                self._include_hosts([])
                return

            # Get a list of tuples of uuid and host for each instance in group
            member_hosts = sg_utils.get_hosts_for_each_member(
                context, group, members=instances, has_host_only=False)
            # Add unique uuid and host tuples to all_groups_member_info
            all_member_hosts = all_member_hosts.union(set(member_hosts))

        # If this is a create operation, the VM will only be in one group.
        # The group_hosts property in filter_properties should be used for
        # filtering, because it is updated during multi-instance deploy each
        # iteration through the loop of instances.
        exclude_hosts = None
        if operation == ego_constants.FILTER_WORKING_SCOPE_CREATE:
            exclude_hosts = filter_properties.get('group_hosts', [])
        else:
            # Hosts to exclude are any hosts in the all_member_hosts list,
            # except the host of the instance in the request.
            exclude_hosts = set([member_info[1] for member_info in
                                 all_member_hosts if member_info[1]])
            if source_host in exclude_hosts:
                exclude_hosts.remove(source_host)

        host_to_hyper_map = self._get_hosts_to_hypervisors_map(context)
        exclude_hosts = [host_to_hyper_map[host]
                         for host in exclude_hosts
                         if host in host_to_hyper_map]
        self._exclude_hosts(list(exclude_hosts))

        LOG.debug("Setting excluded hosts in %(filter)s (%(hosts)s)." %
                  dict(filter=self.__class__.__name__,
                       hosts=self._exclude_hosts))


class PowerVCServerGroupAffinityFilter(PowerVCBaseFilter):
    """Schedule the instance on a different host from a set of group
    hosts. Overrides EGO filter since OpenStack only supports
    one affinity rule per VM.
    """
    def __init__(self):
        super(PowerVCServerGroupAffinityFilter, self).__init__()
        self.policy_name = 'affinity'
        self._set_filter_scope(
            [ego_constants.FILTER_WORKING_SCOPE_CREATE,
             ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION,
             # NOTE(jwcroppe): If cold migration is implemented in PowerVC,
             # cold migration vm_states and task_states will need
             # to be added to the PRHOHIBITED_VM_STATES and
             # PROHIBITED_TASK_STATES in
             # powervc_nova.common.server_group_utils.py. This will
             # prevent requests for a VM in a group from being
             # scheduled if another VM in the same group is currently
             # in the process of a cold migration.
             ego_constants.FILTER_WORKING_SCOPE_COLD_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_EVACUATE,
             ego_constants.FILTER_WORKING_SCOPE_UNSHELVE])
        self._set_error_message(_("The attempted operation would violate an "
                                  "affinity collocation rule or there is "
                                  "an ongoing operation for a "
                                  "different virtual machine in the same "
                                  "affinity collocation rule as the virtual "
                                  "machine in the scheduling request. This "
                                  "ongoing operation could result in the "
                                  "other virtual machine relocating to a new "
                                  "host and violating the collocation rule. "
                                  "Wait until ongoing operations complete for "
                                  "all virtual machines in the same "
                                  "collocation rules as the virtual machine "
                                  "in the request, then try the request "
                                  "again."))

    @logcall
    def parse(self, filter_properties, all_host_attrs):
        # Return if ignore_collocaiton_rules conf option is True
        if sg_utils.ignore_collocation_rules():
            return
        context = filter_properties.get('context')
        request_spec = filter_properties.get('request_spec', {})
        instance_uuid = request_spec.get('instance_id')
        operation = request_spec.get('operation_type')
        group_hosts = filter_properties.get('group_hosts', [])

        # Get affinity groups this instance belongs to
        groups = sg_utils.get_groups_for_instance_by_policy(
            context, instance_uuid, self.policy_name)

        # Only invoke the filter if instance is a member of any
        # 'anti-affinity' groups
        if not groups:
            return

        # If this is a create operation, the VM will only be in one group.
        # The group_hosts property in filter_properties should be used for
        # filtering, because it is updated during multi-instance deploy each
        # iteration through the loop of instances.
        include_hosts = None
        if operation == ego_constants.FILTER_WORKING_SCOPE_CREATE:
            group_hosts = filter_properties.get('group_hosts', [])
            # Get instances belonging to the group
            instances = sg_utils.get_members(context, groups[0].uuid)

            # Make sure no other instances in the group are undergoing create,
            # rebuild, or migrate operations. Fail the create request if there
            # are any, so potential race conditions that would violate the
            # policy of the group are blocked.
            has_ongoing_operation = sg_utils.has_ongoing_operations_in_group(
                groups[0].uuid, instance_uuid, instances, operation,
                request_spec=request_spec)
            if has_ongoing_operation:
                include_hosts = []
            elif group_hosts:
                include_hosts = group_hosts
        else:
            all_member_hosts = set()
            only_member_in_every_group = True
            for group in groups:
                # Get back instance objects for each member in the group
                instances = sg_utils.get_members(context, group.uuid)
                # If this is the only instance in the group
                if len(instances) > 1:
                    only_member_in_every_group = False

                # Get a list of tuples of instance uuid and host for each
                # instance in group
                member_hosts = sg_utils.get_hosts_for_each_member(
                    context, group, members=instances, has_host_only=False)
                # Add unique tuples of instance uuid and host tuples to
                # all_groups_member_info.
                all_member_hosts = all_member_hosts.union(set(member_hosts))

            # If this is the only instance in the group for every group,
            # it can be placed on any host. Otherwise, use all_member_hosts.
            if not only_member_in_every_group and all_member_hosts:
                # Hosts to include are any hosts in the all_member_hosts list.
                # Extract host names from the tuples.
                include_hosts = set([member_info[1] for member_info in
                                     all_member_hosts if member_info[1]])

        if include_hosts is not None:
            host_to_hyper_map = self._get_hosts_to_hypervisors_map(context)
            include_hosts = [host_to_hyper_map[host]
                             for host in include_hosts
                             if host in host_to_hyper_map]
            self._include_hosts(list(include_hosts))

        LOG.debug("Setting included hosts in %(filter)s (%(hosts)s)." %
                  dict(filter=self.__class__.__name__,
                       hosts=self._include_hosts))
