# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from oslo_log import log as logging

from nova.ibm.ego import ego_constants

from powervc_nova.common import aggregate_utils as agg_utils
from powervc_nova.scheduler.filters import PowerVCBaseFilter
from powervc_nova import logcall

LOG = logging.getLogger(__name__)


class PowerVCHostAggregateFilter(PowerVCBaseFilter):
    """
    This EGO scheduler filter is required to keep VMs within a host aggregate
    domain, particularly for VMs that were created prior to PowerVC 1.2.3 and
    on-boarded VMs--essentially because these VMs won't have the reserveon(x)
    such that x != 'group_all' specified (i.e., because they were booted
    without the host_aggregate_id=x hint), which would otherwise allow the VM
    to be migrated anywhere within the environment.
    """

    def __init__(self):
        super(PowerVCHostAggregateFilter, self).__init__()
        self._set_filter_scope(
            [ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_COLD_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_EVACUATE,
             ego_constants.FILTER_WORKING_SCOPE_UNSHELVE])

    @logcall
    def parse(self, filter_properties, all_host_attrs):
        def generate_expr(agg_id):
            return '(%s <in> $host_aggregate_ids)' % str(agg_id)

        request_spec = filter_properties.get('request_spec', {})
        context = filter_properties.get('context')
        instance_props = request_spec.get('instance_properties', {})
        source_host = instance_props.get('host')

        # Retrieve a list of aggregates in which the host resides.
        agg_ids = agg_utils.get_aggregate_ids_for_host(context, source_host)
        if not agg_ids:
            # Host wasn't a member of any aggregates; no filtering necessary.
            return

        # Generate the resource requirement string.
        expressions = [generate_expr(agg_id) for agg_id in agg_ids]

        req_str = '||'.join(expressions)
        self._select(req_str)
        LOG.debug("After passing the %(filter)s, the resource requirement "
                  "expression is (%(exp)s)." % dict(
                      filter=self.__class__.__name__, exp=req_str))
