# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import six

from oslo_log import log as logging

from nova.ibm.ego import ego_constants

from powervc_nova.scheduler.filters import PowerVCBaseFilter
from powervc_nova.virt.powerkvm import constants as pkvmc
from powervc_nova import logcall
from powervc_nova import _

LOG = logging.getLogger(__name__)

MIGRATION_VERSION_MATRIX = {
    '3': ['2'],           # 3.* -> 2.* is not supported
    '2.1.0': ['3.1'],     # 2.1.0.* -> 3.1.* is not supported
    '2.1.1': ['2.1.0']    # 2.1.1.* -> 2.1.0.* is not supported
}


def convert_version_to_list(version):
    '''
    Takes in a version as a string (e.g., "1.2.3") and converts it to a list of
    integers (e.g., ['1', '2', '3']).
    '''
    return [i for i in version.split('.')]


class PowerKVMMigrationVersionFilter(PowerVCBaseFilter):
    """
    This EGO scheduler filter that ensures VMs are not migrated between
    incompatible PowerKVM hosts, such as from PowerKVM 3 to PowerKVM 2.
    """

    def __init__(self):
        super(PowerKVMMigrationVersionFilter, self).__init__()
        self._set_filter_scope(
            [ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION])

    @logcall
    def parse(self, filter_properties, all_host_attrs):
        if not self._allow_oper_for_powerkvm(filter_properties,
                                             all_host_attrs):
            LOG.debug("Ignoring %s..." % self.__class__.__name__)
            return

        request_spec = filter_properties.get('request_spec', {})
        instance_properties = request_spec.get('instance_properties', {})
        source_hostname = instance_properties.get('ego_host')
        inst_uuid = instance_properties.get('uuid')
        host_data = self._get_source_host_data(instance_properties,
                                               all_host_attrs)

        source_pkvm_version = host_data.get('stats', {}).get(
            pkvmc.IBM_POWERKVM_VERSION)

        if not source_pkvm_version:
            LOG.warning(_('The IBM PowerKVM version for source host '
                          '%(source_host)s is None; therefore, no hosts '
                          'will be excluded based on the IBM PowerKVM '
                          'version.')
                        % {'source_host': source_hostname})
            return

        # list of versions which a VM on the source powerkvm version cannot
        # migrate to
        invalid_versions = self._get_invalid_versions(source_pkvm_version)

        if not invalid_versions:
            # There are no versions of KVM which this VM cannot migrate to, so
            # just return early
            return

        if '*' in invalid_versions:
            self._set_error_message(
                _('No suitable host could be found because the source host '
                  '(%(source_host)s) of the instance (%(instance)s) has an '
                  'IBM PowerKVM version of %(version)s, which does not '
                  'support migration.') %
                {'source_host': source_hostname,
                 'instance': inst_uuid,
                 'version': source_pkvm_version})
            self._include_hosts([])
            return

        exclude_hosts = []

        # Iterate over all hosts and check if they are valid destinations
        for hostname, host_data in six.iteritems(all_host_attrs):
            pkvm_version = host_data.get('stats', {}).get(
                pkvmc.IBM_POWERKVM_VERSION)
            if not pkvm_version:
                LOG.warning(_('The IBM PowerKVM version for host %(hostname)s '
                              'is None; therefore, this host will not be '
                              'excluded based on the IBM PowerKVM version.')
                            % {'hostname': hostname})
                continue

            if not self._can_migrate(pkvm_version, invalid_versions):
                LOG.debug("Filtered host (%(host)s) for instance (%(inst)s) "
                          "because instance is on a PowerKVM host "
                          "(%(source_host)s) with version %(source_version)s "
                          "which cannot be migrated to a PowerKVM host with "
                          "version %(target_version)s." %
                          {'host': hostname,
                           'inst': inst_uuid,
                           'source_host': source_hostname,
                           'source_version': source_pkvm_version,
                           'target_version': pkvm_version})
                exclude_hosts.append(hostname)
                continue

        self._exclude_hosts(exclude_hosts)

    def _get_invalid_versions(self, source_powerkvm_version):
        '''
        Given a source host version, determine all versions of PowerKVM which
        a VM on the source host cannot migrate to.

        :param source_powerkvm_version: string representing the source host
                                        PowerKVM OS verison, such as "3.1.0"
        :returns: list of strings representing all versions of PowerKVM which
                  a VM on the source host cannot migrate to, such as:
                      ["2", "1.4"]
                  If there are no restricted versions, then an empty list is
                  returned. If there is an error while parsing the source
                  PowerKVM version, None is returned.
        '''
        try:
            source_powerkvm_version = convert_version_to_list(
                source_powerkvm_version)
        except Exception as e:
            LOG.error(_('The IBM PowerKVM version was in an unexpected '
                        'format: %s; therefore, no filtering based on '
                        'PowerKVM version will be performed.') %
                      source_powerkvm_version)
            LOG.exception(e)
            return None

        invalid_versions = []

        # generate all possible versions that could be in the support matrix
        # for example, for ['1', '2', '3'], then generate:
        #    ['1'], ['1', '2'], ['1', '2', '3']
        possible_source_versions = [source_powerkvm_version[0:i] for i in
                                    range(1, len(source_powerkvm_version) + 1)]

        # append an asterisk for the case when not a single version can be
        # migrated to a specific version
        possible_source_versions.append(['*'])

        for possible_source_version in possible_source_versions:
            version_key = '.'.join(possible_source_version)
            if version_key in MIGRATION_VERSION_MATRIX:
                value = MIGRATION_VERSION_MATRIX[version_key]
                if isinstance(value, list):
                    invalid_versions.extend(value)
                else:
                    invalid_versions.append(value)

        return invalid_versions

    def _can_migrate(self, target_powerkvm_version, invalid_versions):
        '''
        Determines if the target host PowerKVM version is in the list of all
        PowerKVM versions that are not valid migration targets, and if so,
        returns that the VM cannot migrate to the target host.

        :param: target_powerkvm_version: string representing the target host's
                                         PowerKVM version, such as "3.1.0"
        :param: invalid_versions: list of strings representing all PowerKVM
                                  versions that are not valid migration targets
        :returns: False if the target host PowerKVM version is in the list of
                  invalid versions, or a base version of the target version (
                  e.g., 3.2 for 3.2.1) is in the list of invalid versions.
                  True otherwise.  Also returns True if there is an error
                  parsing the target PowerKVM version string.
        '''
        if '*' in invalid_versions:
            return False

        try:
            target_powerkvm_version = convert_version_to_list(
                target_powerkvm_version)
        except Exception as e:
            LOG.error(_('The IBM PowerKVM version was in an unexpected '
                        'format: %s') % target_powerkvm_version)
            LOG.exception(e)
            # default to allowing the migration to take place
            return True

        possible_target_versions = [target_powerkvm_version[0:i] for i in
                                    range(len(target_powerkvm_version) + 1)]

        for possible_target_version in possible_target_versions:
            version_key = '.'.join(possible_target_version)
            if version_key in invalid_versions:
                return False

        return True
