# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from oslo_log import log as logging
from nova import utils

from nova.ibm.ego import ego_constants

from powervc_nova.scheduler.filters import PowerVCBaseFilter
from powervc_nova.virt.ibmpowervm.common import constants as pvmc
from powervc_nova import logcall
from powervc_nova import _

LOG = logging.getLogger(__name__)


class PowerVMGuestCapableFilter(PowerVCBaseFilter):
    """
    This EGO scheduler filter ensures that the PowerVM host supports running
    the virtual machine's operating system type.  For example, there are
    PowerLinux servers that are only capable of running Linux guest operating
    systems and not AIX or IBM i.
    """

    run_filter_once_per_request = True

    def __init__(self):
        super(PowerVMGuestCapableFilter, self).__init__()
        self._set_filter_scope(
            [ego_constants.FILTER_WORKING_SCOPE_CREATE,
             ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_COLD_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_EVACUATE,
             ego_constants.FILTER_WORKING_SCOPE_UNSHELVE])

    @logcall
    def parse(self, filter_properties, all_host_attrs):
        if not self._allow_oper_for_powervm(filter_properties, all_host_attrs):
            LOG.debug("Ignoring %s..." % self.__class__.__name__)
            return

        request_spec = filter_properties.get('request_spec', {})
        image = request_spec.get('image', {})
        image_name = image.get('name', '<?>')
        image_props = image.get('properties', {})
        os_distro = image_props.get('os_distro')

        if os_distro is None:
            LOG.debug("There was no 'os_distro' image property set for "
                      "image '(%(image_name)s)'; checking system metadata."
                      % dict(image_name=image_name))
            instance_props = request_spec.get('instance_properties', {})
            sys_meta = utils.instance_sys_meta(instance_props) or {}
            os_distro = sys_meta.get('image_os_distro')
            if os_distro is None:
                return

        os_distro = os_distro.lower()
        req_str = None

        if os_distro in ('rhel', 'sles', 'ubuntu'):
            req_str = "$%(cap)s == True" % dict(cap=pvmc.LINUX_CAPABLE)
        elif os_distro == 'aix':
            req_str = "$%(cap)s == True" % dict(cap=pvmc.AIX_CAPABLE)
        elif os_distro == 'ibmi':
            req_str = "$%(cap)s == True" % dict(cap=pvmc.IBMI_CAPABLE)
        else:
            LOG.warning(_("Unexpected value for 'os_distro' image property "
                          "set for image (%(image_name)s); no filtering to "
                          "perform.") % dict(image_name=image_name))

        if req_str is not None:
            self._select(req_str)
            LOG.debug("After passing the %(filter)s, the resource requirement "
                      "expression is (%(exp)s)." % dict(
                          filter=self.__class__.__name__, exp=req_str))
