# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from oslo_log import log as logging

from nova.ibm.ego import ego_constants

from powervc_nova.network.neutronv2 import ibmpowervm_api
from powervc_nova.scheduler.filters import PowerVCBaseFilter
from powervc_nova import logcall
from powervc_nova import _

LOG = logging.getLogger(__name__)


class PowerVMNetworkFilter(PowerVCBaseFilter):
    """
    This EGO scheduler filter eliminates candidate hosts whose logical network
    settings are not compatible with the respective LPAR. The filter is
    applicable for deploy and migration operations.
    """

    run_filter_once_per_request = True

    def __init__(self, pvm_api=None):
        """
        :param pvm_api: Optional API (a way for tests to override the API)
        """
        super(PowerVMNetworkFilter, self).__init__()
        self._set_filter_scope(
            [ego_constants.FILTER_WORKING_SCOPE_CREATE,
             ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_COLD_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_EVACUATE,
             ego_constants.FILTER_WORKING_SCOPE_UNSHELVE])
        self._set_error_message(_("No suitable host could be found due to the "
                                  "network configuration. Please ensure that "
                                  "PowerVC's network SEA mappings are "
                                  "properly defined and retry your request."))
        self._powervm_api = (pvm_api if pvm_api is not None
                             else ibmpowervm_api.PowerVMAPI())

    @logcall
    def parse(self, filter_properties, all_host_attrs):
        if not self._allow_oper_for_powervm(filter_properties, all_host_attrs):
            LOG.debug("Ignoring %s..." % self.__class__.__name__)
            return

        request_spec = filter_properties.get('request_spec', {})
        instance_props = request_spec.get('instance_properties', {})
        requested_networks = filter_properties.get('requested_networks')
        source_host = instance_props.get('host')
        viable_hosts = self._get_results(filter_properties.get('context'),
                                         instance_props.get('uuid'),
                                         requested_networks,
                                         source_host)
        xform_viable_hosts = []
        hosts = []

        # Transform the dictionary of dictionaries into a list of hosts
        for host_data in viable_hosts.values():
            xform_viable_hosts.append(host_data.get('host'))

        # Build a list of hosts that satisfy the network requirements
        for host in all_host_attrs.keys():
            if host in xform_viable_hosts:
                hosts.append(host)
            else:
                LOG.debug(("Filtered host (%s) for instance (%s) because of "
                           "network deficiencies.") %
                          (host, instance_props.get('uuid')))

        self._include_hosts(hosts)

    def _get_results(self, context, instance_uuid, requested_networks,
                     source_host):
        ret = {}
        res = self._powervm_api.get_viable_hosts(
            context, instance_uuid, requested_networks=requested_networks,
            source_host=source_host)
        host_to_hyper_map = self._get_hosts_to_hypervisors_map(context)

        # Convert the host names to hypervisor hostnames
        for host in res.keys():
            hyper_hostname = host_to_hyper_map[host]
            ret[hyper_hostname] = {'host': hyper_hostname}

        return ret
