# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import six

from oslo_log import log as logging

from nova.ibm.ego import ego_constants
from nova.scheduler.ibm.ego import ego_resreq_utils as ego_utils

from powervc_nova.common import shared_processor_pools_utils as spp_utils
from powervc_nova.objects.compute import dom as compute_dom
from powervc_nova.scheduler.filters import PowerVCBaseFilter
from powervc_nova.virt.ibmpowervm.common import constants as pvmc
from powervc_nova.virt.ibmpowervm.common import proc_utils
from powervc_nova import logcall
from powervc_nova import _

LOG = logging.getLogger(__name__)


class PowerVMSharedProcPoolFilter(PowerVCBaseFilter):
    """
    This EGO scheduler filter ensures that if an image or virtual machine
    requires a particular shared processor pool (identified by its name),
    that only hosts with (a) an exact match on pool name; and (b) that the
    respective pool has enough capacity to house the virtual machine.
    """

    run_filter_once_per_request = True

    def __init__(self):
        super(PowerVMSharedProcPoolFilter, self).__init__()
        self._set_filter_scope(
            [ego_constants.FILTER_WORKING_SCOPE_CREATE,
             ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_COLD_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_EVACUATE,
             ego_constants.FILTER_WORKING_SCOPE_UNSHELVE])
        self._set_error_message(_("No suitable host could be found for the "
                                  "virtual machine that has a matching shared "
                                  "processor pool name and has sufficient "
                                  "processor capacity remaining; please check "
                                  "your shared processor pool names and their "
                                  "respective capacities and retry the "
                                  "operation."))

    @logcall
    def parse(self, filter_properties, all_host_attrs):
        if not self._allow_oper_for_powervm(filter_properties, all_host_attrs):
            LOG.debug("Ignoring %s..." % self.__class__.__name__)
            return

        context = filter_properties.get('context')
        request_spec = filter_properties.get('request_spec', {})
        instance_props = request_spec.get('instance_properties', {})
        inst_uuid = instance_props.get('uuid')
        operation_type = request_spec.get('operation_type')
        is_deploy = (operation_type ==
                     ego_constants.FILTER_WORKING_SCOPE_CREATE)
        proc_pool_name = None
        vcpu_demand = None

        if is_deploy:
            # For deploy operations, we need to grab the processor pool name
            # from the flavor because that's all we have to go upon.
            instance_type = filter_properties.get('instance_type') or {}
            extra_spec = instance_type.get('extra_specs') or {}
            proc_pool_name = extra_spec.get(
                'powervm:shared_proc_pool_name') or pvmc.DEFAULT_PROC_POOL
            # Determine the effective vcpu demand for the deploy; try to grab
            # the flavor extra_spec data and fall back to the flavor proper
            # if powervm:proc_units is unspecified.
            flavor_proc_units = extra_spec.get('powervm:proc_units')
            if flavor_proc_units is not None:
                vcpu_demand = float(flavor_proc_units)
            else:
                vcpu_demand = proc_utils.get_num_proc_units(
                    instance_type.get('vcpus') or 0)
        else:
            # For any non-deploy operation, we need to grab the processor pool
            # name from the power_specs because of potential out-of-band
            # changes.
            inst_fact = compute_dom.ManagedInstanceFactory.get_factory()
            inst = inst_fact.find_instance_by_uuid(context, inst_uuid) or {}
            power_specs = inst.get('power_specs') or {}
            proc_pool_name = (power_specs.get('shared_proc_pool_name')
                              or pvmc.DEFAULT_PROC_POOL)
            vcpu_demand = ego_utils.get_instance_vcpus(
                inst, ego_constants.POWERVM_HYPERVISOR_TYPE)

        if proc_pool_name == pvmc.DEFAULT_PROC_POOL:
            # All hosts support the default pool; don't bother filtering.
            LOG.debug("Because the default shared processor pool is assigned "
                      "to the instance (%s), no filtering is necessary."
                      % inst_uuid)
            return

        include_hosts = []

        # Iterate over all the hosts and check whether they have any processor
        # pools with a matching name and sufficient space to contain the VM.
        for hostname, host_data in six.iteritems(all_host_attrs):
            spp_data = spp_utils.get_shared_proc_pool(context, hostname,
                                                      proc_pool_name)
            if not spp_data:
                # There wasn't any matching pool name found; exclude this host.
                continue
            avail_capacity = float(spp_data.get('available_proc_units') or -1)
            LOG.debug("Determined that shared processor pool "
                      "'%(host)s:%(pool)s' has %(avail)s processor units "
                      "available and the demand by instance '%(inst)s' is "
                      "%(demand)s processor units."
                      % dict(host=hostname, pool=proc_pool_name,
                             avail=avail_capacity, inst=inst_uuid,
                             demand=vcpu_demand))
            if avail_capacity >= vcpu_demand:
                # The pool has enough capacity; include this host in the list.
                include_hosts.append(hostname)

        self._include_hosts(include_hosts)
