#
#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""
IBM PowerVC Storage utilities for retrieving and parsing HMC (K2)
response data for topology resources.
"""

from oslo_config.cfg import CONF
from oslo_log import log as logging
import six
import time

from powervc_nova import _
from powervc_nova.storage import topo_common
import powervc_nova.virt.ibmpowervm.hmc.host as dvr_host
from powervc_nova.virt.ibmpowervm.hmc.k2oprguard import K2OperGuard
from powervc_nova.virt.ibmpowervm.vif.hmc import K2_MNG_SYS_READ_SEC
from powervc_oslo.k2.k2operator import K2Error
from powervc_nova.storage import exception as stgex


LOG = logging.getLogger(__name__)
CONF.import_opt('ibmpowervm_register_ssps', 'powervc_nova.virt.ibmpowervm.hmc')
CONF.import_opt('ibmpowervm_ssp_hmc_version_check',
                'powervc_nova.virt.ibmpowervm.hmc')

VFCM_PORTELEMT_LEN = 5
XAG_DEFAULT = [None]
XAG_VIOS = ['ViosStorage']
XAG_VIOS_FC_STORAGE = ['ViosFCMapping', 'ViosStorage']


class StorageTopologyHmc(topo_common.StorageTopologyCommon):

    """ HMC-managed compute node specific collector """

    def __init__(self, host_name=None, context=None):
        super(StorageTopologyHmc, self).__init__(host_name, context)
        self.host_type = 'hmc'
        self.clusters_supported = None
        LOG.debug("Initialized HMC topology collector: %s" % host_name)

    def get_live_topology(self, skip_cluster=False, max_cache_age=-1):
        """
        Build a host topology dictionary using data from K2.
        If host_name is None, then build topology for all hosts. In normal
        operations, this API is not called for all hosts, but the flow
        allows it because of the original impl and for future expansion.
        """
        # Initialize data vars
        topo_factory = self.dom.HostStorageTopologyFactory.get_factory()
        static_topo = topo_factory.get_static_topology()
        topology = {'hosts': {}, 'skip-update-clusters': skip_cluster,
                    'vios_by_id': {}, 'vios_by_uuid': {},
                    'vios_num_trusted': 0,
                    'seq': static_topo.bump_topology_sequence(),
                    'static_topo': static_topo,
                    'nova_host': dvr_host.compute_host,
                    'max_cache_age': max_cache_age}

        if topology['nova_host']:
            # We can assume we are being called from within the compute
            # driver process. In this case, the compute host decides it's own
            # HMC and whether it can fall back to a different one.
            # Grab the IP for logging purposes.
            hmc_obj = dvr_host.compute_host.select_active_hmc()
            hmc_ip = hmc_obj.ipaddr if hmc_obj else 'unknown'
        else:
            msg = _("Improper call for method. The dvr_host.compute_host is "
                    "not set.")
            LOG.error(msg)
            return topology

        # TODO: cleanup any remaining obsolete code for API process caller.
        LOG.debug("HMC: access_ip=%s, passed host_name=%s"
                  % (hmc_ip, self.host_name))
        try:
            # Get the K2 Operator client.
            oper_guard = K2OperGuard(host=topology['nova_host'])

            # Use the operator guard as the K2 operator for all downstream
            # activity. This will maintain a single transaction for compute.
            with oper_guard as oper:
                if oper is None:
                    # Log the error and continue to next HMC if possible.
                    m = (_("Could not get an HMC K2 Operator instance "
                           "for managed host '%s'.") % self.host_name)
                    LOG.error(m)
                    raise stgex.IBMPowerVCStorageError(error=m)

                ##############################################################
                # Per topology collection run (per host), we have paired
                # the K2 calls down to two at most:
                # 1. Request a VIOS Feed for the host with a single extended
                #    group of [ViosStorage].
                # 2. If needed, request a Cluster feed, to get VIOS
                #    membership information per defined cluster.
                #
                # Based on the K2 behavior seen during PVC 1.2.1 development,
                # the order of the above calls has been switched to be the
                # order listed above. The reason is that the Cluster call is
                # generally more expensive and it is a wasted call if all
                # of the VIOSes are not in a good state (including RMC). If
                # the set of VIOSes are not healthy, then the information
                # returned from K2 will be stale with respect to information
                # the HMC receives by talking to the VIOS operating system.
                # And this includes cluster membership. Even if everything
                # is configured properly, we have seen the set of host VIOSes
                # spontaneously go into an rmc_state of 'busy', sometimes
                # remaining in that state, and sometimes recovering quickly.
                # In either case, it is not worthwhile for the periodic task
                # to go after cluster membership information during this cycle.
                # We will leave the membership in the nova DB as is to
                # avoid thrashing, even though from a technical standpoint,
                # the VIOSes may not be members if they don't see one another
                # because of a slow response due to connectivity or VIO
                # Server busy-ness.
                ##############################################################
                # 1. Get the topology for host, which is essentially the
                #    VIOSs topology, and add it to the topology dictionary.
                ##############################################################
                host_dict = {'name': self.host_name,
                             'uuid': topology['nova_host'].uuid}
                topology['hosts'][self.host_name] = host_dict
                self.collect_host_vios_feeds(oper, topology)

                ##############################################################
                # 2. Get VIOS cluster membership and add to topology.
                #
                #    The method will no-op for a variety of cluster
                #    skip conditions. Note that the skip conditions of
                #    the called method here handle most cases for the
                #    defect 19372. And the remaining timing window case
                #    between the VIOS and Cluster feed calls is handled
                #    by the computed 'trust_seq' property added to VIOS
                #    topology entries.
                ##############################################################
                self.map_vios_to_cluster(oper, topology)
        except Exception as ex:
            LOG.exception(ex)
            msg = _("There was a problem getting host or Virtual I/O "
                    "Server information from HMC '%s'." % hmc_ip)
            LOG.warn(msg)
            topology['error'] = [msg, six.text_type(ex)]

        LOG.debug("Acquired HOST STORAGE TOPOLOGY from K2 for VIOS IDs %s. "
                  "is_error_set=%s" % (topology['vios_by_id'].keys(),
                                       ('error' in topology)))
        return topology

    def map_vios_to_cluster(self, oper, topology):
        """
        Given a K2 operator connection to an HMC, retrieve all the VIOS
        Clusters and loop over the member VIOSes, add the membership info
        to the passed topology data structure. This mapping
        will subsequently be used in the DB reconciliation to update the
        cluster membership field on the VIOS DB resource if needed.
        """
        if topology['skip-update-clusters']:
            LOG.debug("Skip getting Cluster feed info. Could be during "
                      "deploy.")
            return
        if topology['vios_num_trusted'] == 0:
            LOG.debug("No VIOSes are reporting a trusted state for collection "
                      "sequence %d. Skip getting Cluster feed." %
                      topology['seq'])
            topology['skip-update-clusters'] = True
            return
        # See if Registering Clusters is disabled in this Environment, if it
        # is then it isn't worth the Performance hit to query Clusters from K2
        if not CONF.ibmpowervm_register_ssps or \
           self.cluster_registration_verification(oper, topology) is False:
            LOG.debug("Cluster registration is disabled, so no HMC cluster "
                      "retrieval.")
            topology['skip-update-clusters'] = True
            return

        LOG.debug("Retrieving all clusters and VIOS members for current hmc.")
        # Get the root cluster object from the K2 API.
        # NOTE: In HMC 810, the Cluster info does not have events and so the
        #       caching is not event based. The K2 operator will just employ
        #       a small fixed timeout for the Cluster cache.
        try:
            k2resp = k2_read(oper, 'Cluster', xag=XAG_DEFAULT,
                             age=topology['max_cache_age'])
        except Exception as ex:
            # Don't pass in 'topology' to avoid the 'error' key being set
            # in the dictionary. Instead, we set the special flag to skip
            # updating the cluster info below. We don't want to skip
            # everything.
            self.log_k2ex_msg(ex, _("Cannot retrieve any VIOS clusters."),
                              None)
            topology['skip-update-clusters'] = True
            return

        if len(k2resp.feed.entries) == 0:
            LOG.debug("No cluster entries returned from HMC.")
        cluster_members = {}
        static_topo = topology['static_topo']

        # Traverse the hierarchy structure to find the VIOS objects.
        # The structure is:
        # k2resp --> entry --> Node --> Node --> VirtualIOServer
        for i in range(len(k2resp.feed.entries)):
            cluster = k2resp.feed.entries[i]
            cluster_id = cluster.element.findtext('./ClusterID')
            if cluster_id is None or len(cluster_id) == 0:
                LOG.warn(_("HMC Problem: Cluster entry does not have a "
                           "ClusterID Skipping. Entry is '%s'.") %
                         cluster.gettext())
                continue
            cluster_members[cluster_id] = set()
            cluster_name = cluster.element.findtext('./ClusterName')
            LOG.debug("Cluster entry #%d: id=%s, display_name=%s." %
                      (i, cluster_id, cluster_name))
            backend_id = cluster.properties['id']
            viosElements = cluster.element.findall(
                './Node/Node/VirtualIOServer')
            if not viosElements:
                # This may not be possible.
                LOG.info(_("The cluster '%s' has no member VIOSs reporting.") %
                         cluster_name)
            for j in range(len(viosElements)):
                vios_elmt = viosElements[j]
                href = vios_elmt.get('href')
                if href is None:
                    LOG.warn(_("HMC Problem: VirtualIOServer element has no "
                               "'href': %s." % str(vios_elmt)))
                    continue

                LOG.debug("Found href for VirtualIOServer member #%d: %s."
                          % (j, str(href)))
                # Split the URL into elements.
                href_components = href.split('/')
                # Ensure the list has at least one element.
                if len(href_components) > 0:
                    # We only care about the last element in
                    # the list, which should be the uuid.
                    vios_uuid = href_components[-1]
                    LOG.debug("Found uuid: '%s'." % vios_uuid)
                    vios_cluster = {'cluster_id': cluster_id,
                                    'cluster_name': cluster_name,
                                    'backend_id': backend_id}
                    if vios_uuid in topology['vios_by_uuid']:
                        vios = topology['vios_by_uuid'][vios_uuid]
                        vios['vios_cluster'] = vios_cluster
                        LOG.debug("VIOS entry for SSP cluster: %s " % vios)
                        static_topo.set_cluster_for_vios(vios['id'],
                                                         cluster_id,
                                                         cluster_name)
                        if not vios['trusted']:
                            # For debug purposes only
                            LOG.debug("VIOS is reporting cluster membership "
                                      "even though its state information in a "
                                      "separate K2 response has it marked not "
                                      "trusted: %s." % vios)
                    else:
                        LOG.debug("Clustered VIOS uuid %s not under host for"
                                  "this "
                                  "topology collection. Cluster = %s" %
                                  (vios_uuid, vios_cluster))
            # end for each vios element

        # Log a missing cluster.
        for cluster in static_topo.cluster_keyed.values():
            if cluster['set_cluster_seq'] != static_topo.sequence_num:
                LOG.info(_("The cluster %(cluster_name)s, inventoried during "
                           "prior topology collection sequence %(old)d, did "
                           "not report members this time [%(new)d].") %
                         dict(cluster_name=cluster['display_name'],
                              old=cluster['set_cluster_seq'],
                              new=static_topo.sequence_num))
        LOG.debug("Tracked cluster membership: %s." %
                  static_topo.cluster_keyed)

    def collect_host_vios_feeds(self, oper, topology):
        """
        We get the feed for all its VIOSes, then loop over the
        feed entries. This flow is for when we know the uuid of the host.
        The topology dictionary passed is updated with the collected and parsed
        information.
        """
        host_ref = topology['nova_host'].mtm_serial
        LOG.debug('Obtaining VIOS feed for Managed System: %s' % host_ref)
        k2resp = None
        try:
            k2resp = k2_read(oper, 'ManagedSystem',
                             rootId=topology['nova_host'].uuid,
                             childType='VirtualIOServer',
                             timeout=K2_MNG_SYS_READ_SEC,
                             xag=XAG_VIOS_FC_STORAGE,
                             age=topology['max_cache_age'])
        except Exception as ex:
            self.log_k2ex_msg(ex, _("Cannot get VIOSs for host %s.") %
                              host_ref, topology)
            return topology

        if k2resp is None:
            # should not get here
            msg = _("Managed System K2 response was none for host: "
                    "%s.") % host_ref
            LOG.error(msg)
            topology['error'] = [msg]
            return topology

        if not k2resp.feed or not k2resp.feed.entries:
            LOG.info(_("No VIOS entries for host: %s") % host_ref)
            return topology

        # Loop over all vioses found
        num = 0
        for vios_entry in k2resp.feed.entries:
            num += 1
            # call the helper to get a dict for this vios entry
            vios_info = self.get_vios_info(vios_entry)
            vios_info['id'] = self.get_vios_id(host_ref, vios_info)
            # Add the vios entry to both indexes
            # TODO: One of the following can be removed now that the VIOS
            #       UUID is persisted as 'id' in the nova DB. Needs careful
            #       code combing however.
            topology['vios_by_id'][vios_info['id']] = vios_info
            topology['vios_by_uuid'][vios_info['uuid']] = vios_info
            if 'state' in vios_info and vios_info['state'] == "running" and\
                    'rmc_state' in vios_info and\
                    vios_info['rmc_state'] == "active":
                topology['vios_num_trusted'] += 1
                vios_info['trusted'] = True
            else:
                # Some VIOS information like cluster membership and FC ports
                # should not be trusted if state info is not all OK.
                vios_info['trusted'] = False
            # Note the trust state marking below only makes a difference if the
            # vios is being tracked as a prior cluster member.
            if not topology['skip-update-clusters']:
                # The static_topo structure is not syncronized, so we skip
                # changing it if the 'skip' value is present, which deploy
                # processing will always set. This avoids periodic task
                # collections colliding with deploy collections.
                topology['static_topo'].mark_vios_trust_state(
                    vios_info['id'], vios_info['trusted'])
            LOG.debug("Added VIOS dict entry to topology: %s" % vios_info)
        LOG.debug("Returning %d trusted VIOS entries out of %d collected for "
                  "compute host %s." % (topology['vios_num_trusted'], num,
                                        host_ref))
        return topology

    def get_vios_info(self, vios_entry):
        """
        Get the specific data for the VIOS from the K2 VIOS entry.
        """
        # Initialize the empty dictionary.
        vios_info = {}
        if not vios_entry:
            LOG.warn(_("No Virtual I/O Servers returned in HMC response."))
            return vios_info

        vios_info['k2element'] = vios_entry  # track for later logging
        # Get the Partition Name Element
        partitionNameElement = \
            vios_entry.element.find('./PartitionName')
        # If the Partition Name element exists
        if partitionNameElement is not None:
            # Save the partition name into the vios info object.
            vios_info['name'] = partitionNameElement.gettext()
        else:
            vios_info['name'] = None

        vios_info['uuid'] = uuid = vios_entry.properties['id']
        LOG.debug("Processing VIOS partition '%s' with uuid: %s" %
                  (vios_info['name'], uuid))
        # Get the Partition State Element
        partitionStateElement = \
            vios_entry.element.find('./PartitionState')

        if partitionStateElement is not None:
            # Save the partition State into the vios info object.
            vios_info['state'] = partitionStateElement.gettext()

        # Get the RMC State Element
        RMCStateElement = \
            vios_entry.element.find('./ResourceMonitoringControlState')

        if RMCStateElement is not None:
            # Save the RMC State into the vios info object.
            vios_info['rmc_state'] = RMCStateElement.gettext()

        # Get the Partition ID Element.
        partitionIDElement = vios_entry.element.find('./PartitionID')
        # If the partition ID Element exists
        if partitionIDElement is not None:
            # Save the partition id to the vios info object.
            vios_info['lpar_id'] = partitionIDElement.gettext()

            # For each VIOS, also call the helper method to extract
            # the FC Port info from call already done.
            LOG.debug("Getting FC ports for vios '%s'." % uuid)
            vios_info['fcports'] = self.parse_fcports(vios_entry, vios_info)
        else:
            LOG.warn(_("HMC Problem: No PartitionID element for VIOS '%s'.") %
                     uuid)

        return vios_info

    def parse_fcports(self, vios, vios_info):
        """
        Get a list of the FC Ports related to a specific VIOS.
        'vios' passed in is the K2 response VIOS entry of a previous call.
        We traverse the hierarchy of objects to find the desired
        information.
        """
        LOG.debug('Extracting FC Ports from VIOS K2 entry.')
        # Create an empty dictionary for the FC port mapping (UDID-->port info)
        fcports = {}
        # Traverse the hierarchy to find the FC port information.
        # First, check under PhysicalFibreChannelAdapter
        adapterPath = str('./PartitionIOConfiguration/ProfileIOSlots/'
                          'ProfileIOSlot/AssociatedIOSlot/RelatedIOAdapter/'
                          'PhysicalFibreChannelAdapter')
        fcAdapters = vios.element.findall(adapterPath)
        if fcAdapters:
            LOG.debug('Found element(s): %s.' % adapterPath)
            for fcAdapter in fcAdapters:
                LOG.debug("Found adapter element: %s " %
                          fcAdapter.toxmlstring())
                portElements = fcAdapter.findall(
                    './PhysicalFibreChannelPorts/PhysicalFibreChannelPort')
                if portElements:
                    for portElement in portElements:
                        LOG.debug('Found PhysicalFibreChannelPort element.')
                        self._parse_single_fc_port(portElement, fcports,
                                                   fcAdapter)
                else:
                    LOG.debug("PhysicalFibreChannelPort elements not found "
                              "for vios '%s'." % vios_info['name'])
        else:
            LOG.debug("PhysicalFibreChannelAdapter elements not found for "
                      "vios '%s'." % vios_info['name'])

            # Check for the parent RelatedIOAdapters as we have seen cases
            # where these exist for the FC adapters, but the FC adapter
            # children elements are not returned. We can log this case.
            ioAdaptPath = str('./PartitionIOConfiguration/ProfileIOSlots/'
                              'ProfileIOSlot/AssociatedIOSlot/RelatedIOAdapt'
                              'er/IOAdapter')
            ioAdapters = vios.element.findall(ioAdaptPath)
            io_num = 0
            if ioAdapters:
                vios_info['IOAdapters'] = []
                for ioAdapter in ioAdapters:
                    io_num += 1
                    if ioAdapter.find('./DeviceName') is not None:
                        adapterID = ioAdapter.find('./DeviceName').gettext()
                        desc = ioAdapter.find('./Description').gettext()
                        vios_info['IOAdapters'].append(adapterID)
                        if 'Fibre' in desc:
                            LOG.debug("#%d: Found FC IOAdapter element: %s" %
                                      (io_num, ioAdapter.toxmlstring()))

        # For each port find the number of vfc mappings
        virtualFibreChannelMappingsElement = vios.element.find(
            './VirtualFibreChannelMappings')

        if virtualFibreChannelMappingsElement is not None:
            portmap = {}
            LOG.debug('Found VirtualFibreChannelMappings element.')
            virtualFibreChannelMappingElements = \
                virtualFibreChannelMappingsElement.findall(
                    './VirtualFibreChannelMapping')
            for virtualFibreChannelMappingElement in \
                    virtualFibreChannelMappingElements:
                LOG.debug('Found VirtualFibreChannelMapping element.')

                portElement = virtualFibreChannelMappingElement.find('./Port')
                # The VirtualFibreChannelMapping contains FC ports and the
                # ports in the server/client adapter virtual mapping. Only the
                # matches that have Metadata and Port are truly portElement.
                if (portElement is not None and
                    len(virtualFibreChannelMappingElement) ==
                        VFCM_PORTELEMT_LEN):
                    LOG.debug('Found Port element: %s.' %
                              portElement.toxmlstring())
                    portmap = self._update_mappings_fc_port(portElement,
                                                            portmap)

            for portId in portmap.keys():
                LOG.debug('Found mappings for port: %s.' % portId)
                if portId not in fcports:
                    LOG.warn(_("Did not collect information for FC port "
                               "%(portid)s, so cannot ascertain client "
                               "adapter mapping.") % {'portid': portId})
                    continue
                fcport = fcports[portId]
                fcport['mapped_vfcs'] = portmap[portId]
                if 'total_vports' in fcport and (
                        fcport['mapped_vfcs'] >= (fcport['total_vports'] -
                                                  topo_common.PORT_BUFFER)):
                    if self.use_mapped_vfc_scheduling:
                        LOG.debug("For port %s, using mapped vFC scheduling, "
                                  "so set npiv_status to Full." % fcport)
                        fcport['npiv_status'] = topo_common.PORT_STATUS_FULL
                    else:
                        LOG.debug("For port %s, mapped_vfcs is at or above "
                                  "total_vports, but not using vFC scheduling."
                                  % (fcport))
        # Return the mapping of all the ports for the VIOS.
        LOG.debug("Finished parsing FC ports on VIOS: %s" % vios_info['name'])
        return fcports

    def _update_mappings_fc_port(self, portElement, portmap):
        """
        Update the vfc mapping count for each port
        """
        portIdElement = portElement.find('./UniqueDeviceID')
        if portIdElement is not None:
            portId = portIdElement.gettext()
            if portId not in portmap:
                portmap = {portId: 1}
            else:
                map_count = portmap.get(portId, 0)
                portmap[portId] = map_count + 1

        return portmap

    def _parse_single_fc_port(self, portElement, fcports, adapterElement=None):
        """
        If portElement is a valid, NPIV-capable FCPort, add its information
        to fcports as a new dictionary port entry.
        NOTE: This function modifies 'fcports' by adding an element for a
        supported FC port, rather than returning a dictionary to add later.
        Could refactor in the future to be more openstack-esque.
        """
        uniqueDeviceIDElement = portElement.find('./UniqueDeviceID')
        portNameElement = portElement.find('./PortName')
        # AdapterUUID have been removed from schema. Keep it
        # as a placeholder.
        # adapter_uuidElement = portElement.find('./AdapterUUID')
        wwpnElement = portElement.find('./WWPN')
        total_portsElement = portElement.find('./TotalPorts')
        avail_portsElement = portElement.find('./AvailablePorts')
        is_NPIV_adapter = True
        # use exception to find invalid attribute.
        try:
            if adapterElement is not None:
                # adapter id is not needed for PowerVC at this time,
                # it is part of the DB DTO for the fcport object, and if the
                # parent adapter element is passed, we use the DeviceName from
                # that.
                adapterID = adapterElement.find('./DeviceName').gettext()
                adapter_desc = adapterElement.find('./Description').gettext()
                is_NPIV_adapter = self.is_npiv_adapter_description(
                    adapter_desc)
            else:
                adapterID = '[unknown]'
            # The portElement contains non-NPIV ports which
            # has no TotalPorts and AvailablePorts attributes.
            # Filter out non-NPIV ports other wise VFC map
            # will fail.
            fcport = {'id': uniqueDeviceIDElement.gettext(),
                      'name': portNameElement.gettext(),
                      'adapter_id': adapterID,
                      'wwpn': wwpnElement.gettext(),
                      'status': topo_common.PORT_STATUS_UNKNOWN,
                      'npiv_status': topo_common.PORT_STATUS_UNKNOWN
                      }
            assert len(fcport['id']) > 0
            if fcport['wwpn']:
                # Note: if empty wwpn, then status will be unknown
                # WARNING! We need to be very careful in checking k2 elements.
                #   The following code line is not equivalent to:
                #        if total_portsElement and avail_portsElement
                #   This is because these K2 elements have a __len()__ method
                #   that evaluates to 0, meaning they have no nested entries.
                #   But the objects still exists!
                if total_portsElement is not None and avail_portsElement\
                        is not None:
                    fcport['total_vports'] = int(total_portsElement.gettext())
                    fcport['available_vports'] = int(
                        avail_portsElement.gettext())
                    fcport['status'] = topo_common.PORT_STATUS_OK
                    # TODO: Instead of constant PORT_BUFFER in check below,
                    #       could a config option value to give a buffer,
                    #       reducing the chance of port over-commitment.
                    if fcport['available_vports'] > topo_common.PORT_BUFFER:
                        # put number of VFCs left on the port for OK status as
                        # this info can be split out and used for API info.
                        fcport['npiv_status'] = topo_common.PORT_STATUS_OK
                    else:
                        fcport['npiv_status'] = topo_common.PORT_STATUS_FULL
                else:
                    LOG.debug("Not a currently supported FC Port since "
                              "TotalPorts and AvailablePorts are missing "
                              "from the element. is_NPIV_adapter=%s " %
                              is_NPIV_adapter)
                    if is_NPIV_adapter:
                        fcport['npiv_status'] = topo_common.PORT_STATUS_OFFLINE
                        fcport['status'] = topo_common.PORT_STATUS_OFFLINE
                    else:
                        fcport['npiv_status'] =\
                            topo_common.PORT_NPIV_STATUS_UNS_ADAPTER
                        fcport['status'] = topo_common.PORT_STATUS_OK

            LOG.debug("Adding FC port: %(fcport)s" % locals())
            fcports[fcport['id']] = fcport
        except Exception as ex:
            LOG.exception(ex)
            # One of the required port elements are missing.
            LOG.warn(_("HMC Problem: Found non-NPIV capable port due to "
                       "unexpected format. Skip portElement: %(portelem)s") %
                     {'portelem': portElement.toxmlstring()})

    def log_k2ex_msg(self, ex, prefix, topology):
        """ LOG K2 exception and extracted message. Return NLS message """
        LOG.exception(ex)
        detail = {}
        k2msg = _("None")
        if isinstance(ex, K2Error) and ex.k2response:
            detail['Request_headers'] = ex.k2response.reqheaders
            detail['Response_headers'] = ex.k2response.headers
            detail['Response_body'] = ex.k2response.body
            detail['Response_status'] = ex.k2response.status
            if hasattr(ex.k2response, 'k2err'):
                m = ex.k2response.k2err.find('./Message')
                if m is not None:
                    k2msg = m.text
        msg = _("%(prefix)s ***K2 Operator Error***: %(ex_msg)s  [K2 Error "
                "body Message: %(k2msg)s]") %\
            dict(prefix=prefix, ex_msg=ex, k2msg=k2msg)
        LOG.error(msg)
        if detail:
            LOG.error(_("Error details: %s") % detail)
        if topology is not None:
            if 'error' in topology:
                topology['error'].append(msg)
            else:
                topology['error'] = [msg]
        return msg

    def cluster_registration_verification(self, oper, topology):
        """Method to verify the HMC level is good for cluster reg. support"""
        if not CONF.ibmpowervm_ssp_hmc_version_check:
            return True
        if topology['static_topo'].clusters_supported is not None:
            return topology['static_topo'].clusters_supported
        else:
            try:
                # Get the ManagementConsole information from the K2 API.
                mc_info = oper.read('ManagementConsole')
                # Parse the feed for the HMC version info
                ver_info = mc_info.feed.entries[0].element.find('VersionInfo')
                # Get the maintenance number from the version info
                maintenance = ver_info.findtext('Maintenance')
                # Get the minor number from the version info
                minor = ver_info.findtext('Minor')
                # Get the version number from the version info
                version = ver_info.findtext('Version')
                # Combine to get the HMC release version
                release = version + maintenance + minor
                LOG.info(_("Management Console release is '%s'.") % release)
                is_satisfied = (int(release) >= 810)
                topology['static_topo'].clusters_supported = is_satisfied
                return is_satisfied
            except Exception as exc:
                LOG.exception(exc)
                LOG.error("There was an error getting the HMC release version")
                return True

    def get_vios_id(self, host_name, vios_dict):
        """ Generate the VIOS ID for REST API consumption based on K2 data """
        return vios_dict['uuid']


def k2_read(operator, rootType, rootId=None, childType=None,
            suffixType=None, suffixParm=None, timeout=-1, age=-1, xag=[]):
    """
    We provide our own read abstraction interface on top of K2's here.
    This can be enhanced as needed, but it is mainly for our own custom
    logging.
    NOTES: This method will not be called if cached K2 feeds are being used
           from the compute driver.
           If this is called from the compute periodic task, then the
           operator should encapsulate a common session that is always
           re-used.
           The operator may do its own internal caching so an over-the-wire
           request may not be performed.
    """
    # Log the call
    LOG.debug("\nK2_OPER_REQ - rootType=%(rootType)s, rootId=%(rootId)s, "
              "childType=%(childType)s, suffixType=%(suffixType)s, "
              "suffixParm=%(suffixParm)s, timeout=%(timeout)d, age=%(age)d, "
              "xag=%(xag)s, ...operator_obj=%(operator)s" % locals())
    # Make the K2 call.
    ticks = time.time()
    resp = operator.read(rootType, rootId=rootId, childType=childType,
                         suffixType=suffixType, suffixParm=suffixParm,
                         timeout=timeout, age=age, xag=xag)
    # Set up timing object
    global logTiming
    try:
        logTiming
    except NameError:
        logTiming = restTiming()  # init if it does not exist yet
    # Log that the response was received. Note there is a special logger
    # that can be enabled for the operator to log the response itself.
    logTiming.log_return(time.time() - ticks, operator, rootType, rootId,
                         childType, suffixType)
    # return
    return resp


class restTiming():

    """ Simple class to track and log the timing of a K2 rest call """

    def __init__(self):
        self.hashdict = {}

    def log_return(self, seconds, operator, rootType,
                   rootId=None, childType=None, suffixType=None):
        # Create a hash key from a few pieces of info so that we can
        # compare query timings between 'types' of calls. Note that
        # operator.session.host will contain the HMC IP/hostname.
        pieces = [operator._k2oper.session.host, rootType, rootId, childType,
                  suffixType]
        key = ''.join(map(lambda x: str(x) + "_" if x is not None else "*",
                          pieces))
        if key not in self.hashdict:
            self.hashdict[key] = {'samples': 1, 'sum': seconds}
        else:
            self.hashdict[key]['samples'] += 1
            self.hashdict[key]['sum'] += seconds
        avg = self.hashdict[key]['sum'] / self.hashdict[key]['samples']
        LOG.debug("\nK2_OPER_Timing=%.3f   Call_hash=%s, average=%.3f, "
                  "avg_delta=%+.3f, entry=%s.\n" %
                  (seconds, key, avg, seconds - avg, self.hashdict[key]))
