# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
import six
import time

from oslo_config import cfg
from oslo_log import log as logging
from nova.compute import vm_states
from nova import rpc
from nova import context
from nova import conductor
from powervc_nova import _

LOG = logging.getLogger(__name__)

EVENT_TYPE_ERROR = 'ERROR'
EVENT_TYPE_WARN = 'WARN'

BOOKKEEPER = dict(cleanup_timer=0)


def set_instance_error_state_and_notify(instance):
    """
    Set an instance to ERROR state and send out a notification
    """
    # set instance to error state. Instance
    # could be a dictionary when this method is called during
    # virtual machine delete process.
    instance.vm_state = vm_states.ERROR
    instance.task_state = None
    instance.save()
    instance_name = instance['name']
    host_name = instance['host']
    instance_id = instance['uuid']
    LOG.warn(_('Unable to find virtual machine %(inst_name)s '
               'on host %(host)s. Set state to ERROR')
             % {'inst_name': instance_name,
                'host': host_name})
    # Send event notification
    note = {'event_type': 'compute.instance.log',
            'msg': _('Unable to find virtual machine {instance_name} on '
                     'host {host_name} with id {instance_id}. An operation '
                     'might have been performed on the virtual machine outside'
                     ' of PowerVC or the deploy of the virtual machine failed.'
                     'The virtual machine is now set to Error state in the '
                     'database.'),
            'instance_name': instance_name,
            'host_name': host_name,
            'instance_id': instance_id}
    notifier = rpc.get_notifier(service='compute', host=host_name)
    notifier.warn(context.get_admin_context(), 'compute.instance.log',
                  note)


def instance_error_seen_multiple_time(instance, repeats):
    """
    Count the number of times InstanceNotFound exceptions have been
    raised for an instance. Return True if threshold reached.
    Threshold is reached if number of exceptions raised within
    a time interval is greater than or equals to a fixed value.
    """
    global BOOKKEEPER
    ret = False
    try:
        timing_window = cfg.CONF.sync_power_state_interval * (repeats + 1)
        instance_uuid = instance['uuid']
        cached_info = BOOKKEEPER.get(instance_uuid)
        current_time = time.time()
        if cached_info is not None:
            cached_time = cached_info['timestamp']
            if current_time - cached_time > timing_window:
                BOOKKEEPER[instance_uuid]['count'] = 1
                BOOKKEEPER[instance_uuid]['timestamp'] = current_time
            else:
                count = cached_info['count']
                if count >= repeats:
                    LOG.debug('Threshold reached for instance uuid %s' %
                              instance_uuid)
                    ret = True
                    del BOOKKEEPER[instance_uuid]
                else:
                    BOOKKEEPER[instance_uuid]['count'] = count + 1
        else:
            BOOKKEEPER[instance_uuid] = dict(count=1,
                                             timestamp=current_time)
        if current_time - BOOKKEEPER['cleanup_timer'] > 43200:
            # Clean up cache every 12 hours
            BOOKKEEPER['cleanup_timer'] = current_time
            _clean_up_cache()
    except Exception as e:
        LOG.info(_('Exception encountered while counting '
                   'occurances of instance not found exceptions %s')
                 % six.text_type(e))
    return ret


def instance_found_remove_from_cache(instance):
    """Removes the instance from the NotFound Bookeeper when Found"""
    global BOOKKEEPER
    BOOKKEEPER.pop(instance['uuid'], None)


def _clean_up_cache():
    """Clean up bookkeeper cache every so often"""
    global BOOKKEEPER
    list_of_keys = BOOKKEEPER.keys()
    expire_time = time.time() - 7200
    for key in list_of_keys:
        try:
            if key != 'cleanup_timer':
                cached_time = BOOKKEEPER[key]['timestamp']
                if cached_time < expire_time:
                    # Remove entries older than 2 hours
                    del BOOKKEEPER[key]
        except Exception:
            pass
