# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Utility class for updating the PowerSpec Data in the DB for IVM and HMC"""

import six
import hashlib

from nova import notifications as notifyutil
from nova import rpc
from nova import exception
from nova.compute import power_state
from nova.compute import task_states
from nova.compute import vm_states
from nova.compute import flavors
from nova.context import get_admin_context
from nova.objects import flavor
from oslo_config import cfg
from oslo_log import log as logging

from powervc_nova.objects.compute.dom import ManagedInstanceFactory,\
    ManagedInstance
from powervc_nova.virt.ibmpowervm.common.constants import DEFAULT_PROC_POOL
from powervc_nova import _

LOG = logging.getLogger(__name__)
POWER_SPECS_SYS_META_MAP = {
    'proc_units': 'vcpus', 'memory_mb': 'memory_mb'}

CONF = cfg.CONF


def update_instance_power_specs(context, virtapi, instance, power_specs):
    """Helper Method to update the Power-specific Attributes in the Database"""
    # stg defect 44740
    # make sure the instance passed in is ManagedInstance object
    managed_inst = instance
    if not isinstance(instance, ManagedInstance):
        inst_fact = ManagedInstanceFactory.get_factory()
        managed_inst = inst_fact.find_instance_by_uuid(
            context, instance.uuid)
    old_power_spec = managed_inst.get('power_specs')
    old_power_spec = old_power_spec if not None else {}
    LOG.debug('update_instance_power_specs: new=%s old=%s' %
              (power_specs, old_power_spec))
    # Figure out if there is any Pending Action for this Instance
    _handle_pending_action(context, managed_inst, power_specs)

    # defect 21441
    # move the type conversion up, so we can avoid unnecessary update
    # defect 17007
    # convert uncapped value from string to boolean
    uncapped = power_specs.get('uncapped')
    if uncapped is not None:
        if isinstance(uncapped, str):
            if uncapped.lower() == 'true':
                power_specs['uncapped'] = True
            else:
                power_specs['uncapped'] = False

    srr_cap = power_specs.get('srr_capability')
    if srr_cap is not None:
        if isinstance(srr_cap, str):
            if srr_cap.lower() == 'true':
                power_specs['srr_capability'] = True
            else:
                power_specs['srr_capability'] = False

    # convert dedicated_proc to string as in HMC
    ded_proc = power_specs.get('dedicated_proc')
    if ded_proc is not None:
        power_specs['dedicated_proc'] = six.text_type(ded_proc).lower()

    # check the instance system metadata as well
    db_inst_sys_metadata = managed_inst.get('system_metadata', {})
    sys_metadata_ch = _sys_metadata_updates(db_inst_sys_metadata,
                                            power_specs)

    # Determine if any of the PowerSpecs actually changed
    for key in power_specs.keys():
        if _compare_spec_value(power_specs.get(key), old_power_spec.get(key)):
            power_specs.pop(key, None)
        else:
            LOG.debug(
                six.text_type("spec update needed for instance %s: key %s"
                              " new: %s old: %s") %
                (managed_inst['display_name'],
                 key, power_specs.get(key),
                 old_power_spec.get(key)))

    # If something did really change, then update the PowerSpecs in the DB
    # We also skip the update if the Instance is currently resizing/changing
    if ((len(power_specs) > 0 or len(sys_metadata_ch) > 0)
            and not _is_instance_resizing(managed_inst)
            and not _is_instance_spawning(managed_inst)):
        updates = {}
        if len(power_specs) > 0:
            updates['power_specs'] = power_specs

        # Temporarily duplicate the info on the System Meta-data
        _duplicate_system_metadata(updates, sys_metadata_ch)
        oob_changes = _handle_out_of_band_updates(context, managed_inst,
                                                  updates)
        # Update the Instance in the DB through the Instance DOM Model
        managed_inst.update(updates)
        managed_inst.save()
        # Send a Notification to the Scheduler with the Instance Updates
        _notify_instance_update_spec(context, managed_inst, updates)
        # If there were Out-of-band Changes, then send a Notification
        if oob_changes:
            _notify_out_of_band_change(context, managed_inst)


def _notify_instance_update_spec(context, instance, updates):
    """Internal Helper Method to send notifications if allocations changed"""
    LOG.debug(six.text_type('notify instance update %s %s') %
              (instance['display_name'], updates))
    updates = updates.copy()
    # Parse out the Old and New Meta-Data and PowerSpecs to merge them together
    new_pwrspecs = updates.pop('power_specs', {})
    new_metadata = updates.pop('system_metadata', {})
    old_metadata = instance.get('system_metadata', {}).copy()
    old_pwrspecs = instance.get('power_specs', {}).copy()
    # Merge the Old and New MetaData and PowerSpecs together for the Event
    old_pwrspecs.update(new_pwrspecs)
    old_metadata.update(new_metadata)
    # Use the Utility to Build the base Instance, and add MetaData/PowerSpecs
    info = notifyutil.info_from_instance(context, instance, None, None)
    info.update(dict(system_metadata=old_metadata, power_specs=old_pwrspecs))
    info.update(updates)
    # Now we can send the Notification to the Scheduler that the Specs changed
    notifier = rpc.get_notifier(service='compute', host=instance['host'])
    notifier.info(context, 'compute.instance.update.spec', info)


def _handle_out_of_band_updates(context, instance, updates):
    """Helper method to Update and Notify when Out-of-Band Changes"""
    # updates only contains attributes with different values between
    # old and new power_specs
    db_vcpus = instance.get('vcpus')
    db_memmb = instance.get('memory_mb')
    db_procs = instance.get('power_specs', {}).get('proc_units')
    db_srrcp = instance.get('power_specs', {}).get('srr_capability')
    db_srrstate = instance.get('power_specs', {}).get('srr_state')
    db_ameef = instance.get('power_specs', {}).get('ame_expansion_factor')
    db_procpoolname = instance.get('power_specs', {}).get(
        'shared_proc_pool_name')
    db_procpoolid = instance.get('power_specs', {}).get(
        'shared_proc_pool_id')
    # db_procmode helps us determine whether proc pool info change
    # originate from deploy or not
    db_procmode = instance.get('power_specs', {}).get(
        'dedicated_proc')
    update_pwr_specs = updates.get('power_specs', {})
    drv_vcpus = update_pwr_specs.get('vcpus')
    drv_memmb = update_pwr_specs.get('memory_mb')
    drv_procs = update_pwr_specs.get('proc_units')
    drv_srrcp = update_pwr_specs.get('srr_capability')
    drv_srrstate = update_pwr_specs.get('srr_state')
    drv_ameef = instance.get('power_specs', {}).get('ame_expansion_factor')
    drv_procpoolname = update_pwr_specs.get('shared_proc_pool_name')
    drv_procpoolid = update_pwr_specs.get('shared_proc_pool_id')
    # If the Driver gave a different value than the DB, update the DB to match
    if db_vcpus and drv_vcpus and not _compare_spec_value(db_vcpus, drv_vcpus):
        updates['vcpus'] = drv_vcpus
    if db_memmb and drv_memmb and not _compare_spec_value(db_memmb, drv_memmb):
        updates['memory_mb'] = drv_memmb
    if db_srrcp and drv_srrcp and not _compare_spec_value(db_srrcp, drv_srrcp):
        updates['srr_capability'] = drv_srrcp
    if db_srrstate and drv_srrstate and not _compare_spec_value(db_srrstate,
                                                                drv_srrstate):
        updates['srr_state'] = drv_srrstate
    # for other attributes, if it's not set, we can determine it's originated
    # from deploy and not to send out notification.
    # this is not true for proc pool
    # it's possible that we don't have shared processor pool if vm is
    # configured with dedicated processors
    if (db_procmode and 'shared_proc_pool_id' in update_pwr_specs
            and not _compare_spec_value(db_procpoolid, drv_procpoolid)):
        # additional checking for backup/restore scenario
        # if restored prior from 1.2.3.2, it won't have the shared proc pool
        # info in the DB.
        # so, if db value is not None and driver value is not set to 0,
        # it should be considered as out of band changes.
        if db_procpoolid and drv_procpoolid != 0:
            updates['proc_pool_id'] = drv_procpoolid
    if (db_procmode and 'shared_proc_pool_name' in update_pwr_specs
            and not _compare_spec_value(db_procpoolname, drv_procpoolname)):
        # additional checking for backup/restore scenario
        # if restored prior from 1.2.3.2, it won't have the shared proc pool
        # info in the DB.
        # so, if db value is not None and driver value is not set to
        # DefaultPool, it should be considered as out of band changes.
        if db_procpoolname and drv_procpoolname != DEFAULT_PROC_POOL:
            updates['proc_pool_name'] = drv_procpoolname
    if db_ameef and drv_ameef and not _compare_spec_value(db_ameef, drv_ameef):
        updates['ame_expansion_factor'] = drv_ameef
    ch_for_updates = ['vcpus', 'memory_mb', 'srr_capability',
                      'proc_pool_id', 'proc_pool_name',
                      'ame_expansion_factor']
    changed = (any(p in updates for p in ch_for_updates) or
               (db_procs and drv_procs and not _compare_spec_value(db_procs,
                                                                   drv_procs)))
    if changed:
        LOG.info(six.text_type('out of band change updates for instance %s '
                               ': %s db_procs: %s '
                               'drv_procs: %s')
                 % (instance['display_name'],
                    updates, db_procs, drv_procs))
    return changed


def _handle_pending_action(context, instance, power_specs):
    """Helper method to Update the Pending Action flag for Awaiting Reboot"""
    vm_state = instance.get('vm_state', '')
    pwr_state = power_specs.pop('power_state', 0)
    active_states = [vm_states.ACTIVE, vm_states.RESIZED]
    pending_action = power_specs.pop('pending_action', None)
    attribs = ['memory_mb', 'min_memory_mb', 'max_memory_mb',
               'proc_units', 'min_proc_units', 'max_proc_units',
               'vcpus', 'min_vcpus', 'max_vcpus']
    power_specs['pending_action'] = None
    no_pending_action_update = False
    # If the VM State is in-progress we want to wait until it finishes,
    # also if the VM is Stopped then obviously a reboot isn't needed
    if vm_state not in active_states or pwr_state != power_state.RUNNING:
        no_pending_action_update = True
    # If we were provided a Pending Action already, don't need to check it
    if pending_action is not None:
        power_specs['pending_action'] = pending_action
        no_pending_action_update = True

    # Compare the Pending/Current Values to see if a Reboot is still needed
    for attr in attribs:
        # remove attributes that's not stored in instance db
        run_attr = power_specs.pop('run_' + attr, None)
        pend_attr = power_specs.pop('pending_' + attr, None)
        # no need to check for pending actions
        if not no_pending_action_update:
            cur_attr = power_specs.get(attr)
            # Compare the Current and Pending Value to see if they are
            # different
            if cur_attr and pend_attr and cur_attr != pend_attr:
                power_specs['pending_action'] = 'awaiting_restart'
            # If this is a min/max, there is only pending, no running value
            if attr.startswith('min_') or attr.startswith('max_'):
                continue
            # If Running Value is 0, the VM is stopped so exclude
            if str(run_attr) == '0' or str(run_attr) == '0.00':
                continue
            # Compare the Current and Running Value to see if they are
            # different
            if cur_attr and run_attr and cur_attr != run_attr:
                power_specs['pending_action'] = 'awaiting_restart'


def _notify_out_of_band_change(context, instance):
    """Helper method to Notify when Out-of-Band Changes occur"""
    # First write out a statement to the log file saying it changed
    LOG.info(_('Resource allocation of instance %(instance_name)s '
               'on host %(host_name)s has been modified.') %
             {'instance_name': instance['display_name'],
              'host_name': instance['host']})
    # Construct a translatable message to log and send in the notification
    msg = _('Resource allocation of instance {instance_name} '
            'on host {host_name} has been modified.')
    # Construct the full payload for the message that is being sent
    info = {'msg': msg, 'instance_id': instance['uuid']}
    info['instance_name'] = instance['display_name']
    info['host_name'] = instance['host']
    # Send the notification that the Allocations changed Out-of-Band
    notifier = rpc.get_notifier(service='compute', host=instance['host'])
    notifier.info(context, 'compute.instance.log', info)


def _is_instance_resizing(instance):
    """Helper method to see if the Instance is currently being resized"""
    resize_states = [task_states.RESIZE_PREP, task_states.RESIZE_MIGRATING,
                     task_states.RESIZE_MIGRATED, task_states.RESIZE_FINISH]
    # If the VM State is Resized then it has just finished resizing
    if instance['vm_state'] == vm_states.RESIZED:
        return True
    # Otherwise see if the Task State is one of the Resizing Tasks
    return instance['task_state'] in resize_states


def _is_instance_spawning(instance):
    """Helper method to see if the Instance is currently being deployed"""
    spawn_states = [task_states.SCHEDULING,
                    task_states.BLOCK_DEVICE_MAPPING,
                    task_states.NETWORKING,
                    task_states.SPAWNING]
    # If the VM State is BUILDING then it's in the middle of deploying
    if instance['vm_state'] == vm_states.BUILDING:
        return True
    # Otherwise see if the Task State is one of the spawning Tasks
    return instance['task_state'] in spawn_states


def _compare_spec_value(oldval, newval):
    """Helper method to compare if the old and new spec value are the same"""
    # If both are None, then obviously they are the same value
    if oldval is None and newval is None:
        return True
    # We checked if both are None, so if either is None they aren't equal
    if oldval is None or newval is None:
        return False
    # We need to do something special if it is a float specified
    if isinstance(oldval, float) or isinstance(newval, float):
        oldval, newval = ("%0.2f" % float(oldval), "%0.2f" % float(newval))
    # Otherwise we can do the comparison as string values
    return six.text_type(oldval) == six.text_type(newval)


def _duplicate_system_metadata(updates, sys_metadata_changes):
    """synchronize instance system metadata with power_specs changes.
    Not every power_specs attributes needs to duplicated to instance system
    metadata. It would require detail explanation why additional attributes
    are needed to duplicate between power_specs and system metadata.
    :param updates: list of instance attributes we will update
    :param sys_metadata_changes: This dict includes system metadata needs to
    be updated because it's out of sync with power_specs. It's use to close
    up the scenario where system_metadata is override with old value due to
    instance.save() in powering on/off.
    """
    # TODO: Remove this when Scheduler/REST API's aren't using Meta-data
    system_metadata = dict(updates_only=True)
    power_specs = updates.get('power_specs', {})
    # Loop through the PowerSpecs, mapping to their equivalent Meta-data
    for key, val in six.iteritems(power_specs):
        if key in POWER_SPECS_SYS_META_MAP:
            system_metadata[POWER_SPECS_SYS_META_MAP[key]] = str(val)
    for key, val in six.iteritems(sys_metadata_changes):
        if system_metadata.get(key) != val:
            LOG.warn(_('system_metadata %(key)s is not in sync with '
                       'power_specs. current: %(curr)s new: %(new)s') %
                     {'key': key,
                      'curr': system_metadata.get(key),
                      'new': val})
            system_metadata[key] = val
    updates['system_metadata'] = system_metadata


def _sys_metadata_updates(curr_sys_metadata, latest_power_specs):
    """Compare instance system metadata with power_specs to close up any
    cases where power_specs is updated but not system_metadata
    :param curr_sys_metadata: current instance system metadata
    :param latest_power_specs: power_specs info used for update
    """
    changes = {}
    for spec_key, sys_metadata_key in six.iteritems(POWER_SPECS_SYS_META_MAP):
        sys_meta_val = curr_sys_metadata.get(sys_metadata_key)
        change = latest_power_specs.get(spec_key)
        if not sys_meta_val and not change:
            continue
        if not _compare_spec_value(sys_meta_val, change):
            changes[sys_metadata_key] = str(change) if change else ''
    return changes


def _check_os_distro_is_aix_for_ame_resize(context, flavor, inst_distro,
                                           inst_name, curr_exp_factor=None):
    """Validate the AME factor and override if necessary

    Override the AME factor if the OS is unknown and AME is attempted to be
    set. If AME is set then the expansion factor is allowed to be changed
    regardless of the OS. If coming from the API and the AME factor is
    anything other than 0 for non AIX VMs, override and set to 0.

    :param context: Context
    :param flavor: The new flavor object
    :param inst_distro: The OS distribution of the VM being resized
    :param inst_name: The name of the instance being resized
    :param curr_exp_factor: Current expansion factor as a string
    :returns new_flav: The flavor object to use
    """
    new_flav = flavor
    req_exp_factor = (
        flavor['extra_specs'].get('powervm:ame_expansion_factor'))
    if req_exp_factor is None:
        return new_flav
    # Convert expansion factor to float for validation
    exp_fact_float = round(float(req_exp_factor), 2)
    curr_fact_float = (0.0 if curr_exp_factor is None else
                       round(float(curr_exp_factor), 2))
    # In the case of onboarded VMs we might not know the OS, so if the
    # LPAR currently doesn't have AME set and the new flavor does
    # then override the flavor because the OS may not be AIX
    if inst_distro is None:
        if not curr_fact_float and exp_fact_float != 0:
            flavor['extra_specs']['powervm:ame_expansion_factor'] = '0.0'
            updated_flavor = _get_flavor_from_existing(context, flavor)
            LOG.warn(_('Overriding AME factor to 0 for instance: '
                       '%(instance_name)s.') %
                     {'instance_name': inst_name})
            info = {'msg': _(
                'The desired active memory expansion factor for '
                'virtual machine {instance_name} will not be set. '
                'This is because the operating system is unknown. '
                'The process for bringing the virtual machine under '
                'management by PowerVC must be completed.'),
                'instance_name': inst_name}
            notifier = rpc.get_notifier(service='compute',
                                        host=CONF.host)
            notifier.warn(get_admin_context(), 'compute.instance.log',
                          info)
            new_flav = updated_flavor
    # Otherwise if the OS isn't AIX but we know it and the flavor
    # is setting the expansion factor, then override the flavor.
    elif inst_distro != 'aix' and exp_fact_float != 0:
        flavor['extra_specs']['powervm:ame_expansion_factor'] = '0.0'
        updated_flavor = _get_flavor_from_existing(context, flavor)
        LOG.warn(_('Overriding AME factor to 0 for instance: '
                   '%(instance_name)s.') %
                 {'instance_name': inst_name})
        info = {'msg': _(
            'The active memory expansion factor for virtual '
            'machine {instance_name} will not be set because '
            'the OS distribution is not AIX.'),
            'instance_name': inst_name}
        notifier = rpc.get_notifier(service='compute',
                                    host=CONF.host)
        notifier.warn(get_admin_context(), 'compute.instance.log',
                      info)
        new_flav = updated_flavor
    return new_flav


def _get_flavor_from_existing(context, flavor_obj):
    """Find or create a flavor from a flavor object

    Using the attributes from an existing flavor object that may have been
    modified, check to see if a flavor with the same stats already exists.
    If not create a new one with the new stats.

    :param context: context
    :param flavor_obj: The flavor object to get attributes from
    :returns inst_type: Either an existing or newly created flavor object
    """

    vcpus = flavor_obj.get('vcpus', 0)
    memory_mb = flavor_obj.get('memory_mb', 0)
    root_gb = flavor_obj.get('root_gb', 0)
    ephem_gb = flavor_obj.get('ephemeral_gb', 0)

    #We want to create a unique/consistent Flavor Name based on the Values
    text_name = ("auto-flavor-%(vcpus)s-%(memory_mb)s-"
                 "%(root_gb)s-%(ephem_gb)s" % {'vcpus': vcpus,
                 'memory_mb': memory_mb, 'root_gb': root_gb,
                 'ephem_gb': ephem_gb})
    #Loop through the Extra Specs provided, adding them to the Name
    extra_specs = flavor_obj.get('extra_specs', {})
    sorted_keys = sorted(extra_specs)
    for key in sorted_keys:
        text_name = '-'.join([text_name, key, str(extra_specs[key])])
    #Create an MD5 Hash to use for the Name of the Flavor
    flavor_name = hashlib.md5(text_name).hexdigest()
    try:
        #Query the DB to see if there is an existing Flavor we can reuse
        inst_type = flavors.get_flavor_by_name(flavor_name, context)
    except exception.FlavorNotFoundByName:
        LOG.debug("Flavor '%s' not found. Creating a new one.", flavor_name)
        #If there isn't a matching Flavor, we will have to create a new one
        inst_type = flavor.Flavor(
            context=context,
            name=flavor_name, memory_mb=memory_mb, vcpus=vcpus,
            root_gb=root_gb, ephemeral_gb=ephem_gb,
            flavorid=flavor_name, swap=0, is_public=False,
            projects=[context.project_id],
            extra_specs=extra_specs)
        inst_type.create()
    return inst_type
