#
#
# Copyright 2013, 2016 IBM Corp.
# All Rights Reserved.
# Copyright 2011 Justin Santa Barbara
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Virtualization driver for PowerVM.

Implements all virtual driver functions supported by the HMC driver
"""
import copy
import eventlet
import os
import six
import sys

from eventlet import greenio
from eventlet import greenthread
from eventlet import patcher

from nova import context as ctx
from nova import exception as nova_exception
from nova import network as net_api
from nova.network import base_api
from nova.network import model as net_model
from nova import objects
from nova import rpc
from nova.compute import power_state
from nova.compute import task_states
from nova.compute import vm_states
from nova.objects import instance as instance_obj
from oslo_utils import excutils
from oslo_utils import importutils
from oslo_serialization import jsonutils
from oslo_log import log as logging
from nova.virt import driver
from nova.virt import event as virtevent
from nova.virt import hardware

from oslo_config import cfg

from oslo_concurrency import lockutils

from powervc_oslo.config import data_utils
from powervc_oslo.k2.k2operator import K2Error, K2HttpError
from powervc_oslo.k2 import k2operator as k2o

from powervc_nova import return_cached_driver_data
from powervc_nova import cache_param_data
from powervc_nova.compute import api as pvc_compute_rpcapi
from powervc_nova.compute import exception as pvc_exc
from powervc_nova.objects.compute import dom as compute_dom
from powervc_nova.objects.storage import dom as storage_dom
from powervc_nova.objects.compute.dom import SharedProcessorPoolFactory
from powervc_nova.objects import migrate_data as mig_obj

from powervc_nova.virt import discovery_driver as disc_driver
from powervc_nova.compute import notify_messages as notify_msg
from powervc_nova.compute import send_ui_notification
from powervc_nova.compute.kvmhmc_manager import WAITING_ON_HOST
from powervc_nova.virt import instance_error_seen_multiple_time
from powervc_nova.virt import instance_found_remove_from_cache
from powervc_nova.virt import set_instance_error_state_and_notify
from powervc_nova.virt.ibmpowervm.common import exception as pvm_exc
from powervc_nova.virt.ibmpowervm.common import migrate_utils
from powervc_nova.virt.ibmpowervm.common import power_spec_utils
from powervc_nova.virt.ibmpowervm.common import constants as pvc_consts
from powervc_nova.virt.ibmpowervm.common.image import Image
from powervc_nova.virt.ibmpowervm.common.image import ImageEntry

from powervc_nova.virt.ibmpowervm.common import constants as commonconsts
from powervc_nova import logcall, logcall_log_instance
from powervc_nova import _

from powervc_nova.virt.ibmpowervm.hmc.exception import IBMPowerMACNotSupported
import powervc_nova.virt.ibmpowervm.hmc.host as pvm_host
import powervc_nova.virt.ibmpowervm.hmc.hmc as pvm_hmc
import powervc_nova.virt.ibmpowervm.hmc.lpar as pvm_lpar
import powervc_nova.virt.ibmpowervm.hmc.migrate as pvm_migrate
import powervc_nova.virt.ibmpowervm.hmc.vscsi as pvm_vscsi
import powervc_nova.virt.ibmpowervm.hmc.remote_restart as pvm_rebuild
from powervc_nova.virt.ibmpowervm.hmc import event as hmcevent
from powervc_nova.virt.ibmpowervm.hmc.import_vm import ImportVM
from powervc_nova.virt.ibmpowervm.hmc.k2oprguard import K2OperGuard
from powervc_nova.virt.ibmpowervm.hmc.k2 import contains_msg_in_k2_err
from powervc_nova.virt.ibmpowervm.hmc.k2.mgmt_console_wrapper \
    import HMCVersion
from powervc_nova.virt.ibmpowervm.hmc.k2.scsi_map_processor \
    import SCSIMapProcessor
from powervc_nova.virt.ibmpowervm.hmc.k2 import storage_map_processor
from powervc_nova.virt.ibmpowervm.hmc.k2 import fc_map_processor
import powervc_nova.virt.ibmpowervm.hmc.k2.constants as k2const
from powervc_nova.virt.ibmpowervm.hmc.k2.jobentry_wrapper \
    import JobEntryWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2.msentry_wrapper \
    import ManagedSystemEntryWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2.lparentry_wrapper \
    import LPAREntryWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2.shr_prc_pool_wrapper \
    import ShrPrcPoolWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2.viosentry_wrapper \
    import VIOSEntryWrapper
from powervc_nova.virt.ibmpowervm.hmc.virtualoptical import VirtualOptical

from powervc_nova.network.ibmpowervm import net_assn_cleanup
import random
from nova.objects import block_device as blkdev_obj

from powervc_nova.virt.ibmpowervm.hmc import enterprise_pool as hmc_ep

native_Queue = patcher.original("Queue")

LOG = logging.getLogger(__name__)
CONF = cfg.CONF
# Task state for VM if host is down

k2_session_logon = k2o.K2Session._logon
stop_k2_communications = False

log_warning = logging.BaseLoggerAdapter.warning


def k2o_logon_choke(self):
    """
    This method is used as a choke point for all K2O logon failure.
    _logon is called implicitly by k2session.request if we have not
    log on or session expired.
    If a 401 is returned it means we could not successfully log onto
    the HMC.  In that case we need to stop all tasks from trying additional
    log ons so we don't lock out the HMC user id.
    """
    # If k2 communication are shut off just wait
    global stop_k2_communications
    while stop_k2_communications:
        try:
            greenthread.sleep(60)
        finally:
            LOG.error(_('Received 401 unauthorized error from HMC %s.')
                      % self.host)
            sys.exit()
    try:
        LOG.debug("K2 choke point called")
        # Call K2O request method
        k2_session_logon(self)
    except k2o.K2Error as e:
        # Handle special situations
        response = e.k2response
        if (response is not None and
                response.status == k2const.HTTP_STATUS_UNAUTHORIZED):
            try:
                stop_k2_communications = True

                # Send any notifications
                LOG.error(_('The compute process is ending.'))
                info = {'msg': _('Unable to access HMC {hmc} due to '
                        'an authorization error.'),
                        'hmc': self.host}
                context = ctx.get_admin_context()
                notifier = rpc.get_notifier(service='compute', host=CONF.host)
                notifier.error(context, 'compute.instance.log', info)
                # This is to trigger the health monitoring to update
                # HMC health
                # NOTE: potential issue if the string is translated.
                LOG.error(_('Received 401 unauthorized error from '
                            'HMC %(host)s. Response: %(response)s')
                          % {'host': self.host, 'response': response.body})
                rpc_api = pvc_compute_rpcapi.PowerVCComputeRPCAPI()
                rpc_api.end_compute_with_hmc(context, {'hmc_ip': self.host})
                # give it some time to sent out the rpc cast
                greenthread.sleep(5)
            finally:
                sys.exit()
        raise e


def log_exc_as_warning(self, msg, *args, **kwargs):
    """
    Log exception as WARNING instead of ERROR.
    This helps reduce the 'ERROR' noises in the log files.
    """
    if 'exc_info' not in kwargs:
        kwargs['exc_info'] = True
    log_warning(self, msg, *args, **kwargs)


class PowerVMDriver(driver.ComputeDriver, disc_driver.ComputeDiscoveryDriver):

    capabilities = {
        "has_imagecache": False,
        "supports_recreate": True,
        "supports_migrate_to_same_host": False
    }
    """PowerVM HMC Implementation of Compute Driver."""

    def __init__(self, virtapi):
        super(PowerVMDriver, self).__init__(virtapi)
        self.host_list = pvm_host.HostList()
        self.hmc_list = pvm_hmc.HMCList()
        self.migrate = {}
        self.rebuild_dict = {}
        self.network_topo = None
        k2o.K2Session._logon = k2o_logon_choke
        # override LOG.exception to log as WARNING instead
        logging.BaseLoggerAdapter.exception = log_exc_as_warning

    @logcall
    def init_host(self, host):
        """Initialize anything that is necessary for the driver to function,
        including catching up with currently running VM's on the given host."""

        mtm_serial = CONF.host
        # The host_id format is: 8246L1C_SN0604C6A
        # host_uniqueid = 8776e2ab-403a-359c-bb09-162038300b6b
        context = ctx.get_admin_context()
        host_uuid = CONF.powervm_mgr_uuid

        # initialize the host and hmcs for the host
        host_obj = pvm_host.Host(uuid=host_uuid, mtm_serial=mtm_serial,
                                 host_id=mtm_serial, virtapi=self.virtapi)
        pvm_hmc.get_consoles_for_host_sys(self.virtapi, mtm_serial,
                                          self.hmc_list, host_obj)
        # Retrieve the Instances from the Database to Load the UUID Cache
        instfact = compute_dom.ManagedInstanceFactory.get_factory()
        dbinsts = instfact.find_all_instances_by_host(context, CONF.host)

        # If this is right after the switch from PVM, do data conversion
        if CONF.powervm_switch_type == 'pvm':
            self.__convert_data_for_host_switch(context, dbinsts)

        # use K2 level caching
        with K2OperGuard(host=host_obj) as k2oper:
            lpars_resp = k2oper.read(k2const.MGT_SYS, rootId=host_uuid,
                                     childType=k2const.LPAR)
        host_obj.load_UUID_cache(dbinsts, lpars_resp.feed)

        # initialize vif driver
        host_obj.init_vif_driver()
        host_obj.enable_long_term_monitoring()

        self.host_list.add(mtm_serial, host_obj)
        pvm_host.compute_host = host_obj
        # Perform the Network Topology on Initialization to make sure is valid
        topo_class = CONF.powervm_vif_topo_driver
        with K2OperGuard(host=host_obj) as k2opr:
            topo_drv = importutils.import_object(topo_class,
                                                 host_name=host,
                                                 operator=k2opr)
        try:
            self.network_topo = topo_drv.get_host_network_topology(context)
        except Exception as e:
            # Just log the exception and continue on if Network Topology failed
            LOG.error(_("Reconciling Host-Network Topology failed."))
            LOG.exception(e)

        # Collect the enterprise pool data for the host. Note that this is
        # susceptible to a race condition with a set mobile cores operation,
        # but the likelihood that this occurs as well as the fact it will be
        # resolved by the conductor's periodic task means that it makes more
        # sense to ensure the DB is updated immediately after a host is
        # registered.
        try:
            LOG.info(_('Collecting Host-Enterprise-Pool Topology...'))
            hmc_ep.sync_pool_for_host(context, host_obj)
            LOG.info(_('Done Collecting Host-Enterprise-Pool Topology.'))
        except Exception as e:
            # Just log the exception and continue
            LOG.error(_("Reconciling Host-Enterprise-Pool Topology failed."))
            LOG.exception(e)
        self._init_events()

        # Log that the Initialization of the Host has completed
        LOG.info(_('Initialization of host %s has completed.') % mtm_serial)
        LOG.debug("Configured host %s" % pvm_host.compute_host.uuid)

    @logcall
    def _resolve_instance(self, instance,
                          force_refresh=False, update_db=False,
                          lpars_resp=None, log_notfound=True):

        def find_lpar(entries, uuids):
            for entry in entries:
                for uuid in uuids:
                    if uuid.lower() == entry.properties['id'].lower():
                        return entry, uuid

            return None, None

        name = instance['name']
        # Migration operations to the target thread cannot use instance['host']
        # because that is a reference to the source host
        host_name = CONF.host
        uuid = None
        entry = None
        try:
            host = self.host_list.get_list().get(host_name)
            # See if the power_specs are available and if it has
            # the powervm uuid that we need
            uuids = []
            pwr_spec_uuid = None
            try:
                pwr_spec_uuid = instance.get('power_specs').get('vm_uuid')
                if pwr_spec_uuid is not None:
                    uuids.append(pwr_spec_uuid)
            except Exception:
                pass

            # The power spec may not be correct. In the case of a deploy
            # reschedule it will have an old value, so use our cache and
            # see which one we find on the system.
            cache_uuid = host.pvm_uuid_cache.get_pvm_uuid(name)
            if cache_uuid is not None and cache_uuid not in uuids:
                uuids.append(cache_uuid)

            # if lpars_resp is passed in, we use the feed to find the entry
            if lpars_resp is not None:
                entry, uuid = find_lpar(lpars_resp.feed.entries, uuids)
            else:
                # use K2 level caching
                last_exc = None
                with K2OperGuard(host=host) as k2oper:
                    age = 0 if force_refresh else -1
                    for lpar_uuid in uuids:
                        try:
                            lpars_resp = k2oper.read(k2const.MGT_SYS,
                                                     rootId=host.uuid,
                                                     childType=k2const.LPAR,
                                                     childId=lpar_uuid,
                                                     age=age)
                        except K2Error as e:
                            # it's possible that we cannot find the k2response
                            # for the given uuid if out-of-band operations
                            # has been performed
                            # check whether we received 404 status
                            response = e.k2response
                            if (response is not None and
                                    response.status ==
                                    k2const.HTTP_STATUS_NOT_FOUND):
                                LOG.info(_('LPAR with uuid %s is not found.') %
                                         lpar_uuid)
                            else:
                                # remember the exception in case there are
                                # multiple uuids
                                last_exc = e
                        if lpars_resp is not None:
                            uuid = lpar_uuid
                            entry = lpars_resp.entry
                            break
                # if we cannot find the lpar entry, and there was an exception
                if entry is None and last_exc is not None:
                    raise last_exc

            # Since the OpenStack Name may not match the actual LPAR Name
            if entry is not None:
                lpar_name = LPAREntryWrapper(entry).get_lpar_name()
                # See if the name has changed
                if update_db and lpar_name != name:
                    # Update our mapping for the name in the db and cache
                    admin_context = ctx.get_admin_context()
                    instfact = compute_dom.ManagedInstanceFactory.get_factory()
                    managed_inst = instfact.find_instance_by_uuid(
                        admin_context, instance['uuid'])
                    managed_inst.power_specs = dict(vm_name=lpar_name)
                    managed_inst.save()
                    host.pvm_uuid_cache.rename(name, lpar_name, uuid)

                # Always return the actual lpar name
                name = lpar_name
            elif log_notfound:
                # cannot find an entry matching to the uuid
                LOG.warning(_('Virtual machine %(name)s cannot be found. '
                              'LPAR UUID is: %(uuid)s.')
                            % {'name': name,
                               'uuid': uuids})
                # Don't log full feed for a missing LPAR, that could be normal
        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.exception(_("Error resolving instance: %s")
                              % six.text_type(e))
        return name, host, uuid, entry

    @logcall
    @cache_param_data([{'lpars_feed': ['get_lpars_feed']}])
    def get_info(self, instance, lpars_feed=None):
        """Get the current status of an instance, by name (not ID!)

        Returns a dict containing:

        :state:           the running state, one of the power_state codes
        :max_mem:         (int) the maximum memory in KBytes allowed
        :mem:             (int) the memory in KBytes used by the domain
        :num_cpu:         (int) the number of virtual CPUs for the domain
        :cpu_time:        (int) the CPU time used in nanoseconds
        """

        name, node, uuid, entry = self._resolve_instance(instance,
                                                         update_db=True,
                                                         lpars_resp=lpars_feed)
        LOG.debug("Get info for %s on host %s." % (name, CONF.host))

        if entry is not None:
            # If current task state is waiting on host, it
            # needs to be cleared, since an LPAR entry was
            # successfully obtained
            if instance['task_state'] == WAITING_ON_HOST:
                self._queue_event(
                    hmcevent.LifecycleEvent(instance['uuid'],
                                            hmcevent.EVENT_CLEAR_TASK))
            info = pvm_lpar.get_info(entry)
            # Remove the Instance from the NotFound Cache when it is reported
            instance_found_remove_from_cache(instance)
        else:
            info = hardware.InstanceInfo(state=power_state.NOSTATE)
            host_state = node.get_state()
            self._handle_null_entry(instance, host_state)
        return info

    def _handle_null_entry(self, instance, state):
        """Method to handle _resolve_instance() returning None for
        the entry. If it was unable to be obtained because the host
        was not operational, then issue a host down event to update
        the vm task state to WAITING_ON_HOST and to update the vm state
        to stopped if the host was powered off. Otherwise, if the host
        is up, set the instance to error state.
        :param instance: db instance
        :param state: state of host
        """
        instance_name = instance['name']
        LOG.debug('LPAR %s not found' % instance_name)
        inst_vm_state = instance['vm_state']
        inst_task_state = instance['task_state']
        task_state_ignore_list = [task_states.MIGRATING, task_states.DELETING,
                                  task_states.REBUILDING,
                                  task_states.REBUILD_BLOCK_DEVICE_MAPPING,
                                  task_states.REBUILD_SPAWNING]
        repeats = CONF.number_of_exceptions_needed_for_missing_inst_hmc

        # If host is not running, issue a host down event. This will set
        # set the VM's task state to WAITING_ON_HOST and the VM's state
        # to stopped if the host went to the powered off state.
        if state not in ('operating', 'standby'):
            if inst_task_state is None and inst_vm_state not in (
                    vm_states.ERROR, vm_states.BUILDING):
                # If there is no pending task, issue the host down event
                LOG.debug('Issuing a host down event for vm %s' %
                          instance_name)
                self._queue_event(hmcevent.LifecycleEvent(
                                  instance['uuid'], hmcevent.EVENT_HOST_DOWN,
                                  host_state=state))

        # If host is okay, and we truly can't find entry for
        # instance and VM is already in error,
        # clear the task state.
        elif (inst_task_state == WAITING_ON_HOST and
              inst_vm_state == vm_states.ERROR):
            self._queue_event(hmcevent.LifecycleEvent(
                              instance['uuid'], hmcevent.EVENT_CLEAR_TASK))

        # If host is okay, and we truly can't find entry for
        # instance, set state to error.
        elif (inst_task_state not in task_state_ignore_list and
              inst_vm_state in (vm_states.ACTIVE, vm_states.STOPPED) and
              instance_error_seen_multiple_time(instance, repeats)):
            set_instance_error_state_and_notify(instance)

    @logcall
    def list_instances(self):
        """
        Return the names of all the instances known to the virtualization
        layer, as a list.
        """
        inst_list = []

        # For each host, get the feed cache and list the LPARs
        interval = CONF.ibmpowervm_instance_cache_interval
        for node in self.host_list.itervalues():
            # use K2 level caching
            with K2OperGuard(host=node) as k2oper:
                # need to specify the age attirbute
                try:
                    lpars_resp = k2oper.read(k2const.MGT_SYS, rootId=node.uuid,
                                             childType=k2const.LPAR,
                                             age=interval)
                except K2Error as e:
                    # stg defect 19632
                    # this is a work around for SW256810
                    # If we received HSCL8011, it means a LPAR has been
                    # deleted in the middle of GET LPARS request.
                    # We will retry the GET request again.
                    # If we continue receiving this error after retry,
                    # K2 should really fix this problem in 810
                    if contains_msg_in_k2_err(e, 'HSCL8011'):
                        lpars_resp = k2oper.read(k2const.MGT_SYS,
                                                 rootId=node.uuid,
                                                 childType=k2const.LPAR,
                                                 age=interval)
                    else:
                        with excutils.save_and_reraise_exception():
                            # If error retrieving managed system's LPARs from
                            # K2, issue a host down event to update the task
                            # state of all the instances on this host to
                            # WAITING_ON_HOST.
                            db_instances = node.get_instances_from_db()
                            for instance in db_instances:
                                if (instance['task_state'] is None and
                                    instance['vm_state'] not in (
                                        vm_states.ERROR, vm_states.BUILDING)):
                                    self._queue_event(hmcevent.LifecycleEvent(
                                        instance['uuid'],
                                        hmcevent.EVENT_HOST_DOWN))

            # If no lpars came back from K2, then any instances in the
            # database should be handled an put into appropriate states.
            host_state = node.get_state()
            if lpars_resp.feed.entries == []:
                db_instances = node.get_instances_from_db()
                for instance in db_instances:
                    self._handle_null_entry(instance, host_state)

            # Catch exceptions from each host, don't let one
            # host crater the request
            try:
                # refresh the uuid cache
                node.pvm_uuid_cache.load_cache_from_feed(lpars_resp.feed,
                                                         refresh_only=True)
                # Get the LPARs from the feed
                for entry in lpars_resp.feed.entries:
                    lpar_uuid = entry.properties['id']
                    lpar_name = (
                        entry.element.findtext(k2const.XP_PARTITION_NAME))
                    # Look up the name that OpenStack knows the LPAR as
                    # We will default to the Name from K2 if not found
                    inst_name = node.pvm_uuid_cache.get_pvm_name(lpar_uuid)
                    # Default to using the K2 Name, if it isn't managed
                    if inst_name is None:
                        inst_name = lpar_name
                    inst_list.append(inst_name)
            except Exception as e:
                LOG.exception(_("List of virtual machines failed: %s")
                              % six.text_type(e))

        return inst_list

    def vopt_cleanup(self, instance_list):
        """ remove virtual opticals for given instances """

        host_name = CONF.host
        host = self.host_list.get_list().get(host_name)
        vopt = VirtualOptical()
        vopt.timed_cleanup(host, instance_list)

    @logcall_log_instance
    def spawn(self, context, instance, image_meta, injected_files,
              admin_password, network_info=None, block_device_info=None):
        """
        Create a new instance/VM/domain on the virtualization platform.

        Once this successfully completes, the instance should be
        running (power_state.RUNNING).

        If this fails, any partial instance should be completely
        cleaned up, and the virtualization platform should be in the state
        that it was before this call began.

        :param context: security context
        :param instance: Instance object as returned by DB layer.
                         This function should use the data there to guide
                         the creation of the new instance.
        :param image_meta: image object returned by nova.image.glance that
                           defines the image from which to boot this instance
        :param injected_files: User files to inject into instance.
        :param admin_password: Administrator password to set in instance.
        :param network_info:
           :py:meth:`~nova.network.manager.NetworkManager.get_instance_nw_info`
        :param block_device_info: Information about block devices to be
                                  attached to the instance.
        """
        mesg = (_('Deploying'))
        self.log_operation(mesg, instance)
        host_name = instance['host']
        host = self.host_list.get_list().get(host_name)
        if host is None:
            LOG.error(_('Unable to extract information for host %(host_name)s'
                        ' from the virtual machine configuration: %(inst)s,'
                        ' %(list)s') %
                      {'host_name': host_name,
                       'inst': str(instance),
                       'list': str(self.host_list.get_list())})
        try:
            callback = block_device_info['powervc_blockdev_callbacks']
            validate_volumes = callback['validate_volumes']
            attach_volumes = callback['attach_volumes']
            validate_attachment = callback['validate_attachment']
        except Exception:
            validate_volumes = None
            attach_volumes = None
            validate_attachment = None

        pvm_lpar.spawn(context, self.virtapi, host, instance, image_meta,
                       injected_files, admin_password,
                       network_info, block_device_info,
                       validate_volumes, attach_volumes,
                       validate_attachment)
        return

    @logcall_log_instance
    def destroy(self, context, instance, network_info, block_device_info=None,
                destroy_disks=True, migrate_data=None):
        """Destroy (shutdown and delete) the specified instance.

        If the instance is not found (for example if networking failed), this
        function should still succeed.  It's probably a good idea to log a
        warning in that case.

        :param instance: Instance object as returned by DB layer.
        :param network_info:
           :py:meth:`~nova.network.manager.NetworkManager.get_instance_nw_info`
        :param block_device_info: Information about block devices that should
                                  be detached from the instance.
        :param destroy_disks: Indicates if disks should be destroyed

    """
        mesg = (_('Deleting'))
        self.log_operation(mesg, instance)
        lpar_name, host, lpar_uuid, entry = \
            self._resolve_instance(instance)
        pvm_lpar.destroy(host,
                         lpar_name,
                         lpar_uuid,
                         entry,
                         destroy_disks)
        # Unplug VIFs from VIOS SEA
        if network_info is not None:
            LOG.debug("network_info: %s" % network_info)
            self.unplug_vifs(instance, network_info)

    @logcall
    def cleanup_disks(self, context, connection_info, instance, vios_name):
        """ cleanup the hdisk left after vscsi mapping"
        """
        lpar_name, host, lpar_uuid, entry = self._resolve_instance(instance)
        pvm_vscsi.cleanup_disks(host,
                                instance,
                                lpar_name,
                                connection_info, context, vios_name)

    @logcall
    def call_terminate_connection(self, context, instance, volume_id,
                                  connector, conninfo=None, clean=True):
        """ cleanup the hdisk left after vscsi mapping"
        """
        lpar_name, host, lpar_uuid, entry = self._resolve_instance(instance)
        pvm_vscsi.call_terminate_connection(host, context, instance, lpar_name,
                                            volume_id, connector, conninfo,
                                            clean=clean)

    @logcall
    def attach_volume(self, context, connection_info, instance, mountpoint,
                      disk_bus=None, device_type=None, encryption=None):
        """Attach the disk to the instance at mountpoint using info."""
        mesg = (_('Attaching volume'))
        self.log_operation(mesg, instance)
        lpar_name, host, lpar_uuid, entry = self._resolve_instance(instance)
        pvm_lpar.attach_volume(host,
                               instance,
                               lpar_name,
                               lpar_uuid,
                               entry,
                               connection_info, context)

    @logcall
    def detach_volume(self, connection_info,
                      instance, mountpoint, encryption=None):
        """Detach the disk attached to the instance."""
        mesg = (_('Detaching volume'))
        self.log_operation(mesg, instance)
        lpar_name, host, lpar_uuid, entry = self._resolve_instance(instance)
        pvm_lpar.detach_volume(host,
                               lpar_name,
                               lpar_uuid,
                               entry,
                               connection_info)

    @logcall
    def get_volume_mapped_entries(self, volume_udid, vios_name):
        """Get the number of SCSI mappings associated for the disk."""
        LOG.debug('Entering get_volume_mapped_entries')
        host = self.host_list.get_list().get(CONF.host)
        num_mapping = pvm_vscsi.get_volume_mapped_entries(host,
                                                          volume_udid,
                                                          vios_name)
        return num_mapping

    @logcall
    def get_volume_connector(self, instance, isVscsi=False):
        """Get connector information for the instance for attaching to volumes.
        """
        lpar_name, host, lpar_uuid, entry = self._resolve_instance(instance)

        # Since get_volume_connector() get called during delete instance,
        # we need to handle the situation if instance no longer exists.
        # If we cannot find the instance, just return None.
        if (lpar_uuid is None or entry is None) and (not isVscsi):
            LOG.debug('Instance not found. '
                      'LPAR UUID: %(lpar_uuid)s. Entry: %(entry)s.' %
                      {'lpar_uuid': lpar_uuid, 'entry': entry})
            return None
        return pvm_lpar.get_volume_connector(host, instance,
                                             lpar_name, lpar_uuid, entry)

    @logcall_log_instance
    def snapshot(self, context, instance, image_id, update_task_state):
        """
        Snapshots the specified instance.

        :param context: security context
        :param instance: Instance object as returned by DB layer.
        :param image_id: Reference to a pre-created image that will
                         hold the snapshot.
        :param update_task_state: Function reference that allows for updates
                                  to the instance task state.
        """
        mesg = (_('Taking Snapshot'))
        self.log_operation(mesg, instance)
        host_name = instance['host']
        host = self.host_list.get_list().get(host_name)
        image = Image()
        image_entry = ImageEntry(context, image_id)

        def get_lpar_state():
            """
            returns the state of the lpar based on constants contained
            in nova.compute.power_state
            :param host: Host object for retrieving host information
            :param instance: Instance object as returned by DB layer.
            """
            state = None
            with K2OperGuard(host=host) as oper:
                lpar_name = instance['name']
                lpar_uuid = host.pvm_uuid_cache.get_pvm_uuid(lpar_name)
                lpar_resp = oper.read(k2const.MGT_SYS,
                                      rootId=host.uuid,
                                      childType=k2const.LPAR,
                                      childId=lpar_uuid,
                                      age=0)
                lpar_wrapper = LPAREntryWrapper(lpar_resp.entry)
                state = lpar_wrapper.get_lpar_state()
            return state

        def remove_opticals():
            optical = VirtualOptical()
            # forcing the timed cleanup that would normally happen after 1 hour
            optical.timed_cleanup(host, [instance['uuid']])

        image.capture(context, instance, image_entry, update_task_state,
                      get_lpar_state, remove_opticals)

    @logcall
    def power_off(self, instance, timeout=0, retry_interval=0):
        """
        Power off the specified instance.

        :param instance: Instance object as returned by DB layer.
        """
        try:
            mesg = (_('Powering off'))
            self.log_operation(mesg, instance)
            name, host, uuid, entry = (
                self._resolve_instance(instance, force_refresh=True))
            if entry is None:
                LOG.error(_('Virtual machine cannot be found. The '
                            'operation cannot proceed.'
                            'LPAR name: %(lpar_name)s. '
                            'LPAR UUID: %(lpar_uuid)s.') %
                          {'lpar_name': name, 'lpar_uuid': uuid})
                raise pvm_exc.IBMPowerVMLPARInstanceNotFound(
                    instance_name=name)
            lpar_wrapper = LPAREntryWrapper(entry)
            lpar_state = lpar_wrapper.get_lpar_state()
            if lpar_state is not power_state.SHUTDOWN:
                with K2OperGuard(host=host) as opr:
                    pvm_lpar.\
                        power_on_off(opr, uuid,
                                     k2const.SUFFIX_PARM_POWER_OFF,
                                     host, entry,
                                     timeout=CONF.ibmpowervm_power_off_timeout)
                    context = ctx.get_admin_context()
                    self.reset_inst_rmc_state(context, instance)
        except Exception:
            with excutils.save_and_reraise_exception():
                context = ctx.get_admin_context()
                # set instance to error state
                instance.vm_state = vm_states.ERROR
                instance.task_state = None
                instance.save()

    @logcall
    def power_on(self, context, instance, network_info,
                 block_device_info=None):
        """
        Power on the specified instance.

        :param context: security context
        :param instance: Instance object as returned by DB layer.
        :param network_info: not used
        :param block_device_info: not used
        """
        try:
            if context is None:
                context = ctx.get_admin_context()
            mesg = (_('Powering on'))
            self.log_operation(mesg, instance)
            name, host, uuid, entry = (
                self._resolve_instance(instance, force_refresh=True))
            if entry is None:
                LOG.error(_('Virtual machine cannot be found. The '
                            'operation cannot proceed.'
                            'LPAR name: %(lpar_name)s. '
                            'LPAR UUID: %(lpar_uuid)s.') %
                          {'lpar_name': name, 'lpar_uuid': uuid})
                raise pvm_exc.IBMPowerVMLPARInstanceNotFound(
                    instance_name=name)
            lpar_wrapper = LPAREntryWrapper(entry)
            lpar_state = lpar_wrapper.get_lpar_state()
            if lpar_state is not power_state.RUNNING:
                with K2OperGuard(host=host) as opr:
                    pvm_lpar.power_on_off(opr, uuid,
                                          k2const.SUFFIX_PARM_POWER_ON,
                                          host)
        except Exception:
            with excutils.save_and_reraise_exception():
                # set instance to error state
                instance.vm_state = vm_states.ERROR
                instance.task_state = None
                instance.save()

    @logcall
    def reboot(self, context, instance, network_info, reboot_type,
               block_device_info=None, bad_volumes_callback=None):
        """Reboot the specified instance.

        :param context: security context
        :param instance: Instance object as returned by DB layer.
        :param network_info: not used
        :param reboot_type: Either a HARD or SOFT reboot
        :param block_device_info: not used
        :param bad_volumes_callback: not used
        """
        try:
            if context is None:
                context = ctx.get_admin_context()
            mesg = (_('Rebooting'))
            self.log_operation(mesg, instance)
            name, host, uuid, entry = (
                self._resolve_instance(instance, force_refresh=True))
            if entry is None:
                LOG.error(_('Virtual machine cannot be found. The '
                            'operation cannot proceed.'
                            'LPAR name: %(lpar_name)s. '
                            'LPAR UUID: %(lpar_uuid)s.') %
                          {'lpar_name': name, 'lpar_uuid': uuid})
                raise pvm_exc.IBMPowerVMLPARInstanceNotFound(
                    instance_name=name)
            lpar_wrapper = LPAREntryWrapper(entry)
            lpar_state = lpar_wrapper.get_lpar_state()
            if lpar_state is not power_state.RUNNING:
                try:
                    # Attempt to power on the VM
                    with K2OperGuard(host=host) as opr:
                        pvm_lpar.power_on_off(opr, uuid,
                                              k2const.SUFFIX_PARM_POWER_ON,
                                              host)
                except Exception as exc:
                    LOG.exception(exc)
                    error = (_('Cannot restart virtual machine %(lpar_name)s. '
                               'The virtual machine is not in '
                               'active state.') % {'lpar_name': name})
                    LOG.error(error)
                    raise nova_exception.InstanceInvalidState(error)
            else:
                force = False
                if reboot_type == 'HARD':
                    force = True
                with K2OperGuard(host=host) as opr:
                    pvm_lpar.\
                        power_on_off(opr, uuid,
                                     k2const.SUFFIX_PARM_POWER_OFF,
                                     host, entry,
                                     force_immediate=force,
                                     restart=True,
                                     timeout=CONF.ibmpowervm_power_off_timeout)
            self.reset_inst_rmc_state(context, instance)
        except Exception as error:
            # Raise the error and nova.compute.manager will update
            # the vm to error state as needed
            raise nova_exception.InstanceRebootFailure(six.text_type(error))

    @logcall
    @return_cached_driver_data()
    def get_available_resource(self, nodename):
        """Retrieve resource information.

        This method is called when nova-compute launches, and
        as part of a periodic task

        :param nodename:
            node which the caller want to get resources from
            a driver that manages only one node can safely ignore this
        :returns: Dictionary describing resources
        """

        if nodename is None:
            # This only happens when we have just one node
            node = self.host_list.get_list().values()[0]
        else:
            node = self.host_list.get(nodename)
        data = node.get_host_stats(refresh=False)
        return data

    @logcall
    def get_host_uptime(self):
        """Returns the result of calling "uptime" on the target host."""
        raise NotImplementedError()

    @logcall
    def plug_vifs(self, instance, network_info):
        """Plug VIFs into networks."""
        host_name = instance['host']
        host = self.host_list.get_list().get(host_name)
        host.plug_vifs(instance, network_info)

    @logcall
    def unplug_vifs(self, instance, network_info):
        """Unplug VIFs from networks."""
        host_name = CONF.host
        host = self.host_list.get_list().get(host_name)
        host.unplug_vifs(instance, network_info)

    @logcall
    def get_host_cpu_stats(self):
        stats = dict()
        host_name = CONF.host
        node = self.host_list.get_list().get(host_name)

        if node is None:
            LOG.error(_('Host %s not found in hosts list.') % host_name)
            return stats

        context = ctx.get_admin_context()
        stats = node.get_host_utilization_data(context)
        if stats is None:
            return None

        # get one of the host's VIOSes vios id; use first one as default
        hostfact = compute_dom.ManagedHostFactory.get_factory()
        managed_host = hostfact.find_host_by_name(context, host_name)

        if managed_host is not None:
            all_vioses = managed_host.vioservers
        else:
            LOG.warning(_('Host %s not found by factory.') % host_name)
            return stats

        if not all_vioses:
            LOG.warning(_('The host %s has an empty VIOS list.') % host_name)
            return stats

        frequency = 0
        for vios in all_vioses:
            frequency = node.get_cpu_frequency(context, vios.lpar_id)
            if frequency != 0:
                # Got the frequency from this specific VIOS, so break
                break
        else:
            LOG.warning(_('CPU frequency for host %s could not be obtained '
                          'from any of its VIOSes; a value of 0 MHz will be '
                          'used for the CPU frequency.') % host_name)

        stats['frequency'] = frequency
        return stats

    @logcall
    def sync_enterprise_pool_for_host(self, context):
        """Syncs enterprise pool info corresponding to this host."""
        node = self.host_list.get_list().get(CONF.host)
        if node is None:
            raise Exception(_('Host %s not found in hosts list.') % CONF.host)
        return hmc_ep.sync_pool_for_host(context, node)

    @logcall
    def set_mobile_cores(self, context, desired_mobile_cores):
        node = self.host_list.get_list().get(CONF.host)
        if node is None:
            raise Exception(_('Host %s not found in hosts list.') % CONF.host)
        return hmc_ep.set_mobile_cores(context, desired_mobile_cores, node)

    @logcall
    def get_host_ip_addr(self):
        """Retrieves the IP address of the hypervisor host."""
        # TODO kps: need to resolve the correct host rather than
        # defaulting to the first
        host = self.host_list.get_list().values()[0]
        # use K2 level caching
        with K2OperGuard(host=host) as k2oper:
            mgt_sys_resp = k2oper.read(k2const.MGT_SYS, rootId=host.uuid)
        host_wrapper = ManagedSystemEntryWrapper(mgt_sys_resp.entry)
        ip_addr = host_wrapper.get_host_ip_address()
        return ip_addr

    @logcall
    def get_available_nodes(self):
        """Returns nodenames of all nodes managed by the compute service.

        This method is for multi compute-nodes support. If a driver supports
        multi compute-nodes, this method returns a list of nodenames managed
        by the service. Otherwise, this method should return
        [hypervisor_hostname].
        """
        nodes = []
        for node in self.host_list.itervalues():
            nodes.append(node.mtm_serial)
        return nodes

    def legacy_nwinfo(self):
        """
        Indicate if the driver requires the legacy network_info format.
        """
        return False

    # Do not use logcall because passwords will be printed to the logs
    @send_ui_notification(error_message=notify_msg.LPM_ERROR_DEST,
                          success_message=None,
                          log_exception=True)
    def check_can_live_migrate_destination(self, ctxt, instance_ref,
                                           src_compute_info, dst_compute_info,
                                           block_migration=False,
                                           disk_over_commit=False):
        """
        Validate the destination host is capable of live partition migration.

        :param ctxt: security context
        :param instance_ref: instance to be migrated
        :param src_compute_info: source host information
        :param dst_compute_info: destination host information
        :param block_migration: if true, prepare for block migration
        :param disk_over_commit: if true, allow disk over commit
        :returns dest_check_data: dictionary containing destination data
        """
        # Do not use logcall because the enter/exit parms contain a password
        LOG.debug("Entering check_can_live_migrate_destination/driver.py: "
                  "ctxt=%s, instance_ref=%s, src_compute_info=%s, "
                  "dst_compute_info=%s, block_migration=%s, "
                  "disk_over_commit=%s" %
                  (ctxt, instance_ref,
                   src_compute_info,
                   dst_compute_info,
                   block_migration,
                   disk_over_commit))

        tgt_mtms_name = dst_compute_info.get('hypervisor_hostname')
        tgt_sys_name = CONF.host_display_name
        host = self.host_list.get_list().get(tgt_mtms_name)
        inst_uuid = instance_ref['uuid']
        inst_name = instance_ref['display_name']

        # Print an informational trace message to the logs
        msg = (_('Validating destination host %(host)s for live migration')
               % {'host': tgt_sys_name})
        self.log_operation(msg, instance_ref)

        # Verify the instance is not already migrating.  Also verify the host
        # is not already participating in migration as a source.
        for key in self.migrate:
            if key == inst_uuid:
                error = (_("Migration of %(lpar)s is already in progress.")
                         % {'lpar': inst_name})
                LOG.exception(error)
                raise pvm_exc.IBMPowerVMMigrationInProgress(lpar=inst_name)
            migrating_src_host = self.migrate[key].lpar.get('host')
            migrating_tgt_host = self.migrate[key].host.host_id
            if migrating_src_host == migrating_tgt_host:
                error = (_("Cannot migrate %(inst)s to %(host)s because that "
                           "host is currently serving as the source in a "
                           "separate migration operation for virtual machine "
                           "%(vm)s.")
                         % {'inst': inst_name,
                            'host': tgt_sys_name,
                            'vm': self.migrate[key].lpar.get('display_name')})
                raise nova_exception.MigrationPreCheckError(reason=error)

        try:
            # This is the entry point for migration on the destination thread.
            # Get a new migrate object for the target.
            self.migrate[inst_uuid] = pvm_migrate.Migrate(instance_ref, host)

            dest_check_data = self.migrate[inst_uuid]. \
                check_can_live_migrate_destination()
        except Exception:
            with excutils.save_and_reraise_exception():
                self.migration_cleanup(ctxt, instance_ref)

        # Identify the destination management console in Neo/HMC migration
        dst_managed_by = migrate_utils.get_dest_managed_by(dst_compute_info)
        dest_check_data.managed_by = dst_managed_by
        # Do not use logcall because the enter/exit parms contain a password
        logged_dest_data = copy.deepcopy(dest_check_data)
        logged_dest_data.dest_cred = '********'
        LOG.debug("Exiting check_can_live_migrate_destination/driver.py: "
                  "dest_check_data=%s" % logged_dest_data)
        return dest_check_data

    # Do not use logcall because passwords will be printed to the logs
    @send_ui_notification(error_message=notify_msg.LPM_ERROR,
                          success_message=None, migrate_op=True,
                          log_exception=True)
    def check_can_live_migrate_source(self, ctxt, instance_ref,
                                      dest_check_data, block_device_info=None):
        """
        Validate the source host is capable of live partition migration.

        :param context: security context
        :param instance_ref: instance to be migrated
        :param dest_check_data: results from check_can_live_migrate_destination
        :returns migrate_data: dictionary containing source and
            destination data for migration
        """
        # Do not use logcall because the enter/exit parms contain a password
        logged_dest_data = copy.deepcopy(dest_check_data)
        logged_dest_data.dest_cred = '********'
        LOG.debug("Entering check_can_live_migrate_source/driver.py: "
                  "ctxt=%s, instance_ref=%s, "
                  "dest_check_data=%s" % (ctxt, instance_ref,
                                          logged_dest_data))

        src_mtms_name = instance_ref['host']
        src_sys_name = CONF.host_display_name
        inst_uuid = instance_ref['uuid']
        inst_name = instance_ref['display_name']
        k2_name, host, k2_uuid, lpar_entry = \
            self._resolve_instance(instance_ref)

        # Print an informational trace message to the logs
        msg = (_('Validating source host %(host)s for live migration')
               % {'host': src_sys_name})
        self.log_operation(msg, instance_ref)

        # See issue 8707 - There are rare cases (e.g. RPC timeout)
        # where the target thread gets cleaned up, but the source doesn't.
        # If the key still exists in the source migrate dictionary at this
        # point we'll know it didn't exist in the target dictionary.  We can
        # safely clean it up knowing it wasn't cleaned up due to a previous
        # failure.
        if inst_uuid in self.migrate:
            LOG.debug("An entry exists in the migrate dictionary for %s "
                      "on the source thread. This wasn't cleaned up on "
                      "a previous migration failure. Cleaning up now."
                      % inst_name)
            self.migration_cleanup(ctxt, instance_ref, source=True)

        # Verify the host is not already participating
        # in migration as a target.
        for key in self.migrate:
            migrating_src_host = self.migrate[key].lpar.get('host')
            if src_mtms_name != migrating_src_host:
                error = (_("Cannot migrate %(inst)s from %(host)s because "
                           "that host is currently serving as the target "
                           "in a separate migration operation for "
                           "virtual machine %(vm)s.")
                         % {'inst': inst_name,
                            'host': src_sys_name,
                            'vm': self.migrate[key].lpar.get('display_name')})
                raise nova_exception.MigrationPreCheckError(reason=error)

        try:
            # This is the entry point for migration on the source thread.
            # Get a new migrate object for the source
            self.migrate[inst_uuid] = pvm_migrate.Migrate(instance_ref, host,
                                                          lpar_entry, k2_uuid)
            self.migrate[inst_uuid].host.source_host_for_migration = True

            migrate_data = self.migrate[inst_uuid]. \
                check_can_live_migrate_source(ctxt, dest_check_data)

            self._update_compute_node_stats_source_host_for_migration(
                ctxt, self.migrate[inst_uuid].host.host_id, True)
        except Exception:
            with excutils.save_and_reraise_exception():
                self.migration_cleanup(ctxt, instance_ref, source=True)

        # Do not use logcall because the enter/exit parms contain a password
        logged_mig_data = copy.deepcopy(migrate_data)
        logged_mig_data.dest_cred = '********'
        LOG.debug("Exiting check_can_live_migrate_source/driver.py: "
                  "migrate_data=%s" % logged_mig_data)
        return migrate_data

    def _update_compute_node_stats_source_host_for_migration(
            self, context, host, source_host_for_migration):
        '''
        Update the compute node in the DB with a stats entry of
        source_host_for_migration set to True immediately instead of waiting
        for a periodic task to pick up the change and send out a notification
        that the update was done so that the scheduler can be aware of this and
        update its attributes immediately.

        :param context: Nova context object
        :param host: String representing the host name
        :param source_host_for_migration: Boolean value to set the field to
        '''
        compute_nodes = (objects.ComputeNodeList.get_all_by_host(
            context, host) or [])

        if not compute_nodes:
            msg = (_('Compute node not found for host %s.') % host)
            LOG.error(msg)
            raise nova_exception.NotFound(message=msg)

        node = compute_nodes[0]
        stats = copy.deepcopy(node.stats)
        stats[pvc_consts.SOURCE_HOST_FOR_MIGRATION] = source_host_for_migration
        node.stats = stats
        node.save()

        # Send out a notification that the node was updated
        notifier = rpc.get_notifier('compute', host)
        event_type = 'compute.node.update.end'

        service = objects.Service.get_by_compute_host(context, node.host)

        info = {'host': host,
                'service_id': service.id,
                'compute_node_id': node.id,
                'hypervisor_hostname': node.hypervisor_hostname,
                'stats': node.stats}
        notifier.info(context, event_type, info)

    def check_can_live_migrate_destination_cleanup(self, ctxt,
                                                   dest_check_data):
        """
        Performs any necessary cleanup on the destination
        host prior to live partition migration.

        :param ctxt: security context
        :param dest_check_data: results from check_can_live_migrate_destination
        """
        pass

    @logcall
    @send_ui_notification(error_message=notify_msg.LPM_ERROR,
                          success_message=None, migrate_op=True,
                          log_exception=True)
    def pre_live_migration(self, context, instance, block_device_info,
                           network_info, disk_info, migrate_data=None):
        """
        Perfoms any required prerequisites on the destination
        host prior to live partition migration.

        :param context: security context
        :param instance: instance to be migrated
        :param block_device_info: instance block device information
        :param network_info: instance network information
        :param migrate_data: implementation specific data dictionary
        """
        inst_uuid = instance['uuid']

        # Print an informational trace message to the logs
        msg = (_('Preparing destination host %(host)s for live migration')
               % {'host': migrate_data.dest_sys_name})
        self.log_operation(msg, instance)

        bdms = blkdev_obj.BlockDeviceMappingList.\
            get_by_instance_uuid(context, instance['uuid'])
        return self.migrate[inst_uuid].pre_live_migration(context, instance,
                                                          bdms, network_info,
                                                          migrate_data)

    def ensure_filtering_rules_for_instance(self, instance, network_info,
                                            time_module=None):
        """
        Sets up networking filter rules for the destination host
        prior to live partition migration.

        :params instance_ref: nova.db.sqlalchemy.models.Instance object
        """
        pass

    @logcall
    @send_ui_notification(error_message=notify_msg.LPM_ERROR,
                          success_message=None, migrate_op=True,
                          log_exception=True)
    def live_migration(self, ctxt, instance_ref, dest,
                       post_method, recover_method,
                       block_migration=False, migrate_data=None):
        """
        Live migrates a partition from one host to another.

        :param ctxt: security context
        :param instance_ref: instance to be migrated.
        :param dest: destination host
        :param post_method: post operation method.
            nova.compute.manager.post_live_migration.
        :param recover_method: recovery method when any exception occurs.
            nova.compute.manager.recover_live_migration.
        :param block_migration: if true, migrate VM disk.
        :param migrate_data: implementation specific data dictionary.
        """

        src_sys_name = migrate_data.src_sys_name
        tgt_sys_name = migrate_data.dest_sys_name
        inst_name = instance_ref['display_name']
        inst_uuid = instance_ref['uuid']

        # Print an informational trace message to the logs
        msg = (_('Migrating virtual machine %(inst)s from host %(src)s '
                 'to host %(tgt)s.')
               % {'inst': inst_name,
                  'src': src_sys_name,
                  'tgt': tgt_sys_name})
        self.log_operation(msg, instance_ref)

        self.migrate[inst_uuid].live_migration(ctxt, instance_ref, dest,
                                               post_method, recover_method,
                                               block_migration,
                                               migrate_data)

    @logcall
    @send_ui_notification(error_message=notify_msg.LPM_ERROR,
                          success_message=None, migrate_op=True,
                          log_exception=True)
    def post_live_migration(self, ctxt, instance_ref, block_device_info,
                            migrate_data=None):
        """
        Performs post operations on the source host
        following a successful live migration.

        :param ctxt: security context
        :param instance_ref: migrated instance
        :param block_device_info: instance volume block device info
        :param migrate_data: implementation specific data dictionary.
        """
        # Print an informational trace message to the logs
        msg = (_('Cleaning up source host after successful live migration'))
        self.log_operation(msg, instance_ref)

        inst_uuid = instance_ref['uuid']

        if migrate_data is not None:
            # Unmap and remove all the vscsi volumes of an instance
            # from source host using the pre_migrate_bdms.
            try:
                bdmaps_source_host = migrate_data.pre_migrate_bdms
                self.migrate[inst_uuid].\
                    _unmap_vscsi_volumes_from_host(ctxt, instance_ref,
                                                   bdmaps_source_host)
                LOG.info(_("Removed volume connections from source host."))
            except Exception as ex:
                # No need to fail migration in this case.
                LOG.warn(_("Failed to remove volume connections "
                           "due to error: %s") %
                         six.text_type(ex))

        # Remove the migrate class object for the source host on a
        # successful migration.  On a failing migration it will
        # be cleaned up via the rollback path in the manager code.
        self.migration_cleanup(ctxt, instance_ref, source=True)

    @logcall
    @send_ui_notification(error_message=notify_msg.LPM_ERROR,
                          success_message=notify_msg.LPM_SUCCESS,
                          migrate_op=True, log_exception=True)
    def post_live_migration_at_destination(self, ctxt, instance_ref,
                                           network_info,
                                           block_migration=False,
                                           block_device_info=None,
                                           migrate_data=None):
        """
        Performs post operations on the destination host
        following a successful live migration.

        :param ctxt: security context
        :param instance_ref: migrated instance
        :param network_info: dictionary of network info for instance
        :param block_migration: boolean for block migration
        :param block_device_info: instance volume block device info
        :param migrate_data: implementation specific data dictionary.
        """
        # Print an informational trace message to the logs
        msg = (_('Cleaning up destination host after successful '
                 'live migration'))
        self.log_operation(msg, instance_ref)

        inst_name = instance_ref['display_name']
        inst_uuid = instance_ref['uuid']

        # Get the host mtms names to query k2 caches
        src_mtms_name = instance_ref['host']
        tgt_mtms_name = CONF.host
        try:
            k2_uuid = self._resolve_instance_k2_uuid_for_migrate(
                inst_uuid, src_mtms_name, tgt_mtms_name)
            if k2_uuid is None:
                error = (_("Virtual machine %(inst)s from %(host)s "
                           "cannot be found in the "
                           "database.") % {'inst': inst_name,
                                           'host': src_mtms_name})
                LOG.exception(error)
                raise pvm_exc.IBMPowerVMMigrationFailed(error)

            # Do post migration processing for the destination host
            if inst_uuid in self.migrate:
                self.migrate[inst_uuid].\
                    post_live_migration_at_destination(ctxt, k2_uuid)
            else:
                LOG.warn(_("Instance %s not found in the current migrate "
                           "list") % inst_uuid)
        except Exception:
            # Just save and raise the exception so the finally will run.
            # The notify decorator will handle the notify
            with excutils.save_and_reraise_exception():
                pass
        finally:
            # Remove the migrate class object for the destination host
            # on a successful migration
            self.migration_cleanup(ctxt, instance_ref)

    @logcall
    def attach_interface(self, instance, image_meta, vif):
        """
        Attach an interface to the instance.
        """
        # Create a thread to go off and do the attach async
        greenthread.spawn(self._async_attach_interface, instance, vif)

    @lockutils.synchronized('powervm_attach_ifc', 'nova-attachifc-powervm')
    def _async_attach_interface(self, instance, vif):
        '''
        This method will asynchronously do the attach of the network interface.
        It must be done asynchronously because all other platforms are very
        quick, but since PowerVM takes longer...if done sync'd the user can
        cause many issues in the REST API (perhaps even a timeout).

        We will use bumblebee to send messages on status.
        '''
        notifier = rpc.get_notifier(service='compute', host=CONF.host)
        napi = net_api.API()
        admin_context = ctx.get_admin_context()
        lpar_name, host, lpar_uuid, entry = self._resolve_instance(
            instance, force_refresh=True)

        # Notify the start of this.
        msg = (_("The network interface attach action has started on Host "
                 "%(host)s.  This operation might take some time to "
                 "complete.  Do not modify the network interface until a "
                 "notification has been sent that indicates operation "
                 "completion.") %
               {'host': host.host_display_name})
        info = {'msg': msg, 'host_name': CONF.host_display_name}
        notifier.info(admin_context, 'compute.instance.log', info)

        try:
            # First need to check to see whether or not we support the
            # attachment of this on the system, or if the network is set to
            # not allow it then we should give them the correct message.
            net_id = vif.get('network').get('id')
            nets = [[net_id, None, None]]
            viable_hosts = napi.get_viable_hosts(admin_context,
                                                 instance['uuid'],
                                                 requested_networks=nets)
            local_host_viable = False
            for viable_host in viable_hosts:
                if viable_host == host.host_id:
                    local_host_viable = True
                    break

            if not local_host_viable:
                msg = _("The system configuration does not allow "
                        "attaching this network to the host %(host)s.  "
                        "This could be because currently this network "
                        "has been configured to not allow attachments or "
                        "the RMC state is inactive on the VIOS. "
                        "To resolve this, update the network and "
                        "configure the the host's Shared Ethernet "
                        "Adapter. ") %\
                    {'host': host.host_display_name}
                LOG.error(msg)
                info = {'msg': msg, 'host_name': CONF.host_display_name}
                notifier.error(admin_context, 'compute.instance.log', info)
                raise nova_exception.InterfaceAttachFailed(
                    instance_uuid=instance['uuid'])

            # If RMC is down and instance is running, can't proceed
            # But for IBM i, there is no need to check RMC
            lpar = LPAREntryWrapper(entry)
            if ((lpar.get_lpar_type() != pvc_consts.IBMI_OS_TYPE and
                 lpar.get_rmc_state().lower() != 'active' and
                 lpar.get_lpar_state() != power_state.SHUTDOWN)):
                msg = _('Cannot add a virtual Ethernet adapter on virtual '
                        'machine %(lpar_name)s because the RMC state is '
                        '%(rmc_state)s.  Ensure that the RMC is active on the '
                        'virtual machine, or shut down the virtual machine '
                        'before attempting to add a virtual Ethernet '
                        'adapter.') %\
                    {'lpar_name': lpar_name, 'rmc_state': lpar.get_rmc_state()}
                LOG.error(msg)
                info = {'msg': msg, 'host_name': CONF.host_display_name}
                notifier.error(admin_context, 'compute.instance.log', info)
                raise nova_exception.InterfaceAttachFailed(
                    instance_uuid=instance['uuid'])

            try:

                # If PowerVC created this port, do not preserve on delete
                port = napi.show_port(admin_context, vif.get('id'))
                name = port['port'].get('name', '')
                if name.startswith('powervc_attached'):
                    # Clear the preserve flag
                    vif['preserve_on_delete'] = False

                    # Check to see if any other VIFs are on this instance
                    cur_vifs = napi.get_instance_nw_info(admin_context,
                                                         instance)
                    if not isinstance(cur_vifs, list):
                        cur_vifs = []

                    # Remove / readd the VIF we're updating
                    cur_vifs[:] = [d for d in cur_vifs if d.get('id') !=
                                   vif.get('id')]
                    cur_vifs.append(vif)

                    # Convert VIF list to NetworkInfos
                    net_infos = cur_vifs
                    if not isinstance(net_infos, net_model.NetworkInfo):
                        net_infos = net_model.NetworkInfo(cur_vifs)

                    # Update VIF list
                    base_api.update_instance_cache_with_nw_info(
                        napi, admin_context, instance, net_infos)

                host.plug_vifs(instance, [vif])
                pvm_lpar.attach_interfaces(admin_context,
                                           host,
                                           None,  # pvm_lpar gets host_stats
                                           entry,
                                           instance,
                                           [vif],
                                           False)
            except K2HttpError as e:
                msg = _('Cannot create a virtual Ethernet adapter on the '
                        'virtual machine %(lpar_name)s because the HMC '
                        'operation for create failed with an HTTP error. '
                        'Refer to the logs for more information before '
                        'you attempt to add the virtual Ethernet adapter '
                        'from the UI.') % {'lpar_name': lpar_name}
                LOG.error(msg)
                info = {'msg': msg, 'host_name': CONF.host_display_name}
                notifier.error(admin_context, 'compute.instance.log', info)
                raise nova_exception.InterfaceAttachFailed(
                    instance_uuid=instance['uuid'])
            except IBMPowerMACNotSupported as e:
                # Failed on P6, send to bumblebee
                notifier.error(admin_context, 'compute.instance.log',
                               {'msg': e.format_message()})
                raise nova_exception.InterfaceAttachFailed(
                    instance_uuid=instance['uuid'])
            # Notify the successful operation
            msg = (_("The network interface attach action has completed for "
                     "Host %(host)s.  Manual configuration of the interface "
                     "is required on the system %(lpar)s.") %
                   {'host': host.host_display_name, 'lpar': lpar_name})
            info = {'msg': msg, 'host_name': CONF.host_display_name,
                    'instance_id': instance['uuid']}
            notifier.info(admin_context, 'compute.instance.log', info)
        except:
            napi.deallocate_port_for_instance(admin_context, instance,
                                              vif.get('id'))
            with excutils.save_and_reraise_exception():
                pass

    @logcall
    def detach_interface(self, instance, vif):
        """
        Detach an interface from the instance.
        """
        notifier = rpc.get_notifier(service='compute', host=CONF.host)
        lpar_name, host, lpar_uuid, entry = self._resolve_instance(instance)
        admin_context = ctx.get_admin_context()

        # Notify the start of this.
        msg = (_("The network interface detach action has started on Host "
                 "%(host)s.  This operation may take some time to complete.  "
                 "A subsequent message will be displayed when the operation "
                 "completes.") %
               {'host': host.host_display_name})
        info = {'msg': msg, 'host_name': CONF.host_display_name}
        notifier.info(admin_context, 'compute.instance.log', info)

        # If RMC is down and instance is running, can't proceed
        # But for IBM i, there is no need to check RMC
        lpar = LPAREntryWrapper(entry)
        if ((lpar.get_lpar_type() != pvc_consts.IBMI_OS_TYPE and
             lpar.get_rmc_state().lower() != 'active' and
             lpar.get_lpar_state() != power_state.SHUTDOWN)):
            msg = _('Cannot delete a virtual Ethernet adapter on virtual '
                    'machine %(lpar_name)s because the RMC state is '
                    '%(rmc_state)s.  Ensure that the RMC is active on the '
                    'virtual machine, or shut down the virtual machine '
                    'before attempting to delete a virtual Ethernet '
                    'adapter.') % {'lpar_name': lpar_name,
                                   'rmc_state': lpar.get_rmc_state()}
            LOG.error(msg)
            info = {'msg': msg, 'host_name': CONF.host_display_name}
            notifier.error(admin_context, 'compute.instance.log', info)
            raise nova_exception.InterfaceDetachFailed(
                instance_uuid=instance['uuid'])

        try:
            pvm_lpar.detach_interfaces(ctx.get_admin_context(),
                                       host,
                                       entry,
                                       instance,
                                       [vif])
            host.unplug_vifs(instance, [vif])
        except K2HttpError:
            msg = _('Cannot delete a virtual Ethernet adapter on the '
                    'virtual machine %(lpar_name)s because the HMC '
                    'operation for delete failed with an HTTP error. '
                    'Refer to the logs for more information and '
                    'ensure that the virtual Ethernet adapter is '
                    'manually removed after you attempt to delete '
                    'it from the UI.') % {'lpar_name': lpar_name}
            LOG.error(msg)
            info = {'msg': msg, 'host_name': CONF.host_display_name}
            notifier.error(admin_context, 'compute.instance.log', info)
            raise nova_exception.InterfaceDetachFailed(
                instance_uuid=instance['uuid'])

        msg = (_("The network interface detach action has completed on Host "
                 "%(host)s.") %
               {'host': host.host_display_name})
        info = {'msg': msg, 'host_name': CONF.host_display_name}
        notifier.info(admin_context, 'compute.instance.log', info)

    @logcall
    @send_ui_notification(error_message=notify_msg.LPM_ERROR_DEST,
                          success_message=None,
                          log_exception=True)
    def rollback_live_migration_at_destination(self, context, instance,
                                               network_info,
                                               block_device_info,
                                               destroy_disks=True,
                                               migrate_data=None):
        """
        Performs post operations on the destination host
        following a failing live migration.

        :param context: security context
        :param instance: failed migrated instance
        :param network_info: dictionary of network info for instance
        :param block_device_info: instance volume block device info
        """
        # Print an informational trace message to the logs
        msg = (_('Cleaning up destination host after failed live migration'))
        self.log_operation(msg, instance)

        inst_uuid = instance['uuid']

        # Get the host mtms names to query k2 caches
        src_mtms_name = instance['host']
        tgt_mtms_name = CONF.host

        # Unmap and remove all the vscsi volumes of an instance
        # from target host.
        try:
            bdms = blkdev_obj.BlockDeviceMappingList.\
                get_by_instance_uuid(context, inst_uuid)
            self.migrate[inst_uuid].\
                _unmap_vscsi_volumes_from_host(context, instance, bdms)
            LOG.info(_("Removed volume connections from destination host."))
        except Exception as ex:
            # No need to throw exception in this case.
            LOG.warn(_("Failed to remove volume connections "
                       "due to error: %s") % six.text_type(ex))

        try:
            k2_uuid = self._resolve_instance_k2_uuid_for_migrate(
                inst_uuid, src_mtms_name, tgt_mtms_name)
            if k2_uuid is not None:
                self.migrate[inst_uuid]. \
                    rollback_live_migration_at_destination(k2_uuid)
        except Exception:
            # Just save and raise the exception so the finally will run.
            # The notify decorator will handle the notify
            with excutils.save_and_reraise_exception():
                pass
        finally:
            # Remove the migrate class object for the destination host
            # on a failing migration
            self.migration_cleanup(context, instance)

    @logcall
    def _resolve_instance_k2_uuid_for_migrate(self, lpar_uuid,
                                              source_host_name,
                                              target_host_name):
        """
        Gets the k2 uuid from the DB for an instance given the instance uuid
        and host. Will attempt to use the source host first, if not found
        there, the lpar may already be on target or in transition (hence the
        retry loop)

        :param lpar_uuid: openstack uuid
        :param host_name: name of source host
        :param host_name: name of target host
        :return: uuid if found else None
        """
        k2_uuid = None
        host_name = source_host_name
        for count in range(10):
            while k2_uuid is None:
                # Retrieve the Instances from the Database for given host
                instfact = compute_dom.ManagedInstanceFactory.get_factory()
                db_instances = instfact.find_all_instances_by_host(
                    ctx.get_admin_context(), host_name)
                for db_instance in db_instances:
                    if lpar_uuid == db_instance.get('uuid'):
                        k2_uuid = db_instance.get(
                            'power_specs', {}).get('vm_uuid')
                        break
                # If we just tried the source
                if host_name == source_host_name:
                    # Try target (if k2_uuid still None)
                    host_name = target_host_name
                else:
                    host_name = source_host_name
                    break

            if k2_uuid is None:
                # Progressively sleep a little longer
                greenthread.sleep(count)
            else:
                break

        return k2_uuid

    @logcall
    def migration_cleanup(self, context, instance, source=False):
        """
        Update the migration object dictionary. If this is the source migration
        dictionary, and we're removing the last migrate object, we can clear
        the 'source_host_for_migration' flag.

        Note: We call this from the manager level when migration validation
        fails on the source to clean up the target allowing
        another migration to this lpar

        :param context: Nova context object
        :param instance: the instance reference
        :param source: boolean indicating we're cleaning the source dict
        """
        inst_uuid = instance['uuid']
        inst_name = instance['display_name']

        if inst_uuid in self.migrate:

            if source:
                LOG.debug("Removing source host migration object "
                          "for %s" % inst_name)

                # Clear source host flag on host object
                if len(self.migrate) == 1:
                    LOG.debug("%s is the last lpar migrating from host %s. "
                              "Clearing source_host_for_migration flag." %
                              (inst_name,
                               self.migrate[inst_uuid].host_name))
                    self.migrate[inst_uuid].host. \
                        source_host_for_migration = False

                    self._update_compute_node_stats_source_host_for_migration(
                        context, self.migrate[inst_uuid].host.host_id, False)
            else:
                LOG.debug("Removing target host migration object "
                          "for %s" % inst_name)

            # Remove host migration object for instance
            self.migrate.pop(inst_uuid)

    def remove_volume_device_at_destination(self, context, instance, cinfo):
        """ clean up volume device at destination during Rollback

        :param context: security context
        :param instance: instance dictionary of failed migration
        :param cinfo: block_disk_map['connection_info']

        Called from manager.py, and needed only for IVM migration.
        """
        pass

    def unfilter_instance(self, instance, network_info):
        """
        Stop filtering instance.

        :param instance: dictionary of instance info
        :param network_info: dictionary of network info for instance
        """
        pass

    def discover_instances(self, context):
        """
        Returns a list of all of the VM's that exist on the given Host.
        For each VM the driver needs to return a dictionary containing
        the following attributes:
            name:      The Name of the VM defined on the Hypervisor
            state:     The State of the VM, matching the power_state definition
            uuid:      The UUID of the VM when created thru OS (Optional)
            hostuuid:  The Hypervisor-specific UUID of the VM (Optional)
            supported: Dictionary stating whether the VM can be managed
              status:  Whether or not it is supported
              reasons: List of Dictionaries as to why it isn't supported
                id:      Message ID to key off of for the reason
                message: Message Text as to why it isn't supported

        :param context:   The security context for the query
        """
        disc_insts = []
        for node in self.host_list.itervalues():
            importvm = ImportVM(node)
            info = importvm.list_lpar_for_onboard(context)
            disc_insts += info
        return disc_insts

    def query_instances(self, context, instances):
        """
        Returns a list of VM's (matching those specified on input) with
        enough additional details about each VM to be able to import the
        VM's into the Nova Database such as OpenStack can start managing.
        For each VM the driver needs to return a dictionary containing
        the following attributes:
            uuid:      The UUID of the VM when created thru OS
            name:      The Name of the VM defined on the Hypervisor
            state:     The State of the VM, matching the power_state definition
            hostuuid:  The Hypervisor-specific UUID of the VM (Optional)
            vcpus:     The number of Virtual CPU's currently allocated
            memory_mb: The amount of Memory currently allocated
            root_gb:   The amount of storage currently allocated
            ephemeral_gb:  The amount of ephemeral storage currently allocated
            flavor:    Dictionary containing the profile/desired allocations
              vcpus:     The number of Virtual CPU's defined for allocation
              memory_mb: The amount of Memory defined for allocation
              root_gb:   The amount of storage defined for allocation
              ephemeral_gb:  The ephemeral storage defined for allocation
              powervm:proc_units:  The number of proc_units defined (Optional)
            volumes:   List of dictionary objects with the attached volumes:
              uuid:      The UUID of the Volume when created thru OS (Optional)
              name:      The Name of the Volume defined on the Backend
              host:      The registered name of the Storage Provider
              provider_location:  The Volume ID known on the Backend (Optional)
            network_ports: List of dictionary objects with the network ports:
              mac_address:  The MAC Address of the Port allocated to the VM
              status:       The Status of the Port, matching the definition
              segmentation_id: The VLAN ID for the Network the Port is part of
              physical_network: The Physical Network the Port is part of

        :param context:   The security context for the query
        :param instances: A list of dictionary objects for each VM containing:
            uuid:      The UUID of the VM when created thru OS
            name:      The Name of the VM defined on the Hypervisor
            hostuuid:  The Hypervisor-specific UUID of the VM (Optional)
        """
        disc_insts = []
        for node in self.host_list.itervalues():
            importvm = ImportVM(node)
            info = importvm.query_onboarding_instances(context, instances)
            disc_insts += info
        return disc_insts

    def inventory_instances(self, context, instances):
        """
        Provides a mechanism for the Driver to gather Inventory-related
        information for the Instances provided off of the Hypervisor at
        periodic intervals.  The Driver is free from there to populate
        the information directly in the Database rather than return it.

        :param context: The security context for the query
        :param instances: A list of dictionary objects for each VM containing:
            uuid:      The UUID of the VM when created thru OS
            name:      The Name of the VM defined on the Hypervisor
            hostuuid:  The Hypervisor-specific UUID of the VM
        """
        # First we will update the UUID/Name Map to pick up Imported LPAR's
        for node in self.host_list.itervalues():
            node.refresh_UUID_cache(instances)

        host_name = CONF.host
        host = self.host_list.get_list().get(host_name)
        # defect 21496
        # performance enhancement:
        # Perform 1 GET of lpar feed, and re-use the feed to resolve instances
        lpars_resp = None
        shared_proc_pools = {}
        try:
            with K2OperGuard(host=host) as k2oper:
                lpars_resp = k2oper.read(k2const.MGT_SYS, rootId=host.uuid,
                                         childType=k2const.LPAR)
                shared_proc_pools = host.get_shared_proc_pools_id_name()
        except Exception as e:
            # just log the exception
            LOG.exception(e)

        # query the vios to get the latest slot information
        # for each lpar
        vios_entries = None
        try:
            with K2OperGuard(host=host) as k2oper:
                vioses_resp = k2oper.read(k2const.MGT_SYS,
                                          rootId=host.uuid,
                                          childType=k2const.VIOS,
                                          xag=k2const.ALL_VIOS_EXT_PROPS)
                vios_entries = vioses_resp.feed.entries
        except Exception as e:
            #just log the exception
            LOG.exception(e)

        # Loop through each of the Instances getting the PowerSpecs for them
        for instance in instances:
            try:
                name, node, uuid, entry = self._resolve_instance(
                    instance, lpars_resp=lpars_resp, log_notfound=False)
                if entry is not None:
                    # If current task state is waiting on host, it needs to be
                    # cleared, since an LPAR entry was successfully obtained
                    if instance['task_state'] == WAITING_ON_HOST:
                        self._queue_event(
                            hmcevent.LifecycleEvent(instance['uuid'],
                                                    hmcevent.EVENT_CLEAR_TASK))
                    # Call the Wrapper to retrieve the PowerSpecs for the LPAR
                    power_spec = pvm_lpar.get_power_specs(entry)
                    # add the pool name info
                    proc_pool_id = power_spec.get('shared_proc_pool_id')
                    if proc_pool_id is not None:
                        power_spec['shared_proc_pool_name'] = \
                            shared_proc_pools.get(proc_pool_id)
                    else:
                        power_spec['shared_proc_pool_name'] = None
                        power_spec['shared_proc_pool_id'] = 0
                    # Call the Common Utility to update PowerSpecs in the DB
                    power_spec_utils.update_instance_power_specs(
                        context, self.virtapi, instance, power_spec.copy())
                    lpar_id = LPAREntryWrapper(entry).get_lpar_id()
                    #check if the lpar slot and wwpn info needs to be updated
                    self._update_lpar_slot_info(context, instance,
                                                vios_entries, lpar_id)
            # If an error occurred, just log it and move on to the next one
            except Exception as exc:
                LOG.warn(_('Error retrieving PowerSpecs for %s') %
                         instance['name'])
                LOG.exception(exc)

    def _update_lpar_slot_info(self, context, instance, vios_entries, lpar_id):
        """
        For instance which are upgraded or managed the lpar slot information
        needs to be updated.
        """
        fc_map = fc_map_processor.FCMapProcessor()

        sys_metadata = instance["system_metadata"]
        lpar_slot_found = False
        for key in sys_metadata:
            if "vios_" in key:
                lpar_slot_found = True
                return

        LOG.info("lpar slot data found %(found)s for %(instance)s" %
                 dict(found=lpar_slot_found, instance=instance["name"]))

        if vios_entries:
            for vios_entry in vios_entries:
                vios_wrapper = VIOSEntryWrapper(vios_entry)
                vios_name = vios_wrapper.get_vios_name()
                # Gather the VFC mapping data needed to construct the
                # VirtualFCMappings parameter.
                fc_client_map_data = fc_map.extract_client_adapter_data(
                    vios_entry, lpar_id)
                if fc_client_map_data is not None:
                    LOG.debug("FC client mapping data for vios %(vios)s is"
                              " %(data)s." % {'vios': vios_name,
                                              'data': fc_client_map_data})
                    migrate_utils.set_lpar_slot_info_from_fc_data(
                        instance, fc_client_map_data)

            sys_meta = instance.get("system_metadata")
            new_values = {key: sys_meta[key] for key in sys_meta
                          if migrate_utils.VIOS_KEY in key}
            managed_inst = instance
            if not isinstance(instance, compute_dom.ManagedInstance):
                inst_fact = compute_dom.ManagedInstanceFactory.get_factory()
                managed_inst = inst_fact.find_instance_by_uuid(
                    context, instance['uuid'])
            managed_inst.system_metadata.update(new_values)
            managed_inst.save()

    def inventory_host(self, context):
        """
        Provides a mechanism for the Driver to gather Inventory-related
        information for the Host at periodic intervals.  The Driver is
        free from there to populate the information directly in the
        Database and should not return any information from the method.

        :param context: The security context for the query
        """
        topo_class = CONF.powervm_vif_topo_driver
        host = self.host_list.get_list().get(CONF.host)
        with K2OperGuard(host=host) as k2opr:
            topo_drv = importutils.import_object(topo_class,
                                                 host_name=CONF.host,
                                                 operator=k2opr)
        network_topo, self.network_topo = (self.network_topo, None)
        # Collect the Network Topology Information and update it in the DB
        try:
            LOG.info(_('Collecting Host-Network Topology...'))
            # We retrieved the Network Topology on init_host to verify the
            # configuration is valid, so we will use that one if it is cached
            if network_topo is None:
                network_topo = topo_drv.get_host_network_topology(context)
            # Call save on the Host Network Topology to update it in the DB
            network_topo.save()

            self._cleanup_neutron_ports(context)

            LOG.info(_('Done Collecting Host-Network Topology.'))
        except Exception as e:
            # Just log the exception and continue on if Network Topology failed
            LOG.error(_("Reconciling Host-Network Topology failed."))
            LOG.exception(e)

        # Collect the Storage Topology Information and update it in the DB
        try:
            LOG.info(_('Collecting Host-Storage Topology...'))
            topofact = storage_dom.HostStorageTopologyFactory.get_factory()
            stg_topo = topofact.construct_storage_topology(context,
                                                           CONF.host)
            topo_helper = stg_topo.construct_hmc_helper(CONF.host)
            topo_helper.reconcile_topology()
            LOG.info(_('Done Collecting Host-Storage Topology.'))
        except Exception as e:
            # Just log the exception and continue on if Storage Topology failed
            LOG.error(_("Reconciling Host-Storage Topology failed."))
            LOG.exception(e)

        # Collect the Shared processor pool information of the host
        try:
            LOG.info(_('Collecting Host-Shared proc pool information...'))
            # Updates shared proc pool info from k2 to DB
            self._update_spp_from_k2_to_db(context, host)
            LOG.info(_('Done Collecting Host-Shared proc pool information...'))
        except Exception as e:
            LOG.error(_("Reconciling Host-Shared proc pool "
                        "information failed."))
            LOG.exception(e)

    def fetch_hmc_mtmsn(self, access_ip, user_id, password):
        '''
        Use K2 to fetch the machine type, model, and serial number for an HMC
        '''
        # Get a K2 Session to Log-on/Connect to the HMC's K2 API's
        session = k2o.K2Session(access_ip, user_id, password)
        operator = k2o.K2Operator(session)
        try:
            # Retrieve the list of /ManagementConsole resources from K2
            resp = operator.read('ManagementConsole')
        except Exception as exc:
            # If the exception is a K2 Error, print out the K2 Response
            if isinstance(exc, k2o.K2Error):
                resp, uri = (exc.k2response, '/ManagementConsole')
                body = resp.body if resp is not None else None
                LOG.warn(_("Error reading K2 %(uri)s: %(exc)s") %
                         dict(uri=uri, exc=six.text_type(body)))
            raise exc
        # Retrieve the MTMSN (machine type model and serial number)
        mtmsn_elem = resp.feed.entries[0].element.find(
            'MachineTypeModelAndSerialNumber')
        machine_type = mtmsn_elem.findtext('MachineType')
        model = mtmsn_elem.findtext('Model')
        serial_number = mtmsn_elem.findtext('SerialNumber')
        return "{0}-{1} {2}".format(machine_type, model, serial_number)

    def _update_spp_from_k2_to_db(self, context, host):
        '''
        This method reads shared proc pool information of the host from k2
        Creates a SharedProcPool dom object, constructs a HostComputeTopology
        Saves the updated shared proc pool information to DB.
        '''
        with K2OperGuard(host=host) as k2oper:
            shr_prc_pools_resp = k2oper.read(k2const.MGT_SYS,
                                             rootId=host.uuid,
                                             childType=k2const.SHR_PRC_PL)
        proc_pools = []
        for entry in shr_prc_pools_resp.feed.entries:
            shr_prc_pool_wrapper = ShrPrcPoolWrapper(entry)
            pool_id = shr_prc_pool_wrapper.get_pool_id()
            args = {}
            args['pool_name'] = shr_prc_pool_wrapper.get_pool_name()
            args['pool_id'] = int(pool_id)
            args['host_name'] = CONF.host
            args['available_proc_units'] = \
                float(shr_prc_pool_wrapper.get_avail_proc_units())
            if pool_id != '0':
                args['maximum_proc_units'] = \
                    float(shr_prc_pool_wrapper.get_max_proc_units())
                args['reserved_proc_units'] = \
                    float(shr_prc_pool_wrapper.get_curr_rsrv_proc_units())
            else:
                args['maximum_proc_units'] = 0.0
                args['reserved_proc_units'] = 0.0
            spp = compute_dom.\
                SharedProcessorPool(context,
                                    SharedProcessorPoolFactory,
                                    None, **args)
            proc_pools.append(spp)

        # Get the ResourceTransaction (cf. DB2 session) for the nova db
        hct_factory = compute_dom.HostComputeTopologyFactory.get_factory()
        hct_topo = hct_factory.\
            construct_compute_topology(context, host_name=CONF.host,
                                       shared_processor_pools=proc_pools)
        hct_topo.save()

    @logcall
    def is_inventory_needed(self, context):
        '''
        This method will query K2 for each VIOS to see if any of them had
        an RMC state change from what our DB says.  This is meant to be a FAST
        check.  If any do, then return True.
        '''
        # Query all the VIOSes for this host
        hostfact = compute_dom.ManagedHostFactory.get_factory()
        all_vioses = hostfact.find_host_by_name(context, CONF.host).vioservers

        # Make a copy of all the VIOSes' lpar_ids so we can be sure all get
        # processed.  If there are more or less VIOSes in the db vs the
        # ManagedSystem k2response, then we obviously have a reconcile to do.
        db_vios_lparids = [vios.lpar_id for vios in all_vioses]

        # Get the ManagedSystem k2response
        host = self.host_list.get_list().values()[0]
        with K2OperGuard(host=host) as k2oper:
            mgt_sys_resp = k2oper.read(rootType=k2const.MGT_SYS,
                                       rootId=host.uuid)
        host_wrapper = ManagedSystemEntryWrapper(mgt_sys_resp.entry)

        # Get all the VIOSes associated with this ManagedSystem
        vios_links = host_wrapper.get_vios_links()

        # First easy check... do the number of vioses match?  If not, we need
        # a full inventory because one may have been added or removed.
        if len(vios_links) != len(all_vioses):
            LOG.debug('Number of vioses don\'t match.')
            # Do a full read of VIOSes so the cache is updated for reconcile,
            # which will be called by our invoker upon getting our True return.
            self._k2_read_vios(host)
            return True

        # Now check each VIOS to see if we have one whose live RMC state
        # doesn't match what's in the DB.
        for vios_link in vios_links:
            # Make sure there's an actual link in this element.
            if not vios_link:
                # Something wrong!  Whatever went wrong will go wrong again
                # in the full inventory, so there's no reason to initiate that
                # now.  Better to have it happen every 5 minutes instead of
                # every 20 seconds.
                LOG.warn(_('Found blank link in AssociatedVirtualIOServers '
                           'list.'))
                return False

            live_lparid = 0
            live_rmcstate = None
            with K2OperGuard(host=host) as k2op:
                # Parse out the VIOS uuid
                vios_uuid = vios_link.split('/')[-1]

                # Get the lparid and rmc state of this vios.  On 810 HMC we
                # could get all quick properties in one call, but 780 HMC
                # doesn't support that.  So do them one at a time to maintain
                # compatibility with 780.
                try:
                    live_lparid = int(k2op.read(rootType='VirtualIOServer',
                                                rootId=vios_uuid,
                                                suffixType='quick',
                                                suffixParm='PartitionID').body)
                    live_rmcstate = str(k2op.read(rootType='VirtualIOServer',
                                                  rootId=vios_uuid,
                                                  suffixType='quick',
                                                  suffixParm='RMCState')
                                        .body).lower().replace('"', '')
                except Exception as e:
                    # Again, this is something we'll hit in full inventory.  No
                    # sense in hitting it over and over in quick inventory.
                    LOG.warn(_('Exception getting quick properties "RMCState" '
                               'and "PartitionID": %s'), six.text_type(e))
                    return False

            # Check that we got valid lparid and rmcstate back
            if live_lparid == 0 or not live_rmcstate:
                # Again, this is something we'll hit in full inventory.  No
                # sense in hitting it over and over in quick inventory.
                LOG.warn(_('Found bad lpar_id (%(lparid)d) or rmcstate '
                           '(%(rmcstate)s). Exiting')
                         % {'lparid': live_lparid,
                            'rmcstate': live_rmcstate})
                return False

            # Finally, let's look for the live vios in the db
            for vios in all_vioses:
                db_lparid = int(vios.lpar_id)
                db_rmcstate = vios.rmc_state.lower() \
                    if vios.rmc_state else None
                # Is this our vios?
                if db_lparid == live_lparid:
                    if db_rmcstate != live_rmcstate:
                        # RMC states don't match, return True
                        LOG.info(_('Found RMC discrepancy on lpar %(lparid)d. '
                                   'Live rmcstate: %(live_state)s, db '
                                   'rmcstate: %(db_state)s.') %
                                 {'lparid': live_lparid,
                                  'live_state': live_rmcstate,
                                  'db_state': db_rmcstate})
                        # Do a full read of VIOSes so the cache is updated for
                        # reconcile, which will be called by our invoker upon
                        # getting our True return.
                        self._k2_read_vios(host)
                        return True
                    else:
                        # RMC matched, done with this live vios
                        LOG.debug('RMC matched for lpar %d' % live_lparid)

                        # Remove the lparid from the list
                        db_vios_lparids.remove(live_lparid)
                        break

        # Outside the loops.  If we got here, we processed all the live VIOSes
        # we found and none required a full inventory (or we'd have returned
        # already).  One last sanity check.  If we didn't remove all the
        # lparids from the db_vios_lparids, then that means there was some
        # vios in the db that didn't get matched to a live vios.  Return true
        # if this is the case.
        if len(db_vios_lparids) > 0:
            LOG.warn(_('Processed all live vioses yet didn\'t find lparids %d')
                     % db_vios_lparids)
            # Do a full read of VIOSes so the cache is updated for reconcile,
            # which will be called by our invoker upon getting our True return.
            self._k2_read_vios(host)
            return True

        return False

    @logcall
    def _k2_read_vios(self, host):
        '''
        This method will do a couple K2 reads of all the VIOSes on the host.
        This is done so that the cache is updated for the subsequent network
        and storage host reconciles get accurate information.  Since caching
        is done separately for xag vs no-xag reads, we'll do both reads here.

        :param host: host we're running on
        '''
        with K2OperGuard(host=host) as k2oper:
            # This read (without xag) is for the network inventory.
            k2oper.read(rootType=k2const.MGT_SYS,
                        rootId=host.uuid,
                        childType=k2const.VIOS,
                        age=0, xag=k2const.ALL_VIOS_EXT_PROPS)
            # Do another read for storage (with xag).
            k2oper.read(rootType=k2const.MGT_SYS,
                        rootId=host.uuid,
                        childType=k2const.VIOS,
                        xag=['ViosStorage'],
                        age=0)

    def macs_for_instance(self, instance):
        """Generates set of valid MAC addresses for an HMC instance."""
        # NOTE(vish): We would prefer to use 0xfe here to ensure that linux
        #             bridge mac addresses don't change, but it appears to
        #             conflict with libvirt, so we use the next highest octet
        #             that has the unicast and locally administered bits set
        #             properly: 0xfa.
        #             Discussion: https://bugs.launchpad.net/nova/+bug/921838
        # NOTE(safron): For POWER7 PowerVM, the MAC address can be directly
        #               set, so the list generated by this method will be
        #               used.  For POWER6 PowerVM, the MAC address cannot be
        #               set, so the Neutron ports will be updated with the
        #               correct MAC addresses after the LPAR has been created.
        #               The final 2 hex values represent a value between 32 and
        #               64 so we can assign it as the slot id on the system for
        #               both POWER6 and POWER7.
        #
        #               FA:xx:xx:xx:xx:[32-64]

        macs = set()
        mac_base = [0xfa,
                    random.randint(0x00, 0xff),
                    random.randint(0x00, 0xff),
                    random.randint(0x00, 0xff),
                    random.randint(0x00, 0xff),
                    random.randint(0x00, 0x00)]
        for n in range(32, 64):
            mac_base[5] = n
            macs.add(':'.join(map(lambda x: "%02x" % x, mac_base)))

        return macs

    @logcall_log_instance
    def migrate_disk_and_power_off(self, context, instance, dest,
                                   instance_type, network_info,
                                   block_devices_info=None,
                                   timeout=0, retry_interval=0):
        """Transfers the disk of a running instance in multiple phases, turning
           off the instance before the end.

        :returns: disk_info dictionary that is passed as the
                  disk_info parameter to finish_migration
                  on the destination nova-compute host
        """
        disk_info = {}

        name, host, uuid, entry = self._resolve_instance(instance)

        if self.get_host_ip_addr() != dest:
            raise NotImplementedError()
        else:
            mesg = (_('Resizing'))
            self.log_operation(mesg, instance)
            try:
                pvm_lpar.resize(context, host, uuid, instance, instance_type,
                                network_info, block_devices_info)
            except Exception as exc:
                with excutils.save_and_reraise_exception():
                    LOG.exception(exc)
        return disk_info

    @logcall
    def finish_migration(self, context, migration, instance, disk_info,
                         network_info, image_meta, resize_instance,
                         block_device_info=None, power_on=True):
        """Completes a resize, turning on the migrated instance.

        :param context: the context for the migration/resize
        :param migration: the migrate/resize information
        :param instance: the instance being migrated/resized
        :param disk_info: the newly transferred disk information
        :param network_info:
           :py:meth:`~nova.network.manager.NetworkManager.get_instance_nw_info`
        :param image_meta: image object returned by nova.image.glance that
                           defines the image from which this instance
                           was created
        :param resize_instance: True if the instance is being resized,
                                False otherwise
        :param block_device_info: instance volume block device info
        :param power_on: True if the instance should be powered on, False
                         otherwise
        """

        if (migration['dest_compute'] != migration['source_compute']):
            raise NotImplementedError()

    @logcall
    @send_ui_notification(error_message=notify_msg.RR_ERROR_DEST,
                          success_message=None,
                          log_exception=True)
    def rebuild(self, context, instance, image_meta,
                injected_files, admin_password, bdms, detach_block_devices,
                attach_block_devices, network_info=None,
                recreate=False, block_device_info=None,
                preserve_ephemeral=False):
        """Remote restarts the vm when user invokes evacuation

        :param context: security context
        :param instance: instance to be remote restarted.
        :param image_meta: image meta data.
        :param injected_files: activation files
        :param admin_password: admin password.
        :param bdms: Block device mapping.
        :param detach_block_devices: Detached block devices.
        :param attach_block_devices: block devices to attach.
        :param block_device_info: Block device info.
        :param network_info: Network info to recreate.
        :param preserve_ephemeral: To preserve the ephemeral disk.
        :param recreate: True if host supports rebuild with in-memory.
        :param source_host: Source host of the instance.
        """

        mesg = (_('Rebuilding'))
        self.log_operation(mesg, instance)

        # need to set task_state since manager code is expecting
        # it to be rebuild_spawning after exit this method
        instance.task_state = task_states.REBUILD_SPAWNING
        instance.save(expected_task_state=[task_states.REBUILDING])

        source_host = instance['host']
        inst_uuid = instance['uuid']
        inst_name = instance['display_name']
        name, host, uuid, entry = self._resolve_instance(instance)
        pwr_spec_uuid = instance.get('power_specs').get('vm_uuid')

        try:
            # Verify the instance is not already remote restarting. Verify
            # the host is not already participating in migration as a source.
            if inst_uuid in self.rebuild_dict:
                error = (_("Remote restart of %(lpar)s is already in"
                           " progress.") % {'lpar': inst_name})
                LOG.exception(error)
                raise pvm_exc.\
                    IBMPowerVMRemoteRestartInProgress(lpar=inst_name)

            # This is the entry point for remote restart operation.
            # Get a new migrate object for the target.
            self.rebuild_dict[inst_uuid] = pvm_rebuild.\
                RemoteRestart(instance, host, entry, pwr_spec_uuid)

            # Check if the destination host supports remote restart
            # and has sufficient resources.
            self.rebuild_dict[inst_uuid].\
                check_can_rebuild_on_destination(context, instance)

            # Check if the instance has the remote restart capability
            # and in eligible state to be remote restarted.
            self.rebuild_dict[inst_uuid].check_can_rebuild_instance(context,
                                                                    instance)
            # Copying original bdms prior to updating it
            src_bdmaps = copy.deepcopy(bdms)
            # Perform vscsi storage mappings.
            self.rebuild_dict[inst_uuid].pre_remote_restart(context, instance,
                                                            bdms, network_info)

            # Actual remote restart job.
            rr_done = self.rebuild_dict[inst_uuid]._rebuild(context, instance,
                                                            source_host)

        except Exception as exc:
            with excutils.save_and_reraise_exception():
                self.rebuild_dict.pop(inst_uuid, None)
                LOG.exception(exc)

        if rr_done:
            # If there is no exception and RR success
            # After successful remote restart unmap the volume from sourcehost
            self.rebuild_dict[inst_uuid].\
                _post_process(context, instance, src_bdmaps)

        # Pop the instance from the dict.
        self.rebuild_dict.pop(inst_uuid, None)
        mesg = (_('Rebuilding completed'))
        self.log_operation(mesg, instance)
        return

    @logcall
    def confirm_migration(self, migration, instance, network_info):
        """Confirms a resize, destroying the source VM."""
        pass

    def _get_resize_name(self, instance_name):
        """Rename the instance to be migrated to avoid naming conflicts

        :param instance_name: name of instance to be migrated
        :returns: the new instance name
        """
        name_tag = 'rsz_'

        # if the current name would overflow with new tag
        if ((len(instance_name) + len(name_tag)) > 31):
            # remove enough chars for the tag to fit
            num_chars = len(name_tag)
            old_name = instance_name[num_chars:]
        else:
            old_name = instance_name

        return ''.join([name_tag, old_name])

    @logcall
    def finish_revert_migration(self, instance, network_info,
                                block_device_info=None):
        """Finish reverting a resize, powering back on the instance."""
        pass

    def monitor_instances(self, context, instances):
        """
        Returns details about the VM's specified that can be considered
        an accurate inventory of the VM that is periodically collected.
        For each VM the driver needs to return a dictionary containing
        the following attributes:
            uuid:      The UUID of the Instance that was passed in
            name:      The Name of the Instance that was passed in
            cpu_utilization: The percent of CPU that is being used

        :param context: The security context for the query
        :param instances: A list of dictionary objects for each VM containing:
            uuid:      The UUID of the VM when created thru OpenStack
            name:      The Name of the VM when created thru OpenStack
            vm_uuid:   The Hypervisor-specific UUID of the VM (Optional)
        """
        data = []
        for node in self.host_list.itervalues():
            info = node.collect_instances_utilization_data(context)

            for inst in instances:
                name = inst['name']
                os_uuid = inst['uuid']
                vm_uuid = node.pvm_uuid_cache.get_pvm_uuid(name)
                # Fix for defect 7146, check for None
                if vm_uuid is None:
                    continue

                LOG.debug("Data for lookup: %s %s %s " % (str(name),
                                                          str(os_uuid),
                                                          str(vm_uuid)))
                # Account for PCM json which doesn not have dashes in
                # LPAR IDs
                if vm_uuid not in info:
                    vm_uuid = vm_uuid.replace('-', '')

                if vm_uuid in info:
                    cpu_util = info.get(vm_uuid)['cpu_utilization']
                    data.append({'uuid': os_uuid,
                                 'name': name,
                                 'cpu_util': (cpu_util * 100)})
                else:
                    LOG.debug("CPU data not found for: %s" % inst)

        return data

    @logcall
    def prep_instance_for_attach(self, context, instance, storage_type):
        lpar_name, host, lpar_uuid, entry = self._resolve_instance(instance)
        pvm_lpar.prep_instance_for_attach(host,
                                          instance,
                                          lpar_name,
                                          lpar_uuid,
                                          entry,
                                          storage_type, context)

    def evacuate_cleanup_handled_by_hmc(self, host):
        """
        Check if the RR cleanup is handled by HMC using the lsrrstartlpar
        command. This command is available only with HMC 850 and above.
        The response is expected as auto_cleanup_enabled=1 or 0
        :param host: the host object for compute host
        """
        with K2OperGuard(host=host) as k2oper:
            try:
                chk_hmc_cleanup_cmd = ('lsrrstartlpar -r mc')
                # Get the K2 response for the CLIRunner command
                k2resp = k2oper.read(k2const.MGT_CONSOLE, None,
                                     suffixType=k2const.SUFFIX_TYPE_DO,
                                     suffixParm=k2const.SUFFIX_PARM_CLI_RUNNER)
                # Create the job parameters
                job_wrapper = JobEntryWrapper(k2resp.entry)
                ack_parm = 'acknowledgeThisAPIMayGoAwayInTheFuture'
                job_parms = [job_wrapper.create_job_parameter
                             ('cmd', chk_hmc_cleanup_cmd),
                             job_wrapper.create_job_parameter(ack_parm,
                                                              'true')]
                job_wrapper.run_job(k2oper, None, job_parms=job_parms)
                results = job_wrapper.get_job_results_as_dict()
            except pvm_exc.IBMPowerVMJobRequestFailed as error:
                LOG.info(_('CLIRunner Error: %s') % error)
                return False
        value = results.get('result')
        LOG.debug('lsrrstartlpar CLI output: %s' % value)
        if value:
            conf_value = value.split("=")
            if conf_value[1] == '1':
                return True
        return False

    def handle_evacuated_instances(self, context, db_instances, drv_instances):
        """Handle the case when an instance is migrated out of band
           on HMC
        """
        host_name = CONF.host
        host = self.host_list.get_list().get(host_name)
        # If we were asked not to handle the out-of-band changes,
        # Or if the HMC is configured perform the clean up, just return
        if not CONF.ibmpowervm_handle_out_of_band_migration:
            return

        # Check if the RR clean up is handled by HMC. It supported only from
        # HMC 850 so specifically check for that.
        host_stats = host.get_host_stats(refresh=False)
        hmc_ver_obj = HMCVersion(host_stats['stats'].
                                 get(commonconsts.HMC_VERSION, None))
        if hmc_ver_obj is not None and hmc_ver_obj >= (HMCVersion('V8R8.5.0')):
            if self.evacuate_cleanup_handled_by_hmc(host):
                LOG.info(_('Evacuation cleanup handled by HMC, skip cleanup.'))
                return
        else:
            LOG.info(_('Performing Evacuation cleanup from PowerVC.'))

        # Create a list of instance uuids from the DB
        managed_instance_uuids = []
        for db_instance in db_instances:
            k2_uuid = db_instance.get('power_specs', {}).get('vm_uuid')
            # we don't use the k2_uuid here to do any GET or POST,
            # so store the uuid as lowercases
            if k2_uuid is not None:
                managed_instance_uuids.append(k2_uuid.lower())

        vios_entries = {}
        parsed_scsi_maps = {}
        with K2OperGuard(host=host) as k2oper:
            # This read is to check VIOS RMC activity and VIOS to LPAR mappings
            vioses_resp = k2oper.read(rootType=k2const.MGT_SYS,
                                      rootId=host.uuid, childType=k2const.VIOS,
                                      xag=[k2const.VIOS_VSCSI_MAP_EXT_PROP,
                                           k2const.VIOS_VFC_MAP_EXT_PROP])
        for entry in vioses_resp.feed.entries:
            wrapper = VIOSEntryWrapper(entry)
            vios_entries[wrapper.get_uuid()] = [wrapper.is_rmc_active(),
                                                entry.element.toxmlstring(),
                                                wrapper.get_vios_name(),
                                                entry.element]
            # Create empty scsi maps to be populated only in post RR cleanup
            parsed_scsi_maps[wrapper.get_uuid()] = None
        # Iterate through LPARs on the host. Handle those LPAR
        # not in DB
        all_instances = None
        srr_cleanup_states = ['Remote_Restarted', 'Source_Side_Cleanup_Failed']
        host_state_dict = {}
        host_uuid_dict = {}
        host_state = host.get_state()
        # use K2 level caching
        with K2OperGuard(host=host) as k2oper:
            lpars_resp = k2oper.read(k2const.MGT_SYS, rootId=host.uuid,
                                     childType=k2const.LPAR)
        for entry in lpars_resp.feed.entries:
            try:
                lpar_wrapper = LPAREntryWrapper(entry)
                # Retrieve LPAR UUID. Note that PowerVM UUID
                # (not PowerVC UUID) is returned
                uuid = lpar_wrapper.get_uuid()
                name = lpar_wrapper.get_lpar_name()
                srr_state = lpar_wrapper.get_srr_state()
                if uuid is None:
                    LOG.info(_('Virtual machine UUID: None. '
                               'Hypervisor partition name: %s') % name)
                    continue
                if uuid.lower() not in managed_instance_uuids:
                    if all_instances is None:
                        filters = {'deleted': False}
                        all_instances = instance_obj.InstanceList.\
                            get_by_filters(context,
                                           filters)
                    # Go through the list to find the one with matching UUID
                    for instance in all_instances:
                        host_in_db = instance.host
                        uuid_in_db = instance.get('power_specs', {}).\
                            get('vm_uuid')
                        if (uuid_in_db is not None and uuid is not None
                                and uuid.lower() == uuid_in_db.lower()
                                and host_in_db != host_name
                                and host_in_db is not None
                                and srr_state not in srr_cleanup_states):
                            node = self.get_available_nodes()[0]
                            vm_info = self.get_info(instance)
                            # Update UUID cache to pick up UUID of the LPAR
                            # that is being migrated to this host
                            if host is not None and lpars_resp is not None:
                                host.pvm_uuid_cache.\
                                    load_cache_from_feed(lpars_resp.feed)
                            # Check K2 for source host state to determine
                            # whether remote restart or migration was performed
                            # NOTE: As of 2Q15 remote restart across HMCs is
                            # not supported. This will have to be changed if
                            # the support is desired.
                            k2oper = host.hmcs[0].get_k2oper()
                            if host_in_db not in host_state_dict:
                                hostfact = compute_dom.ManagedHostFactory.\
                                    get_factory()
                                src_host = hostfact.\
                                    find_host_by_name(context, host_in_db)
                                src_host_uuid = src_host.host_uniqueid
                                host_uuid_dict[host_in_db] = src_host_uuid
                                try:
                                    src_sys_resp = k2oper.\
                                        read(k2const.MGT_SYS,
                                             rootId=src_host_uuid)
                                    host_wrapper = ManagedSystemEntryWrapper(
                                        src_sys_resp.entry)
                                    host_state_dict[host_in_db] = \
                                        host_wrapper.get_system_state()
                                except Exception as e:
                                    LOG.exception(e)
                                    LOG.error(_('Error obtaining source host '
                                                'information.'))
                            migrate_utils.verify_out_of_band_evacuation(
                                context, instance, uuid, uuid_in_db,
                                host_in_db, host_name, vm_info, node,
                                src_host_state=host_state_dict.get(host_in_db),
                                src_host_uuid=host_uuid_dict[host_in_db],
                                k2oper=k2oper)
                        # Cleanup for remote restarted VMs on source host
                        if (uuid_in_db is not None and uuid is not None
                                and uuid.lower() == uuid_in_db.lower()
                                and srr_state in srr_cleanup_states
                                and host_state == 'operating'):
                            # Check that the VIOSes the LPAR is using has an
                            # active RMC state
                            for entry in vios_entries:
                                # [1]: VIOS entry as string from K2
                                if uuid in vios_entries[entry][1]:
                                    # [0]: Boolean for RMC activity
                                    if not vios_entries[entry][0]:
                                        break
                            else:
                                self.prepare_and_invoke_rmdev(vios_entries,
                                                              parsed_scsi_maps,
                                                              lpar_wrapper,
                                                              host)
                                pvm_lpar.rr_cleanup(context, host, uuid,
                                                    instance, name)

            except Exception as e:
                LOG.exception(e)
                LOG.error(_('handle_evacuated_instances encounters '
                            'error: %s') % six.text_type(e))

    def prepare_and_invoke_rmdev(self, vios_entries, parsed_scsi_maps,
                                 lpar_wrapper, host):
        # Remove any hdisks from the VIOS
        for entry in vios_entries:
            # Parse the VIOS entry elements for SCSI
            # mappings just once and save.
            # [3]: The entry.element object
            if not parsed_scsi_maps[entry]:
                parsed_scsi_maps[entry] =\
                    self.parse_vios_scsi_map(
                        vios_entries[entry][3])
        for vios in parsed_scsi_maps:
            cid = lpar_wrapper.get_lpar_id()
            if cid is None:
                break
            disks = parsed_scsi_maps[vios].disk_info_by_lpar.get(cid, [])
            pvtypes = [storage_map_processor.VDISK, storage_map_processor.VOPT]
            if disks is None:
                continue
            for disk in disks:
                if disk.storage_type in pvtypes:
                    continue
                # [2]: The VIOS name
                vntry = vios_entries[vios][2]
                pvm_lpar._invoke_rmdev(host, host.host_display_name, vntry,
                                       disk.backing_device, vios)

    def parse_vios_scsi_map(self, element):
        """
        Return VIOSStorageInfo from parsed VirtualFibreChannelMappings.
        """
        scsi_map = SCSIMapProcessor()
        return scsi_map.parse_and_populate(element)

    def log_operation(self, op, instance):
        """
        Log entry point of user initiated operation to aid debugging
        """
        LOG.info(_('Operation: %(op)s. Virtual machine display name: '
                   '%(display_name)s, name: %(name)s, UUID: %(uuid)s') %
                 {'op': op, 'display_name': instance['display_name'],
                  'name': instance['name'], 'uuid': instance['uuid']})

    def cleanup_network_associations(self, context, host):
        """
        cleans up network associations of deleted networks
        """
        net_assn_cleanup.cleanup_network_associations(
            context, host_name=host)

    def log_free_disk(self, is_audit):
        """
        Called from resource tracker to log free disk on host
        :param is_audit: is this an audit log (Bool)
        :param resources: Host resources (Dict)
        """
        # No local storage on host for HMC
        pass

    def is_instance_on_host(self, instance_uuid, instance_name):
        """
        Check if an instance exists on this host
        """
        try:
            host_name = CONF.host
            host = self.host_list.get_list().get(host_name)
            with K2OperGuard(host=host) as k2oper:
                lparid = k2oper.read(rootType='ManagedSystem',
                                     rootId=host.uuid,
                                     childType='LogicalPartition',
                                     childId=instance_uuid,
                                     suffixType='quick',
                                     suffixParm='PartitionID').body
        except Exception as ex:
            lparid = None
            LOG.debug('Failed to find out if virtual machine %(name)s is '
                      'being managed by this host. Exception: %(exp)s' %
                      {'name': instance_name, 'exp': ex})
        return True if lparid is not None else False

    def _dispatch_thread(self):
        """Dispatches async events.

        This is a green thread which waits for events to
        arrive. This then dispatches the events to the
        compute manager.
        """

        while True:
            self._dispatch_events()

    def _queue_event(self, event):
        """Puts an event on the queue for dispatch.

        This method is called by the native event thread to
        put events on the queue for later dispatch by the
        green thread. Any use of logging APIs is forbidden.
        """

        if self._event_queue is None:
            return

        # Queue the event...
        self._event_queue.put(event)

        # ...then wakeup the green thread to dispatch it
        c = ' '.encode()
        self._event_notify_send.write(c)
        self._event_notify_send.flush()

    def _dispatch_events(self):
        """Wait for & dispatch events from native thread

        Blocks until native thread indicates some events
        are ready. Then dispatches all queued events.
        """

        # Wait to be notified that there are some
        # events pending
        try:
            _c = self._event_notify_recv.read(1)
            assert _c
        except ValueError:
            return  # will be raised when pipe is closed

        # Process as many events as possible without
        # blocking
        while not self._event_queue.empty():
            try:
                event = self._event_queue.get(block=False)
                if isinstance(event, virtevent.LifecycleEvent):
                    self.emit_event(event)
            except native_Queue.Empty:
                pass

    def _init_events_pipe(self):
        """Create a self-pipe for the native thread to synchronize on.

        This code is taken from the eventlet tpool module, under terms
        of the Apache License v2.0.
        """

        self._event_queue = native_Queue.Queue()
        rpipe, wpipe = os.pipe()
        self._event_notify_send = greenio.GreenPipe(wpipe, 'wb', 0)
        self._event_notify_recv = greenio.GreenPipe(rpipe, 'rb', 0)

    def _init_events(self):
        """Initializes the events subsystem.

        This requires running a native thread to provide the
        event loop integration. This forwards events
        to a green thread which does the actual dispatching.
        """

        self._init_events_pipe()

        LOG.debug("Starting green dispatch thread")
        eventlet.spawn(self._dispatch_thread)

    def instance_exists(self, instance_id):
        """Checks existence of an instance on the host.
        Override the super since it's now using list_instance_uuids
        """
        return instance_id in self.list_instances()

    def get_lpars_feed(self):
        lpars_resp = None
        try:
            host_name = CONF.host
            host = self.host_list.get_list().get(host_name)
            with K2OperGuard(host=host) as k2oper:
                lpars_resp = k2oper.read(k2const.MGT_SYS, rootId=host.uuid,
                                         childType=k2const.LPAR)
        except Exception as e:
            # just log the exception
            LOG.exception(e)
        return lpars_resp

    @logcall
    def set_host_dead(self, context, nodename, dead=True, target_host=None):

        # This only happens when we have just one node
        if nodename is None:
            node = self.host_list.get_list().values()[0]
        else:
            node = self.host_list.get(nodename)
        data = node.get_host_stats(refresh=False)
        if dead:
            # if the host is not capable of SRR, fail the operation
            if not data['stats'][pvc_consts.PVM_SIMP_REMOTE_RESTART_CAPABLE]:
                raise pvc_exc.IBMPowerVCNotAllowed(
                    _('Host %s is not capable of remote restart.') % nodename)
            # if the hypervisor is at error state, the stats might only contain
            # 1 value
            if (data['stats'][commonconsts.HYPERVISOR_STATE]
                    in commonconsts.RR_HYP_STATES):
                # if it's already mark dead in the stats, check the target
                # host is the same in the stats
                if data['stats'].get(commonconsts.HOST_DEAD, False):
                    # reject the request
                    raise pvc_exc.IBMPowerVCNotAllowed(
                        _('Remote restart operation for all virtual machines '
                          'under host %s is currently in progress.') %
                        nodename)
                else:
                    # update the cached host_statas
                    node.cached_host_stats.set_host_dead(
                        dead=True, target_host=target_host)
            else:
                # Raise exception on cannot set host_dead if the host is not
                # in error state
                LOG.info(_('The host %s is not in a remote restartable '
                           'state. Current host state: %s.') %
                         (nodename,
                          data['stats'][commonconsts.HYPERVISOR_STATE]))
                raise pvc_exc.IBMPowerVCNotAllowed(
                    _('Remote restart is allowed only '
                      'when the source host is down.'))
        else:
            node.cached_host_stats.set_host_dead(dead=False)
        # extract the updated stats from cacahe
        data = node.cached_host_stats.get_stats()
        LOG.info(six.text_type(data))
        return data

    def reset_inst_rmc_state(self, context, instance):
        """
        Reset instance power_specs rmc_state value to inactive when
        power off or reboot operation is performed.
        """
        try:
            power_spec_utils.update_instance_power_specs(
                context, self.virtapi, instance,
                {'rmc_state': 'inactive'})
        except Exception as e:
            LOG.exception(e)

    def unmanage_host(self, context):
        self.host_list.get_list().get(CONF.host).hmcs[0].shutdown()

    def instance_on_disk(self, instance):
        """
        For PowerVM migration and remote restart support, this is always
        True.
        """
        return True

    def __convert_data_for_host_switch(self, context, dbinsts):
        """Helper method to convert data when switching from PVM to HMC"""
        try:
            cfile = '/etc/nova/nova-%s.conf' % CONF.host
            LOG.info(_("Converting data for switch from PVM to HMC..."))
            # Loop through the VM's converting the BDM's and System MetaData
            for dbinst in dbinsts:
                vmname = dbinst.power_specs.get('vm_name')
                bdms = blkdev_obj.BlockDeviceMappingList.\
                    get_by_instance_uuid(context, dbinst.uuid)
                # Loop through the BDM's converting the necessary data
                for bdm in bdms:
                    if not bdm.connection_info:
                        continue
                    # Parse the Connection Info needed to use
                    cinfo = jsonutils.loads(bdm.connection_info)
                    data = cinfo.get('data', {})
                    ctype = data.get('connection-type')
                    tgmap = data.get('initiator_target_map')
                    # If this is NPIV, then populate the BDM Volume Connector
                    if ctype == 'npiv':
                        if data.get('volume_connector') or not tgmap:
                            continue
                        vconn, phy_to_virt, wwpns = dict(), dict(), list()
                        # Loop through the MetaData getting the mappings
                        for key, val in dbinst.system_metadata.iteritems():
                            if key.startswith('npiv_adpt_wwpns_'):
                                val = val.split(',')
                                phy_to_virt[val[0]] = val[1:]
                                wwpns.extend(val[1:])
                        # Upate the Volume Connector in the BDM and save
                        vconn['phy_to_virt_initiators'] = phy_to_virt
                        vconn['wwpns'], vconn['host'] = wwpns, vmname
                        vconn['connection_type'] = ['npiv']
                        data['volume_connector'] = dict(connector=vconn)
                        bdm.connection_info = jsonutils.dumps(cinfo)
                        bdm.save()
                    # If this is vSCSI, then populate the BDM Volume Connector
                    if ctype == 'pv_vscsi':
                        if cinfo.get('volume_connector'):
                            continue
                        # Upate the Volume Connector in the BDM and save
                        vconn = dict(connector=dict(host=vmname))
                        vconn['connector']['connection_type'] = ['pv_vscsi']
                        vconn['connection-type'] = 'pv_vscsi'
                        cinfo['volume_connector'] = vconn
                        cinfo['vscsi_target_lun'] = []
                        if 'vios_data' in data:
                            vios_data = data.get('vios_data')
                            for vios_host in vios_data:
                                vios_info = vios_data[vios_host]
                                it_map = vios_info['initiator_target_map']
                                i_wwpns = it_map.keys()
                                t_wwpns = []
                                for it_key in i_wwpns:
                                    t_wwpns.extend(it_map[it_key])
                                lun = vios_info['target_lun']
                                vscsi_dict = dict(vios_wwpn=i_wwpns,
                                                  target_lun=lun,
                                                  vios_host=vios_host,
                                                  target_wwn=t_wwpns)
                                cinfo['vscsi_target_lun'].append(vscsi_dict)
                        bdm.connection_info = jsonutils.dumps(cinfo)
                        bdm.save()
            # Since the conversion worked, delete the property from the file
            data_utils.del_openstack_config_value(cfile, 'powervm_switch_type')
        except Exception as exc:
            LOG.warn(_("Error converting data for switch from PVM: %s") % exc)
            LOG.exception(exc)
