# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import re

from eventlet import greenthread
from nova.context import get_admin_context
from oslo_log import log as logging

from oslo_config import cfg

from powervc_oslo.k2.k2operator import K2SSLError
from powervc_nova.objects.compute import dom as compute_dom
from powervc_nova.virt.ibmpowervm.hmc.k2owrappers import K2OperWrap, K2SessWrap
from powervc_nova import logcall
from powervc_nova import _
import six

LOG = logging.getLogger(__name__)

CONF = cfg.CONF


@logcall
def get_consoles_for_host_sys(virtapi, mtm_serial, hmc_list, host_sys):
    """
    For this host create the hmc object if there isn't one already
    """

    # get the hmcs from the db
    context = get_admin_context()
    hmcfact = compute_dom.ManagementConsoleFactory.get_factory()
    reg_hmcs = hmcfact.find_all_hmcs_by_host(context, mtm_serial)
    # loop through each hmc returned
    for reg_hmc in reg_hmcs:
        ip_adr = reg_hmc['access_ip']
        console = hmc_list.get(ip_adr)
        # create the hmc if it's not already in the hmcs list
        if console is None:
            # Prevent password from being printed in log
            save_pw = reg_hmc.get('password')
            if save_pw is not None:
                reg_hmc['password'] = '<SANITIZED>'
            LOG.debug("Adding HMC to list: %s" % reg_hmc)
            if save_pw is not None:
                reg_hmc['password'] = save_pw
            console = HMC(ip_adr, reg_hmc['user_id'],
                          reg_hmc['password'])
            hmc_list.add(ip_adr, console)
        host_sys.hmcs.append(console)

        # TODO: Should we confirm this here?
        console.confirm_manages_host(mtm_serial)


# Managed Console (HMC)
class HMC(object):
    """
    This object represents an HMC that the compute driver uses
    to manage the managed system (or "host")
    """
    def __init__(self, ipaddr, user, cred):
        """
        To initialize an HMC we need:

        :param ipaddr: The ip address of the HMC
        :param user: The user to log on as
        :param cred: The credential to use

        """
        self.ipaddr = ipaddr
        self.user = user
        self.cred = cred
        self._usable = True
        self._k2oper = None

    def __str__(self):
        conn = 'established' if self._k2oper else 'none'
        str_ = ('HMC(ipaddr=%s, user=%s, usable=%s, conn=%s)' %
                (self.ipaddr, self.user, self._usable, conn))
        return str_

    def __repr__(self):
        return self.__str__()

    @logcall
    def set_usable(self):
        """
        Mark this HMC as usable which means to the best of our
        knowledge it's up and running.
        """
        self._usable = True

    @logcall
    def set_unusable(self):
        """
        Mark this HMC as unusable which means the last time we checked
        it was not up and running.  Use an alternate if possible.
        """
        self._usable = False

    def is_usable(self):
        """
        Returns the whether the HMC is usable or not.
        """
        return self._usable

    def confirm_manages_host(self, mtm_serial):
        """
        Check the systems that this HMC manages and return whether
        this particular system is managed by this HMC.
        """

        # TODO:
        return True

    def get_k2oper(self):
        """
        Build a k2 operator to connect to this HMC.

        Returns a k2 operator
        """
        if self._k2oper is None:
            max_retry_count = CONF.ibmpowervm_hmc_connection_retries
            count = 1
            k2conn = None
            while k2conn is None:
                try:
                    k2conn = K2SessWrap(self.ipaddr, self.user,
                                        self.cred, timeout=CONF.
                                        ibmpowervm_comm_timeout)
                    LOG.info(_('Successfully established connection with HMC'))
                except Exception as e:
                    if max_retry_count > 0 and count >= max_retry_count:
                        LOG.error(_('Exceeded maximum number of retries %d to '
                                    'establish connection with HMC.')
                                  % max_retry_count)
                        raise e
                    if isinstance(e, K2SSLError):
                        # Check whether it's because certificate error.
                        # if so, we won't retry
                        # certificate exception looks like this:
                        # [Errno 1] _ssl.c:492: error:14090086:SSL routines:
                        # SSL3_GET_SERVER_CERTIFICATE:certificate verify failed
                        certificate_err = re.search(':14090086:',
                                                    six.text_type(e))
                        if certificate_err:
                            LOG.error(_('Failed to establish a connection with'
                                        ' the HMC due to a certificate error.'
                                        ))
                            raise e
                    LOG.warning(_("Failed to establish session with HMC. "
                                  "Retrying, count = %s") % count)
                    count = count + 1
                    greenthread.sleep(60)
            self._k2oper = K2OperWrap(k2conn)
        return self._k2oper

    def shutdown(self):
        self._k2oper.shutdown()


class HMCList(object):
    """
    Keeps track of the HMCs for the driver
    """
    def __init__(self):
        self._list = {}
        pass

    def get(self, ip_adr):
        return self._list.get(ip_adr)

    def add(self, ip_adr, hmc):
        self._list[ip_adr] = hmc
