# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import copy
import math
import logging as python_logging
import pprint
import re
import time

from oslo_config import cfg

from nova import context

from oslo_utils import importutils
from oslo_serialization import jsonutils
from oslo_log import log as logging

from powervc_nova import _
from powervc_nova import logcall
from powervc_nova.objects.compute import dom as compute_dom
from powervc_nova.virt.ibmpowervm.hmc import exception as hmcexception

import powervc_nova.virt.ibmpowervm.hmc.k2.constants as k2const
from powervc_nova.virt.ibmpowervm.hmc.k2oprguard import K2OperGuard
import six
from powervc_nova.virt.ibmpowervm.hmc.k2.cpu_util_processor \
    import CpuUtilProcessor
from powervc_nova.virt.ibmpowervm.hmc.k2.jobentry_wrapper \
    import JobEntryWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2.lparentry_wrapper \
    import LPAREntryWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2.mgmt_console_wrapper \
    import HMCVersion, MgmtConsoleWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2 import msentry_wrapper
from powervc_nova.virt.ibmpowervm.hmc.k2.viosentry_wrapper \
    import VIOSEntryWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2.monitor_pref_processor \
    import MonitorPrefProcessor
from powervc_nova.virt.ibmpowervm.hmc.k2.shr_prc_pool_wrapper \
    import ShrPrcPoolWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2 import enterprise_pool_member_wrapper\
    as ep_member_wrap

import powervc_nova.virt.ibmpowervm.common.constants as commonconsts
from powervc_nova.virt.ibmpowervm.common.constants import \
    REMOTE_RESTART_TARGET_HOST, HOST_DEAD
from powervc_nova.virt.ibmpowervm.common import proc_utils
from powervc_oslo.k2.k2operator import K2SSLError

from powervc_nova.virt.ibmpowervm.common import enterprise_pool as common_ep
from powervc_nova.virt.ibmpowervm.hmc import enterprise_pool as hmc_ep


LOG = logging.getLogger(__name__)


host_opts = [
    cfg.IntOpt('ibmpowervm_common_max_mem_mb',
               default=32768,
               help='A commonly used value for max memory MB of an instance '
               'during deploy to help calculate the reserved memory on the '
               'host'),
    cfg.IntOpt('ibmpowervm_initialize_host_retries',
               default=60,
               help='Number of attempts to initialize the lpar feed'),
    cfg.StrOpt('ibmpowervm_mover_service_partitions',
               default='',
               help='A comma separated list of VIOS IP addresses to use as '
               'the mover service partitions for live partition migration. '
               'If you have more than one network adapter on your VIOS, and '
               'want migrations to run on the faster network set the value '
               'to the IP address of the faster network. For dual VIOS '
               'environments you can list an MSP for both VIOS and one will '
               'randomly be chosen for the migration operation.'),
    cfg.IntOpt('ibmpowervm_cpu_frequency_cache_ttl',
               default=1800,
               help='Number of seconds that the cached CPU frequency value '
               'should last before it is fetched again from the VIOS.'),
    cfg.BoolOpt('nonclassic_supported', default=False,
                help='Whether or not HMC Non-Classic Mode is supported')
]

LOG = logging.getLogger(__name__)

CONF = cfg.CONF
CONF.register_opts(host_opts)

compute_host = None

#
# stats list specific to this driver. It will be added to the
# COMPUTE_STATS common list for values that should be returned in
# the 'stats' value of the Host.get_host_stats table
#
EXTRA_COMPUTE_STATS = [
    'max_migration_ops_supported',
    'mover_service_partitions',
    commonconsts.HMC_VERSION]

FAKE_STG_SIZE = (1 << 21)
#
#
# def get_list(oper):
#    k2resp = oper.read(k2const.MGT_SYS)
#    return k2resp


class HostFeed(object):

    def __init__(self):
        self.resp = None

    def load(self, oper):
        self.resp = oper.read(k2const.MGT_SYS)

    def find_entry_by_mtm_serial(self, mtm_serial):
        xpath = './MachineTypeModelAndSerialNumber/SerialNumber'
        machine, model, serial = Host.parse_mtm(mtm_serial)
        entries = self.resp.feed.findentries(xpath, serial)
        if entries is None:
            return None

        # Confirm same model and type
        machine_xpath = './MachineTypeModelAndSerialNumber/MachineType'
        model_xpath = './MachineTypeModelAndSerialNumber/Model'
        for entry in entries:
            entry_machine = entry.element.findtext(machine_xpath)
            entry_model = entry.element.findtext(model_xpath)
            if entry_machine == machine and entry_model == model:
                return entry

        # No matching MTM Serial was found
        return None


class Host(object):

    reserved_memory_overhead = None

    # cpu_info that will be returned by build_host_stats_from_entry()
    host_stats_cpu_info = jsonutils.dumps({'vendor': 'ibm', 'arch': 'ppc64'})

    class HostStatsCache(object):
        def __init__(self, mgt_etag, prc_pool_etag, epool_member_etag, stats):
            self.mgt_etag = mgt_etag
            self.prc_pool_etag = prc_pool_etag
            self.epool_member_etag = epool_member_etag
            self.stats = copy.deepcopy(stats)
            self.host_dead = False
            self.host_dead_target_host = None

        def is_valid(self, mgt_etag, prc_pool_etag, epool_member_etag,
                     instances_res_totals):
            # Compares etags with cached etags for managed system and processor
            # pool elements. Also checks if used vs reserved values have
            # shifted as a result of managing or unmananging VMs by looking
            # at the value for managed memory.
            # also need to compare the number of remote restart in progress
            if (mgt_etag == self.mgt_etag and
                prc_pool_etag == self.prc_pool_etag and
                epool_member_etag == self.epool_member_etag and
                self.stats is not None and
                instances_res_totals['managed_memory'] ==
                    self.stats['memory_mb_used'] and
                '%.2f' % (instances_res_totals['managed_proc_units']) ==
                    self.stats['proc_units_used'] and
                (self.stats.get('stats') is not None and
                 instances_res_totals[commonconsts.NUM_RR_IN_PROGRESS] ==
                    self.stats['stats'].get(commonconsts.NUM_RR_IN_PROGRESS))):
                return True
            return False

        def get_stats(self):
            return copy.deepcopy(self.stats)

        def set_host_dead(self, dead=True, target_host=None):
            self.host_dead = dead
            # it's possible that self.stats is None if compute service
            # is started up with hypervisor state not in 'operating' or
            # 'standby'
            if self.stats is not None:
                self.stats['stats'][HOST_DEAD] = dead
            if dead:
                self.host_dead_target_host = target_host
                if target_host and self.stats is not None:
                    self.stats['stats'][REMOTE_RESTART_TARGET_HOST] = \
                        target_host
                elif (not target_host and self.stats is not None and
                      self.stats.get('stats') is not None):
                    self.stats['stats'].pop(REMOTE_RESTART_TARGET_HOST,
                                            None)
            else:
                self.host_dead_target_host = None
                if (self.stats is not None and
                        self.stats.get('stats') is not None):
                    self.stats['stats'].pop(REMOTE_RESTART_TARGET_HOST,
                                            None)

    def __init__(self, mtm_serial=None, uuid=None, host_id=None, virtapi=None):
        self.host_id = host_id
        self.mtm_serial = mtm_serial
        self.host_display_name = CONF.host_display_name
        self.uuid = uuid
        self.hmcs = []
        self.pvm_uuid_cache = None
        self.cached_host_stats = self.HostStatsCache(0, 0, 0, None)
        self.source_host_for_migration = False
        self.vif_driver = None
        self.virtapi = virtapi
        # Tuple representing the cached CPU frequency, where the first entry in
        # the tuple is the frequency itself and the second entry is the time at
        # which the cached CPU frequency is deemed expired by the TTL
        self._cached_cpu_frequency = (None, None)
        self.cpu_util_processor = CpuUtilProcessor()

    def __str__(self):
        str_ = ('Host(host_id=%s, mtm_serial=%s, uuid=%s)'
                % (self.host_id, self.mtm_serial, self.uuid))
        return str_

    def __repr__(self):
        return self.__str__()

    @property
    def source_host_for_migration(self):
        return self._source_host_for_migration

    @source_host_for_migration.setter
    def source_host_for_migration(self, val):
        self._source_host_for_migration = val
        # Copy it into host cache
        if self.cached_host_stats and self.cached_host_stats.stats:
            self.cached_host_stats.stats['source_host_for_migration'] = val

    def load_UUID_cache(self, db_instances, lpar_feed):
        self.pvm_uuid_cache = Instance_to_PVM_UUID_Map()
        # First we will load the UUID Map from the HMC, for a catch all
        self.pvm_uuid_cache.load_cache_from_feed(lpar_feed)
        # Second we will load the UUID Map from the DB, to map the OS name
        self.pvm_uuid_cache.load_cache_from_db(db_instances)

    def refresh_UUID_cache(self, db_instances):
        self.pvm_uuid_cache.load_cache_from_db(db_instances)

    def select_active_hmc(self):
        for console in self.hmcs:
            if console.is_usable():
                return console
        return None

    def reset_hmcs(self):
        for console in self.hmcs:
            console.set_usable()

    @staticmethod
    def parse_mtm(mtm_serial):
        mtm, serial = mtm_serial.split('_', 1)
        model = mtm[4:7]
        machine = mtm[0:4]
        return machine, model, serial

    @logcall
    def init_vif_driver(self):
        with K2OperGuard(host=self) as k2opr:
            vif_conf = CONF.powervm_vif_driver
            self.vif_driver = importutils.import_object(vif_conf,
                                                        self.mtm_serial,
                                                        k2opr,
                                                        self.mtm_serial)

            # Need to set the HMC VIF Driver operator for the out of band
            # operations that occur outside of the host trigger.
            from powervc_nova.virt.ibmpowervm.vif import hmc as vif_hmc
            vif_hmc.K2_OPER = k2opr

    @staticmethod
    def create_lpar_on_host(oper, lpar_attr, uuid):
        oper.create(lpar_attr, k2const.MGT_SYS, uuid, k2const.LPAR)

    def create_lpar(self, oper, lpar_attr):
        Host.create_lpar_on_host(oper, lpar_attr, self.uuid)

    def build_host_stats_from_entry(self, entry, epool_member_entry,
                                    instances_res_totals):
        """
        From a ManagedSystem entry and enterprise pool member entry build the
        host_stats dictionary
        """

        msentry = msentry_wrapper.ManagedSystemEntryWrapper(entry)

        epool_member_entry_wrapper = None
        if epool_member_entry is not None:
            epool_member_entry_wrapper = ep_member_wrap.\
                EnterprisePoolMemberEntryWrapper(epool_member_entry)

        hostname = self.mtm_serial
        # Get state of hypervisor
        state = msentry.get_system_state()
        # Some attributes require PHYP to obtain. If state is not good,
        # we won't be able to get all host stats, so return with
        # current state set in data, and DB will keep existing good
        # stats.
        if state != 'operating' and state != 'standby':
            LOG.debug(("State of host %s is %s. Unable to retrieve "
                       "host stats.") % (hostname, state))
            data = {'hypervisor_hostname': hostname,
                    'stats': {commonconsts.HYPERVISOR_STATE: state}}
            return data

        data = {}
        data['system_name'] = msentry.get_system_name()
        data['hypervisor_type'] = commonconsts.IBM_POWERVM_HYPERVISOR_TYPE
        data['hypervisor_version'] = (
            commonconsts.IBM_POWERVM_HYPERVISOR_VERSION)
        data['supported_instances'] = commonconsts.POWERVM_SUPPORTED_INSTANCES
        data['hypervisor_hostname'] = hostname

        capabilities = msentry.get_capabilities()
        data['capabilities'] = capabilities

        data['cpu_info'] = self.host_stats_cpu_info

        data['proc_units'] = msentry.get_proc_units_configurable()
        data['proc_units_used'] = '%.2f' % (
            instances_res_totals['managed_proc_units'])
        shr_prc_pool_total = self.get_proc_units_from_shr_prc_pools()
        data['proc_units_reserved'] = '%.2f' % (
            instances_res_totals['unmanaged_proc_units'] +
            shr_prc_pool_total)

        data['vcpus'] = int(math.ceil(float(data['proc_units'])))
        data['vcpus_used'] = int(math.ceil(float(data['proc_units_used'])))
        data['vcpus_reserved'] = int(
            math.ceil(float(data['proc_units_reserved'])))

        data['max_sys_procs_limit'] = msentry.get_max_sys_procs_limit()
        data['max_procs_per_aix_linux_lpar'] = (
            msentry.get_max_procs_per_aix_linux_lpar())
        data['max_sys_vcpus_limit'] = msentry.get_max_sys_vcpus_limit()
        data['max_vcpus_per_aix_linux_lpar'] = (
            msentry.get_max_vcpus_per_aix_linux_lpar())
        data['mover_service_partitions'] = (
            CONF.ibmpowervm_mover_service_partitions)
        data['lmb_size'] = msentry.get_memory_region_size()

        data['host_memory_total'] = msentry.get_memory_configurable()
        data['host_memory_free'] = msentry.get_memory_free()
        # host_memory_rserved = memory allocated to instances on the host
        # not managed by PowerVC, memory overhead required during instance
        # creation by PHYP, and memory allocated to firmware on the host
        data['host_memory_reserved'] = (
            instances_res_totals['unmanaged_memory'] +
            self.get_memory_overhead(data['lmb_size']) +
            msentry.get_firmware_memory())

        data['memory_mb'] = data['host_memory_total']
        data['memory_mb_used'] = instances_res_totals['managed_memory']

        # local_gb is not used for HMC.
        # Return a large number for local_gb to resolve
        # warning messages that are continuously logged. Issue 6960.
        data['local_gb'] = FAKE_STG_SIZE
        data['local_gb_used'] = 0

        data['initiator'] = 'none'

        data['extras'] = ''
        data["numa_topology"] = None

        # Migration data
        data['source_host_for_migration'] = self.source_host_for_migration
        # The database needs key value pairs where the value is a string.
        # Compatibility modes is a list of strings - covert to string.
        data['compatibility_modes'] = ",".join(msentry.get_proc_compat_modes())
        data.update(msentry.get_migration_data())

        # determine whether host suport little endian guest based on
        # proc compatibility mode
        le_guest_cap = proc_utils.le_guest_support_from_proc_compat(
            data['compatibility_modes'])
        data['capabilities'][commonconsts.LE_GUEST_CAPABLE] = le_guest_cap
        # VIOS' virtual slots information
        # The information is used by EGO to determine whether it can
        # allocate virtual slots on the VIOS
        data['vios_max_virtual_slots'] = 40

        stats = {}
        # Enterprise pool member stats (CoD)
        if epool_member_entry_wrapper is not None:
            stats[commonconsts.COD_ENABLED] = True
            stats[commonconsts.COD_INACTIVE_CORES] = \
                epool_member_entry_wrapper.inactive_cores
            stats[commonconsts.COD_MOBILE_CORES] = \
                epool_member_entry_wrapper.mobile_cores
            stats[commonconsts.COD_UNRETURNED_MOBILE_CORES] = \
                epool_member_entry_wrapper.unreturned_mobile_cores
        else:
            stats[commonconsts.COD_ENABLED] = False
            stats[commonconsts.COD_INACTIVE_CORES] = 0
            stats[commonconsts.COD_MOBILE_CORES] = 0
            stats[commonconsts.COD_UNRETURNED_MOBILE_CORES] = 0
        # end enterprise pool (CoD) stats
        stats[commonconsts.HYPERVISOR_STATE] = state
        stats[commonconsts.MANAGED_BY] = commonconsts.HMC_PLATFORM
        # update remote restart concurrent support for 840 HMC
        # you cannot upgrade HMC major version without reboot hmc,
        # so it's okay don't cache the console object for comparison
        hmc_version = None
        with K2OperGuard(host=self) as k2opr:
            mgt_console_resp = k2opr.read(k2const.MGT_CONSOLE)
            mgt_console = MgmtConsoleWrapper(mgt_console_resp.feed.entries[0])
            hmc_base_version = mgt_console.get_hmc_base_version()
            hmc_version = HMCVersion(hmc_base_version)
            stats[commonconsts.HMC_VERSION] = hmc_base_version
        if hmc_version is not None and hmc_version >= (HMCVersion('V8R8.4.0')):
            stats[commonconsts.PRS_MAX_PARA_REBUILDS] = \
                commonconsts.MAX_PARA_RR_SRC_840
            stats[commonconsts.PRS_MAX_PARA_DEST_REBUILDS] = \
                commonconsts.MAX_PARA_RR_DEST_840
        else:
            stats[commonconsts.PRS_MAX_PARA_REBUILDS] = \
                CONF.prs_max_parallel_rebuilds_hmc
            stats[commonconsts.PRS_MAX_PARA_DEST_REBUILDS] = \
                CONF.prs_max_parallel_dest_rebuilds_hmc
        stats[commonconsts.NUM_RR_IN_PROGRESS] = \
            instances_res_totals[commonconsts.NUM_RR_IN_PROGRESS]
        stats[HOST_DEAD] = False
        stats[REMOTE_RESTART_TARGET_HOST] = None
        if capabilities:
            capable_list = [commonconsts.ACTIVE_LPAR_MOBILITY_CAPABLE,
                            commonconsts.INACTIVE_LPAR_MOBILITY_CAPABLE,
                            commonconsts.PVM_SIMP_REMOTE_RESTART_CAPABLE,
                            commonconsts.ACTIVE_MEMORY_EXPANSION_CAPABLE,
                            commonconsts.CUSTOM_MAC_ADDR_CAPABLE,
                            commonconsts.IBMi_LPAR_MOBILITY_CAPABLE,
                            commonconsts.IBMi_RESTRICTED_CAPABLE,
                            commonconsts.LE_GUEST_CAPABLE,
                            commonconsts.AIX_CAPABLE,
                            commonconsts.IBMI_CAPABLE,
                            commonconsts.LINUX_CAPABLE,
                            commonconsts.SHR_PROC_POOL_CAPABLE]
            for key in capable_list:
                if key in capabilities:
                    stats[key] = capabilities[key]

        # If the current Manager isn't the Master update the stats accordingly
        self._update_stats_for_non_master(entry, stats)

        # Stats should only contain a specific list
        # of data attributes as defined in COMPUTE_STATS
        # If additional properties should be in 'stats',
        # then those properties must be added to the
        # COMPUTE_STATS list
        missing_stats = []
        for key in commonconsts.COMPUTE_STATS + EXTRA_COMPUTE_STATS:
            if key in data:
                stats[key] = data[key]
            else:
                missing_stats.append(key)

        data['stats'] = stats

        return data

    def get_cpu_frequency(self, context, viosid):
        current_time = int(time.time())

        if self._cached_cpu_frequency[0]:
            expires_at = self._cached_cpu_frequency[1]
            if ((expires_at - current_time) >
                    CONF.ibmpowervm_cpu_frequency_cache_ttl):
                # This handles the case in which the system clock was set back
                # (e.g., for daylight savings adjustments); we don't want the
                # cache to be 'stuck' while time catches up, so just force the
                # cached entry to expire and cache a new value.
                expires_at = current_time
            if expires_at > current_time:
                # i.e., the cached value's TTL has not expired.
                return self._cached_cpu_frequency[0]

        with K2OperGuard(host=self) as k2opr:
            k2resp = k2opr.read(
                k2const.MGT_CONSOLE, None, suffixType=k2const.SUFFIX_TYPE_DO,
                suffixParm='CLIRunner')
            job_wrapper = JobEntryWrapper(k2resp.entry)

            run_cmd = ('viosvrcmd -m \"%(host)s\" --id \"%(vios)s\" '
                       '-c \"lsdev -dev proc0 -attr frequency\"' %
                       {'host': self.host_display_name,
                        'vios': str(viosid)})
            # Create job parameters
            job_parms = [job_wrapper.create_job_parameter(
                'cmd', run_cmd)]
            job_parms.append(job_wrapper.create_job_parameter(
                'acknowledgeThisAPIMayGoAwayInTheFuture', 'true'))

            try:
                job_wrapper.run_job(k2opr, None, job_parms=job_parms,
                                    timeout=120)
                results = job_wrapper.get_job_results_as_dict()
            except Exception as error:
                LOG.error(_("Error obtaining CPU frequency for host "
                            "%(host)s and VIOS %(vios)s: %(error)s.") %
                          {'host': CONF.host, 'vios': str(viosid),
                           'error': error})
                LOG.debug("Defaulting required CPU frequency for host %(host)s"
                          " to %(default)d Mhz." % {'host': CONF.host,
                                                    'default': 0})
                return 0

            value = results.get('result')
            val = value.split()
            pval = ''
            if len(val) == 2 and val[0] == 'value':
                pval = val[1]

            cpu_freq = long(long(pval) / math.pow(10, 6))

            # Update the cached CPU freq value
            self._cached_cpu_frequency = (
                cpu_freq,
                current_time + CONF.ibmpowervm_cpu_frequency_cache_ttl)

            return cpu_freq

    def get_host_stats(self, refresh=False):
        """
        Return currently known host stats. If host stats is not
        available, return a dict with state of 'error', but does
        not raise exception.
        """
        try:
            host_state = {'state': None}
            data = self.get_host_stats_internal(refresh, host_state)
        except Exception as e:
            if isinstance(e, K2SSLError):
                # Check whether it's because certificate error.
                # if so, we will set the reason to invalid_certificate

                certificate_err = re.search('certificate', six.text_type(e))
                state = "failed authentication"
                if certificate_err:
                    LOG.error(_('Failed to establish a connection with'
                                ' the HMC due to a certificate error.'))
                    state = "invalid_certificate"
            else:
                state = host_state.get('state')
                if state is None:
                    state = 'error'
            data = {'hypervisor_hostname': self.mtm_serial,
                    'stats': {commonconsts.HYPERVISOR_STATE: state}}
            data['stats'][HOST_DEAD] =\
                self.cached_host_stats.host_dead
            srr_cap = commonconsts.PVM_SIMP_REMOTE_RESTART_CAPABLE
            if host_state.get(srr_cap):
                data['stats'][srr_cap] = host_state.get(srr_cap)
            if self.cached_host_stats.host_dead:
                if self.cached_host_stats.host_dead_target_host:
                    data['stats'][REMOTE_RESTART_TARGET_HOST] =\
                        self.cached_host_stats.host_dead_target_host
        return data

    def get_host_stats_internal(self, refresh=False, host_state=None):
        """
        Return currently known host stats. If host stats is not
        available, raise an exception. This method is meant to be
        called only within hmc component.
        """
        shr_prc_pools_resp = None
        mgt_sys_elem = None
        try:
            with K2OperGuard(host=self) as k2opr:
                # use K2 level caching
                mgt_sys_elem = k2opr.read(k2const.MGT_SYS, rootId=self.uuid)
                shr_prc_pools_resp = k2opr.read(k2const.MGT_SYS,
                                                rootId=self.uuid,
                                                childType=k2const.SHR_PRC_PL)
        except Exception as e:
            if mgt_sys_elem is None:
                LOG.debug('mgt_sys_elem is None in the host object')
            elif shr_prc_pools_resp is None:
                LOG.debug('shr_prc_pools is None in the host object')
            else:
                LOG.error(_("Could not get managed system %(host)s. "
                            "K2 returned with error: %(error)s") %
                          {'host': self.host_display_name,
                           'error': six.text_type(e)})
            raise e

        epool_member = None
        uuids = common_ep.get_pool_member_uuids(
            context.get_admin_context(), self.host_id)
        if uuids is not None:
            pool_uuid, member_uuid = uuids
            epool_member = hmc_ep.get_member(self, pool_uuid, member_uuid)

        # Since we have the UID, we get just one entry back
        if mgt_sys_elem is None:
            LOG.debug('mgt_sys_elem k2 response is None in the host object.')
            raise hmcexception.IBMPowerVMIncompleteHostStatistics(
                host=self.host_display_name)

        # Get used memory and processor unit totals from all instances
        # managed in the database, and used memory and processor unit
        # totals from all unmanaged instances
        inst_res_totals_stats = self.calculate_used_values_stats_from_insts()

        # We now have all the data we need, check if it was updated at all
        mgt_etag = mgt_sys_elem.headers.get('etag')
        prc_pool_etag = shr_prc_pools_resp.headers.get('etag')

        epool_member_etag = 0
        if epool_member is not None:
            epool_member_etag = epool_member.headers.get('etag')

        if (self.cached_host_stats.is_valid(mgt_etag, prc_pool_etag,
                                            epool_member_etag,
                                            inst_res_totals_stats)):
            LOG.debug('Returned cached host stats.')

            host_stats = self.cached_host_stats.get_stats()

        else:
            entry = mgt_sys_elem.entry
            epool_member_entry = None
            if epool_member is not None:
                epool_member_entry = epool_member.entry

            # Now build the stats from the entry
            host_stats = self.build_host_stats_from_entry(
                entry, epool_member_entry, inst_res_totals_stats)
            stats = host_stats.get('stats', {})
            state = stats.get('hypervisor_state')
            # If the only thing in the stats is the state and is error, this
            # means the stats is incomplete, raise exception so we won't
            # get key error later on.
            if len(stats) == 1 and state not in ('operating', 'standby'):
                if host_state is not None:
                    host_state['state'] = state
                    # need to populate the remote restart capability at least
                    # so, we can perform remote restart
                    stats = {}
                    msentry = msentry_wrapper.ManagedSystemEntryWrapper(entry)
                    capabilities = msentry.get_capabilities()
                    srr_cap = commonconsts.PVM_SIMP_REMOTE_RESTART_CAPABLE
                    if capabilities and srr_cap in capabilities:
                        host_state[srr_cap] = capabilities[srr_cap]
                raise hmcexception.IBMPowerVMIncompleteHostStatistics(
                    host=self.host_display_name)
            self.cached_host_stats = self.HostStatsCache(mgt_etag,
                                                         prc_pool_etag,
                                                         epool_member_etag,
                                                         host_stats)
            if LOG.logger and LOG.logger.isEnabledFor(python_logging.DEBUG):
                LOG.debug('host_stats=%s' %
                          pprint.PrettyPrinter(indent=1).pformat(host_stats))

        return host_stats

    @logcall
    def enable_long_term_monitoring(self):
        try:
            console = self.select_active_hmc()
            k2oper = console.get_k2oper()

            mon_pref_processor = MonitorPrefProcessor()
            mon_pref_processor.enable_ltm(k2oper, self.uuid)
        except Exception:
            LOG.error(_('Error starting long term monitoring for %s.') %
                      str(self.mtm_serial), exc_info=True)

    def collect_instances_utilization_data(self, context):
        util_data = {}
        console = self.select_active_hmc()
        k2oper = console.get_k2oper()

        # Get the cpu utilization data for instances on cec
        data = self.cpu_util_processor.get_cpu_utilization_data(
            k2oper, self.uuid)
        for k2_lpar_uuid, cpu_util_data in six.iteritems(data):
            if k2_lpar_uuid in util_data:
                util_data.get(k2_lpar_uuid)['cpu_utilization'] = cpu_util_data
            else:
                util_data[k2_lpar_uuid] = {'cpu_utilization': cpu_util_data}
        # could someday get get other types of data such as IO
        return util_data

    def get_host_utilization_data(self, context):
        console = self.select_active_hmc()
        k2oper = console.get_k2oper()

        # It will be sum of all the util data
        data = self.cpu_util_processor.get_host_utilization(k2oper, self.uuid)
        return data

    def plug_vifs(self, instance, network_info):
        self.vif_driver.plug(instance, network_info)

    def unplug_vifs(self, instance, network_info):
        self.vif_driver.unplug(instance, network_info)

    def get_proc_units_from_shr_prc_pools(self):
        """
        Loop through the shared processor pools in the cache
        and total up the processing units reserved to each one
        other than the default pool
        :returns: float: total processing units reserved to other pools
        """
        total_proc_units = 0.0
        # use K2 level caching
        with K2OperGuard(host=self) as k2oper:
            shr_prc_pools_resp = k2oper.read(k2const.MGT_SYS, rootId=self.uuid,
                                             childType=k2const.SHR_PRC_PL)
        for entry in shr_prc_pools_resp.feed.entries:
            shr_prc_pool_wrapper = ShrPrcPoolWrapper(entry)
            pool_id = shr_prc_pool_wrapper.get_pool_id()
            if pool_id != '0':
                rsrv_proc_units = float(
                    shr_prc_pool_wrapper.get_curr_rsrv_proc_units())
                total_proc_units += rsrv_proc_units
        return total_proc_units

    def get_instances_from_db(self):
        """Returns a list of db instances for this host."""
        instfact = compute_dom.ManagedInstanceFactory.get_factory()
        db_instances = instfact.find_all_instances_by_host(
            context.get_admin_context(), CONF.host)
        return db_instances

    def calculate_used_values_stats_from_insts(self):
        """
        Finds instances on host that are not present in the database,
        and calculates totals for unmanaged memory and unmanaged proc units.
        Calculates totals for managed memory and managed proc units
        for instances in database.
        Also keep track of number VM with remote restart status.
        :returns: dictionary with the following keys:
                  unmanaged_memory
                  unmanaged_proc_units
                  managed_memory
                  managed_proc_units
                  remote_restarts_in_progress
        """
        res_totals = {'unmanaged_memory': 0,
                      'unmanaged_proc_units': 0,
                      'managed_memory': 0,
                      'managed_proc_units': 0,
                      commonconsts.NUM_RR_IN_PROGRESS: 0}

        lpars_resp = None
        vioses_resp = None
        # use K2 level caching
        try:
            with K2OperGuard(host=self) as k2oper:
                lpars_resp = k2oper.read(k2const.MGT_SYS, rootId=self.uuid,
                                         childType=k2const.LPAR)
                vioses_resp = k2oper.read(k2const.MGT_SYS, rootId=self.uuid,
                                          childType=k2const.VIOS, xag=[None])
        except Exception:
            LOG.debug("Cannot extract K2 feed")

        if self.virtapi is None or lpars_resp is None:
            LOG.debug("Cannot determine reserved values without "
                      "virtapi(%s) and lpar_resp(%s)" %
                      (self.virtapi, lpars_resp))
            return res_totals

        # Create a list of instance uuids from the DB
        managed_instance_uuids = []
        db_instances = self.get_instances_from_db()
        for db_instance in db_instances:
            vm_uuid = db_instance.get('power_specs', {}).get('vm_uuid')
            # we don't use the k2_uuid here to do any GET or POST,
            # so store the uuid as lowercases
            if vm_uuid is not None:
                managed_instance_uuids.append(vm_uuid.lower())

        # Iterate through LPARs on the host and check if they are
        # in the DB or not.
        for entry in lpars_resp.feed.entries:
            lpar_wrapper = LPAREntryWrapper(entry)
            uuid = lpar_wrapper.get_uuid()
            memory = int(lpar_wrapper.get_current_mem())
            has_shared_proc = lpar_wrapper.get_current_proc_mode() == "false"
            if has_shared_proc:
                proc_units = float(lpar_wrapper.get_current_proc_units())
            else:
                proc_units = float(lpar_wrapper.get_current_procs())
            # If managed add memory and proc units to managed totals
            if uuid.lower() in managed_instance_uuids:
                res_totals['managed_memory'] += memory
                res_totals['managed_proc_units'] += proc_units
            # If unmanaged add memory and proc units to unmanaged totals
            else:
                res_totals['unmanaged_memory'] += memory
                res_totals['unmanaged_proc_units'] += proc_units
            # check vm's srr state to see srr is in progress
            if (lpar_wrapper.get_srr_state() in
                ['Destination_Remote_Restarting',
                 'Destination_Remote_Restarted']):
                res_totals[commonconsts.NUM_RR_IN_PROGRESS] += 1

        # Add mem and procs for VIOS Partitions to unmanaged total
        if vioses_resp is None:
            LOG.debug("Cannot determine reserved values for"
                      " vios partitions %s" % (vioses_resp))
            return res_totals

        for entry in vioses_resp.feed.entries:
            vios_entry_wrapper = VIOSEntryWrapper(entry)
            memory = int(vios_entry_wrapper.get_current_mem())
            has_shared_proc = (
                vios_entry_wrapper.get_current_proc_mode() == "false")
            if has_shared_proc:
                proc_units = float(vios_entry_wrapper.get_current_proc_units())
            else:
                proc_units = float(vios_entry_wrapper.get_current_procs())
            res_totals['unmanaged_memory'] += memory
            res_totals['unmanaged_proc_units'] += proc_units

        return res_totals

    def get_memory_overhead(self, lmb_size):
        if Host.reserved_memory_overhead is None:
            reserved_mem_info = self.calculate_memory_overhead_on_host(
                lmb_size=lmb_size)
            Host.reserved_memory_overhead = reserved_mem_info[0]
        return Host.reserved_memory_overhead

    @logcall
    def calculate_memory_overhead_on_host(self, reserved_mem_data={},
                                          lmb_size=None, default=512):
        """
        A certain amount of additional memory, such as memory for firmware,
        is required from the host during the creation of an instance or while
        changing an instance's memory specifcations. This method queries the
        host to get the reserved PHYP memory needed for an LPAR. The
        calculation is based off of the instance's max memory requested,
        network and I/O adapter configurations, and the host's HPT ratio.
        The job response contains a value for total memory required to create
        or change the LPAR, which is desired memory plus reserved PHYP memory.

        :parm reserve_mem_data: dictionary with values for job params
        :parm default: default value to use for required memory overhead
                       value if there was an error with the job
        :returns tuple of ints PHYP memory required. (Required memory -
                 desired memory) and current free memory on the host
        """

        # If desired memory and maximum memory are not known, this query is
        # part of calculating host stats, and specific configurations of an
        # instance is not known. This will use the config option for a default
        # maximum memory.
        desired_mem = reserved_mem_data.get('desired_mem', 512)
        max_mem = reserved_mem_data.get('max_mem',
                                        CONF.ibmpowervm_common_max_mem_mb)

        # If lmb size is given, round max mem up to be a  multiple
        # of lmb size. If max_mem is 0, max_mem will be set to lmb size.
        if lmb_size is not None:
            max_mem = int(math.ceil((max_mem or 1) / float(lmb_size)) *
                          int(lmb_size))
        num_virt_eth_adapter = reserved_mem_data.get('num_virt_eth_adapters')

        with K2OperGuard(host=self) as k2opr:
            k2resp = k2opr.read(
                k2const.MGT_SYS, self.uuid, suffixType=k2const.SUFFIX_TYPE_DO,
                suffixParm=k2const.SUFFIX_PARM_QUERY_RESERVE_MEM)
            job_wrapper = JobEntryWrapper(k2resp.entry)

            # Create job parameters
            job_parms = [job_wrapper.create_job_parameter(
                'LogicalPartitionEnvironment', 'AIX/Linux')]
            job_parms.append(job_wrapper.create_job_parameter(
                'DesiredMemory', str(desired_mem)))
            job_parms.append(job_wrapper.create_job_parameter(
                'MaximumMemory', str(max_mem)))
            # If number of virtual ethernet adapters is provided, use it,
            # otherwise it will be defaulted to 2.
            if num_virt_eth_adapter is not None:
                job_parms.append(job_wrapper.create_job_parameter(
                    'NumberOfVirtualEthernetAdapter',
                    str(num_virt_eth_adapter)))
            job_parms.append(job_wrapper.create_job_parameter(
                'NumberOfVirtualSCSIAdapter', '1'))

            try:
                job_wrapper.run_job(k2opr, self.uuid, job_parms=job_parms,
                                    timeout=120)
                results = job_wrapper.get_job_results_as_dict()
            except Exception as error:
                LOG.error(_("Error obtaining host memory overhead for host "
                            "'%(host)s' %(error)s.") %
                          {'host': CONF.host, 'error': error})
                LOG.debug("Defaulting required memory overhead for host '%s'"
                          " to %d MB" % (CONF.host, default))
                return default, None
            required_mem = results.get('RequiredMemory')
            avail_mem = results.get('CurrentAvailableSystemMemory')
            if required_mem is not None:
                overhead = int(required_mem) - desired_mem
            else:
                overhead = default
            if avail_mem is not None:
                avail_mem = int(avail_mem)

            return overhead, avail_mem

    def get_state(self):
        """
        Get state of the host.
        :returns state as a string. Handles all exceptions to return None.
        """
        mgt_sys_resp = None
        try:
            with K2OperGuard(host=self) as k2opr:
                # Always refresh managed system, since we need
                # the up to date host state, otherwise we risk
                # the instance going into error state
                mgt_sys_resp = k2opr.read(k2const.MGT_SYS, rootId=self.uuid,
                                          age=0)
        except Exception as e:
            if mgt_sys_resp is None:
                LOG.debug("mgt_sys_elem is None in the host object")
            else:
                LOG.error(_("Could not get managed system info for"
                            " %(managed_sys)s. K2 returned with error:"
                            " %(error)s") % {'managed_sys': self.mtm_serial,
                                             'error': six.text_type(e)})
            return None

        # Since we have the UID, we get just one entry back
        if mgt_sys_resp is None:
            LOG.debug("mgt_sys_elem k2 response"
                      " is None in the host object")
            return None
        else:
            host_entry = mgt_sys_resp.entry
            msentry = msentry_wrapper.ManagedSystemEntryWrapper(host_entry)
            # Get state
            state = msentry.get_system_state()
            return state

    def get_shared_proc_pools(self, refresh=True):
        """
        Extract shared processor pool info from HMC.
        :refresh to request to hmc for fresh data
        :return dictionary containing shared processor pool information.
        e.g
        {'SharedPool01': {'id': 1,
                          'max_proc_units': 2,
                          'reserved_proc_units': 0.5,
                          'available_proc_units': 1.0},
        'SharedPool02': {'id': 2,
                          'max_proc_units': 0,
                          'reserved_proc_units': 0.0,
                          'available_proc_units': 0.0}
        }
        """
        with K2OperGuard(host=self) as k2oper:
            age = 0 if refresh else -1
            shr_prc_pools_resp = k2oper.read(k2const.MGT_SYS, rootId=self.uuid,
                                             childType=k2const.SHR_PRC_PL,
                                             age=age)
        proc_pools = {}
        for entry in shr_prc_pools_resp.feed.entries:
            shr_prc_pool_wrapper = ShrPrcPoolWrapper(entry)
            pool_id = shr_prc_pool_wrapper.get_pool_id()
            if pool_id != '0':
                proc_pools[shr_prc_pool_wrapper.get_pool_name()] = dict(
                    id=int(pool_id),
                    max_proc_units=float(
                        shr_prc_pool_wrapper.get_max_proc_units()),
                    reserved_proc_units=float(
                        shr_prc_pool_wrapper.get_curr_rsrv_proc_units()),
                    available_proc_units=float(
                        shr_prc_pool_wrapper.get_avail_proc_units()))
            else:
                proc_pools[shr_prc_pool_wrapper.get_pool_name()] = dict(
                    id=int(pool_id))
        return proc_pools

    def get_shared_proc_pools_id_name(self, refresh=False):
        """
        Extract shared processor pool id and name from HMC.
        :return dictionary containing shared processor pool id and name.
        e.g
        {1: 'SharedPool01',
         2: 'SharedPool02'
        }
        """
        with K2OperGuard(host=self) as k2oper:
            age = 0 if refresh else -1
            shr_prc_pools_resp = k2oper.read(k2const.MGT_SYS, rootId=self.uuid,
                                             childType=k2const.SHR_PRC_PL,
                                             age=age)
        proc_pools = {}
        for entry in shr_prc_pools_resp.feed.entries:
            shr_prc_pool_wrapper = ShrPrcPoolWrapper(entry)
            pool_id = int(shr_prc_pool_wrapper.get_pool_id())
            proc_pools[pool_id] = shr_prc_pool_wrapper.get_pool_name()
        return proc_pools

    def get_shared_proc_pool_id(self, spp_name, refresh=True):
        """
        Extract shared processor pool id from given name from HMC.
        :return id of the shared proc pool.
        e.g
        {1 if 'SharedPool01',
         2 if SharedPool02'
        """
        with K2OperGuard(host=self) as k2oper:
            age = 0 if refresh else -1
            shr_prc_pools_resp = k2oper.read(k2const.MGT_SYS, rootId=self.uuid,
                                             childType=k2const.SHR_PRC_PL,
                                             age=age)
        for entry in shr_prc_pools_resp.feed.findentries(
                './PoolName', spp_name):
            return entry.element.find('./PoolID').gettext()
        return None

    def _update_stats_for_non_master(self, entry, stats):
        """Helper method to update the stats if the Manager isn't the Master"""
        stats['maintenance_mode_needed'] = 'False'
        clsc, slave = 'IsClassicHMCManagement', 'IsNotPowerVMManagementMaster'
        is_classic = entry.element.findtext(clsc) != 'false'
        is_slave = entry.element.findtext(slave) == 'true'
        # If NovaLink is the master or the HMC is not in Classic Mode, then
        # we want to change the State and put the host in Maintenance Mode.
        # Note: We have an override configuration property for the classic
        # mode check to allow for development/test to mange in non-classic
        if is_slave or (not is_classic and not CONF.nonclassic_supported):
            stats['hypervisor_state'] = 'not master - use pvm to manage'
            stats['maintenance_mode_needed'] = 'True'


class HostList(object):

    '''
    Keep track of the hosts configured for this driver
    '''

    def __init__(self):
        self._list = {}
        pass

    def get(self, name):
        return self._list.get(name)

    def add(self, name, host_sys):
        self._list[name] = host_sys

    def get_list(self):
        return self._list

    def itervalues(self):
        return self._list.itervalues()


class Instance_to_PVM_UUID_Map(object):

    """
    Setup a cache of OS instance name to PowerVM UUID mappings.

    Initially we use the instance name to map to an LPAR name to establish
    the mapping.  Once we have a mapping then we can save it in a DB and
    retrieve it from there.
    """

    def __init__(self):
        self._cache = {}
        # use lowercase uuid for reverse lookup
        # NOTE: we don't need to do the same for name uuid lookup
        # since we want to query HMC using the UUID being returned
        self._reverse_cache = {}

    def rename(self, name, new_name, uid):
        # For a rename we need to patch in the new name
        # without screwing up the reverse map
        self._cache[new_name] = uid
        self._reverse_cache[uid.lower()] = new_name
        try:
            del self._cache[name]
        except Exception:
            pass

    def add(self, name, uid):
        self._cache[name] = uid
        self._reverse_cache[uid.lower()] = name

    def remove(self, name):
        try:
            uuid = self._cache.get(name)
            del self._cache[name]
            if uuid is not None:
                del self._reverse_cache[uuid.lower()]
        except Exception as e:
            LOG.error(_('Cannot remove %s from the UUID cache') % name)
            LOG.exception(_('Error removing instance from UUID cache: %s')
                          % six.text_type(e))

    def get_pvm_uuid(self, instance_name):
        """ Returns the UUID for the given Name from the UUID/Name Map """
        return self._cache.get(instance_name)

    def get_pvm_name(self, instance_uuid):
        """ Returns the Name for the given UUID from the UUID/Name Map """
        return self._reverse_cache.get(instance_uuid.lower())

    def load_cache_from_feed(self, feed, refresh_only=False):
        # Loop through getting the name and uuid
        for entry in feed.entries:
            uuid = entry.properties['id']
            name = entry.element.findtext('./PartitionName')
            if not refresh_only or not self.get_pvm_uuid(name):
                self.add(name, uuid)
            else:
                # if there is already an entry with the name, don't overwrite
                # db instance name has priority
                self._reverse_cache[uuid.lower()] = name

    def load_cache_from_db(self, db_instances):
        """ Update the UUID Cache with the Instance Info from the DB """
        for db_instance in db_instances:
            name = db_instance.get('name')
            uuid = db_instance.get('power_specs', {}).get('vm_uuid')
            # If the K2 UUID was stored in the DB, update the UUID Map for it
            if name is not None and uuid is not None:
                self.add(name, uuid)
