# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
import six
from oslo_log import log as logging

from oslo_config import cfg

from powervc_nova import logcall

from powervc_nova.virt.ibmpowervm.hmc.k2oprguard import K2OperGuard
from powervc_nova.virt.ibmpowervm.hmc import lpar

from powervc_nova.virt.ibmpowervm.hmc.k2 import constants as k2const
from powervc_nova.virt.ibmpowervm.hmc.k2.fc_map_processor \
    import FCMapProcessor
from powervc_nova.virt.ibmpowervm.hmc.k2.lparentry_wrapper \
    import LPAREntryWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2.viosentry_wrapper \
    import VIOSEntryWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2.scsi_map_processor \
    import SCSIMapProcessor
from powervc_nova.virt.ibmpowervm.hmc.k2.jobentry_wrapper \
    import JobEntryWrapper
from powervc_nova.virt.ibmpowervm.common import exception\
    as common_exception
from powervc_nova.virt.ibmpowervm.common.power_spec_utils \
    import POWER_SPECS_SYS_META_MAP
from oslo_concurrency.lockutils import synchronized
import base64
import xml.etree.ElementTree as ElementTree
from xml.etree.ElementTree import Element
from powervc_nova.virt.ibmpowervm.hmc.storage import vios_storage_info

from powervc_nova.virt.ibmpowervm.vif.hmc.vm_import_hmc \
    import ImportVMNetworkHMC
from powervc_nova.virt.ibmpowervm.common.constants \
    import CONNECTION_TYPE_NPIV, CONNECTION_TYPE_VSCSI
from powervc_nova import _

LOG = logging.getLogger(__name__)

REASONS = 'reasons'
STATUS = 'status'

SUPPORTED = 'supported'
NOT_SUPPORTED = 'not_supported'

CONF = cfg.CONF

NON_PV_ATTACHED = _('This virtual machine cannot be managed by '
                    'PowerVC because it has a non SAN-backed '
                    'physical volume attached to the storage '
                    'controller by virtual SCSI. '
                    'PowerVC manages NPIV-attached volumes and '
                    'SAN-backed physical volumes that are attached '
                    'by virtual SCSI. To manage this virtual machine '
                    'with PowerVC, remove all non SAN-backed '
                    'physical volumes that are attached by virtual '
                    'SCSI.')

INCOMP_STG_FOR_IMPORT = _('This virtual machine is not a candidate'
                          ' for management, either because it is not'
                          ' enabled for NPIV or because it has no'
                          ' volumes attached. PowerVC can only manage'
                          ' existing virtual machines that are using'
                          ' NPIV, SSP, or virtual SCSI'
                          ' and have at least a boot volume'
                          ' attached.')

PORT_NOT_ATTACHED = _('This virtual machine is not a candidate'
                      ' for management because the virtual'
                      ' machine has not been enabled for NPIV.'
                      ' It means that the virtual Fibre'
                      ' Channel client adapters on the virtual'
                      ' machine and virtual Fibre Channel'
                      ' server adapters on the Virtual I/O'
                      ' Server are mapped to a physical Fibre'
                      ' Channel port.')


class ImportVM(object):
    """
    Object to handle import vm functions for a given host
    """
    def __init__(self, host):
        self.host = host
        self.vios_id_uuid_lookup = {}
        # for performance purpose, we will keep track of VirtualIOServer
        # VirtualSCSIMappings and VirtualFibreChannelMappings we had parsed.
        # key is the vio partition id
        self._parsed_scsi_maps = {}
        self._parsed_fc_maps = {}

    @logcall
    def _refresh_vios_info(self, vioses_resp):
        """
        Reset and refresh vios information
        """
        self.vios_id_uuid_lookup.clear()
        self._parsed_scsi_maps.clear()
        self._parsed_fc_maps.clear()
        # construct the vios_id_uuid_lookup

        for entry in vioses_resp.feed.entries:
            vios_wrapper = VIOSEntryWrapper(entry)
            self.vios_id_uuid_lookup[vios_wrapper.get_vios_id()] = vios_wrapper
            scsi_map = SCSIMapProcessor()
            self._parsed_scsi_maps[vios_wrapper.get_vios_id()] = (
                scsi_map.parse_and_populate(entry.element))
            fc_map = FCMapProcessor()
            self._parsed_fc_maps[vios_wrapper.get_vios_id()] = (
                fc_map.parse_and_populate(entry.element))

    def get_vioses_for_lpar(self, lpar_id):
        """Helper method to get the VIOS's that this LPAR has Disks thru"""
        vios_ids = list()
        # Loop through the VIOS's seeing which ones have Disks for this LPAR
        for vios_id in self.vios_id_uuid_lookup:
            found_disk = False
            fc_storage = self._parsed_fc_maps.get(vios_id)
            scsi_storage = self._parsed_scsi_maps.get(vios_id)
            # See if there are any FC Disks for this LPAR on the VIOS
            if fc_storage is not None:
                disks = fc_storage.disk_info_by_lpar.get(lpar_id)
                if disks is not None and len(disks) > 0:
                    found_disk = True
            # See if there are any SCSI Disks for this LPAR on the VIOS
            if scsi_storage is not None:
                disks = scsi_storage.disk_info_by_lpar.get(lpar_id)
                if disks is not None and len(disks) > 0:
                    found_disk = True
            # Only add the VIOS if there was a Disk found matching this LPAR
            if found_disk:
                vios_ids.append(str(vios_id))
        return ','.join(vios_ids)

    def get_parsed_scsi_map_for_vios(self, vios_lpar_id):
        """
        Get VIOSStorageInfo from parsed VirtualSCSIMappings.
        Create one if there isn't a parsed result for the VIOS
        """
        data = self._parsed_scsi_maps.get(vios_lpar_id)
        if data is None:
            scsi_map = SCSIMapProcessor()
            k2entry = self.vios_id_uuid_lookup.get(vios_lpar_id)._k2entry
            self._parsed_scsi_maps[vios_lpar_id] = (
                scsi_map.parse_and_populate(k2entry.element))
        return self._parsed_scsi_maps[vios_lpar_id]

    def get_parsed_fc_map_for_vios(self, vios_lpar_id):
        """
        Get VIOSStorageInfo from parsed VirtualFibreChannelMappings.
        Create one if there isn't a parsed result for the VIOS
        """
        data = self._parsed_fc_maps.get(vios_lpar_id)
        if data is None:
            fc_map = FCMapProcessor()
            k2entry = self.vios_id_uuid_lookup.get(vios_lpar_id)._k2entry
            self._parsed_fc_maps[vios_lpar_id] = (
                fc_map.parse_and_populate(k2entry.element))
        return self._parsed_fc_maps[vios_lpar_id]

    def extract_scsi_storage(self, lpar_id):
        """
        Extract all the scsi storage from all the known VIOS for given
        lpar id.
        :returns:  list of VIOSDiskInfo
        """
        # iterate through the VIOS
        storages = []
        for scsi_storage in self._parsed_scsi_maps.itervalues():
            # extract storage info for the particular lpar
            disks = scsi_storage.disk_info_by_lpar.get(lpar_id)
            if disks is not None and len(disks) > 0:
                storages += disks
        return None if len(storages) == 0 else storages

    def extract_fc_storage(self, lpar_id):
        """
        Extract all the fc storage from all the known VIOS for given
        lpar id
        :returns:  list of VIOSDiskInfo
        """
        # iterate through the VIOS
        storages = []
        for fc_storage in self._parsed_fc_maps.itervalues():
            # extract storage info for the particular lpar
            disks = fc_storage.disk_info_by_lpar.get(lpar_id)
            if disks is not None and len(disks) > 0:
                storages += disks
        return None if len(storages) == 0 else storages

    @logcall
    def interrogate_lpar_storage(self, oper, lpar_uuid, lpar_id):
        """
        Iterate through the storages allocated for the partition to verify
        it's eligible for onboarding.
        """
        # find vios hosting scsi storage
        # vios_hosting_scsi = (
        #     vio_adapter_utils.get_vios_lpar_id_hosting_scsi(oper,
        #                                                     lpar_uuid))

        # # find vios hostings fc storage
        # vios_hosting_fc = (
        #     vio_adapter_utils.get_vios_lpar_id_hosting_fc(oper,
        #                                                   lpar_uuid))
        status = SUPPORTED
        reasons = []
        valid_scsi_cfg = True
        npiv_connector_data = None
        vscsi_connector_data = None
        # find the storages hosted by vios
        # find scsi storage first
        scsi_storage = self.extract_scsi_storage(lpar_id)
        if scsi_storage is not None:
            # make sure it's not using local or physical volume storage
            contains_lv = vios_storage_info.contains_storage_types(
                scsi_storage, [vios_storage_info.VDISK_TYPE])
            valid_scsi_cfg = False if contains_lv else valid_scsi_cfg
            LOG.debug('scsi_storage %s' % scsi_storage)

        if not valid_scsi_cfg:
            status = NOT_SUPPORTED
            reasons.append(NON_PV_ATTACHED)
        else:
            # find fc storage
            fc_storage = self.extract_fc_storage(lpar_id)
            LOG.debug('fc_storage_data %s' % fc_storage)
            npiv_connector_data = self._convert_to_volume_connector_info(
                fc_storage, CONNECTION_TYPE_NPIV)
            vscsi_connector_data = self._convert_to_volume_connector_info(
                scsi_storage, CONNECTION_TYPE_VSCSI)

            combined_data = []
            if scsi_storage is None and fc_storage is None:
                status = NOT_SUPPORTED
                reasons.append(INCOMP_STG_FOR_IMPORT)
            elif scsi_storage is None:
                # make sure it's mapped to physical port
                if (npiv_connector_data and
                        npiv_connector_data.get('phy_to_virt_initiators')):
                    combined_data = fc_storage
                else:
                    status = NOT_SUPPORTED
                    reasons.append(PORT_NOT_ATTACHED)
            elif fc_storage is None:
                # check for local and pv storage
                combined_data = scsi_storage
            else:
                # check for local and pv storage
                combined_data += (scsi_storage + fc_storage)

            # extract disks from combined_data
            if status == SUPPORTED:
                disks = vios_storage_info.guess_boot_order(combined_data)
                LOG.debug('disk orders for lpar_id %d %s' % (lpar_id, disks))
                # the output from guess_boot_order will exclude
                # virtual media
                # so, we will make sure that disks or connector_data
                # is not empty
                # Also, add another check for whether PowerVC supports SSP
                if ((npiv_connector_data and fc_storage) or
                        (disks and vscsi_connector_data) or
                        (disks and CONF.ibmpowervm_register_ssps)):
                    connector_data = {}
                    connector_data['vscsi_connector_data'] =  \
                        vscsi_connector_data
                    connector_data['npiv_connector_data'] = npiv_connector_data
                    return {STATUS: SUPPORTED,
                            'data': disks,
                            'volume_connector': connector_data}
                else:
                    status = NOT_SUPPORTED
                    reasons.append(INCOMP_STG_FOR_IMPORT)

        return {STATUS: status,
                REASONS: reasons}

    @logcall
    def interrogate_lpar_misc(self, lpar_wrapper):
        """
        Check various partition properties and resources to make sure the
        configuration is supported by PowerVC.
        - check for PartitionType - Support AIX/Linux and IBM i
        - check for SharedProcPoolID - Only default pool
        - check for SharedMemmoryEnabled - Make sure it's false
        """
        status = SUPPORTED
        reasons = []
        lpar_type = lpar_wrapper.get_lpar_type()
        if lpar_type != 'AIX/Linux' and lpar_type != 'OS400':
            status = NOT_SUPPORTED
            reasons.append(_('This virtual machine is not a candidate for '
                             'management because you can manage only '
                             'virtual machines that are configured for AIX,'
                             ' Linux, or IBM i operating systems.'))

        shared_mem_enabled = lpar_wrapper.get_current_mem_share_enabled()
        if shared_mem_enabled is not None and shared_mem_enabled:
            status = NOT_SUPPORTED
            reasons.append(_('This virtual machine is not a candidate for '
                             'management because it is configured to use a '
                             'shared processor memory pool other than the '
                             'default.  You can only manage virtual machines '
                             'that use the default shared pool.'))
        return {STATUS: status,
                REASONS: reasons}

    @logcall
    def discover_info_for_onboarding(self, k2opr, vif_driver, entry):
        """
        Information needed for presenting users a list of partitions to onboard
        instances into PowerVC
        """
        lpar_wrapper = LPAREntryWrapper(entry)
        lpar_uuid = lpar_wrapper.get_uuid()

        state = lpar_wrapper.get_lpar_state()
        lpar_id = lpar_wrapper.get_lpar_id()
        name = lpar_wrapper.get_lpar_name()

        # call to network & storage to update the
        # status field
        # call network component to see whether lpar is supported
        status = 'supported'
        reasons = []

        # check for various lpar info
        lpar_result = self.interrogate_lpar_misc(lpar_wrapper)
        if lpar_result.get(STATUS) == NOT_SUPPORTED:
            status = NOT_SUPPORTED
            reasons += lpar_result.get(REASONS)

        LOG.debug('network import')
        net_prep_steps, vif_reasons, adapters = \
            vif_driver.interrogate_vm(lpar_uuid)
        if vif_reasons is not None and len(vif_reasons) > 0:
            status = NOT_SUPPORTED
            reasons += vif_reasons
        LOG.debug('network import end %s' % vif_reasons)

        # check for storage
        storage_result = self.interrogate_lpar_storage(
            k2opr, lpar_uuid, lpar_id)
        if storage_result.get(STATUS) == NOT_SUPPORTED:
            status = NOT_SUPPORTED
            reasons += storage_result.get(REASONS)

        support = {'status': status}
        if status == NOT_SUPPORTED:
            support['reasons'] = reasons

        power_specs = {'vm_id': lpar_id,
                       'vm_uuid': lpar_uuid,
                       'vm_name': name}

        lpar_info = {'uuid': lpar_uuid.lower(),
                     'name': name,
                     'state': state,
                     'power_specs': power_specs,
                     'support': support,
                     'adapters': adapters}
        return lpar_info

    @logcall
    def extract_info_for_onboarding(self, k2opr, vif_driver, entry):
        """
        Extract information needed by onboarding a partition into PowerVC
        """
        lpar_wrapper = LPAREntryWrapper(entry)
        lpar_uuid = lpar_wrapper.get_uuid()

        state = lpar_wrapper.get_lpar_state()
        lpar_id = lpar_wrapper.get_lpar_id()
        lpar_type = lpar_wrapper.get_lpar_type()
        lpar_name = lpar_wrapper.get_lpar_name()
        root_gb, ephemeral_gb = 0, 0
        memory_mb = lpar_wrapper.get_current_mem()
        proc_mode = lpar_wrapper.get_current_proc_mode()

        if proc_mode == 'true':
            vcpu = lpar_wrapper.get_current_procs()
            proc_units = lpar_wrapper.get_current_procs()
        else:
            vcpu = lpar_wrapper.get_current_vcpus()
            proc_units = lpar_wrapper.get_current_proc_units()

        # this need to be inserted into flavor
        avail_priority = lpar_wrapper.get_avail_priority()
        srr_cap = lpar_wrapper.get_srr_capability()
        srr_state = lpar_wrapper.get_srr_state()

        # Need to update to include volume and disk size info
        volumes = []
        ports = []
        disks = None
        vol_connector = None
        # call network component to  update the ports field
        vif_ports = vif_driver.list_ports(lpar_uuid)
        for port in vif_ports:
            # extract the info we needed
            ports.append({'mac_address': port.get('mac_address'),
                          'status': port.get('status'),
                          'vswitch': port.get('vswitch'),
                          'segmentation_id':
                          port.get('provider:segmentation_id'),
                          'physical_network':
                          port.get('physical_network')})
        # validate & extract storage info
        storage_result = self.interrogate_lpar_storage(
            k2opr, lpar_uuid, lpar_id)
        if storage_result.get(STATUS) == SUPPORTED:
            # extract the disk info
            disks = storage_result.get('data')
            vol_connector = storage_result.get('volume_connector')
            vol_connector['host'] = lpar_name

        vios_ids = self.get_vioses_for_lpar(lpar_id)
        vios_ids = vios_ids.split(',')
        # parse through the disk information
        if disks is not None and len(disks) > 0:
            # the disk is alerady sorted, grab the first disk
            # to populate the root_gb and ephemeral_gb
            root_gb = disks[0].storage_info.get('disk_size')
            root_gb = 0 if root_gb is None else root_gb
            ephemeral_gb = 0
            for i in xrange(len(disks)):
                storage_info = disks[i].storage_info
                k2udid = disks[i].storage_udid
                vol = {'name': disks[i].backing_device,
                       'unique_device_id': disks[i].storage_uid,
                       'k2udid': disks[i].storage_udid
                       }
                if k2udid is not None:
                    # get the wwpns of vios
                    vios_wwpns = {}
                    for vios in vios_ids:
                        vios_wrapper = self.vios_id_uuid_lookup[int(vios)]
                        if vios_wrapper is None:
                            LOG.debug("Vioswrapper not found for vios %s "
                                      % vios)
                            continue
                        wwpn_lst = vios_wrapper.get_vios_wwpns()
                        vios_wwpns[vios] = wwpn_lst
                    vol['vios_wwpns'] = vios_wwpns
                    pg83 = self._get_disk_inventory(vios_ids, k2udid)
                    if pg83 is not None:
                        vol['pg83NAA'] = pg83
                if storage_info.get('wwpns') is not None:
                    vol['wwpns'] = storage_info.get('wwpns')
                volumes.append(vol)

        # Call the Wrapper to retrieve the PowerSpecs for the LPAR
        power_specs = lpar.get_power_specs(entry)
        power_specs['vios_ids'] = self.get_vioses_for_lpar(lpar_id)
        power_specs['lpar_type'] = lpar_type

        spp_id = power_specs.get('shared_proc_pool_id')
        if spp_id is not None:
            shared_proc_pools = self.host.get_shared_proc_pools_id_name()
            power_specs['shared_proc_pool_name'] = \
                shared_proc_pools.get(spp_id)

        # duplicate the system_metadata
        sys_meta = {}
        for key, val in six.iteritems(power_specs):
            if key in POWER_SPECS_SYS_META_MAP:
                sys_meta[POWER_SPECS_SYS_META_MAP[key]] = str(val)
        lpar_info = {'uuid': lpar_uuid.lower(),
                     'name': lpar_name,
                     'state': state,
                     'architecture': 'ppc64',
                     'power_specs': power_specs,
                     'system_metadata': sys_meta,
                     'vcpus': vcpu,
                     'memory_mb': memory_mb,
                     'root_gb': root_gb,
                     'ephemeral_gb': ephemeral_gb,
                     'volumes': volumes,
                     'ports': ports,
                     'powervm:availability_priority': avail_priority,
                     'volume_connector': vol_connector}
        return lpar_info

    @logcall
    def list_lpar_for_onboard(self, context):
        """
        Generate the data needed by the discovery_driver.discover_instances
        for onboarding
        """
        ret_list = []

        # dictionary with PartitionID as key and onboarding dictionary as
        # value
        lpar_data_dict = {}
        # Force a refresh of the LPAR and VIOS Cache before on-boarding
        with K2OperGuard(host=self.host) as k2opr:
            # use K2 level caching
            vioses_resp = k2opr.read(k2const.MGT_SYS, rootId=self.host.uuid,
                                     childType=k2const.VIOS,
                                     xag=k2const.ALL_VIOS_EXT_PROPS)
            self._refresh_vios_info(vioses_resp)
            lpars_resp = k2opr.read(k2const.MGT_SYS, rootId=self.host.uuid,
                                    childType=k2const.LPAR)
            # Construct the VIF Import Driver to use to get the Network info
            vif_driver = ImportVMNetworkHMC(k2opr, k2resp=lpars_resp)
            # Loop through the list of LPAR's retrieving the on-boarding info
            for entry in lpars_resp.feed.entries:
                lpar_info = self.discover_info_for_onboarding(
                    k2opr, vif_driver, entry)
                lpar_id = lpar_info['power_specs']['vm_id']
                lpar_data_dict[lpar_id] = lpar_info

        # Transfer all the values from the LPAR Dictionary to the LPAR List
        for val in lpar_data_dict.itervalues():
            ret_list.append(val)
        return ret_list

    @logcall
    def query_onboarding_instances(self, context, instances):
        """
        Generate the data needed by the discovery_driver.query_instances
        for onboarding.
        """
        ret_list = []
        lpar_data_dict = {}
        # Construct a dictionary for the Name to UUID lookup
        name_uuid_lookup = dict((inst['name'], inst['uuid'])
                                for inst in instances)
        # Force a refresh of the LPAR and VIOS Cache before on-boarding
        with K2OperGuard(host=self.host) as k2opr:
            # use K2 level caching
            vioses_resp = k2opr.read(k2const.MGT_SYS, rootId=self.host.uuid,
                                     childType=k2const.VIOS, age=0,
                                     xag=k2const.ALL_VIOS_EXT_PROPS)
            self._refresh_vios_info(vioses_resp)
            lpars_resp = k2opr.read(k2const.MGT_SYS, rootId=self.host.uuid,
                                    childType=k2const.LPAR, age=0)
            # Construct the VIF Import Driver to use to get the Network info
            vif_driver = ImportVMNetworkHMC(k2opr, k2resp=lpars_resp)

            for entry in lpars_resp.feed.entries:
                lpar_wrapper = LPAREntryWrapper(entry)
                name = lpar_wrapper.get_lpar_name()
                if name not in name_uuid_lookup:
                    continue
                lpar_info = self.extract_info_for_onboarding(
                    k2opr, vif_driver, entry)
                lpar_id = lpar_info['power_specs']['vm_id']
                prof_path = lpar_wrapper.get_lpar_prof_path()
                LOG.debug('prof_path %s' % prof_path)

                prof_resp = k2opr.readbyhref(prof_path)
                lpar_flavor = {}
                if prof_resp is not None:
                    lpar_flavor = lpar.extract_flavor_from_prof(
                        prof_resp.entry)

                avail_prior_key = 'powervm:availability_priority'
                lpar_flavor[avail_prior_key] = lpar_info.pop(avail_prior_key)
                lpar_info['flavor'] = lpar_flavor
                lpar_info['name'] = name
                lpar_data_dict[lpar_id] = lpar_info

        # Transfer all the values from the LPAR Dictionary to the LPAR List
        for val in lpar_data_dict.itervalues():
            ret_list.append(val)
        return ret_list

    def _convert_to_volume_connector_info(self, data, connection_type):
        """
        Convert storage data to volume connector information that can be
        consumed by cinder to clean up zoning setup.
        """
        vol_conn = {}
        vol_conn['connection_type'] = [connection_type]
        if data:
            # iterate through the storage info
            stg_info = [entry.storage_info for entry in data]
            if stg_info:
                for stg_entry in stg_info:
                    # extract client wwpns
                    wwpns = stg_entry.get('wwpns')
                    if wwpns:
                        if vol_conn.get('wwpns') is None:
                            vol_conn['wwpns'] = []
                        vol_conn['wwpns'] += wwpns
                        if vol_conn.get('phy_to_virt_initiators') is None:
                            vol_conn['phy_to_virt_initiators'] = {}
                        # extract port wwpn
                        port_wwpn = stg_entry.get('port_wwpn')
                        if port_wwpn:
                            vol_conn['phy_to_virt_initiators'][port_wwpn] = (
                                wwpns)
        return vol_conn

    @logcall
    def _invoke_lua_inventory(self, vios_uuid, vios_name, inputXML):
        @synchronized('stg-mapping-lock' + vios_uuid)
        def _call_lua_inventory(vios_uuid, vios_name, inputXML):
            pg83_naa = None
            with K2OperGuard(host=self.host) as oper:
                try:
                    # Get the K2 response for the LuaRecovery command
                    k2resp = oper.vios_read(
                        k2const.VIOS, vios_uuid,
                        suffixType=k2const.SUFFIX_TYPE_DO,
                        suffixParm=k2const.SUFFIX_PARM_LUA_RECOVERY,
                        vios_name=vios_name)
                    job_wrapper = JobEntryWrapper(k2resp.entry)
                    job_parms = None
                    job_parms = [job_wrapper.create_job_parameter(
                        'inputXML', inputXML)]
                    LOG.info(_("INPUT XML:%s"), inputXML)
                    # Execute the K2 LUA Inventory call.
                    job_wrapper.run_job(oper, vios_uuid, job_parms=job_parms)
                    result = job_wrapper.get_job_results_as_dict()
                    pg83_naa = self._process_lua_inv_result(result)
                    if pg83_naa is not None:
                        return pg83_naa
                except common_exception.IBMPowerVMJobRequestFailed as error:
                    LOG.warn(_('Error: %s') % error)
            return pg83_naa

        return _call_lua_inventory(vios_uuid, vios_name, inputXML)

    def _get_disk_inventory(self, vios_ids, k2udid):
        """
        Return the page 83 NAA for the given K2 id
        To retrieve the NAA information we need to make the
        VIOS passthrough LUA inventory K2 call. Based on the VIOS
        k2udid passed VIOS will returned the Page 83 NAA descriptor
        information.
        We perform the below steps in this function
        a. Create the INPUT XML for the LUA Inventory call
        b. Fire the K2 LUA Inventory call
        c. Process the K2 Response Output XML to fetch the NAA Descriptor.

        :param vios_ids: VIOS ids information of the attached hdisk for
        which we need the descriptor information. We can have the same
        disk discovered on multiple VIOSes. In that case we process each
        vios and break if the NAA descriptor is found. Since NAA is same
        for a particular disk across all VIOSs.

        :param k2udid: This is the K2 Unique ID descriptor which is generated
        by VIOS when the device is discovered on the VIOS during LUA discovery.
        VIOS uses the same K2udid to fetch the pg83 NAA information.
        #Todo: In future we may not need this call since VIOS will be storing
        returning the Pg83 NAA information when we fetch the VIOS
        Extended properties.
        """

        pg83_naa = None
        for vios in vios_ids:
            idoc = self._create_lua_query_inv_xml(k2udid)
            inputXML = ElementTree.tostring(idoc)
            vios_wrapper = self.vios_id_uuid_lookup[int(vios)]
            if vios_wrapper is None:
                LOG.debug("Vioswrapper not found for vios %s " % vios)
                continue
            vios_name = vios_wrapper.get_vios_name()
            vios_uuid = vios_wrapper.get_vios_uuid()
            pg83_naa = self._invoke_lua_inventory(vios_uuid,
                                                  vios_name, inputXML)
            if pg83_naa is not None:
                return pg83_naa

            LOG.warn(_("PG83 NAA pg83_naa= %(naa)s information cannot be "
                       "retrieved for disk %(id)s"), {'naa': pg83_naa,
                                                      'id': k2udid})

        return pg83_naa

    @logcall
    def _create_lua_query_inv_xml(self, k2udid):
        """Create an lua query inventory input XML
           Sample XML
           <VIO xmlns="http://ausgsa.austin.ibm.com/projects/v/vios/schema
           /1.21 version="1.21"><Request action_str="QUERY_INVENTORY">
           <InventoryRequest inventoryType="base" >
           <VioTypeFilter type="PV" />
           <VioUdidFilter udid="01M0lCTTE4MTggICAgICBGQVN0VDYwMEEwQjg
            wMDA2RTQ0ODIwMDAwQjIxRDUzRTBCODg0" />
           </InventoryRequest></Request></VIO>
           Note: Tag inventoryType should be "base" for pg83NAA
        """
        xmlns = "http://ausgsa.austin.ibm.com/projects/v/vios/schema/1.21"
        root = Element('VIO', attrib={'xmlns': xmlns, 'version': "1.21"})
        request = Element('Request', attrib={'action_str': "QUERY_INVENTORY"})
        inventory = Element('InventoryRequest',
                            attrib={'inventoryType': "base"})
        VioTypeFilter = Element('VioTypeFilter', attrib={'type': "PV"})
        VioUdidFilter = Element('VioUdidFilter', attrib={'udid': k2udid})
        inventory.append(VioTypeFilter)
        inventory.append(VioUdidFilter)
        request.append(inventory)
        root.append(request)
        return root

    @logcall
    def _process_lua_inv_result(self, result):
        """ Processes the Output XML returned by LUARecovery
        SAMPLE Response
        <VIO version="1.21" xmlns="http://ausgsa.austin.ibm.com/projects
        /v/vios/schema/1.21"><Response><InventoryResponse viosId=
        "8233-E8B02061AAEP" sequence="1" inventoryType="base"
        eventLogOn="true">
        <PhysicalVolume udid="01M0lCTTE4MTggICAgICBGQVN0VDYw
        MEEwQjgwMDA2RTQ0ODIwMDAwM0M1MjUzRUIzMDdC" name="hdisk2"
        description="MPIO DS5100/5300 Disk"><PhysicalVolume_base
        capacity="12288" locationCode="U5877.001.0081004-P1-C3-T2-
        W205600A0B86E4A5A-L17000000000000" unique_id="3E213600A0B80
        006E448200003C5253EB307B0F1818  FAStT03IBMfcp" vendor_id="IBM"
        product_id="1818   in_use="no" state="ONLINE"
        SCSIReserve="single_path" MpioAlgorithm="fail_over"
        reserveKey="none"
        descriptor="NjAwQTBCODAwMDZFNDQ4MjAwMDAzQzUyNTNFQjMwN0I="
        desType="NAA"><southbound udid="1cU5877.001.0081004-P1-C3-T1"/>
        <southbound udid="1cU5877.001.0081004-P1-C3-T2"/>
        </PhysicalVolume_base></PhysicalVolume></InventoryResponse></Response>

        Note: We are processing the descriptor only if the desType="NAA"
        """
        naa = None
        result = result['OutputXML']
        LOG.debug("_process_lua_inv_result OutputXML: %s" % result)
        if result is not None:
            idx = result.find('</VIO>')
            res = result.find('<Response')
            result = result[res:idx]
            LOG.debug("Output Response updated result : %s" % result)
            root = ElementTree.fromstring(result)
            for pv in root.getiterator('PhysicalVolume_base'):
                if ('desType' in pv.attrib.keys() and
                        pv.attrib['desType'] == "NAA"):
                    # Don't fail if we can't decode. Go on
                    # without pg83. VIOS must have passed
                    # invalid data. Won't be able to use pg83
                    # for future vscsi lua_discoveries.
                    try:
                        pg83_naa = pv.attrib['descriptor']
                        naa = base64.b64decode(pg83_naa)
                        LOG.info("PG83 NAA Found: %(naa)s and"
                                 "decode %(pg83)s" % {'naa': pg83_naa,
                                                      'pg83': naa})
                    except Exception as ex:
                        LOG.info(_("Failed decode pg83 %s" % pv))
                        LOG.exception(ex)
                elif 'desType' in pv.attrib.keys():
                    LOG.debug("Available desTypes %s" % pv.attrib['desType'])
        return naa
