# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import json
import six

from oslo_log import log as logging
from powervc_nova import _
from powervc_nova.virt.ibmpowervm.hmc.k2 import monitor_pref_processor

MANAGED_SYSTEM_TYPE = 'ManagedSystem'
RAW_METRICS_TYPE = 'RawMetrics'
LONG_TERM_MONITOR_TYPE = 'LongTermMonitor'
PCM_SERVICE = 'pcm'

LOG = logging.getLogger(__name__)


class CpuUtilProcessor(object):
    """
    This class is used to collect and calculate the CPU utilization
    of all LPARs (VMs) on a given CEC (host).
    """
    def __init__(self):
        # This represents the current state of cycles spent on the system.
        # These are used to figure out usage statistics.  As such, they are
        # tied to the start of the nova compute process.
        #
        # - idle: Total idle cycles on the compute host.
        # - kernel: How many cycles the hypervisor has consumed.  Not a direct
        #           analogy to KVM
        # - user: The amount of time spent by the VM's themselves.
        # - iowait: Not used in PowerVM, but needed for nova.
        # - frequency: The CPU frequency
        self.host_total_data = {'idle': 0, 'kernel': 0,
                                'user': 0, 'iowait': 0,
                                'frequency': 0}

    def get_t0_t1_sample_pair(self, oper, cec_uuid):
        try:
            ltm_response = CpuUtilProcessor._get_managed_system_ltm(oper,
                                                                    cec_uuid)

            if ltm_response is None:
                return None, None

            phyp_links = []
            if ltm_response.status == 200:
                for entry in ltm_response.feed.entries:
                    category = entry.properties.get('category', '')

                    if category == 'phyp':
                        phyp_links.append(entry.properties.get('link', ''))
            elif ltm_response.status == 204:
                # Handle Out-of-band HMC restarts.
                LOG.info(_('Response status was 204. This could be '
                           'due to the HMC being restarted, which causes '
                           'the LTM collection to be disabled on the host; '
                           'now re-enabling LTM collection for host %s '
                           'on the HMC.') % str(cec_uuid))
                mon_pref_processor = (
                    monitor_pref_processor.MonitorPrefProcessor())
                mon_pref_processor.enable_ltm(oper, cec_uuid)
                return None, None
            else:
                LOG.info(_('Response status was not 200; instead, it was: %s.')
                         % str(ltm_response.status))
                return None, None

            # sort by time stamp
            phyp_links.sort()
            num_links = len(phyp_links)
            if num_links > 1:
                data_t0_link = phyp_links[num_links - 2]
                data_t1_link = phyp_links[num_links - 1]
                data_t0_json = json.loads(oper.readbyhref(data_t0_link).body)
                data_t1_json = json.loads(oper.readbyhref(data_t1_link).body)
                return data_t0_json, data_t1_json
            else:
                LOG.info(_('The number of PHYP entries is not > 1; '
                           'instead, it was: %s.') % str(num_links))
                return None, None
        except Exception:
            LOG.error(_('Error processing cpu_util data.'), exc_info=True)

    def get_cpu_utilization_data(self, oper, cec_uuid):
        cpu_util_data = {}

        try:
            data_t0_json, data_t1_json = (
                self.get_t0_t1_sample_pair(oper, cec_uuid))
            if not all((data_t0_json, data_t1_json)):
                return cpu_util_data
            data_lpar_t0 = CpuUtilProcessor.\
                _parse_ltm_raw_data(data_t0_json, is_vios=False,
                                    is_host=False)
            data_lpar_t1 = CpuUtilProcessor.\
                _parse_ltm_raw_data(data_t1_json, is_vios=False,
                                    is_host=False)

            cpu_util_data_subResult = CpuUtilProcessor.\
                _calculate_result_pair(data_lpar_t0, data_lpar_t1,
                                       is_host=False)
            cpu_util_data.update(cpu_util_data_subResult)
        except Exception:
            LOG.error(_('Error processing cpu_util data.'), exc_info=True)

        return cpu_util_data

    @staticmethod
    def _calculate_result_pair(data_lpar_t0, data_lpar_t1, is_host=False):
        cpu_util_data = {}
        for key, value_t0 in six.iteritems(data_lpar_t0):
            value_t1 = data_lpar_t1.get(key, None)
            # data for the LPAR must be at both collection points
            if value_t1 is not None:
                cpu_util = \
                    CpuUtilProcessor._calc_cpu_utilization(value_t0,
                                                           value_t1,
                                                           is_host)
                # avoid negative values for CPU util
                if cpu_util < 0:
                    cpu_util = 0
                cpu_util_data[value_t0['uuid']] = float(cpu_util)
        return cpu_util_data

    @staticmethod
    def _get_managed_system_ltm(operator, cec_uuid):
        try:
            raw_data_response = operator.read(MANAGED_SYSTEM_TYPE,
                                              cec_uuid,
                                              RAW_METRICS_TYPE,
                                              LONG_TERM_MONITOR_TYPE,
                                              service=PCM_SERVICE)
            return raw_data_response
        except Exception:
            # Removing exc_info flag so that Logger does not print the
            # complete exception stack. For RTC defect 10175
            LOG.error(_("Unable to get managed system LTM property for "
                        "uuid: %s") % str(cec_uuid))
            return None

    @staticmethod
    def _parse_ltm_raw_data(data, is_vios=False, is_host=False):
        data_timestamp = data['systemUtil']['utilSample']['timeStamp']
        if is_host:
            time_based_cycles = \
                data['systemUtil']['utilSample']['timeBasedCycles']
        if is_vios:
            utils = data['systemUtil']['utilSample']['viosUtil']
        else:
            utils = data['systemUtil']['utilSample']['lparsUtil']
        results = {}
        for util in utils:
            uncapped_cycles = util['processor']['utilizedUnCappedProcCycles']
            capped_cycles = util['processor']['utilizedCappedProcCycles']
            if is_host and (uncapped_cycles == 0 and capped_cycles == 0):
                # LPAR has just been migrated. Ignore.
                continue
            result = {}
            result['timestamp'] = data_timestamp
            result['uuid'] = util['uuid']
            result['name'] = util['name']
            if is_host:
                # For host CPU utilization, timeBasedCycles is used for
                # calculation.
                result['timeBasedCycles'] = time_based_cycles
            else:
                # Extract entitledProcCycles. This is the number of cycles
                # LPAR is entitled to.
                result['entitledProcCycles'] = \
                    util['processor']['entitledProcCycles']
            result['utilizedCappedProcCycles'] = capped_cycles
            result['utilizedUnCappedProcCycles'] = uncapped_cycles
            result['idleProcCycles'] = \
                util['processor']['idleProcCycles']
            # Extract donatedProcCycles. This is the number of cycles
            # LPAR can donate its cycles to another LPAR.
            result['donatedProcCycles'] = \
                util['processor']['donatedProcCycles']

            results[result['uuid']] = result

        return results

    @staticmethod
    def _calc_cpu_utilization(data_t0, data_t1, is_host=False):
        # The formula for CPU utilization is kept same  as proposed
        # Ceilometer's powervm inspector :
        # https://review.openstack.org/#/c/197720/6/ceilometer_powervm/
        # compute/virt/powervm/inspector.py

        if is_host:
            cs_t0 = data_t0['timeBasedCycles']
            cs_t1 = data_t1['timeBasedCycles']
        else:
            cs_t0 = data_t0['entitledProcCycles']
            cs_t1 = data_t1['entitledProcCycles']

        cycles = float(cs_t1 - cs_t0)
        if cycles == 0:
            return 0

        cap_cs_t0 = data_t0['utilizedCappedProcCycles']
        uncap_cs_t0 = data_t0['utilizedUnCappedProcCycles']
        idle_cs_t0 = data_t0['idleProcCycles']
        donated_cs_t0 = data_t0['donatedProcCycles']

        cap_cs_t1 = data_t1['utilizedCappedProcCycles']
        uncap_cs_t1 = data_t1['utilizedUnCappedProcCycles']
        idle_cs_t1 = data_t1['idleProcCycles']
        donated_cs_t1 = data_t1['donatedProcCycles']

        # This formula is for both dedicated and shared CPUs
        #
        # For shared CPUs, the idle CPU is always 0
        #
        # For dedicated CPU, idle_cpu is reported, so that will
        # be greater than zero and we have to deduct that when
        # determining CPU Utilization.
        #
        # When the CPU mode is dedicated and share_idle_procs is set,
        # donated CPU could be greater than 0.  If it is set to
        # not shared, then it will be 0.
        cpu_cycles_used = float((cap_cs_t1 - cap_cs_t0) +
                                (uncap_cs_t1 - uncap_cs_t0) -
                                (donated_cs_t1 - donated_cs_t0) -
                                (idle_cs_t1 - idle_cs_t0))
        cpu_util = cpu_cycles_used / cycles
        return cpu_util

    @staticmethod
    def _get_firware_utilization(data_t0_json, data_t1_json):
        fw = data_t0_json['systemUtil']['utilSample']['systemFirmware']
        # Return 0 if firmware 'utilizedProcCycles' is not part of K2 response.
        # Firmware's cycles are only reported from CEC 780. So assume 0 for
        # all other releases as this is mostly negligible value.
        fw_cycles_t0 = fw.get('utilizedProcCycles', 0)
        fw = data_t1_json['systemUtil']['utilSample']['systemFirmware']
        fw_cycles_t1 = fw.get('utilizedProcCycles', 0)
        time_based_cycles_t0 = \
            data_t0_json['systemUtil']['utilSample']['timeBasedCycles']
        time_based_cycles_t1 = \
            data_t1_json['systemUtil']['utilSample']['timeBasedCycles']
        cycles = float(time_based_cycles_t1 - time_based_cycles_t0)
        fw_cycles_used = float(fw_cycles_t1 - fw_cycles_t0)
        fw_util = fw_cycles_used / cycles
        return {'firmware': fw_util}

    def get_host_utilization(self, oper, cec_uuid):
        '''Evaluates and returns host's CPU utilization data.
        The returned stats is in line with what Nova monitor framework expects.
        '''
        try:
            data_t0_json, data_t1_json = (
                self.get_t0_t1_sample_pair(oper, cec_uuid))
            if not all((data_t0_json, data_t1_json)):
                return None
            # Check that total activated cores at T0 and T1 are same.
            util_data_t0 = data_t0_json['systemUtil']['utilSample']
            total_cores_t0 = util_data_t0['processor']['configurableProcUnits']
            util_data_t1 = data_t1_json['systemUtil']['utilSample']
            total_cores_t1 = util_data_t1['processor']['configurableProcUnits']
            # Get firmware utilization data
            fw_data_t0 = (
                data_t0_json['systemUtil']['utilSample']['systemFirmware'])
            fw_data_t1 = (
                data_t1_json['systemUtil']['utilSample']['systemFirmware'])
            # Get cycles per core at T1 and T0
            time_based_cycles_t0 = (
                data_t0_json['systemUtil']['utilSample']['timeBasedCycles'])
            time_based_cycles_t1 = (
                data_t1_json['systemUtil']['utilSample']['timeBasedCycles'])
            if total_cores_t0 != total_cores_t1:
                LOG.info(_("The total number of activated cores at T0 "
                           "(%(total_cores_t0)d) is not the same as the "
                           "total number of activated cores at T1 "
                           "(%(total_cores_t1)d); therefore skipping host "
                           "CPU utilization evaluation for this interval."
                           % {'total_cores_t0': total_cores_t0,
                              'total_cores_t1': total_cores_t1}))
                return None
            # Get the delta cycles w.r.t user, kernel and total.
            delta_user_cycles = self._gather_delta_user_cycles(data_t0_json,
                                                               data_t1_json)
            delta_firmware_cycles = self._calculate_delta_firmware_cycles(
                fw_data_t0, fw_data_t1)
            delta_total_cycles = self._calculate_delta_total_cycles(
                time_based_cycles_t0, time_based_cycles_t1, total_cores_t1)
            if delta_user_cycles + delta_firmware_cycles > delta_total_cycles:
                LOG.warning(_(
                    "Host CPU Metrics determined that the total cycles "
                    "reported was less than the used cycles. This indicates "
                    "an issue with the PCM data. Please investigate the "
                    "results.\n"
                    "Total Delta Cycles: %(tot_cycles)d\n"
                    "User Delta Cycles: %(user_cycles)d\n"
                    "Firmware Delta Cycles: %(fw_cycles)d"
                    % {'tot_cycles': delta_total_cycles,
                       'fw_cycles': delta_firmware_cycles,
                       'user_cycles': delta_user_cycles}))
                delta_total_cycles = delta_user_cycles + delta_firmware_cycles
            delta_idle_cycles = (
                delta_total_cycles - delta_user_cycles - delta_firmware_cycles)

            LOG.debug("Delta User cycles : %(delta_user)d\n"
                      "Delta Firmware cycles : %(delta_firmware)d\n"
                      "Total activated cores : %(active_cores)d\n"
                      "Delta Total cycles : %(delta_total)d"
                      % {'delta_user': delta_user_cycles,
                         'delta_firmware': delta_firmware_cycles,
                         'active_cores': total_cores_t1,
                         'delta_total': delta_total_cycles})
            self.host_total_data['user'] += delta_user_cycles
            self.host_total_data['kernel'] += delta_firmware_cycles
            self.host_total_data['idle'] += delta_idle_cycles
        except Exception:
            LOG.error(_('Error processing cpu_util data.'), exc_info=True)

        return self.host_total_data

    def _gather_delta_user_cycles(self, data_t0_json, data_t1_json):
        """The estimated delta user cycles.

        The sample data includes information about how much CPU has been used
        by workloads and the Virtual I/O Servers.  There is not one global
        counter that can be used to obtain the CPU spent cycles.

        This method will calculate the delta utilized cycles by all
        Lpars and Vioses between T0 and T1.

        :return: Estimated cycles spent on workload (including VMs and Virtual
                 I/O Server).
        """
        # Get the utilization of all LPARs
        data_lpar_t0 = (CpuUtilProcessor._parse_ltm_raw_data(
            data_t0_json, is_vios=False, is_host=True))
        data_lpar_t1 = (CpuUtilProcessor._parse_ltm_raw_data(
            data_t1_json, is_vios=False, is_host=True))
        # Get the utilization of all VIOSes
        data_vios_t0 = (CpuUtilProcessor._parse_ltm_raw_data(
            data_t0_json, is_vios=True, is_host=True))
        data_vios_t1 = (CpuUtilProcessor._parse_ltm_raw_data(
            data_t1_json, is_vios=True, is_host=True))
        # Get the delta utilization
        all_lpars_delta_util_cs = self._calculate_delta_user_cycles(
            data_lpar_t0, data_lpar_t1)
        all_vios_delta_util_cs = self._calculate_delta_user_cycles(
            data_vios_t0, data_vios_t1)
        return (all_lpars_delta_util_cs + all_vios_delta_util_cs)

    def _calculate_delta_user_cycles(self, data_lpar_t0, data_lpar_t1):
        '''This method calculates the delta user cycles between T1 and T0.
        '''
        delta_user_cycles = 0
        for key, value_t0 in six.iteritems(data_lpar_t0):
            value_t1 = data_lpar_t1.get(key, None)
            # data for the LPAR must be at both collection points
            if value_t1 is not None:
                cap_cs_t1 = value_t1['utilizedCappedProcCycles']
                uncap_cs_t1 = value_t1['utilizedUnCappedProcCycles']
                donated_cs_t1 = value_t1['donatedProcCycles']
                cap_cs_t0 = value_t0['utilizedCappedProcCycles']
                uncap_cs_t0 = value_t0['utilizedUnCappedProcCycles']
                donated_cs_t0 = value_t0['donatedProcCycles']
                # Utilized = Capped + Uncapped - Donated
                lpar_utilized_cs_t1 = cap_cs_t1 + uncap_cs_t1 - donated_cs_t1
                lpar_utilized_cs_t0 = cap_cs_t0 + uncap_cs_t0 - donated_cs_t0
                delta_user_cycles += lpar_utilized_cs_t1 - lpar_utilized_cs_t0
        return delta_user_cycles

    def _calculate_delta_firmware_cycles(self, fw_data_t0, fw_data_t1):
        '''This method calculates the delta firmware cycles between T1 and T0
        '''
        fw_cycles_t1 = fw_data_t1.get('utilizedProcCycles', 0)
        fw_cycles_t0 = fw_data_t0.get('utilizedProcCycles', 0)
        return fw_cycles_t1 - fw_cycles_t0

    def _calculate_delta_total_cycles(
            self, time_based_cycles_t0, time_based_cycles_t1, total_cores):
        '''This method gets the delta total cycles between T1 and T0.
        '''
        tot_cycles_t1 = total_cores * time_based_cycles_t1
        tot_cycles_t0 = total_cores * time_based_cycles_t0
        return tot_cycles_t1 - tot_cycles_t0
