# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from powervc_nova import _

from powervc_nova.virt.ibmpowervm.common import constants
from powervc_nova.virt.ibmpowervm.hmc.k2.entry_wrapper import EntryWrapper

from oslo_log import log as logging

LOG = logging.getLogger(__name__)


COMPLIANCE_STATE = 'ComplianceState'
POOL_ID = 'PoolID'
POOL_NAME = 'PoolName'
TOTAL_MOBILE_COD_PROC_UNITS = 'TotalMobileCoDProcUnits'
AVAILABLE_MOBILE_COD_PROC_UNITS = 'AvailableMobileCoDProcUnits'
UNRETURNED_MOBILE_COD_PROC_UNITS = 'UnreturnedMobileCoDProcUnits'
MGMT_CONSOLES = 'PowerEnterprisePoolManagementConsoles'
MGMT_CONSOLE = 'PowerEnterprisePoolManagementConsole'
IS_MASTER_CONSOLE = 'IsMasterConsole'
MGMT_CONSOLE_MTMSN = 'ManagementConsoleMachineTypeModelSerialNumber'
MACHINE_TYPE = 'MachineType'
MODEL = 'Model'
SERIAL_NUMBER = 'SerialNumber'


class EnterprisePoolEntryWrapper(EntryWrapper):

    def __init__(self, k2entry):
        super(EnterprisePoolEntryWrapper, self).__init__(k2entry)

    @property
    def compliance_state(self):
        return self.get_parm_value(COMPLIANCE_STATE,
                                   constants.EPOOL_OUT_OF_COMPLIANCE)

    @property
    def unreturned_mobile_cores(self):
        return int(self.get_parm_value(UNRETURNED_MOBILE_COD_PROC_UNITS, 0))

    @property
    def pool_id(self):
        return self.get_parm_value(POOL_ID)

    @property
    def pool_name(self):
        return self.get_parm_value(POOL_NAME)

    @property
    def total_mobile_cores(self):
        return int(self.get_parm_value(TOTAL_MOBILE_COD_PROC_UNITS, 0))

    @property
    def available_mobile_cores(self):
        return int(self.get_parm_value(AVAILABLE_MOBILE_COD_PROC_UNITS, 0))

    @property
    def master_console_mtmsn(self):
        mgmt_consoles_element = self._k2find(MGMT_CONSOLES)

        if mgmt_consoles_element is None:
            self.log_missing_value(MGMT_CONSOLES)
            return None

        mgmt_console_elements = mgmt_consoles_element.findall(MGMT_CONSOLE)

        mtmsn = None
        for mgmt_console_element in mgmt_console_elements:
            is_master_console = mgmt_console_element.findtext(
                IS_MASTER_CONSOLE)

            if not is_master_console:
                LOG.warn(_('Missing %(is_master_console)s parameter.') %
                         {'is_master_console': IS_MASTER_CONSOLE})
                continue

            if is_master_console.lower() != 'true':
                # We don't care about any HMCs that are not the master
                continue

            mtmsn_element = mgmt_console_element.find(MGMT_CONSOLE_MTMSN)

            if not mtmsn_element:
                LOG.warn(_('Missing %(mgmt_console_mtmsn)s parameter.') %
                         {'mgmt_console_mtmsn': MGMT_CONSOLE_MTMSN})
                continue

            machine_type = mtmsn_element.findtext(MACHINE_TYPE)
            model = mtmsn_element.findtext(MODEL)
            serial_number = mtmsn_element.findtext(SERIAL_NUMBER)

            if (machine_type is None or model is None or
                    serial_number is None):
                LOG.warn(_('Missing parameter that is part of MTMSN (machine '
                           'type, model, serial number): machine_type='
                           '%(type)s, model=%(model)s, serial_number='
                           '%(serial)s.') %
                         {'type': machine_type,
                          'model': model,
                          'serial': serial_number})
                continue

            mtmsn = '{0}-{1} {2}'.format(machine_type, model,
                                         serial_number)
            break
        else:
            LOG.error(_('Unable to determine master management console '
                        'MTMSN (machine type, model, serial number) from '
                        '%(identifier)s because no %(param)s was found to be '
                        'marked as the master console for the pool.') %
                      {'identifier': self.get_type_and_uuid(),
                       'param': MGMT_CONSOLE})

        return mtmsn
