# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from oslo_log import log as logging
from powervc_nova.virt.ibmpowervm.common import exception
from powervc_nova import _

LOG = logging.getLogger(__name__)

# Managed System Constants
DELIM = '/'
ROOT = './'

hmc_sharing_mode_to_nova = {'sre idle proces': 'share_idle_procs',
                            'sre idle procs active': 'share_idle_procs_active',
                            'sre idle procs always': 'share_idle_procs_always',
                            'keep idle procs': 'keep_idle_procs',
                            'uncapped': 'uncapped', 'capped': 'capped'}


_deprecated_properties = {}
"""
    A dictionary of properties which have been deprecated by schema changes
    in the HMC.  The key is the current property and the value is a list
    of prior property names.  If a call is made to retrieve a property and
    it doesn't exist, then the associated _deprecated_properties will be
    used.
"""


def add_deprecated_property_name(new_property_name, old_property_name):
    """
    Add an alternate property name for a given property.
    Sometimes it is necessary to communicate with multiple versions
    of an HMC server.  By creating a list of alternative property names,
    the lookup of a value can try alternatives if the main proeprty is
    not found.
    This method may be called multiple times to add multiple alternatives
    for a property.
    @param new_property_name The current version of the property name.
    @param old_property_name A prior version of the property.  If the
        new_property_name does not exist in the entry, then the
        old_property_name will be checked.
    """

    if new_property_name is None or old_property_name is None:
        return

    if new_property_name in _deprecated_properties:
        deprecated_list = _deprecated_properties[new_property_name]
        deprecated_list.append(old_property_name)
    else:
        deprecated_list = [old_property_name]
        _deprecated_properties[new_property_name] = deprecated_list


def get_deprecated_property_names(new_property_name):
    """
    Return the list of deprecated property names for the specified
    property (if any) set by add_deprecated_property_name.
    @return A list of string property names if one exists.  None otherwise
    """

    if new_property_name in _deprecated_properties:
        return _deprecated_properties[new_property_name]

    return None


class EntryWrapper(object):

    def __init__(self, k2entry):
        if k2entry is None:
            raise exception.IBMPowerVMEmptyK2Entry()
        self._k2entry = k2entry

    def _k2find(self, property_name, use_find_all=False):
        if use_find_all:
            found_value = self._k2entry.element.findall(property_name)
        else:
            found_value = self._k2entry.element.find(property_name)

        if found_value is not None:
            return found_value

        # The property was not found. See if there is a deprecated value
        deprecated_names = get_deprecated_property_names(property_name)
        if deprecated_names is None:
            return None  # No other properties to try

        for deprecated_name in deprecated_names:
            if use_find_all:
                found_value = self._k2entry.element.findall(deprecated_name)
            else:
                found_value = self._k2entry.element.find(deprecated_name)

            if found_value is not None:
                return found_value

        return None  # Did not find value

    def set_parm_value(self, property_name, value):
        element_value = self._k2find(property_name)
        if element_value is None:
            self.log_missing_value(property_name)
            return

        element_value.settext(value)

    def get_parm_value(self, property_name, default=None):
        element_value = self._k2find(property_name)
        if element_value is None:
            self.log_missing_value(property_name)
            return default

        text = element_value.gettext()
        if text is None:
            return default

        if type(text) is str:
            text = text.strip()

        return text

    def get_parm_values(self, property_name):
        """
        Gets a list of values from K2

        param property_name: property to return
        returns: list of strings containing property values
        """
        values = []
        elements = self._k2find(property_name, use_find_all=True)
        for element in elements:
            values.append(element.gettext())
        return values

    def get_parm_value_bool(self, property_name, default='*unset*'):
        """
        Gets the boolean value of a K2 property.

        @param property_name: property to return
        @param default: The value to return if the property is not found
            in the K2 data
        @return: If the property exists in the K2 data and has the value
            'true' or 'false' (case-insensitive), then the corresponding
            boolean value will be returned.
            If the property does not exist, then the default value will be
            returned if specified, otherwise False will be returned.
        """
        value = self.get_parm_value(property_name)
        if value is None and default != '*unset*':
            # The caller has set a default value.  Return their default
            # value if the K2 data does not have their property.
            return default

        if value:
            value = value.lower()

        return value == 'true'

    def get_parm_value_int(self, property_name, default=0):
        """
        Gets the integer value of a K2 property

        param property_name: property to return
        returns: integer value of property. Otherwise, default value
        """
        value = self.get_parm_value(property_name)
        if value is None or len(value) == 0:
            return default

        try:
            int_value = int(value)
            return int_value
        except ValueError:
            message = _("Cannot convert %(property_name)s='%(value)s' "
                        "to an integer in object %(k2object)s") % \
                {"property_name": property_name,
                 "value": value,
                 "k2object": self.get_type_and_uuid()}

            LOG.error(message)

        return default

    def get_uuid(self):
        """
        Returns the uuid of the K2 entry.
        """
        if self._k2entry is None:
            return None

        uuid = self._k2entry.properties['id']
        return uuid

    def get_k2_type(self):
        """
        Returns the K2 type of the entry from the tag value
        of the attribute. (ManagedServer, LogicalPartition, etc)
        """

        tag = self._k2entry.element.gettag()
        if tag is None:
            return None

        return tag

    def get_type_and_uuid(self):
        """
        Return the type and uuid of this entry together in one string.
        This is useful for error messages, logging, etc.
        """
        entry_type = self.get_k2_type()
        uuid = self.get_uuid()

        if entry_type is None:
            entry_type = "UnknownType"

        if uuid is None:
            uuid = "UnknownUUID"

        return entry_type + ":" + uuid

    def log_missing_value(self, param):
        error_message = _('The expected parameter of %(param)s '
                          'was not found in %(identifier)s') % \
            {"param": param,
             "identifier": self.get_type_and_uuid()}
        LOG.warn(error_message)
