# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import copy

from oslo_concurrency.lockutils import synchronized
from oslo_log import log as logging

from oslo_config import cfg

from powervc_oslo.k2.k2operator import K2Element
from powervc_oslo.k2.k2operator import K2Error

from powervc_nova import logcall
from powervc_nova import _

from powervc_nova.virt.ibmpowervm.hmc import exception as hmc_ex

from powervc_nova.virt.ibmpowervm.hmc.k2 import constants as k2const
from powervc_nova.virt.ibmpowervm.hmc.k2 import storage_map_processor
from powervc_nova.virt.ibmpowervm.hmc.k2.storage_map_processor import \
    StorageMapProcessor

from powervc_nova.virt.ibmpowervm.hmc.storage.vios_storage_info \
    import VIOSStorageInfo
from powervc_nova.virt.ibmpowervm.hmc.storage.virtual_storage_adapter import \
    VirtualFCServerAdapter
from powervc_nova.virt.ibmpowervm.hmc.storage.virtual_storage_adapter import \
    VirtualFCClientAdapter

CONF = cfg.CONF

LOG = logging.getLogger(__name__)

ROOT = storage_map_processor.ROOT
DELIM = storage_map_processor.DELIM
ETAG = storage_map_processor.ETAG
ASSOC_LPAR = storage_map_processor.ASSOC_LPAR
PARTITION_ID = storage_map_processor.PARTITION_ID
STORAGE_TYPE = storage_map_processor.STORAGE_TYPE
SERVER_ADAPTER = storage_map_processor.SERVER_ADAPTER
CLIENT_ADAPTER = storage_map_processor.CLIENT_ADAPTER
LOCAL_LPAR_ID = storage_map_processor.LOCAL_LPAR_ID
VIR_SLOT_NUM = storage_map_processor.VIR_SLOT_NUM
CONN_LPAR_ID = storage_map_processor.CONN_LPAR_ID
CONN_SLOT_NUM = storage_map_processor.CONN_SLOT_NUM
ADAPTER_NAME = storage_map_processor.ADAPTER_NAME
BACKING_DEV = storage_map_processor.BACKING_DEV
STORAGE = storage_map_processor.STORAGE
VOL_SIZE = storage_map_processor.VOL_SIZE
VOL_UID = storage_map_processor.VOL_UID
UNIT_SIZE = storage_map_processor.UNIT_SIZE
UDID = storage_map_processor.UDID
TARGET_DEV = storage_map_processor.TARGET_DEV
TARGET_NAME_REL_PATH = storage_map_processor.TARGET_NAME_REL_PATH
LU_ADDR_REL_PATH = storage_map_processor.LU_ADDR_REL_PATH
PV = storage_map_processor.PV
LU = storage_map_processor.LU
WWPNS = storage_map_processor.WWPNS
PORT = 'Port'
PORT_NAME = 'PortName'
PORT_WWPN = 'WWPN'
SERVER_ADAPTER_PATH = storage_map_processor.SERVER_ADAPTER_PATH
CLIENT_ADAPTER_PATH = storage_map_processor.CLIENT_ADAPTER_PATH
CLIENT_WWPNS_REL_PATH = storage_map_processor.CLIENT_WWPNS_REL_PATH
PORT_WWPN_REL_PATH = ROOT + PORT + DELIM + PORT_WWPN
CLIENT_VIR_SLOT_NUM_REL_PATH = storage_map_processor. \
    CLIENT_VIR_SLOT_NUM_REL_PATH
CLIENT_CONN_LPAR_REL_PATH = storage_map_processor.CLIENT_CONN_LPAR_REL_PATH

POST_RETRIES = CONF.ibmpowervm_update_collision_retries

EXT_FC_MAP = storage_map_processor.EXT_FC_MAP


class FCMapProcessor(StorageMapProcessor):

    def __init__(self):
        super(FCMapProcessor, self).__init__(storage_map_processor.VFC)

    def populate_client_adapter(self, lpar_id, slot_num, remote_lpar,
                                remote_slot):
        return VirtualFCClientAdapter(lpar_id, slot_num, remote_lpar,
                                      remote_slot)

    def populate_server_adapter(self, lpar_id, slot_num, remote_lpar,
                                remote_slot):
        return VirtualFCServerAdapter(lpar_id, slot_num, remote_lpar,
                                      remote_slot)

    @logcall
    def parse_and_populate(self, vios_elem, map_entries=None):
        """
        Parse the storage map and populate VIOStorageInfo.
        This information is used for onboarding support.

`       :param vios_elem: A k2 element representing a vios.
        :param map_entries: storage map entries for the vios. If not provided
                            all the storage will be populated.

        :returns: vios storage information
        """
        if vios_elem is not None:

            if map_entries is None:
                map_entries = vios_elem.findall(ROOT + self.map_collection +
                                                DELIM + self.map_type)

            if map_entries is not None:
                vios_storages = VIOSStorageInfo(
                    int(vios_elem.findtext(ROOT + PARTITION_ID)))

                for map_entry in map_entries:
                    # extract elements from VirtualSCSIMapping
                    # i.e, ClientAdapter, ServerAdapter, Storage,
                    # TargetDevice, etc
                    server_adapter = None
                    client_adapter = None
                    # vfchost
                    vadapter = None
                    # port name like fcs0
                    td_name = None
                    wwpns = None
                    storage_info = {}

                    for elem in map_entry.iter():
                        # if it's client or server adapter element,
                        # extract slot info
                        if (elem.gettag() == SERVER_ADAPTER or
                                elem.gettag() == CLIENT_ADAPTER):
                            lpar_id = None
                            slot_num = None
                            remote_lpar = None
                            remote_slot = None
                            adapter_attrs = elem.iter()
                            if adapter_attrs is None:
                                continue
                            for attr in adapter_attrs:
                                field_name = attr.gettag()
                                field_val = attr.gettext()
                                # only extract info if value is set
                                if field_val:
                                    if field_name == LOCAL_LPAR_ID:
                                        lpar_id = int(field_val)
                                    elif field_name == VIR_SLOT_NUM:
                                        slot_num = int(field_val)
                                    elif (field_name ==
                                          CONN_LPAR_ID):
                                        remote_lpar = int(field_val)
                                    elif field_name == CONN_SLOT_NUM:
                                        remote_slot = int(field_val)
                                    elif field_name == ADAPTER_NAME:
                                        vadapter = field_val
                                    elif field_name == WWPNS:
                                        wwpns = field_val

                            if elem.gettag() == SERVER_ADAPTER:
                                # It's acceptable to configure a server adapter
                                # to accept any client lpar or remote slot num.
                                # In that case, HMC would not return remote
                                # lpar field.
                                # However, the remote slot would set to
                                # 65535 (0xffff). We will override the value to
                                # remote lpar to 65535 as well.
                                if remote_lpar is None:
                                    remote_lpar = 65535
                                server_adapter = self.populate_server_adapter(
                                    lpar_id, slot_num, remote_lpar,
                                    remote_slot)
                            else:
                                client_adapter = self.populate_client_adapter(
                                    lpar_id, slot_num, remote_lpar,
                                    remote_slot)
                        elif elem.gettag() == PORT:
                            # extract the device attributes
                            td_name = elem.findtext(ROOT + PORT_NAME)
                            unique_id = elem.findtext(ROOT + UDID)
                            port_wwpn = elem.findtext(ROOT + PORT_WWPN)
                            storage_info['storage_uid'] = unique_id
                            storage_info['port_wwpn'] = port_wwpn

                    if client_adapter is None and not td_name:
                        continue
                    storage_info['target_device_name'] = td_name
                    storage_info['wwpns'] = (
                        None if wwpns is None else wwpns.split(' '))
                    storage_info['adapter_name'] = vadapter

                    vios_storages.add(storage_info, server_adapter,
                                      client_adapter)

                    LOG.debug('storage_info %s' % storage_info)
                return vios_storages
        return None

    @logcall
    def attach_storage(self, oper, cec_uuid, vios_uuid,
                       client_lpar_id, client_slot_num,
                       backing_dev, storage_type=None,
                       add_info=None, resp=None, lpar_uuid=None):
        """
        To attach storage to the specified client lpar and slot via k2.
        It will try to find an entry matching to specified parameters.
        If response is not passed in, we will extract the latest via
        k2operator.
        :param oper: K2Operator that will be used for GET and POST
        :param cec_uuid: k2 uuid of the managed system
        :param vios_uuid: k2 uuid of the virtual io server
        :param client_lpar_id: partition ID of the client lpar
        :param client_slot_num: virtual slot number of the client adapter
        :param backing_dev: physical port name (i.e fcs2)
        :param storage_type: N/A for virtual fibre channel adapter
        :param add_info: N/A for virtual fibre channel adapter
        :param resp: VirtualIOServer K2 entry. Provided for performance
                     purpose. If it's not provided, we will issue a GET to
                     extract the most up-to-date VirtualIOServer output
        :param lpar_uuid: This is used when we cannot find an adapter to
                          attach. We will create one and attach storage.
        :return VirtualIOServer entry from the K2 POST response
        """
        def _create_new_entry(map_entry, dev):
            """
            This method create a new map entry for the attached storage
            """
            # make copy of server & client adapter and port
            server_adapter = copy.deepcopy(map_entry.find(SERVER_ADAPTER_PATH))
            client_adapter = copy.deepcopy(map_entry.find(CLIENT_ADAPTER_PATH))
            associated_lpar = copy.deepcopy(map_entry.find(ROOT + ASSOC_LPAR))

            schema_attr = k2const.DEFAULT_SCHEMA_ATTR
            # construct the port element
            port = K2Element(PORT, attrib=schema_attr,
                             children=[K2Element(PORT_NAME, text=dev)])
            new_map_entry = K2Element(self.map_type,
                                      attrib=schema_attr,
                                      children=[associated_lpar,
                                                client_adapter,
                                                port,
                                                server_adapter])
            return new_map_entry

        def _attach_storage(k2resp):
            """
            Map physical port to specified virtual fibre channel adapter.
            If the physical is already mapped to an adapter in the given
            client partition. It's a NOP.
            """

            def _find_match_entries():
                '''
                Find the matching entries
                Returns:
                storage_map
                entries found
                Will return None, None if a physical port is already mapped
                '''
                storage_map = vios_resp.entry.element.find(ROOT +
                                                           self.map_collection)

                # check whether there is already a fc zoned to the phy port
                mapped_entries = self.find_map_entries(
                    storage_map, client_lpar_id,
                    client_slot_num,
                    backing_dev=backing_dev,
                    with_valid_client_adapter=True)
                if mapped_entries:
                    LOG.debug('port %s is already mapped' % backing_dev)
                    return None, None

                # find the entries with matching info
                entries = self.find_map_entries(storage_map, client_lpar_id,
                                                client_slot_num,
                                                with_valid_client_adapter=True,
                                                without_port_name=True)
                # return the stg map and entries we found
                return storage_map, entries

            # start, check if we need to get the vios info
            vios_resp, etag, ext_props = (
                self.initial_vios_load(k2resp, [], oper, cec_uuid,
                                       vios_uuid, xag=None))

            read_again = True
            while True:
                # We should have a usable vios_resp now
                storage_map, entries = _find_match_entries()
                # if the storage map is None then it's already mapped
                if storage_map is None:
                    return None

                # if we cannot find an entry, the vios_resp probably is not up
                # to date, we need to do a get it again with age=0
                if not entries and read_again:
                    vios_resp, etag, ext_props = self.load_vios(oper, cec_uuid,
                                                                vios_uuid,
                                                                xag=None,
                                                                age=0)
                    read_again = False
                else:
                    # Found a good storage map and found entries, or no retries
                    break

            if entries:
                new_elem = _create_new_entry(entries[0], backing_dev)

                # remove old entry
                storage_map.remove(entries[0])
                storage_map.append(new_elem)
                try:
                    return self.k2oper_update_vios(oper, cec_uuid, vios_uuid,
                                                   etag, vios_resp,
                                                   ext_props=ext_props)
                except K2Error as e:
                    self._check_update_failed(e.k2response, vios_uuid)
                    # if we get here then it's an unexpected error
                    self._log_k2err(vios_resp.entry.element, e)
                    raise e
            else:
                # dump out k2 request and response for debug
                if vios_resp is not None:
                    LOG.error(_('Unable to find adapters for attaching '
                                'storage. vios_resp: %s')
                              % vios_resp.body)
                else:
                    LOG.error(_('Unable to find adapters for attaching '
                                'storage.'))
                oper.dump_logs()
                # raise exception
                raise hmc_ex.IBMPowerVMInvalidVIOConfig()

        @synchronized('stg-mapping-lock' + vios_uuid)
        def _call_attach_storage():
            """
            With built-in retry logics if we get back etag mismatch error.
            """
            k2resp = resp
            for retries in xrange(POST_RETRIES + 1):
                try:
                    return _attach_storage(k2resp)
                except hmc_ex.IBMPowerVMUpdateFailed as e:
                    LOG.debug('Retry  _call_attach_storage %s' % retries)
                    if retries == POST_RETRIES:
                        # dump all the k2 requests & responses accumulated
                        # since the wrapper would not log single
                        # HTTP_STATUS_ETAG_MISMATCH failure
                        oper.dump_logs()
                        raise e
                    else:
                        k2resp = None
                except Exception as e:
                    LOG.exception(e)
                    raise e

        try:
            return _call_attach_storage()
        except hmc_ex.IBMPowerVMInvalidVIOConfig as e:
            # have to do it here otherwise we would have deadlock situation
            if lpar_uuid:
                # create and attach adapter instead
                lpar_href = oper.buildhref(rootType='LogicalPartition',
                                           rootId=lpar_uuid)
                return self.create_mixed_adapter_pairs(
                    oper,
                    cec_uuid,
                    vios_uuid,
                    lpar_href,
                    client_lpar_id,
                    num_fc=1,
                    resp=resp,
                    fc_backing_devs=[backing_dev])
            else:
                raise e

    @logcall
    def extract_wwpn_mappings(self, vios_entry, client_lpar_id):
        """
        Extract physical wwpns and virtual wwpns mapping.
        :param vios_entry: VirtualIOServer K2 entry
        :param client_lpar_id: client partition ID
        :return dictionary containing 'wwpns': list of virtual wwpns
                and 'phy_to_virt_initiators': dict of physical wwpn
                                              virtual wwpns.
        """
        storage_map = vios_entry.element.find(ROOT + self.map_collection)

        entries = self.find_map_entries(storage_map, client_lpar_id,
                                        with_valid_client_adapter=True,
                                        without_port_name=False)
        client_wwpn_list = []
        phy_port_map = {}
        if entries is None:
            return None
        else:
            for fc_mapping in entries:
                wwpns = fc_mapping.findtext(CLIENT_WWPNS_REL_PATH)
                phy_port_wwpn = fc_mapping.findtext(PORT_WWPN_REL_PATH)
                LOG.debug('wwpns %s phy_port_wwpn %s' % (wwpns, phy_port_wwpn))
                wwpn_list = None if wwpns is None else wwpns.split(' ')
                client_wwpn_list += wwpn_list
                if phy_port_wwpn is not None:
                    phy_port_map[phy_port_wwpn] = wwpn_list
        return {'phy_to_virt_initiators': phy_port_map,
                'wwpns': client_wwpn_list}

    @logcall
    def extract_client_adapter_data(self, vios_entry, client_lpar_id):
        """
        Extract data from the client adapter including WWPNs, slot, and
        vios ID.
        :param vios_entry: VirtualIOServer K2 entry
        :param client_lpar_id: client partition ID
        :return list of dictionaries containing
            'phys_wwpn': physical port wwpn
            'client_slot': client FC adapter slot
            'vios_id': connecting vios ID
        :return None if no entries are found
        """
        storage_map = vios_entry.element.find(ROOT + self.map_collection)

        entries = self.find_map_entries(storage_map, client_lpar_id,
                                        with_valid_client_adapter=True,
                                        without_port_name=False)
        if entries is None or not entries:
            return None
        else:
            fc_data = []
            for fc_mapping in entries:
                slot_number = fc_mapping.findtext(CLIENT_VIR_SLOT_NUM_REL_PATH)
                vios_id = fc_mapping.findtext(CLIENT_CONN_LPAR_REL_PATH)
                phy_port_wwpn = fc_mapping.findtext(PORT_WWPN_REL_PATH)
                adapter_data = {'phys_wwpn': phy_port_wwpn,
                                'client_slot': slot_number,
                                'vios_id': vios_id}
                fc_data.append(adapter_data)
        return fc_data
