# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from oslo_config import cfg

from oslo_log import log as logging
from oslo_concurrency.lockutils import synchronized

from powervc_oslo.k2.k2operator import K2Element
from powervc_oslo.k2.k2operator import K2Error

from powervc_nova import logcall
from powervc_nova import _
from powervc_nova.virt.ibmpowervm.hmc.k2 import constants as k2const
from powervc_nova.virt.ibmpowervm.hmc.k2.storage_map_processor \
    import StorageMapProcessor
from powervc_nova.virt.ibmpowervm.hmc import exception as hmc_ex

LOG = logging.getLogger(__name__)

DELIM = '/'
ROOT = './'
ETAG = 'etag'

MEDIA_REPOSITORIES = 'MediaRepositories'
VIRTUAL_MEDIA_REPOS_ELEM = 'VirtualMediaRepository'
REPOSITORY_NAME = 'RepositoryName'
REPOSITORY_SIZE = 'RepositorySize'
OPTICAL_MEDIA = 'OpticalMedia'
VIRTUAL_OPTICAL_MEDIA = 'VirtualOpticalMedia'
MEDIA_NAME = 'MediaName'
MEDIA_REPOSITORIES_PATH = ROOT + MEDIA_REPOSITORIES
VIRT_MEDIA_REPOSITORY_PATH = MEDIA_REPOSITORIES_PATH + DELIM \
    + VIRTUAL_MEDIA_REPOS_ELEM
OPTICAL_MEDIA_PATH = VIRT_MEDIA_REPOSITORY_PATH + DELIM \
    + OPTICAL_MEDIA

CONF = cfg.CONF
POST_RETRIES = CONF.ibmpowervm_update_collision_retries
K2_EXT_PROPS = [k2const.VIOS_STORAGE_EXT_PROP]


class MediaRepositoryProcessor(object):
    """
    This class provides method to parse, extrabulate, and manipuate K2
    MediaRepositories
    """
    def __init__(self):
        """
        Initialize attributes.
        """

    @logcall
    def create_media_repository(
            self, oper, cec_uuid, vios_uuid, volume_group, size, resp=None):
        """
        Create a media repository.
        """
        def _findStoragePoolByGroup():
            volgroups_resp = oper.vios_read(k2const.VIOS, rootId=vios_uuid,
                                            childType=k2const.VOL_GROUP,
                                            age=0, vios_name=vios_uuid)
            LOG.info(_('Looking for VolumeGroup %s') % volume_group)
            volgroup_entries = volgroups_resp.feed.findentries(
                './GroupName', volume_group)
            found = False
            volgroup_entry = None
            if volgroup_entries:
                volgroup_entry = volgroup_entries[0]
                found = True
                LOG.info(_('Found VolumeGroup on VIOS %s') % vios_uuid)

            return volgroup_entry, found

        def _create_media_repository(k2resp):
            schema_attr = k2const.DEFAULT_SCHEMA_ATTR

            # building media repository entry
            new_media_repos = (
                K2Element(VIRTUAL_MEDIA_REPOS_ELEM,
                          attrib=schema_attr,
                          children=[K2Element(REPOSITORY_NAME,
                                              text='powervc'),
                                    K2Element(REPOSITORY_SIZE,
                                              text=str(size))]))
            # finding volume group
            volgroup_entry, found = _findStoragePoolByGroup()

            if (found):
                LOG.info(_('Creating Media Repository in %s') % volume_group)

                repositories = volgroup_entry.element.find(
                    MEDIA_REPOSITORIES_PATH)
                if (repositories):
                    # check for existing repository again.
                    repo = repositories.find(VIRTUAL_MEDIA_REPOS_ELEM)
                    if repo is not None:
                        LOG.info(_('Media Repository exists'))
                        return volgroup_entry

                volgroup_uuid = volgroup_entry.properties['id']
                # get cached entry to get etag
                volgroup_resp = oper.vios_read(k2const.VIOS, rootId=vios_uuid,
                                               childType=k2const.VOL_GROUP,
                                               childId=volgroup_uuid,
                                               vios_name=vios_uuid)
                etag = volgroup_resp.headers.get(ETAG)
                if etag is None:
                    # must be prior to 810 HMC, so need to GET again without
                    # cache for etag
                    volgroup_resp = oper.vios_read(k2const.VIOS,
                                                   rootId=vios_uuid,
                                                   childType=k2const.VOL_GROUP,
                                                   childId=volgroup_uuid,
                                                   age=0, vios_name=vios_uuid)
                    etag = volgroup_resp.headers.get(ETAG)
                if etag is None:
                    # raise exception
                    LOG.error(_('etag not found'))
                    raise hmc_ex.IBMPowerVMInvalidVIOConfig()

                if (not repositories):
                    repositories = K2Element(MEDIA_REPOSITORIES,
                                             attrib=schema_attr)
                    root = volgroup_resp.entry.element
                    # TODO: insert index not hardcoded
                    root.insert(7, repositories)

                # append the new entry at the end
                repositories.append(new_media_repos)
                try:
                    ret_resp = oper.update(volgroup_resp.entry.element,
                                           etag,
                                           k2const.VIOS,
                                           rootId=vios_uuid,
                                           childType=k2const.VOL_GROUP,
                                           childId=volgroup_uuid)
                except K2Error as e:
                    if e.k2response and e.k2response.status == (
                            k2const.HTTP_STATUS_ETAG_MISMATCH):
                        raise hmc_ex.IBMPowerVMUpdateFailed(uuid=vios_uuid)
                    else:
                        self._log_k2err(volgroup_resp.entry.element, e)
                        raise e
            else:
                LOG.error(_('Failed to find Volume Group %(vg_name)s '
                            'in Virtual I/O Server data.')
                          % {'vg_name': volume_group}
                          )
                raise hmc_ex.IBMPowerVMVGMissing(
                    volume_group_name=volume_group, uuid=vios_uuid)

            return ret_resp

        @synchronized('stg-mapping-lock' + vios_uuid)
        def _call_create_media_repository():
            """
            With built-in retry logic if we get back etag mismatch error.
            """
            k2resp = resp
            for retries in xrange(POST_RETRIES + 1):
                try:
                    return _create_media_repository(k2resp)
                except (hmc_ex.IBMPowerVMUpdateFailed,
                        hmc_ex.IBMPowerVIOSIsTooBusy) as e:
                    LOG.debug('Retry _create_media_repository %s'
                              % retries)
                    if retries == POST_RETRIES:
                        # dump all the k2 requests & responses accumulated
                        # since the wrapper would not log single
                        # HTTP_STATUS_ETAG_MISMATCH failure
                        oper.dump_logs()
                        raise e
                    else:
                        k2resp = None
                except Exception as e:
                    LOG.exception(e)
                    raise e

        return _call_create_media_repository()

    def delete_optical_media(self, oper, cec_uuid, vios_uuid,
                             vopt_name, resp=None):
        """
        delete optical media from a media repository.
        """
        def _find_vopt(optical_media, virtopt_name):
            media_elem = None
            for media in optical_media.findall(VIRTUAL_OPTICAL_MEDIA):
                optical_name = media.findtext(MEDIA_NAME)
                if (virtopt_name == optical_name):
                    media_elem = media
                    break
            return media_elem

        def _delete_media_in_repository(k2resp):
            smp = StorageMapProcessor()
            vios_resp, etag, ext_props = smp.initial_vios_load(
                k2resp, [], oper, cec_uuid, vios_uuid, xag=K2_EXT_PROPS)
            # finding volume group with Media Repository
            links = vios_resp.entry.element.findall('./StoragePools/link')
            media = None
            for j in range(len(links)):
                pool_resp = oper.readbyhref(links[j].getattrib()['href'])
                media = pool_resp.entry.element.find(
                    OPTICAL_MEDIA_PATH)
                if (media):
                    break
            if not media:
                # raise exception
                # TODO new exception?
                LOG.error(_('No media repository found on VIOS: %s')
                          % vios_uuid)
                raise hmc_ex.IBMPowerVMInvalidVIOConfig()

            vopt_name_list = (
                vopt_name if isinstance(vopt_name, list) else
                [vopt_name])
            delete_needed = False
            for name in vopt_name_list:
                # find virtual optical to delete
                elem_to_delete = _find_vopt(media, name)
                if (elem_to_delete):
                    media.remove(elem_to_delete)
                    delete_needed = True
                else:
                    LOG.warning(_('virtual optical: %s not found') % vopt_name)

            if delete_needed:
                etag = pool_resp.headers.get(ETAG)
                if etag is None:
                    # raise exception
                    # TODO: SW215351
                    # raise hmc_ex.IBMPowerVMInvalidVIOConfig()
                    LOG.error(_('etag not found'))
                pool_id = pool_resp.entry.properties['id']
                try:
                    oper.update(pool_resp.entry.element,
                                etag,
                                k2const.VIOS,
                                rootId=vios_uuid,
                                childType=k2const.VOL_GROUP,
                                childId=pool_id)
                except K2Error as e:
                    if e.k2response and e.k2response.status == (
                            k2const.HTTP_STATUS_ETAG_MISMATCH):
                        raise hmc_ex.IBMPowerVMUpdateFailed(uuid=pool_id)
                    else:
                        self._log_k2err(pool_resp.entry.element, e)
                        raise e

        @synchronized('stg-mapping-lock' + vios_uuid)
        def _call_delete_media_in_repository():
            """
            With built-in retry logic if we get back etag mismatch error.
            """
            k2resp = resp
            for retries in xrange(POST_RETRIES + 1):
                try:
                    return _delete_media_in_repository(k2resp)
                except (hmc_ex.IBMPowerVMUpdateFailed,
                        hmc_ex.IBMPowerVIOSIsTooBusy) as e:
                    LOG.debug('Retry _delete_media_in_repository %s'
                              % retries)
                    if retries == POST_RETRIES:
                        # dump all the k2 requests & responses accumulated
                        # since the wrapper would not log single
                        # HTTP_STATUS_ETAG_MISMATCH failure
                        oper.dump_logs()
                        raise e
                    else:
                        k2resp = None
                except Exception as e:
                    LOG.exception(e)
                    raise e

        _call_delete_media_in_repository()

    def _log_k2err(self, req, k2err):
        LOG.error(_('Failed to modify object with request %s.')
                  % req.toxmlstring() if req else '')
        resp = (k2err.k2response.body
                if k2err and k2err.k2response and k2err.k2response.body
                else '')
        LOG.error(_('Failed to modify object with response %s.') % resp)
