#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
import six
from oslo_log import log as logging
from powervc_nova import _

MANAGED_SYSTEM_TYPE = 'ManagedSystem'
RAW_METRICS_TYPE = 'RawMetrics'
LONG_TERM_MONITOR_TYPE = 'LongTermMonitor'
COMPUTE_LTM_ENABLED_PROPERTY = 'ComputeLTMEnabled'
LTM_ENABLED_PROPERTY = 'LongTermMonitorEnabled'
PREFERENCES_TYPE = 'preferences'
PCM_SERVICE = 'pcm'

LOG = logging.getLogger(__name__)


class MonitorPrefProcessor(object):
    """
    This class is used to enable or disable the long term monitoring
    of a CEC (host)
    """
    def __init__(self):
        pass

    def get_ltm_preference(self, oper, cec_uuid):
        pref_resp = oper.read(MANAGED_SYSTEM_TYPE,
                              cec_uuid,
                              PREFERENCES_TYPE,
                              service=PCM_SERVICE)
        # defect 23495
        # Since ComputeLTMEnabled is not available for HMC 780
        # we would need to use LongTermMonitorEnabled for HMC 780
        # The idea is is to check for existence of ComputeLTMEnabled
        # as a first check, failing that we need to default to
        # LongTermMonitorEnabled which is enabled for both 780 & 810
        try:
            long_term_monitoring_enabled = \
                pref_resp.feed.entries[0].element.\
                find(COMPUTE_LTM_ENABLED_PROPERTY)
            if long_term_monitoring_enabled is not None:
                return long_term_monitoring_enabled.gettext()
            else:
                long_term_monitoring_enabled = \
                    pref_resp.feed.entries[0].element.\
                    find(LTM_ENABLED_PROPERTY)
                if long_term_monitoring_enabled is not None:
                    return long_term_monitoring_enabled.gettext()
        except Exception as e:
            LOG.warning(_('Error reading ltm property %s') %
                        six.text_type(e), exc_info=True)
        return 'false'

    def enable_ltm(self, oper, cec_uuid):
        self._set_ltm_preference(oper, cec_uuid, 'true')

    def disable_ltm(self, oper, cec_uuid):
        self._set_ltm_preference(oper, cec_uuid, 'false')

    def _set_ltm_preference(self, oper, cec_uuid, value):
        if value is None:
            return

        value_to_set = value
        if str(value_to_set).lower() == 'true':
            value_to_set = 'true'
        else:
            value_to_set = 'false'

        try:
            pref_resp = oper.read(MANAGED_SYSTEM_TYPE,
                                  cec_uuid,
                                  PREFERENCES_TYPE,
                                  service=PCM_SERVICE)
            # defect 23495
            # Since ComputeLTMEnabled is not available for HMC 780
            # we would need to use LongTermMonitorEnabled for HMC 780
            # The idea is is to check for existence of ComputeLTMEnabled
            # as a first check, failing that we need to default to
            # LongTermMonitorEnabled which is enabled for both 780 & 810
            long_term_monitoring_enabled = \
                pref_resp.feed.entries[0].element.\
                find(COMPUTE_LTM_ENABLED_PROPERTY)
            if long_term_monitoring_enabled is None:
                long_term_monitoring_enabled = \
                    pref_resp.feed.entries[0].element.\
                    find(LTM_ENABLED_PROPERTY)

            if long_term_monitoring_enabled.gettext() != value:
                LOG.debug('Setting the ltm property to ' + str(value))
                long_term_monitoring_enabled.settext(value)

                try:
                    oper.update(pref_resp.feed.entries[0].element,
                                pref_resp.headers['etag'],
                                MANAGED_SYSTEM_TYPE,
                                cec_uuid,
                                PREFERENCES_TYPE,
                                service=PCM_SERVICE)
                except Exception:
                    LOG.warning(_('Error updating ltm property to %s') %
                                str(value), exc_info=True)
            else:
                LOG.debug('LTM property was already ' + str(value))
        except Exception:
            LOG.error(_('Failed to set ltm property to %s') % str(value),
                      exc_info=True)
