# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import copy
import re

from oslo_concurrency.lockutils import synchronized
from oslo_log import log as logging
from powervc_nova import _

from oslo_config import cfg

from powervc_oslo.k2.k2operator import K2Element
from powervc_oslo.k2.k2operator import K2Error

from powervc_nova import logcall

from powervc_nova.virt.ibmpowervm.hmc import exception as hmc_ex

from powervc_nova.virt.ibmpowervm.hmc.k2 import constants as k2const
from powervc_nova.virt.ibmpowervm.hmc.k2 import storage_map_processor
from powervc_nova.virt.ibmpowervm.hmc.k2.storage_map_processor import \
    StorageMapProcessor

from powervc_nova.virt.ibmpowervm.hmc.storage.vios_storage_info \
    import VIOSStorageInfo
from powervc_nova.virt.ibmpowervm.hmc.storage.virtual_storage_adapter import \
    VirtualSCSIServerAdapter
from powervc_nova.virt.ibmpowervm.hmc.storage.virtual_storage_adapter import \
    VirtualSCSIClientAdapter

CONF = cfg.CONF

LOG = logging.getLogger(__name__)

ROOT = storage_map_processor.ROOT
DELIM = storage_map_processor.DELIM
ETAG = storage_map_processor.ETAG
ASSOC_LPAR = storage_map_processor.ASSOC_LPAR
PARTITION_ID = storage_map_processor.PARTITION_ID
STORAGE_TYPE = storage_map_processor.STORAGE_TYPE
SERVER_ADAPTER = storage_map_processor.SERVER_ADAPTER
CLIENT_ADAPTER = storage_map_processor.CLIENT_ADAPTER
LOCAL_LPAR_ID = storage_map_processor.LOCAL_LPAR_ID
VIR_SLOT_NUM = storage_map_processor.VIR_SLOT_NUM
REMOTE_LPAR_ID = storage_map_processor.REMOTE_LPAR_ID
REMOTE_SLOT_NUM = storage_map_processor.REMOTE_SLOT_NUM
ADAPTER_NAME = storage_map_processor.ADAPTER_NAME
BACKING_DEV = storage_map_processor.BACKING_DEV
STORAGE = storage_map_processor.STORAGE
VOL_SIZE = storage_map_processor.VOL_SIZE
VOL_UID = storage_map_processor.VOL_UID
UNIT_SIZE = storage_map_processor.UNIT_SIZE
UDID = storage_map_processor.UDID
TARGET_DEV = storage_map_processor.TARGET_DEV
TARGET_NAME_REL_PATH = storage_map_processor.TARGET_NAME_REL_PATH
LU_ADDR_REL_PATH = storage_map_processor.LU_ADDR_REL_PATH
PV = storage_map_processor.PV
LU = storage_map_processor.LU
VOPT = storage_map_processor.VOPT
MEDIA_NAME = storage_map_processor.MEDIA_NAME
CLIENT_LPAR_REL_PATH = storage_map_processor.CLIENT_LPAR_REL_PATH
SERVER_ADAPTER_PATH = storage_map_processor.SERVER_ADAPTER_PATH
CLIENT_ADAPTER_PATH = storage_map_processor.CLIENT_ADAPTER_PATH
SERVER_BACKING_DEV_REL_PATH = storage_map_processor.SERVER_BACKING_DEV_REL_PATH
VDISK_TYPE = storage_map_processor.VDISK_TYPE
PV_TYPE = storage_map_processor.PV_TYPE
LU_TYPE = storage_map_processor.LU_TYPE
VOPT_TYPE = storage_map_processor.VOPT_TYPE
VDISK = storage_map_processor.VDISK
DISK_NAME = storage_map_processor.DISK_NAME
VOL_NAME = storage_map_processor.VOL_NAME
UNIT_NAME = storage_map_processor.UNIT_NAME
MEDIA_NAME = storage_map_processor.MEDIA_NAME

VOPT_MEDIA_PATH = ROOT + STORAGE + DELIM + VOPT + DELIM + MEDIA_NAME
RESERVE_POLICY = storage_map_processor.RESERVE_POLICY
CLIENT_VIR_SLOT_NUM_REL_PATH = storage_map_processor. \
    CLIENT_VIR_SLOT_NUM_REL_PATH
CLIENT_REMOTE_LPAR_REL_PATH = storage_map_processor.CLIENT_REMOTE_LPAR_REL_PATH

POST_RETRIES = CONF.ibmpowervm_update_collision_retries

EXT_SCSI_MAP = storage_map_processor.EXT_SCSI_MAP


class SCSIMapProcessor(StorageMapProcessor):

    def __init__(self):
        super(SCSIMapProcessor, self).__init__(storage_map_processor.SCSI)

    @logcall
    def contains_logical_volume(self, virt_scsi_element):
        """
        param virt_scsi_element: The K2 element for a virtual scsi mapping

        Returns true if this vscsi element contains a logical volume
        """
        result = False

        logical_volumes = virt_scsi_element.findall('./Storage/VirtualDisk')
        if logical_volumes:
            result = True

        return result

    def populate_client_adapter(self, lpar_id, slot_num, remote_lpar,
                                remote_slot):
        return VirtualSCSIClientAdapter(lpar_id, slot_num, remote_lpar,
                                        remote_slot)

    def populate_server_adapter(self, lpar_id, slot_num, remote_lpar,
                                remote_slot):
        return VirtualSCSIServerAdapter(lpar_id, slot_num, remote_lpar,
                                        remote_slot)

    @logcall
    def parse_and_populate(self, vios_elem, map_entries=None):
        """
        Parse the storage map and populate VIOStorageInfo.
        This information is used for onboarding support.

`       :param vios_elem: A k2 element representing a vios.
        :param map_entries: storage map entries for the vios. If not provided
                            all the storage will be populated.

        :returns: vios storage information
        """
        if vios_elem is not None:

            if map_entries is None:
                map_entries = vios_elem.findall(ROOT + self.map_collection +
                                                DELIM + self.map_type)
            if map_entries is not None:
                vios_storages = VIOSStorageInfo(
                    int(vios_elem.findtext(ROOT + PARTITION_ID)))

                for map_entry in map_entries:
                    # if there is no storage type, no point parsing this
                    # element
                    storage_type_elem = map_entry.find(ROOT + STORAGE)
                    if storage_type_elem is None:
                        continue
                    # extract elements from VirtualSCSIMapping
                    # i.e, ClientAdapter, ServerAdapter, Storage,
                    # TargetDevice, etc
                    server_adapter = None
                    client_adapter = None
                    vadapter = None
                    td_type = self._extract_storage_type(storage_type_elem)
                    # if there is no storage type, no point parsing this
                    # element
                    if td_type is None:
                        continue
                    td_name = None
                    backing_dev_name = None
                    wwpns = None
                    storage_info = {}

                    for elem in map_entry.iter():
                        # if it's client or server adapter element,
                        # extract slot info
                        if (elem.gettag() == SERVER_ADAPTER or
                                elem.gettag() == CLIENT_ADAPTER):
                            lpar_id = None
                            slot_num = None
                            remote_lpar = None
                            remote_slot = None
                            adapter_attrs = elem.iter()
                            if adapter_attrs is None:
                                continue
                            for attr in adapter_attrs:
                                field_name = attr.gettag()
                                field_val = attr.gettext()
                                # only extract info if value is set
                                if field_val:
                                    if field_name == LOCAL_LPAR_ID:
                                        lpar_id = int(field_val)
                                    elif field_name == VIR_SLOT_NUM:
                                        slot_num = int(field_val)
                                    elif (field_name ==
                                          REMOTE_LPAR_ID):
                                        remote_lpar = int(field_val)
                                    elif field_name == REMOTE_SLOT_NUM:
                                        remote_slot = int(field_val)
                                    elif field_name == ADAPTER_NAME:
                                        vadapter = field_val
                                    elif field_name == BACKING_DEV:
                                        backing_dev_name = field_val

                            if elem.gettag() == SERVER_ADAPTER:
                                # It's acceptable to configure a server adapter
                                # to accept any client lpar or remote slot num.
                                # In that case, HMC would not return remote
                                # lpar field.
                                # However, the remote slot would set to
                                # 65535 (0xffff). We will override the value to
                                # remote lpar to 65535 as well.
                                if remote_lpar is None:
                                    remote_lpar = 65535
                                server_adapter = self.populate_server_adapter(
                                    lpar_id, slot_num, remote_lpar,
                                    remote_slot)
                            else:
                                client_adapter = self.populate_client_adapter(
                                    lpar_id, slot_num, remote_lpar,
                                    remote_slot)
                        elif elem.gettag() == STORAGE:
                            # travese Storage element
                            for storage_elem in elem.iter():
                                if storage_elem.gettag() == PV:
                                    disk_size = int(storage_elem.findtext(
                                        VOL_SIZE))
                                    unique_id = storage_elem.findtext(
                                        VOL_UID)
                                    reserve_policy = storage_elem.findtext(
                                        RESERVE_POLICY)
                                    udid = storage_elem.findtext(UDID)
                                    # convert to GB
                                    disk_size = disk_size / 1024
                                    storage_info['disk_size'] = disk_size
                                    # strip out the SAN volume uid
                                    if unique_id:
                                        match = re.search(
                                            r'^[0-9]{5}([0-9A-F]{32}).+$',
                                            unique_id)
                                        # only update unique_id if found match
                                        if match and match.lastindex >= 1:
                                            unique_id = match.group(1)
                                    storage_info['storage_uid'] = unique_id
                                    storage_info['policy'] = reserve_policy
                                    storage_info['storage_udid'] = udid
                                elif storage_elem.gettag() == LU:
                                    sizestr = storage_elem.findtext(UNIT_SIZE)
                                    # If there is a decimal point, then strip
                                    # it off so we don't blow up converting
                                    # to int
                                    if sizestr.find('.') > 0:
                                        sizestr = sizestr[:sizestr.find('.')]
                                    disk_size = int(sizestr)
                                    udid = storage_elem.findtext(UDID)
                                    storage_info['disk_size'] = disk_size
                                    storage_info['storage_uid'] = udid
                                else:
                                    LOG.debug('Not supported storage')
                            # extract the disk size
                        elif elem.gettag() == TARGET_DEV:
                            # extract the device attributes
                            td_name = elem.findtext(TARGET_NAME_REL_PATH)
                            luaddr = elem.findtext(LU_ADDR_REL_PATH)
                            storage_info['lu_addr'] = luaddr

                    storage_info['backing_device'] = backing_dev_name
                    storage_info['storage_type'] = td_type
                    storage_info['target_device_name'] = td_name
                    storage_info['wwpns'] = wwpns
                    storage_info['adapter_name'] = vadapter

                    vios_storages.add(storage_info, server_adapter,
                                      client_adapter)

                    LOG.debug('storage_info %s' % storage_info)
                return vios_storages
        return None

    @logcall
    def find_attached_medias(self, vios_entry, lpar_id):
        """
        Iterate through VirtualSCSIMappings to find attach optical medias
        for the given lpar_id.
        """
        map_entries = vios_entry.findall(ROOT + self.map_collection +
                                         DELIM + self.map_type)
        media_names = []

        # handle whether input is int or string
        if isinstance(lpar_id, int):
            client_lpar_id = str(lpar_id)
        else:
            client_lpar_id = lpar_id

        if map_entries:
            for entry in map_entries:
                found_match_lpar = False
                if client_lpar_id is not None:
                    client_lpar = entry.findtext(CLIENT_LPAR_REL_PATH)
                    if (client_lpar is not None and
                            client_lpar == client_lpar_id):
                        LOG.debug('found matching client adapter %s' % entry)
                        found_match_lpar = True
                    elif client_lpar is None:
                        # this would occurred if the client lpar has been
                        # deleted or only server adapter is created.
                        # we need to check the server adapter to find a match
                        server_remote_lpar_id = (
                            entry.findtext(self.serv_remote_lpar_path))
                        # if we found the ServerAdapter's remote lpar ID,
                        # make sure it matches to client_lpar_id.
                        if (server_remote_lpar_id is not None and
                                server_remote_lpar_id == client_lpar_id):
                            LOG.debug('found matching client adapter %s'
                                      % entry)
                            found_match_lpar = True

                if found_match_lpar:
                    opt_med_name = entry.findtext(VOPT_MEDIA_PATH)
                    if opt_med_name:
                        media_names.append(opt_med_name)
        return media_names

    @logcall
    def attach_storage(self, oper, cec_uuid, vios_uuid,
                       client_lpar_id, client_slot_num,
                       backing_dev, storage_type, add_info=None, resp=None,
                       lpar_uuid=None):
        """
        To attach storage to the specified client lpar and slot via k2.
        It will try to find an entry matching
        If response is not passed in, we will extract the latest via
        k2operator.
        :param oper: K2Operator that will be used for GET and POST
        :param cec_uuid: k2 uuid of the managed system
        :param vios_uuid: k2 uuid of the virtual io server
        :param client_lpar_id: partition ID of the client lpar
        :param client_slot_num: virtual slot number of the client adapter
        :param backing_dev: volume name
        :param storage_type: LogicalUnit
        :param add_info: logical unit's UDID
        :param resp: VirtualIOServer K2 entry. Provided for performance
                     purpose. If it's not provided, we will issue a GET to
                     extract the most up-to-date VirtualIOServer output
        :param lpar_uuid: This is used when we cannot find an adapter to
                          attach. We will create one and attach storage.
        :return VirtualIOServer entry from the K2 POST response
        """
        def _create_new_entry(map_entry):
            """
            This method create a new map entry for the attached storage
            """
            # make copy of server & client adapter
            server_adapter = copy.deepcopy(map_entry.find(SERVER_ADAPTER_PATH))
            client_adapter = copy.deepcopy(map_entry.find(CLIENT_ADAPTER_PATH))
            associated_lpar = copy.deepcopy(map_entry.find(
                './AssociatedPartition'))
            if associated_lpar is None:
                associated_lpar = copy.deepcopy(map_entry.find(
                    ROOT + ASSOC_LPAR))

            schema_attr = k2const.DEFAULT_SCHEMA_ATTR
            # building client adapter
            storage = storage_map_processor.create_storage_element(
                backing_dev, storage_type, add_info)

            new_map_entry = K2Element(self.map_type,
                                      attrib=schema_attr,
                                      children=[associated_lpar,
                                                client_adapter,
                                                server_adapter, storage])
            #                                    storage_type_elem])
            #  Add above line back if you are on HMC build older than 1338A

            return new_map_entry

        def _attach_storage(k2resp):

            # start, check if we need to get the vios info
            vios_resp, etag, ext_props = (
                self.initial_vios_load(k2resp, [], oper, cec_uuid,
                                       vios_uuid, xag=None))

            read_again = True
            while True:
                # We should have a usable vios_resp now
                storage_map = vios_resp.entry.element.find(ROOT +
                                                           self.map_collection)

                # find the entries with matching info
                entries = self.find_map_entries(storage_map, client_lpar_id,
                                                client_slot_num,
                                                with_valid_client_adapter=True)

                # if we cannot find an entry, the vios_resp probably is not up
                # to date, we need to do a get it again with age=0
                if not entries and read_again:
                    vios_resp, etag, ext_props = self.load_vios(oper, cec_uuid,
                                                                vios_uuid,
                                                                xag=None,
                                                                age=0)
                    read_again = False
                else:
                    # Found entries or we're not retrying
                    break

            if entries:
                new_elem = _create_new_entry(entries[0])
                # remove old entries
                for elem in entries:
                    # make sure there is no storage attached
                    # remove the old entries
                    curr_backing_dev = elem.findtext(
                        SERVER_BACKING_DEV_REL_PATH)
                    if not curr_backing_dev:
                        storage_map.remove(elem)
                storage_map.append(new_elem)
                try:
                    return self.k2oper_update_vios(oper, cec_uuid, vios_uuid,
                                                   etag, vios_resp,
                                                   ext_props=ext_props)
                except K2Error as e:
                    self._check_update_failed(e.k2response, vios_uuid)
                    # if we get here then it's an unexpected error
                    self._log_k2err(vios_resp.entry.element, e)
                    raise e
            else:
                # dump out k2 request and response for debug
                if vios_resp is not None:
                    LOG.error(_('Unable to find adapters for attaching '
                                'storage. vios_resp: %s')
                              % vios_resp.body)
                else:
                    LOG.error(_('Unable to find adapters for attaching '
                                'storage.'))

                oper.dump_logs()
                # raise exception
                raise hmc_ex.IBMPowerVMInvalidVIOConfig()

        @synchronized('stg-mapping-lock' + vios_uuid)
        def _call_attach_storage():
            """
            With built-in retry logics if we get back etag mismatch error.
            """
            k2resp = resp
            for retries in xrange(POST_RETRIES + 1):
                try:
                    return _attach_storage(k2resp)
                except hmc_ex.IBMPowerVMUpdateFailed as e:
                    LOG.debug('Retry  _call_attach_storage %s' % retries)
                    if retries == POST_RETRIES:
                        # dump all the k2 requests & responses accumulated
                        # since the wrapper would not log single
                        # HTTP_STATUS_ETAG_MISMATCH failure
                        oper.dump_logs()
                        raise e
                    else:
                        k2resp = None
                except Exception as e:
                    raise e

        try:
            return _call_attach_storage()
        except hmc_ex.IBMPowerVMInvalidVIOConfig as e:
            # have to do it here otherwise we would have deadlock situation
            if lpar_uuid:
                # create and attach adapter instead
                lpar_href = oper.buildhref(rootType='LogicalPartition',
                                           rootId=lpar_uuid)
                scsi_info = (backing_dev, storage_type, add_info)
                return self.create_mixed_adapter_pairs(
                    oper, cec_uuid, vios_uuid,
                    lpar_href, client_lpar_id, num_scsi=1, resp=resp,
                    scsi_backing_devs=[scsi_info])
            else:
                raise e

    def _extract_storage_type(self, scsi_map_elem):
        tag = None
        if len(scsi_map_elem) > 0:
            # there should only be 1 children under Storage element
            tag = scsi_map_elem[0].gettag()
        return tag

    @logcall
    def extract_client_adapter_data(self, vios_entry, client_lpar_id):
        """
        Extract data from the client adapter including slot and
        vios ID.
        :param vios_entry: VirtualIOServer K2 entry
        :param client_lpar_id: client partition ID
        :return list of dictionaries containing
            'client_slot': client FC adapter slot
            'vios_id': connecting vios ID
        :return None if no entries are found
        """
        storage_map = vios_entry.element.find(ROOT + self.map_collection)

        entries = self.find_map_entries(storage_map, client_lpar_id,
                                        with_valid_client_adapter=True,
                                        without_port_name=False)
        if entries is None or not entries:
            return None
        else:
            scsi_data = []
            for scsi_mapping in entries:
                slot_number = scsi_mapping.findtext(
                    CLIENT_VIR_SLOT_NUM_REL_PATH)
                vios_id = scsi_mapping.findtext(CLIENT_REMOTE_LPAR_REL_PATH)
                adapter_data = {'client_slot': slot_number,
                                'vios_id': vios_id}
                scsi_data.append(adapter_data)
        return scsi_data
