# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import re

from powervc_nova import _

from powervc_nova.virt.ibmpowervm.hmc.k2.entry_wrapper import EntryWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2.entry_wrapper import DELIM
from powervc_nova.virt.ibmpowervm.hmc.k2.entry_wrapper import ROOT
from powervc_nova.virt.ibmpowervm.hmc import exception
import six

from oslo_log import log as logging

LOG = logging.getLogger(__name__)

PARTITION_NAME = ROOT + 'PartitionName'
VIOS_ID = ROOT + 'PartitionID'
VIOS_UUID = ROOT + 'PartitionUUID'
PARTITION_STATE = ROOT + 'PartitionState'
ZERO = '0'
FALSE = 'false'

VIOS_LICENSE = ROOT + 'VirtualIOServerLicenseAccepted'

# Fiber Channel
VIRT_FIBRE_CHANNEL_COLLECTION = ROOT + 'VirtualFibreChannelMappings'
VIRT_FIBRE_CHANNEL_MAPPING = (VIRT_FIBRE_CHANNEL_COLLECTION + DELIM +
                              'VirtualFibreChannelMapping')
CLIENT_ADAPTER = VIRT_FIBRE_CHANNEL_MAPPING + DELIM + 'ClientAdapter'
WWPNS = CLIENT_ADAPTER + DELIM + 'WWPNs'

# Virtual SCSI
VIRT_SCSI_MAPPINGS = ROOT + 'VirtualSCSIMappings'

# Media Repository
MEDIA_REPOSITORY = ROOT + 'MediaRepositories'
VIRT_MEDIA_REPOSITORY = MEDIA_REPOSITORY + DELIM + 'VirtualMediaRepository'

# Physical Volume
PHYSICAL_VOLUMES = ROOT + 'PhysicalVolumes'
PHYSICAL_VOLUME = PHYSICAL_VOLUMES + DELIM + 'PhysicalVolume'
VOLUME_UUID = 'VolumeUniqueID'
VOLUME_NAME = 'VolumeName'
RESERVE_POLICY = 'ReservePolicy'
VOLUME_UDID = 'UniqueDeviceID'
# Mem
MEM_CONFIG = ROOT + 'PartitionMemoryConfiguration'
CURR_MEM = MEM_CONFIG + DELIM + 'CurrentMemory'

# Procs
PROC_CONFIG = ROOT + 'PartitionProcessorConfiguration'
CURR_USE_DED_PROCS = PROC_CONFIG + DELIM + 'CurrentHasDedicatedProcessors'

# Dedicated proc
CURR_DED_PROC_CONFIG = (PROC_CONFIG + DELIM +
                        'CurrentDedicatedProcessorConfiguration')
CURR_PROCS = CURR_DED_PROC_CONFIG + DELIM + 'CurrentProcessors'

# Shared proc
CURR_SHARED_PROC_CONFIG = (PROC_CONFIG + DELIM +
                           'CurrentSharedProcessorConfiguration')
CURR_PROC_UNITS = CURR_SHARED_PROC_CONFIG + DELIM + 'CurrentProcessingUnits'

# Shared Ethernet Adapters
SHARED_ETHERNET_ADAPTERS = ROOT + 'SharedEthernetAdapters'
SHARED_ETHERNET_ADAPTER = (SHARED_ETHERNET_ADAPTERS + DELIM +
                           'SharedEthernetAdapter')
FREE_ETHERNET_ADAPTERS = ROOT + 'FreeEthenetBackingDevicesForSEA'
IO_ADAPTER_CHOICE = FREE_ETHERNET_ADAPTERS + DELIM + 'IOAdapterChoice'
ETHERNET_BACKING_DEVICE = IO_ADAPTER_CHOICE + DELIM + 'EthernetBackingDevice'
IP_INTERFACE = 'IPInterface'
IP_ADDRESS = 'IPAddress'

# LPM
MOVER_SERVICE_PARTITION = ROOT + 'MoverServicePartition'
RMC_STATE = 'ResourceMonitoringControlState'
RMC_IPADDR = 'ResourceMonitoringIPAddress'
# profile
ASSOC_PROF = ROOT + 'AssociatedPartitionProfile'
LPAR_PROFS = ROOT + 'PartitionProfiles'
PROF_LINK = LPAR_PROFS + DELIM + 'link'
# VIOS FC Port
FC_ADAPTER_PATH = (ROOT + 'PartitionIOConfiguration' + DELIM +
                   'ProfileIOSlots' + DELIM + 'ProfileIOSlot' + DELIM +
                   'AssociatedIOSlot' + DELIM + 'RelatedIOAdapter' + DELIM +
                   'PhysicalFibreChannelAdapter')


class VIOSEntryWrapper(EntryWrapper):

    def get_vios_name(self):
        return self.get_parm_value(PARTITION_NAME)

    def get_vios_id(self):
        return int(self.get_parm_value(VIOS_ID, ZERO))

    def get_vios_uuid(self):
        return self.get_parm_value(VIOS_UUID)

    def get_vios_state(self):
        return self.get_parm_value(PARTITION_STATE)

    def is_running(self):
        return self.get_vios_state() == 'running'

    def get_rmc_state(self):
        return self.get_parm_value(RMC_STATE)

    def get_rmc_ipaddr(self):
        return self.get_parm_value(RMC_IPADDR)

    def is_rmc_active(self):
        return self.get_rmc_state() == 'active'

    def get_virtual_scsi_mappings(self):
        return self._k2entry.element.find(VIRT_SCSI_MAPPINGS)

    def get_media_repository(self):
        return self._k2entry.element.find(VIRT_MEDIA_REPOSITORY)

    def get_wwpns(self):
        """
        returns: A list of the wwpn pairs for the vios
        """
        return self.get_parm_values(WWPNS)

    def license_accepted(self):
        return self.get_parm_value_bool(VIOS_LICENSE, default=True)

    def get_hdisk_reserve_policy(self, disk_uuid):
        """
        Get the reserve policy for an hdisk

        param disk_uuid: The uuid of the hdisk
        returns: The reserve policy or None if the disk isn't found
        """
        policy = None

        # Get all the physical volume elements and look for a diskname match
        volumes = self._k2entry.element.findall(PHYSICAL_VOLUME)
        for volume in volumes:
            vol_uuid = volume.findtext(VOLUME_UUID)
            match = re.search(r'^[0-9]{5}([0-9A-F]{32}).+$', vol_uuid)

            if match and match.group(1) == disk_uuid:
                policy = volume.findtext(RESERVE_POLICY)
                break

        return policy

    def get_hdisk_from_uid(self, disk_uuid):
        """
        Get the hdisk name from the volume uid

        param disk_uuid: The uuid of the hdisk
        returns: The associated hdisk name
        """
        name = None

        # Get all the physical volume elements and look for a diskname match
        volumes = self._k2entry.element.findall(PHYSICAL_VOLUME)
        for volume in volumes:
            vol_uuid = volume.findtext(ROOT + VOLUME_UDID)
            if vol_uuid:
                LOG.debug('get_hdisk_from_uid match: %s' %
                          six.text_type(vol_uuid))
                LOG.debug('get_hdisk_from_uid disk_uuid: %s' %
                          six.text_type(disk_uuid))
                if vol_uuid == disk_uuid:
                    name = volume.findtext(VOLUME_NAME)
                    break
        # toss execption if we didn't have a name.
        if name is None:
            LOG.info(_("Did not find the disk identifier: %(uid)s "
                       " in the list of hdisks from the "
                       "Virutal I/O server: %(volumes)s") %
                     {'uid': disk_uuid,
                      'volumes': volumes})
            raise exception.IBMPowerVMVolumeLookupFailed(
                uuid=six.text_type(disk_uuid))
        return name

    def get_current_mem(self):
        return self.get_parm_value(CURR_MEM, ZERO)

    def get_current_proc_mode(self):
        # Returns true if dedicated or false if shared
        return self.get_parm_value(CURR_USE_DED_PROCS, FALSE).lower()

    def get_current_procs(self):
        return self.get_parm_value(CURR_PROCS, ZERO)

    def get_current_proc_units(self):
        return self.get_parm_value(CURR_PROC_UNITS, ZERO)

    def is_mover_service_partition(self):
        return self.get_parm_value_bool(MOVER_SERVICE_PARTITION, False)

    def get_ip_addresses(self):
        """
        Returns a list of IP addresses assigned to the VIOS
        """
        ip_list = []

        # Get all the shared ethernet adapters and free
        # ethernet devices and pull the IPs
        seas = self._k2entry.element.findall(SHARED_ETHERNET_ADAPTER)
        free_eths = self._k2entry.element.findall(ETHERNET_BACKING_DEVICE)
        for eth in seas + free_eths:
            ip = eth.findtext(IP_INTERFACE + DELIM + IP_ADDRESS)
            if ip and ip not in ip_list:
                ip_list.append(ip)

        return ip_list

    def get_vios_wwpns(self):
        wwpn_list = []

        fcAdapters = self._k2entry.element.findall(FC_ADAPTER_PATH)
        if fcAdapters:
            LOG.debug('Found element(s): %s.' % FC_ADAPTER_PATH)
            for fcAdapter in fcAdapters:
                portElements = fcAdapter.findall(
                    './PhysicalFibreChannelPorts/PhysicalFibreChannelPort')
                if portElements:
                    for portElement in portElements:
                        LOG.debug('Found PhysicalFibreChannelPort element.')
                        wwpnElement = portElement.find('./WWPN')
                        if wwpnElement is not None:
                            wwpn_list.append(wwpnElement.gettext())
                else:
                    LOG.debug("PhysicalFibreChannelPort elements not found")
        return wwpn_list

    def get_lpar_prof_path(self):
        """
        Find the last activated lpar profile. If there isn't one, pick
        the first profile.
        """
        prof_elem = self._k2entry.element.find(ASSOC_PROF)
        prof_path = None
        if prof_elem is not None:
            prof_path = prof_elem.getattrib()['href']
        if prof_path is None:
            prof_elem = self._k2entry.element.find(PROF_LINK)
            if prof_elem is not None:
                prof_path = prof_elem.getattrib()['href']

        return prof_path
