# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import copy
import functools
import six

from eventlet import greenthread
from oslo_config import cfg

from collections import deque
from logging import DEBUG
from oslo_log import log as logging
from powervc_oslo.k2 import k2operator as k2o
import nova.context as os_ctx

from powervc_nova import _
from powervc_nova import strong_store
from powervc_nova.virt.ibmpowervm.hmc import exception as hmc_ex

from powervc_nova.virt.ibmpowervm.hmc.k2 import constants as k2const
from powervc_nova.virt.ibmpowervm.hmc.k2 import contains_msg_in_k2_err


ibmpowervm_opts = [
    cfg.BoolOpt('ibmpowervm_log_k2o_debug',
                default=False,
                help='Determines whether K2 request/responses are logged.'),
]


CONF = cfg.CONF
CONF.register_opts(ibmpowervm_opts)

LOG = logging.getLogger(__name__)

log_store = strong_store

# list of K2 busy error code prefix
BUSY_ERR_CODE_PREFIX = ['HSCL3205']
NOT_IN_OPERATING_STATE = 'HSCL9010'
# list of unexpected HSCL errors that should be caught and passed to the UI
# in response to prohibited user actions
UNEXPECTED_HSCL_ERRORS = ['HSCLA4D1']
VIOS_BUSY_ERR_MSG = ("is currently too busy to process update operations")


def save_extra_parms_to_strong_store(parms):
    """
    Decorator to save the extra parameters specified in the list.
    to strong store. It will pop those extra parameters from
    the input as well.
    If there is already an existing parameter in the strong store,
    we will not pass any of the extra parameters.
    :param    parms: List of extra paameter names
    """
    def func_parms(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            # adding the parms to strong store
            conflicts = False
            parms_saved = {}
            if parms:
                for p in parms:
                    if p in kwargs:
                        # remove the parameter from the input
                        val = kwargs.pop(p)
                        # if there is already attribute in strong store
                        # with the same parameter name,
                        # set conflicts to True
                        if hasattr(strong_store, p):
                            LOG.warn(_('Parameter conflicts in the strong '
                                       'store : %s') % p)
                            conflicts = True
                        else:
                            # keep track extra parameters
                            parms_saved[p] = val
                # If there is no parameter conflict, copy the extra parameters
                # to strong store.
                # Otherwise, ignore all extra parameter passed in.
                # Treat them as noop
                if not conflicts:
                    for p in parms_saved:
                        setattr(strong_store, p, parms_saved[p])
            try:
                return f(*args, **kwargs)
            finally:
                # remove the parms from strong_store
                if parms and not conflicts:
                    for p in parms_saved:
                        delattr(strong_store, p)

        return wrapper
    return func_parms


def receive_extra_params_from_strong_store(parms):
    """
    Decorator to read the extra parameters specified in the list.
    from strong store. This is the corresponding method to
    save_extra_parms_to_strong_store.
    :param    parms: List of paameter names to read from strong store.
    """
    def func_parms(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            if parms:
                extra_input = {}
                # only read extra parameters if we can find them all
                found_all_extra_parms = True
                for p in parms:
                    if hasattr(strong_store, p):
                        # read extra parameters from strong store
                        extra_input[p] = getattr(strong_store, p)
                    else:
                        found_all_extra_parms = False
                if found_all_extra_parms:
                    for p in extra_input:
                        kwargs[p] = extra_input[p]
            return f(*args, **kwargs)
        return wrapper
    return func_parms


def init_thread_log():
    if not hasattr(log_store, 'k2olog'):
        log_store.k2olog = deque(maxlen=20)


def add_thread_log(entry):
    deck = log_store.k2olog
    deck.append(entry)


def dump_thread_log():
    def format_request(request):
        body = None
        # Parse the arguments if we're passed a tuple else its a string
        if isinstance(request, tuple):
            req_args = request[0]
            req_kwargs = request[1]
            dump = dict(method=req_args[0], path=req_args[1])
            for key in req_kwargs:
                if key == 'body':
                    # special format for body
                    body = req_kwargs.get('body')
                elif key == 'headers':
                    # deep copy the header and change what we can't dump
                    headers = copy.deepcopy(req_kwargs.get(key))
                    if 'X-API-Session' in headers:
                        headers['X-API-Session'] = '<SENSITIVE>'
                    dump[key] = str(headers)
                else:
                    log_entry = req_kwargs.get(key)
                    dump[key] = str(log_entry)
        else:
            dump = request

        # Dump all fields besides the body
        LOG.info(_('REQUEST: %s') % dump)
        # Now dump the full body
        if body is not None:
            LOG.info(body)

    def format_response(response):
        body = None
        # Parse the arguments if we're passed a dict else its a string
        if isinstance(response, dict):
            dump = {}
            for key in response:
                if key == 'body':
                    # special format for body
                    body = response.get('body')
                else:
                    log_entry = response.get(key)
                    dump[key] = str(log_entry)
        else:
            dump = response

        # Dump all fields besides the body
        LOG.info(_('RESPONSE: %s') % dump)
        # Now dump the full body
        if body is not None:
            LOG.info(body)

    # Pop each entry out of the log until it's empty
    try:
        while True:
            entry = log_store.k2olog.popleft()
            request = entry.get('request')
            if request is not None:
                format_request(request)

            response = entry.get('response')
            if response is not None:
                format_response(response)
    except IndexError:
        pass


class K2OperWrap(object):

    def __init__(self, *args, **kwargs):
        init_thread_log()
        self.is_shutdown = False
        self._k2oper = k2o.K2Operator(*args, **kwargs)

    def buildhref(self, *args, **kwargs):
        self._check_shutdown()
        return self._k2oper.buildhref(*args, **kwargs)

    def buildjobparm(self, *args, **kwargs):
        self._check_shutdown()
        return self._k2oper.buildjobparm(*args, **kwargs)

    def create(self, *args, **kwargs):
        self._check_shutdown()
        return self._k2oper.create(*args, **kwargs)

    def createbypath(self, *args, **kwargs):
        self._check_shutdown()
        return self._k2oper.createbypath(*args, **kwargs)

    def createjob(self, *args, **kwargs):
        self._check_shutdown()
        return self._k2oper.createjob(*args, **kwargs)

    def delete(self, *args, **kwargs):
        self._check_shutdown()
        return self._k2oper.delete(*args, **kwargs)

    def deletebyhref(self, *args, **kwargs):
        self._check_shutdown()
        return self._k2oper.deletebyhref(*args, **kwargs)

    def deletebypath(self, *args, **kwargs):
        self._check_shutdown()
        self._k2oper.deletebypath(*args, **kwargs)

    def downloadfile(self, *args, **kwargs):
        self._check_shutdown()
        return self._k2oper.downloadfile(*args, **kwargs)

    def extendpath(self, *args, **kwargs):
        self._check_shutdown()
        self._k2oper.extendpath(*args, **kwargs)

    def invalidate_cache_elem(self, *args, **kwargs):
        self._check_shutdown()
        return self._k2oper.invalidate_cache_elem(*args, **kwargs)

    def invalidate_cache_elem_by_path(self, *args, **kwargs):
        self._check_shutdown()
        return self._k2oper.invalidate_cache_elem_by_path(*args, **kwargs)

    def read(self, *args, **kwargs):
        self._check_shutdown()
        xag = kwargs.get('xag')
        service = kwargs.get('service')
        suffix_type = kwargs.get('suffixType')
        if suffix_type is None:
            if len(args) > 4:
                suffix_type = args[4]
        # if the extended properties is empty and not pcm service
        if (not xag and (service is None or service == 'uom') and
                suffix_type is None):
            if len(args) > 0:
                obj_type = args[0]
            else:
                obj_type = kwargs.get('rootType')
            child_type = kwargs.get('childType')
            if child_type is None and len(args) > 2:
                child_type = args[2]
            obj_type = obj_type if child_type is None else child_type
            # we don't utilize any of the extended properties offered
            # under ManagedSystem and LogicalPartition
            if obj_type == k2const.MGT_SYS or obj_type == k2const.LPAR:
                kwargs['xag'] = [None]
        return self._k2oper.read(*args, **kwargs)

    def readbyhref(self, *args, **kwargs):
        self._check_shutdown()
        return self._k2oper.readbyhref(*args, **kwargs)

    def readbypath(self, *args, **kwargs):
        self._check_shutdown()
        return self._k2oper.readbypath(*args, **kwargs)

    def readjob(self, *args, **kwargs):
        self._check_shutdown()
        return self._k2oper.readjob(*args, **kwargs)

    def readjobs(self, *args, **kwargs):
        self._check_shutdown()
        return self._k2oper.readjobs(*args, **kwargs)

    def shutdown(self, *args, **kwargs):
        """Force K2 log off and set is_shutdown."""
        self._k2oper = None
        self.is_shutdown = True

    @save_extra_parms_to_strong_store(['exp_errs'])
    def update(self, *args, **kwargs):
        self._check_shutdown()
        return self._k2oper.update(*args, **kwargs)

    def updatebypath(self, *args, **kwargs):
        self._check_shutdown()
        return self._k2oper.updatebypath(*args, **kwargs)

    def uploadfile(self, *args, **kwargs):
        self._check_shutdown()
        return self._k2oper.uploadfile(*args, **kwargs)

    def vios_read(self, *args, **kwargs):
        # This is a wrapper around read to use for
        # vios reads to handle exceptions from K2
        vios_name = None
        if 'vios_name' in kwargs:
            vios_name = kwargs.pop('vios_name')
        try:
            return self.read(*args, **kwargs)
        except Exception as e:
            if (contains_msg_in_k2_err(e, VIOS_BUSY_ERR_MSG) and
                    vios_name is not None):
                raise hmc_ex.IBMPowerVIOSIsTooBusy(vios_name=vios_name)
            else:
                raise hmc_ex.IBMPowerVIOSError()

    def _check_shutdown(self):
        if self.is_shutdown:
            raise hmc_ex.IBMPowerVMHMCSessionHasLoggedOff(host=CONF.host)

    def dump_logs(self):
        self._k2oper.session.dump_logs()


class K2SessWrap(k2o.K2Session):

    # We can't log this since it always has a password
    def __init__(self, *args, **kwargs):
        init_thread_log()
        super(K2SessWrap, self).__init__(*args, **kwargs)

    # Override the K2O request method to save logs around it.
    @receive_extra_params_from_strong_store(['exp_errs'])
    def request(self, *args, **kwargs):

        def log(sensitive, type_, value):
            if sensitive:
                value = '<SENSITIVE>'
            add_thread_log({type_: value})

        exp_errs = None
        if 'exp_errs' in kwargs:
            exp_errs = kwargs.pop('exp_errs')
        init_thread_log()
        sensitive = kwargs.get('sensitive', False)
        log(sensitive, 'request', (args, kwargs))

        # Get the context of this request and set audit memento
        try:
            user = kwargs.get('auditmemento')
            if user is None:
                context = getattr(strong_store, 'context', None)
                if context and os_ctx.is_user_context(context):
                    kwargs['auditmemento'] = context.user_name

        except Exception as e:
            # Ignore all exception if we can't process the memento
            pass

        response = None
        system_busy = True
        retry_count = 0
        found_exp_err = False
        while system_busy and retry_count < CONF.ibmpowervm_sys_busy_retries:
            try:
                # reset the system_busy
                system_busy = False
                # Call K2O request method
                response = super(K2SessWrap, self).request(*args, **kwargs)
            except k2o.K2Error as e:
                response = e.k2response
                # Initialize flags
                etag_mismatch_err = not_found = False
                if response is not None:
                    log(sensitive, 'response', response.__dict__)
                    if response.status == k2const.HTTP_STATUS_ETAG_MISMATCH:
                        etag_mismatch_err = True
                    elif response.status == k2const.HTTP_STATUS_NOT_FOUND:
                        not_found = True
                    else:
                        # check whether we received HSCL3205
                        # HSCL3205 indicating the managed system is busy,
                        # try the operation later
                        if response.body is not None:
                            resp_body = six.text_type(response.body)
                            if NOT_IN_OPERATING_STATE in resp_body:
                                raise hmc_ex.IBMPowerVMHostNotOperating()
                            if any(code in resp_body
                                   for code in BUSY_ERR_CODE_PREFIX):
                                system_busy = True
                                # increase the wait in every loop
                                greenthread.sleep(10 + 10 * retry_count)
                                LOG.debug('The system is busy. Retries count:'
                                          ' %s' % retry_count)
                                retry_count += 1
                            if exp_errs:
                                if any(code in resp_body for code in exp_errs):
                                    # clear the expect errors
                                    LOG.info(_('Encountered expected error'
                                               ' from K2: %s.') % exp_errs)
                                    if isinstance(e, k2o.K2HttpError):
                                        LOG.error(six.text_type(
                                            response.k2err.findtext(
                                                'Message')))
                                    else:
                                        LOG.error(six.text_type(resp_body))
                                    found_exp_err = True
                            # Extracts error message from response body and
                            # passes it to the UI for display
                            if any(error in resp_body for error in
                                   UNEXPECTED_HSCL_ERRORS):
                                errMsg = ((resp_body.partition('HSCL')[2]).
                                          partition(' ')[2]).splitlines()[0]
                                ex_args = {'err_message': errMsg}
                                LOG.error(_('HSCL error encountered: %s')
                                          % errMsg)
                                raise hmc_ex.IBMPowerVMHSCLError(**ex_args)

                # if system_busy and still within the retry count
                # continue on with the loop
                if (system_busy and
                        retry_count < CONF.ibmpowervm_sys_busy_retries):
                    continue

                if system_busy:
                    # exhaust the retry
                    raise hmc_ex.IBMPowerVMHostBusy()
                # Don't dump k2 req/resp for condition that might be expected.
                # With etag mismatch error, we often would retry the request
                # don't dump the log at this time
                if not (etag_mismatch_err or not_found or found_exp_err):
                    dump_thread_log()
                raise e

        # Log the response
        response_attributes = dict(response.__dict__)
        # Remove feed and entry from our dictionary of response attributes.
        # This avoids keeping heavy ElementTree objects in memory for FFDC
        # since FFDC only dumps the string version of the body for responses.
        response_attributes.pop('entry', None)
        response_attributes.pop('feed', None)

        log(sensitive, 'response', response_attributes)

        # Log if debug is on
        if CONF.ibmpowervm_log_k2o_debug and LOG.logger.isEnabledFor(DEBUG):
            dump_thread_log()

        return response

    def dump_logs(self):
        """
        With etag mismatch error (412), we will retry number of times. We
        don't want to dump_thread_log on every 412 failure. This
        method is called after we exhaust the retries.
        """
        dump_thread_log()
