# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import logging

from powervc_nova import logcall
from powervc_nova.virt.ibmpowervm.hmc.k2oprguard import K2OperGuard
from powervc_nova.virt.ibmpowervm.hmc.k2 import constants as k2const
from powervc_nova.virt.ibmpowervm.hmc.k2.viosentry_wrapper \
    import VIOSEntryWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2.storage_map_processor \
    import StorageMapProcessor
from powervc_nova.virt.ibmpowervm.hmc.k2.media_repository_processor \
    import MediaRepositoryProcessor


from oslo_config import cfg

LOG = logging.getLogger(__name__)
mr_opts = [
    cfg.StrOpt('media_rep_volume_group',
               default='rootvg',
               help='Default volume group to use for the VIOS'
               ' media repository'),
    cfg.StrOpt('media_rep_size',
               default='1',
               help='The size for the media repository in GB.')
]

CONF = cfg.CONF
CONF.register_opts(mr_opts)

K2_EXT_PROPS = [k2const.VIOS_STORAGE_EXT_PROP]


class MediaRepository(object):
    '''
    classdocs
    '''
    def __init__(self, host, vios_uuid):
        '''
        Constructor
        '''
        self.host = host
        self.vios_uuid = vios_uuid
        self.vios_wrapper = None
        self.media_repos = None

    @logcall
    def exists(self):
        if (not self.vios_wrapper):
            with K2OperGuard(host=self.host) as oper:
                smp = StorageMapProcessor()
                vios_resp, etag, ext_props = smp.load_vios(
                    oper, self.host.uuid, self.vios_uuid, xag=K2_EXT_PROPS)

            self.vios_wrapper = VIOSEntryWrapper(vios_resp.entry)
            self.media_repos = self.vios_wrapper.get_media_repository()

        if (self.media_repos):
            return True
        return False

    @logcall
    def create(self):
        """
        Creates the media repository on the VIOS.
        """
        volume_group = CONF.media_rep_volume_group
        size = CONF.media_rep_size
        media_repo_processor = MediaRepositoryProcessor()
        smp = StorageMapProcessor()
        with K2OperGuard(host=self.host) as oper:
            media_repo_processor.create_media_repository(
                oper, self.host.uuid, self.vios_uuid,
                volume_group, size, resp=None)
            # refresh VIOS info
            # need to make sure we don't use cache
            vios_resp, etag, ext_props = smp.load_vios(
                oper, self.host.uuid, self.vios_uuid, xag=K2_EXT_PROPS, age=0)

            self.vios_wrapper = VIOSEntryWrapper(vios_resp.entry)
            self.media_repos = self.vios_wrapper.get_media_repository()

    @logcall
    def get_free_space(self):
        """
        Get the free space available in the VIOS media repository.

        :returns: free space in media repository in GB
        """
        size_elem = self.media_repos.find('RepositorySize')
        size = float(size_elem.gettext())
        media_sizes = self.media_repos.findall('Size')
        for entry in media_sizes:
            media_size = float(entry.gettext())
            if media_size == 0:
                # setting media size to smallest granularity of K2 API value
                media_size = 0.0000005
            size = size - media_size

        return size

    def get_vios_name(self):
        """
        Return name of VIOS used in this repository class
        """
        if (self.vios_wrapper):
            return self.vios_wrapper.get_vios_name()
        return None
