# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import random
import time
import six
import copy
import re

from eventlet import greenthread

from nova.compute import power_state
from nova.exception import MigrationPreCheckError
from oslo_utils import excutils
from oslo_log import log as logging

from powervc_oslo.k2.k2operator import K2Error
from powervc_oslo.k2.k2operator import K2Element
from powervc_keystone.encrypthandler import EncryptHandler
from powervc_nova import logcall
from powervc_nova.storage import exception as stgex
from powervc_nova.storage import storage_mapping_api
from powervc_nova.storage import exception as stgex, storage_utils as su
from powervc_nova.virt.ibmpowervm.common import exception, volume_utils
from powervc_nova.virt.ibmpowervm.common import migrate_utils as utils
from powervc_nova.virt.ibmpowervm.common import lpar_state_checker
from powervc_nova.virt.ibmpowervm.common.constants import IBMI_OS_TYPE, \
    WWPN_SYSTEM_METADATA_KEY, FABRIC_STATE_METADATA_KEY, FS_INST_MAPPED, \
    FS_MIGRATING, NOVALINK_PLATFORM
from powervc_nova.virt.ibmpowervm.hmc.k2 import constants as k2const
from powervc_nova.virt.ibmpowervm.hmc.k2 import fc_map_processor
from powervc_nova.virt.ibmpowervm.hmc.k2 import jobentry_wrapper
from powervc_nova.objects.storage.dom import CHK_ALL, CHK_BOOT
from powervc_nova.virt.ibmpowervm.hmc.k2.lparentry_wrapper \
    import LPAREntryWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2.msentry_wrapper \
    import ManagedSystemEntryWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2 import scsi_map_processor
from powervc_nova.virt.ibmpowervm.hmc.k2 import storage_map_processor
from powervc_nova.virt.ibmpowervm.hmc.k2 import viosentry_wrapper
from powervc_nova.virt.ibmpowervm.hmc.k2oprguard import K2OperGuard
from powervc_nova.virt.ibmpowervm.hmc import virtualoptical
from powervc_nova import _

from powervc_nova.storage.storage_mapping_api import StorageMapping
from powervc_nova.virt.ibmpowervm.common.constants \
    import CONNECTION_TYPE_VSCSI
import powervc_nova.virt.ibmpowervm.hmc.lpar as lpar
import powervc_nova.virt.ibmpowervm.hmc.vscsi as pv_vscsi
from powervc_nova.volume import cinder as volume
from oslo_serialization import jsonutils
from powervc_nova.virt.ibmpowervm.hmc.k2.viosentry_wrapper \
    import VIOSEntryWrapper
from powervc_nova.conductor import api as powervc_conductor
from powervc_nova.objects import migrate_data as mig_obj

from oslo_config import cfg

LOG = logging.getLogger(__name__)

attrib = {'schemaVersion': 'V1_3_1', 'ksv': 'V1_3_1'}

ibmpowervm_opts = [
    cfg.IntOpt('ibmpowervm_live_migration_timeout',
               default=3600,
               help='Timeout in seconds for live migration task.')
]

CONF = cfg.CONF
CONF.register_opts(ibmpowervm_opts)


class Migrate(object):

    def __init__(self, instance, host, lpar_entry=None, lpar_uuid=None):
        """
        :param instance: lpar instance to be migrated
        :param host: The host object
        :param lpar_entry: The lpar entry from the K2 feed
        :param lpar_uuid: The K2 uuid of the lpar
        """
        # Get the host data
        # Refresh the stats to get live data about number of migrations running
        self.host = host
        # use K2 level caching
        with K2OperGuard(host=self.host) as k2oper:
            mgt_sys_resp = k2oper.read(k2const.MGT_SYS, rootId=self.host.uuid)
        self.host_entry = mgt_sys_resp.entry
        self.host_wrapper = ManagedSystemEntryWrapper(self.host_entry)
        self.host_stats = host.get_host_stats_internal(refresh=True)
        self.host_uuid = self.host_wrapper.get_uuid()
        self.host_name = self.host_wrapper.get_system_name()

        # Get the lpar data
        self.lpar = instance
        self.lpar_entry = lpar_entry
        self.lpar_uuid = lpar_uuid
        self.lpar_wrapper = None

        # Get the user MSP values which is a comma delimited string
        self.msps = self.host_stats.get('mover_service_partitions')
        if self.msps:
            self.msps = self.msps.split(',')

        # We'll always need to pull the VIOS data for the source.
        # We need to pull it for the target if there are MSP addresses.
        if self.msps or lpar_entry is not None:
            if lpar_entry is not None:
                self.lpar_wrapper = LPAREntryWrapper(self.lpar_entry)

            # Get vios data for the source migration methods.
            # Need to populate the vios cache first by refreshing.
            # use K2 level caching
            with K2OperGuard(host=self.host) as k2oper:
                vioses_resp = k2oper.read(k2const.MGT_SYS,
                                          rootId=self.host.uuid,
                                          childType=k2const.VIOS,
                                          xag=k2const.ALL_VIOS_EXT_PROPS)
            self.vios_entries = vioses_resp.feed.entries

        # Get the hmc
        self.hmc = host.select_active_hmc()

        # Get volume api
        self.volume_api = volume.PowerVCVolumeAPI()
        self.conductor_api = powervc_conductor.PowerVCConductorAPI()

    # Do not use logcall because passwords will be printed to the logs
    def check_can_live_migrate_destination(self):
        """
        1) Validates the lpar is a candidate for live
           migration to the target host.
        2) Collects data about the destination host for live
           partition migration to be used in future method calls

        :returns dest_check_data: dictionary of target host data
        """
        # Do not use logcall because the enter/exit parms contain a password
        LOG.debug("Entering check_can_live_migrate_destination/migrate.py")

        lpar_name = self.lpar['display_name']
        lpar_state = self.lpar.get('power_state')
        power_specs = self.lpar.get('power_specs')
        operating_system = power_specs['operating_system']

        # Verify processor compatibility mode
        lpar_compat_mode = self._get_lpar_proc_compat_mode(lpar_state)
        host_compat_modes = self.host_stats.get('compatibility_modes')
        utils.verify_proc_compat_mode(lpar_compat_mode, host_compat_modes,
                                      lpar_name)

        is_ded = power_specs.get('dedicated_proc')
        lpar_spp = None
        if is_ded is not None and is_ded.lower() == 'false':
            lpar_spp = power_specs.get('shared_proc_pool_name')
            lpar_spp_id = power_specs.get('shared_proc_pool_id')
        if lpar_spp is not None:
            if lpar_spp_id == 0:
                host_spp_id = 0
            else:
                host_spp_id = self.host.\
                    get_shared_proc_pool_id(spp_name=lpar_spp)
                utils.verify_shared_proc_pool(lpar_spp, host_spp_id, lpar_name)

        # Check pre-requisite on target host for IBM i
        if operating_system.startswith('IBM i'):
            # For IBM i, target host must have IBM i mobility capability
            capability = self.host_stats.get('capabilities')
            utils.verify_ibmi_mobility_capable_for_migration(capability,
                                                             lpar_name)
            utils.verify_ibmi_proc_compat_mode(host_compat_modes, lpar_name)

        # Verify the target host can support another migration
        mig_allowed, mig_running = self._get_migration_stats(lpar_state)
        utils.verify_host_capacity_for_migration(mig_allowed,
                                                 mig_running,
                                                 lpar_name,
                                                 self.host_name)

        # Create dictionary of target host data.  The nova manager code
        # looks for the information under the 'migrate_data' key when
        # it consolidates target and source dictionaries.
        migrate_data = mig_obj.PowerVCLiveMigrateData()
        migrate_data.dest_sys_name = self.host_name
        migrate_data.dest_mtms_name = self.host.host_id
        migrate_data.dest_ip = self.hmc.ipaddr
        migrate_data.dest_user_id = self.hmc.user
        migrate_data.dest_lmb = self.host_stats.get('lmb_size')
        migrate_data.dest_proc_compat = \
            self.host_stats.get('compatibility_modes')

        # add dest_spp_id only if the lpar has the spp id
        # and use host spp id of the same spp on target host
        if lpar_spp is not None:
            migrate_data.dest_spp_id = host_spp_id

        # Pick a mover service partition for the destination if available
        if self.msps:

            # Pull the valid MSPs off of the VIOS
            valid_msps = []
            for vios_entry in self.vios_entries:
                vios_wrapper = viosentry_wrapper.VIOSEntryWrapper(vios_entry)
                if vios_wrapper.is_mover_service_partition():
                    valid_msps += vios_wrapper.get_ip_addresses()

            migrate_data.dest_msp = self._choose_msp_address(
                self.msps, valid_msps)
            LOG.debug("MSP IP address chosen for destination host: %s"
                      % migrate_data.dest_msp)

        # Add the password outside of the 'migrate_data' dictionary.
        # This ensures it is only "seen" by the source validation, and will
        # not propagate through all of the migration methods.
        # The password must be encrypted via STG defect 22189.
        encrypted_pwd = EncryptHandler().encode(self.hmc.cred)
        migrate_data.dest_cred = encrypted_pwd
        # Do not use logcall because the enter/exit parms contain a password
        logged_dest_data = copy.deepcopy(migrate_data)
        logged_dest_data.dest_cred = '********'
        LOG.debug("Exiting check_can_live_migrate_destination/migrate.py: "
                  "dest_check_data=%s" % migrate_data.dest_cred)
        return migrate_data

    # Do not use logcall because passwords will be printed to the logs
    def check_can_live_migrate_source(self, ctxt, dest_check_data,
                                      live_migration=True):
        """
        1) Validates the lpar is a candidate for live
           migration on the source host.
        2) Establishes an SSH connection to a remote HMC if
           necessary
        3) Collects data about the source host for live
           partition migration to be used in future method calls

        :param dest_check_data: dictionary of target host data
        :param live_migration: Indicates whether this was called from the
                               live or cold migration flows.
        :returns migrate_data: dictionary of source host data
        """
        # Do not use logcall because the enter/exit parms contain a password
        logged_dest_data = copy.deepcopy(dest_check_data)
        logged_dest_data.dest_cred = '********'
        LOG.debug("Entering check_can_live_migrate_source/migrate.py: "
                  "dest_check_data=%s, "
                  "live_migration=%s" % (logged_dest_data,
                                         live_migration))

        # Get lpar data
        lpar_id = self.lpar_wrapper.get_lpar_id()
        lpar_name = self.lpar['display_name']
        lpar_state = self.lpar.get('power_state')
        lpar_type = self.lpar_wrapper.get_lpar_type()

        host_compat_modes = self.host_stats.get('compatibility_modes')

        # Check pre-requisite for IBM i
        if IBMI_OS_TYPE == lpar_type:
            # For IBM i, the lpar must be set as IsRestrictedIO as true
            iorestricted = self.lpar_wrapper.get_restrictedio()
            utils.verify_ibmi_iorestricted_for_migration(iorestricted,
                                                         lpar_name)

            # For IBM i, source host must have IBM i mobility capability
            capability = self.host_stats.get('capabilities')
            utils.verify_ibmi_mobility_capable_for_migration(capability,
                                                             lpar_name)
            # Check host proc_compat_mode for IBM i. Only Power7 and above
            # is OK. But Power8 is not support temporarily
            utils.verify_ibmi_proc_compat_mode(host_compat_modes, lpar_name)

        # Verify the source host can support another migration
        mig_allowed, mig_running = self._get_migration_stats(lpar_state)
        utils.verify_host_capacity_for_migration(mig_allowed,
                                                 mig_running,
                                                 lpar_name,
                                                 self.host_name)

        # Validate logical memory block sizes
        source_lmb = self.host_stats.get('lmb_size')
        target_lmb = int(dest_check_data.dest_lmb)
        utils.verify_logical_memory_block_size(source_lmb, target_lmb,
                                               lpar_name)

        # Validate RMC and DLPAR for live migration
        if (live_migration):
            dlpar, rmc_state = self.lpar_wrapper.check_dlpar_connectivity()
            # For IBM i, there is no need to verify rmc state
            if IBMI_OS_TYPE != lpar_type:
                utils.verify_rmc_state(rmc_state, lpar_name)
            utils.verify_dlpar_enabled(dlpar, lpar_name)

        # Validate the storage
        valid_msps = []
        for vios_entry in self.vios_entries:

            vios_wrapper = viosentry_wrapper.VIOSEntryWrapper(vios_entry)
            scsi_map = scsi_map_processor.SCSIMapProcessor()
            scsi_map_entries = scsi_map.find_map_entries(
                vios_wrapper.get_virtual_scsi_mappings(), lpar_id)

            # If the user defined a specific MSP to use
            # we need to pull the list of valid MSPs for each VIOS
            if self.msps:
                if vios_wrapper.is_mover_service_partition():
                    valid_msps += vios_wrapper.get_ip_addresses()

            # Check the storage mapping for lpar logical volumes
            for virt_scsi_element in scsi_map_entries:
                if (scsi_map.contains_logical_volume(virt_scsi_element)):
                    error = (_("Cannot migrate %(lpar)s because it " +
                               "contains a logical volume.")
                             % {'lpar': lpar_name})
                    LOG.exception(error)
                    raise MigrationPreCheckError(reason=error)

            # Verify the reserve policy
            scsi_storage = scsi_map.parse_and_populate(vios_entry.element,
                                                       scsi_map_entries)
            if scsi_storage is not None:
                disk_info = scsi_storage.disk_info_by_lpar.get(lpar_id)
                if disk_info is not None:
                    for disk in disk_info:
                        storage_type = disk.storage_type
                        if storage_type == storage_map_processor.PV_TYPE:
                            reserve_policy = disk.storage_info.get('policy')
                            disk_name = disk.backing_device
                            utils.verify_reserve_policy(reserve_policy,
                                                        lpar_name, disk_name)

        migrate_data = copy.deepcopy(dest_check_data)
        managed_by = None
        if hasattr(dest_check_data, 'managed_by'):
            managed_by = dest_check_data.managed_by
            if managed_by == NOVALINK_PLATFORM:
                # Grab HMC public key to exchange with Novalink host
                pub_key = self._get_hmc_public_key()
                migrate_data.public_key = pub_key
            else:
                # Establish an SSH connection to a remote HMC if necessary
                self._ensure_hmc_to_hmc_connection(dest_check_data)

        # Create dictionary of source data.  We don't need to add the target
        # data as the manager will consolidate it for us.
        migrate_data.refresh_conn_info = False
        migrate_data.src_sys_name = self.host_name

        # Pick a valid mover service partition for the source if available
        if self.msps:
            migrate_data.src_msp = self._choose_msp_address(
                self.msps, valid_msps)
            LOG.debug("MSP IP address chosen for source host: %s"
                      % migrate_data.src_msp)

        if managed_by == NOVALINK_PLATFORM:
            # Update instance's system_metadata with the fabric meta-data
            LOG.debug("Updating instance system_metada with fabric meta-data")
            npiv_fabric_port_maps, npiv_fabric_slots_map, vios_slot_peers = \
                self._extract_instance_fabric_metadata(ctxt)
            for fabric in npiv_fabric_port_maps.iterkeys():
                port_maps = []
                for phy_wwpn, virt_wwpns in \
                        npiv_fabric_port_maps.get(fabric).iteritems():
                    port_map = (phy_wwpn, " ".join(virt_wwpns))
                    port_maps.append(port_map)
                self._set_fabric_meta(fabric, port_maps)
                self._set_fabric_state(fabric, FS_INST_MAPPED)

            self.lpar.save()
            LOG.debug("Instance's system_metadata is updated with "
                      "fabric meta-data")

            # Update migrate_data with npiv_fabric_client_slots.
            # Save the slots which are placed on the same VIOS as
            # peers.
            #     Ex: [[2,3], [4,5]]
            # Where slots 2 and 3 are on one VIOS and 4,5 are on
            # the second.
            vol_data = {}
            for fabric in npiv_fabric_slots_map.iterkeys():
                vol_data['src_npiv_fabric_slots_' + fabric] = \
                    (jsonutils.dumps(npiv_fabric_slots_map.get(fabric)))
                vol_data['src_vios_peer_slots'] = \
                    (jsonutils.dumps(vios_slot_peers))
            migrate_data.vol_data = vol_data

        logged_mig_data = copy.deepcopy(migrate_data)
        logged_mig_data.dest_cred = '********'
        LOG.debug("Exiting check_can_live_migrate_source/migrate.py: "
                  "migrate_data=%s" % logged_mig_data)
        return migrate_data

    def _set_fabric_state(self, fabric, state):
        """Sets the fabric state into the instance's system metadata.
        :param fabric: The name of the fabric
        :param state: state of the fabric which needs to be set

        To-Do: This method is taken from nova-powervm code.
               Need to see a way to get rid of this duplicate code.

         Possible Valid States:
         FS_UNMAPPED: Initial state unmapped.
         FS_INST_MAPPED: Fabric is mapped with the nova instance.
        """
        meta_key = FABRIC_STATE_METADATA_KEY + '_' + fabric
        LOG.info(_("Setting Fabric state=%(st)s for instance=%(inst)s") %
                 {'st': state, 'inst': self.lpar.name})
        self.lpar.system_metadata[meta_key] = state

    def _set_fabric_meta(self, fabric, port_map):
        """Sets the port map into the instance's system metadata.

        The system metadata will store per-fabric port maps that link the
        physical ports to the virtual ports.  This is needed for the async
        nature between the wwpns call (get_volume_connector) and the
        connect_volume (spawn).

        To-Do: This method is taken from nova-powervm code.
               Need to see a way to get rid of this duplicate code.

        :param fabric: The name of the fabric.
        :param port_map: The port map.
        """

        # We will store the metadata in comma-separated strings with up to 4
        # three-token pairs. Each set of three comprises the Physical Port
        # WWPN followed by the two Virtual Port WWPNs:
        # Ex:
        # npiv_wwpn_adpt_A:
        #     "p_wwpn1,v_wwpn1,v_wwpn2,p_wwpn2,v_wwpn3,v_wwpn4,..."
        # npiv_wwpn_adpt_A_2:
        #     "p_wwpn5,v_wwpn9,vwwpn_10,p_wwpn6,..."

        meta_elems = []
        for p_wwpn, v_wwpn in port_map:
            meta_elems.append(p_wwpn)
            meta_elems.extend(v_wwpn.split())

        LOG.debug(("Fabric %(fabric)s wwpn metadata will be set to "
                   "%(meta)s for instance %(inst)s") %
                  {'fabric': fabric, 'meta': ",".join(meta_elems),
                   'inst': self.lpar.name})

        # Clear out the original metadata.  We may be reducing the number of
        # keys (ex. reschedule) so we need to just delete what we had before
        # we add something new.
        meta_key_root = WWPN_SYSTEM_METADATA_KEY + '_' + fabric
        for key in tuple(self.lpar.system_metadata.keys()):
            if key.startswith(meta_key_root):
                del self.lpar.system_metadata[key]

        # Build up the mapping for the new keys.
        fabric_id_iter = 1
        meta_key = meta_key_root
        key_len = len(meta_key)

        for key in range(self._get_num_keys(port_map)):
            start_elem = 12 * (fabric_id_iter - 1)
            meta_value = ",".join(meta_elems[start_elem:start_elem + 12])
            self.lpar.system_metadata[meta_key] = meta_value
            # If this is not the first time through, replace the end else cat
            if fabric_id_iter > 1:
                fabric_id_iter += 1
                meta_key = meta_key.replace(meta_key[key_len:],
                                            "_%s" % fabric_id_iter)
            else:
                fabric_id_iter += 1
                meta_key = meta_key + "_%s" % fabric_id_iter

    def _get_num_keys(self, port_map):
        """Returns the number of keys we need to generate"""
        # Keys will have up to 4 mapping pairs so we determine based on that
        if len(port_map) % 4 > 0:
            return int(len(port_map) / 4 + 1)
        else:
            return int(len(port_map) / 4)

    @logcall
    def pre_live_migration(self, context, instance, bdms, network_info,
                           migrate_data):
        """
        Prepare the destination vios for live migration

        Call plug_vifs() on target host to enable VLAN bridging for VM.
        Also establish mappings between vscsi volumes of an instance to
        dest host.

        NOTES:
        If there are any VLANs or disks attached to VM that are not
        managed by PowerVC, user needs to manually establish mapping on
        the target host before LPM started.

        :param instance: lpar instance to be migrated
        :param network_info: instance network information
        """

        try:
            # Establish hostmapping of ephemeral and cinder volumes on the
            # target host.
            pre_migration_data = self._map_vscsi_volumes_to_target_host(
                context, instance, bdms, migrate_data)
            LOG.info(_("Completed configuration of required "
                       "external storage."))
        except Exception as e:
            error = six.text_type(e)
            inst_name = instance['display_name']
            msg = (_("Failed during the configuration of required external "
                     "storage on destination host for virtual machine "
                     "%(inst_name)s. Error: %(error)s") % locals())
            LOG.exception(msg, instance=self.lpar)
            raise exception.IBMPowerVMMigrationFailed(msg)

        #migrate from novalink build the vfc and vscsi mappings
        src_managed_by = None
        if migrate_data:
            if hasattr(migrate_data, 'src_managed_by'):
                src_managed_by = migrate_data.src_managed_by
        if src_managed_by and src_managed_by == NOVALINK_PLATFORM:
            vios_scsi_data = {}
            vios_fc_data = self._generate_fc_data_for_novalink(migrate_data)
            stg_mappings = self._generate_storage_mapping_strings(
                context, migrate_data, vios_fc_data, vios_scsi_data)
            if stg_mappings and stg_mappings.get('fc_mappings'):
                vfc_map = jsonutils.\
                    dumps(stg_mappings['fc_mappings'].split(','))
                if hasattr(pre_migration_data, 'vol_data'):
                    pre_migration_data.vol_data['vfc_lpm_mappings'] = \
                        vfc_map
                else:
                    vol_data = {'vfc_lpm_mappings': vfc_map}
                    pre_migration_data.vol_data = vol_data

        try:
            # Plug VLAN on the target VIOSes
            self.host.plug_vifs(instance, network_info)
        except Exception as e:
            error = six.text_type(e)
            inst_name = instance['display_name']
            msg = (_("Failed to enable VLAN bridging "
                     "for virtual machine %(inst_name)s."
                     " %(error)s") % locals())
            LOG.exception(msg)
            raise exception.IBMPowerVMMigrationFailed(msg)

        if hasattr(migrate_data, 'dest_sys_name'):
            pre_migration_data.dest_sys_name = migrate_data.dest_sys_name
        # Add public key to HMC during NovaLink to HMC migration
        if hasattr(migrate_data, 'public_key'):
            pub_key = migrate_data.public_key
            self._add_novalink_public_key(migrate_data, pub_key)
        return pre_migration_data

    @logcall
    def _generate_fc_data_for_novalink(self, migrate_data):
        """
        Generate the fc data dict with the info from novalink
        migrate data.

        The fc data is a list of,
        {'phys_wwpn': phy_port_wwpn,
         'client_slot': slot_number,
         'vios_id': vios_id}
        for each physical port and client slot on Novalink
        """
        merge_fc_data = []
        if not migrate_data:
            return None
        vol_data = migrate_data.vol_data
        for key in vol_data.keys():
            if key.startswith('src_npiv_fabric_vios_slots_'):
                fabric_vios_slots = jsonutils.loads(vol_data[key])
                LOG.debug("fabric_vios_slots %s" % fabric_vios_slots)
                vios_fc_data = [self.__parse_vios_wwpn_slot(x) for x in
                                fabric_vios_slots]
                for x in vios_fc_data:
                    merge_fc_data.append(x)
                LOG.debug("fc_data %s" % vios_fc_data)
        return merge_fc_data

    def __parse_vios_wwpn_slot(self, merged_str):
        """
        Helper method to return fc data dict from the
        merged string
        Input : <vios_id>_<pwwpn>_<slot>
        output: {'phys_wwpn': phy_port_wwpn,
                 'client_slot': slot_number,
                 'vios_id': vios_id}
        """
        parsed_lst = merged_str.split('_')
        fc_data = {'phys_wwpn': parsed_lst[1],
                   'client_slot': parsed_lst[2],
                   'vios_id': parsed_lst[0]}
        return fc_data

    @logcall
    def _map_vscsi_volumes_to_target_host(self, context, instance, bdms,
                                          migrate_data):
        """
        Establish hostmaps for cinder volumes on the target host
        before live migration.
        This function is running on target host.

        :param instance: VM instance
        :param bdms: bdms for the cinder volumes on source host.
        :param migrate_data: source and target host related migration data.
        """

        LOG.debug("instance_uuid=%s, number of BDMs=%d." %
                  (instance['uuid'], 0 if not bdms else len(bdms)))
        scg_map = StorageMapping()
        # Determine what connectivity type check is needed so that
        # the connectivity call does not raise an unnecessary exception.
        if volume_utils.PowerVCNovaVirtDriverVolumeAPI().\
                get_instance_volumes(context, instance, 'data'):
            chk_conn_type = CHK_ALL
        else:
            chk_conn_type = CHK_BOOT
        scg_info = scg_map.get_connectivity_info(context, instance['uuid'],
                                                 self.host.host_id,
                                                 instance, live_data=False,
                                                 chk_conn_type=chk_conn_type)

        # extract scg class to six.text_type to avoid UnicodeEncodeError
        scg_class = scg_info.pop('scg')
        LOG.debug('scg_info: %s, scg_class: %s' %
                  (scg_info, six.text_type(scg_class)), instance=self.lpar)
        scg_info['scg'] = scg_class

        scg_backend_host_group = scg_info.get('backend_host_group')
        if scg_backend_host_group is not None:
            LOG.info(_("This LPAR belongs to an SCG which is using "
                       "Global-Host-Group '%s'.") % scg_backend_host_group,
                     instance=self.lpar)
        else:
            LOG.info(_("This LPAR belongs to an SCG which does not use any "
                       "Global-Host-Group."), instance=self.lpar)

        if(not bdms or not migrate_data):
            # no attached cinder volumes.
            LOG.info(_("No attached cinder volumes for processing."),
                     instance=self.lpar)
            return

        LOG.info(_("Processing volume mappings to target host..."),
                 instance=self.lpar)
        # Establish connection mappings for every VSCSI volume against
        # all target VIOS's and invoke lua_recovery.
        pre_migration_data = mig_obj.PowerVCLiveMigrateData()
        for bdm in bdms:
            LOG.debug("Processing block device mapping. bdm=%s." % bdm,
                      instance=self.lpar)
            connection_info = bdm['connection_info']
            if connection_info is None:
                LOG.error(_("BDM connection info is none. Skipping it."
                            "bdm=%s.") % bdm, instance=self.lpar)
                continue
            else:
                if connection_info and isinstance(connection_info, basestring):
                    connection_info = jsonutils.loads(connection_info)

            volume_id = connection_info.get('data').get('volume_id')
            volume_type = connection_info.get('driver_volume_type')
            if volume_type != 'fibre_channel':
                LOG.debug("Volume type is not fibre_channel. So, skipping "
                          "volume %s." % volume_id, instance=self.lpar)
                continue

            connection_type = lpar._get_volume_connection_type(connection_info)
            if connection_type is None:
                msg = (_("Could not identify the connection type of a volume. "
                         "Volume: %s.") % volume_id)
                LOG.error(_("Could not identify the connection type of a "
                            "volume. volume: %(volume_id)s. "
                            "connection_info: %(connection_info)s.")
                          % {'volume_id': volume_id,
                             'connection_info': connection_info},
                          instance=self.lpar
                          )

            if connection_type == CONNECTION_TYPE_VSCSI:
                LOG.info(_("Processing a vSCSI volume. volume=%s.") %
                         volume_id, instance=self.lpar)
                connector = lpar.\
                    get_volume_connector(self.host, instance,
                                         self.lpar['display_name'],
                                         self.lpar_uuid,
                                         self.lpar_entry)
                if connector is None:
                    msg = (_("Cannot get volume connector to "
                             "virtual machine: %s.") %
                           self.lpar['display_name'])
                    LOG.error(msg, instance=self.lpar)
                    raise Exception(msg)

                vios_ports_list = connector.pop('wwpn_vscsi')
                LOG.debug("List of VIOS'es and their WWPNS maps %s." %
                          vios_ports_list, instance=self.lpar)
                # Process the connection on all vios's
                pre_migration_data = self._process_vscsi_conn(context,
                                                              vios_ports_list,
                                                              connector,
                                                              connection_info,
                                                              bdm)

        return migrate_data

    @logcall
    def _process_vscsi_conn(self, context, vios_ports_list, vol_connector,
                            connection_info, bdm):
        """
        Establishes hostmap for cinder volume and the target VIOS's and invokes
        lua_recovery on the target VIOS's for hdisks to reflect on VIOS's.

        This function is running on target host.

        :param vios_ports_list: a list of maps where each map contains the
                                vios_name, vios_uuid and wwpns for a VIOS.
        :param vol_connector: volume connector for the VM.
        :param connection_info: connection info for the volume.
        :param bdm: block device mapping
        """
        backend_host_group = vol_connector.get('backend_host_group')
        volume_id = connection_info.get('data').get('volume_id')

        connector = copy.deepcopy(vol_connector)
        if 'connection_type' not in connector:
            connector.update({'connection_type': [CONNECTION_TYPE_VSCSI]})

        # process the connection on all vios's of the Host
        vscsi_lun_list = []
        pre_mig_data = {}
        for vios in vios_ports_list:
            connector['wwpns'] = vios['wwpns']
            connector['host'] = vios['vios_host']
            c_info = self._process_vscsi_conn_on_vios(context, vios, connector,
                                                      connection_info,
                                                      backend_host_group)
            for vios_target_lun in c_info['vscsi_target_lun']:
                vscsi_lun_list.append(vios_target_lun)
            # Update migrate data for Novalink to HMC migration
            pre_mig_data['vscsi-' + volume_id] = c_info['data']['target_UDID']

        # Update BDM entry in DB with new Connection_info.
        #
        # Modify connection info with new 'vscsi_target_lun', and volume
        # connector. Volume connector would have 'wwpns' of only one VIOS
        # of the host. The callers of detach, terminate_connection etc...
        # would loop through the wwpns of all the vios'es available
        # in 'vscsi_target_lun'.
        c_info['vscsi_target_lun'] = vscsi_lun_list
        if 'serial' not in c_info:
            c_info['serial'] = volume_id
        volume_connector = {'connection-type': CONNECTION_TYPE_VSCSI,
                            'connector': connector}
        c_info.update({'volume_connector': volume_connector})
        bdm.connection_info = jsonutils.dumps(c_info)
        bdm.save()
        LOG.debug("Updated BDM db entry for volume %(volume_id)s. "
                  "Updated BDM: %(bdm)s." % {'volume_id': volume_id,
                                             'bdm': bdm})
        return pre_mig_data

    @logcall
    def _process_vscsi_conn_on_vios(self, context, vios_wwpns, connector,
                                    connection_info, backend_host_group):
        """
        Establishes hostmap for cinder volume and the target VIOS and invokes
        lua_recovery on the target VIOS for hdisks to reflect on VIOS.

        This function is running on target host.

        :param vios_wwpns: a map that contains the vios_name, vios_uuid and
                           wwpns for that VIOS.
        :param connector: volume connector for the VM.
        :param connection_info: connection info for the volume.
        :param backend_host_group: Name of the global host group.
        :returns conn_info: returns conn_info for the volume for that VIOS.
        """
        volume_id = connection_info.get('data').get('volume_id')

        # As we are doing exclusively for vscsi
        # if any 'phy_to_virt_initiators' are there
        # in connector just pop them
        if 'phy_to_virt_initiators' in connector:
            connector.pop('phy_to_virt_initiators')

        if backend_host_group is None:
            LOG.info(_("Establishing mapping between volume %(volume)s "
                       "and VIOS %(vios)s.") %
                     {'volume': volume_id, 'vios': vios_wwpns['vios_name']},
                     instance=self.lpar)
            c_info = self.volume_api.initialize_connection(context, volume_id,
                                                           connector)
            if c_info is None:
                msg = (_("Connection information is not returned for volume "
                         "%(volume_id)s. connector: %(connector)s.") %
                       {'volume_id': volume_id, 'connector': connector})
                LOG.error(msg, instance=self.lpar)
                raise Exception(msg)
        else:
            # If connection_info is None, we would not reach here.
            c_info = copy.deepcopy(connection_info)

        vios_target_lun = {'vios_name': vios_wwpns['vios_name'],
                           'target_lun': c_info['data']['target_lun'],
                           'vios_wwpn': vios_wwpns['wwpns'],
                           'target_wwpn': c_info['data']['target_wwn'],
                           'vios_host': connector['host']
                           }
        if 'ip' in connector:
            vios_target_lun['vios_ip'] = connector['ip']
        c_info['vscsi_target_lun'] = [vios_target_lun]

        status, dev_name, udid = lpar.\
            _invoke_lua_recovery(self.host, c_info,
                                 vios_wwpns['vios_uuid'],
                                 self.lpar,
                                 vios_wwpns['vios_name'])
        if dev_name is not None:
            if status in [k2const.LUA_STATUS_DEVICE_AVIALABLE,
                          k2const.LUA_STATUS_DEVICE_IN_USE,
                          k2const.LUA_STATUS_FOUND_ITL_ERR]:
                LOG.info(_("Discovered hdisk %(hdisk)s on vios %(vios)s for "
                           "volume %(volume_id)s. Status code: %(status)s.") %
                         {'hdisk': dev_name, 'vios': vios_wwpns['vios_name'],
                          'volume_id': volume_id, 'status': status},
                         instance=self.lpar)
            else:
                msg = (_("Cannot proceed with the hdisk %(hdisk)s discovered "
                         "on vios %(vios)s for volume %(volume_id)s. "
                         "Status: %(status)s.") %
                       {'hdisk': dev_name, 'vios': vios_wwpns['vios_name'],
                        'volume_id': volume_id, 'status': status})
                LOG.error(msg, instance=self.lpar)
                raise Exception(msg)
        else:
            msg = (_("Failed to discover hdisk on %(vios)s for volume "
                     "%(volume_id)s. Status: %(status)s.") %
                   {'vios': vios_wwpns['vios_name'], 'volume_id': volume_id,
                    'status': status})
            LOG.error(msg, instance=self.lpar)
            raise Exception(msg)

        c_info['data']['target_UDID'] = udid

        # Check to see if we discovered the correct volume on the VIOS.
        # TODO: workaround till target_UDID is added in BDM of novalink
        expected_udid = connection_info.get('data').get('target_UDID', None)
        if expected_udid is not None:
            if udid != expected_udid:
                msg = (_("UDID of the discovered volume on vios %(vios)s "
                         "did not match with the expected UDID for volume "
                         "%(volume_id)s. UDID=%(udid)s. Expected "
                         "UDID=%(expected_udid)s.") %
                       {'vios': vios_wwpns['vios_name'],
                        'volume_id': volume_id,
                        'udid': udid, 'expected_udid': expected_udid})
                LOG.error(msg, instance=self.lpar)
                raise Exception(msg)

        return c_info

    @logcall
    def _remove_volume_device_from_vios(self, vios_name, vios_uuid,
                                        volume_id, volume_uid):
        """
        Removes volume device from VIOS.

        :param vios_name: name of the VIOS.
        :param vios_uuid: UUID of the VIOS.
        :param volume_id: ID of the volume.
        :param volume_uid: UID of the volume (part of the UUID of the
                           volume on storage).
        """
        LOG.info(_("Removing volume device from vios %(vios)s for volume "
                   "%(volume_id)s. volume_uid: %(volume_uid)s.") %
                 {'vios': vios_name, 'volume_id': volume_id,
                  'volume_uid': volume_uid}, instance=self.lpar)

        if vios_name is None or vios_uuid is None or volume_id is None or \
                volume_uid is None:
            LOG.error(_("Cannot detach and remove volume from vios. "
                        "At least one of the required properties is missing."),
                      instance=self.lpar)
            return

        with K2OperGuard(host=self.host) as oper:
            try:
                # extract vios info
                # Get vios data - need to populate the vios cache first
                # by refreshing
                vioses_resp = oper.read(k2const.MGT_SYS, rootId=self.host.uuid,
                                        childType=k2const.VIOS, age=0,
                                        xag=k2const.ALL_VIOS_EXT_PROPS)

                vios_entries = vioses_resp.feed.entries

                # Find the vios entry for our VIOS.
                vios_entry = None
                for entry in vios_entries:
                    vios_id = entry.properties['id']
                    if vios_id == vios_uuid:
                        vios_entry = entry
                        break

                if vios_entry is None:
                    LOG.error(_("No matching entry found for vios "
                                "%(vios_name)s. vios_uuid: %(vios_uuid)s.") %
                              {'vios_name': vios_name, 'vios_uuid': vios_uuid},
                              instance=self.lpar)
                    return

                vios_wrapper = VIOSEntryWrapper(vios_entry)
                vios_name = vios_wrapper.get_vios_name()
                backing_dev = None
                LOG.debug("vios uuid = %s" % vios_uuid, instance=self.lpar)

                try:
                    LOG.debug("Getting hdisk name from vios %(vios_name)s "
                              "for volume %(volume_id)s. "
                              "volume_uid: %(volume_uid)s."
                              % {'volume_id': volume_id,
                                 'volume_uid': volume_uid,
                                 'vios_name': vios_name}, instance=self.lpar)

                    backing_dev = vios_wrapper.get_hdisk_from_uid(volume_uid)

                    LOG.info(_("%(hdisk)s is the vscsi backing device on vios "
                               "%(vios_name)s for volume %(volume_id)s. "
                               "volume_uid: %(volume_uid)s.") %
                             {'hdisk': backing_dev,
                              'volume_id': volume_id,
                              'volume_uid': volume_uid,
                              'vios_name': vios_name}, instance=self.lpar)

                    if not backing_dev:
                        LOG.info(_("No mapped device found on vios "
                                   "%(vios_name)s for volume %(volume_id)s. "
                                   "volume_uid: %(volume_uid)s ")
                                 % {'volume_id': volume_id,
                                    'volume_uid': volume_uid,
                                    'vios_name': vios_name},
                                 instance=self.lpar)
                except Exception as e:
                    LOG.error(_("Failed to get associated disk on vios "
                                "%(vios_name)s for volume %(volume_id)s. "
                                "volume_uid: %(volume_uid)s. Error=%(error)s")
                              % {'volume_id': volume_id,
                                 'volume_uid': volume_uid,
                                 'vios_name': vios_name,
                                 'error': six.text_type(e)},
                              instance=self.lpar)

                if backing_dev:
                    storage_map = scsi_map_processor.SCSIMapProcessor()
                    lpar_wrapper = LPAREntryWrapper(self.lpar_entry)
                    lpar_id = lpar_wrapper.get_lpar_id()
                    LOG.info(_("Detaching and deleting adapter "
                               "on vios %(vios)s. Backing_device=%(dev)s.")
                             % {'vios': vios_name, 'dev': backing_dev},
                             instance=self.lpar)
                    storage_map.\
                        detach_and_delete_adapter(oper, self.host.uuid,
                                                  vios_uuid,
                                                  client_lpar_id=lpar_id,
                                                  backing_dev=backing_dev)
                    if CONF.enable_rmdev:
                        LOG.info(_("Removing volume device %(dev)s on "
                                   "vios %(vios_name)s.") %
                                 {'dev': backing_dev,
                                  'vios_name': vios_name}, instance=self.lpar)
                        job_status = lpar.\
                            _invoke_rmdev(self.host,
                                          CONF.host_display_name,
                                          vios_name,
                                          backing_dev,
                                          vios_uuid)
                        LOG.info(_("Status=%(status)s "
                                   "Task='Remove device %(volume)s from "
                                   "vios %(vios)s'.") %
                                 {'volume': backing_dev,
                                  'vios': vios_name,
                                  'status': job_status}, instance=self.lpar)
            except Exception as e:
                # It is okay to move on even if it fails to clean-up.
                LOG.warn(six.text_type(e), instance=self.lpar)
                pass

    @logcall
    def _unmap_vscsi_volumes_from_host(self, context, instance, bdmaps):
        """
        This function is running on source host/target host based
        on its caller.

        :param instance: VM instance
        :param bdmaps: bdmaps for the cinder volumes on host
        """
        LOG.debug("instance_uuid=%s, number of BDM maps=%d." %
                  (instance['uuid'], 0 if not bdmaps else len(bdmaps)))
        scg_map = StorageMapping()
        # Since we are unmapping the whole group of mapped volumes, we
        # would not want the following call to raise a connectivity
        # exception for the case where the host would be excluded from
        # either boot or data connectivity now. So we use default CHK_EITHER
        # to be the most lenient. Should consider a CHK_NONE type
        # suppor in the future.
        scg_info = scg_map.get_connectivity_info(context, instance['uuid'],
                                                 self.host.host_id,
                                                 instance, live_data=False)
        # extract scg class to six.text_type to avoid UnicodeEncodeError
        scg_class = scg_info.pop('scg')
        LOG.debug('scg_info: %s, scg_class: %s' %
                  (scg_info, six.text_type(scg_class)), instance=self.lpar)
        scg_info['scg'] = scg_class

        backend_host_group = scg_info.get('backend_host_group')
        LOG.info(_("Backend_host_group: %s.") % backend_host_group,
                 instance=self.lpar)

        # continue to process attached cinder volumes if any
        if bdmaps is None:
            # no attached cinder volume.
            LOG.info(_("No attached cinder volumes for processing."),
                     instance=self.lpar)
            return

        LOG.info(_("Processing volume mapping removals from host..."),
                 instance=self.lpar)
        for bdm in bdmaps:
            connection_info = bdm['connection_info']
            if connection_info is None:
                LOG.error(_("Connection_info is missing in bdm. bdm=%s.")
                          % bdm, instance=self.lpar)
                continue
            else:
                if connection_info and isinstance(connection_info, basestring):
                    connection_info = jsonutils.loads(connection_info)

            volume_id = connection_info.get('data').get('volume_id')
            volume_type = connection_info.get('driver_volume_type')
            if volume_type != 'fibre_channel':
                LOG.debug("Volume type is not fibre_channel."
                          " So, skipping volume %s." % volume_id,
                          instance=self.lpar)
                continue

            connection_type = lpar._get_volume_connection_type(connection_info)
            if connection_type is None:
                msg = (_("Could not identify the connection type of a volume. "
                         "Volume: %s.") % volume_id)
                LOG.error(_("Could not identify the connection type of a "
                            "volume. volume: %(volume_id)s. "
                            "connection_info: %(connection_info)s.")
                          % {'volume_id': volume_id,
                             'connection_info': connection_info},
                          instance=self.lpar
                          )
            if connection_type == CONNECTION_TYPE_VSCSI:
                LOG.info(_("Processing a vscsi volume. volume=%s.") %
                         volume_id, instance=self.lpar)
                connector = lpar.\
                    get_volume_connector(self.host, instance,
                                         self.lpar['display_name'],
                                         self.lpar_uuid,
                                         self.lpar_entry)
                if connector is None:
                    msg = (_("Cannot get volume connector to "
                             "virtual machine: %s.") %
                           self.lpar['display_name'])
                    LOG.warn(msg, instance=self.lpar)
                    continue

                volume_id = connection_info.get('data').get('volume_id')
                vios_ports_list = connector.get('wwpn_vscsi')
                LOG.debug("List of VIOS'es and their WWPNS maps %s." %
                          vios_ports_list, instance=self.lpar)
                # Loop through VIOS's and remove volumes and vol. connections.
                if 'connection_type' not in connector:
                    connector.update({'connection_type':
                                      [CONNECTION_TYPE_VSCSI]})

                for vios_wwpns in vios_ports_list:
                    try:
                        connector['wwpns'] = vios_wwpns['wwpns']
                        connector['host'] = vios_wwpns['vios_host']
                        # Detach adapters and remove volume device from VIOS.
                        volume_uid = connection_info.get('data').\
                            get('target_UDID')
                        vios_name = vios_wwpns['vios_name']
                        vios_uuid = vios_wwpns['vios_uuid']
                        self._remove_volume_device_from_vios(vios_name,
                                                             vios_uuid,
                                                             volume_id,
                                                             volume_uid)

                    except Exception as e:
                        # It is okay to move on even if it fails to remove
                        # volume device or fails to remove volume connection.
                        LOG.exception(_("Error: %s.") % six.text_type(e),
                                      instance=self.lpar)
                        pass

                # call terminate_connection once, which in turn parses
                # for each vios
                try:
                    if backend_host_group is None:
                        pv_vscsi.call_terminate_connection(
                            self.host, context, instance,
                            self.lpar['display_name'], volume_id,
                            connector, connection_info,
                            clean=CONF.migration_cleanup)
                except Exception as e:
                    LOG.exception(_("Failed to terminate connection %s") %
                                  six.text_type(e), instance=self.lpar)

    @logcall
    def live_migration(self, ctxt, instance, dest,
                       post_method, recover_method,
                       block_migration, migrate_data):
        """
        1) Issues K2 command to migrate the lpar
        2) Will recover the migration and rollback on failure

        :param ctxt: security context
        :param instance: lpar instance to be migrated
        :param dest: destination host
        :param post_method: post operation method.
        :param recover_method: recovery method when any exception occurs.
        :param block_migration: if true, migrate VM disk.
        :param migrate_data: dictionary of source and target host data
        """

        inst_name = self.lpar['display_name']

        try:
            stg_mappings = self._prep_source_for_migration(ctxt, migrate_data)
            if 'fc_mappings' in stg_mappings:
                migrate_data.fc_mappings = stg_mappings['fc_mappings']
            if 'scsi_mappings' in stg_mappings:
                migrate_data.scsi_mappings = stg_mappings['scsi_mappings']
            # If available, use the 'vfc_lpm_mappings' from
            # pre_live_migration_result.
            # Note: Only NovaLink managed destination hosts
            #       would provide 'vfc_lpm_mappings' in the
            #       pre_live_migration_result.
            if hasattr(migrate_data, 'vol_data'):
                pre_lm_result = migrate_data.vol_data
                vfc_lpm_mappings = pre_lm_result.get('vfc_lpm_mappings')
                if vfc_lpm_mappings:
                    vfc_lpm_mappings = jsonutils.loads(vfc_lpm_mappings)
                    migrate_data.fc_mappings = ",".join(vfc_lpm_mappings)
                    LOG.info(_("FC mappings for live migration=%s") %
                             migrate_data.fc_mappings)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                recover_method(ctxt, instance, dest,
                               block_migration, migrate_data)

        with K2OperGuard(host=self.host) as k2oper:
            start_time = time.time()
            try:
                self._migrate(k2oper, migrate_data)
            except exception.IBMPowerVMJobRequestTimedOut as exc:
                LOG.exception(exc)
                try:
                    LOG.info(_("Live migration task for virtual machine "
                               "%(lpar)s timed out. "
                               "Aborting migration.") %
                             {'lpar': inst_name})
                    self._migrate_abort(k2oper)
                except Exception as e:
                    LOG.exception(_("Live migration abort for virtual machine "
                                    "%(lpar)s failed with error '%(error)s'") %
                                  {'lpar': inst_name,
                                   'error': six.text_type(e)})
                # Rollback the migration on the source
                self._rollback_live_migration(k2oper)
                recover_method(ctxt, instance, dest,
                               block_migration, migrate_data)
                error = self._intercept_hmc_exception(exc)
                raise exception.IBMPowerVMMigrationFailed(error)
            except Exception as e:
                LOG.exception(e)
                # Rollback the migration on the source
                self._rollback_live_migration(k2oper)
                recover_method(ctxt, instance, dest,
                               block_migration, migrate_data)
                error = self._intercept_hmc_exception(e)
                raise exception.IBMPowerVMMigrationFailed(error)

            # After the migration task completes we want to continuously
            # parse the state of the lpar waiting for the migration to
            # complete.  We will timeout after a configurable period of time.
            self._refresh_lpar_on_host(k2oper)
            if self.lpar_wrapper:
                timeout = CONF.ibmpowervm_live_migration_timeout
                migr_state = self.lpar_wrapper.get_migration_state()
                while migr_state != 'Migration Complete':
                    curr_time = time.time()
                    if (curr_time - start_time) > timeout:
                        try:
                            LOG.info(_("Live migration for virtual machine "
                                       "%(lpar)s timed out. "
                                       "Aborting migration.")
                                     % {'lpar': inst_name})
                            self._migrate_abort(k2oper)
                        except Exception as e:
                            LOG.exception(_("Live migration abort for virtual "
                                            "machine %(lpar)s failed with "
                                            "error '%(error)s'") %
                                          {'lpar': inst_name,
                                           'error': six.text_type(e)})
                        # Try Migrate Recovery
                        self._rollback_live_migration(k2oper)
                        recover_method(ctxt, instance, dest,
                                       block_migration, migrate_data)
                        error = (_("Live migration for virtual machine "
                                   "%(lpar)s timed out.")
                                 % {'lpar': inst_name})
                        LOG.exception(error)
                        raise exception.IBMPowerVMMigrationFailed(error)

                    # Try to read the lpar. If it still exists check the state.
                    self._refresh_lpar_on_host(k2oper)
                    if self.lpar_wrapper:
                        migr_state = self.lpar_wrapper.get_migration_state()
                    else:
                        break

                    greenthread.sleep(1)

            # Migration is now complete.  Before running post ops we need to
            # Wait for the lpar to be removed from the source host
            while self.lpar_entry:
                self._refresh_lpar_on_host(k2oper)
                greenthread.sleep(1)

        # Post method to update host in OpenStack and finish live-migration
        post_method(ctxt, instance, dest, block_migration, migrate_data)

    @logcall
    def rollback_live_migration_at_destination(self, k2_uuid):
        """
        Rollback the migration on the destination if the
        virtual machine exists there.

        :param lpar_uuid: k2 uuid of lpar being migrated
        """
        self.lpar_uuid = k2_uuid
        lpar_name = self.lpar['display_name']

        # Update the destination host lpars to see if the lpar migrated
        self._refresh_lpar_cache()

        # Recover the lpar if it exists.
        if self.lpar_wrapper is not None:

            try:
                LOG.warn(_("Virtual machine %(inst)s exists on destination "
                           "host %(host)s following a failing migration.")
                         % {'inst': lpar_name,
                            'host': self.host_name})
                with K2OperGuard(host=self.host) as k2oper:
                    self._rollback_live_migration(k2oper)
            except Exception:
                LOG.warn(_("Unable to recover virtual machine %(inst)s from "
                           "destination host "
                           "%(host)s.") % {'inst': lpar_name,
                                           'host': self.host_name})

    @logcall
    def post_live_migration_at_destination(self, ctxt, k2_uuid):
        """
        Refreshes the lpar cache to pick up the entry on
        the destination host, then adds the lpar uuid to
        the uuid cache. Also spawns a thread to wait for
        RMC to become active so the lpar cache and the GUI
        have the most current RMC state.

        :param ctxt: security context
        :param lpar_uuid: k2 uuid of lpar being migrated
        """
        self.lpar_uuid = k2_uuid
        lpar_name = self.lpar['display_name']

        timeout = CONF.ibmpowervm_live_migration_timeout
        start_time = time.time()
        self._refresh_lpar_cache()
        while self.lpar_wrapper is None:
            curr_time = time.time()
            if (curr_time - start_time) > timeout:
                # timed out, instance not on destination
                error = (_("Timed out waiting for virtual machine "
                           "%(inst)s to arrive on destination host "
                           "%(host)s.") % {'inst': lpar_name,
                                           'host': self.host_name})
                LOG.exception(error)
                raise exception.IBMPowerVMMigrationFailed(error)
            greenthread.sleep(2)
            self._refresh_lpar_cache()

        # Add the new uuid of target lpar to uuid cache
        self.host.pvm_uuid_cache.add(self.lpar['name'], self.lpar_uuid)

        # clear the lpar slot info for vfc mappings. The new
        # mapping gets updated as part of inventory instance
        utils.clear_lpar_slot_info(self.lpar)
        # For AIX/Linux Spawn a thread to check for error ref code
        if IBMI_OS_TYPE != self.lpar_wrapper.get_lpar_type():
            greenthread.spawn(self._check_vm_for_error_ref_codes, ctxt)

    @logcall
    def _migrate(self, k2oper, migrate_data):
        """
        Performs the migration of the lpar

        :param k2oper: K2 operator
        :param migrate_data: dictionary of source host data
        """
        target_host = migrate_data.dest_sys_name
        target_user = migrate_data.dest_user_id
        target_ip = migrate_data.dest_ip
        source_ip = self.hmc.ipaddr

        # Get the K2 response for the migrate command
        k2resp = k2oper.read(k2const.LPAR, self.lpar_uuid,
                             suffixType=k2const.SUFFIX_TYPE_DO,
                             suffixParm=k2const.SUFFIX_PARM_MIGRATE)

        # Create the job parameters
        job_wrapper = jobentry_wrapper.JobEntryWrapper(k2resp.entry)

        job_parms = [job_wrapper.create_job_parameter(
            'TargetManagedSystemName', target_host)]
        if source_ip != target_ip:
            job_parms.append(job_wrapper.create_job_parameter(
                'TargetRemoteHMCIPAddress', target_ip))
            job_parms.append(job_wrapper.create_job_parameter(
                'TargetRemoteHMCUserID', target_user))

        # Create the virtual FC mappings parameter
        if hasattr(migrate_data, 'fc_mappings'):
            fc_mappings = migrate_data.fc_mappings
            if fc_mappings:
                job_parms.append(job_wrapper.create_job_parameter(
                    'VirtualFCMappings', fc_mappings))

        # Create the virtual SCSI mappings parameter
        if hasattr(migrate_data, 'scsi_mappings'):
            scsi_mappings = migrate_data.scsi_mappings
            if scsi_mappings:
                job_parms.append(job_wrapper.create_job_parameter(
                    'VirtualSCSIMappings', scsi_mappings))

        # Create the source/target MSP inputs
        if hasattr(migrate_data, 'dest_msp'):
            dest_msp = migrate_data.dest_msp
            job_parms.append(job_wrapper.create_job_parameter(
                'DestMSPIPaddr', dest_msp))

        if hasattr(migrate_data, 'src_msp'):
            src_msp = migrate_data.src_msp
            if src_msp:
                job_parms.append(job_wrapper.create_job_parameter(
                    'SourceMSPIPaddr', src_msp))

        if hasattr(migrate_data, 'dest_spp_id'):
            spp_id = migrate_data.dest_spp_id
            if spp_id:
                job_parms.append(job_wrapper.create_job_parameter(
                    'SharedProcPoolID', six.text_type(spp_id)))

        # Submit the job
        job_wrapper.run_job(k2oper, self.lpar_uuid,
                            job_parms=job_parms, host_uuid=self.host_uuid)

    @logcall
    def _rollback_live_migration(self, k2oper):
        """
        Calls migrate recover.  If first attempt fails, it
        is retried with the force option.

        :param k2oper: K2 operator
        """
        lpar_name = self.lpar['display_name']
        try:
            LOG.info(_("Live migration of virtual machine %(lpar)s "
                       "failed. Initiating rollback.")
                     % {'lpar': lpar_name})
            self._migrate_recover(k2oper)
        except Exception as recover_err:
            try:
                LOG.debug("Migration recovery for virtual machine "
                          "%(lpar)s failed. Retrying with force "
                          "option." % {'lpar': lpar_name})
                LOG.exception(recover_err)
                self._migrate_recover(k2oper, force=True)
            except Exception as fe:
                LOG.debug("Rollback of virtual machine %(lpar)s "
                          "failed." % {'lpar': lpar_name})
                LOG.exception(fe)

    @logcall
    def _migrate_recover(self, k2oper, force=False):
        """
        Performs migration recovery of the lpar

        :param k2oper: K2 operator
        :param force: force option
        """

        # Get the K2 response for the migrate command
        k2resp = k2oper.read(k2const.LPAR, self.lpar_uuid,
                             suffixType=k2const.SUFFIX_TYPE_DO,
                             suffixParm=k2const.SUFFIX_PARM_MIGRATE_RECOVER)

        # Submit the job
        job_wrapper = jobentry_wrapper.JobEntryWrapper(k2resp.entry)
        job_parms = None
        if force:
            job_parms = [job_wrapper.create_job_parameter('Force', 'true')]
        # need to provide the host uuid
        # since _migrate_recover only get called from source compute host,
        # self.host_uuid can be used
        job_wrapper.run_job(k2oper, self.lpar_uuid, job_parms=job_parms,
                            host_uuid=self.host_uuid)

    @logcall
    def _migrate_abort(self, k2oper):
        """
        Performs migration abort of the lpar

        :param k2oper: K2 operator
        """

        # Get the K2 response for the migrate abort command
        k2resp = k2oper.read(k2const.LPAR, self.lpar_uuid,
                             suffixType=k2const.SUFFIX_TYPE_DO,
                             suffixParm=k2const.SUFFIX_PARM_MIGRATE_ABORT)

        # Submit the job
        job_wrapper = jobentry_wrapper.JobEntryWrapper(k2resp.entry)
        job_parms = None
        # need to provide the host uuid
        # since _migrate_abort only get called from source compute host,
        # self.host_uuid can be used
        job_wrapper.run_job(k2oper, self.lpar_uuid, job_parms=job_parms,
                            host_uuid=self.host_uuid)

    @logcall
    def _choose_msp_address(self, user_msps, valid_msps):
        """
        Picks an MSP address to use from a list of valid addresses.

        :param user_msps: A list of MSP addresses given by the user
        :param valid_msps: A list of valid MSP addresses for all VIOS on
                           the host
        :returns msp: An MSP IP address
        """
        # We need to verify the list of addresses provided by the user
        # are valid before randomly selecting one of them.
        valid_user_msps = []
        for msp in user_msps:
            if msp in valid_msps:
                if msp not in valid_user_msps:
                    valid_user_msps.append(msp)
            else:
                LOG.warn(_('The mover service partition %(msp)s is '
                           'invalid for host %(host)s. '
                           'It will be ignored.') % {'msp': msp,
                                                     'host': self.host_name})

        choice = ''
        if valid_user_msps:
            choice = random.choice(valid_user_msps)
        return choice

    @logcall
    def _get_lpar_proc_compat_mode(self, lpar_state):
        """
        Checks the lpar state and returns either the lpar's current
        or pending processor compatibility mode

        :param lpar_state: The power state of the lpar
        :returns lpar_compat_mode: The lpar's compatibility mode
                                   or 'None' if it's not found
        """
        lpar_compat_mode = None
        power_specs = self.lpar.get('power_specs')

        if lpar_state == power_state.RUNNING:
            lpar_compat_mode = power_specs.get(
                'current_compatibility_mode')
        else:
            lpar_compat_mode = power_specs.get(
                'desired_compatibility_mode')

        return lpar_compat_mode

    @logcall
    def _get_migration_stats(self, lpar_state):
        """
        Checks the lpar state and returns the host stats for number
        of migrations supported and number in progress

        :param lpar_state: The power state of the lpar
        :returns mig_allowed: Number of migrations supported by the host
        :returns mig_running: Number of migrations in progress
        """
        mig_allowed = self.host_stats.get('stats'). \
            get('active_migrations_supported')
        mig_running = self.host_stats.get('stats'). \
            get('active_migrations_in_progress')
        if lpar_state != power_state.RUNNING:
            mig_allowed = self.host_stats.get('stats'). \
                get('inactive_migrations_supported')
            mig_running = self.host_stats.get('stats'). \
                get('inactive_migrations_in_progress')

        return mig_allowed, mig_running

    @logcall
    def _prep_source_for_migration(self, ctxt, migrate_data):
        """
        Prepares the source host for migration

        1) detaches any optical devices
        2) removes any orphaned adapters
        3) gathers data for the fc and vscsi mapping strings
           if destination host is not managed by novalink.

        :param ctxt: security context
        :param migrate_data: dictionary of source and target host data
        :returns stg_mappings: dictionary containing mapping strings for both
                               fc and scsi mappings
        """
        scsi_map = scsi_map_processor.SCSIMapProcessor()
        fc_map = fc_map_processor.FCMapProcessor()

        lpar_id = self.lpar_wrapper.get_lpar_id()
        vios_fc_data = []
        vios_scsi_data = []

        for vios_entry in self.vios_entries:

            vios_wrapper = viosentry_wrapper.VIOSEntryWrapper(vios_entry)
            vios_uuid = vios_wrapper.get_uuid()
            vios_name = vios_wrapper.get_vios_name()

            # Detach the vopt devices off of the instance
            opticals = scsi_map.find_attached_medias(vios_entry.element,
                                                     lpar_id)
            if opticals:
                LOG.debug("VIOS %(vios)s (%(uuid)s) maps the following "
                          "optical devices to the lpar: %(opts)s. They will "
                          "be detached prior to migration."
                          % {'vios': vios_name,
                             'uuid': vios_uuid,
                             'opts': opticals})
                self._detach_opticals(vios_uuid, vios_name, lpar_id)

            # Clean up server storage adapters that do not have a matching
            # client adapter.  This will clean up both scsi and fc adapters.
            orphans = scsi_map.check_for_orphaned_adapters(vios_entry, lpar_id,
                                                           ignore_type=True)
            if orphans:
                LOG.debug("Removing server storage adapters from vios "
                          "%(vios)s (%(uuid)s) that do not have a matching "
                          "client adapter for lpar id %(id)s."
                          % {'vios': vios_name,
                             'uuid': vios_uuid,
                             'id': lpar_id})
                self._remove_orphans(scsi_map, vios_uuid, vios_name, lpar_id)

            # Gather the VFC mapping data needed to construct the
            # VirtualFCMappings parameter.
            fc_client_map_data = fc_map.extract_client_adapter_data(vios_entry,
                                                                    lpar_id)
            if fc_client_map_data is not None:
                LOG.debug("FC client mapping data for vios %(vios)s is "
                          "%(data)s." % {'vios': vios_name,
                                         'data': fc_client_map_data})
                vios_fc_data += fc_client_map_data

            # Gather the VSCSI mapping data needed to construct the
            # VirtualSCSIMappings parameter.
            scsi_client_map_data = scsi_map.extract_client_adapter_data(
                vios_entry, lpar_id)
            if scsi_client_map_data is not None:
                LOG.debug("SCSI client mapping data for vios %(vios)s is "
                          "%(data)s." % {'vios': vios_name,
                                         'data': scsi_client_map_data})
                vios_scsi_data += scsi_client_map_data

        stg_mappings = {}
        # If it is migration to NovaLink, we do not need to generate
        # the storage mapping strings here. We will use the 'vfc_lpm_mappings'
        # generated by novalink.
        #
        # Assumption: We do not really use vscsi_lpm_mappings(except ssp case)
        #             with the migrate job.
        dest_managed_by = None
        if hasattr(migrate_data, 'dest_managed_by'):
            dest_managed_by = migrate_data.dest_managed_by
        if dest_managed_by is None or dest_managed_by != NOVALINK_PLATFORM:
            stg_mappings = self.\
                _generate_storage_mapping_strings(ctxt, migrate_data,
                                                  vios_fc_data,
                                                  vios_scsi_data)
        return stg_mappings

    @logcall
    def _extract_instance_fabric_metadata(self, ctxt):
        """
        Extract instance's fabric meta-data.

        :param ctxt: security context
        :returns port_maps, slots, vios_slot_peers: dictionary containing
            fabric-portmaps ,fabric-slots and vios_slot_peers.
            Ex:
               fabric-portmaps {'A' : {'phy_wwpn1': ['vwwpn1', 'vwwpn2']..}
            The fabric slots represents the slots used by a given fabric
            connection
               fabric-slots {'A': [3,4], 'B':[5,6]}
            The vios slot peer gives the slots placed on the same vios.
               vios_slot_peer [[3,5], [4,6]]
        """
        fc_map = fc_map_processor.FCMapProcessor()

        lpar_id = self.lpar_wrapper.get_lpar_id()
        client_vfc_list = []
        phy_wwpn_virt_wwpn_mappings = {}
        vios_slot_peers = []
        for vios_entry in self.vios_entries:
            vios_wrapper = viosentry_wrapper.VIOSEntryWrapper(vios_entry)
            vios_name = vios_wrapper.get_vios_name()

            # Gather the VFC mapping data needed to construct the
            # VirtualFCMappings parameter.
            fc_client_map_data = fc_map.extract_client_adapter_data(vios_entry,
                                                                    lpar_id)

            if fc_client_map_data is not None:
                LOG.debug("FC client mapping data for vios %(vios)s is "
                          "%(data)s." % {'vios': vios_name,
                                         'data': fc_client_map_data})
                client_vfc_list += fc_client_map_data
                # Save the client slots used on a particular VIOS
                vios_slots = [fmap.get("client_slot") for fmap in
                              fc_client_map_data if fmap.get("client_slot")]
                vios_slot_peers.append(vios_slots)

            fc_wwpn_mappings = fc_map.extract_wwpn_mappings(vios_entry,
                                                            lpar_id)

            if fc_wwpn_mappings is not None:
                phy_to_virt_initiators = fc_wwpn_mappings.\
                    get('phy_to_virt_initiators')
                phy_wwpn_virt_wwpn_mappings.update(phy_to_virt_initiators)
        LOG.debug("Physical wwpns to Virtual wwpns map=%s"
                  % phy_wwpn_virt_wwpn_mappings)

        fc_phy_wwpns = [vfc['phys_wwpn'] for vfc in client_vfc_list]
        fc_phy_wwpns_to_fabric = su.get_db_wwpns_to_fabric_map(ctxt,
                                                               fc_phy_wwpns)
        LOG.debug("Physical wwpns to Fabric map=%s" % fc_phy_wwpns_to_fabric)
        fc_fabric_to_phy_wwpns = {}
        for wwpn, fabric in fc_phy_wwpns_to_fabric.iteritems():
            fc_fabric_to_phy_wwpns[fabric] = \
                fc_fabric_to_phy_wwpns.get(fabric, [])
            fc_fabric_to_phy_wwpns[fabric].append(wwpn)
        LOG.debug("Fabric to Physical wwpns map=%s" % fc_fabric_to_phy_wwpns)

        # Extract client slots per fabric.
        npiv_fabric_slots_map = {}
        for fabric in fc_fabric_to_phy_wwpns.iterkeys():
            client_slots = []
            phy_wwpns_mapped_to_fabric = fc_fabric_to_phy_wwpns.get(fabric)
            LOG.debug("Physical wwpns mapped to fabric %(fabric) are "
                      " %(phy_wwpns)",
                      {'fabric': fabric,
                       'phy_wwpns': phy_wwpns_mapped_to_fabric})
            for vfc in client_vfc_list:
                if (vfc['phys_wwpn'] and vfc['phys_wwpn']
                        in phy_wwpns_mapped_to_fabric):
                    if vfc['client_slot']:
                        client_slots.append(vfc['client_slot'])
            npiv_fabric_slots_map[fabric] = client_slots
        LOG.debug("FC slots map is %s" % npiv_fabric_slots_map)

        # Extract port-maps(phy_wwpns and virt wwpns) per fabric
        npiv_port_maps = {}
        for fabric in fc_fabric_to_phy_wwpns.iterkeys():
            LOG.debug("Extracting port-maps for fabric %s" % fabric)
            phy_virt_mappings = {}
            for phy_wwpn in fc_phy_wwpns:
                if phy_wwpn in fc_fabric_to_phy_wwpns.get(fabric):
                    phy_virt_mappings[phy_wwpn] = \
                        phy_wwpn_virt_wwpn_mappings.get(phy_wwpn)
            npiv_port_maps[fabric] = phy_virt_mappings
        LOG.debug("FC ports map is %s" % npiv_port_maps)

        return (npiv_port_maps, npiv_fabric_slots_map, vios_slot_peers)

    @logcall
    def _detach_opticals(self, vios_uuid, vios_name, lpar_id):
        """
        Detatch the virtual optical devices from the VIOS prior to live
        migration

        :param vios_uuid: UUID of vios containing a vopt to detach
        :param vios_name: Name of vios containing a vopt to detach
        :param lpar_id: lpar ID of migrating partition
        """
        try:
            optical = virtualoptical.VirtualOptical()
            optical.detach_all_vopt_from_vios(self.host, vios_uuid,
                                              lpar_id, self.lpar_uuid)
        except Exception as e:
            lpar_name = self.lpar['display_name']
            error = (_("Failed to detach the virtual optical "
                     "device for virtual machine %(lpar_name)s "
                     "from VIOS %(vios_name)s. Retry the migration. "
                     "If you still see this error, remove the virtual "
                     "optical device from your HMC manually and retry "
                     "the migration.") % locals())
            LOG.exception(error)
            LOG.exception(e)
            raise exception.IBMPowerVMMigrationFailed(error)

    @logcall
    def _remove_orphans(self, scsi_map, vios_uuid, vios_name, lpar_id):
        """
        Remove any orphaned fc/vscsi adapters from the VIOS prior to live
        migration

        :scsi_map: storage mapping wrapper
        :param vios_uuid: UUID of vios containing an orphan
        :param vios_name: Name of vios containing a vopt to detach
        :param lpar_id: lpar ID of migrating partition
        """
        try:
            with K2OperGuard(host=self.host) as k2oper:
                scsi_map.cleanup_orphaned_adapters(
                    k2oper, self.host_uuid, vios_uuid,
                    client_lpar_id=lpar_id, ignore_type=True)
        except Exception as e:
            lpar_name = self.lpar['display_name']
            error = (_("One or more server adapters on VIOS %(vios)s "
                       "for host %(host)s are mapped to the migrating "
                       "virtual machine %(lpar)s without a "
                       "corresponding client adapter. PowerVC failed "
                       "while attempting to remove them. From the "
                       "HMC, manually remove any mappings for "
                       "adapters that are not mapped to both the VIOS "
                       "and the virtual machine. Then, try the "
                       "migration again.") % {'vios': vios_name,
                                              'host': self.host_name,
                                              'lpar': lpar_name})
            LOG.exception(error)
            LOG.exception(e)
            raise exception.IBMPowerVMMigrationFailed(error)

    @logcall
    def _generate_storage_mapping_strings(self, context, migrate_data,
                                          vios_fc_data, vios_scsi_data):
        """
        Generater the vfc and vscsi mapping strings to pass to the migrate
        command.

        The vfc mapping parameter is a comma delimited list of
        virtual fibre channel mappings for the target vios (one for each
        source VFC adapter) in the form
        <client slot num>//<target vios ID>//<vfc port name>
        (e.g. 2//1//fcs0,3//2//fcs1)

        The vscsi mapping parameter is similar, but in the form
        <client slot num>//<target vios ID>
        (e.g. 5//1,6//2

        :param context: security context
        :param migrate_data: dictionary of source and target host data
        :param vios_fc_data: list of fc data used to generate a mapping string
        :param vios_scsi_data: list of scsi data used to
                               generate a mapping string
        :returns stg_mappings: dictionary containing mapping strings for both
                               fc and scsi mappings
        """
        # Generate the string to pass to VirtualFCMappings
        stg_map = storage_mapping_api.StorageMapping()
        tgt_mtms = None
        tgt_sys = None
        if hasattr(migrate_data, 'dest_mtms_name'):
            tgt_mtms = migrate_data.dest_mtms_name
        if hasattr(migrate_data, 'dest_sys_name'):
            tgt_sys = migrate_data.dest_sys_name
        try:
            mappings = stg_map.get_hmc_migrate_map_strings(context,
                                                           vios_fc_data,
                                                           vios_scsi_data,
                                                           self.lpar,
                                                           tgt_mtms)
        except stgex.IBMPowerVCStorageError:
            # IBMPowerVCStorageError details has been written
            # in the compute log by SCG logging. translate it to the compute
            # migrate exception. Otherwise, it will be truncated
            # by instance fault, which is only 256 characters long.
            error = (_("Failed to generate Virtual Fibre Channel mapping on "
                       "the destination host %(dest_host)s based on storage "
                       "connectivity group configuration. For detailed "
                       "information about the migration failure, see "
                       "the nova compute log file "
                       "(/var/log/nova/nova-compute-%(host)s.log).")
                     % {'dest_host': tgt_sys,
                        'host': self.host.host_id})
            LOG.exception(error)
            raise exception.IBMPowerVMMigrationFailed(error)

        stg_mappings = {'fc_mappings': mappings['npiv'],
                        'scsi_mappings': mappings['ibm_ssp']}
        return stg_mappings

    @logcall
    def _refresh_lpar_on_host(self, k2oper):
        """
        Refreshes the lpar entry if the lpar still exists on the host

        :param k2oper: k2 operator
        """
        self.lpar_entry = None
        self.lpar_wrapper = None

        try:
            k2resp = k2oper.read(k2const.MGT_SYS, self.host_uuid,
                                 k2const.LPAR, self.lpar_uuid, age=0)
            self.lpar_entry = k2resp.entry
            self.lpar_wrapper = LPAREntryWrapper(self.lpar_entry)

        except K2Error as e:
            if (e.k2response and
                    e.k2response.status == (k2const.HTTP_STATUS_NOT_FOUND)):
                LOG.debug("Lpar with uuid %s no longer exists on host "
                          "%s. This is expected on the source host "
                          "following a successful migration, or on the "
                          "target host following a failing migration " %
                          (self.lpar_uuid, self.host_name))
            else:
                LOG.exception(e)

    @logcall
    def _refresh_lpar_cache(self):
        """
        Forces a refresh on the lpar cache to bring in the
        newly migrated lpar. Parses the feed using the k2
        uuid.
        """
        inst_name = self.lpar['display_name']
        host_name = self.host_name

        # force the lpar cache to update to pull in target lpar
        lpars_resp = None
        try:
            # use K2 level caching
            with K2OperGuard(host=self.host) as k2oper:
                lpars_resp = k2oper.read(k2const.MGT_SYS,
                                         rootId=self.host.uuid,
                                         childType=k2const.LPAR, age=0)
        except Exception as e:
            LOG.warn(_("An error occurred while searching for virtual "
                       "machine %(inst)s on host %(host)s.")
                     % {'inst': inst_name,
                        'host': host_name})
            LOG.exception(e)

        self.lpar_entry = None
        self.lpar_wrapper = None

        # Parse the feed to get the lpar entry by uuid
        if lpars_resp and lpars_resp.feed is not None:
            entries = lpars_resp.feed.entries
            for entry in entries:
                lpar_wrapper = LPAREntryWrapper(entry)
                if self.lpar_uuid == lpar_wrapper.get_uuid():
                    self.lpar_entry = entry
                    self.lpar_wrapper = LPAREntryWrapper(self.lpar_entry)
                    return

        # lpar not found on host
        LOG.warn(_("Migrated virtual machine %(inst)s cannot be found on "
                   "host %(host)s.") % {'inst': inst_name,
                                        'host': host_name})

    @logcall
    def _check_vm_for_error_ref_codes(self, ctxt):
        """
        Check VM for error ref codes
        :param ctxt: security context
        """

        lpar_state_checker.check_lpar_state(
            ctxt, self.lpar, self.host, self.lpar_uuid)

    @logcall
    def _intercept_hmc_exception(self, exception):
        """
        HMC error messages are not user friendly.  We want to inspect
        any error messages that come from the HMC, and print a helpful
        message. By default we'll indicate the HMC details can be found
        in the log file.

        :param exception: The exception from the HMC
        :returns error: An error message for the GUI
        """
        see_log_file_msg = (_("For detailed information about the "
                              "failure, see the nova compute log file "
                              "(/var/log/nova/nova-compute-%(suffix)s.log).")
                            % {'suffix': self.host.host_id})

        # Get the message from the HMC and log it
        if hasattr(exception, 'message'):
            hmc_err_msg = six.text_type(exception.message)
        else:
            hmc_err_msg = six.text_type(exception)
        LOG.exception(exception)

        # Query the error messages
        if hmc_err_msg.find("HSCLA369") >= 0:    # Issue 7620

            # Get the processor compatibility mode
            lpar_state = self.lpar.get('power_state')
            lpar_compat_mode = self._get_lpar_proc_compat_mode(lpar_state)

            error = (_("The virtual machine was deployed with a processor "
                       "compatibility mode of POWER7 or POWER8, "
                       "which was later changed to %s. The virtual machine "
                       "cannot be migrated to a "
                       "POWER6 host because the MAC address was generated by "
                       "PowerVC during deploy. To migrate to a POWER6 host, "
                       "the MAC address must be generated by the HMC. To "
                       "migrate from a POWER7 or POWER8 to a POWER6 host, "
                       "you must initially deploy to a POWER7 or POWER8 "
                       "system with the processor compatibility mode set to a "
                       "POWER6 derivative, or initially deploy to POWER6 "
                       "host.")
                     % lpar_compat_mode)

        # Issue 8066
        elif hmc_err_msg.find("HSCLA31B") >= 0:
            error = (_("The virtual fibre channel adapter for the virtual "
                       "machine that you want to migrate does not have a "
                       "valid mapping to the VIOS on the destination host. "
                       "Ensure that the Fibre Channel ports are assigned to "
                       "registered fabrics for all of your hosts. Refer to "
                       "the Configuration page in PowerVC. Then, try the "
                       "migration again. %s" % see_log_file_msg))

        # Check for failure due to orphaned adapters
        elif ((hmc_err_msg.find("HSCL400A") >= 0) and
                (hmc_err_msg.find("Client mapped to") >= 0) and
                (hmc_err_msg.find("more than once") >= 0)):
            error = (_("One or more server adapters on the VIOS for the source"
                       " host are mapped to the migrating virtual machine "
                       "without a corresponding client adapter. From the HMC, "
                       "manually remove any mappings for adapters that are not"
                       " mapped to both the VIOS and the virtual machine. "
                       "Then, try the migration again. %s" % see_log_file_msg))

        # Issue 9108
        # stg defect 44404
        elif (hmc_err_msg.find("HSCL03EA") >= 0 or
              hmc_err_msg.find("HSCL363A") >= 0):
            error = (_("There is an insufficient number of processors "
                       "available on the destination host to activate the "
                       "migrated virtual machine.  Either modify the existing "
                       "virtual machine to use less resource, or deactivate "
                       "any unused virtual machines consuming available "
                       "resource on the destination host. Then, try the "
                       "migration again. %s" % see_log_file_msg))

        # Issue 9462
        elif hmc_err_msg.find("HSCL03EC") >= 0:
            error = (_("There is insufficient memory available on the "
                       "destination host to activate the migrated virtual "
                       "machine.  Either modify the existing virtual machine "
                       "to use less resource, or deactivate any unused "
                       "virtual machines consuming available resource on the "
                       "destination host. Then, try the migration "
                       "again. %s" % see_log_file_msg))

        elif ((hmc_err_msg.find("HSCL400A") >= 0) and
              (hmc_err_msg.find("Inactive needs to be preceeded by at least "
                                "1 successful live migration for wwpn "
                                "propogation") >= 0) and
              (hmc_err_msg.find("Do not have wwpn") >= 0)):
            error = (_("One or more NPIV attached volumes are not "
                       "discovered on the virtual machine that you want to "
                       "migrate. Either log on to the virtual machine and "
                       "discover the NPIV attached volumes, or reboot the "
                       "virtual machine. Then, try the migration again. %s" %
                       see_log_file_msg))

        elif ((hmc_err_msg.find("HSCLB603") >= 0) and
              (hmc_err_msg.find("A partition migration operation to a managed "
                                "system that is currently the source managed "
                                "system in an outgoing partition migration "
                                "operation is not allowed.") >= 0)):
            error = (_("The destination host is currently serving as the "
                       "source in a separate migration operation. "
                       "Please try the migration again later. %s" %
                       see_log_file_msg))

        elif (hmc_err_msg.find("HSCLA319") >= 0):
            error = (_("The Virtual I/O Server (VIOS) partitions on the "
                       "destination managed system cannot host the migrating"
                       " partition's virtual fibre channel client adapter. %s"
                       % see_log_file_msg))

        elif ((hmc_err_msg.find("HSCLA281") >= 0) or
              (hmc_err_msg.find("HSCLB07C") >= 0)):
            # No need to search for HSCLB07C error message because
            # what we provide by default is suffificent.
            extracted_error = self._extract_hmc_error_by_code(hmc_err_msg,
                                                              "HSCLA281")
            if extracted_error is None:
                error = (_("The operation to query virtual I/O server "
                           "partition for completion has failed. %s" %
                           see_log_file_msg))
            else:
                error = (_("%(extracted_error)s %(see_log_file_msg)s") %
                         {'extracted_error': extracted_error,
                          'see_log_file_msg': see_log_file_msg})

        elif hmc_err_msg.find("HSCL3202") >= 0:
            error = (_("The HMC timed out while trying to acquire a lock. "
                       "Restart the Hardware Management Console (HMC). If the "
                       "problem persists, contact your next level of support "
                       "or your service provider."))

        else:
            error = (_("The migration command issued to the HMC failed. "
                       "Ensure that source and target Virtual I/O Servers are "
                       "running and with an RMC status of Active. %s"
                       % see_log_file_msg))

        return error

    def _extract_hmc_error_by_code(self, exception_msg, error_code):
        """
        Extracts the HMC error (by HMC error code) from the exception message.

        :param exception_msg: exception message thrown by HMC.
        :param error_code: HMC error code we are interested in.
        """
        pattern = "(.*)" + error_code + "(.*)\n(.*)"
        try:
            match = re.search(pattern, exception_msg)
            if match is not None:
                return error_code + match.group(2)
        except Exception as e:
            LOG.debug("Failed to retrieve hmc error by error code. Error: %s" %
                      six.text_type(e))

        return None

    def _get_hmc_public_key(self):
        """
        If this is HMC to Novalink migration, we need to get the HMC public
        key and pass the key to the target NovaLink host.

        :param dest_check_data: dictionary of target host data
        """
        LOG.debug("Entering _get_hmc_public_key/migrate.py")

        # Get the source IP address
        source_ip = self.hmc.ipaddr
        with K2OperGuard(host=self.host) as k2oper:
            try:
                # Get the K2 response for the management console
                k2resp = k2oper.read(k2const.MGT_CONSOLE)
                ssh_key = k2resp.feed.entries[0].element.find(
                    'PublicSSHKeyValue')
                public_key = ssh_key.gettext()
            except Exception as e:
                error = six.text_type(e)
                message = (_("Unable to get public key of source HMC "
                             "with IP address: %(source_ip)s. %(error)s")
                           % {'source_ip': source_ip, 'error': error})
                LOG.exception(message)
                raise MigrationPreCheckError(reason=message)

        return public_key

    def _add_novalink_public_key(self, migrate_data, novalink_pub_key):
        """
        If this is Novalink to HMC migration, we need to add the Novalink
        public key and pass the key to the target HMC host.

        :param dest_check_data: dictionary of target host data
        """
        LOG.debug("Entering _add_novalink_public_key/migrate.py")

        # Get the source host name and target IP addresses
        found = False
        source_name = migrate_data.src_sys_name
        target_ip = self.hmc.ipaddr

        with K2OperGuard(host=self.host) as k2oper:
            try:
                # Get the K2 response for the management console
                k2resp = k2oper.read(k2const.MGT_CONSOLE)
                console = k2resp.feed.entries[0]
                console_id = console.properties['id']
                console_etag = console.properties['etag']

                # Discover if NovaLink public key is already an authorized key
                authorized_keys = console.element.find('AuthorizedKeysValue')
                if authorized_keys is not None:
                    for auth_key in authorized_keys.findall('AuthorizedKey'):
                        if auth_key.gettext() == novalink_pub_key:
                            LOG.debug('Found a matching publickey.')
                            found = True

                # Add Novalink public key to HMC's authorized list of keys
                if not found:
                    LOG.debug('Did not find a matching public key.')
                    public_key = K2Element('AuthorizedKey',
                                           text=str(novalink_pub_key))
                    if authorized_keys is None:
                        # Need to insert new element after the
                        # PublicSSHKeyValue element, or else an
                        # XML validation error will occur
                        index = -1
                        for i in range(len(console.element)):
                            if (console.element[i].gettag() ==
                                    'PublicSSHKeyValue'):
                                index = i + 1
                                break

                        authorized_keys = K2Element('AuthorizedKeysValue',
                                                    attrib=attrib,
                                                    children=[public_key])
                        console.element.insert(index, authorized_keys)
                    else:
                        authorized_keys.append(public_key)
                    # Update the list of authorized keys for the
                    # management console
                    k2oper.update(k2resp.feed.entries[0].element, console_etag,
                                  k2const.MGT_CONSOLE, console_id)
            except Exception as e:
                error = six.text_type(e)
                message = (_("Unable to add public key of Novalink host "
                             "%(source_name)s to target HMC with IP address: "
                             "%(target_ip)s. %(error)s")
                           % {'source_name': source_name,
                              'target_ip': target_ip, 'error': error})
                LOG.exception(message)
                raise MigrationPreCheckError(reason=message)

    # Do not use logcall because passwords will be printed to the logs
    def _ensure_hmc_to_hmc_connection(self, dest_check_data):
        """
        If this is HMC to HMC migration we need to establish the
        SSH connection to the target HMC. If we're migrating
        between hosts managed by the same HMC, we do not need
        to set up an SSH connection

        :param dest_check_data: dictionary of target host data
        """
        # Do not use logcall because the enter/exit parms contain a password
        logged_dest_data = copy.deepcopy(dest_check_data)
        logged_dest_data.dest_cred = '********'
        LOG.debug("Entering _ensure_hmc_to_hmc_connection/migrate.py: "
                  "dest_check_data=%s" % logged_dest_data)

        # Get the source and target IP addresses
        source_ip = self.hmc.ipaddr
        dest_ip = dest_check_data.dest_ip

        # Set up the SSH connection if the source and target
        # have different IP addresses.
        if source_ip != dest_ip:
            LOG.debug(("Setup ssh key exchange between source HMC ip:" +
                       "%(source_ip)s and destination hmc ip: %(dest_ip)s") %
                      locals())
            user = dest_check_data.dest_user_id

            # The password is stored encrypted (STG defect 22189)
            encrypted_pwd = dest_check_data.dest_cred
            password = EncryptHandler().decode(encrypted_pwd)

            with K2OperGuard(host=self.host) as k2oper:
                try:
                    # Test to see if an exchange is required
                    self._exchange_ssh_keys(k2oper, dest_ip, user,
                                            password, test=True)
                except exception.IBMPowerVMJobRequestFailed as exc:
                    # Failed the test, do the exchange
                    try:
                        self._exchange_ssh_keys(
                            k2oper, dest_ip, user, password)
                    except exception.IBMPowerVMJobRequestFailed as exc:
                        error = six.text_type(exc)
                        message = (_(
                            "Unable to exchange SSH keys between the "
                            "source HMC with IP address: %(source_ip)s "
                            "and the destination HMC with IP address: "
                            "%(dest_ip)s. %(error)s") % locals())
                        LOG.exception(message)
                        raise MigrationPreCheckError(reason=message)

        # Do not use logcall because the enter/exit parms contain a password
        LOG.debug("Exiting _ensure_hmc_to_hmc_connection/migrate.py")

    def _exchange_ssh_keys(self, k2oper, ip, user, password, test=False):
        """
        Exchange ssh keys with the target hmc

        :param k2oper: K2 operator
        :param ip: ip address of target hmc
        :param user: hmc user id
        :param password: hmc user password
        :param test: if true, add test option
        """
        if test:
            ssh_cmd = 'mkauthkeys --ip ' + ip + ' -u ' + user + ' --test'
        else:
            ssh_cmd = ('mkauthkeys --ip ' + ip + ' --passwd ' + password +
                       ' -u ' + user)
        pwd_removed = ssh_cmd.replace(password, '********')
        LOG.debug(("Exchange ssh keys command: %s") % pwd_removed)

        # Get the K2 response for the CLIRunner command
        k2resp = k2oper.read(k2const.MGT_CONSOLE, None,
                             suffixType=k2const.SUFFIX_TYPE_DO,
                             suffixParm=k2const.SUFFIX_PARM_CLI_RUNNER)

        # Create the job parameters
        job_wrapper = jobentry_wrapper.JobEntryWrapper(k2resp.entry)
        ack_parm = 'acknowledgeThisAPIMayGoAwayInTheFuture'
        job_parms = [job_wrapper.create_job_parameter('cmd', ssh_cmd),
                     job_wrapper.create_job_parameter(ack_parm, 'true')]
        job_wrapper.run_job(k2oper, None, job_parms=job_parms,
                            sensitive=True)
