# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import six
import copy
import re

from eventlet import greenthread

from oslo_log import log as logging
from nova.objects import compute_node
from powervc_nova import logcall
from powervc_nova.objects.compute import dom as compute_dom
from powervc_nova.virt.ibmpowervm.common import exception, volume_utils,\
    migrate_utils
from powervc_nova.virt.ibmpowervm.hmc.k2 import constants as k2const
from powervc_nova.virt.ibmpowervm.hmc.k2 import jobentry_wrapper
from powervc_nova.virt.ibmpowervm.hmc.k2.msentry_wrapper \
    import ManagedSystemEntryWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2 import scsi_map_processor
from powervc_nova.virt.ibmpowervm.hmc.k2 import fc_map_processor
from powervc_nova.virt.ibmpowervm.hmc.k2 import viosentry_wrapper
from powervc_nova.virt.ibmpowervm.hmc.k2oprguard import K2OperGuard
from powervc_nova.storage import exception as stgex
from powervc_nova import _

from powervc_nova.storage.storage_mapping_api import StorageMapping
from powervc_nova.virt.ibmpowervm.common.constants \
    import CONNECTION_TYPE_VSCSI, PVM_SIMP_REMOTE_RESTART_CAPABLE, \
    PRS_MAX_PARA_DEST_REBUILDS, NUM_RR_IN_PROGRESS, CHK_ALL, CHK_BOOT, \
    HYPERVISOR_STATE, CONNECTION_TYPE_NPIV
import powervc_nova.virt.ibmpowervm.hmc.lpar as lpar
from powervc_nova.volume import cinder as volume
from oslo_serialization import jsonutils
from powervc_nova.virt.ibmpowervm.hmc.k2.viosentry_wrapper \
    import VIOSEntryWrapper
from powervc_nova.conductor import api as powervc_conductor

from oslo_config import cfg

LOG = logging.getLogger(__name__)

ibmpowervm_opts = [
    cfg.IntOpt('ibmpowervm_live_remote_restart_timeout',
               default=3600,
               help='Timeout in seconds for remote restart task.')
]

CONF = cfg.CONF
CONF.register_opts(ibmpowervm_opts)


class RemoteRestart(object):

    def __init__(self, instance, host, entry=None, lpar_uuid=None):
        """
        :param instance: lpar instance to be remote restarted
        :param host: The host object
        :param lpar_entry: The lpar entry from the K2 feed
        :param lpar_uuid: The K2 uuid of the lpar
        """
        # Get the host data
        # Refresh the stats to get live data about number of remote restart
        # operations running
        self.host = host
        # use K2 level caching
        with K2OperGuard(host=self.host) as k2oper:
            mgt_sys_resp = k2oper.read(k2const.MGT_SYS, rootId=self.host.uuid)
        self.host_entry = mgt_sys_resp.entry
        self.host_wrapper = ManagedSystemEntryWrapper(self.host_entry)
        self.host_stats = host.get_host_stats_internal(refresh=True)
        self.host_uuid = host.uuid
        self.host_name = self.host_wrapper.get_system_name()

        # Get the lpar data
        self.lpar = instance
        self.lpar_entry = None
        self.lpar_uuid = lpar_uuid
        self.lpar_wrapper = None

        # Get the hmc
        self.hmc = host.select_active_hmc()

        # npiv_mappings
        self.npiv_map_str = None

        # Get volume api
        self.volume_api = volume.PowerVCVolumeAPI()
        self.conductor_api = powervc_conductor.PowerVCConductorAPI()

    # Do not use logcall because passwords will be printed to the logs
    def check_can_rebuild_on_destination(self, ctxt, instance_ref):
        """
        1) Validate the destination host is capable of remote restart
            operation.
        2) Collects data about the destination host for partition
           remote restart operation to be used in future method calls

        :returns dest_check_data: dictionary of target host data
        """
        tgt_sys_name = CONF.host_display_name

        # Do not use logcall because the enter/exit parms contain a password
        LOG.debug("Entering"
                  "check_can_rebuild_on_destination/remote_restart.py"
                  "ctxt=%s, instance_ref=%s, "
                  "dst_compute_info=%s" % (ctxt, instance_ref,
                                           CONF.host))

        msg = (_('Validating destination host %(host)s for remote restart')
               % {'host': tgt_sys_name})
        LOG.info(msg)

        lpar_name = self.lpar['display_name']

        host_cap = self.host_stats.get('stats').\
            get(PVM_SIMP_REMOTE_RESTART_CAPABLE)
        if host_cap is not True:
            msg = (_("Destination host %(host)s is not capable of remote "
                     "restart") % {'host': self.host_name})
            raise exception.IBMPowerVMRemoteRestartFailed(error=msg)

        # check we can find the lpar via uuid
        # This is how we validate the source and destination hosts
        # are managed by the same HMC
        source_host = instance_ref['host']
        no_connection = self._check_no_connection(ctxt, source_host)
        if not no_connection:
            lpar_entry = None
            try:
                with K2OperGuard(host=self.host) as k2oper:
                    k2resp = k2oper.read(k2const.LPAR, self.lpar_uuid)
                    lpar_entry = k2resp.entry
            except Exception:
                # problem extracting the LPAR
                # ignore the error here since we will raise a different
                # exception
                pass
            if lpar_entry is None:
                msg = (_("Both the source and destination hosts must be "
                         "managed by the same HMC."))
                raise exception.IBMPowerVMRemoteRestartFailed(error=msg)

        # Verify the target host can support another remote restart operation
        # CW: Remove the check for concurrent remote restart in progress.
        # the stats information could be out of sync. Also, it's best to
        # rely on the HMC error coming back from the remote restart
        # job

        # Create dictionary of target host data.  The nova manager code
        # looks for the information under the 'rebuild_data' key when
        # it consolidates target and source dictionaries.
        rebuild_data = {'dest_sys_name': self.host_name,
                        'dest_mtms_name': self.host.host_id,
                        'dest_lmb': self.host_stats.get('lmb_size')}

        return rebuild_data

    def _check_no_connection(self, context, host):
        """
        Return true is the state of the host is "no connection"

        :param context: security context
        :param host: host name
        :returns no_connection: True if state of host is "no connection"
        """
        no_connection = False
        compute_nodes = (compute_node.ComputeNodeList.get_all_by_host(
            context, host) or [])
        if compute_nodes:
            node = compute_nodes[0]
            if node.stats.get(HYPERVISOR_STATE) == 'no connection':
                no_connection = True
        return no_connection

    def check_can_rebuild_instance(self, ctxt, instance_ref):
        """
        Validate the instance is capable of remote restart operation.
        Validate the instance is in Remote Restartable state.

        :param ctxt: security context
        :param instance_ref: instance to be migrated
        """

        inst_name = instance_ref['display_name']

        msg = (_('Validating instance %(instance)s for remote restart')
               % {'instance': inst_name})
        LOG.info(msg)

        if instance_ref.get('power_specs').get('srr_capability') is not True:
            msg = (_("Instance %(lpar)s simplified remote restart capability"
                     " is not set to True.") % {'lpar': inst_name})
            raise exception.IBMPowerVMRemoteRestartFailed(error=msg)

        rr_state = 'Remote_Restartable'
        inst_rr_state = instance_ref.get('power_specs').get('srr_state')
        if inst_rr_state is not None:
            if inst_rr_state.lower() != rr_state.lower():
                msg = (_("Instance %(lpar)s with remote restart state "
                         "'%(state)s' is not Remote Restartable.")
                       % {'lpar': inst_name, 'state': inst_rr_state})
                raise exception.IBMPowerVMRemoteRestartFailed(error=msg)

        return

    @logcall
    def verify_host_capacity_for_rebuild(self, rebuild_allowed,
                                         rebuild_running,
                                         instance_name, host_name):
        """
        The host can only support a limited number of concurrent remote
        restart operations. Verify there is room for another.

        :param rebuild_allowed: The number of remote restarts allowed
        :param rebuild_running: The number of remote restart in progress
        :param instance_name: The name of the instance (for logging)
        :param host_name: The name of the host (for logging)
        """
        # if rebuild_running >= cfg.CONF.prs_max_parallel_dest_rebuilds_hmc:
        if rebuild_running >= rebuild_allowed:
            error = (_("Cannot remote restart %(inst)s because host %(host)s "
                       "only allows %(allowed)s concurrent remote restart"
                       " operations and %(running)s remote restart operations"
                       " are currently running.") %
                     {'inst': instance_name,
                      'host': host_name,
                      'allowed': rebuild_allowed,
                      'running': rebuild_running})
            LOG.exception(error)
            raise exception.IBMPowerVMRemoteRestartFailed(error)

    @logcall
    def pre_remote_restart(self, context, instance, bdms, network_info):
        """
        Prepare the destination vios for remote restart

        Call plug_vifs() on target host to enable VLAN bridging for VM.
        Also establish mappings between vscsi volumes of an instance to
        dest host.

        NOTES:
        If there are any VLANs or disks attached to VM that are not
        managed by PowerVC, user needs to manually establish mapping on
        the target host before remote restart started.

        :param instance: lpar instance to be remote restarted
        :param network_info: instance network information
        """

        self._check_npiv_port_logged_in(context, instance, bdms)

        try:
            # Establish hostmapping of ephemeral and cinder volumes on the
            # target host.
            self._map_vscsi_volumes_to_target_host(context, instance, bdms)
            LOG.info(_("Completed configuration of required "
                       "external storage."))
        except Exception as e:
            error = six.text_type(e)
            inst_name = instance['display_name']
            msg = (_("Failed during the configuration of required external "
                     "storage on destination host for virtual machine "
                     "%(inst_name)s. Error: %(error)s") % locals())
            LOG.exception(msg, instance=self.lpar)
            raise exception.IBMPowerVMRemoteRestartFailed(error=msg)

        try:
            self._generate_npiv_mappings(context, instance)
        except Exception as e:
            error = six.text_type(e)
            inst_name = instance['display_name']
            msg = (_("Failed to generate required mapping for npiv "
                     "storage on destination host for virtual machine "
                     "%(inst_name)s. Error: %(error)s") % locals())
            LOG.exception(msg, instance=self.lpar)
            raise exception.IBMPowerVMRemoteRestartFailed(error=msg)

        try:
            # Plug VLAN on the target VIOSes
            self.host.plug_vifs(instance, network_info)
        except Exception as e:
            error = six.text_type(e)
            inst_name = instance['display_name']
            msg = (_("Failed to enable VLAN bridging "
                     "for virtual machine %(inst_name)s."
                     " %(error)s") % locals())
            LOG.exception(msg)
            raise exception.IBMPowerVMRemoteRestartFailed(error=msg)

    @logcall
    def _check_npiv_port_logged_in(self, context, instance, bdms):
        """
        Checks to see if any of the npiv port of the instance is logged in to
        the fabric. If it is, an exception is thrown and remote restart does
        not proceed

        :param context: service context
        :param instance: The instance being remote restarted
        :param bdms: All the BDMs associated with the instance
        """
        if not bdms:
            # no attached cinder volumes.
            LOG.info(_("No attached cinder volumes for processing."),
                     instance=self.lpar)
            return

        for bdm in bdms:
            connection_info = self._process_bdm(bdm)
            volume_id = connection_info.get('data').get('volume_id') \
                if connection_info and 'data' in connection_info else None
            connection_type = lpar._get_volume_connection_type(connection_info)
            if connection_type is None:
                LOG.error(_("Could not identify the connection type of a "
                            "volume. volume: %(volume_id)s. "
                            "connection_info: %(connection_info)s.")
                          % {'volume_id': volume_id,
                             'connection_info': connection_info},
                          instance=self.lpar)
                continue
            if connection_type == CONNECTION_TYPE_NPIV:
                vol_conn = connection_info.get("volume_connector")
                if not vol_conn:
                    LOG.info(_("Volume_connector is missing from bdm "
                               "mapping for instance %(vm_id)s and volume "
                               "%(vol_id)s." % dict(vm_id=instance['id'],
                                                    vol_id=volume_id)))
                    continue
                conn = vol_conn.get('connector')
                wwpns = conn.get('wwpns') if conn else None
                if wwpns:
                    fabAPI = volume.PowerVCFabricAPI()
                    active_map = fabAPI.map_ports_to_fabrics(
                        context, wwpns)
                    # p = port wwpn and f = fabric identifier in
                    # composition.
                    active_wwpns = ([p for p, f in active_map.items()
                                    if f is not None])
                    if active_wwpns:
                        msg = (_("One or more WWPNs %(wwpns)s of NPIV volume "
                                 "%(vol_id)s attached to Virtual machine are "
                                 "found active on the managed fabric. A "
                                 "running Virtual Machine can not be "
                                 "restarted.") %
                               {'wwpns': active_wwpns,
                                'vol_id': volume_id})
                        LOG.error(msg, instance=self.lpar)
                        raise exception.IBMPowerVMRemoteRestartFailed(
                            error=msg)
                    else:
                        LOG.debug("No active wwpns for instance %s" %
                                  instance['id'])

    @logcall
    def _generate_npiv_mappings(self, context, instance):
        fc_data = migrate_utils.get_lpar_slot_info(instance)
        stg_map = StorageMapping()
        dest_host_name = self.host.host_id

        try:
            mapping_dict = stg_map.get_hmc_migrate_map_strings(
                context, fc_data, {}, self.lpar, dest_host_name)
            self.npiv_map_str = mapping_dict.get('npiv')
        except stgex.IBMPowerVCStorageError:
            # IBMPowerVCStorageError details has been written
            # in the compute log by SCG logging. translate it to the compute
            # migrate exception. Otherwise, it will be truncated
            # by instance fault, which is only 256 characters long.
            error = (_("Failed to generate Virtual Fibre Channel mapping on "
                       "the destination host %(dest_host)s based on storage "
                       "connectivity group configuration. For detailed "
                       "information about the migration failure, see "
                       "the nova compute log file "
                       "(/var/log/nova/nova-compute-%(dest_host)s.log).")
                     % {'dest_host': dest_host_name})
            LOG.exception(error)
            raise exception.IBMPowerVMMigrationFailed(error)

    @logcall
    def _map_vscsi_volumes_to_target_host(self, context, instance, bdms):
        """
        Establish hostmaps for cinder volumes on the target host
        before remote restart operation.
        This function is running on target host.

        :param instance: VM instance
        :param bdms: bdms for the cinder volumes on source host.
        """

        LOG.debug("instance_uuid=%s, number of BDMs=%d." %
                  (instance['uuid'], 0 if not bdms else len(bdms)))
        scg_map = StorageMapping()
        # Determine what connectivity type check is needed so that
        # the connectivity call does not raise an unnecessary exception.
        if volume_utils.PowerVCNovaVirtDriverVolumeAPI().\
                get_instance_volumes(context, instance, 'data'):
            chk_conn_type = CHK_ALL
        else:
            chk_conn_type = CHK_BOOT
        scg_info = scg_map.get_connectivity_info(context, instance['uuid'],
                                                 self.host.host_id,
                                                 instance, live_data=False,
                                                 chk_conn_type=chk_conn_type)

        # extract scg class to six.text_type to avoid UnicodeEncodeError
        scg_class = scg_info.pop('scg')
        LOG.debug('scg_info: %s, scg_class: %s' %
                  (scg_info, six.text_type(scg_class)), instance=self.lpar)
        scg_info['scg'] = scg_class

        scg_backend_host_group = scg_info.get('backend_host_group')
        if scg_backend_host_group is not None:
            LOG.info(_("This LPAR belongs to an SCG which is using "
                       "Global-Host-Group '%s'.") % scg_backend_host_group,
                     instance=self.lpar)
        else:
            LOG.info(_("This LPAR belongs to an SCG which does not use any "
                       "Global-Host-Group."), instance=self.lpar)

        if not bdms:
            # no attached cinder volumes.
            LOG.info(_("No attached cinder volumes for processing."),
                     instance=self.lpar)
            return

        LOG.info(_("Processing volume mappings to target host..."),
                 instance=self.lpar)
        # Establish connection mappings for every VSCSI volume against
        # all target VIOS's and invoke lua_recovery.
        for bdm in bdms:
            connection_info = self._process_bdm(bdm)
            volume_id = connection_info.get('data').get('volume_id') \
                if connection_info and 'data' in connection_info else None
            connection_type = lpar._get_volume_connection_type(connection_info)
            if connection_type is None:
                msg = (_("Could not identify the connection type of a volume. "
                         "Volume: %s.") % volume_id)
                LOG.error(_("Could not identify the connection type of a "
                            "volume. volume: %(volume_id)s. "
                            "connection_info: %(connection_info)s.")
                          % {'volume_id': volume_id,
                             'connection_info': connection_info},
                          instance=self.lpar)

            if connection_type == CONNECTION_TYPE_VSCSI:
                LOG.info(_("Processing a vSCSI volume. volume=%s.") %
                         volume_id, instance=self.lpar)
                connector = lpar.\
                    get_volume_connector(self.host, instance,
                                         self.lpar['display_name'],
                                         self.lpar_uuid,
                                         self.lpar_entry)
                if connector is None:
                    msg = (_("Cannot get volume connector to "
                             "virtual machine: %s.") %
                           self.lpar['display_name'])
                    LOG.error(msg, instance=self.lpar)
                    raise Exception(msg)

                vios_ports_list = connector.pop('wwpn_vscsi')
                LOG.debug("List of VIOS'es and their WWPNS maps %s." %
                          vios_ports_list, instance=self.lpar)
                # Process the connection on all vios's
                self._process_vscsi_conn(context, vios_ports_list, connector,
                                         connection_info, bdm)

    @logcall
    def _process_bdm(self, bdm):
        """
        Returns the connection info from the bdm object.

        :param bdm: The BDM to fetch connection info from
        :returns: Connection_info in the BDM
        """
        LOG.debug("Processing block device mapping. bdm=%s." % bdm,
                  instance=self.lpar)
        connection_info = bdm['connection_info']
        if connection_info is None:
            LOG.error(_("BDM connection info is none. Skipping it."
                        "bdm=%s.") % bdm, instance=self.lpar)
            return None
        else:
            if connection_info and isinstance(connection_info, basestring):
                connection_info = jsonutils.loads(connection_info)

        volume_id = connection_info.get('data').get('volume_id')
        volume_type = connection_info.get('driver_volume_type')
        if volume_type != 'fibre_channel':
            LOG.debug("Volume type is not fibre_channel. So, skipping "
                      "volume %s." % volume_id, instance=self.lpar)
            return None

        return connection_info

    @logcall
    def _process_vscsi_conn(self, context, vios_ports_list, vol_connector,
                            connection_info, bdm):
        """
        Establishes hostmap for cinder volume and the target VIOS's and invokes
        lua_recovery on the target VIOS's for hdisks to reflect on VIOS's.

        This function is running on target host.

        :param vios_ports_list: a list of maps where each map contains the
                                vios_name, vios_uuid and wwpns for a VIOS.
        :param vol_connector: volume connector for the VM.
        :param connection_info: connection info for the volume.
        :param bdm: block device mapping
        """
        backend_host_group = vol_connector.get('backend_host_group')
        volume_id = connection_info.get('data').get('volume_id')

        connector = copy.deepcopy(vol_connector)
        if 'connection_type' not in connector:
            connector.update({'connection_type': [CONNECTION_TYPE_VSCSI]})

        # process the connection on all vios's of the Host
        vscsi_lun_list = []
        for vios in vios_ports_list:
            connector['wwpns'] = vios['wwpns']
            connector['host'] = vios['vios_host']
            c_info = self._process_vscsi_conn_on_vios(context, vios, connector,
                                                      connection_info,
                                                      backend_host_group)
            for vios_target_lun in c_info['vscsi_target_lun']:
                vscsi_lun_list.append(vios_target_lun)

        # Update BDM entry in DB with new Connection_info.
        #
        # Modify connection info with new 'vscsi_target_lun', and volume
        # connector. Volume connector would have 'wwpns' of only one VIOS
        # of the host. The callers of detach, terminate_connection etc...
        # would loop through the wwpns of all the vios'es available
        # in 'vscsi_target_lun'.
        c_info['vscsi_target_lun'] = vscsi_lun_list
        if 'serial' not in c_info:
            c_info['serial'] = volume_id
        volume_connector = {'connection-type': CONNECTION_TYPE_VSCSI,
                            'connector': connector}
        c_info.update({'volume_connector': volume_connector})
        bdm.connection_info = jsonutils.dumps(c_info)
        bdm.save()
        LOG.debug("Updated BDM db entry for volume %(volume_id)s. "
                  "Updated BDM: %(bdm)s." % {'volume_id': volume_id,
                                             'bdm': bdm})

    @logcall
    def _process_vscsi_conn_on_vios(self, context, vios_wwpns, connector,
                                    connection_info, backend_host_group):
        """
        Establishes hostmap for cinder volume and the target VIOS and invokes
        lua_recovery on the target VIOS for hdisks to reflect on VIOS.

        This function is running on target host.

        :param vios_wwpns: a map that contains the vios_name, vios_uuid and
                           wwpns for that VIOS.
        :param connector: volume connector for the VM.
        :param connection_info: connection info for the volume.
        :param backend_host_group: Name of the global host group.
        :returns conn_info: returns conn_info for the volume for that VIOS.
        """
        volume_id = connection_info.get('data').get('volume_id')

        # As we are doing exclusively for vscsi
        # if any 'phy_to_virt_initiators' are there
        # in connector just pop them
        if 'phy_to_virt_initiators' in connector:
            connector.pop('phy_to_virt_initiators')

        if backend_host_group is None:
            LOG.info(_("Establishing mapping between volume %(volume)s "
                       "and VIOS %(vios)s.") %
                     {'volume': volume_id, 'vios': vios_wwpns['vios_name']},
                     instance=self.lpar)
            c_info = self.volume_api.initialize_connection(context, volume_id,
                                                           connector)
            if c_info is None:
                msg = (_("Connection information is not returned for volume "
                         "%(volume_id)s. connector: %(connector)s.") %
                       {'volume_id': volume_id, 'connector': connector})
                LOG.error(msg, instance=self.lpar)
                raise Exception(msg)
        else:
            # If connection_info is None, we would not reach here.
            c_info = copy.deepcopy(connection_info)

        vios_target_lun = {'vios_name': vios_wwpns['vios_name'],
                           'target_lun': c_info['data']['target_lun'],
                           'vios_wwpn': vios_wwpns['wwpns'],
                           'target_wwpn': c_info['data']['target_wwn'],
                           'vios_host': connector['host']
                           }
        if 'ip' in connector:
            vios_target_lun['vios_ip'] = connector['ip']
        c_info['vscsi_target_lun'] = [vios_target_lun]

        status, dev_name, udid = lpar.\
            _invoke_lua_recovery(self.host, c_info,
                                 vios_wwpns['vios_uuid'],
                                 self.lpar,
                                 vios_wwpns['vios_name'])
        if dev_name is not None:
            if status in [k2const.LUA_STATUS_DEVICE_AVIALABLE,
                          k2const.LUA_STATUS_DEVICE_IN_USE,
                          k2const.LUA_STATUS_FOUND_ITL_ERR]:
                LOG.info(_("Discovered hdisk %(hdisk)s on vios %(vios)s for "
                           "volume %(volume_id)s. Status code: %(status)s.") %
                         {'hdisk': dev_name, 'vios': vios_wwpns['vios_name'],
                          'volume_id': volume_id, 'status': status},
                         instance=self.lpar)
            else:
                msg = (_("Cannot proceed with the hdisk %(hdisk)s discovered "
                         "on vios %(vios)s for volume %(volume_id)s. "
                         "Status: %(status)s.") %
                       {'hdisk': dev_name, 'vios': vios_wwpns['vios_name'],
                        'volume_id': volume_id, 'status': status})
                LOG.error(msg, instance=self.lpar)
                raise Exception(msg)
        else:
            msg = (_("Failed to discover hdisk on %(vios)s for volume "
                     "%(volume_id)s. Status: %(status)s.") %
                   {'vios': vios_wwpns['vios_name'], 'volume_id': volume_id,
                    'status': status})
            LOG.error(msg, instance=self.lpar)
            raise Exception(msg)

        c_info['data']['target_UDID'] = udid

        # Check to see if we discovered the correct volume on the VIOS.
        expected_udid = connection_info['data']['target_UDID']
        if udid != expected_udid:
            msg = (_("UDID of the discovered volume on vios %(vios)s did not "
                     "match with the expected UDID for volume %(volume_id)s. "
                     "UDID=%(udid)s. Expected UDID=%(expected_udid)s.") %
                   {'vios': vios_wwpns['vios_name'], 'volume_id': volume_id,
                    'udid': udid, 'expected_udid': expected_udid})
            LOG.error(msg, instance=self.lpar)
            raise Exception(msg)

        return c_info

    @logcall
    def _remove_volume_device_from_vios(self, vios_name, vios_uuid,
                                        volume_id, volume_uid):
        """
        Removes volume device from VIOS.

        :param vios_name: name of the VIOS.
        :param vios_uuid: UUID of the VIOS.
        :param volume_id: ID of the volume.
        :param volume_uid: UID of the volume (part of the UUID of the
                           volume on storage).
        """
        LOG.info(_("Removing volume device from vios %(vios)s for volume "
                   "%(volume_id)s. volume_uid: %(volume_uid)s.") %
                 {'vios': vios_name, 'volume_id': volume_id,
                  'volume_uid': volume_uid}, instance=self.lpar)

        if vios_name is None or vios_uuid is None or volume_id is None or \
                volume_uid is None:
            LOG.error(_("Cannot detach and remove volume from vios. "
                        "At least one of the required properties is missing."),
                      instance=self.lpar)
            return

        with K2OperGuard(host=self.host) as oper:
            try:
                # extract vios info
                # Get vios data - need to populate the vios cache first
                # by refreshing
                vioses_resp = oper.read(k2const.MGT_SYS, rootId=self.host.uuid,
                                        childType=k2const.VIOS, age=0,
                                        xag=k2const.ALL_VIOS_EXT_PROPS)

                vios_entries = vioses_resp.feed.entries

                # Find the vios entry for our VIOS.
                vios_entry = None
                for entry in vios_entries:
                    vios_id = entry.properties['id']
                    if vios_id == vios_uuid:
                        vios_entry = entry
                        break

                if vios_entry is None:
                    LOG.error(_("No matching entry found for vios "
                                "%(vios_name)s. vios_uuid: %(vios_uuid)s.") %
                              {'vios_name': vios_name, 'vios_uuid': vios_uuid},
                              instance=self.lpar)
                    return

                vios_wrapper = VIOSEntryWrapper(vios_entry)
                vios_name = vios_wrapper.get_vios_name()
                backing_dev = None
                LOG.debug("vios uuid = %s" % vios_uuid, instance=self.lpar)

                try:
                    LOG.debug("Getting hdisk name from vios %(vios_name)s "
                              "for volume %(volume_id)s. "
                              "volume_uid: %(volume_uid)s."
                              % {'volume_id': volume_id,
                                 'volume_uid': volume_uid,
                                 'vios_name': vios_name}, instance=self.lpar)

                    backing_dev = vios_wrapper.get_hdisk_from_uid(volume_uid)

                    LOG.info(_("%(hdisk)s is the vscsi backing device on vios "
                               "%(vios_name)s for volume %(volume_id)s. "
                               "volume_uid: %(volume_uid)s.") %
                             {'hdisk': backing_dev,
                              'volume_id': volume_id,
                              'volume_uid': volume_uid,
                              'vios_name': vios_name}, instance=self.lpar)

                    if not backing_dev:
                        LOG.info(_("No mapped device found on vios "
                                   "%(vios_name)s for volume %(volume_id)s. "
                                   "volume_uid: %(volume_uid)s ")
                                 % {'volume_id': volume_id,
                                    'volume_uid': volume_uid,
                                    'vios_name': vios_name},
                                 instance=self.lpar)
                except Exception as e:
                    LOG.error(_("Failed to get associated disk on vios "
                                "%(vios_name)s for volume %(volume_id)s. "
                                "volume_uid: %(volume_uid)s. Error=%(error)s")
                              % {'volume_id': volume_id,
                                 'volume_uid': volume_uid,
                                 'vios_name': vios_name,
                                 'error': six.text_type(e)},
                              instance=self.lpar)

                if backing_dev:
                    storage_map = scsi_map_processor.SCSIMapProcessor()
                    LOG.info(_("Detaching and deleting adapter "
                               "on vios %(vios)s. Backing_device=%(dev)s.")
                             % {'vios': vios_name, 'dev': backing_dev},
                             instance=self.lpar)
                    storage_map.\
                        detach_and_delete_adapter(oper, self.host.uuid,
                                                  vios_uuid,
                                                  backing_dev=backing_dev)
                    if CONF.enable_rmdev:
                        LOG.info(_("Removing volume device %(dev)s on "
                                   "vios %(vios_name)s.") %
                                 {'dev': backing_dev,
                                  'vios_name': vios_name}, instance=self.lpar)
                        job_status = lpar.\
                            _invoke_rmdev(self.host,
                                          CONF.host_display_name,
                                          vios_name,
                                          backing_dev,
                                          vios_uuid)
                        LOG.info(_("Status=%(status)s "
                                   "Task='Remove device %(volume)s from "
                                   "vios %(vios)s'.") %
                                 {'volume': backing_dev,
                                  'vios': vios_name,
                                  'status': job_status}, instance=self.lpar)
            except Exception as e:
                # It is okay to move on even if it fails to clean-up.
                LOG.warn(six.text_type(e), instance=self.lpar)
                pass

    @logcall
    def _post_process(self, context, instance, bdmap):
        # unmap the vscsi volumes on cinder
        self._unmap_vscsi_volumes_from_host(context, instance, bdmap)
        migrate_utils.clear_lpar_slot_info(instance)
        # save the new slot number and ports
        # need to validate the impact on performance
        with K2OperGuard(host=self.host) as k2oper:
            vioses_resp = k2oper.read(k2const.MGT_SYS,
                                      rootId=self.host.uuid,
                                      childType=k2const.VIOS,
                                      xag=k2const.ALL_VIOS_EXT_PROPS)
            vios_entries = vioses_resp.feed.entries
            fc_map = fc_map_processor.FCMapProcessor()
            if vios_entries:
                for vios_entry in vios_entries:
                    vios_wrapper = viosentry_wrapper.VIOSEntryWrapper(
                        vios_entry)
                    vios_name = vios_wrapper.get_vios_name()
                    # Gather the VFC mapping data needed to construct the
                    # VirtualFCMappings parameter.
                    lpar_id = self.lpar_uuid
                    fc_client_map_data = fc_map.extract_client_adapter_data(
                        vios_entry, lpar_id)
                    if fc_client_map_data is not None:
                        LOG.debug("FC client mapping data for vios %(vios)s is"
                                  " %(data)s." % {'vios': vios_name,
                                                  'data': fc_client_map_data})
                        migrate_utils.set_lpar_slot_info_from_fc_data(
                            instance, fc_client_map_data)

    @logcall
    def _unmap_vscsi_volumes_from_host(self, context, instance, bdmaps):
        """
        This function is running on source host/target host based
        on its caller.

        :param context: context
        :param instance: VM instance
        :param bdmaps: bdmaps for the cinder volumes on source host
        """

        if bdmaps is None:
            # no attached cinder volume.
            LOG.info(_("No attached cinder volumes for processing."),
                     instance=self.lpar)
            return

        # continue to process attached cinder volumes if any
        LOG.info(_("Processing volume mapping removals from source host..."),
                 instance=self.lpar)

        # Ignoring the backend_host_group check compared to migrate cleanup
        # as in remote restart case source host is not up and running.

        for bdm in bdmaps:
            connection_info = bdm['connection_info']
            if connection_info is None:
                LOG.error(_("Connection_info is missing in bdm. bdm=%s.")
                          % bdm, instance=self.lpar)
                continue
            else:
                if connection_info and isinstance(connection_info, basestring):
                    connection_info = jsonutils.loads(connection_info)

            volume_id = connection_info.get('data').get('volume_id')
            volume_type = connection_info.get('driver_volume_type')
            if volume_type != 'fibre_channel':
                LOG.debug("Volume type is not fibre_channel."
                          " So, skipping volume %s." % volume_id,
                          instance=self.lpar)
                continue

            connection_type = lpar._get_volume_connection_type(connection_info)
            if connection_type is None:
                LOG.error(_("Could not identify the connection type of a "
                            "volume. volume: %(volume_id)s. "
                            "connection_info: %(connection_info)s.")
                          % {'volume_id': volume_id,
                             'connection_info': connection_info},
                          instance=self.lpar
                          )
                continue
            if CONNECTION_TYPE_VSCSI in connection_type:
                vol_connector = connection_info.get("volume_connector")
                vscsi_target_wwpn = []
                if 'wwpn_vscsi' in vol_connector:
                    vscsi_target_wwpn = vol_connector.get('wwpn_vscsi')
                elif connection_info is not None:
                    vscsi_target_wwpn = connection_info.get('vscsi_target_lun')
                wwpn_list = []
                for idx in vscsi_target_wwpn:
                    if('vios_wwpn' in idx):
                        wwpn_list.append(idx.get('vios_wwpn'))
                    elif ('source_wwn' in idx):
                        wwpn_list.append(idx.get('source_wwn'))
                    else:
                        wwpn_list.append(idx.get('wwpns'))
                    LOG.info(_("Processing a vscsi volume. volume=%s.") %
                             volume_id, instance=self.lpar)
                connector = vol_connector.get('connector')

                for wwpn in wwpn_list:
                    connector['wwpns'] = wwpn
                    try:
                        self.volume_api.terminate_connection(context,
                                                             volume_id,
                                                             connector)
                    except Exception as e:
                        # It is okay to move on even if it fails to remove
                        # volume device or fails to remove volume connection.
                        LOG.exception(_("Error: %s.") % six.text_type(e),
                                      instance=self.lpar)
                        pass
        return True

    @logcall
    def _k2_rebuild_job(self, k2oper, source_host, target_host,
                        no_connection=False):
        """
        Creates remote restart k2 request job

        :param k2oper: K2 operator
        :param source_host: Source host
        :param target_host: Destination host
        :param no_connection: if true, source host has no connection to FSP
        """

        # Get the K2 response for the remote restart command
        # SW390979: remote restart job with ManagedSystem UUID in the URI
        # is not supported
        k2resp = k2oper.read(k2const.LPAR, self.lpar_uuid,
                             suffixType=k2const.SUFFIX_TYPE_DO,
                             suffixParm=k2const.SUFFIX_PARM_REMOTE_RESTART)

        # Create the job parameters
        job_wrapper = jobentry_wrapper.JobEntryWrapper(k2resp.entry)
        tar_UUID = 'targetManagedSystemUUID'
        job_parms = [job_wrapper.create_job_parameter(tar_UUID,
                                                      self.host_uuid)]
        # job_parms.append(job_wrapper.create_job_parameter(tar_name,
        #                                                 target_host))
        job_parms.append(job_wrapper.create_job_parameter('managedSystem',
                                                          source_host))
        job_parms.append(job_wrapper.create_job_parameter('operation',
                                                          'restart'))
        if no_connection:
            job_parms.append(job_wrapper.create_job_parameter('noconnection',
                                                              'true'))

        if self.npiv_map_str:
            job_parms.append(job_wrapper.create_job_parameter(
                'virtualFCMappings', self.npiv_map_str))
        # Submit the job
        try:
            job_wrapper.run_job(k2oper, self.lpar_uuid,
                                job_parms=job_parms)
        except exception.IBMPowerVMJobRequestFailed as e:
            error_msg = six.text_type(e)
            if 'HSCLB907' in error_msg:
                k2resp = k2oper.read(
                    k2const.LPAR, self.lpar_uuid,
                    suffixType=k2const.SUFFIX_TYPE_DO,
                    suffixParm=k2const.SUFFIX_PARM_REMOTE_RESTART)

                # Create the job parameters
                job_wrapper = jobentry_wrapper.JobEntryWrapper(k2resp.entry)
                tar_UUID = 'targetManagedSystemUUID'
                job_parms = [job_wrapper.create_job_parameter(tar_UUID,
                                                              self.host_uuid)]
                job_parms.append(job_wrapper.
                                 create_job_parameter('managedSystem',
                                                      source_host))
                job_parms.append(job_wrapper.create_job_parameter('operation',
                                                                  'restart'))
                if no_connection:
                    job_parms.append(job_wrapper.
                                     create_job_parameter('noconnection',
                                                          'true'))
                # sleep and retry
                greenthread.sleep(120)
                job_wrapper.run_job(k2oper, self.lpar_uuid,
                                    job_parms=job_parms)
            else:
                raise e

    @logcall
    def _rebuild(self, ctxt, instance, source_host):
        """
        Performs the remote restart of the lpar
        :param ctxt: Context object
        :param instance: Instance details object
        :param source_host: Source host of the lpar
        """

        inst_name = instance['display_name']
        target_host = instance['host']
        hostfact = compute_dom.ManagedHostFactory.get_factory()
        src_host = hostfact.find_host_by_name(ctxt, source_host)
        src_hst_uuid = src_host['host_uniqueid']

        # FSP-less Remote Restart
        no_connection = self._check_no_connection(ctxt, source_host)

        with K2OperGuard(host=self.host) as k2oper:
            try:
                self._k2_rebuild_job(k2oper, source_host, target_host,
                                     no_connection)
            except exception.IBMPowerVMJobRequestTimedOut as exc:
                LOG.exception(exc)
                try:
                    LOG.info(_("Remote restart task for virtual machine "
                               "%(lpar)s timed out. "
                               "Aborting remote restart.") %
                             {'lpar': inst_name})
                    self._rebuild_abort(k2oper, source_host, src_hst_uuid)
                except Exception as e:
                    LOG.exception(_("Remote restart abort for virtual machine "
                                    "%(lpar)s failed with error '%(error)s'") %
                                  {'lpar': inst_name,
                                   'error': six.text_type(e)})
                self._rebuild_recover(k2oper, source_host, src_hst_uuid, False)
                error = self._intercept_hmc_exception(exc)
                raise exception.IBMPowerVMRemoteRestartFailed(error)
            except Exception as e:
                # LOG.exception(e)
                self._rebuild_recover(k2oper, source_host, src_hst_uuid, False)
                LOG.exception(_("Remote restart task for virtual machine "
                                "%(lpar)s failed with error '%(error)s'") %
                              {'lpar': inst_name,
                               'error': six.text_type(e)})
                error = self._intercept_hmc_exception(e)
                raise exception.IBMPowerVMRemoteRestartFailed(error)
        return True

    @logcall
    def _rebuild_recover(self, k2oper, source_host, src_hst_uuid, force=False):
        """
        Performs remote restrart operation recovery of the lpar

        :param k2oper: K2 operator
        :param force: force option
        """
        try:
            # Get the K2 response for the migrate command
            k2resp = k2oper.read(k2const.LPAR, self.lpar_uuid,
                                 suffixType=k2const.SUFFIX_TYPE_DO,
                                 suffixParm=k2const.SUFFIX_PARM_REMOTE_RESTART)

            # Create the job parameters
            job_wrapper = jobentry_wrapper.JobEntryWrapper(k2resp.entry)

            job_parms = [job_wrapper.create_job_parameter('managedSystem',
                                                          source_host)]
            tar_UUID = 'targetManagedSystemUUID'
            job_parms.append(job_wrapper.create_job_parameter(tar_UUID,
                                                              self.host_uuid))

            job_parms.append(job_wrapper.create_job_parameter('operation',
                                                              'recover'))
            if force:
                job_parms.append(job_wrapper.create_job_parameter('force',
                                                                  'true'))

            LOG.debug('Submitting K2 remote restart recover job '
                      'for VM: %s' % self.lpar_uuid)
            # Submit the job
            job_wrapper.run_job(k2oper, self.lpar_uuid, job_parms=job_parms,
                                host_uuid=src_hst_uuid)
            LOG.info(_('Remote restart recovery for virtual machine '
                       '%(lpar)s is completed successfully.') %
                     {'lpar': self.lpar_uuid})
        except exception.IBMPowerVMJobRequestFailed as error:
            LOG.exception(_('Error: %s') % error)
            error_msg = six.text_type(error)
            if 'HSCLB906' in error_msg:
                # no need to run recover job since the VM is already
                # in remote restartable state
                LOG.info(_('Recover of Remote restart for virtual machine '
                           '%(lpar)s is not needed.') %
                         {'lpar': self.lpar_uuid})
        except Exception as e:
            LOG.exception(_("Recover of Remote restart for virtual machine"
                            " %(lpar)s failed with error '%(error)s'") %
                          {'lpar': self.lpar_uuid,
                           'error': six.text_type(e)})
            try:
                if not force:
                    self._rebuild_recover(k2oper, source_host, force=True)
            except Exception as ex:
                LOG.exception(_("Recover of Remote restart for virtual machine"
                                " %(lpar)s failed with error '%(error)s'") %
                              {'lpar': self.lpar_uuid,
                               'error': six.text_type(ex)})

    @logcall
    def _rebuild_abort(self, k2oper, source_host, src_hst_uuid):
        """
        Performs abort of remote restart of the lpar

        :param k2oper: K2 operator
        """

        # Get the K2 response for the migrate abort command
        k2resp = k2oper.read(k2const.LPAR, self.lpar_uuid,
                             suffixType=k2const.SUFFIX_TYPE_DO,
                             suffixParm=k2const.SUFFIX_PARM_REMOTE_RESTART)

        # Create the job parameters
        job_wrapper = jobentry_wrapper.JobEntryWrapper(k2resp.entry)

        job_parms = [job_wrapper.create_job_parameter('managedSystem',
                                                      source_host)]
        tar_UUID = 'targetManagedSystemUUID'
        job_parms.append(job_wrapper.create_job_parameter(tar_UUID,
                                                          self.host_uuid))

        job_parms.append(job_wrapper.create_job_parameter('operation',
                                                          'cancel'))

        LOG.debug('Submitting K2 remote restart abort job '
                  'for VM: %s' % self.lpar_uuid)
        # Submit the job
        job_wrapper.run_job(k2oper, self.lpar_uuid, job_parms=job_parms,
                            host_uuid=src_hst_uuid)

    @logcall
    def _get_rebuild_stats(self):
        """
        Checks the lpar state and returns the host stats for number
        of remote restart operations supported and number in progress

        :returns rr_allowed: Number of remote restart operations supported by
                             the host
        :returns rr_running: Number of remote restart operations in progress
        """
        rr_allowed = self.host_stats.get('stats'). \
            get(PRS_MAX_PARA_DEST_REBUILDS)
        rr_running = self.host_stats.get('stats'). \
            get(NUM_RR_IN_PROGRESS)

        return rr_allowed, rr_running

    @logcall
    def _intercept_hmc_exception(self, exception):
        """
        HMC error messages are not user friendly.  We want to inspect
        any error messages that come from the HMC, and print a helpful
        message. By default we'll indicate the HMC details can be found
        in the log file.

        :param exception: The exception from the HMC
        :returns error: An error message for the GUI
        """
        see_log_file_msg = (_("For detailed information about the "
                              "failure, see the nova compute log file "
                              "(/var/log/nova/nova-compute-%(suffix)s.log).")
                            % {'suffix': self.host.host_id})

        # Get the message from the HMC and log it
        if hasattr(exception, 'message'):
            hmc_err_msg = six.text_type(exception.message)
        else:
            hmc_err_msg = six.text_type(exception)
        LOG.exception(exception)

        # Query the error messages
        if ((hmc_err_msg.find("HSCLA281") >= 0) or
                (hmc_err_msg.find("HSCLB07C") >= 0)):
            # No need to search for HSCLB07C error message because
            # what we provide by default is suffificent.
            extracted_error = self._extract_hmc_error_by_code(hmc_err_msg,
                                                              "HSCLA281")
            if extracted_error is None:
                error = (_("The operation to query virtual I/O server "
                           "partition for completion has failed. %s" %
                           see_log_file_msg))
            else:
                error = (_("%(extracted_error)s %(see_log_file_msg)s") %
                         {'extracted_error': extracted_error,
                          'see_log_file_msg': see_log_file_msg})

        else:
            error = (_("The remote restart operation issued to the HMC failed."
                       "Ensure that source and target hosts supports "
                       "simplified remote restart capability. %s"
                       % see_log_file_msg))

        return error

    def _extract_hmc_error_by_code(self, exception_msg, error_code):
        """
        Extracts the HMC error (by HMC error code) from the exception message.

        :param exception_msg: exception message thrown by HMC.
        :param error_code: HMC error code we are interested in.
        """
        pattern = "(.*)" + error_code + "(.*)\n(.*)"
        try:
            match = re.search(pattern, exception_msg)
            if match is not None:
                return error_code + match.group(2)
        except Exception as e:
            LOG.debug("Failed to retrieve hmc error by error code. Error: %s" %
                      six.text_type(e))

        return None

    def _exchange_ssh_keys(self, k2oper, ip, user, password, test=False):
        """
        Exchange ssh keys with the target hmc

        :param k2oper: K2 operator
        :param ip: ip address of target hmc
        :param user: hmc user id
        :param password: hmc user password
        :param test: if true, add test option
        """
        if test:
            ssh_cmd = 'mkauthkeys --ip ' + ip + ' -u ' + user + ' --test'
        else:
            ssh_cmd = ('mkauthkeys --ip ' + ip + ' --passwd ' + password +
                       ' -u ' + user)
        pwd_removed = ssh_cmd.replace(password, '********')
        LOG.debug(("Exchange ssh keys command: %s") % pwd_removed)

        # Get the K2 response for the CLIRunner command
        k2resp = k2oper.read(k2const.MGT_CONSOLE, None,
                             suffixType=k2const.SUFFIX_TYPE_DO,
                             suffixParm=k2const.SUFFIX_PARM_CLI_RUNNER)

        # Create the job parameters
        job_wrapper = jobentry_wrapper.JobEntryWrapper(k2resp.entry)
        ack_parm = 'acknowledgeThisAPIMayGoAwayInTheFuture'
        job_parms = [job_wrapper.create_job_parameter('cmd', ssh_cmd),
                     job_wrapper.create_job_parameter(ack_parm, 'true')]
        job_wrapper.run_job(k2oper, None, job_parms=job_parms,
                            sensitive=True)
