# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
from oslo_log import log as logging

from powervc_nova.virt.ibmpowervm.hmc import exception as hmc_ex
from powervc_nova.virt.ibmpowervm.hmc.storage.virtual_io_adapter import \
    VirtualIOAdapter

LOG = logging.getLogger(__name__)


class VirtualStorageAdapter(VirtualIOAdapter):
    """
    Object represents virtual io adapter (storage type) information
    queried from HMC.
    The goal is only to store attributes needed by PowerVC.
    We should use the child class like VirtualSCSIClientAdapter.
    """
    def __init__(self, lpar_id, slot_num=None,
                 remote_lpar_id=None, remote_slot=None):
        """
        Power allows input of 0xFFFF for remote_lpar_id & remote_slot
        to indicate any client adapter can connect to the server. So it's
        signal for such of use if both remote_lpar_id & remote_slot are None.
        However, if remote_slot are None, it's used to signal
        HMC to create corresponding adapter on the particular remote_lpar_id.
        :param lpar_id: lpar_id owning the virtual adapter
        :param slot_num: virtual slot number used by the virtual adapter. If
                         it's None, HMC will pick the next available slot to
                         use.
        :param remote_lpar_id: lpar id for the corresponding server/client
                               adapter pair.
        :param slot_num: virtual slot number for the corresponding
                         server/client adapter pair.
        """
        # make sure if remote_lpar_id is None, remote_slot_num must be None
        # as well
        if remote_lpar_id is None and remote_slot is not None:
            raise hmc_ex.IBMPowerVMInvalidVIOConfig()
        super(VirtualStorageAdapter, self).__init__(lpar_id, slot_num)

        self.remote_lpar_id = remote_lpar_id
        self.remote_slot_num = remote_slot

    def __repr__(self):
        return ('VirtualStorageAdapter lpar_id: %s '
                'slot_num: %s remote_lpar_id: %s remote_slot: %s' %
                (self.lpar_id, self.slot_num, self.remote_lpar_id,
                 self.remote_slot_num))


class VirtualSCSIServerAdapter(VirtualStorageAdapter):
    """
    Virtual SCSI server adapter.
    """
    def __init__(self, lpar_id, slot_num,
                 remote_lpar_id=None, remote_slot=None):
        super(VirtualSCSIServerAdapter, self).__init__(
            lpar_id, slot_num, remote_lpar_id, remote_slot)


class VirtualSCSIClientAdapter(VirtualStorageAdapter):
    """
    Virtual SCSI client adapter
    """
    def __init__(self, lpar_id, slot_num,
                 remote_lpar_id=None, remote_slot=None):
        super(VirtualSCSIClientAdapter, self).__init__(
            lpar_id, slot_num, remote_lpar_id, remote_slot)


class VirtualFCServerAdapter(VirtualStorageAdapter):
    """
    Virtual Fibre Channel server adapter
    """
    def __init__(self, lpar_id, slot_num,
                 remote_lpar_id=None, remote_slot=None):
        super(VirtualFCServerAdapter, self).__init__(
            lpar_id, slot_num, remote_lpar_id, remote_slot)


class VirtualFCClientAdapter(VirtualStorageAdapter):
    """
    Virtual Fibre Channel client adapter
    """
    def __init__(self, lpar_id, slot_num,
                 remote_lpar_id=None, remote_slot=None, wwpns=None):
        super(VirtualFCClientAdapter, self).__init__(
            lpar_id, slot_num, remote_lpar_id, remote_slot)
        self.wwpns = wwpns
