# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import copy
import six
from powervc_nova import logcall
from powervc_nova.objects import block_device as pvc_blkdev
from powervc_nova.objects.compute import dom as compute_dom
from oslo_log import log as logging
from oslo_config import cfg
from powervc_nova.volume import cinder as volume_cinder
import powervc_nova.virt.ibmpowervm.hmc.lpar as lpar
from powervc_nova.virt.ibmpowervm.hmc.k2oprguard import K2OperGuard
import powervc_nova.virt.ibmpowervm.hmc.k2.constants as k2const
from powervc_nova.virt.ibmpowervm.hmc.k2.viosentry_wrapper \
    import VIOSEntryWrapper
from powervc_nova.virt.ibmpowervm.hmc.k2.scsi_map_processor \
    import SCSIMapProcessor
from powervc_nova import _
from powervc_nova.virt.ibmpowervm.hmc import exception
from powervc_nova.virt.ibmpowervm.common.constants \
    import CONNECTION_TYPE_VSCSI

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


def _get_connection_type(connector):
    if (connector is not None and 'connection_type' in connector
            and (connector['connection_type'] is not None)):
        return connector['connection_type'][0]


@logcall
def _update_connection_itls(connection, vol_itl_list, vios_name, vios_wwpns):
    """ For onboarded volumes we do not have the vios names
    in the connection info so connection info needs to be updated
    to have the same which can be used for lua discovery
    """
    # validate connection
    if (connection is None
            or 'vscsi_target_lun' not in connection):
        LOG.warn(_("connection info does not have the ITL List"))

    vscsi_itl_list = copy.deepcopy(vol_itl_list)
    connection['vscsi_target_lun'] = []
    for itl in vscsi_itl_list:
        match = list(set(vios_wwpns) & set(itl['vios_wwpn']))
        LOG.debug("Vios wwpn %(vios)s itl wwpn %(itl)s and "
                  "matched WWPN %(wwpn)s Updating vios_name %(name)s: " %
                  {'vios': vios_wwpns, 'itl': itl['vios_wwpn'],
                   'wwpn': match, 'name': vios_name})
        if match:
            itl['vios_name'] = vios_name
            itl['vios_wwpn'] = match
        else:
            itl['vios_name'] = None
        connection['vscsi_target_lun'].append(itl)

    return


def _get_vios_entries(host):
    # Get vios data for the terminate_connection methods.
    # Need to populate the vios cache first by refreshing.
    # use K2 level caching
    with K2OperGuard(host=host) as k2oper:
        try:
            vioses_resp = k2oper.vios_read(k2const.MGT_SYS, rootId=host.uuid,
                                           childType=k2const.VIOS, age=0,
                                           xag=k2const.ALL_VIOS_EXT_PROPS)
            vios_entries = vioses_resp.feed.entries
            if vios_entries is None:
                LOG.error(_('No VIOS Found'))
                raise exception.IBMPowerVMVIOSNotFound()

            return vios_entries
        except Exception as e:
            LOG.error(_('Failed to get the vios response: %s')
                      % six.text_type(e))


@logcall
def cleanup_disks(host, lpar_name, connection_info,
                  wwpn_list, vios_entries, dev=None):
    """ cleanup routine to remove the left hdisk devices on the
        VIOS. It performs the below steps
        1. Iterate through the VIOS
        2. Invoke LUA discovery using the ITLs from the connection info
        3. If LUA discovery returns with the status 6 that means
        hdisk is properly removed from the VIOS
        4. For any other status code it means device still exists
        and might be used by some vhost
        5. No exception is thrown in case cleanup fails.
    """
    LOG.debug('cleanup_disks on host %(host)s. '
              'connection_info = %(conn_info)s wwpn_list = %(wwpns)s' %
              {'host': host, 'conn_info': connection_info,
               'wwpns': wwpn_list})

    if 'vscsi_target_lun' not in connection_info:
        LOG.info(_("Missing vscsi_target_lun in connection_info. "
                   "Cleanup disk cannot be performed. Connection "
                   "information: %s") % connection_info)
        return

    vol_itl_list = connection_info['vscsi_target_lun'] if 'vscsi_target_lun'\
        in connection_info else None

    volume_id = connection_info.get('data').get('volume_id') if 'volume_id' in\
        connection_info['data'] else None

    for vios_entry in vios_entries:
        vios_wrapper = VIOSEntryWrapper(vios_entry)
        vios_wwpns = vios_wrapper.get_vios_wwpns()
        vios_name = vios_wrapper.get_vios_name()
        vios_uuid = vios_entry.properties['id']
        found = _is_wwpns_found(wwpn_list, vios_wwpns)
        if not found:
            LOG.info(_("Cleanup: No WWPNS matched on vios %(name)s. "
                       "vios_wwpn=%(vios)s,  wwpn_list=%(wwpn)s") %
                     {'name': six.text_type(vios_name),
                      'vios': vios_wwpns, 'wwpn': wwpn_list})
            continue
        # Check if the device name exist and if its not a powerpath device
        # then remove the disk directly no need to perform lua-discovery
        if dev and "hdiskpower" not in dev:
            lpar._invoke_rmdev(host, CONF.host_display_name,
                               vios_name, dev, vios_uuid)
            continue

        if ('vscsi_target_lun' not in connection_info or
            'vios_name' not in connection_info['vscsi_target_lun'] or
            [True for i in connection_info['vscsi_target_lun']
             if 'vios_name' not in i or i['vios_name'] != vios_name]):
                _update_connection_itls(connection_info, vol_itl_list,
                                        vios_name, vios_wwpns)
        try:
            status, dev_name = lpar._invoke_lua_recovery(
                host, connection_info, vios_uuid,
                lpar_name, vios_name)[:2]
            if status == k2const.LUA_STATUS_INCORRECT_ITL:
                msg = (_("Successfully cleaned volume %(vol)s "
                         "for %(lpar)s on VIOS %(vios)s") %
                       {'vol': volume_id, 'lpar': lpar_name,
                        'vios': vios_name})
                LOG.info(msg)
            else:
                LOG.info(_("Cannot cleanup since device %(dev)s for "
                           "volume %(volume)s on %(vios)s "
                           "is discovered with status: %(status)s."),
                         {'dev': dev_name, 'volume': volume_id,
                          'vios': vios_name, 'status': status})
        except Exception as e:
            msg = (_("Failed to cleanup volume %(vol)s for %(lpar)s "
                     "on %(vios)s with reason: %(msg)s") %
                   {'vol': volume_id, 'lpar': lpar_name,
                    'vios': vios_name, 'msg': six.text_type(e)})
            LOG.warn(msg)
    return


def _get_bdms_for_volume_mapped_host(context, volume_id, connector):
    # go get the bdms for the volume_id mapped to same host
    bdm_list = list()
    bdms = pvc_blkdev.PowerVCBlockDeviceMappingList.get_by_volume_id(
        context, volume_id)

    for bdm in bdms:
        instfact = compute_dom.ManagedInstanceFactory.get_factory()
        inst = instfact.find_instance_by_uuid(context,
                                              bdm['instance_uuid'])
        if connector['host'] == inst['host']:
            LOG.debug("Shared instance %(inst)s on %(host)s" %
                      {'inst': bdm['instance_uuid'],
                       'host': connector['host']})
            bdm_list.append(bdm)
    return bdm_list


@logcall
def call_terminate_connection(host, context, instance, lpar_name, volume_id,
                              connector, conninfo=None, clean=True):
    """Wrapper to the terminate_connector for vSCSI shared volumes
    Check1: Check if the volume has more than one mapped BDM entries
    for the same host. If multiple entries exist skip terminate
    connections
    Check 2: There might be some VMs which are not managed by PowerVC
    using the same volumes. So check if there are any VirtualSCSI mappings
    exist on the host for the given volume.
    If exist then skip terminate connections
    For any other case proceed with the terminate connection.
    if clean= True then perform the LUA discovery cleanup
    """
    volume_api = volume_cinder.PowerVCVolumeAPI()
    if ((_get_connection_type(connector) == CONNECTION_TYPE_VSCSI) or
            (connector is not None and 'wwpn_vscsi' in connector)):
        if 'phy_to_virt_initiators' in connector and\
                not connector['phy_to_virt_initiators']:
            connector.pop('phy_to_virt_initiators', None)

        bdm_list = _get_bdms_for_volume_mapped_host(context,
                                                    volume_id,
                                                    connector)
        if len(bdm_list) > 1:
            LOG.info(_("The volume with ID %(vol)s is mapped to more than one "
                       "virtual machine instance, so detach processing will "
                       "be skipped for this instance. The shared count is "
                       "%(count)d.") %
                     {'vol': volume_id, 'count': len(bdm_list)})
            return

        # Onboarded Volume check for mapped entries.
        # check if the volume is mapped to any lpars on the host
        if conninfo and 'target_UDID' in conninfo['data']:
            volume_uid = conninfo.get('data').get('target_UDID')
            vscsi_target_wwpn = None
            if 'wwpn_vscsi' in connector:
                vscsi_target_wwpn = connector['wwpn_vscsi']
            elif conninfo is not None:
                vscsi_target_wwpn = conninfo['vscsi_target_lun']
            wwpn_list = []
            wwpn_done_list = []
            for idx in vscsi_target_wwpn:
                if('vios_wwpn' in idx):
                    wwpn_list.append(idx['vios_wwpn'])
                else:
                    wwpn_list.append(idx['wwpns'])
            try:
                vios_entries = _get_vios_entries(host)
                mappings = get_volume_mapped_entries(host, volume_uid,
                                                     wwpn_list, vios_entries)
                LOG.info(_("mappings returned %s") % mappings)
                for entry in mappings:
                    if ('map_count' in entry):
                        if entry['map_count'] > 0:
                            LOG.info(_("Volume is shared with %(cnt)s mappings"
                                       " to non-managed LPARs on vios WWPNs "
                                       "%(vios)s. Skipping terminate "
                                       "connection.") %
                                     {'cnt': entry['map_count'],
                                      'vios': entry['vios_wwpns']})
                            continue
                        if entry['map_count'] < 0:
                            LOG.info(_("Volume %(vol)s is not mapped. The VIOS"
                                       "wwpns not matched are: %(vios)s. "
                                       "Skipping terminate connection.") %
                                     {'vol': volume_id,
                                      'vios': entry['vios_wwpns']})
                            continue
                        # TODO: Comment on why the logic here seems to indicate
                        # that terminate connection is called when there is
                        # 0 mappings or an exception occurred obtaining them.
                        # Is that because the current managed connection
                        # is somehow not included in the count?
                        connector['wwpns'] = entry['vios_wwpns']
                        connector['host'] = entry['vios_host']
                        connector['ip'] = entry['vios_ip']
                        LOG.info(_("Calling terminate connection for: %s") %
                                 connector['wwpns'])
                        volume_api.terminate_connection(context,
                                                        volume_id, connector)
                        wwpn_done_list.append(entry['vios_wwpns'])
                        if clean:
                            wwpn_clean = []
                            wwpn_clean.append(connector['wwpns'])
                            dev = None
                            if 'backing_dev' in entry:
                                dev = entry['backing_dev']
                            cleanup_disks(host, lpar_name, conninfo,
                                          wwpn_clean, vios_entries, dev)
            except Exception as e:
                LOG.info(_("Call Terminate connection Failed"))
                LOG.exception(e)
            if wwpn_list == wwpn_done_list:
                return
            still_need_terminate = []
            initiators_terminated = set([x for y in wwpn_done_list for x in y])
            # wwpn_list may contain partial initiators that have already been
            # unmapped in a previous call. Double check if the initiators
            # from wwpn_list have already been called.
            still_need_terminate = [wl for wl in wwpn_list
                                    if not (set(wl) & initiators_terminated)]
            try:
                for wwpns in still_need_terminate:
                    connector['wwpns'] = wwpns
                    if conninfo is not None and 'vscsi_target_lun' in conninfo:
                        try:
                            vscsi_target_lun = conninfo['vscsi_target_lun']
                            entry = [x for x in vscsi_target_lun if
                                     set(x['vios_wwpn']) == set(wwpns)]
                            if len(entry) > 0:
                                if 'vios_host' in entry[0]:
                                    connector['host'] = entry[0]['vios_host']
                                if 'vios_ip' in entry[0]:
                                    connector['ip'] = entry[0]['vios_ip']
                            else:
                                # Drivers normally use the wwpns to lookup
                                # the host mapping entry. But pop the IP
                                # Should a driver require IP in the future
                                # for detach then this will need to be
                                # revisited
                                if 'ip' in connector:
                                    connector.pop("ip", None)
                        except Exception as e:
                            # don't fail just because something failed
                            # we still want to call terminate connection
                            LOG.info(_("Failed to find host or ip for vios"
                                       " ingnoring the following execption"))
                            LOG.exception(e)
                    LOG.info(_("Calling terminate connection for VSCSI: %s") %
                             connector['wwpns'])
                    volume_api.terminate_connection(context,
                                                    volume_id, connector)
            except Exception as e:
                LOG.info(_("Call Terminate connection Failed for VSCSI"))
                LOG.exception(e)
                pass
    else:
        volume_api.terminate_connection(context, volume_id, connector)


@logcall
def _is_wwpns_found(wwpn_list, vios_wwpn):
    """
    Check if the wwpns set is found in the list
    if found returns true
    """
    for wwpns in wwpn_list:
        match = list(set(vios_wwpn) & set(wwpns))
        if match:
            LOG.info(_("wwpns found"))
            return True
    return False


@logcall
def get_volume_mapped_entries(host, volume_id, wwpn_list, vios_entries):
    """
    Returns the number of VSCSI mappings for the specified volume
    :param lpar_uuid uuid of the lpar
    :param volume_id K2udid of the volume
    :param vios_name
    """
    LOG.info(_('get_volume_mapping: %(vol)s on %(wwpn)s'), {'vol': volume_id,
                                                            'wwpn': wwpn_list})
    backing_dev = None
    map_list = []
    CONNECTOR_HOST_LEN = 55
    for vios_entry in vios_entries:
        vios_wrapper = VIOSEntryWrapper(vios_entry)
        vios_wwpns = vios_wrapper.get_vios_wwpns()
        vios_name = vios_wrapper.get_vios_name()
        vios_lpar_id = vios_wrapper.get_vios_id()
        vios_host = host.host_id + "_" + str(vios_lpar_id)
        vios_host = vios_host[:CONNECTOR_HOST_LEN]
        vios_ip = vios_wrapper.get_rmc_ipaddr()

        found = _is_wwpns_found(wwpn_list, vios_wwpns)
        if not found:
            LOG.info(_("No WWPNS matched for vios %(name)s. "
                       "vios_wwpn=%(vios)s  wwpn_list=%(wwpn)s") %
                     {'name': six.text_type(vios_name),
                      'vios': vios_wwpns, 'wwpn': wwpn_list})
            map_data = {'vios_wwpns': vios_wwpns, 'map_count': -1,
                        'vios_host': vios_host, 'vios_ip': vios_ip}
            map_list.append(map_data)
            continue
        num_map_found = 0
        try:
            backing_dev = vios_wrapper.get_hdisk_from_uid(volume_id)
            if backing_dev:
                LOG.info(_("Device %(backing_dev)s found device on vios "
                           "%(vios_name)s for volume_id %(volume_id)s. ") %
                         {'backing_dev': backing_dev,
                          'vios_name': vios_name,
                          'volume_id': volume_id})
                storage_map = SCSIMapProcessor()
                scsi_map_entries = storage_map.find_map_entries(
                    vios_wrapper.get_virtual_scsi_mappings(), None, None,
                    backing_dev)
                if scsi_map_entries:
                    LOG.debug("Number of mapped entries found: %s" %
                              len(scsi_map_entries))
                    num_map_found = len(scsi_map_entries)

                map_data = {'vios_wwpns': vios_wwpns,
                            'backing_dev': backing_dev,
                            'vios_uuid': vios_wrapper.get_vios_id(),
                            'vios_name': vios_name,
                            'map_count': num_map_found,
                            'vios_host': vios_host, 'vios_ip': vios_ip}
                map_list.append(map_data)

        except Exception as e:
            LOG.info(_("No associated disk device on vios "
                       "%(vios_name)s for volume %(volume_id)s. "
                       "Error: %(error)s")
                     % {'vios_name': vios_name,
                        'volume_id': volume_id,
                        'error': six.text_type(e)})
            map_list.append({'vios_wwpns': vios_wwpns,
                             'vios_uuid': vios_wrapper.get_vios_id(),
                             'vios_name': vios_name,
                             'vios_host': vios_host,
                             'vios_ip': vios_ip,
                             'map_count': 0})
    return map_list
