# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
from networking_powervm.plugins.ibm.agent.powervm import utils as pvm_utils
from nova import context
from nova import exception as nova_exception
from nova.compute import vm_states
from nova_powervm.virt.powervm import vm as base_vm
from nova_powervm.virt.powervm.vif import PvmSeaVifDriver as vifdriver
from nova_powervm.virt.powervm import vif as novapowervmvif
from oslo_config import cfg
from oslo_log import log as logging
from powervc_nova.network.ibmpowervm import adapter_mapping as mapping_task
from powervc_nova.objects.network import dom as network_dom
from pypowervm.tasks import cna as cna_task
from pypowervm.utils import lpar_builder as lpar_bldr

pvm_opts = [
    cfg.StrOpt('powervm_mgr_uuid',
               default='',
               help='The unique id of the host to manage'),
    cfg.IntOpt('prs_max_parallel_rebuilds_novalink',
               default=10,
               help='Integer value that indicates the max number of '
               'concurrent remote restarts that will occur for a given '
               'source host. Required stat that applies to PowerVM'),
    cfg.IntOpt('prs_max_parallel_dest_rebuilds_novalink',
               default=10,
               help='Integer value that indicates the max number of '
               'concurrent remote restarts that will occur for a given '
               'target host. Required stat that applies to PowerVM'),
    cfg.IntOpt('number_of_exceptions_needed_for_missing_inst_pvm',
               default=2,
               help='Number of instance not found exceptions must be received'
               'before reporting the instance as missing'),
]

CONF = cfg.CONF
# TODO: We need to register this under powervm.
#       However, this option is set during register of new host
#       for HMC and Novalink, and several changes need to be
#       made in the registration code.
#       Defer this change for the registration support.
CONF.register_opts(pvm_opts)
CONF.import_opt('host', 'nova.netconf')

LOG = logging.getLogger(__name__)


def monkey_patch_pvm_valid_key():
    base_vm.VMBuilder._is_powervc_valid_key = \
        base_vm.VMBuilder._is_pvm_valid_key

    # The community driver will not support AME (at the moment) so this line
    # below will make the community VMBuilder put the AME EF into the list
    # of valid attributes for the pypowervm LPARBuilder where AME is supported
    base_vm.VMBuilder._ATTRS_MAP['powervm:ame_expansion_factor'] = (
        lpar_bldr.AME_FACTOR)

    def _is_powervc_valid_key(self, key):
        """Will return if this is a valid PowerVM key.

        :param key: The powervm key.
        :return: True if valid key.  False if non-powervm key and should be
                 skipped.  Raises an InvalidAttribute exception if is an
                 unknown PowerVM key.
        """
        _POWERVC_ATTR_MAP = {'powervm:storage_connectivity_group': None,
                             'powervm:boot_volume_type': None}
        # If not a powervm key, then it is not 'pvm_valid'
        if not key.startswith('powervm:'):
            return False

        if key in self._ATTRS_MAP.keys():
            return True

        if key in _POWERVC_ATTR_MAP.keys():
            return False

        # If the key starts with 'powervm:image_volume_type_', let it pass
        # as this is the prefix for volume type keys to be used during
        # multi-volume image deploy.
        if key.startswith('powervm:image_volume_type_'):
            return False

        exc = nova_exception.InvalidAttribute(attr=key)
        LOG.exception(exc)
        raise exc

    base_vm.VMBuilder._is_pvm_valid_key = _is_powervc_valid_key

monkey_patch_pvm_valid_key()


PVC_VIF_MAPPING = {'pvc_sea': 'nova_powervm.virt.powervm.vif.PvmSeaVifDriver',
                   'ovs': 'nova_powervm.virt.powervm.vif.PvmOvsVifDriver'}
novapowervmvif.VIF_MAPPING = PVC_VIF_MAPPING


def monkey_patch_plug():
    """
    We need to monkey patch the nova_powervm.virt.powervm.vm.crt_vif
    method to provide support for non-default vswitch and set the
    proper vswitch on the CNAs created during VM deploys.
    """
    #base_vm.powervc_mod_crt_vif = \
    #    base_vm.crt_vif
    vifdriver.powervc_mod_plug = \
        vifdriver.plug

    def powervc_mod_plug(self, vif, rebuild_slot_num):
        """Will create a Client Network Adapter on the system.
        :param vif: The nova VIF that describes the ethernet interface.
        :param rebuild_slot_num: The slot number for a rebuild request.

        :return: The created network adapter.
        """
        if not vif:
            LOG.info(_('No vif passed in'))
            return

        lpar_uuid = base_vm.get_pvm_uuid(self.instance)
        mac = vif['address']

        if rebuild_slot_num is not None:
            vif['slot_id'] = rebuild_slot_num
        else:
            # If this is a pre-created port (as determined by presence of
            # slot_id in the vif), then manually find the next available slot
            if (vif.get('slot_id') is
                not None) or (not self.instance.vm_state ==
                              vm_states.BUILDING):
                # Calculate slot to use
                slot = 0
                for cna in pvm_utils.list_cnas(self.adapter,
                                               self.host_uuid, lpar_uuid):
                    slot = cna.slot if cna.slot > slot else slot
                if slot > 0:
                    vif['slot_id'] = slot + 1
            else:
                # Not pre-created, get it from the mac
                vif['slot_id'] = int(mac[-2:], 16)

        state = self.instance.vm_state
        if (not vif.get('slot_id')) and (state == 'stopped' or
                                         state == 'error' or
                                         state == 'active'):
            # VM rebuild scenario
            vif['slot_id'] = int(mac[-2:], 16)

        LOG.debug('mac = %s, slot_id = %s' % (mac, vif['slot_id']))

        # CNA's require a VLAN
        vlan = vif['meta']['vlan']

        # Find the vswitch name for the cna using network association
        netafact = network_dom.NetworkAssociationFactory.get_factory()
        ctx = context.get_admin_context()
        neta = netafact.find_network_assoc(
            ctx, CONF.host, vif['network']['id'])

        sea = None
        if neta is None:
            # Network association didn't exist, build the default
            hostfact = network_dom.ManagedHostFactory.get_factory()
            host_dom = hostfact.find_host_by_name(ctx, CONF.host)
            adapt_map_task = mapping_task.AdapterMappingTaskClass()
            sea = adapt_map_task.build_default_network_assn(
                ctx, host_dom, vlan, vif['network']['id'])
        else:
            # Network association was found, use it.
            sea = neta.sea

        if sea is None:
            raise nova_exception.InterfaceAttachFailed(
                instance_uuid=self.instance['uuid'])

        vswitch_name = sea.primary_vea.vswitch_name

        return cna_task.crt_cna(
            self.adapter, self.host_uuid, lpar_uuid,
            vlan, vswitch=vswitch_name,
            crt_vswitch=False, slot_num=vif['slot_id'], mac_addr=mac)

    #base_vm.crt_vif = \
    #    powervc_mod_crt_vif
    vifdriver.plug = \
        powervc_mod_plug

monkey_patch_plug()


def convert_pypowervm_proc_compat(proc_compat_str):
    # To be consistent with HMC driver implementation and usability
    # need to convert _Plus with +
    # e.g Power6_Plus to Power6+
    return proc_compat_str.replace('_Plus', '+')
