# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
import copy
from datetime import datetime
from eventlet import greenthread
import random
import six
import time
from nova import context as ctx
from nova import exception
from nova import network as net_api
from nova.network import model as net_model
from nova.network import base_api
from nova.objects import block_device as blkdev_obj
from nova.objects import instance as inst_objs
from nova import rpc
from nova.virt import block_device as blk_dev
from nova.compute import task_states
from nova.compute import vm_states
from nova_powervm.virt.powervm import driver
from nova_powervm.virt.powervm import vm
from oslo_concurrency import lockutils
from oslo_serialization import jsonutils
from oslo_utils import encodeutils
from powervc_nova import logcall_log_instance
from powervc_nova.network.neutronv2 import ibmpowervm_api as pvc_net_api
from powervc_nova.objects.network import dom as network_dom
from powervc_nova.virt import instance_error_seen_multiple_time
from powervc_nova.virt import instance_found_remove_from_cache
from powervc_nova.virt import set_instance_error_state_and_notify
from powervc_nova.virt.ibmpowervm.common.constants import MIGRATING_STATES,\
    NOVALINK_PLATFORM, HMC_PLATFORM
from powervc_nova.virt.ibmpowervm.common import exception as common_exception
from powervc_nova.virt.ibmpowervm.common import image
from powervc_nova.virt.ibmpowervm.common import migrate_utils
from powervc_nova.virt.ibmpowervm.pvm import enterprise_pool as pvm_ep

from nova_powervm.virt.powervm import media
from pypowervm import adapter as pvm_apt
from pypowervm.exceptions import ConnectionError, TimeoutError
from pypowervm.helpers import log_helper as log_hlp
from pypowervm.tasks import network_bridger
from pypowervm.wrappers import managed_system as pvm_ms
from pypowervm.wrappers import shared_proc_pool as pvm_spp
import pypowervm.wrappers.managed_system as pvm_system
import pypowervm.wrappers.network as pvm_net
from pypowervm.utils import retry as pvm_retry
from pypowervm.utils import validation as vldn
from pypowervm.wrappers import virtual_io_server as pvm_vios
from pypowervm import const as pvm_const
from powervc_nova.virt.ibmpowervm.common import power_spec_utils
from powervc_nova.virt.ibmpowervm.pvm.import_vm import ImportVM
from powervc_nova.virt.ibmpowervm.pvm.metrics import LparMetrics
from nova import exception as nova_exception
from powervc_nova.virt.ibmpowervm.common import constants as pvc_consts
from nova_powervm.virt.powervm.tasks.network import\
    VirtualInterfaceUnplugException
from nova.compute import power_state

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils, importutils
from nova import volume
from nova.objects import flavor as flavor_obj
from powervc_oslo.config import data_utils

from nova_powervm.virt.powervm import vios
from nova_powervm.virt.powervm.tasks import storage as tf_stg
from nova_powervm.virt.powervm.tasks import vm as tf_vm
from nova_powervm.virt.powervm import vios as pvm_vios_util
from nova_powervm.virt.powervm import slot
from networking_powervm.plugins.ibm.agent.powervm import utils as nw_utils
from powervc_nova.objects.compute import dom as compute_dom
from powervc_nova.objects.compute.dom import SharedProcessorPoolFactory
from powervc_nova.objects.storage import dom as storage_dom
from powervc_nova.storage.storage_mapping_api import StorageMapping
import powervc_nova.virt.ibmpowervm.pvm.exception as pvm_exception
from powervc_nova.virt import discovery_driver as disc_driver
import powervc_nova.virt.ibmpowervm.pvm.host as pvm_host
from powervc_nova.virt.ibmpowervm.pvm import ephemeral_disk
from powervc_nova.virt.ibmpowervm.pvm import lpar as pvc_lpar
from powervc_nova import logcall
from ceilometer_powervm.compute.virt.powervm.inspector import PowerVMInspector
from powervc_nova.virt.ibmpowervm.pvm import live_migration as lpm
from powervc_nova.virt.ibmpowervm.pvm.tasks import vm as tf_pvc_vm
from powervc_nova import _

from taskflow import engines as tf_eng
from taskflow.patterns import linear_flow as tf_lf

PVC_FC_STATERGY_MAPPING = {'npiv': 'powervc_nova.virt.ibmpowervm.pvm.'
                           'volume.npiv.PowerVCNpivVolumeAdapter',
                           'vscsi': 'powervc_nova.virt.ibmpowervm.pvm.'
                           'volume.vscsi.PowerVCVscsiVolumeAdapter',
                           'pv_vscsi': 'powervc_nova.virt.ibmpowervm.'
                           'pvm.volume.vscsi.PowerVCVscsiVolumeAdapter'
                           }
pvm_opts = [
    cfg.StrOpt('ip', default=None,
               help='The IP address of the Novalink partition.'),
    cfg.StrOpt('username', default=None,
               help='The username of the Novalink partition.'),
    cfg.StrOpt('password', default=None, secret=True,
               help='The password for the Novalink partition.')
]

CONF = cfg.CONF
CONF.register_opts(pvm_opts, group='pvm')

LOG = logging.getLogger(__name__)


class PowerVMDriver(driver.PowerVMDriver, disc_driver.ComputeDiscoveryDriver):

    def __init__(self, virtapi):
        super(PowerVMDriver, self).__init__(virtapi)
        self.network_topo = None
        self.volume_api = volume.API()
        self._instances_managed_on_startup = False

    @logcall
    def attach_interface(self, instance, image_meta, vif):
        """
        Attach an interface to the instance.
        """
        # Create a thread to go off and do the attach async
        greenthread.spawn(self._async_attach_interface, instance, vif)

    @lockutils.synchronized('powervm_attach_ifc', 'nova-attachifc-powervm')
    def _async_attach_interface(self, instance, vif):
        '''
        This method will asynchronously do the attach of the network interface.
        It must be done asynchronously because few other platforms are very
        quick, but since Novalink takes longer...if done sync'd the user can
        cause many issues in the REST API (perhaps even a timeout).

        We will use bumblebee to send messages on status.
        '''
        notifier = rpc.get_notifier(service='compute', host=CONF.host)
        napi = net_api.API()
        admin_context = ctx.get_admin_context()

        # Notify the start of this.
        msg = (_("The network interface attach action has started on host "
                 "%(host)s.  This operation might take some time to "
                 "complete.  Do not modify the network interface until a "
                 "notification has been sent that indicates operation "
                 "completion.") %
               {'host': CONF.host_display_name})
        info = {'msg': msg, 'host_name': CONF.host_display_name}
        notifier.info(admin_context, 'compute.instance.log', info)

        try:
            # Continue on to the original attach code in the community.
            super(PowerVMDriver, self).plug_vifs(instance, [vif])

            # If PowerVC created this port, do not preserve on delete
            admin_context = ctx.get_admin_context()
            napi = net_api.API()
            port = napi.show_port(admin_context, vif.get('id'))
            name = port['port'].get('name', '')
            if name.startswith('powervc_attached'):
                # Clear the preserve flag
                vif['preserve_on_delete'] = False

                # Check to see if any other VIFs are already on this instance
                cur_vifs = napi.get_instance_nw_info(admin_context, instance)
                if not isinstance(cur_vifs, list):
                    cur_vifs = []

                # Remove / readd the VIF we're updating
                cur_vifs[:] = [d for d in cur_vifs if d.get('id') !=
                               vif.get('id')]
                cur_vifs.append(vif)

                # Convert VIF list to NetworkInfos
                net_infos = cur_vifs
                if not isinstance(net_infos, net_model.NetworkInfo):
                    net_infos = net_model.NetworkInfo(cur_vifs)

                # Update VIF list
                base_api.update_instance_cache_with_nw_info(
                    napi, admin_context, instance, net_infos)
        except Exception as e:
                LOG.exception(e)
                msg = _('Cannot create a virtual Ethernet adapter on the '
                        'virtual machine %(lpar_name)s because the Novalink '
                        'operation for create failed with an error. '
                        'Refer to the logs for more information before '
                        'you attempt to add the virtual Ethernet adapter '
                        'from the UI.') % {'lpar_name': instance.name}
                LOG.error(msg)
                info = {'msg': msg, 'host_name': CONF.host_display_name}
                notifier.error(admin_context, 'compute.instance.log', info)
                raise nova_exception.VirtualInterfaceCreateException()

        # Notify the successful operation
        msg = (_("The network interface attach action has completed for "
                 "host %(host)s.  Manual configuration of the interface "
                 "is required on the system %(lpar)s.") %
               {'host': CONF.host_display_name, 'lpar': instance.name})
        info = {'msg': msg, 'host_name': CONF.host_display_name,
                'instance_id': instance['uuid']}
        notifier.info(admin_context, 'compute.instance.log', info)

    @logcall
    def detach_interface(self, instance, vif):
        """
        Detach an interface from the instance.
        """
        notifier = rpc.get_notifier(service='compute', host=CONF.host)
        admin_context = ctx.get_admin_context()

        # Notify the start of this.
        msg = (_("The network interface detach action has started on host "
                 "%(host)s.  This operation may take some time to complete.  "
                 "A subsequent message will be displayed when the operation "
                 "completes.") %
               {'host': CONF.host_display_name})
        info = {'msg': msg, 'host_name': CONF.host_display_name}
        notifier.info(admin_context, 'compute.instance.log', info)

        try:
            # Continue on to the original detach code in the community.
            super(PowerVMDriver, self).unplug_vifs(instance, [vif])
        except Exception as e:
            LOG.exception(e)
            msg = (_('Cannot delete a virtual Ethernet adapter on the '
                     'virtual machine %(lpar_name)s because the Novalink '
                     'operation for delete failed. Refer to the logs '
                     'for more information and ensure that the virtual '
                     'Ethernet adapter is manually removed after you '
                     'attempt to delete it from the UI.') %
                   {'lpar_name': instance.name})
            LOG.error(msg)
            info = {'msg': msg, 'host_name': CONF.host_display_name}
            notifier.error(admin_context, 'compute.instance.log', info)
            raise VirtualInterfaceUnplugException()

        msg = (_("The network interface detach action has completed on host "
                 "%(host)s.") %
               {'host': CONF.host_display_name})
        info = {'msg': msg, 'host_name': CONF.host_display_name}
        notifier.info(admin_context, 'compute.instance.log', info)

    @logcall
    def init_host(self, host):
        """Initialize anything that is necessary for the driver to function,
        including catching up with currently running VM's on the given host."""
        context = ctx.get_admin_context()
        # Get the instances up front to be able to fix-up necessary data
        instfact = compute_dom.ManagedInstanceFactory.get_factory()
        dbinsts = instfact.find_all_instances_by_host(context, CONF.host)
        self._instances_managed_on_startup = len(dbinsts) > 0
        # After we get the Managed Instances, then we can call the parent
        super(PowerVMDriver, self).init_host(host)
        # Only need to do anything if there are any instances in the DB
        if len(dbinsts) > 0:
            # If this is right after the switch from HMC, do data conversion
            if CONF.powervm_switch_type == 'hmc':
                self.__convert_data_for_host_switch(context, dbinsts)
            # Next update the slot information for the missing instances
            self._update_slot_info(context, dbinsts)

        # Perform the Network Topology on Initialization to make sure is valid
        topo_class = CONF.powervm_vif_topo_driver
        topo_drv = importutils.import_object(topo_class,
                                             host_name=CONF.host,
                                             host_uuid=self.host_uuid,
                                             adapter=self.adapter)
        self.network_topo = topo_drv.get_host_network_topology(context)
        # Find the network VIF driver
        vif_class = CONF.powervm_vif_driver
        self.vif_driver = importutils.import_object(vif_class,
                                                    host_name=CONF.host)

        # Collect the enterprise pool data for the host. Note that this is
        # susceptible to a race condition with a set mobile cores operation,
        # but the likelihood that this occurs as well as the fact it will be
        # resolved by the conductor's periodic task means that it makes more
        # sense to ensure the DB is updated immediately after a host is
        # registered.
        try:
            LOG.info(_('Collecting Host-Enterprise-Pool Topology...'))
            pvm_ep.sync_pool_for_host(context, self.adapter, CONF.host)
            LOG.info(_('Done Collecting Host-Enterprise-Pool Topology.'))
        except Exception as e:
            # Just log the exception and continue
            LOG.error(_("Reconciling Host-Enterprise-Pool Topology failed."))
            LOG.exception(e)

    def _setup_rebuild_store(self):
        """Override the setup of Swift store to wait for it to start"""
        store = CONF.powervm.nvram_store.lower()
        if store != 'none':
            self.store_api = importutils.import_object(
                driver.NVRAM_NS + driver.NVRAM_APIS[store])
            # Wait for Swift to startup before doing initialization
            if self._instances_managed_on_startup:
                self._wait_for_swift_to_start()
            # Events will be handled once the nvram_mgr is set.
            self.nvram_mgr = driver.nvram_manager.NvramManager(
                self.store_api, self.adapter, self.host_uuid)
            # Do host startup for NVRAM for existing VMs on the host
            if self._instances_managed_on_startup:
                driver.n_utils.spawn(self._nvram_host_startup)

    def _wait_for_swift_to_start(self):
        """Helper method to wait for Swift to startup if it is down"""
        cur_wait_time, max_wait_time, interval = 0, 300, 15
        LOG.info(_("Waiting for the Swift Service to startup..."))
        # Wait for up to the maximum wait time for Swift to start
        while cur_wait_time < max_wait_time:
            start = datetime.now()
            # If the Swift Service is ready, then just return now
            if self._is_swift_service_ready():
                return
            # Determine how long we waited on the client call itself
            delta = (datetime.now() - start).total_seconds()
            delta = delta if delta > 0 else 0
            # If we waited less than 15 seconds, then delay ourselves
            if delta < interval:
                time.sleep(interval - delta)
            cur_wait_time += delta if delta > interval else interval
        # If we timed out waiting for it to start, raise an exception
        raise exception.NovaException(
            _("Timed out waiting for Swift to startup"))

    def _is_swift_service_ready(self):
        """Helper method to see if the Swift Service is started/ready"""
        try:
            try:
                results = self.store_api._run_operation(
                    'list', options={'long': True})
            # The method signature changed between nova-powervm builds
            except TypeError:
                results = self.store_api._run_operation(
                    None, 'list', options={'long': True})
            # If the results say not successful, then it isn't ready
            for result in results:
                if not result['success']:
                    return False
            # Add a dummy object, to confirm the object-server is up
            self.store_api._store('powervc_version', '', '1.3.1.0')
            return True
        # If we got an exception we will treat as not being ready
        except:
            return False

    def get_info(self, instance):
        """Override get_info to set the Instance to Error when Not Found"""
        try:
            info = super(PowerVMDriver, self).get_info(instance)
            # Need to get a property to do the query to determine if not found
            _x = info.state
            # Remove the Instance from the NotFound Cache when it is reported
            instance_found_remove_from_cache(instance)
            return info
        except exception.InstanceNotFound:
            # Set the instance to error/notify when not found consecutive times
            vm_state = instance['vm_state']
            task_state = instance['task_state']
            vm_state_allowed = [vm_states.ACTIVE, vm_states.STOPPED]
            repeats = CONF.number_of_exceptions_needed_for_missing_inst_pvm
            # Ignore setting to error if in certain states or a hiccup
            if (((vm_state in vm_state_allowed) and task_state is None and
                 instance_error_seen_multiple_time(instance, repeats))):
                set_instance_error_state_and_notify(instance)
            raise

    def _get_adapter(self):
        """Override this method to allow PowerVC to run nova compute
        remote from the Novalink partition.
        Also add handling where pvm is not ready for accept REST request.
        """
        if CONF.pvm.ip and CONF.pvm.username and CONF.pvm.password:
            self.session = pvm_apt.Session(host=CONF.pvm.ip,
                                           username=CONF.pvm.username,
                                           password=CONF.pvm.password,
                                           certpath=None)
            self.adapter = pvm_apt.Adapter(self.session,
                                           helpers=log_hlp.log_helper)
        else:
            super(PowerVMDriver, self)._get_adapter()

    @logcall_log_instance
    def snapshot(self, context, instance, image_id, update_task_state):
        """Snapshots the specified instance.

        :param context: security context
        :param instance: nova.objects.instance.Instance
        :param image_id: Reference to a pre-created image that will
                         hold the snapshot.
        """
        def get_lpar_state():
            return self.get_info(instance).state

        def remove_opticals():
            lpar_uuid = vm.get_pvm_uuid(instance)
            media_builder = media.ConfigDrivePowerVM(
                self.adapter, self.host_uuid)
            media_builder.dlt_vopt(lpar_uuid)

        img = image.Image()
        image_entry = image.ImageEntry(context, image_id)
        img.capture(context, instance, image_entry, update_task_state,
                    get_lpar_state, remove_opticals)

    def get_available_resource(self, nodename):
        """Retrieve resource information.

        This method is called when nova-compute launches, and
        as part of a periodic task

        :param nodename:
            node which the caller want to get resources from
            a driver that manages only one node can safely ignore this
        :returns: Dictionary describing resources
        """

        data = super(PowerVMDriver, self).get_available_resource(nodename)
        new_data = pvm_host.get_host_stats(self.adapter)
        # copy over the info
        for k, v in new_data.iteritems():
            data[k] = v
        # Safety-check supported_instances until nova-powervm is on Mitaka
        supp_inst = data.get('supported_instances')
        if supp_inst and not isinstance(supp_inst, list):
            data['supported_instances'] = jsonutils.loads(supp_inst)
        # Add in the Management Version so the UI can tell whether to update
        data['stats']['mgmt_version'] = data_utils.get_powervc_version()
        return data

    def get_available_nodes(self, refresh=False):
        """Override of nova-powervm get_available_nodes. The community driver
        returns the wrapper mtms_str with a *, this is not what the scheduler
        expects so we return with the proper value set at registration.
        """

        return [CONF.host]

    def inventory_host(self, context):
        """
        Provides a mechanism for the Driver to gather Inventory-related
        information for the Host at periodic intervals.  The Driver is
        free from there to populate the information directly in the
        Database and should not return any information from the method.

        :param context: The security context for the query
        """
        topo_class = CONF.powervm_vif_topo_driver
        topo_drv = importutils.import_object(topo_class,
                                             host_name=CONF.host,
                                             host_uuid=self.host_uuid,
                                             adapter=self.adapter)
        network_topo, self.network_topo = (self.network_topo, None)
        # Collect the Network Topology Information and update it in the DB
        try:
            LOG.info(_('Collecting Host-Network Topology...'))
            # We retrieved the Network Topology on init_host to verify the
            # configuration is valid, so we will use that one if it is cached
            if network_topo is None:
                network_topo = topo_drv.get_host_network_topology(context)
            # Call save on the Host Network Topology to update it in the DB
            network_topo.save()

            self._cleanup_neutron_ports(context)

            LOG.info(_('Done Collecting Host-Network Topology.'))
        except Exception as e:
            # Just log the exception and continue on if Network Topology failed
            LOG.error(_("Reconciling Host-Network Topology failed."))
            LOG.exception(e)

        # Collect the Storage Topology Information and update it in the DB
        try:
            LOG.info(_('Collecting Host-Storage Topology...'))
            topofact = storage_dom.HostStorageTopologyFactory.get_factory()
            stg_topo = topofact.construct_storage_topology(
                context, CONF.host)
            topo_helper = stg_topo.construct_pvm_helper(
                CONF.host, self.host_uuid, self.adapter)
            topo_helper.reconcile_topology()
            LOG.info(_('Done Collecting Host-Storage Topology.'))
        except Exception as e:
            # Just log the exception and continue on if Storage Topology failed
            LOG.error(_("Reconciling Host-Storage Topology failed."))
            LOG.exception(e)

        # Update shared processor pool information
        try:
            LOG.info(_('Collecting Host-Shared proc pool information...'))
            # Updates shared proc pool info from k2 to DB
            self._sync_shared_proc_pools(context)
            LOG.info(_('Done Collecting Host-Shared proc pool information...'))
        except Exception as e:
            LOG.error(_("Reconciling Host-Shared proc pool "
                        "information failed."))
            LOG.exception(e)

    def _sync_shared_proc_pools(self, context):
        # Extract all shared processor pools info.
        pool_wraps = pvm_spp.SharedProcPool.get(self.adapter,
                                                parent_type=pvm_ms.System,
                                                parent_uuid=self.host_uuid)
        proc_pools = []
        for pool in pool_wraps:
            args = dict(pool_name=pool.name,
                        pool_id=pool.id,
                        host_name=CONF.host,
                        maximum_proc_units=pool.max_proc_units,
                        reserved_proc_units=pool.curr_rsrv_proc_units,
                        available_proc_units=pool.avail_proc_units)
            spp = compute_dom.\
                SharedProcessorPool(context,
                                    SharedProcessorPoolFactory,
                                    None, **args)
            proc_pools.append(spp)
        # Get the ResourceTransaction (cf. DB2 session) for the nova db
        hct_factory = compute_dom.HostComputeTopologyFactory.get_factory()
        hct_topo = hct_factory.\
            construct_compute_topology(context, host_name=CONF.host,
                                       shared_processor_pools=proc_pools)
        hct_topo.save()

    def get_volume_connector(self, instance):
        # Get the scg for the given instance and add the
        # volume adapter object to the connector
        scg_map = StorageMapping()
        context = ctx.get_admin_context()
        connectivity_types = None
        scg_info = scg = None
        LOG.debug("Instance %s: task_state=%s, power_state=%d" %
                  (instance.name, instance.task_state, instance.power_state))
        if instance.task_state in [task_states.DELETING,
                                   task_states.MIGRATING,
                                   task_states.RESIZE_MIGRATING,
                                   task_states.RESIZE_FINISH]:
            # Don't compute the full connectivity information. Just get the
            # SCG and pull in the set of connectivity types.
            scg = scg_map.get_scg_from_instance_meta(
                context, instance, {'messages': []})
            if not scg:
                # handle onboarded case.
                response = scg_map.determine_scg(context, instance=instance,
                                                 host=CONF.host,
                                                 raise_err=True)
                LOG.debug("Determine SCG messages: %s" % response['messages'])
                scg = response['scg']
            if scg:
                ctypes = set(scg.boot_connectivity)
                ctypes.update(scg.data_connectivity)
                connectivity_types = list(ctypes)
                LOG.debug("connectivity types from SCG: %s" % ctypes)
                boot_types = scg.boot_connectivity
                data_types = scg.data_connectivity
        if not connectivity_types:
            scg_info = scg_map.get_connectivity_info(context,
                                                     instance['uuid'],
                                                     CONF.host,
                                                     instance)
            connectivity_types = scg_info.get('connectivity_types')
            boot_types = scg_info.get('boot_conn_types')
            data_types = scg_info.get('data_conn_types')

        if not connectivity_types:
            if not scg_info and scg:
                scg_info = scg.to_dict()
            LOG.error(_('Missing volume connector because the storage '
                        'connectivity group is None or does not contain '
                        'supported connectivity type. %s') % scg_info)
            return None

        connector = {'host': CONF.host}
        # build a dict for each connection type supported by the instance scg
        for conn_type in connectivity_types:
            # community driver looks for vscsi whereas powervc understands
            # pv_vscsi
            ctype = 'vscsi' if conn_type == 'pv_vscsi' else conn_type
            conn_adap = self._get_inst_vol_adpt(context, instance,
                                                {'connection-type': ctype})
            conn_dict = dict(connection_type=ctype,
                             vol_adap=conn_adap)
            # check if boot_conn type
            if boot_types and conn_type in boot_types:
                conn_dict['is_boot_type'] = True
            if data_types and conn_type in data_types:
                conn_dict['is_data_type'] = True

            connector[conn_type] = conn_dict

        return connector

    def _get_inst_vol_adpt(self, context, instance, conn_info=None,
                           stg_ftsk=None):
        """Method for returning volume driver object by connection type"""
        # If we have connection-type, we can just use that
        LOG.debug("conn_info=%s" % six.text_type(conn_info))
        drv_type = self._get_drv_type_by_conn_info(conn_info)
        if not drv_type:
            # Get the driver type through the scg
            drv_type = self._get_drv_type_by_scg(context, instance, conn_info)

        # could not find drv_type in conn_info and SCg. Throw exception
        if drv_type is None:
            raise pvm_exception.IBMPowerVMVolumeAdapterNotFound(
                uuid=instance['uuid'])

        adp_type = PVC_FC_STATERGY_MAPPING[drv_type]
        vol_cls = importutils.import_class(adp_type)
        vol_drv = vol_cls(self.adapter, self.host_uuid, instance, conn_info,
                          stg_ftsk=stg_ftsk)
        return vol_drv

    def _get_inst_xag(self, instance, bdms, recreate=False):
        """
        Override the inst xag to get from the drv type bdm
        """
        # If it's a recreate we need everything.
        if recreate:
            return [pvm_const.XAG.VIO_SMAP, pvm_const.XAG.VIO_FMAP,
                    pvm_const.XAG.VIO_STOR]
        # All operations for deploy/destroy require scsi by default.  This is
        # either vopt, local/SSP disks, etc...
        # xags is also used for scrubbing left over storage server adapters.
        # we need to include fc_mapping by default as well
        xags = {pvm_const.XAG.VIO_SMAP,
                pvm_const.XAG.VIO_FMAP}
        if not bdms:
            LOG.debug('Instance XAGs for VM %(inst)s is %(xags)s.',
                      {'inst': instance.name,
                       'xags': ','.join([x for x in xags])})
            return list(xags)

        for bdm in bdms:
            conn_info = bdm.get('connection_info')
            drv_type = self._get_drv_type_by_conn_info(conn_info)
            if drv_type:
                adp_type = PVC_FC_STATERGY_MAPPING[drv_type]
                vol_cls = importutils.import_class(adp_type)
                xags.update(set(vol_cls.min_xags()))

        LOG.debug('Instance XAGs for VM %(inst)s is %(xags)s.',
                  {'inst': instance.name,
                   'xags': ','.join([x for x in xags])})
        return list(xags)

    def _get_drv_type_by_conn_info(self, conn_info):
        drv_type = None
        if conn_info:
            if conn_info.get('connection-type') is not None:
                drv_type = conn_info.get('connection-type')
            elif ((conn_info.get('data') and
                   conn_info['data'].get('connection-type'))):
                drv_type = conn_info['data'].get('connection-type')
            elif ((conn_info.get('volume_connector') and
                   conn_info['volume_connector'].get('connection-type'))):
                drv_type = conn_info['volume_connector'].get('connection-type')
        return drv_type

    def _get_drv_type_by_scg(self, context, instance, conn_info):
        """Helper Method to get volume driver type through the SCG"""
        scg_map = StorageMapping()
        if not conn_info.get('data') or not conn_info['data'].get('volume_id'):
            return None
        # Get the volume to determine if it is a boot volume or not
        volume_id = conn_info['data']['volume_id']
        volume = self.volume_api.get(context, volume_id)
        is_boot = volume.get('is_boot_volume')
        # Get the scg for the instance
        scg_info = scg_map.get_connectivity_info(context, instance['uuid'],
                                                 CONF.host, instance)
        # Use is_boot to get the driver type from the scg
        drv_types = (scg_info.get('boot_conn_types') if is_boot else
                     scg_info.get('data_conn_types'))
        LOG.debug("is_boot=%s, drv_types=%s" % (is_boot, drv_types))
        # Now we need to pick a single type from the list. The following
        # logic only works because novalink does not support SSP yet.
        # TODO: When novalink supports both SSP and NPIV data disks, then
        #       we need to pick a preference or consider this case of
        #       missing the connection type in the BDM an error.
        filtered_types = [t for t in drv_types if t != 'ibm_ssp']
        return filtered_types[0]

    def _get_disk_adapter(self):
        self.disk_dvr = ephemeral_disk.NoEphemeralDisk()

    @logcall
    def sync_enterprise_pool_for_host(self, context):
        """Syncs enterprise pool info corresponding to this host."""
        return pvm_ep.sync_pool_for_host(context, self.adapter, CONF.host)

    @logcall
    def set_mobile_cores(self, context, desired_mobile_cores):
        """Sets the number of mobile cores for this host."""
        return pvm_ep.set_mobile_cores(context, desired_mobile_cores,
                                       CONF.host, self.adapter)

    @logcall
    @lockutils.synchronized('change_powervm_vifs', 'nova-change-vif-powervm-')
    def change_vif_sea(self, context, data):
        """Moves a vif from one sea to another, based on data struct sent in.
        """
        # List all the networks we're to update to point to the new sea
        net_ids = [data['net_id']]
        net_ids.extend(data.get('peers', []))

        # Remember the current mapping in case we need to roll back
        netafact = network_dom.NetworkAssociationFactory.get_factory()
        old_sea = ''
        for net_id in net_ids:
            neta = netafact.find_network_assoc(context, CONF.host, net_id)
            if neta and neta.sea:
                old_sea = neta.sea.name
                break

        # Find the right SEA to use
        vios_wraps = pvm_vios.VIOS.wrap(
            self.adapter.read(
                root_type=pvm_ms.System.schema_type, root_id=self.host_uuid,
                child_type=pvm_vios.VIOS.schema_type,
                xag=[pvm_const.XAG.VIO_NET]))
        nb_wraps = pvm_net.NetBridge.wrap(
            self.adapter.read(root_type=pvm_system.System.schema_type,
                              root_id=self.host_uuid,
                              child_type=pvm_net.NetBridge.schema_type))
        nb_uuid = None
        for vios_wrap in vios_wraps:
            if vios_wrap.id == data['new_lpar_id']:
                for nb_wrap in nb_wraps:
                    for sea in nb_wrap.seas:
                        if sea.dev_name == data['new_sea'] and\
                                sea.vio_uri == vios_wrap.related_href:
                            nb_uuid = nb_wrap.uuid
                            break
                break

        # Gather the proper info to put out a useful info log
        napi = net_api.API()
        neutron_obj = napi.get(context, data['net_id'])
        net_name = neutron_obj.name if neutron_obj else data['net_id']
        LOG.info(_("Network %(net)s is being modified to use Shared "
                   "Ethernet Adapter (SEA) %(new_sea)s instead of SEA "
                   "%(old_sea)s."),
                 {'net': net_name, 'old_sea': old_sea,
                  'new_sea': data['new_sea']})

        # Now update the network bridge
        net_bridge_updated = False
        notifier = rpc.get_notifier(service='compute', host=CONF.host)
        try:
            network_bridger.ensure_vlan_on_nb(
                adapter=self.adapter, host_uuid=self.host_uuid,
                nb_uuid=nb_uuid, vlan_id=data['vlan'])
            net_bridge_updated = True

            # Inform the user of success
            msg = (_("Network %(net)s was successfully modified to use "
                     "Shared Ethernet Adapter (SEA) %(new_sea)s instead of "
                     "SEA %(old_sea)s on host %(host)s. Connectivity was "
                     "restored on all virtual machines that use the network")
                   % {'host': CONF.host, 'net': net_name, 'old_sea': old_sea,
                      'new_sea': data['new_sea']})
            info = {'msg': msg, 'sea_name': data['new_sea'],
                    'host_name': CONF.host}
            LOG.info(msg)
            notifier.info(context, 'compute.instance.log', info)
        except Exception as e:
            LOG.exception(e)

            # Log and inform user of failure
            msg = (_("The adapter change on Host %(host)s for Network "
                     "%(net)s failed to finish successfully.  The system has"
                     " returned the following error message: %(msg)s.  "
                     "Please review the issue and retry the command.") %
                   {'host': CONF.host, 'msg': e.message, 'net': net_name})
            warn = {'msg': msg, 'sea_name': data['new_sea'],
                    'host_name': CONF.host}
            LOG.warn(msg)
            notifier.warn(context, 'compute.instance.log', warn)

        # Update the network association
        if net_bridge_updated:
            self.__update_na_and_peers_in_db(
                context, data['new_lpar_id'], [(data['new_sea'], net_ids)])

    @logcall
    def migration_cleanup(self, context, instance, source=False):
        pass

    def __update_na_and_peers_in_db(self, context, lpar_id, sea_na_tuple_list):
        """
        Updates the network associations for the net_ids with a new SEA.

        :param context: The user context making the request
        :param lpar_id: The lpar housing the sea
        :param sea_na_tuple_list: A list of SEA/NA tuples to update.
        """
        netafact = network_dom.NetworkAssociationFactory.get_factory()
        hostfact = network_dom.ManagedHostFactory.get_factory()
        transaction = netafact.get_resource_transaction()
        with transaction.begin(subtransactions=True):
            host_dom = hostfact.find_host_by_name(
                context, CONF.host, transaction)

            for ttuple in sea_na_tuple_list:
                sea = ttuple[0]
                net_ids = ttuple[1]
                sea_dom = host_dom.find_adapter(lpar_id, sea)
                for net_id in net_ids:
                    netafact.update_network_assoc(
                        context, CONF.host, net_id, sea_dom, transaction)

    def discover_instances(self, context):
        """
        Returns a list of all of the VM's that exist on the given Host.
        For each VM the driver needs to return a dictionary containing
        the following attributes:
            name:      The Name of the VM defined on the Hypervisor
            state:     The State of the VM, matching the power_state definition
            uuid:      The UUID of the VM when created thru OS (Optional)
            hostuuid:  The Hypervisor-specific UUID of the VM (Optional)
            supported: Dictionary stating whether the VM can be managed
              status:  Whether or not it is supported
              reasons: List of Dictionaries as to why it isn't supported
                id:      Message ID to key off of for the reason
                message: Message Text as to why it isn't supported

        :param context:   The security context for the query
        """
        disc_insts = []
        instances = vm.get_lpars(self.adapter)
        importvm = ImportVM(self.host_uuid, self.adapter)
        info = importvm.list_lpar_for_onboard(context, instances)
        disc_insts += info
        return disc_insts

    def query_instances(self, context, instances):
        """
        Returns a list of VM's (matching those specified on input) with
        enough additional details about each VM to be able to import the
        VM's into the Nova Database such as OpenStack can start managing.
        For each VM the driver needs to return a dictionary containing
        the following attributes:
            uuid:      The UUID of the VM when created thru OS
            name:      The Name of the VM defined on the Hypervisor
            state:     The State of the VM, matching the power_state definition
            hostuuid:  The Hypervisor-specific UUID of the VM (Optional)
            vcpus:     The number of Virtual CPU's currently allocated
            memory_mb: The amount of Memory currently allocated
            root_gb:   The amount of storage currently allocated
            ephemeral_gb:  The amount of ephemeral storage currently allocated
            flavor:    Dictionary containing the profile/desired allocations
              vcpus:     The number of Virtual CPU's defined for allocation
              memory_mb: The amount of Memory defined for allocation
              root_gb:   The amount of storage defined for allocation
              ephemeral_gb:  The ephemeral storage defined for allocation
              powervm:proc_units:  The number of proc_units defined (Optional)
            volumes:   List of dictionary objects with the attached volumes:
              uuid:      The UUID of the Volume when created thru OS (Optional)
              name:      The Name of the Volume defined on the Backend
              host:      The registered name of the Storage Provider
              provider_location:  The Volume ID known on the Backend (Optional)
            network_ports: List of dictionary objects with the network ports:
              mac_address:  The MAC Address of the Port allocated to the VM
              status:       The Status of the Port, matching the definition
              segmentation_id: The VLAN ID for the Network the Port is part of
              physical_network: The Physical Network the Port is part of

        :param context:   The security context for the query
        :param instances: A list of dictionary objects for each VM containing:
            uuid:      The UUID of the VM when created thru OS
            name:      The Name of the VM defined on the Hypervisor
            hostuuid:  The Hypervisor-specific UUID of the VM (Optional)
        """
        disc_insts = []
        importvm = ImportVM(self.host_uuid, self.adapter)
        info = importvm.query_onboarding_instances(context, instances)
        disc_insts += info
        return disc_insts

    def inventory_instances(self, context, instances):
        """
        Provides a mechanism for the Driver to gather Inventory-related
        information for the managed LPARs at periodic intervals.
        The Driver is free from there to populate the information directly
        in the Database and should not return any information from
        the method.

        :param context: The security context for the query
        :param instances: the db instances for which the stats need to be
        updates
        """
        # No reason to even collect the inventory if no instances on host
        if not instances:
            return
        LOG.info(_('Collecting instance inventory.'))
        shared_proc_pools_id_names = pvm_host.get_shared_proc_pools_id_name(
            self.adapter, self.host_uuid)
        # Grab the LPAR feed from the hypervisor and map the instance UUIDs to
        # their wrappers. This way we only go out to the network once.
        lpars_map = {lpar.uuid: lpar for lpar in vm.get_lpars(self.adapter)}
        # Since there may be db entries not on the hypervisor filter those out
        instances = [i for i in instances if vm.get_pvm_uuid(i) in lpars_map]
        # Collect the instance inventory and stats
        # Loop through each of the Instances getting the PowerSpecs for them
        for instance in instances:
            try:
                pvm_uuid = vm.get_pvm_uuid(instance)
                entry = lpars_map[pvm_uuid]

                lpar_stats = pvc_lpar.get_lpar_power_stats(entry)
                # Add the shared processor pool info
                proc_pool_id = lpar_stats.get('shared_proc_pool_id')
                if proc_pool_id is not None:
                    if proc_pool_id == 0:
                        proc_pool_name = pvm_spp.DEFAULT_POOL_DISPLAY_NAME
                    else:
                        proc_pool_name = shared_proc_pools_id_names.\
                            get(proc_pool_id)
                else:
                    proc_pool_name = None
                    lpar_stats['shared_proc_pool_id'] = 0

                lpar_stats['shared_proc_pool_name'] = proc_pool_name
                # Call the Common Utility to update PowerSpecs in the DB
                power_spec_utils.update_instance_power_specs(
                    context, self.virtapi, instance, lpar_stats.copy())

            except Exception as e:
                # log the exception and continue with next LPAR
                LOG.warn(_('Error retrieving PowerSpecs for %s') %
                         instance.name)
                LOG.exception(e)
        # Update the slot information for the missing instances
        self._update_slot_info(context, instances)

    def monitor_instances(self, context, instances):
        """
        Returns details about the VM's specified that can be considered
        an accurate inventory of the VM that is periodically collected.
        For each VM the driver needs to return a dictionary containing
        the following attributes:
            uuid:      The UUID of the Instance that was passed in
            name:      The Name of the Instance that was passed in
            cpu_utilization: The percent of CPU that is being used
            mem_utilization: Represents the percentage of real page frames
                            that are currently available on the VMM free list.

        :param context: The security context for the query
        :param instances: A list of dictionary objects for each VM containing:
            uuid:      The UUID of the VM when created thru OpenStack
            name:      The Name of the VM when created thru OpenStack
            vm_uuid:   The Hypervisor-specific UUID of the VM (Optional)
        """
        LOG.info(_('Collecting instances cpu utilization data'))
        data = []
        if self.adapter is None:
            # If periodic tasks starts before host is init.
            # Should not come here.
            LOG.warn(_('Cannot collect metrics in this interval as adapter '
                       'is not yet initialized.'))
            return data
        lpar_metrics = LparMetrics.get_instance(self.adapter)
        for instance in instances:
            cpu_util, mem_util_dict = None, None
            try:
                # Get CPU util. This may throw exception from Inspector
                # if it can't find VM.
                cpu_util = lpar_metrics.get_cpu_util_for_lpar(instance)
            except Exception:
                LOG.warn(_('Error retrieving cpu utilization for %s') %
                         instance.name)
            try:
                # Get memory utilization. This can throw exception from
                # pypowervm
                mem_util_dict = lpar_metrics.get_mem_util_for_lpar(instance)
            except Exception:
                LOG.warn(_('Error retrieving mem utilization for %s') %
                         instance.name)
            metrics_dict = dict(uuid=instance["uuid"],
                                name=instance["name"])
            if cpu_util is not None:
                self._update_metrics_dict(
                    metrics_dict, 'cpu_util', cpu_util.util)
            if mem_util_dict:
                self._update_metrics_dict(
                    metrics_dict, 'mem_util',
                    mem_util_dict.get('mem_util'))
                self._update_metrics_dict(
                    metrics_dict, 'pg_in_rate',
                    mem_util_dict.get('pg_in_rate'))
                self._update_metrics_dict(
                    metrics_dict, 'pg_out_rate',
                    mem_util_dict.get('pg_out_rate'))
                self._update_metrics_dict(
                    metrics_dict, 'swap_in_rate',
                    mem_util_dict.get('swap_in_rate'))
                self._update_metrics_dict(
                    metrics_dict, 'swap_out_rate',
                    mem_util_dict.get('swap_out_rate'))
                data.append(metrics_dict)
        LOG.debug("Returning utilization metrics " + unicode(data))
        return data

    def _update_metrics_dict(self, metrics_dict, metric_key, metric_val):
        if metric_val is not None:
            metrics_dict.update({metric_key: metric_val})

    def spawn(self, context, instance, image_meta, injected_files,
              admin_password, network_info=None, block_device_info=None):

        admin_ctx = ctx.get_admin_context(read_deleted='yes')
        flavor = (
            flavor_obj.Flavor.get_by_id(admin_ctx,
                                        instance.instance_type_id))

        # The following attach is needs to done until we get
        # community bug https://bugs.launchpad.net/nova/+bug/1488111 fixed.
        # This bug prevents deploy reschedule from happening when
        # initialize_connection on volumes fails.
        # Until that is fixed this attach is being done in the virt driver
        # spawn method.
        bdms = self._extract_bdm(block_device_info)

        blk_dev.attach_block_devices(bdms, context, instance,
                                     self.volume_api, self,
                                     do_check_attach=True)

        super(PowerVMDriver, self).spawn(context, instance, image_meta,
                                         injected_files, admin_password,
                                         network_info=network_info,
                                         block_device_info=block_device_info,
                                         flavor=flavor)

    def _forbid_ssp(self, instance):
        """Helper method for forbidding LPM if instance's scg uses 'ibm-ssp'
           connectivity.

           Raise MigrationPreCheckError exception if instance's scg uses
           'ibm-ssp' connectivity.
        """
        scg_map = StorageMapping()
        context = ctx.get_admin_context()
        scg = scg_map.get_scg_from_instance_meta(context, instance,
                                                 {'messages': []})
        if scg is not None:
            if (pvc_consts.CONNECTION_TYPE_SSP in scg.get('boot_connectivity')
                    or pvc_consts.CONNECTION_TYPE_SSP in
                    scg.get('data_connectivity')):
                msg = (_("Cannot migrate instance %s because SSP connectivity "
                         "is not supported for NovaLink.")
                       % instance.name)
                raise exception.MigrationPreCheckError(reason=msg)

    def migrate_disk_and_power_off(self, context, instance, dest,
                                   flavor, network_info,
                                   block_device_info=None,
                                   timeout=0, retry_interval=0):
        """Transfers the disk of a running instance in multiple phases, turning
           off the instance before the end. In case of the resize operation
           instance would not be turned off.
        :param context: the context for the migration/resize
        :param instance: nova.objects.instance.Instance
        :param dest: destination host
        :param network_info:
           :py:meth:`~nova.network.manager.NetworkManager.get_instance_nw_info`
        :param block_device_info: instance volume block device info
        :param timeout: time to wait for GuestOS to shutdown
        :param retry_interval: How often to signal guest while
                               waiting for it to shutdown

        :returns: disk_info dictionary that is passed as the
                  disk_info parameter to finish_migration
                  on the destination nova-compute host
        """
        disk_info = {}

        # We may be passed a flavor that is in dict format, but the
        # downstream code is expecting an object, so convert it.
        if flavor and not isinstance(flavor, flavor_obj.Flavor):
            flav_obj = flavor_obj.Flavor.get_by_id(context, flavor['id'])
        else:
            flav_obj = flavor

        if flav_obj and flav_obj.root_gb < instance.root_gb:
            raise exception.InstanceFaultRollback(
                exception.ResizeError(reason=_('Cannot reduce disk size.')))

        if dest == self.get_host_ip_addr():
            self._log_operation('resize', instance)
        else:
            self._log_operation('migration', instance)
            raise NotImplementedError()

        pvm_inst_uuid = vm.get_pvm_uuid(instance)

        # Define the migrate flow
        flow = tf_lf.Flow("resize_vm")

        if flav_obj.root_gb > instance.root_gb:
            # Resize the root disk
            flow.add(tf_stg.ExtendDisk(self.disk_dvr, context, instance,
                                       dict(type='boot'), flav_obj.root_gb))

        # We rename the VM to help identify if this is a resize and so it's
        # easy to see the VM is being migrated from pvmctl.  We use the resize
        # name so we don't destroy it on a revert when it's on the same host.
        new_name = self._gen_resize_name(instance, same_host=True)
        flow.add(tf_vm.Rename(self.adapter, self.host_uuid, instance,
                              new_name))
        try:
            tf_eng.run(flow)
        except Exception as e:
            raise exception.InstanceFaultRollback(e)

        return disk_info

    def pre_live_migration(self, context, instance, block_device_info,
                           network_info, disk_info, migrate_data=None):
        """Prepare an instance for live migration

        :param context: security context
        :param instance: nova.objects.instance.Instance object
        :param block_device_info: instance block device information
        :param network_info: instance network information
        :param disk_info: instance disk information
        :param migrate_data: implementation specific data dict.
        """
        ret_val = super(PowerVMDriver, self).pre_live_migration(
            context=context,
            instance=instance,
            block_device_info=block_device_info,
            network_info=network_info,
            disk_info=disk_info,
            migrate_data=migrate_data)

        self._ensure_vlan_plugged(context, instance, network_info)
        return ret_val

    @lockutils.synchronized('ensure_vlan_plug', 'nova-ensure-vlan-plug')
    def _ensure_vlan_plugged(self, context, instance, network_info):
        """
        Ensure the VLANs specified in the network_info are plugged on the host.

        :param context: security context
        :param instance: nova.objects.instance.Instance object
        :param network_info: instance network information
        """
        try:
            # Get the network DOM
            self.hostfact = network_dom.ManagedHostFactory.get_factory()
            host_dom = self.hostfact.find_host_by_name(context=context,
                                                       host_name=CONF.host)

            # Call pypowervm to get the Network Bridges for this host
            response_feed = self.adapter.read(
                root_type=pvm_system.System.schema_type,
                root_id=self.host_uuid,
                child_type=pvm_net.NetBridge.schema_type)
            bridge_wrap = pvm_net.NetBridge.wrap(response_feed)
            if not bridge_wrap:
                msg = _('Error retrieving Network Bridges for instance %s') % \
                    instance['name']
                raise Exception(msg)

            # Go through each network the VM uses
            for vif in network_info:
                # Find the VLAN and SEA for this VIF
                vswitch, vlan_id, sea, needs_provision =\
                    self.vif_driver._parse_vif(host=host_dom,
                                               vif=vif,
                                               here_for_plug=True)
                if not vlan_id:
                    msg = _('Error retrieving vlan_id for %s') % \
                        instance['name']
                    raise Exception(msg)
                if not sea:
                    msg = _('Error retrieving SEA for instance %(name)s and '
                            'vlan_id %(vlan_id)s') % \
                        {'name': instance['name'],
                         'vlan_id': vlan_id}
                    raise Exception(msg)

                # Find the Network Bridge with the specified SEA
                net_bridge_uuid = None
                for bridge_w in bridge_wrap:
                    for sea_w in bridge_w.seas:
                        if(sea.name == sea_w.dev_name and
                           sea.vioserver.id in sea_w.vio_uri):
                            net_bridge_uuid = bridge_w.uuid
                            break
                if not net_bridge_uuid:
                    msg = _('Error finding Network Bridge for instance '
                            '%(name)s and vlan_id %(vlan_id)s and SEA '
                            '%(sea_name)s') % {'name': instance['name'],
                                               'vlan_id': vlan_id,
                                               'sea_name': sea.name}
                    raise Exception(msg)

                # Call pypowervm to verify VLAN is plugged
                network_bridger.ensure_vlan_on_nb(adapter=self.adapter,
                                                  host_uuid=self.host_uuid,
                                                  nb_uuid=net_bridge_uuid,
                                                  vlan_id=vlan_id)
        except Exception as e:
            error = six.text_type(e)
            inst_name = instance['name']
            msg = (_("Failed to enable VLAN bridging for virtual machine "
                     "%(inst_name)s. %(error)s") % locals())
            LOG.exception(e)
            raise common_exception.IBMPowerVMMigrationFailed(msg)

    def finish_migration(self, context, migration, instance, disk_info,
                         network_info, image_meta, resize_instance,
                         block_device_info=None, power_on=True):
        """Completes a resize or cold migration.

        :param context: the context for the migration/resize
        :param migration: the migrate/resize information
        :param instance: nova.objects.instance.Instance being migrated/resized
        :param disk_info: the newly transferred disk information
        :param network_info:
           :py:meth:`~nova.network.manager.NetworkManager.get_instance_nw_info`
        :param nova.objects.ImageMeta image_meta:
            The metadata of the image of the instance.
        :param resize_instance: True if the instance disks are being resized,
                                False otherwise
        :param block_device_info: instance volume block device info
        :param power_on: True if the instance should be powered on, False
                         otherwise
        """
        # See if this was to the same host
        same_host = migration.source_compute == migration.dest_compute

        if same_host:
            self._log_operation('finish resize', instance)
        else:
            self._log_operation('finish migration', instance)

        # Get the new flavor
        flav_obj = flavor_obj.Flavor.get_by_id(
            context, migration.new_instance_type_id)

        # Define the flow
        flow = tf_lf.Flow("finish_resize")

        if same_host:
            # This is just a resize.
            new_name = self._gen_resize_name(instance, same_host=True)
            flow.add(tf_pvc_vm.Resize(context, self.adapter, self.host_wrapper,
                                      instance, flav_obj, name=new_name))

        try:
            tf_eng.run(flow)
        except Exception as e:
            emsg = six.text_type(e)
            LOG.exception(_('Resize Error: %s') % emsg)
            # Check if the exception is due to insufficient memory
            if ('PVME01050403-0569' in emsg):
                LOG.error(_('Error resizing virtual machine %(name)s.'
                            ' Exception: %(exc)s') %
                          {'name': instance.display_name, 'exc': emsg})
                ex_args = {'instance_name': instance.display_name}
                raise pvm_exception.IBMPowerVMOSRemMemFailed(**ex_args)
            raise exception.InstanceFaultRollback(e)

    def finish_revert_migration(self, context, instance, network_info,
                                block_device_info=None, power_on=True):
        """Finish reverting a resize.

        :param context: the context for the finish_revert_migration
        :param instance: nova.objects.instance.Instance being migrated/resized
        :param network_info:
           :py:meth:`~nova.network.manager.NetworkManager.get_instance_nw_info`
        :param block_device_info: instance volume block device info
        :param power_on: True if the instance should be powered on, False
                         otherwise
        """
        self._log_operation('revert resize', instance)
        # TODO(IBM): What to do here?  Do we want to recreate the LPAR
        # Or just change the settings back to the flavor?

        # Get the flavor from the instance, so we can revert it
        admin_ctx = ctx.get_admin_context(read_deleted='yes')
        flav_obj = (
            flavor_obj.Flavor.get_by_id(admin_ctx,
                                        instance.instance_type_id))
        # TODO(IBM)  Get the entry once for update
        vm.update(self.adapter, self.host_uuid, instance, flav_obj)

    def post_live_migration_at_source(self, context, instance, network_info):
        """Overriding to delete slot information of the instance only
           when vm is migrated from Novalink to HMC, as it is not needed
           any more
        """
        super(PowerVMDriver, self).post_live_migration_at_source(
            context, instance, network_info)
        mig = self.live_migrations[instance.uuid]
        dest_host_type = mig.mig_data.managed_by
        if dest_host_type == HMC_PLATFORM:
            try:
                inst_slot_map = slot.SwiftSlotManager(self.store_api,
                                                      instance=instance)
                inst_slot_map.delete()
            except Exception as ex:
                LOG.exception(_("Unable to delete slot information %s "
                                % six.text_type(ex)))

    def post_live_migration_at_destination(self, context, instance,
                                           network_info,
                                           block_migration=False,
                                           block_device_info=None):
        """Overriding to populate the NVRAM after migration is done"""
        # Refresh the instance to ensure fabric information is present
        # in the system metadata
        instance.refresh()
        super(PowerVMDriver, self).post_live_migration_at_destination(
            context, instance, network_info,
            block_migration, block_device_info)
        # If the OpenStack and Hypervisor UUID are different (besides the
        # first bit), that means the OpenStack UUID needs to be fixed up
        # so that they match for NovaLink to handle, but since the migrate
        # doesn't do this until the very end, we need to account for it now
        power_specs = instance.get('power_specs', dict())
        vm_uuid = power_specs.get('vm_uuid').lower()
        if instance.uuid[1:] != vm_uuid[1:]:
            instance = copy.deepcopy(instance)
            instance.uuid = vm_uuid
        # We now want to populate the NVRAM when we are done with the
        # migration both to cover HMC-to-NovaLink issues and in case
        # some reason the event doesn't come in for the NVRAM update
        self.nvram_mgr.store(instance, immediate=True)

    @logcall
    def finish_onboard(self, context,
                       instance,
                       preparation,
                       network_info,
                       block_device_info):
        # temporary workaround to add system metadata for npiv imported VM
        super(PowerVMDriver, self).\
            finish_onboard(context,
                           instance,
                           preparation,
                           network_info,
                           block_device_info)
        # check the bdm to make sure it has npiv info
        # if so, extract the phy_to_virt_initiator from volume connector
        npiv_phy_to_virt_init = {}
        if block_device_info and block_device_info.get('block_device_mapping'):
            for mapping in block_device_info.get('block_device_mapping'):
                conn_info = mapping.get('connection_info')
                if conn_info and conn_info.get('volume_connector'):
                    vol_conn = conn_info.get('volume_connector')
                    if vol_conn.get('connection-type') == 'npiv':
                        if vol_conn.get('connector'):
                            initiator = vol_conn.get('connector').get(
                                'phy_to_virt_initiators')
                            if initiator:
                                for phy_fc in initiator:
                                    phy_fc_l = phy_fc.lower()
                                    virt_fcs = initiator.get(phy_fc)
                                    if virt_fcs:
                                        if phy_fc_l in npiv_phy_to_virt_init:
                                            npiv_phy_to_virt_init[phy_fc_l]\
                                                .extend(virt_fcs)
                                        else:
                                            npiv_phy_to_virt_init[phy_fc_l] = \
                                                copy.deepcopy(virt_fcs)
        # populate the portmap for setting the instance_system_metadata
        # based on phy_to_virt_initiator and fabric info from
        # npiv volume driver.
        connector = self.get_volume_connector(instance)
        fab_port_maps = {}
        LOG.debug("npiv_phy_to_virt_init=%s" % npiv_phy_to_virt_init)
        if connector and 'npiv' in connector:
            npiv_drv = connector.get('npiv').get('vol_adap')
            if npiv_drv is not None:

                # populate the fabric system metadata info.
                # We pass ignore_ports_per_fab so that the full list of
                # ports is returned and not a subset based on the
                # SCG ports_per_fabric_npiv property, which defaults to 1.
                scg_info = npiv_drv._get_scg_info(
                    ignore_ports_per_fab=True)
                scg_fab_port_map = scg_info['fab_port_map']
                db_vioses = scg_info['scg'].vioservers
                for vios in db_vioses:
                    for fcport in vios.fcports:
                        p_wwpn = fcport.wwpn.lower()
                        if p_wwpn in npiv_phy_to_virt_init:
                            if not fcport.fabric or "None" in fcport.fabric:
                                msg = (_("The following port is not added to "
                                         "the system metadata for the VM, "
                                         "since the port does not have "
                                         "connectivity to a fabric. "
                                         "NPIV instance: %(name)s, "
                                         "port: %(port)s") %
                                       dict(port=p_wwpn,
                                            name=instance['name']))
                                LOG.warn(msg)
                                continue
                            if fcport.fabric not in fab_port_maps:
                                fab_port_maps[fcport.fabric] = {}
                            if (not fcport.enabled or
                                (fcport.fc_protocol != 'any' and
                                 fcport.fc_protocol != 'npiv')):
                                msg = (_("The VM uses an NPIV connectivity "
                                         "mapping to physical port disallowed "
                                         "by PowerVC. NPIV instance: "
                                         "%(name)s, port: %(port)s") %
                                       dict(port=p_wwpn,
                                            name=instance['name']))
                                LOG.warn(msg)

                            virt_fcs = npiv_phy_to_virt_init.get(p_wwpn)
                            # get the unique vfc. Do not use SET
                            # here as we need to preserve the order
                            uniq_vfcs = []
                            for vfc in virt_fcs:
                                if not uniq_vfcs.count(vfc):
                                    uniq_vfcs.append(vfc)

                            fab_port_maps[fcport.fabric][p_wwpn] =\
                                uniq_vfcs
                            LOG.debug('fabric=%s, p_wwpn_l=%s, set of '
                                      'virtual initiators=%s' %
                                      (fcport.fabric, p_wwpn, virt_fcs))
                if not fab_port_maps:
                    LOG.warn(_("Onboarding NPIV instance %(name)s, but no "
                               "physical to virtual port maps found. The "
                               "storage connectivity group mapping data is "
                               "%(scg)s and the correlating set of mapping "
                               "data is: %(data)s.") %
                             {'name': instance['name'],
                              'scg': scg_fab_port_map,
                              'data': npiv_phy_to_virt_init})
                # set the system metadata
                for fabric in fab_port_maps:
                    port_map = fab_port_maps.get(fabric)
                    if not port_map:
                        LOG.debug("No maps for fabric %s. Continue." % fabric)
                        continue
                    port_map_reformated = []
                    for p_wwpn, v_wwpns in six.iteritems(port_map):
                        port_map_reformated.append(p_wwpn)
                        port_map_reformated.extend(v_wwpns)
                    if not port_map_reformated:
                        LOG.warn(_("Onboarding NPIV instance %(name)s, but no "
                                   "correlated maps found for fabric %(fab)s. "
                                   "The storage connectivity group mapping "
                                   "data is %(scg)s and the mapping "
                                   "data is: %(data)s.") %
                                 {'name': instance['name'],
                                  'scg': scg_fab_port_map, 'fab': fabric,
                                  'data': npiv_phy_to_virt_init})
                        # Can we fail the onboard here instead of continue?
                        continue
                    state_key = npiv_drv._sys_fabric_state_key(fabric)
                    map_key = npiv_drv._sys_meta_fabric_key(fabric)
                    instance.system_metadata[state_key] = 'inst_mapped'
                    instance.system_metadata[map_key] = \
                        ','.join(port_map_reformated)
                    LOG.info(_('During onboard of virtual machine %(name)s, '
                               'store NPIV mapping of %(key)s to %(val)s.') %
                             {'name': instance['name'], 'key': map_key,
                              'val': instance.system_metadata[map_key]})
            else:
                LOG.warn(_('Unable to obtain the NPIV volume adapter.'))
        # Defect 52457 - Collect the NVRAM for the intsance
        if self.nvram_mgr is not None:
            self.nvram_mgr.store(instance, immediate=True)
            LOG.debug(_('Storing NVRAM as part of onboard for instance: '
                        '%s.'),
                      instance.name, instance=instance)
        else:
            LOG.error(_('NVRAM manager is not configured. NVRAM not '
                        'collected during onboard for instance: %s.'),
                      instance.name, instance=instance)
        instance.save()

    def macs_for_instance(self, instance):
        """Generates set of valid MAC addresses for a Novalink instance."""
        # NOTE(vish): We would prefer to use 0xfe here to ensure that linux
        #             bridge mac addresses don't change, but it appears to
        #             conflict with libvirt, so we use the next highest octet
        #             that has the unicast and locally administered bits set
        #             properly: 0xfa.
        #             Discussion: https://bugs.launchpad.net/nova/+bug/921838
        # NOTE(safron): For POWER7 PowerVM, the MAC address can be directly
        #               set, so the list generated by this method will be
        #               used.  For POWER6 PowerVM, the MAC address cannot be
        #               set, so the Neutron ports will be updated with the
        #               correct MAC addresses after the LPAR has been created.
        #               The final 2 hex values represent a value between 32 and
        #               64 so we can assign it as the slot id on the system for
        #               both POWER6 and POWER7.
        #
        #               FA:xx:xx:xx:xx:[32-64]

        macs = set()
        mac_base = [0xfa,
                    random.randint(0x00, 0xff),
                    random.randint(0x00, 0xff),
                    random.randint(0x00, 0xff),
                    random.randint(0x00, 0xff),
                    random.randint(0x00, 0x00)]
        for n in range(32, 64):
            mac_base[5] = n
            macs.add(':'.join(map(lambda x: "%02x" % x, mac_base)))

        return macs

    #######################################################
    #        LPM related Overrides/Extensions             #
    #######################################################
    def check_can_live_migrate_destination(self, ctxt, instance_ref,
                                           src_compute_info, dst_compute_info,
                                           block_migration=False,
                                           disk_over_commit=False):
        """
        Validate the destination host is capable of live partition migration.

        :param ctxt: security context
        :param instance_ref: instance to be migrated
        :param src_compute_info: source host information
        :param dst_compute_info: destination host information
        :param block_migration: if true, prepare for block migration
        :param disk_over_commit: if true, allow disk over commit
        :returns dest_check_data: dictionary containing destination data
        """
        LOG.debug("Entering check_can_live_migrate_destination/driver.py: "
                  "ctxt=%s, instance_ref=%s, src_compute_info=%s, "
                  "dst_compute_info=%s, block_migration=%s, "
                  "disk_over_commit=%s" %
                  (ctxt, instance_ref,
                   src_compute_info,
                   dst_compute_info,
                   block_migration,
                   disk_over_commit))
        # Block LPM if VM's SCG is using'ibm_ssp' connectivity.
        self._forbid_ssp(instance_ref)

        # Block LPM if dest host doesn't have needed network config
        self._verify_dest_nets(ctxt, instance_ref, src_compute_info,
                               dst_compute_info)

        # Adding attribute memory region size to source host info
        src_compute_info['stats']['memory_region_size'] = \
            src_compute_info['stats']['lmb_size']
        mig = lpm.PowerVCLiveMigrationDest(self, instance_ref)
        self.live_migrations[instance_ref.uuid] = mig
        dest_data = mig.check_destination(ctxt, src_compute_info,
                                          dst_compute_info)
        # dest_data = super(PowerVMDriver, self). \
        #    check_can_live_migrate_destination(
        #        ctxt, instance_ref, src_compute_info, dst_compute_info,
        #        block_migration=block_migration,
        #        disk_over_commit=disk_over_commit)
        # Overriding the dest_ip to the PVM host instead of proxy.
        # Though it is not required in production env, having this in
        # place will be an advantage.
        if CONF.pvm.ip:
            dest_data.dest_ip = CONF.pvm.ip
        # we should only pass in the CONF.host since this attribute
        # is used for migration job
        if not hasattr(dest_data, 'dest_sys_name'):
            # CONF.host should always be set
            dest_data.dest_sys_name = CONF.host

        # If migrating from HMC management console, we need to
        # add the necessary migration data from HMC to PVM.
        dest_data.dest_lmb = int(dst_compute_info['stats']['lmb_size'])
        dst_managed_by = migrate_utils.get_dest_managed_by(dst_compute_info)
        dest_data.dest_managed_by = dst_managed_by
        dest_data.managed_by = dst_managed_by
        # Check if shared processor pool is used and if that pool exists
        # on the destination host.
        power_specs = instance_ref.get('power_specs', {})
        is_ded = power_specs.get('dedicated_proc')
        if is_ded is not None and is_ded.lower() == 'false':
            shared_proc_pool_name = power_specs.get('shared_proc_pool_name')
            shared_proc_pool_id = power_specs.get('shared_proc_pool_id')
            if shared_proc_pool_id is not None:
                if shared_proc_pool_id == 0:
                    dest_spp_id = 0
                else:
                    dest_spp_id = pvm_host.\
                        get_spp_id(self.adapter, self.host_uuid,
                                   shared_proc_pool_name)
                    if dest_spp_id is None:
                        error = (_("Cannot migrate %(inst)s because the target"
                                   " host does not have the shared processor "
                                   "pool named %(pool)s.") %
                                 {'inst': instance_ref.name,
                                  'pool': shared_proc_pool_name})
                        raise nova_exception.\
                            MigrationPreCheckError(reason=error)
                dest_data.dest_spp_id = dest_spp_id

        # Not logging as nova-powervm driver logging
        return dest_data

    def check_can_live_migrate_source(self, context, instance,
                                      dest_check_data, block_device_info=None):
        """Check if it is possible to execute live migration.

        This checks if the live migration can succeed, based on the
        results from check_can_live_migrate_destination.

        :param context: security context
        :param instance: nova.db.sqlalchemy.models.Instance
        :param dest_check_data: result of check_can_live_migrate_destination
        :param block_device_info: result of _get_instance_block_device_info
        :returns: a dict containing migration info (hypervisor-dependent)
        """
        LOG.info(_("Checking live migration capability on source host."),
                 instance=instance)
        # ToDo HMC should pass the host_mig_data as well
        # since this value is used to get the LPM capability.
        if not hasattr(dest_check_data, 'host_mig_data'):
            dest_check_data.host_mig_data = {}

        mig = lpm.PowerVCLiveMigrationSrc(self, instance, dest_check_data)
        self.live_migrations[instance.uuid] = mig

        # Get a volume driver for each volume
        vol_drvs = self._build_vol_drivers(context, instance,
                                           block_device_info)

        src_data = mig.check_source(context, block_device_info, vol_drvs)
        host_name = encodeutils.safe_decode(CONF.host_display_name)
        src_data.src_sys_name = host_name
        src_data.src_managed_by = NOVALINK_PLATFORM
        return src_data

    def _verify_dest_nets(self, ctxt, instance_ref, src_compute_info,
                          dst_compute_info):
        """Helper method for forbidding LPM if destination's networks are not
           compatible with source networks.

          :param ctxt: security context
          :param instance_ref: instance to be migrated
          :param src_compute_info: source host information
          :param dst_compute_info: destination host information
           Raise MigrationPreCheckError exception if dest network problems are
           found.
        """
        try:
            _net_api = pvc_net_api.PowerVMAPI()
            hosts = _net_api.get_viable_hosts(
                context=ctxt,
                instance_uuid=instance_ref.uuid,
                requested_networks=None,
                source_host=src_compute_info['host'])
            if dst_compute_info['host'] not in hosts:
                raise Exception()
        except Exception:
            msg = (_("Cannot migrate instance %s because source and target "
                     "hosts have incompatible network configurations.  Verify "
                     "vswitches match and needed VLANs are bridged.")
                   % instance_ref.name)
            raise exception.MigrationPreCheckError(reason=msg)

    def handle_evacuated_instances(self, context, db_instances, drv_instances):
        """Handle the case when an instance is migrated out of band
           on Novalink
        """
        # If we were asked not to handle the out-of-band changes, just return
        if not CONF.ibmpowervm_handle_out_of_band_migration:
            return
        host_name = CONF.host

        # Create a list of instance uuids from the DB
        managed_instance_uuids = [vm.get_pvm_uuid(db_instance)
                                  for db_instance in db_instances]

        # Iterate through LPARs on the host. Handle those LPAR
        # not in DB
        lpars = vm.get_lpars(self.adapter)
        all_instances = None
        for lpar in lpars:
            try:
                # Retrieve LPAR UUID. Note that PowerVM UUID
                # (not PowerVC UUID) is returned
                uuid = lpar.uuid
                # if the uuid is not part of the instance managed by this
                # host, it's likely an out of band operation has been done
                # to this VM.
                if uuid not in managed_instance_uuids:
                    # populate the non deleted instance in the DB if it's not
                    # populated already
                    if all_instances is None:
                        filters = {'deleted': False}
                        all_instances = inst_objs.InstanceList.\
                            get_by_filters(context,
                                           filters)
                    # Go through the list to find the one with matching UUID
                    for instance in all_instances:
                        # extract the host information in the db
                        host_in_db = instance.host
                        uuid_in_db = vm.get_pvm_uuid(instance)
                        # if uuid matched up but host info doesn't,
                        # figure out what operation is done
                        if (uuid.lower() == uuid_in_db.lower() and
                            host_in_db != host_name and host_in_db
                                is not None):
                            node = self.get_available_nodes()[0]
                            vm_info = self.get_info(instance)
                            # update additional info to determine out of band
                            # operation
                            self._update_vm_info_with_oob(vm_info)

                            migrate_utils.verify_out_of_band_evacuation(
                                context, instance, uuid, uuid_in_db,
                                host_in_db, host_name, vm_info, node,
                                src_host_state=None,
                                src_host_uuid=None,
                                k2oper=None)

            except Exception as e:
                LOG.exception(e)
                LOG.error(_('handle_evacuated_instances encounters '
                            'error: %s') % six.text_type(e))

    def is_instance_on_host(self, instance_uuid, instance_name):
        """
        Check if an instance exists on this host
        """
        # all other exception will bubble up
        try:
            # If we're able to get the property, then it exists.
            vm.get_vm_id(self.adapter, instance_uuid)
            return True
        except exception.InstanceNotFound:
            return False

    def _update_vm_info_with_oob(self, vm_info):
        """Update get_info return with addtional attributes to help
        PowerVC to determine VM is in process of migration,
        remote restart, etc
        :param vm_info: return object from get_info.
        """
        # extract the PartitionState
        lpar_state = vm_info._get_property('PartitionState').lower()
        vm_info.instance_migrating = lpar_state in MIGRATING_STATES

    def __convert_data_for_host_switch(self, context, dbinsts):
        """Helper method to convert data when switching from HMC to PVM"""
        try:
            LOG.info(_("Converting data for switch from HMC to PVM..."))
            # Get the Fabrics for each of the WWPN's for the FC Ports
            fcfact = storage_dom.FcPortFactory.get_factory()
            fcports = fcfact.find_all_fcports_by_host(context, CONF.host)
            fabmap = dict([(p.wwpn.lower(), p.fabric) for p in fcports])
            # Loop through the VM's converting the BDM's and System MetaData
            for dbinst in dbinsts:
                bdms, sysmeta = blkdev_obj.BlockDeviceMappingList.\
                    get_by_instance_uuid(context, dbinst.uuid), dict()
                # Loop through the BDM's converting the necessary data
                for bdm in bdms:
                    if not bdm.connection_info:
                        continue
                    # Parse the Connection Info needed to use
                    cinfo = jsonutils.loads(bdm.connection_info)
                    ctype = cinfo.get('data', {}).get('connection-type')
                    vconn = cinfo.get('data', {}).get('volume_connector')
                    conn = vconn.get('connector', {}) if vconn else {}
                    # If the Connector isn't in the Data, check up a level
                    if not conn:
                        vconn = cinfo.get('volume_connector')
                        conn = vconn.get('connector', {}) if vconn else {}
                    # If the Connector isn't wrappered, check stand-alon
                    if not conn:
                        conn = cinfo.get('connector') or {}
                    # Check for the alternative name for the connection type
                    if not ctype:
                        ctype = cinfo.get('data', {}).get('connector_type')
                    # Also check in the volume_connector if didnt' find in data
                    if not ctype:
                        ctype = vconn.get('connection-type') if vconn else None
                    # If this is NPIV, then populate the BDM Volume Connector
                    if ctype == 'npiv' and conn:
                        if not conn.get('phy_to_virt_initiators'):
                            continue
                        # Loop through the Physical to Virtual Mappings
                        for p, v in conn['phy_to_virt_initiators'].iteritems():
                            fab, p = fabmap.get(p.lower()), p.lower()
                            # Skip if for some reason the Fabric isn't set
                            if fab:
                                wwpns = '%s,%s' % (p, ','.join(v).upper())
                                sysmeta['fabric_state_' + fab] = 'inst_mapped'
                                sysmeta['npiv_adpt_wwpns_' + fab] = wwpns
                # If we got some System MetaData to update, do that now
                if sysmeta:
                    sysmeta = dict(dbinst.system_metadata, **sysmeta)
                    dbinst.system_metadata = sysmeta
                    dbinst.save()
            # Since the conversion worked, delete the property from the file
            data_utils.del_openstack_config_value(
                '/etc/nova/nova.conf', 'powervm_switch_type')
        except Exception as exc:
            LOG.warn(_("Error converting data for switch from HMC: %s") % exc)
            LOG.exception(exc)

    def _update_slot_info(self, context, dbinsts):
        """Check if instance has slot information stored in swift.
        If not it might be
        1. Just upgraded from a previous release or
        2. Host might be reconfigured from hmc
        So update the slot information of vfc/vea/scsi of this instance
        in swift.
        :param context: context.
        """
        container = CONF.powervm.swift_container
        vios_wraps, swift_obj_names = None, list()
        LOG.info(_("Updating slot information..."))
        try:
            # There can only be Object Names for the Container if it exists
            if container in self.store_api._get_container_names():
                swift_obj_names = self.store_api._get_object_names(container)
            for instance in dbinsts:
                inst_key = "%s_slot_map" % instance.uuid
                if instance.uuid not in swift_obj_names:
                    # Store the NVRAM data
                    if self.nvram_mgr is not None:
                        LOG.debug(_('NVRAM not found for instance: '
                                    '%s. Storing NVRAM data.'),
                                  instance.name, instance=instance)
                        self.nvram_mgr.store(instance, immediate=True)

                if inst_key in swift_obj_names:
                    continue
                if not vios_wraps:
                    stg_ftask = pvm_vios_util.build_tx_feed_task(
                        self.adapter, self.host_uuid)
                    vios_wraps = stg_ftask.feed
                self._update_instance_slot_info(context, instance, vios_wraps)
            LOG.info(_("Updating slot information is done."))
        except Exception as exc:
            LOG.warn(_("Error updating slot information: %s") % exc)
            LOG.exception(exc)

    def _update_instance_slot_info(self, context, instance, vios_wraps=None):
        try:
            inst_slot_map = slot.SwiftSlotManager(
                self.store_api, instance=instance)
        except Exception as ex:
            LOG.exception(_("Unable to retrieve slot information "
                            "during update_instance_slot_info %s "
                            % six.text_type(ex)))
            return
        if not vios_wraps:
            tg_ftask = pvm_vios_util.build_tx_feed_task(
                self.adapter, self.host_uuid)
            vios_wraps = tg_ftask.feed

        lpar_uuid = instance.get('power_specs', {}).get('vm_uuid')
        cna_list = nw_utils.list_cnas(
            self.adapter, self.host_uuid, lpar_uuid=lpar_uuid)
        pvc_lpar.update_instance_slot_map(
            context, instance, vios_wraps, cna_list, inst_slot_map)
