# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from functools import partial

from oslo_log import log as logging
from oslo_utils import excutils
from pypowervm.utils import transaction
from pypowervm import const as pvmc
from pypowervm import exceptions as pvm_exc
from pypowervm.wrappers import enterprise_pool as ep_wrap

from powervc_nova import _
from powervc_nova.virt.ibmpowervm.common import enterprise_pool as common_ep

LOG = logging.getLogger(__name__)


def sync_pool_for_host(context, adapter, hostname):
    """Syncs pool info corresponding to a specific host managed by NovaLink.

    Since this code is ran from each individual compute driver process, it will
    only sync membership for this specific host (that corresponds to hostname).
    It will also sync the overall pool properties, such as the available mobile
    cores, total mobile cores, etc., that aren't tied to a specific member.

    :param context: The security context for the query
    :param adapter: pypowervm adapter object
    :param hostname: The hostname to sync information for
    """
    # Fetch all the pools from pypowervm
    pools = []
    try:
        pools = ep_wrap.Pool.get(adapter)
    except pvm_exc.HttpError as e:
        with excutils.save_and_reraise_exception() as exc_ctxt:
            if e.response.status == pvmc.HTTPStatus.INTERNAL_ERROR:
                if e.response.body and 'REST9301 ' in e.response.body:
                    # If "REST9301 " is in the response's body, then we know
                    # that all of the HMCs attached to the NovaLink host are
                    # not compatible with NovaLink hosts and Power Enterprise
                    # Pools. This is likely to occur if the HMC is running 840,
                    # which does not support NovaLink CoD / enterprise pools at
                    # all, but the NovaLink version does support enterprise
                    # pools. Log an info message and do not reraise the
                    # exception.
                    LOG.info(_('The HMCs attached to NovaLink host %s either '
                               'do not support Power Enterprise Pools with '
                               'NovaLink hosts or they are not compatible '
                               'with the version of NovaLink on the host.') %
                             hostname)
                    exc_ctxt.reraise = False

    pool_wrapper = None
    pool_member_uuid = None

    if pools:
        # Determine if this host belongs to any of the pools
        for pool in pools:
            pool_member_uuid = _find_pool_member_uuid(adapter, hostname,
                                                      pool.uuid)
            if pool_member_uuid:
                # We found a pool member corresponding to the host, so break
                pool_wrapper = pool
                break

    pool_dict = {}
    if pool_wrapper:
        # PowerVC expects the HMC MTMS to be in a specific format
        master_console_mtms = pool_wrapper.master_console_mtms
        master_console_mtms_str = (
            '%(machine_type)s-%(model)s %(serial_num)s' %
            {'machine_type': master_console_mtms.machine_type,
             'model': master_console_mtms.model,
             'serial_num': master_console_mtms.serial})
        pool_dict = {
            'uuid': pool_wrapper.uuid,
            'name': pool_wrapper.name,
            'master_hmc_mtmsn': master_console_mtms_str,
            'available_mobile_cores': pool_wrapper.avail_mobile_procs,
            'total_mobile_cores': pool_wrapper.total_mobile_procs,
            'compliance_state': pool_wrapper.compliance_state}

    common_ep.sync_pool_for_host(context, pool_dict, pool_member_uuid,
                                 hostname)


def _find_pool_member_uuid(adapter, hostname, pool_uuid):
    """Returns the pool member UUID of the host if it is a member of the pool.

    :param adapter: pypowervm adapter object
    :param hostname: The hostname of the host of interest
    :param pool_uuid: The UUID of the pool to search for a relevant member.
    :returns: The pool member UUID of the host if it is a member of the pool
              corresponding to pool_uuid, None otherwise.
    """
    # Fetch all the pool members from pypowervm
    pool_members = ep_wrap.PoolMember.getter(adapter,
                                             parent_class=ep_wrap.Pool,
                                             parent_uuid=pool_uuid).get()

    if not pool_members:
        return None

    for member in pool_members:
        mtms = common_ep.convert_mtms_to_hostname(member.sys_mtms.machine_type,
                                                  member.sys_mtms.model,
                                                  member.sys_mtms.serial)
        if mtms == hostname:
            # Found a member that corresponds to the host
            return member.uuid

    # No member was found corresponding to the host
    return None


def get_member(adapter, pool_uuid, member_uuid):
    """Get pool member with specified pool UUID and member UUID.

    :param adapter: pypowervm adapter object
    :param pool_uuid: UUID of the pool that member UUID is in
    :param member_uuid: UUID of the member of interest
    :returns: pypowervm PoolMember wrapper object
    """
    return ep_wrap.PoolMember.getter(adapter, entry_uuid=member_uuid,
                                     parent_class=ep_wrap.Pool,
                                     parent_uuid=pool_uuid).get()


def update_mobile_cores(adapter, pool_uuid, member_uuid, desired_mobile_cores):
    """Update the pool member's num of mobile cores via pypowervm.

    :param adapter: pypowervm adapter object
    :param pool_uuid: UUID of the enterprise pool
    :param member_uuid: UUID of the enterprise pool member
    :param desired_mobile_cores: Int number of desired value for mobile cores
    :returns: Tuple of (unreturned mobile cores, mobile cores, inactive cores)
              corresponding to the updated pool member.
    """
    member_getter = ep_wrap.PoolMember.getter(adapter, entry_uuid=member_uuid,
                                              parent_class=ep_wrap.Pool,
                                              parent_uuid=pool_uuid)

    @transaction.entry_transaction
    def run_update(pool_member):
        current_value = pool_member.mobile_procs
        if current_value != desired_mobile_cores:
            pool_member.mobile_procs = desired_mobile_cores
            pool_member = pool_member.update()
        return pool_member

    updated_member = run_update(member_getter)

    return (updated_member.unret_mobile_procs,
            updated_member.mobile_procs,
            updated_member.inactive_procs)


def set_mobile_cores(context, desired_mobile_cores, hostname, adapter):
    """Sets the mobile cores that a host is using to the desired value.

    :param context: The security context for the query
    :param desired_mobile_cores: the number of mobile cores that the host
                                 should use
    :param adapter: pypowervm adapter object
    :returns: Tuple of (result, reason). For example, if the mobile cores
              were successfully updated, then a tuple of (True,) will be
              returned. Otherwise, it would return a tuple of
              (False, 'Reason for failure').
    """
    update_mobile_cores_partial = partial(update_mobile_cores, adapter=adapter)

    return common_ep.set_mobile_cores(context, desired_mobile_cores, hostname,
                                      update_mobile_cores_partial)
