# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from ceilometer_powervm.compute.virt.powervm import inspector
from oslo_log import log as logging
from pypowervm.tasks.monitor import util as pvm_mon_util
from pypowervm.utils import uuid as pvm_uuid
from pypowervm.wrappers import managed_system as pvm_ms

LOG = logging.getLogger(__name__)


class LparMetrics(object):

    INSTANCE = None

    def __init__(self, adpt):
        self.adpt = adpt
        # Get the host system UUID
        host_uuid = self._get_host_uuid()
        # Ensure that metrics gathering is running for the host.
        pvm_mon_util.ensure_ltm_monitors(self.adpt, host_uuid)
        # Get the VM Metric Utility
        self.vm_metrics = pvm_mon_util.LparMetricCache(self.adpt, host_uuid)
        try:
            # For defect 45290. Gobble exception for registration to proceed.
            self._cpu_util_inspector = inspector.PowerVMInspector()
        except Exception:
            self._cpu_util_inspector = None
            LOG.error(_("Unable to initialize PowerVMInspector."),
                      exc_info=True)

    @staticmethod
    def get_instance(adpt):
        """Returns the singleton to this class. This is to avoid initializing
           the class every time.
        """
        if LparMetrics.INSTANCE is None:
            LparMetrics.INSTANCE = LparMetrics(adpt)
        return LparMetrics.INSTANCE

    @staticmethod
    def _puuid(instance):
        """Derives the PowerVM UUID for an instance.

        :param instance: The OpenStack instance object.
        :return: The PowerVM API's UUID for the instance.
        """
        return pvm_uuid.convert_uuid_to_pvm(instance.uuid).upper()

    def _get_host_uuid(self):
        """Returns the Host system's UUID for pypowervm.

        The pypowervm API needs a UUID of the server that it is managing.  This
        method returns the UUID of that host.

        :param adpt: The pypowervm adapter.
        :return: The UUID of the host system.
        """
        hosts = pvm_ms.System.get(self.adpt)
        if len(hosts) != 1:
            raise Exception(_("Expected exactly one host; found %d."),
                            len(hosts))
        LOG.debug("Host UUID: %s" % hosts[0].uuid)
        return hosts[0].uuid

    def get_mem_util_for_lpar(self, instance):
        """Get memory utilization of LPAR

        :param instance: the target instance
        :return: the memory util of lpar
        """
        uuid = self._puuid(instance)
        cur_date, cur_metric = self.vm_metrics.get_latest_metric(uuid)
        # If the current metric is none, then the instance can not be found in
        # the sample set.  An error should be raised.
        if cur_metric is None:
            raise Exception(
                _('VM %s not found in PowerVM Metrics Sample.') %
                instance.name)
        # Get the metrics
        lpar_memory_metrics = cur_metric.memory
        return self._validate_memory_metrics(lpar_memory_metrics)

    def _validate_memory_metrics(self, lpar_memory_metrics):
        """Validates the different memory metrics. No metric can be negative.
        If a metric is found to be negative, then None is returned for that
        metric. Memory utilization percent should be between 0,100.
        """
        lpar_mem_pct_free = getattr(lpar_memory_metrics, 'pct_real_mem_free')
        vm_pg_in_rate = getattr(lpar_memory_metrics, 'vm_pg_in_rate')
        vm_pg_out_rate = getattr(lpar_memory_metrics, 'vm_pg_out_rate')
        vm_pg_swap_in_rate = getattr(lpar_memory_metrics, 'vm_pg_swap_in_rate')
        vm_pg_swap_out_rate = getattr(lpar_memory_metrics,
                                      'vm_pg_swap_out_rate')
        lpar_mem_util = None
        # Utilization is 100 - FreeMemory
        if 0 <= lpar_mem_pct_free <= 100:
            lpar_mem_util = (100 - lpar_mem_pct_free)
        vm_pg_in_rate = self._validate_non_negative(vm_pg_in_rate)
        vm_pg_out_rate = self._validate_non_negative(vm_pg_out_rate)
        vm_pg_swap_in_rate = self._validate_non_negative(vm_pg_swap_in_rate)
        vm_pg_swap_out_rate = self._validate_non_negative(vm_pg_swap_out_rate)
        lpar_mem_metrics_dict = {'mem_util': lpar_mem_util,
                                 'pg_in_rate': vm_pg_in_rate,
                                 'pg_out_rate': vm_pg_out_rate,
                                 'swap_in_rate': vm_pg_swap_in_rate,
                                 'swap_out_rate': vm_pg_swap_out_rate}
        return lpar_mem_metrics_dict

    def _validate_non_negative(self, val):
        return val if val >= 0 else None

    def get_cpu_util_for_lpar(self, instance):
        """Get the CPU utilization of LPAR from PowerVMInspector"""
        if self._cpu_util_inspector is None:
            return None
        # Hack to pass instance id & name to NovaLink ceilometer inspector
        instance_obj = PVMInst()
        instance_obj.name = instance["name"]
        instance_obj.id = instance["uuid"]
        cpu_util = self._cpu_util_inspector.inspect_cpu_util(
            instance_obj, duration=None)
        return cpu_util


class PVMInst(object):
    def __init__(self):
        self.id = None
        self.name = None
