# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from nova.context import get_admin_context
from nova_powervm.virt.powervm.volume import npiv as parent_npiv
from oslo_log import log as logging
from oslo_config import cfg
from oslo_serialization import jsonutils
from pypowervm.tasks import vfc_mapper as pvm_vfcm
from nova.compute import task_states
from string import ascii_uppercase

from powervc_nova.storage import exception as stgex
from powervc_nova.storage.storage_mapping_api import StorageMapping


LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class PowerVCNpivVolumeAdapter(parent_npiv.NPIVVolumeAdapter):

    def __init__(self, adapter, host_uuid, instance, connection_info,
                 stg_ftsk=None):
        super(PowerVCNpivVolumeAdapter, self).__init__(adapter, host_uuid,
                                                       instance,
                                                       connection_info,
                                                       stg_ftsk=stg_ftsk)
        self.scg_info = None

    def wwpns(self):
        """ PowerVC override to pull in the full set of initiator wwpns """
        # First call the super
        wwpns = set(super(PowerVCNpivVolumeAdapter, self).wwpns())
        LOG.debug("Initiator wwpns from parent adapter: %s" % wwpns)
        if not wwpns:
            return []
        # We need to use both the primary and standby initiators. It is
        # helpful because it allows everything to be mapped up-front for
        # NPIV (rather than in separate initialize_connections), and it
        # is needed for at least VMAX on a Liberty base.
        fab_states = {}
        for fabric in self._fabric_names():
            fab_states[fabric] = self._get_fabric_state(fabric)  # for logging
            maps = self._get_fabric_meta(fabric)
            if not maps:
                LOG.debug("Unexpected no maps for fabric %s" % fabric)
                continue
            for __, v_wwpns in maps:
                wwpns.update(v_wwpns.split())
        exp_wwpns = list(wwpns)
        exp_wwpns.sort()
        LOG.info(_("[wwpns] Instance '%(inst)s' in task_state '%(t_state)s' "
                   "and power_state '%(p_state)d', the fabric states are "
                   "%(fab_states)s' and the expanded wwpns are: %(wwpns)s.") %
                 {'inst': self.instance.name,
                  't_state': self.instance.task_state,
                  'p_state': self.instance.power_state,
                  'fab_states': fab_states, 'wwpns': exp_wwpns})
        return exp_wwpns

    def _ports_per_fabric(self):
        """Returns the number of virtual ports that should be used per fabric
           for all VIOS to be connected. PowerVC overrides this method to
           use SCG determined info rather than a static config file value.
        """
        LOG.debug("Instance %s: task_state=%s, power_state=%d" %
                  (self.instance.name, self.instance.task_state,
                   self.instance.power_state))
        scg_fab_map = self._get_scg_info()['fab_port_map']
        num_ports = 0
        # All fabrics should connect to same number of ports, but just in
        # case, pick the fabric with the minimum number of ports and
        # return that number for ports_per_fabric.
        for fab in scg_fab_map:
            if num_ports == 0 or len(scg_fab_map[fab]) < num_ports:
                num_ports = len(scg_fab_map[fab])
                LOG.debug("Per fabric %s, setting num_ports=%d" %
                          (fab, num_ports))

        LOG.debug("Return %d ports across VIOSs." % num_ports)
        return num_ports

    def _fabric_names(self):
        """Returns a list of the fabric names."""
        LOG.debug("Instance %s: task_state=%s, power_state=%d" %
                  (self.instance.name, self.instance.task_state,
                   self.instance.power_state))
        loc = None
        if self.instance.task_state in [task_states.DELETING,
                                        task_states.MIGRATING,
                                        task_states.RESIZE_MIGRATING,
                                        task_states.RESIZE_FINISH]:
            # This code path will not go through SCG get_connectivity
            # processing, and so should not raise SCG specific exceptions
            # when we are not adding a disk.
            fabrics = []
            for fab in ascii_uppercase:
                if self._get_fabric_state(fab) in [parent_npiv.FS_INST_MAPPED,
                                                   parent_npiv.FS_MIGRATING]:
                    fabrics.append(fab)
            loc = 'Instance'
        else:
            # For remote-restart, it is on a different host, so assuming we
            # do need to compute the connectivity information through SCG(?)
            fabrics = self._get_scg_info()['fab_port_map'].keys()
            loc = 'SCG'

        LOG.info(_("Fabrics determined from %(loc)s: %(fabs)s") %
                 {'loc': loc, 'fabs': fabrics})
        return fabrics

    def _fabric_ports(self, fabric_name):
        """Returns a list of WWPNs for the fabric's physical ports."""
        fab_state = self._get_fabric_state(fabric_name)
        if (fab_state == parent_npiv.FS_INST_MAPPED and
                not self._is_migration_wwpn(fab_state)):
            fmeta = self._get_fabric_meta(fabric_name)
            ports = [m[0] for m in fmeta]
            loc = 'metadata'
            LOG.debug("Getting fabric port data %s from inst_mapped store: %s"
                      % (ports, fmeta))
        else:
            LOG.debug("Fabric_state=%s. Get fabric ports from conn info..." %
                      fab_state)
            scg_info = self._get_scg_info()
            loc = "SCG %s" % scg_info['scg'].display_name
            if fabric_name in scg_info['fab_port_map']:
                plist = scg_info['fab_port_map'][fabric_name]
                ports = [p[0] for p in plist]
            else:
                LOG.debug("Fabric %s not in available list." % fabric_name)
                ports = []
        LOG.info(_("[fabric_ports] Instance '%(inst)s' in task_state "
                   "'%(t_state)s' and power_state '%(p_state)d', the fabric "
                   "'%(fab)s' with fabric state '%(fab_state)s' reports ports "
                   "from '%(loc)s': %(ports)s") %
                 {'inst': self.instance.name,
                  't_state': self.instance.task_state,
                  'p_state': self.instance.power_state, 'fab': fabric_name,
                  'fab_state': fab_state, 'loc': loc, 'ports': ports})
        return ports

    def _get_scg_info(self, ignore_ports_per_fab=False):
        if self.scg_info:
            LOG.info(_("Fabric to physical port map already computed."))
            return self.scg_info
        LOG.debug("Obtaining SCG info for instance: %s" % self.instance.name)
        scg_map = StorageMapping()

        # Ignore SCG health in the delete & resize flows
        skip_health_check = True if self.instance.task_state in\
            [task_states.DELETING, task_states.RESIZE_MIGRATING,
             task_states.RESIZE_FINISH] else False
        fab_port_map = {}
        fab_phys_map = {}  # for logging only
        if not skip_health_check:
            # The following call will raise connectivity errors if encountered.
            scg_info = scg_map.get_connectivity_info(get_admin_context(),
                                                     self.instance['uuid'],
                                                     CONF.host,
                                                     self.instance,
                                                     live_data=False)
            # Get the needed ports-per-fabric-per-vios value....we do not call
            # the private helper to do this since the helper calls this method.
            scg_ports_per = scg_info['scg'].ports_per_fabric_npiv
            for v in scg_info.get('vios_list'):
                if 'ports' in v.keys():
                    for fab in v['npiv_fab_order']:
                        if fab not in v['ports']:
                            LOG.debug("Skip %s for VIOS %s." %
                                      (fab, v['name']))
                            continue
                        if fab not in fab_port_map:
                            fab_port_map[fab] = []
                            fab_phys_map[fab] = []
                        ports_per_fab = scg_ports_per
                        for p in v['ports'][fab]:
                            fab_port_map[fab].append(
                                (p['wwpn'].upper(), pvm_vfcm._FUSED_ANY_WWPN))
                            fab_phys_map[fab].append(p['wwpn'].lower())
                            ports_per_fab -= 1
                            LOG.debug("VIOS=%s, Fabric=%s, ports-left-to-"
                                      "choose= %s after adding port %s." %
                                      (v['name'], fab,
                                       '*' if ignore_ports_per_fab else
                                       str(ports_per_fab), p['wwpn'].upper()))
                            if ports_per_fab == 0 and not ignore_ports_per_fab:
                                break
                        if ports_per_fab > 0 and not ignore_ports_per_fab:
                            # in general this should not happen, because the
                            # connectivity info processing checks for this
                            err = (_("Not enough ports on vios %(vios)s and "
                                     "fabric %(fab)s to satisfy the ports per"
                                     " fabric requirement: %(num)d.")
                                   % {"vios": v['name'], 'fab': fab,
                                      'num': scg_ports_per})
                            raise stgex.IBMPowerVCStorageError(error=err)
                        if (scg_info['scg'].fabric_set_req == 'exactly_one' and
                                not ignore_ports_per_fab):
                            LOG.debug("exactly_one semantic for fabrics - "
                                      "leaving loop after handling fabric "
                                      "%s." %
                                      fab)
                            break
            if not fab_port_map:
                err = _("The storage connectivity group '%s' does not report "
                        "any allowable fabrics for NPIV connectivity.") %\
                    scg_info['scg_name']
                raise stgex.IBMPowerVCStorageError(error=err)
        else:
            # In case of delete flow ignore the SCG health and return all the
            # pwwns in a fabric
            scg_info = scg_map.determine_scg(
                get_admin_context(), instance=self.instance,
                host=CONF.host, raise_err=True)
            scg = scg_info['scg']
            db_vioses = scg.vioservers
            for vios in db_vioses:
                for fcport in vios.fcports:
                    p_wwpn = fcport.wwpn.upper()
                    fab = fcport.fabric
                    if fab:
                        if fab not in fab_port_map:
                            fab_port_map[fab] = []
                            fab_phys_map[fab] = []
                        fab_port_map[fab].append(
                            (p_wwpn, pvm_vfcm._FUSED_ANY_WWPN))
                        fab_phys_map[fab].append(p_wwpn.lower())

        scg_info['fab_port_map'] = fab_port_map
        self.scg_info = scg_info
        LOG.info(_("Physical port selection by fabric: %(fab_map)s") %
                 {'fab_map': fab_phys_map})
        return scg_info

    def pre_live_migration_on_source(self, mig_data):
        """Override community method to pass in the vios id along with slot
        number for novalink to hmc migration
        """
        #TODO: check if the vios to client slot mapping is needed in the
        #community novalink code as well so that this can be contributed
        #back .RTC task # 49618
        fabrics = self._fabric_names()
        vios_wraps = self.stg_ftsk.feed

        for fabric in fabrics:
            npiv_port_maps = self._get_fabric_meta(fabric)
            if not npiv_port_maps:
                continue
            vios_slot_lst = []
            client_slots = []
            for port_map in npiv_port_maps:
                vios_w, vfc_map = pvm_vfcm.find_vios_for_vfc_wwpns(
                    vios_wraps, port_map[1].split())
                client_slot_num = vfc_map.client_adapter.lpar_slot_num
                client_slots.append(client_slot_num)
                phy_wwpn = port_map[0]
                if client_slot_num and phy_wwpn:
                    vios_pwwpn_slot = str(vios_w.id) + '_' + phy_wwpn + '_' +\
                        str(client_slot_num)
                    vios_slot_lst.append(vios_pwwpn_slot)

            # Set the client slots into the fabric data to pass to the
            # destination.
            mig_data['src_npiv_fabric_slots_%s' % fabric] = (
                jsonutils.dumps(client_slots))
            mig_data['src_npiv_fabric_vios_slots_%s' % fabric] = (
                jsonutils.dumps(vios_slot_lst))

    def host_name(self):
        """Derives the host name that should be used for the storage device.

        :return: The host name.
        """
        return self.instance.name
