# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from nova.i18n import _
from oslo_log import log as logging
from oslo_config import cfg
from nova.context import get_admin_context
from nova.compute import task_states
from nova_powervm.virt.powervm.volume import vscsi
from pypowervm.tasks import scsi_mapper as tsk_map
from pypowervm.wrappers import storage as pvm_stor
from powervc_nova.storage.storage_mapping_api import StorageMapping
from powervc_nova.virt.ibmpowervm.common import exception
from powervc_nova.storage.topo_common import StorageTopologyCommon
from powervc_nova.objects.compute import dom as compute_dom
from powervc_nova.objects.storage.dom import FcPortFactory

CONF = cfg.CONF
LOG = logging.getLogger(__name__)
CONNECTOR_HOST_LEN = 55


class PowerVCVscsiVolumeAdapter(vscsi.VscsiVolumeAdapter):

    """The vSCSI implementation of the Volume Adapter.

    vSCSI is the internal mechanism to link a given hdisk on the Virtual
    I/O Server to a Virtual Machine.  This volume driver will take the
    information from the driver and link it to a given virtual machine.
    """

    def __init__(self, adapter, host_uuid, instance, connection_info,
                 stg_ftsk=None):
        super(PowerVCVscsiVolumeAdapter, self).__init__(adapter, host_uuid,
                                                        instance,
                                                        connection_info,
                                                        stg_ftsk=stg_ftsk)

    def _validate_vios_on_connection(self, num_vioses_found):
        """
        override the community method to check lua discovery happened
        on minimum number of vios as per the SCG
        """
        scg_map = StorageMapping()
        context = get_admin_context()
        scg = scg_map.get_scg_from_instance_meta(context, self.instance,
                                                 {'messages': []})
        if not scg:
            response = scg_map.determine_scg(context, instance=self.instance,
                                             host=CONF.host, raise_err=True)
            LOG.debug("Determine SCG messages: %s" % response['messages'])
            scg = response['scg']

        if scg:
            min_vios = scg.vios_redundancy
        else:
            scg_info = scg_map.get_connectivity_info(context,
                                                     self.instance['uuid'],
                                                     self.instance['host'],
                                                     self.instance,
                                                     live_data=False)
            min_vios = scg_info.get('min_vioses_per_vol', 1)

        if num_vioses_found < min_vios:
            # A valid hdisk was not found so log and exit
            msg = (_("Failure to connect the volume on required"
                     " number of vioses"))
            LOG.error(msg)
            raise exception.IBMPowerVMAllocStgForMinVIOSFailed(
                **dict(instance_name=self.instance.name,
                       min_vios=min_vios,
                       reasons=msg))

    def wwpns(self):
        """Override this function to return wwpns of only
            those vios which are selected in SCG.
        :param adapter: The pypowervm API adapter.
        :param host_uuid: The UUID of the host for the pypowervm adapter.
        :param instance: The nova instance.
        :returns: Dict having keys as vios id and values as wwpns of vios
        """
        LOG.debug("Instance %s: task_state=%s, power_state=%d" %
                  (self.instance.name, self.instance.task_state,
                   self.instance.power_state))
        # TODO : this condition needs to be revisited
        # when novalink supports remote restart
        if self.instance.task_state in [task_states.DELETING]:
            chk_conn_type = None
        else:
            chk_conn_type = 'either'
        scg_map = StorageMapping()
        scg_info = scg_map.get_connectivity_info(get_admin_context(),
                                                 self.instance['uuid'],
                                                 CONF.host,
                                                 self.instance,
                                                 live_data=False,
                                                 chk_conn_type=chk_conn_type)
        vios_wwpn = {}
        for v in scg_info.get('vios_list'):
            for port in v['vscsi:ports']:
                vios_host_name = self.__get_vios_host_name(v['lpar_id'])
                if vios_host_name in vios_wwpn:
                    vios_wwpn[vios_host_name].append(port['wwpn'])
                else:
                    vios_wwpn[vios_host_name] = [port['wwpn']]

        if not chk_conn_type:
            # For the cases where chk_conn_type is set to None, the SCG may
            # not report all the wwpns that were used to initially attach
            # the volume(s). So we go to the feed task to get all active
            # wwpns for VIOSes in the SCG.
            hosts = [h for h in scg_info['scg'].to_dict()['host_list']
                     if h['name'] == CONF.host]
            scg_vios_lst = [v['id'] for v in hosts[0]['vios_list']]
            LOG.debug("Get full active WWPN list from SCG VIOSs: %s. "
                      "SCG-based wwpn list is: %s" % (scg_vios_lst,
                                                      vios_wwpn))
            vios_wwpn = dict()
            # Instead of feed task approach, just try the DB. This took
            # one quarter second as opposed to feed taking a full second
            # and not having the right xargs to get the physical ports.
            db_vioses = [v for v in scg_info['scg'].vioservers
                         if v.id in scg_vios_lst]
            topo = StorageTopologyCommon(CONF.host, get_admin_context())
            fcport_factory = FcPortFactory.get_factory()
            for db_vios in db_vioses:
                vios_host_name = self.__get_vios_host_name(
                    db_vios.lpar_id)
                vios_wwpn[vios_host_name] = []
                v_dict = topo._get_dbvios_dict(
                    db_vios, {}, None, fcport_factory, 'skip', None, None)
                for port in v_dict['fcport_list']:
                    if port['status'] == 'OK':
                        vios_wwpn[vios_host_name].append(port['wwpn'])

        LOG.info(_("wwpns for vios in selected SCG are %s " % vios_wwpn))
        return vios_wwpn

    def __get_vios_host_name(self, vios_lpar_id):
        """
        Returns the cecName_viosLparId as the vios host name
        which will be used to create an entry on the storage
        """
        vios_host = CONF.host + "_" + str(vios_lpar_id)
        return vios_host

    def _get_hdisk_itls(self, vios_w):
        """
        Override novw_powervm implementation to return itl only for
        vios specified in the SCG. This is set in the vios_data dict as
        part of initialize_connection
        """
        i_wwpns = []
        t_wwpns = []
        lun = None
        if 'vios_data' in self.connection_info['data']:
            vios_data = self.connection_info['data']['vios_data']
            vios_lpar_id = vios_w.id
            vios_host_name = self.__get_vios_host_name(vios_lpar_id)
            if vios_host_name in vios_data.keys():
                it_map = vios_data[vios_host_name]['initiator_target_map']
                i_wwpns = it_map.keys()
                t_wwpns = []
                for it_key in i_wwpns:
                    t_wwpns.extend(it_map[it_key])
                lun = vios_data[vios_host_name]['target_lun']
        elif 'vscsi_target_lun' in self.connection_info:
            # in case of onboarded VM vios_target_lun is there
            # get required data from it for the given vios
            vios_pfc_wwpn_list = vios_w.get_active_pfc_wwpns()
            vios_pfc_wwpn_list = [x.upper() for x in vios_pfc_wwpn_list]
            for vios_target_lun in self.connection_info['vscsi_target_lun']:
                match = list(set(vios_pfc_wwpn_list) &
                             set(vios_target_lun['vios_wwpn']))
                if match:
                    t_wwpns = vios_target_lun['target_wwpn']
                    i_wwpns = vios_target_lun['vios_wwpn']
                    lun = vios_target_lun['target_lun']
                    break
        return i_wwpns, t_wwpns, lun

    def _add_remove_hdisk(self, vio_wrap, device_name, stg_ftsk=None):
        """
        Override the community method to check for shared hdisk
        before removing it.
        :param stg_ftsk: The feed task to add to. If None, then self.stg_ftsk
        """
        # TODO push this into the community
        multiattach = self._check_host_mappings(vio_wrap, device_name)
        if multiattach:
            # this is shared volume set the volume_id to multi attach
            # in the instance system metadata
            volume_id = self.connection_info['data']['volume_id']
            vios_host_name = self.__get_vios_host_name(vio_wrap.id)
            self._set_shared_vol_key(volume_id, vios_host_name)
        else:
            s = super(PowerVCVscsiVolumeAdapter, self)
            s._add_remove_hdisk(vio_wrap, device_name, stg_ftsk)

    def _check_host_mappings(self, vios_w, device_name):
        """
        Return true is there are multiple instances using the given hdisk
        """
        vios_scsi_mappings = next(v.scsi_mappings for v in self.stg_ftsk.feed
                                  if v.uuid == vios_w.uuid)
        mappings = tsk_map.find_maps(
            vios_scsi_mappings, None,
            tsk_map.gen_match_func(pvm_stor.PV, names=[device_name]))

        map_dict = {mapping.client_lpar_href: mapping.backing_storage.name
                    for mapping in mappings}
        LOG.info("Mappings returned %s" % map_dict)
        # the mapping is still present as the task feed removes
        # the mapping later
        return len(map_dict) > 1

    def _set_shared_vol_key(self, volume_id, vios_host_name):
        """
        Set the shared volume key for given vios and volume,
        in instance system metadata, so that it can later be
        used before terminate_connection call
        """
        # TODO another way is to query the bdm and put
        # this in connection_info and save bdm. That does
        # require a call to the conductor api. May be we can
        # move to bdm once we have this in community
        meta_key = vios_host_name + volume_id + "_shared"
        self.instance.system_metadata[meta_key] = "True"

    def _get_shared_vol_key(self, volume_id, vios_host_name):
        """
        Get the shared volume key for given volume and
        vios
        """
        meta_key = vios_host_name + volume_id + "_shared"
        if self.instance.system_metadata.get(meta_key) is None:
            LOG.debug("No multiattach key set %(vol)s %(vios)s" %
                      dict(vol=volume_id, vios=vios_host_name))
            self.instance.system_metadata[meta_key] = False

        return self.instance.system_metadata[meta_key] == "True"

    def _get_finish_onboard_status(self, context):
        """
        Return true if finish onboard is pending else
        False to indicate finish onboard is complete
        """
        instfact = compute_dom.ManagedInstanceFactory.get_factory()
        managed_inst = instfact.find_instance_by_uuid(
            context, self.instance.uuid)
        power_specs = managed_inst.power_specs
        return power_specs.get('pending_onboard_actions', False) if \
            power_specs else False
