# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import logging

from powervc_nova import _
from powervc_nova.virt.ibmpowervm.vif import vm_import
from powervc_nova.virt.ibmpowervm.vif.pvm import topo_pvm
from powervc_nova.virt.ibmpowervm.vif.pvm import utils

from pypowervm.wrappers import logical_partition as pvm_lpar
from pypowervm.wrappers import network as pvm_net

from powervc_nova.virt.ibmpowervm.vif.common import ras
from powervc_nova import logcall

from nova_powervm.virt.powervm import vm

from oslo_config import cfg

LOG = logging.getLogger(__name__)
auto_switch = cfg.StrOpt('auto_switch', default='MGMTSWITCH',
                         help='Name of the switch auto-added and '
                         'it will be ignored while onboarding.')

CONF = cfg.CONF
CONF.register_opt(auto_switch, None, False)


class ImportVMNetworkPVM(vm_import.ImportVMNetworkPowerVM):
    """
    PVM Implementation of the VM Import function.
    """

    def __init__(self, host_name, adapter, uuid):
        """
        ImportVMNetworkPVM constructor

        :param adapter: Communication channel to interface with PowerVM host
        :param uuid: uuid of the host
        """
        super(ImportVMNetworkPVM, self).__init__()

        self.hostname = host_name
        self.host_uuid = uuid
        self.adapter = adapter

    @logcall
    def list_ports(self, lpar_entry):
        """
        get lpar for given lpar id
        find all client network adapters for the lpar
        get orphaned pairs
        for every client network adapter :
            get vea.
            from vea, get slot number, mac address, segmentatin id, status,
            physical network, vswitch, tagged, isveth (always true), is_qbg,
            isorphan and add to a list.
        return the list...
        """
        # A return list of dictionary objects.
        ret_adapter_list = []

        cna_resp = self.adapter.read(pvm_lpar.LPAR.schema_type,
                                     root_id=lpar_entry.uuid,
                                     child_type=pvm_net.CNA.schema_type)
        cnas = pvm_net.CNA.wrap(cna_resp)
        orphaned_pairs = self._get_orphan_pairs()
        for cna in cnas:
            adapter_dict = {}
            adapter_dict['slot_num'] = cna.slot
            adapter_dict['mac_address'] = cna.mac
            adapter_dict['provider:segmentation_id'] = cna.pvid
            adapter_dict['status'] = 'Available'
            adapter_dict['physical_network'] = 'default'
            vswitchid = cna.vswitch_id
            switchname = utils.\
                _get_vswitch_name_for_id(self.adapter,
                                         self.host_uuid,
                                         vswitchid)
            if switchname == CONF.auto_switch:
                continue
            adapter_dict['vswitch'] = switchname

            # Elements for internal disection
            if (cna.tagged_vlans):
                adapter_dict['tagged'] = 'true'
            else:
                adapter_dict['tagged'] = 'false'

            adapter_dict['is_veth'] = True

            qbg_type = cna.vsi_type_id
            if qbg_type is not None and len(qbg_type) > 0:
                adapter_dict['is_qbg'] = True
            else:
                adapter_dict['is_qbg'] = False

            # Determine whether the adapter is orphaned
            orphan = False
            vlanid = int(adapter_dict['provider:segmentation_id'])
            pair = '%(vs)s:%(vlan)d' % {'vs': adapter_dict['vswitch'],
                                        'vlan': vlanid}
            if pair in orphaned_pairs:
                orphan = True
            adapter_dict['is_orphan'] = orphan

            ret_adapter_list.append(adapter_dict)

        return ret_adapter_list

    @logcall
    def interrogate_vm(self, lpar_entry):
        """
        Will perform validation of the LPAR to determine whether or not it
        can be imported.

        :param lpar_instance: The LPAR Instance
        :returns: 1.An empty list to adhere to global interface
                  2.A list of reasons why the LPAR may not be imported.
        """
        # Store for all the warnings
        warnings_list = []

        # List all the adapters
        adapters = self.list_ports(lpar_entry)

        # For each adapter, do interrogation.
        for adapter in adapters:
            # Validate it IS a Virtual Ethernet Adapter.
            if not self._is_veth_adapter(adapter):
                msg = ras.msg('info', 'IMPORT_NON_VIRTUAL_PORT')
                warnings_list.append(msg)
                continue

            # Validate that it is not a Qbg adapter
            if self._is_qbg_veth(adapter):
                msg = ras.msg('info', 'IMPORT_QBG_PORT')
                warnings_list.append(msg)
                continue

            # Validate that it has only a single VLAN (no addl_vlans)
            if not self._has_only_pvid(adapter):
                msg = ras.msg('info', 'IMPORT_ADDL_VLANS')
                warnings_list.append(msg)
                continue

            # Validate the adapter is not on an orphaned VLAN
            if self._is_orphan_vlan(adapter):
                msg = ras.msg('info', 'IMPORT_ORPHAN_VLANS')
                warnings_list.append(msg)
                continue

        return [], warnings_list, adapters

    @logcall
    def _is_veth_adapter(self, adapter):
        """
        Determines whether or not the adapter is a Virtual Ethernet Adapter.

        :param adapter: The adapter object.  Will vary by platform.
        :returns: True if it is a Virtual Ethernet Adapter.
        """

        # Set via _list_net_adapters method
        return adapter['is_veth']

    @logcall
    def _is_qbg_veth(self, adapter):
        """
        Determines whether or not the adapter is a Qbg Virtual Ethernet
        Adapter.

        :param adapter: The adapter.
        :returns: True if it is a Qbg enabled adapter.
        """
        return adapter['is_qbg']

    @logcall
    def _has_only_pvid(self, adapter):
        """
        Determines whether or not the adapter only has a PVID (no additional
        VLANs).

        :param adapter: The adapter.
        :returns: True if it has only a PVID.  False if it has 'additional'
                  VLANs
        """
        return adapter['tagged'] == 'false'

    @logcall
    def _is_orphan_vlan(self, adapter):
        """
        Determines whether or not the adapter has a PVID that does not have an
        associated SEA, and thus is considered an "orphan".

        :param adapter: The adapter.
        :returns: True if the adapters PVID is an orphan, otherwise False.
        """
        return adapter['is_orphan']
