#
#
# Copyright 2014, 2015 IBM Corp.
# All Rights Reserved.
#
# Copyright 2014 IBM Corp.
# Copyright (c) 2012-2013 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from nova.virt.libvirt import config as libvirt_conf

LOG = logging.getLogger(__name__)

"""
Extends nova.virt.libvirt.config to add additional guest
   attributes
"""


class ConfigGuest(libvirt_conf.LibvirtConfigGuest):
    """Extends the libvirt.config.LibvirtConfigGuest class"""

    def __init__(self, **kwargs):
        super(ConfigGuest, self).__init__(**kwargs)
        self.os = None
        self.interfaces = []

    def parse_dom(self, xmldoc):
        super(ConfigGuest, self).parse_dom(xmldoc)
        # Clear out devices, and populate using PowerVC methods
        self.devices = []
        for c in xmldoc.getchildren():
            if c.tag == 'os':
                obj = ConfigGuestOS()
                obj.parse_dom(c)
                self.os = obj
            if c.tag == 'devices':
                for d in c.getchildren():
                    if d.tag == 'disk':
                        obj = ConfigGuestDisk()
                        obj.parse_dom(d)
                        self.devices.append(obj)
                    elif d.tag == 'hostdev' and d.get('type') == 'pci':
                        obj = libvirt_conf.LibvirtConfigGuestHostdevPCI()
                        obj.parse_dom(d)
                        self.devices.append(obj)
                    elif d.tag == 'interface':
                        obj = ConfigGuestInterface()
                        obj.parse_dom(d)
                        self.add_interface(obj)
                    elif d.tag == 'graphics' and d.get('type') == 'vnc':
                        obj = ConfigGuestVNC()
                        obj.parse_dom(d)
                        self.devices.append(obj)

    def add_interface(self, interface):
        self.interfaces.append(interface)


class ConfigGuestOS(libvirt_conf.LibvirtConfigObject):
    """New class for getting the 'os' element from the guest."""

    def __init__(self, **kwargs):
        super(ConfigGuestOS, self).__init__(root_name='os',
                                            **kwargs)
        self.os_arch = None

    def parse_dom(self, xmldoc):
        super(ConfigGuestOS, self).parse_dom(xmldoc)

        for c in xmldoc.getchildren():
            if c.tag == "type":
                self.os_arch = c.get("arch")


class ConfigGuestInterface(libvirt_conf.LibvirtConfigGuestInterface):
    """Extend LibvirtConfigGuestInterface class to add a
       parse_dom method."""

    def parse_dom(self, xmldoc):
        super(ConfigGuestInterface, self).parse_dom(xmldoc)

        self.net_type = xmldoc.get('type')

        for c in xmldoc.getchildren():
            if c.tag == 'source':
                if self.net_type == 'bridge':
                    self.source_dev = c.get('bridge')
                elif self.net_type == 'network':
                    self.source_dev = c.get('network')
            elif c.tag == 'target':
                self.target_dev = c.get('dev')
            elif c.tag == 'mac':
                self.mac_address = c.get('address')
            elif c.tag == 'model':
                self.model = c.get('type')
            elif c.tag == 'virtualport':
                self.vporttype = c.get('type')


class ConfigGuestDisk(libvirt_conf.LibvirtConfigGuestDisk):
    """Extend LibvirtConfigGuestDisk class to extend the
       parse_dom method."""

    def parse_dom(self, xmldoc):
        super(ConfigGuestDisk, self).parse_dom(xmldoc)
        self.source_device = xmldoc.get('device')


class ConfigGuestVNC(libvirt_conf.LibvirtConfigGuestDevice):
    """Extend LibvirtConfigGuestDevice class to extend the format_dom and
       parse_dom method."""

    def __init__(self, **kwargs):
        super(ConfigGuestVNC, self).__init__(root_name="graphics", **kwargs)
        self.type = 'vnc'
        self.port = None
        self.autoport = True
        self.keymap = None
        self.listen = None
        self.listen_type = None
        # Verified binding to all interfaces issue
        # found by the Bandit Scans
        self.listen_address = '0.0.0.0'  # nosec

    def format_dom(self):
        dev = super(ConfigGuestVNC, self).format_dom()

        dev.set("type", self.type)
        if self.autoport:
            dev.set("autoport", "yes")
        else:
            dev.set("autoport", "no")
        if self.port:
            dev.set("port", self.port)
        if self.keymap:
            dev.set("keymap", self.keymap)
        if self.listen:
            dev.set("listen", self.listen)

        return dev

    def parse_dom(self, xmldoc):
        super(ConfigGuestVNC, self).parse_dom(xmldoc)

        self.type = xmldoc.get('type')
        self.port = xmldoc.get('port')
        self.autoport = xmldoc.get('autoport')
        self.keymap = xmldoc.get('keymap')
        self.listen = xmldoc.get('listen')

        for c in xmldoc.getchildren():
            if c.tag == 'listen':
                self.listen_type = c.get('type')
                if self.listen_type == 'address':
                    self.listen_address = c.get('address')
                if self.listen_type == 'network':
                    self.listen_address = c.get('network')
