# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Processor related utilities"""

import math
import six

from nova import utils
from oslo_log import log as logging

from powervc_nova import _
from powervc_nova.virt.powerkvm import exception as pkvmexception

LOG = logging.getLogger(__name__)

# Constant for requesting no SMT (i.e., 1 hardware thread only)
IBM_POWERKVM_SMT_OFF = -1
# Maximum number of threads per sub-core if SMT is enabled
THREADS_PER_CORE = 8

IBM_POWERKVM_VERSION_FILE = '/etc/os-release'


def get_smt_value(flavor_extra_specs, raise_exception=False,
                  instance_name=None):
    """Get the SMT value from the flavor's extra_specs and if it's not present,
       return a default value. The values could be -1 (off), 2, 4, 8, 16, etc.

       :param flavor_extra_specs: the flavor's extra_specs (dict)
       :param raise_exception: raise exception if error. (bool)
       :param instance_name: passed in when raise_exception=True to use in
                             exception message. (string)
       :returns: the SMT value being requested by the flavor, or a default
                 value if the key wasn't defined or improperly defined (int)
    """
    def _raise_smt_invalid(smt_val, instance_name):
        """Raise PowerKVMInvalidSMT exception"""

        # Create a list of all powers of 2 between 2, and
        # THREADS_PER_CORE. This will be all the valid values
        # for enabling SMT.
        valid_values = [2 ** (i + 1) for i in
                        range(int(math.log(THREADS_PER_CORE, 2)))]
        ex_args = {'smt': smt_val,
                   'instance_name': instance_name,
                   'values': str(valid_values)}
        raise pkvmexception.PowerKVMInvalidSMT(**ex_args)

    def _is_power_of_2(val):
        """Checks to see if the value is a power of 2 (larger than 1)."""

        # Check to see if the value only has 1 bit set (i.e., a power of 2)
        return val > 1 and ((val & (val - 1)) == 0)

    smt_val = flavor_extra_specs.get('powerkvm:smt', None)
    if smt_val is None:
        return IBM_POWERKVM_SMT_OFF

    try:
        ret = int(smt_val)
        if not _is_power_of_2(ret) and ret != IBM_POWERKVM_SMT_OFF:
            if raise_exception:
                _raise_smt_invalid(smt_val, instance_name)
            else:
                ret = IBM_POWERKVM_SMT_OFF
                LOG.warn(_("The value of powerkvm:smt must be a power "
                           "of 2 (greater than 1); assuming 'SMT off' mode "
                           "instead."))
    except ValueError:
        if raise_exception:
            _raise_smt_invalid(smt_val, instance_name)
        else:
            ret = IBM_POWERKVM_SMT_OFF
            LOG.warn(_("Could not parse powerkvm:smt value '%(smt)s' as an "
                       "integer; assuming 'SMT off' mode instead.") %
                     {'smt': str(smt_val)})
    return ret


def get_sockets_and_cores_on_host():
    """
    This method is used to send the lscpu command, which will return the number
    of sockets on the host.
    :returns: number of socket on host
    """
    try:
        outputs, stderr = utils.execute('env', 'LC_ALL=C', 'LANG=C',
                                        'lscpu', run_as_root=True,
                                        check_exit_code=False)
        socket = 0
        cores = 0
        if stderr and stderr != "":
            return 1
        else:
            for output in outputs.split('\n'):
                if output.split(':')[0] == 'Socket(s)':
                    socket = int(output.split(':')[1])
                if output.split(':')[0] == 'Core(s) per socket':
                    cores = int(output.split(':')[1])

            return socket, (socket * cores)
    except object as e:
        LOG.error(_("Could not run command 'lscpu'; error: '%s'")
                  % six.text_type(e))
        raise


def skip_smt(flavor_extra_specs, vcpus, instance_name, host_stats):
    hw_cpu_cores = flavor_extra_specs.get('hw:cpu_cores', None)
    hw_cpu_socket = flavor_extra_specs.get('hw:cpu_sockets', None)
    hw_cpu_threads = flavor_extra_specs.get('hw:cpu_threads', None)
    skip_smt = False

    if hw_cpu_cores:
        try:
            hw_cpu_cores = int(hw_cpu_cores)
            validate_cores(hw_cpu_cores, instance_name, host_stats)
            skip_smt = True
        except ValueError:
            ex_args = {'cores': hw_cpu_cores,
                       'instance_name': instance_name}
            raise pkvmexception.PowerKVMInvalidCoreValue(**ex_args)

    if hw_cpu_socket:
        try:
            hw_cpu_socket = int(hw_cpu_socket)
            validate_sockets(hw_cpu_socket, instance_name, host_stats)
            skip_smt = True
        except ValueError:
            ex_args = {'sockets': hw_cpu_socket,
                       'instance_name': instance_name}
            raise pkvmexception.PowerKVMInvalidSocketValue(**ex_args)

    if hw_cpu_threads:
        try:
            hw_cpu_threads = int(hw_cpu_threads)
            validate_threads(hw_cpu_threads, instance_name, host_stats)
            skip_smt = True
        except ValueError:
            ex_args = {'threads': hw_cpu_threads,
                       'instance_name': instance_name}
            raise pkvmexception.PowerKVMInvalidThreadValue(**ex_args)

    if hw_cpu_cores and hw_cpu_socket and hw_cpu_threads:
        total_vcpu = hw_cpu_cores * hw_cpu_socket * hw_cpu_threads

        if vcpus == total_vcpu:
            skip_smt = True
        else:
            ex_args = {'sockets': hw_cpu_socket,
                       'cores': hw_cpu_cores,
                       'threads': hw_cpu_threads,
                       'vcpus': vcpus,
                       'instance_name': instance_name}
            raise pkvmexception.PowerKVMInValidCPUConfiguration(**ex_args)
    return skip_smt


def validate_cores(cores, instance_name, host_stats):
    max_cores = host_stats['stats']['num_cores']
    if cores > 0 and cores <= max_cores:
        return True
    else:
        ex_args = {'cores': cores,
                   'instance_name': instance_name,
                   'max_cores': max_cores}
        raise pkvmexception.PowerKVMCoresGreaterThanMax(**ex_args)


def validate_sockets(sockets, instance_name, host_stats):
    max_sockets = host_stats['stats']['num_sockets']
    if sockets > 0 and sockets <= max_sockets:
        return True
    else:
        ex_args = {'sockets': sockets,
                   'max_sockets': max_sockets,
                   'instance_name': instance_name}
        raise pkvmexception.PowerKVMSocketsGreaterThanMax(**ex_args)


def validate_threads(threads, instance_name, host_stats):
    max_threads = host_stats['stats']['threads_per_core']

    def _is_powerof_2(threads):
        return ((threads & (threads - 1)) == 0)

    if threads > 0 and threads <= max_threads and _is_powerof_2(threads):
        return True
    else:
        ex_args = {'threads': threads,
                   'max_threads': max_threads,
                   'instance_name': instance_name}
        raise pkvmexception.PowerKVMInValidThreadConfiguration(**ex_args)


def validate_smt(smt_val, vcpus, instance_name, host_stats):
    """Validate the smt value passed in with the vcpus in the
       flavor and the max_smt_per_guest host_stat.

       :param smt_val: smt value to validate (int)
       :param vcpus: vcpus in flavor to validate (int)
       :param instance_name: name of instance for exception messages (str)
       :param host_stats: host_stats to validate smt against (dict)
    """

    # vcpus must be a multiple of SMT
    if vcpus % smt_val != 0:
        ex_args = {'smt': smt_val,
                   'vcpus': vcpus,
                   'instance_name': instance_name}
        raise pkvmexception.PowerKVMVCPUSNotValidWithSMT(**ex_args)

    # SMT must be less than the host value for max_smt_per_guest.
    # Scheduler already performed this check, so this situation
    # should never occur, this is just to double check.
    max_smt = host_stats['stats']['max_smt_per_guest']
    if smt_val > max_smt:
        ex_args = {'smt': smt_val,
                   'max_smt': max_smt,
                   'instance_name': instance_name}
        raise pkvmexception.PowerKVMSMTGreaterThanMax(**ex_args)


def run_ppc64_cpu_subcores_command():
    """
    This method is used to send the ppc64_cpu --subcores-per-core
    command, which will return the number of subcores per core.
    :returns: number of subcores per core (int)
    """
    try:
        out, stderr = utils.execute('ppc64_cpu', '--subcores-per-core',
                                    'LC_ALL=C', 'LANG=C', run_as_root=True,
                                    check_exit_code=False)
        if stderr and stderr != "":
            # Split core is not supported on this system; return 1
            # subcore per core.
            return 1
        else:
            # Output is in the format of 'Subcores per core: 1'.
            # Extract the value.
            return int(out.split(':')[1])
    except object as e:
        LOG.error(_("Could not run command 'ppc64_cpu --subcores-per-core'; "
                    "error: '%s'") % six.text_type(e))
        raise


def get_powerkvm_version():
    '''
    Attempts to extract the version of the PowerKVM operating system by looking
    at the /etc/os-release file.  The expected format of this file is:

    # cat /etc/os-release
    NAME="IBM_PowerKVM"
    VERSION="2.1.1"
    ID=ibm_powerkvm
    VERSION_ID="2.1.1"
    BUILD_ID="35.0-service"
    PRETTY_NAME="IBM_PowerKVM 2.1.1"
    ANSI_COLOR="0;34"
    CPE_NAME="cpe:/o:ibm:service:pkvm2_1_1"

    The sed command is used to extract the contents between the quotes
    associated with the VERSION="<contents>".

    If there is any problem parsing this file, an error will be logged and None
    will be returned.
    '''
    version = None
    try:
        out, stderr = utils.execute('/usr/bin/sed', '-n',
                                    's/^VERSION="\(.*\)"/\\1/p',
                                    IBM_POWERKVM_VERSION_FILE)
        if stderr and stderr != "":
            LOG.error(_('Command used to parse the IBM PowerKVM version file '
                        '(%(file)s) returned an error: %(error)s') %
                      {'file': IBM_POWERKVM_VERSION_FILE,
                       'error': stderr})
        else:
            version = out.strip()
            if not version:
                LOG.error(_('IBM PowerKVM version file %s is in an unexpected '
                            'format.') % IBM_POWERKVM_VERSION_FILE)
    except Exception as e:
        LOG.error(_('Error while running command used to parse the IBM '
                    'PowerKVM version file (%(file)s).') %
                  {'file': IBM_POWERKVM_VERSION_FILE})
        LOG.exception(e)

    return version
