# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import copy
import six

from nova import context as nova_context
from nova import utils
from oslo_log import log as logging
from oslo_config import cfg
from powervc_nova.network.powerkvm.agent \
    import ovs_operation_runner as op_runner
from powervc_nova.network.powerkvm.agent.common import exception as exp
from powervc_nova.network.powerkvm.agent.commandlet import CommandExecutor
from powervc_nova.network.powerkvm.agent.commandlet import DOMObjectConverter
from powervc_nova.objects.network import dom_kvm
from powervc_nova import _
from powervc_nova import logcall
from powervc_nova.virt import vm_net_import
from powervc_oslo.network.powerkvm_flow_mapping import \
    VMFlowMapping

LOG = logging.getLogger(__name__)

CONF = cfg.CONF
CONF.import_opt('integration_bridge', 'powervc_nova.network.powerkvm.agent')


class ImportVMNetworkKVM(vm_net_import.ImportVMNetwork):
    """
    Object to handle network configuration required for the import of
    kvm-based VM into PowerVC environment.
    """

    # This variable is used to store tags for vswitch/vlan combinations that
    # are generated by this module because the actual flow for the vswitch/vlan
    # has not yet been created.  If there is no match when checking the flows
    # on the host, this dictionary will be checked to see if we have already
    # handled this vswitch/vlan combination and, if so, that tag value will
    # be returned instead of generating a new tag.  if no match is found,
    # a new tag will be generated and stored with the vswitch and vlan
    # in this dictionary.
    #
    # In the event we do find a match in the flows that exist
    # on the host, we will check this dictionary to see if it contains an
    # entry for that vswitch/vlan combination and will remove it since
    # it is no longer necessary for us to store since the correct tag value
    # can now be retrieved directly from the flows on the host.
    generated_tag_values = {}

    def __init__(self, libvirt_conn):
        dom_converter = DOMObjectConverter()
        self.dom = dom_converter.get_dom(CONF.host)

        # Get the integration bridge info
        self.intbr = CommandExecutor.send_vsctl_command().get(
            CONF.integration_bridge)
        if not self.intbr:
            raise Exception(_("The ovs-vsctl information is missing "
                              "for the integration vswitch. Ensure "
                              "that the integration bridge exists "
                              "and try again."))

        self.libvirt_conn = libvirt_conn

        # Get all the flow information as well
        self.intbr_flows = CommandExecutor.send_ofctl_dump_flows(
            CONF.integration_bridge)
        self.intbr_flow_ports = \
            CommandExecutor.send_ofctl_show_order_by_port_id(
                CONF.integration_bridge)

        LOG.debug('%(intbr)s\nintbr_flows: %(flows)s' %
                  {'intbr': self.intbr,
                   'flows': self.intbr_flows})

    def _get_vmid(self, vm_name):
        """
        Retrieve the id of the domain with the specified name

        :param vm_name: Name of the vm to return the id for
        :returns: id of the specified vm
        """
        # Get it from the map
        vm_id = self.virsh_domain_map.get(vm_name)

        # If it's not in the map, it was probably added after the map was
        # generated.  Rare, but we should handle it.  Regenerate the map.
        if not vm_id:
            self.virsh_domain_map = \
                CommandExecutor.send_virsh_list(self.libvirt_conn)
            vm_id = self.virsh_domain_map.get(vm_name)

        return vm_id

    def _is_connected_to_linuxbridge(self, interface):
        # Find the dom component representing the source bridge.  Note, the
        # method name find_port_or_ovs is misleading because it actually will
        # return things in the unused_components list, which includes linux
        # bridges
        LOG.debug("interface data for linux bridge check: %s"
                  % str(interface))
        return isinstance(self.dom.find_port_or_ovs(interface.get('source')),
                          dom_kvm.LinuxBridge)

    def _is_kimchi_bridge(self, interface):
        """
        Checks to see if the interface is a "bridge" in the kimchi sense.  A
        kimchi bridge can actually look one of two ways.  The first way is when
        a guest has been created but has yet to be started for the first time.
        In this case, the original type is "network".  If it HAS been started,
        then the original type will be "direct".  By virtue of the fact that
        list_ports() didn't blow up, we know a "direct" or "network" original
        type here is valid.
        """
        LOG.debug("checking interface for kimchi bridge: %s" % interface)
        return interface.get('orig_type') in ("direct", "network")

    def _is_directly_connected_to_internal_vswitch(self, interface):
        LOG.debug("interface data for internal bridge check: %s"
                  % str(interface))
        return CONF.integration_bridge == interface.get('source')

    def _is_virtio(self, interface):
        LOG.debug("interface data for virtio check: %s"
                  % str(interface))
        return interface.get('model') == 'virtio'

    @logcall
    def list_ports(self, vm_name):
        '''
        This method will list all the interfaces on the passed in vm and
        return a list of dictionaries that look like the following:

        [ {"interface": "tapxxx", "type": "bridge", "source": "br-int",
           "model": "virtio", "mac": "fa:16:3e:49:80:4a",
           "segmentation_id": "5", "physical_network": "default"},
          {"interface": "tapyyy", "type": "bridge", "source": "br-int",
           "model": "virtio", "mac": "fa:16:3e:49:80:4b",
           "segmentation_id": "-1", "physical_network": "second"},
          ...
        ]

        NOTE: A segmentation_id of -1 indicates a flat network.

        :param vm_id: name of the vm/domain
        :returns: list of interfaces
        '''
        if vm_name:
            try:
                interfaces = \
                    CommandExecutor.send_virsh_domiflist(self.libvirt_conn,
                                                         vm_name)
            except exp.IBMPowerKVMNonBridgeInterfaces:
                raise

            # Now, find the external vid and external vswitch for each
            # interface that has them.
            for interface in interfaces:
                LOG.debug(str(interface))

                # We set the status to active.
                interface['status'] = 'active'

                found_vswitch, found_vlan = self._find_vswitch_vlan(interface)

                # This could be due to the device being a linux bridge.
                if not found_vswitch and not found_vlan:
                    # Should check if it has a known VLAN quick, as that would
                    # be set on the imported data information.
                    if interface.get('known_vlan'):
                        key = 'provider:segmentation_id'
                        interface[key] = interface.get('known_vlan')
                    continue

                # Grab the vlan (use 'segmentation_id' to match vm_importer).
                # If it doesn't exist, it's a flat network, which we'll
                # signify with a -1.
                key = 'provider:segmentation_id'
                interface[key] = found_vlan
                interface['physical_network'] = found_vswitch
                interface['vswitch'] = found_vswitch

            return interfaces
        else:
            LOG.warn(_('No interfaces found for Virtual Machine %s') % vm_name)
            return list()

    def _find_vswitch_vlan(self, interface):
        """
        For the specified interface, compute and return the flow from
        the internal vswitch through to the external vswitch and
        its connection with the outside world

        :param interface: interface object
        :returns: flow as dictionary
        """
        if not interface:
            LOG.error(_('No interface passed in to _find_vswitch_vlan'))
            return None, None

        # See if the interface is on the integration bridge (note, it could
        # be on a different vswitch or on linuxbridge)
        tag = 0
        for port in self.intbr:
            if interface.get('interface') in self.intbr[port]['interface']:
                tag = self.intbr[port].get('tag')
                break

        if tag == 0:
            # Investigate the mapping file to find the ext vswitch and vlanid

            # make sure we are looking at the latest mapping data
            VMFlowMapping.refresh_vm_flow_map()
            tap_device_data = VMFlowMapping.\
                get_tap_device_data(interface.get('interface'))
            if tap_device_data:
                vswitch_name = tap_device_data[VMFlowMapping.
                                               VM_EXTERNAL_VSWITCH_KEY]
                vlan_id = tap_device_data[VMFlowMapping.VM_EXTERNAL_VLAN_KEY]
                vlan_id = '-1' if vlan_id == 'flat' else vlan_id
                return vswitch_name, vlan_id

        if tag == 0:
            return None, None

        # We find the interface's flow by matching the flow's internal vid
        # to the interface's tag
        correct_flow = None
        for flow in self.intbr_flows:
            if flow.get('mod_vlan_vid') == tag:
                correct_flow = flow

        if correct_flow:
            # Grab the vlan (use 'segmentation_id' to match vm_importer).
            # If it doesn't exist, it's a flat network, which we'll
            # signify with a -1.
            vlan = (correct_flow.get('dl_vlan')
                    if correct_flow.get('dl_vlan')
                    else '-1')

            # Now find the external vswitch this interface is potentially
            # on.
            in_port = correct_flow.get('in_port')
            if not in_port:
                return None, None

            host_ovs = self.intbr_flow_ports.get(in_port)
            return host_ovs, vlan

        # Didn't find a matching flow.
        LOG.warn('No flow found for interface %s', interface.get('interface'))
        return None, None

    @logcall
    def interrogate_vm(self, vm_name):
        '''
        This method will loop through the given VM's interfaces and decide
        if the VM is importable or not.  The return is a tuple listing which
        interfaces would need to be moved onto an internal vswitch before the
        VM could be imported and/or the reasons the VM is not importable at
        all.  Note, if both lists are empty, that means the VM is importable
        as-is.  If reasons_not_importable is not empty, then the VM is not
        importable regardless of the state of interfaces_to_move.

        :param vm_name: VM name as shown in 'virsh list'
        :returns: (interfaces_to_move, reasons_not_importable)
        '''
        # Assume the VM is importable until proven it isn't.
        interfaces_to_move = list()
        reasons_not_importable = list()

        try:
            interfaces = self.list_ports(vm_name)
        except exp.IBMPowerKVMNonBridgeInterfaces as nbi_ex:
            reasons_not_importable.append(
                six.text_type(nbi_ex).encode('utf-8'))
            interfaces = list()
        linux_bridges = []

        for interface in interfaces:
            if self._is_connected_to_linuxbridge(interface):
                LOG.debug("Interface %s on linuxbridge needs preparation" %
                          interface.get('interface'))
                interfaces_to_move.append(interface)
                linux_bridge = \
                    self.dom.find_port_or_ovs(interface.get('source'))
                if linux_bridge not in linux_bridges:
                    linux_bridges.append(linux_bridge)
                continue
            if self._is_kimchi_bridge(interface):
                LOG.debug("Interface %s on direct bridge needs preparation" %
                          interface.get('interface'))
                interfaces_to_move.append(interface)
                continue
            if not self._is_virtio(interface):
                msg = _("This virtual machine is not a candidate for "
                        "management because the virtual machine's interface "
                        "model is not set to 'virtio'. PowerVC can only "
                        "manage virtual machines if they have an interface "
                        "model of virtio.")
                LOG.warn("%s - %s" % (vm_name, msg))
                reasons_not_importable.append(msg)
                continue
            if not self._is_directly_connected_to_internal_vswitch(interface):
                msg = _("This virtual machine is not a candidate for "
                        "management because the virtual machine is not "
                        "connected to the integration switch. PowerVC can "
                        "only manage virtual machines if they are connected "
                        "to the integration switch.")
                LOG.warn("%s - %s" % (vm_name, msg))
                reasons_not_importable.append(msg)
                continue
            if not interface.get('physical_network'):
                LOG.debug("Interface for physical network check: %s"
                          % str(interface))
                msg = _("This virtual machine is not a candidate for "
                        "management because the virtual machine's interface "
                        "is not directly connected to an external vswitch. "
                        "PowerVC can only manage virtual machines if they "
                        "are directly connected to an external vswitch.")
                LOG.warn("%s - %s" % (vm_name, msg))
                reasons_not_importable.append(msg)
                continue

        if len(linux_bridges) > 1:
            msg = _("This virtual machine is not a candidate for management "
                    "because the virtual is connected to multiple linux "
                    "bridges. PowerVC can only manage virtual machines if "
                    "they are connected to a single linux bridge.")
            LOG.warn(msg)
            reasons_not_importable.append(msg)

        net_prep_steps = []
        if len(interfaces_to_move) > 0:
            # Build comma separated string
            int_csv = ', '.join([dev.get('source') for dev in
                                 interfaces_to_move if dev.get('source')])
            # If we built something
            if int_csv:
                actions = [_('Connect the virtual machine to the Open vSwitch '
                             'and optionally move interface%(plural)s %(int)s '
                             'to the Open vSwitch.') % {'plural': 's' if ',' in
                                                        int_csv else '',
                                                        'int': int_csv}]
            else:
                actions = [_('Connect the virtual machine to the Open '
                             'vSwitch.')]
            net_prep_steps = [{'name': 'linux_bridge_fix',
                               'actions': actions}]
        return (net_prep_steps, reasons_not_importable, interfaces)

    @logcall
    def prepare_vm_network_for_onboard(self, interfaces, preparation, vm_name):
        '''
        Execute methods required to do the following:
            1. Update VM configuration XML (via libvirt python api) to move
               interface connected to linuxbridge over to integration bridge.
               This update is done so the device UUID does not change.
            2. If external_vswitch_name does not exist, create it.  If
               move_linux_bridge_adapters is True, specify the physical
               adapters on the create so they get moved to the new vswitch.
               If the flag is True but external_vswitch_name already exists,
               just update external_vswitch_name to move the physical adapters
               to it.  Note, if external_vswitch_name already contains
               adapter(s), those will be bonded with the adapter(s) being
               moved over.

            NOTE: Compute should create the neutron ports (via
            allocate_for_instance) and this, in turn, should create the needed
            flows between the integration bridge and external vswitch for the
            given network.  It should also update the flow mapper file.

        :param interfaces: A list of LibvirtConfigGuestInterface objects
        :param preparation: Dictionary describing how to prepare the VM.
            Should look like this:
                {
                    "move_network_components": true,
                    "network_vswitch": "default",
                    "servers": [
                        "d61c1e9f-4c42-40f0-8a83-59e7d34d4ca5"
                    ]
                }
        :returns: An updated version of the list of interface objects passed
            in as interfaces that have been updated to configure the VM
            interfaces to the internal integration bridge.
        '''
        # If we didn't get input indicating something needs to be done,
        # just return what we were given.
        if not interfaces or not preparation:
            return interfaces

        # Create a DOM object representing the new state we want.
        desired_dom = dom_kvm.HostOVSNetworkConfig(self.dom.host_name)

        # We need to keep track of the linux bridges we're moving off of so
        # we can pass them on the create/update of the OVS to move the
        # physical adapters.
        orig_bridges = list()

        interface_vlan_dict = {}

        # Find interfaces attached to linux bridges.
        for interface in interfaces:
            # already connected to integration bridge, don't need to do
            # anything
            if interface.source_dev == CONF.integration_bridge:
                continue

            # Find the bridge based on network type
            bridge = None

            vlan = -1

            # don't run this if the source device is already the integration
            # bridge
            if interface.net_type == 'bridge':
                # Find the bridge in the existing dom
                bridge = self.dom.find_port_or_ovs(interface.source_dev)

                vlan = CommandExecutor._find_vlan(interface.source_dev)
            elif interface.net_type == 'network':
                # Grab the appropriate bridge for type network
                vlan, s = \
                    CommandExecutor.find_network_info(self.libvirt_conn,
                                                      interface.source_dev)

                # Now find it in the dom.  It's probably an unused_component
                bridge = self.dom.find_port_or_ovs(s)

            # If we found a bridge, add it to the list
            if bridge and bridge not in orig_bridges:
                orig_bridges.append(bridge)
                # Update source dev for returned list
                interface.source_dev = CONF.integration_bridge

                # add interface/vlan combination to dictionary
                if vlan == -1:
                    vlan = VMFlowMapping.FLAT_NETWORK_VALUE
                interface_vlan_dict[interface.mac_address] = \
                    {VMFlowMapping.VM_EXTERNAL_VLAN_KEY: vlan,
                     VMFlowMapping.VM_EXTERNAL_VSWITCH_KEY: ""}

        # Now ensure the OVS they want exists.
        desired_ovs = self.dom.find_ovs_obj(preparation.get('network_vswitch',
                                                            'default'))
        if not desired_ovs:
            # Doesn't exist yet, create it
            desired_ovs = dom_kvm.OpenVSwitch(
                preparation.get('network_vswitch', 'default'))
        else:
            # It did exist.  Make a copy of it so we don't update the original
            desired_ovs = copy.deepcopy(desired_ovs)

        # Put it in the desired dom
        desired_dom.ovs_list.append(desired_ovs)

        # Do they want to move adapters over?
        if preparation.get('move_network_components', False):
            # Find the ovs port we'll add the LB's to
            if desired_ovs.ovs_port_list:
                desired_port = desired_ovs.ovs_port_list[0]
            else:
                # None yet.  Create one and add to ovs.
                desired_port = dom_kvm.OVSPort('%s-port' % desired_ovs.name)
                desired_ovs.ovs_port_list.append(desired_port)

            for bridge in orig_bridges:
                # Add this linux bridge as a port on the ovs.  This will cause
                # the micro op code to move all physical adapters on the LB
                # to the ovs.
                if bridge not in desired_port.port_list:
                    desired_port.port_list.append(bridge)

        # We are now ready to call micro ops to create/update the ovs.
        ops_runner = op_runner.OVSOperationRunner(self.dom, True, False)
        if not self.dom.find_ovs_obj(desired_ovs.name):
            # OVS doesn't exist, create it.
            ops_runner.create_host_ovs(nova_context.get_admin_context(),
                                       desired_dom)
        elif preparation.get('move_network_components', False):
            # OVS exists, and they want to move ports, so update it.
            ops_runner.update_host_ovs(nova_context.get_admin_context(),
                                       desired_dom)

        # create flow mapping file for VM
        for interface in interface_vlan_dict.iterkeys():
            interface_vlan_dict[interface][VMFlowMapping.
                                           VM_EXTERNAL_VSWITCH_KEY] = \
                desired_ovs.name

        # create file to save data
        filename = "%s/%s%s" % (VMFlowMapping.VM_FLOW_MAP_DIR, vm_name,
                                VMFlowMapping.IMPORT_VM_FLOW_MAP_FILE_SUFFIX)

        try:
            VMFlowMapping.save_flow_mapping_for_imported_vm(
                vm_name,
                **interface_vlan_dict)

            # change owner of file to neutron
            filename = "%s/%s%s" % \
                (VMFlowMapping.IMPORT_VM_FLOW_MAP_DIR,
                 vm_name,
                 VMFlowMapping.IMPORT_VM_FLOW_MAP_FILE_SUFFIX)

            utils.execute("chown", "neutron", filename,
                          run_as_root=True)
        except Exception as e:
            LOG.exception(six.text_type(e))
            LOG.warning(_("Unable to create flow mapping file for %s") %
                        vm_name)

        # Return the updated interfaces
        return interfaces

    @logcall
    def undo_prepare_vm_network_for_onboard(self, orig_interfaces,
                                            preparation):
        """
        Rollback / undo networking changes that were done during a prepare.
        """
        # This was not committed to in the 4Q14 release by the network team,
        # adding this stub as a placeholder for potential future
        # implementation.
        pass
