# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import os
import six
import stat
from lxml import etree

from nova import utils
from oslo_serialization import jsonutils
from oslo_log import log as logging
from oslo_utils import units
from nova.virt import images
from nova.virt.disk import api as disk_api
from nova.virt.libvirt import utils as libvirt_utils
from nova.virt.libvirt.guest import LIBVIRT_POWER_STATE

from oslo_config import cfg

from powervc_nova import _
from powervc_nova.virt.powerkvm.vif import vm_import as network_import
from powervc_nova.virt.powerkvm import config as pkvm_config
from powervc_nova.virt.powerkvm import exception as pkvmexception

CONF = cfg.CONF
LOG = logging.getLogger(__name__)

PERSISTENT = 1
BOOT = 'boot'
CDROM = 'cdrom'

# support status
SUPPORTED = 'supported'
NOT_SUPPORTED = 'not_supported'
PREPARATION = 'preparation'

# preparation actions
MOVE_DISK = _('Move boot disk location')
UPDATE_PERMISSION = _('Update boot disk permissions')
UPDATE_BACKING_PERMISSION = _('Update backing file permissions')
UPDATE_BOOT_DISK_NAME_ISO = _('Update boot disk name to disk.local')
UPDATE_BOOT_DISK_NAME_QCOW2 = _('Update boot disk name to disk')
CONVERT_BOOT_DISK_FORMAT_TO_QCOW2 = _('Convert boot disk format to qcow2')
COPY_CDROM_DEV = _('Copy CDROM device file')
UPDATE_CDROM_PERMISSION = _('Update CDROM device permissions')
UPDATE_CDROM_BACKING_PERMISSION = _('Update CDROM backing file permissions')
UPDATE_CDROM_DEV_NAME_CONFIG = _('Update CDROM device name to disk.config')
UPDATE_CDROM_DEV_NAME = _('Update CDROM device name to disk')
MERGE_BACKING_FILE = _('Merge boot disk backing file')
MERGE_CDROM_BACKING_FILE = _('Merge CDROM device backing file')


class ImportVM(object):
    """
    Object to handle import vm functions for a given host
    """
    def __init__(self, conn):
        self._conn = conn

    def list_instances_for_onboard(self, drv_inst_names):
        """
        Generate the data needed by the discovery_driver.discover_instances
        for onboarding

        :param drv_inst_names: List of names of all guests on the host
        :returns: List of dictionaries for each instance specifying
                  their onboard support status and other details.
        """
        disc_insts = []

        pkvm_net = None
        for name in drv_inst_names:
            disc_inst = {}
            # Get the domain object and XML for the guest
            domain = self._conn.lookupByName(name)
            xml = domain.XMLDesc(0)
            xml_doc = etree.fromstring(xml)
            guest_cfg = pkvm_config.ConfigGuest()
            guest_cfg.parse_dom(xml_doc)

            uuid = domain.UUIDString()
            status = SUPPORTED
            steps = []
            reasons = []

            disc_inst['name'] = name
            disc_inst['state'] = LIBVIRT_POWER_STATE[domain.state()[0]]
            disc_inst['uuid'] = uuid
            disc_inst['support'] = {}

            # Check network support of this VM
            if not pkvm_net:
                pkvm_net = network_import.ImportVMNetworkKVM(self._conn)
            (net_prep_steps, network_warnings, adapters) = \
                pkvm_net.interrogate_vm(name)
            disc_inst['adapters'] = adapters
            if net_prep_steps and len(net_prep_steps) > 0:
                status = PREPARATION
                steps.extend(net_prep_steps)
            if network_warnings:
                status = NOT_SUPPORTED
                reasons.extend(network_warnings)

            # Check support of properties for this VM
            vm_props = self._interrogate_vm_properties(domain, guest_cfg)
            if vm_props['status'] == NOT_SUPPORTED:
                status = NOT_SUPPORTED
                reasons.extend(vm_props['reasons'])

            # Check storage support of this VM
            disk_info = self._interrogate_vm_storage(uuid, guest_cfg)
            # Set the import status, but don't override if current
            # status is not supported.
            if status != NOT_SUPPORTED and disk_info['status'] == PREPARATION:
                status = PREPARATION
                steps.extend(disk_info['steps'])
            if disk_info['status'] == NOT_SUPPORTED:
                status = NOT_SUPPORTED
                reasons.extend(disk_info['reasons'])

            disc_inst['support']['status'] = status
            # Add steps and reasons if this instance
            # is not automatically supported
            if status == NOT_SUPPORTED:
                disc_inst['support']['reasons'] = reasons
            if status == PREPARATION:
                disc_inst['support']['steps'] = steps
            disc_insts.append(disc_inst)

        return disc_insts

    def prepare_instance_for_onboard(self, domain, preparation, orig_state):
        """Prepare an instance for onboarding.

        :param domain: The domain object of the VM
        :param preparation: Dictionary describing how to prepare the VM
        :param orig_state: The state the VM was in before the onboarding
                           process powered off the VM.
        """

        # These flags will track progress of preparation so we can rollback
        # all changes in the event if an exception is raised
        network_preped = False
        cdrom_preped = False
        boot_disk_preped = False

        try:
            # Get domain XML
            xml = domain.XMLDesc(0)
            doc = etree.fromstring(xml)
            guest_cfg = pkvm_config.ConfigGuest()
            guest_cfg.parse_dom(doc)

            # Check if there is a boot device and get info for it
            # The boot device should always be the first device
            uuid = domain.UUIDString()
            instance_path = os.path.join(CONF.instances_path, uuid)

            # Get list of non boot disks
            disks, status_info = self._get_disk_devices(guest_cfg.devices)

            # Do any necessary preparation on the cdrom device.
            # Preparation on the cdrom device must be done before boot device,
            # in order to determine disk format. This has to be run in a loop,
            # since some preparation checks in the interrogate method can't be
            # executed until previous steps have been prepared.
            status = PREPARATION
            cdrom_rollback_details = {}
            cdrom_actions_performed = []
            count = 0
            while status == PREPARATION:
                # This if condition should never be true, but this exit is
                # here as a precaution so we never get in an infinite loop.
                if count >= 10:
                    ex_args = {
                        'path': cdrom_rollback_details.get('orig_source_path')}
                    raise pkvmexception.PowerKVMOnboardDiskPrepFailed(
                        **ex_args)
                # Check for any necessary preparation
                cdrom_dev_info = self._interrogate_cdrom_dev(
                    disks, instance_path,
                    rollback_details=cdrom_rollback_details)
                status = cdrom_dev_info['status']
                if status == PREPARATION:
                    cdrom_preped = True
                    # Prepare the CDROM disk
                    disks = self._prepare_cdrom_dev(
                        disks, instance_path, cdrom_dev_info,
                        cdrom_actions_performed,
                        guest_cfg.devices[0].driver_format)
                count += 1

            if status == NOT_SUPPORTED:
                reasons = cdrom_dev_info['reasons']
                ex_args = {'name': domain.name(),
                           'reasons': ' ,'.join(reasons)}
                LOG.error(_("Preparation was attempted on virtual"
                            " machine %(name)s, but it was determined"
                            " that this virtual machine is not a"
                            " candidate for management for the following"
                            " reason(s): %(reasons)s") % ex_args)
                raise pkvmexception.PowerKVMOnboardPrepFailed(**ex_args)

            # Do any necessary preparation on the boot device. This has to be
            # run in a loop, since some preparation checks in the interrogate
            # method can't be executed until previous steps have been prepared.
            status = PREPARATION
            rollback_details = {}
            actions_performed = []
            count = 0
            while status == PREPARATION:
                # This if condition should never be true, but this exit is
                # here as a precaution so we never get in an infinite loop.
                if count >= 10:
                    ex_args = {
                        'path': rollback_details.get('orig_source_path')}
                    raise pkvmexception.PowerKVMOnboardDiskPrepFailed(
                        **ex_args)
                # Check for any necessary preparation
                boot_info = self._interrogate_boot_disk(
                    guest_cfg.devices, instance_path,
                    cdrom_dev_info['disk_format'],
                    rollback_details=rollback_details)
                status = boot_info['status']
                if status == PREPARATION:
                    boot_disk_preped = True
                    # Prepare the boot disk
                    guest_cfg.devices = self._prepare_boot_disk(
                        guest_cfg.devices, instance_path, boot_info,
                        actions_performed)
                count += 1

            if status == NOT_SUPPORTED:
                reasons = boot_info['reasons']
                ex_args = {'name': domain.name(),
                           'reasons': ' ,'.join(reasons)}
                LOG.error(_("Preparation was attempted on virtual"
                            " machine %(name)s, but it was determined"
                            " that this virtual machine is not a"
                            " candidate for management for the following"
                            " reason(s): %(reasons)s") % ex_args)
                raise pkvmexception.PowerKVMOnboardPrepFailed(**ex_args)

            # Prepare network
            network_preped = True
            pkvm_net = network_import.ImportVMNetworkKVM(self._conn)
            guest_cfg.interfaces = pkvm_net.prepare_vm_network_for_onboard(
                guest_cfg.interfaces, preparation, domain.name())

            # Testing path - if the API caller set the undo flag, undo changes
            if preparation.get('rollback', False):
                raise Exception('Initiating caller requested rollback.')

        except Exception as e:
            LOG.exception(_("Preparation of virtual machine %(name)s failed."
                            " Attempting roll back. %(ex)s") %
                          {'name': domain.name(), 'ex': six.text_type(e)})
            if boot_disk_preped:
                self._undo_prepare_boot_disk(
                    guest_cfg.devices[0].source_path,
                    rollback_details, actions_performed,
                    guest_cfg.devices[0].driver_format)
            if cdrom_preped:
                for disk in guest_cfg.devices:
                    if (disk.root_name == 'disk' and
                        disk.source_type == 'file' and
                            disk.source_device == 'cdrom'):
                        cdrom_source_path = disk.source_path
                self._undo_prepare_cdrom_disk(cdrom_source_path,
                                              cdrom_rollback_details,
                                              cdrom_actions_performed)

            if network_preped:
                pkvm_net.undo_prepare_vm_network_for_onboard("",
                                                             preparation)
            # Re-raise the exception so the caller can undo any preparation
            raise

        # Extract info for the instance
        vm_info = self._extract_info_for_onboarding(
            pkvm_net, domain, guest_cfg, orig_state)

        LOG.info(_("Virtual machine %(name)s to be managed has"
                   " the following properties: %(vm_info)s") % {
                 'name': domain.name(), 'vm_info': vm_info})

        return vm_info

    def query_onboarding_instances(self, instances):
        """
        Generate the data needed by the discovery_driver.query_instances
        for onboarding.

        :param instances: A list of instance dictionaries
        :returns: A list of dictionaries for each VM containing all
                  the necessary instance specifications to bring
                  the VM onboard.
        """
        vm_list = []

        pkvm_net = network_import.ImportVMNetworkKVM(self._conn)

        for instance in instances:
            try:
                domain = self._conn.lookupByName(instance['name'])
            except Exception as ex:
                # This shouldn't happen unless the instance was
                # deleted between the list_instances phase and this
                # one. Handle exceptions, so the import operation does not
                # fail if one of the instances cannot be found.
                msg = (_("Error from libvirt while looking up "
                         "%(instance_name)s: %(ex)s")
                       % {'instance_name': instance['name'],
                          'ex': ex})
                LOG.warn(_("Could not import VM %(name)s. %(msg)s.") %
                         {'name': instance['name'], 'msg': msg})
                continue

            # Get guest config object
            xml = domain.XMLDesc(0)
            doc = etree.fromstring(xml)
            guest_cfg = pkvm_config.ConfigGuest()
            guest_cfg.parse_dom(doc)

            # Extract detailed info for each instance
            vm_info = self._extract_info_for_onboarding(
                pkvm_net, domain, guest_cfg)
            vm_list.append(vm_info)

            LOG.info(_("Virtual machine %(name)s to be managed has"
                       " the following properties: %(vm_info)s") % {
                     'name': domain.name(), 'vm_info': vm_info})
        return vm_list

    def _extract_info_for_onboarding(self, pkvm_net, domain, guest_cfg,
                                     orig_state=None):
        """
        Extract information needed for onboarding a VM into PowerVC
        :param pkvm_net: A ImportVMNetworkKVM object for the VM
        :param domain: A libvirt guest domain object for the VM
        :param orig_state: The state the VM was in before the onboarding
                           process powered off the VM.
        :returns: A dictionary of specifications for this VM including
                  system_metadata, flavor, volume, and network info.
        """

        # Set basic info
        uuid = domain.UUIDString()
        name = domain.name()
        state = LIBVIRT_POWER_STATE[domain.state()[0]]
        if orig_state:
            state = orig_state
        power_specs = dict(vm_uuid=uuid, vm_name=name)

        info = domain.info()
        memory_mb = int(info[2] / units.Ki)
        vcpus = info[3]
        arch = None
        smt = None
        disk_size = 0
        disk_format = None
        config_drive = None

        # Get other vm properties
        vm_props = self._interrogate_vm_properties(domain, guest_cfg)
        if vm_props['status'] == SUPPORTED:
            arch = vm_props.get('arch')
            smt = vm_props.get('threads')

        # Get instance storage info
        disk_info = self._interrogate_vm_storage(uuid, guest_cfg)
        if disk_info['status'] == SUPPORTED:
            disk_size = disk_info['virt_disk_size']
            disk_format = disk_info['disk_format']
            config_drive = disk_info['config_drive']

        # Check vnc support of this VM
        vnc_device = self._get_vnc_device(guest_cfg.devices)
        vnc_dict = {}
        if vnc_device:
            vnc_dict['vnc_port'] = vnc_device.port
            vnc_dict['vnc_autoport'] = vnc_device.autoport
            vnc_dict['vnc_keymap'] = vnc_device.keymap
            vnc_dict['vnc_listen'] = vnc_device.listen
            vnc_dict['vnc_listen_type'] = vnc_device.listen_type
            if vnc_device.listen_type == 'address':
                vnc_dict['vnc_listen_address'] = vnc_device.listen_address
            if vnc_device.listen_type == 'network':
                vnc_dict['vnc_listen_network'] = vnc_device.listen_network

        power_specs['vnc_info'] = jsonutils.dumps(vnc_dict)

        # Set system_metadata
        system_metadata = {'vcpus': vcpus,
                           'image_hw_vif_model': 'virtio',
                           'disk_format': disk_format}

        # Set flavor
        flavor = {'vcpus': vcpus,
                  'memory_mb': memory_mb}
        if disk_format == 'iso':
            flavor['ephemeral_gb'] = disk_size
        else:
            flavor['root_gb'] = disk_size
        if smt is not None and smt != 1:
            flavor['powerkvm:smt'] = smt
        # extract cpu info for extra_specs
        if guest_cfg.cpu:
            extra_specs = {}
            extra_specs['hw:cpu_cores'] = guest_cfg.cpu.cores
            extra_specs['hw:cpu_sockets'] = guest_cfg.cpu.sockets
            extra_specs['hw:cpu_threads'] = guest_cfg.cpu.threads
            flavor['extra_specs'] = extra_specs

        # Get the networking information
        port_data = []
        net_ifs = pkvm_net.list_ports(name)
        for net_if in net_ifs:
            # extract the info we needed.  We can safely assume that if
            # we are at this point, there are no bridges.  They were prepared.
            port_data.append({'mac_address': net_if.get('mac_address'),
                              'status': net_if.get('status'),
                              'vswitch': net_if.get('vswitch'),
                              'segmentation_id':
                              net_if.get('provider:segmentation_id', -1),
                              'physical_network':
                              net_if.get('physical_network')})

        # Get attached disk information
        volumes = []
        for disk in guest_cfg.devices:
            if (disk.root_name != 'disk'):
                continue
            # parse to get uid
            # '/dev/disk/by-id/wwn-0x60050760008580058800000000000f30' to
            # '60050760008580058800000000000f30'
            disktype = disk.source_type
            if disktype != 'block':
                continue
            diskname = disk.serial
            diskuid = disk.source_path
            diskuid = diskuid[22:]
            vol = {'name': diskname,
                   'unique_device_id': diskuid}
            # Provide the Target Device from the XML as the mount point
            if disk.target_dev:
                vol['mountpoint'] = '/dev/' + disk.target_dev
            volumes.append(vol)

        # Set vm_info to return
        vm_info = {'name': name,
                   'uuid': uuid,
                   'state': state,
                   'architecture': arch,
                   'memory_mb': memory_mb,
                   'vcpus': vcpus,
                   'ephemeral_gb': 0,
                   'root_gb': 0,
                   'system_metadata': system_metadata,
                   'flavor': flavor,
                   'power_specs': power_specs,
                   'volumes': volumes,
                   'ports': port_data}

        # Only set config_drive if it is True. Libvirt
        # expects it not to be set, if False
        if config_drive:
            vm_info['config_drive'] = config_drive

        if disk_format == 'iso':
            vm_info['ephemeral_gb'] = disk_size
        else:
            vm_info['root_gb'] = disk_size
        # we need to get the mountpoint used for the boot disk
        boot_targetdev = guest_cfg.devices[0].target_dev
        vm_info['root_device_name'] = '/dev/' + boot_targetdev

        return vm_info

    def _interrogate_vm_properties(self, domain, guest_cfg):
        """
        Get information on VM properties needed for onboarding
        a VM into PowerVC.
        :param domain: The libvirt domain object for the VM
        :param guest_cfg: The LibvirtConfigGuest object for the VM
        :returns: A dictionary of info on the VM for this VM, and
                  whether or not it is supported for onboarding.
        """
        reasons = []
        status = SUPPORTED

        # Check if VM is persistent:
        if domain.isPersistent() != PERSISTENT:
            status = NOT_SUPPORTED
            reasons.append(_("This virtual machine is not a candidate"
                             " for management because it is transient. "
                             "PowerVC can only manage virtual machines "
                             "that have a domain that is defined."))

        # Get architecture
        arch = None
        supported_arch = ['ppc64', 'ppc64le']
        if guest_cfg.os is not None:
            arch = guest_cfg.os.os_arch
        if arch is not None and arch not in supported_arch:
            status = NOT_SUPPORTED
            reasons.append(_("This virtual machine is not a candidate"
                             " for management because the operating system"
                             " architecture of '%s' is not supported. PowerVC "
                             "can only manage virtual machines that have a "
                             "'ppc64' or 'ppc64le' architecture.") % arch)

        # Get threads from CPU topology
        threads = None
        supported_threads = [1, 2, 4, 8]
        if (guest_cfg.cpu is not None and
                guest_cfg.cpu.threads is not None):
            threads = int(guest_cfg.cpu.threads)
        # If threads is set and is an unsupported value,
        # return not supported.
        if (threads is not None and
                threads not in supported_threads):
            status = NOT_SUPPORTED
            reasons.append(_("This virtual machine is not a candidate for "
                             "management because the CPU topology of "
                             "%(threads)d is not supported. PowerVC can "
                             "only manage virtual machines that have threads "
                             "set to a power of 2 and cannot be greater "
                             "than %(max)d.") %
                           {'threads': threads, 'max': supported_threads[-1]})

        if status == SUPPORTED:
            vm_info = {'status': status, 'arch': arch, 'threads': threads}
        else:
            vm_info = {'status': status, 'reasons': reasons}
        return vm_info

    def _interrogate_vm_storage(self, uuid, guest_cfg):
        """
        Get information on the boot volume and attached volumes
        needed for onboarding a VM into PowerVC.
        :param uuid: The uuid for the VM
        :param guest_cfg: The LibvirtConfigGuest object for the VM
        :returns: A dictionary of info on the volumes for this VM. If 'status'
                 is SUPPORTED, volume specifications will be returned.
                 If 'status' is PREPARATION, steps will be returned. If
                 'status' is UNSUPPORTED, reasons will be returned.
        """

        status = set()
        reasons = []
        steps = []
        disk_format = None
        config_drive = None

        # Get the storage path to be used for this VM
        # based on the storage path the user specified
        # during host registration and the VM's UUID.
        instance_path = os.path.join(CONF.instances_path, uuid)

        # Get list of non boot disks
        disks, status_info = self._get_disk_devices(guest_cfg.devices)
        status.add(status_info['status'])
        if status_info['status'] == NOT_SUPPORTED:
            reasons.extend(status_info['reasons'])

        # Check if there is a CDROM device, and interrogate it
        cdrom_dev_info = self._interrogate_cdrom_dev(disks, instance_path)
        cdrom_dev_status = cdrom_dev_info['status']
        status.add(cdrom_dev_status)
        if cdrom_dev_status == NOT_SUPPORTED:
            reasons.extend(cdrom_dev_info['reasons'])
        elif cdrom_dev_status == PREPARATION:
            steps.append(cdrom_dev_info['steps'])
        else:
            disk_format = cdrom_dev_info['disk_format']
            config_drive = cdrom_dev_info['config_drive']

        # Check if there is a boot device and get info for it
        # The boot device should always be the first device
        boot_info = self._interrogate_boot_disk(guest_cfg.devices,
                                                instance_path, disk_format)

        boot_disk_status = boot_info['status']
        status.add(boot_disk_status)
        if boot_disk_status == NOT_SUPPORTED:
            reasons.extend(boot_info['reasons'])
        if boot_disk_status == PREPARATION:
            steps.append(boot_info['steps'])

        if NOT_SUPPORTED in status:
            disk_info = {'status': NOT_SUPPORTED, 'reasons': reasons}
        elif PREPARATION in status:
            disk_info = {'status': PREPARATION, 'steps': steps}
        else:
            disk_info = {'status': SUPPORTED,
                         'virt_disk_size': boot_info['virt_disk_size'],
                         'disk_format': disk_format,
                         'config_drive': config_drive}

        return disk_info

    def _get_disk_devices(self, devices):
        """Helper method to extract the non boot disk config objects
           from the device config objects.

        :param devices: list of device config objects
        :returns: (disks, status_info) where disks is the list of
                  disk objects and status_info is whether or not
                  the VM is supported for import.
        """
        disks = []
        status_info = {'status': SUPPORTED,
                       'reasons': []}
        found_boot_disk = False

        for device in devices:
            if (device.root_name != 'disk'):
                continue
            # If this is a boot disk, check if it is the only boot disk
            elif (device.source_type == 'file' and
                  device.source_device == 'disk'):
                if found_boot_disk:
                    status_info['status'] = NOT_SUPPORTED
                    unsupported_msg =\
                        _("This virtual machine is not a candidate for "
                          "management because the virtual machine has "
                          "multiple boot disks. PowerVC can only manage "
                          "virtual machines with single boot disk.")
                    if unsupported_msg not in status_info['reasons']:
                        status_info['reasons'].append(unsupported_msg)
                found_boot_disk = True
            elif (device.source_type == 'volume' and
                  device.source_device == 'disk'):
                status_info['status'] = NOT_SUPPORTED
                status_info['reasons'].append(
                    _("This virtual machine is not a candidate for management "
                      "because the virtual machine has storage pool based "
                      "disks attached. PowerVC can only manage virtual "
                      "machines if the attached data volumes are directly "
                      "through iSCSI."))
            elif (device.source_type == 'volume' and
                  device.source_device == 'lun'):
                status_info['status'] = NOT_SUPPORTED
                status_info['reasons'].append(
                    _("This virtual machine is not a candidate for management "
                      "because the virtual machine has storage pool based "
                      "disks attached. PowerVC can only manage virtual "
                      "machines if the attached data volumes are directly "
                      "through iSCSI."))
            else:
                disks.append(device)

        return (disks, status_info)

    def _get_vnc_device(self, devices):
        """Helper method to extract the vnc graphics config objects
           from the device config objects.

        :param devices: list of device config objects
        :returns: (vnc_devices, status_info) where vnc devices list and
                  status_info is whether or not the VM is supported for import.
        """
        vnc_device = None

        for device in devices:
            if (device.root_name == 'graphics' and device.type == 'vnc'):
                return device

        return vnc_device

    def _interrogate_boot_disk(self, devices, instance_path, disk_format,
                               rollback_details=None):
        """
        Get information on the boot volume needed for
        onboarding a VM into PowerVC.
        :param devices: A list of devices in the LibvirtConfigGuest
                        object for the VM
        :param instance_path: The path where this boot disk should be stored
        :param disk_format: string, 'iso' or 'qcow2'
        :param rollback_details: dictionary of details needed for rollback of a
                                 failed prepare.
        :returns: A dictionary of info on the boot disk for this VM. If
                 'status' is SUPPORTED, volume specifications will be returned.
                 If 'status' is PREPARATION, steps will be returned. If
                 'status' is UNSUPPORTED, reasons will be returned.
        """

        status = set()
        reasons = []
        steps = {'name': 'storage_path_update',
                 'actions': []}
        virt_disk_size = None

        # Check if this device is a boot disk
        if (devices == [] or devices[0].root_name != 'disk'):
            return {'status': NOT_SUPPORTED,
                    'reasons':
                    [_("This virtual machine is not a candidate for "
                       "management because the first device in the guest "
                       "XML is not a boot disk. PowerVC can only manage "
                       "virtual machines if their first device is a boot "
                       "disk.")]}

        disk = devices[0]
        if (disk.target_dev is None):
            return {'status': NOT_SUPPORTED,
                    'reasons': [_("This virtual machine is not a candidate "
                                  "for management because the first disk "
                                  "device in the guest XML does not have a "
                                  "target device. PowerVC can only manage "
                                  "virtual machines if their first disk "
                                  "device has a target device.")]}
        if (disk.target_bus != 'virtio'):
            return {'status': NOT_SUPPORTED,
                    'reasons': [_("This virtual machine is not a candidate "
                                  "for management because the first disk "
                                  "device in the guest XML does not have a "
                                  "target bus of 'virtio'. PowerVC can only "
                                  "manage virtual machines if the target "
                                  "bus is 'virtio'.")]}
        if (disk.source_type != 'file' or disk.source_device != 'disk' or
                disk.driver_name != 'qemu' or (
                    disk.driver_format != 'qcow2' and
                    disk.driver_format != 'raw')):
            return {'status': NOT_SUPPORTED,
                    'reasons': [_("This virtual machine is not a candidate"
                                  " for management because the virtual"
                                  " machine does not have a bootable volume"
                                  " as its first disk device. PowerVC can"
                                  " only manage virtual machines if their"
                                  " first disk device is a qemu qcow2 or qemu"
                                  " raw disk of type 'file'.")]}

        if disk.driver_format == 'raw':
            status.add(PREPARATION)
            steps['actions'].append(CONVERT_BOOT_DISK_FORMAT_TO_QCOW2)

        # This is a boot disk, see if the path is correct
        source_path = disk.source_path
        if rollback_details is not None:
            orig_source_path = rollback_details.get('orig_source_path')
            # Only set original source path if it is not already set.
            if orig_source_path is None:
                rollback_details['orig_source_path'] = source_path
        steps['path'] = source_path
        path_status = self._check_disk_paths(source_path, instance_path, BOOT,
                                             rollback_details)
        if path_status['status'] == NOT_SUPPORTED:
            status.add(NOT_SUPPORTED)
            reasons.extend(path_status['reasons'])
        elif path_status['status'] == PREPARATION:
            status.add(PREPARATION)
            steps['actions'].extend(path_status['actions'])
        else:
            # The path is okay

            # If disk_format is set, check the name of the disk. If it is None,
            # then the VM already has a preparation status. The name will be
            # re-checked during preparation.
            if disk_format is not None:
                action = self._check_file_name(
                    source_path, BOOT, disk_format=disk_format)
                if action:
                    status.add(PREPARATION)
                    steps['actions'].append(action)

            # Get the virtual disk size
            try:
                virt_disk_size = disk_api.get_disk_size(source_path)
            except Exception as e:
                status.add(NOT_SUPPORTED)
                reasons.append(_("Error retrieving size of boot disk: %s")
                               % six.text_type(e))
            # If virtual size could be obtained, convert from bytes to GB
            if status == set():
                if virt_disk_size is None:
                    status.add(NOT_SUPPORTED)
                    reasons.append(_("Error retrieving size of boot"
                                     " disk."))
                else:
                    virt_disk_size = int(virt_disk_size / units.Gi)

        # Return info based on status list. Not supported trumps other
        # status values. Preparation trumps supported.
        if NOT_SUPPORTED in status:
            boot_info = {'status': NOT_SUPPORTED,
                         'reasons': reasons}
        elif PREPARATION in status:
            boot_info = {'status': PREPARATION,
                         'steps': steps}
        else:
            boot_info = {'status': SUPPORTED,
                         'virt_disk_size': virt_disk_size}

        return boot_info

    def _interrogate_cdrom_dev(self, disks, instance_path,
                               rollback_details=None):
        """
        Looks for a cdrom device. If there is one, checks that there is only
        one, and that the source path is in the correct path with valid
        permissions.
        :param disks: A list of disk config objects after the boot disk
        :param instance_path: The path where this boot disk should be stored
        :param rollback_details: dictionary of details needed for rollback of a
                                 failed prepare.
        :returns: A dictionary of info on the boot disk for this VM.
                 If 'status' is UNSUPPORTED, reasons will be returned.
                 If 'status' is PREPARATION, steps will be returned.
                 Otherwise, a 'status' of SUPPORTED will be returned.
        """
        status = set()
        reasons = []
        steps = {'name': 'cdrom_dev_update',
                 'actions': []}
        cdrom_dev = None

        for disk in disks:
            # If there is a cdrom device present, check that there is only one.
            if disk.source_type == 'network' and disk.source_device == 'cdrom':
                return {'status': NOT_SUPPORTED,
                        'reasons': [_("This virtual machine is not a "
                                      "candidate for management because "
                                      "the virtual machine was created "
                                      "from a remote image. PowerVC can "
                                      "only manage virtual machines if "
                                      "were created from local images.")]}
            elif disk.source_type == 'file' and disk.source_device == 'cdrom':
                if disk.target_bus and disk.target_bus != 'scsi':
                    return {'status': NOT_SUPPORTED,
                            'reasons': [_("This virtual machine is not a "
                                          "candidate for management because "
                                          "the virtual machine has a CDROM "
                                          "device with an unsupported target "
                                          "bus type. PowerVC can only manage "
                                          "virtual machines if the CDROM "
                                          "device has a target bus type of "
                                          "'scsi'.")]}
                if cdrom_dev is None:
                    cdrom_dev = disk
                else:
                    return {'status': NOT_SUPPORTED,
                            'reasons': [_("This virtual machine is not a "
                                          "candidate for management because "
                                          "the virtual machine has multiple "
                                          "CDROM devices. PowerVC can only "
                                          "manage virtual machines if they "
                                          "have at most one CDROM device.")]}
        # There is no CDROM device
        if cdrom_dev is None:
            return {'status': SUPPORTED,
                    'disk_format': 'qcow2',
                    'config_drive': False}

        # A cdrom device was found, interrogate it.
        source_path = cdrom_dev.source_path

        # If there is no path, return supported. This virtual machine can
        # still be onboarded, but the CDROM device will be removed from the XML
        if source_path is None:
            return {'status': SUPPORTED,
                    'disk_format': 'qcow2',
                    'config_drive': False}

        if rollback_details is not None:
            orig_source_path = rollback_details.get('orig_source_path')
            # Only set original source path if it is not already set.
            if orig_source_path is None:
                rollback_details['orig_source_path'] = source_path
        steps['path'] = source_path

        # Validate the source path of the CDROM
        path_status = self._check_disk_paths(source_path, instance_path,
                                             CDROM, rollback_details)
        if path_status['status'] == NOT_SUPPORTED:
            status.add(NOT_SUPPORTED)
            reasons.extend(path_status['reasons'])
        elif path_status['status'] == PREPARATION:
            status.add(PREPARATION)
            steps['actions'].extend(path_status['actions'])
        else:
            # Get back disk format and config drive info
            disk_format_info = self._get_disk_format(source_path)
            if disk_format_info['status'] == NOT_SUPPORTED:
                status.add(NOT_SUPPORTED)
                reasons.extend(disk_format_info['reasons'])
            else:
                # Use the disk format info to check if preparation
                # on the name of the cdrom device is neeeded.
                action = self._check_file_name(
                    source_path, CDROM,
                    config_drive=disk_format_info['config_drive'])
                if action:
                    status.add(PREPARATION)
                    steps['actions'].append(action)

        # Return info based on status list. Not supported trumps other
        # status values. Preparation trumps supported.
        if NOT_SUPPORTED in status:
            cdrom_dev_info = {'status': NOT_SUPPORTED,
                              'reasons': reasons}
        elif PREPARATION in status:
            cdrom_dev_info = {'status': PREPARATION,
                              'steps': steps}
        else:
            cdrom_dev_info = {'status': SUPPORTED,
                              'disk_format': disk_format_info['disk_format'],
                              'config_drive': disk_format_info['config_drive']}

        return cdrom_dev_info

    def _check_file_name(self, source_path, disk_type,
                         disk_format=None, config_drive=None):
        action = None
        file_name = os.path.basename(source_path)

        if disk_type == BOOT:
            if disk_format == 'iso' and file_name != 'disk.local':
                action = UPDATE_BOOT_DISK_NAME_ISO
            elif disk_format == 'qcow2' and file_name != 'disk':
                action = UPDATE_BOOT_DISK_NAME_QCOW2

        elif disk_type == CDROM:
            if config_drive is True and file_name != 'disk.config':
                action = UPDATE_CDROM_DEV_NAME_CONFIG
            elif config_drive is False and file_name != 'disk':
                action = UPDATE_CDROM_DEV_NAME

        return action

    def _get_disk_format(self, source_path):
        # Determine if this is a config drive
        out, err = utils.execute('file', source_path)
        if 'config-2' in out:
            if 'ISO' in out:
                config_drive = True
                disk_format = 'qcow2'
            else:
                # If config isn't ISO, ex. VFAT,
                # this is an unsupported config drive
                return {'status': NOT_SUPPORTED,
                        'reasons': [(_("This virtual machine is not a "
                                       "candidate for management because the "
                                       "virtual machine has a configuration  "
                                       "drive format that is not supported. "
                                       "PowerVC can only manage virtual "
                                       "machines if the configuration drive "
                                       "format is ISO 9660."))]}
        else:
            # The CDROM device is not a config-drive, so this is not qcow2.
            config_drive = False
            disk_format = 'iso'

        return {'status': SUPPORTED,
                'disk_format': disk_format,
                'config_drive': config_drive}

    def _check_disk_paths(self, source_path, instance_path, disk_type,
                          rollback_details=None):
        """
        Helper method to check the source file path and backing
        file path, if the source file has a backing file. It will
        check if the file in the path exists, if PowerVC has
        permission to it, and if it is in the correct path.

        :param source_path: disk's source path from domain XML
        :param instance_path: The path where this boot disk should be stored
        :param disk_type: string, either 'boot' or 'cdrom'. Determines if
                          interrogating a boot disk or cdrom device.
        :param rollback_details: dictionary of details that are
                                 needed for rollback of a failed prepare.
        :returns: A status dictionary. If 'status' is PREPARATION, actions
                  will be returned. If 'status' is UNSUPPORTED, reasons
                  will be returned. Otherwise SUPPORTED will be returned.
        """
        status = set()
        reasons = []
        actions = []
        source_path_ok = True
        if rollback_details is None:
            rollback_details = {}

        # If the disk does not exist, or it is in a directory
        # that we do not have permission to, return unsupported
        if not os.path.exists(source_path):
            if disk_type == BOOT:
                reason = (_("This virtual machine is not a candidate for "
                            "management because the virtual machine's boot "
                            "disk %s was not found or PowerVC does not have "
                            "permission to its path. PowerVC can only manage "
                            "virtual machines if they have an accessible "
                            "boot disk.") % source_path)
            elif disk_type == CDROM:
                reason = (_("This virtual machine is not a candidate for "
                            "management because the virtual machine's CDROM "
                            "device %s was not found or PowerVC does not have "
                            "permission to its path. PowerVC can only manage "
                            "virtual machines if they have an accessible "
                            "CDROM device.") % source_path)
            return {'status': NOT_SUPPORTED,
                    'reasons': [reason]}

        # If the disk is not in the correct path, it will need to be moved
        # in the preparation phase of VM import
        if os.path.dirname(source_path) != instance_path:
            status.add(PREPARATION)
            if disk_type == BOOT:
                actions.append(MOVE_DISK)
            elif disk_type == CDROM:
                # The cdrom device should not be moved in case other
                # virtual machines share the same file. A symbolic
                # link should be created for it.
                actions.append(COPY_CDROM_DEV)
        # If no read permission to source path, it will need
        # to be given the correct permissions in the preparation phase
        if not os.access(source_path, os.R_OK):
            status.add(PREPARATION)
            if disk_type == BOOT:
                actions.append(UPDATE_PERMISSION)
            elif disk_type == CDROM:
                actions.append(UPDATE_CDROM_PERMISSION)
            source_path_ok = False

            # Set original owner and permissions info
            # in rollback dictionary if not already set.
            if rollback_details.get('orig_perm') is None:
                rollback_details['orig_perm'] = (
                    oct(os.stat(source_path)[stat.ST_MODE])[-3:])
            if rollback_details.get('orig_owner') is None:
                rollback_details['orig_owner'] = (
                    os.stat(source_path)[stat.ST_UID])
            if rollback_details.get('orig_group') is None:
                rollback_details['orig_group'] = (
                    os.stat(source_path)[stat.ST_GID])

        # If the source path is okay, check if there is a qcow2 backing file.
        if source_path_ok:
            backing_file = libvirt_utils.get_disk_backing_file(
                source_path, basename=False)
            # There may not be a backing file, but if there is, check if
            # it exists still and that PowerVC has access.
            if backing_file:
                if not os.path.exists(backing_file):
                    status.add(NOT_SUPPORTED)
                    if disk_type == BOOT:
                        reasons.append(_("This virtual machine is not a "
                                         "candidate for management because "
                                         "the virtual machine's boot disk's "
                                         "backing file %s was not found or "
                                         "PowerVC does not have permission to "
                                         "its path. PowerVC can only manage "
                                         "virtual machines if they have an "
                                         "accessible boot disk's backing "
                                         "file.") % backing_file)
                    elif disk_type == CDROM:
                        reasons.append(_("This virtual machine is not a cand"
                                         "idate for management because the "
                                         "virtual machine's CDROM device's "
                                         "backing file %s was not found or "
                                         "PowerVC does not have permission to "
                                         "its path. PowerVC can only manage "
                                         "virtual machines if they have an "
                                         "accessible CDROM device's backing "
                                         "file.") % backing_file)
                elif NOT_SUPPORTED not in status:
                    have_access = os.access(backing_file, os.R_OK)
                    common_prefix = os.path.commonprefix(
                        [CONF.instances_path, backing_file])
                    # If no read permission to backing file, it will need to
                    # be given the correct permissions in the preparation phase
                    if not have_access:
                        status.add(PREPARATION)
                        if disk_type == BOOT:
                            actions.append(UPDATE_BACKING_PERMISSION)
                        if disk_type == CDROM:
                            actions.append(UPDATE_CDROM_BACKING_PERMISSION)
                    # If backing file does not exist somewhere in the
                    # instances_path file system, then the backing file needs
                    # to be merged. This is to protect the case of onboarding
                    # a VM into a shared storage env that was originally
                    # deployed in local storage. Otherwise, if the VM is later
                    # migrated, the destination host would not have access to
                    # the backing in the local path on the origin host.
                    if common_prefix != CONF.instances_path:
                        status.add(PREPARATION)
                        if disk_type == BOOT:
                            actions.append(MERGE_BACKING_FILE)
                        if disk_type == CDROM:
                            actions.append(MERGE_CDROM_BACKING_FILE)

                    # Set original owner and permissions info
                    # in rollback dictionary if not already set.
                    if rollback_details.get('orig_backing_perm') is None:
                        rollback_details['orig_backing_perm'] = (
                            oct(os.stat(backing_file)[stat.ST_MODE])[-3:])
                    if rollback_details.get('orig_backing_owner') is None:
                        rollback_details['orig_backing_owner'] = (
                            os.stat(backing_file)[stat.ST_UID])
                    if rollback_details.get('orig_backing_group') is None:
                        rollback_details['orig_backing_group'] = (
                            os.stat(backing_file)[stat.ST_GID])

        if NOT_SUPPORTED in status:
            return {'status': NOT_SUPPORTED,
                    'reasons': reasons}
        elif PREPARATION in status:
            prep_dictionary = {'status': PREPARATION, 'actions': actions}
            return prep_dictionary
        else:
            return {'status': SUPPORTED}

    def _prepare_boot_disk(self, devices, instance_path, boot_info,
                           actions_performed):
        """
        Do all the necessary preparation on the boot disk of the VM.
        Checks for the following preparation actions:
        MOVE_DISK = 'move boot disk'
        UPDATE_PERMISSION = 'update boot disk permissions'
        UPDATE_BACKING_PERMISSION = 'update backing file permissions'
        CONVERT_BOOT_DISK_FORMAT_TO_QCOW2 = 'convert boot disk format to qcow2'
        UPDATE_BOOT_DISK_NAME_ISO = 'update boot disk name to disk.local'
        UPDATE_BOOT_DISK_NAME_QCOW2 = 'update boot disk name to disk'
        MERGE_BACKING_FILE = 'merge boot disk backing file'

        :param devices: A list of devices in the LibvirtConfigGuest object for
                        the VM.
        :param instance_path: The storage path for the instance that the boot
                              disk should be located in.
        :param boot_info: A dictionary with preparation steps required on the
                          boot disk.
        :param actions_performed: A list to keep track of all the preparation
                                  actions that have been performed.

        :returns: A list of devices in the LibvirtConfigGuest object for the VM
                  with updated config properties.
        """
        steps = boot_info['steps']
        if steps:
            actions = steps['actions']
            source_path = steps['path']
            if UPDATE_PERMISSION in actions:
                libvirt_utils.chown(source_path, "qemu")
                utils.execute('chmod', '644', source_path, run_as_root=True)
                actions_performed.append(UPDATE_PERMISSION)
            if UPDATE_BACKING_PERMISSION in actions:
                backing_file = libvirt_utils.get_disk_backing_file(
                    source_path, basename=False)
                libvirt_utils.chown(backing_file, "qemu")
                utils.execute('chmod', '644', backing_file,
                              run_as_root=True)
                actions_performed.append(UPDATE_BACKING_PERMISSION)
            if (CONVERT_BOOT_DISK_FORMAT_TO_QCOW2 in actions or
                    MERGE_BACKING_FILE in actions):
                is_merge = MERGE_BACKING_FILE in actions
                self._convert_disk_format(source_path,
                                          devices[0].driver_format, 'qcow2',
                                          is_merge=is_merge)
                devices[0].driver_format = 'qcow2'
                # If converting from raw to qcow2, this step can be backed out.
                # on prepare failure. If merging the backing file into the
                # qcow2 image, don't back out.
                if not is_merge:
                    actions_performed.append(CONVERT_BOOT_DISK_FORMAT_TO_QCOW2)
            if UPDATE_BOOT_DISK_NAME_ISO in actions:
                new_source_path = instance_path + '/disk.local'
                utils.execute("mv", source_path, new_source_path,
                              run_as_root=True)
                source_path = new_source_path
                actions_performed.append(UPDATE_BOOT_DISK_NAME_ISO)
            elif UPDATE_BOOT_DISK_NAME_QCOW2 in actions:
                new_source_path = instance_path + '/disk'
                utils.execute("mv", source_path, new_source_path,
                              run_as_root=True)
                source_path = new_source_path
                actions_performed.append(UPDATE_BOOT_DISK_NAME_QCOW2)
            elif MOVE_DISK in actions:
                utils.execute("mv", source_path, instance_path,
                              run_as_root=True)
                source_path = os.path.join(instance_path,
                                           os.path.basename(source_path))
                actions_performed.append(MOVE_DISK)
            # Change the boot disk source path to instance path
            devices[0].source_path = source_path

        return devices

    def _undo_prepare_boot_disk(self, new_source_path, rollback_details,
                                actions, input_format):
        """
        Undo boot disk preparation, used when the prepare/onboard flow fails.
        The roll back will be performed in opposite order of prepare, saving
        the file permission changes for last.

        :param new_source_path: The final path to the boot disk after
                                preparation.
        :param rollback_details: A dictionary of boot disk info stored before
                                 the prepare began.
        :param actions: A list of actions that were performed and need to be
                        rolled back.
        :param input_format: The format of the disk.
        """
        # Without details and actions, we can't undo
        if not rollback_details or not actions:
            LOG.warn(_("Could not find data to undo boot disk prepare for"
                       " '%(path)s'.") % {'path': new_source_path})
            return

        # Look up the original path of the boot disk
        source_path = rollback_details['orig_source_path']
        LOG.info(_("Rolling back preparation on boot disk '%s'") % source_path)

        # Return the boot disk to the state it was in before preparation began
        try:
            # Rename the boot disk to the orig name in the orig directory
            if (MOVE_DISK in actions or UPDATE_BOOT_DISK_NAME_ISO in actions or
                    UPDATE_BOOT_DISK_NAME_QCOW2 in actions):
                utils.execute("mv", new_source_path, source_path,
                              run_as_root=True)
            # Convert the boot disk back to raw format
            if CONVERT_BOOT_DISK_FORMAT_TO_QCOW2 in actions:
                self._convert_disk_format(source_path, input_format, 'raw',
                                          is_merge=False)
            # Put back the original owner and permission of the backing file
            if UPDATE_BACKING_PERMISSION in actions:
                backing_file = libvirt_utils.get_disk_backing_file(
                    source_path, basename=False)
                # If backing_file does not exist, then it was merged
                # during preparation
                if backing_file:
                    libvirt_utils.chown(backing_file,
                                        rollback_details['orig_backing_owner'])
                    utils.execute('chmod',
                                  rollback_details['orig_backing_perm'],
                                  backing_file, run_as_root=True)
            # Put back the original owner and permission of the boot disk
            if UPDATE_PERMISSION in actions:
                libvirt_utils.chown(source_path,
                                    rollback_details['orig_owner'])
                utils.execute('chmod', rollback_details['orig_perm'],
                              source_path, run_as_root=True)
        except Exception as e:
            # If there was an error rolling back, log a warning.
            LOG.warn(_("Unable to rollback all of the virtual machine"
                       " management preparation steps done on boot disk"
                       " '%(path)s': %(ex)s") % {'path': source_path,
                                                 'ex': six.text_type(e)})

    def _prepare_cdrom_dev(self, disks, instance_path, cdrom_dev_info,
                           actions_performed, input_format):
        """
        Do all the necassary prepartion on the cdrom device of the VM.
        Checks for the following preparation actions:
        COPY_CDROM_DEV = 'copy cdrom device file'
        UPDATE_CDROM_PERMISSION = 'update cdrom device permissions'
        UPDATE_CDROM_BACKING_PERMISSION =
                'update cdrom backing file permissions'
        UPDATE_CDROM_DEV_NAME_CONFIG =
                'update cdrom device name to disk.config'
        UPDATE_CDROM_DEV_NAME = 'update cdrom device name to disk'
        MERGE_CDROM_BACKING_FILE = 'merge CDROM disk backing file'

        :param instance_path: The storage path for the instance that the boot
                              disk should be located in.
        :param cdrcom_dev_info: A dictionary with preparation steps required
                                on the boot disk.
        :param actions_performed: A list to keep track of all the preparation
                                  actions that have been performed.
        :param input_format: The format of the disk.
        """
        steps = cdrom_dev_info['steps']
        if steps:
            actions = steps['actions']
            source_path = steps['path']
            if UPDATE_CDROM_PERMISSION in actions:
                libvirt_utils.chown(source_path, "qemu")
                utils.execute('chmod', '644', source_path, run_as_root=True)
                actions_performed.append(UPDATE_CDROM_PERMISSION)
            if UPDATE_CDROM_BACKING_PERMISSION in actions:
                backing_file = libvirt_utils.get_disk_backing_file(
                    source_path, basename=False)
                libvirt_utils.chown(backing_file, "qemu")
                utils.execute('chmod', '644', backing_file,
                              run_as_root=True)
                actions_performed.append(UPDATE_CDROM_BACKING_PERMISSION)
            if MERGE_CDROM_BACKING_FILE in actions:
                self._convert_disk_format(source_path, input_format, 'qcow2')
            if UPDATE_CDROM_DEV_NAME_CONFIG in actions:
                new_source_path = instance_path + '/disk.config'
                utils.execute("mv", source_path, new_source_path,
                              run_as_root=True)
                source_path = new_source_path
                actions_performed.append(UPDATE_CDROM_DEV_NAME_CONFIG)
            elif UPDATE_CDROM_DEV_NAME in actions:
                new_source_path = instance_path + '/disk'
                utils.execute("mv", source_path, new_source_path,
                              run_as_root=True)
                source_path = new_source_path
                actions_performed.append(UPDATE_CDROM_DEV_NAME)
            elif COPY_CDROM_DEV in actions:
                utils.execute("cp", source_path, instance_path,
                              run_as_root=True)
                source_path = os.path.join(instance_path,
                                           os.path.basename(source_path))
                actions_performed.append(COPY_CDROM_DEV)
            for disk in disks:
                # Change the cdrom device source path to new path
                if disk.source_type == 'file' and \
                        disk.source_device == 'cdrom':
                    disk.source_path = source_path
        return disks

    def _undo_prepare_cdrom_disk(self, new_source_path, rollback_details,
                                 actions):
        """
        Undo CDROM disk preparation, used when the prepare/onboard flow fails.
        The roll back will be performed in opposite order of prepare, saving
        the file permission changes for last.

        :param new_source_path: The final path to the CDROM disk after
                                preparation.
        :param rollback_details: A dictionary of CDROM disk info before the
                                 prepare began.
        :param actions: A list of actions that were performed and need to be
                        rolled back.
        """
        # Without details and actions, we can't undo
        if not rollback_details or not actions:
            LOG.warn(_("Could not find data to undo CDROM device prepare for"
                       " '%(path)s'.") % {'path': new_source_path})
            return

        # Look up the original path of the boot disk
        source_path = rollback_details['orig_source_path']

        LOG.info(_("Rolling back preparation on cdrom disk '%s'") %
                 source_path)
        # Return the boot disk to the state it was in before preparation began
        try:
            # If the cdrom disk was not copied, but was renamed
            # then it needs to be renamed back to the original name.
            if (COPY_CDROM_DEV not in actions and (
                UPDATE_CDROM_DEV_NAME_CONFIG in actions or
                    UPDATE_CDROM_DEV_NAME in actions)):
                utils.execute('mv', new_source_path, source_path,
                              run_as_root=True)
            # Put back the original owner and permission of the
            # backing file
            if UPDATE_CDROM_BACKING_PERMISSION in actions:
                backing_file = libvirt_utils.get_disk_backing_file(
                    source_path, basename=False)
                # If backing_file does not exist, then it was merged
                # during preparation
                if backing_file:
                    libvirt_utils.chown(backing_file,
                                        rollback_details['orig_backing_owner'])
                    utils.execute('chmod',
                                  rollback_details['orig_backing_perm'],
                                  backing_file, run_as_root=True)
            # Put back the original owner and permission of the cdrom disk
            if UPDATE_CDROM_PERMISSION in actions:
                libvirt_utils.chown(source_path,
                                    rollback_details['orig_owner'])
                utils.execute('chmod', rollback_details['orig_perm'],
                              source_path, run_as_root=True)
        except Exception as e:
            # If there was an error rolling back, log a warning.
            LOG.warn(_("Unable to rollback all of the virtual machine"
                       " management preparation steps done on CDROM device"
                       " '%(path)s': %(ex)s") % {'path': source_path,
                                                 'ex': six.text_type(e)})

    def _convert_disk_format(self, source_path, input_format,
                             output_format, is_merge=True):
        """
        Helper method to convert the disk format to another format. Note
        that this method is very defensive to prevent the possibility of losing
        disk data of existing VMs while attempting to onboard. A temporary
        file at location `staged_path` is used to ensure the integrity of the
        original disk. The errors caught below are not likely to occur.

        :param source_path: disk's source path from domain XML
        :param input_format: convert_image without input format is a security
                             risk.
        :param output_format: string, likely 'qcow2'. Determines the format
                              to be converted to.
        :param is_merge: True if the purpose of this conversion is to merge the
                         backing file. Used for more detailed logging.

        :returns: Nothing. If error during conversion, a warning is generated.
                  An exception should be raised in the future.
        """
        staged_path = '%s.converted' % source_path
        try:
            images.convert_image(source_path, staged_path, input_format,
                                 output_format, run_as_root=True)
        except Exception as e:
            if is_merge:
                LOG.error(_("The merge of the backing file for disk "
                            "'%(path)s' failed: %(err)s") %
                          {'path': source_path, 'err': six.text_type(e)})
                raise pkvmexception.PowerKVMOnboardDiskPrepFailed(
                    path=source_path)
            else:
                LOG.error(_("Conversion of disk '%(path)s' to '%(format)s'"
                            " format failed: %(err)s") %
                          {'path': source_path, 'format': output_format,
                           'err': six.text_type(e)})
                raise pkvmexception.PowerKVMOnboardDiskPrepFailed(
                    path=source_path)
        else:
            # Verify that the conversion was successful
            if self._disk_conversion_successful(staged_path, output_format):
                self._delete_disk_file(source_path)
                self._rename_disk_file(staged_path, source_path)
                libvirt_utils.chown(source_path, 'qemu')
            else:
                self._delete_disk_file(staged_path)
                if is_merge:
                    LOG.error(_("The merge of the backing file for disk "
                                "'%(path)s' failed.") % {'path': source_path})
                    raise pkvmexception.PowerKVMOnboardDiskPrepFailed(
                        path=source_path)
                else:
                    LOG.error(_("Conversion of disk '%(path)s' to '%(format)s'"
                                " format finished but did not produce a disk"
                                " of correct format.") %
                              {'path': source_path, 'format': output_format})
                    raise pkvmexception.PowerKVMOnboardDiskPrepFailed(
                        path=source_path)

    def _disk_conversion_successful(self, path, expected_format):
        """
        Checks if the disk image at location `path` is of the expected format,
        returning the boolean result.
        """
        try:
            img_info = images.qemu_img_info(path, format=expected_format)
        except Exception:
            msg = (_("Failed to get QEMU image info from disk at location "
                     "'%(path)s' after doing conversion.  Assuming that the "
                     "conversion was not successful.")
                   % {'path': path})
            LOG.warn(msg)
            return False

        return img_info.file_format == expected_format

    def _delete_disk_file(self, path):
        try:
            utils.execute('rm', '-f', path, run_as_root=True)
        except Exception as e:
            LOG.error(_("Failed to delete disk file at location '%(path)s': "
                        "%(err)s.") % {'path': path, 'err': six.text_type(e)})
            raise pkvmexception.PowerKVMOnboardDiskPrepFailed(
                path=path)

    def _rename_disk_file(self, old, new):
        try:
            utils.execute('mv', old, new, run_as_root=True)
        except Exception as e:
            LOG.error(_("Failed to rename disk file from '%(old)s' to "
                        "'%(new)s': %(err)s.") % {'old': old, 'new': new,
                                                  'err': six.text_type(e)})
            raise pkvmexception.PowerKVMOnboardDiskPrepFailed(
                path=old)
