#
#
# Copyright 2014, 2015 IBM Corp.
# All Rights Reserved.
#
# Copyright 2011 OpenStack Foundation
# (c) Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from nova import utils, exception
from oslo_log import log as logging
from nova.virt.libvirt.volume import iscsi as volume_iscsi
from nova.i18n import _LW
import os
import time
import six
import glob
from powervc_nova import _


LOG = logging.getLogger(__name__)


class PowerKVMSCSIVolumeDriver(volume_iscsi.LibvirtISCSIVolumeDriver):

    """PowerKVM Implementation of ISCSI Volume Driver."""

    def __init__(self, connection):
        super(PowerKVMSCSIVolumeDriver, self).__init__(connection)

    def _get_host_device(self, iscsi_properties):
        """
        extending _get_host_device for path /dev/disk/by-id/
        """
        if 'target_UID' in iscsi_properties:
            LOG.debug(_('target_UID found in iscsi_properties...'))
            return ("/dev/disk/by-id/wwn-0x%s" %
                    iscsi_properties['target_UID'].lower())
        else:
            LOG.info(_('The target_UID was not specified in iscsi_properties.'
                       ' Virt volume driver will be used instead.'))
            return self.__get_host_device(iscsi_properties)

    def __get_host_device(self, transport_properties):
        """Find device path in devtemfs."""
        device = ("ip-%s-iscsi-%s-lun-%s" %
                  (transport_properties['target_portal'],
                   transport_properties['target_iqn'],
                   transport_properties.get('target_lun', 0)))
        if self._get_transport() == "default":
            return ("/dev/disk/by-path/%s" % device)
        else:
            host_device = None
            look_for_device = glob.glob('/dev/disk/by-path/*%s' % device)
            if look_for_device:
                host_device = look_for_device[0]
            return host_device

    def connect_volume(self, connection_info, disk_info):
        """Override method for powerkvm attaching an iSCSI volume to
           a host. This will contain any extra pre/post operations we
           need to do to handle stale device links."""
        iscsi_props = connection_info['data']
        log_data = {k: v for k, v in iscsi_props.iteritems() if k in
                    ['device_path', 'target_UID', 'target_iqn',
                     'target_portal', 'volume_id', 'target_lun']}
        LOG.debug("PowerKVMSCSIVolumeDriver: connect_volume() - iscsi_"
                  "data=%s, disk_info=%s." % (log_data, disk_info))
        if self.connector.use_multipath or 'target_UID' not in iscsi_props:
            # We don't do anything for this case now.
            # Call the parent connect_volume method.
            return super(PowerKVMSCSIVolumeDriver,
                         self).connect_volume(connection_info, disk_info)

        # Non-multipath - do workaround to handle stale device links.
        host_device = self._get_host_device(iscsi_props)
        by_path_device = self._get_bypath_device(iscsi_props)
        pre = "/dev/disk/by-id"
        LOG.debug("host_device=%s, by_path_device=%s, num_retries=%s" %
                  (host_device, by_path_device,
                   self.connector.device_scan_attempts))
        LOG.info(_("PowerKVMSCSIVolumeDriver: Check for stale device links "
                   "for volume ID %(vol_id)s with backend UID %(uid)s "
                   "attached as LUN %(lun)s.") %
                 dict(vol_id=iscsi_props['volume_id'],
                      uid=iscsi_props['target_UID'],
                      lun=iscsi_props['target_lun']))
        try:
            # 1. Check for expected by-path device, then correlate to by-id
            #    and look for stale links.
            if os.path.exists(by_path_device):
                sdX_path = os.path.realpath(by_path_device)
                by_ids = [f for f in os.listdir(pre) if
                          f.startswith('wwn-') and
                          os.path.realpath(pre + "/" + f) == sdX_path]
                for by_id in by_ids:
                    by_id_path = pre + '/' + by_id
                    # dump debug for this stuff
                    is_readable = self._is_readable(by_id_path)
                    used_disks = self.connection._get_all_block_devices()
                    LOG.debug("used block disks=%s." % used_disks)
                    if by_id_path in used_disks:
                        in_use = True
                    else:
                        in_use = False
                    # Do check for good link or not
                    if by_id_path == host_device:
                        # Already seems to be attached - we can let normal
                        # processing go through and discover.
                        LOG.info(_("The volume device path '%(path)s' already "
                                   "appears attached to this host.") %
                                 dict(path=by_id_path))
                        # use info from debug data above for additional chks?
                    elif not is_readable or not in_use:
                        # stale assumed case.
                        LOG.info(_("The by-path device link '%(by_path)s' "
                                   "corresponds to existing by-id device "
                                   "link '%(by_id)s via SCSI device "
                                   "%(dev)s. Because the UID is not a match "
                                   "for '%(uid)s', the device is considered "
                                   "stale and will be removed before the "
                                   "attach processing for volume ID %(vol)s "
                                   "continues. Readable check=%(readable)s. "
                                   "In-use check=%(in_use)s.") %
                                 dict(by_path=by_path_device, by_id=by_id_path,
                                      dev=sdX_path,
                                      uid=iscsi_props['target_UID'],
                                      vol=iscsi_props['volume_id'],
                                      readable=is_readable, in_use=in_use))
                        self._delete_dev_and_check(by_path_device)
                    else:
                        LOG.warn(_("There is a conflicting iSCSI by-path "
                                   "device link '%(by_path)s' for new "
                                   "attach request for volume UID %(uid)s, "
                                   "But the device cannot be removed because "
                                   "it is currently readable and the device "
                                   "path '%(by_id)s' is in-use by a virtual "
                                   "machine.") %
                                 dict(by_path=by_path_device,
                                      uid=iscsi_props['target_UID'],
                                      by_id=by_id_path))
            # 2. Now check the expected by-id path itself and see if it is
            #    stale.
            if os.path.exists(host_device):
                sdX_path = os.path.realpath(host_device)
                LOG.debug("The expected by-id device path already exists: "
                          "%s. It points at device %s. Doing read check" %
                          (host_device, sdX_path))
                if not self._is_readable(host_device):
                    LOG.info(_("The attach request expects a host by-id "
                               "path of %(path)s, which already exists "
                               "but is not readable. The device %(dev)s will "
                               "be removed and normal discovery attempted "
                               "for the mapped volume.") %
                             dict(path=host_device, dev=sdX_path))
                    self._delete_dev_and_check(host_device)
            # 3. Thirdly, check for by-path device links that match the
            #    iscsi portal prefix and LUN suffix. There may be an existing
            #    link on a different node (canister) that is stale and will
            #    prevent iscsiadm from discovering the LUN on a different
            #    node.
            for link in [f for f in os.listdir('/dev/disk/by-path') if
                         f.endswith("lun-%s" % iscsi_props['target_lun'])]:
                by_path_device = '/dev/disk/by-path/' + link
                sdX_path = os.path.realpath(by_path_device)
                LOG.debug("Checking %s --> %s." % (by_path_device, sdX_path))
                if not sdX_path or (sdX_path == by_path_device):
                    LOG.info(_("Remove conflicting by-path '%s' that does "
                               "not link to any device.") % by_path_device)
                    # Note: unless the offending file or link is owned by
                    #       the nova user, the following will probably fail.
                    #       But this case has not prevented volume discovery
                    #       in my tests. The 'rm' command is currently not
                    #       configured for rootwrap. Adding it would allow
                    #       a utils.execute('rm') call to be made.
                    os.remove(by_path_device)
                elif not self._is_readable(by_path_device) or\
                        not self._is_readable(sdX_path):
                    LOG.info(_("The host device %(device)s may "
                               "conflict with the attach request for %(id)s "
                               "via the by-path link %(link)s. The device "
                               "will be removed.") %
                             dict(device=sdX_path, id=iscsi_props['volume_id'],
                                  link=by_path_device))
                    self._delete_dev_and_check(by_path_device)
                else:
                    LOG.info(_("The host device %(device)s "
                               "with the attach request for %(id)s "
                               "via the by-path link %(link)s."
                               "is already mapped.") %
                             dict(device=sdX_path,
                                  id=iscsi_props['volume_id'],
                                  link=by_path_device))

        except Exception as ex:
            LOG.exception(ex)
            LOG.warn(_("An error occurred while trying to determine or "
                       "cleanup any pre-existing stale device links for "
                       "iSCSI attachment properties %(iscsi)s and host "
                       "device %(device)s. The exception was logged. "
                       "Continuing on to LibvirtISCSIVolumeDriver volume "
                       "connection.") % dict(iscsi=log_data,
                                             device=host_device))

        # Call the parent connect_volume method.
        LOG.info(_("PowerKVMSCSIVolumeDriver: Discovering mapped volume with "
                   "UID %s.") % iscsi_props['target_UID'])
        conf = self.connect_volume_helper(connection_info, disk_info)
        return conf

    @utils.synchronized('connect_volume')
    def connect_volume_helper(self, connection_info, disk_info):
        """Attach the volume to instance_name."""
        """
          PowerVC clone of the libvirt iscsi connect_volume() method.
          We do an extra block of code here in 1.2.3.
          This method is called from our connect_volume() override which
          is doing pre-attach checks and stale disk handling.
          TODO: remove this if extra logic gets pushed to community.
        """
        iscsi_properties = connection_info['data']

        # multipath installed, discovering other targets if available
        # multipath should be configured on the nova-compute node,
        # in order to fit storage vendor
        out = None
        if self.connector.use_multipath:
            out = self.connector._run_iscsiadm_discover(iscsi_properties)

            # There are two types of iSCSI multipath devices.  One which shares
            # the same iqn between multiple portals, and the other which use
            # different iqns on different portals.  Try to identify the type by
            # checking the iscsiadm output if the iqn is used by multiple
            # portals.  If it is, it's the former, so use the supplied iqn.
            # Otherwise, it's the latter, so try the ip,iqn combinations to
            # find the targets which constitutes the multipath device.
            ips_iqns = self.connector.\
                _get_target_portals_from_iscsiadm_output(out)
            same_portal = False
            all_portals = set()
            match_portals = set()
            for ip, iqn in ips_iqns:
                all_portals.add(ip)
                if iqn == iscsi_properties['target_iqn']:
                    match_portals.add(ip)
            if len(all_portals) == len(match_portals):
                same_portal = True

            for ip, iqn in ips_iqns:
                props = iscsi_properties.copy()
                props['target_portal'] = ip.split(",")[0]
                if not same_portal:
                    props['target_iqn'] = iqn
                self.connector._connect_to_iscsi_portal(props)

            self.connector._rescan_iscsi()
        else:
            self.connector._connect_to_iscsi_portal(iscsi_properties)

            # Detect new/resized LUNs for existing sessions
            self.connector._run_iscsiadm(iscsi_properties, ("--rescan",))

        host_device = self._get_host_device(iscsi_properties)

        # The /dev/disk/by-path/... node is not always present immediately
        # TODO(justinsb): This retry-with-delay is a pattern, move to utils?
        tries = 0
        disk_dev = disk_info['dev']

        # Check host_device only when transport is used, since otherwise it is
        # directly derived from properties. Only needed for unit tests
        while ((self._get_transport() != "default" and not host_device) or not
               os.path.exists(host_device)):
            if tries >= self.connector.device_scan_attempts:
                raise exception.NovaException(_("iSCSI device not found at %s")
                                              % (host_device))

            LOG.warn(_LW("ISCSI volume %(disk_dev)s not yet found at host "
                         "device %(host_dev)s. "
                         "Will rescan & retry.  Try number: %(tries)s"),
                     {'disk_dev': disk_dev, 'host_dev': host_device,
                      'tries': tries})
            # **********************************************************
            # BEGIN added block to disconnect from iscsi portal and
            # reconnect before running the --rescan again. There are
            # dual-iscsi target node situations where a newly attached
            # volume is not found until this is done. Do this on the
            # second time through the loop.
            # **********************************************************
            if tries == 1:
                LOG.info(_("Attempt to disconnect and reconnect to the "
                           "portal before continuing with retry loop."))
                self.connector._disconnect_from_iscsi_portal(iscsi_properties)
                LOG.debug("Now re-connect...")
                self.connector._connect_to_iscsi_portal(iscsi_properties)
            # *************************************
            # END added block for PowerVC
            # *************************************
            # The rescan isn't documented as being necessary(?), but it helps
            self.connector._run_iscsiadm(iscsi_properties, ("--rescan",))

            # For offloaded open-iscsi transports, host_device cannot be
            # guessed unlike iscsi_tcp where it can be obtained from
            # properties, so try and get it again.
            if not host_device and self._get_transport() != "default":
                host_device = self._get_host_device(iscsi_properties)

            tries = tries + 1
            if not host_device or not os.path.exists(host_device):
                time.sleep(tries ** 2)

        if tries != 0:
            LOG.debug("Found iSCSI node %(disk_dev)s "
                      "(after %(tries)s rescans)",
                      {'disk_dev': disk_dev,
                       'tries': tries})

        if self.connector.use_multipath:
            # we use the multipath device instead of the single path device
            self.connector._rescan_multipath()

            multipath_device = self.connector.\
                _get_multipath_device_name(host_device)

            if multipath_device is not None:
                host_device = multipath_device
                connection_info['data']['multipath_id'] = \
                    multipath_device.split('/')[-1]

        connection_info['data']['device_path'] = host_device

    def _get_bypath_device(self, iscsi_properties):
        """ private method copy of what the parent class should have for
            getting the host device
        """
        return ("/dev/disk/by-path/ip-%s-iscsi-%s-lun-%s" %
                (iscsi_properties['target_portal'],
                 iscsi_properties['target_iqn'],
                 iscsi_properties.get('target_lun', 0)))

    def _is_readable(self, host_device):
        """ Return whether we consider the block device readable or not"""
        try:
            cmd = ['dd', 'if=' + host_device, 'of=/dev/null', 'bs=512',
                   'count=1', 'iflag=direct']
            (out, err) = utils.execute(*cmd, run_as_root=True)
            if 'error' in out or 'error' in err or '(0 B)' in err:
                LOG.warn(_("Host device %(device)s was found, but is not "
                           "readable. Read output is: stdout=%(out)s, "
                           "stderr=%(err)s") %
                         dict(device=host_device, out=out, err=err))
                return False
            LOG.debug("Reading from device successful: %s" % cmd)
            return True
        except Exception as ex:
            LOG.info(_("Error reading from %(device)s: Error is '%(err)s'") %
                     dict(device=host_device, err=six.text_type(ex)))
            return False

    def _delete_dev_and_check(self, device_path):
        LOG.debug("_delete_dev_and_check() %s " % device_path)
        dev_name = self.connector._linuxscsi.get_name_from_path(device_path)
        self.connector._linuxscsi.remove_scsi_device(dev_name)
        if os.path.exists(device_path):
            LOG.debug("Waiting 3 seconds for removal sync.")
            # wait a couple seconds so give /dev/disk links a
            # chance to sync up with /dev/sdX device removal.
            time.sleep(3)
            if os.path.exists(device_path):
                LOG.warn(_("Device path identified as stale earlier "
                           "still exists: %s") % device_path)

    @utils.synchronized('connect_volume')
    def disconnect_volume(self, connection_info, disk_dev):
        """Detach the volume from instance_name.
           Copy parent class method and change device
           compare by considering by-id path from domain xml and the
           device it links to.
           NOTE: This method completely overrides the one in
                 virt/libvirt/volume.py
        """
        iscsi_properties = connection_info['data']
        by_id_path = self._get_host_device(iscsi_properties)
        multipath_device = None
        if self.connector.use_multipath:
            multipath_device = self.connector.\
                _get_multipath_device_name(by_id_path)
        # Call the parent's parent (LibvirtBaseVolumeDriver) disconnect_volume
        # for housekeeping items.
        super(volume_iscsi.LibvirtISCSIVolumeDriver,
              self).disconnect_volume(connection_info, disk_dev)

        if self.connector.use_multipath and multipath_device:
            return self._disconnect_volume_multipath_iscsi(iscsi_properties,
                                                           multipath_device)
        this_device = os.path.realpath(by_id_path)
        deleted_dev = None
        LOG.info(_("PowerKVMSCSIVolumeDriver disconnect_volume processing "
                   "for %(by_id)s --> %(sdx)s.") % dict(by_id=by_id_path,
                                                        sdx=this_device))
        iscsi_prefix = ("ip-%s-iscsi-%s-lun-" %
                        (iscsi_properties['target_portal'],
                         iscsi_properties['target_iqn']))
        try:
            vm_devices = [os.path.realpath(i) for i in
                          self.connection._get_all_block_devices()]
            # Delete device if LUN is not in use by another instance
            if this_device not in vm_devices:
                # Do the removal even if disconnecting from the iscsi portal
                # later - for consistency.
                LOG.info(_("Remove unused host device %(device)s on LUN "
                           "%(lun)s") %
                         dict(device=this_device,
                              lun=iscsi_properties['target_lun']))
                self._delete_dev_and_check(by_id_path)
                deleted_dev = this_device
            else:
                LOG.warn(_("Could not remove host device %(device)s since the "
                           "by-id link %(by_id)s is in use by a virtual "
                           "machine.") %
                         dict(device=this_device, by_id=by_id_path))
            # If no by-path devices are connected from the same iscsi portal
            # and node, then we can disconnect from the portal. We also check
            # device path readability and whether it is in-use to make the best
            # choice here.
            at_least_one_connected = False
            for by_path in os.listdir("/dev/disk/by-path"):
                sdx_path = os.path.realpath("/dev/disk/by-path/" + by_path)
                if deleted_dev and deleted_dev == sdx_path:
                    LOG.debug("Skip deleted dev %s --> %s" % (by_path,
                                                              sdx_path))
                    continue
                if by_path.startswith(iscsi_prefix):
                    LOG.debug("Found prefix matching by_path link %s to "
                              "device %s." % (by_path, sdx_path))
                    if self._is_readable("/dev/disk/by-path/" + by_path) or\
                            sdx_path in vm_devices:
                        at_least_one_connected = True
                        break
                    else:
                        LOG.debug("Device %s is not readable or not in-use."
                                  % by_path)
            if not at_least_one_connected:
                LOG.info(_(
                    "No remaining connected paths for iscsi portal and "
                    "IQN prefix %(prefix)s. Disconnect from the portal.")
                    % dict(prefix=iscsi_prefix))
                self.connector._disconnect_from_iscsi_portal(iscsi_properties)
        except Exception as ex:
            LOG.exception(ex)
            # Per STG defect 24039, one of the problem iSCSI cleanup
            # scenarios is that the volume detach from the VM was successful.
            # But then the disconnect_volume() from the host fails
            # for some reason. If an exception is raised, the user
            # gets a detach failure, when in fact the VM 'detach' part of the
            # operation happened and is not rolled back, leaving the console
            # VM out of sync with the XML definition. Instead we output a
            # warning message and continue. This also fixes an error path
            # on the attach flow. If the libvirt operation fails on the
            # attach, then a disconnect_volume operation is done to roll
            # back the connection on the host. If that rollback fails, then
            # the user sees a 'detach' failure in the console (when they
            # requested an 'attach'. Loging a warning here avoids that
            # until nova libvirt driver fixes their exception handling.
            data = {k: v for k, v in iscsi_properties.iteritems() if k in
                    ['device_path', 'target_UID', 'target_iqn',
                     'target_portal', 'volume_id', 'target_lun']}
            m = _("There was an error trying to disconnect the "
                  "volume from the host. The volume was detached from the "
                  "virtual machine, so processing will continue. If the "
                  "mapping to the host from the backend storage device is "
                  "cleaned up successfully, no manual cleanup of the stale "
                  "device on the host should be needed. The connection "
                  "information is %(data)s and the disconnect problem was: "
                  "%(err)s")
            LOG.warn(m % dict(data=data, err=ex))
        LOG.debug("PowerKVMSCSIVolumeDriver: Completed actionable items for "
                  "disconnect of block device path %s." % by_id_path)
